/****************************************************************
 *								*
 * Copyright (c) 2017 Fidelity National Information		*
 * Services, Inc. and/or its subsidiaries. All rights reserved.	*
 *								*
 *	This source code contains the intellectual property	*
 *	of its copyright holder(s), and is made available	*
 *	under a license.  If you do not know the terms of	*
 *	the license, please stop and do not read further.	*
 *								*
 ****************************************************************/

/* All additions to the end please */

TAB_JPL_TRC_REC("JnlPoolWrite Sleep      ", jnl_pool_write_sleep)	  /* # of SLEEP calls in "jnl_pool_write" */
TAB_JPL_TRC_REC("Phs2CmtArrayFull Sleep  ", repl_phs2cmt_arrayfull_sleep) /* # of SLEEP calls in "UPDATE_JPL_RSRV_WRITE_ADDR" */
TAB_JPL_TRC_REC("ReplPhs2Clnup IsPrcAlv  ", repl_phase2_cleanup_isprcalv) /* # of "is_proc_alive" calls in "repl_phase2_cleanup" */
