/****************************************************************
 *								*
 *	Copyright 2013, 2014 Fidelity Information Services, Inc	*
 *								*
 *	This source code contains the intellectual property	*
 *	of its copyright holder(s), and is made available	*
 *	under a license.  If you do not know the terms of	*
 *	the license, please stop and do not read further.	*
 *								*
 ****************************************************************/
#ifndef GDS_RUNDOWN_ERR_CLEANUP_INCLUDED
#define GDS_RUNDOWN_ERR_CLEANUP_INCLUDED

void gds_rundown_err_cleanup(boolean_t have_standalone_access);

#endif
