/****************************************************************
 *
 * Copyright (c) 2005-2018 Fidelity National Information	*
 * Services, Inc. and/or its subsidiaries. All rights reserved.	*
 *								*
 *	This source code contains the intellectual property	*
 *	of its copyright holder(s), and is made available	*
 *	under a license.  If you do not know the terms of	*
 *	the license, please stop and do not read further.	*
 *								*
 ****************************************************************/
#ifndef VALIDATE_CHECKSUM_H_INCLUDED
#define VALIDATE_CHECKSUM_H_INCLUDED

boolean_t mur_validate_checksum(jnl_ctl_list *jctl);

#endif
