/****************************************************************
 *								*
 * Copyright (c) 2001-2016 Fidelity National Information	*
 * Services, Inc. and/or its subsidiaries. All rights reserved.	*
 *								*
 *	This source code contains the intellectual property	*
 *	of its copyright holder(s), and is made available	*
 *	under a license.  If you do not know the terms of	*
 *	the license, please stop and do not read further.	*
 *								*
 ****************************************************************/

/* toktyp.h */

#define TK_ERROR 1
#define TK_EOL 2
#define TK_EOR 3
#define TK_SPACE 4
#define TK_ATSIGN 5
#define TK_IDENT 6
#define TK_NUMLIT 7
#define TK_INTLIT 8
#define TK_CIRCUMFLEX 9
#define TK_COMMA 10
#define TK_LPAREN 11
#define TK_RPAREN 12
#define TK_PLUS 13
#define TK_MINUS 14
#define TK_ASTERISK 15
#define TK_SLASH 16
#define TK_BACKSLASH 17
#define TK_UNDERSCORE 18
#define TK_HASH 19
#define TK_APOSTROPHE 20
#define TK_COLON 21
#define TK_QUOTE 22
#define TK_EQUAL 23
#define TK_GREATER 24
#define TK_LESS 25
#define TK_LBRACKET 26
#define TK_RBRACKET 27
#define TK_QUESTION 28
#define TK_AMPERSAND 29
#define TK_EXCLAIMATION 30
#define TK_NEQUAL 31
#define TK_NGREATER 32
#define TK_NLESS 33
#define TK_NLBRACKET 34
#define TK_NRBRACKET 35
#define TK_NQUESTION 36
#define TK_NAMPERSAND 37
#define TK_NEXCLAIMATION 38
#define TK_PERCENT 39
#define TK_UPPER 40
#define TK_LOWER 41
#define TK_PERIOD 42
#define TK_DIGIT 43
#define TK_DOLLAR 44
#define TK_SEMICOLON 45
#define TK_STRLIT 46
#define TK_VBAR 47
#define TK_EXPONENT 48
#define TK_SORTS_AFTER 49
#define TK_NSORTS_AFTER 50
#define TK_ATHASH 51
#define TK_SQLFIELDNAME 52
#define TK_CR 53

#define	NUM_ASCII_CHARS	128
#define	NUM_CHARS	256	/* number of single-byte characters (includes ASCII and non-ASCII) */
