/****************************************************************
 *								*
 *	Copyright 2001, 2009 Fidelity Information Services, Inc	*
 *								*
 *	This source code contains the intellectual property	*
 *	of its copyright holder(s), and is made available	*
 *	under a license.  If you do not know the terms of	*
 *	the license, please stop and do not read further.	*
 *								*
 ****************************************************************/

/* defines product name and version for licensing */

#include "mdef.h"

LITDEF char	cm_prd_name[] = "GT.CM" ;
LITDEF int4	cm_prd_len    = SIZEOF(cm_prd_name) - 1 ;
LITDEF char	cm_ver_name[] = "V210";
LITDEF int4	cm_ver_len    = SIZEOF(cm_ver_name) - 1 ;
