/****************************************************************
 *								*
 *	Copyright 2003 Sanchez Computer Associates, Inc.	*
 *								*
 *	This source code contains the intellectual property	*
 *	of its copyright holder(s), and is made available	*
 *	under a license.  If you do not know the terms of	*
 *	the license, please stop and do not read further.	*
 *								*
 ****************************************************************/

#ifndef REPL_TR_GOOD_H
#define REPL_TR_GOOD_H

boolean_t repl_tr_good(uchar_ptr_t tr, int tr_len, seq_num seqno);

#endif /* REPL_TR_GOOD_H */
