/****************************************************************
 *								*
 *	Copyright 2003, 2005 Fidelity Information Services, Inc	*
 *								*
 *	This source code contains the intellectual property	*
 *	of its copyright holder(s), and is made available	*
 *	under a license.  If you do not know the terms of	*
 *	the license, please stop and do not read further.	*
 *								*
 ****************************************************************/

#ifndef FILE_HEAD_READ_INCLUDED
#define FILE_HEAD_READ_INCLUDED

boolean_t file_head_read(char *, sgmnt_data_ptr_t, int4);

#endif /* FILE_HEAD_READ_INCLUDED */
