/****************************************************************
 *								*
 *	Copyright 2007 Fidelity Information Services, Inc	*
 *								*
 *	This source code contains the intellectual property	*
 *	of its copyright holder(s), and is made available	*
 *	under a license.  If you do not know the terms of	*
 *	the license, please stop and do not read further.	*
 *								*
 ****************************************************************/

/* 8-bit register definitions: */
REGDEF(AL,0),
REGDEF(CL,1),
REGDEF(DL,2),
REGDEF(BL,3),

REGDEF(AH,4),
REGDEF(CH,5),
REGDEF(DH,6),
REGDEF(BH,7),
