/****************************************************************
 *								*
 *	Copyright 2005 Fidelity Information Services, Inc	*
 *								*
 *	This source code contains the intellectual property	*
 *	of its copyright holder(s), and is made available	*
 *	under a license.  If you do not know the terms of	*
 *	the license, please stop and do not read further.	*
 *								*
 ****************************************************************/

#ifndef DESIRED_DB_FORMAT_SET_DEFINED

/* prototypes */

int4	desired_db_format_set(gd_region *reg, enum db_ver new_db_format, char *command_name);

#define DESIRED_DB_FORMAT_SET_DEFINED

#endif
