/****************************************************************
 *								*
 *	Copyright 2009 Fidelity Information Services, Inc	*
 *								*
 *	This source code contains the intellectual property	*
 *	of its copyright holder(s), and is made available	*
 *	under a license.  If you do not know the terms of	*
 *	the license, please stop and do not read further.	*
 *								*
 ****************************************************************/
#ifndef MU_INT_WAIT_RDONLY_H
#define MU_INT_WAIT_RDONLY_H

boolean_t	mu_int_wait_rdonly(sgmnt_addrs *, char	*);

#endif
