/****************************************************************
 *								*
 *	Copyright 2001 Sanchez Computer Associates, Inc.	*
 *								*
 *	This source code contains the intellectual property	*
 *	of its copyright holder(s), and is made available	*
 *	under a license.  If you do not know the terms of	*
 *	the license, please stop and do not read further.	*
 *								*
 ****************************************************************/

#define	TRM_MODIFIERS	0
#define	TRM_EDITMODE	1
#define	TRM_NOTYPEAHD	2
#define	TRM_READSYNC	4
#define	TRM_PROMPT	8
#define	TRM_PASTHRU	16
#define	TRM_ESCTRMOVR 	32
#define	TRM_ESCAPE	64
#define TRM_NOECHO      128
#define TRM_CONVERT	256
