/****************************************************************
 *								*
 *	Copyright 2005, 2007 Fidelity Information Services, Inc	*
 *								*
 *	This source code contains the intellectual property	*
 *	of its copyright holder(s), and is made available	*
 *	under a license.  If you do not know the terms of	*
 *	the license, please stop and do not read further.	*
 *								*
 ****************************************************************/

#ifndef V15_GDSROOT_H
#define V15_GDSROOT_H

typedef uint4	v15_trans_num;
typedef uint4   v15_time_t;

#endif
