/****************************************************************
 *                                                              *
 *      Copyright 2007 Fidelity Information Services, Inc       *
 *                                                              *
 *      This source code contains the intellectual property     *
 *      of its copyright holder(s), and is made available       *
 *      under a license.  If you do not know the terms of       *
 *      the license, please stop and do not read further.       *
 *                                                              *
 ****************************************************************/

#ifndef SOCKINTSTATS_DEFINED
#define SOCKINTSTATS_DEFINED

void sockint_stats(void);

#endif
