#################################################################
#								#
# Copyright (c) 2001-2020 Fidelity National Information		#
# Services, Inc. and/or its subsidiaries. All rights reserved.	#
#								#
#	This source code contains the intellectual property	#
#	of its copyright holder(s), and is made available	#
#	under a license.  If you do not know the terms of	#
#	the license, please stop and do not read further.	#
#								#
#################################################################

#
############################################
#
#
#	versions - set default GT.M versions
#
#
############################################

#	gtm_curpro is the current production version
setenv	gtm_curpro	"V63013"
#	gtm_verno is the current production version
setenv	gtm_verno	$gtm_curpro
