/****************************************************************
 *								*
 *	Copyright 2009 Fidelity Information Services, Inc	*
 *								*
 *	This source code contains the intellectual property	*
 *	of its copyright holder(s), and is made available	*
 *	under a license.  If you do not know the terms of	*
 *	the license, please stop and do not read further.	*
 *								*
 ****************************************************************/

#ifndef   	PUSH_LVVAL_H_
# define   	PUSH_LVVAL_H_

lv_val *push_lvval(mval *arg1);

#endif 	    /* !PUSH_LVVAL_H_ */
