/****************************************************************
 *								*
 *	Copyright 2001, 2007 Fidelity Information Services, Inc	*
 *								*
 *	This source code contains the intellectual property	*
 *	of its copyright holder(s), and is made available	*
 *	under a license.  If you do not know the terms of	*
 *	the license, please stop and do not read further.	*
 *								*
 ****************************************************************/

#ifndef GTCM_JNL_SWITCHED
#define GTCM_JNL_SWITCHED

void gtcm_jnl_switched(gd_region *reg);

#endif /* GTCM_JNL_SWITCHED */
