/****************************************************************
 *								*
 * Copyright (c) 2001-2019 Fidelity National Information	*
 * Services, Inc. and/or its subsidiaries. All rights reserved.	*
 *								*
 *	This source code contains the intellectual property	*
 *	of its copyright holder(s), and is made available	*
 *	under a license.  If you do not know the terms of	*
 *	the license, please stop and do not read further.	*
 *								*
 ****************************************************************/

#ifndef GTM_FFS_INCLUDED
#define GTM_FFS_INCLUDED

#include "gdsroot.h"

block_id gtm_ffs(block_id offset, uchar_ptr_t addr, block_id size);

#endif /* GTM_FFS_INCLUDED */
