/****************************************************************
 *                                                              *
 *      Copyright 2012 Fidelity Information Services, Inc *
 *                                                              *
 *      This source code contains the intellectual property     *
 *      of its copyright holder(s), and is made available       *
 *      under a license.  If you do not know the terms of       *
 *      the license, please stop and do not read further.       *
 *                                                              *
 ****************************************************************/

#ifndef T_RECYCLED2FREE_DEFINED
#define T_RECYCLED2FREE_DEFINED

boolean_t t_recycled2free(srch_blk_status *blkhist);

#endif
