/****************************************************************
 *								*
 *	Copyright 2003 Sanchez Computer Associates, Inc.	*
 *								*
 *	This source code contains the intellectual property	*
 *	of its copyright holder(s), and is made available	*
 *	under a license.  If you do not know the terms of	*
 *	the license, please stop and do not read further.	*
 *								*
 ****************************************************************/

#ifndef REPL_DEBUG
#define REPL_DEBUG_DEFINED
#undef REPL_DBG_H
#undef REPL_DEBUG_ONLY
#undef REPL_DPRINT1
#undef REPL_DPRINT2
#undef REPL_DPRINT3
#undef REPL_DPRINT4
#undef REPL_DPRINT5
#undef REPL_DPRINT6
#undef REPL_EXTRA_DEBUG_ONLY
#undef REPL_EXTRA_DPRINT1
#undef REPL_EXTRA_DPRINT2
#undef REPL_EXTRA_DPRINT3
#undef REPL_EXTRA_DPRINT4
#undef REPL_EXTRA_DPRINT5
#undef REPL_EXTRA_DPRINT6
#define REPL_DEBUG
#include "repl_dbg.h"
#endif

