/****************************************************************
 *								*
 * Copyright (c) 2015-2016 Fidelity National Information	*
 * Services, Inc. and/or its subsidiaries. All rights reserved.	*
 *								*
 *	This source code contains the intellectual property	*
 *	of its copyright holder(s), and is made available	*
 *	under a license.  If you do not know the terms of	*
 *	the license, please stop and do not read further.	*
 *								*
 ****************************************************************/
#ifndef MU_REORG_ENCRYPT_DEFINED
#define MU_REORG_ENCRYPT_DEFINED

void mupip_reorg_encrypt(void);

#define MUPIP_REORG_IN_PROG_FALSE		0
#define MUPIP_REORG_IN_PROG_TRUE		1
#define MUPIP_REORG_IN_PROG_LOCAL_DSK_READ	2

#endif
