/****************************************************************
 *								*
 *	Copyright 2001 Sanchez Computer Associates, Inc.	*
 *								*
 *	This source code contains the intellectual property	*
 *	of its copyright holder(s), and is made available	*
 *	under a license.  If you do not know the terms of	*
 *	the license, please stop and do not read further.	*
 *								*
 ****************************************************************/

#define VDAT_LIT_SIZE (3)
#define VDAT_BDISP_SIZE (2)
#define VDAT_WDISP_SIZE (2)
#define VDAT_LDISP_SIZE (2)
#define VDAT_AP_SIZE (4)
#define VDAT_R9_SIZE (4)
#define VDAT_R8_SIZE (4)
#define VDAT_GR_SIZE (2)
#define VDAT_IMMED_SIZE (3)
#define VDAT_R11_SIZE (5)
#define VDAT_GTMLITERAL_SIZE (11)
#define VDAT_DEF_SIZE (1)
