/****************************************************************
 *								*
 *	Copyright 2007 Fidelity Information Services, Inc	*
 *								*
 *	This source code contains the intellectual property	*
 *	of its copyright holder(s), and is made available	*
 *	under a license.  If you do not know the terms of	*
 *	the license, please stop and do not read further.	*
 *								*
 ****************************************************************/

/* Mod field values for 16-bit addressing forms: */
I386_MOD(BASE,0)
I386_MOD(BASE_DISP_8,1)
I386_MOD(BASE_DISP_16,2)
I386_MOD(REGISTER,3)
