/****************************************************************
 *								*
 *	Copyright 2001 Sanchez Computer Associates, Inc.	*
 *								*
 *	This source code contains the intellectual property	*
 *	of its copyright holder(s), and is made available	*
 *	under a license.  If you do not know the terms of	*
 *	the license, please stop and do not read further.	*
 *								*
 ****************************************************************/

#define OBJ_EMIT_BUF_SIZE	512
#define OBJ_SYM_BUF_SIZE	512

/* PSECT index values used during code generation: */
#define GTM_CODE		0
#define GTM_LITERALS		1
#define GTM_RNAMESAAAAB 	2
#define GTM_LINKAGE		3
#define GTM_LASTPSECT		4
