/****************************************************************
 *								*
 *	Copyright 2001, 2013 Fidelity Information Services, Inc	*
 *								*
 *	This source code contains the intellectual property	*
 *	of its copyright holder(s), and is made available	*
 *	under a license.  If you do not know the terms of	*
 *	the license, please stop and do not read further.	*
 *								*
 ****************************************************************/

#ifndef NAMELOOK_INCLUDED
#define NAMELOOK_INCLUDED
int int_namelook(const uint4 offset_tab[], const nametabent *name_tab, char *str, int len);

#endif /* NAMELOOK_INCLUDED */
