/****************************************************************
 *								*
 *	Copyright 2005 Fidelity Information Services, LLC.	*
 *								*
 *	This source code contains the intellectual property	*
 *	of its copyright holder(s), and is made available	*
 *	under a license.  If you do not know the terms of	*
 *	the license, please stop and do not read further.	*
 *								*
 ****************************************************************/

#define V15_GDS_LABEL 	"GDSDYNUNX02" 	/* GT.M V4 label */
