/****************************************************************
 *                                                              *
 *      Copyright 2007 Fidelity Information Services, Inc       *
 *                                                              *
 *      This source code contains the intellectual property     *
 *      of its copyright holder(s), and is made available       *
 *      under a license.  If you do not know the terms of       *
 *      the license, please stop and do not read further.       *
 *                                                              *
 ****************************************************************/

REGDEF(RAX,0),
REGDEF(RCX,1),
REGDEF(RDX,2),
REGDEF(RBX,3),

REGDEF(RSP,4),
REGDEF(RBP,5),
REGDEF(RSI,6),
REGDEF(RDI,7),

REGDEF(R8,8),
REGDEF(FIRST_64_BIT,8) = I386_REG_R8,
REGDEF(R9,9),
REGDEF(R10,10),
REGDEF(R11,11),
REGDEF(R12,12),
REGDEF(R13,13),
REGDEF(R14,14),
REGDEF(R15,15),
