/****************************************************************
 *								*
 *	Copyright 2001, 2014 Fidelity Information Services, Inc	*
 *								*
 *	This source code contains the intellectual property	*
 *	of its copyright holder(s), and is made available	*
 *	under a license.  If you do not know the terms of	*
 *	the license, please stop and do not read further.	*
 *								*
 ****************************************************************/

/*
 * IMAGE_TABLE_ENTRY(image_type, image-name)
 */
IMAGE_TABLE_ENTRY (INVALID_IMAGE,         "INVALID_IMAGE")
IMAGE_TABLE_ENTRY (GTM_IMAGE,             "GT.M")
IMAGE_TABLE_ENTRY (MUPIP_IMAGE,           "MUPIP")
IMAGE_TABLE_ENTRY (DSE_IMAGE,             "DSE")
IMAGE_TABLE_ENTRY (LKE_IMAGE,             "LKE")
IMAGE_TABLE_ENTRY (GTMSECSHR_IMAGE,       "GTMSECSHR")
IMAGE_TABLE_ENTRY (GTCM_SERVER_IMAGE,     "GTCM_SERVER")
IMAGE_TABLE_ENTRY (GTCM_GNP_SERVER_IMAGE, "GTCM_GNP_SERVER")
IMAGE_TABLE_ENTRY (DBCERTIFY_IMAGE,       "DBCERTIFY")
