/****************************************************************
 *								*
 *	Copyright 2001, 2004 Sanchez Computer Associates, Inc.	*
 *								*
 *	This source code contains the intellectual property	*
 *	of its copyright holder(s), and is made available	*
 *	under a license.  If you do not know the terms of	*
 *	the license, please stop and do not read further.	*
 *								*
 ****************************************************************/

#ifndef CG_VAR_H_INCLUDED
#define CG_VAR_H_INCLUDED

void cg_var(mvar *v, var_tabent **p);
void ind_cg_var(mvar *v, var_tabent **p);

#endif
