/****************************************************************
 *								*
 * Copyright (c) 2001-2018 Fidelity National Information	*
 * Services, Inc. and/or its subsidiaries. All rights reserved.	*
 *								*
 *	This source code contains the intellectual property	*
 *	of its copyright holder(s), and is made available	*
 *	under a license.  If you do not know the terms of	*
 *	the license, please stop and do not read further.	*
 *								*
 ****************************************************************/

#ifndef MLK_PRCBLK_DELETE_DEFINED

/* Declare parms for mlk_prcblk_delete.c */

void mlk_prcblk_delete(mlk_pvtctl_ptr_t pctl,
		       mlk_shrblk_ptr_t d,
		       uint4 pid);


#define MLK_PRCBLK_DELETE_DEFINED

#endif
