/****************************************************************
 *								*
 *	Copyright 2002 Sanchez Computer Associates, Inc.	*
 *								*
 *	This source code contains the intellectual property	*
 *	of its copyright holder(s), and is made available	*
 *	under a license.  If you do not know the terms of	*
 *	the license, please stop and do not read further.	*
 *								*
 ****************************************************************/

#ifndef FIVE_BIT_H_INCLUDED
#define FIVE_BIT_H_INCLUDED

unsigned short five_bit(unsigned char *src);

#endif /* FIVE_BIT_H_INCLUDED */
