/****************************************************************
 *								*
 * Copyright (c) 2007-2018 Fidelity National Information	*
 * Services, Inc. and/or its subsidiaries. All rights reserved.	*
 *								*
 *	This source code contains the intellectual property	*
 *	of its copyright holder(s), and is made available	*
 *	under a license.  If you do not know the terms of	*
 *	the license, please stop and do not read further.	*
 *								*
 ****************************************************************/

#ifndef WCS_WRITE_IN_PROGRESS_WAIT_H_INCLUDED
#define WCS_WRITE_IN_PROGRESS_WAIT_H_INCLUDED

boolean_t	wcs_write_in_progress_wait(node_local_ptr_t cnl, cache_rec_ptr_t cr, wbtest_code_t wbox_test_code);

#endif
