/****************************************************************
 *								*
 *	Copyright 2006, 2009 Fidelity Information Services, Inc	*
 *								*
 *	This source code contains the intellectual property	*
 *	of its copyright holder(s), and is made available	*
 *	under a license.  If you do not know the terms of	*
 *	the license, please stop and do not read further.	*
 *								*
 ****************************************************************/

ICU_DEF(u_getVersion)
ICU_DEF(u_isalpha)
ICU_DEF(u_islower)
ICU_DEF(u_isupper)
ICU_DEF(u_istitle)
ICU_DEF(u_iscntrl)
ICU_DEF(u_ispunct)
ICU_DEF(u_isdigit)
ICU_DEF(u_isspace)
ICU_DEF(u_isblank)
ICU_DEF(u_isprint)
ICU_DEF(u_getIntPropertyValue)
ICU_DEF(u_strFromUTF8)
ICU_DEF(u_strToUTF8)
ICU_DEF(u_strToLower)
ICU_DEF(u_strToUpper)
ICU_DEF(u_strToTitle)
ICU_DEF(u_strlen)
ICU_DEF(u_toupper)
ICU_DEF(u_finit)
ICU_DEF(u_fgets)
ICU_DEF(u_fclose)
ICU_DEF(u_feof)
ICU_DEF(ucnv_open)
ICU_DEF(ucnv_close)
ICU_DEF(ucnv_convertEx)
ICU_DEF(ucnv_getMaxCharSize)
ICU_DEF(ucnv_getMinCharSize)
ICU_DEF(ucnv_setToUCallBack)
ICU_DEF(ucnv_getName)
ICU_DEF(u_charType)
