/****************************************************************
 *								*
 *	Copyright 2001, 2010 Fidelity Information Services, Inc	*
 *								*
 *	This source code contains the intellectual property	*
 *	of its copyright holder(s), and is made available	*
 *	under a license.  If you do not know the terms of	*
 *	the license, please stop and do not read further.	*
 *								*
 ****************************************************************/

#ifndef ADVANCEWINDOW_included
#define ADVANCEWINDOW_included

void advancewindow(void);
#ifdef GTM_TRIGGER
void advwindw_hash_in_mname_allowed(void);
#endif
#endif /* ADVANCEWINDOW_included */
