/****************************************************************
 *								*
 *	Copyright 2012 Fidelity Information Services, Inc	*
 *								*
 *	This source code contains the intellectual property	*
 *	of its copyright holder(s), and is made available	*
 *	under a license.  If you do not know the terms of	*
 *	the license, please stop and do not read further.	*
 *								*
 ****************************************************************/

#ifndef GTM_MVAE_INCLUDED
#define GTM_MVAE_INCLUDED

void *gtm_memcpy_validate_and_execute(void *target, const void *src, size_t len);

#endif
