/****************************************************************
 *								*
 *	Copyright 2012 Fidelity Information Services, Inc	*
 *								*
 *	This source code contains the intellectual property	*
 *	of its copyright holder(s), and is made available	*
 *	under a license.  If you do not know the terms of	*
 *	the license, please stop and do not read further.	*
 *								*
 ****************************************************************/

#ifndef LOCK_STR_TO_BUFF_H
#define LOCK_STR_TO_BUFF_H

void		lock_str_to_buff(mlk_pvtblk *pvt_ptr1, char *outbuff, int outbuff_len);

#endif
