/****************************************************************
 *								*
 * Copyright (c) 2001-2019 Fidelity National Information	*
 * Services, Inc. and/or its subsidiaries. All rights reserved.	*
 *								*
 *	This source code contains the intellectual property	*
 *	of its copyright holder(s), and is made available	*
 *	under a license.  If you do not know the terms of	*
 *	the license, please stop and do not read further.	*
 *								*
 ****************************************************************/

#ifndef T_END_DEFINED
#define T_END_DEFINED

#define TN_NOT_SPECIFIED	(MAX_TN_ANY + 1)
trans_num t_end(srch_hist *hist1, srch_hist *hist2, trans_num ctn);

#endif
