/****************************************************************
 *								*
 *	Copyright 2010 Fidelity Information Services, Inc	*
 *								*
 *	This source code contains the intellectual property	*
 *	of its copyright holder(s), and is made available	*
 *	under a license.  If you do not know the terms of	*
 *	the license, please stop and do not read further.	*
 *								*
 ****************************************************************/

#ifndef MUPIP_LOAD_TRIGGERFILE_INCLUDED
#define MUPIP_LOAD_TRIGGERFILE_INCLUDED

void mupip_load_triggerfile(char *trigger_filename, int trigger_filename_len, char p_m);

#endif /* MUPIP_LOAD_TRIGGERFILE_INCLUDED */
