/****************************************************************
 *								*
 *	Copyright 2010 Fidelity Information Services, Inc	*
 *								*
 *	This source code contains the intellectual property	*
 *	of its copyright holder(s), and is made available	*
 *	under a license.  If you do not know the terms of	*
 *	the license, please stop and do not read further.	*
 *								*
 ****************************************************************/

#ifndef TRIGGER__FILL_XECUTE_BUFFER_H_INCLUDED
#define TRIGGER__FILL_XECUTE_BUFFER_H_INCLUDED

int trigger_fill_xecute_buffer(gv_trigger_t *trigdsc);

#endif
