/****************************************************************
 *								*
 *	Copyright 2001, 2002 Sanchez Computer Associates, Inc.	*
 *								*
 *	This source code contains the intellectual property	*
 *	of its copyright holder(s), and is made available	*
 *	under a license.  If you do not know the terms of	*
 *	the license, please stop and do not read further.	*
 *								*
 ****************************************************************/

#ifndef GET_COMMAND_LINE_INCLUDED
#define GET_COMMAND_LINE_INCLUDED

void get_command_line(mval *result, boolean_t zcmd_line);

#endif /* GET_COMMAND_LINE_INCLUDED */
