/****************************************************************
 *								*
 *	Copyright 2001 Sanchez Computer Associates, Inc.	*
 *								*
 *	This source code contains the intellectual property	*
 *	of its copyright holder(s), and is made available	*
 *	under a license.  If you do not know the terms of	*
 *	the license, please stop and do not read further.	*
 *								*
 ****************************************************************/

#ifndef CVTTIME_INCLUDED
#define CVTTIME_INCLUDED

int4 cvttime(mval *src, int4 *tim);

#endif /* CVTTIME_INCLUDED */
