;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;								;
; Copyright (c) 2010-2020 Fidelity National Information 	;
; Services, Inc. and/or its subsidiaries. All rights reserved.	;
;								;
;	This source code contains the intellectual property	;
;	of its copyright holder(s), and is made available	;
;	under a license.  If you do not know the terms of	;
;	the license, please stop and do not read further.	;
;								;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;
; Generated by scantypedefs.m at 19:04:41 on 2020-09-28
; Build done with GT.M version: GT.M V6.3-014 Linux x86_64
;
; Environment variables during run:
;   $gtm_dist: /usr/library/V63014/pro
;   $gtm_exe: /usr/library/V63014/pro
;   $gtm_src: /usr/library/V63014/src
;   $gtm_inc: /usr/library/V63014/inc
;
; Note this file should not be manually invoked
;
	Write "GTM-E-GTMSDFILE This routine (",$TEXT(+0),") should not be manually invoked",!
	Quit
;
; Entry point used by gtmpcat and others to define GTM structure fields
;
Init
	; GT.M structure and field definitions
	Set gtmsdver="1.2.0"
	Set gtmsdtyp="pro"
	;
	;
	Set gtmtypes("ABS_TIME")="struct"
	Set gtmtypes("ABS_TIME",0)=2
	Set gtmtypes("ABS_TIME","len")=16
	Set gtmtypes("ABS_TIME",1,"name")="ABS_TIME.at_sec"
	Set gtmtypes("ABS_TIME",1,"off")=0
	Set gtmtypes("ABS_TIME",1,"len")=8
	Set gtmtypes("ABS_TIME",1,"type")="long"
	Set gtmtypfldindx("ABS_TIME","at_sec")=1
	Set gtmtypes("ABS_TIME",2,"name")="ABS_TIME.at_usec"
	Set gtmtypes("ABS_TIME",2,"off")=8
	Set gtmtypes("ABS_TIME",2,"len")=8
	Set gtmtypes("ABS_TIME",2,"type")="long"
	Set gtmtypfldindx("ABS_TIME","at_usec")=2
	;
	Set gtmtypes("BFILE")="struct"
	Set gtmtypes("BFILE",0)=10
	Set gtmtypes("BFILE","len")=56
	Set gtmtypes("BFILE",1,"name")="BFILE.fd"
	Set gtmtypes("BFILE",1,"off")=0
	Set gtmtypes("BFILE",1,"len")=4
	Set gtmtypes("BFILE",1,"type")="int"
	Set gtmtypfldindx("BFILE","fd")=1
	Set gtmtypes("BFILE",2,"name")="BFILE.path"
	Set gtmtypes("BFILE",2,"off")=8
	Set gtmtypes("BFILE",2,"len")=8
	Set gtmtypes("BFILE",2,"type")="addr"
	Set gtmtypfldindx("BFILE","path")=2
	Set gtmtypes("BFILE",3,"name")="BFILE.oflag"
	Set gtmtypes("BFILE",3,"off")=16
	Set gtmtypes("BFILE",3,"len")=4
	Set gtmtypes("BFILE",3,"type")="int"
	Set gtmtypfldindx("BFILE","oflag")=3
	Set gtmtypes("BFILE",4,"name")="BFILE.mode"
	Set gtmtypes("BFILE",4,"off")=20
	Set gtmtypes("BFILE",4,"len")=4
	Set gtmtypes("BFILE",4,"type")="int"
	Set gtmtypfldindx("BFILE","mode")=4
	Set gtmtypes("BFILE",5,"name")="BFILE.blksiz"
	Set gtmtypes("BFILE",5,"off")=24
	Set gtmtypes("BFILE",5,"len")=4
	Set gtmtypes("BFILE",5,"type")="int"
	Set gtmtypfldindx("BFILE","blksiz")=5
	Set gtmtypes("BFILE",6,"name")="BFILE.bufsiz"
	Set gtmtypes("BFILE",6,"off")=28
	Set gtmtypes("BFILE",6,"len")=4
	Set gtmtypes("BFILE",6,"type")="int"
	Set gtmtypfldindx("BFILE","bufsiz")=6
	Set gtmtypes("BFILE",7,"name")="BFILE.buf"
	Set gtmtypes("BFILE",7,"off")=32
	Set gtmtypes("BFILE",7,"len")=8
	Set gtmtypes("BFILE",7,"type")="addr"
	Set gtmtypfldindx("BFILE","buf")=7
	Set gtmtypes("BFILE",8,"name")="BFILE.bptr"
	Set gtmtypes("BFILE",8,"off")=40
	Set gtmtypes("BFILE",8,"len")=8
	Set gtmtypes("BFILE",8,"type")="addr"
	Set gtmtypfldindx("BFILE","bptr")=8
	Set gtmtypes("BFILE",9,"name")="BFILE.remaining"
	Set gtmtypes("BFILE",9,"off")=48
	Set gtmtypes("BFILE",9,"len")=4
	Set gtmtypes("BFILE",9,"type")="int"
	Set gtmtypfldindx("BFILE","remaining")=9
	Set gtmtypes("BFILE",10,"name")="BFILE.write_mode"
	Set gtmtypes("BFILE",10,"off")=52
	Set gtmtypes("BFILE",10,"len")=4
	Set gtmtypes("BFILE",10,"type")="int"
	Set gtmtypfldindx("BFILE","write_mode")=10
	;
	Set gtmtypes("CLI_ENTRY")="struct"
	Set gtmtypes("CLI_ENTRY",0)=9
	Set gtmtypes("CLI_ENTRY","len")=96
	Set gtmtypes("CLI_ENTRY",1,"name")="CLI_ENTRY.name"
	Set gtmtypes("CLI_ENTRY",1,"off")=0
	Set gtmtypes("CLI_ENTRY",1,"len")=25
	Set gtmtypes("CLI_ENTRY",1,"type")="char"
	Set gtmtypfldindx("CLI_ENTRY","name")=1
	Set gtmtypes("CLI_ENTRY",2,"name")="CLI_ENTRY.func"
	Set gtmtypes("CLI_ENTRY",2,"off")=32
	Set gtmtypes("CLI_ENTRY",2,"len")=8
	Set gtmtypes("CLI_ENTRY",2,"type")="addr"
	Set gtmtypfldindx("CLI_ENTRY","func")=2
	Set gtmtypes("CLI_ENTRY",3,"name")="CLI_ENTRY.parms"
	Set gtmtypes("CLI_ENTRY",3,"off")=40
	Set gtmtypes("CLI_ENTRY",3,"len")=8
	Set gtmtypes("CLI_ENTRY",3,"type")="addr"
	Set gtmtypfldindx("CLI_ENTRY","parms")=3
	Set gtmtypes("CLI_ENTRY",4,"name")="CLI_ENTRY.parm_values"
	Set gtmtypes("CLI_ENTRY",4,"off")=48
	Set gtmtypes("CLI_ENTRY",4,"len")=8
	Set gtmtypes("CLI_ENTRY",4,"type")="addr"
	Set gtmtypfldindx("CLI_ENTRY","parm_values")=4
	Set gtmtypes("CLI_ENTRY",5,"name")="CLI_ENTRY.qual_vals"
	Set gtmtypes("CLI_ENTRY",5,"off")=56
	Set gtmtypes("CLI_ENTRY",5,"len")=8
	Set gtmtypes("CLI_ENTRY",5,"type")="addr"
	Set gtmtypfldindx("CLI_ENTRY","qual_vals")=5
	Set gtmtypes("CLI_ENTRY",6,"name")="CLI_ENTRY.disallow_func"
	Set gtmtypes("CLI_ENTRY",6,"off")=64
	Set gtmtypes("CLI_ENTRY",6,"len")=8
	Set gtmtypes("CLI_ENTRY",6,"type")="addr"
	Set gtmtypfldindx("CLI_ENTRY","disallow_func")=6
	Set gtmtypes("CLI_ENTRY",7,"name")="CLI_ENTRY.dfault_str"
	Set gtmtypes("CLI_ENTRY",7,"off")=72
	Set gtmtypes("CLI_ENTRY",7,"len")=8
	Set gtmtypes("CLI_ENTRY",7,"type")="addr"
	Set gtmtypfldindx("CLI_ENTRY","dfault_str")=7
	Set gtmtypes("CLI_ENTRY",8,"name")="CLI_ENTRY.max_parms"
	Set gtmtypes("CLI_ENTRY",8,"off")=82
	Set gtmtypes("CLI_ENTRY",8,"len")=2
	Set gtmtypes("CLI_ENTRY",8,"type")="unsigned-short"
	Set gtmtypfldindx("CLI_ENTRY","max_parms")=8
	Set gtmtypes("CLI_ENTRY",9,"name")="CLI_ENTRY.pval_str"
	Set gtmtypes("CLI_ENTRY",9,"off")=88
	Set gtmtypes("CLI_ENTRY",9,"len")=8
	Set gtmtypes("CLI_ENTRY",9,"type")="addr"
	Set gtmtypfldindx("CLI_ENTRY","pval_str")=9
	;
	Set gtmtypes("CLI_PARM")="struct"
	Set gtmtypes("CLI_PARM",0)=3
	Set gtmtypes("CLI_PARM","len")=56
	Set gtmtypes("CLI_PARM",1,"name")="CLI_PARM.name"
	Set gtmtypes("CLI_PARM",1,"off")=0
	Set gtmtypes("CLI_PARM",1,"len")=25
	Set gtmtypes("CLI_PARM",1,"type")="char"
	Set gtmtypfldindx("CLI_PARM","name")=1
	Set gtmtypes("CLI_PARM",2,"name")="CLI_PARM.prompt"
	Set gtmtypes("CLI_PARM",2,"off")=25
	Set gtmtypes("CLI_PARM",2,"len")=25
	Set gtmtypes("CLI_PARM",2,"type")="char"
	Set gtmtypfldindx("CLI_PARM","prompt")=2
	Set gtmtypes("CLI_PARM",3,"name")="CLI_PARM.parm_required"
	Set gtmtypes("CLI_PARM",3,"off")=52
	Set gtmtypes("CLI_PARM",3,"len")=4
	Set gtmtypes("CLI_PARM",3,"type")="boolean_t"
	Set gtmtypfldindx("CLI_PARM","parm_required")=3
	;
	Set gtmtypes("Dev_param_pair")="struct"
	Set gtmtypes("Dev_param_pair",0)=2
	Set gtmtypes("Dev_param_pair","len")=16
	Set gtmtypes("Dev_param_pair",1,"name")="Dev_param_pair.name"
	Set gtmtypes("Dev_param_pair",1,"off")=0
	Set gtmtypes("Dev_param_pair",1,"len")=8
	Set gtmtypes("Dev_param_pair",1,"type")="addr"
	Set gtmtypfldindx("Dev_param_pair","name")=1
	Set gtmtypes("Dev_param_pair",2,"name")="Dev_param_pair.definition"
	Set gtmtypes("Dev_param_pair",2,"off")=8
	Set gtmtypes("Dev_param_pair",2,"len")=8
	Set gtmtypes("Dev_param_pair",2,"type")="addr"
	Set gtmtypfldindx("Dev_param_pair","definition")=2
	;
	Set gtmtypes("Dev_param_pairs")="struct"
	Set gtmtypes("Dev_param_pairs",0)=4
	Set gtmtypes("Dev_param_pairs","len")=56
	Set gtmtypes("Dev_param_pairs",1,"name")="Dev_param_pairs.num_pairs"
	Set gtmtypes("Dev_param_pairs",1,"off")=0
	Set gtmtypes("Dev_param_pairs",1,"len")=4
	Set gtmtypes("Dev_param_pairs",1,"type")="int"
	Set gtmtypfldindx("Dev_param_pairs","num_pairs")=1
	Set gtmtypes("Dev_param_pairs",2,"name")="Dev_param_pairs.pairs"
	Set gtmtypes("Dev_param_pairs",2,"off")=8
	Set gtmtypes("Dev_param_pairs",2,"len")=48
	Set gtmtypes("Dev_param_pairs",2,"type")="Dev_param_pair"
	Set gtmtypfldindx("Dev_param_pairs","pairs")=2
	Set gtmtypes("Dev_param_pairs",2,"dim")=3
	Set gtmtypes("Dev_param_pairs",3,"name")="Dev_param_pairs.pairs[0].name"
	Set gtmtypes("Dev_param_pairs",3,"off")=8
	Set gtmtypes("Dev_param_pairs",3,"len")=8
	Set gtmtypes("Dev_param_pairs",3,"type")="addr"
	Set gtmtypfldindx("Dev_param_pairs","pairs[0].name")=3
	Set gtmtypes("Dev_param_pairs",4,"name")="Dev_param_pairs.pairs[0].definition"
	Set gtmtypes("Dev_param_pairs",4,"off")=16
	Set gtmtypes("Dev_param_pairs",4,"len")=8
	Set gtmtypes("Dev_param_pairs",4,"type")="addr"
	Set gtmtypfldindx("Dev_param_pairs","pairs[0].definition")=4
	;
	Set gtmtypes("GT_TIMER")="struct"
	Set gtmtypes("GT_TIMER",0)=14
	Set gtmtypes("GT_TIMER","len")=80
	Set gtmtypes("GT_TIMER",1,"name")="GT_TIMER.expir_time"
	Set gtmtypes("GT_TIMER",1,"off")=0
	Set gtmtypes("GT_TIMER",1,"len")=16
	Set gtmtypes("GT_TIMER",1,"type")="ABS_TIME"
	Set gtmtypfldindx("GT_TIMER","expir_time")=1
	Set gtmtypes("GT_TIMER",2,"name")="GT_TIMER.expir_time.at_sec"
	Set gtmtypes("GT_TIMER",2,"off")=0
	Set gtmtypes("GT_TIMER",2,"len")=8
	Set gtmtypes("GT_TIMER",2,"type")="long"
	Set gtmtypfldindx("GT_TIMER","expir_time.at_sec")=2
	Set gtmtypes("GT_TIMER",3,"name")="GT_TIMER.expir_time.at_usec"
	Set gtmtypes("GT_TIMER",3,"off")=8
	Set gtmtypes("GT_TIMER",3,"len")=8
	Set gtmtypes("GT_TIMER",3,"type")="long"
	Set gtmtypfldindx("GT_TIMER","expir_time.at_usec")=3
	Set gtmtypes("GT_TIMER",4,"name")="GT_TIMER.start_time"
	Set gtmtypes("GT_TIMER",4,"off")=16
	Set gtmtypes("GT_TIMER",4,"len")=16
	Set gtmtypes("GT_TIMER",4,"type")="ABS_TIME"
	Set gtmtypfldindx("GT_TIMER","start_time")=4
	Set gtmtypes("GT_TIMER",5,"name")="GT_TIMER.start_time.at_sec"
	Set gtmtypes("GT_TIMER",5,"off")=16
	Set gtmtypes("GT_TIMER",5,"len")=8
	Set gtmtypes("GT_TIMER",5,"type")="long"
	Set gtmtypfldindx("GT_TIMER","start_time.at_sec")=5
	Set gtmtypes("GT_TIMER",6,"name")="GT_TIMER.start_time.at_usec"
	Set gtmtypes("GT_TIMER",6,"off")=24
	Set gtmtypes("GT_TIMER",6,"len")=8
	Set gtmtypes("GT_TIMER",6,"type")="long"
	Set gtmtypfldindx("GT_TIMER","start_time.at_usec")=6
	Set gtmtypes("GT_TIMER",7,"name")="GT_TIMER.handler"
	Set gtmtypes("GT_TIMER",7,"off")=32
	Set gtmtypes("GT_TIMER",7,"len")=8
	Set gtmtypes("GT_TIMER",7,"type")="addr"
	Set gtmtypfldindx("GT_TIMER","handler")=7
	Set gtmtypes("GT_TIMER",8,"name")="GT_TIMER.next"
	Set gtmtypes("GT_TIMER",8,"off")=40
	Set gtmtypes("GT_TIMER",8,"len")=8
	Set gtmtypes("GT_TIMER",8,"type")="addr"
	Set gtmtypfldindx("GT_TIMER","next")=8
	Set gtmtypes("GT_TIMER",9,"name")="GT_TIMER.tid"
	Set gtmtypes("GT_TIMER",9,"off")=48
	Set gtmtypes("GT_TIMER",9,"len")=8
	Set gtmtypes("GT_TIMER",9,"type")="intptr_t"
	Set gtmtypfldindx("GT_TIMER","tid")=9
	Set gtmtypes("GT_TIMER",10,"name")="GT_TIMER.safe"
	Set gtmtypes("GT_TIMER",10,"off")=56
	Set gtmtypes("GT_TIMER",10,"len")=4
	Set gtmtypes("GT_TIMER",10,"type")="int"
	Set gtmtypfldindx("GT_TIMER","safe")=10
	Set gtmtypes("GT_TIMER",11,"name")="GT_TIMER.block_int"
	Set gtmtypes("GT_TIMER",11,"off")=60
	Set gtmtypes("GT_TIMER",11,"len")=4
	Set gtmtypes("GT_TIMER",11,"type")="int"
	Set gtmtypfldindx("GT_TIMER","block_int")=11
	Set gtmtypes("GT_TIMER",12,"name")="GT_TIMER.hd_len_max"
	Set gtmtypes("GT_TIMER",12,"off")=64
	Set gtmtypes("GT_TIMER",12,"len")=4
	Set gtmtypes("GT_TIMER",12,"type")="int"
	Set gtmtypfldindx("GT_TIMER","hd_len_max")=12
	Set gtmtypes("GT_TIMER",13,"name")="GT_TIMER.hd_len"
	Set gtmtypes("GT_TIMER",13,"off")=68
	Set gtmtypes("GT_TIMER",13,"len")=4
	Set gtmtypes("GT_TIMER",13,"type")="int"
	Set gtmtypfldindx("GT_TIMER","hd_len")=13
	Set gtmtypes("GT_TIMER",14,"name")="GT_TIMER.hd_data"
	Set gtmtypes("GT_TIMER",14,"off")=72
	Set gtmtypes("GT_TIMER",14,"len")=1
	Set gtmtypes("GT_TIMER",14,"type")="char"
	Set gtmtypfldindx("GT_TIMER","hd_data")=14
	;
	Set gtmtypes("IN_PARMS")="struct"
	Set gtmtypes("IN_PARMS",0)=5
	Set gtmtypes("IN_PARMS","len")=32
	Set gtmtypes("IN_PARMS",1,"name")="IN_PARMS.argc"
	Set gtmtypes("IN_PARMS",1,"off")=0
	Set gtmtypes("IN_PARMS",1,"len")=4
	Set gtmtypes("IN_PARMS",1,"type")="int"
	Set gtmtypfldindx("IN_PARMS","argc")=1
	Set gtmtypes("IN_PARMS",2,"name")="IN_PARMS.argv"
	Set gtmtypes("IN_PARMS",2,"off")=8
	Set gtmtypes("IN_PARMS",2,"len")=8
	Set gtmtypes("IN_PARMS",2,"type")="addr"
	Set gtmtypfldindx("IN_PARMS","argv")=2
	Set gtmtypes("IN_PARMS",3,"name")="IN_PARMS.tp"
	Set gtmtypes("IN_PARMS",3,"off")=16
	Set gtmtypes("IN_PARMS",3,"len")=8
	Set gtmtypes("IN_PARMS",3,"type")="addr"
	Set gtmtypfldindx("IN_PARMS","tp")=3
	Set gtmtypes("IN_PARMS",4,"name")="IN_PARMS.buflen"
	Set gtmtypes("IN_PARMS",4,"off")=24
	Set gtmtypes("IN_PARMS",4,"len")=4
	Set gtmtypes("IN_PARMS",4,"type")="int"
	Set gtmtypfldindx("IN_PARMS","buflen")=4
	Set gtmtypes("IN_PARMS",5,"name")="IN_PARMS.in_str"
	Set gtmtypes("IN_PARMS",5,"off")=28
	Set gtmtypes("IN_PARMS",5,"len")=0
	Set gtmtypes("IN_PARMS",5,"type")="char"
	Set gtmtypfldindx("IN_PARMS","in_str")=5
	;
	Set gtmtypes("activelv_dbg_t")="struct"
	Set gtmtypes("activelv_dbg_t",0)=8
	Set gtmtypes("activelv_dbg_t","len")=56
	Set gtmtypes("activelv_dbg_t",1,"name")="activelv_dbg_t.active_lv"
	Set gtmtypes("activelv_dbg_t",1,"off")=0
	Set gtmtypes("activelv_dbg_t",1,"len")=8
	Set gtmtypes("activelv_dbg_t",1,"type")="addr"
	Set gtmtypfldindx("activelv_dbg_t","active_lv")=1
	Set gtmtypes("activelv_dbg_t",2,"name")="activelv_dbg_t.newlv"
	Set gtmtypes("activelv_dbg_t",2,"off")=8
	Set gtmtypes("activelv_dbg_t",2,"len")=8
	Set gtmtypes("activelv_dbg_t",2,"type")="addr"
	Set gtmtypfldindx("activelv_dbg_t","newlv")=2
	Set gtmtypes("activelv_dbg_t",3,"name")="activelv_dbg_t.frame_pointer"
	Set gtmtypes("activelv_dbg_t",3,"off")=16
	Set gtmtypes("activelv_dbg_t",3,"len")=8
	Set gtmtypes("activelv_dbg_t",3,"type")="addr"
	Set gtmtypfldindx("activelv_dbg_t","frame_pointer")=3
	Set gtmtypes("activelv_dbg_t",4,"name")="activelv_dbg_t.curr_symval"
	Set gtmtypes("activelv_dbg_t",4,"off")=24
	Set gtmtypes("activelv_dbg_t",4,"len")=8
	Set gtmtypes("activelv_dbg_t",4,"type")="addr"
	Set gtmtypfldindx("activelv_dbg_t","curr_symval")=4
	Set gtmtypes("activelv_dbg_t",5,"name")="activelv_dbg_t.mpc"
	Set gtmtypes("activelv_dbg_t",5,"off")=32
	Set gtmtypes("activelv_dbg_t",5,"len")=8
	Set gtmtypes("activelv_dbg_t",5,"type")="addr"
	Set gtmtypfldindx("activelv_dbg_t","mpc")=5
	Set gtmtypes("activelv_dbg_t",6,"name")="activelv_dbg_t.ctxt"
	Set gtmtypes("activelv_dbg_t",6,"off")=40
	Set gtmtypes("activelv_dbg_t",6,"len")=8
	Set gtmtypes("activelv_dbg_t",6,"type")="addr"
	Set gtmtypfldindx("activelv_dbg_t","ctxt")=6
	Set gtmtypes("activelv_dbg_t",7,"name")="activelv_dbg_t.count"
	Set gtmtypes("activelv_dbg_t",7,"off")=48
	Set gtmtypes("activelv_dbg_t",7,"len")=4
	Set gtmtypes("activelv_dbg_t",7,"type")="unsigned-int"
	Set gtmtypfldindx("activelv_dbg_t","count")=7
	Set gtmtypes("activelv_dbg_t",8,"name")="activelv_dbg_t.type"
	Set gtmtypes("activelv_dbg_t",8,"off")=52
	Set gtmtypes("activelv_dbg_t",8,"len")=4
	Set gtmtypes("activelv_dbg_t",8,"type")="unsigned-int"
	Set gtmtypfldindx("activelv_dbg_t","type")=8
	;
	Set gtmtypes("backup_reg_list")="struct"
	Set gtmtypes("backup_reg_list",0)=18
	Set gtmtypes("backup_reg_list","len")=344
	Set gtmtypes("backup_reg_list",1,"name")="backup_reg_list.fPtr"
	Set gtmtypes("backup_reg_list",1,"off")=0
	Set gtmtypes("backup_reg_list",1,"len")=8
	Set gtmtypes("backup_reg_list",1,"type")="addr"
	Set gtmtypfldindx("backup_reg_list","fPtr")=1
	Set gtmtypes("backup_reg_list",2,"name")="backup_reg_list.reg"
	Set gtmtypes("backup_reg_list",2,"off")=8
	Set gtmtypes("backup_reg_list",2,"len")=8
	Set gtmtypes("backup_reg_list",2,"type")="addr"
	Set gtmtypfldindx("backup_reg_list","reg")=2
	Set gtmtypes("backup_reg_list",3,"name")="backup_reg_list.unique_file_id"
	Set gtmtypes("backup_reg_list",3,"off")=16
	Set gtmtypes("backup_reg_list",3,"len")=16
	Set gtmtypes("backup_reg_list",3,"type")="unix_file_id"
	Set gtmtypfldindx("backup_reg_list","unique_file_id")=3
	Set gtmtypes("backup_reg_list",4,"name")="backup_reg_list.unique_file_id.inode"
	Set gtmtypes("backup_reg_list",4,"off")=16
	Set gtmtypes("backup_reg_list",4,"len")=8
	Set gtmtypes("backup_reg_list",4,"type")="ino_t"
	Set gtmtypfldindx("backup_reg_list","unique_file_id.inode")=4
	Set gtmtypes("backup_reg_list",5,"name")="backup_reg_list.unique_file_id.device"
	Set gtmtypes("backup_reg_list",5,"off")=24
	Set gtmtypes("backup_reg_list",5,"len")=8
	Set gtmtypes("backup_reg_list",5,"type")="dev_t"
	Set gtmtypfldindx("backup_reg_list","unique_file_id.device")=5
	Set gtmtypes("backup_reg_list",6,"name")="backup_reg_list.backup_file"
	Set gtmtypes("backup_reg_list",6,"off")=32
	Set gtmtypes("backup_reg_list",6,"len")=16
	Set gtmtypes("backup_reg_list",6,"type")="mstr"
	Set gtmtypfldindx("backup_reg_list","backup_file")=6
	Set gtmtypes("backup_reg_list",7,"name")="backup_reg_list.backup_file.char_len"
	Set gtmtypes("backup_reg_list",7,"off")=32
	Set gtmtypes("backup_reg_list",7,"len")=4
	Set gtmtypes("backup_reg_list",7,"type")="unsigned-int"
	Set gtmtypfldindx("backup_reg_list","backup_file.char_len")=7
	Set gtmtypes("backup_reg_list",8,"name")="backup_reg_list.backup_file.len"
	Set gtmtypes("backup_reg_list",8,"off")=36
	Set gtmtypes("backup_reg_list",8,"len")=4
	Set gtmtypes("backup_reg_list",8,"type")="int"
	Set gtmtypfldindx("backup_reg_list","backup_file.len")=8
	Set gtmtypes("backup_reg_list",9,"name")="backup_reg_list.backup_file.addr"
	Set gtmtypes("backup_reg_list",9,"off")=40
	Set gtmtypes("backup_reg_list",9,"len")=8
	Set gtmtypes("backup_reg_list",9,"type")="addr"
	Set gtmtypfldindx("backup_reg_list","backup_file.addr")=9
	Set gtmtypes("backup_reg_list",10,"name")="backup_reg_list.crashcnt"
	Set gtmtypes("backup_reg_list",10,"off")=48
	Set gtmtypes("backup_reg_list",10,"len")=2
	Set gtmtypes("backup_reg_list",10,"type")="short"
	Set gtmtypfldindx("backup_reg_list","crashcnt")=10
	Set gtmtypes("backup_reg_list",11,"name")="backup_reg_list.filler"
	Set gtmtypes("backup_reg_list",11,"off")=50
	Set gtmtypes("backup_reg_list",11,"len")=2
	Set gtmtypes("backup_reg_list",11,"type")="short"
	Set gtmtypfldindx("backup_reg_list","filler")=11
	Set gtmtypes("backup_reg_list",12,"name")="backup_reg_list.not_this_time"
	Set gtmtypes("backup_reg_list",12,"off")=52
	Set gtmtypes("backup_reg_list",12,"len")=4
	Set gtmtypes("backup_reg_list",12,"type")="int"
	Set gtmtypfldindx("backup_reg_list","not_this_time")=12
	Set gtmtypes("backup_reg_list",13,"name")="backup_reg_list.backup_to"
	Set gtmtypes("backup_reg_list",13,"off")=56
	Set gtmtypes("backup_reg_list",13,"len")=4
	Set gtmtypes("backup_reg_list",13,"type")="int"
	Set gtmtypfldindx("backup_reg_list","backup_to")=13
	Set gtmtypes("backup_reg_list",14,"name")="backup_reg_list.backup_hdr"
	Set gtmtypes("backup_reg_list",14,"off")=64
	Set gtmtypes("backup_reg_list",14,"len")=8
	Set gtmtypes("backup_reg_list",14,"type")="addr"
	Set gtmtypfldindx("backup_reg_list","backup_hdr")=14
	Set gtmtypes("backup_reg_list",15,"name")="backup_reg_list.tn"
	Set gtmtypes("backup_reg_list",15,"off")=72
	Set gtmtypes("backup_reg_list",15,"len")=8
	Set gtmtypes("backup_reg_list",15,"type")="uint64_t"
	Set gtmtypfldindx("backup_reg_list","tn")=15
	Set gtmtypes("backup_reg_list",16,"name")="backup_reg_list.last_blk_at_last_bkup"
	Set gtmtypes("backup_reg_list",16,"off")=80
	Set gtmtypes("backup_reg_list",16,"len")=4
	Set gtmtypes("backup_reg_list",16,"type")="int"
	Set gtmtypfldindx("backup_reg_list","last_blk_at_last_bkup")=16
	Set gtmtypes("backup_reg_list",17,"name")="backup_reg_list.backup_fd"
	Set gtmtypes("backup_reg_list",17,"off")=84
	Set gtmtypes("backup_reg_list",17,"len")=4
	Set gtmtypes("backup_reg_list",17,"type")="int"
	Set gtmtypfldindx("backup_reg_list","backup_fd")=17
	Set gtmtypes("backup_reg_list",18,"name")="backup_reg_list.backup_tempfile"
	Set gtmtypes("backup_reg_list",18,"off")=88
	Set gtmtypes("backup_reg_list",18,"len")=256
	Set gtmtypes("backup_reg_list",18,"type")="char"
	Set gtmtypfldindx("backup_reg_list","backup_tempfile")=18
	;
	Set gtmtypes("blk_hdr")="struct"
	Set gtmtypes("blk_hdr",0)=5
	Set gtmtypes("blk_hdr","len")=16
	Set gtmtypes("blk_hdr",1,"name")="blk_hdr.bver"
	Set gtmtypes("blk_hdr",1,"off")=0
	Set gtmtypes("blk_hdr",1,"len")=2
	Set gtmtypes("blk_hdr",1,"type")="unsigned-short"
	Set gtmtypfldindx("blk_hdr","bver")=1
	Set gtmtypes("blk_hdr",2,"name")="blk_hdr.filler"
	Set gtmtypes("blk_hdr",2,"off")=2
	Set gtmtypes("blk_hdr",2,"len")=1
	Set gtmtypes("blk_hdr",2,"type")="unsigned-char"
	Set gtmtypfldindx("blk_hdr","filler")=2
	Set gtmtypes("blk_hdr",3,"name")="blk_hdr.levl"
	Set gtmtypes("blk_hdr",3,"off")=3
	Set gtmtypes("blk_hdr",3,"len")=1
	Set gtmtypes("blk_hdr",3,"type")="unsigned-char"
	Set gtmtypfldindx("blk_hdr","levl")=3
	Set gtmtypes("blk_hdr",4,"name")="blk_hdr.bsiz"
	Set gtmtypes("blk_hdr",4,"off")=4
	Set gtmtypes("blk_hdr",4,"len")=4
	Set gtmtypes("blk_hdr",4,"type")="unsigned-int"
	Set gtmtypfldindx("blk_hdr","bsiz")=4
	Set gtmtypes("blk_hdr",5,"name")="blk_hdr.tn"
	Set gtmtypes("blk_hdr",5,"off")=8
	Set gtmtypes("blk_hdr",5,"len")=8
	Set gtmtypes("blk_hdr",5,"type")="uint64_t"
	Set gtmtypfldindx("blk_hdr","tn")=5
	;
	Set gtmtypes("blk_segment")="struct"
	Set gtmtypes("blk_segment",0)=2
	Set gtmtypes("blk_segment","len")=16
	Set gtmtypes("blk_segment",1,"name")="blk_segment.addr"
	Set gtmtypes("blk_segment",1,"off")=0
	Set gtmtypes("blk_segment",1,"len")=8
	Set gtmtypes("blk_segment",1,"type")="addr"
	Set gtmtypfldindx("blk_segment","addr")=1
	Set gtmtypes("blk_segment",2,"name")="blk_segment.len"
	Set gtmtypes("blk_segment",2,"off")=8
	Set gtmtypes("blk_segment",2,"len")=8
	Set gtmtypes("blk_segment",2,"type")="uint64_t"
	Set gtmtypfldindx("blk_segment","len")=2
	;
	Set gtmtypes("block_info")="struct"
	Set gtmtypes("block_info",0)=20
	Set gtmtypes("block_info","len")=128
	Set gtmtypes("block_info",1,"name")="block_info.tn"
	Set gtmtypes("block_info",1,"off")=0
	Set gtmtypes("block_info",1,"len")=4
	Set gtmtypes("block_info",1,"type")="v15_trans_num"
	Set gtmtypfldindx("block_info","tn")=1
	Set gtmtypes("block_info",2,"name")="block_info.blk_num"
	Set gtmtypes("block_info",2,"off")=4
	Set gtmtypes("block_info",2,"len")=4
	Set gtmtypes("block_info",2,"type")="int"
	Set gtmtypfldindx("block_info","blk_num")=2
	Set gtmtypes("block_info",3,"name")="block_info.usage"
	Set gtmtypes("block_info",3,"off")=8
	Set gtmtypes("block_info",3,"len")=4
	Set gtmtypes("block_info",3,"type")="int"
	Set gtmtypfldindx("block_info","usage")=3
	Set gtmtypes("block_info",4,"name")="block_info.blk_type"
	Set gtmtypes("block_info",4,"off")=12
	Set gtmtypes("block_info",4,"len")=4
	Set gtmtypes("block_info",4,"type")="int"
	Set gtmtypfldindx("block_info","blk_type")=4
	Set gtmtypes("block_info",5,"name")="block_info.upd_addr"
	Set gtmtypes("block_info",5,"off")=16
	Set gtmtypes("block_info",5,"len")=8
	Set gtmtypes("block_info",5,"type")="addr"
	Set gtmtypfldindx("block_info","upd_addr")=5
	Set gtmtypes("block_info",6,"name")="block_info.found_in_cache"
	Set gtmtypes("block_info",6,"off")=24
	Set gtmtypes("block_info",6,"len")=4
	Set gtmtypes("block_info",6,"type")="boolean_t"
	Set gtmtypfldindx("block_info","found_in_cache")=6
	Set gtmtypes("block_info",7,"name")="block_info.old_buff"
	Set gtmtypes("block_info",7,"off")=32
	Set gtmtypes("block_info",7,"len")=8
	Set gtmtypes("block_info",7,"type")="addr"
	Set gtmtypfldindx("block_info","old_buff")=7
	Set gtmtypes("block_info",8,"name")="block_info.new_buff"
	Set gtmtypes("block_info",8,"off")=40
	Set gtmtypes("block_info",8,"len")=8
	Set gtmtypes("block_info",8,"type")="addr"
	Set gtmtypfldindx("block_info","new_buff")=8
	Set gtmtypes("block_info",9,"name")="block_info.prev_rec"
	Set gtmtypes("block_info",9,"off")=48
	Set gtmtypes("block_info",9,"len")=8
	Set gtmtypes("block_info",9,"type")="addr"
	Set gtmtypfldindx("block_info","prev_rec")=9
	Set gtmtypes("block_info",10,"name")="block_info.prev_match"
	Set gtmtypes("block_info",10,"off")=56
	Set gtmtypes("block_info",10,"len")=4
	Set gtmtypes("block_info",10,"type")="unsigned-int"
	Set gtmtypfldindx("block_info","prev_match")=10
	Set gtmtypes("block_info",11,"name")="block_info.curr_rec"
	Set gtmtypes("block_info",11,"off")=64
	Set gtmtypes("block_info",11,"len")=8
	Set gtmtypes("block_info",11,"type")="addr"
	Set gtmtypfldindx("block_info","curr_rec")=11
	Set gtmtypes("block_info",12,"name")="block_info.curr_match"
	Set gtmtypes("block_info",12,"off")=72
	Set gtmtypes("block_info",12,"len")=4
	Set gtmtypes("block_info",12,"type")="unsigned-int"
	Set gtmtypfldindx("block_info","curr_match")=12
	Set gtmtypes("block_info",13,"name")="block_info.curr_blk_key"
	Set gtmtypes("block_info",13,"off")=80
	Set gtmtypes("block_info",13,"len")=8
	Set gtmtypes("block_info",13,"type")="addr"
	Set gtmtypfldindx("block_info","curr_blk_key")=13
	Set gtmtypes("block_info",14,"name")="block_info.prev_blk_key"
	Set gtmtypes("block_info",14,"off")=88
	Set gtmtypes("block_info",14,"len")=8
	Set gtmtypes("block_info",14,"type")="addr"
	Set gtmtypfldindx("block_info","prev_blk_key")=14
	Set gtmtypes("block_info",15,"name")="block_info.ins_rec"
	Set gtmtypes("block_info",15,"off")=96
	Set gtmtypes("block_info",15,"len")=16
	Set gtmtypes("block_info",15,"type")="dbc_inserted_rec"
	Set gtmtypfldindx("block_info","ins_rec")=15
	Set gtmtypes("block_info",16,"name")="block_info.ins_rec.ins_key"
	Set gtmtypes("block_info",16,"off")=96
	Set gtmtypes("block_info",16,"len")=8
	Set gtmtypes("block_info",16,"type")="addr"
	Set gtmtypfldindx("block_info","ins_rec.ins_key")=16
	Set gtmtypes("block_info",17,"name")="block_info.ins_rec.blk_id"
	Set gtmtypes("block_info",17,"off")=104
	Set gtmtypes("block_info",17,"len")=4
	Set gtmtypes("block_info",17,"type")="int"
	Set gtmtypfldindx("block_info","ins_rec.blk_id")=17
	Set gtmtypes("block_info",18,"name")="block_info.blk_len"
	Set gtmtypes("block_info",18,"off")=112
	Set gtmtypes("block_info",18,"len")=4
	Set gtmtypes("block_info",18,"type")="int"
	Set gtmtypfldindx("block_info","blk_len")=18
	Set gtmtypes("block_info",19,"name")="block_info.blk_levl"
	Set gtmtypes("block_info",19,"off")=116
	Set gtmtypes("block_info",19,"len")=4
	Set gtmtypes("block_info",19,"type")="int"
	Set gtmtypfldindx("block_info","blk_levl")=19
	Set gtmtypes("block_info",20,"name")="block_info.ins_blk_id_p"
	Set gtmtypes("block_info",20,"off")=120
	Set gtmtypes("block_info",20,"len")=8
	Set gtmtypes("block_info",20,"type")="addr"
	Set gtmtypfldindx("block_info","ins_blk_id_p")=20
	;
	Set gtmtypes("bt_rec")="struct"
	Set gtmtypes("bt_rec",0)=13
	Set gtmtypes("bt_rec","len")=64
	Set gtmtypes("bt_rec",1,"name")="bt_rec.blkque"
	Set gtmtypes("bt_rec",1,"off")=0
	Set gtmtypes("bt_rec",1,"len")=16
	Set gtmtypes("bt_rec",1,"type")="struct"
	Set gtmtypfldindx("bt_rec","blkque")=1
	Set gtmtypes("bt_rec",2,"name")="bt_rec.blkque.fl"
	Set gtmtypes("bt_rec",2,"off")=0
	Set gtmtypes("bt_rec",2,"len")=8
	Set gtmtypes("bt_rec",2,"type")="intptr_t"
	Set gtmtypfldindx("bt_rec","blkque.fl")=2
	Set gtmtypes("bt_rec",3,"name")="bt_rec.blkque.bl"
	Set gtmtypes("bt_rec",3,"off")=8
	Set gtmtypes("bt_rec",3,"len")=8
	Set gtmtypes("bt_rec",3,"type")="intptr_t"
	Set gtmtypfldindx("bt_rec","blkque.bl")=3
	Set gtmtypes("bt_rec",4,"name")="bt_rec.tnque"
	Set gtmtypes("bt_rec",4,"off")=16
	Set gtmtypes("bt_rec",4,"len")=16
	Set gtmtypes("bt_rec",4,"type")="struct"
	Set gtmtypfldindx("bt_rec","tnque")=4
	Set gtmtypes("bt_rec",5,"name")="bt_rec.tnque.fl"
	Set gtmtypes("bt_rec",5,"off")=16
	Set gtmtypes("bt_rec",5,"len")=8
	Set gtmtypes("bt_rec",5,"type")="intptr_t"
	Set gtmtypfldindx("bt_rec","tnque.fl")=5
	Set gtmtypes("bt_rec",6,"name")="bt_rec.tnque.bl"
	Set gtmtypes("bt_rec",6,"off")=24
	Set gtmtypes("bt_rec",6,"len")=8
	Set gtmtypes("bt_rec",6,"type")="intptr_t"
	Set gtmtypfldindx("bt_rec","tnque.bl")=6
	Set gtmtypes("bt_rec",7,"name")="bt_rec.tn"
	Set gtmtypes("bt_rec",7,"off")=32
	Set gtmtypes("bt_rec",7,"len")=8
	Set gtmtypes("bt_rec",7,"type")="uint64_t"
	Set gtmtypfldindx("bt_rec","tn")=7
	Set gtmtypes("bt_rec",8,"name")="bt_rec.killtn"
	Set gtmtypes("bt_rec",8,"off")=40
	Set gtmtypes("bt_rec",8,"len")=8
	Set gtmtypes("bt_rec",8,"type")="uint64_t"
	Set gtmtypfldindx("bt_rec","killtn")=8
	Set gtmtypes("bt_rec",9,"name")="bt_rec.blk"
	Set gtmtypes("bt_rec",9,"off")=48
	Set gtmtypes("bt_rec",9,"len")=4
	Set gtmtypes("bt_rec",9,"type")="int"
	Set gtmtypfldindx("bt_rec","blk")=9
	Set gtmtypes("bt_rec",10,"name")="bt_rec.cache_index"
	Set gtmtypes("bt_rec",10,"off")=52
	Set gtmtypes("bt_rec",10,"len")=4
	Set gtmtypes("bt_rec",10,"type")="int"
	Set gtmtypfldindx("bt_rec","cache_index")=10
	Set gtmtypes("bt_rec",11,"name")="bt_rec.flushing"
	Set gtmtypes("bt_rec",11,"off")=56
	Set gtmtypes("bt_rec",11,"len")=1
	Set gtmtypes("bt_rec",11,"type")="char"
	Set gtmtypfldindx("bt_rec","flushing")=11
	Set gtmtypes("bt_rec",12,"name")="bt_rec.filler"
	Set gtmtypes("bt_rec",12,"off")=57
	Set gtmtypes("bt_rec",12,"len")=3
	Set gtmtypes("bt_rec",12,"type")="char"
	Set gtmtypfldindx("bt_rec","filler")=12
	Set gtmtypes("bt_rec",13,"name")="bt_rec.filler_int4"
	Set gtmtypes("bt_rec",13,"off")=60
	Set gtmtypes("bt_rec",13,"len")=4
	Set gtmtypes("bt_rec",13,"type")="int"
	Set gtmtypfldindx("bt_rec","filler_int4")=13
	;
	Set gtmtypes("buddy_list")="struct"
	Set gtmtypes("buddy_list",0)=9
	Set gtmtypes("buddy_list","len")=56
	Set gtmtypes("buddy_list",1,"name")="buddy_list.ptrArray"
	Set gtmtypes("buddy_list",1,"off")=0
	Set gtmtypes("buddy_list",1,"len")=8
	Set gtmtypes("buddy_list",1,"type")="addr"
	Set gtmtypfldindx("buddy_list","ptrArray")=1
	Set gtmtypes("buddy_list",2,"name")="buddy_list.elemSize"
	Set gtmtypes("buddy_list",2,"off")=8
	Set gtmtypes("buddy_list",2,"len")=4
	Set gtmtypes("buddy_list",2,"type")="int"
	Set gtmtypfldindx("buddy_list","elemSize")=2
	Set gtmtypes("buddy_list",3,"name")="buddy_list.initAlloc"
	Set gtmtypes("buddy_list",3,"off")=12
	Set gtmtypes("buddy_list",3,"len")=4
	Set gtmtypes("buddy_list",3,"type")="int"
	Set gtmtypfldindx("buddy_list","initAlloc")=3
	Set gtmtypes("buddy_list",4,"name")="buddy_list.initAllocBits"
	Set gtmtypes("buddy_list",4,"off")=16
	Set gtmtypes("buddy_list",4,"len")=4
	Set gtmtypes("buddy_list",4,"type")="int"
	Set gtmtypfldindx("buddy_list","initAllocBits")=4
	Set gtmtypes("buddy_list",5,"name")="buddy_list.nElems"
	Set gtmtypes("buddy_list",5,"off")=20
	Set gtmtypes("buddy_list",5,"len")=4
	Set gtmtypes("buddy_list",5,"type")="int"
	Set gtmtypfldindx("buddy_list","nElems")=5
	Set gtmtypes("buddy_list",6,"name")="buddy_list.cumulMaxElems"
	Set gtmtypes("buddy_list",6,"off")=24
	Set gtmtypes("buddy_list",6,"len")=4
	Set gtmtypes("buddy_list",6,"type")="int"
	Set gtmtypfldindx("buddy_list","cumulMaxElems")=6
	Set gtmtypes("buddy_list",7,"name")="buddy_list.ptrArrayCurr"
	Set gtmtypes("buddy_list",7,"off")=32
	Set gtmtypes("buddy_list",7,"len")=8
	Set gtmtypes("buddy_list",7,"type")="addr"
	Set gtmtypfldindx("buddy_list","ptrArrayCurr")=7
	Set gtmtypes("buddy_list",8,"name")="buddy_list.nextFreePtr"
	Set gtmtypes("buddy_list",8,"off")=40
	Set gtmtypes("buddy_list",8,"len")=8
	Set gtmtypes("buddy_list",8,"type")="addr"
	Set gtmtypfldindx("buddy_list","nextFreePtr")=8
	Set gtmtypes("buddy_list",9,"name")="buddy_list.free_que"
	Set gtmtypes("buddy_list",9,"off")=48
	Set gtmtypes("buddy_list",9,"len")=8
	Set gtmtypes("buddy_list",9,"type")="addr"
	Set gtmtypfldindx("buddy_list","free_que")=9
	;
	Set gtmtypes("cache_entry")="struct"
	Set gtmtypes("cache_entry",0)=9
	Set gtmtypes("cache_entry","len")=48
	Set gtmtypes("cache_entry",1,"name")="cache_entry.obj"
	Set gtmtypes("cache_entry",1,"off")=0
	Set gtmtypes("cache_entry",1,"len")=16
	Set gtmtypes("cache_entry",1,"type")="mstr"
	Set gtmtypfldindx("cache_entry","obj")=1
	Set gtmtypes("cache_entry",2,"name")="cache_entry.obj.char_len"
	Set gtmtypes("cache_entry",2,"off")=0
	Set gtmtypes("cache_entry",2,"len")=4
	Set gtmtypes("cache_entry",2,"type")="unsigned-int"
	Set gtmtypfldindx("cache_entry","obj.char_len")=2
	Set gtmtypes("cache_entry",3,"name")="cache_entry.obj.len"
	Set gtmtypes("cache_entry",3,"off")=4
	Set gtmtypes("cache_entry",3,"len")=4
	Set gtmtypes("cache_entry",3,"type")="int"
	Set gtmtypfldindx("cache_entry","obj.len")=3
	Set gtmtypes("cache_entry",4,"name")="cache_entry.obj.addr"
	Set gtmtypes("cache_entry",4,"off")=8
	Set gtmtypes("cache_entry",4,"len")=8
	Set gtmtypes("cache_entry",4,"type")="addr"
	Set gtmtypfldindx("cache_entry","obj.addr")=4
	Set gtmtypes("cache_entry",5,"name")="cache_entry.src"
	Set gtmtypes("cache_entry",5,"off")=16
	Set gtmtypes("cache_entry",5,"len")=24
	Set gtmtypes("cache_entry",5,"type")="icode_str"
	Set gtmtypfldindx("cache_entry","src")=5
	Set gtmtypes("cache_entry",6,"name")="cache_entry.src.str"
	Set gtmtypes("cache_entry",6,"off")=16
	Set gtmtypes("cache_entry",6,"len")=16
	Set gtmtypes("cache_entry",6,"type")="mstr"
	Set gtmtypfldindx("cache_entry","src.str")=6
	Set gtmtypes("cache_entry",7,"name")="cache_entry.src.code"
	Set gtmtypes("cache_entry",7,"off")=32
	Set gtmtypes("cache_entry",7,"len")=4
	Set gtmtypes("cache_entry",7,"type")="unsigned-int"
	Set gtmtypfldindx("cache_entry","src.code")=7
	Set gtmtypes("cache_entry",8,"name")="cache_entry.refcnt"
	Set gtmtypes("cache_entry",8,"off")=40
	Set gtmtypes("cache_entry",8,"len")=4
	Set gtmtypes("cache_entry",8,"type")="int"
	Set gtmtypfldindx("cache_entry","refcnt")=8
	Set gtmtypes("cache_entry",9,"name")="cache_entry.zb_refcnt"
	Set gtmtypes("cache_entry",9,"off")=44
	Set gtmtypes("cache_entry",9,"len")=4
	Set gtmtypes("cache_entry",9,"type")="int"
	Set gtmtypfldindx("cache_entry","zb_refcnt")=9
	;
	Set gtmtypes("cache_que_head")="struct"
	Set gtmtypes("cache_que_head",0)=8
	Set gtmtypes("cache_que_head","len")=24
	Set gtmtypes("cache_que_head",1,"name")="cache_que_head.fl"
	Set gtmtypes("cache_que_head",1,"off")=0
	Set gtmtypes("cache_que_head",1,"len")=8
	Set gtmtypes("cache_que_head",1,"type")="intptr_t"
	Set gtmtypfldindx("cache_que_head","fl")=1
	Set gtmtypes("cache_que_head",2,"name")="cache_que_head.bl"
	Set gtmtypes("cache_que_head",2,"off")=8
	Set gtmtypes("cache_que_head",2,"len")=8
	Set gtmtypes("cache_que_head",2,"type")="intptr_t"
	Set gtmtypfldindx("cache_que_head","bl")=2
	Set gtmtypes("cache_que_head",3,"name")="cache_que_head.latch"
	Set gtmtypes("cache_que_head",3,"off")=16
	Set gtmtypes("cache_que_head",3,"len")=8
	Set gtmtypes("cache_que_head",3,"type")="global_latch_t"
	Set gtmtypfldindx("cache_que_head","latch")=3
	Set gtmtypes("cache_que_head",4,"name")="cache_que_head.latch.u"
	Set gtmtypes("cache_que_head",4,"off")=16
	Set gtmtypes("cache_que_head",4,"len")=8
	Set gtmtypes("cache_que_head",4,"type")="union"
	Set gtmtypfldindx("cache_que_head","latch.u")=4
	Set gtmtypes("cache_que_head",5,"name")="cache_que_head.latch.u.pid_imgcnt"
	Set gtmtypes("cache_que_head",5,"off")=16
	Set gtmtypes("cache_que_head",5,"len")=8
	Set gtmtypes("cache_que_head",5,"type")="uint64_t"
	Set gtmtypfldindx("cache_que_head","latch.u.pid_imgcnt")=5
	Set gtmtypes("cache_que_head",6,"name")="cache_que_head.latch.u.parts"
	Set gtmtypes("cache_que_head",6,"off")=16
	Set gtmtypes("cache_que_head",6,"len")=8
	Set gtmtypes("cache_que_head",6,"type")="struct"
	Set gtmtypfldindx("cache_que_head","latch.u.parts")=6
	Set gtmtypes("cache_que_head",7,"name")="cache_que_head.latch.u.parts.latch_pid"
	Set gtmtypes("cache_que_head",7,"off")=16
	Set gtmtypes("cache_que_head",7,"len")=4
	Set gtmtypes("cache_que_head",7,"type")="int"
	Set gtmtypfldindx("cache_que_head","latch.u.parts.latch_pid")=7
	Set gtmtypes("cache_que_head",8,"name")="cache_que_head.latch.u.parts.latch_word"
	Set gtmtypes("cache_que_head",8,"off")=20
	Set gtmtypes("cache_que_head",8,"len")=4
	Set gtmtypes("cache_que_head",8,"type")="int"
	Set gtmtypfldindx("cache_que_head","latch.u.parts.latch_word")=8
	;
	Set gtmtypes("cache_que_heads")="struct"
	Set gtmtypes("cache_que_heads",0)=52
	Set gtmtypes("cache_que_heads","len")=264
	Set gtmtypes("cache_que_heads",1,"name")="cache_que_heads.cacheq_wip"
	Set gtmtypes("cache_que_heads",1,"off")=0
	Set gtmtypes("cache_que_heads",1,"len")=24
	Set gtmtypes("cache_que_heads",1,"type")="cache_que_head"
	Set gtmtypfldindx("cache_que_heads","cacheq_wip")=1
	Set gtmtypes("cache_que_heads",2,"name")="cache_que_heads.cacheq_wip.fl"
	Set gtmtypes("cache_que_heads",2,"off")=0
	Set gtmtypes("cache_que_heads",2,"len")=8
	Set gtmtypes("cache_que_heads",2,"type")="intptr_t"
	Set gtmtypfldindx("cache_que_heads","cacheq_wip.fl")=2
	Set gtmtypes("cache_que_heads",3,"name")="cache_que_heads.cacheq_wip.bl"
	Set gtmtypes("cache_que_heads",3,"off")=8
	Set gtmtypes("cache_que_heads",3,"len")=8
	Set gtmtypes("cache_que_heads",3,"type")="intptr_t"
	Set gtmtypfldindx("cache_que_heads","cacheq_wip.bl")=3
	Set gtmtypes("cache_que_heads",4,"name")="cache_que_heads.cacheq_wip.latch"
	Set gtmtypes("cache_que_heads",4,"off")=16
	Set gtmtypes("cache_que_heads",4,"len")=8
	Set gtmtypes("cache_que_heads",4,"type")="global_latch_t"
	Set gtmtypfldindx("cache_que_heads","cacheq_wip.latch")=4
	Set gtmtypes("cache_que_heads",5,"name")="cache_que_heads.cacheq_wip.latch.u"
	Set gtmtypes("cache_que_heads",5,"off")=16
	Set gtmtypes("cache_que_heads",5,"len")=8
	Set gtmtypes("cache_que_heads",5,"type")="union"
	Set gtmtypfldindx("cache_que_heads","cacheq_wip.latch.u")=5
	Set gtmtypes("cache_que_heads",6,"name")="cache_que_heads.cacheq_wip.latch.u.pid_imgcnt"
	Set gtmtypes("cache_que_heads",6,"off")=16
	Set gtmtypes("cache_que_heads",6,"len")=8
	Set gtmtypes("cache_que_heads",6,"type")="uint64_t"
	Set gtmtypfldindx("cache_que_heads","cacheq_wip.latch.u.pid_imgcnt")=6
	Set gtmtypes("cache_que_heads",7,"name")="cache_que_heads.cacheq_wip.latch.u.parts"
	Set gtmtypes("cache_que_heads",7,"off")=16
	Set gtmtypes("cache_que_heads",7,"len")=8
	Set gtmtypes("cache_que_heads",7,"type")="struct"
	Set gtmtypfldindx("cache_que_heads","cacheq_wip.latch.u.parts")=7
	Set gtmtypes("cache_que_heads",8,"name")="cache_que_heads.cacheq_wip.latch.u.parts.latch_pid"
	Set gtmtypes("cache_que_heads",8,"off")=16
	Set gtmtypes("cache_que_heads",8,"len")=4
	Set gtmtypes("cache_que_heads",8,"type")="int"
	Set gtmtypfldindx("cache_que_heads","cacheq_wip.latch.u.parts.latch_pid")=8
	Set gtmtypes("cache_que_heads",9,"name")="cache_que_heads.cacheq_wip.latch.u.parts.latch_word"
	Set gtmtypes("cache_que_heads",9,"off")=20
	Set gtmtypes("cache_que_heads",9,"len")=4
	Set gtmtypes("cache_que_heads",9,"type")="int"
	Set gtmtypfldindx("cache_que_heads","cacheq_wip.latch.u.parts.latch_word")=9
	Set gtmtypes("cache_que_heads",10,"name")="cache_que_heads.cacheq_active"
	Set gtmtypes("cache_que_heads",10,"off")=24
	Set gtmtypes("cache_que_heads",10,"len")=24
	Set gtmtypes("cache_que_heads",10,"type")="cache_que_head"
	Set gtmtypfldindx("cache_que_heads","cacheq_active")=10
	Set gtmtypes("cache_que_heads",11,"name")="cache_que_heads.cacheq_active.fl"
	Set gtmtypes("cache_que_heads",11,"off")=24
	Set gtmtypes("cache_que_heads",11,"len")=8
	Set gtmtypes("cache_que_heads",11,"type")="intptr_t"
	Set gtmtypfldindx("cache_que_heads","cacheq_active.fl")=11
	Set gtmtypes("cache_que_heads",12,"name")="cache_que_heads.cacheq_active.bl"
	Set gtmtypes("cache_que_heads",12,"off")=32
	Set gtmtypes("cache_que_heads",12,"len")=8
	Set gtmtypes("cache_que_heads",12,"type")="intptr_t"
	Set gtmtypfldindx("cache_que_heads","cacheq_active.bl")=12
	Set gtmtypes("cache_que_heads",13,"name")="cache_que_heads.cacheq_active.latch"
	Set gtmtypes("cache_que_heads",13,"off")=40
	Set gtmtypes("cache_que_heads",13,"len")=8
	Set gtmtypes("cache_que_heads",13,"type")="global_latch_t"
	Set gtmtypfldindx("cache_que_heads","cacheq_active.latch")=13
	Set gtmtypes("cache_que_heads",14,"name")="cache_que_heads.cacheq_active.latch.u"
	Set gtmtypes("cache_que_heads",14,"off")=40
	Set gtmtypes("cache_que_heads",14,"len")=8
	Set gtmtypes("cache_que_heads",14,"type")="union"
	Set gtmtypfldindx("cache_que_heads","cacheq_active.latch.u")=14
	Set gtmtypes("cache_que_heads",15,"name")="cache_que_heads.cacheq_active.latch.u.pid_imgcnt"
	Set gtmtypes("cache_que_heads",15,"off")=40
	Set gtmtypes("cache_que_heads",15,"len")=8
	Set gtmtypes("cache_que_heads",15,"type")="uint64_t"
	Set gtmtypfldindx("cache_que_heads","cacheq_active.latch.u.pid_imgcnt")=15
	Set gtmtypes("cache_que_heads",16,"name")="cache_que_heads.cacheq_active.latch.u.parts"
	Set gtmtypes("cache_que_heads",16,"off")=40
	Set gtmtypes("cache_que_heads",16,"len")=8
	Set gtmtypes("cache_que_heads",16,"type")="struct"
	Set gtmtypfldindx("cache_que_heads","cacheq_active.latch.u.parts")=16
	Set gtmtypes("cache_que_heads",17,"name")="cache_que_heads.cacheq_active.latch.u.parts.latch_pid"
	Set gtmtypes("cache_que_heads",17,"off")=40
	Set gtmtypes("cache_que_heads",17,"len")=4
	Set gtmtypes("cache_que_heads",17,"type")="int"
	Set gtmtypfldindx("cache_que_heads","cacheq_active.latch.u.parts.latch_pid")=17
	Set gtmtypes("cache_que_heads",18,"name")="cache_que_heads.cacheq_active.latch.u.parts.latch_word"
	Set gtmtypes("cache_que_heads",18,"off")=44
	Set gtmtypes("cache_que_heads",18,"len")=4
	Set gtmtypes("cache_que_heads",18,"type")="int"
	Set gtmtypfldindx("cache_que_heads","cacheq_active.latch.u.parts.latch_word")=18
	Set gtmtypes("cache_que_heads",19,"name")="cache_que_heads.cache_array"
	Set gtmtypes("cache_que_heads",19,"off")=48
	Set gtmtypes("cache_que_heads",19,"len")=216
	Set gtmtypes("cache_que_heads",19,"type")="cache_rec"
	Set gtmtypfldindx("cache_que_heads","cache_array")=19
	Set gtmtypes("cache_que_heads",20,"name")="cache_que_heads.cache_array[0].blkque"
	Set gtmtypes("cache_que_heads",20,"off")=48
	Set gtmtypes("cache_que_heads",20,"len")=16
	Set gtmtypes("cache_que_heads",20,"type")="struct"
	Set gtmtypfldindx("cache_que_heads","cache_array[0].blkque")=20
	Set gtmtypes("cache_que_heads",21,"name")="cache_que_heads.cache_array[0].blkque.fl"
	Set gtmtypes("cache_que_heads",21,"off")=48
	Set gtmtypes("cache_que_heads",21,"len")=8
	Set gtmtypes("cache_que_heads",21,"type")="intptr_t"
	Set gtmtypfldindx("cache_que_heads","cache_array[0].blkque.fl")=21
	Set gtmtypes("cache_que_heads",22,"name")="cache_que_heads.cache_array[0].blkque.bl"
	Set gtmtypes("cache_que_heads",22,"off")=56
	Set gtmtypes("cache_que_heads",22,"len")=8
	Set gtmtypes("cache_que_heads",22,"type")="intptr_t"
	Set gtmtypfldindx("cache_que_heads","cache_array[0].blkque.bl")=22
	Set gtmtypes("cache_que_heads",23,"name")="cache_que_heads.cache_array[0].state_que"
	Set gtmtypes("cache_que_heads",23,"off")=64
	Set gtmtypes("cache_que_heads",23,"len")=16
	Set gtmtypes("cache_que_heads",23,"type")="struct"
	Set gtmtypfldindx("cache_que_heads","cache_array[0].state_que")=23
	Set gtmtypes("cache_que_heads",24,"name")="cache_que_heads.cache_array[0].state_que.fl"
	Set gtmtypes("cache_que_heads",24,"off")=64
	Set gtmtypes("cache_que_heads",24,"len")=8
	Set gtmtypes("cache_que_heads",24,"type")="intptr_t"
	Set gtmtypfldindx("cache_que_heads","cache_array[0].state_que.fl")=24
	Set gtmtypes("cache_que_heads",25,"name")="cache_que_heads.cache_array[0].state_que.bl"
	Set gtmtypes("cache_que_heads",25,"off")=72
	Set gtmtypes("cache_que_heads",25,"len")=8
	Set gtmtypes("cache_que_heads",25,"type")="intptr_t"
	Set gtmtypfldindx("cache_que_heads","cache_array[0].state_que.bl")=25
	Set gtmtypes("cache_que_heads",26,"name")="cache_que_heads.cache_array[0].interlock"
	Set gtmtypes("cache_que_heads",26,"off")=80
	Set gtmtypes("cache_que_heads",26,"len")=4
	Set gtmtypes("cache_que_heads",26,"type")="union"
	Set gtmtypfldindx("cache_que_heads","cache_array[0].interlock")=26
	Set gtmtypes("cache_que_heads",27,"name")="cache_que_heads.cache_array[0].interlock.semaphore"
	Set gtmtypes("cache_que_heads",27,"off")=80
	Set gtmtypes("cache_que_heads",27,"len")=2
	Set gtmtypes("cache_que_heads",27,"type")="short"
	Set gtmtypfldindx("cache_que_heads","cache_array[0].interlock.semaphore")=27
	Set gtmtypes("cache_que_heads",28,"name")="cache_que_heads.cache_array[0].interlock.latch"
	Set gtmtypes("cache_que_heads",28,"off")=80
	Set gtmtypes("cache_que_heads",28,"len")=4
	Set gtmtypes("cache_que_heads",28,"type")="int"
	Set gtmtypfldindx("cache_que_heads","cache_array[0].interlock.latch")=28
	Set gtmtypes("cache_que_heads",29,"name")="cache_que_heads.cache_array[0].blk"
	Set gtmtypes("cache_que_heads",29,"off")=84
	Set gtmtypes("cache_que_heads",29,"len")=4
	Set gtmtypes("cache_que_heads",29,"type")="int"
	Set gtmtypfldindx("cache_que_heads","cache_array[0].blk")=29
	Set gtmtypes("cache_que_heads",30,"name")="cache_que_heads.cache_array[0].refer"
	Set gtmtypes("cache_que_heads",30,"off")=88
	Set gtmtypes("cache_que_heads",30,"len")=4
	Set gtmtypes("cache_que_heads",30,"type")="unsigned-int"
	Set gtmtypfldindx("cache_que_heads","cache_array[0].refer")=30
	Set gtmtypes("cache_que_heads",31,"name")="cache_que_heads.cache_array[0].ondsk_blkver"
	Set gtmtypes("cache_que_heads",31,"off")=92
	Set gtmtypes("cache_que_heads",31,"len")=4
	Set gtmtypes("cache_que_heads",31,"type")="int"
	Set gtmtypfldindx("cache_que_heads","cache_array[0].ondsk_blkver")=31
	Set gtmtypes("cache_que_heads",32,"name")="cache_que_heads.cache_array[0].dirty"
	Set gtmtypes("cache_que_heads",32,"off")=96
	Set gtmtypes("cache_que_heads",32,"len")=8
	Set gtmtypes("cache_que_heads",32,"type")="uint64_t"
	Set gtmtypfldindx("cache_que_heads","cache_array[0].dirty")=32
	Set gtmtypes("cache_que_heads",33,"name")="cache_que_heads.cache_array[0].flushed_dirty_tn"
	Set gtmtypes("cache_que_heads",33,"off")=104
	Set gtmtypes("cache_que_heads",33,"len")=8
	Set gtmtypes("cache_que_heads",33,"type")="uint64_t"
	Set gtmtypfldindx("cache_que_heads","cache_array[0].flushed_dirty_tn")=33
	Set gtmtypes("cache_que_heads",34,"name")="cache_que_heads.cache_array[0].tn"
	Set gtmtypes("cache_que_heads",34,"off")=112
	Set gtmtypes("cache_que_heads",34,"len")=8
	Set gtmtypes("cache_que_heads",34,"type")="uint64_t"
	Set gtmtypfldindx("cache_que_heads","cache_array[0].tn")=34
	Set gtmtypes("cache_que_heads",35,"name")="cache_que_heads.cache_array[0].bt_index"
	Set gtmtypes("cache_que_heads",35,"off")=120
	Set gtmtypes("cache_que_heads",35,"len")=8
	Set gtmtypes("cache_que_heads",35,"type")="intptr_t"
	Set gtmtypfldindx("cache_que_heads","cache_array[0].bt_index")=35
	Set gtmtypes("cache_que_heads",36,"name")="cache_que_heads.cache_array[0].buffaddr"
	Set gtmtypes("cache_que_heads",36,"off")=128
	Set gtmtypes("cache_que_heads",36,"len")=8
	Set gtmtypes("cache_que_heads",36,"type")="intptr_t"
	Set gtmtypfldindx("cache_que_heads","cache_array[0].buffaddr")=36
	Set gtmtypes("cache_que_heads",37,"name")="cache_que_heads.cache_array[0].twin"
	Set gtmtypes("cache_que_heads",37,"off")=136
	Set gtmtypes("cache_que_heads",37,"len")=8
	Set gtmtypes("cache_que_heads",37,"type")="intptr_t"
	Set gtmtypfldindx("cache_que_heads","cache_array[0].twin")=37
	Set gtmtypes("cache_que_heads",38,"name")="cache_que_heads.cache_array[0].jnl_addr"
	Set gtmtypes("cache_que_heads",38,"off")=144
	Set gtmtypes("cache_que_heads",38,"len")=4
	Set gtmtypes("cache_que_heads",38,"type")="unsigned-int"
	Set gtmtypfldindx("cache_que_heads","cache_array[0].jnl_addr")=38
	Set gtmtypes("cache_que_heads",39,"name")="cache_que_heads.cache_array[0].stopped"
	Set gtmtypes("cache_que_heads",39,"off")=148
	Set gtmtypes("cache_que_heads",39,"len")=4
	Set gtmtypes("cache_que_heads",39,"type")="boolean_t"
	Set gtmtypfldindx("cache_que_heads","cache_array[0].stopped")=39
	Set gtmtypes("cache_que_heads",40,"name")="cache_que_heads.cache_array[0].rip_latch"
	Set gtmtypes("cache_que_heads",40,"off")=152
	Set gtmtypes("cache_que_heads",40,"len")=8
	Set gtmtypes("cache_que_heads",40,"type")="global_latch_t"
	Set gtmtypfldindx("cache_que_heads","cache_array[0].rip_latch")=40
	Set gtmtypes("cache_que_heads",41,"name")="cache_que_heads.cache_array[0].data_invalid"
	Set gtmtypes("cache_que_heads",41,"off")=160
	Set gtmtypes("cache_que_heads",41,"len")=4
	Set gtmtypes("cache_que_heads",41,"type")="unsigned-int"
	Set gtmtypfldindx("cache_que_heads","cache_array[0].data_invalid")=41
	Set gtmtypes("cache_que_heads",42,"name")="cache_que_heads.cache_array[0].epid"
	Set gtmtypes("cache_que_heads",42,"off")=164
	Set gtmtypes("cache_que_heads",42,"len")=4
	Set gtmtypes("cache_que_heads",42,"type")="int"
	Set gtmtypfldindx("cache_que_heads","cache_array[0].epid")=42
	Set gtmtypes("cache_que_heads",43,"name")="cache_que_heads.cache_array[0].cycle"
	Set gtmtypes("cache_que_heads",43,"off")=168
	Set gtmtypes("cache_que_heads",43,"len")=4
	Set gtmtypes("cache_que_heads",43,"type")="int"
	Set gtmtypfldindx("cache_que_heads","cache_array[0].cycle")=43
	Set gtmtypes("cache_que_heads",44,"name")="cache_que_heads.cache_array[0].r_epid"
	Set gtmtypes("cache_que_heads",44,"off")=172
	Set gtmtypes("cache_que_heads",44,"len")=4
	Set gtmtypes("cache_que_heads",44,"type")="int"
	Set gtmtypfldindx("cache_que_heads","cache_array[0].r_epid")=44
	Set gtmtypes("cache_que_heads",45,"name")="cache_que_heads.cache_array[0].aiocb"
	Set gtmtypes("cache_que_heads",45,"off")=176
	Set gtmtypes("cache_que_heads",45,"len")=72
	Set gtmtypes("cache_que_heads",45,"type")="struct"
	Set gtmtypfldindx("cache_que_heads","cache_array[0].aiocb")=45
	Set gtmtypes("cache_que_heads",46,"name")="cache_que_heads.cache_array[0].read_in_progress"
	Set gtmtypes("cache_que_heads",46,"off")=248
	Set gtmtypes("cache_que_heads",46,"len")=4
	Set gtmtypes("cache_que_heads",46,"type")="int"
	Set gtmtypfldindx("cache_que_heads","cache_array[0].read_in_progress")=46
	Set gtmtypes("cache_que_heads",47,"name")="cache_que_heads.cache_array[0].in_tend"
	Set gtmtypes("cache_que_heads",47,"off")=252
	Set gtmtypes("cache_que_heads",47,"len")=4
	Set gtmtypes("cache_que_heads",47,"type")="unsigned-int"
	Set gtmtypfldindx("cache_que_heads","cache_array[0].in_tend")=47
	Set gtmtypes("cache_que_heads",48,"name")="cache_que_heads.cache_array[0].in_cw_set"
	Set gtmtypes("cache_que_heads",48,"off")=256
	Set gtmtypes("cache_que_heads",48,"len")=4
	Set gtmtypes("cache_que_heads",48,"type")="unsigned-int"
	Set gtmtypfldindx("cache_que_heads","cache_array[0].in_cw_set")=48
	Set gtmtypes("cache_que_heads",49,"name")="cache_que_heads.cache_array[0].wip_is_encr_buf"
	Set gtmtypes("cache_que_heads",49,"off")=260
	Set gtmtypes("cache_que_heads",49,"len")=1
	Set gtmtypes("cache_que_heads",49,"type")="char"
	Set gtmtypfldindx("cache_que_heads","cache_array[0].wip_is_encr_buf")=49
	Set gtmtypes("cache_que_heads",50,"name")="cache_que_heads.cache_array[0].backup_cr_is_twin"
	Set gtmtypes("cache_que_heads",50,"off")=261
	Set gtmtypes("cache_que_heads",50,"len")=1
	Set gtmtypes("cache_que_heads",50,"type")="char"
	Set gtmtypfldindx("cache_que_heads","cache_array[0].backup_cr_is_twin")=50
	Set gtmtypes("cache_que_heads",51,"name")="cache_que_heads.cache_array[0].aio_issued"
	Set gtmtypes("cache_que_heads",51,"off")=262
	Set gtmtypes("cache_que_heads",51,"len")=1
	Set gtmtypes("cache_que_heads",51,"type")="char"
	Set gtmtypfldindx("cache_que_heads","cache_array[0].aio_issued")=51
	Set gtmtypes("cache_que_heads",52,"name")="cache_que_heads.cache_array[0].needs_first_write"
	Set gtmtypes("cache_que_heads",52,"off")=263
	Set gtmtypes("cache_que_heads",52,"len")=1
	Set gtmtypes("cache_que_heads",52,"type")="char"
	Set gtmtypfldindx("cache_que_heads","cache_array[0].needs_first_write")=52
	;
	Set gtmtypes("cache_rec")="struct"
	Set gtmtypes("cache_rec",0)=38
	Set gtmtypes("cache_rec","len")=216
	Set gtmtypes("cache_rec",1,"name")="cache_rec.blkque"
	Set gtmtypes("cache_rec",1,"off")=0
	Set gtmtypes("cache_rec",1,"len")=16
	Set gtmtypes("cache_rec",1,"type")="struct"
	Set gtmtypfldindx("cache_rec","blkque")=1
	Set gtmtypes("cache_rec",2,"name")="cache_rec.blkque.fl"
	Set gtmtypes("cache_rec",2,"off")=0
	Set gtmtypes("cache_rec",2,"len")=8
	Set gtmtypes("cache_rec",2,"type")="intptr_t"
	Set gtmtypfldindx("cache_rec","blkque.fl")=2
	Set gtmtypes("cache_rec",3,"name")="cache_rec.blkque.bl"
	Set gtmtypes("cache_rec",3,"off")=8
	Set gtmtypes("cache_rec",3,"len")=8
	Set gtmtypes("cache_rec",3,"type")="intptr_t"
	Set gtmtypfldindx("cache_rec","blkque.bl")=3
	Set gtmtypes("cache_rec",4,"name")="cache_rec.state_que"
	Set gtmtypes("cache_rec",4,"off")=16
	Set gtmtypes("cache_rec",4,"len")=16
	Set gtmtypes("cache_rec",4,"type")="struct"
	Set gtmtypfldindx("cache_rec","state_que")=4
	Set gtmtypes("cache_rec",5,"name")="cache_rec.state_que.fl"
	Set gtmtypes("cache_rec",5,"off")=16
	Set gtmtypes("cache_rec",5,"len")=8
	Set gtmtypes("cache_rec",5,"type")="intptr_t"
	Set gtmtypfldindx("cache_rec","state_que.fl")=5
	Set gtmtypes("cache_rec",6,"name")="cache_rec.state_que.bl"
	Set gtmtypes("cache_rec",6,"off")=24
	Set gtmtypes("cache_rec",6,"len")=8
	Set gtmtypes("cache_rec",6,"type")="intptr_t"
	Set gtmtypfldindx("cache_rec","state_que.bl")=6
	Set gtmtypes("cache_rec",7,"name")="cache_rec.interlock"
	Set gtmtypes("cache_rec",7,"off")=32
	Set gtmtypes("cache_rec",7,"len")=4
	Set gtmtypes("cache_rec",7,"type")="union"
	Set gtmtypfldindx("cache_rec","interlock")=7
	Set gtmtypes("cache_rec",8,"name")="cache_rec.interlock.semaphore"
	Set gtmtypes("cache_rec",8,"off")=32
	Set gtmtypes("cache_rec",8,"len")=2
	Set gtmtypes("cache_rec",8,"type")="short"
	Set gtmtypfldindx("cache_rec","interlock.semaphore")=8
	Set gtmtypes("cache_rec",9,"name")="cache_rec.interlock.latch"
	Set gtmtypes("cache_rec",9,"off")=32
	Set gtmtypes("cache_rec",9,"len")=4
	Set gtmtypes("cache_rec",9,"type")="int"
	Set gtmtypfldindx("cache_rec","interlock.latch")=9
	Set gtmtypes("cache_rec",10,"name")="cache_rec.blk"
	Set gtmtypes("cache_rec",10,"off")=36
	Set gtmtypes("cache_rec",10,"len")=4
	Set gtmtypes("cache_rec",10,"type")="int"
	Set gtmtypfldindx("cache_rec","blk")=10
	Set gtmtypes("cache_rec",11,"name")="cache_rec.refer"
	Set gtmtypes("cache_rec",11,"off")=40
	Set gtmtypes("cache_rec",11,"len")=4
	Set gtmtypes("cache_rec",11,"type")="unsigned-int"
	Set gtmtypfldindx("cache_rec","refer")=11
	Set gtmtypes("cache_rec",12,"name")="cache_rec.ondsk_blkver"
	Set gtmtypes("cache_rec",12,"off")=44
	Set gtmtypes("cache_rec",12,"len")=4
	Set gtmtypes("cache_rec",12,"type")="int"
	Set gtmtypfldindx("cache_rec","ondsk_blkver")=12
	Set gtmtypes("cache_rec",13,"name")="cache_rec.dirty"
	Set gtmtypes("cache_rec",13,"off")=48
	Set gtmtypes("cache_rec",13,"len")=8
	Set gtmtypes("cache_rec",13,"type")="uint64_t"
	Set gtmtypfldindx("cache_rec","dirty")=13
	Set gtmtypes("cache_rec",14,"name")="cache_rec.flushed_dirty_tn"
	Set gtmtypes("cache_rec",14,"off")=56
	Set gtmtypes("cache_rec",14,"len")=8
	Set gtmtypes("cache_rec",14,"type")="uint64_t"
	Set gtmtypfldindx("cache_rec","flushed_dirty_tn")=14
	Set gtmtypes("cache_rec",15,"name")="cache_rec.tn"
	Set gtmtypes("cache_rec",15,"off")=64
	Set gtmtypes("cache_rec",15,"len")=8
	Set gtmtypes("cache_rec",15,"type")="uint64_t"
	Set gtmtypfldindx("cache_rec","tn")=15
	Set gtmtypes("cache_rec",16,"name")="cache_rec.bt_index"
	Set gtmtypes("cache_rec",16,"off")=72
	Set gtmtypes("cache_rec",16,"len")=8
	Set gtmtypes("cache_rec",16,"type")="intptr_t"
	Set gtmtypfldindx("cache_rec","bt_index")=16
	Set gtmtypes("cache_rec",17,"name")="cache_rec.buffaddr"
	Set gtmtypes("cache_rec",17,"off")=80
	Set gtmtypes("cache_rec",17,"len")=8
	Set gtmtypes("cache_rec",17,"type")="intptr_t"
	Set gtmtypfldindx("cache_rec","buffaddr")=17
	Set gtmtypes("cache_rec",18,"name")="cache_rec.twin"
	Set gtmtypes("cache_rec",18,"off")=88
	Set gtmtypes("cache_rec",18,"len")=8
	Set gtmtypes("cache_rec",18,"type")="intptr_t"
	Set gtmtypfldindx("cache_rec","twin")=18
	Set gtmtypes("cache_rec",19,"name")="cache_rec.jnl_addr"
	Set gtmtypes("cache_rec",19,"off")=96
	Set gtmtypes("cache_rec",19,"len")=4
	Set gtmtypes("cache_rec",19,"type")="unsigned-int"
	Set gtmtypfldindx("cache_rec","jnl_addr")=19
	Set gtmtypes("cache_rec",20,"name")="cache_rec.stopped"
	Set gtmtypes("cache_rec",20,"off")=100
	Set gtmtypes("cache_rec",20,"len")=4
	Set gtmtypes("cache_rec",20,"type")="boolean_t"
	Set gtmtypfldindx("cache_rec","stopped")=20
	Set gtmtypes("cache_rec",21,"name")="cache_rec.rip_latch"
	Set gtmtypes("cache_rec",21,"off")=104
	Set gtmtypes("cache_rec",21,"len")=8
	Set gtmtypes("cache_rec",21,"type")="global_latch_t"
	Set gtmtypfldindx("cache_rec","rip_latch")=21
	Set gtmtypes("cache_rec",22,"name")="cache_rec.rip_latch.u"
	Set gtmtypes("cache_rec",22,"off")=104
	Set gtmtypes("cache_rec",22,"len")=8
	Set gtmtypes("cache_rec",22,"type")="union"
	Set gtmtypfldindx("cache_rec","rip_latch.u")=22
	Set gtmtypes("cache_rec",23,"name")="cache_rec.rip_latch.u.pid_imgcnt"
	Set gtmtypes("cache_rec",23,"off")=104
	Set gtmtypes("cache_rec",23,"len")=8
	Set gtmtypes("cache_rec",23,"type")="uint64_t"
	Set gtmtypfldindx("cache_rec","rip_latch.u.pid_imgcnt")=23
	Set gtmtypes("cache_rec",24,"name")="cache_rec.rip_latch.u.parts"
	Set gtmtypes("cache_rec",24,"off")=104
	Set gtmtypes("cache_rec",24,"len")=8
	Set gtmtypes("cache_rec",24,"type")="struct"
	Set gtmtypfldindx("cache_rec","rip_latch.u.parts")=24
	Set gtmtypes("cache_rec",25,"name")="cache_rec.rip_latch.u.parts.latch_pid"
	Set gtmtypes("cache_rec",25,"off")=104
	Set gtmtypes("cache_rec",25,"len")=4
	Set gtmtypes("cache_rec",25,"type")="int"
	Set gtmtypfldindx("cache_rec","rip_latch.u.parts.latch_pid")=25
	Set gtmtypes("cache_rec",26,"name")="cache_rec.rip_latch.u.parts.latch_word"
	Set gtmtypes("cache_rec",26,"off")=108
	Set gtmtypes("cache_rec",26,"len")=4
	Set gtmtypes("cache_rec",26,"type")="int"
	Set gtmtypfldindx("cache_rec","rip_latch.u.parts.latch_word")=26
	Set gtmtypes("cache_rec",27,"name")="cache_rec.data_invalid"
	Set gtmtypes("cache_rec",27,"off")=112
	Set gtmtypes("cache_rec",27,"len")=4
	Set gtmtypes("cache_rec",27,"type")="unsigned-int"
	Set gtmtypfldindx("cache_rec","data_invalid")=27
	Set gtmtypes("cache_rec",28,"name")="cache_rec.epid"
	Set gtmtypes("cache_rec",28,"off")=116
	Set gtmtypes("cache_rec",28,"len")=4
	Set gtmtypes("cache_rec",28,"type")="int"
	Set gtmtypfldindx("cache_rec","epid")=28
	Set gtmtypes("cache_rec",29,"name")="cache_rec.cycle"
	Set gtmtypes("cache_rec",29,"off")=120
	Set gtmtypes("cache_rec",29,"len")=4
	Set gtmtypes("cache_rec",29,"type")="int"
	Set gtmtypfldindx("cache_rec","cycle")=29
	Set gtmtypes("cache_rec",30,"name")="cache_rec.r_epid"
	Set gtmtypes("cache_rec",30,"off")=124
	Set gtmtypes("cache_rec",30,"len")=4
	Set gtmtypes("cache_rec",30,"type")="int"
	Set gtmtypfldindx("cache_rec","r_epid")=30
	Set gtmtypes("cache_rec",31,"name")="cache_rec.aiocb"
	Set gtmtypes("cache_rec",31,"off")=128
	Set gtmtypes("cache_rec",31,"len")=72
	Set gtmtypes("cache_rec",31,"type")="struct"
	Set gtmtypfldindx("cache_rec","aiocb")=31
	Set gtmtypes("cache_rec",32,"name")="cache_rec.read_in_progress"
	Set gtmtypes("cache_rec",32,"off")=200
	Set gtmtypes("cache_rec",32,"len")=4
	Set gtmtypes("cache_rec",32,"type")="int"
	Set gtmtypfldindx("cache_rec","read_in_progress")=32
	Set gtmtypes("cache_rec",33,"name")="cache_rec.in_tend"
	Set gtmtypes("cache_rec",33,"off")=204
	Set gtmtypes("cache_rec",33,"len")=4
	Set gtmtypes("cache_rec",33,"type")="unsigned-int"
	Set gtmtypfldindx("cache_rec","in_tend")=33
	Set gtmtypes("cache_rec",34,"name")="cache_rec.in_cw_set"
	Set gtmtypes("cache_rec",34,"off")=208
	Set gtmtypes("cache_rec",34,"len")=4
	Set gtmtypes("cache_rec",34,"type")="unsigned-int"
	Set gtmtypfldindx("cache_rec","in_cw_set")=34
	Set gtmtypes("cache_rec",35,"name")="cache_rec.wip_is_encr_buf"
	Set gtmtypes("cache_rec",35,"off")=212
	Set gtmtypes("cache_rec",35,"len")=1
	Set gtmtypes("cache_rec",35,"type")="char"
	Set gtmtypfldindx("cache_rec","wip_is_encr_buf")=35
	Set gtmtypes("cache_rec",36,"name")="cache_rec.backup_cr_is_twin"
	Set gtmtypes("cache_rec",36,"off")=213
	Set gtmtypes("cache_rec",36,"len")=1
	Set gtmtypes("cache_rec",36,"type")="char"
	Set gtmtypfldindx("cache_rec","backup_cr_is_twin")=36
	Set gtmtypes("cache_rec",37,"name")="cache_rec.aio_issued"
	Set gtmtypes("cache_rec",37,"off")=214
	Set gtmtypes("cache_rec",37,"len")=1
	Set gtmtypes("cache_rec",37,"type")="char"
	Set gtmtypfldindx("cache_rec","aio_issued")=37
	Set gtmtypes("cache_rec",38,"name")="cache_rec.needs_first_write"
	Set gtmtypes("cache_rec",38,"off")=215
	Set gtmtypes("cache_rec",38,"len")=1
	Set gtmtypes("cache_rec",38,"type")="char"
	Set gtmtypfldindx("cache_rec","needs_first_write")=38
	;
	Set gtmtypes("cache_state_rec")="struct"
	Set gtmtypes("cache_state_rec",0)=35
	Set gtmtypes("cache_state_rec","len")=200
	Set gtmtypes("cache_state_rec",1,"name")="cache_state_rec.state_que"
	Set gtmtypes("cache_state_rec",1,"off")=0
	Set gtmtypes("cache_state_rec",1,"len")=16
	Set gtmtypes("cache_state_rec",1,"type")="struct"
	Set gtmtypfldindx("cache_state_rec","state_que")=1
	Set gtmtypes("cache_state_rec",2,"name")="cache_state_rec.state_que.fl"
	Set gtmtypes("cache_state_rec",2,"off")=0
	Set gtmtypes("cache_state_rec",2,"len")=8
	Set gtmtypes("cache_state_rec",2,"type")="intptr_t"
	Set gtmtypfldindx("cache_state_rec","state_que.fl")=2
	Set gtmtypes("cache_state_rec",3,"name")="cache_state_rec.state_que.bl"
	Set gtmtypes("cache_state_rec",3,"off")=8
	Set gtmtypes("cache_state_rec",3,"len")=8
	Set gtmtypes("cache_state_rec",3,"type")="intptr_t"
	Set gtmtypfldindx("cache_state_rec","state_que.bl")=3
	Set gtmtypes("cache_state_rec",4,"name")="cache_state_rec.interlock"
	Set gtmtypes("cache_state_rec",4,"off")=16
	Set gtmtypes("cache_state_rec",4,"len")=4
	Set gtmtypes("cache_state_rec",4,"type")="union"
	Set gtmtypfldindx("cache_state_rec","interlock")=4
	Set gtmtypes("cache_state_rec",5,"name")="cache_state_rec.interlock.semaphore"
	Set gtmtypes("cache_state_rec",5,"off")=16
	Set gtmtypes("cache_state_rec",5,"len")=2
	Set gtmtypes("cache_state_rec",5,"type")="short"
	Set gtmtypfldindx("cache_state_rec","interlock.semaphore")=5
	Set gtmtypes("cache_state_rec",6,"name")="cache_state_rec.interlock.latch"
	Set gtmtypes("cache_state_rec",6,"off")=16
	Set gtmtypes("cache_state_rec",6,"len")=4
	Set gtmtypes("cache_state_rec",6,"type")="int"
	Set gtmtypfldindx("cache_state_rec","interlock.latch")=6
	Set gtmtypes("cache_state_rec",7,"name")="cache_state_rec.blk"
	Set gtmtypes("cache_state_rec",7,"off")=20
	Set gtmtypes("cache_state_rec",7,"len")=4
	Set gtmtypes("cache_state_rec",7,"type")="int"
	Set gtmtypfldindx("cache_state_rec","blk")=7
	Set gtmtypes("cache_state_rec",8,"name")="cache_state_rec.refer"
	Set gtmtypes("cache_state_rec",8,"off")=24
	Set gtmtypes("cache_state_rec",8,"len")=4
	Set gtmtypes("cache_state_rec",8,"type")="unsigned-int"
	Set gtmtypfldindx("cache_state_rec","refer")=8
	Set gtmtypes("cache_state_rec",9,"name")="cache_state_rec.ondsk_blkver"
	Set gtmtypes("cache_state_rec",9,"off")=28
	Set gtmtypes("cache_state_rec",9,"len")=4
	Set gtmtypes("cache_state_rec",9,"type")="int"
	Set gtmtypfldindx("cache_state_rec","ondsk_blkver")=9
	Set gtmtypes("cache_state_rec",10,"name")="cache_state_rec.dirty"
	Set gtmtypes("cache_state_rec",10,"off")=32
	Set gtmtypes("cache_state_rec",10,"len")=8
	Set gtmtypes("cache_state_rec",10,"type")="uint64_t"
	Set gtmtypfldindx("cache_state_rec","dirty")=10
	Set gtmtypes("cache_state_rec",11,"name")="cache_state_rec.flushed_dirty_tn"
	Set gtmtypes("cache_state_rec",11,"off")=40
	Set gtmtypes("cache_state_rec",11,"len")=8
	Set gtmtypes("cache_state_rec",11,"type")="uint64_t"
	Set gtmtypfldindx("cache_state_rec","flushed_dirty_tn")=11
	Set gtmtypes("cache_state_rec",12,"name")="cache_state_rec.tn"
	Set gtmtypes("cache_state_rec",12,"off")=48
	Set gtmtypes("cache_state_rec",12,"len")=8
	Set gtmtypes("cache_state_rec",12,"type")="uint64_t"
	Set gtmtypfldindx("cache_state_rec","tn")=12
	Set gtmtypes("cache_state_rec",13,"name")="cache_state_rec.bt_index"
	Set gtmtypes("cache_state_rec",13,"off")=56
	Set gtmtypes("cache_state_rec",13,"len")=8
	Set gtmtypes("cache_state_rec",13,"type")="intptr_t"
	Set gtmtypfldindx("cache_state_rec","bt_index")=13
	Set gtmtypes("cache_state_rec",14,"name")="cache_state_rec.buffaddr"
	Set gtmtypes("cache_state_rec",14,"off")=64
	Set gtmtypes("cache_state_rec",14,"len")=8
	Set gtmtypes("cache_state_rec",14,"type")="intptr_t"
	Set gtmtypfldindx("cache_state_rec","buffaddr")=14
	Set gtmtypes("cache_state_rec",15,"name")="cache_state_rec.twin"
	Set gtmtypes("cache_state_rec",15,"off")=72
	Set gtmtypes("cache_state_rec",15,"len")=8
	Set gtmtypes("cache_state_rec",15,"type")="intptr_t"
	Set gtmtypfldindx("cache_state_rec","twin")=15
	Set gtmtypes("cache_state_rec",16,"name")="cache_state_rec.jnl_addr"
	Set gtmtypes("cache_state_rec",16,"off")=80
	Set gtmtypes("cache_state_rec",16,"len")=4
	Set gtmtypes("cache_state_rec",16,"type")="unsigned-int"
	Set gtmtypfldindx("cache_state_rec","jnl_addr")=16
	Set gtmtypes("cache_state_rec",17,"name")="cache_state_rec.stopped"
	Set gtmtypes("cache_state_rec",17,"off")=84
	Set gtmtypes("cache_state_rec",17,"len")=4
	Set gtmtypes("cache_state_rec",17,"type")="boolean_t"
	Set gtmtypfldindx("cache_state_rec","stopped")=17
	Set gtmtypes("cache_state_rec",18,"name")="cache_state_rec.rip_latch"
	Set gtmtypes("cache_state_rec",18,"off")=88
	Set gtmtypes("cache_state_rec",18,"len")=8
	Set gtmtypes("cache_state_rec",18,"type")="global_latch_t"
	Set gtmtypfldindx("cache_state_rec","rip_latch")=18
	Set gtmtypes("cache_state_rec",19,"name")="cache_state_rec.rip_latch.u"
	Set gtmtypes("cache_state_rec",19,"off")=88
	Set gtmtypes("cache_state_rec",19,"len")=8
	Set gtmtypes("cache_state_rec",19,"type")="union"
	Set gtmtypfldindx("cache_state_rec","rip_latch.u")=19
	Set gtmtypes("cache_state_rec",20,"name")="cache_state_rec.rip_latch.u.pid_imgcnt"
	Set gtmtypes("cache_state_rec",20,"off")=88
	Set gtmtypes("cache_state_rec",20,"len")=8
	Set gtmtypes("cache_state_rec",20,"type")="uint64_t"
	Set gtmtypfldindx("cache_state_rec","rip_latch.u.pid_imgcnt")=20
	Set gtmtypes("cache_state_rec",21,"name")="cache_state_rec.rip_latch.u.parts"
	Set gtmtypes("cache_state_rec",21,"off")=88
	Set gtmtypes("cache_state_rec",21,"len")=8
	Set gtmtypes("cache_state_rec",21,"type")="struct"
	Set gtmtypfldindx("cache_state_rec","rip_latch.u.parts")=21
	Set gtmtypes("cache_state_rec",22,"name")="cache_state_rec.rip_latch.u.parts.latch_pid"
	Set gtmtypes("cache_state_rec",22,"off")=88
	Set gtmtypes("cache_state_rec",22,"len")=4
	Set gtmtypes("cache_state_rec",22,"type")="int"
	Set gtmtypfldindx("cache_state_rec","rip_latch.u.parts.latch_pid")=22
	Set gtmtypes("cache_state_rec",23,"name")="cache_state_rec.rip_latch.u.parts.latch_word"
	Set gtmtypes("cache_state_rec",23,"off")=92
	Set gtmtypes("cache_state_rec",23,"len")=4
	Set gtmtypes("cache_state_rec",23,"type")="int"
	Set gtmtypfldindx("cache_state_rec","rip_latch.u.parts.latch_word")=23
	Set gtmtypes("cache_state_rec",24,"name")="cache_state_rec.data_invalid"
	Set gtmtypes("cache_state_rec",24,"off")=96
	Set gtmtypes("cache_state_rec",24,"len")=4
	Set gtmtypes("cache_state_rec",24,"type")="unsigned-int"
	Set gtmtypfldindx("cache_state_rec","data_invalid")=24
	Set gtmtypes("cache_state_rec",25,"name")="cache_state_rec.epid"
	Set gtmtypes("cache_state_rec",25,"off")=100
	Set gtmtypes("cache_state_rec",25,"len")=4
	Set gtmtypes("cache_state_rec",25,"type")="int"
	Set gtmtypfldindx("cache_state_rec","epid")=25
	Set gtmtypes("cache_state_rec",26,"name")="cache_state_rec.cycle"
	Set gtmtypes("cache_state_rec",26,"off")=104
	Set gtmtypes("cache_state_rec",26,"len")=4
	Set gtmtypes("cache_state_rec",26,"type")="int"
	Set gtmtypfldindx("cache_state_rec","cycle")=26
	Set gtmtypes("cache_state_rec",27,"name")="cache_state_rec.r_epid"
	Set gtmtypes("cache_state_rec",27,"off")=108
	Set gtmtypes("cache_state_rec",27,"len")=4
	Set gtmtypes("cache_state_rec",27,"type")="int"
	Set gtmtypfldindx("cache_state_rec","r_epid")=27
	Set gtmtypes("cache_state_rec",28,"name")="cache_state_rec.aiocb"
	Set gtmtypes("cache_state_rec",28,"off")=112
	Set gtmtypes("cache_state_rec",28,"len")=72
	Set gtmtypes("cache_state_rec",28,"type")="struct"
	Set gtmtypfldindx("cache_state_rec","aiocb")=28
	Set gtmtypes("cache_state_rec",29,"name")="cache_state_rec.read_in_progress"
	Set gtmtypes("cache_state_rec",29,"off")=184
	Set gtmtypes("cache_state_rec",29,"len")=4
	Set gtmtypes("cache_state_rec",29,"type")="int"
	Set gtmtypfldindx("cache_state_rec","read_in_progress")=29
	Set gtmtypes("cache_state_rec",30,"name")="cache_state_rec.in_tend"
	Set gtmtypes("cache_state_rec",30,"off")=188
	Set gtmtypes("cache_state_rec",30,"len")=4
	Set gtmtypes("cache_state_rec",30,"type")="unsigned-int"
	Set gtmtypfldindx("cache_state_rec","in_tend")=30
	Set gtmtypes("cache_state_rec",31,"name")="cache_state_rec.in_cw_set"
	Set gtmtypes("cache_state_rec",31,"off")=192
	Set gtmtypes("cache_state_rec",31,"len")=4
	Set gtmtypes("cache_state_rec",31,"type")="unsigned-int"
	Set gtmtypfldindx("cache_state_rec","in_cw_set")=31
	Set gtmtypes("cache_state_rec",32,"name")="cache_state_rec.wip_is_encr_buf"
	Set gtmtypes("cache_state_rec",32,"off")=196
	Set gtmtypes("cache_state_rec",32,"len")=1
	Set gtmtypes("cache_state_rec",32,"type")="char"
	Set gtmtypfldindx("cache_state_rec","wip_is_encr_buf")=32
	Set gtmtypes("cache_state_rec",33,"name")="cache_state_rec.backup_cr_is_twin"
	Set gtmtypes("cache_state_rec",33,"off")=197
	Set gtmtypes("cache_state_rec",33,"len")=1
	Set gtmtypes("cache_state_rec",33,"type")="char"
	Set gtmtypfldindx("cache_state_rec","backup_cr_is_twin")=33
	Set gtmtypes("cache_state_rec",34,"name")="cache_state_rec.aio_issued"
	Set gtmtypes("cache_state_rec",34,"off")=198
	Set gtmtypes("cache_state_rec",34,"len")=1
	Set gtmtypes("cache_state_rec",34,"type")="char"
	Set gtmtypfldindx("cache_state_rec","aio_issued")=34
	Set gtmtypes("cache_state_rec",35,"name")="cache_state_rec.needs_first_write"
	Set gtmtypes("cache_state_rec",35,"off")=199
	Set gtmtypes("cache_state_rec",35,"len")=1
	Set gtmtypes("cache_state_rec",35,"type")="char"
	Set gtmtypfldindx("cache_state_rec","needs_first_write")=35
	;
	Set gtmtypes("callin_entry_list")="struct"
	Set gtmtypes("callin_entry_list",0)=14
	Set gtmtypes("callin_entry_list","len")=64
	Set gtmtypes("callin_entry_list",1,"name")="callin_entry_list.label_ref"
	Set gtmtypes("callin_entry_list",1,"off")=0
	Set gtmtypes("callin_entry_list",1,"len")=16
	Set gtmtypes("callin_entry_list",1,"type")="mstr"
	Set gtmtypfldindx("callin_entry_list","label_ref")=1
	Set gtmtypes("callin_entry_list",2,"name")="callin_entry_list.label_ref.char_len"
	Set gtmtypes("callin_entry_list",2,"off")=0
	Set gtmtypes("callin_entry_list",2,"len")=4
	Set gtmtypes("callin_entry_list",2,"type")="unsigned-int"
	Set gtmtypfldindx("callin_entry_list","label_ref.char_len")=2
	Set gtmtypes("callin_entry_list",3,"name")="callin_entry_list.label_ref.len"
	Set gtmtypes("callin_entry_list",3,"off")=4
	Set gtmtypes("callin_entry_list",3,"len")=4
	Set gtmtypes("callin_entry_list",3,"type")="int"
	Set gtmtypfldindx("callin_entry_list","label_ref.len")=3
	Set gtmtypes("callin_entry_list",4,"name")="callin_entry_list.label_ref.addr"
	Set gtmtypes("callin_entry_list",4,"off")=8
	Set gtmtypes("callin_entry_list",4,"len")=8
	Set gtmtypes("callin_entry_list",4,"type")="addr"
	Set gtmtypfldindx("callin_entry_list","label_ref.addr")=4
	Set gtmtypes("callin_entry_list",5,"name")="callin_entry_list.call_name"
	Set gtmtypes("callin_entry_list",5,"off")=16
	Set gtmtypes("callin_entry_list",5,"len")=16
	Set gtmtypes("callin_entry_list",5,"type")="mstr"
	Set gtmtypfldindx("callin_entry_list","call_name")=5
	Set gtmtypes("callin_entry_list",6,"name")="callin_entry_list.call_name.char_len"
	Set gtmtypes("callin_entry_list",6,"off")=16
	Set gtmtypes("callin_entry_list",6,"len")=4
	Set gtmtypes("callin_entry_list",6,"type")="unsigned-int"
	Set gtmtypfldindx("callin_entry_list","call_name.char_len")=6
	Set gtmtypes("callin_entry_list",7,"name")="callin_entry_list.call_name.len"
	Set gtmtypes("callin_entry_list",7,"off")=20
	Set gtmtypes("callin_entry_list",7,"len")=4
	Set gtmtypes("callin_entry_list",7,"type")="int"
	Set gtmtypfldindx("callin_entry_list","call_name.len")=7
	Set gtmtypes("callin_entry_list",8,"name")="callin_entry_list.call_name.addr"
	Set gtmtypes("callin_entry_list",8,"off")=24
	Set gtmtypes("callin_entry_list",8,"len")=8
	Set gtmtypes("callin_entry_list",8,"type")="addr"
	Set gtmtypfldindx("callin_entry_list","call_name.addr")=8
	Set gtmtypes("callin_entry_list",9,"name")="callin_entry_list.input_mask"
	Set gtmtypes("callin_entry_list",9,"off")=32
	Set gtmtypes("callin_entry_list",9,"len")=4
	Set gtmtypes("callin_entry_list",9,"type")="unsigned-int"
	Set gtmtypfldindx("callin_entry_list","input_mask")=9
	Set gtmtypes("callin_entry_list",10,"name")="callin_entry_list.output_mask"
	Set gtmtypes("callin_entry_list",10,"off")=36
	Set gtmtypes("callin_entry_list",10,"len")=4
	Set gtmtypes("callin_entry_list",10,"type")="unsigned-int"
	Set gtmtypfldindx("callin_entry_list","output_mask")=10
	Set gtmtypes("callin_entry_list",11,"name")="callin_entry_list.argcnt"
	Set gtmtypes("callin_entry_list",11,"off")=40
	Set gtmtypes("callin_entry_list",11,"len")=2
	Set gtmtypes("callin_entry_list",11,"type")="unsigned-short"
	Set gtmtypfldindx("callin_entry_list","argcnt")=11
	Set gtmtypes("callin_entry_list",12,"name")="callin_entry_list.return_type"
	Set gtmtypes("callin_entry_list",12,"off")=44
	Set gtmtypes("callin_entry_list",12,"len")=4
	Set gtmtypes("callin_entry_list",12,"type")="int"
	Set gtmtypfldindx("callin_entry_list","return_type")=12
	Set gtmtypes("callin_entry_list",13,"name")="callin_entry_list.parms"
	Set gtmtypes("callin_entry_list",13,"off")=48
	Set gtmtypes("callin_entry_list",13,"len")=8
	Set gtmtypes("callin_entry_list",13,"type")="addr"
	Set gtmtypfldindx("callin_entry_list","parms")=13
	Set gtmtypes("callin_entry_list",14,"name")="callin_entry_list.next_entry"
	Set gtmtypes("callin_entry_list",14,"off")=56
	Set gtmtypes("callin_entry_list",14,"len")=8
	Set gtmtypes("callin_entry_list",14,"type")="addr"
	Set gtmtypfldindx("callin_entry_list","next_entry")=14
	;
	Set gtmtypes("casemap_t")="struct"
	Set gtmtypes("casemap_t",0)=3
	Set gtmtypes("casemap_t","len")=24
	Set gtmtypes("casemap_t",1,"name")="casemap_t.code"
	Set gtmtypes("casemap_t",1,"off")=0
	Set gtmtypes("casemap_t",1,"len")=8
	Set gtmtypes("casemap_t",1,"type")="addr"
	Set gtmtypfldindx("casemap_t","code")=1
	Set gtmtypes("casemap_t",2,"name")="casemap_t.m"
	Set gtmtypes("casemap_t",2,"off")=8
	Set gtmtypes("casemap_t",2,"len")=8
	Set gtmtypes("casemap_t",2,"type")="void"
	Set gtmtypfldindx("casemap_t","m")=2
	Set gtmtypes("casemap_t",3,"name")="casemap_t.u"
	Set gtmtypes("casemap_t",3,"off")=16
	Set gtmtypes("casemap_t",3,"len")=8
	Set gtmtypes("casemap_t",3,"type")="int32_t"
	Set gtmtypfldindx("casemap_t","u")=3
	;
	Set gtmtypes("ci_name_descriptor")="struct"
	Set gtmtypes("ci_name_descriptor",0)=2
	Set gtmtypes("ci_name_descriptor","len")=24
	Set gtmtypes("ci_name_descriptor",1,"name")="ci_name_descriptor.rtn_name"
	Set gtmtypes("ci_name_descriptor",1,"off")=0
	Set gtmtypes("ci_name_descriptor",1,"len")=16
	Set gtmtypes("ci_name_descriptor",1,"type")="gtm_string_t"
	Set gtmtypfldindx("ci_name_descriptor","rtn_name")=1
	Set gtmtypes("ci_name_descriptor",2,"name")="ci_name_descriptor.handle"
	Set gtmtypes("ci_name_descriptor",2,"off")=16
	Set gtmtypes("ci_name_descriptor",2,"len")=8
	Set gtmtypes("ci_name_descriptor",2,"type")="addr"
	Set gtmtypfldindx("ci_name_descriptor","handle")=2
	;
	Set gtmtypes("clb_stat")="struct"
	Set gtmtypes("clb_stat",0)=10
	Set gtmtypes("clb_stat","len")=32
	Set gtmtypes("clb_stat",1,"name")="clb_stat.read"
	Set gtmtypes("clb_stat",1,"off")=0
	Set gtmtypes("clb_stat",1,"len")=16
	Set gtmtypes("clb_stat",1,"type")="struct"
	Set gtmtypfldindx("clb_stat","read")=1
	Set gtmtypes("clb_stat",2,"name")="clb_stat.read.msgs"
	Set gtmtypes("clb_stat",2,"off")=0
	Set gtmtypes("clb_stat",2,"len")=4
	Set gtmtypes("clb_stat",2,"type")="unsigned-int"
	Set gtmtypfldindx("clb_stat","read.msgs")=2
	Set gtmtypes("clb_stat",3,"name")="clb_stat.read.errors"
	Set gtmtypes("clb_stat",3,"off")=4
	Set gtmtypes("clb_stat",3,"len")=4
	Set gtmtypes("clb_stat",3,"type")="unsigned-int"
	Set gtmtypfldindx("clb_stat","read.errors")=3
	Set gtmtypes("clb_stat",4,"name")="clb_stat.read.bytes"
	Set gtmtypes("clb_stat",4,"off")=8
	Set gtmtypes("clb_stat",4,"len")=4
	Set gtmtypes("clb_stat",4,"type")="unsigned-int"
	Set gtmtypfldindx("clb_stat","read.bytes")=4
	Set gtmtypes("clb_stat",5,"name")="clb_stat.read.last_error"
	Set gtmtypes("clb_stat",5,"off")=12
	Set gtmtypes("clb_stat",5,"len")=4
	Set gtmtypes("clb_stat",5,"type")="unsigned-int"
	Set gtmtypfldindx("clb_stat","read.last_error")=5
	Set gtmtypes("clb_stat",6,"name")="clb_stat.write"
	Set gtmtypes("clb_stat",6,"off")=16
	Set gtmtypes("clb_stat",6,"len")=16
	Set gtmtypes("clb_stat",6,"type")="struct"
	Set gtmtypfldindx("clb_stat","write")=6
	Set gtmtypes("clb_stat",7,"name")="clb_stat.write.msgs"
	Set gtmtypes("clb_stat",7,"off")=16
	Set gtmtypes("clb_stat",7,"len")=4
	Set gtmtypes("clb_stat",7,"type")="unsigned-int"
	Set gtmtypfldindx("clb_stat","write.msgs")=7
	Set gtmtypes("clb_stat",8,"name")="clb_stat.write.errors"
	Set gtmtypes("clb_stat",8,"off")=20
	Set gtmtypes("clb_stat",8,"len")=4
	Set gtmtypes("clb_stat",8,"type")="unsigned-int"
	Set gtmtypfldindx("clb_stat","write.errors")=8
	Set gtmtypes("clb_stat",9,"name")="clb_stat.write.bytes"
	Set gtmtypes("clb_stat",9,"off")=24
	Set gtmtypes("clb_stat",9,"len")=4
	Set gtmtypes("clb_stat",9,"type")="unsigned-int"
	Set gtmtypfldindx("clb_stat","write.bytes")=9
	Set gtmtypes("clb_stat",10,"name")="clb_stat.write.last_error"
	Set gtmtypes("clb_stat",10,"off")=28
	Set gtmtypes("clb_stat",10,"len")=4
	Set gtmtypes("clb_stat",10,"type")="unsigned-int"
	Set gtmtypfldindx("clb_stat","write.last_error")=10
	;
	Set gtmtypes("clear_confirm")="struct"
	Set gtmtypes("clear_confirm",0)=2
	Set gtmtypes("clear_confirm","len")=2
	Set gtmtypes("clear_confirm",1,"name")="clear_confirm.code"
	Set gtmtypes("clear_confirm",1,"off")=0
	Set gtmtypes("clear_confirm",1,"len")=1
	Set gtmtypes("clear_confirm",1,"type")="char"
	Set gtmtypfldindx("clear_confirm","code")=1
	Set gtmtypes("clear_confirm",2,"name")="clear_confirm.clear"
	Set gtmtypes("clear_confirm",2,"off")=1
	Set gtmtypes("clear_confirm",2,"len")=1
	Set gtmtypes("clear_confirm",2,"type")="char"
	Set gtmtypfldindx("clear_confirm","clear")=2
	;
	Set gtmtypes("clear_reply")="struct"
	Set gtmtypes("clear_reply",0)=5
	Set gtmtypes("clear_reply","len")=268
	Set gtmtypes("clear_reply",1,"name")="clear_reply.code"
	Set gtmtypes("clear_reply",1,"off")=0
	Set gtmtypes("clear_reply",1,"len")=1
	Set gtmtypes("clear_reply",1,"type")="char"
	Set gtmtypfldindx("clear_reply","code")=1
	Set gtmtypes("clear_reply",2,"name")="clear_reply.filler"
	Set gtmtypes("clear_reply",2,"off")=1
	Set gtmtypes("clear_reply",2,"len")=3
	Set gtmtypes("clear_reply",2,"type")="char"
	Set gtmtypfldindx("clear_reply","filler")=2
	Set gtmtypes("clear_reply",3,"name")="clear_reply.status"
	Set gtmtypes("clear_reply",3,"off")=4
	Set gtmtypes("clear_reply",3,"len")=4
	Set gtmtypes("clear_reply",3,"type")="int"
	Set gtmtypfldindx("clear_reply","status")=3
	Set gtmtypes("clear_reply",4,"name")="clear_reply.locknamelength"
	Set gtmtypes("clear_reply",4,"off")=8
	Set gtmtypes("clear_reply",4,"len")=4
	Set gtmtypes("clear_reply",4,"type")="int"
	Set gtmtypfldindx("clear_reply","locknamelength")=4
	Set gtmtypes("clear_reply",5,"name")="clear_reply.lockname"
	Set gtmtypes("clear_reply",5,"off")=12
	Set gtmtypes("clear_reply",5,"len")=256
	Set gtmtypes("clear_reply",5,"type")="char"
	Set gtmtypfldindx("clear_reply","lockname")=5
	;
	Set gtmtypes("clear_request")="struct"
	Set gtmtypes("clear_request",0)=7
	Set gtmtypes("clear_request","len")=44
	Set gtmtypes("clear_request",1,"name")="clear_request.code"
	Set gtmtypes("clear_request",1,"off")=0
	Set gtmtypes("clear_request",1,"len")=1
	Set gtmtypes("clear_request",1,"type")="char"
	Set gtmtypfldindx("clear_request","code")=1
	Set gtmtypes("clear_request",2,"name")="clear_request.rnum"
	Set gtmtypes("clear_request",2,"off")=1
	Set gtmtypes("clear_request",2,"len")=1
	Set gtmtypes("clear_request",2,"type")="char"
	Set gtmtypfldindx("clear_request","rnum")=2
	Set gtmtypes("clear_request",3,"name")="clear_request.all"
	Set gtmtypes("clear_request",3,"off")=2
	Set gtmtypes("clear_request",3,"len")=1
	Set gtmtypes("clear_request",3,"type")="char"
	Set gtmtypfldindx("clear_request","all")=3
	Set gtmtypes("clear_request",4,"name")="clear_request.interactive"
	Set gtmtypes("clear_request",4,"off")=3
	Set gtmtypes("clear_request",4,"len")=1
	Set gtmtypes("clear_request",4,"type")="char"
	Set gtmtypfldindx("clear_request","interactive")=4
	Set gtmtypes("clear_request",5,"name")="clear_request.pid"
	Set gtmtypes("clear_request",5,"off")=4
	Set gtmtypes("clear_request",5,"len")=4
	Set gtmtypes("clear_request",5,"type")="int"
	Set gtmtypfldindx("clear_request","pid")=5
	Set gtmtypes("clear_request",6,"name")="clear_request.nodelength"
	Set gtmtypes("clear_request",6,"off")=8
	Set gtmtypes("clear_request",6,"len")=1
	Set gtmtypes("clear_request",6,"type")="char"
	Set gtmtypfldindx("clear_request","nodelength")=6
	Set gtmtypes("clear_request",7,"name")="clear_request.node"
	Set gtmtypes("clear_request",7,"off")=9
	Set gtmtypes("clear_request",7,"len")=32
	Set gtmtypes("clear_request",7,"type")="char"
	Set gtmtypfldindx("clear_request","node")=7
	;
	Set gtmtypes("cm_lckblklck")="struct"
	Set gtmtypes("cm_lckblklck",0)=8
	Set gtmtypes("cm_lckblklck","len")=56
	Set gtmtypes("cm_lckblklck",1,"name")="cm_lckblklck.next"
	Set gtmtypes("cm_lckblklck",1,"off")=0
	Set gtmtypes("cm_lckblklck",1,"len")=8
	Set gtmtypes("cm_lckblklck",1,"type")="addr"
	Set gtmtypfldindx("cm_lckblklck","next")=1
	Set gtmtypes("cm_lckblklck",2,"name")="cm_lckblklck.last"
	Set gtmtypes("cm_lckblklck",2,"off")=8
	Set gtmtypes("cm_lckblklck",2,"len")=8
	Set gtmtypes("cm_lckblklck",2,"type")="addr"
	Set gtmtypfldindx("cm_lckblklck","last")=2
	Set gtmtypes("cm_lckblklck",3,"name")="cm_lckblklck.node"
	Set gtmtypes("cm_lckblklck",3,"off")=16
	Set gtmtypes("cm_lckblklck",3,"len")=8
	Set gtmtypes("cm_lckblklck",3,"type")="addr"
	Set gtmtypfldindx("cm_lckblklck","node")=3
	Set gtmtypes("cm_lckblklck",4,"name")="cm_lckblklck.prc"
	Set gtmtypes("cm_lckblklck",4,"off")=24
	Set gtmtypes("cm_lckblklck",4,"len")=8
	Set gtmtypes("cm_lckblklck",4,"type")="addr"
	Set gtmtypfldindx("cm_lckblklck","prc")=4
	Set gtmtypes("cm_lckblklck",5,"name")="cm_lckblklck.sequence"
	Set gtmtypes("cm_lckblklck",5,"off")=32
	Set gtmtypes("cm_lckblklck",5,"len")=4
	Set gtmtypes("cm_lckblklck",5,"type")="unsigned-int"
	Set gtmtypfldindx("cm_lckblklck","sequence")=5
	Set gtmtypes("cm_lckblklck",6,"name")="cm_lckblklck.blktime"
	Set gtmtypes("cm_lckblklck",6,"off")=40
	Set gtmtypes("cm_lckblklck",6,"len")=16
	Set gtmtypes("cm_lckblklck",6,"type")="ABS_TIME"
	Set gtmtypfldindx("cm_lckblklck","blktime")=6
	Set gtmtypes("cm_lckblklck",7,"name")="cm_lckblklck.blktime.at_sec"
	Set gtmtypes("cm_lckblklck",7,"off")=40
	Set gtmtypes("cm_lckblklck",7,"len")=8
	Set gtmtypes("cm_lckblklck",7,"type")="long"
	Set gtmtypfldindx("cm_lckblklck","blktime.at_sec")=7
	Set gtmtypes("cm_lckblklck",8,"name")="cm_lckblklck.blktime.at_usec"
	Set gtmtypes("cm_lckblklck",8,"off")=48
	Set gtmtypes("cm_lckblklck",8,"len")=8
	Set gtmtypes("cm_lckblklck",8,"type")="long"
	Set gtmtypfldindx("cm_lckblklck","blktime.at_usec")=8
	;
	Set gtmtypes("cm_lckblkprc")="struct"
	Set gtmtypes("cm_lckblkprc",0)=5
	Set gtmtypes("cm_lckblkprc","len")=40
	Set gtmtypes("cm_lckblkprc",1,"name")="cm_lckblkprc.next"
	Set gtmtypes("cm_lckblkprc",1,"off")=0
	Set gtmtypes("cm_lckblkprc",1,"len")=8
	Set gtmtypes("cm_lckblkprc",1,"type")="addr"
	Set gtmtypfldindx("cm_lckblkprc","next")=1
	Set gtmtypes("cm_lckblkprc",2,"name")="cm_lckblkprc.last"
	Set gtmtypes("cm_lckblkprc",2,"off")=8
	Set gtmtypes("cm_lckblkprc",2,"len")=8
	Set gtmtypes("cm_lckblkprc",2,"type")="addr"
	Set gtmtypfldindx("cm_lckblkprc","last")=2
	Set gtmtypes("cm_lckblkprc",3,"name")="cm_lckblkprc.user"
	Set gtmtypes("cm_lckblkprc",3,"off")=16
	Set gtmtypes("cm_lckblkprc",3,"len")=8
	Set gtmtypes("cm_lckblkprc",3,"type")="addr"
	Set gtmtypfldindx("cm_lckblkprc","user")=3
	Set gtmtypes("cm_lckblkprc",4,"name")="cm_lckblkprc.blocked"
	Set gtmtypes("cm_lckblkprc",4,"off")=24
	Set gtmtypes("cm_lckblkprc",4,"len")=8
	Set gtmtypes("cm_lckblkprc",4,"type")="addr"
	Set gtmtypfldindx("cm_lckblkprc","blocked")=4
	Set gtmtypes("cm_lckblkprc",5,"name")="cm_lckblkprc.blk_sequence"
	Set gtmtypes("cm_lckblkprc",5,"off")=32
	Set gtmtypes("cm_lckblkprc",5,"len")=4
	Set gtmtypes("cm_lckblkprc",5,"type")="unsigned-int"
	Set gtmtypfldindx("cm_lckblkprc","blk_sequence")=5
	;
	Set gtmtypes("cm_lckblkreg")="struct"
	Set gtmtypes("cm_lckblkreg",0)=4
	Set gtmtypes("cm_lckblkreg","len")=32
	Set gtmtypes("cm_lckblkreg",1,"name")="cm_lckblkreg.region"
	Set gtmtypes("cm_lckblkreg",1,"off")=0
	Set gtmtypes("cm_lckblkreg",1,"len")=8
	Set gtmtypes("cm_lckblkreg",1,"type")="addr"
	Set gtmtypfldindx("cm_lckblkreg","region")=1
	Set gtmtypes("cm_lckblkreg",2,"name")="cm_lckblkreg.next"
	Set gtmtypes("cm_lckblkreg",2,"off")=8
	Set gtmtypes("cm_lckblkreg",2,"len")=8
	Set gtmtypes("cm_lckblkreg",2,"type")="addr"
	Set gtmtypfldindx("cm_lckblkreg","next")=2
	Set gtmtypes("cm_lckblkreg",3,"name")="cm_lckblkreg.lock"
	Set gtmtypes("cm_lckblkreg",3,"off")=16
	Set gtmtypes("cm_lckblkreg",3,"len")=8
	Set gtmtypes("cm_lckblkreg",3,"type")="addr"
	Set gtmtypfldindx("cm_lckblkreg","lock")=3
	Set gtmtypes("cm_lckblkreg",4,"name")="cm_lckblkreg.pass"
	Set gtmtypes("cm_lckblkreg",4,"off")=24
	Set gtmtypes("cm_lckblkreg",4,"len")=4
	Set gtmtypes("cm_lckblkreg",4,"type")="unsigned-int"
	Set gtmtypfldindx("cm_lckblkreg","pass")=4
	;
	Set gtmtypes("cm_lk_response")="struct"
	Set gtmtypes("cm_lk_response",0)=2
	Set gtmtypes("cm_lk_response","len")=16
	Set gtmtypes("cm_lk_response",1,"name")="cm_lk_response.next"
	Set gtmtypes("cm_lk_response",1,"off")=0
	Set gtmtypes("cm_lk_response",1,"len")=8
	Set gtmtypes("cm_lk_response",1,"type")="addr"
	Set gtmtypfldindx("cm_lk_response","next")=1
	Set gtmtypes("cm_lk_response",2,"name")="cm_lk_response.response"
	Set gtmtypes("cm_lk_response",2,"off")=8
	Set gtmtypes("cm_lk_response",2,"len")=8
	Set gtmtypes("cm_lk_response",2,"type")="addr"
	Set gtmtypfldindx("cm_lk_response","response")=2
	;
	Set gtmtypes("cm_region_head")="struct"
	Set gtmtypes("cm_region_head",0)=11
	Set gtmtypes("cm_region_head","len")=80
	Set gtmtypes("cm_region_head",1,"name")="cm_region_head.head"
	Set gtmtypes("cm_region_head",1,"off")=0
	Set gtmtypes("cm_region_head",1,"len")=24
	Set gtmtypes("cm_region_head",1,"type")="que_head"
	Set gtmtypfldindx("cm_region_head","head")=1
	Set gtmtypes("cm_region_head",2,"name")="cm_region_head.head.fl"
	Set gtmtypes("cm_region_head",2,"off")=0
	Set gtmtypes("cm_region_head",2,"len")=8
	Set gtmtypes("cm_region_head",2,"type")="intptr_t"
	Set gtmtypfldindx("cm_region_head","head.fl")=2
	Set gtmtypes("cm_region_head",3,"name")="cm_region_head.head.bl"
	Set gtmtypes("cm_region_head",3,"off")=8
	Set gtmtypes("cm_region_head",3,"len")=8
	Set gtmtypes("cm_region_head",3,"type")="intptr_t"
	Set gtmtypfldindx("cm_region_head","head.bl")=3
	Set gtmtypes("cm_region_head",4,"name")="cm_region_head.head.latch"
	Set gtmtypes("cm_region_head",4,"off")=16
	Set gtmtypes("cm_region_head",4,"len")=8
	Set gtmtypes("cm_region_head",4,"type")="global_latch_t"
	Set gtmtypfldindx("cm_region_head","head.latch")=4
	Set gtmtypes("cm_region_head",5,"name")="cm_region_head.next"
	Set gtmtypes("cm_region_head",5,"off")=24
	Set gtmtypes("cm_region_head",5,"len")=8
	Set gtmtypes("cm_region_head",5,"type")="addr"
	Set gtmtypfldindx("cm_region_head","next")=5
	Set gtmtypes("cm_region_head",6,"name")="cm_region_head.last"
	Set gtmtypes("cm_region_head",6,"off")=32
	Set gtmtypes("cm_region_head",6,"len")=8
	Set gtmtypes("cm_region_head",6,"type")="addr"
	Set gtmtypfldindx("cm_region_head","last")=6
	Set gtmtypes("cm_region_head",7,"name")="cm_region_head.connect_ptr"
	Set gtmtypes("cm_region_head",7,"off")=40
	Set gtmtypes("cm_region_head",7,"len")=8
	Set gtmtypes("cm_region_head",7,"type")="addr"
	Set gtmtypfldindx("cm_region_head","connect_ptr")=7
	Set gtmtypes("cm_region_head",8,"name")="cm_region_head.reg"
	Set gtmtypes("cm_region_head",8,"off")=48
	Set gtmtypes("cm_region_head",8,"len")=8
	Set gtmtypes("cm_region_head",8,"type")="addr"
	Set gtmtypfldindx("cm_region_head","reg")=8
	Set gtmtypes("cm_region_head",9,"name")="cm_region_head.wakeup"
	Set gtmtypes("cm_region_head",9,"off")=56
	Set gtmtypes("cm_region_head",9,"len")=8
	Set gtmtypes("cm_region_head",9,"type")="uint64_t"
	Set gtmtypfldindx("cm_region_head","wakeup")=9
	Set gtmtypes("cm_region_head",10,"name")="cm_region_head.refcnt"
	Set gtmtypes("cm_region_head",10,"off")=64
	Set gtmtypes("cm_region_head",10,"len")=4
	Set gtmtypes("cm_region_head",10,"type")="unsigned-int"
	Set gtmtypfldindx("cm_region_head","refcnt")=10
	Set gtmtypes("cm_region_head",11,"name")="cm_region_head.reg_hash"
	Set gtmtypes("cm_region_head",11,"off")=72
	Set gtmtypes("cm_region_head",11,"len")=8
	Set gtmtypes("cm_region_head",11,"type")="addr"
	Set gtmtypfldindx("cm_region_head","reg_hash")=11
	;
	Set gtmtypes("cm_region_list")="struct"
	Set gtmtypes("cm_region_list",0)=14
	Set gtmtypes("cm_region_list","len")=72
	Set gtmtypes("cm_region_list",1,"name")="cm_region_list.regque"
	Set gtmtypes("cm_region_list",1,"off")=0
	Set gtmtypes("cm_region_list",1,"len")=16
	Set gtmtypes("cm_region_list",1,"type")="que_ent"
	Set gtmtypfldindx("cm_region_list","regque")=1
	Set gtmtypes("cm_region_list",2,"name")="cm_region_list.regque.fl"
	Set gtmtypes("cm_region_list",2,"off")=0
	Set gtmtypes("cm_region_list",2,"len")=8
	Set gtmtypes("cm_region_list",2,"type")="intptr_t"
	Set gtmtypfldindx("cm_region_list","regque.fl")=2
	Set gtmtypes("cm_region_list",3,"name")="cm_region_list.regque.bl"
	Set gtmtypes("cm_region_list",3,"off")=8
	Set gtmtypes("cm_region_list",3,"len")=8
	Set gtmtypes("cm_region_list",3,"type")="intptr_t"
	Set gtmtypfldindx("cm_region_list","regque.bl")=3
	Set gtmtypes("cm_region_list",4,"name")="cm_region_list.next"
	Set gtmtypes("cm_region_list",4,"off")=16
	Set gtmtypes("cm_region_list",4,"len")=8
	Set gtmtypes("cm_region_list",4,"type")="addr"
	Set gtmtypfldindx("cm_region_list","next")=4
	Set gtmtypes("cm_region_list",5,"name")="cm_region_list.regnum"
	Set gtmtypes("cm_region_list",5,"off")=24
	Set gtmtypes("cm_region_list",5,"len")=1
	Set gtmtypes("cm_region_list",5,"type")="unsigned-char"
	Set gtmtypfldindx("cm_region_list","regnum")=5
	Set gtmtypes("cm_region_list",6,"name")="cm_region_list.oper"
	Set gtmtypes("cm_region_list",6,"off")=25
	Set gtmtypes("cm_region_list",6,"len")=1
	Set gtmtypes("cm_region_list",6,"type")="unsigned-char"
	Set gtmtypfldindx("cm_region_list","oper")=6
	Set gtmtypes("cm_region_list",7,"name")="cm_region_list.lks_this_cmd"
	Set gtmtypes("cm_region_list",7,"off")=26
	Set gtmtypes("cm_region_list",7,"len")=2
	Set gtmtypes("cm_region_list",7,"type")="unsigned-short"
	Set gtmtypfldindx("cm_region_list","lks_this_cmd")=7
	Set gtmtypes("cm_region_list",8,"name")="cm_region_list.reqnode"
	Set gtmtypes("cm_region_list",8,"off")=28
	Set gtmtypes("cm_region_list",8,"len")=1
	Set gtmtypes("cm_region_list",8,"type")="char"
	Set gtmtypfldindx("cm_region_list","reqnode")=8
	Set gtmtypes("cm_region_list",9,"name")="cm_region_list.filler"
	Set gtmtypes("cm_region_list",9,"off")=29
	Set gtmtypes("cm_region_list",9,"len")=3
	Set gtmtypes("cm_region_list",9,"type")="char"
	Set gtmtypfldindx("cm_region_list","filler")=9
	Set gtmtypes("cm_region_list",10,"name")="cm_region_list.reghead"
	Set gtmtypes("cm_region_list",10,"off")=32
	Set gtmtypes("cm_region_list",10,"len")=8
	Set gtmtypes("cm_region_list",10,"type")="addr"
	Set gtmtypfldindx("cm_region_list","reghead")=10
	Set gtmtypes("cm_region_list",11,"name")="cm_region_list.cs"
	Set gtmtypes("cm_region_list",11,"off")=40
	Set gtmtypes("cm_region_list",11,"len")=8
	Set gtmtypes("cm_region_list",11,"type")="addr"
	Set gtmtypfldindx("cm_region_list","cs")=11
	Set gtmtypes("cm_region_list",12,"name")="cm_region_list.blkd"
	Set gtmtypes("cm_region_list",12,"off")=48
	Set gtmtypes("cm_region_list",12,"len")=8
	Set gtmtypes("cm_region_list",12,"type")="addr"
	Set gtmtypfldindx("cm_region_list","blkd")=12
	Set gtmtypes("cm_region_list",13,"name")="cm_region_list.lockdata"
	Set gtmtypes("cm_region_list",13,"off")=56
	Set gtmtypes("cm_region_list",13,"len")=8
	Set gtmtypes("cm_region_list",13,"type")="addr"
	Set gtmtypfldindx("cm_region_list","lockdata")=13
	Set gtmtypes("cm_region_list",14,"name")="cm_region_list.pini_addr"
	Set gtmtypes("cm_region_list",14,"off")=64
	Set gtmtypes("cm_region_list",14,"len")=4
	Set gtmtypes("cm_region_list",14,"type")="unsigned-int"
	Set gtmtypfldindx("cm_region_list","pini_addr")=14
	;
	Set gtmtypes("coll_hdr")="struct"
	Set gtmtypes("coll_hdr",0)=4
	Set gtmtypes("coll_hdr","len")=4
	Set gtmtypes("coll_hdr",1,"name")="coll_hdr.act"
	Set gtmtypes("coll_hdr",1,"off")=0
	Set gtmtypes("coll_hdr",1,"len")=1
	Set gtmtypes("coll_hdr",1,"type")="unsigned-char"
	Set gtmtypfldindx("coll_hdr","act")=1
	Set gtmtypes("coll_hdr",2,"name")="coll_hdr.nct"
	Set gtmtypes("coll_hdr",2,"off")=1
	Set gtmtypes("coll_hdr",2,"len")=1
	Set gtmtypes("coll_hdr",2,"type")="unsigned-char"
	Set gtmtypfldindx("coll_hdr","nct")=2
	Set gtmtypes("coll_hdr",3,"name")="coll_hdr.ver"
	Set gtmtypes("coll_hdr",3,"off")=2
	Set gtmtypes("coll_hdr",3,"len")=1
	Set gtmtypes("coll_hdr",3,"type")="unsigned-char"
	Set gtmtypfldindx("coll_hdr","ver")=3
	Set gtmtypes("coll_hdr",4,"name")="coll_hdr.pad"
	Set gtmtypes("coll_hdr",4,"off")=3
	Set gtmtypes("coll_hdr",4,"len")=1
	Set gtmtypes("coll_hdr",4,"type")="unsigned-char"
	Set gtmtypfldindx("coll_hdr","pad")=4
	;
	Set gtmtypes("collseq")="struct"
	Set gtmtypes("collseq",0)=8
	Set gtmtypes("collseq","len")=64
	Set gtmtypes("collseq",1,"name")="collseq.flink"
	Set gtmtypes("collseq",1,"off")=0
	Set gtmtypes("collseq",1,"len")=8
	Set gtmtypes("collseq",1,"type")="addr"
	Set gtmtypfldindx("collseq","flink")=1
	Set gtmtypes("collseq",2,"name")="collseq.act"
	Set gtmtypes("collseq",2,"off")=8
	Set gtmtypes("collseq",2,"len")=4
	Set gtmtypes("collseq",2,"type")="int"
	Set gtmtypfldindx("collseq","act")=2
	Set gtmtypes("collseq",3,"name")="collseq.xform"
	Set gtmtypes("collseq",3,"off")=16
	Set gtmtypes("collseq",3,"len")=8
	Set gtmtypes("collseq",3,"type")="addr"
	Set gtmtypfldindx("collseq","xform")=3
	Set gtmtypes("collseq",4,"name")="collseq.xback"
	Set gtmtypes("collseq",4,"off")=24
	Set gtmtypes("collseq",4,"len")=8
	Set gtmtypes("collseq",4,"type")="addr"
	Set gtmtypfldindx("collseq","xback")=4
	Set gtmtypes("collseq",5,"name")="collseq.xutil"
	Set gtmtypes("collseq",5,"off")=32
	Set gtmtypes("collseq",5,"len")=8
	Set gtmtypes("collseq",5,"type")="addr"
	Set gtmtypfldindx("collseq","xutil")=5
	Set gtmtypes("collseq",6,"name")="collseq.version"
	Set gtmtypes("collseq",6,"off")=40
	Set gtmtypes("collseq",6,"len")=8
	Set gtmtypes("collseq",6,"type")="addr"
	Set gtmtypfldindx("collseq","version")=6
	Set gtmtypes("collseq",7,"name")="collseq.verify"
	Set gtmtypes("collseq",7,"off")=48
	Set gtmtypes("collseq",7,"len")=8
	Set gtmtypes("collseq",7,"type")="addr"
	Set gtmtypfldindx("collseq","verify")=7
	Set gtmtypes("collseq",8,"name")="collseq.argtype"
	Set gtmtypes("collseq",8,"off")=56
	Set gtmtypes("collseq",8,"len")=4
	Set gtmtypes("collseq",8,"type")="int"
	Set gtmtypfldindx("collseq","argtype")=8
	;
	Set gtmtypes("command_qualifier")="struct"
	Set gtmtypes("command_qualifier",0)=25
	Set gtmtypes("command_qualifier","len")=136
	Set gtmtypes("command_qualifier",1,"name")="command_qualifier.qlf"
	Set gtmtypes("command_qualifier",1,"off")=0
	Set gtmtypes("command_qualifier",1,"len")=4
	Set gtmtypes("command_qualifier",1,"type")="unsigned-int"
	Set gtmtypfldindx("command_qualifier","qlf")=1
	Set gtmtypes("command_qualifier",2,"name")="command_qualifier.object_file"
	Set gtmtypes("command_qualifier",2,"off")=8
	Set gtmtypes("command_qualifier",2,"len")=32
	Set gtmtypes("command_qualifier",2,"type")="mval"
	Set gtmtypfldindx("command_qualifier","object_file")=2
	Set gtmtypes("command_qualifier",3,"name")="command_qualifier.object_file.mvtype"
	Set gtmtypes("command_qualifier",3,"off")=8
	Set gtmtypes("command_qualifier",3,"len")=2
	Set gtmtypes("command_qualifier",3,"type")="unsigned-short"
	Set gtmtypfldindx("command_qualifier","object_file.mvtype")=3
	Set gtmtypes("command_qualifier",4,"name")="command_qualifier.object_file.fnpc_indx"
	Set gtmtypes("command_qualifier",4,"off")=11
	Set gtmtypes("command_qualifier",4,"len")=1
	Set gtmtypes("command_qualifier",4,"type")="unsigned-char"
	Set gtmtypfldindx("command_qualifier","object_file.fnpc_indx")=4
	Set gtmtypes("command_qualifier",5,"name")="command_qualifier.object_file.utfcgr_indx"
	Set gtmtypes("command_qualifier",5,"off")=12
	Set gtmtypes("command_qualifier",5,"len")=4
	Set gtmtypes("command_qualifier",5,"type")="unsigned-int"
	Set gtmtypfldindx("command_qualifier","object_file.utfcgr_indx")=5
	Set gtmtypes("command_qualifier",6,"name")="command_qualifier.object_file.m"
	Set gtmtypes("command_qualifier",6,"off")=16
	Set gtmtypes("command_qualifier",6,"len")=8
	Set gtmtypes("command_qualifier",6,"type")="int"
	Set gtmtypfldindx("command_qualifier","object_file.m")=6
	Set gtmtypes("command_qualifier",6,"dim")=2
	Set gtmtypes("command_qualifier",7,"name")="command_qualifier.object_file.str"
	Set gtmtypes("command_qualifier",7,"off")=24
	Set gtmtypes("command_qualifier",7,"len")=16
	Set gtmtypes("command_qualifier",7,"type")="mstr"
	Set gtmtypfldindx("command_qualifier","object_file.str")=7
	Set gtmtypes("command_qualifier",8,"name")="command_qualifier.list_file"
	Set gtmtypes("command_qualifier",8,"off")=40
	Set gtmtypes("command_qualifier",8,"len")=32
	Set gtmtypes("command_qualifier",8,"type")="mval"
	Set gtmtypfldindx("command_qualifier","list_file")=8
	Set gtmtypes("command_qualifier",9,"name")="command_qualifier.list_file.mvtype"
	Set gtmtypes("command_qualifier",9,"off")=40
	Set gtmtypes("command_qualifier",9,"len")=2
	Set gtmtypes("command_qualifier",9,"type")="unsigned-short"
	Set gtmtypfldindx("command_qualifier","list_file.mvtype")=9
	Set gtmtypes("command_qualifier",10,"name")="command_qualifier.list_file.fnpc_indx"
	Set gtmtypes("command_qualifier",10,"off")=43
	Set gtmtypes("command_qualifier",10,"len")=1
	Set gtmtypes("command_qualifier",10,"type")="unsigned-char"
	Set gtmtypfldindx("command_qualifier","list_file.fnpc_indx")=10
	Set gtmtypes("command_qualifier",11,"name")="command_qualifier.list_file.utfcgr_indx"
	Set gtmtypes("command_qualifier",11,"off")=44
	Set gtmtypes("command_qualifier",11,"len")=4
	Set gtmtypes("command_qualifier",11,"type")="unsigned-int"
	Set gtmtypfldindx("command_qualifier","list_file.utfcgr_indx")=11
	Set gtmtypes("command_qualifier",12,"name")="command_qualifier.list_file.m"
	Set gtmtypes("command_qualifier",12,"off")=48
	Set gtmtypes("command_qualifier",12,"len")=8
	Set gtmtypes("command_qualifier",12,"type")="int"
	Set gtmtypfldindx("command_qualifier","list_file.m")=12
	Set gtmtypes("command_qualifier",12,"dim")=2
	Set gtmtypes("command_qualifier",13,"name")="command_qualifier.list_file.str"
	Set gtmtypes("command_qualifier",13,"off")=56
	Set gtmtypes("command_qualifier",13,"len")=16
	Set gtmtypes("command_qualifier",13,"type")="mstr"
	Set gtmtypfldindx("command_qualifier","list_file.str")=13
	Set gtmtypes("command_qualifier",14,"name")="command_qualifier.ceprep_file"
	Set gtmtypes("command_qualifier",14,"off")=72
	Set gtmtypes("command_qualifier",14,"len")=32
	Set gtmtypes("command_qualifier",14,"type")="mval"
	Set gtmtypfldindx("command_qualifier","ceprep_file")=14
	Set gtmtypes("command_qualifier",15,"name")="command_qualifier.ceprep_file.mvtype"
	Set gtmtypes("command_qualifier",15,"off")=72
	Set gtmtypes("command_qualifier",15,"len")=2
	Set gtmtypes("command_qualifier",15,"type")="unsigned-short"
	Set gtmtypfldindx("command_qualifier","ceprep_file.mvtype")=15
	Set gtmtypes("command_qualifier",16,"name")="command_qualifier.ceprep_file.fnpc_indx"
	Set gtmtypes("command_qualifier",16,"off")=75
	Set gtmtypes("command_qualifier",16,"len")=1
	Set gtmtypes("command_qualifier",16,"type")="unsigned-char"
	Set gtmtypfldindx("command_qualifier","ceprep_file.fnpc_indx")=16
	Set gtmtypes("command_qualifier",17,"name")="command_qualifier.ceprep_file.utfcgr_indx"
	Set gtmtypes("command_qualifier",17,"off")=76
	Set gtmtypes("command_qualifier",17,"len")=4
	Set gtmtypes("command_qualifier",17,"type")="unsigned-int"
	Set gtmtypfldindx("command_qualifier","ceprep_file.utfcgr_indx")=17
	Set gtmtypes("command_qualifier",18,"name")="command_qualifier.ceprep_file.m"
	Set gtmtypes("command_qualifier",18,"off")=80
	Set gtmtypes("command_qualifier",18,"len")=8
	Set gtmtypes("command_qualifier",18,"type")="int"
	Set gtmtypfldindx("command_qualifier","ceprep_file.m")=18
	Set gtmtypes("command_qualifier",18,"dim")=2
	Set gtmtypes("command_qualifier",19,"name")="command_qualifier.ceprep_file.str"
	Set gtmtypes("command_qualifier",19,"off")=88
	Set gtmtypes("command_qualifier",19,"len")=16
	Set gtmtypes("command_qualifier",19,"type")="mstr"
	Set gtmtypfldindx("command_qualifier","ceprep_file.str")=19
	Set gtmtypes("command_qualifier",20,"name")="command_qualifier.rtnname"
	Set gtmtypes("command_qualifier",20,"off")=104
	Set gtmtypes("command_qualifier",20,"len")=32
	Set gtmtypes("command_qualifier",20,"type")="mval"
	Set gtmtypfldindx("command_qualifier","rtnname")=20
	Set gtmtypes("command_qualifier",21,"name")="command_qualifier.rtnname.mvtype"
	Set gtmtypes("command_qualifier",21,"off")=104
	Set gtmtypes("command_qualifier",21,"len")=2
	Set gtmtypes("command_qualifier",21,"type")="unsigned-short"
	Set gtmtypfldindx("command_qualifier","rtnname.mvtype")=21
	Set gtmtypes("command_qualifier",22,"name")="command_qualifier.rtnname.fnpc_indx"
	Set gtmtypes("command_qualifier",22,"off")=107
	Set gtmtypes("command_qualifier",22,"len")=1
	Set gtmtypes("command_qualifier",22,"type")="unsigned-char"
	Set gtmtypfldindx("command_qualifier","rtnname.fnpc_indx")=22
	Set gtmtypes("command_qualifier",23,"name")="command_qualifier.rtnname.utfcgr_indx"
	Set gtmtypes("command_qualifier",23,"off")=108
	Set gtmtypes("command_qualifier",23,"len")=4
	Set gtmtypes("command_qualifier",23,"type")="unsigned-int"
	Set gtmtypfldindx("command_qualifier","rtnname.utfcgr_indx")=23
	Set gtmtypes("command_qualifier",24,"name")="command_qualifier.rtnname.m"
	Set gtmtypes("command_qualifier",24,"off")=112
	Set gtmtypes("command_qualifier",24,"len")=8
	Set gtmtypes("command_qualifier",24,"type")="int"
	Set gtmtypfldindx("command_qualifier","rtnname.m")=24
	Set gtmtypes("command_qualifier",24,"dim")=2
	Set gtmtypes("command_qualifier",25,"name")="command_qualifier.rtnname.str"
	Set gtmtypes("command_qualifier",25,"off")=120
	Set gtmtypes("command_qualifier",25,"len")=16
	Set gtmtypes("command_qualifier",25,"type")="mstr"
	Set gtmtypfldindx("command_qualifier","rtnname.str")=25
	;
	Set gtmtypes("compswap_time_field")="struct"
	Set gtmtypes("compswap_time_field",0)=7
	Set gtmtypes("compswap_time_field","len")=24
	Set gtmtypes("compswap_time_field",1,"name")="compswap_time_field.time_latch"
	Set gtmtypes("compswap_time_field",1,"off")=0
	Set gtmtypes("compswap_time_field",1,"len")=8
	Set gtmtypes("compswap_time_field",1,"type")="global_latch_t"
	Set gtmtypfldindx("compswap_time_field","time_latch")=1
	Set gtmtypes("compswap_time_field",2,"name")="compswap_time_field.time_latch.u"
	Set gtmtypes("compswap_time_field",2,"off")=0
	Set gtmtypes("compswap_time_field",2,"len")=8
	Set gtmtypes("compswap_time_field",2,"type")="union"
	Set gtmtypfldindx("compswap_time_field","time_latch.u")=2
	Set gtmtypes("compswap_time_field",3,"name")="compswap_time_field.time_latch.u.pid_imgcnt"
	Set gtmtypes("compswap_time_field",3,"off")=0
	Set gtmtypes("compswap_time_field",3,"len")=8
	Set gtmtypes("compswap_time_field",3,"type")="uint64_t"
	Set gtmtypfldindx("compswap_time_field","time_latch.u.pid_imgcnt")=3
	Set gtmtypes("compswap_time_field",4,"name")="compswap_time_field.time_latch.u.parts"
	Set gtmtypes("compswap_time_field",4,"off")=0
	Set gtmtypes("compswap_time_field",4,"len")=8
	Set gtmtypes("compswap_time_field",4,"type")="struct"
	Set gtmtypfldindx("compswap_time_field","time_latch.u.parts")=4
	Set gtmtypes("compswap_time_field",5,"name")="compswap_time_field.time_latch.u.parts.latch_pid"
	Set gtmtypes("compswap_time_field",5,"off")=0
	Set gtmtypes("compswap_time_field",5,"len")=4
	Set gtmtypes("compswap_time_field",5,"type")="int"
	Set gtmtypfldindx("compswap_time_field","time_latch.u.parts.latch_pid")=5
	Set gtmtypes("compswap_time_field",6,"name")="compswap_time_field.time_latch.u.parts.latch_word"
	Set gtmtypes("compswap_time_field",6,"off")=4
	Set gtmtypes("compswap_time_field",6,"len")=4
	Set gtmtypes("compswap_time_field",6,"type")="int"
	Set gtmtypfldindx("compswap_time_field","time_latch.u.parts.latch_word")=6
	Set gtmtypes("compswap_time_field",7,"name")="compswap_time_field.hp_latch_space"
	Set gtmtypes("compswap_time_field",7,"off")=8
	Set gtmtypes("compswap_time_field",7,"len")=16
	Set gtmtypes("compswap_time_field",7,"type")="int"
	Set gtmtypfldindx("compswap_time_field","hp_latch_space")=7
	Set gtmtypes("compswap_time_field",7,"dim")=4
	;
	Set gtmtypes("condition_handler")="struct"
	Set gtmtypes("condition_handler",0)=6
	Set gtmtypes("condition_handler","len")=232
	Set gtmtypes("condition_handler",1,"name")="condition_handler.save_active_ch"
	Set gtmtypes("condition_handler",1,"off")=0
	Set gtmtypes("condition_handler",1,"len")=8
	Set gtmtypes("condition_handler",1,"type")="addr"
	Set gtmtypfldindx("condition_handler","save_active_ch")=1
	Set gtmtypes("condition_handler",2,"name")="condition_handler.ch_active"
	Set gtmtypes("condition_handler",2,"off")=8
	Set gtmtypes("condition_handler",2,"len")=4
	Set gtmtypes("condition_handler",2,"type")="boolean_t"
	Set gtmtypfldindx("condition_handler","ch_active")=2
	Set gtmtypes("condition_handler",3,"name")="condition_handler.dollar_tlevel"
	Set gtmtypes("condition_handler",3,"off")=12
	Set gtmtypes("condition_handler",3,"len")=4
	Set gtmtypes("condition_handler",3,"type")="unsigned-int"
	Set gtmtypfldindx("condition_handler","dollar_tlevel")=3
	Set gtmtypes("condition_handler",4,"name")="condition_handler.ch"
	Set gtmtypes("condition_handler",4,"off")=16
	Set gtmtypes("condition_handler",4,"len")=8
	Set gtmtypes("condition_handler",4,"type")="addr"
	Set gtmtypfldindx("condition_handler","ch")=4
	Set gtmtypes("condition_handler",5,"name")="condition_handler.jmp"
	Set gtmtypes("condition_handler",5,"off")=24
	Set gtmtypes("condition_handler",5,"len")=200
	Set gtmtypes("condition_handler",5,"type")="jmp_buf"
	Set gtmtypfldindx("condition_handler","jmp")=5
	Set gtmtypes("condition_handler",6,"name")="condition_handler.intrpt_ok_state"
	Set gtmtypes("condition_handler",6,"off")=224
	Set gtmtypes("condition_handler",6,"len")=4
	Set gtmtypes("condition_handler",6,"type")="int"
	Set gtmtypfldindx("condition_handler","intrpt_ok_state")=6
	;
	Set gtmtypes("connection_struct")="struct"
	Set gtmtypes("connection_struct",0)=26
	Set gtmtypes("connection_struct","len")=2160
	Set gtmtypes("connection_struct",1,"name")="connection_struct.qent"
	Set gtmtypes("connection_struct",1,"off")=0
	Set gtmtypes("connection_struct",1,"len")=16
	Set gtmtypes("connection_struct",1,"type")="que_ent"
	Set gtmtypfldindx("connection_struct","qent")=1
	Set gtmtypes("connection_struct",2,"name")="connection_struct.qent.fl"
	Set gtmtypes("connection_struct",2,"off")=0
	Set gtmtypes("connection_struct",2,"len")=8
	Set gtmtypes("connection_struct",2,"type")="intptr_t"
	Set gtmtypfldindx("connection_struct","qent.fl")=2
	Set gtmtypes("connection_struct",3,"name")="connection_struct.qent.bl"
	Set gtmtypes("connection_struct",3,"off")=8
	Set gtmtypes("connection_struct",3,"len")=8
	Set gtmtypes("connection_struct",3,"type")="intptr_t"
	Set gtmtypfldindx("connection_struct","qent.bl")=3
	Set gtmtypes("connection_struct",4,"name")="connection_struct.region_root"
	Set gtmtypes("connection_struct",4,"off")=16
	Set gtmtypes("connection_struct",4,"len")=8
	Set gtmtypes("connection_struct",4,"type")="addr"
	Set gtmtypfldindx("connection_struct","region_root")=4
	Set gtmtypes("connection_struct",5,"name")="connection_struct.current_region"
	Set gtmtypes("connection_struct",5,"off")=24
	Set gtmtypes("connection_struct",5,"len")=8
	Set gtmtypes("connection_struct",5,"type")="addr"
	Set gtmtypfldindx("connection_struct","current_region")=5
	Set gtmtypes("connection_struct",6,"name")="connection_struct.clb_ptr"
	Set gtmtypes("connection_struct",6,"off")=32
	Set gtmtypes("connection_struct",6,"len")=8
	Set gtmtypes("connection_struct",6,"type")="addr"
	Set gtmtypfldindx("connection_struct","clb_ptr")=6
	Set gtmtypes("connection_struct",7,"name")="connection_struct.state"
	Set gtmtypes("connection_struct",7,"off")=40
	Set gtmtypes("connection_struct",7,"len")=1
	Set gtmtypes("connection_struct",7,"type")="unsigned-char"
	Set gtmtypfldindx("connection_struct","state")=7
	Set gtmtypes("connection_struct",8,"name")="connection_struct.new_msg"
	Set gtmtypes("connection_struct",8,"off")=41
	Set gtmtypes("connection_struct",8,"len")=1
	Set gtmtypes("connection_struct",8,"type")="unsigned-char"
	Set gtmtypfldindx("connection_struct","new_msg")=8
	Set gtmtypes("connection_struct",9,"name")="connection_struct.maxregnum"
	Set gtmtypes("connection_struct",9,"off")=42
	Set gtmtypes("connection_struct",9,"len")=1
	Set gtmtypes("connection_struct",9,"type")="unsigned-char"
	Set gtmtypfldindx("connection_struct","maxregnum")=9
	Set gtmtypes("connection_struct",10,"name")="connection_struct.waiting_in_queue"
	Set gtmtypes("connection_struct",10,"off")=43
	Set gtmtypes("connection_struct",10,"len")=1
	Set gtmtypes("connection_struct",10,"type")="char"
	Set gtmtypfldindx("connection_struct","waiting_in_queue")=10
	Set gtmtypes("connection_struct",11,"name")="connection_struct.connect"
	Set gtmtypes("connection_struct",11,"off")=48
	Set gtmtypes("connection_struct",11,"len")=16
	Set gtmtypes("connection_struct",11,"type")="struct"
	Set gtmtypfldindx("connection_struct","connect")=11
	Set gtmtypes("connection_struct",12,"name")="connection_struct.lastact"
	Set gtmtypes("connection_struct",12,"off")=64
	Set gtmtypes("connection_struct",12,"len")=8
	Set gtmtypes("connection_struct",12,"type")="time_t"
	Set gtmtypfldindx("connection_struct","lastact")=12
	Set gtmtypes("connection_struct",13,"name")="connection_struct.stats"
	Set gtmtypes("connection_struct",13,"off")=72
	Set gtmtypes("connection_struct",13,"len")=4
	Set gtmtypes("connection_struct",13,"type")="unsigned-int"
	Set gtmtypfldindx("connection_struct","stats")=13
	Set gtmtypes("connection_struct",14,"name")="connection_struct.procnum"
	Set gtmtypes("connection_struct",14,"off")=76
	Set gtmtypes("connection_struct",14,"len")=2
	Set gtmtypes("connection_struct",14,"type")="unsigned-short"
	Set gtmtypfldindx("connection_struct","procnum")=14
	Set gtmtypes("connection_struct",15,"name")="connection_struct.transnum"
	Set gtmtypes("connection_struct",15,"off")=78
	Set gtmtypes("connection_struct",15,"len")=2
	Set gtmtypes("connection_struct",15,"type")="unsigned-short"
	Set gtmtypfldindx("connection_struct","transnum")=15
	Set gtmtypes("connection_struct",16,"name")="connection_struct.lk_cancel"
	Set gtmtypes("connection_struct",16,"off")=80
	Set gtmtypes("connection_struct",16,"len")=2
	Set gtmtypes("connection_struct",16,"type")="unsigned-short"
	Set gtmtypfldindx("connection_struct","lk_cancel")=16
	Set gtmtypes("connection_struct",17,"name")="connection_struct.last_cancelled"
	Set gtmtypes("connection_struct",17,"off")=82
	Set gtmtypes("connection_struct",17,"len")=2
	Set gtmtypes("connection_struct",17,"type")="unsigned-short"
	Set gtmtypfldindx("connection_struct","last_cancelled")=17
	Set gtmtypes("connection_struct",18,"name")="connection_struct.int_cancel"
	Set gtmtypes("connection_struct",18,"off")=84
	Set gtmtypes("connection_struct",18,"len")=2
	Set gtmtypes("connection_struct",18,"type")="struct"
	Set gtmtypfldindx("connection_struct","int_cancel")=18
	Set gtmtypes("connection_struct",19,"name")="connection_struct.int_cancel.laflag"
	Set gtmtypes("connection_struct",19,"off")=84
	Set gtmtypes("connection_struct",19,"len")=1
	Set gtmtypes("connection_struct",19,"type")="unsigned-char"
	Set gtmtypfldindx("connection_struct","int_cancel.laflag")=19
	Set gtmtypes("connection_struct",20,"name")="connection_struct.int_cancel.transnum"
	Set gtmtypes("connection_struct",20,"off")=85
	Set gtmtypes("connection_struct",20,"len")=1
	Set gtmtypes("connection_struct",20,"type")="unsigned-char"
	Set gtmtypfldindx("connection_struct","int_cancel.transnum")=20
	Set gtmtypes("connection_struct",21,"name")="connection_struct.pvec"
	Set gtmtypes("connection_struct",21,"off")=88
	Set gtmtypes("connection_struct",21,"len")=8
	Set gtmtypes("connection_struct",21,"type")="addr"
	Set gtmtypfldindx("connection_struct","pvec")=21
	Set gtmtypes("connection_struct",22,"name")="connection_struct.query_is_queryget"
	Set gtmtypes("connection_struct",22,"off")=96
	Set gtmtypes("connection_struct",22,"len")=4
	Set gtmtypes("connection_struct",22,"type")="boolean_t"
	Set gtmtypfldindx("connection_struct","query_is_queryget")=22
	Set gtmtypes("connection_struct",23,"name")="connection_struct.err_compat"
	Set gtmtypes("connection_struct",23,"off")=100
	Set gtmtypes("connection_struct",23,"len")=4
	Set gtmtypes("connection_struct",23,"type")="boolean_t"
	Set gtmtypfldindx("connection_struct","err_compat")=23
	Set gtmtypes("connection_struct",24,"name")="connection_struct.cli_supp_allowexisting_stdnullcoll"
	Set gtmtypes("connection_struct",24,"off")=104
	Set gtmtypes("connection_struct",24,"len")=4
	Set gtmtypes("connection_struct",24,"type")="boolean_t"
	Set gtmtypfldindx("connection_struct","cli_supp_allowexisting_stdnullcoll")=24
	Set gtmtypes("connection_struct",25,"name")="connection_struct.client_supports_long_names"
	Set gtmtypes("connection_struct",25,"off")=108
	Set gtmtypes("connection_struct",25,"len")=4
	Set gtmtypes("connection_struct",25,"type")="boolean_t"
	Set gtmtypfldindx("connection_struct","client_supports_long_names")=25
	Set gtmtypes("connection_struct",26,"name")="connection_struct.region_array"
	Set gtmtypes("connection_struct",26,"off")=112
	Set gtmtypes("connection_struct",26,"len")=2048
	Set gtmtypes("connection_struct",26,"type")="addr"
	Set gtmtypfldindx("connection_struct","region_array")=26
	Set gtmtypes("connection_struct",26,"dim")=256
	;
	Set gtmtypes("crit_trace")="struct"
	Set gtmtypes("crit_trace",0)=4
	Set gtmtypes("crit_trace","len")=24
	Set gtmtypes("crit_trace",1,"name")="crit_trace.call_from"
	Set gtmtypes("crit_trace",1,"off")=0
	Set gtmtypes("crit_trace",1,"len")=8
	Set gtmtypes("crit_trace",1,"type")="caddr_t"
	Set gtmtypfldindx("crit_trace","call_from")=1
	Set gtmtypes("crit_trace",2,"name")="crit_trace.crit_act"
	Set gtmtypes("crit_trace",2,"off")=8
	Set gtmtypes("crit_trace",2,"len")=4
	Set gtmtypes("crit_trace",2,"type")="int"
	Set gtmtypfldindx("crit_trace","crit_act")=2
	Set gtmtypes("crit_trace",3,"name")="crit_trace.epid"
	Set gtmtypes("crit_trace",3,"off")=12
	Set gtmtypes("crit_trace",3,"len")=4
	Set gtmtypes("crit_trace",3,"type")="int"
	Set gtmtypfldindx("crit_trace","epid")=3
	Set gtmtypes("crit_trace",4,"name")="crit_trace.curr_tn"
	Set gtmtypes("crit_trace",4,"off")=16
	Set gtmtypes("crit_trace",4,"len")=8
	Set gtmtypes("crit_trace",4,"type")="uint64_t"
	Set gtmtypfldindx("crit_trace","curr_tn")=4
	;
	Set gtmtypes("cvs_MD5_CTX")="struct"
	Set gtmtypes("cvs_MD5_CTX",0)=3
	Set gtmtypes("cvs_MD5_CTX","len")=112
	Set gtmtypes("cvs_MD5_CTX",1,"name")="cvs_MD5_CTX.buf"
	Set gtmtypes("cvs_MD5_CTX",1,"off")=0
	Set gtmtypes("cvs_MD5_CTX",1,"len")=32
	Set gtmtypes("cvs_MD5_CTX",1,"type")="unsigned-long"
	Set gtmtypfldindx("cvs_MD5_CTX","buf")=1
	Set gtmtypes("cvs_MD5_CTX",1,"dim")=4
	Set gtmtypes("cvs_MD5_CTX",2,"name")="cvs_MD5_CTX.bits"
	Set gtmtypes("cvs_MD5_CTX",2,"off")=32
	Set gtmtypes("cvs_MD5_CTX",2,"len")=16
	Set gtmtypes("cvs_MD5_CTX",2,"type")="unsigned-long"
	Set gtmtypfldindx("cvs_MD5_CTX","bits")=2
	Set gtmtypes("cvs_MD5_CTX",2,"dim")=2
	Set gtmtypes("cvs_MD5_CTX",3,"name")="cvs_MD5_CTX.in"
	Set gtmtypes("cvs_MD5_CTX",3,"off")=48
	Set gtmtypes("cvs_MD5_CTX",3,"len")=64
	Set gtmtypes("cvs_MD5_CTX",3,"type")="unsigned-char"
	Set gtmtypfldindx("cvs_MD5_CTX","in")=3
	;
	Set gtmtypes("cw_set_element")="struct"
	Set gtmtypes("cw_set_element",0)=33
	Set gtmtypes("cw_set_element","len")=192
	Set gtmtypes("cw_set_element",1,"name")="cw_set_element.tn"
	Set gtmtypes("cw_set_element",1,"off")=0
	Set gtmtypes("cw_set_element",1,"len")=8
	Set gtmtypes("cw_set_element",1,"type")="uint64_t"
	Set gtmtypfldindx("cw_set_element","tn")=1
	Set gtmtypes("cw_set_element",2,"name")="cw_set_element.old_block"
	Set gtmtypes("cw_set_element",2,"off")=8
	Set gtmtypes("cw_set_element",2,"len")=8
	Set gtmtypes("cw_set_element",2,"type")="addr"
	Set gtmtypfldindx("cw_set_element","old_block")=2
	Set gtmtypes("cw_set_element",3,"name")="cw_set_element.cr"
	Set gtmtypes("cw_set_element",3,"off")=16
	Set gtmtypes("cw_set_element",3,"len")=8
	Set gtmtypes("cw_set_element",3,"type")="addr"
	Set gtmtypfldindx("cw_set_element","cr")=3
	Set gtmtypes("cw_set_element",4,"name")="cw_set_element.next_cw_set"
	Set gtmtypes("cw_set_element",4,"off")=24
	Set gtmtypes("cw_set_element",4,"len")=8
	Set gtmtypes("cw_set_element",4,"type")="addr"
	Set gtmtypfldindx("cw_set_element","next_cw_set")=4
	Set gtmtypes("cw_set_element",5,"name")="cw_set_element.prev_cw_set"
	Set gtmtypes("cw_set_element",5,"off")=32
	Set gtmtypes("cw_set_element",5,"len")=8
	Set gtmtypes("cw_set_element",5,"type")="addr"
	Set gtmtypfldindx("cw_set_element","prev_cw_set")=5
	Set gtmtypes("cw_set_element",6,"name")="cw_set_element.high_tlevel"
	Set gtmtypes("cw_set_element",6,"off")=40
	Set gtmtypes("cw_set_element",6,"len")=8
	Set gtmtypes("cw_set_element",6,"type")="addr"
	Set gtmtypfldindx("cw_set_element","high_tlevel")=6
	Set gtmtypes("cw_set_element",7,"name")="cw_set_element.low_tlevel"
	Set gtmtypes("cw_set_element",7,"off")=48
	Set gtmtypes("cw_set_element",7,"len")=8
	Set gtmtypes("cw_set_element",7,"type")="addr"
	Set gtmtypfldindx("cw_set_element","low_tlevel")=7
	Set gtmtypes("cw_set_element",8,"name")="cw_set_element.jnl_freeaddr"
	Set gtmtypes("cw_set_element",8,"off")=56
	Set gtmtypes("cw_set_element",8,"len")=4
	Set gtmtypes("cw_set_element",8,"type")="unsigned-int"
	Set gtmtypfldindx("cw_set_element","jnl_freeaddr")=8
	Set gtmtypes("cw_set_element",9,"name")="cw_set_element.write_type"
	Set gtmtypes("cw_set_element",9,"off")=60
	Set gtmtypes("cw_set_element",9,"len")=4
	Set gtmtypes("cw_set_element",9,"type")="unsigned-int"
	Set gtmtypfldindx("cw_set_element","write_type")=9
	Set gtmtypes("cw_set_element",10,"name")="cw_set_element.recompute_list_head"
	Set gtmtypes("cw_set_element",10,"off")=64
	Set gtmtypes("cw_set_element",10,"len")=8
	Set gtmtypes("cw_set_element",10,"type")="addr"
	Set gtmtypfldindx("cw_set_element","recompute_list_head")=10
	Set gtmtypes("cw_set_element",11,"name")="cw_set_element.recompute_list_tail"
	Set gtmtypes("cw_set_element",11,"off")=72
	Set gtmtypes("cw_set_element",11,"len")=8
	Set gtmtypes("cw_set_element",11,"type")="addr"
	Set gtmtypfldindx("cw_set_element","recompute_list_tail")=11
	Set gtmtypes("cw_set_element",12,"name")="cw_set_element.mode"
	Set gtmtypes("cw_set_element",12,"off")=80
	Set gtmtypes("cw_set_element",12,"len")=4
	Set gtmtypes("cw_set_element",12,"type")="int"
	Set gtmtypfldindx("cw_set_element","mode")=12
	Set gtmtypes("cw_set_element",13,"name")="cw_set_element.blk"
	Set gtmtypes("cw_set_element",13,"off")=84
	Set gtmtypes("cw_set_element",13,"len")=4
	Set gtmtypes("cw_set_element",13,"type")="int"
	Set gtmtypfldindx("cw_set_element","blk")=13
	Set gtmtypes("cw_set_element",14,"name")="cw_set_element.upd_addr"
	Set gtmtypes("cw_set_element",14,"off")=88
	Set gtmtypes("cw_set_element",14,"len")=8
	Set gtmtypes("cw_set_element",14,"type")="addr"
	Set gtmtypfldindx("cw_set_element","upd_addr")=14
	Set gtmtypes("cw_set_element",15,"name")="cw_set_element.new_buff"
	Set gtmtypes("cw_set_element",15,"off")=96
	Set gtmtypes("cw_set_element",15,"len")=8
	Set gtmtypes("cw_set_element",15,"type")="addr"
	Set gtmtypfldindx("cw_set_element","new_buff")=15
	Set gtmtypes("cw_set_element",16,"name")="cw_set_element.blk_target"
	Set gtmtypes("cw_set_element",16,"off")=104
	Set gtmtypes("cw_set_element",16,"len")=8
	Set gtmtypes("cw_set_element",16,"type")="addr"
	Set gtmtypfldindx("cw_set_element","blk_target")=16
	Set gtmtypes("cw_set_element",17,"name")="cw_set_element.cycle"
	Set gtmtypes("cw_set_element",17,"off")=112
	Set gtmtypes("cw_set_element",17,"len")=4
	Set gtmtypes("cw_set_element",17,"type")="int"
	Set gtmtypfldindx("cw_set_element","cycle")=17
	Set gtmtypes("cw_set_element",18,"name")="cw_set_element.first_off"
	Set gtmtypes("cw_set_element",18,"off")=116
	Set gtmtypes("cw_set_element",18,"len")=4
	Set gtmtypes("cw_set_element",18,"type")="unsigned-int"
	Set gtmtypfldindx("cw_set_element","first_off")=18
	Set gtmtypes("cw_set_element",19,"name")="cw_set_element.ins_off"
	Set gtmtypes("cw_set_element",19,"off")=120
	Set gtmtypes("cw_set_element",19,"len")=4
	Set gtmtypes("cw_set_element",19,"type")="unsigned-int"
	Set gtmtypfldindx("cw_set_element","ins_off")=19
	Set gtmtypes("cw_set_element",20,"name")="cw_set_element.next_off"
	Set gtmtypes("cw_set_element",20,"off")=124
	Set gtmtypes("cw_set_element",20,"len")=4
	Set gtmtypes("cw_set_element",20,"type")="unsigned-int"
	Set gtmtypfldindx("cw_set_element","next_off")=20
	Set gtmtypes("cw_set_element",21,"name")="cw_set_element.index"
	Set gtmtypes("cw_set_element",21,"off")=128
	Set gtmtypes("cw_set_element",21,"len")=4
	Set gtmtypes("cw_set_element",21,"type")="int"
	Set gtmtypfldindx("cw_set_element","index")=21
	Set gtmtypes("cw_set_element",22,"name")="cw_set_element.reference_cnt"
	Set gtmtypes("cw_set_element",22,"off")=132
	Set gtmtypes("cw_set_element",22,"len")=4
	Set gtmtypes("cw_set_element",22,"type")="int"
	Set gtmtypfldindx("cw_set_element","reference_cnt")=22
	Set gtmtypes("cw_set_element",23,"name")="cw_set_element.level"
	Set gtmtypes("cw_set_element",23,"off")=136
	Set gtmtypes("cw_set_element",23,"len")=4
	Set gtmtypes("cw_set_element",23,"type")="int"
	Set gtmtypfldindx("cw_set_element","level")=23
	Set gtmtypes("cw_set_element",24,"name")="cw_set_element.done"
	Set gtmtypes("cw_set_element",24,"off")=140
	Set gtmtypes("cw_set_element",24,"len")=4
	Set gtmtypes("cw_set_element",24,"type")="boolean_t"
	Set gtmtypfldindx("cw_set_element","done")=24
	Set gtmtypes("cw_set_element",25,"name")="cw_set_element.first_copy"
	Set gtmtypes("cw_set_element",25,"off")=144
	Set gtmtypes("cw_set_element",25,"len")=4
	Set gtmtypes("cw_set_element",25,"type")="boolean_t"
	Set gtmtypfldindx("cw_set_element","first_copy")=25
	Set gtmtypes("cw_set_element",26,"name")="cw_set_element.forward_process"
	Set gtmtypes("cw_set_element",26,"off")=148
	Set gtmtypes("cw_set_element",26,"len")=4
	Set gtmtypes("cw_set_element",26,"type")="boolean_t"
	Set gtmtypfldindx("cw_set_element","forward_process")=26
	Set gtmtypes("cw_set_element",27,"name")="cw_set_element.t_level"
	Set gtmtypes("cw_set_element",27,"off")=152
	Set gtmtypes("cw_set_element",27,"len")=4
	Set gtmtypes("cw_set_element",27,"type")="unsigned-int"
	Set gtmtypfldindx("cw_set_element","t_level")=27
	Set gtmtypes("cw_set_element",28,"name")="cw_set_element.ondsk_blkver"
	Set gtmtypes("cw_set_element",28,"off")=156
	Set gtmtypes("cw_set_element",28,"len")=4
	Set gtmtypes("cw_set_element",28,"type")="int"
	Set gtmtypfldindx("cw_set_element","ondsk_blkver")=28
	Set gtmtypes("cw_set_element",29,"name")="cw_set_element.old_mode"
	Set gtmtypes("cw_set_element",29,"off")=160
	Set gtmtypes("cw_set_element",29,"len")=4
	Set gtmtypes("cw_set_element",29,"type")="int"
	Set gtmtypfldindx("cw_set_element","old_mode")=29
	Set gtmtypes("cw_set_element",30,"name")="cw_set_element.undo_next_off"
	Set gtmtypes("cw_set_element",30,"off")=164
	Set gtmtypes("cw_set_element",30,"len")=8
	Set gtmtypes("cw_set_element",30,"type")="unsigned-int"
	Set gtmtypfldindx("cw_set_element","undo_next_off")=30
	Set gtmtypes("cw_set_element",30,"dim")=2
	Set gtmtypes("cw_set_element",31,"name")="cw_set_element.undo_offset"
	Set gtmtypes("cw_set_element",31,"off")=172
	Set gtmtypes("cw_set_element",31,"len")=8
	Set gtmtypes("cw_set_element",31,"type")="unsigned-int"
	Set gtmtypfldindx("cw_set_element","undo_offset")=31
	Set gtmtypes("cw_set_element",31,"dim")=2
	Set gtmtypes("cw_set_element",32,"name")="cw_set_element.blk_checksum"
	Set gtmtypes("cw_set_element",32,"off")=180
	Set gtmtypes("cw_set_element",32,"len")=4
	Set gtmtypes("cw_set_element",32,"type")="unsigned-int"
	Set gtmtypfldindx("cw_set_element","blk_checksum")=32
	Set gtmtypes("cw_set_element",33,"name")="cw_set_element.blk_prior_state"
	Set gtmtypes("cw_set_element",33,"off")=184
	Set gtmtypes("cw_set_element",33,"len")=4
	Set gtmtypes("cw_set_element",33,"type")="unsigned-int"
	Set gtmtypfldindx("cw_set_element","blk_prior_state")=33
	;
	Set gtmtypes("d_rm_struct")="struct"
	Set gtmtypes("d_rm_struct",0)=91
	Set gtmtypes("d_rm_struct","len")=488
	Set gtmtypes("d_rm_struct",1,"name")="d_rm_struct.fixed"
	Set gtmtypes("d_rm_struct",1,"off")=0
	Set gtmtypes("d_rm_struct",1,"len")=4
	Set gtmtypes("d_rm_struct",1,"type")="boolean_t"
	Set gtmtypfldindx("d_rm_struct","fixed")=1
	Set gtmtypes("d_rm_struct",2,"name")="d_rm_struct.read_only"
	Set gtmtypes("d_rm_struct",2,"off")=4
	Set gtmtypes("d_rm_struct",2,"len")=4
	Set gtmtypes("d_rm_struct",2,"type")="boolean_t"
	Set gtmtypfldindx("d_rm_struct","read_only")=2
	Set gtmtypes("d_rm_struct",3,"name")="d_rm_struct.write_only"
	Set gtmtypes("d_rm_struct",3,"off")=8
	Set gtmtypes("d_rm_struct",3,"len")=4
	Set gtmtypes("d_rm_struct",3,"type")="boolean_t"
	Set gtmtypfldindx("d_rm_struct","write_only")=3
	Set gtmtypes("d_rm_struct",4,"name")="d_rm_struct.stream"
	Set gtmtypes("d_rm_struct",4,"off")=12
	Set gtmtypes("d_rm_struct",4,"len")=4
	Set gtmtypes("d_rm_struct",4,"type")="boolean_t"
	Set gtmtypfldindx("d_rm_struct","stream")=4
	Set gtmtypes("d_rm_struct",5,"name")="d_rm_struct.fifo"
	Set gtmtypes("d_rm_struct",5,"off")=16
	Set gtmtypes("d_rm_struct",5,"len")=4
	Set gtmtypes("d_rm_struct",5,"type")="boolean_t"
	Set gtmtypfldindx("d_rm_struct","fifo")=5
	Set gtmtypes("d_rm_struct",6,"name")="d_rm_struct.is_pipe"
	Set gtmtypes("d_rm_struct",6,"off")=20
	Set gtmtypes("d_rm_struct",6,"len")=4
	Set gtmtypes("d_rm_struct",6,"type")="boolean_t"
	Set gtmtypfldindx("d_rm_struct","is_pipe")=6
	Set gtmtypes("d_rm_struct",7,"name")="d_rm_struct.independent"
	Set gtmtypes("d_rm_struct",7,"off")=24
	Set gtmtypes("d_rm_struct",7,"len")=4
	Set gtmtypes("d_rm_struct",7,"type")="boolean_t"
	Set gtmtypfldindx("d_rm_struct","independent")=7
	Set gtmtypes("d_rm_struct",8,"name")="d_rm_struct.parse"
	Set gtmtypes("d_rm_struct",8,"off")=28
	Set gtmtypes("d_rm_struct",8,"len")=4
	Set gtmtypes("d_rm_struct",8,"type")="boolean_t"
	Set gtmtypfldindx("d_rm_struct","parse")=8
	Set gtmtypes("d_rm_struct",9,"name")="d_rm_struct.done_1st_read"
	Set gtmtypes("d_rm_struct",9,"off")=32
	Set gtmtypes("d_rm_struct",9,"len")=4
	Set gtmtypes("d_rm_struct",9,"type")="boolean_t"
	Set gtmtypfldindx("d_rm_struct","done_1st_read")=9
	Set gtmtypes("d_rm_struct",10,"name")="d_rm_struct.done_1st_write"
	Set gtmtypes("d_rm_struct",10,"off")=36
	Set gtmtypes("d_rm_struct",10,"len")=4
	Set gtmtypes("d_rm_struct",10,"type")="boolean_t"
	Set gtmtypfldindx("d_rm_struct","done_1st_write")=10
	Set gtmtypes("d_rm_struct",11,"name")="d_rm_struct.crlast"
	Set gtmtypes("d_rm_struct",11,"off")=40
	Set gtmtypes("d_rm_struct",11,"len")=4
	Set gtmtypes("d_rm_struct",11,"type")="boolean_t"
	Set gtmtypfldindx("d_rm_struct","crlast")=11
	Set gtmtypes("d_rm_struct",12,"name")="d_rm_struct.def_width"
	Set gtmtypes("d_rm_struct",12,"off")=44
	Set gtmtypes("d_rm_struct",12,"len")=4
	Set gtmtypes("d_rm_struct",12,"type")="boolean_t"
	Set gtmtypfldindx("d_rm_struct","def_width")=12
	Set gtmtypes("d_rm_struct",13,"name")="d_rm_struct.def_recsize"
	Set gtmtypes("d_rm_struct",13,"off")=48
	Set gtmtypes("d_rm_struct",13,"len")=4
	Set gtmtypes("d_rm_struct",13,"type")="boolean_t"
	Set gtmtypfldindx("d_rm_struct","def_recsize")=13
	Set gtmtypes("d_rm_struct",14,"name")="d_rm_struct.bom_read_one_done"
	Set gtmtypes("d_rm_struct",14,"off")=52
	Set gtmtypes("d_rm_struct",14,"len")=4
	Set gtmtypes("d_rm_struct",14,"type")="boolean_t"
	Set gtmtypfldindx("d_rm_struct","bom_read_one_done")=14
	Set gtmtypes("d_rm_struct",15,"name")="d_rm_struct.follow"
	Set gtmtypes("d_rm_struct",15,"off")=56
	Set gtmtypes("d_rm_struct",15,"len")=4
	Set gtmtypes("d_rm_struct",15,"type")="boolean_t"
	Set gtmtypfldindx("d_rm_struct","follow")=15
	Set gtmtypes("d_rm_struct",16,"name")="d_rm_struct.no_destroy"
	Set gtmtypes("d_rm_struct",16,"off")=60
	Set gtmtypes("d_rm_struct",16,"len")=4
	Set gtmtypes("d_rm_struct",16,"type")="boolean_t"
	Set gtmtypfldindx("d_rm_struct","no_destroy")=16
	Set gtmtypes("d_rm_struct",17,"name")="d_rm_struct.bom_checked"
	Set gtmtypes("d_rm_struct",17,"off")=64
	Set gtmtypes("d_rm_struct",17,"len")=4
	Set gtmtypes("d_rm_struct",17,"type")="boolean_t"
	Set gtmtypfldindx("d_rm_struct","bom_checked")=17
	Set gtmtypes("d_rm_struct",18,"name")="d_rm_struct.pipe_save_state"
	Set gtmtypes("d_rm_struct",18,"off")=72
	Set gtmtypes("d_rm_struct",18,"len")=56
	Set gtmtypes("d_rm_struct",18,"type")="pipe_interrupt"
	Set gtmtypfldindx("d_rm_struct","pipe_save_state")=18
	Set gtmtypes("d_rm_struct",19,"name")="d_rm_struct.pipe_save_state.end_time"
	Set gtmtypes("d_rm_struct",19,"off")=72
	Set gtmtypes("d_rm_struct",19,"len")=16
	Set gtmtypes("d_rm_struct",19,"type")="ABS_TIME"
	Set gtmtypfldindx("d_rm_struct","pipe_save_state.end_time")=19
	Set gtmtypes("d_rm_struct",20,"name")="d_rm_struct.pipe_save_state.who_saved"
	Set gtmtypes("d_rm_struct",20,"off")=88
	Set gtmtypes("d_rm_struct",20,"len")=4
	Set gtmtypes("d_rm_struct",20,"type")="int"
	Set gtmtypfldindx("d_rm_struct","pipe_save_state.who_saved")=20
	Set gtmtypes("d_rm_struct",21,"name")="d_rm_struct.pipe_save_state.max_bufflen"
	Set gtmtypes("d_rm_struct",21,"off")=92
	Set gtmtypes("d_rm_struct",21,"len")=4
	Set gtmtypes("d_rm_struct",21,"type")="int"
	Set gtmtypfldindx("d_rm_struct","pipe_save_state.max_bufflen")=21
	Set gtmtypes("d_rm_struct",22,"name")="d_rm_struct.pipe_save_state.bytes_read"
	Set gtmtypes("d_rm_struct",22,"off")=96
	Set gtmtypes("d_rm_struct",22,"len")=4
	Set gtmtypes("d_rm_struct",22,"type")="int"
	Set gtmtypfldindx("d_rm_struct","pipe_save_state.bytes_read")=22
	Set gtmtypes("d_rm_struct",23,"name")="d_rm_struct.pipe_save_state.bytes2read"
	Set gtmtypes("d_rm_struct",23,"off")=100
	Set gtmtypes("d_rm_struct",23,"len")=4
	Set gtmtypes("d_rm_struct",23,"type")="int"
	Set gtmtypfldindx("d_rm_struct","pipe_save_state.bytes2read")=23
	Set gtmtypes("d_rm_struct",24,"name")="d_rm_struct.pipe_save_state.char_count"
	Set gtmtypes("d_rm_struct",24,"off")=104
	Set gtmtypes("d_rm_struct",24,"len")=4
	Set gtmtypes("d_rm_struct",24,"type")="int"
	Set gtmtypfldindx("d_rm_struct","pipe_save_state.char_count")=24
	Set gtmtypes("d_rm_struct",25,"name")="d_rm_struct.pipe_save_state.bytes_count"
	Set gtmtypes("d_rm_struct",25,"off")=108
	Set gtmtypes("d_rm_struct",25,"len")=4
	Set gtmtypes("d_rm_struct",25,"type")="int"
	Set gtmtypfldindx("d_rm_struct","pipe_save_state.bytes_count")=25
	Set gtmtypes("d_rm_struct",26,"name")="d_rm_struct.pipe_save_state.add_bytes"
	Set gtmtypes("d_rm_struct",26,"off")=112
	Set gtmtypes("d_rm_struct",26,"len")=4
	Set gtmtypes("d_rm_struct",26,"type")="int"
	Set gtmtypfldindx("d_rm_struct","pipe_save_state.add_bytes")=26
	Set gtmtypes("d_rm_struct",27,"name")="d_rm_struct.pipe_save_state.end_time_valid"
	Set gtmtypes("d_rm_struct",27,"off")=116
	Set gtmtypes("d_rm_struct",27,"len")=4
	Set gtmtypes("d_rm_struct",27,"type")="boolean_t"
	Set gtmtypfldindx("d_rm_struct","pipe_save_state.end_time_valid")=27
	Set gtmtypes("d_rm_struct",28,"name")="d_rm_struct.pipe_save_state.newpipe"
	Set gtmtypes("d_rm_struct",28,"off")=120
	Set gtmtypes("d_rm_struct",28,"len")=8
	Set gtmtypes("d_rm_struct",28,"type")="addr"
	Set gtmtypfldindx("d_rm_struct","pipe_save_state.newpipe")=28
	Set gtmtypes("d_rm_struct",29,"name")="d_rm_struct.mupintr"
	Set gtmtypes("d_rm_struct",29,"off")=128
	Set gtmtypes("d_rm_struct",29,"len")=4
	Set gtmtypes("d_rm_struct",29,"type")="boolean_t"
	Set gtmtypfldindx("d_rm_struct","mupintr")=29
	Set gtmtypes("d_rm_struct",30,"name")="d_rm_struct.lastop"
	Set gtmtypes("d_rm_struct",30,"off")=132
	Set gtmtypes("d_rm_struct",30,"len")=4
	Set gtmtypes("d_rm_struct",30,"type")="unsigned-int"
	Set gtmtypfldindx("d_rm_struct","lastop")=30
	Set gtmtypes("d_rm_struct",31,"name")="d_rm_struct.fildes"
	Set gtmtypes("d_rm_struct",31,"off")=136
	Set gtmtypes("d_rm_struct",31,"len")=4
	Set gtmtypes("d_rm_struct",31,"type")="int"
	Set gtmtypfldindx("d_rm_struct","fildes")=31
	Set gtmtypes("d_rm_struct",32,"name")="d_rm_struct.read_fildes"
	Set gtmtypes("d_rm_struct",32,"off")=140
	Set gtmtypes("d_rm_struct",32,"len")=4
	Set gtmtypes("d_rm_struct",32,"type")="int"
	Set gtmtypfldindx("d_rm_struct","read_fildes")=32
	Set gtmtypes("d_rm_struct",33,"name")="d_rm_struct.read_filstr"
	Set gtmtypes("d_rm_struct",33,"off")=144
	Set gtmtypes("d_rm_struct",33,"len")=8
	Set gtmtypes("d_rm_struct",33,"type")="addr"
	Set gtmtypfldindx("d_rm_struct","read_filstr")=33
	Set gtmtypes("d_rm_struct",34,"name")="d_rm_struct.stderr_child"
	Set gtmtypes("d_rm_struct",34,"off")=152
	Set gtmtypes("d_rm_struct",34,"len")=8
	Set gtmtypes("d_rm_struct",34,"type")="addr"
	Set gtmtypfldindx("d_rm_struct","stderr_child")=34
	Set gtmtypes("d_rm_struct",35,"name")="d_rm_struct.stderr_parent"
	Set gtmtypes("d_rm_struct",35,"off")=160
	Set gtmtypes("d_rm_struct",35,"len")=8
	Set gtmtypes("d_rm_struct",35,"type")="addr"
	Set gtmtypfldindx("d_rm_struct","stderr_parent")=35
	Set gtmtypes("d_rm_struct",36,"name")="d_rm_struct.pipe_pid"
	Set gtmtypes("d_rm_struct",36,"off")=168
	Set gtmtypes("d_rm_struct",36,"len")=4
	Set gtmtypes("d_rm_struct",36,"type")="pid_t"
	Set gtmtypfldindx("d_rm_struct","pipe_pid")=36
	Set gtmtypes("d_rm_struct",37,"name")="d_rm_struct.dev_param_pairs"
	Set gtmtypes("d_rm_struct",37,"off")=176
	Set gtmtypes("d_rm_struct",37,"len")=56
	Set gtmtypes("d_rm_struct",37,"type")="Dev_param_pairs"
	Set gtmtypfldindx("d_rm_struct","dev_param_pairs")=37
	Set gtmtypes("d_rm_struct",38,"name")="d_rm_struct.dev_param_pairs.num_pairs"
	Set gtmtypes("d_rm_struct",38,"off")=176
	Set gtmtypes("d_rm_struct",38,"len")=4
	Set gtmtypes("d_rm_struct",38,"type")="int"
	Set gtmtypfldindx("d_rm_struct","dev_param_pairs.num_pairs")=38
	Set gtmtypes("d_rm_struct",39,"name")="d_rm_struct.dev_param_pairs.pairs"
	Set gtmtypes("d_rm_struct",39,"off")=184
	Set gtmtypes("d_rm_struct",39,"len")=48
	Set gtmtypes("d_rm_struct",39,"type")="Dev_param_pair"
	Set gtmtypfldindx("d_rm_struct","dev_param_pairs.pairs")=39
	Set gtmtypes("d_rm_struct",39,"dim")=3
	Set gtmtypes("d_rm_struct",40,"name")="d_rm_struct.dev_param_pairs.pairs[0].name"
	Set gtmtypes("d_rm_struct",40,"off")=184
	Set gtmtypes("d_rm_struct",40,"len")=8
	Set gtmtypes("d_rm_struct",40,"type")="addr"
	Set gtmtypfldindx("d_rm_struct","dev_param_pairs.pairs[0].name")=40
	Set gtmtypes("d_rm_struct",41,"name")="d_rm_struct.dev_param_pairs.pairs[0].definition"
	Set gtmtypes("d_rm_struct",41,"off")=192
	Set gtmtypes("d_rm_struct",41,"len")=8
	Set gtmtypes("d_rm_struct",41,"type")="addr"
	Set gtmtypfldindx("d_rm_struct","dev_param_pairs.pairs[0].definition")=41
	Set gtmtypes("d_rm_struct",42,"name")="d_rm_struct.bufsize"
	Set gtmtypes("d_rm_struct",42,"off")=232
	Set gtmtypes("d_rm_struct",42,"len")=4
	Set gtmtypes("d_rm_struct",42,"type")="int"
	Set gtmtypfldindx("d_rm_struct","bufsize")=42
	Set gtmtypes("d_rm_struct",43,"name")="d_rm_struct.outbufsize"
	Set gtmtypes("d_rm_struct",43,"off")=236
	Set gtmtypes("d_rm_struct",43,"len")=4
	Set gtmtypes("d_rm_struct",43,"type")="int"
	Set gtmtypfldindx("d_rm_struct","outbufsize")=43
	Set gtmtypes("d_rm_struct",44,"name")="d_rm_struct.recordsize"
	Set gtmtypes("d_rm_struct",44,"off")=240
	Set gtmtypes("d_rm_struct",44,"len")=4
	Set gtmtypes("d_rm_struct",44,"type")="int"
	Set gtmtypfldindx("d_rm_struct","recordsize")=44
	Set gtmtypes("d_rm_struct",45,"name")="d_rm_struct.padchar"
	Set gtmtypes("d_rm_struct",45,"off")=244
	Set gtmtypes("d_rm_struct",45,"len")=4
	Set gtmtypes("d_rm_struct",45,"type")="int"
	Set gtmtypfldindx("d_rm_struct","padchar")=45
	Set gtmtypes("d_rm_struct",46,"name")="d_rm_struct.fol_bytes_read"
	Set gtmtypes("d_rm_struct",46,"off")=248
	Set gtmtypes("d_rm_struct",46,"len")=4
	Set gtmtypes("d_rm_struct",46,"type")="int"
	Set gtmtypfldindx("d_rm_struct","fol_bytes_read")=46
	Set gtmtypes("d_rm_struct",47,"name")="d_rm_struct.last_was_timeout"
	Set gtmtypes("d_rm_struct",47,"off")=252
	Set gtmtypes("d_rm_struct",47,"len")=4
	Set gtmtypes("d_rm_struct",47,"type")="int"
	Set gtmtypfldindx("d_rm_struct","last_was_timeout")=47
	Set gtmtypes("d_rm_struct",48,"name")="d_rm_struct.orig_bytes_already_read"
	Set gtmtypes("d_rm_struct",48,"off")=256
	Set gtmtypes("d_rm_struct",48,"len")=4
	Set gtmtypes("d_rm_struct",48,"type")="int"
	Set gtmtypfldindx("d_rm_struct","orig_bytes_already_read")=48
	Set gtmtypes("d_rm_struct",49,"name")="d_rm_struct.out_bytes"
	Set gtmtypes("d_rm_struct",49,"off")=260
	Set gtmtypes("d_rm_struct",49,"len")=4
	Set gtmtypes("d_rm_struct",49,"type")="int"
	Set gtmtypfldindx("d_rm_struct","out_bytes")=49
	Set gtmtypes("d_rm_struct",50,"name")="d_rm_struct.bom_buf_cnt"
	Set gtmtypes("d_rm_struct",50,"off")=264
	Set gtmtypes("d_rm_struct",50,"len")=4
	Set gtmtypes("d_rm_struct",50,"type")="unsigned-int"
	Set gtmtypfldindx("d_rm_struct","bom_buf_cnt")=50
	Set gtmtypes("d_rm_struct",51,"name")="d_rm_struct.bom_buf_off"
	Set gtmtypes("d_rm_struct",51,"off")=268
	Set gtmtypes("d_rm_struct",51,"len")=4
	Set gtmtypes("d_rm_struct",51,"type")="unsigned-int"
	Set gtmtypfldindx("d_rm_struct","bom_buf_off")=51
	Set gtmtypes("d_rm_struct",52,"name")="d_rm_struct.bom_num_bytes"
	Set gtmtypes("d_rm_struct",52,"off")=272
	Set gtmtypes("d_rm_struct",52,"len")=4
	Set gtmtypes("d_rm_struct",52,"type")="unsigned-int"
	Set gtmtypfldindx("d_rm_struct","bom_num_bytes")=52
	Set gtmtypes("d_rm_struct",53,"name")="d_rm_struct.bom_buf"
	Set gtmtypes("d_rm_struct",53,"off")=276
	Set gtmtypes("d_rm_struct",53,"len")=4
	Set gtmtypes("d_rm_struct",53,"type")="unsigned-char"
	Set gtmtypfldindx("d_rm_struct","bom_buf")=53
	Set gtmtypes("d_rm_struct",54,"name")="d_rm_struct.inbuf"
	Set gtmtypes("d_rm_struct",54,"off")=280
	Set gtmtypes("d_rm_struct",54,"len")=8
	Set gtmtypes("d_rm_struct",54,"type")="addr"
	Set gtmtypfldindx("d_rm_struct","inbuf")=54
	Set gtmtypes("d_rm_struct",55,"name")="d_rm_struct.inbuf_pos"
	Set gtmtypes("d_rm_struct",55,"off")=288
	Set gtmtypes("d_rm_struct",55,"len")=8
	Set gtmtypes("d_rm_struct",55,"type")="addr"
	Set gtmtypfldindx("d_rm_struct","inbuf_pos")=55
	Set gtmtypes("d_rm_struct",56,"name")="d_rm_struct.inbuf_off"
	Set gtmtypes("d_rm_struct",56,"off")=296
	Set gtmtypes("d_rm_struct",56,"len")=8
	Set gtmtypes("d_rm_struct",56,"type")="addr"
	Set gtmtypfldindx("d_rm_struct","inbuf_off")=56
	Set gtmtypes("d_rm_struct",57,"name")="d_rm_struct.inbuf_top"
	Set gtmtypes("d_rm_struct",57,"off")=304
	Set gtmtypes("d_rm_struct",57,"len")=8
	Set gtmtypes("d_rm_struct",57,"type")="addr"
	Set gtmtypfldindx("d_rm_struct","inbuf_top")=57
	Set gtmtypes("d_rm_struct",58,"name")="d_rm_struct.outbuf"
	Set gtmtypes("d_rm_struct",58,"off")=312
	Set gtmtypes("d_rm_struct",58,"len")=8
	Set gtmtypes("d_rm_struct",58,"type")="addr"
	Set gtmtypfldindx("d_rm_struct","outbuf")=58
	Set gtmtypes("d_rm_struct",59,"name")="d_rm_struct.filstr"
	Set gtmtypes("d_rm_struct",59,"off")=320
	Set gtmtypes("d_rm_struct",59,"len")=8
	Set gtmtypes("d_rm_struct",59,"type")="addr"
	Set gtmtypfldindx("d_rm_struct","filstr")=59
	Set gtmtypes("d_rm_struct",60,"name")="d_rm_struct.file_pos"
	Set gtmtypes("d_rm_struct",60,"off")=328
	Set gtmtypes("d_rm_struct",60,"len")=8
	Set gtmtypes("d_rm_struct",60,"type")="off_t"
	Set gtmtypfldindx("d_rm_struct","file_pos")=60
	Set gtmtypes("d_rm_struct",61,"name")="d_rm_struct.pipe_buff_size"
	Set gtmtypes("d_rm_struct",61,"off")=336
	Set gtmtypes("d_rm_struct",61,"len")=8
	Set gtmtypes("d_rm_struct",61,"type")="long"
	Set gtmtypfldindx("d_rm_struct","pipe_buff_size")=61
	Set gtmtypes("d_rm_struct",62,"name")="d_rm_struct.tmp_buffer"
	Set gtmtypes("d_rm_struct",62,"off")=344
	Set gtmtypes("d_rm_struct",62,"len")=8
	Set gtmtypes("d_rm_struct",62,"type")="addr"
	Set gtmtypfldindx("d_rm_struct","tmp_buffer")=62
	Set gtmtypes("d_rm_struct",63,"name")="d_rm_struct.tot_bytes_in_buffer"
	Set gtmtypes("d_rm_struct",63,"off")=352
	Set gtmtypes("d_rm_struct",63,"len")=4
	Set gtmtypes("d_rm_struct",63,"type")="int"
	Set gtmtypfldindx("d_rm_struct","tot_bytes_in_buffer")=63
	Set gtmtypes("d_rm_struct",64,"name")="d_rm_struct.start_pos"
	Set gtmtypes("d_rm_struct",64,"off")=356
	Set gtmtypes("d_rm_struct",64,"len")=4
	Set gtmtypes("d_rm_struct",64,"type")="int"
	Set gtmtypfldindx("d_rm_struct","start_pos")=64
	Set gtmtypes("d_rm_struct",65,"name")="d_rm_struct.write_occurred"
	Set gtmtypes("d_rm_struct",65,"off")=360
	Set gtmtypes("d_rm_struct",65,"len")=4
	Set gtmtypes("d_rm_struct",65,"type")="boolean_t"
	Set gtmtypfldindx("d_rm_struct","write_occurred")=65
	Set gtmtypes("d_rm_struct",66,"name")="d_rm_struct.read_occurred"
	Set gtmtypes("d_rm_struct",66,"off")=364
	Set gtmtypes("d_rm_struct",66,"len")=4
	Set gtmtypes("d_rm_struct",66,"type")="boolean_t"
	Set gtmtypfldindx("d_rm_struct","read_occurred")=66
	Set gtmtypes("d_rm_struct",67,"name")="d_rm_struct.input_encrypted"
	Set gtmtypes("d_rm_struct",67,"off")=368
	Set gtmtypes("d_rm_struct",67,"len")=4
	Set gtmtypes("d_rm_struct",67,"type")="boolean_t"
	Set gtmtypfldindx("d_rm_struct","input_encrypted")=67
	Set gtmtypes("d_rm_struct",68,"name")="d_rm_struct.output_encrypted"
	Set gtmtypes("d_rm_struct",68,"off")=372
	Set gtmtypes("d_rm_struct",68,"len")=4
	Set gtmtypes("d_rm_struct",68,"type")="boolean_t"
	Set gtmtypfldindx("d_rm_struct","output_encrypted")=68
	Set gtmtypes("d_rm_struct",69,"name")="d_rm_struct.input_iv"
	Set gtmtypes("d_rm_struct",69,"off")=376
	Set gtmtypes("d_rm_struct",69,"len")=16
	Set gtmtypes("d_rm_struct",69,"type")="mstr"
	Set gtmtypfldindx("d_rm_struct","input_iv")=69
	Set gtmtypes("d_rm_struct",70,"name")="d_rm_struct.input_iv.char_len"
	Set gtmtypes("d_rm_struct",70,"off")=376
	Set gtmtypes("d_rm_struct",70,"len")=4
	Set gtmtypes("d_rm_struct",70,"type")="unsigned-int"
	Set gtmtypfldindx("d_rm_struct","input_iv.char_len")=70
	Set gtmtypes("d_rm_struct",71,"name")="d_rm_struct.input_iv.len"
	Set gtmtypes("d_rm_struct",71,"off")=380
	Set gtmtypes("d_rm_struct",71,"len")=4
	Set gtmtypes("d_rm_struct",71,"type")="int"
	Set gtmtypfldindx("d_rm_struct","input_iv.len")=71
	Set gtmtypes("d_rm_struct",72,"name")="d_rm_struct.input_iv.addr"
	Set gtmtypes("d_rm_struct",72,"off")=384
	Set gtmtypes("d_rm_struct",72,"len")=8
	Set gtmtypes("d_rm_struct",72,"type")="addr"
	Set gtmtypfldindx("d_rm_struct","input_iv.addr")=72
	Set gtmtypes("d_rm_struct",73,"name")="d_rm_struct.output_iv"
	Set gtmtypes("d_rm_struct",73,"off")=392
	Set gtmtypes("d_rm_struct",73,"len")=16
	Set gtmtypes("d_rm_struct",73,"type")="mstr"
	Set gtmtypfldindx("d_rm_struct","output_iv")=73
	Set gtmtypes("d_rm_struct",74,"name")="d_rm_struct.output_iv.char_len"
	Set gtmtypes("d_rm_struct",74,"off")=392
	Set gtmtypes("d_rm_struct",74,"len")=4
	Set gtmtypes("d_rm_struct",74,"type")="unsigned-int"
	Set gtmtypfldindx("d_rm_struct","output_iv.char_len")=74
	Set gtmtypes("d_rm_struct",75,"name")="d_rm_struct.output_iv.len"
	Set gtmtypes("d_rm_struct",75,"off")=396
	Set gtmtypes("d_rm_struct",75,"len")=4
	Set gtmtypes("d_rm_struct",75,"type")="int"
	Set gtmtypfldindx("d_rm_struct","output_iv.len")=75
	Set gtmtypes("d_rm_struct",76,"name")="d_rm_struct.output_iv.addr"
	Set gtmtypes("d_rm_struct",76,"off")=400
	Set gtmtypes("d_rm_struct",76,"len")=8
	Set gtmtypes("d_rm_struct",76,"type")="addr"
	Set gtmtypfldindx("d_rm_struct","output_iv.addr")=76
	Set gtmtypes("d_rm_struct",77,"name")="d_rm_struct.input_key"
	Set gtmtypes("d_rm_struct",77,"off")=408
	Set gtmtypes("d_rm_struct",77,"len")=16
	Set gtmtypes("d_rm_struct",77,"type")="mstr"
	Set gtmtypfldindx("d_rm_struct","input_key")=77
	Set gtmtypes("d_rm_struct",78,"name")="d_rm_struct.input_key.char_len"
	Set gtmtypes("d_rm_struct",78,"off")=408
	Set gtmtypes("d_rm_struct",78,"len")=4
	Set gtmtypes("d_rm_struct",78,"type")="unsigned-int"
	Set gtmtypfldindx("d_rm_struct","input_key.char_len")=78
	Set gtmtypes("d_rm_struct",79,"name")="d_rm_struct.input_key.len"
	Set gtmtypes("d_rm_struct",79,"off")=412
	Set gtmtypes("d_rm_struct",79,"len")=4
	Set gtmtypes("d_rm_struct",79,"type")="int"
	Set gtmtypfldindx("d_rm_struct","input_key.len")=79
	Set gtmtypes("d_rm_struct",80,"name")="d_rm_struct.input_key.addr"
	Set gtmtypes("d_rm_struct",80,"off")=416
	Set gtmtypes("d_rm_struct",80,"len")=8
	Set gtmtypes("d_rm_struct",80,"type")="addr"
	Set gtmtypfldindx("d_rm_struct","input_key.addr")=80
	Set gtmtypes("d_rm_struct",81,"name")="d_rm_struct.output_key"
	Set gtmtypes("d_rm_struct",81,"off")=424
	Set gtmtypes("d_rm_struct",81,"len")=16
	Set gtmtypes("d_rm_struct",81,"type")="mstr"
	Set gtmtypfldindx("d_rm_struct","output_key")=81
	Set gtmtypes("d_rm_struct",82,"name")="d_rm_struct.output_key.char_len"
	Set gtmtypes("d_rm_struct",82,"off")=424
	Set gtmtypes("d_rm_struct",82,"len")=4
	Set gtmtypes("d_rm_struct",82,"type")="unsigned-int"
	Set gtmtypfldindx("d_rm_struct","output_key.char_len")=82
	Set gtmtypes("d_rm_struct",83,"name")="d_rm_struct.output_key.len"
	Set gtmtypes("d_rm_struct",83,"off")=428
	Set gtmtypes("d_rm_struct",83,"len")=4
	Set gtmtypes("d_rm_struct",83,"type")="int"
	Set gtmtypfldindx("d_rm_struct","output_key.len")=83
	Set gtmtypes("d_rm_struct",84,"name")="d_rm_struct.output_key.addr"
	Set gtmtypes("d_rm_struct",84,"off")=432
	Set gtmtypes("d_rm_struct",84,"len")=8
	Set gtmtypes("d_rm_struct",84,"type")="addr"
	Set gtmtypfldindx("d_rm_struct","output_key.addr")=84
	Set gtmtypes("d_rm_struct",85,"name")="d_rm_struct.input_cipher_handle"
	Set gtmtypes("d_rm_struct",85,"off")=440
	Set gtmtypes("d_rm_struct",85,"len")=8
	Set gtmtypes("d_rm_struct",85,"type")="addr"
	Set gtmtypfldindx("d_rm_struct","input_cipher_handle")=85
	Set gtmtypes("d_rm_struct",86,"name")="d_rm_struct.output_cipher_handle"
	Set gtmtypes("d_rm_struct",86,"off")=448
	Set gtmtypes("d_rm_struct",86,"len")=8
	Set gtmtypes("d_rm_struct",86,"type")="addr"
	Set gtmtypfldindx("d_rm_struct","output_cipher_handle")=86
	Set gtmtypes("d_rm_struct",87,"name")="d_rm_struct.ichset_utf16_variant"
	Set gtmtypes("d_rm_struct",87,"off")=456
	Set gtmtypes("d_rm_struct",87,"len")=4
	Set gtmtypes("d_rm_struct",87,"type")="int"
	Set gtmtypfldindx("d_rm_struct","ichset_utf16_variant")=87
	Set gtmtypes("d_rm_struct",88,"name")="d_rm_struct.ochset_utf16_variant"
	Set gtmtypes("d_rm_struct",88,"off")=460
	Set gtmtypes("d_rm_struct",88,"len")=4
	Set gtmtypes("d_rm_struct",88,"type")="int"
	Set gtmtypfldindx("d_rm_struct","ochset_utf16_variant")=88
	Set gtmtypes("d_rm_struct",89,"name")="d_rm_struct.fsblock_buffer_size"
	Set gtmtypes("d_rm_struct",89,"off")=464
	Set gtmtypes("d_rm_struct",89,"len")=4
	Set gtmtypes("d_rm_struct",89,"type")="unsigned-int"
	Set gtmtypfldindx("d_rm_struct","fsblock_buffer_size")=89
	Set gtmtypes("d_rm_struct",90,"name")="d_rm_struct.fsblock_buffer"
	Set gtmtypes("d_rm_struct",90,"off")=472
	Set gtmtypes("d_rm_struct",90,"len")=8
	Set gtmtypes("d_rm_struct",90,"type")="addr"
	Set gtmtypfldindx("d_rm_struct","fsblock_buffer")=90
	Set gtmtypes("d_rm_struct",91,"name")="d_rm_struct.crlastbuff"
	Set gtmtypes("d_rm_struct",91,"off")=480
	Set gtmtypes("d_rm_struct",91,"len")=4
	Set gtmtypes("d_rm_struct",91,"type")="boolean_t"
	Set gtmtypfldindx("d_rm_struct","crlastbuff")=91
	;
	Set gtmtypes("d_socket_struct")="struct"
	Set gtmtypes("d_socket_struct",0)=19
	Set gtmtypes("d_socket_struct","len")=96
	Set gtmtypes("d_socket_struct",1,"name")="d_socket_struct.sock_save_state"
	Set gtmtypes("d_socket_struct",1,"off")=0
	Set gtmtypes("d_socket_struct",1,"len")=48
	Set gtmtypes("d_socket_struct",1,"type")="socket_interrupt"
	Set gtmtypfldindx("d_socket_struct","sock_save_state")=1
	Set gtmtypes("d_socket_struct",2,"name")="d_socket_struct.sock_save_state.end_time"
	Set gtmtypes("d_socket_struct",2,"off")=0
	Set gtmtypes("d_socket_struct",2,"len")=16
	Set gtmtypes("d_socket_struct",2,"type")="ABS_TIME"
	Set gtmtypfldindx("d_socket_struct","sock_save_state.end_time")=2
	Set gtmtypes("d_socket_struct",3,"name")="d_socket_struct.sock_save_state.who_saved"
	Set gtmtypes("d_socket_struct",3,"off")=16
	Set gtmtypes("d_socket_struct",3,"len")=4
	Set gtmtypes("d_socket_struct",3,"type")="int"
	Set gtmtypfldindx("d_socket_struct","sock_save_state.who_saved")=3
	Set gtmtypes("d_socket_struct",4,"name")="d_socket_struct.sock_save_state.max_bufflen"
	Set gtmtypes("d_socket_struct",4,"off")=20
	Set gtmtypes("d_socket_struct",4,"len")=4
	Set gtmtypes("d_socket_struct",4,"type")="int"
	Set gtmtypfldindx("d_socket_struct","sock_save_state.max_bufflen")=4
	Set gtmtypes("d_socket_struct",5,"name")="d_socket_struct.sock_save_state.bytes_read"
	Set gtmtypes("d_socket_struct",5,"off")=24
	Set gtmtypes("d_socket_struct",5,"len")=4
	Set gtmtypes("d_socket_struct",5,"type")="int"
	Set gtmtypfldindx("d_socket_struct","sock_save_state.bytes_read")=5
	Set gtmtypes("d_socket_struct",6,"name")="d_socket_struct.sock_save_state.chars_read"
	Set gtmtypes("d_socket_struct",6,"off")=28
	Set gtmtypes("d_socket_struct",6,"len")=4
	Set gtmtypes("d_socket_struct",6,"type")="int"
	Set gtmtypfldindx("d_socket_struct","sock_save_state.chars_read")=6
	Set gtmtypes("d_socket_struct",7,"name")="d_socket_struct.sock_save_state.end_time_valid"
	Set gtmtypes("d_socket_struct",7,"off")=32
	Set gtmtypes("d_socket_struct",7,"len")=4
	Set gtmtypes("d_socket_struct",7,"type")="boolean_t"
	Set gtmtypfldindx("d_socket_struct","sock_save_state.end_time_valid")=7
	Set gtmtypes("d_socket_struct",8,"name")="d_socket_struct.sock_save_state.ibfsize_specified"
	Set gtmtypes("d_socket_struct",8,"off")=36
	Set gtmtypes("d_socket_struct",8,"len")=4
	Set gtmtypes("d_socket_struct",8,"type")="boolean_t"
	Set gtmtypfldindx("d_socket_struct","sock_save_state.ibfsize_specified")=8
	Set gtmtypes("d_socket_struct",9,"name")="d_socket_struct.sock_save_state.newdsocket"
	Set gtmtypes("d_socket_struct",9,"off")=40
	Set gtmtypes("d_socket_struct",9,"len")=8
	Set gtmtypes("d_socket_struct",9,"type")="addr"
	Set gtmtypfldindx("d_socket_struct","sock_save_state.newdsocket")=9
	Set gtmtypes("d_socket_struct",10,"name")="d_socket_struct.mupintr"
	Set gtmtypes("d_socket_struct",10,"off")=48
	Set gtmtypes("d_socket_struct",10,"len")=4
	Set gtmtypes("d_socket_struct",10,"type")="boolean_t"
	Set gtmtypfldindx("d_socket_struct","mupintr")=10
	Set gtmtypes("d_socket_struct",11,"name")="d_socket_struct.current_socket"
	Set gtmtypes("d_socket_struct",11,"off")=52
	Set gtmtypes("d_socket_struct",11,"len")=4
	Set gtmtypes("d_socket_struct",11,"type")="int"
	Set gtmtypfldindx("d_socket_struct","current_socket")=11
	Set gtmtypes("d_socket_struct",12,"name")="d_socket_struct.n_socket"
	Set gtmtypes("d_socket_struct",12,"off")=56
	Set gtmtypes("d_socket_struct",12,"len")=4
	Set gtmtypes("d_socket_struct",12,"type")="int"
	Set gtmtypfldindx("d_socket_struct","n_socket")=12
	Set gtmtypes("d_socket_struct",13,"name")="d_socket_struct.waitcycle"
	Set gtmtypes("d_socket_struct",13,"off")=60
	Set gtmtypes("d_socket_struct",13,"len")=4
	Set gtmtypes("d_socket_struct",13,"type")="unsigned-int"
	Set gtmtypfldindx("d_socket_struct","waitcycle")=13
	Set gtmtypes("d_socket_struct",14,"name")="d_socket_struct.ichset_specified"
	Set gtmtypes("d_socket_struct",14,"off")=64
	Set gtmtypes("d_socket_struct",14,"len")=4
	Set gtmtypes("d_socket_struct",14,"type")="boolean_t"
	Set gtmtypfldindx("d_socket_struct","ichset_specified")=14
	Set gtmtypes("d_socket_struct",15,"name")="d_socket_struct.ochset_specified"
	Set gtmtypes("d_socket_struct",15,"off")=68
	Set gtmtypes("d_socket_struct",15,"len")=4
	Set gtmtypes("d_socket_struct",15,"type")="boolean_t"
	Set gtmtypfldindx("d_socket_struct","ochset_specified")=15
	Set gtmtypes("d_socket_struct",16,"name")="d_socket_struct.ichset_utf16_variant"
	Set gtmtypes("d_socket_struct",16,"off")=72
	Set gtmtypes("d_socket_struct",16,"len")=4
	Set gtmtypes("d_socket_struct",16,"type")="int"
	Set gtmtypfldindx("d_socket_struct","ichset_utf16_variant")=16
	Set gtmtypes("d_socket_struct",17,"name")="d_socket_struct.ochset_utf16_variant"
	Set gtmtypes("d_socket_struct",17,"off")=76
	Set gtmtypes("d_socket_struct",17,"len")=4
	Set gtmtypes("d_socket_struct",17,"type")="int"
	Set gtmtypfldindx("d_socket_struct","ochset_utf16_variant")=17
	Set gtmtypes("d_socket_struct",18,"name")="d_socket_struct.iod"
	Set gtmtypes("d_socket_struct",18,"off")=80
	Set gtmtypes("d_socket_struct",18,"len")=8
	Set gtmtypes("d_socket_struct",18,"type")="addr"
	Set gtmtypfldindx("d_socket_struct","iod")=18
	Set gtmtypes("d_socket_struct",19,"name")="d_socket_struct.gtm_socket"
	Set gtmtypes("d_socket_struct",19,"off")=88
	Set gtmtypes("d_socket_struct",19,"len")=8
	Set gtmtypes("d_socket_struct",19,"type")="addr"
	Set gtmtypfldindx("d_socket_struct","gtm_socket")=19
	;
	Set gtmtypes("d_tt_struct")="struct"
	Set gtmtypes("d_tt_struct",0)=48
	Set gtmtypes("d_tt_struct","len")=280
	Set gtmtypes("d_tt_struct",1,"name")="d_tt_struct.in_buf_sz"
	Set gtmtypes("d_tt_struct",1,"off")=0
	Set gtmtypes("d_tt_struct",1,"len")=4
	Set gtmtypes("d_tt_struct",1,"type")="unsigned-int"
	Set gtmtypfldindx("d_tt_struct","in_buf_sz")=1
	Set gtmtypes("d_tt_struct",2,"name")="d_tt_struct.ext_cap"
	Set gtmtypes("d_tt_struct",2,"off")=4
	Set gtmtypes("d_tt_struct",2,"len")=4
	Set gtmtypes("d_tt_struct",2,"type")="unsigned-int"
	Set gtmtypfldindx("d_tt_struct","ext_cap")=2
	Set gtmtypes("d_tt_struct",3,"name")="d_tt_struct.enbld_outofbands"
	Set gtmtypes("d_tt_struct",3,"off")=8
	Set gtmtypes("d_tt_struct",3,"len")=8
	Set gtmtypes("d_tt_struct",3,"type")="io_terminator"
	Set gtmtypfldindx("d_tt_struct","enbld_outofbands")=3
	Set gtmtypes("d_tt_struct",4,"name")="d_tt_struct.enbld_outofbands.x"
	Set gtmtypes("d_tt_struct",4,"off")=8
	Set gtmtypes("d_tt_struct",4,"len")=4
	Set gtmtypes("d_tt_struct",4,"type")="unsigned-int"
	Set gtmtypfldindx("d_tt_struct","enbld_outofbands.x")=4
	Set gtmtypes("d_tt_struct",5,"name")="d_tt_struct.enbld_outofbands.mask"
	Set gtmtypes("d_tt_struct",5,"off")=12
	Set gtmtypes("d_tt_struct",5,"len")=4
	Set gtmtypes("d_tt_struct",5,"type")="unsigned-int"
	Set gtmtypfldindx("d_tt_struct","enbld_outofbands.mask")=5
	Set gtmtypes("d_tt_struct",6,"name")="d_tt_struct.term_ctrl"
	Set gtmtypes("d_tt_struct",6,"off")=16
	Set gtmtypes("d_tt_struct",6,"len")=4
	Set gtmtypes("d_tt_struct",6,"type")="unsigned-int"
	Set gtmtypfldindx("d_tt_struct","term_ctrl")=6
	Set gtmtypes("d_tt_struct",7,"name")="d_tt_struct.mask_term"
	Set gtmtypes("d_tt_struct",7,"off")=20
	Set gtmtypes("d_tt_struct",7,"len")=32
	Set gtmtypes("d_tt_struct",7,"type")="io_termmask"
	Set gtmtypfldindx("d_tt_struct","mask_term")=7
	Set gtmtypes("d_tt_struct",8,"name")="d_tt_struct.mask_term.mask"
	Set gtmtypes("d_tt_struct",8,"off")=20
	Set gtmtypes("d_tt_struct",8,"len")=32
	Set gtmtypes("d_tt_struct",8,"type")="unsigned-int"
	Set gtmtypfldindx("d_tt_struct","mask_term.mask")=8
	Set gtmtypes("d_tt_struct",8,"dim")=8
	Set gtmtypes("d_tt_struct",9,"name")="d_tt_struct.fildes"
	Set gtmtypes("d_tt_struct",9,"off")=52
	Set gtmtypes("d_tt_struct",9,"len")=4
	Set gtmtypes("d_tt_struct",9,"type")="int"
	Set gtmtypfldindx("d_tt_struct","fildes")=9
	Set gtmtypes("d_tt_struct",10,"name")="d_tt_struct.ttio_struct"
	Set gtmtypes("d_tt_struct",10,"off")=56
	Set gtmtypes("d_tt_struct",10,"len")=8
	Set gtmtypes("d_tt_struct",10,"type")="addr"
	Set gtmtypfldindx("d_tt_struct","ttio_struct")=10
	Set gtmtypes("d_tt_struct",11,"name")="d_tt_struct.tt_state_save"
	Set gtmtypes("d_tt_struct",11,"off")=64
	Set gtmtypes("d_tt_struct",11,"len")=136
	Set gtmtypes("d_tt_struct",11,"type")="tt_interrupt"
	Set gtmtypfldindx("d_tt_struct","tt_state_save")=11
	Set gtmtypes("d_tt_struct",12,"name")="d_tt_struct.tt_state_save.who_saved"
	Set gtmtypes("d_tt_struct",12,"off")=64
	Set gtmtypes("d_tt_struct",12,"len")=4
	Set gtmtypes("d_tt_struct",12,"type")="int"
	Set gtmtypfldindx("d_tt_struct","tt_state_save.who_saved")=12
	Set gtmtypes("d_tt_struct",13,"name")="d_tt_struct.tt_state_save.buffer_start"
	Set gtmtypes("d_tt_struct",13,"off")=72
	Set gtmtypes("d_tt_struct",13,"len")=8
	Set gtmtypes("d_tt_struct",13,"type")="addr"
	Set gtmtypfldindx("d_tt_struct","tt_state_save.buffer_start")=13
	Set gtmtypes("d_tt_struct",14,"name")="d_tt_struct.tt_state_save.buffer_32_start"
	Set gtmtypes("d_tt_struct",14,"off")=80
	Set gtmtypes("d_tt_struct",14,"len")=8
	Set gtmtypes("d_tt_struct",14,"type")="addr"
	Set gtmtypfldindx("d_tt_struct","tt_state_save.buffer_32_start")=14
	Set gtmtypes("d_tt_struct",15,"name")="d_tt_struct.tt_state_save.utf8_more"
	Set gtmtypes("d_tt_struct",15,"off")=88
	Set gtmtypes("d_tt_struct",15,"len")=4
	Set gtmtypes("d_tt_struct",15,"type")="int"
	Set gtmtypfldindx("d_tt_struct","tt_state_save.utf8_more")=15
	Set gtmtypes("d_tt_struct",16,"name")="d_tt_struct.tt_state_save.dx"
	Set gtmtypes("d_tt_struct",16,"off")=92
	Set gtmtypes("d_tt_struct",16,"len")=4
	Set gtmtypes("d_tt_struct",16,"type")="int"
	Set gtmtypfldindx("d_tt_struct","tt_state_save.dx")=16
	Set gtmtypes("d_tt_struct",17,"name")="d_tt_struct.tt_state_save.dx_start"
	Set gtmtypes("d_tt_struct",17,"off")=96
	Set gtmtypes("d_tt_struct",17,"len")=4
	Set gtmtypes("d_tt_struct",17,"type")="int"
	Set gtmtypfldindx("d_tt_struct","tt_state_save.dx_start")=17
	Set gtmtypes("d_tt_struct",18,"name")="d_tt_struct.tt_state_save.dx_instr"
	Set gtmtypes("d_tt_struct",18,"off")=100
	Set gtmtypes("d_tt_struct",18,"len")=4
	Set gtmtypes("d_tt_struct",18,"type")="int"
	Set gtmtypfldindx("d_tt_struct","tt_state_save.dx_instr")=18
	Set gtmtypes("d_tt_struct",19,"name")="d_tt_struct.tt_state_save.dx_outlen"
	Set gtmtypes("d_tt_struct",19,"off")=104
	Set gtmtypes("d_tt_struct",19,"len")=4
	Set gtmtypes("d_tt_struct",19,"type")="int"
	Set gtmtypfldindx("d_tt_struct","tt_state_save.dx_outlen")=19
	Set gtmtypes("d_tt_struct",20,"name")="d_tt_struct.tt_state_save.instr"
	Set gtmtypes("d_tt_struct",20,"off")=108
	Set gtmtypes("d_tt_struct",20,"len")=4
	Set gtmtypes("d_tt_struct",20,"type")="int"
	Set gtmtypfldindx("d_tt_struct","tt_state_save.instr")=20
	Set gtmtypes("d_tt_struct",21,"name")="d_tt_struct.tt_state_save.outlen"
	Set gtmtypes("d_tt_struct",21,"off")=112
	Set gtmtypes("d_tt_struct",21,"len")=4
	Set gtmtypes("d_tt_struct",21,"type")="int"
	Set gtmtypfldindx("d_tt_struct","tt_state_save.outlen")=21
	Set gtmtypes("d_tt_struct",22,"name")="d_tt_struct.tt_state_save.index"
	Set gtmtypes("d_tt_struct",22,"off")=116
	Set gtmtypes("d_tt_struct",22,"len")=4
	Set gtmtypes("d_tt_struct",22,"type")="int"
	Set gtmtypfldindx("d_tt_struct","tt_state_save.index")=22
	Set gtmtypes("d_tt_struct",23,"name")="d_tt_struct.tt_state_save.cl"
	Set gtmtypes("d_tt_struct",23,"off")=120
	Set gtmtypes("d_tt_struct",23,"len")=4
	Set gtmtypes("d_tt_struct",23,"type")="int"
	Set gtmtypfldindx("d_tt_struct","tt_state_save.cl")=23
	Set gtmtypes("d_tt_struct",24,"name")="d_tt_struct.tt_state_save.length"
	Set gtmtypes("d_tt_struct",24,"off")=124
	Set gtmtypes("d_tt_struct",24,"len")=4
	Set gtmtypes("d_tt_struct",24,"type")="int"
	Set gtmtypfldindx("d_tt_struct","tt_state_save.length")=24
	Set gtmtypes("d_tt_struct",25,"name")="d_tt_struct.tt_state_save.exp_length"
	Set gtmtypes("d_tt_struct",25,"off")=128
	Set gtmtypes("d_tt_struct",25,"len")=4
	Set gtmtypes("d_tt_struct",25,"type")="int"
	Set gtmtypfldindx("d_tt_struct","tt_state_save.exp_length")=25
	Set gtmtypes("d_tt_struct",26,"name")="d_tt_struct.tt_state_save.insert_mode"
	Set gtmtypes("d_tt_struct",26,"off")=132
	Set gtmtypes("d_tt_struct",26,"len")=4
	Set gtmtypes("d_tt_struct",26,"type")="boolean_t"
	Set gtmtypfldindx("d_tt_struct","tt_state_save.insert_mode")=26
	Set gtmtypes("d_tt_struct",27,"name")="d_tt_struct.tt_state_save.end_time"
	Set gtmtypes("d_tt_struct",27,"off")=136
	Set gtmtypes("d_tt_struct",27,"len")=16
	Set gtmtypes("d_tt_struct",27,"type")="ABS_TIME"
	Set gtmtypfldindx("d_tt_struct","tt_state_save.end_time")=27
	Set gtmtypes("d_tt_struct",28,"name")="d_tt_struct.tt_state_save.more_ptr"
	Set gtmtypes("d_tt_struct",28,"off")=152
	Set gtmtypes("d_tt_struct",28,"len")=8
	Set gtmtypes("d_tt_struct",28,"type")="addr"
	Set gtmtypfldindx("d_tt_struct","tt_state_save.more_ptr")=28
	Set gtmtypes("d_tt_struct",29,"name")="d_tt_struct.tt_state_save.zb_ptr"
	Set gtmtypes("d_tt_struct",29,"off")=160
	Set gtmtypes("d_tt_struct",29,"len")=8
	Set gtmtypes("d_tt_struct",29,"type")="addr"
	Set gtmtypfldindx("d_tt_struct","tt_state_save.zb_ptr")=29
	Set gtmtypes("d_tt_struct",30,"name")="d_tt_struct.tt_state_save.zb_top"
	Set gtmtypes("d_tt_struct",30,"off")=168
	Set gtmtypes("d_tt_struct",30,"len")=8
	Set gtmtypes("d_tt_struct",30,"type")="addr"
	Set gtmtypfldindx("d_tt_struct","tt_state_save.zb_top")=30
	Set gtmtypes("d_tt_struct",31,"name")="d_tt_struct.tt_state_save.escape_length"
	Set gtmtypes("d_tt_struct",31,"off")=176
	Set gtmtypes("d_tt_struct",31,"len")=2
	Set gtmtypes("d_tt_struct",31,"type")="unsigned-short"
	Set gtmtypfldindx("d_tt_struct","tt_state_save.escape_length")=31
	Set gtmtypes("d_tt_struct",32,"name")="d_tt_struct.tt_state_save.escape_sequence"
	Set gtmtypes("d_tt_struct",32,"off")=178
	Set gtmtypes("d_tt_struct",32,"len")=16
	Set gtmtypes("d_tt_struct",32,"type")="unsigned-char"
	Set gtmtypfldindx("d_tt_struct","tt_state_save.escape_sequence")=32
	Set gtmtypes("d_tt_struct",33,"name")="d_tt_struct.tt_state_save.more_buf"
	Set gtmtypes("d_tt_struct",33,"off")=194
	Set gtmtypes("d_tt_struct",33,"len")=5
	Set gtmtypes("d_tt_struct",33,"type")="unsigned-char"
	Set gtmtypfldindx("d_tt_struct","tt_state_save.more_buf")=33
	Set gtmtypes("d_tt_struct",34,"name")="d_tt_struct.mupintr"
	Set gtmtypes("d_tt_struct",34,"off")=200
	Set gtmtypes("d_tt_struct",34,"len")=4
	Set gtmtypes("d_tt_struct",34,"type")="boolean_t"
	Set gtmtypfldindx("d_tt_struct","mupintr")=34
	Set gtmtypes("d_tt_struct",35,"name")="d_tt_struct.ttybuff"
	Set gtmtypes("d_tt_struct",35,"off")=208
	Set gtmtypes("d_tt_struct",35,"len")=8
	Set gtmtypes("d_tt_struct",35,"type")="addr"
	Set gtmtypfldindx("d_tt_struct","ttybuff")=35
	Set gtmtypes("d_tt_struct",36,"name")="d_tt_struct.tbuffp"
	Set gtmtypes("d_tt_struct",36,"off")=216
	Set gtmtypes("d_tt_struct",36,"len")=8
	Set gtmtypes("d_tt_struct",36,"type")="addr"
	Set gtmtypfldindx("d_tt_struct","tbuffp")=36
	Set gtmtypes("d_tt_struct",37,"name")="d_tt_struct.timer_set"
	Set gtmtypes("d_tt_struct",37,"off")=224
	Set gtmtypes("d_tt_struct",37,"len")=4
	Set gtmtypes("d_tt_struct",37,"type")="boolean_t"
	Set gtmtypfldindx("d_tt_struct","timer_set")=37
	Set gtmtypes("d_tt_struct",38,"name")="d_tt_struct.write_active"
	Set gtmtypes("d_tt_struct",38,"off")=228
	Set gtmtypes("d_tt_struct",38,"len")=4
	Set gtmtypes("d_tt_struct",38,"type")="boolean_t"
	Set gtmtypfldindx("d_tt_struct","write_active")=38
	Set gtmtypes("d_tt_struct",39,"name")="d_tt_struct.canonical"
	Set gtmtypes("d_tt_struct",39,"off")=232
	Set gtmtypes("d_tt_struct",39,"len")=4
	Set gtmtypes("d_tt_struct",39,"type")="boolean_t"
	Set gtmtypfldindx("d_tt_struct","canonical")=39
	Set gtmtypes("d_tt_struct",40,"name")="d_tt_struct.recall_buff"
	Set gtmtypes("d_tt_struct",40,"off")=240
	Set gtmtypes("d_tt_struct",40,"len")=16
	Set gtmtypes("d_tt_struct",40,"type")="mstr"
	Set gtmtypfldindx("d_tt_struct","recall_buff")=40
	Set gtmtypes("d_tt_struct",41,"name")="d_tt_struct.recall_buff.char_len"
	Set gtmtypes("d_tt_struct",41,"off")=240
	Set gtmtypes("d_tt_struct",41,"len")=4
	Set gtmtypes("d_tt_struct",41,"type")="unsigned-int"
	Set gtmtypfldindx("d_tt_struct","recall_buff.char_len")=41
	Set gtmtypes("d_tt_struct",42,"name")="d_tt_struct.recall_buff.len"
	Set gtmtypes("d_tt_struct",42,"off")=244
	Set gtmtypes("d_tt_struct",42,"len")=4
	Set gtmtypes("d_tt_struct",42,"type")="int"
	Set gtmtypfldindx("d_tt_struct","recall_buff.len")=42
	Set gtmtypes("d_tt_struct",43,"name")="d_tt_struct.recall_buff.addr"
	Set gtmtypes("d_tt_struct",43,"off")=248
	Set gtmtypes("d_tt_struct",43,"len")=8
	Set gtmtypes("d_tt_struct",43,"type")="addr"
	Set gtmtypfldindx("d_tt_struct","recall_buff.addr")=43
	Set gtmtypes("d_tt_struct",44,"name")="d_tt_struct.recall_size"
	Set gtmtypes("d_tt_struct",44,"off")=256
	Set gtmtypes("d_tt_struct",44,"len")=4
	Set gtmtypes("d_tt_struct",44,"type")="int"
	Set gtmtypfldindx("d_tt_struct","recall_size")=44
	Set gtmtypes("d_tt_struct",45,"name")="d_tt_struct.recall_width"
	Set gtmtypes("d_tt_struct",45,"off")=260
	Set gtmtypes("d_tt_struct",45,"len")=4
	Set gtmtypes("d_tt_struct",45,"type")="int"
	Set gtmtypfldindx("d_tt_struct","recall_width")=45
	Set gtmtypes("d_tt_struct",46,"name")="d_tt_struct.discard_lf"
	Set gtmtypes("d_tt_struct",46,"off")=264
	Set gtmtypes("d_tt_struct",46,"len")=4
	Set gtmtypes("d_tt_struct",46,"type")="boolean_t"
	Set gtmtypfldindx("d_tt_struct","discard_lf")=46
	Set gtmtypes("d_tt_struct",47,"name")="d_tt_struct.default_mask_term"
	Set gtmtypes("d_tt_struct",47,"off")=268
	Set gtmtypes("d_tt_struct",47,"len")=4
	Set gtmtypes("d_tt_struct",47,"type")="boolean_t"
	Set gtmtypfldindx("d_tt_struct","default_mask_term")=47
	Set gtmtypes("d_tt_struct",48,"name")="d_tt_struct.done_1st_read"
	Set gtmtypes("d_tt_struct",48,"off")=272
	Set gtmtypes("d_tt_struct",48,"len")=4
	Set gtmtypes("d_tt_struct",48,"type")="boolean_t"
	Set gtmtypfldindx("d_tt_struct","done_1st_read")=48
	;
	Set gtmtypes("d_us_struct")="struct"
	Set gtmtypes("d_us_struct",0)=1
	Set gtmtypes("d_us_struct","len")=8
	Set gtmtypes("d_us_struct",1,"name")="d_us_struct.disp"
	Set gtmtypes("d_us_struct",1,"off")=0
	Set gtmtypes("d_us_struct",1,"len")=8
	Set gtmtypes("d_us_struct",1,"type")="addr"
	Set gtmtypfldindx("d_us_struct","disp")=1
	;
	Set gtmtypes("date_time")="struct"
	Set gtmtypes("date_time",0)=2
	Set gtmtypes("date_time","len")=8
	Set gtmtypes("date_time",1,"name")="date_time.low"
	Set gtmtypes("date_time",1,"off")=0
	Set gtmtypes("date_time",1,"len")=4
	Set gtmtypes("date_time",1,"type")="unsigned-int"
	Set gtmtypfldindx("date_time","low")=1
	Set gtmtypes("date_time",2,"name")="date_time.high"
	Set gtmtypes("date_time",2,"off")=4
	Set gtmtypes("date_time",2,"len")=4
	Set gtmtypes("date_time",2,"type")="unsigned-int"
	Set gtmtypfldindx("date_time","high")=2
	;
	Set gtmtypes("db_csh_acct_rec")="struct"
	Set gtmtypes("db_csh_acct_rec",0)=2
	Set gtmtypes("db_csh_acct_rec","len")=8
	Set gtmtypes("db_csh_acct_rec",1,"name")="db_csh_acct_rec.curr_count"
	Set gtmtypes("db_csh_acct_rec",1,"off")=0
	Set gtmtypes("db_csh_acct_rec",1,"len")=4
	Set gtmtypes("db_csh_acct_rec",1,"type")="int"
	Set gtmtypfldindx("db_csh_acct_rec","curr_count")=1
	Set gtmtypes("db_csh_acct_rec",2,"name")="db_csh_acct_rec.cumul_count"
	Set gtmtypes("db_csh_acct_rec",2,"off")=4
	Set gtmtypes("db_csh_acct_rec",2,"len")=4
	Set gtmtypes("db_csh_acct_rec",2,"type")="int"
	Set gtmtypfldindx("db_csh_acct_rec","cumul_count")=2
	;
	Set gtmtypes("dbc_gv_key")="struct"
	Set gtmtypes("dbc_gv_key",0)=4
	Set gtmtypes("dbc_gv_key","len")=16
	Set gtmtypes("dbc_gv_key",1,"name")="dbc_gv_key.top"
	Set gtmtypes("dbc_gv_key",1,"off")=0
	Set gtmtypes("dbc_gv_key",1,"len")=4
	Set gtmtypes("dbc_gv_key",1,"type")="unsigned-int"
	Set gtmtypfldindx("dbc_gv_key","top")=1
	Set gtmtypes("dbc_gv_key",2,"name")="dbc_gv_key.end"
	Set gtmtypes("dbc_gv_key",2,"off")=4
	Set gtmtypes("dbc_gv_key",2,"len")=4
	Set gtmtypes("dbc_gv_key",2,"type")="unsigned-int"
	Set gtmtypfldindx("dbc_gv_key","end")=2
	Set gtmtypes("dbc_gv_key",3,"name")="dbc_gv_key.gvn_len"
	Set gtmtypes("dbc_gv_key",3,"off")=8
	Set gtmtypes("dbc_gv_key",3,"len")=4
	Set gtmtypes("dbc_gv_key",3,"type")="unsigned-int"
	Set gtmtypfldindx("dbc_gv_key","gvn_len")=3
	Set gtmtypes("dbc_gv_key",4,"name")="dbc_gv_key.base"
	Set gtmtypes("dbc_gv_key",4,"off")=12
	Set gtmtypes("dbc_gv_key",4,"len")=1
	Set gtmtypes("dbc_gv_key",4,"type")="unsigned-char"
	Set gtmtypfldindx("dbc_gv_key","base")=4
	;
	Set gtmtypes("dbc_inserted_rec")="struct"
	Set gtmtypes("dbc_inserted_rec",0)=2
	Set gtmtypes("dbc_inserted_rec","len")=16
	Set gtmtypes("dbc_inserted_rec",1,"name")="dbc_inserted_rec.ins_key"
	Set gtmtypes("dbc_inserted_rec",1,"off")=0
	Set gtmtypes("dbc_inserted_rec",1,"len")=8
	Set gtmtypes("dbc_inserted_rec",1,"type")="addr"
	Set gtmtypfldindx("dbc_inserted_rec","ins_key")=1
	Set gtmtypes("dbc_inserted_rec",2,"name")="dbc_inserted_rec.blk_id"
	Set gtmtypes("dbc_inserted_rec",2,"off")=8
	Set gtmtypes("dbc_inserted_rec",2,"len")=4
	Set gtmtypes("dbc_inserted_rec",2,"type")="int"
	Set gtmtypfldindx("dbc_inserted_rec","blk_id")=2
	;
	Set gtmtypes("ddp_announce_msg_t")="struct"
	Set gtmtypes("ddp_announce_msg_t",0)=19
	Set gtmtypes("ddp_announce_msg_t","len")=106
	Set gtmtypes("ddp_announce_msg_t",1,"name")="ddp_announce_msg_t.filler0"
	Set gtmtypes("ddp_announce_msg_t",1,"off")=0
	Set gtmtypes("ddp_announce_msg_t",1,"len")=2
	Set gtmtypes("ddp_announce_msg_t",1,"type")="unsigned-short"
	Set gtmtypfldindx("ddp_announce_msg_t","filler0")=1
	Set gtmtypes("ddp_announce_msg_t",2,"name")="ddp_announce_msg_t.code"
	Set gtmtypes("ddp_announce_msg_t",2,"off")=2
	Set gtmtypes("ddp_announce_msg_t",2,"len")=2
	Set gtmtypes("ddp_announce_msg_t",2,"type")="unsigned-char"
	Set gtmtypfldindx("ddp_announce_msg_t","code")=2
	Set gtmtypes("ddp_announce_msg_t",3,"name")="ddp_announce_msg_t.ether_addr"
	Set gtmtypes("ddp_announce_msg_t",3,"off")=4
	Set gtmtypes("ddp_announce_msg_t",3,"len")=6
	Set gtmtypes("ddp_announce_msg_t",3,"type")="unsigned-char"
	Set gtmtypfldindx("ddp_announce_msg_t","ether_addr")=3
	Set gtmtypes("ddp_announce_msg_t",4,"name")="ddp_announce_msg_t.circuit_name"
	Set gtmtypes("ddp_announce_msg_t",4,"off")=10
	Set gtmtypes("ddp_announce_msg_t",4,"len")=2
	Set gtmtypes("ddp_announce_msg_t",4,"type")="unsigned-short"
	Set gtmtypfldindx("ddp_announce_msg_t","circuit_name")=4
	Set gtmtypes("ddp_announce_msg_t",5,"name")="ddp_announce_msg_t.filler1"
	Set gtmtypes("ddp_announce_msg_t",5,"off")=12
	Set gtmtypes("ddp_announce_msg_t",5,"len")=2
	Set gtmtypes("ddp_announce_msg_t",5,"type")="unsigned-short"
	Set gtmtypfldindx("ddp_announce_msg_t","filler1")=5
	Set gtmtypes("ddp_announce_msg_t",6,"name")="ddp_announce_msg_t.filler2"
	Set gtmtypes("ddp_announce_msg_t",6,"off")=14
	Set gtmtypes("ddp_announce_msg_t",6,"len")=2
	Set gtmtypes("ddp_announce_msg_t",6,"type")="unsigned-short"
	Set gtmtypfldindx("ddp_announce_msg_t","filler2")=6
	Set gtmtypes("ddp_announce_msg_t",7,"name")="ddp_announce_msg_t.max_job_no"
	Set gtmtypes("ddp_announce_msg_t",7,"off")=16
	Set gtmtypes("ddp_announce_msg_t",7,"len")=2
	Set gtmtypes("ddp_announce_msg_t",7,"type")="unsigned-short"
	Set gtmtypfldindx("ddp_announce_msg_t","max_job_no")=7
	Set gtmtypes("ddp_announce_msg_t",8,"name")="ddp_announce_msg_t.group_mask"
	Set gtmtypes("ddp_announce_msg_t",8,"off")=18
	Set gtmtypes("ddp_announce_msg_t",8,"len")=2
	Set gtmtypes("ddp_announce_msg_t",8,"type")="unsigned-short"
	Set gtmtypfldindx("ddp_announce_msg_t","group_mask")=8
	Set gtmtypes("ddp_announce_msg_t",9,"name")="ddp_announce_msg_t.advertise_interval"
	Set gtmtypes("ddp_announce_msg_t",9,"off")=20
	Set gtmtypes("ddp_announce_msg_t",9,"len")=1
	Set gtmtypes("ddp_announce_msg_t",9,"type")="unsigned-char"
	Set gtmtypfldindx("ddp_announce_msg_t","advertise_interval")=9
	Set gtmtypes("ddp_announce_msg_t",10,"name")="ddp_announce_msg_t.max_request_credits"
	Set gtmtypes("ddp_announce_msg_t",10,"off")=21
	Set gtmtypes("ddp_announce_msg_t",10,"len")=1
	Set gtmtypes("ddp_announce_msg_t",10,"type")="unsigned-char"
	Set gtmtypfldindx("ddp_announce_msg_t","max_request_credits")=10
	Set gtmtypes("ddp_announce_msg_t",11,"name")="ddp_announce_msg_t.cpu_type"
	Set gtmtypes("ddp_announce_msg_t",11,"off")=22
	Set gtmtypes("ddp_announce_msg_t",11,"len")=1
	Set gtmtypes("ddp_announce_msg_t",11,"type")="unsigned-char"
	Set gtmtypfldindx("ddp_announce_msg_t","cpu_type")=11
	Set gtmtypes("ddp_announce_msg_t",12,"name")="ddp_announce_msg_t.version"
	Set gtmtypes("ddp_announce_msg_t",12,"off")=23
	Set gtmtypes("ddp_announce_msg_t",12,"len")=1
	Set gtmtypes("ddp_announce_msg_t",12,"type")="unsigned-char"
	Set gtmtypfldindx("ddp_announce_msg_t","version")=12
	Set gtmtypes("ddp_announce_msg_t",13,"name")="ddp_announce_msg_t.cpu_load_rating"
	Set gtmtypes("ddp_announce_msg_t",13,"off")=24
	Set gtmtypes("ddp_announce_msg_t",13,"len")=1
	Set gtmtypes("ddp_announce_msg_t",13,"type")="unsigned-char"
	Set gtmtypfldindx("ddp_announce_msg_t","cpu_load_rating")=13
	Set gtmtypes("ddp_announce_msg_t",14,"name")="ddp_announce_msg_t.proto_version"
	Set gtmtypes("ddp_announce_msg_t",14,"off")=25
	Set gtmtypes("ddp_announce_msg_t",14,"len")=1
	Set gtmtypes("ddp_announce_msg_t",14,"type")="unsigned-char"
	Set gtmtypfldindx("ddp_announce_msg_t","proto_version")=14
	Set gtmtypes("ddp_announce_msg_t",15,"name")="ddp_announce_msg_t.node_status"
	Set gtmtypes("ddp_announce_msg_t",15,"off")=26
	Set gtmtypes("ddp_announce_msg_t",15,"len")=1
	Set gtmtypes("ddp_announce_msg_t",15,"type")="unsigned-char"
	Set gtmtypfldindx("ddp_announce_msg_t","node_status")=15
	Set gtmtypes("ddp_announce_msg_t",16,"name")="ddp_announce_msg_t.autoconfigure_version"
	Set gtmtypes("ddp_announce_msg_t",16,"off")=27
	Set gtmtypes("ddp_announce_msg_t",16,"len")=1
	Set gtmtypes("ddp_announce_msg_t",16,"type")="unsigned-char"
	Set gtmtypfldindx("ddp_announce_msg_t","autoconfigure_version")=16
	Set gtmtypes("ddp_announce_msg_t",17,"name")="ddp_announce_msg_t.volset"
	Set gtmtypes("ddp_announce_msg_t",17,"off")=28
	Set gtmtypes("ddp_announce_msg_t",17,"len")=32
	Set gtmtypes("ddp_announce_msg_t",17,"type")="unsigned-short"
	Set gtmtypfldindx("ddp_announce_msg_t","volset")=17
	Set gtmtypes("ddp_announce_msg_t",17,"dim")=16
	Set gtmtypes("ddp_announce_msg_t",18,"name")="ddp_announce_msg_t.filler3"
	Set gtmtypes("ddp_announce_msg_t",18,"off")=60
	Set gtmtypes("ddp_announce_msg_t",18,"len")=44
	Set gtmtypes("ddp_announce_msg_t",18,"type")="unsigned-char"
	Set gtmtypfldindx("ddp_announce_msg_t","filler3")=18
	Set gtmtypes("ddp_announce_msg_t",19,"name")="ddp_announce_msg_t.terminator"
	Set gtmtypes("ddp_announce_msg_t",19,"off")=104
	Set gtmtypes("ddp_announce_msg_t",19,"len")=1
	Set gtmtypes("ddp_announce_msg_t",19,"type")="unsigned-char"
	Set gtmtypfldindx("ddp_announce_msg_t","terminator")=19
	;
	Set gtmtypes("ddp_global_request_t")="struct"
	Set gtmtypes("ddp_global_request_t",0)=6
	Set gtmtypes("ddp_global_request_t","len")=10
	Set gtmtypes("ddp_global_request_t",1,"name")="ddp_global_request_t.naked_size"
	Set gtmtypes("ddp_global_request_t",1,"off")=0
	Set gtmtypes("ddp_global_request_t",1,"len")=1
	Set gtmtypes("ddp_global_request_t",1,"type")="unsigned-char"
	Set gtmtypfldindx("ddp_global_request_t","naked_size")=1
	Set gtmtypes("ddp_global_request_t",2,"name")="ddp_global_request_t.uci"
	Set gtmtypes("ddp_global_request_t",2,"off")=2
	Set gtmtypes("ddp_global_request_t",2,"len")=2
	Set gtmtypes("ddp_global_request_t",2,"type")="unsigned-short"
	Set gtmtypfldindx("ddp_global_request_t","uci")=2
	Set gtmtypes("ddp_global_request_t",3,"name")="ddp_global_request_t.vol"
	Set gtmtypes("ddp_global_request_t",3,"off")=4
	Set gtmtypes("ddp_global_request_t",3,"len")=2
	Set gtmtypes("ddp_global_request_t",3,"type")="unsigned-short"
	Set gtmtypfldindx("ddp_global_request_t","vol")=3
	Set gtmtypes("ddp_global_request_t",4,"name")="ddp_global_request_t.global_type"
	Set gtmtypes("ddp_global_request_t",4,"off")=6
	Set gtmtypes("ddp_global_request_t",4,"len")=1
	Set gtmtypes("ddp_global_request_t",4,"type")="unsigned-char"
	Set gtmtypfldindx("ddp_global_request_t","global_type")=4
	Set gtmtypes("ddp_global_request_t",5,"name")="ddp_global_request_t.global_len"
	Set gtmtypes("ddp_global_request_t",5,"off")=7
	Set gtmtypes("ddp_global_request_t",5,"len")=1
	Set gtmtypes("ddp_global_request_t",5,"type")="unsigned-char"
	Set gtmtypfldindx("ddp_global_request_t","global_len")=5
	Set gtmtypes("ddp_global_request_t",6,"name")="ddp_global_request_t.global"
	Set gtmtypes("ddp_global_request_t",6,"off")=8
	Set gtmtypes("ddp_global_request_t",6,"len")=1
	Set gtmtypes("ddp_global_request_t",6,"type")="unsigned-char"
	Set gtmtypfldindx("ddp_global_request_t","global")=6
	;
	Set gtmtypes("ddp_hdr_t")="struct"
	Set gtmtypes("ddp_hdr_t",0)=11
	Set gtmtypes("ddp_hdr_t","len")=16
	Set gtmtypes("ddp_hdr_t",1,"name")="ddp_hdr_t.trancode"
	Set gtmtypes("ddp_hdr_t",1,"off")=0
	Set gtmtypes("ddp_hdr_t",1,"len")=1
	Set gtmtypes("ddp_hdr_t",1,"type")="unsigned-char"
	Set gtmtypfldindx("ddp_hdr_t","trancode")=1
	Set gtmtypes("ddp_hdr_t",2,"name")="ddp_hdr_t.proto"
	Set gtmtypes("ddp_hdr_t",2,"off")=1
	Set gtmtypes("ddp_hdr_t",2,"len")=1
	Set gtmtypes("ddp_hdr_t",2,"type")="unsigned-char"
	Set gtmtypfldindx("ddp_hdr_t","proto")=2
	Set gtmtypes("ddp_hdr_t",3,"name")="ddp_hdr_t.source_circuit_name"
	Set gtmtypes("ddp_hdr_t",3,"off")=2
	Set gtmtypes("ddp_hdr_t",3,"len")=2
	Set gtmtypes("ddp_hdr_t",3,"type")="unsigned-short"
	Set gtmtypfldindx("ddp_hdr_t","source_circuit_name")=3
	Set gtmtypes("ddp_hdr_t",4,"name")="ddp_hdr_t.source_job_number"
	Set gtmtypes("ddp_hdr_t",4,"off")=4
	Set gtmtypes("ddp_hdr_t",4,"len")=2
	Set gtmtypes("ddp_hdr_t",4,"type")="unsigned-short"
	Set gtmtypfldindx("ddp_hdr_t","source_job_number")=4
	Set gtmtypes("ddp_hdr_t",5,"name")="ddp_hdr_t.remote_circuit_name"
	Set gtmtypes("ddp_hdr_t",5,"off")=6
	Set gtmtypes("ddp_hdr_t",5,"len")=2
	Set gtmtypes("ddp_hdr_t",5,"type")="unsigned-short"
	Set gtmtypfldindx("ddp_hdr_t","remote_circuit_name")=5
	Set gtmtypes("ddp_hdr_t",6,"name")="ddp_hdr_t.remote_job_number"
	Set gtmtypes("ddp_hdr_t",6,"off")=8
	Set gtmtypes("ddp_hdr_t",6,"len")=2
	Set gtmtypes("ddp_hdr_t",6,"type")="unsigned-short"
	Set gtmtypfldindx("ddp_hdr_t","remote_job_number")=6
	Set gtmtypes("ddp_hdr_t",7,"name")="ddp_hdr_t.message_number"
	Set gtmtypes("ddp_hdr_t",7,"off")=10
	Set gtmtypes("ddp_hdr_t",7,"len")=1
	Set gtmtypes("ddp_hdr_t",7,"type")="unsigned-char"
	Set gtmtypfldindx("ddp_hdr_t","message_number")=7
	Set gtmtypes("ddp_hdr_t",8,"name")="ddp_hdr_t.filler1"
	Set gtmtypes("ddp_hdr_t",8,"off")=11
	Set gtmtypes("ddp_hdr_t",8,"len")=1
	Set gtmtypes("ddp_hdr_t",8,"type")="unsigned-char"
	Set gtmtypfldindx("ddp_hdr_t","filler1")=8
	Set gtmtypes("ddp_hdr_t",9,"name")="ddp_hdr_t.message_length"
	Set gtmtypes("ddp_hdr_t",9,"off")=12
	Set gtmtypes("ddp_hdr_t",9,"len")=2
	Set gtmtypes("ddp_hdr_t",9,"type")="unsigned-short"
	Set gtmtypfldindx("ddp_hdr_t","message_length")=9
	Set gtmtypes("ddp_hdr_t",10,"name")="ddp_hdr_t.hdrlen"
	Set gtmtypes("ddp_hdr_t",10,"off")=14
	Set gtmtypes("ddp_hdr_t",10,"len")=1
	Set gtmtypes("ddp_hdr_t",10,"type")="unsigned-char"
	Set gtmtypfldindx("ddp_hdr_t","hdrlen")=10
	Set gtmtypes("ddp_hdr_t",11,"name")="ddp_hdr_t.txt"
	Set gtmtypes("ddp_hdr_t",11,"off")=15
	Set gtmtypes("ddp_hdr_t",11,"len")=1
	Set gtmtypes("ddp_hdr_t",11,"type")="unsigned-char"
	Set gtmtypfldindx("ddp_hdr_t","txt")=11
	;
	Set gtmtypes("ddp_info")="struct"
	Set gtmtypes("ddp_info",0)=2
	Set gtmtypes("ddp_info","len")=4
	Set gtmtypes("ddp_info",1,"name")="ddp_info.uci"
	Set gtmtypes("ddp_info",1,"off")=0
	Set gtmtypes("ddp_info",1,"len")=2
	Set gtmtypes("ddp_info",1,"type")="unsigned-short"
	Set gtmtypfldindx("ddp_info","uci")=1
	Set gtmtypes("ddp_info",2,"name")="ddp_info.volset"
	Set gtmtypes("ddp_info",2,"off")=2
	Set gtmtypes("ddp_info",2,"len")=2
	Set gtmtypes("ddp_info",2,"type")="unsigned-short"
	Set gtmtypfldindx("ddp_info","volset")=2
	;
	Set gtmtypes("delimfmt")="union"
	Set gtmtypes("delimfmt",0)=2
	Set gtmtypes("delimfmt","len")=4
	Set gtmtypes("delimfmt",1,"name")="delimfmt.unichar_val"
	Set gtmtypes("delimfmt",1,"off")=0
	Set gtmtypes("delimfmt",1,"len")=4
	Set gtmtypes("delimfmt",1,"type")="int"
	Set gtmtypfldindx("delimfmt","unichar_val")=1
	Set gtmtypes("delimfmt",2,"name")="delimfmt.unibytes_val"
	Set gtmtypes("delimfmt",2,"off")=0
	Set gtmtypes("delimfmt",2,"len")=4
	Set gtmtypes("delimfmt",2,"type")="unsigned-char"
	Set gtmtypfldindx("delimfmt","unibytes_val")=2
	;
	Set gtmtypes("desc_struct")="struct"
	Set gtmtypes("desc_struct",0)=4
	Set gtmtypes("desc_struct","len")=16
	Set gtmtypes("desc_struct",1,"name")="desc_struct.len"
	Set gtmtypes("desc_struct",1,"off")=0
	Set gtmtypes("desc_struct",1,"len")=2
	Set gtmtypes("desc_struct",1,"type")="unsigned-short"
	Set gtmtypfldindx("desc_struct","len")=1
	Set gtmtypes("desc_struct",2,"name")="desc_struct.fill1"
	Set gtmtypes("desc_struct",2,"off")=2
	Set gtmtypes("desc_struct",2,"len")=1
	Set gtmtypes("desc_struct",2,"type")="unsigned-char"
	Set gtmtypfldindx("desc_struct","fill1")=2
	Set gtmtypes("desc_struct",3,"name")="desc_struct.fill2"
	Set gtmtypes("desc_struct",3,"off")=3
	Set gtmtypes("desc_struct",3,"len")=1
	Set gtmtypes("desc_struct",3,"type")="unsigned-char"
	Set gtmtypfldindx("desc_struct","fill2")=3
	Set gtmtypes("desc_struct",4,"name")="desc_struct.addr"
	Set gtmtypes("desc_struct",4,"off")=8
	Set gtmtypes("desc_struct",4,"len")=8
	Set gtmtypes("desc_struct",4,"type")="addr"
	Set gtmtypfldindx("desc_struct","addr")=4
	;
	Set gtmtypes("dev_ctl_struct")="struct"
	Set gtmtypes("dev_ctl_struct",0)=2
	Set gtmtypes("dev_ctl_struct","len")=2
	Set gtmtypes("dev_ctl_struct",1,"name")="dev_ctl_struct.valid_with"
	Set gtmtypes("dev_ctl_struct",1,"off")=0
	Set gtmtypes("dev_ctl_struct",1,"len")=1
	Set gtmtypes("dev_ctl_struct",1,"type")="unsigned-char"
	Set gtmtypfldindx("dev_ctl_struct","valid_with")=1
	Set gtmtypes("dev_ctl_struct",2,"name")="dev_ctl_struct.source_type"
	Set gtmtypes("dev_ctl_struct",2,"off")=1
	Set gtmtypes("dev_ctl_struct",2,"len")=1
	Set gtmtypes("dev_ctl_struct",2,"type")="unsigned-char"
	Set gtmtypfldindx("dev_ctl_struct","source_type")=2
	;
	Set gtmtypes("dev_dispatch_struct")="struct"
	Set gtmtypes("dev_dispatch_struct",0)=16
	Set gtmtypes("dev_dispatch_struct","len")=128
	Set gtmtypes("dev_dispatch_struct",1,"name")="dev_dispatch_struct.open"
	Set gtmtypes("dev_dispatch_struct",1,"off")=0
	Set gtmtypes("dev_dispatch_struct",1,"len")=8
	Set gtmtypes("dev_dispatch_struct",1,"type")="addr"
	Set gtmtypfldindx("dev_dispatch_struct","open")=1
	Set gtmtypes("dev_dispatch_struct",2,"name")="dev_dispatch_struct.gtm_close"
	Set gtmtypes("dev_dispatch_struct",2,"off")=8
	Set gtmtypes("dev_dispatch_struct",2,"len")=8
	Set gtmtypes("dev_dispatch_struct",2,"type")="addr"
	Set gtmtypfldindx("dev_dispatch_struct","gtm_close")=2
	Set gtmtypes("dev_dispatch_struct",3,"name")="dev_dispatch_struct.use"
	Set gtmtypes("dev_dispatch_struct",3,"off")=16
	Set gtmtypes("dev_dispatch_struct",3,"len")=8
	Set gtmtypes("dev_dispatch_struct",3,"type")="addr"
	Set gtmtypfldindx("dev_dispatch_struct","use")=3
	Set gtmtypes("dev_dispatch_struct",4,"name")="dev_dispatch_struct.read"
	Set gtmtypes("dev_dispatch_struct",4,"off")=24
	Set gtmtypes("dev_dispatch_struct",4,"len")=8
	Set gtmtypes("dev_dispatch_struct",4,"type")="addr"
	Set gtmtypfldindx("dev_dispatch_struct","read")=4
	Set gtmtypes("dev_dispatch_struct",5,"name")="dev_dispatch_struct.rdone"
	Set gtmtypes("dev_dispatch_struct",5,"off")=32
	Set gtmtypes("dev_dispatch_struct",5,"len")=8
	Set gtmtypes("dev_dispatch_struct",5,"type")="addr"
	Set gtmtypfldindx("dev_dispatch_struct","rdone")=5
	Set gtmtypes("dev_dispatch_struct",6,"name")="dev_dispatch_struct.write"
	Set gtmtypes("dev_dispatch_struct",6,"off")=40
	Set gtmtypes("dev_dispatch_struct",6,"len")=8
	Set gtmtypes("dev_dispatch_struct",6,"type")="addr"
	Set gtmtypfldindx("dev_dispatch_struct","write")=6
	Set gtmtypes("dev_dispatch_struct",7,"name")="dev_dispatch_struct.wtone"
	Set gtmtypes("dev_dispatch_struct",7,"off")=48
	Set gtmtypes("dev_dispatch_struct",7,"len")=8
	Set gtmtypes("dev_dispatch_struct",7,"type")="addr"
	Set gtmtypfldindx("dev_dispatch_struct","wtone")=7
	Set gtmtypes("dev_dispatch_struct",8,"name")="dev_dispatch_struct.wteol"
	Set gtmtypes("dev_dispatch_struct",8,"off")=56
	Set gtmtypes("dev_dispatch_struct",8,"len")=8
	Set gtmtypes("dev_dispatch_struct",8,"type")="addr"
	Set gtmtypfldindx("dev_dispatch_struct","wteol")=8
	Set gtmtypes("dev_dispatch_struct",9,"name")="dev_dispatch_struct.wtff"
	Set gtmtypes("dev_dispatch_struct",9,"off")=64
	Set gtmtypes("dev_dispatch_struct",9,"len")=8
	Set gtmtypes("dev_dispatch_struct",9,"type")="addr"
	Set gtmtypfldindx("dev_dispatch_struct","wtff")=9
	Set gtmtypes("dev_dispatch_struct",10,"name")="dev_dispatch_struct.wttab"
	Set gtmtypes("dev_dispatch_struct",10,"off")=72
	Set gtmtypes("dev_dispatch_struct",10,"len")=8
	Set gtmtypes("dev_dispatch_struct",10,"type")="addr"
	Set gtmtypfldindx("dev_dispatch_struct","wttab")=10
	Set gtmtypes("dev_dispatch_struct",11,"name")="dev_dispatch_struct.flush"
	Set gtmtypes("dev_dispatch_struct",11,"off")=80
	Set gtmtypes("dev_dispatch_struct",11,"len")=8
	Set gtmtypes("dev_dispatch_struct",11,"type")="addr"
	Set gtmtypfldindx("dev_dispatch_struct","flush")=11
	Set gtmtypes("dev_dispatch_struct",12,"name")="dev_dispatch_struct.readfl"
	Set gtmtypes("dev_dispatch_struct",12,"off")=88
	Set gtmtypes("dev_dispatch_struct",12,"len")=8
	Set gtmtypes("dev_dispatch_struct",12,"type")="addr"
	Set gtmtypfldindx("dev_dispatch_struct","readfl")=12
	Set gtmtypes("dev_dispatch_struct",13,"name")="dev_dispatch_struct.iocontrol"
	Set gtmtypes("dev_dispatch_struct",13,"off")=96
	Set gtmtypes("dev_dispatch_struct",13,"len")=8
	Set gtmtypes("dev_dispatch_struct",13,"type")="addr"
	Set gtmtypfldindx("dev_dispatch_struct","iocontrol")=13
	Set gtmtypes("dev_dispatch_struct",14,"name")="dev_dispatch_struct.dlr_device"
	Set gtmtypes("dev_dispatch_struct",14,"off")=104
	Set gtmtypes("dev_dispatch_struct",14,"len")=8
	Set gtmtypes("dev_dispatch_struct",14,"type")="addr"
	Set gtmtypfldindx("dev_dispatch_struct","dlr_device")=14
	Set gtmtypes("dev_dispatch_struct",15,"name")="dev_dispatch_struct.dlr_key"
	Set gtmtypes("dev_dispatch_struct",15,"off")=112
	Set gtmtypes("dev_dispatch_struct",15,"len")=8
	Set gtmtypes("dev_dispatch_struct",15,"type")="addr"
	Set gtmtypfldindx("dev_dispatch_struct","dlr_key")=15
	Set gtmtypes("dev_dispatch_struct",16,"name")="dev_dispatch_struct.dlr_zkey"
	Set gtmtypes("dev_dispatch_struct",16,"off")=120
	Set gtmtypes("dev_dispatch_struct",16,"len")=8
	Set gtmtypes("dev_dispatch_struct",16,"type")="addr"
	Set gtmtypfldindx("dev_dispatch_struct","dlr_zkey")=16
	;
	Set gtmtypes("dio_buff_t")="struct"
	Set gtmtypes("dio_buff_t",0)=3
	Set gtmtypes("dio_buff_t","len")=24
	Set gtmtypes("dio_buff_t",1,"name")="dio_buff_t.unaligned_size"
	Set gtmtypes("dio_buff_t",1,"off")=0
	Set gtmtypes("dio_buff_t",1,"len")=4
	Set gtmtypes("dio_buff_t",1,"type")="int"
	Set gtmtypfldindx("dio_buff_t","unaligned_size")=1
	Set gtmtypes("dio_buff_t",2,"name")="dio_buff_t.unaligned"
	Set gtmtypes("dio_buff_t",2,"off")=8
	Set gtmtypes("dio_buff_t",2,"len")=8
	Set gtmtypes("dio_buff_t",2,"type")="addr"
	Set gtmtypfldindx("dio_buff_t","unaligned")=2
	Set gtmtypes("dio_buff_t",3,"name")="dio_buff_t.aligned"
	Set gtmtypes("dio_buff_t",3,"off")=16
	Set gtmtypes("dio_buff_t",3,"len")=8
	Set gtmtypes("dio_buff_t",3,"type")="addr"
	Set gtmtypfldindx("dio_buff_t","aligned")=3
	;
	Set gtmtypes("dm_audit_info")="struct"
	Set gtmtypes("dm_audit_info",0)=7
	Set gtmtypes("dm_audit_info","len")=160
	Set gtmtypes("dm_audit_info",1,"name")="dm_audit_info.conn_type"
	Set gtmtypes("dm_audit_info",1,"off")=0
	Set gtmtypes("dm_audit_info",1,"len")=4
	Set gtmtypes("dm_audit_info",1,"type")="int"
	Set gtmtypfldindx("dm_audit_info","conn_type")=1
	Set gtmtypes("dm_audit_info",2,"name")="dm_audit_info.un_addr"
	Set gtmtypes("dm_audit_info",2,"off")=4
	Set gtmtypes("dm_audit_info",2,"len")=110
	Set gtmtypes("dm_audit_info",2,"type")="struct"
	Set gtmtypfldindx("dm_audit_info","un_addr")=2
	Set gtmtypes("dm_audit_info",3,"name")="dm_audit_info.tcp_addr"
	Set gtmtypes("dm_audit_info",3,"off")=120
	Set gtmtypes("dm_audit_info",3,"len")=8
	Set gtmtypes("dm_audit_info",3,"type")="addr"
	Set gtmtypfldindx("dm_audit_info","tcp_addr")=3
	Set gtmtypes("dm_audit_info",4,"name")="dm_audit_info.sock_fd"
	Set gtmtypes("dm_audit_info",4,"off")=128
	Set gtmtypes("dm_audit_info",4,"len")=4
	Set gtmtypes("dm_audit_info",4,"type")="int"
	Set gtmtypfldindx("dm_audit_info","sock_fd")=4
	Set gtmtypes("dm_audit_info",5,"name")="dm_audit_info.tls_id"
	Set gtmtypes("dm_audit_info",5,"off")=136
	Set gtmtypes("dm_audit_info",5,"len")=8
	Set gtmtypes("dm_audit_info",5,"type")="addr"
	Set gtmtypfldindx("dm_audit_info","tls_id")=5
	Set gtmtypes("dm_audit_info",6,"name")="dm_audit_info.tls_sock"
	Set gtmtypes("dm_audit_info",6,"off")=144
	Set gtmtypes("dm_audit_info",6,"len")=8
	Set gtmtypes("dm_audit_info",6,"type")="addr"
	Set gtmtypfldindx("dm_audit_info","tls_sock")=6
	Set gtmtypes("dm_audit_info",7,"name")="dm_audit_info.initialized"
	Set gtmtypes("dm_audit_info",7,"off")=152
	Set gtmtypes("dm_audit_info",7,"len")=4
	Set gtmtypes("dm_audit_info",7,"type")="boolean_t"
	Set gtmtypfldindx("dm_audit_info","initialized")=7
	;
	Set gtmtypes("dollar_ecode_struct")="struct"
	Set gtmtypes("dollar_ecode_struct",0)=4
	Set gtmtypes("dollar_ecode_struct","len")=16
	Set gtmtypes("dollar_ecode_struct",1,"name")="dollar_ecode_struct.ecode_str"
	Set gtmtypes("dollar_ecode_struct",1,"off")=0
	Set gtmtypes("dollar_ecode_struct",1,"len")=16
	Set gtmtypes("dollar_ecode_struct",1,"type")="mstr"
	Set gtmtypfldindx("dollar_ecode_struct","ecode_str")=1
	Set gtmtypes("dollar_ecode_struct",2,"name")="dollar_ecode_struct.ecode_str.char_len"
	Set gtmtypes("dollar_ecode_struct",2,"off")=0
	Set gtmtypes("dollar_ecode_struct",2,"len")=4
	Set gtmtypes("dollar_ecode_struct",2,"type")="unsigned-int"
	Set gtmtypfldindx("dollar_ecode_struct","ecode_str.char_len")=2
	Set gtmtypes("dollar_ecode_struct",3,"name")="dollar_ecode_struct.ecode_str.len"
	Set gtmtypes("dollar_ecode_struct",3,"off")=4
	Set gtmtypes("dollar_ecode_struct",3,"len")=4
	Set gtmtypes("dollar_ecode_struct",3,"type")="int"
	Set gtmtypfldindx("dollar_ecode_struct","ecode_str.len")=3
	Set gtmtypes("dollar_ecode_struct",4,"name")="dollar_ecode_struct.ecode_str.addr"
	Set gtmtypes("dollar_ecode_struct",4,"off")=8
	Set gtmtypes("dollar_ecode_struct",4,"len")=8
	Set gtmtypes("dollar_ecode_struct",4,"type")="addr"
	Set gtmtypfldindx("dollar_ecode_struct","ecode_str.addr")=4
	;
	Set gtmtypes("dollar_ecode_type")="struct"
	Set gtmtypes("dollar_ecode_type",0)=11
	Set gtmtypes("dollar_ecode_type","len")=80
	Set gtmtypes("dollar_ecode_type",1,"name")="dollar_ecode_type.begin"
	Set gtmtypes("dollar_ecode_type",1,"off")=0
	Set gtmtypes("dollar_ecode_type",1,"len")=8
	Set gtmtypes("dollar_ecode_type",1,"type")="addr"
	Set gtmtypfldindx("dollar_ecode_type","begin")=1
	Set gtmtypes("dollar_ecode_type",2,"name")="dollar_ecode_type.end"
	Set gtmtypes("dollar_ecode_type",2,"off")=8
	Set gtmtypes("dollar_ecode_type",2,"len")=8
	Set gtmtypes("dollar_ecode_type",2,"type")="addr"
	Set gtmtypfldindx("dollar_ecode_type","end")=2
	Set gtmtypes("dollar_ecode_type",3,"name")="dollar_ecode_type.top"
	Set gtmtypes("dollar_ecode_type",3,"off")=16
	Set gtmtypes("dollar_ecode_type",3,"len")=8
	Set gtmtypes("dollar_ecode_type",3,"type")="addr"
	Set gtmtypfldindx("dollar_ecode_type","top")=3
	Set gtmtypes("dollar_ecode_type",4,"name")="dollar_ecode_type.array"
	Set gtmtypes("dollar_ecode_type",4,"off")=24
	Set gtmtypes("dollar_ecode_type",4,"len")=8
	Set gtmtypes("dollar_ecode_type",4,"type")="addr"
	Set gtmtypfldindx("dollar_ecode_type","array")=4
	Set gtmtypes("dollar_ecode_type",5,"name")="dollar_ecode_type.index"
	Set gtmtypes("dollar_ecode_type",5,"off")=32
	Set gtmtypes("dollar_ecode_type",5,"len")=4
	Set gtmtypes("dollar_ecode_type",5,"type")="unsigned-int"
	Set gtmtypfldindx("dollar_ecode_type","index")=5
	Set gtmtypes("dollar_ecode_type",6,"name")="dollar_ecode_type.error_last_ecode"
	Set gtmtypes("dollar_ecode_type",6,"off")=36
	Set gtmtypes("dollar_ecode_type",6,"len")=4
	Set gtmtypes("dollar_ecode_type",6,"type")="int"
	Set gtmtypfldindx("dollar_ecode_type","error_last_ecode")=6
	Set gtmtypes("dollar_ecode_type",7,"name")="dollar_ecode_type.error_last_b_line"
	Set gtmtypes("dollar_ecode_type",7,"off")=40
	Set gtmtypes("dollar_ecode_type",7,"len")=8
	Set gtmtypes("dollar_ecode_type",7,"type")="addr"
	Set gtmtypfldindx("dollar_ecode_type","error_last_b_line")=7
	Set gtmtypes("dollar_ecode_type",8,"name")="dollar_ecode_type.first_ecode_error_frame"
	Set gtmtypes("dollar_ecode_type",8,"off")=48
	Set gtmtypes("dollar_ecode_type",8,"len")=8
	Set gtmtypes("dollar_ecode_type",8,"type")="addr"
	Set gtmtypfldindx("dollar_ecode_type","first_ecode_error_frame")=8
	Set gtmtypes("dollar_ecode_type",9,"name")="dollar_ecode_type.error_rtn_addr"
	Set gtmtypes("dollar_ecode_type",9,"off")=56
	Set gtmtypes("dollar_ecode_type",9,"len")=8
	Set gtmtypes("dollar_ecode_type",9,"type")="addr"
	Set gtmtypfldindx("dollar_ecode_type","error_rtn_addr")=9
	Set gtmtypes("dollar_ecode_type",10,"name")="dollar_ecode_type.error_rtn_ctxt"
	Set gtmtypes("dollar_ecode_type",10,"off")=64
	Set gtmtypes("dollar_ecode_type",10,"len")=8
	Set gtmtypes("dollar_ecode_type",10,"type")="addr"
	Set gtmtypfldindx("dollar_ecode_type","error_rtn_ctxt")=10
	Set gtmtypes("dollar_ecode_type",11,"name")="dollar_ecode_type.error_return_addr"
	Set gtmtypes("dollar_ecode_type",11,"off")=72
	Set gtmtypes("dollar_ecode_type",11,"len")=8
	Set gtmtypes("dollar_ecode_type",11,"type")="void"
	Set gtmtypfldindx("dollar_ecode_type","error_return_addr")=11
	;
	Set gtmtypes("dollar_stack_struct")="struct"
	Set gtmtypes("dollar_stack_struct",0)=13
	Set gtmtypes("dollar_stack_struct","len")=56
	Set gtmtypes("dollar_stack_struct",1,"name")="dollar_stack_struct.mode_str"
	Set gtmtypes("dollar_stack_struct",1,"off")=0
	Set gtmtypes("dollar_stack_struct",1,"len")=16
	Set gtmtypes("dollar_stack_struct",1,"type")="mstr"
	Set gtmtypfldindx("dollar_stack_struct","mode_str")=1
	Set gtmtypes("dollar_stack_struct",2,"name")="dollar_stack_struct.mode_str.char_len"
	Set gtmtypes("dollar_stack_struct",2,"off")=0
	Set gtmtypes("dollar_stack_struct",2,"len")=4
	Set gtmtypes("dollar_stack_struct",2,"type")="unsigned-int"
	Set gtmtypfldindx("dollar_stack_struct","mode_str.char_len")=2
	Set gtmtypes("dollar_stack_struct",3,"name")="dollar_stack_struct.mode_str.len"
	Set gtmtypes("dollar_stack_struct",3,"off")=4
	Set gtmtypes("dollar_stack_struct",3,"len")=4
	Set gtmtypes("dollar_stack_struct",3,"type")="int"
	Set gtmtypfldindx("dollar_stack_struct","mode_str.len")=3
	Set gtmtypes("dollar_stack_struct",4,"name")="dollar_stack_struct.mode_str.addr"
	Set gtmtypes("dollar_stack_struct",4,"off")=8
	Set gtmtypes("dollar_stack_struct",4,"len")=8
	Set gtmtypes("dollar_stack_struct",4,"type")="addr"
	Set gtmtypfldindx("dollar_stack_struct","mode_str.addr")=4
	Set gtmtypes("dollar_stack_struct",5,"name")="dollar_stack_struct.ecode_ptr"
	Set gtmtypes("dollar_stack_struct",5,"off")=16
	Set gtmtypes("dollar_stack_struct",5,"len")=8
	Set gtmtypes("dollar_stack_struct",5,"type")="addr"
	Set gtmtypfldindx("dollar_stack_struct","ecode_ptr")=5
	Set gtmtypes("dollar_stack_struct",6,"name")="dollar_stack_struct.mcode_str"
	Set gtmtypes("dollar_stack_struct",6,"off")=24
	Set gtmtypes("dollar_stack_struct",6,"len")=16
	Set gtmtypes("dollar_stack_struct",6,"type")="mstr"
	Set gtmtypfldindx("dollar_stack_struct","mcode_str")=6
	Set gtmtypes("dollar_stack_struct",7,"name")="dollar_stack_struct.mcode_str.char_len"
	Set gtmtypes("dollar_stack_struct",7,"off")=24
	Set gtmtypes("dollar_stack_struct",7,"len")=4
	Set gtmtypes("dollar_stack_struct",7,"type")="unsigned-int"
	Set gtmtypfldindx("dollar_stack_struct","mcode_str.char_len")=7
	Set gtmtypes("dollar_stack_struct",8,"name")="dollar_stack_struct.mcode_str.len"
	Set gtmtypes("dollar_stack_struct",8,"off")=28
	Set gtmtypes("dollar_stack_struct",8,"len")=4
	Set gtmtypes("dollar_stack_struct",8,"type")="int"
	Set gtmtypfldindx("dollar_stack_struct","mcode_str.len")=8
	Set gtmtypes("dollar_stack_struct",9,"name")="dollar_stack_struct.mcode_str.addr"
	Set gtmtypes("dollar_stack_struct",9,"off")=32
	Set gtmtypes("dollar_stack_struct",9,"len")=8
	Set gtmtypes("dollar_stack_struct",9,"type")="addr"
	Set gtmtypfldindx("dollar_stack_struct","mcode_str.addr")=9
	Set gtmtypes("dollar_stack_struct",10,"name")="dollar_stack_struct.place_str"
	Set gtmtypes("dollar_stack_struct",10,"off")=40
	Set gtmtypes("dollar_stack_struct",10,"len")=16
	Set gtmtypes("dollar_stack_struct",10,"type")="mstr"
	Set gtmtypfldindx("dollar_stack_struct","place_str")=10
	Set gtmtypes("dollar_stack_struct",11,"name")="dollar_stack_struct.place_str.char_len"
	Set gtmtypes("dollar_stack_struct",11,"off")=40
	Set gtmtypes("dollar_stack_struct",11,"len")=4
	Set gtmtypes("dollar_stack_struct",11,"type")="unsigned-int"
	Set gtmtypfldindx("dollar_stack_struct","place_str.char_len")=11
	Set gtmtypes("dollar_stack_struct",12,"name")="dollar_stack_struct.place_str.len"
	Set gtmtypes("dollar_stack_struct",12,"off")=44
	Set gtmtypes("dollar_stack_struct",12,"len")=4
	Set gtmtypes("dollar_stack_struct",12,"type")="int"
	Set gtmtypfldindx("dollar_stack_struct","place_str.len")=12
	Set gtmtypes("dollar_stack_struct",13,"name")="dollar_stack_struct.place_str.addr"
	Set gtmtypes("dollar_stack_struct",13,"off")=48
	Set gtmtypes("dollar_stack_struct",13,"len")=8
	Set gtmtypes("dollar_stack_struct",13,"type")="addr"
	Set gtmtypfldindx("dollar_stack_struct","place_str.addr")=13
	;
	Set gtmtypes("dollar_stack_type")="struct"
	Set gtmtypes("dollar_stack_type",0)=6
	Set gtmtypes("dollar_stack_type","len")=40
	Set gtmtypes("dollar_stack_type",1,"name")="dollar_stack_type.begin"
	Set gtmtypes("dollar_stack_type",1,"off")=0
	Set gtmtypes("dollar_stack_type",1,"len")=8
	Set gtmtypes("dollar_stack_type",1,"type")="addr"
	Set gtmtypfldindx("dollar_stack_type","begin")=1
	Set gtmtypes("dollar_stack_type",2,"name")="dollar_stack_type.end"
	Set gtmtypes("dollar_stack_type",2,"off")=8
	Set gtmtypes("dollar_stack_type",2,"len")=8
	Set gtmtypes("dollar_stack_type",2,"type")="addr"
	Set gtmtypfldindx("dollar_stack_type","end")=2
	Set gtmtypes("dollar_stack_type",3,"name")="dollar_stack_type.top"
	Set gtmtypes("dollar_stack_type",3,"off")=16
	Set gtmtypes("dollar_stack_type",3,"len")=8
	Set gtmtypes("dollar_stack_type",3,"type")="addr"
	Set gtmtypfldindx("dollar_stack_type","top")=3
	Set gtmtypes("dollar_stack_type",4,"name")="dollar_stack_type.array"
	Set gtmtypes("dollar_stack_type",4,"off")=24
	Set gtmtypes("dollar_stack_type",4,"len")=8
	Set gtmtypes("dollar_stack_type",4,"type")="addr"
	Set gtmtypfldindx("dollar_stack_type","array")=4
	Set gtmtypes("dollar_stack_type",5,"name")="dollar_stack_type.index"
	Set gtmtypes("dollar_stack_type",5,"off")=32
	Set gtmtypes("dollar_stack_type",5,"len")=4
	Set gtmtypes("dollar_stack_type",5,"type")="unsigned-int"
	Set gtmtypfldindx("dollar_stack_type","index")=5
	Set gtmtypes("dollar_stack_type",6,"name")="dollar_stack_type.incomplete"
	Set gtmtypes("dollar_stack_type",6,"off")=36
	Set gtmtypes("dollar_stack_type",6,"len")=4
	Set gtmtypes("dollar_stack_type",6,"type")="boolean_t"
	Set gtmtypfldindx("dollar_stack_type","incomplete")=6
	;
	Set gtmtypes("dollar_ztimeout_struct")="struct"
	Set gtmtypes("dollar_ztimeout_struct",0)=15
	Set gtmtypes("dollar_ztimeout_struct","len")=80
	Set gtmtypes("dollar_ztimeout_struct",1,"name")="dollar_ztimeout_struct.ztimeout_vector"
	Set gtmtypes("dollar_ztimeout_struct",1,"off")=0
	Set gtmtypes("dollar_ztimeout_struct",1,"len")=32
	Set gtmtypes("dollar_ztimeout_struct",1,"type")="mval"
	Set gtmtypfldindx("dollar_ztimeout_struct","ztimeout_vector")=1
	Set gtmtypes("dollar_ztimeout_struct",2,"name")="dollar_ztimeout_struct.ztimeout_vector.mvtype"
	Set gtmtypes("dollar_ztimeout_struct",2,"off")=0
	Set gtmtypes("dollar_ztimeout_struct",2,"len")=2
	Set gtmtypes("dollar_ztimeout_struct",2,"type")="unsigned-short"
	Set gtmtypfldindx("dollar_ztimeout_struct","ztimeout_vector.mvtype")=2
	Set gtmtypes("dollar_ztimeout_struct",3,"name")="dollar_ztimeout_struct.ztimeout_vector.fnpc_indx"
	Set gtmtypes("dollar_ztimeout_struct",3,"off")=3
	Set gtmtypes("dollar_ztimeout_struct",3,"len")=1
	Set gtmtypes("dollar_ztimeout_struct",3,"type")="unsigned-char"
	Set gtmtypfldindx("dollar_ztimeout_struct","ztimeout_vector.fnpc_indx")=3
	Set gtmtypes("dollar_ztimeout_struct",4,"name")="dollar_ztimeout_struct.ztimeout_vector.utfcgr_indx"
	Set gtmtypes("dollar_ztimeout_struct",4,"off")=4
	Set gtmtypes("dollar_ztimeout_struct",4,"len")=4
	Set gtmtypes("dollar_ztimeout_struct",4,"type")="unsigned-int"
	Set gtmtypfldindx("dollar_ztimeout_struct","ztimeout_vector.utfcgr_indx")=4
	Set gtmtypes("dollar_ztimeout_struct",5,"name")="dollar_ztimeout_struct.ztimeout_vector.m"
	Set gtmtypes("dollar_ztimeout_struct",5,"off")=8
	Set gtmtypes("dollar_ztimeout_struct",5,"len")=8
	Set gtmtypes("dollar_ztimeout_struct",5,"type")="int"
	Set gtmtypfldindx("dollar_ztimeout_struct","ztimeout_vector.m")=5
	Set gtmtypes("dollar_ztimeout_struct",5,"dim")=2
	Set gtmtypes("dollar_ztimeout_struct",6,"name")="dollar_ztimeout_struct.ztimeout_vector.str"
	Set gtmtypes("dollar_ztimeout_struct",6,"off")=16
	Set gtmtypes("dollar_ztimeout_struct",6,"len")=16
	Set gtmtypes("dollar_ztimeout_struct",6,"type")="mstr"
	Set gtmtypfldindx("dollar_ztimeout_struct","ztimeout_vector.str")=6
	Set gtmtypes("dollar_ztimeout_struct",7,"name")="dollar_ztimeout_struct.ztimeout_seconds"
	Set gtmtypes("dollar_ztimeout_struct",7,"off")=32
	Set gtmtypes("dollar_ztimeout_struct",7,"len")=32
	Set gtmtypes("dollar_ztimeout_struct",7,"type")="mval"
	Set gtmtypfldindx("dollar_ztimeout_struct","ztimeout_seconds")=7
	Set gtmtypes("dollar_ztimeout_struct",8,"name")="dollar_ztimeout_struct.ztimeout_seconds.mvtype"
	Set gtmtypes("dollar_ztimeout_struct",8,"off")=32
	Set gtmtypes("dollar_ztimeout_struct",8,"len")=2
	Set gtmtypes("dollar_ztimeout_struct",8,"type")="unsigned-short"
	Set gtmtypfldindx("dollar_ztimeout_struct","ztimeout_seconds.mvtype")=8
	Set gtmtypes("dollar_ztimeout_struct",9,"name")="dollar_ztimeout_struct.ztimeout_seconds.fnpc_indx"
	Set gtmtypes("dollar_ztimeout_struct",9,"off")=35
	Set gtmtypes("dollar_ztimeout_struct",9,"len")=1
	Set gtmtypes("dollar_ztimeout_struct",9,"type")="unsigned-char"
	Set gtmtypfldindx("dollar_ztimeout_struct","ztimeout_seconds.fnpc_indx")=9
	Set gtmtypes("dollar_ztimeout_struct",10,"name")="dollar_ztimeout_struct.ztimeout_seconds.utfcgr_indx"
	Set gtmtypes("dollar_ztimeout_struct",10,"off")=36
	Set gtmtypes("dollar_ztimeout_struct",10,"len")=4
	Set gtmtypes("dollar_ztimeout_struct",10,"type")="unsigned-int"
	Set gtmtypfldindx("dollar_ztimeout_struct","ztimeout_seconds.utfcgr_indx")=10
	Set gtmtypes("dollar_ztimeout_struct",11,"name")="dollar_ztimeout_struct.ztimeout_seconds.m"
	Set gtmtypes("dollar_ztimeout_struct",11,"off")=40
	Set gtmtypes("dollar_ztimeout_struct",11,"len")=8
	Set gtmtypes("dollar_ztimeout_struct",11,"type")="int"
	Set gtmtypfldindx("dollar_ztimeout_struct","ztimeout_seconds.m")=11
	Set gtmtypes("dollar_ztimeout_struct",11,"dim")=2
	Set gtmtypes("dollar_ztimeout_struct",12,"name")="dollar_ztimeout_struct.ztimeout_seconds.str"
	Set gtmtypes("dollar_ztimeout_struct",12,"off")=48
	Set gtmtypes("dollar_ztimeout_struct",12,"len")=16
	Set gtmtypes("dollar_ztimeout_struct",12,"type")="mstr"
	Set gtmtypfldindx("dollar_ztimeout_struct","ztimeout_seconds.str")=12
	Set gtmtypes("dollar_ztimeout_struct",13,"name")="dollar_ztimeout_struct.end_time"
	Set gtmtypes("dollar_ztimeout_struct",13,"off")=64
	Set gtmtypes("dollar_ztimeout_struct",13,"len")=16
	Set gtmtypes("dollar_ztimeout_struct",13,"type")="ABS_TIME"
	Set gtmtypfldindx("dollar_ztimeout_struct","end_time")=13
	Set gtmtypes("dollar_ztimeout_struct",14,"name")="dollar_ztimeout_struct.end_time.at_sec"
	Set gtmtypes("dollar_ztimeout_struct",14,"off")=64
	Set gtmtypes("dollar_ztimeout_struct",14,"len")=8
	Set gtmtypes("dollar_ztimeout_struct",14,"type")="long"
	Set gtmtypfldindx("dollar_ztimeout_struct","end_time.at_sec")=14
	Set gtmtypes("dollar_ztimeout_struct",15,"name")="dollar_ztimeout_struct.end_time.at_usec"
	Set gtmtypes("dollar_ztimeout_struct",15,"off")=72
	Set gtmtypes("dollar_ztimeout_struct",15,"len")=8
	Set gtmtypes("dollar_ztimeout_struct",15,"type")="long"
	Set gtmtypfldindx("dollar_ztimeout_struct","end_time.at_usec")=15
	;
	Set gtmtypes("dskread_trace")="struct"
	Set gtmtypes("dskread_trace",0)=5
	Set gtmtypes("dskread_trace","len")=32
	Set gtmtypes("dskread_trace",1,"name")="dskread_trace.cr_off"
	Set gtmtypes("dskread_trace",1,"off")=0
	Set gtmtypes("dskread_trace",1,"len")=8
	Set gtmtypes("dskread_trace",1,"type")="intptr_t"
	Set gtmtypfldindx("dskread_trace","cr_off")=1
	Set gtmtypes("dskread_trace",2,"name")="dskread_trace.cr_tn"
	Set gtmtypes("dskread_trace",2,"off")=8
	Set gtmtypes("dskread_trace",2,"len")=8
	Set gtmtypes("dskread_trace",2,"type")="uint64_t"
	Set gtmtypfldindx("dskread_trace","cr_tn")=2
	Set gtmtypes("dskread_trace",3,"name")="dskread_trace.process_id"
	Set gtmtypes("dskread_trace",3,"off")=16
	Set gtmtypes("dskread_trace",3,"len")=4
	Set gtmtypes("dskread_trace",3,"type")="unsigned-int"
	Set gtmtypfldindx("dskread_trace","process_id")=3
	Set gtmtypes("dskread_trace",4,"name")="dskread_trace.blk"
	Set gtmtypes("dskread_trace",4,"off")=20
	Set gtmtypes("dskread_trace",4,"len")=4
	Set gtmtypes("dskread_trace",4,"type")="int"
	Set gtmtypfldindx("dskread_trace","blk")=4
	Set gtmtypes("dskread_trace",5,"name")="dskread_trace.cycle"
	Set gtmtypes("dskread_trace",5,"off")=24
	Set gtmtypes("dskread_trace",5,"len")=4
	Set gtmtypes("dskread_trace",5,"type")="unsigned-int"
	Set gtmtypfldindx("dskread_trace","cycle")=5
	;
	Set gtmtypes("emit_base_info_struct")="struct"
	Set gtmtypes("emit_base_info_struct",0)=12
	Set gtmtypes("emit_base_info_struct","len")=16
	Set gtmtypes("emit_base_info_struct",1,"name")="emit_base_info_struct.modrm_byte"
	Set gtmtypes("emit_base_info_struct",1,"off")=0
	Set gtmtypes("emit_base_info_struct",1,"len")=1
	Set gtmtypes("emit_base_info_struct",1,"type")="modrm_byte_type"
	Set gtmtypfldindx("emit_base_info_struct","modrm_byte")=1
	Set gtmtypes("emit_base_info_struct",2,"name")="emit_base_info_struct.modrm_byte.modrm"
	Set gtmtypes("emit_base_info_struct",2,"off")=0
	Set gtmtypes("emit_base_info_struct",2,"len")=1
	Set gtmtypes("emit_base_info_struct",2,"type")="ModR_M"
	Set gtmtypfldindx("emit_base_info_struct","modrm_byte.modrm")=2
	Set gtmtypes("emit_base_info_struct",3,"name")="emit_base_info_struct.modrm_byte.byte"
	Set gtmtypes("emit_base_info_struct",3,"off")=0
	Set gtmtypes("emit_base_info_struct",3,"len")=1
	Set gtmtypes("emit_base_info_struct",3,"type")="unsigned-char"
	Set gtmtypfldindx("emit_base_info_struct","modrm_byte.byte")=3
	Set gtmtypes("emit_base_info_struct",4,"name")="emit_base_info_struct.sib_byte"
	Set gtmtypes("emit_base_info_struct",4,"off")=1
	Set gtmtypes("emit_base_info_struct",4,"len")=1
	Set gtmtypes("emit_base_info_struct",4,"type")="sib_byte_type"
	Set gtmtypfldindx("emit_base_info_struct","sib_byte")=4
	Set gtmtypes("emit_base_info_struct",5,"name")="emit_base_info_struct.sib_byte.sib"
	Set gtmtypes("emit_base_info_struct",5,"off")=1
	Set gtmtypes("emit_base_info_struct",5,"len")=1
	Set gtmtypes("emit_base_info_struct",5,"type")="SIB"
	Set gtmtypfldindx("emit_base_info_struct","sib_byte.sib")=5
	Set gtmtypes("emit_base_info_struct",6,"name")="emit_base_info_struct.sib_byte.byte"
	Set gtmtypes("emit_base_info_struct",6,"off")=1
	Set gtmtypes("emit_base_info_struct",6,"len")=1
	Set gtmtypes("emit_base_info_struct",6,"type")="unsigned-char"
	Set gtmtypfldindx("emit_base_info_struct","sib_byte.byte")=6
	Set gtmtypes("emit_base_info_struct",7,"name")="emit_base_info_struct.rex"
	Set gtmtypes("emit_base_info_struct",7,"off")=2
	Set gtmtypes("emit_base_info_struct",7,"len")=1
	Set gtmtypes("emit_base_info_struct",7,"type")="char"
	Set gtmtypfldindx("emit_base_info_struct","rex")=7
	Set gtmtypes("emit_base_info_struct",8,"name")="emit_base_info_struct.offset8"
	Set gtmtypes("emit_base_info_struct",8,"off")=3
	Set gtmtypes("emit_base_info_struct",8,"len")=1
	Set gtmtypes("emit_base_info_struct",8,"type")="char"
	Set gtmtypfldindx("emit_base_info_struct","offset8")=8
	Set gtmtypes("emit_base_info_struct",9,"name")="emit_base_info_struct.flags"
	Set gtmtypes("emit_base_info_struct",9,"off")=4
	Set gtmtypes("emit_base_info_struct",9,"len")=1
	Set gtmtypes("emit_base_info_struct",9,"type")="struct"
	Set gtmtypfldindx("emit_base_info_struct","flags")=9
	Set gtmtypes("emit_base_info_struct",10,"name")="emit_base_info_struct.filler"
	Set gtmtypes("emit_base_info_struct",10,"off")=5
	Set gtmtypes("emit_base_info_struct",10,"len")=1
	Set gtmtypes("emit_base_info_struct",10,"type")="char"
	Set gtmtypfldindx("emit_base_info_struct","filler")=10
	Set gtmtypes("emit_base_info_struct",11,"name")="emit_base_info_struct.offset32"
	Set gtmtypes("emit_base_info_struct",11,"off")=8
	Set gtmtypes("emit_base_info_struct",11,"len")=4
	Set gtmtypes("emit_base_info_struct",11,"type")="int"
	Set gtmtypfldindx("emit_base_info_struct","offset32")=11
	Set gtmtypes("emit_base_info_struct",12,"name")="emit_base_info_struct.imm32"
	Set gtmtypes("emit_base_info_struct",12,"off")=12
	Set gtmtypes("emit_base_info_struct",12,"len")=4
	Set gtmtypes("emit_base_info_struct",12,"type")="int"
	Set gtmtypfldindx("emit_base_info_struct","imm32")=12
	;
	Set gtmtypes("enc_handles")="struct"
	Set gtmtypes("enc_handles",0)=2
	Set gtmtypes("enc_handles","len")=16
	Set gtmtypes("enc_handles",1,"name")="enc_handles.encr_key_handle"
	Set gtmtypes("enc_handles",1,"off")=0
	Set gtmtypes("enc_handles",1,"len")=8
	Set gtmtypes("enc_handles",1,"type")="addr"
	Set gtmtypfldindx("enc_handles","encr_key_handle")=1
	Set gtmtypes("enc_handles",2,"name")="enc_handles.encr_key_handle2"
	Set gtmtypes("enc_handles",2,"off")=8
	Set gtmtypes("enc_handles",2,"len")=8
	Set gtmtypes("enc_handles",2,"type")="addr"
	Set gtmtypfldindx("enc_handles","encr_key_handle2")=2
	;
	Set gtmtypes("enc_info_t")="struct"
	Set gtmtypes("enc_info_t",0)=9
	Set gtmtypes("enc_info_t","len")=160
	Set gtmtypes("enc_info_t",1,"name")="enc_info_t.reorg_encrypt_cycle"
	Set gtmtypes("enc_info_t",1,"off")=0
	Set gtmtypes("enc_info_t",1,"len")=4
	Set gtmtypes("enc_info_t",1,"type")="unsigned-int"
	Set gtmtypfldindx("enc_info_t","reorg_encrypt_cycle")=1
	Set gtmtypes("enc_info_t",2,"name")="enc_info_t.is_encrypted"
	Set gtmtypes("enc_info_t",2,"off")=4
	Set gtmtypes("enc_info_t",2,"len")=4
	Set gtmtypes("enc_info_t",2,"type")="unsigned-int"
	Set gtmtypfldindx("enc_info_t","is_encrypted")=2
	Set gtmtypes("enc_info_t",3,"name")="enc_info_t.non_null_iv"
	Set gtmtypes("enc_info_t",3,"off")=8
	Set gtmtypes("enc_info_t",3,"len")=4
	Set gtmtypes("enc_info_t",3,"type")="boolean_t"
	Set gtmtypfldindx("enc_info_t","non_null_iv")=3
	Set gtmtypes("enc_info_t",4,"name")="enc_info_t.encryption_hash_cutoff"
	Set gtmtypes("enc_info_t",4,"off")=12
	Set gtmtypes("enc_info_t",4,"len")=4
	Set gtmtypes("enc_info_t",4,"type")="int"
	Set gtmtypfldindx("enc_info_t","encryption_hash_cutoff")=4
	Set gtmtypes("enc_info_t",5,"name")="enc_info_t.encryption_hash2_start_tn"
	Set gtmtypes("enc_info_t",5,"off")=16
	Set gtmtypes("enc_info_t",5,"len")=8
	Set gtmtypes("enc_info_t",5,"type")="uint64_t"
	Set gtmtypfldindx("enc_info_t","encryption_hash2_start_tn")=5
	Set gtmtypes("enc_info_t",6,"name")="enc_info_t.encryption_hash"
	Set gtmtypes("enc_info_t",6,"off")=24
	Set gtmtypes("enc_info_t",6,"len")=64
	Set gtmtypes("enc_info_t",6,"type")="char"
	Set gtmtypfldindx("enc_info_t","encryption_hash")=6
	Set gtmtypes("enc_info_t",7,"name")="enc_info_t.encryption_hash2"
	Set gtmtypes("enc_info_t",7,"off")=88
	Set gtmtypes("enc_info_t",7,"len")=64
	Set gtmtypes("enc_info_t",7,"type")="char"
	Set gtmtypfldindx("enc_info_t","encryption_hash2")=7
	Set gtmtypes("enc_info_t",8,"name")="enc_info_t.issued_db_init_crypt_warning"
	Set gtmtypes("enc_info_t",8,"off")=152
	Set gtmtypes("enc_info_t",8,"len")=4
	Set gtmtypes("enc_info_t",8,"type")="boolean_t"
	Set gtmtypfldindx("enc_info_t","issued_db_init_crypt_warning")=8
	Set gtmtypes("enc_info_t",9,"name")="enc_info_t.filler"
	Set gtmtypes("enc_info_t",9,"off")=156
	Set gtmtypes("enc_info_t",9,"len")=4
	Set gtmtypes("enc_info_t",9,"type")="unsigned-int"
	Set gtmtypfldindx("enc_info_t","filler")=9
	;
	Set gtmtypes("endian32_struct")="union"
	Set gtmtypes("endian32_struct",0)=4
	Set gtmtypes("endian32_struct","len")=4
	Set gtmtypes("endian32_struct",1,"name")="endian32_struct.shorts"
	Set gtmtypes("endian32_struct",1,"off")=0
	Set gtmtypes("endian32_struct",1,"len")=4
	Set gtmtypes("endian32_struct",1,"type")="struct"
	Set gtmtypfldindx("endian32_struct","shorts")=1
	Set gtmtypes("endian32_struct",2,"name")="endian32_struct.shorts.little_endian"
	Set gtmtypes("endian32_struct",2,"off")=0
	Set gtmtypes("endian32_struct",2,"len")=2
	Set gtmtypes("endian32_struct",2,"type")="unsigned-short"
	Set gtmtypfldindx("endian32_struct","shorts.little_endian")=2
	Set gtmtypes("endian32_struct",3,"name")="endian32_struct.shorts.big_endian"
	Set gtmtypes("endian32_struct",3,"off")=2
	Set gtmtypes("endian32_struct",3,"len")=2
	Set gtmtypes("endian32_struct",3,"type")="unsigned-short"
	Set gtmtypfldindx("endian32_struct","shorts.big_endian")=3
	Set gtmtypes("endian32_struct",4,"name")="endian32_struct.word32"
	Set gtmtypes("endian32_struct",4,"off")=0
	Set gtmtypes("endian32_struct",4,"len")=4
	Set gtmtypes("endian32_struct",4,"type")="unsigned-int"
	Set gtmtypfldindx("endian32_struct","word32")=4
	;
	Set gtmtypes("err_ctl")="struct"
	Set gtmtypes("err_ctl",0)=6
	Set gtmtypes("err_ctl","len")=48
	Set gtmtypes("err_ctl",1,"name")="err_ctl.facnum"
	Set gtmtypes("err_ctl",1,"off")=0
	Set gtmtypes("err_ctl",1,"len")=4
	Set gtmtypes("err_ctl",1,"type")="int"
	Set gtmtypfldindx("err_ctl","facnum")=1
	Set gtmtypes("err_ctl",2,"name")="err_ctl.facname"
	Set gtmtypes("err_ctl",2,"off")=8
	Set gtmtypes("err_ctl",2,"len")=8
	Set gtmtypes("err_ctl",2,"type")="addr"
	Set gtmtypfldindx("err_ctl","facname")=2
	Set gtmtypes("err_ctl",3,"name")="err_ctl.fst_msg"
	Set gtmtypes("err_ctl",3,"off")=16
	Set gtmtypes("err_ctl",3,"len")=8
	Set gtmtypes("err_ctl",3,"type")="addr"
	Set gtmtypfldindx("err_ctl","fst_msg")=3
	Set gtmtypes("err_ctl",4,"name")="err_ctl.msg_cnt"
	Set gtmtypes("err_ctl",4,"off")=24
	Set gtmtypes("err_ctl",4,"len")=4
	Set gtmtypes("err_ctl",4,"type")="int"
	Set gtmtypfldindx("err_ctl","msg_cnt")=4
	Set gtmtypes("err_ctl",5,"name")="err_ctl.undocmsg"
	Set gtmtypes("err_ctl",5,"off")=32
	Set gtmtypes("err_ctl",5,"len")=8
	Set gtmtypes("err_ctl",5,"type")="addr"
	Set gtmtypfldindx("err_ctl","undocmsg")=5
	Set gtmtypes("err_ctl",6,"name")="err_ctl.undocmsg_cnt"
	Set gtmtypes("err_ctl",6,"off")=40
	Set gtmtypes("err_ctl",6,"len")=4
	Set gtmtypes("err_ctl",6,"type")="int"
	Set gtmtypfldindx("err_ctl","undocmsg_cnt")=6
	;
	Set gtmtypes("err_msg")="struct"
	Set gtmtypes("err_msg",0)=4
	Set gtmtypes("err_msg","len")=24
	Set gtmtypes("err_msg",1,"name")="err_msg.tag"
	Set gtmtypes("err_msg",1,"off")=0
	Set gtmtypes("err_msg",1,"len")=8
	Set gtmtypes("err_msg",1,"type")="addr"
	Set gtmtypfldindx("err_msg","tag")=1
	Set gtmtypes("err_msg",2,"name")="err_msg.msg"
	Set gtmtypes("err_msg",2,"off")=8
	Set gtmtypes("err_msg",2,"len")=8
	Set gtmtypes("err_msg",2,"type")="addr"
	Set gtmtypfldindx("err_msg","msg")=2
	Set gtmtypes("err_msg",3,"name")="err_msg.parm_count"
	Set gtmtypes("err_msg",3,"off")=16
	Set gtmtypes("err_msg",3,"len")=4
	Set gtmtypes("err_msg",3,"type")="int"
	Set gtmtypfldindx("err_msg","parm_count")=3
	Set gtmtypes("err_msg",4,"name")="err_msg.mu_int_sev_idx"
	Set gtmtypes("err_msg",4,"off")=20
	Set gtmtypes("err_msg",4,"len")=4
	Set gtmtypes("err_msg",4,"type")="int"
	Set gtmtypfldindx("err_msg","mu_int_sev_idx")=4
	;
	Set gtmtypes("ext_tms")="struct"
	Set gtmtypes("ext_tms",0)=3
	Set gtmtypes("ext_tms","len")=24
	Set gtmtypes("ext_tms",1,"name")="ext_tms.tms_utime"
	Set gtmtypes("ext_tms",1,"off")=0
	Set gtmtypes("ext_tms",1,"len")=8
	Set gtmtypes("ext_tms",1,"type")="uint64_t"
	Set gtmtypfldindx("ext_tms","tms_utime")=1
	Set gtmtypes("ext_tms",2,"name")="ext_tms.tms_stime"
	Set gtmtypes("ext_tms",2,"off")=8
	Set gtmtypes("ext_tms",2,"len")=8
	Set gtmtypes("ext_tms",2,"type")="uint64_t"
	Set gtmtypfldindx("ext_tms","tms_stime")=2
	Set gtmtypes("ext_tms",3,"name")="ext_tms.tms_etime"
	Set gtmtypes("ext_tms",3,"off")=16
	Set gtmtypes("ext_tms",3,"len")=8
	Set gtmtypes("ext_tms",3,"type")="uint64_t"
	Set gtmtypfldindx("ext_tms","tms_etime")=3
	;
	Set gtmtypes("extr_fn_t")="struct"
	Set gtmtypes("extr_fn_t",0)=1
	Set gtmtypes("extr_fn_t","len")=256
	Set gtmtypes("extr_fn_t",1,"name")="extr_fn_t.fn"
	Set gtmtypes("extr_fn_t",1,"off")=0
	Set gtmtypes("extr_fn_t",1,"len")=256
	Set gtmtypes("extr_fn_t",1,"type")="char"
	Set gtmtypfldindx("extr_fn_t","fn")=1
	;
	Set gtmtypes("fdinfo_t")="struct"
	Set gtmtypes("fdinfo_t",0)=2
	Set gtmtypes("fdinfo_t","len")=16
	Set gtmtypes("fdinfo_t",1,"name")="fdinfo_t.reg"
	Set gtmtypes("fdinfo_t",1,"off")=0
	Set gtmtypes("fdinfo_t",1,"len")=8
	Set gtmtypes("fdinfo_t",1,"type")="addr"
	Set gtmtypfldindx("fdinfo_t","reg")=1
	Set gtmtypes("fdinfo_t",2,"name")="fdinfo_t.is_db"
	Set gtmtypes("fdinfo_t",2,"off")=8
	Set gtmtypes("fdinfo_t",2,"len")=4
	Set gtmtypes("fdinfo_t",2,"type")="boolean_t"
	Set gtmtypfldindx("fdinfo_t","is_db")=2
	;
	Set gtmtypes("file_control")="struct"
	Set gtmtypes("file_control",0)=5
	Set gtmtypes("file_control","len")=40
	Set gtmtypes("file_control",1,"name")="file_control.op_buff"
	Set gtmtypes("file_control",1,"off")=0
	Set gtmtypes("file_control",1,"len")=8
	Set gtmtypes("file_control",1,"type")="addr"
	Set gtmtypfldindx("file_control","op_buff")=1
	Set gtmtypes("file_control",2,"name")="file_control.op_pos"
	Set gtmtypes("file_control",2,"off")=8
	Set gtmtypes("file_control",2,"len")=8
	Set gtmtypes("file_control",2,"type")="int64_t"
	Set gtmtypfldindx("file_control","op_pos")=2
	Set gtmtypes("file_control",3,"name")="file_control.op_len"
	Set gtmtypes("file_control",3,"off")=16
	Set gtmtypes("file_control",3,"len")=4
	Set gtmtypes("file_control",3,"type")="int"
	Set gtmtypfldindx("file_control","op_len")=3
	Set gtmtypes("file_control",4,"name")="file_control.file_info"
	Set gtmtypes("file_control",4,"off")=24
	Set gtmtypes("file_control",4,"len")=8
	Set gtmtypes("file_control",4,"type")="addr"
	Set gtmtypfldindx("file_control","file_info")=4
	Set gtmtypes("file_control",5,"name")="file_control.op"
	Set gtmtypes("file_control",5,"off")=32
	Set gtmtypes("file_control",5,"len")=1
	Set gtmtypes("file_control",5,"type")="char"
	Set gtmtypfldindx("file_control","op")=5
	;
	Set gtmtypes("file_info")="struct"
	Set gtmtypes("file_info",0)=5
	Set gtmtypes("file_info","len")=52
	Set gtmtypes("file_info",1,"name")="file_info.fac"
	Set gtmtypes("file_info",1,"off")=0
	Set gtmtypes("file_info",1,"len")=4
	Set gtmtypes("file_info",1,"type")="int"
	Set gtmtypfldindx("file_info","fac")=1
	Set gtmtypes("file_info",2,"name")="file_info.dat"
	Set gtmtypes("file_info",2,"off")=4
	Set gtmtypes("file_info",2,"len")=8
	Set gtmtypes("file_info",2,"type")="short"
	Set gtmtypfldindx("file_info","dat")=2
	Set gtmtypes("file_info",2,"dim")=4
	Set gtmtypes("file_info",3,"name")="file_info.usr"
	Set gtmtypes("file_info",3,"off")=12
	Set gtmtypes("file_info",3,"len")=31
	Set gtmtypes("file_info",3,"type")="char"
	Set gtmtypfldindx("file_info","usr")=3
	Set gtmtypes("file_info",4,"name")="file_info.trm"
	Set gtmtypes("file_info",4,"off")=43
	Set gtmtypes("file_info",4,"len")=7
	Set gtmtypes("file_info",4,"type")="char"
	Set gtmtypfldindx("file_info","trm")=4
	Set gtmtypes("file_info",5,"name")="file_info.filler"
	Set gtmtypes("file_info",5,"off")=50
	Set gtmtypes("file_info",5,"len")=2
	Set gtmtypes("file_info",5,"type")="char"
	Set gtmtypfldindx("file_info","filler")=5
	;
	Set gtmtypes("file_log")="struct"
	Set gtmtypes("file_log",0)=7
	Set gtmtypes("file_log","len")=264
	Set gtmtypes("file_log",1,"name")="file_log.cnt"
	Set gtmtypes("file_log",1,"off")=0
	Set gtmtypes("file_log",1,"len")=4
	Set gtmtypes("file_log",1,"type")="int"
	Set gtmtypfldindx("file_log","cnt")=1
	Set gtmtypes("file_log",2,"name")="file_log.ent"
	Set gtmtypes("file_log",2,"off")=4
	Set gtmtypes("file_log",2,"len")=260
	Set gtmtypes("file_log",2,"type")="file_info"
	Set gtmtypfldindx("file_log","ent")=2
	Set gtmtypes("file_log",2,"dim")=5
	Set gtmtypes("file_log",3,"name")="file_log.ent[0].fac"
	Set gtmtypes("file_log",3,"off")=4
	Set gtmtypes("file_log",3,"len")=4
	Set gtmtypes("file_log",3,"type")="int"
	Set gtmtypfldindx("file_log","ent[0].fac")=3
	Set gtmtypes("file_log",4,"name")="file_log.ent[0].dat"
	Set gtmtypes("file_log",4,"off")=8
	Set gtmtypes("file_log",4,"len")=8
	Set gtmtypes("file_log",4,"type")="short"
	Set gtmtypfldindx("file_log","ent[0].dat")=4
	Set gtmtypes("file_log",4,"dim")=4
	Set gtmtypes("file_log",5,"name")="file_log.ent[0].usr"
	Set gtmtypes("file_log",5,"off")=16
	Set gtmtypes("file_log",5,"len")=31
	Set gtmtypes("file_log",5,"type")="char"
	Set gtmtypfldindx("file_log","ent[0].usr")=5
	Set gtmtypes("file_log",6,"name")="file_log.ent[0].trm"
	Set gtmtypes("file_log",6,"off")=47
	Set gtmtypes("file_log",6,"len")=7
	Set gtmtypes("file_log",6,"type")="char"
	Set gtmtypfldindx("file_log","ent[0].trm")=6
	Set gtmtypes("file_log",7,"name")="file_log.ent[0].filler"
	Set gtmtypes("file_log",7,"off")=54
	Set gtmtypes("file_log",7,"len")=2
	Set gtmtypes("file_log",7,"type")="char"
	Set gtmtypfldindx("file_log","ent[0].filler")=7
	;
	Set gtmtypes("file_pointer")="struct"
	Set gtmtypes("file_pointer",0)=5
	Set gtmtypes("file_pointer","len")=24
	Set gtmtypes("file_pointer",1,"name")="file_pointer.fd"
	Set gtmtypes("file_pointer",1,"off")=0
	Set gtmtypes("file_pointer",1,"len")=4
	Set gtmtypes("file_pointer",1,"type")="int"
	Set gtmtypfldindx("file_pointer","fd")=1
	Set gtmtypes("file_pointer",2,"name")="file_pointer.v"
	Set gtmtypes("file_pointer",2,"off")=8
	Set gtmtypes("file_pointer",2,"len")=16
	Set gtmtypes("file_pointer",2,"type")="mstr"
	Set gtmtypfldindx("file_pointer","v")=2
	Set gtmtypes("file_pointer",3,"name")="file_pointer.v.char_len"
	Set gtmtypes("file_pointer",3,"off")=8
	Set gtmtypes("file_pointer",3,"len")=4
	Set gtmtypes("file_pointer",3,"type")="unsigned-int"
	Set gtmtypfldindx("file_pointer","v.char_len")=3
	Set gtmtypes("file_pointer",4,"name")="file_pointer.v.len"
	Set gtmtypes("file_pointer",4,"off")=12
	Set gtmtypes("file_pointer",4,"len")=4
	Set gtmtypes("file_pointer",4,"type")="int"
	Set gtmtypfldindx("file_pointer","v.len")=4
	Set gtmtypes("file_pointer",5,"name")="file_pointer.v.addr"
	Set gtmtypes("file_pointer",5,"off")=16
	Set gtmtypes("file_pointer",5,"len")=8
	Set gtmtypes("file_pointer",5,"type")="addr"
	Set gtmtypfldindx("file_pointer","v.addr")=5
	;
	Set gtmtypes("fnpc")="struct"
	Set gtmtypes("fnpc",0)=9
	Set gtmtypes("fnpc","len")=360
	Set gtmtypes("fnpc",1,"name")="fnpc.last_str"
	Set gtmtypes("fnpc",1,"off")=0
	Set gtmtypes("fnpc",1,"len")=16
	Set gtmtypes("fnpc",1,"type")="mstr"
	Set gtmtypfldindx("fnpc","last_str")=1
	Set gtmtypes("fnpc",2,"name")="fnpc.last_str.char_len"
	Set gtmtypes("fnpc",2,"off")=0
	Set gtmtypes("fnpc",2,"len")=4
	Set gtmtypes("fnpc",2,"type")="unsigned-int"
	Set gtmtypfldindx("fnpc","last_str.char_len")=2
	Set gtmtypes("fnpc",3,"name")="fnpc.last_str.len"
	Set gtmtypes("fnpc",3,"off")=4
	Set gtmtypes("fnpc",3,"len")=4
	Set gtmtypes("fnpc",3,"type")="int"
	Set gtmtypfldindx("fnpc","last_str.len")=3
	Set gtmtypes("fnpc",4,"name")="fnpc.last_str.addr"
	Set gtmtypes("fnpc",4,"off")=8
	Set gtmtypes("fnpc",4,"len")=8
	Set gtmtypes("fnpc",4,"type")="addr"
	Set gtmtypfldindx("fnpc","last_str.addr")=4
	Set gtmtypes("fnpc",5,"name")="fnpc.delim"
	Set gtmtypes("fnpc",5,"off")=16
	Set gtmtypes("fnpc",5,"len")=4
	Set gtmtypes("fnpc",5,"type")="int"
	Set gtmtypfldindx("fnpc","delim")=5
	Set gtmtypes("fnpc",6,"name")="fnpc.npcs"
	Set gtmtypes("fnpc",6,"off")=20
	Set gtmtypes("fnpc",6,"len")=4
	Set gtmtypes("fnpc",6,"type")="int"
	Set gtmtypfldindx("fnpc","npcs")=6
	Set gtmtypes("fnpc",7,"name")="fnpc.indx"
	Set gtmtypes("fnpc",7,"off")=24
	Set gtmtypes("fnpc",7,"len")=4
	Set gtmtypes("fnpc",7,"type")="int"
	Set gtmtypfldindx("fnpc","indx")=7
	Set gtmtypes("fnpc",8,"name")="fnpc.byte_oriented"
	Set gtmtypes("fnpc",8,"off")=28
	Set gtmtypes("fnpc",8,"len")=4
	Set gtmtypes("fnpc",8,"type")="boolean_t"
	Set gtmtypfldindx("fnpc","byte_oriented")=8
	Set gtmtypes("fnpc",9,"name")="fnpc.pstart"
	Set gtmtypes("fnpc",9,"off")=32
	Set gtmtypes("fnpc",9,"len")=328
	Set gtmtypes("fnpc",9,"type")="unsigned-int"
	Set gtmtypfldindx("fnpc","pstart")=9
	Set gtmtypes("fnpc",9,"dim")=82
	;
	Set gtmtypes("fnpc_area")="struct"
	Set gtmtypes("fnpc_area",0)=12
	Set gtmtypes("fnpc_area","len")=18016
	Set gtmtypes("fnpc_area",1,"name")="fnpc_area.fnpcsteal"
	Set gtmtypes("fnpc_area",1,"off")=0
	Set gtmtypes("fnpc_area",1,"len")=8
	Set gtmtypes("fnpc_area",1,"type")="addr"
	Set gtmtypfldindx("fnpc_area","fnpcsteal")=1
	Set gtmtypes("fnpc_area",2,"name")="fnpc_area.fnpcmax"
	Set gtmtypes("fnpc_area",2,"off")=8
	Set gtmtypes("fnpc_area",2,"len")=8
	Set gtmtypes("fnpc_area",2,"type")="addr"
	Set gtmtypfldindx("fnpc_area","fnpcmax")=2
	Set gtmtypes("fnpc_area",3,"name")="fnpc_area.fnpcs"
	Set gtmtypes("fnpc_area",3,"off")=16
	Set gtmtypes("fnpc_area",3,"len")=18000
	Set gtmtypes("fnpc_area",3,"type")="fnpc"
	Set gtmtypfldindx("fnpc_area","fnpcs")=3
	Set gtmtypes("fnpc_area",3,"dim")=50
	Set gtmtypes("fnpc_area",4,"name")="fnpc_area.fnpcs[0].last_str"
	Set gtmtypes("fnpc_area",4,"off")=16
	Set gtmtypes("fnpc_area",4,"len")=16
	Set gtmtypes("fnpc_area",4,"type")="mstr"
	Set gtmtypfldindx("fnpc_area","fnpcs[0].last_str")=4
	Set gtmtypes("fnpc_area",5,"name")="fnpc_area.fnpcs[0].last_str.char_len"
	Set gtmtypes("fnpc_area",5,"off")=16
	Set gtmtypes("fnpc_area",5,"len")=4
	Set gtmtypes("fnpc_area",5,"type")="unsigned-int"
	Set gtmtypfldindx("fnpc_area","fnpcs[0].last_str.char_len")=5
	Set gtmtypes("fnpc_area",6,"name")="fnpc_area.fnpcs[0].last_str.len"
	Set gtmtypes("fnpc_area",6,"off")=20
	Set gtmtypes("fnpc_area",6,"len")=4
	Set gtmtypes("fnpc_area",6,"type")="int"
	Set gtmtypfldindx("fnpc_area","fnpcs[0].last_str.len")=6
	Set gtmtypes("fnpc_area",7,"name")="fnpc_area.fnpcs[0].last_str.addr"
	Set gtmtypes("fnpc_area",7,"off")=24
	Set gtmtypes("fnpc_area",7,"len")=8
	Set gtmtypes("fnpc_area",7,"type")="addr"
	Set gtmtypfldindx("fnpc_area","fnpcs[0].last_str.addr")=7
	Set gtmtypes("fnpc_area",8,"name")="fnpc_area.fnpcs[0].delim"
	Set gtmtypes("fnpc_area",8,"off")=32
	Set gtmtypes("fnpc_area",8,"len")=4
	Set gtmtypes("fnpc_area",8,"type")="int"
	Set gtmtypfldindx("fnpc_area","fnpcs[0].delim")=8
	Set gtmtypes("fnpc_area",9,"name")="fnpc_area.fnpcs[0].npcs"
	Set gtmtypes("fnpc_area",9,"off")=36
	Set gtmtypes("fnpc_area",9,"len")=4
	Set gtmtypes("fnpc_area",9,"type")="int"
	Set gtmtypfldindx("fnpc_area","fnpcs[0].npcs")=9
	Set gtmtypes("fnpc_area",10,"name")="fnpc_area.fnpcs[0].indx"
	Set gtmtypes("fnpc_area",10,"off")=40
	Set gtmtypes("fnpc_area",10,"len")=4
	Set gtmtypes("fnpc_area",10,"type")="int"
	Set gtmtypfldindx("fnpc_area","fnpcs[0].indx")=10
	Set gtmtypes("fnpc_area",11,"name")="fnpc_area.fnpcs[0].byte_oriented"
	Set gtmtypes("fnpc_area",11,"off")=44
	Set gtmtypes("fnpc_area",11,"len")=4
	Set gtmtypes("fnpc_area",11,"type")="boolean_t"
	Set gtmtypfldindx("fnpc_area","fnpcs[0].byte_oriented")=11
	Set gtmtypes("fnpc_area",12,"name")="fnpc_area.fnpcs[0].pstart"
	Set gtmtypes("fnpc_area",12,"off")=48
	Set gtmtypes("fnpc_area",12,"len")=328
	Set gtmtypes("fnpc_area",12,"type")="unsigned-int"
	Set gtmtypfldindx("fnpc_area","fnpcs[0].pstart")=12
	Set gtmtypes("fnpc_area",12,"dim")=82
	;
	Set gtmtypes("forw_multi_struct")="struct"
	Set gtmtypes("forw_multi_struct",0)=16
	Set gtmtypes("forw_multi_struct","len")=80
	Set gtmtypes("forw_multi_struct",1,"name")="forw_multi_struct.u"
	Set gtmtypes("forw_multi_struct",1,"off")=0
	Set gtmtypes("forw_multi_struct",1,"len")=16
	Set gtmtypes("forw_multi_struct",1,"type")="union"
	Set gtmtypfldindx("forw_multi_struct","u")=1
	Set gtmtypes("forw_multi_struct",2,"name")="forw_multi_struct.u.tabent"
	Set gtmtypes("forw_multi_struct",2,"off")=0
	Set gtmtypes("forw_multi_struct",2,"len")=8
	Set gtmtypes("forw_multi_struct",2,"type")="addr"
	Set gtmtypfldindx("forw_multi_struct","u.tabent")=2
	Set gtmtypes("forw_multi_struct",3,"name")="forw_multi_struct.u.free_que"
	Set gtmtypes("forw_multi_struct",3,"off")=0
	Set gtmtypes("forw_multi_struct",3,"len")=16
	Set gtmtypes("forw_multi_struct",3,"type")="que_ent"
	Set gtmtypfldindx("forw_multi_struct","u.free_que")=3
	Set gtmtypes("forw_multi_struct",4,"name")="forw_multi_struct.u.free_que.fl"
	Set gtmtypes("forw_multi_struct",4,"off")=0
	Set gtmtypes("forw_multi_struct",4,"len")=8
	Set gtmtypes("forw_multi_struct",4,"type")="intptr_t"
	Set gtmtypfldindx("forw_multi_struct","u.free_que.fl")=4
	Set gtmtypes("forw_multi_struct",5,"name")="forw_multi_struct.u.free_que.bl"
	Set gtmtypes("forw_multi_struct",5,"off")=8
	Set gtmtypes("forw_multi_struct",5,"len")=8
	Set gtmtypes("forw_multi_struct",5,"type")="intptr_t"
	Set gtmtypfldindx("forw_multi_struct","u.free_que.bl")=5
	Set gtmtypes("forw_multi_struct",6,"name")="forw_multi_struct.token"
	Set gtmtypes("forw_multi_struct",6,"off")=16
	Set gtmtypes("forw_multi_struct",6,"len")=8
	Set gtmtypes("forw_multi_struct",6,"type")="uint64_t"
	Set gtmtypfldindx("forw_multi_struct","token")=6
	Set gtmtypes("forw_multi_struct",7,"name")="forw_multi_struct.first_tp_rctl"
	Set gtmtypes("forw_multi_struct",7,"off")=24
	Set gtmtypes("forw_multi_struct",7,"len")=8
	Set gtmtypes("forw_multi_struct",7,"type")="addr"
	Set gtmtypfldindx("forw_multi_struct","first_tp_rctl")=7
	Set gtmtypes("forw_multi_struct",8,"name")="forw_multi_struct.next"
	Set gtmtypes("forw_multi_struct",8,"off")=32
	Set gtmtypes("forw_multi_struct",8,"len")=8
	Set gtmtypes("forw_multi_struct",8,"type")="addr"
	Set gtmtypfldindx("forw_multi_struct","next")=8
	Set gtmtypes("forw_multi_struct",9,"name")="forw_multi_struct.shm_forw_multi"
	Set gtmtypes("forw_multi_struct",9,"off")=40
	Set gtmtypes("forw_multi_struct",9,"len")=8
	Set gtmtypes("forw_multi_struct",9,"type")="addr"
	Set gtmtypfldindx("forw_multi_struct","shm_forw_multi")=9
	Set gtmtypes("forw_multi_struct",10,"name")="forw_multi_struct.multi"
	Set gtmtypes("forw_multi_struct",10,"off")=48
	Set gtmtypes("forw_multi_struct",10,"len")=8
	Set gtmtypes("forw_multi_struct",10,"type")="addr"
	Set gtmtypfldindx("forw_multi_struct","multi")=10
	Set gtmtypes("forw_multi_struct",11,"name")="forw_multi_struct.time"
	Set gtmtypes("forw_multi_struct",11,"off")=56
	Set gtmtypes("forw_multi_struct",11,"len")=4
	Set gtmtypes("forw_multi_struct",11,"type")="unsigned-int"
	Set gtmtypfldindx("forw_multi_struct","time")=11
	Set gtmtypes("forw_multi_struct",12,"name")="forw_multi_struct.recstat"
	Set gtmtypes("forw_multi_struct",12,"off")=60
	Set gtmtypes("forw_multi_struct",12,"len")=4
	Set gtmtypes("forw_multi_struct",12,"type")="int"
	Set gtmtypfldindx("forw_multi_struct","recstat")=12
	Set gtmtypes("forw_multi_struct",13,"name")="forw_multi_struct.num_reg_total"
	Set gtmtypes("forw_multi_struct",13,"off")=64
	Set gtmtypes("forw_multi_struct",13,"len")=4
	Set gtmtypes("forw_multi_struct",13,"type")="unsigned-int"
	Set gtmtypfldindx("forw_multi_struct","num_reg_total")=13
	Set gtmtypes("forw_multi_struct",14,"name")="forw_multi_struct.num_reg_seen_backward"
	Set gtmtypes("forw_multi_struct",14,"off")=68
	Set gtmtypes("forw_multi_struct",14,"len")=4
	Set gtmtypes("forw_multi_struct",14,"type")="unsigned-int"
	Set gtmtypfldindx("forw_multi_struct","num_reg_seen_backward")=14
	Set gtmtypes("forw_multi_struct",15,"name")="forw_multi_struct.num_reg_seen_forward"
	Set gtmtypes("forw_multi_struct",15,"off")=72
	Set gtmtypes("forw_multi_struct",15,"len")=4
	Set gtmtypes("forw_multi_struct",15,"type")="unsigned-int"
	Set gtmtypfldindx("forw_multi_struct","num_reg_seen_forward")=15
	Set gtmtypes("forw_multi_struct",16,"name")="forw_multi_struct.no_longer_stuck"
	Set gtmtypes("forw_multi_struct",16,"off")=76
	Set gtmtypes("forw_multi_struct",16,"len")=4
	Set gtmtypes("forw_multi_struct",16,"type")="boolean_t"
	Set gtmtypfldindx("forw_multi_struct","no_longer_stuck")=16
	;
	Set gtmtypes("freeze_multiproc_state")="struct"
	Set gtmtypes("freeze_multiproc_state",0)=10
	Set gtmtypes("freeze_multiproc_state","len")=24
	Set gtmtypes("freeze_multiproc_state",1,"name")="freeze_multiproc_state.ntasks"
	Set gtmtypes("freeze_multiproc_state",1,"off")=0
	Set gtmtypes("freeze_multiproc_state",1,"len")=4
	Set gtmtypes("freeze_multiproc_state",1,"type")="int"
	Set gtmtypfldindx("freeze_multiproc_state","ntasks")=1
	Set gtmtypes("freeze_multiproc_state",2,"name")="freeze_multiproc_state.grab_crit_counter"
	Set gtmtypes("freeze_multiproc_state",2,"off")=4
	Set gtmtypes("freeze_multiproc_state",2,"len")=4
	Set gtmtypes("freeze_multiproc_state",2,"type")="int"
	Set gtmtypfldindx("freeze_multiproc_state","grab_crit_counter")=2
	Set gtmtypes("freeze_multiproc_state",3,"name")="freeze_multiproc_state.region_frozen_counter"
	Set gtmtypes("freeze_multiproc_state",3,"off")=8
	Set gtmtypes("freeze_multiproc_state",3,"len")=4
	Set gtmtypes("freeze_multiproc_state",3,"type")="int"
	Set gtmtypfldindx("freeze_multiproc_state","region_frozen_counter")=3
	Set gtmtypes("freeze_multiproc_state",4,"name")="freeze_multiproc_state.region_frozen_latch"
	Set gtmtypes("freeze_multiproc_state",4,"off")=16
	Set gtmtypes("freeze_multiproc_state",4,"len")=8
	Set gtmtypes("freeze_multiproc_state",4,"type")="global_latch_t"
	Set gtmtypfldindx("freeze_multiproc_state","region_frozen_latch")=4
	Set gtmtypes("freeze_multiproc_state",5,"name")="freeze_multiproc_state.region_frozen_latch.u"
	Set gtmtypes("freeze_multiproc_state",5,"off")=16
	Set gtmtypes("freeze_multiproc_state",5,"len")=8
	Set gtmtypes("freeze_multiproc_state",5,"type")="union"
	Set gtmtypfldindx("freeze_multiproc_state","region_frozen_latch.u")=5
	Set gtmtypes("freeze_multiproc_state",6,"name")="freeze_multiproc_state.region_frozen_latch.u.pid_imgcnt"
	Set gtmtypes("freeze_multiproc_state",6,"off")=16
	Set gtmtypes("freeze_multiproc_state",6,"len")=8
	Set gtmtypes("freeze_multiproc_state",6,"type")="uint64_t"
	Set gtmtypfldindx("freeze_multiproc_state","region_frozen_latch.u.pid_imgcnt")=6
	Set gtmtypes("freeze_multiproc_state",7,"name")="freeze_multiproc_state.region_frozen_latch.u.parts"
	Set gtmtypes("freeze_multiproc_state",7,"off")=16
	Set gtmtypes("freeze_multiproc_state",7,"len")=8
	Set gtmtypes("freeze_multiproc_state",7,"type")="struct"
	Set gtmtypfldindx("freeze_multiproc_state","region_frozen_latch.u.parts")=7
	Set gtmtypes("freeze_multiproc_state",8,"name")="freeze_multiproc_state.region_frozen_latch.u.parts.latch_pid"
	Set gtmtypes("freeze_multiproc_state",8,"off")=16
	Set gtmtypes("freeze_multiproc_state",8,"len")=4
	Set gtmtypes("freeze_multiproc_state",8,"type")="int"
	Set gtmtypfldindx("freeze_multiproc_state","region_frozen_latch.u.parts.latch_pid")=8
	Set gtmtypes("freeze_multiproc_state",9,"name")="freeze_multiproc_state.region_frozen_latch.u.parts.latch_word"
	Set gtmtypes("freeze_multiproc_state",9,"off")=20
	Set gtmtypes("freeze_multiproc_state",9,"len")=4
	Set gtmtypes("freeze_multiproc_state",9,"type")="int"
	Set gtmtypfldindx("freeze_multiproc_state","region_frozen_latch.u.parts.latch_word")=9
	Set gtmtypes("freeze_multiproc_state",10,"name")="freeze_multiproc_state.freeze_ret_array"
	Set gtmtypes("freeze_multiproc_state",10,"off")=24
	Set gtmtypes("freeze_multiproc_state",10,"len")=0
	Set gtmtypes("freeze_multiproc_state",10,"type")="int"
	Set gtmtypfldindx("freeze_multiproc_state","freeze_ret_array")=10
	;
	Set gtmtypes("freeze_reg_mp_state")="struct"
	Set gtmtypes("freeze_reg_mp_state",0)=2
	Set gtmtypes("freeze_reg_mp_state","len")=16
	Set gtmtypes("freeze_reg_mp_state",1,"name")="freeze_reg_mp_state.region_index"
	Set gtmtypes("freeze_reg_mp_state",1,"off")=0
	Set gtmtypes("freeze_reg_mp_state",1,"len")=4
	Set gtmtypes("freeze_reg_mp_state",1,"type")="int"
	Set gtmtypfldindx("freeze_reg_mp_state","region_index")=1
	Set gtmtypes("freeze_reg_mp_state",2,"name")="freeze_reg_mp_state.pfms"
	Set gtmtypes("freeze_reg_mp_state",2,"off")=8
	Set gtmtypes("freeze_reg_mp_state",2,"len")=8
	Set gtmtypes("freeze_reg_mp_state",2,"type")="addr"
	Set gtmtypfldindx("freeze_reg_mp_state","pfms")=2
	;
	Set gtmtypes("ftokhist")="struct"
	Set gtmtypes("ftokhist",0)=3
	Set gtmtypes("ftokhist","len")=16
	Set gtmtypes("ftokhist",1,"name")="ftokhist.ftok_oper"
	Set gtmtypes("ftokhist",1,"off")=0
	Set gtmtypes("ftokhist",1,"len")=4
	Set gtmtypes("ftokhist",1,"type")="int"
	Set gtmtypfldindx("ftokhist","ftok_oper")=1
	Set gtmtypes("ftokhist",2,"name")="ftokhist.process_id"
	Set gtmtypes("ftokhist",2,"off")=4
	Set gtmtypes("ftokhist",2,"len")=4
	Set gtmtypes("ftokhist",2,"type")="unsigned-int"
	Set gtmtypfldindx("ftokhist","process_id")=2
	Set gtmtypes("ftokhist",3,"name")="ftokhist.cr_tn"
	Set gtmtypes("ftokhist",3,"off")=8
	Set gtmtypes("ftokhist",3,"len")=8
	Set gtmtypes("ftokhist",3,"type")="uint64_t"
	Set gtmtypfldindx("ftokhist","cr_tn")=3
	;
	Set gtmtypes("fun_data_type")="struct"
	Set gtmtypes("fun_data_type",0)=2
	Set gtmtypes("fun_data_type","len")=8
	Set gtmtypes("fun_data_type",1,"name")="fun_data_type.opcode"
	Set gtmtypes("fun_data_type",1,"off")=0
	Set gtmtypes("fun_data_type",1,"len")=4
	Set gtmtypes("fun_data_type",1,"type")="unsigned-int"
	Set gtmtypfldindx("fun_data_type","opcode")=1
	Set gtmtypes("fun_data_type",2,"name")="fun_data_type.os_syst"
	Set gtmtypes("fun_data_type",2,"off")=4
	Set gtmtypes("fun_data_type",2,"len")=1
	Set gtmtypes("fun_data_type",2,"type")="char"
	Set gtmtypfldindx("fun_data_type","os_syst")=2
	;
	Set gtmtypes("gcall_args")="struct"
	Set gtmtypes("gcall_args",0)=6
	Set gtmtypes("gcall_args","len")=296
	Set gtmtypes("gcall_args",1,"name")="gcall_args.callargs"
	Set gtmtypes("gcall_args",1,"off")=0
	Set gtmtypes("gcall_args",1,"len")=8
	Set gtmtypes("gcall_args",1,"type")="intptr_t"
	Set gtmtypfldindx("gcall_args","callargs")=1
	Set gtmtypes("gcall_args",2,"name")="gcall_args.truth"
	Set gtmtypes("gcall_args",2,"off")=8
	Set gtmtypes("gcall_args",2,"len")=8
	Set gtmtypes("gcall_args",2,"type")="intptr_t"
	Set gtmtypfldindx("gcall_args","truth")=2
	Set gtmtypes("gcall_args",3,"name")="gcall_args.retval"
	Set gtmtypes("gcall_args",3,"off")=16
	Set gtmtypes("gcall_args",3,"len")=8
	Set gtmtypes("gcall_args",3,"type")="intptr_t"
	Set gtmtypfldindx("gcall_args","retval")=3
	Set gtmtypes("gcall_args",4,"name")="gcall_args.mask"
	Set gtmtypes("gcall_args",4,"off")=24
	Set gtmtypes("gcall_args",4,"len")=8
	Set gtmtypes("gcall_args",4,"type")="intptr_t"
	Set gtmtypfldindx("gcall_args","mask")=4
	Set gtmtypes("gcall_args",5,"name")="gcall_args.argcnt"
	Set gtmtypes("gcall_args",5,"off")=32
	Set gtmtypes("gcall_args",5,"len")=8
	Set gtmtypes("gcall_args",5,"type")="intptr_t"
	Set gtmtypfldindx("gcall_args","argcnt")=5
	Set gtmtypes("gcall_args",6,"name")="gcall_args.argval"
	Set gtmtypes("gcall_args",6,"off")=40
	Set gtmtypes("gcall_args",6,"len")=256
	Set gtmtypes("gcall_args",6,"type")="addr"
	Set gtmtypfldindx("gcall_args","argval")=6
	Set gtmtypes("gcall_args",6,"dim")=32
	;
	Set gtmtypes("gd_addr")="struct"
	Set gtmtypes("gd_addr",0)=21
	Set gtmtypes("gd_addr","len")=120
	Set gtmtypes("gd_addr",1,"name")="gd_addr.local_locks"
	Set gtmtypes("gd_addr",1,"off")=0
	Set gtmtypes("gd_addr",1,"len")=8
	Set gtmtypes("gd_addr",1,"type")="addr"
	Set gtmtypfldindx("gd_addr","local_locks")=1
	Set gtmtypes("gd_addr",2,"name")="gd_addr.max_rec_size"
	Set gtmtypes("gd_addr",2,"off")=8
	Set gtmtypes("gd_addr",2,"len")=4
	Set gtmtypes("gd_addr",2,"type")="int"
	Set gtmtypfldindx("gd_addr","max_rec_size")=2
	Set gtmtypes("gd_addr",3,"name")="gd_addr.n_maps"
	Set gtmtypes("gd_addr",3,"off")=12
	Set gtmtypes("gd_addr",3,"len")=4
	Set gtmtypes("gd_addr",3,"type")="unsigned-int"
	Set gtmtypfldindx("gd_addr","n_maps")=3
	Set gtmtypes("gd_addr",4,"name")="gd_addr.n_regions"
	Set gtmtypes("gd_addr",4,"off")=16
	Set gtmtypes("gd_addr",4,"len")=4
	Set gtmtypes("gd_addr",4,"type")="unsigned-int"
	Set gtmtypfldindx("gd_addr","n_regions")=4
	Set gtmtypes("gd_addr",5,"name")="gd_addr.n_segments"
	Set gtmtypes("gd_addr",5,"off")=20
	Set gtmtypes("gd_addr",5,"len")=4
	Set gtmtypes("gd_addr",5,"type")="unsigned-int"
	Set gtmtypfldindx("gd_addr","n_segments")=5
	Set gtmtypes("gd_addr",6,"name")="gd_addr.n_gblnames"
	Set gtmtypes("gd_addr",6,"off")=24
	Set gtmtypes("gd_addr",6,"len")=4
	Set gtmtypes("gd_addr",6,"type")="unsigned-int"
	Set gtmtypfldindx("gd_addr","n_gblnames")=6
	Set gtmtypes("gd_addr",7,"name")="gd_addr.var_maps_len"
	Set gtmtypes("gd_addr",7,"off")=28
	Set gtmtypes("gd_addr",7,"len")=4
	Set gtmtypes("gd_addr",7,"type")="unsigned-int"
	Set gtmtypfldindx("gd_addr","var_maps_len")=7
	Set gtmtypes("gd_addr",8,"name")="gd_addr.maps"
	Set gtmtypes("gd_addr",8,"off")=32
	Set gtmtypes("gd_addr",8,"len")=8
	Set gtmtypes("gd_addr",8,"type")="addr"
	Set gtmtypfldindx("gd_addr","maps")=8
	Set gtmtypes("gd_addr",9,"name")="gd_addr.regions"
	Set gtmtypes("gd_addr",9,"off")=40
	Set gtmtypes("gd_addr",9,"len")=8
	Set gtmtypes("gd_addr",9,"type")="addr"
	Set gtmtypfldindx("gd_addr","regions")=9
	Set gtmtypes("gd_addr",10,"name")="gd_addr.segments"
	Set gtmtypes("gd_addr",10,"off")=48
	Set gtmtypes("gd_addr",10,"len")=8
	Set gtmtypes("gd_addr",10,"type")="addr"
	Set gtmtypfldindx("gd_addr","segments")=10
	Set gtmtypes("gd_addr",11,"name")="gd_addr.gblnames"
	Set gtmtypes("gd_addr",11,"off")=56
	Set gtmtypes("gd_addr",11,"len")=8
	Set gtmtypes("gd_addr",11,"type")="addr"
	Set gtmtypfldindx("gd_addr","gblnames")=11
	Set gtmtypes("gd_addr",12,"name")="gd_addr.instinfo"
	Set gtmtypes("gd_addr",12,"off")=64
	Set gtmtypes("gd_addr",12,"len")=8
	Set gtmtypes("gd_addr",12,"type")="addr"
	Set gtmtypfldindx("gd_addr","instinfo")=12
	Set gtmtypes("gd_addr",13,"name")="gd_addr.link"
	Set gtmtypes("gd_addr",13,"off")=72
	Set gtmtypes("gd_addr",13,"len")=8
	Set gtmtypes("gd_addr",13,"type")="addr"
	Set gtmtypfldindx("gd_addr","link")=13
	Set gtmtypes("gd_addr",14,"name")="gd_addr.tab_ptr"
	Set gtmtypes("gd_addr",14,"off")=80
	Set gtmtypes("gd_addr",14,"len")=8
	Set gtmtypes("gd_addr",14,"type")="addr"
	Set gtmtypfldindx("gd_addr","tab_ptr")=14
	Set gtmtypes("gd_addr",15,"name")="gd_addr.id"
	Set gtmtypes("gd_addr",15,"off")=88
	Set gtmtypes("gd_addr",15,"len")=8
	Set gtmtypes("gd_addr",15,"type")="addr"
	Set gtmtypfldindx("gd_addr","id")=15
	Set gtmtypes("gd_addr",16,"name")="gd_addr.end"
	Set gtmtypes("gd_addr",16,"off")=96
	Set gtmtypes("gd_addr",16,"len")=8
	Set gtmtypes("gd_addr",16,"type")="uintptr_t"
	Set gtmtypfldindx("gd_addr","end")=16
	Set gtmtypes("gd_addr",17,"name")="gd_addr.has_span_gbls"
	Set gtmtypes("gd_addr",17,"off")=104
	Set gtmtypes("gd_addr",17,"len")=4
	Set gtmtypes("gd_addr",17,"type")="unsigned-int"
	Set gtmtypfldindx("gd_addr","has_span_gbls")=17
	Set gtmtypes("gd_addr",18,"name")="gd_addr.ygs_map_entry_changed"
	Set gtmtypes("gd_addr",18,"off")=108
	Set gtmtypes("gd_addr",18,"len")=1
	Set gtmtypes("gd_addr",18,"type")="char"
	Set gtmtypfldindx("gd_addr","ygs_map_entry_changed")=18
	Set gtmtypes("gd_addr",19,"name")="gd_addr.is_dummy_gbldir"
	Set gtmtypes("gd_addr",19,"off")=109
	Set gtmtypes("gd_addr",19,"len")=1
	Set gtmtypes("gd_addr",19,"type")="char"
	Set gtmtypfldindx("gd_addr","is_dummy_gbldir")=19
	Set gtmtypes("gd_addr",20,"name")="gd_addr.filler"
	Set gtmtypes("gd_addr",20,"off")=110
	Set gtmtypes("gd_addr",20,"len")=2
	Set gtmtypes("gd_addr",20,"type")="char"
	Set gtmtypfldindx("gd_addr","filler")=20
	Set gtmtypes("gd_addr",21,"name")="gd_addr.thread_gdi"
	Set gtmtypes("gd_addr",21,"off")=112
	Set gtmtypes("gd_addr",21,"len")=8
	Set gtmtypes("gd_addr",21,"type")="addr"
	Set gtmtypfldindx("gd_addr","thread_gdi")=21
	;
	Set gtmtypes("gd_binding")="struct"
	Set gtmtypes("gd_binding",0)=8
	Set gtmtypes("gd_binding","len")=24
	Set gtmtypes("gd_binding",1,"name")="gd_binding.gvkey"
	Set gtmtypes("gd_binding",1,"off")=0
	Set gtmtypes("gd_binding",1,"len")=8
	Set gtmtypes("gd_binding",1,"type")="union"
	Set gtmtypfldindx("gd_binding","gvkey")=1
	Set gtmtypes("gd_binding",2,"name")="gd_binding.gvkey.addr"
	Set gtmtypes("gd_binding",2,"off")=0
	Set gtmtypes("gd_binding",2,"len")=8
	Set gtmtypes("gd_binding",2,"type")="addr"
	Set gtmtypfldindx("gd_binding","gvkey.addr")=2
	Set gtmtypes("gd_binding",3,"name")="gd_binding.gvkey.offset"
	Set gtmtypes("gd_binding",3,"off")=0
	Set gtmtypes("gd_binding",3,"len")=4
	Set gtmtypes("gd_binding",3,"type")="unsigned-int"
	Set gtmtypfldindx("gd_binding","gvkey.offset")=3
	Set gtmtypes("gd_binding",4,"name")="gd_binding.reg"
	Set gtmtypes("gd_binding",4,"off")=8
	Set gtmtypes("gd_binding",4,"len")=8
	Set gtmtypes("gd_binding",4,"type")="union"
	Set gtmtypfldindx("gd_binding","reg")=4
	Set gtmtypes("gd_binding",5,"name")="gd_binding.reg.addr"
	Set gtmtypes("gd_binding",5,"off")=8
	Set gtmtypes("gd_binding",5,"len")=8
	Set gtmtypes("gd_binding",5,"type")="addr"
	Set gtmtypfldindx("gd_binding","reg.addr")=5
	Set gtmtypes("gd_binding",6,"name")="gd_binding.reg.offset"
	Set gtmtypes("gd_binding",6,"off")=8
	Set gtmtypes("gd_binding",6,"len")=4
	Set gtmtypes("gd_binding",6,"type")="unsigned-int"
	Set gtmtypfldindx("gd_binding","reg.offset")=6
	Set gtmtypes("gd_binding",7,"name")="gd_binding.gvname_len"
	Set gtmtypes("gd_binding",7,"off")=16
	Set gtmtypes("gd_binding",7,"len")=4
	Set gtmtypes("gd_binding",7,"type")="unsigned-int"
	Set gtmtypfldindx("gd_binding","gvname_len")=7
	Set gtmtypes("gd_binding",8,"name")="gd_binding.gvkey_len"
	Set gtmtypes("gd_binding",8,"off")=20
	Set gtmtypes("gd_binding",8,"len")=4
	Set gtmtypes("gd_binding",8,"type")="unsigned-int"
	Set gtmtypfldindx("gd_binding","gvkey_len")=8
	;
	Set gtmtypes("gd_gblname")="struct"
	Set gtmtypes("gd_gblname",0)=3
	Set gtmtypes("gd_gblname","len")=40
	Set gtmtypes("gd_gblname",1,"name")="gd_gblname.gblname"
	Set gtmtypes("gd_gblname",1,"off")=0
	Set gtmtypes("gd_gblname",1,"len")=32
	Set gtmtypes("gd_gblname",1,"type")="unsigned-char"
	Set gtmtypfldindx("gd_gblname","gblname")=1
	Set gtmtypes("gd_gblname",2,"name")="gd_gblname.act"
	Set gtmtypes("gd_gblname",2,"off")=32
	Set gtmtypes("gd_gblname",2,"len")=4
	Set gtmtypes("gd_gblname",2,"type")="unsigned-int"
	Set gtmtypfldindx("gd_gblname","act")=2
	Set gtmtypes("gd_gblname",3,"name")="gd_gblname.ver"
	Set gtmtypes("gd_gblname",3,"off")=36
	Set gtmtypes("gd_gblname",3,"len")=4
	Set gtmtypes("gd_gblname",3,"type")="unsigned-int"
	Set gtmtypfldindx("gd_gblname","ver")=3
	;
	Set gtmtypes("gd_inst_info")="struct"
	Set gtmtypes("gd_inst_info",0)=1
	Set gtmtypes("gd_inst_info","len")=256
	Set gtmtypes("gd_inst_info",1,"name")="gd_inst_info.instfilename"
	Set gtmtypes("gd_inst_info",1,"off")=0
	Set gtmtypes("gd_inst_info",1,"len")=256
	Set gtmtypes("gd_inst_info",1,"type")="char"
	Set gtmtypfldindx("gd_inst_info","instfilename")=1
	;
	Set gtmtypes("gd_region")="struct"
	Set gtmtypes("gd_region",0)=45
	Set gtmtypes("gd_region","len")=432
	Set gtmtypes("gd_region",1,"name")="gd_region.rname_len"
	Set gtmtypes("gd_region",1,"off")=0
	Set gtmtypes("gd_region",1,"len")=2
	Set gtmtypes("gd_region",1,"type")="unsigned-short"
	Set gtmtypfldindx("gd_region","rname_len")=1
	Set gtmtypes("gd_region",2,"name")="gd_region.rname"
	Set gtmtypes("gd_region",2,"off")=2
	Set gtmtypes("gd_region",2,"len")=32
	Set gtmtypes("gd_region",2,"type")="unsigned-char"
	Set gtmtypfldindx("gd_region","rname")=2
	Set gtmtypes("gd_region",3,"name")="gd_region.max_key_size"
	Set gtmtypes("gd_region",3,"off")=34
	Set gtmtypes("gd_region",3,"len")=2
	Set gtmtypes("gd_region",3,"type")="unsigned-short"
	Set gtmtypfldindx("gd_region","max_key_size")=3
	Set gtmtypes("gd_region",4,"name")="gd_region.max_rec_size"
	Set gtmtypes("gd_region",4,"off")=36
	Set gtmtypes("gd_region",4,"len")=4
	Set gtmtypes("gd_region",4,"type")="unsigned-int"
	Set gtmtypfldindx("gd_region","max_rec_size")=4
	Set gtmtypes("gd_region",5,"name")="gd_region.dyn"
	Set gtmtypes("gd_region",5,"off")=40
	Set gtmtypes("gd_region",5,"len")=8
	Set gtmtypes("gd_region",5,"type")="gd_seg_addr"
	Set gtmtypfldindx("gd_region","dyn")=5
	Set gtmtypes("gd_region",6,"name")="gd_region.dyn.offset"
	Set gtmtypes("gd_region",6,"off")=40
	Set gtmtypes("gd_region",6,"len")=4
	Set gtmtypes("gd_region",6,"type")="int"
	Set gtmtypfldindx("gd_region","dyn.offset")=6
	Set gtmtypes("gd_region",7,"name")="gd_region.dyn.addr"
	Set gtmtypes("gd_region",7,"off")=40
	Set gtmtypes("gd_region",7,"len")=8
	Set gtmtypes("gd_region",7,"type")="addr"
	Set gtmtypfldindx("gd_region","dyn.addr")=7
	Set gtmtypes("gd_region",8,"name")="gd_region.stat"
	Set gtmtypes("gd_region",8,"off")=48
	Set gtmtypes("gd_region",8,"len")=8
	Set gtmtypes("gd_region",8,"type")="gd_seg_addr"
	Set gtmtypfldindx("gd_region","stat")=8
	Set gtmtypes("gd_region",9,"name")="gd_region.stat.offset"
	Set gtmtypes("gd_region",9,"off")=48
	Set gtmtypes("gd_region",9,"len")=4
	Set gtmtypes("gd_region",9,"type")="int"
	Set gtmtypfldindx("gd_region","stat.offset")=9
	Set gtmtypes("gd_region",10,"name")="gd_region.stat.addr"
	Set gtmtypes("gd_region",10,"off")=48
	Set gtmtypes("gd_region",10,"len")=8
	Set gtmtypes("gd_region",10,"type")="addr"
	Set gtmtypfldindx("gd_region","stat.addr")=10
	Set gtmtypes("gd_region",11,"name")="gd_region.open"
	Set gtmtypes("gd_region",11,"off")=56
	Set gtmtypes("gd_region",11,"len")=1
	Set gtmtypes("gd_region",11,"type")="char"
	Set gtmtypfldindx("gd_region","open")=11
	Set gtmtypes("gd_region",12,"name")="gd_region.lock_write"
	Set gtmtypes("gd_region",12,"off")=57
	Set gtmtypes("gd_region",12,"len")=1
	Set gtmtypes("gd_region",12,"type")="char"
	Set gtmtypfldindx("gd_region","lock_write")=12
	Set gtmtypes("gd_region",13,"name")="gd_region.null_subs"
	Set gtmtypes("gd_region",13,"off")=58
	Set gtmtypes("gd_region",13,"len")=1
	Set gtmtypes("gd_region",13,"type")="char"
	Set gtmtypfldindx("gd_region","null_subs")=13
	Set gtmtypes("gd_region",14,"name")="gd_region.jnl_state"
	Set gtmtypes("gd_region",14,"off")=59
	Set gtmtypes("gd_region",14,"len")=1
	Set gtmtypes("gd_region",14,"type")="unsigned-char"
	Set gtmtypfldindx("gd_region","jnl_state")=14
	Set gtmtypes("gd_region",15,"name")="gd_region.jnl_alq"
	Set gtmtypes("gd_region",15,"off")=60
	Set gtmtypes("gd_region",15,"len")=4
	Set gtmtypes("gd_region",15,"type")="unsigned-int"
	Set gtmtypfldindx("gd_region","jnl_alq")=15
	Set gtmtypes("gd_region",16,"name")="gd_region.jnl_deq"
	Set gtmtypes("gd_region",16,"off")=64
	Set gtmtypes("gd_region",16,"len")=4
	Set gtmtypes("gd_region",16,"type")="unsigned-int"
	Set gtmtypfldindx("gd_region","jnl_deq")=16
	Set gtmtypes("gd_region",17,"name")="gd_region.jnl_autoswitchlimit"
	Set gtmtypes("gd_region",17,"off")=68
	Set gtmtypes("gd_region",17,"len")=4
	Set gtmtypes("gd_region",17,"type")="unsigned-int"
	Set gtmtypfldindx("gd_region","jnl_autoswitchlimit")=17
	Set gtmtypes("gd_region",18,"name")="gd_region.jnl_alignsize"
	Set gtmtypes("gd_region",18,"off")=72
	Set gtmtypes("gd_region",18,"len")=4
	Set gtmtypes("gd_region",18,"type")="unsigned-int"
	Set gtmtypfldindx("gd_region","jnl_alignsize")=18
	Set gtmtypes("gd_region",19,"name")="gd_region.jnl_epoch_interval"
	Set gtmtypes("gd_region",19,"off")=76
	Set gtmtypes("gd_region",19,"len")=4
	Set gtmtypes("gd_region",19,"type")="int"
	Set gtmtypfldindx("gd_region","jnl_epoch_interval")=19
	Set gtmtypes("gd_region",20,"name")="gd_region.jnl_sync_io"
	Set gtmtypes("gd_region",20,"off")=80
	Set gtmtypes("gd_region",20,"len")=4
	Set gtmtypes("gd_region",20,"type")="int"
	Set gtmtypfldindx("gd_region","jnl_sync_io")=20
	Set gtmtypes("gd_region",21,"name")="gd_region.jnl_yield_lmt"
	Set gtmtypes("gd_region",21,"off")=84
	Set gtmtypes("gd_region",21,"len")=4
	Set gtmtypes("gd_region",21,"type")="int"
	Set gtmtypfldindx("gd_region","jnl_yield_lmt")=21
	Set gtmtypes("gd_region",22,"name")="gd_region.jnl_buffer_size"
	Set gtmtypes("gd_region",22,"off")=88
	Set gtmtypes("gd_region",22,"len")=4
	Set gtmtypes("gd_region",22,"type")="int"
	Set gtmtypfldindx("gd_region","jnl_buffer_size")=22
	Set gtmtypes("gd_region",23,"name")="gd_region.jnl_before_image"
	Set gtmtypes("gd_region",23,"off")=92
	Set gtmtypes("gd_region",23,"len")=1
	Set gtmtypes("gd_region",23,"type")="char"
	Set gtmtypfldindx("gd_region","jnl_before_image")=23
	Set gtmtypes("gd_region",24,"name")="gd_region.opening"
	Set gtmtypes("gd_region",24,"off")=93
	Set gtmtypes("gd_region",24,"len")=1
	Set gtmtypes("gd_region",24,"type")="char"
	Set gtmtypfldindx("gd_region","opening")=24
	Set gtmtypes("gd_region",25,"name")="gd_region.read_only"
	Set gtmtypes("gd_region",25,"off")=94
	Set gtmtypes("gd_region",25,"len")=1
	Set gtmtypes("gd_region",25,"type")="char"
	Set gtmtypfldindx("gd_region","read_only")=25
	Set gtmtypes("gd_region",26,"name")="gd_region.was_open"
	Set gtmtypes("gd_region",26,"off")=95
	Set gtmtypes("gd_region",26,"len")=1
	Set gtmtypes("gd_region",26,"type")="char"
	Set gtmtypfldindx("gd_region","was_open")=26
	Set gtmtypes("gd_region",27,"name")="gd_region.cmx_regnum"
	Set gtmtypes("gd_region",27,"off")=96
	Set gtmtypes("gd_region",27,"len")=1
	Set gtmtypes("gd_region",27,"type")="unsigned-char"
	Set gtmtypfldindx("gd_region","cmx_regnum")=27
	Set gtmtypes("gd_region",28,"name")="gd_region.def_coll"
	Set gtmtypes("gd_region",28,"off")=97
	Set gtmtypes("gd_region",28,"len")=1
	Set gtmtypes("gd_region",28,"type")="unsigned-char"
	Set gtmtypfldindx("gd_region","def_coll")=28
	Set gtmtypes("gd_region",29,"name")="gd_region.std_null_coll"
	Set gtmtypes("gd_region",29,"off")=98
	Set gtmtypes("gd_region",29,"len")=1
	Set gtmtypes("gd_region",29,"type")="char"
	Set gtmtypfldindx("gd_region","std_null_coll")=29
	Set gtmtypes("gd_region",30,"name")="gd_region.freeze_on_fail"
	Set gtmtypes("gd_region",30,"off")=99
	Set gtmtypes("gd_region",30,"len")=1
	Set gtmtypes("gd_region",30,"type")="char"
	Set gtmtypfldindx("gd_region","freeze_on_fail")=30
	Set gtmtypes("gd_region",31,"name")="gd_region.mumps_can_bypass"
	Set gtmtypes("gd_region",31,"off")=100
	Set gtmtypes("gd_region",31,"len")=1
	Set gtmtypes("gd_region",31,"type")="char"
	Set gtmtypfldindx("gd_region","mumps_can_bypass")=31
	Set gtmtypes("gd_region",32,"name")="gd_region.jnl_file_len"
	Set gtmtypes("gd_region",32,"off")=101
	Set gtmtypes("gd_region",32,"len")=1
	Set gtmtypes("gd_region",32,"type")="unsigned-char"
	Set gtmtypfldindx("gd_region","jnl_file_len")=32
	Set gtmtypes("gd_region",33,"name")="gd_region.jnl_file_name"
	Set gtmtypes("gd_region",33,"off")=102
	Set gtmtypes("gd_region",33,"len")=256
	Set gtmtypes("gd_region",33,"type")="unsigned-char"
	Set gtmtypfldindx("gd_region","jnl_file_name")=33
	Set gtmtypes("gd_region",34,"name")="gd_region.align_filler"
	Set gtmtypes("gd_region",34,"off")=358
	Set gtmtypes("gd_region",34,"len")=2
	Set gtmtypes("gd_region",34,"type")="char"
	Set gtmtypfldindx("gd_region","align_filler")=34
	Set gtmtypes("gd_region",35,"name")="gd_region.node"
	Set gtmtypes("gd_region",35,"off")=360
	Set gtmtypes("gd_region",35,"len")=4
	Set gtmtypes("gd_region",35,"type")="int"
	Set gtmtypfldindx("gd_region","node")=35
	Set gtmtypes("gd_region",36,"name")="gd_region.sec_size"
	Set gtmtypes("gd_region",36,"off")=364
	Set gtmtypes("gd_region",36,"len")=4
	Set gtmtypes("gd_region",36,"type")="int"
	Set gtmtypfldindx("gd_region","sec_size")=36
	Set gtmtypes("gd_region",37,"name")="gd_region.is_spanned"
	Set gtmtypes("gd_region",37,"off")=368
	Set gtmtypes("gd_region",37,"len")=4
	Set gtmtypes("gd_region",37,"type")="unsigned-int"
	Set gtmtypfldindx("gd_region","is_spanned")=37
	Set gtmtypes("gd_region",38,"name")="gd_region.statsDB_reg_index"
	Set gtmtypes("gd_region",38,"off")=372
	Set gtmtypes("gd_region",38,"len")=4
	Set gtmtypes("gd_region",38,"type")="unsigned-int"
	Set gtmtypfldindx("gd_region","statsDB_reg_index")=38
	Set gtmtypes("gd_region",39,"name")="gd_region.epoch_taper"
	Set gtmtypes("gd_region",39,"off")=376
	Set gtmtypes("gd_region",39,"len")=1
	Set gtmtypes("gd_region",39,"type")="char"
	Set gtmtypfldindx("gd_region","epoch_taper")=39
	Set gtmtypes("gd_region",40,"name")="gd_region.reservedDBFlags"
	Set gtmtypes("gd_region",40,"off")=377
	Set gtmtypes("gd_region",40,"len")=1
	Set gtmtypes("gd_region",40,"type")="char"
	Set gtmtypfldindx("gd_region","reservedDBFlags")=40
	Set gtmtypes("gd_region",41,"name")="gd_region.lock_crit_with_db"
	Set gtmtypes("gd_region",41,"off")=378
	Set gtmtypes("gd_region",41,"len")=1
	Set gtmtypes("gd_region",41,"type")="char"
	Set gtmtypfldindx("gd_region","lock_crit_with_db")=41
	Set gtmtypes("gd_region",42,"name")="gd_region.statsDB_setup_started"
	Set gtmtypes("gd_region",42,"off")=379
	Set gtmtypes("gd_region",42,"len")=1
	Set gtmtypes("gd_region",42,"type")="char"
	Set gtmtypfldindx("gd_region","statsDB_setup_started")=42
	Set gtmtypes("gd_region",43,"name")="gd_region.owning_gd"
	Set gtmtypes("gd_region",43,"off")=384
	Set gtmtypes("gd_region",43,"len")=8
	Set gtmtypes("gd_region",43,"type")="addr"
	Set gtmtypfldindx("gd_region","owning_gd")=43
	Set gtmtypes("gd_region",44,"name")="gd_region.statsDB_setup_completed"
	Set gtmtypes("gd_region",44,"off")=392
	Set gtmtypes("gd_region",44,"len")=1
	Set gtmtypes("gd_region",44,"type")="char"
	Set gtmtypfldindx("gd_region","statsDB_setup_completed")=44
	Set gtmtypes("gd_region",45,"name")="gd_region.filler"
	Set gtmtypes("gd_region",45,"off")=393
	Set gtmtypes("gd_region",45,"len")=39
	Set gtmtypes("gd_region",45,"type")="char"
	Set gtmtypfldindx("gd_region","filler")=45
	;
	Set gtmtypes("gd_seg_addr")="union"
	Set gtmtypes("gd_seg_addr",0)=2
	Set gtmtypes("gd_seg_addr","len")=8
	Set gtmtypes("gd_seg_addr",1,"name")="gd_seg_addr.offset"
	Set gtmtypes("gd_seg_addr",1,"off")=0
	Set gtmtypes("gd_seg_addr",1,"len")=4
	Set gtmtypes("gd_seg_addr",1,"type")="int"
	Set gtmtypfldindx("gd_seg_addr","offset")=1
	Set gtmtypes("gd_seg_addr",2,"name")="gd_seg_addr.addr"
	Set gtmtypes("gd_seg_addr",2,"off")=0
	Set gtmtypes("gd_seg_addr",2,"len")=8
	Set gtmtypes("gd_seg_addr",2,"type")="addr"
	Set gtmtypfldindx("gd_seg_addr","addr")=2
	;
	Set gtmtypes("gd_segment")="struct"
	Set gtmtypes("gd_segment",0)=26
	Set gtmtypes("gd_segment","len")=384
	Set gtmtypes("gd_segment",1,"name")="gd_segment.sname_len"
	Set gtmtypes("gd_segment",1,"off")=0
	Set gtmtypes("gd_segment",1,"len")=2
	Set gtmtypes("gd_segment",1,"type")="unsigned-short"
	Set gtmtypfldindx("gd_segment","sname_len")=1
	Set gtmtypes("gd_segment",2,"name")="gd_segment.sname"
	Set gtmtypes("gd_segment",2,"off")=2
	Set gtmtypes("gd_segment",2,"len")=32
	Set gtmtypes("gd_segment",2,"type")="unsigned-char"
	Set gtmtypfldindx("gd_segment","sname")=2
	Set gtmtypes("gd_segment",3,"name")="gd_segment.fname_len"
	Set gtmtypes("gd_segment",3,"off")=34
	Set gtmtypes("gd_segment",3,"len")=2
	Set gtmtypes("gd_segment",3,"type")="unsigned-short"
	Set gtmtypfldindx("gd_segment","fname_len")=3
	Set gtmtypes("gd_segment",4,"name")="gd_segment.fname"
	Set gtmtypes("gd_segment",4,"off")=36
	Set gtmtypes("gd_segment",4,"len")=256
	Set gtmtypes("gd_segment",4,"type")="unsigned-char"
	Set gtmtypfldindx("gd_segment","fname")=4
	Set gtmtypes("gd_segment",5,"name")="gd_segment.blk_size"
	Set gtmtypes("gd_segment",5,"off")=292
	Set gtmtypes("gd_segment",5,"len")=2
	Set gtmtypes("gd_segment",5,"type")="unsigned-short"
	Set gtmtypfldindx("gd_segment","blk_size")=5
	Set gtmtypes("gd_segment",6,"name")="gd_segment.full_blkwrt"
	Set gtmtypes("gd_segment",6,"off")=294
	Set gtmtypes("gd_segment",6,"len")=2
	Set gtmtypes("gd_segment",6,"type")="unsigned-short"
	Set gtmtypfldindx("gd_segment","full_blkwrt")=6
	Set gtmtypes("gd_segment",7,"name")="gd_segment.ext_blk_count"
	Set gtmtypes("gd_segment",7,"off")=296
	Set gtmtypes("gd_segment",7,"len")=4
	Set gtmtypes("gd_segment",7,"type")="unsigned-int"
	Set gtmtypfldindx("gd_segment","ext_blk_count")=7
	Set gtmtypes("gd_segment",8,"name")="gd_segment.allocation"
	Set gtmtypes("gd_segment",8,"off")=300
	Set gtmtypes("gd_segment",8,"len")=4
	Set gtmtypes("gd_segment",8,"type")="unsigned-int"
	Set gtmtypfldindx("gd_segment","allocation")=8
	Set gtmtypes("gd_segment",9,"name")="gd_segment.cm_blk"
	Set gtmtypes("gd_segment",9,"off")=304
	Set gtmtypes("gd_segment",9,"len")=8
	Set gtmtypes("gd_segment",9,"type")="addr"
	Set gtmtypfldindx("gd_segment","cm_blk")=9
	Set gtmtypes("gd_segment",10,"name")="gd_segment.defext"
	Set gtmtypes("gd_segment",10,"off")=312
	Set gtmtypes("gd_segment",10,"len")=4
	Set gtmtypes("gd_segment",10,"type")="unsigned-char"
	Set gtmtypfldindx("gd_segment","defext")=10
	Set gtmtypes("gd_segment",11,"name")="gd_segment.defer_time"
	Set gtmtypes("gd_segment",11,"off")=316
	Set gtmtypes("gd_segment",11,"len")=1
	Set gtmtypes("gd_segment",11,"type")="char"
	Set gtmtypfldindx("gd_segment","defer_time")=11
	Set gtmtypes("gd_segment",12,"name")="gd_segment.file_type"
	Set gtmtypes("gd_segment",12,"off")=317
	Set gtmtypes("gd_segment",12,"len")=1
	Set gtmtypes("gd_segment",12,"type")="unsigned-char"
	Set gtmtypfldindx("gd_segment","file_type")=12
	Set gtmtypes("gd_segment",13,"name")="gd_segment.buckets"
	Set gtmtypes("gd_segment",13,"off")=318
	Set gtmtypes("gd_segment",13,"len")=1
	Set gtmtypes("gd_segment",13,"type")="unsigned-char"
	Set gtmtypfldindx("gd_segment","buckets")=13
	Set gtmtypes("gd_segment",14,"name")="gd_segment.windows"
	Set gtmtypes("gd_segment",14,"off")=319
	Set gtmtypes("gd_segment",14,"len")=1
	Set gtmtypes("gd_segment",14,"type")="unsigned-char"
	Set gtmtypfldindx("gd_segment","windows")=14
	Set gtmtypes("gd_segment",15,"name")="gd_segment.lock_space"
	Set gtmtypes("gd_segment",15,"off")=320
	Set gtmtypes("gd_segment",15,"len")=4
	Set gtmtypes("gd_segment",15,"type")="unsigned-int"
	Set gtmtypfldindx("gd_segment","lock_space")=15
	Set gtmtypes("gd_segment",16,"name")="gd_segment.global_buffers"
	Set gtmtypes("gd_segment",16,"off")=324
	Set gtmtypes("gd_segment",16,"len")=4
	Set gtmtypes("gd_segment",16,"type")="unsigned-int"
	Set gtmtypfldindx("gd_segment","global_buffers")=16
	Set gtmtypes("gd_segment",17,"name")="gd_segment.reserved_bytes"
	Set gtmtypes("gd_segment",17,"off")=328
	Set gtmtypes("gd_segment",17,"len")=4
	Set gtmtypes("gd_segment",17,"type")="unsigned-int"
	Set gtmtypfldindx("gd_segment","reserved_bytes")=17
	Set gtmtypes("gd_segment",18,"name")="gd_segment.mutex_slots"
	Set gtmtypes("gd_segment",18,"off")=332
	Set gtmtypes("gd_segment",18,"len")=4
	Set gtmtypes("gd_segment",18,"type")="unsigned-int"
	Set gtmtypfldindx("gd_segment","mutex_slots")=18
	Set gtmtypes("gd_segment",19,"name")="gd_segment.defer_allocate"
	Set gtmtypes("gd_segment",19,"off")=336
	Set gtmtypes("gd_segment",19,"len")=4
	Set gtmtypes("gd_segment",19,"type")="boolean_t"
	Set gtmtypfldindx("gd_segment","defer_allocate")=19
	Set gtmtypes("gd_segment",20,"name")="gd_segment.acc_meth"
	Set gtmtypes("gd_segment",20,"off")=340
	Set gtmtypes("gd_segment",20,"len")=4
	Set gtmtypes("gd_segment",20,"type")="int"
	Set gtmtypfldindx("gd_segment","acc_meth")=20
	Set gtmtypes("gd_segment",21,"name")="gd_segment.file_cntl"
	Set gtmtypes("gd_segment",21,"off")=344
	Set gtmtypes("gd_segment",21,"len")=8
	Set gtmtypes("gd_segment",21,"type")="addr"
	Set gtmtypfldindx("gd_segment","file_cntl")=21
	Set gtmtypes("gd_segment",22,"name")="gd_segment.repl_list"
	Set gtmtypes("gd_segment",22,"off")=352
	Set gtmtypes("gd_segment",22,"len")=8
	Set gtmtypes("gd_segment",22,"type")="addr"
	Set gtmtypfldindx("gd_segment","repl_list")=22
	Set gtmtypes("gd_segment",23,"name")="gd_segment.is_encrypted"
	Set gtmtypes("gd_segment",23,"off")=360
	Set gtmtypes("gd_segment",23,"len")=4
	Set gtmtypes("gd_segment",23,"type")="unsigned-int"
	Set gtmtypfldindx("gd_segment","is_encrypted")=23
	Set gtmtypes("gd_segment",24,"name")="gd_segment.asyncio"
	Set gtmtypes("gd_segment",24,"off")=364
	Set gtmtypes("gd_segment",24,"len")=4
	Set gtmtypes("gd_segment",24,"type")="boolean_t"
	Set gtmtypfldindx("gd_segment","asyncio")=24
	Set gtmtypes("gd_segment",25,"name")="gd_segment.read_only"
	Set gtmtypes("gd_segment",25,"off")=368
	Set gtmtypes("gd_segment",25,"len")=4
	Set gtmtypes("gd_segment",25,"type")="boolean_t"
	Set gtmtypfldindx("gd_segment","read_only")=25
	Set gtmtypes("gd_segment",26,"name")="gd_segment.filler"
	Set gtmtypes("gd_segment",26,"off")=372
	Set gtmtypes("gd_segment",26,"len")=12
	Set gtmtypes("gd_segment",26,"type")="char"
	Set gtmtypfldindx("gd_segment","filler")=26
	;
	Set gtmtypes("gdr_name")="struct"
	Set gtmtypes("gdr_name",0)=10
	Set gtmtypes("gdr_name","len")=48
	Set gtmtypes("gdr_name",1,"name")="gdr_name.name"
	Set gtmtypes("gdr_name",1,"off")=0
	Set gtmtypes("gdr_name",1,"len")=16
	Set gtmtypes("gdr_name",1,"type")="mstr"
	Set gtmtypfldindx("gdr_name","name")=1
	Set gtmtypes("gdr_name",2,"name")="gdr_name.name.char_len"
	Set gtmtypes("gdr_name",2,"off")=0
	Set gtmtypes("gdr_name",2,"len")=4
	Set gtmtypes("gdr_name",2,"type")="unsigned-int"
	Set gtmtypfldindx("gdr_name","name.char_len")=2
	Set gtmtypes("gdr_name",3,"name")="gdr_name.name.len"
	Set gtmtypes("gdr_name",3,"off")=4
	Set gtmtypes("gdr_name",3,"len")=4
	Set gtmtypes("gdr_name",3,"type")="int"
	Set gtmtypfldindx("gdr_name","name.len")=3
	Set gtmtypes("gdr_name",4,"name")="gdr_name.name.addr"
	Set gtmtypes("gdr_name",4,"off")=8
	Set gtmtypes("gdr_name",4,"len")=8
	Set gtmtypes("gdr_name",4,"type")="addr"
	Set gtmtypfldindx("gdr_name","name.addr")=4
	Set gtmtypes("gdr_name",5,"name")="gdr_name.exp_name"
	Set gtmtypes("gdr_name",5,"off")=16
	Set gtmtypes("gdr_name",5,"len")=16
	Set gtmtypes("gdr_name",5,"type")="mstr"
	Set gtmtypfldindx("gdr_name","exp_name")=5
	Set gtmtypes("gdr_name",6,"name")="gdr_name.exp_name.char_len"
	Set gtmtypes("gdr_name",6,"off")=16
	Set gtmtypes("gdr_name",6,"len")=4
	Set gtmtypes("gdr_name",6,"type")="unsigned-int"
	Set gtmtypfldindx("gdr_name","exp_name.char_len")=6
	Set gtmtypes("gdr_name",7,"name")="gdr_name.exp_name.len"
	Set gtmtypes("gdr_name",7,"off")=20
	Set gtmtypes("gdr_name",7,"len")=4
	Set gtmtypes("gdr_name",7,"type")="int"
	Set gtmtypfldindx("gdr_name","exp_name.len")=7
	Set gtmtypes("gdr_name",8,"name")="gdr_name.exp_name.addr"
	Set gtmtypes("gdr_name",8,"off")=24
	Set gtmtypes("gdr_name",8,"len")=8
	Set gtmtypes("gdr_name",8,"type")="addr"
	Set gtmtypfldindx("gdr_name","exp_name.addr")=8
	Set gtmtypes("gdr_name",9,"name")="gdr_name.link"
	Set gtmtypes("gdr_name",9,"off")=32
	Set gtmtypes("gdr_name",9,"len")=8
	Set gtmtypes("gdr_name",9,"type")="addr"
	Set gtmtypfldindx("gdr_name","link")=9
	Set gtmtypes("gdr_name",10,"name")="gdr_name.gd_ptr"
	Set gtmtypes("gdr_name",10,"off")=40
	Set gtmtypes("gdr_name",10,"len")=8
	Set gtmtypes("gdr_name",10,"type")="addr"
	Set gtmtypfldindx("gdr_name","gd_ptr")=10
	;
	Set gtmtypes("gds_file_id")="struct"
	Set gtmtypes("gds_file_id",0)=3
	Set gtmtypes("gds_file_id","len")=28
	Set gtmtypes("gds_file_id",1,"name")="gds_file_id.dvi"
	Set gtmtypes("gds_file_id",1,"off")=0
	Set gtmtypes("gds_file_id",1,"len")=16
	Set gtmtypes("gds_file_id",1,"type")="char"
	Set gtmtypfldindx("gds_file_id","dvi")=1
	Set gtmtypes("gds_file_id",2,"name")="gds_file_id.did"
	Set gtmtypes("gds_file_id",2,"off")=16
	Set gtmtypes("gds_file_id",2,"len")=6
	Set gtmtypes("gds_file_id",2,"type")="unsigned-short"
	Set gtmtypfldindx("gds_file_id","did")=2
	Set gtmtypes("gds_file_id",2,"dim")=3
	Set gtmtypes("gds_file_id",3,"name")="gds_file_id.fid"
	Set gtmtypes("gds_file_id",3,"off")=22
	Set gtmtypes("gds_file_id",3,"len")=6
	Set gtmtypes("gds_file_id",3,"type")="unsigned-short"
	Set gtmtypfldindx("gds_file_id","fid")=3
	Set gtmtypes("gds_file_id",3,"dim")=3
	;
	Set gtmtypes("gld_dbname_list")="struct"
	Set gtmtypes("gld_dbname_list",0)=3
	Set gtmtypes("gld_dbname_list","len")=24
	Set gtmtypes("gld_dbname_list",1,"name")="gld_dbname_list.next"
	Set gtmtypes("gld_dbname_list",1,"off")=0
	Set gtmtypes("gld_dbname_list",1,"len")=8
	Set gtmtypes("gld_dbname_list",1,"type")="addr"
	Set gtmtypfldindx("gld_dbname_list","next")=1
	Set gtmtypes("gld_dbname_list",2,"name")="gld_dbname_list.db_ctl"
	Set gtmtypes("gld_dbname_list",2,"off")=8
	Set gtmtypes("gld_dbname_list",2,"len")=8
	Set gtmtypes("gld_dbname_list",2,"type")="addr"
	Set gtmtypfldindx("gld_dbname_list","db_ctl")=2
	Set gtmtypes("gld_dbname_list",3,"name")="gld_dbname_list.gd"
	Set gtmtypes("gld_dbname_list",3,"off")=16
	Set gtmtypes("gld_dbname_list",3,"len")=8
	Set gtmtypes("gld_dbname_list",3,"type")="addr"
	Set gtmtypfldindx("gld_dbname_list","gd")=3
	;
	Set gtmtypes("glist")="struct"
	Set gtmtypes("glist",0)=4
	Set gtmtypes("glist","len")=32
	Set gtmtypes("glist",1,"name")="glist.next"
	Set gtmtypes("glist",1,"off")=0
	Set gtmtypes("glist",1,"len")=8
	Set gtmtypes("glist",1,"type")="addr"
	Set gtmtypfldindx("glist","next")=1
	Set gtmtypes("glist",2,"name")="glist.reg"
	Set gtmtypes("glist",2,"off")=8
	Set gtmtypes("glist",2,"len")=8
	Set gtmtypes("glist",2,"type")="addr"
	Set gtmtypfldindx("glist","reg")=2
	Set gtmtypes("glist",3,"name")="glist.gvt"
	Set gtmtypes("glist",3,"off")=16
	Set gtmtypes("glist",3,"len")=8
	Set gtmtypes("glist",3,"type")="addr"
	Set gtmtypfldindx("glist","gvt")=3
	Set gtmtypes("glist",4,"name")="glist.gvnh_reg"
	Set gtmtypes("glist",4,"off")=24
	Set gtmtypes("glist",4,"len")=8
	Set gtmtypes("glist",4,"type")="addr"
	Set gtmtypfldindx("glist","gvnh_reg")=4
	;
	Set gtmtypes("global_dir_path")="struct"
	Set gtmtypes("global_dir_path",0)=3
	Set gtmtypes("global_dir_path","len")=16
	Set gtmtypes("global_dir_path",1,"name")="global_dir_path.block"
	Set gtmtypes("global_dir_path",1,"off")=0
	Set gtmtypes("global_dir_path",1,"len")=4
	Set gtmtypes("global_dir_path",1,"type")="int"
	Set gtmtypfldindx("global_dir_path","block")=1
	Set gtmtypes("global_dir_path",2,"name")="global_dir_path.offset"
	Set gtmtypes("global_dir_path",2,"off")=4
	Set gtmtypes("global_dir_path",2,"len")=4
	Set gtmtypes("global_dir_path",2,"type")="int"
	Set gtmtypfldindx("global_dir_path","offset")=2
	Set gtmtypes("global_dir_path",3,"name")="global_dir_path.next"
	Set gtmtypes("global_dir_path",3,"off")=8
	Set gtmtypes("global_dir_path",3,"len")=8
	Set gtmtypes("global_dir_path",3,"type")="addr"
	Set gtmtypfldindx("global_dir_path","next")=3
	;
	Set gtmtypes("global_latch_t")="struct"
	Set gtmtypes("global_latch_t",0)=5
	Set gtmtypes("global_latch_t","len")=8
	Set gtmtypes("global_latch_t",1,"name")="global_latch_t.u"
	Set gtmtypes("global_latch_t",1,"off")=0
	Set gtmtypes("global_latch_t",1,"len")=8
	Set gtmtypes("global_latch_t",1,"type")="union"
	Set gtmtypfldindx("global_latch_t","u")=1
	Set gtmtypes("global_latch_t",2,"name")="global_latch_t.u.pid_imgcnt"
	Set gtmtypes("global_latch_t",2,"off")=0
	Set gtmtypes("global_latch_t",2,"len")=8
	Set gtmtypes("global_latch_t",2,"type")="uint64_t"
	Set gtmtypfldindx("global_latch_t","u.pid_imgcnt")=2
	Set gtmtypes("global_latch_t",3,"name")="global_latch_t.u.parts"
	Set gtmtypes("global_latch_t",3,"off")=0
	Set gtmtypes("global_latch_t",3,"len")=8
	Set gtmtypes("global_latch_t",3,"type")="struct"
	Set gtmtypfldindx("global_latch_t","u.parts")=3
	Set gtmtypes("global_latch_t",4,"name")="global_latch_t.u.parts.latch_pid"
	Set gtmtypes("global_latch_t",4,"off")=0
	Set gtmtypes("global_latch_t",4,"len")=4
	Set gtmtypes("global_latch_t",4,"type")="int"
	Set gtmtypfldindx("global_latch_t","u.parts.latch_pid")=4
	Set gtmtypes("global_latch_t",5,"name")="global_latch_t.u.parts.latch_word"
	Set gtmtypes("global_latch_t",5,"off")=4
	Set gtmtypes("global_latch_t",5,"len")=4
	Set gtmtypes("global_latch_t",5,"type")="int"
	Set gtmtypfldindx("global_latch_t","u.parts.latch_word")=5
	;
	Set gtmtypes("global_list")="struct"
	Set gtmtypes("global_list",0)=9
	Set gtmtypes("global_list","len")=120
	Set gtmtypes("global_list",1,"name")="global_list.root"
	Set gtmtypes("global_list",1,"off")=0
	Set gtmtypes("global_list",1,"len")=4
	Set gtmtypes("global_list",1,"type")="int"
	Set gtmtypfldindx("global_list","root")=1
	Set gtmtypes("global_list",2,"name")="global_list.link"
	Set gtmtypes("global_list",2,"off")=8
	Set gtmtypes("global_list",2,"len")=8
	Set gtmtypes("global_list",2,"type")="addr"
	Set gtmtypfldindx("global_list","link")=2
	Set gtmtypes("global_list",3,"name")="global_list.path"
	Set gtmtypes("global_list",3,"off")=16
	Set gtmtypes("global_list",3,"len")=32
	Set gtmtypes("global_list",3,"type")="int"
	Set gtmtypfldindx("global_list","path")=3
	Set gtmtypes("global_list",3,"dim")=8
	Set gtmtypes("global_list",4,"name")="global_list.offset"
	Set gtmtypes("global_list",4,"off")=48
	Set gtmtypes("global_list",4,"len")=32
	Set gtmtypes("global_list",4,"type")="unsigned-int"
	Set gtmtypfldindx("global_list","offset")=4
	Set gtmtypes("global_list",4,"dim")=8
	Set gtmtypes("global_list",5,"name")="global_list.nct"
	Set gtmtypes("global_list",5,"off")=80
	Set gtmtypes("global_list",5,"len")=1
	Set gtmtypes("global_list",5,"type")="unsigned-char"
	Set gtmtypfldindx("global_list","nct")=5
	Set gtmtypes("global_list",6,"name")="global_list.act"
	Set gtmtypes("global_list",6,"off")=81
	Set gtmtypes("global_list",6,"len")=1
	Set gtmtypes("global_list",6,"type")="unsigned-char"
	Set gtmtypfldindx("global_list","act")=6
	Set gtmtypes("global_list",7,"name")="global_list.ver"
	Set gtmtypes("global_list",7,"off")=82
	Set gtmtypes("global_list",7,"len")=1
	Set gtmtypes("global_list",7,"type")="unsigned-char"
	Set gtmtypfldindx("global_list","ver")=7
	Set gtmtypes("global_list",8,"name")="global_list.key"
	Set gtmtypes("global_list",8,"off")=83
	Set gtmtypes("global_list",8,"len")=32
	Set gtmtypes("global_list",8,"type")="char"
	Set gtmtypfldindx("global_list","key")=8
	Set gtmtypes("global_list",9,"name")="global_list.keysize"
	Set gtmtypes("global_list",9,"off")=116
	Set gtmtypes("global_list",9,"len")=4
	Set gtmtypes("global_list",9,"type")="int"
	Set gtmtypfldindx("global_list","keysize")=9
	;
	Set gtmtypes("global_root_list")="struct"
	Set gtmtypes("global_root_list",0)=3
	Set gtmtypes("global_root_list","len")=24
	Set gtmtypes("global_root_list",1,"name")="global_root_list.root"
	Set gtmtypes("global_root_list",1,"off")=0
	Set gtmtypes("global_root_list",1,"len")=4
	Set gtmtypes("global_root_list",1,"type")="int"
	Set gtmtypfldindx("global_root_list","root")=1
	Set gtmtypes("global_root_list",2,"name")="global_root_list.dir_path"
	Set gtmtypes("global_root_list",2,"off")=8
	Set gtmtypes("global_root_list",2,"len")=8
	Set gtmtypes("global_root_list",2,"type")="addr"
	Set gtmtypfldindx("global_root_list","dir_path")=2
	Set gtmtypes("global_root_list",3,"name")="global_root_list.link"
	Set gtmtypes("global_root_list",3,"off")=16
	Set gtmtypes("global_root_list",3,"len")=8
	Set gtmtypes("global_root_list",3,"type")="addr"
	Set gtmtypfldindx("global_root_list","link")=3
	;
	Set gtmtypes("global_tlvl_info")="struct"
	Set gtmtypes("global_tlvl_info",0)=9
	Set gtmtypes("global_tlvl_info","len")=64
	Set gtmtypes("global_tlvl_info",1,"name")="global_tlvl_info.next_global_tlvl_info"
	Set gtmtypes("global_tlvl_info",1,"off")=0
	Set gtmtypes("global_tlvl_info",1,"len")=8
	Set gtmtypes("global_tlvl_info",1,"type")="addr"
	Set gtmtypfldindx("global_tlvl_info","next_global_tlvl_info")=1
	Set gtmtypes("global_tlvl_info",2,"name")="global_tlvl_info.global_tlvl_fence_info"
	Set gtmtypes("global_tlvl_info",2,"off")=8
	Set gtmtypes("global_tlvl_info",2,"len")=8
	Set gtmtypes("global_tlvl_info",2,"type")="addr"
	Set gtmtypfldindx("global_tlvl_info","global_tlvl_fence_info")=2
	Set gtmtypes("global_tlvl_info",3,"name")="global_tlvl_info.t_level"
	Set gtmtypes("global_tlvl_info",3,"off")=16
	Set gtmtypes("global_tlvl_info",3,"len")=4
	Set gtmtypes("global_tlvl_info",3,"type")="unsigned-int"
	Set gtmtypfldindx("global_tlvl_info","t_level")=3
	Set gtmtypes("global_tlvl_info",4,"name")="global_tlvl_info.tlvl_cumul_jrec_len"
	Set gtmtypes("global_tlvl_info",4,"off")=20
	Set gtmtypes("global_tlvl_info",4,"len")=4
	Set gtmtypes("global_tlvl_info",4,"type")="unsigned-int"
	Set gtmtypfldindx("global_tlvl_info","tlvl_cumul_jrec_len")=4
	Set gtmtypes("global_tlvl_info",5,"name")="global_tlvl_info.tlvl_tp_ztp_jnl_upd_num"
	Set gtmtypes("global_tlvl_info",5,"off")=24
	Set gtmtypes("global_tlvl_info",5,"len")=4
	Set gtmtypes("global_tlvl_info",5,"type")="unsigned-int"
	Set gtmtypfldindx("global_tlvl_info","tlvl_tp_ztp_jnl_upd_num")=5
	Set gtmtypes("global_tlvl_info",6,"name")="global_tlvl_info.tlvl_prev_ztworm_ptr"
	Set gtmtypes("global_tlvl_info",6,"off")=32
	Set gtmtypes("global_tlvl_info",6,"len")=8
	Set gtmtypes("global_tlvl_info",6,"type")="addr"
	Set gtmtypfldindx("global_tlvl_info","tlvl_prev_ztworm_ptr")=6
	Set gtmtypes("global_tlvl_info",7,"name")="global_tlvl_info.curr_ua"
	Set gtmtypes("global_tlvl_info",7,"off")=40
	Set gtmtypes("global_tlvl_info",7,"len")=8
	Set gtmtypes("global_tlvl_info",7,"type")="addr"
	Set gtmtypfldindx("global_tlvl_info","curr_ua")=7
	Set gtmtypes("global_tlvl_info",8,"name")="global_tlvl_info.upd_array_ptr"
	Set gtmtypes("global_tlvl_info",8,"off")=48
	Set gtmtypes("global_tlvl_info",8,"len")=8
	Set gtmtypes("global_tlvl_info",8,"type")="addr"
	Set gtmtypfldindx("global_tlvl_info","upd_array_ptr")=8
	Set gtmtypes("global_tlvl_info",9,"name")="global_tlvl_info.ztrigbuffLen"
	Set gtmtypes("global_tlvl_info",9,"off")=56
	Set gtmtypes("global_tlvl_info",9,"len")=4
	Set gtmtypes("global_tlvl_info",9,"type")="int"
	Set gtmtypfldindx("global_tlvl_info","ztrigbuffLen")=9
	;
	Set gtmtypes("glvn_pool")="struct"
	Set gtmtypes("glvn_pool",0)=14
	Set gtmtypes("glvn_pool","len")=488
	Set gtmtypes("glvn_pool",1,"name")="glvn_pool.capacity"
	Set gtmtypes("glvn_pool",1,"off")=0
	Set gtmtypes("glvn_pool",1,"len")=4
	Set gtmtypes("glvn_pool",1,"type")="unsigned-int"
	Set gtmtypfldindx("glvn_pool","capacity")=1
	Set gtmtypes("glvn_pool",2,"name")="glvn_pool.top"
	Set gtmtypes("glvn_pool",2,"off")=4
	Set gtmtypes("glvn_pool",2,"len")=4
	Set gtmtypes("glvn_pool",2,"type")="unsigned-int"
	Set gtmtypfldindx("glvn_pool","top")=2
	Set gtmtypes("glvn_pool",3,"name")="glvn_pool.for_slot"
	Set gtmtypes("glvn_pool",3,"off")=8
	Set gtmtypes("glvn_pool",3,"len")=132
	Set gtmtypes("glvn_pool",3,"type")="unsigned-int"
	Set gtmtypfldindx("glvn_pool","for_slot")=3
	Set gtmtypes("glvn_pool",3,"dim")=33
	Set gtmtypes("glvn_pool",4,"name")="glvn_pool.share_slot"
	Set gtmtypes("glvn_pool",4,"off")=140
	Set gtmtypes("glvn_pool",4,"len")=4
	Set gtmtypes("glvn_pool",4,"type")="unsigned-int"
	Set gtmtypfldindx("glvn_pool","share_slot")=4
	Set gtmtypes("glvn_pool",5,"name")="glvn_pool.share_opcode"
	Set gtmtypes("glvn_pool",5,"off")=144
	Set gtmtypes("glvn_pool",5,"len")=4
	Set gtmtypes("glvn_pool",5,"type")="unsigned-int"
	Set gtmtypfldindx("glvn_pool","share_opcode")=5
	Set gtmtypes("glvn_pool",6,"name")="glvn_pool.mval_capacity"
	Set gtmtypes("glvn_pool",6,"off")=148
	Set gtmtypes("glvn_pool",6,"len")=4
	Set gtmtypes("glvn_pool",6,"type")="unsigned-int"
	Set gtmtypfldindx("glvn_pool","mval_capacity")=6
	Set gtmtypes("glvn_pool",7,"name")="glvn_pool.mval_top"
	Set gtmtypes("glvn_pool",7,"off")=152
	Set gtmtypes("glvn_pool",7,"len")=4
	Set gtmtypes("glvn_pool",7,"type")="unsigned-int"
	Set gtmtypfldindx("glvn_pool","mval_top")=7
	Set gtmtypes("glvn_pool",8,"name")="glvn_pool.mval_stack"
	Set gtmtypes("glvn_pool",8,"off")=160
	Set gtmtypes("glvn_pool",8,"len")=8
	Set gtmtypes("glvn_pool",8,"type")="addr"
	Set gtmtypfldindx("glvn_pool","mval_stack")=8
	Set gtmtypes("glvn_pool",9,"name")="glvn_pool.slot"
	Set gtmtypes("glvn_pool",9,"off")=168
	Set gtmtypes("glvn_pool",9,"len")=320
	Set gtmtypes("glvn_pool",9,"type")="glvn_pool_entry"
	Set gtmtypfldindx("glvn_pool","slot")=9
	Set gtmtypes("glvn_pool",10,"name")="glvn_pool.slot[0].sav_opcode"
	Set gtmtypes("glvn_pool",10,"off")=168
	Set gtmtypes("glvn_pool",10,"len")=4
	Set gtmtypes("glvn_pool",10,"type")="unsigned-int"
	Set gtmtypfldindx("glvn_pool","slot[0].sav_opcode")=10
	Set gtmtypes("glvn_pool",11,"name")="glvn_pool.slot[0].mval_top"
	Set gtmtypes("glvn_pool",11,"off")=172
	Set gtmtypes("glvn_pool",11,"len")=4
	Set gtmtypes("glvn_pool",11,"type")="unsigned-int"
	Set gtmtypfldindx("glvn_pool","slot[0].mval_top")=11
	Set gtmtypes("glvn_pool",12,"name")="glvn_pool.slot[0].precursor"
	Set gtmtypes("glvn_pool",12,"off")=176
	Set gtmtypes("glvn_pool",12,"len")=4
	Set gtmtypes("glvn_pool",12,"type")="unsigned-int"
	Set gtmtypfldindx("glvn_pool","slot[0].precursor")=12
	Set gtmtypes("glvn_pool",13,"name")="glvn_pool.slot[0].lvname"
	Set gtmtypes("glvn_pool",13,"off")=184
	Set gtmtypes("glvn_pool",13,"len")=8
	Set gtmtypes("glvn_pool",13,"type")="addr"
	Set gtmtypfldindx("glvn_pool","slot[0].lvname")=13
	Set gtmtypes("glvn_pool",14,"name")="glvn_pool.slot[0].glvn_info"
	Set gtmtypes("glvn_pool",14,"off")=192
	Set gtmtypes("glvn_pool",14,"len")=296
	Set gtmtypes("glvn_pool",14,"type")="gparam_list"
	Set gtmtypfldindx("glvn_pool","slot[0].glvn_info")=14
	;
	Set gtmtypes("glvn_pool_entry")="struct"
	Set gtmtypes("glvn_pool_entry",0)=7
	Set gtmtypes("glvn_pool_entry","len")=320
	Set gtmtypes("glvn_pool_entry",1,"name")="glvn_pool_entry.sav_opcode"
	Set gtmtypes("glvn_pool_entry",1,"off")=0
	Set gtmtypes("glvn_pool_entry",1,"len")=4
	Set gtmtypes("glvn_pool_entry",1,"type")="unsigned-int"
	Set gtmtypfldindx("glvn_pool_entry","sav_opcode")=1
	Set gtmtypes("glvn_pool_entry",2,"name")="glvn_pool_entry.mval_top"
	Set gtmtypes("glvn_pool_entry",2,"off")=4
	Set gtmtypes("glvn_pool_entry",2,"len")=4
	Set gtmtypes("glvn_pool_entry",2,"type")="unsigned-int"
	Set gtmtypfldindx("glvn_pool_entry","mval_top")=2
	Set gtmtypes("glvn_pool_entry",3,"name")="glvn_pool_entry.precursor"
	Set gtmtypes("glvn_pool_entry",3,"off")=8
	Set gtmtypes("glvn_pool_entry",3,"len")=4
	Set gtmtypes("glvn_pool_entry",3,"type")="unsigned-int"
	Set gtmtypfldindx("glvn_pool_entry","precursor")=3
	Set gtmtypes("glvn_pool_entry",4,"name")="glvn_pool_entry.lvname"
	Set gtmtypes("glvn_pool_entry",4,"off")=16
	Set gtmtypes("glvn_pool_entry",4,"len")=8
	Set gtmtypes("glvn_pool_entry",4,"type")="addr"
	Set gtmtypfldindx("glvn_pool_entry","lvname")=4
	Set gtmtypes("glvn_pool_entry",5,"name")="glvn_pool_entry.glvn_info"
	Set gtmtypes("glvn_pool_entry",5,"off")=24
	Set gtmtypes("glvn_pool_entry",5,"len")=296
	Set gtmtypes("glvn_pool_entry",5,"type")="gparam_list"
	Set gtmtypfldindx("glvn_pool_entry","glvn_info")=5
	Set gtmtypes("glvn_pool_entry",6,"name")="glvn_pool_entry.glvn_info.n"
	Set gtmtypes("glvn_pool_entry",6,"off")=24
	Set gtmtypes("glvn_pool_entry",6,"len")=8
	Set gtmtypes("glvn_pool_entry",6,"type")="intptr_t"
	Set gtmtypfldindx("glvn_pool_entry","glvn_info.n")=6
	Set gtmtypes("glvn_pool_entry",7,"name")="glvn_pool_entry.glvn_info.arg"
	Set gtmtypes("glvn_pool_entry",7,"off")=32
	Set gtmtypes("glvn_pool_entry",7,"len")=288
	Set gtmtypes("glvn_pool_entry",7,"type")="addr"
	Set gtmtypfldindx("glvn_pool_entry","glvn_info.arg")=7
	Set gtmtypes("glvn_pool_entry",7,"dim")=36
	;
	Set gtmtypes("gparam_list")="struct"
	Set gtmtypes("gparam_list",0)=2
	Set gtmtypes("gparam_list","len")=296
	Set gtmtypes("gparam_list",1,"name")="gparam_list.n"
	Set gtmtypes("gparam_list",1,"off")=0
	Set gtmtypes("gparam_list",1,"len")=8
	Set gtmtypes("gparam_list",1,"type")="intptr_t"
	Set gtmtypfldindx("gparam_list","n")=1
	Set gtmtypes("gparam_list",2,"name")="gparam_list.arg"
	Set gtmtypes("gparam_list",2,"off")=8
	Set gtmtypes("gparam_list",2,"len")=288
	Set gtmtypes("gparam_list",2,"type")="addr"
	Set gtmtypfldindx("gparam_list","arg")=2
	Set gtmtypes("gparam_list",2,"dim")=36
	;
	Set gtmtypes("gtcm_proto_cpu_info_t")="struct"
	Set gtmtypes("gtcm_proto_cpu_info_t",0)=3
	Set gtmtypes("gtcm_proto_cpu_info_t","len")=24
	Set gtmtypes("gtcm_proto_cpu_info_t",1,"name")="gtcm_proto_cpu_info_t.cpu_in_rel_str"
	Set gtmtypes("gtcm_proto_cpu_info_t",1,"off")=0
	Set gtmtypes("gtcm_proto_cpu_info_t",1,"len")=8
	Set gtmtypes("gtcm_proto_cpu_info_t",1,"type")="addr"
	Set gtmtypfldindx("gtcm_proto_cpu_info_t","cpu_in_rel_str")=1
	Set gtmtypes("gtcm_proto_cpu_info_t",2,"name")="gtcm_proto_cpu_info_t.size_of_cpu_in_rel_str"
	Set gtmtypes("gtcm_proto_cpu_info_t",2,"off")=8
	Set gtmtypes("gtcm_proto_cpu_info_t",2,"len")=4
	Set gtmtypes("gtcm_proto_cpu_info_t",2,"type")="int"
	Set gtmtypfldindx("gtcm_proto_cpu_info_t","size_of_cpu_in_rel_str")=2
	Set gtmtypes("gtcm_proto_cpu_info_t",3,"name")="gtcm_proto_cpu_info_t.proto_cpu"
	Set gtmtypes("gtcm_proto_cpu_info_t",3,"off")=16
	Set gtmtypes("gtcm_proto_cpu_info_t",3,"len")=8
	Set gtmtypes("gtcm_proto_cpu_info_t",3,"type")="addr"
	Set gtmtypfldindx("gtcm_proto_cpu_info_t","proto_cpu")=3
	;
	Set gtmtypes("gtcm_proto_os_info_t")="struct"
	Set gtmtypes("gtcm_proto_os_info_t",0)=3
	Set gtmtypes("gtcm_proto_os_info_t","len")=24
	Set gtmtypes("gtcm_proto_os_info_t",1,"name")="gtcm_proto_os_info_t.os_in_rel_str"
	Set gtmtypes("gtcm_proto_os_info_t",1,"off")=0
	Set gtmtypes("gtcm_proto_os_info_t",1,"len")=8
	Set gtmtypes("gtcm_proto_os_info_t",1,"type")="addr"
	Set gtmtypfldindx("gtcm_proto_os_info_t","os_in_rel_str")=1
	Set gtmtypes("gtcm_proto_os_info_t",2,"name")="gtcm_proto_os_info_t.size_of_os_in_rel_str"
	Set gtmtypes("gtcm_proto_os_info_t",2,"off")=8
	Set gtmtypes("gtcm_proto_os_info_t",2,"len")=4
	Set gtmtypes("gtcm_proto_os_info_t",2,"type")="int"
	Set gtmtypfldindx("gtcm_proto_os_info_t","size_of_os_in_rel_str")=2
	Set gtmtypes("gtcm_proto_os_info_t",3,"name")="gtcm_proto_os_info_t.proto_os"
	Set gtmtypes("gtcm_proto_os_info_t",3,"off")=16
	Set gtmtypes("gtcm_proto_os_info_t",3,"len")=8
	Set gtmtypes("gtcm_proto_os_info_t",3,"type")="addr"
	Set gtmtypfldindx("gtcm_proto_os_info_t","proto_os")=3
	;
	Set gtmtypes("gtm32_descriptor")="struct"
	Set gtmtypes("gtm32_descriptor",0)=3
	Set gtmtypes("gtm32_descriptor","len")=16
	Set gtmtypes("gtm32_descriptor",1,"name")="gtm32_descriptor.len"
	Set gtmtypes("gtm32_descriptor",1,"off")=0
	Set gtmtypes("gtm32_descriptor",1,"len")=4
	Set gtmtypes("gtm32_descriptor",1,"type")="unsigned-int"
	Set gtmtypfldindx("gtm32_descriptor","len")=1
	Set gtmtypes("gtm32_descriptor",2,"name")="gtm32_descriptor.type"
	Set gtmtypes("gtm32_descriptor",2,"off")=4
	Set gtmtypes("gtm32_descriptor",2,"len")=4
	Set gtmtypes("gtm32_descriptor",2,"type")="unsigned-int"
	Set gtmtypfldindx("gtm32_descriptor","type")=2
	Set gtmtypes("gtm32_descriptor",3,"name")="gtm32_descriptor.val"
	Set gtmtypes("gtm32_descriptor",3,"off")=8
	Set gtmtypes("gtm32_descriptor",3,"len")=8
	Set gtmtypes("gtm32_descriptor",3,"type")="addr"
	Set gtmtypfldindx("gtm32_descriptor","val")=3
	;
	Set gtmtypes("gtmImageName")="struct"
	Set gtmtypes("gtmImageName",0)=2
	Set gtmtypes("gtmImageName","len")=16
	Set gtmtypes("gtmImageName",1,"name")="gtmImageName.imageName"
	Set gtmtypes("gtmImageName",1,"off")=0
	Set gtmtypes("gtmImageName",1,"len")=8
	Set gtmtypes("gtmImageName",1,"type")="addr"
	Set gtmtypfldindx("gtmImageName","imageName")=1
	Set gtmtypes("gtmImageName",2,"name")="gtmImageName.imageNameLen"
	Set gtmtypes("gtmImageName",2,"off")=8
	Set gtmtypes("gtmImageName",2,"len")=4
	Set gtmtypes("gtmImageName",2,"type")="int"
	Set gtmtypfldindx("gtmImageName","imageNameLen")=2
	;
	Set gtmtypes("gtm_descriptor")="struct"
	Set gtmtypes("gtm_descriptor",0)=3
	Set gtmtypes("gtm_descriptor","len")=16
	Set gtmtypes("gtm_descriptor",1,"name")="gtm_descriptor.len"
	Set gtmtypes("gtm_descriptor",1,"off")=0
	Set gtmtypes("gtm_descriptor",1,"len")=2
	Set gtmtypes("gtm_descriptor",1,"type")="short"
	Set gtmtypfldindx("gtm_descriptor","len")=1
	Set gtmtypes("gtm_descriptor",2,"name")="gtm_descriptor.type"
	Set gtmtypes("gtm_descriptor",2,"off")=2
	Set gtmtypes("gtm_descriptor",2,"len")=2
	Set gtmtypes("gtm_descriptor",2,"type")="short"
	Set gtmtypfldindx("gtm_descriptor","type")=2
	Set gtmtypes("gtm_descriptor",3,"name")="gtm_descriptor.val"
	Set gtmtypes("gtm_descriptor",3,"off")=8
	Set gtmtypes("gtm_descriptor",3,"len")=8
	Set gtmtypes("gtm_descriptor",3,"type")="addr"
	Set gtmtypfldindx("gtm_descriptor","val")=3
	;
	Set gtmtypes("gtm_id_block")="struct"
	Set gtmtypes("gtm_id_block",0)=3
	Set gtmtypes("gtm_id_block","len")=76
	Set gtmtypes("gtm_id_block",1,"name")="gtm_id_block.sid"
	Set gtmtypes("gtm_id_block",1,"off")=0
	Set gtmtypes("gtm_id_block",1,"len")=64
	Set gtmtypes("gtm_id_block",1,"type")="unsigned-int"
	Set gtmtypfldindx("gtm_id_block","sid")=1
	Set gtmtypes("gtm_id_block",1,"dim")=16
	Set gtmtypes("gtm_id_block",2,"name")="gtm_id_block.exp_date"
	Set gtmtypes("gtm_id_block",2,"off")=64
	Set gtmtypes("gtm_id_block",2,"len")=4
	Set gtmtypes("gtm_id_block",2,"type")="unsigned-int"
	Set gtmtypfldindx("gtm_id_block","exp_date")=2
	Set gtmtypes("gtm_id_block",3,"name")="gtm_id_block.gtm_serial"
	Set gtmtypes("gtm_id_block",3,"off")=68
	Set gtmtypes("gtm_id_block",3,"len")=8
	Set gtmtypes("gtm_id_block",3,"type")="unsigned-char"
	Set gtmtypfldindx("gtm_id_block","gtm_serial")=3
	;
	Set gtmtypes("gtm_id_struct")="struct"
	Set gtmtypes("gtm_id_struct",0)=9
	Set gtmtypes("gtm_id_struct","len")=228
	Set gtmtypes("gtm_id_struct",1,"name")="gtm_id_struct.plaintext"
	Set gtmtypes("gtm_id_struct",1,"off")=0
	Set gtmtypes("gtm_id_struct",1,"len")=76
	Set gtmtypes("gtm_id_struct",1,"type")="gtm_id_block"
	Set gtmtypfldindx("gtm_id_struct","plaintext")=1
	Set gtmtypes("gtm_id_struct",2,"name")="gtm_id_struct.plaintext.sid"
	Set gtmtypes("gtm_id_struct",2,"off")=0
	Set gtmtypes("gtm_id_struct",2,"len")=64
	Set gtmtypes("gtm_id_struct",2,"type")="unsigned-int"
	Set gtmtypfldindx("gtm_id_struct","plaintext.sid")=2
	Set gtmtypes("gtm_id_struct",2,"dim")=16
	Set gtmtypes("gtm_id_struct",3,"name")="gtm_id_struct.plaintext.exp_date"
	Set gtmtypes("gtm_id_struct",3,"off")=64
	Set gtmtypes("gtm_id_struct",3,"len")=4
	Set gtmtypes("gtm_id_struct",3,"type")="unsigned-int"
	Set gtmtypfldindx("gtm_id_struct","plaintext.exp_date")=3
	Set gtmtypes("gtm_id_struct",4,"name")="gtm_id_struct.plaintext.gtm_serial"
	Set gtmtypes("gtm_id_struct",4,"off")=68
	Set gtmtypes("gtm_id_struct",4,"len")=8
	Set gtmtypes("gtm_id_struct",4,"type")="unsigned-char"
	Set gtmtypfldindx("gtm_id_struct","plaintext.gtm_serial")=4
	Set gtmtypes("gtm_id_struct",5,"name")="gtm_id_struct.key"
	Set gtmtypes("gtm_id_struct",5,"off")=76
	Set gtmtypes("gtm_id_struct",5,"len")=76
	Set gtmtypes("gtm_id_struct",5,"type")="unsigned-char"
	Set gtmtypfldindx("gtm_id_struct","key")=5
	Set gtmtypes("gtm_id_struct",6,"name")="gtm_id_struct.cryptext"
	Set gtmtypes("gtm_id_struct",6,"off")=152
	Set gtmtypes("gtm_id_struct",6,"len")=76
	Set gtmtypes("gtm_id_struct",6,"type")="gtm_id_block"
	Set gtmtypfldindx("gtm_id_struct","cryptext")=6
	Set gtmtypes("gtm_id_struct",7,"name")="gtm_id_struct.cryptext.sid"
	Set gtmtypes("gtm_id_struct",7,"off")=152
	Set gtmtypes("gtm_id_struct",7,"len")=64
	Set gtmtypes("gtm_id_struct",7,"type")="unsigned-int"
	Set gtmtypfldindx("gtm_id_struct","cryptext.sid")=7
	Set gtmtypes("gtm_id_struct",7,"dim")=16
	Set gtmtypes("gtm_id_struct",8,"name")="gtm_id_struct.cryptext.exp_date"
	Set gtmtypes("gtm_id_struct",8,"off")=216
	Set gtmtypes("gtm_id_struct",8,"len")=4
	Set gtmtypes("gtm_id_struct",8,"type")="unsigned-int"
	Set gtmtypfldindx("gtm_id_struct","cryptext.exp_date")=8
	Set gtmtypes("gtm_id_struct",9,"name")="gtm_id_struct.cryptext.gtm_serial"
	Set gtmtypes("gtm_id_struct",9,"off")=220
	Set gtmtypes("gtm_id_struct",9,"len")=8
	Set gtmtypes("gtm_id_struct",9,"type")="unsigned-char"
	Set gtmtypfldindx("gtm_id_struct","cryptext.gtm_serial")=9
	;
	Set gtmtypes("gtm_num_range_t")="struct"
	Set gtmtypes("gtm_num_range_t",0)=2
	Set gtmtypes("gtm_num_range_t","len")=8
	Set gtmtypes("gtm_num_range_t",1,"name")="gtm_num_range_t.min"
	Set gtmtypes("gtm_num_range_t",1,"off")=0
	Set gtmtypes("gtm_num_range_t",1,"len")=4
	Set gtmtypes("gtm_num_range_t",1,"type")="unsigned-int"
	Set gtmtypfldindx("gtm_num_range_t","min")=1
	Set gtmtypes("gtm_num_range_t",2,"name")="gtm_num_range_t.max"
	Set gtmtypes("gtm_num_range_t",2,"off")=4
	Set gtmtypes("gtm_num_range_t",2,"len")=4
	Set gtmtypes("gtm_num_range_t",2,"type")="unsigned-int"
	Set gtmtypfldindx("gtm_num_range_t","max")=2
	;
	Set gtmtypes("gtm_rtn_src_chksum_ctx")="struct"
	Set gtmtypes("gtm_rtn_src_chksum_ctx",0)=6
	Set gtmtypes("gtm_rtn_src_chksum_ctx","len")=64
	Set gtmtypes("gtm_rtn_src_chksum_ctx",1,"name")="gtm_rtn_src_chksum_ctx.hash_state"
	Set gtmtypes("gtm_rtn_src_chksum_ctx",1,"off")=0
	Set gtmtypes("gtm_rtn_src_chksum_ctx",1,"len")=40
	Set gtmtypes("gtm_rtn_src_chksum_ctx",1,"type")="hash128_state_t"
	Set gtmtypfldindx("gtm_rtn_src_chksum_ctx","hash_state")=1
	Set gtmtypes("gtm_rtn_src_chksum_ctx",2,"name")="gtm_rtn_src_chksum_ctx.hash_state.h"
	Set gtmtypes("gtm_rtn_src_chksum_ctx",2,"off")=0
	Set gtmtypes("gtm_rtn_src_chksum_ctx",2,"len")=16
	Set gtmtypes("gtm_rtn_src_chksum_ctx",2,"type")="gtm_uint16"
	Set gtmtypfldindx("gtm_rtn_src_chksum_ctx","hash_state.h")=2
	Set gtmtypes("gtm_rtn_src_chksum_ctx",3,"name")="gtm_rtn_src_chksum_ctx.hash_state.c"
	Set gtmtypes("gtm_rtn_src_chksum_ctx",3,"off")=16
	Set gtmtypes("gtm_rtn_src_chksum_ctx",3,"len")=16
	Set gtmtypes("gtm_rtn_src_chksum_ctx",3,"type")="gtm_uint16"
	Set gtmtypfldindx("gtm_rtn_src_chksum_ctx","hash_state.c")=3
	Set gtmtypes("gtm_rtn_src_chksum_ctx",4,"name")="gtm_rtn_src_chksum_ctx.hash_state.carry_bytes"
	Set gtmtypes("gtm_rtn_src_chksum_ctx",4,"off")=32
	Set gtmtypes("gtm_rtn_src_chksum_ctx",4,"len")=4
	Set gtmtypes("gtm_rtn_src_chksum_ctx",4,"type")="int"
	Set gtmtypfldindx("gtm_rtn_src_chksum_ctx","hash_state.carry_bytes")=4
	Set gtmtypes("gtm_rtn_src_chksum_ctx",5,"name")="gtm_rtn_src_chksum_ctx.total_size"
	Set gtmtypes("gtm_rtn_src_chksum_ctx",5,"off")=40
	Set gtmtypes("gtm_rtn_src_chksum_ctx",5,"len")=4
	Set gtmtypes("gtm_rtn_src_chksum_ctx",5,"type")="unsigned-int"
	Set gtmtypfldindx("gtm_rtn_src_chksum_ctx","total_size")=5
	Set gtmtypes("gtm_rtn_src_chksum_ctx",6,"name")="gtm_rtn_src_chksum_ctx.digest"
	Set gtmtypes("gtm_rtn_src_chksum_ctx",6,"off")=44
	Set gtmtypes("gtm_rtn_src_chksum_ctx",6,"len")=16
	Set gtmtypes("gtm_rtn_src_chksum_ctx",6,"type")="unsigned-char"
	Set gtmtypfldindx("gtm_rtn_src_chksum_ctx","digest")=6
	;
	Set gtmtypes("gtm_time8")="union"
	Set gtmtypes("gtm_time8",0)=2
	Set gtmtypes("gtm_time8","len")=8
	Set gtmtypes("gtm_time8",1,"name")="gtm_time8.ctime"
	Set gtmtypes("gtm_time8",1,"off")=0
	Set gtmtypes("gtm_time8",1,"len")=8
	Set gtmtypes("gtm_time8",1,"type")="time_t"
	Set gtmtypfldindx("gtm_time8","ctime")=1
	Set gtmtypes("gtm_time8",2,"name")="gtm_time8.filler"
	Set gtmtypes("gtm_time8",2,"off")=0
	Set gtmtypes("gtm_time8",2,"len")=8
	Set gtmtypes("gtm_time8",2,"type")="int"
	Set gtmtypfldindx("gtm_time8","filler")=2
	Set gtmtypes("gtm_time8",2,"dim")=2
	;
	Set gtmtypes("gtm_tls_conn_info")="struct"
	Set gtmtypes("gtm_tls_conn_info",0)=17
	Set gtmtypes("gtm_tls_conn_info","len")=896
	Set gtmtypes("gtm_tls_conn_info",1,"name")="gtm_tls_conn_info.protocol"
	Set gtmtypes("gtm_tls_conn_info",1,"off")=0
	Set gtmtypes("gtm_tls_conn_info",1,"len")=64
	Set gtmtypes("gtm_tls_conn_info",1,"type")="char"
	Set gtmtypfldindx("gtm_tls_conn_info","protocol")=1
	Set gtmtypes("gtm_tls_conn_info",2,"name")="gtm_tls_conn_info.session_algo"
	Set gtmtypes("gtm_tls_conn_info",2,"off")=64
	Set gtmtypes("gtm_tls_conn_info",2,"len")=64
	Set gtmtypes("gtm_tls_conn_info",2,"type")="char"
	Set gtmtypfldindx("gtm_tls_conn_info","session_algo")=2
	Set gtmtypes("gtm_tls_conn_info",3,"name")="gtm_tls_conn_info.session_id"
	Set gtmtypes("gtm_tls_conn_info",3,"off")=128
	Set gtmtypes("gtm_tls_conn_info",3,"len")=65
	Set gtmtypes("gtm_tls_conn_info",3,"type")="char"
	Set gtmtypfldindx("gtm_tls_conn_info","session_id")=3
	Set gtmtypes("gtm_tls_conn_info",4,"name")="gtm_tls_conn_info.compression"
	Set gtmtypes("gtm_tls_conn_info",4,"off")=200
	Set gtmtypes("gtm_tls_conn_info",4,"len")=8
	Set gtmtypes("gtm_tls_conn_info",4,"type")="addr"
	Set gtmtypfldindx("gtm_tls_conn_info","compression")=4
	Set gtmtypes("gtm_tls_conn_info",5,"name")="gtm_tls_conn_info.secure_renegotiation"
	Set gtmtypes("gtm_tls_conn_info",5,"off")=208
	Set gtmtypes("gtm_tls_conn_info",5,"len")=4
	Set gtmtypes("gtm_tls_conn_info",5,"type")="int"
	Set gtmtypfldindx("gtm_tls_conn_info","secure_renegotiation")=5
	Set gtmtypes("gtm_tls_conn_info",6,"name")="gtm_tls_conn_info.reused"
	Set gtmtypes("gtm_tls_conn_info",6,"off")=212
	Set gtmtypes("gtm_tls_conn_info",6,"len")=4
	Set gtmtypes("gtm_tls_conn_info",6,"type")="int"
	Set gtmtypfldindx("gtm_tls_conn_info","reused")=6
	Set gtmtypes("gtm_tls_conn_info",7,"name")="gtm_tls_conn_info.session_expiry_timeout"
	Set gtmtypes("gtm_tls_conn_info",7,"off")=216
	Set gtmtypes("gtm_tls_conn_info",7,"len")=8
	Set gtmtypes("gtm_tls_conn_info",7,"type")="long"
	Set gtmtypfldindx("gtm_tls_conn_info","session_expiry_timeout")=7
	Set gtmtypes("gtm_tls_conn_info",8,"name")="gtm_tls_conn_info.cert_algo"
	Set gtmtypes("gtm_tls_conn_info",8,"off")=224
	Set gtmtypes("gtm_tls_conn_info",8,"len")=64
	Set gtmtypes("gtm_tls_conn_info",8,"type")="char"
	Set gtmtypfldindx("gtm_tls_conn_info","cert_algo")=8
	Set gtmtypes("gtm_tls_conn_info",9,"name")="gtm_tls_conn_info.cert_nbits"
	Set gtmtypes("gtm_tls_conn_info",9,"off")=288
	Set gtmtypes("gtm_tls_conn_info",9,"len")=4
	Set gtmtypes("gtm_tls_conn_info",9,"type")="int"
	Set gtmtypfldindx("gtm_tls_conn_info","cert_nbits")=9
	Set gtmtypes("gtm_tls_conn_info",10,"name")="gtm_tls_conn_info.subject"
	Set gtmtypes("gtm_tls_conn_info",10,"off")=292
	Set gtmtypes("gtm_tls_conn_info",10,"len")=256
	Set gtmtypes("gtm_tls_conn_info",10,"type")="char"
	Set gtmtypfldindx("gtm_tls_conn_info","subject")=10
	Set gtmtypes("gtm_tls_conn_info",11,"name")="gtm_tls_conn_info.issuer"
	Set gtmtypes("gtm_tls_conn_info",11,"off")=548
	Set gtmtypes("gtm_tls_conn_info",11,"len")=256
	Set gtmtypes("gtm_tls_conn_info",11,"type")="char"
	Set gtmtypfldindx("gtm_tls_conn_info","issuer")=11
	Set gtmtypes("gtm_tls_conn_info",12,"name")="gtm_tls_conn_info.not_before"
	Set gtmtypes("gtm_tls_conn_info",12,"off")=804
	Set gtmtypes("gtm_tls_conn_info",12,"len")=32
	Set gtmtypes("gtm_tls_conn_info",12,"type")="char"
	Set gtmtypfldindx("gtm_tls_conn_info","not_before")=12
	Set gtmtypes("gtm_tls_conn_info",13,"name")="gtm_tls_conn_info.not_after"
	Set gtmtypes("gtm_tls_conn_info",13,"off")=836
	Set gtmtypes("gtm_tls_conn_info",13,"len")=32
	Set gtmtypes("gtm_tls_conn_info",13,"type")="char"
	Set gtmtypfldindx("gtm_tls_conn_info","not_after")=13
	Set gtmtypes("gtm_tls_conn_info",14,"name")="gtm_tls_conn_info.options"
	Set gtmtypes("gtm_tls_conn_info",14,"off")=872
	Set gtmtypes("gtm_tls_conn_info",14,"len")=8
	Set gtmtypes("gtm_tls_conn_info",14,"type")="long"
	Set gtmtypfldindx("gtm_tls_conn_info","options")=14
	Set gtmtypes("gtm_tls_conn_info",15,"name")="gtm_tls_conn_info.renegotiation_pending"
	Set gtmtypes("gtm_tls_conn_info",15,"off")=880
	Set gtmtypes("gtm_tls_conn_info",15,"len")=4
	Set gtmtypes("gtm_tls_conn_info",15,"type")="int"
	Set gtmtypfldindx("gtm_tls_conn_info","renegotiation_pending")=15
	Set gtmtypes("gtm_tls_conn_info",16,"name")="gtm_tls_conn_info.total_renegotiations"
	Set gtmtypes("gtm_tls_conn_info",16,"off")=884
	Set gtmtypes("gtm_tls_conn_info",16,"len")=4
	Set gtmtypes("gtm_tls_conn_info",16,"type")="int"
	Set gtmtypfldindx("gtm_tls_conn_info","total_renegotiations")=16
	Set gtmtypes("gtm_tls_conn_info",17,"name")="gtm_tls_conn_info.verify_mode"
	Set gtmtypes("gtm_tls_conn_info",17,"off")=888
	Set gtmtypes("gtm_tls_conn_info",17,"len")=4
	Set gtmtypes("gtm_tls_conn_info",17,"type")="int"
	Set gtmtypfldindx("gtm_tls_conn_info","verify_mode")=17
	;
	Set gtmtypes("gtm_tls_ctx_t")="struct"
	Set gtmtypes("gtm_tls_ctx_t",0)=6
	Set gtmtypes("gtm_tls_ctx_t","len")=40
	Set gtmtypes("gtm_tls_ctx_t",1,"name")="gtm_tls_ctx_t.flags"
	Set gtmtypes("gtm_tls_ctx_t",1,"off")=0
	Set gtmtypes("gtm_tls_ctx_t",1,"len")=4
	Set gtmtypes("gtm_tls_ctx_t",1,"type")="int"
	Set gtmtypfldindx("gtm_tls_ctx_t","flags")=1
	Set gtmtypes("gtm_tls_ctx_t",2,"name")="gtm_tls_ctx_t.fips_mode"
	Set gtmtypes("gtm_tls_ctx_t",2,"off")=4
	Set gtmtypes("gtm_tls_ctx_t",2,"len")=4
	Set gtmtypes("gtm_tls_ctx_t",2,"type")="int"
	Set gtmtypfldindx("gtm_tls_ctx_t","fips_mode")=2
	Set gtmtypes("gtm_tls_ctx_t",3,"name")="gtm_tls_ctx_t.compile_time_version"
	Set gtmtypes("gtm_tls_ctx_t",3,"off")=8
	Set gtmtypes("gtm_tls_ctx_t",3,"len")=8
	Set gtmtypes("gtm_tls_ctx_t",3,"type")="unsigned-long"
	Set gtmtypfldindx("gtm_tls_ctx_t","compile_time_version")=3
	Set gtmtypes("gtm_tls_ctx_t",4,"name")="gtm_tls_ctx_t.runtime_version"
	Set gtmtypes("gtm_tls_ctx_t",4,"off")=16
	Set gtmtypes("gtm_tls_ctx_t",4,"len")=8
	Set gtmtypes("gtm_tls_ctx_t",4,"type")="unsigned-long"
	Set gtmtypfldindx("gtm_tls_ctx_t","runtime_version")=4
	Set gtmtypes("gtm_tls_ctx_t",5,"name")="gtm_tls_ctx_t.ctx"
	Set gtmtypes("gtm_tls_ctx_t",5,"off")=24
	Set gtmtypes("gtm_tls_ctx_t",5,"len")=8
	Set gtmtypes("gtm_tls_ctx_t",5,"type")="addr"
	Set gtmtypfldindx("gtm_tls_ctx_t","ctx")=5
	Set gtmtypes("gtm_tls_ctx_t",6,"name")="gtm_tls_ctx_t.version"
	Set gtmtypes("gtm_tls_ctx_t",6,"off")=32
	Set gtmtypes("gtm_tls_ctx_t",6,"len")=4
	Set gtmtypes("gtm_tls_ctx_t",6,"type")="int"
	Set gtmtypfldindx("gtm_tls_ctx_t","version")=6
	;
	Set gtmtypes("gtm_tls_socket_t")="struct"
	Set gtmtypes("gtm_tls_socket_t",0)=5
	Set gtmtypes("gtm_tls_socket_t","len")=72
	Set gtmtypes("gtm_tls_socket_t",1,"name")="gtm_tls_socket_t.flags"
	Set gtmtypes("gtm_tls_socket_t",1,"off")=0
	Set gtmtypes("gtm_tls_socket_t",1,"len")=4
	Set gtmtypes("gtm_tls_socket_t",1,"type")="int"
	Set gtmtypfldindx("gtm_tls_socket_t","flags")=1
	Set gtmtypes("gtm_tls_socket_t",2,"name")="gtm_tls_socket_t.ssl"
	Set gtmtypes("gtm_tls_socket_t",2,"off")=8
	Set gtmtypes("gtm_tls_socket_t",2,"len")=8
	Set gtmtypes("gtm_tls_socket_t",2,"type")="addr"
	Set gtmtypfldindx("gtm_tls_socket_t","ssl")=2
	Set gtmtypes("gtm_tls_socket_t",3,"name")="gtm_tls_socket_t.session"
	Set gtmtypes("gtm_tls_socket_t",3,"off")=16
	Set gtmtypes("gtm_tls_socket_t",3,"len")=8
	Set gtmtypes("gtm_tls_socket_t",3,"type")="addr"
	Set gtmtypfldindx("gtm_tls_socket_t","session")=3
	Set gtmtypes("gtm_tls_socket_t",4,"name")="gtm_tls_socket_t.tlsid"
	Set gtmtypes("gtm_tls_socket_t",4,"off")=24
	Set gtmtypes("gtm_tls_socket_t",4,"len")=33
	Set gtmtypes("gtm_tls_socket_t",4,"type")="char"
	Set gtmtypfldindx("gtm_tls_socket_t","tlsid")=4
	Set gtmtypes("gtm_tls_socket_t",5,"name")="gtm_tls_socket_t.gtm_ctx"
	Set gtmtypes("gtm_tls_socket_t",5,"off")=64
	Set gtmtypes("gtm_tls_socket_t",5,"len")=8
	Set gtmtypes("gtm_tls_socket_t",5,"type")="addr"
	Set gtmtypfldindx("gtm_tls_socket_t","gtm_ctx")=5
	;
	Set gtmtypes("gtm_trigger_parms")="struct"
	Set gtmtypes("gtm_trigger_parms",0)=8
	Set gtmtypes("gtm_trigger_parms","len")=64
	Set gtmtypes("gtm_trigger_parms",1,"name")="gtm_trigger_parms.ztoldval_new"
	Set gtmtypes("gtm_trigger_parms",1,"off")=0
	Set gtmtypes("gtm_trigger_parms",1,"len")=8
	Set gtmtypes("gtm_trigger_parms",1,"type")="addr"
	Set gtmtypfldindx("gtm_trigger_parms","ztoldval_new")=1
	Set gtmtypes("gtm_trigger_parms",2,"name")="gtm_trigger_parms.ztvalue_new"
	Set gtmtypes("gtm_trigger_parms",2,"off")=8
	Set gtmtypes("gtm_trigger_parms",2,"len")=8
	Set gtmtypes("gtm_trigger_parms",2,"type")="addr"
	Set gtmtypfldindx("gtm_trigger_parms","ztvalue_new")=2
	Set gtmtypes("gtm_trigger_parms",3,"name")="gtm_trigger_parms.ztdata_new"
	Set gtmtypes("gtm_trigger_parms",3,"off")=16
	Set gtmtypes("gtm_trigger_parms",3,"len")=8
	Set gtmtypes("gtm_trigger_parms",3,"type")="addr"
	Set gtmtypfldindx("gtm_trigger_parms","ztdata_new")=3
	Set gtmtypes("gtm_trigger_parms",4,"name")="gtm_trigger_parms.ztdelim_new"
	Set gtmtypes("gtm_trigger_parms",4,"off")=24
	Set gtmtypes("gtm_trigger_parms",4,"len")=8
	Set gtmtypes("gtm_trigger_parms",4,"type")="addr"
	Set gtmtypfldindx("gtm_trigger_parms","ztdelim_new")=4
	Set gtmtypes("gtm_trigger_parms",5,"name")="gtm_trigger_parms.ztriggerop_new"
	Set gtmtypes("gtm_trigger_parms",5,"off")=32
	Set gtmtypes("gtm_trigger_parms",5,"len")=8
	Set gtmtypes("gtm_trigger_parms",5,"type")="addr"
	Set gtmtypfldindx("gtm_trigger_parms","ztriggerop_new")=5
	Set gtmtypes("gtm_trigger_parms",6,"name")="gtm_trigger_parms.ztupdate_new"
	Set gtmtypes("gtm_trigger_parms",6,"off")=40
	Set gtmtypes("gtm_trigger_parms",6,"len")=8
	Set gtmtypes("gtm_trigger_parms",6,"type")="addr"
	Set gtmtypfldindx("gtm_trigger_parms","ztupdate_new")=6
	Set gtmtypes("gtm_trigger_parms",7,"name")="gtm_trigger_parms.lvvalarray"
	Set gtmtypes("gtm_trigger_parms",7,"off")=48
	Set gtmtypes("gtm_trigger_parms",7,"len")=8
	Set gtmtypes("gtm_trigger_parms",7,"type")="addr"
	Set gtmtypfldindx("gtm_trigger_parms","lvvalarray")=7
	Set gtmtypes("gtm_trigger_parms",8,"name")="gtm_trigger_parms.ztvalue_changed"
	Set gtmtypes("gtm_trigger_parms",8,"off")=56
	Set gtmtypes("gtm_trigger_parms",8,"len")=4
	Set gtmtypes("gtm_trigger_parms",8,"type")="boolean_t"
	Set gtmtypfldindx("gtm_trigger_parms","ztvalue_changed")=8
	;
	Set gtmtypes("gtm_uint16")="struct"
	Set gtmtypes("gtm_uint16",0)=2
	Set gtmtypes("gtm_uint16","len")=16
	Set gtmtypes("gtm_uint16",1,"name")="gtm_uint16.one"
	Set gtmtypes("gtm_uint16",1,"off")=0
	Set gtmtypes("gtm_uint16",1,"len")=8
	Set gtmtypes("gtm_uint16",1,"type")="gtm_uint8"
	Set gtmtypfldindx("gtm_uint16","one")=1
	Set gtmtypes("gtm_uint16",2,"name")="gtm_uint16.two"
	Set gtmtypes("gtm_uint16",2,"off")=8
	Set gtmtypes("gtm_uint16",2,"len")=8
	Set gtmtypes("gtm_uint16",2,"type")="gtm_uint8"
	Set gtmtypfldindx("gtm_uint16","two")=2
	;
	Set gtmtypes("gtmrecv_local_struct")="struct"
	Set gtmtypes("gtmrecv_local_struct",0)=41
	Set gtmtypes("gtmrecv_local_struct","len")=1288
	Set gtmtypes("gtmrecv_local_struct",1,"name")="gtmrecv_local_struct.recv_serv_pid"
	Set gtmtypes("gtmrecv_local_struct",1,"off")=0
	Set gtmtypes("gtmrecv_local_struct",1,"len")=4
	Set gtmtypes("gtmrecv_local_struct",1,"type")="unsigned-int"
	Set gtmtypfldindx("gtmrecv_local_struct","recv_serv_pid")=1
	Set gtmtypes("gtmrecv_local_struct",2,"name")="gtmrecv_local_struct.lastrecvd_time"
	Set gtmtypes("gtmrecv_local_struct",2,"off")=4
	Set gtmtypes("gtmrecv_local_struct",2,"len")=4
	Set gtmtypes("gtmrecv_local_struct",2,"type")="int"
	Set gtmtypfldindx("gtmrecv_local_struct","lastrecvd_time")=2
	Set gtmtypes("gtmrecv_local_struct",3,"name")="gtmrecv_local_struct.statslog"
	Set gtmtypes("gtmrecv_local_struct",3,"off")=8
	Set gtmtypes("gtmrecv_local_struct",3,"len")=4
	Set gtmtypes("gtmrecv_local_struct",3,"type")="unsigned-int"
	Set gtmtypfldindx("gtmrecv_local_struct","statslog")=3
	Set gtmtypes("gtmrecv_local_struct",4,"name")="gtmrecv_local_struct.shutdown"
	Set gtmtypes("gtmrecv_local_struct",4,"off")=12
	Set gtmtypes("gtmrecv_local_struct",4,"len")=4
	Set gtmtypes("gtmrecv_local_struct",4,"type")="unsigned-int"
	Set gtmtypfldindx("gtmrecv_local_struct","shutdown")=4
	Set gtmtypes("gtmrecv_local_struct",5,"name")="gtmrecv_local_struct.shutdown_time"
	Set gtmtypes("gtmrecv_local_struct",5,"off")=16
	Set gtmtypes("gtmrecv_local_struct",5,"len")=4
	Set gtmtypes("gtmrecv_local_struct",5,"type")="int"
	Set gtmtypfldindx("gtmrecv_local_struct","shutdown_time")=5
	Set gtmtypes("gtmrecv_local_struct",6,"name")="gtmrecv_local_struct.listen_port"
	Set gtmtypes("gtmrecv_local_struct",6,"off")=20
	Set gtmtypes("gtmrecv_local_struct",6,"len")=4
	Set gtmtypes("gtmrecv_local_struct",6,"type")="int"
	Set gtmtypfldindx("gtmrecv_local_struct","listen_port")=6
	Set gtmtypes("gtmrecv_local_struct",7,"name")="gtmrecv_local_struct.restart"
	Set gtmtypes("gtmrecv_local_struct",7,"off")=24
	Set gtmtypes("gtmrecv_local_struct",7,"len")=4
	Set gtmtypes("gtmrecv_local_struct",7,"type")="unsigned-int"
	Set gtmtypfldindx("gtmrecv_local_struct","restart")=7
	Set gtmtypes("gtmrecv_local_struct",8,"name")="gtmrecv_local_struct.changelog"
	Set gtmtypes("gtmrecv_local_struct",8,"off")=28
	Set gtmtypes("gtmrecv_local_struct",8,"len")=4
	Set gtmtypes("gtmrecv_local_struct",8,"type")="unsigned-int"
	Set gtmtypfldindx("gtmrecv_local_struct","changelog")=8
	Set gtmtypes("gtmrecv_local_struct",9,"name")="gtmrecv_local_struct.log_interval"
	Set gtmtypes("gtmrecv_local_struct",9,"off")=32
	Set gtmtypes("gtmrecv_local_struct",9,"len")=4
	Set gtmtypes("gtmrecv_local_struct",9,"type")="unsigned-int"
	Set gtmtypfldindx("gtmrecv_local_struct","log_interval")=9
	Set gtmtypes("gtmrecv_local_struct",10,"name")="gtmrecv_local_struct.filter_cmd"
	Set gtmtypes("gtmrecv_local_struct",10,"off")=36
	Set gtmtypes("gtmrecv_local_struct",10,"len")=512
	Set gtmtypes("gtmrecv_local_struct",10,"type")="char"
	Set gtmtypfldindx("gtmrecv_local_struct","filter_cmd")=10
	Set gtmtypes("gtmrecv_local_struct",11,"name")="gtmrecv_local_struct.log_file"
	Set gtmtypes("gtmrecv_local_struct",11,"off")=548
	Set gtmtypes("gtmrecv_local_struct",11,"len")=256
	Set gtmtypes("gtmrecv_local_struct",11,"type")="char"
	Set gtmtypfldindx("gtmrecv_local_struct","log_file")=11
	Set gtmtypes("gtmrecv_local_struct",12,"name")="gtmrecv_local_struct.statslog_file"
	Set gtmtypes("gtmrecv_local_struct",12,"off")=804
	Set gtmtypes("gtmrecv_local_struct",12,"len")=256
	Set gtmtypes("gtmrecv_local_struct",12,"type")="char"
	Set gtmtypfldindx("gtmrecv_local_struct","statslog_file")=12
	Set gtmtypes("gtmrecv_local_struct",13,"name")="gtmrecv_local_struct.remote_side"
	Set gtmtypes("gtmrecv_local_struct",13,"off")=1060
	Set gtmtypes("gtmrecv_local_struct",13,"len")=48
	Set gtmtypes("gtmrecv_local_struct",13,"type")="repl_conn_info_t"
	Set gtmtypfldindx("gtmrecv_local_struct","remote_side")=13
	Set gtmtypes("gtmrecv_local_struct",14,"name")="gtmrecv_local_struct.remote_side.proto_ver"
	Set gtmtypes("gtmrecv_local_struct",14,"off")=1060
	Set gtmtypes("gtmrecv_local_struct",14,"len")=4
	Set gtmtypes("gtmrecv_local_struct",14,"type")="int"
	Set gtmtypfldindx("gtmrecv_local_struct","remote_side.proto_ver")=14
	Set gtmtypes("gtmrecv_local_struct",15,"name")="gtmrecv_local_struct.remote_side.jnl_ver"
	Set gtmtypes("gtmrecv_local_struct",15,"off")=1064
	Set gtmtypes("gtmrecv_local_struct",15,"len")=4
	Set gtmtypes("gtmrecv_local_struct",15,"type")="unsigned-int"
	Set gtmtypfldindx("gtmrecv_local_struct","remote_side.jnl_ver")=15
	Set gtmtypes("gtmrecv_local_struct",16,"name")="gtmrecv_local_struct.remote_side.is_std_null_coll"
	Set gtmtypes("gtmrecv_local_struct",16,"off")=1068
	Set gtmtypes("gtmrecv_local_struct",16,"len")=4
	Set gtmtypes("gtmrecv_local_struct",16,"type")="boolean_t"
	Set gtmtypfldindx("gtmrecv_local_struct","remote_side.is_std_null_coll")=16
	Set gtmtypes("gtmrecv_local_struct",17,"name")="gtmrecv_local_struct.remote_side.trigger_supported"
	Set gtmtypes("gtmrecv_local_struct",17,"off")=1072
	Set gtmtypes("gtmrecv_local_struct",17,"len")=4
	Set gtmtypes("gtmrecv_local_struct",17,"type")="boolean_t"
	Set gtmtypfldindx("gtmrecv_local_struct","remote_side.trigger_supported")=17
	Set gtmtypes("gtmrecv_local_struct",18,"name")="gtmrecv_local_struct.remote_side.cross_endian"
	Set gtmtypes("gtmrecv_local_struct",18,"off")=1076
	Set gtmtypes("gtmrecv_local_struct",18,"len")=4
	Set gtmtypes("gtmrecv_local_struct",18,"type")="boolean_t"
	Set gtmtypfldindx("gtmrecv_local_struct","remote_side.cross_endian")=18
	Set gtmtypes("gtmrecv_local_struct",19,"name")="gtmrecv_local_struct.remote_side.endianness_known"
	Set gtmtypes("gtmrecv_local_struct",19,"off")=1080
	Set gtmtypes("gtmrecv_local_struct",19,"len")=4
	Set gtmtypes("gtmrecv_local_struct",19,"type")="boolean_t"
	Set gtmtypfldindx("gtmrecv_local_struct","remote_side.endianness_known")=19
	Set gtmtypes("gtmrecv_local_struct",20,"name")="gtmrecv_local_struct.remote_side.null_subs_xform"
	Set gtmtypes("gtmrecv_local_struct",20,"off")=1084
	Set gtmtypes("gtmrecv_local_struct",20,"len")=4
	Set gtmtypes("gtmrecv_local_struct",20,"type")="boolean_t"
	Set gtmtypfldindx("gtmrecv_local_struct","remote_side.null_subs_xform")=20
	Set gtmtypes("gtmrecv_local_struct",21,"name")="gtmrecv_local_struct.remote_side.is_supplementary"
	Set gtmtypes("gtmrecv_local_struct",21,"off")=1088
	Set gtmtypes("gtmrecv_local_struct",21,"len")=4
	Set gtmtypes("gtmrecv_local_struct",21,"type")="boolean_t"
	Set gtmtypfldindx("gtmrecv_local_struct","remote_side.is_supplementary")=21
	Set gtmtypes("gtmrecv_local_struct",22,"name")="gtmrecv_local_struct.remote_side.tls_requested"
	Set gtmtypes("gtmrecv_local_struct",22,"off")=1092
	Set gtmtypes("gtmrecv_local_struct",22,"len")=4
	Set gtmtypes("gtmrecv_local_struct",22,"type")="boolean_t"
	Set gtmtypfldindx("gtmrecv_local_struct","remote_side.tls_requested")=22
	Set gtmtypes("gtmrecv_local_struct",23,"name")="gtmrecv_local_struct.remote_side.filler_16"
	Set gtmtypes("gtmrecv_local_struct",23,"off")=1096
	Set gtmtypes("gtmrecv_local_struct",23,"len")=12
	Set gtmtypes("gtmrecv_local_struct",23,"type")="char"
	Set gtmtypfldindx("gtmrecv_local_struct","remote_side.filler_16")=23
	Set gtmtypes("gtmrecv_local_struct",24,"name")="gtmrecv_local_struct.strm_index"
	Set gtmtypes("gtmrecv_local_struct",24,"off")=1108
	Set gtmtypes("gtmrecv_local_struct",24,"len")=4
	Set gtmtypes("gtmrecv_local_struct",24,"type")="int"
	Set gtmtypfldindx("gtmrecv_local_struct","strm_index")=24
	Set gtmtypes("gtmrecv_local_struct",25,"name")="gtmrecv_local_struct.updateresync"
	Set gtmtypes("gtmrecv_local_struct",25,"off")=1112
	Set gtmtypes("gtmrecv_local_struct",25,"len")=4
	Set gtmtypes("gtmrecv_local_struct",25,"type")="boolean_t"
	Set gtmtypfldindx("gtmrecv_local_struct","updateresync")=25
	Set gtmtypes("gtmrecv_local_struct",26,"name")="gtmrecv_local_struct.noresync"
	Set gtmtypes("gtmrecv_local_struct",26,"off")=1116
	Set gtmtypes("gtmrecv_local_struct",26,"len")=4
	Set gtmtypes("gtmrecv_local_struct",26,"type")="boolean_t"
	Set gtmtypfldindx("gtmrecv_local_struct","noresync")=26
	Set gtmtypes("gtmrecv_local_struct",27,"name")="gtmrecv_local_struct.updresync_instfile_fd"
	Set gtmtypes("gtmrecv_local_struct",27,"off")=1120
	Set gtmtypes("gtmrecv_local_struct",27,"len")=4
	Set gtmtypes("gtmrecv_local_struct",27,"type")="int"
	Set gtmtypfldindx("gtmrecv_local_struct","updresync_instfile_fd")=27
	Set gtmtypes("gtmrecv_local_struct",28,"name")="gtmrecv_local_struct.updresync_num_histinfo"
	Set gtmtypes("gtmrecv_local_struct",28,"off")=1124
	Set gtmtypes("gtmrecv_local_struct",28,"len")=4
	Set gtmtypes("gtmrecv_local_struct",28,"type")="int"
	Set gtmtypfldindx("gtmrecv_local_struct","updresync_num_histinfo")=28
	Set gtmtypes("gtmrecv_local_struct",29,"name")="gtmrecv_local_struct.updresync_cross_endian"
	Set gtmtypes("gtmrecv_local_struct",29,"off")=1128
	Set gtmtypes("gtmrecv_local_struct",29,"len")=4
	Set gtmtypes("gtmrecv_local_struct",29,"type")="boolean_t"
	Set gtmtypfldindx("gtmrecv_local_struct","updresync_cross_endian")=29
	Set gtmtypes("gtmrecv_local_struct",30,"name")="gtmrecv_local_struct.updresync_num_histinfo_strm"
	Set gtmtypes("gtmrecv_local_struct",30,"off")=1132
	Set gtmtypes("gtmrecv_local_struct",30,"len")=64
	Set gtmtypes("gtmrecv_local_struct",30,"type")="int"
	Set gtmtypfldindx("gtmrecv_local_struct","updresync_num_histinfo_strm")=30
	Set gtmtypes("gtmrecv_local_struct",30,"dim")=16
	Set gtmtypes("gtmrecv_local_struct",31,"name")="gtmrecv_local_struct.updresync_lms_group"
	Set gtmtypes("gtmrecv_local_struct",31,"off")=1196
	Set gtmtypes("gtmrecv_local_struct",31,"len")=40
	Set gtmtypes("gtmrecv_local_struct",31,"type")="repl_inst_uuid"
	Set gtmtypfldindx("gtmrecv_local_struct","updresync_lms_group")=31
	Set gtmtypes("gtmrecv_local_struct",32,"name")="gtmrecv_local_struct.updresync_lms_group.created_nodename"
	Set gtmtypes("gtmrecv_local_struct",32,"off")=1196
	Set gtmtypes("gtmrecv_local_struct",32,"len")=16
	Set gtmtypes("gtmrecv_local_struct",32,"type")="unsigned-char"
	Set gtmtypfldindx("gtmrecv_local_struct","updresync_lms_group.created_nodename")=32
	Set gtmtypes("gtmrecv_local_struct",33,"name")="gtmrecv_local_struct.updresync_lms_group.this_instname"
	Set gtmtypes("gtmrecv_local_struct",33,"off")=1212
	Set gtmtypes("gtmrecv_local_struct",33,"len")=16
	Set gtmtypes("gtmrecv_local_struct",33,"type")="unsigned-char"
	Set gtmtypfldindx("gtmrecv_local_struct","updresync_lms_group.this_instname")=33
	Set gtmtypes("gtmrecv_local_struct",34,"name")="gtmrecv_local_struct.updresync_lms_group.created_time"
	Set gtmtypes("gtmrecv_local_struct",34,"off")=1228
	Set gtmtypes("gtmrecv_local_struct",34,"len")=4
	Set gtmtypes("gtmrecv_local_struct",34,"type")="unsigned-int"
	Set gtmtypfldindx("gtmrecv_local_struct","updresync_lms_group.created_time")=34
	Set gtmtypes("gtmrecv_local_struct",35,"name")="gtmrecv_local_struct.updresync_lms_group.creator_pid"
	Set gtmtypes("gtmrecv_local_struct",35,"off")=1232
	Set gtmtypes("gtmrecv_local_struct",35,"len")=4
	Set gtmtypes("gtmrecv_local_struct",35,"type")="unsigned-int"
	Set gtmtypfldindx("gtmrecv_local_struct","updresync_lms_group.creator_pid")=35
	Set gtmtypes("gtmrecv_local_struct",36,"name")="gtmrecv_local_struct.updresync_jnl_seqno"
	Set gtmtypes("gtmrecv_local_struct",36,"off")=1240
	Set gtmtypes("gtmrecv_local_struct",36,"len")=8
	Set gtmtypes("gtmrecv_local_struct",36,"type")="uint64_t"
	Set gtmtypfldindx("gtmrecv_local_struct","updresync_jnl_seqno")=36
	Set gtmtypes("gtmrecv_local_struct",37,"name")="gtmrecv_local_struct.remote_lms_group"
	Set gtmtypes("gtmrecv_local_struct",37,"off")=1248
	Set gtmtypes("gtmrecv_local_struct",37,"len")=40
	Set gtmtypes("gtmrecv_local_struct",37,"type")="repl_inst_uuid"
	Set gtmtypfldindx("gtmrecv_local_struct","remote_lms_group")=37
	Set gtmtypes("gtmrecv_local_struct",38,"name")="gtmrecv_local_struct.remote_lms_group.created_nodename"
	Set gtmtypes("gtmrecv_local_struct",38,"off")=1248
	Set gtmtypes("gtmrecv_local_struct",38,"len")=16
	Set gtmtypes("gtmrecv_local_struct",38,"type")="unsigned-char"
	Set gtmtypfldindx("gtmrecv_local_struct","remote_lms_group.created_nodename")=38
	Set gtmtypes("gtmrecv_local_struct",39,"name")="gtmrecv_local_struct.remote_lms_group.this_instname"
	Set gtmtypes("gtmrecv_local_struct",39,"off")=1264
	Set gtmtypes("gtmrecv_local_struct",39,"len")=16
	Set gtmtypes("gtmrecv_local_struct",39,"type")="unsigned-char"
	Set gtmtypfldindx("gtmrecv_local_struct","remote_lms_group.this_instname")=39
	Set gtmtypes("gtmrecv_local_struct",40,"name")="gtmrecv_local_struct.remote_lms_group.created_time"
	Set gtmtypes("gtmrecv_local_struct",40,"off")=1280
	Set gtmtypes("gtmrecv_local_struct",40,"len")=4
	Set gtmtypes("gtmrecv_local_struct",40,"type")="unsigned-int"
	Set gtmtypfldindx("gtmrecv_local_struct","remote_lms_group.created_time")=40
	Set gtmtypes("gtmrecv_local_struct",41,"name")="gtmrecv_local_struct.remote_lms_group.creator_pid"
	Set gtmtypes("gtmrecv_local_struct",41,"off")=1284
	Set gtmtypes("gtmrecv_local_struct",41,"len")=4
	Set gtmtypes("gtmrecv_local_struct",41,"type")="unsigned-int"
	Set gtmtypfldindx("gtmrecv_local_struct","remote_lms_group.creator_pid")=41
	;
	Set gtmtypes("gtmrecv_options_t")="struct"
	Set gtmtypes("gtmrecv_options_t",0)=30
	Set gtmtypes("gtmrecv_options_t","len")=1152
	Set gtmtypes("gtmrecv_options_t",1,"name")="gtmrecv_options_t.start"
	Set gtmtypes("gtmrecv_options_t",1,"off")=0
	Set gtmtypes("gtmrecv_options_t",1,"len")=4
	Set gtmtypes("gtmrecv_options_t",1,"type")="boolean_t"
	Set gtmtypfldindx("gtmrecv_options_t","start")=1
	Set gtmtypes("gtmrecv_options_t",2,"name")="gtmrecv_options_t.shut_down"
	Set gtmtypes("gtmrecv_options_t",2,"off")=4
	Set gtmtypes("gtmrecv_options_t",2,"len")=4
	Set gtmtypes("gtmrecv_options_t",2,"type")="boolean_t"
	Set gtmtypfldindx("gtmrecv_options_t","shut_down")=2
	Set gtmtypes("gtmrecv_options_t",3,"name")="gtmrecv_options_t.checkhealth"
	Set gtmtypes("gtmrecv_options_t",3,"off")=8
	Set gtmtypes("gtmrecv_options_t",3,"len")=4
	Set gtmtypes("gtmrecv_options_t",3,"type")="boolean_t"
	Set gtmtypfldindx("gtmrecv_options_t","checkhealth")=3
	Set gtmtypes("gtmrecv_options_t",4,"name")="gtmrecv_options_t.statslog"
	Set gtmtypes("gtmrecv_options_t",4,"off")=12
	Set gtmtypes("gtmrecv_options_t",4,"len")=4
	Set gtmtypes("gtmrecv_options_t",4,"type")="boolean_t"
	Set gtmtypfldindx("gtmrecv_options_t","statslog")=4
	Set gtmtypes("gtmrecv_options_t",5,"name")="gtmrecv_options_t.showbacklog"
	Set gtmtypes("gtmrecv_options_t",5,"off")=16
	Set gtmtypes("gtmrecv_options_t",5,"len")=4
	Set gtmtypes("gtmrecv_options_t",5,"type")="boolean_t"
	Set gtmtypfldindx("gtmrecv_options_t","showbacklog")=5
	Set gtmtypes("gtmrecv_options_t",6,"name")="gtmrecv_options_t.updateonly"
	Set gtmtypes("gtmrecv_options_t",6,"off")=20
	Set gtmtypes("gtmrecv_options_t",6,"len")=4
	Set gtmtypes("gtmrecv_options_t",6,"type")="boolean_t"
	Set gtmtypfldindx("gtmrecv_options_t","updateonly")=6
	Set gtmtypes("gtmrecv_options_t",7,"name")="gtmrecv_options_t.stopsourcefilter"
	Set gtmtypes("gtmrecv_options_t",7,"off")=24
	Set gtmtypes("gtmrecv_options_t",7,"len")=4
	Set gtmtypes("gtmrecv_options_t",7,"type")="boolean_t"
	Set gtmtypfldindx("gtmrecv_options_t","stopsourcefilter")=7
	Set gtmtypes("gtmrecv_options_t",8,"name")="gtmrecv_options_t.changelog"
	Set gtmtypes("gtmrecv_options_t",8,"off")=28
	Set gtmtypes("gtmrecv_options_t",8,"len")=4
	Set gtmtypes("gtmrecv_options_t",8,"type")="boolean_t"
	Set gtmtypfldindx("gtmrecv_options_t","changelog")=8
	Set gtmtypes("gtmrecv_options_t",9,"name")="gtmrecv_options_t.buffsize"
	Set gtmtypes("gtmrecv_options_t",9,"off")=32
	Set gtmtypes("gtmrecv_options_t",9,"len")=8
	Set gtmtypes("gtmrecv_options_t",9,"type")="uint64_t"
	Set gtmtypfldindx("gtmrecv_options_t","buffsize")=9
	Set gtmtypes("gtmrecv_options_t",10,"name")="gtmrecv_options_t.shutdown_time"
	Set gtmtypes("gtmrecv_options_t",10,"off")=40
	Set gtmtypes("gtmrecv_options_t",10,"len")=4
	Set gtmtypes("gtmrecv_options_t",10,"type")="int"
	Set gtmtypfldindx("gtmrecv_options_t","shutdown_time")=10
	Set gtmtypes("gtmrecv_options_t",11,"name")="gtmrecv_options_t.listen_port"
	Set gtmtypes("gtmrecv_options_t",11,"off")=44
	Set gtmtypes("gtmrecv_options_t",11,"len")=4
	Set gtmtypes("gtmrecv_options_t",11,"type")="int"
	Set gtmtypfldindx("gtmrecv_options_t","listen_port")=11
	Set gtmtypes("gtmrecv_options_t",12,"name")="gtmrecv_options_t.updateresync"
	Set gtmtypes("gtmrecv_options_t",12,"off")=48
	Set gtmtypes("gtmrecv_options_t",12,"len")=4
	Set gtmtypes("gtmrecv_options_t",12,"type")="boolean_t"
	Set gtmtypfldindx("gtmrecv_options_t","updateresync")=12
	Set gtmtypes("gtmrecv_options_t",13,"name")="gtmrecv_options_t.noresync"
	Set gtmtypes("gtmrecv_options_t",13,"off")=52
	Set gtmtypes("gtmrecv_options_t",13,"len")=4
	Set gtmtypes("gtmrecv_options_t",13,"type")="boolean_t"
	Set gtmtypfldindx("gtmrecv_options_t","noresync")=13
	Set gtmtypes("gtmrecv_options_t",14,"name")="gtmrecv_options_t.rcvr_log_interval"
	Set gtmtypes("gtmrecv_options_t",14,"off")=56
	Set gtmtypes("gtmrecv_options_t",14,"len")=4
	Set gtmtypes("gtmrecv_options_t",14,"type")="unsigned-int"
	Set gtmtypfldindx("gtmrecv_options_t","rcvr_log_interval")=14
	Set gtmtypes("gtmrecv_options_t",15,"name")="gtmrecv_options_t.upd_log_interval"
	Set gtmtypes("gtmrecv_options_t",15,"off")=60
	Set gtmtypes("gtmrecv_options_t",15,"len")=4
	Set gtmtypes("gtmrecv_options_t",15,"type")="unsigned-int"
	Set gtmtypfldindx("gtmrecv_options_t","upd_log_interval")=15
	Set gtmtypes("gtmrecv_options_t",16,"name")="gtmrecv_options_t.helpers"
	Set gtmtypes("gtmrecv_options_t",16,"off")=64
	Set gtmtypes("gtmrecv_options_t",16,"len")=4
	Set gtmtypes("gtmrecv_options_t",16,"type")="boolean_t"
	Set gtmtypfldindx("gtmrecv_options_t","helpers")=16
	Set gtmtypes("gtmrecv_options_t",17,"name")="gtmrecv_options_t.reuse_specified"
	Set gtmtypes("gtmrecv_options_t",17,"off")=68
	Set gtmtypes("gtmrecv_options_t",17,"len")=4
	Set gtmtypes("gtmrecv_options_t",17,"type")="boolean_t"
	Set gtmtypfldindx("gtmrecv_options_t","reuse_specified")=17
	Set gtmtypes("gtmrecv_options_t",18,"name")="gtmrecv_options_t.resume_specified"
	Set gtmtypes("gtmrecv_options_t",18,"off")=72
	Set gtmtypes("gtmrecv_options_t",18,"len")=4
	Set gtmtypes("gtmrecv_options_t",18,"type")="boolean_t"
	Set gtmtypfldindx("gtmrecv_options_t","resume_specified")=18
	Set gtmtypes("gtmrecv_options_t",19,"name")="gtmrecv_options_t.initialize_specified"
	Set gtmtypes("gtmrecv_options_t",19,"off")=76
	Set gtmtypes("gtmrecv_options_t",19,"len")=4
	Set gtmtypes("gtmrecv_options_t",19,"type")="boolean_t"
	Set gtmtypfldindx("gtmrecv_options_t","initialize_specified")=19
	Set gtmtypes("gtmrecv_options_t",20,"name")="gtmrecv_options_t.resume_strm_num"
	Set gtmtypes("gtmrecv_options_t",20,"off")=80
	Set gtmtypes("gtmrecv_options_t",20,"len")=4
	Set gtmtypes("gtmrecv_options_t",20,"type")="int"
	Set gtmtypfldindx("gtmrecv_options_t","resume_strm_num")=20
	Set gtmtypes("gtmrecv_options_t",21,"name")="gtmrecv_options_t.n_readers"
	Set gtmtypes("gtmrecv_options_t",21,"off")=84
	Set gtmtypes("gtmrecv_options_t",21,"len")=4
	Set gtmtypes("gtmrecv_options_t",21,"type")="int"
	Set gtmtypfldindx("gtmrecv_options_t","n_readers")=21
	Set gtmtypes("gtmrecv_options_t",22,"name")="gtmrecv_options_t.n_writers"
	Set gtmtypes("gtmrecv_options_t",22,"off")=88
	Set gtmtypes("gtmrecv_options_t",22,"len")=4
	Set gtmtypes("gtmrecv_options_t",22,"type")="int"
	Set gtmtypfldindx("gtmrecv_options_t","n_writers")=22
	Set gtmtypes("gtmrecv_options_t",23,"name")="gtmrecv_options_t.cmplvl"
	Set gtmtypes("gtmrecv_options_t",23,"off")=92
	Set gtmtypes("gtmrecv_options_t",23,"len")=4
	Set gtmtypes("gtmrecv_options_t",23,"type")="int"
	Set gtmtypfldindx("gtmrecv_options_t","cmplvl")=23
	Set gtmtypes("gtmrecv_options_t",24,"name")="gtmrecv_options_t.log_file"
	Set gtmtypes("gtmrecv_options_t",24,"off")=96
	Set gtmtypes("gtmrecv_options_t",24,"len")=256
	Set gtmtypes("gtmrecv_options_t",24,"type")="char"
	Set gtmtypfldindx("gtmrecv_options_t","log_file")=24
	Set gtmtypes("gtmrecv_options_t",25,"name")="gtmrecv_options_t.updresync_instfilename"
	Set gtmtypes("gtmrecv_options_t",25,"off")=352
	Set gtmtypes("gtmrecv_options_t",25,"len")=256
	Set gtmtypes("gtmrecv_options_t",25,"type")="char"
	Set gtmtypfldindx("gtmrecv_options_t","updresync_instfilename")=25
	Set gtmtypes("gtmrecv_options_t",26,"name")="gtmrecv_options_t.filter_cmd"
	Set gtmtypes("gtmrecv_options_t",26,"off")=608
	Set gtmtypes("gtmrecv_options_t",26,"len")=512
	Set gtmtypes("gtmrecv_options_t",26,"type")="char"
	Set gtmtypfldindx("gtmrecv_options_t","filter_cmd")=26
	Set gtmtypes("gtmrecv_options_t",27,"name")="gtmrecv_options_t.reuse_instname"
	Set gtmtypes("gtmrecv_options_t",27,"off")=1120
	Set gtmtypes("gtmrecv_options_t",27,"len")=16
	Set gtmtypes("gtmrecv_options_t",27,"type")="char"
	Set gtmtypfldindx("gtmrecv_options_t","reuse_instname")=27
	Set gtmtypes("gtmrecv_options_t",28,"name")="gtmrecv_options_t.autorollback"
	Set gtmtypes("gtmrecv_options_t",28,"off")=1136
	Set gtmtypes("gtmrecv_options_t",28,"len")=4
	Set gtmtypes("gtmrecv_options_t",28,"type")="boolean_t"
	Set gtmtypfldindx("gtmrecv_options_t","autorollback")=28
	Set gtmtypes("gtmrecv_options_t",29,"name")="gtmrecv_options_t.autorollback_verbose"
	Set gtmtypes("gtmrecv_options_t",29,"off")=1140
	Set gtmtypes("gtmrecv_options_t",29,"len")=4
	Set gtmtypes("gtmrecv_options_t",29,"type")="boolean_t"
	Set gtmtypfldindx("gtmrecv_options_t","autorollback_verbose")=29
	Set gtmtypes("gtmrecv_options_t",30,"name")="gtmrecv_options_t.stopreceiverfilter"
	Set gtmtypes("gtmrecv_options_t",30,"off")=1144
	Set gtmtypes("gtmrecv_options_t",30,"len")=4
	Set gtmtypes("gtmrecv_options_t",30,"type")="boolean_t"
	Set gtmtypfldindx("gtmrecv_options_t","stopreceiverfilter")=30
	;
	Set gtmtypes("gtmsecshr_mesg")="struct"
	Set gtmtypes("gtmsecshr_mesg",0)=16
	Set gtmtypes("gtmsecshr_mesg","len")=4160
	Set gtmtypes("gtmsecshr_mesg",1,"name")="gtmsecshr_mesg.code"
	Set gtmtypes("gtmsecshr_mesg",1,"off")=0
	Set gtmtypes("gtmsecshr_mesg",1,"len")=4
	Set gtmtypes("gtmsecshr_mesg",1,"type")="int"
	Set gtmtypfldindx("gtmsecshr_mesg","code")=1
	Set gtmtypes("gtmsecshr_mesg",2,"name")="gtmsecshr_mesg.comkey"
	Set gtmtypes("gtmsecshr_mesg",2,"off")=4
	Set gtmtypes("gtmsecshr_mesg",2,"len")=4
	Set gtmtypes("gtmsecshr_mesg",2,"type")="unsigned-int"
	Set gtmtypfldindx("gtmsecshr_mesg","comkey")=2
	Set gtmtypes("gtmsecshr_mesg",3,"name")="gtmsecshr_mesg.usesecshr"
	Set gtmtypes("gtmsecshr_mesg",3,"off")=8
	Set gtmtypes("gtmsecshr_mesg",3,"len")=4
	Set gtmtypes("gtmsecshr_mesg",3,"type")="boolean_t"
	Set gtmtypfldindx("gtmsecshr_mesg","usesecshr")=3
	Set gtmtypes("gtmsecshr_mesg",4,"name")="gtmsecshr_mesg.pid"
	Set gtmtypes("gtmsecshr_mesg",4,"off")=12
	Set gtmtypes("gtmsecshr_mesg",4,"len")=4
	Set gtmtypes("gtmsecshr_mesg",4,"type")="pid_t"
	Set gtmtypfldindx("gtmsecshr_mesg","pid")=4
	Set gtmtypes("gtmsecshr_mesg",5,"name")="gtmsecshr_mesg.seqno"
	Set gtmtypes("gtmsecshr_mesg",5,"off")=16
	Set gtmtypes("gtmsecshr_mesg",5,"len")=8
	Set gtmtypes("gtmsecshr_mesg",5,"type")="unsigned-long"
	Set gtmtypfldindx("gtmsecshr_mesg","seqno")=5
	Set gtmtypes("gtmsecshr_mesg",6,"name")="gtmsecshr_mesg.mesg"
	Set gtmtypes("gtmsecshr_mesg",6,"off")=24
	Set gtmtypes("gtmsecshr_mesg",6,"len")=4136
	Set gtmtypes("gtmsecshr_mesg",6,"type")="union"
	Set gtmtypfldindx("gtmsecshr_mesg","mesg")=6
	Set gtmtypes("gtmsecshr_mesg",7,"name")="gtmsecshr_mesg.mesg.id"
	Set gtmtypes("gtmsecshr_mesg",7,"off")=24
	Set gtmtypes("gtmsecshr_mesg",7,"len")=4
	Set gtmtypes("gtmsecshr_mesg",7,"type")="int"
	Set gtmtypfldindx("gtmsecshr_mesg","mesg.id")=7
	Set gtmtypes("gtmsecshr_mesg",8,"name")="gtmsecshr_mesg.mesg.path"
	Set gtmtypes("gtmsecshr_mesg",8,"off")=24
	Set gtmtypes("gtmsecshr_mesg",8,"len")=4097
	Set gtmtypes("gtmsecshr_mesg",8,"type")="char"
	Set gtmtypfldindx("gtmsecshr_mesg","mesg.path")=8
	Set gtmtypes("gtmsecshr_mesg",9,"name")="gtmsecshr_mesg.mesg.db_ipcs"
	Set gtmtypes("gtmsecshr_mesg",9,"off")=24
	Set gtmtypes("gtmsecshr_mesg",9,"len")=4136
	Set gtmtypes("gtmsecshr_mesg",9,"type")="ipcs_mesg"
	Set gtmtypfldindx("gtmsecshr_mesg","mesg.db_ipcs")=9
	Set gtmtypes("gtmsecshr_mesg",10,"name")="gtmsecshr_mesg.mesg.db_ipcs.open_fd_with_o_direct"
	Set gtmtypes("gtmsecshr_mesg",10,"off")=24
	Set gtmtypes("gtmsecshr_mesg",10,"len")=4
	Set gtmtypes("gtmsecshr_mesg",10,"type")="boolean_t"
	Set gtmtypfldindx("gtmsecshr_mesg","mesg.db_ipcs.open_fd_with_o_direct")=10
	Set gtmtypes("gtmsecshr_mesg",11,"name")="gtmsecshr_mesg.mesg.db_ipcs.semid"
	Set gtmtypes("gtmsecshr_mesg",11,"off")=28
	Set gtmtypes("gtmsecshr_mesg",11,"len")=4
	Set gtmtypes("gtmsecshr_mesg",11,"type")="int"
	Set gtmtypfldindx("gtmsecshr_mesg","mesg.db_ipcs.semid")=11
	Set gtmtypes("gtmsecshr_mesg",12,"name")="gtmsecshr_mesg.mesg.db_ipcs.shmid"
	Set gtmtypes("gtmsecshr_mesg",12,"off")=32
	Set gtmtypes("gtmsecshr_mesg",12,"len")=4
	Set gtmtypes("gtmsecshr_mesg",12,"type")="int"
	Set gtmtypfldindx("gtmsecshr_mesg","mesg.db_ipcs.shmid")=12
	Set gtmtypes("gtmsecshr_mesg",13,"name")="gtmsecshr_mesg.mesg.db_ipcs.gt_sem_ctime"
	Set gtmtypes("gtmsecshr_mesg",13,"off")=40
	Set gtmtypes("gtmsecshr_mesg",13,"len")=8
	Set gtmtypes("gtmsecshr_mesg",13,"type")="time_t"
	Set gtmtypfldindx("gtmsecshr_mesg","mesg.db_ipcs.gt_sem_ctime")=13
	Set gtmtypes("gtmsecshr_mesg",14,"name")="gtmsecshr_mesg.mesg.db_ipcs.gt_shm_ctime"
	Set gtmtypes("gtmsecshr_mesg",14,"off")=48
	Set gtmtypes("gtmsecshr_mesg",14,"len")=8
	Set gtmtypes("gtmsecshr_mesg",14,"type")="time_t"
	Set gtmtypfldindx("gtmsecshr_mesg","mesg.db_ipcs.gt_shm_ctime")=14
	Set gtmtypes("gtmsecshr_mesg",15,"name")="gtmsecshr_mesg.mesg.db_ipcs.fn_len"
	Set gtmtypes("gtmsecshr_mesg",15,"off")=56
	Set gtmtypes("gtmsecshr_mesg",15,"len")=4
	Set gtmtypes("gtmsecshr_mesg",15,"type")="unsigned-int"
	Set gtmtypfldindx("gtmsecshr_mesg","mesg.db_ipcs.fn_len")=15
	Set gtmtypes("gtmsecshr_mesg",16,"name")="gtmsecshr_mesg.mesg.db_ipcs.fn"
	Set gtmtypes("gtmsecshr_mesg",16,"off")=60
	Set gtmtypes("gtmsecshr_mesg",16,"len")=4097
	Set gtmtypes("gtmsecshr_mesg",16,"type")="char"
	Set gtmtypfldindx("gtmsecshr_mesg","mesg.db_ipcs.fn")=16
	;
	Set gtmtypes("gtmsiginfo_t")="struct"
	Set gtmtypes("gtmsiginfo_t",0)=8
	Set gtmtypes("gtmsiginfo_t","len")=40
	Set gtmtypes("gtmsiginfo_t",1,"name")="gtmsiginfo_t.int_iadr"
	Set gtmtypes("gtmsiginfo_t",1,"off")=0
	Set gtmtypes("gtmsiginfo_t",1,"len")=8
	Set gtmtypes("gtmsiginfo_t",1,"type")="caddr_t"
	Set gtmtypfldindx("gtmsiginfo_t","int_iadr")=1
	Set gtmtypes("gtmsiginfo_t",2,"name")="gtmsiginfo_t.bad_vadr"
	Set gtmtypes("gtmsiginfo_t",2,"off")=8
	Set gtmtypes("gtmsiginfo_t",2,"len")=8
	Set gtmtypes("gtmsiginfo_t",2,"type")="caddr_t"
	Set gtmtypfldindx("gtmsiginfo_t","bad_vadr")=2
	Set gtmtypes("gtmsiginfo_t",3,"name")="gtmsiginfo_t.sig_err"
	Set gtmtypes("gtmsiginfo_t",3,"off")=16
	Set gtmtypes("gtmsiginfo_t",3,"len")=4
	Set gtmtypes("gtmsiginfo_t",3,"type")="int"
	Set gtmtypfldindx("gtmsiginfo_t","sig_err")=3
	Set gtmtypes("gtmsiginfo_t",4,"name")="gtmsiginfo_t.send_pid"
	Set gtmtypes("gtmsiginfo_t",4,"off")=20
	Set gtmtypes("gtmsiginfo_t",4,"len")=4
	Set gtmtypes("gtmsiginfo_t",4,"type")="pid_t"
	Set gtmtypfldindx("gtmsiginfo_t","send_pid")=4
	Set gtmtypes("gtmsiginfo_t",5,"name")="gtmsiginfo_t.send_uid"
	Set gtmtypes("gtmsiginfo_t",5,"off")=24
	Set gtmtypes("gtmsiginfo_t",5,"len")=4
	Set gtmtypes("gtmsiginfo_t",5,"type")="uid_t"
	Set gtmtypfldindx("gtmsiginfo_t","send_uid")=5
	Set gtmtypes("gtmsiginfo_t",6,"name")="gtmsiginfo_t.subcode"
	Set gtmtypes("gtmsiginfo_t",6,"off")=28
	Set gtmtypes("gtmsiginfo_t",6,"len")=4
	Set gtmtypes("gtmsiginfo_t",6,"type")="int"
	Set gtmtypfldindx("gtmsiginfo_t","subcode")=6
	Set gtmtypes("gtmsiginfo_t",7,"name")="gtmsiginfo_t.infotype"
	Set gtmtypes("gtmsiginfo_t",7,"off")=32
	Set gtmtypes("gtmsiginfo_t",7,"len")=4
	Set gtmtypes("gtmsiginfo_t",7,"type")="int"
	Set gtmtypfldindx("gtmsiginfo_t","infotype")=7
	Set gtmtypes("gtmsiginfo_t",8,"name")="gtmsiginfo_t.signal"
	Set gtmtypes("gtmsiginfo_t",8,"off")=36
	Set gtmtypes("gtmsiginfo_t",8,"len")=4
	Set gtmtypes("gtmsiginfo_t",8,"type")="int"
	Set gtmtypfldindx("gtmsiginfo_t","signal")=8
	;
	Set gtmtypes("gtmsource_local_struct")="struct"
	Set gtmtypes("gtmsource_local_struct",0)=54
	Set gtmtypes("gtmsource_local_struct","len")=1552
	Set gtmtypes("gtmsource_local_struct",1,"name")="gtmsource_local_struct.secondary_instname"
	Set gtmtypes("gtmsource_local_struct",1,"off")=0
	Set gtmtypes("gtmsource_local_struct",1,"len")=16
	Set gtmtypes("gtmsource_local_struct",1,"type")="unsigned-char"
	Set gtmtypfldindx("gtmsource_local_struct","secondary_instname")=1
	Set gtmtypes("gtmsource_local_struct",2,"name")="gtmsource_local_struct.gtmsource_pid"
	Set gtmtypes("gtmsource_local_struct",2,"off")=16
	Set gtmtypes("gtmsource_local_struct",2,"len")=4
	Set gtmtypes("gtmsource_local_struct",2,"type")="unsigned-int"
	Set gtmtypfldindx("gtmsource_local_struct","gtmsource_pid")=2
	Set gtmtypes("gtmsource_local_struct",3,"name")="gtmsource_local_struct.mode"
	Set gtmtypes("gtmsource_local_struct",3,"off")=20
	Set gtmtypes("gtmsource_local_struct",3,"len")=4
	Set gtmtypes("gtmsource_local_struct",3,"type")="int"
	Set gtmtypfldindx("gtmsource_local_struct","mode")=3
	Set gtmtypes("gtmsource_local_struct",4,"name")="gtmsource_local_struct.gtmsource_state"
	Set gtmtypes("gtmsource_local_struct",4,"off")=24
	Set gtmtypes("gtmsource_local_struct",4,"len")=4
	Set gtmtypes("gtmsource_local_struct",4,"type")="int"
	Set gtmtypfldindx("gtmsource_local_struct","gtmsource_state")=4
	Set gtmtypes("gtmsource_local_struct",5,"name")="gtmsource_local_struct.gtmsrc_lcl_array_index"
	Set gtmtypes("gtmsource_local_struct",5,"off")=28
	Set gtmtypes("gtmsource_local_struct",5,"len")=4
	Set gtmtypes("gtmsource_local_struct",5,"type")="int"
	Set gtmtypfldindx("gtmsource_local_struct","gtmsrc_lcl_array_index")=5
	Set gtmtypes("gtmsource_local_struct",6,"name")="gtmsource_local_struct.repl_zlib_cmp_level"
	Set gtmtypes("gtmsource_local_struct",6,"off")=32
	Set gtmtypes("gtmsource_local_struct",6,"len")=4
	Set gtmtypes("gtmsource_local_struct",6,"type")="int"
	Set gtmtypfldindx("gtmsource_local_struct","repl_zlib_cmp_level")=6
	Set gtmtypes("gtmsource_local_struct",7,"name")="gtmsource_local_struct.filler1_align_8"
	Set gtmtypes("gtmsource_local_struct",7,"off")=36
	Set gtmtypes("gtmsource_local_struct",7,"len")=4
	Set gtmtypes("gtmsource_local_struct",7,"type")="unsigned-char"
	Set gtmtypfldindx("gtmsource_local_struct","filler1_align_8")=7
	Set gtmtypes("gtmsource_local_struct",8,"name")="gtmsource_local_struct.read_state"
	Set gtmtypes("gtmsource_local_struct",8,"off")=40
	Set gtmtypes("gtmsource_local_struct",8,"len")=4
	Set gtmtypes("gtmsource_local_struct",8,"type")="int"
	Set gtmtypfldindx("gtmsource_local_struct","read_state")=8
	Set gtmtypes("gtmsource_local_struct",9,"name")="gtmsource_local_struct.read"
	Set gtmtypes("gtmsource_local_struct",9,"off")=48
	Set gtmtypes("gtmsource_local_struct",9,"len")=8
	Set gtmtypes("gtmsource_local_struct",9,"type")="uint64_t"
	Set gtmtypfldindx("gtmsource_local_struct","read")=9
	Set gtmtypes("gtmsource_local_struct",10,"name")="gtmsource_local_struct.remote_side"
	Set gtmtypes("gtmsource_local_struct",10,"off")=56
	Set gtmtypes("gtmsource_local_struct",10,"len")=48
	Set gtmtypes("gtmsource_local_struct",10,"type")="repl_conn_info_t"
	Set gtmtypfldindx("gtmsource_local_struct","remote_side")=10
	Set gtmtypes("gtmsource_local_struct",11,"name")="gtmsource_local_struct.remote_side.proto_ver"
	Set gtmtypes("gtmsource_local_struct",11,"off")=56
	Set gtmtypes("gtmsource_local_struct",11,"len")=4
	Set gtmtypes("gtmsource_local_struct",11,"type")="int"
	Set gtmtypfldindx("gtmsource_local_struct","remote_side.proto_ver")=11
	Set gtmtypes("gtmsource_local_struct",12,"name")="gtmsource_local_struct.remote_side.jnl_ver"
	Set gtmtypes("gtmsource_local_struct",12,"off")=60
	Set gtmtypes("gtmsource_local_struct",12,"len")=4
	Set gtmtypes("gtmsource_local_struct",12,"type")="unsigned-int"
	Set gtmtypfldindx("gtmsource_local_struct","remote_side.jnl_ver")=12
	Set gtmtypes("gtmsource_local_struct",13,"name")="gtmsource_local_struct.remote_side.is_std_null_coll"
	Set gtmtypes("gtmsource_local_struct",13,"off")=64
	Set gtmtypes("gtmsource_local_struct",13,"len")=4
	Set gtmtypes("gtmsource_local_struct",13,"type")="boolean_t"
	Set gtmtypfldindx("gtmsource_local_struct","remote_side.is_std_null_coll")=13
	Set gtmtypes("gtmsource_local_struct",14,"name")="gtmsource_local_struct.remote_side.trigger_supported"
	Set gtmtypes("gtmsource_local_struct",14,"off")=68
	Set gtmtypes("gtmsource_local_struct",14,"len")=4
	Set gtmtypes("gtmsource_local_struct",14,"type")="boolean_t"
	Set gtmtypfldindx("gtmsource_local_struct","remote_side.trigger_supported")=14
	Set gtmtypes("gtmsource_local_struct",15,"name")="gtmsource_local_struct.remote_side.cross_endian"
	Set gtmtypes("gtmsource_local_struct",15,"off")=72
	Set gtmtypes("gtmsource_local_struct",15,"len")=4
	Set gtmtypes("gtmsource_local_struct",15,"type")="boolean_t"
	Set gtmtypfldindx("gtmsource_local_struct","remote_side.cross_endian")=15
	Set gtmtypes("gtmsource_local_struct",16,"name")="gtmsource_local_struct.remote_side.endianness_known"
	Set gtmtypes("gtmsource_local_struct",16,"off")=76
	Set gtmtypes("gtmsource_local_struct",16,"len")=4
	Set gtmtypes("gtmsource_local_struct",16,"type")="boolean_t"
	Set gtmtypfldindx("gtmsource_local_struct","remote_side.endianness_known")=16
	Set gtmtypes("gtmsource_local_struct",17,"name")="gtmsource_local_struct.remote_side.null_subs_xform"
	Set gtmtypes("gtmsource_local_struct",17,"off")=80
	Set gtmtypes("gtmsource_local_struct",17,"len")=4
	Set gtmtypes("gtmsource_local_struct",17,"type")="boolean_t"
	Set gtmtypfldindx("gtmsource_local_struct","remote_side.null_subs_xform")=17
	Set gtmtypes("gtmsource_local_struct",18,"name")="gtmsource_local_struct.remote_side.is_supplementary"
	Set gtmtypes("gtmsource_local_struct",18,"off")=84
	Set gtmtypes("gtmsource_local_struct",18,"len")=4
	Set gtmtypes("gtmsource_local_struct",18,"type")="boolean_t"
	Set gtmtypfldindx("gtmsource_local_struct","remote_side.is_supplementary")=18
	Set gtmtypes("gtmsource_local_struct",19,"name")="gtmsource_local_struct.remote_side.tls_requested"
	Set gtmtypes("gtmsource_local_struct",19,"off")=88
	Set gtmtypes("gtmsource_local_struct",19,"len")=4
	Set gtmtypes("gtmsource_local_struct",19,"type")="boolean_t"
	Set gtmtypfldindx("gtmsource_local_struct","remote_side.tls_requested")=19
	Set gtmtypes("gtmsource_local_struct",20,"name")="gtmsource_local_struct.remote_side.filler_16"
	Set gtmtypes("gtmsource_local_struct",20,"off")=92
	Set gtmtypes("gtmsource_local_struct",20,"len")=12
	Set gtmtypes("gtmsource_local_struct",20,"type")="char"
	Set gtmtypfldindx("gtmsource_local_struct","remote_side.filler_16")=20
	Set gtmtypes("gtmsource_local_struct",21,"name")="gtmsource_local_struct.read_addr"
	Set gtmtypes("gtmsource_local_struct",21,"off")=104
	Set gtmtypes("gtmsource_local_struct",21,"len")=8
	Set gtmtypes("gtmsource_local_struct",21,"type")="uint64_t"
	Set gtmtypfldindx("gtmsource_local_struct","read_addr")=21
	Set gtmtypes("gtmsource_local_struct",22,"name")="gtmsource_local_struct.read_jnl_seqno"
	Set gtmtypes("gtmsource_local_struct",22,"off")=112
	Set gtmtypes("gtmsource_local_struct",22,"len")=8
	Set gtmtypes("gtmsource_local_struct",22,"type")="uint64_t"
	Set gtmtypfldindx("gtmsource_local_struct","read_jnl_seqno")=22
	Set gtmtypes("gtmsource_local_struct",23,"name")="gtmsource_local_struct.connect_jnl_seqno"
	Set gtmtypes("gtmsource_local_struct",23,"off")=120
	Set gtmtypes("gtmsource_local_struct",23,"len")=8
	Set gtmtypes("gtmsource_local_struct",23,"type")="uint64_t"
	Set gtmtypfldindx("gtmsource_local_struct","connect_jnl_seqno")=23
	Set gtmtypes("gtmsource_local_struct",24,"name")="gtmsource_local_struct.num_histinfo"
	Set gtmtypes("gtmsource_local_struct",24,"off")=128
	Set gtmtypes("gtmsource_local_struct",24,"len")=4
	Set gtmtypes("gtmsource_local_struct",24,"type")="int"
	Set gtmtypfldindx("gtmsource_local_struct","num_histinfo")=24
	Set gtmtypes("gtmsource_local_struct",25,"name")="gtmsource_local_struct.next_histinfo_num"
	Set gtmtypes("gtmsource_local_struct",25,"off")=132
	Set gtmtypes("gtmsource_local_struct",25,"len")=4
	Set gtmtypes("gtmsource_local_struct",25,"type")="int"
	Set gtmtypfldindx("gtmsource_local_struct","next_histinfo_num")=25
	Set gtmtypes("gtmsource_local_struct",26,"name")="gtmsource_local_struct.next_histinfo_seqno"
	Set gtmtypes("gtmsource_local_struct",26,"off")=136
	Set gtmtypes("gtmsource_local_struct",26,"len")=8
	Set gtmtypes("gtmsource_local_struct",26,"type")="uint64_t"
	Set gtmtypfldindx("gtmsource_local_struct","next_histinfo_seqno")=26
	Set gtmtypes("gtmsource_local_struct",27,"name")="gtmsource_local_struct.last_flush_resync_seqno"
	Set gtmtypes("gtmsource_local_struct",27,"off")=144
	Set gtmtypes("gtmsource_local_struct",27,"len")=8
	Set gtmtypes("gtmsource_local_struct",27,"type")="uint64_t"
	Set gtmtypfldindx("gtmsource_local_struct","last_flush_resync_seqno")=27
	Set gtmtypes("gtmsource_local_struct",28,"name")="gtmsource_local_struct.send_new_histrec"
	Set gtmtypes("gtmsource_local_struct",28,"off")=152
	Set gtmtypes("gtmsource_local_struct",28,"len")=4
	Set gtmtypes("gtmsource_local_struct",28,"type")="boolean_t"
	Set gtmtypfldindx("gtmsource_local_struct","send_new_histrec")=28
	Set gtmtypes("gtmsource_local_struct",29,"name")="gtmsource_local_struct.send_losttn_complete"
	Set gtmtypes("gtmsource_local_struct",29,"off")=156
	Set gtmtypes("gtmsource_local_struct",29,"len")=4
	Set gtmtypes("gtmsource_local_struct",29,"type")="boolean_t"
	Set gtmtypfldindx("gtmsource_local_struct","send_losttn_complete")=29
	Set gtmtypes("gtmsource_local_struct",30,"name")="gtmsource_local_struct.secondary_host"
	Set gtmtypes("gtmsource_local_struct",30,"off")=160
	Set gtmtypes("gtmsource_local_struct",30,"len")=256
	Set gtmtypes("gtmsource_local_struct",30,"type")="char"
	Set gtmtypfldindx("gtmsource_local_struct","secondary_host")=30
	Set gtmtypes("gtmsource_local_struct",31,"name")="gtmsource_local_struct.secondary_inet_addr"
	Set gtmtypes("gtmsource_local_struct",31,"off")=416
	Set gtmtypes("gtmsource_local_struct",31,"len")=28
	Set gtmtypes("gtmsource_local_struct",31,"type")="union"
	Set gtmtypfldindx("gtmsource_local_struct","secondary_inet_addr")=31
	Set gtmtypes("gtmsource_local_struct",32,"name")="gtmsource_local_struct.secondary_af"
	Set gtmtypes("gtmsource_local_struct",32,"off")=444
	Set gtmtypes("gtmsource_local_struct",32,"len")=4
	Set gtmtypes("gtmsource_local_struct",32,"type")="int"
	Set gtmtypfldindx("gtmsource_local_struct","secondary_af")=32
	Set gtmtypes("gtmsource_local_struct",33,"name")="gtmsource_local_struct.secondary_addrlen"
	Set gtmtypes("gtmsource_local_struct",33,"off")=448
	Set gtmtypes("gtmsource_local_struct",33,"len")=4
	Set gtmtypes("gtmsource_local_struct",33,"type")="int"
	Set gtmtypfldindx("gtmsource_local_struct","secondary_addrlen")=33
	Set gtmtypes("gtmsource_local_struct",34,"name")="gtmsource_local_struct.secondary_port"
	Set gtmtypes("gtmsource_local_struct",34,"off")=452
	Set gtmtypes("gtmsource_local_struct",34,"len")=4
	Set gtmtypes("gtmsource_local_struct",34,"type")="unsigned-int"
	Set gtmtypfldindx("gtmsource_local_struct","secondary_port")=34
	Set gtmtypes("gtmsource_local_struct",35,"name")="gtmsource_local_struct.child_server_running"
	Set gtmtypes("gtmsource_local_struct",35,"off")=456
	Set gtmtypes("gtmsource_local_struct",35,"len")=4
	Set gtmtypes("gtmsource_local_struct",35,"type")="boolean_t"
	Set gtmtypfldindx("gtmsource_local_struct","child_server_running")=35
	Set gtmtypes("gtmsource_local_struct",36,"name")="gtmsource_local_struct.log_interval"
	Set gtmtypes("gtmsource_local_struct",36,"off")=460
	Set gtmtypes("gtmsource_local_struct",36,"len")=4
	Set gtmtypes("gtmsource_local_struct",36,"type")="unsigned-int"
	Set gtmtypfldindx("gtmsource_local_struct","log_interval")=36
	Set gtmtypes("gtmsource_local_struct",37,"name")="gtmsource_local_struct.log_file"
	Set gtmtypes("gtmsource_local_struct",37,"off")=464
	Set gtmtypes("gtmsource_local_struct",37,"len")=256
	Set gtmtypes("gtmsource_local_struct",37,"type")="char"
	Set gtmtypfldindx("gtmsource_local_struct","log_file")=37
	Set gtmtypes("gtmsource_local_struct",38,"name")="gtmsource_local_struct.changelog"
	Set gtmtypes("gtmsource_local_struct",38,"off")=720
	Set gtmtypes("gtmsource_local_struct",38,"len")=4
	Set gtmtypes("gtmsource_local_struct",38,"type")="unsigned-int"
	Set gtmtypfldindx("gtmsource_local_struct","changelog")=38
	Set gtmtypes("gtmsource_local_struct",39,"name")="gtmsource_local_struct.statslog"
	Set gtmtypes("gtmsource_local_struct",39,"off")=724
	Set gtmtypes("gtmsource_local_struct",39,"len")=4
	Set gtmtypes("gtmsource_local_struct",39,"type")="unsigned-int"
	Set gtmtypfldindx("gtmsource_local_struct","statslog")=39
	Set gtmtypes("gtmsource_local_struct",40,"name")="gtmsource_local_struct.statslog_file"
	Set gtmtypes("gtmsource_local_struct",40,"off")=728
	Set gtmtypes("gtmsource_local_struct",40,"len")=256
	Set gtmtypes("gtmsource_local_struct",40,"type")="char"
	Set gtmtypfldindx("gtmsource_local_struct","statslog_file")=40
	Set gtmtypes("gtmsource_local_struct",41,"name")="gtmsource_local_struct.connect_parms"
	Set gtmtypes("gtmsource_local_struct",41,"off")=984
	Set gtmtypes("gtmsource_local_struct",41,"len")=24
	Set gtmtypes("gtmsource_local_struct",41,"type")="int"
	Set gtmtypfldindx("gtmsource_local_struct","connect_parms")=41
	Set gtmtypes("gtmsource_local_struct",41,"dim")=6
	Set gtmtypes("gtmsource_local_struct",42,"name")="gtmsource_local_struct.shutdown"
	Set gtmtypes("gtmsource_local_struct",42,"off")=1008
	Set gtmtypes("gtmsource_local_struct",42,"len")=4
	Set gtmtypes("gtmsource_local_struct",42,"type")="unsigned-int"
	Set gtmtypfldindx("gtmsource_local_struct","shutdown")=42
	Set gtmtypes("gtmsource_local_struct",43,"name")="gtmsource_local_struct.shutdown_time"
	Set gtmtypes("gtmsource_local_struct",43,"off")=1012
	Set gtmtypes("gtmsource_local_struct",43,"len")=4
	Set gtmtypes("gtmsource_local_struct",43,"type")="int"
	Set gtmtypfldindx("gtmsource_local_struct","shutdown_time")=43
	Set gtmtypes("gtmsource_local_struct",44,"name")="gtmsource_local_struct.filter_cmd"
	Set gtmtypes("gtmsource_local_struct",44,"off")=1016
	Set gtmtypes("gtmsource_local_struct",44,"len")=512
	Set gtmtypes("gtmsource_local_struct",44,"type")="char"
	Set gtmtypfldindx("gtmsource_local_struct","filter_cmd")=44
	Set gtmtypes("gtmsource_local_struct",45,"name")="gtmsource_local_struct.gtmsource_srv_latch"
	Set gtmtypes("gtmsource_local_struct",45,"off")=1528
	Set gtmtypes("gtmsource_local_struct",45,"len")=8
	Set gtmtypes("gtmsource_local_struct",45,"type")="global_latch_t"
	Set gtmtypfldindx("gtmsource_local_struct","gtmsource_srv_latch")=45
	Set gtmtypes("gtmsource_local_struct",46,"name")="gtmsource_local_struct.gtmsource_srv_latch.u"
	Set gtmtypes("gtmsource_local_struct",46,"off")=1528
	Set gtmtypes("gtmsource_local_struct",46,"len")=8
	Set gtmtypes("gtmsource_local_struct",46,"type")="union"
	Set gtmtypfldindx("gtmsource_local_struct","gtmsource_srv_latch.u")=46
	Set gtmtypes("gtmsource_local_struct",47,"name")="gtmsource_local_struct.gtmsource_srv_latch.u.pid_imgcnt"
	Set gtmtypes("gtmsource_local_struct",47,"off")=1528
	Set gtmtypes("gtmsource_local_struct",47,"len")=8
	Set gtmtypes("gtmsource_local_struct",47,"type")="uint64_t"
	Set gtmtypfldindx("gtmsource_local_struct","gtmsource_srv_latch.u.pid_imgcnt")=47
	Set gtmtypes("gtmsource_local_struct",48,"name")="gtmsource_local_struct.gtmsource_srv_latch.u.parts"
	Set gtmtypes("gtmsource_local_struct",48,"off")=1528
	Set gtmtypes("gtmsource_local_struct",48,"len")=8
	Set gtmtypes("gtmsource_local_struct",48,"type")="struct"
	Set gtmtypfldindx("gtmsource_local_struct","gtmsource_srv_latch.u.parts")=48
	Set gtmtypes("gtmsource_local_struct",49,"name")="gtmsource_local_struct.gtmsource_srv_latch.u.parts.latch_pid"
	Set gtmtypes("gtmsource_local_struct",49,"off")=1528
	Set gtmtypes("gtmsource_local_struct",49,"len")=4
	Set gtmtypes("gtmsource_local_struct",49,"type")="int"
	Set gtmtypfldindx("gtmsource_local_struct","gtmsource_srv_latch.u.parts.latch_pid")=49
	Set gtmtypes("gtmsource_local_struct",50,"name")="gtmsource_local_struct.gtmsource_srv_latch.u.parts.latch_word"
	Set gtmtypes("gtmsource_local_struct",50,"off")=1532
	Set gtmtypes("gtmsource_local_struct",50,"len")=4
	Set gtmtypes("gtmsource_local_struct",50,"type")="int"
	Set gtmtypfldindx("gtmsource_local_struct","gtmsource_srv_latch.u.parts.latch_word")=50
	Set gtmtypes("gtmsource_local_struct",51,"name")="gtmsource_local_struct.jnlfileonly"
	Set gtmtypes("gtmsource_local_struct",51,"off")=1536
	Set gtmtypes("gtmsource_local_struct",51,"len")=4
	Set gtmtypes("gtmsource_local_struct",51,"type")="boolean_t"
	Set gtmtypfldindx("gtmsource_local_struct","jnlfileonly")=51
	Set gtmtypes("gtmsource_local_struct",52,"name")="gtmsource_local_struct.next_renegotiate_time"
	Set gtmtypes("gtmsource_local_struct",52,"off")=1540
	Set gtmtypes("gtmsource_local_struct",52,"len")=4
	Set gtmtypes("gtmsource_local_struct",52,"type")="unsigned-int"
	Set gtmtypfldindx("gtmsource_local_struct","next_renegotiate_time")=52
	Set gtmtypes("gtmsource_local_struct",53,"name")="gtmsource_local_struct.num_renegotiations"
	Set gtmtypes("gtmsource_local_struct",53,"off")=1544
	Set gtmtypes("gtmsource_local_struct",53,"len")=4
	Set gtmtypes("gtmsource_local_struct",53,"type")="int"
	Set gtmtypfldindx("gtmsource_local_struct","num_renegotiations")=53
	Set gtmtypes("gtmsource_local_struct",54,"name")="gtmsource_local_struct.filler_8byte_align"
	Set gtmtypes("gtmsource_local_struct",54,"off")=1548
	Set gtmtypes("gtmsource_local_struct",54,"len")=4
	Set gtmtypes("gtmsource_local_struct",54,"type")="int"
	Set gtmtypfldindx("gtmsource_local_struct","filler_8byte_align")=54
	;
	Set gtmtypes("gtmsource_options_t")="struct"
	Set gtmtypes("gtmsource_options_t",0)=34
	Set gtmtypes("gtmsource_options_t","len")=2232
	Set gtmtypes("gtmsource_options_t",1,"name")="gtmsource_options_t.start"
	Set gtmtypes("gtmsource_options_t",1,"off")=0
	Set gtmtypes("gtmsource_options_t",1,"len")=4
	Set gtmtypes("gtmsource_options_t",1,"type")="boolean_t"
	Set gtmtypfldindx("gtmsource_options_t","start")=1
	Set gtmtypes("gtmsource_options_t",2,"name")="gtmsource_options_t.shut_down"
	Set gtmtypes("gtmsource_options_t",2,"off")=4
	Set gtmtypes("gtmsource_options_t",2,"len")=4
	Set gtmtypes("gtmsource_options_t",2,"type")="boolean_t"
	Set gtmtypfldindx("gtmsource_options_t","shut_down")=2
	Set gtmtypes("gtmsource_options_t",3,"name")="gtmsource_options_t.activate"
	Set gtmtypes("gtmsource_options_t",3,"off")=8
	Set gtmtypes("gtmsource_options_t",3,"len")=4
	Set gtmtypes("gtmsource_options_t",3,"type")="boolean_t"
	Set gtmtypfldindx("gtmsource_options_t","activate")=3
	Set gtmtypes("gtmsource_options_t",4,"name")="gtmsource_options_t.changelog"
	Set gtmtypes("gtmsource_options_t",4,"off")=12
	Set gtmtypes("gtmsource_options_t",4,"len")=4
	Set gtmtypes("gtmsource_options_t",4,"type")="boolean_t"
	Set gtmtypfldindx("gtmsource_options_t","changelog")=4
	Set gtmtypes("gtmsource_options_t",5,"name")="gtmsource_options_t.checkhealth"
	Set gtmtypes("gtmsource_options_t",5,"off")=16
	Set gtmtypes("gtmsource_options_t",5,"len")=4
	Set gtmtypes("gtmsource_options_t",5,"type")="boolean_t"
	Set gtmtypfldindx("gtmsource_options_t","checkhealth")=5
	Set gtmtypes("gtmsource_options_t",6,"name")="gtmsource_options_t.deactivate"
	Set gtmtypes("gtmsource_options_t",6,"off")=20
	Set gtmtypes("gtmsource_options_t",6,"len")=4
	Set gtmtypes("gtmsource_options_t",6,"type")="boolean_t"
	Set gtmtypfldindx("gtmsource_options_t","deactivate")=6
	Set gtmtypes("gtmsource_options_t",7,"name")="gtmsource_options_t.jnlpool"
	Set gtmtypes("gtmsource_options_t",7,"off")=24
	Set gtmtypes("gtmsource_options_t",7,"len")=4
	Set gtmtypes("gtmsource_options_t",7,"type")="boolean_t"
	Set gtmtypfldindx("gtmsource_options_t","jnlpool")=7
	Set gtmtypes("gtmsource_options_t",8,"name")="gtmsource_options_t.showbacklog"
	Set gtmtypes("gtmsource_options_t",8,"off")=28
	Set gtmtypes("gtmsource_options_t",8,"len")=4
	Set gtmtypes("gtmsource_options_t",8,"type")="boolean_t"
	Set gtmtypfldindx("gtmsource_options_t","showbacklog")=8
	Set gtmtypes("gtmsource_options_t",9,"name")="gtmsource_options_t.statslog"
	Set gtmtypes("gtmsource_options_t",9,"off")=32
	Set gtmtypes("gtmsource_options_t",9,"len")=4
	Set gtmtypes("gtmsource_options_t",9,"type")="boolean_t"
	Set gtmtypfldindx("gtmsource_options_t","statslog")=9
	Set gtmtypes("gtmsource_options_t",10,"name")="gtmsource_options_t.stopsourcefilter"
	Set gtmtypes("gtmsource_options_t",10,"off")=36
	Set gtmtypes("gtmsource_options_t",10,"len")=4
	Set gtmtypes("gtmsource_options_t",10,"type")="boolean_t"
	Set gtmtypfldindx("gtmsource_options_t","stopsourcefilter")=10
	Set gtmtypes("gtmsource_options_t",11,"name")="gtmsource_options_t.rootprimary"
	Set gtmtypes("gtmsource_options_t",11,"off")=40
	Set gtmtypes("gtmsource_options_t",11,"len")=4
	Set gtmtypes("gtmsource_options_t",11,"type")="int"
	Set gtmtypfldindx("gtmsource_options_t","rootprimary")=11
	Set gtmtypes("gtmsource_options_t",12,"name")="gtmsource_options_t.instsecondary"
	Set gtmtypes("gtmsource_options_t",12,"off")=44
	Set gtmtypes("gtmsource_options_t",12,"len")=4
	Set gtmtypes("gtmsource_options_t",12,"type")="boolean_t"
	Set gtmtypfldindx("gtmsource_options_t","instsecondary")=12
	Set gtmtypes("gtmsource_options_t",13,"name")="gtmsource_options_t.needrestart"
	Set gtmtypes("gtmsource_options_t",13,"off")=48
	Set gtmtypes("gtmsource_options_t",13,"len")=4
	Set gtmtypes("gtmsource_options_t",13,"type")="boolean_t"
	Set gtmtypfldindx("gtmsource_options_t","needrestart")=13
	Set gtmtypes("gtmsource_options_t",14,"name")="gtmsource_options_t.losttncomplete"
	Set gtmtypes("gtmsource_options_t",14,"off")=52
	Set gtmtypes("gtmsource_options_t",14,"len")=4
	Set gtmtypes("gtmsource_options_t",14,"type")="boolean_t"
	Set gtmtypfldindx("gtmsource_options_t","losttncomplete")=14
	Set gtmtypes("gtmsource_options_t",15,"name")="gtmsource_options_t.showfreeze"
	Set gtmtypes("gtmsource_options_t",15,"off")=56
	Set gtmtypes("gtmsource_options_t",15,"len")=4
	Set gtmtypes("gtmsource_options_t",15,"type")="boolean_t"
	Set gtmtypfldindx("gtmsource_options_t","showfreeze")=15
	Set gtmtypes("gtmsource_options_t",16,"name")="gtmsource_options_t.setfreeze"
	Set gtmtypes("gtmsource_options_t",16,"off")=60
	Set gtmtypes("gtmsource_options_t",16,"len")=4
	Set gtmtypes("gtmsource_options_t",16,"type")="boolean_t"
	Set gtmtypfldindx("gtmsource_options_t","setfreeze")=16
	Set gtmtypes("gtmsource_options_t",17,"name")="gtmsource_options_t.freezeval"
	Set gtmtypes("gtmsource_options_t",17,"off")=64
	Set gtmtypes("gtmsource_options_t",17,"len")=4
	Set gtmtypes("gtmsource_options_t",17,"type")="boolean_t"
	Set gtmtypfldindx("gtmsource_options_t","freezeval")=17
	Set gtmtypes("gtmsource_options_t",18,"name")="gtmsource_options_t.setcomment"
	Set gtmtypes("gtmsource_options_t",18,"off")=68
	Set gtmtypes("gtmsource_options_t",18,"len")=4
	Set gtmtypes("gtmsource_options_t",18,"type")="boolean_t"
	Set gtmtypfldindx("gtmsource_options_t","setcomment")=18
	Set gtmtypes("gtmsource_options_t",19,"name")="gtmsource_options_t.jnlfileonly"
	Set gtmtypes("gtmsource_options_t",19,"off")=72
	Set gtmtypes("gtmsource_options_t",19,"len")=4
	Set gtmtypes("gtmsource_options_t",19,"type")="boolean_t"
	Set gtmtypfldindx("gtmsource_options_t","jnlfileonly")=19
	Set gtmtypes("gtmsource_options_t",20,"name")="gtmsource_options_t.zerobacklog"
	Set gtmtypes("gtmsource_options_t",20,"off")=76
	Set gtmtypes("gtmsource_options_t",20,"len")=4
	Set gtmtypes("gtmsource_options_t",20,"type")="boolean_t"
	Set gtmtypfldindx("gtmsource_options_t","zerobacklog")=20
	Set gtmtypes("gtmsource_options_t",21,"name")="gtmsource_options_t.cmplvl"
	Set gtmtypes("gtmsource_options_t",21,"off")=80
	Set gtmtypes("gtmsource_options_t",21,"len")=4
	Set gtmtypes("gtmsource_options_t",21,"type")="int"
	Set gtmtypfldindx("gtmsource_options_t","cmplvl")=21
	Set gtmtypes("gtmsource_options_t",22,"name")="gtmsource_options_t.shutdown_time"
	Set gtmtypes("gtmsource_options_t",22,"off")=84
	Set gtmtypes("gtmsource_options_t",22,"len")=4
	Set gtmtypes("gtmsource_options_t",22,"type")="int"
	Set gtmtypfldindx("gtmsource_options_t","shutdown_time")=22
	Set gtmtypes("gtmsource_options_t",23,"name")="gtmsource_options_t.buffsize"
	Set gtmtypes("gtmsource_options_t",23,"off")=88
	Set gtmtypes("gtmsource_options_t",23,"len")=8
	Set gtmtypes("gtmsource_options_t",23,"type")="uint64_t"
	Set gtmtypfldindx("gtmsource_options_t","buffsize")=23
	Set gtmtypes("gtmsource_options_t",24,"name")="gtmsource_options_t.mode"
	Set gtmtypes("gtmsource_options_t",24,"off")=96
	Set gtmtypes("gtmsource_options_t",24,"len")=4
	Set gtmtypes("gtmsource_options_t",24,"type")="int"
	Set gtmtypfldindx("gtmsource_options_t","mode")=24
	Set gtmtypes("gtmsource_options_t",25,"name")="gtmsource_options_t.secondary_port"
	Set gtmtypes("gtmsource_options_t",25,"off")=100
	Set gtmtypes("gtmsource_options_t",25,"len")=4
	Set gtmtypes("gtmsource_options_t",25,"type")="int"
	Set gtmtypfldindx("gtmsource_options_t","secondary_port")=25
	Set gtmtypes("gtmsource_options_t",26,"name")="gtmsource_options_t.src_log_interval"
	Set gtmtypes("gtmsource_options_t",26,"off")=104
	Set gtmtypes("gtmsource_options_t",26,"len")=4
	Set gtmtypes("gtmsource_options_t",26,"type")="unsigned-int"
	Set gtmtypfldindx("gtmsource_options_t","src_log_interval")=26
	Set gtmtypes("gtmsource_options_t",27,"name")="gtmsource_options_t.connect_parms"
	Set gtmtypes("gtmsource_options_t",27,"off")=108
	Set gtmtypes("gtmsource_options_t",27,"len")=24
	Set gtmtypes("gtmsource_options_t",27,"type")="int"
	Set gtmtypfldindx("gtmsource_options_t","connect_parms")=27
	Set gtmtypes("gtmsource_options_t",27,"dim")=6
	Set gtmtypes("gtmsource_options_t",28,"name")="gtmsource_options_t.filter_cmd"
	Set gtmtypes("gtmsource_options_t",28,"off")=132
	Set gtmtypes("gtmsource_options_t",28,"len")=512
	Set gtmtypes("gtmsource_options_t",28,"type")="char"
	Set gtmtypfldindx("gtmsource_options_t","filter_cmd")=28
	Set gtmtypes("gtmsource_options_t",29,"name")="gtmsource_options_t.secondary_host"
	Set gtmtypes("gtmsource_options_t",29,"off")=644
	Set gtmtypes("gtmsource_options_t",29,"len")=256
	Set gtmtypes("gtmsource_options_t",29,"type")="char"
	Set gtmtypfldindx("gtmsource_options_t","secondary_host")=29
	Set gtmtypes("gtmsource_options_t",30,"name")="gtmsource_options_t.log_file"
	Set gtmtypes("gtmsource_options_t",30,"off")=900
	Set gtmtypes("gtmsource_options_t",30,"len")=256
	Set gtmtypes("gtmsource_options_t",30,"type")="char"
	Set gtmtypfldindx("gtmsource_options_t","log_file")=30
	Set gtmtypes("gtmsource_options_t",31,"name")="gtmsource_options_t.secondary_instname"
	Set gtmtypes("gtmsource_options_t",31,"off")=1156
	Set gtmtypes("gtmsource_options_t",31,"len")=16
	Set gtmtypes("gtmsource_options_t",31,"type")="char"
	Set gtmtypfldindx("gtmsource_options_t","secondary_instname")=31
	Set gtmtypes("gtmsource_options_t",32,"name")="gtmsource_options_t.freeze_comment"
	Set gtmtypes("gtmsource_options_t",32,"off")=1172
	Set gtmtypes("gtmsource_options_t",32,"len")=1024
	Set gtmtypes("gtmsource_options_t",32,"type")="char"
	Set gtmtypfldindx("gtmsource_options_t","freeze_comment")=32
	Set gtmtypes("gtmsource_options_t",33,"name")="gtmsource_options_t.tlsid"
	Set gtmtypes("gtmsource_options_t",33,"off")=2196
	Set gtmtypes("gtmsource_options_t",33,"len")=32
	Set gtmtypes("gtmsource_options_t",33,"type")="char"
	Set gtmtypfldindx("gtmsource_options_t","tlsid")=33
	Set gtmtypes("gtmsource_options_t",34,"name")="gtmsource_options_t.renegotiate_interval"
	Set gtmtypes("gtmsource_options_t",34,"off")=2228
	Set gtmtypes("gtmsource_options_t",34,"len")=4
	Set gtmtypes("gtmsource_options_t",34,"type")="int"
	Set gtmtypfldindx("gtmsource_options_t","renegotiate_interval")=34
	;
	Set gtmtypes("gtmsrc_lcl")="struct"
	Set gtmtypes("gtmsrc_lcl",0)=4
	Set gtmtypes("gtmsrc_lcl","len")=64
	Set gtmtypes("gtmsrc_lcl",1,"name")="gtmsrc_lcl.secondary_instname"
	Set gtmtypes("gtmsrc_lcl",1,"off")=0
	Set gtmtypes("gtmsrc_lcl",1,"len")=16
	Set gtmtypes("gtmsrc_lcl",1,"type")="unsigned-char"
	Set gtmtypfldindx("gtmsrc_lcl","secondary_instname")=1
	Set gtmtypes("gtmsrc_lcl",2,"name")="gtmsrc_lcl.resync_seqno"
	Set gtmtypes("gtmsrc_lcl",2,"off")=16
	Set gtmtypes("gtmsrc_lcl",2,"len")=8
	Set gtmtypes("gtmsrc_lcl",2,"type")="uint64_t"
	Set gtmtypfldindx("gtmsrc_lcl","resync_seqno")=2
	Set gtmtypes("gtmsrc_lcl",3,"name")="gtmsrc_lcl.connect_jnl_seqno"
	Set gtmtypes("gtmsrc_lcl",3,"off")=24
	Set gtmtypes("gtmsrc_lcl",3,"len")=8
	Set gtmtypes("gtmsrc_lcl",3,"type")="uint64_t"
	Set gtmtypfldindx("gtmsrc_lcl","connect_jnl_seqno")=3
	Set gtmtypes("gtmsrc_lcl",4,"name")="gtmsrc_lcl.filler_64"
	Set gtmtypes("gtmsrc_lcl",4,"off")=32
	Set gtmtypes("gtmsrc_lcl",4,"len")=32
	Set gtmtypes("gtmsrc_lcl",4,"type")="unsigned-char"
	Set gtmtypfldindx("gtmsrc_lcl","filler_64")=4
	;
	Set gtmtypes("gv_key")="struct"
	Set gtmtypes("gv_key",0)=4
	Set gtmtypes("gv_key","len")=6
	Set gtmtypes("gv_key",1,"name")="gv_key.top"
	Set gtmtypes("gv_key",1,"off")=0
	Set gtmtypes("gv_key",1,"len")=2
	Set gtmtypes("gv_key",1,"type")="unsigned-short"
	Set gtmtypfldindx("gv_key","top")=1
	Set gtmtypes("gv_key",2,"name")="gv_key.end"
	Set gtmtypes("gv_key",2,"off")=2
	Set gtmtypes("gv_key",2,"len")=2
	Set gtmtypes("gv_key",2,"type")="unsigned-short"
	Set gtmtypfldindx("gv_key","end")=2
	Set gtmtypes("gv_key",3,"name")="gv_key.prev"
	Set gtmtypes("gv_key",3,"off")=4
	Set gtmtypes("gv_key",3,"len")=2
	Set gtmtypes("gv_key",3,"type")="unsigned-short"
	Set gtmtypfldindx("gv_key","prev")=3
	Set gtmtypes("gv_key",4,"name")="gv_key.base"
	Set gtmtypes("gv_key",4,"off")=6
	Set gtmtypes("gv_key",4,"len")=0
	Set gtmtypes("gv_key",4,"type")="unsigned-char"
	Set gtmtypfldindx("gv_key","base")=4
	;
	Set gtmtypes("gv_key_buf")="union"
	Set gtmtypes("gv_key_buf",0)=8
	Set gtmtypes("gv_key_buf","len")=1058
	Set gtmtypes("gv_key_buf",1,"name")="gv_key_buf.key"
	Set gtmtypes("gv_key_buf",1,"off")=0
	Set gtmtypes("gv_key_buf",1,"len")=6
	Set gtmtypes("gv_key_buf",1,"type")="gv_key_nobase"
	Set gtmtypfldindx("gv_key_buf","key")=1
	Set gtmtypes("gv_key_buf",2,"name")="gv_key_buf.key.top"
	Set gtmtypes("gv_key_buf",2,"off")=0
	Set gtmtypes("gv_key_buf",2,"len")=2
	Set gtmtypes("gv_key_buf",2,"type")="unsigned-short"
	Set gtmtypfldindx("gv_key_buf","key.top")=2
	Set gtmtypes("gv_key_buf",3,"name")="gv_key_buf.key.end"
	Set gtmtypes("gv_key_buf",3,"off")=2
	Set gtmtypes("gv_key_buf",3,"len")=2
	Set gtmtypes("gv_key_buf",3,"type")="unsigned-short"
	Set gtmtypfldindx("gv_key_buf","key.end")=3
	Set gtmtypes("gv_key_buf",4,"name")="gv_key_buf.key.prev"
	Set gtmtypes("gv_key_buf",4,"off")=4
	Set gtmtypes("gv_key_buf",4,"len")=2
	Set gtmtypes("gv_key_buf",4,"type")="unsigned-short"
	Set gtmtypfldindx("gv_key_buf","key.prev")=4
	Set gtmtypes("gv_key_buf",5,"name")="gv_key_buf.buf"
	Set gtmtypes("gv_key_buf",5,"off")=0
	Set gtmtypes("gv_key_buf",5,"len")=1058
	Set gtmtypes("gv_key_buf",5,"type")="char"
	Set gtmtypfldindx("gv_key_buf","buf")=5
	Set gtmtypes("gv_key_buf",6,"name")="gv_key_buf.split"
	Set gtmtypes("gv_key_buf",6,"off")=0
	Set gtmtypes("gv_key_buf",6,"len")=1058
	Set gtmtypes("gv_key_buf",6,"type")="struct"
	Set gtmtypfldindx("gv_key_buf","split")=6
	Set gtmtypes("gv_key_buf",7,"name")="gv_key_buf.split.gv_key_data"
	Set gtmtypes("gv_key_buf",7,"off")=0
	Set gtmtypes("gv_key_buf",7,"len")=6
	Set gtmtypes("gv_key_buf",7,"type")="char"
	Set gtmtypfldindx("gv_key_buf","split.gv_key_data")=7
	Set gtmtypes("gv_key_buf",8,"name")="gv_key_buf.split.base"
	Set gtmtypes("gv_key_buf",8,"off")=6
	Set gtmtypes("gv_key_buf",8,"len")=1052
	Set gtmtypes("gv_key_buf",8,"type")="char"
	Set gtmtypfldindx("gv_key_buf","split.base")=8
	;
	Set gtmtypes("gv_key_nobase")="struct"
	Set gtmtypes("gv_key_nobase",0)=3
	Set gtmtypes("gv_key_nobase","len")=6
	Set gtmtypes("gv_key_nobase",1,"name")="gv_key_nobase.top"
	Set gtmtypes("gv_key_nobase",1,"off")=0
	Set gtmtypes("gv_key_nobase",1,"len")=2
	Set gtmtypes("gv_key_nobase",1,"type")="unsigned-short"
	Set gtmtypfldindx("gv_key_nobase","top")=1
	Set gtmtypes("gv_key_nobase",2,"name")="gv_key_nobase.end"
	Set gtmtypes("gv_key_nobase",2,"off")=2
	Set gtmtypes("gv_key_nobase",2,"len")=2
	Set gtmtypes("gv_key_nobase",2,"type")="unsigned-short"
	Set gtmtypfldindx("gv_key_nobase","end")=2
	Set gtmtypes("gv_key_nobase",3,"name")="gv_key_nobase.prev"
	Set gtmtypes("gv_key_nobase",3,"off")=4
	Set gtmtypes("gv_key_nobase",3,"len")=2
	Set gtmtypes("gv_key_nobase",3,"type")="unsigned-short"
	Set gtmtypfldindx("gv_key_nobase","prev")=3
	;
	Set gtmtypes("gv_name_and_subscripts")="union"
	Set gtmtypes("gv_name_and_subscripts",0)=2
	Set gtmtypes("gv_name_and_subscripts","len")=140
	Set gtmtypes("gv_name_and_subscripts",1,"name")="gv_name_and_subscripts.lv"
	Set gtmtypes("gv_name_and_subscripts",1,"off")=0
	Set gtmtypes("gv_name_and_subscripts",1,"len")=140
	Set gtmtypes("gv_name_and_subscripts",1,"type")="int"
	Set gtmtypfldindx("gv_name_and_subscripts","lv")=1
	Set gtmtypes("gv_name_and_subscripts",1,"dim")=35
	Set gtmtypes("gv_name_and_subscripts",2,"name")="gv_name_and_subscripts.gv"
	Set gtmtypes("gv_name_and_subscripts",2,"off")=0
	Set gtmtypes("gv_name_and_subscripts",2,"len")=140
	Set gtmtypes("gv_name_and_subscripts",2,"type")="int"
	Set gtmtypfldindx("gv_name_and_subscripts","gv")=2
	Set gtmtypes("gv_name_and_subscripts",2,"dim")=35
	;
	Set gtmtypes("gv_namehead")="struct"
	Set gtmtypes("gv_namehead",0)=42
	Set gtmtypes("gv_namehead","len")=808
	Set gtmtypes("gv_namehead",1,"name")="gv_namehead.first_rec"
	Set gtmtypes("gv_namehead",1,"off")=0
	Set gtmtypes("gv_namehead",1,"len")=8
	Set gtmtypes("gv_namehead",1,"type")="addr"
	Set gtmtypfldindx("gv_namehead","first_rec")=1
	Set gtmtypes("gv_namehead",2,"name")="gv_namehead.last_rec"
	Set gtmtypes("gv_namehead",2,"off")=8
	Set gtmtypes("gv_namehead",2,"len")=8
	Set gtmtypes("gv_namehead",2,"type")="addr"
	Set gtmtypfldindx("gv_namehead","last_rec")=2
	Set gtmtypes("gv_namehead",3,"name")="gv_namehead.next_gvnh"
	Set gtmtypes("gv_namehead",3,"off")=16
	Set gtmtypes("gv_namehead",3,"len")=8
	Set gtmtypes("gv_namehead",3,"type")="addr"
	Set gtmtypfldindx("gv_namehead","next_gvnh")=3
	Set gtmtypes("gv_namehead",4,"name")="gv_namehead.prev_gvnh"
	Set gtmtypes("gv_namehead",4,"off")=24
	Set gtmtypes("gv_namehead",4,"len")=8
	Set gtmtypes("gv_namehead",4,"type")="addr"
	Set gtmtypfldindx("gv_namehead","prev_gvnh")=4
	Set gtmtypes("gv_namehead",5,"name")="gv_namehead.next_tp_gvnh"
	Set gtmtypes("gv_namehead",5,"off")=32
	Set gtmtypes("gv_namehead",5,"len")=8
	Set gtmtypes("gv_namehead",5,"type")="addr"
	Set gtmtypfldindx("gv_namehead","next_tp_gvnh")=5
	Set gtmtypes("gv_namehead",6,"name")="gv_namehead.gd_csa"
	Set gtmtypes("gv_namehead",6,"off")=40
	Set gtmtypes("gv_namehead",6,"len")=8
	Set gtmtypes("gv_namehead",6,"type")="addr"
	Set gtmtypfldindx("gv_namehead","gd_csa")=6
	Set gtmtypes("gv_namehead",7,"name")="gv_namehead.alt_hist"
	Set gtmtypes("gv_namehead",7,"off")=48
	Set gtmtypes("gv_namehead",7,"len")=8
	Set gtmtypes("gv_namehead",7,"type")="addr"
	Set gtmtypfldindx("gv_namehead","alt_hist")=7
	Set gtmtypes("gv_namehead",8,"name")="gv_namehead.collseq"
	Set gtmtypes("gv_namehead",8,"off")=56
	Set gtmtypes("gv_namehead",8,"len")=8
	Set gtmtypes("gv_namehead",8,"type")="addr"
	Set gtmtypfldindx("gv_namehead","collseq")=8
	Set gtmtypes("gv_namehead",9,"name")="gv_namehead.read_local_tn"
	Set gtmtypes("gv_namehead",9,"off")=64
	Set gtmtypes("gv_namehead",9,"len")=8
	Set gtmtypes("gv_namehead",9,"type")="uint64_t"
	Set gtmtypfldindx("gv_namehead","read_local_tn")=9
	Set gtmtypes("gv_namehead",10,"name")="gv_namehead.trig_local_tn"
	Set gtmtypes("gv_namehead",10,"off")=72
	Set gtmtypes("gv_namehead",10,"len")=8
	Set gtmtypes("gv_namehead",10,"type")="uint64_t"
	Set gtmtypfldindx("gv_namehead","trig_local_tn")=10
	Set gtmtypes("gv_namehead",11,"name")="gv_namehead.trig_read_tn"
	Set gtmtypes("gv_namehead",11,"off")=80
	Set gtmtypes("gv_namehead",11,"len")=8
	Set gtmtypes("gv_namehead",11,"type")="uint64_t"
	Set gtmtypfldindx("gv_namehead","trig_read_tn")=11
	Set gtmtypes("gv_namehead",12,"name")="gv_namehead.prev_key"
	Set gtmtypes("gv_namehead",12,"off")=88
	Set gtmtypes("gv_namehead",12,"len")=8
	Set gtmtypes("gv_namehead",12,"type")="addr"
	Set gtmtypfldindx("gv_namehead","prev_key")=12
	Set gtmtypes("gv_namehead",13,"name")="gv_namehead.noisolation"
	Set gtmtypes("gv_namehead",13,"off")=96
	Set gtmtypes("gv_namehead",13,"len")=4
	Set gtmtypes("gv_namehead",13,"type")="boolean_t"
	Set gtmtypfldindx("gv_namehead","noisolation")=13
	Set gtmtypes("gv_namehead",14,"name")="gv_namehead.root"
	Set gtmtypes("gv_namehead",14,"off")=100
	Set gtmtypes("gv_namehead",14,"len")=4
	Set gtmtypes("gv_namehead",14,"type")="int"
	Set gtmtypfldindx("gv_namehead","root")=14
	Set gtmtypes("gv_namehead",15,"name")="gv_namehead.gvname"
	Set gtmtypes("gv_namehead",15,"off")=104
	Set gtmtypes("gv_namehead",15,"len")=24
	Set gtmtypes("gv_namehead",15,"type")="mname_entry"
	Set gtmtypfldindx("gv_namehead","gvname")=15
	Set gtmtypes("gv_namehead",16,"name")="gv_namehead.gvname.var_name"
	Set gtmtypes("gv_namehead",16,"off")=104
	Set gtmtypes("gv_namehead",16,"len")=16
	Set gtmtypes("gv_namehead",16,"type")="mstr"
	Set gtmtypfldindx("gv_namehead","gvname.var_name")=16
	Set gtmtypes("gv_namehead",17,"name")="gv_namehead.gvname.hash_code"
	Set gtmtypes("gv_namehead",17,"off")=120
	Set gtmtypes("gv_namehead",17,"len")=4
	Set gtmtypes("gv_namehead",17,"type")="unsigned-int"
	Set gtmtypfldindx("gv_namehead","gvname.hash_code")=17
	Set gtmtypes("gv_namehead",18,"name")="gv_namehead.gvname.marked"
	Set gtmtypes("gv_namehead",18,"off")=124
	Set gtmtypes("gv_namehead",18,"len")=4
	Set gtmtypes("gv_namehead",18,"type")="boolean_t"
	Set gtmtypfldindx("gv_namehead","gvname.marked")=18
	Set gtmtypes("gv_namehead",19,"name")="gv_namehead.hist"
	Set gtmtypes("gv_namehead",19,"off")=128
	Set gtmtypes("gv_namehead",19,"len")=584
	Set gtmtypes("gv_namehead",19,"type")="srch_hist"
	Set gtmtypfldindx("gv_namehead","hist")=19
	Set gtmtypes("gv_namehead",20,"name")="gv_namehead.hist.depth"
	Set gtmtypes("gv_namehead",20,"off")=128
	Set gtmtypes("gv_namehead",20,"len")=4
	Set gtmtypes("gv_namehead",20,"type")="int"
	Set gtmtypfldindx("gv_namehead","hist.depth")=20
	Set gtmtypes("gv_namehead",21,"name")="gv_namehead.hist.filler"
	Set gtmtypes("gv_namehead",21,"off")=132
	Set gtmtypes("gv_namehead",21,"len")=4
	Set gtmtypes("gv_namehead",21,"type")="int"
	Set gtmtypfldindx("gv_namehead","hist.filler")=21
	Set gtmtypes("gv_namehead",22,"name")="gv_namehead.hist.h"
	Set gtmtypes("gv_namehead",22,"off")=136
	Set gtmtypes("gv_namehead",22,"len")=576
	Set gtmtypes("gv_namehead",22,"type")="srch_blk_status"
	Set gtmtypfldindx("gv_namehead","hist.h")=22
	Set gtmtypes("gv_namehead",22,"dim")=8
	Set gtmtypes("gv_namehead",23,"name")="gv_namehead.regcnt"
	Set gtmtypes("gv_namehead",23,"off")=712
	Set gtmtypes("gv_namehead",23,"len")=4
	Set gtmtypes("gv_namehead",23,"type")="int"
	Set gtmtypfldindx("gv_namehead","regcnt")=23
	Set gtmtypes("gv_namehead",24,"name")="gv_namehead.nct"
	Set gtmtypes("gv_namehead",24,"off")=716
	Set gtmtypes("gv_namehead",24,"len")=4
	Set gtmtypes("gv_namehead",24,"type")="unsigned-int"
	Set gtmtypfldindx("gv_namehead","nct")=24
	Set gtmtypes("gv_namehead",25,"name")="gv_namehead.act"
	Set gtmtypes("gv_namehead",25,"off")=720
	Set gtmtypes("gv_namehead",25,"len")=4
	Set gtmtypes("gv_namehead",25,"type")="unsigned-int"
	Set gtmtypfldindx("gv_namehead","act")=25
	Set gtmtypes("gv_namehead",26,"name")="gv_namehead.ver"
	Set gtmtypes("gv_namehead",26,"off")=724
	Set gtmtypes("gv_namehead",26,"len")=4
	Set gtmtypes("gv_namehead",26,"type")="unsigned-int"
	Set gtmtypfldindx("gv_namehead","ver")=26
	Set gtmtypes("gv_namehead",27,"name")="gv_namehead.act_specified_in_gld"
	Set gtmtypes("gv_namehead",27,"off")=728
	Set gtmtypes("gv_namehead",27,"len")=4
	Set gtmtypes("gv_namehead",27,"type")="boolean_t"
	Set gtmtypfldindx("gv_namehead","act_specified_in_gld")=27
	Set gtmtypes("gv_namehead",28,"name")="gv_namehead.nct_must_be_zero"
	Set gtmtypes("gv_namehead",28,"off")=732
	Set gtmtypes("gv_namehead",28,"len")=4
	Set gtmtypes("gv_namehead",28,"type")="boolean_t"
	Set gtmtypfldindx("gv_namehead","nct_must_be_zero")=28
	Set gtmtypes("gv_namehead",29,"name")="gv_namehead.split_cleanup_needed"
	Set gtmtypes("gv_namehead",29,"off")=736
	Set gtmtypes("gv_namehead",29,"len")=4
	Set gtmtypes("gv_namehead",29,"type")="boolean_t"
	Set gtmtypfldindx("gv_namehead","split_cleanup_needed")=29
	Set gtmtypes("gv_namehead",30,"name")="gv_namehead.last_split_direction"
	Set gtmtypes("gv_namehead",30,"off")=740
	Set gtmtypes("gv_namehead",30,"len")=6
	Set gtmtypes("gv_namehead",30,"type")="char"
	Set gtmtypfldindx("gv_namehead","last_split_direction")=30
	Set gtmtypes("gv_namehead",31,"name")="gv_namehead.filler_8byte_align1"
	Set gtmtypes("gv_namehead",31,"off")=746
	Set gtmtypes("gv_namehead",31,"len")=6
	Set gtmtypes("gv_namehead",31,"type")="char"
	Set gtmtypfldindx("gv_namehead","filler_8byte_align1")=31
	Set gtmtypes("gv_namehead",32,"name")="gv_namehead.last_split_blk_num"
	Set gtmtypes("gv_namehead",32,"off")=752
	Set gtmtypes("gv_namehead",32,"len")=24
	Set gtmtypes("gv_namehead",32,"type")="int"
	Set gtmtypfldindx("gv_namehead","last_split_blk_num")=32
	Set gtmtypes("gv_namehead",32,"dim")=6
	Set gtmtypes("gv_namehead",33,"name")="gv_namehead.gvt_trigger"
	Set gtmtypes("gv_namehead",33,"off")=776
	Set gtmtypes("gv_namehead",33,"len")=8
	Set gtmtypes("gv_namehead",33,"type")="addr"
	Set gtmtypfldindx("gv_namehead","gvt_trigger")=33
	Set gtmtypes("gv_namehead",34,"name")="gv_namehead.db_trigger_cycle"
	Set gtmtypes("gv_namehead",34,"off")=784
	Set gtmtypes("gv_namehead",34,"len")=4
	Set gtmtypes("gv_namehead",34,"type")="unsigned-int"
	Set gtmtypfldindx("gv_namehead","db_trigger_cycle")=34
	Set gtmtypes("gv_namehead",35,"name")="gv_namehead.db_dztrigger_cycle"
	Set gtmtypes("gv_namehead",35,"off")=788
	Set gtmtypes("gv_namehead",35,"len")=4
	Set gtmtypes("gv_namehead",35,"type")="unsigned-int"
	Set gtmtypfldindx("gv_namehead","db_dztrigger_cycle")=35
	Set gtmtypes("gv_namehead",36,"name")="gv_namehead.trig_mismatch_test_done"
	Set gtmtypes("gv_namehead",36,"off")=792
	Set gtmtypes("gv_namehead",36,"len")=4
	Set gtmtypes("gv_namehead",36,"type")="boolean_t"
	Set gtmtypfldindx("gv_namehead","trig_mismatch_test_done")=36
	Set gtmtypes("gv_namehead",37,"name")="gv_namehead.filler_8byte_align2"
	Set gtmtypes("gv_namehead",37,"off")=796
	Set gtmtypes("gv_namehead",37,"len")=4
	Set gtmtypes("gv_namehead",37,"type")="unsigned-int"
	Set gtmtypfldindx("gv_namehead","filler_8byte_align2")=37
	Set gtmtypes("gv_namehead",38,"name")="gv_namehead.filler_clue_end_align"
	Set gtmtypes("gv_namehead",38,"off")=800
	Set gtmtypes("gv_namehead",38,"len")=2
	Set gtmtypes("gv_namehead",38,"type")="unsigned-short"
	Set gtmtypfldindx("gv_namehead","filler_clue_end_align")=38
	Set gtmtypes("gv_namehead",39,"name")="gv_namehead.clue"
	Set gtmtypes("gv_namehead",39,"off")=802
	Set gtmtypes("gv_namehead",39,"len")=6
	Set gtmtypes("gv_namehead",39,"type")="gv_key_nobase"
	Set gtmtypfldindx("gv_namehead","clue")=39
	Set gtmtypes("gv_namehead",40,"name")="gv_namehead.clue.top"
	Set gtmtypes("gv_namehead",40,"off")=802
	Set gtmtypes("gv_namehead",40,"len")=2
	Set gtmtypes("gv_namehead",40,"type")="unsigned-short"
	Set gtmtypfldindx("gv_namehead","clue.top")=40
	Set gtmtypes("gv_namehead",41,"name")="gv_namehead.clue.end"
	Set gtmtypes("gv_namehead",41,"off")=804
	Set gtmtypes("gv_namehead",41,"len")=2
	Set gtmtypes("gv_namehead",41,"type")="unsigned-short"
	Set gtmtypfldindx("gv_namehead","clue.end")=41
	Set gtmtypes("gv_namehead",42,"name")="gv_namehead.clue.prev"
	Set gtmtypes("gv_namehead",42,"off")=806
	Set gtmtypes("gv_namehead",42,"len")=2
	Set gtmtypes("gv_namehead",42,"type")="unsigned-short"
	Set gtmtypfldindx("gv_namehead","clue.prev")=42
	;
	Set gtmtypes("gv_trigger_t")="struct"
	Set gtmtypes("gv_trigger_t",0)=32
	Set gtmtypes("gv_trigger_t","len")=192
	Set gtmtypes("gv_trigger_t",1,"name")="gv_trigger_t.next_set"
	Set gtmtypes("gv_trigger_t",1,"off")=0
	Set gtmtypes("gv_trigger_t",1,"len")=8
	Set gtmtypes("gv_trigger_t",1,"type")="addr"
	Set gtmtypfldindx("gv_trigger_t","next_set")=1
	Set gtmtypes("gv_trigger_t",2,"name")="gv_trigger_t.next_kill"
	Set gtmtypes("gv_trigger_t",2,"off")=8
	Set gtmtypes("gv_trigger_t",2,"len")=8
	Set gtmtypes("gv_trigger_t",2,"type")="addr"
	Set gtmtypfldindx("gv_trigger_t","next_kill")=2
	Set gtmtypes("gv_trigger_t",3,"name")="gv_trigger_t.next_ztrig"
	Set gtmtypes("gv_trigger_t",3,"off")=16
	Set gtmtypes("gv_trigger_t",3,"len")=8
	Set gtmtypes("gv_trigger_t",3,"type")="addr"
	Set gtmtypfldindx("gv_trigger_t","next_ztrig")=3
	Set gtmtypes("gv_trigger_t",4,"name")="gv_trigger_t.cmdmask"
	Set gtmtypes("gv_trigger_t",4,"off")=24
	Set gtmtypes("gv_trigger_t",4,"len")=4
	Set gtmtypes("gv_trigger_t",4,"type")="unsigned-int"
	Set gtmtypfldindx("gv_trigger_t","cmdmask")=4
	Set gtmtypes("gv_trigger_t",5,"name")="gv_trigger_t.numsubs"
	Set gtmtypes("gv_trigger_t",5,"off")=28
	Set gtmtypes("gv_trigger_t",5,"len")=4
	Set gtmtypes("gv_trigger_t",5,"type")="unsigned-int"
	Set gtmtypfldindx("gv_trigger_t","numsubs")=5
	Set gtmtypes("gv_trigger_t",6,"name")="gv_trigger_t.numlvsubs"
	Set gtmtypes("gv_trigger_t",6,"off")=32
	Set gtmtypes("gv_trigger_t",6,"len")=4
	Set gtmtypes("gv_trigger_t",6,"type")="unsigned-int"
	Set gtmtypfldindx("gv_trigger_t","numlvsubs")=6
	Set gtmtypes("gv_trigger_t",7,"name")="gv_trigger_t.numpieces"
	Set gtmtypes("gv_trigger_t",7,"off")=36
	Set gtmtypes("gv_trigger_t",7,"len")=4
	Set gtmtypes("gv_trigger_t",7,"type")="unsigned-int"
	Set gtmtypfldindx("gv_trigger_t","numpieces")=7
	Set gtmtypes("gv_trigger_t",8,"name")="gv_trigger_t.subsarray"
	Set gtmtypes("gv_trigger_t",8,"off")=40
	Set gtmtypes("gv_trigger_t",8,"len")=8
	Set gtmtypes("gv_trigger_t",8,"type")="addr"
	Set gtmtypfldindx("gv_trigger_t","subsarray")=8
	Set gtmtypes("gv_trigger_t",9,"name")="gv_trigger_t.lvindexarray"
	Set gtmtypes("gv_trigger_t",9,"off")=48
	Set gtmtypes("gv_trigger_t",9,"len")=8
	Set gtmtypes("gv_trigger_t",9,"type")="addr"
	Set gtmtypfldindx("gv_trigger_t","lvindexarray")=9
	Set gtmtypes("gv_trigger_t",10,"name")="gv_trigger_t.lvnamearray"
	Set gtmtypes("gv_trigger_t",10,"off")=56
	Set gtmtypes("gv_trigger_t",10,"len")=8
	Set gtmtypes("gv_trigger_t",10,"type")="addr"
	Set gtmtypfldindx("gv_trigger_t","lvnamearray")=10
	Set gtmtypes("gv_trigger_t",11,"name")="gv_trigger_t.piecearray"
	Set gtmtypes("gv_trigger_t",11,"off")=64
	Set gtmtypes("gv_trigger_t",11,"len")=8
	Set gtmtypes("gv_trigger_t",11,"type")="addr"
	Set gtmtypfldindx("gv_trigger_t","piecearray")=11
	Set gtmtypes("gv_trigger_t",12,"name")="gv_trigger_t.rtn_desc"
	Set gtmtypes("gv_trigger_t",12,"off")=72
	Set gtmtypes("gv_trigger_t",12,"len")=24
	Set gtmtypes("gv_trigger_t",12,"type")="rtn_tabent"
	Set gtmtypfldindx("gv_trigger_t","rtn_desc")=12
	Set gtmtypes("gv_trigger_t",13,"name")="gv_trigger_t.rtn_desc.rt_name"
	Set gtmtypes("gv_trigger_t",13,"off")=72
	Set gtmtypes("gv_trigger_t",13,"len")=16
	Set gtmtypes("gv_trigger_t",13,"type")="mstr"
	Set gtmtypfldindx("gv_trigger_t","rtn_desc.rt_name")=13
	Set gtmtypes("gv_trigger_t",14,"name")="gv_trigger_t.rtn_desc.rt_adr"
	Set gtmtypes("gv_trigger_t",14,"off")=88
	Set gtmtypes("gv_trigger_t",14,"len")=8
	Set gtmtypes("gv_trigger_t",14,"type")="addr"
	Set gtmtypfldindx("gv_trigger_t","rtn_desc.rt_adr")=14
	Set gtmtypes("gv_trigger_t",15,"name")="gv_trigger_t.is_zdelim"
	Set gtmtypes("gv_trigger_t",15,"off")=96
	Set gtmtypes("gv_trigger_t",15,"len")=4
	Set gtmtypes("gv_trigger_t",15,"type")="boolean_t"
	Set gtmtypfldindx("gv_trigger_t","is_zdelim")=15
	Set gtmtypes("gv_trigger_t",16,"name")="gv_trigger_t.delimiter"
	Set gtmtypes("gv_trigger_t",16,"off")=104
	Set gtmtypes("gv_trigger_t",16,"len")=32
	Set gtmtypes("gv_trigger_t",16,"type")="mval"
	Set gtmtypfldindx("gv_trigger_t","delimiter")=16
	Set gtmtypes("gv_trigger_t",17,"name")="gv_trigger_t.delimiter.mvtype"
	Set gtmtypes("gv_trigger_t",17,"off")=104
	Set gtmtypes("gv_trigger_t",17,"len")=2
	Set gtmtypes("gv_trigger_t",17,"type")="unsigned-short"
	Set gtmtypfldindx("gv_trigger_t","delimiter.mvtype")=17
	Set gtmtypes("gv_trigger_t",18,"name")="gv_trigger_t.delimiter.fnpc_indx"
	Set gtmtypes("gv_trigger_t",18,"off")=107
	Set gtmtypes("gv_trigger_t",18,"len")=1
	Set gtmtypes("gv_trigger_t",18,"type")="unsigned-char"
	Set gtmtypfldindx("gv_trigger_t","delimiter.fnpc_indx")=18
	Set gtmtypes("gv_trigger_t",19,"name")="gv_trigger_t.delimiter.utfcgr_indx"
	Set gtmtypes("gv_trigger_t",19,"off")=108
	Set gtmtypes("gv_trigger_t",19,"len")=4
	Set gtmtypes("gv_trigger_t",19,"type")="unsigned-int"
	Set gtmtypfldindx("gv_trigger_t","delimiter.utfcgr_indx")=19
	Set gtmtypes("gv_trigger_t",20,"name")="gv_trigger_t.delimiter.m"
	Set gtmtypes("gv_trigger_t",20,"off")=112
	Set gtmtypes("gv_trigger_t",20,"len")=8
	Set gtmtypes("gv_trigger_t",20,"type")="int"
	Set gtmtypfldindx("gv_trigger_t","delimiter.m")=20
	Set gtmtypes("gv_trigger_t",20,"dim")=2
	Set gtmtypes("gv_trigger_t",21,"name")="gv_trigger_t.delimiter.str"
	Set gtmtypes("gv_trigger_t",21,"off")=120
	Set gtmtypes("gv_trigger_t",21,"len")=16
	Set gtmtypes("gv_trigger_t",21,"type")="mstr"
	Set gtmtypfldindx("gv_trigger_t","delimiter.str")=21
	Set gtmtypes("gv_trigger_t",22,"name")="gv_trigger_t.options"
	Set gtmtypes("gv_trigger_t",22,"off")=136
	Set gtmtypes("gv_trigger_t",22,"len")=16
	Set gtmtypes("gv_trigger_t",22,"type")="mstr"
	Set gtmtypfldindx("gv_trigger_t","options")=22
	Set gtmtypes("gv_trigger_t",23,"name")="gv_trigger_t.options.char_len"
	Set gtmtypes("gv_trigger_t",23,"off")=136
	Set gtmtypes("gv_trigger_t",23,"len")=4
	Set gtmtypes("gv_trigger_t",23,"type")="unsigned-int"
	Set gtmtypfldindx("gv_trigger_t","options.char_len")=23
	Set gtmtypes("gv_trigger_t",24,"name")="gv_trigger_t.options.len"
	Set gtmtypes("gv_trigger_t",24,"off")=140
	Set gtmtypes("gv_trigger_t",24,"len")=4
	Set gtmtypes("gv_trigger_t",24,"type")="int"
	Set gtmtypfldindx("gv_trigger_t","options.len")=24
	Set gtmtypes("gv_trigger_t",25,"name")="gv_trigger_t.options.addr"
	Set gtmtypes("gv_trigger_t",25,"off")=144
	Set gtmtypes("gv_trigger_t",25,"len")=8
	Set gtmtypes("gv_trigger_t",25,"type")="addr"
	Set gtmtypfldindx("gv_trigger_t","options.addr")=25
	Set gtmtypes("gv_trigger_t",26,"name")="gv_trigger_t.xecute_str"
	Set gtmtypes("gv_trigger_t",26,"off")=152
	Set gtmtypes("gv_trigger_t",26,"len")=32
	Set gtmtypes("gv_trigger_t",26,"type")="mval"
	Set gtmtypfldindx("gv_trigger_t","xecute_str")=26
	Set gtmtypes("gv_trigger_t",27,"name")="gv_trigger_t.xecute_str.mvtype"
	Set gtmtypes("gv_trigger_t",27,"off")=152
	Set gtmtypes("gv_trigger_t",27,"len")=2
	Set gtmtypes("gv_trigger_t",27,"type")="unsigned-short"
	Set gtmtypfldindx("gv_trigger_t","xecute_str.mvtype")=27
	Set gtmtypes("gv_trigger_t",28,"name")="gv_trigger_t.xecute_str.fnpc_indx"
	Set gtmtypes("gv_trigger_t",28,"off")=155
	Set gtmtypes("gv_trigger_t",28,"len")=1
	Set gtmtypes("gv_trigger_t",28,"type")="unsigned-char"
	Set gtmtypfldindx("gv_trigger_t","xecute_str.fnpc_indx")=28
	Set gtmtypes("gv_trigger_t",29,"name")="gv_trigger_t.xecute_str.utfcgr_indx"
	Set gtmtypes("gv_trigger_t",29,"off")=156
	Set gtmtypes("gv_trigger_t",29,"len")=4
	Set gtmtypes("gv_trigger_t",29,"type")="unsigned-int"
	Set gtmtypfldindx("gv_trigger_t","xecute_str.utfcgr_indx")=29
	Set gtmtypes("gv_trigger_t",30,"name")="gv_trigger_t.xecute_str.m"
	Set gtmtypes("gv_trigger_t",30,"off")=160
	Set gtmtypes("gv_trigger_t",30,"len")=8
	Set gtmtypes("gv_trigger_t",30,"type")="int"
	Set gtmtypfldindx("gv_trigger_t","xecute_str.m")=30
	Set gtmtypes("gv_trigger_t",30,"dim")=2
	Set gtmtypes("gv_trigger_t",31,"name")="gv_trigger_t.xecute_str.str"
	Set gtmtypes("gv_trigger_t",31,"off")=168
	Set gtmtypes("gv_trigger_t",31,"len")=16
	Set gtmtypes("gv_trigger_t",31,"type")="mstr"
	Set gtmtypfldindx("gv_trigger_t","xecute_str.str")=31
	Set gtmtypes("gv_trigger_t",32,"name")="gv_trigger_t.gvt_trigger"
	Set gtmtypes("gv_trigger_t",32,"off")=184
	Set gtmtypes("gv_trigger_t",32,"len")=8
	Set gtmtypes("gv_trigger_t",32,"type")="addr"
	Set gtmtypfldindx("gv_trigger_t","gvt_trigger")=32
	;
	Set gtmtypes("gvargs_t")="struct"
	Set gtmtypes("gvargs_t",0)=2
	Set gtmtypes("gvargs_t","len")=272
	Set gtmtypes("gvargs_t",1,"name")="gvargs_t.count"
	Set gtmtypes("gvargs_t",1,"off")=0
	Set gtmtypes("gvargs_t",1,"len")=8
	Set gtmtypes("gvargs_t",1,"type")="ssize_t"
	Set gtmtypfldindx("gvargs_t","count")=1
	Set gtmtypes("gvargs_t",2,"name")="gvargs_t.args"
	Set gtmtypes("gvargs_t",2,"off")=8
	Set gtmtypes("gvargs_t",2,"len")=264
	Set gtmtypes("gvargs_t",2,"type")="addr"
	Set gtmtypfldindx("gvargs_t","args")=2
	Set gtmtypes("gvargs_t",2,"dim")=33
	;
	Set gtmtypes("gvname_info")="struct"
	Set gtmtypes("gvname_info",0)=9
	Set gtmtypes("gvname_info","len")=72
	Set gtmtypes("gvname_info",1,"name")="gvname_info.s_gv_currkey"
	Set gtmtypes("gvname_info",1,"off")=0
	Set gtmtypes("gvname_info",1,"len")=8
	Set gtmtypes("gvname_info",1,"type")="addr"
	Set gtmtypfldindx("gvname_info","s_gv_currkey")=1
	Set gtmtypes("gvname_info",2,"name")="gvname_info.s_gv_target"
	Set gtmtypes("gvname_info",2,"off")=8
	Set gtmtypes("gvname_info",2,"len")=8
	Set gtmtypes("gvname_info",2,"type")="addr"
	Set gtmtypfldindx("gvname_info","s_gv_target")=2
	Set gtmtypes("gvname_info",3,"name")="gvname_info.s_gv_cur_region"
	Set gtmtypes("gvname_info",3,"off")=16
	Set gtmtypes("gvname_info",3,"len")=8
	Set gtmtypes("gvname_info",3,"type")="addr"
	Set gtmtypfldindx("gvname_info","s_gv_cur_region")=3
	Set gtmtypes("gvname_info",4,"name")="gvname_info.s_cs_addrs"
	Set gtmtypes("gvname_info",4,"off")=24
	Set gtmtypes("gvname_info",4,"len")=8
	Set gtmtypes("gvname_info",4,"type")="addr"
	Set gtmtypfldindx("gvname_info","s_cs_addrs")=4
	Set gtmtypes("gvname_info",5,"name")="gvname_info.s_sgm_info_ptr"
	Set gtmtypes("gvname_info",5,"off")=32
	Set gtmtypes("gvname_info",5,"len")=8
	Set gtmtypes("gvname_info",5,"type")="addr"
	Set gtmtypfldindx("gvname_info","s_sgm_info_ptr")=5
	Set gtmtypes("gvname_info",6,"name")="gvname_info.s_gd_targ_gvnh_reg"
	Set gtmtypes("gvname_info",6,"off")=40
	Set gtmtypes("gvname_info",6,"len")=8
	Set gtmtypes("gvname_info",6,"type")="addr"
	Set gtmtypfldindx("gvname_info","s_gd_targ_gvnh_reg")=6
	Set gtmtypes("gvname_info",7,"name")="gvname_info.s_gd_targ_map"
	Set gtmtypes("gvname_info",7,"off")=48
	Set gtmtypes("gvname_info",7,"len")=8
	Set gtmtypes("gvname_info",7,"type")="addr"
	Set gtmtypfldindx("gvname_info","s_gd_targ_map")=7
	Set gtmtypes("gvname_info",8,"name")="gvname_info.s_gd_targ_addr"
	Set gtmtypes("gvname_info",8,"off")=56
	Set gtmtypes("gvname_info",8,"len")=8
	Set gtmtypes("gvname_info",8,"type")="addr"
	Set gtmtypfldindx("gvname_info","s_gd_targ_addr")=8
	Set gtmtypes("gvname_info",9,"name")="gvname_info.gvkey_nsubs"
	Set gtmtypes("gvname_info",9,"off")=64
	Set gtmtypes("gvname_info",9,"len")=4
	Set gtmtypes("gvname_info",9,"type")="int"
	Set gtmtypfldindx("gvname_info","gvkey_nsubs")=9
	;
	Set gtmtypes("gvnh_reg_t")="struct"
	Set gtmtypes("gvnh_reg_t",0)=5
	Set gtmtypes("gvnh_reg_t","len")=32
	Set gtmtypes("gvnh_reg_t",1,"name")="gvnh_reg_t.gvt"
	Set gtmtypes("gvnh_reg_t",1,"off")=0
	Set gtmtypes("gvnh_reg_t",1,"len")=8
	Set gtmtypes("gvnh_reg_t",1,"type")="addr"
	Set gtmtypfldindx("gvnh_reg_t","gvt")=1
	Set gtmtypes("gvnh_reg_t",2,"name")="gvnh_reg_t.gd_reg"
	Set gtmtypes("gvnh_reg_t",2,"off")=8
	Set gtmtypes("gvnh_reg_t",2,"len")=8
	Set gtmtypes("gvnh_reg_t",2,"type")="addr"
	Set gtmtypfldindx("gvnh_reg_t","gd_reg")=2
	Set gtmtypes("gvnh_reg_t",3,"name")="gvnh_reg_t.gvspan"
	Set gtmtypes("gvnh_reg_t",3,"off")=16
	Set gtmtypes("gvnh_reg_t",3,"len")=8
	Set gtmtypes("gvnh_reg_t",3,"type")="addr"
	Set gtmtypfldindx("gvnh_reg_t","gvspan")=3
	Set gtmtypes("gvnh_reg_t",4,"name")="gvnh_reg_t.act"
	Set gtmtypes("gvnh_reg_t",4,"off")=24
	Set gtmtypes("gvnh_reg_t",4,"len")=4
	Set gtmtypes("gvnh_reg_t",4,"type")="unsigned-int"
	Set gtmtypfldindx("gvnh_reg_t","act")=4
	Set gtmtypes("gvnh_reg_t",5,"name")="gvnh_reg_t.ver"
	Set gtmtypes("gvnh_reg_t",5,"off")=28
	Set gtmtypes("gvnh_reg_t",5,"len")=4
	Set gtmtypes("gvnh_reg_t",5,"type")="unsigned-int"
	Set gtmtypfldindx("gvnh_reg_t","ver")=5
	;
	Set gtmtypes("gvnh_spanreg_t")="struct"
	Set gtmtypes("gvnh_spanreg_t",0)=5
	Set gtmtypes("gvnh_spanreg_t","len")=24
	Set gtmtypes("gvnh_spanreg_t",1,"name")="gvnh_spanreg_t.start_map_index"
	Set gtmtypes("gvnh_spanreg_t",1,"off")=0
	Set gtmtypes("gvnh_spanreg_t",1,"len")=4
	Set gtmtypes("gvnh_spanreg_t",1,"type")="int"
	Set gtmtypfldindx("gvnh_spanreg_t","start_map_index")=1
	Set gtmtypes("gvnh_spanreg_t",2,"name")="gvnh_spanreg_t.end_map_index"
	Set gtmtypes("gvnh_spanreg_t",2,"off")=4
	Set gtmtypes("gvnh_spanreg_t",2,"len")=4
	Set gtmtypes("gvnh_spanreg_t",2,"type")="int"
	Set gtmtypfldindx("gvnh_spanreg_t","end_map_index")=2
	Set gtmtypes("gvnh_spanreg_t",3,"name")="gvnh_spanreg_t.min_reg_index"
	Set gtmtypes("gvnh_spanreg_t",3,"off")=8
	Set gtmtypes("gvnh_spanreg_t",3,"len")=4
	Set gtmtypes("gvnh_spanreg_t",3,"type")="int"
	Set gtmtypfldindx("gvnh_spanreg_t","min_reg_index")=3
	Set gtmtypes("gvnh_spanreg_t",4,"name")="gvnh_spanreg_t.max_reg_index"
	Set gtmtypes("gvnh_spanreg_t",4,"off")=12
	Set gtmtypes("gvnh_spanreg_t",4,"len")=4
	Set gtmtypes("gvnh_spanreg_t",4,"type")="int"
	Set gtmtypfldindx("gvnh_spanreg_t","max_reg_index")=4
	Set gtmtypes("gvnh_spanreg_t",5,"name")="gvnh_spanreg_t.gvt_array"
	Set gtmtypes("gvnh_spanreg_t",5,"off")=16
	Set gtmtypes("gvnh_spanreg_t",5,"len")=8
	Set gtmtypes("gvnh_spanreg_t",5,"type")="addr"
	Set gtmtypfldindx("gvnh_spanreg_t","gvt_array")=5
	;
	Set gtmtypes("gvsavtarg_t")="struct"
	Set gtmtypes("gvsavtarg_t",0)=8
	Set gtmtypes("gvsavtarg_t","len")=48
	Set gtmtypes("gvsavtarg_t",1,"name")="gvsavtarg_t.gv_cur_region"
	Set gtmtypes("gvsavtarg_t",1,"off")=0
	Set gtmtypes("gvsavtarg_t",1,"len")=8
	Set gtmtypes("gvsavtarg_t",1,"type")="addr"
	Set gtmtypfldindx("gvsavtarg_t","gv_cur_region")=1
	Set gtmtypes("gvsavtarg_t",2,"name")="gvsavtarg_t.gv_target"
	Set gtmtypes("gvsavtarg_t",2,"off")=8
	Set gtmtypes("gvsavtarg_t",2,"len")=8
	Set gtmtypes("gvsavtarg_t",2,"type")="addr"
	Set gtmtypfldindx("gvsavtarg_t","gv_target")=2
	Set gtmtypes("gvsavtarg_t",3,"name")="gvsavtarg_t.gd_targ_gvnh_reg"
	Set gtmtypes("gvsavtarg_t",3,"off")=16
	Set gtmtypes("gvsavtarg_t",3,"len")=8
	Set gtmtypes("gvsavtarg_t",3,"type")="addr"
	Set gtmtypfldindx("gvsavtarg_t","gd_targ_gvnh_reg")=3
	Set gtmtypes("gvsavtarg_t",4,"name")="gvsavtarg_t.gd_targ_map"
	Set gtmtypes("gvsavtarg_t",4,"off")=24
	Set gtmtypes("gvsavtarg_t",4,"len")=8
	Set gtmtypes("gvsavtarg_t",4,"type")="addr"
	Set gtmtypfldindx("gvsavtarg_t","gd_targ_map")=4
	Set gtmtypes("gvsavtarg_t",5,"name")="gvsavtarg_t.gv_last_subsc_null"
	Set gtmtypes("gvsavtarg_t",5,"off")=32
	Set gtmtypes("gvsavtarg_t",5,"len")=4
	Set gtmtypes("gvsavtarg_t",5,"type")="boolean_t"
	Set gtmtypfldindx("gvsavtarg_t","gv_last_subsc_null")=5
	Set gtmtypes("gvsavtarg_t",6,"name")="gvsavtarg_t.gv_some_subsc_null"
	Set gtmtypes("gvsavtarg_t",6,"off")=36
	Set gtmtypes("gvsavtarg_t",6,"len")=4
	Set gtmtypes("gvsavtarg_t",6,"type")="boolean_t"
	Set gtmtypfldindx("gvsavtarg_t","gv_some_subsc_null")=6
	Set gtmtypes("gvsavtarg_t",7,"name")="gvsavtarg_t.prev"
	Set gtmtypes("gvsavtarg_t",7,"off")=40
	Set gtmtypes("gvsavtarg_t",7,"len")=4
	Set gtmtypes("gvsavtarg_t",7,"type")="unsigned-int"
	Set gtmtypfldindx("gvsavtarg_t","prev")=7
	Set gtmtypes("gvsavtarg_t",8,"name")="gvsavtarg_t.end"
	Set gtmtypes("gvsavtarg_t",8,"off")=44
	Set gtmtypes("gvsavtarg_t",8,"len")=4
	Set gtmtypes("gvsavtarg_t",8,"type")="unsigned-int"
	Set gtmtypfldindx("gvsavtarg_t","end")=8
	;
	Set gtmtypes("gvstats_rec_csd_t")="struct"
	Set gtmtypes("gvstats_rec_csd_t",0)=74
	Set gtmtypes("gvstats_rec_csd_t","len")=592
	Set gtmtypes("gvstats_rec_csd_t",1,"name")="gvstats_rec_csd_t.n_set"
	Set gtmtypes("gvstats_rec_csd_t",1,"off")=0
	Set gtmtypes("gvstats_rec_csd_t",1,"len")=8
	Set gtmtypes("gvstats_rec_csd_t",1,"type")="uint64_t"
	Set gtmtypfldindx("gvstats_rec_csd_t","n_set")=1
	Set gtmtypes("gvstats_rec_csd_t",2,"name")="gvstats_rec_csd_t.n_kill"
	Set gtmtypes("gvstats_rec_csd_t",2,"off")=8
	Set gtmtypes("gvstats_rec_csd_t",2,"len")=8
	Set gtmtypes("gvstats_rec_csd_t",2,"type")="uint64_t"
	Set gtmtypfldindx("gvstats_rec_csd_t","n_kill")=2
	Set gtmtypes("gvstats_rec_csd_t",3,"name")="gvstats_rec_csd_t.n_get"
	Set gtmtypes("gvstats_rec_csd_t",3,"off")=16
	Set gtmtypes("gvstats_rec_csd_t",3,"len")=8
	Set gtmtypes("gvstats_rec_csd_t",3,"type")="uint64_t"
	Set gtmtypfldindx("gvstats_rec_csd_t","n_get")=3
	Set gtmtypes("gvstats_rec_csd_t",4,"name")="gvstats_rec_csd_t.n_data"
	Set gtmtypes("gvstats_rec_csd_t",4,"off")=24
	Set gtmtypes("gvstats_rec_csd_t",4,"len")=8
	Set gtmtypes("gvstats_rec_csd_t",4,"type")="uint64_t"
	Set gtmtypfldindx("gvstats_rec_csd_t","n_data")=4
	Set gtmtypes("gvstats_rec_csd_t",5,"name")="gvstats_rec_csd_t.n_order"
	Set gtmtypes("gvstats_rec_csd_t",5,"off")=32
	Set gtmtypes("gvstats_rec_csd_t",5,"len")=8
	Set gtmtypes("gvstats_rec_csd_t",5,"type")="uint64_t"
	Set gtmtypfldindx("gvstats_rec_csd_t","n_order")=5
	Set gtmtypes("gvstats_rec_csd_t",6,"name")="gvstats_rec_csd_t.n_zprev"
	Set gtmtypes("gvstats_rec_csd_t",6,"off")=40
	Set gtmtypes("gvstats_rec_csd_t",6,"len")=8
	Set gtmtypes("gvstats_rec_csd_t",6,"type")="uint64_t"
	Set gtmtypfldindx("gvstats_rec_csd_t","n_zprev")=6
	Set gtmtypes("gvstats_rec_csd_t",7,"name")="gvstats_rec_csd_t.n_query"
	Set gtmtypes("gvstats_rec_csd_t",7,"off")=48
	Set gtmtypes("gvstats_rec_csd_t",7,"len")=8
	Set gtmtypes("gvstats_rec_csd_t",7,"type")="uint64_t"
	Set gtmtypfldindx("gvstats_rec_csd_t","n_query")=7
	Set gtmtypes("gvstats_rec_csd_t",8,"name")="gvstats_rec_csd_t.n_lock_success"
	Set gtmtypes("gvstats_rec_csd_t",8,"off")=56
	Set gtmtypes("gvstats_rec_csd_t",8,"len")=8
	Set gtmtypes("gvstats_rec_csd_t",8,"type")="uint64_t"
	Set gtmtypfldindx("gvstats_rec_csd_t","n_lock_success")=8
	Set gtmtypes("gvstats_rec_csd_t",9,"name")="gvstats_rec_csd_t.n_lock_fail"
	Set gtmtypes("gvstats_rec_csd_t",9,"off")=64
	Set gtmtypes("gvstats_rec_csd_t",9,"len")=8
	Set gtmtypes("gvstats_rec_csd_t",9,"type")="uint64_t"
	Set gtmtypfldindx("gvstats_rec_csd_t","n_lock_fail")=9
	Set gtmtypes("gvstats_rec_csd_t",10,"name")="gvstats_rec_csd_t.db_curr_tn"
	Set gtmtypes("gvstats_rec_csd_t",10,"off")=72
	Set gtmtypes("gvstats_rec_csd_t",10,"len")=8
	Set gtmtypes("gvstats_rec_csd_t",10,"type")="uint64_t"
	Set gtmtypfldindx("gvstats_rec_csd_t","db_curr_tn")=10
	Set gtmtypes("gvstats_rec_csd_t",11,"name")="gvstats_rec_csd_t.n_dsk_read"
	Set gtmtypes("gvstats_rec_csd_t",11,"off")=80
	Set gtmtypes("gvstats_rec_csd_t",11,"len")=8
	Set gtmtypes("gvstats_rec_csd_t",11,"type")="uint64_t"
	Set gtmtypfldindx("gvstats_rec_csd_t","n_dsk_read")=11
	Set gtmtypes("gvstats_rec_csd_t",12,"name")="gvstats_rec_csd_t.n_dsk_write"
	Set gtmtypes("gvstats_rec_csd_t",12,"off")=88
	Set gtmtypes("gvstats_rec_csd_t",12,"len")=8
	Set gtmtypes("gvstats_rec_csd_t",12,"type")="uint64_t"
	Set gtmtypfldindx("gvstats_rec_csd_t","n_dsk_write")=12
	Set gtmtypes("gvstats_rec_csd_t",13,"name")="gvstats_rec_csd_t.n_nontp_readwrite"
	Set gtmtypes("gvstats_rec_csd_t",13,"off")=96
	Set gtmtypes("gvstats_rec_csd_t",13,"len")=8
	Set gtmtypes("gvstats_rec_csd_t",13,"type")="uint64_t"
	Set gtmtypfldindx("gvstats_rec_csd_t","n_nontp_readwrite")=13
	Set gtmtypes("gvstats_rec_csd_t",14,"name")="gvstats_rec_csd_t.n_nontp_readonly"
	Set gtmtypes("gvstats_rec_csd_t",14,"off")=104
	Set gtmtypes("gvstats_rec_csd_t",14,"len")=8
	Set gtmtypes("gvstats_rec_csd_t",14,"type")="uint64_t"
	Set gtmtypfldindx("gvstats_rec_csd_t","n_nontp_readonly")=14
	Set gtmtypes("gvstats_rec_csd_t",15,"name")="gvstats_rec_csd_t.n_nontp_blkwrite"
	Set gtmtypes("gvstats_rec_csd_t",15,"off")=112
	Set gtmtypes("gvstats_rec_csd_t",15,"len")=8
	Set gtmtypes("gvstats_rec_csd_t",15,"type")="uint64_t"
	Set gtmtypfldindx("gvstats_rec_csd_t","n_nontp_blkwrite")=15
	Set gtmtypes("gvstats_rec_csd_t",16,"name")="gvstats_rec_csd_t.n_nontp_blkread"
	Set gtmtypes("gvstats_rec_csd_t",16,"off")=120
	Set gtmtypes("gvstats_rec_csd_t",16,"len")=8
	Set gtmtypes("gvstats_rec_csd_t",16,"type")="uint64_t"
	Set gtmtypfldindx("gvstats_rec_csd_t","n_nontp_blkread")=16
	Set gtmtypes("gvstats_rec_csd_t",17,"name")="gvstats_rec_csd_t.n_nontp_retries_0"
	Set gtmtypes("gvstats_rec_csd_t",17,"off")=128
	Set gtmtypes("gvstats_rec_csd_t",17,"len")=8
	Set gtmtypes("gvstats_rec_csd_t",17,"type")="uint64_t"
	Set gtmtypfldindx("gvstats_rec_csd_t","n_nontp_retries_0")=17
	Set gtmtypes("gvstats_rec_csd_t",18,"name")="gvstats_rec_csd_t.n_nontp_retries_1"
	Set gtmtypes("gvstats_rec_csd_t",18,"off")=136
	Set gtmtypes("gvstats_rec_csd_t",18,"len")=8
	Set gtmtypes("gvstats_rec_csd_t",18,"type")="uint64_t"
	Set gtmtypfldindx("gvstats_rec_csd_t","n_nontp_retries_1")=18
	Set gtmtypes("gvstats_rec_csd_t",19,"name")="gvstats_rec_csd_t.n_nontp_retries_2"
	Set gtmtypes("gvstats_rec_csd_t",19,"off")=144
	Set gtmtypes("gvstats_rec_csd_t",19,"len")=8
	Set gtmtypes("gvstats_rec_csd_t",19,"type")="uint64_t"
	Set gtmtypfldindx("gvstats_rec_csd_t","n_nontp_retries_2")=19
	Set gtmtypes("gvstats_rec_csd_t",20,"name")="gvstats_rec_csd_t.n_nontp_retries_3"
	Set gtmtypes("gvstats_rec_csd_t",20,"off")=152
	Set gtmtypes("gvstats_rec_csd_t",20,"len")=8
	Set gtmtypes("gvstats_rec_csd_t",20,"type")="uint64_t"
	Set gtmtypfldindx("gvstats_rec_csd_t","n_nontp_retries_3")=20
	Set gtmtypes("gvstats_rec_csd_t",21,"name")="gvstats_rec_csd_t.n_tp_readwrite"
	Set gtmtypes("gvstats_rec_csd_t",21,"off")=160
	Set gtmtypes("gvstats_rec_csd_t",21,"len")=8
	Set gtmtypes("gvstats_rec_csd_t",21,"type")="uint64_t"
	Set gtmtypfldindx("gvstats_rec_csd_t","n_tp_readwrite")=21
	Set gtmtypes("gvstats_rec_csd_t",22,"name")="gvstats_rec_csd_t.n_tp_readonly"
	Set gtmtypes("gvstats_rec_csd_t",22,"off")=168
	Set gtmtypes("gvstats_rec_csd_t",22,"len")=8
	Set gtmtypes("gvstats_rec_csd_t",22,"type")="uint64_t"
	Set gtmtypfldindx("gvstats_rec_csd_t","n_tp_readonly")=22
	Set gtmtypes("gvstats_rec_csd_t",23,"name")="gvstats_rec_csd_t.n_tp_rolledback"
	Set gtmtypes("gvstats_rec_csd_t",23,"off")=176
	Set gtmtypes("gvstats_rec_csd_t",23,"len")=8
	Set gtmtypes("gvstats_rec_csd_t",23,"type")="uint64_t"
	Set gtmtypfldindx("gvstats_rec_csd_t","n_tp_rolledback")=23
	Set gtmtypes("gvstats_rec_csd_t",24,"name")="gvstats_rec_csd_t.n_tp_blkwrite"
	Set gtmtypes("gvstats_rec_csd_t",24,"off")=184
	Set gtmtypes("gvstats_rec_csd_t",24,"len")=8
	Set gtmtypes("gvstats_rec_csd_t",24,"type")="uint64_t"
	Set gtmtypfldindx("gvstats_rec_csd_t","n_tp_blkwrite")=24
	Set gtmtypes("gvstats_rec_csd_t",25,"name")="gvstats_rec_csd_t.n_tp_blkread"
	Set gtmtypes("gvstats_rec_csd_t",25,"off")=192
	Set gtmtypes("gvstats_rec_csd_t",25,"len")=8
	Set gtmtypes("gvstats_rec_csd_t",25,"type")="uint64_t"
	Set gtmtypfldindx("gvstats_rec_csd_t","n_tp_blkread")=25
	Set gtmtypes("gvstats_rec_csd_t",26,"name")="gvstats_rec_csd_t.n_tp_tot_retries_0"
	Set gtmtypes("gvstats_rec_csd_t",26,"off")=200
	Set gtmtypes("gvstats_rec_csd_t",26,"len")=8
	Set gtmtypes("gvstats_rec_csd_t",26,"type")="uint64_t"
	Set gtmtypfldindx("gvstats_rec_csd_t","n_tp_tot_retries_0")=26
	Set gtmtypes("gvstats_rec_csd_t",27,"name")="gvstats_rec_csd_t.n_tp_tot_retries_1"
	Set gtmtypes("gvstats_rec_csd_t",27,"off")=208
	Set gtmtypes("gvstats_rec_csd_t",27,"len")=8
	Set gtmtypes("gvstats_rec_csd_t",27,"type")="uint64_t"
	Set gtmtypfldindx("gvstats_rec_csd_t","n_tp_tot_retries_1")=27
	Set gtmtypes("gvstats_rec_csd_t",28,"name")="gvstats_rec_csd_t.n_tp_tot_retries_2"
	Set gtmtypes("gvstats_rec_csd_t",28,"off")=216
	Set gtmtypes("gvstats_rec_csd_t",28,"len")=8
	Set gtmtypes("gvstats_rec_csd_t",28,"type")="uint64_t"
	Set gtmtypfldindx("gvstats_rec_csd_t","n_tp_tot_retries_2")=28
	Set gtmtypes("gvstats_rec_csd_t",29,"name")="gvstats_rec_csd_t.n_tp_tot_retries_3"
	Set gtmtypes("gvstats_rec_csd_t",29,"off")=224
	Set gtmtypes("gvstats_rec_csd_t",29,"len")=8
	Set gtmtypes("gvstats_rec_csd_t",29,"type")="uint64_t"
	Set gtmtypfldindx("gvstats_rec_csd_t","n_tp_tot_retries_3")=29
	Set gtmtypes("gvstats_rec_csd_t",30,"name")="gvstats_rec_csd_t.n_tp_tot_retries_4"
	Set gtmtypes("gvstats_rec_csd_t",30,"off")=232
	Set gtmtypes("gvstats_rec_csd_t",30,"len")=8
	Set gtmtypes("gvstats_rec_csd_t",30,"type")="uint64_t"
	Set gtmtypfldindx("gvstats_rec_csd_t","n_tp_tot_retries_4")=30
	Set gtmtypes("gvstats_rec_csd_t",31,"name")="gvstats_rec_csd_t.n_tp_cnflct_retries_0"
	Set gtmtypes("gvstats_rec_csd_t",31,"off")=240
	Set gtmtypes("gvstats_rec_csd_t",31,"len")=8
	Set gtmtypes("gvstats_rec_csd_t",31,"type")="uint64_t"
	Set gtmtypfldindx("gvstats_rec_csd_t","n_tp_cnflct_retries_0")=31
	Set gtmtypes("gvstats_rec_csd_t",32,"name")="gvstats_rec_csd_t.n_tp_cnflct_retries_1"
	Set gtmtypes("gvstats_rec_csd_t",32,"off")=248
	Set gtmtypes("gvstats_rec_csd_t",32,"len")=8
	Set gtmtypes("gvstats_rec_csd_t",32,"type")="uint64_t"
	Set gtmtypfldindx("gvstats_rec_csd_t","n_tp_cnflct_retries_1")=32
	Set gtmtypes("gvstats_rec_csd_t",33,"name")="gvstats_rec_csd_t.n_tp_cnflct_retries_2"
	Set gtmtypes("gvstats_rec_csd_t",33,"off")=256
	Set gtmtypes("gvstats_rec_csd_t",33,"len")=8
	Set gtmtypes("gvstats_rec_csd_t",33,"type")="uint64_t"
	Set gtmtypfldindx("gvstats_rec_csd_t","n_tp_cnflct_retries_2")=33
	Set gtmtypes("gvstats_rec_csd_t",34,"name")="gvstats_rec_csd_t.n_tp_cnflct_retries_3"
	Set gtmtypes("gvstats_rec_csd_t",34,"off")=264
	Set gtmtypes("gvstats_rec_csd_t",34,"len")=8
	Set gtmtypes("gvstats_rec_csd_t",34,"type")="uint64_t"
	Set gtmtypfldindx("gvstats_rec_csd_t","n_tp_cnflct_retries_3")=34
	Set gtmtypes("gvstats_rec_csd_t",35,"name")="gvstats_rec_csd_t.n_tp_cnflct_retries_4"
	Set gtmtypes("gvstats_rec_csd_t",35,"off")=272
	Set gtmtypes("gvstats_rec_csd_t",35,"len")=8
	Set gtmtypes("gvstats_rec_csd_t",35,"type")="uint64_t"
	Set gtmtypfldindx("gvstats_rec_csd_t","n_tp_cnflct_retries_4")=35
	Set gtmtypes("gvstats_rec_csd_t",36,"name")="gvstats_rec_csd_t.n_ztrigger"
	Set gtmtypes("gvstats_rec_csd_t",36,"off")=280
	Set gtmtypes("gvstats_rec_csd_t",36,"len")=8
	Set gtmtypes("gvstats_rec_csd_t",36,"type")="uint64_t"
	Set gtmtypfldindx("gvstats_rec_csd_t","n_ztrigger")=36
	Set gtmtypes("gvstats_rec_csd_t",37,"name")="gvstats_rec_csd_t.n_db_flush"
	Set gtmtypes("gvstats_rec_csd_t",37,"off")=288
	Set gtmtypes("gvstats_rec_csd_t",37,"len")=8
	Set gtmtypes("gvstats_rec_csd_t",37,"type")="uint64_t"
	Set gtmtypfldindx("gvstats_rec_csd_t","n_db_flush")=37
	Set gtmtypes("gvstats_rec_csd_t",38,"name")="gvstats_rec_csd_t.n_db_fsync"
	Set gtmtypes("gvstats_rec_csd_t",38,"off")=296
	Set gtmtypes("gvstats_rec_csd_t",38,"len")=8
	Set gtmtypes("gvstats_rec_csd_t",38,"type")="uint64_t"
	Set gtmtypfldindx("gvstats_rec_csd_t","n_db_fsync")=38
	Set gtmtypes("gvstats_rec_csd_t",39,"name")="gvstats_rec_csd_t.n_jnl_flush"
	Set gtmtypes("gvstats_rec_csd_t",39,"off")=304
	Set gtmtypes("gvstats_rec_csd_t",39,"len")=8
	Set gtmtypes("gvstats_rec_csd_t",39,"type")="uint64_t"
	Set gtmtypfldindx("gvstats_rec_csd_t","n_jnl_flush")=39
	Set gtmtypes("gvstats_rec_csd_t",40,"name")="gvstats_rec_csd_t.n_jnl_fsync"
	Set gtmtypes("gvstats_rec_csd_t",40,"off")=312
	Set gtmtypes("gvstats_rec_csd_t",40,"len")=8
	Set gtmtypes("gvstats_rec_csd_t",40,"type")="uint64_t"
	Set gtmtypfldindx("gvstats_rec_csd_t","n_jnl_fsync")=40
	Set gtmtypes("gvstats_rec_csd_t",41,"name")="gvstats_rec_csd_t.n_jbuff_bytes"
	Set gtmtypes("gvstats_rec_csd_t",41,"off")=320
	Set gtmtypes("gvstats_rec_csd_t",41,"len")=8
	Set gtmtypes("gvstats_rec_csd_t",41,"type")="uint64_t"
	Set gtmtypfldindx("gvstats_rec_csd_t","n_jbuff_bytes")=41
	Set gtmtypes("gvstats_rec_csd_t",42,"name")="gvstats_rec_csd_t.n_jfile_bytes"
	Set gtmtypes("gvstats_rec_csd_t",42,"off")=328
	Set gtmtypes("gvstats_rec_csd_t",42,"len")=8
	Set gtmtypes("gvstats_rec_csd_t",42,"type")="uint64_t"
	Set gtmtypfldindx("gvstats_rec_csd_t","n_jfile_bytes")=42
	Set gtmtypes("gvstats_rec_csd_t",43,"name")="gvstats_rec_csd_t.n_jfile_writes"
	Set gtmtypes("gvstats_rec_csd_t",43,"off")=336
	Set gtmtypes("gvstats_rec_csd_t",43,"len")=8
	Set gtmtypes("gvstats_rec_csd_t",43,"type")="uint64_t"
	Set gtmtypfldindx("gvstats_rec_csd_t","n_jfile_writes")=43
	Set gtmtypes("gvstats_rec_csd_t",44,"name")="gvstats_rec_csd_t.n_jrec_logical"
	Set gtmtypes("gvstats_rec_csd_t",44,"off")=344
	Set gtmtypes("gvstats_rec_csd_t",44,"len")=8
	Set gtmtypes("gvstats_rec_csd_t",44,"type")="uint64_t"
	Set gtmtypfldindx("gvstats_rec_csd_t","n_jrec_logical")=44
	Set gtmtypes("gvstats_rec_csd_t",45,"name")="gvstats_rec_csd_t.n_jrec_pblk"
	Set gtmtypes("gvstats_rec_csd_t",45,"off")=352
	Set gtmtypes("gvstats_rec_csd_t",45,"len")=8
	Set gtmtypes("gvstats_rec_csd_t",45,"type")="uint64_t"
	Set gtmtypfldindx("gvstats_rec_csd_t","n_jrec_pblk")=45
	Set gtmtypes("gvstats_rec_csd_t",46,"name")="gvstats_rec_csd_t.n_jrec_epoch_regular"
	Set gtmtypes("gvstats_rec_csd_t",46,"off")=360
	Set gtmtypes("gvstats_rec_csd_t",46,"len")=8
	Set gtmtypes("gvstats_rec_csd_t",46,"type")="uint64_t"
	Set gtmtypfldindx("gvstats_rec_csd_t","n_jrec_epoch_regular")=46
	Set gtmtypes("gvstats_rec_csd_t",47,"name")="gvstats_rec_csd_t.n_jrec_epoch_idle"
	Set gtmtypes("gvstats_rec_csd_t",47,"off")=368
	Set gtmtypes("gvstats_rec_csd_t",47,"len")=8
	Set gtmtypes("gvstats_rec_csd_t",47,"type")="uint64_t"
	Set gtmtypfldindx("gvstats_rec_csd_t","n_jrec_epoch_idle")=47
	Set gtmtypes("gvstats_rec_csd_t",48,"name")="gvstats_rec_csd_t.n_jrec_other"
	Set gtmtypes("gvstats_rec_csd_t",48,"off")=376
	Set gtmtypes("gvstats_rec_csd_t",48,"len")=8
	Set gtmtypes("gvstats_rec_csd_t",48,"type")="uint64_t"
	Set gtmtypfldindx("gvstats_rec_csd_t","n_jrec_other")=48
	Set gtmtypes("gvstats_rec_csd_t",49,"name")="gvstats_rec_csd_t.n_jnl_extends"
	Set gtmtypes("gvstats_rec_csd_t",49,"off")=384
	Set gtmtypes("gvstats_rec_csd_t",49,"len")=8
	Set gtmtypes("gvstats_rec_csd_t",49,"type")="uint64_t"
	Set gtmtypfldindx("gvstats_rec_csd_t","n_jnl_extends")=49
	Set gtmtypes("gvstats_rec_csd_t",50,"name")="gvstats_rec_csd_t.n_db_extends"
	Set gtmtypes("gvstats_rec_csd_t",50,"off")=392
	Set gtmtypes("gvstats_rec_csd_t",50,"len")=8
	Set gtmtypes("gvstats_rec_csd_t",50,"type")="uint64_t"
	Set gtmtypfldindx("gvstats_rec_csd_t","n_db_extends")=50
	Set gtmtypes("gvstats_rec_csd_t",51,"name")="gvstats_rec_csd_t.n_crit_success"
	Set gtmtypes("gvstats_rec_csd_t",51,"off")=400
	Set gtmtypes("gvstats_rec_csd_t",51,"len")=8
	Set gtmtypes("gvstats_rec_csd_t",51,"type")="uint64_t"
	Set gtmtypfldindx("gvstats_rec_csd_t","n_crit_success")=51
	Set gtmtypes("gvstats_rec_csd_t",52,"name")="gvstats_rec_csd_t.n_crits_in_epch"
	Set gtmtypes("gvstats_rec_csd_t",52,"off")=408
	Set gtmtypes("gvstats_rec_csd_t",52,"len")=8
	Set gtmtypes("gvstats_rec_csd_t",52,"type")="uint64_t"
	Set gtmtypfldindx("gvstats_rec_csd_t","n_crits_in_epch")=52
	Set gtmtypes("gvstats_rec_csd_t",53,"name")="gvstats_rec_csd_t.sq_crit_failed"
	Set gtmtypes("gvstats_rec_csd_t",53,"off")=416
	Set gtmtypes("gvstats_rec_csd_t",53,"len")=8
	Set gtmtypes("gvstats_rec_csd_t",53,"type")="uint64_t"
	Set gtmtypfldindx("gvstats_rec_csd_t","sq_crit_failed")=53
	Set gtmtypes("gvstats_rec_csd_t",54,"name")="gvstats_rec_csd_t.n_crit_failed"
	Set gtmtypes("gvstats_rec_csd_t",54,"off")=424
	Set gtmtypes("gvstats_rec_csd_t",54,"len")=8
	Set gtmtypes("gvstats_rec_csd_t",54,"type")="uint64_t"
	Set gtmtypfldindx("gvstats_rec_csd_t","n_crit_failed")=54
	Set gtmtypes("gvstats_rec_csd_t",55,"name")="gvstats_rec_csd_t.sq_crit_que_slps"
	Set gtmtypes("gvstats_rec_csd_t",55,"off")=432
	Set gtmtypes("gvstats_rec_csd_t",55,"len")=8
	Set gtmtypes("gvstats_rec_csd_t",55,"type")="uint64_t"
	Set gtmtypfldindx("gvstats_rec_csd_t","sq_crit_que_slps")=55
	Set gtmtypes("gvstats_rec_csd_t",56,"name")="gvstats_rec_csd_t.n_crit_que_slps"
	Set gtmtypes("gvstats_rec_csd_t",56,"off")=440
	Set gtmtypes("gvstats_rec_csd_t",56,"len")=8
	Set gtmtypes("gvstats_rec_csd_t",56,"type")="uint64_t"
	Set gtmtypfldindx("gvstats_rec_csd_t","n_crit_que_slps")=56
	Set gtmtypes("gvstats_rec_csd_t",57,"name")="gvstats_rec_csd_t.sq_crit_yields"
	Set gtmtypes("gvstats_rec_csd_t",57,"off")=448
	Set gtmtypes("gvstats_rec_csd_t",57,"len")=8
	Set gtmtypes("gvstats_rec_csd_t",57,"type")="uint64_t"
	Set gtmtypfldindx("gvstats_rec_csd_t","sq_crit_yields")=57
	Set gtmtypes("gvstats_rec_csd_t",58,"name")="gvstats_rec_csd_t.n_crit_yields"
	Set gtmtypes("gvstats_rec_csd_t",58,"off")=456
	Set gtmtypes("gvstats_rec_csd_t",58,"len")=8
	Set gtmtypes("gvstats_rec_csd_t",58,"type")="uint64_t"
	Set gtmtypfldindx("gvstats_rec_csd_t","n_crit_yields")=58
	Set gtmtypes("gvstats_rec_csd_t",59,"name")="gvstats_rec_csd_t.n_clean2dirty"
	Set gtmtypes("gvstats_rec_csd_t",59,"off")=464
	Set gtmtypes("gvstats_rec_csd_t",59,"len")=8
	Set gtmtypes("gvstats_rec_csd_t",59,"type")="uint64_t"
	Set gtmtypfldindx("gvstats_rec_csd_t","n_clean2dirty")=59
	Set gtmtypes("gvstats_rec_csd_t",60,"name")="gvstats_rec_csd_t.n_wait_for_read"
	Set gtmtypes("gvstats_rec_csd_t",60,"off")=472
	Set gtmtypes("gvstats_rec_csd_t",60,"len")=8
	Set gtmtypes("gvstats_rec_csd_t",60,"type")="uint64_t"
	Set gtmtypfldindx("gvstats_rec_csd_t","n_wait_for_read")=60
	Set gtmtypes("gvstats_rec_csd_t",61,"name")="gvstats_rec_csd_t.n_buffer_scarce"
	Set gtmtypes("gvstats_rec_csd_t",61,"off")=480
	Set gtmtypes("gvstats_rec_csd_t",61,"len")=8
	Set gtmtypes("gvstats_rec_csd_t",61,"type")="uint64_t"
	Set gtmtypfldindx("gvstats_rec_csd_t","n_buffer_scarce")=61
	Set gtmtypes("gvstats_rec_csd_t",62,"name")="gvstats_rec_csd_t.n_bt_scarce"
	Set gtmtypes("gvstats_rec_csd_t",62,"off")=488
	Set gtmtypes("gvstats_rec_csd_t",62,"len")=8
	Set gtmtypes("gvstats_rec_csd_t",62,"type")="uint64_t"
	Set gtmtypfldindx("gvstats_rec_csd_t","n_bt_scarce")=62
	Set gtmtypes("gvstats_rec_csd_t",63,"name")="gvstats_rec_csd_t.f_dbext_wait"
	Set gtmtypes("gvstats_rec_csd_t",63,"off")=496
	Set gtmtypes("gvstats_rec_csd_t",63,"len")=8
	Set gtmtypes("gvstats_rec_csd_t",63,"type")="uint64_t"
	Set gtmtypfldindx("gvstats_rec_csd_t","f_dbext_wait")=63
	Set gtmtypes("gvstats_rec_csd_t",64,"name")="gvstats_rec_csd_t.f_bg_wait"
	Set gtmtypes("gvstats_rec_csd_t",64,"off")=504
	Set gtmtypes("gvstats_rec_csd_t",64,"len")=8
	Set gtmtypes("gvstats_rec_csd_t",64,"type")="uint64_t"
	Set gtmtypfldindx("gvstats_rec_csd_t","f_bg_wait")=64
	Set gtmtypes("gvstats_rec_csd_t",65,"name")="gvstats_rec_csd_t.f_jnl_wait"
	Set gtmtypes("gvstats_rec_csd_t",65,"off")=512
	Set gtmtypes("gvstats_rec_csd_t",65,"len")=8
	Set gtmtypes("gvstats_rec_csd_t",65,"type")="uint64_t"
	Set gtmtypfldindx("gvstats_rec_csd_t","f_jnl_wait")=65
	Set gtmtypes("gvstats_rec_csd_t",66,"name")="gvstats_rec_csd_t.f_mlk_wait"
	Set gtmtypes("gvstats_rec_csd_t",66,"off")=520
	Set gtmtypes("gvstats_rec_csd_t",66,"len")=8
	Set gtmtypes("gvstats_rec_csd_t",66,"type")="uint64_t"
	Set gtmtypfldindx("gvstats_rec_csd_t","f_mlk_wait")=66
	Set gtmtypes("gvstats_rec_csd_t",67,"name")="gvstats_rec_csd_t.f_proc_wait"
	Set gtmtypes("gvstats_rec_csd_t",67,"off")=528
	Set gtmtypes("gvstats_rec_csd_t",67,"len")=8
	Set gtmtypes("gvstats_rec_csd_t",67,"type")="uint64_t"
	Set gtmtypfldindx("gvstats_rec_csd_t","f_proc_wait")=67
	Set gtmtypes("gvstats_rec_csd_t",68,"name")="gvstats_rec_csd_t.f_trans_wait"
	Set gtmtypes("gvstats_rec_csd_t",68,"off")=536
	Set gtmtypes("gvstats_rec_csd_t",68,"len")=8
	Set gtmtypes("gvstats_rec_csd_t",68,"type")="uint64_t"
	Set gtmtypfldindx("gvstats_rec_csd_t","f_trans_wait")=68
	Set gtmtypes("gvstats_rec_csd_t",69,"name")="gvstats_rec_csd_t.f_util_wait"
	Set gtmtypes("gvstats_rec_csd_t",69,"off")=544
	Set gtmtypes("gvstats_rec_csd_t",69,"len")=8
	Set gtmtypes("gvstats_rec_csd_t",69,"type")="uint64_t"
	Set gtmtypfldindx("gvstats_rec_csd_t","f_util_wait")=69
	Set gtmtypes("gvstats_rec_csd_t",70,"name")="gvstats_rec_csd_t.f_ws2"
	Set gtmtypes("gvstats_rec_csd_t",70,"off")=552
	Set gtmtypes("gvstats_rec_csd_t",70,"len")=8
	Set gtmtypes("gvstats_rec_csd_t",70,"type")="uint64_t"
	Set gtmtypfldindx("gvstats_rec_csd_t","f_ws2")=70
	Set gtmtypes("gvstats_rec_csd_t",71,"name")="gvstats_rec_csd_t.f_ws12"
	Set gtmtypes("gvstats_rec_csd_t",71,"off")=560
	Set gtmtypes("gvstats_rec_csd_t",71,"len")=8
	Set gtmtypes("gvstats_rec_csd_t",71,"type")="uint64_t"
	Set gtmtypfldindx("gvstats_rec_csd_t","f_ws12")=71
	Set gtmtypes("gvstats_rec_csd_t",72,"name")="gvstats_rec_csd_t.f_ws15"
	Set gtmtypes("gvstats_rec_csd_t",72,"off")=568
	Set gtmtypes("gvstats_rec_csd_t",72,"len")=8
	Set gtmtypes("gvstats_rec_csd_t",72,"type")="uint64_t"
	Set gtmtypfldindx("gvstats_rec_csd_t","f_ws15")=72
	Set gtmtypes("gvstats_rec_csd_t",73,"name")="gvstats_rec_csd_t.f_ws39"
	Set gtmtypes("gvstats_rec_csd_t",73,"off")=576
	Set gtmtypes("gvstats_rec_csd_t",73,"len")=8
	Set gtmtypes("gvstats_rec_csd_t",73,"type")="uint64_t"
	Set gtmtypfldindx("gvstats_rec_csd_t","f_ws39")=73
	Set gtmtypes("gvstats_rec_csd_t",74,"name")="gvstats_rec_csd_t.f_ws47"
	Set gtmtypes("gvstats_rec_csd_t",74,"off")=584
	Set gtmtypes("gvstats_rec_csd_t",74,"len")=8
	Set gtmtypes("gvstats_rec_csd_t",74,"type")="uint64_t"
	Set gtmtypfldindx("gvstats_rec_csd_t","f_ws47")=74
	;
	Set gtmtypes("gvstats_rec_t")="struct"
	Set gtmtypes("gvstats_rec_t",0)=74
	Set gtmtypes("gvstats_rec_t","len")=592
	Set gtmtypes("gvstats_rec_t",1,"name")="gvstats_rec_t.n_set"
	Set gtmtypes("gvstats_rec_t",1,"off")=0
	Set gtmtypes("gvstats_rec_t",1,"len")=8
	Set gtmtypes("gvstats_rec_t",1,"type")="uint64_t"
	Set gtmtypfldindx("gvstats_rec_t","n_set")=1
	Set gtmtypes("gvstats_rec_t",2,"name")="gvstats_rec_t.n_kill"
	Set gtmtypes("gvstats_rec_t",2,"off")=8
	Set gtmtypes("gvstats_rec_t",2,"len")=8
	Set gtmtypes("gvstats_rec_t",2,"type")="uint64_t"
	Set gtmtypfldindx("gvstats_rec_t","n_kill")=2
	Set gtmtypes("gvstats_rec_t",3,"name")="gvstats_rec_t.n_get"
	Set gtmtypes("gvstats_rec_t",3,"off")=16
	Set gtmtypes("gvstats_rec_t",3,"len")=8
	Set gtmtypes("gvstats_rec_t",3,"type")="uint64_t"
	Set gtmtypfldindx("gvstats_rec_t","n_get")=3
	Set gtmtypes("gvstats_rec_t",4,"name")="gvstats_rec_t.n_data"
	Set gtmtypes("gvstats_rec_t",4,"off")=24
	Set gtmtypes("gvstats_rec_t",4,"len")=8
	Set gtmtypes("gvstats_rec_t",4,"type")="uint64_t"
	Set gtmtypfldindx("gvstats_rec_t","n_data")=4
	Set gtmtypes("gvstats_rec_t",5,"name")="gvstats_rec_t.n_order"
	Set gtmtypes("gvstats_rec_t",5,"off")=32
	Set gtmtypes("gvstats_rec_t",5,"len")=8
	Set gtmtypes("gvstats_rec_t",5,"type")="uint64_t"
	Set gtmtypfldindx("gvstats_rec_t","n_order")=5
	Set gtmtypes("gvstats_rec_t",6,"name")="gvstats_rec_t.n_zprev"
	Set gtmtypes("gvstats_rec_t",6,"off")=40
	Set gtmtypes("gvstats_rec_t",6,"len")=8
	Set gtmtypes("gvstats_rec_t",6,"type")="uint64_t"
	Set gtmtypfldindx("gvstats_rec_t","n_zprev")=6
	Set gtmtypes("gvstats_rec_t",7,"name")="gvstats_rec_t.n_query"
	Set gtmtypes("gvstats_rec_t",7,"off")=48
	Set gtmtypes("gvstats_rec_t",7,"len")=8
	Set gtmtypes("gvstats_rec_t",7,"type")="uint64_t"
	Set gtmtypfldindx("gvstats_rec_t","n_query")=7
	Set gtmtypes("gvstats_rec_t",8,"name")="gvstats_rec_t.n_lock_success"
	Set gtmtypes("gvstats_rec_t",8,"off")=56
	Set gtmtypes("gvstats_rec_t",8,"len")=8
	Set gtmtypes("gvstats_rec_t",8,"type")="uint64_t"
	Set gtmtypfldindx("gvstats_rec_t","n_lock_success")=8
	Set gtmtypes("gvstats_rec_t",9,"name")="gvstats_rec_t.n_lock_fail"
	Set gtmtypes("gvstats_rec_t",9,"off")=64
	Set gtmtypes("gvstats_rec_t",9,"len")=8
	Set gtmtypes("gvstats_rec_t",9,"type")="uint64_t"
	Set gtmtypfldindx("gvstats_rec_t","n_lock_fail")=9
	Set gtmtypes("gvstats_rec_t",10,"name")="gvstats_rec_t.db_curr_tn"
	Set gtmtypes("gvstats_rec_t",10,"off")=72
	Set gtmtypes("gvstats_rec_t",10,"len")=8
	Set gtmtypes("gvstats_rec_t",10,"type")="uint64_t"
	Set gtmtypfldindx("gvstats_rec_t","db_curr_tn")=10
	Set gtmtypes("gvstats_rec_t",11,"name")="gvstats_rec_t.n_dsk_read"
	Set gtmtypes("gvstats_rec_t",11,"off")=80
	Set gtmtypes("gvstats_rec_t",11,"len")=8
	Set gtmtypes("gvstats_rec_t",11,"type")="uint64_t"
	Set gtmtypfldindx("gvstats_rec_t","n_dsk_read")=11
	Set gtmtypes("gvstats_rec_t",12,"name")="gvstats_rec_t.n_dsk_write"
	Set gtmtypes("gvstats_rec_t",12,"off")=88
	Set gtmtypes("gvstats_rec_t",12,"len")=8
	Set gtmtypes("gvstats_rec_t",12,"type")="uint64_t"
	Set gtmtypfldindx("gvstats_rec_t","n_dsk_write")=12
	Set gtmtypes("gvstats_rec_t",13,"name")="gvstats_rec_t.n_nontp_readwrite"
	Set gtmtypes("gvstats_rec_t",13,"off")=96
	Set gtmtypes("gvstats_rec_t",13,"len")=8
	Set gtmtypes("gvstats_rec_t",13,"type")="uint64_t"
	Set gtmtypfldindx("gvstats_rec_t","n_nontp_readwrite")=13
	Set gtmtypes("gvstats_rec_t",14,"name")="gvstats_rec_t.n_nontp_readonly"
	Set gtmtypes("gvstats_rec_t",14,"off")=104
	Set gtmtypes("gvstats_rec_t",14,"len")=8
	Set gtmtypes("gvstats_rec_t",14,"type")="uint64_t"
	Set gtmtypfldindx("gvstats_rec_t","n_nontp_readonly")=14
	Set gtmtypes("gvstats_rec_t",15,"name")="gvstats_rec_t.n_nontp_blkwrite"
	Set gtmtypes("gvstats_rec_t",15,"off")=112
	Set gtmtypes("gvstats_rec_t",15,"len")=8
	Set gtmtypes("gvstats_rec_t",15,"type")="uint64_t"
	Set gtmtypfldindx("gvstats_rec_t","n_nontp_blkwrite")=15
	Set gtmtypes("gvstats_rec_t",16,"name")="gvstats_rec_t.n_nontp_blkread"
	Set gtmtypes("gvstats_rec_t",16,"off")=120
	Set gtmtypes("gvstats_rec_t",16,"len")=8
	Set gtmtypes("gvstats_rec_t",16,"type")="uint64_t"
	Set gtmtypfldindx("gvstats_rec_t","n_nontp_blkread")=16
	Set gtmtypes("gvstats_rec_t",17,"name")="gvstats_rec_t.n_nontp_retries_0"
	Set gtmtypes("gvstats_rec_t",17,"off")=128
	Set gtmtypes("gvstats_rec_t",17,"len")=8
	Set gtmtypes("gvstats_rec_t",17,"type")="uint64_t"
	Set gtmtypfldindx("gvstats_rec_t","n_nontp_retries_0")=17
	Set gtmtypes("gvstats_rec_t",18,"name")="gvstats_rec_t.n_nontp_retries_1"
	Set gtmtypes("gvstats_rec_t",18,"off")=136
	Set gtmtypes("gvstats_rec_t",18,"len")=8
	Set gtmtypes("gvstats_rec_t",18,"type")="uint64_t"
	Set gtmtypfldindx("gvstats_rec_t","n_nontp_retries_1")=18
	Set gtmtypes("gvstats_rec_t",19,"name")="gvstats_rec_t.n_nontp_retries_2"
	Set gtmtypes("gvstats_rec_t",19,"off")=144
	Set gtmtypes("gvstats_rec_t",19,"len")=8
	Set gtmtypes("gvstats_rec_t",19,"type")="uint64_t"
	Set gtmtypfldindx("gvstats_rec_t","n_nontp_retries_2")=19
	Set gtmtypes("gvstats_rec_t",20,"name")="gvstats_rec_t.n_nontp_retries_3"
	Set gtmtypes("gvstats_rec_t",20,"off")=152
	Set gtmtypes("gvstats_rec_t",20,"len")=8
	Set gtmtypes("gvstats_rec_t",20,"type")="uint64_t"
	Set gtmtypfldindx("gvstats_rec_t","n_nontp_retries_3")=20
	Set gtmtypes("gvstats_rec_t",21,"name")="gvstats_rec_t.n_tp_readwrite"
	Set gtmtypes("gvstats_rec_t",21,"off")=160
	Set gtmtypes("gvstats_rec_t",21,"len")=8
	Set gtmtypes("gvstats_rec_t",21,"type")="uint64_t"
	Set gtmtypfldindx("gvstats_rec_t","n_tp_readwrite")=21
	Set gtmtypes("gvstats_rec_t",22,"name")="gvstats_rec_t.n_tp_readonly"
	Set gtmtypes("gvstats_rec_t",22,"off")=168
	Set gtmtypes("gvstats_rec_t",22,"len")=8
	Set gtmtypes("gvstats_rec_t",22,"type")="uint64_t"
	Set gtmtypfldindx("gvstats_rec_t","n_tp_readonly")=22
	Set gtmtypes("gvstats_rec_t",23,"name")="gvstats_rec_t.n_tp_rolledback"
	Set gtmtypes("gvstats_rec_t",23,"off")=176
	Set gtmtypes("gvstats_rec_t",23,"len")=8
	Set gtmtypes("gvstats_rec_t",23,"type")="uint64_t"
	Set gtmtypfldindx("gvstats_rec_t","n_tp_rolledback")=23
	Set gtmtypes("gvstats_rec_t",24,"name")="gvstats_rec_t.n_tp_blkwrite"
	Set gtmtypes("gvstats_rec_t",24,"off")=184
	Set gtmtypes("gvstats_rec_t",24,"len")=8
	Set gtmtypes("gvstats_rec_t",24,"type")="uint64_t"
	Set gtmtypfldindx("gvstats_rec_t","n_tp_blkwrite")=24
	Set gtmtypes("gvstats_rec_t",25,"name")="gvstats_rec_t.n_tp_blkread"
	Set gtmtypes("gvstats_rec_t",25,"off")=192
	Set gtmtypes("gvstats_rec_t",25,"len")=8
	Set gtmtypes("gvstats_rec_t",25,"type")="uint64_t"
	Set gtmtypfldindx("gvstats_rec_t","n_tp_blkread")=25
	Set gtmtypes("gvstats_rec_t",26,"name")="gvstats_rec_t.n_tp_tot_retries_0"
	Set gtmtypes("gvstats_rec_t",26,"off")=200
	Set gtmtypes("gvstats_rec_t",26,"len")=8
	Set gtmtypes("gvstats_rec_t",26,"type")="uint64_t"
	Set gtmtypfldindx("gvstats_rec_t","n_tp_tot_retries_0")=26
	Set gtmtypes("gvstats_rec_t",27,"name")="gvstats_rec_t.n_tp_tot_retries_1"
	Set gtmtypes("gvstats_rec_t",27,"off")=208
	Set gtmtypes("gvstats_rec_t",27,"len")=8
	Set gtmtypes("gvstats_rec_t",27,"type")="uint64_t"
	Set gtmtypfldindx("gvstats_rec_t","n_tp_tot_retries_1")=27
	Set gtmtypes("gvstats_rec_t",28,"name")="gvstats_rec_t.n_tp_tot_retries_2"
	Set gtmtypes("gvstats_rec_t",28,"off")=216
	Set gtmtypes("gvstats_rec_t",28,"len")=8
	Set gtmtypes("gvstats_rec_t",28,"type")="uint64_t"
	Set gtmtypfldindx("gvstats_rec_t","n_tp_tot_retries_2")=28
	Set gtmtypes("gvstats_rec_t",29,"name")="gvstats_rec_t.n_tp_tot_retries_3"
	Set gtmtypes("gvstats_rec_t",29,"off")=224
	Set gtmtypes("gvstats_rec_t",29,"len")=8
	Set gtmtypes("gvstats_rec_t",29,"type")="uint64_t"
	Set gtmtypfldindx("gvstats_rec_t","n_tp_tot_retries_3")=29
	Set gtmtypes("gvstats_rec_t",30,"name")="gvstats_rec_t.n_tp_tot_retries_4"
	Set gtmtypes("gvstats_rec_t",30,"off")=232
	Set gtmtypes("gvstats_rec_t",30,"len")=8
	Set gtmtypes("gvstats_rec_t",30,"type")="uint64_t"
	Set gtmtypfldindx("gvstats_rec_t","n_tp_tot_retries_4")=30
	Set gtmtypes("gvstats_rec_t",31,"name")="gvstats_rec_t.n_tp_cnflct_retries_0"
	Set gtmtypes("gvstats_rec_t",31,"off")=240
	Set gtmtypes("gvstats_rec_t",31,"len")=8
	Set gtmtypes("gvstats_rec_t",31,"type")="uint64_t"
	Set gtmtypfldindx("gvstats_rec_t","n_tp_cnflct_retries_0")=31
	Set gtmtypes("gvstats_rec_t",32,"name")="gvstats_rec_t.n_tp_cnflct_retries_1"
	Set gtmtypes("gvstats_rec_t",32,"off")=248
	Set gtmtypes("gvstats_rec_t",32,"len")=8
	Set gtmtypes("gvstats_rec_t",32,"type")="uint64_t"
	Set gtmtypfldindx("gvstats_rec_t","n_tp_cnflct_retries_1")=32
	Set gtmtypes("gvstats_rec_t",33,"name")="gvstats_rec_t.n_tp_cnflct_retries_2"
	Set gtmtypes("gvstats_rec_t",33,"off")=256
	Set gtmtypes("gvstats_rec_t",33,"len")=8
	Set gtmtypes("gvstats_rec_t",33,"type")="uint64_t"
	Set gtmtypfldindx("gvstats_rec_t","n_tp_cnflct_retries_2")=33
	Set gtmtypes("gvstats_rec_t",34,"name")="gvstats_rec_t.n_tp_cnflct_retries_3"
	Set gtmtypes("gvstats_rec_t",34,"off")=264
	Set gtmtypes("gvstats_rec_t",34,"len")=8
	Set gtmtypes("gvstats_rec_t",34,"type")="uint64_t"
	Set gtmtypfldindx("gvstats_rec_t","n_tp_cnflct_retries_3")=34
	Set gtmtypes("gvstats_rec_t",35,"name")="gvstats_rec_t.n_tp_cnflct_retries_4"
	Set gtmtypes("gvstats_rec_t",35,"off")=272
	Set gtmtypes("gvstats_rec_t",35,"len")=8
	Set gtmtypes("gvstats_rec_t",35,"type")="uint64_t"
	Set gtmtypfldindx("gvstats_rec_t","n_tp_cnflct_retries_4")=35
	Set gtmtypes("gvstats_rec_t",36,"name")="gvstats_rec_t.n_ztrigger"
	Set gtmtypes("gvstats_rec_t",36,"off")=280
	Set gtmtypes("gvstats_rec_t",36,"len")=8
	Set gtmtypes("gvstats_rec_t",36,"type")="uint64_t"
	Set gtmtypfldindx("gvstats_rec_t","n_ztrigger")=36
	Set gtmtypes("gvstats_rec_t",37,"name")="gvstats_rec_t.n_db_flush"
	Set gtmtypes("gvstats_rec_t",37,"off")=288
	Set gtmtypes("gvstats_rec_t",37,"len")=8
	Set gtmtypes("gvstats_rec_t",37,"type")="uint64_t"
	Set gtmtypfldindx("gvstats_rec_t","n_db_flush")=37
	Set gtmtypes("gvstats_rec_t",38,"name")="gvstats_rec_t.n_db_fsync"
	Set gtmtypes("gvstats_rec_t",38,"off")=296
	Set gtmtypes("gvstats_rec_t",38,"len")=8
	Set gtmtypes("gvstats_rec_t",38,"type")="uint64_t"
	Set gtmtypfldindx("gvstats_rec_t","n_db_fsync")=38
	Set gtmtypes("gvstats_rec_t",39,"name")="gvstats_rec_t.n_jnl_flush"
	Set gtmtypes("gvstats_rec_t",39,"off")=304
	Set gtmtypes("gvstats_rec_t",39,"len")=8
	Set gtmtypes("gvstats_rec_t",39,"type")="uint64_t"
	Set gtmtypfldindx("gvstats_rec_t","n_jnl_flush")=39
	Set gtmtypes("gvstats_rec_t",40,"name")="gvstats_rec_t.n_jnl_fsync"
	Set gtmtypes("gvstats_rec_t",40,"off")=312
	Set gtmtypes("gvstats_rec_t",40,"len")=8
	Set gtmtypes("gvstats_rec_t",40,"type")="uint64_t"
	Set gtmtypfldindx("gvstats_rec_t","n_jnl_fsync")=40
	Set gtmtypes("gvstats_rec_t",41,"name")="gvstats_rec_t.n_jbuff_bytes"
	Set gtmtypes("gvstats_rec_t",41,"off")=320
	Set gtmtypes("gvstats_rec_t",41,"len")=8
	Set gtmtypes("gvstats_rec_t",41,"type")="uint64_t"
	Set gtmtypfldindx("gvstats_rec_t","n_jbuff_bytes")=41
	Set gtmtypes("gvstats_rec_t",42,"name")="gvstats_rec_t.n_jfile_bytes"
	Set gtmtypes("gvstats_rec_t",42,"off")=328
	Set gtmtypes("gvstats_rec_t",42,"len")=8
	Set gtmtypes("gvstats_rec_t",42,"type")="uint64_t"
	Set gtmtypfldindx("gvstats_rec_t","n_jfile_bytes")=42
	Set gtmtypes("gvstats_rec_t",43,"name")="gvstats_rec_t.n_jfile_writes"
	Set gtmtypes("gvstats_rec_t",43,"off")=336
	Set gtmtypes("gvstats_rec_t",43,"len")=8
	Set gtmtypes("gvstats_rec_t",43,"type")="uint64_t"
	Set gtmtypfldindx("gvstats_rec_t","n_jfile_writes")=43
	Set gtmtypes("gvstats_rec_t",44,"name")="gvstats_rec_t.n_jrec_logical"
	Set gtmtypes("gvstats_rec_t",44,"off")=344
	Set gtmtypes("gvstats_rec_t",44,"len")=8
	Set gtmtypes("gvstats_rec_t",44,"type")="uint64_t"
	Set gtmtypfldindx("gvstats_rec_t","n_jrec_logical")=44
	Set gtmtypes("gvstats_rec_t",45,"name")="gvstats_rec_t.n_jrec_pblk"
	Set gtmtypes("gvstats_rec_t",45,"off")=352
	Set gtmtypes("gvstats_rec_t",45,"len")=8
	Set gtmtypes("gvstats_rec_t",45,"type")="uint64_t"
	Set gtmtypfldindx("gvstats_rec_t","n_jrec_pblk")=45
	Set gtmtypes("gvstats_rec_t",46,"name")="gvstats_rec_t.n_jrec_epoch_regular"
	Set gtmtypes("gvstats_rec_t",46,"off")=360
	Set gtmtypes("gvstats_rec_t",46,"len")=8
	Set gtmtypes("gvstats_rec_t",46,"type")="uint64_t"
	Set gtmtypfldindx("gvstats_rec_t","n_jrec_epoch_regular")=46
	Set gtmtypes("gvstats_rec_t",47,"name")="gvstats_rec_t.n_jrec_epoch_idle"
	Set gtmtypes("gvstats_rec_t",47,"off")=368
	Set gtmtypes("gvstats_rec_t",47,"len")=8
	Set gtmtypes("gvstats_rec_t",47,"type")="uint64_t"
	Set gtmtypfldindx("gvstats_rec_t","n_jrec_epoch_idle")=47
	Set gtmtypes("gvstats_rec_t",48,"name")="gvstats_rec_t.n_jrec_other"
	Set gtmtypes("gvstats_rec_t",48,"off")=376
	Set gtmtypes("gvstats_rec_t",48,"len")=8
	Set gtmtypes("gvstats_rec_t",48,"type")="uint64_t"
	Set gtmtypfldindx("gvstats_rec_t","n_jrec_other")=48
	Set gtmtypes("gvstats_rec_t",49,"name")="gvstats_rec_t.n_jnl_extends"
	Set gtmtypes("gvstats_rec_t",49,"off")=384
	Set gtmtypes("gvstats_rec_t",49,"len")=8
	Set gtmtypes("gvstats_rec_t",49,"type")="uint64_t"
	Set gtmtypfldindx("gvstats_rec_t","n_jnl_extends")=49
	Set gtmtypes("gvstats_rec_t",50,"name")="gvstats_rec_t.n_db_extends"
	Set gtmtypes("gvstats_rec_t",50,"off")=392
	Set gtmtypes("gvstats_rec_t",50,"len")=8
	Set gtmtypes("gvstats_rec_t",50,"type")="uint64_t"
	Set gtmtypfldindx("gvstats_rec_t","n_db_extends")=50
	Set gtmtypes("gvstats_rec_t",51,"name")="gvstats_rec_t.n_crit_success"
	Set gtmtypes("gvstats_rec_t",51,"off")=400
	Set gtmtypes("gvstats_rec_t",51,"len")=8
	Set gtmtypes("gvstats_rec_t",51,"type")="uint64_t"
	Set gtmtypfldindx("gvstats_rec_t","n_crit_success")=51
	Set gtmtypes("gvstats_rec_t",52,"name")="gvstats_rec_t.n_crits_in_epch"
	Set gtmtypes("gvstats_rec_t",52,"off")=408
	Set gtmtypes("gvstats_rec_t",52,"len")=8
	Set gtmtypes("gvstats_rec_t",52,"type")="uint64_t"
	Set gtmtypfldindx("gvstats_rec_t","n_crits_in_epch")=52
	Set gtmtypes("gvstats_rec_t",53,"name")="gvstats_rec_t.sq_crit_failed"
	Set gtmtypes("gvstats_rec_t",53,"off")=416
	Set gtmtypes("gvstats_rec_t",53,"len")=8
	Set gtmtypes("gvstats_rec_t",53,"type")="uint64_t"
	Set gtmtypfldindx("gvstats_rec_t","sq_crit_failed")=53
	Set gtmtypes("gvstats_rec_t",54,"name")="gvstats_rec_t.n_crit_failed"
	Set gtmtypes("gvstats_rec_t",54,"off")=424
	Set gtmtypes("gvstats_rec_t",54,"len")=8
	Set gtmtypes("gvstats_rec_t",54,"type")="uint64_t"
	Set gtmtypfldindx("gvstats_rec_t","n_crit_failed")=54
	Set gtmtypes("gvstats_rec_t",55,"name")="gvstats_rec_t.sq_crit_que_slps"
	Set gtmtypes("gvstats_rec_t",55,"off")=432
	Set gtmtypes("gvstats_rec_t",55,"len")=8
	Set gtmtypes("gvstats_rec_t",55,"type")="uint64_t"
	Set gtmtypfldindx("gvstats_rec_t","sq_crit_que_slps")=55
	Set gtmtypes("gvstats_rec_t",56,"name")="gvstats_rec_t.n_crit_que_slps"
	Set gtmtypes("gvstats_rec_t",56,"off")=440
	Set gtmtypes("gvstats_rec_t",56,"len")=8
	Set gtmtypes("gvstats_rec_t",56,"type")="uint64_t"
	Set gtmtypfldindx("gvstats_rec_t","n_crit_que_slps")=56
	Set gtmtypes("gvstats_rec_t",57,"name")="gvstats_rec_t.sq_crit_yields"
	Set gtmtypes("gvstats_rec_t",57,"off")=448
	Set gtmtypes("gvstats_rec_t",57,"len")=8
	Set gtmtypes("gvstats_rec_t",57,"type")="uint64_t"
	Set gtmtypfldindx("gvstats_rec_t","sq_crit_yields")=57
	Set gtmtypes("gvstats_rec_t",58,"name")="gvstats_rec_t.n_crit_yields"
	Set gtmtypes("gvstats_rec_t",58,"off")=456
	Set gtmtypes("gvstats_rec_t",58,"len")=8
	Set gtmtypes("gvstats_rec_t",58,"type")="uint64_t"
	Set gtmtypfldindx("gvstats_rec_t","n_crit_yields")=58
	Set gtmtypes("gvstats_rec_t",59,"name")="gvstats_rec_t.n_clean2dirty"
	Set gtmtypes("gvstats_rec_t",59,"off")=464
	Set gtmtypes("gvstats_rec_t",59,"len")=8
	Set gtmtypes("gvstats_rec_t",59,"type")="uint64_t"
	Set gtmtypfldindx("gvstats_rec_t","n_clean2dirty")=59
	Set gtmtypes("gvstats_rec_t",60,"name")="gvstats_rec_t.n_wait_for_read"
	Set gtmtypes("gvstats_rec_t",60,"off")=472
	Set gtmtypes("gvstats_rec_t",60,"len")=8
	Set gtmtypes("gvstats_rec_t",60,"type")="uint64_t"
	Set gtmtypfldindx("gvstats_rec_t","n_wait_for_read")=60
	Set gtmtypes("gvstats_rec_t",61,"name")="gvstats_rec_t.n_buffer_scarce"
	Set gtmtypes("gvstats_rec_t",61,"off")=480
	Set gtmtypes("gvstats_rec_t",61,"len")=8
	Set gtmtypes("gvstats_rec_t",61,"type")="uint64_t"
	Set gtmtypfldindx("gvstats_rec_t","n_buffer_scarce")=61
	Set gtmtypes("gvstats_rec_t",62,"name")="gvstats_rec_t.n_bt_scarce"
	Set gtmtypes("gvstats_rec_t",62,"off")=488
	Set gtmtypes("gvstats_rec_t",62,"len")=8
	Set gtmtypes("gvstats_rec_t",62,"type")="uint64_t"
	Set gtmtypfldindx("gvstats_rec_t","n_bt_scarce")=62
	Set gtmtypes("gvstats_rec_t",63,"name")="gvstats_rec_t.f_dbext_wait"
	Set gtmtypes("gvstats_rec_t",63,"off")=496
	Set gtmtypes("gvstats_rec_t",63,"len")=8
	Set gtmtypes("gvstats_rec_t",63,"type")="uint64_t"
	Set gtmtypfldindx("gvstats_rec_t","f_dbext_wait")=63
	Set gtmtypes("gvstats_rec_t",64,"name")="gvstats_rec_t.f_bg_wait"
	Set gtmtypes("gvstats_rec_t",64,"off")=504
	Set gtmtypes("gvstats_rec_t",64,"len")=8
	Set gtmtypes("gvstats_rec_t",64,"type")="uint64_t"
	Set gtmtypfldindx("gvstats_rec_t","f_bg_wait")=64
	Set gtmtypes("gvstats_rec_t",65,"name")="gvstats_rec_t.f_jnl_wait"
	Set gtmtypes("gvstats_rec_t",65,"off")=512
	Set gtmtypes("gvstats_rec_t",65,"len")=8
	Set gtmtypes("gvstats_rec_t",65,"type")="uint64_t"
	Set gtmtypfldindx("gvstats_rec_t","f_jnl_wait")=65
	Set gtmtypes("gvstats_rec_t",66,"name")="gvstats_rec_t.f_mlk_wait"
	Set gtmtypes("gvstats_rec_t",66,"off")=520
	Set gtmtypes("gvstats_rec_t",66,"len")=8
	Set gtmtypes("gvstats_rec_t",66,"type")="uint64_t"
	Set gtmtypfldindx("gvstats_rec_t","f_mlk_wait")=66
	Set gtmtypes("gvstats_rec_t",67,"name")="gvstats_rec_t.f_proc_wait"
	Set gtmtypes("gvstats_rec_t",67,"off")=528
	Set gtmtypes("gvstats_rec_t",67,"len")=8
	Set gtmtypes("gvstats_rec_t",67,"type")="uint64_t"
	Set gtmtypfldindx("gvstats_rec_t","f_proc_wait")=67
	Set gtmtypes("gvstats_rec_t",68,"name")="gvstats_rec_t.f_trans_wait"
	Set gtmtypes("gvstats_rec_t",68,"off")=536
	Set gtmtypes("gvstats_rec_t",68,"len")=8
	Set gtmtypes("gvstats_rec_t",68,"type")="uint64_t"
	Set gtmtypfldindx("gvstats_rec_t","f_trans_wait")=68
	Set gtmtypes("gvstats_rec_t",69,"name")="gvstats_rec_t.f_util_wait"
	Set gtmtypes("gvstats_rec_t",69,"off")=544
	Set gtmtypes("gvstats_rec_t",69,"len")=8
	Set gtmtypes("gvstats_rec_t",69,"type")="uint64_t"
	Set gtmtypfldindx("gvstats_rec_t","f_util_wait")=69
	Set gtmtypes("gvstats_rec_t",70,"name")="gvstats_rec_t.f_ws2"
	Set gtmtypes("gvstats_rec_t",70,"off")=552
	Set gtmtypes("gvstats_rec_t",70,"len")=8
	Set gtmtypes("gvstats_rec_t",70,"type")="uint64_t"
	Set gtmtypfldindx("gvstats_rec_t","f_ws2")=70
	Set gtmtypes("gvstats_rec_t",71,"name")="gvstats_rec_t.f_ws12"
	Set gtmtypes("gvstats_rec_t",71,"off")=560
	Set gtmtypes("gvstats_rec_t",71,"len")=8
	Set gtmtypes("gvstats_rec_t",71,"type")="uint64_t"
	Set gtmtypfldindx("gvstats_rec_t","f_ws12")=71
	Set gtmtypes("gvstats_rec_t",72,"name")="gvstats_rec_t.f_ws15"
	Set gtmtypes("gvstats_rec_t",72,"off")=568
	Set gtmtypes("gvstats_rec_t",72,"len")=8
	Set gtmtypes("gvstats_rec_t",72,"type")="uint64_t"
	Set gtmtypfldindx("gvstats_rec_t","f_ws15")=72
	Set gtmtypes("gvstats_rec_t",73,"name")="gvstats_rec_t.f_ws39"
	Set gtmtypes("gvstats_rec_t",73,"off")=576
	Set gtmtypes("gvstats_rec_t",73,"len")=8
	Set gtmtypes("gvstats_rec_t",73,"type")="uint64_t"
	Set gtmtypfldindx("gvstats_rec_t","f_ws39")=73
	Set gtmtypes("gvstats_rec_t",74,"name")="gvstats_rec_t.f_ws47"
	Set gtmtypes("gvstats_rec_t",74,"off")=584
	Set gtmtypes("gvstats_rec_t",74,"len")=8
	Set gtmtypes("gvstats_rec_t",74,"type")="uint64_t"
	Set gtmtypfldindx("gvstats_rec_t","f_ws47")=74
	;
	Set gtmtypes("gvt_container")="struct"
	Set gtmtypes("gvt_container",0)=4
	Set gtmtypes("gvt_container","len")=32
	Set gtmtypes("gvt_container",1,"name")="gvt_container.gvt_ptr"
	Set gtmtypes("gvt_container",1,"off")=0
	Set gtmtypes("gvt_container",1,"len")=8
	Set gtmtypes("gvt_container",1,"type")="addr"
	Set gtmtypfldindx("gvt_container","gvt_ptr")=1
	Set gtmtypes("gvt_container",2,"name")="gvt_container.gvt_ptr2"
	Set gtmtypes("gvt_container",2,"off")=8
	Set gtmtypes("gvt_container",2,"len")=8
	Set gtmtypes("gvt_container",2,"type")="addr"
	Set gtmtypfldindx("gvt_container","gvt_ptr2")=2
	Set gtmtypes("gvt_container",3,"name")="gvt_container.gd_reg"
	Set gtmtypes("gvt_container",3,"off")=16
	Set gtmtypes("gvt_container",3,"len")=8
	Set gtmtypes("gvt_container",3,"type")="addr"
	Set gtmtypfldindx("gvt_container","gd_reg")=3
	Set gtmtypes("gvt_container",4,"name")="gvt_container.next_gvtc"
	Set gtmtypes("gvt_container",4,"off")=24
	Set gtmtypes("gvt_container",4,"len")=8
	Set gtmtypes("gvt_container",4,"type")="addr"
	Set gtmtypfldindx("gvt_container","next_gvtc")=4
	;
	Set gtmtypes("gvt_trigger_t")="struct"
	Set gtmtypes("gvt_trigger_t",0)=9
	Set gtmtypes("gvt_trigger_t","len")=64
	Set gtmtypes("gvt_trigger_t",1,"name")="gvt_trigger_t.gv_trigger_cycle"
	Set gtmtypes("gvt_trigger_t",1,"off")=0
	Set gtmtypes("gvt_trigger_t",1,"len")=4
	Set gtmtypes("gvt_trigger_t",1,"type")="unsigned-int"
	Set gtmtypfldindx("gvt_trigger_t","gv_trigger_cycle")=1
	Set gtmtypes("gvt_trigger_t",2,"name")="gvt_trigger_t.num_gv_triggers"
	Set gtmtypes("gvt_trigger_t",2,"off")=4
	Set gtmtypes("gvt_trigger_t",2,"len")=4
	Set gtmtypes("gvt_trigger_t",2,"type")="unsigned-int"
	Set gtmtypfldindx("gvt_trigger_t","num_gv_triggers")=2
	Set gtmtypes("gvt_trigger_t",3,"name")="gvt_trigger_t.set_triglist"
	Set gtmtypes("gvt_trigger_t",3,"off")=8
	Set gtmtypes("gvt_trigger_t",3,"len")=8
	Set gtmtypes("gvt_trigger_t",3,"type")="addr"
	Set gtmtypfldindx("gvt_trigger_t","set_triglist")=3
	Set gtmtypes("gvt_trigger_t",4,"name")="gvt_trigger_t.kill_triglist"
	Set gtmtypes("gvt_trigger_t",4,"off")=16
	Set gtmtypes("gvt_trigger_t",4,"len")=8
	Set gtmtypes("gvt_trigger_t",4,"type")="addr"
	Set gtmtypfldindx("gvt_trigger_t","kill_triglist")=4
	Set gtmtypes("gvt_trigger_t",5,"name")="gvt_trigger_t.ztrig_triglist"
	Set gtmtypes("gvt_trigger_t",5,"off")=24
	Set gtmtypes("gvt_trigger_t",5,"len")=8
	Set gtmtypes("gvt_trigger_t",5,"type")="addr"
	Set gtmtypfldindx("gvt_trigger_t","ztrig_triglist")=5
	Set gtmtypes("gvt_trigger_t",6,"name")="gvt_trigger_t.gv_target"
	Set gtmtypes("gvt_trigger_t",6,"off")=32
	Set gtmtypes("gvt_trigger_t",6,"len")=8
	Set gtmtypes("gvt_trigger_t",6,"type")="addr"
	Set gtmtypfldindx("gvt_trigger_t","gv_target")=6
	Set gtmtypes("gvt_trigger_t",7,"name")="gvt_trigger_t.gv_trig_top"
	Set gtmtypes("gvt_trigger_t",7,"off")=40
	Set gtmtypes("gvt_trigger_t",7,"len")=8
	Set gtmtypes("gvt_trigger_t",7,"type")="addr"
	Set gtmtypfldindx("gvt_trigger_t","gv_trig_top")=7
	Set gtmtypes("gvt_trigger_t",8,"name")="gvt_trigger_t.gv_trig_list"
	Set gtmtypes("gvt_trigger_t",8,"off")=48
	Set gtmtypes("gvt_trigger_t",8,"len")=8
	Set gtmtypes("gvt_trigger_t",8,"type")="addr"
	Set gtmtypfldindx("gvt_trigger_t","gv_trig_list")=8
	Set gtmtypes("gvt_trigger_t",9,"name")="gvt_trigger_t.gv_trig_array"
	Set gtmtypes("gvt_trigger_t",9,"off")=56
	Set gtmtypes("gvt_trigger_t",9,"len")=8
	Set gtmtypes("gvt_trigger_t",9,"type")="addr"
	Set gtmtypfldindx("gvt_trigger_t","gv_trig_array")=9
	;
	Set gtmtypes("gvtr_invoke_parms_t")="struct"
	Set gtmtypes("gvtr_invoke_parms_t",0)=3
	Set gtmtypes("gvtr_invoke_parms_t","len")=16
	Set gtmtypes("gvtr_invoke_parms_t",1,"name")="gvtr_invoke_parms_t.gvt_trigger"
	Set gtmtypes("gvtr_invoke_parms_t",1,"off")=0
	Set gtmtypes("gvtr_invoke_parms_t",1,"len")=8
	Set gtmtypes("gvtr_invoke_parms_t",1,"type")="addr"
	Set gtmtypfldindx("gvtr_invoke_parms_t","gvt_trigger")=1
	Set gtmtypes("gvtr_invoke_parms_t",2,"name")="gvtr_invoke_parms_t.gvtr_cmd"
	Set gtmtypes("gvtr_invoke_parms_t",2,"off")=8
	Set gtmtypes("gvtr_invoke_parms_t",2,"len")=4
	Set gtmtypes("gvtr_invoke_parms_t",2,"type")="int"
	Set gtmtypfldindx("gvtr_invoke_parms_t","gvtr_cmd")=2
	Set gtmtypes("gvtr_invoke_parms_t",3,"name")="gvtr_invoke_parms_t.num_triggers_invoked"
	Set gtmtypes("gvtr_invoke_parms_t",3,"off")=12
	Set gtmtypes("gvtr_invoke_parms_t",3,"len")=4
	Set gtmtypes("gvtr_invoke_parms_t",3,"type")="int"
	Set gtmtypfldindx("gvtr_invoke_parms_t","num_triggers_invoked")=3
	;
	Set gtmtypes("gvtr_subs_pattern_t")="struct"
	Set gtmtypes("gvtr_subs_pattern_t",0)=9
	Set gtmtypes("gvtr_subs_pattern_t","len")=48
	Set gtmtypes("gvtr_subs_pattern_t",1,"name")="gvtr_subs_pattern_t.gvtr_subs_type"
	Set gtmtypes("gvtr_subs_pattern_t",1,"off")=0
	Set gtmtypes("gvtr_subs_pattern_t",1,"len")=4
	Set gtmtypes("gvtr_subs_pattern_t",1,"type")="unsigned-int"
	Set gtmtypfldindx("gvtr_subs_pattern_t","gvtr_subs_type")=1
	Set gtmtypes("gvtr_subs_pattern_t",2,"name")="gvtr_subs_pattern_t.filler_8byte_align"
	Set gtmtypes("gvtr_subs_pattern_t",2,"off")=4
	Set gtmtypes("gvtr_subs_pattern_t",2,"len")=4
	Set gtmtypes("gvtr_subs_pattern_t",2,"type")="unsigned-int"
	Set gtmtypfldindx("gvtr_subs_pattern_t","filler_8byte_align")=2
	Set gtmtypes("gvtr_subs_pattern_t",3,"name")="gvtr_subs_pattern_t.next_range"
	Set gtmtypes("gvtr_subs_pattern_t",3,"off")=8
	Set gtmtypes("gvtr_subs_pattern_t",3,"len")=8
	Set gtmtypes("gvtr_subs_pattern_t",3,"type")="addr"
	Set gtmtypfldindx("gvtr_subs_pattern_t","next_range")=3
	Set gtmtypes("gvtr_subs_pattern_t",4,"name")="gvtr_subs_pattern_t.pat_mval"
	Set gtmtypes("gvtr_subs_pattern_t",4,"off")=16
	Set gtmtypes("gvtr_subs_pattern_t",4,"len")=32
	Set gtmtypes("gvtr_subs_pattern_t",4,"type")="mval"
	Set gtmtypfldindx("gvtr_subs_pattern_t","pat_mval")=4
	Set gtmtypes("gvtr_subs_pattern_t",5,"name")="gvtr_subs_pattern_t.pat_mval.mvtype"
	Set gtmtypes("gvtr_subs_pattern_t",5,"off")=16
	Set gtmtypes("gvtr_subs_pattern_t",5,"len")=2
	Set gtmtypes("gvtr_subs_pattern_t",5,"type")="unsigned-short"
	Set gtmtypfldindx("gvtr_subs_pattern_t","pat_mval.mvtype")=5
	Set gtmtypes("gvtr_subs_pattern_t",6,"name")="gvtr_subs_pattern_t.pat_mval.fnpc_indx"
	Set gtmtypes("gvtr_subs_pattern_t",6,"off")=19
	Set gtmtypes("gvtr_subs_pattern_t",6,"len")=1
	Set gtmtypes("gvtr_subs_pattern_t",6,"type")="unsigned-char"
	Set gtmtypfldindx("gvtr_subs_pattern_t","pat_mval.fnpc_indx")=6
	Set gtmtypes("gvtr_subs_pattern_t",7,"name")="gvtr_subs_pattern_t.pat_mval.utfcgr_indx"
	Set gtmtypes("gvtr_subs_pattern_t",7,"off")=20
	Set gtmtypes("gvtr_subs_pattern_t",7,"len")=4
	Set gtmtypes("gvtr_subs_pattern_t",7,"type")="unsigned-int"
	Set gtmtypfldindx("gvtr_subs_pattern_t","pat_mval.utfcgr_indx")=7
	Set gtmtypes("gvtr_subs_pattern_t",8,"name")="gvtr_subs_pattern_t.pat_mval.m"
	Set gtmtypes("gvtr_subs_pattern_t",8,"off")=24
	Set gtmtypes("gvtr_subs_pattern_t",8,"len")=8
	Set gtmtypes("gvtr_subs_pattern_t",8,"type")="int"
	Set gtmtypfldindx("gvtr_subs_pattern_t","pat_mval.m")=8
	Set gtmtypes("gvtr_subs_pattern_t",8,"dim")=2
	Set gtmtypes("gvtr_subs_pattern_t",9,"name")="gvtr_subs_pattern_t.pat_mval.str"
	Set gtmtypes("gvtr_subs_pattern_t",9,"off")=32
	Set gtmtypes("gvtr_subs_pattern_t",9,"len")=16
	Set gtmtypes("gvtr_subs_pattern_t",9,"type")="mstr"
	Set gtmtypfldindx("gvtr_subs_pattern_t","pat_mval.str")=9
	;
	Set gtmtypes("gvtr_subs_point_t")="struct"
	Set gtmtypes("gvtr_subs_point_t",0)=4
	Set gtmtypes("gvtr_subs_point_t","len")=24
	Set gtmtypes("gvtr_subs_point_t",1,"name")="gvtr_subs_point_t.gvtr_subs_type"
	Set gtmtypes("gvtr_subs_point_t",1,"off")=0
	Set gtmtypes("gvtr_subs_point_t",1,"len")=4
	Set gtmtypes("gvtr_subs_point_t",1,"type")="unsigned-int"
	Set gtmtypfldindx("gvtr_subs_point_t","gvtr_subs_type")=1
	Set gtmtypes("gvtr_subs_point_t",2,"name")="gvtr_subs_point_t.len"
	Set gtmtypes("gvtr_subs_point_t",2,"off")=4
	Set gtmtypes("gvtr_subs_point_t",2,"len")=4
	Set gtmtypes("gvtr_subs_point_t",2,"type")="unsigned-int"
	Set gtmtypfldindx("gvtr_subs_point_t","len")=2
	Set gtmtypes("gvtr_subs_point_t",3,"name")="gvtr_subs_point_t.next_range"
	Set gtmtypes("gvtr_subs_point_t",3,"off")=8
	Set gtmtypes("gvtr_subs_point_t",3,"len")=8
	Set gtmtypes("gvtr_subs_point_t",3,"type")="addr"
	Set gtmtypfldindx("gvtr_subs_point_t","next_range")=3
	Set gtmtypes("gvtr_subs_point_t",4,"name")="gvtr_subs_point_t.subs_key"
	Set gtmtypes("gvtr_subs_point_t",4,"off")=16
	Set gtmtypes("gvtr_subs_point_t",4,"len")=8
	Set gtmtypes("gvtr_subs_point_t",4,"type")="addr"
	Set gtmtypfldindx("gvtr_subs_point_t","subs_key")=4
	;
	Set gtmtypes("gvtr_subs_range_t")="struct"
	Set gtmtypes("gvtr_subs_range_t",0)=6
	Set gtmtypes("gvtr_subs_range_t","len")=40
	Set gtmtypes("gvtr_subs_range_t",1,"name")="gvtr_subs_range_t.gvtr_subs_type"
	Set gtmtypes("gvtr_subs_range_t",1,"off")=0
	Set gtmtypes("gvtr_subs_range_t",1,"len")=4
	Set gtmtypes("gvtr_subs_range_t",1,"type")="unsigned-int"
	Set gtmtypfldindx("gvtr_subs_range_t","gvtr_subs_type")=1
	Set gtmtypes("gvtr_subs_range_t",2,"name")="gvtr_subs_range_t.len1"
	Set gtmtypes("gvtr_subs_range_t",2,"off")=4
	Set gtmtypes("gvtr_subs_range_t",2,"len")=4
	Set gtmtypes("gvtr_subs_range_t",2,"type")="unsigned-int"
	Set gtmtypfldindx("gvtr_subs_range_t","len1")=2
	Set gtmtypes("gvtr_subs_range_t",3,"name")="gvtr_subs_range_t.next_range"
	Set gtmtypes("gvtr_subs_range_t",3,"off")=8
	Set gtmtypes("gvtr_subs_range_t",3,"len")=8
	Set gtmtypes("gvtr_subs_range_t",3,"type")="addr"
	Set gtmtypfldindx("gvtr_subs_range_t","next_range")=3
	Set gtmtypes("gvtr_subs_range_t",4,"name")="gvtr_subs_range_t.subs_key1"
	Set gtmtypes("gvtr_subs_range_t",4,"off")=16
	Set gtmtypes("gvtr_subs_range_t",4,"len")=8
	Set gtmtypes("gvtr_subs_range_t",4,"type")="addr"
	Set gtmtypfldindx("gvtr_subs_range_t","subs_key1")=4
	Set gtmtypes("gvtr_subs_range_t",5,"name")="gvtr_subs_range_t.len2"
	Set gtmtypes("gvtr_subs_range_t",5,"off")=24
	Set gtmtypes("gvtr_subs_range_t",5,"len")=4
	Set gtmtypes("gvtr_subs_range_t",5,"type")="unsigned-int"
	Set gtmtypfldindx("gvtr_subs_range_t","len2")=5
	Set gtmtypes("gvtr_subs_range_t",6,"name")="gvtr_subs_range_t.subs_key2"
	Set gtmtypes("gvtr_subs_range_t",6,"off")=32
	Set gtmtypes("gvtr_subs_range_t",6,"len")=8
	Set gtmtypes("gvtr_subs_range_t",6,"type")="addr"
	Set gtmtypfldindx("gvtr_subs_range_t","subs_key2")=6
	;
	Set gtmtypes("gvtr_subs_star_t")="struct"
	Set gtmtypes("gvtr_subs_star_t",0)=3
	Set gtmtypes("gvtr_subs_star_t","len")=16
	Set gtmtypes("gvtr_subs_star_t",1,"name")="gvtr_subs_star_t.gvtr_subs_type"
	Set gtmtypes("gvtr_subs_star_t",1,"off")=0
	Set gtmtypes("gvtr_subs_star_t",1,"len")=4
	Set gtmtypes("gvtr_subs_star_t",1,"type")="unsigned-int"
	Set gtmtypfldindx("gvtr_subs_star_t","gvtr_subs_type")=1
	Set gtmtypes("gvtr_subs_star_t",2,"name")="gvtr_subs_star_t.filler_8byte_align"
	Set gtmtypes("gvtr_subs_star_t",2,"off")=4
	Set gtmtypes("gvtr_subs_star_t",2,"len")=4
	Set gtmtypes("gvtr_subs_star_t",2,"type")="unsigned-int"
	Set gtmtypfldindx("gvtr_subs_star_t","filler_8byte_align")=2
	Set gtmtypes("gvtr_subs_star_t",3,"name")="gvtr_subs_star_t.next_range"
	Set gtmtypes("gvtr_subs_star_t",3,"off")=8
	Set gtmtypes("gvtr_subs_star_t",3,"len")=8
	Set gtmtypes("gvtr_subs_star_t",3,"type")="addr"
	Set gtmtypfldindx("gvtr_subs_star_t","next_range")=3
	;
	Set gtmtypes("gvtr_subs_t")="union"
	Set gtmtypes("gvtr_subs_t",0)=27
	Set gtmtypes("gvtr_subs_t","len")=48
	Set gtmtypes("gvtr_subs_t",1,"name")="gvtr_subs_t.gvtr_subs_type"
	Set gtmtypes("gvtr_subs_t",1,"off")=0
	Set gtmtypes("gvtr_subs_t",1,"len")=4
	Set gtmtypes("gvtr_subs_t",1,"type")="unsigned-int"
	Set gtmtypfldindx("gvtr_subs_t","gvtr_subs_type")=1
	Set gtmtypes("gvtr_subs_t",2,"name")="gvtr_subs_t.gvtr_subs_star"
	Set gtmtypes("gvtr_subs_t",2,"off")=0
	Set gtmtypes("gvtr_subs_t",2,"len")=16
	Set gtmtypes("gvtr_subs_t",2,"type")="gvtr_subs_star_t"
	Set gtmtypfldindx("gvtr_subs_t","gvtr_subs_star")=2
	Set gtmtypes("gvtr_subs_t",3,"name")="gvtr_subs_t.gvtr_subs_star.gvtr_subs_type"
	Set gtmtypes("gvtr_subs_t",3,"off")=0
	Set gtmtypes("gvtr_subs_t",3,"len")=4
	Set gtmtypes("gvtr_subs_t",3,"type")="unsigned-int"
	Set gtmtypfldindx("gvtr_subs_t","gvtr_subs_star.gvtr_subs_type")=3
	Set gtmtypes("gvtr_subs_t",4,"name")="gvtr_subs_t.gvtr_subs_star.filler_8byte_align"
	Set gtmtypes("gvtr_subs_t",4,"off")=4
	Set gtmtypes("gvtr_subs_t",4,"len")=4
	Set gtmtypes("gvtr_subs_t",4,"type")="unsigned-int"
	Set gtmtypfldindx("gvtr_subs_t","gvtr_subs_star.filler_8byte_align")=4
	Set gtmtypes("gvtr_subs_t",5,"name")="gvtr_subs_t.gvtr_subs_star.next_range"
	Set gtmtypes("gvtr_subs_t",5,"off")=8
	Set gtmtypes("gvtr_subs_t",5,"len")=8
	Set gtmtypes("gvtr_subs_t",5,"type")="addr"
	Set gtmtypfldindx("gvtr_subs_t","gvtr_subs_star.next_range")=5
	Set gtmtypes("gvtr_subs_t",6,"name")="gvtr_subs_t.gvtr_subs_point"
	Set gtmtypes("gvtr_subs_t",6,"off")=0
	Set gtmtypes("gvtr_subs_t",6,"len")=24
	Set gtmtypes("gvtr_subs_t",6,"type")="gvtr_subs_point_t"
	Set gtmtypfldindx("gvtr_subs_t","gvtr_subs_point")=6
	Set gtmtypes("gvtr_subs_t",7,"name")="gvtr_subs_t.gvtr_subs_point.gvtr_subs_type"
	Set gtmtypes("gvtr_subs_t",7,"off")=0
	Set gtmtypes("gvtr_subs_t",7,"len")=4
	Set gtmtypes("gvtr_subs_t",7,"type")="unsigned-int"
	Set gtmtypfldindx("gvtr_subs_t","gvtr_subs_point.gvtr_subs_type")=7
	Set gtmtypes("gvtr_subs_t",8,"name")="gvtr_subs_t.gvtr_subs_point.len"
	Set gtmtypes("gvtr_subs_t",8,"off")=4
	Set gtmtypes("gvtr_subs_t",8,"len")=4
	Set gtmtypes("gvtr_subs_t",8,"type")="unsigned-int"
	Set gtmtypfldindx("gvtr_subs_t","gvtr_subs_point.len")=8
	Set gtmtypes("gvtr_subs_t",9,"name")="gvtr_subs_t.gvtr_subs_point.next_range"
	Set gtmtypes("gvtr_subs_t",9,"off")=8
	Set gtmtypes("gvtr_subs_t",9,"len")=8
	Set gtmtypes("gvtr_subs_t",9,"type")="addr"
	Set gtmtypfldindx("gvtr_subs_t","gvtr_subs_point.next_range")=9
	Set gtmtypes("gvtr_subs_t",10,"name")="gvtr_subs_t.gvtr_subs_point.subs_key"
	Set gtmtypes("gvtr_subs_t",10,"off")=16
	Set gtmtypes("gvtr_subs_t",10,"len")=8
	Set gtmtypes("gvtr_subs_t",10,"type")="addr"
	Set gtmtypfldindx("gvtr_subs_t","gvtr_subs_point.subs_key")=10
	Set gtmtypes("gvtr_subs_t",11,"name")="gvtr_subs_t.gvtr_subs_range"
	Set gtmtypes("gvtr_subs_t",11,"off")=0
	Set gtmtypes("gvtr_subs_t",11,"len")=40
	Set gtmtypes("gvtr_subs_t",11,"type")="gvtr_subs_range_t"
	Set gtmtypfldindx("gvtr_subs_t","gvtr_subs_range")=11
	Set gtmtypes("gvtr_subs_t",12,"name")="gvtr_subs_t.gvtr_subs_range.gvtr_subs_type"
	Set gtmtypes("gvtr_subs_t",12,"off")=0
	Set gtmtypes("gvtr_subs_t",12,"len")=4
	Set gtmtypes("gvtr_subs_t",12,"type")="unsigned-int"
	Set gtmtypfldindx("gvtr_subs_t","gvtr_subs_range.gvtr_subs_type")=12
	Set gtmtypes("gvtr_subs_t",13,"name")="gvtr_subs_t.gvtr_subs_range.len1"
	Set gtmtypes("gvtr_subs_t",13,"off")=4
	Set gtmtypes("gvtr_subs_t",13,"len")=4
	Set gtmtypes("gvtr_subs_t",13,"type")="unsigned-int"
	Set gtmtypfldindx("gvtr_subs_t","gvtr_subs_range.len1")=13
	Set gtmtypes("gvtr_subs_t",14,"name")="gvtr_subs_t.gvtr_subs_range.next_range"
	Set gtmtypes("gvtr_subs_t",14,"off")=8
	Set gtmtypes("gvtr_subs_t",14,"len")=8
	Set gtmtypes("gvtr_subs_t",14,"type")="addr"
	Set gtmtypfldindx("gvtr_subs_t","gvtr_subs_range.next_range")=14
	Set gtmtypes("gvtr_subs_t",15,"name")="gvtr_subs_t.gvtr_subs_range.subs_key1"
	Set gtmtypes("gvtr_subs_t",15,"off")=16
	Set gtmtypes("gvtr_subs_t",15,"len")=8
	Set gtmtypes("gvtr_subs_t",15,"type")="addr"
	Set gtmtypfldindx("gvtr_subs_t","gvtr_subs_range.subs_key1")=15
	Set gtmtypes("gvtr_subs_t",16,"name")="gvtr_subs_t.gvtr_subs_range.len2"
	Set gtmtypes("gvtr_subs_t",16,"off")=24
	Set gtmtypes("gvtr_subs_t",16,"len")=4
	Set gtmtypes("gvtr_subs_t",16,"type")="unsigned-int"
	Set gtmtypfldindx("gvtr_subs_t","gvtr_subs_range.len2")=16
	Set gtmtypes("gvtr_subs_t",17,"name")="gvtr_subs_t.gvtr_subs_range.subs_key2"
	Set gtmtypes("gvtr_subs_t",17,"off")=32
	Set gtmtypes("gvtr_subs_t",17,"len")=8
	Set gtmtypes("gvtr_subs_t",17,"type")="addr"
	Set gtmtypfldindx("gvtr_subs_t","gvtr_subs_range.subs_key2")=17
	Set gtmtypes("gvtr_subs_t",18,"name")="gvtr_subs_t.gvtr_subs_pattern"
	Set gtmtypes("gvtr_subs_t",18,"off")=0
	Set gtmtypes("gvtr_subs_t",18,"len")=48
	Set gtmtypes("gvtr_subs_t",18,"type")="gvtr_subs_pattern_t"
	Set gtmtypfldindx("gvtr_subs_t","gvtr_subs_pattern")=18
	Set gtmtypes("gvtr_subs_t",19,"name")="gvtr_subs_t.gvtr_subs_pattern.gvtr_subs_type"
	Set gtmtypes("gvtr_subs_t",19,"off")=0
	Set gtmtypes("gvtr_subs_t",19,"len")=4
	Set gtmtypes("gvtr_subs_t",19,"type")="unsigned-int"
	Set gtmtypfldindx("gvtr_subs_t","gvtr_subs_pattern.gvtr_subs_type")=19
	Set gtmtypes("gvtr_subs_t",20,"name")="gvtr_subs_t.gvtr_subs_pattern.filler_8byte_align"
	Set gtmtypes("gvtr_subs_t",20,"off")=4
	Set gtmtypes("gvtr_subs_t",20,"len")=4
	Set gtmtypes("gvtr_subs_t",20,"type")="unsigned-int"
	Set gtmtypfldindx("gvtr_subs_t","gvtr_subs_pattern.filler_8byte_align")=20
	Set gtmtypes("gvtr_subs_t",21,"name")="gvtr_subs_t.gvtr_subs_pattern.next_range"
	Set gtmtypes("gvtr_subs_t",21,"off")=8
	Set gtmtypes("gvtr_subs_t",21,"len")=8
	Set gtmtypes("gvtr_subs_t",21,"type")="addr"
	Set gtmtypfldindx("gvtr_subs_t","gvtr_subs_pattern.next_range")=21
	Set gtmtypes("gvtr_subs_t",22,"name")="gvtr_subs_t.gvtr_subs_pattern.pat_mval"
	Set gtmtypes("gvtr_subs_t",22,"off")=16
	Set gtmtypes("gvtr_subs_t",22,"len")=32
	Set gtmtypes("gvtr_subs_t",22,"type")="mval"
	Set gtmtypfldindx("gvtr_subs_t","gvtr_subs_pattern.pat_mval")=22
	Set gtmtypes("gvtr_subs_t",23,"name")="gvtr_subs_t.gvtr_subs_pattern.pat_mval.mvtype"
	Set gtmtypes("gvtr_subs_t",23,"off")=16
	Set gtmtypes("gvtr_subs_t",23,"len")=2
	Set gtmtypes("gvtr_subs_t",23,"type")="unsigned-short"
	Set gtmtypfldindx("gvtr_subs_t","gvtr_subs_pattern.pat_mval.mvtype")=23
	Set gtmtypes("gvtr_subs_t",24,"name")="gvtr_subs_t.gvtr_subs_pattern.pat_mval.fnpc_indx"
	Set gtmtypes("gvtr_subs_t",24,"off")=19
	Set gtmtypes("gvtr_subs_t",24,"len")=1
	Set gtmtypes("gvtr_subs_t",24,"type")="unsigned-char"
	Set gtmtypfldindx("gvtr_subs_t","gvtr_subs_pattern.pat_mval.fnpc_indx")=24
	Set gtmtypes("gvtr_subs_t",25,"name")="gvtr_subs_t.gvtr_subs_pattern.pat_mval.utfcgr_indx"
	Set gtmtypes("gvtr_subs_t",25,"off")=20
	Set gtmtypes("gvtr_subs_t",25,"len")=4
	Set gtmtypes("gvtr_subs_t",25,"type")="unsigned-int"
	Set gtmtypfldindx("gvtr_subs_t","gvtr_subs_pattern.pat_mval.utfcgr_indx")=25
	Set gtmtypes("gvtr_subs_t",26,"name")="gvtr_subs_t.gvtr_subs_pattern.pat_mval.m"
	Set gtmtypes("gvtr_subs_t",26,"off")=24
	Set gtmtypes("gvtr_subs_t",26,"len")=8
	Set gtmtypes("gvtr_subs_t",26,"type")="int"
	Set gtmtypfldindx("gvtr_subs_t","gvtr_subs_pattern.pat_mval.m")=26
	Set gtmtypes("gvtr_subs_t",26,"dim")=2
	Set gtmtypes("gvtr_subs_t",27,"name")="gvtr_subs_t.gvtr_subs_pattern.pat_mval.str"
	Set gtmtypes("gvtr_subs_t",27,"off")=32
	Set gtmtypes("gvtr_subs_t",27,"len")=16
	Set gtmtypes("gvtr_subs_t",27,"type")="mstr"
	Set gtmtypfldindx("gvtr_subs_t","gvtr_subs_pattern.pat_mval.str")=27
	;
	Set gtmtypes("gvzwrite_datablk")="struct"
	Set gtmtypes("gvzwrite_datablk",0)=12
	Set gtmtypes("gvzwrite_datablk","len")=64
	Set gtmtypes("gvzwrite_datablk",1,"name")="gvzwrite_datablk.type"
	Set gtmtypes("gvzwrite_datablk",1,"off")=0
	Set gtmtypes("gvzwrite_datablk",1,"len")=4
	Set gtmtypes("gvzwrite_datablk",1,"type")="boolean_t"
	Set gtmtypfldindx("gvzwrite_datablk","type")=1
	Set gtmtypes("gvzwrite_datablk",2,"name")="gvzwrite_datablk.subsc_count"
	Set gtmtypes("gvzwrite_datablk",2,"off")=4
	Set gtmtypes("gvzwrite_datablk",2,"len")=2
	Set gtmtypes("gvzwrite_datablk",2,"type")="unsigned-short"
	Set gtmtypfldindx("gvzwrite_datablk","subsc_count")=2
	Set gtmtypes("gvzwrite_datablk",3,"name")="gvzwrite_datablk.curr_subsc"
	Set gtmtypes("gvzwrite_datablk",3,"off")=6
	Set gtmtypes("gvzwrite_datablk",3,"len")=2
	Set gtmtypes("gvzwrite_datablk",3,"type")="unsigned-short"
	Set gtmtypfldindx("gvzwrite_datablk","curr_subsc")=3
	Set gtmtypes("gvzwrite_datablk",4,"name")="gvzwrite_datablk.fixed"
	Set gtmtypes("gvzwrite_datablk",4,"off")=8
	Set gtmtypes("gvzwrite_datablk",4,"len")=4
	Set gtmtypes("gvzwrite_datablk",4,"type")="boolean_t"
	Set gtmtypfldindx("gvzwrite_datablk","fixed")=4
	Set gtmtypes("gvzwrite_datablk",5,"name")="gvzwrite_datablk.mask"
	Set gtmtypes("gvzwrite_datablk",5,"off")=12
	Set gtmtypes("gvzwrite_datablk",5,"len")=4
	Set gtmtypes("gvzwrite_datablk",5,"type")="unsigned-int"
	Set gtmtypfldindx("gvzwrite_datablk","mask")=5
	Set gtmtypes("gvzwrite_datablk",6,"name")="gvzwrite_datablk.pat"
	Set gtmtypes("gvzwrite_datablk",6,"off")=16
	Set gtmtypes("gvzwrite_datablk",6,"len")=8
	Set gtmtypes("gvzwrite_datablk",6,"type")="addr"
	Set gtmtypfldindx("gvzwrite_datablk","pat")=6
	Set gtmtypes("gvzwrite_datablk",7,"name")="gvzwrite_datablk.old_key"
	Set gtmtypes("gvzwrite_datablk",7,"off")=24
	Set gtmtypes("gvzwrite_datablk",7,"len")=8
	Set gtmtypes("gvzwrite_datablk",7,"type")="addr"
	Set gtmtypfldindx("gvzwrite_datablk","old_key")=7
	Set gtmtypes("gvzwrite_datablk",8,"name")="gvzwrite_datablk.old_targ"
	Set gtmtypes("gvzwrite_datablk",8,"off")=32
	Set gtmtypes("gvzwrite_datablk",8,"len")=8
	Set gtmtypes("gvzwrite_datablk",8,"type")="addr"
	Set gtmtypfldindx("gvzwrite_datablk","old_targ")=8
	Set gtmtypes("gvzwrite_datablk",9,"name")="gvzwrite_datablk.sub"
	Set gtmtypes("gvzwrite_datablk",9,"off")=40
	Set gtmtypes("gvzwrite_datablk",9,"len")=8
	Set gtmtypes("gvzwrite_datablk",9,"type")="addr"
	Set gtmtypfldindx("gvzwrite_datablk","sub")=9
	Set gtmtypes("gvzwrite_datablk",10,"name")="gvzwrite_datablk.gd_reg"
	Set gtmtypes("gvzwrite_datablk",10,"off")=48
	Set gtmtypes("gvzwrite_datablk",10,"len")=8
	Set gtmtypes("gvzwrite_datablk",10,"type")="addr"
	Set gtmtypfldindx("gvzwrite_datablk","gd_reg")=10
	Set gtmtypes("gvzwrite_datablk",11,"name")="gvzwrite_datablk.gv_last_subsc_null"
	Set gtmtypes("gvzwrite_datablk",11,"off")=56
	Set gtmtypes("gvzwrite_datablk",11,"len")=4
	Set gtmtypes("gvzwrite_datablk",11,"type")="boolean_t"
	Set gtmtypfldindx("gvzwrite_datablk","gv_last_subsc_null")=11
	Set gtmtypes("gvzwrite_datablk",12,"name")="gvzwrite_datablk.gv_some_subsc_null"
	Set gtmtypes("gvzwrite_datablk",12,"off")=60
	Set gtmtypes("gvzwrite_datablk",12,"len")=4
	Set gtmtypes("gvzwrite_datablk",12,"type")="boolean_t"
	Set gtmtypfldindx("gvzwrite_datablk","gv_some_subsc_null")=12
	;
	Set gtmtypes("hash128_state_t")="struct"
	Set gtmtypes("hash128_state_t",0)=7
	Set gtmtypes("hash128_state_t","len")=40
	Set gtmtypes("hash128_state_t",1,"name")="hash128_state_t.h"
	Set gtmtypes("hash128_state_t",1,"off")=0
	Set gtmtypes("hash128_state_t",1,"len")=16
	Set gtmtypes("hash128_state_t",1,"type")="gtm_uint16"
	Set gtmtypfldindx("hash128_state_t","h")=1
	Set gtmtypes("hash128_state_t",2,"name")="hash128_state_t.h.one"
	Set gtmtypes("hash128_state_t",2,"off")=0
	Set gtmtypes("hash128_state_t",2,"len")=8
	Set gtmtypes("hash128_state_t",2,"type")="gtm_uint8"
	Set gtmtypfldindx("hash128_state_t","h.one")=2
	Set gtmtypes("hash128_state_t",3,"name")="hash128_state_t.h.two"
	Set gtmtypes("hash128_state_t",3,"off")=8
	Set gtmtypes("hash128_state_t",3,"len")=8
	Set gtmtypes("hash128_state_t",3,"type")="gtm_uint8"
	Set gtmtypfldindx("hash128_state_t","h.two")=3
	Set gtmtypes("hash128_state_t",4,"name")="hash128_state_t.c"
	Set gtmtypes("hash128_state_t",4,"off")=16
	Set gtmtypes("hash128_state_t",4,"len")=16
	Set gtmtypes("hash128_state_t",4,"type")="gtm_uint16"
	Set gtmtypfldindx("hash128_state_t","c")=4
	Set gtmtypes("hash128_state_t",5,"name")="hash128_state_t.c.one"
	Set gtmtypes("hash128_state_t",5,"off")=16
	Set gtmtypes("hash128_state_t",5,"len")=8
	Set gtmtypes("hash128_state_t",5,"type")="gtm_uint8"
	Set gtmtypfldindx("hash128_state_t","c.one")=5
	Set gtmtypes("hash128_state_t",6,"name")="hash128_state_t.c.two"
	Set gtmtypes("hash128_state_t",6,"off")=24
	Set gtmtypes("hash128_state_t",6,"len")=8
	Set gtmtypes("hash128_state_t",6,"type")="gtm_uint8"
	Set gtmtypfldindx("hash128_state_t","c.two")=6
	Set gtmtypes("hash128_state_t",7,"name")="hash128_state_t.carry_bytes"
	Set gtmtypes("hash128_state_t",7,"off")=32
	Set gtmtypes("hash128_state_t",7,"len")=4
	Set gtmtypes("hash128_state_t",7,"type")="int"
	Set gtmtypfldindx("hash128_state_t","carry_bytes")=7
	;
	Set gtmtypes("hash_table_addr")="struct"
	Set gtmtypes("hash_table_addr",0)=15
	Set gtmtypes("hash_table_addr","len")=80
	Set gtmtypes("hash_table_addr",1,"name")="hash_table_addr.base"
	Set gtmtypes("hash_table_addr",1,"off")=0
	Set gtmtypes("hash_table_addr",1,"len")=8
	Set gtmtypes("hash_table_addr",1,"type")="addr"
	Set gtmtypfldindx("hash_table_addr","base")=1
	Set gtmtypes("hash_table_addr",2,"name")="hash_table_addr.top"
	Set gtmtypes("hash_table_addr",2,"off")=8
	Set gtmtypes("hash_table_addr",2,"len")=8
	Set gtmtypes("hash_table_addr",2,"type")="addr"
	Set gtmtypfldindx("hash_table_addr","top")=2
	Set gtmtypes("hash_table_addr",3,"name")="hash_table_addr.spare_base"
	Set gtmtypes("hash_table_addr",3,"off")=16
	Set gtmtypes("hash_table_addr",3,"len")=8
	Set gtmtypes("hash_table_addr",3,"type")="addr"
	Set gtmtypfldindx("hash_table_addr","spare_base")=3
	Set gtmtypes("hash_table_addr",4,"name")="hash_table_addr.entry_passed_thru"
	Set gtmtypes("hash_table_addr",4,"off")=24
	Set gtmtypes("hash_table_addr",4,"len")=8
	Set gtmtypes("hash_table_addr",4,"type")="addr"
	Set gtmtypfldindx("hash_table_addr","entry_passed_thru")=4
	Set gtmtypes("hash_table_addr",5,"name")="hash_table_addr.size"
	Set gtmtypes("hash_table_addr",5,"off")=32
	Set gtmtypes("hash_table_addr",5,"len")=4
	Set gtmtypes("hash_table_addr",5,"type")="unsigned-int"
	Set gtmtypfldindx("hash_table_addr","size")=5
	Set gtmtypes("hash_table_addr",6,"name")="hash_table_addr.initial_size"
	Set gtmtypes("hash_table_addr",6,"off")=36
	Set gtmtypes("hash_table_addr",6,"len")=4
	Set gtmtypes("hash_table_addr",6,"type")="unsigned-int"
	Set gtmtypfldindx("hash_table_addr","initial_size")=6
	Set gtmtypes("hash_table_addr",7,"name")="hash_table_addr.spare_base_size"
	Set gtmtypes("hash_table_addr",7,"off")=40
	Set gtmtypes("hash_table_addr",7,"len")=4
	Set gtmtypes("hash_table_addr",7,"type")="unsigned-int"
	Set gtmtypfldindx("hash_table_addr","spare_base_size")=7
	Set gtmtypes("hash_table_addr",8,"name")="hash_table_addr.count"
	Set gtmtypes("hash_table_addr",8,"off")=44
	Set gtmtypes("hash_table_addr",8,"len")=4
	Set gtmtypes("hash_table_addr",8,"type")="unsigned-int"
	Set gtmtypfldindx("hash_table_addr","count")=8
	Set gtmtypes("hash_table_addr",9,"name")="hash_table_addr.del_count"
	Set gtmtypes("hash_table_addr",9,"off")=48
	Set gtmtypes("hash_table_addr",9,"len")=4
	Set gtmtypes("hash_table_addr",9,"type")="unsigned-int"
	Set gtmtypfldindx("hash_table_addr","del_count")=9
	Set gtmtypes("hash_table_addr",10,"name")="hash_table_addr.exp_trigger_size"
	Set gtmtypes("hash_table_addr",10,"off")=52
	Set gtmtypes("hash_table_addr",10,"len")=4
	Set gtmtypes("hash_table_addr",10,"type")="unsigned-int"
	Set gtmtypfldindx("hash_table_addr","exp_trigger_size")=10
	Set gtmtypes("hash_table_addr",11,"name")="hash_table_addr.cmp_trigger_size"
	Set gtmtypes("hash_table_addr",11,"off")=56
	Set gtmtypes("hash_table_addr",11,"len")=4
	Set gtmtypes("hash_table_addr",11,"type")="unsigned-int"
	Set gtmtypfldindx("hash_table_addr","cmp_trigger_size")=11
	Set gtmtypes("hash_table_addr",12,"name")="hash_table_addr.dont_compact"
	Set gtmtypes("hash_table_addr",12,"off")=60
	Set gtmtypes("hash_table_addr",12,"len")=4
	Set gtmtypes("hash_table_addr",12,"type")="boolean_t"
	Set gtmtypfldindx("hash_table_addr","dont_compact")=12
	Set gtmtypes("hash_table_addr",13,"name")="hash_table_addr.dont_keep_spare_table"
	Set gtmtypes("hash_table_addr",13,"off")=64
	Set gtmtypes("hash_table_addr",13,"len")=4
	Set gtmtypes("hash_table_addr",13,"type")="boolean_t"
	Set gtmtypfldindx("hash_table_addr","dont_keep_spare_table")=13
	Set gtmtypes("hash_table_addr",14,"name")="hash_table_addr.defer_base_release"
	Set gtmtypes("hash_table_addr",14,"off")=68
	Set gtmtypes("hash_table_addr",14,"len")=4
	Set gtmtypes("hash_table_addr",14,"type")="boolean_t"
	Set gtmtypfldindx("hash_table_addr","defer_base_release")=14
	Set gtmtypes("hash_table_addr",15,"name")="hash_table_addr.active"
	Set gtmtypes("hash_table_addr",15,"off")=72
	Set gtmtypes("hash_table_addr",15,"len")=4
	Set gtmtypes("hash_table_addr",15,"type")="boolean_t"
	Set gtmtypfldindx("hash_table_addr","active")=15
	;
	Set gtmtypes("hash_table_int4")="struct"
	Set gtmtypes("hash_table_int4",0)=15
	Set gtmtypes("hash_table_int4","len")=80
	Set gtmtypes("hash_table_int4",1,"name")="hash_table_int4.base"
	Set gtmtypes("hash_table_int4",1,"off")=0
	Set gtmtypes("hash_table_int4",1,"len")=8
	Set gtmtypes("hash_table_int4",1,"type")="addr"
	Set gtmtypfldindx("hash_table_int4","base")=1
	Set gtmtypes("hash_table_int4",2,"name")="hash_table_int4.top"
	Set gtmtypes("hash_table_int4",2,"off")=8
	Set gtmtypes("hash_table_int4",2,"len")=8
	Set gtmtypes("hash_table_int4",2,"type")="addr"
	Set gtmtypfldindx("hash_table_int4","top")=2
	Set gtmtypes("hash_table_int4",3,"name")="hash_table_int4.spare_base"
	Set gtmtypes("hash_table_int4",3,"off")=16
	Set gtmtypes("hash_table_int4",3,"len")=8
	Set gtmtypes("hash_table_int4",3,"type")="addr"
	Set gtmtypfldindx("hash_table_int4","spare_base")=3
	Set gtmtypes("hash_table_int4",4,"name")="hash_table_int4.entry_passed_thru"
	Set gtmtypes("hash_table_int4",4,"off")=24
	Set gtmtypes("hash_table_int4",4,"len")=8
	Set gtmtypes("hash_table_int4",4,"type")="addr"
	Set gtmtypfldindx("hash_table_int4","entry_passed_thru")=4
	Set gtmtypes("hash_table_int4",5,"name")="hash_table_int4.size"
	Set gtmtypes("hash_table_int4",5,"off")=32
	Set gtmtypes("hash_table_int4",5,"len")=4
	Set gtmtypes("hash_table_int4",5,"type")="unsigned-int"
	Set gtmtypfldindx("hash_table_int4","size")=5
	Set gtmtypes("hash_table_int4",6,"name")="hash_table_int4.initial_size"
	Set gtmtypes("hash_table_int4",6,"off")=36
	Set gtmtypes("hash_table_int4",6,"len")=4
	Set gtmtypes("hash_table_int4",6,"type")="unsigned-int"
	Set gtmtypfldindx("hash_table_int4","initial_size")=6
	Set gtmtypes("hash_table_int4",7,"name")="hash_table_int4.spare_base_size"
	Set gtmtypes("hash_table_int4",7,"off")=40
	Set gtmtypes("hash_table_int4",7,"len")=4
	Set gtmtypes("hash_table_int4",7,"type")="unsigned-int"
	Set gtmtypfldindx("hash_table_int4","spare_base_size")=7
	Set gtmtypes("hash_table_int4",8,"name")="hash_table_int4.count"
	Set gtmtypes("hash_table_int4",8,"off")=44
	Set gtmtypes("hash_table_int4",8,"len")=4
	Set gtmtypes("hash_table_int4",8,"type")="unsigned-int"
	Set gtmtypfldindx("hash_table_int4","count")=8
	Set gtmtypes("hash_table_int4",9,"name")="hash_table_int4.del_count"
	Set gtmtypes("hash_table_int4",9,"off")=48
	Set gtmtypes("hash_table_int4",9,"len")=4
	Set gtmtypes("hash_table_int4",9,"type")="unsigned-int"
	Set gtmtypfldindx("hash_table_int4","del_count")=9
	Set gtmtypes("hash_table_int4",10,"name")="hash_table_int4.exp_trigger_size"
	Set gtmtypes("hash_table_int4",10,"off")=52
	Set gtmtypes("hash_table_int4",10,"len")=4
	Set gtmtypes("hash_table_int4",10,"type")="unsigned-int"
	Set gtmtypfldindx("hash_table_int4","exp_trigger_size")=10
	Set gtmtypes("hash_table_int4",11,"name")="hash_table_int4.cmp_trigger_size"
	Set gtmtypes("hash_table_int4",11,"off")=56
	Set gtmtypes("hash_table_int4",11,"len")=4
	Set gtmtypes("hash_table_int4",11,"type")="unsigned-int"
	Set gtmtypfldindx("hash_table_int4","cmp_trigger_size")=11
	Set gtmtypes("hash_table_int4",12,"name")="hash_table_int4.dont_compact"
	Set gtmtypes("hash_table_int4",12,"off")=60
	Set gtmtypes("hash_table_int4",12,"len")=4
	Set gtmtypes("hash_table_int4",12,"type")="boolean_t"
	Set gtmtypfldindx("hash_table_int4","dont_compact")=12
	Set gtmtypes("hash_table_int4",13,"name")="hash_table_int4.dont_keep_spare_table"
	Set gtmtypes("hash_table_int4",13,"off")=64
	Set gtmtypes("hash_table_int4",13,"len")=4
	Set gtmtypes("hash_table_int4",13,"type")="boolean_t"
	Set gtmtypfldindx("hash_table_int4","dont_keep_spare_table")=13
	Set gtmtypes("hash_table_int4",14,"name")="hash_table_int4.defer_base_release"
	Set gtmtypes("hash_table_int4",14,"off")=68
	Set gtmtypes("hash_table_int4",14,"len")=4
	Set gtmtypes("hash_table_int4",14,"type")="boolean_t"
	Set gtmtypfldindx("hash_table_int4","defer_base_release")=14
	Set gtmtypes("hash_table_int4",15,"name")="hash_table_int4.active"
	Set gtmtypes("hash_table_int4",15,"off")=72
	Set gtmtypes("hash_table_int4",15,"len")=4
	Set gtmtypes("hash_table_int4",15,"type")="boolean_t"
	Set gtmtypfldindx("hash_table_int4","active")=15
	;
	Set gtmtypes("hash_table_int8")="struct"
	Set gtmtypes("hash_table_int8",0)=15
	Set gtmtypes("hash_table_int8","len")=80
	Set gtmtypes("hash_table_int8",1,"name")="hash_table_int8.base"
	Set gtmtypes("hash_table_int8",1,"off")=0
	Set gtmtypes("hash_table_int8",1,"len")=8
	Set gtmtypes("hash_table_int8",1,"type")="addr"
	Set gtmtypfldindx("hash_table_int8","base")=1
	Set gtmtypes("hash_table_int8",2,"name")="hash_table_int8.top"
	Set gtmtypes("hash_table_int8",2,"off")=8
	Set gtmtypes("hash_table_int8",2,"len")=8
	Set gtmtypes("hash_table_int8",2,"type")="addr"
	Set gtmtypfldindx("hash_table_int8","top")=2
	Set gtmtypes("hash_table_int8",3,"name")="hash_table_int8.spare_base"
	Set gtmtypes("hash_table_int8",3,"off")=16
	Set gtmtypes("hash_table_int8",3,"len")=8
	Set gtmtypes("hash_table_int8",3,"type")="addr"
	Set gtmtypfldindx("hash_table_int8","spare_base")=3
	Set gtmtypes("hash_table_int8",4,"name")="hash_table_int8.entry_passed_thru"
	Set gtmtypes("hash_table_int8",4,"off")=24
	Set gtmtypes("hash_table_int8",4,"len")=8
	Set gtmtypes("hash_table_int8",4,"type")="addr"
	Set gtmtypfldindx("hash_table_int8","entry_passed_thru")=4
	Set gtmtypes("hash_table_int8",5,"name")="hash_table_int8.size"
	Set gtmtypes("hash_table_int8",5,"off")=32
	Set gtmtypes("hash_table_int8",5,"len")=4
	Set gtmtypes("hash_table_int8",5,"type")="unsigned-int"
	Set gtmtypfldindx("hash_table_int8","size")=5
	Set gtmtypes("hash_table_int8",6,"name")="hash_table_int8.initial_size"
	Set gtmtypes("hash_table_int8",6,"off")=36
	Set gtmtypes("hash_table_int8",6,"len")=4
	Set gtmtypes("hash_table_int8",6,"type")="unsigned-int"
	Set gtmtypfldindx("hash_table_int8","initial_size")=6
	Set gtmtypes("hash_table_int8",7,"name")="hash_table_int8.spare_base_size"
	Set gtmtypes("hash_table_int8",7,"off")=40
	Set gtmtypes("hash_table_int8",7,"len")=4
	Set gtmtypes("hash_table_int8",7,"type")="unsigned-int"
	Set gtmtypfldindx("hash_table_int8","spare_base_size")=7
	Set gtmtypes("hash_table_int8",8,"name")="hash_table_int8.count"
	Set gtmtypes("hash_table_int8",8,"off")=44
	Set gtmtypes("hash_table_int8",8,"len")=4
	Set gtmtypes("hash_table_int8",8,"type")="unsigned-int"
	Set gtmtypfldindx("hash_table_int8","count")=8
	Set gtmtypes("hash_table_int8",9,"name")="hash_table_int8.del_count"
	Set gtmtypes("hash_table_int8",9,"off")=48
	Set gtmtypes("hash_table_int8",9,"len")=4
	Set gtmtypes("hash_table_int8",9,"type")="unsigned-int"
	Set gtmtypfldindx("hash_table_int8","del_count")=9
	Set gtmtypes("hash_table_int8",10,"name")="hash_table_int8.exp_trigger_size"
	Set gtmtypes("hash_table_int8",10,"off")=52
	Set gtmtypes("hash_table_int8",10,"len")=4
	Set gtmtypes("hash_table_int8",10,"type")="unsigned-int"
	Set gtmtypfldindx("hash_table_int8","exp_trigger_size")=10
	Set gtmtypes("hash_table_int8",11,"name")="hash_table_int8.cmp_trigger_size"
	Set gtmtypes("hash_table_int8",11,"off")=56
	Set gtmtypes("hash_table_int8",11,"len")=4
	Set gtmtypes("hash_table_int8",11,"type")="unsigned-int"
	Set gtmtypfldindx("hash_table_int8","cmp_trigger_size")=11
	Set gtmtypes("hash_table_int8",12,"name")="hash_table_int8.dont_compact"
	Set gtmtypes("hash_table_int8",12,"off")=60
	Set gtmtypes("hash_table_int8",12,"len")=4
	Set gtmtypes("hash_table_int8",12,"type")="boolean_t"
	Set gtmtypfldindx("hash_table_int8","dont_compact")=12
	Set gtmtypes("hash_table_int8",13,"name")="hash_table_int8.dont_keep_spare_table"
	Set gtmtypes("hash_table_int8",13,"off")=64
	Set gtmtypes("hash_table_int8",13,"len")=4
	Set gtmtypes("hash_table_int8",13,"type")="boolean_t"
	Set gtmtypfldindx("hash_table_int8","dont_keep_spare_table")=13
	Set gtmtypes("hash_table_int8",14,"name")="hash_table_int8.defer_base_release"
	Set gtmtypes("hash_table_int8",14,"off")=68
	Set gtmtypes("hash_table_int8",14,"len")=4
	Set gtmtypes("hash_table_int8",14,"type")="boolean_t"
	Set gtmtypfldindx("hash_table_int8","defer_base_release")=14
	Set gtmtypes("hash_table_int8",15,"name")="hash_table_int8.active"
	Set gtmtypes("hash_table_int8",15,"off")=72
	Set gtmtypes("hash_table_int8",15,"len")=4
	Set gtmtypes("hash_table_int8",15,"type")="boolean_t"
	Set gtmtypfldindx("hash_table_int8","active")=15
	;
	Set gtmtypes("hash_table_mname")="struct"
	Set gtmtypes("hash_table_mname",0)=15
	Set gtmtypes("hash_table_mname","len")=80
	Set gtmtypes("hash_table_mname",1,"name")="hash_table_mname.base"
	Set gtmtypes("hash_table_mname",1,"off")=0
	Set gtmtypes("hash_table_mname",1,"len")=8
	Set gtmtypes("hash_table_mname",1,"type")="addr"
	Set gtmtypfldindx("hash_table_mname","base")=1
	Set gtmtypes("hash_table_mname",2,"name")="hash_table_mname.top"
	Set gtmtypes("hash_table_mname",2,"off")=8
	Set gtmtypes("hash_table_mname",2,"len")=8
	Set gtmtypes("hash_table_mname",2,"type")="addr"
	Set gtmtypfldindx("hash_table_mname","top")=2
	Set gtmtypes("hash_table_mname",3,"name")="hash_table_mname.spare_base"
	Set gtmtypes("hash_table_mname",3,"off")=16
	Set gtmtypes("hash_table_mname",3,"len")=8
	Set gtmtypes("hash_table_mname",3,"type")="addr"
	Set gtmtypfldindx("hash_table_mname","spare_base")=3
	Set gtmtypes("hash_table_mname",4,"name")="hash_table_mname.entry_passed_thru"
	Set gtmtypes("hash_table_mname",4,"off")=24
	Set gtmtypes("hash_table_mname",4,"len")=8
	Set gtmtypes("hash_table_mname",4,"type")="addr"
	Set gtmtypfldindx("hash_table_mname","entry_passed_thru")=4
	Set gtmtypes("hash_table_mname",5,"name")="hash_table_mname.size"
	Set gtmtypes("hash_table_mname",5,"off")=32
	Set gtmtypes("hash_table_mname",5,"len")=4
	Set gtmtypes("hash_table_mname",5,"type")="unsigned-int"
	Set gtmtypfldindx("hash_table_mname","size")=5
	Set gtmtypes("hash_table_mname",6,"name")="hash_table_mname.initial_size"
	Set gtmtypes("hash_table_mname",6,"off")=36
	Set gtmtypes("hash_table_mname",6,"len")=4
	Set gtmtypes("hash_table_mname",6,"type")="unsigned-int"
	Set gtmtypfldindx("hash_table_mname","initial_size")=6
	Set gtmtypes("hash_table_mname",7,"name")="hash_table_mname.spare_base_size"
	Set gtmtypes("hash_table_mname",7,"off")=40
	Set gtmtypes("hash_table_mname",7,"len")=4
	Set gtmtypes("hash_table_mname",7,"type")="unsigned-int"
	Set gtmtypfldindx("hash_table_mname","spare_base_size")=7
	Set gtmtypes("hash_table_mname",8,"name")="hash_table_mname.count"
	Set gtmtypes("hash_table_mname",8,"off")=44
	Set gtmtypes("hash_table_mname",8,"len")=4
	Set gtmtypes("hash_table_mname",8,"type")="unsigned-int"
	Set gtmtypfldindx("hash_table_mname","count")=8
	Set gtmtypes("hash_table_mname",9,"name")="hash_table_mname.del_count"
	Set gtmtypes("hash_table_mname",9,"off")=48
	Set gtmtypes("hash_table_mname",9,"len")=4
	Set gtmtypes("hash_table_mname",9,"type")="unsigned-int"
	Set gtmtypfldindx("hash_table_mname","del_count")=9
	Set gtmtypes("hash_table_mname",10,"name")="hash_table_mname.exp_trigger_size"
	Set gtmtypes("hash_table_mname",10,"off")=52
	Set gtmtypes("hash_table_mname",10,"len")=4
	Set gtmtypes("hash_table_mname",10,"type")="unsigned-int"
	Set gtmtypfldindx("hash_table_mname","exp_trigger_size")=10
	Set gtmtypes("hash_table_mname",11,"name")="hash_table_mname.cmp_trigger_size"
	Set gtmtypes("hash_table_mname",11,"off")=56
	Set gtmtypes("hash_table_mname",11,"len")=4
	Set gtmtypes("hash_table_mname",11,"type")="unsigned-int"
	Set gtmtypfldindx("hash_table_mname","cmp_trigger_size")=11
	Set gtmtypes("hash_table_mname",12,"name")="hash_table_mname.dont_compact"
	Set gtmtypes("hash_table_mname",12,"off")=60
	Set gtmtypes("hash_table_mname",12,"len")=4
	Set gtmtypes("hash_table_mname",12,"type")="boolean_t"
	Set gtmtypfldindx("hash_table_mname","dont_compact")=12
	Set gtmtypes("hash_table_mname",13,"name")="hash_table_mname.dont_keep_spare_table"
	Set gtmtypes("hash_table_mname",13,"off")=64
	Set gtmtypes("hash_table_mname",13,"len")=4
	Set gtmtypes("hash_table_mname",13,"type")="boolean_t"
	Set gtmtypfldindx("hash_table_mname","dont_keep_spare_table")=13
	Set gtmtypes("hash_table_mname",14,"name")="hash_table_mname.defer_base_release"
	Set gtmtypes("hash_table_mname",14,"off")=68
	Set gtmtypes("hash_table_mname",14,"len")=4
	Set gtmtypes("hash_table_mname",14,"type")="boolean_t"
	Set gtmtypfldindx("hash_table_mname","defer_base_release")=14
	Set gtmtypes("hash_table_mname",15,"name")="hash_table_mname.active"
	Set gtmtypes("hash_table_mname",15,"off")=72
	Set gtmtypes("hash_table_mname",15,"len")=4
	Set gtmtypes("hash_table_mname",15,"type")="boolean_t"
	Set gtmtypfldindx("hash_table_mname","active")=15
	;
	Set gtmtypes("hash_table_objcode")="struct"
	Set gtmtypes("hash_table_objcode",0)=15
	Set gtmtypes("hash_table_objcode","len")=80
	Set gtmtypes("hash_table_objcode",1,"name")="hash_table_objcode.base"
	Set gtmtypes("hash_table_objcode",1,"off")=0
	Set gtmtypes("hash_table_objcode",1,"len")=8
	Set gtmtypes("hash_table_objcode",1,"type")="addr"
	Set gtmtypfldindx("hash_table_objcode","base")=1
	Set gtmtypes("hash_table_objcode",2,"name")="hash_table_objcode.top"
	Set gtmtypes("hash_table_objcode",2,"off")=8
	Set gtmtypes("hash_table_objcode",2,"len")=8
	Set gtmtypes("hash_table_objcode",2,"type")="addr"
	Set gtmtypfldindx("hash_table_objcode","top")=2
	Set gtmtypes("hash_table_objcode",3,"name")="hash_table_objcode.spare_base"
	Set gtmtypes("hash_table_objcode",3,"off")=16
	Set gtmtypes("hash_table_objcode",3,"len")=8
	Set gtmtypes("hash_table_objcode",3,"type")="addr"
	Set gtmtypfldindx("hash_table_objcode","spare_base")=3
	Set gtmtypes("hash_table_objcode",4,"name")="hash_table_objcode.entry_passed_thru"
	Set gtmtypes("hash_table_objcode",4,"off")=24
	Set gtmtypes("hash_table_objcode",4,"len")=8
	Set gtmtypes("hash_table_objcode",4,"type")="addr"
	Set gtmtypfldindx("hash_table_objcode","entry_passed_thru")=4
	Set gtmtypes("hash_table_objcode",5,"name")="hash_table_objcode.size"
	Set gtmtypes("hash_table_objcode",5,"off")=32
	Set gtmtypes("hash_table_objcode",5,"len")=4
	Set gtmtypes("hash_table_objcode",5,"type")="unsigned-int"
	Set gtmtypfldindx("hash_table_objcode","size")=5
	Set gtmtypes("hash_table_objcode",6,"name")="hash_table_objcode.initial_size"
	Set gtmtypes("hash_table_objcode",6,"off")=36
	Set gtmtypes("hash_table_objcode",6,"len")=4
	Set gtmtypes("hash_table_objcode",6,"type")="unsigned-int"
	Set gtmtypfldindx("hash_table_objcode","initial_size")=6
	Set gtmtypes("hash_table_objcode",7,"name")="hash_table_objcode.spare_base_size"
	Set gtmtypes("hash_table_objcode",7,"off")=40
	Set gtmtypes("hash_table_objcode",7,"len")=4
	Set gtmtypes("hash_table_objcode",7,"type")="unsigned-int"
	Set gtmtypfldindx("hash_table_objcode","spare_base_size")=7
	Set gtmtypes("hash_table_objcode",8,"name")="hash_table_objcode.count"
	Set gtmtypes("hash_table_objcode",8,"off")=44
	Set gtmtypes("hash_table_objcode",8,"len")=4
	Set gtmtypes("hash_table_objcode",8,"type")="unsigned-int"
	Set gtmtypfldindx("hash_table_objcode","count")=8
	Set gtmtypes("hash_table_objcode",9,"name")="hash_table_objcode.del_count"
	Set gtmtypes("hash_table_objcode",9,"off")=48
	Set gtmtypes("hash_table_objcode",9,"len")=4
	Set gtmtypes("hash_table_objcode",9,"type")="unsigned-int"
	Set gtmtypfldindx("hash_table_objcode","del_count")=9
	Set gtmtypes("hash_table_objcode",10,"name")="hash_table_objcode.exp_trigger_size"
	Set gtmtypes("hash_table_objcode",10,"off")=52
	Set gtmtypes("hash_table_objcode",10,"len")=4
	Set gtmtypes("hash_table_objcode",10,"type")="unsigned-int"
	Set gtmtypfldindx("hash_table_objcode","exp_trigger_size")=10
	Set gtmtypes("hash_table_objcode",11,"name")="hash_table_objcode.cmp_trigger_size"
	Set gtmtypes("hash_table_objcode",11,"off")=56
	Set gtmtypes("hash_table_objcode",11,"len")=4
	Set gtmtypes("hash_table_objcode",11,"type")="unsigned-int"
	Set gtmtypfldindx("hash_table_objcode","cmp_trigger_size")=11
	Set gtmtypes("hash_table_objcode",12,"name")="hash_table_objcode.dont_compact"
	Set gtmtypes("hash_table_objcode",12,"off")=60
	Set gtmtypes("hash_table_objcode",12,"len")=4
	Set gtmtypes("hash_table_objcode",12,"type")="boolean_t"
	Set gtmtypfldindx("hash_table_objcode","dont_compact")=12
	Set gtmtypes("hash_table_objcode",13,"name")="hash_table_objcode.dont_keep_spare_table"
	Set gtmtypes("hash_table_objcode",13,"off")=64
	Set gtmtypes("hash_table_objcode",13,"len")=4
	Set gtmtypes("hash_table_objcode",13,"type")="boolean_t"
	Set gtmtypfldindx("hash_table_objcode","dont_keep_spare_table")=13
	Set gtmtypes("hash_table_objcode",14,"name")="hash_table_objcode.defer_base_release"
	Set gtmtypes("hash_table_objcode",14,"off")=68
	Set gtmtypes("hash_table_objcode",14,"len")=4
	Set gtmtypes("hash_table_objcode",14,"type")="boolean_t"
	Set gtmtypfldindx("hash_table_objcode","defer_base_release")=14
	Set gtmtypes("hash_table_objcode",15,"name")="hash_table_objcode.active"
	Set gtmtypes("hash_table_objcode",15,"off")=72
	Set gtmtypes("hash_table_objcode",15,"len")=4
	Set gtmtypes("hash_table_objcode",15,"type")="boolean_t"
	Set gtmtypfldindx("hash_table_objcode","active")=15
	;
	Set gtmtypes("hash_table_str")="struct"
	Set gtmtypes("hash_table_str",0)=15
	Set gtmtypes("hash_table_str","len")=80
	Set gtmtypes("hash_table_str",1,"name")="hash_table_str.base"
	Set gtmtypes("hash_table_str",1,"off")=0
	Set gtmtypes("hash_table_str",1,"len")=8
	Set gtmtypes("hash_table_str",1,"type")="addr"
	Set gtmtypfldindx("hash_table_str","base")=1
	Set gtmtypes("hash_table_str",2,"name")="hash_table_str.top"
	Set gtmtypes("hash_table_str",2,"off")=8
	Set gtmtypes("hash_table_str",2,"len")=8
	Set gtmtypes("hash_table_str",2,"type")="addr"
	Set gtmtypfldindx("hash_table_str","top")=2
	Set gtmtypes("hash_table_str",3,"name")="hash_table_str.spare_base"
	Set gtmtypes("hash_table_str",3,"off")=16
	Set gtmtypes("hash_table_str",3,"len")=8
	Set gtmtypes("hash_table_str",3,"type")="addr"
	Set gtmtypfldindx("hash_table_str","spare_base")=3
	Set gtmtypes("hash_table_str",4,"name")="hash_table_str.entry_passed_thru"
	Set gtmtypes("hash_table_str",4,"off")=24
	Set gtmtypes("hash_table_str",4,"len")=8
	Set gtmtypes("hash_table_str",4,"type")="addr"
	Set gtmtypfldindx("hash_table_str","entry_passed_thru")=4
	Set gtmtypes("hash_table_str",5,"name")="hash_table_str.size"
	Set gtmtypes("hash_table_str",5,"off")=32
	Set gtmtypes("hash_table_str",5,"len")=4
	Set gtmtypes("hash_table_str",5,"type")="unsigned-int"
	Set gtmtypfldindx("hash_table_str","size")=5
	Set gtmtypes("hash_table_str",6,"name")="hash_table_str.initial_size"
	Set gtmtypes("hash_table_str",6,"off")=36
	Set gtmtypes("hash_table_str",6,"len")=4
	Set gtmtypes("hash_table_str",6,"type")="unsigned-int"
	Set gtmtypfldindx("hash_table_str","initial_size")=6
	Set gtmtypes("hash_table_str",7,"name")="hash_table_str.spare_base_size"
	Set gtmtypes("hash_table_str",7,"off")=40
	Set gtmtypes("hash_table_str",7,"len")=4
	Set gtmtypes("hash_table_str",7,"type")="unsigned-int"
	Set gtmtypfldindx("hash_table_str","spare_base_size")=7
	Set gtmtypes("hash_table_str",8,"name")="hash_table_str.count"
	Set gtmtypes("hash_table_str",8,"off")=44
	Set gtmtypes("hash_table_str",8,"len")=4
	Set gtmtypes("hash_table_str",8,"type")="unsigned-int"
	Set gtmtypfldindx("hash_table_str","count")=8
	Set gtmtypes("hash_table_str",9,"name")="hash_table_str.del_count"
	Set gtmtypes("hash_table_str",9,"off")=48
	Set gtmtypes("hash_table_str",9,"len")=4
	Set gtmtypes("hash_table_str",9,"type")="unsigned-int"
	Set gtmtypfldindx("hash_table_str","del_count")=9
	Set gtmtypes("hash_table_str",10,"name")="hash_table_str.exp_trigger_size"
	Set gtmtypes("hash_table_str",10,"off")=52
	Set gtmtypes("hash_table_str",10,"len")=4
	Set gtmtypes("hash_table_str",10,"type")="unsigned-int"
	Set gtmtypfldindx("hash_table_str","exp_trigger_size")=10
	Set gtmtypes("hash_table_str",11,"name")="hash_table_str.cmp_trigger_size"
	Set gtmtypes("hash_table_str",11,"off")=56
	Set gtmtypes("hash_table_str",11,"len")=4
	Set gtmtypes("hash_table_str",11,"type")="unsigned-int"
	Set gtmtypfldindx("hash_table_str","cmp_trigger_size")=11
	Set gtmtypes("hash_table_str",12,"name")="hash_table_str.dont_compact"
	Set gtmtypes("hash_table_str",12,"off")=60
	Set gtmtypes("hash_table_str",12,"len")=4
	Set gtmtypes("hash_table_str",12,"type")="boolean_t"
	Set gtmtypfldindx("hash_table_str","dont_compact")=12
	Set gtmtypes("hash_table_str",13,"name")="hash_table_str.dont_keep_spare_table"
	Set gtmtypes("hash_table_str",13,"off")=64
	Set gtmtypes("hash_table_str",13,"len")=4
	Set gtmtypes("hash_table_str",13,"type")="boolean_t"
	Set gtmtypfldindx("hash_table_str","dont_keep_spare_table")=13
	Set gtmtypes("hash_table_str",14,"name")="hash_table_str.defer_base_release"
	Set gtmtypes("hash_table_str",14,"off")=68
	Set gtmtypes("hash_table_str",14,"len")=4
	Set gtmtypes("hash_table_str",14,"type")="boolean_t"
	Set gtmtypfldindx("hash_table_str","defer_base_release")=14
	Set gtmtypes("hash_table_str",15,"name")="hash_table_str.active"
	Set gtmtypes("hash_table_str",15,"off")=72
	Set gtmtypes("hash_table_str",15,"len")=4
	Set gtmtypes("hash_table_str",15,"type")="boolean_t"
	Set gtmtypfldindx("hash_table_str","active")=15
	;
	Set gtmtypes("header_struct")="struct"
	Set gtmtypes("header_struct",0)=2
	Set gtmtypes("header_struct","len")=16
	Set gtmtypes("header_struct",1,"name")="header_struct.label"
	Set gtmtypes("header_struct",1,"off")=0
	Set gtmtypes("header_struct",1,"len")=12
	Set gtmtypes("header_struct",1,"type")="char"
	Set gtmtypfldindx("header_struct","label")=1
	Set gtmtypes("header_struct",2,"name")="header_struct.filesize"
	Set gtmtypes("header_struct",2,"off")=12
	Set gtmtypes("header_struct",2,"len")=4
	Set gtmtypes("header_struct",2,"type")="unsigned"
	Set gtmtypfldindx("header_struct","filesize")=2
	;
	Set gtmtypes("ht_ent_addr")="struct"
	Set gtmtypes("ht_ent_addr",0)=2
	Set gtmtypes("ht_ent_addr","len")=16
	Set gtmtypes("ht_ent_addr",1,"name")="ht_ent_addr.key"
	Set gtmtypes("ht_ent_addr",1,"off")=0
	Set gtmtypes("ht_ent_addr",1,"len")=8
	Set gtmtypes("ht_ent_addr",1,"type")="addr"
	Set gtmtypfldindx("ht_ent_addr","key")=1
	Set gtmtypes("ht_ent_addr",2,"name")="ht_ent_addr.value"
	Set gtmtypes("ht_ent_addr",2,"off")=8
	Set gtmtypes("ht_ent_addr",2,"len")=8
	Set gtmtypes("ht_ent_addr",2,"type")="addr"
	Set gtmtypfldindx("ht_ent_addr","value")=2
	;
	Set gtmtypes("ht_ent_int4")="struct"
	Set gtmtypes("ht_ent_int4",0)=2
	Set gtmtypes("ht_ent_int4","len")=16
	Set gtmtypes("ht_ent_int4",1,"name")="ht_ent_int4.key"
	Set gtmtypes("ht_ent_int4",1,"off")=0
	Set gtmtypes("ht_ent_int4",1,"len")=4
	Set gtmtypes("ht_ent_int4",1,"type")="unsigned-int"
	Set gtmtypfldindx("ht_ent_int4","key")=1
	Set gtmtypes("ht_ent_int4",2,"name")="ht_ent_int4.value"
	Set gtmtypes("ht_ent_int4",2,"off")=8
	Set gtmtypes("ht_ent_int4",2,"len")=8
	Set gtmtypes("ht_ent_int4",2,"type")="addr"
	Set gtmtypfldindx("ht_ent_int4","value")=2
	;
	Set gtmtypes("ht_ent_int8")="struct"
	Set gtmtypes("ht_ent_int8",0)=2
	Set gtmtypes("ht_ent_int8","len")=16
	Set gtmtypes("ht_ent_int8",1,"name")="ht_ent_int8.key"
	Set gtmtypes("ht_ent_int8",1,"off")=0
	Set gtmtypes("ht_ent_int8",1,"len")=8
	Set gtmtypes("ht_ent_int8",1,"type")="uint64_t"
	Set gtmtypfldindx("ht_ent_int8","key")=1
	Set gtmtypes("ht_ent_int8",2,"name")="ht_ent_int8.value"
	Set gtmtypes("ht_ent_int8",2,"off")=8
	Set gtmtypes("ht_ent_int8",2,"len")=8
	Set gtmtypes("ht_ent_int8",2,"type")="addr"
	Set gtmtypfldindx("ht_ent_int8","value")=2
	;
	Set gtmtypes("ht_ent_mname")="struct"
	Set gtmtypes("ht_ent_mname",0)=5
	Set gtmtypes("ht_ent_mname","len")=32
	Set gtmtypes("ht_ent_mname",1,"name")="ht_ent_mname.key"
	Set gtmtypes("ht_ent_mname",1,"off")=0
	Set gtmtypes("ht_ent_mname",1,"len")=24
	Set gtmtypes("ht_ent_mname",1,"type")="mname_entry"
	Set gtmtypfldindx("ht_ent_mname","key")=1
	Set gtmtypes("ht_ent_mname",2,"name")="ht_ent_mname.key.var_name"
	Set gtmtypes("ht_ent_mname",2,"off")=0
	Set gtmtypes("ht_ent_mname",2,"len")=16
	Set gtmtypes("ht_ent_mname",2,"type")="mstr"
	Set gtmtypfldindx("ht_ent_mname","key.var_name")=2
	Set gtmtypes("ht_ent_mname",3,"name")="ht_ent_mname.key.hash_code"
	Set gtmtypes("ht_ent_mname",3,"off")=16
	Set gtmtypes("ht_ent_mname",3,"len")=4
	Set gtmtypes("ht_ent_mname",3,"type")="unsigned-int"
	Set gtmtypfldindx("ht_ent_mname","key.hash_code")=3
	Set gtmtypes("ht_ent_mname",4,"name")="ht_ent_mname.key.marked"
	Set gtmtypes("ht_ent_mname",4,"off")=20
	Set gtmtypes("ht_ent_mname",4,"len")=4
	Set gtmtypes("ht_ent_mname",4,"type")="boolean_t"
	Set gtmtypfldindx("ht_ent_mname","key.marked")=4
	Set gtmtypes("ht_ent_mname",5,"name")="ht_ent_mname.value"
	Set gtmtypes("ht_ent_mname",5,"off")=24
	Set gtmtypes("ht_ent_mname",5,"len")=8
	Set gtmtypes("ht_ent_mname",5,"type")="addr"
	Set gtmtypfldindx("ht_ent_mname","value")=5
	;
	Set gtmtypes("ht_ent_objcode")="struct"
	Set gtmtypes("ht_ent_objcode",0)=4
	Set gtmtypes("ht_ent_objcode","len")=32
	Set gtmtypes("ht_ent_objcode",1,"name")="ht_ent_objcode.key"
	Set gtmtypes("ht_ent_objcode",1,"off")=0
	Set gtmtypes("ht_ent_objcode",1,"len")=24
	Set gtmtypes("ht_ent_objcode",1,"type")="icode_str"
	Set gtmtypfldindx("ht_ent_objcode","key")=1
	Set gtmtypes("ht_ent_objcode",2,"name")="ht_ent_objcode.key.str"
	Set gtmtypes("ht_ent_objcode",2,"off")=0
	Set gtmtypes("ht_ent_objcode",2,"len")=16
	Set gtmtypes("ht_ent_objcode",2,"type")="mstr"
	Set gtmtypfldindx("ht_ent_objcode","key.str")=2
	Set gtmtypes("ht_ent_objcode",3,"name")="ht_ent_objcode.key.code"
	Set gtmtypes("ht_ent_objcode",3,"off")=16
	Set gtmtypes("ht_ent_objcode",3,"len")=4
	Set gtmtypes("ht_ent_objcode",3,"type")="unsigned-int"
	Set gtmtypfldindx("ht_ent_objcode","key.code")=3
	Set gtmtypes("ht_ent_objcode",4,"name")="ht_ent_objcode.value"
	Set gtmtypes("ht_ent_objcode",4,"off")=24
	Set gtmtypes("ht_ent_objcode",4,"len")=8
	Set gtmtypes("ht_ent_objcode",4,"type")="addr"
	Set gtmtypfldindx("ht_ent_objcode","value")=4
	;
	Set gtmtypes("ht_ent_str")="struct"
	Set gtmtypes("ht_ent_str",0)=5
	Set gtmtypes("ht_ent_str","len")=32
	Set gtmtypes("ht_ent_str",1,"name")="ht_ent_str.key"
	Set gtmtypes("ht_ent_str",1,"off")=0
	Set gtmtypes("ht_ent_str",1,"len")=24
	Set gtmtypes("ht_ent_str",1,"type")="stringkey"
	Set gtmtypfldindx("ht_ent_str","key")=1
	Set gtmtypes("ht_ent_str",2,"name")="ht_ent_str.key.str"
	Set gtmtypes("ht_ent_str",2,"off")=0
	Set gtmtypes("ht_ent_str",2,"len")=16
	Set gtmtypes("ht_ent_str",2,"type")="mstr"
	Set gtmtypfldindx("ht_ent_str","key.str")=2
	Set gtmtypes("ht_ent_str",3,"name")="ht_ent_str.key.hash_code"
	Set gtmtypes("ht_ent_str",3,"off")=16
	Set gtmtypes("ht_ent_str",3,"len")=4
	Set gtmtypes("ht_ent_str",3,"type")="unsigned-int"
	Set gtmtypfldindx("ht_ent_str","key.hash_code")=3
	Set gtmtypes("ht_ent_str",4,"name")="ht_ent_str.key.filler"
	Set gtmtypes("ht_ent_str",4,"off")=20
	Set gtmtypes("ht_ent_str",4,"len")=4
	Set gtmtypes("ht_ent_str",4,"type")="int"
	Set gtmtypfldindx("ht_ent_str","key.filler")=4
	Set gtmtypes("ht_ent_str",5,"name")="ht_ent_str.value"
	Set gtmtypes("ht_ent_str",5,"off")=24
	Set gtmtypes("ht_ent_str",5,"len")=8
	Set gtmtypes("ht_ent_str",5,"type")="addr"
	Set gtmtypfldindx("ht_ent_str","value")=5
	;
	Set gtmtypes("icode_str")="struct"
	Set gtmtypes("icode_str",0)=5
	Set gtmtypes("icode_str","len")=24
	Set gtmtypes("icode_str",1,"name")="icode_str.str"
	Set gtmtypes("icode_str",1,"off")=0
	Set gtmtypes("icode_str",1,"len")=16
	Set gtmtypes("icode_str",1,"type")="mstr"
	Set gtmtypfldindx("icode_str","str")=1
	Set gtmtypes("icode_str",2,"name")="icode_str.str.char_len"
	Set gtmtypes("icode_str",2,"off")=0
	Set gtmtypes("icode_str",2,"len")=4
	Set gtmtypes("icode_str",2,"type")="unsigned-int"
	Set gtmtypfldindx("icode_str","str.char_len")=2
	Set gtmtypes("icode_str",3,"name")="icode_str.str.len"
	Set gtmtypes("icode_str",3,"off")=4
	Set gtmtypes("icode_str",3,"len")=4
	Set gtmtypes("icode_str",3,"type")="int"
	Set gtmtypfldindx("icode_str","str.len")=3
	Set gtmtypes("icode_str",4,"name")="icode_str.str.addr"
	Set gtmtypes("icode_str",4,"off")=8
	Set gtmtypes("icode_str",4,"len")=8
	Set gtmtypes("icode_str",4,"type")="addr"
	Set gtmtypfldindx("icode_str","str.addr")=4
	Set gtmtypes("icode_str",5,"name")="icode_str.code"
	Set gtmtypes("icode_str",5,"off")=16
	Set gtmtypes("icode_str",5,"len")=4
	Set gtmtypes("icode_str",5,"type")="unsigned-int"
	Set gtmtypfldindx("icode_str","code")=5
	;
	Set gtmtypes("ihdtyp")="struct"
	Set gtmtypes("ihdtyp",0)=7
	Set gtmtypes("ihdtyp","len")=32
	Set gtmtypes("ihdtyp",1,"name")="ihdtyp.indce"
	Set gtmtypes("ihdtyp",1,"off")=0
	Set gtmtypes("ihdtyp",1,"len")=8
	Set gtmtypes("ihdtyp",1,"type")="addr"
	Set gtmtypfldindx("ihdtyp","indce")=1
	Set gtmtypes("ihdtyp",2,"name")="ihdtyp.vartab_off"
	Set gtmtypes("ihdtyp",2,"off")=8
	Set gtmtypes("ihdtyp",2,"len")=4
	Set gtmtypes("ihdtyp",2,"type")="int"
	Set gtmtypfldindx("ihdtyp","vartab_off")=2
	Set gtmtypes("ihdtyp",3,"name")="ihdtyp.vartab_len"
	Set gtmtypes("ihdtyp",3,"off")=12
	Set gtmtypes("ihdtyp",3,"len")=4
	Set gtmtypes("ihdtyp",3,"type")="int"
	Set gtmtypfldindx("ihdtyp","vartab_len")=3
	Set gtmtypes("ihdtyp",4,"name")="ihdtyp.temp_mvals"
	Set gtmtypes("ihdtyp",4,"off")=16
	Set gtmtypes("ihdtyp",4,"len")=4
	Set gtmtypes("ihdtyp",4,"type")="int"
	Set gtmtypfldindx("ihdtyp","temp_mvals")=4
	Set gtmtypes("ihdtyp",5,"name")="ihdtyp.temp_size"
	Set gtmtypes("ihdtyp",5,"off")=20
	Set gtmtypes("ihdtyp",5,"len")=4
	Set gtmtypes("ihdtyp",5,"type")="int"
	Set gtmtypfldindx("ihdtyp","temp_size")=5
	Set gtmtypes("ihdtyp",6,"name")="ihdtyp.fixup_vals_off"
	Set gtmtypes("ihdtyp",6,"off")=24
	Set gtmtypes("ihdtyp",6,"len")=4
	Set gtmtypes("ihdtyp",6,"type")="int"
	Set gtmtypfldindx("ihdtyp","fixup_vals_off")=6
	Set gtmtypes("ihdtyp",7,"name")="ihdtyp.fixup_vals_num"
	Set gtmtypes("ihdtyp",7,"off")=28
	Set gtmtypes("ihdtyp",7,"len")=4
	Set gtmtypes("ihdtyp",7,"type")="int"
	Set gtmtypfldindx("ihdtyp","fixup_vals_num")=7
	;
	Set gtmtypes("inc_header")="struct"
	Set gtmtypes("inc_header",0)=14
	Set gtmtypes("inc_header","len")=280
	Set gtmtypes("inc_header",1,"name")="inc_header.label"
	Set gtmtypes("inc_header",1,"off")=0
	Set gtmtypes("inc_header",1,"len")=26
	Set gtmtypes("inc_header",1,"type")="char"
	Set gtmtypfldindx("inc_header","label")=1
	Set gtmtypes("inc_header",2,"name")="inc_header.date"
	Set gtmtypes("inc_header",2,"off")=26
	Set gtmtypes("inc_header",2,"len")=14
	Set gtmtypes("inc_header",2,"type")="char"
	Set gtmtypfldindx("inc_header","date")=2
	Set gtmtypes("inc_header",3,"name")="inc_header.reg"
	Set gtmtypes("inc_header",3,"off")=40
	Set gtmtypes("inc_header",3,"len")=31
	Set gtmtypes("inc_header",3,"type")="char"
	Set gtmtypfldindx("inc_header","reg")=3
	Set gtmtypes("inc_header",4,"name")="inc_header.start_tn"
	Set gtmtypes("inc_header",4,"off")=72
	Set gtmtypes("inc_header",4,"len")=8
	Set gtmtypes("inc_header",4,"type")="uint64_t"
	Set gtmtypfldindx("inc_header","start_tn")=4
	Set gtmtypes("inc_header",5,"name")="inc_header.end_tn"
	Set gtmtypes("inc_header",5,"off")=80
	Set gtmtypes("inc_header",5,"len")=8
	Set gtmtypes("inc_header",5,"type")="uint64_t"
	Set gtmtypfldindx("inc_header","end_tn")=5
	Set gtmtypes("inc_header",6,"name")="inc_header.db_total_blks"
	Set gtmtypes("inc_header",6,"off")=88
	Set gtmtypes("inc_header",6,"len")=4
	Set gtmtypes("inc_header",6,"type")="unsigned-int"
	Set gtmtypfldindx("inc_header","db_total_blks")=6
	Set gtmtypes("inc_header",7,"name")="inc_header.blk_size"
	Set gtmtypes("inc_header",7,"off")=92
	Set gtmtypes("inc_header",7,"len")=4
	Set gtmtypes("inc_header",7,"type")="unsigned-int"
	Set gtmtypfldindx("inc_header","blk_size")=7
	Set gtmtypes("inc_header",8,"name")="inc_header.blks_to_upgrd"
	Set gtmtypes("inc_header",8,"off")=96
	Set gtmtypes("inc_header",8,"len")=4
	Set gtmtypes("inc_header",8,"type")="int"
	Set gtmtypfldindx("inc_header","blks_to_upgrd")=8
	Set gtmtypes("inc_header",9,"name")="inc_header.is_encrypted"
	Set gtmtypes("inc_header",9,"off")=100
	Set gtmtypes("inc_header",9,"len")=4
	Set gtmtypes("inc_header",9,"type")="unsigned-int"
	Set gtmtypfldindx("inc_header","is_encrypted")=9
	Set gtmtypes("inc_header",10,"name")="inc_header.encryption_hash"
	Set gtmtypes("inc_header",10,"off")=104
	Set gtmtypes("inc_header",10,"len")=80
	Set gtmtypes("inc_header",10,"type")="char"
	Set gtmtypfldindx("inc_header","encryption_hash")=10
	Set gtmtypes("inc_header",11,"name")="inc_header.encryption_hash2"
	Set gtmtypes("inc_header",11,"off")=184
	Set gtmtypes("inc_header",11,"len")=80
	Set gtmtypes("inc_header",11,"type")="char"
	Set gtmtypfldindx("inc_header","encryption_hash2")=11
	Set gtmtypes("inc_header",12,"name")="inc_header.non_null_iv"
	Set gtmtypes("inc_header",12,"off")=264
	Set gtmtypes("inc_header",12,"len")=4
	Set gtmtypes("inc_header",12,"type")="boolean_t"
	Set gtmtypfldindx("inc_header","non_null_iv")=12
	Set gtmtypes("inc_header",13,"name")="inc_header.encryption_hash_cutoff"
	Set gtmtypes("inc_header",13,"off")=268
	Set gtmtypes("inc_header",13,"len")=4
	Set gtmtypes("inc_header",13,"type")="int"
	Set gtmtypfldindx("inc_header","encryption_hash_cutoff")=13
	Set gtmtypes("inc_header",14,"name")="inc_header.encryption_hash2_start_tn"
	Set gtmtypes("inc_header",14,"off")=272
	Set gtmtypes("inc_header",14,"len")=8
	Set gtmtypes("inc_header",14,"type")="uint64_t"
	Set gtmtypfldindx("inc_header","encryption_hash2_start_tn")=14
	;
	Set gtmtypes("inc_list_struct")="struct"
	Set gtmtypes("inc_list_struct",0)=5
	Set gtmtypes("inc_list_struct","len")=24
	Set gtmtypes("inc_list_struct",1,"name")="inc_list_struct.input_file"
	Set gtmtypes("inc_list_struct",1,"off")=0
	Set gtmtypes("inc_list_struct",1,"len")=16
	Set gtmtypes("inc_list_struct",1,"type")="mstr"
	Set gtmtypfldindx("inc_list_struct","input_file")=1
	Set gtmtypes("inc_list_struct",2,"name")="inc_list_struct.input_file.char_len"
	Set gtmtypes("inc_list_struct",2,"off")=0
	Set gtmtypes("inc_list_struct",2,"len")=4
	Set gtmtypes("inc_list_struct",2,"type")="unsigned-int"
	Set gtmtypfldindx("inc_list_struct","input_file.char_len")=2
	Set gtmtypes("inc_list_struct",3,"name")="inc_list_struct.input_file.len"
	Set gtmtypes("inc_list_struct",3,"off")=4
	Set gtmtypes("inc_list_struct",3,"len")=4
	Set gtmtypes("inc_list_struct",3,"type")="int"
	Set gtmtypfldindx("inc_list_struct","input_file.len")=3
	Set gtmtypes("inc_list_struct",4,"name")="inc_list_struct.input_file.addr"
	Set gtmtypes("inc_list_struct",4,"off")=8
	Set gtmtypes("inc_list_struct",4,"len")=8
	Set gtmtypes("inc_list_struct",4,"type")="addr"
	Set gtmtypfldindx("inc_list_struct","input_file.addr")=4
	Set gtmtypes("inc_list_struct",5,"name")="inc_list_struct.next"
	Set gtmtypes("inc_list_struct",5,"off")=16
	Set gtmtypes("inc_list_struct",5,"len")=8
	Set gtmtypes("inc_list_struct",5,"type")="addr"
	Set gtmtypfldindx("inc_list_struct","next")=5
	;
	Set gtmtypes("inctn_detail_blknum_t")="struct"
	Set gtmtypes("inctn_detail_blknum_t",0)=5
	Set gtmtypes("inctn_detail_blknum_t","len")=16
	Set gtmtypes("inctn_detail_blknum_t",1,"name")="inctn_detail_blknum_t.blknum"
	Set gtmtypes("inctn_detail_blknum_t",1,"off")=0
	Set gtmtypes("inctn_detail_blknum_t",1,"len")=4
	Set gtmtypes("inctn_detail_blknum_t",1,"type")="int"
	Set gtmtypfldindx("inctn_detail_blknum_t","blknum")=1
	Set gtmtypes("inctn_detail_blknum_t",2,"name")="inctn_detail_blknum_t.filler_uint4"
	Set gtmtypes("inctn_detail_blknum_t",2,"off")=4
	Set gtmtypes("inctn_detail_blknum_t",2,"len")=4
	Set gtmtypes("inctn_detail_blknum_t",2,"type")="unsigned-int"
	Set gtmtypfldindx("inctn_detail_blknum_t","filler_uint4")=2
	Set gtmtypes("inctn_detail_blknum_t",3,"name")="inctn_detail_blknum_t.filler_short"
	Set gtmtypes("inctn_detail_blknum_t",3,"off")=8
	Set gtmtypes("inctn_detail_blknum_t",3,"len")=2
	Set gtmtypes("inctn_detail_blknum_t",3,"type")="unsigned-short"
	Set gtmtypfldindx("inctn_detail_blknum_t","filler_short")=3
	Set gtmtypes("inctn_detail_blknum_t",4,"name")="inctn_detail_blknum_t.opcode"
	Set gtmtypes("inctn_detail_blknum_t",4,"off")=10
	Set gtmtypes("inctn_detail_blknum_t",4,"len")=2
	Set gtmtypes("inctn_detail_blknum_t",4,"type")="unsigned-short"
	Set gtmtypfldindx("inctn_detail_blknum_t","opcode")=4
	Set gtmtypes("inctn_detail_blknum_t",5,"name")="inctn_detail_blknum_t.suffix"
	Set gtmtypes("inctn_detail_blknum_t",5,"off")=12
	Set gtmtypes("inctn_detail_blknum_t",5,"len")=4
	Set gtmtypes("inctn_detail_blknum_t",5,"type")="jrec_suffix"
	Set gtmtypfldindx("inctn_detail_blknum_t","suffix")=5
	;
	Set gtmtypes("inctn_detail_blks2upgrd_t")="struct"
	Set gtmtypes("inctn_detail_blks2upgrd_t",0)=5
	Set gtmtypes("inctn_detail_blks2upgrd_t","len")=16
	Set gtmtypes("inctn_detail_blks2upgrd_t",1,"name")="inctn_detail_blks2upgrd_t.blks_to_upgrd_delta"
	Set gtmtypes("inctn_detail_blks2upgrd_t",1,"off")=0
	Set gtmtypes("inctn_detail_blks2upgrd_t",1,"len")=4
	Set gtmtypes("inctn_detail_blks2upgrd_t",1,"type")="int"
	Set gtmtypfldindx("inctn_detail_blks2upgrd_t","blks_to_upgrd_delta")=1
	Set gtmtypes("inctn_detail_blks2upgrd_t",2,"name")="inctn_detail_blks2upgrd_t.filler_uint4"
	Set gtmtypes("inctn_detail_blks2upgrd_t",2,"off")=4
	Set gtmtypes("inctn_detail_blks2upgrd_t",2,"len")=4
	Set gtmtypes("inctn_detail_blks2upgrd_t",2,"type")="unsigned-int"
	Set gtmtypfldindx("inctn_detail_blks2upgrd_t","filler_uint4")=2
	Set gtmtypes("inctn_detail_blks2upgrd_t",3,"name")="inctn_detail_blks2upgrd_t.filler_short"
	Set gtmtypes("inctn_detail_blks2upgrd_t",3,"off")=8
	Set gtmtypes("inctn_detail_blks2upgrd_t",3,"len")=2
	Set gtmtypes("inctn_detail_blks2upgrd_t",3,"type")="unsigned-short"
	Set gtmtypfldindx("inctn_detail_blks2upgrd_t","filler_short")=3
	Set gtmtypes("inctn_detail_blks2upgrd_t",4,"name")="inctn_detail_blks2upgrd_t.opcode"
	Set gtmtypes("inctn_detail_blks2upgrd_t",4,"off")=10
	Set gtmtypes("inctn_detail_blks2upgrd_t",4,"len")=2
	Set gtmtypes("inctn_detail_blks2upgrd_t",4,"type")="unsigned-short"
	Set gtmtypfldindx("inctn_detail_blks2upgrd_t","opcode")=4
	Set gtmtypes("inctn_detail_blks2upgrd_t",5,"name")="inctn_detail_blks2upgrd_t.suffix"
	Set gtmtypes("inctn_detail_blks2upgrd_t",5,"off")=12
	Set gtmtypes("inctn_detail_blks2upgrd_t",5,"len")=4
	Set gtmtypes("inctn_detail_blks2upgrd_t",5,"type")="jrec_suffix"
	Set gtmtypfldindx("inctn_detail_blks2upgrd_t","suffix")=5
	;
	Set gtmtypes("inctn_detail_t")="union"
	Set gtmtypes("inctn_detail_t",0)=12
	Set gtmtypes("inctn_detail_t","len")=16
	Set gtmtypes("inctn_detail_t",1,"name")="inctn_detail_t.blknum_struct"
	Set gtmtypes("inctn_detail_t",1,"off")=0
	Set gtmtypes("inctn_detail_t",1,"len")=16
	Set gtmtypes("inctn_detail_t",1,"type")="inctn_detail_blknum_t"
	Set gtmtypfldindx("inctn_detail_t","blknum_struct")=1
	Set gtmtypes("inctn_detail_t",2,"name")="inctn_detail_t.blknum_struct.blknum"
	Set gtmtypes("inctn_detail_t",2,"off")=0
	Set gtmtypes("inctn_detail_t",2,"len")=4
	Set gtmtypes("inctn_detail_t",2,"type")="int"
	Set gtmtypfldindx("inctn_detail_t","blknum_struct.blknum")=2
	Set gtmtypes("inctn_detail_t",3,"name")="inctn_detail_t.blknum_struct.filler_uint4"
	Set gtmtypes("inctn_detail_t",3,"off")=4
	Set gtmtypes("inctn_detail_t",3,"len")=4
	Set gtmtypes("inctn_detail_t",3,"type")="unsigned-int"
	Set gtmtypfldindx("inctn_detail_t","blknum_struct.filler_uint4")=3
	Set gtmtypes("inctn_detail_t",4,"name")="inctn_detail_t.blknum_struct.filler_short"
	Set gtmtypes("inctn_detail_t",4,"off")=8
	Set gtmtypes("inctn_detail_t",4,"len")=2
	Set gtmtypes("inctn_detail_t",4,"type")="unsigned-short"
	Set gtmtypfldindx("inctn_detail_t","blknum_struct.filler_short")=4
	Set gtmtypes("inctn_detail_t",5,"name")="inctn_detail_t.blknum_struct.opcode"
	Set gtmtypes("inctn_detail_t",5,"off")=10
	Set gtmtypes("inctn_detail_t",5,"len")=2
	Set gtmtypes("inctn_detail_t",5,"type")="unsigned-short"
	Set gtmtypfldindx("inctn_detail_t","blknum_struct.opcode")=5
	Set gtmtypes("inctn_detail_t",6,"name")="inctn_detail_t.blknum_struct.suffix"
	Set gtmtypes("inctn_detail_t",6,"off")=12
	Set gtmtypes("inctn_detail_t",6,"len")=4
	Set gtmtypes("inctn_detail_t",6,"type")="jrec_suffix"
	Set gtmtypfldindx("inctn_detail_t","blknum_struct.suffix")=6
	Set gtmtypes("inctn_detail_t",7,"name")="inctn_detail_t.blks2upgrd_struct"
	Set gtmtypes("inctn_detail_t",7,"off")=0
	Set gtmtypes("inctn_detail_t",7,"len")=16
	Set gtmtypes("inctn_detail_t",7,"type")="inctn_detail_blks2upgrd_t"
	Set gtmtypfldindx("inctn_detail_t","blks2upgrd_struct")=7
	Set gtmtypes("inctn_detail_t",8,"name")="inctn_detail_t.blks2upgrd_struct.blks_to_upgrd_delta"
	Set gtmtypes("inctn_detail_t",8,"off")=0
	Set gtmtypes("inctn_detail_t",8,"len")=4
	Set gtmtypes("inctn_detail_t",8,"type")="int"
	Set gtmtypfldindx("inctn_detail_t","blks2upgrd_struct.blks_to_upgrd_delta")=8
	Set gtmtypes("inctn_detail_t",9,"name")="inctn_detail_t.blks2upgrd_struct.filler_uint4"
	Set gtmtypes("inctn_detail_t",9,"off")=4
	Set gtmtypes("inctn_detail_t",9,"len")=4
	Set gtmtypes("inctn_detail_t",9,"type")="unsigned-int"
	Set gtmtypfldindx("inctn_detail_t","blks2upgrd_struct.filler_uint4")=9
	Set gtmtypes("inctn_detail_t",10,"name")="inctn_detail_t.blks2upgrd_struct.filler_short"
	Set gtmtypes("inctn_detail_t",10,"off")=8
	Set gtmtypes("inctn_detail_t",10,"len")=2
	Set gtmtypes("inctn_detail_t",10,"type")="unsigned-short"
	Set gtmtypfldindx("inctn_detail_t","blks2upgrd_struct.filler_short")=10
	Set gtmtypes("inctn_detail_t",11,"name")="inctn_detail_t.blks2upgrd_struct.opcode"
	Set gtmtypes("inctn_detail_t",11,"off")=10
	Set gtmtypes("inctn_detail_t",11,"len")=2
	Set gtmtypes("inctn_detail_t",11,"type")="unsigned-short"
	Set gtmtypfldindx("inctn_detail_t","blks2upgrd_struct.opcode")=11
	Set gtmtypes("inctn_detail_t",12,"name")="inctn_detail_t.blks2upgrd_struct.suffix"
	Set gtmtypes("inctn_detail_t",12,"off")=12
	Set gtmtypes("inctn_detail_t",12,"len")=4
	Set gtmtypes("inctn_detail_t",12,"type")="jrec_suffix"
	Set gtmtypfldindx("inctn_detail_t","blks2upgrd_struct.suffix")=12
	;
	Set gtmtypes("integ_error_blk_list")="struct"
	Set gtmtypes("integ_error_blk_list",0)=3
	Set gtmtypes("integ_error_blk_list","len")=56
	Set gtmtypes("integ_error_blk_list",1,"name")="integ_error_blk_list.next"
	Set gtmtypes("integ_error_blk_list",1,"off")=0
	Set gtmtypes("integ_error_blk_list",1,"len")=8
	Set gtmtypes("integ_error_blk_list",1,"type")="addr"
	Set gtmtypfldindx("integ_error_blk_list","next")=1
	Set gtmtypes("integ_error_blk_list",2,"name")="integ_error_blk_list.blk_cnt"
	Set gtmtypes("integ_error_blk_list",2,"off")=8
	Set gtmtypes("integ_error_blk_list",2,"len")=4
	Set gtmtypes("integ_error_blk_list",2,"type")="int"
	Set gtmtypfldindx("integ_error_blk_list","blk_cnt")=2
	Set gtmtypes("integ_error_blk_list",3,"name")="integ_error_blk_list.blk_list"
	Set gtmtypes("integ_error_blk_list",3,"off")=12
	Set gtmtypes("integ_error_blk_list",3,"len")=40
	Set gtmtypes("integ_error_blk_list",3,"type")="int"
	Set gtmtypfldindx("integ_error_blk_list","blk_list")=3
	Set gtmtypes("integ_error_blk_list",3,"dim")=10
	;
	Set gtmtypes("io_desc")="struct"
	Set gtmtypes("io_desc",0)=33
	Set gtmtypes("io_desc","len")=320
	Set gtmtypes("io_desc",1,"name")="io_desc.pair"
	Set gtmtypes("io_desc",1,"off")=0
	Set gtmtypes("io_desc",1,"len")=16
	Set gtmtypes("io_desc",1,"type")="io_pair"
	Set gtmtypfldindx("io_desc","pair")=1
	Set gtmtypes("io_desc",2,"name")="io_desc.pair.in"
	Set gtmtypes("io_desc",2,"off")=0
	Set gtmtypes("io_desc",2,"len")=8
	Set gtmtypes("io_desc",2,"type")="addr"
	Set gtmtypfldindx("io_desc","pair.in")=2
	Set gtmtypes("io_desc",3,"name")="io_desc.pair.out"
	Set gtmtypes("io_desc",3,"off")=8
	Set gtmtypes("io_desc",3,"len")=8
	Set gtmtypes("io_desc",3,"type")="addr"
	Set gtmtypfldindx("io_desc","pair.out")=3
	Set gtmtypes("io_desc",4,"name")="io_desc.trans_name"
	Set gtmtypes("io_desc",4,"off")=16
	Set gtmtypes("io_desc",4,"len")=8
	Set gtmtypes("io_desc",4,"type")="addr"
	Set gtmtypfldindx("io_desc","trans_name")=4
	Set gtmtypes("io_desc",5,"name")="io_desc.name"
	Set gtmtypes("io_desc",5,"off")=24
	Set gtmtypes("io_desc",5,"len")=8
	Set gtmtypes("io_desc",5,"type")="addr"
	Set gtmtypfldindx("io_desc","name")=5
	Set gtmtypes("io_desc",6,"name")="io_desc.error_handler"
	Set gtmtypes("io_desc",6,"off")=32
	Set gtmtypes("io_desc",6,"len")=16
	Set gtmtypes("io_desc",6,"type")="mstr"
	Set gtmtypfldindx("io_desc","error_handler")=6
	Set gtmtypes("io_desc",7,"name")="io_desc.error_handler.char_len"
	Set gtmtypes("io_desc",7,"off")=32
	Set gtmtypes("io_desc",7,"len")=4
	Set gtmtypes("io_desc",7,"type")="unsigned-int"
	Set gtmtypfldindx("io_desc","error_handler.char_len")=7
	Set gtmtypes("io_desc",8,"name")="io_desc.error_handler.len"
	Set gtmtypes("io_desc",8,"off")=36
	Set gtmtypes("io_desc",8,"len")=4
	Set gtmtypes("io_desc",8,"type")="int"
	Set gtmtypfldindx("io_desc","error_handler.len")=8
	Set gtmtypes("io_desc",9,"name")="io_desc.error_handler.addr"
	Set gtmtypes("io_desc",9,"off")=40
	Set gtmtypes("io_desc",9,"len")=8
	Set gtmtypes("io_desc",9,"type")="addr"
	Set gtmtypfldindx("io_desc","error_handler.addr")=9
	Set gtmtypes("io_desc",10,"name")="io_desc.length"
	Set gtmtypes("io_desc",10,"off")=48
	Set gtmtypes("io_desc",10,"len")=4
	Set gtmtypes("io_desc",10,"type")="unsigned-int"
	Set gtmtypfldindx("io_desc","length")=10
	Set gtmtypes("io_desc",11,"name")="io_desc.width"
	Set gtmtypes("io_desc",11,"off")=52
	Set gtmtypes("io_desc",11,"len")=4
	Set gtmtypes("io_desc",11,"type")="unsigned-int"
	Set gtmtypfldindx("io_desc","width")=11
	Set gtmtypes("io_desc",12,"name")="io_desc.perm"
	Set gtmtypes("io_desc",12,"off")=56
	Set gtmtypes("io_desc",12,"len")=1
	Set gtmtypes("io_desc",12,"type")="char"
	Set gtmtypfldindx("io_desc","perm")=12
	Set gtmtypes("io_desc",13,"name")="io_desc.wrap"
	Set gtmtypes("io_desc",13,"off")=57
	Set gtmtypes("io_desc",13,"len")=1
	Set gtmtypes("io_desc",13,"type")="char"
	Set gtmtypfldindx("io_desc","wrap")=13
	Set gtmtypes("io_desc",14,"name")="io_desc.fflf"
	Set gtmtypes("io_desc",14,"off")=60
	Set gtmtypes("io_desc",14,"len")=4
	Set gtmtypes("io_desc",14,"type")="boolean_t"
	Set gtmtypfldindx("io_desc","fflf")=14
	Set gtmtypes("io_desc",15,"name")="io_desc.type"
	Set gtmtypes("io_desc",15,"off")=64
	Set gtmtypes("io_desc",15,"len")=4
	Set gtmtypes("io_desc",15,"type")="int"
	Set gtmtypfldindx("io_desc","type")=15
	Set gtmtypes("io_desc",16,"name")="io_desc.state"
	Set gtmtypes("io_desc",16,"off")=68
	Set gtmtypes("io_desc",16,"len")=4
	Set gtmtypes("io_desc",16,"type")="int"
	Set gtmtypfldindx("io_desc","state")=16
	Set gtmtypes("io_desc",17,"name")="io_desc.dollar"
	Set gtmtypes("io_desc",17,"off")=72
	Set gtmtypes("io_desc",17,"len")=208
	Set gtmtypes("io_desc",17,"type")="struct"
	Set gtmtypfldindx("io_desc","dollar")=17
	Set gtmtypes("io_desc",18,"name")="io_desc.dollar.x"
	Set gtmtypes("io_desc",18,"off")=72
	Set gtmtypes("io_desc",18,"len")=4
	Set gtmtypes("io_desc",18,"type")="unsigned-int"
	Set gtmtypfldindx("io_desc","dollar.x")=18
	Set gtmtypes("io_desc",19,"name")="io_desc.dollar.y"
	Set gtmtypes("io_desc",19,"off")=76
	Set gtmtypes("io_desc",19,"len")=4
	Set gtmtypes("io_desc",19,"type")="unsigned-int"
	Set gtmtypfldindx("io_desc","dollar.y")=19
	Set gtmtypes("io_desc",20,"name")="io_desc.dollar.zeof"
	Set gtmtypes("io_desc",20,"off")=80
	Set gtmtypes("io_desc",20,"len")=2
	Set gtmtypes("io_desc",20,"type")="unsigned-short"
	Set gtmtypfldindx("io_desc","dollar.zeof")=20
	Set gtmtypes("io_desc",21,"name")="io_desc.dollar.za"
	Set gtmtypes("io_desc",21,"off")=82
	Set gtmtypes("io_desc",21,"len")=2
	Set gtmtypes("io_desc",21,"type")="unsigned-short"
	Set gtmtypfldindx("io_desc","dollar.za")=21
	Set gtmtypes("io_desc",22,"name")="io_desc.dollar.zb"
	Set gtmtypes("io_desc",22,"off")=84
	Set gtmtypes("io_desc",22,"len")=16
	Set gtmtypes("io_desc",22,"type")="unsigned-char"
	Set gtmtypfldindx("io_desc","dollar.zb")=22
	Set gtmtypes("io_desc",23,"name")="io_desc.dollar.key"
	Set gtmtypes("io_desc",23,"off")=100
	Set gtmtypes("io_desc",23,"len")=80
	Set gtmtypes("io_desc",23,"type")="char"
	Set gtmtypfldindx("io_desc","dollar.key")=23
	Set gtmtypes("io_desc",24,"name")="io_desc.dollar.device"
	Set gtmtypes("io_desc",24,"off")=180
	Set gtmtypes("io_desc",24,"len")=80
	Set gtmtypes("io_desc",24,"type")="char"
	Set gtmtypfldindx("io_desc","dollar.device")=24
	Set gtmtypes("io_desc",25,"name")="io_desc.dollar.devicebuffer"
	Set gtmtypes("io_desc",25,"off")=264
	Set gtmtypes("io_desc",25,"len")=8
	Set gtmtypes("io_desc",25,"type")="addr"
	Set gtmtypfldindx("io_desc","dollar.devicebuffer")=25
	Set gtmtypes("io_desc",26,"name")="io_desc.dollar.devicebufferlen"
	Set gtmtypes("io_desc",26,"off")=272
	Set gtmtypes("io_desc",26,"len")=4
	Set gtmtypes("io_desc",26,"type")="int"
	Set gtmtypfldindx("io_desc","dollar.devicebufferlen")=26
	Set gtmtypes("io_desc",27,"name")="io_desc.esc_state"
	Set gtmtypes("io_desc",27,"off")=280
	Set gtmtypes("io_desc",27,"len")=1
	Set gtmtypes("io_desc",27,"type")="unsigned-char"
	Set gtmtypfldindx("io_desc","esc_state")=27
	Set gtmtypes("io_desc",28,"name")="io_desc.dev_sp"
	Set gtmtypes("io_desc",28,"off")=288
	Set gtmtypes("io_desc",28,"len")=8
	Set gtmtypes("io_desc",28,"type")="addr"
	Set gtmtypfldindx("io_desc","dev_sp")=28
	Set gtmtypes("io_desc",29,"name")="io_desc.disp_ptr"
	Set gtmtypes("io_desc",29,"off")=296
	Set gtmtypes("io_desc",29,"len")=8
	Set gtmtypes("io_desc",29,"type")="addr"
	Set gtmtypfldindx("io_desc","disp_ptr")=29
	Set gtmtypes("io_desc",30,"name")="io_desc.newly_created"
	Set gtmtypes("io_desc",30,"off")=304
	Set gtmtypes("io_desc",30,"len")=4
	Set gtmtypes("io_desc",30,"type")="boolean_t"
	Set gtmtypfldindx("io_desc","newly_created")=30
	Set gtmtypes("io_desc",31,"name")="io_desc.ichset"
	Set gtmtypes("io_desc",31,"off")=308
	Set gtmtypes("io_desc",31,"len")=4
	Set gtmtypes("io_desc",31,"type")="int"
	Set gtmtypfldindx("io_desc","ichset")=31
	Set gtmtypes("io_desc",32,"name")="io_desc.ochset"
	Set gtmtypes("io_desc",32,"off")=312
	Set gtmtypes("io_desc",32,"len")=4
	Set gtmtypes("io_desc",32,"type")="int"
	Set gtmtypfldindx("io_desc","ochset")=32
	Set gtmtypes("io_desc",33,"name")="io_desc.write_filter"
	Set gtmtypes("io_desc",33,"off")=316
	Set gtmtypes("io_desc",33,"len")=4
	Set gtmtypes("io_desc",33,"type")="int"
	Set gtmtypfldindx("io_desc","write_filter")=33
	;
	Set gtmtypes("io_log_name")="struct"
	Set gtmtypes("io_log_name",0)=4
	Set gtmtypes("io_log_name","len")=24
	Set gtmtypes("io_log_name",1,"name")="io_log_name.iod"
	Set gtmtypes("io_log_name",1,"off")=0
	Set gtmtypes("io_log_name",1,"len")=8
	Set gtmtypes("io_log_name",1,"type")="addr"
	Set gtmtypfldindx("io_log_name","iod")=1
	Set gtmtypes("io_log_name",2,"name")="io_log_name.next"
	Set gtmtypes("io_log_name",2,"off")=8
	Set gtmtypes("io_log_name",2,"len")=8
	Set gtmtypes("io_log_name",2,"type")="addr"
	Set gtmtypfldindx("io_log_name","next")=2
	Set gtmtypes("io_log_name",3,"name")="io_log_name.len"
	Set gtmtypes("io_log_name",3,"off")=16
	Set gtmtypes("io_log_name",3,"len")=1
	Set gtmtypes("io_log_name",3,"type")="unsigned-char"
	Set gtmtypfldindx("io_log_name","len")=3
	Set gtmtypes("io_log_name",4,"name")="io_log_name.dollar_io"
	Set gtmtypes("io_log_name",4,"off")=17
	Set gtmtypes("io_log_name",4,"len")=1
	Set gtmtypes("io_log_name",4,"type")="char"
	Set gtmtypfldindx("io_log_name","dollar_io")=4
	;
	Set gtmtypes("io_pair")="struct"
	Set gtmtypes("io_pair",0)=2
	Set gtmtypes("io_pair","len")=16
	Set gtmtypes("io_pair",1,"name")="io_pair.in"
	Set gtmtypes("io_pair",1,"off")=0
	Set gtmtypes("io_pair",1,"len")=8
	Set gtmtypes("io_pair",1,"type")="addr"
	Set gtmtypfldindx("io_pair","in")=1
	Set gtmtypes("io_pair",2,"name")="io_pair.out"
	Set gtmtypes("io_pair",2,"off")=8
	Set gtmtypes("io_pair",2,"len")=8
	Set gtmtypes("io_pair",2,"type")="addr"
	Set gtmtypfldindx("io_pair","out")=2
	;
	Set gtmtypes("io_terminator")="struct"
	Set gtmtypes("io_terminator",0)=2
	Set gtmtypes("io_terminator","len")=8
	Set gtmtypes("io_terminator",1,"name")="io_terminator.x"
	Set gtmtypes("io_terminator",1,"off")=0
	Set gtmtypes("io_terminator",1,"len")=4
	Set gtmtypes("io_terminator",1,"type")="unsigned-int"
	Set gtmtypfldindx("io_terminator","x")=1
	Set gtmtypes("io_terminator",2,"name")="io_terminator.mask"
	Set gtmtypes("io_terminator",2,"off")=4
	Set gtmtypes("io_terminator",2,"len")=4
	Set gtmtypes("io_terminator",2,"type")="unsigned-int"
	Set gtmtypfldindx("io_terminator","mask")=2
	;
	Set gtmtypes("io_termmask")="struct"
	Set gtmtypes("io_termmask",0)=1
	Set gtmtypes("io_termmask","len")=32
	Set gtmtypes("io_termmask",1,"name")="io_termmask.mask"
	Set gtmtypes("io_termmask",1,"off")=0
	Set gtmtypes("io_termmask",1,"len")=32
	Set gtmtypes("io_termmask",1,"type")="unsigned-int"
	Set gtmtypfldindx("io_termmask","mask")=1
	Set gtmtypes("io_termmask",1,"dim")=8
	;
	Set gtmtypes("iosb")="struct"
	Set gtmtypes("iosb",0)=3
	Set gtmtypes("iosb","len")=8
	Set gtmtypes("iosb",1,"name")="iosb.status"
	Set gtmtypes("iosb",1,"off")=0
	Set gtmtypes("iosb",1,"len")=2
	Set gtmtypes("iosb",1,"type")="unsigned-short"
	Set gtmtypfldindx("iosb","status")=1
	Set gtmtypes("iosb",2,"name")="iosb.char_ct"
	Set gtmtypes("iosb",2,"off")=2
	Set gtmtypes("iosb",2,"len")=2
	Set gtmtypes("iosb",2,"type")="unsigned-short"
	Set gtmtypfldindx("iosb","char_ct")=2
	Set gtmtypes("iosb",3,"name")="iosb.dev_dep_info"
	Set gtmtypes("iosb",3,"off")=4
	Set gtmtypes("iosb",3,"len")=4
	Set gtmtypes("iosb",3,"type")="unsigned-int"
	Set gtmtypfldindx("iosb","dev_dep_info")=3
	;
	Set gtmtypes("ipcs_mesg")="struct"
	Set gtmtypes("ipcs_mesg",0)=7
	Set gtmtypes("ipcs_mesg","len")=4136
	Set gtmtypes("ipcs_mesg",1,"name")="ipcs_mesg.open_fd_with_o_direct"
	Set gtmtypes("ipcs_mesg",1,"off")=0
	Set gtmtypes("ipcs_mesg",1,"len")=4
	Set gtmtypes("ipcs_mesg",1,"type")="boolean_t"
	Set gtmtypfldindx("ipcs_mesg","open_fd_with_o_direct")=1
	Set gtmtypes("ipcs_mesg",2,"name")="ipcs_mesg.semid"
	Set gtmtypes("ipcs_mesg",2,"off")=4
	Set gtmtypes("ipcs_mesg",2,"len")=4
	Set gtmtypes("ipcs_mesg",2,"type")="int"
	Set gtmtypfldindx("ipcs_mesg","semid")=2
	Set gtmtypes("ipcs_mesg",3,"name")="ipcs_mesg.shmid"
	Set gtmtypes("ipcs_mesg",3,"off")=8
	Set gtmtypes("ipcs_mesg",3,"len")=4
	Set gtmtypes("ipcs_mesg",3,"type")="int"
	Set gtmtypfldindx("ipcs_mesg","shmid")=3
	Set gtmtypes("ipcs_mesg",4,"name")="ipcs_mesg.gt_sem_ctime"
	Set gtmtypes("ipcs_mesg",4,"off")=16
	Set gtmtypes("ipcs_mesg",4,"len")=8
	Set gtmtypes("ipcs_mesg",4,"type")="time_t"
	Set gtmtypfldindx("ipcs_mesg","gt_sem_ctime")=4
	Set gtmtypes("ipcs_mesg",5,"name")="ipcs_mesg.gt_shm_ctime"
	Set gtmtypes("ipcs_mesg",5,"off")=24
	Set gtmtypes("ipcs_mesg",5,"len")=8
	Set gtmtypes("ipcs_mesg",5,"type")="time_t"
	Set gtmtypfldindx("ipcs_mesg","gt_shm_ctime")=5
	Set gtmtypes("ipcs_mesg",6,"name")="ipcs_mesg.fn_len"
	Set gtmtypes("ipcs_mesg",6,"off")=32
	Set gtmtypes("ipcs_mesg",6,"len")=4
	Set gtmtypes("ipcs_mesg",6,"type")="unsigned-int"
	Set gtmtypfldindx("ipcs_mesg","fn_len")=6
	Set gtmtypes("ipcs_mesg",7,"name")="ipcs_mesg.fn"
	Set gtmtypes("ipcs_mesg",7,"off")=36
	Set gtmtypes("ipcs_mesg",7,"len")=4097
	Set gtmtypes("ipcs_mesg",7,"type")="char"
	Set gtmtypfldindx("ipcs_mesg","fn")=7
	;
	Set gtmtypes("jbuf_phase2_in_prog_t")="struct"
	Set gtmtypes("jbuf_phase2_in_prog_t",0)=12
	Set gtmtypes("jbuf_phase2_in_prog_t","len")=48
	Set gtmtypes("jbuf_phase2_in_prog_t",1,"name")="jbuf_phase2_in_prog_t.curr_tn"
	Set gtmtypes("jbuf_phase2_in_prog_t",1,"off")=0
	Set gtmtypes("jbuf_phase2_in_prog_t",1,"len")=8
	Set gtmtypes("jbuf_phase2_in_prog_t",1,"type")="uint64_t"
	Set gtmtypfldindx("jbuf_phase2_in_prog_t","curr_tn")=1
	Set gtmtypes("jbuf_phase2_in_prog_t",2,"name")="jbuf_phase2_in_prog_t.jnl_seqno"
	Set gtmtypes("jbuf_phase2_in_prog_t",2,"off")=8
	Set gtmtypes("jbuf_phase2_in_prog_t",2,"len")=8
	Set gtmtypes("jbuf_phase2_in_prog_t",2,"type")="uint64_t"
	Set gtmtypfldindx("jbuf_phase2_in_prog_t","jnl_seqno")=2
	Set gtmtypes("jbuf_phase2_in_prog_t",3,"name")="jbuf_phase2_in_prog_t.strm_seqno"
	Set gtmtypes("jbuf_phase2_in_prog_t",3,"off")=16
	Set gtmtypes("jbuf_phase2_in_prog_t",3,"len")=8
	Set gtmtypes("jbuf_phase2_in_prog_t",3,"type")="uint64_t"
	Set gtmtypfldindx("jbuf_phase2_in_prog_t","strm_seqno")=3
	Set gtmtypes("jbuf_phase2_in_prog_t",4,"name")="jbuf_phase2_in_prog_t.process_id"
	Set gtmtypes("jbuf_phase2_in_prog_t",4,"off")=24
	Set gtmtypes("jbuf_phase2_in_prog_t",4,"len")=4
	Set gtmtypes("jbuf_phase2_in_prog_t",4,"type")="unsigned-int"
	Set gtmtypfldindx("jbuf_phase2_in_prog_t","process_id")=4
	Set gtmtypes("jbuf_phase2_in_prog_t",5,"name")="jbuf_phase2_in_prog_t.start_freeaddr"
	Set gtmtypes("jbuf_phase2_in_prog_t",5,"off")=28
	Set gtmtypes("jbuf_phase2_in_prog_t",5,"len")=4
	Set gtmtypes("jbuf_phase2_in_prog_t",5,"type")="unsigned-int"
	Set gtmtypfldindx("jbuf_phase2_in_prog_t","start_freeaddr")=5
	Set gtmtypes("jbuf_phase2_in_prog_t",6,"name")="jbuf_phase2_in_prog_t.tot_jrec_len"
	Set gtmtypes("jbuf_phase2_in_prog_t",6,"off")=32
	Set gtmtypes("jbuf_phase2_in_prog_t",6,"len")=4
	Set gtmtypes("jbuf_phase2_in_prog_t",6,"type")="unsigned-int"
	Set gtmtypfldindx("jbuf_phase2_in_prog_t","tot_jrec_len")=6
	Set gtmtypes("jbuf_phase2_in_prog_t",7,"name")="jbuf_phase2_in_prog_t.pini_addr"
	Set gtmtypes("jbuf_phase2_in_prog_t",7,"off")=36
	Set gtmtypes("jbuf_phase2_in_prog_t",7,"len")=4
	Set gtmtypes("jbuf_phase2_in_prog_t",7,"type")="unsigned-int"
	Set gtmtypfldindx("jbuf_phase2_in_prog_t","pini_addr")=7
	Set gtmtypes("jbuf_phase2_in_prog_t",8,"name")="jbuf_phase2_in_prog_t.jrec_time"
	Set gtmtypes("jbuf_phase2_in_prog_t",8,"off")=40
	Set gtmtypes("jbuf_phase2_in_prog_t",8,"len")=4
	Set gtmtypes("jbuf_phase2_in_prog_t",8,"type")="unsigned-int"
	Set gtmtypfldindx("jbuf_phase2_in_prog_t","jrec_time")=8
	Set gtmtypes("jbuf_phase2_in_prog_t",9,"name")="jbuf_phase2_in_prog_t.in_phase2"
	Set gtmtypes("jbuf_phase2_in_prog_t",9,"off")=44
	Set gtmtypes("jbuf_phase2_in_prog_t",9,"len")=1
	Set gtmtypes("jbuf_phase2_in_prog_t",9,"type")="char"
	Set gtmtypfldindx("jbuf_phase2_in_prog_t","in_phase2")=9
	Set gtmtypes("jbuf_phase2_in_prog_t",10,"name")="jbuf_phase2_in_prog_t.replication"
	Set gtmtypes("jbuf_phase2_in_prog_t",10,"off")=45
	Set gtmtypes("jbuf_phase2_in_prog_t",10,"len")=1
	Set gtmtypes("jbuf_phase2_in_prog_t",10,"type")="char"
	Set gtmtypfldindx("jbuf_phase2_in_prog_t","replication")=10
	Set gtmtypes("jbuf_phase2_in_prog_t",11,"name")="jbuf_phase2_in_prog_t.write_complete"
	Set gtmtypes("jbuf_phase2_in_prog_t",11,"off")=46
	Set gtmtypes("jbuf_phase2_in_prog_t",11,"len")=1
	Set gtmtypes("jbuf_phase2_in_prog_t",11,"type")="char"
	Set gtmtypfldindx("jbuf_phase2_in_prog_t","write_complete")=11
	Set gtmtypes("jbuf_phase2_in_prog_t",12,"name")="jbuf_phase2_in_prog_t.filler_8byte_align"
	Set gtmtypes("jbuf_phase2_in_prog_t",12,"off")=47
	Set gtmtypes("jbuf_phase2_in_prog_t",12,"len")=1
	Set gtmtypes("jbuf_phase2_in_prog_t",12,"type")="char"
	Set gtmtypfldindx("jbuf_phase2_in_prog_t","filler_8byte_align")=12
	;
	Set gtmtypes("jbuf_rsrv_struct_t")="struct"
	Set gtmtypes("jbuf_rsrv_struct_t",0)=5
	Set gtmtypes("jbuf_rsrv_struct_t","len")=24
	Set gtmtypes("jbuf_rsrv_struct_t",1,"name")="jbuf_rsrv_struct_t.alloclen"
	Set gtmtypes("jbuf_rsrv_struct_t",1,"off")=0
	Set gtmtypes("jbuf_rsrv_struct_t",1,"len")=4
	Set gtmtypes("jbuf_rsrv_struct_t",1,"type")="unsigned-int"
	Set gtmtypfldindx("jbuf_rsrv_struct_t","alloclen")=1
	Set gtmtypes("jbuf_rsrv_struct_t",2,"name")="jbuf_rsrv_struct_t.usedlen"
	Set gtmtypes("jbuf_rsrv_struct_t",2,"off")=4
	Set gtmtypes("jbuf_rsrv_struct_t",2,"len")=4
	Set gtmtypes("jbuf_rsrv_struct_t",2,"type")="unsigned-int"
	Set gtmtypfldindx("jbuf_rsrv_struct_t","usedlen")=2
	Set gtmtypes("jbuf_rsrv_struct_t",3,"name")="jbuf_rsrv_struct_t.tot_jrec_len"
	Set gtmtypes("jbuf_rsrv_struct_t",3,"off")=8
	Set gtmtypes("jbuf_rsrv_struct_t",3,"len")=4
	Set gtmtypes("jbuf_rsrv_struct_t",3,"type")="unsigned-int"
	Set gtmtypfldindx("jbuf_rsrv_struct_t","tot_jrec_len")=3
	Set gtmtypes("jbuf_rsrv_struct_t",4,"name")="jbuf_rsrv_struct_t.phase2_commit_index"
	Set gtmtypes("jbuf_rsrv_struct_t",4,"off")=12
	Set gtmtypes("jbuf_rsrv_struct_t",4,"len")=4
	Set gtmtypes("jbuf_rsrv_struct_t",4,"type")="unsigned-int"
	Set gtmtypfldindx("jbuf_rsrv_struct_t","phase2_commit_index")=4
	Set gtmtypes("jbuf_rsrv_struct_t",5,"name")="jbuf_rsrv_struct_t.jrs_array"
	Set gtmtypes("jbuf_rsrv_struct_t",5,"off")=16
	Set gtmtypes("jbuf_rsrv_struct_t",5,"len")=8
	Set gtmtypes("jbuf_rsrv_struct_t",5,"type")="addr"
	Set gtmtypfldindx("jbuf_rsrv_struct_t","jrs_array")=5
	;
	Set gtmtypes("jext_heap_elem_t")="struct"
	Set gtmtypes("jext_heap_elem_t",0)=2
	Set gtmtypes("jext_heap_elem_t","len")=16
	Set gtmtypes("jext_heap_elem_t",1,"name")="jext_heap_elem_t.rctl_index"
	Set gtmtypes("jext_heap_elem_t",1,"off")=0
	Set gtmtypes("jext_heap_elem_t",1,"len")=4
	Set gtmtypes("jext_heap_elem_t",1,"type")="int"
	Set gtmtypfldindx("jext_heap_elem_t","rctl_index")=1
	Set gtmtypes("jext_heap_elem_t",2,"name")="jext_heap_elem_t.jext_rec"
	Set gtmtypes("jext_heap_elem_t",2,"off")=8
	Set gtmtypes("jext_heap_elem_t",2,"len")=8
	Set gtmtypes("jext_heap_elem_t",2,"type")="addr"
	Set gtmtypfldindx("jext_heap_elem_t","jext_rec")=2
	;
	Set gtmtypes("jnl_action")="struct"
	Set gtmtypes("jnl_action",0)=2
	Set gtmtypes("jnl_action","len")=8
	Set gtmtypes("jnl_action",1,"name")="jnl_action.operation"
	Set gtmtypes("jnl_action",1,"off")=0
	Set gtmtypes("jnl_action",1,"len")=4
	Set gtmtypes("jnl_action",1,"type")="int"
	Set gtmtypfldindx("jnl_action","operation")=1
	Set gtmtypes("jnl_action",2,"name")="jnl_action.nodeflags"
	Set gtmtypes("jnl_action",2,"off")=4
	Set gtmtypes("jnl_action",2,"len")=4
	Set gtmtypes("jnl_action",2,"type")="unsigned-int"
	Set gtmtypfldindx("jnl_action","nodeflags")=2
	;
	Set gtmtypes("jnl_buffer")="struct"
	Set gtmtypes("jnl_buffer",0)=85
	Set gtmtypes("jnl_buffer","len")=197360
	Set gtmtypes("jnl_buffer",1,"name")="jnl_buffer.eov_tn"
	Set gtmtypes("jnl_buffer",1,"off")=0
	Set gtmtypes("jnl_buffer",1,"len")=8
	Set gtmtypes("jnl_buffer",1,"type")="uint64_t"
	Set gtmtypfldindx("jnl_buffer","eov_tn")=1
	Set gtmtypes("jnl_buffer",2,"name")="jnl_buffer.epoch_tn"
	Set gtmtypes("jnl_buffer",2,"off")=8
	Set gtmtypes("jnl_buffer",2,"len")=8
	Set gtmtypes("jnl_buffer",2,"type")="uint64_t"
	Set gtmtypfldindx("jnl_buffer","epoch_tn")=2
	Set gtmtypes("jnl_buffer",3,"name")="jnl_buffer.end_seqno"
	Set gtmtypes("jnl_buffer",3,"off")=16
	Set gtmtypes("jnl_buffer",3,"len")=8
	Set gtmtypes("jnl_buffer",3,"type")="uint64_t"
	Set gtmtypfldindx("jnl_buffer","end_seqno")=3
	Set gtmtypes("jnl_buffer",4,"name")="jnl_buffer.strm_end_seqno"
	Set gtmtypes("jnl_buffer",4,"off")=24
	Set gtmtypes("jnl_buffer",4,"len")=128
	Set gtmtypes("jnl_buffer",4,"type")="uint64_t"
	Set gtmtypfldindx("jnl_buffer","strm_end_seqno")=4
	Set gtmtypes("jnl_buffer",4,"dim")=16
	Set gtmtypes("jnl_buffer",5,"name")="jnl_buffer.min_write_size"
	Set gtmtypes("jnl_buffer",5,"off")=152
	Set gtmtypes("jnl_buffer",5,"len")=4
	Set gtmtypes("jnl_buffer",5,"type")="int"
	Set gtmtypfldindx("jnl_buffer","min_write_size")=5
	Set gtmtypes("jnl_buffer",6,"name")="jnl_buffer.max_write_size"
	Set gtmtypes("jnl_buffer",6,"off")=156
	Set gtmtypes("jnl_buffer",6,"len")=4
	Set gtmtypes("jnl_buffer",6,"type")="int"
	Set gtmtypfldindx("jnl_buffer","max_write_size")=6
	Set gtmtypes("jnl_buffer",7,"name")="jnl_buffer.size"
	Set gtmtypes("jnl_buffer",7,"off")=160
	Set gtmtypes("jnl_buffer",7,"len")=4
	Set gtmtypes("jnl_buffer",7,"type")="int"
	Set gtmtypfldindx("jnl_buffer","size")=7
	Set gtmtypes("jnl_buffer",8,"name")="jnl_buffer.epoch_interval"
	Set gtmtypes("jnl_buffer",8,"off")=164
	Set gtmtypes("jnl_buffer",8,"len")=4
	Set gtmtypes("jnl_buffer",8,"type")="int"
	Set gtmtypfldindx("jnl_buffer","epoch_interval")=8
	Set gtmtypes("jnl_buffer",9,"name")="jnl_buffer.before_images"
	Set gtmtypes("jnl_buffer",9,"off")=168
	Set gtmtypes("jnl_buffer",9,"len")=4
	Set gtmtypes("jnl_buffer",9,"type")="boolean_t"
	Set gtmtypfldindx("jnl_buffer","before_images")=9
	Set gtmtypes("jnl_buffer",10,"name")="jnl_buffer.buff_off"
	Set gtmtypes("jnl_buffer",10,"off")=176
	Set gtmtypes("jnl_buffer",10,"len")=8
	Set gtmtypes("jnl_buffer",10,"type")="uintptr_t"
	Set gtmtypfldindx("jnl_buffer","buff_off")=10
	Set gtmtypes("jnl_buffer",11,"name")="jnl_buffer.gtm_free"
	Set gtmtypes("jnl_buffer",11,"off")=184
	Set gtmtypes("jnl_buffer",11,"len")=4
	Set gtmtypes("jnl_buffer",11,"type")="int"
	Set gtmtypfldindx("jnl_buffer","gtm_free")=11
	Set gtmtypes("jnl_buffer",12,"name")="jnl_buffer.rsrv_free"
	Set gtmtypes("jnl_buffer",12,"off")=188
	Set gtmtypes("jnl_buffer",12,"len")=4
	Set gtmtypes("jnl_buffer",12,"type")="int"
	Set gtmtypfldindx("jnl_buffer","rsrv_free")=12
	Set gtmtypes("jnl_buffer",13,"name")="jnl_buffer.freeaddr"
	Set gtmtypes("jnl_buffer",13,"off")=192
	Set gtmtypes("jnl_buffer",13,"len")=4
	Set gtmtypes("jnl_buffer",13,"type")="unsigned-int"
	Set gtmtypfldindx("jnl_buffer","freeaddr")=13
	Set gtmtypes("jnl_buffer",14,"name")="jnl_buffer.rsrv_freeaddr"
	Set gtmtypes("jnl_buffer",14,"off")=196
	Set gtmtypes("jnl_buffer",14,"len")=4
	Set gtmtypes("jnl_buffer",14,"type")="unsigned-int"
	Set gtmtypfldindx("jnl_buffer","rsrv_freeaddr")=14
	Set gtmtypes("jnl_buffer",15,"name")="jnl_buffer.next_align_addr"
	Set gtmtypes("jnl_buffer",15,"off")=200
	Set gtmtypes("jnl_buffer",15,"len")=4
	Set gtmtypes("jnl_buffer",15,"type")="unsigned-int"
	Set gtmtypfldindx("jnl_buffer","next_align_addr")=15
	Set gtmtypes("jnl_buffer",16,"name")="jnl_buffer.end_of_data"
	Set gtmtypes("jnl_buffer",16,"off")=204
	Set gtmtypes("jnl_buffer",16,"len")=4
	Set gtmtypes("jnl_buffer",16,"type")="unsigned-int"
	Set gtmtypfldindx("jnl_buffer","end_of_data")=16
	Set gtmtypes("jnl_buffer",17,"name")="jnl_buffer.filesize"
	Set gtmtypes("jnl_buffer",17,"off")=208
	Set gtmtypes("jnl_buffer",17,"len")=4
	Set gtmtypes("jnl_buffer",17,"type")="unsigned-int"
	Set gtmtypfldindx("jnl_buffer","filesize")=17
	Set gtmtypes("jnl_buffer",18,"name")="jnl_buffer.phase2_commit_index1"
	Set gtmtypes("jnl_buffer",18,"off")=212
	Set gtmtypes("jnl_buffer",18,"len")=4
	Set gtmtypes("jnl_buffer",18,"type")="unsigned-int"
	Set gtmtypfldindx("jnl_buffer","phase2_commit_index1")=18
	Set gtmtypes("jnl_buffer",19,"name")="jnl_buffer.phase2_commit_index2"
	Set gtmtypes("jnl_buffer",19,"off")=216
	Set gtmtypes("jnl_buffer",19,"len")=4
	Set gtmtypes("jnl_buffer",19,"type")="unsigned-int"
	Set gtmtypfldindx("jnl_buffer","phase2_commit_index2")=19
	Set gtmtypes("jnl_buffer",20,"name")="jnl_buffer.phase2_commit_array"
	Set gtmtypes("jnl_buffer",20,"off")=224
	Set gtmtypes("jnl_buffer",20,"len")=196608
	Set gtmtypes("jnl_buffer",20,"type")="jbuf_phase2_in_prog_t"
	Set gtmtypfldindx("jnl_buffer","phase2_commit_array")=20
	Set gtmtypes("jnl_buffer",20,"dim")=4096
	Set gtmtypes("jnl_buffer",21,"name")="jnl_buffer.phase2_commit_array[0].curr_tn"
	Set gtmtypes("jnl_buffer",21,"off")=224
	Set gtmtypes("jnl_buffer",21,"len")=8
	Set gtmtypes("jnl_buffer",21,"type")="uint64_t"
	Set gtmtypfldindx("jnl_buffer","phase2_commit_array[0].curr_tn")=21
	Set gtmtypes("jnl_buffer",22,"name")="jnl_buffer.phase2_commit_array[0].jnl_seqno"
	Set gtmtypes("jnl_buffer",22,"off")=232
	Set gtmtypes("jnl_buffer",22,"len")=8
	Set gtmtypes("jnl_buffer",22,"type")="uint64_t"
	Set gtmtypfldindx("jnl_buffer","phase2_commit_array[0].jnl_seqno")=22
	Set gtmtypes("jnl_buffer",23,"name")="jnl_buffer.phase2_commit_array[0].strm_seqno"
	Set gtmtypes("jnl_buffer",23,"off")=240
	Set gtmtypes("jnl_buffer",23,"len")=8
	Set gtmtypes("jnl_buffer",23,"type")="uint64_t"
	Set gtmtypfldindx("jnl_buffer","phase2_commit_array[0].strm_seqno")=23
	Set gtmtypes("jnl_buffer",24,"name")="jnl_buffer.phase2_commit_array[0].process_id"
	Set gtmtypes("jnl_buffer",24,"off")=248
	Set gtmtypes("jnl_buffer",24,"len")=4
	Set gtmtypes("jnl_buffer",24,"type")="unsigned-int"
	Set gtmtypfldindx("jnl_buffer","phase2_commit_array[0].process_id")=24
	Set gtmtypes("jnl_buffer",25,"name")="jnl_buffer.phase2_commit_array[0].start_freeaddr"
	Set gtmtypes("jnl_buffer",25,"off")=252
	Set gtmtypes("jnl_buffer",25,"len")=4
	Set gtmtypes("jnl_buffer",25,"type")="unsigned-int"
	Set gtmtypfldindx("jnl_buffer","phase2_commit_array[0].start_freeaddr")=25
	Set gtmtypes("jnl_buffer",26,"name")="jnl_buffer.phase2_commit_array[0].tot_jrec_len"
	Set gtmtypes("jnl_buffer",26,"off")=256
	Set gtmtypes("jnl_buffer",26,"len")=4
	Set gtmtypes("jnl_buffer",26,"type")="unsigned-int"
	Set gtmtypfldindx("jnl_buffer","phase2_commit_array[0].tot_jrec_len")=26
	Set gtmtypes("jnl_buffer",27,"name")="jnl_buffer.phase2_commit_array[0].pini_addr"
	Set gtmtypes("jnl_buffer",27,"off")=260
	Set gtmtypes("jnl_buffer",27,"len")=4
	Set gtmtypes("jnl_buffer",27,"type")="unsigned-int"
	Set gtmtypfldindx("jnl_buffer","phase2_commit_array[0].pini_addr")=27
	Set gtmtypes("jnl_buffer",28,"name")="jnl_buffer.phase2_commit_array[0].jrec_time"
	Set gtmtypes("jnl_buffer",28,"off")=264
	Set gtmtypes("jnl_buffer",28,"len")=4
	Set gtmtypes("jnl_buffer",28,"type")="unsigned-int"
	Set gtmtypfldindx("jnl_buffer","phase2_commit_array[0].jrec_time")=28
	Set gtmtypes("jnl_buffer",29,"name")="jnl_buffer.phase2_commit_array[0].in_phase2"
	Set gtmtypes("jnl_buffer",29,"off")=268
	Set gtmtypes("jnl_buffer",29,"len")=1
	Set gtmtypes("jnl_buffer",29,"type")="char"
	Set gtmtypfldindx("jnl_buffer","phase2_commit_array[0].in_phase2")=29
	Set gtmtypes("jnl_buffer",30,"name")="jnl_buffer.phase2_commit_array[0].replication"
	Set gtmtypes("jnl_buffer",30,"off")=269
	Set gtmtypes("jnl_buffer",30,"len")=1
	Set gtmtypes("jnl_buffer",30,"type")="char"
	Set gtmtypfldindx("jnl_buffer","phase2_commit_array[0].replication")=30
	Set gtmtypes("jnl_buffer",31,"name")="jnl_buffer.phase2_commit_array[0].write_complete"
	Set gtmtypes("jnl_buffer",31,"off")=270
	Set gtmtypes("jnl_buffer",31,"len")=1
	Set gtmtypes("jnl_buffer",31,"type")="char"
	Set gtmtypfldindx("jnl_buffer","phase2_commit_array[0].write_complete")=31
	Set gtmtypes("jnl_buffer",32,"name")="jnl_buffer.phase2_commit_array[0].filler_8byte_align"
	Set gtmtypes("jnl_buffer",32,"off")=271
	Set gtmtypes("jnl_buffer",32,"len")=1
	Set gtmtypes("jnl_buffer",32,"type")="char"
	Set gtmtypfldindx("jnl_buffer","phase2_commit_array[0].filler_8byte_align")=32
	Set gtmtypes("jnl_buffer",33,"name")="jnl_buffer.blocked"
	Set gtmtypes("jnl_buffer",33,"off")=196832
	Set gtmtypes("jnl_buffer",33,"len")=4
	Set gtmtypes("jnl_buffer",33,"type")="int"
	Set gtmtypfldindx("jnl_buffer","blocked")=33
	Set gtmtypes("jnl_buffer",34,"name")="jnl_buffer.fsync_dskaddr"
	Set gtmtypes("jnl_buffer",34,"off")=196836
	Set gtmtypes("jnl_buffer",34,"len")=4
	Set gtmtypes("jnl_buffer",34,"type")="unsigned-int"
	Set gtmtypfldindx("jnl_buffer","fsync_dskaddr")=34
	Set gtmtypes("jnl_buffer",35,"name")="jnl_buffer.dsk"
	Set gtmtypes("jnl_buffer",35,"off")=196840
	Set gtmtypes("jnl_buffer",35,"len")=4
	Set gtmtypes("jnl_buffer",35,"type")="int"
	Set gtmtypfldindx("jnl_buffer","dsk")=35
	Set gtmtypes("jnl_buffer",36,"name")="jnl_buffer.wrtsize"
	Set gtmtypes("jnl_buffer",36,"off")=196844
	Set gtmtypes("jnl_buffer",36,"len")=4
	Set gtmtypes("jnl_buffer",36,"type")="int"
	Set gtmtypfldindx("jnl_buffer","wrtsize")=36
	Set gtmtypes("jnl_buffer",37,"name")="jnl_buffer.dskaddr"
	Set gtmtypes("jnl_buffer",37,"off")=196848
	Set gtmtypes("jnl_buffer",37,"len")=4
	Set gtmtypes("jnl_buffer",37,"type")="unsigned-int"
	Set gtmtypfldindx("jnl_buffer","dskaddr")=37
	Set gtmtypes("jnl_buffer",38,"name")="jnl_buffer.now_writer"
	Set gtmtypes("jnl_buffer",38,"off")=196852
	Set gtmtypes("jnl_buffer",38,"len")=4
	Set gtmtypes("jnl_buffer",38,"type")="unsigned-int"
	Set gtmtypfldindx("jnl_buffer","now_writer")=38
	Set gtmtypes("jnl_buffer",39,"name")="jnl_buffer.image_count"
	Set gtmtypes("jnl_buffer",39,"off")=196856
	Set gtmtypes("jnl_buffer",39,"len")=4
	Set gtmtypes("jnl_buffer",39,"type")="unsigned-int"
	Set gtmtypfldindx("jnl_buffer","image_count")=39
	Set gtmtypes("jnl_buffer",40,"name")="jnl_buffer.iosb"
	Set gtmtypes("jnl_buffer",40,"off")=196860
	Set gtmtypes("jnl_buffer",40,"len")=8
	Set gtmtypes("jnl_buffer",40,"type")="struct"
	Set gtmtypfldindx("jnl_buffer","iosb")=40
	Set gtmtypes("jnl_buffer",41,"name")="jnl_buffer.iosb.cond"
	Set gtmtypes("jnl_buffer",41,"off")=196860
	Set gtmtypes("jnl_buffer",41,"len")=2
	Set gtmtypes("jnl_buffer",41,"type")="short"
	Set gtmtypfldindx("jnl_buffer","iosb.cond")=41
	Set gtmtypes("jnl_buffer",42,"name")="jnl_buffer.iosb.length"
	Set gtmtypes("jnl_buffer",42,"off")=196862
	Set gtmtypes("jnl_buffer",42,"len")=2
	Set gtmtypes("jnl_buffer",42,"type")="unsigned-short"
	Set gtmtypfldindx("jnl_buffer","iosb.length")=42
	Set gtmtypes("jnl_buffer",43,"name")="jnl_buffer.iosb.dev_specific"
	Set gtmtypes("jnl_buffer",43,"off")=196864
	Set gtmtypes("jnl_buffer",43,"len")=4
	Set gtmtypes("jnl_buffer",43,"type")="int"
	Set gtmtypfldindx("jnl_buffer","iosb.dev_specific")=43
	Set gtmtypes("jnl_buffer",44,"name")="jnl_buffer.alignsize"
	Set gtmtypes("jnl_buffer",44,"off")=196868
	Set gtmtypes("jnl_buffer",44,"len")=4
	Set gtmtypes("jnl_buffer",44,"type")="unsigned-int"
	Set gtmtypfldindx("jnl_buffer","alignsize")=44
	Set gtmtypes("jnl_buffer",45,"name")="jnl_buffer.eov_timestamp"
	Set gtmtypes("jnl_buffer",45,"off")=196872
	Set gtmtypes("jnl_buffer",45,"len")=4
	Set gtmtypes("jnl_buffer",45,"type")="unsigned-int"
	Set gtmtypfldindx("jnl_buffer","eov_timestamp")=45
	Set gtmtypes("jnl_buffer",46,"name")="jnl_buffer.cycle"
	Set gtmtypes("jnl_buffer",46,"off")=196876
	Set gtmtypes("jnl_buffer",46,"len")=4
	Set gtmtypes("jnl_buffer",46,"type")="unsigned-int"
	Set gtmtypfldindx("jnl_buffer","cycle")=46
	Set gtmtypes("jnl_buffer",47,"name")="jnl_buffer.qiocnt"
	Set gtmtypes("jnl_buffer",47,"off")=196880
	Set gtmtypes("jnl_buffer",47,"len")=4
	Set gtmtypes("jnl_buffer",47,"type")="int"
	Set gtmtypfldindx("jnl_buffer","qiocnt")=47
	Set gtmtypes("jnl_buffer",48,"name")="jnl_buffer.bytcnt"
	Set gtmtypes("jnl_buffer",48,"off")=196884
	Set gtmtypes("jnl_buffer",48,"len")=4
	Set gtmtypes("jnl_buffer",48,"type")="int"
	Set gtmtypfldindx("jnl_buffer","bytcnt")=48
	Set gtmtypes("jnl_buffer",49,"name")="jnl_buffer.errcnt"
	Set gtmtypes("jnl_buffer",49,"off")=196888
	Set gtmtypes("jnl_buffer",49,"len")=4
	Set gtmtypes("jnl_buffer",49,"type")="int"
	Set gtmtypfldindx("jnl_buffer","errcnt")=49
	Set gtmtypes("jnl_buffer",50,"name")="jnl_buffer.reccnt"
	Set gtmtypes("jnl_buffer",50,"off")=196892
	Set gtmtypes("jnl_buffer",50,"len")=144
	Set gtmtypes("jnl_buffer",50,"type")="int"
	Set gtmtypfldindx("jnl_buffer","reccnt")=50
	Set gtmtypes("jnl_buffer",50,"dim")=36
	Set gtmtypes("jnl_buffer",51,"name")="jnl_buffer.filler_align"
	Set gtmtypes("jnl_buffer",51,"off")=197036
	Set gtmtypes("jnl_buffer",51,"len")=24
	Set gtmtypes("jnl_buffer",51,"type")="int"
	Set gtmtypfldindx("jnl_buffer","filler_align")=51
	Set gtmtypes("jnl_buffer",51,"dim")=6
	Set gtmtypes("jnl_buffer",52,"name")="jnl_buffer.prev_jrec_time"
	Set gtmtypes("jnl_buffer",52,"off")=197060
	Set gtmtypes("jnl_buffer",52,"len")=4
	Set gtmtypes("jnl_buffer",52,"type")="unsigned-int"
	Set gtmtypfldindx("jnl_buffer","prev_jrec_time")=52
	Set gtmtypes("jnl_buffer",53,"name")="jnl_buffer.next_epoch_time"
	Set gtmtypes("jnl_buffer",53,"off")=197064
	Set gtmtypes("jnl_buffer",53,"len")=4
	Set gtmtypes("jnl_buffer",53,"type")="unsigned-int"
	Set gtmtypfldindx("jnl_buffer","next_epoch_time")=53
	Set gtmtypes("jnl_buffer",54,"name")="jnl_buffer.need_db_fsync"
	Set gtmtypes("jnl_buffer",54,"off")=197068
	Set gtmtypes("jnl_buffer",54,"len")=4
	Set gtmtypes("jnl_buffer",54,"type")="boolean_t"
	Set gtmtypfldindx("jnl_buffer","need_db_fsync")=54
	Set gtmtypes("jnl_buffer",55,"name")="jnl_buffer.io_in_prog"
	Set gtmtypes("jnl_buffer",55,"off")=197072
	Set gtmtypes("jnl_buffer",55,"len")=4
	Set gtmtypes("jnl_buffer",55,"type")="int"
	Set gtmtypfldindx("jnl_buffer","io_in_prog")=55
	Set gtmtypes("jnl_buffer",56,"name")="jnl_buffer.enospc_errcnt"
	Set gtmtypes("jnl_buffer",56,"off")=197076
	Set gtmtypes("jnl_buffer",56,"len")=4
	Set gtmtypes("jnl_buffer",56,"type")="unsigned-int"
	Set gtmtypfldindx("jnl_buffer","enospc_errcnt")=56
	Set gtmtypes("jnl_buffer",57,"name")="jnl_buffer.max_jrec_len"
	Set gtmtypes("jnl_buffer",57,"off")=197080
	Set gtmtypes("jnl_buffer",57,"len")=4
	Set gtmtypes("jnl_buffer",57,"type")="unsigned-int"
	Set gtmtypfldindx("jnl_buffer","max_jrec_len")=57
	Set gtmtypes("jnl_buffer",58,"name")="jnl_buffer.fs_block_size"
	Set gtmtypes("jnl_buffer",58,"off")=197084
	Set gtmtypes("jnl_buffer",58,"len")=4
	Set gtmtypes("jnl_buffer",58,"type")="unsigned-int"
	Set gtmtypfldindx("jnl_buffer","fs_block_size")=58
	Set gtmtypes("jnl_buffer",59,"name")="jnl_buffer.post_epoch_freeaddr"
	Set gtmtypes("jnl_buffer",59,"off")=197088
	Set gtmtypes("jnl_buffer",59,"len")=4
	Set gtmtypes("jnl_buffer",59,"type")="unsigned-int"
	Set gtmtypfldindx("jnl_buffer","post_epoch_freeaddr")=59
	Set gtmtypes("jnl_buffer",60,"name")="jnl_buffer.last_eof_written"
	Set gtmtypes("jnl_buffer",60,"off")=197092
	Set gtmtypes("jnl_buffer",60,"len")=4
	Set gtmtypes("jnl_buffer",60,"type")="boolean_t"
	Set gtmtypfldindx("jnl_buffer","last_eof_written")=60
	Set gtmtypes("jnl_buffer",61,"name")="jnl_buffer.end_of_data_at_open"
	Set gtmtypes("jnl_buffer",61,"off")=197096
	Set gtmtypes("jnl_buffer",61,"len")=4
	Set gtmtypes("jnl_buffer",61,"type")="unsigned-int"
	Set gtmtypfldindx("jnl_buffer","end_of_data_at_open")=61
	Set gtmtypes("jnl_buffer",62,"name")="jnl_buffer.re_read_dskaddr"
	Set gtmtypes("jnl_buffer",62,"off")=197100
	Set gtmtypes("jnl_buffer",62,"len")=4
	Set gtmtypes("jnl_buffer",62,"type")="unsigned-int"
	Set gtmtypfldindx("jnl_buffer","re_read_dskaddr")=62
	Set gtmtypes("jnl_buffer",63,"name")="jnl_buffer.fill_cacheline0"
	Set gtmtypes("jnl_buffer",63,"off")=197104
	Set gtmtypes("jnl_buffer",63,"len")=56
	Set gtmtypes("jnl_buffer",63,"type")="char"
	Set gtmtypfldindx("jnl_buffer","fill_cacheline0")=63
	Set gtmtypes("jnl_buffer",64,"name")="jnl_buffer.io_in_prog_latch"
	Set gtmtypes("jnl_buffer",64,"off")=197160
	Set gtmtypes("jnl_buffer",64,"len")=8
	Set gtmtypes("jnl_buffer",64,"type")="global_latch_t"
	Set gtmtypfldindx("jnl_buffer","io_in_prog_latch")=64
	Set gtmtypes("jnl_buffer",65,"name")="jnl_buffer.io_in_prog_latch.u"
	Set gtmtypes("jnl_buffer",65,"off")=197160
	Set gtmtypes("jnl_buffer",65,"len")=8
	Set gtmtypes("jnl_buffer",65,"type")="union"
	Set gtmtypfldindx("jnl_buffer","io_in_prog_latch.u")=65
	Set gtmtypes("jnl_buffer",66,"name")="jnl_buffer.io_in_prog_latch.u.pid_imgcnt"
	Set gtmtypes("jnl_buffer",66,"off")=197160
	Set gtmtypes("jnl_buffer",66,"len")=8
	Set gtmtypes("jnl_buffer",66,"type")="uint64_t"
	Set gtmtypfldindx("jnl_buffer","io_in_prog_latch.u.pid_imgcnt")=66
	Set gtmtypes("jnl_buffer",67,"name")="jnl_buffer.io_in_prog_latch.u.parts"
	Set gtmtypes("jnl_buffer",67,"off")=197160
	Set gtmtypes("jnl_buffer",67,"len")=8
	Set gtmtypes("jnl_buffer",67,"type")="struct"
	Set gtmtypfldindx("jnl_buffer","io_in_prog_latch.u.parts")=67
	Set gtmtypes("jnl_buffer",68,"name")="jnl_buffer.io_in_prog_latch.u.parts.latch_pid"
	Set gtmtypes("jnl_buffer",68,"off")=197160
	Set gtmtypes("jnl_buffer",68,"len")=4
	Set gtmtypes("jnl_buffer",68,"type")="int"
	Set gtmtypfldindx("jnl_buffer","io_in_prog_latch.u.parts.latch_pid")=68
	Set gtmtypes("jnl_buffer",69,"name")="jnl_buffer.io_in_prog_latch.u.parts.latch_word"
	Set gtmtypes("jnl_buffer",69,"off")=197164
	Set gtmtypes("jnl_buffer",69,"len")=4
	Set gtmtypes("jnl_buffer",69,"type")="int"
	Set gtmtypfldindx("jnl_buffer","io_in_prog_latch.u.parts.latch_word")=69
	Set gtmtypes("jnl_buffer",70,"name")="jnl_buffer.fill_cacheline1"
	Set gtmtypes("jnl_buffer",70,"off")=197168
	Set gtmtypes("jnl_buffer",70,"len")=56
	Set gtmtypes("jnl_buffer",70,"type")="char"
	Set gtmtypfldindx("jnl_buffer","fill_cacheline1")=70
	Set gtmtypes("jnl_buffer",71,"name")="jnl_buffer.fsync_in_prog_latch"
	Set gtmtypes("jnl_buffer",71,"off")=197224
	Set gtmtypes("jnl_buffer",71,"len")=8
	Set gtmtypes("jnl_buffer",71,"type")="global_latch_t"
	Set gtmtypfldindx("jnl_buffer","fsync_in_prog_latch")=71
	Set gtmtypes("jnl_buffer",72,"name")="jnl_buffer.fsync_in_prog_latch.u"
	Set gtmtypes("jnl_buffer",72,"off")=197224
	Set gtmtypes("jnl_buffer",72,"len")=8
	Set gtmtypes("jnl_buffer",72,"type")="union"
	Set gtmtypfldindx("jnl_buffer","fsync_in_prog_latch.u")=72
	Set gtmtypes("jnl_buffer",73,"name")="jnl_buffer.fsync_in_prog_latch.u.pid_imgcnt"
	Set gtmtypes("jnl_buffer",73,"off")=197224
	Set gtmtypes("jnl_buffer",73,"len")=8
	Set gtmtypes("jnl_buffer",73,"type")="uint64_t"
	Set gtmtypfldindx("jnl_buffer","fsync_in_prog_latch.u.pid_imgcnt")=73
	Set gtmtypes("jnl_buffer",74,"name")="jnl_buffer.fsync_in_prog_latch.u.parts"
	Set gtmtypes("jnl_buffer",74,"off")=197224
	Set gtmtypes("jnl_buffer",74,"len")=8
	Set gtmtypes("jnl_buffer",74,"type")="struct"
	Set gtmtypfldindx("jnl_buffer","fsync_in_prog_latch.u.parts")=74
	Set gtmtypes("jnl_buffer",75,"name")="jnl_buffer.fsync_in_prog_latch.u.parts.latch_pid"
	Set gtmtypes("jnl_buffer",75,"off")=197224
	Set gtmtypes("jnl_buffer",75,"len")=4
	Set gtmtypes("jnl_buffer",75,"type")="int"
	Set gtmtypfldindx("jnl_buffer","fsync_in_prog_latch.u.parts.latch_pid")=75
	Set gtmtypes("jnl_buffer",76,"name")="jnl_buffer.fsync_in_prog_latch.u.parts.latch_word"
	Set gtmtypes("jnl_buffer",76,"off")=197228
	Set gtmtypes("jnl_buffer",76,"len")=4
	Set gtmtypes("jnl_buffer",76,"type")="int"
	Set gtmtypfldindx("jnl_buffer","fsync_in_prog_latch.u.parts.latch_word")=76
	Set gtmtypes("jnl_buffer",77,"name")="jnl_buffer.fill_cacheline2"
	Set gtmtypes("jnl_buffer",77,"off")=197232
	Set gtmtypes("jnl_buffer",77,"len")=56
	Set gtmtypes("jnl_buffer",77,"type")="char"
	Set gtmtypfldindx("jnl_buffer","fill_cacheline2")=77
	Set gtmtypes("jnl_buffer",78,"name")="jnl_buffer.phase2_commit_latch"
	Set gtmtypes("jnl_buffer",78,"off")=197288
	Set gtmtypes("jnl_buffer",78,"len")=8
	Set gtmtypes("jnl_buffer",78,"type")="global_latch_t"
	Set gtmtypfldindx("jnl_buffer","phase2_commit_latch")=78
	Set gtmtypes("jnl_buffer",79,"name")="jnl_buffer.phase2_commit_latch.u"
	Set gtmtypes("jnl_buffer",79,"off")=197288
	Set gtmtypes("jnl_buffer",79,"len")=8
	Set gtmtypes("jnl_buffer",79,"type")="union"
	Set gtmtypfldindx("jnl_buffer","phase2_commit_latch.u")=79
	Set gtmtypes("jnl_buffer",80,"name")="jnl_buffer.phase2_commit_latch.u.pid_imgcnt"
	Set gtmtypes("jnl_buffer",80,"off")=197288
	Set gtmtypes("jnl_buffer",80,"len")=8
	Set gtmtypes("jnl_buffer",80,"type")="uint64_t"
	Set gtmtypfldindx("jnl_buffer","phase2_commit_latch.u.pid_imgcnt")=80
	Set gtmtypes("jnl_buffer",81,"name")="jnl_buffer.phase2_commit_latch.u.parts"
	Set gtmtypes("jnl_buffer",81,"off")=197288
	Set gtmtypes("jnl_buffer",81,"len")=8
	Set gtmtypes("jnl_buffer",81,"type")="struct"
	Set gtmtypfldindx("jnl_buffer","phase2_commit_latch.u.parts")=81
	Set gtmtypes("jnl_buffer",82,"name")="jnl_buffer.phase2_commit_latch.u.parts.latch_pid"
	Set gtmtypes("jnl_buffer",82,"off")=197288
	Set gtmtypes("jnl_buffer",82,"len")=4
	Set gtmtypes("jnl_buffer",82,"type")="int"
	Set gtmtypfldindx("jnl_buffer","phase2_commit_latch.u.parts.latch_pid")=82
	Set gtmtypes("jnl_buffer",83,"name")="jnl_buffer.phase2_commit_latch.u.parts.latch_word"
	Set gtmtypes("jnl_buffer",83,"off")=197292
	Set gtmtypes("jnl_buffer",83,"len")=4
	Set gtmtypes("jnl_buffer",83,"type")="int"
	Set gtmtypfldindx("jnl_buffer","phase2_commit_latch.u.parts.latch_word")=83
	Set gtmtypes("jnl_buffer",84,"name")="jnl_buffer.fill_cacheline3"
	Set gtmtypes("jnl_buffer",84,"off")=197296
	Set gtmtypes("jnl_buffer",84,"len")=56
	Set gtmtypes("jnl_buffer",84,"type")="char"
	Set gtmtypfldindx("jnl_buffer","fill_cacheline3")=84
	Set gtmtypes("jnl_buffer",85,"name")="jnl_buffer.buff"
	Set gtmtypes("jnl_buffer",85,"off")=197352
	Set gtmtypes("jnl_buffer",85,"len")=1
	Set gtmtypes("jnl_buffer",85,"type")="unsigned-char"
	Set gtmtypfldindx("jnl_buffer","buff")=85
	;
	Set gtmtypes("jnl_create_info")="struct"
	Set gtmtypes("jnl_create_info",0)=35
	Set gtmtypes("jnl_create_info","len")=560
	Set gtmtypes("jnl_create_info",1,"name")="jnl_create_info.status"
	Set gtmtypes("jnl_create_info",1,"off")=0
	Set gtmtypes("jnl_create_info",1,"len")=4
	Set gtmtypes("jnl_create_info",1,"type")="int"
	Set gtmtypfldindx("jnl_create_info","status")=1
	Set gtmtypes("jnl_create_info",2,"name")="jnl_create_info.alloc"
	Set gtmtypes("jnl_create_info",2,"off")=4
	Set gtmtypes("jnl_create_info",2,"len")=4
	Set gtmtypes("jnl_create_info",2,"type")="int"
	Set gtmtypfldindx("jnl_create_info","alloc")=2
	Set gtmtypes("jnl_create_info",3,"name")="jnl_create_info.extend"
	Set gtmtypes("jnl_create_info",3,"off")=8
	Set gtmtypes("jnl_create_info",3,"len")=4
	Set gtmtypes("jnl_create_info",3,"type")="int"
	Set gtmtypfldindx("jnl_create_info","extend")=3
	Set gtmtypes("jnl_create_info",4,"name")="jnl_create_info.buffer"
	Set gtmtypes("jnl_create_info",4,"off")=12
	Set gtmtypes("jnl_create_info",4,"len")=4
	Set gtmtypes("jnl_create_info",4,"type")="int"
	Set gtmtypfldindx("jnl_create_info","buffer")=4
	Set gtmtypes("jnl_create_info",5,"name")="jnl_create_info.csd"
	Set gtmtypes("jnl_create_info",5,"off")=16
	Set gtmtypes("jnl_create_info",5,"len")=8
	Set gtmtypes("jnl_create_info",5,"type")="addr"
	Set gtmtypfldindx("jnl_create_info","csd")=5
	Set gtmtypes("jnl_create_info",6,"name")="jnl_create_info.reg_seqno"
	Set gtmtypes("jnl_create_info",6,"off")=24
	Set gtmtypes("jnl_create_info",6,"len")=8
	Set gtmtypes("jnl_create_info",6,"type")="uint64_t"
	Set gtmtypfldindx("jnl_create_info","reg_seqno")=6
	Set gtmtypes("jnl_create_info",7,"name")="jnl_create_info.jnl"
	Set gtmtypes("jnl_create_info",7,"off")=32
	Set gtmtypes("jnl_create_info",7,"len")=256
	Set gtmtypes("jnl_create_info",7,"type")="unsigned-char"
	Set gtmtypfldindx("jnl_create_info","jnl")=7
	Set gtmtypes("jnl_create_info",8,"name")="jnl_create_info.fn"
	Set gtmtypes("jnl_create_info",8,"off")=288
	Set gtmtypes("jnl_create_info",8,"len")=8
	Set gtmtypes("jnl_create_info",8,"type")="addr"
	Set gtmtypfldindx("jnl_create_info","fn")=8
	Set gtmtypes("jnl_create_info",9,"name")="jnl_create_info.max_jrec_len"
	Set gtmtypes("jnl_create_info",9,"off")=296
	Set gtmtypes("jnl_create_info",9,"len")=4
	Set gtmtypes("jnl_create_info",9,"type")="unsigned-int"
	Set gtmtypfldindx("jnl_create_info","max_jrec_len")=9
	Set gtmtypes("jnl_create_info",10,"name")="jnl_create_info.fn_len"
	Set gtmtypes("jnl_create_info",10,"off")=300
	Set gtmtypes("jnl_create_info",10,"len")=2
	Set gtmtypes("jnl_create_info",10,"type")="short"
	Set gtmtypfldindx("jnl_create_info","fn_len")=10
	Set gtmtypes("jnl_create_info",11,"name")="jnl_create_info.jnl_len"
	Set gtmtypes("jnl_create_info",11,"off")=302
	Set gtmtypes("jnl_create_info",11,"len")=2
	Set gtmtypes("jnl_create_info",11,"type")="short"
	Set gtmtypfldindx("jnl_create_info","jnl_len")=11
	Set gtmtypes("jnl_create_info",12,"name")="jnl_create_info.jnl_def_len"
	Set gtmtypes("jnl_create_info",12,"off")=304
	Set gtmtypes("jnl_create_info",12,"len")=2
	Set gtmtypes("jnl_create_info",12,"type")="short"
	Set gtmtypfldindx("jnl_create_info","jnl_def_len")=12
	Set gtmtypes("jnl_create_info",13,"name")="jnl_create_info.before_images"
	Set gtmtypes("jnl_create_info",13,"off")=306
	Set gtmtypes("jnl_create_info",13,"len")=1
	Set gtmtypes("jnl_create_info",13,"type")="char"
	Set gtmtypfldindx("jnl_create_info","before_images")=13
	Set gtmtypes("jnl_create_info",14,"name")="jnl_create_info.filler_bool"
	Set gtmtypes("jnl_create_info",14,"off")=307
	Set gtmtypes("jnl_create_info",14,"len")=1
	Set gtmtypes("jnl_create_info",14,"type")="char"
	Set gtmtypfldindx("jnl_create_info","filler_bool")=14
	Set gtmtypes("jnl_create_info",15,"name")="jnl_create_info.alignsize"
	Set gtmtypes("jnl_create_info",15,"off")=308
	Set gtmtypes("jnl_create_info",15,"len")=4
	Set gtmtypes("jnl_create_info",15,"type")="unsigned-int"
	Set gtmtypfldindx("jnl_create_info","alignsize")=15
	Set gtmtypes("jnl_create_info",16,"name")="jnl_create_info.autoswitchlimit"
	Set gtmtypes("jnl_create_info",16,"off")=312
	Set gtmtypes("jnl_create_info",16,"len")=4
	Set gtmtypes("jnl_create_info",16,"type")="int"
	Set gtmtypfldindx("jnl_create_info","autoswitchlimit")=16
	Set gtmtypes("jnl_create_info",17,"name")="jnl_create_info.epoch_interval"
	Set gtmtypes("jnl_create_info",17,"off")=316
	Set gtmtypes("jnl_create_info",17,"len")=4
	Set gtmtypes("jnl_create_info",17,"type")="int"
	Set gtmtypfldindx("jnl_create_info","epoch_interval")=17
	Set gtmtypes("jnl_create_info",18,"name")="jnl_create_info.prev_jnl"
	Set gtmtypes("jnl_create_info",18,"off")=320
	Set gtmtypes("jnl_create_info",18,"len")=8
	Set gtmtypes("jnl_create_info",18,"type")="addr"
	Set gtmtypfldindx("jnl_create_info","prev_jnl")=18
	Set gtmtypes("jnl_create_info",19,"name")="jnl_create_info.prev_jnl_len"
	Set gtmtypes("jnl_create_info",19,"off")=328
	Set gtmtypes("jnl_create_info",19,"len")=4
	Set gtmtypes("jnl_create_info",19,"type")="int"
	Set gtmtypfldindx("jnl_create_info","prev_jnl_len")=19
	Set gtmtypes("jnl_create_info",20,"name")="jnl_create_info.jnl_state"
	Set gtmtypes("jnl_create_info",20,"off")=332
	Set gtmtypes("jnl_create_info",20,"len")=4
	Set gtmtypes("jnl_create_info",20,"type")="int"
	Set gtmtypfldindx("jnl_create_info","jnl_state")=20
	Set gtmtypes("jnl_create_info",21,"name")="jnl_create_info.repl_state"
	Set gtmtypes("jnl_create_info",21,"off")=336
	Set gtmtypes("jnl_create_info",21,"len")=4
	Set gtmtypes("jnl_create_info",21,"type")="int"
	Set gtmtypfldindx("jnl_create_info","repl_state")=21
	Set gtmtypes("jnl_create_info",22,"name")="jnl_create_info.status2"
	Set gtmtypes("jnl_create_info",22,"off")=340
	Set gtmtypes("jnl_create_info",22,"len")=4
	Set gtmtypes("jnl_create_info",22,"type")="unsigned-int"
	Set gtmtypfldindx("jnl_create_info","status2")=22
	Set gtmtypes("jnl_create_info",23,"name")="jnl_create_info.no_rename"
	Set gtmtypes("jnl_create_info",23,"off")=344
	Set gtmtypes("jnl_create_info",23,"len")=4
	Set gtmtypes("jnl_create_info",23,"type")="boolean_t"
	Set gtmtypfldindx("jnl_create_info","no_rename")=23
	Set gtmtypes("jnl_create_info",24,"name")="jnl_create_info.no_prev_link"
	Set gtmtypes("jnl_create_info",24,"off")=348
	Set gtmtypes("jnl_create_info",24,"len")=4
	Set gtmtypes("jnl_create_info",24,"type")="boolean_t"
	Set gtmtypfldindx("jnl_create_info","no_prev_link")=24
	Set gtmtypes("jnl_create_info",25,"name")="jnl_create_info.blks_to_upgrd"
	Set gtmtypes("jnl_create_info",25,"off")=352
	Set gtmtypes("jnl_create_info",25,"len")=4
	Set gtmtypes("jnl_create_info",25,"type")="int"
	Set gtmtypfldindx("jnl_create_info","blks_to_upgrd")=25
	Set gtmtypes("jnl_create_info",26,"name")="jnl_create_info.checksum"
	Set gtmtypes("jnl_create_info",26,"off")=356
	Set gtmtypes("jnl_create_info",26,"len")=4
	Set gtmtypes("jnl_create_info",26,"type")="unsigned-int"
	Set gtmtypfldindx("jnl_create_info","checksum")=26
	Set gtmtypes("jnl_create_info",27,"name")="jnl_create_info.free_blocks"
	Set gtmtypes("jnl_create_info",27,"off")=360
	Set gtmtypes("jnl_create_info",27,"len")=4
	Set gtmtypes("jnl_create_info",27,"type")="unsigned-int"
	Set gtmtypfldindx("jnl_create_info","free_blocks")=27
	Set gtmtypes("jnl_create_info",28,"name")="jnl_create_info.total_blks"
	Set gtmtypes("jnl_create_info",28,"off")=364
	Set gtmtypes("jnl_create_info",28,"len")=4
	Set gtmtypes("jnl_create_info",28,"type")="unsigned-int"
	Set gtmtypfldindx("jnl_create_info","total_blks")=28
	Set gtmtypes("jnl_create_info",29,"name")="jnl_create_info.is_encrypted"
	Set gtmtypes("jnl_create_info",29,"off")=368
	Set gtmtypes("jnl_create_info",29,"len")=4
	Set gtmtypes("jnl_create_info",29,"type")="int"
	Set gtmtypfldindx("jnl_create_info","is_encrypted")=29
	Set gtmtypes("jnl_create_info",30,"name")="jnl_create_info.encryption_hash"
	Set gtmtypes("jnl_create_info",30,"off")=372
	Set gtmtypes("jnl_create_info",30,"len")=80
	Set gtmtypes("jnl_create_info",30,"type")="char"
	Set gtmtypfldindx("jnl_create_info","encryption_hash")=30
	Set gtmtypes("jnl_create_info",31,"name")="jnl_create_info.encryption_hash2"
	Set gtmtypes("jnl_create_info",31,"off")=452
	Set gtmtypes("jnl_create_info",31,"len")=80
	Set gtmtypes("jnl_create_info",31,"type")="char"
	Set gtmtypfldindx("jnl_create_info","encryption_hash2")=31
	Set gtmtypes("jnl_create_info",32,"name")="jnl_create_info.non_null_iv"
	Set gtmtypes("jnl_create_info",32,"off")=532
	Set gtmtypes("jnl_create_info",32,"len")=4
	Set gtmtypes("jnl_create_info",32,"type")="boolean_t"
	Set gtmtypfldindx("jnl_create_info","non_null_iv")=32
	Set gtmtypes("jnl_create_info",33,"name")="jnl_create_info.encryption_hash_cutoff"
	Set gtmtypes("jnl_create_info",33,"off")=536
	Set gtmtypes("jnl_create_info",33,"len")=4
	Set gtmtypes("jnl_create_info",33,"type")="int"
	Set gtmtypfldindx("jnl_create_info","encryption_hash_cutoff")=33
	Set gtmtypes("jnl_create_info",34,"name")="jnl_create_info.encryption_hash2_start_tn"
	Set gtmtypes("jnl_create_info",34,"off")=544
	Set gtmtypes("jnl_create_info",34,"len")=8
	Set gtmtypes("jnl_create_info",34,"type")="uint64_t"
	Set gtmtypfldindx("jnl_create_info","encryption_hash2_start_tn")=34
	Set gtmtypes("jnl_create_info",35,"name")="jnl_create_info.csa"
	Set gtmtypes("jnl_create_info",35,"off")=552
	Set gtmtypes("jnl_create_info",35,"len")=8
	Set gtmtypes("jnl_create_info",35,"type")="addr"
	Set gtmtypfldindx("jnl_create_info","csa")=35
	;
	Set gtmtypes("jnl_ctl_list")="struct"
	Set gtmtypes("jnl_ctl_list",0)=47
	Set gtmtypes("jnl_ctl_list","len")=640
	Set gtmtypes("jnl_ctl_list",1,"name")="jnl_ctl_list.turn_around_tn"
	Set gtmtypes("jnl_ctl_list",1,"off")=0
	Set gtmtypes("jnl_ctl_list",1,"len")=8
	Set gtmtypes("jnl_ctl_list",1,"type")="uint64_t"
	Set gtmtypfldindx("jnl_ctl_list","turn_around_tn")=1
	Set gtmtypes("jnl_ctl_list",2,"name")="jnl_ctl_list.turn_around_seqno"
	Set gtmtypes("jnl_ctl_list",2,"off")=8
	Set gtmtypes("jnl_ctl_list",2,"len")=8
	Set gtmtypes("jnl_ctl_list",2,"type")="uint64_t"
	Set gtmtypfldindx("jnl_ctl_list","turn_around_seqno")=2
	Set gtmtypes("jnl_ctl_list",3,"name")="jnl_ctl_list.jnl_fn"
	Set gtmtypes("jnl_ctl_list",3,"off")=16
	Set gtmtypes("jnl_ctl_list",3,"len")=256
	Set gtmtypes("jnl_ctl_list",3,"type")="unsigned-char"
	Set gtmtypfldindx("jnl_ctl_list","jnl_fn")=3
	Set gtmtypes("jnl_ctl_list",4,"name")="jnl_ctl_list.jnl_fn_len"
	Set gtmtypes("jnl_ctl_list",4,"off")=272
	Set gtmtypes("jnl_ctl_list",4,"len")=4
	Set gtmtypes("jnl_ctl_list",4,"type")="unsigned-int"
	Set gtmtypfldindx("jnl_ctl_list","jnl_fn_len")=4
	Set gtmtypes("jnl_ctl_list",5,"name")="jnl_ctl_list.jfh"
	Set gtmtypes("jnl_ctl_list",5,"off")=280
	Set gtmtypes("jnl_ctl_list",5,"len")=8
	Set gtmtypes("jnl_ctl_list",5,"type")="addr"
	Set gtmtypfldindx("jnl_ctl_list","jfh")=5
	Set gtmtypes("jnl_ctl_list",6,"name")="jnl_ctl_list.lvrec_time"
	Set gtmtypes("jnl_ctl_list",6,"off")=288
	Set gtmtypes("jnl_ctl_list",6,"len")=4
	Set gtmtypes("jnl_ctl_list",6,"type")="unsigned-int"
	Set gtmtypfldindx("jnl_ctl_list","lvrec_time")=6
	Set gtmtypes("jnl_ctl_list",7,"name")="jnl_ctl_list.lvrec_off"
	Set gtmtypes("jnl_ctl_list",7,"off")=292
	Set gtmtypes("jnl_ctl_list",7,"len")=4
	Set gtmtypes("jnl_ctl_list",7,"type")="unsigned-int"
	Set gtmtypfldindx("jnl_ctl_list","lvrec_off")=7
	Set gtmtypes("jnl_ctl_list",8,"name")="jnl_ctl_list.rec_offset"
	Set gtmtypes("jnl_ctl_list",8,"off")=296
	Set gtmtypes("jnl_ctl_list",8,"len")=4
	Set gtmtypes("jnl_ctl_list",8,"type")="unsigned-int"
	Set gtmtypfldindx("jnl_ctl_list","rec_offset")=8
	Set gtmtypes("jnl_ctl_list",9,"name")="jnl_ctl_list.os_filesize"
	Set gtmtypes("jnl_ctl_list",9,"off")=300
	Set gtmtypes("jnl_ctl_list",9,"len")=4
	Set gtmtypes("jnl_ctl_list",9,"type")="unsigned-int"
	Set gtmtypfldindx("jnl_ctl_list","os_filesize")=9
	Set gtmtypes("jnl_ctl_list",10,"name")="jnl_ctl_list.eof_addr"
	Set gtmtypes("jnl_ctl_list",10,"off")=304
	Set gtmtypes("jnl_ctl_list",10,"len")=4
	Set gtmtypes("jnl_ctl_list",10,"type")="unsigned-int"
	Set gtmtypfldindx("jnl_ctl_list","eof_addr")=10
	Set gtmtypes("jnl_ctl_list",11,"name")="jnl_ctl_list.apply_pblk_stop_offset"
	Set gtmtypes("jnl_ctl_list",11,"off")=308
	Set gtmtypes("jnl_ctl_list",11,"len")=4
	Set gtmtypes("jnl_ctl_list",11,"type")="unsigned-int"
	Set gtmtypfldindx("jnl_ctl_list","apply_pblk_stop_offset")=11
	Set gtmtypes("jnl_ctl_list",12,"name")="jnl_ctl_list.turn_around_offset"
	Set gtmtypes("jnl_ctl_list",12,"off")=312
	Set gtmtypes("jnl_ctl_list",12,"len")=4
	Set gtmtypes("jnl_ctl_list",12,"type")="unsigned-int"
	Set gtmtypfldindx("jnl_ctl_list","turn_around_offset")=12
	Set gtmtypes("jnl_ctl_list",13,"name")="jnl_ctl_list.turn_around_time"
	Set gtmtypes("jnl_ctl_list",13,"off")=316
	Set gtmtypes("jnl_ctl_list",13,"len")=4
	Set gtmtypes("jnl_ctl_list",13,"type")="unsigned-int"
	Set gtmtypfldindx("jnl_ctl_list","turn_around_time")=13
	Set gtmtypes("jnl_ctl_list",14,"name")="jnl_ctl_list.properly_closed"
	Set gtmtypes("jnl_ctl_list",14,"off")=320
	Set gtmtypes("jnl_ctl_list",14,"len")=4
	Set gtmtypes("jnl_ctl_list",14,"type")="boolean_t"
	Set gtmtypfldindx("jnl_ctl_list","properly_closed")=14
	Set gtmtypes("jnl_ctl_list",15,"name")="jnl_ctl_list.tail_analysis"
	Set gtmtypes("jnl_ctl_list",15,"off")=324
	Set gtmtypes("jnl_ctl_list",15,"len")=4
	Set gtmtypes("jnl_ctl_list",15,"type")="boolean_t"
	Set gtmtypfldindx("jnl_ctl_list","tail_analysis")=15
	Set gtmtypes("jnl_ctl_list",16,"name")="jnl_ctl_list.after_end_of_data"
	Set gtmtypes("jnl_ctl_list",16,"off")=328
	Set gtmtypes("jnl_ctl_list",16,"len")=4
	Set gtmtypes("jnl_ctl_list",16,"type")="boolean_t"
	Set gtmtypfldindx("jnl_ctl_list","after_end_of_data")=16
	Set gtmtypes("jnl_ctl_list",17,"name")="jnl_ctl_list.read_only"
	Set gtmtypes("jnl_ctl_list",17,"off")=332
	Set gtmtypes("jnl_ctl_list",17,"len")=4
	Set gtmtypes("jnl_ctl_list",17,"type")="boolean_t"
	Set gtmtypfldindx("jnl_ctl_list","read_only")=17
	Set gtmtypes("jnl_ctl_list",18,"name")="jnl_ctl_list.jnlrec_cnt"
	Set gtmtypes("jnl_ctl_list",18,"off")=336
	Set gtmtypes("jnl_ctl_list",18,"len")=144
	Set gtmtypes("jnl_ctl_list",18,"type")="int"
	Set gtmtypfldindx("jnl_ctl_list","jnlrec_cnt")=18
	Set gtmtypes("jnl_ctl_list",18,"dim")=36
	Set gtmtypes("jnl_ctl_list",19,"name")="jnl_ctl_list.status"
	Set gtmtypes("jnl_ctl_list",19,"off")=480
	Set gtmtypes("jnl_ctl_list",19,"len")=4
	Set gtmtypes("jnl_ctl_list",19,"type")="int"
	Set gtmtypfldindx("jnl_ctl_list","status")=19
	Set gtmtypes("jnl_ctl_list",20,"name")="jnl_ctl_list.status2"
	Set gtmtypes("jnl_ctl_list",20,"off")=484
	Set gtmtypes("jnl_ctl_list",20,"len")=4
	Set gtmtypes("jnl_ctl_list",20,"type")="unsigned-int"
	Set gtmtypfldindx("jnl_ctl_list","status2")=20
	Set gtmtypes("jnl_ctl_list",21,"name")="jnl_ctl_list.channel"
	Set gtmtypes("jnl_ctl_list",21,"off")=488
	Set gtmtypes("jnl_ctl_list",21,"len")=4
	Set gtmtypes("jnl_ctl_list",21,"type")="int"
	Set gtmtypfldindx("jnl_ctl_list","channel")=21
	Set gtmtypes("jnl_ctl_list",22,"name")="jnl_ctl_list.fid"
	Set gtmtypes("jnl_ctl_list",22,"off")=496
	Set gtmtypes("jnl_ctl_list",22,"len")=16
	Set gtmtypes("jnl_ctl_list",22,"type")="unix_file_id"
	Set gtmtypfldindx("jnl_ctl_list","fid")=22
	Set gtmtypes("jnl_ctl_list",23,"name")="jnl_ctl_list.fid.inode"
	Set gtmtypes("jnl_ctl_list",23,"off")=496
	Set gtmtypes("jnl_ctl_list",23,"len")=8
	Set gtmtypes("jnl_ctl_list",23,"type")="ino_t"
	Set gtmtypfldindx("jnl_ctl_list","fid.inode")=23
	Set gtmtypes("jnl_ctl_list",24,"name")="jnl_ctl_list.fid.device"
	Set gtmtypes("jnl_ctl_list",24,"off")=504
	Set gtmtypes("jnl_ctl_list",24,"len")=8
	Set gtmtypes("jnl_ctl_list",24,"type")="dev_t"
	Set gtmtypfldindx("jnl_ctl_list","fid.device")=24
	Set gtmtypes("jnl_ctl_list",25,"name")="jnl_ctl_list.pini_list"
	Set gtmtypes("jnl_ctl_list",25,"off")=512
	Set gtmtypes("jnl_ctl_list",25,"len")=80
	Set gtmtypes("jnl_ctl_list",25,"type")="hash_table_int4"
	Set gtmtypfldindx("jnl_ctl_list","pini_list")=25
	Set gtmtypes("jnl_ctl_list",26,"name")="jnl_ctl_list.pini_list.base"
	Set gtmtypes("jnl_ctl_list",26,"off")=512
	Set gtmtypes("jnl_ctl_list",26,"len")=8
	Set gtmtypes("jnl_ctl_list",26,"type")="addr"
	Set gtmtypfldindx("jnl_ctl_list","pini_list.base")=26
	Set gtmtypes("jnl_ctl_list",27,"name")="jnl_ctl_list.pini_list.top"
	Set gtmtypes("jnl_ctl_list",27,"off")=520
	Set gtmtypes("jnl_ctl_list",27,"len")=8
	Set gtmtypes("jnl_ctl_list",27,"type")="addr"
	Set gtmtypfldindx("jnl_ctl_list","pini_list.top")=27
	Set gtmtypes("jnl_ctl_list",28,"name")="jnl_ctl_list.pini_list.spare_base"
	Set gtmtypes("jnl_ctl_list",28,"off")=528
	Set gtmtypes("jnl_ctl_list",28,"len")=8
	Set gtmtypes("jnl_ctl_list",28,"type")="addr"
	Set gtmtypfldindx("jnl_ctl_list","pini_list.spare_base")=28
	Set gtmtypes("jnl_ctl_list",29,"name")="jnl_ctl_list.pini_list.entry_passed_thru"
	Set gtmtypes("jnl_ctl_list",29,"off")=536
	Set gtmtypes("jnl_ctl_list",29,"len")=8
	Set gtmtypes("jnl_ctl_list",29,"type")="addr"
	Set gtmtypfldindx("jnl_ctl_list","pini_list.entry_passed_thru")=29
	Set gtmtypes("jnl_ctl_list",30,"name")="jnl_ctl_list.pini_list.size"
	Set gtmtypes("jnl_ctl_list",30,"off")=544
	Set gtmtypes("jnl_ctl_list",30,"len")=4
	Set gtmtypes("jnl_ctl_list",30,"type")="unsigned-int"
	Set gtmtypfldindx("jnl_ctl_list","pini_list.size")=30
	Set gtmtypes("jnl_ctl_list",31,"name")="jnl_ctl_list.pini_list.initial_size"
	Set gtmtypes("jnl_ctl_list",31,"off")=548
	Set gtmtypes("jnl_ctl_list",31,"len")=4
	Set gtmtypes("jnl_ctl_list",31,"type")="unsigned-int"
	Set gtmtypfldindx("jnl_ctl_list","pini_list.initial_size")=31
	Set gtmtypes("jnl_ctl_list",32,"name")="jnl_ctl_list.pini_list.spare_base_size"
	Set gtmtypes("jnl_ctl_list",32,"off")=552
	Set gtmtypes("jnl_ctl_list",32,"len")=4
	Set gtmtypes("jnl_ctl_list",32,"type")="unsigned-int"
	Set gtmtypfldindx("jnl_ctl_list","pini_list.spare_base_size")=32
	Set gtmtypes("jnl_ctl_list",33,"name")="jnl_ctl_list.pini_list.count"
	Set gtmtypes("jnl_ctl_list",33,"off")=556
	Set gtmtypes("jnl_ctl_list",33,"len")=4
	Set gtmtypes("jnl_ctl_list",33,"type")="unsigned-int"
	Set gtmtypfldindx("jnl_ctl_list","pini_list.count")=33
	Set gtmtypes("jnl_ctl_list",34,"name")="jnl_ctl_list.pini_list.del_count"
	Set gtmtypes("jnl_ctl_list",34,"off")=560
	Set gtmtypes("jnl_ctl_list",34,"len")=4
	Set gtmtypes("jnl_ctl_list",34,"type")="unsigned-int"
	Set gtmtypfldindx("jnl_ctl_list","pini_list.del_count")=34
	Set gtmtypes("jnl_ctl_list",35,"name")="jnl_ctl_list.pini_list.exp_trigger_size"
	Set gtmtypes("jnl_ctl_list",35,"off")=564
	Set gtmtypes("jnl_ctl_list",35,"len")=4
	Set gtmtypes("jnl_ctl_list",35,"type")="unsigned-int"
	Set gtmtypfldindx("jnl_ctl_list","pini_list.exp_trigger_size")=35
	Set gtmtypes("jnl_ctl_list",36,"name")="jnl_ctl_list.pini_list.cmp_trigger_size"
	Set gtmtypes("jnl_ctl_list",36,"off")=568
	Set gtmtypes("jnl_ctl_list",36,"len")=4
	Set gtmtypes("jnl_ctl_list",36,"type")="unsigned-int"
	Set gtmtypfldindx("jnl_ctl_list","pini_list.cmp_trigger_size")=36
	Set gtmtypes("jnl_ctl_list",37,"name")="jnl_ctl_list.pini_list.dont_compact"
	Set gtmtypes("jnl_ctl_list",37,"off")=572
	Set gtmtypes("jnl_ctl_list",37,"len")=4
	Set gtmtypes("jnl_ctl_list",37,"type")="boolean_t"
	Set gtmtypfldindx("jnl_ctl_list","pini_list.dont_compact")=37
	Set gtmtypes("jnl_ctl_list",38,"name")="jnl_ctl_list.pini_list.dont_keep_spare_table"
	Set gtmtypes("jnl_ctl_list",38,"off")=576
	Set gtmtypes("jnl_ctl_list",38,"len")=4
	Set gtmtypes("jnl_ctl_list",38,"type")="boolean_t"
	Set gtmtypfldindx("jnl_ctl_list","pini_list.dont_keep_spare_table")=38
	Set gtmtypes("jnl_ctl_list",39,"name")="jnl_ctl_list.pini_list.defer_base_release"
	Set gtmtypes("jnl_ctl_list",39,"off")=580
	Set gtmtypes("jnl_ctl_list",39,"len")=4
	Set gtmtypes("jnl_ctl_list",39,"type")="boolean_t"
	Set gtmtypfldindx("jnl_ctl_list","pini_list.defer_base_release")=39
	Set gtmtypes("jnl_ctl_list",40,"name")="jnl_ctl_list.pini_list.active"
	Set gtmtypes("jnl_ctl_list",40,"off")=584
	Set gtmtypes("jnl_ctl_list",40,"len")=4
	Set gtmtypes("jnl_ctl_list",40,"type")="boolean_t"
	Set gtmtypfldindx("jnl_ctl_list","pini_list.active")=40
	Set gtmtypes("jnl_ctl_list",41,"name")="jnl_ctl_list.reg_ctl"
	Set gtmtypes("jnl_ctl_list",41,"off")=592
	Set gtmtypes("jnl_ctl_list",41,"len")=8
	Set gtmtypes("jnl_ctl_list",41,"type")="addr"
	Set gtmtypfldindx("jnl_ctl_list","reg_ctl")=41
	Set gtmtypes("jnl_ctl_list",42,"name")="jnl_ctl_list.next_gen"
	Set gtmtypes("jnl_ctl_list",42,"off")=600
	Set gtmtypes("jnl_ctl_list",42,"len")=8
	Set gtmtypes("jnl_ctl_list",42,"type")="addr"
	Set gtmtypfldindx("jnl_ctl_list","next_gen")=42
	Set gtmtypes("jnl_ctl_list",43,"name")="jnl_ctl_list.prev_gen"
	Set gtmtypes("jnl_ctl_list",43,"off")=608
	Set gtmtypes("jnl_ctl_list",43,"len")=8
	Set gtmtypes("jnl_ctl_list",43,"type")="addr"
	Set gtmtypfldindx("jnl_ctl_list","prev_gen")=43
	Set gtmtypes("jnl_ctl_list",44,"name")="jnl_ctl_list.encr_key_handle"
	Set gtmtypes("jnl_ctl_list",44,"off")=616
	Set gtmtypes("jnl_ctl_list",44,"len")=8
	Set gtmtypes("jnl_ctl_list",44,"type")="addr"
	Set gtmtypfldindx("jnl_ctl_list","encr_key_handle")=44
	Set gtmtypes("jnl_ctl_list",45,"name")="jnl_ctl_list.encr_key_handle2"
	Set gtmtypes("jnl_ctl_list",45,"off")=624
	Set gtmtypes("jnl_ctl_list",45,"len")=8
	Set gtmtypes("jnl_ctl_list",45,"type")="addr"
	Set gtmtypfldindx("jnl_ctl_list","encr_key_handle2")=45
	Set gtmtypes("jnl_ctl_list",46,"name")="jnl_ctl_list.same_encryption_settings"
	Set gtmtypes("jnl_ctl_list",46,"off")=632
	Set gtmtypes("jnl_ctl_list",46,"len")=4
	Set gtmtypes("jnl_ctl_list",46,"type")="boolean_t"
	Set gtmtypfldindx("jnl_ctl_list","same_encryption_settings")=46
	Set gtmtypes("jnl_ctl_list",47,"name")="jnl_ctl_list.turn_around_fullyupgraded"
	Set gtmtypes("jnl_ctl_list",47,"off")=636
	Set gtmtypes("jnl_ctl_list",47,"len")=4
	Set gtmtypes("jnl_ctl_list",47,"type")="boolean_t"
	Set gtmtypfldindx("jnl_ctl_list","turn_around_fullyupgraded")=47
	;
	Set gtmtypes("jnl_fence_control")="struct"
	Set gtmtypes("jnl_fence_control",0)=5
	Set gtmtypes("jnl_fence_control","len")=32
	Set gtmtypes("jnl_fence_control",1,"name")="jnl_fence_control.fence_list"
	Set gtmtypes("jnl_fence_control",1,"off")=0
	Set gtmtypes("jnl_fence_control",1,"len")=8
	Set gtmtypes("jnl_fence_control",1,"type")="addr"
	Set gtmtypfldindx("jnl_fence_control","fence_list")=1
	Set gtmtypes("jnl_fence_control",2,"name")="jnl_fence_control.level"
	Set gtmtypes("jnl_fence_control",2,"off")=8
	Set gtmtypes("jnl_fence_control",2,"len")=4
	Set gtmtypes("jnl_fence_control",2,"type")="int"
	Set gtmtypfldindx("jnl_fence_control","level")=2
	Set gtmtypes("jnl_fence_control",3,"name")="jnl_fence_control.replication"
	Set gtmtypes("jnl_fence_control",3,"off")=12
	Set gtmtypes("jnl_fence_control",3,"len")=4
	Set gtmtypes("jnl_fence_control",3,"type")="boolean_t"
	Set gtmtypfldindx("jnl_fence_control","replication")=3
	Set gtmtypes("jnl_fence_control",4,"name")="jnl_fence_control.token"
	Set gtmtypes("jnl_fence_control",4,"off")=16
	Set gtmtypes("jnl_fence_control",4,"len")=8
	Set gtmtypes("jnl_fence_control",4,"type")="uint64_t"
	Set gtmtypfldindx("jnl_fence_control","token")=4
	Set gtmtypes("jnl_fence_control",5,"name")="jnl_fence_control.strm_seqno"
	Set gtmtypes("jnl_fence_control",5,"off")=24
	Set gtmtypes("jnl_fence_control",5,"len")=8
	Set gtmtypes("jnl_fence_control",5,"type")="uint64_t"
	Set gtmtypfldindx("jnl_fence_control","strm_seqno")=5
	;
	Set gtmtypes("jnl_file_header")="struct"
	Set gtmtypes("jnl_file_header",0)=67
	Set gtmtypes("jnl_file_header","len")=2048
	Set gtmtypes("jnl_file_header",1,"name")="jnl_file_header.label"
	Set gtmtypes("jnl_file_header",1,"off")=0
	Set gtmtypes("jnl_file_header",1,"len")=8
	Set gtmtypes("jnl_file_header",1,"type")="char"
	Set gtmtypfldindx("jnl_file_header","label")=1
	Set gtmtypes("jnl_file_header",2,"name")="jnl_file_header.is_little_endian"
	Set gtmtypes("jnl_file_header",2,"off")=8
	Set gtmtypes("jnl_file_header",2,"len")=1
	Set gtmtypes("jnl_file_header",2,"type")="char"
	Set gtmtypfldindx("jnl_file_header","is_little_endian")=2
	Set gtmtypes("jnl_file_header",3,"name")="jnl_file_header.filler_align8"
	Set gtmtypes("jnl_file_header",3,"off")=9
	Set gtmtypes("jnl_file_header",3,"len")=7
	Set gtmtypes("jnl_file_header",3,"type")="char"
	Set gtmtypfldindx("jnl_file_header","filler_align8")=3
	Set gtmtypes("jnl_file_header",4,"name")="jnl_file_header.who_created"
	Set gtmtypes("jnl_file_header",4,"off")=16
	Set gtmtypes("jnl_file_header",4,"len")=88
	Set gtmtypes("jnl_file_header",4,"type")="jnl_process_vector"
	Set gtmtypfldindx("jnl_file_header","who_created")=4
	Set gtmtypes("jnl_file_header",5,"name")="jnl_file_header.who_created.jpv_pid"
	Set gtmtypes("jnl_file_header",5,"off")=16
	Set gtmtypes("jnl_file_header",5,"len")=4
	Set gtmtypes("jnl_file_header",5,"type")="unsigned-int"
	Set gtmtypfldindx("jnl_file_header","who_created.jpv_pid")=5
	Set gtmtypes("jnl_file_header",6,"name")="jnl_file_header.who_created.jpv_image_count"
	Set gtmtypes("jnl_file_header",6,"off")=20
	Set gtmtypes("jnl_file_header",6,"len")=4
	Set gtmtypes("jnl_file_header",6,"type")="int"
	Set gtmtypfldindx("jnl_file_header","who_created.jpv_image_count")=6
	Set gtmtypes("jnl_file_header",7,"name")="jnl_file_header.who_created.jpv_time"
	Set gtmtypes("jnl_file_header",7,"off")=24
	Set gtmtypes("jnl_file_header",7,"len")=8
	Set gtmtypes("jnl_file_header",7,"type")="int64_t"
	Set gtmtypfldindx("jnl_file_header","who_created.jpv_time")=7
	Set gtmtypes("jnl_file_header",8,"name")="jnl_file_header.who_created.jpv_login_time"
	Set gtmtypes("jnl_file_header",8,"off")=32
	Set gtmtypes("jnl_file_header",8,"len")=8
	Set gtmtypes("jnl_file_header",8,"type")="int64_t"
	Set gtmtypfldindx("jnl_file_header","who_created.jpv_login_time")=8
	Set gtmtypes("jnl_file_header",9,"name")="jnl_file_header.who_created.jpv_node"
	Set gtmtypes("jnl_file_header",9,"off")=40
	Set gtmtypes("jnl_file_header",9,"len")=16
	Set gtmtypes("jnl_file_header",9,"type")="char"
	Set gtmtypfldindx("jnl_file_header","who_created.jpv_node")=9
	Set gtmtypes("jnl_file_header",10,"name")="jnl_file_header.who_created.jpv_user"
	Set gtmtypes("jnl_file_header",10,"off")=56
	Set gtmtypes("jnl_file_header",10,"len")=12
	Set gtmtypes("jnl_file_header",10,"type")="char"
	Set gtmtypfldindx("jnl_file_header","who_created.jpv_user")=10
	Set gtmtypes("jnl_file_header",11,"name")="jnl_file_header.who_created.jpv_prcnam"
	Set gtmtypes("jnl_file_header",11,"off")=68
	Set gtmtypes("jnl_file_header",11,"len")=16
	Set gtmtypes("jnl_file_header",11,"type")="char"
	Set gtmtypfldindx("jnl_file_header","who_created.jpv_prcnam")=11
	Set gtmtypes("jnl_file_header",12,"name")="jnl_file_header.who_created.jpv_terminal"
	Set gtmtypes("jnl_file_header",12,"off")=84
	Set gtmtypes("jnl_file_header",12,"len")=15
	Set gtmtypes("jnl_file_header",12,"type")="char"
	Set gtmtypfldindx("jnl_file_header","who_created.jpv_terminal")=12
	Set gtmtypes("jnl_file_header",13,"name")="jnl_file_header.who_created.jpv_mode"
	Set gtmtypes("jnl_file_header",13,"off")=99
	Set gtmtypes("jnl_file_header",13,"len")=1
	Set gtmtypes("jnl_file_header",13,"type")="unsigned-char"
	Set gtmtypfldindx("jnl_file_header","who_created.jpv_mode")=13
	Set gtmtypes("jnl_file_header",14,"name")="jnl_file_header.who_created.filler"
	Set gtmtypes("jnl_file_header",14,"off")=100
	Set gtmtypes("jnl_file_header",14,"len")=4
	Set gtmtypes("jnl_file_header",14,"type")="int"
	Set gtmtypfldindx("jnl_file_header","who_created.filler")=14
	Set gtmtypes("jnl_file_header",15,"name")="jnl_file_header.who_opened"
	Set gtmtypes("jnl_file_header",15,"off")=104
	Set gtmtypes("jnl_file_header",15,"len")=88
	Set gtmtypes("jnl_file_header",15,"type")="jnl_process_vector"
	Set gtmtypfldindx("jnl_file_header","who_opened")=15
	Set gtmtypes("jnl_file_header",16,"name")="jnl_file_header.who_opened.jpv_pid"
	Set gtmtypes("jnl_file_header",16,"off")=104
	Set gtmtypes("jnl_file_header",16,"len")=4
	Set gtmtypes("jnl_file_header",16,"type")="unsigned-int"
	Set gtmtypfldindx("jnl_file_header","who_opened.jpv_pid")=16
	Set gtmtypes("jnl_file_header",17,"name")="jnl_file_header.who_opened.jpv_image_count"
	Set gtmtypes("jnl_file_header",17,"off")=108
	Set gtmtypes("jnl_file_header",17,"len")=4
	Set gtmtypes("jnl_file_header",17,"type")="int"
	Set gtmtypfldindx("jnl_file_header","who_opened.jpv_image_count")=17
	Set gtmtypes("jnl_file_header",18,"name")="jnl_file_header.who_opened.jpv_time"
	Set gtmtypes("jnl_file_header",18,"off")=112
	Set gtmtypes("jnl_file_header",18,"len")=8
	Set gtmtypes("jnl_file_header",18,"type")="int64_t"
	Set gtmtypfldindx("jnl_file_header","who_opened.jpv_time")=18
	Set gtmtypes("jnl_file_header",19,"name")="jnl_file_header.who_opened.jpv_login_time"
	Set gtmtypes("jnl_file_header",19,"off")=120
	Set gtmtypes("jnl_file_header",19,"len")=8
	Set gtmtypes("jnl_file_header",19,"type")="int64_t"
	Set gtmtypfldindx("jnl_file_header","who_opened.jpv_login_time")=19
	Set gtmtypes("jnl_file_header",20,"name")="jnl_file_header.who_opened.jpv_node"
	Set gtmtypes("jnl_file_header",20,"off")=128
	Set gtmtypes("jnl_file_header",20,"len")=16
	Set gtmtypes("jnl_file_header",20,"type")="char"
	Set gtmtypfldindx("jnl_file_header","who_opened.jpv_node")=20
	Set gtmtypes("jnl_file_header",21,"name")="jnl_file_header.who_opened.jpv_user"
	Set gtmtypes("jnl_file_header",21,"off")=144
	Set gtmtypes("jnl_file_header",21,"len")=12
	Set gtmtypes("jnl_file_header",21,"type")="char"
	Set gtmtypfldindx("jnl_file_header","who_opened.jpv_user")=21
	Set gtmtypes("jnl_file_header",22,"name")="jnl_file_header.who_opened.jpv_prcnam"
	Set gtmtypes("jnl_file_header",22,"off")=156
	Set gtmtypes("jnl_file_header",22,"len")=16
	Set gtmtypes("jnl_file_header",22,"type")="char"
	Set gtmtypfldindx("jnl_file_header","who_opened.jpv_prcnam")=22
	Set gtmtypes("jnl_file_header",23,"name")="jnl_file_header.who_opened.jpv_terminal"
	Set gtmtypes("jnl_file_header",23,"off")=172
	Set gtmtypes("jnl_file_header",23,"len")=15
	Set gtmtypes("jnl_file_header",23,"type")="char"
	Set gtmtypfldindx("jnl_file_header","who_opened.jpv_terminal")=23
	Set gtmtypes("jnl_file_header",24,"name")="jnl_file_header.who_opened.jpv_mode"
	Set gtmtypes("jnl_file_header",24,"off")=187
	Set gtmtypes("jnl_file_header",24,"len")=1
	Set gtmtypes("jnl_file_header",24,"type")="unsigned-char"
	Set gtmtypfldindx("jnl_file_header","who_opened.jpv_mode")=24
	Set gtmtypes("jnl_file_header",25,"name")="jnl_file_header.who_opened.filler"
	Set gtmtypes("jnl_file_header",25,"off")=188
	Set gtmtypes("jnl_file_header",25,"len")=4
	Set gtmtypes("jnl_file_header",25,"type")="int"
	Set gtmtypfldindx("jnl_file_header","who_opened.filler")=25
	Set gtmtypes("jnl_file_header",26,"name")="jnl_file_header.bov_timestamp"
	Set gtmtypes("jnl_file_header",26,"off")=192
	Set gtmtypes("jnl_file_header",26,"len")=8
	Set gtmtypes("jnl_file_header",26,"type")="int64_t"
	Set gtmtypfldindx("jnl_file_header","bov_timestamp")=26
	Set gtmtypes("jnl_file_header",27,"name")="jnl_file_header.eov_timestamp"
	Set gtmtypes("jnl_file_header",27,"off")=200
	Set gtmtypes("jnl_file_header",27,"len")=8
	Set gtmtypes("jnl_file_header",27,"type")="int64_t"
	Set gtmtypfldindx("jnl_file_header","eov_timestamp")=27
	Set gtmtypes("jnl_file_header",28,"name")="jnl_file_header.bov_tn"
	Set gtmtypes("jnl_file_header",28,"off")=208
	Set gtmtypes("jnl_file_header",28,"len")=8
	Set gtmtypes("jnl_file_header",28,"type")="uint64_t"
	Set gtmtypfldindx("jnl_file_header","bov_tn")=28
	Set gtmtypes("jnl_file_header",29,"name")="jnl_file_header.eov_tn"
	Set gtmtypes("jnl_file_header",29,"off")=216
	Set gtmtypes("jnl_file_header",29,"len")=8
	Set gtmtypes("jnl_file_header",29,"type")="uint64_t"
	Set gtmtypfldindx("jnl_file_header","eov_tn")=29
	Set gtmtypes("jnl_file_header",30,"name")="jnl_file_header.start_seqno"
	Set gtmtypes("jnl_file_header",30,"off")=224
	Set gtmtypes("jnl_file_header",30,"len")=8
	Set gtmtypes("jnl_file_header",30,"type")="uint64_t"
	Set gtmtypfldindx("jnl_file_header","start_seqno")=30
	Set gtmtypes("jnl_file_header",31,"name")="jnl_file_header.end_seqno"
	Set gtmtypes("jnl_file_header",31,"off")=232
	Set gtmtypes("jnl_file_header",31,"len")=8
	Set gtmtypes("jnl_file_header",31,"type")="uint64_t"
	Set gtmtypfldindx("jnl_file_header","end_seqno")=31
	Set gtmtypes("jnl_file_header",32,"name")="jnl_file_header.end_of_data"
	Set gtmtypes("jnl_file_header",32,"off")=240
	Set gtmtypes("jnl_file_header",32,"len")=4
	Set gtmtypes("jnl_file_header",32,"type")="unsigned-int"
	Set gtmtypfldindx("jnl_file_header","end_of_data")=32
	Set gtmtypes("jnl_file_header",33,"name")="jnl_file_header.prev_recov_end_of_data"
	Set gtmtypes("jnl_file_header",33,"off")=244
	Set gtmtypes("jnl_file_header",33,"len")=4
	Set gtmtypes("jnl_file_header",33,"type")="unsigned-int"
	Set gtmtypfldindx("jnl_file_header","prev_recov_end_of_data")=33
	Set gtmtypes("jnl_file_header",34,"name")="jnl_file_header.virtual_size"
	Set gtmtypes("jnl_file_header",34,"off")=248
	Set gtmtypes("jnl_file_header",34,"len")=4
	Set gtmtypes("jnl_file_header",34,"type")="unsigned-int"
	Set gtmtypfldindx("jnl_file_header","virtual_size")=34
	Set gtmtypes("jnl_file_header",35,"name")="jnl_file_header.crash"
	Set gtmtypes("jnl_file_header",35,"off")=252
	Set gtmtypes("jnl_file_header",35,"len")=4
	Set gtmtypes("jnl_file_header",35,"type")="boolean_t"
	Set gtmtypfldindx("jnl_file_header","crash")=35
	Set gtmtypes("jnl_file_header",36,"name")="jnl_file_header.recover_interrupted"
	Set gtmtypes("jnl_file_header",36,"off")=256
	Set gtmtypes("jnl_file_header",36,"len")=4
	Set gtmtypes("jnl_file_header",36,"type")="boolean_t"
	Set gtmtypfldindx("jnl_file_header","recover_interrupted")=36
	Set gtmtypes("jnl_file_header",37,"name")="jnl_file_header.turn_around_offset"
	Set gtmtypes("jnl_file_header",37,"off")=260
	Set gtmtypes("jnl_file_header",37,"len")=4
	Set gtmtypes("jnl_file_header",37,"type")="unsigned-int"
	Set gtmtypfldindx("jnl_file_header","turn_around_offset")=37
	Set gtmtypes("jnl_file_header",38,"name")="jnl_file_header.turn_around_time"
	Set gtmtypes("jnl_file_header",38,"off")=264
	Set gtmtypes("jnl_file_header",38,"len")=4
	Set gtmtypes("jnl_file_header",38,"type")="unsigned-int"
	Set gtmtypfldindx("jnl_file_header","turn_around_time")=38
	Set gtmtypes("jnl_file_header",39,"name")="jnl_file_header.before_images"
	Set gtmtypes("jnl_file_header",39,"off")=268
	Set gtmtypes("jnl_file_header",39,"len")=4
	Set gtmtypes("jnl_file_header",39,"type")="boolean_t"
	Set gtmtypfldindx("jnl_file_header","before_images")=39
	Set gtmtypes("jnl_file_header",40,"name")="jnl_file_header.alignsize"
	Set gtmtypes("jnl_file_header",40,"off")=272
	Set gtmtypes("jnl_file_header",40,"len")=4
	Set gtmtypes("jnl_file_header",40,"type")="unsigned-int"
	Set gtmtypfldindx("jnl_file_header","alignsize")=40
	Set gtmtypes("jnl_file_header",41,"name")="jnl_file_header.epoch_interval"
	Set gtmtypes("jnl_file_header",41,"off")=276
	Set gtmtypes("jnl_file_header",41,"len")=4
	Set gtmtypes("jnl_file_header",41,"type")="int"
	Set gtmtypfldindx("jnl_file_header","epoch_interval")=41
	Set gtmtypes("jnl_file_header",42,"name")="jnl_file_header.repl_state"
	Set gtmtypes("jnl_file_header",42,"off")=280
	Set gtmtypes("jnl_file_header",42,"len")=4
	Set gtmtypes("jnl_file_header",42,"type")="int"
	Set gtmtypfldindx("jnl_file_header","repl_state")=42
	Set gtmtypes("jnl_file_header",43,"name")="jnl_file_header.autoswitchlimit"
	Set gtmtypes("jnl_file_header",43,"off")=284
	Set gtmtypes("jnl_file_header",43,"len")=4
	Set gtmtypes("jnl_file_header",43,"type")="unsigned-int"
	Set gtmtypfldindx("jnl_file_header","autoswitchlimit")=43
	Set gtmtypes("jnl_file_header",44,"name")="jnl_file_header.jnl_alq"
	Set gtmtypes("jnl_file_header",44,"off")=288
	Set gtmtypes("jnl_file_header",44,"len")=4
	Set gtmtypes("jnl_file_header",44,"type")="unsigned-int"
	Set gtmtypfldindx("jnl_file_header","jnl_alq")=44
	Set gtmtypes("jnl_file_header",45,"name")="jnl_file_header.jnl_deq"
	Set gtmtypes("jnl_file_header",45,"off")=292
	Set gtmtypes("jnl_file_header",45,"len")=4
	Set gtmtypes("jnl_file_header",45,"type")="unsigned-int"
	Set gtmtypfldindx("jnl_file_header","jnl_deq")=45
	Set gtmtypes("jnl_file_header",46,"name")="jnl_file_header.filler_update_disabled"
	Set gtmtypes("jnl_file_header",46,"off")=296
	Set gtmtypes("jnl_file_header",46,"len")=4
	Set gtmtypes("jnl_file_header",46,"type")="boolean_t"
	Set gtmtypfldindx("jnl_file_header","filler_update_disabled")=46
	Set gtmtypes("jnl_file_header",47,"name")="jnl_file_header.max_jrec_len"
	Set gtmtypes("jnl_file_header",47,"off")=300
	Set gtmtypes("jnl_file_header",47,"len")=4
	Set gtmtypes("jnl_file_header",47,"type")="int"
	Set gtmtypfldindx("jnl_file_header","max_jrec_len")=47
	Set gtmtypes("jnl_file_header",48,"name")="jnl_file_header.data_file_name_length"
	Set gtmtypes("jnl_file_header",48,"off")=304
	Set gtmtypes("jnl_file_header",48,"len")=4
	Set gtmtypes("jnl_file_header",48,"type")="unsigned-int"
	Set gtmtypfldindx("jnl_file_header","data_file_name_length")=48
	Set gtmtypes("jnl_file_header",49,"name")="jnl_file_header.prev_jnl_file_name_length"
	Set gtmtypes("jnl_file_header",49,"off")=308
	Set gtmtypes("jnl_file_header",49,"len")=4
	Set gtmtypes("jnl_file_header",49,"type")="unsigned-int"
	Set gtmtypfldindx("jnl_file_header","prev_jnl_file_name_length")=49
	Set gtmtypes("jnl_file_header",50,"name")="jnl_file_header.next_jnl_file_name_length"
	Set gtmtypes("jnl_file_header",50,"off")=312
	Set gtmtypes("jnl_file_header",50,"len")=4
	Set gtmtypes("jnl_file_header",50,"type")="unsigned-int"
	Set gtmtypfldindx("jnl_file_header","next_jnl_file_name_length")=50
	Set gtmtypes("jnl_file_header",51,"name")="jnl_file_header.checksum"
	Set gtmtypes("jnl_file_header",51,"off")=316
	Set gtmtypes("jnl_file_header",51,"len")=4
	Set gtmtypes("jnl_file_header",51,"type")="unsigned-int"
	Set gtmtypfldindx("jnl_file_header","checksum")=51
	Set gtmtypes("jnl_file_header",52,"name")="jnl_file_header.prev_recov_blks_to_upgrd_adjust"
	Set gtmtypes("jnl_file_header",52,"off")=320
	Set gtmtypes("jnl_file_header",52,"len")=4
	Set gtmtypes("jnl_file_header",52,"type")="unsigned-int"
	Set gtmtypfldindx("jnl_file_header","prev_recov_blks_to_upgrd_adjust")=52
	Set gtmtypes("jnl_file_header",53,"name")="jnl_file_header.data_file_name"
	Set gtmtypes("jnl_file_header",53,"off")=324
	Set gtmtypes("jnl_file_header",53,"len")=256
	Set gtmtypes("jnl_file_header",53,"type")="unsigned-char"
	Set gtmtypfldindx("jnl_file_header","data_file_name")=53
	Set gtmtypes("jnl_file_header",54,"name")="jnl_file_header.prev_jnl_file_name"
	Set gtmtypes("jnl_file_header",54,"off")=580
	Set gtmtypes("jnl_file_header",54,"len")=256
	Set gtmtypes("jnl_file_header",54,"type")="unsigned-char"
	Set gtmtypfldindx("jnl_file_header","prev_jnl_file_name")=54
	Set gtmtypes("jnl_file_header",55,"name")="jnl_file_header.next_jnl_file_name"
	Set gtmtypes("jnl_file_header",55,"off")=836
	Set gtmtypes("jnl_file_header",55,"len")=256
	Set gtmtypes("jnl_file_header",55,"type")="unsigned-char"
	Set gtmtypfldindx("jnl_file_header","next_jnl_file_name")=55
	Set gtmtypes("jnl_file_header",56,"name")="jnl_file_header.is_encrypted"
	Set gtmtypes("jnl_file_header",56,"off")=1092
	Set gtmtypes("jnl_file_header",56,"len")=4
	Set gtmtypes("jnl_file_header",56,"type")="int"
	Set gtmtypfldindx("jnl_file_header","is_encrypted")=56
	Set gtmtypes("jnl_file_header",57,"name")="jnl_file_header.encryption_hash"
	Set gtmtypes("jnl_file_header",57,"off")=1096
	Set gtmtypes("jnl_file_header",57,"len")=80
	Set gtmtypes("jnl_file_header",57,"type")="char"
	Set gtmtypfldindx("jnl_file_header","encryption_hash")=57
	Set gtmtypes("jnl_file_header",58,"name")="jnl_file_header.encryption_hash2"
	Set gtmtypes("jnl_file_header",58,"off")=1176
	Set gtmtypes("jnl_file_header",58,"len")=80
	Set gtmtypes("jnl_file_header",58,"type")="char"
	Set gtmtypfldindx("jnl_file_header","encryption_hash2")=58
	Set gtmtypes("jnl_file_header",59,"name")="jnl_file_header.non_null_iv"
	Set gtmtypes("jnl_file_header",59,"off")=1256
	Set gtmtypes("jnl_file_header",59,"len")=4
	Set gtmtypes("jnl_file_header",59,"type")="boolean_t"
	Set gtmtypfldindx("jnl_file_header","non_null_iv")=59
	Set gtmtypes("jnl_file_header",60,"name")="jnl_file_header.encryption_hash_cutoff"
	Set gtmtypes("jnl_file_header",60,"off")=1260
	Set gtmtypes("jnl_file_header",60,"len")=4
	Set gtmtypes("jnl_file_header",60,"type")="int"
	Set gtmtypfldindx("jnl_file_header","encryption_hash_cutoff")=60
	Set gtmtypes("jnl_file_header",61,"name")="jnl_file_header.encryption_hash2_start_tn"
	Set gtmtypes("jnl_file_header",61,"off")=1264
	Set gtmtypes("jnl_file_header",61,"len")=8
	Set gtmtypes("jnl_file_header",61,"type")="uint64_t"
	Set gtmtypfldindx("jnl_file_header","encryption_hash2_start_tn")=61
	Set gtmtypes("jnl_file_header",62,"name")="jnl_file_header.encr_filler"
	Set gtmtypes("jnl_file_header",62,"off")=1272
	Set gtmtypes("jnl_file_header",62,"len")=80
	Set gtmtypes("jnl_file_header",62,"type")="char"
	Set gtmtypfldindx("jnl_file_header","encr_filler")=62
	Set gtmtypes("jnl_file_header",63,"name")="jnl_file_header.strm_start_seqno"
	Set gtmtypes("jnl_file_header",63,"off")=1352
	Set gtmtypes("jnl_file_header",63,"len")=128
	Set gtmtypes("jnl_file_header",63,"type")="uint64_t"
	Set gtmtypfldindx("jnl_file_header","strm_start_seqno")=63
	Set gtmtypes("jnl_file_header",63,"dim")=16
	Set gtmtypes("jnl_file_header",64,"name")="jnl_file_header.strm_end_seqno"
	Set gtmtypes("jnl_file_header",64,"off")=1480
	Set gtmtypes("jnl_file_header",64,"len")=128
	Set gtmtypes("jnl_file_header",64,"type")="uint64_t"
	Set gtmtypfldindx("jnl_file_header","strm_end_seqno")=64
	Set gtmtypes("jnl_file_header",64,"dim")=16
	Set gtmtypes("jnl_file_header",65,"name")="jnl_file_header.last_eof_written"
	Set gtmtypes("jnl_file_header",65,"off")=1608
	Set gtmtypes("jnl_file_header",65,"len")=4
	Set gtmtypes("jnl_file_header",65,"type")="boolean_t"
	Set gtmtypfldindx("jnl_file_header","last_eof_written")=65
	Set gtmtypes("jnl_file_header",66,"name")="jnl_file_header.is_not_latest_jnl"
	Set gtmtypes("jnl_file_header",66,"off")=1612
	Set gtmtypes("jnl_file_header",66,"len")=4
	Set gtmtypes("jnl_file_header",66,"type")="boolean_t"
	Set gtmtypfldindx("jnl_file_header","is_not_latest_jnl")=66
	Set gtmtypes("jnl_file_header",67,"name")="jnl_file_header.filler"
	Set gtmtypes("jnl_file_header",67,"off")=1616
	Set gtmtypes("jnl_file_header",67,"len")=432
	Set gtmtypes("jnl_file_header",67,"type")="char"
	Set gtmtypfldindx("jnl_file_header","filler")=67
	;
	Set gtmtypes("jnl_format_buffer")="struct"
	Set gtmtypes("jnl_format_buffer",0)=14
	Set gtmtypes("jnl_format_buffer","len")=80
	Set gtmtypes("jnl_format_buffer",1,"name")="jnl_format_buffer.free_que"
	Set gtmtypes("jnl_format_buffer",1,"off")=0
	Set gtmtypes("jnl_format_buffer",1,"len")=16
	Set gtmtypes("jnl_format_buffer",1,"type")="que_ent"
	Set gtmtypfldindx("jnl_format_buffer","free_que")=1
	Set gtmtypes("jnl_format_buffer",2,"name")="jnl_format_buffer.free_que.fl"
	Set gtmtypes("jnl_format_buffer",2,"off")=0
	Set gtmtypes("jnl_format_buffer",2,"len")=8
	Set gtmtypes("jnl_format_buffer",2,"type")="intptr_t"
	Set gtmtypfldindx("jnl_format_buffer","free_que.fl")=2
	Set gtmtypes("jnl_format_buffer",3,"name")="jnl_format_buffer.free_que.bl"
	Set gtmtypes("jnl_format_buffer",3,"off")=8
	Set gtmtypes("jnl_format_buffer",3,"len")=8
	Set gtmtypes("jnl_format_buffer",3,"type")="intptr_t"
	Set gtmtypfldindx("jnl_format_buffer","free_que.bl")=3
	Set gtmtypes("jnl_format_buffer",4,"name")="jnl_format_buffer.next"
	Set gtmtypes("jnl_format_buffer",4,"off")=16
	Set gtmtypes("jnl_format_buffer",4,"len")=8
	Set gtmtypes("jnl_format_buffer",4,"type")="addr"
	Set gtmtypfldindx("jnl_format_buffer","next")=4
	Set gtmtypes("jnl_format_buffer",5,"name")="jnl_format_buffer.prev"
	Set gtmtypes("jnl_format_buffer",5,"off")=24
	Set gtmtypes("jnl_format_buffer",5,"len")=8
	Set gtmtypes("jnl_format_buffer",5,"type")="addr"
	Set gtmtypfldindx("jnl_format_buffer","prev")=5
	Set gtmtypes("jnl_format_buffer",6,"name")="jnl_format_buffer.rectype"
	Set gtmtypes("jnl_format_buffer",6,"off")=32
	Set gtmtypes("jnl_format_buffer",6,"len")=4
	Set gtmtypes("jnl_format_buffer",6,"type")="int"
	Set gtmtypfldindx("jnl_format_buffer","rectype")=6
	Set gtmtypes("jnl_format_buffer",7,"name")="jnl_format_buffer.record_size"
	Set gtmtypes("jnl_format_buffer",7,"off")=36
	Set gtmtypes("jnl_format_buffer",7,"len")=4
	Set gtmtypes("jnl_format_buffer",7,"type")="int"
	Set gtmtypfldindx("jnl_format_buffer","record_size")=7
	Set gtmtypes("jnl_format_buffer",8,"name")="jnl_format_buffer.hi_water_bsize"
	Set gtmtypes("jnl_format_buffer",8,"off")=40
	Set gtmtypes("jnl_format_buffer",8,"len")=4
	Set gtmtypes("jnl_format_buffer",8,"type")="int"
	Set gtmtypfldindx("jnl_format_buffer","hi_water_bsize")=8
	Set gtmtypes("jnl_format_buffer",9,"name")="jnl_format_buffer.buff"
	Set gtmtypes("jnl_format_buffer",9,"off")=48
	Set gtmtypes("jnl_format_buffer",9,"len")=8
	Set gtmtypes("jnl_format_buffer",9,"type")="addr"
	Set gtmtypfldindx("jnl_format_buffer","buff")=9
	Set gtmtypes("jnl_format_buffer",10,"name")="jnl_format_buffer.checksum"
	Set gtmtypes("jnl_format_buffer",10,"off")=56
	Set gtmtypes("jnl_format_buffer",10,"len")=4
	Set gtmtypes("jnl_format_buffer",10,"type")="unsigned-int"
	Set gtmtypfldindx("jnl_format_buffer","checksum")=10
	Set gtmtypes("jnl_format_buffer",11,"name")="jnl_format_buffer.ja"
	Set gtmtypes("jnl_format_buffer",11,"off")=60
	Set gtmtypes("jnl_format_buffer",11,"len")=8
	Set gtmtypes("jnl_format_buffer",11,"type")="jnl_action"
	Set gtmtypfldindx("jnl_format_buffer","ja")=11
	Set gtmtypes("jnl_format_buffer",12,"name")="jnl_format_buffer.ja.operation"
	Set gtmtypes("jnl_format_buffer",12,"off")=60
	Set gtmtypes("jnl_format_buffer",12,"len")=4
	Set gtmtypes("jnl_format_buffer",12,"type")="int"
	Set gtmtypfldindx("jnl_format_buffer","ja.operation")=12
	Set gtmtypes("jnl_format_buffer",13,"name")="jnl_format_buffer.ja.nodeflags"
	Set gtmtypes("jnl_format_buffer",13,"off")=64
	Set gtmtypes("jnl_format_buffer",13,"len")=4
	Set gtmtypes("jnl_format_buffer",13,"type")="unsigned-int"
	Set gtmtypfldindx("jnl_format_buffer","ja.nodeflags")=13
	Set gtmtypes("jnl_format_buffer",14,"name")="jnl_format_buffer.alt_buff"
	Set gtmtypes("jnl_format_buffer",14,"off")=72
	Set gtmtypes("jnl_format_buffer",14,"len")=8
	Set gtmtypes("jnl_format_buffer",14,"type")="addr"
	Set gtmtypfldindx("jnl_format_buffer","alt_buff")=14
	;
	Set gtmtypes("jnl_gbls_t")="struct"
	Set gtmtypes("jnl_gbls_t",0)=21
	Set gtmtypes("jnl_gbls_t","len")=104
	Set gtmtypes("jnl_gbls_t",1,"name")="jnl_gbls_t.mur_jrec_seqno"
	Set gtmtypes("jnl_gbls_t",1,"off")=0
	Set gtmtypes("jnl_gbls_t",1,"len")=8
	Set gtmtypes("jnl_gbls_t",1,"type")="uint64_t"
	Set gtmtypfldindx("jnl_gbls_t","mur_jrec_seqno")=1
	Set gtmtypes("jnl_gbls_t",2,"name")="jnl_gbls_t.mur_jrec_strm_seqno"
	Set gtmtypes("jnl_gbls_t",2,"off")=8
	Set gtmtypes("jnl_gbls_t",2,"len")=8
	Set gtmtypes("jnl_gbls_t",2,"type")="uint64_t"
	Set gtmtypfldindx("jnl_gbls_t","mur_jrec_strm_seqno")=2
	Set gtmtypes("jnl_gbls_t",3,"name")="jnl_gbls_t.filler_short"
	Set gtmtypes("jnl_gbls_t",3,"off")=16
	Set gtmtypes("jnl_gbls_t",3,"len")=2
	Set gtmtypes("jnl_gbls_t",3,"type")="unsigned-short"
	Set gtmtypfldindx("jnl_gbls_t","filler_short")=3
	Set gtmtypes("jnl_gbls_t",4,"name")="jnl_gbls_t.mur_jrec_participants"
	Set gtmtypes("jnl_gbls_t",4,"off")=18
	Set gtmtypes("jnl_gbls_t",4,"len")=2
	Set gtmtypes("jnl_gbls_t",4,"type")="unsigned-short"
	Set gtmtypfldindx("jnl_gbls_t","mur_jrec_participants")=4
	Set gtmtypes("jnl_gbls_t",5,"name")="jnl_gbls_t.gbl_jrec_time"
	Set gtmtypes("jnl_gbls_t",5,"off")=20
	Set gtmtypes("jnl_gbls_t",5,"len")=4
	Set gtmtypes("jnl_gbls_t",5,"type")="unsigned-int"
	Set gtmtypfldindx("jnl_gbls_t","gbl_jrec_time")=5
	Set gtmtypes("jnl_gbls_t",6,"name")="jnl_gbls_t.mur_tp_resolve_time"
	Set gtmtypes("jnl_gbls_t",6,"off")=24
	Set gtmtypes("jnl_gbls_t",6,"len")=4
	Set gtmtypes("jnl_gbls_t",6,"type")="unsigned-int"
	Set gtmtypfldindx("jnl_gbls_t","mur_tp_resolve_time")=6
	Set gtmtypes("jnl_gbls_t",7,"name")="jnl_gbls_t.forw_phase_recovery"
	Set gtmtypes("jnl_gbls_t",7,"off")=28
	Set gtmtypes("jnl_gbls_t",7,"len")=4
	Set gtmtypes("jnl_gbls_t",7,"type")="boolean_t"
	Set gtmtypfldindx("jnl_gbls_t","forw_phase_recovery")=7
	Set gtmtypes("jnl_gbls_t",8,"name")="jnl_gbls_t.mur_rollback"
	Set gtmtypes("jnl_gbls_t",8,"off")=32
	Set gtmtypes("jnl_gbls_t",8,"len")=4
	Set gtmtypes("jnl_gbls_t",8,"type")="boolean_t"
	Set gtmtypfldindx("jnl_gbls_t","mur_rollback")=8
	Set gtmtypes("jnl_gbls_t",9,"name")="jnl_gbls_t.mupip_journal"
	Set gtmtypes("jnl_gbls_t",9,"off")=36
	Set gtmtypes("jnl_gbls_t",9,"len")=4
	Set gtmtypes("jnl_gbls_t",9,"type")="boolean_t"
	Set gtmtypfldindx("jnl_gbls_t","mupip_journal")=9
	Set gtmtypes("jnl_gbls_t",10,"name")="jnl_gbls_t.dont_reset_gbl_jrec_time"
	Set gtmtypes("jnl_gbls_t",10,"off")=40
	Set gtmtypes("jnl_gbls_t",10,"len")=4
	Set gtmtypes("jnl_gbls_t",10,"type")="boolean_t"
	Set gtmtypfldindx("jnl_gbls_t","dont_reset_gbl_jrec_time")=10
	Set gtmtypes("jnl_gbls_t",11,"name")="jnl_gbls_t.mur_pini_addr_reset_fnptr"
	Set gtmtypes("jnl_gbls_t",11,"off")=48
	Set gtmtypes("jnl_gbls_t",11,"len")=8
	Set gtmtypes("jnl_gbls_t",11,"type")="void"
	Set gtmtypfldindx("jnl_gbls_t","mur_pini_addr_reset_fnptr")=11
	Set gtmtypes("jnl_gbls_t",12,"name")="jnl_gbls_t.cumul_jnl_rec_len"
	Set gtmtypes("jnl_gbls_t",12,"off")=56
	Set gtmtypes("jnl_gbls_t",12,"len")=4
	Set gtmtypes("jnl_gbls_t",12,"type")="unsigned-int"
	Set gtmtypfldindx("jnl_gbls_t","cumul_jnl_rec_len")=12
	Set gtmtypes("jnl_gbls_t",13,"name")="jnl_gbls_t.wait_for_jnl_hard"
	Set gtmtypes("jnl_gbls_t",13,"off")=60
	Set gtmtypes("jnl_gbls_t",13,"len")=4
	Set gtmtypes("jnl_gbls_t",13,"type")="boolean_t"
	Set gtmtypfldindx("jnl_gbls_t","wait_for_jnl_hard")=13
	Set gtmtypes("jnl_gbls_t",14,"name")="jnl_gbls_t.tp_ztp_jnl_upd_num"
	Set gtmtypes("jnl_gbls_t",14,"off")=64
	Set gtmtypes("jnl_gbls_t",14,"len")=4
	Set gtmtypes("jnl_gbls_t",14,"type")="unsigned-int"
	Set gtmtypfldindx("jnl_gbls_t","tp_ztp_jnl_upd_num")=14
	Set gtmtypes("jnl_gbls_t",15,"name")="jnl_gbls_t.mur_jrec_nodeflags"
	Set gtmtypes("jnl_gbls_t",15,"off")=68
	Set gtmtypes("jnl_gbls_t",15,"len")=4
	Set gtmtypes("jnl_gbls_t",15,"type")="unsigned-int"
	Set gtmtypfldindx("jnl_gbls_t","mur_jrec_nodeflags")=15
	Set gtmtypes("jnl_gbls_t",16,"name")="jnl_gbls_t.prev_ztworm_ptr"
	Set gtmtypes("jnl_gbls_t",16,"off")=72
	Set gtmtypes("jnl_gbls_t",16,"len")=8
	Set gtmtypes("jnl_gbls_t",16,"type")="addr"
	Set gtmtypfldindx("jnl_gbls_t","prev_ztworm_ptr")=16
	Set gtmtypes("jnl_gbls_t",17,"name")="jnl_gbls_t.save_ztworm_ptr"
	Set gtmtypes("jnl_gbls_t",17,"off")=80
	Set gtmtypes("jnl_gbls_t",17,"len")=8
	Set gtmtypes("jnl_gbls_t",17,"type")="addr"
	Set gtmtypfldindx("jnl_gbls_t","save_ztworm_ptr")=17
	Set gtmtypes("jnl_gbls_t",18,"name")="jnl_gbls_t.onlnrlbk"
	Set gtmtypes("jnl_gbls_t",18,"off")=88
	Set gtmtypes("jnl_gbls_t",18,"len")=4
	Set gtmtypes("jnl_gbls_t",18,"type")="boolean_t"
	Set gtmtypfldindx("jnl_gbls_t","onlnrlbk")=18
	Set gtmtypes("jnl_gbls_t",19,"name")="jnl_gbls_t.mur_extract"
	Set gtmtypes("jnl_gbls_t",19,"off")=92
	Set gtmtypes("jnl_gbls_t",19,"len")=4
	Set gtmtypes("jnl_gbls_t",19,"type")="boolean_t"
	Set gtmtypfldindx("jnl_gbls_t","mur_extract")=19
	Set gtmtypes("jnl_gbls_t",20,"name")="jnl_gbls_t.save_dont_reset_gbl_jrec_time"
	Set gtmtypes("jnl_gbls_t",20,"off")=96
	Set gtmtypes("jnl_gbls_t",20,"len")=4
	Set gtmtypes("jnl_gbls_t",20,"type")="boolean_t"
	Set gtmtypfldindx("jnl_gbls_t","save_dont_reset_gbl_jrec_time")=20
	Set gtmtypes("jnl_gbls_t",21,"name")="jnl_gbls_t.mur_update"
	Set gtmtypes("jnl_gbls_t",21,"off")=100
	Set gtmtypes("jnl_gbls_t",21,"len")=4
	Set gtmtypes("jnl_gbls_t",21,"type")="boolean_t"
	Set gtmtypfldindx("jnl_gbls_t","mur_update")=21
	;
	Set gtmtypes("jnl_private_control")="struct"
	Set gtmtypes("jnl_private_control",0)=21
	Set gtmtypes("jnl_private_control","len")=104
	Set gtmtypes("jnl_private_control",1,"name")="jnl_private_control.jnl_buff"
	Set gtmtypes("jnl_private_control",1,"off")=0
	Set gtmtypes("jnl_private_control",1,"len")=8
	Set gtmtypes("jnl_private_control",1,"type")="addr"
	Set gtmtypfldindx("jnl_private_control","jnl_buff")=1
	Set gtmtypes("jnl_private_control",2,"name")="jnl_private_control.region"
	Set gtmtypes("jnl_private_control",2,"off")=8
	Set gtmtypes("jnl_private_control",2,"len")=8
	Set gtmtypes("jnl_private_control",2,"type")="addr"
	Set gtmtypfldindx("jnl_private_control","region")=2
	Set gtmtypes("jnl_private_control",3,"name")="jnl_private_control.channel"
	Set gtmtypes("jnl_private_control",3,"off")=16
	Set gtmtypes("jnl_private_control",3,"len")=4
	Set gtmtypes("jnl_private_control",3,"type")="int"
	Set gtmtypfldindx("jnl_private_control","channel")=3
	Set gtmtypes("jnl_private_control",4,"name")="jnl_private_control.old_channel"
	Set gtmtypes("jnl_private_control",4,"off")=20
	Set gtmtypes("jnl_private_control",4,"len")=4
	Set gtmtypes("jnl_private_control",4,"type")="int"
	Set gtmtypfldindx("jnl_private_control","old_channel")=4
	Set gtmtypes("jnl_private_control",5,"name")="jnl_private_control.fileid"
	Set gtmtypes("jnl_private_control",5,"off")=24
	Set gtmtypes("jnl_private_control",5,"len")=16
	Set gtmtypes("jnl_private_control",5,"type")="unix_file_id"
	Set gtmtypfldindx("jnl_private_control","fileid")=5
	Set gtmtypes("jnl_private_control",6,"name")="jnl_private_control.fileid.inode"
	Set gtmtypes("jnl_private_control",6,"off")=24
	Set gtmtypes("jnl_private_control",6,"len")=8
	Set gtmtypes("jnl_private_control",6,"type")="ino_t"
	Set gtmtypfldindx("jnl_private_control","fileid.inode")=6
	Set gtmtypes("jnl_private_control",7,"name")="jnl_private_control.fileid.device"
	Set gtmtypes("jnl_private_control",7,"off")=32
	Set gtmtypes("jnl_private_control",7,"len")=8
	Set gtmtypes("jnl_private_control",7,"type")="dev_t"
	Set gtmtypfldindx("jnl_private_control","fileid.device")=7
	Set gtmtypes("jnl_private_control",8,"name")="jnl_private_control.pini_addr"
	Set gtmtypes("jnl_private_control",8,"off")=40
	Set gtmtypes("jnl_private_control",8,"len")=4
	Set gtmtypes("jnl_private_control",8,"type")="unsigned-int"
	Set gtmtypfldindx("jnl_private_control","pini_addr")=8
	Set gtmtypes("jnl_private_control",9,"name")="jnl_private_control.new_freeaddr"
	Set gtmtypes("jnl_private_control",9,"off")=44
	Set gtmtypes("jnl_private_control",9,"len")=4
	Set gtmtypes("jnl_private_control",9,"type")="unsigned-int"
	Set gtmtypfldindx("jnl_private_control","new_freeaddr")=9
	Set gtmtypes("jnl_private_control",10,"name")="jnl_private_control.phase2_freeaddr"
	Set gtmtypes("jnl_private_control",10,"off")=48
	Set gtmtypes("jnl_private_control",10,"len")=4
	Set gtmtypes("jnl_private_control",10,"type")="unsigned-int"
	Set gtmtypfldindx("jnl_private_control","phase2_freeaddr")=10
	Set gtmtypes("jnl_private_control",11,"name")="jnl_private_control.phase2_free"
	Set gtmtypes("jnl_private_control",11,"off")=52
	Set gtmtypes("jnl_private_control",11,"len")=4
	Set gtmtypes("jnl_private_control",11,"type")="unsigned-int"
	Set gtmtypfldindx("jnl_private_control","phase2_free")=11
	Set gtmtypes("jnl_private_control",12,"name")="jnl_private_control.in_jnl_phase2_salvage"
	Set gtmtypes("jnl_private_control",12,"off")=56
	Set gtmtypes("jnl_private_control",12,"len")=4
	Set gtmtypes("jnl_private_control",12,"type")="boolean_t"
	Set gtmtypfldindx("jnl_private_control","in_jnl_phase2_salvage")=12
	Set gtmtypes("jnl_private_control",13,"name")="jnl_private_control.status"
	Set gtmtypes("jnl_private_control",13,"off")=60
	Set gtmtypes("jnl_private_control",13,"len")=4
	Set gtmtypes("jnl_private_control",13,"type")="int"
	Set gtmtypfldindx("jnl_private_control","status")=13
	Set gtmtypes("jnl_private_control",14,"name")="jnl_private_control.qio_active"
	Set gtmtypes("jnl_private_control",14,"off")=64
	Set gtmtypes("jnl_private_control",14,"len")=4
	Set gtmtypes("jnl_private_control",14,"type")="boolean_t"
	Set gtmtypfldindx("jnl_private_control","qio_active")=14
	Set gtmtypes("jnl_private_control",15,"name")="jnl_private_control.fd_mismatch"
	Set gtmtypes("jnl_private_control",15,"off")=68
	Set gtmtypes("jnl_private_control",15,"len")=4
	Set gtmtypes("jnl_private_control",15,"type")="boolean_t"
	Set gtmtypfldindx("jnl_private_control","fd_mismatch")=15
	Set gtmtypes("jnl_private_control",16,"name")="jnl_private_control.sync_io"
	Set gtmtypes("jnl_private_control",16,"off")=72
	Set gtmtypes("jnl_private_control",16,"len")=4
	Set gtmtypes("jnl_private_control",16,"type")="boolean_t"
	Set gtmtypfldindx("jnl_private_control","sync_io")=16
	Set gtmtypes("jnl_private_control",17,"name")="jnl_private_control.error_reported"
	Set gtmtypes("jnl_private_control",17,"off")=76
	Set gtmtypes("jnl_private_control",17,"len")=4
	Set gtmtypes("jnl_private_control",17,"type")="boolean_t"
	Set gtmtypfldindx("jnl_private_control","error_reported")=17
	Set gtmtypes("jnl_private_control",18,"name")="jnl_private_control.status2"
	Set gtmtypes("jnl_private_control",18,"off")=80
	Set gtmtypes("jnl_private_control",18,"len")=4
	Set gtmtypes("jnl_private_control",18,"type")="unsigned-int"
	Set gtmtypfldindx("jnl_private_control","status2")=18
	Set gtmtypes("jnl_private_control",19,"name")="jnl_private_control.cycle"
	Set gtmtypes("jnl_private_control",19,"off")=84
	Set gtmtypes("jnl_private_control",19,"len")=4
	Set gtmtypes("jnl_private_control",19,"type")="unsigned-int"
	Set gtmtypfldindx("jnl_private_control","cycle")=19
	Set gtmtypes("jnl_private_control",20,"name")="jnl_private_control.err_str"
	Set gtmtypes("jnl_private_control",20,"off")=88
	Set gtmtypes("jnl_private_control",20,"len")=8
	Set gtmtypes("jnl_private_control",20,"type")="addr"
	Set gtmtypfldindx("jnl_private_control","err_str")=20
	Set gtmtypes("jnl_private_control",21,"name")="jnl_private_control.curr_tn"
	Set gtmtypes("jnl_private_control",21,"off")=96
	Set gtmtypes("jnl_private_control",21,"len")=8
	Set gtmtypes("jnl_private_control",21,"type")="uint64_t"
	Set gtmtypfldindx("jnl_private_control","curr_tn")=21
	;
	Set gtmtypes("jnl_process_vector")="struct"
	Set gtmtypes("jnl_process_vector",0)=10
	Set gtmtypes("jnl_process_vector","len")=88
	Set gtmtypes("jnl_process_vector",1,"name")="jnl_process_vector.jpv_pid"
	Set gtmtypes("jnl_process_vector",1,"off")=0
	Set gtmtypes("jnl_process_vector",1,"len")=4
	Set gtmtypes("jnl_process_vector",1,"type")="unsigned-int"
	Set gtmtypfldindx("jnl_process_vector","jpv_pid")=1
	Set gtmtypes("jnl_process_vector",2,"name")="jnl_process_vector.jpv_image_count"
	Set gtmtypes("jnl_process_vector",2,"off")=4
	Set gtmtypes("jnl_process_vector",2,"len")=4
	Set gtmtypes("jnl_process_vector",2,"type")="int"
	Set gtmtypfldindx("jnl_process_vector","jpv_image_count")=2
	Set gtmtypes("jnl_process_vector",3,"name")="jnl_process_vector.jpv_time"
	Set gtmtypes("jnl_process_vector",3,"off")=8
	Set gtmtypes("jnl_process_vector",3,"len")=8
	Set gtmtypes("jnl_process_vector",3,"type")="int64_t"
	Set gtmtypfldindx("jnl_process_vector","jpv_time")=3
	Set gtmtypes("jnl_process_vector",4,"name")="jnl_process_vector.jpv_login_time"
	Set gtmtypes("jnl_process_vector",4,"off")=16
	Set gtmtypes("jnl_process_vector",4,"len")=8
	Set gtmtypes("jnl_process_vector",4,"type")="int64_t"
	Set gtmtypfldindx("jnl_process_vector","jpv_login_time")=4
	Set gtmtypes("jnl_process_vector",5,"name")="jnl_process_vector.jpv_node"
	Set gtmtypes("jnl_process_vector",5,"off")=24
	Set gtmtypes("jnl_process_vector",5,"len")=16
	Set gtmtypes("jnl_process_vector",5,"type")="char"
	Set gtmtypfldindx("jnl_process_vector","jpv_node")=5
	Set gtmtypes("jnl_process_vector",6,"name")="jnl_process_vector.jpv_user"
	Set gtmtypes("jnl_process_vector",6,"off")=40
	Set gtmtypes("jnl_process_vector",6,"len")=12
	Set gtmtypes("jnl_process_vector",6,"type")="char"
	Set gtmtypfldindx("jnl_process_vector","jpv_user")=6
	Set gtmtypes("jnl_process_vector",7,"name")="jnl_process_vector.jpv_prcnam"
	Set gtmtypes("jnl_process_vector",7,"off")=52
	Set gtmtypes("jnl_process_vector",7,"len")=16
	Set gtmtypes("jnl_process_vector",7,"type")="char"
	Set gtmtypfldindx("jnl_process_vector","jpv_prcnam")=7
	Set gtmtypes("jnl_process_vector",8,"name")="jnl_process_vector.jpv_terminal"
	Set gtmtypes("jnl_process_vector",8,"off")=68
	Set gtmtypes("jnl_process_vector",8,"len")=15
	Set gtmtypes("jnl_process_vector",8,"type")="char"
	Set gtmtypfldindx("jnl_process_vector","jpv_terminal")=8
	Set gtmtypes("jnl_process_vector",9,"name")="jnl_process_vector.jpv_mode"
	Set gtmtypes("jnl_process_vector",9,"off")=83
	Set gtmtypes("jnl_process_vector",9,"len")=1
	Set gtmtypes("jnl_process_vector",9,"type")="unsigned-char"
	Set gtmtypfldindx("jnl_process_vector","jpv_mode")=9
	Set gtmtypes("jnl_process_vector",10,"name")="jnl_process_vector.filler"
	Set gtmtypes("jnl_process_vector",10,"off")=84
	Set gtmtypes("jnl_process_vector",10,"len")=4
	Set gtmtypes("jnl_process_vector",10,"type")="int"
	Set gtmtypfldindx("jnl_process_vector","filler")=10
	;
	Set gtmtypes("jnl_record")="union"
	Set gtmtypes("jnl_record",0)=100
	Set gtmtypes("jnl_record","len")=208
	Set gtmtypes("jnl_record",1,"name")="jnl_record.prefix"
	Set gtmtypes("jnl_record",1,"off")=0
	Set gtmtypes("jnl_record",1,"len")=24
	Set gtmtypes("jnl_record",1,"type")="jrec_prefix"
	Set gtmtypfldindx("jnl_record","prefix")=1
	Set gtmtypes("jnl_record",2,"name")="jnl_record.prefix.pini_addr"
	Set gtmtypes("jnl_record",2,"off")=4
	Set gtmtypes("jnl_record",2,"len")=4
	Set gtmtypes("jnl_record",2,"type")="unsigned-int"
	Set gtmtypfldindx("jnl_record","prefix.pini_addr")=2
	Set gtmtypes("jnl_record",3,"name")="jnl_record.prefix.time"
	Set gtmtypes("jnl_record",3,"off")=8
	Set gtmtypes("jnl_record",3,"len")=4
	Set gtmtypes("jnl_record",3,"type")="unsigned-int"
	Set gtmtypfldindx("jnl_record","prefix.time")=3
	Set gtmtypes("jnl_record",4,"name")="jnl_record.prefix.checksum"
	Set gtmtypes("jnl_record",4,"off")=12
	Set gtmtypes("jnl_record",4,"len")=4
	Set gtmtypes("jnl_record",4,"type")="unsigned-int"
	Set gtmtypfldindx("jnl_record","prefix.checksum")=4
	Set gtmtypes("jnl_record",5,"name")="jnl_record.prefix.tn"
	Set gtmtypes("jnl_record",5,"off")=16
	Set gtmtypes("jnl_record",5,"len")=8
	Set gtmtypes("jnl_record",5,"type")="uint64_t"
	Set gtmtypfldindx("jnl_record","prefix.tn")=5
	Set gtmtypes("jnl_record",6,"name")="jnl_record.jrec_set_kill"
	Set gtmtypes("jnl_record",6,"off")=0
	Set gtmtypes("jnl_record",6,"len")=56
	Set gtmtypes("jnl_record",6,"type")="struct_jrec_upd"
	Set gtmtypfldindx("jnl_record","jrec_set_kill")=6
	Set gtmtypes("jnl_record",7,"name")="jnl_record.jrec_set_kill.prefix"
	Set gtmtypes("jnl_record",7,"off")=0
	Set gtmtypes("jnl_record",7,"len")=24
	Set gtmtypes("jnl_record",7,"type")="jrec_prefix"
	Set gtmtypfldindx("jnl_record","jrec_set_kill.prefix")=7
	Set gtmtypes("jnl_record",8,"name")="jnl_record.jrec_set_kill.token_seq"
	Set gtmtypes("jnl_record",8,"off")=24
	Set gtmtypes("jnl_record",8,"len")=8
	Set gtmtypes("jnl_record",8,"type")="token_seq_t"
	Set gtmtypfldindx("jnl_record","jrec_set_kill.token_seq")=8
	Set gtmtypes("jnl_record",9,"name")="jnl_record.jrec_set_kill.strm_seqno"
	Set gtmtypes("jnl_record",9,"off")=32
	Set gtmtypes("jnl_record",9,"len")=8
	Set gtmtypes("jnl_record",9,"type")="uint64_t"
	Set gtmtypfldindx("jnl_record","jrec_set_kill.strm_seqno")=9
	Set gtmtypes("jnl_record",10,"name")="jnl_record.jrec_set_kill.update_num"
	Set gtmtypes("jnl_record",10,"off")=40
	Set gtmtypes("jnl_record",10,"len")=4
	Set gtmtypes("jnl_record",10,"type")="unsigned-int"
	Set gtmtypfldindx("jnl_record","jrec_set_kill.update_num")=10
	Set gtmtypes("jnl_record",11,"name")="jnl_record.jrec_set_kill.filler_short"
	Set gtmtypes("jnl_record",11,"off")=44
	Set gtmtypes("jnl_record",11,"len")=2
	Set gtmtypes("jnl_record",11,"type")="unsigned-short"
	Set gtmtypfldindx("jnl_record","jrec_set_kill.filler_short")=11
	Set gtmtypes("jnl_record",12,"name")="jnl_record.jrec_set_kill.num_participants"
	Set gtmtypes("jnl_record",12,"off")=46
	Set gtmtypes("jnl_record",12,"len")=2
	Set gtmtypes("jnl_record",12,"type")="unsigned-short"
	Set gtmtypfldindx("jnl_record","jrec_set_kill.num_participants")=12
	Set gtmtypes("jnl_record",13,"name")="jnl_record.jrec_set_kill.mumps_node"
	Set gtmtypes("jnl_record",13,"off")=48
	Set gtmtypes("jnl_record",13,"len")=8
	Set gtmtypes("jnl_record",13,"type")="jnl_string"
	Set gtmtypfldindx("jnl_record","jrec_set_kill.mumps_node")=13
	Set gtmtypes("jnl_record",14,"name")="jnl_record.jrec_ztworm"
	Set gtmtypes("jnl_record",14,"off")=0
	Set gtmtypes("jnl_record",14,"len")=56
	Set gtmtypes("jnl_record",14,"type")="struct_jrec_ztworm"
	Set gtmtypfldindx("jnl_record","jrec_ztworm")=14
	Set gtmtypes("jnl_record",15,"name")="jnl_record.jrec_ztworm.prefix"
	Set gtmtypes("jnl_record",15,"off")=0
	Set gtmtypes("jnl_record",15,"len")=24
	Set gtmtypes("jnl_record",15,"type")="jrec_prefix"
	Set gtmtypfldindx("jnl_record","jrec_ztworm.prefix")=15
	Set gtmtypes("jnl_record",16,"name")="jnl_record.jrec_ztworm.token_seq"
	Set gtmtypes("jnl_record",16,"off")=24
	Set gtmtypes("jnl_record",16,"len")=8
	Set gtmtypes("jnl_record",16,"type")="token_seq_t"
	Set gtmtypfldindx("jnl_record","jrec_ztworm.token_seq")=16
	Set gtmtypes("jnl_record",17,"name")="jnl_record.jrec_ztworm.strm_seqno"
	Set gtmtypes("jnl_record",17,"off")=32
	Set gtmtypes("jnl_record",17,"len")=8
	Set gtmtypes("jnl_record",17,"type")="uint64_t"
	Set gtmtypfldindx("jnl_record","jrec_ztworm.strm_seqno")=17
	Set gtmtypes("jnl_record",18,"name")="jnl_record.jrec_ztworm.update_num"
	Set gtmtypes("jnl_record",18,"off")=40
	Set gtmtypes("jnl_record",18,"len")=4
	Set gtmtypes("jnl_record",18,"type")="unsigned-int"
	Set gtmtypfldindx("jnl_record","jrec_ztworm.update_num")=18
	Set gtmtypes("jnl_record",19,"name")="jnl_record.jrec_ztworm.filler_short"
	Set gtmtypes("jnl_record",19,"off")=44
	Set gtmtypes("jnl_record",19,"len")=2
	Set gtmtypes("jnl_record",19,"type")="unsigned-short"
	Set gtmtypfldindx("jnl_record","jrec_ztworm.filler_short")=19
	Set gtmtypes("jnl_record",20,"name")="jnl_record.jrec_ztworm.num_participants"
	Set gtmtypes("jnl_record",20,"off")=46
	Set gtmtypes("jnl_record",20,"len")=2
	Set gtmtypes("jnl_record",20,"type")="unsigned-short"
	Set gtmtypfldindx("jnl_record","jrec_ztworm.num_participants")=20
	Set gtmtypes("jnl_record",21,"name")="jnl_record.jrec_ztworm.ztworm_str"
	Set gtmtypes("jnl_record",21,"off")=48
	Set gtmtypes("jnl_record",21,"len")=8
	Set gtmtypes("jnl_record",21,"type")="jnl_string"
	Set gtmtypfldindx("jnl_record","jrec_ztworm.ztworm_str")=21
	Set gtmtypes("jnl_record",22,"name")="jnl_record.jrec_lgtrig"
	Set gtmtypes("jnl_record",22,"off")=0
	Set gtmtypes("jnl_record",22,"len")=56
	Set gtmtypes("jnl_record",22,"type")="struct_jrec_lgtrig"
	Set gtmtypfldindx("jnl_record","jrec_lgtrig")=22
	Set gtmtypes("jnl_record",23,"name")="jnl_record.jrec_lgtrig.prefix"
	Set gtmtypes("jnl_record",23,"off")=0
	Set gtmtypes("jnl_record",23,"len")=24
	Set gtmtypes("jnl_record",23,"type")="jrec_prefix"
	Set gtmtypfldindx("jnl_record","jrec_lgtrig.prefix")=23
	Set gtmtypes("jnl_record",24,"name")="jnl_record.jrec_lgtrig.token_seq"
	Set gtmtypes("jnl_record",24,"off")=24
	Set gtmtypes("jnl_record",24,"len")=8
	Set gtmtypes("jnl_record",24,"type")="token_seq_t"
	Set gtmtypfldindx("jnl_record","jrec_lgtrig.token_seq")=24
	Set gtmtypes("jnl_record",25,"name")="jnl_record.jrec_lgtrig.strm_seqno"
	Set gtmtypes("jnl_record",25,"off")=32
	Set gtmtypes("jnl_record",25,"len")=8
	Set gtmtypes("jnl_record",25,"type")="uint64_t"
	Set gtmtypfldindx("jnl_record","jrec_lgtrig.strm_seqno")=25
	Set gtmtypes("jnl_record",26,"name")="jnl_record.jrec_lgtrig.update_num"
	Set gtmtypes("jnl_record",26,"off")=40
	Set gtmtypes("jnl_record",26,"len")=4
	Set gtmtypes("jnl_record",26,"type")="unsigned-int"
	Set gtmtypfldindx("jnl_record","jrec_lgtrig.update_num")=26
	Set gtmtypes("jnl_record",27,"name")="jnl_record.jrec_lgtrig.filler_short"
	Set gtmtypes("jnl_record",27,"off")=44
	Set gtmtypes("jnl_record",27,"len")=2
	Set gtmtypes("jnl_record",27,"type")="unsigned-short"
	Set gtmtypfldindx("jnl_record","jrec_lgtrig.filler_short")=27
	Set gtmtypes("jnl_record",28,"name")="jnl_record.jrec_lgtrig.num_participants"
	Set gtmtypes("jnl_record",28,"off")=46
	Set gtmtypes("jnl_record",28,"len")=2
	Set gtmtypes("jnl_record",28,"type")="unsigned-short"
	Set gtmtypfldindx("jnl_record","jrec_lgtrig.num_participants")=28
	Set gtmtypes("jnl_record",29,"name")="jnl_record.jrec_lgtrig.lgtrig_str"
	Set gtmtypes("jnl_record",29,"off")=48
	Set gtmtypes("jnl_record",29,"len")=8
	Set gtmtypes("jnl_record",29,"type")="jnl_string"
	Set gtmtypfldindx("jnl_record","jrec_lgtrig.lgtrig_str")=29
	Set gtmtypes("jnl_record",30,"name")="jnl_record.jrec_pblk"
	Set gtmtypes("jnl_record",30,"off")=0
	Set gtmtypes("jnl_record",30,"len")=48
	Set gtmtypes("jnl_record",30,"type")="struct_jrec_blk"
	Set gtmtypfldindx("jnl_record","jrec_pblk")=30
	Set gtmtypes("jnl_record",31,"name")="jnl_record.jrec_pblk.prefix"
	Set gtmtypes("jnl_record",31,"off")=0
	Set gtmtypes("jnl_record",31,"len")=24
	Set gtmtypes("jnl_record",31,"type")="jrec_prefix"
	Set gtmtypfldindx("jnl_record","jrec_pblk.prefix")=31
	Set gtmtypes("jnl_record",32,"name")="jnl_record.jrec_pblk.blknum"
	Set gtmtypes("jnl_record",32,"off")=24
	Set gtmtypes("jnl_record",32,"len")=4
	Set gtmtypes("jnl_record",32,"type")="int"
	Set gtmtypfldindx("jnl_record","jrec_pblk.blknum")=32
	Set gtmtypes("jnl_record",33,"name")="jnl_record.jrec_pblk.bsiz"
	Set gtmtypes("jnl_record",33,"off")=28
	Set gtmtypes("jnl_record",33,"len")=4
	Set gtmtypes("jnl_record",33,"type")="unsigned-int"
	Set gtmtypfldindx("jnl_record","jrec_pblk.bsiz")=33
	Set gtmtypes("jnl_record",34,"name")="jnl_record.jrec_pblk.ondsk_blkver"
	Set gtmtypes("jnl_record",34,"off")=32
	Set gtmtypes("jnl_record",34,"len")=4
	Set gtmtypes("jnl_record",34,"type")="int"
	Set gtmtypfldindx("jnl_record","jrec_pblk.ondsk_blkver")=34
	Set gtmtypes("jnl_record",35,"name")="jnl_record.jrec_pblk.filler"
	Set gtmtypes("jnl_record",35,"off")=36
	Set gtmtypes("jnl_record",35,"len")=4
	Set gtmtypes("jnl_record",35,"type")="int"
	Set gtmtypfldindx("jnl_record","jrec_pblk.filler")=35
	Set gtmtypes("jnl_record",36,"name")="jnl_record.jrec_pblk.blk_contents"
	Set gtmtypes("jnl_record",36,"off")=40
	Set gtmtypes("jnl_record",36,"len")=1
	Set gtmtypes("jnl_record",36,"type")="char"
	Set gtmtypfldindx("jnl_record","jrec_pblk.blk_contents")=36
	Set gtmtypes("jnl_record",37,"name")="jnl_record.jrec_aimg"
	Set gtmtypes("jnl_record",37,"off")=0
	Set gtmtypes("jnl_record",37,"len")=48
	Set gtmtypes("jnl_record",37,"type")="struct_jrec_blk"
	Set gtmtypfldindx("jnl_record","jrec_aimg")=37
	Set gtmtypes("jnl_record",38,"name")="jnl_record.jrec_aimg.prefix"
	Set gtmtypes("jnl_record",38,"off")=0
	Set gtmtypes("jnl_record",38,"len")=24
	Set gtmtypes("jnl_record",38,"type")="jrec_prefix"
	Set gtmtypfldindx("jnl_record","jrec_aimg.prefix")=38
	Set gtmtypes("jnl_record",39,"name")="jnl_record.jrec_aimg.blknum"
	Set gtmtypes("jnl_record",39,"off")=24
	Set gtmtypes("jnl_record",39,"len")=4
	Set gtmtypes("jnl_record",39,"type")="int"
	Set gtmtypfldindx("jnl_record","jrec_aimg.blknum")=39
	Set gtmtypes("jnl_record",40,"name")="jnl_record.jrec_aimg.bsiz"
	Set gtmtypes("jnl_record",40,"off")=28
	Set gtmtypes("jnl_record",40,"len")=4
	Set gtmtypes("jnl_record",40,"type")="unsigned-int"
	Set gtmtypfldindx("jnl_record","jrec_aimg.bsiz")=40
	Set gtmtypes("jnl_record",41,"name")="jnl_record.jrec_aimg.ondsk_blkver"
	Set gtmtypes("jnl_record",41,"off")=32
	Set gtmtypes("jnl_record",41,"len")=4
	Set gtmtypes("jnl_record",41,"type")="int"
	Set gtmtypfldindx("jnl_record","jrec_aimg.ondsk_blkver")=41
	Set gtmtypes("jnl_record",42,"name")="jnl_record.jrec_aimg.filler"
	Set gtmtypes("jnl_record",42,"off")=36
	Set gtmtypes("jnl_record",42,"len")=4
	Set gtmtypes("jnl_record",42,"type")="int"
	Set gtmtypfldindx("jnl_record","jrec_aimg.filler")=42
	Set gtmtypes("jnl_record",43,"name")="jnl_record.jrec_aimg.blk_contents"
	Set gtmtypes("jnl_record",43,"off")=40
	Set gtmtypes("jnl_record",43,"len")=1
	Set gtmtypes("jnl_record",43,"type")="char"
	Set gtmtypfldindx("jnl_record","jrec_aimg.blk_contents")=43
	Set gtmtypes("jnl_record",44,"name")="jnl_record.jrec_align"
	Set gtmtypes("jnl_record",44,"off")=0
	Set gtmtypes("jnl_record",44,"len")=12
	Set gtmtypes("jnl_record",44,"type")="struct_jrec_align"
	Set gtmtypfldindx("jnl_record","jrec_align")=44
	Set gtmtypes("jnl_record",45,"name")="jnl_record.jrec_align.checksum"
	Set gtmtypes("jnl_record",45,"off")=4
	Set gtmtypes("jnl_record",45,"len")=4
	Set gtmtypes("jnl_record",45,"type")="unsigned-int"
	Set gtmtypfldindx("jnl_record","jrec_align.checksum")=45
	Set gtmtypes("jnl_record",46,"name")="jnl_record.jrec_align.time"
	Set gtmtypes("jnl_record",46,"off")=8
	Set gtmtypes("jnl_record",46,"len")=4
	Set gtmtypes("jnl_record",46,"type")="unsigned-int"
	Set gtmtypfldindx("jnl_record","jrec_align.time")=46
	Set gtmtypes("jnl_record",47,"name")="jnl_record.jrec_tcom"
	Set gtmtypes("jnl_record",47,"off")=0
	Set gtmtypes("jnl_record",47,"len")=56
	Set gtmtypes("jnl_record",47,"type")="struct_jrec_tcom"
	Set gtmtypfldindx("jnl_record","jrec_tcom")=47
	Set gtmtypes("jnl_record",48,"name")="jnl_record.jrec_tcom.prefix"
	Set gtmtypes("jnl_record",48,"off")=0
	Set gtmtypes("jnl_record",48,"len")=24
	Set gtmtypes("jnl_record",48,"type")="jrec_prefix"
	Set gtmtypfldindx("jnl_record","jrec_tcom.prefix")=48
	Set gtmtypes("jnl_record",49,"name")="jnl_record.jrec_tcom.token_seq"
	Set gtmtypes("jnl_record",49,"off")=24
	Set gtmtypes("jnl_record",49,"len")=8
	Set gtmtypes("jnl_record",49,"type")="token_seq_t"
	Set gtmtypfldindx("jnl_record","jrec_tcom.token_seq")=49
	Set gtmtypes("jnl_record",50,"name")="jnl_record.jrec_tcom.strm_seqno"
	Set gtmtypes("jnl_record",50,"off")=32
	Set gtmtypes("jnl_record",50,"len")=8
	Set gtmtypes("jnl_record",50,"type")="uint64_t"
	Set gtmtypfldindx("jnl_record","jrec_tcom.strm_seqno")=50
	Set gtmtypes("jnl_record",51,"name")="jnl_record.jrec_tcom.filler_short"
	Set gtmtypes("jnl_record",51,"off")=40
	Set gtmtypes("jnl_record",51,"len")=2
	Set gtmtypes("jnl_record",51,"type")="unsigned-short"
	Set gtmtypfldindx("jnl_record","jrec_tcom.filler_short")=51
	Set gtmtypes("jnl_record",52,"name")="jnl_record.jrec_tcom.num_participants"
	Set gtmtypes("jnl_record",52,"off")=42
	Set gtmtypes("jnl_record",52,"len")=2
	Set gtmtypes("jnl_record",52,"type")="unsigned-short"
	Set gtmtypfldindx("jnl_record","jrec_tcom.num_participants")=52
	Set gtmtypes("jnl_record",53,"name")="jnl_record.jrec_tcom.jnl_tid"
	Set gtmtypes("jnl_record",53,"off")=44
	Set gtmtypes("jnl_record",53,"len")=8
	Set gtmtypes("jnl_record",53,"type")="char"
	Set gtmtypfldindx("jnl_record","jrec_tcom.jnl_tid")=53
	Set gtmtypes("jnl_record",54,"name")="jnl_record.jrec_tcom.suffix"
	Set gtmtypes("jnl_record",54,"off")=52
	Set gtmtypes("jnl_record",54,"len")=4
	Set gtmtypes("jnl_record",54,"type")="jrec_suffix"
	Set gtmtypfldindx("jnl_record","jrec_tcom.suffix")=54
	Set gtmtypes("jnl_record",55,"name")="jnl_record.jrec_ztcom"
	Set gtmtypes("jnl_record",55,"off")=0
	Set gtmtypes("jnl_record",55,"len")=48
	Set gtmtypes("jnl_record",55,"type")="struct_jrec_ztcom"
	Set gtmtypfldindx("jnl_record","jrec_ztcom")=55
	Set gtmtypes("jnl_record",56,"name")="jnl_record.jrec_ztcom.prefix"
	Set gtmtypes("jnl_record",56,"off")=0
	Set gtmtypes("jnl_record",56,"len")=24
	Set gtmtypes("jnl_record",56,"type")="jrec_prefix"
	Set gtmtypfldindx("jnl_record","jrec_ztcom.prefix")=56
	Set gtmtypes("jnl_record",57,"name")="jnl_record.jrec_ztcom.token"
	Set gtmtypes("jnl_record",57,"off")=24
	Set gtmtypes("jnl_record",57,"len")=8
	Set gtmtypes("jnl_record",57,"type")="uint64_t"
	Set gtmtypfldindx("jnl_record","jrec_ztcom.token")=57
	Set gtmtypes("jnl_record",58,"name")="jnl_record.jrec_ztcom.filler_8bytes"
	Set gtmtypes("jnl_record",58,"off")=32
	Set gtmtypes("jnl_record",58,"len")=8
	Set gtmtypes("jnl_record",58,"type")="uint64_t"
	Set gtmtypfldindx("jnl_record","jrec_ztcom.filler_8bytes")=58
	Set gtmtypes("jnl_record",59,"name")="jnl_record.jrec_ztcom.filler_short"
	Set gtmtypes("jnl_record",59,"off")=40
	Set gtmtypes("jnl_record",59,"len")=2
	Set gtmtypes("jnl_record",59,"type")="unsigned-short"
	Set gtmtypfldindx("jnl_record","jrec_ztcom.filler_short")=59
	Set gtmtypes("jnl_record",60,"name")="jnl_record.jrec_ztcom.participants"
	Set gtmtypes("jnl_record",60,"off")=42
	Set gtmtypes("jnl_record",60,"len")=2
	Set gtmtypes("jnl_record",60,"type")="unsigned-short"
	Set gtmtypfldindx("jnl_record","jrec_ztcom.participants")=60
	Set gtmtypes("jnl_record",61,"name")="jnl_record.jrec_ztcom.suffix"
	Set gtmtypes("jnl_record",61,"off")=44
	Set gtmtypes("jnl_record",61,"len")=4
	Set gtmtypes("jnl_record",61,"type")="jrec_suffix"
	Set gtmtypfldindx("jnl_record","jrec_ztcom.suffix")=61
	Set gtmtypes("jnl_record",62,"name")="jnl_record.jrec_inctn"
	Set gtmtypes("jnl_record",62,"off")=0
	Set gtmtypes("jnl_record",62,"len")=40
	Set gtmtypes("jnl_record",62,"type")="struct_jrec_inctn"
	Set gtmtypfldindx("jnl_record","jrec_inctn")=62
	Set gtmtypes("jnl_record",63,"name")="jnl_record.jrec_inctn.prefix"
	Set gtmtypes("jnl_record",63,"off")=0
	Set gtmtypes("jnl_record",63,"len")=24
	Set gtmtypes("jnl_record",63,"type")="jrec_prefix"
	Set gtmtypfldindx("jnl_record","jrec_inctn.prefix")=63
	Set gtmtypes("jnl_record",64,"name")="jnl_record.jrec_inctn.detail"
	Set gtmtypes("jnl_record",64,"off")=24
	Set gtmtypes("jnl_record",64,"len")=16
	Set gtmtypes("jnl_record",64,"type")="inctn_detail_t"
	Set gtmtypfldindx("jnl_record","jrec_inctn.detail")=64
	Set gtmtypes("jnl_record",65,"name")="jnl_record.jrec_pini"
	Set gtmtypes("jnl_record",65,"off")=0
	Set gtmtypes("jnl_record",65,"len")=208
	Set gtmtypes("jnl_record",65,"type")="struct_jrec_pini"
	Set gtmtypfldindx("jnl_record","jrec_pini")=65
	Set gtmtypes("jnl_record",66,"name")="jnl_record.jrec_pini.prefix"
	Set gtmtypes("jnl_record",66,"off")=0
	Set gtmtypes("jnl_record",66,"len")=24
	Set gtmtypes("jnl_record",66,"type")="jrec_prefix"
	Set gtmtypfldindx("jnl_record","jrec_pini.prefix")=66
	Set gtmtypes("jnl_record",67,"name")="jnl_record.jrec_pini.process_vector"
	Set gtmtypes("jnl_record",67,"off")=24
	Set gtmtypes("jnl_record",67,"len")=176
	Set gtmtypes("jnl_record",67,"type")="jnl_process_vector"
	Set gtmtypfldindx("jnl_record","jrec_pini.process_vector")=67
	Set gtmtypes("jnl_record",67,"dim")=2
	Set gtmtypes("jnl_record",68,"name")="jnl_record.jrec_pini.filler"
	Set gtmtypes("jnl_record",68,"off")=200
	Set gtmtypes("jnl_record",68,"len")=4
	Set gtmtypes("jnl_record",68,"type")="int"
	Set gtmtypfldindx("jnl_record","jrec_pini.filler")=68
	Set gtmtypes("jnl_record",69,"name")="jnl_record.jrec_pini.suffix"
	Set gtmtypes("jnl_record",69,"off")=204
	Set gtmtypes("jnl_record",69,"len")=4
	Set gtmtypes("jnl_record",69,"type")="jrec_suffix"
	Set gtmtypfldindx("jnl_record","jrec_pini.suffix")=69
	Set gtmtypes("jnl_record",70,"name")="jnl_record.jrec_pfin"
	Set gtmtypes("jnl_record",70,"off")=0
	Set gtmtypes("jnl_record",70,"len")=32
	Set gtmtypes("jnl_record",70,"type")="struct_jrec_pfin"
	Set gtmtypfldindx("jnl_record","jrec_pfin")=70
	Set gtmtypes("jnl_record",71,"name")="jnl_record.jrec_pfin.prefix"
	Set gtmtypes("jnl_record",71,"off")=0
	Set gtmtypes("jnl_record",71,"len")=24
	Set gtmtypes("jnl_record",71,"type")="jrec_prefix"
	Set gtmtypfldindx("jnl_record","jrec_pfin.prefix")=71
	Set gtmtypes("jnl_record",72,"name")="jnl_record.jrec_pfin.filler"
	Set gtmtypes("jnl_record",72,"off")=24
	Set gtmtypes("jnl_record",72,"len")=4
	Set gtmtypes("jnl_record",72,"type")="unsigned-int"
	Set gtmtypfldindx("jnl_record","jrec_pfin.filler")=72
	Set gtmtypes("jnl_record",73,"name")="jnl_record.jrec_pfin.suffix"
	Set gtmtypes("jnl_record",73,"off")=28
	Set gtmtypes("jnl_record",73,"len")=4
	Set gtmtypes("jnl_record",73,"type")="jrec_suffix"
	Set gtmtypfldindx("jnl_record","jrec_pfin.suffix")=73
	Set gtmtypes("jnl_record",74,"name")="jnl_record.jrec_null"
	Set gtmtypes("jnl_record",74,"off")=0
	Set gtmtypes("jnl_record",74,"len")=48
	Set gtmtypes("jnl_record",74,"type")="struct_jrec_null"
	Set gtmtypfldindx("jnl_record","jrec_null")=74
	Set gtmtypes("jnl_record",75,"name")="jnl_record.jrec_null.prefix"
	Set gtmtypes("jnl_record",75,"off")=0
	Set gtmtypes("jnl_record",75,"len")=24
	Set gtmtypes("jnl_record",75,"type")="jrec_prefix"
	Set gtmtypfldindx("jnl_record","jrec_null.prefix")=75
	Set gtmtypes("jnl_record",76,"name")="jnl_record.jrec_null.jnl_seqno"
	Set gtmtypes("jnl_record",76,"off")=24
	Set gtmtypes("jnl_record",76,"len")=8
	Set gtmtypes("jnl_record",76,"type")="uint64_t"
	Set gtmtypfldindx("jnl_record","jrec_null.jnl_seqno")=76
	Set gtmtypes("jnl_record",77,"name")="jnl_record.jrec_null.strm_seqno"
	Set gtmtypes("jnl_record",77,"off")=32
	Set gtmtypes("jnl_record",77,"len")=8
	Set gtmtypes("jnl_record",77,"type")="uint64_t"
	Set gtmtypfldindx("jnl_record","jrec_null.strm_seqno")=77
	Set gtmtypes("jnl_record",78,"name")="jnl_record.jrec_null.filler"
	Set gtmtypes("jnl_record",78,"off")=40
	Set gtmtypes("jnl_record",78,"len")=4
	Set gtmtypes("jnl_record",78,"type")="unsigned-int"
	Set gtmtypfldindx("jnl_record","jrec_null.filler")=78
	Set gtmtypes("jnl_record",79,"name")="jnl_record.jrec_null.suffix"
	Set gtmtypes("jnl_record",79,"off")=44
	Set gtmtypes("jnl_record",79,"len")=4
	Set gtmtypes("jnl_record",79,"type")="jrec_suffix"
	Set gtmtypfldindx("jnl_record","jrec_null.suffix")=79
	Set gtmtypes("jnl_record",80,"name")="jnl_record.jrec_epoch"
	Set gtmtypes("jnl_record",80,"off")=0
	Set gtmtypes("jnl_record",80,"len")=184
	Set gtmtypes("jnl_record",80,"type")="struct_jrec_epoch"
	Set gtmtypfldindx("jnl_record","jrec_epoch")=80
	Set gtmtypes("jnl_record",81,"name")="jnl_record.jrec_epoch.prefix"
	Set gtmtypes("jnl_record",81,"off")=0
	Set gtmtypes("jnl_record",81,"len")=24
	Set gtmtypes("jnl_record",81,"type")="jrec_prefix"
	Set gtmtypfldindx("jnl_record","jrec_epoch.prefix")=81
	Set gtmtypes("jnl_record",82,"name")="jnl_record.jrec_epoch.jnl_seqno"
	Set gtmtypes("jnl_record",82,"off")=24
	Set gtmtypes("jnl_record",82,"len")=8
	Set gtmtypes("jnl_record",82,"type")="uint64_t"
	Set gtmtypfldindx("jnl_record","jrec_epoch.jnl_seqno")=82
	Set gtmtypes("jnl_record",83,"name")="jnl_record.jrec_epoch.blks_to_upgrd"
	Set gtmtypes("jnl_record",83,"off")=32
	Set gtmtypes("jnl_record",83,"len")=4
	Set gtmtypes("jnl_record",83,"type")="unsigned-int"
	Set gtmtypfldindx("jnl_record","jrec_epoch.blks_to_upgrd")=83
	Set gtmtypes("jnl_record",84,"name")="jnl_record.jrec_epoch.free_blocks"
	Set gtmtypes("jnl_record",84,"off")=36
	Set gtmtypes("jnl_record",84,"len")=4
	Set gtmtypes("jnl_record",84,"type")="unsigned-int"
	Set gtmtypfldindx("jnl_record","jrec_epoch.free_blocks")=84
	Set gtmtypes("jnl_record",85,"name")="jnl_record.jrec_epoch.total_blks"
	Set gtmtypes("jnl_record",85,"off")=40
	Set gtmtypes("jnl_record",85,"len")=4
	Set gtmtypes("jnl_record",85,"type")="unsigned-int"
	Set gtmtypfldindx("jnl_record","jrec_epoch.total_blks")=85
	Set gtmtypes("jnl_record",86,"name")="jnl_record.jrec_epoch.fully_upgraded"
	Set gtmtypes("jnl_record",86,"off")=44
	Set gtmtypes("jnl_record",86,"len")=4
	Set gtmtypes("jnl_record",86,"type")="boolean_t"
	Set gtmtypfldindx("jnl_record","jrec_epoch.fully_upgraded")=86
	Set gtmtypes("jnl_record",87,"name")="jnl_record.jrec_epoch.strm_seqno"
	Set gtmtypes("jnl_record",87,"off")=48
	Set gtmtypes("jnl_record",87,"len")=128
	Set gtmtypes("jnl_record",87,"type")="uint64_t"
	Set gtmtypfldindx("jnl_record","jrec_epoch.strm_seqno")=87
	Set gtmtypes("jnl_record",87,"dim")=16
	Set gtmtypes("jnl_record",88,"name")="jnl_record.jrec_epoch.filler"
	Set gtmtypes("jnl_record",88,"off")=176
	Set gtmtypes("jnl_record",88,"len")=4
	Set gtmtypes("jnl_record",88,"type")="unsigned-int"
	Set gtmtypfldindx("jnl_record","jrec_epoch.filler")=88
	Set gtmtypes("jnl_record",89,"name")="jnl_record.jrec_epoch.suffix"
	Set gtmtypes("jnl_record",89,"off")=180
	Set gtmtypes("jnl_record",89,"len")=4
	Set gtmtypes("jnl_record",89,"type")="jrec_suffix"
	Set gtmtypfldindx("jnl_record","jrec_epoch.suffix")=89
	Set gtmtypes("jnl_record",90,"name")="jnl_record.jrec_eof"
	Set gtmtypes("jnl_record",90,"off")=0
	Set gtmtypes("jnl_record",90,"len")=40
	Set gtmtypes("jnl_record",90,"type")="struct_jrec_eof"
	Set gtmtypfldindx("jnl_record","jrec_eof")=90
	Set gtmtypes("jnl_record",91,"name")="jnl_record.jrec_eof.prefix"
	Set gtmtypes("jnl_record",91,"off")=0
	Set gtmtypes("jnl_record",91,"len")=24
	Set gtmtypes("jnl_record",91,"type")="jrec_prefix"
	Set gtmtypfldindx("jnl_record","jrec_eof.prefix")=91
	Set gtmtypes("jnl_record",92,"name")="jnl_record.jrec_eof.jnl_seqno"
	Set gtmtypes("jnl_record",92,"off")=24
	Set gtmtypes("jnl_record",92,"len")=8
	Set gtmtypes("jnl_record",92,"type")="uint64_t"
	Set gtmtypfldindx("jnl_record","jrec_eof.jnl_seqno")=92
	Set gtmtypes("jnl_record",93,"name")="jnl_record.jrec_eof.filler"
	Set gtmtypes("jnl_record",93,"off")=32
	Set gtmtypes("jnl_record",93,"len")=4
	Set gtmtypes("jnl_record",93,"type")="unsigned-int"
	Set gtmtypfldindx("jnl_record","jrec_eof.filler")=93
	Set gtmtypes("jnl_record",94,"name")="jnl_record.jrec_eof.suffix"
	Set gtmtypes("jnl_record",94,"off")=36
	Set gtmtypes("jnl_record",94,"len")=4
	Set gtmtypes("jnl_record",94,"type")="jrec_suffix"
	Set gtmtypfldindx("jnl_record","jrec_eof.suffix")=94
	Set gtmtypes("jnl_record",95,"name")="jnl_record.jrec_trunc"
	Set gtmtypes("jnl_record",95,"off")=0
	Set gtmtypes("jnl_record",95,"len")=40
	Set gtmtypes("jnl_record",95,"type")="struct_jrec_trunc"
	Set gtmtypfldindx("jnl_record","jrec_trunc")=95
	Set gtmtypes("jnl_record",96,"name")="jnl_record.jrec_trunc.prefix"
	Set gtmtypes("jnl_record",96,"off")=0
	Set gtmtypes("jnl_record",96,"len")=24
	Set gtmtypes("jnl_record",96,"type")="jrec_prefix"
	Set gtmtypfldindx("jnl_record","jrec_trunc.prefix")=96
	Set gtmtypes("jnl_record",97,"name")="jnl_record.jrec_trunc.orig_total_blks"
	Set gtmtypes("jnl_record",97,"off")=24
	Set gtmtypes("jnl_record",97,"len")=4
	Set gtmtypes("jnl_record",97,"type")="unsigned-int"
	Set gtmtypfldindx("jnl_record","jrec_trunc.orig_total_blks")=97
	Set gtmtypes("jnl_record",98,"name")="jnl_record.jrec_trunc.orig_free_blocks"
	Set gtmtypes("jnl_record",98,"off")=28
	Set gtmtypes("jnl_record",98,"len")=4
	Set gtmtypes("jnl_record",98,"type")="unsigned-int"
	Set gtmtypfldindx("jnl_record","jrec_trunc.orig_free_blocks")=98
	Set gtmtypes("jnl_record",99,"name")="jnl_record.jrec_trunc.total_blks_after_trunc"
	Set gtmtypes("jnl_record",99,"off")=32
	Set gtmtypes("jnl_record",99,"len")=4
	Set gtmtypes("jnl_record",99,"type")="unsigned-int"
	Set gtmtypfldindx("jnl_record","jrec_trunc.total_blks_after_trunc")=99
	Set gtmtypes("jnl_record",100,"name")="jnl_record.jrec_trunc.suffix"
	Set gtmtypes("jnl_record",100,"off")=36
	Set gtmtypes("jnl_record",100,"len")=4
	Set gtmtypes("jnl_record",100,"type")="jrec_suffix"
	Set gtmtypfldindx("jnl_record","jrec_trunc.suffix")=100
	;
	Set gtmtypes("jnl_string")="struct"
	Set gtmtypes("jnl_string",0)=1
	Set gtmtypes("jnl_string","len")=8
	Set gtmtypes("jnl_string",1,"name")="jnl_string.text"
	Set gtmtypes("jnl_string",1,"off")=4
	Set gtmtypes("jnl_string",1,"len")=1
	Set gtmtypes("jnl_string",1,"type")="char"
	Set gtmtypfldindx("jnl_string","text")=1
	;
	Set gtmtypes("jnldata_hdr_struct")="struct"
	Set gtmtypes("jnldata_hdr_struct",0)=2
	Set gtmtypes("jnldata_hdr_struct","len")=8
	Set gtmtypes("jnldata_hdr_struct",1,"name")="jnldata_hdr_struct.jnldata_len"
	Set gtmtypes("jnldata_hdr_struct",1,"off")=0
	Set gtmtypes("jnldata_hdr_struct",1,"len")=4
	Set gtmtypes("jnldata_hdr_struct",1,"type")="unsigned-int"
	Set gtmtypfldindx("jnldata_hdr_struct","jnldata_len")=1
	Set gtmtypes("jnldata_hdr_struct",2,"name")="jnldata_hdr_struct.prev_jnldata_len"
	Set gtmtypes("jnldata_hdr_struct",2,"off")=4
	Set gtmtypes("jnldata_hdr_struct",2,"len")=4
	Set gtmtypes("jnldata_hdr_struct",2,"type")="unsigned-int"
	Set gtmtypfldindx("jnldata_hdr_struct","prev_jnldata_len")=2
	;
	Set gtmtypes("jnlext_multi_t")="struct"
	Set gtmtypes("jnlext_multi_t",0)=7
	Set gtmtypes("jnlext_multi_t","len")=32
	Set gtmtypes("jnlext_multi_t",1,"name")="jnlext_multi_t.time"
	Set gtmtypes("jnlext_multi_t",1,"off")=0
	Set gtmtypes("jnlext_multi_t",1,"len")=4
	Set gtmtypes("jnlext_multi_t",1,"type")="unsigned-int"
	Set gtmtypfldindx("jnlext_multi_t","time")=1
	Set gtmtypes("jnlext_multi_t",2,"name")="jnlext_multi_t.token_seq"
	Set gtmtypes("jnlext_multi_t",2,"off")=8
	Set gtmtypes("jnlext_multi_t",2,"len")=8
	Set gtmtypes("jnlext_multi_t",2,"type")="token_seq_t"
	Set gtmtypfldindx("jnlext_multi_t","token_seq")=2
	Set gtmtypes("jnlext_multi_t",3,"name")="jnlext_multi_t.token_seq.jnl_seqno"
	Set gtmtypes("jnlext_multi_t",3,"off")=8
	Set gtmtypes("jnlext_multi_t",3,"len")=8
	Set gtmtypes("jnlext_multi_t",3,"type")="uint64_t"
	Set gtmtypfldindx("jnlext_multi_t","token_seq.jnl_seqno")=3
	Set gtmtypes("jnlext_multi_t",4,"name")="jnlext_multi_t.token_seq.token"
	Set gtmtypes("jnlext_multi_t",4,"off")=8
	Set gtmtypes("jnlext_multi_t",4,"len")=8
	Set gtmtypes("jnlext_multi_t",4,"type")="uint64_t"
	Set gtmtypfldindx("jnlext_multi_t","token_seq.token")=4
	Set gtmtypes("jnlext_multi_t",5,"name")="jnlext_multi_t.update_num"
	Set gtmtypes("jnlext_multi_t",5,"off")=16
	Set gtmtypes("jnlext_multi_t",5,"len")=4
	Set gtmtypes("jnlext_multi_t",5,"type")="unsigned-int"
	Set gtmtypfldindx("jnlext_multi_t","update_num")=5
	Set gtmtypes("jnlext_multi_t",6,"name")="jnlext_multi_t.num_more_reg"
	Set gtmtypes("jnlext_multi_t",6,"off")=20
	Set gtmtypes("jnlext_multi_t",6,"len")=4
	Set gtmtypes("jnlext_multi_t",6,"type")="unsigned-int"
	Set gtmtypfldindx("jnlext_multi_t","num_more_reg")=6
	Set gtmtypes("jnlext_multi_t",7,"name")="jnlext_multi_t.size"
	Set gtmtypes("jnlext_multi_t",7,"off")=24
	Set gtmtypes("jnlext_multi_t",7,"len")=8
	Set gtmtypes("jnlext_multi_t",7,"type")="size_t"
	Set gtmtypfldindx("jnlext_multi_t","size")=7
	;
	Set gtmtypes("jnlpool_addrs")="struct"
	Set gtmtypes("jnlpool_addrs",0)=22
	Set gtmtypes("jnlpool_addrs","len")=136
	Set gtmtypes("jnlpool_addrs",1,"name")="jnlpool_addrs.jnlpool_ctl"
	Set gtmtypes("jnlpool_addrs",1,"off")=0
	Set gtmtypes("jnlpool_addrs",1,"len")=8
	Set gtmtypes("jnlpool_addrs",1,"type")="addr"
	Set gtmtypfldindx("jnlpool_addrs","jnlpool_ctl")=1
	Set gtmtypes("jnlpool_addrs",2,"name")="jnlpool_addrs.jnlpool_dummy_reg"
	Set gtmtypes("jnlpool_addrs",2,"off")=8
	Set gtmtypes("jnlpool_addrs",2,"len")=8
	Set gtmtypes("jnlpool_addrs",2,"type")="addr"
	Set gtmtypfldindx("jnlpool_addrs","jnlpool_dummy_reg")=2
	Set gtmtypes("jnlpool_addrs",3,"name")="jnlpool_addrs.gtmsource_local"
	Set gtmtypes("jnlpool_addrs",3,"off")=16
	Set gtmtypes("jnlpool_addrs",3,"len")=8
	Set gtmtypes("jnlpool_addrs",3,"type")="addr"
	Set gtmtypfldindx("jnlpool_addrs","gtmsource_local")=3
	Set gtmtypes("jnlpool_addrs",4,"name")="jnlpool_addrs.gtmsource_local_array"
	Set gtmtypes("jnlpool_addrs",4,"off")=24
	Set gtmtypes("jnlpool_addrs",4,"len")=8
	Set gtmtypes("jnlpool_addrs",4,"type")="addr"
	Set gtmtypfldindx("jnlpool_addrs","gtmsource_local_array")=4
	Set gtmtypes("jnlpool_addrs",5,"name")="jnlpool_addrs.repl_inst_filehdr"
	Set gtmtypes("jnlpool_addrs",5,"off")=32
	Set gtmtypes("jnlpool_addrs",5,"len")=8
	Set gtmtypes("jnlpool_addrs",5,"type")="addr"
	Set gtmtypfldindx("jnlpool_addrs","repl_inst_filehdr")=5
	Set gtmtypes("jnlpool_addrs",6,"name")="jnlpool_addrs.gtmsrc_lcl_array"
	Set gtmtypes("jnlpool_addrs",6,"off")=40
	Set gtmtypes("jnlpool_addrs",6,"len")=8
	Set gtmtypes("jnlpool_addrs",6,"type")="addr"
	Set gtmtypfldindx("jnlpool_addrs","gtmsrc_lcl_array")=6
	Set gtmtypes("jnlpool_addrs",7,"name")="jnlpool_addrs.jnldata_base"
	Set gtmtypes("jnlpool_addrs",7,"off")=48
	Set gtmtypes("jnlpool_addrs",7,"len")=8
	Set gtmtypes("jnlpool_addrs",7,"type")="addr"
	Set gtmtypfldindx("jnlpool_addrs","jnldata_base")=7
	Set gtmtypes("jnlpool_addrs",8,"name")="jnlpool_addrs.jrs"
	Set gtmtypes("jnlpool_addrs",8,"off")=56
	Set gtmtypes("jnlpool_addrs",8,"len")=40
	Set gtmtypes("jnlpool_addrs",8,"type")="jpl_rsrv_struct_t"
	Set gtmtypfldindx("jnlpool_addrs","jrs")=8
	Set gtmtypes("jnlpool_addrs",9,"name")="jnlpool_addrs.jrs.start_write_addr"
	Set gtmtypes("jnlpool_addrs",9,"off")=56
	Set gtmtypes("jnlpool_addrs",9,"len")=8
	Set gtmtypes("jnlpool_addrs",9,"type")="uint64_t"
	Set gtmtypfldindx("jnlpool_addrs","jrs.start_write_addr")=9
	Set gtmtypes("jnlpool_addrs",10,"name")="jnlpool_addrs.jrs.cur_write_addr"
	Set gtmtypes("jnlpool_addrs",10,"off")=64
	Set gtmtypes("jnlpool_addrs",10,"len")=8
	Set gtmtypes("jnlpool_addrs",10,"type")="uint64_t"
	Set gtmtypfldindx("jnlpool_addrs","jrs.cur_write_addr")=10
	Set gtmtypes("jnlpool_addrs",11,"name")="jnlpool_addrs.jrs.tot_jrec_len"
	Set gtmtypes("jnlpool_addrs",11,"off")=72
	Set gtmtypes("jnlpool_addrs",11,"len")=4
	Set gtmtypes("jnlpool_addrs",11,"type")="unsigned-int"
	Set gtmtypfldindx("jnlpool_addrs","jrs.tot_jrec_len")=11
	Set gtmtypes("jnlpool_addrs",12,"name")="jnlpool_addrs.jrs.write_total"
	Set gtmtypes("jnlpool_addrs",12,"off")=76
	Set gtmtypes("jnlpool_addrs",12,"len")=4
	Set gtmtypes("jnlpool_addrs",12,"type")="unsigned-int"
	Set gtmtypfldindx("jnlpool_addrs","jrs.write_total")=12
	Set gtmtypes("jnlpool_addrs",13,"name")="jnlpool_addrs.jrs.memcpy_skipped"
	Set gtmtypes("jnlpool_addrs",13,"off")=80
	Set gtmtypes("jnlpool_addrs",13,"len")=4
	Set gtmtypes("jnlpool_addrs",13,"type")="boolean_t"
	Set gtmtypfldindx("jnlpool_addrs","jrs.memcpy_skipped")=13
	Set gtmtypes("jnlpool_addrs",14,"name")="jnlpool_addrs.jrs.phase2_commit_index"
	Set gtmtypes("jnlpool_addrs",14,"off")=84
	Set gtmtypes("jnlpool_addrs",14,"len")=4
	Set gtmtypes("jnlpool_addrs",14,"type")="unsigned-int"
	Set gtmtypfldindx("jnlpool_addrs","jrs.phase2_commit_index")=14
	Set gtmtypes("jnlpool_addrs",15,"name")="jnlpool_addrs.jrs.num_tcoms"
	Set gtmtypes("jnlpool_addrs",15,"off")=88
	Set gtmtypes("jnlpool_addrs",15,"len")=4
	Set gtmtypes("jnlpool_addrs",15,"type")="unsigned-int"
	Set gtmtypfldindx("jnlpool_addrs","jrs.num_tcoms")=15
	Set gtmtypes("jnlpool_addrs",16,"name")="jnlpool_addrs.jrs.filler_8byte_align"
	Set gtmtypes("jnlpool_addrs",16,"off")=92
	Set gtmtypes("jnlpool_addrs",16,"len")=4
	Set gtmtypes("jnlpool_addrs",16,"type")="char"
	Set gtmtypfldindx("jnlpool_addrs","jrs.filler_8byte_align")=16
	Set gtmtypes("jnlpool_addrs",17,"name")="jnlpool_addrs.pool_init"
	Set gtmtypes("jnlpool_addrs",17,"off")=96
	Set gtmtypes("jnlpool_addrs",17,"len")=4
	Set gtmtypes("jnlpool_addrs",17,"type")="boolean_t"
	Set gtmtypfldindx("jnlpool_addrs","pool_init")=17
	Set gtmtypes("jnlpool_addrs",18,"name")="jnlpool_addrs.recv_pool"
	Set gtmtypes("jnlpool_addrs",18,"off")=100
	Set gtmtypes("jnlpool_addrs",18,"len")=4
	Set gtmtypes("jnlpool_addrs",18,"type")="boolean_t"
	Set gtmtypfldindx("jnlpool_addrs","recv_pool")=18
	Set gtmtypes("jnlpool_addrs",19,"name")="jnlpool_addrs.relaxed"
	Set gtmtypes("jnlpool_addrs",19,"off")=104
	Set gtmtypes("jnlpool_addrs",19,"len")=4
	Set gtmtypes("jnlpool_addrs",19,"type")="boolean_t"
	Set gtmtypfldindx("jnlpool_addrs","relaxed")=19
	Set gtmtypes("jnlpool_addrs",20,"name")="jnlpool_addrs.next"
	Set gtmtypes("jnlpool_addrs",20,"off")=112
	Set gtmtypes("jnlpool_addrs",20,"len")=8
	Set gtmtypes("jnlpool_addrs",20,"type")="addr"
	Set gtmtypfldindx("jnlpool_addrs","next")=20
	Set gtmtypes("jnlpool_addrs",21,"name")="jnlpool_addrs.gd_instinfo"
	Set gtmtypes("jnlpool_addrs",21,"off")=120
	Set gtmtypes("jnlpool_addrs",21,"len")=8
	Set gtmtypes("jnlpool_addrs",21,"type")="addr"
	Set gtmtypfldindx("jnlpool_addrs","gd_instinfo")=21
	Set gtmtypes("jnlpool_addrs",22,"name")="jnlpool_addrs.gd_ptr"
	Set gtmtypes("jnlpool_addrs",22,"off")=128
	Set gtmtypes("jnlpool_addrs",22,"len")=8
	Set gtmtypes("jnlpool_addrs",22,"type")="addr"
	Set gtmtypfldindx("jnlpool_addrs","gd_ptr")=22
	;
	Set gtmtypes("jnlpool_ctl_struct")="struct"
	Set gtmtypes("jnlpool_ctl_struct",0)=72
	Set gtmtypes("jnlpool_ctl_struct","len")=659216
	Set gtmtypes("jnlpool_ctl_struct",1,"name")="jnlpool_ctl_struct.jnlpool_id"
	Set gtmtypes("jnlpool_ctl_struct",1,"off")=0
	Set gtmtypes("jnlpool_ctl_struct",1,"len")=312
	Set gtmtypes("jnlpool_ctl_struct",1,"type")="replpool_identifier"
	Set gtmtypfldindx("jnlpool_ctl_struct","jnlpool_id")=1
	Set gtmtypes("jnlpool_ctl_struct",2,"name")="jnlpool_ctl_struct.jnlpool_id.label"
	Set gtmtypes("jnlpool_ctl_struct",2,"off")=0
	Set gtmtypes("jnlpool_ctl_struct",2,"len")=12
	Set gtmtypes("jnlpool_ctl_struct",2,"type")="unsigned-char"
	Set gtmtypfldindx("jnlpool_ctl_struct","jnlpool_id.label")=2
	Set gtmtypes("jnlpool_ctl_struct",3,"name")="jnlpool_ctl_struct.jnlpool_id.pool_type"
	Set gtmtypes("jnlpool_ctl_struct",3,"off")=12
	Set gtmtypes("jnlpool_ctl_struct",3,"len")=1
	Set gtmtypes("jnlpool_ctl_struct",3,"type")="char"
	Set gtmtypfldindx("jnlpool_ctl_struct","jnlpool_id.pool_type")=3
	Set gtmtypes("jnlpool_ctl_struct",4,"name")="jnlpool_ctl_struct.jnlpool_id.now_running"
	Set gtmtypes("jnlpool_ctl_struct",4,"off")=13
	Set gtmtypes("jnlpool_ctl_struct",4,"len")=36
	Set gtmtypes("jnlpool_ctl_struct",4,"type")="char"
	Set gtmtypfldindx("jnlpool_ctl_struct","jnlpool_id.now_running")=4
	Set gtmtypes("jnlpool_ctl_struct",5,"name")="jnlpool_ctl_struct.jnlpool_id.repl_pool_key_filler"
	Set gtmtypes("jnlpool_ctl_struct",5,"off")=52
	Set gtmtypes("jnlpool_ctl_struct",5,"len")=4
	Set gtmtypes("jnlpool_ctl_struct",5,"type")="int"
	Set gtmtypfldindx("jnlpool_ctl_struct","jnlpool_id.repl_pool_key_filler")=5
	Set gtmtypes("jnlpool_ctl_struct",6,"name")="jnlpool_ctl_struct.jnlpool_id.instfilename"
	Set gtmtypes("jnlpool_ctl_struct",6,"off")=56
	Set gtmtypes("jnlpool_ctl_struct",6,"len")=256
	Set gtmtypes("jnlpool_ctl_struct",6,"type")="char"
	Set gtmtypfldindx("jnlpool_ctl_struct","jnlpool_id.instfilename")=6
	Set gtmtypes("jnlpool_ctl_struct",7,"name")="jnlpool_ctl_struct.critical_off"
	Set gtmtypes("jnlpool_ctl_struct",7,"off")=312
	Set gtmtypes("jnlpool_ctl_struct",7,"len")=8
	Set gtmtypes("jnlpool_ctl_struct",7,"type")="intptr_t"
	Set gtmtypfldindx("jnlpool_ctl_struct","critical_off")=7
	Set gtmtypes("jnlpool_ctl_struct",8,"name")="jnlpool_ctl_struct.filehdr_off"
	Set gtmtypes("jnlpool_ctl_struct",8,"off")=320
	Set gtmtypes("jnlpool_ctl_struct",8,"len")=8
	Set gtmtypes("jnlpool_ctl_struct",8,"type")="intptr_t"
	Set gtmtypfldindx("jnlpool_ctl_struct","filehdr_off")=8
	Set gtmtypes("jnlpool_ctl_struct",9,"name")="jnlpool_ctl_struct.srclcl_array_off"
	Set gtmtypes("jnlpool_ctl_struct",9,"off")=328
	Set gtmtypes("jnlpool_ctl_struct",9,"len")=8
	Set gtmtypes("jnlpool_ctl_struct",9,"type")="intptr_t"
	Set gtmtypfldindx("jnlpool_ctl_struct","srclcl_array_off")=9
	Set gtmtypes("jnlpool_ctl_struct",10,"name")="jnlpool_ctl_struct.sourcelocal_array_off"
	Set gtmtypes("jnlpool_ctl_struct",10,"off")=336
	Set gtmtypes("jnlpool_ctl_struct",10,"len")=8
	Set gtmtypes("jnlpool_ctl_struct",10,"type")="intptr_t"
	Set gtmtypfldindx("jnlpool_ctl_struct","sourcelocal_array_off")=10
	Set gtmtypes("jnlpool_ctl_struct",11,"name")="jnlpool_ctl_struct.start_jnl_seqno"
	Set gtmtypes("jnlpool_ctl_struct",11,"off")=344
	Set gtmtypes("jnlpool_ctl_struct",11,"len")=8
	Set gtmtypes("jnlpool_ctl_struct",11,"type")="uint64_t"
	Set gtmtypfldindx("jnlpool_ctl_struct","start_jnl_seqno")=11
	Set gtmtypes("jnlpool_ctl_struct",12,"name")="jnlpool_ctl_struct.jnldata_base_off"
	Set gtmtypes("jnlpool_ctl_struct",12,"off")=352
	Set gtmtypes("jnlpool_ctl_struct",12,"len")=8
	Set gtmtypes("jnlpool_ctl_struct",12,"type")="uint64_t"
	Set gtmtypfldindx("jnlpool_ctl_struct","jnldata_base_off")=12
	Set gtmtypes("jnlpool_ctl_struct",13,"name")="jnlpool_ctl_struct.jnlpool_size"
	Set gtmtypes("jnlpool_ctl_struct",13,"off")=360
	Set gtmtypes("jnlpool_ctl_struct",13,"len")=8
	Set gtmtypes("jnlpool_ctl_struct",13,"type")="uint64_t"
	Set gtmtypfldindx("jnlpool_ctl_struct","jnlpool_size")=13
	Set gtmtypes("jnlpool_ctl_struct",14,"name")="jnlpool_ctl_struct.jnl_seqno"
	Set gtmtypes("jnlpool_ctl_struct",14,"off")=368
	Set gtmtypes("jnlpool_ctl_struct",14,"len")=8
	Set gtmtypes("jnlpool_ctl_struct",14,"type")="uint64_t"
	Set gtmtypfldindx("jnlpool_ctl_struct","jnl_seqno")=14
	Set gtmtypes("jnlpool_ctl_struct",15,"name")="jnlpool_ctl_struct.last_histinfo_seqno"
	Set gtmtypes("jnlpool_ctl_struct",15,"off")=376
	Set gtmtypes("jnlpool_ctl_struct",15,"len")=8
	Set gtmtypes("jnlpool_ctl_struct",15,"type")="uint64_t"
	Set gtmtypfldindx("jnlpool_ctl_struct","last_histinfo_seqno")=15
	Set gtmtypes("jnlpool_ctl_struct",16,"name")="jnlpool_ctl_struct.max_zqgblmod_seqno"
	Set gtmtypes("jnlpool_ctl_struct",16,"off")=384
	Set gtmtypes("jnlpool_ctl_struct",16,"len")=8
	Set gtmtypes("jnlpool_ctl_struct",16,"type")="uint64_t"
	Set gtmtypfldindx("jnlpool_ctl_struct","max_zqgblmod_seqno")=16
	Set gtmtypes("jnlpool_ctl_struct",17,"name")="jnlpool_ctl_struct.strm_seqno"
	Set gtmtypes("jnlpool_ctl_struct",17,"off")=392
	Set gtmtypes("jnlpool_ctl_struct",17,"len")=128
	Set gtmtypes("jnlpool_ctl_struct",17,"type")="uint64_t"
	Set gtmtypfldindx("jnlpool_ctl_struct","strm_seqno")=17
	Set gtmtypes("jnlpool_ctl_struct",17,"dim")=16
	Set gtmtypes("jnlpool_ctl_struct",18,"name")="jnlpool_ctl_struct.this_side"
	Set gtmtypes("jnlpool_ctl_struct",18,"off")=520
	Set gtmtypes("jnlpool_ctl_struct",18,"len")=48
	Set gtmtypes("jnlpool_ctl_struct",18,"type")="repl_conn_info_t"
	Set gtmtypfldindx("jnlpool_ctl_struct","this_side")=18
	Set gtmtypes("jnlpool_ctl_struct",19,"name")="jnlpool_ctl_struct.this_side.proto_ver"
	Set gtmtypes("jnlpool_ctl_struct",19,"off")=520
	Set gtmtypes("jnlpool_ctl_struct",19,"len")=4
	Set gtmtypes("jnlpool_ctl_struct",19,"type")="int"
	Set gtmtypfldindx("jnlpool_ctl_struct","this_side.proto_ver")=19
	Set gtmtypes("jnlpool_ctl_struct",20,"name")="jnlpool_ctl_struct.this_side.jnl_ver"
	Set gtmtypes("jnlpool_ctl_struct",20,"off")=524
	Set gtmtypes("jnlpool_ctl_struct",20,"len")=4
	Set gtmtypes("jnlpool_ctl_struct",20,"type")="unsigned-int"
	Set gtmtypfldindx("jnlpool_ctl_struct","this_side.jnl_ver")=20
	Set gtmtypes("jnlpool_ctl_struct",21,"name")="jnlpool_ctl_struct.this_side.is_std_null_coll"
	Set gtmtypes("jnlpool_ctl_struct",21,"off")=528
	Set gtmtypes("jnlpool_ctl_struct",21,"len")=4
	Set gtmtypes("jnlpool_ctl_struct",21,"type")="boolean_t"
	Set gtmtypfldindx("jnlpool_ctl_struct","this_side.is_std_null_coll")=21
	Set gtmtypes("jnlpool_ctl_struct",22,"name")="jnlpool_ctl_struct.this_side.trigger_supported"
	Set gtmtypes("jnlpool_ctl_struct",22,"off")=532
	Set gtmtypes("jnlpool_ctl_struct",22,"len")=4
	Set gtmtypes("jnlpool_ctl_struct",22,"type")="boolean_t"
	Set gtmtypfldindx("jnlpool_ctl_struct","this_side.trigger_supported")=22
	Set gtmtypes("jnlpool_ctl_struct",23,"name")="jnlpool_ctl_struct.this_side.cross_endian"
	Set gtmtypes("jnlpool_ctl_struct",23,"off")=536
	Set gtmtypes("jnlpool_ctl_struct",23,"len")=4
	Set gtmtypes("jnlpool_ctl_struct",23,"type")="boolean_t"
	Set gtmtypfldindx("jnlpool_ctl_struct","this_side.cross_endian")=23
	Set gtmtypes("jnlpool_ctl_struct",24,"name")="jnlpool_ctl_struct.this_side.endianness_known"
	Set gtmtypes("jnlpool_ctl_struct",24,"off")=540
	Set gtmtypes("jnlpool_ctl_struct",24,"len")=4
	Set gtmtypes("jnlpool_ctl_struct",24,"type")="boolean_t"
	Set gtmtypfldindx("jnlpool_ctl_struct","this_side.endianness_known")=24
	Set gtmtypes("jnlpool_ctl_struct",25,"name")="jnlpool_ctl_struct.this_side.null_subs_xform"
	Set gtmtypes("jnlpool_ctl_struct",25,"off")=544
	Set gtmtypes("jnlpool_ctl_struct",25,"len")=4
	Set gtmtypes("jnlpool_ctl_struct",25,"type")="boolean_t"
	Set gtmtypfldindx("jnlpool_ctl_struct","this_side.null_subs_xform")=25
	Set gtmtypes("jnlpool_ctl_struct",26,"name")="jnlpool_ctl_struct.this_side.is_supplementary"
	Set gtmtypes("jnlpool_ctl_struct",26,"off")=548
	Set gtmtypes("jnlpool_ctl_struct",26,"len")=4
	Set gtmtypes("jnlpool_ctl_struct",26,"type")="boolean_t"
	Set gtmtypfldindx("jnlpool_ctl_struct","this_side.is_supplementary")=26
	Set gtmtypes("jnlpool_ctl_struct",27,"name")="jnlpool_ctl_struct.this_side.tls_requested"
	Set gtmtypes("jnlpool_ctl_struct",27,"off")=552
	Set gtmtypes("jnlpool_ctl_struct",27,"len")=4
	Set gtmtypes("jnlpool_ctl_struct",27,"type")="boolean_t"
	Set gtmtypfldindx("jnlpool_ctl_struct","this_side.tls_requested")=27
	Set gtmtypes("jnlpool_ctl_struct",28,"name")="jnlpool_ctl_struct.this_side.filler_16"
	Set gtmtypes("jnlpool_ctl_struct",28,"off")=556
	Set gtmtypes("jnlpool_ctl_struct",28,"len")=12
	Set gtmtypes("jnlpool_ctl_struct",28,"type")="char"
	Set gtmtypfldindx("jnlpool_ctl_struct","this_side.filler_16")=28
	Set gtmtypes("jnlpool_ctl_struct",29,"name")="jnlpool_ctl_struct.write_addr"
	Set gtmtypes("jnlpool_ctl_struct",29,"off")=568
	Set gtmtypes("jnlpool_ctl_struct",29,"len")=8
	Set gtmtypes("jnlpool_ctl_struct",29,"type")="uint64_t"
	Set gtmtypfldindx("jnlpool_ctl_struct","write_addr")=29
	Set gtmtypes("jnlpool_ctl_struct",30,"name")="jnlpool_ctl_struct.rsrv_write_addr"
	Set gtmtypes("jnlpool_ctl_struct",30,"off")=576
	Set gtmtypes("jnlpool_ctl_struct",30,"len")=8
	Set gtmtypes("jnlpool_ctl_struct",30,"type")="uint64_t"
	Set gtmtypfldindx("jnlpool_ctl_struct","rsrv_write_addr")=30
	Set gtmtypes("jnlpool_ctl_struct",31,"name")="jnlpool_ctl_struct.upd_disabled"
	Set gtmtypes("jnlpool_ctl_struct",31,"off")=584
	Set gtmtypes("jnlpool_ctl_struct",31,"len")=4
	Set gtmtypes("jnlpool_ctl_struct",31,"type")="boolean_t"
	Set gtmtypfldindx("jnlpool_ctl_struct","upd_disabled")=31
	Set gtmtypes("jnlpool_ctl_struct",32,"name")="jnlpool_ctl_struct.lastwrite_len"
	Set gtmtypes("jnlpool_ctl_struct",32,"off")=588
	Set gtmtypes("jnlpool_ctl_struct",32,"len")=4
	Set gtmtypes("jnlpool_ctl_struct",32,"type")="unsigned-int"
	Set gtmtypfldindx("jnlpool_ctl_struct","lastwrite_len")=32
	Set gtmtypes("jnlpool_ctl_struct",33,"name")="jnlpool_ctl_struct.send_losttn_complete"
	Set gtmtypes("jnlpool_ctl_struct",33,"off")=592
	Set gtmtypes("jnlpool_ctl_struct",33,"len")=4
	Set gtmtypes("jnlpool_ctl_struct",33,"type")="boolean_t"
	Set gtmtypfldindx("jnlpool_ctl_struct","send_losttn_complete")=33
	Set gtmtypes("jnlpool_ctl_struct",34,"name")="jnlpool_ctl_struct.primary_instname"
	Set gtmtypes("jnlpool_ctl_struct",34,"off")=596
	Set gtmtypes("jnlpool_ctl_struct",34,"len")=16
	Set gtmtypes("jnlpool_ctl_struct",34,"type")="unsigned-char"
	Set gtmtypfldindx("jnlpool_ctl_struct","primary_instname")=34
	Set gtmtypes("jnlpool_ctl_struct",35,"name")="jnlpool_ctl_struct.gtmrecv_pid"
	Set gtmtypes("jnlpool_ctl_struct",35,"off")=612
	Set gtmtypes("jnlpool_ctl_struct",35,"len")=4
	Set gtmtypes("jnlpool_ctl_struct",35,"type")="unsigned-int"
	Set gtmtypfldindx("jnlpool_ctl_struct","gtmrecv_pid")=35
	Set gtmtypes("jnlpool_ctl_struct",36,"name")="jnlpool_ctl_struct.prev_jnlseqno_time"
	Set gtmtypes("jnlpool_ctl_struct",36,"off")=616
	Set gtmtypes("jnlpool_ctl_struct",36,"len")=4
	Set gtmtypes("jnlpool_ctl_struct",36,"type")="unsigned-int"
	Set gtmtypfldindx("jnlpool_ctl_struct","prev_jnlseqno_time")=36
	Set gtmtypes("jnlpool_ctl_struct",37,"name")="jnlpool_ctl_struct.pool_initialized"
	Set gtmtypes("jnlpool_ctl_struct",37,"off")=620
	Set gtmtypes("jnlpool_ctl_struct",37,"len")=4
	Set gtmtypes("jnlpool_ctl_struct",37,"type")="boolean_t"
	Set gtmtypfldindx("jnlpool_ctl_struct","pool_initialized")=37
	Set gtmtypes("jnlpool_ctl_struct",38,"name")="jnlpool_ctl_struct.jnlpool_creator_pid"
	Set gtmtypes("jnlpool_ctl_struct",38,"off")=624
	Set gtmtypes("jnlpool_ctl_struct",38,"len")=4
	Set gtmtypes("jnlpool_ctl_struct",38,"type")="unsigned-int"
	Set gtmtypfldindx("jnlpool_ctl_struct","jnlpool_creator_pid")=38
	Set gtmtypes("jnlpool_ctl_struct",39,"name")="jnlpool_ctl_struct.onln_rlbk_pid"
	Set gtmtypes("jnlpool_ctl_struct",39,"off")=628
	Set gtmtypes("jnlpool_ctl_struct",39,"len")=4
	Set gtmtypes("jnlpool_ctl_struct",39,"type")="unsigned-int"
	Set gtmtypfldindx("jnlpool_ctl_struct","onln_rlbk_pid")=39
	Set gtmtypes("jnlpool_ctl_struct",40,"name")="jnlpool_ctl_struct.onln_rlbk_cycle"
	Set gtmtypes("jnlpool_ctl_struct",40,"off")=632
	Set gtmtypes("jnlpool_ctl_struct",40,"len")=4
	Set gtmtypes("jnlpool_ctl_struct",40,"type")="unsigned-int"
	Set gtmtypfldindx("jnlpool_ctl_struct","onln_rlbk_cycle")=40
	Set gtmtypes("jnlpool_ctl_struct",41,"name")="jnlpool_ctl_struct.freeze"
	Set gtmtypes("jnlpool_ctl_struct",41,"off")=636
	Set gtmtypes("jnlpool_ctl_struct",41,"len")=4
	Set gtmtypes("jnlpool_ctl_struct",41,"type")="boolean_t"
	Set gtmtypfldindx("jnlpool_ctl_struct","freeze")=41
	Set gtmtypes("jnlpool_ctl_struct",42,"name")="jnlpool_ctl_struct.freeze_comment"
	Set gtmtypes("jnlpool_ctl_struct",42,"off")=640
	Set gtmtypes("jnlpool_ctl_struct",42,"len")=1024
	Set gtmtypes("jnlpool_ctl_struct",42,"type")="char"
	Set gtmtypfldindx("jnlpool_ctl_struct","freeze_comment")=42
	Set gtmtypes("jnlpool_ctl_struct",43,"name")="jnlpool_ctl_struct.instfreeze_environ_inited"
	Set gtmtypes("jnlpool_ctl_struct",43,"off")=1664
	Set gtmtypes("jnlpool_ctl_struct",43,"len")=4
	Set gtmtypes("jnlpool_ctl_struct",43,"type")="boolean_t"
	Set gtmtypfldindx("jnlpool_ctl_struct","instfreeze_environ_inited")=43
	Set gtmtypes("jnlpool_ctl_struct",44,"name")="jnlpool_ctl_struct.merrors_array"
	Set gtmtypes("jnlpool_ctl_struct",44,"off")=1668
	Set gtmtypes("jnlpool_ctl_struct",44,"len")=2048
	Set gtmtypes("jnlpool_ctl_struct",44,"type")="unsigned-char"
	Set gtmtypfldindx("jnlpool_ctl_struct","merrors_array")=44
	Set gtmtypes("jnlpool_ctl_struct",45,"name")="jnlpool_ctl_struct.ftok_counter_halted"
	Set gtmtypes("jnlpool_ctl_struct",45,"off")=3716
	Set gtmtypes("jnlpool_ctl_struct",45,"len")=4
	Set gtmtypes("jnlpool_ctl_struct",45,"type")="boolean_t"
	Set gtmtypfldindx("jnlpool_ctl_struct","ftok_counter_halted")=45
	Set gtmtypes("jnlpool_ctl_struct",46,"name")="jnlpool_ctl_struct.phase2_commit_index1"
	Set gtmtypes("jnlpool_ctl_struct",46,"off")=3720
	Set gtmtypes("jnlpool_ctl_struct",46,"len")=4
	Set gtmtypes("jnlpool_ctl_struct",46,"type")="unsigned-int"
	Set gtmtypfldindx("jnlpool_ctl_struct","phase2_commit_index1")=46
	Set gtmtypes("jnlpool_ctl_struct",47,"name")="jnlpool_ctl_struct.phase2_commit_index2"
	Set gtmtypes("jnlpool_ctl_struct",47,"off")=3724
	Set gtmtypes("jnlpool_ctl_struct",47,"len")=4
	Set gtmtypes("jnlpool_ctl_struct",47,"type")="unsigned-int"
	Set gtmtypfldindx("jnlpool_ctl_struct","phase2_commit_index2")=47
	Set gtmtypes("jnlpool_ctl_struct",48,"name")="jnlpool_ctl_struct.filler_16bytealign1"
	Set gtmtypes("jnlpool_ctl_struct",48,"off")=3728
	Set gtmtypes("jnlpool_ctl_struct",48,"len")=16
	Set gtmtypes("jnlpool_ctl_struct",48,"type")="char"
	Set gtmtypfldindx("jnlpool_ctl_struct","filler_16bytealign1")=48
	Set gtmtypes("jnlpool_ctl_struct",49,"name")="jnlpool_ctl_struct.jnl_pool_write_sleep"
	Set gtmtypes("jnlpool_ctl_struct",49,"off")=3744
	Set gtmtypes("jnlpool_ctl_struct",49,"len")=16
	Set gtmtypes("jnlpool_ctl_struct",49,"type")="jpl_trc_rec_t"
	Set gtmtypfldindx("jnlpool_ctl_struct","jnl_pool_write_sleep")=49
	Set gtmtypes("jnlpool_ctl_struct",50,"name")="jnlpool_ctl_struct.jnl_pool_write_sleep.cntr"
	Set gtmtypes("jnlpool_ctl_struct",50,"off")=3744
	Set gtmtypes("jnlpool_ctl_struct",50,"len")=8
	Set gtmtypes("jnlpool_ctl_struct",50,"type")="uint64_t"
	Set gtmtypfldindx("jnlpool_ctl_struct","jnl_pool_write_sleep.cntr")=50
	Set gtmtypes("jnlpool_ctl_struct",51,"name")="jnlpool_ctl_struct.jnl_pool_write_sleep.seqno"
	Set gtmtypes("jnlpool_ctl_struct",51,"off")=3752
	Set gtmtypes("jnlpool_ctl_struct",51,"len")=8
	Set gtmtypes("jnlpool_ctl_struct",51,"type")="uint64_t"
	Set gtmtypfldindx("jnlpool_ctl_struct","jnl_pool_write_sleep.seqno")=51
	Set gtmtypes("jnlpool_ctl_struct",52,"name")="jnlpool_ctl_struct.repl_phs2cmt_arrayfull_sleep"
	Set gtmtypes("jnlpool_ctl_struct",52,"off")=3760
	Set gtmtypes("jnlpool_ctl_struct",52,"len")=16
	Set gtmtypes("jnlpool_ctl_struct",52,"type")="jpl_trc_rec_t"
	Set gtmtypfldindx("jnlpool_ctl_struct","repl_phs2cmt_arrayfull_sleep")=52
	Set gtmtypes("jnlpool_ctl_struct",53,"name")="jnlpool_ctl_struct.repl_phs2cmt_arrayfull_sleep.cntr"
	Set gtmtypes("jnlpool_ctl_struct",53,"off")=3760
	Set gtmtypes("jnlpool_ctl_struct",53,"len")=8
	Set gtmtypes("jnlpool_ctl_struct",53,"type")="uint64_t"
	Set gtmtypfldindx("jnlpool_ctl_struct","repl_phs2cmt_arrayfull_sleep.cntr")=53
	Set gtmtypes("jnlpool_ctl_struct",54,"name")="jnlpool_ctl_struct.repl_phs2cmt_arrayfull_sleep.seqno"
	Set gtmtypes("jnlpool_ctl_struct",54,"off")=3768
	Set gtmtypes("jnlpool_ctl_struct",54,"len")=8
	Set gtmtypes("jnlpool_ctl_struct",54,"type")="uint64_t"
	Set gtmtypfldindx("jnlpool_ctl_struct","repl_phs2cmt_arrayfull_sleep.seqno")=54
	Set gtmtypes("jnlpool_ctl_struct",55,"name")="jnlpool_ctl_struct.repl_phase2_cleanup_isprcalv"
	Set gtmtypes("jnlpool_ctl_struct",55,"off")=3776
	Set gtmtypes("jnlpool_ctl_struct",55,"len")=16
	Set gtmtypes("jnlpool_ctl_struct",55,"type")="jpl_trc_rec_t"
	Set gtmtypfldindx("jnlpool_ctl_struct","repl_phase2_cleanup_isprcalv")=55
	Set gtmtypes("jnlpool_ctl_struct",56,"name")="jnlpool_ctl_struct.repl_phase2_cleanup_isprcalv.cntr"
	Set gtmtypes("jnlpool_ctl_struct",56,"off")=3776
	Set gtmtypes("jnlpool_ctl_struct",56,"len")=8
	Set gtmtypes("jnlpool_ctl_struct",56,"type")="uint64_t"
	Set gtmtypfldindx("jnlpool_ctl_struct","repl_phase2_cleanup_isprcalv.cntr")=56
	Set gtmtypes("jnlpool_ctl_struct",57,"name")="jnlpool_ctl_struct.repl_phase2_cleanup_isprcalv.seqno"
	Set gtmtypes("jnlpool_ctl_struct",57,"off")=3784
	Set gtmtypes("jnlpool_ctl_struct",57,"len")=8
	Set gtmtypes("jnlpool_ctl_struct",57,"type")="uint64_t"
	Set gtmtypfldindx("jnlpool_ctl_struct","repl_phase2_cleanup_isprcalv.seqno")=57
	Set gtmtypes("jnlpool_ctl_struct",58,"name")="jnlpool_ctl_struct.phase2_commit_array"
	Set gtmtypes("jnlpool_ctl_struct",58,"off")=3792
	Set gtmtypes("jnlpool_ctl_struct",58,"len")=655360
	Set gtmtypes("jnlpool_ctl_struct",58,"type")="jpl_phase2_in_prog_t"
	Set gtmtypfldindx("jnlpool_ctl_struct","phase2_commit_array")=58
	Set gtmtypes("jnlpool_ctl_struct",58,"dim")=16384
	Set gtmtypes("jnlpool_ctl_struct",59,"name")="jnlpool_ctl_struct.phase2_commit_array[0].jnl_seqno"
	Set gtmtypes("jnlpool_ctl_struct",59,"off")=3792
	Set gtmtypes("jnlpool_ctl_struct",59,"len")=8
	Set gtmtypes("jnlpool_ctl_struct",59,"type")="uint64_t"
	Set gtmtypfldindx("jnlpool_ctl_struct","phase2_commit_array[0].jnl_seqno")=59
	Set gtmtypes("jnlpool_ctl_struct",60,"name")="jnlpool_ctl_struct.phase2_commit_array[0].strm_seqno"
	Set gtmtypes("jnlpool_ctl_struct",60,"off")=3800
	Set gtmtypes("jnlpool_ctl_struct",60,"len")=8
	Set gtmtypes("jnlpool_ctl_struct",60,"type")="uint64_t"
	Set gtmtypfldindx("jnlpool_ctl_struct","phase2_commit_array[0].strm_seqno")=60
	Set gtmtypes("jnlpool_ctl_struct",61,"name")="jnlpool_ctl_struct.phase2_commit_array[0].start_write_addr"
	Set gtmtypes("jnlpool_ctl_struct",61,"off")=3808
	Set gtmtypes("jnlpool_ctl_struct",61,"len")=8
	Set gtmtypes("jnlpool_ctl_struct",61,"type")="uint64_t"
	Set gtmtypfldindx("jnlpool_ctl_struct","phase2_commit_array[0].start_write_addr")=61
	Set gtmtypes("jnlpool_ctl_struct",62,"name")="jnlpool_ctl_struct.phase2_commit_array[0].process_id"
	Set gtmtypes("jnlpool_ctl_struct",62,"off")=3816
	Set gtmtypes("jnlpool_ctl_struct",62,"len")=4
	Set gtmtypes("jnlpool_ctl_struct",62,"type")="unsigned-int"
	Set gtmtypfldindx("jnlpool_ctl_struct","phase2_commit_array[0].process_id")=62
	Set gtmtypes("jnlpool_ctl_struct",63,"name")="jnlpool_ctl_struct.phase2_commit_array[0].tot_jrec_len"
	Set gtmtypes("jnlpool_ctl_struct",63,"off")=3820
	Set gtmtypes("jnlpool_ctl_struct",63,"len")=4
	Set gtmtypes("jnlpool_ctl_struct",63,"type")="unsigned-int"
	Set gtmtypfldindx("jnlpool_ctl_struct","phase2_commit_array[0].tot_jrec_len")=63
	Set gtmtypes("jnlpool_ctl_struct",64,"name")="jnlpool_ctl_struct.phase2_commit_array[0].prev_jrec_len"
	Set gtmtypes("jnlpool_ctl_struct",64,"off")=3824
	Set gtmtypes("jnlpool_ctl_struct",64,"len")=4
	Set gtmtypes("jnlpool_ctl_struct",64,"type")="unsigned-int"
	Set gtmtypfldindx("jnlpool_ctl_struct","phase2_commit_array[0].prev_jrec_len")=64
	Set gtmtypes("jnlpool_ctl_struct",65,"name")="jnlpool_ctl_struct.phase2_commit_array[0].write_complete"
	Set gtmtypes("jnlpool_ctl_struct",65,"off")=3828
	Set gtmtypes("jnlpool_ctl_struct",65,"len")=4
	Set gtmtypes("jnlpool_ctl_struct",65,"type")="boolean_t"
	Set gtmtypfldindx("jnlpool_ctl_struct","phase2_commit_array[0].write_complete")=65
	Set gtmtypes("jnlpool_ctl_struct",66,"name")="jnlpool_ctl_struct.fill_cacheline0"
	Set gtmtypes("jnlpool_ctl_struct",66,"off")=659152
	Set gtmtypes("jnlpool_ctl_struct",66,"len")=56
	Set gtmtypes("jnlpool_ctl_struct",66,"type")="char"
	Set gtmtypfldindx("jnlpool_ctl_struct","fill_cacheline0")=66
	Set gtmtypes("jnlpool_ctl_struct",67,"name")="jnlpool_ctl_struct.phase2_commit_latch"
	Set gtmtypes("jnlpool_ctl_struct",67,"off")=659208
	Set gtmtypes("jnlpool_ctl_struct",67,"len")=8
	Set gtmtypes("jnlpool_ctl_struct",67,"type")="global_latch_t"
	Set gtmtypfldindx("jnlpool_ctl_struct","phase2_commit_latch")=67
	Set gtmtypes("jnlpool_ctl_struct",68,"name")="jnlpool_ctl_struct.phase2_commit_latch.u"
	Set gtmtypes("jnlpool_ctl_struct",68,"off")=659208
	Set gtmtypes("jnlpool_ctl_struct",68,"len")=8
	Set gtmtypes("jnlpool_ctl_struct",68,"type")="union"
	Set gtmtypfldindx("jnlpool_ctl_struct","phase2_commit_latch.u")=68
	Set gtmtypes("jnlpool_ctl_struct",69,"name")="jnlpool_ctl_struct.phase2_commit_latch.u.pid_imgcnt"
	Set gtmtypes("jnlpool_ctl_struct",69,"off")=659208
	Set gtmtypes("jnlpool_ctl_struct",69,"len")=8
	Set gtmtypes("jnlpool_ctl_struct",69,"type")="uint64_t"
	Set gtmtypfldindx("jnlpool_ctl_struct","phase2_commit_latch.u.pid_imgcnt")=69
	Set gtmtypes("jnlpool_ctl_struct",70,"name")="jnlpool_ctl_struct.phase2_commit_latch.u.parts"
	Set gtmtypes("jnlpool_ctl_struct",70,"off")=659208
	Set gtmtypes("jnlpool_ctl_struct",70,"len")=8
	Set gtmtypes("jnlpool_ctl_struct",70,"type")="struct"
	Set gtmtypfldindx("jnlpool_ctl_struct","phase2_commit_latch.u.parts")=70
	Set gtmtypes("jnlpool_ctl_struct",71,"name")="jnlpool_ctl_struct.phase2_commit_latch.u.parts.latch_pid"
	Set gtmtypes("jnlpool_ctl_struct",71,"off")=659208
	Set gtmtypes("jnlpool_ctl_struct",71,"len")=4
	Set gtmtypes("jnlpool_ctl_struct",71,"type")="int"
	Set gtmtypfldindx("jnlpool_ctl_struct","phase2_commit_latch.u.parts.latch_pid")=71
	Set gtmtypes("jnlpool_ctl_struct",72,"name")="jnlpool_ctl_struct.phase2_commit_latch.u.parts.latch_word"
	Set gtmtypes("jnlpool_ctl_struct",72,"off")=659212
	Set gtmtypes("jnlpool_ctl_struct",72,"len")=4
	Set gtmtypes("jnlpool_ctl_struct",72,"type")="int"
	Set gtmtypfldindx("jnlpool_ctl_struct","phase2_commit_latch.u.parts.latch_word")=72
	;
	Set gtmtypes("job_arg_msg")="struct"
	Set gtmtypes("job_arg_msg",0)=2
	Set gtmtypes("job_arg_msg","len")=8200
	Set gtmtypes("job_arg_msg",1,"name")="job_arg_msg.len"
	Set gtmtypes("job_arg_msg",1,"off")=0
	Set gtmtypes("job_arg_msg",1,"len")=8
	Set gtmtypes("job_arg_msg",1,"type")="ssize_t"
	Set gtmtypfldindx("job_arg_msg","len")=1
	Set gtmtypes("job_arg_msg",2,"name")="job_arg_msg.data"
	Set gtmtypes("job_arg_msg",2,"off")=8
	Set gtmtypes("job_arg_msg",2,"len")=8192
	Set gtmtypes("job_arg_msg",2,"type")="char"
	Set gtmtypfldindx("job_arg_msg","data")=2
	;
	Set gtmtypes("job_param_str")="struct"
	Set gtmtypes("job_param_str",0)=2
	Set gtmtypes("job_param_str","len")=1032
	Set gtmtypes("job_param_str",1,"name")="job_param_str.len"
	Set gtmtypes("job_param_str",1,"off")=0
	Set gtmtypes("job_param_str",1,"len")=8
	Set gtmtypes("job_param_str",1,"type")="size_t"
	Set gtmtypfldindx("job_param_str","len")=1
	Set gtmtypes("job_param_str",2,"name")="job_param_str.buffer"
	Set gtmtypes("job_param_str",2,"off")=8
	Set gtmtypes("job_param_str",2,"len")=1024
	Set gtmtypes("job_param_str",2,"type")="char"
	Set gtmtypfldindx("job_param_str","buffer")=2
	;
	Set gtmtypes("job_params_type")="struct"
	Set gtmtypes("job_params_type",0)=10
	Set gtmtypes("job_params_type","len")=9344
	Set gtmtypes("job_params_type",1,"name")="job_params_type.params"
	Set gtmtypes("job_params_type",1,"off")=0
	Set gtmtypes("job_params_type",1,"len")=8264
	Set gtmtypes("job_params_type",1,"type")="struct"
	Set gtmtypfldindx("job_params_type","params")=1
	Set gtmtypes("job_params_type",2,"name")="job_params_type.cmdline"
	Set gtmtypes("job_params_type",2,"off")=8264
	Set gtmtypes("job_params_type",2,"len")=1032
	Set gtmtypes("job_params_type",2,"type")="job_param_str"
	Set gtmtypfldindx("job_params_type","cmdline")=2
	Set gtmtypes("job_params_type",3,"name")="job_params_type.cmdline.len"
	Set gtmtypes("job_params_type",3,"off")=8264
	Set gtmtypes("job_params_type",3,"len")=8
	Set gtmtypes("job_params_type",3,"type")="size_t"
	Set gtmtypfldindx("job_params_type","cmdline.len")=3
	Set gtmtypes("job_params_type",4,"name")="job_params_type.cmdline.buffer"
	Set gtmtypes("job_params_type",4,"off")=8272
	Set gtmtypes("job_params_type",4,"len")=1024
	Set gtmtypes("job_params_type",4,"type")="char"
	Set gtmtypfldindx("job_params_type","cmdline.buffer")=4
	Set gtmtypes("job_params_type",5,"name")="job_params_type.parms"
	Set gtmtypes("job_params_type",5,"off")=9296
	Set gtmtypes("job_params_type",5,"len")=8
	Set gtmtypes("job_params_type",5,"type")="addr"
	Set gtmtypfldindx("job_params_type","parms")=5
	Set gtmtypes("job_params_type",6,"name")="job_params_type.input_prebuffer_size"
	Set gtmtypes("job_params_type",6,"off")=9304
	Set gtmtypes("job_params_type",6,"len")=8
	Set gtmtypes("job_params_type",6,"type")="size_t"
	Set gtmtypfldindx("job_params_type","input_prebuffer_size")=6
	Set gtmtypes("job_params_type",7,"name")="job_params_type.input_prebuffer"
	Set gtmtypes("job_params_type",7,"off")=9312
	Set gtmtypes("job_params_type",7,"len")=8
	Set gtmtypes("job_params_type",7,"type")="addr"
	Set gtmtypfldindx("job_params_type","input_prebuffer")=7
	Set gtmtypes("job_params_type",8,"name")="job_params_type.passcurlvn"
	Set gtmtypes("job_params_type",8,"off")=9320
	Set gtmtypes("job_params_type",8,"len")=4
	Set gtmtypes("job_params_type",8,"type")="boolean_t"
	Set gtmtypfldindx("job_params_type","passcurlvn")=8
	Set gtmtypes("job_params_type",9,"name")="job_params_type.curlvn_buffer_ptr"
	Set gtmtypes("job_params_type",9,"off")=9328
	Set gtmtypes("job_params_type",9,"len")=8
	Set gtmtypes("job_params_type",9,"type")="addr"
	Set gtmtypfldindx("job_params_type","curlvn_buffer_ptr")=9
	Set gtmtypes("job_params_type",10,"name")="job_params_type.curlvn_buffer_size"
	Set gtmtypes("job_params_type",10,"off")=9336
	Set gtmtypes("job_params_type",10,"len")=8
	Set gtmtypes("job_params_type",10,"type")="size_t"
	Set gtmtypfldindx("job_params_type","curlvn_buffer_size")=10
	;
	Set gtmtypes("job_parm")="struct"
	Set gtmtypes("job_parm",0)=2
	Set gtmtypes("job_parm","len")=16
	Set gtmtypes("job_parm",1,"name")="job_parm.parm"
	Set gtmtypes("job_parm",1,"off")=0
	Set gtmtypes("job_parm",1,"len")=8
	Set gtmtypes("job_parm",1,"type")="addr"
	Set gtmtypfldindx("job_parm","parm")=1
	Set gtmtypes("job_parm",2,"name")="job_parm.next"
	Set gtmtypes("job_parm",2,"off")=8
	Set gtmtypes("job_parm",2,"len")=8
	Set gtmtypes("job_parm",2,"type")="addr"
	Set gtmtypfldindx("job_parm","next")=2
	;
	Set gtmtypes("joberr_msg")="struct"
	Set gtmtypes("joberr_msg",0)=2
	Set gtmtypes("joberr_msg","len")=16
	Set gtmtypes("joberr_msg",1,"name")="joberr_msg.msg"
	Set gtmtypes("joberr_msg",1,"off")=0
	Set gtmtypes("joberr_msg",1,"len")=8
	Set gtmtypes("joberr_msg",1,"type")="addr"
	Set gtmtypfldindx("joberr_msg","msg")=1
	Set gtmtypes("joberr_msg",2,"name")="joberr_msg.len"
	Set gtmtypes("joberr_msg",2,"off")=8
	Set gtmtypes("joberr_msg",2,"len")=4
	Set gtmtypes("joberr_msg",2,"type")="int"
	Set gtmtypfldindx("joberr_msg","len")=2
	;
	Set gtmtypes("jpl_phase2_in_prog_t")="struct"
	Set gtmtypes("jpl_phase2_in_prog_t",0)=7
	Set gtmtypes("jpl_phase2_in_prog_t","len")=40
	Set gtmtypes("jpl_phase2_in_prog_t",1,"name")="jpl_phase2_in_prog_t.jnl_seqno"
	Set gtmtypes("jpl_phase2_in_prog_t",1,"off")=0
	Set gtmtypes("jpl_phase2_in_prog_t",1,"len")=8
	Set gtmtypes("jpl_phase2_in_prog_t",1,"type")="uint64_t"
	Set gtmtypfldindx("jpl_phase2_in_prog_t","jnl_seqno")=1
	Set gtmtypes("jpl_phase2_in_prog_t",2,"name")="jpl_phase2_in_prog_t.strm_seqno"
	Set gtmtypes("jpl_phase2_in_prog_t",2,"off")=8
	Set gtmtypes("jpl_phase2_in_prog_t",2,"len")=8
	Set gtmtypes("jpl_phase2_in_prog_t",2,"type")="uint64_t"
	Set gtmtypfldindx("jpl_phase2_in_prog_t","strm_seqno")=2
	Set gtmtypes("jpl_phase2_in_prog_t",3,"name")="jpl_phase2_in_prog_t.start_write_addr"
	Set gtmtypes("jpl_phase2_in_prog_t",3,"off")=16
	Set gtmtypes("jpl_phase2_in_prog_t",3,"len")=8
	Set gtmtypes("jpl_phase2_in_prog_t",3,"type")="uint64_t"
	Set gtmtypfldindx("jpl_phase2_in_prog_t","start_write_addr")=3
	Set gtmtypes("jpl_phase2_in_prog_t",4,"name")="jpl_phase2_in_prog_t.process_id"
	Set gtmtypes("jpl_phase2_in_prog_t",4,"off")=24
	Set gtmtypes("jpl_phase2_in_prog_t",4,"len")=4
	Set gtmtypes("jpl_phase2_in_prog_t",4,"type")="unsigned-int"
	Set gtmtypfldindx("jpl_phase2_in_prog_t","process_id")=4
	Set gtmtypes("jpl_phase2_in_prog_t",5,"name")="jpl_phase2_in_prog_t.tot_jrec_len"
	Set gtmtypes("jpl_phase2_in_prog_t",5,"off")=28
	Set gtmtypes("jpl_phase2_in_prog_t",5,"len")=4
	Set gtmtypes("jpl_phase2_in_prog_t",5,"type")="unsigned-int"
	Set gtmtypfldindx("jpl_phase2_in_prog_t","tot_jrec_len")=5
	Set gtmtypes("jpl_phase2_in_prog_t",6,"name")="jpl_phase2_in_prog_t.prev_jrec_len"
	Set gtmtypes("jpl_phase2_in_prog_t",6,"off")=32
	Set gtmtypes("jpl_phase2_in_prog_t",6,"len")=4
	Set gtmtypes("jpl_phase2_in_prog_t",6,"type")="unsigned-int"
	Set gtmtypfldindx("jpl_phase2_in_prog_t","prev_jrec_len")=6
	Set gtmtypes("jpl_phase2_in_prog_t",7,"name")="jpl_phase2_in_prog_t.write_complete"
	Set gtmtypes("jpl_phase2_in_prog_t",7,"off")=36
	Set gtmtypes("jpl_phase2_in_prog_t",7,"len")=4
	Set gtmtypes("jpl_phase2_in_prog_t",7,"type")="boolean_t"
	Set gtmtypfldindx("jpl_phase2_in_prog_t","write_complete")=7
	;
	Set gtmtypes("jpl_rsrv_struct_t")="struct"
	Set gtmtypes("jpl_rsrv_struct_t",0)=8
	Set gtmtypes("jpl_rsrv_struct_t","len")=40
	Set gtmtypes("jpl_rsrv_struct_t",1,"name")="jpl_rsrv_struct_t.start_write_addr"
	Set gtmtypes("jpl_rsrv_struct_t",1,"off")=0
	Set gtmtypes("jpl_rsrv_struct_t",1,"len")=8
	Set gtmtypes("jpl_rsrv_struct_t",1,"type")="uint64_t"
	Set gtmtypfldindx("jpl_rsrv_struct_t","start_write_addr")=1
	Set gtmtypes("jpl_rsrv_struct_t",2,"name")="jpl_rsrv_struct_t.cur_write_addr"
	Set gtmtypes("jpl_rsrv_struct_t",2,"off")=8
	Set gtmtypes("jpl_rsrv_struct_t",2,"len")=8
	Set gtmtypes("jpl_rsrv_struct_t",2,"type")="uint64_t"
	Set gtmtypfldindx("jpl_rsrv_struct_t","cur_write_addr")=2
	Set gtmtypes("jpl_rsrv_struct_t",3,"name")="jpl_rsrv_struct_t.tot_jrec_len"
	Set gtmtypes("jpl_rsrv_struct_t",3,"off")=16
	Set gtmtypes("jpl_rsrv_struct_t",3,"len")=4
	Set gtmtypes("jpl_rsrv_struct_t",3,"type")="unsigned-int"
	Set gtmtypfldindx("jpl_rsrv_struct_t","tot_jrec_len")=3
	Set gtmtypes("jpl_rsrv_struct_t",4,"name")="jpl_rsrv_struct_t.write_total"
	Set gtmtypes("jpl_rsrv_struct_t",4,"off")=20
	Set gtmtypes("jpl_rsrv_struct_t",4,"len")=4
	Set gtmtypes("jpl_rsrv_struct_t",4,"type")="unsigned-int"
	Set gtmtypfldindx("jpl_rsrv_struct_t","write_total")=4
	Set gtmtypes("jpl_rsrv_struct_t",5,"name")="jpl_rsrv_struct_t.memcpy_skipped"
	Set gtmtypes("jpl_rsrv_struct_t",5,"off")=24
	Set gtmtypes("jpl_rsrv_struct_t",5,"len")=4
	Set gtmtypes("jpl_rsrv_struct_t",5,"type")="boolean_t"
	Set gtmtypfldindx("jpl_rsrv_struct_t","memcpy_skipped")=5
	Set gtmtypes("jpl_rsrv_struct_t",6,"name")="jpl_rsrv_struct_t.phase2_commit_index"
	Set gtmtypes("jpl_rsrv_struct_t",6,"off")=28
	Set gtmtypes("jpl_rsrv_struct_t",6,"len")=4
	Set gtmtypes("jpl_rsrv_struct_t",6,"type")="unsigned-int"
	Set gtmtypfldindx("jpl_rsrv_struct_t","phase2_commit_index")=6
	Set gtmtypes("jpl_rsrv_struct_t",7,"name")="jpl_rsrv_struct_t.num_tcoms"
	Set gtmtypes("jpl_rsrv_struct_t",7,"off")=32
	Set gtmtypes("jpl_rsrv_struct_t",7,"len")=4
	Set gtmtypes("jpl_rsrv_struct_t",7,"type")="unsigned-int"
	Set gtmtypfldindx("jpl_rsrv_struct_t","num_tcoms")=7
	Set gtmtypes("jpl_rsrv_struct_t",8,"name")="jpl_rsrv_struct_t.filler_8byte_align"
	Set gtmtypes("jpl_rsrv_struct_t",8,"off")=36
	Set gtmtypes("jpl_rsrv_struct_t",8,"len")=4
	Set gtmtypes("jpl_rsrv_struct_t",8,"type")="char"
	Set gtmtypfldindx("jpl_rsrv_struct_t","filler_8byte_align")=8
	;
	Set gtmtypes("jpl_trc_rec_t")="struct"
	Set gtmtypes("jpl_trc_rec_t",0)=2
	Set gtmtypes("jpl_trc_rec_t","len")=16
	Set gtmtypes("jpl_trc_rec_t",1,"name")="jpl_trc_rec_t.cntr"
	Set gtmtypes("jpl_trc_rec_t",1,"off")=0
	Set gtmtypes("jpl_trc_rec_t",1,"len")=8
	Set gtmtypes("jpl_trc_rec_t",1,"type")="uint64_t"
	Set gtmtypfldindx("jpl_trc_rec_t","cntr")=1
	Set gtmtypes("jpl_trc_rec_t",2,"name")="jpl_trc_rec_t.seqno"
	Set gtmtypes("jpl_trc_rec_t",2,"off")=8
	Set gtmtypes("jpl_trc_rec_t",2,"len")=8
	Set gtmtypes("jpl_trc_rec_t",2,"type")="uint64_t"
	Set gtmtypfldindx("jpl_trc_rec_t","seqno")=2
	;
	Set gtmtypes("jrec_prefix")="struct"
	Set gtmtypes("jrec_prefix",0)=4
	Set gtmtypes("jrec_prefix","len")=24
	Set gtmtypes("jrec_prefix",1,"name")="jrec_prefix.pini_addr"
	Set gtmtypes("jrec_prefix",1,"off")=4
	Set gtmtypes("jrec_prefix",1,"len")=4
	Set gtmtypes("jrec_prefix",1,"type")="unsigned-int"
	Set gtmtypfldindx("jrec_prefix","pini_addr")=1
	Set gtmtypes("jrec_prefix",2,"name")="jrec_prefix.time"
	Set gtmtypes("jrec_prefix",2,"off")=8
	Set gtmtypes("jrec_prefix",2,"len")=4
	Set gtmtypes("jrec_prefix",2,"type")="unsigned-int"
	Set gtmtypfldindx("jrec_prefix","time")=2
	Set gtmtypes("jrec_prefix",3,"name")="jrec_prefix.checksum"
	Set gtmtypes("jrec_prefix",3,"off")=12
	Set gtmtypes("jrec_prefix",3,"len")=4
	Set gtmtypes("jrec_prefix",3,"type")="unsigned-int"
	Set gtmtypfldindx("jrec_prefix","checksum")=3
	Set gtmtypes("jrec_prefix",4,"name")="jrec_prefix.tn"
	Set gtmtypes("jrec_prefix",4,"off")=16
	Set gtmtypes("jrec_prefix",4,"len")=8
	Set gtmtypes("jrec_prefix",4,"type")="uint64_t"
	Set gtmtypfldindx("jrec_prefix","tn")=4
	;
	Set gtmtypes("jrec_rsrv_elem_t")="struct"
	Set gtmtypes("jrec_rsrv_elem_t",0)=3
	Set gtmtypes("jrec_rsrv_elem_t","len")=16
	Set gtmtypes("jrec_rsrv_elem_t",1,"name")="jrec_rsrv_elem_t.rectype"
	Set gtmtypes("jrec_rsrv_elem_t",1,"off")=0
	Set gtmtypes("jrec_rsrv_elem_t",1,"len")=4
	Set gtmtypes("jrec_rsrv_elem_t",1,"type")="int"
	Set gtmtypfldindx("jrec_rsrv_elem_t","rectype")=1
	Set gtmtypes("jrec_rsrv_elem_t",2,"name")="jrec_rsrv_elem_t.reclen"
	Set gtmtypes("jrec_rsrv_elem_t",2,"off")=4
	Set gtmtypes("jrec_rsrv_elem_t",2,"len")=4
	Set gtmtypes("jrec_rsrv_elem_t",2,"type")="unsigned-int"
	Set gtmtypfldindx("jrec_rsrv_elem_t","reclen")=2
	Set gtmtypes("jrec_rsrv_elem_t",3,"name")="jrec_rsrv_elem_t.param1"
	Set gtmtypes("jrec_rsrv_elem_t",3,"off")=8
	Set gtmtypes("jrec_rsrv_elem_t",3,"len")=8
	Set gtmtypes("jrec_rsrv_elem_t",3,"type")="addr"
	Set gtmtypfldindx("jrec_rsrv_elem_t","param1")=3
	;
	Set gtmtypes("key_cum_value")="struct"
	Set gtmtypes("key_cum_value",0)=14
	Set gtmtypes("key_cum_value","len")=1088
	Set gtmtypes("key_cum_value",1,"name")="key_cum_value.keybuf"
	Set gtmtypes("key_cum_value",1,"off")=0
	Set gtmtypes("key_cum_value",1,"len")=1058
	Set gtmtypes("key_cum_value",1,"type")="gv_key_buf"
	Set gtmtypfldindx("key_cum_value","keybuf")=1
	Set gtmtypes("key_cum_value",2,"name")="key_cum_value.keybuf.key"
	Set gtmtypes("key_cum_value",2,"off")=0
	Set gtmtypes("key_cum_value",2,"len")=6
	Set gtmtypes("key_cum_value",2,"type")="gv_key_nobase"
	Set gtmtypfldindx("key_cum_value","keybuf.key")=2
	Set gtmtypes("key_cum_value",3,"name")="key_cum_value.keybuf.key.top"
	Set gtmtypes("key_cum_value",3,"off")=0
	Set gtmtypes("key_cum_value",3,"len")=2
	Set gtmtypes("key_cum_value",3,"type")="unsigned-short"
	Set gtmtypfldindx("key_cum_value","keybuf.key.top")=3
	Set gtmtypes("key_cum_value",4,"name")="key_cum_value.keybuf.key.end"
	Set gtmtypes("key_cum_value",4,"off")=2
	Set gtmtypes("key_cum_value",4,"len")=2
	Set gtmtypes("key_cum_value",4,"type")="unsigned-short"
	Set gtmtypfldindx("key_cum_value","keybuf.key.end")=4
	Set gtmtypes("key_cum_value",5,"name")="key_cum_value.keybuf.key.prev"
	Set gtmtypes("key_cum_value",5,"off")=4
	Set gtmtypes("key_cum_value",5,"len")=2
	Set gtmtypes("key_cum_value",5,"type")="unsigned-short"
	Set gtmtypfldindx("key_cum_value","keybuf.key.prev")=5
	Set gtmtypes("key_cum_value",6,"name")="key_cum_value.keybuf.buf"
	Set gtmtypes("key_cum_value",6,"off")=0
	Set gtmtypes("key_cum_value",6,"len")=1058
	Set gtmtypes("key_cum_value",6,"type")="char"
	Set gtmtypfldindx("key_cum_value","keybuf.buf")=6
	Set gtmtypes("key_cum_value",7,"name")="key_cum_value.keybuf.split"
	Set gtmtypes("key_cum_value",7,"off")=0
	Set gtmtypes("key_cum_value",7,"len")=1058
	Set gtmtypes("key_cum_value",7,"type")="struct"
	Set gtmtypfldindx("key_cum_value","keybuf.split")=7
	Set gtmtypes("key_cum_value",8,"name")="key_cum_value.keybuf.split.gv_key_data"
	Set gtmtypes("key_cum_value",8,"off")=0
	Set gtmtypes("key_cum_value",8,"len")=6
	Set gtmtypes("key_cum_value",8,"type")="char"
	Set gtmtypfldindx("key_cum_value","keybuf.split.gv_key_data")=8
	Set gtmtypes("key_cum_value",9,"name")="key_cum_value.keybuf.split.base"
	Set gtmtypes("key_cum_value",9,"off")=6
	Set gtmtypes("key_cum_value",9,"len")=1052
	Set gtmtypes("key_cum_value",9,"type")="char"
	Set gtmtypfldindx("key_cum_value","keybuf.split.base")=9
	Set gtmtypes("key_cum_value",10,"name")="key_cum_value.value"
	Set gtmtypes("key_cum_value",10,"off")=1064
	Set gtmtypes("key_cum_value",10,"len")=16
	Set gtmtypes("key_cum_value",10,"type")="mstr"
	Set gtmtypfldindx("key_cum_value","value")=10
	Set gtmtypes("key_cum_value",11,"name")="key_cum_value.value.char_len"
	Set gtmtypes("key_cum_value",11,"off")=1064
	Set gtmtypes("key_cum_value",11,"len")=4
	Set gtmtypes("key_cum_value",11,"type")="unsigned-int"
	Set gtmtypfldindx("key_cum_value","value.char_len")=11
	Set gtmtypes("key_cum_value",12,"name")="key_cum_value.value.len"
	Set gtmtypes("key_cum_value",12,"off")=1068
	Set gtmtypes("key_cum_value",12,"len")=4
	Set gtmtypes("key_cum_value",12,"type")="int"
	Set gtmtypfldindx("key_cum_value","value.len")=12
	Set gtmtypes("key_cum_value",13,"name")="key_cum_value.value.addr"
	Set gtmtypes("key_cum_value",13,"off")=1072
	Set gtmtypes("key_cum_value",13,"len")=8
	Set gtmtypes("key_cum_value",13,"type")="addr"
	Set gtmtypfldindx("key_cum_value","value.addr")=13
	Set gtmtypes("key_cum_value",14,"name")="key_cum_value.next"
	Set gtmtypes("key_cum_value",14,"off")=1080
	Set gtmtypes("key_cum_value",14,"len")=8
	Set gtmtypes("key_cum_value",14,"type")="addr"
	Set gtmtypfldindx("key_cum_value","next")=14
	;
	Set gtmtypes("kill_set")="struct"
	Set gtmtypes("kill_set",0)=3
	Set gtmtypes("kill_set","len")=1016
	Set gtmtypes("kill_set",1,"name")="kill_set.next_kill_set"
	Set gtmtypes("kill_set",1,"off")=0
	Set gtmtypes("kill_set",1,"len")=8
	Set gtmtypes("kill_set",1,"type")="addr"
	Set gtmtypfldindx("kill_set","next_kill_set")=1
	Set gtmtypes("kill_set",2,"name")="kill_set.used"
	Set gtmtypes("kill_set",2,"off")=8
	Set gtmtypes("kill_set",2,"len")=4
	Set gtmtypes("kill_set",2,"type")="int"
	Set gtmtypfldindx("kill_set","used")=2
	Set gtmtypes("kill_set",3,"name")="kill_set.blk"
	Set gtmtypes("kill_set",3,"off")=12
	Set gtmtypes("kill_set",3,"len")=1004
	Set gtmtypes("kill_set",3,"type")="blk_ident"
	Set gtmtypfldindx("kill_set","blk")=3
	Set gtmtypes("kill_set",3,"dim")=251
	;
	Set gtmtypes("lab_tabent")="struct"
	Set gtmtypes("lab_tabent",0)=7
	Set gtmtypes("lab_tabent","len")=32
	Set gtmtypes("lab_tabent",1,"name")="lab_tabent.lab_name"
	Set gtmtypes("lab_tabent",1,"off")=0
	Set gtmtypes("lab_tabent",1,"len")=16
	Set gtmtypes("lab_tabent",1,"type")="mstr"
	Set gtmtypfldindx("lab_tabent","lab_name")=1
	Set gtmtypes("lab_tabent",2,"name")="lab_tabent.lab_name.char_len"
	Set gtmtypes("lab_tabent",2,"off")=0
	Set gtmtypes("lab_tabent",2,"len")=4
	Set gtmtypes("lab_tabent",2,"type")="unsigned-int"
	Set gtmtypfldindx("lab_tabent","lab_name.char_len")=2
	Set gtmtypes("lab_tabent",3,"name")="lab_tabent.lab_name.len"
	Set gtmtypes("lab_tabent",3,"off")=4
	Set gtmtypes("lab_tabent",3,"len")=4
	Set gtmtypes("lab_tabent",3,"type")="int"
	Set gtmtypfldindx("lab_tabent","lab_name.len")=3
	Set gtmtypes("lab_tabent",4,"name")="lab_tabent.lab_name.addr"
	Set gtmtypes("lab_tabent",4,"off")=8
	Set gtmtypes("lab_tabent",4,"len")=8
	Set gtmtypes("lab_tabent",4,"type")="addr"
	Set gtmtypfldindx("lab_tabent","lab_name.addr")=4
	Set gtmtypes("lab_tabent",5,"name")="lab_tabent.lnr_adr"
	Set gtmtypes("lab_tabent",5,"off")=16
	Set gtmtypes("lab_tabent",5,"len")=8
	Set gtmtypes("lab_tabent",5,"type")="addr"
	Set gtmtypfldindx("lab_tabent","lnr_adr")=5
	Set gtmtypes("lab_tabent",6,"name")="lab_tabent.has_parms"
	Set gtmtypes("lab_tabent",6,"off")=24
	Set gtmtypes("lab_tabent",6,"len")=4
	Set gtmtypes("lab_tabent",6,"type")="boolean_t"
	Set gtmtypfldindx("lab_tabent","has_parms")=6
	Set gtmtypes("lab_tabent",7,"name")="lab_tabent.filler"
	Set gtmtypes("lab_tabent",7,"off")=28
	Set gtmtypes("lab_tabent",7,"len")=4
	Set gtmtypes("lab_tabent",7,"type")="int"
	Set gtmtypfldindx("lab_tabent","filler")=7
	;
	Set gtmtypes("link_info")="struct"
	Set gtmtypes("link_info",0)=19
	Set gtmtypes("link_info","len")=80
	Set gtmtypes("link_info",1,"name")="link_info.neterr"
	Set gtmtypes("link_info",1,"off")=0
	Set gtmtypes("link_info",1,"len")=1
	Set gtmtypes("link_info",1,"type")="unsigned-char"
	Set gtmtypfldindx("link_info","neterr")=1
	Set gtmtypes("link_info",2,"name")="link_info.lck_info"
	Set gtmtypes("link_info",2,"off")=1
	Set gtmtypes("link_info",2,"len")=1
	Set gtmtypes("link_info",2,"type")="unsigned-char"
	Set gtmtypfldindx("link_info","lck_info")=2
	Set gtmtypes("link_info",3,"name")="link_info.lnk_active"
	Set gtmtypes("link_info",3,"off")=2
	Set gtmtypes("link_info",3,"len")=1
	Set gtmtypes("link_info",3,"type")="unsigned-char"
	Set gtmtypfldindx("link_info","lnk_active")=3
	Set gtmtypes("link_info",4,"name")="link_info.filler"
	Set gtmtypes("link_info",4,"off")=3
	Set gtmtypes("link_info",4,"len")=1
	Set gtmtypes("link_info",4,"type")="char"
	Set gtmtypfldindx("link_info","filler")=4
	Set gtmtypes("link_info",5,"name")="link_info.netlocks"
	Set gtmtypes("link_info",5,"off")=8
	Set gtmtypes("link_info",5,"len")=8
	Set gtmtypes("link_info",5,"type")="addr"
	Set gtmtypfldindx("link_info","netlocks")=5
	Set gtmtypes("link_info",6,"name")="link_info.procnum"
	Set gtmtypes("link_info",6,"off")=16
	Set gtmtypes("link_info",6,"len")=2
	Set gtmtypes("link_info",6,"type")="unsigned-short"
	Set gtmtypfldindx("link_info","procnum")=6
	Set gtmtypes("link_info",7,"name")="link_info.buffered_count"
	Set gtmtypes("link_info",7,"off")=18
	Set gtmtypes("link_info",7,"len")=2
	Set gtmtypes("link_info",7,"type")="unsigned-short"
	Set gtmtypfldindx("link_info","buffered_count")=7
	Set gtmtypes("link_info",8,"name")="link_info.buffer_size"
	Set gtmtypes("link_info",8,"off")=20
	Set gtmtypes("link_info",8,"len")=2
	Set gtmtypes("link_info",8,"type")="unsigned-short"
	Set gtmtypfldindx("link_info","buffer_size")=8
	Set gtmtypes("link_info",9,"name")="link_info.buffer_used"
	Set gtmtypes("link_info",9,"off")=22
	Set gtmtypes("link_info",9,"len")=2
	Set gtmtypes("link_info",9,"type")="unsigned-short"
	Set gtmtypfldindx("link_info","buffer_used")=9
	Set gtmtypes("link_info",10,"name")="link_info.buffer"
	Set gtmtypes("link_info",10,"off")=24
	Set gtmtypes("link_info",10,"len")=8
	Set gtmtypes("link_info",10,"type")="addr"
	Set gtmtypfldindx("link_info","buffer")=10
	Set gtmtypes("link_info",11,"name")="link_info.convert_byteorder"
	Set gtmtypes("link_info",11,"off")=32
	Set gtmtypes("link_info",11,"len")=4
	Set gtmtypes("link_info",11,"type")="boolean_t"
	Set gtmtypfldindx("link_info","convert_byteorder")=11
	Set gtmtypes("link_info",12,"name")="link_info.query_is_queryget"
	Set gtmtypes("link_info",12,"off")=36
	Set gtmtypes("link_info",12,"len")=4
	Set gtmtypes("link_info",12,"type")="boolean_t"
	Set gtmtypfldindx("link_info","query_is_queryget")=12
	Set gtmtypes("link_info",13,"name")="link_info.err_compat"
	Set gtmtypes("link_info",13,"off")=40
	Set gtmtypes("link_info",13,"len")=4
	Set gtmtypes("link_info",13,"type")="boolean_t"
	Set gtmtypfldindx("link_info","err_compat")=13
	Set gtmtypes("link_info",14,"name")="link_info.lk_response"
	Set gtmtypes("link_info",14,"off")=48
	Set gtmtypes("link_info",14,"len")=16
	Set gtmtypes("link_info",14,"type")="cm_lk_response"
	Set gtmtypfldindx("link_info","lk_response")=14
	Set gtmtypes("link_info",15,"name")="link_info.lk_response.next"
	Set gtmtypes("link_info",15,"off")=48
	Set gtmtypes("link_info",15,"len")=8
	Set gtmtypes("link_info",15,"type")="addr"
	Set gtmtypfldindx("link_info","lk_response.next")=15
	Set gtmtypes("link_info",16,"name")="link_info.lk_response.response"
	Set gtmtypes("link_info",16,"off")=56
	Set gtmtypes("link_info",16,"len")=8
	Set gtmtypes("link_info",16,"type")="addr"
	Set gtmtypfldindx("link_info","lk_response.response")=16
	Set gtmtypes("link_info",17,"name")="link_info.server_supports_dollar_incr"
	Set gtmtypes("link_info",17,"off")=64
	Set gtmtypes("link_info",17,"len")=4
	Set gtmtypes("link_info",17,"type")="boolean_t"
	Set gtmtypfldindx("link_info","server_supports_dollar_incr")=17
	Set gtmtypes("link_info",18,"name")="link_info.server_supports_std_null_coll"
	Set gtmtypes("link_info",18,"off")=68
	Set gtmtypes("link_info",18,"len")=4
	Set gtmtypes("link_info",18,"type")="boolean_t"
	Set gtmtypfldindx("link_info","server_supports_std_null_coll")=18
	Set gtmtypes("link_info",19,"name")="link_info.server_supports_long_names"
	Set gtmtypes("link_info",19,"off")=72
	Set gtmtypes("link_info",19,"len")=4
	Set gtmtypes("link_info",19,"type")="boolean_t"
	Set gtmtypfldindx("link_info","server_supports_long_names")=19
	;
	Set gtmtypes("linkage_entry")="struct"
	Set gtmtypes("linkage_entry",0)=3
	Set gtmtypes("linkage_entry","len")=24
	Set gtmtypes("linkage_entry",1,"name")="linkage_entry.next"
	Set gtmtypes("linkage_entry",1,"off")=0
	Set gtmtypes("linkage_entry",1,"len")=8
	Set gtmtypes("linkage_entry",1,"type")="addr"
	Set gtmtypfldindx("linkage_entry","next")=1
	Set gtmtypes("linkage_entry",2,"name")="linkage_entry.symbol"
	Set gtmtypes("linkage_entry",2,"off")=8
	Set gtmtypes("linkage_entry",2,"len")=8
	Set gtmtypes("linkage_entry",2,"type")="addr"
	Set gtmtypfldindx("linkage_entry","symbol")=2
	Set gtmtypes("linkage_entry",3,"name")="linkage_entry.lit_offset"
	Set gtmtypes("linkage_entry",3,"off")=16
	Set gtmtypes("linkage_entry",3,"len")=8
	Set gtmtypes("linkage_entry",3,"type")="size_t"
	Set gtmtypfldindx("linkage_entry","lit_offset")=3
	;
	Set gtmtypes("list_params")="struct"
	Set gtmtypes("list_params",0)=4
	Set gtmtypes("list_params","len")=8
	Set gtmtypes("list_params",1,"name")="list_params.page"
	Set gtmtypes("list_params",1,"off")=0
	Set gtmtypes("list_params",1,"len")=2
	Set gtmtypes("list_params",1,"type")="unsigned-short"
	Set gtmtypfldindx("list_params","page")=1
	Set gtmtypes("list_params",2,"name")="list_params.list_line"
	Set gtmtypes("list_params",2,"off")=2
	Set gtmtypes("list_params",2,"len")=2
	Set gtmtypes("list_params",2,"type")="unsigned-short"
	Set gtmtypfldindx("list_params","list_line")=2
	Set gtmtypes("list_params",3,"name")="list_params.lines_per_page"
	Set gtmtypes("list_params",3,"off")=4
	Set gtmtypes("list_params",3,"len")=2
	Set gtmtypes("list_params",3,"type")="unsigned-short"
	Set gtmtypfldindx("list_params","lines_per_page")=3
	Set gtmtypes("list_params",4,"name")="list_params.space"
	Set gtmtypes("list_params",4,"off")=6
	Set gtmtypes("list_params",4,"len")=2
	Set gtmtypes("list_params",4,"type")="unsigned-short"
	Set gtmtypfldindx("list_params","space")=4
	;
	Set gtmtypes("lnk_tabent")="struct"
	Set gtmtypes("lnk_tabent",0)=1
	Set gtmtypes("lnk_tabent","len")=8
	Set gtmtypes("lnk_tabent",1,"name")="lnk_tabent.ext_ref"
	Set gtmtypes("lnk_tabent",1,"off")=0
	Set gtmtypes("lnk_tabent",1,"len")=8
	Set gtmtypes("lnk_tabent",1,"type")="addr"
	Set gtmtypfldindx("lnk_tabent","ext_ref")=1
	;
	Set gtmtypes("lnk_tabent_proxy")="struct"
	Set gtmtypes("lnk_tabent_proxy",0)=4
	Set gtmtypes("lnk_tabent_proxy","len")=24
	Set gtmtypes("lnk_tabent_proxy",1,"name")="lnk_tabent_proxy.rtnhdr_adr"
	Set gtmtypes("lnk_tabent_proxy",1,"off")=0
	Set gtmtypes("lnk_tabent_proxy",1,"len")=8
	Set gtmtypes("lnk_tabent_proxy",1,"type")="addr"
	Set gtmtypfldindx("lnk_tabent_proxy","rtnhdr_adr")=1
	Set gtmtypes("lnk_tabent_proxy",2,"name")="lnk_tabent_proxy.lnr_adr"
	Set gtmtypes("lnk_tabent_proxy",2,"off")=8
	Set gtmtypes("lnk_tabent_proxy",2,"len")=8
	Set gtmtypes("lnk_tabent_proxy",2,"type")="addr"
	Set gtmtypfldindx("lnk_tabent_proxy","lnr_adr")=2
	Set gtmtypes("lnk_tabent_proxy",3,"name")="lnk_tabent_proxy.has_parms"
	Set gtmtypes("lnk_tabent_proxy",3,"off")=16
	Set gtmtypes("lnk_tabent_proxy",3,"len")=4
	Set gtmtypes("lnk_tabent_proxy",3,"type")="boolean_t"
	Set gtmtypfldindx("lnk_tabent_proxy","has_parms")=3
	Set gtmtypes("lnk_tabent_proxy",4,"name")="lnk_tabent_proxy.filler1"
	Set gtmtypes("lnk_tabent_proxy",4,"off")=20
	Set gtmtypes("lnk_tabent_proxy",4,"len")=4
	Set gtmtypes("lnk_tabent_proxy",4,"type")="int"
	Set gtmtypfldindx("lnk_tabent_proxy","filler1")=4
	;
	Set gtmtypes("lockhist")="struct"
	Set gtmtypes("lockhist",0)=5
	Set gtmtypes("lockhist","len")=32
	Set gtmtypes("lockhist",1,"name")="lockhist.lock_addr"
	Set gtmtypes("lockhist",1,"off")=0
	Set gtmtypes("lockhist",1,"len")=8
	Set gtmtypes("lockhist",1,"type")="addr"
	Set gtmtypfldindx("lockhist","lock_addr")=1
	Set gtmtypes("lockhist",2,"name")="lockhist.lock_callr"
	Set gtmtypes("lockhist",2,"off")=8
	Set gtmtypes("lockhist",2,"len")=8
	Set gtmtypes("lockhist",2,"type")="caddr_t"
	Set gtmtypfldindx("lockhist","lock_callr")=2
	Set gtmtypes("lockhist",3,"name")="lockhist.lock_pid"
	Set gtmtypes("lockhist",3,"off")=16
	Set gtmtypes("lockhist",3,"len")=4
	Set gtmtypes("lockhist",3,"type")="int"
	Set gtmtypfldindx("lockhist","lock_pid")=3
	Set gtmtypes("lockhist",4,"name")="lockhist.loop_cnt"
	Set gtmtypes("lockhist",4,"off")=20
	Set gtmtypes("lockhist",4,"len")=4
	Set gtmtypes("lockhist",4,"type")="int"
	Set gtmtypfldindx("lockhist","loop_cnt")=4
	Set gtmtypes("lockhist",5,"name")="lockhist.lock_op"
	Set gtmtypes("lockhist",5,"off")=24
	Set gtmtypes("lockhist",5,"len")=4
	Set gtmtypes("lockhist",5,"type")="char"
	Set gtmtypfldindx("lockhist","lock_op")=5
	;
	Set gtmtypes("long_list")="struct"
	Set gtmtypes("long_list",0)=3
	Set gtmtypes("long_list","len")=16
	Set gtmtypes("long_list",1,"name")="long_list.next"
	Set gtmtypes("long_list",1,"off")=0
	Set gtmtypes("long_list",1,"len")=8
	Set gtmtypes("long_list",1,"type")="addr"
	Set gtmtypfldindx("long_list","next")=1
	Set gtmtypes("long_list",2,"name")="long_list.num"
	Set gtmtypes("long_list",2,"off")=8
	Set gtmtypes("long_list",2,"len")=4
	Set gtmtypes("long_list",2,"type")="unsigned-int"
	Set gtmtypfldindx("long_list","num")=2
	Set gtmtypes("long_list",3,"name")="long_list.exclude"
	Set gtmtypes("long_list",3,"off")=12
	Set gtmtypes("long_list",3,"len")=1
	Set gtmtypes("long_list",3,"type")="char"
	Set gtmtypfldindx("long_list","exclude")=3
	;
	Set gtmtypes("long_long_list")="struct"
	Set gtmtypes("long_long_list",0)=3
	Set gtmtypes("long_long_list","len")=24
	Set gtmtypes("long_long_list",1,"name")="long_long_list.next"
	Set gtmtypes("long_long_list",1,"off")=0
	Set gtmtypes("long_long_list",1,"len")=8
	Set gtmtypes("long_long_list",1,"type")="addr"
	Set gtmtypfldindx("long_long_list","next")=1
	Set gtmtypes("long_long_list",2,"name")="long_long_list.seqno"
	Set gtmtypes("long_long_list",2,"off")=8
	Set gtmtypes("long_long_list",2,"len")=8
	Set gtmtypes("long_long_list",2,"type")="uint64_t"
	Set gtmtypfldindx("long_long_list","seqno")=2
	Set gtmtypes("long_long_list",3,"name")="long_long_list.exclude"
	Set gtmtypes("long_long_list",3,"off")=16
	Set gtmtypes("long_long_list",3,"len")=4
	Set gtmtypes("long_long_list",3,"type")="boolean_t"
	Set gtmtypfldindx("long_long_list","exclude")=3
	;
	Set gtmtypes("lvTree")="struct"
	Set gtmtypes("lvTree",0)=10
	Set gtmtypes("lvTree","len")=56
	Set gtmtypes("lvTree",1,"name")="lvTree.ident"
	Set gtmtypes("lvTree",1,"off")=0
	Set gtmtypes("lvTree",1,"len")=2
	Set gtmtypes("lvTree",1,"type")="unsigned-short"
	Set gtmtypfldindx("lvTree","ident")=1
	Set gtmtypes("lvTree",2,"name")="lvTree.sbs_depth"
	Set gtmtypes("lvTree",2,"off")=2
	Set gtmtypes("lvTree",2,"len")=2
	Set gtmtypes("lvTree",2,"type")="unsigned-short"
	Set gtmtypfldindx("lvTree","sbs_depth")=2
	Set gtmtypes("lvTree",3,"name")="lvTree.avl_height"
	Set gtmtypes("lvTree",3,"off")=4
	Set gtmtypes("lvTree",3,"len")=4
	Set gtmtypes("lvTree",3,"type")="unsigned-int"
	Set gtmtypfldindx("lvTree","avl_height")=3
	Set gtmtypes("lvTree",4,"name")="lvTree.base_lv"
	Set gtmtypes("lvTree",4,"off")=8
	Set gtmtypes("lvTree",4,"len")=8
	Set gtmtypes("lvTree",4,"type")="addr"
	Set gtmtypfldindx("lvTree","base_lv")=4
	Set gtmtypes("lvTree",5,"name")="lvTree.avl_root"
	Set gtmtypes("lvTree",5,"off")=16
	Set gtmtypes("lvTree",5,"len")=8
	Set gtmtypes("lvTree",5,"type")="addr"
	Set gtmtypfldindx("lvTree","avl_root")=5
	Set gtmtypes("lvTree",6,"name")="lvTree.sbs_parent"
	Set gtmtypes("lvTree",6,"off")=24
	Set gtmtypes("lvTree",6,"len")=8
	Set gtmtypes("lvTree",6,"type")="addr"
	Set gtmtypfldindx("lvTree","sbs_parent")=6
	Set gtmtypes("lvTree",7,"name")="lvTree.lastLookup"
	Set gtmtypes("lvTree",7,"off")=32
	Set gtmtypes("lvTree",7,"len")=24
	Set gtmtypes("lvTree",7,"type")="treeSrchStatus"
	Set gtmtypfldindx("lvTree","lastLookup")=7
	Set gtmtypes("lvTree",8,"name")="lvTree.lastLookup.lastNodeLookedUp"
	Set gtmtypes("lvTree",8,"off")=32
	Set gtmtypes("lvTree",8,"len")=8
	Set gtmtypes("lvTree",8,"type")="addr"
	Set gtmtypfldindx("lvTree","lastLookup.lastNodeLookedUp")=8
	Set gtmtypes("lvTree",9,"name")="lvTree.lastLookup.lastNodeMin"
	Set gtmtypes("lvTree",9,"off")=40
	Set gtmtypes("lvTree",9,"len")=8
	Set gtmtypes("lvTree",9,"type")="addr"
	Set gtmtypfldindx("lvTree","lastLookup.lastNodeMin")=9
	Set gtmtypes("lvTree",10,"name")="lvTree.lastLookup.lastNodeMax"
	Set gtmtypes("lvTree",10,"off")=48
	Set gtmtypes("lvTree",10,"len")=8
	Set gtmtypes("lvTree",10,"type")="addr"
	Set gtmtypfldindx("lvTree","lastLookup.lastNodeMax")=10
	;
	Set gtmtypes("lvTreeNode")="struct"
	Set gtmtypes("lvTreeNode",0)=16
	Set gtmtypes("lvTreeNode","len")=88
	Set gtmtypes("lvTreeNode",1,"name")="lvTreeNode.v"
	Set gtmtypes("lvTreeNode",1,"off")=0
	Set gtmtypes("lvTreeNode",1,"len")=32
	Set gtmtypes("lvTreeNode",1,"type")="mval"
	Set gtmtypfldindx("lvTreeNode","v")=1
	Set gtmtypes("lvTreeNode",2,"name")="lvTreeNode.v.mvtype"
	Set gtmtypes("lvTreeNode",2,"off")=0
	Set gtmtypes("lvTreeNode",2,"len")=2
	Set gtmtypes("lvTreeNode",2,"type")="unsigned-short"
	Set gtmtypfldindx("lvTreeNode","v.mvtype")=2
	Set gtmtypes("lvTreeNode",3,"name")="lvTreeNode.v.fnpc_indx"
	Set gtmtypes("lvTreeNode",3,"off")=3
	Set gtmtypes("lvTreeNode",3,"len")=1
	Set gtmtypes("lvTreeNode",3,"type")="unsigned-char"
	Set gtmtypfldindx("lvTreeNode","v.fnpc_indx")=3
	Set gtmtypes("lvTreeNode",4,"name")="lvTreeNode.v.utfcgr_indx"
	Set gtmtypes("lvTreeNode",4,"off")=4
	Set gtmtypes("lvTreeNode",4,"len")=4
	Set gtmtypes("lvTreeNode",4,"type")="unsigned-int"
	Set gtmtypfldindx("lvTreeNode","v.utfcgr_indx")=4
	Set gtmtypes("lvTreeNode",5,"name")="lvTreeNode.v.m"
	Set gtmtypes("lvTreeNode",5,"off")=8
	Set gtmtypes("lvTreeNode",5,"len")=8
	Set gtmtypes("lvTreeNode",5,"type")="int"
	Set gtmtypfldindx("lvTreeNode","v.m")=5
	Set gtmtypes("lvTreeNode",5,"dim")=2
	Set gtmtypes("lvTreeNode",6,"name")="lvTreeNode.v.str"
	Set gtmtypes("lvTreeNode",6,"off")=16
	Set gtmtypes("lvTreeNode",6,"len")=16
	Set gtmtypes("lvTreeNode",6,"type")="mstr"
	Set gtmtypfldindx("lvTreeNode","v.str")=6
	Set gtmtypes("lvTreeNode",7,"name")="lvTreeNode.sbs_child"
	Set gtmtypes("lvTreeNode",7,"off")=32
	Set gtmtypes("lvTreeNode",7,"len")=8
	Set gtmtypes("lvTreeNode",7,"type")="addr"
	Set gtmtypfldindx("lvTreeNode","sbs_child")=7
	Set gtmtypes("lvTreeNode",8,"name")="lvTreeNode.tree_parent"
	Set gtmtypes("lvTreeNode",8,"off")=40
	Set gtmtypes("lvTreeNode",8,"len")=8
	Set gtmtypes("lvTreeNode",8,"type")="addr"
	Set gtmtypfldindx("lvTreeNode","tree_parent")=8
	Set gtmtypes("lvTreeNode",9,"name")="lvTreeNode.key_mvtype"
	Set gtmtypes("lvTreeNode",9,"off")=48
	Set gtmtypes("lvTreeNode",9,"len")=2
	Set gtmtypes("lvTreeNode",9,"type")="unsigned-short"
	Set gtmtypfldindx("lvTreeNode","key_mvtype")=9
	Set gtmtypes("lvTreeNode",10,"name")="lvTreeNode.balance"
	Set gtmtypes("lvTreeNode",10,"off")=50
	Set gtmtypes("lvTreeNode",10,"len")=1
	Set gtmtypes("lvTreeNode",10,"type")="char"
	Set gtmtypfldindx("lvTreeNode","balance")=10
	Set gtmtypes("lvTreeNode",11,"name")="lvTreeNode.descent_dir"
	Set gtmtypes("lvTreeNode",11,"off")=51
	Set gtmtypes("lvTreeNode",11,"len")=1
	Set gtmtypes("lvTreeNode",11,"type")="unsigned-char"
	Set gtmtypfldindx("lvTreeNode","descent_dir")=11
	Set gtmtypes("lvTreeNode",12,"name")="lvTreeNode.key_len"
	Set gtmtypes("lvTreeNode",12,"off")=52
	Set gtmtypes("lvTreeNode",12,"len")=4
	Set gtmtypes("lvTreeNode",12,"type")="unsigned-int"
	Set gtmtypfldindx("lvTreeNode","key_len")=12
	Set gtmtypes("lvTreeNode",13,"name")="lvTreeNode.key_addr"
	Set gtmtypes("lvTreeNode",13,"off")=56
	Set gtmtypes("lvTreeNode",13,"len")=8
	Set gtmtypes("lvTreeNode",13,"type")="addr"
	Set gtmtypfldindx("lvTreeNode","key_addr")=13
	Set gtmtypes("lvTreeNode",14,"name")="lvTreeNode.avl_left"
	Set gtmtypes("lvTreeNode",14,"off")=64
	Set gtmtypes("lvTreeNode",14,"len")=8
	Set gtmtypes("lvTreeNode",14,"type")="addr"
	Set gtmtypfldindx("lvTreeNode","avl_left")=14
	Set gtmtypes("lvTreeNode",15,"name")="lvTreeNode.avl_right"
	Set gtmtypes("lvTreeNode",15,"off")=72
	Set gtmtypes("lvTreeNode",15,"len")=8
	Set gtmtypes("lvTreeNode",15,"type")="addr"
	Set gtmtypfldindx("lvTreeNode","avl_right")=15
	Set gtmtypes("lvTreeNode",16,"name")="lvTreeNode.avl_parent"
	Set gtmtypes("lvTreeNode",16,"off")=80
	Set gtmtypes("lvTreeNode",16,"len")=8
	Set gtmtypes("lvTreeNode",16,"type")="addr"
	Set gtmtypfldindx("lvTreeNode","avl_parent")=16
	;
	Set gtmtypes("lvTreeNodeNum")="struct"
	Set gtmtypes("lvTreeNodeNum",0)=20
	Set gtmtypes("lvTreeNodeNum","len")=88
	Set gtmtypes("lvTreeNodeNum",1,"name")="lvTreeNodeNum.v"
	Set gtmtypes("lvTreeNodeNum",1,"off")=0
	Set gtmtypes("lvTreeNodeNum",1,"len")=32
	Set gtmtypes("lvTreeNodeNum",1,"type")="mval"
	Set gtmtypfldindx("lvTreeNodeNum","v")=1
	Set gtmtypes("lvTreeNodeNum",2,"name")="lvTreeNodeNum.v.mvtype"
	Set gtmtypes("lvTreeNodeNum",2,"off")=0
	Set gtmtypes("lvTreeNodeNum",2,"len")=2
	Set gtmtypes("lvTreeNodeNum",2,"type")="unsigned-short"
	Set gtmtypfldindx("lvTreeNodeNum","v.mvtype")=2
	Set gtmtypes("lvTreeNodeNum",3,"name")="lvTreeNodeNum.v.fnpc_indx"
	Set gtmtypes("lvTreeNodeNum",3,"off")=3
	Set gtmtypes("lvTreeNodeNum",3,"len")=1
	Set gtmtypes("lvTreeNodeNum",3,"type")="unsigned-char"
	Set gtmtypfldindx("lvTreeNodeNum","v.fnpc_indx")=3
	Set gtmtypes("lvTreeNodeNum",4,"name")="lvTreeNodeNum.v.utfcgr_indx"
	Set gtmtypes("lvTreeNodeNum",4,"off")=4
	Set gtmtypes("lvTreeNodeNum",4,"len")=4
	Set gtmtypes("lvTreeNodeNum",4,"type")="unsigned-int"
	Set gtmtypfldindx("lvTreeNodeNum","v.utfcgr_indx")=4
	Set gtmtypes("lvTreeNodeNum",5,"name")="lvTreeNodeNum.v.m"
	Set gtmtypes("lvTreeNodeNum",5,"off")=8
	Set gtmtypes("lvTreeNodeNum",5,"len")=8
	Set gtmtypes("lvTreeNodeNum",5,"type")="int"
	Set gtmtypfldindx("lvTreeNodeNum","v.m")=5
	Set gtmtypes("lvTreeNodeNum",5,"dim")=2
	Set gtmtypes("lvTreeNodeNum",6,"name")="lvTreeNodeNum.v.str"
	Set gtmtypes("lvTreeNodeNum",6,"off")=16
	Set gtmtypes("lvTreeNodeNum",6,"len")=16
	Set gtmtypes("lvTreeNodeNum",6,"type")="mstr"
	Set gtmtypfldindx("lvTreeNodeNum","v.str")=6
	Set gtmtypes("lvTreeNodeNum",7,"name")="lvTreeNodeNum.sbs_child"
	Set gtmtypes("lvTreeNodeNum",7,"off")=32
	Set gtmtypes("lvTreeNodeNum",7,"len")=8
	Set gtmtypes("lvTreeNodeNum",7,"type")="addr"
	Set gtmtypfldindx("lvTreeNodeNum","sbs_child")=7
	Set gtmtypes("lvTreeNodeNum",8,"name")="lvTreeNodeNum.tree_parent"
	Set gtmtypes("lvTreeNodeNum",8,"off")=40
	Set gtmtypes("lvTreeNodeNum",8,"len")=8
	Set gtmtypes("lvTreeNodeNum",8,"type")="addr"
	Set gtmtypfldindx("lvTreeNodeNum","tree_parent")=8
	Set gtmtypes("lvTreeNodeNum",9,"name")="lvTreeNodeNum.key_mvtype"
	Set gtmtypes("lvTreeNodeNum",9,"off")=48
	Set gtmtypes("lvTreeNodeNum",9,"len")=2
	Set gtmtypes("lvTreeNodeNum",9,"type")="unsigned-short"
	Set gtmtypfldindx("lvTreeNodeNum","key_mvtype")=9
	Set gtmtypes("lvTreeNodeNum",10,"name")="lvTreeNodeNum.balance"
	Set gtmtypes("lvTreeNodeNum",10,"off")=50
	Set gtmtypes("lvTreeNodeNum",10,"len")=1
	Set gtmtypes("lvTreeNodeNum",10,"type")="char"
	Set gtmtypfldindx("lvTreeNodeNum","balance")=10
	Set gtmtypes("lvTreeNodeNum",11,"name")="lvTreeNodeNum.descent_dir"
	Set gtmtypes("lvTreeNodeNum",11,"off")=51
	Set gtmtypes("lvTreeNodeNum",11,"len")=1
	Set gtmtypes("lvTreeNodeNum",11,"type")="unsigned-char"
	Set gtmtypfldindx("lvTreeNodeNum","descent_dir")=11
	Set gtmtypes("lvTreeNodeNum",12,"name")="lvTreeNodeNum.key_flags"
	Set gtmtypes("lvTreeNodeNum",12,"off")=52
	Set gtmtypes("lvTreeNodeNum",12,"len")=4
	Set gtmtypes("lvTreeNodeNum",12,"type")="union"
	Set gtmtypfldindx("lvTreeNodeNum","key_flags")=12
	Set gtmtypes("lvTreeNodeNum",13,"name")="lvTreeNodeNum.key_flags.key_bytes"
	Set gtmtypes("lvTreeNodeNum",13,"off")=52
	Set gtmtypes("lvTreeNodeNum",13,"len")=1
	Set gtmtypes("lvTreeNodeNum",13,"type")="struct"
	Set gtmtypfldindx("lvTreeNodeNum","key_flags.key_bytes")=13
	Set gtmtypes("lvTreeNodeNum",14,"name")="lvTreeNodeNum.key_flags.key_bytes.key_sgne"
	Set gtmtypes("lvTreeNodeNum",14,"off")=52
	Set gtmtypes("lvTreeNodeNum",14,"len")=1
	Set gtmtypes("lvTreeNodeNum",14,"type")="unsigned-char"
	Set gtmtypfldindx("lvTreeNodeNum","key_flags.key_bytes.key_sgne")=14
	Set gtmtypes("lvTreeNodeNum",15,"name")="lvTreeNodeNum.key_flags.key_bits"
	Set gtmtypes("lvTreeNodeNum",15,"off")=52
	Set gtmtypes("lvTreeNodeNum",15,"len")=4
	Set gtmtypes("lvTreeNodeNum",15,"type")="struct"
	Set gtmtypfldindx("lvTreeNodeNum","key_flags.key_bits")=15
	Set gtmtypes("lvTreeNodeNum",16,"name")="lvTreeNodeNum.key_m0"
	Set gtmtypes("lvTreeNodeNum",16,"off")=56
	Set gtmtypes("lvTreeNodeNum",16,"len")=4
	Set gtmtypes("lvTreeNodeNum",16,"type")="int"
	Set gtmtypfldindx("lvTreeNodeNum","key_m0")=16
	Set gtmtypes("lvTreeNodeNum",17,"name")="lvTreeNodeNum.key_m1"
	Set gtmtypes("lvTreeNodeNum",17,"off")=60
	Set gtmtypes("lvTreeNodeNum",17,"len")=4
	Set gtmtypes("lvTreeNodeNum",17,"type")="int"
	Set gtmtypfldindx("lvTreeNodeNum","key_m1")=17
	Set gtmtypes("lvTreeNodeNum",18,"name")="lvTreeNodeNum.avl_left"
	Set gtmtypes("lvTreeNodeNum",18,"off")=64
	Set gtmtypes("lvTreeNodeNum",18,"len")=8
	Set gtmtypes("lvTreeNodeNum",18,"type")="addr"
	Set gtmtypfldindx("lvTreeNodeNum","avl_left")=18
	Set gtmtypes("lvTreeNodeNum",19,"name")="lvTreeNodeNum.avl_right"
	Set gtmtypes("lvTreeNodeNum",19,"off")=72
	Set gtmtypes("lvTreeNodeNum",19,"len")=8
	Set gtmtypes("lvTreeNodeNum",19,"type")="addr"
	Set gtmtypfldindx("lvTreeNodeNum","avl_right")=19
	Set gtmtypes("lvTreeNodeNum",20,"name")="lvTreeNodeNum.avl_parent"
	Set gtmtypes("lvTreeNodeNum",20,"off")=80
	Set gtmtypes("lvTreeNodeNum",20,"len")=8
	Set gtmtypes("lvTreeNodeNum",20,"type")="addr"
	Set gtmtypfldindx("lvTreeNodeNum","avl_parent")=20
	;
	Set gtmtypes("lv_blk")="struct"
	Set gtmtypes("lv_blk",0)=3
	Set gtmtypes("lv_blk","len")=16
	Set gtmtypes("lv_blk",1,"name")="lv_blk.next"
	Set gtmtypes("lv_blk",1,"off")=0
	Set gtmtypes("lv_blk",1,"len")=8
	Set gtmtypes("lv_blk",1,"type")="addr"
	Set gtmtypfldindx("lv_blk","next")=1
	Set gtmtypes("lv_blk",2,"name")="lv_blk.numAlloc"
	Set gtmtypes("lv_blk",2,"off")=8
	Set gtmtypes("lv_blk",2,"len")=4
	Set gtmtypes("lv_blk",2,"type")="unsigned-int"
	Set gtmtypfldindx("lv_blk","numAlloc")=2
	Set gtmtypes("lv_blk",3,"name")="lv_blk.numUsed"
	Set gtmtypes("lv_blk",3,"off")=12
	Set gtmtypes("lv_blk",3,"len")=4
	Set gtmtypes("lv_blk",3,"type")="unsigned-int"
	Set gtmtypfldindx("lv_blk","numUsed")=3
	;
	Set gtmtypes("lv_val")="struct"
	Set gtmtypes("lv_val",0)=24
	Set gtmtypes("lv_val","len")=80
	Set gtmtypes("lv_val",1,"name")="lv_val.v"
	Set gtmtypes("lv_val",1,"off")=0
	Set gtmtypes("lv_val",1,"len")=32
	Set gtmtypes("lv_val",1,"type")="mval"
	Set gtmtypfldindx("lv_val","v")=1
	Set gtmtypes("lv_val",2,"name")="lv_val.v.mvtype"
	Set gtmtypes("lv_val",2,"off")=0
	Set gtmtypes("lv_val",2,"len")=2
	Set gtmtypes("lv_val",2,"type")="unsigned-short"
	Set gtmtypfldindx("lv_val","v.mvtype")=2
	Set gtmtypes("lv_val",3,"name")="lv_val.v.fnpc_indx"
	Set gtmtypes("lv_val",3,"off")=3
	Set gtmtypes("lv_val",3,"len")=1
	Set gtmtypes("lv_val",3,"type")="unsigned-char"
	Set gtmtypfldindx("lv_val","v.fnpc_indx")=3
	Set gtmtypes("lv_val",4,"name")="lv_val.v.utfcgr_indx"
	Set gtmtypes("lv_val",4,"off")=4
	Set gtmtypes("lv_val",4,"len")=4
	Set gtmtypes("lv_val",4,"type")="unsigned-int"
	Set gtmtypfldindx("lv_val","v.utfcgr_indx")=4
	Set gtmtypes("lv_val",5,"name")="lv_val.v.m"
	Set gtmtypes("lv_val",5,"off")=8
	Set gtmtypes("lv_val",5,"len")=8
	Set gtmtypes("lv_val",5,"type")="int"
	Set gtmtypfldindx("lv_val","v.m")=5
	Set gtmtypes("lv_val",5,"dim")=2
	Set gtmtypes("lv_val",6,"name")="lv_val.v.str"
	Set gtmtypes("lv_val",6,"off")=16
	Set gtmtypes("lv_val",6,"len")=16
	Set gtmtypes("lv_val",6,"type")="mstr"
	Set gtmtypfldindx("lv_val","v.str")=6
	Set gtmtypes("lv_val",7,"name")="lv_val.ptrs"
	Set gtmtypes("lv_val",7,"off")=32
	Set gtmtypes("lv_val",7,"len")=16
	Set gtmtypes("lv_val",7,"type")="union"
	Set gtmtypfldindx("lv_val","ptrs")=7
	Set gtmtypes("lv_val",8,"name")="lv_val.ptrs.val_ent"
	Set gtmtypes("lv_val",8,"off")=32
	Set gtmtypes("lv_val",8,"len")=16
	Set gtmtypes("lv_val",8,"type")="struct"
	Set gtmtypfldindx("lv_val","ptrs.val_ent")=8
	Set gtmtypes("lv_val",9,"name")="lv_val.ptrs.val_ent.children"
	Set gtmtypes("lv_val",9,"off")=32
	Set gtmtypes("lv_val",9,"len")=8
	Set gtmtypes("lv_val",9,"type")="addr"
	Set gtmtypfldindx("lv_val","ptrs.val_ent.children")=9
	Set gtmtypes("lv_val",10,"name")="lv_val.ptrs.val_ent.parent"
	Set gtmtypes("lv_val",10,"off")=40
	Set gtmtypes("lv_val",10,"len")=8
	Set gtmtypes("lv_val",10,"type")="union"
	Set gtmtypfldindx("lv_val","ptrs.val_ent.parent")=10
	Set gtmtypes("lv_val",11,"name")="lv_val.ptrs.val_ent.parent.sym"
	Set gtmtypes("lv_val",11,"off")=40
	Set gtmtypes("lv_val",11,"len")=8
	Set gtmtypes("lv_val",11,"type")="addr"
	Set gtmtypfldindx("lv_val","ptrs.val_ent.parent.sym")=11
	Set gtmtypes("lv_val",12,"name")="lv_val.ptrs.val_ent.parent.sbs_tree"
	Set gtmtypes("lv_val",12,"off")=40
	Set gtmtypes("lv_val",12,"len")=8
	Set gtmtypes("lv_val",12,"type")="addr"
	Set gtmtypfldindx("lv_val","ptrs.val_ent.parent.sbs_tree")=12
	Set gtmtypes("lv_val",13,"name")="lv_val.ptrs.free_ent"
	Set gtmtypes("lv_val",13,"off")=32
	Set gtmtypes("lv_val",13,"len")=8
	Set gtmtypes("lv_val",13,"type")="struct"
	Set gtmtypfldindx("lv_val","ptrs.free_ent")=13
	Set gtmtypes("lv_val",14,"name")="lv_val.ptrs.free_ent.next_free"
	Set gtmtypes("lv_val",14,"off")=32
	Set gtmtypes("lv_val",14,"len")=8
	Set gtmtypes("lv_val",14,"type")="addr"
	Set gtmtypfldindx("lv_val","ptrs.free_ent.next_free")=14
	Set gtmtypes("lv_val",15,"name")="lv_val.ptrs.copy_loc"
	Set gtmtypes("lv_val",15,"off")=32
	Set gtmtypes("lv_val",15,"len")=8
	Set gtmtypes("lv_val",15,"type")="struct"
	Set gtmtypfldindx("lv_val","ptrs.copy_loc")=15
	Set gtmtypes("lv_val",16,"name")="lv_val.ptrs.copy_loc.newtablv"
	Set gtmtypes("lv_val",16,"off")=32
	Set gtmtypes("lv_val",16,"len")=8
	Set gtmtypes("lv_val",16,"type")="addr"
	Set gtmtypfldindx("lv_val","ptrs.copy_loc.newtablv")=16
	Set gtmtypes("lv_val",17,"name")="lv_val.stats"
	Set gtmtypes("lv_val",17,"off")=48
	Set gtmtypes("lv_val",17,"len")=16
	Set gtmtypes("lv_val",17,"type")="struct"
	Set gtmtypfldindx("lv_val","stats")=17
	Set gtmtypes("lv_val",18,"name")="lv_val.stats.trefcnt"
	Set gtmtypes("lv_val",18,"off")=48
	Set gtmtypes("lv_val",18,"len")=4
	Set gtmtypes("lv_val",18,"type")="int"
	Set gtmtypfldindx("lv_val","stats.trefcnt")=18
	Set gtmtypes("lv_val",19,"name")="lv_val.stats.crefcnt"
	Set gtmtypes("lv_val",19,"off")=52
	Set gtmtypes("lv_val",19,"len")=4
	Set gtmtypes("lv_val",19,"type")="int"
	Set gtmtypfldindx("lv_val","stats.crefcnt")=19
	Set gtmtypes("lv_val",20,"name")="lv_val.stats.tstartcycle"
	Set gtmtypes("lv_val",20,"off")=56
	Set gtmtypes("lv_val",20,"len")=4
	Set gtmtypes("lv_val",20,"type")="unsigned-int"
	Set gtmtypfldindx("lv_val","stats.tstartcycle")=20
	Set gtmtypes("lv_val",21,"name")="lv_val.stats.lvtaskcycle"
	Set gtmtypes("lv_val",21,"off")=60
	Set gtmtypes("lv_val",21,"len")=4
	Set gtmtypes("lv_val",21,"type")="unsigned-int"
	Set gtmtypfldindx("lv_val","stats.lvtaskcycle")=21
	Set gtmtypes("lv_val",22,"name")="lv_val.has_aliascont"
	Set gtmtypes("lv_val",22,"off")=64
	Set gtmtypes("lv_val",22,"len")=4
	Set gtmtypes("lv_val",22,"type")="boolean_t"
	Set gtmtypfldindx("lv_val","has_aliascont")=22
	Set gtmtypes("lv_val",23,"name")="lv_val.lvmon_mark"
	Set gtmtypes("lv_val",23,"off")=68
	Set gtmtypes("lv_val",23,"len")=4
	Set gtmtypes("lv_val",23,"type")="boolean_t"
	Set gtmtypfldindx("lv_val","lvmon_mark")=23
	Set gtmtypes("lv_val",24,"name")="lv_val.tp_var"
	Set gtmtypes("lv_val",24,"off")=72
	Set gtmtypes("lv_val",24,"len")=8
	Set gtmtypes("lv_val",24,"type")="addr"
	Set gtmtypfldindx("lv_val","tp_var")=24
	;
	Set gtmtypes("lv_xnew_ref")="struct"
	Set gtmtypes("lv_xnew_ref",0)=2
	Set gtmtypes("lv_xnew_ref","len")=16
	Set gtmtypes("lv_xnew_ref",1,"name")="lv_xnew_ref.next"
	Set gtmtypes("lv_xnew_ref",1,"off")=0
	Set gtmtypes("lv_xnew_ref",1,"len")=8
	Set gtmtypes("lv_xnew_ref",1,"type")="addr"
	Set gtmtypfldindx("lv_xnew_ref","next")=1
	Set gtmtypes("lv_xnew_ref",2,"name")="lv_xnew_ref.lvval"
	Set gtmtypes("lv_xnew_ref",2,"off")=8
	Set gtmtypes("lv_xnew_ref",2,"len")=8
	Set gtmtypes("lv_xnew_ref",2,"type")="addr"
	Set gtmtypfldindx("lv_xnew_ref","lvval")=2
	;
	Set gtmtypes("lv_xnew_var")="struct"
	Set gtmtypes("lv_xnew_var",0)=6
	Set gtmtypes("lv_xnew_var","len")=40
	Set gtmtypes("lv_xnew_var",1,"name")="lv_xnew_var.next"
	Set gtmtypes("lv_xnew_var",1,"off")=0
	Set gtmtypes("lv_xnew_var",1,"len")=8
	Set gtmtypes("lv_xnew_var",1,"type")="addr"
	Set gtmtypfldindx("lv_xnew_var","next")=1
	Set gtmtypes("lv_xnew_var",2,"name")="lv_xnew_var.key"
	Set gtmtypes("lv_xnew_var",2,"off")=8
	Set gtmtypes("lv_xnew_var",2,"len")=24
	Set gtmtypes("lv_xnew_var",2,"type")="mname_entry"
	Set gtmtypfldindx("lv_xnew_var","key")=2
	Set gtmtypes("lv_xnew_var",3,"name")="lv_xnew_var.key.var_name"
	Set gtmtypes("lv_xnew_var",3,"off")=8
	Set gtmtypes("lv_xnew_var",3,"len")=16
	Set gtmtypes("lv_xnew_var",3,"type")="mstr"
	Set gtmtypfldindx("lv_xnew_var","key.var_name")=3
	Set gtmtypes("lv_xnew_var",4,"name")="lv_xnew_var.key.hash_code"
	Set gtmtypes("lv_xnew_var",4,"off")=24
	Set gtmtypes("lv_xnew_var",4,"len")=4
	Set gtmtypes("lv_xnew_var",4,"type")="unsigned-int"
	Set gtmtypfldindx("lv_xnew_var","key.hash_code")=4
	Set gtmtypes("lv_xnew_var",5,"name")="lv_xnew_var.key.marked"
	Set gtmtypes("lv_xnew_var",5,"off")=28
	Set gtmtypes("lv_xnew_var",5,"len")=4
	Set gtmtypes("lv_xnew_var",5,"type")="boolean_t"
	Set gtmtypfldindx("lv_xnew_var","key.marked")=5
	Set gtmtypes("lv_xnew_var",6,"name")="lv_xnew_var.lvval"
	Set gtmtypes("lv_xnew_var",6,"off")=32
	Set gtmtypes("lv_xnew_var",6,"len")=8
	Set gtmtypes("lv_xnew_var",6,"type")="addr"
	Set gtmtypfldindx("lv_xnew_var","lvval")=6
	;
	Set gtmtypes("lvmon_value_ent")="struct"
	Set gtmtypes("lvmon_value_ent",0)=31
	Set gtmtypes("lvmon_value_ent","len")=104
	Set gtmtypes("lvmon_value_ent",1,"name")="lvmon_value_ent.varlvval"
	Set gtmtypes("lvmon_value_ent",1,"off")=0
	Set gtmtypes("lvmon_value_ent",1,"len")=80
	Set gtmtypes("lvmon_value_ent",1,"type")="lv_val"
	Set gtmtypfldindx("lvmon_value_ent","varlvval")=1
	Set gtmtypes("lvmon_value_ent",2,"name")="lvmon_value_ent.varlvval.v"
	Set gtmtypes("lvmon_value_ent",2,"off")=0
	Set gtmtypes("lvmon_value_ent",2,"len")=32
	Set gtmtypes("lvmon_value_ent",2,"type")="mval"
	Set gtmtypfldindx("lvmon_value_ent","varlvval.v")=2
	Set gtmtypes("lvmon_value_ent",3,"name")="lvmon_value_ent.varlvval.v.mvtype"
	Set gtmtypes("lvmon_value_ent",3,"off")=0
	Set gtmtypes("lvmon_value_ent",3,"len")=2
	Set gtmtypes("lvmon_value_ent",3,"type")="unsigned-short"
	Set gtmtypfldindx("lvmon_value_ent","varlvval.v.mvtype")=3
	Set gtmtypes("lvmon_value_ent",4,"name")="lvmon_value_ent.varlvval.v.fnpc_indx"
	Set gtmtypes("lvmon_value_ent",4,"off")=3
	Set gtmtypes("lvmon_value_ent",4,"len")=1
	Set gtmtypes("lvmon_value_ent",4,"type")="unsigned-char"
	Set gtmtypfldindx("lvmon_value_ent","varlvval.v.fnpc_indx")=4
	Set gtmtypes("lvmon_value_ent",5,"name")="lvmon_value_ent.varlvval.v.utfcgr_indx"
	Set gtmtypes("lvmon_value_ent",5,"off")=4
	Set gtmtypes("lvmon_value_ent",5,"len")=4
	Set gtmtypes("lvmon_value_ent",5,"type")="unsigned-int"
	Set gtmtypfldindx("lvmon_value_ent","varlvval.v.utfcgr_indx")=5
	Set gtmtypes("lvmon_value_ent",6,"name")="lvmon_value_ent.varlvval.v.m"
	Set gtmtypes("lvmon_value_ent",6,"off")=8
	Set gtmtypes("lvmon_value_ent",6,"len")=8
	Set gtmtypes("lvmon_value_ent",6,"type")="int"
	Set gtmtypfldindx("lvmon_value_ent","varlvval.v.m")=6
	Set gtmtypes("lvmon_value_ent",6,"dim")=2
	Set gtmtypes("lvmon_value_ent",7,"name")="lvmon_value_ent.varlvval.v.str"
	Set gtmtypes("lvmon_value_ent",7,"off")=16
	Set gtmtypes("lvmon_value_ent",7,"len")=16
	Set gtmtypes("lvmon_value_ent",7,"type")="mstr"
	Set gtmtypfldindx("lvmon_value_ent","varlvval.v.str")=7
	Set gtmtypes("lvmon_value_ent",8,"name")="lvmon_value_ent.varlvval.ptrs"
	Set gtmtypes("lvmon_value_ent",8,"off")=32
	Set gtmtypes("lvmon_value_ent",8,"len")=16
	Set gtmtypes("lvmon_value_ent",8,"type")="union"
	Set gtmtypfldindx("lvmon_value_ent","varlvval.ptrs")=8
	Set gtmtypes("lvmon_value_ent",9,"name")="lvmon_value_ent.varlvval.ptrs.val_ent"
	Set gtmtypes("lvmon_value_ent",9,"off")=32
	Set gtmtypes("lvmon_value_ent",9,"len")=16
	Set gtmtypes("lvmon_value_ent",9,"type")="struct"
	Set gtmtypfldindx("lvmon_value_ent","varlvval.ptrs.val_ent")=9
	Set gtmtypes("lvmon_value_ent",10,"name")="lvmon_value_ent.varlvval.ptrs.val_ent.children"
	Set gtmtypes("lvmon_value_ent",10,"off")=32
	Set gtmtypes("lvmon_value_ent",10,"len")=8
	Set gtmtypes("lvmon_value_ent",10,"type")="addr"
	Set gtmtypfldindx("lvmon_value_ent","varlvval.ptrs.val_ent.children")=10
	Set gtmtypes("lvmon_value_ent",11,"name")="lvmon_value_ent.varlvval.ptrs.val_ent.parent"
	Set gtmtypes("lvmon_value_ent",11,"off")=40
	Set gtmtypes("lvmon_value_ent",11,"len")=8
	Set gtmtypes("lvmon_value_ent",11,"type")="union"
	Set gtmtypfldindx("lvmon_value_ent","varlvval.ptrs.val_ent.parent")=11
	Set gtmtypes("lvmon_value_ent",12,"name")="lvmon_value_ent.varlvval.ptrs.val_ent.parent.sym"
	Set gtmtypes("lvmon_value_ent",12,"off")=40
	Set gtmtypes("lvmon_value_ent",12,"len")=8
	Set gtmtypes("lvmon_value_ent",12,"type")="addr"
	Set gtmtypfldindx("lvmon_value_ent","varlvval.ptrs.val_ent.parent.sym")=12
	Set gtmtypes("lvmon_value_ent",13,"name")="lvmon_value_ent.varlvval.ptrs.val_ent.parent.sbs_tree"
	Set gtmtypes("lvmon_value_ent",13,"off")=40
	Set gtmtypes("lvmon_value_ent",13,"len")=8
	Set gtmtypes("lvmon_value_ent",13,"type")="addr"
	Set gtmtypfldindx("lvmon_value_ent","varlvval.ptrs.val_ent.parent.sbs_tree")=13
	Set gtmtypes("lvmon_value_ent",14,"name")="lvmon_value_ent.varlvval.ptrs.free_ent"
	Set gtmtypes("lvmon_value_ent",14,"off")=32
	Set gtmtypes("lvmon_value_ent",14,"len")=8
	Set gtmtypes("lvmon_value_ent",14,"type")="struct"
	Set gtmtypfldindx("lvmon_value_ent","varlvval.ptrs.free_ent")=14
	Set gtmtypes("lvmon_value_ent",15,"name")="lvmon_value_ent.varlvval.ptrs.free_ent.next_free"
	Set gtmtypes("lvmon_value_ent",15,"off")=32
	Set gtmtypes("lvmon_value_ent",15,"len")=8
	Set gtmtypes("lvmon_value_ent",15,"type")="addr"
	Set gtmtypfldindx("lvmon_value_ent","varlvval.ptrs.free_ent.next_free")=15
	Set gtmtypes("lvmon_value_ent",16,"name")="lvmon_value_ent.varlvval.ptrs.copy_loc"
	Set gtmtypes("lvmon_value_ent",16,"off")=32
	Set gtmtypes("lvmon_value_ent",16,"len")=8
	Set gtmtypes("lvmon_value_ent",16,"type")="struct"
	Set gtmtypfldindx("lvmon_value_ent","varlvval.ptrs.copy_loc")=16
	Set gtmtypes("lvmon_value_ent",17,"name")="lvmon_value_ent.varlvval.ptrs.copy_loc.newtablv"
	Set gtmtypes("lvmon_value_ent",17,"off")=32
	Set gtmtypes("lvmon_value_ent",17,"len")=8
	Set gtmtypes("lvmon_value_ent",17,"type")="addr"
	Set gtmtypfldindx("lvmon_value_ent","varlvval.ptrs.copy_loc.newtablv")=17
	Set gtmtypes("lvmon_value_ent",18,"name")="lvmon_value_ent.varlvval.stats"
	Set gtmtypes("lvmon_value_ent",18,"off")=48
	Set gtmtypes("lvmon_value_ent",18,"len")=16
	Set gtmtypes("lvmon_value_ent",18,"type")="struct"
	Set gtmtypfldindx("lvmon_value_ent","varlvval.stats")=18
	Set gtmtypes("lvmon_value_ent",19,"name")="lvmon_value_ent.varlvval.stats.trefcnt"
	Set gtmtypes("lvmon_value_ent",19,"off")=48
	Set gtmtypes("lvmon_value_ent",19,"len")=4
	Set gtmtypes("lvmon_value_ent",19,"type")="int"
	Set gtmtypfldindx("lvmon_value_ent","varlvval.stats.trefcnt")=19
	Set gtmtypes("lvmon_value_ent",20,"name")="lvmon_value_ent.varlvval.stats.crefcnt"
	Set gtmtypes("lvmon_value_ent",20,"off")=52
	Set gtmtypes("lvmon_value_ent",20,"len")=4
	Set gtmtypes("lvmon_value_ent",20,"type")="int"
	Set gtmtypfldindx("lvmon_value_ent","varlvval.stats.crefcnt")=20
	Set gtmtypes("lvmon_value_ent",21,"name")="lvmon_value_ent.varlvval.stats.tstartcycle"
	Set gtmtypes("lvmon_value_ent",21,"off")=56
	Set gtmtypes("lvmon_value_ent",21,"len")=4
	Set gtmtypes("lvmon_value_ent",21,"type")="unsigned-int"
	Set gtmtypfldindx("lvmon_value_ent","varlvval.stats.tstartcycle")=21
	Set gtmtypes("lvmon_value_ent",22,"name")="lvmon_value_ent.varlvval.stats.lvtaskcycle"
	Set gtmtypes("lvmon_value_ent",22,"off")=60
	Set gtmtypes("lvmon_value_ent",22,"len")=4
	Set gtmtypes("lvmon_value_ent",22,"type")="unsigned-int"
	Set gtmtypfldindx("lvmon_value_ent","varlvval.stats.lvtaskcycle")=22
	Set gtmtypes("lvmon_value_ent",23,"name")="lvmon_value_ent.varlvval.has_aliascont"
	Set gtmtypes("lvmon_value_ent",23,"off")=64
	Set gtmtypes("lvmon_value_ent",23,"len")=4
	Set gtmtypes("lvmon_value_ent",23,"type")="boolean_t"
	Set gtmtypfldindx("lvmon_value_ent","varlvval.has_aliascont")=23
	Set gtmtypes("lvmon_value_ent",24,"name")="lvmon_value_ent.varlvval.lvmon_mark"
	Set gtmtypes("lvmon_value_ent",24,"off")=68
	Set gtmtypes("lvmon_value_ent",24,"len")=4
	Set gtmtypes("lvmon_value_ent",24,"type")="boolean_t"
	Set gtmtypfldindx("lvmon_value_ent","varlvval.lvmon_mark")=24
	Set gtmtypes("lvmon_value_ent",25,"name")="lvmon_value_ent.varlvval.tp_var"
	Set gtmtypes("lvmon_value_ent",25,"off")=72
	Set gtmtypes("lvmon_value_ent",25,"len")=8
	Set gtmtypes("lvmon_value_ent",25,"type")="addr"
	Set gtmtypfldindx("lvmon_value_ent","varlvval.tp_var")=25
	Set gtmtypes("lvmon_value_ent",26,"name")="lvmon_value_ent.varvalue"
	Set gtmtypes("lvmon_value_ent",26,"off")=80
	Set gtmtypes("lvmon_value_ent",26,"len")=16
	Set gtmtypes("lvmon_value_ent",26,"type")="mstr"
	Set gtmtypfldindx("lvmon_value_ent","varvalue")=26
	Set gtmtypes("lvmon_value_ent",27,"name")="lvmon_value_ent.varvalue.char_len"
	Set gtmtypes("lvmon_value_ent",27,"off")=80
	Set gtmtypes("lvmon_value_ent",27,"len")=4
	Set gtmtypes("lvmon_value_ent",27,"type")="unsigned-int"
	Set gtmtypfldindx("lvmon_value_ent","varvalue.char_len")=27
	Set gtmtypes("lvmon_value_ent",28,"name")="lvmon_value_ent.varvalue.len"
	Set gtmtypes("lvmon_value_ent",28,"off")=84
	Set gtmtypes("lvmon_value_ent",28,"len")=4
	Set gtmtypes("lvmon_value_ent",28,"type")="int"
	Set gtmtypfldindx("lvmon_value_ent","varvalue.len")=28
	Set gtmtypes("lvmon_value_ent",29,"name")="lvmon_value_ent.varvalue.addr"
	Set gtmtypes("lvmon_value_ent",29,"off")=88
	Set gtmtypes("lvmon_value_ent",29,"len")=8
	Set gtmtypes("lvmon_value_ent",29,"type")="addr"
	Set gtmtypfldindx("lvmon_value_ent","varvalue.addr")=29
	Set gtmtypes("lvmon_value_ent",30,"name")="lvmon_value_ent.alloclen"
	Set gtmtypes("lvmon_value_ent",30,"off")=96
	Set gtmtypes("lvmon_value_ent",30,"len")=4
	Set gtmtypes("lvmon_value_ent",30,"type")="unsigned-int"
	Set gtmtypfldindx("lvmon_value_ent","alloclen")=30
	Set gtmtypes("lvmon_value_ent",31,"name")="lvmon_value_ent.filler"
	Set gtmtypes("lvmon_value_ent",31,"off")=100
	Set gtmtypes("lvmon_value_ent",31,"len")=4
	Set gtmtypes("lvmon_value_ent",31,"type")="int"
	Set gtmtypfldindx("lvmon_value_ent","filler")=31
	;
	Set gtmtypes("lvmon_var")="struct"
	Set gtmtypes("lvmon_var",0)=38
	Set gtmtypes("lvmon_var","len")=248
	Set gtmtypes("lvmon_var",1,"name")="lvmon_var.lvmv"
	Set gtmtypes("lvmon_var",1,"off")=0
	Set gtmtypes("lvmon_var",1,"len")=24
	Set gtmtypes("lvmon_var",1,"type")="mname_entry"
	Set gtmtypfldindx("lvmon_var","lvmv")=1
	Set gtmtypes("lvmon_var",2,"name")="lvmon_var.lvmv.var_name"
	Set gtmtypes("lvmon_var",2,"off")=0
	Set gtmtypes("lvmon_var",2,"len")=16
	Set gtmtypes("lvmon_var",2,"type")="mstr"
	Set gtmtypfldindx("lvmon_var","lvmv.var_name")=2
	Set gtmtypes("lvmon_var",3,"name")="lvmon_var.lvmv.hash_code"
	Set gtmtypes("lvmon_var",3,"off")=16
	Set gtmtypes("lvmon_var",3,"len")=4
	Set gtmtypes("lvmon_var",3,"type")="unsigned-int"
	Set gtmtypfldindx("lvmon_var","lvmv.hash_code")=3
	Set gtmtypes("lvmon_var",4,"name")="lvmon_var.lvmv.marked"
	Set gtmtypes("lvmon_var",4,"off")=20
	Set gtmtypes("lvmon_var",4,"len")=4
	Set gtmtypes("lvmon_var",4,"type")="boolean_t"
	Set gtmtypfldindx("lvmon_var","lvmv.marked")=4
	Set gtmtypes("lvmon_var",5,"name")="lvmon_var.varlvadr"
	Set gtmtypes("lvmon_var",5,"off")=24
	Set gtmtypes("lvmon_var",5,"len")=8
	Set gtmtypes("lvmon_var",5,"type")="addr"
	Set gtmtypfldindx("lvmon_var","varlvadr")=5
	Set gtmtypes("lvmon_var",6,"name")="lvmon_var.curr_symval_cycle"
	Set gtmtypes("lvmon_var",6,"off")=32
	Set gtmtypes("lvmon_var",6,"len")=8
	Set gtmtypes("lvmon_var",6,"type")="gtm_int8"
	Set gtmtypfldindx("lvmon_var","curr_symval_cycle")=6
	Set gtmtypes("lvmon_var",7,"name")="lvmon_var.values"
	Set gtmtypes("lvmon_var",7,"off")=40
	Set gtmtypes("lvmon_var",7,"len")=208
	Set gtmtypes("lvmon_var",7,"type")="lvmon_value_ent"
	Set gtmtypfldindx("lvmon_var","values")=7
	Set gtmtypes("lvmon_var",7,"dim")=2
	Set gtmtypes("lvmon_var",8,"name")="lvmon_var.values[0].varlvval"
	Set gtmtypes("lvmon_var",8,"off")=40
	Set gtmtypes("lvmon_var",8,"len")=80
	Set gtmtypes("lvmon_var",8,"type")="lv_val"
	Set gtmtypfldindx("lvmon_var","values[0].varlvval")=8
	Set gtmtypes("lvmon_var",9,"name")="lvmon_var.values[0].varlvval.v"
	Set gtmtypes("lvmon_var",9,"off")=40
	Set gtmtypes("lvmon_var",9,"len")=32
	Set gtmtypes("lvmon_var",9,"type")="mval"
	Set gtmtypfldindx("lvmon_var","values[0].varlvval.v")=9
	Set gtmtypes("lvmon_var",10,"name")="lvmon_var.values[0].varlvval.v.mvtype"
	Set gtmtypes("lvmon_var",10,"off")=40
	Set gtmtypes("lvmon_var",10,"len")=2
	Set gtmtypes("lvmon_var",10,"type")="unsigned-short"
	Set gtmtypfldindx("lvmon_var","values[0].varlvval.v.mvtype")=10
	Set gtmtypes("lvmon_var",11,"name")="lvmon_var.values[0].varlvval.v.fnpc_indx"
	Set gtmtypes("lvmon_var",11,"off")=43
	Set gtmtypes("lvmon_var",11,"len")=1
	Set gtmtypes("lvmon_var",11,"type")="unsigned-char"
	Set gtmtypfldindx("lvmon_var","values[0].varlvval.v.fnpc_indx")=11
	Set gtmtypes("lvmon_var",12,"name")="lvmon_var.values[0].varlvval.v.utfcgr_indx"
	Set gtmtypes("lvmon_var",12,"off")=44
	Set gtmtypes("lvmon_var",12,"len")=4
	Set gtmtypes("lvmon_var",12,"type")="unsigned-int"
	Set gtmtypfldindx("lvmon_var","values[0].varlvval.v.utfcgr_indx")=12
	Set gtmtypes("lvmon_var",13,"name")="lvmon_var.values[0].varlvval.v.m"
	Set gtmtypes("lvmon_var",13,"off")=48
	Set gtmtypes("lvmon_var",13,"len")=8
	Set gtmtypes("lvmon_var",13,"type")="int"
	Set gtmtypfldindx("lvmon_var","values[0].varlvval.v.m")=13
	Set gtmtypes("lvmon_var",13,"dim")=2
	Set gtmtypes("lvmon_var",14,"name")="lvmon_var.values[0].varlvval.v.str"
	Set gtmtypes("lvmon_var",14,"off")=56
	Set gtmtypes("lvmon_var",14,"len")=16
	Set gtmtypes("lvmon_var",14,"type")="mstr"
	Set gtmtypfldindx("lvmon_var","values[0].varlvval.v.str")=14
	Set gtmtypes("lvmon_var",15,"name")="lvmon_var.values[0].varlvval.ptrs"
	Set gtmtypes("lvmon_var",15,"off")=72
	Set gtmtypes("lvmon_var",15,"len")=16
	Set gtmtypes("lvmon_var",15,"type")="union"
	Set gtmtypfldindx("lvmon_var","values[0].varlvval.ptrs")=15
	Set gtmtypes("lvmon_var",16,"name")="lvmon_var.values[0].varlvval.ptrs.val_ent"
	Set gtmtypes("lvmon_var",16,"off")=72
	Set gtmtypes("lvmon_var",16,"len")=16
	Set gtmtypes("lvmon_var",16,"type")="struct"
	Set gtmtypfldindx("lvmon_var","values[0].varlvval.ptrs.val_ent")=16
	Set gtmtypes("lvmon_var",17,"name")="lvmon_var.values[0].varlvval.ptrs.val_ent.children"
	Set gtmtypes("lvmon_var",17,"off")=72
	Set gtmtypes("lvmon_var",17,"len")=8
	Set gtmtypes("lvmon_var",17,"type")="addr"
	Set gtmtypfldindx("lvmon_var","values[0].varlvval.ptrs.val_ent.children")=17
	Set gtmtypes("lvmon_var",18,"name")="lvmon_var.values[0].varlvval.ptrs.val_ent.parent"
	Set gtmtypes("lvmon_var",18,"off")=80
	Set gtmtypes("lvmon_var",18,"len")=8
	Set gtmtypes("lvmon_var",18,"type")="union"
	Set gtmtypfldindx("lvmon_var","values[0].varlvval.ptrs.val_ent.parent")=18
	Set gtmtypes("lvmon_var",19,"name")="lvmon_var.values[0].varlvval.ptrs.val_ent.parent.sym"
	Set gtmtypes("lvmon_var",19,"off")=80
	Set gtmtypes("lvmon_var",19,"len")=8
	Set gtmtypes("lvmon_var",19,"type")="addr"
	Set gtmtypfldindx("lvmon_var","values[0].varlvval.ptrs.val_ent.parent.sym")=19
	Set gtmtypes("lvmon_var",20,"name")="lvmon_var.values[0].varlvval.ptrs.val_ent.parent.sbs_tree"
	Set gtmtypes("lvmon_var",20,"off")=80
	Set gtmtypes("lvmon_var",20,"len")=8
	Set gtmtypes("lvmon_var",20,"type")="addr"
	Set gtmtypfldindx("lvmon_var","values[0].varlvval.ptrs.val_ent.parent.sbs_tree")=20
	Set gtmtypes("lvmon_var",21,"name")="lvmon_var.values[0].varlvval.ptrs.free_ent"
	Set gtmtypes("lvmon_var",21,"off")=72
	Set gtmtypes("lvmon_var",21,"len")=8
	Set gtmtypes("lvmon_var",21,"type")="struct"
	Set gtmtypfldindx("lvmon_var","values[0].varlvval.ptrs.free_ent")=21
	Set gtmtypes("lvmon_var",22,"name")="lvmon_var.values[0].varlvval.ptrs.free_ent.next_free"
	Set gtmtypes("lvmon_var",22,"off")=72
	Set gtmtypes("lvmon_var",22,"len")=8
	Set gtmtypes("lvmon_var",22,"type")="addr"
	Set gtmtypfldindx("lvmon_var","values[0].varlvval.ptrs.free_ent.next_free")=22
	Set gtmtypes("lvmon_var",23,"name")="lvmon_var.values[0].varlvval.ptrs.copy_loc"
	Set gtmtypes("lvmon_var",23,"off")=72
	Set gtmtypes("lvmon_var",23,"len")=8
	Set gtmtypes("lvmon_var",23,"type")="struct"
	Set gtmtypfldindx("lvmon_var","values[0].varlvval.ptrs.copy_loc")=23
	Set gtmtypes("lvmon_var",24,"name")="lvmon_var.values[0].varlvval.ptrs.copy_loc.newtablv"
	Set gtmtypes("lvmon_var",24,"off")=72
	Set gtmtypes("lvmon_var",24,"len")=8
	Set gtmtypes("lvmon_var",24,"type")="addr"
	Set gtmtypfldindx("lvmon_var","values[0].varlvval.ptrs.copy_loc.newtablv")=24
	Set gtmtypes("lvmon_var",25,"name")="lvmon_var.values[0].varlvval.stats"
	Set gtmtypes("lvmon_var",25,"off")=88
	Set gtmtypes("lvmon_var",25,"len")=16
	Set gtmtypes("lvmon_var",25,"type")="struct"
	Set gtmtypfldindx("lvmon_var","values[0].varlvval.stats")=25
	Set gtmtypes("lvmon_var",26,"name")="lvmon_var.values[0].varlvval.stats.trefcnt"
	Set gtmtypes("lvmon_var",26,"off")=88
	Set gtmtypes("lvmon_var",26,"len")=4
	Set gtmtypes("lvmon_var",26,"type")="int"
	Set gtmtypfldindx("lvmon_var","values[0].varlvval.stats.trefcnt")=26
	Set gtmtypes("lvmon_var",27,"name")="lvmon_var.values[0].varlvval.stats.crefcnt"
	Set gtmtypes("lvmon_var",27,"off")=92
	Set gtmtypes("lvmon_var",27,"len")=4
	Set gtmtypes("lvmon_var",27,"type")="int"
	Set gtmtypfldindx("lvmon_var","values[0].varlvval.stats.crefcnt")=27
	Set gtmtypes("lvmon_var",28,"name")="lvmon_var.values[0].varlvval.stats.tstartcycle"
	Set gtmtypes("lvmon_var",28,"off")=96
	Set gtmtypes("lvmon_var",28,"len")=4
	Set gtmtypes("lvmon_var",28,"type")="unsigned-int"
	Set gtmtypfldindx("lvmon_var","values[0].varlvval.stats.tstartcycle")=28
	Set gtmtypes("lvmon_var",29,"name")="lvmon_var.values[0].varlvval.stats.lvtaskcycle"
	Set gtmtypes("lvmon_var",29,"off")=100
	Set gtmtypes("lvmon_var",29,"len")=4
	Set gtmtypes("lvmon_var",29,"type")="unsigned-int"
	Set gtmtypfldindx("lvmon_var","values[0].varlvval.stats.lvtaskcycle")=29
	Set gtmtypes("lvmon_var",30,"name")="lvmon_var.values[0].varlvval.has_aliascont"
	Set gtmtypes("lvmon_var",30,"off")=104
	Set gtmtypes("lvmon_var",30,"len")=4
	Set gtmtypes("lvmon_var",30,"type")="boolean_t"
	Set gtmtypfldindx("lvmon_var","values[0].varlvval.has_aliascont")=30
	Set gtmtypes("lvmon_var",31,"name")="lvmon_var.values[0].varlvval.lvmon_mark"
	Set gtmtypes("lvmon_var",31,"off")=108
	Set gtmtypes("lvmon_var",31,"len")=4
	Set gtmtypes("lvmon_var",31,"type")="boolean_t"
	Set gtmtypfldindx("lvmon_var","values[0].varlvval.lvmon_mark")=31
	Set gtmtypes("lvmon_var",32,"name")="lvmon_var.values[0].varlvval.tp_var"
	Set gtmtypes("lvmon_var",32,"off")=112
	Set gtmtypes("lvmon_var",32,"len")=8
	Set gtmtypes("lvmon_var",32,"type")="addr"
	Set gtmtypfldindx("lvmon_var","values[0].varlvval.tp_var")=32
	Set gtmtypes("lvmon_var",33,"name")="lvmon_var.values[0].varvalue"
	Set gtmtypes("lvmon_var",33,"off")=120
	Set gtmtypes("lvmon_var",33,"len")=16
	Set gtmtypes("lvmon_var",33,"type")="mstr"
	Set gtmtypfldindx("lvmon_var","values[0].varvalue")=33
	Set gtmtypes("lvmon_var",34,"name")="lvmon_var.values[0].varvalue.char_len"
	Set gtmtypes("lvmon_var",34,"off")=120
	Set gtmtypes("lvmon_var",34,"len")=4
	Set gtmtypes("lvmon_var",34,"type")="unsigned-int"
	Set gtmtypfldindx("lvmon_var","values[0].varvalue.char_len")=34
	Set gtmtypes("lvmon_var",35,"name")="lvmon_var.values[0].varvalue.len"
	Set gtmtypes("lvmon_var",35,"off")=124
	Set gtmtypes("lvmon_var",35,"len")=4
	Set gtmtypes("lvmon_var",35,"type")="int"
	Set gtmtypfldindx("lvmon_var","values[0].varvalue.len")=35
	Set gtmtypes("lvmon_var",36,"name")="lvmon_var.values[0].varvalue.addr"
	Set gtmtypes("lvmon_var",36,"off")=128
	Set gtmtypes("lvmon_var",36,"len")=8
	Set gtmtypes("lvmon_var",36,"type")="addr"
	Set gtmtypfldindx("lvmon_var","values[0].varvalue.addr")=36
	Set gtmtypes("lvmon_var",37,"name")="lvmon_var.values[0].alloclen"
	Set gtmtypes("lvmon_var",37,"off")=136
	Set gtmtypes("lvmon_var",37,"len")=4
	Set gtmtypes("lvmon_var",37,"type")="unsigned-int"
	Set gtmtypfldindx("lvmon_var","values[0].alloclen")=37
	Set gtmtypes("lvmon_var",38,"name")="lvmon_var.values[0].filler"
	Set gtmtypes("lvmon_var",38,"off")=140
	Set gtmtypes("lvmon_var",38,"len")=4
	Set gtmtypes("lvmon_var",38,"type")="int"
	Set gtmtypfldindx("lvmon_var","values[0].filler")=38
	;
	Set gtmtypes("lvname_info")="struct"
	Set gtmtypes("lvname_info",0)=4
	Set gtmtypes("lvname_info","len")=280
	Set gtmtypes("lvname_info",1,"name")="lvname_info.total_lv_subs"
	Set gtmtypes("lvname_info",1,"off")=0
	Set gtmtypes("lvname_info",1,"len")=8
	Set gtmtypes("lvname_info",1,"type")="intptr_t"
	Set gtmtypfldindx("lvname_info","total_lv_subs")=1
	Set gtmtypes("lvname_info",2,"name")="lvname_info.start_lvp"
	Set gtmtypes("lvname_info",2,"off")=8
	Set gtmtypes("lvname_info",2,"len")=8
	Set gtmtypes("lvname_info",2,"type")="addr"
	Set gtmtypfldindx("lvname_info","start_lvp")=2
	Set gtmtypes("lvname_info",3,"name")="lvname_info.lv_subs"
	Set gtmtypes("lvname_info",3,"off")=16
	Set gtmtypes("lvname_info",3,"len")=256
	Set gtmtypes("lvname_info",3,"type")="addr"
	Set gtmtypfldindx("lvname_info","lv_subs")=3
	Set gtmtypes("lvname_info",3,"dim")=32
	Set gtmtypes("lvname_info",4,"name")="lvname_info.end_lvp"
	Set gtmtypes("lvname_info",4,"off")=272
	Set gtmtypes("lvname_info",4,"len")=8
	Set gtmtypes("lvname_info",4,"type")="addr"
	Set gtmtypfldindx("lvname_info","end_lvp")=4
	;
	Set gtmtypes("lvzwrite_datablk")="struct"
	Set gtmtypes("lvzwrite_datablk",0)=10
	Set gtmtypes("lvzwrite_datablk","len")=56
	Set gtmtypes("lvzwrite_datablk",1,"name")="lvzwrite_datablk.zwr_intype"
	Set gtmtypes("lvzwrite_datablk",1,"off")=0
	Set gtmtypes("lvzwrite_datablk",1,"len")=4
	Set gtmtypes("lvzwrite_datablk",1,"type")="int"
	Set gtmtypfldindx("lvzwrite_datablk","zwr_intype")=1
	Set gtmtypes("lvzwrite_datablk",2,"name")="lvzwrite_datablk.fixed"
	Set gtmtypes("lvzwrite_datablk",2,"off")=4
	Set gtmtypes("lvzwrite_datablk",2,"len")=4
	Set gtmtypes("lvzwrite_datablk",2,"type")="boolean_t"
	Set gtmtypfldindx("lvzwrite_datablk","fixed")=2
	Set gtmtypes("lvzwrite_datablk",3,"name")="lvzwrite_datablk.zav_added"
	Set gtmtypes("lvzwrite_datablk",3,"off")=8
	Set gtmtypes("lvzwrite_datablk",3,"len")=4
	Set gtmtypes("lvzwrite_datablk",3,"type")="boolean_t"
	Set gtmtypfldindx("lvzwrite_datablk","zav_added")=3
	Set gtmtypes("lvzwrite_datablk",4,"name")="lvzwrite_datablk.subsc_count"
	Set gtmtypes("lvzwrite_datablk",4,"off")=12
	Set gtmtypes("lvzwrite_datablk",4,"len")=2
	Set gtmtypes("lvzwrite_datablk",4,"type")="unsigned-short"
	Set gtmtypfldindx("lvzwrite_datablk","subsc_count")=4
	Set gtmtypes("lvzwrite_datablk",5,"name")="lvzwrite_datablk.curr_subsc"
	Set gtmtypes("lvzwrite_datablk",5,"off")=14
	Set gtmtypes("lvzwrite_datablk",5,"len")=2
	Set gtmtypes("lvzwrite_datablk",5,"type")="unsigned-short"
	Set gtmtypfldindx("lvzwrite_datablk","curr_subsc")=5
	Set gtmtypes("lvzwrite_datablk",6,"name")="lvzwrite_datablk.mask"
	Set gtmtypes("lvzwrite_datablk",6,"off")=16
	Set gtmtypes("lvzwrite_datablk",6,"len")=4
	Set gtmtypes("lvzwrite_datablk",6,"type")="unsigned-int"
	Set gtmtypfldindx("lvzwrite_datablk","mask")=6
	Set gtmtypes("lvzwrite_datablk",7,"name")="lvzwrite_datablk.pat"
	Set gtmtypes("lvzwrite_datablk",7,"off")=24
	Set gtmtypes("lvzwrite_datablk",7,"len")=8
	Set gtmtypes("lvzwrite_datablk",7,"type")="addr"
	Set gtmtypfldindx("lvzwrite_datablk","pat")=7
	Set gtmtypes("lvzwrite_datablk",8,"name")="lvzwrite_datablk.curr_name"
	Set gtmtypes("lvzwrite_datablk",8,"off")=32
	Set gtmtypes("lvzwrite_datablk",8,"len")=8
	Set gtmtypes("lvzwrite_datablk",8,"type")="addr"
	Set gtmtypfldindx("lvzwrite_datablk","curr_name")=8
	Set gtmtypes("lvzwrite_datablk",9,"name")="lvzwrite_datablk.sub"
	Set gtmtypes("lvzwrite_datablk",9,"off")=40
	Set gtmtypes("lvzwrite_datablk",9,"len")=8
	Set gtmtypes("lvzwrite_datablk",9,"type")="addr"
	Set gtmtypfldindx("lvzwrite_datablk","sub")=9
	Set gtmtypes("lvzwrite_datablk",10,"name")="lvzwrite_datablk.prev"
	Set gtmtypes("lvzwrite_datablk",10,"off")=48
	Set gtmtypes("lvzwrite_datablk",10,"len")=8
	Set gtmtypes("lvzwrite_datablk",10,"type")="addr"
	Set gtmtypfldindx("lvzwrite_datablk","prev")=10
	;
	Set gtmtypes("mcalloc_hdr")="struct"
	Set gtmtypes("mcalloc_hdr",0)=4
	Set gtmtypes("mcalloc_hdr","len")=24
	Set gtmtypes("mcalloc_hdr",1,"name")="mcalloc_hdr.link"
	Set gtmtypes("mcalloc_hdr",1,"off")=0
	Set gtmtypes("mcalloc_hdr",1,"len")=8
	Set gtmtypes("mcalloc_hdr",1,"type")="addr"
	Set gtmtypfldindx("mcalloc_hdr","link")=1
	Set gtmtypes("mcalloc_hdr",2,"name")="mcalloc_hdr.size"
	Set gtmtypes("mcalloc_hdr",2,"off")=8
	Set gtmtypes("mcalloc_hdr",2,"len")=4
	Set gtmtypes("mcalloc_hdr",2,"type")="int"
	Set gtmtypfldindx("mcalloc_hdr","size")=2
	Set gtmtypes("mcalloc_hdr",3,"name")="mcalloc_hdr.filler"
	Set gtmtypes("mcalloc_hdr",3,"off")=12
	Set gtmtypes("mcalloc_hdr",3,"len")=4
	Set gtmtypes("mcalloc_hdr",3,"type")="int"
	Set gtmtypfldindx("mcalloc_hdr","filler")=3
	Set gtmtypes("mcalloc_hdr",4,"name")="mcalloc_hdr.data"
	Set gtmtypes("mcalloc_hdr",4,"off")=16
	Set gtmtypes("mcalloc_hdr",4,"len")=1
	Set gtmtypes("mcalloc_hdr",4,"type")="char"
	Set gtmtypfldindx("mcalloc_hdr","data")=4
	;
	Set gtmtypes("mem_list")="struct"
	Set gtmtypes("mem_list",0)=5
	Set gtmtypes("mem_list","len")=40
	Set gtmtypes("mem_list",1,"name")="mem_list.addr"
	Set gtmtypes("mem_list",1,"off")=0
	Set gtmtypes("mem_list",1,"len")=8
	Set gtmtypes("mem_list",1,"type")="addr"
	Set gtmtypfldindx("mem_list","addr")=1
	Set gtmtypes("mem_list",2,"name")="mem_list.pages"
	Set gtmtypes("mem_list",2,"off")=8
	Set gtmtypes("mem_list",2,"len")=4
	Set gtmtypes("mem_list",2,"type")="unsigned-int"
	Set gtmtypfldindx("mem_list","pages")=2
	Set gtmtypes("mem_list",3,"name")="mem_list.next"
	Set gtmtypes("mem_list",3,"off")=16
	Set gtmtypes("mem_list",3,"len")=8
	Set gtmtypes("mem_list",3,"type")="addr"
	Set gtmtypfldindx("mem_list","next")=3
	Set gtmtypes("mem_list",4,"name")="mem_list.prev"
	Set gtmtypes("mem_list",4,"off")=24
	Set gtmtypes("mem_list",4,"len")=8
	Set gtmtypes("mem_list",4,"type")="addr"
	Set gtmtypfldindx("mem_list","prev")=4
	Set gtmtypes("mem_list",5,"name")="mem_list.gtm_free"
	Set gtmtypes("mem_list",5,"off")=32
	Set gtmtypes("mem_list",5,"len")=1
	Set gtmtypes("mem_list",5,"type")="char"
	Set gtmtypfldindx("mem_list","gtm_free")=5
	;
	Set gtmtypes("merge_glvn_struct")="struct"
	Set gtmtypes("merge_glvn_struct",0)=2
	Set gtmtypes("merge_glvn_struct","len")=32
	Set gtmtypes("merge_glvn_struct",1,"name")="merge_glvn_struct.gblp"
	Set gtmtypes("merge_glvn_struct",1,"off")=0
	Set gtmtypes("merge_glvn_struct",1,"len")=16
	Set gtmtypes("merge_glvn_struct",1,"type")="addr"
	Set gtmtypfldindx("merge_glvn_struct","gblp")=1
	Set gtmtypes("merge_glvn_struct",1,"dim")=2
	Set gtmtypes("merge_glvn_struct",2,"name")="merge_glvn_struct.lclp"
	Set gtmtypes("merge_glvn_struct",2,"off")=16
	Set gtmtypes("merge_glvn_struct",2,"len")=16
	Set gtmtypes("merge_glvn_struct",2,"type")="addr"
	Set gtmtypfldindx("merge_glvn_struct","lclp")=2
	Set gtmtypes("merge_glvn_struct",2,"dim")=2
	;
	Set gtmtypes("mident_fixed")="struct"
	Set gtmtypes("mident_fixed",0)=1
	Set gtmtypes("mident_fixed","len")=32
	Set gtmtypes("mident_fixed",1,"name")="mident_fixed.c"
	Set gtmtypes("mident_fixed",1,"off")=0
	Set gtmtypes("mident_fixed",1,"len")=32
	Set gtmtypes("mident_fixed",1,"type")="char"
	Set gtmtypfldindx("mident_fixed","c")=1
	;
	Set gtmtypes("mlabel")="struct"
	Set gtmtypes("mlabel",0)=9
	Set gtmtypes("mlabel","len")=48
	Set gtmtypes("mlabel",1,"name")="mlabel.lson"
	Set gtmtypes("mlabel",1,"off")=0
	Set gtmtypes("mlabel",1,"len")=8
	Set gtmtypes("mlabel",1,"type")="addr"
	Set gtmtypfldindx("mlabel","lson")=1
	Set gtmtypes("mlabel",2,"name")="mlabel.rson"
	Set gtmtypes("mlabel",2,"off")=8
	Set gtmtypes("mlabel",2,"len")=8
	Set gtmtypes("mlabel",2,"type")="addr"
	Set gtmtypfldindx("mlabel","rson")=2
	Set gtmtypes("mlabel",3,"name")="mlabel.ml"
	Set gtmtypes("mlabel",3,"off")=16
	Set gtmtypes("mlabel",3,"len")=8
	Set gtmtypes("mlabel",3,"type")="addr"
	Set gtmtypfldindx("mlabel","ml")=3
	Set gtmtypes("mlabel",4,"name")="mlabel.mvname"
	Set gtmtypes("mlabel",4,"off")=24
	Set gtmtypes("mlabel",4,"len")=16
	Set gtmtypes("mlabel",4,"type")="mstr"
	Set gtmtypfldindx("mlabel","mvname")=4
	Set gtmtypes("mlabel",5,"name")="mlabel.mvname.char_len"
	Set gtmtypes("mlabel",5,"off")=24
	Set gtmtypes("mlabel",5,"len")=4
	Set gtmtypes("mlabel",5,"type")="unsigned-int"
	Set gtmtypfldindx("mlabel","mvname.char_len")=5
	Set gtmtypes("mlabel",6,"name")="mlabel.mvname.len"
	Set gtmtypes("mlabel",6,"off")=28
	Set gtmtypes("mlabel",6,"len")=4
	Set gtmtypes("mlabel",6,"type")="int"
	Set gtmtypfldindx("mlabel","mvname.len")=6
	Set gtmtypes("mlabel",7,"name")="mlabel.mvname.addr"
	Set gtmtypes("mlabel",7,"off")=32
	Set gtmtypes("mlabel",7,"len")=8
	Set gtmtypes("mlabel",7,"type")="addr"
	Set gtmtypfldindx("mlabel","mvname.addr")=7
	Set gtmtypes("mlabel",8,"name")="mlabel.formalcnt"
	Set gtmtypes("mlabel",8,"off")=40
	Set gtmtypes("mlabel",8,"len")=4
	Set gtmtypes("mlabel",8,"type")="int"
	Set gtmtypfldindx("mlabel","formalcnt")=8
	Set gtmtypes("mlabel",9,"name")="mlabel.gbl"
	Set gtmtypes("mlabel",9,"off")=44
	Set gtmtypes("mlabel",9,"len")=4
	Set gtmtypes("mlabel",9,"type")="boolean_t"
	Set gtmtypfldindx("mlabel","gbl")=9
	;
	Set gtmtypes("mline")="struct"
	Set gtmtypes("mline",0)=7
	Set gtmtypes("mline","len")=48
	Set gtmtypes("mline",1,"name")="mline.parent"
	Set gtmtypes("mline",1,"off")=0
	Set gtmtypes("mline",1,"len")=8
	Set gtmtypes("mline",1,"type")="addr"
	Set gtmtypfldindx("mline","parent")=1
	Set gtmtypes("mline",2,"name")="mline.sibling"
	Set gtmtypes("mline",2,"off")=8
	Set gtmtypes("mline",2,"len")=8
	Set gtmtypes("mline",2,"type")="addr"
	Set gtmtypfldindx("mline","sibling")=2
	Set gtmtypes("mline",3,"name")="mline.child"
	Set gtmtypes("mline",3,"off")=16
	Set gtmtypes("mline",3,"len")=8
	Set gtmtypes("mline",3,"type")="addr"
	Set gtmtypfldindx("mline","child")=3
	Set gtmtypes("mline",4,"name")="mline.externalentry"
	Set gtmtypes("mline",4,"off")=24
	Set gtmtypes("mline",4,"len")=8
	Set gtmtypes("mline",4,"type")="addr"
	Set gtmtypfldindx("mline","externalentry")=4
	Set gtmtypes("mline",5,"name")="mline.line_number"
	Set gtmtypes("mline",5,"off")=32
	Set gtmtypes("mline",5,"len")=4
	Set gtmtypes("mline",5,"type")="unsigned-int"
	Set gtmtypfldindx("mline","line_number")=5
	Set gtmtypes("mline",6,"name")="mline.table"
	Set gtmtypes("mline",6,"off")=36
	Set gtmtypes("mline",6,"len")=4
	Set gtmtypes("mline",6,"type")="boolean_t"
	Set gtmtypfldindx("mline","table")=6
	Set gtmtypes("mline",7,"name")="mline.block_ok"
	Set gtmtypes("mline",7,"off")=40
	Set gtmtypes("mline",7,"len")=4
	Set gtmtypes("mline",7,"type")="boolean_t"
	Set gtmtypfldindx("mline","block_ok")=7
	;
	Set gtmtypes("mliteral")="struct"
	Set gtmtypes("mliteral",0)=11
	Set gtmtypes("mliteral","len")=64
	Set gtmtypes("mliteral",1,"name")="mliteral.que"
	Set gtmtypes("mliteral",1,"off")=0
	Set gtmtypes("mliteral",1,"len")=16
	Set gtmtypes("mliteral",1,"type")="struct"
	Set gtmtypfldindx("mliteral","que")=1
	Set gtmtypes("mliteral",2,"name")="mliteral.que.fl"
	Set gtmtypes("mliteral",2,"off")=0
	Set gtmtypes("mliteral",2,"len")=8
	Set gtmtypes("mliteral",2,"type")="addr"
	Set gtmtypfldindx("mliteral","que.fl")=2
	Set gtmtypes("mliteral",3,"name")="mliteral.que.bl"
	Set gtmtypes("mliteral",3,"off")=8
	Set gtmtypes("mliteral",3,"len")=8
	Set gtmtypes("mliteral",3,"type")="addr"
	Set gtmtypfldindx("mliteral","que.bl")=3
	Set gtmtypes("mliteral",4,"name")="mliteral.rt_addr"
	Set gtmtypes("mliteral",4,"off")=16
	Set gtmtypes("mliteral",4,"len")=8
	Set gtmtypes("mliteral",4,"type")="intptr_t"
	Set gtmtypfldindx("mliteral","rt_addr")=4
	Set gtmtypes("mliteral",5,"name")="mliteral.reference_count"
	Set gtmtypes("mliteral",5,"off")=24
	Set gtmtypes("mliteral",5,"len")=4
	Set gtmtypes("mliteral",5,"type")="int"
	Set gtmtypfldindx("mliteral","reference_count")=5
	Set gtmtypes("mliteral",6,"name")="mliteral.v"
	Set gtmtypes("mliteral",6,"off")=32
	Set gtmtypes("mliteral",6,"len")=32
	Set gtmtypes("mliteral",6,"type")="mval"
	Set gtmtypfldindx("mliteral","v")=6
	Set gtmtypes("mliteral",7,"name")="mliteral.v.mvtype"
	Set gtmtypes("mliteral",7,"off")=32
	Set gtmtypes("mliteral",7,"len")=2
	Set gtmtypes("mliteral",7,"type")="unsigned-short"
	Set gtmtypfldindx("mliteral","v.mvtype")=7
	Set gtmtypes("mliteral",8,"name")="mliteral.v.fnpc_indx"
	Set gtmtypes("mliteral",8,"off")=35
	Set gtmtypes("mliteral",8,"len")=1
	Set gtmtypes("mliteral",8,"type")="unsigned-char"
	Set gtmtypfldindx("mliteral","v.fnpc_indx")=8
	Set gtmtypes("mliteral",9,"name")="mliteral.v.utfcgr_indx"
	Set gtmtypes("mliteral",9,"off")=36
	Set gtmtypes("mliteral",9,"len")=4
	Set gtmtypes("mliteral",9,"type")="unsigned-int"
	Set gtmtypfldindx("mliteral","v.utfcgr_indx")=9
	Set gtmtypes("mliteral",10,"name")="mliteral.v.m"
	Set gtmtypes("mliteral",10,"off")=40
	Set gtmtypes("mliteral",10,"len")=8
	Set gtmtypes("mliteral",10,"type")="int"
	Set gtmtypfldindx("mliteral","v.m")=10
	Set gtmtypes("mliteral",10,"dim")=2
	Set gtmtypes("mliteral",11,"name")="mliteral.v.str"
	Set gtmtypes("mliteral",11,"off")=48
	Set gtmtypes("mliteral",11,"len")=16
	Set gtmtypes("mliteral",11,"type")="mstr"
	Set gtmtypfldindx("mliteral","v.str")=11
	;
	Set gtmtypes("mlk_ctldata")="struct"
	Set gtmtypes("mlk_ctldata",0)=26
	Set gtmtypes("mlk_ctldata","len")=136
	Set gtmtypes("mlk_ctldata",1,"name")="mlk_ctldata.prcfree"
	Set gtmtypes("mlk_ctldata",1,"off")=0
	Set gtmtypes("mlk_ctldata",1,"len")=8
	Set gtmtypes("mlk_ctldata",1,"type")="intptr_t"
	Set gtmtypfldindx("mlk_ctldata","prcfree")=1
	Set gtmtypes("mlk_ctldata",2,"name")="mlk_ctldata.blkbase"
	Set gtmtypes("mlk_ctldata",2,"off")=8
	Set gtmtypes("mlk_ctldata",2,"len")=8
	Set gtmtypes("mlk_ctldata",2,"type")="intptr_t"
	Set gtmtypfldindx("mlk_ctldata","blkbase")=2
	Set gtmtypes("mlk_ctldata",3,"name")="mlk_ctldata.blkfree"
	Set gtmtypes("mlk_ctldata",3,"off")=16
	Set gtmtypes("mlk_ctldata",3,"len")=8
	Set gtmtypes("mlk_ctldata",3,"type")="intptr_t"
	Set gtmtypfldindx("mlk_ctldata","blkfree")=3
	Set gtmtypes("mlk_ctldata",4,"name")="mlk_ctldata.blkhash"
	Set gtmtypes("mlk_ctldata",4,"off")=24
	Set gtmtypes("mlk_ctldata",4,"len")=8
	Set gtmtypes("mlk_ctldata",4,"type")="intptr_t"
	Set gtmtypfldindx("mlk_ctldata","blkhash")=4
	Set gtmtypes("mlk_ctldata",5,"name")="mlk_ctldata.blkroot"
	Set gtmtypes("mlk_ctldata",5,"off")=32
	Set gtmtypes("mlk_ctldata",5,"len")=8
	Set gtmtypes("mlk_ctldata",5,"type")="intptr_t"
	Set gtmtypfldindx("mlk_ctldata","blkroot")=5
	Set gtmtypes("mlk_ctldata",6,"name")="mlk_ctldata.subbase"
	Set gtmtypes("mlk_ctldata",6,"off")=40
	Set gtmtypes("mlk_ctldata",6,"len")=8
	Set gtmtypes("mlk_ctldata",6,"type")="intptr_t"
	Set gtmtypfldindx("mlk_ctldata","subbase")=6
	Set gtmtypes("mlk_ctldata",7,"name")="mlk_ctldata.subfree"
	Set gtmtypes("mlk_ctldata",7,"off")=48
	Set gtmtypes("mlk_ctldata",7,"len")=8
	Set gtmtypes("mlk_ctldata",7,"type")="intptr_t"
	Set gtmtypfldindx("mlk_ctldata","subfree")=7
	Set gtmtypes("mlk_ctldata",8,"name")="mlk_ctldata.subtop"
	Set gtmtypes("mlk_ctldata",8,"off")=56
	Set gtmtypes("mlk_ctldata",8,"len")=8
	Set gtmtypes("mlk_ctldata",8,"type")="intptr_t"
	Set gtmtypfldindx("mlk_ctldata","subtop")=8
	Set gtmtypes("mlk_ctldata",9,"name")="mlk_ctldata.wakeups"
	Set gtmtypes("mlk_ctldata",9,"off")=64
	Set gtmtypes("mlk_ctldata",9,"len")=8
	Set gtmtypes("mlk_ctldata",9,"type")="uint64_t"
	Set gtmtypfldindx("mlk_ctldata","wakeups")=9
	Set gtmtypes("mlk_ctldata",10,"name")="mlk_ctldata.max_prccnt"
	Set gtmtypes("mlk_ctldata",10,"off")=72
	Set gtmtypes("mlk_ctldata",10,"len")=4
	Set gtmtypes("mlk_ctldata",10,"type")="unsigned-int"
	Set gtmtypfldindx("mlk_ctldata","max_prccnt")=10
	Set gtmtypes("mlk_ctldata",11,"name")="mlk_ctldata.max_blkcnt"
	Set gtmtypes("mlk_ctldata",11,"off")=76
	Set gtmtypes("mlk_ctldata",11,"len")=4
	Set gtmtypes("mlk_ctldata",11,"type")="unsigned-int"
	Set gtmtypfldindx("mlk_ctldata","max_blkcnt")=11
	Set gtmtypes("mlk_ctldata",12,"name")="mlk_ctldata.num_blkhash"
	Set gtmtypes("mlk_ctldata",12,"off")=80
	Set gtmtypes("mlk_ctldata",12,"len")=4
	Set gtmtypes("mlk_ctldata",12,"type")="unsigned-int"
	Set gtmtypfldindx("mlk_ctldata","num_blkhash")=12
	Set gtmtypes("mlk_ctldata",13,"name")="mlk_ctldata.prccnt"
	Set gtmtypes("mlk_ctldata",13,"off")=84
	Set gtmtypes("mlk_ctldata",13,"len")=4
	Set gtmtypes("mlk_ctldata",13,"type")="int"
	Set gtmtypfldindx("mlk_ctldata","prccnt")=13
	Set gtmtypes("mlk_ctldata",14,"name")="mlk_ctldata.blkcnt"
	Set gtmtypes("mlk_ctldata",14,"off")=88
	Set gtmtypes("mlk_ctldata",14,"len")=4
	Set gtmtypes("mlk_ctldata",14,"type")="int"
	Set gtmtypfldindx("mlk_ctldata","blkcnt")=14
	Set gtmtypes("mlk_ctldata",15,"name")="mlk_ctldata.lockspacefull_logged"
	Set gtmtypes("mlk_ctldata",15,"off")=92
	Set gtmtypes("mlk_ctldata",15,"len")=4
	Set gtmtypes("mlk_ctldata",15,"type")="boolean_t"
	Set gtmtypfldindx("mlk_ctldata","lockspacefull_logged")=15
	Set gtmtypes("mlk_ctldata",16,"name")="mlk_ctldata.gc_needed"
	Set gtmtypes("mlk_ctldata",16,"off")=96
	Set gtmtypes("mlk_ctldata",16,"len")=4
	Set gtmtypes("mlk_ctldata",16,"type")="boolean_t"
	Set gtmtypfldindx("mlk_ctldata","gc_needed")=16
	Set gtmtypes("mlk_ctldata",17,"name")="mlk_ctldata.resize_needed"
	Set gtmtypes("mlk_ctldata",17,"off")=100
	Set gtmtypes("mlk_ctldata",17,"len")=4
	Set gtmtypes("mlk_ctldata",17,"type")="boolean_t"
	Set gtmtypfldindx("mlk_ctldata","resize_needed")=17
	Set gtmtypes("mlk_ctldata",18,"name")="mlk_ctldata.rehash_needed"
	Set gtmtypes("mlk_ctldata",18,"off")=104
	Set gtmtypes("mlk_ctldata",18,"len")=4
	Set gtmtypes("mlk_ctldata",18,"type")="boolean_t"
	Set gtmtypfldindx("mlk_ctldata","rehash_needed")=18
	Set gtmtypes("mlk_ctldata",19,"name")="mlk_ctldata.lock_gc_in_progress"
	Set gtmtypes("mlk_ctldata",19,"off")=112
	Set gtmtypes("mlk_ctldata",19,"len")=8
	Set gtmtypes("mlk_ctldata",19,"type")="global_latch_t"
	Set gtmtypfldindx("mlk_ctldata","lock_gc_in_progress")=19
	Set gtmtypes("mlk_ctldata",20,"name")="mlk_ctldata.lock_gc_in_progress.u"
	Set gtmtypes("mlk_ctldata",20,"off")=112
	Set gtmtypes("mlk_ctldata",20,"len")=8
	Set gtmtypes("mlk_ctldata",20,"type")="union"
	Set gtmtypfldindx("mlk_ctldata","lock_gc_in_progress.u")=20
	Set gtmtypes("mlk_ctldata",21,"name")="mlk_ctldata.lock_gc_in_progress.u.pid_imgcnt"
	Set gtmtypes("mlk_ctldata",21,"off")=112
	Set gtmtypes("mlk_ctldata",21,"len")=8
	Set gtmtypes("mlk_ctldata",21,"type")="uint64_t"
	Set gtmtypfldindx("mlk_ctldata","lock_gc_in_progress.u.pid_imgcnt")=21
	Set gtmtypes("mlk_ctldata",22,"name")="mlk_ctldata.lock_gc_in_progress.u.parts"
	Set gtmtypes("mlk_ctldata",22,"off")=112
	Set gtmtypes("mlk_ctldata",22,"len")=8
	Set gtmtypes("mlk_ctldata",22,"type")="struct"
	Set gtmtypfldindx("mlk_ctldata","lock_gc_in_progress.u.parts")=22
	Set gtmtypes("mlk_ctldata",23,"name")="mlk_ctldata.lock_gc_in_progress.u.parts.latch_pid"
	Set gtmtypes("mlk_ctldata",23,"off")=112
	Set gtmtypes("mlk_ctldata",23,"len")=4
	Set gtmtypes("mlk_ctldata",23,"type")="int"
	Set gtmtypfldindx("mlk_ctldata","lock_gc_in_progress.u.parts.latch_pid")=23
	Set gtmtypes("mlk_ctldata",24,"name")="mlk_ctldata.lock_gc_in_progress.u.parts.latch_word"
	Set gtmtypes("mlk_ctldata",24,"off")=116
	Set gtmtypes("mlk_ctldata",24,"len")=4
	Set gtmtypes("mlk_ctldata",24,"type")="int"
	Set gtmtypfldindx("mlk_ctldata","lock_gc_in_progress.u.parts.latch_word")=24
	Set gtmtypes("mlk_ctldata",25,"name")="mlk_ctldata.hash_seed"
	Set gtmtypes("mlk_ctldata",25,"off")=120
	Set gtmtypes("mlk_ctldata",25,"len")=8
	Set gtmtypes("mlk_ctldata",25,"type")="gtm_uint8"
	Set gtmtypfldindx("mlk_ctldata","hash_seed")=25
	Set gtmtypes("mlk_ctldata",26,"name")="mlk_ctldata.hash_shmid"
	Set gtmtypes("mlk_ctldata",26,"off")=128
	Set gtmtypes("mlk_ctldata",26,"len")=4
	Set gtmtypes("mlk_ctldata",26,"type")="int"
	Set gtmtypfldindx("mlk_ctldata","hash_shmid")=26
	;
	Set gtmtypes("mlk_prcblk")="struct"
	Set gtmtypes("mlk_prcblk",0)=4
	Set gtmtypes("mlk_prcblk","len")=16
	Set gtmtypes("mlk_prcblk",1,"name")="mlk_prcblk.next"
	Set gtmtypes("mlk_prcblk",1,"off")=0
	Set gtmtypes("mlk_prcblk",1,"len")=8
	Set gtmtypes("mlk_prcblk",1,"type")="intptr_t"
	Set gtmtypfldindx("mlk_prcblk","next")=1
	Set gtmtypes("mlk_prcblk",2,"name")="mlk_prcblk.process_id"
	Set gtmtypes("mlk_prcblk",2,"off")=8
	Set gtmtypes("mlk_prcblk",2,"len")=4
	Set gtmtypes("mlk_prcblk",2,"type")="unsigned-int"
	Set gtmtypfldindx("mlk_prcblk","process_id")=2
	Set gtmtypes("mlk_prcblk",3,"name")="mlk_prcblk.ref_cnt"
	Set gtmtypes("mlk_prcblk",3,"off")=12
	Set gtmtypes("mlk_prcblk",3,"len")=2
	Set gtmtypes("mlk_prcblk",3,"type")="short"
	Set gtmtypfldindx("mlk_prcblk","ref_cnt")=3
	Set gtmtypes("mlk_prcblk",4,"name")="mlk_prcblk.filler_4byte"
	Set gtmtypes("mlk_prcblk",4,"off")=14
	Set gtmtypes("mlk_prcblk",4,"len")=2
	Set gtmtypes("mlk_prcblk",4,"type")="short"
	Set gtmtypfldindx("mlk_prcblk","filler_4byte")=4
	;
	Set gtmtypes("mlk_pvtblk")="struct"
	Set gtmtypes("mlk_pvtblk",0)=22
	Set gtmtypes("mlk_pvtblk","len")=120
	Set gtmtypes("mlk_pvtblk",1,"name")="mlk_pvtblk.pvtctl"
	Set gtmtypes("mlk_pvtblk",1,"off")=0
	Set gtmtypes("mlk_pvtblk",1,"len")=48
	Set gtmtypes("mlk_pvtblk",1,"type")="mlk_pvtctl"
	Set gtmtypfldindx("mlk_pvtblk","pvtctl")=1
	Set gtmtypes("mlk_pvtblk",2,"name")="mlk_pvtblk.pvtctl.region"
	Set gtmtypes("mlk_pvtblk",2,"off")=0
	Set gtmtypes("mlk_pvtblk",2,"len")=8
	Set gtmtypes("mlk_pvtblk",2,"type")="addr"
	Set gtmtypfldindx("mlk_pvtblk","pvtctl.region")=2
	Set gtmtypes("mlk_pvtblk",3,"name")="mlk_pvtblk.pvtctl.csa"
	Set gtmtypes("mlk_pvtblk",3,"off")=8
	Set gtmtypes("mlk_pvtblk",3,"len")=8
	Set gtmtypes("mlk_pvtblk",3,"type")="addr"
	Set gtmtypfldindx("mlk_pvtblk","pvtctl.csa")=3
	Set gtmtypes("mlk_pvtblk",4,"name")="mlk_pvtblk.pvtctl.ctl"
	Set gtmtypes("mlk_pvtblk",4,"off")=16
	Set gtmtypes("mlk_pvtblk",4,"len")=8
	Set gtmtypes("mlk_pvtblk",4,"type")="addr"
	Set gtmtypfldindx("mlk_pvtblk","pvtctl.ctl")=4
	Set gtmtypes("mlk_pvtblk",5,"name")="mlk_pvtblk.pvtctl.shrblk"
	Set gtmtypes("mlk_pvtblk",5,"off")=24
	Set gtmtypes("mlk_pvtblk",5,"len")=8
	Set gtmtypes("mlk_pvtblk",5,"type")="addr"
	Set gtmtypfldindx("mlk_pvtblk","pvtctl.shrblk")=5
	Set gtmtypes("mlk_pvtblk",6,"name")="mlk_pvtblk.pvtctl.shrhash"
	Set gtmtypes("mlk_pvtblk",6,"off")=32
	Set gtmtypes("mlk_pvtblk",6,"len")=8
	Set gtmtypes("mlk_pvtblk",6,"type")="addr"
	Set gtmtypfldindx("mlk_pvtblk","pvtctl.shrhash")=6
	Set gtmtypes("mlk_pvtblk",7,"name")="mlk_pvtblk.pvtctl.shrhash_size"
	Set gtmtypes("mlk_pvtblk",7,"off")=40
	Set gtmtypes("mlk_pvtblk",7,"len")=4
	Set gtmtypes("mlk_pvtblk",7,"type")="unsigned-int"
	Set gtmtypfldindx("mlk_pvtblk","pvtctl.shrhash_size")=7
	Set gtmtypes("mlk_pvtblk",8,"name")="mlk_pvtblk.pvtctl.hash_fail_cnt"
	Set gtmtypes("mlk_pvtblk",8,"off")=44
	Set gtmtypes("mlk_pvtblk",8,"len")=4
	Set gtmtypes("mlk_pvtblk",8,"type")="unsigned-int"
	Set gtmtypfldindx("mlk_pvtblk","pvtctl.hash_fail_cnt")=8
	Set gtmtypes("mlk_pvtblk",9,"name")="mlk_pvtblk.nodptr"
	Set gtmtypes("mlk_pvtblk",9,"off")=48
	Set gtmtypes("mlk_pvtblk",9,"len")=8
	Set gtmtypes("mlk_pvtblk",9,"type")="addr"
	Set gtmtypfldindx("mlk_pvtblk","nodptr")=9
	Set gtmtypes("mlk_pvtblk",10,"name")="mlk_pvtblk.blocked"
	Set gtmtypes("mlk_pvtblk",10,"off")=56
	Set gtmtypes("mlk_pvtblk",10,"len")=8
	Set gtmtypes("mlk_pvtblk",10,"type")="addr"
	Set gtmtypfldindx("mlk_pvtblk","blocked")=10
	Set gtmtypes("mlk_pvtblk",11,"name")="mlk_pvtblk.next"
	Set gtmtypes("mlk_pvtblk",11,"off")=64
	Set gtmtypes("mlk_pvtblk",11,"len")=8
	Set gtmtypes("mlk_pvtblk",11,"type")="addr"
	Set gtmtypfldindx("mlk_pvtblk","next")=11
	Set gtmtypes("mlk_pvtblk",12,"name")="mlk_pvtblk.sequence"
	Set gtmtypes("mlk_pvtblk",12,"off")=72
	Set gtmtypes("mlk_pvtblk",12,"len")=4
	Set gtmtypes("mlk_pvtblk",12,"type")="unsigned-int"
	Set gtmtypfldindx("mlk_pvtblk","sequence")=12
	Set gtmtypes("mlk_pvtblk",13,"name")="mlk_pvtblk.blk_sequence"
	Set gtmtypes("mlk_pvtblk",13,"off")=76
	Set gtmtypes("mlk_pvtblk",13,"len")=4
	Set gtmtypes("mlk_pvtblk",13,"type")="unsigned-int"
	Set gtmtypfldindx("mlk_pvtblk","blk_sequence")=13
	Set gtmtypes("mlk_pvtblk",14,"name")="mlk_pvtblk.tp"
	Set gtmtypes("mlk_pvtblk",14,"off")=80
	Set gtmtypes("mlk_pvtblk",14,"len")=8
	Set gtmtypes("mlk_pvtblk",14,"type")="addr"
	Set gtmtypfldindx("mlk_pvtblk","tp")=14
	Set gtmtypes("mlk_pvtblk",15,"name")="mlk_pvtblk.hash_seed"
	Set gtmtypes("mlk_pvtblk",15,"off")=88
	Set gtmtypes("mlk_pvtblk",15,"len")=8
	Set gtmtypes("mlk_pvtblk",15,"type")="gtm_uint8"
	Set gtmtypfldindx("mlk_pvtblk","hash_seed")=15
	Set gtmtypes("mlk_pvtblk",16,"name")="mlk_pvtblk.nref_length"
	Set gtmtypes("mlk_pvtblk",16,"off")=96
	Set gtmtypes("mlk_pvtblk",16,"len")=4
	Set gtmtypes("mlk_pvtblk",16,"type")="unsigned-int"
	Set gtmtypfldindx("mlk_pvtblk","nref_length")=16
	Set gtmtypes("mlk_pvtblk",17,"name")="mlk_pvtblk.subscript_cnt"
	Set gtmtypes("mlk_pvtblk",17,"off")=100
	Set gtmtypes("mlk_pvtblk",17,"len")=2
	Set gtmtypes("mlk_pvtblk",17,"type")="unsigned-short"
	Set gtmtypfldindx("mlk_pvtblk","subscript_cnt")=17
	Set gtmtypes("mlk_pvtblk",18,"name")="mlk_pvtblk.trans"
	Set gtmtypes("mlk_pvtblk",18,"off")=104
	Set gtmtypes("mlk_pvtblk",18,"len")=1
	Set gtmtypes("mlk_pvtblk",18,"type")="unsigned-char"
	Set gtmtypfldindx("mlk_pvtblk","trans")=18
	Set gtmtypes("mlk_pvtblk",19,"name")="mlk_pvtblk.translev"
	Set gtmtypes("mlk_pvtblk",19,"off")=105
	Set gtmtypes("mlk_pvtblk",19,"len")=1
	Set gtmtypes("mlk_pvtblk",19,"type")="unsigned-char"
	Set gtmtypfldindx("mlk_pvtblk","translev")=19
	Set gtmtypes("mlk_pvtblk",20,"name")="mlk_pvtblk.old"
	Set gtmtypes("mlk_pvtblk",20,"off")=106
	Set gtmtypes("mlk_pvtblk",20,"len")=1
	Set gtmtypes("mlk_pvtblk",20,"type")="unsigned-char"
	Set gtmtypfldindx("mlk_pvtblk","old")=20
	Set gtmtypes("mlk_pvtblk",21,"name")="mlk_pvtblk.filler"
	Set gtmtypes("mlk_pvtblk",21,"off")=107
	Set gtmtypes("mlk_pvtblk",21,"len")=5
	Set gtmtypes("mlk_pvtblk",21,"type")="unsigned-char"
	Set gtmtypfldindx("mlk_pvtblk","filler")=21
	Set gtmtypes("mlk_pvtblk",22,"name")="mlk_pvtblk.value"
	Set gtmtypes("mlk_pvtblk",22,"off")=112
	Set gtmtypes("mlk_pvtblk",22,"len")=1
	Set gtmtypes("mlk_pvtblk",22,"type")="unsigned-char"
	Set gtmtypfldindx("mlk_pvtblk","value")=22
	;
	Set gtmtypes("mlk_pvtctl")="struct"
	Set gtmtypes("mlk_pvtctl",0)=7
	Set gtmtypes("mlk_pvtctl","len")=48
	Set gtmtypes("mlk_pvtctl",1,"name")="mlk_pvtctl.region"
	Set gtmtypes("mlk_pvtctl",1,"off")=0
	Set gtmtypes("mlk_pvtctl",1,"len")=8
	Set gtmtypes("mlk_pvtctl",1,"type")="addr"
	Set gtmtypfldindx("mlk_pvtctl","region")=1
	Set gtmtypes("mlk_pvtctl",2,"name")="mlk_pvtctl.csa"
	Set gtmtypes("mlk_pvtctl",2,"off")=8
	Set gtmtypes("mlk_pvtctl",2,"len")=8
	Set gtmtypes("mlk_pvtctl",2,"type")="addr"
	Set gtmtypfldindx("mlk_pvtctl","csa")=2
	Set gtmtypes("mlk_pvtctl",3,"name")="mlk_pvtctl.ctl"
	Set gtmtypes("mlk_pvtctl",3,"off")=16
	Set gtmtypes("mlk_pvtctl",3,"len")=8
	Set gtmtypes("mlk_pvtctl",3,"type")="addr"
	Set gtmtypfldindx("mlk_pvtctl","ctl")=3
	Set gtmtypes("mlk_pvtctl",4,"name")="mlk_pvtctl.shrblk"
	Set gtmtypes("mlk_pvtctl",4,"off")=24
	Set gtmtypes("mlk_pvtctl",4,"len")=8
	Set gtmtypes("mlk_pvtctl",4,"type")="addr"
	Set gtmtypfldindx("mlk_pvtctl","shrblk")=4
	Set gtmtypes("mlk_pvtctl",5,"name")="mlk_pvtctl.shrhash"
	Set gtmtypes("mlk_pvtctl",5,"off")=32
	Set gtmtypes("mlk_pvtctl",5,"len")=8
	Set gtmtypes("mlk_pvtctl",5,"type")="addr"
	Set gtmtypfldindx("mlk_pvtctl","shrhash")=5
	Set gtmtypes("mlk_pvtctl",6,"name")="mlk_pvtctl.shrhash_size"
	Set gtmtypes("mlk_pvtctl",6,"off")=40
	Set gtmtypes("mlk_pvtctl",6,"len")=4
	Set gtmtypes("mlk_pvtctl",6,"type")="unsigned-int"
	Set gtmtypfldindx("mlk_pvtctl","shrhash_size")=6
	Set gtmtypes("mlk_pvtctl",7,"name")="mlk_pvtctl.hash_fail_cnt"
	Set gtmtypes("mlk_pvtctl",7,"off")=44
	Set gtmtypes("mlk_pvtctl",7,"len")=4
	Set gtmtypes("mlk_pvtctl",7,"type")="unsigned-int"
	Set gtmtypfldindx("mlk_pvtctl","hash_fail_cnt")=7
	;
	Set gtmtypes("mlk_shrblk")="struct"
	Set gtmtypes("mlk_shrblk",0)=12
	Set gtmtypes("mlk_shrblk","len")=88
	Set gtmtypes("mlk_shrblk",1,"name")="mlk_shrblk.value"
	Set gtmtypes("mlk_shrblk",1,"off")=0
	Set gtmtypes("mlk_shrblk",1,"len")=8
	Set gtmtypes("mlk_shrblk",1,"type")="intptr_t"
	Set gtmtypfldindx("mlk_shrblk","value")=1
	Set gtmtypes("mlk_shrblk",2,"name")="mlk_shrblk.parent"
	Set gtmtypes("mlk_shrblk",2,"off")=8
	Set gtmtypes("mlk_shrblk",2,"len")=8
	Set gtmtypes("mlk_shrblk",2,"type")="intptr_t"
	Set gtmtypfldindx("mlk_shrblk","parent")=2
	Set gtmtypes("mlk_shrblk",3,"name")="mlk_shrblk.children"
	Set gtmtypes("mlk_shrblk",3,"off")=16
	Set gtmtypes("mlk_shrblk",3,"len")=8
	Set gtmtypes("mlk_shrblk",3,"type")="intptr_t"
	Set gtmtypfldindx("mlk_shrblk","children")=3
	Set gtmtypes("mlk_shrblk",4,"name")="mlk_shrblk.lsib"
	Set gtmtypes("mlk_shrblk",4,"off")=24
	Set gtmtypes("mlk_shrblk",4,"len")=8
	Set gtmtypes("mlk_shrblk",4,"type")="intptr_t"
	Set gtmtypfldindx("mlk_shrblk","lsib")=4
	Set gtmtypes("mlk_shrblk",5,"name")="mlk_shrblk.rsib"
	Set gtmtypes("mlk_shrblk",5,"off")=32
	Set gtmtypes("mlk_shrblk",5,"len")=8
	Set gtmtypes("mlk_shrblk",5,"type")="intptr_t"
	Set gtmtypfldindx("mlk_shrblk","rsib")=5
	Set gtmtypes("mlk_shrblk",6,"name")="mlk_shrblk.pending"
	Set gtmtypes("mlk_shrblk",6,"off")=40
	Set gtmtypes("mlk_shrblk",6,"len")=8
	Set gtmtypes("mlk_shrblk",6,"type")="intptr_t"
	Set gtmtypfldindx("mlk_shrblk","pending")=6
	Set gtmtypes("mlk_shrblk",7,"name")="mlk_shrblk.owner"
	Set gtmtypes("mlk_shrblk",7,"off")=48
	Set gtmtypes("mlk_shrblk",7,"len")=4
	Set gtmtypes("mlk_shrblk",7,"type")="unsigned-int"
	Set gtmtypfldindx("mlk_shrblk","owner")=7
	Set gtmtypes("mlk_shrblk",8,"name")="mlk_shrblk.sequence"
	Set gtmtypes("mlk_shrblk",8,"off")=52
	Set gtmtypes("mlk_shrblk",8,"len")=4
	Set gtmtypes("mlk_shrblk",8,"type")="unsigned-int"
	Set gtmtypfldindx("mlk_shrblk","sequence")=8
	Set gtmtypes("mlk_shrblk",9,"name")="mlk_shrblk.hash"
	Set gtmtypes("mlk_shrblk",9,"off")=56
	Set gtmtypes("mlk_shrblk",9,"len")=4
	Set gtmtypes("mlk_shrblk",9,"type")="unsigned-int"
	Set gtmtypfldindx("mlk_shrblk","hash")=9
	Set gtmtypes("mlk_shrblk",10,"name")="mlk_shrblk.auxpid"
	Set gtmtypes("mlk_shrblk",10,"off")=60
	Set gtmtypes("mlk_shrblk",10,"len")=4
	Set gtmtypes("mlk_shrblk",10,"type")="int"
	Set gtmtypfldindx("mlk_shrblk","auxpid")=10
	Set gtmtypes("mlk_shrblk",11,"name")="mlk_shrblk.auxowner"
	Set gtmtypes("mlk_shrblk",11,"off")=64
	Set gtmtypes("mlk_shrblk",11,"len")=8
	Set gtmtypes("mlk_shrblk",11,"type")="uintptr_t"
	Set gtmtypfldindx("mlk_shrblk","auxowner")=11
	Set gtmtypes("mlk_shrblk",12,"name")="mlk_shrblk.auxnode"
	Set gtmtypes("mlk_shrblk",12,"off")=72
	Set gtmtypes("mlk_shrblk",12,"len")=16
	Set gtmtypes("mlk_shrblk",12,"type")="unsigned-char"
	Set gtmtypfldindx("mlk_shrblk","auxnode")=12
	;
	Set gtmtypes("mlk_shrhash")="struct"
	Set gtmtypes("mlk_shrhash",0)=3
	Set gtmtypes("mlk_shrhash","len")=12
	Set gtmtypes("mlk_shrhash",1,"name")="mlk_shrhash.shrblk_idx"
	Set gtmtypes("mlk_shrhash",1,"off")=0
	Set gtmtypes("mlk_shrhash",1,"len")=4
	Set gtmtypes("mlk_shrhash",1,"type")="unsigned-int"
	Set gtmtypfldindx("mlk_shrhash","shrblk_idx")=1
	Set gtmtypes("mlk_shrhash",2,"name")="mlk_shrhash.usedmap"
	Set gtmtypes("mlk_shrhash",2,"off")=4
	Set gtmtypes("mlk_shrhash",2,"len")=4
	Set gtmtypes("mlk_shrhash",2,"type")="unsigned-int"
	Set gtmtypfldindx("mlk_shrhash","usedmap")=2
	Set gtmtypes("mlk_shrhash",3,"name")="mlk_shrhash.hash"
	Set gtmtypes("mlk_shrhash",3,"off")=8
	Set gtmtypes("mlk_shrhash",3,"len")=4
	Set gtmtypes("mlk_shrhash",3,"type")="unsigned-int"
	Set gtmtypfldindx("mlk_shrhash","hash")=3
	;
	Set gtmtypes("mlk_shrsub")="struct"
	Set gtmtypes("mlk_shrsub",0)=3
	Set gtmtypes("mlk_shrsub","len")=16
	Set gtmtypes("mlk_shrsub",1,"name")="mlk_shrsub.backpointer"
	Set gtmtypes("mlk_shrsub",1,"off")=0
	Set gtmtypes("mlk_shrsub",1,"len")=8
	Set gtmtypes("mlk_shrsub",1,"type")="intptr_t"
	Set gtmtypfldindx("mlk_shrsub","backpointer")=1
	Set gtmtypes("mlk_shrsub",2,"name")="mlk_shrsub.length"
	Set gtmtypes("mlk_shrsub",2,"off")=8
	Set gtmtypes("mlk_shrsub",2,"len")=1
	Set gtmtypes("mlk_shrsub",2,"type")="unsigned-char"
	Set gtmtypfldindx("mlk_shrsub","length")=2
	Set gtmtypes("mlk_shrsub",3,"name")="mlk_shrsub.data"
	Set gtmtypes("mlk_shrsub",3,"off")=9
	Set gtmtypes("mlk_shrsub",3,"len")=1
	Set gtmtypes("mlk_shrsub",3,"type")="unsigned-char"
	Set gtmtypfldindx("mlk_shrsub","data")=3
	;
	Set gtmtypes("mlk_stats_t")="struct"
	Set gtmtypes("mlk_stats_t",0)=2
	Set gtmtypes("mlk_stats_t","len")=16
	Set gtmtypes("mlk_stats_t",1,"name")="mlk_stats_t.n_user_locks_success"
	Set gtmtypes("mlk_stats_t",1,"off")=0
	Set gtmtypes("mlk_stats_t",1,"len")=8
	Set gtmtypes("mlk_stats_t",1,"type")="uint64_t"
	Set gtmtypfldindx("mlk_stats_t","n_user_locks_success")=1
	Set gtmtypes("mlk_stats_t",2,"name")="mlk_stats_t.n_user_locks_fail"
	Set gtmtypes("mlk_stats_t",2,"off")=8
	Set gtmtypes("mlk_stats_t",2,"len")=8
	Set gtmtypes("mlk_stats_t",2,"type")="uint64_t"
	Set gtmtypfldindx("mlk_stats_t","n_user_locks_fail")=2
	;
	Set gtmtypes("mlk_tp")="struct"
	Set gtmtypes("mlk_tp",0)=1
	Set gtmtypes("mlk_tp","len")=16
	Set gtmtypes("mlk_tp",1,"name")="mlk_tp.next"
	Set gtmtypes("mlk_tp",1,"off")=0
	Set gtmtypes("mlk_tp",1,"len")=8
	Set gtmtypes("mlk_tp",1,"type")="addr"
	Set gtmtypfldindx("mlk_tp","next")=1
	;
	Set gtmtypes("mmseg")="struct"
	Set gtmtypes("mmseg",0)=3
	Set gtmtypes("mmseg","len")=24
	Set gtmtypes("mmseg",1,"name")="mmseg.next"
	Set gtmtypes("mmseg",1,"off")=0
	Set gtmtypes("mmseg",1,"len")=8
	Set gtmtypes("mmseg",1,"type")="addr"
	Set gtmtypfldindx("mmseg","next")=1
	Set gtmtypes("mmseg",2,"name")="mmseg.begin"
	Set gtmtypes("mmseg",2,"off")=8
	Set gtmtypes("mmseg",2,"len")=8
	Set gtmtypes("mmseg",2,"type")="addr"
	Set gtmtypfldindx("mmseg","begin")=2
	Set gtmtypes("mmseg",3,"name")="mmseg.end"
	Set gtmtypes("mmseg",3,"off")=16
	Set gtmtypes("mmseg",3,"len")=8
	Set gtmtypes("mmseg",3,"type")="addr"
	Set gtmtypfldindx("mmseg","end")=3
	;
	Set gtmtypes("mname_entry")="struct"
	Set gtmtypes("mname_entry",0)=6
	Set gtmtypes("mname_entry","len")=24
	Set gtmtypes("mname_entry",1,"name")="mname_entry.var_name"
	Set gtmtypes("mname_entry",1,"off")=0
	Set gtmtypes("mname_entry",1,"len")=16
	Set gtmtypes("mname_entry",1,"type")="mstr"
	Set gtmtypfldindx("mname_entry","var_name")=1
	Set gtmtypes("mname_entry",2,"name")="mname_entry.var_name.char_len"
	Set gtmtypes("mname_entry",2,"off")=0
	Set gtmtypes("mname_entry",2,"len")=4
	Set gtmtypes("mname_entry",2,"type")="unsigned-int"
	Set gtmtypfldindx("mname_entry","var_name.char_len")=2
	Set gtmtypes("mname_entry",3,"name")="mname_entry.var_name.len"
	Set gtmtypes("mname_entry",3,"off")=4
	Set gtmtypes("mname_entry",3,"len")=4
	Set gtmtypes("mname_entry",3,"type")="int"
	Set gtmtypfldindx("mname_entry","var_name.len")=3
	Set gtmtypes("mname_entry",4,"name")="mname_entry.var_name.addr"
	Set gtmtypes("mname_entry",4,"off")=8
	Set gtmtypes("mname_entry",4,"len")=8
	Set gtmtypes("mname_entry",4,"type")="addr"
	Set gtmtypfldindx("mname_entry","var_name.addr")=4
	Set gtmtypes("mname_entry",5,"name")="mname_entry.hash_code"
	Set gtmtypes("mname_entry",5,"off")=16
	Set gtmtypes("mname_entry",5,"len")=4
	Set gtmtypes("mname_entry",5,"type")="unsigned-int"
	Set gtmtypfldindx("mname_entry","hash_code")=5
	Set gtmtypes("mname_entry",6,"name")="mname_entry.marked"
	Set gtmtypes("mname_entry",6,"off")=20
	Set gtmtypes("mname_entry",6,"len")=4
	Set gtmtypes("mname_entry",6,"type")="boolean_t"
	Set gtmtypfldindx("mname_entry","marked")=6
	;
	Set gtmtypes("modrm_byte_type")="union"
	Set gtmtypes("modrm_byte_type",0)=2
	Set gtmtypes("modrm_byte_type","len")=1
	Set gtmtypes("modrm_byte_type",1,"name")="modrm_byte_type.modrm"
	Set gtmtypes("modrm_byte_type",1,"off")=0
	Set gtmtypes("modrm_byte_type",1,"len")=1
	Set gtmtypes("modrm_byte_type",1,"type")="ModR_M"
	Set gtmtypfldindx("modrm_byte_type","modrm")=1
	Set gtmtypes("modrm_byte_type",2,"name")="modrm_byte_type.byte"
	Set gtmtypes("modrm_byte_type",2,"off")=0
	Set gtmtypes("modrm_byte_type",2,"len")=1
	Set gtmtypes("modrm_byte_type",2,"type")="unsigned-char"
	Set gtmtypfldindx("modrm_byte_type","byte")=2
	;
	Set gtmtypes("mprof_stack_frame")="struct"
	Set gtmtypes("mprof_stack_frame",0)=7
	Set gtmtypes("mprof_stack_frame","len")=88
	Set gtmtypes("mprof_stack_frame",1,"name")="mprof_stack_frame.prev"
	Set gtmtypes("mprof_stack_frame",1,"off")=0
	Set gtmtypes("mprof_stack_frame",1,"len")=8
	Set gtmtypes("mprof_stack_frame",1,"type")="addr"
	Set gtmtypfldindx("mprof_stack_frame","prev")=1
	Set gtmtypes("mprof_stack_frame",2,"name")="mprof_stack_frame.rout_name"
	Set gtmtypes("mprof_stack_frame",2,"off")=8
	Set gtmtypes("mprof_stack_frame",2,"len")=8
	Set gtmtypes("mprof_stack_frame",2,"type")="addr"
	Set gtmtypfldindx("mprof_stack_frame","rout_name")=2
	Set gtmtypes("mprof_stack_frame",3,"name")="mprof_stack_frame.label_name"
	Set gtmtypes("mprof_stack_frame",3,"off")=16
	Set gtmtypes("mprof_stack_frame",3,"len")=8
	Set gtmtypes("mprof_stack_frame",3,"type")="addr"
	Set gtmtypfldindx("mprof_stack_frame","label_name")=3
	Set gtmtypes("mprof_stack_frame",4,"name")="mprof_stack_frame.curr_node"
	Set gtmtypes("mprof_stack_frame",4,"off")=24
	Set gtmtypes("mprof_stack_frame",4,"len")=8
	Set gtmtypes("mprof_stack_frame",4,"type")="addr"
	Set gtmtypfldindx("mprof_stack_frame","curr_node")=4
	Set gtmtypes("mprof_stack_frame",5,"name")="mprof_stack_frame.start"
	Set gtmtypes("mprof_stack_frame",5,"off")=32
	Set gtmtypes("mprof_stack_frame",5,"len")=24
	Set gtmtypes("mprof_stack_frame",5,"type")="struct"
	Set gtmtypfldindx("mprof_stack_frame","start")=5
	Set gtmtypes("mprof_stack_frame",6,"name")="mprof_stack_frame.carryover"
	Set gtmtypes("mprof_stack_frame",6,"off")=56
	Set gtmtypes("mprof_stack_frame",6,"len")=24
	Set gtmtypes("mprof_stack_frame",6,"type")="struct"
	Set gtmtypfldindx("mprof_stack_frame","carryover")=6
	Set gtmtypes("mprof_stack_frame",7,"name")="mprof_stack_frame.dummy_stack_count"
	Set gtmtypes("mprof_stack_frame",7,"off")=80
	Set gtmtypes("mprof_stack_frame",7,"len")=4
	Set gtmtypes("mprof_stack_frame",7,"type")="int"
	Set gtmtypfldindx("mprof_stack_frame","dummy_stack_count")=7
	;
	Set gtmtypes("mprof_tree")="struct"
	Set gtmtypes("mprof_tree",0)=14
	Set gtmtypes("mprof_tree","len")=96
	Set gtmtypes("mprof_tree",1,"name")="mprof_tree.e"
	Set gtmtypes("mprof_tree",1,"off")=0
	Set gtmtypes("mprof_tree",1,"len")=64
	Set gtmtypes("mprof_tree",1,"type")="trace_entry"
	Set gtmtypfldindx("mprof_tree","e")=1
	Set gtmtypes("mprof_tree",2,"name")="mprof_tree.e.rout_name"
	Set gtmtypes("mprof_tree",2,"off")=0
	Set gtmtypes("mprof_tree",2,"len")=8
	Set gtmtypes("mprof_tree",2,"type")="addr"
	Set gtmtypfldindx("mprof_tree","e.rout_name")=2
	Set gtmtypes("mprof_tree",3,"name")="mprof_tree.e.label_name"
	Set gtmtypes("mprof_tree",3,"off")=8
	Set gtmtypes("mprof_tree",3,"len")=8
	Set gtmtypes("mprof_tree",3,"type")="addr"
	Set gtmtypfldindx("mprof_tree","e.label_name")=3
	Set gtmtypes("mprof_tree",4,"name")="mprof_tree.e.line_num"
	Set gtmtypes("mprof_tree",4,"off")=16
	Set gtmtypes("mprof_tree",4,"len")=4
	Set gtmtypes("mprof_tree",4,"type")="int"
	Set gtmtypfldindx("mprof_tree","e.line_num")=4
	Set gtmtypes("mprof_tree",5,"name")="mprof_tree.e.count"
	Set gtmtypes("mprof_tree",5,"off")=20
	Set gtmtypes("mprof_tree",5,"len")=4
	Set gtmtypes("mprof_tree",5,"type")="unsigned"
	Set gtmtypfldindx("mprof_tree","e.count")=5
	Set gtmtypes("mprof_tree",6,"name")="mprof_tree.e.sys_time"
	Set gtmtypes("mprof_tree",6,"off")=24
	Set gtmtypes("mprof_tree",6,"len")=8
	Set gtmtypes("mprof_tree",6,"type")="uint64_t"
	Set gtmtypfldindx("mprof_tree","e.sys_time")=6
	Set gtmtypes("mprof_tree",7,"name")="mprof_tree.e.usr_time"
	Set gtmtypes("mprof_tree",7,"off")=32
	Set gtmtypes("mprof_tree",7,"len")=8
	Set gtmtypes("mprof_tree",7,"type")="uint64_t"
	Set gtmtypfldindx("mprof_tree","e.usr_time")=7
	Set gtmtypes("mprof_tree",8,"name")="mprof_tree.e.elp_time"
	Set gtmtypes("mprof_tree",8,"off")=40
	Set gtmtypes("mprof_tree",8,"len")=8
	Set gtmtypes("mprof_tree",8,"type")="uint64_t"
	Set gtmtypfldindx("mprof_tree","e.elp_time")=8
	Set gtmtypes("mprof_tree",9,"name")="mprof_tree.e.loop_level"
	Set gtmtypes("mprof_tree",9,"off")=48
	Set gtmtypes("mprof_tree",9,"len")=4
	Set gtmtypes("mprof_tree",9,"type")="int"
	Set gtmtypfldindx("mprof_tree","e.loop_level")=9
	Set gtmtypes("mprof_tree",10,"name")="mprof_tree.e.raddr"
	Set gtmtypes("mprof_tree",10,"off")=56
	Set gtmtypes("mprof_tree",10,"len")=8
	Set gtmtypes("mprof_tree",10,"type")="addr"
	Set gtmtypfldindx("mprof_tree","e.raddr")=10
	Set gtmtypes("mprof_tree",11,"name")="mprof_tree.link"
	Set gtmtypes("mprof_tree",11,"off")=64
	Set gtmtypes("mprof_tree",11,"len")=16
	Set gtmtypes("mprof_tree",11,"type")="addr"
	Set gtmtypfldindx("mprof_tree","link")=11
	Set gtmtypes("mprof_tree",11,"dim")=2
	Set gtmtypes("mprof_tree",12,"name")="mprof_tree.loop_link"
	Set gtmtypes("mprof_tree",12,"off")=80
	Set gtmtypes("mprof_tree",12,"len")=8
	Set gtmtypes("mprof_tree",12,"type")="addr"
	Set gtmtypfldindx("mprof_tree","loop_link")=12
	Set gtmtypes("mprof_tree",13,"name")="mprof_tree.desc_dir"
	Set gtmtypes("mprof_tree",13,"off")=88
	Set gtmtypes("mprof_tree",13,"len")=4
	Set gtmtypes("mprof_tree",13,"type")="int"
	Set gtmtypfldindx("mprof_tree","desc_dir")=13
	Set gtmtypes("mprof_tree",14,"name")="mprof_tree.ins_path_hint"
	Set gtmtypes("mprof_tree",14,"off")=92
	Set gtmtypes("mprof_tree",14,"len")=4
	Set gtmtypes("mprof_tree",14,"type")="int"
	Set gtmtypfldindx("mprof_tree","ins_path_hint")=14
	;
	Set gtmtypes("mprof_wrapper")="struct"
	Set gtmtypes("mprof_wrapper",0)=24
	Set gtmtypes("mprof_wrapper","len")=1424
	Set gtmtypes("mprof_wrapper",1,"name")="mprof_wrapper.tprev"
	Set gtmtypes("mprof_wrapper",1,"off")=0
	Set gtmtypes("mprof_wrapper",1,"len")=24
	Set gtmtypes("mprof_wrapper",1,"type")="struct"
	Set gtmtypfldindx("mprof_wrapper","tprev")=1
	Set gtmtypes("mprof_wrapper",2,"name")="mprof_wrapper.tcurr"
	Set gtmtypes("mprof_wrapper",2,"off")=24
	Set gtmtypes("mprof_wrapper",2,"len")=24
	Set gtmtypes("mprof_wrapper",2,"type")="struct"
	Set gtmtypfldindx("mprof_wrapper","tcurr")=2
	Set gtmtypes("mprof_wrapper",3,"name")="mprof_wrapper.head_tblnd"
	Set gtmtypes("mprof_wrapper",3,"off")=48
	Set gtmtypes("mprof_wrapper",3,"len")=8
	Set gtmtypes("mprof_wrapper",3,"type")="addr"
	Set gtmtypfldindx("mprof_wrapper","head_tblnd")=3
	Set gtmtypes("mprof_wrapper",4,"name")="mprof_wrapper.curr_tblnd"
	Set gtmtypes("mprof_wrapper",4,"off")=56
	Set gtmtypes("mprof_wrapper",4,"len")=8
	Set gtmtypes("mprof_wrapper",4,"type")="addr"
	Set gtmtypfldindx("mprof_wrapper","curr_tblnd")=4
	Set gtmtypes("mprof_wrapper",5,"name")="mprof_wrapper.curr_num_subscripts"
	Set gtmtypes("mprof_wrapper",5,"off")=64
	Set gtmtypes("mprof_wrapper",5,"len")=4
	Set gtmtypes("mprof_wrapper",5,"type")="int"
	Set gtmtypfldindx("mprof_wrapper","curr_num_subscripts")=5
	Set gtmtypes("mprof_wrapper",6,"name")="mprof_wrapper.pcavailptr"
	Set gtmtypes("mprof_wrapper",6,"off")=72
	Set gtmtypes("mprof_wrapper",6,"len")=8
	Set gtmtypes("mprof_wrapper",6,"type")="addr"
	Set gtmtypfldindx("mprof_wrapper","pcavailptr")=6
	Set gtmtypes("mprof_wrapper",7,"name")="mprof_wrapper.pcavailbase"
	Set gtmtypes("mprof_wrapper",7,"off")=80
	Set gtmtypes("mprof_wrapper",7,"len")=8
	Set gtmtypes("mprof_wrapper",7,"type")="addr"
	Set gtmtypfldindx("mprof_wrapper","pcavailbase")=7
	Set gtmtypes("mprof_wrapper",8,"name")="mprof_wrapper.pcavail"
	Set gtmtypes("mprof_wrapper",8,"off")=88
	Set gtmtypes("mprof_wrapper",8,"len")=4
	Set gtmtypes("mprof_wrapper",8,"type")="int"
	Set gtmtypfldindx("mprof_wrapper","pcavail")=8
	Set gtmtypes("mprof_wrapper",9,"name")="mprof_wrapper.is_tracing_ini"
	Set gtmtypes("mprof_wrapper",9,"off")=92
	Set gtmtypes("mprof_wrapper",9,"len")=4
	Set gtmtypes("mprof_wrapper",9,"type")="boolean_t"
	Set gtmtypfldindx("mprof_wrapper","is_tracing_ini")=9
	Set gtmtypes("mprof_wrapper",10,"name")="mprof_wrapper.subsc"
	Set gtmtypes("mprof_wrapper",10,"off")=96
	Set gtmtypes("mprof_wrapper",10,"len")=1024
	Set gtmtypes("mprof_wrapper",10,"type")="mval"
	Set gtmtypfldindx("mprof_wrapper","subsc")=10
	Set gtmtypes("mprof_wrapper",10,"dim")=32
	Set gtmtypes("mprof_wrapper",11,"name")="mprof_wrapper.subsc[0].mvtype"
	Set gtmtypes("mprof_wrapper",11,"off")=96
	Set gtmtypes("mprof_wrapper",11,"len")=2
	Set gtmtypes("mprof_wrapper",11,"type")="unsigned-short"
	Set gtmtypfldindx("mprof_wrapper","subsc[0].mvtype")=11
	Set gtmtypes("mprof_wrapper",12,"name")="mprof_wrapper.subsc[0].fnpc_indx"
	Set gtmtypes("mprof_wrapper",12,"off")=99
	Set gtmtypes("mprof_wrapper",12,"len")=1
	Set gtmtypes("mprof_wrapper",12,"type")="unsigned-char"
	Set gtmtypfldindx("mprof_wrapper","subsc[0].fnpc_indx")=12
	Set gtmtypes("mprof_wrapper",13,"name")="mprof_wrapper.subsc[0].utfcgr_indx"
	Set gtmtypes("mprof_wrapper",13,"off")=100
	Set gtmtypes("mprof_wrapper",13,"len")=4
	Set gtmtypes("mprof_wrapper",13,"type")="unsigned-int"
	Set gtmtypfldindx("mprof_wrapper","subsc[0].utfcgr_indx")=13
	Set gtmtypes("mprof_wrapper",14,"name")="mprof_wrapper.subsc[0].m"
	Set gtmtypes("mprof_wrapper",14,"off")=104
	Set gtmtypes("mprof_wrapper",14,"len")=8
	Set gtmtypes("mprof_wrapper",14,"type")="int"
	Set gtmtypfldindx("mprof_wrapper","subsc[0].m")=14
	Set gtmtypes("mprof_wrapper",14,"dim")=2
	Set gtmtypes("mprof_wrapper",15,"name")="mprof_wrapper.subsc[0].str"
	Set gtmtypes("mprof_wrapper",15,"off")=112
	Set gtmtypes("mprof_wrapper",15,"len")=16
	Set gtmtypes("mprof_wrapper",15,"type")="mstr"
	Set gtmtypfldindx("mprof_wrapper","subsc[0].str")=15
	Set gtmtypes("mprof_wrapper",16,"name")="mprof_wrapper.gvargs"
	Set gtmtypes("mprof_wrapper",16,"off")=1120
	Set gtmtypes("mprof_wrapper",16,"len")=272
	Set gtmtypes("mprof_wrapper",16,"type")="gvargs_t"
	Set gtmtypfldindx("mprof_wrapper","gvargs")=16
	Set gtmtypes("mprof_wrapper",17,"name")="mprof_wrapper.gvargs.count"
	Set gtmtypes("mprof_wrapper",17,"off")=1120
	Set gtmtypes("mprof_wrapper",17,"len")=8
	Set gtmtypes("mprof_wrapper",17,"type")="ssize_t"
	Set gtmtypfldindx("mprof_wrapper","gvargs.count")=17
	Set gtmtypes("mprof_wrapper",18,"name")="mprof_wrapper.gvargs.args"
	Set gtmtypes("mprof_wrapper",18,"off")=1128
	Set gtmtypes("mprof_wrapper",18,"len")=264
	Set gtmtypes("mprof_wrapper",18,"type")="addr"
	Set gtmtypfldindx("mprof_wrapper","gvargs.args")=18
	Set gtmtypes("mprof_wrapper",18,"dim")=33
	Set gtmtypes("mprof_wrapper",19,"name")="mprof_wrapper.gbl_to_fill"
	Set gtmtypes("mprof_wrapper",19,"off")=1392
	Set gtmtypes("mprof_wrapper",19,"len")=32
	Set gtmtypes("mprof_wrapper",19,"type")="mval"
	Set gtmtypfldindx("mprof_wrapper","gbl_to_fill")=19
	Set gtmtypes("mprof_wrapper",20,"name")="mprof_wrapper.gbl_to_fill.mvtype"
	Set gtmtypes("mprof_wrapper",20,"off")=1392
	Set gtmtypes("mprof_wrapper",20,"len")=2
	Set gtmtypes("mprof_wrapper",20,"type")="unsigned-short"
	Set gtmtypfldindx("mprof_wrapper","gbl_to_fill.mvtype")=20
	Set gtmtypes("mprof_wrapper",21,"name")="mprof_wrapper.gbl_to_fill.fnpc_indx"
	Set gtmtypes("mprof_wrapper",21,"off")=1395
	Set gtmtypes("mprof_wrapper",21,"len")=1
	Set gtmtypes("mprof_wrapper",21,"type")="unsigned-char"
	Set gtmtypfldindx("mprof_wrapper","gbl_to_fill.fnpc_indx")=21
	Set gtmtypes("mprof_wrapper",22,"name")="mprof_wrapper.gbl_to_fill.utfcgr_indx"
	Set gtmtypes("mprof_wrapper",22,"off")=1396
	Set gtmtypes("mprof_wrapper",22,"len")=4
	Set gtmtypes("mprof_wrapper",22,"type")="unsigned-int"
	Set gtmtypfldindx("mprof_wrapper","gbl_to_fill.utfcgr_indx")=22
	Set gtmtypes("mprof_wrapper",23,"name")="mprof_wrapper.gbl_to_fill.m"
	Set gtmtypes("mprof_wrapper",23,"off")=1400
	Set gtmtypes("mprof_wrapper",23,"len")=8
	Set gtmtypes("mprof_wrapper",23,"type")="int"
	Set gtmtypfldindx("mprof_wrapper","gbl_to_fill.m")=23
	Set gtmtypes("mprof_wrapper",23,"dim")=2
	Set gtmtypes("mprof_wrapper",24,"name")="mprof_wrapper.gbl_to_fill.str"
	Set gtmtypes("mprof_wrapper",24,"off")=1408
	Set gtmtypes("mprof_wrapper",24,"len")=16
	Set gtmtypes("mprof_wrapper",24,"type")="mstr"
	Set gtmtypfldindx("mprof_wrapper","gbl_to_fill.str")=24
	;
	Set gtmtypes("mstr")="struct"
	Set gtmtypes("mstr",0)=3
	Set gtmtypes("mstr","len")=16
	Set gtmtypes("mstr",1,"name")="mstr.char_len"
	Set gtmtypes("mstr",1,"off")=0
	Set gtmtypes("mstr",1,"len")=4
	Set gtmtypes("mstr",1,"type")="unsigned-int"
	Set gtmtypfldindx("mstr","char_len")=1
	Set gtmtypes("mstr",2,"name")="mstr.len"
	Set gtmtypes("mstr",2,"off")=4
	Set gtmtypes("mstr",2,"len")=4
	Set gtmtypes("mstr",2,"type")="int"
	Set gtmtypfldindx("mstr","len")=2
	Set gtmtypes("mstr",3,"name")="mstr.addr"
	Set gtmtypes("mstr",3,"off")=8
	Set gtmtypes("mstr",3,"len")=8
	Set gtmtypes("mstr",3,"type")="addr"
	Set gtmtypfldindx("mstr","addr")=3
	;
	Set gtmtypes("mu_extr_stats")="struct"
	Set gtmtypes("mu_extr_stats",0)=4
	Set gtmtypes("mu_extr_stats","len")=24
	Set gtmtypes("mu_extr_stats",1,"name")="mu_extr_stats.recknt"
	Set gtmtypes("mu_extr_stats",1,"off")=0
	Set gtmtypes("mu_extr_stats",1,"len")=8
	Set gtmtypes("mu_extr_stats",1,"type")="uint64_t"
	Set gtmtypfldindx("mu_extr_stats","recknt")=1
	Set gtmtypes("mu_extr_stats",2,"name")="mu_extr_stats.reclen"
	Set gtmtypes("mu_extr_stats",2,"off")=8
	Set gtmtypes("mu_extr_stats",2,"len")=4
	Set gtmtypes("mu_extr_stats",2,"type")="unsigned-int"
	Set gtmtypfldindx("mu_extr_stats","reclen")=2
	Set gtmtypes("mu_extr_stats",3,"name")="mu_extr_stats.keylen"
	Set gtmtypes("mu_extr_stats",3,"off")=12
	Set gtmtypes("mu_extr_stats",3,"len")=4
	Set gtmtypes("mu_extr_stats",3,"type")="unsigned-int"
	Set gtmtypfldindx("mu_extr_stats","keylen")=3
	Set gtmtypes("mu_extr_stats",4,"name")="mu_extr_stats.datalen"
	Set gtmtypes("mu_extr_stats",4,"off")=16
	Set gtmtypes("mu_extr_stats",4,"len")=4
	Set gtmtypes("mu_extr_stats",4,"type")="unsigned-int"
	Set gtmtypfldindx("mu_extr_stats","datalen")=4
	;
	Set gtmtypes("mu_set_rlist")="struct"
	Set gtmtypes("mu_set_rlist",0)=11
	Set gtmtypes("mu_set_rlist","len")=80
	Set gtmtypes("mu_set_rlist",1,"name")="mu_set_rlist.fPtr"
	Set gtmtypes("mu_set_rlist",1,"off")=0
	Set gtmtypes("mu_set_rlist",1,"len")=8
	Set gtmtypes("mu_set_rlist",1,"type")="addr"
	Set gtmtypfldindx("mu_set_rlist","fPtr")=1
	Set gtmtypes("mu_set_rlist",2,"name")="mu_set_rlist.reg"
	Set gtmtypes("mu_set_rlist",2,"off")=8
	Set gtmtypes("mu_set_rlist",2,"len")=8
	Set gtmtypes("mu_set_rlist",2,"type")="addr"
	Set gtmtypfldindx("mu_set_rlist","reg")=2
	Set gtmtypes("mu_set_rlist",3,"name")="mu_set_rlist.unique_id"
	Set gtmtypes("mu_set_rlist",3,"off")=16
	Set gtmtypes("mu_set_rlist",3,"len")=20
	Set gtmtypes("mu_set_rlist",3,"type")="char"
	Set gtmtypfldindx("mu_set_rlist","unique_id")=3
	Set gtmtypes("mu_set_rlist",4,"name")="mu_set_rlist.fid_index"
	Set gtmtypes("mu_set_rlist",4,"off")=36
	Set gtmtypes("mu_set_rlist",4,"len")=4
	Set gtmtypes("mu_set_rlist",4,"type")="int"
	Set gtmtypfldindx("mu_set_rlist","fid_index")=4
	Set gtmtypes("mu_set_rlist",5,"name")="mu_set_rlist.state"
	Set gtmtypes("mu_set_rlist",5,"off")=40
	Set gtmtypes("mu_set_rlist",5,"len")=4
	Set gtmtypes("mu_set_rlist",5,"type")="int"
	Set gtmtypfldindx("mu_set_rlist","state")=5
	Set gtmtypes("mu_set_rlist",6,"name")="mu_set_rlist.sd"
	Set gtmtypes("mu_set_rlist",6,"off")=48
	Set gtmtypes("mu_set_rlist",6,"len")=8
	Set gtmtypes("mu_set_rlist",6,"type")="addr"
	Set gtmtypfldindx("mu_set_rlist","sd")=6
	Set gtmtypes("mu_set_rlist",7,"name")="mu_set_rlist.exclusive"
	Set gtmtypes("mu_set_rlist",7,"off")=56
	Set gtmtypes("mu_set_rlist",7,"len")=1
	Set gtmtypes("mu_set_rlist",7,"type")="char"
	Set gtmtypfldindx("mu_set_rlist","exclusive")=7
	Set gtmtypes("mu_set_rlist",8,"name")="mu_set_rlist.fd"
	Set gtmtypes("mu_set_rlist",8,"off")=60
	Set gtmtypes("mu_set_rlist",8,"len")=4
	Set gtmtypes("mu_set_rlist",8,"type")="int"
	Set gtmtypfldindx("mu_set_rlist","fd")=8
	Set gtmtypes("mu_set_rlist",9,"name")="mu_set_rlist.jnl_new_state"
	Set gtmtypes("mu_set_rlist",9,"off")=64
	Set gtmtypes("mu_set_rlist",9,"len")=4
	Set gtmtypes("mu_set_rlist",9,"type")="int"
	Set gtmtypfldindx("mu_set_rlist","jnl_new_state")=9
	Set gtmtypes("mu_set_rlist",10,"name")="mu_set_rlist.repl_new_state"
	Set gtmtypes("mu_set_rlist",10,"off")=68
	Set gtmtypes("mu_set_rlist",10,"len")=4
	Set gtmtypes("mu_set_rlist",10,"type")="int"
	Set gtmtypfldindx("mu_set_rlist","repl_new_state")=10
	Set gtmtypes("mu_set_rlist",11,"name")="mu_set_rlist.before_images"
	Set gtmtypes("mu_set_rlist",11,"off")=72
	Set gtmtypes("mu_set_rlist",11,"len")=4
	Set gtmtypes("mu_set_rlist",11,"type")="boolean_t"
	Set gtmtypfldindx("mu_set_rlist","before_images")=11
	;
	Set gtmtypes("muinc_blk_hdr")="struct"
	Set gtmtypes("muinc_blk_hdr",0)=9
	Set gtmtypes("muinc_blk_hdr","len")=32
	Set gtmtypes("muinc_blk_hdr",1,"name")="muinc_blk_hdr.filler_8byte"
	Set gtmtypes("muinc_blk_hdr",1,"off")=0
	Set gtmtypes("muinc_blk_hdr",1,"len")=8
	Set gtmtypes("muinc_blk_hdr",1,"type")="char"
	Set gtmtypfldindx("muinc_blk_hdr","filler_8byte")=1
	Set gtmtypes("muinc_blk_hdr",2,"name")="muinc_blk_hdr.blktype"
	Set gtmtypes("muinc_blk_hdr",2,"off")=8
	Set gtmtypes("muinc_blk_hdr",2,"len")=4
	Set gtmtypes("muinc_blk_hdr",2,"type")="int"
	Set gtmtypfldindx("muinc_blk_hdr","blktype")=2
	Set gtmtypes("muinc_blk_hdr",3,"name")="muinc_blk_hdr.blkid"
	Set gtmtypes("muinc_blk_hdr",3,"off")=12
	Set gtmtypes("muinc_blk_hdr",3,"len")=4
	Set gtmtypes("muinc_blk_hdr",3,"type")="int"
	Set gtmtypfldindx("muinc_blk_hdr","blkid")=3
	Set gtmtypes("muinc_blk_hdr",4,"name")="muinc_blk_hdr.use"
	Set gtmtypes("muinc_blk_hdr",4,"off")=16
	Set gtmtypes("muinc_blk_hdr",4,"len")=4
	Set gtmtypes("muinc_blk_hdr",4,"type")="union"
	Set gtmtypfldindx("muinc_blk_hdr","use")=4
	Set gtmtypes("muinc_blk_hdr",5,"name")="muinc_blk_hdr.use.bkup"
	Set gtmtypes("muinc_blk_hdr",5,"off")=16
	Set gtmtypes("muinc_blk_hdr",5,"len")=4
	Set gtmtypes("muinc_blk_hdr",5,"type")="struct"
	Set gtmtypfldindx("muinc_blk_hdr","use.bkup")=5
	Set gtmtypes("muinc_blk_hdr",6,"name")="muinc_blk_hdr.use.bkup.ondsk_blkver"
	Set gtmtypes("muinc_blk_hdr",6,"off")=16
	Set gtmtypes("muinc_blk_hdr",6,"len")=4
	Set gtmtypes("muinc_blk_hdr",6,"type")="int"
	Set gtmtypfldindx("muinc_blk_hdr","use.bkup.ondsk_blkver")=6
	Set gtmtypes("muinc_blk_hdr",7,"name")="muinc_blk_hdr.holder_pid"
	Set gtmtypes("muinc_blk_hdr",7,"off")=20
	Set gtmtypes("muinc_blk_hdr",7,"len")=4
	Set gtmtypes("muinc_blk_hdr",7,"type")="pid_t"
	Set gtmtypfldindx("muinc_blk_hdr","holder_pid")=7
	Set gtmtypes("muinc_blk_hdr",8,"name")="muinc_blk_hdr.valid_data"
	Set gtmtypes("muinc_blk_hdr",8,"off")=24
	Set gtmtypes("muinc_blk_hdr",8,"len")=4
	Set gtmtypes("muinc_blk_hdr",8,"type")="boolean_t"
	Set gtmtypfldindx("muinc_blk_hdr","valid_data")=8
	Set gtmtypes("muinc_blk_hdr",9,"name")="muinc_blk_hdr.image_count"
	Set gtmtypes("muinc_blk_hdr",9,"off")=28
	Set gtmtypes("muinc_blk_hdr",9,"len")=4
	Set gtmtypes("muinc_blk_hdr",9,"type")="int"
	Set gtmtypfldindx("muinc_blk_hdr","image_count")=9
	;
	Set gtmtypes("multi_proc_shm_hdr_t")="struct"
	Set gtmtypes("multi_proc_shm_hdr_t",0)=22
	Set gtmtypes("multi_proc_shm_hdr_t","len")=12080
	Set gtmtypes("multi_proc_shm_hdr_t",1,"name")="multi_proc_shm_hdr_t.multi_proc_latch"
	Set gtmtypes("multi_proc_shm_hdr_t",1,"off")=0
	Set gtmtypes("multi_proc_shm_hdr_t",1,"len")=8
	Set gtmtypes("multi_proc_shm_hdr_t",1,"type")="global_latch_t"
	Set gtmtypfldindx("multi_proc_shm_hdr_t","multi_proc_latch")=1
	Set gtmtypes("multi_proc_shm_hdr_t",2,"name")="multi_proc_shm_hdr_t.multi_proc_latch.u"
	Set gtmtypes("multi_proc_shm_hdr_t",2,"off")=0
	Set gtmtypes("multi_proc_shm_hdr_t",2,"len")=8
	Set gtmtypes("multi_proc_shm_hdr_t",2,"type")="union"
	Set gtmtypfldindx("multi_proc_shm_hdr_t","multi_proc_latch.u")=2
	Set gtmtypes("multi_proc_shm_hdr_t",3,"name")="multi_proc_shm_hdr_t.multi_proc_latch.u.pid_imgcnt"
	Set gtmtypes("multi_proc_shm_hdr_t",3,"off")=0
	Set gtmtypes("multi_proc_shm_hdr_t",3,"len")=8
	Set gtmtypes("multi_proc_shm_hdr_t",3,"type")="uint64_t"
	Set gtmtypfldindx("multi_proc_shm_hdr_t","multi_proc_latch.u.pid_imgcnt")=3
	Set gtmtypes("multi_proc_shm_hdr_t",4,"name")="multi_proc_shm_hdr_t.multi_proc_latch.u.parts"
	Set gtmtypes("multi_proc_shm_hdr_t",4,"off")=0
	Set gtmtypes("multi_proc_shm_hdr_t",4,"len")=8
	Set gtmtypes("multi_proc_shm_hdr_t",4,"type")="struct"
	Set gtmtypfldindx("multi_proc_shm_hdr_t","multi_proc_latch.u.parts")=4
	Set gtmtypes("multi_proc_shm_hdr_t",5,"name")="multi_proc_shm_hdr_t.multi_proc_latch.u.parts.latch_pid"
	Set gtmtypes("multi_proc_shm_hdr_t",5,"off")=0
	Set gtmtypes("multi_proc_shm_hdr_t",5,"len")=4
	Set gtmtypes("multi_proc_shm_hdr_t",5,"type")="int"
	Set gtmtypfldindx("multi_proc_shm_hdr_t","multi_proc_latch.u.parts.latch_pid")=5
	Set gtmtypes("multi_proc_shm_hdr_t",6,"name")="multi_proc_shm_hdr_t.multi_proc_latch.u.parts.latch_word"
	Set gtmtypes("multi_proc_shm_hdr_t",6,"off")=4
	Set gtmtypes("multi_proc_shm_hdr_t",6,"len")=4
	Set gtmtypes("multi_proc_shm_hdr_t",6,"type")="int"
	Set gtmtypfldindx("multi_proc_shm_hdr_t","multi_proc_latch.u.parts.latch_word")=6
	Set gtmtypes("multi_proc_shm_hdr_t",7,"name")="multi_proc_shm_hdr_t.forced_multi_proc_exit"
	Set gtmtypes("multi_proc_shm_hdr_t",7,"off")=8
	Set gtmtypes("multi_proc_shm_hdr_t",7,"len")=4
	Set gtmtypes("multi_proc_shm_hdr_t",7,"type")="boolean_t"
	Set gtmtypfldindx("multi_proc_shm_hdr_t","forced_multi_proc_exit")=7
	Set gtmtypes("multi_proc_shm_hdr_t",8,"name")="multi_proc_shm_hdr_t.wait_done"
	Set gtmtypes("multi_proc_shm_hdr_t",8,"off")=12
	Set gtmtypes("multi_proc_shm_hdr_t",8,"len")=4
	Set gtmtypes("multi_proc_shm_hdr_t",8,"type")="boolean_t"
	Set gtmtypfldindx("multi_proc_shm_hdr_t","wait_done")=8
	Set gtmtypes("multi_proc_shm_hdr_t",9,"name")="multi_proc_shm_hdr_t.shmid"
	Set gtmtypes("multi_proc_shm_hdr_t",9,"off")=16
	Set gtmtypes("multi_proc_shm_hdr_t",9,"len")=4
	Set gtmtypes("multi_proc_shm_hdr_t",9,"type")="int"
	Set gtmtypfldindx("multi_proc_shm_hdr_t","shmid")=9
	Set gtmtypes("multi_proc_shm_hdr_t",10,"name")="multi_proc_shm_hdr_t.procs_created"
	Set gtmtypes("multi_proc_shm_hdr_t",10,"off")=20
	Set gtmtypes("multi_proc_shm_hdr_t",10,"len")=4
	Set gtmtypes("multi_proc_shm_hdr_t",10,"type")="int"
	Set gtmtypfldindx("multi_proc_shm_hdr_t","procs_created")=10
	Set gtmtypes("multi_proc_shm_hdr_t",11,"name")="multi_proc_shm_hdr_t.parent_pid"
	Set gtmtypes("multi_proc_shm_hdr_t",11,"off")=24
	Set gtmtypes("multi_proc_shm_hdr_t",11,"len")=4
	Set gtmtypes("multi_proc_shm_hdr_t",11,"type")="int"
	Set gtmtypfldindx("multi_proc_shm_hdr_t","parent_pid")=11
	Set gtmtypes("multi_proc_shm_hdr_t",12,"name")="multi_proc_shm_hdr_t.next_task"
	Set gtmtypes("multi_proc_shm_hdr_t",12,"off")=28
	Set gtmtypes("multi_proc_shm_hdr_t",12,"len")=4
	Set gtmtypes("multi_proc_shm_hdr_t",12,"type")="int"
	Set gtmtypfldindx("multi_proc_shm_hdr_t","next_task")=12
	Set gtmtypes("multi_proc_shm_hdr_t",13,"name")="multi_proc_shm_hdr_t.pid"
	Set gtmtypes("multi_proc_shm_hdr_t",13,"off")=32
	Set gtmtypes("multi_proc_shm_hdr_t",13,"len")=4000
	Set gtmtypes("multi_proc_shm_hdr_t",13,"type")="pid_t"
	Set gtmtypfldindx("multi_proc_shm_hdr_t","pid")=13
	Set gtmtypes("multi_proc_shm_hdr_t",13,"dim")=1000
	Set gtmtypes("multi_proc_shm_hdr_t",14,"name")="multi_proc_shm_hdr_t.orig_pid"
	Set gtmtypes("multi_proc_shm_hdr_t",14,"off")=4032
	Set gtmtypes("multi_proc_shm_hdr_t",14,"len")=4000
	Set gtmtypes("multi_proc_shm_hdr_t",14,"type")="pid_t"
	Set gtmtypfldindx("multi_proc_shm_hdr_t","orig_pid")=14
	Set gtmtypes("multi_proc_shm_hdr_t",14,"dim")=1000
	Set gtmtypes("multi_proc_shm_hdr_t",15,"name")="multi_proc_shm_hdr_t.wait_stat"
	Set gtmtypes("multi_proc_shm_hdr_t",15,"off")=8032
	Set gtmtypes("multi_proc_shm_hdr_t",15,"len")=4000
	Set gtmtypes("multi_proc_shm_hdr_t",15,"type")="int"
	Set gtmtypfldindx("multi_proc_shm_hdr_t","wait_stat")=15
	Set gtmtypes("multi_proc_shm_hdr_t",15,"dim")=1000
	Set gtmtypes("multi_proc_shm_hdr_t",16,"name")="multi_proc_shm_hdr_t.fnptr"
	Set gtmtypes("multi_proc_shm_hdr_t",16,"off")=12032
	Set gtmtypes("multi_proc_shm_hdr_t",16,"len")=8
	Set gtmtypes("multi_proc_shm_hdr_t",16,"type")="gtm_multi_proc_fnptr_t"
	Set gtmtypfldindx("multi_proc_shm_hdr_t","fnptr")=16
	Set gtmtypes("multi_proc_shm_hdr_t",17,"name")="multi_proc_shm_hdr_t.pvt_ret_array"
	Set gtmtypes("multi_proc_shm_hdr_t",17,"off")=12040
	Set gtmtypes("multi_proc_shm_hdr_t",17,"len")=8
	Set gtmtypes("multi_proc_shm_hdr_t",17,"type")="addr"
	Set gtmtypfldindx("multi_proc_shm_hdr_t","pvt_ret_array")=17
	Set gtmtypes("multi_proc_shm_hdr_t",18,"name")="multi_proc_shm_hdr_t.shm_ret_array"
	Set gtmtypes("multi_proc_shm_hdr_t",18,"off")=12048
	Set gtmtypes("multi_proc_shm_hdr_t",18,"len")=8
	Set gtmtypes("multi_proc_shm_hdr_t",18,"type")="addr"
	Set gtmtypfldindx("multi_proc_shm_hdr_t","shm_ret_array")=18
	Set gtmtypes("multi_proc_shm_hdr_t",19,"name")="multi_proc_shm_hdr_t.parm_array"
	Set gtmtypes("multi_proc_shm_hdr_t",19,"off")=12056
	Set gtmtypes("multi_proc_shm_hdr_t",19,"len")=8
	Set gtmtypes("multi_proc_shm_hdr_t",19,"type")="addr"
	Set gtmtypfldindx("multi_proc_shm_hdr_t","parm_array")=19
	Set gtmtypes("multi_proc_shm_hdr_t",20,"name")="multi_proc_shm_hdr_t.ntasks"
	Set gtmtypes("multi_proc_shm_hdr_t",20,"off")=12064
	Set gtmtypes("multi_proc_shm_hdr_t",20,"len")=4
	Set gtmtypes("multi_proc_shm_hdr_t",20,"type")="int"
	Set gtmtypfldindx("multi_proc_shm_hdr_t","ntasks")=20
	Set gtmtypes("multi_proc_shm_hdr_t",21,"name")="multi_proc_shm_hdr_t.max_procs"
	Set gtmtypes("multi_proc_shm_hdr_t",21,"off")=12068
	Set gtmtypes("multi_proc_shm_hdr_t",21,"len")=4
	Set gtmtypes("multi_proc_shm_hdr_t",21,"type")="int"
	Set gtmtypfldindx("multi_proc_shm_hdr_t","max_procs")=21
	Set gtmtypes("multi_proc_shm_hdr_t",22,"name")="multi_proc_shm_hdr_t.parmElemSize"
	Set gtmtypes("multi_proc_shm_hdr_t",22,"off")=12072
	Set gtmtypes("multi_proc_shm_hdr_t",22,"len")=4
	Set gtmtypes("multi_proc_shm_hdr_t",22,"type")="int"
	Set gtmtypfldindx("multi_proc_shm_hdr_t","parmElemSize")=22
	;
	Set gtmtypes("multi_struct")="struct"
	Set gtmtypes("multi_struct",0)=6
	Set gtmtypes("multi_struct","len")=32
	Set gtmtypes("multi_struct",1,"name")="multi_struct.token"
	Set gtmtypes("multi_struct",1,"off")=0
	Set gtmtypes("multi_struct",1,"len")=8
	Set gtmtypes("multi_struct",1,"type")="uint64_t"
	Set gtmtypfldindx("multi_struct","token")=1
	Set gtmtypes("multi_struct",2,"name")="multi_struct.time"
	Set gtmtypes("multi_struct",2,"off")=8
	Set gtmtypes("multi_struct",2,"len")=4
	Set gtmtypes("multi_struct",2,"type")="unsigned-int"
	Set gtmtypfldindx("multi_struct","time")=2
	Set gtmtypes("multi_struct",3,"name")="multi_struct.partner"
	Set gtmtypes("multi_struct",3,"off")=12
	Set gtmtypes("multi_struct",3,"len")=4
	Set gtmtypes("multi_struct",3,"type")="unsigned-int"
	Set gtmtypfldindx("multi_struct","partner")=3
	Set gtmtypes("multi_struct",4,"name")="multi_struct.tot_partner"
	Set gtmtypes("multi_struct",4,"off")=16
	Set gtmtypes("multi_struct",4,"len")=4
	Set gtmtypes("multi_struct",4,"type")="unsigned-int"
	Set gtmtypfldindx("multi_struct","tot_partner")=4
	Set gtmtypes("multi_struct",5,"name")="multi_struct.fence"
	Set gtmtypes("multi_struct",5,"off")=20
	Set gtmtypes("multi_struct",5,"len")=4
	Set gtmtypes("multi_struct",5,"type")="int"
	Set gtmtypfldindx("multi_struct","fence")=5
	Set gtmtypes("multi_struct",6,"name")="multi_struct.next"
	Set gtmtypes("multi_struct",6,"off")=24
	Set gtmtypes("multi_struct",6,"len")=8
	Set gtmtypes("multi_struct",6,"type")="addr"
	Set gtmtypfldindx("multi_struct","next")=6
	;
	Set gtmtypes("mur_back_opt_t")="struct"
	Set gtmtypes("mur_back_opt_t",0)=4
	Set gtmtypes("mur_back_opt_t","len")=24
	Set gtmtypes("mur_back_opt_t",1,"name")="mur_back_opt_t.jctl"
	Set gtmtypes("mur_back_opt_t",1,"off")=0
	Set gtmtypes("mur_back_opt_t",1,"len")=8
	Set gtmtypes("mur_back_opt_t",1,"type")="addr"
	Set gtmtypfldindx("mur_back_opt_t","jctl")=1
	Set gtmtypes("mur_back_opt_t",2,"name")="mur_back_opt_t.rec_token_seq"
	Set gtmtypes("mur_back_opt_t",2,"off")=8
	Set gtmtypes("mur_back_opt_t",2,"len")=8
	Set gtmtypes("mur_back_opt_t",2,"type")="uint64_t"
	Set gtmtypfldindx("mur_back_opt_t","rec_token_seq")=2
	Set gtmtypes("mur_back_opt_t",3,"name")="mur_back_opt_t.first_epoch"
	Set gtmtypes("mur_back_opt_t",3,"off")=16
	Set gtmtypes("mur_back_opt_t",3,"len")=4
	Set gtmtypes("mur_back_opt_t",3,"type")="boolean_t"
	Set gtmtypfldindx("mur_back_opt_t","first_epoch")=3
	Set gtmtypes("mur_back_opt_t",4,"name")="mur_back_opt_t.status"
	Set gtmtypes("mur_back_opt_t",4,"off")=20
	Set gtmtypes("mur_back_opt_t",4,"len")=4
	Set gtmtypes("mur_back_opt_t",4,"type")="unsigned-int"
	Set gtmtypfldindx("mur_back_opt_t","status")=4
	;
	Set gtmtypes("mur_buff_desc_t")="struct"
	Set gtmtypes("mur_buff_desc_t",0)=7
	Set gtmtypes("mur_buff_desc_t","len")=48
	Set gtmtypes("mur_buff_desc_t",1,"name")="mur_buff_desc_t.base"
	Set gtmtypes("mur_buff_desc_t",1,"off")=0
	Set gtmtypes("mur_buff_desc_t",1,"len")=8
	Set gtmtypes("mur_buff_desc_t",1,"type")="addr"
	Set gtmtypfldindx("mur_buff_desc_t","base")=1
	Set gtmtypes("mur_buff_desc_t",2,"name")="mur_buff_desc_t.top"
	Set gtmtypes("mur_buff_desc_t",2,"off")=8
	Set gtmtypes("mur_buff_desc_t",2,"len")=8
	Set gtmtypes("mur_buff_desc_t",2,"type")="addr"
	Set gtmtypfldindx("mur_buff_desc_t","top")=2
	Set gtmtypes("mur_buff_desc_t",3,"name")="mur_buff_desc_t.blen"
	Set gtmtypes("mur_buff_desc_t",3,"off")=16
	Set gtmtypes("mur_buff_desc_t",3,"len")=4
	Set gtmtypes("mur_buff_desc_t",3,"type")="unsigned-int"
	Set gtmtypfldindx("mur_buff_desc_t","blen")=3
	Set gtmtypes("mur_buff_desc_t",4,"name")="mur_buff_desc_t.dskaddr"
	Set gtmtypes("mur_buff_desc_t",4,"off")=20
	Set gtmtypes("mur_buff_desc_t",4,"len")=4
	Set gtmtypes("mur_buff_desc_t",4,"type")="unsigned-int"
	Set gtmtypfldindx("mur_buff_desc_t","dskaddr")=4
	Set gtmtypes("mur_buff_desc_t",5,"name")="mur_buff_desc_t.read_in_progress"
	Set gtmtypes("mur_buff_desc_t",5,"off")=24
	Set gtmtypes("mur_buff_desc_t",5,"len")=4
	Set gtmtypes("mur_buff_desc_t",5,"type")="boolean_t"
	Set gtmtypfldindx("mur_buff_desc_t","read_in_progress")=5
	Set gtmtypes("mur_buff_desc_t",6,"name")="mur_buff_desc_t.aiocbp"
	Set gtmtypes("mur_buff_desc_t",6,"off")=32
	Set gtmtypes("mur_buff_desc_t",6,"len")=8
	Set gtmtypes("mur_buff_desc_t",6,"type")="addr"
	Set gtmtypfldindx("mur_buff_desc_t","aiocbp")=6
	Set gtmtypes("mur_buff_desc_t",7,"name")="mur_buff_desc_t.rip_channel"
	Set gtmtypes("mur_buff_desc_t",7,"off")=40
	Set gtmtypes("mur_buff_desc_t",7,"len")=4
	Set gtmtypes("mur_buff_desc_t",7,"type")="int"
	Set gtmtypfldindx("mur_buff_desc_t","rip_channel")=7
	;
	Set gtmtypes("mur_gbls_t")="struct"
	Set gtmtypes("mur_gbls_t",0)=75
	Set gtmtypes("mur_gbls_t","len")=528
	Set gtmtypes("mur_gbls_t",1,"name")="mur_gbls_t.repl_standalone"
	Set gtmtypes("mur_gbls_t",1,"off")=0
	Set gtmtypes("mur_gbls_t",1,"len")=4
	Set gtmtypes("mur_gbls_t",1,"type")="boolean_t"
	Set gtmtypfldindx("mur_gbls_t","repl_standalone")=1
	Set gtmtypes("mur_gbls_t",2,"name")="mur_gbls_t.clean_exit"
	Set gtmtypes("mur_gbls_t",2,"off")=4
	Set gtmtypes("mur_gbls_t",2,"len")=4
	Set gtmtypes("mur_gbls_t",2,"type")="boolean_t"
	Set gtmtypfldindx("mur_gbls_t","clean_exit")=2
	Set gtmtypes("mur_gbls_t",3,"name")="mur_gbls_t.ok_to_update_db"
	Set gtmtypes("mur_gbls_t",3,"off")=8
	Set gtmtypes("mur_gbls_t",3,"len")=4
	Set gtmtypes("mur_gbls_t",3,"type")="boolean_t"
	Set gtmtypfldindx("mur_gbls_t","ok_to_update_db")=3
	Set gtmtypes("mur_gbls_t",4,"name")="mur_gbls_t.intrpt_recovery"
	Set gtmtypes("mur_gbls_t",4,"off")=12
	Set gtmtypes("mur_gbls_t",4,"len")=4
	Set gtmtypes("mur_gbls_t",4,"type")="boolean_t"
	Set gtmtypfldindx("mur_gbls_t","intrpt_recovery")=4
	Set gtmtypes("mur_gbls_t",5,"name")="mur_gbls_t.reg_total"
	Set gtmtypes("mur_gbls_t",5,"off")=16
	Set gtmtypes("mur_gbls_t",5,"len")=4
	Set gtmtypes("mur_gbls_t",5,"type")="int"
	Set gtmtypfldindx("mur_gbls_t","reg_total")=5
	Set gtmtypes("mur_gbls_t",6,"name")="mur_gbls_t.reg_full_total"
	Set gtmtypes("mur_gbls_t",6,"off")=20
	Set gtmtypes("mur_gbls_t",6,"len")=4
	Set gtmtypes("mur_gbls_t",6,"type")="int"
	Set gtmtypfldindx("mur_gbls_t","reg_full_total")=6
	Set gtmtypes("mur_gbls_t",7,"name")="mur_gbls_t.regcnt_remaining"
	Set gtmtypes("mur_gbls_t",7,"off")=24
	Set gtmtypes("mur_gbls_t",7,"len")=4
	Set gtmtypes("mur_gbls_t",7,"type")="int"
	Set gtmtypfldindx("mur_gbls_t","regcnt_remaining")=7
	Set gtmtypes("mur_gbls_t",8,"name")="mur_gbls_t.err_cnt"
	Set gtmtypes("mur_gbls_t",8,"off")=28
	Set gtmtypes("mur_gbls_t",8,"len")=4
	Set gtmtypes("mur_gbls_t",8,"type")="int"
	Set gtmtypfldindx("mur_gbls_t","err_cnt")=8
	Set gtmtypes("mur_gbls_t",9,"name")="mur_gbls_t.wrn_count"
	Set gtmtypes("mur_gbls_t",9,"off")=32
	Set gtmtypes("mur_gbls_t",9,"len")=4
	Set gtmtypes("mur_gbls_t",9,"type")="int"
	Set gtmtypfldindx("mur_gbls_t","wrn_count")=9
	Set gtmtypes("mur_gbls_t",10,"name")="mur_gbls_t.broken_cnt"
	Set gtmtypes("mur_gbls_t",10,"off")=36
	Set gtmtypes("mur_gbls_t",10,"len")=4
	Set gtmtypes("mur_gbls_t",10,"type")="int"
	Set gtmtypfldindx("mur_gbls_t","broken_cnt")=10
	Set gtmtypes("mur_gbls_t",11,"name")="mur_gbls_t.max_extr_record_length"
	Set gtmtypes("mur_gbls_t",11,"off")=40
	Set gtmtypes("mur_gbls_t",11,"len")=4
	Set gtmtypes("mur_gbls_t",11,"type")="int"
	Set gtmtypfldindx("mur_gbls_t","max_extr_record_length")=11
	Set gtmtypes("mur_gbls_t",12,"name")="mur_gbls_t.resync_seqno"
	Set gtmtypes("mur_gbls_t",12,"off")=48
	Set gtmtypes("mur_gbls_t",12,"len")=8
	Set gtmtypes("mur_gbls_t",12,"type")="uint64_t"
	Set gtmtypfldindx("mur_gbls_t","resync_seqno")=12
	Set gtmtypes("mur_gbls_t",13,"name")="mur_gbls_t.consist_jnl_seqno"
	Set gtmtypes("mur_gbls_t",13,"off")=56
	Set gtmtypes("mur_gbls_t",13,"len")=8
	Set gtmtypes("mur_gbls_t",13,"type")="uint64_t"
	Set gtmtypfldindx("mur_gbls_t","consist_jnl_seqno")=13
	Set gtmtypes("mur_gbls_t",14,"name")="mur_gbls_t.losttn_seqno"
	Set gtmtypes("mur_gbls_t",14,"off")=64
	Set gtmtypes("mur_gbls_t",14,"len")=8
	Set gtmtypes("mur_gbls_t",14,"type")="uint64_t"
	Set gtmtypfldindx("mur_gbls_t","losttn_seqno")=14
	Set gtmtypes("mur_gbls_t",15,"name")="mur_gbls_t.min_broken_seqno"
	Set gtmtypes("mur_gbls_t",15,"off")=72
	Set gtmtypes("mur_gbls_t",15,"len")=8
	Set gtmtypes("mur_gbls_t",15,"type")="uint64_t"
	Set gtmtypfldindx("mur_gbls_t","min_broken_seqno")=15
	Set gtmtypes("mur_gbls_t",16,"name")="mur_gbls_t.min_broken_time"
	Set gtmtypes("mur_gbls_t",16,"off")=80
	Set gtmtypes("mur_gbls_t",16,"len")=4
	Set gtmtypes("mur_gbls_t",16,"type")="unsigned-int"
	Set gtmtypfldindx("mur_gbls_t","min_broken_time")=16
	Set gtmtypes("mur_gbls_t",17,"name")="mur_gbls_t.token_table"
	Set gtmtypes("mur_gbls_t",17,"off")=88
	Set gtmtypes("mur_gbls_t",17,"len")=80
	Set gtmtypes("mur_gbls_t",17,"type")="hash_table_int8"
	Set gtmtypfldindx("mur_gbls_t","token_table")=17
	Set gtmtypes("mur_gbls_t",18,"name")="mur_gbls_t.token_table.base"
	Set gtmtypes("mur_gbls_t",18,"off")=88
	Set gtmtypes("mur_gbls_t",18,"len")=8
	Set gtmtypes("mur_gbls_t",18,"type")="addr"
	Set gtmtypfldindx("mur_gbls_t","token_table.base")=18
	Set gtmtypes("mur_gbls_t",19,"name")="mur_gbls_t.token_table.top"
	Set gtmtypes("mur_gbls_t",19,"off")=96
	Set gtmtypes("mur_gbls_t",19,"len")=8
	Set gtmtypes("mur_gbls_t",19,"type")="addr"
	Set gtmtypfldindx("mur_gbls_t","token_table.top")=19
	Set gtmtypes("mur_gbls_t",20,"name")="mur_gbls_t.token_table.spare_base"
	Set gtmtypes("mur_gbls_t",20,"off")=104
	Set gtmtypes("mur_gbls_t",20,"len")=8
	Set gtmtypes("mur_gbls_t",20,"type")="addr"
	Set gtmtypfldindx("mur_gbls_t","token_table.spare_base")=20
	Set gtmtypes("mur_gbls_t",21,"name")="mur_gbls_t.token_table.entry_passed_thru"
	Set gtmtypes("mur_gbls_t",21,"off")=112
	Set gtmtypes("mur_gbls_t",21,"len")=8
	Set gtmtypes("mur_gbls_t",21,"type")="addr"
	Set gtmtypfldindx("mur_gbls_t","token_table.entry_passed_thru")=21
	Set gtmtypes("mur_gbls_t",22,"name")="mur_gbls_t.token_table.size"
	Set gtmtypes("mur_gbls_t",22,"off")=120
	Set gtmtypes("mur_gbls_t",22,"len")=4
	Set gtmtypes("mur_gbls_t",22,"type")="unsigned-int"
	Set gtmtypfldindx("mur_gbls_t","token_table.size")=22
	Set gtmtypes("mur_gbls_t",23,"name")="mur_gbls_t.token_table.initial_size"
	Set gtmtypes("mur_gbls_t",23,"off")=124
	Set gtmtypes("mur_gbls_t",23,"len")=4
	Set gtmtypes("mur_gbls_t",23,"type")="unsigned-int"
	Set gtmtypfldindx("mur_gbls_t","token_table.initial_size")=23
	Set gtmtypes("mur_gbls_t",24,"name")="mur_gbls_t.token_table.spare_base_size"
	Set gtmtypes("mur_gbls_t",24,"off")=128
	Set gtmtypes("mur_gbls_t",24,"len")=4
	Set gtmtypes("mur_gbls_t",24,"type")="unsigned-int"
	Set gtmtypfldindx("mur_gbls_t","token_table.spare_base_size")=24
	Set gtmtypes("mur_gbls_t",25,"name")="mur_gbls_t.token_table.count"
	Set gtmtypes("mur_gbls_t",25,"off")=132
	Set gtmtypes("mur_gbls_t",25,"len")=4
	Set gtmtypes("mur_gbls_t",25,"type")="unsigned-int"
	Set gtmtypfldindx("mur_gbls_t","token_table.count")=25
	Set gtmtypes("mur_gbls_t",26,"name")="mur_gbls_t.token_table.del_count"
	Set gtmtypes("mur_gbls_t",26,"off")=136
	Set gtmtypes("mur_gbls_t",26,"len")=4
	Set gtmtypes("mur_gbls_t",26,"type")="unsigned-int"
	Set gtmtypfldindx("mur_gbls_t","token_table.del_count")=26
	Set gtmtypes("mur_gbls_t",27,"name")="mur_gbls_t.token_table.exp_trigger_size"
	Set gtmtypes("mur_gbls_t",27,"off")=140
	Set gtmtypes("mur_gbls_t",27,"len")=4
	Set gtmtypes("mur_gbls_t",27,"type")="unsigned-int"
	Set gtmtypfldindx("mur_gbls_t","token_table.exp_trigger_size")=27
	Set gtmtypes("mur_gbls_t",28,"name")="mur_gbls_t.token_table.cmp_trigger_size"
	Set gtmtypes("mur_gbls_t",28,"off")=144
	Set gtmtypes("mur_gbls_t",28,"len")=4
	Set gtmtypes("mur_gbls_t",28,"type")="unsigned-int"
	Set gtmtypfldindx("mur_gbls_t","token_table.cmp_trigger_size")=28
	Set gtmtypes("mur_gbls_t",29,"name")="mur_gbls_t.token_table.dont_compact"
	Set gtmtypes("mur_gbls_t",29,"off")=148
	Set gtmtypes("mur_gbls_t",29,"len")=4
	Set gtmtypes("mur_gbls_t",29,"type")="boolean_t"
	Set gtmtypfldindx("mur_gbls_t","token_table.dont_compact")=29
	Set gtmtypes("mur_gbls_t",30,"name")="mur_gbls_t.token_table.dont_keep_spare_table"
	Set gtmtypes("mur_gbls_t",30,"off")=152
	Set gtmtypes("mur_gbls_t",30,"len")=4
	Set gtmtypes("mur_gbls_t",30,"type")="boolean_t"
	Set gtmtypfldindx("mur_gbls_t","token_table.dont_keep_spare_table")=30
	Set gtmtypes("mur_gbls_t",31,"name")="mur_gbls_t.token_table.defer_base_release"
	Set gtmtypes("mur_gbls_t",31,"off")=156
	Set gtmtypes("mur_gbls_t",31,"len")=4
	Set gtmtypes("mur_gbls_t",31,"type")="boolean_t"
	Set gtmtypfldindx("mur_gbls_t","token_table.defer_base_release")=31
	Set gtmtypes("mur_gbls_t",32,"name")="mur_gbls_t.token_table.active"
	Set gtmtypes("mur_gbls_t",32,"off")=160
	Set gtmtypes("mur_gbls_t",32,"len")=4
	Set gtmtypes("mur_gbls_t",32,"type")="boolean_t"
	Set gtmtypfldindx("mur_gbls_t","token_table.active")=32
	Set gtmtypes("mur_gbls_t",33,"name")="mur_gbls_t.forw_token_table"
	Set gtmtypes("mur_gbls_t",33,"off")=168
	Set gtmtypes("mur_gbls_t",33,"len")=80
	Set gtmtypes("mur_gbls_t",33,"type")="hash_table_int8"
	Set gtmtypfldindx("mur_gbls_t","forw_token_table")=33
	Set gtmtypes("mur_gbls_t",34,"name")="mur_gbls_t.forw_token_table.base"
	Set gtmtypes("mur_gbls_t",34,"off")=168
	Set gtmtypes("mur_gbls_t",34,"len")=8
	Set gtmtypes("mur_gbls_t",34,"type")="addr"
	Set gtmtypfldindx("mur_gbls_t","forw_token_table.base")=34
	Set gtmtypes("mur_gbls_t",35,"name")="mur_gbls_t.forw_token_table.top"
	Set gtmtypes("mur_gbls_t",35,"off")=176
	Set gtmtypes("mur_gbls_t",35,"len")=8
	Set gtmtypes("mur_gbls_t",35,"type")="addr"
	Set gtmtypfldindx("mur_gbls_t","forw_token_table.top")=35
	Set gtmtypes("mur_gbls_t",36,"name")="mur_gbls_t.forw_token_table.spare_base"
	Set gtmtypes("mur_gbls_t",36,"off")=184
	Set gtmtypes("mur_gbls_t",36,"len")=8
	Set gtmtypes("mur_gbls_t",36,"type")="addr"
	Set gtmtypfldindx("mur_gbls_t","forw_token_table.spare_base")=36
	Set gtmtypes("mur_gbls_t",37,"name")="mur_gbls_t.forw_token_table.entry_passed_thru"
	Set gtmtypes("mur_gbls_t",37,"off")=192
	Set gtmtypes("mur_gbls_t",37,"len")=8
	Set gtmtypes("mur_gbls_t",37,"type")="addr"
	Set gtmtypfldindx("mur_gbls_t","forw_token_table.entry_passed_thru")=37
	Set gtmtypes("mur_gbls_t",38,"name")="mur_gbls_t.forw_token_table.size"
	Set gtmtypes("mur_gbls_t",38,"off")=200
	Set gtmtypes("mur_gbls_t",38,"len")=4
	Set gtmtypes("mur_gbls_t",38,"type")="unsigned-int"
	Set gtmtypfldindx("mur_gbls_t","forw_token_table.size")=38
	Set gtmtypes("mur_gbls_t",39,"name")="mur_gbls_t.forw_token_table.initial_size"
	Set gtmtypes("mur_gbls_t",39,"off")=204
	Set gtmtypes("mur_gbls_t",39,"len")=4
	Set gtmtypes("mur_gbls_t",39,"type")="unsigned-int"
	Set gtmtypfldindx("mur_gbls_t","forw_token_table.initial_size")=39
	Set gtmtypes("mur_gbls_t",40,"name")="mur_gbls_t.forw_token_table.spare_base_size"
	Set gtmtypes("mur_gbls_t",40,"off")=208
	Set gtmtypes("mur_gbls_t",40,"len")=4
	Set gtmtypes("mur_gbls_t",40,"type")="unsigned-int"
	Set gtmtypfldindx("mur_gbls_t","forw_token_table.spare_base_size")=40
	Set gtmtypes("mur_gbls_t",41,"name")="mur_gbls_t.forw_token_table.count"
	Set gtmtypes("mur_gbls_t",41,"off")=212
	Set gtmtypes("mur_gbls_t",41,"len")=4
	Set gtmtypes("mur_gbls_t",41,"type")="unsigned-int"
	Set gtmtypfldindx("mur_gbls_t","forw_token_table.count")=41
	Set gtmtypes("mur_gbls_t",42,"name")="mur_gbls_t.forw_token_table.del_count"
	Set gtmtypes("mur_gbls_t",42,"off")=216
	Set gtmtypes("mur_gbls_t",42,"len")=4
	Set gtmtypes("mur_gbls_t",42,"type")="unsigned-int"
	Set gtmtypfldindx("mur_gbls_t","forw_token_table.del_count")=42
	Set gtmtypes("mur_gbls_t",43,"name")="mur_gbls_t.forw_token_table.exp_trigger_size"
	Set gtmtypes("mur_gbls_t",43,"off")=220
	Set gtmtypes("mur_gbls_t",43,"len")=4
	Set gtmtypes("mur_gbls_t",43,"type")="unsigned-int"
	Set gtmtypfldindx("mur_gbls_t","forw_token_table.exp_trigger_size")=43
	Set gtmtypes("mur_gbls_t",44,"name")="mur_gbls_t.forw_token_table.cmp_trigger_size"
	Set gtmtypes("mur_gbls_t",44,"off")=224
	Set gtmtypes("mur_gbls_t",44,"len")=4
	Set gtmtypes("mur_gbls_t",44,"type")="unsigned-int"
	Set gtmtypfldindx("mur_gbls_t","forw_token_table.cmp_trigger_size")=44
	Set gtmtypes("mur_gbls_t",45,"name")="mur_gbls_t.forw_token_table.dont_compact"
	Set gtmtypes("mur_gbls_t",45,"off")=228
	Set gtmtypes("mur_gbls_t",45,"len")=4
	Set gtmtypes("mur_gbls_t",45,"type")="boolean_t"
	Set gtmtypfldindx("mur_gbls_t","forw_token_table.dont_compact")=45
	Set gtmtypes("mur_gbls_t",46,"name")="mur_gbls_t.forw_token_table.dont_keep_spare_table"
	Set gtmtypes("mur_gbls_t",46,"off")=232
	Set gtmtypes("mur_gbls_t",46,"len")=4
	Set gtmtypes("mur_gbls_t",46,"type")="boolean_t"
	Set gtmtypfldindx("mur_gbls_t","forw_token_table.dont_keep_spare_table")=46
	Set gtmtypes("mur_gbls_t",47,"name")="mur_gbls_t.forw_token_table.defer_base_release"
	Set gtmtypes("mur_gbls_t",47,"off")=236
	Set gtmtypes("mur_gbls_t",47,"len")=4
	Set gtmtypes("mur_gbls_t",47,"type")="boolean_t"
	Set gtmtypfldindx("mur_gbls_t","forw_token_table.defer_base_release")=47
	Set gtmtypes("mur_gbls_t",48,"name")="mur_gbls_t.forw_token_table.active"
	Set gtmtypes("mur_gbls_t",48,"off")=240
	Set gtmtypes("mur_gbls_t",48,"len")=4
	Set gtmtypes("mur_gbls_t",48,"type")="boolean_t"
	Set gtmtypfldindx("mur_gbls_t","forw_token_table.active")=48
	Set gtmtypes("mur_gbls_t",49,"name")="mur_gbls_t.multi_list"
	Set gtmtypes("mur_gbls_t",49,"off")=248
	Set gtmtypes("mur_gbls_t",49,"len")=8
	Set gtmtypes("mur_gbls_t",49,"type")="addr"
	Set gtmtypfldindx("mur_gbls_t","multi_list")=49
	Set gtmtypes("mur_gbls_t",50,"name")="mur_gbls_t.forw_multi_list"
	Set gtmtypes("mur_gbls_t",50,"off")=256
	Set gtmtypes("mur_gbls_t",50,"len")=8
	Set gtmtypes("mur_gbls_t",50,"type")="addr"
	Set gtmtypfldindx("mur_gbls_t","forw_multi_list")=50
	Set gtmtypes("mur_gbls_t",51,"name")="mur_gbls_t.pini_buddy_list"
	Set gtmtypes("mur_gbls_t",51,"off")=264
	Set gtmtypes("mur_gbls_t",51,"len")=8
	Set gtmtypes("mur_gbls_t",51,"type")="addr"
	Set gtmtypfldindx("mur_gbls_t","pini_buddy_list")=51
	Set gtmtypes("mur_gbls_t",52,"name")="mur_gbls_t.extr_buff"
	Set gtmtypes("mur_gbls_t",52,"off")=272
	Set gtmtypes("mur_gbls_t",52,"len")=8
	Set gtmtypes("mur_gbls_t",52,"type")="addr"
	Set gtmtypfldindx("mur_gbls_t","extr_buff")=52
	Set gtmtypes("mur_gbls_t",53,"name")="mur_gbls_t.prc_vec"
	Set gtmtypes("mur_gbls_t",53,"off")=280
	Set gtmtypes("mur_gbls_t",53,"len")=8
	Set gtmtypes("mur_gbls_t",53,"type")="addr"
	Set gtmtypfldindx("mur_gbls_t","prc_vec")=53
	Set gtmtypes("mur_gbls_t",54,"name")="mur_gbls_t.remote_side"
	Set gtmtypes("mur_gbls_t",54,"off")=288
	Set gtmtypes("mur_gbls_t",54,"len")=48
	Set gtmtypes("mur_gbls_t",54,"type")="repl_conn_info_t"
	Set gtmtypfldindx("mur_gbls_t","remote_side")=54
	Set gtmtypes("mur_gbls_t",55,"name")="mur_gbls_t.remote_side.proto_ver"
	Set gtmtypes("mur_gbls_t",55,"off")=288
	Set gtmtypes("mur_gbls_t",55,"len")=4
	Set gtmtypes("mur_gbls_t",55,"type")="int"
	Set gtmtypfldindx("mur_gbls_t","remote_side.proto_ver")=55
	Set gtmtypes("mur_gbls_t",56,"name")="mur_gbls_t.remote_side.jnl_ver"
	Set gtmtypes("mur_gbls_t",56,"off")=292
	Set gtmtypes("mur_gbls_t",56,"len")=4
	Set gtmtypes("mur_gbls_t",56,"type")="unsigned-int"
	Set gtmtypfldindx("mur_gbls_t","remote_side.jnl_ver")=56
	Set gtmtypes("mur_gbls_t",57,"name")="mur_gbls_t.remote_side.is_std_null_coll"
	Set gtmtypes("mur_gbls_t",57,"off")=296
	Set gtmtypes("mur_gbls_t",57,"len")=4
	Set gtmtypes("mur_gbls_t",57,"type")="boolean_t"
	Set gtmtypfldindx("mur_gbls_t","remote_side.is_std_null_coll")=57
	Set gtmtypes("mur_gbls_t",58,"name")="mur_gbls_t.remote_side.trigger_supported"
	Set gtmtypes("mur_gbls_t",58,"off")=300
	Set gtmtypes("mur_gbls_t",58,"len")=4
	Set gtmtypes("mur_gbls_t",58,"type")="boolean_t"
	Set gtmtypfldindx("mur_gbls_t","remote_side.trigger_supported")=58
	Set gtmtypes("mur_gbls_t",59,"name")="mur_gbls_t.remote_side.cross_endian"
	Set gtmtypes("mur_gbls_t",59,"off")=304
	Set gtmtypes("mur_gbls_t",59,"len")=4
	Set gtmtypes("mur_gbls_t",59,"type")="boolean_t"
	Set gtmtypfldindx("mur_gbls_t","remote_side.cross_endian")=59
	Set gtmtypes("mur_gbls_t",60,"name")="mur_gbls_t.remote_side.endianness_known"
	Set gtmtypes("mur_gbls_t",60,"off")=308
	Set gtmtypes("mur_gbls_t",60,"len")=4
	Set gtmtypes("mur_gbls_t",60,"type")="boolean_t"
	Set gtmtypfldindx("mur_gbls_t","remote_side.endianness_known")=60
	Set gtmtypes("mur_gbls_t",61,"name")="mur_gbls_t.remote_side.null_subs_xform"
	Set gtmtypes("mur_gbls_t",61,"off")=312
	Set gtmtypes("mur_gbls_t",61,"len")=4
	Set gtmtypes("mur_gbls_t",61,"type")="boolean_t"
	Set gtmtypfldindx("mur_gbls_t","remote_side.null_subs_xform")=61
	Set gtmtypes("mur_gbls_t",62,"name")="mur_gbls_t.remote_side.is_supplementary"
	Set gtmtypes("mur_gbls_t",62,"off")=316
	Set gtmtypes("mur_gbls_t",62,"len")=4
	Set gtmtypes("mur_gbls_t",62,"type")="boolean_t"
	Set gtmtypfldindx("mur_gbls_t","remote_side.is_supplementary")=62
	Set gtmtypes("mur_gbls_t",63,"name")="mur_gbls_t.remote_side.tls_requested"
	Set gtmtypes("mur_gbls_t",63,"off")=320
	Set gtmtypes("mur_gbls_t",63,"len")=4
	Set gtmtypes("mur_gbls_t",63,"type")="boolean_t"
	Set gtmtypfldindx("mur_gbls_t","remote_side.tls_requested")=63
	Set gtmtypes("mur_gbls_t",64,"name")="mur_gbls_t.remote_side.filler_16"
	Set gtmtypes("mur_gbls_t",64,"off")=324
	Set gtmtypes("mur_gbls_t",64,"len")=12
	Set gtmtypes("mur_gbls_t",64,"type")="char"
	Set gtmtypfldindx("mur_gbls_t","remote_side.filler_16")=64
	Set gtmtypes("mur_gbls_t",65,"name")="mur_gbls_t.was_rootprimary"
	Set gtmtypes("mur_gbls_t",65,"off")=336
	Set gtmtypes("mur_gbls_t",65,"len")=4
	Set gtmtypes("mur_gbls_t",65,"type")="boolean_t"
	Set gtmtypfldindx("mur_gbls_t","was_rootprimary")=65
	Set gtmtypes("mur_gbls_t",66,"name")="mur_gbls_t.resync_strm_index"
	Set gtmtypes("mur_gbls_t",66,"off")=340
	Set gtmtypes("mur_gbls_t",66,"len")=4
	Set gtmtypes("mur_gbls_t",66,"type")="int"
	Set gtmtypfldindx("mur_gbls_t","resync_strm_index")=66
	Set gtmtypes("mur_gbls_t",67,"name")="mur_gbls_t.resync_strm_seqno"
	Set gtmtypes("mur_gbls_t",67,"off")=344
	Set gtmtypes("mur_gbls_t",67,"len")=128
	Set gtmtypes("mur_gbls_t",67,"type")="uint64_t"
	Set gtmtypfldindx("mur_gbls_t","resync_strm_seqno")=67
	Set gtmtypes("mur_gbls_t",67,"dim")=16
	Set gtmtypes("mur_gbls_t",68,"name")="mur_gbls_t.resync_strm_seqno_nonzero"
	Set gtmtypes("mur_gbls_t",68,"off")=472
	Set gtmtypes("mur_gbls_t",68,"len")=4
	Set gtmtypes("mur_gbls_t",68,"type")="boolean_t"
	Set gtmtypfldindx("mur_gbls_t","resync_strm_seqno_nonzero")=68
	Set gtmtypes("mur_gbls_t",69,"name")="mur_gbls_t.incr_onln_rlbk_cycle"
	Set gtmtypes("mur_gbls_t",69,"off")=476
	Set gtmtypes("mur_gbls_t",69,"len")=4
	Set gtmtypes("mur_gbls_t",69,"type")="boolean_t"
	Set gtmtypfldindx("mur_gbls_t","incr_onln_rlbk_cycle")=69
	Set gtmtypes("mur_gbls_t",70,"name")="mur_gbls_t.incr_db_rlbkd_cycle"
	Set gtmtypes("mur_gbls_t",70,"off")=480
	Set gtmtypes("mur_gbls_t",70,"len")=4
	Set gtmtypes("mur_gbls_t",70,"type")="boolean_t"
	Set gtmtypfldindx("mur_gbls_t","incr_db_rlbkd_cycle")=70
	Set gtmtypes("mur_gbls_t",71,"name")="mur_gbls_t.thr_array"
	Set gtmtypes("mur_gbls_t",71,"off")=488
	Set gtmtypes("mur_gbls_t",71,"len")=8
	Set gtmtypes("mur_gbls_t",71,"type")="addr"
	Set gtmtypfldindx("mur_gbls_t","thr_array")=71
	Set gtmtypes("mur_gbls_t",72,"name")="mur_gbls_t.ret_array"
	Set gtmtypes("mur_gbls_t",72,"off")=496
	Set gtmtypes("mur_gbls_t",72,"len")=8
	Set gtmtypes("mur_gbls_t",72,"type")="addr"
	Set gtmtypfldindx("mur_gbls_t","ret_array")=72
	Set gtmtypes("mur_gbls_t",73,"name")="mur_gbls_t.mur_state"
	Set gtmtypes("mur_gbls_t",73,"off")=504
	Set gtmtypes("mur_gbls_t",73,"len")=4
	Set gtmtypes("mur_gbls_t",73,"type")="int"
	Set gtmtypfldindx("mur_gbls_t","mur_state")=73
	Set gtmtypes("mur_gbls_t",74,"name")="mur_gbls_t.adjusted_resolve_time"
	Set gtmtypes("mur_gbls_t",74,"off")=508
	Set gtmtypes("mur_gbls_t",74,"len")=4
	Set gtmtypes("mur_gbls_t",74,"type")="unsigned-int"
	Set gtmtypfldindx("mur_gbls_t","adjusted_resolve_time")=74
	Set gtmtypes("mur_gbls_t",75,"name")="mur_gbls_t.filenotcreate_displayed"
	Set gtmtypes("mur_gbls_t",75,"off")=512
	Set gtmtypes("mur_gbls_t",75,"len")=12
	Set gtmtypes("mur_gbls_t",75,"type")="boolean_t"
	Set gtmtypfldindx("mur_gbls_t","filenotcreate_displayed")=75
	Set gtmtypes("mur_gbls_t",75,"dim")=3
	;
	Set gtmtypes("mur_opt_struct")="struct"
	Set gtmtypes("mur_opt_struct",0)=45
	Set gtmtypes("mur_opt_struct","len")=288
	Set gtmtypes("mur_opt_struct",1,"name")="mur_opt_struct.lookback_time"
	Set gtmtypes("mur_opt_struct",1,"off")=0
	Set gtmtypes("mur_opt_struct",1,"len")=8
	Set gtmtypes("mur_opt_struct",1,"type")="int64_t"
	Set gtmtypfldindx("mur_opt_struct","lookback_time")=1
	Set gtmtypes("mur_opt_struct",2,"name")="mur_opt_struct.before_time"
	Set gtmtypes("mur_opt_struct",2,"off")=8
	Set gtmtypes("mur_opt_struct",2,"len")=8
	Set gtmtypes("mur_opt_struct",2,"type")="int64_t"
	Set gtmtypfldindx("mur_opt_struct","before_time")=2
	Set gtmtypes("mur_opt_struct",3,"name")="mur_opt_struct.since_time"
	Set gtmtypes("mur_opt_struct",3,"off")=16
	Set gtmtypes("mur_opt_struct",3,"len")=8
	Set gtmtypes("mur_opt_struct",3,"type")="int64_t"
	Set gtmtypfldindx("mur_opt_struct","since_time")=3
	Set gtmtypes("mur_opt_struct",4,"name")="mur_opt_struct.after_time"
	Set gtmtypes("mur_opt_struct",4,"off")=24
	Set gtmtypes("mur_opt_struct",4,"len")=8
	Set gtmtypes("mur_opt_struct",4,"type")="int64_t"
	Set gtmtypfldindx("mur_opt_struct","after_time")=4
	Set gtmtypes("mur_opt_struct",5,"name")="mur_opt_struct.fences"
	Set gtmtypes("mur_opt_struct",5,"off")=32
	Set gtmtypes("mur_opt_struct",5,"len")=4
	Set gtmtypes("mur_opt_struct",5,"type")="int"
	Set gtmtypfldindx("mur_opt_struct","fences")=5
	Set gtmtypes("mur_opt_struct",6,"name")="mur_opt_struct.error_limit"
	Set gtmtypes("mur_opt_struct",6,"off")=36
	Set gtmtypes("mur_opt_struct",6,"len")=4
	Set gtmtypes("mur_opt_struct",6,"type")="int"
	Set gtmtypfldindx("mur_opt_struct","error_limit")=6
	Set gtmtypes("mur_opt_struct",7,"name")="mur_opt_struct.fetchresync_port"
	Set gtmtypes("mur_opt_struct",7,"off")=40
	Set gtmtypes("mur_opt_struct",7,"len")=4
	Set gtmtypes("mur_opt_struct",7,"type")="int"
	Set gtmtypfldindx("mur_opt_struct","fetchresync_port")=7
	Set gtmtypes("mur_opt_struct",8,"name")="mur_opt_struct.show"
	Set gtmtypes("mur_opt_struct",8,"off")=44
	Set gtmtypes("mur_opt_struct",8,"len")=4
	Set gtmtypes("mur_opt_struct",8,"type")="int"
	Set gtmtypfldindx("mur_opt_struct","show")=8
	Set gtmtypes("mur_opt_struct",9,"name")="mur_opt_struct.lookback_opers"
	Set gtmtypes("mur_opt_struct",9,"off")=48
	Set gtmtypes("mur_opt_struct",9,"len")=4
	Set gtmtypes("mur_opt_struct",9,"type")="int"
	Set gtmtypfldindx("mur_opt_struct","lookback_opers")=9
	Set gtmtypes("mur_opt_struct",10,"name")="mur_opt_struct.forward"
	Set gtmtypes("mur_opt_struct",10,"off")=52
	Set gtmtypes("mur_opt_struct",10,"len")=4
	Set gtmtypes("mur_opt_struct",10,"type")="boolean_t"
	Set gtmtypfldindx("mur_opt_struct","forward")=10
	Set gtmtypes("mur_opt_struct",11,"name")="mur_opt_struct.update"
	Set gtmtypes("mur_opt_struct",11,"off")=56
	Set gtmtypes("mur_opt_struct",11,"len")=4
	Set gtmtypes("mur_opt_struct",11,"type")="boolean_t"
	Set gtmtypfldindx("mur_opt_struct","update")=11
	Set gtmtypes("mur_opt_struct",12,"name")="mur_opt_struct.corruptdb"
	Set gtmtypes("mur_opt_struct",12,"off")=60
	Set gtmtypes("mur_opt_struct",12,"len")=4
	Set gtmtypes("mur_opt_struct",12,"type")="boolean_t"
	Set gtmtypfldindx("mur_opt_struct","corruptdb")=12
	Set gtmtypes("mur_opt_struct",13,"name")="mur_opt_struct.rollback"
	Set gtmtypes("mur_opt_struct",13,"off")=64
	Set gtmtypes("mur_opt_struct",13,"len")=4
	Set gtmtypes("mur_opt_struct",13,"type")="boolean_t"
	Set gtmtypfldindx("mur_opt_struct","rollback")=13
	Set gtmtypes("mur_opt_struct",14,"name")="mur_opt_struct.rollback_losttnonly"
	Set gtmtypes("mur_opt_struct",14,"off")=68
	Set gtmtypes("mur_opt_struct",14,"len")=4
	Set gtmtypes("mur_opt_struct",14,"type")="boolean_t"
	Set gtmtypfldindx("mur_opt_struct","rollback_losttnonly")=14
	Set gtmtypes("mur_opt_struct",15,"name")="mur_opt_struct.verify"
	Set gtmtypes("mur_opt_struct",15,"off")=72
	Set gtmtypes("mur_opt_struct",15,"len")=4
	Set gtmtypes("mur_opt_struct",15,"type")="boolean_t"
	Set gtmtypfldindx("mur_opt_struct","verify")=15
	Set gtmtypes("mur_opt_struct",16,"name")="mur_opt_struct.verify_specified"
	Set gtmtypes("mur_opt_struct",16,"off")=76
	Set gtmtypes("mur_opt_struct",16,"len")=4
	Set gtmtypes("mur_opt_struct",16,"type")="boolean_t"
	Set gtmtypfldindx("mur_opt_struct","verify_specified")=16
	Set gtmtypes("mur_opt_struct",17,"name")="mur_opt_struct.before_time_specified"
	Set gtmtypes("mur_opt_struct",17,"off")=80
	Set gtmtypes("mur_opt_struct",17,"len")=4
	Set gtmtypes("mur_opt_struct",17,"type")="boolean_t"
	Set gtmtypfldindx("mur_opt_struct","before_time_specified")=17
	Set gtmtypes("mur_opt_struct",18,"name")="mur_opt_struct.since_time_specified"
	Set gtmtypes("mur_opt_struct",18,"off")=84
	Set gtmtypes("mur_opt_struct",18,"len")=4
	Set gtmtypes("mur_opt_struct",18,"type")="boolean_t"
	Set gtmtypfldindx("mur_opt_struct","since_time_specified")=18
	Set gtmtypes("mur_opt_struct",19,"name")="mur_opt_struct.resync_specified"
	Set gtmtypes("mur_opt_struct",19,"off")=88
	Set gtmtypes("mur_opt_struct",19,"len")=4
	Set gtmtypes("mur_opt_struct",19,"type")="boolean_t"
	Set gtmtypfldindx("mur_opt_struct","resync_specified")=19
	Set gtmtypes("mur_opt_struct",20,"name")="mur_opt_struct.lookback_time_specified"
	Set gtmtypes("mur_opt_struct",20,"off")=92
	Set gtmtypes("mur_opt_struct",20,"len")=4
	Set gtmtypes("mur_opt_struct",20,"type")="boolean_t"
	Set gtmtypfldindx("mur_opt_struct","lookback_time_specified")=20
	Set gtmtypes("mur_opt_struct",21,"name")="mur_opt_struct.lookback_opers_specified"
	Set gtmtypes("mur_opt_struct",21,"off")=96
	Set gtmtypes("mur_opt_struct",21,"len")=4
	Set gtmtypes("mur_opt_struct",21,"type")="boolean_t"
	Set gtmtypfldindx("mur_opt_struct","lookback_opers_specified")=21
	Set gtmtypes("mur_opt_struct",22,"name")="mur_opt_struct.interactive"
	Set gtmtypes("mur_opt_struct",22,"off")=100
	Set gtmtypes("mur_opt_struct",22,"len")=4
	Set gtmtypes("mur_opt_struct",22,"type")="boolean_t"
	Set gtmtypfldindx("mur_opt_struct","interactive")=22
	Set gtmtypes("mur_opt_struct",23,"name")="mur_opt_struct.selection"
	Set gtmtypes("mur_opt_struct",23,"off")=104
	Set gtmtypes("mur_opt_struct",23,"len")=4
	Set gtmtypes("mur_opt_struct",23,"type")="boolean_t"
	Set gtmtypfldindx("mur_opt_struct","selection")=23
	Set gtmtypes("mur_opt_struct",24,"name")="mur_opt_struct.apply_after_image"
	Set gtmtypes("mur_opt_struct",24,"off")=108
	Set gtmtypes("mur_opt_struct",24,"len")=4
	Set gtmtypes("mur_opt_struct",24,"type")="boolean_t"
	Set gtmtypfldindx("mur_opt_struct","apply_after_image")=24
	Set gtmtypes("mur_opt_struct",25,"name")="mur_opt_struct.chain"
	Set gtmtypes("mur_opt_struct",25,"off")=112
	Set gtmtypes("mur_opt_struct",25,"len")=4
	Set gtmtypes("mur_opt_struct",25,"type")="boolean_t"
	Set gtmtypfldindx("mur_opt_struct","chain")=25
	Set gtmtypes("mur_opt_struct",26,"name")="mur_opt_struct.notncheck"
	Set gtmtypes("mur_opt_struct",26,"off")=116
	Set gtmtypes("mur_opt_struct",26,"len")=4
	Set gtmtypes("mur_opt_struct",26,"type")="boolean_t"
	Set gtmtypfldindx("mur_opt_struct","notncheck")=26
	Set gtmtypes("mur_opt_struct",27,"name")="mur_opt_struct.verbose"
	Set gtmtypes("mur_opt_struct",27,"off")=120
	Set gtmtypes("mur_opt_struct",27,"len")=4
	Set gtmtypes("mur_opt_struct",27,"type")="boolean_t"
	Set gtmtypfldindx("mur_opt_struct","verbose")=27
	Set gtmtypes("mur_opt_struct",28,"name")="mur_opt_struct.log"
	Set gtmtypes("mur_opt_struct",28,"off")=124
	Set gtmtypes("mur_opt_struct",28,"len")=4
	Set gtmtypes("mur_opt_struct",28,"type")="boolean_t"
	Set gtmtypfldindx("mur_opt_struct","log")=28
	Set gtmtypes("mur_opt_struct",29,"name")="mur_opt_struct.detail"
	Set gtmtypes("mur_opt_struct",29,"off")=128
	Set gtmtypes("mur_opt_struct",29,"len")=4
	Set gtmtypes("mur_opt_struct",29,"type")="boolean_t"
	Set gtmtypfldindx("mur_opt_struct","detail")=29
	Set gtmtypes("mur_opt_struct",30,"name")="mur_opt_struct.extract_full"
	Set gtmtypes("mur_opt_struct",30,"off")=132
	Set gtmtypes("mur_opt_struct",30,"len")=4
	Set gtmtypes("mur_opt_struct",30,"type")="boolean_t"
	Set gtmtypfldindx("mur_opt_struct","extract_full")=30
	Set gtmtypes("mur_opt_struct",31,"name")="mur_opt_struct.show_head_only"
	Set gtmtypes("mur_opt_struct",31,"off")=136
	Set gtmtypes("mur_opt_struct",31,"len")=4
	Set gtmtypes("mur_opt_struct",31,"type")="boolean_t"
	Set gtmtypfldindx("mur_opt_struct","show_head_only")=31
	Set gtmtypes("mur_opt_struct",32,"name")="mur_opt_struct.extr"
	Set gtmtypes("mur_opt_struct",32,"off")=140
	Set gtmtypes("mur_opt_struct",32,"len")=12
	Set gtmtypes("mur_opt_struct",32,"type")="boolean_t"
	Set gtmtypfldindx("mur_opt_struct","extr")=32
	Set gtmtypes("mur_opt_struct",32,"dim")=3
	Set gtmtypes("mur_opt_struct",33,"name")="mur_opt_struct.transaction"
	Set gtmtypes("mur_opt_struct",33,"off")=152
	Set gtmtypes("mur_opt_struct",33,"len")=1
	Set gtmtypes("mur_opt_struct",33,"type")="char"
	Set gtmtypfldindx("mur_opt_struct","transaction")=33
	Set gtmtypes("mur_opt_struct",34,"name")="mur_opt_struct.redirect"
	Set gtmtypes("mur_opt_struct",34,"off")=160
	Set gtmtypes("mur_opt_struct",34,"len")=8
	Set gtmtypes("mur_opt_struct",34,"type")="addr"
	Set gtmtypfldindx("mur_opt_struct","redirect")=34
	Set gtmtypes("mur_opt_struct",35,"name")="mur_opt_struct.user"
	Set gtmtypes("mur_opt_struct",35,"off")=168
	Set gtmtypes("mur_opt_struct",35,"len")=8
	Set gtmtypes("mur_opt_struct",35,"type")="addr"
	Set gtmtypfldindx("mur_opt_struct","user")=35
	Set gtmtypes("mur_opt_struct",36,"name")="mur_opt_struct.database"
	Set gtmtypes("mur_opt_struct",36,"off")=176
	Set gtmtypes("mur_opt_struct",36,"len")=8
	Set gtmtypes("mur_opt_struct",36,"type")="addr"
	Set gtmtypfldindx("mur_opt_struct","database")=36
	Set gtmtypes("mur_opt_struct",37,"name")="mur_opt_struct.patterns"
	Set gtmtypes("mur_opt_struct",37,"off")=184
	Set gtmtypes("mur_opt_struct",37,"len")=8
	Set gtmtypes("mur_opt_struct",37,"type")="addr"
	Set gtmtypfldindx("mur_opt_struct","patterns")=37
	Set gtmtypes("mur_opt_struct",38,"name")="mur_opt_struct.global"
	Set gtmtypes("mur_opt_struct",38,"off")=192
	Set gtmtypes("mur_opt_struct",38,"len")=8
	Set gtmtypes("mur_opt_struct",38,"type")="addr"
	Set gtmtypfldindx("mur_opt_struct","global")=38
	Set gtmtypes("mur_opt_struct",39,"name")="mur_opt_struct.process"
	Set gtmtypes("mur_opt_struct",39,"off")=200
	Set gtmtypes("mur_opt_struct",39,"len")=8
	Set gtmtypes("mur_opt_struct",39,"type")="addr"
	Set gtmtypfldindx("mur_opt_struct","process")=39
	Set gtmtypes("mur_opt_struct",40,"name")="mur_opt_struct.id"
	Set gtmtypes("mur_opt_struct",40,"off")=208
	Set gtmtypes("mur_opt_struct",40,"len")=8
	Set gtmtypes("mur_opt_struct",40,"type")="addr"
	Set gtmtypfldindx("mur_opt_struct","id")=40
	Set gtmtypes("mur_opt_struct",41,"name")="mur_opt_struct.seqno"
	Set gtmtypes("mur_opt_struct",41,"off")=216
	Set gtmtypes("mur_opt_struct",41,"len")=8
	Set gtmtypes("mur_opt_struct",41,"type")="addr"
	Set gtmtypfldindx("mur_opt_struct","seqno")=41
	Set gtmtypes("mur_opt_struct",42,"name")="mur_opt_struct.extr_fn"
	Set gtmtypes("mur_opt_struct",42,"off")=224
	Set gtmtypes("mur_opt_struct",42,"len")=24
	Set gtmtypes("mur_opt_struct",42,"type")="addr"
	Set gtmtypfldindx("mur_opt_struct","extr_fn")=42
	Set gtmtypes("mur_opt_struct",42,"dim")=3
	Set gtmtypes("mur_opt_struct",43,"name")="mur_opt_struct.extr_fn_len"
	Set gtmtypes("mur_opt_struct",43,"off")=248
	Set gtmtypes("mur_opt_struct",43,"len")=12
	Set gtmtypes("mur_opt_struct",43,"type")="int"
	Set gtmtypfldindx("mur_opt_struct","extr_fn_len")=43
	Set gtmtypes("mur_opt_struct",43,"dim")=3
	Set gtmtypes("mur_opt_struct",44,"name")="mur_opt_struct.extr_fn_is_stdout"
	Set gtmtypes("mur_opt_struct",44,"off")=260
	Set gtmtypes("mur_opt_struct",44,"len")=12
	Set gtmtypes("mur_opt_struct",44,"type")="boolean_t"
	Set gtmtypfldindx("mur_opt_struct","extr_fn_is_stdout")=44
	Set gtmtypes("mur_opt_struct",44,"dim")=3
	Set gtmtypes("mur_opt_struct",45,"name")="mur_opt_struct.extr_fn_is_devnull"
	Set gtmtypes("mur_opt_struct",45,"off")=272
	Set gtmtypes("mur_opt_struct",45,"len")=12
	Set gtmtypes("mur_opt_struct",45,"type")="boolean_t"
	Set gtmtypfldindx("mur_opt_struct","extr_fn_is_devnull")=45
	Set gtmtypes("mur_opt_struct",45,"dim")=3
	;
	Set gtmtypes("mur_read_desc_t")="struct"
	Set gtmtypes("mur_read_desc_t",0)=33
	Set gtmtypes("mur_read_desc_t","len")=264
	Set gtmtypes("mur_read_desc_t",1,"name")="mur_read_desc_t.blocksize"
	Set gtmtypes("mur_read_desc_t",1,"off")=0
	Set gtmtypes("mur_read_desc_t",1,"len")=4
	Set gtmtypes("mur_read_desc_t",1,"type")="int"
	Set gtmtypfldindx("mur_read_desc_t","blocksize")=1
	Set gtmtypes("mur_read_desc_t",2,"name")="mur_read_desc_t.alloc_base"
	Set gtmtypes("mur_read_desc_t",2,"off")=8
	Set gtmtypes("mur_read_desc_t",2,"len")=8
	Set gtmtypes("mur_read_desc_t",2,"type")="addr"
	Set gtmtypfldindx("mur_read_desc_t","alloc_base")=2
	Set gtmtypes("mur_read_desc_t",3,"name")="mur_read_desc_t.alloc_len"
	Set gtmtypes("mur_read_desc_t",3,"off")=16
	Set gtmtypes("mur_read_desc_t",3,"len")=4
	Set gtmtypes("mur_read_desc_t",3,"type")="int"
	Set gtmtypfldindx("mur_read_desc_t","alloc_len")=3
	Set gtmtypes("mur_read_desc_t",4,"name")="mur_read_desc_t.random_buff"
	Set gtmtypes("mur_read_desc_t",4,"off")=24
	Set gtmtypes("mur_read_desc_t",4,"len")=48
	Set gtmtypes("mur_read_desc_t",4,"type")="mur_buff_desc_t"
	Set gtmtypfldindx("mur_read_desc_t","random_buff")=4
	Set gtmtypes("mur_read_desc_t",5,"name")="mur_read_desc_t.random_buff.base"
	Set gtmtypes("mur_read_desc_t",5,"off")=24
	Set gtmtypes("mur_read_desc_t",5,"len")=8
	Set gtmtypes("mur_read_desc_t",5,"type")="addr"
	Set gtmtypfldindx("mur_read_desc_t","random_buff.base")=5
	Set gtmtypes("mur_read_desc_t",6,"name")="mur_read_desc_t.random_buff.top"
	Set gtmtypes("mur_read_desc_t",6,"off")=32
	Set gtmtypes("mur_read_desc_t",6,"len")=8
	Set gtmtypes("mur_read_desc_t",6,"type")="addr"
	Set gtmtypfldindx("mur_read_desc_t","random_buff.top")=6
	Set gtmtypes("mur_read_desc_t",7,"name")="mur_read_desc_t.random_buff.blen"
	Set gtmtypes("mur_read_desc_t",7,"off")=40
	Set gtmtypes("mur_read_desc_t",7,"len")=4
	Set gtmtypes("mur_read_desc_t",7,"type")="unsigned-int"
	Set gtmtypfldindx("mur_read_desc_t","random_buff.blen")=7
	Set gtmtypes("mur_read_desc_t",8,"name")="mur_read_desc_t.random_buff.dskaddr"
	Set gtmtypes("mur_read_desc_t",8,"off")=44
	Set gtmtypes("mur_read_desc_t",8,"len")=4
	Set gtmtypes("mur_read_desc_t",8,"type")="unsigned-int"
	Set gtmtypfldindx("mur_read_desc_t","random_buff.dskaddr")=8
	Set gtmtypes("mur_read_desc_t",9,"name")="mur_read_desc_t.random_buff.read_in_progress"
	Set gtmtypes("mur_read_desc_t",9,"off")=48
	Set gtmtypes("mur_read_desc_t",9,"len")=4
	Set gtmtypes("mur_read_desc_t",9,"type")="boolean_t"
	Set gtmtypfldindx("mur_read_desc_t","random_buff.read_in_progress")=9
	Set gtmtypes("mur_read_desc_t",10,"name")="mur_read_desc_t.random_buff.aiocbp"
	Set gtmtypes("mur_read_desc_t",10,"off")=56
	Set gtmtypes("mur_read_desc_t",10,"len")=8
	Set gtmtypes("mur_read_desc_t",10,"type")="addr"
	Set gtmtypfldindx("mur_read_desc_t","random_buff.aiocbp")=10
	Set gtmtypes("mur_read_desc_t",11,"name")="mur_read_desc_t.random_buff.rip_channel"
	Set gtmtypes("mur_read_desc_t",11,"off")=64
	Set gtmtypes("mur_read_desc_t",11,"len")=4
	Set gtmtypes("mur_read_desc_t",11,"type")="int"
	Set gtmtypfldindx("mur_read_desc_t","random_buff.rip_channel")=11
	Set gtmtypes("mur_read_desc_t",12,"name")="mur_read_desc_t.aux_buff1"
	Set gtmtypes("mur_read_desc_t",12,"off")=72
	Set gtmtypes("mur_read_desc_t",12,"len")=8
	Set gtmtypes("mur_read_desc_t",12,"type")="addr"
	Set gtmtypfldindx("mur_read_desc_t","aux_buff1")=12
	Set gtmtypes("mur_read_desc_t",13,"name")="mur_read_desc_t.seq_buff"
	Set gtmtypes("mur_read_desc_t",13,"off")=80
	Set gtmtypes("mur_read_desc_t",13,"len")=96
	Set gtmtypes("mur_read_desc_t",13,"type")="mur_buff_desc_t"
	Set gtmtypfldindx("mur_read_desc_t","seq_buff")=13
	Set gtmtypes("mur_read_desc_t",13,"dim")=2
	Set gtmtypes("mur_read_desc_t",14,"name")="mur_read_desc_t.seq_buff[0].base"
	Set gtmtypes("mur_read_desc_t",14,"off")=80
	Set gtmtypes("mur_read_desc_t",14,"len")=8
	Set gtmtypes("mur_read_desc_t",14,"type")="addr"
	Set gtmtypfldindx("mur_read_desc_t","seq_buff[0].base")=14
	Set gtmtypes("mur_read_desc_t",15,"name")="mur_read_desc_t.seq_buff[0].top"
	Set gtmtypes("mur_read_desc_t",15,"off")=88
	Set gtmtypes("mur_read_desc_t",15,"len")=8
	Set gtmtypes("mur_read_desc_t",15,"type")="addr"
	Set gtmtypfldindx("mur_read_desc_t","seq_buff[0].top")=15
	Set gtmtypes("mur_read_desc_t",16,"name")="mur_read_desc_t.seq_buff[0].blen"
	Set gtmtypes("mur_read_desc_t",16,"off")=96
	Set gtmtypes("mur_read_desc_t",16,"len")=4
	Set gtmtypes("mur_read_desc_t",16,"type")="unsigned-int"
	Set gtmtypfldindx("mur_read_desc_t","seq_buff[0].blen")=16
	Set gtmtypes("mur_read_desc_t",17,"name")="mur_read_desc_t.seq_buff[0].dskaddr"
	Set gtmtypes("mur_read_desc_t",17,"off")=100
	Set gtmtypes("mur_read_desc_t",17,"len")=4
	Set gtmtypes("mur_read_desc_t",17,"type")="unsigned-int"
	Set gtmtypfldindx("mur_read_desc_t","seq_buff[0].dskaddr")=17
	Set gtmtypes("mur_read_desc_t",18,"name")="mur_read_desc_t.seq_buff[0].read_in_progress"
	Set gtmtypes("mur_read_desc_t",18,"off")=104
	Set gtmtypes("mur_read_desc_t",18,"len")=4
	Set gtmtypes("mur_read_desc_t",18,"type")="boolean_t"
	Set gtmtypfldindx("mur_read_desc_t","seq_buff[0].read_in_progress")=18
	Set gtmtypes("mur_read_desc_t",19,"name")="mur_read_desc_t.seq_buff[0].aiocbp"
	Set gtmtypes("mur_read_desc_t",19,"off")=112
	Set gtmtypes("mur_read_desc_t",19,"len")=8
	Set gtmtypes("mur_read_desc_t",19,"type")="addr"
	Set gtmtypfldindx("mur_read_desc_t","seq_buff[0].aiocbp")=19
	Set gtmtypes("mur_read_desc_t",20,"name")="mur_read_desc_t.seq_buff[0].rip_channel"
	Set gtmtypes("mur_read_desc_t",20,"off")=120
	Set gtmtypes("mur_read_desc_t",20,"len")=4
	Set gtmtypes("mur_read_desc_t",20,"type")="int"
	Set gtmtypfldindx("mur_read_desc_t","seq_buff[0].rip_channel")=20
	Set gtmtypes("mur_read_desc_t",21,"name")="mur_read_desc_t.aux_buff2"
	Set gtmtypes("mur_read_desc_t",21,"off")=176
	Set gtmtypes("mur_read_desc_t",21,"len")=48
	Set gtmtypes("mur_read_desc_t",21,"type")="mur_buff_desc_t"
	Set gtmtypfldindx("mur_read_desc_t","aux_buff2")=21
	Set gtmtypes("mur_read_desc_t",22,"name")="mur_read_desc_t.aux_buff2.base"
	Set gtmtypes("mur_read_desc_t",22,"off")=176
	Set gtmtypes("mur_read_desc_t",22,"len")=8
	Set gtmtypes("mur_read_desc_t",22,"type")="addr"
	Set gtmtypfldindx("mur_read_desc_t","aux_buff2.base")=22
	Set gtmtypes("mur_read_desc_t",23,"name")="mur_read_desc_t.aux_buff2.top"
	Set gtmtypes("mur_read_desc_t",23,"off")=184
	Set gtmtypes("mur_read_desc_t",23,"len")=8
	Set gtmtypes("mur_read_desc_t",23,"type")="addr"
	Set gtmtypfldindx("mur_read_desc_t","aux_buff2.top")=23
	Set gtmtypes("mur_read_desc_t",24,"name")="mur_read_desc_t.aux_buff2.blen"
	Set gtmtypes("mur_read_desc_t",24,"off")=192
	Set gtmtypes("mur_read_desc_t",24,"len")=4
	Set gtmtypes("mur_read_desc_t",24,"type")="unsigned-int"
	Set gtmtypfldindx("mur_read_desc_t","aux_buff2.blen")=24
	Set gtmtypes("mur_read_desc_t",25,"name")="mur_read_desc_t.aux_buff2.dskaddr"
	Set gtmtypes("mur_read_desc_t",25,"off")=196
	Set gtmtypes("mur_read_desc_t",25,"len")=4
	Set gtmtypes("mur_read_desc_t",25,"type")="unsigned-int"
	Set gtmtypfldindx("mur_read_desc_t","aux_buff2.dskaddr")=25
	Set gtmtypes("mur_read_desc_t",26,"name")="mur_read_desc_t.aux_buff2.read_in_progress"
	Set gtmtypes("mur_read_desc_t",26,"off")=200
	Set gtmtypes("mur_read_desc_t",26,"len")=4
	Set gtmtypes("mur_read_desc_t",26,"type")="boolean_t"
	Set gtmtypfldindx("mur_read_desc_t","aux_buff2.read_in_progress")=26
	Set gtmtypes("mur_read_desc_t",27,"name")="mur_read_desc_t.aux_buff2.aiocbp"
	Set gtmtypes("mur_read_desc_t",27,"off")=208
	Set gtmtypes("mur_read_desc_t",27,"len")=8
	Set gtmtypes("mur_read_desc_t",27,"type")="addr"
	Set gtmtypfldindx("mur_read_desc_t","aux_buff2.aiocbp")=27
	Set gtmtypes("mur_read_desc_t",28,"name")="mur_read_desc_t.aux_buff2.rip_channel"
	Set gtmtypes("mur_read_desc_t",28,"off")=216
	Set gtmtypes("mur_read_desc_t",28,"len")=4
	Set gtmtypes("mur_read_desc_t",28,"type")="int"
	Set gtmtypfldindx("mur_read_desc_t","aux_buff2.rip_channel")=28
	Set gtmtypes("mur_read_desc_t",29,"name")="mur_read_desc_t.buff_index"
	Set gtmtypes("mur_read_desc_t",29,"off")=224
	Set gtmtypes("mur_read_desc_t",29,"len")=4
	Set gtmtypes("mur_read_desc_t",29,"type")="int"
	Set gtmtypfldindx("mur_read_desc_t","buff_index")=29
	Set gtmtypes("mur_read_desc_t",30,"name")="mur_read_desc_t.cur_buff"
	Set gtmtypes("mur_read_desc_t",30,"off")=232
	Set gtmtypes("mur_read_desc_t",30,"len")=8
	Set gtmtypes("mur_read_desc_t",30,"type")="addr"
	Set gtmtypfldindx("mur_read_desc_t","cur_buff")=30
	Set gtmtypes("mur_read_desc_t",31,"name")="mur_read_desc_t.sec_buff"
	Set gtmtypes("mur_read_desc_t",31,"off")=240
	Set gtmtypes("mur_read_desc_t",31,"len")=8
	Set gtmtypes("mur_read_desc_t",31,"type")="addr"
	Set gtmtypfldindx("mur_read_desc_t","sec_buff")=31
	Set gtmtypes("mur_read_desc_t",32,"name")="mur_read_desc_t.jnlrec"
	Set gtmtypes("mur_read_desc_t",32,"off")=248
	Set gtmtypes("mur_read_desc_t",32,"len")=8
	Set gtmtypes("mur_read_desc_t",32,"type")="addr"
	Set gtmtypfldindx("mur_read_desc_t","jnlrec")=32
	Set gtmtypes("mur_read_desc_t",33,"name")="mur_read_desc_t.jreclen"
	Set gtmtypes("mur_read_desc_t",33,"off")=256
	Set gtmtypes("mur_read_desc_t",33,"len")=4
	Set gtmtypes("mur_read_desc_t",33,"type")="unsigned-int"
	Set gtmtypfldindx("mur_read_desc_t","jreclen")=33
	;
	Set gtmtypes("mur_shm_hdr_t")="struct"
	Set gtmtypes("mur_shm_hdr_t",0)=15
	Set gtmtypes("mur_shm_hdr_t","len")=832
	Set gtmtypes("mur_shm_hdr_t",1,"name")="mur_shm_hdr_t.shm_rctl_start"
	Set gtmtypes("mur_shm_hdr_t",1,"off")=0
	Set gtmtypes("mur_shm_hdr_t",1,"len")=8
	Set gtmtypes("mur_shm_hdr_t",1,"type")="addr"
	Set gtmtypfldindx("mur_shm_hdr_t","shm_rctl_start")=1
	Set gtmtypes("mur_shm_hdr_t",2,"name")="mur_shm_hdr_t.hash_bucket_start"
	Set gtmtypes("mur_shm_hdr_t",2,"off")=8
	Set gtmtypes("mur_shm_hdr_t",2,"len")=8
	Set gtmtypes("mur_shm_hdr_t",2,"type")="addr"
	Set gtmtypfldindx("mur_shm_hdr_t","hash_bucket_start")=2
	Set gtmtypes("mur_shm_hdr_t",3,"name")="mur_shm_hdr_t.shm_forw_multi_start"
	Set gtmtypes("mur_shm_hdr_t",3,"off")=16
	Set gtmtypes("mur_shm_hdr_t",3,"len")=8
	Set gtmtypes("mur_shm_hdr_t",3,"type")="addr"
	Set gtmtypfldindx("mur_shm_hdr_t","shm_forw_multi_start")=3
	Set gtmtypes("mur_shm_hdr_t",4,"name")="mur_shm_hdr_t.forw_multi_free"
	Set gtmtypes("mur_shm_hdr_t",4,"off")=24
	Set gtmtypes("mur_shm_hdr_t",4,"len")=16
	Set gtmtypes("mur_shm_hdr_t",4,"type")="que_ent"
	Set gtmtypfldindx("mur_shm_hdr_t","forw_multi_free")=4
	Set gtmtypes("mur_shm_hdr_t",5,"name")="mur_shm_hdr_t.forw_multi_free.fl"
	Set gtmtypes("mur_shm_hdr_t",5,"off")=24
	Set gtmtypes("mur_shm_hdr_t",5,"len")=8
	Set gtmtypes("mur_shm_hdr_t",5,"type")="intptr_t"
	Set gtmtypfldindx("mur_shm_hdr_t","forw_multi_free.fl")=5
	Set gtmtypes("mur_shm_hdr_t",6,"name")="mur_shm_hdr_t.forw_multi_free.bl"
	Set gtmtypes("mur_shm_hdr_t",6,"off")=32
	Set gtmtypes("mur_shm_hdr_t",6,"len")=8
	Set gtmtypes("mur_shm_hdr_t",6,"type")="intptr_t"
	Set gtmtypfldindx("mur_shm_hdr_t","forw_multi_free.bl")=6
	Set gtmtypes("mur_shm_hdr_t",7,"name")="mur_shm_hdr_t.mur_latch"
	Set gtmtypes("mur_shm_hdr_t",7,"off")=40
	Set gtmtypes("mur_shm_hdr_t",7,"len")=8
	Set gtmtypes("mur_shm_hdr_t",7,"type")="global_latch_t"
	Set gtmtypfldindx("mur_shm_hdr_t","mur_latch")=7
	Set gtmtypes("mur_shm_hdr_t",8,"name")="mur_shm_hdr_t.mur_latch.u"
	Set gtmtypes("mur_shm_hdr_t",8,"off")=40
	Set gtmtypes("mur_shm_hdr_t",8,"len")=8
	Set gtmtypes("mur_shm_hdr_t",8,"type")="union"
	Set gtmtypfldindx("mur_shm_hdr_t","mur_latch.u")=8
	Set gtmtypes("mur_shm_hdr_t",9,"name")="mur_shm_hdr_t.mur_latch.u.pid_imgcnt"
	Set gtmtypes("mur_shm_hdr_t",9,"off")=40
	Set gtmtypes("mur_shm_hdr_t",9,"len")=8
	Set gtmtypes("mur_shm_hdr_t",9,"type")="uint64_t"
	Set gtmtypfldindx("mur_shm_hdr_t","mur_latch.u.pid_imgcnt")=9
	Set gtmtypes("mur_shm_hdr_t",10,"name")="mur_shm_hdr_t.mur_latch.u.parts"
	Set gtmtypes("mur_shm_hdr_t",10,"off")=40
	Set gtmtypes("mur_shm_hdr_t",10,"len")=8
	Set gtmtypes("mur_shm_hdr_t",10,"type")="struct"
	Set gtmtypfldindx("mur_shm_hdr_t","mur_latch.u.parts")=10
	Set gtmtypes("mur_shm_hdr_t",11,"name")="mur_shm_hdr_t.mur_latch.u.parts.latch_pid"
	Set gtmtypes("mur_shm_hdr_t",11,"off")=40
	Set gtmtypes("mur_shm_hdr_t",11,"len")=4
	Set gtmtypes("mur_shm_hdr_t",11,"type")="int"
	Set gtmtypfldindx("mur_shm_hdr_t","mur_latch.u.parts.latch_pid")=11
	Set gtmtypes("mur_shm_hdr_t",12,"name")="mur_shm_hdr_t.mur_latch.u.parts.latch_word"
	Set gtmtypes("mur_shm_hdr_t",12,"off")=44
	Set gtmtypes("mur_shm_hdr_t",12,"len")=4
	Set gtmtypes("mur_shm_hdr_t",12,"type")="int"
	Set gtmtypfldindx("mur_shm_hdr_t","mur_latch.u.parts.latch_word")=12
	Set gtmtypes("mur_shm_hdr_t",13,"name")="mur_shm_hdr_t.extr_fn_len"
	Set gtmtypes("mur_shm_hdr_t",13,"off")=48
	Set gtmtypes("mur_shm_hdr_t",13,"len")=12
	Set gtmtypes("mur_shm_hdr_t",13,"type")="int"
	Set gtmtypfldindx("mur_shm_hdr_t","extr_fn_len")=13
	Set gtmtypes("mur_shm_hdr_t",13,"dim")=3
	Set gtmtypes("mur_shm_hdr_t",14,"name")="mur_shm_hdr_t.extr_fn"
	Set gtmtypes("mur_shm_hdr_t",14,"off")=60
	Set gtmtypes("mur_shm_hdr_t",14,"len")=768
	Set gtmtypes("mur_shm_hdr_t",14,"type")="extr_fn_t"
	Set gtmtypfldindx("mur_shm_hdr_t","extr_fn")=14
	Set gtmtypes("mur_shm_hdr_t",14,"dim")=3
	Set gtmtypes("mur_shm_hdr_t",15,"name")="mur_shm_hdr_t.extr_fn[0].fn"
	Set gtmtypes("mur_shm_hdr_t",15,"off")=60
	Set gtmtypes("mur_shm_hdr_t",15,"len")=256
	Set gtmtypes("mur_shm_hdr_t",15,"type")="char"
	Set gtmtypfldindx("mur_shm_hdr_t","extr_fn[0].fn")=15
	;
	Set gtmtypes("mutex_que_entry")="struct"
	Set gtmtypes("mutex_que_entry",0)=8
	Set gtmtypes("mutex_que_entry","len")=72
	Set gtmtypes("mutex_que_entry",1,"name")="mutex_que_entry.que"
	Set gtmtypes("mutex_que_entry",1,"off")=0
	Set gtmtypes("mutex_que_entry",1,"len")=16
	Set gtmtypes("mutex_que_entry",1,"type")="struct"
	Set gtmtypfldindx("mutex_que_entry","que")=1
	Set gtmtypes("mutex_que_entry",2,"name")="mutex_que_entry.que.fl"
	Set gtmtypes("mutex_que_entry",2,"off")=0
	Set gtmtypes("mutex_que_entry",2,"len")=8
	Set gtmtypes("mutex_que_entry",2,"type")="intptr_t"
	Set gtmtypfldindx("mutex_que_entry","que.fl")=2
	Set gtmtypes("mutex_que_entry",3,"name")="mutex_que_entry.que.bl"
	Set gtmtypes("mutex_que_entry",3,"off")=8
	Set gtmtypes("mutex_que_entry",3,"len")=8
	Set gtmtypes("mutex_que_entry",3,"type")="intptr_t"
	Set gtmtypfldindx("mutex_que_entry","que.bl")=3
	Set gtmtypes("mutex_que_entry",4,"name")="mutex_que_entry.pid"
	Set gtmtypes("mutex_que_entry",4,"off")=16
	Set gtmtypes("mutex_que_entry",4,"len")=4
	Set gtmtypes("mutex_que_entry",4,"type")="int"
	Set gtmtypfldindx("mutex_que_entry","pid")=4
	Set gtmtypes("mutex_que_entry",5,"name")="mutex_que_entry.super_crit"
	Set gtmtypes("mutex_que_entry",5,"off")=24
	Set gtmtypes("mutex_que_entry",5,"len")=8
	Set gtmtypes("mutex_que_entry",5,"type")="addr"
	Set gtmtypfldindx("mutex_que_entry","super_crit")=5
	Set gtmtypes("mutex_que_entry",6,"name")="mutex_que_entry.mutex_wake_instance"
	Set gtmtypes("mutex_que_entry",6,"off")=32
	Set gtmtypes("mutex_que_entry",6,"len")=4
	Set gtmtypes("mutex_que_entry",6,"type")="int"
	Set gtmtypfldindx("mutex_que_entry","mutex_wake_instance")=6
	Set gtmtypes("mutex_que_entry",7,"name")="mutex_que_entry.filler1"
	Set gtmtypes("mutex_que_entry",7,"off")=36
	Set gtmtypes("mutex_que_entry",7,"len")=4
	Set gtmtypes("mutex_que_entry",7,"type")="int"
	Set gtmtypfldindx("mutex_que_entry","filler1")=7
	Set gtmtypes("mutex_que_entry",8,"name")="mutex_que_entry.mutex_wake_msem"
	Set gtmtypes("mutex_que_entry",8,"off")=40
	Set gtmtypes("mutex_que_entry",8,"len")=32
	Set gtmtypes("mutex_que_entry",8,"type")="sem_t"
	Set gtmtypfldindx("mutex_que_entry","mutex_wake_msem")=8
	;
	Set gtmtypes("mutex_que_head")="struct"
	Set gtmtypes("mutex_que_head",0)=9
	Set gtmtypes("mutex_que_head","len")=24
	Set gtmtypes("mutex_que_head",1,"name")="mutex_que_head.que"
	Set gtmtypes("mutex_que_head",1,"off")=0
	Set gtmtypes("mutex_que_head",1,"len")=16
	Set gtmtypes("mutex_que_head",1,"type")="struct"
	Set gtmtypfldindx("mutex_que_head","que")=1
	Set gtmtypes("mutex_que_head",2,"name")="mutex_que_head.que.fl"
	Set gtmtypes("mutex_que_head",2,"off")=0
	Set gtmtypes("mutex_que_head",2,"len")=8
	Set gtmtypes("mutex_que_head",2,"type")="intptr_t"
	Set gtmtypfldindx("mutex_que_head","que.fl")=2
	Set gtmtypes("mutex_que_head",3,"name")="mutex_que_head.que.bl"
	Set gtmtypes("mutex_que_head",3,"off")=8
	Set gtmtypes("mutex_que_head",3,"len")=8
	Set gtmtypes("mutex_que_head",3,"type")="intptr_t"
	Set gtmtypfldindx("mutex_que_head","que.bl")=3
	Set gtmtypes("mutex_que_head",4,"name")="mutex_que_head.latch"
	Set gtmtypes("mutex_que_head",4,"off")=16
	Set gtmtypes("mutex_que_head",4,"len")=8
	Set gtmtypes("mutex_que_head",4,"type")="global_latch_t"
	Set gtmtypfldindx("mutex_que_head","latch")=4
	Set gtmtypes("mutex_que_head",5,"name")="mutex_que_head.latch.u"
	Set gtmtypes("mutex_que_head",5,"off")=16
	Set gtmtypes("mutex_que_head",5,"len")=8
	Set gtmtypes("mutex_que_head",5,"type")="union"
	Set gtmtypfldindx("mutex_que_head","latch.u")=5
	Set gtmtypes("mutex_que_head",6,"name")="mutex_que_head.latch.u.pid_imgcnt"
	Set gtmtypes("mutex_que_head",6,"off")=16
	Set gtmtypes("mutex_que_head",6,"len")=8
	Set gtmtypes("mutex_que_head",6,"type")="uint64_t"
	Set gtmtypfldindx("mutex_que_head","latch.u.pid_imgcnt")=6
	Set gtmtypes("mutex_que_head",7,"name")="mutex_que_head.latch.u.parts"
	Set gtmtypes("mutex_que_head",7,"off")=16
	Set gtmtypes("mutex_que_head",7,"len")=8
	Set gtmtypes("mutex_que_head",7,"type")="struct"
	Set gtmtypfldindx("mutex_que_head","latch.u.parts")=7
	Set gtmtypes("mutex_que_head",8,"name")="mutex_que_head.latch.u.parts.latch_pid"
	Set gtmtypes("mutex_que_head",8,"off")=16
	Set gtmtypes("mutex_que_head",8,"len")=4
	Set gtmtypes("mutex_que_head",8,"type")="int"
	Set gtmtypfldindx("mutex_que_head","latch.u.parts.latch_pid")=8
	Set gtmtypes("mutex_que_head",9,"name")="mutex_que_head.latch.u.parts.latch_word"
	Set gtmtypes("mutex_que_head",9,"off")=20
	Set gtmtypes("mutex_que_head",9,"len")=4
	Set gtmtypes("mutex_que_head",9,"type")="int"
	Set gtmtypfldindx("mutex_que_head","latch.u.parts.latch_word")=9
	;
	Set gtmtypes("mutex_spin_parms_struct")="struct"
	Set gtmtypes("mutex_spin_parms_struct",0)=4
	Set gtmtypes("mutex_spin_parms_struct","len")=16
	Set gtmtypes("mutex_spin_parms_struct",1,"name")="mutex_spin_parms_struct.mutex_hard_spin_count"
	Set gtmtypes("mutex_spin_parms_struct",1,"off")=0
	Set gtmtypes("mutex_spin_parms_struct",1,"len")=4
	Set gtmtypes("mutex_spin_parms_struct",1,"type")="int"
	Set gtmtypfldindx("mutex_spin_parms_struct","mutex_hard_spin_count")=1
	Set gtmtypes("mutex_spin_parms_struct",2,"name")="mutex_spin_parms_struct.mutex_sleep_spin_count"
	Set gtmtypes("mutex_spin_parms_struct",2,"off")=4
	Set gtmtypes("mutex_spin_parms_struct",2,"len")=4
	Set gtmtypes("mutex_spin_parms_struct",2,"type")="int"
	Set gtmtypfldindx("mutex_spin_parms_struct","mutex_sleep_spin_count")=2
	Set gtmtypes("mutex_spin_parms_struct",3,"name")="mutex_spin_parms_struct.mutex_spin_sleep_mask"
	Set gtmtypes("mutex_spin_parms_struct",3,"off")=8
	Set gtmtypes("mutex_spin_parms_struct",3,"len")=4
	Set gtmtypes("mutex_spin_parms_struct",3,"type")="int"
	Set gtmtypfldindx("mutex_spin_parms_struct","mutex_spin_sleep_mask")=3
	Set gtmtypes("mutex_spin_parms_struct",4,"name")="mutex_spin_parms_struct.mutex_que_entry_space_size"
	Set gtmtypes("mutex_spin_parms_struct",4,"off")=12
	Set gtmtypes("mutex_spin_parms_struct",4,"len")=4
	Set gtmtypes("mutex_spin_parms_struct",4,"type")="int"
	Set gtmtypfldindx("mutex_spin_parms_struct","mutex_que_entry_space_size")=4
	;
	Set gtmtypes("mutex_struct")="struct"
	Set gtmtypes("mutex_struct",0)=49
	Set gtmtypes("mutex_struct","len")=384
	Set gtmtypes("mutex_struct",1,"name")="mutex_struct.crit_cycle"
	Set gtmtypes("mutex_struct",1,"off")=0
	Set gtmtypes("mutex_struct",1,"len")=4
	Set gtmtypes("mutex_struct",1,"type")="unsigned-int"
	Set gtmtypfldindx("mutex_struct","crit_cycle")=1
	Set gtmtypes("mutex_struct",2,"name")="mutex_struct.semaphore"
	Set gtmtypes("mutex_struct",2,"off")=8
	Set gtmtypes("mutex_struct",2,"len")=8
	Set gtmtypes("mutex_struct",2,"type")="global_latch_t"
	Set gtmtypfldindx("mutex_struct","semaphore")=2
	Set gtmtypes("mutex_struct",3,"name")="mutex_struct.semaphore.u"
	Set gtmtypes("mutex_struct",3,"off")=8
	Set gtmtypes("mutex_struct",3,"len")=8
	Set gtmtypes("mutex_struct",3,"type")="union"
	Set gtmtypfldindx("mutex_struct","semaphore.u")=3
	Set gtmtypes("mutex_struct",4,"name")="mutex_struct.semaphore.u.pid_imgcnt"
	Set gtmtypes("mutex_struct",4,"off")=8
	Set gtmtypes("mutex_struct",4,"len")=8
	Set gtmtypes("mutex_struct",4,"type")="uint64_t"
	Set gtmtypfldindx("mutex_struct","semaphore.u.pid_imgcnt")=4
	Set gtmtypes("mutex_struct",5,"name")="mutex_struct.semaphore.u.parts"
	Set gtmtypes("mutex_struct",5,"off")=8
	Set gtmtypes("mutex_struct",5,"len")=8
	Set gtmtypes("mutex_struct",5,"type")="struct"
	Set gtmtypfldindx("mutex_struct","semaphore.u.parts")=5
	Set gtmtypes("mutex_struct",6,"name")="mutex_struct.semaphore.u.parts.latch_pid"
	Set gtmtypes("mutex_struct",6,"off")=8
	Set gtmtypes("mutex_struct",6,"len")=4
	Set gtmtypes("mutex_struct",6,"type")="int"
	Set gtmtypfldindx("mutex_struct","semaphore.u.parts.latch_pid")=6
	Set gtmtypes("mutex_struct",7,"name")="mutex_struct.semaphore.u.parts.latch_word"
	Set gtmtypes("mutex_struct",7,"off")=12
	Set gtmtypes("mutex_struct",7,"len")=4
	Set gtmtypes("mutex_struct",7,"type")="int"
	Set gtmtypfldindx("mutex_struct","semaphore.u.parts.latch_word")=7
	Set gtmtypes("mutex_struct",8,"name")="mutex_struct.fill_cacheline2"
	Set gtmtypes("mutex_struct",8,"off")=16
	Set gtmtypes("mutex_struct",8,"len")=48
	Set gtmtypes("mutex_struct",8,"type")="char"
	Set gtmtypfldindx("mutex_struct","fill_cacheline2")=8
	Set gtmtypes("mutex_struct",9,"name")="mutex_struct.crashcnt"
	Set gtmtypes("mutex_struct",9,"off")=64
	Set gtmtypes("mutex_struct",9,"len")=4
	Set gtmtypes("mutex_struct",9,"type")="int"
	Set gtmtypfldindx("mutex_struct","crashcnt")=9
	Set gtmtypes("mutex_struct",10,"name")="mutex_struct.crashcnt_latch"
	Set gtmtypes("mutex_struct",10,"off")=72
	Set gtmtypes("mutex_struct",10,"len")=8
	Set gtmtypes("mutex_struct",10,"type")="global_latch_t"
	Set gtmtypfldindx("mutex_struct","crashcnt_latch")=10
	Set gtmtypes("mutex_struct",11,"name")="mutex_struct.crashcnt_latch.u"
	Set gtmtypes("mutex_struct",11,"off")=72
	Set gtmtypes("mutex_struct",11,"len")=8
	Set gtmtypes("mutex_struct",11,"type")="union"
	Set gtmtypfldindx("mutex_struct","crashcnt_latch.u")=11
	Set gtmtypes("mutex_struct",12,"name")="mutex_struct.crashcnt_latch.u.pid_imgcnt"
	Set gtmtypes("mutex_struct",12,"off")=72
	Set gtmtypes("mutex_struct",12,"len")=8
	Set gtmtypes("mutex_struct",12,"type")="uint64_t"
	Set gtmtypfldindx("mutex_struct","crashcnt_latch.u.pid_imgcnt")=12
	Set gtmtypes("mutex_struct",13,"name")="mutex_struct.crashcnt_latch.u.parts"
	Set gtmtypes("mutex_struct",13,"off")=72
	Set gtmtypes("mutex_struct",13,"len")=8
	Set gtmtypes("mutex_struct",13,"type")="struct"
	Set gtmtypfldindx("mutex_struct","crashcnt_latch.u.parts")=13
	Set gtmtypes("mutex_struct",14,"name")="mutex_struct.crashcnt_latch.u.parts.latch_pid"
	Set gtmtypes("mutex_struct",14,"off")=72
	Set gtmtypes("mutex_struct",14,"len")=4
	Set gtmtypes("mutex_struct",14,"type")="int"
	Set gtmtypfldindx("mutex_struct","crashcnt_latch.u.parts.latch_pid")=14
	Set gtmtypes("mutex_struct",15,"name")="mutex_struct.crashcnt_latch.u.parts.latch_word"
	Set gtmtypes("mutex_struct",15,"off")=76
	Set gtmtypes("mutex_struct",15,"len")=4
	Set gtmtypes("mutex_struct",15,"type")="int"
	Set gtmtypfldindx("mutex_struct","crashcnt_latch.u.parts.latch_word")=15
	Set gtmtypes("mutex_struct",16,"name")="mutex_struct.fill_cacheline4"
	Set gtmtypes("mutex_struct",16,"off")=80
	Set gtmtypes("mutex_struct",16,"len")=48
	Set gtmtypes("mutex_struct",16,"type")="char"
	Set gtmtypfldindx("mutex_struct","fill_cacheline4")=16
	Set gtmtypes("mutex_struct",17,"name")="mutex_struct.stuckexec"
	Set gtmtypes("mutex_struct",17,"off")=128
	Set gtmtypes("mutex_struct",17,"len")=24
	Set gtmtypes("mutex_struct",17,"type")="compswap_time_field"
	Set gtmtypfldindx("mutex_struct","stuckexec")=17
	Set gtmtypes("mutex_struct",18,"name")="mutex_struct.stuckexec.time_latch"
	Set gtmtypes("mutex_struct",18,"off")=128
	Set gtmtypes("mutex_struct",18,"len")=8
	Set gtmtypes("mutex_struct",18,"type")="global_latch_t"
	Set gtmtypfldindx("mutex_struct","stuckexec.time_latch")=18
	Set gtmtypes("mutex_struct",19,"name")="mutex_struct.stuckexec.time_latch.u"
	Set gtmtypes("mutex_struct",19,"off")=128
	Set gtmtypes("mutex_struct",19,"len")=8
	Set gtmtypes("mutex_struct",19,"type")="union"
	Set gtmtypfldindx("mutex_struct","stuckexec.time_latch.u")=19
	Set gtmtypes("mutex_struct",20,"name")="mutex_struct.stuckexec.time_latch.u.pid_imgcnt"
	Set gtmtypes("mutex_struct",20,"off")=128
	Set gtmtypes("mutex_struct",20,"len")=8
	Set gtmtypes("mutex_struct",20,"type")="uint64_t"
	Set gtmtypfldindx("mutex_struct","stuckexec.time_latch.u.pid_imgcnt")=20
	Set gtmtypes("mutex_struct",21,"name")="mutex_struct.stuckexec.time_latch.u.parts"
	Set gtmtypes("mutex_struct",21,"off")=128
	Set gtmtypes("mutex_struct",21,"len")=8
	Set gtmtypes("mutex_struct",21,"type")="struct"
	Set gtmtypfldindx("mutex_struct","stuckexec.time_latch.u.parts")=21
	Set gtmtypes("mutex_struct",22,"name")="mutex_struct.stuckexec.time_latch.u.parts.latch_pid"
	Set gtmtypes("mutex_struct",22,"off")=128
	Set gtmtypes("mutex_struct",22,"len")=4
	Set gtmtypes("mutex_struct",22,"type")="int"
	Set gtmtypfldindx("mutex_struct","stuckexec.time_latch.u.parts.latch_pid")=22
	Set gtmtypes("mutex_struct",23,"name")="mutex_struct.stuckexec.time_latch.u.parts.latch_word"
	Set gtmtypes("mutex_struct",23,"off")=132
	Set gtmtypes("mutex_struct",23,"len")=4
	Set gtmtypes("mutex_struct",23,"type")="int"
	Set gtmtypfldindx("mutex_struct","stuckexec.time_latch.u.parts.latch_word")=23
	Set gtmtypes("mutex_struct",24,"name")="mutex_struct.stuckexec.hp_latch_space"
	Set gtmtypes("mutex_struct",24,"off")=136
	Set gtmtypes("mutex_struct",24,"len")=16
	Set gtmtypes("mutex_struct",24,"type")="int"
	Set gtmtypfldindx("mutex_struct","stuckexec.hp_latch_space")=24
	Set gtmtypes("mutex_struct",24,"dim")=4
	Set gtmtypes("mutex_struct",25,"name")="mutex_struct.fill_cacheline5"
	Set gtmtypes("mutex_struct",25,"off")=152
	Set gtmtypes("mutex_struct",25,"len")=40
	Set gtmtypes("mutex_struct",25,"type")="char"
	Set gtmtypfldindx("mutex_struct","fill_cacheline5")=25
	Set gtmtypes("mutex_struct",26,"name")="mutex_struct.queslots"
	Set gtmtypes("mutex_struct",26,"off")=192
	Set gtmtypes("mutex_struct",26,"len")=4
	Set gtmtypes("mutex_struct",26,"type")="int"
	Set gtmtypfldindx("mutex_struct","queslots")=26
	Set gtmtypes("mutex_struct",27,"name")="mutex_struct.fill_cacheline7"
	Set gtmtypes("mutex_struct",27,"off")=196
	Set gtmtypes("mutex_struct",27,"len")=56
	Set gtmtypes("mutex_struct",27,"type")="char"
	Set gtmtypfldindx("mutex_struct","fill_cacheline7")=27
	Set gtmtypes("mutex_struct",28,"name")="mutex_struct.prochead"
	Set gtmtypes("mutex_struct",28,"off")=256
	Set gtmtypes("mutex_struct",28,"len")=24
	Set gtmtypes("mutex_struct",28,"type")="mutex_que_head"
	Set gtmtypfldindx("mutex_struct","prochead")=28
	Set gtmtypes("mutex_struct",29,"name")="mutex_struct.prochead.que"
	Set gtmtypes("mutex_struct",29,"off")=256
	Set gtmtypes("mutex_struct",29,"len")=16
	Set gtmtypes("mutex_struct",29,"type")="struct"
	Set gtmtypfldindx("mutex_struct","prochead.que")=29
	Set gtmtypes("mutex_struct",30,"name")="mutex_struct.prochead.que.fl"
	Set gtmtypes("mutex_struct",30,"off")=256
	Set gtmtypes("mutex_struct",30,"len")=8
	Set gtmtypes("mutex_struct",30,"type")="intptr_t"
	Set gtmtypfldindx("mutex_struct","prochead.que.fl")=30
	Set gtmtypes("mutex_struct",31,"name")="mutex_struct.prochead.que.bl"
	Set gtmtypes("mutex_struct",31,"off")=264
	Set gtmtypes("mutex_struct",31,"len")=8
	Set gtmtypes("mutex_struct",31,"type")="intptr_t"
	Set gtmtypfldindx("mutex_struct","prochead.que.bl")=31
	Set gtmtypes("mutex_struct",32,"name")="mutex_struct.prochead.latch"
	Set gtmtypes("mutex_struct",32,"off")=272
	Set gtmtypes("mutex_struct",32,"len")=8
	Set gtmtypes("mutex_struct",32,"type")="global_latch_t"
	Set gtmtypfldindx("mutex_struct","prochead.latch")=32
	Set gtmtypes("mutex_struct",33,"name")="mutex_struct.prochead.latch.u"
	Set gtmtypes("mutex_struct",33,"off")=272
	Set gtmtypes("mutex_struct",33,"len")=8
	Set gtmtypes("mutex_struct",33,"type")="union"
	Set gtmtypfldindx("mutex_struct","prochead.latch.u")=33
	Set gtmtypes("mutex_struct",34,"name")="mutex_struct.prochead.latch.u.pid_imgcnt"
	Set gtmtypes("mutex_struct",34,"off")=272
	Set gtmtypes("mutex_struct",34,"len")=8
	Set gtmtypes("mutex_struct",34,"type")="uint64_t"
	Set gtmtypfldindx("mutex_struct","prochead.latch.u.pid_imgcnt")=34
	Set gtmtypes("mutex_struct",35,"name")="mutex_struct.prochead.latch.u.parts"
	Set gtmtypes("mutex_struct",35,"off")=272
	Set gtmtypes("mutex_struct",35,"len")=8
	Set gtmtypes("mutex_struct",35,"type")="struct"
	Set gtmtypfldindx("mutex_struct","prochead.latch.u.parts")=35
	Set gtmtypes("mutex_struct",36,"name")="mutex_struct.prochead.latch.u.parts.latch_pid"
	Set gtmtypes("mutex_struct",36,"off")=272
	Set gtmtypes("mutex_struct",36,"len")=4
	Set gtmtypes("mutex_struct",36,"type")="int"
	Set gtmtypfldindx("mutex_struct","prochead.latch.u.parts.latch_pid")=36
	Set gtmtypes("mutex_struct",37,"name")="mutex_struct.prochead.latch.u.parts.latch_word"
	Set gtmtypes("mutex_struct",37,"off")=276
	Set gtmtypes("mutex_struct",37,"len")=4
	Set gtmtypes("mutex_struct",37,"type")="int"
	Set gtmtypfldindx("mutex_struct","prochead.latch.u.parts.latch_word")=37
	Set gtmtypes("mutex_struct",38,"name")="mutex_struct.fill_cacheline8"
	Set gtmtypes("mutex_struct",38,"off")=280
	Set gtmtypes("mutex_struct",38,"len")=40
	Set gtmtypes("mutex_struct",38,"type")="char"
	Set gtmtypfldindx("mutex_struct","fill_cacheline8")=38
	Set gtmtypes("mutex_struct",39,"name")="mutex_struct.freehead"
	Set gtmtypes("mutex_struct",39,"off")=320
	Set gtmtypes("mutex_struct",39,"len")=24
	Set gtmtypes("mutex_struct",39,"type")="mutex_que_head"
	Set gtmtypfldindx("mutex_struct","freehead")=39
	Set gtmtypes("mutex_struct",40,"name")="mutex_struct.freehead.que"
	Set gtmtypes("mutex_struct",40,"off")=320
	Set gtmtypes("mutex_struct",40,"len")=16
	Set gtmtypes("mutex_struct",40,"type")="struct"
	Set gtmtypfldindx("mutex_struct","freehead.que")=40
	Set gtmtypes("mutex_struct",41,"name")="mutex_struct.freehead.que.fl"
	Set gtmtypes("mutex_struct",41,"off")=320
	Set gtmtypes("mutex_struct",41,"len")=8
	Set gtmtypes("mutex_struct",41,"type")="intptr_t"
	Set gtmtypfldindx("mutex_struct","freehead.que.fl")=41
	Set gtmtypes("mutex_struct",42,"name")="mutex_struct.freehead.que.bl"
	Set gtmtypes("mutex_struct",42,"off")=328
	Set gtmtypes("mutex_struct",42,"len")=8
	Set gtmtypes("mutex_struct",42,"type")="intptr_t"
	Set gtmtypfldindx("mutex_struct","freehead.que.bl")=42
	Set gtmtypes("mutex_struct",43,"name")="mutex_struct.freehead.latch"
	Set gtmtypes("mutex_struct",43,"off")=336
	Set gtmtypes("mutex_struct",43,"len")=8
	Set gtmtypes("mutex_struct",43,"type")="global_latch_t"
	Set gtmtypfldindx("mutex_struct","freehead.latch")=43
	Set gtmtypes("mutex_struct",44,"name")="mutex_struct.freehead.latch.u"
	Set gtmtypes("mutex_struct",44,"off")=336
	Set gtmtypes("mutex_struct",44,"len")=8
	Set gtmtypes("mutex_struct",44,"type")="union"
	Set gtmtypfldindx("mutex_struct","freehead.latch.u")=44
	Set gtmtypes("mutex_struct",45,"name")="mutex_struct.freehead.latch.u.pid_imgcnt"
	Set gtmtypes("mutex_struct",45,"off")=336
	Set gtmtypes("mutex_struct",45,"len")=8
	Set gtmtypes("mutex_struct",45,"type")="uint64_t"
	Set gtmtypfldindx("mutex_struct","freehead.latch.u.pid_imgcnt")=45
	Set gtmtypes("mutex_struct",46,"name")="mutex_struct.freehead.latch.u.parts"
	Set gtmtypes("mutex_struct",46,"off")=336
	Set gtmtypes("mutex_struct",46,"len")=8
	Set gtmtypes("mutex_struct",46,"type")="struct"
	Set gtmtypfldindx("mutex_struct","freehead.latch.u.parts")=46
	Set gtmtypes("mutex_struct",47,"name")="mutex_struct.freehead.latch.u.parts.latch_pid"
	Set gtmtypes("mutex_struct",47,"off")=336
	Set gtmtypes("mutex_struct",47,"len")=4
	Set gtmtypes("mutex_struct",47,"type")="int"
	Set gtmtypfldindx("mutex_struct","freehead.latch.u.parts.latch_pid")=47
	Set gtmtypes("mutex_struct",48,"name")="mutex_struct.freehead.latch.u.parts.latch_word"
	Set gtmtypes("mutex_struct",48,"off")=340
	Set gtmtypes("mutex_struct",48,"len")=4
	Set gtmtypes("mutex_struct",48,"type")="int"
	Set gtmtypfldindx("mutex_struct","freehead.latch.u.parts.latch_word")=48
	Set gtmtypes("mutex_struct",49,"name")="mutex_struct.fill_cacheline9"
	Set gtmtypes("mutex_struct",49,"off")=344
	Set gtmtypes("mutex_struct",49,"len")=40
	Set gtmtypes("mutex_struct",49,"type")="char"
	Set gtmtypfldindx("mutex_struct","fill_cacheline9")=49
	;
	Set gtmtypes("mutex_wake_msg_t")="struct"
	Set gtmtypes("mutex_wake_msg_t",0)=2
	Set gtmtypes("mutex_wake_msg_t","len")=8
	Set gtmtypes("mutex_wake_msg_t",1,"name")="mutex_wake_msg_t.pid"
	Set gtmtypes("mutex_wake_msg_t",1,"off")=0
	Set gtmtypes("mutex_wake_msg_t",1,"len")=4
	Set gtmtypes("mutex_wake_msg_t",1,"type")="unsigned-int"
	Set gtmtypfldindx("mutex_wake_msg_t","pid")=1
	Set gtmtypes("mutex_wake_msg_t",2,"name")="mutex_wake_msg_t.mutex_wake_instance"
	Set gtmtypes("mutex_wake_msg_t",2,"off")=4
	Set gtmtypes("mutex_wake_msg_t",2,"len")=4
	Set gtmtypes("mutex_wake_msg_t",2,"type")="int"
	Set gtmtypfldindx("mutex_wake_msg_t","mutex_wake_instance")=2
	;
	Set gtmtypes("mv_stent")="struct"
	Set gtmtypes("mv_stent",0)=91
	Set gtmtypes("mv_stent","len")=224
	Set gtmtypes("mv_stent",1,"name")="mv_stent.mv_st_cont"
	Set gtmtypes("mv_stent",1,"off")=8
	Set gtmtypes("mv_stent",1,"len")=216
	Set gtmtypes("mv_stent",1,"type")="union"
	Set gtmtypfldindx("mv_stent","mv_st_cont")=1
	Set gtmtypes("mv_stent",2,"name")="mv_stent.mv_st_cont.mvs_mval"
	Set gtmtypes("mv_stent",2,"off")=8
	Set gtmtypes("mv_stent",2,"len")=32
	Set gtmtypes("mv_stent",2,"type")="mval"
	Set gtmtypfldindx("mv_stent","mv_st_cont.mvs_mval")=2
	Set gtmtypes("mv_stent",3,"name")="mv_stent.mv_st_cont.mvs_mval.mvtype"
	Set gtmtypes("mv_stent",3,"off")=8
	Set gtmtypes("mv_stent",3,"len")=2
	Set gtmtypes("mv_stent",3,"type")="unsigned-short"
	Set gtmtypfldindx("mv_stent","mv_st_cont.mvs_mval.mvtype")=3
	Set gtmtypes("mv_stent",4,"name")="mv_stent.mv_st_cont.mvs_mval.fnpc_indx"
	Set gtmtypes("mv_stent",4,"off")=11
	Set gtmtypes("mv_stent",4,"len")=1
	Set gtmtypes("mv_stent",4,"type")="unsigned-char"
	Set gtmtypfldindx("mv_stent","mv_st_cont.mvs_mval.fnpc_indx")=4
	Set gtmtypes("mv_stent",5,"name")="mv_stent.mv_st_cont.mvs_mval.utfcgr_indx"
	Set gtmtypes("mv_stent",5,"off")=12
	Set gtmtypes("mv_stent",5,"len")=4
	Set gtmtypes("mv_stent",5,"type")="unsigned-int"
	Set gtmtypfldindx("mv_stent","mv_st_cont.mvs_mval.utfcgr_indx")=5
	Set gtmtypes("mv_stent",6,"name")="mv_stent.mv_st_cont.mvs_mval.m"
	Set gtmtypes("mv_stent",6,"off")=16
	Set gtmtypes("mv_stent",6,"len")=8
	Set gtmtypes("mv_stent",6,"type")="int"
	Set gtmtypfldindx("mv_stent","mv_st_cont.mvs_mval.m")=6
	Set gtmtypes("mv_stent",6,"dim")=2
	Set gtmtypes("mv_stent",7,"name")="mv_stent.mv_st_cont.mvs_mval.str"
	Set gtmtypes("mv_stent",7,"off")=24
	Set gtmtypes("mv_stent",7,"len")=16
	Set gtmtypes("mv_stent",7,"type")="mstr"
	Set gtmtypfldindx("mv_stent","mv_st_cont.mvs_mval.str")=7
	Set gtmtypes("mv_stent",8,"name")="mv_stent.mv_st_cont.mvs_lvval"
	Set gtmtypes("mv_stent",8,"off")=8
	Set gtmtypes("mv_stent",8,"len")=8
	Set gtmtypes("mv_stent",8,"type")="addr"
	Set gtmtypfldindx("mv_stent","mv_st_cont.mvs_lvval")=8
	Set gtmtypes("mv_stent",9,"name")="mv_stent.mv_st_cont.mvs_msav"
	Set gtmtypes("mv_stent",9,"off")=8
	Set gtmtypes("mv_stent",9,"len")=40
	Set gtmtypes("mv_stent",9,"type")="struct"
	Set gtmtypfldindx("mv_stent","mv_st_cont.mvs_msav")=9
	Set gtmtypes("mv_stent",10,"name")="mv_stent.mv_st_cont.mvs_msav.v"
	Set gtmtypes("mv_stent",10,"off")=8
	Set gtmtypes("mv_stent",10,"len")=32
	Set gtmtypes("mv_stent",10,"type")="mval"
	Set gtmtypfldindx("mv_stent","mv_st_cont.mvs_msav.v")=10
	Set gtmtypes("mv_stent",11,"name")="mv_stent.mv_st_cont.mvs_msav.v.mvtype"
	Set gtmtypes("mv_stent",11,"off")=8
	Set gtmtypes("mv_stent",11,"len")=2
	Set gtmtypes("mv_stent",11,"type")="unsigned-short"
	Set gtmtypfldindx("mv_stent","mv_st_cont.mvs_msav.v.mvtype")=11
	Set gtmtypes("mv_stent",12,"name")="mv_stent.mv_st_cont.mvs_msav.v.fnpc_indx"
	Set gtmtypes("mv_stent",12,"off")=11
	Set gtmtypes("mv_stent",12,"len")=1
	Set gtmtypes("mv_stent",12,"type")="unsigned-char"
	Set gtmtypfldindx("mv_stent","mv_st_cont.mvs_msav.v.fnpc_indx")=12
	Set gtmtypes("mv_stent",13,"name")="mv_stent.mv_st_cont.mvs_msav.v.utfcgr_indx"
	Set gtmtypes("mv_stent",13,"off")=12
	Set gtmtypes("mv_stent",13,"len")=4
	Set gtmtypes("mv_stent",13,"type")="unsigned-int"
	Set gtmtypfldindx("mv_stent","mv_st_cont.mvs_msav.v.utfcgr_indx")=13
	Set gtmtypes("mv_stent",14,"name")="mv_stent.mv_st_cont.mvs_msav.v.m"
	Set gtmtypes("mv_stent",14,"off")=16
	Set gtmtypes("mv_stent",14,"len")=8
	Set gtmtypes("mv_stent",14,"type")="int"
	Set gtmtypfldindx("mv_stent","mv_st_cont.mvs_msav.v.m")=14
	Set gtmtypes("mv_stent",14,"dim")=2
	Set gtmtypes("mv_stent",15,"name")="mv_stent.mv_st_cont.mvs_msav.v.str"
	Set gtmtypes("mv_stent",15,"off")=24
	Set gtmtypes("mv_stent",15,"len")=16
	Set gtmtypes("mv_stent",15,"type")="mstr"
	Set gtmtypfldindx("mv_stent","mv_st_cont.mvs_msav.v.str")=15
	Set gtmtypes("mv_stent",16,"name")="mv_stent.mv_st_cont.mvs_msav.addr"
	Set gtmtypes("mv_stent",16,"off")=40
	Set gtmtypes("mv_stent",16,"len")=8
	Set gtmtypes("mv_stent",16,"type")="addr"
	Set gtmtypfldindx("mv_stent","mv_st_cont.mvs_msav.addr")=16
	Set gtmtypes("mv_stent",17,"name")="mv_stent.mv_st_cont.mvs_stab"
	Set gtmtypes("mv_stent",17,"off")=8
	Set gtmtypes("mv_stent",17,"len")=8
	Set gtmtypes("mv_stent",17,"type")="addr"
	Set gtmtypfldindx("mv_stent","mv_st_cont.mvs_stab")=17
	Set gtmtypes("mv_stent",18,"name")="mv_stent.mv_st_cont.mvs_iarr"
	Set gtmtypes("mv_stent",18,"off")=8
	Set gtmtypes("mv_stent",18,"len")=16
	Set gtmtypes("mv_stent",18,"type")="struct"
	Set gtmtypfldindx("mv_stent","mv_st_cont.mvs_iarr")=18
	Set gtmtypes("mv_stent",19,"name")="mv_stent.mv_st_cont.mvs_iarr.iarr_mvals"
	Set gtmtypes("mv_stent",19,"off")=8
	Set gtmtypes("mv_stent",19,"len")=2
	Set gtmtypes("mv_stent",19,"type")="unsigned-short"
	Set gtmtypfldindx("mv_stent","mv_st_cont.mvs_iarr.iarr_mvals")=19
	Set gtmtypes("mv_stent",20,"name")="mv_stent.mv_st_cont.mvs_iarr.iarr_base"
	Set gtmtypes("mv_stent",20,"off")=16
	Set gtmtypes("mv_stent",20,"len")=8
	Set gtmtypes("mv_stent",20,"type")="addr"
	Set gtmtypfldindx("mv_stent","mv_st_cont.mvs_iarr.iarr_base")=20
	Set gtmtypes("mv_stent",21,"name")="mv_stent.mv_st_cont.mvs_stck"
	Set gtmtypes("mv_stent",21,"off")=8
	Set gtmtypes("mv_stent",21,"len")=24
	Set gtmtypes("mv_stent",21,"type")="struct"
	Set gtmtypfldindx("mv_stent","mv_st_cont.mvs_stck")=21
	Set gtmtypes("mv_stent",22,"name")="mv_stent.mv_st_cont.mvs_stck.mvs_stck_addr"
	Set gtmtypes("mv_stent",22,"off")=8
	Set gtmtypes("mv_stent",22,"len")=8
	Set gtmtypes("mv_stent",22,"type")="addr"
	Set gtmtypfldindx("mv_stent","mv_st_cont.mvs_stck.mvs_stck_addr")=22
	Set gtmtypes("mv_stent",23,"name")="mv_stent.mv_st_cont.mvs_stck.mvs_stck_val"
	Set gtmtypes("mv_stent",23,"off")=16
	Set gtmtypes("mv_stent",23,"len")=8
	Set gtmtypes("mv_stent",23,"type")="addr"
	Set gtmtypfldindx("mv_stent","mv_st_cont.mvs_stck.mvs_stck_val")=23
	Set gtmtypes("mv_stent",24,"name")="mv_stent.mv_st_cont.mvs_stck.mvs_stck_size"
	Set gtmtypes("mv_stent",24,"off")=24
	Set gtmtypes("mv_stent",24,"len")=4
	Set gtmtypes("mv_stent",24,"type")="int"
	Set gtmtypfldindx("mv_stent","mv_st_cont.mvs_stck.mvs_stck_size")=24
	Set gtmtypes("mv_stent",25,"name")="mv_stent.mv_st_cont.mvs_ntab"
	Set gtmtypes("mv_stent",25,"off")=8
	Set gtmtypes("mv_stent",25,"len")=16
	Set gtmtypes("mv_stent",25,"type")="mvs_ntab_struct"
	Set gtmtypfldindx("mv_stent","mv_st_cont.mvs_ntab")=25
	Set gtmtypes("mv_stent",26,"name")="mv_stent.mv_st_cont.mvs_ntab.hte_addr"
	Set gtmtypes("mv_stent",26,"off")=8
	Set gtmtypes("mv_stent",26,"len")=8
	Set gtmtypes("mv_stent",26,"type")="addr"
	Set gtmtypfldindx("mv_stent","mv_st_cont.mvs_ntab.hte_addr")=26
	Set gtmtypes("mv_stent",27,"name")="mv_stent.mv_st_cont.mvs_ntab.save_value"
	Set gtmtypes("mv_stent",27,"off")=16
	Set gtmtypes("mv_stent",27,"len")=8
	Set gtmtypes("mv_stent",27,"type")="addr"
	Set gtmtypfldindx("mv_stent","mv_st_cont.mvs_ntab.save_value")=27
	Set gtmtypes("mv_stent",28,"name")="mv_stent.mv_st_cont.mvs_zintdev"
	Set gtmtypes("mv_stent",28,"off")=8
	Set gtmtypes("mv_stent",28,"len")=40
	Set gtmtypes("mv_stent",28,"type")="mvs_zintdev_struct"
	Set gtmtypfldindx("mv_stent","mv_st_cont.mvs_zintdev")=28
	Set gtmtypes("mv_stent",29,"name")="mv_stent.mv_st_cont.mvs_zintdev.io_ptr"
	Set gtmtypes("mv_stent",29,"off")=8
	Set gtmtypes("mv_stent",29,"len")=8
	Set gtmtypes("mv_stent",29,"type")="addr"
	Set gtmtypfldindx("mv_stent","mv_st_cont.mvs_zintdev.io_ptr")=29
	Set gtmtypes("mv_stent",30,"name")="mv_stent.mv_st_cont.mvs_zintdev.buffer_valid"
	Set gtmtypes("mv_stent",30,"off")=16
	Set gtmtypes("mv_stent",30,"len")=4
	Set gtmtypes("mv_stent",30,"type")="boolean_t"
	Set gtmtypfldindx("mv_stent","mv_st_cont.mvs_zintdev.buffer_valid")=30
	Set gtmtypes("mv_stent",31,"name")="mv_stent.mv_st_cont.mvs_zintdev.curr_sp_buffer"
	Set gtmtypes("mv_stent",31,"off")=24
	Set gtmtypes("mv_stent",31,"len")=16
	Set gtmtypes("mv_stent",31,"type")="mstr"
	Set gtmtypfldindx("mv_stent","mv_st_cont.mvs_zintdev.curr_sp_buffer")=31
	Set gtmtypes("mv_stent",32,"name")="mv_stent.mv_st_cont.mvs_zintdev.socketptr"
	Set gtmtypes("mv_stent",32,"off")=40
	Set gtmtypes("mv_stent",32,"len")=8
	Set gtmtypes("mv_stent",32,"type")="addr"
	Set gtmtypfldindx("mv_stent","mv_st_cont.mvs_zintdev.socketptr")=32
	Set gtmtypes("mv_stent",33,"name")="mv_stent.mv_st_cont.mvs_pval"
	Set gtmtypes("mv_stent",33,"off")=8
	Set gtmtypes("mv_stent",33,"len")=24
	Set gtmtypes("mv_stent",33,"type")="mvs_pval_struct"
	Set gtmtypfldindx("mv_stent","mv_st_cont.mvs_pval")=33
	Set gtmtypes("mv_stent",34,"name")="mv_stent.mv_st_cont.mvs_pval.mvs_val"
	Set gtmtypes("mv_stent",34,"off")=8
	Set gtmtypes("mv_stent",34,"len")=8
	Set gtmtypes("mv_stent",34,"type")="addr"
	Set gtmtypfldindx("mv_stent","mv_st_cont.mvs_pval.mvs_val")=34
	Set gtmtypes("mv_stent",35,"name")="mv_stent.mv_st_cont.mvs_pval.mvs_ptab"
	Set gtmtypes("mv_stent",35,"off")=16
	Set gtmtypes("mv_stent",35,"len")=16
	Set gtmtypes("mv_stent",35,"type")="mvs_ntab_struct"
	Set gtmtypfldindx("mv_stent","mv_st_cont.mvs_pval.mvs_ptab")=35
	Set gtmtypes("mv_stent",36,"name")="mv_stent.mv_st_cont.mvs_nval"
	Set gtmtypes("mv_stent",36,"off")=8
	Set gtmtypes("mv_stent",36,"len")=24
	Set gtmtypes("mv_stent",36,"type")="mvs_nval_struct"
	Set gtmtypfldindx("mv_stent","mv_st_cont.mvs_nval")=36
	Set gtmtypes("mv_stent",37,"name")="mv_stent.mv_st_cont.mvs_nval.mvs_val"
	Set gtmtypes("mv_stent",37,"off")=8
	Set gtmtypes("mv_stent",37,"len")=8
	Set gtmtypes("mv_stent",37,"type")="addr"
	Set gtmtypfldindx("mv_stent","mv_st_cont.mvs_nval.mvs_val")=37
	Set gtmtypes("mv_stent",38,"name")="mv_stent.mv_st_cont.mvs_nval.mvs_ptab"
	Set gtmtypes("mv_stent",38,"off")=16
	Set gtmtypes("mv_stent",38,"len")=16
	Set gtmtypes("mv_stent",38,"type")="mvs_ntab_struct"
	Set gtmtypfldindx("mv_stent","mv_st_cont.mvs_nval.mvs_ptab")=38
	Set gtmtypes("mv_stent",39,"name")="mv_stent.mv_st_cont.mvs_zintr"
	Set gtmtypes("mv_stent",39,"off")=8
	Set gtmtypes("mv_stent",39,"len")=184
	Set gtmtypes("mv_stent",39,"type")="mvs_zintr_struct"
	Set gtmtypfldindx("mv_stent","mv_st_cont.mvs_zintr")=39
	Set gtmtypes("mv_stent",40,"name")="mv_stent.mv_st_cont.mvs_zintr.saved_dollar_truth"
	Set gtmtypes("mv_stent",40,"off")=8
	Set gtmtypes("mv_stent",40,"len")=4
	Set gtmtypes("mv_stent",40,"type")="boolean_t"
	Set gtmtypfldindx("mv_stent","mv_st_cont.mvs_zintr.saved_dollar_truth")=40
	Set gtmtypes("mv_stent",41,"name")="mv_stent.mv_st_cont.mvs_zintr.filler"
	Set gtmtypes("mv_stent",41,"off")=12
	Set gtmtypes("mv_stent",41,"len")=4
	Set gtmtypes("mv_stent",41,"type")="int"
	Set gtmtypfldindx("mv_stent","mv_st_cont.mvs_zintr.filler")=41
	Set gtmtypes("mv_stent",42,"name")="mv_stent.mv_st_cont.mvs_zintr.savtarg"
	Set gtmtypes("mv_stent",42,"off")=16
	Set gtmtypes("mv_stent",42,"len")=32
	Set gtmtypes("mv_stent",42,"type")="mval"
	Set gtmtypfldindx("mv_stent","mv_st_cont.mvs_zintr.savtarg")=42
	Set gtmtypes("mv_stent",43,"name")="mv_stent.mv_st_cont.mvs_zintr.savextref"
	Set gtmtypes("mv_stent",43,"off")=48
	Set gtmtypes("mv_stent",43,"len")=16
	Set gtmtypes("mv_stent",43,"type")="mstr"
	Set gtmtypfldindx("mv_stent","mv_st_cont.mvs_zintr.savextref")=43
	Set gtmtypes("mv_stent",44,"name")="mv_stent.mv_st_cont.mvs_zintr.error_frame_save"
	Set gtmtypes("mv_stent",44,"off")=64
	Set gtmtypes("mv_stent",44,"len")=8
	Set gtmtypes("mv_stent",44,"type")="addr"
	Set gtmtypfldindx("mv_stent","mv_st_cont.mvs_zintr.error_frame_save")=44
	Set gtmtypes("mv_stent",45,"name")="mv_stent.mv_st_cont.mvs_zintr.dollar_ecode_save"
	Set gtmtypes("mv_stent",45,"off")=72
	Set gtmtypes("mv_stent",45,"len")=80
	Set gtmtypes("mv_stent",45,"type")="dollar_ecode_type"
	Set gtmtypfldindx("mv_stent","mv_st_cont.mvs_zintr.dollar_ecode_save")=45
	Set gtmtypes("mv_stent",46,"name")="mv_stent.mv_st_cont.mvs_zintr.dollar_stack_save"
	Set gtmtypes("mv_stent",46,"off")=152
	Set gtmtypes("mv_stent",46,"len")=40
	Set gtmtypes("mv_stent",46,"type")="dollar_stack_type"
	Set gtmtypfldindx("mv_stent","mv_st_cont.mvs_zintr.dollar_stack_save")=46
	Set gtmtypes("mv_stent",47,"name")="mv_stent.mv_st_cont.mvs_trigr"
	Set gtmtypes("mv_stent",47,"off")=8
	Set gtmtypes("mv_stent",47,"len")=216
	Set gtmtypes("mv_stent",47,"type")="mvs_trigr_struct"
	Set gtmtypfldindx("mv_stent","mv_st_cont.mvs_trigr")=47
	Set gtmtypes("mv_stent",48,"name")="mv_stent.mv_st_cont.mvs_trigr.saved_dollar_truth"
	Set gtmtypes("mv_stent",48,"off")=8
	Set gtmtypes("mv_stent",48,"len")=4
	Set gtmtypes("mv_stent",48,"type")="boolean_t"
	Set gtmtypfldindx("mv_stent","mv_st_cont.mvs_trigr.saved_dollar_truth")=48
	Set gtmtypes("mv_stent",49,"name")="mv_stent.mv_st_cont.mvs_trigr.filler"
	Set gtmtypes("mv_stent",49,"off")=12
	Set gtmtypes("mv_stent",49,"len")=4
	Set gtmtypes("mv_stent",49,"type")="int"
	Set gtmtypfldindx("mv_stent","mv_st_cont.mvs_trigr.filler")=49
	Set gtmtypes("mv_stent",50,"name")="mv_stent.mv_st_cont.mvs_trigr.savtarg"
	Set gtmtypes("mv_stent",50,"off")=16
	Set gtmtypes("mv_stent",50,"len")=32
	Set gtmtypes("mv_stent",50,"type")="mval"
	Set gtmtypfldindx("mv_stent","mv_st_cont.mvs_trigr.savtarg")=50
	Set gtmtypes("mv_stent",51,"name")="mv_stent.mv_st_cont.mvs_trigr.savextref"
	Set gtmtypes("mv_stent",51,"off")=48
	Set gtmtypes("mv_stent",51,"len")=16
	Set gtmtypes("mv_stent",51,"type")="mstr"
	Set gtmtypfldindx("mv_stent","mv_st_cont.mvs_trigr.savextref")=51
	Set gtmtypes("mv_stent",52,"name")="mv_stent.mv_st_cont.mvs_trigr.ztvalue_changed_ptr"
	Set gtmtypes("mv_stent",52,"off")=64
	Set gtmtypes("mv_stent",52,"len")=8
	Set gtmtypes("mv_stent",52,"type")="addr"
	Set gtmtypfldindx("mv_stent","mv_st_cont.mvs_trigr.ztvalue_changed_ptr")=52
	Set gtmtypes("mv_stent",53,"name")="mv_stent.mv_st_cont.mvs_trigr.ztvalue_save"
	Set gtmtypes("mv_stent",53,"off")=72
	Set gtmtypes("mv_stent",53,"len")=8
	Set gtmtypes("mv_stent",53,"type")="addr"
	Set gtmtypfldindx("mv_stent","mv_st_cont.mvs_trigr.ztvalue_save")=53
	Set gtmtypes("mv_stent",54,"name")="mv_stent.mv_st_cont.mvs_trigr.ztname_save"
	Set gtmtypes("mv_stent",54,"off")=80
	Set gtmtypes("mv_stent",54,"len")=8
	Set gtmtypes("mv_stent",54,"type")="addr"
	Set gtmtypfldindx("mv_stent","mv_st_cont.mvs_trigr.ztname_save")=54
	Set gtmtypes("mv_stent",55,"name")="mv_stent.mv_st_cont.mvs_trigr.ztdata_save"
	Set gtmtypes("mv_stent",55,"off")=88
	Set gtmtypes("mv_stent",55,"len")=8
	Set gtmtypes("mv_stent",55,"type")="addr"
	Set gtmtypfldindx("mv_stent","mv_st_cont.mvs_trigr.ztdata_save")=55
	Set gtmtypes("mv_stent",56,"name")="mv_stent.mv_st_cont.mvs_trigr.ztdelim_save"
	Set gtmtypes("mv_stent",56,"off")=96
	Set gtmtypes("mv_stent",56,"len")=8
	Set gtmtypes("mv_stent",56,"type")="addr"
	Set gtmtypfldindx("mv_stent","mv_st_cont.mvs_trigr.ztdelim_save")=56
	Set gtmtypes("mv_stent",57,"name")="mv_stent.mv_st_cont.mvs_trigr.ztoldval_save"
	Set gtmtypes("mv_stent",57,"off")=104
	Set gtmtypes("mv_stent",57,"len")=8
	Set gtmtypes("mv_stent",57,"type")="addr"
	Set gtmtypfldindx("mv_stent","mv_st_cont.mvs_trigr.ztoldval_save")=57
	Set gtmtypes("mv_stent",58,"name")="mv_stent.mv_st_cont.mvs_trigr.ztriggerop_save"
	Set gtmtypes("mv_stent",58,"off")=112
	Set gtmtypes("mv_stent",58,"len")=8
	Set gtmtypes("mv_stent",58,"type")="addr"
	Set gtmtypfldindx("mv_stent","mv_st_cont.mvs_trigr.ztriggerop_save")=58
	Set gtmtypes("mv_stent",59,"name")="mv_stent.mv_st_cont.mvs_trigr.ztupdate_save"
	Set gtmtypes("mv_stent",59,"off")=120
	Set gtmtypes("mv_stent",59,"len")=8
	Set gtmtypes("mv_stent",59,"type")="addr"
	Set gtmtypfldindx("mv_stent","mv_st_cont.mvs_trigr.ztupdate_save")=59
	Set gtmtypes("mv_stent",60,"name")="mv_stent.mv_st_cont.mvs_trigr.ctxt_save"
	Set gtmtypes("mv_stent",60,"off")=128
	Set gtmtypes("mv_stent",60,"len")=8
	Set gtmtypes("mv_stent",60,"type")="addr"
	Set gtmtypfldindx("mv_stent","mv_st_cont.mvs_trigr.ctxt_save")=60
	Set gtmtypes("mv_stent",61,"name")="mv_stent.mv_st_cont.mvs_trigr.gtm_trigger_depth_save"
	Set gtmtypes("mv_stent",61,"off")=136
	Set gtmtypes("mv_stent",61,"len")=4
	Set gtmtypes("mv_stent",61,"type")="int"
	Set gtmtypfldindx("mv_stent","mv_st_cont.mvs_trigr.gtm_trigger_depth_save")=61
	Set gtmtypes("mv_stent",62,"name")="mv_stent.mv_st_cont.mvs_trigr.dollar_etrap_save"
	Set gtmtypes("mv_stent",62,"off")=144
	Set gtmtypes("mv_stent",62,"len")=32
	Set gtmtypes("mv_stent",62,"type")="mval"
	Set gtmtypfldindx("mv_stent","mv_st_cont.mvs_trigr.dollar_etrap_save")=62
	Set gtmtypes("mv_stent",63,"name")="mv_stent.mv_st_cont.mvs_trigr.dollar_ztrap_save"
	Set gtmtypes("mv_stent",63,"off")=176
	Set gtmtypes("mv_stent",63,"len")=32
	Set gtmtypes("mv_stent",63,"type")="mval"
	Set gtmtypfldindx("mv_stent","mv_st_cont.mvs_trigr.dollar_ztrap_save")=63
	Set gtmtypes("mv_stent",64,"name")="mv_stent.mv_st_cont.mvs_trigr.ztrap_explicit_null_save"
	Set gtmtypes("mv_stent",64,"off")=208
	Set gtmtypes("mv_stent",64,"len")=4
	Set gtmtypes("mv_stent",64,"type")="boolean_t"
	Set gtmtypfldindx("mv_stent","mv_st_cont.mvs_trigr.ztrap_explicit_null_save")=64
	Set gtmtypes("mv_stent",65,"name")="mv_stent.mv_st_cont.mvs_trigr.mumps_status_save"
	Set gtmtypes("mv_stent",65,"off")=212
	Set gtmtypes("mv_stent",65,"len")=4
	Set gtmtypes("mv_stent",65,"type")="int"
	Set gtmtypfldindx("mv_stent","mv_st_cont.mvs_trigr.mumps_status_save")=65
	Set gtmtypes("mv_stent",66,"name")="mv_stent.mv_st_cont.mvs_trigr.run_time_save"
	Set gtmtypes("mv_stent",66,"off")=216
	Set gtmtypes("mv_stent",66,"len")=4
	Set gtmtypes("mv_stent",66,"type")="boolean_t"
	Set gtmtypfldindx("mv_stent","mv_st_cont.mvs_trigr.run_time_save")=66
	Set gtmtypes("mv_stent",67,"name")="mv_stent.mv_st_cont.mvs_tp_holder"
	Set gtmtypes("mv_stent",67,"off")=8
	Set gtmtypes("mv_stent",67,"len")=40
	Set gtmtypes("mv_stent",67,"type")="mvs_tphold_struct"
	Set gtmtypfldindx("mv_stent","mv_st_cont.mvs_tp_holder")=67
	Set gtmtypes("mv_stent",68,"name")="mv_stent.mv_st_cont.mvs_tp_holder.tphold_tlevel"
	Set gtmtypes("mv_stent",68,"off")=8
	Set gtmtypes("mv_stent",68,"len")=4
	Set gtmtypes("mv_stent",68,"type")="unsigned-int"
	Set gtmtypfldindx("mv_stent","mv_st_cont.mvs_tp_holder.tphold_tlevel")=68
	Set gtmtypes("mv_stent",69,"name")="mv_stent.mv_st_cont.mvs_tp_holder.ztwormhole_save"
	Set gtmtypes("mv_stent",69,"off")=16
	Set gtmtypes("mv_stent",69,"len")=32
	Set gtmtypes("mv_stent",69,"type")="mval"
	Set gtmtypfldindx("mv_stent","mv_st_cont.mvs_tp_holder.ztwormhole_save")=69
	Set gtmtypes("mv_stent",70,"name")="mv_stent.mv_st_cont.mvs_rstrtpc"
	Set gtmtypes("mv_stent",70,"off")=8
	Set gtmtypes("mv_stent",70,"len")=16
	Set gtmtypes("mv_stent",70,"type")="mvs_rstrtpc_struct"
	Set gtmtypfldindx("mv_stent","mv_st_cont.mvs_rstrtpc")=70
	Set gtmtypes("mv_stent",71,"name")="mv_stent.mv_st_cont.mvs_rstrtpc.restart_pc_save"
	Set gtmtypes("mv_stent",71,"off")=8
	Set gtmtypes("mv_stent",71,"len")=8
	Set gtmtypes("mv_stent",71,"type")="addr"
	Set gtmtypfldindx("mv_stent","mv_st_cont.mvs_rstrtpc.restart_pc_save")=71
	Set gtmtypes("mv_stent",72,"name")="mv_stent.mv_st_cont.mvs_rstrtpc.restart_ctxt_save"
	Set gtmtypes("mv_stent",72,"off")=16
	Set gtmtypes("mv_stent",72,"len")=8
	Set gtmtypes("mv_stent",72,"type")="addr"
	Set gtmtypfldindx("mv_stent","mv_st_cont.mvs_rstrtpc.restart_ctxt_save")=72
	Set gtmtypes("mv_stent",73,"name")="mv_stent.mv_st_cont.mvs_mrgzwrsv"
	Set gtmtypes("mv_stent",73,"off")=8
	Set gtmtypes("mv_stent",73,"len")=56
	Set gtmtypes("mv_stent",73,"type")="mvs_mrgzwrsv_struct"
	Set gtmtypfldindx("mv_stent","mv_st_cont.mvs_mrgzwrsv")=73
	Set gtmtypes("mv_stent",74,"name")="mv_stent.mv_st_cont.mvs_mrgzwrsv.save_merge_args"
	Set gtmtypes("mv_stent",74,"off")=8
	Set gtmtypes("mv_stent",74,"len")=4
	Set gtmtypes("mv_stent",74,"type")="int"
	Set gtmtypfldindx("mv_stent","mv_st_cont.mvs_mrgzwrsv.save_merge_args")=74
	Set gtmtypes("mv_stent",75,"name")="mv_stent.mv_st_cont.mvs_mrgzwrsv.save_zwrtacindx"
	Set gtmtypes("mv_stent",75,"off")=12
	Set gtmtypes("mv_stent",75,"len")=4
	Set gtmtypes("mv_stent",75,"type")="unsigned-int"
	Set gtmtypfldindx("mv_stent","mv_st_cont.mvs_mrgzwrsv.save_zwrtacindx")=75
	Set gtmtypes("mv_stent",76,"name")="mv_stent.mv_st_cont.mvs_mrgzwrsv.save_in_zwrite"
	Set gtmtypes("mv_stent",76,"off")=16
	Set gtmtypes("mv_stent",76,"len")=4
	Set gtmtypes("mv_stent",76,"type")="boolean_t"
	Set gtmtypfldindx("mv_stent","mv_st_cont.mvs_mrgzwrsv.save_in_zwrite")=76
	Set gtmtypes("mv_stent",77,"name")="mv_stent.mv_st_cont.mvs_mrgzwrsv.filler"
	Set gtmtypes("mv_stent",77,"off")=20
	Set gtmtypes("mv_stent",77,"len")=4
	Set gtmtypes("mv_stent",77,"type")="int"
	Set gtmtypfldindx("mv_stent","mv_st_cont.mvs_mrgzwrsv.filler")=77
	Set gtmtypes("mv_stent",78,"name")="mv_stent.mv_st_cont.mvs_mrgzwrsv.save_mglvnp"
	Set gtmtypes("mv_stent",78,"off")=24
	Set gtmtypes("mv_stent",78,"len")=8
	Set gtmtypes("mv_stent",78,"type")="addr"
	Set gtmtypfldindx("mv_stent","mv_st_cont.mvs_mrgzwrsv.save_mglvnp")=78
	Set gtmtypes("mv_stent",79,"name")="mv_stent.mv_st_cont.mvs_mrgzwrsv.save_gvzwrite_block"
	Set gtmtypes("mv_stent",79,"off")=32
	Set gtmtypes("mv_stent",79,"len")=8
	Set gtmtypes("mv_stent",79,"type")="addr"
	Set gtmtypfldindx("mv_stent","mv_st_cont.mvs_mrgzwrsv.save_gvzwrite_block")=79
	Set gtmtypes("mv_stent",80,"name")="mv_stent.mv_st_cont.mvs_mrgzwrsv.save_lvzwrite_block"
	Set gtmtypes("mv_stent",80,"off")=40
	Set gtmtypes("mv_stent",80,"len")=8
	Set gtmtypes("mv_stent",80,"type")="addr"
	Set gtmtypfldindx("mv_stent","mv_st_cont.mvs_mrgzwrsv.save_lvzwrite_block")=80
	Set gtmtypes("mv_stent",81,"name")="mv_stent.mv_st_cont.mvs_mrgzwrsv.save_zwr_output"
	Set gtmtypes("mv_stent",81,"off")=48
	Set gtmtypes("mv_stent",81,"len")=8
	Set gtmtypes("mv_stent",81,"type")="addr"
	Set gtmtypfldindx("mv_stent","mv_st_cont.mvs_mrgzwrsv.save_zwr_output")=81
	Set gtmtypes("mv_stent",82,"name")="mv_stent.mv_st_cont.mvs_mrgzwrsv.save_zwrhtab"
	Set gtmtypes("mv_stent",82,"off")=56
	Set gtmtypes("mv_stent",82,"len")=8
	Set gtmtypes("mv_stent",82,"type")="addr"
	Set gtmtypfldindx("mv_stent","mv_st_cont.mvs_mrgzwrsv.save_zwrhtab")=82
	Set gtmtypes("mv_stent",83,"name")="mv_stent.mv_st_cont.mvs_zintcmd"
	Set gtmtypes("mv_stent",83,"off")=8
	Set gtmtypes("mv_stent",83,"len")=56
	Set gtmtypes("mv_stent",83,"type")="mvs_zintcmd_struct"
	Set gtmtypfldindx("mv_stent","mv_st_cont.mvs_zintcmd")=83
	Set gtmtypes("mv_stent",84,"name")="mv_stent.mv_st_cont.mvs_zintcmd.command"
	Set gtmtypes("mv_stent",84,"off")=8
	Set gtmtypes("mv_stent",84,"len")=4
	Set gtmtypes("mv_stent",84,"type")="int"
	Set gtmtypfldindx("mv_stent","mv_st_cont.mvs_zintcmd.command")=84
	Set gtmtypes("mv_stent",85,"name")="mv_stent.mv_st_cont.mvs_zintcmd.end_or_remain"
	Set gtmtypes("mv_stent",85,"off")=16
	Set gtmtypes("mv_stent",85,"len")=16
	Set gtmtypes("mv_stent",85,"type")="ABS_TIME"
	Set gtmtypfldindx("mv_stent","mv_st_cont.mvs_zintcmd.end_or_remain")=85
	Set gtmtypes("mv_stent",86,"name")="mv_stent.mv_st_cont.mvs_zintcmd.restart_pc_check"
	Set gtmtypes("mv_stent",86,"off")=32
	Set gtmtypes("mv_stent",86,"len")=8
	Set gtmtypes("mv_stent",86,"type")="addr"
	Set gtmtypfldindx("mv_stent","mv_st_cont.mvs_zintcmd.restart_pc_check")=86
	Set gtmtypes("mv_stent",87,"name")="mv_stent.mv_st_cont.mvs_zintcmd.restart_ctxt_check"
	Set gtmtypes("mv_stent",87,"off")=40
	Set gtmtypes("mv_stent",87,"len")=8
	Set gtmtypes("mv_stent",87,"type")="addr"
	Set gtmtypfldindx("mv_stent","mv_st_cont.mvs_zintcmd.restart_ctxt_check")=87
	Set gtmtypes("mv_stent",88,"name")="mv_stent.mv_st_cont.mvs_zintcmd.restart_pc_prior"
	Set gtmtypes("mv_stent",88,"off")=48
	Set gtmtypes("mv_stent",88,"len")=8
	Set gtmtypes("mv_stent",88,"type")="addr"
	Set gtmtypfldindx("mv_stent","mv_st_cont.mvs_zintcmd.restart_pc_prior")=88
	Set gtmtypes("mv_stent",89,"name")="mv_stent.mv_st_cont.mvs_zintcmd.restart_ctxt_prior"
	Set gtmtypes("mv_stent",89,"off")=56
	Set gtmtypes("mv_stent",89,"len")=8
	Set gtmtypes("mv_stent",89,"type")="addr"
	Set gtmtypfldindx("mv_stent","mv_st_cont.mvs_zintcmd.restart_ctxt_prior")=89
	Set gtmtypes("mv_stent",90,"name")="mv_stent.mv_st_cont.mvs_tval"
	Set gtmtypes("mv_stent",90,"off")=8
	Set gtmtypes("mv_stent",90,"len")=4
	Set gtmtypes("mv_stent",90,"type")="int"
	Set gtmtypfldindx("mv_stent","mv_st_cont.mvs_tval")=90
	Set gtmtypes("mv_stent",91,"name")="mv_stent.mv_st_cont.mvs_storig"
	Set gtmtypes("mv_stent",91,"off")=8
	Set gtmtypes("mv_stent",91,"len")=4
	Set gtmtypes("mv_stent",91,"type")="int"
	Set gtmtypfldindx("mv_stent","mv_st_cont.mvs_storig")=91
	;
	Set gtmtypes("mval")="struct"
	Set gtmtypes("mval",0)=8
	Set gtmtypes("mval","len")=32
	Set gtmtypes("mval",1,"name")="mval.mvtype"
	Set gtmtypes("mval",1,"off")=0
	Set gtmtypes("mval",1,"len")=2
	Set gtmtypes("mval",1,"type")="unsigned-short"
	Set gtmtypfldindx("mval","mvtype")=1
	Set gtmtypes("mval",2,"name")="mval.fnpc_indx"
	Set gtmtypes("mval",2,"off")=3
	Set gtmtypes("mval",2,"len")=1
	Set gtmtypes("mval",2,"type")="unsigned-char"
	Set gtmtypfldindx("mval","fnpc_indx")=2
	Set gtmtypes("mval",3,"name")="mval.utfcgr_indx"
	Set gtmtypes("mval",3,"off")=4
	Set gtmtypes("mval",3,"len")=4
	Set gtmtypes("mval",3,"type")="unsigned-int"
	Set gtmtypfldindx("mval","utfcgr_indx")=3
	Set gtmtypes("mval",4,"name")="mval.m"
	Set gtmtypes("mval",4,"off")=8
	Set gtmtypes("mval",4,"len")=8
	Set gtmtypes("mval",4,"type")="int"
	Set gtmtypfldindx("mval","m")=4
	Set gtmtypes("mval",4,"dim")=2
	Set gtmtypes("mval",5,"name")="mval.str"
	Set gtmtypes("mval",5,"off")=16
	Set gtmtypes("mval",5,"len")=16
	Set gtmtypes("mval",5,"type")="mstr"
	Set gtmtypfldindx("mval","str")=5
	Set gtmtypes("mval",6,"name")="mval.str.char_len"
	Set gtmtypes("mval",6,"off")=16
	Set gtmtypes("mval",6,"len")=4
	Set gtmtypes("mval",6,"type")="unsigned-int"
	Set gtmtypfldindx("mval","str.char_len")=6
	Set gtmtypes("mval",7,"name")="mval.str.len"
	Set gtmtypes("mval",7,"off")=20
	Set gtmtypes("mval",7,"len")=4
	Set gtmtypes("mval",7,"type")="int"
	Set gtmtypfldindx("mval","str.len")=7
	Set gtmtypes("mval",8,"name")="mval.str.addr"
	Set gtmtypes("mval",8,"off")=24
	Set gtmtypes("mval",8,"len")=8
	Set gtmtypes("mval",8,"type")="addr"
	Set gtmtypfldindx("mval","str.addr")=8
	;
	Set gtmtypes("mval_b")="struct"
	Set gtmtypes("mval_b",0)=9
	Set gtmtypes("mval_b","len")=32
	Set gtmtypes("mval_b",1,"name")="mval_b.mvtype"
	Set gtmtypes("mval_b",1,"off")=0
	Set gtmtypes("mval_b",1,"len")=2
	Set gtmtypes("mval_b",1,"type")="unsigned-short"
	Set gtmtypfldindx("mval_b","mvtype")=1
	Set gtmtypes("mval_b",2,"name")="mval_b.sgne"
	Set gtmtypes("mval_b",2,"off")=2
	Set gtmtypes("mval_b",2,"len")=1
	Set gtmtypes("mval_b",2,"type")="unsigned-char"
	Set gtmtypfldindx("mval_b","sgne")=2
	Set gtmtypes("mval_b",3,"name")="mval_b.fnpc_indx"
	Set gtmtypes("mval_b",3,"off")=3
	Set gtmtypes("mval_b",3,"len")=1
	Set gtmtypes("mval_b",3,"type")="unsigned-char"
	Set gtmtypfldindx("mval_b","fnpc_indx")=3
	Set gtmtypes("mval_b",4,"name")="mval_b.utfcgr_indx"
	Set gtmtypes("mval_b",4,"off")=4
	Set gtmtypes("mval_b",4,"len")=4
	Set gtmtypes("mval_b",4,"type")="unsigned-int"
	Set gtmtypfldindx("mval_b","utfcgr_indx")=4
	Set gtmtypes("mval_b",5,"name")="mval_b.m"
	Set gtmtypes("mval_b",5,"off")=8
	Set gtmtypes("mval_b",5,"len")=8
	Set gtmtypes("mval_b",5,"type")="int"
	Set gtmtypfldindx("mval_b","m")=5
	Set gtmtypes("mval_b",5,"dim")=2
	Set gtmtypes("mval_b",6,"name")="mval_b.str"
	Set gtmtypes("mval_b",6,"off")=16
	Set gtmtypes("mval_b",6,"len")=16
	Set gtmtypes("mval_b",6,"type")="mstr"
	Set gtmtypfldindx("mval_b","str")=6
	Set gtmtypes("mval_b",7,"name")="mval_b.str.char_len"
	Set gtmtypes("mval_b",7,"off")=16
	Set gtmtypes("mval_b",7,"len")=4
	Set gtmtypes("mval_b",7,"type")="unsigned-int"
	Set gtmtypfldindx("mval_b","str.char_len")=7
	Set gtmtypes("mval_b",8,"name")="mval_b.str.len"
	Set gtmtypes("mval_b",8,"off")=20
	Set gtmtypes("mval_b",8,"len")=4
	Set gtmtypes("mval_b",8,"type")="int"
	Set gtmtypfldindx("mval_b","str.len")=8
	Set gtmtypes("mval_b",9,"name")="mval_b.str.addr"
	Set gtmtypes("mval_b",9,"off")=24
	Set gtmtypes("mval_b",9,"len")=8
	Set gtmtypes("mval_b",9,"type")="addr"
	Set gtmtypfldindx("mval_b","str.addr")=9
	;
	Set gtmtypes("mvar")="struct"
	Set gtmtypes("mvar",0)=8
	Set gtmtypes("mvar","len")=48
	Set gtmtypes("mvar",1,"name")="mvar.lson"
	Set gtmtypes("mvar",1,"off")=0
	Set gtmtypes("mvar",1,"len")=8
	Set gtmtypes("mvar",1,"type")="addr"
	Set gtmtypfldindx("mvar","lson")=1
	Set gtmtypes("mvar",2,"name")="mvar.rson"
	Set gtmtypes("mvar",2,"off")=8
	Set gtmtypes("mvar",2,"len")=8
	Set gtmtypes("mvar",2,"type")="addr"
	Set gtmtypfldindx("mvar","rson")=2
	Set gtmtypes("mvar",3,"name")="mvar.mvidx"
	Set gtmtypes("mvar",3,"off")=16
	Set gtmtypes("mvar",3,"len")=4
	Set gtmtypes("mvar",3,"type")="int"
	Set gtmtypfldindx("mvar","mvidx")=3
	Set gtmtypes("mvar",4,"name")="mvar.mvname"
	Set gtmtypes("mvar",4,"off")=24
	Set gtmtypes("mvar",4,"len")=16
	Set gtmtypes("mvar",4,"type")="mstr"
	Set gtmtypfldindx("mvar","mvname")=4
	Set gtmtypes("mvar",5,"name")="mvar.mvname.char_len"
	Set gtmtypes("mvar",5,"off")=24
	Set gtmtypes("mvar",5,"len")=4
	Set gtmtypes("mvar",5,"type")="unsigned-int"
	Set gtmtypfldindx("mvar","mvname.char_len")=5
	Set gtmtypes("mvar",6,"name")="mvar.mvname.len"
	Set gtmtypes("mvar",6,"off")=28
	Set gtmtypes("mvar",6,"len")=4
	Set gtmtypes("mvar",6,"type")="int"
	Set gtmtypfldindx("mvar","mvname.len")=6
	Set gtmtypes("mvar",7,"name")="mvar.mvname.addr"
	Set gtmtypes("mvar",7,"off")=32
	Set gtmtypes("mvar",7,"len")=8
	Set gtmtypes("mvar",7,"type")="addr"
	Set gtmtypfldindx("mvar","mvname.addr")=7
	Set gtmtypes("mvar",8,"name")="mvar.last_fetch"
	Set gtmtypes("mvar",8,"off")=40
	Set gtmtypes("mvar",8,"len")=8
	Set gtmtypes("mvar",8,"type")="addr"
	Set gtmtypfldindx("mvar","last_fetch")=8
	;
	Set gtmtypes("mvax")="struct"
	Set gtmtypes("mvax",0)=4
	Set gtmtypes("mvax","len")=32
	Set gtmtypes("mvax",1,"name")="mvax.last"
	Set gtmtypes("mvax",1,"off")=0
	Set gtmtypes("mvax",1,"len")=8
	Set gtmtypes("mvax",1,"type")="addr"
	Set gtmtypfldindx("mvax","last")=1
	Set gtmtypes("mvax",2,"name")="mvax.next"
	Set gtmtypes("mvax",2,"off")=8
	Set gtmtypes("mvax",2,"len")=8
	Set gtmtypes("mvax",2,"type")="addr"
	Set gtmtypfldindx("mvax","next")=2
	Set gtmtypes("mvax",3,"name")="mvax.var"
	Set gtmtypes("mvax",3,"off")=16
	Set gtmtypes("mvax",3,"len")=8
	Set gtmtypes("mvax",3,"type")="addr"
	Set gtmtypfldindx("mvax","var")=3
	Set gtmtypes("mvax",4,"name")="mvax.mvidx"
	Set gtmtypes("mvax",4,"off")=24
	Set gtmtypes("mvax",4,"len")=4
	Set gtmtypes("mvax",4,"type")="int"
	Set gtmtypfldindx("mvax","mvidx")=4
	;
	Set gtmtypes("mvs_mrgzwrsv_struct")="struct"
	Set gtmtypes("mvs_mrgzwrsv_struct",0)=9
	Set gtmtypes("mvs_mrgzwrsv_struct","len")=56
	Set gtmtypes("mvs_mrgzwrsv_struct",1,"name")="mvs_mrgzwrsv_struct.save_merge_args"
	Set gtmtypes("mvs_mrgzwrsv_struct",1,"off")=0
	Set gtmtypes("mvs_mrgzwrsv_struct",1,"len")=4
	Set gtmtypes("mvs_mrgzwrsv_struct",1,"type")="int"
	Set gtmtypfldindx("mvs_mrgzwrsv_struct","save_merge_args")=1
	Set gtmtypes("mvs_mrgzwrsv_struct",2,"name")="mvs_mrgzwrsv_struct.save_zwrtacindx"
	Set gtmtypes("mvs_mrgzwrsv_struct",2,"off")=4
	Set gtmtypes("mvs_mrgzwrsv_struct",2,"len")=4
	Set gtmtypes("mvs_mrgzwrsv_struct",2,"type")="unsigned-int"
	Set gtmtypfldindx("mvs_mrgzwrsv_struct","save_zwrtacindx")=2
	Set gtmtypes("mvs_mrgzwrsv_struct",3,"name")="mvs_mrgzwrsv_struct.save_in_zwrite"
	Set gtmtypes("mvs_mrgzwrsv_struct",3,"off")=8
	Set gtmtypes("mvs_mrgzwrsv_struct",3,"len")=4
	Set gtmtypes("mvs_mrgzwrsv_struct",3,"type")="boolean_t"
	Set gtmtypfldindx("mvs_mrgzwrsv_struct","save_in_zwrite")=3
	Set gtmtypes("mvs_mrgzwrsv_struct",4,"name")="mvs_mrgzwrsv_struct.filler"
	Set gtmtypes("mvs_mrgzwrsv_struct",4,"off")=12
	Set gtmtypes("mvs_mrgzwrsv_struct",4,"len")=4
	Set gtmtypes("mvs_mrgzwrsv_struct",4,"type")="int"
	Set gtmtypfldindx("mvs_mrgzwrsv_struct","filler")=4
	Set gtmtypes("mvs_mrgzwrsv_struct",5,"name")="mvs_mrgzwrsv_struct.save_mglvnp"
	Set gtmtypes("mvs_mrgzwrsv_struct",5,"off")=16
	Set gtmtypes("mvs_mrgzwrsv_struct",5,"len")=8
	Set gtmtypes("mvs_mrgzwrsv_struct",5,"type")="addr"
	Set gtmtypfldindx("mvs_mrgzwrsv_struct","save_mglvnp")=5
	Set gtmtypes("mvs_mrgzwrsv_struct",6,"name")="mvs_mrgzwrsv_struct.save_gvzwrite_block"
	Set gtmtypes("mvs_mrgzwrsv_struct",6,"off")=24
	Set gtmtypes("mvs_mrgzwrsv_struct",6,"len")=8
	Set gtmtypes("mvs_mrgzwrsv_struct",6,"type")="addr"
	Set gtmtypfldindx("mvs_mrgzwrsv_struct","save_gvzwrite_block")=6
	Set gtmtypes("mvs_mrgzwrsv_struct",7,"name")="mvs_mrgzwrsv_struct.save_lvzwrite_block"
	Set gtmtypes("mvs_mrgzwrsv_struct",7,"off")=32
	Set gtmtypes("mvs_mrgzwrsv_struct",7,"len")=8
	Set gtmtypes("mvs_mrgzwrsv_struct",7,"type")="addr"
	Set gtmtypfldindx("mvs_mrgzwrsv_struct","save_lvzwrite_block")=7
	Set gtmtypes("mvs_mrgzwrsv_struct",8,"name")="mvs_mrgzwrsv_struct.save_zwr_output"
	Set gtmtypes("mvs_mrgzwrsv_struct",8,"off")=40
	Set gtmtypes("mvs_mrgzwrsv_struct",8,"len")=8
	Set gtmtypes("mvs_mrgzwrsv_struct",8,"type")="addr"
	Set gtmtypfldindx("mvs_mrgzwrsv_struct","save_zwr_output")=8
	Set gtmtypes("mvs_mrgzwrsv_struct",9,"name")="mvs_mrgzwrsv_struct.save_zwrhtab"
	Set gtmtypes("mvs_mrgzwrsv_struct",9,"off")=48
	Set gtmtypes("mvs_mrgzwrsv_struct",9,"len")=8
	Set gtmtypes("mvs_mrgzwrsv_struct",9,"type")="addr"
	Set gtmtypfldindx("mvs_mrgzwrsv_struct","save_zwrhtab")=9
	;
	Set gtmtypes("mvs_ntab_struct")="struct"
	Set gtmtypes("mvs_ntab_struct",0)=2
	Set gtmtypes("mvs_ntab_struct","len")=16
	Set gtmtypes("mvs_ntab_struct",1,"name")="mvs_ntab_struct.hte_addr"
	Set gtmtypes("mvs_ntab_struct",1,"off")=0
	Set gtmtypes("mvs_ntab_struct",1,"len")=8
	Set gtmtypes("mvs_ntab_struct",1,"type")="addr"
	Set gtmtypfldindx("mvs_ntab_struct","hte_addr")=1
	Set gtmtypes("mvs_ntab_struct",2,"name")="mvs_ntab_struct.save_value"
	Set gtmtypes("mvs_ntab_struct",2,"off")=8
	Set gtmtypes("mvs_ntab_struct",2,"len")=8
	Set gtmtypes("mvs_ntab_struct",2,"type")="addr"
	Set gtmtypfldindx("mvs_ntab_struct","save_value")=2
	;
	Set gtmtypes("mvs_nval_struct")="struct"
	Set gtmtypes("mvs_nval_struct",0)=4
	Set gtmtypes("mvs_nval_struct","len")=24
	Set gtmtypes("mvs_nval_struct",1,"name")="mvs_nval_struct.mvs_val"
	Set gtmtypes("mvs_nval_struct",1,"off")=0
	Set gtmtypes("mvs_nval_struct",1,"len")=8
	Set gtmtypes("mvs_nval_struct",1,"type")="addr"
	Set gtmtypfldindx("mvs_nval_struct","mvs_val")=1
	Set gtmtypes("mvs_nval_struct",2,"name")="mvs_nval_struct.mvs_ptab"
	Set gtmtypes("mvs_nval_struct",2,"off")=8
	Set gtmtypes("mvs_nval_struct",2,"len")=16
	Set gtmtypes("mvs_nval_struct",2,"type")="mvs_ntab_struct"
	Set gtmtypfldindx("mvs_nval_struct","mvs_ptab")=2
	Set gtmtypes("mvs_nval_struct",3,"name")="mvs_nval_struct.mvs_ptab.hte_addr"
	Set gtmtypes("mvs_nval_struct",3,"off")=8
	Set gtmtypes("mvs_nval_struct",3,"len")=8
	Set gtmtypes("mvs_nval_struct",3,"type")="addr"
	Set gtmtypfldindx("mvs_nval_struct","mvs_ptab.hte_addr")=3
	Set gtmtypes("mvs_nval_struct",4,"name")="mvs_nval_struct.mvs_ptab.save_value"
	Set gtmtypes("mvs_nval_struct",4,"off")=16
	Set gtmtypes("mvs_nval_struct",4,"len")=8
	Set gtmtypes("mvs_nval_struct",4,"type")="addr"
	Set gtmtypfldindx("mvs_nval_struct","mvs_ptab.save_value")=4
	;
	Set gtmtypes("mvs_pval_struct")="struct"
	Set gtmtypes("mvs_pval_struct",0)=4
	Set gtmtypes("mvs_pval_struct","len")=24
	Set gtmtypes("mvs_pval_struct",1,"name")="mvs_pval_struct.mvs_val"
	Set gtmtypes("mvs_pval_struct",1,"off")=0
	Set gtmtypes("mvs_pval_struct",1,"len")=8
	Set gtmtypes("mvs_pval_struct",1,"type")="addr"
	Set gtmtypfldindx("mvs_pval_struct","mvs_val")=1
	Set gtmtypes("mvs_pval_struct",2,"name")="mvs_pval_struct.mvs_ptab"
	Set gtmtypes("mvs_pval_struct",2,"off")=8
	Set gtmtypes("mvs_pval_struct",2,"len")=16
	Set gtmtypes("mvs_pval_struct",2,"type")="mvs_ntab_struct"
	Set gtmtypfldindx("mvs_pval_struct","mvs_ptab")=2
	Set gtmtypes("mvs_pval_struct",3,"name")="mvs_pval_struct.mvs_ptab.hte_addr"
	Set gtmtypes("mvs_pval_struct",3,"off")=8
	Set gtmtypes("mvs_pval_struct",3,"len")=8
	Set gtmtypes("mvs_pval_struct",3,"type")="addr"
	Set gtmtypfldindx("mvs_pval_struct","mvs_ptab.hte_addr")=3
	Set gtmtypes("mvs_pval_struct",4,"name")="mvs_pval_struct.mvs_ptab.save_value"
	Set gtmtypes("mvs_pval_struct",4,"off")=16
	Set gtmtypes("mvs_pval_struct",4,"len")=8
	Set gtmtypes("mvs_pval_struct",4,"type")="addr"
	Set gtmtypfldindx("mvs_pval_struct","mvs_ptab.save_value")=4
	;
	Set gtmtypes("mvs_rstrtpc_struct")="struct"
	Set gtmtypes("mvs_rstrtpc_struct",0)=2
	Set gtmtypes("mvs_rstrtpc_struct","len")=16
	Set gtmtypes("mvs_rstrtpc_struct",1,"name")="mvs_rstrtpc_struct.restart_pc_save"
	Set gtmtypes("mvs_rstrtpc_struct",1,"off")=0
	Set gtmtypes("mvs_rstrtpc_struct",1,"len")=8
	Set gtmtypes("mvs_rstrtpc_struct",1,"type")="addr"
	Set gtmtypfldindx("mvs_rstrtpc_struct","restart_pc_save")=1
	Set gtmtypes("mvs_rstrtpc_struct",2,"name")="mvs_rstrtpc_struct.restart_ctxt_save"
	Set gtmtypes("mvs_rstrtpc_struct",2,"off")=8
	Set gtmtypes("mvs_rstrtpc_struct",2,"len")=8
	Set gtmtypes("mvs_rstrtpc_struct",2,"type")="addr"
	Set gtmtypfldindx("mvs_rstrtpc_struct","restart_ctxt_save")=2
	;
	Set gtmtypes("mvs_tphold_struct")="struct"
	Set gtmtypes("mvs_tphold_struct",0)=10
	Set gtmtypes("mvs_tphold_struct","len")=40
	Set gtmtypes("mvs_tphold_struct",1,"name")="mvs_tphold_struct.tphold_tlevel"
	Set gtmtypes("mvs_tphold_struct",1,"off")=0
	Set gtmtypes("mvs_tphold_struct",1,"len")=4
	Set gtmtypes("mvs_tphold_struct",1,"type")="unsigned-int"
	Set gtmtypfldindx("mvs_tphold_struct","tphold_tlevel")=1
	Set gtmtypes("mvs_tphold_struct",2,"name")="mvs_tphold_struct.ztwormhole_save"
	Set gtmtypes("mvs_tphold_struct",2,"off")=8
	Set gtmtypes("mvs_tphold_struct",2,"len")=32
	Set gtmtypes("mvs_tphold_struct",2,"type")="mval"
	Set gtmtypfldindx("mvs_tphold_struct","ztwormhole_save")=2
	Set gtmtypes("mvs_tphold_struct",3,"name")="mvs_tphold_struct.ztwormhole_save.mvtype"
	Set gtmtypes("mvs_tphold_struct",3,"off")=8
	Set gtmtypes("mvs_tphold_struct",3,"len")=2
	Set gtmtypes("mvs_tphold_struct",3,"type")="unsigned-short"
	Set gtmtypfldindx("mvs_tphold_struct","ztwormhole_save.mvtype")=3
	Set gtmtypes("mvs_tphold_struct",4,"name")="mvs_tphold_struct.ztwormhole_save.fnpc_indx"
	Set gtmtypes("mvs_tphold_struct",4,"off")=11
	Set gtmtypes("mvs_tphold_struct",4,"len")=1
	Set gtmtypes("mvs_tphold_struct",4,"type")="unsigned-char"
	Set gtmtypfldindx("mvs_tphold_struct","ztwormhole_save.fnpc_indx")=4
	Set gtmtypes("mvs_tphold_struct",5,"name")="mvs_tphold_struct.ztwormhole_save.utfcgr_indx"
	Set gtmtypes("mvs_tphold_struct",5,"off")=12
	Set gtmtypes("mvs_tphold_struct",5,"len")=4
	Set gtmtypes("mvs_tphold_struct",5,"type")="unsigned-int"
	Set gtmtypfldindx("mvs_tphold_struct","ztwormhole_save.utfcgr_indx")=5
	Set gtmtypes("mvs_tphold_struct",6,"name")="mvs_tphold_struct.ztwormhole_save.m"
	Set gtmtypes("mvs_tphold_struct",6,"off")=16
	Set gtmtypes("mvs_tphold_struct",6,"len")=8
	Set gtmtypes("mvs_tphold_struct",6,"type")="int"
	Set gtmtypfldindx("mvs_tphold_struct","ztwormhole_save.m")=6
	Set gtmtypes("mvs_tphold_struct",6,"dim")=2
	Set gtmtypes("mvs_tphold_struct",7,"name")="mvs_tphold_struct.ztwormhole_save.str"
	Set gtmtypes("mvs_tphold_struct",7,"off")=24
	Set gtmtypes("mvs_tphold_struct",7,"len")=16
	Set gtmtypes("mvs_tphold_struct",7,"type")="mstr"
	Set gtmtypfldindx("mvs_tphold_struct","ztwormhole_save.str")=7
	Set gtmtypes("mvs_tphold_struct",8,"name")="mvs_tphold_struct.ztwormhole_save.str.char_len"
	Set gtmtypes("mvs_tphold_struct",8,"off")=24
	Set gtmtypes("mvs_tphold_struct",8,"len")=4
	Set gtmtypes("mvs_tphold_struct",8,"type")="unsigned-int"
	Set gtmtypfldindx("mvs_tphold_struct","ztwormhole_save.str.char_len")=8
	Set gtmtypes("mvs_tphold_struct",9,"name")="mvs_tphold_struct.ztwormhole_save.str.len"
	Set gtmtypes("mvs_tphold_struct",9,"off")=28
	Set gtmtypes("mvs_tphold_struct",9,"len")=4
	Set gtmtypes("mvs_tphold_struct",9,"type")="int"
	Set gtmtypfldindx("mvs_tphold_struct","ztwormhole_save.str.len")=9
	Set gtmtypes("mvs_tphold_struct",10,"name")="mvs_tphold_struct.ztwormhole_save.str.addr"
	Set gtmtypes("mvs_tphold_struct",10,"off")=32
	Set gtmtypes("mvs_tphold_struct",10,"len")=8
	Set gtmtypes("mvs_tphold_struct",10,"type")="addr"
	Set gtmtypfldindx("mvs_tphold_struct","ztwormhole_save.str.addr")=10
	;
	Set gtmtypes("mvs_trigr_struct")="struct"
	Set gtmtypes("mvs_trigr_struct",0)=46
	Set gtmtypes("mvs_trigr_struct","len")=216
	Set gtmtypes("mvs_trigr_struct",1,"name")="mvs_trigr_struct.saved_dollar_truth"
	Set gtmtypes("mvs_trigr_struct",1,"off")=0
	Set gtmtypes("mvs_trigr_struct",1,"len")=4
	Set gtmtypes("mvs_trigr_struct",1,"type")="boolean_t"
	Set gtmtypfldindx("mvs_trigr_struct","saved_dollar_truth")=1
	Set gtmtypes("mvs_trigr_struct",2,"name")="mvs_trigr_struct.filler"
	Set gtmtypes("mvs_trigr_struct",2,"off")=4
	Set gtmtypes("mvs_trigr_struct",2,"len")=4
	Set gtmtypes("mvs_trigr_struct",2,"type")="int"
	Set gtmtypfldindx("mvs_trigr_struct","filler")=2
	Set gtmtypes("mvs_trigr_struct",3,"name")="mvs_trigr_struct.savtarg"
	Set gtmtypes("mvs_trigr_struct",3,"off")=8
	Set gtmtypes("mvs_trigr_struct",3,"len")=32
	Set gtmtypes("mvs_trigr_struct",3,"type")="mval"
	Set gtmtypfldindx("mvs_trigr_struct","savtarg")=3
	Set gtmtypes("mvs_trigr_struct",4,"name")="mvs_trigr_struct.savtarg.mvtype"
	Set gtmtypes("mvs_trigr_struct",4,"off")=8
	Set gtmtypes("mvs_trigr_struct",4,"len")=2
	Set gtmtypes("mvs_trigr_struct",4,"type")="unsigned-short"
	Set gtmtypfldindx("mvs_trigr_struct","savtarg.mvtype")=4
	Set gtmtypes("mvs_trigr_struct",5,"name")="mvs_trigr_struct.savtarg.fnpc_indx"
	Set gtmtypes("mvs_trigr_struct",5,"off")=11
	Set gtmtypes("mvs_trigr_struct",5,"len")=1
	Set gtmtypes("mvs_trigr_struct",5,"type")="unsigned-char"
	Set gtmtypfldindx("mvs_trigr_struct","savtarg.fnpc_indx")=5
	Set gtmtypes("mvs_trigr_struct",6,"name")="mvs_trigr_struct.savtarg.utfcgr_indx"
	Set gtmtypes("mvs_trigr_struct",6,"off")=12
	Set gtmtypes("mvs_trigr_struct",6,"len")=4
	Set gtmtypes("mvs_trigr_struct",6,"type")="unsigned-int"
	Set gtmtypfldindx("mvs_trigr_struct","savtarg.utfcgr_indx")=6
	Set gtmtypes("mvs_trigr_struct",7,"name")="mvs_trigr_struct.savtarg.m"
	Set gtmtypes("mvs_trigr_struct",7,"off")=16
	Set gtmtypes("mvs_trigr_struct",7,"len")=8
	Set gtmtypes("mvs_trigr_struct",7,"type")="int"
	Set gtmtypfldindx("mvs_trigr_struct","savtarg.m")=7
	Set gtmtypes("mvs_trigr_struct",7,"dim")=2
	Set gtmtypes("mvs_trigr_struct",8,"name")="mvs_trigr_struct.savtarg.str"
	Set gtmtypes("mvs_trigr_struct",8,"off")=24
	Set gtmtypes("mvs_trigr_struct",8,"len")=16
	Set gtmtypes("mvs_trigr_struct",8,"type")="mstr"
	Set gtmtypfldindx("mvs_trigr_struct","savtarg.str")=8
	Set gtmtypes("mvs_trigr_struct",9,"name")="mvs_trigr_struct.savtarg.str.char_len"
	Set gtmtypes("mvs_trigr_struct",9,"off")=24
	Set gtmtypes("mvs_trigr_struct",9,"len")=4
	Set gtmtypes("mvs_trigr_struct",9,"type")="unsigned-int"
	Set gtmtypfldindx("mvs_trigr_struct","savtarg.str.char_len")=9
	Set gtmtypes("mvs_trigr_struct",10,"name")="mvs_trigr_struct.savtarg.str.len"
	Set gtmtypes("mvs_trigr_struct",10,"off")=28
	Set gtmtypes("mvs_trigr_struct",10,"len")=4
	Set gtmtypes("mvs_trigr_struct",10,"type")="int"
	Set gtmtypfldindx("mvs_trigr_struct","savtarg.str.len")=10
	Set gtmtypes("mvs_trigr_struct",11,"name")="mvs_trigr_struct.savtarg.str.addr"
	Set gtmtypes("mvs_trigr_struct",11,"off")=32
	Set gtmtypes("mvs_trigr_struct",11,"len")=8
	Set gtmtypes("mvs_trigr_struct",11,"type")="addr"
	Set gtmtypfldindx("mvs_trigr_struct","savtarg.str.addr")=11
	Set gtmtypes("mvs_trigr_struct",12,"name")="mvs_trigr_struct.savextref"
	Set gtmtypes("mvs_trigr_struct",12,"off")=40
	Set gtmtypes("mvs_trigr_struct",12,"len")=16
	Set gtmtypes("mvs_trigr_struct",12,"type")="mstr"
	Set gtmtypfldindx("mvs_trigr_struct","savextref")=12
	Set gtmtypes("mvs_trigr_struct",13,"name")="mvs_trigr_struct.savextref.char_len"
	Set gtmtypes("mvs_trigr_struct",13,"off")=40
	Set gtmtypes("mvs_trigr_struct",13,"len")=4
	Set gtmtypes("mvs_trigr_struct",13,"type")="unsigned-int"
	Set gtmtypfldindx("mvs_trigr_struct","savextref.char_len")=13
	Set gtmtypes("mvs_trigr_struct",14,"name")="mvs_trigr_struct.savextref.len"
	Set gtmtypes("mvs_trigr_struct",14,"off")=44
	Set gtmtypes("mvs_trigr_struct",14,"len")=4
	Set gtmtypes("mvs_trigr_struct",14,"type")="int"
	Set gtmtypfldindx("mvs_trigr_struct","savextref.len")=14
	Set gtmtypes("mvs_trigr_struct",15,"name")="mvs_trigr_struct.savextref.addr"
	Set gtmtypes("mvs_trigr_struct",15,"off")=48
	Set gtmtypes("mvs_trigr_struct",15,"len")=8
	Set gtmtypes("mvs_trigr_struct",15,"type")="addr"
	Set gtmtypfldindx("mvs_trigr_struct","savextref.addr")=15
	Set gtmtypes("mvs_trigr_struct",16,"name")="mvs_trigr_struct.ztvalue_changed_ptr"
	Set gtmtypes("mvs_trigr_struct",16,"off")=56
	Set gtmtypes("mvs_trigr_struct",16,"len")=8
	Set gtmtypes("mvs_trigr_struct",16,"type")="addr"
	Set gtmtypfldindx("mvs_trigr_struct","ztvalue_changed_ptr")=16
	Set gtmtypes("mvs_trigr_struct",17,"name")="mvs_trigr_struct.ztvalue_save"
	Set gtmtypes("mvs_trigr_struct",17,"off")=64
	Set gtmtypes("mvs_trigr_struct",17,"len")=8
	Set gtmtypes("mvs_trigr_struct",17,"type")="addr"
	Set gtmtypfldindx("mvs_trigr_struct","ztvalue_save")=17
	Set gtmtypes("mvs_trigr_struct",18,"name")="mvs_trigr_struct.ztname_save"
	Set gtmtypes("mvs_trigr_struct",18,"off")=72
	Set gtmtypes("mvs_trigr_struct",18,"len")=8
	Set gtmtypes("mvs_trigr_struct",18,"type")="addr"
	Set gtmtypfldindx("mvs_trigr_struct","ztname_save")=18
	Set gtmtypes("mvs_trigr_struct",19,"name")="mvs_trigr_struct.ztdata_save"
	Set gtmtypes("mvs_trigr_struct",19,"off")=80
	Set gtmtypes("mvs_trigr_struct",19,"len")=8
	Set gtmtypes("mvs_trigr_struct",19,"type")="addr"
	Set gtmtypfldindx("mvs_trigr_struct","ztdata_save")=19
	Set gtmtypes("mvs_trigr_struct",20,"name")="mvs_trigr_struct.ztdelim_save"
	Set gtmtypes("mvs_trigr_struct",20,"off")=88
	Set gtmtypes("mvs_trigr_struct",20,"len")=8
	Set gtmtypes("mvs_trigr_struct",20,"type")="addr"
	Set gtmtypfldindx("mvs_trigr_struct","ztdelim_save")=20
	Set gtmtypes("mvs_trigr_struct",21,"name")="mvs_trigr_struct.ztoldval_save"
	Set gtmtypes("mvs_trigr_struct",21,"off")=96
	Set gtmtypes("mvs_trigr_struct",21,"len")=8
	Set gtmtypes("mvs_trigr_struct",21,"type")="addr"
	Set gtmtypfldindx("mvs_trigr_struct","ztoldval_save")=21
	Set gtmtypes("mvs_trigr_struct",22,"name")="mvs_trigr_struct.ztriggerop_save"
	Set gtmtypes("mvs_trigr_struct",22,"off")=104
	Set gtmtypes("mvs_trigr_struct",22,"len")=8
	Set gtmtypes("mvs_trigr_struct",22,"type")="addr"
	Set gtmtypfldindx("mvs_trigr_struct","ztriggerop_save")=22
	Set gtmtypes("mvs_trigr_struct",23,"name")="mvs_trigr_struct.ztupdate_save"
	Set gtmtypes("mvs_trigr_struct",23,"off")=112
	Set gtmtypes("mvs_trigr_struct",23,"len")=8
	Set gtmtypes("mvs_trigr_struct",23,"type")="addr"
	Set gtmtypfldindx("mvs_trigr_struct","ztupdate_save")=23
	Set gtmtypes("mvs_trigr_struct",24,"name")="mvs_trigr_struct.ctxt_save"
	Set gtmtypes("mvs_trigr_struct",24,"off")=120
	Set gtmtypes("mvs_trigr_struct",24,"len")=8
	Set gtmtypes("mvs_trigr_struct",24,"type")="addr"
	Set gtmtypfldindx("mvs_trigr_struct","ctxt_save")=24
	Set gtmtypes("mvs_trigr_struct",25,"name")="mvs_trigr_struct.gtm_trigger_depth_save"
	Set gtmtypes("mvs_trigr_struct",25,"off")=128
	Set gtmtypes("mvs_trigr_struct",25,"len")=4
	Set gtmtypes("mvs_trigr_struct",25,"type")="int"
	Set gtmtypfldindx("mvs_trigr_struct","gtm_trigger_depth_save")=25
	Set gtmtypes("mvs_trigr_struct",26,"name")="mvs_trigr_struct.dollar_etrap_save"
	Set gtmtypes("mvs_trigr_struct",26,"off")=136
	Set gtmtypes("mvs_trigr_struct",26,"len")=32
	Set gtmtypes("mvs_trigr_struct",26,"type")="mval"
	Set gtmtypfldindx("mvs_trigr_struct","dollar_etrap_save")=26
	Set gtmtypes("mvs_trigr_struct",27,"name")="mvs_trigr_struct.dollar_etrap_save.mvtype"
	Set gtmtypes("mvs_trigr_struct",27,"off")=136
	Set gtmtypes("mvs_trigr_struct",27,"len")=2
	Set gtmtypes("mvs_trigr_struct",27,"type")="unsigned-short"
	Set gtmtypfldindx("mvs_trigr_struct","dollar_etrap_save.mvtype")=27
	Set gtmtypes("mvs_trigr_struct",28,"name")="mvs_trigr_struct.dollar_etrap_save.fnpc_indx"
	Set gtmtypes("mvs_trigr_struct",28,"off")=139
	Set gtmtypes("mvs_trigr_struct",28,"len")=1
	Set gtmtypes("mvs_trigr_struct",28,"type")="unsigned-char"
	Set gtmtypfldindx("mvs_trigr_struct","dollar_etrap_save.fnpc_indx")=28
	Set gtmtypes("mvs_trigr_struct",29,"name")="mvs_trigr_struct.dollar_etrap_save.utfcgr_indx"
	Set gtmtypes("mvs_trigr_struct",29,"off")=140
	Set gtmtypes("mvs_trigr_struct",29,"len")=4
	Set gtmtypes("mvs_trigr_struct",29,"type")="unsigned-int"
	Set gtmtypfldindx("mvs_trigr_struct","dollar_etrap_save.utfcgr_indx")=29
	Set gtmtypes("mvs_trigr_struct",30,"name")="mvs_trigr_struct.dollar_etrap_save.m"
	Set gtmtypes("mvs_trigr_struct",30,"off")=144
	Set gtmtypes("mvs_trigr_struct",30,"len")=8
	Set gtmtypes("mvs_trigr_struct",30,"type")="int"
	Set gtmtypfldindx("mvs_trigr_struct","dollar_etrap_save.m")=30
	Set gtmtypes("mvs_trigr_struct",30,"dim")=2
	Set gtmtypes("mvs_trigr_struct",31,"name")="mvs_trigr_struct.dollar_etrap_save.str"
	Set gtmtypes("mvs_trigr_struct",31,"off")=152
	Set gtmtypes("mvs_trigr_struct",31,"len")=16
	Set gtmtypes("mvs_trigr_struct",31,"type")="mstr"
	Set gtmtypfldindx("mvs_trigr_struct","dollar_etrap_save.str")=31
	Set gtmtypes("mvs_trigr_struct",32,"name")="mvs_trigr_struct.dollar_etrap_save.str.char_len"
	Set gtmtypes("mvs_trigr_struct",32,"off")=152
	Set gtmtypes("mvs_trigr_struct",32,"len")=4
	Set gtmtypes("mvs_trigr_struct",32,"type")="unsigned-int"
	Set gtmtypfldindx("mvs_trigr_struct","dollar_etrap_save.str.char_len")=32
	Set gtmtypes("mvs_trigr_struct",33,"name")="mvs_trigr_struct.dollar_etrap_save.str.len"
	Set gtmtypes("mvs_trigr_struct",33,"off")=156
	Set gtmtypes("mvs_trigr_struct",33,"len")=4
	Set gtmtypes("mvs_trigr_struct",33,"type")="int"
	Set gtmtypfldindx("mvs_trigr_struct","dollar_etrap_save.str.len")=33
	Set gtmtypes("mvs_trigr_struct",34,"name")="mvs_trigr_struct.dollar_etrap_save.str.addr"
	Set gtmtypes("mvs_trigr_struct",34,"off")=160
	Set gtmtypes("mvs_trigr_struct",34,"len")=8
	Set gtmtypes("mvs_trigr_struct",34,"type")="addr"
	Set gtmtypfldindx("mvs_trigr_struct","dollar_etrap_save.str.addr")=34
	Set gtmtypes("mvs_trigr_struct",35,"name")="mvs_trigr_struct.dollar_ztrap_save"
	Set gtmtypes("mvs_trigr_struct",35,"off")=168
	Set gtmtypes("mvs_trigr_struct",35,"len")=32
	Set gtmtypes("mvs_trigr_struct",35,"type")="mval"
	Set gtmtypfldindx("mvs_trigr_struct","dollar_ztrap_save")=35
	Set gtmtypes("mvs_trigr_struct",36,"name")="mvs_trigr_struct.dollar_ztrap_save.mvtype"
	Set gtmtypes("mvs_trigr_struct",36,"off")=168
	Set gtmtypes("mvs_trigr_struct",36,"len")=2
	Set gtmtypes("mvs_trigr_struct",36,"type")="unsigned-short"
	Set gtmtypfldindx("mvs_trigr_struct","dollar_ztrap_save.mvtype")=36
	Set gtmtypes("mvs_trigr_struct",37,"name")="mvs_trigr_struct.dollar_ztrap_save.fnpc_indx"
	Set gtmtypes("mvs_trigr_struct",37,"off")=171
	Set gtmtypes("mvs_trigr_struct",37,"len")=1
	Set gtmtypes("mvs_trigr_struct",37,"type")="unsigned-char"
	Set gtmtypfldindx("mvs_trigr_struct","dollar_ztrap_save.fnpc_indx")=37
	Set gtmtypes("mvs_trigr_struct",38,"name")="mvs_trigr_struct.dollar_ztrap_save.utfcgr_indx"
	Set gtmtypes("mvs_trigr_struct",38,"off")=172
	Set gtmtypes("mvs_trigr_struct",38,"len")=4
	Set gtmtypes("mvs_trigr_struct",38,"type")="unsigned-int"
	Set gtmtypfldindx("mvs_trigr_struct","dollar_ztrap_save.utfcgr_indx")=38
	Set gtmtypes("mvs_trigr_struct",39,"name")="mvs_trigr_struct.dollar_ztrap_save.m"
	Set gtmtypes("mvs_trigr_struct",39,"off")=176
	Set gtmtypes("mvs_trigr_struct",39,"len")=8
	Set gtmtypes("mvs_trigr_struct",39,"type")="int"
	Set gtmtypfldindx("mvs_trigr_struct","dollar_ztrap_save.m")=39
	Set gtmtypes("mvs_trigr_struct",39,"dim")=2
	Set gtmtypes("mvs_trigr_struct",40,"name")="mvs_trigr_struct.dollar_ztrap_save.str"
	Set gtmtypes("mvs_trigr_struct",40,"off")=184
	Set gtmtypes("mvs_trigr_struct",40,"len")=16
	Set gtmtypes("mvs_trigr_struct",40,"type")="mstr"
	Set gtmtypfldindx("mvs_trigr_struct","dollar_ztrap_save.str")=40
	Set gtmtypes("mvs_trigr_struct",41,"name")="mvs_trigr_struct.dollar_ztrap_save.str.char_len"
	Set gtmtypes("mvs_trigr_struct",41,"off")=184
	Set gtmtypes("mvs_trigr_struct",41,"len")=4
	Set gtmtypes("mvs_trigr_struct",41,"type")="unsigned-int"
	Set gtmtypfldindx("mvs_trigr_struct","dollar_ztrap_save.str.char_len")=41
	Set gtmtypes("mvs_trigr_struct",42,"name")="mvs_trigr_struct.dollar_ztrap_save.str.len"
	Set gtmtypes("mvs_trigr_struct",42,"off")=188
	Set gtmtypes("mvs_trigr_struct",42,"len")=4
	Set gtmtypes("mvs_trigr_struct",42,"type")="int"
	Set gtmtypfldindx("mvs_trigr_struct","dollar_ztrap_save.str.len")=42
	Set gtmtypes("mvs_trigr_struct",43,"name")="mvs_trigr_struct.dollar_ztrap_save.str.addr"
	Set gtmtypes("mvs_trigr_struct",43,"off")=192
	Set gtmtypes("mvs_trigr_struct",43,"len")=8
	Set gtmtypes("mvs_trigr_struct",43,"type")="addr"
	Set gtmtypfldindx("mvs_trigr_struct","dollar_ztrap_save.str.addr")=43
	Set gtmtypes("mvs_trigr_struct",44,"name")="mvs_trigr_struct.ztrap_explicit_null_save"
	Set gtmtypes("mvs_trigr_struct",44,"off")=200
	Set gtmtypes("mvs_trigr_struct",44,"len")=4
	Set gtmtypes("mvs_trigr_struct",44,"type")="boolean_t"
	Set gtmtypfldindx("mvs_trigr_struct","ztrap_explicit_null_save")=44
	Set gtmtypes("mvs_trigr_struct",45,"name")="mvs_trigr_struct.mumps_status_save"
	Set gtmtypes("mvs_trigr_struct",45,"off")=204
	Set gtmtypes("mvs_trigr_struct",45,"len")=4
	Set gtmtypes("mvs_trigr_struct",45,"type")="int"
	Set gtmtypfldindx("mvs_trigr_struct","mumps_status_save")=45
	Set gtmtypes("mvs_trigr_struct",46,"name")="mvs_trigr_struct.run_time_save"
	Set gtmtypes("mvs_trigr_struct",46,"off")=208
	Set gtmtypes("mvs_trigr_struct",46,"len")=4
	Set gtmtypes("mvs_trigr_struct",46,"type")="boolean_t"
	Set gtmtypfldindx("mvs_trigr_struct","run_time_save")=46
	;
	Set gtmtypes("mvs_zintcmd_struct")="struct"
	Set gtmtypes("mvs_zintcmd_struct",0)=8
	Set gtmtypes("mvs_zintcmd_struct","len")=56
	Set gtmtypes("mvs_zintcmd_struct",1,"name")="mvs_zintcmd_struct.command"
	Set gtmtypes("mvs_zintcmd_struct",1,"off")=0
	Set gtmtypes("mvs_zintcmd_struct",1,"len")=4
	Set gtmtypes("mvs_zintcmd_struct",1,"type")="int"
	Set gtmtypfldindx("mvs_zintcmd_struct","command")=1
	Set gtmtypes("mvs_zintcmd_struct",2,"name")="mvs_zintcmd_struct.end_or_remain"
	Set gtmtypes("mvs_zintcmd_struct",2,"off")=8
	Set gtmtypes("mvs_zintcmd_struct",2,"len")=16
	Set gtmtypes("mvs_zintcmd_struct",2,"type")="ABS_TIME"
	Set gtmtypfldindx("mvs_zintcmd_struct","end_or_remain")=2
	Set gtmtypes("mvs_zintcmd_struct",3,"name")="mvs_zintcmd_struct.end_or_remain.at_sec"
	Set gtmtypes("mvs_zintcmd_struct",3,"off")=8
	Set gtmtypes("mvs_zintcmd_struct",3,"len")=8
	Set gtmtypes("mvs_zintcmd_struct",3,"type")="long"
	Set gtmtypfldindx("mvs_zintcmd_struct","end_or_remain.at_sec")=3
	Set gtmtypes("mvs_zintcmd_struct",4,"name")="mvs_zintcmd_struct.end_or_remain.at_usec"
	Set gtmtypes("mvs_zintcmd_struct",4,"off")=16
	Set gtmtypes("mvs_zintcmd_struct",4,"len")=8
	Set gtmtypes("mvs_zintcmd_struct",4,"type")="long"
	Set gtmtypfldindx("mvs_zintcmd_struct","end_or_remain.at_usec")=4
	Set gtmtypes("mvs_zintcmd_struct",5,"name")="mvs_zintcmd_struct.restart_pc_check"
	Set gtmtypes("mvs_zintcmd_struct",5,"off")=24
	Set gtmtypes("mvs_zintcmd_struct",5,"len")=8
	Set gtmtypes("mvs_zintcmd_struct",5,"type")="addr"
	Set gtmtypfldindx("mvs_zintcmd_struct","restart_pc_check")=5
	Set gtmtypes("mvs_zintcmd_struct",6,"name")="mvs_zintcmd_struct.restart_ctxt_check"
	Set gtmtypes("mvs_zintcmd_struct",6,"off")=32
	Set gtmtypes("mvs_zintcmd_struct",6,"len")=8
	Set gtmtypes("mvs_zintcmd_struct",6,"type")="addr"
	Set gtmtypfldindx("mvs_zintcmd_struct","restart_ctxt_check")=6
	Set gtmtypes("mvs_zintcmd_struct",7,"name")="mvs_zintcmd_struct.restart_pc_prior"
	Set gtmtypes("mvs_zintcmd_struct",7,"off")=40
	Set gtmtypes("mvs_zintcmd_struct",7,"len")=8
	Set gtmtypes("mvs_zintcmd_struct",7,"type")="addr"
	Set gtmtypfldindx("mvs_zintcmd_struct","restart_pc_prior")=7
	Set gtmtypes("mvs_zintcmd_struct",8,"name")="mvs_zintcmd_struct.restart_ctxt_prior"
	Set gtmtypes("mvs_zintcmd_struct",8,"off")=48
	Set gtmtypes("mvs_zintcmd_struct",8,"len")=8
	Set gtmtypes("mvs_zintcmd_struct",8,"type")="addr"
	Set gtmtypfldindx("mvs_zintcmd_struct","restart_ctxt_prior")=8
	;
	Set gtmtypes("mvs_zintdev_struct")="struct"
	Set gtmtypes("mvs_zintdev_struct",0)=7
	Set gtmtypes("mvs_zintdev_struct","len")=40
	Set gtmtypes("mvs_zintdev_struct",1,"name")="mvs_zintdev_struct.io_ptr"
	Set gtmtypes("mvs_zintdev_struct",1,"off")=0
	Set gtmtypes("mvs_zintdev_struct",1,"len")=8
	Set gtmtypes("mvs_zintdev_struct",1,"type")="addr"
	Set gtmtypfldindx("mvs_zintdev_struct","io_ptr")=1
	Set gtmtypes("mvs_zintdev_struct",2,"name")="mvs_zintdev_struct.buffer_valid"
	Set gtmtypes("mvs_zintdev_struct",2,"off")=8
	Set gtmtypes("mvs_zintdev_struct",2,"len")=4
	Set gtmtypes("mvs_zintdev_struct",2,"type")="boolean_t"
	Set gtmtypfldindx("mvs_zintdev_struct","buffer_valid")=2
	Set gtmtypes("mvs_zintdev_struct",3,"name")="mvs_zintdev_struct.curr_sp_buffer"
	Set gtmtypes("mvs_zintdev_struct",3,"off")=16
	Set gtmtypes("mvs_zintdev_struct",3,"len")=16
	Set gtmtypes("mvs_zintdev_struct",3,"type")="mstr"
	Set gtmtypfldindx("mvs_zintdev_struct","curr_sp_buffer")=3
	Set gtmtypes("mvs_zintdev_struct",4,"name")="mvs_zintdev_struct.curr_sp_buffer.char_len"
	Set gtmtypes("mvs_zintdev_struct",4,"off")=16
	Set gtmtypes("mvs_zintdev_struct",4,"len")=4
	Set gtmtypes("mvs_zintdev_struct",4,"type")="unsigned-int"
	Set gtmtypfldindx("mvs_zintdev_struct","curr_sp_buffer.char_len")=4
	Set gtmtypes("mvs_zintdev_struct",5,"name")="mvs_zintdev_struct.curr_sp_buffer.len"
	Set gtmtypes("mvs_zintdev_struct",5,"off")=20
	Set gtmtypes("mvs_zintdev_struct",5,"len")=4
	Set gtmtypes("mvs_zintdev_struct",5,"type")="int"
	Set gtmtypfldindx("mvs_zintdev_struct","curr_sp_buffer.len")=5
	Set gtmtypes("mvs_zintdev_struct",6,"name")="mvs_zintdev_struct.curr_sp_buffer.addr"
	Set gtmtypes("mvs_zintdev_struct",6,"off")=24
	Set gtmtypes("mvs_zintdev_struct",6,"len")=8
	Set gtmtypes("mvs_zintdev_struct",6,"type")="addr"
	Set gtmtypfldindx("mvs_zintdev_struct","curr_sp_buffer.addr")=6
	Set gtmtypes("mvs_zintdev_struct",7,"name")="mvs_zintdev_struct.socketptr"
	Set gtmtypes("mvs_zintdev_struct",7,"off")=32
	Set gtmtypes("mvs_zintdev_struct",7,"len")=8
	Set gtmtypes("mvs_zintdev_struct",7,"type")="addr"
	Set gtmtypfldindx("mvs_zintdev_struct","socketptr")=7
	;
	Set gtmtypes("mvs_zintr_struct")="struct"
	Set gtmtypes("mvs_zintr_struct",0)=35
	Set gtmtypes("mvs_zintr_struct","len")=184
	Set gtmtypes("mvs_zintr_struct",1,"name")="mvs_zintr_struct.saved_dollar_truth"
	Set gtmtypes("mvs_zintr_struct",1,"off")=0
	Set gtmtypes("mvs_zintr_struct",1,"len")=4
	Set gtmtypes("mvs_zintr_struct",1,"type")="boolean_t"
	Set gtmtypfldindx("mvs_zintr_struct","saved_dollar_truth")=1
	Set gtmtypes("mvs_zintr_struct",2,"name")="mvs_zintr_struct.filler"
	Set gtmtypes("mvs_zintr_struct",2,"off")=4
	Set gtmtypes("mvs_zintr_struct",2,"len")=4
	Set gtmtypes("mvs_zintr_struct",2,"type")="int"
	Set gtmtypfldindx("mvs_zintr_struct","filler")=2
	Set gtmtypes("mvs_zintr_struct",3,"name")="mvs_zintr_struct.savtarg"
	Set gtmtypes("mvs_zintr_struct",3,"off")=8
	Set gtmtypes("mvs_zintr_struct",3,"len")=32
	Set gtmtypes("mvs_zintr_struct",3,"type")="mval"
	Set gtmtypfldindx("mvs_zintr_struct","savtarg")=3
	Set gtmtypes("mvs_zintr_struct",4,"name")="mvs_zintr_struct.savtarg.mvtype"
	Set gtmtypes("mvs_zintr_struct",4,"off")=8
	Set gtmtypes("mvs_zintr_struct",4,"len")=2
	Set gtmtypes("mvs_zintr_struct",4,"type")="unsigned-short"
	Set gtmtypfldindx("mvs_zintr_struct","savtarg.mvtype")=4
	Set gtmtypes("mvs_zintr_struct",5,"name")="mvs_zintr_struct.savtarg.fnpc_indx"
	Set gtmtypes("mvs_zintr_struct",5,"off")=11
	Set gtmtypes("mvs_zintr_struct",5,"len")=1
	Set gtmtypes("mvs_zintr_struct",5,"type")="unsigned-char"
	Set gtmtypfldindx("mvs_zintr_struct","savtarg.fnpc_indx")=5
	Set gtmtypes("mvs_zintr_struct",6,"name")="mvs_zintr_struct.savtarg.utfcgr_indx"
	Set gtmtypes("mvs_zintr_struct",6,"off")=12
	Set gtmtypes("mvs_zintr_struct",6,"len")=4
	Set gtmtypes("mvs_zintr_struct",6,"type")="unsigned-int"
	Set gtmtypfldindx("mvs_zintr_struct","savtarg.utfcgr_indx")=6
	Set gtmtypes("mvs_zintr_struct",7,"name")="mvs_zintr_struct.savtarg.m"
	Set gtmtypes("mvs_zintr_struct",7,"off")=16
	Set gtmtypes("mvs_zintr_struct",7,"len")=8
	Set gtmtypes("mvs_zintr_struct",7,"type")="int"
	Set gtmtypfldindx("mvs_zintr_struct","savtarg.m")=7
	Set gtmtypes("mvs_zintr_struct",7,"dim")=2
	Set gtmtypes("mvs_zintr_struct",8,"name")="mvs_zintr_struct.savtarg.str"
	Set gtmtypes("mvs_zintr_struct",8,"off")=24
	Set gtmtypes("mvs_zintr_struct",8,"len")=16
	Set gtmtypes("mvs_zintr_struct",8,"type")="mstr"
	Set gtmtypfldindx("mvs_zintr_struct","savtarg.str")=8
	Set gtmtypes("mvs_zintr_struct",9,"name")="mvs_zintr_struct.savtarg.str.char_len"
	Set gtmtypes("mvs_zintr_struct",9,"off")=24
	Set gtmtypes("mvs_zintr_struct",9,"len")=4
	Set gtmtypes("mvs_zintr_struct",9,"type")="unsigned-int"
	Set gtmtypfldindx("mvs_zintr_struct","savtarg.str.char_len")=9
	Set gtmtypes("mvs_zintr_struct",10,"name")="mvs_zintr_struct.savtarg.str.len"
	Set gtmtypes("mvs_zintr_struct",10,"off")=28
	Set gtmtypes("mvs_zintr_struct",10,"len")=4
	Set gtmtypes("mvs_zintr_struct",10,"type")="int"
	Set gtmtypfldindx("mvs_zintr_struct","savtarg.str.len")=10
	Set gtmtypes("mvs_zintr_struct",11,"name")="mvs_zintr_struct.savtarg.str.addr"
	Set gtmtypes("mvs_zintr_struct",11,"off")=32
	Set gtmtypes("mvs_zintr_struct",11,"len")=8
	Set gtmtypes("mvs_zintr_struct",11,"type")="addr"
	Set gtmtypfldindx("mvs_zintr_struct","savtarg.str.addr")=11
	Set gtmtypes("mvs_zintr_struct",12,"name")="mvs_zintr_struct.savextref"
	Set gtmtypes("mvs_zintr_struct",12,"off")=40
	Set gtmtypes("mvs_zintr_struct",12,"len")=16
	Set gtmtypes("mvs_zintr_struct",12,"type")="mstr"
	Set gtmtypfldindx("mvs_zintr_struct","savextref")=12
	Set gtmtypes("mvs_zintr_struct",13,"name")="mvs_zintr_struct.savextref.char_len"
	Set gtmtypes("mvs_zintr_struct",13,"off")=40
	Set gtmtypes("mvs_zintr_struct",13,"len")=4
	Set gtmtypes("mvs_zintr_struct",13,"type")="unsigned-int"
	Set gtmtypfldindx("mvs_zintr_struct","savextref.char_len")=13
	Set gtmtypes("mvs_zintr_struct",14,"name")="mvs_zintr_struct.savextref.len"
	Set gtmtypes("mvs_zintr_struct",14,"off")=44
	Set gtmtypes("mvs_zintr_struct",14,"len")=4
	Set gtmtypes("mvs_zintr_struct",14,"type")="int"
	Set gtmtypfldindx("mvs_zintr_struct","savextref.len")=14
	Set gtmtypes("mvs_zintr_struct",15,"name")="mvs_zintr_struct.savextref.addr"
	Set gtmtypes("mvs_zintr_struct",15,"off")=48
	Set gtmtypes("mvs_zintr_struct",15,"len")=8
	Set gtmtypes("mvs_zintr_struct",15,"type")="addr"
	Set gtmtypfldindx("mvs_zintr_struct","savextref.addr")=15
	Set gtmtypes("mvs_zintr_struct",16,"name")="mvs_zintr_struct.error_frame_save"
	Set gtmtypes("mvs_zintr_struct",16,"off")=56
	Set gtmtypes("mvs_zintr_struct",16,"len")=8
	Set gtmtypes("mvs_zintr_struct",16,"type")="addr"
	Set gtmtypfldindx("mvs_zintr_struct","error_frame_save")=16
	Set gtmtypes("mvs_zintr_struct",17,"name")="mvs_zintr_struct.dollar_ecode_save"
	Set gtmtypes("mvs_zintr_struct",17,"off")=64
	Set gtmtypes("mvs_zintr_struct",17,"len")=80
	Set gtmtypes("mvs_zintr_struct",17,"type")="dollar_ecode_type"
	Set gtmtypfldindx("mvs_zintr_struct","dollar_ecode_save")=17
	Set gtmtypes("mvs_zintr_struct",18,"name")="mvs_zintr_struct.dollar_ecode_save.begin"
	Set gtmtypes("mvs_zintr_struct",18,"off")=64
	Set gtmtypes("mvs_zintr_struct",18,"len")=8
	Set gtmtypes("mvs_zintr_struct",18,"type")="addr"
	Set gtmtypfldindx("mvs_zintr_struct","dollar_ecode_save.begin")=18
	Set gtmtypes("mvs_zintr_struct",19,"name")="mvs_zintr_struct.dollar_ecode_save.end"
	Set gtmtypes("mvs_zintr_struct",19,"off")=72
	Set gtmtypes("mvs_zintr_struct",19,"len")=8
	Set gtmtypes("mvs_zintr_struct",19,"type")="addr"
	Set gtmtypfldindx("mvs_zintr_struct","dollar_ecode_save.end")=19
	Set gtmtypes("mvs_zintr_struct",20,"name")="mvs_zintr_struct.dollar_ecode_save.top"
	Set gtmtypes("mvs_zintr_struct",20,"off")=80
	Set gtmtypes("mvs_zintr_struct",20,"len")=8
	Set gtmtypes("mvs_zintr_struct",20,"type")="addr"
	Set gtmtypfldindx("mvs_zintr_struct","dollar_ecode_save.top")=20
	Set gtmtypes("mvs_zintr_struct",21,"name")="mvs_zintr_struct.dollar_ecode_save.array"
	Set gtmtypes("mvs_zintr_struct",21,"off")=88
	Set gtmtypes("mvs_zintr_struct",21,"len")=8
	Set gtmtypes("mvs_zintr_struct",21,"type")="addr"
	Set gtmtypfldindx("mvs_zintr_struct","dollar_ecode_save.array")=21
	Set gtmtypes("mvs_zintr_struct",22,"name")="mvs_zintr_struct.dollar_ecode_save.index"
	Set gtmtypes("mvs_zintr_struct",22,"off")=96
	Set gtmtypes("mvs_zintr_struct",22,"len")=4
	Set gtmtypes("mvs_zintr_struct",22,"type")="unsigned-int"
	Set gtmtypfldindx("mvs_zintr_struct","dollar_ecode_save.index")=22
	Set gtmtypes("mvs_zintr_struct",23,"name")="mvs_zintr_struct.dollar_ecode_save.error_last_ecode"
	Set gtmtypes("mvs_zintr_struct",23,"off")=100
	Set gtmtypes("mvs_zintr_struct",23,"len")=4
	Set gtmtypes("mvs_zintr_struct",23,"type")="int"
	Set gtmtypfldindx("mvs_zintr_struct","dollar_ecode_save.error_last_ecode")=23
	Set gtmtypes("mvs_zintr_struct",24,"name")="mvs_zintr_struct.dollar_ecode_save.error_last_b_line"
	Set gtmtypes("mvs_zintr_struct",24,"off")=104
	Set gtmtypes("mvs_zintr_struct",24,"len")=8
	Set gtmtypes("mvs_zintr_struct",24,"type")="addr"
	Set gtmtypfldindx("mvs_zintr_struct","dollar_ecode_save.error_last_b_line")=24
	Set gtmtypes("mvs_zintr_struct",25,"name")="mvs_zintr_struct.dollar_ecode_save.first_ecode_error_frame"
	Set gtmtypes("mvs_zintr_struct",25,"off")=112
	Set gtmtypes("mvs_zintr_struct",25,"len")=8
	Set gtmtypes("mvs_zintr_struct",25,"type")="addr"
	Set gtmtypfldindx("mvs_zintr_struct","dollar_ecode_save.first_ecode_error_frame")=25
	Set gtmtypes("mvs_zintr_struct",26,"name")="mvs_zintr_struct.dollar_ecode_save.error_rtn_addr"
	Set gtmtypes("mvs_zintr_struct",26,"off")=120
	Set gtmtypes("mvs_zintr_struct",26,"len")=8
	Set gtmtypes("mvs_zintr_struct",26,"type")="addr"
	Set gtmtypfldindx("mvs_zintr_struct","dollar_ecode_save.error_rtn_addr")=26
	Set gtmtypes("mvs_zintr_struct",27,"name")="mvs_zintr_struct.dollar_ecode_save.error_rtn_ctxt"
	Set gtmtypes("mvs_zintr_struct",27,"off")=128
	Set gtmtypes("mvs_zintr_struct",27,"len")=8
	Set gtmtypes("mvs_zintr_struct",27,"type")="addr"
	Set gtmtypfldindx("mvs_zintr_struct","dollar_ecode_save.error_rtn_ctxt")=27
	Set gtmtypes("mvs_zintr_struct",28,"name")="mvs_zintr_struct.dollar_ecode_save.error_return_addr"
	Set gtmtypes("mvs_zintr_struct",28,"off")=136
	Set gtmtypes("mvs_zintr_struct",28,"len")=8
	Set gtmtypes("mvs_zintr_struct",28,"type")="void"
	Set gtmtypfldindx("mvs_zintr_struct","dollar_ecode_save.error_return_addr")=28
	Set gtmtypes("mvs_zintr_struct",29,"name")="mvs_zintr_struct.dollar_stack_save"
	Set gtmtypes("mvs_zintr_struct",29,"off")=144
	Set gtmtypes("mvs_zintr_struct",29,"len")=40
	Set gtmtypes("mvs_zintr_struct",29,"type")="dollar_stack_type"
	Set gtmtypfldindx("mvs_zintr_struct","dollar_stack_save")=29
	Set gtmtypes("mvs_zintr_struct",30,"name")="mvs_zintr_struct.dollar_stack_save.begin"
	Set gtmtypes("mvs_zintr_struct",30,"off")=144
	Set gtmtypes("mvs_zintr_struct",30,"len")=8
	Set gtmtypes("mvs_zintr_struct",30,"type")="addr"
	Set gtmtypfldindx("mvs_zintr_struct","dollar_stack_save.begin")=30
	Set gtmtypes("mvs_zintr_struct",31,"name")="mvs_zintr_struct.dollar_stack_save.end"
	Set gtmtypes("mvs_zintr_struct",31,"off")=152
	Set gtmtypes("mvs_zintr_struct",31,"len")=8
	Set gtmtypes("mvs_zintr_struct",31,"type")="addr"
	Set gtmtypfldindx("mvs_zintr_struct","dollar_stack_save.end")=31
	Set gtmtypes("mvs_zintr_struct",32,"name")="mvs_zintr_struct.dollar_stack_save.top"
	Set gtmtypes("mvs_zintr_struct",32,"off")=160
	Set gtmtypes("mvs_zintr_struct",32,"len")=8
	Set gtmtypes("mvs_zintr_struct",32,"type")="addr"
	Set gtmtypfldindx("mvs_zintr_struct","dollar_stack_save.top")=32
	Set gtmtypes("mvs_zintr_struct",33,"name")="mvs_zintr_struct.dollar_stack_save.array"
	Set gtmtypes("mvs_zintr_struct",33,"off")=168
	Set gtmtypes("mvs_zintr_struct",33,"len")=8
	Set gtmtypes("mvs_zintr_struct",33,"type")="addr"
	Set gtmtypfldindx("mvs_zintr_struct","dollar_stack_save.array")=33
	Set gtmtypes("mvs_zintr_struct",34,"name")="mvs_zintr_struct.dollar_stack_save.index"
	Set gtmtypes("mvs_zintr_struct",34,"off")=176
	Set gtmtypes("mvs_zintr_struct",34,"len")=4
	Set gtmtypes("mvs_zintr_struct",34,"type")="unsigned-int"
	Set gtmtypfldindx("mvs_zintr_struct","dollar_stack_save.index")=34
	Set gtmtypes("mvs_zintr_struct",35,"name")="mvs_zintr_struct.dollar_stack_save.incomplete"
	Set gtmtypes("mvs_zintr_struct",35,"off")=180
	Set gtmtypes("mvs_zintr_struct",35,"len")=4
	Set gtmtypes("mvs_zintr_struct",35,"type")="boolean_t"
	Set gtmtypfldindx("mvs_zintr_struct","dollar_stack_save.incomplete")=35
	;
	Set gtmtypes("nametabent")="struct"
	Set gtmtypes("nametabent",0)=2
	Set gtmtypes("nametabent","len")=32
	Set gtmtypes("nametabent",1,"name")="nametabent.len"
	Set gtmtypes("nametabent",1,"off")=0
	Set gtmtypes("nametabent",1,"len")=1
	Set gtmtypes("nametabent",1,"type")="char"
	Set gtmtypfldindx("nametabent","len")=1
	Set gtmtypes("nametabent",2,"name")="nametabent.name"
	Set gtmtypes("nametabent",2,"off")=1
	Set gtmtypes("nametabent",2,"len")=31
	Set gtmtypes("nametabent",2,"type")="char"
	Set gtmtypfldindx("nametabent","name")=2
	;
	Set gtmtypes("node_local")="struct"
	Set gtmtypes("node_local",0)=263
	Set gtmtypes("node_local","len")=113488
	Set gtmtypes("node_local",1,"name")="node_local.label"
	Set gtmtypes("node_local",1,"off")=0
	Set gtmtypes("node_local",1,"len")=12
	Set gtmtypes("node_local",1,"type")="unsigned-char"
	Set gtmtypfldindx("node_local","label")=1
	Set gtmtypes("node_local",2,"name")="node_local.fname"
	Set gtmtypes("node_local",2,"off")=12
	Set gtmtypes("node_local",2,"len")=256
	Set gtmtypes("node_local",2,"type")="unsigned-char"
	Set gtmtypfldindx("node_local","fname")=2
	Set gtmtypes("node_local",3,"name")="node_local.now_running"
	Set gtmtypes("node_local",3,"off")=268
	Set gtmtypes("node_local",3,"len")=36
	Set gtmtypes("node_local",3,"type")="char"
	Set gtmtypfldindx("node_local","now_running")=3
	Set gtmtypes("node_local",4,"name")="node_local.machine_name"
	Set gtmtypes("node_local",4,"off")=304
	Set gtmtypes("node_local",4,"len")=256
	Set gtmtypes("node_local",4,"type")="char"
	Set gtmtypfldindx("node_local","machine_name")=4
	Set gtmtypes("node_local",5,"name")="node_local.bt_header_off"
	Set gtmtypes("node_local",5,"off")=560
	Set gtmtypes("node_local",5,"len")=8
	Set gtmtypes("node_local",5,"type")="intptr_t"
	Set gtmtypfldindx("node_local","bt_header_off")=5
	Set gtmtypes("node_local",6,"name")="node_local.bt_base_off"
	Set gtmtypes("node_local",6,"off")=568
	Set gtmtypes("node_local",6,"len")=8
	Set gtmtypes("node_local",6,"type")="intptr_t"
	Set gtmtypfldindx("node_local","bt_base_off")=6
	Set gtmtypes("node_local",7,"name")="node_local.th_base_off"
	Set gtmtypes("node_local",7,"off")=576
	Set gtmtypes("node_local",7,"len")=8
	Set gtmtypes("node_local",7,"type")="intptr_t"
	Set gtmtypfldindx("node_local","th_base_off")=7
	Set gtmtypes("node_local",8,"name")="node_local.cache_off"
	Set gtmtypes("node_local",8,"off")=584
	Set gtmtypes("node_local",8,"len")=8
	Set gtmtypes("node_local",8,"type")="intptr_t"
	Set gtmtypfldindx("node_local","cache_off")=8
	Set gtmtypes("node_local",9,"name")="node_local.cur_lru_cache_rec_off"
	Set gtmtypes("node_local",9,"off")=592
	Set gtmtypes("node_local",9,"len")=8
	Set gtmtypes("node_local",9,"type")="intptr_t"
	Set gtmtypfldindx("node_local","cur_lru_cache_rec_off")=9
	Set gtmtypes("node_local",10,"name")="node_local.critical"
	Set gtmtypes("node_local",10,"off")=600
	Set gtmtypes("node_local",10,"len")=8
	Set gtmtypes("node_local",10,"type")="intptr_t"
	Set gtmtypfldindx("node_local","critical")=10
	Set gtmtypes("node_local",11,"name")="node_local.jnl_buff"
	Set gtmtypes("node_local",11,"off")=608
	Set gtmtypes("node_local",11,"len")=8
	Set gtmtypes("node_local",11,"type")="intptr_t"
	Set gtmtypfldindx("node_local","jnl_buff")=11
	Set gtmtypes("node_local",12,"name")="node_local.shmpool_buffer"
	Set gtmtypes("node_local",12,"off")=616
	Set gtmtypes("node_local",12,"len")=8
	Set gtmtypes("node_local",12,"type")="intptr_t"
	Set gtmtypfldindx("node_local","shmpool_buffer")=12
	Set gtmtypes("node_local",13,"name")="node_local.lock_addrs"
	Set gtmtypes("node_local",13,"off")=624
	Set gtmtypes("node_local",13,"len")=8
	Set gtmtypes("node_local",13,"type")="intptr_t"
	Set gtmtypfldindx("node_local","lock_addrs")=13
	Set gtmtypes("node_local",14,"name")="node_local.hdr"
	Set gtmtypes("node_local",14,"off")=632
	Set gtmtypes("node_local",14,"len")=8
	Set gtmtypes("node_local",14,"type")="intptr_t"
	Set gtmtypfldindx("node_local","hdr")=14
	Set gtmtypes("node_local",15,"name")="node_local.in_crit"
	Set gtmtypes("node_local",15,"off")=640
	Set gtmtypes("node_local",15,"len")=4
	Set gtmtypes("node_local",15,"type")="int"
	Set gtmtypfldindx("node_local","in_crit")=15
	Set gtmtypes("node_local",16,"name")="node_local.in_reinit"
	Set gtmtypes("node_local",16,"off")=644
	Set gtmtypes("node_local",16,"len")=4
	Set gtmtypes("node_local",16,"type")="int"
	Set gtmtypfldindx("node_local","in_reinit")=16
	Set gtmtypes("node_local",17,"name")="node_local.ccp_cycle"
	Set gtmtypes("node_local",17,"off")=648
	Set gtmtypes("node_local",17,"len")=2
	Set gtmtypes("node_local",17,"type")="unsigned-short"
	Set gtmtypfldindx("node_local","ccp_cycle")=17
	Set gtmtypes("node_local",18,"name")="node_local.filler"
	Set gtmtypes("node_local",18,"off")=650
	Set gtmtypes("node_local",18,"len")=2
	Set gtmtypes("node_local",18,"type")="unsigned-short"
	Set gtmtypfldindx("node_local","filler")=18
	Set gtmtypes("node_local",19,"name")="node_local.ccp_crit_blocked"
	Set gtmtypes("node_local",19,"off")=652
	Set gtmtypes("node_local",19,"len")=4
	Set gtmtypes("node_local",19,"type")="boolean_t"
	Set gtmtypfldindx("node_local","ccp_crit_blocked")=19
	Set gtmtypes("node_local",20,"name")="node_local.ccp_state"
	Set gtmtypes("node_local",20,"off")=656
	Set gtmtypes("node_local",20,"len")=4
	Set gtmtypes("node_local",20,"type")="int"
	Set gtmtypfldindx("node_local","ccp_state")=20
	Set gtmtypes("node_local",21,"name")="node_local.ccp_jnl_closed"
	Set gtmtypes("node_local",21,"off")=660
	Set gtmtypes("node_local",21,"len")=4
	Set gtmtypes("node_local",21,"type")="boolean_t"
	Set gtmtypfldindx("node_local","ccp_jnl_closed")=21
	Set gtmtypes("node_local",22,"name")="node_local.glob_sec_init"
	Set gtmtypes("node_local",22,"off")=664
	Set gtmtypes("node_local",22,"len")=4
	Set gtmtypes("node_local",22,"type")="boolean_t"
	Set gtmtypfldindx("node_local","glob_sec_init")=22
	Set gtmtypes("node_local",23,"name")="node_local.wtstart_pid"
	Set gtmtypes("node_local",23,"off")=668
	Set gtmtypes("node_local",23,"len")=32
	Set gtmtypes("node_local",23,"type")="unsigned-int"
	Set gtmtypfldindx("node_local","wtstart_pid")=23
	Set gtmtypes("node_local",23,"dim")=8
	Set gtmtypes("node_local",24,"name")="node_local.wc_blocked"
	Set gtmtypes("node_local",24,"off")=700
	Set gtmtypes("node_local",24,"len")=4
	Set gtmtypes("node_local",24,"type")="int"
	Set gtmtypfldindx("node_local","wc_blocked")=24
	Set gtmtypes("node_local",25,"name")="node_local.wc_var_lock"
	Set gtmtypes("node_local",25,"off")=704
	Set gtmtypes("node_local",25,"len")=8
	Set gtmtypes("node_local",25,"type")="global_latch_t"
	Set gtmtypfldindx("node_local","wc_var_lock")=25
	Set gtmtypes("node_local",26,"name")="node_local.wc_var_lock.u"
	Set gtmtypes("node_local",26,"off")=704
	Set gtmtypes("node_local",26,"len")=8
	Set gtmtypes("node_local",26,"type")="union"
	Set gtmtypfldindx("node_local","wc_var_lock.u")=26
	Set gtmtypes("node_local",27,"name")="node_local.wc_var_lock.u.pid_imgcnt"
	Set gtmtypes("node_local",27,"off")=704
	Set gtmtypes("node_local",27,"len")=8
	Set gtmtypes("node_local",27,"type")="uint64_t"
	Set gtmtypfldindx("node_local","wc_var_lock.u.pid_imgcnt")=27
	Set gtmtypes("node_local",28,"name")="node_local.wc_var_lock.u.parts"
	Set gtmtypes("node_local",28,"off")=704
	Set gtmtypes("node_local",28,"len")=8
	Set gtmtypes("node_local",28,"type")="struct"
	Set gtmtypfldindx("node_local","wc_var_lock.u.parts")=28
	Set gtmtypes("node_local",29,"name")="node_local.wc_var_lock.u.parts.latch_pid"
	Set gtmtypes("node_local",29,"off")=704
	Set gtmtypes("node_local",29,"len")=4
	Set gtmtypes("node_local",29,"type")="int"
	Set gtmtypfldindx("node_local","wc_var_lock.u.parts.latch_pid")=29
	Set gtmtypes("node_local",30,"name")="node_local.wc_var_lock.u.parts.latch_word"
	Set gtmtypes("node_local",30,"off")=708
	Set gtmtypes("node_local",30,"len")=4
	Set gtmtypes("node_local",30,"type")="int"
	Set gtmtypfldindx("node_local","wc_var_lock.u.parts.latch_word")=30
	Set gtmtypes("node_local",31,"name")="node_local.fill_cacheline1"
	Set gtmtypes("node_local",31,"off")=712
	Set gtmtypes("node_local",31,"len")=56
	Set gtmtypes("node_local",31,"type")="char"
	Set gtmtypfldindx("node_local","fill_cacheline1")=31
	Set gtmtypes("node_local",32,"name")="node_local.db_latch"
	Set gtmtypes("node_local",32,"off")=768
	Set gtmtypes("node_local",32,"len")=8
	Set gtmtypes("node_local",32,"type")="global_latch_t"
	Set gtmtypfldindx("node_local","db_latch")=32
	Set gtmtypes("node_local",33,"name")="node_local.db_latch.u"
	Set gtmtypes("node_local",33,"off")=768
	Set gtmtypes("node_local",33,"len")=8
	Set gtmtypes("node_local",33,"type")="union"
	Set gtmtypfldindx("node_local","db_latch.u")=33
	Set gtmtypes("node_local",34,"name")="node_local.db_latch.u.pid_imgcnt"
	Set gtmtypes("node_local",34,"off")=768
	Set gtmtypes("node_local",34,"len")=8
	Set gtmtypes("node_local",34,"type")="uint64_t"
	Set gtmtypfldindx("node_local","db_latch.u.pid_imgcnt")=34
	Set gtmtypes("node_local",35,"name")="node_local.db_latch.u.parts"
	Set gtmtypes("node_local",35,"off")=768
	Set gtmtypes("node_local",35,"len")=8
	Set gtmtypes("node_local",35,"type")="struct"
	Set gtmtypfldindx("node_local","db_latch.u.parts")=35
	Set gtmtypes("node_local",36,"name")="node_local.db_latch.u.parts.latch_pid"
	Set gtmtypes("node_local",36,"off")=768
	Set gtmtypes("node_local",36,"len")=4
	Set gtmtypes("node_local",36,"type")="int"
	Set gtmtypfldindx("node_local","db_latch.u.parts.latch_pid")=36
	Set gtmtypes("node_local",37,"name")="node_local.db_latch.u.parts.latch_word"
	Set gtmtypes("node_local",37,"off")=772
	Set gtmtypes("node_local",37,"len")=4
	Set gtmtypes("node_local",37,"type")="int"
	Set gtmtypfldindx("node_local","db_latch.u.parts.latch_word")=37
	Set gtmtypes("node_local",38,"name")="node_local.fill_cacheline2"
	Set gtmtypes("node_local",38,"off")=776
	Set gtmtypes("node_local",38,"len")=56
	Set gtmtypes("node_local",38,"type")="char"
	Set gtmtypfldindx("node_local","fill_cacheline2")=38
	Set gtmtypes("node_local",39,"name")="node_local.cache_hits"
	Set gtmtypes("node_local",39,"off")=832
	Set gtmtypes("node_local",39,"len")=4
	Set gtmtypes("node_local",39,"type")="int"
	Set gtmtypfldindx("node_local","cache_hits")=39
	Set gtmtypes("node_local",40,"name")="node_local.wc_in_free"
	Set gtmtypes("node_local",40,"off")=836
	Set gtmtypes("node_local",40,"len")=4
	Set gtmtypes("node_local",40,"type")="int"
	Set gtmtypfldindx("node_local","wc_in_free")=40
	Set gtmtypes("node_local",41,"name")="node_local.wcs_timers"
	Set gtmtypes("node_local",41,"off")=840
	Set gtmtypes("node_local",41,"len")=4
	Set gtmtypes("node_local",41,"type")="int"
	Set gtmtypfldindx("node_local","wcs_timers")=41
	Set gtmtypes("node_local",42,"name")="node_local.fill_cacheline3"
	Set gtmtypes("node_local",42,"off")=844
	Set gtmtypes("node_local",42,"len")=60
	Set gtmtypes("node_local",42,"type")="char"
	Set gtmtypfldindx("node_local","fill_cacheline3")=42
	Set gtmtypes("node_local",43,"name")="node_local.wcs_active_lvl"
	Set gtmtypes("node_local",43,"off")=904
	Set gtmtypes("node_local",43,"len")=4
	Set gtmtypes("node_local",43,"type")="int"
	Set gtmtypfldindx("node_local","wcs_active_lvl")=43
	Set gtmtypes("node_local",44,"name")="node_local.fill_cacheline4"
	Set gtmtypes("node_local",44,"off")=908
	Set gtmtypes("node_local",44,"len")=60
	Set gtmtypes("node_local",44,"type")="char"
	Set gtmtypfldindx("node_local","fill_cacheline4")=44
	Set gtmtypes("node_local",45,"name")="node_local.wcs_staleness"
	Set gtmtypes("node_local",45,"off")=968
	Set gtmtypes("node_local",45,"len")=4
	Set gtmtypes("node_local",45,"type")="int"
	Set gtmtypfldindx("node_local","wcs_staleness")=45
	Set gtmtypes("node_local",46,"name")="node_local.fill_cacheline5"
	Set gtmtypes("node_local",46,"off")=972
	Set gtmtypes("node_local",46,"len")=60
	Set gtmtypes("node_local",46,"type")="char"
	Set gtmtypfldindx("node_local","fill_cacheline5")=46
	Set gtmtypes("node_local",47,"name")="node_local.ref_cnt"
	Set gtmtypes("node_local",47,"off")=1032
	Set gtmtypes("node_local",47,"len")=4
	Set gtmtypes("node_local",47,"type")="int"
	Set gtmtypfldindx("node_local","ref_cnt")=47
	Set gtmtypes("node_local",48,"name")="node_local.fill_cacheline6"
	Set gtmtypes("node_local",48,"off")=1036
	Set gtmtypes("node_local",48,"len")=60
	Set gtmtypes("node_local",48,"type")="char"
	Set gtmtypfldindx("node_local","fill_cacheline6")=48
	Set gtmtypes("node_local",49,"name")="node_local.intent_wtstart"
	Set gtmtypes("node_local",49,"off")=1096
	Set gtmtypes("node_local",49,"len")=4
	Set gtmtypes("node_local",49,"type")="int"
	Set gtmtypfldindx("node_local","intent_wtstart")=49
	Set gtmtypes("node_local",50,"name")="node_local.fill_cacheline7"
	Set gtmtypes("node_local",50,"off")=1100
	Set gtmtypes("node_local",50,"len")=60
	Set gtmtypes("node_local",50,"type")="char"
	Set gtmtypfldindx("node_local","fill_cacheline7")=50
	Set gtmtypes("node_local",51,"name")="node_local.in_wtstart"
	Set gtmtypes("node_local",51,"off")=1160
	Set gtmtypes("node_local",51,"len")=4
	Set gtmtypes("node_local",51,"type")="int"
	Set gtmtypfldindx("node_local","in_wtstart")=51
	Set gtmtypes("node_local",52,"name")="node_local.fill_cacheline8"
	Set gtmtypes("node_local",52,"off")=1164
	Set gtmtypes("node_local",52,"len")=60
	Set gtmtypes("node_local",52,"type")="char"
	Set gtmtypfldindx("node_local","fill_cacheline8")=52
	Set gtmtypes("node_local",53,"name")="node_local.wcs_phase2_commit_pidcnt"
	Set gtmtypes("node_local",53,"off")=1224
	Set gtmtypes("node_local",53,"len")=4
	Set gtmtypes("node_local",53,"type")="int"
	Set gtmtypfldindx("node_local","wcs_phase2_commit_pidcnt")=53
	Set gtmtypes("node_local",54,"name")="node_local.fill_cacheline9"
	Set gtmtypes("node_local",54,"off")=1228
	Set gtmtypes("node_local",54,"len")=60
	Set gtmtypes("node_local",54,"type")="char"
	Set gtmtypfldindx("node_local","fill_cacheline9")=54
	Set gtmtypes("node_local",55,"name")="node_local.wcs_wip_lvl"
	Set gtmtypes("node_local",55,"off")=1288
	Set gtmtypes("node_local",55,"len")=4
	Set gtmtypes("node_local",55,"type")="int"
	Set gtmtypfldindx("node_local","wcs_wip_lvl")=55
	Set gtmtypes("node_local",56,"name")="node_local.fill_cacheline10"
	Set gtmtypes("node_local",56,"off")=1292
	Set gtmtypes("node_local",56,"len")=60
	Set gtmtypes("node_local",56,"type")="char"
	Set gtmtypfldindx("node_local","fill_cacheline10")=56
	Set gtmtypes("node_local",57,"name")="node_local.wtfini_in_prog"
	Set gtmtypes("node_local",57,"off")=1352
	Set gtmtypes("node_local",57,"len")=4
	Set gtmtypes("node_local",57,"type")="int"
	Set gtmtypfldindx("node_local","wtfini_in_prog")=57
	Set gtmtypes("node_local",58,"name")="node_local.freezer_waited_for_kip"
	Set gtmtypes("node_local",58,"off")=1356
	Set gtmtypes("node_local",58,"len")=4
	Set gtmtypes("node_local",58,"type")="boolean_t"
	Set gtmtypfldindx("node_local","freezer_waited_for_kip")=58
	Set gtmtypes("node_local",59,"name")="node_local.mm_extender_pid"
	Set gtmtypes("node_local",59,"off")=1360
	Set gtmtypes("node_local",59,"len")=4
	Set gtmtypes("node_local",59,"type")="int"
	Set gtmtypfldindx("node_local","mm_extender_pid")=59
	Set gtmtypes("node_local",60,"name")="node_local.highest_lbm_blk_changed"
	Set gtmtypes("node_local",60,"off")=1364
	Set gtmtypes("node_local",60,"len")=4
	Set gtmtypes("node_local",60,"type")="int"
	Set gtmtypfldindx("node_local","highest_lbm_blk_changed")=60
	Set gtmtypes("node_local",61,"name")="node_local.nbb"
	Set gtmtypes("node_local",61,"off")=1368
	Set gtmtypes("node_local",61,"len")=4
	Set gtmtypes("node_local",61,"type")="int"
	Set gtmtypfldindx("node_local","nbb")=61
	Set gtmtypes("node_local",62,"name")="node_local.lockhist_idx"
	Set gtmtypes("node_local",62,"off")=1372
	Set gtmtypes("node_local",62,"len")=4
	Set gtmtypes("node_local",62,"type")="int"
	Set gtmtypfldindx("node_local","lockhist_idx")=62
	Set gtmtypes("node_local",63,"name")="node_local.crit_ops_index"
	Set gtmtypes("node_local",63,"off")=1376
	Set gtmtypes("node_local",63,"len")=4
	Set gtmtypes("node_local",63,"type")="int"
	Set gtmtypfldindx("node_local","crit_ops_index")=63
	Set gtmtypes("node_local",64,"name")="node_local.dskread_ops_index"
	Set gtmtypes("node_local",64,"off")=1380
	Set gtmtypes("node_local",64,"len")=4
	Set gtmtypes("node_local",64,"type")="int"
	Set gtmtypfldindx("node_local","dskread_ops_index")=64
	Set gtmtypes("node_local",65,"name")="node_local.ftok_ops_index"
	Set gtmtypes("node_local",65,"off")=1384
	Set gtmtypes("node_local",65,"len")=4
	Set gtmtypes("node_local",65,"type")="int"
	Set gtmtypfldindx("node_local","ftok_ops_index")=65
	Set gtmtypes("node_local",66,"name")="node_local.wcs_ops_index"
	Set gtmtypes("node_local",66,"off")=1388
	Set gtmtypes("node_local",66,"len")=4
	Set gtmtypes("node_local",66,"type")="int"
	Set gtmtypfldindx("node_local","wcs_ops_index")=66
	Set gtmtypes("node_local",67,"name")="node_local.lockhists"
	Set gtmtypes("node_local",67,"off")=1392
	Set gtmtypes("node_local",67,"len")=16384
	Set gtmtypes("node_local",67,"type")="lockhist"
	Set gtmtypfldindx("node_local","lockhists")=67
	Set gtmtypes("node_local",67,"dim")=512
	Set gtmtypes("node_local",68,"name")="node_local.lockhists[0].lock_addr"
	Set gtmtypes("node_local",68,"off")=1392
	Set gtmtypes("node_local",68,"len")=8
	Set gtmtypes("node_local",68,"type")="addr"
	Set gtmtypfldindx("node_local","lockhists[0].lock_addr")=68
	Set gtmtypes("node_local",69,"name")="node_local.lockhists[0].lock_callr"
	Set gtmtypes("node_local",69,"off")=1400
	Set gtmtypes("node_local",69,"len")=8
	Set gtmtypes("node_local",69,"type")="caddr_t"
	Set gtmtypfldindx("node_local","lockhists[0].lock_callr")=69
	Set gtmtypes("node_local",70,"name")="node_local.lockhists[0].lock_pid"
	Set gtmtypes("node_local",70,"off")=1408
	Set gtmtypes("node_local",70,"len")=4
	Set gtmtypes("node_local",70,"type")="int"
	Set gtmtypfldindx("node_local","lockhists[0].lock_pid")=70
	Set gtmtypes("node_local",71,"name")="node_local.lockhists[0].loop_cnt"
	Set gtmtypes("node_local",71,"off")=1412
	Set gtmtypes("node_local",71,"len")=4
	Set gtmtypes("node_local",71,"type")="int"
	Set gtmtypfldindx("node_local","lockhists[0].loop_cnt")=71
	Set gtmtypes("node_local",72,"name")="node_local.lockhists[0].lock_op"
	Set gtmtypes("node_local",72,"off")=1416
	Set gtmtypes("node_local",72,"len")=4
	Set gtmtypes("node_local",72,"type")="char"
	Set gtmtypfldindx("node_local","lockhists[0].lock_op")=72
	Set gtmtypes("node_local",73,"name")="node_local.crit_ops_array"
	Set gtmtypes("node_local",73,"off")=17776
	Set gtmtypes("node_local",73,"len")=12288
	Set gtmtypes("node_local",73,"type")="crit_trace"
	Set gtmtypfldindx("node_local","crit_ops_array")=73
	Set gtmtypes("node_local",73,"dim")=512
	Set gtmtypes("node_local",74,"name")="node_local.crit_ops_array[0].call_from"
	Set gtmtypes("node_local",74,"off")=17776
	Set gtmtypes("node_local",74,"len")=8
	Set gtmtypes("node_local",74,"type")="caddr_t"
	Set gtmtypfldindx("node_local","crit_ops_array[0].call_from")=74
	Set gtmtypes("node_local",75,"name")="node_local.crit_ops_array[0].crit_act"
	Set gtmtypes("node_local",75,"off")=17784
	Set gtmtypes("node_local",75,"len")=4
	Set gtmtypes("node_local",75,"type")="int"
	Set gtmtypfldindx("node_local","crit_ops_array[0].crit_act")=75
	Set gtmtypes("node_local",76,"name")="node_local.crit_ops_array[0].epid"
	Set gtmtypes("node_local",76,"off")=17788
	Set gtmtypes("node_local",76,"len")=4
	Set gtmtypes("node_local",76,"type")="int"
	Set gtmtypfldindx("node_local","crit_ops_array[0].epid")=76
	Set gtmtypes("node_local",77,"name")="node_local.crit_ops_array[0].curr_tn"
	Set gtmtypes("node_local",77,"off")=17792
	Set gtmtypes("node_local",77,"len")=8
	Set gtmtypes("node_local",77,"type")="uint64_t"
	Set gtmtypfldindx("node_local","crit_ops_array[0].curr_tn")=77
	Set gtmtypes("node_local",78,"name")="node_local.dskread_ops_array"
	Set gtmtypes("node_local",78,"off")=30064
	Set gtmtypes("node_local",78,"len")=16384
	Set gtmtypes("node_local",78,"type")="dskread_trace"
	Set gtmtypfldindx("node_local","dskread_ops_array")=78
	Set gtmtypes("node_local",78,"dim")=512
	Set gtmtypes("node_local",79,"name")="node_local.dskread_ops_array[0].cr_off"
	Set gtmtypes("node_local",79,"off")=30064
	Set gtmtypes("node_local",79,"len")=8
	Set gtmtypes("node_local",79,"type")="intptr_t"
	Set gtmtypfldindx("node_local","dskread_ops_array[0].cr_off")=79
	Set gtmtypes("node_local",80,"name")="node_local.dskread_ops_array[0].cr_tn"
	Set gtmtypes("node_local",80,"off")=30072
	Set gtmtypes("node_local",80,"len")=8
	Set gtmtypes("node_local",80,"type")="uint64_t"
	Set gtmtypfldindx("node_local","dskread_ops_array[0].cr_tn")=80
	Set gtmtypes("node_local",81,"name")="node_local.dskread_ops_array[0].process_id"
	Set gtmtypes("node_local",81,"off")=30080
	Set gtmtypes("node_local",81,"len")=4
	Set gtmtypes("node_local",81,"type")="unsigned-int"
	Set gtmtypfldindx("node_local","dskread_ops_array[0].process_id")=81
	Set gtmtypes("node_local",82,"name")="node_local.dskread_ops_array[0].blk"
	Set gtmtypes("node_local",82,"off")=30084
	Set gtmtypes("node_local",82,"len")=4
	Set gtmtypes("node_local",82,"type")="int"
	Set gtmtypfldindx("node_local","dskread_ops_array[0].blk")=82
	Set gtmtypes("node_local",83,"name")="node_local.dskread_ops_array[0].cycle"
	Set gtmtypes("node_local",83,"off")=30088
	Set gtmtypes("node_local",83,"len")=4
	Set gtmtypes("node_local",83,"type")="unsigned-int"
	Set gtmtypfldindx("node_local","dskread_ops_array[0].cycle")=83
	Set gtmtypes("node_local",84,"name")="node_local.wcs_ops_array"
	Set gtmtypes("node_local",84,"off")=46448
	Set gtmtypes("node_local",84,"len")=57344
	Set gtmtypes("node_local",84,"type")="wcs_ops_trace_t"
	Set gtmtypfldindx("node_local","wcs_ops_array")=84
	Set gtmtypes("node_local",84,"dim")=1024
	Set gtmtypes("node_local",85,"name")="node_local.wcs_ops_array[0].db_tn"
	Set gtmtypes("node_local",85,"off")=46448
	Set gtmtypes("node_local",85,"len")=8
	Set gtmtypes("node_local",85,"type")="uint64_t"
	Set gtmtypfldindx("node_local","wcs_ops_array[0].db_tn")=85
	Set gtmtypes("node_local",86,"name")="node_local.wcs_ops_array[0].process_id"
	Set gtmtypes("node_local",86,"off")=46456
	Set gtmtypes("node_local",86,"len")=4
	Set gtmtypes("node_local",86,"type")="unsigned-int"
	Set gtmtypfldindx("node_local","wcs_ops_array[0].process_id")=86
	Set gtmtypes("node_local",87,"name")="node_local.wcs_ops_array[0].wc_in_free"
	Set gtmtypes("node_local",87,"off")=46460
	Set gtmtypes("node_local",87,"len")=4
	Set gtmtypes("node_local",87,"type")="unsigned-int"
	Set gtmtypfldindx("node_local","wcs_ops_array[0].wc_in_free")=87
	Set gtmtypes("node_local",88,"name")="node_local.wcs_ops_array[0].wcs_active_lvl"
	Set gtmtypes("node_local",88,"off")=46464
	Set gtmtypes("node_local",88,"len")=4
	Set gtmtypes("node_local",88,"type")="unsigned-int"
	Set gtmtypfldindx("node_local","wcs_ops_array[0].wcs_active_lvl")=88
	Set gtmtypes("node_local",89,"name")="node_local.wcs_ops_array[0].wcs_wip_lvl"
	Set gtmtypes("node_local",89,"off")=46468
	Set gtmtypes("node_local",89,"len")=4
	Set gtmtypes("node_local",89,"type")="unsigned-int"
	Set gtmtypfldindx("node_local","wcs_ops_array[0].wcs_wip_lvl")=89
	Set gtmtypes("node_local",90,"name")="node_local.wcs_ops_array[0].type"
	Set gtmtypes("node_local",90,"off")=46472
	Set gtmtypes("node_local",90,"len")=4
	Set gtmtypes("node_local",90,"type")="int"
	Set gtmtypfldindx("node_local","wcs_ops_array[0].type")=90
	Set gtmtypes("node_local",91,"name")="node_local.wcs_ops_array[0].blk"
	Set gtmtypes("node_local",91,"off")=46476
	Set gtmtypes("node_local",91,"len")=4
	Set gtmtypes("node_local",91,"type")="int"
	Set gtmtypfldindx("node_local","wcs_ops_array[0].blk")=91
	Set gtmtypes("node_local",92,"name")="node_local.wcs_ops_array[0].cr_off"
	Set gtmtypes("node_local",92,"off")=46480
	Set gtmtypes("node_local",92,"len")=8
	Set gtmtypes("node_local",92,"type")="intptr_t"
	Set gtmtypfldindx("node_local","wcs_ops_array[0].cr_off")=92
	Set gtmtypes("node_local",93,"name")="node_local.wcs_ops_array[0].cr_dirty"
	Set gtmtypes("node_local",93,"off")=46488
	Set gtmtypes("node_local",93,"len")=8
	Set gtmtypes("node_local",93,"type")="uint64_t"
	Set gtmtypfldindx("node_local","wcs_ops_array[0].cr_dirty")=93
	Set gtmtypes("node_local",94,"name")="node_local.wcs_ops_array[0].detail1"
	Set gtmtypes("node_local",94,"off")=46496
	Set gtmtypes("node_local",94,"len")=4
	Set gtmtypes("node_local",94,"type")="unsigned-int"
	Set gtmtypfldindx("node_local","wcs_ops_array[0].detail1")=94
	Set gtmtypes("node_local",95,"name")="node_local.wcs_ops_array[0].detail2"
	Set gtmtypes("node_local",95,"off")=46500
	Set gtmtypes("node_local",95,"len")=4
	Set gtmtypes("node_local",95,"type")="unsigned-int"
	Set gtmtypfldindx("node_local","wcs_ops_array[0].detail2")=95
	Set gtmtypes("node_local",96,"name")="node_local.unique_id"
	Set gtmtypes("node_local",96,"off")=103792
	Set gtmtypes("node_local",96,"len")=24
	Set gtmtypes("node_local",96,"type")="unique_file_id"
	Set gtmtypfldindx("node_local","unique_id")=96
	Set gtmtypes("node_local",97,"name")="node_local.unique_id.uid"
	Set gtmtypes("node_local",97,"off")=103792
	Set gtmtypes("node_local",97,"len")=16
	Set gtmtypes("node_local",97,"type")="unix_file_id"
	Set gtmtypfldindx("node_local","unique_id.uid")=97
	Set gtmtypes("node_local",98,"name")="node_local.unique_id.file_id"
	Set gtmtypes("node_local",98,"off")=103792
	Set gtmtypes("node_local",98,"len")=20
	Set gtmtypes("node_local",98,"type")="char"
	Set gtmtypfldindx("node_local","unique_id.file_id")=98
	Set gtmtypes("node_local",99,"name")="node_local.owner_node"
	Set gtmtypes("node_local",99,"off")=103816
	Set gtmtypes("node_local",99,"len")=4
	Set gtmtypes("node_local",99,"type")="unsigned-int"
	Set gtmtypfldindx("node_local","owner_node")=99
	Set gtmtypes("node_local",100,"name")="node_local.wcsflu_pid"
	Set gtmtypes("node_local",100,"off")=103820
	Set gtmtypes("node_local",100,"len")=4
	Set gtmtypes("node_local",100,"type")="int"
	Set gtmtypfldindx("node_local","wcsflu_pid")=100
	Set gtmtypes("node_local",101,"name")="node_local.creation_date_time4"
	Set gtmtypes("node_local",101,"off")=103824
	Set gtmtypes("node_local",101,"len")=4
	Set gtmtypes("node_local",101,"type")="int"
	Set gtmtypfldindx("node_local","creation_date_time4")=101
	Set gtmtypes("node_local",102,"name")="node_local.inhibit_kills"
	Set gtmtypes("node_local",102,"off")=103828
	Set gtmtypes("node_local",102,"len")=4
	Set gtmtypes("node_local",102,"type")="int"
	Set gtmtypfldindx("node_local","inhibit_kills")=102
	Set gtmtypes("node_local",103,"name")="node_local.remove_shm"
	Set gtmtypes("node_local",103,"off")=103832
	Set gtmtypes("node_local",103,"len")=4
	Set gtmtypes("node_local",103,"type")="boolean_t"
	Set gtmtypfldindx("node_local","remove_shm")=103
	Set gtmtypes("node_local",104,"name")="node_local.jnl_file"
	Set gtmtypes("node_local",104,"off")=103840
	Set gtmtypes("node_local",104,"len")=32
	Set gtmtypes("node_local",104,"type")="union"
	Set gtmtypfldindx("node_local","jnl_file")=104
	Set gtmtypes("node_local",105,"name")="node_local.jnl_file.jnl_file_id"
	Set gtmtypes("node_local",105,"off")=103840
	Set gtmtypes("node_local",105,"len")=28
	Set gtmtypes("node_local",105,"type")="gds_file_id"
	Set gtmtypfldindx("node_local","jnl_file.jnl_file_id")=105
	Set gtmtypes("node_local",106,"name")="node_local.jnl_file.jnl_file_id.dvi"
	Set gtmtypes("node_local",106,"off")=103840
	Set gtmtypes("node_local",106,"len")=16
	Set gtmtypes("node_local",106,"type")="char"
	Set gtmtypfldindx("node_local","jnl_file.jnl_file_id.dvi")=106
	Set gtmtypes("node_local",107,"name")="node_local.jnl_file.jnl_file_id.did"
	Set gtmtypes("node_local",107,"off")=103856
	Set gtmtypes("node_local",107,"len")=6
	Set gtmtypes("node_local",107,"type")="unsigned-short"
	Set gtmtypfldindx("node_local","jnl_file.jnl_file_id.did")=107
	Set gtmtypes("node_local",107,"dim")=3
	Set gtmtypes("node_local",108,"name")="node_local.jnl_file.jnl_file_id.fid"
	Set gtmtypes("node_local",108,"off")=103862
	Set gtmtypes("node_local",108,"len")=6
	Set gtmtypes("node_local",108,"type")="unsigned-short"
	Set gtmtypfldindx("node_local","jnl_file.jnl_file_id.fid")=108
	Set gtmtypes("node_local",108,"dim")=3
	Set gtmtypes("node_local",109,"name")="node_local.jnl_file.u"
	Set gtmtypes("node_local",109,"off")=103840
	Set gtmtypes("node_local",109,"len")=16
	Set gtmtypes("node_local",109,"type")="unix_file_id"
	Set gtmtypfldindx("node_local","jnl_file.u")=109
	Set gtmtypes("node_local",110,"name")="node_local.jnl_file.u.inode"
	Set gtmtypes("node_local",110,"off")=103840
	Set gtmtypes("node_local",110,"len")=8
	Set gtmtypes("node_local",110,"type")="ino_t"
	Set gtmtypfldindx("node_local","jnl_file.u.inode")=110
	Set gtmtypes("node_local",111,"name")="node_local.jnl_file.u.device"
	Set gtmtypes("node_local",111,"off")=103848
	Set gtmtypes("node_local",111,"len")=8
	Set gtmtypes("node_local",111,"type")="dev_t"
	Set gtmtypfldindx("node_local","jnl_file.u.device")=111
	Set gtmtypes("node_local",112,"name")="node_local.donotflush_dbjnl"
	Set gtmtypes("node_local",112,"off")=103872
	Set gtmtypes("node_local",112,"len")=4
	Set gtmtypes("node_local",112,"type")="boolean_t"
	Set gtmtypfldindx("node_local","donotflush_dbjnl")=112
	Set gtmtypes("node_local",113,"name")="node_local.n_pre_read"
	Set gtmtypes("node_local",113,"off")=103876
	Set gtmtypes("node_local",113,"len")=4
	Set gtmtypes("node_local",113,"type")="int"
	Set gtmtypfldindx("node_local","n_pre_read")=113
	Set gtmtypes("node_local",114,"name")="node_local.replinstfilename"
	Set gtmtypes("node_local",114,"off")=103880
	Set gtmtypes("node_local",114,"len")=256
	Set gtmtypes("node_local",114,"type")="char"
	Set gtmtypfldindx("node_local","replinstfilename")=114
	Set gtmtypes("node_local",115,"name")="node_local.statsdb_fname"
	Set gtmtypes("node_local",115,"off")=104136
	Set gtmtypes("node_local",115,"len")=256
	Set gtmtypes("node_local",115,"type")="char"
	Set gtmtypfldindx("node_local","statsdb_fname")=115
	Set gtmtypes("node_local",116,"name")="node_local.gvstats_rec"
	Set gtmtypes("node_local",116,"off")=104392
	Set gtmtypes("node_local",116,"len")=592
	Set gtmtypes("node_local",116,"type")="gvstats_rec_t"
	Set gtmtypfldindx("node_local","gvstats_rec")=116
	Set gtmtypes("node_local",117,"name")="node_local.gvstats_rec.n_set"
	Set gtmtypes("node_local",117,"off")=104392
	Set gtmtypes("node_local",117,"len")=8
	Set gtmtypes("node_local",117,"type")="uint64_t"
	Set gtmtypfldindx("node_local","gvstats_rec.n_set")=117
	Set gtmtypes("node_local",118,"name")="node_local.gvstats_rec.n_kill"
	Set gtmtypes("node_local",118,"off")=104400
	Set gtmtypes("node_local",118,"len")=8
	Set gtmtypes("node_local",118,"type")="uint64_t"
	Set gtmtypfldindx("node_local","gvstats_rec.n_kill")=118
	Set gtmtypes("node_local",119,"name")="node_local.gvstats_rec.n_get"
	Set gtmtypes("node_local",119,"off")=104408
	Set gtmtypes("node_local",119,"len")=8
	Set gtmtypes("node_local",119,"type")="uint64_t"
	Set gtmtypfldindx("node_local","gvstats_rec.n_get")=119
	Set gtmtypes("node_local",120,"name")="node_local.gvstats_rec.n_data"
	Set gtmtypes("node_local",120,"off")=104416
	Set gtmtypes("node_local",120,"len")=8
	Set gtmtypes("node_local",120,"type")="uint64_t"
	Set gtmtypfldindx("node_local","gvstats_rec.n_data")=120
	Set gtmtypes("node_local",121,"name")="node_local.gvstats_rec.n_order"
	Set gtmtypes("node_local",121,"off")=104424
	Set gtmtypes("node_local",121,"len")=8
	Set gtmtypes("node_local",121,"type")="uint64_t"
	Set gtmtypfldindx("node_local","gvstats_rec.n_order")=121
	Set gtmtypes("node_local",122,"name")="node_local.gvstats_rec.n_zprev"
	Set gtmtypes("node_local",122,"off")=104432
	Set gtmtypes("node_local",122,"len")=8
	Set gtmtypes("node_local",122,"type")="uint64_t"
	Set gtmtypfldindx("node_local","gvstats_rec.n_zprev")=122
	Set gtmtypes("node_local",123,"name")="node_local.gvstats_rec.n_query"
	Set gtmtypes("node_local",123,"off")=104440
	Set gtmtypes("node_local",123,"len")=8
	Set gtmtypes("node_local",123,"type")="uint64_t"
	Set gtmtypfldindx("node_local","gvstats_rec.n_query")=123
	Set gtmtypes("node_local",124,"name")="node_local.gvstats_rec.n_lock_success"
	Set gtmtypes("node_local",124,"off")=104448
	Set gtmtypes("node_local",124,"len")=8
	Set gtmtypes("node_local",124,"type")="uint64_t"
	Set gtmtypfldindx("node_local","gvstats_rec.n_lock_success")=124
	Set gtmtypes("node_local",125,"name")="node_local.gvstats_rec.n_lock_fail"
	Set gtmtypes("node_local",125,"off")=104456
	Set gtmtypes("node_local",125,"len")=8
	Set gtmtypes("node_local",125,"type")="uint64_t"
	Set gtmtypfldindx("node_local","gvstats_rec.n_lock_fail")=125
	Set gtmtypes("node_local",126,"name")="node_local.gvstats_rec.db_curr_tn"
	Set gtmtypes("node_local",126,"off")=104464
	Set gtmtypes("node_local",126,"len")=8
	Set gtmtypes("node_local",126,"type")="uint64_t"
	Set gtmtypfldindx("node_local","gvstats_rec.db_curr_tn")=126
	Set gtmtypes("node_local",127,"name")="node_local.gvstats_rec.n_dsk_read"
	Set gtmtypes("node_local",127,"off")=104472
	Set gtmtypes("node_local",127,"len")=8
	Set gtmtypes("node_local",127,"type")="uint64_t"
	Set gtmtypfldindx("node_local","gvstats_rec.n_dsk_read")=127
	Set gtmtypes("node_local",128,"name")="node_local.gvstats_rec.n_dsk_write"
	Set gtmtypes("node_local",128,"off")=104480
	Set gtmtypes("node_local",128,"len")=8
	Set gtmtypes("node_local",128,"type")="uint64_t"
	Set gtmtypfldindx("node_local","gvstats_rec.n_dsk_write")=128
	Set gtmtypes("node_local",129,"name")="node_local.gvstats_rec.n_nontp_readwrite"
	Set gtmtypes("node_local",129,"off")=104488
	Set gtmtypes("node_local",129,"len")=8
	Set gtmtypes("node_local",129,"type")="uint64_t"
	Set gtmtypfldindx("node_local","gvstats_rec.n_nontp_readwrite")=129
	Set gtmtypes("node_local",130,"name")="node_local.gvstats_rec.n_nontp_readonly"
	Set gtmtypes("node_local",130,"off")=104496
	Set gtmtypes("node_local",130,"len")=8
	Set gtmtypes("node_local",130,"type")="uint64_t"
	Set gtmtypfldindx("node_local","gvstats_rec.n_nontp_readonly")=130
	Set gtmtypes("node_local",131,"name")="node_local.gvstats_rec.n_nontp_blkwrite"
	Set gtmtypes("node_local",131,"off")=104504
	Set gtmtypes("node_local",131,"len")=8
	Set gtmtypes("node_local",131,"type")="uint64_t"
	Set gtmtypfldindx("node_local","gvstats_rec.n_nontp_blkwrite")=131
	Set gtmtypes("node_local",132,"name")="node_local.gvstats_rec.n_nontp_blkread"
	Set gtmtypes("node_local",132,"off")=104512
	Set gtmtypes("node_local",132,"len")=8
	Set gtmtypes("node_local",132,"type")="uint64_t"
	Set gtmtypfldindx("node_local","gvstats_rec.n_nontp_blkread")=132
	Set gtmtypes("node_local",133,"name")="node_local.gvstats_rec.n_nontp_retries_0"
	Set gtmtypes("node_local",133,"off")=104520
	Set gtmtypes("node_local",133,"len")=8
	Set gtmtypes("node_local",133,"type")="uint64_t"
	Set gtmtypfldindx("node_local","gvstats_rec.n_nontp_retries_0")=133
	Set gtmtypes("node_local",134,"name")="node_local.gvstats_rec.n_nontp_retries_1"
	Set gtmtypes("node_local",134,"off")=104528
	Set gtmtypes("node_local",134,"len")=8
	Set gtmtypes("node_local",134,"type")="uint64_t"
	Set gtmtypfldindx("node_local","gvstats_rec.n_nontp_retries_1")=134
	Set gtmtypes("node_local",135,"name")="node_local.gvstats_rec.n_nontp_retries_2"
	Set gtmtypes("node_local",135,"off")=104536
	Set gtmtypes("node_local",135,"len")=8
	Set gtmtypes("node_local",135,"type")="uint64_t"
	Set gtmtypfldindx("node_local","gvstats_rec.n_nontp_retries_2")=135
	Set gtmtypes("node_local",136,"name")="node_local.gvstats_rec.n_nontp_retries_3"
	Set gtmtypes("node_local",136,"off")=104544
	Set gtmtypes("node_local",136,"len")=8
	Set gtmtypes("node_local",136,"type")="uint64_t"
	Set gtmtypfldindx("node_local","gvstats_rec.n_nontp_retries_3")=136
	Set gtmtypes("node_local",137,"name")="node_local.gvstats_rec.n_tp_readwrite"
	Set gtmtypes("node_local",137,"off")=104552
	Set gtmtypes("node_local",137,"len")=8
	Set gtmtypes("node_local",137,"type")="uint64_t"
	Set gtmtypfldindx("node_local","gvstats_rec.n_tp_readwrite")=137
	Set gtmtypes("node_local",138,"name")="node_local.gvstats_rec.n_tp_readonly"
	Set gtmtypes("node_local",138,"off")=104560
	Set gtmtypes("node_local",138,"len")=8
	Set gtmtypes("node_local",138,"type")="uint64_t"
	Set gtmtypfldindx("node_local","gvstats_rec.n_tp_readonly")=138
	Set gtmtypes("node_local",139,"name")="node_local.gvstats_rec.n_tp_rolledback"
	Set gtmtypes("node_local",139,"off")=104568
	Set gtmtypes("node_local",139,"len")=8
	Set gtmtypes("node_local",139,"type")="uint64_t"
	Set gtmtypfldindx("node_local","gvstats_rec.n_tp_rolledback")=139
	Set gtmtypes("node_local",140,"name")="node_local.gvstats_rec.n_tp_blkwrite"
	Set gtmtypes("node_local",140,"off")=104576
	Set gtmtypes("node_local",140,"len")=8
	Set gtmtypes("node_local",140,"type")="uint64_t"
	Set gtmtypfldindx("node_local","gvstats_rec.n_tp_blkwrite")=140
	Set gtmtypes("node_local",141,"name")="node_local.gvstats_rec.n_tp_blkread"
	Set gtmtypes("node_local",141,"off")=104584
	Set gtmtypes("node_local",141,"len")=8
	Set gtmtypes("node_local",141,"type")="uint64_t"
	Set gtmtypfldindx("node_local","gvstats_rec.n_tp_blkread")=141
	Set gtmtypes("node_local",142,"name")="node_local.gvstats_rec.n_tp_tot_retries_0"
	Set gtmtypes("node_local",142,"off")=104592
	Set gtmtypes("node_local",142,"len")=8
	Set gtmtypes("node_local",142,"type")="uint64_t"
	Set gtmtypfldindx("node_local","gvstats_rec.n_tp_tot_retries_0")=142
	Set gtmtypes("node_local",143,"name")="node_local.gvstats_rec.n_tp_tot_retries_1"
	Set gtmtypes("node_local",143,"off")=104600
	Set gtmtypes("node_local",143,"len")=8
	Set gtmtypes("node_local",143,"type")="uint64_t"
	Set gtmtypfldindx("node_local","gvstats_rec.n_tp_tot_retries_1")=143
	Set gtmtypes("node_local",144,"name")="node_local.gvstats_rec.n_tp_tot_retries_2"
	Set gtmtypes("node_local",144,"off")=104608
	Set gtmtypes("node_local",144,"len")=8
	Set gtmtypes("node_local",144,"type")="uint64_t"
	Set gtmtypfldindx("node_local","gvstats_rec.n_tp_tot_retries_2")=144
	Set gtmtypes("node_local",145,"name")="node_local.gvstats_rec.n_tp_tot_retries_3"
	Set gtmtypes("node_local",145,"off")=104616
	Set gtmtypes("node_local",145,"len")=8
	Set gtmtypes("node_local",145,"type")="uint64_t"
	Set gtmtypfldindx("node_local","gvstats_rec.n_tp_tot_retries_3")=145
	Set gtmtypes("node_local",146,"name")="node_local.gvstats_rec.n_tp_tot_retries_4"
	Set gtmtypes("node_local",146,"off")=104624
	Set gtmtypes("node_local",146,"len")=8
	Set gtmtypes("node_local",146,"type")="uint64_t"
	Set gtmtypfldindx("node_local","gvstats_rec.n_tp_tot_retries_4")=146
	Set gtmtypes("node_local",147,"name")="node_local.gvstats_rec.n_tp_cnflct_retries_0"
	Set gtmtypes("node_local",147,"off")=104632
	Set gtmtypes("node_local",147,"len")=8
	Set gtmtypes("node_local",147,"type")="uint64_t"
	Set gtmtypfldindx("node_local","gvstats_rec.n_tp_cnflct_retries_0")=147
	Set gtmtypes("node_local",148,"name")="node_local.gvstats_rec.n_tp_cnflct_retries_1"
	Set gtmtypes("node_local",148,"off")=104640
	Set gtmtypes("node_local",148,"len")=8
	Set gtmtypes("node_local",148,"type")="uint64_t"
	Set gtmtypfldindx("node_local","gvstats_rec.n_tp_cnflct_retries_1")=148
	Set gtmtypes("node_local",149,"name")="node_local.gvstats_rec.n_tp_cnflct_retries_2"
	Set gtmtypes("node_local",149,"off")=104648
	Set gtmtypes("node_local",149,"len")=8
	Set gtmtypes("node_local",149,"type")="uint64_t"
	Set gtmtypfldindx("node_local","gvstats_rec.n_tp_cnflct_retries_2")=149
	Set gtmtypes("node_local",150,"name")="node_local.gvstats_rec.n_tp_cnflct_retries_3"
	Set gtmtypes("node_local",150,"off")=104656
	Set gtmtypes("node_local",150,"len")=8
	Set gtmtypes("node_local",150,"type")="uint64_t"
	Set gtmtypfldindx("node_local","gvstats_rec.n_tp_cnflct_retries_3")=150
	Set gtmtypes("node_local",151,"name")="node_local.gvstats_rec.n_tp_cnflct_retries_4"
	Set gtmtypes("node_local",151,"off")=104664
	Set gtmtypes("node_local",151,"len")=8
	Set gtmtypes("node_local",151,"type")="uint64_t"
	Set gtmtypfldindx("node_local","gvstats_rec.n_tp_cnflct_retries_4")=151
	Set gtmtypes("node_local",152,"name")="node_local.gvstats_rec.n_ztrigger"
	Set gtmtypes("node_local",152,"off")=104672
	Set gtmtypes("node_local",152,"len")=8
	Set gtmtypes("node_local",152,"type")="uint64_t"
	Set gtmtypfldindx("node_local","gvstats_rec.n_ztrigger")=152
	Set gtmtypes("node_local",153,"name")="node_local.gvstats_rec.n_db_flush"
	Set gtmtypes("node_local",153,"off")=104680
	Set gtmtypes("node_local",153,"len")=8
	Set gtmtypes("node_local",153,"type")="uint64_t"
	Set gtmtypfldindx("node_local","gvstats_rec.n_db_flush")=153
	Set gtmtypes("node_local",154,"name")="node_local.gvstats_rec.n_db_fsync"
	Set gtmtypes("node_local",154,"off")=104688
	Set gtmtypes("node_local",154,"len")=8
	Set gtmtypes("node_local",154,"type")="uint64_t"
	Set gtmtypfldindx("node_local","gvstats_rec.n_db_fsync")=154
	Set gtmtypes("node_local",155,"name")="node_local.gvstats_rec.n_jnl_flush"
	Set gtmtypes("node_local",155,"off")=104696
	Set gtmtypes("node_local",155,"len")=8
	Set gtmtypes("node_local",155,"type")="uint64_t"
	Set gtmtypfldindx("node_local","gvstats_rec.n_jnl_flush")=155
	Set gtmtypes("node_local",156,"name")="node_local.gvstats_rec.n_jnl_fsync"
	Set gtmtypes("node_local",156,"off")=104704
	Set gtmtypes("node_local",156,"len")=8
	Set gtmtypes("node_local",156,"type")="uint64_t"
	Set gtmtypfldindx("node_local","gvstats_rec.n_jnl_fsync")=156
	Set gtmtypes("node_local",157,"name")="node_local.gvstats_rec.n_jbuff_bytes"
	Set gtmtypes("node_local",157,"off")=104712
	Set gtmtypes("node_local",157,"len")=8
	Set gtmtypes("node_local",157,"type")="uint64_t"
	Set gtmtypfldindx("node_local","gvstats_rec.n_jbuff_bytes")=157
	Set gtmtypes("node_local",158,"name")="node_local.gvstats_rec.n_jfile_bytes"
	Set gtmtypes("node_local",158,"off")=104720
	Set gtmtypes("node_local",158,"len")=8
	Set gtmtypes("node_local",158,"type")="uint64_t"
	Set gtmtypfldindx("node_local","gvstats_rec.n_jfile_bytes")=158
	Set gtmtypes("node_local",159,"name")="node_local.gvstats_rec.n_jfile_writes"
	Set gtmtypes("node_local",159,"off")=104728
	Set gtmtypes("node_local",159,"len")=8
	Set gtmtypes("node_local",159,"type")="uint64_t"
	Set gtmtypfldindx("node_local","gvstats_rec.n_jfile_writes")=159
	Set gtmtypes("node_local",160,"name")="node_local.gvstats_rec.n_jrec_logical"
	Set gtmtypes("node_local",160,"off")=104736
	Set gtmtypes("node_local",160,"len")=8
	Set gtmtypes("node_local",160,"type")="uint64_t"
	Set gtmtypfldindx("node_local","gvstats_rec.n_jrec_logical")=160
	Set gtmtypes("node_local",161,"name")="node_local.gvstats_rec.n_jrec_pblk"
	Set gtmtypes("node_local",161,"off")=104744
	Set gtmtypes("node_local",161,"len")=8
	Set gtmtypes("node_local",161,"type")="uint64_t"
	Set gtmtypfldindx("node_local","gvstats_rec.n_jrec_pblk")=161
	Set gtmtypes("node_local",162,"name")="node_local.gvstats_rec.n_jrec_epoch_regular"
	Set gtmtypes("node_local",162,"off")=104752
	Set gtmtypes("node_local",162,"len")=8
	Set gtmtypes("node_local",162,"type")="uint64_t"
	Set gtmtypfldindx("node_local","gvstats_rec.n_jrec_epoch_regular")=162
	Set gtmtypes("node_local",163,"name")="node_local.gvstats_rec.n_jrec_epoch_idle"
	Set gtmtypes("node_local",163,"off")=104760
	Set gtmtypes("node_local",163,"len")=8
	Set gtmtypes("node_local",163,"type")="uint64_t"
	Set gtmtypfldindx("node_local","gvstats_rec.n_jrec_epoch_idle")=163
	Set gtmtypes("node_local",164,"name")="node_local.gvstats_rec.n_jrec_other"
	Set gtmtypes("node_local",164,"off")=104768
	Set gtmtypes("node_local",164,"len")=8
	Set gtmtypes("node_local",164,"type")="uint64_t"
	Set gtmtypfldindx("node_local","gvstats_rec.n_jrec_other")=164
	Set gtmtypes("node_local",165,"name")="node_local.gvstats_rec.n_jnl_extends"
	Set gtmtypes("node_local",165,"off")=104776
	Set gtmtypes("node_local",165,"len")=8
	Set gtmtypes("node_local",165,"type")="uint64_t"
	Set gtmtypfldindx("node_local","gvstats_rec.n_jnl_extends")=165
	Set gtmtypes("node_local",166,"name")="node_local.gvstats_rec.n_db_extends"
	Set gtmtypes("node_local",166,"off")=104784
	Set gtmtypes("node_local",166,"len")=8
	Set gtmtypes("node_local",166,"type")="uint64_t"
	Set gtmtypfldindx("node_local","gvstats_rec.n_db_extends")=166
	Set gtmtypes("node_local",167,"name")="node_local.gvstats_rec.n_crit_success"
	Set gtmtypes("node_local",167,"off")=104792
	Set gtmtypes("node_local",167,"len")=8
	Set gtmtypes("node_local",167,"type")="uint64_t"
	Set gtmtypfldindx("node_local","gvstats_rec.n_crit_success")=167
	Set gtmtypes("node_local",168,"name")="node_local.gvstats_rec.n_crits_in_epch"
	Set gtmtypes("node_local",168,"off")=104800
	Set gtmtypes("node_local",168,"len")=8
	Set gtmtypes("node_local",168,"type")="uint64_t"
	Set gtmtypfldindx("node_local","gvstats_rec.n_crits_in_epch")=168
	Set gtmtypes("node_local",169,"name")="node_local.gvstats_rec.sq_crit_failed"
	Set gtmtypes("node_local",169,"off")=104808
	Set gtmtypes("node_local",169,"len")=8
	Set gtmtypes("node_local",169,"type")="uint64_t"
	Set gtmtypfldindx("node_local","gvstats_rec.sq_crit_failed")=169
	Set gtmtypes("node_local",170,"name")="node_local.gvstats_rec.n_crit_failed"
	Set gtmtypes("node_local",170,"off")=104816
	Set gtmtypes("node_local",170,"len")=8
	Set gtmtypes("node_local",170,"type")="uint64_t"
	Set gtmtypfldindx("node_local","gvstats_rec.n_crit_failed")=170
	Set gtmtypes("node_local",171,"name")="node_local.gvstats_rec.sq_crit_que_slps"
	Set gtmtypes("node_local",171,"off")=104824
	Set gtmtypes("node_local",171,"len")=8
	Set gtmtypes("node_local",171,"type")="uint64_t"
	Set gtmtypfldindx("node_local","gvstats_rec.sq_crit_que_slps")=171
	Set gtmtypes("node_local",172,"name")="node_local.gvstats_rec.n_crit_que_slps"
	Set gtmtypes("node_local",172,"off")=104832
	Set gtmtypes("node_local",172,"len")=8
	Set gtmtypes("node_local",172,"type")="uint64_t"
	Set gtmtypfldindx("node_local","gvstats_rec.n_crit_que_slps")=172
	Set gtmtypes("node_local",173,"name")="node_local.gvstats_rec.sq_crit_yields"
	Set gtmtypes("node_local",173,"off")=104840
	Set gtmtypes("node_local",173,"len")=8
	Set gtmtypes("node_local",173,"type")="uint64_t"
	Set gtmtypfldindx("node_local","gvstats_rec.sq_crit_yields")=173
	Set gtmtypes("node_local",174,"name")="node_local.gvstats_rec.n_crit_yields"
	Set gtmtypes("node_local",174,"off")=104848
	Set gtmtypes("node_local",174,"len")=8
	Set gtmtypes("node_local",174,"type")="uint64_t"
	Set gtmtypfldindx("node_local","gvstats_rec.n_crit_yields")=174
	Set gtmtypes("node_local",175,"name")="node_local.gvstats_rec.n_clean2dirty"
	Set gtmtypes("node_local",175,"off")=104856
	Set gtmtypes("node_local",175,"len")=8
	Set gtmtypes("node_local",175,"type")="uint64_t"
	Set gtmtypfldindx("node_local","gvstats_rec.n_clean2dirty")=175
	Set gtmtypes("node_local",176,"name")="node_local.gvstats_rec.n_wait_for_read"
	Set gtmtypes("node_local",176,"off")=104864
	Set gtmtypes("node_local",176,"len")=8
	Set gtmtypes("node_local",176,"type")="uint64_t"
	Set gtmtypfldindx("node_local","gvstats_rec.n_wait_for_read")=176
	Set gtmtypes("node_local",177,"name")="node_local.gvstats_rec.n_buffer_scarce"
	Set gtmtypes("node_local",177,"off")=104872
	Set gtmtypes("node_local",177,"len")=8
	Set gtmtypes("node_local",177,"type")="uint64_t"
	Set gtmtypfldindx("node_local","gvstats_rec.n_buffer_scarce")=177
	Set gtmtypes("node_local",178,"name")="node_local.gvstats_rec.n_bt_scarce"
	Set gtmtypes("node_local",178,"off")=104880
	Set gtmtypes("node_local",178,"len")=8
	Set gtmtypes("node_local",178,"type")="uint64_t"
	Set gtmtypfldindx("node_local","gvstats_rec.n_bt_scarce")=178
	Set gtmtypes("node_local",179,"name")="node_local.gvstats_rec.f_dbext_wait"
	Set gtmtypes("node_local",179,"off")=104888
	Set gtmtypes("node_local",179,"len")=8
	Set gtmtypes("node_local",179,"type")="uint64_t"
	Set gtmtypfldindx("node_local","gvstats_rec.f_dbext_wait")=179
	Set gtmtypes("node_local",180,"name")="node_local.gvstats_rec.f_bg_wait"
	Set gtmtypes("node_local",180,"off")=104896
	Set gtmtypes("node_local",180,"len")=8
	Set gtmtypes("node_local",180,"type")="uint64_t"
	Set gtmtypfldindx("node_local","gvstats_rec.f_bg_wait")=180
	Set gtmtypes("node_local",181,"name")="node_local.gvstats_rec.f_jnl_wait"
	Set gtmtypes("node_local",181,"off")=104904
	Set gtmtypes("node_local",181,"len")=8
	Set gtmtypes("node_local",181,"type")="uint64_t"
	Set gtmtypfldindx("node_local","gvstats_rec.f_jnl_wait")=181
	Set gtmtypes("node_local",182,"name")="node_local.gvstats_rec.f_mlk_wait"
	Set gtmtypes("node_local",182,"off")=104912
	Set gtmtypes("node_local",182,"len")=8
	Set gtmtypes("node_local",182,"type")="uint64_t"
	Set gtmtypfldindx("node_local","gvstats_rec.f_mlk_wait")=182
	Set gtmtypes("node_local",183,"name")="node_local.gvstats_rec.f_proc_wait"
	Set gtmtypes("node_local",183,"off")=104920
	Set gtmtypes("node_local",183,"len")=8
	Set gtmtypes("node_local",183,"type")="uint64_t"
	Set gtmtypfldindx("node_local","gvstats_rec.f_proc_wait")=183
	Set gtmtypes("node_local",184,"name")="node_local.gvstats_rec.f_trans_wait"
	Set gtmtypes("node_local",184,"off")=104928
	Set gtmtypes("node_local",184,"len")=8
	Set gtmtypes("node_local",184,"type")="uint64_t"
	Set gtmtypfldindx("node_local","gvstats_rec.f_trans_wait")=184
	Set gtmtypes("node_local",185,"name")="node_local.gvstats_rec.f_util_wait"
	Set gtmtypes("node_local",185,"off")=104936
	Set gtmtypes("node_local",185,"len")=8
	Set gtmtypes("node_local",185,"type")="uint64_t"
	Set gtmtypfldindx("node_local","gvstats_rec.f_util_wait")=185
	Set gtmtypes("node_local",186,"name")="node_local.gvstats_rec.f_ws2"
	Set gtmtypes("node_local",186,"off")=104944
	Set gtmtypes("node_local",186,"len")=8
	Set gtmtypes("node_local",186,"type")="uint64_t"
	Set gtmtypfldindx("node_local","gvstats_rec.f_ws2")=186
	Set gtmtypes("node_local",187,"name")="node_local.gvstats_rec.f_ws12"
	Set gtmtypes("node_local",187,"off")=104952
	Set gtmtypes("node_local",187,"len")=8
	Set gtmtypes("node_local",187,"type")="uint64_t"
	Set gtmtypfldindx("node_local","gvstats_rec.f_ws12")=187
	Set gtmtypes("node_local",188,"name")="node_local.gvstats_rec.f_ws15"
	Set gtmtypes("node_local",188,"off")=104960
	Set gtmtypes("node_local",188,"len")=8
	Set gtmtypes("node_local",188,"type")="uint64_t"
	Set gtmtypfldindx("node_local","gvstats_rec.f_ws15")=188
	Set gtmtypes("node_local",189,"name")="node_local.gvstats_rec.f_ws39"
	Set gtmtypes("node_local",189,"off")=104968
	Set gtmtypes("node_local",189,"len")=8
	Set gtmtypes("node_local",189,"type")="uint64_t"
	Set gtmtypfldindx("node_local","gvstats_rec.f_ws39")=189
	Set gtmtypes("node_local",190,"name")="node_local.gvstats_rec.f_ws47"
	Set gtmtypes("node_local",190,"off")=104976
	Set gtmtypes("node_local",190,"len")=8
	Set gtmtypes("node_local",190,"type")="uint64_t"
	Set gtmtypfldindx("node_local","gvstats_rec.f_ws47")=190
	Set gtmtypes("node_local",191,"name")="node_local.last_wcsflu_tn"
	Set gtmtypes("node_local",191,"off")=104984
	Set gtmtypes("node_local",191,"len")=8
	Set gtmtypes("node_local",191,"type")="uint64_t"
	Set gtmtypfldindx("node_local","last_wcsflu_tn")=191
	Set gtmtypes("node_local",192,"name")="node_local.last_wcs_recover_tn"
	Set gtmtypes("node_local",192,"off")=104992
	Set gtmtypes("node_local",192,"len")=8
	Set gtmtypes("node_local",192,"type")="uint64_t"
	Set gtmtypfldindx("node_local","last_wcs_recover_tn")=192
	Set gtmtypes("node_local",193,"name")="node_local.encrypt_glo_buff_off"
	Set gtmtypes("node_local",193,"off")=105000
	Set gtmtypes("node_local",193,"len")=8
	Set gtmtypes("node_local",193,"type")="intptr_t"
	Set gtmtypfldindx("node_local","encrypt_glo_buff_off")=193
	Set gtmtypes("node_local",194,"name")="node_local.snapshot_crit_latch"
	Set gtmtypes("node_local",194,"off")=105008
	Set gtmtypes("node_local",194,"len")=8
	Set gtmtypes("node_local",194,"type")="global_latch_t"
	Set gtmtypfldindx("node_local","snapshot_crit_latch")=194
	Set gtmtypes("node_local",195,"name")="node_local.snapshot_crit_latch.u"
	Set gtmtypes("node_local",195,"off")=105008
	Set gtmtypes("node_local",195,"len")=8
	Set gtmtypes("node_local",195,"type")="union"
	Set gtmtypfldindx("node_local","snapshot_crit_latch.u")=195
	Set gtmtypes("node_local",196,"name")="node_local.snapshot_crit_latch.u.pid_imgcnt"
	Set gtmtypes("node_local",196,"off")=105008
	Set gtmtypes("node_local",196,"len")=8
	Set gtmtypes("node_local",196,"type")="uint64_t"
	Set gtmtypfldindx("node_local","snapshot_crit_latch.u.pid_imgcnt")=196
	Set gtmtypes("node_local",197,"name")="node_local.snapshot_crit_latch.u.parts"
	Set gtmtypes("node_local",197,"off")=105008
	Set gtmtypes("node_local",197,"len")=8
	Set gtmtypes("node_local",197,"type")="struct"
	Set gtmtypfldindx("node_local","snapshot_crit_latch.u.parts")=197
	Set gtmtypes("node_local",198,"name")="node_local.snapshot_crit_latch.u.parts.latch_pid"
	Set gtmtypes("node_local",198,"off")=105008
	Set gtmtypes("node_local",198,"len")=4
	Set gtmtypes("node_local",198,"type")="int"
	Set gtmtypfldindx("node_local","snapshot_crit_latch.u.parts.latch_pid")=198
	Set gtmtypes("node_local",199,"name")="node_local.snapshot_crit_latch.u.parts.latch_word"
	Set gtmtypes("node_local",199,"off")=105012
	Set gtmtypes("node_local",199,"len")=4
	Set gtmtypes("node_local",199,"type")="int"
	Set gtmtypfldindx("node_local","snapshot_crit_latch.u.parts.latch_word")=199
	Set gtmtypes("node_local",200,"name")="node_local.ss_shmid"
	Set gtmtypes("node_local",200,"off")=105016
	Set gtmtypes("node_local",200,"len")=8
	Set gtmtypes("node_local",200,"type")="long"
	Set gtmtypfldindx("node_local","ss_shmid")=200
	Set gtmtypes("node_local",201,"name")="node_local.ss_shmcycle"
	Set gtmtypes("node_local",201,"off")=105024
	Set gtmtypes("node_local",201,"len")=4
	Set gtmtypes("node_local",201,"type")="unsigned-int"
	Set gtmtypfldindx("node_local","ss_shmcycle")=201
	Set gtmtypes("node_local",202,"name")="node_local.snapshot_in_prog"
	Set gtmtypes("node_local",202,"off")=105028
	Set gtmtypes("node_local",202,"len")=4
	Set gtmtypes("node_local",202,"type")="boolean_t"
	Set gtmtypfldindx("node_local","snapshot_in_prog")=202
	Set gtmtypes("node_local",203,"name")="node_local.num_snapshots_in_effect"
	Set gtmtypes("node_local",203,"off")=105032
	Set gtmtypes("node_local",203,"len")=4
	Set gtmtypes("node_local",203,"type")="unsigned-int"
	Set gtmtypfldindx("node_local","num_snapshots_in_effect")=203
	Set gtmtypes("node_local",204,"name")="node_local.wbox_test_seq_num"
	Set gtmtypes("node_local",204,"off")=105036
	Set gtmtypes("node_local",204,"len")=4
	Set gtmtypes("node_local",204,"type")="unsigned-int"
	Set gtmtypfldindx("node_local","wbox_test_seq_num")=204
	Set gtmtypes("node_local",205,"name")="node_local.freeze_online"
	Set gtmtypes("node_local",205,"off")=105040
	Set gtmtypes("node_local",205,"len")=4
	Set gtmtypes("node_local",205,"type")="unsigned-int"
	Set gtmtypfldindx("node_local","freeze_online")=205
	Set gtmtypes("node_local",206,"name")="node_local.kip_pid_array"
	Set gtmtypes("node_local",206,"off")=105044
	Set gtmtypes("node_local",206,"len")=32
	Set gtmtypes("node_local",206,"type")="unsigned-int"
	Set gtmtypfldindx("node_local","kip_pid_array")=206
	Set gtmtypes("node_local",206,"dim")=8
	Set gtmtypes("node_local",207,"name")="node_local.sec_size"
	Set gtmtypes("node_local",207,"off")=105080
	Set gtmtypes("node_local",207,"len")=8
	Set gtmtypes("node_local",207,"type")="uint64_t"
	Set gtmtypfldindx("node_local","sec_size")=207
	Set gtmtypes("node_local",208,"name")="node_local.jnlpool_shmid"
	Set gtmtypes("node_local",208,"off")=105088
	Set gtmtypes("node_local",208,"len")=4
	Set gtmtypes("node_local",208,"type")="int"
	Set gtmtypfldindx("node_local","jnlpool_shmid")=208
	Set gtmtypes("node_local",209,"name")="node_local.trunc_pid"
	Set gtmtypes("node_local",209,"off")=105092
	Set gtmtypes("node_local",209,"len")=4
	Set gtmtypes("node_local",209,"type")="unsigned-int"
	Set gtmtypfldindx("node_local","trunc_pid")=209
	Set gtmtypes("node_local",210,"name")="node_local.highest_lbm_with_busy_blk"
	Set gtmtypes("node_local",210,"off")=105096
	Set gtmtypes("node_local",210,"len")=4
	Set gtmtypes("node_local",210,"type")="int"
	Set gtmtypfldindx("node_local","highest_lbm_with_busy_blk")=210
	Set gtmtypes("node_local",211,"name")="node_local.ftok_ops_array"
	Set gtmtypes("node_local",211,"off")=105104
	Set gtmtypes("node_local",211,"len")=8192
	Set gtmtypes("node_local",211,"type")="ftokhist"
	Set gtmtypfldindx("node_local","ftok_ops_array")=211
	Set gtmtypes("node_local",211,"dim")=512
	Set gtmtypes("node_local",212,"name")="node_local.ftok_ops_array[0].ftok_oper"
	Set gtmtypes("node_local",212,"off")=105104
	Set gtmtypes("node_local",212,"len")=4
	Set gtmtypes("node_local",212,"type")="int"
	Set gtmtypfldindx("node_local","ftok_ops_array[0].ftok_oper")=212
	Set gtmtypes("node_local",213,"name")="node_local.ftok_ops_array[0].process_id"
	Set gtmtypes("node_local",213,"off")=105108
	Set gtmtypes("node_local",213,"len")=4
	Set gtmtypes("node_local",213,"type")="unsigned-int"
	Set gtmtypfldindx("node_local","ftok_ops_array[0].process_id")=213
	Set gtmtypes("node_local",214,"name")="node_local.ftok_ops_array[0].cr_tn"
	Set gtmtypes("node_local",214,"off")=105112
	Set gtmtypes("node_local",214,"len")=8
	Set gtmtypes("node_local",214,"type")="uint64_t"
	Set gtmtypfldindx("node_local","ftok_ops_array[0].cr_tn")=214
	Set gtmtypes("node_local",215,"name")="node_local.root_search_cycle"
	Set gtmtypes("node_local",215,"off")=113296
	Set gtmtypes("node_local",215,"len")=4
	Set gtmtypes("node_local",215,"type")="unsigned-int"
	Set gtmtypfldindx("node_local","root_search_cycle")=215
	Set gtmtypes("node_local",216,"name")="node_local.onln_rlbk_cycle"
	Set gtmtypes("node_local",216,"off")=113300
	Set gtmtypes("node_local",216,"len")=4
	Set gtmtypes("node_local",216,"type")="unsigned-int"
	Set gtmtypfldindx("node_local","onln_rlbk_cycle")=216
	Set gtmtypes("node_local",217,"name")="node_local.db_onln_rlbkd_cycle"
	Set gtmtypes("node_local",217,"off")=113304
	Set gtmtypes("node_local",217,"len")=4
	Set gtmtypes("node_local",217,"type")="unsigned-int"
	Set gtmtypfldindx("node_local","db_onln_rlbkd_cycle")=217
	Set gtmtypes("node_local",218,"name")="node_local.onln_rlbk_pid"
	Set gtmtypes("node_local",218,"off")=113308
	Set gtmtypes("node_local",218,"len")=4
	Set gtmtypes("node_local",218,"type")="unsigned-int"
	Set gtmtypfldindx("node_local","onln_rlbk_pid")=218
	Set gtmtypes("node_local",219,"name")="node_local.dbrndwn_ftok_skip"
	Set gtmtypes("node_local",219,"off")=113312
	Set gtmtypes("node_local",219,"len")=4
	Set gtmtypes("node_local",219,"type")="unsigned-int"
	Set gtmtypfldindx("node_local","dbrndwn_ftok_skip")=219
	Set gtmtypes("node_local",220,"name")="node_local.dbrndwn_access_skip"
	Set gtmtypes("node_local",220,"off")=113316
	Set gtmtypes("node_local",220,"len")=4
	Set gtmtypes("node_local",220,"type")="unsigned-int"
	Set gtmtypfldindx("node_local","dbrndwn_access_skip")=220
	Set gtmtypes("node_local",221,"name")="node_local.fastinteg_in_prog"
	Set gtmtypes("node_local",221,"off")=113320
	Set gtmtypes("node_local",221,"len")=4
	Set gtmtypes("node_local",221,"type")="boolean_t"
	Set gtmtypfldindx("node_local","fastinteg_in_prog")=221
	Set gtmtypes("node_local",222,"name")="node_local.wtstart_errcnt"
	Set gtmtypes("node_local",222,"off")=113324
	Set gtmtypes("node_local",222,"len")=4
	Set gtmtypes("node_local",222,"type")="unsigned-int"
	Set gtmtypfldindx("node_local","wtstart_errcnt")=222
	Set gtmtypes("node_local",223,"name")="node_local.fake_db_enospc"
	Set gtmtypes("node_local",223,"off")=113328
	Set gtmtypes("node_local",223,"len")=4
	Set gtmtypes("node_local",223,"type")="boolean_t"
	Set gtmtypfldindx("node_local","fake_db_enospc")=223
	Set gtmtypes("node_local",224,"name")="node_local.fake_jnl_enospc"
	Set gtmtypes("node_local",224,"off")=113332
	Set gtmtypes("node_local",224,"len")=4
	Set gtmtypes("node_local",224,"type")="boolean_t"
	Set gtmtypfldindx("node_local","fake_jnl_enospc")=224
	Set gtmtypes("node_local",225,"name")="node_local.jnl_writes"
	Set gtmtypes("node_local",225,"off")=113336
	Set gtmtypes("node_local",225,"len")=4
	Set gtmtypes("node_local",225,"type")="unsigned-int"
	Set gtmtypfldindx("node_local","jnl_writes")=225
	Set gtmtypes("node_local",226,"name")="node_local.db_writes"
	Set gtmtypes("node_local",226,"off")=113340
	Set gtmtypes("node_local",226,"len")=4
	Set gtmtypes("node_local",226,"type")="unsigned-int"
	Set gtmtypfldindx("node_local","db_writes")=226
	Set gtmtypes("node_local",227,"name")="node_local.doing_epoch"
	Set gtmtypes("node_local",227,"off")=113344
	Set gtmtypes("node_local",227,"len")=4
	Set gtmtypes("node_local",227,"type")="boolean_t"
	Set gtmtypfldindx("node_local","doing_epoch")=227
	Set gtmtypes("node_local",228,"name")="node_local.epoch_taper_start_dbuffs"
	Set gtmtypes("node_local",228,"off")=113348
	Set gtmtypes("node_local",228,"len")=4
	Set gtmtypes("node_local",228,"type")="unsigned-int"
	Set gtmtypfldindx("node_local","epoch_taper_start_dbuffs")=228
	Set gtmtypes("node_local",229,"name")="node_local.epoch_taper_need_fsync"
	Set gtmtypes("node_local",229,"off")=113352
	Set gtmtypes("node_local",229,"len")=4
	Set gtmtypes("node_local",229,"type")="boolean_t"
	Set gtmtypfldindx("node_local","epoch_taper_need_fsync")=229
	Set gtmtypes("node_local",230,"name")="node_local.wt_pid_array"
	Set gtmtypes("node_local",230,"off")=113356
	Set gtmtypes("node_local",230,"len")=16
	Set gtmtypes("node_local",230,"type")="unsigned-int"
	Set gtmtypfldindx("node_local","wt_pid_array")=230
	Set gtmtypes("node_local",230,"dim")=4
	Set gtmtypes("node_local",231,"name")="node_local.reorg_encrypt_pid"
	Set gtmtypes("node_local",231,"off")=113372
	Set gtmtypes("node_local",231,"len")=4
	Set gtmtypes("node_local",231,"type")="unsigned-int"
	Set gtmtypfldindx("node_local","reorg_encrypt_pid")=231
	Set gtmtypes("node_local",232,"name")="node_local.reorg_encrypt_cycle"
	Set gtmtypes("node_local",232,"off")=113376
	Set gtmtypes("node_local",232,"len")=4
	Set gtmtypes("node_local",232,"type")="unsigned-int"
	Set gtmtypfldindx("node_local","reorg_encrypt_cycle")=232
	Set gtmtypes("node_local",233,"name")="node_local.mupip_extract_count"
	Set gtmtypes("node_local",233,"off")=113380
	Set gtmtypes("node_local",233,"len")=4
	Set gtmtypes("node_local",233,"type")="unsigned-int"
	Set gtmtypfldindx("node_local","mupip_extract_count")=233
	Set gtmtypes("node_local",234,"name")="node_local.saved_acc_meth"
	Set gtmtypes("node_local",234,"off")=113384
	Set gtmtypes("node_local",234,"len")=4
	Set gtmtypes("node_local",234,"type")="int"
	Set gtmtypfldindx("node_local","saved_acc_meth")=234
	Set gtmtypes("node_local",235,"name")="node_local.saved_blk_size"
	Set gtmtypes("node_local",235,"off")=113388
	Set gtmtypes("node_local",235,"len")=4
	Set gtmtypes("node_local",235,"type")="int"
	Set gtmtypfldindx("node_local","saved_blk_size")=235
	Set gtmtypes("node_local",236,"name")="node_local.saved_lock_space_size"
	Set gtmtypes("node_local",236,"off")=113392
	Set gtmtypes("node_local",236,"len")=4
	Set gtmtypes("node_local",236,"type")="unsigned-int"
	Set gtmtypfldindx("node_local","saved_lock_space_size")=236
	Set gtmtypes("node_local",237,"name")="node_local.saved_jnl_buffer_size"
	Set gtmtypes("node_local",237,"off")=113396
	Set gtmtypes("node_local",237,"len")=4
	Set gtmtypes("node_local",237,"type")="int"
	Set gtmtypfldindx("node_local","saved_jnl_buffer_size")=237
	Set gtmtypes("node_local",238,"name")="node_local.update_underway_tn"
	Set gtmtypes("node_local",238,"off")=113400
	Set gtmtypes("node_local",238,"len")=8
	Set gtmtypes("node_local",238,"type")="uint64_t"
	Set gtmtypfldindx("node_local","update_underway_tn")=238
	Set gtmtypes("node_local",239,"name")="node_local.lastwriterbypas_msg_issued"
	Set gtmtypes("node_local",239,"off")=113408
	Set gtmtypes("node_local",239,"len")=4
	Set gtmtypes("node_local",239,"type")="boolean_t"
	Set gtmtypfldindx("node_local","lastwriterbypas_msg_issued")=239
	Set gtmtypes("node_local",240,"name")="node_local.first_writer_seen"
	Set gtmtypes("node_local",240,"off")=113412
	Set gtmtypes("node_local",240,"len")=4
	Set gtmtypes("node_local",240,"type")="boolean_t"
	Set gtmtypfldindx("node_local","first_writer_seen")=240
	Set gtmtypes("node_local",241,"name")="node_local.first_nonbypas_writer_seen"
	Set gtmtypes("node_local",241,"off")=113416
	Set gtmtypes("node_local",241,"len")=4
	Set gtmtypes("node_local",241,"type")="boolean_t"
	Set gtmtypfldindx("node_local","first_nonbypas_writer_seen")=241
	Set gtmtypes("node_local",242,"name")="node_local.ftok_counter_halted"
	Set gtmtypes("node_local",242,"off")=113420
	Set gtmtypes("node_local",242,"len")=4
	Set gtmtypes("node_local",242,"type")="boolean_t"
	Set gtmtypfldindx("node_local","ftok_counter_halted")=242
	Set gtmtypes("node_local",243,"name")="node_local.access_counter_halted"
	Set gtmtypes("node_local",243,"off")=113424
	Set gtmtypes("node_local",243,"len")=4
	Set gtmtypes("node_local",243,"type")="boolean_t"
	Set gtmtypfldindx("node_local","access_counter_halted")=243
	Set gtmtypes("node_local",244,"name")="node_local.statsdb_created"
	Set gtmtypes("node_local",244,"off")=113428
	Set gtmtypes("node_local",244,"len")=4
	Set gtmtypes("node_local",244,"type")="boolean_t"
	Set gtmtypfldindx("node_local","statsdb_created")=244
	Set gtmtypes("node_local",245,"name")="node_local.statsdb_fname_len"
	Set gtmtypes("node_local",245,"off")=113432
	Set gtmtypes("node_local",245,"len")=4
	Set gtmtypes("node_local",245,"type")="unsigned-int"
	Set gtmtypfldindx("node_local","statsdb_fname_len")=245
	Set gtmtypes("node_local",246,"name")="node_local.statsdb_rundown_clean"
	Set gtmtypes("node_local",246,"off")=113436
	Set gtmtypes("node_local",246,"len")=4
	Set gtmtypes("node_local",246,"type")="boolean_t"
	Set gtmtypfldindx("node_local","statsdb_rundown_clean")=246
	Set gtmtypes("node_local",247,"name")="node_local.statsdb_cur_error"
	Set gtmtypes("node_local",247,"off")=113440
	Set gtmtypes("node_local",247,"len")=4
	Set gtmtypes("node_local",247,"type")="int"
	Set gtmtypfldindx("node_local","statsdb_cur_error")=247
	Set gtmtypes("node_local",248,"name")="node_local.statsdb_error_cycle"
	Set gtmtypes("node_local",248,"off")=113444
	Set gtmtypes("node_local",248,"len")=4
	Set gtmtypes("node_local",248,"type")="unsigned-int"
	Set gtmtypfldindx("node_local","statsdb_error_cycle")=248
	Set gtmtypes("node_local",249,"name")="node_local.freeze_latch"
	Set gtmtypes("node_local",249,"off")=113448
	Set gtmtypes("node_local",249,"len")=8
	Set gtmtypes("node_local",249,"type")="global_latch_t"
	Set gtmtypfldindx("node_local","freeze_latch")=249
	Set gtmtypes("node_local",250,"name")="node_local.freeze_latch.u"
	Set gtmtypes("node_local",250,"off")=113448
	Set gtmtypes("node_local",250,"len")=8
	Set gtmtypes("node_local",250,"type")="union"
	Set gtmtypfldindx("node_local","freeze_latch.u")=250
	Set gtmtypes("node_local",251,"name")="node_local.freeze_latch.u.pid_imgcnt"
	Set gtmtypes("node_local",251,"off")=113448
	Set gtmtypes("node_local",251,"len")=8
	Set gtmtypes("node_local",251,"type")="uint64_t"
	Set gtmtypfldindx("node_local","freeze_latch.u.pid_imgcnt")=251
	Set gtmtypes("node_local",252,"name")="node_local.freeze_latch.u.parts"
	Set gtmtypes("node_local",252,"off")=113448
	Set gtmtypes("node_local",252,"len")=8
	Set gtmtypes("node_local",252,"type")="struct"
	Set gtmtypfldindx("node_local","freeze_latch.u.parts")=252
	Set gtmtypes("node_local",253,"name")="node_local.freeze_latch.u.parts.latch_pid"
	Set gtmtypes("node_local",253,"off")=113448
	Set gtmtypes("node_local",253,"len")=4
	Set gtmtypes("node_local",253,"type")="int"
	Set gtmtypfldindx("node_local","freeze_latch.u.parts.latch_pid")=253
	Set gtmtypes("node_local",254,"name")="node_local.freeze_latch.u.parts.latch_word"
	Set gtmtypes("node_local",254,"off")=113452
	Set gtmtypes("node_local",254,"len")=4
	Set gtmtypes("node_local",254,"type")="int"
	Set gtmtypfldindx("node_local","freeze_latch.u.parts.latch_word")=254
	Set gtmtypes("node_local",255,"name")="node_local.wcs_buffs_freed"
	Set gtmtypes("node_local",255,"off")=113456
	Set gtmtypes("node_local",255,"len")=8
	Set gtmtypes("node_local",255,"type")="uint64_t"
	Set gtmtypfldindx("node_local","wcs_buffs_freed")=255
	Set gtmtypes("node_local",256,"name")="node_local.dskspace_next_fire"
	Set gtmtypes("node_local",256,"off")=113464
	Set gtmtypes("node_local",256,"len")=8
	Set gtmtypes("node_local",256,"type")="uint64_t"
	Set gtmtypfldindx("node_local","dskspace_next_fire")=256
	Set gtmtypes("node_local",257,"name")="node_local.lock_crit"
	Set gtmtypes("node_local",257,"off")=113472
	Set gtmtypes("node_local",257,"len")=8
	Set gtmtypes("node_local",257,"type")="global_latch_t"
	Set gtmtypfldindx("node_local","lock_crit")=257
	Set gtmtypes("node_local",258,"name")="node_local.lock_crit.u"
	Set gtmtypes("node_local",258,"off")=113472
	Set gtmtypes("node_local",258,"len")=8
	Set gtmtypes("node_local",258,"type")="union"
	Set gtmtypfldindx("node_local","lock_crit.u")=258
	Set gtmtypes("node_local",259,"name")="node_local.lock_crit.u.pid_imgcnt"
	Set gtmtypes("node_local",259,"off")=113472
	Set gtmtypes("node_local",259,"len")=8
	Set gtmtypes("node_local",259,"type")="uint64_t"
	Set gtmtypfldindx("node_local","lock_crit.u.pid_imgcnt")=259
	Set gtmtypes("node_local",260,"name")="node_local.lock_crit.u.parts"
	Set gtmtypes("node_local",260,"off")=113472
	Set gtmtypes("node_local",260,"len")=8
	Set gtmtypes("node_local",260,"type")="struct"
	Set gtmtypfldindx("node_local","lock_crit.u.parts")=260
	Set gtmtypes("node_local",261,"name")="node_local.lock_crit.u.parts.latch_pid"
	Set gtmtypes("node_local",261,"off")=113472
	Set gtmtypes("node_local",261,"len")=4
	Set gtmtypes("node_local",261,"type")="int"
	Set gtmtypfldindx("node_local","lock_crit.u.parts.latch_pid")=261
	Set gtmtypes("node_local",262,"name")="node_local.lock_crit.u.parts.latch_word"
	Set gtmtypes("node_local",262,"off")=113476
	Set gtmtypes("node_local",262,"len")=4
	Set gtmtypes("node_local",262,"type")="int"
	Set gtmtypfldindx("node_local","lock_crit.u.parts.latch_word")=262
	Set gtmtypes("node_local",263,"name")="node_local.tp_hint"
	Set gtmtypes("node_local",263,"off")=113480
	Set gtmtypes("node_local",263,"len")=4
	Set gtmtypes("node_local",263,"type")="int"
	Set gtmtypfldindx("node_local","tp_hint")=263
	;
	Set gtmtypes("noisolation_element")="struct"
	Set gtmtypes("noisolation_element",0)=2
	Set gtmtypes("noisolation_element","len")=16
	Set gtmtypes("noisolation_element",1,"name")="noisolation_element.gvnh"
	Set gtmtypes("noisolation_element",1,"off")=0
	Set gtmtypes("noisolation_element",1,"len")=8
	Set gtmtypes("noisolation_element",1,"type")="addr"
	Set gtmtypfldindx("noisolation_element","gvnh")=1
	Set gtmtypes("noisolation_element",2,"name")="noisolation_element.next"
	Set gtmtypes("noisolation_element",2,"off")=8
	Set gtmtypes("noisolation_element",2,"len")=8
	Set gtmtypes("noisolation_element",2,"type")="addr"
	Set gtmtypfldindx("noisolation_element","next")=2
	;
	Set gtmtypes("noisolation_list")="struct"
	Set gtmtypes("noisolation_list",0)=2
	Set gtmtypes("noisolation_list","len")=16
	Set gtmtypes("noisolation_list",1,"name")="noisolation_list.type"
	Set gtmtypes("noisolation_list",1,"off")=0
	Set gtmtypes("noisolation_list",1,"len")=4
	Set gtmtypes("noisolation_list",1,"type")="int"
	Set gtmtypfldindx("noisolation_list","type")=1
	Set gtmtypes("noisolation_list",2,"name")="noisolation_list.gvnh_list"
	Set gtmtypes("noisolation_list",2,"off")=8
	Set gtmtypes("noisolation_list",2,"len")=8
	Set gtmtypes("noisolation_list",2,"type")="addr"
	Set gtmtypfldindx("noisolation_list","gvnh_list")=2
	;
	Set gtmtypes("non_native_uint8")="struct"
	Set gtmtypes("non_native_uint8",0)=1
	Set gtmtypes("non_native_uint8","len")=8
	Set gtmtypes("non_native_uint8",1,"name")="non_native_uint8.value"
	Set gtmtypes("non_native_uint8",1,"off")=0
	Set gtmtypes("non_native_uint8",1,"len")=8
	Set gtmtypes("non_native_uint8",1,"type")="unsigned-int"
	Set gtmtypfldindx("non_native_uint8","value")=1
	Set gtmtypes("non_native_uint8",1,"dim")=2
	;
	Set gtmtypes("octabstruct")="struct"
	Set gtmtypes("octabstruct",0)=1
	Set gtmtypes("octabstruct","len")=4
	Set gtmtypes("octabstruct",1,"name")="octabstruct.octype"
	Set gtmtypes("octabstruct",1,"off")=0
	Set gtmtypes("octabstruct",1,"len")=4
	Set gtmtypes("octabstruct",1,"type")="unsigned-int"
	Set gtmtypfldindx("octabstruct","octype")=1
	;
	Set gtmtypes("omi_cl_stat")="struct"
	Set gtmtypes("omi_cl_stat",0)=3
	Set gtmtypes("omi_cl_stat","len")=12
	Set gtmtypes("omi_cl_stat",1,"name")="omi_cl_stat.conn"
	Set gtmtypes("omi_cl_stat",1,"off")=0
	Set gtmtypes("omi_cl_stat",1,"len")=4
	Set gtmtypes("omi_cl_stat",1,"type")="int"
	Set gtmtypfldindx("omi_cl_stat","conn")=1
	Set gtmtypes("omi_cl_stat",2,"name")="omi_cl_stat.disc"
	Set gtmtypes("omi_cl_stat",2,"off")=4
	Set gtmtypes("omi_cl_stat",2,"len")=4
	Set gtmtypes("omi_cl_stat",2,"type")="int"
	Set gtmtypfldindx("omi_cl_stat","disc")=2
	Set gtmtypes("omi_cl_stat",3,"name")="omi_cl_stat.clos"
	Set gtmtypes("omi_cl_stat",3,"off")=8
	Set gtmtypes("omi_cl_stat",3,"len")=4
	Set gtmtypes("omi_cl_stat",3,"type")="int"
	Set gtmtypfldindx("omi_cl_stat","clos")=3
	;
	Set gtmtypes("omi_cn_stat")="struct"
	Set gtmtypes("omi_cn_stat",0)=8
	Set gtmtypes("omi_cn_stat","len")=440
	Set gtmtypes("omi_cn_stat",1,"name")="omi_cn_stat.id"
	Set gtmtypes("omi_cn_stat",1,"off")=0
	Set gtmtypes("omi_cn_stat",1,"len")=4
	Set gtmtypes("omi_cn_stat",1,"type")="int"
	Set gtmtypfldindx("omi_cn_stat","id")=1
	Set gtmtypes("omi_cn_stat",2,"name")="omi_cn_stat.start"
	Set gtmtypes("omi_cn_stat",2,"off")=8
	Set gtmtypes("omi_cn_stat",2,"len")=8
	Set gtmtypes("omi_cn_stat",2,"type")="time_t"
	Set gtmtypfldindx("omi_cn_stat","start")=2
	Set gtmtypes("omi_cn_stat",3,"name")="omi_cn_stat.ai"
	Set gtmtypes("omi_cn_stat",3,"off")=16
	Set gtmtypes("omi_cn_stat",3,"len")=48
	Set gtmtypes("omi_cn_stat",3,"type")="struct"
	Set gtmtypfldindx("omi_cn_stat","ai")=3
	Set gtmtypes("omi_cn_stat",4,"name")="omi_cn_stat.sas"
	Set gtmtypes("omi_cn_stat",4,"off")=64
	Set gtmtypes("omi_cn_stat",4,"len")=128
	Set gtmtypes("omi_cn_stat",4,"type")="struct"
	Set gtmtypfldindx("omi_cn_stat","sas")=4
	Set gtmtypes("omi_cn_stat",5,"name")="omi_cn_stat.bytes_recv"
	Set gtmtypes("omi_cn_stat",5,"off")=192
	Set gtmtypes("omi_cn_stat",5,"len")=4
	Set gtmtypes("omi_cn_stat",5,"type")="int"
	Set gtmtypfldindx("omi_cn_stat","bytes_recv")=5
	Set gtmtypes("omi_cn_stat",6,"name")="omi_cn_stat.bytes_send"
	Set gtmtypes("omi_cn_stat",6,"off")=196
	Set gtmtypes("omi_cn_stat",6,"len")=4
	Set gtmtypes("omi_cn_stat",6,"type")="int"
	Set gtmtypfldindx("omi_cn_stat","bytes_send")=6
	Set gtmtypes("omi_cn_stat",7,"name")="omi_cn_stat.xact"
	Set gtmtypes("omi_cn_stat",7,"off")=200
	Set gtmtypes("omi_cn_stat",7,"len")=136
	Set gtmtypes("omi_cn_stat",7,"type")="int"
	Set gtmtypfldindx("omi_cn_stat","xact")=7
	Set gtmtypes("omi_cn_stat",7,"dim")=34
	Set gtmtypes("omi_cn_stat",8,"name")="omi_cn_stat.errs"
	Set gtmtypes("omi_cn_stat",8,"off")=336
	Set gtmtypes("omi_cn_stat",8,"len")=100
	Set gtmtypes("omi_cn_stat",8,"type")="int"
	Set gtmtypfldindx("omi_cn_stat","errs")=8
	Set gtmtypes("omi_cn_stat",8,"dim")=25
	;
	Set gtmtypes("omi_conn_ll")="struct"
	Set gtmtypes("omi_conn_ll",0)=16
	Set gtmtypes("omi_conn_ll","len")=480
	Set gtmtypes("omi_conn_ll",1,"name")="omi_conn_ll.nve"
	Set gtmtypes("omi_conn_ll",1,"off")=0
	Set gtmtypes("omi_conn_ll",1,"len")=4
	Set gtmtypes("omi_conn_ll",1,"type")="int"
	Set gtmtypfldindx("omi_conn_ll","nve")=1
	Set gtmtypes("omi_conn_ll",2,"name")="omi_conn_ll.head"
	Set gtmtypes("omi_conn_ll",2,"off")=8
	Set gtmtypes("omi_conn_ll",2,"len")=8
	Set gtmtypes("omi_conn_ll",2,"type")="addr"
	Set gtmtypfldindx("omi_conn_ll","head")=2
	Set gtmtypes("omi_conn_ll",3,"name")="omi_conn_ll.tail"
	Set gtmtypes("omi_conn_ll",3,"off")=16
	Set gtmtypes("omi_conn_ll",3,"len")=8
	Set gtmtypes("omi_conn_ll",3,"type")="addr"
	Set gtmtypfldindx("omi_conn_ll","tail")=3
	Set gtmtypes("omi_conn_ll",4,"name")="omi_conn_ll.stats"
	Set gtmtypes("omi_conn_ll",4,"off")=24
	Set gtmtypes("omi_conn_ll",4,"len")=12
	Set gtmtypes("omi_conn_ll",4,"type")="omi_cl_stat"
	Set gtmtypfldindx("omi_conn_ll","stats")=4
	Set gtmtypes("omi_conn_ll",5,"name")="omi_conn_ll.stats.conn"
	Set gtmtypes("omi_conn_ll",5,"off")=24
	Set gtmtypes("omi_conn_ll",5,"len")=4
	Set gtmtypes("omi_conn_ll",5,"type")="int"
	Set gtmtypfldindx("omi_conn_ll","stats.conn")=5
	Set gtmtypes("omi_conn_ll",6,"name")="omi_conn_ll.stats.disc"
	Set gtmtypes("omi_conn_ll",6,"off")=28
	Set gtmtypes("omi_conn_ll",6,"len")=4
	Set gtmtypes("omi_conn_ll",6,"type")="int"
	Set gtmtypfldindx("omi_conn_ll","stats.disc")=6
	Set gtmtypes("omi_conn_ll",7,"name")="omi_conn_ll.stats.clos"
	Set gtmtypes("omi_conn_ll",7,"off")=32
	Set gtmtypes("omi_conn_ll",7,"len")=4
	Set gtmtypes("omi_conn_ll",7,"type")="int"
	Set gtmtypfldindx("omi_conn_ll","stats.clos")=7
	Set gtmtypes("omi_conn_ll",8,"name")="omi_conn_ll.st_cn"
	Set gtmtypes("omi_conn_ll",8,"off")=40
	Set gtmtypes("omi_conn_ll",8,"len")=440
	Set gtmtypes("omi_conn_ll",8,"type")="omi_cn_stat"
	Set gtmtypfldindx("omi_conn_ll","st_cn")=8
	Set gtmtypes("omi_conn_ll",9,"name")="omi_conn_ll.st_cn.id"
	Set gtmtypes("omi_conn_ll",9,"off")=40
	Set gtmtypes("omi_conn_ll",9,"len")=4
	Set gtmtypes("omi_conn_ll",9,"type")="int"
	Set gtmtypfldindx("omi_conn_ll","st_cn.id")=9
	Set gtmtypes("omi_conn_ll",10,"name")="omi_conn_ll.st_cn.start"
	Set gtmtypes("omi_conn_ll",10,"off")=48
	Set gtmtypes("omi_conn_ll",10,"len")=8
	Set gtmtypes("omi_conn_ll",10,"type")="time_t"
	Set gtmtypfldindx("omi_conn_ll","st_cn.start")=10
	Set gtmtypes("omi_conn_ll",11,"name")="omi_conn_ll.st_cn.ai"
	Set gtmtypes("omi_conn_ll",11,"off")=56
	Set gtmtypes("omi_conn_ll",11,"len")=48
	Set gtmtypes("omi_conn_ll",11,"type")="struct"
	Set gtmtypfldindx("omi_conn_ll","st_cn.ai")=11
	Set gtmtypes("omi_conn_ll",12,"name")="omi_conn_ll.st_cn.sas"
	Set gtmtypes("omi_conn_ll",12,"off")=104
	Set gtmtypes("omi_conn_ll",12,"len")=128
	Set gtmtypes("omi_conn_ll",12,"type")="struct"
	Set gtmtypfldindx("omi_conn_ll","st_cn.sas")=12
	Set gtmtypes("omi_conn_ll",13,"name")="omi_conn_ll.st_cn.bytes_recv"
	Set gtmtypes("omi_conn_ll",13,"off")=232
	Set gtmtypes("omi_conn_ll",13,"len")=4
	Set gtmtypes("omi_conn_ll",13,"type")="int"
	Set gtmtypfldindx("omi_conn_ll","st_cn.bytes_recv")=13
	Set gtmtypes("omi_conn_ll",14,"name")="omi_conn_ll.st_cn.bytes_send"
	Set gtmtypes("omi_conn_ll",14,"off")=236
	Set gtmtypes("omi_conn_ll",14,"len")=4
	Set gtmtypes("omi_conn_ll",14,"type")="int"
	Set gtmtypfldindx("omi_conn_ll","st_cn.bytes_send")=14
	Set gtmtypes("omi_conn_ll",15,"name")="omi_conn_ll.st_cn.xact"
	Set gtmtypes("omi_conn_ll",15,"off")=240
	Set gtmtypes("omi_conn_ll",15,"len")=136
	Set gtmtypes("omi_conn_ll",15,"type")="int"
	Set gtmtypfldindx("omi_conn_ll","st_cn.xact")=15
	Set gtmtypes("omi_conn_ll",15,"dim")=34
	Set gtmtypes("omi_conn_ll",16,"name")="omi_conn_ll.st_cn.errs"
	Set gtmtypes("omi_conn_ll",16,"off")=376
	Set gtmtypes("omi_conn_ll",16,"len")=100
	Set gtmtypes("omi_conn_ll",16,"type")="int"
	Set gtmtypfldindx("omi_conn_ll","st_cn.errs")=16
	Set gtmtypes("omi_conn_ll",16,"dim")=25
	;
	Set gtmtypes("omi_err_hdr")="struct"
	Set gtmtypes("omi_err_hdr",0)=3
	Set gtmtypes("omi_err_hdr","len")=6
	Set gtmtypes("omi_err_hdr",1,"name")="omi_err_hdr.class"
	Set gtmtypes("omi_err_hdr",1,"off")=0
	Set gtmtypes("omi_err_hdr",1,"len")=2
	Set gtmtypes("omi_err_hdr",1,"type")="unsigned-short"
	Set gtmtypfldindx("omi_err_hdr","class")=1
	Set gtmtypes("omi_err_hdr",2,"name")="omi_err_hdr.type"
	Set gtmtypes("omi_err_hdr",2,"off")=2
	Set gtmtypes("omi_err_hdr",2,"len")=1
	Set gtmtypes("omi_err_hdr",2,"type")="unsigned-char"
	Set gtmtypfldindx("omi_err_hdr","type")=2
	Set gtmtypes("omi_err_hdr",3,"name")="omi_err_hdr.modifier"
	Set gtmtypes("omi_err_hdr",3,"off")=4
	Set gtmtypes("omi_err_hdr",3,"len")=2
	Set gtmtypes("omi_err_hdr",3,"type")="unsigned-short"
	Set gtmtypfldindx("omi_err_hdr","modifier")=3
	;
	Set gtmtypes("omi_hist_rec")="struct"
	Set gtmtypes("omi_hist_rec",0)=7
	Set gtmtypes("omi_hist_rec","len")=16416
	Set gtmtypes("omi_hist_rec",1,"name")="omi_hist_rec.conn"
	Set gtmtypes("omi_hist_rec",1,"off")=0
	Set gtmtypes("omi_hist_rec",1,"len")=4
	Set gtmtypes("omi_hist_rec",1,"type")="int"
	Set gtmtypfldindx("omi_hist_rec","conn")=1
	Set gtmtypes("omi_hist_rec",2,"name")="omi_hist_rec.timestamp"
	Set gtmtypes("omi_hist_rec",2,"off")=8
	Set gtmtypes("omi_hist_rec",2,"len")=8
	Set gtmtypes("omi_hist_rec",2,"type")="time_t"
	Set gtmtypfldindx("omi_hist_rec","timestamp")=2
	Set gtmtypes("omi_hist_rec",3,"name")="omi_hist_rec.toobigflag"
	Set gtmtypes("omi_hist_rec",3,"off")=16
	Set gtmtypes("omi_hist_rec",3,"len")=4
	Set gtmtypes("omi_hist_rec",3,"type")="int"
	Set gtmtypfldindx("omi_hist_rec","toobigflag")=3
	Set gtmtypes("omi_hist_rec",4,"name")="omi_hist_rec.req"
	Set gtmtypes("omi_hist_rec",4,"off")=20
	Set gtmtypes("omi_hist_rec",4,"len")=8192
	Set gtmtypes("omi_hist_rec",4,"type")="char"
	Set gtmtypfldindx("omi_hist_rec","req")=4
	Set gtmtypes("omi_hist_rec",5,"name")="omi_hist_rec.rsp"
	Set gtmtypes("omi_hist_rec",5,"off")=8212
	Set gtmtypes("omi_hist_rec",5,"len")=8192
	Set gtmtypes("omi_hist_rec",5,"type")="char"
	Set gtmtypfldindx("omi_hist_rec","rsp")=5
	Set gtmtypes("omi_hist_rec",6,"name")="omi_hist_rec.req_len"
	Set gtmtypes("omi_hist_rec",6,"off")=16404
	Set gtmtypes("omi_hist_rec",6,"len")=4
	Set gtmtypes("omi_hist_rec",6,"type")="int"
	Set gtmtypfldindx("omi_hist_rec","req_len")=6
	Set gtmtypes("omi_hist_rec",7,"name")="omi_hist_rec.rsp_len"
	Set gtmtypes("omi_hist_rec",7,"off")=16408
	Set gtmtypes("omi_hist_rec",7,"len")=4
	Set gtmtypes("omi_hist_rec",7,"type")="int"
	Set gtmtypfldindx("omi_hist_rec","rsp_len")=7
	;
	Set gtmtypes("omi_li")="union"
	Set gtmtypes("omi_li",0)=2
	Set gtmtypes("omi_li","len")=2
	Set gtmtypes("omi_li",1,"name")="omi_li.octet"
	Set gtmtypes("omi_li",1,"off")=0
	Set gtmtypes("omi_li",1,"len")=2
	Set gtmtypes("omi_li",1,"type")="char"
	Set gtmtypfldindx("omi_li","octet")=1
	Set gtmtypes("omi_li",2,"name")="omi_li.value"
	Set gtmtypes("omi_li",2,"off")=0
	Set gtmtypes("omi_li",2,"len")=2
	Set gtmtypes("omi_li",2,"type")="unsigned-short"
	Set gtmtypfldindx("omi_li","value")=2
	;
	Set gtmtypes("omi_req_hdr")="struct"
	Set gtmtypes("omi_req_hdr",0)=18
	Set gtmtypes("omi_req_hdr","len")=12
	Set gtmtypes("omi_req_hdr",1,"name")="omi_req_hdr.op_class"
	Set gtmtypes("omi_req_hdr",1,"off")=0
	Set gtmtypes("omi_req_hdr",1,"len")=2
	Set gtmtypes("omi_req_hdr",1,"type")="omi_li"
	Set gtmtypfldindx("omi_req_hdr","op_class")=1
	Set gtmtypes("omi_req_hdr",2,"name")="omi_req_hdr.op_class.octet"
	Set gtmtypes("omi_req_hdr",2,"off")=0
	Set gtmtypes("omi_req_hdr",2,"len")=2
	Set gtmtypes("omi_req_hdr",2,"type")="char"
	Set gtmtypfldindx("omi_req_hdr","op_class.octet")=2
	Set gtmtypes("omi_req_hdr",3,"name")="omi_req_hdr.op_class.value"
	Set gtmtypes("omi_req_hdr",3,"off")=0
	Set gtmtypes("omi_req_hdr",3,"len")=2
	Set gtmtypes("omi_req_hdr",3,"type")="unsigned-short"
	Set gtmtypfldindx("omi_req_hdr","op_class.value")=3
	Set gtmtypes("omi_req_hdr",4,"name")="omi_req_hdr.op_type"
	Set gtmtypes("omi_req_hdr",4,"off")=2
	Set gtmtypes("omi_req_hdr",4,"len")=1
	Set gtmtypes("omi_req_hdr",4,"type")="omi_si"
	Set gtmtypfldindx("omi_req_hdr","op_type")=4
	Set gtmtypes("omi_req_hdr",5,"name")="omi_req_hdr.op_type.octet"
	Set gtmtypes("omi_req_hdr",5,"off")=2
	Set gtmtypes("omi_req_hdr",5,"len")=1
	Set gtmtypes("omi_req_hdr",5,"type")="char"
	Set gtmtypfldindx("omi_req_hdr","op_type.octet")=5
	Set gtmtypes("omi_req_hdr",6,"name")="omi_req_hdr.op_type.value"
	Set gtmtypes("omi_req_hdr",6,"off")=2
	Set gtmtypes("omi_req_hdr",6,"len")=1
	Set gtmtypes("omi_req_hdr",6,"type")="unsigned-char"
	Set gtmtypfldindx("omi_req_hdr","op_type.value")=6
	Set gtmtypes("omi_req_hdr",7,"name")="omi_req_hdr.user"
	Set gtmtypes("omi_req_hdr",7,"off")=4
	Set gtmtypes("omi_req_hdr",7,"len")=2
	Set gtmtypes("omi_req_hdr",7,"type")="omi_li"
	Set gtmtypfldindx("omi_req_hdr","user")=7
	Set gtmtypes("omi_req_hdr",8,"name")="omi_req_hdr.user.octet"
	Set gtmtypes("omi_req_hdr",8,"off")=4
	Set gtmtypes("omi_req_hdr",8,"len")=2
	Set gtmtypes("omi_req_hdr",8,"type")="char"
	Set gtmtypfldindx("omi_req_hdr","user.octet")=8
	Set gtmtypes("omi_req_hdr",9,"name")="omi_req_hdr.user.value"
	Set gtmtypes("omi_req_hdr",9,"off")=4
	Set gtmtypes("omi_req_hdr",9,"len")=2
	Set gtmtypes("omi_req_hdr",9,"type")="unsigned-short"
	Set gtmtypfldindx("omi_req_hdr","user.value")=9
	Set gtmtypes("omi_req_hdr",10,"name")="omi_req_hdr.group"
	Set gtmtypes("omi_req_hdr",10,"off")=6
	Set gtmtypes("omi_req_hdr",10,"len")=2
	Set gtmtypes("omi_req_hdr",10,"type")="omi_li"
	Set gtmtypfldindx("omi_req_hdr","group")=10
	Set gtmtypes("omi_req_hdr",11,"name")="omi_req_hdr.group.octet"
	Set gtmtypes("omi_req_hdr",11,"off")=6
	Set gtmtypes("omi_req_hdr",11,"len")=2
	Set gtmtypes("omi_req_hdr",11,"type")="char"
	Set gtmtypfldindx("omi_req_hdr","group.octet")=11
	Set gtmtypes("omi_req_hdr",12,"name")="omi_req_hdr.group.value"
	Set gtmtypes("omi_req_hdr",12,"off")=6
	Set gtmtypes("omi_req_hdr",12,"len")=2
	Set gtmtypes("omi_req_hdr",12,"type")="unsigned-short"
	Set gtmtypfldindx("omi_req_hdr","group.value")=12
	Set gtmtypes("omi_req_hdr",13,"name")="omi_req_hdr.seq"
	Set gtmtypes("omi_req_hdr",13,"off")=8
	Set gtmtypes("omi_req_hdr",13,"len")=2
	Set gtmtypes("omi_req_hdr",13,"type")="omi_li"
	Set gtmtypfldindx("omi_req_hdr","seq")=13
	Set gtmtypes("omi_req_hdr",14,"name")="omi_req_hdr.seq.octet"
	Set gtmtypes("omi_req_hdr",14,"off")=8
	Set gtmtypes("omi_req_hdr",14,"len")=2
	Set gtmtypes("omi_req_hdr",14,"type")="char"
	Set gtmtypfldindx("omi_req_hdr","seq.octet")=14
	Set gtmtypes("omi_req_hdr",15,"name")="omi_req_hdr.seq.value"
	Set gtmtypes("omi_req_hdr",15,"off")=8
	Set gtmtypes("omi_req_hdr",15,"len")=2
	Set gtmtypes("omi_req_hdr",15,"type")="unsigned-short"
	Set gtmtypfldindx("omi_req_hdr","seq.value")=15
	Set gtmtypes("omi_req_hdr",16,"name")="omi_req_hdr.ref"
	Set gtmtypes("omi_req_hdr",16,"off")=10
	Set gtmtypes("omi_req_hdr",16,"len")=2
	Set gtmtypes("omi_req_hdr",16,"type")="omi_li"
	Set gtmtypfldindx("omi_req_hdr","ref")=16
	Set gtmtypes("omi_req_hdr",17,"name")="omi_req_hdr.ref.octet"
	Set gtmtypes("omi_req_hdr",17,"off")=10
	Set gtmtypes("omi_req_hdr",17,"len")=2
	Set gtmtypes("omi_req_hdr",17,"type")="char"
	Set gtmtypfldindx("omi_req_hdr","ref.octet")=17
	Set gtmtypes("omi_req_hdr",18,"name")="omi_req_hdr.ref.value"
	Set gtmtypes("omi_req_hdr",18,"off")=10
	Set gtmtypes("omi_req_hdr",18,"len")=2
	Set gtmtypes("omi_req_hdr",18,"type")="unsigned-short"
	Set gtmtypfldindx("omi_req_hdr","ref.value")=18
	;
	Set gtmtypes("omi_si")="union"
	Set gtmtypes("omi_si",0)=2
	Set gtmtypes("omi_si","len")=1
	Set gtmtypes("omi_si",1,"name")="omi_si.octet"
	Set gtmtypes("omi_si",1,"off")=0
	Set gtmtypes("omi_si",1,"len")=1
	Set gtmtypes("omi_si",1,"type")="char"
	Set gtmtypfldindx("omi_si","octet")=1
	Set gtmtypes("omi_si",2,"name")="omi_si.value"
	Set gtmtypes("omi_si",2,"off")=0
	Set gtmtypes("omi_si",2,"len")=1
	Set gtmtypes("omi_si",2,"type")="unsigned-char"
	Set gtmtypfldindx("omi_si","value")=2
	;
	Set gtmtypes("omi_vi")="union"
	Set gtmtypes("omi_vi",0)=2
	Set gtmtypes("omi_vi","len")=4
	Set gtmtypes("omi_vi",1,"name")="omi_vi.octet"
	Set gtmtypes("omi_vi",1,"off")=0
	Set gtmtypes("omi_vi",1,"len")=4
	Set gtmtypes("omi_vi",1,"type")="char"
	Set gtmtypfldindx("omi_vi","octet")=1
	Set gtmtypes("omi_vi",2,"name")="omi_vi.value"
	Set gtmtypes("omi_vi",2,"off")=0
	Set gtmtypes("omi_vi",2,"len")=4
	Set gtmtypes("omi_vi",2,"type")="unsigned-int"
	Set gtmtypfldindx("omi_vi","value")=2
	;
	Set gtmtypes("onln_rlbk_reg_list")="struct"
	Set gtmtypes("onln_rlbk_reg_list",0)=7
	Set gtmtypes("onln_rlbk_reg_list","len")=48
	Set gtmtypes("onln_rlbk_reg_list",1,"name")="onln_rlbk_reg_list.fPtr"
	Set gtmtypes("onln_rlbk_reg_list",1,"off")=0
	Set gtmtypes("onln_rlbk_reg_list",1,"len")=8
	Set gtmtypes("onln_rlbk_reg_list",1,"type")="addr"
	Set gtmtypfldindx("onln_rlbk_reg_list","fPtr")=1
	Set gtmtypes("onln_rlbk_reg_list",2,"name")="onln_rlbk_reg_list.reg"
	Set gtmtypes("onln_rlbk_reg_list",2,"off")=8
	Set gtmtypes("onln_rlbk_reg_list",2,"len")=8
	Set gtmtypes("onln_rlbk_reg_list",2,"type")="addr"
	Set gtmtypfldindx("onln_rlbk_reg_list","reg")=2
	Set gtmtypes("onln_rlbk_reg_list",3,"name")="onln_rlbk_reg_list.unique_file_id"
	Set gtmtypes("onln_rlbk_reg_list",3,"off")=16
	Set gtmtypes("onln_rlbk_reg_list",3,"len")=16
	Set gtmtypes("onln_rlbk_reg_list",3,"type")="unix_file_id"
	Set gtmtypfldindx("onln_rlbk_reg_list","unique_file_id")=3
	Set gtmtypes("onln_rlbk_reg_list",4,"name")="onln_rlbk_reg_list.unique_file_id.inode"
	Set gtmtypes("onln_rlbk_reg_list",4,"off")=16
	Set gtmtypes("onln_rlbk_reg_list",4,"len")=8
	Set gtmtypes("onln_rlbk_reg_list",4,"type")="ino_t"
	Set gtmtypfldindx("onln_rlbk_reg_list","unique_file_id.inode")=4
	Set gtmtypes("onln_rlbk_reg_list",5,"name")="onln_rlbk_reg_list.unique_file_id.device"
	Set gtmtypes("onln_rlbk_reg_list",5,"off")=24
	Set gtmtypes("onln_rlbk_reg_list",5,"len")=8
	Set gtmtypes("onln_rlbk_reg_list",5,"type")="dev_t"
	Set gtmtypfldindx("onln_rlbk_reg_list","unique_file_id.device")=5
	Set gtmtypes("onln_rlbk_reg_list",6,"name")="onln_rlbk_reg_list.fid_index"
	Set gtmtypes("onln_rlbk_reg_list",6,"off")=32
	Set gtmtypes("onln_rlbk_reg_list",6,"len")=4
	Set gtmtypes("onln_rlbk_reg_list",6,"type")="int"
	Set gtmtypfldindx("onln_rlbk_reg_list","fid_index")=6
	Set gtmtypes("onln_rlbk_reg_list",7,"name")="onln_rlbk_reg_list.rctl"
	Set gtmtypes("onln_rlbk_reg_list",7,"off")=40
	Set gtmtypes("onln_rlbk_reg_list",7,"len")=8
	Set gtmtypes("onln_rlbk_reg_list",7,"type")="addr"
	Set gtmtypfldindx("onln_rlbk_reg_list","rctl")=7
	;
	Set gtmtypes("open_relinkctl_sgm")="struct"
	Set gtmtypes("open_relinkctl_sgm",0)=16
	Set gtmtypes("open_relinkctl_sgm","len")=536
	Set gtmtypes("open_relinkctl_sgm",1,"name")="open_relinkctl_sgm.next"
	Set gtmtypes("open_relinkctl_sgm",1,"off")=0
	Set gtmtypes("open_relinkctl_sgm",1,"len")=8
	Set gtmtypes("open_relinkctl_sgm",1,"type")="addr"
	Set gtmtypfldindx("open_relinkctl_sgm","next")=1
	Set gtmtypes("open_relinkctl_sgm",2,"name")="open_relinkctl_sgm.zro_entry_name"
	Set gtmtypes("open_relinkctl_sgm",2,"off")=8
	Set gtmtypes("open_relinkctl_sgm",2,"len")=16
	Set gtmtypes("open_relinkctl_sgm",2,"type")="mstr"
	Set gtmtypfldindx("open_relinkctl_sgm","zro_entry_name")=2
	Set gtmtypes("open_relinkctl_sgm",3,"name")="open_relinkctl_sgm.zro_entry_name.char_len"
	Set gtmtypes("open_relinkctl_sgm",3,"off")=8
	Set gtmtypes("open_relinkctl_sgm",3,"len")=4
	Set gtmtypes("open_relinkctl_sgm",3,"type")="unsigned-int"
	Set gtmtypfldindx("open_relinkctl_sgm","zro_entry_name.char_len")=3
	Set gtmtypes("open_relinkctl_sgm",4,"name")="open_relinkctl_sgm.zro_entry_name.len"
	Set gtmtypes("open_relinkctl_sgm",4,"off")=12
	Set gtmtypes("open_relinkctl_sgm",4,"len")=4
	Set gtmtypes("open_relinkctl_sgm",4,"type")="int"
	Set gtmtypfldindx("open_relinkctl_sgm","zro_entry_name.len")=4
	Set gtmtypes("open_relinkctl_sgm",5,"name")="open_relinkctl_sgm.zro_entry_name.addr"
	Set gtmtypes("open_relinkctl_sgm",5,"off")=16
	Set gtmtypes("open_relinkctl_sgm",5,"len")=8
	Set gtmtypes("open_relinkctl_sgm",5,"type")="addr"
	Set gtmtypfldindx("open_relinkctl_sgm","zro_entry_name.addr")=5
	Set gtmtypes("open_relinkctl_sgm",6,"name")="open_relinkctl_sgm.relinkctl_path"
	Set gtmtypes("open_relinkctl_sgm",6,"off")=24
	Set gtmtypes("open_relinkctl_sgm",6,"len")=8
	Set gtmtypes("open_relinkctl_sgm",6,"type")="addr"
	Set gtmtypfldindx("open_relinkctl_sgm","relinkctl_path")=6
	Set gtmtypes("open_relinkctl_sgm",7,"name")="open_relinkctl_sgm.n_records"
	Set gtmtypes("open_relinkctl_sgm",7,"off")=32
	Set gtmtypes("open_relinkctl_sgm",7,"len")=4
	Set gtmtypes("open_relinkctl_sgm",7,"type")="unsigned-int"
	Set gtmtypfldindx("open_relinkctl_sgm","n_records")=7
	Set gtmtypes("open_relinkctl_sgm",8,"name")="open_relinkctl_sgm.locked"
	Set gtmtypes("open_relinkctl_sgm",8,"off")=36
	Set gtmtypes("open_relinkctl_sgm",8,"len")=4
	Set gtmtypes("open_relinkctl_sgm",8,"type")="boolean_t"
	Set gtmtypfldindx("open_relinkctl_sgm","locked")=8
	Set gtmtypes("open_relinkctl_sgm",9,"name")="open_relinkctl_sgm.hdr"
	Set gtmtypes("open_relinkctl_sgm",9,"off")=40
	Set gtmtypes("open_relinkctl_sgm",9,"len")=8
	Set gtmtypes("open_relinkctl_sgm",9,"type")="addr"
	Set gtmtypfldindx("open_relinkctl_sgm","hdr")=9
	Set gtmtypes("open_relinkctl_sgm",10,"name")="open_relinkctl_sgm.rec_base"
	Set gtmtypes("open_relinkctl_sgm",10,"off")=48
	Set gtmtypes("open_relinkctl_sgm",10,"len")=8
	Set gtmtypes("open_relinkctl_sgm",10,"type")="addr"
	Set gtmtypfldindx("open_relinkctl_sgm","rec_base")=10
	Set gtmtypes("open_relinkctl_sgm",11,"name")="open_relinkctl_sgm.shm_hashbase"
	Set gtmtypes("open_relinkctl_sgm",11,"off")=56
	Set gtmtypes("open_relinkctl_sgm",11,"len")=8
	Set gtmtypes("open_relinkctl_sgm",11,"type")="addr"
	Set gtmtypfldindx("open_relinkctl_sgm","shm_hashbase")=11
	Set gtmtypes("open_relinkctl_sgm",12,"name")="open_relinkctl_sgm.rtnobj_shm_base"
	Set gtmtypes("open_relinkctl_sgm",12,"off")=64
	Set gtmtypes("open_relinkctl_sgm",12,"len")=304
	Set gtmtypes("open_relinkctl_sgm",12,"type")="addr"
	Set gtmtypfldindx("open_relinkctl_sgm","rtnobj_shm_base")=12
	Set gtmtypes("open_relinkctl_sgm",12,"dim")=38
	Set gtmtypes("open_relinkctl_sgm",13,"name")="open_relinkctl_sgm.rtnobj_shmid"
	Set gtmtypes("open_relinkctl_sgm",13,"off")=368
	Set gtmtypes("open_relinkctl_sgm",13,"len")=152
	Set gtmtypes("open_relinkctl_sgm",13,"type")="int"
	Set gtmtypfldindx("open_relinkctl_sgm","rtnobj_shmid")=13
	Set gtmtypes("open_relinkctl_sgm",13,"dim")=38
	Set gtmtypes("open_relinkctl_sgm",14,"name")="open_relinkctl_sgm.fd"
	Set gtmtypes("open_relinkctl_sgm",14,"off")=520
	Set gtmtypes("open_relinkctl_sgm",14,"len")=4
	Set gtmtypes("open_relinkctl_sgm",14,"type")="int"
	Set gtmtypfldindx("open_relinkctl_sgm","fd")=14
	Set gtmtypes("open_relinkctl_sgm",15,"name")="open_relinkctl_sgm.rtnobj_min_shm_index"
	Set gtmtypes("open_relinkctl_sgm",15,"off")=524
	Set gtmtypes("open_relinkctl_sgm",15,"len")=4
	Set gtmtypes("open_relinkctl_sgm",15,"type")="int"
	Set gtmtypfldindx("open_relinkctl_sgm","rtnobj_min_shm_index")=15
	Set gtmtypes("open_relinkctl_sgm",16,"name")="open_relinkctl_sgm.rtnobj_max_shm_index"
	Set gtmtypes("open_relinkctl_sgm",16,"off")=528
	Set gtmtypes("open_relinkctl_sgm",16,"len")=4
	Set gtmtypes("open_relinkctl_sgm",16,"type")="int"
	Set gtmtypfldindx("open_relinkctl_sgm","rtnobj_max_shm_index")=16
	;
	Set gtmtypes("open_shlib")="struct"
	Set gtmtypes("open_shlib",0)=3
	Set gtmtypes("open_shlib","len")=272
	Set gtmtypes("open_shlib",1,"name")="open_shlib.next"
	Set gtmtypes("open_shlib",1,"off")=0
	Set gtmtypes("open_shlib",1,"len")=8
	Set gtmtypes("open_shlib",1,"type")="addr"
	Set gtmtypfldindx("open_shlib","next")=1
	Set gtmtypes("open_shlib",2,"name")="open_shlib.shlib_handle"
	Set gtmtypes("open_shlib",2,"off")=8
	Set gtmtypes("open_shlib",2,"len")=8
	Set gtmtypes("open_shlib",2,"type")="addr"
	Set gtmtypfldindx("open_shlib","shlib_handle")=2
	Set gtmtypes("open_shlib",3,"name")="open_shlib.shlib_name"
	Set gtmtypes("open_shlib",3,"off")=16
	Set gtmtypes("open_shlib",3,"len")=256
	Set gtmtypes("open_shlib",3,"type")="char"
	Set gtmtypfldindx("open_shlib","shlib_name")=3
	;
	Set gtmtypes("oprtype")="struct"
	Set gtmtypes("oprtype",0)=15
	Set gtmtypes("oprtype","len")=16
	Set gtmtypes("oprtype",1,"name")="oprtype.oprclass"
	Set gtmtypes("oprtype",1,"off")=0
	Set gtmtypes("oprtype",1,"len")=4
	Set gtmtypes("oprtype",1,"type")="int"
	Set gtmtypfldindx("oprtype","oprclass")=1
	Set gtmtypes("oprtype",2,"name")="oprtype.oprval"
	Set gtmtypes("oprtype",2,"off")=8
	Set gtmtypes("oprtype",2,"len")=8
	Set gtmtypes("oprtype",2,"type")="union"
	Set gtmtypfldindx("oprtype","oprval")=2
	Set gtmtypes("oprtype",3,"name")="oprtype.oprval.indr"
	Set gtmtypes("oprtype",3,"off")=8
	Set gtmtypes("oprtype",3,"len")=8
	Set gtmtypes("oprtype",3,"type")="addr"
	Set gtmtypfldindx("oprtype","oprval.indr")=3
	Set gtmtypes("oprtype",4,"name")="oprtype.oprval.tref"
	Set gtmtypes("oprtype",4,"off")=8
	Set gtmtypes("oprtype",4,"len")=8
	Set gtmtypes("oprtype",4,"type")="addr"
	Set gtmtypfldindx("oprtype","oprval.tref")=4
	Set gtmtypes("oprtype",5,"name")="oprtype.oprval.tsize"
	Set gtmtypes("oprtype",5,"off")=8
	Set gtmtypes("oprtype",5,"len")=8
	Set gtmtypes("oprtype",5,"type")="addr"
	Set gtmtypfldindx("oprtype","oprval.tsize")=5
	Set gtmtypes("oprtype",6,"name")="oprtype.oprval.lab"
	Set gtmtypes("oprtype",6,"off")=8
	Set gtmtypes("oprtype",6,"len")=8
	Set gtmtypes("oprtype",6,"type")="addr"
	Set gtmtypfldindx("oprtype","oprval.lab")=6
	Set gtmtypes("oprtype",7,"name")="oprtype.oprval.mlin"
	Set gtmtypes("oprtype",7,"off")=8
	Set gtmtypes("oprtype",7,"len")=8
	Set gtmtypes("oprtype",7,"type")="addr"
	Set gtmtypfldindx("oprtype","oprval.mlin")=7
	Set gtmtypes("oprtype",8,"name")="oprtype.oprval.mlit"
	Set gtmtypes("oprtype",8,"off")=8
	Set gtmtypes("oprtype",8,"len")=8
	Set gtmtypes("oprtype",8,"type")="addr"
	Set gtmtypfldindx("oprtype","oprval.mlit")=8
	Set gtmtypes("oprtype",9,"name")="oprtype.oprval.cdlt"
	Set gtmtypes("oprtype",9,"off")=8
	Set gtmtypes("oprtype",9,"len")=8
	Set gtmtypes("oprtype",9,"type")="addr"
	Set gtmtypfldindx("oprtype","oprval.cdlt")=9
	Set gtmtypes("oprtype",10,"name")="oprtype.oprval.cdidx"
	Set gtmtypes("oprtype",10,"off")=8
	Set gtmtypes("oprtype",10,"len")=8
	Set gtmtypes("oprtype",10,"type")="addr"
	Set gtmtypfldindx("oprtype","oprval.cdidx")=10
	Set gtmtypes("oprtype",11,"name")="oprtype.oprval.vref"
	Set gtmtypes("oprtype",11,"off")=8
	Set gtmtypes("oprtype",11,"len")=8
	Set gtmtypes("oprtype",11,"type")="addr"
	Set gtmtypfldindx("oprtype","oprval.vref")=11
	Set gtmtypes("oprtype",12,"name")="oprtype.oprval.temp"
	Set gtmtypes("oprtype",12,"off")=8
	Set gtmtypes("oprtype",12,"len")=4
	Set gtmtypes("oprtype",12,"type")="int"
	Set gtmtypfldindx("oprtype","oprval.temp")=12
	Set gtmtypes("oprtype",13,"name")="oprtype.oprval.ilit"
	Set gtmtypes("oprtype",13,"off")=8
	Set gtmtypes("oprtype",13,"len")=4
	Set gtmtypes("oprtype",13,"type")="int"
	Set gtmtypfldindx("oprtype","oprval.ilit")=13
	Set gtmtypes("oprtype",14,"name")="oprtype.oprval.offset"
	Set gtmtypes("oprtype",14,"off")=8
	Set gtmtypes("oprtype",14,"len")=4
	Set gtmtypes("oprtype",14,"type")="int"
	Set gtmtypfldindx("oprtype","oprval.offset")=14
	Set gtmtypes("oprtype",15,"name")="oprtype.oprval.vreg"
	Set gtmtypes("oprtype",15,"off")=8
	Set gtmtypes("oprtype",15,"len")=1
	Set gtmtypes("oprtype",15,"type")="unsigned-char"
	Set gtmtypfldindx("oprtype","oprval.vreg")=15
	;
	Set gtmtypes("p1hdr")="struct"
	Set gtmtypes("p1hdr",0)=16
	Set gtmtypes("p1hdr","len")=512
	Set gtmtypes("p1hdr",1,"name")="p1hdr.p1hdr_tag"
	Set gtmtypes("p1hdr",1,"off")=0
	Set gtmtypes("p1hdr",1,"len")=8
	Set gtmtypes("p1hdr",1,"type")="unsigned-char"
	Set gtmtypfldindx("p1hdr","p1hdr_tag")=1
	Set gtmtypes("p1hdr",2,"name")="p1hdr.tn"
	Set gtmtypes("p1hdr",2,"off")=8
	Set gtmtypes("p1hdr",2,"len")=4
	Set gtmtypes("p1hdr",2,"type")="v15_trans_num"
	Set gtmtypfldindx("p1hdr","tn")=2
	Set gtmtypes("p1hdr",3,"name")="p1hdr.blk_count"
	Set gtmtypes("p1hdr",3,"off")=12
	Set gtmtypes("p1hdr",3,"len")=4
	Set gtmtypes("p1hdr",3,"type")="int"
	Set gtmtypfldindx("p1hdr","blk_count")=3
	Set gtmtypes("p1hdr",4,"name")="p1hdr.tot_blocks"
	Set gtmtypes("p1hdr",4,"off")=16
	Set gtmtypes("p1hdr",4,"len")=4
	Set gtmtypes("p1hdr",4,"type")="int"
	Set gtmtypfldindx("p1hdr","tot_blocks")=4
	Set gtmtypes("p1hdr",5,"name")="p1hdr.dt_leaf_cnt"
	Set gtmtypes("p1hdr",5,"off")=20
	Set gtmtypes("p1hdr",5,"len")=4
	Set gtmtypes("p1hdr",5,"type")="int"
	Set gtmtypfldindx("p1hdr","dt_leaf_cnt")=5
	Set gtmtypes("p1hdr",6,"name")="p1hdr.dt_index_cnt"
	Set gtmtypes("p1hdr",6,"off")=24
	Set gtmtypes("p1hdr",6,"len")=4
	Set gtmtypes("p1hdr",6,"type")="int"
	Set gtmtypfldindx("p1hdr","dt_index_cnt")=6
	Set gtmtypes("p1hdr",7,"name")="p1hdr.gvt_leaf_cnt"
	Set gtmtypes("p1hdr",7,"off")=28
	Set gtmtypes("p1hdr",7,"len")=4
	Set gtmtypes("p1hdr",7,"type")="int"
	Set gtmtypfldindx("p1hdr","gvt_leaf_cnt")=7
	Set gtmtypes("p1hdr",8,"name")="p1hdr.gvt_index_cnt"
	Set gtmtypes("p1hdr",8,"off")=32
	Set gtmtypes("p1hdr",8,"len")=4
	Set gtmtypes("p1hdr",8,"type")="int"
	Set gtmtypfldindx("p1hdr","gvt_index_cnt")=8
	Set gtmtypes("p1hdr",9,"name")="p1hdr.regname"
	Set gtmtypes("p1hdr",9,"off")=36
	Set gtmtypes("p1hdr",9,"len")=32
	Set gtmtypes("p1hdr",9,"type")="unsigned-char"
	Set gtmtypfldindx("p1hdr","regname")=9
	Set gtmtypes("p1hdr",10,"name")="p1hdr.dbfn"
	Set gtmtypes("p1hdr",10,"off")=68
	Set gtmtypes("p1hdr",10,"len")=256
	Set gtmtypes("p1hdr",10,"type")="unsigned-char"
	Set gtmtypfldindx("p1hdr","dbfn")=10
	Set gtmtypes("p1hdr",11,"name")="p1hdr.uid_len"
	Set gtmtypes("p1hdr",11,"off")=324
	Set gtmtypes("p1hdr",11,"len")=4
	Set gtmtypes("p1hdr",11,"type")="int"
	Set gtmtypfldindx("p1hdr","uid_len")=11
	Set gtmtypes("p1hdr",12,"name")="p1hdr.unique_id"
	Set gtmtypes("p1hdr",12,"off")=328
	Set gtmtypes("p1hdr",12,"len")=24
	Set gtmtypes("p1hdr",12,"type")="unique_file_id"
	Set gtmtypfldindx("p1hdr","unique_id")=12
	Set gtmtypes("p1hdr",13,"name")="p1hdr.unique_id.uid"
	Set gtmtypes("p1hdr",13,"off")=328
	Set gtmtypes("p1hdr",13,"len")=16
	Set gtmtypes("p1hdr",13,"type")="unix_file_id"
	Set gtmtypfldindx("p1hdr","unique_id.uid")=13
	Set gtmtypes("p1hdr",14,"name")="p1hdr.unique_id.file_id"
	Set gtmtypes("p1hdr",14,"off")=328
	Set gtmtypes("p1hdr",14,"len")=20
	Set gtmtypes("p1hdr",14,"type")="char"
	Set gtmtypfldindx("p1hdr","unique_id.file_id")=14
	Set gtmtypes("p1hdr",15,"name")="p1hdr.fillx"
	Set gtmtypes("p1hdr",15,"off")=352
	Set gtmtypes("p1hdr",15,"len")=8
	Set gtmtypes("p1hdr",15,"type")="char"
	Set gtmtypfldindx("p1hdr","fillx")=15
	Set gtmtypes("p1hdr",16,"name")="p1hdr.fill512"
	Set gtmtypes("p1hdr",16,"off")=360
	Set gtmtypes("p1hdr",16,"len")=152
	Set gtmtypes("p1hdr",16,"type")="char"
	Set gtmtypfldindx("p1hdr","fill512")=16
	;
	Set gtmtypes("p1rec")="struct"
	Set gtmtypes("p1rec",0)=5
	Set gtmtypes("p1rec","len")=20
	Set gtmtypes("p1rec",1,"name")="p1rec.tn"
	Set gtmtypes("p1rec",1,"off")=0
	Set gtmtypes("p1rec",1,"len")=4
	Set gtmtypes("p1rec",1,"type")="v15_trans_num"
	Set gtmtypfldindx("p1rec","tn")=1
	Set gtmtypes("p1rec",2,"name")="p1rec.blk_num"
	Set gtmtypes("p1rec",2,"off")=4
	Set gtmtypes("p1rec",2,"len")=4
	Set gtmtypes("p1rec",2,"type")="int"
	Set gtmtypfldindx("p1rec","blk_num")=2
	Set gtmtypes("p1rec",3,"name")="p1rec.blk_type"
	Set gtmtypes("p1rec",3,"off")=8
	Set gtmtypes("p1rec",3,"len")=4
	Set gtmtypes("p1rec",3,"type")="int"
	Set gtmtypfldindx("p1rec","blk_type")=3
	Set gtmtypes("p1rec",4,"name")="p1rec.blk_levl"
	Set gtmtypes("p1rec",4,"off")=12
	Set gtmtypes("p1rec",4,"len")=4
	Set gtmtypes("p1rec",4,"type")="int"
	Set gtmtypfldindx("p1rec","blk_levl")=4
	Set gtmtypes("p1rec",5,"name")="p1rec.akey_len"
	Set gtmtypes("p1rec",5,"off")=16
	Set gtmtypes("p1rec",5,"len")=4
	Set gtmtypes("p1rec",5,"type")="int"
	Set gtmtypfldindx("p1rec","akey_len")=5
	;
	Set gtmtypes("parm_pool")="struct"
	Set gtmtypes("parm_pool",0)=8
	Set gtmtypes("parm_pool","len")=16
	Set gtmtypes("parm_pool",1,"name")="parm_pool.capacity"
	Set gtmtypes("parm_pool",1,"off")=0
	Set gtmtypes("parm_pool",1,"len")=4
	Set gtmtypes("parm_pool",1,"type")="unsigned-int"
	Set gtmtypfldindx("parm_pool","capacity")=1
	Set gtmtypes("parm_pool",2,"name")="parm_pool.start_idx"
	Set gtmtypes("parm_pool",2,"off")=4
	Set gtmtypes("parm_pool",2,"len")=4
	Set gtmtypes("parm_pool",2,"type")="unsigned-int"
	Set gtmtypfldindx("parm_pool","start_idx")=2
	Set gtmtypes("parm_pool",3,"name")="parm_pool.parms"
	Set gtmtypes("parm_pool",3,"off")=8
	Set gtmtypes("parm_pool",3,"len")=8
	Set gtmtypes("parm_pool",3,"type")="parm_slot"
	Set gtmtypfldindx("parm_pool","parms")=3
	Set gtmtypes("parm_pool",4,"name")="parm_pool.parms[0].mask_and_cnt"
	Set gtmtypes("parm_pool",4,"off")=8
	Set gtmtypes("parm_pool",4,"len")=8
	Set gtmtypes("parm_pool",4,"type")="struct"
	Set gtmtypfldindx("parm_pool","parms[0].mask_and_cnt")=4
	Set gtmtypes("parm_pool",5,"name")="parm_pool.parms[0].mask_and_cnt.mask"
	Set gtmtypes("parm_pool",5,"off")=8
	Set gtmtypes("parm_pool",5,"len")=4
	Set gtmtypes("parm_pool",5,"type")="unsigned-int"
	Set gtmtypfldindx("parm_pool","parms[0].mask_and_cnt.mask")=5
	Set gtmtypes("parm_pool",6,"name")="parm_pool.parms[0].mask_and_cnt.actualcnt"
	Set gtmtypes("parm_pool",6,"off")=12
	Set gtmtypes("parm_pool",6,"len")=4
	Set gtmtypes("parm_pool",6,"type")="unsigned-int"
	Set gtmtypfldindx("parm_pool","parms[0].mask_and_cnt.actualcnt")=6
	Set gtmtypes("parm_pool",7,"name")="parm_pool.parms[0].frame"
	Set gtmtypes("parm_pool",7,"off")=8
	Set gtmtypes("parm_pool",7,"len")=8
	Set gtmtypes("parm_pool",7,"type")="addr"
	Set gtmtypfldindx("parm_pool","parms[0].frame")=7
	Set gtmtypes("parm_pool",8,"name")="parm_pool.parms[0].actuallist"
	Set gtmtypes("parm_pool",8,"off")=8
	Set gtmtypes("parm_pool",8,"len")=8
	Set gtmtypes("parm_pool",8,"type")="addr"
	Set gtmtypfldindx("parm_pool","parms[0].actuallist")=8
	;
	Set gtmtypes("parm_slot")="union"
	Set gtmtypes("parm_slot",0)=5
	Set gtmtypes("parm_slot","len")=8
	Set gtmtypes("parm_slot",1,"name")="parm_slot.mask_and_cnt"
	Set gtmtypes("parm_slot",1,"off")=0
	Set gtmtypes("parm_slot",1,"len")=8
	Set gtmtypes("parm_slot",1,"type")="struct"
	Set gtmtypfldindx("parm_slot","mask_and_cnt")=1
	Set gtmtypes("parm_slot",2,"name")="parm_slot.mask_and_cnt.mask"
	Set gtmtypes("parm_slot",2,"off")=0
	Set gtmtypes("parm_slot",2,"len")=4
	Set gtmtypes("parm_slot",2,"type")="unsigned-int"
	Set gtmtypfldindx("parm_slot","mask_and_cnt.mask")=2
	Set gtmtypes("parm_slot",3,"name")="parm_slot.mask_and_cnt.actualcnt"
	Set gtmtypes("parm_slot",3,"off")=4
	Set gtmtypes("parm_slot",3,"len")=4
	Set gtmtypes("parm_slot",3,"type")="unsigned-int"
	Set gtmtypfldindx("parm_slot","mask_and_cnt.actualcnt")=3
	Set gtmtypes("parm_slot",4,"name")="parm_slot.frame"
	Set gtmtypes("parm_slot",4,"off")=0
	Set gtmtypes("parm_slot",4,"len")=8
	Set gtmtypes("parm_slot",4,"type")="addr"
	Set gtmtypfldindx("parm_slot","frame")=4
	Set gtmtypes("parm_slot",5,"name")="parm_slot.actuallist"
	Set gtmtypes("parm_slot",5,"off")=0
	Set gtmtypes("parm_slot",5,"len")=8
	Set gtmtypes("parm_slot",5,"type")="addr"
	Set gtmtypfldindx("parm_slot","actuallist")=5
	;
	Set gtmtypes("parmblk_struct")="struct"
	Set gtmtypes("parmblk_struct",0)=7
	Set gtmtypes("parmblk_struct","len")=304
	Set gtmtypes("parmblk_struct",1,"name")="parmblk_struct.ci_rtn"
	Set gtmtypes("parmblk_struct",1,"off")=0
	Set gtmtypes("parmblk_struct",1,"len")=8
	Set gtmtypes("parmblk_struct",1,"type")="addr"
	Set gtmtypfldindx("parmblk_struct","ci_rtn")=1
	Set gtmtypes("parmblk_struct",2,"name")="parmblk_struct.argcnt"
	Set gtmtypes("parmblk_struct",2,"off")=8
	Set gtmtypes("parmblk_struct",2,"len")=4
	Set gtmtypes("parmblk_struct",2,"type")="int"
	Set gtmtypfldindx("parmblk_struct","argcnt")=2
	Set gtmtypes("parmblk_struct",3,"name")="parmblk_struct.rtnaddr"
	Set gtmtypes("parmblk_struct",3,"off")=16
	Set gtmtypes("parmblk_struct",3,"len")=8
	Set gtmtypes("parmblk_struct",3,"type")="addr"
	Set gtmtypfldindx("parmblk_struct","rtnaddr")=3
	Set gtmtypes("parmblk_struct",4,"name")="parmblk_struct.labaddr"
	Set gtmtypes("parmblk_struct",4,"off")=24
	Set gtmtypes("parmblk_struct",4,"len")=8
	Set gtmtypes("parmblk_struct",4,"type")="addr"
	Set gtmtypfldindx("parmblk_struct","labaddr")=4
	Set gtmtypes("parmblk_struct",5,"name")="parmblk_struct.retaddr"
	Set gtmtypes("parmblk_struct",5,"off")=32
	Set gtmtypes("parmblk_struct",5,"len")=8
	Set gtmtypes("parmblk_struct",5,"type")="addr"
	Set gtmtypfldindx("parmblk_struct","retaddr")=5
	Set gtmtypes("parmblk_struct",6,"name")="parmblk_struct.mask"
	Set gtmtypes("parmblk_struct",6,"off")=40
	Set gtmtypes("parmblk_struct",6,"len")=4
	Set gtmtypes("parmblk_struct",6,"type")="int"
	Set gtmtypfldindx("parmblk_struct","mask")=6
	Set gtmtypes("parmblk_struct",7,"name")="parmblk_struct.args"
	Set gtmtypes("parmblk_struct",7,"off")=48
	Set gtmtypes("parmblk_struct",7,"len")=256
	Set gtmtypes("parmblk_struct",7,"type")="addr"
	Set gtmtypfldindx("parmblk_struct","args")=7
	Set gtmtypes("parmblk_struct",7,"dim")=32
	;
	Set gtmtypes("parse_blk")="struct"
	Set gtmtypes("parse_blk",0)=17
	Set gtmtypes("parse_blk","len")=88
	Set gtmtypes("parse_blk",1,"name")="parse_blk.b_esl"
	Set gtmtypes("parse_blk",1,"off")=0
	Set gtmtypes("parse_blk",1,"len")=1
	Set gtmtypes("parse_blk",1,"type")="unsigned-char"
	Set gtmtypfldindx("parse_blk","b_esl")=1
	Set gtmtypes("parse_blk",2,"name")="parse_blk.b_node"
	Set gtmtypes("parse_blk",2,"off")=1
	Set gtmtypes("parse_blk",2,"len")=1
	Set gtmtypes("parse_blk",2,"type")="unsigned-char"
	Set gtmtypfldindx("parse_blk","b_node")=2
	Set gtmtypes("parse_blk",3,"name")="parse_blk.b_dir"
	Set gtmtypes("parse_blk",3,"off")=2
	Set gtmtypes("parse_blk",3,"len")=1
	Set gtmtypes("parse_blk",3,"type")="unsigned-char"
	Set gtmtypfldindx("parse_blk","b_dir")=3
	Set gtmtypes("parse_blk",4,"name")="parse_blk.b_name"
	Set gtmtypes("parse_blk",4,"off")=3
	Set gtmtypes("parse_blk",4,"len")=1
	Set gtmtypes("parse_blk",4,"type")="unsigned-char"
	Set gtmtypfldindx("parse_blk","b_name")=4
	Set gtmtypes("parse_blk",5,"name")="parse_blk.b_ext"
	Set gtmtypes("parse_blk",5,"off")=4
	Set gtmtypes("parse_blk",5,"len")=1
	Set gtmtypes("parse_blk",5,"type")="unsigned-char"
	Set gtmtypfldindx("parse_blk","b_ext")=5
	Set gtmtypes("parse_blk",6,"name")="parse_blk.def1_size"
	Set gtmtypes("parse_blk",6,"off")=5
	Set gtmtypes("parse_blk",6,"len")=1
	Set gtmtypes("parse_blk",6,"type")="unsigned-char"
	Set gtmtypfldindx("parse_blk","def1_size")=6
	Set gtmtypes("parse_blk",7,"name")="parse_blk.def1_buf"
	Set gtmtypes("parse_blk",7,"off")=8
	Set gtmtypes("parse_blk",7,"len")=8
	Set gtmtypes("parse_blk",7,"type")="addr"
	Set gtmtypfldindx("parse_blk","def1_buf")=7
	Set gtmtypes("parse_blk",8,"name")="parse_blk.def2_size"
	Set gtmtypes("parse_blk",8,"off")=16
	Set gtmtypes("parse_blk",8,"len")=1
	Set gtmtypes("parse_blk",8,"type")="unsigned-char"
	Set gtmtypfldindx("parse_blk","def2_size")=8
	Set gtmtypes("parse_blk",9,"name")="parse_blk.def2_buf"
	Set gtmtypes("parse_blk",9,"off")=24
	Set gtmtypes("parse_blk",9,"len")=8
	Set gtmtypes("parse_blk",9,"type")="addr"
	Set gtmtypfldindx("parse_blk","def2_buf")=9
	Set gtmtypes("parse_blk",10,"name")="parse_blk.buff_size"
	Set gtmtypes("parse_blk",10,"off")=32
	Set gtmtypes("parse_blk",10,"len")=1
	Set gtmtypes("parse_blk",10,"type")="unsigned-char"
	Set gtmtypfldindx("parse_blk","buff_size")=10
	Set gtmtypes("parse_blk",11,"name")="parse_blk.buffer"
	Set gtmtypes("parse_blk",11,"off")=40
	Set gtmtypes("parse_blk",11,"len")=8
	Set gtmtypes("parse_blk",11,"type")="addr"
	Set gtmtypfldindx("parse_blk","buffer")=11
	Set gtmtypes("parse_blk",12,"name")="parse_blk.fnb"
	Set gtmtypes("parse_blk",12,"off")=48
	Set gtmtypes("parse_blk",12,"len")=4
	Set gtmtypes("parse_blk",12,"type")="int"
	Set gtmtypfldindx("parse_blk","fnb")=12
	Set gtmtypes("parse_blk",13,"name")="parse_blk.fop"
	Set gtmtypes("parse_blk",13,"off")=52
	Set gtmtypes("parse_blk",13,"len")=4
	Set gtmtypes("parse_blk",13,"type")="int"
	Set gtmtypfldindx("parse_blk","fop")=13
	Set gtmtypes("parse_blk",14,"name")="parse_blk.l_node"
	Set gtmtypes("parse_blk",14,"off")=56
	Set gtmtypes("parse_blk",14,"len")=8
	Set gtmtypes("parse_blk",14,"type")="addr"
	Set gtmtypfldindx("parse_blk","l_node")=14
	Set gtmtypes("parse_blk",15,"name")="parse_blk.l_dir"
	Set gtmtypes("parse_blk",15,"off")=64
	Set gtmtypes("parse_blk",15,"len")=8
	Set gtmtypes("parse_blk",15,"type")="addr"
	Set gtmtypfldindx("parse_blk","l_dir")=15
	Set gtmtypes("parse_blk",16,"name")="parse_blk.l_name"
	Set gtmtypes("parse_blk",16,"off")=72
	Set gtmtypes("parse_blk",16,"len")=8
	Set gtmtypes("parse_blk",16,"type")="addr"
	Set gtmtypfldindx("parse_blk","l_name")=16
	Set gtmtypes("parse_blk",17,"name")="parse_blk.l_ext"
	Set gtmtypes("parse_blk",17,"off")=80
	Set gtmtypes("parse_blk",17,"len")=8
	Set gtmtypes("parse_blk",17,"type")="addr"
	Set gtmtypfldindx("parse_blk","l_ext")=17
	;
	Set gtmtypes("parse_save_block")="struct"
	Set gtmtypes("parse_save_block",0)=29
	Set gtmtypes("parse_save_block","len")=272
	Set gtmtypes("parse_save_block",1,"name")="parse_save_block.source_error_found"
	Set gtmtypes("parse_save_block",1,"off")=0
	Set gtmtypes("parse_save_block",1,"len")=4
	Set gtmtypes("parse_save_block",1,"type")="boolean_t"
	Set gtmtypfldindx("parse_save_block","source_error_found")=1
	Set gtmtypes("parse_save_block",2,"name")="parse_save_block.source_column"
	Set gtmtypes("parse_save_block",2,"off")=4
	Set gtmtypes("parse_save_block",2,"len")=4
	Set gtmtypes("parse_save_block",2,"type")="int"
	Set gtmtypfldindx("parse_save_block","source_column")=2
	Set gtmtypes("parse_save_block",3,"name")="parse_save_block.director_ident_len"
	Set gtmtypes("parse_save_block",3,"off")=8
	Set gtmtypes("parse_save_block",3,"len")=4
	Set gtmtypes("parse_save_block",3,"type")="int"
	Set gtmtypfldindx("parse_save_block","director_ident_len")=3
	Set gtmtypes("parse_save_block",4,"name")="parse_save_block.block_level"
	Set gtmtypes("parse_save_block",4,"off")=12
	Set gtmtypes("parse_save_block",4,"len")=4
	Set gtmtypes("parse_save_block",4,"type")="int"
	Set gtmtypfldindx("parse_save_block","block_level")=4
	Set gtmtypes("parse_save_block",5,"name")="parse_save_block.source_len"
	Set gtmtypes("parse_save_block",5,"off")=16
	Set gtmtypes("parse_save_block",5,"len")=4
	Set gtmtypes("parse_save_block",5,"type")="int"
	Set gtmtypfldindx("parse_save_block","source_len")=5
	Set gtmtypes("parse_save_block",6,"name")="parse_save_block.director_mval"
	Set gtmtypes("parse_save_block",6,"off")=24
	Set gtmtypes("parse_save_block",6,"len")=32
	Set gtmtypes("parse_save_block",6,"type")="mval"
	Set gtmtypfldindx("parse_save_block","director_mval")=6
	Set gtmtypes("parse_save_block",7,"name")="parse_save_block.director_mval.mvtype"
	Set gtmtypes("parse_save_block",7,"off")=24
	Set gtmtypes("parse_save_block",7,"len")=2
	Set gtmtypes("parse_save_block",7,"type")="unsigned-short"
	Set gtmtypfldindx("parse_save_block","director_mval.mvtype")=7
	Set gtmtypes("parse_save_block",8,"name")="parse_save_block.director_mval.fnpc_indx"
	Set gtmtypes("parse_save_block",8,"off")=27
	Set gtmtypes("parse_save_block",8,"len")=1
	Set gtmtypes("parse_save_block",8,"type")="unsigned-char"
	Set gtmtypfldindx("parse_save_block","director_mval.fnpc_indx")=8
	Set gtmtypes("parse_save_block",9,"name")="parse_save_block.director_mval.utfcgr_indx"
	Set gtmtypes("parse_save_block",9,"off")=28
	Set gtmtypes("parse_save_block",9,"len")=4
	Set gtmtypes("parse_save_block",9,"type")="unsigned-int"
	Set gtmtypfldindx("parse_save_block","director_mval.utfcgr_indx")=9
	Set gtmtypes("parse_save_block",10,"name")="parse_save_block.director_mval.m"
	Set gtmtypes("parse_save_block",10,"off")=32
	Set gtmtypes("parse_save_block",10,"len")=8
	Set gtmtypes("parse_save_block",10,"type")="int"
	Set gtmtypfldindx("parse_save_block","director_mval.m")=10
	Set gtmtypes("parse_save_block",10,"dim")=2
	Set gtmtypes("parse_save_block",11,"name")="parse_save_block.director_mval.str"
	Set gtmtypes("parse_save_block",11,"off")=40
	Set gtmtypes("parse_save_block",11,"len")=16
	Set gtmtypes("parse_save_block",11,"type")="mstr"
	Set gtmtypfldindx("parse_save_block","director_mval.str")=11
	Set gtmtypes("parse_save_block",12,"name")="parse_save_block.director_mval.str.char_len"
	Set gtmtypes("parse_save_block",12,"off")=40
	Set gtmtypes("parse_save_block",12,"len")=4
	Set gtmtypes("parse_save_block",12,"type")="unsigned-int"
	Set gtmtypfldindx("parse_save_block","director_mval.str.char_len")=12
	Set gtmtypes("parse_save_block",13,"name")="parse_save_block.director_mval.str.len"
	Set gtmtypes("parse_save_block",13,"off")=44
	Set gtmtypes("parse_save_block",13,"len")=4
	Set gtmtypes("parse_save_block",13,"type")="int"
	Set gtmtypfldindx("parse_save_block","director_mval.str.len")=13
	Set gtmtypes("parse_save_block",14,"name")="parse_save_block.director_mval.str.addr"
	Set gtmtypes("parse_save_block",14,"off")=48
	Set gtmtypes("parse_save_block",14,"len")=8
	Set gtmtypes("parse_save_block",14,"type")="addr"
	Set gtmtypfldindx("parse_save_block","director_mval.str.addr")=14
	Set gtmtypes("parse_save_block",15,"name")="parse_save_block.ident_buffer"
	Set gtmtypes("parse_save_block",15,"off")=56
	Set gtmtypes("parse_save_block",15,"len")=63
	Set gtmtypes("parse_save_block",15,"type")="unsigned-char"
	Set gtmtypfldindx("parse_save_block","ident_buffer")=15
	Set gtmtypes("parse_save_block",16,"name")="parse_save_block.director_token"
	Set gtmtypes("parse_save_block",16,"off")=119
	Set gtmtypes("parse_save_block",16,"len")=1
	Set gtmtypes("parse_save_block",16,"type")="char"
	Set gtmtypfldindx("parse_save_block","director_token")=16
	Set gtmtypes("parse_save_block",17,"name")="parse_save_block.lexical_ptr"
	Set gtmtypes("parse_save_block",17,"off")=120
	Set gtmtypes("parse_save_block",17,"len")=8
	Set gtmtypes("parse_save_block",17,"type")="addr"
	Set gtmtypfldindx("parse_save_block","lexical_ptr")=17
	Set gtmtypes("parse_save_block",18,"name")="parse_save_block.window_token"
	Set gtmtypes("parse_save_block",18,"off")=128
	Set gtmtypes("parse_save_block",18,"len")=1
	Set gtmtypes("parse_save_block",18,"type")="char"
	Set gtmtypfldindx("parse_save_block","window_token")=18
	Set gtmtypes("parse_save_block",19,"name")="parse_save_block.pos_in_chain"
	Set gtmtypes("parse_save_block",19,"off")=136
	Set gtmtypes("parse_save_block",19,"len")=136
	Set gtmtypes("parse_save_block",19,"type")="triple"
	Set gtmtypfldindx("parse_save_block","pos_in_chain")=19
	Set gtmtypes("parse_save_block",20,"name")="parse_save_block.pos_in_chain.opcode"
	Set gtmtypes("parse_save_block",20,"off")=136
	Set gtmtypes("parse_save_block",20,"len")=4
	Set gtmtypes("parse_save_block",20,"type")="unsigned-int"
	Set gtmtypfldindx("parse_save_block","pos_in_chain.opcode")=20
	Set gtmtypes("parse_save_block",21,"name")="parse_save_block.pos_in_chain.exorder"
	Set gtmtypes("parse_save_block",21,"off")=144
	Set gtmtypes("parse_save_block",21,"len")=16
	Set gtmtypes("parse_save_block",21,"type")="struct"
	Set gtmtypfldindx("parse_save_block","pos_in_chain.exorder")=21
	Set gtmtypes("parse_save_block",22,"name")="parse_save_block.pos_in_chain.exorder.fl"
	Set gtmtypes("parse_save_block",22,"off")=144
	Set gtmtypes("parse_save_block",22,"len")=8
	Set gtmtypes("parse_save_block",22,"type")="addr"
	Set gtmtypfldindx("parse_save_block","pos_in_chain.exorder.fl")=22
	Set gtmtypes("parse_save_block",23,"name")="parse_save_block.pos_in_chain.exorder.bl"
	Set gtmtypes("parse_save_block",23,"off")=152
	Set gtmtypes("parse_save_block",23,"len")=8
	Set gtmtypes("parse_save_block",23,"type")="addr"
	Set gtmtypfldindx("parse_save_block","pos_in_chain.exorder.bl")=23
	Set gtmtypes("parse_save_block",24,"name")="parse_save_block.pos_in_chain.backptr"
	Set gtmtypes("parse_save_block",24,"off")=160
	Set gtmtypes("parse_save_block",24,"len")=24
	Set gtmtypes("parse_save_block",24,"type")="tbp"
	Set gtmtypfldindx("parse_save_block","pos_in_chain.backptr")=24
	Set gtmtypes("parse_save_block",25,"name")="parse_save_block.pos_in_chain.jmplist"
	Set gtmtypes("parse_save_block",25,"off")=184
	Set gtmtypes("parse_save_block",25,"len")=24
	Set gtmtypes("parse_save_block",25,"type")="tbp"
	Set gtmtypfldindx("parse_save_block","pos_in_chain.jmplist")=25
	Set gtmtypes("parse_save_block",26,"name")="parse_save_block.pos_in_chain.src"
	Set gtmtypes("parse_save_block",26,"off")=208
	Set gtmtypes("parse_save_block",26,"len")=8
	Set gtmtypes("parse_save_block",26,"type")="source_address"
	Set gtmtypfldindx("parse_save_block","pos_in_chain.src")=26
	Set gtmtypes("parse_save_block",27,"name")="parse_save_block.pos_in_chain.rtaddr"
	Set gtmtypes("parse_save_block",27,"off")=216
	Set gtmtypes("parse_save_block",27,"len")=4
	Set gtmtypes("parse_save_block",27,"type")="int"
	Set gtmtypfldindx("parse_save_block","pos_in_chain.rtaddr")=27
	Set gtmtypes("parse_save_block",28,"name")="parse_save_block.pos_in_chain.operand"
	Set gtmtypes("parse_save_block",28,"off")=224
	Set gtmtypes("parse_save_block",28,"len")=32
	Set gtmtypes("parse_save_block",28,"type")="oprtype"
	Set gtmtypfldindx("parse_save_block","pos_in_chain.operand")=28
	Set gtmtypes("parse_save_block",28,"dim")=2
	Set gtmtypes("parse_save_block",29,"name")="parse_save_block.pos_in_chain.destination"
	Set gtmtypes("parse_save_block",29,"off")=256
	Set gtmtypes("parse_save_block",29,"len")=16
	Set gtmtypes("parse_save_block",29,"type")="oprtype"
	Set gtmtypfldindx("parse_save_block","pos_in_chain.destination")=29
	;
	Set gtmtypes("pat_strlit")="struct"
	Set gtmtypes("pat_strlit",0)=4
	Set gtmtypes("pat_strlit","len")=3956
	Set gtmtypes("pat_strlit",1,"name")="pat_strlit.bytelen"
	Set gtmtypes("pat_strlit",1,"off")=0
	Set gtmtypes("pat_strlit",1,"len")=4
	Set gtmtypes("pat_strlit",1,"type")="int"
	Set gtmtypfldindx("pat_strlit","bytelen")=1
	Set gtmtypes("pat_strlit",2,"name")="pat_strlit.charlen"
	Set gtmtypes("pat_strlit",2,"off")=4
	Set gtmtypes("pat_strlit",2,"len")=4
	Set gtmtypes("pat_strlit",2,"type")="int"
	Set gtmtypfldindx("pat_strlit","charlen")=2
	Set gtmtypes("pat_strlit",3,"name")="pat_strlit.flags"
	Set gtmtypes("pat_strlit",3,"off")=8
	Set gtmtypes("pat_strlit",3,"len")=4
	Set gtmtypes("pat_strlit",3,"type")="unsigned-int"
	Set gtmtypfldindx("pat_strlit","flags")=3
	Set gtmtypes("pat_strlit",4,"name")="pat_strlit.buff"
	Set gtmtypes("pat_strlit",4,"off")=12
	Set gtmtypes("pat_strlit",4,"len")=3941
	Set gtmtypes("pat_strlit",4,"type")="unsigned-char"
	Set gtmtypfldindx("pat_strlit","buff")=4
	;
	Set gtmtypes("pattern")="struct"
	Set gtmtypes("pattern",0)=7
	Set gtmtypes("pattern","len")=48
	Set gtmtypes("pattern",1,"name")="pattern.flink"
	Set gtmtypes("pattern",1,"off")=0
	Set gtmtypes("pattern",1,"len")=8
	Set gtmtypes("pattern",1,"type")="addr"
	Set gtmtypfldindx("pattern","flink")=1
	Set gtmtypes("pattern",2,"name")="pattern.typemask"
	Set gtmtypes("pattern",2,"off")=8
	Set gtmtypes("pattern",2,"len")=8
	Set gtmtypes("pattern",2,"type")="addr"
	Set gtmtypfldindx("pattern","typemask")=2
	Set gtmtypes("pattern",3,"name")="pattern.patYZnam"
	Set gtmtypes("pattern",3,"off")=16
	Set gtmtypes("pattern",3,"len")=8
	Set gtmtypes("pattern",3,"type")="addr"
	Set gtmtypfldindx("pattern","patYZnam")=3
	Set gtmtypes("pattern",4,"name")="pattern.patYZlen"
	Set gtmtypes("pattern",4,"off")=24
	Set gtmtypes("pattern",4,"len")=8
	Set gtmtypes("pattern",4,"type")="addr"
	Set gtmtypfldindx("pattern","patYZlen")=4
	Set gtmtypes("pattern",5,"name")="pattern.patYZnum"
	Set gtmtypes("pattern",5,"off")=32
	Set gtmtypes("pattern",5,"len")=4
	Set gtmtypes("pattern",5,"type")="int"
	Set gtmtypfldindx("pattern","patYZnum")=5
	Set gtmtypes("pattern",6,"name")="pattern.namlen"
	Set gtmtypes("pattern",6,"off")=36
	Set gtmtypes("pattern",6,"len")=4
	Set gtmtypes("pattern",6,"type")="int"
	Set gtmtypfldindx("pattern","namlen")=6
	Set gtmtypes("pattern",7,"name")="pattern.name"
	Set gtmtypes("pattern",7,"off")=40
	Set gtmtypes("pattern",7,"len")=2
	Set gtmtypes("pattern",7,"type")="char"
	Set gtmtypfldindx("pattern","name")=7
	;
	Set gtmtypes("phase2_wait_trace_t")="struct"
	Set gtmtypes("phase2_wait_trace_t",0)=2
	Set gtmtypes("phase2_wait_trace_t","len")=16
	Set gtmtypes("phase2_wait_trace_t",1,"name")="phase2_wait_trace_t.blocking_pid"
	Set gtmtypes("phase2_wait_trace_t",1,"off")=0
	Set gtmtypes("phase2_wait_trace_t",1,"len")=4
	Set gtmtypes("phase2_wait_trace_t",1,"type")="unsigned-int"
	Set gtmtypfldindx("phase2_wait_trace_t","blocking_pid")=1
	Set gtmtypes("phase2_wait_trace_t",2,"name")="phase2_wait_trace_t.cr"
	Set gtmtypes("phase2_wait_trace_t",2,"off")=8
	Set gtmtypes("phase2_wait_trace_t",2,"len")=8
	Set gtmtypes("phase2_wait_trace_t",2,"type")="addr"
	Set gtmtypfldindx("phase2_wait_trace_t","cr")=2
	;
	Set gtmtypes("phase_static_area")="struct"
	Set gtmtypes("phase_static_area",0)=86
	Set gtmtypes("phase_static_area","len")=11496
	Set gtmtypes("phase_static_area",1,"name")="phase_static_area.hint_lcl"
	Set gtmtypes("phase_static_area",1,"off")=0
	Set gtmtypes("phase_static_area",1,"len")=4
	Set gtmtypes("phase_static_area",1,"type")="int"
	Set gtmtypfldindx("phase_static_area","hint_lcl")=1
	Set gtmtypes("phase_static_area",2,"name")="phase_static_area.outfd"
	Set gtmtypes("phase_static_area",2,"off")=4
	Set gtmtypes("phase_static_area",2,"len")=4
	Set gtmtypes("phase_static_area",2,"type")="int"
	Set gtmtypfldindx("phase_static_area","outfd")=2
	Set gtmtypes("phase_static_area",3,"name")="phase_static_area.blks_processed"
	Set gtmtypes("phase_static_area",3,"off")=8
	Set gtmtypes("phase_static_area",3,"len")=4
	Set gtmtypes("phase_static_area",3,"type")="int"
	Set gtmtypfldindx("phase_static_area","blks_processed")=3
	Set gtmtypes("phase_static_area",4,"name")="phase_static_area.blks_bypassed"
	Set gtmtypes("phase_static_area",4,"off")=12
	Set gtmtypes("phase_static_area",4,"len")=4
	Set gtmtypes("phase_static_area",4,"type")="int"
	Set gtmtypfldindx("phase_static_area","blks_bypassed")=4
	Set gtmtypes("phase_static_area",5,"name")="phase_static_area.blks_too_big"
	Set gtmtypes("phase_static_area",5,"off")=16
	Set gtmtypes("phase_static_area",5,"len")=4
	Set gtmtypes("phase_static_area",5,"type")="int"
	Set gtmtypfldindx("phase_static_area","blks_too_big")=5
	Set gtmtypes("phase_static_area",6,"name")="phase_static_area.blks_read"
	Set gtmtypes("phase_static_area",6,"off")=20
	Set gtmtypes("phase_static_area",6,"len")=4
	Set gtmtypes("phase_static_area",6,"type")="int"
	Set gtmtypfldindx("phase_static_area","blks_read")=6
	Set gtmtypes("phase_static_area",7,"name")="phase_static_area.blks_cached"
	Set gtmtypes("phase_static_area",7,"off")=24
	Set gtmtypes("phase_static_area",7,"len")=4
	Set gtmtypes("phase_static_area",7,"type")="int"
	Set gtmtypfldindx("phase_static_area","blks_cached")=7
	Set gtmtypes("phase_static_area",8,"name")="phase_static_area.blks_updated"
	Set gtmtypes("phase_static_area",8,"off")=28
	Set gtmtypes("phase_static_area",8,"len")=4
	Set gtmtypes("phase_static_area",8,"type")="int"
	Set gtmtypfldindx("phase_static_area","blks_updated")=8
	Set gtmtypes("phase_static_area",9,"name")="phase_static_area.blks_created"
	Set gtmtypes("phase_static_area",9,"off")=32
	Set gtmtypes("phase_static_area",9,"len")=4
	Set gtmtypes("phase_static_area",9,"type")="int"
	Set gtmtypfldindx("phase_static_area","blks_created")=9
	Set gtmtypes("phase_static_area",10,"name")="phase_static_area.dtlvl0"
	Set gtmtypes("phase_static_area",10,"off")=36
	Set gtmtypes("phase_static_area",10,"len")=4
	Set gtmtypes("phase_static_area",10,"type")="int"
	Set gtmtypfldindx("phase_static_area","dtlvl0")=10
	Set gtmtypes("phase_static_area",11,"name")="phase_static_area.dtlvln0"
	Set gtmtypes("phase_static_area",11,"off")=40
	Set gtmtypes("phase_static_area",11,"len")=4
	Set gtmtypes("phase_static_area",11,"type")="int"
	Set gtmtypfldindx("phase_static_area","dtlvln0")=11
	Set gtmtypes("phase_static_area",12,"name")="phase_static_area.gvtlvl0"
	Set gtmtypes("phase_static_area",12,"off")=44
	Set gtmtypes("phase_static_area",12,"len")=4
	Set gtmtypes("phase_static_area",12,"type")="int"
	Set gtmtypfldindx("phase_static_area","gvtlvl0")=12
	Set gtmtypes("phase_static_area",13,"name")="phase_static_area.gvtlvln0"
	Set gtmtypes("phase_static_area",13,"off")=48
	Set gtmtypes("phase_static_area",13,"len")=4
	Set gtmtypes("phase_static_area",13,"type")="int"
	Set gtmtypfldindx("phase_static_area","gvtlvln0")=13
	Set gtmtypes("phase_static_area",14,"name")="phase_static_area.gvtrchildren"
	Set gtmtypes("phase_static_area",14,"off")=52
	Set gtmtypes("phase_static_area",14,"len")=4
	Set gtmtypes("phase_static_area",14,"type")="int"
	Set gtmtypfldindx("phase_static_area","gvtrchildren")=14
	Set gtmtypes("phase_static_area",15,"name")="phase_static_area.blk_process_errors"
	Set gtmtypes("phase_static_area",15,"off")=56
	Set gtmtypes("phase_static_area",15,"len")=4
	Set gtmtypes("phase_static_area",15,"type")="int"
	Set gtmtypfldindx("phase_static_area","blk_process_errors")=15
	Set gtmtypes("phase_static_area",16,"name")="phase_static_area.gvtroot_rchildren_cnt"
	Set gtmtypes("phase_static_area",16,"off")=60
	Set gtmtypes("phase_static_area",16,"len")=4
	Set gtmtypes("phase_static_area",16,"type")="int"
	Set gtmtypfldindx("phase_static_area","gvtroot_rchildren_cnt")=16
	Set gtmtypes("phase_static_area",17,"name")="phase_static_area.local_bit_map_cnt"
	Set gtmtypes("phase_static_area",17,"off")=64
	Set gtmtypes("phase_static_area",17,"len")=4
	Set gtmtypes("phase_static_area",17,"type")="int"
	Set gtmtypfldindx("phase_static_area","local_bit_map_cnt")=17
	Set gtmtypes("phase_static_area",18,"name")="phase_static_area.blocks_to_process"
	Set gtmtypes("phase_static_area",18,"off")=68
	Set gtmtypes("phase_static_area",18,"len")=4
	Set gtmtypes("phase_static_area",18,"type")="unsigned-int"
	Set gtmtypfldindx("phase_static_area","blocks_to_process")=18
	Set gtmtypes("phase_static_area",19,"name")="phase_static_area.tmpcmdfile_len"
	Set gtmtypes("phase_static_area",19,"off")=72
	Set gtmtypes("phase_static_area",19,"len")=4
	Set gtmtypes("phase_static_area",19,"type")="int"
	Set gtmtypfldindx("phase_static_area","tmpcmdfile_len")=19
	Set gtmtypes("phase_static_area",20,"name")="phase_static_area.tmprsltfile_len"
	Set gtmtypes("phase_static_area",20,"off")=76
	Set gtmtypes("phase_static_area",20,"len")=4
	Set gtmtypes("phase_static_area",20,"type")="int"
	Set gtmtypfldindx("phase_static_area","tmprsltfile_len")=20
	Set gtmtypes("phase_static_area",21,"name")="phase_static_area.max_blk_len"
	Set gtmtypes("phase_static_area",21,"off")=80
	Set gtmtypes("phase_static_area",21,"len")=4
	Set gtmtypes("phase_static_area",21,"type")="unsigned-int"
	Set gtmtypfldindx("phase_static_area","max_blk_len")=21
	Set gtmtypes("phase_static_area",22,"name")="phase_static_area.max_rec_len"
	Set gtmtypes("phase_static_area",22,"off")=84
	Set gtmtypes("phase_static_area",22,"len")=4
	Set gtmtypes("phase_static_area",22,"type")="unsigned-int"
	Set gtmtypfldindx("phase_static_area","max_rec_len")=22
	Set gtmtypes("phase_static_area",23,"name")="phase_static_area.report_only"
	Set gtmtypes("phase_static_area",23,"off")=88
	Set gtmtypes("phase_static_area",23,"len")=4
	Set gtmtypes("phase_static_area",23,"type")="boolean_t"
	Set gtmtypfldindx("phase_static_area","report_only")=23
	Set gtmtypes("phase_static_area",24,"name")="phase_static_area.detail"
	Set gtmtypes("phase_static_area",24,"off")=92
	Set gtmtypes("phase_static_area",24,"len")=4
	Set gtmtypes("phase_static_area",24,"type")="boolean_t"
	Set gtmtypfldindx("phase_static_area","detail")=24
	Set gtmtypes("phase_static_area",25,"name")="phase_static_area.bsu_keys"
	Set gtmtypes("phase_static_area",25,"off")=96
	Set gtmtypes("phase_static_area",25,"len")=4
	Set gtmtypes("phase_static_area",25,"type")="boolean_t"
	Set gtmtypfldindx("phase_static_area","bsu_keys")=25
	Set gtmtypes("phase_static_area",26,"name")="phase_static_area.final"
	Set gtmtypes("phase_static_area",26,"off")=100
	Set gtmtypes("phase_static_area",26,"len")=4
	Set gtmtypes("phase_static_area",26,"type")="boolean_t"
	Set gtmtypfldindx("phase_static_area","final")=26
	Set gtmtypes("phase_static_area",27,"name")="phase_static_area.phase_one"
	Set gtmtypes("phase_static_area",27,"off")=104
	Set gtmtypes("phase_static_area",27,"len")=4
	Set gtmtypes("phase_static_area",27,"type")="boolean_t"
	Set gtmtypfldindx("phase_static_area","phase_one")=27
	Set gtmtypes("phase_static_area",28,"name")="phase_static_area.dbc_debug"
	Set gtmtypes("phase_static_area",28,"off")=108
	Set gtmtypes("phase_static_area",28,"len")=4
	Set gtmtypes("phase_static_area",28,"type")="boolean_t"
	Set gtmtypfldindx("phase_static_area","dbc_debug")=28
	Set gtmtypes("phase_static_area",29,"name")="phase_static_area.tmp_file_names_gend"
	Set gtmtypes("phase_static_area",29,"off")=112
	Set gtmtypes("phase_static_area",29,"len")=4
	Set gtmtypes("phase_static_area",29,"type")="boolean_t"
	Set gtmtypfldindx("phase_static_area","tmp_file_names_gend")=29
	Set gtmtypes("phase_static_area",30,"name")="phase_static_area.keep_temp_files"
	Set gtmtypes("phase_static_area",30,"off")=116
	Set gtmtypes("phase_static_area",30,"len")=4
	Set gtmtypes("phase_static_area",30,"type")="boolean_t"
	Set gtmtypfldindx("phase_static_area","keep_temp_files")=30
	Set gtmtypes("phase_static_area",31,"name")="phase_static_area.sem_inf"
	Set gtmtypes("phase_static_area",31,"off")=120
	Set gtmtypes("phase_static_area",31,"len")=24
	Set gtmtypes("phase_static_area",31,"type")="sem_info"
	Set gtmtypfldindx("phase_static_area","sem_inf")=31
	Set gtmtypes("phase_static_area",31,"dim")=3
	Set gtmtypes("phase_static_area",32,"name")="phase_static_area.sem_inf[0].ftok_key"
	Set gtmtypes("phase_static_area",32,"off")=120
	Set gtmtypes("phase_static_area",32,"len")=4
	Set gtmtypes("phase_static_area",32,"type")="int"
	Set gtmtypfldindx("phase_static_area","sem_inf[0].ftok_key")=32
	Set gtmtypes("phase_static_area",33,"name")="phase_static_area.sem_inf[0].sem_id"
	Set gtmtypes("phase_static_area",33,"off")=124
	Set gtmtypes("phase_static_area",33,"len")=4
	Set gtmtypes("phase_static_area",33,"type")="int"
	Set gtmtypfldindx("phase_static_area","sem_inf[0].sem_id")=33
	Set gtmtypes("phase_static_area",34,"name")="phase_static_area.dbc_critical"
	Set gtmtypes("phase_static_area",34,"off")=144
	Set gtmtypes("phase_static_area",34,"len")=4
	Set gtmtypes("phase_static_area",34,"type")="boolean_t"
	Set gtmtypfldindx("phase_static_area","dbc_critical")=34
	Set gtmtypes("phase_static_area",35,"name")="phase_static_area.dbc_fhdr_dirty"
	Set gtmtypes("phase_static_area",35,"off")=148
	Set gtmtypes("phase_static_area",35,"len")=4
	Set gtmtypes("phase_static_area",35,"type")="boolean_t"
	Set gtmtypfldindx("phase_static_area","dbc_fhdr_dirty")=35
	Set gtmtypes("phase_static_area",36,"name")="phase_static_area.curr_lbmap_buff"
	Set gtmtypes("phase_static_area",36,"off")=152
	Set gtmtypes("phase_static_area",36,"len")=8
	Set gtmtypes("phase_static_area",36,"type")="addr"
	Set gtmtypfldindx("phase_static_area","curr_lbmap_buff")=36
	Set gtmtypes("phase_static_area",37,"name")="phase_static_area.block_buff"
	Set gtmtypes("phase_static_area",37,"off")=160
	Set gtmtypes("phase_static_area",37,"len")=8
	Set gtmtypes("phase_static_area",37,"type")="addr"
	Set gtmtypfldindx("phase_static_area","block_buff")=37
	Set gtmtypes("phase_static_area",38,"name")="phase_static_area.util_cmd_buff"
	Set gtmtypes("phase_static_area",38,"off")=168
	Set gtmtypes("phase_static_area",38,"len")=256
	Set gtmtypes("phase_static_area",38,"type")="unsigned-char"
	Set gtmtypfldindx("phase_static_area","util_cmd_buff")=38
	Set gtmtypes("phase_static_area",39,"name")="phase_static_area.blk_set"
	Set gtmtypes("phase_static_area",39,"off")=424
	Set gtmtypes("phase_static_area",39,"len")=8
	Set gtmtypes("phase_static_area",39,"type")="addr"
	Set gtmtypfldindx("phase_static_area","blk_set")=39
	Set gtmtypes("phase_static_area",40,"name")="phase_static_area.block_depth"
	Set gtmtypes("phase_static_area",40,"off")=432
	Set gtmtypes("phase_static_area",40,"len")=4
	Set gtmtypes("phase_static_area",40,"type")="int"
	Set gtmtypfldindx("phase_static_area","block_depth")=40
	Set gtmtypes("phase_static_area",41,"name")="phase_static_area.block_depth_hwm"
	Set gtmtypes("phase_static_area",41,"off")=436
	Set gtmtypes("phase_static_area",41,"len")=4
	Set gtmtypes("phase_static_area",41,"type")="int"
	Set gtmtypfldindx("phase_static_area","block_depth_hwm")=41
	Set gtmtypes("phase_static_area",42,"name")="phase_static_area.tcfp"
	Set gtmtypes("phase_static_area",42,"off")=440
	Set gtmtypes("phase_static_area",42,"len")=8
	Set gtmtypes("phase_static_area",42,"type")="addr"
	Set gtmtypfldindx("phase_static_area","tcfp")=42
	Set gtmtypes("phase_static_area",43,"name")="phase_static_area.trfp"
	Set gtmtypes("phase_static_area",43,"off")=448
	Set gtmtypes("phase_static_area",43,"len")=8
	Set gtmtypes("phase_static_area",43,"type")="addr"
	Set gtmtypfldindx("phase_static_area","trfp")=43
	Set gtmtypes("phase_static_area",44,"name")="phase_static_area.hint_blk"
	Set gtmtypes("phase_static_area",44,"off")=456
	Set gtmtypes("phase_static_area",44,"len")=4
	Set gtmtypes("phase_static_area",44,"type")="int"
	Set gtmtypfldindx("phase_static_area","hint_blk")=44
	Set gtmtypes("phase_static_area",45,"name")="phase_static_area.ofhdr"
	Set gtmtypes("phase_static_area",45,"off")=464
	Set gtmtypes("phase_static_area",45,"len")=512
	Set gtmtypes("phase_static_area",45,"type")="p1hdr"
	Set gtmtypfldindx("phase_static_area","ofhdr")=45
	Set gtmtypes("phase_static_area",46,"name")="phase_static_area.ofhdr.p1hdr_tag"
	Set gtmtypes("phase_static_area",46,"off")=464
	Set gtmtypes("phase_static_area",46,"len")=8
	Set gtmtypes("phase_static_area",46,"type")="unsigned-char"
	Set gtmtypfldindx("phase_static_area","ofhdr.p1hdr_tag")=46
	Set gtmtypes("phase_static_area",47,"name")="phase_static_area.ofhdr.tn"
	Set gtmtypes("phase_static_area",47,"off")=472
	Set gtmtypes("phase_static_area",47,"len")=4
	Set gtmtypes("phase_static_area",47,"type")="v15_trans_num"
	Set gtmtypfldindx("phase_static_area","ofhdr.tn")=47
	Set gtmtypes("phase_static_area",48,"name")="phase_static_area.ofhdr.blk_count"
	Set gtmtypes("phase_static_area",48,"off")=476
	Set gtmtypes("phase_static_area",48,"len")=4
	Set gtmtypes("phase_static_area",48,"type")="int"
	Set gtmtypfldindx("phase_static_area","ofhdr.blk_count")=48
	Set gtmtypes("phase_static_area",49,"name")="phase_static_area.ofhdr.tot_blocks"
	Set gtmtypes("phase_static_area",49,"off")=480
	Set gtmtypes("phase_static_area",49,"len")=4
	Set gtmtypes("phase_static_area",49,"type")="int"
	Set gtmtypfldindx("phase_static_area","ofhdr.tot_blocks")=49
	Set gtmtypes("phase_static_area",50,"name")="phase_static_area.ofhdr.dt_leaf_cnt"
	Set gtmtypes("phase_static_area",50,"off")=484
	Set gtmtypes("phase_static_area",50,"len")=4
	Set gtmtypes("phase_static_area",50,"type")="int"
	Set gtmtypfldindx("phase_static_area","ofhdr.dt_leaf_cnt")=50
	Set gtmtypes("phase_static_area",51,"name")="phase_static_area.ofhdr.dt_index_cnt"
	Set gtmtypes("phase_static_area",51,"off")=488
	Set gtmtypes("phase_static_area",51,"len")=4
	Set gtmtypes("phase_static_area",51,"type")="int"
	Set gtmtypfldindx("phase_static_area","ofhdr.dt_index_cnt")=51
	Set gtmtypes("phase_static_area",52,"name")="phase_static_area.ofhdr.gvt_leaf_cnt"
	Set gtmtypes("phase_static_area",52,"off")=492
	Set gtmtypes("phase_static_area",52,"len")=4
	Set gtmtypes("phase_static_area",52,"type")="int"
	Set gtmtypfldindx("phase_static_area","ofhdr.gvt_leaf_cnt")=52
	Set gtmtypes("phase_static_area",53,"name")="phase_static_area.ofhdr.gvt_index_cnt"
	Set gtmtypes("phase_static_area",53,"off")=496
	Set gtmtypes("phase_static_area",53,"len")=4
	Set gtmtypes("phase_static_area",53,"type")="int"
	Set gtmtypfldindx("phase_static_area","ofhdr.gvt_index_cnt")=53
	Set gtmtypes("phase_static_area",54,"name")="phase_static_area.ofhdr.regname"
	Set gtmtypes("phase_static_area",54,"off")=500
	Set gtmtypes("phase_static_area",54,"len")=32
	Set gtmtypes("phase_static_area",54,"type")="unsigned-char"
	Set gtmtypfldindx("phase_static_area","ofhdr.regname")=54
	Set gtmtypes("phase_static_area",55,"name")="phase_static_area.ofhdr.dbfn"
	Set gtmtypes("phase_static_area",55,"off")=532
	Set gtmtypes("phase_static_area",55,"len")=256
	Set gtmtypes("phase_static_area",55,"type")="unsigned-char"
	Set gtmtypfldindx("phase_static_area","ofhdr.dbfn")=55
	Set gtmtypes("phase_static_area",56,"name")="phase_static_area.ofhdr.uid_len"
	Set gtmtypes("phase_static_area",56,"off")=788
	Set gtmtypes("phase_static_area",56,"len")=4
	Set gtmtypes("phase_static_area",56,"type")="int"
	Set gtmtypfldindx("phase_static_area","ofhdr.uid_len")=56
	Set gtmtypes("phase_static_area",57,"name")="phase_static_area.ofhdr.unique_id"
	Set gtmtypes("phase_static_area",57,"off")=792
	Set gtmtypes("phase_static_area",57,"len")=24
	Set gtmtypes("phase_static_area",57,"type")="unique_file_id"
	Set gtmtypfldindx("phase_static_area","ofhdr.unique_id")=57
	Set gtmtypes("phase_static_area",58,"name")="phase_static_area.ofhdr.unique_id.uid"
	Set gtmtypes("phase_static_area",58,"off")=792
	Set gtmtypes("phase_static_area",58,"len")=16
	Set gtmtypes("phase_static_area",58,"type")="unix_file_id"
	Set gtmtypfldindx("phase_static_area","ofhdr.unique_id.uid")=58
	Set gtmtypes("phase_static_area",59,"name")="phase_static_area.ofhdr.unique_id.file_id"
	Set gtmtypes("phase_static_area",59,"off")=792
	Set gtmtypes("phase_static_area",59,"len")=20
	Set gtmtypes("phase_static_area",59,"type")="char"
	Set gtmtypfldindx("phase_static_area","ofhdr.unique_id.file_id")=59
	Set gtmtypes("phase_static_area",60,"name")="phase_static_area.ofhdr.fillx"
	Set gtmtypes("phase_static_area",60,"off")=816
	Set gtmtypes("phase_static_area",60,"len")=8
	Set gtmtypes("phase_static_area",60,"type")="char"
	Set gtmtypfldindx("phase_static_area","ofhdr.fillx")=60
	Set gtmtypes("phase_static_area",61,"name")="phase_static_area.ofhdr.fill512"
	Set gtmtypes("phase_static_area",61,"off")=824
	Set gtmtypes("phase_static_area",61,"len")=152
	Set gtmtypes("phase_static_area",61,"type")="char"
	Set gtmtypfldindx("phase_static_area","ofhdr.fill512")=61
	Set gtmtypes("phase_static_area",62,"name")="phase_static_area.rhdr"
	Set gtmtypes("phase_static_area",62,"off")=976
	Set gtmtypes("phase_static_area",62,"len")=20
	Set gtmtypes("phase_static_area",62,"type")="p1rec"
	Set gtmtypfldindx("phase_static_area","rhdr")=62
	Set gtmtypes("phase_static_area",63,"name")="phase_static_area.rhdr.tn"
	Set gtmtypes("phase_static_area",63,"off")=976
	Set gtmtypes("phase_static_area",63,"len")=4
	Set gtmtypes("phase_static_area",63,"type")="v15_trans_num"
	Set gtmtypfldindx("phase_static_area","rhdr.tn")=63
	Set gtmtypes("phase_static_area",64,"name")="phase_static_area.rhdr.blk_num"
	Set gtmtypes("phase_static_area",64,"off")=980
	Set gtmtypes("phase_static_area",64,"len")=4
	Set gtmtypes("phase_static_area",64,"type")="int"
	Set gtmtypfldindx("phase_static_area","rhdr.blk_num")=64
	Set gtmtypes("phase_static_area",65,"name")="phase_static_area.rhdr.blk_type"
	Set gtmtypes("phase_static_area",65,"off")=984
	Set gtmtypes("phase_static_area",65,"len")=4
	Set gtmtypes("phase_static_area",65,"type")="int"
	Set gtmtypfldindx("phase_static_area","rhdr.blk_type")=65
	Set gtmtypes("phase_static_area",66,"name")="phase_static_area.rhdr.blk_levl"
	Set gtmtypes("phase_static_area",66,"off")=988
	Set gtmtypes("phase_static_area",66,"len")=4
	Set gtmtypes("phase_static_area",66,"type")="int"
	Set gtmtypfldindx("phase_static_area","rhdr.blk_levl")=66
	Set gtmtypes("phase_static_area",67,"name")="phase_static_area.rhdr.akey_len"
	Set gtmtypes("phase_static_area",67,"off")=992
	Set gtmtypes("phase_static_area",67,"len")=4
	Set gtmtypes("phase_static_area",67,"type")="int"
	Set gtmtypfldindx("phase_static_area","rhdr.akey_len")=67
	Set gtmtypes("phase_static_area",68,"name")="phase_static_area.gvtroot_rchildren"
	Set gtmtypes("phase_static_area",68,"off")=996
	Set gtmtypes("phase_static_area",68,"len")=160
	Set gtmtypes("phase_static_area",68,"type")="p1rec"
	Set gtmtypfldindx("phase_static_area","gvtroot_rchildren")=68
	Set gtmtypes("phase_static_area",68,"dim")=8
	Set gtmtypes("phase_static_area",69,"name")="phase_static_area.gvtroot_rchildren[0].tn"
	Set gtmtypes("phase_static_area",69,"off")=996
	Set gtmtypes("phase_static_area",69,"len")=4
	Set gtmtypes("phase_static_area",69,"type")="v15_trans_num"
	Set gtmtypfldindx("phase_static_area","gvtroot_rchildren[0].tn")=69
	Set gtmtypes("phase_static_area",70,"name")="phase_static_area.gvtroot_rchildren[0].blk_num"
	Set gtmtypes("phase_static_area",70,"off")=1000
	Set gtmtypes("phase_static_area",70,"len")=4
	Set gtmtypes("phase_static_area",70,"type")="int"
	Set gtmtypfldindx("phase_static_area","gvtroot_rchildren[0].blk_num")=70
	Set gtmtypes("phase_static_area",71,"name")="phase_static_area.gvtroot_rchildren[0].blk_type"
	Set gtmtypes("phase_static_area",71,"off")=1004
	Set gtmtypes("phase_static_area",71,"len")=4
	Set gtmtypes("phase_static_area",71,"type")="int"
	Set gtmtypfldindx("phase_static_area","gvtroot_rchildren[0].blk_type")=71
	Set gtmtypes("phase_static_area",72,"name")="phase_static_area.gvtroot_rchildren[0].blk_levl"
	Set gtmtypes("phase_static_area",72,"off")=1008
	Set gtmtypes("phase_static_area",72,"len")=4
	Set gtmtypes("phase_static_area",72,"type")="int"
	Set gtmtypfldindx("phase_static_area","gvtroot_rchildren[0].blk_levl")=72
	Set gtmtypes("phase_static_area",73,"name")="phase_static_area.gvtroot_rchildren[0].akey_len"
	Set gtmtypes("phase_static_area",73,"off")=1012
	Set gtmtypes("phase_static_area",73,"len")=4
	Set gtmtypes("phase_static_area",73,"type")="int"
	Set gtmtypfldindx("phase_static_area","gvtroot_rchildren[0].akey_len")=73
	Set gtmtypes("phase_static_area",74,"name")="phase_static_area.dbc_gv_cur_region"
	Set gtmtypes("phase_static_area",74,"off")=1160
	Set gtmtypes("phase_static_area",74,"len")=8
	Set gtmtypes("phase_static_area",74,"type")="addr"
	Set gtmtypfldindx("phase_static_area","dbc_gv_cur_region")=74
	Set gtmtypes("phase_static_area",75,"name")="phase_static_area.dbc_cs_data"
	Set gtmtypes("phase_static_area",75,"off")=1168
	Set gtmtypes("phase_static_area",75,"len")=8
	Set gtmtypes("phase_static_area",75,"type")="v15_sgmnt_data_ptr_t"
	Set gtmtypfldindx("phase_static_area","dbc_cs_data")=75
	Set gtmtypes("phase_static_area",76,"name")="phase_static_area.first_rec_key"
	Set gtmtypes("phase_static_area",76,"off")=1176
	Set gtmtypes("phase_static_area",76,"len")=8
	Set gtmtypes("phase_static_area",76,"type")="addr"
	Set gtmtypfldindx("phase_static_area","first_rec_key")=76
	Set gtmtypes("phase_static_area",77,"name")="phase_static_area.fc"
	Set gtmtypes("phase_static_area",77,"off")=1184
	Set gtmtypes("phase_static_area",77,"len")=8
	Set gtmtypes("phase_static_area",77,"type")="addr"
	Set gtmtypfldindx("phase_static_area","fc")=77
	Set gtmtypes("phase_static_area",78,"name")="phase_static_area.iebl"
	Set gtmtypes("phase_static_area",78,"off")=1192
	Set gtmtypes("phase_static_area",78,"len")=8
	Set gtmtypes("phase_static_area",78,"type")="addr"
	Set gtmtypfldindx("phase_static_area","iebl")=78
	Set gtmtypes("phase_static_area",79,"name")="phase_static_area.gvn_key"
	Set gtmtypes("phase_static_area",79,"off")=1200
	Set gtmtypes("phase_static_area",79,"len")=8
	Set gtmtypes("phase_static_area",79,"type")="addr"
	Set gtmtypfldindx("phase_static_area","gvn_key")=79
	Set gtmtypes("phase_static_area",80,"name")="phase_static_area.max_key"
	Set gtmtypes("phase_static_area",80,"off")=1208
	Set gtmtypes("phase_static_area",80,"len")=8
	Set gtmtypes("phase_static_area",80,"type")="addr"
	Set gtmtypfldindx("phase_static_area","max_key")=80
	Set gtmtypes("phase_static_area",81,"name")="phase_static_area.outfn"
	Set gtmtypes("phase_static_area",81,"off")=1216
	Set gtmtypes("phase_static_area",81,"len")=256
	Set gtmtypes("phase_static_area",81,"type")="unsigned-char"
	Set gtmtypfldindx("phase_static_area","outfn")=81
	Set gtmtypes("phase_static_area",82,"name")="phase_static_area.regname"
	Set gtmtypes("phase_static_area",82,"off")=1472
	Set gtmtypes("phase_static_area",82,"len")=32
	Set gtmtypes("phase_static_area",82,"type")="unsigned-char"
	Set gtmtypfldindx("phase_static_area","regname")=82
	Set gtmtypes("phase_static_area",83,"name")="phase_static_area.rslt_buff"
	Set gtmtypes("phase_static_area",83,"off")=1504
	Set gtmtypes("phase_static_area",83,"len")=9219
	Set gtmtypes("phase_static_area",83,"type")="unsigned-char"
	Set gtmtypfldindx("phase_static_area","rslt_buff")=83
	Set gtmtypes("phase_static_area",84,"name")="phase_static_area.tmpcmdfile"
	Set gtmtypes("phase_static_area",84,"off")=10723
	Set gtmtypes("phase_static_area",84,"len")=256
	Set gtmtypes("phase_static_area",84,"type")="unsigned-char"
	Set gtmtypfldindx("phase_static_area","tmpcmdfile")=84
	Set gtmtypes("phase_static_area",85,"name")="phase_static_area.tmprsltfile"
	Set gtmtypes("phase_static_area",85,"off")=10979
	Set gtmtypes("phase_static_area",85,"len")=256
	Set gtmtypes("phase_static_area",85,"type")="unsigned-char"
	Set gtmtypfldindx("phase_static_area","tmprsltfile")=85
	Set gtmtypes("phase_static_area",86,"name")="phase_static_area.tmpfiledir"
	Set gtmtypes("phase_static_area",86,"off")=11235
	Set gtmtypes("phase_static_area",86,"len")=256
	Set gtmtypes("phase_static_area",86,"type")="unsigned-char"
	Set gtmtypfldindx("phase_static_area","tmpfiledir")=86
	;
	Set gtmtypes("pini_list_struct")="struct"
	Set gtmtypes("pini_list_struct",0)=26
	Set gtmtypes("pini_list_struct","len")=200
	Set gtmtypes("pini_list_struct",1,"name")="pini_list_struct.pini_addr"
	Set gtmtypes("pini_list_struct",1,"off")=0
	Set gtmtypes("pini_list_struct",1,"len")=4
	Set gtmtypes("pini_list_struct",1,"type")="unsigned-int"
	Set gtmtypfldindx("pini_list_struct","pini_addr")=1
	Set gtmtypes("pini_list_struct",2,"name")="pini_list_struct.new_pini_addr"
	Set gtmtypes("pini_list_struct",2,"off")=4
	Set gtmtypes("pini_list_struct",2,"len")=4
	Set gtmtypes("pini_list_struct",2,"type")="unsigned-int"
	Set gtmtypfldindx("pini_list_struct","new_pini_addr")=2
	Set gtmtypes("pini_list_struct",3,"name")="pini_list_struct.jpv"
	Set gtmtypes("pini_list_struct",3,"off")=8
	Set gtmtypes("pini_list_struct",3,"len")=88
	Set gtmtypes("pini_list_struct",3,"type")="jnl_process_vector"
	Set gtmtypfldindx("pini_list_struct","jpv")=3
	Set gtmtypes("pini_list_struct",4,"name")="pini_list_struct.jpv.jpv_pid"
	Set gtmtypes("pini_list_struct",4,"off")=8
	Set gtmtypes("pini_list_struct",4,"len")=4
	Set gtmtypes("pini_list_struct",4,"type")="unsigned-int"
	Set gtmtypfldindx("pini_list_struct","jpv.jpv_pid")=4
	Set gtmtypes("pini_list_struct",5,"name")="pini_list_struct.jpv.jpv_image_count"
	Set gtmtypes("pini_list_struct",5,"off")=12
	Set gtmtypes("pini_list_struct",5,"len")=4
	Set gtmtypes("pini_list_struct",5,"type")="int"
	Set gtmtypfldindx("pini_list_struct","jpv.jpv_image_count")=5
	Set gtmtypes("pini_list_struct",6,"name")="pini_list_struct.jpv.jpv_time"
	Set gtmtypes("pini_list_struct",6,"off")=16
	Set gtmtypes("pini_list_struct",6,"len")=8
	Set gtmtypes("pini_list_struct",6,"type")="int64_t"
	Set gtmtypfldindx("pini_list_struct","jpv.jpv_time")=6
	Set gtmtypes("pini_list_struct",7,"name")="pini_list_struct.jpv.jpv_login_time"
	Set gtmtypes("pini_list_struct",7,"off")=24
	Set gtmtypes("pini_list_struct",7,"len")=8
	Set gtmtypes("pini_list_struct",7,"type")="int64_t"
	Set gtmtypfldindx("pini_list_struct","jpv.jpv_login_time")=7
	Set gtmtypes("pini_list_struct",8,"name")="pini_list_struct.jpv.jpv_node"
	Set gtmtypes("pini_list_struct",8,"off")=32
	Set gtmtypes("pini_list_struct",8,"len")=16
	Set gtmtypes("pini_list_struct",8,"type")="char"
	Set gtmtypfldindx("pini_list_struct","jpv.jpv_node")=8
	Set gtmtypes("pini_list_struct",9,"name")="pini_list_struct.jpv.jpv_user"
	Set gtmtypes("pini_list_struct",9,"off")=48
	Set gtmtypes("pini_list_struct",9,"len")=12
	Set gtmtypes("pini_list_struct",9,"type")="char"
	Set gtmtypfldindx("pini_list_struct","jpv.jpv_user")=9
	Set gtmtypes("pini_list_struct",10,"name")="pini_list_struct.jpv.jpv_prcnam"
	Set gtmtypes("pini_list_struct",10,"off")=60
	Set gtmtypes("pini_list_struct",10,"len")=16
	Set gtmtypes("pini_list_struct",10,"type")="char"
	Set gtmtypfldindx("pini_list_struct","jpv.jpv_prcnam")=10
	Set gtmtypes("pini_list_struct",11,"name")="pini_list_struct.jpv.jpv_terminal"
	Set gtmtypes("pini_list_struct",11,"off")=76
	Set gtmtypes("pini_list_struct",11,"len")=15
	Set gtmtypes("pini_list_struct",11,"type")="char"
	Set gtmtypfldindx("pini_list_struct","jpv.jpv_terminal")=11
	Set gtmtypes("pini_list_struct",12,"name")="pini_list_struct.jpv.jpv_mode"
	Set gtmtypes("pini_list_struct",12,"off")=91
	Set gtmtypes("pini_list_struct",12,"len")=1
	Set gtmtypes("pini_list_struct",12,"type")="unsigned-char"
	Set gtmtypfldindx("pini_list_struct","jpv.jpv_mode")=12
	Set gtmtypes("pini_list_struct",13,"name")="pini_list_struct.jpv.filler"
	Set gtmtypes("pini_list_struct",13,"off")=92
	Set gtmtypes("pini_list_struct",13,"len")=4
	Set gtmtypes("pini_list_struct",13,"type")="int"
	Set gtmtypfldindx("pini_list_struct","jpv.filler")=13
	Set gtmtypes("pini_list_struct",14,"name")="pini_list_struct.origjpv"
	Set gtmtypes("pini_list_struct",14,"off")=96
	Set gtmtypes("pini_list_struct",14,"len")=88
	Set gtmtypes("pini_list_struct",14,"type")="jnl_process_vector"
	Set gtmtypfldindx("pini_list_struct","origjpv")=14
	Set gtmtypes("pini_list_struct",15,"name")="pini_list_struct.origjpv.jpv_pid"
	Set gtmtypes("pini_list_struct",15,"off")=96
	Set gtmtypes("pini_list_struct",15,"len")=4
	Set gtmtypes("pini_list_struct",15,"type")="unsigned-int"
	Set gtmtypfldindx("pini_list_struct","origjpv.jpv_pid")=15
	Set gtmtypes("pini_list_struct",16,"name")="pini_list_struct.origjpv.jpv_image_count"
	Set gtmtypes("pini_list_struct",16,"off")=100
	Set gtmtypes("pini_list_struct",16,"len")=4
	Set gtmtypes("pini_list_struct",16,"type")="int"
	Set gtmtypfldindx("pini_list_struct","origjpv.jpv_image_count")=16
	Set gtmtypes("pini_list_struct",17,"name")="pini_list_struct.origjpv.jpv_time"
	Set gtmtypes("pini_list_struct",17,"off")=104
	Set gtmtypes("pini_list_struct",17,"len")=8
	Set gtmtypes("pini_list_struct",17,"type")="int64_t"
	Set gtmtypfldindx("pini_list_struct","origjpv.jpv_time")=17
	Set gtmtypes("pini_list_struct",18,"name")="pini_list_struct.origjpv.jpv_login_time"
	Set gtmtypes("pini_list_struct",18,"off")=112
	Set gtmtypes("pini_list_struct",18,"len")=8
	Set gtmtypes("pini_list_struct",18,"type")="int64_t"
	Set gtmtypfldindx("pini_list_struct","origjpv.jpv_login_time")=18
	Set gtmtypes("pini_list_struct",19,"name")="pini_list_struct.origjpv.jpv_node"
	Set gtmtypes("pini_list_struct",19,"off")=120
	Set gtmtypes("pini_list_struct",19,"len")=16
	Set gtmtypes("pini_list_struct",19,"type")="char"
	Set gtmtypfldindx("pini_list_struct","origjpv.jpv_node")=19
	Set gtmtypes("pini_list_struct",20,"name")="pini_list_struct.origjpv.jpv_user"
	Set gtmtypes("pini_list_struct",20,"off")=136
	Set gtmtypes("pini_list_struct",20,"len")=12
	Set gtmtypes("pini_list_struct",20,"type")="char"
	Set gtmtypfldindx("pini_list_struct","origjpv.jpv_user")=20
	Set gtmtypes("pini_list_struct",21,"name")="pini_list_struct.origjpv.jpv_prcnam"
	Set gtmtypes("pini_list_struct",21,"off")=148
	Set gtmtypes("pini_list_struct",21,"len")=16
	Set gtmtypes("pini_list_struct",21,"type")="char"
	Set gtmtypfldindx("pini_list_struct","origjpv.jpv_prcnam")=21
	Set gtmtypes("pini_list_struct",22,"name")="pini_list_struct.origjpv.jpv_terminal"
	Set gtmtypes("pini_list_struct",22,"off")=164
	Set gtmtypes("pini_list_struct",22,"len")=15
	Set gtmtypes("pini_list_struct",22,"type")="char"
	Set gtmtypfldindx("pini_list_struct","origjpv.jpv_terminal")=22
	Set gtmtypes("pini_list_struct",23,"name")="pini_list_struct.origjpv.jpv_mode"
	Set gtmtypes("pini_list_struct",23,"off")=179
	Set gtmtypes("pini_list_struct",23,"len")=1
	Set gtmtypes("pini_list_struct",23,"type")="unsigned-char"
	Set gtmtypfldindx("pini_list_struct","origjpv.jpv_mode")=23
	Set gtmtypes("pini_list_struct",24,"name")="pini_list_struct.origjpv.filler"
	Set gtmtypes("pini_list_struct",24,"off")=180
	Set gtmtypes("pini_list_struct",24,"len")=4
	Set gtmtypes("pini_list_struct",24,"type")="int"
	Set gtmtypfldindx("pini_list_struct","origjpv.filler")=24
	Set gtmtypes("pini_list_struct",25,"name")="pini_list_struct.pini_jpv_time"
	Set gtmtypes("pini_list_struct",25,"off")=184
	Set gtmtypes("pini_list_struct",25,"len")=8
	Set gtmtypes("pini_list_struct",25,"type")="int64_t"
	Set gtmtypfldindx("pini_list_struct","pini_jpv_time")=25
	Set gtmtypes("pini_list_struct",26,"name")="pini_list_struct.state"
	Set gtmtypes("pini_list_struct",26,"off")=192
	Set gtmtypes("pini_list_struct",26,"len")=4
	Set gtmtypes("pini_list_struct",26,"type")="int"
	Set gtmtypfldindx("pini_list_struct","state")=26
	;
	Set gtmtypes("pipe_interrupt")="struct"
	Set gtmtypes("pipe_interrupt",0)=12
	Set gtmtypes("pipe_interrupt","len")=56
	Set gtmtypes("pipe_interrupt",1,"name")="pipe_interrupt.end_time"
	Set gtmtypes("pipe_interrupt",1,"off")=0
	Set gtmtypes("pipe_interrupt",1,"len")=16
	Set gtmtypes("pipe_interrupt",1,"type")="ABS_TIME"
	Set gtmtypfldindx("pipe_interrupt","end_time")=1
	Set gtmtypes("pipe_interrupt",2,"name")="pipe_interrupt.end_time.at_sec"
	Set gtmtypes("pipe_interrupt",2,"off")=0
	Set gtmtypes("pipe_interrupt",2,"len")=8
	Set gtmtypes("pipe_interrupt",2,"type")="long"
	Set gtmtypfldindx("pipe_interrupt","end_time.at_sec")=2
	Set gtmtypes("pipe_interrupt",3,"name")="pipe_interrupt.end_time.at_usec"
	Set gtmtypes("pipe_interrupt",3,"off")=8
	Set gtmtypes("pipe_interrupt",3,"len")=8
	Set gtmtypes("pipe_interrupt",3,"type")="long"
	Set gtmtypfldindx("pipe_interrupt","end_time.at_usec")=3
	Set gtmtypes("pipe_interrupt",4,"name")="pipe_interrupt.who_saved"
	Set gtmtypes("pipe_interrupt",4,"off")=16
	Set gtmtypes("pipe_interrupt",4,"len")=4
	Set gtmtypes("pipe_interrupt",4,"type")="int"
	Set gtmtypfldindx("pipe_interrupt","who_saved")=4
	Set gtmtypes("pipe_interrupt",5,"name")="pipe_interrupt.max_bufflen"
	Set gtmtypes("pipe_interrupt",5,"off")=20
	Set gtmtypes("pipe_interrupt",5,"len")=4
	Set gtmtypes("pipe_interrupt",5,"type")="int"
	Set gtmtypfldindx("pipe_interrupt","max_bufflen")=5
	Set gtmtypes("pipe_interrupt",6,"name")="pipe_interrupt.bytes_read"
	Set gtmtypes("pipe_interrupt",6,"off")=24
	Set gtmtypes("pipe_interrupt",6,"len")=4
	Set gtmtypes("pipe_interrupt",6,"type")="int"
	Set gtmtypfldindx("pipe_interrupt","bytes_read")=6
	Set gtmtypes("pipe_interrupt",7,"name")="pipe_interrupt.bytes2read"
	Set gtmtypes("pipe_interrupt",7,"off")=28
	Set gtmtypes("pipe_interrupt",7,"len")=4
	Set gtmtypes("pipe_interrupt",7,"type")="int"
	Set gtmtypfldindx("pipe_interrupt","bytes2read")=7
	Set gtmtypes("pipe_interrupt",8,"name")="pipe_interrupt.char_count"
	Set gtmtypes("pipe_interrupt",8,"off")=32
	Set gtmtypes("pipe_interrupt",8,"len")=4
	Set gtmtypes("pipe_interrupt",8,"type")="int"
	Set gtmtypfldindx("pipe_interrupt","char_count")=8
	Set gtmtypes("pipe_interrupt",9,"name")="pipe_interrupt.bytes_count"
	Set gtmtypes("pipe_interrupt",9,"off")=36
	Set gtmtypes("pipe_interrupt",9,"len")=4
	Set gtmtypes("pipe_interrupt",9,"type")="int"
	Set gtmtypfldindx("pipe_interrupt","bytes_count")=9
	Set gtmtypes("pipe_interrupt",10,"name")="pipe_interrupt.add_bytes"
	Set gtmtypes("pipe_interrupt",10,"off")=40
	Set gtmtypes("pipe_interrupt",10,"len")=4
	Set gtmtypes("pipe_interrupt",10,"type")="int"
	Set gtmtypfldindx("pipe_interrupt","add_bytes")=10
	Set gtmtypes("pipe_interrupt",11,"name")="pipe_interrupt.end_time_valid"
	Set gtmtypes("pipe_interrupt",11,"off")=44
	Set gtmtypes("pipe_interrupt",11,"len")=4
	Set gtmtypes("pipe_interrupt",11,"type")="boolean_t"
	Set gtmtypfldindx("pipe_interrupt","end_time_valid")=11
	Set gtmtypes("pipe_interrupt",12,"name")="pipe_interrupt.newpipe"
	Set gtmtypes("pipe_interrupt",12,"off")=48
	Set gtmtypes("pipe_interrupt",12,"len")=8
	Set gtmtypes("pipe_interrupt",12,"type")="addr"
	Set gtmtypfldindx("pipe_interrupt","newpipe")=12
	;
	Set gtmtypes("plength")="struct"
	Set gtmtypes("plength",0)=7
	Set gtmtypes("plength","len")=4
	Set gtmtypes("plength",1,"name")="plength.p"
	Set gtmtypes("plength",1,"off")=0
	Set gtmtypes("plength",1,"len")=4
	Set gtmtypes("plength",1,"type")="union"
	Set gtmtypfldindx("plength","p")=1
	Set gtmtypes("plength",2,"name")="plength.p.pint"
	Set gtmtypes("plength",2,"off")=0
	Set gtmtypes("plength",2,"len")=4
	Set gtmtypes("plength",2,"type")="int"
	Set gtmtypfldindx("plength","p.pint")=2
	Set gtmtypes("plength",3,"name")="plength.p.pblk"
	Set gtmtypes("plength",3,"off")=0
	Set gtmtypes("plength",3,"len")=4
	Set gtmtypes("plength",3,"type")="struct"
	Set gtmtypfldindx("plength","p.pblk")=3
	Set gtmtypes("plength",4,"name")="plength.p.pblk.b_esl"
	Set gtmtypes("plength",4,"off")=0
	Set gtmtypes("plength",4,"len")=1
	Set gtmtypes("plength",4,"type")="unsigned-char"
	Set gtmtypfldindx("plength","p.pblk.b_esl")=4
	Set gtmtypes("plength",5,"name")="plength.p.pblk.b_dir"
	Set gtmtypes("plength",5,"off")=1
	Set gtmtypes("plength",5,"len")=1
	Set gtmtypes("plength",5,"type")="unsigned-char"
	Set gtmtypfldindx("plength","p.pblk.b_dir")=5
	Set gtmtypes("plength",6,"name")="plength.p.pblk.b_name"
	Set gtmtypes("plength",6,"off")=2
	Set gtmtypes("plength",6,"len")=1
	Set gtmtypes("plength",6,"type")="unsigned-char"
	Set gtmtypfldindx("plength","p.pblk.b_name")=6
	Set gtmtypes("plength",7,"name")="plength.p.pblk.b_ext"
	Set gtmtypes("plength",7,"off")=3
	Set gtmtypes("plength",7,"len")=1
	Set gtmtypes("plength",7,"type")="unsigned-char"
	Set gtmtypfldindx("plength","p.pblk.b_ext")=7
	;
	Set gtmtypes("pre_v5_mident")="struct"
	Set gtmtypes("pre_v5_mident",0)=1
	Set gtmtypes("pre_v5_mident","len")=8
	Set gtmtypes("pre_v5_mident",1,"name")="pre_v5_mident.c"
	Set gtmtypes("pre_v5_mident",1,"off")=0
	Set gtmtypes("pre_v5_mident",1,"len")=8
	Set gtmtypes("pre_v5_mident",1,"type")="char"
	Set gtmtypfldindx("pre_v5_mident","c")=1
	;
	Set gtmtypes("probecrit_rec_t")="struct"
	Set gtmtypes("probecrit_rec_t",0)=7
	Set gtmtypes("probecrit_rec_t","len")=56
	Set gtmtypes("probecrit_rec_t",1,"name")="probecrit_rec_t.t_get_crit"
	Set gtmtypes("probecrit_rec_t",1,"off")=0
	Set gtmtypes("probecrit_rec_t",1,"len")=8
	Set gtmtypes("probecrit_rec_t",1,"type")="uint64_t"
	Set gtmtypfldindx("probecrit_rec_t","t_get_crit")=1
	Set gtmtypes("probecrit_rec_t",2,"name")="probecrit_rec_t.p_crit_failed"
	Set gtmtypes("probecrit_rec_t",2,"off")=8
	Set gtmtypes("probecrit_rec_t",2,"len")=8
	Set gtmtypes("probecrit_rec_t",2,"type")="uint64_t"
	Set gtmtypfldindx("probecrit_rec_t","p_crit_failed")=2
	Set gtmtypes("probecrit_rec_t",3,"name")="probecrit_rec_t.p_crit_que_slps"
	Set gtmtypes("probecrit_rec_t",3,"off")=16
	Set gtmtypes("probecrit_rec_t",3,"len")=8
	Set gtmtypes("probecrit_rec_t",3,"type")="uint64_t"
	Set gtmtypfldindx("probecrit_rec_t","p_crit_que_slps")=3
	Set gtmtypes("probecrit_rec_t",4,"name")="probecrit_rec_t.p_crit_yields"
	Set gtmtypes("probecrit_rec_t",4,"off")=24
	Set gtmtypes("probecrit_rec_t",4,"len")=8
	Set gtmtypes("probecrit_rec_t",4,"type")="uint64_t"
	Set gtmtypfldindx("probecrit_rec_t","p_crit_yields")=4
	Set gtmtypes("probecrit_rec_t",5,"name")="probecrit_rec_t.p_crit_que_full"
	Set gtmtypes("probecrit_rec_t",5,"off")=32
	Set gtmtypes("probecrit_rec_t",5,"len")=8
	Set gtmtypes("probecrit_rec_t",5,"type")="uint64_t"
	Set gtmtypfldindx("probecrit_rec_t","p_crit_que_full")=5
	Set gtmtypes("probecrit_rec_t",6,"name")="probecrit_rec_t.p_crit_que_slots"
	Set gtmtypes("probecrit_rec_t",6,"off")=40
	Set gtmtypes("probecrit_rec_t",6,"len")=8
	Set gtmtypes("probecrit_rec_t",6,"type")="uint64_t"
	Set gtmtypfldindx("probecrit_rec_t","p_crit_que_slots")=6
	Set gtmtypes("probecrit_rec_t",7,"name")="probecrit_rec_t.p_crit_success"
	Set gtmtypes("probecrit_rec_t",7,"off")=48
	Set gtmtypes("probecrit_rec_t",7,"len")=8
	Set gtmtypes("probecrit_rec_t",7,"type")="uint64_t"
	Set gtmtypfldindx("probecrit_rec_t","p_crit_success")=7
	;
	Set gtmtypes("protocol_msg")="struct"
	Set gtmtypes("protocol_msg",0)=1
	Set gtmtypes("protocol_msg","len")=33
	Set gtmtypes("protocol_msg",1,"name")="protocol_msg.msg"
	Set gtmtypes("protocol_msg",1,"off")=0
	Set gtmtypes("protocol_msg",1,"len")=33
	Set gtmtypes("protocol_msg",1,"type")="char"
	Set gtmtypfldindx("protocol_msg","msg")=1
	;
	Set gtmtypes("pte_csh")="struct"
	Set gtmtypes("pte_csh",0)=6
	Set gtmtypes("pte_csh","len")=32
	Set gtmtypes("pte_csh",1,"name")="pte_csh.patptr"
	Set gtmtypes("pte_csh",1,"off")=0
	Set gtmtypes("pte_csh",1,"len")=8
	Set gtmtypes("pte_csh",1,"type")="addr"
	Set gtmtypfldindx("pte_csh","patptr")=1
	Set gtmtypes("pte_csh",2,"name")="pte_csh.strptr"
	Set gtmtypes("pte_csh",2,"off")=8
	Set gtmtypes("pte_csh",2,"len")=8
	Set gtmtypes("pte_csh",2,"type")="addr"
	Set gtmtypfldindx("pte_csh","strptr")=2
	Set gtmtypes("pte_csh",3,"name")="pte_csh.charlen"
	Set gtmtypes("pte_csh",3,"off")=16
	Set gtmtypes("pte_csh",3,"len")=4
	Set gtmtypes("pte_csh",3,"type")="int"
	Set gtmtypfldindx("pte_csh","charlen")=3
	Set gtmtypes("pte_csh",4,"name")="pte_csh.repcnt"
	Set gtmtypes("pte_csh",4,"off")=20
	Set gtmtypes("pte_csh",4,"len")=4
	Set gtmtypes("pte_csh",4,"type")="int"
	Set gtmtypfldindx("pte_csh","repcnt")=4
	Set gtmtypes("pte_csh",5,"name")="pte_csh.count"
	Set gtmtypes("pte_csh",5,"off")=24
	Set gtmtypes("pte_csh",5,"len")=4
	Set gtmtypes("pte_csh",5,"type")="unsigned-int"
	Set gtmtypfldindx("pte_csh","count")=5
	Set gtmtypes("pte_csh",6,"name")="pte_csh.match"
	Set gtmtypes("pte_csh",6,"off")=28
	Set gtmtypes("pte_csh",6,"len")=4
	Set gtmtypes("pte_csh",6,"type")="boolean_t"
	Set gtmtypfldindx("pte_csh","match")=6
	;
	Set gtmtypes("pth_mutex_struct")="struct"
	Set gtmtypes("pth_mutex_struct",0)=3
	Set gtmtypes("pth_mutex_struct","len")=48
	Set gtmtypes("pth_mutex_struct",1,"name")="pth_mutex_struct.crit_cycle"
	Set gtmtypes("pth_mutex_struct",1,"off")=0
	Set gtmtypes("pth_mutex_struct",1,"len")=4
	Set gtmtypes("pth_mutex_struct",1,"type")="unsigned-int"
	Set gtmtypfldindx("pth_mutex_struct","crit_cycle")=1
	Set gtmtypes("pth_mutex_struct",2,"name")="pth_mutex_struct.stuck_cycle"
	Set gtmtypes("pth_mutex_struct",2,"off")=4
	Set gtmtypes("pth_mutex_struct",2,"len")=4
	Set gtmtypes("pth_mutex_struct",2,"type")="unsigned-int"
	Set gtmtypfldindx("pth_mutex_struct","stuck_cycle")=2
	Set gtmtypes("pth_mutex_struct",3,"name")="pth_mutex_struct.mutex"
	Set gtmtypes("pth_mutex_struct",3,"off")=8
	Set gtmtypes("pth_mutex_struct",3,"len")=40
	Set gtmtypes("pth_mutex_struct",3,"type")="pthread_mutex_t"
	Set gtmtypfldindx("pth_mutex_struct","mutex")=3
	;
	Set gtmtypes("ptstr")="struct"
	Set gtmtypes("ptstr",0)=2
	Set gtmtypes("ptstr","len")=16388
	Set gtmtypes("ptstr",1,"name")="ptstr.len"
	Set gtmtypes("ptstr",1,"off")=0
	Set gtmtypes("ptstr",1,"len")=4
	Set gtmtypes("ptstr",1,"type")="int"
	Set gtmtypfldindx("ptstr","len")=1
	Set gtmtypes("ptstr",2,"name")="ptstr.buff"
	Set gtmtypes("ptstr",2,"off")=4
	Set gtmtypes("ptstr",2,"len")=16384
	Set gtmtypes("ptstr",2,"type")="unsigned-int"
	Set gtmtypfldindx("ptstr","buff")=2
	Set gtmtypes("ptstr",2,"dim")=4096
	;
	Set gtmtypes("qio_iosb")="struct"
	Set gtmtypes("qio_iosb",0)=5
	Set gtmtypes("qio_iosb","len")=12
	Set gtmtypes("qio_iosb",1,"name")="qio_iosb.xfer_count"
	Set gtmtypes("qio_iosb",1,"off")=0
	Set gtmtypes("qio_iosb",1,"len")=2
	Set gtmtypes("qio_iosb",1,"type")="unsigned-short"
	Set gtmtypfldindx("qio_iosb","xfer_count")=1
	Set gtmtypes("qio_iosb",2,"name")="qio_iosb.len_len"
	Set gtmtypes("qio_iosb",2,"off")=4
	Set gtmtypes("qio_iosb",2,"len")=4
	Set gtmtypes("qio_iosb",2,"type")="int"
	Set gtmtypfldindx("qio_iosb","len_len")=2
	Set gtmtypes("qio_iosb",3,"name")="qio_iosb.u"
	Set gtmtypes("qio_iosb",3,"off")=8
	Set gtmtypes("qio_iosb",3,"len")=2
	Set gtmtypes("qio_iosb",3,"type")="union"
	Set gtmtypfldindx("qio_iosb","u")=3
	Set gtmtypes("qio_iosb",4,"name")="qio_iosb.u.len"
	Set gtmtypes("qio_iosb",4,"off")=8
	Set gtmtypes("qio_iosb",4,"len")=2
	Set gtmtypes("qio_iosb",4,"type")="unsigned-short"
	Set gtmtypfldindx("qio_iosb","u.len")=4
	Set gtmtypes("qio_iosb",5,"name")="qio_iosb.u.lenbuf"
	Set gtmtypes("qio_iosb",5,"off")=8
	Set gtmtypes("qio_iosb",5,"len")=2
	Set gtmtypes("qio_iosb",5,"type")="char"
	Set gtmtypfldindx("qio_iosb","u.lenbuf")=5
	;
	Set gtmtypes("que_ent")="struct"
	Set gtmtypes("que_ent",0)=2
	Set gtmtypes("que_ent","len")=16
	Set gtmtypes("que_ent",1,"name")="que_ent.fl"
	Set gtmtypes("que_ent",1,"off")=0
	Set gtmtypes("que_ent",1,"len")=8
	Set gtmtypes("que_ent",1,"type")="intptr_t"
	Set gtmtypfldindx("que_ent","fl")=1
	Set gtmtypes("que_ent",2,"name")="que_ent.bl"
	Set gtmtypes("que_ent",2,"off")=8
	Set gtmtypes("que_ent",2,"len")=8
	Set gtmtypes("que_ent",2,"type")="intptr_t"
	Set gtmtypfldindx("que_ent","bl")=2
	;
	Set gtmtypes("que_head")="struct"
	Set gtmtypes("que_head",0)=8
	Set gtmtypes("que_head","len")=24
	Set gtmtypes("que_head",1,"name")="que_head.fl"
	Set gtmtypes("que_head",1,"off")=0
	Set gtmtypes("que_head",1,"len")=8
	Set gtmtypes("que_head",1,"type")="intptr_t"
	Set gtmtypfldindx("que_head","fl")=1
	Set gtmtypes("que_head",2,"name")="que_head.bl"
	Set gtmtypes("que_head",2,"off")=8
	Set gtmtypes("que_head",2,"len")=8
	Set gtmtypes("que_head",2,"type")="intptr_t"
	Set gtmtypfldindx("que_head","bl")=2
	Set gtmtypes("que_head",3,"name")="que_head.latch"
	Set gtmtypes("que_head",3,"off")=16
	Set gtmtypes("que_head",3,"len")=8
	Set gtmtypes("que_head",3,"type")="global_latch_t"
	Set gtmtypfldindx("que_head","latch")=3
	Set gtmtypes("que_head",4,"name")="que_head.latch.u"
	Set gtmtypes("que_head",4,"off")=16
	Set gtmtypes("que_head",4,"len")=8
	Set gtmtypes("que_head",4,"type")="union"
	Set gtmtypfldindx("que_head","latch.u")=4
	Set gtmtypes("que_head",5,"name")="que_head.latch.u.pid_imgcnt"
	Set gtmtypes("que_head",5,"off")=16
	Set gtmtypes("que_head",5,"len")=8
	Set gtmtypes("que_head",5,"type")="uint64_t"
	Set gtmtypfldindx("que_head","latch.u.pid_imgcnt")=5
	Set gtmtypes("que_head",6,"name")="que_head.latch.u.parts"
	Set gtmtypes("que_head",6,"off")=16
	Set gtmtypes("que_head",6,"len")=8
	Set gtmtypes("que_head",6,"type")="struct"
	Set gtmtypfldindx("que_head","latch.u.parts")=6
	Set gtmtypes("que_head",7,"name")="que_head.latch.u.parts.latch_pid"
	Set gtmtypes("que_head",7,"off")=16
	Set gtmtypes("que_head",7,"len")=4
	Set gtmtypes("que_head",7,"type")="int"
	Set gtmtypfldindx("que_head","latch.u.parts.latch_pid")=7
	Set gtmtypes("que_head",8,"name")="que_head.latch.u.parts.latch_word"
	Set gtmtypes("que_head",8,"off")=20
	Set gtmtypes("que_head",8,"len")=4
	Set gtmtypes("que_head",8,"type")="int"
	Set gtmtypfldindx("que_head","latch.u.parts.latch_word")=8
	;
	Set gtmtypes("rc_aq_hdr")="struct"
	Set gtmtypes("rc_aq_hdr",0)=18
	Set gtmtypes("rc_aq_hdr","len")=14
	Set gtmtypes("rc_aq_hdr",1,"name")="rc_aq_hdr.len"
	Set gtmtypes("rc_aq_hdr",1,"off")=0
	Set gtmtypes("rc_aq_hdr",1,"len")=2
	Set gtmtypes("rc_aq_hdr",1,"type")="rc_word"
	Set gtmtypfldindx("rc_aq_hdr","len")=1
	Set gtmtypes("rc_aq_hdr",2,"name")="rc_aq_hdr.len.octet"
	Set gtmtypes("rc_aq_hdr",2,"off")=0
	Set gtmtypes("rc_aq_hdr",2,"len")=2
	Set gtmtypes("rc_aq_hdr",2,"type")="char"
	Set gtmtypfldindx("rc_aq_hdr","len.octet")=2
	Set gtmtypes("rc_aq_hdr",3,"name")="rc_aq_hdr.len.value"
	Set gtmtypes("rc_aq_hdr",3,"off")=0
	Set gtmtypes("rc_aq_hdr",3,"len")=2
	Set gtmtypes("rc_aq_hdr",3,"type")="unsigned-short"
	Set gtmtypfldindx("rc_aq_hdr","len.value")=3
	Set gtmtypes("rc_aq_hdr",4,"name")="rc_aq_hdr.typ"
	Set gtmtypes("rc_aq_hdr",4,"off")=2
	Set gtmtypes("rc_aq_hdr",4,"len")=2
	Set gtmtypes("rc_aq_hdr",4,"type")="rc_word"
	Set gtmtypfldindx("rc_aq_hdr","typ")=4
	Set gtmtypes("rc_aq_hdr",5,"name")="rc_aq_hdr.typ.octet"
	Set gtmtypes("rc_aq_hdr",5,"off")=2
	Set gtmtypes("rc_aq_hdr",5,"len")=2
	Set gtmtypes("rc_aq_hdr",5,"type")="char"
	Set gtmtypfldindx("rc_aq_hdr","typ.octet")=5
	Set gtmtypes("rc_aq_hdr",6,"name")="rc_aq_hdr.typ.value"
	Set gtmtypes("rc_aq_hdr",6,"off")=2
	Set gtmtypes("rc_aq_hdr",6,"len")=2
	Set gtmtypes("rc_aq_hdr",6,"type")="unsigned-short"
	Set gtmtypfldindx("rc_aq_hdr","typ.value")=6
	Set gtmtypes("rc_aq_hdr",7,"name")="rc_aq_hdr.erc"
	Set gtmtypes("rc_aq_hdr",7,"off")=4
	Set gtmtypes("rc_aq_hdr",7,"len")=2
	Set gtmtypes("rc_aq_hdr",7,"type")="rc_word"
	Set gtmtypfldindx("rc_aq_hdr","erc")=7
	Set gtmtypes("rc_aq_hdr",8,"name")="rc_aq_hdr.erc.octet"
	Set gtmtypes("rc_aq_hdr",8,"off")=4
	Set gtmtypes("rc_aq_hdr",8,"len")=2
	Set gtmtypes("rc_aq_hdr",8,"type")="char"
	Set gtmtypfldindx("rc_aq_hdr","erc.octet")=8
	Set gtmtypes("rc_aq_hdr",9,"name")="rc_aq_hdr.erc.value"
	Set gtmtypes("rc_aq_hdr",9,"off")=4
	Set gtmtypes("rc_aq_hdr",9,"len")=2
	Set gtmtypes("rc_aq_hdr",9,"type")="unsigned-short"
	Set gtmtypfldindx("rc_aq_hdr","erc.value")=9
	Set gtmtypes("rc_aq_hdr",10,"name")="rc_aq_hdr.pid1"
	Set gtmtypes("rc_aq_hdr",10,"off")=6
	Set gtmtypes("rc_aq_hdr",10,"len")=2
	Set gtmtypes("rc_aq_hdr",10,"type")="rc_word"
	Set gtmtypfldindx("rc_aq_hdr","pid1")=10
	Set gtmtypes("rc_aq_hdr",11,"name")="rc_aq_hdr.pid1.octet"
	Set gtmtypes("rc_aq_hdr",11,"off")=6
	Set gtmtypes("rc_aq_hdr",11,"len")=2
	Set gtmtypes("rc_aq_hdr",11,"type")="char"
	Set gtmtypfldindx("rc_aq_hdr","pid1.octet")=11
	Set gtmtypes("rc_aq_hdr",12,"name")="rc_aq_hdr.pid1.value"
	Set gtmtypes("rc_aq_hdr",12,"off")=6
	Set gtmtypes("rc_aq_hdr",12,"len")=2
	Set gtmtypes("rc_aq_hdr",12,"type")="unsigned-short"
	Set gtmtypfldindx("rc_aq_hdr","pid1.value")=12
	Set gtmtypes("rc_aq_hdr",13,"name")="rc_aq_hdr.pid2"
	Set gtmtypes("rc_aq_hdr",13,"off")=8
	Set gtmtypes("rc_aq_hdr",13,"len")=2
	Set gtmtypes("rc_aq_hdr",13,"type")="rc_word"
	Set gtmtypfldindx("rc_aq_hdr","pid2")=13
	Set gtmtypes("rc_aq_hdr",14,"name")="rc_aq_hdr.pid2.octet"
	Set gtmtypes("rc_aq_hdr",14,"off")=8
	Set gtmtypes("rc_aq_hdr",14,"len")=2
	Set gtmtypes("rc_aq_hdr",14,"type")="char"
	Set gtmtypfldindx("rc_aq_hdr","pid2.octet")=14
	Set gtmtypes("rc_aq_hdr",15,"name")="rc_aq_hdr.pid2.value"
	Set gtmtypes("rc_aq_hdr",15,"off")=8
	Set gtmtypes("rc_aq_hdr",15,"len")=2
	Set gtmtypes("rc_aq_hdr",15,"type")="unsigned-short"
	Set gtmtypfldindx("rc_aq_hdr","pid2.value")=15
	Set gtmtypes("rc_aq_hdr",16,"name")="rc_aq_hdr.xdsid"
	Set gtmtypes("rc_aq_hdr",16,"off")=10
	Set gtmtypes("rc_aq_hdr",16,"len")=4
	Set gtmtypes("rc_aq_hdr",16,"type")="rc_xdsid"
	Set gtmtypfldindx("rc_aq_hdr","xdsid")=16
	Set gtmtypes("rc_aq_hdr",17,"name")="rc_aq_hdr.xdsid.dsid"
	Set gtmtypes("rc_aq_hdr",17,"off")=10
	Set gtmtypes("rc_aq_hdr",17,"len")=2
	Set gtmtypes("rc_aq_hdr",17,"type")="rc_word"
	Set gtmtypfldindx("rc_aq_hdr","xdsid.dsid")=17
	Set gtmtypes("rc_aq_hdr",18,"name")="rc_aq_hdr.xdsid.node"
	Set gtmtypes("rc_aq_hdr",18,"off")=12
	Set gtmtypes("rc_aq_hdr",18,"len")=2
	Set gtmtypes("rc_aq_hdr",18,"type")="rc_word"
	Set gtmtypfldindx("rc_aq_hdr","xdsid.node")=18
	;
	Set gtmtypes("rc_byte")="union"
	Set gtmtypes("rc_byte",0)=2
	Set gtmtypes("rc_byte","len")=1
	Set gtmtypes("rc_byte",1,"name")="rc_byte.octet"
	Set gtmtypes("rc_byte",1,"off")=0
	Set gtmtypes("rc_byte",1,"len")=1
	Set gtmtypes("rc_byte",1,"type")="char"
	Set gtmtypfldindx("rc_byte","octet")=1
	Set gtmtypes("rc_byte",2,"name")="rc_byte.value"
	Set gtmtypes("rc_byte",2,"off")=0
	Set gtmtypes("rc_byte",2,"len")=1
	Set gtmtypes("rc_byte",2,"type")="unsigned-char"
	Set gtmtypfldindx("rc_byte","value")=2
	;
	Set gtmtypes("rc_cp_table")="struct"
	Set gtmtypes("rc_cp_table",0)=5
	Set gtmtypes("rc_cp_table","len")=2056
	Set gtmtypes("rc_cp_table",1,"name")="rc_cp_table.ring_buff"
	Set gtmtypes("rc_cp_table",1,"off")=0
	Set gtmtypes("rc_cp_table",1,"len")=2048
	Set gtmtypes("rc_cp_table",1,"type")="int"
	Set gtmtypfldindx("rc_cp_table","ring_buff")=1
	Set gtmtypes("rc_cp_table",1,"dim")=512
	Set gtmtypes("rc_cp_table",2,"name")="rc_cp_table.cpsync"
	Set gtmtypes("rc_cp_table",2,"off")=2048
	Set gtmtypes("rc_cp_table",2,"len")=2
	Set gtmtypes("rc_cp_table",2,"type")="unsigned-short"
	Set gtmtypfldindx("rc_cp_table","cpsync")=2
	Set gtmtypes("rc_cp_table",3,"name")="rc_cp_table.cpvfy"
	Set gtmtypes("rc_cp_table",3,"off")=2050
	Set gtmtypes("rc_cp_table",3,"len")=2
	Set gtmtypes("rc_cp_table",3,"type")="unsigned-short"
	Set gtmtypfldindx("rc_cp_table","cpvfy")=3
	Set gtmtypes("rc_cp_table",4,"name")="rc_cp_table.index"
	Set gtmtypes("rc_cp_table",4,"off")=2052
	Set gtmtypes("rc_cp_table",4,"len")=2
	Set gtmtypes("rc_cp_table",4,"type")="short"
	Set gtmtypfldindx("rc_cp_table","index")=4
	Set gtmtypes("rc_cp_table",5,"name")="rc_cp_table.server_count"
	Set gtmtypes("rc_cp_table",5,"off")=2054
	Set gtmtypes("rc_cp_table",5,"len")=2
	Set gtmtypes("rc_cp_table",5,"type")="short"
	Set gtmtypfldindx("rc_cp_table","server_count")=5
	;
	Set gtmtypes("rc_dsid_list")="struct"
	Set gtmtypes("rc_dsid_list",0)=4
	Set gtmtypes("rc_dsid_list","len")=32
	Set gtmtypes("rc_dsid_list",1,"name")="rc_dsid_list.dsid"
	Set gtmtypes("rc_dsid_list",1,"off")=0
	Set gtmtypes("rc_dsid_list",1,"len")=2
	Set gtmtypes("rc_dsid_list",1,"type")="short"
	Set gtmtypfldindx("rc_dsid_list","dsid")=1
	Set gtmtypes("rc_dsid_list",2,"name")="rc_dsid_list.fname"
	Set gtmtypes("rc_dsid_list",2,"off")=8
	Set gtmtypes("rc_dsid_list",2,"len")=8
	Set gtmtypes("rc_dsid_list",2,"type")="addr"
	Set gtmtypfldindx("rc_dsid_list","fname")=2
	Set gtmtypes("rc_dsid_list",3,"name")="rc_dsid_list.gda"
	Set gtmtypes("rc_dsid_list",3,"off")=16
	Set gtmtypes("rc_dsid_list",3,"len")=8
	Set gtmtypes("rc_dsid_list",3,"type")="addr"
	Set gtmtypfldindx("rc_dsid_list","gda")=3
	Set gtmtypes("rc_dsid_list",4,"name")="rc_dsid_list.next"
	Set gtmtypes("rc_dsid_list",4,"off")=24
	Set gtmtypes("rc_dsid_list",4,"len")=8
	Set gtmtypes("rc_dsid_list",4,"type")="addr"
	Set gtmtypfldindx("rc_dsid_list","next")=4
	;
	Set gtmtypes("rc_hist_rec")="struct"
	Set gtmtypes("rc_hist_rec",0)=7
	Set gtmtypes("rc_hist_rec","len")=8224
	Set gtmtypes("rc_hist_rec",1,"name")="rc_hist_rec.conn"
	Set gtmtypes("rc_hist_rec",1,"off")=0
	Set gtmtypes("rc_hist_rec",1,"len")=4
	Set gtmtypes("rc_hist_rec",1,"type")="int"
	Set gtmtypfldindx("rc_hist_rec","conn")=1
	Set gtmtypes("rc_hist_rec",2,"name")="rc_hist_rec.timestamp"
	Set gtmtypes("rc_hist_rec",2,"off")=8
	Set gtmtypes("rc_hist_rec",2,"len")=8
	Set gtmtypes("rc_hist_rec",2,"type")="time_t"
	Set gtmtypfldindx("rc_hist_rec","timestamp")=2
	Set gtmtypes("rc_hist_rec",3,"name")="rc_hist_rec.toobigflag"
	Set gtmtypes("rc_hist_rec",3,"off")=16
	Set gtmtypes("rc_hist_rec",3,"len")=4
	Set gtmtypes("rc_hist_rec",3,"type")="int"
	Set gtmtypfldindx("rc_hist_rec","toobigflag")=3
	Set gtmtypes("rc_hist_rec",4,"name")="rc_hist_rec.req"
	Set gtmtypes("rc_hist_rec",4,"off")=20
	Set gtmtypes("rc_hist_rec",4,"len")=4096
	Set gtmtypes("rc_hist_rec",4,"type")="char"
	Set gtmtypfldindx("rc_hist_rec","req")=4
	Set gtmtypes("rc_hist_rec",5,"name")="rc_hist_rec.rsp"
	Set gtmtypes("rc_hist_rec",5,"off")=4116
	Set gtmtypes("rc_hist_rec",5,"len")=4096
	Set gtmtypes("rc_hist_rec",5,"type")="char"
	Set gtmtypfldindx("rc_hist_rec","rsp")=5
	Set gtmtypes("rc_hist_rec",6,"name")="rc_hist_rec.req_len"
	Set gtmtypes("rc_hist_rec",6,"off")=8212
	Set gtmtypes("rc_hist_rec",6,"len")=4
	Set gtmtypes("rc_hist_rec",6,"type")="int"
	Set gtmtypfldindx("rc_hist_rec","req_len")=6
	Set gtmtypes("rc_hist_rec",7,"name")="rc_hist_rec.rsp_len"
	Set gtmtypes("rc_hist_rec",7,"off")=8216
	Set gtmtypes("rc_hist_rec",7,"len")=4
	Set gtmtypes("rc_hist_rec",7,"type")="int"
	Set gtmtypfldindx("rc_hist_rec","rsp_len")=7
	;
	Set gtmtypes("rc_kill")="struct"
	Set gtmtypes("rc_kill",0)=8
	Set gtmtypes("rc_kill","len")=20
	Set gtmtypes("rc_kill",1,"name")="rc_kill.hdr"
	Set gtmtypes("rc_kill",1,"off")=0
	Set gtmtypes("rc_kill",1,"len")=14
	Set gtmtypes("rc_kill",1,"type")="rc_q_hdr"
	Set gtmtypfldindx("rc_kill","hdr")=1
	Set gtmtypes("rc_kill",2,"name")="rc_kill.hdr.r"
	Set gtmtypes("rc_kill",2,"off")=0
	Set gtmtypes("rc_kill",2,"len")=14
	Set gtmtypes("rc_kill",2,"type")="rc_rq_hdr"
	Set gtmtypfldindx("rc_kill","hdr.r")=2
	Set gtmtypes("rc_kill",3,"name")="rc_kill.hdr.a"
	Set gtmtypes("rc_kill",3,"off")=0
	Set gtmtypes("rc_kill",3,"len")=14
	Set gtmtypes("rc_kill",3,"type")="rc_aq_hdr"
	Set gtmtypfldindx("rc_kill","hdr.a")=3
	Set gtmtypes("rc_kill",4,"name")="rc_kill.xnsid"
	Set gtmtypes("rc_kill",4,"off")=14
	Set gtmtypes("rc_kill",4,"len")=4
	Set gtmtypes("rc_kill",4,"type")="rc_xnsid"
	Set gtmtypfldindx("rc_kill","xnsid")=4
	Set gtmtypes("rc_kill",5,"name")="rc_kill.xnsid.value"
	Set gtmtypes("rc_kill",5,"off")=14
	Set gtmtypes("rc_kill",5,"len")=4
	Set gtmtypes("rc_kill",5,"type")="char"
	Set gtmtypfldindx("rc_kill","xnsid.value")=5
	Set gtmtypes("rc_kill",6,"name")="rc_kill.key"
	Set gtmtypes("rc_kill",6,"off")=18
	Set gtmtypes("rc_kill",6,"len")=2
	Set gtmtypes("rc_kill",6,"type")="rc_sbkey"
	Set gtmtypfldindx("rc_kill","key")=6
	Set gtmtypes("rc_kill",7,"name")="rc_kill.key.len"
	Set gtmtypes("rc_kill",7,"off")=18
	Set gtmtypes("rc_kill",7,"len")=1
	Set gtmtypes("rc_kill",7,"type")="rc_byte"
	Set gtmtypfldindx("rc_kill","key.len")=7
	Set gtmtypes("rc_kill",8,"name")="rc_kill.key.key"
	Set gtmtypes("rc_kill",8,"off")=19
	Set gtmtypes("rc_kill",8,"len")=1
	Set gtmtypes("rc_kill",8,"type")="char"
	Set gtmtypfldindx("rc_kill","key.key")=8
	;
	Set gtmtypes("rc_lknam")="struct"
	Set gtmtypes("rc_lknam",0)=9
	Set gtmtypes("rc_lknam","len")=8
	Set gtmtypes("rc_lknam",1,"name")="rc_lknam.xdsid"
	Set gtmtypes("rc_lknam",1,"off")=0
	Set gtmtypes("rc_lknam",1,"len")=4
	Set gtmtypes("rc_lknam",1,"type")="rc_xdsid"
	Set gtmtypfldindx("rc_lknam","xdsid")=1
	Set gtmtypes("rc_lknam",2,"name")="rc_lknam.xdsid.dsid"
	Set gtmtypes("rc_lknam",2,"off")=0
	Set gtmtypes("rc_lknam",2,"len")=2
	Set gtmtypes("rc_lknam",2,"type")="rc_word"
	Set gtmtypfldindx("rc_lknam","xdsid.dsid")=2
	Set gtmtypes("rc_lknam",3,"name")="rc_lknam.xdsid.node"
	Set gtmtypes("rc_lknam",3,"off")=2
	Set gtmtypes("rc_lknam",3,"len")=2
	Set gtmtypes("rc_lknam",3,"type")="rc_word"
	Set gtmtypfldindx("rc_lknam","xdsid.node")=3
	Set gtmtypes("rc_lknam",4,"name")="rc_lknam.node_handle"
	Set gtmtypes("rc_lknam",4,"off")=4
	Set gtmtypes("rc_lknam",4,"len")=2
	Set gtmtypes("rc_lknam",4,"type")="rc_word"
	Set gtmtypfldindx("rc_lknam","node_handle")=4
	Set gtmtypes("rc_lknam",5,"name")="rc_lknam.node_handle.octet"
	Set gtmtypes("rc_lknam",5,"off")=4
	Set gtmtypes("rc_lknam",5,"len")=2
	Set gtmtypes("rc_lknam",5,"type")="char"
	Set gtmtypfldindx("rc_lknam","node_handle.octet")=5
	Set gtmtypes("rc_lknam",6,"name")="rc_lknam.node_handle.value"
	Set gtmtypes("rc_lknam",6,"off")=4
	Set gtmtypes("rc_lknam",6,"len")=2
	Set gtmtypes("rc_lknam",6,"type")="unsigned-short"
	Set gtmtypfldindx("rc_lknam","node_handle.value")=6
	Set gtmtypes("rc_lknam",7,"name")="rc_lknam.sb_key"
	Set gtmtypes("rc_lknam",7,"off")=6
	Set gtmtypes("rc_lknam",7,"len")=2
	Set gtmtypes("rc_lknam",7,"type")="rc_sbkey"
	Set gtmtypfldindx("rc_lknam","sb_key")=7
	Set gtmtypes("rc_lknam",8,"name")="rc_lknam.sb_key.len"
	Set gtmtypes("rc_lknam",8,"off")=6
	Set gtmtypes("rc_lknam",8,"len")=1
	Set gtmtypes("rc_lknam",8,"type")="rc_byte"
	Set gtmtypfldindx("rc_lknam","sb_key.len")=8
	Set gtmtypes("rc_lknam",9,"name")="rc_lknam.sb_key.key"
	Set gtmtypes("rc_lknam",9,"off")=7
	Set gtmtypes("rc_lknam",9,"len")=1
	Set gtmtypes("rc_lknam",9,"type")="char"
	Set gtmtypfldindx("rc_lknam","sb_key.key")=9
	;
	Set gtmtypes("rc_lword")="union"
	Set gtmtypes("rc_lword",0)=2
	Set gtmtypes("rc_lword","len")=4
	Set gtmtypes("rc_lword",1,"name")="rc_lword.octet"
	Set gtmtypes("rc_lword",1,"off")=0
	Set gtmtypes("rc_lword",1,"len")=4
	Set gtmtypes("rc_lword",1,"type")="char"
	Set gtmtypfldindx("rc_lword","octet")=1
	Set gtmtypes("rc_lword",2,"name")="rc_lword.value"
	Set gtmtypes("rc_lword",2,"off")=0
	Set gtmtypes("rc_lword",2,"len")=4
	Set gtmtypes("rc_lword",2,"type")="unsigned-int"
	Set gtmtypfldindx("rc_lword","value")=2
	;
	Set gtmtypes("rc_q_hdr")="union"
	Set gtmtypes("rc_q_hdr",0)=26
	Set gtmtypes("rc_q_hdr","len")=14
	Set gtmtypes("rc_q_hdr",1,"name")="rc_q_hdr.r"
	Set gtmtypes("rc_q_hdr",1,"off")=0
	Set gtmtypes("rc_q_hdr",1,"len")=14
	Set gtmtypes("rc_q_hdr",1,"type")="rc_rq_hdr"
	Set gtmtypfldindx("rc_q_hdr","r")=1
	Set gtmtypes("rc_q_hdr",2,"name")="rc_q_hdr.r.len"
	Set gtmtypes("rc_q_hdr",2,"off")=0
	Set gtmtypes("rc_q_hdr",2,"len")=2
	Set gtmtypes("rc_q_hdr",2,"type")="rc_word"
	Set gtmtypfldindx("rc_q_hdr","r.len")=2
	Set gtmtypes("rc_q_hdr",3,"name")="rc_q_hdr.r.typ"
	Set gtmtypes("rc_q_hdr",3,"off")=2
	Set gtmtypes("rc_q_hdr",3,"len")=2
	Set gtmtypes("rc_q_hdr",3,"type")="rc_word"
	Set gtmtypfldindx("rc_q_hdr","r.typ")=3
	Set gtmtypes("rc_q_hdr",4,"name")="rc_q_hdr.r.fmd"
	Set gtmtypes("rc_q_hdr",4,"off")=4
	Set gtmtypes("rc_q_hdr",4,"len")=2
	Set gtmtypes("rc_q_hdr",4,"type")="rc_word"
	Set gtmtypfldindx("rc_q_hdr","r.fmd")=4
	Set gtmtypes("rc_q_hdr",5,"name")="rc_q_hdr.r.pid1"
	Set gtmtypes("rc_q_hdr",5,"off")=6
	Set gtmtypes("rc_q_hdr",5,"len")=2
	Set gtmtypes("rc_q_hdr",5,"type")="rc_word"
	Set gtmtypfldindx("rc_q_hdr","r.pid1")=5
	Set gtmtypes("rc_q_hdr",6,"name")="rc_q_hdr.r.pid2"
	Set gtmtypes("rc_q_hdr",6,"off")=8
	Set gtmtypes("rc_q_hdr",6,"len")=2
	Set gtmtypes("rc_q_hdr",6,"type")="rc_word"
	Set gtmtypfldindx("rc_q_hdr","r.pid2")=6
	Set gtmtypes("rc_q_hdr",7,"name")="rc_q_hdr.r.xdsid"
	Set gtmtypes("rc_q_hdr",7,"off")=10
	Set gtmtypes("rc_q_hdr",7,"len")=4
	Set gtmtypes("rc_q_hdr",7,"type")="rc_xdsid"
	Set gtmtypfldindx("rc_q_hdr","r.xdsid")=7
	Set gtmtypes("rc_q_hdr",8,"name")="rc_q_hdr.a"
	Set gtmtypes("rc_q_hdr",8,"off")=0
	Set gtmtypes("rc_q_hdr",8,"len")=14
	Set gtmtypes("rc_q_hdr",8,"type")="rc_aq_hdr"
	Set gtmtypfldindx("rc_q_hdr","a")=8
	Set gtmtypes("rc_q_hdr",9,"name")="rc_q_hdr.a.len"
	Set gtmtypes("rc_q_hdr",9,"off")=0
	Set gtmtypes("rc_q_hdr",9,"len")=2
	Set gtmtypes("rc_q_hdr",9,"type")="rc_word"
	Set gtmtypfldindx("rc_q_hdr","a.len")=9
	Set gtmtypes("rc_q_hdr",10,"name")="rc_q_hdr.a.len.octet"
	Set gtmtypes("rc_q_hdr",10,"off")=0
	Set gtmtypes("rc_q_hdr",10,"len")=2
	Set gtmtypes("rc_q_hdr",10,"type")="char"
	Set gtmtypfldindx("rc_q_hdr","a.len.octet")=10
	Set gtmtypes("rc_q_hdr",11,"name")="rc_q_hdr.a.len.value"
	Set gtmtypes("rc_q_hdr",11,"off")=0
	Set gtmtypes("rc_q_hdr",11,"len")=2
	Set gtmtypes("rc_q_hdr",11,"type")="unsigned-short"
	Set gtmtypfldindx("rc_q_hdr","a.len.value")=11
	Set gtmtypes("rc_q_hdr",12,"name")="rc_q_hdr.a.typ"
	Set gtmtypes("rc_q_hdr",12,"off")=2
	Set gtmtypes("rc_q_hdr",12,"len")=2
	Set gtmtypes("rc_q_hdr",12,"type")="rc_word"
	Set gtmtypfldindx("rc_q_hdr","a.typ")=12
	Set gtmtypes("rc_q_hdr",13,"name")="rc_q_hdr.a.typ.octet"
	Set gtmtypes("rc_q_hdr",13,"off")=2
	Set gtmtypes("rc_q_hdr",13,"len")=2
	Set gtmtypes("rc_q_hdr",13,"type")="char"
	Set gtmtypfldindx("rc_q_hdr","a.typ.octet")=13
	Set gtmtypes("rc_q_hdr",14,"name")="rc_q_hdr.a.typ.value"
	Set gtmtypes("rc_q_hdr",14,"off")=2
	Set gtmtypes("rc_q_hdr",14,"len")=2
	Set gtmtypes("rc_q_hdr",14,"type")="unsigned-short"
	Set gtmtypfldindx("rc_q_hdr","a.typ.value")=14
	Set gtmtypes("rc_q_hdr",15,"name")="rc_q_hdr.a.erc"
	Set gtmtypes("rc_q_hdr",15,"off")=4
	Set gtmtypes("rc_q_hdr",15,"len")=2
	Set gtmtypes("rc_q_hdr",15,"type")="rc_word"
	Set gtmtypfldindx("rc_q_hdr","a.erc")=15
	Set gtmtypes("rc_q_hdr",16,"name")="rc_q_hdr.a.erc.octet"
	Set gtmtypes("rc_q_hdr",16,"off")=4
	Set gtmtypes("rc_q_hdr",16,"len")=2
	Set gtmtypes("rc_q_hdr",16,"type")="char"
	Set gtmtypfldindx("rc_q_hdr","a.erc.octet")=16
	Set gtmtypes("rc_q_hdr",17,"name")="rc_q_hdr.a.erc.value"
	Set gtmtypes("rc_q_hdr",17,"off")=4
	Set gtmtypes("rc_q_hdr",17,"len")=2
	Set gtmtypes("rc_q_hdr",17,"type")="unsigned-short"
	Set gtmtypfldindx("rc_q_hdr","a.erc.value")=17
	Set gtmtypes("rc_q_hdr",18,"name")="rc_q_hdr.a.pid1"
	Set gtmtypes("rc_q_hdr",18,"off")=6
	Set gtmtypes("rc_q_hdr",18,"len")=2
	Set gtmtypes("rc_q_hdr",18,"type")="rc_word"
	Set gtmtypfldindx("rc_q_hdr","a.pid1")=18
	Set gtmtypes("rc_q_hdr",19,"name")="rc_q_hdr.a.pid1.octet"
	Set gtmtypes("rc_q_hdr",19,"off")=6
	Set gtmtypes("rc_q_hdr",19,"len")=2
	Set gtmtypes("rc_q_hdr",19,"type")="char"
	Set gtmtypfldindx("rc_q_hdr","a.pid1.octet")=19
	Set gtmtypes("rc_q_hdr",20,"name")="rc_q_hdr.a.pid1.value"
	Set gtmtypes("rc_q_hdr",20,"off")=6
	Set gtmtypes("rc_q_hdr",20,"len")=2
	Set gtmtypes("rc_q_hdr",20,"type")="unsigned-short"
	Set gtmtypfldindx("rc_q_hdr","a.pid1.value")=20
	Set gtmtypes("rc_q_hdr",21,"name")="rc_q_hdr.a.pid2"
	Set gtmtypes("rc_q_hdr",21,"off")=8
	Set gtmtypes("rc_q_hdr",21,"len")=2
	Set gtmtypes("rc_q_hdr",21,"type")="rc_word"
	Set gtmtypfldindx("rc_q_hdr","a.pid2")=21
	Set gtmtypes("rc_q_hdr",22,"name")="rc_q_hdr.a.pid2.octet"
	Set gtmtypes("rc_q_hdr",22,"off")=8
	Set gtmtypes("rc_q_hdr",22,"len")=2
	Set gtmtypes("rc_q_hdr",22,"type")="char"
	Set gtmtypfldindx("rc_q_hdr","a.pid2.octet")=22
	Set gtmtypes("rc_q_hdr",23,"name")="rc_q_hdr.a.pid2.value"
	Set gtmtypes("rc_q_hdr",23,"off")=8
	Set gtmtypes("rc_q_hdr",23,"len")=2
	Set gtmtypes("rc_q_hdr",23,"type")="unsigned-short"
	Set gtmtypfldindx("rc_q_hdr","a.pid2.value")=23
	Set gtmtypes("rc_q_hdr",24,"name")="rc_q_hdr.a.xdsid"
	Set gtmtypes("rc_q_hdr",24,"off")=10
	Set gtmtypes("rc_q_hdr",24,"len")=4
	Set gtmtypes("rc_q_hdr",24,"type")="rc_xdsid"
	Set gtmtypfldindx("rc_q_hdr","a.xdsid")=24
	Set gtmtypes("rc_q_hdr",25,"name")="rc_q_hdr.a.xdsid.dsid"
	Set gtmtypes("rc_q_hdr",25,"off")=10
	Set gtmtypes("rc_q_hdr",25,"len")=2
	Set gtmtypes("rc_q_hdr",25,"type")="rc_word"
	Set gtmtypfldindx("rc_q_hdr","a.xdsid.dsid")=25
	Set gtmtypes("rc_q_hdr",26,"name")="rc_q_hdr.a.xdsid.node"
	Set gtmtypes("rc_q_hdr",26,"off")=12
	Set gtmtypes("rc_q_hdr",26,"len")=2
	Set gtmtypes("rc_q_hdr",26,"type")="rc_word"
	Set gtmtypfldindx("rc_q_hdr","a.xdsid.node")=26
	;
	Set gtmtypes("rc_req_getp")="struct"
	Set gtmtypes("rc_req_getp",0)=31
	Set gtmtypes("rc_req_getp","len")=20
	Set gtmtypes("rc_req_getp",1,"name")="rc_req_getp.hdr"
	Set gtmtypes("rc_req_getp",1,"off")=0
	Set gtmtypes("rc_req_getp",1,"len")=14
	Set gtmtypes("rc_req_getp",1,"type")="rc_q_hdr"
	Set gtmtypfldindx("rc_req_getp","hdr")=1
	Set gtmtypes("rc_req_getp",2,"name")="rc_req_getp.hdr.r"
	Set gtmtypes("rc_req_getp",2,"off")=0
	Set gtmtypes("rc_req_getp",2,"len")=14
	Set gtmtypes("rc_req_getp",2,"type")="rc_rq_hdr"
	Set gtmtypfldindx("rc_req_getp","hdr.r")=2
	Set gtmtypes("rc_req_getp",3,"name")="rc_req_getp.hdr.r.len"
	Set gtmtypes("rc_req_getp",3,"off")=0
	Set gtmtypes("rc_req_getp",3,"len")=2
	Set gtmtypes("rc_req_getp",3,"type")="rc_word"
	Set gtmtypfldindx("rc_req_getp","hdr.r.len")=3
	Set gtmtypes("rc_req_getp",4,"name")="rc_req_getp.hdr.r.typ"
	Set gtmtypes("rc_req_getp",4,"off")=2
	Set gtmtypes("rc_req_getp",4,"len")=2
	Set gtmtypes("rc_req_getp",4,"type")="rc_word"
	Set gtmtypfldindx("rc_req_getp","hdr.r.typ")=4
	Set gtmtypes("rc_req_getp",5,"name")="rc_req_getp.hdr.r.fmd"
	Set gtmtypes("rc_req_getp",5,"off")=4
	Set gtmtypes("rc_req_getp",5,"len")=2
	Set gtmtypes("rc_req_getp",5,"type")="rc_word"
	Set gtmtypfldindx("rc_req_getp","hdr.r.fmd")=5
	Set gtmtypes("rc_req_getp",6,"name")="rc_req_getp.hdr.r.pid1"
	Set gtmtypes("rc_req_getp",6,"off")=6
	Set gtmtypes("rc_req_getp",6,"len")=2
	Set gtmtypes("rc_req_getp",6,"type")="rc_word"
	Set gtmtypfldindx("rc_req_getp","hdr.r.pid1")=6
	Set gtmtypes("rc_req_getp",7,"name")="rc_req_getp.hdr.r.pid2"
	Set gtmtypes("rc_req_getp",7,"off")=8
	Set gtmtypes("rc_req_getp",7,"len")=2
	Set gtmtypes("rc_req_getp",7,"type")="rc_word"
	Set gtmtypfldindx("rc_req_getp","hdr.r.pid2")=7
	Set gtmtypes("rc_req_getp",8,"name")="rc_req_getp.hdr.r.xdsid"
	Set gtmtypes("rc_req_getp",8,"off")=10
	Set gtmtypes("rc_req_getp",8,"len")=4
	Set gtmtypes("rc_req_getp",8,"type")="rc_xdsid"
	Set gtmtypfldindx("rc_req_getp","hdr.r.xdsid")=8
	Set gtmtypes("rc_req_getp",9,"name")="rc_req_getp.hdr.a"
	Set gtmtypes("rc_req_getp",9,"off")=0
	Set gtmtypes("rc_req_getp",9,"len")=14
	Set gtmtypes("rc_req_getp",9,"type")="rc_aq_hdr"
	Set gtmtypfldindx("rc_req_getp","hdr.a")=9
	Set gtmtypes("rc_req_getp",10,"name")="rc_req_getp.hdr.a.len"
	Set gtmtypes("rc_req_getp",10,"off")=0
	Set gtmtypes("rc_req_getp",10,"len")=2
	Set gtmtypes("rc_req_getp",10,"type")="rc_word"
	Set gtmtypfldindx("rc_req_getp","hdr.a.len")=10
	Set gtmtypes("rc_req_getp",11,"name")="rc_req_getp.hdr.a.len.octet"
	Set gtmtypes("rc_req_getp",11,"off")=0
	Set gtmtypes("rc_req_getp",11,"len")=2
	Set gtmtypes("rc_req_getp",11,"type")="char"
	Set gtmtypfldindx("rc_req_getp","hdr.a.len.octet")=11
	Set gtmtypes("rc_req_getp",12,"name")="rc_req_getp.hdr.a.len.value"
	Set gtmtypes("rc_req_getp",12,"off")=0
	Set gtmtypes("rc_req_getp",12,"len")=2
	Set gtmtypes("rc_req_getp",12,"type")="unsigned-short"
	Set gtmtypfldindx("rc_req_getp","hdr.a.len.value")=12
	Set gtmtypes("rc_req_getp",13,"name")="rc_req_getp.hdr.a.typ"
	Set gtmtypes("rc_req_getp",13,"off")=2
	Set gtmtypes("rc_req_getp",13,"len")=2
	Set gtmtypes("rc_req_getp",13,"type")="rc_word"
	Set gtmtypfldindx("rc_req_getp","hdr.a.typ")=13
	Set gtmtypes("rc_req_getp",14,"name")="rc_req_getp.hdr.a.typ.octet"
	Set gtmtypes("rc_req_getp",14,"off")=2
	Set gtmtypes("rc_req_getp",14,"len")=2
	Set gtmtypes("rc_req_getp",14,"type")="char"
	Set gtmtypfldindx("rc_req_getp","hdr.a.typ.octet")=14
	Set gtmtypes("rc_req_getp",15,"name")="rc_req_getp.hdr.a.typ.value"
	Set gtmtypes("rc_req_getp",15,"off")=2
	Set gtmtypes("rc_req_getp",15,"len")=2
	Set gtmtypes("rc_req_getp",15,"type")="unsigned-short"
	Set gtmtypfldindx("rc_req_getp","hdr.a.typ.value")=15
	Set gtmtypes("rc_req_getp",16,"name")="rc_req_getp.hdr.a.erc"
	Set gtmtypes("rc_req_getp",16,"off")=4
	Set gtmtypes("rc_req_getp",16,"len")=2
	Set gtmtypes("rc_req_getp",16,"type")="rc_word"
	Set gtmtypfldindx("rc_req_getp","hdr.a.erc")=16
	Set gtmtypes("rc_req_getp",17,"name")="rc_req_getp.hdr.a.erc.octet"
	Set gtmtypes("rc_req_getp",17,"off")=4
	Set gtmtypes("rc_req_getp",17,"len")=2
	Set gtmtypes("rc_req_getp",17,"type")="char"
	Set gtmtypfldindx("rc_req_getp","hdr.a.erc.octet")=17
	Set gtmtypes("rc_req_getp",18,"name")="rc_req_getp.hdr.a.erc.value"
	Set gtmtypes("rc_req_getp",18,"off")=4
	Set gtmtypes("rc_req_getp",18,"len")=2
	Set gtmtypes("rc_req_getp",18,"type")="unsigned-short"
	Set gtmtypfldindx("rc_req_getp","hdr.a.erc.value")=18
	Set gtmtypes("rc_req_getp",19,"name")="rc_req_getp.hdr.a.pid1"
	Set gtmtypes("rc_req_getp",19,"off")=6
	Set gtmtypes("rc_req_getp",19,"len")=2
	Set gtmtypes("rc_req_getp",19,"type")="rc_word"
	Set gtmtypfldindx("rc_req_getp","hdr.a.pid1")=19
	Set gtmtypes("rc_req_getp",20,"name")="rc_req_getp.hdr.a.pid1.octet"
	Set gtmtypes("rc_req_getp",20,"off")=6
	Set gtmtypes("rc_req_getp",20,"len")=2
	Set gtmtypes("rc_req_getp",20,"type")="char"
	Set gtmtypfldindx("rc_req_getp","hdr.a.pid1.octet")=20
	Set gtmtypes("rc_req_getp",21,"name")="rc_req_getp.hdr.a.pid1.value"
	Set gtmtypes("rc_req_getp",21,"off")=6
	Set gtmtypes("rc_req_getp",21,"len")=2
	Set gtmtypes("rc_req_getp",21,"type")="unsigned-short"
	Set gtmtypfldindx("rc_req_getp","hdr.a.pid1.value")=21
	Set gtmtypes("rc_req_getp",22,"name")="rc_req_getp.hdr.a.pid2"
	Set gtmtypes("rc_req_getp",22,"off")=8
	Set gtmtypes("rc_req_getp",22,"len")=2
	Set gtmtypes("rc_req_getp",22,"type")="rc_word"
	Set gtmtypfldindx("rc_req_getp","hdr.a.pid2")=22
	Set gtmtypes("rc_req_getp",23,"name")="rc_req_getp.hdr.a.pid2.octet"
	Set gtmtypes("rc_req_getp",23,"off")=8
	Set gtmtypes("rc_req_getp",23,"len")=2
	Set gtmtypes("rc_req_getp",23,"type")="char"
	Set gtmtypfldindx("rc_req_getp","hdr.a.pid2.octet")=23
	Set gtmtypes("rc_req_getp",24,"name")="rc_req_getp.hdr.a.pid2.value"
	Set gtmtypes("rc_req_getp",24,"off")=8
	Set gtmtypes("rc_req_getp",24,"len")=2
	Set gtmtypes("rc_req_getp",24,"type")="unsigned-short"
	Set gtmtypfldindx("rc_req_getp","hdr.a.pid2.value")=24
	Set gtmtypes("rc_req_getp",25,"name")="rc_req_getp.hdr.a.xdsid"
	Set gtmtypes("rc_req_getp",25,"off")=10
	Set gtmtypes("rc_req_getp",25,"len")=4
	Set gtmtypes("rc_req_getp",25,"type")="rc_xdsid"
	Set gtmtypfldindx("rc_req_getp","hdr.a.xdsid")=25
	Set gtmtypes("rc_req_getp",26,"name")="rc_req_getp.hdr.a.xdsid.dsid"
	Set gtmtypes("rc_req_getp",26,"off")=10
	Set gtmtypes("rc_req_getp",26,"len")=2
	Set gtmtypes("rc_req_getp",26,"type")="rc_word"
	Set gtmtypfldindx("rc_req_getp","hdr.a.xdsid.dsid")=26
	Set gtmtypes("rc_req_getp",27,"name")="rc_req_getp.hdr.a.xdsid.node"
	Set gtmtypes("rc_req_getp",27,"off")=12
	Set gtmtypes("rc_req_getp",27,"len")=2
	Set gtmtypes("rc_req_getp",27,"type")="rc_word"
	Set gtmtypfldindx("rc_req_getp","hdr.a.xdsid.node")=27
	Set gtmtypes("rc_req_getp",28,"name")="rc_req_getp.pageaddr"
	Set gtmtypes("rc_req_getp",28,"off")=14
	Set gtmtypes("rc_req_getp",28,"len")=4
	Set gtmtypes("rc_req_getp",28,"type")="char"
	Set gtmtypfldindx("rc_req_getp","pageaddr")=28
	Set gtmtypes("rc_req_getp",29,"name")="rc_req_getp.offset"
	Set gtmtypes("rc_req_getp",29,"off")=18
	Set gtmtypes("rc_req_getp",29,"len")=2
	Set gtmtypes("rc_req_getp",29,"type")="rc_word"
	Set gtmtypfldindx("rc_req_getp","offset")=29
	Set gtmtypes("rc_req_getp",30,"name")="rc_req_getp.offset.octet"
	Set gtmtypes("rc_req_getp",30,"off")=18
	Set gtmtypes("rc_req_getp",30,"len")=2
	Set gtmtypes("rc_req_getp",30,"type")="char"
	Set gtmtypfldindx("rc_req_getp","offset.octet")=30
	Set gtmtypes("rc_req_getp",31,"name")="rc_req_getp.offset.value"
	Set gtmtypes("rc_req_getp",31,"off")=18
	Set gtmtypes("rc_req_getp",31,"len")=2
	Set gtmtypes("rc_req_getp",31,"type")="unsigned-short"
	Set gtmtypfldindx("rc_req_getp","offset.value")=31
	;
	Set gtmtypes("rc_req_getr")="struct"
	Set gtmtypes("rc_req_getr",0)=30
	Set gtmtypes("rc_req_getr","len")=16
	Set gtmtypes("rc_req_getr",1,"name")="rc_req_getr.hdr"
	Set gtmtypes("rc_req_getr",1,"off")=0
	Set gtmtypes("rc_req_getr",1,"len")=14
	Set gtmtypes("rc_req_getr",1,"type")="rc_q_hdr"
	Set gtmtypfldindx("rc_req_getr","hdr")=1
	Set gtmtypes("rc_req_getr",2,"name")="rc_req_getr.hdr.r"
	Set gtmtypes("rc_req_getr",2,"off")=0
	Set gtmtypes("rc_req_getr",2,"len")=14
	Set gtmtypes("rc_req_getr",2,"type")="rc_rq_hdr"
	Set gtmtypfldindx("rc_req_getr","hdr.r")=2
	Set gtmtypes("rc_req_getr",3,"name")="rc_req_getr.hdr.r.len"
	Set gtmtypes("rc_req_getr",3,"off")=0
	Set gtmtypes("rc_req_getr",3,"len")=2
	Set gtmtypes("rc_req_getr",3,"type")="rc_word"
	Set gtmtypfldindx("rc_req_getr","hdr.r.len")=3
	Set gtmtypes("rc_req_getr",4,"name")="rc_req_getr.hdr.r.typ"
	Set gtmtypes("rc_req_getr",4,"off")=2
	Set gtmtypes("rc_req_getr",4,"len")=2
	Set gtmtypes("rc_req_getr",4,"type")="rc_word"
	Set gtmtypfldindx("rc_req_getr","hdr.r.typ")=4
	Set gtmtypes("rc_req_getr",5,"name")="rc_req_getr.hdr.r.fmd"
	Set gtmtypes("rc_req_getr",5,"off")=4
	Set gtmtypes("rc_req_getr",5,"len")=2
	Set gtmtypes("rc_req_getr",5,"type")="rc_word"
	Set gtmtypfldindx("rc_req_getr","hdr.r.fmd")=5
	Set gtmtypes("rc_req_getr",6,"name")="rc_req_getr.hdr.r.pid1"
	Set gtmtypes("rc_req_getr",6,"off")=6
	Set gtmtypes("rc_req_getr",6,"len")=2
	Set gtmtypes("rc_req_getr",6,"type")="rc_word"
	Set gtmtypfldindx("rc_req_getr","hdr.r.pid1")=6
	Set gtmtypes("rc_req_getr",7,"name")="rc_req_getr.hdr.r.pid2"
	Set gtmtypes("rc_req_getr",7,"off")=8
	Set gtmtypes("rc_req_getr",7,"len")=2
	Set gtmtypes("rc_req_getr",7,"type")="rc_word"
	Set gtmtypfldindx("rc_req_getr","hdr.r.pid2")=7
	Set gtmtypes("rc_req_getr",8,"name")="rc_req_getr.hdr.r.xdsid"
	Set gtmtypes("rc_req_getr",8,"off")=10
	Set gtmtypes("rc_req_getr",8,"len")=4
	Set gtmtypes("rc_req_getr",8,"type")="rc_xdsid"
	Set gtmtypfldindx("rc_req_getr","hdr.r.xdsid")=8
	Set gtmtypes("rc_req_getr",9,"name")="rc_req_getr.hdr.a"
	Set gtmtypes("rc_req_getr",9,"off")=0
	Set gtmtypes("rc_req_getr",9,"len")=14
	Set gtmtypes("rc_req_getr",9,"type")="rc_aq_hdr"
	Set gtmtypfldindx("rc_req_getr","hdr.a")=9
	Set gtmtypes("rc_req_getr",10,"name")="rc_req_getr.hdr.a.len"
	Set gtmtypes("rc_req_getr",10,"off")=0
	Set gtmtypes("rc_req_getr",10,"len")=2
	Set gtmtypes("rc_req_getr",10,"type")="rc_word"
	Set gtmtypfldindx("rc_req_getr","hdr.a.len")=10
	Set gtmtypes("rc_req_getr",11,"name")="rc_req_getr.hdr.a.len.octet"
	Set gtmtypes("rc_req_getr",11,"off")=0
	Set gtmtypes("rc_req_getr",11,"len")=2
	Set gtmtypes("rc_req_getr",11,"type")="char"
	Set gtmtypfldindx("rc_req_getr","hdr.a.len.octet")=11
	Set gtmtypes("rc_req_getr",12,"name")="rc_req_getr.hdr.a.len.value"
	Set gtmtypes("rc_req_getr",12,"off")=0
	Set gtmtypes("rc_req_getr",12,"len")=2
	Set gtmtypes("rc_req_getr",12,"type")="unsigned-short"
	Set gtmtypfldindx("rc_req_getr","hdr.a.len.value")=12
	Set gtmtypes("rc_req_getr",13,"name")="rc_req_getr.hdr.a.typ"
	Set gtmtypes("rc_req_getr",13,"off")=2
	Set gtmtypes("rc_req_getr",13,"len")=2
	Set gtmtypes("rc_req_getr",13,"type")="rc_word"
	Set gtmtypfldindx("rc_req_getr","hdr.a.typ")=13
	Set gtmtypes("rc_req_getr",14,"name")="rc_req_getr.hdr.a.typ.octet"
	Set gtmtypes("rc_req_getr",14,"off")=2
	Set gtmtypes("rc_req_getr",14,"len")=2
	Set gtmtypes("rc_req_getr",14,"type")="char"
	Set gtmtypfldindx("rc_req_getr","hdr.a.typ.octet")=14
	Set gtmtypes("rc_req_getr",15,"name")="rc_req_getr.hdr.a.typ.value"
	Set gtmtypes("rc_req_getr",15,"off")=2
	Set gtmtypes("rc_req_getr",15,"len")=2
	Set gtmtypes("rc_req_getr",15,"type")="unsigned-short"
	Set gtmtypfldindx("rc_req_getr","hdr.a.typ.value")=15
	Set gtmtypes("rc_req_getr",16,"name")="rc_req_getr.hdr.a.erc"
	Set gtmtypes("rc_req_getr",16,"off")=4
	Set gtmtypes("rc_req_getr",16,"len")=2
	Set gtmtypes("rc_req_getr",16,"type")="rc_word"
	Set gtmtypfldindx("rc_req_getr","hdr.a.erc")=16
	Set gtmtypes("rc_req_getr",17,"name")="rc_req_getr.hdr.a.erc.octet"
	Set gtmtypes("rc_req_getr",17,"off")=4
	Set gtmtypes("rc_req_getr",17,"len")=2
	Set gtmtypes("rc_req_getr",17,"type")="char"
	Set gtmtypfldindx("rc_req_getr","hdr.a.erc.octet")=17
	Set gtmtypes("rc_req_getr",18,"name")="rc_req_getr.hdr.a.erc.value"
	Set gtmtypes("rc_req_getr",18,"off")=4
	Set gtmtypes("rc_req_getr",18,"len")=2
	Set gtmtypes("rc_req_getr",18,"type")="unsigned-short"
	Set gtmtypfldindx("rc_req_getr","hdr.a.erc.value")=18
	Set gtmtypes("rc_req_getr",19,"name")="rc_req_getr.hdr.a.pid1"
	Set gtmtypes("rc_req_getr",19,"off")=6
	Set gtmtypes("rc_req_getr",19,"len")=2
	Set gtmtypes("rc_req_getr",19,"type")="rc_word"
	Set gtmtypfldindx("rc_req_getr","hdr.a.pid1")=19
	Set gtmtypes("rc_req_getr",20,"name")="rc_req_getr.hdr.a.pid1.octet"
	Set gtmtypes("rc_req_getr",20,"off")=6
	Set gtmtypes("rc_req_getr",20,"len")=2
	Set gtmtypes("rc_req_getr",20,"type")="char"
	Set gtmtypfldindx("rc_req_getr","hdr.a.pid1.octet")=20
	Set gtmtypes("rc_req_getr",21,"name")="rc_req_getr.hdr.a.pid1.value"
	Set gtmtypes("rc_req_getr",21,"off")=6
	Set gtmtypes("rc_req_getr",21,"len")=2
	Set gtmtypes("rc_req_getr",21,"type")="unsigned-short"
	Set gtmtypfldindx("rc_req_getr","hdr.a.pid1.value")=21
	Set gtmtypes("rc_req_getr",22,"name")="rc_req_getr.hdr.a.pid2"
	Set gtmtypes("rc_req_getr",22,"off")=8
	Set gtmtypes("rc_req_getr",22,"len")=2
	Set gtmtypes("rc_req_getr",22,"type")="rc_word"
	Set gtmtypfldindx("rc_req_getr","hdr.a.pid2")=22
	Set gtmtypes("rc_req_getr",23,"name")="rc_req_getr.hdr.a.pid2.octet"
	Set gtmtypes("rc_req_getr",23,"off")=8
	Set gtmtypes("rc_req_getr",23,"len")=2
	Set gtmtypes("rc_req_getr",23,"type")="char"
	Set gtmtypfldindx("rc_req_getr","hdr.a.pid2.octet")=23
	Set gtmtypes("rc_req_getr",24,"name")="rc_req_getr.hdr.a.pid2.value"
	Set gtmtypes("rc_req_getr",24,"off")=8
	Set gtmtypes("rc_req_getr",24,"len")=2
	Set gtmtypes("rc_req_getr",24,"type")="unsigned-short"
	Set gtmtypfldindx("rc_req_getr","hdr.a.pid2.value")=24
	Set gtmtypes("rc_req_getr",25,"name")="rc_req_getr.hdr.a.xdsid"
	Set gtmtypes("rc_req_getr",25,"off")=10
	Set gtmtypes("rc_req_getr",25,"len")=4
	Set gtmtypes("rc_req_getr",25,"type")="rc_xdsid"
	Set gtmtypfldindx("rc_req_getr","hdr.a.xdsid")=25
	Set gtmtypes("rc_req_getr",26,"name")="rc_req_getr.hdr.a.xdsid.dsid"
	Set gtmtypes("rc_req_getr",26,"off")=10
	Set gtmtypes("rc_req_getr",26,"len")=2
	Set gtmtypes("rc_req_getr",26,"type")="rc_word"
	Set gtmtypfldindx("rc_req_getr","hdr.a.xdsid.dsid")=26
	Set gtmtypes("rc_req_getr",27,"name")="rc_req_getr.hdr.a.xdsid.node"
	Set gtmtypes("rc_req_getr",27,"off")=12
	Set gtmtypes("rc_req_getr",27,"len")=2
	Set gtmtypes("rc_req_getr",27,"type")="rc_word"
	Set gtmtypfldindx("rc_req_getr","hdr.a.xdsid.node")=27
	Set gtmtypes("rc_req_getr",28,"name")="rc_req_getr.key"
	Set gtmtypes("rc_req_getr",28,"off")=14
	Set gtmtypes("rc_req_getr",28,"len")=2
	Set gtmtypes("rc_req_getr",28,"type")="rc_sbkey"
	Set gtmtypfldindx("rc_req_getr","key")=28
	Set gtmtypes("rc_req_getr",29,"name")="rc_req_getr.key.len"
	Set gtmtypes("rc_req_getr",29,"off")=14
	Set gtmtypes("rc_req_getr",29,"len")=1
	Set gtmtypes("rc_req_getr",29,"type")="rc_byte"
	Set gtmtypfldindx("rc_req_getr","key.len")=29
	Set gtmtypes("rc_req_getr",30,"name")="rc_req_getr.key.key"
	Set gtmtypes("rc_req_getr",30,"off")=15
	Set gtmtypes("rc_req_getr",30,"len")=1
	Set gtmtypes("rc_req_getr",30,"type")="char"
	Set gtmtypfldindx("rc_req_getr","key.key")=30
	;
	Set gtmtypes("rc_req_lock")="struct"
	Set gtmtypes("rc_req_lock",0)=40
	Set gtmtypes("rc_req_lock","len")=24
	Set gtmtypes("rc_req_lock",1,"name")="rc_req_lock.hdr"
	Set gtmtypes("rc_req_lock",1,"off")=0
	Set gtmtypes("rc_req_lock",1,"len")=14
	Set gtmtypes("rc_req_lock",1,"type")="rc_q_hdr"
	Set gtmtypfldindx("rc_req_lock","hdr")=1
	Set gtmtypes("rc_req_lock",2,"name")="rc_req_lock.hdr.r"
	Set gtmtypes("rc_req_lock",2,"off")=0
	Set gtmtypes("rc_req_lock",2,"len")=14
	Set gtmtypes("rc_req_lock",2,"type")="rc_rq_hdr"
	Set gtmtypfldindx("rc_req_lock","hdr.r")=2
	Set gtmtypes("rc_req_lock",3,"name")="rc_req_lock.hdr.r.len"
	Set gtmtypes("rc_req_lock",3,"off")=0
	Set gtmtypes("rc_req_lock",3,"len")=2
	Set gtmtypes("rc_req_lock",3,"type")="rc_word"
	Set gtmtypfldindx("rc_req_lock","hdr.r.len")=3
	Set gtmtypes("rc_req_lock",4,"name")="rc_req_lock.hdr.r.typ"
	Set gtmtypes("rc_req_lock",4,"off")=2
	Set gtmtypes("rc_req_lock",4,"len")=2
	Set gtmtypes("rc_req_lock",4,"type")="rc_word"
	Set gtmtypfldindx("rc_req_lock","hdr.r.typ")=4
	Set gtmtypes("rc_req_lock",5,"name")="rc_req_lock.hdr.r.fmd"
	Set gtmtypes("rc_req_lock",5,"off")=4
	Set gtmtypes("rc_req_lock",5,"len")=2
	Set gtmtypes("rc_req_lock",5,"type")="rc_word"
	Set gtmtypfldindx("rc_req_lock","hdr.r.fmd")=5
	Set gtmtypes("rc_req_lock",6,"name")="rc_req_lock.hdr.r.pid1"
	Set gtmtypes("rc_req_lock",6,"off")=6
	Set gtmtypes("rc_req_lock",6,"len")=2
	Set gtmtypes("rc_req_lock",6,"type")="rc_word"
	Set gtmtypfldindx("rc_req_lock","hdr.r.pid1")=6
	Set gtmtypes("rc_req_lock",7,"name")="rc_req_lock.hdr.r.pid2"
	Set gtmtypes("rc_req_lock",7,"off")=8
	Set gtmtypes("rc_req_lock",7,"len")=2
	Set gtmtypes("rc_req_lock",7,"type")="rc_word"
	Set gtmtypfldindx("rc_req_lock","hdr.r.pid2")=7
	Set gtmtypes("rc_req_lock",8,"name")="rc_req_lock.hdr.r.xdsid"
	Set gtmtypes("rc_req_lock",8,"off")=10
	Set gtmtypes("rc_req_lock",8,"len")=4
	Set gtmtypes("rc_req_lock",8,"type")="rc_xdsid"
	Set gtmtypfldindx("rc_req_lock","hdr.r.xdsid")=8
	Set gtmtypes("rc_req_lock",9,"name")="rc_req_lock.hdr.a"
	Set gtmtypes("rc_req_lock",9,"off")=0
	Set gtmtypes("rc_req_lock",9,"len")=14
	Set gtmtypes("rc_req_lock",9,"type")="rc_aq_hdr"
	Set gtmtypfldindx("rc_req_lock","hdr.a")=9
	Set gtmtypes("rc_req_lock",10,"name")="rc_req_lock.hdr.a.len"
	Set gtmtypes("rc_req_lock",10,"off")=0
	Set gtmtypes("rc_req_lock",10,"len")=2
	Set gtmtypes("rc_req_lock",10,"type")="rc_word"
	Set gtmtypfldindx("rc_req_lock","hdr.a.len")=10
	Set gtmtypes("rc_req_lock",11,"name")="rc_req_lock.hdr.a.len.octet"
	Set gtmtypes("rc_req_lock",11,"off")=0
	Set gtmtypes("rc_req_lock",11,"len")=2
	Set gtmtypes("rc_req_lock",11,"type")="char"
	Set gtmtypfldindx("rc_req_lock","hdr.a.len.octet")=11
	Set gtmtypes("rc_req_lock",12,"name")="rc_req_lock.hdr.a.len.value"
	Set gtmtypes("rc_req_lock",12,"off")=0
	Set gtmtypes("rc_req_lock",12,"len")=2
	Set gtmtypes("rc_req_lock",12,"type")="unsigned-short"
	Set gtmtypfldindx("rc_req_lock","hdr.a.len.value")=12
	Set gtmtypes("rc_req_lock",13,"name")="rc_req_lock.hdr.a.typ"
	Set gtmtypes("rc_req_lock",13,"off")=2
	Set gtmtypes("rc_req_lock",13,"len")=2
	Set gtmtypes("rc_req_lock",13,"type")="rc_word"
	Set gtmtypfldindx("rc_req_lock","hdr.a.typ")=13
	Set gtmtypes("rc_req_lock",14,"name")="rc_req_lock.hdr.a.typ.octet"
	Set gtmtypes("rc_req_lock",14,"off")=2
	Set gtmtypes("rc_req_lock",14,"len")=2
	Set gtmtypes("rc_req_lock",14,"type")="char"
	Set gtmtypfldindx("rc_req_lock","hdr.a.typ.octet")=14
	Set gtmtypes("rc_req_lock",15,"name")="rc_req_lock.hdr.a.typ.value"
	Set gtmtypes("rc_req_lock",15,"off")=2
	Set gtmtypes("rc_req_lock",15,"len")=2
	Set gtmtypes("rc_req_lock",15,"type")="unsigned-short"
	Set gtmtypfldindx("rc_req_lock","hdr.a.typ.value")=15
	Set gtmtypes("rc_req_lock",16,"name")="rc_req_lock.hdr.a.erc"
	Set gtmtypes("rc_req_lock",16,"off")=4
	Set gtmtypes("rc_req_lock",16,"len")=2
	Set gtmtypes("rc_req_lock",16,"type")="rc_word"
	Set gtmtypfldindx("rc_req_lock","hdr.a.erc")=16
	Set gtmtypes("rc_req_lock",17,"name")="rc_req_lock.hdr.a.erc.octet"
	Set gtmtypes("rc_req_lock",17,"off")=4
	Set gtmtypes("rc_req_lock",17,"len")=2
	Set gtmtypes("rc_req_lock",17,"type")="char"
	Set gtmtypfldindx("rc_req_lock","hdr.a.erc.octet")=17
	Set gtmtypes("rc_req_lock",18,"name")="rc_req_lock.hdr.a.erc.value"
	Set gtmtypes("rc_req_lock",18,"off")=4
	Set gtmtypes("rc_req_lock",18,"len")=2
	Set gtmtypes("rc_req_lock",18,"type")="unsigned-short"
	Set gtmtypfldindx("rc_req_lock","hdr.a.erc.value")=18
	Set gtmtypes("rc_req_lock",19,"name")="rc_req_lock.hdr.a.pid1"
	Set gtmtypes("rc_req_lock",19,"off")=6
	Set gtmtypes("rc_req_lock",19,"len")=2
	Set gtmtypes("rc_req_lock",19,"type")="rc_word"
	Set gtmtypfldindx("rc_req_lock","hdr.a.pid1")=19
	Set gtmtypes("rc_req_lock",20,"name")="rc_req_lock.hdr.a.pid1.octet"
	Set gtmtypes("rc_req_lock",20,"off")=6
	Set gtmtypes("rc_req_lock",20,"len")=2
	Set gtmtypes("rc_req_lock",20,"type")="char"
	Set gtmtypfldindx("rc_req_lock","hdr.a.pid1.octet")=20
	Set gtmtypes("rc_req_lock",21,"name")="rc_req_lock.hdr.a.pid1.value"
	Set gtmtypes("rc_req_lock",21,"off")=6
	Set gtmtypes("rc_req_lock",21,"len")=2
	Set gtmtypes("rc_req_lock",21,"type")="unsigned-short"
	Set gtmtypfldindx("rc_req_lock","hdr.a.pid1.value")=21
	Set gtmtypes("rc_req_lock",22,"name")="rc_req_lock.hdr.a.pid2"
	Set gtmtypes("rc_req_lock",22,"off")=8
	Set gtmtypes("rc_req_lock",22,"len")=2
	Set gtmtypes("rc_req_lock",22,"type")="rc_word"
	Set gtmtypfldindx("rc_req_lock","hdr.a.pid2")=22
	Set gtmtypes("rc_req_lock",23,"name")="rc_req_lock.hdr.a.pid2.octet"
	Set gtmtypes("rc_req_lock",23,"off")=8
	Set gtmtypes("rc_req_lock",23,"len")=2
	Set gtmtypes("rc_req_lock",23,"type")="char"
	Set gtmtypfldindx("rc_req_lock","hdr.a.pid2.octet")=23
	Set gtmtypes("rc_req_lock",24,"name")="rc_req_lock.hdr.a.pid2.value"
	Set gtmtypes("rc_req_lock",24,"off")=8
	Set gtmtypes("rc_req_lock",24,"len")=2
	Set gtmtypes("rc_req_lock",24,"type")="unsigned-short"
	Set gtmtypfldindx("rc_req_lock","hdr.a.pid2.value")=24
	Set gtmtypes("rc_req_lock",25,"name")="rc_req_lock.hdr.a.xdsid"
	Set gtmtypes("rc_req_lock",25,"off")=10
	Set gtmtypes("rc_req_lock",25,"len")=4
	Set gtmtypes("rc_req_lock",25,"type")="rc_xdsid"
	Set gtmtypfldindx("rc_req_lock","hdr.a.xdsid")=25
	Set gtmtypes("rc_req_lock",26,"name")="rc_req_lock.hdr.a.xdsid.dsid"
	Set gtmtypes("rc_req_lock",26,"off")=10
	Set gtmtypes("rc_req_lock",26,"len")=2
	Set gtmtypes("rc_req_lock",26,"type")="rc_word"
	Set gtmtypfldindx("rc_req_lock","hdr.a.xdsid.dsid")=26
	Set gtmtypes("rc_req_lock",27,"name")="rc_req_lock.hdr.a.xdsid.node"
	Set gtmtypes("rc_req_lock",27,"off")=12
	Set gtmtypes("rc_req_lock",27,"len")=2
	Set gtmtypes("rc_req_lock",27,"type")="rc_word"
	Set gtmtypfldindx("rc_req_lock","hdr.a.xdsid.node")=27
	Set gtmtypes("rc_req_lock",28,"name")="rc_req_lock.nlocks"
	Set gtmtypes("rc_req_lock",28,"off")=14
	Set gtmtypes("rc_req_lock",28,"len")=2
	Set gtmtypes("rc_req_lock",28,"type")="rc_word"
	Set gtmtypfldindx("rc_req_lock","nlocks")=28
	Set gtmtypes("rc_req_lock",29,"name")="rc_req_lock.nlocks.octet"
	Set gtmtypes("rc_req_lock",29,"off")=14
	Set gtmtypes("rc_req_lock",29,"len")=2
	Set gtmtypes("rc_req_lock",29,"type")="char"
	Set gtmtypfldindx("rc_req_lock","nlocks.octet")=29
	Set gtmtypes("rc_req_lock",30,"name")="rc_req_lock.nlocks.value"
	Set gtmtypes("rc_req_lock",30,"off")=14
	Set gtmtypes("rc_req_lock",30,"len")=2
	Set gtmtypes("rc_req_lock",30,"type")="unsigned-short"
	Set gtmtypfldindx("rc_req_lock","nlocks.value")=30
	Set gtmtypes("rc_req_lock",31,"name")="rc_req_lock.dlocks"
	Set gtmtypes("rc_req_lock",31,"off")=16
	Set gtmtypes("rc_req_lock",31,"len")=8
	Set gtmtypes("rc_req_lock",31,"type")="rc_lknam"
	Set gtmtypfldindx("rc_req_lock","dlocks")=31
	Set gtmtypes("rc_req_lock",32,"name")="rc_req_lock.dlocks[0].xdsid"
	Set gtmtypes("rc_req_lock",32,"off")=16
	Set gtmtypes("rc_req_lock",32,"len")=4
	Set gtmtypes("rc_req_lock",32,"type")="rc_xdsid"
	Set gtmtypfldindx("rc_req_lock","dlocks[0].xdsid")=32
	Set gtmtypes("rc_req_lock",33,"name")="rc_req_lock.dlocks[0].xdsid.dsid"
	Set gtmtypes("rc_req_lock",33,"off")=16
	Set gtmtypes("rc_req_lock",33,"len")=2
	Set gtmtypes("rc_req_lock",33,"type")="rc_word"
	Set gtmtypfldindx("rc_req_lock","dlocks[0].xdsid.dsid")=33
	Set gtmtypes("rc_req_lock",34,"name")="rc_req_lock.dlocks[0].xdsid.node"
	Set gtmtypes("rc_req_lock",34,"off")=18
	Set gtmtypes("rc_req_lock",34,"len")=2
	Set gtmtypes("rc_req_lock",34,"type")="rc_word"
	Set gtmtypfldindx("rc_req_lock","dlocks[0].xdsid.node")=34
	Set gtmtypes("rc_req_lock",35,"name")="rc_req_lock.dlocks[0].node_handle"
	Set gtmtypes("rc_req_lock",35,"off")=20
	Set gtmtypes("rc_req_lock",35,"len")=2
	Set gtmtypes("rc_req_lock",35,"type")="rc_word"
	Set gtmtypfldindx("rc_req_lock","dlocks[0].node_handle")=35
	Set gtmtypes("rc_req_lock",36,"name")="rc_req_lock.dlocks[0].node_handle.octet"
	Set gtmtypes("rc_req_lock",36,"off")=20
	Set gtmtypes("rc_req_lock",36,"len")=2
	Set gtmtypes("rc_req_lock",36,"type")="char"
	Set gtmtypfldindx("rc_req_lock","dlocks[0].node_handle.octet")=36
	Set gtmtypes("rc_req_lock",37,"name")="rc_req_lock.dlocks[0].node_handle.value"
	Set gtmtypes("rc_req_lock",37,"off")=20
	Set gtmtypes("rc_req_lock",37,"len")=2
	Set gtmtypes("rc_req_lock",37,"type")="unsigned-short"
	Set gtmtypfldindx("rc_req_lock","dlocks[0].node_handle.value")=37
	Set gtmtypes("rc_req_lock",38,"name")="rc_req_lock.dlocks[0].sb_key"
	Set gtmtypes("rc_req_lock",38,"off")=22
	Set gtmtypes("rc_req_lock",38,"len")=2
	Set gtmtypes("rc_req_lock",38,"type")="rc_sbkey"
	Set gtmtypfldindx("rc_req_lock","dlocks[0].sb_key")=38
	Set gtmtypes("rc_req_lock",39,"name")="rc_req_lock.dlocks[0].sb_key.len"
	Set gtmtypes("rc_req_lock",39,"off")=22
	Set gtmtypes("rc_req_lock",39,"len")=1
	Set gtmtypes("rc_req_lock",39,"type")="rc_byte"
	Set gtmtypfldindx("rc_req_lock","dlocks[0].sb_key.len")=39
	Set gtmtypes("rc_req_lock",40,"name")="rc_req_lock.dlocks[0].sb_key.key"
	Set gtmtypes("rc_req_lock",40,"off")=23
	Set gtmtypes("rc_req_lock",40,"len")=1
	Set gtmtypes("rc_req_lock",40,"type")="char"
	Set gtmtypfldindx("rc_req_lock","dlocks[0].sb_key.key")=40
	;
	Set gtmtypes("rc_req_logn")="struct"
	Set gtmtypes("rc_req_logn",0)=29
	Set gtmtypes("rc_req_logn","len")=250
	Set gtmtypes("rc_req_logn",1,"name")="rc_req_logn.hdr"
	Set gtmtypes("rc_req_logn",1,"off")=0
	Set gtmtypes("rc_req_logn",1,"len")=14
	Set gtmtypes("rc_req_logn",1,"type")="rc_q_hdr"
	Set gtmtypfldindx("rc_req_logn","hdr")=1
	Set gtmtypes("rc_req_logn",2,"name")="rc_req_logn.hdr.r"
	Set gtmtypes("rc_req_logn",2,"off")=0
	Set gtmtypes("rc_req_logn",2,"len")=14
	Set gtmtypes("rc_req_logn",2,"type")="rc_rq_hdr"
	Set gtmtypfldindx("rc_req_logn","hdr.r")=2
	Set gtmtypes("rc_req_logn",3,"name")="rc_req_logn.hdr.r.len"
	Set gtmtypes("rc_req_logn",3,"off")=0
	Set gtmtypes("rc_req_logn",3,"len")=2
	Set gtmtypes("rc_req_logn",3,"type")="rc_word"
	Set gtmtypfldindx("rc_req_logn","hdr.r.len")=3
	Set gtmtypes("rc_req_logn",4,"name")="rc_req_logn.hdr.r.typ"
	Set gtmtypes("rc_req_logn",4,"off")=2
	Set gtmtypes("rc_req_logn",4,"len")=2
	Set gtmtypes("rc_req_logn",4,"type")="rc_word"
	Set gtmtypfldindx("rc_req_logn","hdr.r.typ")=4
	Set gtmtypes("rc_req_logn",5,"name")="rc_req_logn.hdr.r.fmd"
	Set gtmtypes("rc_req_logn",5,"off")=4
	Set gtmtypes("rc_req_logn",5,"len")=2
	Set gtmtypes("rc_req_logn",5,"type")="rc_word"
	Set gtmtypfldindx("rc_req_logn","hdr.r.fmd")=5
	Set gtmtypes("rc_req_logn",6,"name")="rc_req_logn.hdr.r.pid1"
	Set gtmtypes("rc_req_logn",6,"off")=6
	Set gtmtypes("rc_req_logn",6,"len")=2
	Set gtmtypes("rc_req_logn",6,"type")="rc_word"
	Set gtmtypfldindx("rc_req_logn","hdr.r.pid1")=6
	Set gtmtypes("rc_req_logn",7,"name")="rc_req_logn.hdr.r.pid2"
	Set gtmtypes("rc_req_logn",7,"off")=8
	Set gtmtypes("rc_req_logn",7,"len")=2
	Set gtmtypes("rc_req_logn",7,"type")="rc_word"
	Set gtmtypfldindx("rc_req_logn","hdr.r.pid2")=7
	Set gtmtypes("rc_req_logn",8,"name")="rc_req_logn.hdr.r.xdsid"
	Set gtmtypes("rc_req_logn",8,"off")=10
	Set gtmtypes("rc_req_logn",8,"len")=4
	Set gtmtypes("rc_req_logn",8,"type")="rc_xdsid"
	Set gtmtypfldindx("rc_req_logn","hdr.r.xdsid")=8
	Set gtmtypes("rc_req_logn",9,"name")="rc_req_logn.hdr.a"
	Set gtmtypes("rc_req_logn",9,"off")=0
	Set gtmtypes("rc_req_logn",9,"len")=14
	Set gtmtypes("rc_req_logn",9,"type")="rc_aq_hdr"
	Set gtmtypfldindx("rc_req_logn","hdr.a")=9
	Set gtmtypes("rc_req_logn",10,"name")="rc_req_logn.hdr.a.len"
	Set gtmtypes("rc_req_logn",10,"off")=0
	Set gtmtypes("rc_req_logn",10,"len")=2
	Set gtmtypes("rc_req_logn",10,"type")="rc_word"
	Set gtmtypfldindx("rc_req_logn","hdr.a.len")=10
	Set gtmtypes("rc_req_logn",11,"name")="rc_req_logn.hdr.a.len.octet"
	Set gtmtypes("rc_req_logn",11,"off")=0
	Set gtmtypes("rc_req_logn",11,"len")=2
	Set gtmtypes("rc_req_logn",11,"type")="char"
	Set gtmtypfldindx("rc_req_logn","hdr.a.len.octet")=11
	Set gtmtypes("rc_req_logn",12,"name")="rc_req_logn.hdr.a.len.value"
	Set gtmtypes("rc_req_logn",12,"off")=0
	Set gtmtypes("rc_req_logn",12,"len")=2
	Set gtmtypes("rc_req_logn",12,"type")="unsigned-short"
	Set gtmtypfldindx("rc_req_logn","hdr.a.len.value")=12
	Set gtmtypes("rc_req_logn",13,"name")="rc_req_logn.hdr.a.typ"
	Set gtmtypes("rc_req_logn",13,"off")=2
	Set gtmtypes("rc_req_logn",13,"len")=2
	Set gtmtypes("rc_req_logn",13,"type")="rc_word"
	Set gtmtypfldindx("rc_req_logn","hdr.a.typ")=13
	Set gtmtypes("rc_req_logn",14,"name")="rc_req_logn.hdr.a.typ.octet"
	Set gtmtypes("rc_req_logn",14,"off")=2
	Set gtmtypes("rc_req_logn",14,"len")=2
	Set gtmtypes("rc_req_logn",14,"type")="char"
	Set gtmtypfldindx("rc_req_logn","hdr.a.typ.octet")=14
	Set gtmtypes("rc_req_logn",15,"name")="rc_req_logn.hdr.a.typ.value"
	Set gtmtypes("rc_req_logn",15,"off")=2
	Set gtmtypes("rc_req_logn",15,"len")=2
	Set gtmtypes("rc_req_logn",15,"type")="unsigned-short"
	Set gtmtypfldindx("rc_req_logn","hdr.a.typ.value")=15
	Set gtmtypes("rc_req_logn",16,"name")="rc_req_logn.hdr.a.erc"
	Set gtmtypes("rc_req_logn",16,"off")=4
	Set gtmtypes("rc_req_logn",16,"len")=2
	Set gtmtypes("rc_req_logn",16,"type")="rc_word"
	Set gtmtypfldindx("rc_req_logn","hdr.a.erc")=16
	Set gtmtypes("rc_req_logn",17,"name")="rc_req_logn.hdr.a.erc.octet"
	Set gtmtypes("rc_req_logn",17,"off")=4
	Set gtmtypes("rc_req_logn",17,"len")=2
	Set gtmtypes("rc_req_logn",17,"type")="char"
	Set gtmtypfldindx("rc_req_logn","hdr.a.erc.octet")=17
	Set gtmtypes("rc_req_logn",18,"name")="rc_req_logn.hdr.a.erc.value"
	Set gtmtypes("rc_req_logn",18,"off")=4
	Set gtmtypes("rc_req_logn",18,"len")=2
	Set gtmtypes("rc_req_logn",18,"type")="unsigned-short"
	Set gtmtypfldindx("rc_req_logn","hdr.a.erc.value")=18
	Set gtmtypes("rc_req_logn",19,"name")="rc_req_logn.hdr.a.pid1"
	Set gtmtypes("rc_req_logn",19,"off")=6
	Set gtmtypes("rc_req_logn",19,"len")=2
	Set gtmtypes("rc_req_logn",19,"type")="rc_word"
	Set gtmtypfldindx("rc_req_logn","hdr.a.pid1")=19
	Set gtmtypes("rc_req_logn",20,"name")="rc_req_logn.hdr.a.pid1.octet"
	Set gtmtypes("rc_req_logn",20,"off")=6
	Set gtmtypes("rc_req_logn",20,"len")=2
	Set gtmtypes("rc_req_logn",20,"type")="char"
	Set gtmtypfldindx("rc_req_logn","hdr.a.pid1.octet")=20
	Set gtmtypes("rc_req_logn",21,"name")="rc_req_logn.hdr.a.pid1.value"
	Set gtmtypes("rc_req_logn",21,"off")=6
	Set gtmtypes("rc_req_logn",21,"len")=2
	Set gtmtypes("rc_req_logn",21,"type")="unsigned-short"
	Set gtmtypfldindx("rc_req_logn","hdr.a.pid1.value")=21
	Set gtmtypes("rc_req_logn",22,"name")="rc_req_logn.hdr.a.pid2"
	Set gtmtypes("rc_req_logn",22,"off")=8
	Set gtmtypes("rc_req_logn",22,"len")=2
	Set gtmtypes("rc_req_logn",22,"type")="rc_word"
	Set gtmtypfldindx("rc_req_logn","hdr.a.pid2")=22
	Set gtmtypes("rc_req_logn",23,"name")="rc_req_logn.hdr.a.pid2.octet"
	Set gtmtypes("rc_req_logn",23,"off")=8
	Set gtmtypes("rc_req_logn",23,"len")=2
	Set gtmtypes("rc_req_logn",23,"type")="char"
	Set gtmtypfldindx("rc_req_logn","hdr.a.pid2.octet")=23
	Set gtmtypes("rc_req_logn",24,"name")="rc_req_logn.hdr.a.pid2.value"
	Set gtmtypes("rc_req_logn",24,"off")=8
	Set gtmtypes("rc_req_logn",24,"len")=2
	Set gtmtypes("rc_req_logn",24,"type")="unsigned-short"
	Set gtmtypfldindx("rc_req_logn","hdr.a.pid2.value")=24
	Set gtmtypes("rc_req_logn",25,"name")="rc_req_logn.hdr.a.xdsid"
	Set gtmtypes("rc_req_logn",25,"off")=10
	Set gtmtypes("rc_req_logn",25,"len")=4
	Set gtmtypes("rc_req_logn",25,"type")="rc_xdsid"
	Set gtmtypfldindx("rc_req_logn","hdr.a.xdsid")=25
	Set gtmtypes("rc_req_logn",26,"name")="rc_req_logn.hdr.a.xdsid.dsid"
	Set gtmtypes("rc_req_logn",26,"off")=10
	Set gtmtypes("rc_req_logn",26,"len")=2
	Set gtmtypes("rc_req_logn",26,"type")="rc_word"
	Set gtmtypfldindx("rc_req_logn","hdr.a.xdsid.dsid")=26
	Set gtmtypes("rc_req_logn",27,"name")="rc_req_logn.hdr.a.xdsid.node"
	Set gtmtypes("rc_req_logn",27,"off")=12
	Set gtmtypes("rc_req_logn",27,"len")=2
	Set gtmtypes("rc_req_logn",27,"type")="rc_word"
	Set gtmtypfldindx("rc_req_logn","hdr.a.xdsid.node")=27
	Set gtmtypes("rc_req_logn",28,"name")="rc_req_logn.license_num"
	Set gtmtypes("rc_req_logn",28,"off")=14
	Set gtmtypes("rc_req_logn",28,"len")=12
	Set gtmtypes("rc_req_logn",28,"type")="char"
	Set gtmtypfldindx("rc_req_logn","license_num")=28
	Set gtmtypes("rc_req_logn",29,"name")="rc_req_logn.license_blk"
	Set gtmtypes("rc_req_logn",29,"off")=26
	Set gtmtypes("rc_req_logn",29,"len")=224
	Set gtmtypes("rc_req_logn",29,"type")="char"
	Set gtmtypfldindx("rc_req_logn","license_blk")=29
	;
	Set gtmtypes("rc_rq_hdr")="struct"
	Set gtmtypes("rc_rq_hdr",0)=18
	Set gtmtypes("rc_rq_hdr","len")=14
	Set gtmtypes("rc_rq_hdr",1,"name")="rc_rq_hdr.len"
	Set gtmtypes("rc_rq_hdr",1,"off")=0
	Set gtmtypes("rc_rq_hdr",1,"len")=2
	Set gtmtypes("rc_rq_hdr",1,"type")="rc_word"
	Set gtmtypfldindx("rc_rq_hdr","len")=1
	Set gtmtypes("rc_rq_hdr",2,"name")="rc_rq_hdr.len.octet"
	Set gtmtypes("rc_rq_hdr",2,"off")=0
	Set gtmtypes("rc_rq_hdr",2,"len")=2
	Set gtmtypes("rc_rq_hdr",2,"type")="char"
	Set gtmtypfldindx("rc_rq_hdr","len.octet")=2
	Set gtmtypes("rc_rq_hdr",3,"name")="rc_rq_hdr.len.value"
	Set gtmtypes("rc_rq_hdr",3,"off")=0
	Set gtmtypes("rc_rq_hdr",3,"len")=2
	Set gtmtypes("rc_rq_hdr",3,"type")="unsigned-short"
	Set gtmtypfldindx("rc_rq_hdr","len.value")=3
	Set gtmtypes("rc_rq_hdr",4,"name")="rc_rq_hdr.typ"
	Set gtmtypes("rc_rq_hdr",4,"off")=2
	Set gtmtypes("rc_rq_hdr",4,"len")=2
	Set gtmtypes("rc_rq_hdr",4,"type")="rc_word"
	Set gtmtypfldindx("rc_rq_hdr","typ")=4
	Set gtmtypes("rc_rq_hdr",5,"name")="rc_rq_hdr.typ.octet"
	Set gtmtypes("rc_rq_hdr",5,"off")=2
	Set gtmtypes("rc_rq_hdr",5,"len")=2
	Set gtmtypes("rc_rq_hdr",5,"type")="char"
	Set gtmtypfldindx("rc_rq_hdr","typ.octet")=5
	Set gtmtypes("rc_rq_hdr",6,"name")="rc_rq_hdr.typ.value"
	Set gtmtypes("rc_rq_hdr",6,"off")=2
	Set gtmtypes("rc_rq_hdr",6,"len")=2
	Set gtmtypes("rc_rq_hdr",6,"type")="unsigned-short"
	Set gtmtypfldindx("rc_rq_hdr","typ.value")=6
	Set gtmtypes("rc_rq_hdr",7,"name")="rc_rq_hdr.fmd"
	Set gtmtypes("rc_rq_hdr",7,"off")=4
	Set gtmtypes("rc_rq_hdr",7,"len")=2
	Set gtmtypes("rc_rq_hdr",7,"type")="rc_word"
	Set gtmtypfldindx("rc_rq_hdr","fmd")=7
	Set gtmtypes("rc_rq_hdr",8,"name")="rc_rq_hdr.fmd.octet"
	Set gtmtypes("rc_rq_hdr",8,"off")=4
	Set gtmtypes("rc_rq_hdr",8,"len")=2
	Set gtmtypes("rc_rq_hdr",8,"type")="char"
	Set gtmtypfldindx("rc_rq_hdr","fmd.octet")=8
	Set gtmtypes("rc_rq_hdr",9,"name")="rc_rq_hdr.fmd.value"
	Set gtmtypes("rc_rq_hdr",9,"off")=4
	Set gtmtypes("rc_rq_hdr",9,"len")=2
	Set gtmtypes("rc_rq_hdr",9,"type")="unsigned-short"
	Set gtmtypfldindx("rc_rq_hdr","fmd.value")=9
	Set gtmtypes("rc_rq_hdr",10,"name")="rc_rq_hdr.pid1"
	Set gtmtypes("rc_rq_hdr",10,"off")=6
	Set gtmtypes("rc_rq_hdr",10,"len")=2
	Set gtmtypes("rc_rq_hdr",10,"type")="rc_word"
	Set gtmtypfldindx("rc_rq_hdr","pid1")=10
	Set gtmtypes("rc_rq_hdr",11,"name")="rc_rq_hdr.pid1.octet"
	Set gtmtypes("rc_rq_hdr",11,"off")=6
	Set gtmtypes("rc_rq_hdr",11,"len")=2
	Set gtmtypes("rc_rq_hdr",11,"type")="char"
	Set gtmtypfldindx("rc_rq_hdr","pid1.octet")=11
	Set gtmtypes("rc_rq_hdr",12,"name")="rc_rq_hdr.pid1.value"
	Set gtmtypes("rc_rq_hdr",12,"off")=6
	Set gtmtypes("rc_rq_hdr",12,"len")=2
	Set gtmtypes("rc_rq_hdr",12,"type")="unsigned-short"
	Set gtmtypfldindx("rc_rq_hdr","pid1.value")=12
	Set gtmtypes("rc_rq_hdr",13,"name")="rc_rq_hdr.pid2"
	Set gtmtypes("rc_rq_hdr",13,"off")=8
	Set gtmtypes("rc_rq_hdr",13,"len")=2
	Set gtmtypes("rc_rq_hdr",13,"type")="rc_word"
	Set gtmtypfldindx("rc_rq_hdr","pid2")=13
	Set gtmtypes("rc_rq_hdr",14,"name")="rc_rq_hdr.pid2.octet"
	Set gtmtypes("rc_rq_hdr",14,"off")=8
	Set gtmtypes("rc_rq_hdr",14,"len")=2
	Set gtmtypes("rc_rq_hdr",14,"type")="char"
	Set gtmtypfldindx("rc_rq_hdr","pid2.octet")=14
	Set gtmtypes("rc_rq_hdr",15,"name")="rc_rq_hdr.pid2.value"
	Set gtmtypes("rc_rq_hdr",15,"off")=8
	Set gtmtypes("rc_rq_hdr",15,"len")=2
	Set gtmtypes("rc_rq_hdr",15,"type")="unsigned-short"
	Set gtmtypfldindx("rc_rq_hdr","pid2.value")=15
	Set gtmtypes("rc_rq_hdr",16,"name")="rc_rq_hdr.xdsid"
	Set gtmtypes("rc_rq_hdr",16,"off")=10
	Set gtmtypes("rc_rq_hdr",16,"len")=4
	Set gtmtypes("rc_rq_hdr",16,"type")="rc_xdsid"
	Set gtmtypfldindx("rc_rq_hdr","xdsid")=16
	Set gtmtypes("rc_rq_hdr",17,"name")="rc_rq_hdr.xdsid.dsid"
	Set gtmtypes("rc_rq_hdr",17,"off")=10
	Set gtmtypes("rc_rq_hdr",17,"len")=2
	Set gtmtypes("rc_rq_hdr",17,"type")="rc_word"
	Set gtmtypfldindx("rc_rq_hdr","xdsid.dsid")=17
	Set gtmtypes("rc_rq_hdr",18,"name")="rc_rq_hdr.xdsid.node"
	Set gtmtypes("rc_rq_hdr",18,"off")=12
	Set gtmtypes("rc_rq_hdr",18,"len")=2
	Set gtmtypes("rc_rq_hdr",18,"type")="rc_word"
	Set gtmtypfldindx("rc_rq_hdr","xdsid.node")=18
	;
	Set gtmtypes("rc_rsp_logn")="struct"
	Set gtmtypes("rc_rsp_logn",0)=41
	Set gtmtypes("rc_rsp_logn","len")=250
	Set gtmtypes("rc_rsp_logn",1,"name")="rc_rsp_logn.hdr"
	Set gtmtypes("rc_rsp_logn",1,"off")=0
	Set gtmtypes("rc_rsp_logn",1,"len")=14
	Set gtmtypes("rc_rsp_logn",1,"type")="rc_q_hdr"
	Set gtmtypfldindx("rc_rsp_logn","hdr")=1
	Set gtmtypes("rc_rsp_logn",2,"name")="rc_rsp_logn.hdr.r"
	Set gtmtypes("rc_rsp_logn",2,"off")=0
	Set gtmtypes("rc_rsp_logn",2,"len")=14
	Set gtmtypes("rc_rsp_logn",2,"type")="rc_rq_hdr"
	Set gtmtypfldindx("rc_rsp_logn","hdr.r")=2
	Set gtmtypes("rc_rsp_logn",3,"name")="rc_rsp_logn.hdr.r.len"
	Set gtmtypes("rc_rsp_logn",3,"off")=0
	Set gtmtypes("rc_rsp_logn",3,"len")=2
	Set gtmtypes("rc_rsp_logn",3,"type")="rc_word"
	Set gtmtypfldindx("rc_rsp_logn","hdr.r.len")=3
	Set gtmtypes("rc_rsp_logn",4,"name")="rc_rsp_logn.hdr.r.typ"
	Set gtmtypes("rc_rsp_logn",4,"off")=2
	Set gtmtypes("rc_rsp_logn",4,"len")=2
	Set gtmtypes("rc_rsp_logn",4,"type")="rc_word"
	Set gtmtypfldindx("rc_rsp_logn","hdr.r.typ")=4
	Set gtmtypes("rc_rsp_logn",5,"name")="rc_rsp_logn.hdr.r.fmd"
	Set gtmtypes("rc_rsp_logn",5,"off")=4
	Set gtmtypes("rc_rsp_logn",5,"len")=2
	Set gtmtypes("rc_rsp_logn",5,"type")="rc_word"
	Set gtmtypfldindx("rc_rsp_logn","hdr.r.fmd")=5
	Set gtmtypes("rc_rsp_logn",6,"name")="rc_rsp_logn.hdr.r.pid1"
	Set gtmtypes("rc_rsp_logn",6,"off")=6
	Set gtmtypes("rc_rsp_logn",6,"len")=2
	Set gtmtypes("rc_rsp_logn",6,"type")="rc_word"
	Set gtmtypfldindx("rc_rsp_logn","hdr.r.pid1")=6
	Set gtmtypes("rc_rsp_logn",7,"name")="rc_rsp_logn.hdr.r.pid2"
	Set gtmtypes("rc_rsp_logn",7,"off")=8
	Set gtmtypes("rc_rsp_logn",7,"len")=2
	Set gtmtypes("rc_rsp_logn",7,"type")="rc_word"
	Set gtmtypfldindx("rc_rsp_logn","hdr.r.pid2")=7
	Set gtmtypes("rc_rsp_logn",8,"name")="rc_rsp_logn.hdr.r.xdsid"
	Set gtmtypes("rc_rsp_logn",8,"off")=10
	Set gtmtypes("rc_rsp_logn",8,"len")=4
	Set gtmtypes("rc_rsp_logn",8,"type")="rc_xdsid"
	Set gtmtypfldindx("rc_rsp_logn","hdr.r.xdsid")=8
	Set gtmtypes("rc_rsp_logn",9,"name")="rc_rsp_logn.hdr.a"
	Set gtmtypes("rc_rsp_logn",9,"off")=0
	Set gtmtypes("rc_rsp_logn",9,"len")=14
	Set gtmtypes("rc_rsp_logn",9,"type")="rc_aq_hdr"
	Set gtmtypfldindx("rc_rsp_logn","hdr.a")=9
	Set gtmtypes("rc_rsp_logn",10,"name")="rc_rsp_logn.hdr.a.len"
	Set gtmtypes("rc_rsp_logn",10,"off")=0
	Set gtmtypes("rc_rsp_logn",10,"len")=2
	Set gtmtypes("rc_rsp_logn",10,"type")="rc_word"
	Set gtmtypfldindx("rc_rsp_logn","hdr.a.len")=10
	Set gtmtypes("rc_rsp_logn",11,"name")="rc_rsp_logn.hdr.a.len.octet"
	Set gtmtypes("rc_rsp_logn",11,"off")=0
	Set gtmtypes("rc_rsp_logn",11,"len")=2
	Set gtmtypes("rc_rsp_logn",11,"type")="char"
	Set gtmtypfldindx("rc_rsp_logn","hdr.a.len.octet")=11
	Set gtmtypes("rc_rsp_logn",12,"name")="rc_rsp_logn.hdr.a.len.value"
	Set gtmtypes("rc_rsp_logn",12,"off")=0
	Set gtmtypes("rc_rsp_logn",12,"len")=2
	Set gtmtypes("rc_rsp_logn",12,"type")="unsigned-short"
	Set gtmtypfldindx("rc_rsp_logn","hdr.a.len.value")=12
	Set gtmtypes("rc_rsp_logn",13,"name")="rc_rsp_logn.hdr.a.typ"
	Set gtmtypes("rc_rsp_logn",13,"off")=2
	Set gtmtypes("rc_rsp_logn",13,"len")=2
	Set gtmtypes("rc_rsp_logn",13,"type")="rc_word"
	Set gtmtypfldindx("rc_rsp_logn","hdr.a.typ")=13
	Set gtmtypes("rc_rsp_logn",14,"name")="rc_rsp_logn.hdr.a.typ.octet"
	Set gtmtypes("rc_rsp_logn",14,"off")=2
	Set gtmtypes("rc_rsp_logn",14,"len")=2
	Set gtmtypes("rc_rsp_logn",14,"type")="char"
	Set gtmtypfldindx("rc_rsp_logn","hdr.a.typ.octet")=14
	Set gtmtypes("rc_rsp_logn",15,"name")="rc_rsp_logn.hdr.a.typ.value"
	Set gtmtypes("rc_rsp_logn",15,"off")=2
	Set gtmtypes("rc_rsp_logn",15,"len")=2
	Set gtmtypes("rc_rsp_logn",15,"type")="unsigned-short"
	Set gtmtypfldindx("rc_rsp_logn","hdr.a.typ.value")=15
	Set gtmtypes("rc_rsp_logn",16,"name")="rc_rsp_logn.hdr.a.erc"
	Set gtmtypes("rc_rsp_logn",16,"off")=4
	Set gtmtypes("rc_rsp_logn",16,"len")=2
	Set gtmtypes("rc_rsp_logn",16,"type")="rc_word"
	Set gtmtypfldindx("rc_rsp_logn","hdr.a.erc")=16
	Set gtmtypes("rc_rsp_logn",17,"name")="rc_rsp_logn.hdr.a.erc.octet"
	Set gtmtypes("rc_rsp_logn",17,"off")=4
	Set gtmtypes("rc_rsp_logn",17,"len")=2
	Set gtmtypes("rc_rsp_logn",17,"type")="char"
	Set gtmtypfldindx("rc_rsp_logn","hdr.a.erc.octet")=17
	Set gtmtypes("rc_rsp_logn",18,"name")="rc_rsp_logn.hdr.a.erc.value"
	Set gtmtypes("rc_rsp_logn",18,"off")=4
	Set gtmtypes("rc_rsp_logn",18,"len")=2
	Set gtmtypes("rc_rsp_logn",18,"type")="unsigned-short"
	Set gtmtypfldindx("rc_rsp_logn","hdr.a.erc.value")=18
	Set gtmtypes("rc_rsp_logn",19,"name")="rc_rsp_logn.hdr.a.pid1"
	Set gtmtypes("rc_rsp_logn",19,"off")=6
	Set gtmtypes("rc_rsp_logn",19,"len")=2
	Set gtmtypes("rc_rsp_logn",19,"type")="rc_word"
	Set gtmtypfldindx("rc_rsp_logn","hdr.a.pid1")=19
	Set gtmtypes("rc_rsp_logn",20,"name")="rc_rsp_logn.hdr.a.pid1.octet"
	Set gtmtypes("rc_rsp_logn",20,"off")=6
	Set gtmtypes("rc_rsp_logn",20,"len")=2
	Set gtmtypes("rc_rsp_logn",20,"type")="char"
	Set gtmtypfldindx("rc_rsp_logn","hdr.a.pid1.octet")=20
	Set gtmtypes("rc_rsp_logn",21,"name")="rc_rsp_logn.hdr.a.pid1.value"
	Set gtmtypes("rc_rsp_logn",21,"off")=6
	Set gtmtypes("rc_rsp_logn",21,"len")=2
	Set gtmtypes("rc_rsp_logn",21,"type")="unsigned-short"
	Set gtmtypfldindx("rc_rsp_logn","hdr.a.pid1.value")=21
	Set gtmtypes("rc_rsp_logn",22,"name")="rc_rsp_logn.hdr.a.pid2"
	Set gtmtypes("rc_rsp_logn",22,"off")=8
	Set gtmtypes("rc_rsp_logn",22,"len")=2
	Set gtmtypes("rc_rsp_logn",22,"type")="rc_word"
	Set gtmtypfldindx("rc_rsp_logn","hdr.a.pid2")=22
	Set gtmtypes("rc_rsp_logn",23,"name")="rc_rsp_logn.hdr.a.pid2.octet"
	Set gtmtypes("rc_rsp_logn",23,"off")=8
	Set gtmtypes("rc_rsp_logn",23,"len")=2
	Set gtmtypes("rc_rsp_logn",23,"type")="char"
	Set gtmtypfldindx("rc_rsp_logn","hdr.a.pid2.octet")=23
	Set gtmtypes("rc_rsp_logn",24,"name")="rc_rsp_logn.hdr.a.pid2.value"
	Set gtmtypes("rc_rsp_logn",24,"off")=8
	Set gtmtypes("rc_rsp_logn",24,"len")=2
	Set gtmtypes("rc_rsp_logn",24,"type")="unsigned-short"
	Set gtmtypfldindx("rc_rsp_logn","hdr.a.pid2.value")=24
	Set gtmtypes("rc_rsp_logn",25,"name")="rc_rsp_logn.hdr.a.xdsid"
	Set gtmtypes("rc_rsp_logn",25,"off")=10
	Set gtmtypes("rc_rsp_logn",25,"len")=4
	Set gtmtypes("rc_rsp_logn",25,"type")="rc_xdsid"
	Set gtmtypfldindx("rc_rsp_logn","hdr.a.xdsid")=25
	Set gtmtypes("rc_rsp_logn",26,"name")="rc_rsp_logn.hdr.a.xdsid.dsid"
	Set gtmtypes("rc_rsp_logn",26,"off")=10
	Set gtmtypes("rc_rsp_logn",26,"len")=2
	Set gtmtypes("rc_rsp_logn",26,"type")="rc_word"
	Set gtmtypfldindx("rc_rsp_logn","hdr.a.xdsid.dsid")=26
	Set gtmtypes("rc_rsp_logn",27,"name")="rc_rsp_logn.hdr.a.xdsid.node"
	Set gtmtypes("rc_rsp_logn",27,"off")=12
	Set gtmtypes("rc_rsp_logn",27,"len")=2
	Set gtmtypes("rc_rsp_logn",27,"type")="rc_word"
	Set gtmtypfldindx("rc_rsp_logn","hdr.a.xdsid.node")=27
	Set gtmtypes("rc_rsp_logn",28,"name")="rc_rsp_logn.version"
	Set gtmtypes("rc_rsp_logn",28,"off")=14
	Set gtmtypes("rc_rsp_logn",28,"len")=2
	Set gtmtypes("rc_rsp_logn",28,"type")="rc_word"
	Set gtmtypfldindx("rc_rsp_logn","version")=28
	Set gtmtypes("rc_rsp_logn",29,"name")="rc_rsp_logn.version.octet"
	Set gtmtypes("rc_rsp_logn",29,"off")=14
	Set gtmtypes("rc_rsp_logn",29,"len")=2
	Set gtmtypes("rc_rsp_logn",29,"type")="char"
	Set gtmtypfldindx("rc_rsp_logn","version.octet")=29
	Set gtmtypes("rc_rsp_logn",30,"name")="rc_rsp_logn.version.value"
	Set gtmtypes("rc_rsp_logn",30,"off")=14
	Set gtmtypes("rc_rsp_logn",30,"len")=2
	Set gtmtypes("rc_rsp_logn",30,"type")="unsigned-short"
	Set gtmtypfldindx("rc_rsp_logn","version.value")=30
	Set gtmtypes("rc_rsp_logn",31,"name")="rc_rsp_logn.method"
	Set gtmtypes("rc_rsp_logn",31,"off")=16
	Set gtmtypes("rc_rsp_logn",31,"len")=2
	Set gtmtypes("rc_rsp_logn",31,"type")="rc_word"
	Set gtmtypfldindx("rc_rsp_logn","method")=31
	Set gtmtypes("rc_rsp_logn",32,"name")="rc_rsp_logn.method.octet"
	Set gtmtypes("rc_rsp_logn",32,"off")=16
	Set gtmtypes("rc_rsp_logn",32,"len")=2
	Set gtmtypes("rc_rsp_logn",32,"type")="char"
	Set gtmtypfldindx("rc_rsp_logn","method.octet")=32
	Set gtmtypes("rc_rsp_logn",33,"name")="rc_rsp_logn.method.value"
	Set gtmtypes("rc_rsp_logn",33,"off")=16
	Set gtmtypes("rc_rsp_logn",33,"len")=2
	Set gtmtypes("rc_rsp_logn",33,"type")="unsigned-short"
	Set gtmtypfldindx("rc_rsp_logn","method.value")=33
	Set gtmtypes("rc_rsp_logn",34,"name")="rc_rsp_logn.session"
	Set gtmtypes("rc_rsp_logn",34,"off")=18
	Set gtmtypes("rc_rsp_logn",34,"len")=2
	Set gtmtypes("rc_rsp_logn",34,"type")="rc_word"
	Set gtmtypfldindx("rc_rsp_logn","session")=34
	Set gtmtypes("rc_rsp_logn",35,"name")="rc_rsp_logn.session.octet"
	Set gtmtypes("rc_rsp_logn",35,"off")=18
	Set gtmtypes("rc_rsp_logn",35,"len")=2
	Set gtmtypes("rc_rsp_logn",35,"type")="char"
	Set gtmtypfldindx("rc_rsp_logn","session.octet")=35
	Set gtmtypes("rc_rsp_logn",36,"name")="rc_rsp_logn.session.value"
	Set gtmtypes("rc_rsp_logn",36,"off")=18
	Set gtmtypes("rc_rsp_logn",36,"len")=2
	Set gtmtypes("rc_rsp_logn",36,"type")="unsigned-short"
	Set gtmtypfldindx("rc_rsp_logn","session.value")=36
	Set gtmtypes("rc_rsp_logn",37,"name")="rc_rsp_logn.date"
	Set gtmtypes("rc_rsp_logn",37,"off")=20
	Set gtmtypes("rc_rsp_logn",37,"len")=2
	Set gtmtypes("rc_rsp_logn",37,"type")="rc_word"
	Set gtmtypfldindx("rc_rsp_logn","date")=37
	Set gtmtypes("rc_rsp_logn",38,"name")="rc_rsp_logn.date.octet"
	Set gtmtypes("rc_rsp_logn",38,"off")=20
	Set gtmtypes("rc_rsp_logn",38,"len")=2
	Set gtmtypes("rc_rsp_logn",38,"type")="char"
	Set gtmtypfldindx("rc_rsp_logn","date.octet")=38
	Set gtmtypes("rc_rsp_logn",39,"name")="rc_rsp_logn.date.value"
	Set gtmtypes("rc_rsp_logn",39,"off")=20
	Set gtmtypes("rc_rsp_logn",39,"len")=2
	Set gtmtypes("rc_rsp_logn",39,"type")="unsigned-short"
	Set gtmtypfldindx("rc_rsp_logn","date.value")=39
	Set gtmtypes("rc_rsp_logn",40,"name")="rc_rsp_logn.time"
	Set gtmtypes("rc_rsp_logn",40,"off")=22
	Set gtmtypes("rc_rsp_logn",40,"len")=4
	Set gtmtypes("rc_rsp_logn",40,"type")="char"
	Set gtmtypfldindx("rc_rsp_logn","time")=40
	Set gtmtypes("rc_rsp_logn",41,"name")="rc_rsp_logn.license_blk"
	Set gtmtypes("rc_rsp_logn",41,"off")=26
	Set gtmtypes("rc_rsp_logn",41,"len")=224
	Set gtmtypes("rc_rsp_logn",41,"type")="char"
	Set gtmtypfldindx("rc_rsp_logn","license_blk")=41
	;
	Set gtmtypes("rc_rsp_page")="struct"
	Set gtmtypes("rc_rsp_page",0)=53
	Set gtmtypes("rc_rsp_page","len")=34
	Set gtmtypes("rc_rsp_page",1,"name")="rc_rsp_page.hdr"
	Set gtmtypes("rc_rsp_page",1,"off")=0
	Set gtmtypes("rc_rsp_page",1,"len")=14
	Set gtmtypes("rc_rsp_page",1,"type")="rc_q_hdr"
	Set gtmtypfldindx("rc_rsp_page","hdr")=1
	Set gtmtypes("rc_rsp_page",2,"name")="rc_rsp_page.hdr.r"
	Set gtmtypes("rc_rsp_page",2,"off")=0
	Set gtmtypes("rc_rsp_page",2,"len")=14
	Set gtmtypes("rc_rsp_page",2,"type")="rc_rq_hdr"
	Set gtmtypfldindx("rc_rsp_page","hdr.r")=2
	Set gtmtypes("rc_rsp_page",3,"name")="rc_rsp_page.hdr.r.len"
	Set gtmtypes("rc_rsp_page",3,"off")=0
	Set gtmtypes("rc_rsp_page",3,"len")=2
	Set gtmtypes("rc_rsp_page",3,"type")="rc_word"
	Set gtmtypfldindx("rc_rsp_page","hdr.r.len")=3
	Set gtmtypes("rc_rsp_page",4,"name")="rc_rsp_page.hdr.r.typ"
	Set gtmtypes("rc_rsp_page",4,"off")=2
	Set gtmtypes("rc_rsp_page",4,"len")=2
	Set gtmtypes("rc_rsp_page",4,"type")="rc_word"
	Set gtmtypfldindx("rc_rsp_page","hdr.r.typ")=4
	Set gtmtypes("rc_rsp_page",5,"name")="rc_rsp_page.hdr.r.fmd"
	Set gtmtypes("rc_rsp_page",5,"off")=4
	Set gtmtypes("rc_rsp_page",5,"len")=2
	Set gtmtypes("rc_rsp_page",5,"type")="rc_word"
	Set gtmtypfldindx("rc_rsp_page","hdr.r.fmd")=5
	Set gtmtypes("rc_rsp_page",6,"name")="rc_rsp_page.hdr.r.pid1"
	Set gtmtypes("rc_rsp_page",6,"off")=6
	Set gtmtypes("rc_rsp_page",6,"len")=2
	Set gtmtypes("rc_rsp_page",6,"type")="rc_word"
	Set gtmtypfldindx("rc_rsp_page","hdr.r.pid1")=6
	Set gtmtypes("rc_rsp_page",7,"name")="rc_rsp_page.hdr.r.pid2"
	Set gtmtypes("rc_rsp_page",7,"off")=8
	Set gtmtypes("rc_rsp_page",7,"len")=2
	Set gtmtypes("rc_rsp_page",7,"type")="rc_word"
	Set gtmtypfldindx("rc_rsp_page","hdr.r.pid2")=7
	Set gtmtypes("rc_rsp_page",8,"name")="rc_rsp_page.hdr.r.xdsid"
	Set gtmtypes("rc_rsp_page",8,"off")=10
	Set gtmtypes("rc_rsp_page",8,"len")=4
	Set gtmtypes("rc_rsp_page",8,"type")="rc_xdsid"
	Set gtmtypfldindx("rc_rsp_page","hdr.r.xdsid")=8
	Set gtmtypes("rc_rsp_page",9,"name")="rc_rsp_page.hdr.a"
	Set gtmtypes("rc_rsp_page",9,"off")=0
	Set gtmtypes("rc_rsp_page",9,"len")=14
	Set gtmtypes("rc_rsp_page",9,"type")="rc_aq_hdr"
	Set gtmtypfldindx("rc_rsp_page","hdr.a")=9
	Set gtmtypes("rc_rsp_page",10,"name")="rc_rsp_page.hdr.a.len"
	Set gtmtypes("rc_rsp_page",10,"off")=0
	Set gtmtypes("rc_rsp_page",10,"len")=2
	Set gtmtypes("rc_rsp_page",10,"type")="rc_word"
	Set gtmtypfldindx("rc_rsp_page","hdr.a.len")=10
	Set gtmtypes("rc_rsp_page",11,"name")="rc_rsp_page.hdr.a.len.octet"
	Set gtmtypes("rc_rsp_page",11,"off")=0
	Set gtmtypes("rc_rsp_page",11,"len")=2
	Set gtmtypes("rc_rsp_page",11,"type")="char"
	Set gtmtypfldindx("rc_rsp_page","hdr.a.len.octet")=11
	Set gtmtypes("rc_rsp_page",12,"name")="rc_rsp_page.hdr.a.len.value"
	Set gtmtypes("rc_rsp_page",12,"off")=0
	Set gtmtypes("rc_rsp_page",12,"len")=2
	Set gtmtypes("rc_rsp_page",12,"type")="unsigned-short"
	Set gtmtypfldindx("rc_rsp_page","hdr.a.len.value")=12
	Set gtmtypes("rc_rsp_page",13,"name")="rc_rsp_page.hdr.a.typ"
	Set gtmtypes("rc_rsp_page",13,"off")=2
	Set gtmtypes("rc_rsp_page",13,"len")=2
	Set gtmtypes("rc_rsp_page",13,"type")="rc_word"
	Set gtmtypfldindx("rc_rsp_page","hdr.a.typ")=13
	Set gtmtypes("rc_rsp_page",14,"name")="rc_rsp_page.hdr.a.typ.octet"
	Set gtmtypes("rc_rsp_page",14,"off")=2
	Set gtmtypes("rc_rsp_page",14,"len")=2
	Set gtmtypes("rc_rsp_page",14,"type")="char"
	Set gtmtypfldindx("rc_rsp_page","hdr.a.typ.octet")=14
	Set gtmtypes("rc_rsp_page",15,"name")="rc_rsp_page.hdr.a.typ.value"
	Set gtmtypes("rc_rsp_page",15,"off")=2
	Set gtmtypes("rc_rsp_page",15,"len")=2
	Set gtmtypes("rc_rsp_page",15,"type")="unsigned-short"
	Set gtmtypfldindx("rc_rsp_page","hdr.a.typ.value")=15
	Set gtmtypes("rc_rsp_page",16,"name")="rc_rsp_page.hdr.a.erc"
	Set gtmtypes("rc_rsp_page",16,"off")=4
	Set gtmtypes("rc_rsp_page",16,"len")=2
	Set gtmtypes("rc_rsp_page",16,"type")="rc_word"
	Set gtmtypfldindx("rc_rsp_page","hdr.a.erc")=16
	Set gtmtypes("rc_rsp_page",17,"name")="rc_rsp_page.hdr.a.erc.octet"
	Set gtmtypes("rc_rsp_page",17,"off")=4
	Set gtmtypes("rc_rsp_page",17,"len")=2
	Set gtmtypes("rc_rsp_page",17,"type")="char"
	Set gtmtypfldindx("rc_rsp_page","hdr.a.erc.octet")=17
	Set gtmtypes("rc_rsp_page",18,"name")="rc_rsp_page.hdr.a.erc.value"
	Set gtmtypes("rc_rsp_page",18,"off")=4
	Set gtmtypes("rc_rsp_page",18,"len")=2
	Set gtmtypes("rc_rsp_page",18,"type")="unsigned-short"
	Set gtmtypfldindx("rc_rsp_page","hdr.a.erc.value")=18
	Set gtmtypes("rc_rsp_page",19,"name")="rc_rsp_page.hdr.a.pid1"
	Set gtmtypes("rc_rsp_page",19,"off")=6
	Set gtmtypes("rc_rsp_page",19,"len")=2
	Set gtmtypes("rc_rsp_page",19,"type")="rc_word"
	Set gtmtypfldindx("rc_rsp_page","hdr.a.pid1")=19
	Set gtmtypes("rc_rsp_page",20,"name")="rc_rsp_page.hdr.a.pid1.octet"
	Set gtmtypes("rc_rsp_page",20,"off")=6
	Set gtmtypes("rc_rsp_page",20,"len")=2
	Set gtmtypes("rc_rsp_page",20,"type")="char"
	Set gtmtypfldindx("rc_rsp_page","hdr.a.pid1.octet")=20
	Set gtmtypes("rc_rsp_page",21,"name")="rc_rsp_page.hdr.a.pid1.value"
	Set gtmtypes("rc_rsp_page",21,"off")=6
	Set gtmtypes("rc_rsp_page",21,"len")=2
	Set gtmtypes("rc_rsp_page",21,"type")="unsigned-short"
	Set gtmtypfldindx("rc_rsp_page","hdr.a.pid1.value")=21
	Set gtmtypes("rc_rsp_page",22,"name")="rc_rsp_page.hdr.a.pid2"
	Set gtmtypes("rc_rsp_page",22,"off")=8
	Set gtmtypes("rc_rsp_page",22,"len")=2
	Set gtmtypes("rc_rsp_page",22,"type")="rc_word"
	Set gtmtypfldindx("rc_rsp_page","hdr.a.pid2")=22
	Set gtmtypes("rc_rsp_page",23,"name")="rc_rsp_page.hdr.a.pid2.octet"
	Set gtmtypes("rc_rsp_page",23,"off")=8
	Set gtmtypes("rc_rsp_page",23,"len")=2
	Set gtmtypes("rc_rsp_page",23,"type")="char"
	Set gtmtypfldindx("rc_rsp_page","hdr.a.pid2.octet")=23
	Set gtmtypes("rc_rsp_page",24,"name")="rc_rsp_page.hdr.a.pid2.value"
	Set gtmtypes("rc_rsp_page",24,"off")=8
	Set gtmtypes("rc_rsp_page",24,"len")=2
	Set gtmtypes("rc_rsp_page",24,"type")="unsigned-short"
	Set gtmtypfldindx("rc_rsp_page","hdr.a.pid2.value")=24
	Set gtmtypes("rc_rsp_page",25,"name")="rc_rsp_page.hdr.a.xdsid"
	Set gtmtypes("rc_rsp_page",25,"off")=10
	Set gtmtypes("rc_rsp_page",25,"len")=4
	Set gtmtypes("rc_rsp_page",25,"type")="rc_xdsid"
	Set gtmtypfldindx("rc_rsp_page","hdr.a.xdsid")=25
	Set gtmtypes("rc_rsp_page",26,"name")="rc_rsp_page.hdr.a.xdsid.dsid"
	Set gtmtypes("rc_rsp_page",26,"off")=10
	Set gtmtypes("rc_rsp_page",26,"len")=2
	Set gtmtypes("rc_rsp_page",26,"type")="rc_word"
	Set gtmtypfldindx("rc_rsp_page","hdr.a.xdsid.dsid")=26
	Set gtmtypes("rc_rsp_page",27,"name")="rc_rsp_page.hdr.a.xdsid.node"
	Set gtmtypes("rc_rsp_page",27,"off")=12
	Set gtmtypes("rc_rsp_page",27,"len")=2
	Set gtmtypes("rc_rsp_page",27,"type")="rc_word"
	Set gtmtypfldindx("rc_rsp_page","hdr.a.xdsid.node")=27
	Set gtmtypes("rc_rsp_page",28,"name")="rc_rsp_page.pageaddr"
	Set gtmtypes("rc_rsp_page",28,"off")=14
	Set gtmtypes("rc_rsp_page",28,"len")=4
	Set gtmtypes("rc_rsp_page",28,"type")="char"
	Set gtmtypfldindx("rc_rsp_page","pageaddr")=28
	Set gtmtypes("rc_rsp_page",29,"name")="rc_rsp_page.frag_offset"
	Set gtmtypes("rc_rsp_page",29,"off")=18
	Set gtmtypes("rc_rsp_page",29,"len")=2
	Set gtmtypes("rc_rsp_page",29,"type")="rc_word"
	Set gtmtypfldindx("rc_rsp_page","frag_offset")=29
	Set gtmtypes("rc_rsp_page",30,"name")="rc_rsp_page.frag_offset.octet"
	Set gtmtypes("rc_rsp_page",30,"off")=18
	Set gtmtypes("rc_rsp_page",30,"len")=2
	Set gtmtypes("rc_rsp_page",30,"type")="char"
	Set gtmtypfldindx("rc_rsp_page","frag_offset.octet")=30
	Set gtmtypes("rc_rsp_page",31,"name")="rc_rsp_page.frag_offset.value"
	Set gtmtypes("rc_rsp_page",31,"off")=18
	Set gtmtypes("rc_rsp_page",31,"len")=2
	Set gtmtypes("rc_rsp_page",31,"type")="unsigned-short"
	Set gtmtypfldindx("rc_rsp_page","frag_offset.value")=31
	Set gtmtypes("rc_rsp_page",32,"name")="rc_rsp_page.size_return"
	Set gtmtypes("rc_rsp_page",32,"off")=20
	Set gtmtypes("rc_rsp_page",32,"len")=2
	Set gtmtypes("rc_rsp_page",32,"type")="rc_word"
	Set gtmtypfldindx("rc_rsp_page","size_return")=32
	Set gtmtypes("rc_rsp_page",33,"name")="rc_rsp_page.size_return.octet"
	Set gtmtypes("rc_rsp_page",33,"off")=20
	Set gtmtypes("rc_rsp_page",33,"len")=2
	Set gtmtypes("rc_rsp_page",33,"type")="char"
	Set gtmtypfldindx("rc_rsp_page","size_return.octet")=33
	Set gtmtypes("rc_rsp_page",34,"name")="rc_rsp_page.size_return.value"
	Set gtmtypes("rc_rsp_page",34,"off")=20
	Set gtmtypes("rc_rsp_page",34,"len")=2
	Set gtmtypes("rc_rsp_page",34,"type")="unsigned-short"
	Set gtmtypfldindx("rc_rsp_page","size_return.value")=34
	Set gtmtypes("rc_rsp_page",35,"name")="rc_rsp_page.size_remain"
	Set gtmtypes("rc_rsp_page",35,"off")=22
	Set gtmtypes("rc_rsp_page",35,"len")=2
	Set gtmtypes("rc_rsp_page",35,"type")="rc_word"
	Set gtmtypfldindx("rc_rsp_page","size_remain")=35
	Set gtmtypes("rc_rsp_page",36,"name")="rc_rsp_page.size_remain.octet"
	Set gtmtypes("rc_rsp_page",36,"off")=22
	Set gtmtypes("rc_rsp_page",36,"len")=2
	Set gtmtypes("rc_rsp_page",36,"type")="char"
	Set gtmtypfldindx("rc_rsp_page","size_remain.octet")=36
	Set gtmtypes("rc_rsp_page",37,"name")="rc_rsp_page.size_remain.value"
	Set gtmtypes("rc_rsp_page",37,"off")=22
	Set gtmtypes("rc_rsp_page",37,"len")=2
	Set gtmtypes("rc_rsp_page",37,"type")="unsigned-short"
	Set gtmtypfldindx("rc_rsp_page","size_remain.value")=37
	Set gtmtypes("rc_rsp_page",38,"name")="rc_rsp_page.before"
	Set gtmtypes("rc_rsp_page",38,"off")=24
	Set gtmtypes("rc_rsp_page",38,"len")=2
	Set gtmtypes("rc_rsp_page",38,"type")="rc_word"
	Set gtmtypfldindx("rc_rsp_page","before")=38
	Set gtmtypes("rc_rsp_page",39,"name")="rc_rsp_page.before.octet"
	Set gtmtypes("rc_rsp_page",39,"off")=24
	Set gtmtypes("rc_rsp_page",39,"len")=2
	Set gtmtypes("rc_rsp_page",39,"type")="char"
	Set gtmtypfldindx("rc_rsp_page","before.octet")=39
	Set gtmtypes("rc_rsp_page",40,"name")="rc_rsp_page.before.value"
	Set gtmtypes("rc_rsp_page",40,"off")=24
	Set gtmtypes("rc_rsp_page",40,"len")=2
	Set gtmtypes("rc_rsp_page",40,"type")="unsigned-short"
	Set gtmtypfldindx("rc_rsp_page","before.value")=40
	Set gtmtypes("rc_rsp_page",41,"name")="rc_rsp_page.after"
	Set gtmtypes("rc_rsp_page",41,"off")=26
	Set gtmtypes("rc_rsp_page",41,"len")=2
	Set gtmtypes("rc_rsp_page",41,"type")="rc_word"
	Set gtmtypfldindx("rc_rsp_page","after")=41
	Set gtmtypes("rc_rsp_page",42,"name")="rc_rsp_page.after.octet"
	Set gtmtypes("rc_rsp_page",42,"off")=26
	Set gtmtypes("rc_rsp_page",42,"len")=2
	Set gtmtypes("rc_rsp_page",42,"type")="char"
	Set gtmtypfldindx("rc_rsp_page","after.octet")=42
	Set gtmtypes("rc_rsp_page",43,"name")="rc_rsp_page.after.value"
	Set gtmtypes("rc_rsp_page",43,"off")=26
	Set gtmtypes("rc_rsp_page",43,"len")=2
	Set gtmtypes("rc_rsp_page",43,"type")="unsigned-short"
	Set gtmtypfldindx("rc_rsp_page","after.value")=43
	Set gtmtypes("rc_rsp_page",44,"name")="rc_rsp_page.xcc"
	Set gtmtypes("rc_rsp_page",44,"off")=28
	Set gtmtypes("rc_rsp_page",44,"len")=2
	Set gtmtypes("rc_rsp_page",44,"type")="rc_word"
	Set gtmtypfldindx("rc_rsp_page","xcc")=44
	Set gtmtypes("rc_rsp_page",45,"name")="rc_rsp_page.xcc.octet"
	Set gtmtypes("rc_rsp_page",45,"off")=28
	Set gtmtypes("rc_rsp_page",45,"len")=2
	Set gtmtypes("rc_rsp_page",45,"type")="char"
	Set gtmtypfldindx("rc_rsp_page","xcc.octet")=45
	Set gtmtypes("rc_rsp_page",46,"name")="rc_rsp_page.xcc.value"
	Set gtmtypes("rc_rsp_page",46,"off")=28
	Set gtmtypes("rc_rsp_page",46,"len")=2
	Set gtmtypes("rc_rsp_page",46,"type")="unsigned-short"
	Set gtmtypfldindx("rc_rsp_page","xcc.value")=46
	Set gtmtypes("rc_rsp_page",47,"name")="rc_rsp_page.rstatus"
	Set gtmtypes("rc_rsp_page",47,"off")=30
	Set gtmtypes("rc_rsp_page",47,"len")=1
	Set gtmtypes("rc_rsp_page",47,"type")="rc_byte"
	Set gtmtypfldindx("rc_rsp_page","rstatus")=47
	Set gtmtypes("rc_rsp_page",48,"name")="rc_rsp_page.rstatus.octet"
	Set gtmtypes("rc_rsp_page",48,"off")=30
	Set gtmtypes("rc_rsp_page",48,"len")=1
	Set gtmtypes("rc_rsp_page",48,"type")="char"
	Set gtmtypfldindx("rc_rsp_page","rstatus.octet")=48
	Set gtmtypes("rc_rsp_page",49,"name")="rc_rsp_page.rstatus.value"
	Set gtmtypes("rc_rsp_page",49,"off")=30
	Set gtmtypes("rc_rsp_page",49,"len")=1
	Set gtmtypes("rc_rsp_page",49,"type")="unsigned-char"
	Set gtmtypfldindx("rc_rsp_page","rstatus.value")=49
	Set gtmtypes("rc_rsp_page",50,"name")="rc_rsp_page.zcode"
	Set gtmtypes("rc_rsp_page",50,"off")=31
	Set gtmtypes("rc_rsp_page",50,"len")=1
	Set gtmtypes("rc_rsp_page",50,"type")="rc_byte"
	Set gtmtypfldindx("rc_rsp_page","zcode")=50
	Set gtmtypes("rc_rsp_page",51,"name")="rc_rsp_page.zcode.octet"
	Set gtmtypes("rc_rsp_page",51,"off")=31
	Set gtmtypes("rc_rsp_page",51,"len")=1
	Set gtmtypes("rc_rsp_page",51,"type")="char"
	Set gtmtypfldindx("rc_rsp_page","zcode.octet")=51
	Set gtmtypes("rc_rsp_page",52,"name")="rc_rsp_page.zcode.value"
	Set gtmtypes("rc_rsp_page",52,"off")=31
	Set gtmtypes("rc_rsp_page",52,"len")=1
	Set gtmtypes("rc_rsp_page",52,"type")="unsigned-char"
	Set gtmtypfldindx("rc_rsp_page","zcode.value")=52
	Set gtmtypes("rc_rsp_page",53,"name")="rc_rsp_page.page"
	Set gtmtypes("rc_rsp_page",53,"off")=32
	Set gtmtypes("rc_rsp_page",53,"len")=1
	Set gtmtypes("rc_rsp_page",53,"type")="char"
	Set gtmtypfldindx("rc_rsp_page","page")=53
	;
	Set gtmtypes("rc_sbkey")="struct"
	Set gtmtypes("rc_sbkey",0)=4
	Set gtmtypes("rc_sbkey","len")=2
	Set gtmtypes("rc_sbkey",1,"name")="rc_sbkey.len"
	Set gtmtypes("rc_sbkey",1,"off")=0
	Set gtmtypes("rc_sbkey",1,"len")=1
	Set gtmtypes("rc_sbkey",1,"type")="rc_byte"
	Set gtmtypfldindx("rc_sbkey","len")=1
	Set gtmtypes("rc_sbkey",2,"name")="rc_sbkey.len.octet"
	Set gtmtypes("rc_sbkey",2,"off")=0
	Set gtmtypes("rc_sbkey",2,"len")=1
	Set gtmtypes("rc_sbkey",2,"type")="char"
	Set gtmtypfldindx("rc_sbkey","len.octet")=2
	Set gtmtypes("rc_sbkey",3,"name")="rc_sbkey.len.value"
	Set gtmtypes("rc_sbkey",3,"off")=0
	Set gtmtypes("rc_sbkey",3,"len")=1
	Set gtmtypes("rc_sbkey",3,"type")="unsigned-char"
	Set gtmtypfldindx("rc_sbkey","len.value")=3
	Set gtmtypes("rc_sbkey",4,"name")="rc_sbkey.key"
	Set gtmtypes("rc_sbkey",4,"off")=1
	Set gtmtypes("rc_sbkey",4,"len")=1
	Set gtmtypes("rc_sbkey",4,"type")="char"
	Set gtmtypfldindx("rc_sbkey","key")=4
	;
	Set gtmtypes("rc_set")="struct"
	Set gtmtypes("rc_set",0)=34
	Set gtmtypes("rc_set","len")=20
	Set gtmtypes("rc_set",1,"name")="rc_set.hdr"
	Set gtmtypes("rc_set",1,"off")=0
	Set gtmtypes("rc_set",1,"len")=14
	Set gtmtypes("rc_set",1,"type")="rc_q_hdr"
	Set gtmtypfldindx("rc_set","hdr")=1
	Set gtmtypes("rc_set",2,"name")="rc_set.hdr.r"
	Set gtmtypes("rc_set",2,"off")=0
	Set gtmtypes("rc_set",2,"len")=14
	Set gtmtypes("rc_set",2,"type")="rc_rq_hdr"
	Set gtmtypfldindx("rc_set","hdr.r")=2
	Set gtmtypes("rc_set",3,"name")="rc_set.hdr.r.len"
	Set gtmtypes("rc_set",3,"off")=0
	Set gtmtypes("rc_set",3,"len")=2
	Set gtmtypes("rc_set",3,"type")="rc_word"
	Set gtmtypfldindx("rc_set","hdr.r.len")=3
	Set gtmtypes("rc_set",4,"name")="rc_set.hdr.r.typ"
	Set gtmtypes("rc_set",4,"off")=2
	Set gtmtypes("rc_set",4,"len")=2
	Set gtmtypes("rc_set",4,"type")="rc_word"
	Set gtmtypfldindx("rc_set","hdr.r.typ")=4
	Set gtmtypes("rc_set",5,"name")="rc_set.hdr.r.fmd"
	Set gtmtypes("rc_set",5,"off")=4
	Set gtmtypes("rc_set",5,"len")=2
	Set gtmtypes("rc_set",5,"type")="rc_word"
	Set gtmtypfldindx("rc_set","hdr.r.fmd")=5
	Set gtmtypes("rc_set",6,"name")="rc_set.hdr.r.pid1"
	Set gtmtypes("rc_set",6,"off")=6
	Set gtmtypes("rc_set",6,"len")=2
	Set gtmtypes("rc_set",6,"type")="rc_word"
	Set gtmtypfldindx("rc_set","hdr.r.pid1")=6
	Set gtmtypes("rc_set",7,"name")="rc_set.hdr.r.pid2"
	Set gtmtypes("rc_set",7,"off")=8
	Set gtmtypes("rc_set",7,"len")=2
	Set gtmtypes("rc_set",7,"type")="rc_word"
	Set gtmtypfldindx("rc_set","hdr.r.pid2")=7
	Set gtmtypes("rc_set",8,"name")="rc_set.hdr.r.xdsid"
	Set gtmtypes("rc_set",8,"off")=10
	Set gtmtypes("rc_set",8,"len")=4
	Set gtmtypes("rc_set",8,"type")="rc_xdsid"
	Set gtmtypfldindx("rc_set","hdr.r.xdsid")=8
	Set gtmtypes("rc_set",9,"name")="rc_set.hdr.a"
	Set gtmtypes("rc_set",9,"off")=0
	Set gtmtypes("rc_set",9,"len")=14
	Set gtmtypes("rc_set",9,"type")="rc_aq_hdr"
	Set gtmtypfldindx("rc_set","hdr.a")=9
	Set gtmtypes("rc_set",10,"name")="rc_set.hdr.a.len"
	Set gtmtypes("rc_set",10,"off")=0
	Set gtmtypes("rc_set",10,"len")=2
	Set gtmtypes("rc_set",10,"type")="rc_word"
	Set gtmtypfldindx("rc_set","hdr.a.len")=10
	Set gtmtypes("rc_set",11,"name")="rc_set.hdr.a.len.octet"
	Set gtmtypes("rc_set",11,"off")=0
	Set gtmtypes("rc_set",11,"len")=2
	Set gtmtypes("rc_set",11,"type")="char"
	Set gtmtypfldindx("rc_set","hdr.a.len.octet")=11
	Set gtmtypes("rc_set",12,"name")="rc_set.hdr.a.len.value"
	Set gtmtypes("rc_set",12,"off")=0
	Set gtmtypes("rc_set",12,"len")=2
	Set gtmtypes("rc_set",12,"type")="unsigned-short"
	Set gtmtypfldindx("rc_set","hdr.a.len.value")=12
	Set gtmtypes("rc_set",13,"name")="rc_set.hdr.a.typ"
	Set gtmtypes("rc_set",13,"off")=2
	Set gtmtypes("rc_set",13,"len")=2
	Set gtmtypes("rc_set",13,"type")="rc_word"
	Set gtmtypfldindx("rc_set","hdr.a.typ")=13
	Set gtmtypes("rc_set",14,"name")="rc_set.hdr.a.typ.octet"
	Set gtmtypes("rc_set",14,"off")=2
	Set gtmtypes("rc_set",14,"len")=2
	Set gtmtypes("rc_set",14,"type")="char"
	Set gtmtypfldindx("rc_set","hdr.a.typ.octet")=14
	Set gtmtypes("rc_set",15,"name")="rc_set.hdr.a.typ.value"
	Set gtmtypes("rc_set",15,"off")=2
	Set gtmtypes("rc_set",15,"len")=2
	Set gtmtypes("rc_set",15,"type")="unsigned-short"
	Set gtmtypfldindx("rc_set","hdr.a.typ.value")=15
	Set gtmtypes("rc_set",16,"name")="rc_set.hdr.a.erc"
	Set gtmtypes("rc_set",16,"off")=4
	Set gtmtypes("rc_set",16,"len")=2
	Set gtmtypes("rc_set",16,"type")="rc_word"
	Set gtmtypfldindx("rc_set","hdr.a.erc")=16
	Set gtmtypes("rc_set",17,"name")="rc_set.hdr.a.erc.octet"
	Set gtmtypes("rc_set",17,"off")=4
	Set gtmtypes("rc_set",17,"len")=2
	Set gtmtypes("rc_set",17,"type")="char"
	Set gtmtypfldindx("rc_set","hdr.a.erc.octet")=17
	Set gtmtypes("rc_set",18,"name")="rc_set.hdr.a.erc.value"
	Set gtmtypes("rc_set",18,"off")=4
	Set gtmtypes("rc_set",18,"len")=2
	Set gtmtypes("rc_set",18,"type")="unsigned-short"
	Set gtmtypfldindx("rc_set","hdr.a.erc.value")=18
	Set gtmtypes("rc_set",19,"name")="rc_set.hdr.a.pid1"
	Set gtmtypes("rc_set",19,"off")=6
	Set gtmtypes("rc_set",19,"len")=2
	Set gtmtypes("rc_set",19,"type")="rc_word"
	Set gtmtypfldindx("rc_set","hdr.a.pid1")=19
	Set gtmtypes("rc_set",20,"name")="rc_set.hdr.a.pid1.octet"
	Set gtmtypes("rc_set",20,"off")=6
	Set gtmtypes("rc_set",20,"len")=2
	Set gtmtypes("rc_set",20,"type")="char"
	Set gtmtypfldindx("rc_set","hdr.a.pid1.octet")=20
	Set gtmtypes("rc_set",21,"name")="rc_set.hdr.a.pid1.value"
	Set gtmtypes("rc_set",21,"off")=6
	Set gtmtypes("rc_set",21,"len")=2
	Set gtmtypes("rc_set",21,"type")="unsigned-short"
	Set gtmtypfldindx("rc_set","hdr.a.pid1.value")=21
	Set gtmtypes("rc_set",22,"name")="rc_set.hdr.a.pid2"
	Set gtmtypes("rc_set",22,"off")=8
	Set gtmtypes("rc_set",22,"len")=2
	Set gtmtypes("rc_set",22,"type")="rc_word"
	Set gtmtypfldindx("rc_set","hdr.a.pid2")=22
	Set gtmtypes("rc_set",23,"name")="rc_set.hdr.a.pid2.octet"
	Set gtmtypes("rc_set",23,"off")=8
	Set gtmtypes("rc_set",23,"len")=2
	Set gtmtypes("rc_set",23,"type")="char"
	Set gtmtypfldindx("rc_set","hdr.a.pid2.octet")=23
	Set gtmtypes("rc_set",24,"name")="rc_set.hdr.a.pid2.value"
	Set gtmtypes("rc_set",24,"off")=8
	Set gtmtypes("rc_set",24,"len")=2
	Set gtmtypes("rc_set",24,"type")="unsigned-short"
	Set gtmtypfldindx("rc_set","hdr.a.pid2.value")=24
	Set gtmtypes("rc_set",25,"name")="rc_set.hdr.a.xdsid"
	Set gtmtypes("rc_set",25,"off")=10
	Set gtmtypes("rc_set",25,"len")=4
	Set gtmtypes("rc_set",25,"type")="rc_xdsid"
	Set gtmtypfldindx("rc_set","hdr.a.xdsid")=25
	Set gtmtypes("rc_set",26,"name")="rc_set.hdr.a.xdsid.dsid"
	Set gtmtypes("rc_set",26,"off")=10
	Set gtmtypes("rc_set",26,"len")=2
	Set gtmtypes("rc_set",26,"type")="rc_word"
	Set gtmtypfldindx("rc_set","hdr.a.xdsid.dsid")=26
	Set gtmtypes("rc_set",27,"name")="rc_set.hdr.a.xdsid.node"
	Set gtmtypes("rc_set",27,"off")=12
	Set gtmtypes("rc_set",27,"len")=2
	Set gtmtypes("rc_set",27,"type")="rc_word"
	Set gtmtypfldindx("rc_set","hdr.a.xdsid.node")=27
	Set gtmtypes("rc_set",28,"name")="rc_set.xnsid"
	Set gtmtypes("rc_set",28,"off")=14
	Set gtmtypes("rc_set",28,"len")=4
	Set gtmtypes("rc_set",28,"type")="rc_xnsid"
	Set gtmtypfldindx("rc_set","xnsid")=28
	Set gtmtypes("rc_set",29,"name")="rc_set.xnsid.value"
	Set gtmtypes("rc_set",29,"off")=14
	Set gtmtypes("rc_set",29,"len")=4
	Set gtmtypes("rc_set",29,"type")="char"
	Set gtmtypfldindx("rc_set","xnsid.value")=29
	Set gtmtypes("rc_set",30,"name")="rc_set.key"
	Set gtmtypes("rc_set",30,"off")=18
	Set gtmtypes("rc_set",30,"len")=2
	Set gtmtypes("rc_set",30,"type")="rc_sbkey"
	Set gtmtypfldindx("rc_set","key")=30
	Set gtmtypes("rc_set",31,"name")="rc_set.key.len"
	Set gtmtypes("rc_set",31,"off")=18
	Set gtmtypes("rc_set",31,"len")=1
	Set gtmtypes("rc_set",31,"type")="rc_byte"
	Set gtmtypfldindx("rc_set","key.len")=31
	Set gtmtypes("rc_set",32,"name")="rc_set.key.len.octet"
	Set gtmtypes("rc_set",32,"off")=18
	Set gtmtypes("rc_set",32,"len")=1
	Set gtmtypes("rc_set",32,"type")="char"
	Set gtmtypfldindx("rc_set","key.len.octet")=32
	Set gtmtypes("rc_set",33,"name")="rc_set.key.len.value"
	Set gtmtypes("rc_set",33,"off")=18
	Set gtmtypes("rc_set",33,"len")=1
	Set gtmtypes("rc_set",33,"type")="unsigned-char"
	Set gtmtypfldindx("rc_set","key.len.value")=33
	Set gtmtypes("rc_set",34,"name")="rc_set.key.key"
	Set gtmtypes("rc_set",34,"off")=19
	Set gtmtypes("rc_set",34,"len")=1
	Set gtmtypes("rc_set",34,"type")="char"
	Set gtmtypfldindx("rc_set","key.key")=34
	;
	Set gtmtypes("rc_swstr")="struct"
	Set gtmtypes("rc_swstr",0)=4
	Set gtmtypes("rc_swstr","len")=4
	Set gtmtypes("rc_swstr",1,"name")="rc_swstr.len"
	Set gtmtypes("rc_swstr",1,"off")=0
	Set gtmtypes("rc_swstr",1,"len")=2
	Set gtmtypes("rc_swstr",1,"type")="rc_word"
	Set gtmtypfldindx("rc_swstr","len")=1
	Set gtmtypes("rc_swstr",2,"name")="rc_swstr.len.octet"
	Set gtmtypes("rc_swstr",2,"off")=0
	Set gtmtypes("rc_swstr",2,"len")=2
	Set gtmtypes("rc_swstr",2,"type")="char"
	Set gtmtypfldindx("rc_swstr","len.octet")=2
	Set gtmtypes("rc_swstr",3,"name")="rc_swstr.len.value"
	Set gtmtypes("rc_swstr",3,"off")=0
	Set gtmtypes("rc_swstr",3,"len")=2
	Set gtmtypes("rc_swstr",3,"type")="unsigned-short"
	Set gtmtypfldindx("rc_swstr","len.value")=3
	Set gtmtypes("rc_swstr",4,"name")="rc_swstr.str"
	Set gtmtypes("rc_swstr",4,"off")=2
	Set gtmtypes("rc_swstr",4,"len")=1
	Set gtmtypes("rc_swstr",4,"type")="char"
	Set gtmtypfldindx("rc_swstr","str")=4
	;
	Set gtmtypes("rc_word")="union"
	Set gtmtypes("rc_word",0)=2
	Set gtmtypes("rc_word","len")=2
	Set gtmtypes("rc_word",1,"name")="rc_word.octet"
	Set gtmtypes("rc_word",1,"off")=0
	Set gtmtypes("rc_word",1,"len")=2
	Set gtmtypes("rc_word",1,"type")="char"
	Set gtmtypfldindx("rc_word","octet")=1
	Set gtmtypes("rc_word",2,"name")="rc_word.value"
	Set gtmtypes("rc_word",2,"off")=0
	Set gtmtypes("rc_word",2,"len")=2
	Set gtmtypes("rc_word",2,"type")="unsigned-short"
	Set gtmtypfldindx("rc_word","value")=2
	;
	Set gtmtypes("rc_xblk_hdr")="struct"
	Set gtmtypes("rc_xblk_hdr",0)=45
	Set gtmtypes("rc_xblk_hdr","len")=64
	Set gtmtypes("rc_xblk_hdr",1,"name")="rc_xblk_hdr.end"
	Set gtmtypes("rc_xblk_hdr",1,"off")=0
	Set gtmtypes("rc_xblk_hdr",1,"len")=2
	Set gtmtypes("rc_xblk_hdr",1,"type")="rc_word"
	Set gtmtypfldindx("rc_xblk_hdr","end")=1
	Set gtmtypes("rc_xblk_hdr",2,"name")="rc_xblk_hdr.end.octet"
	Set gtmtypes("rc_xblk_hdr",2,"off")=0
	Set gtmtypes("rc_xblk_hdr",2,"len")=2
	Set gtmtypes("rc_xblk_hdr",2,"type")="char"
	Set gtmtypfldindx("rc_xblk_hdr","end.octet")=2
	Set gtmtypes("rc_xblk_hdr",3,"name")="rc_xblk_hdr.end.value"
	Set gtmtypes("rc_xblk_hdr",3,"off")=0
	Set gtmtypes("rc_xblk_hdr",3,"len")=2
	Set gtmtypes("rc_xblk_hdr",3,"type")="unsigned-short"
	Set gtmtypfldindx("rc_xblk_hdr","end.value")=3
	Set gtmtypes("rc_xblk_hdr",4,"name")="rc_xblk_hdr.gtm_free"
	Set gtmtypes("rc_xblk_hdr",4,"off")=2
	Set gtmtypes("rc_xblk_hdr",4,"len")=2
	Set gtmtypes("rc_xblk_hdr",4,"type")="rc_word"
	Set gtmtypfldindx("rc_xblk_hdr","gtm_free")=4
	Set gtmtypes("rc_xblk_hdr",5,"name")="rc_xblk_hdr.gtm_free.octet"
	Set gtmtypes("rc_xblk_hdr",5,"off")=2
	Set gtmtypes("rc_xblk_hdr",5,"len")=2
	Set gtmtypes("rc_xblk_hdr",5,"type")="char"
	Set gtmtypfldindx("rc_xblk_hdr","gtm_free.octet")=5
	Set gtmtypes("rc_xblk_hdr",6,"name")="rc_xblk_hdr.gtm_free.value"
	Set gtmtypes("rc_xblk_hdr",6,"off")=2
	Set gtmtypes("rc_xblk_hdr",6,"len")=2
	Set gtmtypes("rc_xblk_hdr",6,"type")="unsigned-short"
	Set gtmtypfldindx("rc_xblk_hdr","gtm_free.value")=6
	Set gtmtypes("rc_xblk_hdr",7,"name")="rc_xblk_hdr.method"
	Set gtmtypes("rc_xblk_hdr",7,"off")=4
	Set gtmtypes("rc_xblk_hdr",7,"len")=1
	Set gtmtypes("rc_xblk_hdr",7,"type")="rc_byte"
	Set gtmtypfldindx("rc_xblk_hdr","method")=7
	Set gtmtypes("rc_xblk_hdr",8,"name")="rc_xblk_hdr.method.octet"
	Set gtmtypes("rc_xblk_hdr",8,"off")=4
	Set gtmtypes("rc_xblk_hdr",8,"len")=1
	Set gtmtypes("rc_xblk_hdr",8,"type")="char"
	Set gtmtypfldindx("rc_xblk_hdr","method.octet")=8
	Set gtmtypes("rc_xblk_hdr",9,"name")="rc_xblk_hdr.method.value"
	Set gtmtypes("rc_xblk_hdr",9,"off")=4
	Set gtmtypes("rc_xblk_hdr",9,"len")=1
	Set gtmtypes("rc_xblk_hdr",9,"type")="unsigned-char"
	Set gtmtypfldindx("rc_xblk_hdr","method.value")=9
	Set gtmtypes("rc_xblk_hdr",10,"name")="rc_xblk_hdr.alert"
	Set gtmtypes("rc_xblk_hdr",10,"off")=5
	Set gtmtypes("rc_xblk_hdr",10,"len")=1
	Set gtmtypes("rc_xblk_hdr",10,"type")="rc_byte"
	Set gtmtypfldindx("rc_xblk_hdr","alert")=10
	Set gtmtypes("rc_xblk_hdr",11,"name")="rc_xblk_hdr.alert.octet"
	Set gtmtypes("rc_xblk_hdr",11,"off")=5
	Set gtmtypes("rc_xblk_hdr",11,"len")=1
	Set gtmtypes("rc_xblk_hdr",11,"type")="char"
	Set gtmtypfldindx("rc_xblk_hdr","alert.octet")=11
	Set gtmtypes("rc_xblk_hdr",12,"name")="rc_xblk_hdr.alert.value"
	Set gtmtypes("rc_xblk_hdr",12,"off")=5
	Set gtmtypes("rc_xblk_hdr",12,"len")=1
	Set gtmtypes("rc_xblk_hdr",12,"type")="unsigned-char"
	Set gtmtypfldindx("rc_xblk_hdr","alert.value")=12
	Set gtmtypes("rc_xblk_hdr",13,"name")="rc_xblk_hdr.sync"
	Set gtmtypes("rc_xblk_hdr",13,"off")=6
	Set gtmtypes("rc_xblk_hdr",13,"len")=2
	Set gtmtypes("rc_xblk_hdr",13,"type")="rc_word"
	Set gtmtypfldindx("rc_xblk_hdr","sync")=13
	Set gtmtypes("rc_xblk_hdr",14,"name")="rc_xblk_hdr.sync.octet"
	Set gtmtypes("rc_xblk_hdr",14,"off")=6
	Set gtmtypes("rc_xblk_hdr",14,"len")=2
	Set gtmtypes("rc_xblk_hdr",14,"type")="char"
	Set gtmtypfldindx("rc_xblk_hdr","sync.octet")=14
	Set gtmtypes("rc_xblk_hdr",15,"name")="rc_xblk_hdr.sync.value"
	Set gtmtypes("rc_xblk_hdr",15,"off")=6
	Set gtmtypes("rc_xblk_hdr",15,"len")=2
	Set gtmtypes("rc_xblk_hdr",15,"type")="unsigned-short"
	Set gtmtypfldindx("rc_xblk_hdr","sync.value")=15
	Set gtmtypes("rc_xblk_hdr",16,"name")="rc_xblk_hdr.cpt_tab"
	Set gtmtypes("rc_xblk_hdr",16,"off")=8
	Set gtmtypes("rc_xblk_hdr",16,"len")=2
	Set gtmtypes("rc_xblk_hdr",16,"type")="rc_word"
	Set gtmtypfldindx("rc_xblk_hdr","cpt_tab")=16
	Set gtmtypes("rc_xblk_hdr",17,"name")="rc_xblk_hdr.cpt_tab.octet"
	Set gtmtypes("rc_xblk_hdr",17,"off")=8
	Set gtmtypes("rc_xblk_hdr",17,"len")=2
	Set gtmtypes("rc_xblk_hdr",17,"type")="char"
	Set gtmtypfldindx("rc_xblk_hdr","cpt_tab.octet")=17
	Set gtmtypes("rc_xblk_hdr",18,"name")="rc_xblk_hdr.cpt_tab.value"
	Set gtmtypes("rc_xblk_hdr",18,"off")=8
	Set gtmtypes("rc_xblk_hdr",18,"len")=2
	Set gtmtypes("rc_xblk_hdr",18,"type")="unsigned-short"
	Set gtmtypfldindx("rc_xblk_hdr","cpt_tab.value")=18
	Set gtmtypes("rc_xblk_hdr",19,"name")="rc_xblk_hdr.cpt_siz"
	Set gtmtypes("rc_xblk_hdr",19,"off")=10
	Set gtmtypes("rc_xblk_hdr",19,"len")=2
	Set gtmtypes("rc_xblk_hdr",19,"type")="rc_word"
	Set gtmtypfldindx("rc_xblk_hdr","cpt_siz")=19
	Set gtmtypes("rc_xblk_hdr",20,"name")="rc_xblk_hdr.cpt_siz.octet"
	Set gtmtypes("rc_xblk_hdr",20,"off")=10
	Set gtmtypes("rc_xblk_hdr",20,"len")=2
	Set gtmtypes("rc_xblk_hdr",20,"type")="char"
	Set gtmtypfldindx("rc_xblk_hdr","cpt_siz.octet")=20
	Set gtmtypes("rc_xblk_hdr",21,"name")="rc_xblk_hdr.cpt_siz.value"
	Set gtmtypes("rc_xblk_hdr",21,"off")=10
	Set gtmtypes("rc_xblk_hdr",21,"len")=2
	Set gtmtypes("rc_xblk_hdr",21,"type")="unsigned-short"
	Set gtmtypfldindx("rc_xblk_hdr","cpt_siz.value")=21
	Set gtmtypes("rc_xblk_hdr",22,"name")="rc_xblk_hdr.aq_res"
	Set gtmtypes("rc_xblk_hdr",22,"off")=12
	Set gtmtypes("rc_xblk_hdr",22,"len")=2
	Set gtmtypes("rc_xblk_hdr",22,"type")="rc_word"
	Set gtmtypfldindx("rc_xblk_hdr","aq_res")=22
	Set gtmtypes("rc_xblk_hdr",23,"name")="rc_xblk_hdr.aq_res.octet"
	Set gtmtypes("rc_xblk_hdr",23,"off")=12
	Set gtmtypes("rc_xblk_hdr",23,"len")=2
	Set gtmtypes("rc_xblk_hdr",23,"type")="char"
	Set gtmtypfldindx("rc_xblk_hdr","aq_res.octet")=23
	Set gtmtypes("rc_xblk_hdr",24,"name")="rc_xblk_hdr.aq_res.value"
	Set gtmtypes("rc_xblk_hdr",24,"off")=12
	Set gtmtypes("rc_xblk_hdr",24,"len")=2
	Set gtmtypes("rc_xblk_hdr",24,"type")="unsigned-short"
	Set gtmtypfldindx("rc_xblk_hdr","aq_res.value")=24
	Set gtmtypes("rc_xblk_hdr",25,"name")="rc_xblk_hdr.resp_max"
	Set gtmtypes("rc_xblk_hdr",25,"off")=14
	Set gtmtypes("rc_xblk_hdr",25,"len")=2
	Set gtmtypes("rc_xblk_hdr",25,"type")="rc_word"
	Set gtmtypfldindx("rc_xblk_hdr","resp_max")=25
	Set gtmtypes("rc_xblk_hdr",26,"name")="rc_xblk_hdr.resp_max.octet"
	Set gtmtypes("rc_xblk_hdr",26,"off")=14
	Set gtmtypes("rc_xblk_hdr",26,"len")=2
	Set gtmtypes("rc_xblk_hdr",26,"type")="char"
	Set gtmtypfldindx("rc_xblk_hdr","resp_max.octet")=26
	Set gtmtypes("rc_xblk_hdr",27,"name")="rc_xblk_hdr.resp_max.value"
	Set gtmtypes("rc_xblk_hdr",27,"off")=14
	Set gtmtypes("rc_xblk_hdr",27,"len")=2
	Set gtmtypes("rc_xblk_hdr",27,"type")="unsigned-short"
	Set gtmtypfldindx("rc_xblk_hdr","resp_max.value")=27
	Set gtmtypes("rc_xblk_hdr",28,"name")="rc_xblk_hdr.client"
	Set gtmtypes("rc_xblk_hdr",28,"off")=16
	Set gtmtypes("rc_xblk_hdr",28,"len")=4
	Set gtmtypes("rc_xblk_hdr",28,"type")="rc_lword"
	Set gtmtypfldindx("rc_xblk_hdr","client")=28
	Set gtmtypes("rc_xblk_hdr",29,"name")="rc_xblk_hdr.client.octet"
	Set gtmtypes("rc_xblk_hdr",29,"off")=16
	Set gtmtypes("rc_xblk_hdr",29,"len")=4
	Set gtmtypes("rc_xblk_hdr",29,"type")="char"
	Set gtmtypfldindx("rc_xblk_hdr","client.octet")=29
	Set gtmtypes("rc_xblk_hdr",30,"name")="rc_xblk_hdr.client.value"
	Set gtmtypes("rc_xblk_hdr",30,"off")=16
	Set gtmtypes("rc_xblk_hdr",30,"len")=4
	Set gtmtypes("rc_xblk_hdr",30,"type")="unsigned-int"
	Set gtmtypfldindx("rc_xblk_hdr","client.value")=30
	Set gtmtypes("rc_xblk_hdr",31,"name")="rc_xblk_hdr.alive"
	Set gtmtypes("rc_xblk_hdr",31,"off")=20
	Set gtmtypes("rc_xblk_hdr",31,"len")=1
	Set gtmtypes("rc_xblk_hdr",31,"type")="rc_byte"
	Set gtmtypfldindx("rc_xblk_hdr","alive")=31
	Set gtmtypes("rc_xblk_hdr",32,"name")="rc_xblk_hdr.alive.octet"
	Set gtmtypes("rc_xblk_hdr",32,"off")=20
	Set gtmtypes("rc_xblk_hdr",32,"len")=1
	Set gtmtypes("rc_xblk_hdr",32,"type")="char"
	Set gtmtypfldindx("rc_xblk_hdr","alive.octet")=32
	Set gtmtypes("rc_xblk_hdr",33,"name")="rc_xblk_hdr.alive.value"
	Set gtmtypes("rc_xblk_hdr",33,"off")=20
	Set gtmtypes("rc_xblk_hdr",33,"len")=1
	Set gtmtypes("rc_xblk_hdr",33,"type")="unsigned-char"
	Set gtmtypfldindx("rc_xblk_hdr","alive.value")=33
	Set gtmtypes("rc_xblk_hdr",34,"name")="rc_xblk_hdr.filler0"
	Set gtmtypes("rc_xblk_hdr",34,"off")=21
	Set gtmtypes("rc_xblk_hdr",34,"len")=1
	Set gtmtypes("rc_xblk_hdr",34,"type")="char"
	Set gtmtypfldindx("rc_xblk_hdr","filler0")=34
	Set gtmtypes("rc_xblk_hdr",35,"name")="rc_xblk_hdr.err_aq"
	Set gtmtypes("rc_xblk_hdr",35,"off")=22
	Set gtmtypes("rc_xblk_hdr",35,"len")=2
	Set gtmtypes("rc_xblk_hdr",35,"type")="rc_word"
	Set gtmtypfldindx("rc_xblk_hdr","err_aq")=35
	Set gtmtypes("rc_xblk_hdr",36,"name")="rc_xblk_hdr.err_aq.octet"
	Set gtmtypes("rc_xblk_hdr",36,"off")=22
	Set gtmtypes("rc_xblk_hdr",36,"len")=2
	Set gtmtypes("rc_xblk_hdr",36,"type")="char"
	Set gtmtypfldindx("rc_xblk_hdr","err_aq.octet")=36
	Set gtmtypes("rc_xblk_hdr",37,"name")="rc_xblk_hdr.err_aq.value"
	Set gtmtypes("rc_xblk_hdr",37,"off")=22
	Set gtmtypes("rc_xblk_hdr",37,"len")=2
	Set gtmtypes("rc_xblk_hdr",37,"type")="unsigned-short"
	Set gtmtypfldindx("rc_xblk_hdr","err_aq.value")=37
	Set gtmtypes("rc_xblk_hdr",38,"name")="rc_xblk_hdr.last_aq"
	Set gtmtypes("rc_xblk_hdr",38,"off")=24
	Set gtmtypes("rc_xblk_hdr",38,"len")=2
	Set gtmtypes("rc_xblk_hdr",38,"type")="rc_word"
	Set gtmtypfldindx("rc_xblk_hdr","last_aq")=38
	Set gtmtypes("rc_xblk_hdr",39,"name")="rc_xblk_hdr.last_aq.octet"
	Set gtmtypes("rc_xblk_hdr",39,"off")=24
	Set gtmtypes("rc_xblk_hdr",39,"len")=2
	Set gtmtypes("rc_xblk_hdr",39,"type")="char"
	Set gtmtypfldindx("rc_xblk_hdr","last_aq.octet")=39
	Set gtmtypes("rc_xblk_hdr",40,"name")="rc_xblk_hdr.last_aq.value"
	Set gtmtypes("rc_xblk_hdr",40,"off")=24
	Set gtmtypes("rc_xblk_hdr",40,"len")=2
	Set gtmtypes("rc_xblk_hdr",40,"type")="unsigned-short"
	Set gtmtypfldindx("rc_xblk_hdr","last_aq.value")=40
	Set gtmtypes("rc_xblk_hdr",41,"name")="rc_xblk_hdr.big_endian"
	Set gtmtypes("rc_xblk_hdr",41,"off")=26
	Set gtmtypes("rc_xblk_hdr",41,"len")=1
	Set gtmtypes("rc_xblk_hdr",41,"type")="rc_byte"
	Set gtmtypfldindx("rc_xblk_hdr","big_endian")=41
	Set gtmtypes("rc_xblk_hdr",42,"name")="rc_xblk_hdr.big_endian.octet"
	Set gtmtypes("rc_xblk_hdr",42,"off")=26
	Set gtmtypes("rc_xblk_hdr",42,"len")=1
	Set gtmtypes("rc_xblk_hdr",42,"type")="char"
	Set gtmtypfldindx("rc_xblk_hdr","big_endian.octet")=42
	Set gtmtypes("rc_xblk_hdr",43,"name")="rc_xblk_hdr.big_endian.value"
	Set gtmtypes("rc_xblk_hdr",43,"off")=26
	Set gtmtypes("rc_xblk_hdr",43,"len")=1
	Set gtmtypes("rc_xblk_hdr",43,"type")="unsigned-char"
	Set gtmtypfldindx("rc_xblk_hdr","big_endian.value")=43
	Set gtmtypes("rc_xblk_hdr",44,"name")="rc_xblk_hdr.filler1"
	Set gtmtypes("rc_xblk_hdr",44,"off")=27
	Set gtmtypes("rc_xblk_hdr",44,"len")=5
	Set gtmtypes("rc_xblk_hdr",44,"type")="char"
	Set gtmtypfldindx("rc_xblk_hdr","filler1")=44
	Set gtmtypes("rc_xblk_hdr",45,"name")="rc_xblk_hdr.asm1"
	Set gtmtypes("rc_xblk_hdr",45,"off")=32
	Set gtmtypes("rc_xblk_hdr",45,"len")=32
	Set gtmtypes("rc_xblk_hdr",45,"type")="char"
	Set gtmtypfldindx("rc_xblk_hdr","asm1")=45
	;
	Set gtmtypes("rc_xdsid")="struct"
	Set gtmtypes("rc_xdsid",0)=6
	Set gtmtypes("rc_xdsid","len")=4
	Set gtmtypes("rc_xdsid",1,"name")="rc_xdsid.dsid"
	Set gtmtypes("rc_xdsid",1,"off")=0
	Set gtmtypes("rc_xdsid",1,"len")=2
	Set gtmtypes("rc_xdsid",1,"type")="rc_word"
	Set gtmtypfldindx("rc_xdsid","dsid")=1
	Set gtmtypes("rc_xdsid",2,"name")="rc_xdsid.dsid.octet"
	Set gtmtypes("rc_xdsid",2,"off")=0
	Set gtmtypes("rc_xdsid",2,"len")=2
	Set gtmtypes("rc_xdsid",2,"type")="char"
	Set gtmtypfldindx("rc_xdsid","dsid.octet")=2
	Set gtmtypes("rc_xdsid",3,"name")="rc_xdsid.dsid.value"
	Set gtmtypes("rc_xdsid",3,"off")=0
	Set gtmtypes("rc_xdsid",3,"len")=2
	Set gtmtypes("rc_xdsid",3,"type")="unsigned-short"
	Set gtmtypfldindx("rc_xdsid","dsid.value")=3
	Set gtmtypes("rc_xdsid",4,"name")="rc_xdsid.node"
	Set gtmtypes("rc_xdsid",4,"off")=2
	Set gtmtypes("rc_xdsid",4,"len")=2
	Set gtmtypes("rc_xdsid",4,"type")="rc_word"
	Set gtmtypfldindx("rc_xdsid","node")=4
	Set gtmtypes("rc_xdsid",5,"name")="rc_xdsid.node.octet"
	Set gtmtypes("rc_xdsid",5,"off")=2
	Set gtmtypes("rc_xdsid",5,"len")=2
	Set gtmtypes("rc_xdsid",5,"type")="char"
	Set gtmtypfldindx("rc_xdsid","node.octet")=5
	Set gtmtypes("rc_xdsid",6,"name")="rc_xdsid.node.value"
	Set gtmtypes("rc_xdsid",6,"off")=2
	Set gtmtypes("rc_xdsid",6,"len")=2
	Set gtmtypes("rc_xdsid",6,"type")="unsigned-short"
	Set gtmtypfldindx("rc_xdsid","node.value")=6
	;
	Set gtmtypes("rc_xnsid")="struct"
	Set gtmtypes("rc_xnsid",0)=1
	Set gtmtypes("rc_xnsid","len")=4
	Set gtmtypes("rc_xnsid",1,"name")="rc_xnsid.value"
	Set gtmtypes("rc_xnsid",1,"off")=0
	Set gtmtypes("rc_xnsid",1,"len")=4
	Set gtmtypes("rc_xnsid",1,"type")="char"
	Set gtmtypfldindx("rc_xnsid","value")=1
	;
	Set gtmtypes("rec_hdr")="struct"
	Set gtmtypes("rec_hdr",0)=3
	Set gtmtypes("rec_hdr","len")=4
	Set gtmtypes("rec_hdr",1,"name")="rec_hdr.rsiz"
	Set gtmtypes("rec_hdr",1,"off")=0
	Set gtmtypes("rec_hdr",1,"len")=2
	Set gtmtypes("rec_hdr",1,"type")="unsigned-short"
	Set gtmtypfldindx("rec_hdr","rsiz")=1
	Set gtmtypes("rec_hdr",2,"name")="rec_hdr.cmpc"
	Set gtmtypes("rec_hdr",2,"off")=2
	Set gtmtypes("rec_hdr",2,"len")=1
	Set gtmtypes("rec_hdr",2,"type")="unsigned-char"
	Set gtmtypfldindx("rec_hdr","cmpc")=2
	Set gtmtypes("rec_hdr",3,"name")="rec_hdr.cmpc2"
	Set gtmtypes("rec_hdr",3,"off")=3
	Set gtmtypes("rec_hdr",3,"len")=1
	Set gtmtypes("rec_hdr",3,"type")="unsigned-char"
	Set gtmtypfldindx("rec_hdr","cmpc2")=3
	;
	Set gtmtypes("recvpool_addrs")="struct"
	Set gtmtypes("recvpool_addrs",0)=6
	Set gtmtypes("recvpool_addrs","len")=48
	Set gtmtypes("recvpool_addrs",1,"name")="recvpool_addrs.recvpool_ctl"
	Set gtmtypes("recvpool_addrs",1,"off")=0
	Set gtmtypes("recvpool_addrs",1,"len")=8
	Set gtmtypes("recvpool_addrs",1,"type")="addr"
	Set gtmtypfldindx("recvpool_addrs","recvpool_ctl")=1
	Set gtmtypes("recvpool_addrs",2,"name")="recvpool_addrs.upd_proc_local"
	Set gtmtypes("recvpool_addrs",2,"off")=8
	Set gtmtypes("recvpool_addrs",2,"len")=8
	Set gtmtypes("recvpool_addrs",2,"type")="addr"
	Set gtmtypfldindx("recvpool_addrs","upd_proc_local")=2
	Set gtmtypes("recvpool_addrs",3,"name")="recvpool_addrs.gtmrecv_local"
	Set gtmtypes("recvpool_addrs",3,"off")=16
	Set gtmtypes("recvpool_addrs",3,"len")=8
	Set gtmtypes("recvpool_addrs",3,"type")="addr"
	Set gtmtypfldindx("recvpool_addrs","gtmrecv_local")=3
	Set gtmtypes("recvpool_addrs",4,"name")="recvpool_addrs.upd_helper_ctl"
	Set gtmtypes("recvpool_addrs",4,"off")=24
	Set gtmtypes("recvpool_addrs",4,"len")=8
	Set gtmtypes("recvpool_addrs",4,"type")="addr"
	Set gtmtypfldindx("recvpool_addrs","upd_helper_ctl")=4
	Set gtmtypes("recvpool_addrs",5,"name")="recvpool_addrs.recvdata_base"
	Set gtmtypes("recvpool_addrs",5,"off")=32
	Set gtmtypes("recvpool_addrs",5,"len")=8
	Set gtmtypes("recvpool_addrs",5,"type")="addr"
	Set gtmtypfldindx("recvpool_addrs","recvdata_base")=5
	Set gtmtypes("recvpool_addrs",6,"name")="recvpool_addrs.recvpool_dummy_reg"
	Set gtmtypes("recvpool_addrs",6,"off")=40
	Set gtmtypes("recvpool_addrs",6,"len")=8
	Set gtmtypes("recvpool_addrs",6,"type")="addr"
	Set gtmtypfldindx("recvpool_addrs","recvpool_dummy_reg")=6
	;
	Set gtmtypes("recvpool_ctl_struct")="struct"
	Set gtmtypes("recvpool_ctl_struct",0)=87
	Set gtmtypes("recvpool_ctl_struct","len")=6024
	Set gtmtypes("recvpool_ctl_struct",1,"name")="recvpool_ctl_struct.recvpool_id"
	Set gtmtypes("recvpool_ctl_struct",1,"off")=0
	Set gtmtypes("recvpool_ctl_struct",1,"len")=312
	Set gtmtypes("recvpool_ctl_struct",1,"type")="replpool_identifier"
	Set gtmtypfldindx("recvpool_ctl_struct","recvpool_id")=1
	Set gtmtypes("recvpool_ctl_struct",2,"name")="recvpool_ctl_struct.recvpool_id.label"
	Set gtmtypes("recvpool_ctl_struct",2,"off")=0
	Set gtmtypes("recvpool_ctl_struct",2,"len")=12
	Set gtmtypes("recvpool_ctl_struct",2,"type")="unsigned-char"
	Set gtmtypfldindx("recvpool_ctl_struct","recvpool_id.label")=2
	Set gtmtypes("recvpool_ctl_struct",3,"name")="recvpool_ctl_struct.recvpool_id.pool_type"
	Set gtmtypes("recvpool_ctl_struct",3,"off")=12
	Set gtmtypes("recvpool_ctl_struct",3,"len")=1
	Set gtmtypes("recvpool_ctl_struct",3,"type")="char"
	Set gtmtypfldindx("recvpool_ctl_struct","recvpool_id.pool_type")=3
	Set gtmtypes("recvpool_ctl_struct",4,"name")="recvpool_ctl_struct.recvpool_id.now_running"
	Set gtmtypes("recvpool_ctl_struct",4,"off")=13
	Set gtmtypes("recvpool_ctl_struct",4,"len")=36
	Set gtmtypes("recvpool_ctl_struct",4,"type")="char"
	Set gtmtypfldindx("recvpool_ctl_struct","recvpool_id.now_running")=4
	Set gtmtypes("recvpool_ctl_struct",5,"name")="recvpool_ctl_struct.recvpool_id.repl_pool_key_filler"
	Set gtmtypes("recvpool_ctl_struct",5,"off")=52
	Set gtmtypes("recvpool_ctl_struct",5,"len")=4
	Set gtmtypes("recvpool_ctl_struct",5,"type")="int"
	Set gtmtypfldindx("recvpool_ctl_struct","recvpool_id.repl_pool_key_filler")=5
	Set gtmtypes("recvpool_ctl_struct",6,"name")="recvpool_ctl_struct.recvpool_id.instfilename"
	Set gtmtypes("recvpool_ctl_struct",6,"off")=56
	Set gtmtypes("recvpool_ctl_struct",6,"len")=256
	Set gtmtypes("recvpool_ctl_struct",6,"type")="char"
	Set gtmtypfldindx("recvpool_ctl_struct","recvpool_id.instfilename")=6
	Set gtmtypes("recvpool_ctl_struct",7,"name")="recvpool_ctl_struct.jnl_seqno"
	Set gtmtypes("recvpool_ctl_struct",7,"off")=312
	Set gtmtypes("recvpool_ctl_struct",7,"len")=8
	Set gtmtypes("recvpool_ctl_struct",7,"type")="uint64_t"
	Set gtmtypfldindx("recvpool_ctl_struct","jnl_seqno")=7
	Set gtmtypes("recvpool_ctl_struct",8,"name")="recvpool_ctl_struct.old_jnl_seqno"
	Set gtmtypes("recvpool_ctl_struct",8,"off")=320
	Set gtmtypes("recvpool_ctl_struct",8,"len")=8
	Set gtmtypes("recvpool_ctl_struct",8,"type")="uint64_t"
	Set gtmtypfldindx("recvpool_ctl_struct","old_jnl_seqno")=8
	Set gtmtypes("recvpool_ctl_struct",9,"name")="recvpool_ctl_struct.this_side"
	Set gtmtypes("recvpool_ctl_struct",9,"off")=328
	Set gtmtypes("recvpool_ctl_struct",9,"len")=48
	Set gtmtypes("recvpool_ctl_struct",9,"type")="repl_conn_info_t"
	Set gtmtypfldindx("recvpool_ctl_struct","this_side")=9
	Set gtmtypes("recvpool_ctl_struct",10,"name")="recvpool_ctl_struct.this_side.proto_ver"
	Set gtmtypes("recvpool_ctl_struct",10,"off")=328
	Set gtmtypes("recvpool_ctl_struct",10,"len")=4
	Set gtmtypes("recvpool_ctl_struct",10,"type")="int"
	Set gtmtypfldindx("recvpool_ctl_struct","this_side.proto_ver")=10
	Set gtmtypes("recvpool_ctl_struct",11,"name")="recvpool_ctl_struct.this_side.jnl_ver"
	Set gtmtypes("recvpool_ctl_struct",11,"off")=332
	Set gtmtypes("recvpool_ctl_struct",11,"len")=4
	Set gtmtypes("recvpool_ctl_struct",11,"type")="unsigned-int"
	Set gtmtypfldindx("recvpool_ctl_struct","this_side.jnl_ver")=11
	Set gtmtypes("recvpool_ctl_struct",12,"name")="recvpool_ctl_struct.this_side.is_std_null_coll"
	Set gtmtypes("recvpool_ctl_struct",12,"off")=336
	Set gtmtypes("recvpool_ctl_struct",12,"len")=4
	Set gtmtypes("recvpool_ctl_struct",12,"type")="boolean_t"
	Set gtmtypfldindx("recvpool_ctl_struct","this_side.is_std_null_coll")=12
	Set gtmtypes("recvpool_ctl_struct",13,"name")="recvpool_ctl_struct.this_side.trigger_supported"
	Set gtmtypes("recvpool_ctl_struct",13,"off")=340
	Set gtmtypes("recvpool_ctl_struct",13,"len")=4
	Set gtmtypes("recvpool_ctl_struct",13,"type")="boolean_t"
	Set gtmtypfldindx("recvpool_ctl_struct","this_side.trigger_supported")=13
	Set gtmtypes("recvpool_ctl_struct",14,"name")="recvpool_ctl_struct.this_side.cross_endian"
	Set gtmtypes("recvpool_ctl_struct",14,"off")=344
	Set gtmtypes("recvpool_ctl_struct",14,"len")=4
	Set gtmtypes("recvpool_ctl_struct",14,"type")="boolean_t"
	Set gtmtypfldindx("recvpool_ctl_struct","this_side.cross_endian")=14
	Set gtmtypes("recvpool_ctl_struct",15,"name")="recvpool_ctl_struct.this_side.endianness_known"
	Set gtmtypes("recvpool_ctl_struct",15,"off")=348
	Set gtmtypes("recvpool_ctl_struct",15,"len")=4
	Set gtmtypes("recvpool_ctl_struct",15,"type")="boolean_t"
	Set gtmtypfldindx("recvpool_ctl_struct","this_side.endianness_known")=15
	Set gtmtypes("recvpool_ctl_struct",16,"name")="recvpool_ctl_struct.this_side.null_subs_xform"
	Set gtmtypes("recvpool_ctl_struct",16,"off")=352
	Set gtmtypes("recvpool_ctl_struct",16,"len")=4
	Set gtmtypes("recvpool_ctl_struct",16,"type")="boolean_t"
	Set gtmtypfldindx("recvpool_ctl_struct","this_side.null_subs_xform")=16
	Set gtmtypes("recvpool_ctl_struct",17,"name")="recvpool_ctl_struct.this_side.is_supplementary"
	Set gtmtypes("recvpool_ctl_struct",17,"off")=356
	Set gtmtypes("recvpool_ctl_struct",17,"len")=4
	Set gtmtypes("recvpool_ctl_struct",17,"type")="boolean_t"
	Set gtmtypfldindx("recvpool_ctl_struct","this_side.is_supplementary")=17
	Set gtmtypes("recvpool_ctl_struct",18,"name")="recvpool_ctl_struct.this_side.tls_requested"
	Set gtmtypes("recvpool_ctl_struct",18,"off")=360
	Set gtmtypes("recvpool_ctl_struct",18,"len")=4
	Set gtmtypes("recvpool_ctl_struct",18,"type")="boolean_t"
	Set gtmtypfldindx("recvpool_ctl_struct","this_side.tls_requested")=18
	Set gtmtypes("recvpool_ctl_struct",19,"name")="recvpool_ctl_struct.this_side.filler_16"
	Set gtmtypes("recvpool_ctl_struct",19,"off")=364
	Set gtmtypes("recvpool_ctl_struct",19,"len")=12
	Set gtmtypes("recvpool_ctl_struct",19,"type")="char"
	Set gtmtypfldindx("recvpool_ctl_struct","this_side.filler_16")=19
	Set gtmtypes("recvpool_ctl_struct",20,"name")="recvpool_ctl_struct.recvdata_base_off"
	Set gtmtypes("recvpool_ctl_struct",20,"off")=376
	Set gtmtypes("recvpool_ctl_struct",20,"len")=8
	Set gtmtypes("recvpool_ctl_struct",20,"type")="uint64_t"
	Set gtmtypfldindx("recvpool_ctl_struct","recvdata_base_off")=20
	Set gtmtypes("recvpool_ctl_struct",21,"name")="recvpool_ctl_struct.recvpool_size"
	Set gtmtypes("recvpool_ctl_struct",21,"off")=384
	Set gtmtypes("recvpool_ctl_struct",21,"len")=8
	Set gtmtypes("recvpool_ctl_struct",21,"type")="uint64_t"
	Set gtmtypfldindx("recvpool_ctl_struct","recvpool_size")=21
	Set gtmtypes("recvpool_ctl_struct",22,"name")="recvpool_ctl_struct.write"
	Set gtmtypes("recvpool_ctl_struct",22,"off")=392
	Set gtmtypes("recvpool_ctl_struct",22,"len")=8
	Set gtmtypes("recvpool_ctl_struct",22,"type")="uint64_t"
	Set gtmtypfldindx("recvpool_ctl_struct","write")=22
	Set gtmtypes("recvpool_ctl_struct",23,"name")="recvpool_ctl_struct.write_wrap"
	Set gtmtypes("recvpool_ctl_struct",23,"off")=400
	Set gtmtypes("recvpool_ctl_struct",23,"len")=8
	Set gtmtypes("recvpool_ctl_struct",23,"type")="uint64_t"
	Set gtmtypfldindx("recvpool_ctl_struct","write_wrap")=23
	Set gtmtypes("recvpool_ctl_struct",24,"name")="recvpool_ctl_struct.wrapped"
	Set gtmtypes("recvpool_ctl_struct",24,"off")=408
	Set gtmtypes("recvpool_ctl_struct",24,"len")=4
	Set gtmtypes("recvpool_ctl_struct",24,"type")="unsigned-int"
	Set gtmtypfldindx("recvpool_ctl_struct","wrapped")=24
	Set gtmtypes("recvpool_ctl_struct",25,"name")="recvpool_ctl_struct.initialized"
	Set gtmtypes("recvpool_ctl_struct",25,"off")=412
	Set gtmtypes("recvpool_ctl_struct",25,"len")=4
	Set gtmtypes("recvpool_ctl_struct",25,"type")="unsigned-int"
	Set gtmtypfldindx("recvpool_ctl_struct","initialized")=25
	Set gtmtypes("recvpool_ctl_struct",26,"name")="recvpool_ctl_struct.fresh_start"
	Set gtmtypes("recvpool_ctl_struct",26,"off")=416
	Set gtmtypes("recvpool_ctl_struct",26,"len")=4
	Set gtmtypes("recvpool_ctl_struct",26,"type")="unsigned-int"
	Set gtmtypfldindx("recvpool_ctl_struct","fresh_start")=26
	Set gtmtypes("recvpool_ctl_struct",27,"name")="recvpool_ctl_struct.last_rcvd_histinfo"
	Set gtmtypes("recvpool_ctl_struct",27,"off")=424
	Set gtmtypes("recvpool_ctl_struct",27,"len")=160
	Set gtmtypes("recvpool_ctl_struct",27,"type")="repl_histinfo"
	Set gtmtypfldindx("recvpool_ctl_struct","last_rcvd_histinfo")=27
	Set gtmtypes("recvpool_ctl_struct",28,"name")="recvpool_ctl_struct.last_rcvd_histinfo.root_primary_instname"
	Set gtmtypes("recvpool_ctl_struct",28,"off")=424
	Set gtmtypes("recvpool_ctl_struct",28,"len")=16
	Set gtmtypes("recvpool_ctl_struct",28,"type")="unsigned-char"
	Set gtmtypfldindx("recvpool_ctl_struct","last_rcvd_histinfo.root_primary_instname")=28
	Set gtmtypes("recvpool_ctl_struct",29,"name")="recvpool_ctl_struct.last_rcvd_histinfo.start_seqno"
	Set gtmtypes("recvpool_ctl_struct",29,"off")=440
	Set gtmtypes("recvpool_ctl_struct",29,"len")=8
	Set gtmtypes("recvpool_ctl_struct",29,"type")="uint64_t"
	Set gtmtypfldindx("recvpool_ctl_struct","last_rcvd_histinfo.start_seqno")=29
	Set gtmtypes("recvpool_ctl_struct",30,"name")="recvpool_ctl_struct.last_rcvd_histinfo.strm_seqno"
	Set gtmtypes("recvpool_ctl_struct",30,"off")=448
	Set gtmtypes("recvpool_ctl_struct",30,"len")=8
	Set gtmtypes("recvpool_ctl_struct",30,"type")="uint64_t"
	Set gtmtypfldindx("recvpool_ctl_struct","last_rcvd_histinfo.strm_seqno")=30
	Set gtmtypes("recvpool_ctl_struct",31,"name")="recvpool_ctl_struct.last_rcvd_histinfo.root_primary_cycle"
	Set gtmtypes("recvpool_ctl_struct",31,"off")=456
	Set gtmtypes("recvpool_ctl_struct",31,"len")=4
	Set gtmtypes("recvpool_ctl_struct",31,"type")="unsigned-int"
	Set gtmtypfldindx("recvpool_ctl_struct","last_rcvd_histinfo.root_primary_cycle")=31
	Set gtmtypes("recvpool_ctl_struct",32,"name")="recvpool_ctl_struct.last_rcvd_histinfo.creator_pid"
	Set gtmtypes("recvpool_ctl_struct",32,"off")=460
	Set gtmtypes("recvpool_ctl_struct",32,"len")=4
	Set gtmtypes("recvpool_ctl_struct",32,"type")="unsigned-int"
	Set gtmtypfldindx("recvpool_ctl_struct","last_rcvd_histinfo.creator_pid")=32
	Set gtmtypes("recvpool_ctl_struct",33,"name")="recvpool_ctl_struct.last_rcvd_histinfo.created_time"
	Set gtmtypes("recvpool_ctl_struct",33,"off")=464
	Set gtmtypes("recvpool_ctl_struct",33,"len")=4
	Set gtmtypes("recvpool_ctl_struct",33,"type")="unsigned-int"
	Set gtmtypfldindx("recvpool_ctl_struct","last_rcvd_histinfo.created_time")=33
	Set gtmtypes("recvpool_ctl_struct",34,"name")="recvpool_ctl_struct.last_rcvd_histinfo.histinfo_num"
	Set gtmtypes("recvpool_ctl_struct",34,"off")=468
	Set gtmtypes("recvpool_ctl_struct",34,"len")=4
	Set gtmtypes("recvpool_ctl_struct",34,"type")="int"
	Set gtmtypfldindx("recvpool_ctl_struct","last_rcvd_histinfo.histinfo_num")=34
	Set gtmtypes("recvpool_ctl_struct",35,"name")="recvpool_ctl_struct.last_rcvd_histinfo.prev_histinfo_num"
	Set gtmtypes("recvpool_ctl_struct",35,"off")=472
	Set gtmtypes("recvpool_ctl_struct",35,"len")=4
	Set gtmtypes("recvpool_ctl_struct",35,"type")="int"
	Set gtmtypfldindx("recvpool_ctl_struct","last_rcvd_histinfo.prev_histinfo_num")=35
	Set gtmtypes("recvpool_ctl_struct",36,"name")="recvpool_ctl_struct.last_rcvd_histinfo.strm_index"
	Set gtmtypes("recvpool_ctl_struct",36,"off")=476
	Set gtmtypes("recvpool_ctl_struct",36,"len")=1
	Set gtmtypes("recvpool_ctl_struct",36,"type")="char"
	Set gtmtypfldindx("recvpool_ctl_struct","last_rcvd_histinfo.strm_index")=36
	Set gtmtypes("recvpool_ctl_struct",37,"name")="recvpool_ctl_struct.last_rcvd_histinfo.history_type"
	Set gtmtypes("recvpool_ctl_struct",37,"off")=477
	Set gtmtypes("recvpool_ctl_struct",37,"len")=1
	Set gtmtypes("recvpool_ctl_struct",37,"type")="char"
	Set gtmtypfldindx("recvpool_ctl_struct","last_rcvd_histinfo.history_type")=37
	Set gtmtypes("recvpool_ctl_struct",38,"name")="recvpool_ctl_struct.last_rcvd_histinfo.filler_8"
	Set gtmtypes("recvpool_ctl_struct",38,"off")=478
	Set gtmtypes("recvpool_ctl_struct",38,"len")=2
	Set gtmtypes("recvpool_ctl_struct",38,"type")="char"
	Set gtmtypfldindx("recvpool_ctl_struct","last_rcvd_histinfo.filler_8")=38
	Set gtmtypes("recvpool_ctl_struct",39,"name")="recvpool_ctl_struct.last_rcvd_histinfo.lms_group"
	Set gtmtypes("recvpool_ctl_struct",39,"off")=480
	Set gtmtypes("recvpool_ctl_struct",39,"len")=40
	Set gtmtypes("recvpool_ctl_struct",39,"type")="repl_inst_uuid"
	Set gtmtypfldindx("recvpool_ctl_struct","last_rcvd_histinfo.lms_group")=39
	Set gtmtypes("recvpool_ctl_struct",40,"name")="recvpool_ctl_struct.last_rcvd_histinfo.last_histinfo_num"
	Set gtmtypes("recvpool_ctl_struct",40,"off")=520
	Set gtmtypes("recvpool_ctl_struct",40,"len")=64
	Set gtmtypes("recvpool_ctl_struct",40,"type")="int"
	Set gtmtypfldindx("recvpool_ctl_struct","last_rcvd_histinfo.last_histinfo_num")=40
	Set gtmtypes("recvpool_ctl_struct",40,"dim")=16
	Set gtmtypes("recvpool_ctl_struct",41,"name")="recvpool_ctl_struct.last_valid_histinfo"
	Set gtmtypes("recvpool_ctl_struct",41,"off")=584
	Set gtmtypes("recvpool_ctl_struct",41,"len")=160
	Set gtmtypes("recvpool_ctl_struct",41,"type")="repl_histinfo"
	Set gtmtypfldindx("recvpool_ctl_struct","last_valid_histinfo")=41
	Set gtmtypes("recvpool_ctl_struct",42,"name")="recvpool_ctl_struct.last_valid_histinfo.root_primary_instname"
	Set gtmtypes("recvpool_ctl_struct",42,"off")=584
	Set gtmtypes("recvpool_ctl_struct",42,"len")=16
	Set gtmtypes("recvpool_ctl_struct",42,"type")="unsigned-char"
	Set gtmtypfldindx("recvpool_ctl_struct","last_valid_histinfo.root_primary_instname")=42
	Set gtmtypes("recvpool_ctl_struct",43,"name")="recvpool_ctl_struct.last_valid_histinfo.start_seqno"
	Set gtmtypes("recvpool_ctl_struct",43,"off")=600
	Set gtmtypes("recvpool_ctl_struct",43,"len")=8
	Set gtmtypes("recvpool_ctl_struct",43,"type")="uint64_t"
	Set gtmtypfldindx("recvpool_ctl_struct","last_valid_histinfo.start_seqno")=43
	Set gtmtypes("recvpool_ctl_struct",44,"name")="recvpool_ctl_struct.last_valid_histinfo.strm_seqno"
	Set gtmtypes("recvpool_ctl_struct",44,"off")=608
	Set gtmtypes("recvpool_ctl_struct",44,"len")=8
	Set gtmtypes("recvpool_ctl_struct",44,"type")="uint64_t"
	Set gtmtypfldindx("recvpool_ctl_struct","last_valid_histinfo.strm_seqno")=44
	Set gtmtypes("recvpool_ctl_struct",45,"name")="recvpool_ctl_struct.last_valid_histinfo.root_primary_cycle"
	Set gtmtypes("recvpool_ctl_struct",45,"off")=616
	Set gtmtypes("recvpool_ctl_struct",45,"len")=4
	Set gtmtypes("recvpool_ctl_struct",45,"type")="unsigned-int"
	Set gtmtypfldindx("recvpool_ctl_struct","last_valid_histinfo.root_primary_cycle")=45
	Set gtmtypes("recvpool_ctl_struct",46,"name")="recvpool_ctl_struct.last_valid_histinfo.creator_pid"
	Set gtmtypes("recvpool_ctl_struct",46,"off")=620
	Set gtmtypes("recvpool_ctl_struct",46,"len")=4
	Set gtmtypes("recvpool_ctl_struct",46,"type")="unsigned-int"
	Set gtmtypfldindx("recvpool_ctl_struct","last_valid_histinfo.creator_pid")=46
	Set gtmtypes("recvpool_ctl_struct",47,"name")="recvpool_ctl_struct.last_valid_histinfo.created_time"
	Set gtmtypes("recvpool_ctl_struct",47,"off")=624
	Set gtmtypes("recvpool_ctl_struct",47,"len")=4
	Set gtmtypes("recvpool_ctl_struct",47,"type")="unsigned-int"
	Set gtmtypfldindx("recvpool_ctl_struct","last_valid_histinfo.created_time")=47
	Set gtmtypes("recvpool_ctl_struct",48,"name")="recvpool_ctl_struct.last_valid_histinfo.histinfo_num"
	Set gtmtypes("recvpool_ctl_struct",48,"off")=628
	Set gtmtypes("recvpool_ctl_struct",48,"len")=4
	Set gtmtypes("recvpool_ctl_struct",48,"type")="int"
	Set gtmtypfldindx("recvpool_ctl_struct","last_valid_histinfo.histinfo_num")=48
	Set gtmtypes("recvpool_ctl_struct",49,"name")="recvpool_ctl_struct.last_valid_histinfo.prev_histinfo_num"
	Set gtmtypes("recvpool_ctl_struct",49,"off")=632
	Set gtmtypes("recvpool_ctl_struct",49,"len")=4
	Set gtmtypes("recvpool_ctl_struct",49,"type")="int"
	Set gtmtypfldindx("recvpool_ctl_struct","last_valid_histinfo.prev_histinfo_num")=49
	Set gtmtypes("recvpool_ctl_struct",50,"name")="recvpool_ctl_struct.last_valid_histinfo.strm_index"
	Set gtmtypes("recvpool_ctl_struct",50,"off")=636
	Set gtmtypes("recvpool_ctl_struct",50,"len")=1
	Set gtmtypes("recvpool_ctl_struct",50,"type")="char"
	Set gtmtypfldindx("recvpool_ctl_struct","last_valid_histinfo.strm_index")=50
	Set gtmtypes("recvpool_ctl_struct",51,"name")="recvpool_ctl_struct.last_valid_histinfo.history_type"
	Set gtmtypes("recvpool_ctl_struct",51,"off")=637
	Set gtmtypes("recvpool_ctl_struct",51,"len")=1
	Set gtmtypes("recvpool_ctl_struct",51,"type")="char"
	Set gtmtypfldindx("recvpool_ctl_struct","last_valid_histinfo.history_type")=51
	Set gtmtypes("recvpool_ctl_struct",52,"name")="recvpool_ctl_struct.last_valid_histinfo.filler_8"
	Set gtmtypes("recvpool_ctl_struct",52,"off")=638
	Set gtmtypes("recvpool_ctl_struct",52,"len")=2
	Set gtmtypes("recvpool_ctl_struct",52,"type")="char"
	Set gtmtypfldindx("recvpool_ctl_struct","last_valid_histinfo.filler_8")=52
	Set gtmtypes("recvpool_ctl_struct",53,"name")="recvpool_ctl_struct.last_valid_histinfo.lms_group"
	Set gtmtypes("recvpool_ctl_struct",53,"off")=640
	Set gtmtypes("recvpool_ctl_struct",53,"len")=40
	Set gtmtypes("recvpool_ctl_struct",53,"type")="repl_inst_uuid"
	Set gtmtypfldindx("recvpool_ctl_struct","last_valid_histinfo.lms_group")=53
	Set gtmtypes("recvpool_ctl_struct",54,"name")="recvpool_ctl_struct.last_valid_histinfo.last_histinfo_num"
	Set gtmtypes("recvpool_ctl_struct",54,"off")=680
	Set gtmtypes("recvpool_ctl_struct",54,"len")=64
	Set gtmtypes("recvpool_ctl_struct",54,"type")="int"
	Set gtmtypfldindx("recvpool_ctl_struct","last_valid_histinfo.last_histinfo_num")=54
	Set gtmtypes("recvpool_ctl_struct",54,"dim")=16
	Set gtmtypes("recvpool_ctl_struct",55,"name")="recvpool_ctl_struct.last_rcvd_strm_histinfo"
	Set gtmtypes("recvpool_ctl_struct",55,"off")=744
	Set gtmtypes("recvpool_ctl_struct",55,"len")=2560
	Set gtmtypes("recvpool_ctl_struct",55,"type")="repl_histinfo"
	Set gtmtypfldindx("recvpool_ctl_struct","last_rcvd_strm_histinfo")=55
	Set gtmtypes("recvpool_ctl_struct",55,"dim")=16
	Set gtmtypes("recvpool_ctl_struct",56,"name")="recvpool_ctl_struct.last_rcvd_strm_histinfo[0].root_primary_instname"
	Set gtmtypes("recvpool_ctl_struct",56,"off")=744
	Set gtmtypes("recvpool_ctl_struct",56,"len")=16
	Set gtmtypes("recvpool_ctl_struct",56,"type")="unsigned-char"
	Set gtmtypfldindx("recvpool_ctl_struct","last_rcvd_strm_histinfo[0].root_primary_instname")=56
	Set gtmtypes("recvpool_ctl_struct",57,"name")="recvpool_ctl_struct.last_rcvd_strm_histinfo[0].start_seqno"
	Set gtmtypes("recvpool_ctl_struct",57,"off")=760
	Set gtmtypes("recvpool_ctl_struct",57,"len")=8
	Set gtmtypes("recvpool_ctl_struct",57,"type")="uint64_t"
	Set gtmtypfldindx("recvpool_ctl_struct","last_rcvd_strm_histinfo[0].start_seqno")=57
	Set gtmtypes("recvpool_ctl_struct",58,"name")="recvpool_ctl_struct.last_rcvd_strm_histinfo[0].strm_seqno"
	Set gtmtypes("recvpool_ctl_struct",58,"off")=768
	Set gtmtypes("recvpool_ctl_struct",58,"len")=8
	Set gtmtypes("recvpool_ctl_struct",58,"type")="uint64_t"
	Set gtmtypfldindx("recvpool_ctl_struct","last_rcvd_strm_histinfo[0].strm_seqno")=58
	Set gtmtypes("recvpool_ctl_struct",59,"name")="recvpool_ctl_struct.last_rcvd_strm_histinfo[0].root_primary_cycle"
	Set gtmtypes("recvpool_ctl_struct",59,"off")=776
	Set gtmtypes("recvpool_ctl_struct",59,"len")=4
	Set gtmtypes("recvpool_ctl_struct",59,"type")="unsigned-int"
	Set gtmtypfldindx("recvpool_ctl_struct","last_rcvd_strm_histinfo[0].root_primary_cycle")=59
	Set gtmtypes("recvpool_ctl_struct",60,"name")="recvpool_ctl_struct.last_rcvd_strm_histinfo[0].creator_pid"
	Set gtmtypes("recvpool_ctl_struct",60,"off")=780
	Set gtmtypes("recvpool_ctl_struct",60,"len")=4
	Set gtmtypes("recvpool_ctl_struct",60,"type")="unsigned-int"
	Set gtmtypfldindx("recvpool_ctl_struct","last_rcvd_strm_histinfo[0].creator_pid")=60
	Set gtmtypes("recvpool_ctl_struct",61,"name")="recvpool_ctl_struct.last_rcvd_strm_histinfo[0].created_time"
	Set gtmtypes("recvpool_ctl_struct",61,"off")=784
	Set gtmtypes("recvpool_ctl_struct",61,"len")=4
	Set gtmtypes("recvpool_ctl_struct",61,"type")="unsigned-int"
	Set gtmtypfldindx("recvpool_ctl_struct","last_rcvd_strm_histinfo[0].created_time")=61
	Set gtmtypes("recvpool_ctl_struct",62,"name")="recvpool_ctl_struct.last_rcvd_strm_histinfo[0].histinfo_num"
	Set gtmtypes("recvpool_ctl_struct",62,"off")=788
	Set gtmtypes("recvpool_ctl_struct",62,"len")=4
	Set gtmtypes("recvpool_ctl_struct",62,"type")="int"
	Set gtmtypfldindx("recvpool_ctl_struct","last_rcvd_strm_histinfo[0].histinfo_num")=62
	Set gtmtypes("recvpool_ctl_struct",63,"name")="recvpool_ctl_struct.last_rcvd_strm_histinfo[0].prev_histinfo_num"
	Set gtmtypes("recvpool_ctl_struct",63,"off")=792
	Set gtmtypes("recvpool_ctl_struct",63,"len")=4
	Set gtmtypes("recvpool_ctl_struct",63,"type")="int"
	Set gtmtypfldindx("recvpool_ctl_struct","last_rcvd_strm_histinfo[0].prev_histinfo_num")=63
	Set gtmtypes("recvpool_ctl_struct",64,"name")="recvpool_ctl_struct.last_rcvd_strm_histinfo[0].strm_index"
	Set gtmtypes("recvpool_ctl_struct",64,"off")=796
	Set gtmtypes("recvpool_ctl_struct",64,"len")=1
	Set gtmtypes("recvpool_ctl_struct",64,"type")="char"
	Set gtmtypfldindx("recvpool_ctl_struct","last_rcvd_strm_histinfo[0].strm_index")=64
	Set gtmtypes("recvpool_ctl_struct",65,"name")="recvpool_ctl_struct.last_rcvd_strm_histinfo[0].history_type"
	Set gtmtypes("recvpool_ctl_struct",65,"off")=797
	Set gtmtypes("recvpool_ctl_struct",65,"len")=1
	Set gtmtypes("recvpool_ctl_struct",65,"type")="char"
	Set gtmtypfldindx("recvpool_ctl_struct","last_rcvd_strm_histinfo[0].history_type")=65
	Set gtmtypes("recvpool_ctl_struct",66,"name")="recvpool_ctl_struct.last_rcvd_strm_histinfo[0].filler_8"
	Set gtmtypes("recvpool_ctl_struct",66,"off")=798
	Set gtmtypes("recvpool_ctl_struct",66,"len")=2
	Set gtmtypes("recvpool_ctl_struct",66,"type")="char"
	Set gtmtypfldindx("recvpool_ctl_struct","last_rcvd_strm_histinfo[0].filler_8")=66
	Set gtmtypes("recvpool_ctl_struct",67,"name")="recvpool_ctl_struct.last_rcvd_strm_histinfo[0].lms_group"
	Set gtmtypes("recvpool_ctl_struct",67,"off")=800
	Set gtmtypes("recvpool_ctl_struct",67,"len")=40
	Set gtmtypes("recvpool_ctl_struct",67,"type")="repl_inst_uuid"
	Set gtmtypfldindx("recvpool_ctl_struct","last_rcvd_strm_histinfo[0].lms_group")=67
	Set gtmtypes("recvpool_ctl_struct",68,"name")="recvpool_ctl_struct.last_rcvd_strm_histinfo[0].last_histinfo_num"
	Set gtmtypes("recvpool_ctl_struct",68,"off")=840
	Set gtmtypes("recvpool_ctl_struct",68,"len")=64
	Set gtmtypes("recvpool_ctl_struct",68,"type")="int"
	Set gtmtypfldindx("recvpool_ctl_struct","last_rcvd_strm_histinfo[0].last_histinfo_num")=68
	Set gtmtypes("recvpool_ctl_struct",68,"dim")=16
	Set gtmtypes("recvpool_ctl_struct",69,"name")="recvpool_ctl_struct.last_valid_strm_histinfo"
	Set gtmtypes("recvpool_ctl_struct",69,"off")=3304
	Set gtmtypes("recvpool_ctl_struct",69,"len")=2560
	Set gtmtypes("recvpool_ctl_struct",69,"type")="repl_histinfo"
	Set gtmtypfldindx("recvpool_ctl_struct","last_valid_strm_histinfo")=69
	Set gtmtypes("recvpool_ctl_struct",69,"dim")=16
	Set gtmtypes("recvpool_ctl_struct",70,"name")="recvpool_ctl_struct.last_valid_strm_histinfo[0].root_primary_instname"
	Set gtmtypes("recvpool_ctl_struct",70,"off")=3304
	Set gtmtypes("recvpool_ctl_struct",70,"len")=16
	Set gtmtypes("recvpool_ctl_struct",70,"type")="unsigned-char"
	Set gtmtypfldindx("recvpool_ctl_struct","last_valid_strm_histinfo[0].root_primary_instname")=70
	Set gtmtypes("recvpool_ctl_struct",71,"name")="recvpool_ctl_struct.last_valid_strm_histinfo[0].start_seqno"
	Set gtmtypes("recvpool_ctl_struct",71,"off")=3320
	Set gtmtypes("recvpool_ctl_struct",71,"len")=8
	Set gtmtypes("recvpool_ctl_struct",71,"type")="uint64_t"
	Set gtmtypfldindx("recvpool_ctl_struct","last_valid_strm_histinfo[0].start_seqno")=71
	Set gtmtypes("recvpool_ctl_struct",72,"name")="recvpool_ctl_struct.last_valid_strm_histinfo[0].strm_seqno"
	Set gtmtypes("recvpool_ctl_struct",72,"off")=3328
	Set gtmtypes("recvpool_ctl_struct",72,"len")=8
	Set gtmtypes("recvpool_ctl_struct",72,"type")="uint64_t"
	Set gtmtypfldindx("recvpool_ctl_struct","last_valid_strm_histinfo[0].strm_seqno")=72
	Set gtmtypes("recvpool_ctl_struct",73,"name")="recvpool_ctl_struct.last_valid_strm_histinfo[0].root_primary_cycle"
	Set gtmtypes("recvpool_ctl_struct",73,"off")=3336
	Set gtmtypes("recvpool_ctl_struct",73,"len")=4
	Set gtmtypes("recvpool_ctl_struct",73,"type")="unsigned-int"
	Set gtmtypfldindx("recvpool_ctl_struct","last_valid_strm_histinfo[0].root_primary_cycle")=73
	Set gtmtypes("recvpool_ctl_struct",74,"name")="recvpool_ctl_struct.last_valid_strm_histinfo[0].creator_pid"
	Set gtmtypes("recvpool_ctl_struct",74,"off")=3340
	Set gtmtypes("recvpool_ctl_struct",74,"len")=4
	Set gtmtypes("recvpool_ctl_struct",74,"type")="unsigned-int"
	Set gtmtypfldindx("recvpool_ctl_struct","last_valid_strm_histinfo[0].creator_pid")=74
	Set gtmtypes("recvpool_ctl_struct",75,"name")="recvpool_ctl_struct.last_valid_strm_histinfo[0].created_time"
	Set gtmtypes("recvpool_ctl_struct",75,"off")=3344
	Set gtmtypes("recvpool_ctl_struct",75,"len")=4
	Set gtmtypes("recvpool_ctl_struct",75,"type")="unsigned-int"
	Set gtmtypfldindx("recvpool_ctl_struct","last_valid_strm_histinfo[0].created_time")=75
	Set gtmtypes("recvpool_ctl_struct",76,"name")="recvpool_ctl_struct.last_valid_strm_histinfo[0].histinfo_num"
	Set gtmtypes("recvpool_ctl_struct",76,"off")=3348
	Set gtmtypes("recvpool_ctl_struct",76,"len")=4
	Set gtmtypes("recvpool_ctl_struct",76,"type")="int"
	Set gtmtypfldindx("recvpool_ctl_struct","last_valid_strm_histinfo[0].histinfo_num")=76
	Set gtmtypes("recvpool_ctl_struct",77,"name")="recvpool_ctl_struct.last_valid_strm_histinfo[0].prev_histinfo_num"
	Set gtmtypes("recvpool_ctl_struct",77,"off")=3352
	Set gtmtypes("recvpool_ctl_struct",77,"len")=4
	Set gtmtypes("recvpool_ctl_struct",77,"type")="int"
	Set gtmtypfldindx("recvpool_ctl_struct","last_valid_strm_histinfo[0].prev_histinfo_num")=77
	Set gtmtypes("recvpool_ctl_struct",78,"name")="recvpool_ctl_struct.last_valid_strm_histinfo[0].strm_index"
	Set gtmtypes("recvpool_ctl_struct",78,"off")=3356
	Set gtmtypes("recvpool_ctl_struct",78,"len")=1
	Set gtmtypes("recvpool_ctl_struct",78,"type")="char"
	Set gtmtypfldindx("recvpool_ctl_struct","last_valid_strm_histinfo[0].strm_index")=78
	Set gtmtypes("recvpool_ctl_struct",79,"name")="recvpool_ctl_struct.last_valid_strm_histinfo[0].history_type"
	Set gtmtypes("recvpool_ctl_struct",79,"off")=3357
	Set gtmtypes("recvpool_ctl_struct",79,"len")=1
	Set gtmtypes("recvpool_ctl_struct",79,"type")="char"
	Set gtmtypfldindx("recvpool_ctl_struct","last_valid_strm_histinfo[0].history_type")=79
	Set gtmtypes("recvpool_ctl_struct",80,"name")="recvpool_ctl_struct.last_valid_strm_histinfo[0].filler_8"
	Set gtmtypes("recvpool_ctl_struct",80,"off")=3358
	Set gtmtypes("recvpool_ctl_struct",80,"len")=2
	Set gtmtypes("recvpool_ctl_struct",80,"type")="char"
	Set gtmtypfldindx("recvpool_ctl_struct","last_valid_strm_histinfo[0].filler_8")=80
	Set gtmtypes("recvpool_ctl_struct",81,"name")="recvpool_ctl_struct.last_valid_strm_histinfo[0].lms_group"
	Set gtmtypes("recvpool_ctl_struct",81,"off")=3360
	Set gtmtypes("recvpool_ctl_struct",81,"len")=40
	Set gtmtypes("recvpool_ctl_struct",81,"type")="repl_inst_uuid"
	Set gtmtypfldindx("recvpool_ctl_struct","last_valid_strm_histinfo[0].lms_group")=81
	Set gtmtypes("recvpool_ctl_struct",82,"name")="recvpool_ctl_struct.last_valid_strm_histinfo[0].last_histinfo_num"
	Set gtmtypes("recvpool_ctl_struct",82,"off")=3400
	Set gtmtypes("recvpool_ctl_struct",82,"len")=64
	Set gtmtypes("recvpool_ctl_struct",82,"type")="int"
	Set gtmtypfldindx("recvpool_ctl_struct","last_valid_strm_histinfo[0].last_histinfo_num")=82
	Set gtmtypes("recvpool_ctl_struct",82,"dim")=16
	Set gtmtypes("recvpool_ctl_struct",83,"name")="recvpool_ctl_struct.is_valid_strm_histinfo"
	Set gtmtypes("recvpool_ctl_struct",83,"off")=5864
	Set gtmtypes("recvpool_ctl_struct",83,"len")=64
	Set gtmtypes("recvpool_ctl_struct",83,"type")="boolean_t"
	Set gtmtypfldindx("recvpool_ctl_struct","is_valid_strm_histinfo")=83
	Set gtmtypes("recvpool_ctl_struct",83,"dim")=16
	Set gtmtypes("recvpool_ctl_struct",84,"name")="recvpool_ctl_struct.max_strm_histinfo"
	Set gtmtypes("recvpool_ctl_struct",84,"off")=5928
	Set gtmtypes("recvpool_ctl_struct",84,"len")=4
	Set gtmtypes("recvpool_ctl_struct",84,"type")="unsigned-int"
	Set gtmtypfldindx("recvpool_ctl_struct","max_strm_histinfo")=84
	Set gtmtypes("recvpool_ctl_struct",85,"name")="recvpool_ctl_struct.insert_strm_histinfo"
	Set gtmtypes("recvpool_ctl_struct",85,"off")=5932
	Set gtmtypes("recvpool_ctl_struct",85,"len")=4
	Set gtmtypes("recvpool_ctl_struct",85,"type")="boolean_t"
	Set gtmtypfldindx("recvpool_ctl_struct","insert_strm_histinfo")=85
	Set gtmtypes("recvpool_ctl_struct",86,"name")="recvpool_ctl_struct.write_updated_ctl"
	Set gtmtypes("recvpool_ctl_struct",86,"off")=5936
	Set gtmtypes("recvpool_ctl_struct",86,"len")=40
	Set gtmtypes("recvpool_ctl_struct",86,"type")="pthread_mutex_t"
	Set gtmtypfldindx("recvpool_ctl_struct","write_updated_ctl")=86
	Set gtmtypes("recvpool_ctl_struct",87,"name")="recvpool_ctl_struct.write_updated"
	Set gtmtypes("recvpool_ctl_struct",87,"off")=5976
	Set gtmtypes("recvpool_ctl_struct",87,"len")=48
	Set gtmtypes("recvpool_ctl_struct",87,"type")="pthread_cond_t"
	Set gtmtypfldindx("recvpool_ctl_struct","write_updated")=87
	;
	Set gtmtypes("redirect_list")="struct"
	Set gtmtypes("redirect_list",0)=5
	Set gtmtypes("redirect_list","len")=32
	Set gtmtypes("redirect_list",1,"name")="redirect_list.next"
	Set gtmtypes("redirect_list",1,"off")=0
	Set gtmtypes("redirect_list",1,"len")=8
	Set gtmtypes("redirect_list",1,"type")="addr"
	Set gtmtypfldindx("redirect_list","next")=1
	Set gtmtypes("redirect_list",2,"name")="redirect_list.org_name_len"
	Set gtmtypes("redirect_list",2,"off")=8
	Set gtmtypes("redirect_list",2,"len")=4
	Set gtmtypes("redirect_list",2,"type")="unsigned-int"
	Set gtmtypfldindx("redirect_list","org_name_len")=2
	Set gtmtypes("redirect_list",3,"name")="redirect_list.new_name_len"
	Set gtmtypes("redirect_list",3,"off")=12
	Set gtmtypes("redirect_list",3,"len")=4
	Set gtmtypes("redirect_list",3,"type")="unsigned-int"
	Set gtmtypfldindx("redirect_list","new_name_len")=3
	Set gtmtypes("redirect_list",4,"name")="redirect_list.org_name"
	Set gtmtypes("redirect_list",4,"off")=16
	Set gtmtypes("redirect_list",4,"len")=8
	Set gtmtypes("redirect_list",4,"type")="addr"
	Set gtmtypfldindx("redirect_list","org_name")=4
	Set gtmtypes("redirect_list",5,"name")="redirect_list.new_name"
	Set gtmtypes("redirect_list",5,"off")=24
	Set gtmtypes("redirect_list",5,"len")=8
	Set gtmtypes("redirect_list",5,"type")="addr"
	Set gtmtypfldindx("redirect_list","new_name")=5
	;
	Set gtmtypes("redo_root_search_context")="struct"
	Set gtmtypes("redo_root_search_context",0)=18
	Set gtmtypes("redo_root_search_context","len")=1112
	Set gtmtypes("redo_root_search_context",1,"name")="redo_root_search_context.t_fail_hist"
	Set gtmtypes("redo_root_search_context",1,"off")=0
	Set gtmtypes("redo_root_search_context",1,"len")=5
	Set gtmtypes("redo_root_search_context",1,"type")="unsigned-char"
	Set gtmtypfldindx("redo_root_search_context","t_fail_hist")=1
	Set gtmtypes("redo_root_search_context",2,"name")="redo_root_search_context.t_tries"
	Set gtmtypes("redo_root_search_context",2,"off")=8
	Set gtmtypes("redo_root_search_context",2,"len")=4
	Set gtmtypes("redo_root_search_context",2,"type")="unsigned-int"
	Set gtmtypfldindx("redo_root_search_context","t_tries")=2
	Set gtmtypes("redo_root_search_context",3,"name")="redo_root_search_context.prev_t_tries"
	Set gtmtypes("redo_root_search_context",3,"off")=12
	Set gtmtypes("redo_root_search_context",3,"len")=4
	Set gtmtypes("redo_root_search_context",3,"type")="unsigned-int"
	Set gtmtypfldindx("redo_root_search_context","prev_t_tries")=3
	Set gtmtypes("redo_root_search_context",4,"name")="redo_root_search_context.inctn_opcode"
	Set gtmtypes("redo_root_search_context",4,"off")=16
	Set gtmtypes("redo_root_search_context",4,"len")=4
	Set gtmtypes("redo_root_search_context",4,"type")="int"
	Set gtmtypfldindx("redo_root_search_context","inctn_opcode")=4
	Set gtmtypes("redo_root_search_context",5,"name")="redo_root_search_context.start_tn"
	Set gtmtypes("redo_root_search_context",5,"off")=24
	Set gtmtypes("redo_root_search_context",5,"len")=8
	Set gtmtypes("redo_root_search_context",5,"type")="uint64_t"
	Set gtmtypfldindx("redo_root_search_context","start_tn")=5
	Set gtmtypes("redo_root_search_context",6,"name")="redo_root_search_context.update_trans"
	Set gtmtypes("redo_root_search_context",6,"off")=32
	Set gtmtypes("redo_root_search_context",6,"len")=4
	Set gtmtypes("redo_root_search_context",6,"type")="unsigned-int"
	Set gtmtypfldindx("redo_root_search_context","update_trans")=6
	Set gtmtypes("redo_root_search_context",7,"name")="redo_root_search_context.t_err"
	Set gtmtypes("redo_root_search_context",7,"off")=36
	Set gtmtypes("redo_root_search_context",7,"len")=4
	Set gtmtypes("redo_root_search_context",7,"type")="unsigned-int"
	Set gtmtypfldindx("redo_root_search_context","t_err")=7
	Set gtmtypes("redo_root_search_context",8,"name")="redo_root_search_context.hold_onto_crit"
	Set gtmtypes("redo_root_search_context",8,"off")=40
	Set gtmtypes("redo_root_search_context",8,"len")=4
	Set gtmtypes("redo_root_search_context",8,"type")="boolean_t"
	Set gtmtypfldindx("redo_root_search_context","hold_onto_crit")=8
	Set gtmtypes("redo_root_search_context",9,"name")="redo_root_search_context.currkey"
	Set gtmtypes("redo_root_search_context",9,"off")=44
	Set gtmtypes("redo_root_search_context",9,"len")=1058
	Set gtmtypes("redo_root_search_context",9,"type")="gv_key_buf"
	Set gtmtypfldindx("redo_root_search_context","currkey")=9
	Set gtmtypes("redo_root_search_context",10,"name")="redo_root_search_context.currkey.key"
	Set gtmtypes("redo_root_search_context",10,"off")=44
	Set gtmtypes("redo_root_search_context",10,"len")=6
	Set gtmtypes("redo_root_search_context",10,"type")="gv_key_nobase"
	Set gtmtypfldindx("redo_root_search_context","currkey.key")=10
	Set gtmtypes("redo_root_search_context",11,"name")="redo_root_search_context.currkey.key.top"
	Set gtmtypes("redo_root_search_context",11,"off")=44
	Set gtmtypes("redo_root_search_context",11,"len")=2
	Set gtmtypes("redo_root_search_context",11,"type")="unsigned-short"
	Set gtmtypfldindx("redo_root_search_context","currkey.key.top")=11
	Set gtmtypes("redo_root_search_context",12,"name")="redo_root_search_context.currkey.key.end"
	Set gtmtypes("redo_root_search_context",12,"off")=46
	Set gtmtypes("redo_root_search_context",12,"len")=2
	Set gtmtypes("redo_root_search_context",12,"type")="unsigned-short"
	Set gtmtypfldindx("redo_root_search_context","currkey.key.end")=12
	Set gtmtypes("redo_root_search_context",13,"name")="redo_root_search_context.currkey.key.prev"
	Set gtmtypes("redo_root_search_context",13,"off")=48
	Set gtmtypes("redo_root_search_context",13,"len")=2
	Set gtmtypes("redo_root_search_context",13,"type")="unsigned-short"
	Set gtmtypfldindx("redo_root_search_context","currkey.key.prev")=13
	Set gtmtypes("redo_root_search_context",14,"name")="redo_root_search_context.currkey.buf"
	Set gtmtypes("redo_root_search_context",14,"off")=44
	Set gtmtypes("redo_root_search_context",14,"len")=1058
	Set gtmtypes("redo_root_search_context",14,"type")="char"
	Set gtmtypfldindx("redo_root_search_context","currkey.buf")=14
	Set gtmtypes("redo_root_search_context",15,"name")="redo_root_search_context.currkey.split"
	Set gtmtypes("redo_root_search_context",15,"off")=44
	Set gtmtypes("redo_root_search_context",15,"len")=1058
	Set gtmtypes("redo_root_search_context",15,"type")="struct"
	Set gtmtypfldindx("redo_root_search_context","currkey.split")=15
	Set gtmtypes("redo_root_search_context",16,"name")="redo_root_search_context.currkey.split.gv_key_data"
	Set gtmtypes("redo_root_search_context",16,"off")=44
	Set gtmtypes("redo_root_search_context",16,"len")=6
	Set gtmtypes("redo_root_search_context",16,"type")="char"
	Set gtmtypfldindx("redo_root_search_context","currkey.split.gv_key_data")=16
	Set gtmtypes("redo_root_search_context",17,"name")="redo_root_search_context.currkey.split.base"
	Set gtmtypes("redo_root_search_context",17,"off")=50
	Set gtmtypes("redo_root_search_context",17,"len")=1052
	Set gtmtypes("redo_root_search_context",17,"type")="char"
	Set gtmtypfldindx("redo_root_search_context","currkey.split.base")=17
	Set gtmtypes("redo_root_search_context",18,"name")="redo_root_search_context.gv_currkey"
	Set gtmtypes("redo_root_search_context",18,"off")=1104
	Set gtmtypes("redo_root_search_context",18,"len")=8
	Set gtmtypes("redo_root_search_context",18,"type")="addr"
	Set gtmtypfldindx("redo_root_search_context","gv_currkey")=18
	;
	Set gtmtypes("reg_ctl_list")="struct"
	Set gtmtypes("reg_ctl_list",0)=64
	Set gtmtypes("reg_ctl_list","len")=488
	Set gtmtypes("reg_ctl_list",1,"name")="reg_ctl_list.db_tn"
	Set gtmtypes("reg_ctl_list",1,"off")=0
	Set gtmtypes("reg_ctl_list",1,"len")=8
	Set gtmtypes("reg_ctl_list",1,"type")="uint64_t"
	Set gtmtypfldindx("reg_ctl_list","db_tn")=1
	Set gtmtypes("reg_ctl_list",2,"name")="reg_ctl_list.csd"
	Set gtmtypes("reg_ctl_list",2,"off")=8
	Set gtmtypes("reg_ctl_list",2,"len")=8
	Set gtmtypes("reg_ctl_list",2,"type")="addr"
	Set gtmtypfldindx("reg_ctl_list","csd")=2
	Set gtmtypes("reg_ctl_list",3,"name")="reg_ctl_list.gd"
	Set gtmtypes("reg_ctl_list",3,"off")=16
	Set gtmtypes("reg_ctl_list",3,"len")=8
	Set gtmtypes("reg_ctl_list",3,"type")="addr"
	Set gtmtypfldindx("reg_ctl_list","gd")=3
	Set gtmtypes("reg_ctl_list",4,"name")="reg_ctl_list.csa"
	Set gtmtypes("reg_ctl_list",4,"off")=24
	Set gtmtypes("reg_ctl_list",4,"len")=8
	Set gtmtypes("reg_ctl_list",4,"type")="addr"
	Set gtmtypfldindx("reg_ctl_list","csa")=4
	Set gtmtypes("reg_ctl_list",5,"name")="reg_ctl_list.sgm_info_ptr"
	Set gtmtypes("reg_ctl_list",5,"off")=32
	Set gtmtypes("reg_ctl_list",5,"len")=8
	Set gtmtypes("reg_ctl_list",5,"type")="addr"
	Set gtmtypfldindx("reg_ctl_list","sgm_info_ptr")=5
	Set gtmtypes("reg_ctl_list",6,"name")="reg_ctl_list.db_ctl"
	Set gtmtypes("reg_ctl_list",6,"off")=40
	Set gtmtypes("reg_ctl_list",6,"len")=8
	Set gtmtypes("reg_ctl_list",6,"type")="addr"
	Set gtmtypfldindx("reg_ctl_list","db_ctl")=6
	Set gtmtypes("reg_ctl_list",7,"name")="reg_ctl_list.jctl"
	Set gtmtypes("reg_ctl_list",7,"off")=48
	Set gtmtypes("reg_ctl_list",7,"len")=8
	Set gtmtypes("reg_ctl_list",7,"type")="addr"
	Set gtmtypfldindx("reg_ctl_list","jctl")=7
	Set gtmtypes("reg_ctl_list",8,"name")="reg_ctl_list.jctl_head"
	Set gtmtypes("reg_ctl_list",8,"off")=56
	Set gtmtypes("reg_ctl_list",8,"len")=8
	Set gtmtypes("reg_ctl_list",8,"type")="addr"
	Set gtmtypfldindx("reg_ctl_list","jctl_head")=8
	Set gtmtypes("reg_ctl_list",9,"name")="reg_ctl_list.jctl_apply_pblk"
	Set gtmtypes("reg_ctl_list",9,"off")=64
	Set gtmtypes("reg_ctl_list",9,"len")=8
	Set gtmtypes("reg_ctl_list",9,"type")="addr"
	Set gtmtypfldindx("reg_ctl_list","jctl_apply_pblk")=9
	Set gtmtypes("reg_ctl_list",10,"name")="reg_ctl_list.jctl_turn_around"
	Set gtmtypes("reg_ctl_list",10,"off")=72
	Set gtmtypes("reg_ctl_list",10,"len")=8
	Set gtmtypes("reg_ctl_list",10,"type")="addr"
	Set gtmtypfldindx("reg_ctl_list","jctl_turn_around")=10
	Set gtmtypes("reg_ctl_list",11,"name")="reg_ctl_list.jctl_alt_head"
	Set gtmtypes("reg_ctl_list",11,"off")=80
	Set gtmtypes("reg_ctl_list",11,"len")=8
	Set gtmtypes("reg_ctl_list",11,"type")="addr"
	Set gtmtypfldindx("reg_ctl_list","jctl_alt_head")=11
	Set gtmtypes("reg_ctl_list",12,"name")="reg_ctl_list.jctl_error"
	Set gtmtypes("reg_ctl_list",12,"off")=88
	Set gtmtypes("reg_ctl_list",12,"len")=8
	Set gtmtypes("reg_ctl_list",12,"type")="addr"
	Set gtmtypfldindx("reg_ctl_list","jctl_error")=12
	Set gtmtypes("reg_ctl_list",13,"name")="reg_ctl_list.gvntab"
	Set gtmtypes("reg_ctl_list",13,"off")=96
	Set gtmtypes("reg_ctl_list",13,"len")=80
	Set gtmtypes("reg_ctl_list",13,"type")="hash_table_mname"
	Set gtmtypfldindx("reg_ctl_list","gvntab")=13
	Set gtmtypes("reg_ctl_list",14,"name")="reg_ctl_list.gvntab.base"
	Set gtmtypes("reg_ctl_list",14,"off")=96
	Set gtmtypes("reg_ctl_list",14,"len")=8
	Set gtmtypes("reg_ctl_list",14,"type")="addr"
	Set gtmtypfldindx("reg_ctl_list","gvntab.base")=14
	Set gtmtypes("reg_ctl_list",15,"name")="reg_ctl_list.gvntab.top"
	Set gtmtypes("reg_ctl_list",15,"off")=104
	Set gtmtypes("reg_ctl_list",15,"len")=8
	Set gtmtypes("reg_ctl_list",15,"type")="addr"
	Set gtmtypfldindx("reg_ctl_list","gvntab.top")=15
	Set gtmtypes("reg_ctl_list",16,"name")="reg_ctl_list.gvntab.spare_base"
	Set gtmtypes("reg_ctl_list",16,"off")=112
	Set gtmtypes("reg_ctl_list",16,"len")=8
	Set gtmtypes("reg_ctl_list",16,"type")="addr"
	Set gtmtypfldindx("reg_ctl_list","gvntab.spare_base")=16
	Set gtmtypes("reg_ctl_list",17,"name")="reg_ctl_list.gvntab.entry_passed_thru"
	Set gtmtypes("reg_ctl_list",17,"off")=120
	Set gtmtypes("reg_ctl_list",17,"len")=8
	Set gtmtypes("reg_ctl_list",17,"type")="addr"
	Set gtmtypfldindx("reg_ctl_list","gvntab.entry_passed_thru")=17
	Set gtmtypes("reg_ctl_list",18,"name")="reg_ctl_list.gvntab.size"
	Set gtmtypes("reg_ctl_list",18,"off")=128
	Set gtmtypes("reg_ctl_list",18,"len")=4
	Set gtmtypes("reg_ctl_list",18,"type")="unsigned-int"
	Set gtmtypfldindx("reg_ctl_list","gvntab.size")=18
	Set gtmtypes("reg_ctl_list",19,"name")="reg_ctl_list.gvntab.initial_size"
	Set gtmtypes("reg_ctl_list",19,"off")=132
	Set gtmtypes("reg_ctl_list",19,"len")=4
	Set gtmtypes("reg_ctl_list",19,"type")="unsigned-int"
	Set gtmtypfldindx("reg_ctl_list","gvntab.initial_size")=19
	Set gtmtypes("reg_ctl_list",20,"name")="reg_ctl_list.gvntab.spare_base_size"
	Set gtmtypes("reg_ctl_list",20,"off")=136
	Set gtmtypes("reg_ctl_list",20,"len")=4
	Set gtmtypes("reg_ctl_list",20,"type")="unsigned-int"
	Set gtmtypfldindx("reg_ctl_list","gvntab.spare_base_size")=20
	Set gtmtypes("reg_ctl_list",21,"name")="reg_ctl_list.gvntab.count"
	Set gtmtypes("reg_ctl_list",21,"off")=140
	Set gtmtypes("reg_ctl_list",21,"len")=4
	Set gtmtypes("reg_ctl_list",21,"type")="unsigned-int"
	Set gtmtypfldindx("reg_ctl_list","gvntab.count")=21
	Set gtmtypes("reg_ctl_list",22,"name")="reg_ctl_list.gvntab.del_count"
	Set gtmtypes("reg_ctl_list",22,"off")=144
	Set gtmtypes("reg_ctl_list",22,"len")=4
	Set gtmtypes("reg_ctl_list",22,"type")="unsigned-int"
	Set gtmtypfldindx("reg_ctl_list","gvntab.del_count")=22
	Set gtmtypes("reg_ctl_list",23,"name")="reg_ctl_list.gvntab.exp_trigger_size"
	Set gtmtypes("reg_ctl_list",23,"off")=148
	Set gtmtypes("reg_ctl_list",23,"len")=4
	Set gtmtypes("reg_ctl_list",23,"type")="unsigned-int"
	Set gtmtypfldindx("reg_ctl_list","gvntab.exp_trigger_size")=23
	Set gtmtypes("reg_ctl_list",24,"name")="reg_ctl_list.gvntab.cmp_trigger_size"
	Set gtmtypes("reg_ctl_list",24,"off")=152
	Set gtmtypes("reg_ctl_list",24,"len")=4
	Set gtmtypes("reg_ctl_list",24,"type")="unsigned-int"
	Set gtmtypfldindx("reg_ctl_list","gvntab.cmp_trigger_size")=24
	Set gtmtypes("reg_ctl_list",25,"name")="reg_ctl_list.gvntab.dont_compact"
	Set gtmtypes("reg_ctl_list",25,"off")=156
	Set gtmtypes("reg_ctl_list",25,"len")=4
	Set gtmtypes("reg_ctl_list",25,"type")="boolean_t"
	Set gtmtypfldindx("reg_ctl_list","gvntab.dont_compact")=25
	Set gtmtypes("reg_ctl_list",26,"name")="reg_ctl_list.gvntab.dont_keep_spare_table"
	Set gtmtypes("reg_ctl_list",26,"off")=160
	Set gtmtypes("reg_ctl_list",26,"len")=4
	Set gtmtypes("reg_ctl_list",26,"type")="boolean_t"
	Set gtmtypfldindx("reg_ctl_list","gvntab.dont_keep_spare_table")=26
	Set gtmtypes("reg_ctl_list",27,"name")="reg_ctl_list.gvntab.defer_base_release"
	Set gtmtypes("reg_ctl_list",27,"off")=164
	Set gtmtypes("reg_ctl_list",27,"len")=4
	Set gtmtypes("reg_ctl_list",27,"type")="boolean_t"
	Set gtmtypfldindx("reg_ctl_list","gvntab.defer_base_release")=27
	Set gtmtypes("reg_ctl_list",28,"name")="reg_ctl_list.gvntab.active"
	Set gtmtypes("reg_ctl_list",28,"off")=168
	Set gtmtypes("reg_ctl_list",28,"len")=4
	Set gtmtypes("reg_ctl_list",28,"type")="boolean_t"
	Set gtmtypfldindx("reg_ctl_list","gvntab.active")=28
	Set gtmtypes("reg_ctl_list",29,"name")="reg_ctl_list.lvrec_time"
	Set gtmtypes("reg_ctl_list",29,"off")=176
	Set gtmtypes("reg_ctl_list",29,"len")=4
	Set gtmtypes("reg_ctl_list",29,"type")="unsigned-int"
	Set gtmtypfldindx("reg_ctl_list","lvrec_time")=29
	Set gtmtypes("reg_ctl_list",30,"name")="reg_ctl_list.jnl_state"
	Set gtmtypes("reg_ctl_list",30,"off")=180
	Set gtmtypes("reg_ctl_list",30,"len")=4
	Set gtmtypes("reg_ctl_list",30,"type")="int"
	Set gtmtypfldindx("reg_ctl_list","jnl_state")=30
	Set gtmtypes("reg_ctl_list",31,"name")="reg_ctl_list.repl_state"
	Set gtmtypes("reg_ctl_list",31,"off")=184
	Set gtmtypes("reg_ctl_list",31,"len")=4
	Set gtmtypes("reg_ctl_list",31,"type")="int"
	Set gtmtypfldindx("reg_ctl_list","repl_state")=31
	Set gtmtypes("reg_ctl_list",32,"name")="reg_ctl_list.lookback_count"
	Set gtmtypes("reg_ctl_list",32,"off")=188
	Set gtmtypes("reg_ctl_list",32,"len")=4
	Set gtmtypes("reg_ctl_list",32,"type")="int"
	Set gtmtypfldindx("reg_ctl_list","lookback_count")=32
	Set gtmtypes("reg_ctl_list",33,"name")="reg_ctl_list.before_image"
	Set gtmtypes("reg_ctl_list",33,"off")=192
	Set gtmtypes("reg_ctl_list",33,"len")=4
	Set gtmtypes("reg_ctl_list",33,"type")="boolean_t"
	Set gtmtypfldindx("reg_ctl_list","before_image")=33
	Set gtmtypes("reg_ctl_list",34,"name")="reg_ctl_list.standalone"
	Set gtmtypes("reg_ctl_list",34,"off")=196
	Set gtmtypes("reg_ctl_list",34,"len")=4
	Set gtmtypes("reg_ctl_list",34,"type")="boolean_t"
	Set gtmtypfldindx("reg_ctl_list","standalone")=34
	Set gtmtypes("reg_ctl_list",35,"name")="reg_ctl_list.recov_interrupted"
	Set gtmtypes("reg_ctl_list",35,"off")=200
	Set gtmtypes("reg_ctl_list",35,"len")=4
	Set gtmtypes("reg_ctl_list",35,"type")="boolean_t"
	Set gtmtypfldindx("reg_ctl_list","recov_interrupted")=35
	Set gtmtypes("reg_ctl_list",36,"name")="reg_ctl_list.jfh_recov_interrupted"
	Set gtmtypes("reg_ctl_list",36,"off")=204
	Set gtmtypes("reg_ctl_list",36,"len")=4
	Set gtmtypes("reg_ctl_list",36,"type")="boolean_t"
	Set gtmtypfldindx("reg_ctl_list","jfh_recov_interrupted")=36
	Set gtmtypes("reg_ctl_list",37,"name")="reg_ctl_list.blks_to_upgrd_adjust"
	Set gtmtypes("reg_ctl_list",37,"off")=208
	Set gtmtypes("reg_ctl_list",37,"len")=4
	Set gtmtypes("reg_ctl_list",37,"type")="int"
	Set gtmtypfldindx("reg_ctl_list","blks_to_upgrd_adjust")=37
	Set gtmtypes("reg_ctl_list",38,"name")="reg_ctl_list.mur_plst"
	Set gtmtypes("reg_ctl_list",38,"off")=216
	Set gtmtypes("reg_ctl_list",38,"len")=8
	Set gtmtypes("reg_ctl_list",38,"type")="addr"
	Set gtmtypfldindx("reg_ctl_list","mur_plst")=38
	Set gtmtypes("reg_ctl_list",39,"name")="reg_ctl_list.mur_desc"
	Set gtmtypes("reg_ctl_list",39,"off")=224
	Set gtmtypes("reg_ctl_list",39,"len")=8
	Set gtmtypes("reg_ctl_list",39,"type")="addr"
	Set gtmtypfldindx("reg_ctl_list","mur_desc")=39
	Set gtmtypes("reg_ctl_list",40,"name")="reg_ctl_list.db_updated"
	Set gtmtypes("reg_ctl_list",40,"off")=232
	Set gtmtypes("reg_ctl_list",40,"len")=4
	Set gtmtypes("reg_ctl_list",40,"type")="boolean_t"
	Set gtmtypfldindx("reg_ctl_list","db_updated")=40
	Set gtmtypes("reg_ctl_list",41,"name")="reg_ctl_list.forw_eof_seen"
	Set gtmtypes("reg_ctl_list",41,"off")=236
	Set gtmtypes("reg_ctl_list",41,"len")=4
	Set gtmtypes("reg_ctl_list",41,"type")="boolean_t"
	Set gtmtypfldindx("reg_ctl_list","forw_eof_seen")=41
	Set gtmtypes("reg_ctl_list",42,"name")="reg_ctl_list.process_losttn"
	Set gtmtypes("reg_ctl_list",42,"off")=240
	Set gtmtypes("reg_ctl_list",42,"len")=4
	Set gtmtypes("reg_ctl_list",42,"type")="boolean_t"
	Set gtmtypfldindx("reg_ctl_list","process_losttn")=42
	Set gtmtypes("reg_ctl_list",43,"name")="reg_ctl_list.last_tn"
	Set gtmtypes("reg_ctl_list",43,"off")=248
	Set gtmtypes("reg_ctl_list",43,"len")=8
	Set gtmtypes("reg_ctl_list",43,"type")="uint64_t"
	Set gtmtypfldindx("reg_ctl_list","last_tn")=43
	Set gtmtypes("reg_ctl_list",44,"name")="reg_ctl_list.next_rctl"
	Set gtmtypes("reg_ctl_list",44,"off")=256
	Set gtmtypes("reg_ctl_list",44,"len")=8
	Set gtmtypes("reg_ctl_list",44,"type")="addr"
	Set gtmtypfldindx("reg_ctl_list","next_rctl")=44
	Set gtmtypes("reg_ctl_list",45,"name")="reg_ctl_list.prev_rctl"
	Set gtmtypes("reg_ctl_list",45,"off")=264
	Set gtmtypes("reg_ctl_list",45,"len")=8
	Set gtmtypes("reg_ctl_list",45,"type")="addr"
	Set gtmtypfldindx("reg_ctl_list","prev_rctl")=45
	Set gtmtypes("reg_ctl_list",46,"name")="reg_ctl_list.next_tp_rctl"
	Set gtmtypes("reg_ctl_list",46,"off")=272
	Set gtmtypes("reg_ctl_list",46,"len")=8
	Set gtmtypes("reg_ctl_list",46,"type")="addr"
	Set gtmtypfldindx("reg_ctl_list","next_tp_rctl")=46
	Set gtmtypes("reg_ctl_list",47,"name")="reg_ctl_list.prev_tp_rctl"
	Set gtmtypes("reg_ctl_list",47,"off")=280
	Set gtmtypes("reg_ctl_list",47,"len")=8
	Set gtmtypes("reg_ctl_list",47,"type")="addr"
	Set gtmtypfldindx("reg_ctl_list","prev_tp_rctl")=47
	Set gtmtypes("reg_ctl_list",48,"name")="reg_ctl_list.forw_multi"
	Set gtmtypes("reg_ctl_list",48,"off")=288
	Set gtmtypes("reg_ctl_list",48,"len")=8
	Set gtmtypes("reg_ctl_list",48,"type")="addr"
	Set gtmtypfldindx("reg_ctl_list","forw_multi")=48
	Set gtmtypes("reg_ctl_list",49,"name")="reg_ctl_list.initialized"
	Set gtmtypes("reg_ctl_list",49,"off")=296
	Set gtmtypes("reg_ctl_list",49,"len")=4
	Set gtmtypes("reg_ctl_list",49,"type")="boolean_t"
	Set gtmtypfldindx("reg_ctl_list","initialized")=49
	Set gtmtypes("reg_ctl_list",50,"name")="reg_ctl_list.db_present"
	Set gtmtypes("reg_ctl_list",50,"off")=300
	Set gtmtypes("reg_ctl_list",50,"len")=4
	Set gtmtypes("reg_ctl_list",50,"type")="boolean_t"
	Set gtmtypfldindx("reg_ctl_list","db_present")=50
	Set gtmtypes("reg_ctl_list",51,"name")="reg_ctl_list.this_pid_is_owner"
	Set gtmtypes("reg_ctl_list",51,"off")=304
	Set gtmtypes("reg_ctl_list",51,"len")=4
	Set gtmtypes("reg_ctl_list",51,"type")="boolean_t"
	Set gtmtypfldindx("reg_ctl_list","this_pid_is_owner")=51
	Set gtmtypes("reg_ctl_list",52,"name")="reg_ctl_list.file_info"
	Set gtmtypes("reg_ctl_list",52,"off")=312
	Set gtmtypes("reg_ctl_list",52,"len")=24
	Set gtmtypes("reg_ctl_list",52,"type")="addr"
	Set gtmtypfldindx("reg_ctl_list","file_info")=52
	Set gtmtypes("reg_ctl_list",52,"dim")=3
	Set gtmtypes("reg_ctl_list",53,"name")="reg_ctl_list.extr_file_created"
	Set gtmtypes("reg_ctl_list",53,"off")=336
	Set gtmtypes("reg_ctl_list",53,"len")=12
	Set gtmtypes("reg_ctl_list",53,"type")="boolean_t"
	Set gtmtypfldindx("reg_ctl_list","extr_file_created")=53
	Set gtmtypes("reg_ctl_list",53,"dim")=3
	Set gtmtypes("reg_ctl_list",54,"name")="reg_ctl_list.jnlext_multi_list_size"
	Set gtmtypes("reg_ctl_list",54,"off")=348
	Set gtmtypes("reg_ctl_list",54,"len")=12
	Set gtmtypes("reg_ctl_list",54,"type")="int"
	Set gtmtypfldindx("reg_ctl_list","jnlext_multi_list_size")=54
	Set gtmtypes("reg_ctl_list",54,"dim")=3
	Set gtmtypes("reg_ctl_list",55,"name")="reg_ctl_list.jnlext_multi_list"
	Set gtmtypes("reg_ctl_list",55,"off")=360
	Set gtmtypes("reg_ctl_list",55,"len")=24
	Set gtmtypes("reg_ctl_list",55,"type")="addr"
	Set gtmtypfldindx("reg_ctl_list","jnlext_multi_list")=55
	Set gtmtypes("reg_ctl_list",55,"dim")=3
	Set gtmtypes("reg_ctl_list",56,"name")="reg_ctl_list.last_jext_rec"
	Set gtmtypes("reg_ctl_list",56,"off")=384
	Set gtmtypes("reg_ctl_list",56,"len")=24
	Set gtmtypes("reg_ctl_list",56,"type")="addr"
	Set gtmtypfldindx("reg_ctl_list","last_jext_rec")=56
	Set gtmtypes("reg_ctl_list",56,"dim")=3
	Set gtmtypes("reg_ctl_list",57,"name")="reg_ctl_list.jnlext_shm_list"
	Set gtmtypes("reg_ctl_list",57,"off")=408
	Set gtmtypes("reg_ctl_list",57,"len")=24
	Set gtmtypes("reg_ctl_list",57,"type")="addr"
	Set gtmtypfldindx("reg_ctl_list","jnlext_shm_list")=57
	Set gtmtypes("reg_ctl_list",57,"dim")=3
	Set gtmtypes("reg_ctl_list",58,"name")="reg_ctl_list.extr_fn_len_orig"
	Set gtmtypes("reg_ctl_list",58,"off")=432
	Set gtmtypes("reg_ctl_list",58,"len")=12
	Set gtmtypes("reg_ctl_list",58,"type")="int"
	Set gtmtypfldindx("reg_ctl_list","extr_fn_len_orig")=58
	Set gtmtypes("reg_ctl_list",58,"dim")=3
	Set gtmtypes("reg_ctl_list",59,"name")="reg_ctl_list.last_jext_logical_rec"
	Set gtmtypes("reg_ctl_list",59,"off")=444
	Set gtmtypes("reg_ctl_list",59,"len")=12
	Set gtmtypes("reg_ctl_list",59,"type")="boolean_t"
	Set gtmtypfldindx("reg_ctl_list","last_jext_logical_rec")=59
	Set gtmtypes("reg_ctl_list",59,"dim")=3
	Set gtmtypes("reg_ctl_list",60,"name")="reg_ctl_list.dio_buff"
	Set gtmtypes("reg_ctl_list",60,"off")=456
	Set gtmtypes("reg_ctl_list",60,"len")=24
	Set gtmtypes("reg_ctl_list",60,"type")="dio_buff_t"
	Set gtmtypfldindx("reg_ctl_list","dio_buff")=60
	Set gtmtypes("reg_ctl_list",61,"name")="reg_ctl_list.dio_buff.unaligned_size"
	Set gtmtypes("reg_ctl_list",61,"off")=456
	Set gtmtypes("reg_ctl_list",61,"len")=4
	Set gtmtypes("reg_ctl_list",61,"type")="int"
	Set gtmtypfldindx("reg_ctl_list","dio_buff.unaligned_size")=61
	Set gtmtypes("reg_ctl_list",62,"name")="reg_ctl_list.dio_buff.unaligned"
	Set gtmtypes("reg_ctl_list",62,"off")=464
	Set gtmtypes("reg_ctl_list",62,"len")=8
	Set gtmtypes("reg_ctl_list",62,"type")="addr"
	Set gtmtypfldindx("reg_ctl_list","dio_buff.unaligned")=62
	Set gtmtypes("reg_ctl_list",63,"name")="reg_ctl_list.dio_buff.aligned"
	Set gtmtypes("reg_ctl_list",63,"off")=472
	Set gtmtypes("reg_ctl_list",63,"len")=8
	Set gtmtypes("reg_ctl_list",63,"type")="addr"
	Set gtmtypfldindx("reg_ctl_list","dio_buff.aligned")=63
	Set gtmtypes("reg_ctl_list",64,"name")="reg_ctl_list.region_index"
	Set gtmtypes("reg_ctl_list",64,"off")=480
	Set gtmtypes("reg_ctl_list",64,"len")=4
	Set gtmtypes("reg_ctl_list",64,"type")="int"
	Set gtmtypfldindx("reg_ctl_list","region_index")=64
	;
	Set gtmtypes("reg_jrec_info_t")="struct"
	Set gtmtypes("reg_jrec_info_t",0)=2
	Set gtmtypes("reg_jrec_info_t","len")=16
	Set gtmtypes("reg_jrec_info_t",1,"name")="reg_jrec_info_t.working_offset"
	Set gtmtypes("reg_jrec_info_t",1,"off")=0
	Set gtmtypes("reg_jrec_info_t",1,"len")=8
	Set gtmtypes("reg_jrec_info_t",1,"type")="long"
	Set gtmtypfldindx("reg_jrec_info_t","working_offset")=1
	Set gtmtypes("reg_jrec_info_t",2,"name")="reg_jrec_info_t.end"
	Set gtmtypes("reg_jrec_info_t",2,"off")=8
	Set gtmtypes("reg_jrec_info_t",2,"len")=8
	Set gtmtypes("reg_jrec_info_t",2,"type")="long"
	Set gtmtypfldindx("reg_jrec_info_t","end")=2
	;
	Set gtmtypes("relinkctl_data")="struct"
	Set gtmtypes("relinkctl_data",0)=11
	Set gtmtypes("relinkctl_data","len")=336
	Set gtmtypes("relinkctl_data",1,"name")="relinkctl_data.n_records"
	Set gtmtypes("relinkctl_data",1,"off")=0
	Set gtmtypes("relinkctl_data",1,"len")=4
	Set gtmtypes("relinkctl_data",1,"type")="unsigned-int"
	Set gtmtypfldindx("relinkctl_data","n_records")=1
	Set gtmtypes("relinkctl_data",2,"name")="relinkctl_data.nattached"
	Set gtmtypes("relinkctl_data",2,"off")=4
	Set gtmtypes("relinkctl_data",2,"len")=4
	Set gtmtypes("relinkctl_data",2,"type")="int"
	Set gtmtypfldindx("relinkctl_data","nattached")=2
	Set gtmtypes("relinkctl_data",3,"name")="relinkctl_data.relinkctl_shmid"
	Set gtmtypes("relinkctl_data",3,"off")=8
	Set gtmtypes("relinkctl_data",3,"len")=4
	Set gtmtypes("relinkctl_data",3,"type")="int"
	Set gtmtypfldindx("relinkctl_data","relinkctl_shmid")=3
	Set gtmtypes("relinkctl_data",4,"name")="relinkctl_data.relinkctl_shmlen"
	Set gtmtypes("relinkctl_data",4,"off")=12
	Set gtmtypes("relinkctl_data",4,"len")=4
	Set gtmtypes("relinkctl_data",4,"type")="unsigned-int"
	Set gtmtypfldindx("relinkctl_data","relinkctl_shmlen")=4
	Set gtmtypes("relinkctl_data",5,"name")="relinkctl_data.file_deleted"
	Set gtmtypes("relinkctl_data",5,"off")=16
	Set gtmtypes("relinkctl_data",5,"len")=4
	Set gtmtypes("relinkctl_data",5,"type")="int"
	Set gtmtypfldindx("relinkctl_data","file_deleted")=5
	Set gtmtypes("relinkctl_data",6,"name")="relinkctl_data.initialized"
	Set gtmtypes("relinkctl_data",6,"off")=20
	Set gtmtypes("relinkctl_data",6,"len")=4
	Set gtmtypes("relinkctl_data",6,"type")="unsigned-int"
	Set gtmtypfldindx("relinkctl_data","initialized")=6
	Set gtmtypes("relinkctl_data",7,"name")="relinkctl_data.zro_entry_name"
	Set gtmtypes("relinkctl_data",7,"off")=24
	Set gtmtypes("relinkctl_data",7,"len")=256
	Set gtmtypes("relinkctl_data",7,"type")="char"
	Set gtmtypfldindx("relinkctl_data","zro_entry_name")=7
	Set gtmtypes("relinkctl_data",8,"name")="relinkctl_data.zro_entry_name_len"
	Set gtmtypes("relinkctl_data",8,"off")=280
	Set gtmtypes("relinkctl_data",8,"len")=4
	Set gtmtypes("relinkctl_data",8,"type")="int"
	Set gtmtypfldindx("relinkctl_data","zro_entry_name_len")=8
	Set gtmtypes("relinkctl_data",9,"name")="relinkctl_data.relinkctl_max_rtn_entries"
	Set gtmtypes("relinkctl_data",9,"off")=284
	Set gtmtypes("relinkctl_data",9,"len")=4
	Set gtmtypes("relinkctl_data",9,"type")="int"
	Set gtmtypfldindx("relinkctl_data","relinkctl_max_rtn_entries")=9
	Set gtmtypes("relinkctl_data",10,"name")="relinkctl_data.relinkctl_hash_buckets"
	Set gtmtypes("relinkctl_data",10,"off")=288
	Set gtmtypes("relinkctl_data",10,"len")=4
	Set gtmtypes("relinkctl_data",10,"type")="int"
	Set gtmtypfldindx("relinkctl_data","relinkctl_hash_buckets")=10
	Set gtmtypes("relinkctl_data",11,"name")="relinkctl_data.exclu"
	Set gtmtypes("relinkctl_data",11,"off")=296
	Set gtmtypes("relinkctl_data",11,"len")=40
	Set gtmtypes("relinkctl_data",11,"type")="pthread_mutex_t"
	Set gtmtypfldindx("relinkctl_data","exclu")=11
	;
	Set gtmtypes("relinkrec_t")="struct"
	Set gtmtypes("relinkrec_t",0)=16
	Set gtmtypes("relinkrec_t","len")=88
	Set gtmtypes("relinkrec_t",1,"name")="relinkrec_t.rtnname_fixed"
	Set gtmtypes("relinkrec_t",1,"off")=0
	Set gtmtypes("relinkrec_t",1,"len")=32
	Set gtmtypes("relinkrec_t",1,"type")="mident_fixed"
	Set gtmtypfldindx("relinkrec_t","rtnname_fixed")=1
	Set gtmtypes("relinkrec_t",2,"name")="relinkrec_t.rtnname_fixed.c"
	Set gtmtypes("relinkrec_t",2,"off")=0
	Set gtmtypes("relinkrec_t",2,"len")=32
	Set gtmtypes("relinkrec_t",2,"type")="char"
	Set gtmtypfldindx("relinkrec_t","rtnname_fixed.c")=2
	Set gtmtypes("relinkrec_t",3,"name")="relinkrec_t.cycle"
	Set gtmtypes("relinkrec_t",3,"off")=32
	Set gtmtypes("relinkrec_t",3,"len")=4
	Set gtmtypes("relinkrec_t",3,"type")="unsigned-int"
	Set gtmtypfldindx("relinkrec_t","cycle")=3
	Set gtmtypes("relinkrec_t",4,"name")="relinkrec_t.hashindex_fl"
	Set gtmtypes("relinkrec_t",4,"off")=36
	Set gtmtypes("relinkrec_t",4,"len")=4
	Set gtmtypes("relinkrec_t",4,"type")="unsigned-int"
	Set gtmtypfldindx("relinkrec_t","hashindex_fl")=4
	Set gtmtypes("relinkrec_t",5,"name")="relinkrec_t.numvers"
	Set gtmtypes("relinkrec_t",5,"off")=40
	Set gtmtypes("relinkrec_t",5,"len")=4
	Set gtmtypes("relinkrec_t",5,"type")="unsigned-int"
	Set gtmtypfldindx("relinkrec_t","numvers")=5
	Set gtmtypes("relinkrec_t",6,"name")="relinkrec_t.filler_8byte_align"
	Set gtmtypes("relinkrec_t",6,"off")=44
	Set gtmtypes("relinkrec_t",6,"len")=4
	Set gtmtypes("relinkrec_t",6,"type")="unsigned-int"
	Set gtmtypfldindx("relinkrec_t","filler_8byte_align")=6
	Set gtmtypes("relinkrec_t",7,"name")="relinkrec_t.objLen"
	Set gtmtypes("relinkrec_t",7,"off")=48
	Set gtmtypes("relinkrec_t",7,"len")=8
	Set gtmtypes("relinkrec_t",7,"type")="uint64_t"
	Set gtmtypfldindx("relinkrec_t","objLen")=7
	Set gtmtypes("relinkrec_t",8,"name")="relinkrec_t.usedLen"
	Set gtmtypes("relinkrec_t",8,"off")=56
	Set gtmtypes("relinkrec_t",8,"len")=8
	Set gtmtypes("relinkrec_t",8,"type")="uint64_t"
	Set gtmtypfldindx("relinkrec_t","usedLen")=8
	Set gtmtypes("relinkrec_t",9,"name")="relinkrec_t.rtnobj_shm_offset"
	Set gtmtypes("relinkrec_t",9,"off")=64
	Set gtmtypes("relinkrec_t",9,"len")=8
	Set gtmtypes("relinkrec_t",9,"type")="uint64_t"
	Set gtmtypfldindx("relinkrec_t","rtnobj_shm_offset")=9
	Set gtmtypes("relinkrec_t",10,"name")="relinkrec_t.rtnobj_latch"
	Set gtmtypes("relinkrec_t",10,"off")=72
	Set gtmtypes("relinkrec_t",10,"len")=8
	Set gtmtypes("relinkrec_t",10,"type")="global_latch_t"
	Set gtmtypfldindx("relinkrec_t","rtnobj_latch")=10
	Set gtmtypes("relinkrec_t",11,"name")="relinkrec_t.rtnobj_latch.u"
	Set gtmtypes("relinkrec_t",11,"off")=72
	Set gtmtypes("relinkrec_t",11,"len")=8
	Set gtmtypes("relinkrec_t",11,"type")="union"
	Set gtmtypfldindx("relinkrec_t","rtnobj_latch.u")=11
	Set gtmtypes("relinkrec_t",12,"name")="relinkrec_t.rtnobj_latch.u.pid_imgcnt"
	Set gtmtypes("relinkrec_t",12,"off")=72
	Set gtmtypes("relinkrec_t",12,"len")=8
	Set gtmtypes("relinkrec_t",12,"type")="uint64_t"
	Set gtmtypfldindx("relinkrec_t","rtnobj_latch.u.pid_imgcnt")=12
	Set gtmtypes("relinkrec_t",13,"name")="relinkrec_t.rtnobj_latch.u.parts"
	Set gtmtypes("relinkrec_t",13,"off")=72
	Set gtmtypes("relinkrec_t",13,"len")=8
	Set gtmtypes("relinkrec_t",13,"type")="struct"
	Set gtmtypfldindx("relinkrec_t","rtnobj_latch.u.parts")=13
	Set gtmtypes("relinkrec_t",14,"name")="relinkrec_t.rtnobj_latch.u.parts.latch_pid"
	Set gtmtypes("relinkrec_t",14,"off")=72
	Set gtmtypes("relinkrec_t",14,"len")=4
	Set gtmtypes("relinkrec_t",14,"type")="int"
	Set gtmtypfldindx("relinkrec_t","rtnobj_latch.u.parts.latch_pid")=14
	Set gtmtypes("relinkrec_t",15,"name")="relinkrec_t.rtnobj_latch.u.parts.latch_word"
	Set gtmtypes("relinkrec_t",15,"off")=76
	Set gtmtypes("relinkrec_t",15,"len")=4
	Set gtmtypes("relinkrec_t",15,"type")="int"
	Set gtmtypfldindx("relinkrec_t","rtnobj_latch.u.parts.latch_word")=15
	Set gtmtypes("relinkrec_t",16,"name")="relinkrec_t.objhash"
	Set gtmtypes("relinkrec_t",16,"off")=80
	Set gtmtypes("relinkrec_t",16,"len")=8
	Set gtmtypes("relinkrec_t",16,"type")="uint64_t"
	Set gtmtypfldindx("relinkrec_t","objhash")=16
	;
	Set gtmtypes("relinkshm_hdr_t")="struct"
	Set gtmtypes("relinkshm_hdr_t",0)=21
	Set gtmtypes("relinkshm_hdr_t","len")=36712
	Set gtmtypes("relinkshm_hdr_t",1,"name")="relinkshm_hdr_t.relinkctl_fname"
	Set gtmtypes("relinkshm_hdr_t",1,"off")=0
	Set gtmtypes("relinkshm_hdr_t",1,"len")=4097
	Set gtmtypes("relinkshm_hdr_t",1,"type")="char"
	Set gtmtypfldindx("relinkshm_hdr_t","relinkctl_fname")=1
	Set gtmtypes("relinkshm_hdr_t",2,"name")="relinkshm_hdr_t.min_shm_index"
	Set gtmtypes("relinkshm_hdr_t",2,"off")=4100
	Set gtmtypes("relinkshm_hdr_t",2,"len")=4
	Set gtmtypes("relinkshm_hdr_t",2,"type")="int"
	Set gtmtypfldindx("relinkshm_hdr_t","min_shm_index")=2
	Set gtmtypes("relinkshm_hdr_t",3,"name")="relinkshm_hdr_t.rtnobj_min_shm_index"
	Set gtmtypes("relinkshm_hdr_t",3,"off")=4104
	Set gtmtypes("relinkshm_hdr_t",3,"len")=4
	Set gtmtypes("relinkshm_hdr_t",3,"type")="int"
	Set gtmtypfldindx("relinkshm_hdr_t","rtnobj_min_shm_index")=3
	Set gtmtypes("relinkshm_hdr_t",4,"name")="relinkshm_hdr_t.rtnobj_max_shm_index"
	Set gtmtypes("relinkshm_hdr_t",4,"off")=4108
	Set gtmtypes("relinkshm_hdr_t",4,"len")=4
	Set gtmtypes("relinkshm_hdr_t",4,"type")="int"
	Set gtmtypfldindx("relinkshm_hdr_t","rtnobj_max_shm_index")=4
	Set gtmtypes("relinkshm_hdr_t",5,"name")="relinkshm_hdr_t.rndwn_adjusted_nattch"
	Set gtmtypes("relinkshm_hdr_t",5,"off")=4112
	Set gtmtypes("relinkshm_hdr_t",5,"len")=4
	Set gtmtypes("relinkshm_hdr_t",5,"type")="boolean_t"
	Set gtmtypfldindx("relinkshm_hdr_t","rndwn_adjusted_nattch")=5
	Set gtmtypes("relinkshm_hdr_t",6,"name")="relinkshm_hdr_t.skip_rundown_check"
	Set gtmtypes("relinkshm_hdr_t",6,"off")=4116
	Set gtmtypes("relinkshm_hdr_t",6,"len")=4
	Set gtmtypes("relinkshm_hdr_t",6,"type")="boolean_t"
	Set gtmtypfldindx("relinkshm_hdr_t","skip_rundown_check")=6
	Set gtmtypes("relinkshm_hdr_t",7,"name")="relinkshm_hdr_t.rtnobj_shmhdr"
	Set gtmtypes("relinkshm_hdr_t",7,"off")=4120
	Set gtmtypes("relinkshm_hdr_t",7,"len")=32528
	Set gtmtypes("relinkshm_hdr_t",7,"type")="rtnobjshm_hdr_t"
	Set gtmtypfldindx("relinkshm_hdr_t","rtnobj_shmhdr")=7
	Set gtmtypes("relinkshm_hdr_t",7,"dim")=38
	Set gtmtypes("relinkshm_hdr_t",8,"name")="relinkshm_hdr_t.rtnobj_shmhdr[0].freeList"
	Set gtmtypes("relinkshm_hdr_t",8,"off")=4120
	Set gtmtypes("relinkshm_hdr_t",8,"len")=816
	Set gtmtypes("relinkshm_hdr_t",8,"type")="que_ent"
	Set gtmtypfldindx("relinkshm_hdr_t","rtnobj_shmhdr[0].freeList")=8
	Set gtmtypes("relinkshm_hdr_t",8,"dim")=51
	Set gtmtypes("relinkshm_hdr_t",9,"name")="relinkshm_hdr_t.rtnobj_shmhdr[0].rtnobj_min_free_index"
	Set gtmtypes("relinkshm_hdr_t",9,"off")=4936
	Set gtmtypes("relinkshm_hdr_t",9,"len")=4
	Set gtmtypes("relinkshm_hdr_t",9,"type")="int"
	Set gtmtypfldindx("relinkshm_hdr_t","rtnobj_shmhdr[0].rtnobj_min_free_index")=9
	Set gtmtypes("relinkshm_hdr_t",10,"name")="relinkshm_hdr_t.rtnobj_shmhdr[0].rtnobj_max_free_index"
	Set gtmtypes("relinkshm_hdr_t",10,"off")=4940
	Set gtmtypes("relinkshm_hdr_t",10,"len")=4
	Set gtmtypes("relinkshm_hdr_t",10,"type")="int"
	Set gtmtypfldindx("relinkshm_hdr_t","rtnobj_shmhdr[0].rtnobj_max_free_index")=10
	Set gtmtypes("relinkshm_hdr_t",11,"name")="relinkshm_hdr_t.rtnobj_shmhdr[0].rtnobj_shmid"
	Set gtmtypes("relinkshm_hdr_t",11,"off")=4944
	Set gtmtypes("relinkshm_hdr_t",11,"len")=4
	Set gtmtypes("relinkshm_hdr_t",11,"type")="int"
	Set gtmtypfldindx("relinkshm_hdr_t","rtnobj_shmhdr[0].rtnobj_shmid")=11
	Set gtmtypes("relinkshm_hdr_t",12,"name")="relinkshm_hdr_t.rtnobj_shmhdr[0].real_len"
	Set gtmtypes("relinkshm_hdr_t",12,"off")=4952
	Set gtmtypes("relinkshm_hdr_t",12,"len")=8
	Set gtmtypes("relinkshm_hdr_t",12,"type")="uint64_t"
	Set gtmtypfldindx("relinkshm_hdr_t","rtnobj_shmhdr[0].real_len")=12
	Set gtmtypes("relinkshm_hdr_t",13,"name")="relinkshm_hdr_t.rtnobj_shmhdr[0].used_len"
	Set gtmtypes("relinkshm_hdr_t",13,"off")=4960
	Set gtmtypes("relinkshm_hdr_t",13,"len")=8
	Set gtmtypes("relinkshm_hdr_t",13,"type")="uint64_t"
	Set gtmtypfldindx("relinkshm_hdr_t","rtnobj_shmhdr[0].used_len")=13
	Set gtmtypes("relinkshm_hdr_t",14,"name")="relinkshm_hdr_t.rtnobj_shmhdr[0].shm_len"
	Set gtmtypes("relinkshm_hdr_t",14,"off")=4968
	Set gtmtypes("relinkshm_hdr_t",14,"len")=8
	Set gtmtypes("relinkshm_hdr_t",14,"type")="uint64_t"
	Set gtmtypfldindx("relinkshm_hdr_t","rtnobj_shmhdr[0].shm_len")=14
	Set gtmtypes("relinkshm_hdr_t",15,"name")="relinkshm_hdr_t.relinkctl_latch"
	Set gtmtypes("relinkshm_hdr_t",15,"off")=36648
	Set gtmtypes("relinkshm_hdr_t",15,"len")=8
	Set gtmtypes("relinkshm_hdr_t",15,"type")="global_latch_t"
	Set gtmtypfldindx("relinkshm_hdr_t","relinkctl_latch")=15
	Set gtmtypes("relinkshm_hdr_t",16,"name")="relinkshm_hdr_t.relinkctl_latch.u"
	Set gtmtypes("relinkshm_hdr_t",16,"off")=36648
	Set gtmtypes("relinkshm_hdr_t",16,"len")=8
	Set gtmtypes("relinkshm_hdr_t",16,"type")="union"
	Set gtmtypfldindx("relinkshm_hdr_t","relinkctl_latch.u")=16
	Set gtmtypes("relinkshm_hdr_t",17,"name")="relinkshm_hdr_t.relinkctl_latch.u.pid_imgcnt"
	Set gtmtypes("relinkshm_hdr_t",17,"off")=36648
	Set gtmtypes("relinkshm_hdr_t",17,"len")=8
	Set gtmtypes("relinkshm_hdr_t",17,"type")="uint64_t"
	Set gtmtypfldindx("relinkshm_hdr_t","relinkctl_latch.u.pid_imgcnt")=17
	Set gtmtypes("relinkshm_hdr_t",18,"name")="relinkshm_hdr_t.relinkctl_latch.u.parts"
	Set gtmtypes("relinkshm_hdr_t",18,"off")=36648
	Set gtmtypes("relinkshm_hdr_t",18,"len")=8
	Set gtmtypes("relinkshm_hdr_t",18,"type")="struct"
	Set gtmtypfldindx("relinkshm_hdr_t","relinkctl_latch.u.parts")=18
	Set gtmtypes("relinkshm_hdr_t",19,"name")="relinkshm_hdr_t.relinkctl_latch.u.parts.latch_pid"
	Set gtmtypes("relinkshm_hdr_t",19,"off")=36648
	Set gtmtypes("relinkshm_hdr_t",19,"len")=4
	Set gtmtypes("relinkshm_hdr_t",19,"type")="int"
	Set gtmtypfldindx("relinkshm_hdr_t","relinkctl_latch.u.parts.latch_pid")=19
	Set gtmtypes("relinkshm_hdr_t",20,"name")="relinkshm_hdr_t.relinkctl_latch.u.parts.latch_word"
	Set gtmtypes("relinkshm_hdr_t",20,"off")=36652
	Set gtmtypes("relinkshm_hdr_t",20,"len")=4
	Set gtmtypes("relinkshm_hdr_t",20,"type")="int"
	Set gtmtypfldindx("relinkshm_hdr_t","relinkctl_latch.u.parts.latch_word")=20
	Set gtmtypes("relinkshm_hdr_t",21,"name")="relinkshm_hdr_t.fill_cacheline1"
	Set gtmtypes("relinkshm_hdr_t",21,"off")=36656
	Set gtmtypes("relinkshm_hdr_t",21,"len")=56
	Set gtmtypes("relinkshm_hdr_t",21,"type")="char"
	Set gtmtypfldindx("relinkshm_hdr_t","fill_cacheline1")=21
	;
	Set gtmtypes("repl_badtrans_msg_t")="struct"
	Set gtmtypes("repl_badtrans_msg_t",0)=4
	Set gtmtypes("repl_badtrans_msg_t","len")=32
	Set gtmtypes("repl_badtrans_msg_t",1,"name")="repl_badtrans_msg_t.type"
	Set gtmtypes("repl_badtrans_msg_t",1,"off")=0
	Set gtmtypes("repl_badtrans_msg_t",1,"len")=4
	Set gtmtypes("repl_badtrans_msg_t",1,"type")="int"
	Set gtmtypfldindx("repl_badtrans_msg_t","type")=1
	Set gtmtypes("repl_badtrans_msg_t",2,"name")="repl_badtrans_msg_t.len"
	Set gtmtypes("repl_badtrans_msg_t",2,"off")=4
	Set gtmtypes("repl_badtrans_msg_t",2,"len")=4
	Set gtmtypes("repl_badtrans_msg_t",2,"type")="int"
	Set gtmtypfldindx("repl_badtrans_msg_t","len")=2
	Set gtmtypes("repl_badtrans_msg_t",3,"name")="repl_badtrans_msg_t.start_seqno"
	Set gtmtypes("repl_badtrans_msg_t",3,"off")=8
	Set gtmtypes("repl_badtrans_msg_t",3,"len")=8
	Set gtmtypes("repl_badtrans_msg_t",3,"type")="uint64_t"
	Set gtmtypfldindx("repl_badtrans_msg_t","start_seqno")=3
	Set gtmtypes("repl_badtrans_msg_t",4,"name")="repl_badtrans_msg_t.filler_32"
	Set gtmtypes("repl_badtrans_msg_t",4,"off")=16
	Set gtmtypes("repl_badtrans_msg_t",4,"len")=16
	Set gtmtypes("repl_badtrans_msg_t",4,"type")="char"
	Set gtmtypfldindx("repl_badtrans_msg_t","filler_32")=4
	;
	Set gtmtypes("repl_buff_desc")="struct"
	Set gtmtypes("repl_buff_desc",0)=7
	Set gtmtypes("repl_buff_desc","len")=40
	Set gtmtypes("repl_buff_desc",1,"name")="repl_buff_desc.recbuff"
	Set gtmtypes("repl_buff_desc",1,"off")=0
	Set gtmtypes("repl_buff_desc",1,"len")=8
	Set gtmtypes("repl_buff_desc",1,"type")="addr"
	Set gtmtypfldindx("repl_buff_desc","recbuff")=1
	Set gtmtypes("repl_buff_desc",2,"name")="repl_buff_desc.reclen"
	Set gtmtypes("repl_buff_desc",2,"off")=8
	Set gtmtypes("repl_buff_desc",2,"len")=4
	Set gtmtypes("repl_buff_desc",2,"type")="int"
	Set gtmtypfldindx("repl_buff_desc","reclen")=2
	Set gtmtypes("repl_buff_desc",3,"name")="repl_buff_desc.recaddr"
	Set gtmtypes("repl_buff_desc",3,"off")=12
	Set gtmtypes("repl_buff_desc",3,"len")=4
	Set gtmtypes("repl_buff_desc",3,"type")="unsigned-int"
	Set gtmtypfldindx("repl_buff_desc","recaddr")=3
	Set gtmtypes("repl_buff_desc",4,"name")="repl_buff_desc.readaddr"
	Set gtmtypes("repl_buff_desc",4,"off")=16
	Set gtmtypes("repl_buff_desc",4,"len")=4
	Set gtmtypes("repl_buff_desc",4,"type")="unsigned-int"
	Set gtmtypfldindx("repl_buff_desc","readaddr")=4
	Set gtmtypes("repl_buff_desc",5,"name")="repl_buff_desc.buffremaining"
	Set gtmtypes("repl_buff_desc",5,"off")=20
	Set gtmtypes("repl_buff_desc",5,"len")=4
	Set gtmtypes("repl_buff_desc",5,"type")="unsigned-int"
	Set gtmtypfldindx("repl_buff_desc","buffremaining")=5
	Set gtmtypes("repl_buff_desc",6,"name")="repl_buff_desc.base_buff"
	Set gtmtypes("repl_buff_desc",6,"off")=24
	Set gtmtypes("repl_buff_desc",6,"len")=8
	Set gtmtypes("repl_buff_desc",6,"type")="addr"
	Set gtmtypfldindx("repl_buff_desc","base_buff")=6
	Set gtmtypes("repl_buff_desc",7,"name")="repl_buff_desc.base"
	Set gtmtypes("repl_buff_desc",7,"off")=32
	Set gtmtypes("repl_buff_desc",7,"len")=8
	Set gtmtypes("repl_buff_desc",7,"type")="addr"
	Set gtmtypfldindx("repl_buff_desc","base")=7
	;
	Set gtmtypes("repl_buff_t")="struct"
	Set gtmtypes("repl_buff_t",0)=11
	Set gtmtypes("repl_buff_t","len")=104
	Set gtmtypes("repl_buff_t",1,"name")="repl_buff_t.buffindex"
	Set gtmtypes("repl_buff_t",1,"off")=0
	Set gtmtypes("repl_buff_t",1,"len")=4
	Set gtmtypes("repl_buff_t",1,"type")="int"
	Set gtmtypfldindx("repl_buff_t","buffindex")=1
	Set gtmtypes("repl_buff_t",2,"name")="repl_buff_t.buff"
	Set gtmtypes("repl_buff_t",2,"off")=8
	Set gtmtypes("repl_buff_t",2,"len")=80
	Set gtmtypes("repl_buff_t",2,"type")="repl_buff_desc"
	Set gtmtypfldindx("repl_buff_t","buff")=2
	Set gtmtypes("repl_buff_t",2,"dim")=2
	Set gtmtypes("repl_buff_t",3,"name")="repl_buff_t.buff[0].recbuff"
	Set gtmtypes("repl_buff_t",3,"off")=8
	Set gtmtypes("repl_buff_t",3,"len")=8
	Set gtmtypes("repl_buff_t",3,"type")="addr"
	Set gtmtypfldindx("repl_buff_t","buff[0].recbuff")=3
	Set gtmtypes("repl_buff_t",4,"name")="repl_buff_t.buff[0].reclen"
	Set gtmtypes("repl_buff_t",4,"off")=16
	Set gtmtypes("repl_buff_t",4,"len")=4
	Set gtmtypes("repl_buff_t",4,"type")="int"
	Set gtmtypfldindx("repl_buff_t","buff[0].reclen")=4
	Set gtmtypes("repl_buff_t",5,"name")="repl_buff_t.buff[0].recaddr"
	Set gtmtypes("repl_buff_t",5,"off")=20
	Set gtmtypes("repl_buff_t",5,"len")=4
	Set gtmtypes("repl_buff_t",5,"type")="unsigned-int"
	Set gtmtypfldindx("repl_buff_t","buff[0].recaddr")=5
	Set gtmtypes("repl_buff_t",6,"name")="repl_buff_t.buff[0].readaddr"
	Set gtmtypes("repl_buff_t",6,"off")=24
	Set gtmtypes("repl_buff_t",6,"len")=4
	Set gtmtypes("repl_buff_t",6,"type")="unsigned-int"
	Set gtmtypfldindx("repl_buff_t","buff[0].readaddr")=6
	Set gtmtypes("repl_buff_t",7,"name")="repl_buff_t.buff[0].buffremaining"
	Set gtmtypes("repl_buff_t",7,"off")=28
	Set gtmtypes("repl_buff_t",7,"len")=4
	Set gtmtypes("repl_buff_t",7,"type")="unsigned-int"
	Set gtmtypfldindx("repl_buff_t","buff[0].buffremaining")=7
	Set gtmtypes("repl_buff_t",8,"name")="repl_buff_t.buff[0].base_buff"
	Set gtmtypes("repl_buff_t",8,"off")=32
	Set gtmtypes("repl_buff_t",8,"len")=8
	Set gtmtypes("repl_buff_t",8,"type")="addr"
	Set gtmtypfldindx("repl_buff_t","buff[0].base_buff")=8
	Set gtmtypes("repl_buff_t",9,"name")="repl_buff_t.buff[0].base"
	Set gtmtypes("repl_buff_t",9,"off")=40
	Set gtmtypes("repl_buff_t",9,"len")=8
	Set gtmtypes("repl_buff_t",9,"type")="addr"
	Set gtmtypfldindx("repl_buff_t","buff[0].base")=9
	Set gtmtypes("repl_buff_t",10,"name")="repl_buff_t.fc"
	Set gtmtypes("repl_buff_t",10,"off")=88
	Set gtmtypes("repl_buff_t",10,"len")=8
	Set gtmtypes("repl_buff_t",10,"type")="addr"
	Set gtmtypfldindx("repl_buff_t","fc")=10
	Set gtmtypes("repl_buff_t",11,"name")="repl_buff_t.backctl"
	Set gtmtypes("repl_buff_t",11,"off")=96
	Set gtmtypes("repl_buff_t",11,"len")=8
	Set gtmtypes("repl_buff_t",11,"type")="addr"
	Set gtmtypfldindx("repl_buff_t","backctl")=11
	;
	Set gtmtypes("repl_cmpinfo_msg_t")="struct"
	Set gtmtypes("repl_cmpinfo_msg_t",0)=7
	Set gtmtypes("repl_cmpinfo_msg_t","len")=528
	Set gtmtypes("repl_cmpinfo_msg_t",1,"name")="repl_cmpinfo_msg_t.type"
	Set gtmtypes("repl_cmpinfo_msg_t",1,"off")=0
	Set gtmtypes("repl_cmpinfo_msg_t",1,"len")=4
	Set gtmtypes("repl_cmpinfo_msg_t",1,"type")="int"
	Set gtmtypfldindx("repl_cmpinfo_msg_t","type")=1
	Set gtmtypes("repl_cmpinfo_msg_t",2,"name")="repl_cmpinfo_msg_t.len"
	Set gtmtypes("repl_cmpinfo_msg_t",2,"off")=4
	Set gtmtypes("repl_cmpinfo_msg_t",2,"len")=4
	Set gtmtypes("repl_cmpinfo_msg_t",2,"type")="int"
	Set gtmtypfldindx("repl_cmpinfo_msg_t","len")=2
	Set gtmtypes("repl_cmpinfo_msg_t",3,"name")="repl_cmpinfo_msg_t.datalen"
	Set gtmtypes("repl_cmpinfo_msg_t",3,"off")=8
	Set gtmtypes("repl_cmpinfo_msg_t",3,"len")=4
	Set gtmtypes("repl_cmpinfo_msg_t",3,"type")="int"
	Set gtmtypfldindx("repl_cmpinfo_msg_t","datalen")=3
	Set gtmtypes("repl_cmpinfo_msg_t",4,"name")="repl_cmpinfo_msg_t.proto_ver"
	Set gtmtypes("repl_cmpinfo_msg_t",4,"off")=12
	Set gtmtypes("repl_cmpinfo_msg_t",4,"len")=1
	Set gtmtypes("repl_cmpinfo_msg_t",4,"type")="char"
	Set gtmtypfldindx("repl_cmpinfo_msg_t","proto_ver")=4
	Set gtmtypes("repl_cmpinfo_msg_t",5,"name")="repl_cmpinfo_msg_t.filler_16"
	Set gtmtypes("repl_cmpinfo_msg_t",5,"off")=13
	Set gtmtypes("repl_cmpinfo_msg_t",5,"len")=3
	Set gtmtypes("repl_cmpinfo_msg_t",5,"type")="char"
	Set gtmtypfldindx("repl_cmpinfo_msg_t","filler_16")=5
	Set gtmtypes("repl_cmpinfo_msg_t",6,"name")="repl_cmpinfo_msg_t.data"
	Set gtmtypes("repl_cmpinfo_msg_t",6,"off")=16
	Set gtmtypes("repl_cmpinfo_msg_t",6,"len")=256
	Set gtmtypes("repl_cmpinfo_msg_t",6,"type")="char"
	Set gtmtypfldindx("repl_cmpinfo_msg_t","data")=6
	Set gtmtypes("repl_cmpinfo_msg_t",7,"name")="repl_cmpinfo_msg_t.overflowdata"
	Set gtmtypes("repl_cmpinfo_msg_t",7,"off")=272
	Set gtmtypes("repl_cmpinfo_msg_t",7,"len")=256
	Set gtmtypes("repl_cmpinfo_msg_t",7,"type")="char"
	Set gtmtypfldindx("repl_cmpinfo_msg_t","overflowdata")=7
	;
	Set gtmtypes("repl_cmpmsg_t")="struct"
	Set gtmtypes("repl_cmpmsg_t",0)=5
	Set gtmtypes("repl_cmpmsg_t","len")=32
	Set gtmtypes("repl_cmpmsg_t",1,"name")="repl_cmpmsg_t.type"
	Set gtmtypes("repl_cmpmsg_t",1,"off")=0
	Set gtmtypes("repl_cmpmsg_t",1,"len")=4
	Set gtmtypes("repl_cmpmsg_t",1,"type")="int"
	Set gtmtypfldindx("repl_cmpmsg_t","type")=1
	Set gtmtypes("repl_cmpmsg_t",2,"name")="repl_cmpmsg_t.len"
	Set gtmtypes("repl_cmpmsg_t",2,"off")=4
	Set gtmtypes("repl_cmpmsg_t",2,"len")=4
	Set gtmtypes("repl_cmpmsg_t",2,"type")="int"
	Set gtmtypfldindx("repl_cmpmsg_t","len")=2
	Set gtmtypes("repl_cmpmsg_t",3,"name")="repl_cmpmsg_t.uncmplen"
	Set gtmtypes("repl_cmpmsg_t",3,"off")=8
	Set gtmtypes("repl_cmpmsg_t",3,"len")=4
	Set gtmtypes("repl_cmpmsg_t",3,"type")="int"
	Set gtmtypfldindx("repl_cmpmsg_t","uncmplen")=3
	Set gtmtypes("repl_cmpmsg_t",4,"name")="repl_cmpmsg_t.cmplen"
	Set gtmtypes("repl_cmpmsg_t",4,"off")=12
	Set gtmtypes("repl_cmpmsg_t",4,"len")=4
	Set gtmtypes("repl_cmpmsg_t",4,"type")="int"
	Set gtmtypfldindx("repl_cmpmsg_t","cmplen")=4
	Set gtmtypes("repl_cmpmsg_t",5,"name")="repl_cmpmsg_t.msg"
	Set gtmtypes("repl_cmpmsg_t",5,"off")=16
	Set gtmtypes("repl_cmpmsg_t",5,"len")=16
	Set gtmtypes("repl_cmpmsg_t",5,"type")="unsigned-char"
	Set gtmtypfldindx("repl_cmpmsg_t","msg")=5
	;
	Set gtmtypes("repl_conn_info_t")="struct"
	Set gtmtypes("repl_conn_info_t",0)=10
	Set gtmtypes("repl_conn_info_t","len")=48
	Set gtmtypes("repl_conn_info_t",1,"name")="repl_conn_info_t.proto_ver"
	Set gtmtypes("repl_conn_info_t",1,"off")=0
	Set gtmtypes("repl_conn_info_t",1,"len")=4
	Set gtmtypes("repl_conn_info_t",1,"type")="int"
	Set gtmtypfldindx("repl_conn_info_t","proto_ver")=1
	Set gtmtypes("repl_conn_info_t",2,"name")="repl_conn_info_t.jnl_ver"
	Set gtmtypes("repl_conn_info_t",2,"off")=4
	Set gtmtypes("repl_conn_info_t",2,"len")=4
	Set gtmtypes("repl_conn_info_t",2,"type")="unsigned-int"
	Set gtmtypfldindx("repl_conn_info_t","jnl_ver")=2
	Set gtmtypes("repl_conn_info_t",3,"name")="repl_conn_info_t.is_std_null_coll"
	Set gtmtypes("repl_conn_info_t",3,"off")=8
	Set gtmtypes("repl_conn_info_t",3,"len")=4
	Set gtmtypes("repl_conn_info_t",3,"type")="boolean_t"
	Set gtmtypfldindx("repl_conn_info_t","is_std_null_coll")=3
	Set gtmtypes("repl_conn_info_t",4,"name")="repl_conn_info_t.trigger_supported"
	Set gtmtypes("repl_conn_info_t",4,"off")=12
	Set gtmtypes("repl_conn_info_t",4,"len")=4
	Set gtmtypes("repl_conn_info_t",4,"type")="boolean_t"
	Set gtmtypfldindx("repl_conn_info_t","trigger_supported")=4
	Set gtmtypes("repl_conn_info_t",5,"name")="repl_conn_info_t.cross_endian"
	Set gtmtypes("repl_conn_info_t",5,"off")=16
	Set gtmtypes("repl_conn_info_t",5,"len")=4
	Set gtmtypes("repl_conn_info_t",5,"type")="boolean_t"
	Set gtmtypfldindx("repl_conn_info_t","cross_endian")=5
	Set gtmtypes("repl_conn_info_t",6,"name")="repl_conn_info_t.endianness_known"
	Set gtmtypes("repl_conn_info_t",6,"off")=20
	Set gtmtypes("repl_conn_info_t",6,"len")=4
	Set gtmtypes("repl_conn_info_t",6,"type")="boolean_t"
	Set gtmtypfldindx("repl_conn_info_t","endianness_known")=6
	Set gtmtypes("repl_conn_info_t",7,"name")="repl_conn_info_t.null_subs_xform"
	Set gtmtypes("repl_conn_info_t",7,"off")=24
	Set gtmtypes("repl_conn_info_t",7,"len")=4
	Set gtmtypes("repl_conn_info_t",7,"type")="boolean_t"
	Set gtmtypfldindx("repl_conn_info_t","null_subs_xform")=7
	Set gtmtypes("repl_conn_info_t",8,"name")="repl_conn_info_t.is_supplementary"
	Set gtmtypes("repl_conn_info_t",8,"off")=28
	Set gtmtypes("repl_conn_info_t",8,"len")=4
	Set gtmtypes("repl_conn_info_t",8,"type")="boolean_t"
	Set gtmtypfldindx("repl_conn_info_t","is_supplementary")=8
	Set gtmtypes("repl_conn_info_t",9,"name")="repl_conn_info_t.tls_requested"
	Set gtmtypes("repl_conn_info_t",9,"off")=32
	Set gtmtypes("repl_conn_info_t",9,"len")=4
	Set gtmtypes("repl_conn_info_t",9,"type")="boolean_t"
	Set gtmtypfldindx("repl_conn_info_t","tls_requested")=9
	Set gtmtypes("repl_conn_info_t",10,"name")="repl_conn_info_t.filler_16"
	Set gtmtypes("repl_conn_info_t",10,"off")=36
	Set gtmtypes("repl_conn_info_t",10,"len")=12
	Set gtmtypes("repl_conn_info_t",10,"type")="char"
	Set gtmtypfldindx("repl_conn_info_t","filler_16")=10
	;
	Set gtmtypes("repl_ctl_element")="struct"
	Set gtmtypes("repl_ctl_element",0)=22
	Set gtmtypes("repl_ctl_element","len")=384
	Set gtmtypes("repl_ctl_element",1,"name")="repl_ctl_element.reg"
	Set gtmtypes("repl_ctl_element",1,"off")=0
	Set gtmtypes("repl_ctl_element",1,"len")=8
	Set gtmtypes("repl_ctl_element",1,"type")="addr"
	Set gtmtypfldindx("repl_ctl_element","reg")=1
	Set gtmtypes("repl_ctl_element",2,"name")="repl_ctl_element.repl_buff"
	Set gtmtypes("repl_ctl_element",2,"off")=8
	Set gtmtypes("repl_ctl_element",2,"len")=8
	Set gtmtypes("repl_ctl_element",2,"type")="addr"
	Set gtmtypfldindx("repl_ctl_element","repl_buff")=2
	Set gtmtypes("repl_ctl_element",3,"name")="repl_ctl_element.min_seqno"
	Set gtmtypes("repl_ctl_element",3,"off")=16
	Set gtmtypes("repl_ctl_element",3,"len")=8
	Set gtmtypes("repl_ctl_element",3,"type")="uint64_t"
	Set gtmtypfldindx("repl_ctl_element","min_seqno")=3
	Set gtmtypes("repl_ctl_element",4,"name")="repl_ctl_element.max_seqno"
	Set gtmtypes("repl_ctl_element",4,"off")=24
	Set gtmtypes("repl_ctl_element",4,"len")=8
	Set gtmtypes("repl_ctl_element",4,"type")="uint64_t"
	Set gtmtypfldindx("repl_ctl_element","max_seqno")=4
	Set gtmtypes("repl_ctl_element",5,"name")="repl_ctl_element.min_seqno_dskaddr"
	Set gtmtypes("repl_ctl_element",5,"off")=32
	Set gtmtypes("repl_ctl_element",5,"len")=4
	Set gtmtypes("repl_ctl_element",5,"type")="unsigned-int"
	Set gtmtypfldindx("repl_ctl_element","min_seqno_dskaddr")=5
	Set gtmtypes("repl_ctl_element",6,"name")="repl_ctl_element.max_seqno_dskaddr"
	Set gtmtypes("repl_ctl_element",6,"off")=36
	Set gtmtypes("repl_ctl_element",6,"len")=4
	Set gtmtypes("repl_ctl_element",6,"type")="unsigned-int"
	Set gtmtypfldindx("repl_ctl_element","max_seqno_dskaddr")=6
	Set gtmtypes("repl_ctl_element",7,"name")="repl_ctl_element.max_seqno_eof_addr"
	Set gtmtypes("repl_ctl_element",7,"off")=40
	Set gtmtypes("repl_ctl_element",7,"len")=4
	Set gtmtypes("repl_ctl_element",7,"type")="unsigned-int"
	Set gtmtypfldindx("repl_ctl_element","max_seqno_eof_addr")=7
	Set gtmtypes("repl_ctl_element",8,"name")="repl_ctl_element.filler_4byte"
	Set gtmtypes("repl_ctl_element",8,"off")=44
	Set gtmtypes("repl_ctl_element",8,"len")=4
	Set gtmtypes("repl_ctl_element",8,"type")="unsigned-int"
	Set gtmtypfldindx("repl_ctl_element","filler_4byte")=8
	Set gtmtypes("repl_ctl_element",9,"name")="repl_ctl_element.seqno"
	Set gtmtypes("repl_ctl_element",9,"off")=48
	Set gtmtypes("repl_ctl_element",9,"len")=8
	Set gtmtypes("repl_ctl_element",9,"type")="uint64_t"
	Set gtmtypfldindx("repl_ctl_element","seqno")=9
	Set gtmtypes("repl_ctl_element",10,"name")="repl_ctl_element.tn"
	Set gtmtypes("repl_ctl_element",10,"off")=56
	Set gtmtypes("repl_ctl_element",10,"len")=8
	Set gtmtypes("repl_ctl_element",10,"type")="uint64_t"
	Set gtmtypfldindx("repl_ctl_element","tn")=10
	Set gtmtypes("repl_ctl_element",11,"name")="repl_ctl_element.file_state"
	Set gtmtypes("repl_ctl_element",11,"off")=64
	Set gtmtypes("repl_ctl_element",11,"len")=4
	Set gtmtypes("repl_ctl_element",11,"type")="int"
	Set gtmtypfldindx("repl_ctl_element","file_state")=11
	Set gtmtypes("repl_ctl_element",12,"name")="repl_ctl_element.lookback"
	Set gtmtypes("repl_ctl_element",12,"off")=68
	Set gtmtypes("repl_ctl_element",12,"len")=4
	Set gtmtypes("repl_ctl_element",12,"type")="boolean_t"
	Set gtmtypfldindx("repl_ctl_element","lookback")=12
	Set gtmtypes("repl_ctl_element",13,"name")="repl_ctl_element.first_read_done"
	Set gtmtypes("repl_ctl_element",13,"off")=72
	Set gtmtypes("repl_ctl_element",13,"len")=4
	Set gtmtypes("repl_ctl_element",13,"type")="boolean_t"
	Set gtmtypfldindx("repl_ctl_element","first_read_done")=13
	Set gtmtypes("repl_ctl_element",14,"name")="repl_ctl_element.eof_addr_final"
	Set gtmtypes("repl_ctl_element",14,"off")=76
	Set gtmtypes("repl_ctl_element",14,"len")=4
	Set gtmtypes("repl_ctl_element",14,"type")="boolean_t"
	Set gtmtypfldindx("repl_ctl_element","eof_addr_final")=14
	Set gtmtypes("repl_ctl_element",15,"name")="repl_ctl_element.max_seqno_final"
	Set gtmtypes("repl_ctl_element",15,"off")=80
	Set gtmtypes("repl_ctl_element",15,"len")=4
	Set gtmtypes("repl_ctl_element",15,"type")="boolean_t"
	Set gtmtypfldindx("repl_ctl_element","max_seqno_final")=15
	Set gtmtypes("repl_ctl_element",16,"name")="repl_ctl_element.jnl_fn_len"
	Set gtmtypes("repl_ctl_element",16,"off")=84
	Set gtmtypes("repl_ctl_element",16,"len")=4
	Set gtmtypes("repl_ctl_element",16,"type")="int"
	Set gtmtypfldindx("repl_ctl_element","jnl_fn_len")=16
	Set gtmtypes("repl_ctl_element",17,"name")="repl_ctl_element.jnl_fn"
	Set gtmtypes("repl_ctl_element",17,"off")=88
	Set gtmtypes("repl_ctl_element",17,"len")=256
	Set gtmtypes("repl_ctl_element",17,"type")="char"
	Set gtmtypfldindx("repl_ctl_element","jnl_fn")=17
	Set gtmtypes("repl_ctl_element",18,"name")="repl_ctl_element.repl_rctl"
	Set gtmtypes("repl_ctl_element",18,"off")=344
	Set gtmtypes("repl_ctl_element",18,"len")=8
	Set gtmtypes("repl_ctl_element",18,"type")="addr"
	Set gtmtypfldindx("repl_ctl_element","repl_rctl")=18
	Set gtmtypes("repl_ctl_element",19,"name")="repl_ctl_element.prev"
	Set gtmtypes("repl_ctl_element",19,"off")=352
	Set gtmtypes("repl_ctl_element",19,"len")=8
	Set gtmtypes("repl_ctl_element",19,"type")="addr"
	Set gtmtypfldindx("repl_ctl_element","prev")=19
	Set gtmtypes("repl_ctl_element",20,"name")="repl_ctl_element.next"
	Set gtmtypes("repl_ctl_element",20,"off")=360
	Set gtmtypes("repl_ctl_element",20,"len")=8
	Set gtmtypes("repl_ctl_element",20,"type")="addr"
	Set gtmtypfldindx("repl_ctl_element","next")=20
	Set gtmtypes("repl_ctl_element",21,"name")="repl_ctl_element.encr_key_handle"
	Set gtmtypes("repl_ctl_element",21,"off")=368
	Set gtmtypes("repl_ctl_element",21,"len")=8
	Set gtmtypes("repl_ctl_element",21,"type")="addr"
	Set gtmtypfldindx("repl_ctl_element","encr_key_handle")=21
	Set gtmtypes("repl_ctl_element",22,"name")="repl_ctl_element.encr_key_handle2"
	Set gtmtypes("repl_ctl_element",22,"off")=376
	Set gtmtypes("repl_ctl_element",22,"len")=8
	Set gtmtypes("repl_ctl_element",22,"type")="addr"
	Set gtmtypfldindx("repl_ctl_element","encr_key_handle2")=22
	;
	Set gtmtypes("repl_file_control_t")="struct"
	Set gtmtypes("repl_file_control_t",0)=8
	Set gtmtypes("repl_file_control_t","len")=48
	Set gtmtypes("repl_file_control_t",1,"name")="repl_file_control_t.eof_addr"
	Set gtmtypes("repl_file_control_t",1,"off")=0
	Set gtmtypes("repl_file_control_t",1,"len")=4
	Set gtmtypes("repl_file_control_t",1,"type")="unsigned-int"
	Set gtmtypfldindx("repl_file_control_t","eof_addr")=1
	Set gtmtypes("repl_file_control_t",2,"name")="repl_file_control_t.fs_block_size"
	Set gtmtypes("repl_file_control_t",2,"off")=4
	Set gtmtypes("repl_file_control_t",2,"len")=4
	Set gtmtypes("repl_file_control_t",2,"type")="unsigned-int"
	Set gtmtypfldindx("repl_file_control_t","fs_block_size")=2
	Set gtmtypes("repl_file_control_t",3,"name")="repl_file_control_t.jfh_base"
	Set gtmtypes("repl_file_control_t",3,"off")=8
	Set gtmtypes("repl_file_control_t",3,"len")=8
	Set gtmtypes("repl_file_control_t",3,"type")="addr"
	Set gtmtypfldindx("repl_file_control_t","jfh_base")=3
	Set gtmtypes("repl_file_control_t",4,"name")="repl_file_control_t.jfh"
	Set gtmtypes("repl_file_control_t",4,"off")=16
	Set gtmtypes("repl_file_control_t",4,"len")=8
	Set gtmtypes("repl_file_control_t",4,"type")="addr"
	Set gtmtypfldindx("repl_file_control_t","jfh")=4
	Set gtmtypes("repl_file_control_t",5,"name")="repl_file_control_t.fd"
	Set gtmtypes("repl_file_control_t",5,"off")=24
	Set gtmtypes("repl_file_control_t",5,"len")=4
	Set gtmtypes("repl_file_control_t",5,"type")="int"
	Set gtmtypfldindx("repl_file_control_t","fd")=5
	Set gtmtypes("repl_file_control_t",6,"name")="repl_file_control_t.id"
	Set gtmtypes("repl_file_control_t",6,"off")=32
	Set gtmtypes("repl_file_control_t",6,"len")=16
	Set gtmtypes("repl_file_control_t",6,"type")="unix_file_id"
	Set gtmtypfldindx("repl_file_control_t","id")=6
	Set gtmtypes("repl_file_control_t",7,"name")="repl_file_control_t.id.inode"
	Set gtmtypes("repl_file_control_t",7,"off")=32
	Set gtmtypes("repl_file_control_t",7,"len")=8
	Set gtmtypes("repl_file_control_t",7,"type")="ino_t"
	Set gtmtypfldindx("repl_file_control_t","id.inode")=7
	Set gtmtypes("repl_file_control_t",8,"name")="repl_file_control_t.id.device"
	Set gtmtypes("repl_file_control_t",8,"off")=40
	Set gtmtypes("repl_file_control_t",8,"len")=8
	Set gtmtypes("repl_file_control_t",8,"type")="dev_t"
	Set gtmtypfldindx("repl_file_control_t","id.device")=8
	;
	Set gtmtypes("repl_heartbeat_msg_t")="struct"
	Set gtmtypes("repl_heartbeat_msg_t",0)=5
	Set gtmtypes("repl_heartbeat_msg_t","len")=32
	Set gtmtypes("repl_heartbeat_msg_t",1,"name")="repl_heartbeat_msg_t.type"
	Set gtmtypes("repl_heartbeat_msg_t",1,"off")=0
	Set gtmtypes("repl_heartbeat_msg_t",1,"len")=4
	Set gtmtypes("repl_heartbeat_msg_t",1,"type")="int"
	Set gtmtypfldindx("repl_heartbeat_msg_t","type")=1
	Set gtmtypes("repl_heartbeat_msg_t",2,"name")="repl_heartbeat_msg_t.len"
	Set gtmtypes("repl_heartbeat_msg_t",2,"off")=4
	Set gtmtypes("repl_heartbeat_msg_t",2,"len")=4
	Set gtmtypes("repl_heartbeat_msg_t",2,"type")="int"
	Set gtmtypfldindx("repl_heartbeat_msg_t","len")=2
	Set gtmtypes("repl_heartbeat_msg_t",3,"name")="repl_heartbeat_msg_t.ack_seqno"
	Set gtmtypes("repl_heartbeat_msg_t",3,"off")=8
	Set gtmtypes("repl_heartbeat_msg_t",3,"len")=8
	Set gtmtypes("repl_heartbeat_msg_t",3,"type")="unsigned-char"
	Set gtmtypfldindx("repl_heartbeat_msg_t","ack_seqno")=3
	Set gtmtypes("repl_heartbeat_msg_t",4,"name")="repl_heartbeat_msg_t.ack_time"
	Set gtmtypes("repl_heartbeat_msg_t",4,"off")=16
	Set gtmtypes("repl_heartbeat_msg_t",4,"len")=4
	Set gtmtypes("repl_heartbeat_msg_t",4,"type")="unsigned-char"
	Set gtmtypfldindx("repl_heartbeat_msg_t","ack_time")=4
	Set gtmtypes("repl_heartbeat_msg_t",5,"name")="repl_heartbeat_msg_t.filler_32"
	Set gtmtypes("repl_heartbeat_msg_t",5,"off")=20
	Set gtmtypes("repl_heartbeat_msg_t",5,"len")=12
	Set gtmtypes("repl_heartbeat_msg_t",5,"type")="unsigned-char"
	Set gtmtypfldindx("repl_heartbeat_msg_t","filler_32")=5
	;
	Set gtmtypes("repl_heartbeat_que_entry_t")="struct"
	Set gtmtypes("repl_heartbeat_que_entry_t",0)=9
	Set gtmtypes("repl_heartbeat_que_entry_t","len")=48
	Set gtmtypes("repl_heartbeat_que_entry_t",1,"name")="repl_heartbeat_que_entry_t.que"
	Set gtmtypes("repl_heartbeat_que_entry_t",1,"off")=0
	Set gtmtypes("repl_heartbeat_que_entry_t",1,"len")=16
	Set gtmtypes("repl_heartbeat_que_entry_t",1,"type")="struct"
	Set gtmtypfldindx("repl_heartbeat_que_entry_t","que")=1
	Set gtmtypes("repl_heartbeat_que_entry_t",2,"name")="repl_heartbeat_que_entry_t.que.fl"
	Set gtmtypes("repl_heartbeat_que_entry_t",2,"off")=0
	Set gtmtypes("repl_heartbeat_que_entry_t",2,"len")=8
	Set gtmtypes("repl_heartbeat_que_entry_t",2,"type")="intptr_t"
	Set gtmtypfldindx("repl_heartbeat_que_entry_t","que.fl")=2
	Set gtmtypes("repl_heartbeat_que_entry_t",3,"name")="repl_heartbeat_que_entry_t.que.bl"
	Set gtmtypes("repl_heartbeat_que_entry_t",3,"off")=8
	Set gtmtypes("repl_heartbeat_que_entry_t",3,"len")=8
	Set gtmtypes("repl_heartbeat_que_entry_t",3,"type")="intptr_t"
	Set gtmtypfldindx("repl_heartbeat_que_entry_t","que.bl")=3
	Set gtmtypes("repl_heartbeat_que_entry_t",4,"name")="repl_heartbeat_que_entry_t.heartbeat"
	Set gtmtypes("repl_heartbeat_que_entry_t",4,"off")=16
	Set gtmtypes("repl_heartbeat_que_entry_t",4,"len")=32
	Set gtmtypes("repl_heartbeat_que_entry_t",4,"type")="repl_heartbeat_msg_t"
	Set gtmtypfldindx("repl_heartbeat_que_entry_t","heartbeat")=4
	Set gtmtypes("repl_heartbeat_que_entry_t",5,"name")="repl_heartbeat_que_entry_t.heartbeat.type"
	Set gtmtypes("repl_heartbeat_que_entry_t",5,"off")=16
	Set gtmtypes("repl_heartbeat_que_entry_t",5,"len")=4
	Set gtmtypes("repl_heartbeat_que_entry_t",5,"type")="int"
	Set gtmtypfldindx("repl_heartbeat_que_entry_t","heartbeat.type")=5
	Set gtmtypes("repl_heartbeat_que_entry_t",6,"name")="repl_heartbeat_que_entry_t.heartbeat.len"
	Set gtmtypes("repl_heartbeat_que_entry_t",6,"off")=20
	Set gtmtypes("repl_heartbeat_que_entry_t",6,"len")=4
	Set gtmtypes("repl_heartbeat_que_entry_t",6,"type")="int"
	Set gtmtypfldindx("repl_heartbeat_que_entry_t","heartbeat.len")=6
	Set gtmtypes("repl_heartbeat_que_entry_t",7,"name")="repl_heartbeat_que_entry_t.heartbeat.ack_seqno"
	Set gtmtypes("repl_heartbeat_que_entry_t",7,"off")=24
	Set gtmtypes("repl_heartbeat_que_entry_t",7,"len")=8
	Set gtmtypes("repl_heartbeat_que_entry_t",7,"type")="unsigned-char"
	Set gtmtypfldindx("repl_heartbeat_que_entry_t","heartbeat.ack_seqno")=7
	Set gtmtypes("repl_heartbeat_que_entry_t",8,"name")="repl_heartbeat_que_entry_t.heartbeat.ack_time"
	Set gtmtypes("repl_heartbeat_que_entry_t",8,"off")=32
	Set gtmtypes("repl_heartbeat_que_entry_t",8,"len")=4
	Set gtmtypes("repl_heartbeat_que_entry_t",8,"type")="unsigned-char"
	Set gtmtypfldindx("repl_heartbeat_que_entry_t","heartbeat.ack_time")=8
	Set gtmtypes("repl_heartbeat_que_entry_t",9,"name")="repl_heartbeat_que_entry_t.heartbeat.filler_32"
	Set gtmtypes("repl_heartbeat_que_entry_t",9,"off")=36
	Set gtmtypes("repl_heartbeat_que_entry_t",9,"len")=12
	Set gtmtypes("repl_heartbeat_que_entry_t",9,"type")="unsigned-char"
	Set gtmtypfldindx("repl_heartbeat_que_entry_t","heartbeat.filler_32")=9
	;
	Set gtmtypes("repl_histinfo")="struct"
	Set gtmtypes("repl_histinfo",0)=17
	Set gtmtypes("repl_histinfo","len")=160
	Set gtmtypes("repl_histinfo",1,"name")="repl_histinfo.root_primary_instname"
	Set gtmtypes("repl_histinfo",1,"off")=0
	Set gtmtypes("repl_histinfo",1,"len")=16
	Set gtmtypes("repl_histinfo",1,"type")="unsigned-char"
	Set gtmtypfldindx("repl_histinfo","root_primary_instname")=1
	Set gtmtypes("repl_histinfo",2,"name")="repl_histinfo.start_seqno"
	Set gtmtypes("repl_histinfo",2,"off")=16
	Set gtmtypes("repl_histinfo",2,"len")=8
	Set gtmtypes("repl_histinfo",2,"type")="uint64_t"
	Set gtmtypfldindx("repl_histinfo","start_seqno")=2
	Set gtmtypes("repl_histinfo",3,"name")="repl_histinfo.strm_seqno"
	Set gtmtypes("repl_histinfo",3,"off")=24
	Set gtmtypes("repl_histinfo",3,"len")=8
	Set gtmtypes("repl_histinfo",3,"type")="uint64_t"
	Set gtmtypfldindx("repl_histinfo","strm_seqno")=3
	Set gtmtypes("repl_histinfo",4,"name")="repl_histinfo.root_primary_cycle"
	Set gtmtypes("repl_histinfo",4,"off")=32
	Set gtmtypes("repl_histinfo",4,"len")=4
	Set gtmtypes("repl_histinfo",4,"type")="unsigned-int"
	Set gtmtypfldindx("repl_histinfo","root_primary_cycle")=4
	Set gtmtypes("repl_histinfo",5,"name")="repl_histinfo.creator_pid"
	Set gtmtypes("repl_histinfo",5,"off")=36
	Set gtmtypes("repl_histinfo",5,"len")=4
	Set gtmtypes("repl_histinfo",5,"type")="unsigned-int"
	Set gtmtypfldindx("repl_histinfo","creator_pid")=5
	Set gtmtypes("repl_histinfo",6,"name")="repl_histinfo.created_time"
	Set gtmtypes("repl_histinfo",6,"off")=40
	Set gtmtypes("repl_histinfo",6,"len")=4
	Set gtmtypes("repl_histinfo",6,"type")="unsigned-int"
	Set gtmtypfldindx("repl_histinfo","created_time")=6
	Set gtmtypes("repl_histinfo",7,"name")="repl_histinfo.histinfo_num"
	Set gtmtypes("repl_histinfo",7,"off")=44
	Set gtmtypes("repl_histinfo",7,"len")=4
	Set gtmtypes("repl_histinfo",7,"type")="int"
	Set gtmtypfldindx("repl_histinfo","histinfo_num")=7
	Set gtmtypes("repl_histinfo",8,"name")="repl_histinfo.prev_histinfo_num"
	Set gtmtypes("repl_histinfo",8,"off")=48
	Set gtmtypes("repl_histinfo",8,"len")=4
	Set gtmtypes("repl_histinfo",8,"type")="int"
	Set gtmtypfldindx("repl_histinfo","prev_histinfo_num")=8
	Set gtmtypes("repl_histinfo",9,"name")="repl_histinfo.strm_index"
	Set gtmtypes("repl_histinfo",9,"off")=52
	Set gtmtypes("repl_histinfo",9,"len")=1
	Set gtmtypes("repl_histinfo",9,"type")="char"
	Set gtmtypfldindx("repl_histinfo","strm_index")=9
	Set gtmtypes("repl_histinfo",10,"name")="repl_histinfo.history_type"
	Set gtmtypes("repl_histinfo",10,"off")=53
	Set gtmtypes("repl_histinfo",10,"len")=1
	Set gtmtypes("repl_histinfo",10,"type")="char"
	Set gtmtypfldindx("repl_histinfo","history_type")=10
	Set gtmtypes("repl_histinfo",11,"name")="repl_histinfo.filler_8"
	Set gtmtypes("repl_histinfo",11,"off")=54
	Set gtmtypes("repl_histinfo",11,"len")=2
	Set gtmtypes("repl_histinfo",11,"type")="char"
	Set gtmtypfldindx("repl_histinfo","filler_8")=11
	Set gtmtypes("repl_histinfo",12,"name")="repl_histinfo.lms_group"
	Set gtmtypes("repl_histinfo",12,"off")=56
	Set gtmtypes("repl_histinfo",12,"len")=40
	Set gtmtypes("repl_histinfo",12,"type")="repl_inst_uuid"
	Set gtmtypfldindx("repl_histinfo","lms_group")=12
	Set gtmtypes("repl_histinfo",13,"name")="repl_histinfo.lms_group.created_nodename"
	Set gtmtypes("repl_histinfo",13,"off")=56
	Set gtmtypes("repl_histinfo",13,"len")=16
	Set gtmtypes("repl_histinfo",13,"type")="unsigned-char"
	Set gtmtypfldindx("repl_histinfo","lms_group.created_nodename")=13
	Set gtmtypes("repl_histinfo",14,"name")="repl_histinfo.lms_group.this_instname"
	Set gtmtypes("repl_histinfo",14,"off")=72
	Set gtmtypes("repl_histinfo",14,"len")=16
	Set gtmtypes("repl_histinfo",14,"type")="unsigned-char"
	Set gtmtypfldindx("repl_histinfo","lms_group.this_instname")=14
	Set gtmtypes("repl_histinfo",15,"name")="repl_histinfo.lms_group.created_time"
	Set gtmtypes("repl_histinfo",15,"off")=88
	Set gtmtypes("repl_histinfo",15,"len")=4
	Set gtmtypes("repl_histinfo",15,"type")="unsigned-int"
	Set gtmtypfldindx("repl_histinfo","lms_group.created_time")=15
	Set gtmtypes("repl_histinfo",16,"name")="repl_histinfo.lms_group.creator_pid"
	Set gtmtypes("repl_histinfo",16,"off")=92
	Set gtmtypes("repl_histinfo",16,"len")=4
	Set gtmtypes("repl_histinfo",16,"type")="unsigned-int"
	Set gtmtypfldindx("repl_histinfo","lms_group.creator_pid")=16
	Set gtmtypes("repl_histinfo",17,"name")="repl_histinfo.last_histinfo_num"
	Set gtmtypes("repl_histinfo",17,"off")=96
	Set gtmtypes("repl_histinfo",17,"len")=64
	Set gtmtypes("repl_histinfo",17,"type")="int"
	Set gtmtypfldindx("repl_histinfo","last_histinfo_num")=17
	Set gtmtypes("repl_histinfo",17,"dim")=16
	;
	Set gtmtypes("repl_histinfo1_msg_t")="struct"
	Set gtmtypes("repl_histinfo1_msg_t",0)=4
	Set gtmtypes("repl_histinfo1_msg_t","len")=32
	Set gtmtypes("repl_histinfo1_msg_t",1,"name")="repl_histinfo1_msg_t.type"
	Set gtmtypes("repl_histinfo1_msg_t",1,"off")=0
	Set gtmtypes("repl_histinfo1_msg_t",1,"len")=4
	Set gtmtypes("repl_histinfo1_msg_t",1,"type")="int"
	Set gtmtypfldindx("repl_histinfo1_msg_t","type")=1
	Set gtmtypes("repl_histinfo1_msg_t",2,"name")="repl_histinfo1_msg_t.len"
	Set gtmtypes("repl_histinfo1_msg_t",2,"off")=4
	Set gtmtypes("repl_histinfo1_msg_t",2,"len")=4
	Set gtmtypes("repl_histinfo1_msg_t",2,"type")="int"
	Set gtmtypfldindx("repl_histinfo1_msg_t","len")=2
	Set gtmtypes("repl_histinfo1_msg_t",3,"name")="repl_histinfo1_msg_t.start_seqno"
	Set gtmtypes("repl_histinfo1_msg_t",3,"off")=8
	Set gtmtypes("repl_histinfo1_msg_t",3,"len")=8
	Set gtmtypes("repl_histinfo1_msg_t",3,"type")="uint64_t"
	Set gtmtypfldindx("repl_histinfo1_msg_t","start_seqno")=3
	Set gtmtypes("repl_histinfo1_msg_t",4,"name")="repl_histinfo1_msg_t.instname"
	Set gtmtypes("repl_histinfo1_msg_t",4,"off")=16
	Set gtmtypes("repl_histinfo1_msg_t",4,"len")=16
	Set gtmtypes("repl_histinfo1_msg_t",4,"type")="unsigned-char"
	Set gtmtypfldindx("repl_histinfo1_msg_t","instname")=4
	;
	Set gtmtypes("repl_histinfo2_msg_t")="struct"
	Set gtmtypes("repl_histinfo2_msg_t",0)=6
	Set gtmtypes("repl_histinfo2_msg_t","len")=32
	Set gtmtypes("repl_histinfo2_msg_t",1,"name")="repl_histinfo2_msg_t.type"
	Set gtmtypes("repl_histinfo2_msg_t",1,"off")=0
	Set gtmtypes("repl_histinfo2_msg_t",1,"len")=4
	Set gtmtypes("repl_histinfo2_msg_t",1,"type")="int"
	Set gtmtypfldindx("repl_histinfo2_msg_t","type")=1
	Set gtmtypes("repl_histinfo2_msg_t",2,"name")="repl_histinfo2_msg_t.len"
	Set gtmtypes("repl_histinfo2_msg_t",2,"off")=4
	Set gtmtypes("repl_histinfo2_msg_t",2,"len")=4
	Set gtmtypes("repl_histinfo2_msg_t",2,"type")="int"
	Set gtmtypfldindx("repl_histinfo2_msg_t","len")=2
	Set gtmtypes("repl_histinfo2_msg_t",3,"name")="repl_histinfo2_msg_t.start_seqno"
	Set gtmtypes("repl_histinfo2_msg_t",3,"off")=8
	Set gtmtypes("repl_histinfo2_msg_t",3,"len")=8
	Set gtmtypes("repl_histinfo2_msg_t",3,"type")="uint64_t"
	Set gtmtypfldindx("repl_histinfo2_msg_t","start_seqno")=3
	Set gtmtypes("repl_histinfo2_msg_t",4,"name")="repl_histinfo2_msg_t.cycle"
	Set gtmtypes("repl_histinfo2_msg_t",4,"off")=16
	Set gtmtypes("repl_histinfo2_msg_t",4,"len")=4
	Set gtmtypes("repl_histinfo2_msg_t",4,"type")="unsigned-int"
	Set gtmtypfldindx("repl_histinfo2_msg_t","cycle")=4
	Set gtmtypes("repl_histinfo2_msg_t",5,"name")="repl_histinfo2_msg_t.histinfo_num"
	Set gtmtypes("repl_histinfo2_msg_t",5,"off")=20
	Set gtmtypes("repl_histinfo2_msg_t",5,"len")=4
	Set gtmtypes("repl_histinfo2_msg_t",5,"type")="int"
	Set gtmtypfldindx("repl_histinfo2_msg_t","histinfo_num")=5
	Set gtmtypes("repl_histinfo2_msg_t",6,"name")="repl_histinfo2_msg_t.filler_32"
	Set gtmtypes("repl_histinfo2_msg_t",6,"off")=24
	Set gtmtypes("repl_histinfo2_msg_t",6,"len")=8
	Set gtmtypes("repl_histinfo2_msg_t",6,"type")="char"
	Set gtmtypfldindx("repl_histinfo2_msg_t","filler_32")=6
	;
	Set gtmtypes("repl_histinfo_msg_t")="struct"
	Set gtmtypes("repl_histinfo_msg_t",0)=20
	Set gtmtypes("repl_histinfo_msg_t","len")=168
	Set gtmtypes("repl_histinfo_msg_t",1,"name")="repl_histinfo_msg_t.type"
	Set gtmtypes("repl_histinfo_msg_t",1,"off")=0
	Set gtmtypes("repl_histinfo_msg_t",1,"len")=4
	Set gtmtypes("repl_histinfo_msg_t",1,"type")="int"
	Set gtmtypfldindx("repl_histinfo_msg_t","type")=1
	Set gtmtypes("repl_histinfo_msg_t",2,"name")="repl_histinfo_msg_t.len"
	Set gtmtypes("repl_histinfo_msg_t",2,"off")=4
	Set gtmtypes("repl_histinfo_msg_t",2,"len")=4
	Set gtmtypes("repl_histinfo_msg_t",2,"type")="int"
	Set gtmtypfldindx("repl_histinfo_msg_t","len")=2
	Set gtmtypes("repl_histinfo_msg_t",3,"name")="repl_histinfo_msg_t.history"
	Set gtmtypes("repl_histinfo_msg_t",3,"off")=8
	Set gtmtypes("repl_histinfo_msg_t",3,"len")=160
	Set gtmtypes("repl_histinfo_msg_t",3,"type")="repl_histinfo"
	Set gtmtypfldindx("repl_histinfo_msg_t","history")=3
	Set gtmtypes("repl_histinfo_msg_t",4,"name")="repl_histinfo_msg_t.history.root_primary_instname"
	Set gtmtypes("repl_histinfo_msg_t",4,"off")=8
	Set gtmtypes("repl_histinfo_msg_t",4,"len")=16
	Set gtmtypes("repl_histinfo_msg_t",4,"type")="unsigned-char"
	Set gtmtypfldindx("repl_histinfo_msg_t","history.root_primary_instname")=4
	Set gtmtypes("repl_histinfo_msg_t",5,"name")="repl_histinfo_msg_t.history.start_seqno"
	Set gtmtypes("repl_histinfo_msg_t",5,"off")=24
	Set gtmtypes("repl_histinfo_msg_t",5,"len")=8
	Set gtmtypes("repl_histinfo_msg_t",5,"type")="uint64_t"
	Set gtmtypfldindx("repl_histinfo_msg_t","history.start_seqno")=5
	Set gtmtypes("repl_histinfo_msg_t",6,"name")="repl_histinfo_msg_t.history.strm_seqno"
	Set gtmtypes("repl_histinfo_msg_t",6,"off")=32
	Set gtmtypes("repl_histinfo_msg_t",6,"len")=8
	Set gtmtypes("repl_histinfo_msg_t",6,"type")="uint64_t"
	Set gtmtypfldindx("repl_histinfo_msg_t","history.strm_seqno")=6
	Set gtmtypes("repl_histinfo_msg_t",7,"name")="repl_histinfo_msg_t.history.root_primary_cycle"
	Set gtmtypes("repl_histinfo_msg_t",7,"off")=40
	Set gtmtypes("repl_histinfo_msg_t",7,"len")=4
	Set gtmtypes("repl_histinfo_msg_t",7,"type")="unsigned-int"
	Set gtmtypfldindx("repl_histinfo_msg_t","history.root_primary_cycle")=7
	Set gtmtypes("repl_histinfo_msg_t",8,"name")="repl_histinfo_msg_t.history.creator_pid"
	Set gtmtypes("repl_histinfo_msg_t",8,"off")=44
	Set gtmtypes("repl_histinfo_msg_t",8,"len")=4
	Set gtmtypes("repl_histinfo_msg_t",8,"type")="unsigned-int"
	Set gtmtypfldindx("repl_histinfo_msg_t","history.creator_pid")=8
	Set gtmtypes("repl_histinfo_msg_t",9,"name")="repl_histinfo_msg_t.history.created_time"
	Set gtmtypes("repl_histinfo_msg_t",9,"off")=48
	Set gtmtypes("repl_histinfo_msg_t",9,"len")=4
	Set gtmtypes("repl_histinfo_msg_t",9,"type")="unsigned-int"
	Set gtmtypfldindx("repl_histinfo_msg_t","history.created_time")=9
	Set gtmtypes("repl_histinfo_msg_t",10,"name")="repl_histinfo_msg_t.history.histinfo_num"
	Set gtmtypes("repl_histinfo_msg_t",10,"off")=52
	Set gtmtypes("repl_histinfo_msg_t",10,"len")=4
	Set gtmtypes("repl_histinfo_msg_t",10,"type")="int"
	Set gtmtypfldindx("repl_histinfo_msg_t","history.histinfo_num")=10
	Set gtmtypes("repl_histinfo_msg_t",11,"name")="repl_histinfo_msg_t.history.prev_histinfo_num"
	Set gtmtypes("repl_histinfo_msg_t",11,"off")=56
	Set gtmtypes("repl_histinfo_msg_t",11,"len")=4
	Set gtmtypes("repl_histinfo_msg_t",11,"type")="int"
	Set gtmtypfldindx("repl_histinfo_msg_t","history.prev_histinfo_num")=11
	Set gtmtypes("repl_histinfo_msg_t",12,"name")="repl_histinfo_msg_t.history.strm_index"
	Set gtmtypes("repl_histinfo_msg_t",12,"off")=60
	Set gtmtypes("repl_histinfo_msg_t",12,"len")=1
	Set gtmtypes("repl_histinfo_msg_t",12,"type")="char"
	Set gtmtypfldindx("repl_histinfo_msg_t","history.strm_index")=12
	Set gtmtypes("repl_histinfo_msg_t",13,"name")="repl_histinfo_msg_t.history.history_type"
	Set gtmtypes("repl_histinfo_msg_t",13,"off")=61
	Set gtmtypes("repl_histinfo_msg_t",13,"len")=1
	Set gtmtypes("repl_histinfo_msg_t",13,"type")="char"
	Set gtmtypfldindx("repl_histinfo_msg_t","history.history_type")=13
	Set gtmtypes("repl_histinfo_msg_t",14,"name")="repl_histinfo_msg_t.history.filler_8"
	Set gtmtypes("repl_histinfo_msg_t",14,"off")=62
	Set gtmtypes("repl_histinfo_msg_t",14,"len")=2
	Set gtmtypes("repl_histinfo_msg_t",14,"type")="char"
	Set gtmtypfldindx("repl_histinfo_msg_t","history.filler_8")=14
	Set gtmtypes("repl_histinfo_msg_t",15,"name")="repl_histinfo_msg_t.history.lms_group"
	Set gtmtypes("repl_histinfo_msg_t",15,"off")=64
	Set gtmtypes("repl_histinfo_msg_t",15,"len")=40
	Set gtmtypes("repl_histinfo_msg_t",15,"type")="repl_inst_uuid"
	Set gtmtypfldindx("repl_histinfo_msg_t","history.lms_group")=15
	Set gtmtypes("repl_histinfo_msg_t",16,"name")="repl_histinfo_msg_t.history.lms_group.created_nodename"
	Set gtmtypes("repl_histinfo_msg_t",16,"off")=64
	Set gtmtypes("repl_histinfo_msg_t",16,"len")=16
	Set gtmtypes("repl_histinfo_msg_t",16,"type")="unsigned-char"
	Set gtmtypfldindx("repl_histinfo_msg_t","history.lms_group.created_nodename")=16
	Set gtmtypes("repl_histinfo_msg_t",17,"name")="repl_histinfo_msg_t.history.lms_group.this_instname"
	Set gtmtypes("repl_histinfo_msg_t",17,"off")=80
	Set gtmtypes("repl_histinfo_msg_t",17,"len")=16
	Set gtmtypes("repl_histinfo_msg_t",17,"type")="unsigned-char"
	Set gtmtypfldindx("repl_histinfo_msg_t","history.lms_group.this_instname")=17
	Set gtmtypes("repl_histinfo_msg_t",18,"name")="repl_histinfo_msg_t.history.lms_group.created_time"
	Set gtmtypes("repl_histinfo_msg_t",18,"off")=96
	Set gtmtypes("repl_histinfo_msg_t",18,"len")=4
	Set gtmtypes("repl_histinfo_msg_t",18,"type")="unsigned-int"
	Set gtmtypfldindx("repl_histinfo_msg_t","history.lms_group.created_time")=18
	Set gtmtypes("repl_histinfo_msg_t",19,"name")="repl_histinfo_msg_t.history.lms_group.creator_pid"
	Set gtmtypes("repl_histinfo_msg_t",19,"off")=100
	Set gtmtypes("repl_histinfo_msg_t",19,"len")=4
	Set gtmtypes("repl_histinfo_msg_t",19,"type")="unsigned-int"
	Set gtmtypfldindx("repl_histinfo_msg_t","history.lms_group.creator_pid")=19
	Set gtmtypes("repl_histinfo_msg_t",20,"name")="repl_histinfo_msg_t.history.last_histinfo_num"
	Set gtmtypes("repl_histinfo_msg_t",20,"off")=104
	Set gtmtypes("repl_histinfo_msg_t",20,"len")=64
	Set gtmtypes("repl_histinfo_msg_t",20,"type")="int"
	Set gtmtypfldindx("repl_histinfo_msg_t","history.last_histinfo_num")=20
	Set gtmtypes("repl_histinfo_msg_t",20,"dim")=16
	;
	Set gtmtypes("repl_histrec_jnl_t")="struct"
	Set gtmtypes("repl_histrec_jnl_t",0)=19
	Set gtmtypes("repl_histrec_jnl_t","len")=168
	Set gtmtypes("repl_histrec_jnl_t",1,"name")="repl_histrec_jnl_t.filler_8byte_align"
	Set gtmtypes("repl_histrec_jnl_t",1,"off")=4
	Set gtmtypes("repl_histrec_jnl_t",1,"len")=4
	Set gtmtypes("repl_histrec_jnl_t",1,"type")="unsigned-int"
	Set gtmtypfldindx("repl_histrec_jnl_t","filler_8byte_align")=1
	Set gtmtypes("repl_histrec_jnl_t",2,"name")="repl_histrec_jnl_t.histcontent"
	Set gtmtypes("repl_histrec_jnl_t",2,"off")=8
	Set gtmtypes("repl_histrec_jnl_t",2,"len")=160
	Set gtmtypes("repl_histrec_jnl_t",2,"type")="repl_histinfo"
	Set gtmtypfldindx("repl_histrec_jnl_t","histcontent")=2
	Set gtmtypes("repl_histrec_jnl_t",3,"name")="repl_histrec_jnl_t.histcontent.root_primary_instname"
	Set gtmtypes("repl_histrec_jnl_t",3,"off")=8
	Set gtmtypes("repl_histrec_jnl_t",3,"len")=16
	Set gtmtypes("repl_histrec_jnl_t",3,"type")="unsigned-char"
	Set gtmtypfldindx("repl_histrec_jnl_t","histcontent.root_primary_instname")=3
	Set gtmtypes("repl_histrec_jnl_t",4,"name")="repl_histrec_jnl_t.histcontent.start_seqno"
	Set gtmtypes("repl_histrec_jnl_t",4,"off")=24
	Set gtmtypes("repl_histrec_jnl_t",4,"len")=8
	Set gtmtypes("repl_histrec_jnl_t",4,"type")="uint64_t"
	Set gtmtypfldindx("repl_histrec_jnl_t","histcontent.start_seqno")=4
	Set gtmtypes("repl_histrec_jnl_t",5,"name")="repl_histrec_jnl_t.histcontent.strm_seqno"
	Set gtmtypes("repl_histrec_jnl_t",5,"off")=32
	Set gtmtypes("repl_histrec_jnl_t",5,"len")=8
	Set gtmtypes("repl_histrec_jnl_t",5,"type")="uint64_t"
	Set gtmtypfldindx("repl_histrec_jnl_t","histcontent.strm_seqno")=5
	Set gtmtypes("repl_histrec_jnl_t",6,"name")="repl_histrec_jnl_t.histcontent.root_primary_cycle"
	Set gtmtypes("repl_histrec_jnl_t",6,"off")=40
	Set gtmtypes("repl_histrec_jnl_t",6,"len")=4
	Set gtmtypes("repl_histrec_jnl_t",6,"type")="unsigned-int"
	Set gtmtypfldindx("repl_histrec_jnl_t","histcontent.root_primary_cycle")=6
	Set gtmtypes("repl_histrec_jnl_t",7,"name")="repl_histrec_jnl_t.histcontent.creator_pid"
	Set gtmtypes("repl_histrec_jnl_t",7,"off")=44
	Set gtmtypes("repl_histrec_jnl_t",7,"len")=4
	Set gtmtypes("repl_histrec_jnl_t",7,"type")="unsigned-int"
	Set gtmtypfldindx("repl_histrec_jnl_t","histcontent.creator_pid")=7
	Set gtmtypes("repl_histrec_jnl_t",8,"name")="repl_histrec_jnl_t.histcontent.created_time"
	Set gtmtypes("repl_histrec_jnl_t",8,"off")=48
	Set gtmtypes("repl_histrec_jnl_t",8,"len")=4
	Set gtmtypes("repl_histrec_jnl_t",8,"type")="unsigned-int"
	Set gtmtypfldindx("repl_histrec_jnl_t","histcontent.created_time")=8
	Set gtmtypes("repl_histrec_jnl_t",9,"name")="repl_histrec_jnl_t.histcontent.histinfo_num"
	Set gtmtypes("repl_histrec_jnl_t",9,"off")=52
	Set gtmtypes("repl_histrec_jnl_t",9,"len")=4
	Set gtmtypes("repl_histrec_jnl_t",9,"type")="int"
	Set gtmtypfldindx("repl_histrec_jnl_t","histcontent.histinfo_num")=9
	Set gtmtypes("repl_histrec_jnl_t",10,"name")="repl_histrec_jnl_t.histcontent.prev_histinfo_num"
	Set gtmtypes("repl_histrec_jnl_t",10,"off")=56
	Set gtmtypes("repl_histrec_jnl_t",10,"len")=4
	Set gtmtypes("repl_histrec_jnl_t",10,"type")="int"
	Set gtmtypfldindx("repl_histrec_jnl_t","histcontent.prev_histinfo_num")=10
	Set gtmtypes("repl_histrec_jnl_t",11,"name")="repl_histrec_jnl_t.histcontent.strm_index"
	Set gtmtypes("repl_histrec_jnl_t",11,"off")=60
	Set gtmtypes("repl_histrec_jnl_t",11,"len")=1
	Set gtmtypes("repl_histrec_jnl_t",11,"type")="char"
	Set gtmtypfldindx("repl_histrec_jnl_t","histcontent.strm_index")=11
	Set gtmtypes("repl_histrec_jnl_t",12,"name")="repl_histrec_jnl_t.histcontent.history_type"
	Set gtmtypes("repl_histrec_jnl_t",12,"off")=61
	Set gtmtypes("repl_histrec_jnl_t",12,"len")=1
	Set gtmtypes("repl_histrec_jnl_t",12,"type")="char"
	Set gtmtypfldindx("repl_histrec_jnl_t","histcontent.history_type")=12
	Set gtmtypes("repl_histrec_jnl_t",13,"name")="repl_histrec_jnl_t.histcontent.filler_8"
	Set gtmtypes("repl_histrec_jnl_t",13,"off")=62
	Set gtmtypes("repl_histrec_jnl_t",13,"len")=2
	Set gtmtypes("repl_histrec_jnl_t",13,"type")="char"
	Set gtmtypfldindx("repl_histrec_jnl_t","histcontent.filler_8")=13
	Set gtmtypes("repl_histrec_jnl_t",14,"name")="repl_histrec_jnl_t.histcontent.lms_group"
	Set gtmtypes("repl_histrec_jnl_t",14,"off")=64
	Set gtmtypes("repl_histrec_jnl_t",14,"len")=40
	Set gtmtypes("repl_histrec_jnl_t",14,"type")="repl_inst_uuid"
	Set gtmtypfldindx("repl_histrec_jnl_t","histcontent.lms_group")=14
	Set gtmtypes("repl_histrec_jnl_t",15,"name")="repl_histrec_jnl_t.histcontent.lms_group.created_nodename"
	Set gtmtypes("repl_histrec_jnl_t",15,"off")=64
	Set gtmtypes("repl_histrec_jnl_t",15,"len")=16
	Set gtmtypes("repl_histrec_jnl_t",15,"type")="unsigned-char"
	Set gtmtypfldindx("repl_histrec_jnl_t","histcontent.lms_group.created_nodename")=15
	Set gtmtypes("repl_histrec_jnl_t",16,"name")="repl_histrec_jnl_t.histcontent.lms_group.this_instname"
	Set gtmtypes("repl_histrec_jnl_t",16,"off")=80
	Set gtmtypes("repl_histrec_jnl_t",16,"len")=16
	Set gtmtypes("repl_histrec_jnl_t",16,"type")="unsigned-char"
	Set gtmtypfldindx("repl_histrec_jnl_t","histcontent.lms_group.this_instname")=16
	Set gtmtypes("repl_histrec_jnl_t",17,"name")="repl_histrec_jnl_t.histcontent.lms_group.created_time"
	Set gtmtypes("repl_histrec_jnl_t",17,"off")=96
	Set gtmtypes("repl_histrec_jnl_t",17,"len")=4
	Set gtmtypes("repl_histrec_jnl_t",17,"type")="unsigned-int"
	Set gtmtypfldindx("repl_histrec_jnl_t","histcontent.lms_group.created_time")=17
	Set gtmtypes("repl_histrec_jnl_t",18,"name")="repl_histrec_jnl_t.histcontent.lms_group.creator_pid"
	Set gtmtypes("repl_histrec_jnl_t",18,"off")=100
	Set gtmtypes("repl_histrec_jnl_t",18,"len")=4
	Set gtmtypes("repl_histrec_jnl_t",18,"type")="unsigned-int"
	Set gtmtypfldindx("repl_histrec_jnl_t","histcontent.lms_group.creator_pid")=18
	Set gtmtypes("repl_histrec_jnl_t",19,"name")="repl_histrec_jnl_t.histcontent.last_histinfo_num"
	Set gtmtypes("repl_histrec_jnl_t",19,"off")=104
	Set gtmtypes("repl_histrec_jnl_t",19,"len")=64
	Set gtmtypes("repl_histrec_jnl_t",19,"type")="int"
	Set gtmtypfldindx("repl_histrec_jnl_t","histcontent.last_histinfo_num")=19
	Set gtmtypes("repl_histrec_jnl_t",19,"dim")=16
	;
	Set gtmtypes("repl_histrec_msg_t")="struct"
	Set gtmtypes("repl_histrec_msg_t",0)=22
	Set gtmtypes("repl_histrec_msg_t","len")=176
	Set gtmtypes("repl_histrec_msg_t",1,"name")="repl_histrec_msg_t.type"
	Set gtmtypes("repl_histrec_msg_t",1,"off")=0
	Set gtmtypes("repl_histrec_msg_t",1,"len")=4
	Set gtmtypes("repl_histrec_msg_t",1,"type")="int"
	Set gtmtypfldindx("repl_histrec_msg_t","type")=1
	Set gtmtypes("repl_histrec_msg_t",2,"name")="repl_histrec_msg_t.len"
	Set gtmtypes("repl_histrec_msg_t",2,"off")=4
	Set gtmtypes("repl_histrec_msg_t",2,"len")=4
	Set gtmtypes("repl_histrec_msg_t",2,"type")="int"
	Set gtmtypfldindx("repl_histrec_msg_t","len")=2
	Set gtmtypes("repl_histrec_msg_t",3,"name")="repl_histrec_msg_t.histjrec"
	Set gtmtypes("repl_histrec_msg_t",3,"off")=8
	Set gtmtypes("repl_histrec_msg_t",3,"len")=168
	Set gtmtypes("repl_histrec_msg_t",3,"type")="repl_histrec_jnl_t"
	Set gtmtypfldindx("repl_histrec_msg_t","histjrec")=3
	Set gtmtypes("repl_histrec_msg_t",4,"name")="repl_histrec_msg_t.histjrec.filler_8byte_align"
	Set gtmtypes("repl_histrec_msg_t",4,"off")=12
	Set gtmtypes("repl_histrec_msg_t",4,"len")=4
	Set gtmtypes("repl_histrec_msg_t",4,"type")="unsigned-int"
	Set gtmtypfldindx("repl_histrec_msg_t","histjrec.filler_8byte_align")=4
	Set gtmtypes("repl_histrec_msg_t",5,"name")="repl_histrec_msg_t.histjrec.histcontent"
	Set gtmtypes("repl_histrec_msg_t",5,"off")=16
	Set gtmtypes("repl_histrec_msg_t",5,"len")=160
	Set gtmtypes("repl_histrec_msg_t",5,"type")="repl_histinfo"
	Set gtmtypfldindx("repl_histrec_msg_t","histjrec.histcontent")=5
	Set gtmtypes("repl_histrec_msg_t",6,"name")="repl_histrec_msg_t.histjrec.histcontent.root_primary_instname"
	Set gtmtypes("repl_histrec_msg_t",6,"off")=16
	Set gtmtypes("repl_histrec_msg_t",6,"len")=16
	Set gtmtypes("repl_histrec_msg_t",6,"type")="unsigned-char"
	Set gtmtypfldindx("repl_histrec_msg_t","histjrec.histcontent.root_primary_instname")=6
	Set gtmtypes("repl_histrec_msg_t",7,"name")="repl_histrec_msg_t.histjrec.histcontent.start_seqno"
	Set gtmtypes("repl_histrec_msg_t",7,"off")=32
	Set gtmtypes("repl_histrec_msg_t",7,"len")=8
	Set gtmtypes("repl_histrec_msg_t",7,"type")="uint64_t"
	Set gtmtypfldindx("repl_histrec_msg_t","histjrec.histcontent.start_seqno")=7
	Set gtmtypes("repl_histrec_msg_t",8,"name")="repl_histrec_msg_t.histjrec.histcontent.strm_seqno"
	Set gtmtypes("repl_histrec_msg_t",8,"off")=40
	Set gtmtypes("repl_histrec_msg_t",8,"len")=8
	Set gtmtypes("repl_histrec_msg_t",8,"type")="uint64_t"
	Set gtmtypfldindx("repl_histrec_msg_t","histjrec.histcontent.strm_seqno")=8
	Set gtmtypes("repl_histrec_msg_t",9,"name")="repl_histrec_msg_t.histjrec.histcontent.root_primary_cycle"
	Set gtmtypes("repl_histrec_msg_t",9,"off")=48
	Set gtmtypes("repl_histrec_msg_t",9,"len")=4
	Set gtmtypes("repl_histrec_msg_t",9,"type")="unsigned-int"
	Set gtmtypfldindx("repl_histrec_msg_t","histjrec.histcontent.root_primary_cycle")=9
	Set gtmtypes("repl_histrec_msg_t",10,"name")="repl_histrec_msg_t.histjrec.histcontent.creator_pid"
	Set gtmtypes("repl_histrec_msg_t",10,"off")=52
	Set gtmtypes("repl_histrec_msg_t",10,"len")=4
	Set gtmtypes("repl_histrec_msg_t",10,"type")="unsigned-int"
	Set gtmtypfldindx("repl_histrec_msg_t","histjrec.histcontent.creator_pid")=10
	Set gtmtypes("repl_histrec_msg_t",11,"name")="repl_histrec_msg_t.histjrec.histcontent.created_time"
	Set gtmtypes("repl_histrec_msg_t",11,"off")=56
	Set gtmtypes("repl_histrec_msg_t",11,"len")=4
	Set gtmtypes("repl_histrec_msg_t",11,"type")="unsigned-int"
	Set gtmtypfldindx("repl_histrec_msg_t","histjrec.histcontent.created_time")=11
	Set gtmtypes("repl_histrec_msg_t",12,"name")="repl_histrec_msg_t.histjrec.histcontent.histinfo_num"
	Set gtmtypes("repl_histrec_msg_t",12,"off")=60
	Set gtmtypes("repl_histrec_msg_t",12,"len")=4
	Set gtmtypes("repl_histrec_msg_t",12,"type")="int"
	Set gtmtypfldindx("repl_histrec_msg_t","histjrec.histcontent.histinfo_num")=12
	Set gtmtypes("repl_histrec_msg_t",13,"name")="repl_histrec_msg_t.histjrec.histcontent.prev_histinfo_num"
	Set gtmtypes("repl_histrec_msg_t",13,"off")=64
	Set gtmtypes("repl_histrec_msg_t",13,"len")=4
	Set gtmtypes("repl_histrec_msg_t",13,"type")="int"
	Set gtmtypfldindx("repl_histrec_msg_t","histjrec.histcontent.prev_histinfo_num")=13
	Set gtmtypes("repl_histrec_msg_t",14,"name")="repl_histrec_msg_t.histjrec.histcontent.strm_index"
	Set gtmtypes("repl_histrec_msg_t",14,"off")=68
	Set gtmtypes("repl_histrec_msg_t",14,"len")=1
	Set gtmtypes("repl_histrec_msg_t",14,"type")="char"
	Set gtmtypfldindx("repl_histrec_msg_t","histjrec.histcontent.strm_index")=14
	Set gtmtypes("repl_histrec_msg_t",15,"name")="repl_histrec_msg_t.histjrec.histcontent.history_type"
	Set gtmtypes("repl_histrec_msg_t",15,"off")=69
	Set gtmtypes("repl_histrec_msg_t",15,"len")=1
	Set gtmtypes("repl_histrec_msg_t",15,"type")="char"
	Set gtmtypfldindx("repl_histrec_msg_t","histjrec.histcontent.history_type")=15
	Set gtmtypes("repl_histrec_msg_t",16,"name")="repl_histrec_msg_t.histjrec.histcontent.filler_8"
	Set gtmtypes("repl_histrec_msg_t",16,"off")=70
	Set gtmtypes("repl_histrec_msg_t",16,"len")=2
	Set gtmtypes("repl_histrec_msg_t",16,"type")="char"
	Set gtmtypfldindx("repl_histrec_msg_t","histjrec.histcontent.filler_8")=16
	Set gtmtypes("repl_histrec_msg_t",17,"name")="repl_histrec_msg_t.histjrec.histcontent.lms_group"
	Set gtmtypes("repl_histrec_msg_t",17,"off")=72
	Set gtmtypes("repl_histrec_msg_t",17,"len")=40
	Set gtmtypes("repl_histrec_msg_t",17,"type")="repl_inst_uuid"
	Set gtmtypfldindx("repl_histrec_msg_t","histjrec.histcontent.lms_group")=17
	Set gtmtypes("repl_histrec_msg_t",18,"name")="repl_histrec_msg_t.histjrec.histcontent.lms_group.created_nodename"
	Set gtmtypes("repl_histrec_msg_t",18,"off")=72
	Set gtmtypes("repl_histrec_msg_t",18,"len")=16
	Set gtmtypes("repl_histrec_msg_t",18,"type")="unsigned-char"
	Set gtmtypfldindx("repl_histrec_msg_t","histjrec.histcontent.lms_group.created_nodename")=18
	Set gtmtypes("repl_histrec_msg_t",19,"name")="repl_histrec_msg_t.histjrec.histcontent.lms_group.this_instname"
	Set gtmtypes("repl_histrec_msg_t",19,"off")=88
	Set gtmtypes("repl_histrec_msg_t",19,"len")=16
	Set gtmtypes("repl_histrec_msg_t",19,"type")="unsigned-char"
	Set gtmtypfldindx("repl_histrec_msg_t","histjrec.histcontent.lms_group.this_instname")=19
	Set gtmtypes("repl_histrec_msg_t",20,"name")="repl_histrec_msg_t.histjrec.histcontent.lms_group.created_time"
	Set gtmtypes("repl_histrec_msg_t",20,"off")=104
	Set gtmtypes("repl_histrec_msg_t",20,"len")=4
	Set gtmtypes("repl_histrec_msg_t",20,"type")="unsigned-int"
	Set gtmtypfldindx("repl_histrec_msg_t","histjrec.histcontent.lms_group.created_time")=20
	Set gtmtypes("repl_histrec_msg_t",21,"name")="repl_histrec_msg_t.histjrec.histcontent.lms_group.creator_pid"
	Set gtmtypes("repl_histrec_msg_t",21,"off")=108
	Set gtmtypes("repl_histrec_msg_t",21,"len")=4
	Set gtmtypes("repl_histrec_msg_t",21,"type")="unsigned-int"
	Set gtmtypfldindx("repl_histrec_msg_t","histjrec.histcontent.lms_group.creator_pid")=21
	Set gtmtypes("repl_histrec_msg_t",22,"name")="repl_histrec_msg_t.histjrec.histcontent.last_histinfo_num"
	Set gtmtypes("repl_histrec_msg_t",22,"off")=112
	Set gtmtypes("repl_histrec_msg_t",22,"len")=64
	Set gtmtypes("repl_histrec_msg_t",22,"type")="int"
	Set gtmtypfldindx("repl_histrec_msg_t","histjrec.histcontent.last_histinfo_num")=22
	Set gtmtypes("repl_histrec_msg_t",22,"dim")=16
	;
	Set gtmtypes("repl_inst_hdr")="struct"
	Set gtmtypes("repl_inst_hdr",0)=41
	Set gtmtypes("repl_inst_hdr","len")=1024
	Set gtmtypes("repl_inst_hdr",1,"name")="repl_inst_hdr.label"
	Set gtmtypes("repl_inst_hdr",1,"off")=0
	Set gtmtypes("repl_inst_hdr",1,"len")=12
	Set gtmtypes("repl_inst_hdr",1,"type")="unsigned-char"
	Set gtmtypfldindx("repl_inst_hdr","label")=1
	Set gtmtypes("repl_inst_hdr",2,"name")="repl_inst_hdr.replinst_minorver"
	Set gtmtypes("repl_inst_hdr",2,"off")=12
	Set gtmtypes("repl_inst_hdr",2,"len")=1
	Set gtmtypes("repl_inst_hdr",2,"type")="unsigned-char"
	Set gtmtypfldindx("repl_inst_hdr","replinst_minorver")=2
	Set gtmtypes("repl_inst_hdr",3,"name")="repl_inst_hdr.is_little_endian"
	Set gtmtypes("repl_inst_hdr",3,"off")=13
	Set gtmtypes("repl_inst_hdr",3,"len")=1
	Set gtmtypes("repl_inst_hdr",3,"type")="unsigned-char"
	Set gtmtypfldindx("repl_inst_hdr","is_little_endian")=3
	Set gtmtypes("repl_inst_hdr",4,"name")="repl_inst_hdr.is_64bit"
	Set gtmtypes("repl_inst_hdr",4,"off")=14
	Set gtmtypes("repl_inst_hdr",4,"len")=1
	Set gtmtypes("repl_inst_hdr",4,"type")="unsigned-char"
	Set gtmtypfldindx("repl_inst_hdr","is_64bit")=4
	Set gtmtypes("repl_inst_hdr",5,"name")="repl_inst_hdr.filler_16"
	Set gtmtypes("repl_inst_hdr",5,"off")=15
	Set gtmtypes("repl_inst_hdr",5,"len")=1
	Set gtmtypes("repl_inst_hdr",5,"type")="unsigned-char"
	Set gtmtypfldindx("repl_inst_hdr","filler_16")=5
	Set gtmtypes("repl_inst_hdr",6,"name")="repl_inst_hdr.jnlpool_semid"
	Set gtmtypes("repl_inst_hdr",6,"off")=16
	Set gtmtypes("repl_inst_hdr",6,"len")=4
	Set gtmtypes("repl_inst_hdr",6,"type")="int"
	Set gtmtypfldindx("repl_inst_hdr","jnlpool_semid")=6
	Set gtmtypes("repl_inst_hdr",7,"name")="repl_inst_hdr.jnlpool_shmid"
	Set gtmtypes("repl_inst_hdr",7,"off")=20
	Set gtmtypes("repl_inst_hdr",7,"len")=4
	Set gtmtypes("repl_inst_hdr",7,"type")="int"
	Set gtmtypfldindx("repl_inst_hdr","jnlpool_shmid")=7
	Set gtmtypes("repl_inst_hdr",8,"name")="repl_inst_hdr.recvpool_semid"
	Set gtmtypes("repl_inst_hdr",8,"off")=24
	Set gtmtypes("repl_inst_hdr",8,"len")=4
	Set gtmtypes("repl_inst_hdr",8,"type")="int"
	Set gtmtypfldindx("repl_inst_hdr","recvpool_semid")=8
	Set gtmtypes("repl_inst_hdr",9,"name")="repl_inst_hdr.recvpool_shmid"
	Set gtmtypes("repl_inst_hdr",9,"off")=28
	Set gtmtypes("repl_inst_hdr",9,"len")=4
	Set gtmtypes("repl_inst_hdr",9,"type")="int"
	Set gtmtypfldindx("repl_inst_hdr","recvpool_shmid")=9
	Set gtmtypes("repl_inst_hdr",10,"name")="repl_inst_hdr.jnlpool_semid_ctime"
	Set gtmtypes("repl_inst_hdr",10,"off")=32
	Set gtmtypes("repl_inst_hdr",10,"len")=8
	Set gtmtypes("repl_inst_hdr",10,"type")="time_t"
	Set gtmtypfldindx("repl_inst_hdr","jnlpool_semid_ctime")=10
	Set gtmtypes("repl_inst_hdr",11,"name")="repl_inst_hdr.jnlpool_shmid_ctime"
	Set gtmtypes("repl_inst_hdr",11,"off")=40
	Set gtmtypes("repl_inst_hdr",11,"len")=8
	Set gtmtypes("repl_inst_hdr",11,"type")="time_t"
	Set gtmtypfldindx("repl_inst_hdr","jnlpool_shmid_ctime")=11
	Set gtmtypes("repl_inst_hdr",12,"name")="repl_inst_hdr.recvpool_semid_ctime"
	Set gtmtypes("repl_inst_hdr",12,"off")=48
	Set gtmtypes("repl_inst_hdr",12,"len")=8
	Set gtmtypes("repl_inst_hdr",12,"type")="time_t"
	Set gtmtypfldindx("repl_inst_hdr","recvpool_semid_ctime")=12
	Set gtmtypes("repl_inst_hdr",13,"name")="repl_inst_hdr.recvpool_shmid_ctime"
	Set gtmtypes("repl_inst_hdr",13,"off")=56
	Set gtmtypes("repl_inst_hdr",13,"len")=8
	Set gtmtypes("repl_inst_hdr",13,"type")="time_t"
	Set gtmtypfldindx("repl_inst_hdr","recvpool_shmid_ctime")=13
	Set gtmtypes("repl_inst_hdr",14,"name")="repl_inst_hdr.inst_info"
	Set gtmtypes("repl_inst_hdr",14,"off")=64
	Set gtmtypes("repl_inst_hdr",14,"len")=40
	Set gtmtypes("repl_inst_hdr",14,"type")="repl_inst_uuid"
	Set gtmtypfldindx("repl_inst_hdr","inst_info")=14
	Set gtmtypes("repl_inst_hdr",15,"name")="repl_inst_hdr.inst_info.created_nodename"
	Set gtmtypes("repl_inst_hdr",15,"off")=64
	Set gtmtypes("repl_inst_hdr",15,"len")=16
	Set gtmtypes("repl_inst_hdr",15,"type")="unsigned-char"
	Set gtmtypfldindx("repl_inst_hdr","inst_info.created_nodename")=15
	Set gtmtypes("repl_inst_hdr",16,"name")="repl_inst_hdr.inst_info.this_instname"
	Set gtmtypes("repl_inst_hdr",16,"off")=80
	Set gtmtypes("repl_inst_hdr",16,"len")=16
	Set gtmtypes("repl_inst_hdr",16,"type")="unsigned-char"
	Set gtmtypfldindx("repl_inst_hdr","inst_info.this_instname")=16
	Set gtmtypes("repl_inst_hdr",17,"name")="repl_inst_hdr.inst_info.created_time"
	Set gtmtypes("repl_inst_hdr",17,"off")=96
	Set gtmtypes("repl_inst_hdr",17,"len")=4
	Set gtmtypes("repl_inst_hdr",17,"type")="unsigned-int"
	Set gtmtypfldindx("repl_inst_hdr","inst_info.created_time")=17
	Set gtmtypes("repl_inst_hdr",18,"name")="repl_inst_hdr.inst_info.creator_pid"
	Set gtmtypes("repl_inst_hdr",18,"off")=100
	Set gtmtypes("repl_inst_hdr",18,"len")=4
	Set gtmtypes("repl_inst_hdr",18,"type")="unsigned-int"
	Set gtmtypfldindx("repl_inst_hdr","inst_info.creator_pid")=18
	Set gtmtypes("repl_inst_hdr",19,"name")="repl_inst_hdr.lms_group_info"
	Set gtmtypes("repl_inst_hdr",19,"off")=104
	Set gtmtypes("repl_inst_hdr",19,"len")=40
	Set gtmtypes("repl_inst_hdr",19,"type")="repl_inst_uuid"
	Set gtmtypfldindx("repl_inst_hdr","lms_group_info")=19
	Set gtmtypes("repl_inst_hdr",20,"name")="repl_inst_hdr.lms_group_info.created_nodename"
	Set gtmtypes("repl_inst_hdr",20,"off")=104
	Set gtmtypes("repl_inst_hdr",20,"len")=16
	Set gtmtypes("repl_inst_hdr",20,"type")="unsigned-char"
	Set gtmtypfldindx("repl_inst_hdr","lms_group_info.created_nodename")=20
	Set gtmtypes("repl_inst_hdr",21,"name")="repl_inst_hdr.lms_group_info.this_instname"
	Set gtmtypes("repl_inst_hdr",21,"off")=120
	Set gtmtypes("repl_inst_hdr",21,"len")=16
	Set gtmtypes("repl_inst_hdr",21,"type")="unsigned-char"
	Set gtmtypfldindx("repl_inst_hdr","lms_group_info.this_instname")=21
	Set gtmtypes("repl_inst_hdr",22,"name")="repl_inst_hdr.lms_group_info.created_time"
	Set gtmtypes("repl_inst_hdr",22,"off")=136
	Set gtmtypes("repl_inst_hdr",22,"len")=4
	Set gtmtypes("repl_inst_hdr",22,"type")="unsigned-int"
	Set gtmtypfldindx("repl_inst_hdr","lms_group_info.created_time")=22
	Set gtmtypes("repl_inst_hdr",23,"name")="repl_inst_hdr.lms_group_info.creator_pid"
	Set gtmtypes("repl_inst_hdr",23,"off")=140
	Set gtmtypes("repl_inst_hdr",23,"len")=4
	Set gtmtypes("repl_inst_hdr",23,"type")="unsigned-int"
	Set gtmtypfldindx("repl_inst_hdr","lms_group_info.creator_pid")=23
	Set gtmtypes("repl_inst_hdr",24,"name")="repl_inst_hdr.jnl_seqno"
	Set gtmtypes("repl_inst_hdr",24,"off")=144
	Set gtmtypes("repl_inst_hdr",24,"len")=8
	Set gtmtypes("repl_inst_hdr",24,"type")="uint64_t"
	Set gtmtypfldindx("repl_inst_hdr","jnl_seqno")=24
	Set gtmtypes("repl_inst_hdr",25,"name")="repl_inst_hdr.root_primary_cycle"
	Set gtmtypes("repl_inst_hdr",25,"off")=152
	Set gtmtypes("repl_inst_hdr",25,"len")=4
	Set gtmtypes("repl_inst_hdr",25,"type")="unsigned-int"
	Set gtmtypfldindx("repl_inst_hdr","root_primary_cycle")=25
	Set gtmtypes("repl_inst_hdr",26,"name")="repl_inst_hdr.num_histinfo"
	Set gtmtypes("repl_inst_hdr",26,"off")=156
	Set gtmtypes("repl_inst_hdr",26,"len")=4
	Set gtmtypes("repl_inst_hdr",26,"type")="int"
	Set gtmtypfldindx("repl_inst_hdr","num_histinfo")=26
	Set gtmtypes("repl_inst_hdr",27,"name")="repl_inst_hdr.num_alloc_histinfo"
	Set gtmtypes("repl_inst_hdr",27,"off")=160
	Set gtmtypes("repl_inst_hdr",27,"len")=4
	Set gtmtypes("repl_inst_hdr",27,"type")="int"
	Set gtmtypfldindx("repl_inst_hdr","num_alloc_histinfo")=27
	Set gtmtypes("repl_inst_hdr",28,"name")="repl_inst_hdr.crash"
	Set gtmtypes("repl_inst_hdr",28,"off")=164
	Set gtmtypes("repl_inst_hdr",28,"len")=4
	Set gtmtypes("repl_inst_hdr",28,"type")="boolean_t"
	Set gtmtypfldindx("repl_inst_hdr","crash")=28
	Set gtmtypes("repl_inst_hdr",29,"name")="repl_inst_hdr.was_rootprimary"
	Set gtmtypes("repl_inst_hdr",29,"off")=168
	Set gtmtypes("repl_inst_hdr",29,"len")=4
	Set gtmtypes("repl_inst_hdr",29,"type")="boolean_t"
	Set gtmtypfldindx("repl_inst_hdr","was_rootprimary")=29
	Set gtmtypes("repl_inst_hdr",30,"name")="repl_inst_hdr.is_supplementary"
	Set gtmtypes("repl_inst_hdr",30,"off")=172
	Set gtmtypes("repl_inst_hdr",30,"len")=4
	Set gtmtypes("repl_inst_hdr",30,"type")="boolean_t"
	Set gtmtypfldindx("repl_inst_hdr","is_supplementary")=30
	Set gtmtypes("repl_inst_hdr",31,"name")="repl_inst_hdr.last_histinfo_num"
	Set gtmtypes("repl_inst_hdr",31,"off")=176
	Set gtmtypes("repl_inst_hdr",31,"len")=64
	Set gtmtypes("repl_inst_hdr",31,"type")="int"
	Set gtmtypfldindx("repl_inst_hdr","last_histinfo_num")=31
	Set gtmtypes("repl_inst_hdr",31,"dim")=16
	Set gtmtypes("repl_inst_hdr",32,"name")="repl_inst_hdr.strm_seqno"
	Set gtmtypes("repl_inst_hdr",32,"off")=240
	Set gtmtypes("repl_inst_hdr",32,"len")=128
	Set gtmtypes("repl_inst_hdr",32,"type")="uint64_t"
	Set gtmtypfldindx("repl_inst_hdr","strm_seqno")=32
	Set gtmtypes("repl_inst_hdr",32,"dim")=16
	Set gtmtypes("repl_inst_hdr",33,"name")="repl_inst_hdr.strm_group_info"
	Set gtmtypes("repl_inst_hdr",33,"off")=368
	Set gtmtypes("repl_inst_hdr",33,"len")=600
	Set gtmtypes("repl_inst_hdr",33,"type")="repl_inst_uuid"
	Set gtmtypfldindx("repl_inst_hdr","strm_group_info")=33
	Set gtmtypes("repl_inst_hdr",33,"dim")=15
	Set gtmtypes("repl_inst_hdr",34,"name")="repl_inst_hdr.strm_group_info[0].created_nodename"
	Set gtmtypes("repl_inst_hdr",34,"off")=368
	Set gtmtypes("repl_inst_hdr",34,"len")=16
	Set gtmtypes("repl_inst_hdr",34,"type")="unsigned-char"
	Set gtmtypfldindx("repl_inst_hdr","strm_group_info[0].created_nodename")=34
	Set gtmtypes("repl_inst_hdr",35,"name")="repl_inst_hdr.strm_group_info[0].this_instname"
	Set gtmtypes("repl_inst_hdr",35,"off")=384
	Set gtmtypes("repl_inst_hdr",35,"len")=16
	Set gtmtypes("repl_inst_hdr",35,"type")="unsigned-char"
	Set gtmtypfldindx("repl_inst_hdr","strm_group_info[0].this_instname")=35
	Set gtmtypes("repl_inst_hdr",36,"name")="repl_inst_hdr.strm_group_info[0].created_time"
	Set gtmtypes("repl_inst_hdr",36,"off")=400
	Set gtmtypes("repl_inst_hdr",36,"len")=4
	Set gtmtypes("repl_inst_hdr",36,"type")="unsigned-int"
	Set gtmtypfldindx("repl_inst_hdr","strm_group_info[0].created_time")=36
	Set gtmtypes("repl_inst_hdr",37,"name")="repl_inst_hdr.strm_group_info[0].creator_pid"
	Set gtmtypes("repl_inst_hdr",37,"off")=404
	Set gtmtypes("repl_inst_hdr",37,"len")=4
	Set gtmtypes("repl_inst_hdr",37,"type")="unsigned-int"
	Set gtmtypfldindx("repl_inst_hdr","strm_group_info[0].creator_pid")=37
	Set gtmtypes("repl_inst_hdr",38,"name")="repl_inst_hdr.file_corrupt"
	Set gtmtypes("repl_inst_hdr",38,"off")=968
	Set gtmtypes("repl_inst_hdr",38,"len")=4
	Set gtmtypes("repl_inst_hdr",38,"type")="boolean_t"
	Set gtmtypfldindx("repl_inst_hdr","file_corrupt")=38
	Set gtmtypes("repl_inst_hdr",39,"name")="repl_inst_hdr.filler_ftok_counter_halted"
	Set gtmtypes("repl_inst_hdr",39,"off")=972
	Set gtmtypes("repl_inst_hdr",39,"len")=4
	Set gtmtypes("repl_inst_hdr",39,"type")="boolean_t"
	Set gtmtypfldindx("repl_inst_hdr","filler_ftok_counter_halted")=39
	Set gtmtypes("repl_inst_hdr",40,"name")="repl_inst_hdr.qdbrundown"
	Set gtmtypes("repl_inst_hdr",40,"off")=976
	Set gtmtypes("repl_inst_hdr",40,"len")=4
	Set gtmtypes("repl_inst_hdr",40,"type")="boolean_t"
	Set gtmtypfldindx("repl_inst_hdr","qdbrundown")=40
	Set gtmtypes("repl_inst_hdr",41,"name")="repl_inst_hdr.filler_1024"
	Set gtmtypes("repl_inst_hdr",41,"off")=980
	Set gtmtypes("repl_inst_hdr",41,"len")=44
	Set gtmtypes("repl_inst_hdr",41,"type")="unsigned-char"
	Set gtmtypfldindx("repl_inst_hdr","filler_1024")=41
	;
	Set gtmtypes("repl_inst_uuid")="struct"
	Set gtmtypes("repl_inst_uuid",0)=4
	Set gtmtypes("repl_inst_uuid","len")=40
	Set gtmtypes("repl_inst_uuid",1,"name")="repl_inst_uuid.created_nodename"
	Set gtmtypes("repl_inst_uuid",1,"off")=0
	Set gtmtypes("repl_inst_uuid",1,"len")=16
	Set gtmtypes("repl_inst_uuid",1,"type")="unsigned-char"
	Set gtmtypfldindx("repl_inst_uuid","created_nodename")=1
	Set gtmtypes("repl_inst_uuid",2,"name")="repl_inst_uuid.this_instname"
	Set gtmtypes("repl_inst_uuid",2,"off")=16
	Set gtmtypes("repl_inst_uuid",2,"len")=16
	Set gtmtypes("repl_inst_uuid",2,"type")="unsigned-char"
	Set gtmtypfldindx("repl_inst_uuid","this_instname")=2
	Set gtmtypes("repl_inst_uuid",3,"name")="repl_inst_uuid.created_time"
	Set gtmtypes("repl_inst_uuid",3,"off")=32
	Set gtmtypes("repl_inst_uuid",3,"len")=4
	Set gtmtypes("repl_inst_uuid",3,"type")="unsigned-int"
	Set gtmtypfldindx("repl_inst_uuid","created_time")=3
	Set gtmtypes("repl_inst_uuid",4,"name")="repl_inst_uuid.creator_pid"
	Set gtmtypes("repl_inst_uuid",4,"off")=36
	Set gtmtypes("repl_inst_uuid",4,"len")=4
	Set gtmtypes("repl_inst_uuid",4,"type")="unsigned-int"
	Set gtmtypfldindx("repl_inst_uuid","creator_pid")=4
	;
	Set gtmtypes("repl_instinfo_msg_t")="struct"
	Set gtmtypes("repl_instinfo_msg_t",0)=11
	Set gtmtypes("repl_instinfo_msg_t","len")=80
	Set gtmtypes("repl_instinfo_msg_t",1,"name")="repl_instinfo_msg_t.type"
	Set gtmtypes("repl_instinfo_msg_t",1,"off")=0
	Set gtmtypes("repl_instinfo_msg_t",1,"len")=4
	Set gtmtypes("repl_instinfo_msg_t",1,"type")="int"
	Set gtmtypfldindx("repl_instinfo_msg_t","type")=1
	Set gtmtypes("repl_instinfo_msg_t",2,"name")="repl_instinfo_msg_t.len"
	Set gtmtypes("repl_instinfo_msg_t",2,"off")=4
	Set gtmtypes("repl_instinfo_msg_t",2,"len")=4
	Set gtmtypes("repl_instinfo_msg_t",2,"type")="int"
	Set gtmtypfldindx("repl_instinfo_msg_t","len")=2
	Set gtmtypes("repl_instinfo_msg_t",3,"name")="repl_instinfo_msg_t.instname"
	Set gtmtypes("repl_instinfo_msg_t",3,"off")=8
	Set gtmtypes("repl_instinfo_msg_t",3,"len")=16
	Set gtmtypes("repl_instinfo_msg_t",3,"type")="unsigned-char"
	Set gtmtypfldindx("repl_instinfo_msg_t","instname")=3
	Set gtmtypes("repl_instinfo_msg_t",4,"name")="repl_instinfo_msg_t.strm_jnl_seqno"
	Set gtmtypes("repl_instinfo_msg_t",4,"off")=24
	Set gtmtypes("repl_instinfo_msg_t",4,"len")=8
	Set gtmtypes("repl_instinfo_msg_t",4,"type")="uint64_t"
	Set gtmtypfldindx("repl_instinfo_msg_t","strm_jnl_seqno")=4
	Set gtmtypes("repl_instinfo_msg_t",5,"name")="repl_instinfo_msg_t.lms_group_info"
	Set gtmtypes("repl_instinfo_msg_t",5,"off")=32
	Set gtmtypes("repl_instinfo_msg_t",5,"len")=40
	Set gtmtypes("repl_instinfo_msg_t",5,"type")="repl_inst_uuid"
	Set gtmtypfldindx("repl_instinfo_msg_t","lms_group_info")=5
	Set gtmtypes("repl_instinfo_msg_t",6,"name")="repl_instinfo_msg_t.lms_group_info.created_nodename"
	Set gtmtypes("repl_instinfo_msg_t",6,"off")=32
	Set gtmtypes("repl_instinfo_msg_t",6,"len")=16
	Set gtmtypes("repl_instinfo_msg_t",6,"type")="unsigned-char"
	Set gtmtypfldindx("repl_instinfo_msg_t","lms_group_info.created_nodename")=6
	Set gtmtypes("repl_instinfo_msg_t",7,"name")="repl_instinfo_msg_t.lms_group_info.this_instname"
	Set gtmtypes("repl_instinfo_msg_t",7,"off")=48
	Set gtmtypes("repl_instinfo_msg_t",7,"len")=16
	Set gtmtypes("repl_instinfo_msg_t",7,"type")="unsigned-char"
	Set gtmtypfldindx("repl_instinfo_msg_t","lms_group_info.this_instname")=7
	Set gtmtypes("repl_instinfo_msg_t",8,"name")="repl_instinfo_msg_t.lms_group_info.created_time"
	Set gtmtypes("repl_instinfo_msg_t",8,"off")=64
	Set gtmtypes("repl_instinfo_msg_t",8,"len")=4
	Set gtmtypes("repl_instinfo_msg_t",8,"type")="unsigned-int"
	Set gtmtypfldindx("repl_instinfo_msg_t","lms_group_info.created_time")=8
	Set gtmtypes("repl_instinfo_msg_t",9,"name")="repl_instinfo_msg_t.lms_group_info.creator_pid"
	Set gtmtypes("repl_instinfo_msg_t",9,"off")=68
	Set gtmtypes("repl_instinfo_msg_t",9,"len")=4
	Set gtmtypes("repl_instinfo_msg_t",9,"type")="unsigned-int"
	Set gtmtypfldindx("repl_instinfo_msg_t","lms_group_info.creator_pid")=9
	Set gtmtypes("repl_instinfo_msg_t",10,"name")="repl_instinfo_msg_t.was_rootprimary"
	Set gtmtypes("repl_instinfo_msg_t",10,"off")=72
	Set gtmtypes("repl_instinfo_msg_t",10,"len")=1
	Set gtmtypes("repl_instinfo_msg_t",10,"type")="unsigned-char"
	Set gtmtypfldindx("repl_instinfo_msg_t","was_rootprimary")=10
	Set gtmtypes("repl_instinfo_msg_t",11,"name")="repl_instinfo_msg_t.filler_32"
	Set gtmtypes("repl_instinfo_msg_t",11,"off")=73
	Set gtmtypes("repl_instinfo_msg_t",11,"len")=7
	Set gtmtypes("repl_instinfo_msg_t",11,"type")="char"
	Set gtmtypfldindx("repl_instinfo_msg_t","filler_32")=11
	;
	Set gtmtypes("repl_logfile_info_msg_t")="struct"
	Set gtmtypes("repl_logfile_info_msg_t",0)=7
	Set gtmtypes("repl_logfile_info_msg_t","len")=1056
	Set gtmtypes("repl_logfile_info_msg_t",1,"name")="repl_logfile_info_msg_t.type"
	Set gtmtypes("repl_logfile_info_msg_t",1,"off")=0
	Set gtmtypes("repl_logfile_info_msg_t",1,"len")=4
	Set gtmtypes("repl_logfile_info_msg_t",1,"type")="int"
	Set gtmtypfldindx("repl_logfile_info_msg_t","type")=1
	Set gtmtypes("repl_logfile_info_msg_t",2,"name")="repl_logfile_info_msg_t.len"
	Set gtmtypes("repl_logfile_info_msg_t",2,"off")=4
	Set gtmtypes("repl_logfile_info_msg_t",2,"len")=4
	Set gtmtypes("repl_logfile_info_msg_t",2,"type")="int"
	Set gtmtypfldindx("repl_logfile_info_msg_t","len")=2
	Set gtmtypes("repl_logfile_info_msg_t",3,"name")="repl_logfile_info_msg_t.fullpath_len"
	Set gtmtypes("repl_logfile_info_msg_t",3,"off")=8
	Set gtmtypes("repl_logfile_info_msg_t",3,"len")=4
	Set gtmtypes("repl_logfile_info_msg_t",3,"type")="int"
	Set gtmtypfldindx("repl_logfile_info_msg_t","fullpath_len")=3
	Set gtmtypes("repl_logfile_info_msg_t",4,"name")="repl_logfile_info_msg_t.pid"
	Set gtmtypes("repl_logfile_info_msg_t",4,"off")=12
	Set gtmtypes("repl_logfile_info_msg_t",4,"len")=4
	Set gtmtypes("repl_logfile_info_msg_t",4,"type")="unsigned-int"
	Set gtmtypfldindx("repl_logfile_info_msg_t","pid")=4
	Set gtmtypes("repl_logfile_info_msg_t",5,"name")="repl_logfile_info_msg_t.proto_ver"
	Set gtmtypes("repl_logfile_info_msg_t",5,"off")=16
	Set gtmtypes("repl_logfile_info_msg_t",5,"len")=1
	Set gtmtypes("repl_logfile_info_msg_t",5,"type")="char"
	Set gtmtypfldindx("repl_logfile_info_msg_t","proto_ver")=5
	Set gtmtypes("repl_logfile_info_msg_t",6,"name")="repl_logfile_info_msg_t.filler_32"
	Set gtmtypes("repl_logfile_info_msg_t",6,"off")=17
	Set gtmtypes("repl_logfile_info_msg_t",6,"len")=15
	Set gtmtypes("repl_logfile_info_msg_t",6,"type")="char"
	Set gtmtypfldindx("repl_logfile_info_msg_t","filler_32")=6
	Set gtmtypes("repl_logfile_info_msg_t",7,"name")="repl_logfile_info_msg_t.fullpath"
	Set gtmtypes("repl_logfile_info_msg_t",7,"off")=32
	Set gtmtypes("repl_logfile_info_msg_t",7,"len")=1024
	Set gtmtypes("repl_logfile_info_msg_t",7,"type")="char"
	Set gtmtypfldindx("repl_logfile_info_msg_t","fullpath")=7
	;
	Set gtmtypes("repl_msg_t")="struct"
	Set gtmtypes("repl_msg_t",0)=3
	Set gtmtypes("repl_msg_t","len")=32
	Set gtmtypes("repl_msg_t",1,"name")="repl_msg_t.type"
	Set gtmtypes("repl_msg_t",1,"off")=0
	Set gtmtypes("repl_msg_t",1,"len")=4
	Set gtmtypes("repl_msg_t",1,"type")="int"
	Set gtmtypfldindx("repl_msg_t","type")=1
	Set gtmtypes("repl_msg_t",2,"name")="repl_msg_t.len"
	Set gtmtypes("repl_msg_t",2,"off")=4
	Set gtmtypes("repl_msg_t",2,"len")=4
	Set gtmtypes("repl_msg_t",2,"type")="int"
	Set gtmtypfldindx("repl_msg_t","len")=2
	Set gtmtypes("repl_msg_t",3,"name")="repl_msg_t.msg"
	Set gtmtypes("repl_msg_t",3,"off")=8
	Set gtmtypes("repl_msg_t",3,"len")=24
	Set gtmtypes("repl_msg_t",3,"type")="unsigned-char"
	Set gtmtypfldindx("repl_msg_t","msg")=3
	;
	Set gtmtypes("repl_needhistinfo_msg_t")="struct"
	Set gtmtypes("repl_needhistinfo_msg_t",0)=6
	Set gtmtypes("repl_needhistinfo_msg_t","len")=32
	Set gtmtypes("repl_needhistinfo_msg_t",1,"name")="repl_needhistinfo_msg_t.type"
	Set gtmtypes("repl_needhistinfo_msg_t",1,"off")=0
	Set gtmtypes("repl_needhistinfo_msg_t",1,"len")=4
	Set gtmtypes("repl_needhistinfo_msg_t",1,"type")="int"
	Set gtmtypfldindx("repl_needhistinfo_msg_t","type")=1
	Set gtmtypes("repl_needhistinfo_msg_t",2,"name")="repl_needhistinfo_msg_t.len"
	Set gtmtypes("repl_needhistinfo_msg_t",2,"off")=4
	Set gtmtypes("repl_needhistinfo_msg_t",2,"len")=4
	Set gtmtypes("repl_needhistinfo_msg_t",2,"type")="int"
	Set gtmtypfldindx("repl_needhistinfo_msg_t","len")=2
	Set gtmtypes("repl_needhistinfo_msg_t",3,"name")="repl_needhistinfo_msg_t.seqno"
	Set gtmtypes("repl_needhistinfo_msg_t",3,"off")=8
	Set gtmtypes("repl_needhistinfo_msg_t",3,"len")=8
	Set gtmtypes("repl_needhistinfo_msg_t",3,"type")="uint64_t"
	Set gtmtypfldindx("repl_needhistinfo_msg_t","seqno")=3
	Set gtmtypes("repl_needhistinfo_msg_t",4,"name")="repl_needhistinfo_msg_t.strm_num"
	Set gtmtypes("repl_needhistinfo_msg_t",4,"off")=16
	Set gtmtypes("repl_needhistinfo_msg_t",4,"len")=4
	Set gtmtypes("repl_needhistinfo_msg_t",4,"type")="int"
	Set gtmtypfldindx("repl_needhistinfo_msg_t","strm_num")=4
	Set gtmtypes("repl_needhistinfo_msg_t",5,"name")="repl_needhistinfo_msg_t.histinfo_num"
	Set gtmtypes("repl_needhistinfo_msg_t",5,"off")=20
	Set gtmtypes("repl_needhistinfo_msg_t",5,"len")=4
	Set gtmtypes("repl_needhistinfo_msg_t",5,"type")="int"
	Set gtmtypfldindx("repl_needhistinfo_msg_t","histinfo_num")=5
	Set gtmtypes("repl_needhistinfo_msg_t",6,"name")="repl_needhistinfo_msg_t.filler_32"
	Set gtmtypes("repl_needhistinfo_msg_t",6,"off")=24
	Set gtmtypes("repl_needhistinfo_msg_t",6,"len")=8
	Set gtmtypes("repl_needhistinfo_msg_t",6,"type")="char"
	Set gtmtypfldindx("repl_needhistinfo_msg_t","filler_32")=6
	;
	Set gtmtypes("repl_needinst_msg_t")="struct"
	Set gtmtypes("repl_needinst_msg_t",0)=13
	Set gtmtypes("repl_needinst_msg_t","len")=72
	Set gtmtypes("repl_needinst_msg_t",1,"name")="repl_needinst_msg_t.type"
	Set gtmtypes("repl_needinst_msg_t",1,"off")=0
	Set gtmtypes("repl_needinst_msg_t",1,"len")=4
	Set gtmtypes("repl_needinst_msg_t",1,"type")="int"
	Set gtmtypfldindx("repl_needinst_msg_t","type")=1
	Set gtmtypes("repl_needinst_msg_t",2,"name")="repl_needinst_msg_t.len"
	Set gtmtypes("repl_needinst_msg_t",2,"off")=4
	Set gtmtypes("repl_needinst_msg_t",2,"len")=4
	Set gtmtypes("repl_needinst_msg_t",2,"type")="int"
	Set gtmtypfldindx("repl_needinst_msg_t","len")=2
	Set gtmtypes("repl_needinst_msg_t",3,"name")="repl_needinst_msg_t.instname"
	Set gtmtypes("repl_needinst_msg_t",3,"off")=8
	Set gtmtypes("repl_needinst_msg_t",3,"len")=16
	Set gtmtypes("repl_needinst_msg_t",3,"type")="unsigned-char"
	Set gtmtypfldindx("repl_needinst_msg_t","instname")=3
	Set gtmtypes("repl_needinst_msg_t",4,"name")="repl_needinst_msg_t.lms_group_info"
	Set gtmtypes("repl_needinst_msg_t",4,"off")=24
	Set gtmtypes("repl_needinst_msg_t",4,"len")=40
	Set gtmtypes("repl_needinst_msg_t",4,"type")="repl_inst_uuid"
	Set gtmtypfldindx("repl_needinst_msg_t","lms_group_info")=4
	Set gtmtypes("repl_needinst_msg_t",5,"name")="repl_needinst_msg_t.lms_group_info.created_nodename"
	Set gtmtypes("repl_needinst_msg_t",5,"off")=24
	Set gtmtypes("repl_needinst_msg_t",5,"len")=16
	Set gtmtypes("repl_needinst_msg_t",5,"type")="unsigned-char"
	Set gtmtypfldindx("repl_needinst_msg_t","lms_group_info.created_nodename")=5
	Set gtmtypes("repl_needinst_msg_t",6,"name")="repl_needinst_msg_t.lms_group_info.this_instname"
	Set gtmtypes("repl_needinst_msg_t",6,"off")=40
	Set gtmtypes("repl_needinst_msg_t",6,"len")=16
	Set gtmtypes("repl_needinst_msg_t",6,"type")="unsigned-char"
	Set gtmtypfldindx("repl_needinst_msg_t","lms_group_info.this_instname")=6
	Set gtmtypes("repl_needinst_msg_t",7,"name")="repl_needinst_msg_t.lms_group_info.created_time"
	Set gtmtypes("repl_needinst_msg_t",7,"off")=56
	Set gtmtypes("repl_needinst_msg_t",7,"len")=4
	Set gtmtypes("repl_needinst_msg_t",7,"type")="unsigned-int"
	Set gtmtypfldindx("repl_needinst_msg_t","lms_group_info.created_time")=7
	Set gtmtypes("repl_needinst_msg_t",8,"name")="repl_needinst_msg_t.lms_group_info.creator_pid"
	Set gtmtypes("repl_needinst_msg_t",8,"off")=60
	Set gtmtypes("repl_needinst_msg_t",8,"len")=4
	Set gtmtypes("repl_needinst_msg_t",8,"type")="unsigned-int"
	Set gtmtypfldindx("repl_needinst_msg_t","lms_group_info.creator_pid")=8
	Set gtmtypes("repl_needinst_msg_t",9,"name")="repl_needinst_msg_t.proto_ver"
	Set gtmtypes("repl_needinst_msg_t",9,"off")=64
	Set gtmtypes("repl_needinst_msg_t",9,"len")=1
	Set gtmtypes("repl_needinst_msg_t",9,"type")="char"
	Set gtmtypfldindx("repl_needinst_msg_t","proto_ver")=9
	Set gtmtypes("repl_needinst_msg_t",10,"name")="repl_needinst_msg_t.is_rootprimary"
	Set gtmtypes("repl_needinst_msg_t",10,"off")=65
	Set gtmtypes("repl_needinst_msg_t",10,"len")=1
	Set gtmtypes("repl_needinst_msg_t",10,"type")="char"
	Set gtmtypfldindx("repl_needinst_msg_t","is_rootprimary")=10
	Set gtmtypes("repl_needinst_msg_t",11,"name")="repl_needinst_msg_t.is_supplementary"
	Set gtmtypes("repl_needinst_msg_t",11,"off")=66
	Set gtmtypes("repl_needinst_msg_t",11,"len")=1
	Set gtmtypes("repl_needinst_msg_t",11,"type")="char"
	Set gtmtypfldindx("repl_needinst_msg_t","is_supplementary")=11
	Set gtmtypes("repl_needinst_msg_t",12,"name")="repl_needinst_msg_t.jnl_ver"
	Set gtmtypes("repl_needinst_msg_t",12,"off")=67
	Set gtmtypes("repl_needinst_msg_t",12,"len")=1
	Set gtmtypes("repl_needinst_msg_t",12,"type")="char"
	Set gtmtypfldindx("repl_needinst_msg_t","jnl_ver")=12
	Set gtmtypes("repl_needinst_msg_t",13,"name")="repl_needinst_msg_t.filler_32"
	Set gtmtypes("repl_needinst_msg_t",13,"off")=68
	Set gtmtypes("repl_needinst_msg_t",13,"len")=4
	Set gtmtypes("repl_needinst_msg_t",13,"type")="char"
	Set gtmtypfldindx("repl_needinst_msg_t","filler_32")=13
	;
	Set gtmtypes("repl_needstrminfo_msg_t")="struct"
	Set gtmtypes("repl_needstrminfo_msg_t",0)=4
	Set gtmtypes("repl_needstrminfo_msg_t","len")=32
	Set gtmtypes("repl_needstrminfo_msg_t",1,"name")="repl_needstrminfo_msg_t.type"
	Set gtmtypes("repl_needstrminfo_msg_t",1,"off")=0
	Set gtmtypes("repl_needstrminfo_msg_t",1,"len")=4
	Set gtmtypes("repl_needstrminfo_msg_t",1,"type")="int"
	Set gtmtypfldindx("repl_needstrminfo_msg_t","type")=1
	Set gtmtypes("repl_needstrminfo_msg_t",2,"name")="repl_needstrminfo_msg_t.len"
	Set gtmtypes("repl_needstrminfo_msg_t",2,"off")=4
	Set gtmtypes("repl_needstrminfo_msg_t",2,"len")=4
	Set gtmtypes("repl_needstrminfo_msg_t",2,"type")="int"
	Set gtmtypfldindx("repl_needstrminfo_msg_t","len")=2
	Set gtmtypes("repl_needstrminfo_msg_t",3,"name")="repl_needstrminfo_msg_t.seqno"
	Set gtmtypes("repl_needstrminfo_msg_t",3,"off")=8
	Set gtmtypes("repl_needstrminfo_msg_t",3,"len")=8
	Set gtmtypes("repl_needstrminfo_msg_t",3,"type")="uint64_t"
	Set gtmtypfldindx("repl_needstrminfo_msg_t","seqno")=3
	Set gtmtypes("repl_needstrminfo_msg_t",4,"name")="repl_needstrminfo_msg_t.filler_32"
	Set gtmtypes("repl_needstrminfo_msg_t",4,"off")=16
	Set gtmtypes("repl_needstrminfo_msg_t",4,"len")=16
	Set gtmtypes("repl_needstrminfo_msg_t",4,"type")="char"
	Set gtmtypfldindx("repl_needstrminfo_msg_t","filler_32")=4
	;
	Set gtmtypes("repl_old_instinfo_msg_t")="struct"
	Set gtmtypes("repl_old_instinfo_msg_t",0)=5
	Set gtmtypes("repl_old_instinfo_msg_t","len")=32
	Set gtmtypes("repl_old_instinfo_msg_t",1,"name")="repl_old_instinfo_msg_t.type"
	Set gtmtypes("repl_old_instinfo_msg_t",1,"off")=0
	Set gtmtypes("repl_old_instinfo_msg_t",1,"len")=4
	Set gtmtypes("repl_old_instinfo_msg_t",1,"type")="int"
	Set gtmtypfldindx("repl_old_instinfo_msg_t","type")=1
	Set gtmtypes("repl_old_instinfo_msg_t",2,"name")="repl_old_instinfo_msg_t.len"
	Set gtmtypes("repl_old_instinfo_msg_t",2,"off")=4
	Set gtmtypes("repl_old_instinfo_msg_t",2,"len")=4
	Set gtmtypes("repl_old_instinfo_msg_t",2,"type")="int"
	Set gtmtypfldindx("repl_old_instinfo_msg_t","len")=2
	Set gtmtypes("repl_old_instinfo_msg_t",3,"name")="repl_old_instinfo_msg_t.instname"
	Set gtmtypes("repl_old_instinfo_msg_t",3,"off")=8
	Set gtmtypes("repl_old_instinfo_msg_t",3,"len")=16
	Set gtmtypes("repl_old_instinfo_msg_t",3,"type")="unsigned-char"
	Set gtmtypfldindx("repl_old_instinfo_msg_t","instname")=3
	Set gtmtypes("repl_old_instinfo_msg_t",4,"name")="repl_old_instinfo_msg_t.was_rootprimary"
	Set gtmtypes("repl_old_instinfo_msg_t",4,"off")=24
	Set gtmtypes("repl_old_instinfo_msg_t",4,"len")=1
	Set gtmtypes("repl_old_instinfo_msg_t",4,"type")="unsigned-char"
	Set gtmtypfldindx("repl_old_instinfo_msg_t","was_rootprimary")=4
	Set gtmtypes("repl_old_instinfo_msg_t",5,"name")="repl_old_instinfo_msg_t.filler_32"
	Set gtmtypes("repl_old_instinfo_msg_t",5,"off")=25
	Set gtmtypes("repl_old_instinfo_msg_t",5,"len")=7
	Set gtmtypes("repl_old_instinfo_msg_t",5,"type")="char"
	Set gtmtypfldindx("repl_old_instinfo_msg_t","filler_32")=5
	;
	Set gtmtypes("repl_old_needinst_msg_t")="struct"
	Set gtmtypes("repl_old_needinst_msg_t",0)=7
	Set gtmtypes("repl_old_needinst_msg_t","len")=32
	Set gtmtypes("repl_old_needinst_msg_t",1,"name")="repl_old_needinst_msg_t.type"
	Set gtmtypes("repl_old_needinst_msg_t",1,"off")=0
	Set gtmtypes("repl_old_needinst_msg_t",1,"len")=4
	Set gtmtypes("repl_old_needinst_msg_t",1,"type")="int"
	Set gtmtypfldindx("repl_old_needinst_msg_t","type")=1
	Set gtmtypes("repl_old_needinst_msg_t",2,"name")="repl_old_needinst_msg_t.len"
	Set gtmtypes("repl_old_needinst_msg_t",2,"off")=4
	Set gtmtypes("repl_old_needinst_msg_t",2,"len")=4
	Set gtmtypes("repl_old_needinst_msg_t",2,"type")="int"
	Set gtmtypfldindx("repl_old_needinst_msg_t","len")=2
	Set gtmtypes("repl_old_needinst_msg_t",3,"name")="repl_old_needinst_msg_t.instname"
	Set gtmtypes("repl_old_needinst_msg_t",3,"off")=8
	Set gtmtypes("repl_old_needinst_msg_t",3,"len")=16
	Set gtmtypes("repl_old_needinst_msg_t",3,"type")="unsigned-char"
	Set gtmtypfldindx("repl_old_needinst_msg_t","instname")=3
	Set gtmtypes("repl_old_needinst_msg_t",4,"name")="repl_old_needinst_msg_t.proto_ver"
	Set gtmtypes("repl_old_needinst_msg_t",4,"off")=24
	Set gtmtypes("repl_old_needinst_msg_t",4,"len")=1
	Set gtmtypes("repl_old_needinst_msg_t",4,"type")="char"
	Set gtmtypfldindx("repl_old_needinst_msg_t","proto_ver")=4
	Set gtmtypes("repl_old_needinst_msg_t",5,"name")="repl_old_needinst_msg_t.node_endianness"
	Set gtmtypes("repl_old_needinst_msg_t",5,"off")=25
	Set gtmtypes("repl_old_needinst_msg_t",5,"len")=1
	Set gtmtypes("repl_old_needinst_msg_t",5,"type")="char"
	Set gtmtypfldindx("repl_old_needinst_msg_t","node_endianness")=5
	Set gtmtypes("repl_old_needinst_msg_t",6,"name")="repl_old_needinst_msg_t.is_rootprimary"
	Set gtmtypes("repl_old_needinst_msg_t",6,"off")=26
	Set gtmtypes("repl_old_needinst_msg_t",6,"len")=1
	Set gtmtypes("repl_old_needinst_msg_t",6,"type")="char"
	Set gtmtypfldindx("repl_old_needinst_msg_t","is_rootprimary")=6
	Set gtmtypes("repl_old_needinst_msg_t",7,"name")="repl_old_needinst_msg_t.filler_32"
	Set gtmtypes("repl_old_needinst_msg_t",7,"off")=27
	Set gtmtypes("repl_old_needinst_msg_t",7,"len")=5
	Set gtmtypes("repl_old_needinst_msg_t",7,"type")="char"
	Set gtmtypfldindx("repl_old_needinst_msg_t","filler_32")=7
	;
	Set gtmtypes("repl_old_triple_jnl_t")="struct"
	Set gtmtypes("repl_old_triple_jnl_t",0)=4
	Set gtmtypes("repl_old_triple_jnl_t","len")=48
	Set gtmtypes("repl_old_triple_jnl_t",1,"name")="repl_old_triple_jnl_t.cycle"
	Set gtmtypes("repl_old_triple_jnl_t",1,"off")=4
	Set gtmtypes("repl_old_triple_jnl_t",1,"len")=4
	Set gtmtypes("repl_old_triple_jnl_t",1,"type")="unsigned-int"
	Set gtmtypfldindx("repl_old_triple_jnl_t","cycle")=1
	Set gtmtypes("repl_old_triple_jnl_t",2,"name")="repl_old_triple_jnl_t.start_seqno"
	Set gtmtypes("repl_old_triple_jnl_t",2,"off")=8
	Set gtmtypes("repl_old_triple_jnl_t",2,"len")=8
	Set gtmtypes("repl_old_triple_jnl_t",2,"type")="uint64_t"
	Set gtmtypfldindx("repl_old_triple_jnl_t","start_seqno")=2
	Set gtmtypes("repl_old_triple_jnl_t",3,"name")="repl_old_triple_jnl_t.instname"
	Set gtmtypes("repl_old_triple_jnl_t",3,"off")=16
	Set gtmtypes("repl_old_triple_jnl_t",3,"len")=16
	Set gtmtypes("repl_old_triple_jnl_t",3,"type")="unsigned-char"
	Set gtmtypfldindx("repl_old_triple_jnl_t","instname")=3
	Set gtmtypes("repl_old_triple_jnl_t",4,"name")="repl_old_triple_jnl_t.rcvd_from_instname"
	Set gtmtypes("repl_old_triple_jnl_t",4,"off")=32
	Set gtmtypes("repl_old_triple_jnl_t",4,"len")=16
	Set gtmtypes("repl_old_triple_jnl_t",4,"type")="unsigned-char"
	Set gtmtypfldindx("repl_old_triple_jnl_t","rcvd_from_instname")=4
	;
	Set gtmtypes("repl_old_triple_msg_t")="struct"
	Set gtmtypes("repl_old_triple_msg_t",0)=7
	Set gtmtypes("repl_old_triple_msg_t","len")=56
	Set gtmtypes("repl_old_triple_msg_t",1,"name")="repl_old_triple_msg_t.type"
	Set gtmtypes("repl_old_triple_msg_t",1,"off")=0
	Set gtmtypes("repl_old_triple_msg_t",1,"len")=4
	Set gtmtypes("repl_old_triple_msg_t",1,"type")="int"
	Set gtmtypfldindx("repl_old_triple_msg_t","type")=1
	Set gtmtypes("repl_old_triple_msg_t",2,"name")="repl_old_triple_msg_t.len"
	Set gtmtypes("repl_old_triple_msg_t",2,"off")=4
	Set gtmtypes("repl_old_triple_msg_t",2,"len")=4
	Set gtmtypes("repl_old_triple_msg_t",2,"type")="int"
	Set gtmtypfldindx("repl_old_triple_msg_t","len")=2
	Set gtmtypes("repl_old_triple_msg_t",3,"name")="repl_old_triple_msg_t.triplecontent"
	Set gtmtypes("repl_old_triple_msg_t",3,"off")=8
	Set gtmtypes("repl_old_triple_msg_t",3,"len")=48
	Set gtmtypes("repl_old_triple_msg_t",3,"type")="repl_old_triple_jnl_t"
	Set gtmtypfldindx("repl_old_triple_msg_t","triplecontent")=3
	Set gtmtypes("repl_old_triple_msg_t",4,"name")="repl_old_triple_msg_t.triplecontent.cycle"
	Set gtmtypes("repl_old_triple_msg_t",4,"off")=12
	Set gtmtypes("repl_old_triple_msg_t",4,"len")=4
	Set gtmtypes("repl_old_triple_msg_t",4,"type")="unsigned-int"
	Set gtmtypfldindx("repl_old_triple_msg_t","triplecontent.cycle")=4
	Set gtmtypes("repl_old_triple_msg_t",5,"name")="repl_old_triple_msg_t.triplecontent.start_seqno"
	Set gtmtypes("repl_old_triple_msg_t",5,"off")=16
	Set gtmtypes("repl_old_triple_msg_t",5,"len")=8
	Set gtmtypes("repl_old_triple_msg_t",5,"type")="uint64_t"
	Set gtmtypfldindx("repl_old_triple_msg_t","triplecontent.start_seqno")=5
	Set gtmtypes("repl_old_triple_msg_t",6,"name")="repl_old_triple_msg_t.triplecontent.instname"
	Set gtmtypes("repl_old_triple_msg_t",6,"off")=24
	Set gtmtypes("repl_old_triple_msg_t",6,"len")=16
	Set gtmtypes("repl_old_triple_msg_t",6,"type")="unsigned-char"
	Set gtmtypfldindx("repl_old_triple_msg_t","triplecontent.instname")=6
	Set gtmtypes("repl_old_triple_msg_t",7,"name")="repl_old_triple_msg_t.triplecontent.rcvd_from_instname"
	Set gtmtypes("repl_old_triple_msg_t",7,"off")=40
	Set gtmtypes("repl_old_triple_msg_t",7,"len")=16
	Set gtmtypes("repl_old_triple_msg_t",7,"type")="unsigned-char"
	Set gtmtypfldindx("repl_old_triple_msg_t","triplecontent.rcvd_from_instname")=7
	;
	Set gtmtypes("repl_rctl_elem_t")="struct"
	Set gtmtypes("repl_rctl_elem_t",0)=4
	Set gtmtypes("repl_rctl_elem_t","len")=32
	Set gtmtypes("repl_rctl_elem_t",1,"name")="repl_rctl_elem_t.prev"
	Set gtmtypes("repl_rctl_elem_t",1,"off")=0
	Set gtmtypes("repl_rctl_elem_t",1,"len")=8
	Set gtmtypes("repl_rctl_elem_t",1,"type")="addr"
	Set gtmtypfldindx("repl_rctl_elem_t","prev")=1
	Set gtmtypes("repl_rctl_elem_t",2,"name")="repl_rctl_elem_t.next"
	Set gtmtypes("repl_rctl_elem_t",2,"off")=8
	Set gtmtypes("repl_rctl_elem_t",2,"len")=8
	Set gtmtypes("repl_rctl_elem_t",2,"type")="addr"
	Set gtmtypfldindx("repl_rctl_elem_t","next")=2
	Set gtmtypes("repl_rctl_elem_t",3,"name")="repl_rctl_elem_t.ctl_start"
	Set gtmtypes("repl_rctl_elem_t",3,"off")=16
	Set gtmtypes("repl_rctl_elem_t",3,"len")=8
	Set gtmtypes("repl_rctl_elem_t",3,"type")="addr"
	Set gtmtypfldindx("repl_rctl_elem_t","ctl_start")=3
	Set gtmtypes("repl_rctl_elem_t",4,"name")="repl_rctl_elem_t.read_complete"
	Set gtmtypes("repl_rctl_elem_t",4,"off")=24
	Set gtmtypes("repl_rctl_elem_t",4,"len")=4
	Set gtmtypes("repl_rctl_elem_t",4,"type")="boolean_t"
	Set gtmtypfldindx("repl_rctl_elem_t","read_complete")=4
	;
	Set gtmtypes("repl_resync_msg_t")="struct"
	Set gtmtypes("repl_resync_msg_t",0)=7
	Set gtmtypes("repl_resync_msg_t","len")=32
	Set gtmtypes("repl_resync_msg_t",1,"name")="repl_resync_msg_t.type"
	Set gtmtypes("repl_resync_msg_t",1,"off")=0
	Set gtmtypes("repl_resync_msg_t",1,"len")=4
	Set gtmtypes("repl_resync_msg_t",1,"type")="int"
	Set gtmtypfldindx("repl_resync_msg_t","type")=1
	Set gtmtypes("repl_resync_msg_t",2,"name")="repl_resync_msg_t.len"
	Set gtmtypes("repl_resync_msg_t",2,"off")=4
	Set gtmtypes("repl_resync_msg_t",2,"len")=4
	Set gtmtypes("repl_resync_msg_t",2,"type")="int"
	Set gtmtypfldindx("repl_resync_msg_t","len")=2
	Set gtmtypes("repl_resync_msg_t",3,"name")="repl_resync_msg_t.resync_seqno"
	Set gtmtypes("repl_resync_msg_t",3,"off")=8
	Set gtmtypes("repl_resync_msg_t",3,"len")=8
	Set gtmtypes("repl_resync_msg_t",3,"type")="uint64_t"
	Set gtmtypfldindx("repl_resync_msg_t","resync_seqno")=3
	Set gtmtypes("repl_resync_msg_t",4,"name")="repl_resync_msg_t.proto_ver"
	Set gtmtypes("repl_resync_msg_t",4,"off")=16
	Set gtmtypes("repl_resync_msg_t",4,"len")=1
	Set gtmtypes("repl_resync_msg_t",4,"type")="char"
	Set gtmtypfldindx("repl_resync_msg_t","proto_ver")=4
	Set gtmtypes("repl_resync_msg_t",5,"name")="repl_resync_msg_t.node_endianness"
	Set gtmtypes("repl_resync_msg_t",5,"off")=17
	Set gtmtypes("repl_resync_msg_t",5,"len")=1
	Set gtmtypes("repl_resync_msg_t",5,"type")="char"
	Set gtmtypfldindx("repl_resync_msg_t","node_endianness")=5
	Set gtmtypes("repl_resync_msg_t",6,"name")="repl_resync_msg_t.is_supplementary"
	Set gtmtypes("repl_resync_msg_t",6,"off")=18
	Set gtmtypes("repl_resync_msg_t",6,"len")=1
	Set gtmtypes("repl_resync_msg_t",6,"type")="char"
	Set gtmtypfldindx("repl_resync_msg_t","is_supplementary")=6
	Set gtmtypes("repl_resync_msg_t",7,"name")="repl_resync_msg_t.filler_32"
	Set gtmtypes("repl_resync_msg_t",7,"off")=19
	Set gtmtypes("repl_resync_msg_t",7,"len")=13
	Set gtmtypes("repl_resync_msg_t",7,"type")="char"
	Set gtmtypfldindx("repl_resync_msg_t","filler_32")=7
	;
	Set gtmtypes("repl_start_msg_t")="struct"
	Set gtmtypes("repl_start_msg_t",0)=9
	Set gtmtypes("repl_start_msg_t","len")=32
	Set gtmtypes("repl_start_msg_t",1,"name")="repl_start_msg_t.type"
	Set gtmtypes("repl_start_msg_t",1,"off")=0
	Set gtmtypes("repl_start_msg_t",1,"len")=4
	Set gtmtypes("repl_start_msg_t",1,"type")="int"
	Set gtmtypfldindx("repl_start_msg_t","type")=1
	Set gtmtypes("repl_start_msg_t",2,"name")="repl_start_msg_t.len"
	Set gtmtypes("repl_start_msg_t",2,"off")=4
	Set gtmtypes("repl_start_msg_t",2,"len")=4
	Set gtmtypes("repl_start_msg_t",2,"type")="int"
	Set gtmtypfldindx("repl_start_msg_t","len")=2
	Set gtmtypes("repl_start_msg_t",3,"name")="repl_start_msg_t.start_seqno"
	Set gtmtypes("repl_start_msg_t",3,"off")=8
	Set gtmtypes("repl_start_msg_t",3,"len")=8
	Set gtmtypes("repl_start_msg_t",3,"type")="unsigned-char"
	Set gtmtypfldindx("repl_start_msg_t","start_seqno")=3
	Set gtmtypes("repl_start_msg_t",4,"name")="repl_start_msg_t.start_flags"
	Set gtmtypes("repl_start_msg_t",4,"off")=16
	Set gtmtypes("repl_start_msg_t",4,"len")=4
	Set gtmtypes("repl_start_msg_t",4,"type")="unsigned-int"
	Set gtmtypfldindx("repl_start_msg_t","start_flags")=4
	Set gtmtypes("repl_start_msg_t",5,"name")="repl_start_msg_t.jnl_ver"
	Set gtmtypes("repl_start_msg_t",5,"off")=20
	Set gtmtypes("repl_start_msg_t",5,"len")=1
	Set gtmtypes("repl_start_msg_t",5,"type")="unsigned-char"
	Set gtmtypfldindx("repl_start_msg_t","jnl_ver")=5
	Set gtmtypes("repl_start_msg_t",6,"name")="repl_start_msg_t.proto_ver"
	Set gtmtypes("repl_start_msg_t",6,"off")=21
	Set gtmtypes("repl_start_msg_t",6,"len")=1
	Set gtmtypes("repl_start_msg_t",6,"type")="char"
	Set gtmtypfldindx("repl_start_msg_t","proto_ver")=6
	Set gtmtypes("repl_start_msg_t",7,"name")="repl_start_msg_t.node_endianness"
	Set gtmtypes("repl_start_msg_t",7,"off")=22
	Set gtmtypes("repl_start_msg_t",7,"len")=1
	Set gtmtypes("repl_start_msg_t",7,"type")="char"
	Set gtmtypfldindx("repl_start_msg_t","node_endianness")=7
	Set gtmtypes("repl_start_msg_t",8,"name")="repl_start_msg_t.is_supplementary"
	Set gtmtypes("repl_start_msg_t",8,"off")=23
	Set gtmtypes("repl_start_msg_t",8,"len")=1
	Set gtmtypes("repl_start_msg_t",8,"type")="char"
	Set gtmtypfldindx("repl_start_msg_t","is_supplementary")=8
	Set gtmtypes("repl_start_msg_t",9,"name")="repl_start_msg_t.filler_32"
	Set gtmtypes("repl_start_msg_t",9,"off")=24
	Set gtmtypes("repl_start_msg_t",9,"len")=8
	Set gtmtypes("repl_start_msg_t",9,"type")="char"
	Set gtmtypfldindx("repl_start_msg_t","filler_32")=9
	;
	Set gtmtypes("repl_start_reply_msg_t")="struct"
	Set gtmtypes("repl_start_reply_msg_t",0)=9
	Set gtmtypes("repl_start_reply_msg_t","len")=32
	Set gtmtypes("repl_start_reply_msg_t",1,"name")="repl_start_reply_msg_t.type"
	Set gtmtypes("repl_start_reply_msg_t",1,"off")=0
	Set gtmtypes("repl_start_reply_msg_t",1,"len")=4
	Set gtmtypes("repl_start_reply_msg_t",1,"type")="int"
	Set gtmtypfldindx("repl_start_reply_msg_t","type")=1
	Set gtmtypes("repl_start_reply_msg_t",2,"name")="repl_start_reply_msg_t.len"
	Set gtmtypes("repl_start_reply_msg_t",2,"off")=4
	Set gtmtypes("repl_start_reply_msg_t",2,"len")=4
	Set gtmtypes("repl_start_reply_msg_t",2,"type")="int"
	Set gtmtypfldindx("repl_start_reply_msg_t","len")=2
	Set gtmtypes("repl_start_reply_msg_t",3,"name")="repl_start_reply_msg_t.start_seqno"
	Set gtmtypes("repl_start_reply_msg_t",3,"off")=8
	Set gtmtypes("repl_start_reply_msg_t",3,"len")=8
	Set gtmtypes("repl_start_reply_msg_t",3,"type")="unsigned-char"
	Set gtmtypfldindx("repl_start_reply_msg_t","start_seqno")=3
	Set gtmtypes("repl_start_reply_msg_t",4,"name")="repl_start_reply_msg_t.jnl_ver"
	Set gtmtypes("repl_start_reply_msg_t",4,"off")=16
	Set gtmtypes("repl_start_reply_msg_t",4,"len")=1
	Set gtmtypes("repl_start_reply_msg_t",4,"type")="unsigned-char"
	Set gtmtypfldindx("repl_start_reply_msg_t","jnl_ver")=4
	Set gtmtypes("repl_start_reply_msg_t",5,"name")="repl_start_reply_msg_t.start_flags"
	Set gtmtypes("repl_start_reply_msg_t",5,"off")=17
	Set gtmtypes("repl_start_reply_msg_t",5,"len")=4
	Set gtmtypes("repl_start_reply_msg_t",5,"type")="char"
	Set gtmtypfldindx("repl_start_reply_msg_t","start_flags")=5
	Set gtmtypes("repl_start_reply_msg_t",6,"name")="repl_start_reply_msg_t.proto_ver"
	Set gtmtypes("repl_start_reply_msg_t",6,"off")=21
	Set gtmtypes("repl_start_reply_msg_t",6,"len")=1
	Set gtmtypes("repl_start_reply_msg_t",6,"type")="char"
	Set gtmtypfldindx("repl_start_reply_msg_t","proto_ver")=6
	Set gtmtypes("repl_start_reply_msg_t",7,"name")="repl_start_reply_msg_t.node_endianness"
	Set gtmtypes("repl_start_reply_msg_t",7,"off")=22
	Set gtmtypes("repl_start_reply_msg_t",7,"len")=1
	Set gtmtypes("repl_start_reply_msg_t",7,"type")="char"
	Set gtmtypfldindx("repl_start_reply_msg_t","node_endianness")=7
	Set gtmtypes("repl_start_reply_msg_t",8,"name")="repl_start_reply_msg_t.is_supplementary"
	Set gtmtypes("repl_start_reply_msg_t",8,"off")=23
	Set gtmtypes("repl_start_reply_msg_t",8,"len")=1
	Set gtmtypes("repl_start_reply_msg_t",8,"type")="char"
	Set gtmtypfldindx("repl_start_reply_msg_t","is_supplementary")=8
	Set gtmtypes("repl_start_reply_msg_t",9,"name")="repl_start_reply_msg_t.filler_32"
	Set gtmtypes("repl_start_reply_msg_t",9,"off")=24
	Set gtmtypes("repl_start_reply_msg_t",9,"len")=8
	Set gtmtypes("repl_start_reply_msg_t",9,"type")="char"
	Set gtmtypfldindx("repl_start_reply_msg_t","filler_32")=9
	;
	Set gtmtypes("repl_strminfo_msg_t")="struct"
	Set gtmtypes("repl_strminfo_msg_t",0)=3
	Set gtmtypes("repl_strminfo_msg_t","len")=72
	Set gtmtypes("repl_strminfo_msg_t",1,"name")="repl_strminfo_msg_t.type"
	Set gtmtypes("repl_strminfo_msg_t",1,"off")=0
	Set gtmtypes("repl_strminfo_msg_t",1,"len")=4
	Set gtmtypes("repl_strminfo_msg_t",1,"type")="int"
	Set gtmtypfldindx("repl_strminfo_msg_t","type")=1
	Set gtmtypes("repl_strminfo_msg_t",2,"name")="repl_strminfo_msg_t.len"
	Set gtmtypes("repl_strminfo_msg_t",2,"off")=4
	Set gtmtypes("repl_strminfo_msg_t",2,"len")=4
	Set gtmtypes("repl_strminfo_msg_t",2,"type")="int"
	Set gtmtypfldindx("repl_strminfo_msg_t","len")=2
	Set gtmtypes("repl_strminfo_msg_t",3,"name")="repl_strminfo_msg_t.last_histinfo_num"
	Set gtmtypes("repl_strminfo_msg_t",3,"off")=8
	Set gtmtypes("repl_strminfo_msg_t",3,"len")=64
	Set gtmtypes("repl_strminfo_msg_t",3,"type")="int"
	Set gtmtypfldindx("repl_strminfo_msg_t","last_histinfo_num")=3
	Set gtmtypes("repl_strminfo_msg_t",3,"dim")=16
	;
	Set gtmtypes("repl_tls_info_t")="struct"
	Set gtmtypes("repl_tls_info_t",0)=6
	Set gtmtypes("repl_tls_info_t","len")=56
	Set gtmtypes("repl_tls_info_t",1,"name")="repl_tls_info_t.id"
	Set gtmtypes("repl_tls_info_t",1,"off")=0
	Set gtmtypes("repl_tls_info_t",1,"len")=32
	Set gtmtypes("repl_tls_info_t",1,"type")="char"
	Set gtmtypfldindx("repl_tls_info_t","id")=1
	Set gtmtypes("repl_tls_info_t",2,"name")="repl_tls_info_t.plaintext_fallback"
	Set gtmtypes("repl_tls_info_t",2,"off")=32
	Set gtmtypes("repl_tls_info_t",2,"len")=4
	Set gtmtypes("repl_tls_info_t",2,"type")="boolean_t"
	Set gtmtypfldindx("repl_tls_info_t","plaintext_fallback")=2
	Set gtmtypes("repl_tls_info_t",3,"name")="repl_tls_info_t.enabled"
	Set gtmtypes("repl_tls_info_t",3,"off")=36
	Set gtmtypes("repl_tls_info_t",3,"len")=4
	Set gtmtypes("repl_tls_info_t",3,"type")="boolean_t"
	Set gtmtypfldindx("repl_tls_info_t","enabled")=3
	Set gtmtypes("repl_tls_info_t",4,"name")="repl_tls_info_t.notls_retry"
	Set gtmtypes("repl_tls_info_t",4,"off")=40
	Set gtmtypes("repl_tls_info_t",4,"len")=4
	Set gtmtypes("repl_tls_info_t",4,"type")="boolean_t"
	Set gtmtypfldindx("repl_tls_info_t","notls_retry")=4
	Set gtmtypes("repl_tls_info_t",5,"name")="repl_tls_info_t.renegotiate_state"
	Set gtmtypes("repl_tls_info_t",5,"off")=44
	Set gtmtypes("repl_tls_info_t",5,"len")=4
	Set gtmtypes("repl_tls_info_t",5,"type")="int"
	Set gtmtypfldindx("repl_tls_info_t","renegotiate_state")=5
	Set gtmtypes("repl_tls_info_t",6,"name")="repl_tls_info_t.sock"
	Set gtmtypes("repl_tls_info_t",6,"off")=48
	Set gtmtypes("repl_tls_info_t",6,"len")=8
	Set gtmtypes("repl_tls_info_t",6,"type")="addr"
	Set gtmtypfldindx("repl_tls_info_t","sock")=6
	;
	Set gtmtypes("repl_tlsinfo_msg_t")="struct"
	Set gtmtypes("repl_tlsinfo_msg_t",0)=5
	Set gtmtypes("repl_tlsinfo_msg_t","len")=32
	Set gtmtypes("repl_tlsinfo_msg_t",1,"name")="repl_tlsinfo_msg_t.type"
	Set gtmtypes("repl_tlsinfo_msg_t",1,"off")=0
	Set gtmtypes("repl_tlsinfo_msg_t",1,"len")=4
	Set gtmtypes("repl_tlsinfo_msg_t",1,"type")="int"
	Set gtmtypfldindx("repl_tlsinfo_msg_t","type")=1
	Set gtmtypes("repl_tlsinfo_msg_t",2,"name")="repl_tlsinfo_msg_t.len"
	Set gtmtypes("repl_tlsinfo_msg_t",2,"off")=4
	Set gtmtypes("repl_tlsinfo_msg_t",2,"len")=4
	Set gtmtypes("repl_tlsinfo_msg_t",2,"type")="int"
	Set gtmtypfldindx("repl_tlsinfo_msg_t","len")=2
	Set gtmtypes("repl_tlsinfo_msg_t",3,"name")="repl_tlsinfo_msg_t.API_version"
	Set gtmtypes("repl_tlsinfo_msg_t",3,"off")=8
	Set gtmtypes("repl_tlsinfo_msg_t",3,"len")=4
	Set gtmtypes("repl_tlsinfo_msg_t",3,"type")="unsigned-int"
	Set gtmtypfldindx("repl_tlsinfo_msg_t","API_version")=3
	Set gtmtypes("repl_tlsinfo_msg_t",4,"name")="repl_tlsinfo_msg_t.library_version"
	Set gtmtypes("repl_tlsinfo_msg_t",4,"off")=12
	Set gtmtypes("repl_tlsinfo_msg_t",4,"len")=4
	Set gtmtypes("repl_tlsinfo_msg_t",4,"type")="unsigned-int"
	Set gtmtypfldindx("repl_tlsinfo_msg_t","library_version")=4
	Set gtmtypes("repl_tlsinfo_msg_t",5,"name")="repl_tlsinfo_msg_t.filler_32"
	Set gtmtypes("repl_tlsinfo_msg_t",5,"off")=16
	Set gtmtypes("repl_tlsinfo_msg_t",5,"len")=16
	Set gtmtypes("repl_tlsinfo_msg_t",5,"type")="char"
	Set gtmtypfldindx("repl_tlsinfo_msg_t","filler_32")=5
	;
	Set gtmtypes("replgbl_t")="struct"
	Set gtmtypes("replgbl_t",0)=3
	Set gtmtypes("replgbl_t","len")=24
	Set gtmtypes("replgbl_t",1,"name")="replgbl_t.trig_replic_warning_issued"
	Set gtmtypes("replgbl_t",1,"off")=0
	Set gtmtypes("replgbl_t",1,"len")=4
	Set gtmtypes("replgbl_t",1,"type")="boolean_t"
	Set gtmtypfldindx("replgbl_t","trig_replic_warning_issued")=1
	Set gtmtypes("replgbl_t",2,"name")="replgbl_t.trig_replic_suspect_seqno"
	Set gtmtypes("replgbl_t",2,"off")=8
	Set gtmtypes("replgbl_t",2,"len")=8
	Set gtmtypes("replgbl_t",2,"type")="uint64_t"
	Set gtmtypfldindx("replgbl_t","trig_replic_suspect_seqno")=2
	Set gtmtypes("replgbl_t",3,"name")="replgbl_t.jnl_release_timeout"
	Set gtmtypes("replgbl_t",3,"off")=16
	Set gtmtypes("replgbl_t",3,"len")=4
	Set gtmtypes("replgbl_t",3,"type")="int"
	Set gtmtypfldindx("replgbl_t","jnl_release_timeout")=3
	;
	Set gtmtypes("replpool_identifier")="struct"
	Set gtmtypes("replpool_identifier",0)=5
	Set gtmtypes("replpool_identifier","len")=312
	Set gtmtypes("replpool_identifier",1,"name")="replpool_identifier.label"
	Set gtmtypes("replpool_identifier",1,"off")=0
	Set gtmtypes("replpool_identifier",1,"len")=12
	Set gtmtypes("replpool_identifier",1,"type")="unsigned-char"
	Set gtmtypfldindx("replpool_identifier","label")=1
	Set gtmtypes("replpool_identifier",2,"name")="replpool_identifier.pool_type"
	Set gtmtypes("replpool_identifier",2,"off")=12
	Set gtmtypes("replpool_identifier",2,"len")=1
	Set gtmtypes("replpool_identifier",2,"type")="char"
	Set gtmtypfldindx("replpool_identifier","pool_type")=2
	Set gtmtypes("replpool_identifier",3,"name")="replpool_identifier.now_running"
	Set gtmtypes("replpool_identifier",3,"off")=13
	Set gtmtypes("replpool_identifier",3,"len")=36
	Set gtmtypes("replpool_identifier",3,"type")="char"
	Set gtmtypfldindx("replpool_identifier","now_running")=3
	Set gtmtypes("replpool_identifier",4,"name")="replpool_identifier.repl_pool_key_filler"
	Set gtmtypes("replpool_identifier",4,"off")=52
	Set gtmtypes("replpool_identifier",4,"len")=4
	Set gtmtypes("replpool_identifier",4,"type")="int"
	Set gtmtypfldindx("replpool_identifier","repl_pool_key_filler")=4
	Set gtmtypes("replpool_identifier",5,"name")="replpool_identifier.instfilename"
	Set gtmtypes("replpool_identifier",5,"off")=56
	Set gtmtypes("replpool_identifier",5,"len")=256
	Set gtmtypes("replpool_identifier",5,"type")="char"
	Set gtmtypfldindx("replpool_identifier","instfilename")=5
	;
	Set gtmtypes("rhdtyp")="struct"
	Set gtmtypes("rhdtyp",0)=48
	Set gtmtypes("rhdtyp","len")=312
	Set gtmtypes("rhdtyp",1,"name")="rhdtyp.jsb"
	Set gtmtypes("rhdtyp",1,"off")=0
	Set gtmtypes("rhdtyp",1,"len")=16
	Set gtmtypes("rhdtyp",1,"type")="char"
	Set gtmtypfldindx("rhdtyp","jsb")=1
	Set gtmtypes("rhdtyp",2,"name")="rhdtyp.shlib_handle"
	Set gtmtypes("rhdtyp",2,"off")=16
	Set gtmtypes("rhdtyp",2,"len")=8
	Set gtmtypes("rhdtyp",2,"type")="addr"
	Set gtmtypfldindx("rhdtyp","shlib_handle")=2
	Set gtmtypes("rhdtyp",3,"name")="rhdtyp.src_full_name"
	Set gtmtypes("rhdtyp",3,"off")=24
	Set gtmtypes("rhdtyp",3,"len")=16
	Set gtmtypes("rhdtyp",3,"type")="mstr"
	Set gtmtypfldindx("rhdtyp","src_full_name")=3
	Set gtmtypes("rhdtyp",4,"name")="rhdtyp.src_full_name.char_len"
	Set gtmtypes("rhdtyp",4,"off")=24
	Set gtmtypes("rhdtyp",4,"len")=4
	Set gtmtypes("rhdtyp",4,"type")="unsigned-int"
	Set gtmtypfldindx("rhdtyp","src_full_name.char_len")=4
	Set gtmtypes("rhdtyp",5,"name")="rhdtyp.src_full_name.len"
	Set gtmtypes("rhdtyp",5,"off")=28
	Set gtmtypes("rhdtyp",5,"len")=4
	Set gtmtypes("rhdtyp",5,"type")="int"
	Set gtmtypfldindx("rhdtyp","src_full_name.len")=5
	Set gtmtypes("rhdtyp",6,"name")="rhdtyp.src_full_name.addr"
	Set gtmtypes("rhdtyp",6,"off")=32
	Set gtmtypes("rhdtyp",6,"len")=8
	Set gtmtypes("rhdtyp",6,"type")="addr"
	Set gtmtypfldindx("rhdtyp","src_full_name.addr")=6
	Set gtmtypes("rhdtyp",7,"name")="rhdtyp.compiler_qlf"
	Set gtmtypes("rhdtyp",7,"off")=40
	Set gtmtypes("rhdtyp",7,"len")=4
	Set gtmtypes("rhdtyp",7,"type")="unsigned-int"
	Set gtmtypfldindx("rhdtyp","compiler_qlf")=7
	Set gtmtypes("rhdtyp",8,"name")="rhdtyp.objlabel"
	Set gtmtypes("rhdtyp",8,"off")=44
	Set gtmtypes("rhdtyp",8,"len")=4
	Set gtmtypes("rhdtyp",8,"type")="unsigned-int"
	Set gtmtypfldindx("rhdtyp","objlabel")=8
	Set gtmtypes("rhdtyp",9,"name")="rhdtyp.routine_name"
	Set gtmtypes("rhdtyp",9,"off")=48
	Set gtmtypes("rhdtyp",9,"len")=16
	Set gtmtypes("rhdtyp",9,"type")="mstr"
	Set gtmtypfldindx("rhdtyp","routine_name")=9
	Set gtmtypes("rhdtyp",10,"name")="rhdtyp.routine_name.char_len"
	Set gtmtypes("rhdtyp",10,"off")=48
	Set gtmtypes("rhdtyp",10,"len")=4
	Set gtmtypes("rhdtyp",10,"type")="unsigned-int"
	Set gtmtypfldindx("rhdtyp","routine_name.char_len")=10
	Set gtmtypes("rhdtyp",11,"name")="rhdtyp.routine_name.len"
	Set gtmtypes("rhdtyp",11,"off")=52
	Set gtmtypes("rhdtyp",11,"len")=4
	Set gtmtypes("rhdtyp",11,"type")="int"
	Set gtmtypfldindx("rhdtyp","routine_name.len")=11
	Set gtmtypes("rhdtyp",12,"name")="rhdtyp.routine_name.addr"
	Set gtmtypes("rhdtyp",12,"off")=56
	Set gtmtypes("rhdtyp",12,"len")=8
	Set gtmtypes("rhdtyp",12,"type")="addr"
	Set gtmtypfldindx("rhdtyp","routine_name.addr")=12
	Set gtmtypes("rhdtyp",13,"name")="rhdtyp.vartab_adr"
	Set gtmtypes("rhdtyp",13,"off")=64
	Set gtmtypes("rhdtyp",13,"len")=8
	Set gtmtypes("rhdtyp",13,"type")="addr"
	Set gtmtypfldindx("rhdtyp","vartab_adr")=13
	Set gtmtypes("rhdtyp",14,"name")="rhdtyp.vartab_len"
	Set gtmtypes("rhdtyp",14,"off")=72
	Set gtmtypes("rhdtyp",14,"len")=4
	Set gtmtypes("rhdtyp",14,"type")="int"
	Set gtmtypfldindx("rhdtyp","vartab_len")=14
	Set gtmtypes("rhdtyp",15,"name")="rhdtyp.labtab_adr"
	Set gtmtypes("rhdtyp",15,"off")=80
	Set gtmtypes("rhdtyp",15,"len")=8
	Set gtmtypes("rhdtyp",15,"type")="addr"
	Set gtmtypfldindx("rhdtyp","labtab_adr")=15
	Set gtmtypes("rhdtyp",16,"name")="rhdtyp.labtab_len"
	Set gtmtypes("rhdtyp",16,"off")=88
	Set gtmtypes("rhdtyp",16,"len")=4
	Set gtmtypes("rhdtyp",16,"type")="int"
	Set gtmtypfldindx("rhdtyp","labtab_len")=16
	Set gtmtypes("rhdtyp",17,"name")="rhdtyp.lnrtab_adr"
	Set gtmtypes("rhdtyp",17,"off")=96
	Set gtmtypes("rhdtyp",17,"len")=8
	Set gtmtypes("rhdtyp",17,"type")="addr"
	Set gtmtypfldindx("rhdtyp","lnrtab_adr")=17
	Set gtmtypes("rhdtyp",18,"name")="rhdtyp.lnrtab_len"
	Set gtmtypes("rhdtyp",18,"off")=104
	Set gtmtypes("rhdtyp",18,"len")=4
	Set gtmtypes("rhdtyp",18,"type")="int"
	Set gtmtypfldindx("rhdtyp","lnrtab_len")=18
	Set gtmtypes("rhdtyp",19,"name")="rhdtyp.literal_text_adr"
	Set gtmtypes("rhdtyp",19,"off")=112
	Set gtmtypes("rhdtyp",19,"len")=8
	Set gtmtypes("rhdtyp",19,"type")="addr"
	Set gtmtypfldindx("rhdtyp","literal_text_adr")=19
	Set gtmtypes("rhdtyp",20,"name")="rhdtyp.literal_text_len"
	Set gtmtypes("rhdtyp",20,"off")=120
	Set gtmtypes("rhdtyp",20,"len")=4
	Set gtmtypes("rhdtyp",20,"type")="int"
	Set gtmtypfldindx("rhdtyp","literal_text_len")=20
	Set gtmtypes("rhdtyp",21,"name")="rhdtyp.shared_object"
	Set gtmtypes("rhdtyp",21,"off")=124
	Set gtmtypes("rhdtyp",21,"len")=4
	Set gtmtypes("rhdtyp",21,"type")="boolean_t"
	Set gtmtypfldindx("rhdtyp","shared_object")=21
	Set gtmtypes("rhdtyp",22,"name")="rhdtyp.literal_adr"
	Set gtmtypes("rhdtyp",22,"off")=128
	Set gtmtypes("rhdtyp",22,"len")=8
	Set gtmtypes("rhdtyp",22,"type")="addr"
	Set gtmtypfldindx("rhdtyp","literal_adr")=22
	Set gtmtypes("rhdtyp",23,"name")="rhdtyp.literal_len"
	Set gtmtypes("rhdtyp",23,"off")=136
	Set gtmtypes("rhdtyp",23,"len")=4
	Set gtmtypes("rhdtyp",23,"type")="int"
	Set gtmtypfldindx("rhdtyp","literal_len")=23
	Set gtmtypes("rhdtyp",24,"name")="rhdtyp.linkage_adr"
	Set gtmtypes("rhdtyp",24,"off")=144
	Set gtmtypes("rhdtyp",24,"len")=8
	Set gtmtypes("rhdtyp",24,"type")="addr"
	Set gtmtypfldindx("rhdtyp","linkage_adr")=24
	Set gtmtypes("rhdtyp",25,"name")="rhdtyp.linkage_len"
	Set gtmtypes("rhdtyp",25,"off")=152
	Set gtmtypes("rhdtyp",25,"len")=4
	Set gtmtypes("rhdtyp",25,"type")="int"
	Set gtmtypfldindx("rhdtyp","linkage_len")=25
	Set gtmtypes("rhdtyp",26,"name")="rhdtyp.rel_table_off"
	Set gtmtypes("rhdtyp",26,"off")=156
	Set gtmtypes("rhdtyp",26,"len")=4
	Set gtmtypes("rhdtyp",26,"type")="int"
	Set gtmtypfldindx("rhdtyp","rel_table_off")=26
	Set gtmtypes("rhdtyp",27,"name")="rhdtyp.sym_table_off"
	Set gtmtypes("rhdtyp",27,"off")=160
	Set gtmtypes("rhdtyp",27,"len")=4
	Set gtmtypes("rhdtyp",27,"type")="int"
	Set gtmtypfldindx("rhdtyp","sym_table_off")=27
	Set gtmtypes("rhdtyp",28,"name")="rhdtyp.rtn_relinked"
	Set gtmtypes("rhdtyp",28,"off")=164
	Set gtmtypes("rhdtyp",28,"len")=4
	Set gtmtypes("rhdtyp",28,"type")="boolean_t"
	Set gtmtypfldindx("rhdtyp","rtn_relinked")=28
	Set gtmtypes("rhdtyp",29,"name")="rhdtyp.shared_ptext_adr"
	Set gtmtypes("rhdtyp",29,"off")=168
	Set gtmtypes("rhdtyp",29,"len")=8
	Set gtmtypes("rhdtyp",29,"type")="addr"
	Set gtmtypfldindx("rhdtyp","shared_ptext_adr")=29
	Set gtmtypes("rhdtyp",30,"name")="rhdtyp.ptext_adr"
	Set gtmtypes("rhdtyp",30,"off")=176
	Set gtmtypes("rhdtyp",30,"len")=8
	Set gtmtypes("rhdtyp",30,"type")="addr"
	Set gtmtypfldindx("rhdtyp","ptext_adr")=30
	Set gtmtypes("rhdtyp",31,"name")="rhdtyp.ptext_end_adr"
	Set gtmtypes("rhdtyp",31,"off")=184
	Set gtmtypes("rhdtyp",31,"len")=8
	Set gtmtypes("rhdtyp",31,"type")="addr"
	Set gtmtypfldindx("rhdtyp","ptext_end_adr")=31
	Set gtmtypes("rhdtyp",32,"name")="rhdtyp.checksum"
	Set gtmtypes("rhdtyp",32,"off")=192
	Set gtmtypes("rhdtyp",32,"len")=4
	Set gtmtypes("rhdtyp",32,"type")="int"
	Set gtmtypfldindx("rhdtyp","checksum")=32
	Set gtmtypes("rhdtyp",33,"name")="rhdtyp.temp_mvals"
	Set gtmtypes("rhdtyp",33,"off")=196
	Set gtmtypes("rhdtyp",33,"len")=4
	Set gtmtypes("rhdtyp",33,"type")="int"
	Set gtmtypfldindx("rhdtyp","temp_mvals")=33
	Set gtmtypes("rhdtyp",34,"name")="rhdtyp.temp_size"
	Set gtmtypes("rhdtyp",34,"off")=200
	Set gtmtypes("rhdtyp",34,"len")=4
	Set gtmtypes("rhdtyp",34,"type")="int"
	Set gtmtypfldindx("rhdtyp","temp_size")=34
	Set gtmtypes("rhdtyp",35,"name")="rhdtyp.has_ZBREAK"
	Set gtmtypes("rhdtyp",35,"off")=204
	Set gtmtypes("rhdtyp",35,"len")=4
	Set gtmtypes("rhdtyp",35,"type")="boolean_t"
	Set gtmtypfldindx("rhdtyp","has_ZBREAK")=35
	Set gtmtypes("rhdtyp",36,"name")="rhdtyp.current_rhead_adr"
	Set gtmtypes("rhdtyp",36,"off")=208
	Set gtmtypes("rhdtyp",36,"len")=8
	Set gtmtypes("rhdtyp",36,"type")="addr"
	Set gtmtypfldindx("rhdtyp","current_rhead_adr")=36
	Set gtmtypes("rhdtyp",37,"name")="rhdtyp.old_rhead_adr"
	Set gtmtypes("rhdtyp",37,"off")=216
	Set gtmtypes("rhdtyp",37,"len")=8
	Set gtmtypes("rhdtyp",37,"type")="addr"
	Set gtmtypfldindx("rhdtyp","old_rhead_adr")=37
	Set gtmtypes("rhdtyp",38,"name")="rhdtyp.trigr_handle"
	Set gtmtypes("rhdtyp",38,"off")=224
	Set gtmtypes("rhdtyp",38,"len")=8
	Set gtmtypes("rhdtyp",38,"type")="addr"
	Set gtmtypfldindx("rhdtyp","trigr_handle")=38
	Set gtmtypes("rhdtyp",39,"name")="rhdtyp.checksum_128"
	Set gtmtypes("rhdtyp",39,"off")=232
	Set gtmtypes("rhdtyp",39,"len")=16
	Set gtmtypes("rhdtyp",39,"type")="unsigned-char"
	Set gtmtypfldindx("rhdtyp","checksum_128")=39
	Set gtmtypes("rhdtyp",40,"name")="rhdtyp.active_rhead_adr"
	Set gtmtypes("rhdtyp",40,"off")=248
	Set gtmtypes("rhdtyp",40,"len")=8
	Set gtmtypes("rhdtyp",40,"type")="addr"
	Set gtmtypfldindx("rhdtyp","active_rhead_adr")=40
	Set gtmtypes("rhdtyp",41,"name")="rhdtyp.source_code"
	Set gtmtypes("rhdtyp",41,"off")=256
	Set gtmtypes("rhdtyp",41,"len")=8
	Set gtmtypes("rhdtyp",41,"type")="addr"
	Set gtmtypfldindx("rhdtyp","source_code")=41
	Set gtmtypes("rhdtyp",42,"name")="rhdtyp.zhist"
	Set gtmtypes("rhdtyp",42,"off")=264
	Set gtmtypes("rhdtyp",42,"len")=8
	Set gtmtypes("rhdtyp",42,"type")="addr"
	Set gtmtypfldindx("rhdtyp","zhist")=42
	Set gtmtypes("rhdtyp",43,"name")="rhdtyp.objhash"
	Set gtmtypes("rhdtyp",43,"off")=272
	Set gtmtypes("rhdtyp",43,"len")=8
	Set gtmtypes("rhdtyp",43,"type")="uint64_t"
	Set gtmtypfldindx("rhdtyp","objhash")=43
	Set gtmtypes("rhdtyp",44,"name")="rhdtyp.lbltext_ptr"
	Set gtmtypes("rhdtyp",44,"off")=280
	Set gtmtypes("rhdtyp",44,"len")=8
	Set gtmtypes("rhdtyp",44,"type")="addr"
	Set gtmtypfldindx("rhdtyp","lbltext_ptr")=44
	Set gtmtypes("rhdtyp",45,"name")="rhdtyp.object_len"
	Set gtmtypes("rhdtyp",45,"off")=288
	Set gtmtypes("rhdtyp",45,"len")=4
	Set gtmtypes("rhdtyp",45,"type")="unsigned-int"
	Set gtmtypfldindx("rhdtyp","object_len")=45
	Set gtmtypes("rhdtyp",46,"name")="rhdtyp.routine_source_offset"
	Set gtmtypes("rhdtyp",46,"off")=292
	Set gtmtypes("rhdtyp",46,"len")=4
	Set gtmtypes("rhdtyp",46,"type")="unsigned-int"
	Set gtmtypfldindx("rhdtyp","routine_source_offset")=46
	Set gtmtypes("rhdtyp",47,"name")="rhdtyp.linkage_names"
	Set gtmtypes("rhdtyp",47,"off")=296
	Set gtmtypes("rhdtyp",47,"len")=8
	Set gtmtypes("rhdtyp",47,"type")="addr"
	Set gtmtypfldindx("rhdtyp","linkage_names")=47
	Set gtmtypes("rhdtyp",48,"name")="rhdtyp.relinkctl_bkptr"
	Set gtmtypes("rhdtyp",48,"off")=304
	Set gtmtypes("rhdtyp",48,"len")=8
	Set gtmtypes("rhdtyp",48,"type")="addr"
	Set gtmtypfldindx("rhdtyp","relinkctl_bkptr")=48
	;
	Set gtmtypes("routine_source")="struct"
	Set gtmtypes("routine_source",0)=7
	Set gtmtypes("routine_source","len")=32
	Set gtmtypes("routine_source",1,"name")="routine_source.srcrecs"
	Set gtmtypes("routine_source",1,"off")=0
	Set gtmtypes("routine_source",1,"len")=4
	Set gtmtypes("routine_source",1,"type")="int"
	Set gtmtypfldindx("routine_source","srcrecs")=1
	Set gtmtypes("routine_source",2,"name")="routine_source.srcstat"
	Set gtmtypes("routine_source",2,"off")=4
	Set gtmtypes("routine_source",2,"len")=4
	Set gtmtypes("routine_source",2,"type")="unsigned-int"
	Set gtmtypfldindx("routine_source","srcstat")=2
	Set gtmtypes("routine_source",3,"name")="routine_source.srcbuff"
	Set gtmtypes("routine_source",3,"off")=8
	Set gtmtypes("routine_source",3,"len")=8
	Set gtmtypes("routine_source",3,"type")="addr"
	Set gtmtypfldindx("routine_source","srcbuff")=3
	Set gtmtypes("routine_source",4,"name")="routine_source.srclines"
	Set gtmtypes("routine_source",4,"off")=16
	Set gtmtypes("routine_source",4,"len")=16
	Set gtmtypes("routine_source",4,"type")="mstr"
	Set gtmtypfldindx("routine_source","srclines")=4
	Set gtmtypes("routine_source",5,"name")="routine_source.srclines[0].char_len"
	Set gtmtypes("routine_source",5,"off")=16
	Set gtmtypes("routine_source",5,"len")=4
	Set gtmtypes("routine_source",5,"type")="unsigned-int"
	Set gtmtypfldindx("routine_source","srclines[0].char_len")=5
	Set gtmtypes("routine_source",6,"name")="routine_source.srclines[0].len"
	Set gtmtypes("routine_source",6,"off")=20
	Set gtmtypes("routine_source",6,"len")=4
	Set gtmtypes("routine_source",6,"type")="int"
	Set gtmtypfldindx("routine_source","srclines[0].len")=6
	Set gtmtypes("routine_source",7,"name")="routine_source.srclines[0].addr"
	Set gtmtypes("routine_source",7,"off")=24
	Set gtmtypes("routine_source",7,"len")=8
	Set gtmtypes("routine_source",7,"type")="addr"
	Set gtmtypfldindx("routine_source","srclines[0].addr")=7
	;
	Set gtmtypes("rtn_tabent")="struct"
	Set gtmtypes("rtn_tabent",0)=5
	Set gtmtypes("rtn_tabent","len")=24
	Set gtmtypes("rtn_tabent",1,"name")="rtn_tabent.rt_name"
	Set gtmtypes("rtn_tabent",1,"off")=0
	Set gtmtypes("rtn_tabent",1,"len")=16
	Set gtmtypes("rtn_tabent",1,"type")="mstr"
	Set gtmtypfldindx("rtn_tabent","rt_name")=1
	Set gtmtypes("rtn_tabent",2,"name")="rtn_tabent.rt_name.char_len"
	Set gtmtypes("rtn_tabent",2,"off")=0
	Set gtmtypes("rtn_tabent",2,"len")=4
	Set gtmtypes("rtn_tabent",2,"type")="unsigned-int"
	Set gtmtypfldindx("rtn_tabent","rt_name.char_len")=2
	Set gtmtypes("rtn_tabent",3,"name")="rtn_tabent.rt_name.len"
	Set gtmtypes("rtn_tabent",3,"off")=4
	Set gtmtypes("rtn_tabent",3,"len")=4
	Set gtmtypes("rtn_tabent",3,"type")="int"
	Set gtmtypfldindx("rtn_tabent","rt_name.len")=3
	Set gtmtypes("rtn_tabent",4,"name")="rtn_tabent.rt_name.addr"
	Set gtmtypes("rtn_tabent",4,"off")=8
	Set gtmtypes("rtn_tabent",4,"len")=8
	Set gtmtypes("rtn_tabent",4,"type")="addr"
	Set gtmtypfldindx("rtn_tabent","rt_name.addr")=4
	Set gtmtypes("rtn_tabent",5,"name")="rtn_tabent.rt_adr"
	Set gtmtypes("rtn_tabent",5,"off")=16
	Set gtmtypes("rtn_tabent",5,"len")=8
	Set gtmtypes("rtn_tabent",5,"type")="addr"
	Set gtmtypfldindx("rtn_tabent","rt_adr")=5
	;
	Set gtmtypes("rtnobj_hdr_t")="struct"
	Set gtmtypes("rtnobj_hdr_t",0)=13
	Set gtmtypes("rtnobj_hdr_t","len")=48
	Set gtmtypes("rtnobj_hdr_t",1,"name")="rtnobj_hdr_t.queueIndex"
	Set gtmtypes("rtnobj_hdr_t",1,"off")=0
	Set gtmtypes("rtnobj_hdr_t",1,"len")=2
	Set gtmtypes("rtnobj_hdr_t",1,"type")="unsigned-short"
	Set gtmtypfldindx("rtnobj_hdr_t","queueIndex")=1
	Set gtmtypes("rtnobj_hdr_t",2,"name")="rtnobj_hdr_t.state"
	Set gtmtypes("rtnobj_hdr_t",2,"off")=2
	Set gtmtypes("rtnobj_hdr_t",2,"len")=1
	Set gtmtypes("rtnobj_hdr_t",2,"type")="unsigned-char"
	Set gtmtypfldindx("rtnobj_hdr_t","state")=2
	Set gtmtypes("rtnobj_hdr_t",3,"name")="rtnobj_hdr_t.initialized"
	Set gtmtypes("rtnobj_hdr_t",3,"off")=3
	Set gtmtypes("rtnobj_hdr_t",3,"len")=1
	Set gtmtypes("rtnobj_hdr_t",3,"type")="unsigned-char"
	Set gtmtypfldindx("rtnobj_hdr_t","initialized")=3
	Set gtmtypes("rtnobj_hdr_t",4,"name")="rtnobj_hdr_t.refcnt"
	Set gtmtypes("rtnobj_hdr_t",4,"off")=4
	Set gtmtypes("rtnobj_hdr_t",4,"len")=4
	Set gtmtypes("rtnobj_hdr_t",4,"type")="int"
	Set gtmtypfldindx("rtnobj_hdr_t","refcnt")=4
	Set gtmtypes("rtnobj_hdr_t",5,"name")="rtnobj_hdr_t.objhash"
	Set gtmtypes("rtnobj_hdr_t",5,"off")=8
	Set gtmtypes("rtnobj_hdr_t",5,"len")=8
	Set gtmtypes("rtnobj_hdr_t",5,"type")="uint64_t"
	Set gtmtypfldindx("rtnobj_hdr_t","objhash")=5
	Set gtmtypes("rtnobj_hdr_t",6,"name")="rtnobj_hdr_t.next_rtnobj_shm_offset"
	Set gtmtypes("rtnobj_hdr_t",6,"off")=16
	Set gtmtypes("rtnobj_hdr_t",6,"len")=8
	Set gtmtypes("rtnobj_hdr_t",6,"type")="uint64_t"
	Set gtmtypfldindx("rtnobj_hdr_t","next_rtnobj_shm_offset")=6
	Set gtmtypes("rtnobj_hdr_t",7,"name")="rtnobj_hdr_t.relinkctl_index"
	Set gtmtypes("rtnobj_hdr_t",7,"off")=24
	Set gtmtypes("rtnobj_hdr_t",7,"len")=4
	Set gtmtypes("rtnobj_hdr_t",7,"type")="unsigned-int"
	Set gtmtypfldindx("rtnobj_hdr_t","relinkctl_index")=7
	Set gtmtypes("rtnobj_hdr_t",8,"name")="rtnobj_hdr_t.objLen"
	Set gtmtypes("rtnobj_hdr_t",8,"off")=28
	Set gtmtypes("rtnobj_hdr_t",8,"len")=4
	Set gtmtypes("rtnobj_hdr_t",8,"type")="unsigned-int"
	Set gtmtypfldindx("rtnobj_hdr_t","objLen")=8
	Set gtmtypes("rtnobj_hdr_t",9,"name")="rtnobj_hdr_t.userStorage"
	Set gtmtypes("rtnobj_hdr_t",9,"off")=32
	Set gtmtypes("rtnobj_hdr_t",9,"len")=16
	Set gtmtypes("rtnobj_hdr_t",9,"type")="union"
	Set gtmtypfldindx("rtnobj_hdr_t","userStorage")=9
	Set gtmtypes("rtnobj_hdr_t",10,"name")="rtnobj_hdr_t.userStorage.freePtr"
	Set gtmtypes("rtnobj_hdr_t",10,"off")=32
	Set gtmtypes("rtnobj_hdr_t",10,"len")=16
	Set gtmtypes("rtnobj_hdr_t",10,"type")="que_ent"
	Set gtmtypfldindx("rtnobj_hdr_t","userStorage.freePtr")=10
	Set gtmtypes("rtnobj_hdr_t",11,"name")="rtnobj_hdr_t.userStorage.freePtr.fl"
	Set gtmtypes("rtnobj_hdr_t",11,"off")=32
	Set gtmtypes("rtnobj_hdr_t",11,"len")=8
	Set gtmtypes("rtnobj_hdr_t",11,"type")="intptr_t"
	Set gtmtypfldindx("rtnobj_hdr_t","userStorage.freePtr.fl")=11
	Set gtmtypes("rtnobj_hdr_t",12,"name")="rtnobj_hdr_t.userStorage.freePtr.bl"
	Set gtmtypes("rtnobj_hdr_t",12,"off")=40
	Set gtmtypes("rtnobj_hdr_t",12,"len")=8
	Set gtmtypes("rtnobj_hdr_t",12,"type")="intptr_t"
	Set gtmtypfldindx("rtnobj_hdr_t","userStorage.freePtr.bl")=12
	Set gtmtypes("rtnobj_hdr_t",13,"name")="rtnobj_hdr_t.userStorage.userStart"
	Set gtmtypes("rtnobj_hdr_t",13,"off")=32
	Set gtmtypes("rtnobj_hdr_t",13,"len")=1
	Set gtmtypes("rtnobj_hdr_t",13,"type")="unsigned-char"
	Set gtmtypfldindx("rtnobj_hdr_t","userStorage.userStart")=13
	;
	Set gtmtypes("rtnobjshm_hdr_t")="struct"
	Set gtmtypes("rtnobjshm_hdr_t",0)=9
	Set gtmtypes("rtnobjshm_hdr_t","len")=856
	Set gtmtypes("rtnobjshm_hdr_t",1,"name")="rtnobjshm_hdr_t.freeList"
	Set gtmtypes("rtnobjshm_hdr_t",1,"off")=0
	Set gtmtypes("rtnobjshm_hdr_t",1,"len")=816
	Set gtmtypes("rtnobjshm_hdr_t",1,"type")="que_ent"
	Set gtmtypfldindx("rtnobjshm_hdr_t","freeList")=1
	Set gtmtypes("rtnobjshm_hdr_t",1,"dim")=51
	Set gtmtypes("rtnobjshm_hdr_t",2,"name")="rtnobjshm_hdr_t.freeList[0].fl"
	Set gtmtypes("rtnobjshm_hdr_t",2,"off")=0
	Set gtmtypes("rtnobjshm_hdr_t",2,"len")=8
	Set gtmtypes("rtnobjshm_hdr_t",2,"type")="intptr_t"
	Set gtmtypfldindx("rtnobjshm_hdr_t","freeList[0].fl")=2
	Set gtmtypes("rtnobjshm_hdr_t",3,"name")="rtnobjshm_hdr_t.freeList[0].bl"
	Set gtmtypes("rtnobjshm_hdr_t",3,"off")=8
	Set gtmtypes("rtnobjshm_hdr_t",3,"len")=8
	Set gtmtypes("rtnobjshm_hdr_t",3,"type")="intptr_t"
	Set gtmtypfldindx("rtnobjshm_hdr_t","freeList[0].bl")=3
	Set gtmtypes("rtnobjshm_hdr_t",4,"name")="rtnobjshm_hdr_t.rtnobj_min_free_index"
	Set gtmtypes("rtnobjshm_hdr_t",4,"off")=816
	Set gtmtypes("rtnobjshm_hdr_t",4,"len")=4
	Set gtmtypes("rtnobjshm_hdr_t",4,"type")="int"
	Set gtmtypfldindx("rtnobjshm_hdr_t","rtnobj_min_free_index")=4
	Set gtmtypes("rtnobjshm_hdr_t",5,"name")="rtnobjshm_hdr_t.rtnobj_max_free_index"
	Set gtmtypes("rtnobjshm_hdr_t",5,"off")=820
	Set gtmtypes("rtnobjshm_hdr_t",5,"len")=4
	Set gtmtypes("rtnobjshm_hdr_t",5,"type")="int"
	Set gtmtypfldindx("rtnobjshm_hdr_t","rtnobj_max_free_index")=5
	Set gtmtypes("rtnobjshm_hdr_t",6,"name")="rtnobjshm_hdr_t.rtnobj_shmid"
	Set gtmtypes("rtnobjshm_hdr_t",6,"off")=824
	Set gtmtypes("rtnobjshm_hdr_t",6,"len")=4
	Set gtmtypes("rtnobjshm_hdr_t",6,"type")="int"
	Set gtmtypfldindx("rtnobjshm_hdr_t","rtnobj_shmid")=6
	Set gtmtypes("rtnobjshm_hdr_t",7,"name")="rtnobjshm_hdr_t.real_len"
	Set gtmtypes("rtnobjshm_hdr_t",7,"off")=832
	Set gtmtypes("rtnobjshm_hdr_t",7,"len")=8
	Set gtmtypes("rtnobjshm_hdr_t",7,"type")="uint64_t"
	Set gtmtypfldindx("rtnobjshm_hdr_t","real_len")=7
	Set gtmtypes("rtnobjshm_hdr_t",8,"name")="rtnobjshm_hdr_t.used_len"
	Set gtmtypes("rtnobjshm_hdr_t",8,"off")=840
	Set gtmtypes("rtnobjshm_hdr_t",8,"len")=8
	Set gtmtypes("rtnobjshm_hdr_t",8,"type")="uint64_t"
	Set gtmtypfldindx("rtnobjshm_hdr_t","used_len")=8
	Set gtmtypes("rtnobjshm_hdr_t",9,"name")="rtnobjshm_hdr_t.shm_len"
	Set gtmtypes("rtnobjshm_hdr_t",9,"off")=848
	Set gtmtypes("rtnobjshm_hdr_t",9,"len")=8
	Set gtmtypes("rtnobjshm_hdr_t",9,"type")="uint64_t"
	Set gtmtypfldindx("rtnobjshm_hdr_t","shm_len")=9
	;
	Set gtmtypes("save_se")="struct"
	Set gtmtypes("save_se",0)=15
	Set gtmtypes("save_se","len")=160
	Set gtmtypes("save_se",1,"name")="save_se.expr_start"
	Set gtmtypes("save_se",1,"off")=0
	Set gtmtypes("save_se",1,"len")=8
	Set gtmtypes("save_se",1,"type")="addr"
	Set gtmtypfldindx("save_se","expr_start")=1
	Set gtmtypes("save_se",2,"name")="save_se.expr_start_orig"
	Set gtmtypes("save_se",2,"off")=8
	Set gtmtypes("save_se",2,"len")=8
	Set gtmtypes("save_se",2,"type")="addr"
	Set gtmtypfldindx("save_se","expr_start_orig")=2
	Set gtmtypes("save_se",3,"name")="save_se.shift_side_effects"
	Set gtmtypes("save_se",3,"off")=16
	Set gtmtypes("save_se",3,"len")=4
	Set gtmtypes("save_se",3,"type")="boolean_t"
	Set gtmtypfldindx("save_se","shift_side_effects")=3
	Set gtmtypes("save_se",4,"name")="save_se.saw_side_effect"
	Set gtmtypes("save_se",4,"off")=20
	Set gtmtypes("save_se",4,"len")=4
	Set gtmtypes("save_se",4,"type")="boolean_t"
	Set gtmtypfldindx("save_se","saw_side_effect")=4
	Set gtmtypes("save_se",5,"name")="save_se.tmpchain"
	Set gtmtypes("save_se",5,"off")=24
	Set gtmtypes("save_se",5,"len")=136
	Set gtmtypes("save_se",5,"type")="triple"
	Set gtmtypfldindx("save_se","tmpchain")=5
	Set gtmtypes("save_se",6,"name")="save_se.tmpchain.opcode"
	Set gtmtypes("save_se",6,"off")=24
	Set gtmtypes("save_se",6,"len")=4
	Set gtmtypes("save_se",6,"type")="unsigned-int"
	Set gtmtypfldindx("save_se","tmpchain.opcode")=6
	Set gtmtypes("save_se",7,"name")="save_se.tmpchain.exorder"
	Set gtmtypes("save_se",7,"off")=32
	Set gtmtypes("save_se",7,"len")=16
	Set gtmtypes("save_se",7,"type")="struct"
	Set gtmtypfldindx("save_se","tmpchain.exorder")=7
	Set gtmtypes("save_se",8,"name")="save_se.tmpchain.exorder.fl"
	Set gtmtypes("save_se",8,"off")=32
	Set gtmtypes("save_se",8,"len")=8
	Set gtmtypes("save_se",8,"type")="addr"
	Set gtmtypfldindx("save_se","tmpchain.exorder.fl")=8
	Set gtmtypes("save_se",9,"name")="save_se.tmpchain.exorder.bl"
	Set gtmtypes("save_se",9,"off")=40
	Set gtmtypes("save_se",9,"len")=8
	Set gtmtypes("save_se",9,"type")="addr"
	Set gtmtypfldindx("save_se","tmpchain.exorder.bl")=9
	Set gtmtypes("save_se",10,"name")="save_se.tmpchain.backptr"
	Set gtmtypes("save_se",10,"off")=48
	Set gtmtypes("save_se",10,"len")=24
	Set gtmtypes("save_se",10,"type")="tbp"
	Set gtmtypfldindx("save_se","tmpchain.backptr")=10
	Set gtmtypes("save_se",11,"name")="save_se.tmpchain.jmplist"
	Set gtmtypes("save_se",11,"off")=72
	Set gtmtypes("save_se",11,"len")=24
	Set gtmtypes("save_se",11,"type")="tbp"
	Set gtmtypfldindx("save_se","tmpchain.jmplist")=11
	Set gtmtypes("save_se",12,"name")="save_se.tmpchain.src"
	Set gtmtypes("save_se",12,"off")=96
	Set gtmtypes("save_se",12,"len")=8
	Set gtmtypes("save_se",12,"type")="source_address"
	Set gtmtypfldindx("save_se","tmpchain.src")=12
	Set gtmtypes("save_se",13,"name")="save_se.tmpchain.rtaddr"
	Set gtmtypes("save_se",13,"off")=104
	Set gtmtypes("save_se",13,"len")=4
	Set gtmtypes("save_se",13,"type")="int"
	Set gtmtypfldindx("save_se","tmpchain.rtaddr")=13
	Set gtmtypes("save_se",14,"name")="save_se.tmpchain.operand"
	Set gtmtypes("save_se",14,"off")=112
	Set gtmtypes("save_se",14,"len")=32
	Set gtmtypes("save_se",14,"type")="oprtype"
	Set gtmtypfldindx("save_se","tmpchain.operand")=14
	Set gtmtypes("save_se",14,"dim")=2
	Set gtmtypes("save_se",15,"name")="save_se.tmpchain.destination"
	Set gtmtypes("save_se",15,"off")=144
	Set gtmtypes("save_se",15,"len")=16
	Set gtmtypes("save_se",15,"type")="oprtype"
	Set gtmtypfldindx("save_se","tmpchain.destination")=15
	;
	Set gtmtypes("save_strct")="struct"
	Set gtmtypes("save_strct",0)=5
	Set gtmtypes("save_strct","len")=40
	Set gtmtypes("save_strct",1,"name")="save_strct.blk"
	Set gtmtypes("save_strct",1,"off")=0
	Set gtmtypes("save_strct",1,"len")=4
	Set gtmtypes("save_strct",1,"type")="int"
	Set gtmtypfldindx("save_strct","blk")=1
	Set gtmtypes("save_strct",2,"name")="save_strct.bp"
	Set gtmtypes("save_strct",2,"off")=8
	Set gtmtypes("save_strct",2,"len")=8
	Set gtmtypes("save_strct",2,"type")="addr"
	Set gtmtypfldindx("save_strct","bp")=2
	Set gtmtypes("save_strct",3,"name")="save_strct.region"
	Set gtmtypes("save_strct",3,"off")=16
	Set gtmtypes("save_strct",3,"len")=8
	Set gtmtypes("save_strct",3,"type")="addr"
	Set gtmtypfldindx("save_strct","region")=3
	Set gtmtypes("save_strct",4,"name")="save_strct.comment"
	Set gtmtypes("save_strct",4,"off")=24
	Set gtmtypes("save_strct",4,"len")=8
	Set gtmtypes("save_strct",4,"type")="addr"
	Set gtmtypfldindx("save_strct","comment")=4
	Set gtmtypes("save_strct",5,"name")="save_strct.ver"
	Set gtmtypes("save_strct",5,"off")=32
	Set gtmtypes("save_strct",5,"len")=2
	Set gtmtypes("save_strct",5,"type")="short"
	Set gtmtypfldindx("save_strct","ver")=5
	;
	Set gtmtypes("save_xfer_entry")="struct"
	Set gtmtypes("save_xfer_entry",0)=4
	Set gtmtypes("save_xfer_entry","len")=32
	Set gtmtypes("save_xfer_entry",1,"name")="save_xfer_entry.next"
	Set gtmtypes("save_xfer_entry",1,"off")=0
	Set gtmtypes("save_xfer_entry",1,"len")=8
	Set gtmtypes("save_xfer_entry",1,"type")="addr"
	Set gtmtypfldindx("save_xfer_entry","next")=1
	Set gtmtypes("save_xfer_entry",2,"name")="save_xfer_entry.outofband"
	Set gtmtypes("save_xfer_entry",2,"off")=8
	Set gtmtypes("save_xfer_entry",2,"len")=4
	Set gtmtypes("save_xfer_entry",2,"type")="int"
	Set gtmtypfldindx("save_xfer_entry","outofband")=2
	Set gtmtypes("save_xfer_entry",3,"name")="save_xfer_entry.set_fn"
	Set gtmtypes("save_xfer_entry",3,"off")=16
	Set gtmtypes("save_xfer_entry",3,"len")=8
	Set gtmtypes("save_xfer_entry",3,"type")="addr"
	Set gtmtypfldindx("save_xfer_entry","set_fn")=3
	Set gtmtypes("save_xfer_entry",4,"name")="save_xfer_entry.param_val"
	Set gtmtypes("save_xfer_entry",4,"off")=24
	Set gtmtypes("save_xfer_entry",4,"len")=4
	Set gtmtypes("save_xfer_entry",4,"type")="int"
	Set gtmtypfldindx("save_xfer_entry","param_val")=4
	;
	Set gtmtypes("select_list")="struct"
	Set gtmtypes("select_list",0)=5
	Set gtmtypes("select_list","len")=24
	Set gtmtypes("select_list",1,"name")="select_list.next"
	Set gtmtypes("select_list",1,"off")=0
	Set gtmtypes("select_list",1,"len")=8
	Set gtmtypes("select_list",1,"type")="addr"
	Set gtmtypfldindx("select_list","next")=1
	Set gtmtypes("select_list",2,"name")="select_list.buff"
	Set gtmtypes("select_list",2,"off")=8
	Set gtmtypes("select_list",2,"len")=8
	Set gtmtypes("select_list",2,"type")="addr"
	Set gtmtypfldindx("select_list","buff")=2
	Set gtmtypes("select_list",3,"name")="select_list.len"
	Set gtmtypes("select_list",3,"off")=16
	Set gtmtypes("select_list",3,"len")=2
	Set gtmtypes("select_list",3,"type")="short"
	Set gtmtypfldindx("select_list","len")=3
	Set gtmtypes("select_list",4,"name")="select_list.exclude"
	Set gtmtypes("select_list",4,"off")=18
	Set gtmtypes("select_list",4,"len")=1
	Set gtmtypes("select_list",4,"type")="char"
	Set gtmtypfldindx("select_list","exclude")=4
	Set gtmtypes("select_list",5,"name")="select_list.has_wildcard"
	Set gtmtypes("select_list",5,"off")=20
	Set gtmtypes("select_list",5,"len")=4
	Set gtmtypes("select_list",5,"type")="boolean_t"
	Set gtmtypfldindx("select_list","has_wildcard")=5
	;
	Set gtmtypes("sem_info")="struct"
	Set gtmtypes("sem_info",0)=2
	Set gtmtypes("sem_info","len")=8
	Set gtmtypes("sem_info",1,"name")="sem_info.ftok_key"
	Set gtmtypes("sem_info",1,"off")=0
	Set gtmtypes("sem_info",1,"len")=4
	Set gtmtypes("sem_info",1,"type")="int"
	Set gtmtypfldindx("sem_info","ftok_key")=1
	Set gtmtypes("sem_info",2,"name")="sem_info.sem_id"
	Set gtmtypes("sem_info",2,"off")=4
	Set gtmtypes("sem_info",2,"len")=4
	Set gtmtypes("sem_info",2,"type")="int"
	Set gtmtypfldindx("sem_info","sem_id")=2
	;
	Set gtmtypes("semid_queue_elem")="struct"
	Set gtmtypes("semid_queue_elem",0)=2
	Set gtmtypes("semid_queue_elem","len")=16
	Set gtmtypes("semid_queue_elem",1,"name")="semid_queue_elem.semid"
	Set gtmtypes("semid_queue_elem",1,"off")=0
	Set gtmtypes("semid_queue_elem",1,"len")=4
	Set gtmtypes("semid_queue_elem",1,"type")="int"
	Set gtmtypfldindx("semid_queue_elem","semid")=1
	Set gtmtypes("semid_queue_elem",2,"name")="semid_queue_elem.prev"
	Set gtmtypes("semid_queue_elem",2,"off")=8
	Set gtmtypes("semid_queue_elem",2,"len")=8
	Set gtmtypes("semid_queue_elem",2,"type")="addr"
	Set gtmtypfldindx("semid_queue_elem","prev")=2
	;
	Set gtmtypes("semwait_status_t")="struct"
	Set gtmtypes("semwait_status_t",0)=7
	Set gtmtypes("semwait_status_t","len")=40
	Set gtmtypes("semwait_status_t",1,"name")="semwait_status_t.line_no"
	Set gtmtypes("semwait_status_t",1,"off")=0
	Set gtmtypes("semwait_status_t",1,"len")=4
	Set gtmtypes("semwait_status_t",1,"type")="int"
	Set gtmtypfldindx("semwait_status_t","line_no")=1
	Set gtmtypes("semwait_status_t",2,"name")="semwait_status_t.save_errno"
	Set gtmtypes("semwait_status_t",2,"off")=4
	Set gtmtypes("semwait_status_t",2,"len")=4
	Set gtmtypes("semwait_status_t",2,"type")="int"
	Set gtmtypfldindx("semwait_status_t","save_errno")=2
	Set gtmtypes("semwait_status_t",3,"name")="semwait_status_t.status1"
	Set gtmtypes("semwait_status_t",3,"off")=8
	Set gtmtypes("semwait_status_t",3,"len")=4
	Set gtmtypes("semwait_status_t",3,"type")="int"
	Set gtmtypfldindx("semwait_status_t","status1")=3
	Set gtmtypes("semwait_status_t",4,"name")="semwait_status_t.status2"
	Set gtmtypes("semwait_status_t",4,"off")=12
	Set gtmtypes("semwait_status_t",4,"len")=4
	Set gtmtypes("semwait_status_t",4,"type")="int"
	Set gtmtypfldindx("semwait_status_t","status2")=4
	Set gtmtypes("semwait_status_t",5,"name")="semwait_status_t.sem_pid"
	Set gtmtypes("semwait_status_t",5,"off")=16
	Set gtmtypes("semwait_status_t",5,"len")=4
	Set gtmtypes("semwait_status_t",5,"type")="int"
	Set gtmtypfldindx("semwait_status_t","sem_pid")=5
	Set gtmtypes("semwait_status_t",6,"name")="semwait_status_t.module"
	Set gtmtypes("semwait_status_t",6,"off")=24
	Set gtmtypes("semwait_status_t",6,"len")=8
	Set gtmtypes("semwait_status_t",6,"type")="addr"
	Set gtmtypfldindx("semwait_status_t","module")=6
	Set gtmtypes("semwait_status_t",7,"name")="semwait_status_t.op"
	Set gtmtypes("semwait_status_t",7,"off")=32
	Set gtmtypes("semwait_status_t",7,"len")=4
	Set gtmtypes("semwait_status_t",7,"type")="int"
	Set gtmtypfldindx("semwait_status_t","op")=7
	;
	Set gtmtypes("set_jnl_options")="struct"
	Set gtmtypes("set_jnl_options",0)=16
	Set gtmtypes("set_jnl_options","len")=64
	Set gtmtypes("set_jnl_options",1,"name")="set_jnl_options.cli_journal"
	Set gtmtypes("set_jnl_options",1,"off")=0
	Set gtmtypes("set_jnl_options",1,"len")=4
	Set gtmtypes("set_jnl_options",1,"type")="int"
	Set gtmtypfldindx("set_jnl_options","cli_journal")=1
	Set gtmtypes("set_jnl_options",2,"name")="set_jnl_options.cli_enable"
	Set gtmtypes("set_jnl_options",2,"off")=4
	Set gtmtypes("set_jnl_options",2,"len")=4
	Set gtmtypes("set_jnl_options",2,"type")="int"
	Set gtmtypfldindx("set_jnl_options","cli_enable")=2
	Set gtmtypes("set_jnl_options",3,"name")="set_jnl_options.cli_on"
	Set gtmtypes("set_jnl_options",3,"off")=8
	Set gtmtypes("set_jnl_options",3,"len")=4
	Set gtmtypes("set_jnl_options",3,"type")="int"
	Set gtmtypfldindx("set_jnl_options","cli_on")=3
	Set gtmtypes("set_jnl_options",4,"name")="set_jnl_options.cli_replic_on"
	Set gtmtypes("set_jnl_options",4,"off")=12
	Set gtmtypes("set_jnl_options",4,"len")=4
	Set gtmtypes("set_jnl_options",4,"type")="int"
	Set gtmtypfldindx("set_jnl_options","cli_replic_on")=4
	Set gtmtypes("set_jnl_options",5,"name")="set_jnl_options.alignsize_specified"
	Set gtmtypes("set_jnl_options",5,"off")=16
	Set gtmtypes("set_jnl_options",5,"len")=4
	Set gtmtypes("set_jnl_options",5,"type")="boolean_t"
	Set gtmtypfldindx("set_jnl_options","alignsize_specified")=5
	Set gtmtypes("set_jnl_options",6,"name")="set_jnl_options.allocation_specified"
	Set gtmtypes("set_jnl_options",6,"off")=20
	Set gtmtypes("set_jnl_options",6,"len")=4
	Set gtmtypes("set_jnl_options",6,"type")="boolean_t"
	Set gtmtypfldindx("set_jnl_options","allocation_specified")=6
	Set gtmtypes("set_jnl_options",7,"name")="set_jnl_options.autoswitchlimit_specified"
	Set gtmtypes("set_jnl_options",7,"off")=24
	Set gtmtypes("set_jnl_options",7,"len")=4
	Set gtmtypes("set_jnl_options",7,"type")="boolean_t"
	Set gtmtypfldindx("set_jnl_options","autoswitchlimit_specified")=7
	Set gtmtypes("set_jnl_options",8,"name")="set_jnl_options.image_type_specified"
	Set gtmtypes("set_jnl_options",8,"off")=28
	Set gtmtypes("set_jnl_options",8,"len")=4
	Set gtmtypes("set_jnl_options",8,"type")="boolean_t"
	Set gtmtypfldindx("set_jnl_options","image_type_specified")=8
	Set gtmtypes("set_jnl_options",9,"name")="set_jnl_options.buffer_size_specified"
	Set gtmtypes("set_jnl_options",9,"off")=32
	Set gtmtypes("set_jnl_options",9,"len")=4
	Set gtmtypes("set_jnl_options",9,"type")="boolean_t"
	Set gtmtypfldindx("set_jnl_options","buffer_size_specified")=9
	Set gtmtypes("set_jnl_options",10,"name")="set_jnl_options.epoch_interval_specified"
	Set gtmtypes("set_jnl_options",10,"off")=36
	Set gtmtypes("set_jnl_options",10,"len")=4
	Set gtmtypes("set_jnl_options",10,"type")="boolean_t"
	Set gtmtypfldindx("set_jnl_options","epoch_interval_specified")=10
	Set gtmtypes("set_jnl_options",11,"name")="set_jnl_options.extension_specified"
	Set gtmtypes("set_jnl_options",11,"off")=40
	Set gtmtypes("set_jnl_options",11,"len")=4
	Set gtmtypes("set_jnl_options",11,"type")="boolean_t"
	Set gtmtypfldindx("set_jnl_options","extension_specified")=11
	Set gtmtypes("set_jnl_options",12,"name")="set_jnl_options.filename_specified"
	Set gtmtypes("set_jnl_options",12,"off")=44
	Set gtmtypes("set_jnl_options",12,"len")=4
	Set gtmtypes("set_jnl_options",12,"type")="boolean_t"
	Set gtmtypfldindx("set_jnl_options","filename_specified")=12
	Set gtmtypes("set_jnl_options",13,"name")="set_jnl_options.sync_io_specified"
	Set gtmtypes("set_jnl_options",13,"off")=48
	Set gtmtypes("set_jnl_options",13,"len")=4
	Set gtmtypes("set_jnl_options",13,"type")="boolean_t"
	Set gtmtypfldindx("set_jnl_options","sync_io_specified")=13
	Set gtmtypes("set_jnl_options",14,"name")="set_jnl_options.yield_limit_specified"
	Set gtmtypes("set_jnl_options",14,"off")=52
	Set gtmtypes("set_jnl_options",14,"len")=4
	Set gtmtypes("set_jnl_options",14,"type")="boolean_t"
	Set gtmtypfldindx("set_jnl_options","yield_limit_specified")=14
	Set gtmtypes("set_jnl_options",15,"name")="set_jnl_options.sync_io"
	Set gtmtypes("set_jnl_options",15,"off")=56
	Set gtmtypes("set_jnl_options",15,"len")=4
	Set gtmtypes("set_jnl_options",15,"type")="boolean_t"
	Set gtmtypfldindx("set_jnl_options","sync_io")=15
	Set gtmtypes("set_jnl_options",16,"name")="set_jnl_options.yield_limit"
	Set gtmtypes("set_jnl_options",16,"off")=60
	Set gtmtypes("set_jnl_options",16,"len")=4
	Set gtmtypes("set_jnl_options",16,"type")="int"
	Set gtmtypfldindx("set_jnl_options","yield_limit")=16
	;
	Set gtmtypes("sgbg_addrs")="struct"
	Set gtmtypes("sgbg_addrs",0)=1
	Set gtmtypes("sgbg_addrs","len")=8
	Set gtmtypes("sgbg_addrs",1,"name")="sgbg_addrs.cache_state"
	Set gtmtypes("sgbg_addrs",1,"off")=0
	Set gtmtypes("sgbg_addrs",1,"len")=8
	Set gtmtypes("sgbg_addrs",1,"type")="addr"
	Set gtmtypfldindx("sgbg_addrs","cache_state")=1
	;
	Set gtmtypes("sgm_info")="struct"
	Set gtmtypes("sgm_info",0)=40
	Set gtmtypes("sgm_info","len")=272
	Set gtmtypes("sgm_info",1,"name")="sgm_info.next_sgm_info"
	Set gtmtypes("sgm_info",1,"off")=0
	Set gtmtypes("sgm_info",1,"len")=8
	Set gtmtypes("sgm_info",1,"type")="addr"
	Set gtmtypfldindx("sgm_info","next_sgm_info")=1
	Set gtmtypes("sgm_info",2,"name")="sgm_info.next_tp_si_by_ftok"
	Set gtmtypes("sgm_info",2,"off")=8
	Set gtmtypes("sgm_info",2,"len")=8
	Set gtmtypes("sgm_info",2,"type")="addr"
	Set gtmtypfldindx("sgm_info","next_tp_si_by_ftok")=2
	Set gtmtypes("sgm_info",3,"name")="sgm_info.first_tp_hist"
	Set gtmtypes("sgm_info",3,"off")=16
	Set gtmtypes("sgm_info",3,"len")=8
	Set gtmtypes("sgm_info",3,"type")="addr"
	Set gtmtypfldindx("sgm_info","first_tp_hist")=3
	Set gtmtypes("sgm_info",4,"name")="sgm_info.last_tp_hist"
	Set gtmtypes("sgm_info",4,"off")=24
	Set gtmtypes("sgm_info",4,"len")=8
	Set gtmtypes("sgm_info",4,"type")="addr"
	Set gtmtypfldindx("sgm_info","last_tp_hist")=4
	Set gtmtypes("sgm_info",5,"name")="sgm_info.blks_in_use"
	Set gtmtypes("sgm_info",5,"off")=32
	Set gtmtypes("sgm_info",5,"len")=8
	Set gtmtypes("sgm_info",5,"type")="addr"
	Set gtmtypfldindx("sgm_info","blks_in_use")=5
	Set gtmtypes("sgm_info",6,"name")="sgm_info.start_tn"
	Set gtmtypes("sgm_info",6,"off")=40
	Set gtmtypes("sgm_info",6,"len")=8
	Set gtmtypes("sgm_info",6,"type")="uint64_t"
	Set gtmtypfldindx("sgm_info","start_tn")=6
	Set gtmtypes("sgm_info",7,"name")="sgm_info.gv_cur_region"
	Set gtmtypes("sgm_info",7,"off")=48
	Set gtmtypes("sgm_info",7,"len")=8
	Set gtmtypes("sgm_info",7,"type")="addr"
	Set gtmtypfldindx("sgm_info","gv_cur_region")=7
	Set gtmtypes("sgm_info",8,"name")="sgm_info.update_trans"
	Set gtmtypes("sgm_info",8,"off")=56
	Set gtmtypes("sgm_info",8,"len")=4
	Set gtmtypes("sgm_info",8,"type")="unsigned-int"
	Set gtmtypfldindx("sgm_info","update_trans")=8
	Set gtmtypes("sgm_info",9,"name")="sgm_info.first_cw_set"
	Set gtmtypes("sgm_info",9,"off")=64
	Set gtmtypes("sgm_info",9,"len")=8
	Set gtmtypes("sgm_info",9,"type")="addr"
	Set gtmtypfldindx("sgm_info","first_cw_set")=9
	Set gtmtypes("sgm_info",10,"name")="sgm_info.last_cw_set"
	Set gtmtypes("sgm_info",10,"off")=72
	Set gtmtypes("sgm_info",10,"len")=8
	Set gtmtypes("sgm_info",10,"type")="addr"
	Set gtmtypfldindx("sgm_info","last_cw_set")=10
	Set gtmtypes("sgm_info",11,"name")="sgm_info.first_cw_bitmap"
	Set gtmtypes("sgm_info",11,"off")=80
	Set gtmtypes("sgm_info",11,"len")=8
	Set gtmtypes("sgm_info",11,"type")="addr"
	Set gtmtypfldindx("sgm_info","first_cw_bitmap")=11
	Set gtmtypes("sgm_info",12,"name")="sgm_info.cw_set_list"
	Set gtmtypes("sgm_info",12,"off")=88
	Set gtmtypes("sgm_info",12,"len")=8
	Set gtmtypes("sgm_info",12,"type")="addr"
	Set gtmtypfldindx("sgm_info","cw_set_list")=12
	Set gtmtypes("sgm_info",13,"name")="sgm_info.tlvl_cw_set_list"
	Set gtmtypes("sgm_info",13,"off")=96
	Set gtmtypes("sgm_info",13,"len")=8
	Set gtmtypes("sgm_info",13,"type")="addr"
	Set gtmtypfldindx("sgm_info","tlvl_cw_set_list")=13
	Set gtmtypes("sgm_info",14,"name")="sgm_info.new_buff_list"
	Set gtmtypes("sgm_info",14,"off")=104
	Set gtmtypes("sgm_info",14,"len")=8
	Set gtmtypes("sgm_info",14,"type")="addr"
	Set gtmtypfldindx("sgm_info","new_buff_list")=14
	Set gtmtypes("sgm_info",15,"name")="sgm_info.recompute_list"
	Set gtmtypes("sgm_info",15,"off")=112
	Set gtmtypes("sgm_info",15,"len")=8
	Set gtmtypes("sgm_info",15,"type")="addr"
	Set gtmtypfldindx("sgm_info","recompute_list")=15
	Set gtmtypes("sgm_info",16,"name")="sgm_info.tlvl_info_list"
	Set gtmtypes("sgm_info",16,"off")=120
	Set gtmtypes("sgm_info",16,"len")=8
	Set gtmtypes("sgm_info",16,"type")="addr"
	Set gtmtypfldindx("sgm_info","tlvl_info_list")=16
	Set gtmtypes("sgm_info",17,"name")="sgm_info.cr_array"
	Set gtmtypes("sgm_info",17,"off")=128
	Set gtmtypes("sgm_info",17,"len")=8
	Set gtmtypes("sgm_info",17,"type")="addr"
	Set gtmtypfldindx("sgm_info","cr_array")=17
	Set gtmtypes("sgm_info",18,"name")="sgm_info.tp_csd"
	Set gtmtypes("sgm_info",18,"off")=136
	Set gtmtypes("sgm_info",18,"len")=8
	Set gtmtypes("sgm_info",18,"type")="addr"
	Set gtmtypfldindx("sgm_info","tp_csd")=18
	Set gtmtypes("sgm_info",19,"name")="sgm_info.tp_csa"
	Set gtmtypes("sgm_info",19,"off")=144
	Set gtmtypes("sgm_info",19,"len")=8
	Set gtmtypes("sgm_info",19,"type")="addr"
	Set gtmtypfldindx("sgm_info","tp_csa")=19
	Set gtmtypes("sgm_info",20,"name")="sgm_info.kill_set_head"
	Set gtmtypes("sgm_info",20,"off")=152
	Set gtmtypes("sgm_info",20,"len")=8
	Set gtmtypes("sgm_info",20,"type")="addr"
	Set gtmtypfldindx("sgm_info","kill_set_head")=20
	Set gtmtypes("sgm_info",21,"name")="sgm_info.kill_set_tail"
	Set gtmtypes("sgm_info",21,"off")=160
	Set gtmtypes("sgm_info",21,"len")=8
	Set gtmtypes("sgm_info",21,"type")="addr"
	Set gtmtypfldindx("sgm_info","kill_set_tail")=21
	Set gtmtypes("sgm_info",22,"name")="sgm_info.tlvl_info_head"
	Set gtmtypes("sgm_info",22,"off")=168
	Set gtmtypes("sgm_info",22,"len")=8
	Set gtmtypes("sgm_info",22,"type")="addr"
	Set gtmtypfldindx("sgm_info","tlvl_info_head")=22
	Set gtmtypes("sgm_info",23,"name")="sgm_info.jnl_head"
	Set gtmtypes("sgm_info",23,"off")=176
	Set gtmtypes("sgm_info",23,"len")=8
	Set gtmtypes("sgm_info",23,"type")="addr"
	Set gtmtypfldindx("sgm_info","jnl_head")=23
	Set gtmtypes("sgm_info",24,"name")="sgm_info.jnl_tail"
	Set gtmtypes("sgm_info",24,"off")=184
	Set gtmtypes("sgm_info",24,"len")=8
	Set gtmtypes("sgm_info",24,"type")="addr"
	Set gtmtypfldindx("sgm_info","jnl_tail")=24
	Set gtmtypes("sgm_info",25,"name")="sgm_info.format_buff_list"
	Set gtmtypes("sgm_info",25,"off")=192
	Set gtmtypes("sgm_info",25,"len")=8
	Set gtmtypes("sgm_info",25,"type")="addr"
	Set gtmtypfldindx("sgm_info","format_buff_list")=25
	Set gtmtypes("sgm_info",26,"name")="sgm_info.jnl_list"
	Set gtmtypes("sgm_info",26,"off")=200
	Set gtmtypes("sgm_info",26,"len")=8
	Set gtmtypes("sgm_info",26,"type")="addr"
	Set gtmtypfldindx("sgm_info","jnl_list")=26
	Set gtmtypes("sgm_info",27,"name")="sgm_info.cw_set_depth"
	Set gtmtypes("sgm_info",27,"off")=208
	Set gtmtypes("sgm_info",27,"len")=4
	Set gtmtypes("sgm_info",27,"type")="int"
	Set gtmtypfldindx("sgm_info","cw_set_depth")=27
	Set gtmtypes("sgm_info",28,"name")="sgm_info.cr_array_index"
	Set gtmtypes("sgm_info",28,"off")=212
	Set gtmtypes("sgm_info",28,"len")=4
	Set gtmtypes("sgm_info",28,"type")="int"
	Set gtmtypfldindx("sgm_info","cr_array_index")=28
	Set gtmtypes("sgm_info",29,"name")="sgm_info.num_of_blks"
	Set gtmtypes("sgm_info",29,"off")=216
	Set gtmtypes("sgm_info",29,"len")=4
	Set gtmtypes("sgm_info",29,"type")="int"
	Set gtmtypfldindx("sgm_info","num_of_blks")=29
	Set gtmtypes("sgm_info",30,"name")="sgm_info.tp_hist_size"
	Set gtmtypes("sgm_info",30,"off")=220
	Set gtmtypes("sgm_info",30,"len")=4
	Set gtmtypes("sgm_info",30,"type")="int"
	Set gtmtypfldindx("sgm_info","tp_hist_size")=30
	Set gtmtypes("sgm_info",31,"name")="sgm_info.cur_tp_hist_size"
	Set gtmtypes("sgm_info",31,"off")=224
	Set gtmtypes("sgm_info",31,"len")=4
	Set gtmtypes("sgm_info",31,"type")="int"
	Set gtmtypfldindx("sgm_info","cur_tp_hist_size")=31
	Set gtmtypes("sgm_info",32,"name")="sgm_info.total_jnl_rec_size"
	Set gtmtypes("sgm_info",32,"off")=228
	Set gtmtypes("sgm_info",32,"len")=4
	Set gtmtypes("sgm_info",32,"type")="int"
	Set gtmtypfldindx("sgm_info","total_jnl_rec_size")=32
	Set gtmtypes("sgm_info",33,"name")="sgm_info.cr_array_size"
	Set gtmtypes("sgm_info",33,"off")=232
	Set gtmtypes("sgm_info",33,"len")=4
	Set gtmtypes("sgm_info",33,"type")="int"
	Set gtmtypfldindx("sgm_info","cr_array_size")=33
	Set gtmtypes("sgm_info",34,"name")="sgm_info.tp_set_sgm_done"
	Set gtmtypes("sgm_info",34,"off")=236
	Set gtmtypes("sgm_info",34,"len")=4
	Set gtmtypes("sgm_info",34,"type")="boolean_t"
	Set gtmtypfldindx("sgm_info","tp_set_sgm_done")=34
	Set gtmtypes("sgm_info",35,"name")="sgm_info.crash_count"
	Set gtmtypes("sgm_info",35,"off")=240
	Set gtmtypes("sgm_info",35,"len")=4
	Set gtmtypes("sgm_info",35,"type")="int"
	Set gtmtypfldindx("sgm_info","crash_count")=35
	Set gtmtypes("sgm_info",36,"name")="sgm_info.backup_block_saved"
	Set gtmtypes("sgm_info",36,"off")=244
	Set gtmtypes("sgm_info",36,"len")=4
	Set gtmtypes("sgm_info",36,"type")="boolean_t"
	Set gtmtypfldindx("sgm_info","backup_block_saved")=36
	Set gtmtypes("sgm_info",37,"name")="sgm_info.kip_csa"
	Set gtmtypes("sgm_info",37,"off")=248
	Set gtmtypes("sgm_info",37,"len")=8
	Set gtmtypes("sgm_info",37,"type")="addr"
	Set gtmtypfldindx("sgm_info","kip_csa")=37
	Set gtmtypes("sgm_info",38,"name")="sgm_info.tmp_cw_set_depth"
	Set gtmtypes("sgm_info",38,"off")=256
	Set gtmtypes("sgm_info",38,"len")=4
	Set gtmtypes("sgm_info",38,"type")="int"
	Set gtmtypfldindx("sgm_info","tmp_cw_set_depth")=38
	Set gtmtypes("sgm_info",39,"name")="sgm_info.tot_jrec_size"
	Set gtmtypes("sgm_info",39,"off")=260
	Set gtmtypes("sgm_info",39,"len")=4
	Set gtmtypes("sgm_info",39,"type")="unsigned-int"
	Set gtmtypfldindx("sgm_info","tot_jrec_size")=39
	Set gtmtypes("sgm_info",40,"name")="sgm_info.jbuf_rsrv_ptr"
	Set gtmtypes("sgm_info",40,"off")=264
	Set gtmtypes("sgm_info",40,"len")=8
	Set gtmtypes("sgm_info",40,"type")="addr"
	Set gtmtypfldindx("sgm_info","jbuf_rsrv_ptr")=40
	;
	Set gtmtypes("sgmm_addrs")="struct"
	Set gtmtypes("sgmm_addrs",0)=1
	Set gtmtypes("sgmm_addrs","len")=4
	Set gtmtypes("sgmm_addrs",1,"name")="sgmm_addrs.filler"
	Set gtmtypes("sgmm_addrs",1,"off")=0
	Set gtmtypes("sgmm_addrs",1,"len")=4
	Set gtmtypes("sgmm_addrs",1,"type")="int"
	Set gtmtypfldindx("sgmm_addrs","filler")=1
	;
	Set gtmtypes("sgmnt_addrs")="struct"
	Set gtmtypes("sgmnt_addrs",0)=174
	Set gtmtypes("sgmnt_addrs","len")=1152
	Set gtmtypes("sgmnt_addrs",1,"name")="sgmnt_addrs.hdr"
	Set gtmtypes("sgmnt_addrs",1,"off")=0
	Set gtmtypes("sgmnt_addrs",1,"len")=8
	Set gtmtypes("sgmnt_addrs",1,"type")="addr"
	Set gtmtypfldindx("sgmnt_addrs","hdr")=1
	Set gtmtypes("sgmnt_addrs",2,"name")="sgmnt_addrs.bmm"
	Set gtmtypes("sgmnt_addrs",2,"off")=8
	Set gtmtypes("sgmnt_addrs",2,"len")=8
	Set gtmtypes("sgmnt_addrs",2,"type")="addr"
	Set gtmtypfldindx("sgmnt_addrs","bmm")=2
	Set gtmtypes("sgmnt_addrs",3,"name")="sgmnt_addrs.wc"
	Set gtmtypes("sgmnt_addrs",3,"off")=16
	Set gtmtypes("sgmnt_addrs",3,"len")=8
	Set gtmtypes("sgmnt_addrs",3,"type")="addr"
	Set gtmtypfldindx("sgmnt_addrs","wc")=3
	Set gtmtypes("sgmnt_addrs",4,"name")="sgmnt_addrs.bt_header"
	Set gtmtypes("sgmnt_addrs",4,"off")=24
	Set gtmtypes("sgmnt_addrs",4,"len")=8
	Set gtmtypes("sgmnt_addrs",4,"type")="addr"
	Set gtmtypfldindx("sgmnt_addrs","bt_header")=4
	Set gtmtypes("sgmnt_addrs",5,"name")="sgmnt_addrs.bt_base"
	Set gtmtypes("sgmnt_addrs",5,"off")=32
	Set gtmtypes("sgmnt_addrs",5,"len")=8
	Set gtmtypes("sgmnt_addrs",5,"type")="addr"
	Set gtmtypfldindx("sgmnt_addrs","bt_base")=5
	Set gtmtypes("sgmnt_addrs",6,"name")="sgmnt_addrs.th_base"
	Set gtmtypes("sgmnt_addrs",6,"off")=40
	Set gtmtypes("sgmnt_addrs",6,"len")=8
	Set gtmtypes("sgmnt_addrs",6,"type")="addr"
	Set gtmtypfldindx("sgmnt_addrs","th_base")=6
	Set gtmtypes("sgmnt_addrs",7,"name")="sgmnt_addrs.ti"
	Set gtmtypes("sgmnt_addrs",7,"off")=48
	Set gtmtypes("sgmnt_addrs",7,"len")=8
	Set gtmtypes("sgmnt_addrs",7,"type")="addr"
	Set gtmtypfldindx("sgmnt_addrs","ti")=7
	Set gtmtypes("sgmnt_addrs",8,"name")="sgmnt_addrs.nl"
	Set gtmtypes("sgmnt_addrs",8,"off")=56
	Set gtmtypes("sgmnt_addrs",8,"len")=8
	Set gtmtypes("sgmnt_addrs",8,"type")="addr"
	Set gtmtypfldindx("sgmnt_addrs","nl")=8
	Set gtmtypes("sgmnt_addrs",9,"name")="sgmnt_addrs.critical"
	Set gtmtypes("sgmnt_addrs",9,"off")=64
	Set gtmtypes("sgmnt_addrs",9,"len")=8
	Set gtmtypes("sgmnt_addrs",9,"type")="addr"
	Set gtmtypfldindx("sgmnt_addrs","critical")=9
	Set gtmtypes("sgmnt_addrs",10,"name")="sgmnt_addrs.shmpool_buffer"
	Set gtmtypes("sgmnt_addrs",10,"off")=72
	Set gtmtypes("sgmnt_addrs",10,"len")=8
	Set gtmtypes("sgmnt_addrs",10,"type")="addr"
	Set gtmtypfldindx("sgmnt_addrs","shmpool_buffer")=10
	Set gtmtypes("sgmnt_addrs",11,"name")="sgmnt_addrs.db_addrs"
	Set gtmtypes("sgmnt_addrs",11,"off")=80
	Set gtmtypes("sgmnt_addrs",11,"len")=16
	Set gtmtypes("sgmnt_addrs",11,"type")="addr"
	Set gtmtypfldindx("sgmnt_addrs","db_addrs")=11
	Set gtmtypes("sgmnt_addrs",11,"dim")=2
	Set gtmtypes("sgmnt_addrs",12,"name")="sgmnt_addrs.mlkctl"
	Set gtmtypes("sgmnt_addrs",12,"off")=96
	Set gtmtypes("sgmnt_addrs",12,"len")=8
	Set gtmtypes("sgmnt_addrs",12,"type")="addr"
	Set gtmtypfldindx("sgmnt_addrs","mlkctl")=12
	Set gtmtypes("sgmnt_addrs",13,"name")="sgmnt_addrs.mlkctl_len"
	Set gtmtypes("sgmnt_addrs",13,"off")=104
	Set gtmtypes("sgmnt_addrs",13,"len")=4
	Set gtmtypes("sgmnt_addrs",13,"type")="unsigned-int"
	Set gtmtypfldindx("sgmnt_addrs","mlkctl_len")=13
	Set gtmtypes("sgmnt_addrs",14,"name")="sgmnt_addrs.dir_tree"
	Set gtmtypes("sgmnt_addrs",14,"off")=112
	Set gtmtypes("sgmnt_addrs",14,"len")=8
	Set gtmtypes("sgmnt_addrs",14,"type")="addr"
	Set gtmtypfldindx("sgmnt_addrs","dir_tree")=14
	Set gtmtypes("sgmnt_addrs",15,"name")="sgmnt_addrs.hasht_tree"
	Set gtmtypes("sgmnt_addrs",15,"off")=120
	Set gtmtypes("sgmnt_addrs",15,"len")=8
	Set gtmtypes("sgmnt_addrs",15,"type")="addr"
	Set gtmtypfldindx("sgmnt_addrs","hasht_tree")=15
	Set gtmtypes("sgmnt_addrs",16,"name")="sgmnt_addrs.next_fenced"
	Set gtmtypes("sgmnt_addrs",16,"off")=128
	Set gtmtypes("sgmnt_addrs",16,"len")=8
	Set gtmtypes("sgmnt_addrs",16,"type")="addr"
	Set gtmtypfldindx("sgmnt_addrs","next_fenced")=16
	Set gtmtypes("sgmnt_addrs",17,"name")="sgmnt_addrs.jnl"
	Set gtmtypes("sgmnt_addrs",17,"off")=136
	Set gtmtypes("sgmnt_addrs",17,"len")=8
	Set gtmtypes("sgmnt_addrs",17,"type")="addr"
	Set gtmtypfldindx("sgmnt_addrs","jnl")=17
	Set gtmtypes("sgmnt_addrs",18,"name")="sgmnt_addrs.sgm_info_ptr"
	Set gtmtypes("sgmnt_addrs",18,"off")=144
	Set gtmtypes("sgmnt_addrs",18,"len")=8
	Set gtmtypes("sgmnt_addrs",18,"type")="addr"
	Set gtmtypfldindx("sgmnt_addrs","sgm_info_ptr")=18
	Set gtmtypes("sgmnt_addrs",19,"name")="sgmnt_addrs.region"
	Set gtmtypes("sgmnt_addrs",19,"off")=152
	Set gtmtypes("sgmnt_addrs",19,"len")=8
	Set gtmtypes("sgmnt_addrs",19,"type")="addr"
	Set gtmtypfldindx("sgmnt_addrs","region")=19
	Set gtmtypes("sgmnt_addrs",20,"name")="sgmnt_addrs.gvt_hashtab"
	Set gtmtypes("sgmnt_addrs",20,"off")=160
	Set gtmtypes("sgmnt_addrs",20,"len")=8
	Set gtmtypes("sgmnt_addrs",20,"type")="addr"
	Set gtmtypfldindx("sgmnt_addrs","gvt_hashtab")=20
	Set gtmtypes("sgmnt_addrs",21,"name")="sgmnt_addrs.miscptr"
	Set gtmtypes("sgmnt_addrs",21,"off")=168
	Set gtmtypes("sgmnt_addrs",21,"len")=8
	Set gtmtypes("sgmnt_addrs",21,"type")="addr"
	Set gtmtypfldindx("sgmnt_addrs","miscptr")=21
	Set gtmtypes("sgmnt_addrs",22,"name")="sgmnt_addrs.next_csa"
	Set gtmtypes("sgmnt_addrs",22,"off")=176
	Set gtmtypes("sgmnt_addrs",22,"len")=8
	Set gtmtypes("sgmnt_addrs",22,"type")="addr"
	Set gtmtypfldindx("sgmnt_addrs","next_csa")=22
	Set gtmtypes("sgmnt_addrs",23,"name")="sgmnt_addrs.encr_key_handle"
	Set gtmtypes("sgmnt_addrs",23,"off")=184
	Set gtmtypes("sgmnt_addrs",23,"len")=8
	Set gtmtypes("sgmnt_addrs",23,"type")="addr"
	Set gtmtypfldindx("sgmnt_addrs","encr_key_handle")=23
	Set gtmtypes("sgmnt_addrs",24,"name")="sgmnt_addrs.encr_key_handle2"
	Set gtmtypes("sgmnt_addrs",24,"off")=192
	Set gtmtypes("sgmnt_addrs",24,"len")=8
	Set gtmtypes("sgmnt_addrs",24,"type")="addr"
	Set gtmtypfldindx("sgmnt_addrs","encr_key_handle2")=24
	Set gtmtypes("sgmnt_addrs",25,"name")="sgmnt_addrs.encr_ptr"
	Set gtmtypes("sgmnt_addrs",25,"off")=200
	Set gtmtypes("sgmnt_addrs",25,"len")=8
	Set gtmtypes("sgmnt_addrs",25,"type")="addr"
	Set gtmtypfldindx("sgmnt_addrs","encr_ptr")=25
	Set gtmtypes("sgmnt_addrs",26,"name")="sgmnt_addrs.ss_ctx"
	Set gtmtypes("sgmnt_addrs",26,"off")=208
	Set gtmtypes("sgmnt_addrs",26,"len")=8
	Set gtmtypes("sgmnt_addrs",26,"type")="addr"
	Set gtmtypfldindx("sgmnt_addrs","ss_ctx")=26
	Set gtmtypes("sgmnt_addrs",27,"name")="sgmnt_addrs.acc_meth"
	Set gtmtypes("sgmnt_addrs",27,"off")=216
	Set gtmtypes("sgmnt_addrs",27,"len")=8
	Set gtmtypes("sgmnt_addrs",27,"type")="union"
	Set gtmtypfldindx("sgmnt_addrs","acc_meth")=27
	Set gtmtypes("sgmnt_addrs",28,"name")="sgmnt_addrs.acc_meth.mm"
	Set gtmtypes("sgmnt_addrs",28,"off")=216
	Set gtmtypes("sgmnt_addrs",28,"len")=4
	Set gtmtypes("sgmnt_addrs",28,"type")="sgmm_addrs"
	Set gtmtypfldindx("sgmnt_addrs","acc_meth.mm")=28
	Set gtmtypes("sgmnt_addrs",29,"name")="sgmnt_addrs.acc_meth.mm.filler"
	Set gtmtypes("sgmnt_addrs",29,"off")=216
	Set gtmtypes("sgmnt_addrs",29,"len")=4
	Set gtmtypes("sgmnt_addrs",29,"type")="int"
	Set gtmtypfldindx("sgmnt_addrs","acc_meth.mm.filler")=29
	Set gtmtypes("sgmnt_addrs",30,"name")="sgmnt_addrs.acc_meth.bg"
	Set gtmtypes("sgmnt_addrs",30,"off")=216
	Set gtmtypes("sgmnt_addrs",30,"len")=8
	Set gtmtypes("sgmnt_addrs",30,"type")="sgbg_addrs"
	Set gtmtypfldindx("sgmnt_addrs","acc_meth.bg")=30
	Set gtmtypes("sgmnt_addrs",31,"name")="sgmnt_addrs.acc_meth.bg.cache_state"
	Set gtmtypes("sgmnt_addrs",31,"off")=216
	Set gtmtypes("sgmnt_addrs",31,"len")=8
	Set gtmtypes("sgmnt_addrs",31,"type")="addr"
	Set gtmtypfldindx("sgmnt_addrs","acc_meth.bg.cache_state")=31
	Set gtmtypes("sgmnt_addrs",32,"name")="sgmnt_addrs.gvstats_rec_p"
	Set gtmtypes("sgmnt_addrs",32,"off")=224
	Set gtmtypes("sgmnt_addrs",32,"len")=8
	Set gtmtypes("sgmnt_addrs",32,"type")="addr"
	Set gtmtypfldindx("sgmnt_addrs","gvstats_rec_p")=32
	Set gtmtypes("sgmnt_addrs",33,"name")="sgmnt_addrs.gvstats_rec"
	Set gtmtypes("sgmnt_addrs",33,"off")=232
	Set gtmtypes("sgmnt_addrs",33,"len")=592
	Set gtmtypes("sgmnt_addrs",33,"type")="gvstats_rec_t"
	Set gtmtypfldindx("sgmnt_addrs","gvstats_rec")=33
	Set gtmtypes("sgmnt_addrs",34,"name")="sgmnt_addrs.gvstats_rec.n_set"
	Set gtmtypes("sgmnt_addrs",34,"off")=232
	Set gtmtypes("sgmnt_addrs",34,"len")=8
	Set gtmtypes("sgmnt_addrs",34,"type")="uint64_t"
	Set gtmtypfldindx("sgmnt_addrs","gvstats_rec.n_set")=34
	Set gtmtypes("sgmnt_addrs",35,"name")="sgmnt_addrs.gvstats_rec.n_kill"
	Set gtmtypes("sgmnt_addrs",35,"off")=240
	Set gtmtypes("sgmnt_addrs",35,"len")=8
	Set gtmtypes("sgmnt_addrs",35,"type")="uint64_t"
	Set gtmtypfldindx("sgmnt_addrs","gvstats_rec.n_kill")=35
	Set gtmtypes("sgmnt_addrs",36,"name")="sgmnt_addrs.gvstats_rec.n_get"
	Set gtmtypes("sgmnt_addrs",36,"off")=248
	Set gtmtypes("sgmnt_addrs",36,"len")=8
	Set gtmtypes("sgmnt_addrs",36,"type")="uint64_t"
	Set gtmtypfldindx("sgmnt_addrs","gvstats_rec.n_get")=36
	Set gtmtypes("sgmnt_addrs",37,"name")="sgmnt_addrs.gvstats_rec.n_data"
	Set gtmtypes("sgmnt_addrs",37,"off")=256
	Set gtmtypes("sgmnt_addrs",37,"len")=8
	Set gtmtypes("sgmnt_addrs",37,"type")="uint64_t"
	Set gtmtypfldindx("sgmnt_addrs","gvstats_rec.n_data")=37
	Set gtmtypes("sgmnt_addrs",38,"name")="sgmnt_addrs.gvstats_rec.n_order"
	Set gtmtypes("sgmnt_addrs",38,"off")=264
	Set gtmtypes("sgmnt_addrs",38,"len")=8
	Set gtmtypes("sgmnt_addrs",38,"type")="uint64_t"
	Set gtmtypfldindx("sgmnt_addrs","gvstats_rec.n_order")=38
	Set gtmtypes("sgmnt_addrs",39,"name")="sgmnt_addrs.gvstats_rec.n_zprev"
	Set gtmtypes("sgmnt_addrs",39,"off")=272
	Set gtmtypes("sgmnt_addrs",39,"len")=8
	Set gtmtypes("sgmnt_addrs",39,"type")="uint64_t"
	Set gtmtypfldindx("sgmnt_addrs","gvstats_rec.n_zprev")=39
	Set gtmtypes("sgmnt_addrs",40,"name")="sgmnt_addrs.gvstats_rec.n_query"
	Set gtmtypes("sgmnt_addrs",40,"off")=280
	Set gtmtypes("sgmnt_addrs",40,"len")=8
	Set gtmtypes("sgmnt_addrs",40,"type")="uint64_t"
	Set gtmtypfldindx("sgmnt_addrs","gvstats_rec.n_query")=40
	Set gtmtypes("sgmnt_addrs",41,"name")="sgmnt_addrs.gvstats_rec.n_lock_success"
	Set gtmtypes("sgmnt_addrs",41,"off")=288
	Set gtmtypes("sgmnt_addrs",41,"len")=8
	Set gtmtypes("sgmnt_addrs",41,"type")="uint64_t"
	Set gtmtypfldindx("sgmnt_addrs","gvstats_rec.n_lock_success")=41
	Set gtmtypes("sgmnt_addrs",42,"name")="sgmnt_addrs.gvstats_rec.n_lock_fail"
	Set gtmtypes("sgmnt_addrs",42,"off")=296
	Set gtmtypes("sgmnt_addrs",42,"len")=8
	Set gtmtypes("sgmnt_addrs",42,"type")="uint64_t"
	Set gtmtypfldindx("sgmnt_addrs","gvstats_rec.n_lock_fail")=42
	Set gtmtypes("sgmnt_addrs",43,"name")="sgmnt_addrs.gvstats_rec.db_curr_tn"
	Set gtmtypes("sgmnt_addrs",43,"off")=304
	Set gtmtypes("sgmnt_addrs",43,"len")=8
	Set gtmtypes("sgmnt_addrs",43,"type")="uint64_t"
	Set gtmtypfldindx("sgmnt_addrs","gvstats_rec.db_curr_tn")=43
	Set gtmtypes("sgmnt_addrs",44,"name")="sgmnt_addrs.gvstats_rec.n_dsk_read"
	Set gtmtypes("sgmnt_addrs",44,"off")=312
	Set gtmtypes("sgmnt_addrs",44,"len")=8
	Set gtmtypes("sgmnt_addrs",44,"type")="uint64_t"
	Set gtmtypfldindx("sgmnt_addrs","gvstats_rec.n_dsk_read")=44
	Set gtmtypes("sgmnt_addrs",45,"name")="sgmnt_addrs.gvstats_rec.n_dsk_write"
	Set gtmtypes("sgmnt_addrs",45,"off")=320
	Set gtmtypes("sgmnt_addrs",45,"len")=8
	Set gtmtypes("sgmnt_addrs",45,"type")="uint64_t"
	Set gtmtypfldindx("sgmnt_addrs","gvstats_rec.n_dsk_write")=45
	Set gtmtypes("sgmnt_addrs",46,"name")="sgmnt_addrs.gvstats_rec.n_nontp_readwrite"
	Set gtmtypes("sgmnt_addrs",46,"off")=328
	Set gtmtypes("sgmnt_addrs",46,"len")=8
	Set gtmtypes("sgmnt_addrs",46,"type")="uint64_t"
	Set gtmtypfldindx("sgmnt_addrs","gvstats_rec.n_nontp_readwrite")=46
	Set gtmtypes("sgmnt_addrs",47,"name")="sgmnt_addrs.gvstats_rec.n_nontp_readonly"
	Set gtmtypes("sgmnt_addrs",47,"off")=336
	Set gtmtypes("sgmnt_addrs",47,"len")=8
	Set gtmtypes("sgmnt_addrs",47,"type")="uint64_t"
	Set gtmtypfldindx("sgmnt_addrs","gvstats_rec.n_nontp_readonly")=47
	Set gtmtypes("sgmnt_addrs",48,"name")="sgmnt_addrs.gvstats_rec.n_nontp_blkwrite"
	Set gtmtypes("sgmnt_addrs",48,"off")=344
	Set gtmtypes("sgmnt_addrs",48,"len")=8
	Set gtmtypes("sgmnt_addrs",48,"type")="uint64_t"
	Set gtmtypfldindx("sgmnt_addrs","gvstats_rec.n_nontp_blkwrite")=48
	Set gtmtypes("sgmnt_addrs",49,"name")="sgmnt_addrs.gvstats_rec.n_nontp_blkread"
	Set gtmtypes("sgmnt_addrs",49,"off")=352
	Set gtmtypes("sgmnt_addrs",49,"len")=8
	Set gtmtypes("sgmnt_addrs",49,"type")="uint64_t"
	Set gtmtypfldindx("sgmnt_addrs","gvstats_rec.n_nontp_blkread")=49
	Set gtmtypes("sgmnt_addrs",50,"name")="sgmnt_addrs.gvstats_rec.n_nontp_retries_0"
	Set gtmtypes("sgmnt_addrs",50,"off")=360
	Set gtmtypes("sgmnt_addrs",50,"len")=8
	Set gtmtypes("sgmnt_addrs",50,"type")="uint64_t"
	Set gtmtypfldindx("sgmnt_addrs","gvstats_rec.n_nontp_retries_0")=50
	Set gtmtypes("sgmnt_addrs",51,"name")="sgmnt_addrs.gvstats_rec.n_nontp_retries_1"
	Set gtmtypes("sgmnt_addrs",51,"off")=368
	Set gtmtypes("sgmnt_addrs",51,"len")=8
	Set gtmtypes("sgmnt_addrs",51,"type")="uint64_t"
	Set gtmtypfldindx("sgmnt_addrs","gvstats_rec.n_nontp_retries_1")=51
	Set gtmtypes("sgmnt_addrs",52,"name")="sgmnt_addrs.gvstats_rec.n_nontp_retries_2"
	Set gtmtypes("sgmnt_addrs",52,"off")=376
	Set gtmtypes("sgmnt_addrs",52,"len")=8
	Set gtmtypes("sgmnt_addrs",52,"type")="uint64_t"
	Set gtmtypfldindx("sgmnt_addrs","gvstats_rec.n_nontp_retries_2")=52
	Set gtmtypes("sgmnt_addrs",53,"name")="sgmnt_addrs.gvstats_rec.n_nontp_retries_3"
	Set gtmtypes("sgmnt_addrs",53,"off")=384
	Set gtmtypes("sgmnt_addrs",53,"len")=8
	Set gtmtypes("sgmnt_addrs",53,"type")="uint64_t"
	Set gtmtypfldindx("sgmnt_addrs","gvstats_rec.n_nontp_retries_3")=53
	Set gtmtypes("sgmnt_addrs",54,"name")="sgmnt_addrs.gvstats_rec.n_tp_readwrite"
	Set gtmtypes("sgmnt_addrs",54,"off")=392
	Set gtmtypes("sgmnt_addrs",54,"len")=8
	Set gtmtypes("sgmnt_addrs",54,"type")="uint64_t"
	Set gtmtypfldindx("sgmnt_addrs","gvstats_rec.n_tp_readwrite")=54
	Set gtmtypes("sgmnt_addrs",55,"name")="sgmnt_addrs.gvstats_rec.n_tp_readonly"
	Set gtmtypes("sgmnt_addrs",55,"off")=400
	Set gtmtypes("sgmnt_addrs",55,"len")=8
	Set gtmtypes("sgmnt_addrs",55,"type")="uint64_t"
	Set gtmtypfldindx("sgmnt_addrs","gvstats_rec.n_tp_readonly")=55
	Set gtmtypes("sgmnt_addrs",56,"name")="sgmnt_addrs.gvstats_rec.n_tp_rolledback"
	Set gtmtypes("sgmnt_addrs",56,"off")=408
	Set gtmtypes("sgmnt_addrs",56,"len")=8
	Set gtmtypes("sgmnt_addrs",56,"type")="uint64_t"
	Set gtmtypfldindx("sgmnt_addrs","gvstats_rec.n_tp_rolledback")=56
	Set gtmtypes("sgmnt_addrs",57,"name")="sgmnt_addrs.gvstats_rec.n_tp_blkwrite"
	Set gtmtypes("sgmnt_addrs",57,"off")=416
	Set gtmtypes("sgmnt_addrs",57,"len")=8
	Set gtmtypes("sgmnt_addrs",57,"type")="uint64_t"
	Set gtmtypfldindx("sgmnt_addrs","gvstats_rec.n_tp_blkwrite")=57
	Set gtmtypes("sgmnt_addrs",58,"name")="sgmnt_addrs.gvstats_rec.n_tp_blkread"
	Set gtmtypes("sgmnt_addrs",58,"off")=424
	Set gtmtypes("sgmnt_addrs",58,"len")=8
	Set gtmtypes("sgmnt_addrs",58,"type")="uint64_t"
	Set gtmtypfldindx("sgmnt_addrs","gvstats_rec.n_tp_blkread")=58
	Set gtmtypes("sgmnt_addrs",59,"name")="sgmnt_addrs.gvstats_rec.n_tp_tot_retries_0"
	Set gtmtypes("sgmnt_addrs",59,"off")=432
	Set gtmtypes("sgmnt_addrs",59,"len")=8
	Set gtmtypes("sgmnt_addrs",59,"type")="uint64_t"
	Set gtmtypfldindx("sgmnt_addrs","gvstats_rec.n_tp_tot_retries_0")=59
	Set gtmtypes("sgmnt_addrs",60,"name")="sgmnt_addrs.gvstats_rec.n_tp_tot_retries_1"
	Set gtmtypes("sgmnt_addrs",60,"off")=440
	Set gtmtypes("sgmnt_addrs",60,"len")=8
	Set gtmtypes("sgmnt_addrs",60,"type")="uint64_t"
	Set gtmtypfldindx("sgmnt_addrs","gvstats_rec.n_tp_tot_retries_1")=60
	Set gtmtypes("sgmnt_addrs",61,"name")="sgmnt_addrs.gvstats_rec.n_tp_tot_retries_2"
	Set gtmtypes("sgmnt_addrs",61,"off")=448
	Set gtmtypes("sgmnt_addrs",61,"len")=8
	Set gtmtypes("sgmnt_addrs",61,"type")="uint64_t"
	Set gtmtypfldindx("sgmnt_addrs","gvstats_rec.n_tp_tot_retries_2")=61
	Set gtmtypes("sgmnt_addrs",62,"name")="sgmnt_addrs.gvstats_rec.n_tp_tot_retries_3"
	Set gtmtypes("sgmnt_addrs",62,"off")=456
	Set gtmtypes("sgmnt_addrs",62,"len")=8
	Set gtmtypes("sgmnt_addrs",62,"type")="uint64_t"
	Set gtmtypfldindx("sgmnt_addrs","gvstats_rec.n_tp_tot_retries_3")=62
	Set gtmtypes("sgmnt_addrs",63,"name")="sgmnt_addrs.gvstats_rec.n_tp_tot_retries_4"
	Set gtmtypes("sgmnt_addrs",63,"off")=464
	Set gtmtypes("sgmnt_addrs",63,"len")=8
	Set gtmtypes("sgmnt_addrs",63,"type")="uint64_t"
	Set gtmtypfldindx("sgmnt_addrs","gvstats_rec.n_tp_tot_retries_4")=63
	Set gtmtypes("sgmnt_addrs",64,"name")="sgmnt_addrs.gvstats_rec.n_tp_cnflct_retries_0"
	Set gtmtypes("sgmnt_addrs",64,"off")=472
	Set gtmtypes("sgmnt_addrs",64,"len")=8
	Set gtmtypes("sgmnt_addrs",64,"type")="uint64_t"
	Set gtmtypfldindx("sgmnt_addrs","gvstats_rec.n_tp_cnflct_retries_0")=64
	Set gtmtypes("sgmnt_addrs",65,"name")="sgmnt_addrs.gvstats_rec.n_tp_cnflct_retries_1"
	Set gtmtypes("sgmnt_addrs",65,"off")=480
	Set gtmtypes("sgmnt_addrs",65,"len")=8
	Set gtmtypes("sgmnt_addrs",65,"type")="uint64_t"
	Set gtmtypfldindx("sgmnt_addrs","gvstats_rec.n_tp_cnflct_retries_1")=65
	Set gtmtypes("sgmnt_addrs",66,"name")="sgmnt_addrs.gvstats_rec.n_tp_cnflct_retries_2"
	Set gtmtypes("sgmnt_addrs",66,"off")=488
	Set gtmtypes("sgmnt_addrs",66,"len")=8
	Set gtmtypes("sgmnt_addrs",66,"type")="uint64_t"
	Set gtmtypfldindx("sgmnt_addrs","gvstats_rec.n_tp_cnflct_retries_2")=66
	Set gtmtypes("sgmnt_addrs",67,"name")="sgmnt_addrs.gvstats_rec.n_tp_cnflct_retries_3"
	Set gtmtypes("sgmnt_addrs",67,"off")=496
	Set gtmtypes("sgmnt_addrs",67,"len")=8
	Set gtmtypes("sgmnt_addrs",67,"type")="uint64_t"
	Set gtmtypfldindx("sgmnt_addrs","gvstats_rec.n_tp_cnflct_retries_3")=67
	Set gtmtypes("sgmnt_addrs",68,"name")="sgmnt_addrs.gvstats_rec.n_tp_cnflct_retries_4"
	Set gtmtypes("sgmnt_addrs",68,"off")=504
	Set gtmtypes("sgmnt_addrs",68,"len")=8
	Set gtmtypes("sgmnt_addrs",68,"type")="uint64_t"
	Set gtmtypfldindx("sgmnt_addrs","gvstats_rec.n_tp_cnflct_retries_4")=68
	Set gtmtypes("sgmnt_addrs",69,"name")="sgmnt_addrs.gvstats_rec.n_ztrigger"
	Set gtmtypes("sgmnt_addrs",69,"off")=512
	Set gtmtypes("sgmnt_addrs",69,"len")=8
	Set gtmtypes("sgmnt_addrs",69,"type")="uint64_t"
	Set gtmtypfldindx("sgmnt_addrs","gvstats_rec.n_ztrigger")=69
	Set gtmtypes("sgmnt_addrs",70,"name")="sgmnt_addrs.gvstats_rec.n_db_flush"
	Set gtmtypes("sgmnt_addrs",70,"off")=520
	Set gtmtypes("sgmnt_addrs",70,"len")=8
	Set gtmtypes("sgmnt_addrs",70,"type")="uint64_t"
	Set gtmtypfldindx("sgmnt_addrs","gvstats_rec.n_db_flush")=70
	Set gtmtypes("sgmnt_addrs",71,"name")="sgmnt_addrs.gvstats_rec.n_db_fsync"
	Set gtmtypes("sgmnt_addrs",71,"off")=528
	Set gtmtypes("sgmnt_addrs",71,"len")=8
	Set gtmtypes("sgmnt_addrs",71,"type")="uint64_t"
	Set gtmtypfldindx("sgmnt_addrs","gvstats_rec.n_db_fsync")=71
	Set gtmtypes("sgmnt_addrs",72,"name")="sgmnt_addrs.gvstats_rec.n_jnl_flush"
	Set gtmtypes("sgmnt_addrs",72,"off")=536
	Set gtmtypes("sgmnt_addrs",72,"len")=8
	Set gtmtypes("sgmnt_addrs",72,"type")="uint64_t"
	Set gtmtypfldindx("sgmnt_addrs","gvstats_rec.n_jnl_flush")=72
	Set gtmtypes("sgmnt_addrs",73,"name")="sgmnt_addrs.gvstats_rec.n_jnl_fsync"
	Set gtmtypes("sgmnt_addrs",73,"off")=544
	Set gtmtypes("sgmnt_addrs",73,"len")=8
	Set gtmtypes("sgmnt_addrs",73,"type")="uint64_t"
	Set gtmtypfldindx("sgmnt_addrs","gvstats_rec.n_jnl_fsync")=73
	Set gtmtypes("sgmnt_addrs",74,"name")="sgmnt_addrs.gvstats_rec.n_jbuff_bytes"
	Set gtmtypes("sgmnt_addrs",74,"off")=552
	Set gtmtypes("sgmnt_addrs",74,"len")=8
	Set gtmtypes("sgmnt_addrs",74,"type")="uint64_t"
	Set gtmtypfldindx("sgmnt_addrs","gvstats_rec.n_jbuff_bytes")=74
	Set gtmtypes("sgmnt_addrs",75,"name")="sgmnt_addrs.gvstats_rec.n_jfile_bytes"
	Set gtmtypes("sgmnt_addrs",75,"off")=560
	Set gtmtypes("sgmnt_addrs",75,"len")=8
	Set gtmtypes("sgmnt_addrs",75,"type")="uint64_t"
	Set gtmtypfldindx("sgmnt_addrs","gvstats_rec.n_jfile_bytes")=75
	Set gtmtypes("sgmnt_addrs",76,"name")="sgmnt_addrs.gvstats_rec.n_jfile_writes"
	Set gtmtypes("sgmnt_addrs",76,"off")=568
	Set gtmtypes("sgmnt_addrs",76,"len")=8
	Set gtmtypes("sgmnt_addrs",76,"type")="uint64_t"
	Set gtmtypfldindx("sgmnt_addrs","gvstats_rec.n_jfile_writes")=76
	Set gtmtypes("sgmnt_addrs",77,"name")="sgmnt_addrs.gvstats_rec.n_jrec_logical"
	Set gtmtypes("sgmnt_addrs",77,"off")=576
	Set gtmtypes("sgmnt_addrs",77,"len")=8
	Set gtmtypes("sgmnt_addrs",77,"type")="uint64_t"
	Set gtmtypfldindx("sgmnt_addrs","gvstats_rec.n_jrec_logical")=77
	Set gtmtypes("sgmnt_addrs",78,"name")="sgmnt_addrs.gvstats_rec.n_jrec_pblk"
	Set gtmtypes("sgmnt_addrs",78,"off")=584
	Set gtmtypes("sgmnt_addrs",78,"len")=8
	Set gtmtypes("sgmnt_addrs",78,"type")="uint64_t"
	Set gtmtypfldindx("sgmnt_addrs","gvstats_rec.n_jrec_pblk")=78
	Set gtmtypes("sgmnt_addrs",79,"name")="sgmnt_addrs.gvstats_rec.n_jrec_epoch_regular"
	Set gtmtypes("sgmnt_addrs",79,"off")=592
	Set gtmtypes("sgmnt_addrs",79,"len")=8
	Set gtmtypes("sgmnt_addrs",79,"type")="uint64_t"
	Set gtmtypfldindx("sgmnt_addrs","gvstats_rec.n_jrec_epoch_regular")=79
	Set gtmtypes("sgmnt_addrs",80,"name")="sgmnt_addrs.gvstats_rec.n_jrec_epoch_idle"
	Set gtmtypes("sgmnt_addrs",80,"off")=600
	Set gtmtypes("sgmnt_addrs",80,"len")=8
	Set gtmtypes("sgmnt_addrs",80,"type")="uint64_t"
	Set gtmtypfldindx("sgmnt_addrs","gvstats_rec.n_jrec_epoch_idle")=80
	Set gtmtypes("sgmnt_addrs",81,"name")="sgmnt_addrs.gvstats_rec.n_jrec_other"
	Set gtmtypes("sgmnt_addrs",81,"off")=608
	Set gtmtypes("sgmnt_addrs",81,"len")=8
	Set gtmtypes("sgmnt_addrs",81,"type")="uint64_t"
	Set gtmtypfldindx("sgmnt_addrs","gvstats_rec.n_jrec_other")=81
	Set gtmtypes("sgmnt_addrs",82,"name")="sgmnt_addrs.gvstats_rec.n_jnl_extends"
	Set gtmtypes("sgmnt_addrs",82,"off")=616
	Set gtmtypes("sgmnt_addrs",82,"len")=8
	Set gtmtypes("sgmnt_addrs",82,"type")="uint64_t"
	Set gtmtypfldindx("sgmnt_addrs","gvstats_rec.n_jnl_extends")=82
	Set gtmtypes("sgmnt_addrs",83,"name")="sgmnt_addrs.gvstats_rec.n_db_extends"
	Set gtmtypes("sgmnt_addrs",83,"off")=624
	Set gtmtypes("sgmnt_addrs",83,"len")=8
	Set gtmtypes("sgmnt_addrs",83,"type")="uint64_t"
	Set gtmtypfldindx("sgmnt_addrs","gvstats_rec.n_db_extends")=83
	Set gtmtypes("sgmnt_addrs",84,"name")="sgmnt_addrs.gvstats_rec.n_crit_success"
	Set gtmtypes("sgmnt_addrs",84,"off")=632
	Set gtmtypes("sgmnt_addrs",84,"len")=8
	Set gtmtypes("sgmnt_addrs",84,"type")="uint64_t"
	Set gtmtypfldindx("sgmnt_addrs","gvstats_rec.n_crit_success")=84
	Set gtmtypes("sgmnt_addrs",85,"name")="sgmnt_addrs.gvstats_rec.n_crits_in_epch"
	Set gtmtypes("sgmnt_addrs",85,"off")=640
	Set gtmtypes("sgmnt_addrs",85,"len")=8
	Set gtmtypes("sgmnt_addrs",85,"type")="uint64_t"
	Set gtmtypfldindx("sgmnt_addrs","gvstats_rec.n_crits_in_epch")=85
	Set gtmtypes("sgmnt_addrs",86,"name")="sgmnt_addrs.gvstats_rec.sq_crit_failed"
	Set gtmtypes("sgmnt_addrs",86,"off")=648
	Set gtmtypes("sgmnt_addrs",86,"len")=8
	Set gtmtypes("sgmnt_addrs",86,"type")="uint64_t"
	Set gtmtypfldindx("sgmnt_addrs","gvstats_rec.sq_crit_failed")=86
	Set gtmtypes("sgmnt_addrs",87,"name")="sgmnt_addrs.gvstats_rec.n_crit_failed"
	Set gtmtypes("sgmnt_addrs",87,"off")=656
	Set gtmtypes("sgmnt_addrs",87,"len")=8
	Set gtmtypes("sgmnt_addrs",87,"type")="uint64_t"
	Set gtmtypfldindx("sgmnt_addrs","gvstats_rec.n_crit_failed")=87
	Set gtmtypes("sgmnt_addrs",88,"name")="sgmnt_addrs.gvstats_rec.sq_crit_que_slps"
	Set gtmtypes("sgmnt_addrs",88,"off")=664
	Set gtmtypes("sgmnt_addrs",88,"len")=8
	Set gtmtypes("sgmnt_addrs",88,"type")="uint64_t"
	Set gtmtypfldindx("sgmnt_addrs","gvstats_rec.sq_crit_que_slps")=88
	Set gtmtypes("sgmnt_addrs",89,"name")="sgmnt_addrs.gvstats_rec.n_crit_que_slps"
	Set gtmtypes("sgmnt_addrs",89,"off")=672
	Set gtmtypes("sgmnt_addrs",89,"len")=8
	Set gtmtypes("sgmnt_addrs",89,"type")="uint64_t"
	Set gtmtypfldindx("sgmnt_addrs","gvstats_rec.n_crit_que_slps")=89
	Set gtmtypes("sgmnt_addrs",90,"name")="sgmnt_addrs.gvstats_rec.sq_crit_yields"
	Set gtmtypes("sgmnt_addrs",90,"off")=680
	Set gtmtypes("sgmnt_addrs",90,"len")=8
	Set gtmtypes("sgmnt_addrs",90,"type")="uint64_t"
	Set gtmtypfldindx("sgmnt_addrs","gvstats_rec.sq_crit_yields")=90
	Set gtmtypes("sgmnt_addrs",91,"name")="sgmnt_addrs.gvstats_rec.n_crit_yields"
	Set gtmtypes("sgmnt_addrs",91,"off")=688
	Set gtmtypes("sgmnt_addrs",91,"len")=8
	Set gtmtypes("sgmnt_addrs",91,"type")="uint64_t"
	Set gtmtypfldindx("sgmnt_addrs","gvstats_rec.n_crit_yields")=91
	Set gtmtypes("sgmnt_addrs",92,"name")="sgmnt_addrs.gvstats_rec.n_clean2dirty"
	Set gtmtypes("sgmnt_addrs",92,"off")=696
	Set gtmtypes("sgmnt_addrs",92,"len")=8
	Set gtmtypes("sgmnt_addrs",92,"type")="uint64_t"
	Set gtmtypfldindx("sgmnt_addrs","gvstats_rec.n_clean2dirty")=92
	Set gtmtypes("sgmnt_addrs",93,"name")="sgmnt_addrs.gvstats_rec.n_wait_for_read"
	Set gtmtypes("sgmnt_addrs",93,"off")=704
	Set gtmtypes("sgmnt_addrs",93,"len")=8
	Set gtmtypes("sgmnt_addrs",93,"type")="uint64_t"
	Set gtmtypfldindx("sgmnt_addrs","gvstats_rec.n_wait_for_read")=93
	Set gtmtypes("sgmnt_addrs",94,"name")="sgmnt_addrs.gvstats_rec.n_buffer_scarce"
	Set gtmtypes("sgmnt_addrs",94,"off")=712
	Set gtmtypes("sgmnt_addrs",94,"len")=8
	Set gtmtypes("sgmnt_addrs",94,"type")="uint64_t"
	Set gtmtypfldindx("sgmnt_addrs","gvstats_rec.n_buffer_scarce")=94
	Set gtmtypes("sgmnt_addrs",95,"name")="sgmnt_addrs.gvstats_rec.n_bt_scarce"
	Set gtmtypes("sgmnt_addrs",95,"off")=720
	Set gtmtypes("sgmnt_addrs",95,"len")=8
	Set gtmtypes("sgmnt_addrs",95,"type")="uint64_t"
	Set gtmtypfldindx("sgmnt_addrs","gvstats_rec.n_bt_scarce")=95
	Set gtmtypes("sgmnt_addrs",96,"name")="sgmnt_addrs.gvstats_rec.f_dbext_wait"
	Set gtmtypes("sgmnt_addrs",96,"off")=728
	Set gtmtypes("sgmnt_addrs",96,"len")=8
	Set gtmtypes("sgmnt_addrs",96,"type")="uint64_t"
	Set gtmtypfldindx("sgmnt_addrs","gvstats_rec.f_dbext_wait")=96
	Set gtmtypes("sgmnt_addrs",97,"name")="sgmnt_addrs.gvstats_rec.f_bg_wait"
	Set gtmtypes("sgmnt_addrs",97,"off")=736
	Set gtmtypes("sgmnt_addrs",97,"len")=8
	Set gtmtypes("sgmnt_addrs",97,"type")="uint64_t"
	Set gtmtypfldindx("sgmnt_addrs","gvstats_rec.f_bg_wait")=97
	Set gtmtypes("sgmnt_addrs",98,"name")="sgmnt_addrs.gvstats_rec.f_jnl_wait"
	Set gtmtypes("sgmnt_addrs",98,"off")=744
	Set gtmtypes("sgmnt_addrs",98,"len")=8
	Set gtmtypes("sgmnt_addrs",98,"type")="uint64_t"
	Set gtmtypfldindx("sgmnt_addrs","gvstats_rec.f_jnl_wait")=98
	Set gtmtypes("sgmnt_addrs",99,"name")="sgmnt_addrs.gvstats_rec.f_mlk_wait"
	Set gtmtypes("sgmnt_addrs",99,"off")=752
	Set gtmtypes("sgmnt_addrs",99,"len")=8
	Set gtmtypes("sgmnt_addrs",99,"type")="uint64_t"
	Set gtmtypfldindx("sgmnt_addrs","gvstats_rec.f_mlk_wait")=99
	Set gtmtypes("sgmnt_addrs",100,"name")="sgmnt_addrs.gvstats_rec.f_proc_wait"
	Set gtmtypes("sgmnt_addrs",100,"off")=760
	Set gtmtypes("sgmnt_addrs",100,"len")=8
	Set gtmtypes("sgmnt_addrs",100,"type")="uint64_t"
	Set gtmtypfldindx("sgmnt_addrs","gvstats_rec.f_proc_wait")=100
	Set gtmtypes("sgmnt_addrs",101,"name")="sgmnt_addrs.gvstats_rec.f_trans_wait"
	Set gtmtypes("sgmnt_addrs",101,"off")=768
	Set gtmtypes("sgmnt_addrs",101,"len")=8
	Set gtmtypes("sgmnt_addrs",101,"type")="uint64_t"
	Set gtmtypfldindx("sgmnt_addrs","gvstats_rec.f_trans_wait")=101
	Set gtmtypes("sgmnt_addrs",102,"name")="sgmnt_addrs.gvstats_rec.f_util_wait"
	Set gtmtypes("sgmnt_addrs",102,"off")=776
	Set gtmtypes("sgmnt_addrs",102,"len")=8
	Set gtmtypes("sgmnt_addrs",102,"type")="uint64_t"
	Set gtmtypfldindx("sgmnt_addrs","gvstats_rec.f_util_wait")=102
	Set gtmtypes("sgmnt_addrs",103,"name")="sgmnt_addrs.gvstats_rec.f_ws2"
	Set gtmtypes("sgmnt_addrs",103,"off")=784
	Set gtmtypes("sgmnt_addrs",103,"len")=8
	Set gtmtypes("sgmnt_addrs",103,"type")="uint64_t"
	Set gtmtypfldindx("sgmnt_addrs","gvstats_rec.f_ws2")=103
	Set gtmtypes("sgmnt_addrs",104,"name")="sgmnt_addrs.gvstats_rec.f_ws12"
	Set gtmtypes("sgmnt_addrs",104,"off")=792
	Set gtmtypes("sgmnt_addrs",104,"len")=8
	Set gtmtypes("sgmnt_addrs",104,"type")="uint64_t"
	Set gtmtypfldindx("sgmnt_addrs","gvstats_rec.f_ws12")=104
	Set gtmtypes("sgmnt_addrs",105,"name")="sgmnt_addrs.gvstats_rec.f_ws15"
	Set gtmtypes("sgmnt_addrs",105,"off")=800
	Set gtmtypes("sgmnt_addrs",105,"len")=8
	Set gtmtypes("sgmnt_addrs",105,"type")="uint64_t"
	Set gtmtypfldindx("sgmnt_addrs","gvstats_rec.f_ws15")=105
	Set gtmtypes("sgmnt_addrs",106,"name")="sgmnt_addrs.gvstats_rec.f_ws39"
	Set gtmtypes("sgmnt_addrs",106,"off")=808
	Set gtmtypes("sgmnt_addrs",106,"len")=8
	Set gtmtypes("sgmnt_addrs",106,"type")="uint64_t"
	Set gtmtypfldindx("sgmnt_addrs","gvstats_rec.f_ws39")=106
	Set gtmtypes("sgmnt_addrs",107,"name")="sgmnt_addrs.gvstats_rec.f_ws47"
	Set gtmtypes("sgmnt_addrs",107,"off")=816
	Set gtmtypes("sgmnt_addrs",107,"len")=8
	Set gtmtypes("sgmnt_addrs",107,"type")="uint64_t"
	Set gtmtypfldindx("sgmnt_addrs","gvstats_rec.f_ws47")=107
	Set gtmtypes("sgmnt_addrs",108,"name")="sgmnt_addrs.dbsync_timer_tn"
	Set gtmtypes("sgmnt_addrs",108,"off")=824
	Set gtmtypes("sgmnt_addrs",108,"len")=8
	Set gtmtypes("sgmnt_addrs",108,"type")="uint64_t"
	Set gtmtypfldindx("sgmnt_addrs","dbsync_timer_tn")=108
	Set gtmtypes("sgmnt_addrs",109,"name")="sgmnt_addrs.our_midnite"
	Set gtmtypes("sgmnt_addrs",109,"off")=832
	Set gtmtypes("sgmnt_addrs",109,"len")=8
	Set gtmtypes("sgmnt_addrs",109,"type")="addr"
	Set gtmtypfldindx("sgmnt_addrs","our_midnite")=109
	Set gtmtypes("sgmnt_addrs",110,"name")="sgmnt_addrs.fullblockwrite_len"
	Set gtmtypes("sgmnt_addrs",110,"off")=840
	Set gtmtypes("sgmnt_addrs",110,"len")=8
	Set gtmtypes("sgmnt_addrs",110,"type")="size_t"
	Set gtmtypfldindx("sgmnt_addrs","fullblockwrite_len")=110
	Set gtmtypes("sgmnt_addrs",111,"name")="sgmnt_addrs.our_lru_cache_rec_off"
	Set gtmtypes("sgmnt_addrs",111,"off")=848
	Set gtmtypes("sgmnt_addrs",111,"len")=8
	Set gtmtypes("sgmnt_addrs",111,"type")="intptr_t"
	Set gtmtypfldindx("sgmnt_addrs","our_lru_cache_rec_off")=111
	Set gtmtypes("sgmnt_addrs",112,"name")="sgmnt_addrs.total_blks"
	Set gtmtypes("sgmnt_addrs",112,"off")=856
	Set gtmtypes("sgmnt_addrs",112,"len")=4
	Set gtmtypes("sgmnt_addrs",112,"type")="unsigned-int"
	Set gtmtypfldindx("sgmnt_addrs","total_blks")=112
	Set gtmtypes("sgmnt_addrs",113,"name")="sgmnt_addrs.prev_free_blks"
	Set gtmtypes("sgmnt_addrs",113,"off")=860
	Set gtmtypes("sgmnt_addrs",113,"len")=4
	Set gtmtypes("sgmnt_addrs",113,"type")="unsigned-int"
	Set gtmtypfldindx("sgmnt_addrs","prev_free_blks")=113
	Set gtmtypes("sgmnt_addrs",114,"name")="sgmnt_addrs.timer"
	Set gtmtypes("sgmnt_addrs",114,"off")=864
	Set gtmtypes("sgmnt_addrs",114,"len")=4
	Set gtmtypes("sgmnt_addrs",114,"type")="unsigned-int"
	Set gtmtypfldindx("sgmnt_addrs","timer")=114
	Set gtmtypes("sgmnt_addrs",115,"name")="sgmnt_addrs.in_wtstart"
	Set gtmtypes("sgmnt_addrs",115,"off")=868
	Set gtmtypes("sgmnt_addrs",115,"len")=4
	Set gtmtypes("sgmnt_addrs",115,"type")="unsigned-int"
	Set gtmtypfldindx("sgmnt_addrs","in_wtstart")=115
	Set gtmtypes("sgmnt_addrs",116,"name")="sgmnt_addrs.now_crit"
	Set gtmtypes("sgmnt_addrs",116,"off")=872
	Set gtmtypes("sgmnt_addrs",116,"len")=4
	Set gtmtypes("sgmnt_addrs",116,"type")="unsigned-int"
	Set gtmtypfldindx("sgmnt_addrs","now_crit")=116
	Set gtmtypes("sgmnt_addrs",117,"name")="sgmnt_addrs.wbuf_dqd"
	Set gtmtypes("sgmnt_addrs",117,"off")=876
	Set gtmtypes("sgmnt_addrs",117,"len")=4
	Set gtmtypes("sgmnt_addrs",117,"type")="unsigned-int"
	Set gtmtypfldindx("sgmnt_addrs","wbuf_dqd")=117
	Set gtmtypes("sgmnt_addrs",118,"name")="sgmnt_addrs.stale_defer"
	Set gtmtypes("sgmnt_addrs",118,"off")=880
	Set gtmtypes("sgmnt_addrs",118,"len")=4
	Set gtmtypes("sgmnt_addrs",118,"type")="unsigned-int"
	Set gtmtypfldindx("sgmnt_addrs","stale_defer")=118
	Set gtmtypes("sgmnt_addrs",119,"name")="sgmnt_addrs.freeze"
	Set gtmtypes("sgmnt_addrs",119,"off")=884
	Set gtmtypes("sgmnt_addrs",119,"len")=4
	Set gtmtypes("sgmnt_addrs",119,"type")="boolean_t"
	Set gtmtypfldindx("sgmnt_addrs","freeze")=119
	Set gtmtypes("sgmnt_addrs",120,"name")="sgmnt_addrs.dbsync_timer"
	Set gtmtypes("sgmnt_addrs",120,"off")=888
	Set gtmtypes("sgmnt_addrs",120,"len")=4
	Set gtmtypes("sgmnt_addrs",120,"type")="boolean_t"
	Set gtmtypfldindx("sgmnt_addrs","dbsync_timer")=120
	Set gtmtypes("sgmnt_addrs",121,"name")="sgmnt_addrs.reorg_last_dest"
	Set gtmtypes("sgmnt_addrs",121,"off")=892
	Set gtmtypes("sgmnt_addrs",121,"len")=4
	Set gtmtypes("sgmnt_addrs",121,"type")="int"
	Set gtmtypfldindx("sgmnt_addrs","reorg_last_dest")=121
	Set gtmtypes("sgmnt_addrs",122,"name")="sgmnt_addrs.jnl_before_image"
	Set gtmtypes("sgmnt_addrs",122,"off")=896
	Set gtmtypes("sgmnt_addrs",122,"len")=4
	Set gtmtypes("sgmnt_addrs",122,"type")="boolean_t"
	Set gtmtypfldindx("sgmnt_addrs","jnl_before_image")=122
	Set gtmtypes("sgmnt_addrs",123,"name")="sgmnt_addrs.read_write"
	Set gtmtypes("sgmnt_addrs",123,"off")=900
	Set gtmtypes("sgmnt_addrs",123,"len")=4
	Set gtmtypes("sgmnt_addrs",123,"type")="boolean_t"
	Set gtmtypfldindx("sgmnt_addrs","read_write")=123
	Set gtmtypes("sgmnt_addrs",124,"name")="sgmnt_addrs.orig_read_write"
	Set gtmtypes("sgmnt_addrs",124,"off")=904
	Set gtmtypes("sgmnt_addrs",124,"len")=4
	Set gtmtypes("sgmnt_addrs",124,"type")="boolean_t"
	Set gtmtypfldindx("sgmnt_addrs","orig_read_write")=124
	Set gtmtypes("sgmnt_addrs",125,"name")="sgmnt_addrs.persistent_freeze"
	Set gtmtypes("sgmnt_addrs",125,"off")=908
	Set gtmtypes("sgmnt_addrs",125,"len")=4
	Set gtmtypes("sgmnt_addrs",125,"type")="boolean_t"
	Set gtmtypfldindx("sgmnt_addrs","persistent_freeze")=125
	Set gtmtypes("sgmnt_addrs",126,"name")="sgmnt_addrs.pblk_align_jrecsize"
	Set gtmtypes("sgmnt_addrs",126,"off")=912
	Set gtmtypes("sgmnt_addrs",126,"len")=4
	Set gtmtypes("sgmnt_addrs",126,"type")="int"
	Set gtmtypfldindx("sgmnt_addrs","pblk_align_jrecsize")=126
	Set gtmtypes("sgmnt_addrs",127,"name")="sgmnt_addrs.min_total_tpjnl_rec_size"
	Set gtmtypes("sgmnt_addrs",127,"off")=916
	Set gtmtypes("sgmnt_addrs",127,"len")=4
	Set gtmtypes("sgmnt_addrs",127,"type")="int"
	Set gtmtypfldindx("sgmnt_addrs","min_total_tpjnl_rec_size")=127
	Set gtmtypes("sgmnt_addrs",128,"name")="sgmnt_addrs.min_total_nontpjnl_rec_size"
	Set gtmtypes("sgmnt_addrs",128,"off")=920
	Set gtmtypes("sgmnt_addrs",128,"len")=4
	Set gtmtypes("sgmnt_addrs",128,"type")="int"
	Set gtmtypfldindx("sgmnt_addrs","min_total_nontpjnl_rec_size")=128
	Set gtmtypes("sgmnt_addrs",129,"name")="sgmnt_addrs.jnl_state"
	Set gtmtypes("sgmnt_addrs",129,"off")=924
	Set gtmtypes("sgmnt_addrs",129,"len")=4
	Set gtmtypes("sgmnt_addrs",129,"type")="int"
	Set gtmtypfldindx("sgmnt_addrs","jnl_state")=129
	Set gtmtypes("sgmnt_addrs",130,"name")="sgmnt_addrs.repl_state"
	Set gtmtypes("sgmnt_addrs",130,"off")=928
	Set gtmtypes("sgmnt_addrs",130,"len")=4
	Set gtmtypes("sgmnt_addrs",130,"type")="int"
	Set gtmtypfldindx("sgmnt_addrs","repl_state")=130
	Set gtmtypes("sgmnt_addrs",131,"name")="sgmnt_addrs.crit_check_cycle"
	Set gtmtypes("sgmnt_addrs",131,"off")=932
	Set gtmtypes("sgmnt_addrs",131,"len")=4
	Set gtmtypes("sgmnt_addrs",131,"type")="unsigned-int"
	Set gtmtypfldindx("sgmnt_addrs","crit_check_cycle")=131
	Set gtmtypes("sgmnt_addrs",132,"name")="sgmnt_addrs.backup_in_prog"
	Set gtmtypes("sgmnt_addrs",132,"off")=936
	Set gtmtypes("sgmnt_addrs",132,"len")=4
	Set gtmtypes("sgmnt_addrs",132,"type")="int"
	Set gtmtypfldindx("sgmnt_addrs","backup_in_prog")=132
	Set gtmtypes("sgmnt_addrs",133,"name")="sgmnt_addrs.snapshot_in_prog"
	Set gtmtypes("sgmnt_addrs",133,"off")=940
	Set gtmtypes("sgmnt_addrs",133,"len")=4
	Set gtmtypes("sgmnt_addrs",133,"type")="boolean_t"
	Set gtmtypfldindx("sgmnt_addrs","snapshot_in_prog")=133
	Set gtmtypes("sgmnt_addrs",134,"name")="sgmnt_addrs.ref_cnt"
	Set gtmtypes("sgmnt_addrs",134,"off")=944
	Set gtmtypes("sgmnt_addrs",134,"len")=4
	Set gtmtypes("sgmnt_addrs",134,"type")="int"
	Set gtmtypfldindx("sgmnt_addrs","ref_cnt")=134
	Set gtmtypes("sgmnt_addrs",135,"name")="sgmnt_addrs.fid_index"
	Set gtmtypes("sgmnt_addrs",135,"off")=948
	Set gtmtypes("sgmnt_addrs",135,"len")=4
	Set gtmtypes("sgmnt_addrs",135,"type")="int"
	Set gtmtypfldindx("sgmnt_addrs","fid_index")=135
	Set gtmtypes("sgmnt_addrs",136,"name")="sgmnt_addrs.filler"
	Set gtmtypes("sgmnt_addrs",136,"off")=952
	Set gtmtypes("sgmnt_addrs",136,"len")=4
	Set gtmtypes("sgmnt_addrs",136,"type")="char"
	Set gtmtypfldindx("sgmnt_addrs","filler")=136
	Set gtmtypes("sgmnt_addrs",137,"name")="sgmnt_addrs.regnum"
	Set gtmtypes("sgmnt_addrs",137,"off")=956
	Set gtmtypes("sgmnt_addrs",137,"len")=4
	Set gtmtypes("sgmnt_addrs",137,"type")="int"
	Set gtmtypfldindx("sgmnt_addrs","regnum")=137
	Set gtmtypes("sgmnt_addrs",138,"name")="sgmnt_addrs.n_pre_read_trigger"
	Set gtmtypes("sgmnt_addrs",138,"off")=960
	Set gtmtypes("sgmnt_addrs",138,"len")=4
	Set gtmtypes("sgmnt_addrs",138,"type")="int"
	Set gtmtypfldindx("sgmnt_addrs","n_pre_read_trigger")=138
	Set gtmtypes("sgmnt_addrs",139,"name")="sgmnt_addrs.jnlpool_validate_check"
	Set gtmtypes("sgmnt_addrs",139,"off")=964
	Set gtmtypes("sgmnt_addrs",139,"len")=4
	Set gtmtypes("sgmnt_addrs",139,"type")="unsigned-int"
	Set gtmtypfldindx("sgmnt_addrs","jnlpool_validate_check")=139
	Set gtmtypes("sgmnt_addrs",140,"name")="sgmnt_addrs.regcnt"
	Set gtmtypes("sgmnt_addrs",140,"off")=968
	Set gtmtypes("sgmnt_addrs",140,"len")=4
	Set gtmtypes("sgmnt_addrs",140,"type")="int"
	Set gtmtypfldindx("sgmnt_addrs","regcnt")=140
	Set gtmtypes("sgmnt_addrs",141,"name")="sgmnt_addrs.t_commit_crit"
	Set gtmtypes("sgmnt_addrs",141,"off")=972
	Set gtmtypes("sgmnt_addrs",141,"len")=4
	Set gtmtypes("sgmnt_addrs",141,"type")="boolean_t"
	Set gtmtypfldindx("sgmnt_addrs","t_commit_crit")=141
	Set gtmtypes("sgmnt_addrs",142,"name")="sgmnt_addrs.wcs_pidcnt_incremented"
	Set gtmtypes("sgmnt_addrs",142,"off")=976
	Set gtmtypes("sgmnt_addrs",142,"len")=4
	Set gtmtypes("sgmnt_addrs",142,"type")="boolean_t"
	Set gtmtypfldindx("sgmnt_addrs","wcs_pidcnt_incremented")=142
	Set gtmtypes("sgmnt_addrs",143,"name")="sgmnt_addrs.incr_db_trigger_cycle"
	Set gtmtypes("sgmnt_addrs",143,"off")=980
	Set gtmtypes("sgmnt_addrs",143,"len")=4
	Set gtmtypes("sgmnt_addrs",143,"type")="boolean_t"
	Set gtmtypfldindx("sgmnt_addrs","incr_db_trigger_cycle")=143
	Set gtmtypes("sgmnt_addrs",144,"name")="sgmnt_addrs.db_trigger_cycle"
	Set gtmtypes("sgmnt_addrs",144,"off")=984
	Set gtmtypes("sgmnt_addrs",144,"len")=4
	Set gtmtypes("sgmnt_addrs",144,"type")="unsigned-int"
	Set gtmtypfldindx("sgmnt_addrs","db_trigger_cycle")=144
	Set gtmtypes("sgmnt_addrs",145,"name")="sgmnt_addrs.db_dztrigger_cycle"
	Set gtmtypes("sgmnt_addrs",145,"off")=988
	Set gtmtypes("sgmnt_addrs",145,"len")=4
	Set gtmtypes("sgmnt_addrs",145,"type")="unsigned-int"
	Set gtmtypfldindx("sgmnt_addrs","db_dztrigger_cycle")=145
	Set gtmtypes("sgmnt_addrs",146,"name")="sgmnt_addrs.hold_onto_crit"
	Set gtmtypes("sgmnt_addrs",146,"off")=992
	Set gtmtypes("sgmnt_addrs",146,"len")=4
	Set gtmtypes("sgmnt_addrs",146,"type")="boolean_t"
	Set gtmtypfldindx("sgmnt_addrs","hold_onto_crit")=146
	Set gtmtypes("sgmnt_addrs",147,"name")="sgmnt_addrs.dse_crit_seize_done"
	Set gtmtypes("sgmnt_addrs",147,"off")=996
	Set gtmtypes("sgmnt_addrs",147,"len")=4
	Set gtmtypes("sgmnt_addrs",147,"type")="boolean_t"
	Set gtmtypfldindx("sgmnt_addrs","dse_crit_seize_done")=147
	Set gtmtypes("sgmnt_addrs",148,"name")="sgmnt_addrs.gbuff_limit"
	Set gtmtypes("sgmnt_addrs",148,"off")=1000
	Set gtmtypes("sgmnt_addrs",148,"len")=4
	Set gtmtypes("sgmnt_addrs",148,"type")="int"
	Set gtmtypfldindx("sgmnt_addrs","gbuff_limit")=148
	Set gtmtypes("sgmnt_addrs",149,"name")="sgmnt_addrs.root_search_cycle"
	Set gtmtypes("sgmnt_addrs",149,"off")=1004
	Set gtmtypes("sgmnt_addrs",149,"len")=4
	Set gtmtypes("sgmnt_addrs",149,"type")="unsigned-int"
	Set gtmtypfldindx("sgmnt_addrs","root_search_cycle")=149
	Set gtmtypes("sgmnt_addrs",150,"name")="sgmnt_addrs.onln_rlbk_cycle"
	Set gtmtypes("sgmnt_addrs",150,"off")=1008
	Set gtmtypes("sgmnt_addrs",150,"len")=4
	Set gtmtypes("sgmnt_addrs",150,"type")="unsigned-int"
	Set gtmtypfldindx("sgmnt_addrs","onln_rlbk_cycle")=150
	Set gtmtypes("sgmnt_addrs",151,"name")="sgmnt_addrs.db_onln_rlbkd_cycle"
	Set gtmtypes("sgmnt_addrs",151,"off")=1012
	Set gtmtypes("sgmnt_addrs",151,"len")=4
	Set gtmtypes("sgmnt_addrs",151,"type")="unsigned-int"
	Set gtmtypfldindx("sgmnt_addrs","db_onln_rlbkd_cycle")=151
	Set gtmtypes("sgmnt_addrs",152,"name")="sgmnt_addrs.reservedDBFlags"
	Set gtmtypes("sgmnt_addrs",152,"off")=1016
	Set gtmtypes("sgmnt_addrs",152,"len")=4
	Set gtmtypes("sgmnt_addrs",152,"type")="unsigned-int"
	Set gtmtypfldindx("sgmnt_addrs","reservedDBFlags")=152
	Set gtmtypes("sgmnt_addrs",153,"name")="sgmnt_addrs.read_only_fs"
	Set gtmtypes("sgmnt_addrs",153,"off")=1020
	Set gtmtypes("sgmnt_addrs",153,"len")=4
	Set gtmtypes("sgmnt_addrs",153,"type")="boolean_t"
	Set gtmtypfldindx("sgmnt_addrs","read_only_fs")=153
	Set gtmtypes("sgmnt_addrs",154,"name")="sgmnt_addrs.crit_probe"
	Set gtmtypes("sgmnt_addrs",154,"off")=1024
	Set gtmtypes("sgmnt_addrs",154,"len")=4
	Set gtmtypes("sgmnt_addrs",154,"type")="boolean_t"
	Set gtmtypfldindx("sgmnt_addrs","crit_probe")=154
	Set gtmtypes("sgmnt_addrs",155,"name")="sgmnt_addrs.canceled_flush_timer"
	Set gtmtypes("sgmnt_addrs",155,"off")=1028
	Set gtmtypes("sgmnt_addrs",155,"len")=4
	Set gtmtypes("sgmnt_addrs",155,"type")="boolean_t"
	Set gtmtypfldindx("sgmnt_addrs","canceled_flush_timer")=155
	Set gtmtypes("sgmnt_addrs",156,"name")="sgmnt_addrs.probecrit_rec"
	Set gtmtypes("sgmnt_addrs",156,"off")=1032
	Set gtmtypes("sgmnt_addrs",156,"len")=56
	Set gtmtypes("sgmnt_addrs",156,"type")="probecrit_rec_t"
	Set gtmtypfldindx("sgmnt_addrs","probecrit_rec")=156
	Set gtmtypes("sgmnt_addrs",157,"name")="sgmnt_addrs.probecrit_rec.t_get_crit"
	Set gtmtypes("sgmnt_addrs",157,"off")=1032
	Set gtmtypes("sgmnt_addrs",157,"len")=8
	Set gtmtypes("sgmnt_addrs",157,"type")="uint64_t"
	Set gtmtypfldindx("sgmnt_addrs","probecrit_rec.t_get_crit")=157
	Set gtmtypes("sgmnt_addrs",158,"name")="sgmnt_addrs.probecrit_rec.p_crit_failed"
	Set gtmtypes("sgmnt_addrs",158,"off")=1040
	Set gtmtypes("sgmnt_addrs",158,"len")=8
	Set gtmtypes("sgmnt_addrs",158,"type")="uint64_t"
	Set gtmtypfldindx("sgmnt_addrs","probecrit_rec.p_crit_failed")=158
	Set gtmtypes("sgmnt_addrs",159,"name")="sgmnt_addrs.probecrit_rec.p_crit_que_slps"
	Set gtmtypes("sgmnt_addrs",159,"off")=1048
	Set gtmtypes("sgmnt_addrs",159,"len")=8
	Set gtmtypes("sgmnt_addrs",159,"type")="uint64_t"
	Set gtmtypfldindx("sgmnt_addrs","probecrit_rec.p_crit_que_slps")=159
	Set gtmtypes("sgmnt_addrs",160,"name")="sgmnt_addrs.probecrit_rec.p_crit_yields"
	Set gtmtypes("sgmnt_addrs",160,"off")=1056
	Set gtmtypes("sgmnt_addrs",160,"len")=8
	Set gtmtypes("sgmnt_addrs",160,"type")="uint64_t"
	Set gtmtypfldindx("sgmnt_addrs","probecrit_rec.p_crit_yields")=160
	Set gtmtypes("sgmnt_addrs",161,"name")="sgmnt_addrs.probecrit_rec.p_crit_que_full"
	Set gtmtypes("sgmnt_addrs",161,"off")=1064
	Set gtmtypes("sgmnt_addrs",161,"len")=8
	Set gtmtypes("sgmnt_addrs",161,"type")="uint64_t"
	Set gtmtypfldindx("sgmnt_addrs","probecrit_rec.p_crit_que_full")=161
	Set gtmtypes("sgmnt_addrs",162,"name")="sgmnt_addrs.probecrit_rec.p_crit_que_slots"
	Set gtmtypes("sgmnt_addrs",162,"off")=1072
	Set gtmtypes("sgmnt_addrs",162,"len")=8
	Set gtmtypes("sgmnt_addrs",162,"type")="uint64_t"
	Set gtmtypfldindx("sgmnt_addrs","probecrit_rec.p_crit_que_slots")=162
	Set gtmtypes("sgmnt_addrs",163,"name")="sgmnt_addrs.probecrit_rec.p_crit_success"
	Set gtmtypes("sgmnt_addrs",163,"off")=1080
	Set gtmtypes("sgmnt_addrs",163,"len")=8
	Set gtmtypes("sgmnt_addrs",163,"type")="uint64_t"
	Set gtmtypfldindx("sgmnt_addrs","probecrit_rec.p_crit_success")=163
	Set gtmtypes("sgmnt_addrs",164,"name")="sgmnt_addrs.lock_crit_with_db"
	Set gtmtypes("sgmnt_addrs",164,"off")=1088
	Set gtmtypes("sgmnt_addrs",164,"len")=4
	Set gtmtypes("sgmnt_addrs",164,"type")="boolean_t"
	Set gtmtypfldindx("sgmnt_addrs","lock_crit_with_db")=164
	Set gtmtypes("sgmnt_addrs",165,"name")="sgmnt_addrs.needs_post_freeze_online_clean"
	Set gtmtypes("sgmnt_addrs",165,"off")=1092
	Set gtmtypes("sgmnt_addrs",165,"len")=4
	Set gtmtypes("sgmnt_addrs",165,"type")="boolean_t"
	Set gtmtypfldindx("sgmnt_addrs","needs_post_freeze_online_clean")=165
	Set gtmtypes("sgmnt_addrs",166,"name")="sgmnt_addrs.needs_post_freeze_flushsync"
	Set gtmtypes("sgmnt_addrs",166,"off")=1096
	Set gtmtypes("sgmnt_addrs",166,"len")=4
	Set gtmtypes("sgmnt_addrs",166,"type")="boolean_t"
	Set gtmtypfldindx("sgmnt_addrs","needs_post_freeze_flushsync")=166
	Set gtmtypes("sgmnt_addrs",167,"name")="sgmnt_addrs.tp_hint"
	Set gtmtypes("sgmnt_addrs",167,"off")=1100
	Set gtmtypes("sgmnt_addrs",167,"len")=4
	Set gtmtypes("sgmnt_addrs",167,"type")="int"
	Set gtmtypfldindx("sgmnt_addrs","tp_hint")=167
	Set gtmtypes("sgmnt_addrs",168,"name")="sgmnt_addrs.tp_in_use"
	Set gtmtypes("sgmnt_addrs",168,"off")=1104
	Set gtmtypes("sgmnt_addrs",168,"len")=4
	Set gtmtypes("sgmnt_addrs",168,"type")="boolean_t"
	Set gtmtypfldindx("sgmnt_addrs","tp_in_use")=168
	Set gtmtypes("sgmnt_addrs",169,"name")="sgmnt_addrs.statsDB_setup_completed"
	Set gtmtypes("sgmnt_addrs",169,"off")=1108
	Set gtmtypes("sgmnt_addrs",169,"len")=4
	Set gtmtypes("sgmnt_addrs",169,"type")="boolean_t"
	Set gtmtypfldindx("sgmnt_addrs","statsDB_setup_completed")=169
	Set gtmtypes("sgmnt_addrs",170,"name")="sgmnt_addrs.gd_instinfo"
	Set gtmtypes("sgmnt_addrs",170,"off")=1112
	Set gtmtypes("sgmnt_addrs",170,"len")=8
	Set gtmtypes("sgmnt_addrs",170,"type")="addr"
	Set gtmtypfldindx("sgmnt_addrs","gd_instinfo")=170
	Set gtmtypes("sgmnt_addrs",171,"name")="sgmnt_addrs.gd_ptr"
	Set gtmtypes("sgmnt_addrs",171,"off")=1120
	Set gtmtypes("sgmnt_addrs",171,"len")=8
	Set gtmtypes("sgmnt_addrs",171,"type")="addr"
	Set gtmtypfldindx("sgmnt_addrs","gd_ptr")=171
	Set gtmtypes("sgmnt_addrs",172,"name")="sgmnt_addrs.jnlpool"
	Set gtmtypes("sgmnt_addrs",172,"off")=1128
	Set gtmtypes("sgmnt_addrs",172,"len")=8
	Set gtmtypes("sgmnt_addrs",172,"type")="addr"
	Set gtmtypfldindx("sgmnt_addrs","jnlpool")=172
	Set gtmtypes("sgmnt_addrs",173,"name")="sgmnt_addrs.mlkhash"
	Set gtmtypes("sgmnt_addrs",173,"off")=1136
	Set gtmtypes("sgmnt_addrs",173,"len")=8
	Set gtmtypes("sgmnt_addrs",173,"type")="addr"
	Set gtmtypfldindx("sgmnt_addrs","mlkhash")=173
	Set gtmtypes("sgmnt_addrs",174,"name")="sgmnt_addrs.mlkhash_shmid"
	Set gtmtypes("sgmnt_addrs",174,"off")=1144
	Set gtmtypes("sgmnt_addrs",174,"len")=4
	Set gtmtypes("sgmnt_addrs",174,"type")="int"
	Set gtmtypfldindx("sgmnt_addrs","mlkhash_shmid")=174
	;
	Set gtmtypes("sgmnt_data")="struct"
	Set gtmtypes("sgmnt_data",0)=611
	Set gtmtypes("sgmnt_data","len")=8192
	Set gtmtypes("sgmnt_data",1,"name")="sgmnt_data.label"
	Set gtmtypes("sgmnt_data",1,"off")=0
	Set gtmtypes("sgmnt_data",1,"len")=12
	Set gtmtypes("sgmnt_data",1,"type")="unsigned-char"
	Set gtmtypfldindx("sgmnt_data","label")=1
	Set gtmtypes("sgmnt_data",2,"name")="sgmnt_data.blk_size"
	Set gtmtypes("sgmnt_data",2,"off")=12
	Set gtmtypes("sgmnt_data",2,"len")=4
	Set gtmtypes("sgmnt_data",2,"type")="int"
	Set gtmtypfldindx("sgmnt_data","blk_size")=2
	Set gtmtypes("sgmnt_data",3,"name")="sgmnt_data.master_map_len"
	Set gtmtypes("sgmnt_data",3,"off")=16
	Set gtmtypes("sgmnt_data",3,"len")=4
	Set gtmtypes("sgmnt_data",3,"type")="int"
	Set gtmtypfldindx("sgmnt_data","master_map_len")=3
	Set gtmtypes("sgmnt_data",4,"name")="sgmnt_data.bplmap"
	Set gtmtypes("sgmnt_data",4,"off")=20
	Set gtmtypes("sgmnt_data",4,"len")=4
	Set gtmtypes("sgmnt_data",4,"type")="int"
	Set gtmtypfldindx("sgmnt_data","bplmap")=4
	Set gtmtypes("sgmnt_data",5,"name")="sgmnt_data.start_vbn"
	Set gtmtypes("sgmnt_data",5,"off")=24
	Set gtmtypes("sgmnt_data",5,"len")=4
	Set gtmtypes("sgmnt_data",5,"type")="int"
	Set gtmtypfldindx("sgmnt_data","start_vbn")=5
	Set gtmtypes("sgmnt_data",6,"name")="sgmnt_data.acc_meth"
	Set gtmtypes("sgmnt_data",6,"off")=28
	Set gtmtypes("sgmnt_data",6,"len")=4
	Set gtmtypes("sgmnt_data",6,"type")="int"
	Set gtmtypfldindx("sgmnt_data","acc_meth")=6
	Set gtmtypes("sgmnt_data",7,"name")="sgmnt_data.max_bts"
	Set gtmtypes("sgmnt_data",7,"off")=32
	Set gtmtypes("sgmnt_data",7,"len")=4
	Set gtmtypes("sgmnt_data",7,"type")="unsigned-int"
	Set gtmtypfldindx("sgmnt_data","max_bts")=7
	Set gtmtypes("sgmnt_data",8,"name")="sgmnt_data.n_bts"
	Set gtmtypes("sgmnt_data",8,"off")=36
	Set gtmtypes("sgmnt_data",8,"len")=4
	Set gtmtypes("sgmnt_data",8,"type")="int"
	Set gtmtypfldindx("sgmnt_data","n_bts")=8
	Set gtmtypes("sgmnt_data",9,"name")="sgmnt_data.bt_buckets"
	Set gtmtypes("sgmnt_data",9,"off")=40
	Set gtmtypes("sgmnt_data",9,"len")=4
	Set gtmtypes("sgmnt_data",9,"type")="int"
	Set gtmtypfldindx("sgmnt_data","bt_buckets")=9
	Set gtmtypes("sgmnt_data",10,"name")="sgmnt_data.reserved_bytes"
	Set gtmtypes("sgmnt_data",10,"off")=44
	Set gtmtypes("sgmnt_data",10,"len")=4
	Set gtmtypes("sgmnt_data",10,"type")="int"
	Set gtmtypfldindx("sgmnt_data","reserved_bytes")=10
	Set gtmtypes("sgmnt_data",11,"name")="sgmnt_data.max_rec_size"
	Set gtmtypes("sgmnt_data",11,"off")=48
	Set gtmtypes("sgmnt_data",11,"len")=4
	Set gtmtypes("sgmnt_data",11,"type")="int"
	Set gtmtypfldindx("sgmnt_data","max_rec_size")=11
	Set gtmtypes("sgmnt_data",12,"name")="sgmnt_data.max_key_size"
	Set gtmtypes("sgmnt_data",12,"off")=52
	Set gtmtypes("sgmnt_data",12,"len")=4
	Set gtmtypes("sgmnt_data",12,"type")="int"
	Set gtmtypfldindx("sgmnt_data","max_key_size")=12
	Set gtmtypes("sgmnt_data",13,"name")="sgmnt_data.lock_space_size"
	Set gtmtypes("sgmnt_data",13,"off")=56
	Set gtmtypes("sgmnt_data",13,"len")=4
	Set gtmtypes("sgmnt_data",13,"type")="unsigned-int"
	Set gtmtypfldindx("sgmnt_data","lock_space_size")=13
	Set gtmtypes("sgmnt_data",14,"name")="sgmnt_data.extension_size"
	Set gtmtypes("sgmnt_data",14,"off")=60
	Set gtmtypes("sgmnt_data",14,"len")=4
	Set gtmtypes("sgmnt_data",14,"type")="unsigned-int"
	Set gtmtypfldindx("sgmnt_data","extension_size")=14
	Set gtmtypes("sgmnt_data",15,"name")="sgmnt_data.def_coll"
	Set gtmtypes("sgmnt_data",15,"off")=64
	Set gtmtypes("sgmnt_data",15,"len")=4
	Set gtmtypes("sgmnt_data",15,"type")="unsigned-int"
	Set gtmtypfldindx("sgmnt_data","def_coll")=15
	Set gtmtypes("sgmnt_data",16,"name")="sgmnt_data.def_coll_ver"
	Set gtmtypes("sgmnt_data",16,"off")=68
	Set gtmtypes("sgmnt_data",16,"len")=4
	Set gtmtypes("sgmnt_data",16,"type")="unsigned-int"
	Set gtmtypfldindx("sgmnt_data","def_coll_ver")=16
	Set gtmtypes("sgmnt_data",17,"name")="sgmnt_data.std_null_coll"
	Set gtmtypes("sgmnt_data",17,"off")=72
	Set gtmtypes("sgmnt_data",17,"len")=4
	Set gtmtypes("sgmnt_data",17,"type")="boolean_t"
	Set gtmtypfldindx("sgmnt_data","std_null_coll")=17
	Set gtmtypes("sgmnt_data",18,"name")="sgmnt_data.null_subs"
	Set gtmtypes("sgmnt_data",18,"off")=76
	Set gtmtypes("sgmnt_data",18,"len")=4
	Set gtmtypes("sgmnt_data",18,"type")="int"
	Set gtmtypfldindx("sgmnt_data","null_subs")=18
	Set gtmtypes("sgmnt_data",19,"name")="sgmnt_data.free_space"
	Set gtmtypes("sgmnt_data",19,"off")=80
	Set gtmtypes("sgmnt_data",19,"len")=4
	Set gtmtypes("sgmnt_data",19,"type")="unsigned-int"
	Set gtmtypfldindx("sgmnt_data","free_space")=19
	Set gtmtypes("sgmnt_data",20,"name")="sgmnt_data.mutex_spin_parms"
	Set gtmtypes("sgmnt_data",20,"off")=84
	Set gtmtypes("sgmnt_data",20,"len")=16
	Set gtmtypes("sgmnt_data",20,"type")="mutex_spin_parms_struct"
	Set gtmtypfldindx("sgmnt_data","mutex_spin_parms")=20
	Set gtmtypes("sgmnt_data",21,"name")="sgmnt_data.mutex_spin_parms.mutex_hard_spin_count"
	Set gtmtypes("sgmnt_data",21,"off")=84
	Set gtmtypes("sgmnt_data",21,"len")=4
	Set gtmtypes("sgmnt_data",21,"type")="int"
	Set gtmtypfldindx("sgmnt_data","mutex_spin_parms.mutex_hard_spin_count")=21
	Set gtmtypes("sgmnt_data",22,"name")="sgmnt_data.mutex_spin_parms.mutex_sleep_spin_count"
	Set gtmtypes("sgmnt_data",22,"off")=88
	Set gtmtypes("sgmnt_data",22,"len")=4
	Set gtmtypes("sgmnt_data",22,"type")="int"
	Set gtmtypfldindx("sgmnt_data","mutex_spin_parms.mutex_sleep_spin_count")=22
	Set gtmtypes("sgmnt_data",23,"name")="sgmnt_data.mutex_spin_parms.mutex_spin_sleep_mask"
	Set gtmtypes("sgmnt_data",23,"off")=92
	Set gtmtypes("sgmnt_data",23,"len")=4
	Set gtmtypes("sgmnt_data",23,"type")="int"
	Set gtmtypfldindx("sgmnt_data","mutex_spin_parms.mutex_spin_sleep_mask")=23
	Set gtmtypes("sgmnt_data",24,"name")="sgmnt_data.mutex_spin_parms.mutex_que_entry_space_size"
	Set gtmtypes("sgmnt_data",24,"off")=96
	Set gtmtypes("sgmnt_data",24,"len")=4
	Set gtmtypes("sgmnt_data",24,"type")="int"
	Set gtmtypfldindx("sgmnt_data","mutex_spin_parms.mutex_que_entry_space_size")=24
	Set gtmtypes("sgmnt_data",25,"name")="sgmnt_data.max_update_array_size"
	Set gtmtypes("sgmnt_data",25,"off")=100
	Set gtmtypes("sgmnt_data",25,"len")=4
	Set gtmtypes("sgmnt_data",25,"type")="int"
	Set gtmtypfldindx("sgmnt_data","max_update_array_size")=25
	Set gtmtypes("sgmnt_data",26,"name")="sgmnt_data.max_non_bm_update_array_size"
	Set gtmtypes("sgmnt_data",26,"off")=104
	Set gtmtypes("sgmnt_data",26,"len")=4
	Set gtmtypes("sgmnt_data",26,"type")="int"
	Set gtmtypfldindx("sgmnt_data","max_non_bm_update_array_size")=26
	Set gtmtypes("sgmnt_data",27,"name")="sgmnt_data.file_corrupt"
	Set gtmtypes("sgmnt_data",27,"off")=108
	Set gtmtypes("sgmnt_data",27,"len")=4
	Set gtmtypes("sgmnt_data",27,"type")="boolean_t"
	Set gtmtypfldindx("sgmnt_data","file_corrupt")=27
	Set gtmtypes("sgmnt_data",28,"name")="sgmnt_data.minor_dbver"
	Set gtmtypes("sgmnt_data",28,"off")=112
	Set gtmtypes("sgmnt_data",28,"len")=4
	Set gtmtypes("sgmnt_data",28,"type")="int"
	Set gtmtypfldindx("sgmnt_data","minor_dbver")=28
	Set gtmtypes("sgmnt_data",29,"name")="sgmnt_data.jnl_checksum"
	Set gtmtypes("sgmnt_data",29,"off")=116
	Set gtmtypes("sgmnt_data",29,"len")=4
	Set gtmtypes("sgmnt_data",29,"type")="unsigned-int"
	Set gtmtypfldindx("sgmnt_data","jnl_checksum")=29
	Set gtmtypes("sgmnt_data",30,"name")="sgmnt_data.wcs_phase2_commit_wait_spincnt"
	Set gtmtypes("sgmnt_data",30,"off")=120
	Set gtmtypes("sgmnt_data",30,"len")=4
	Set gtmtypes("sgmnt_data",30,"type")="unsigned-int"
	Set gtmtypfldindx("sgmnt_data","wcs_phase2_commit_wait_spincnt")=30
	Set gtmtypes("sgmnt_data",31,"name")="sgmnt_data.last_mdb_ver"
	Set gtmtypes("sgmnt_data",31,"off")=124
	Set gtmtypes("sgmnt_data",31,"len")=4
	Set gtmtypes("sgmnt_data",31,"type")="int"
	Set gtmtypfldindx("sgmnt_data","last_mdb_ver")=31
	Set gtmtypes("sgmnt_data",32,"name")="sgmnt_data.filler_created"
	Set gtmtypes("sgmnt_data",32,"off")=128
	Set gtmtypes("sgmnt_data",32,"len")=52
	Set gtmtypes("sgmnt_data",32,"type")="char"
	Set gtmtypfldindx("sgmnt_data","filler_created")=32
	Set gtmtypes("sgmnt_data",33,"name")="sgmnt_data.createinprogress"
	Set gtmtypes("sgmnt_data",33,"off")=180
	Set gtmtypes("sgmnt_data",33,"len")=4
	Set gtmtypes("sgmnt_data",33,"type")="boolean_t"
	Set gtmtypfldindx("sgmnt_data","createinprogress")=33
	Set gtmtypes("sgmnt_data",34,"name")="sgmnt_data.creation_time4"
	Set gtmtypes("sgmnt_data",34,"off")=184
	Set gtmtypes("sgmnt_data",34,"len")=4
	Set gtmtypes("sgmnt_data",34,"type")="int"
	Set gtmtypfldindx("sgmnt_data","creation_time4")=34
	Set gtmtypes("sgmnt_data",35,"name")="sgmnt_data.reservedDBFlags"
	Set gtmtypes("sgmnt_data",35,"off")=188
	Set gtmtypes("sgmnt_data",35,"len")=4
	Set gtmtypes("sgmnt_data",35,"type")="unsigned-int"
	Set gtmtypfldindx("sgmnt_data","reservedDBFlags")=35
	Set gtmtypes("sgmnt_data",36,"name")="sgmnt_data.max_tn"
	Set gtmtypes("sgmnt_data",36,"off")=192
	Set gtmtypes("sgmnt_data",36,"len")=8
	Set gtmtypes("sgmnt_data",36,"type")="uint64_t"
	Set gtmtypfldindx("sgmnt_data","max_tn")=36
	Set gtmtypes("sgmnt_data",37,"name")="sgmnt_data.max_tn_warn"
	Set gtmtypes("sgmnt_data",37,"off")=200
	Set gtmtypes("sgmnt_data",37,"len")=8
	Set gtmtypes("sgmnt_data",37,"type")="uint64_t"
	Set gtmtypfldindx("sgmnt_data","max_tn_warn")=37
	Set gtmtypes("sgmnt_data",38,"name")="sgmnt_data.last_inc_backup"
	Set gtmtypes("sgmnt_data",38,"off")=208
	Set gtmtypes("sgmnt_data",38,"len")=8
	Set gtmtypes("sgmnt_data",38,"type")="uint64_t"
	Set gtmtypfldindx("sgmnt_data","last_inc_backup")=38
	Set gtmtypes("sgmnt_data",39,"name")="sgmnt_data.last_com_backup"
	Set gtmtypes("sgmnt_data",39,"off")=216
	Set gtmtypes("sgmnt_data",39,"len")=8
	Set gtmtypes("sgmnt_data",39,"type")="uint64_t"
	Set gtmtypfldindx("sgmnt_data","last_com_backup")=39
	Set gtmtypes("sgmnt_data",40,"name")="sgmnt_data.last_rec_backup"
	Set gtmtypes("sgmnt_data",40,"off")=224
	Set gtmtypes("sgmnt_data",40,"len")=8
	Set gtmtypes("sgmnt_data",40,"type")="uint64_t"
	Set gtmtypfldindx("sgmnt_data","last_rec_backup")=40
	Set gtmtypes("sgmnt_data",41,"name")="sgmnt_data.last_inc_bkup_last_blk"
	Set gtmtypes("sgmnt_data",41,"off")=232
	Set gtmtypes("sgmnt_data",41,"len")=4
	Set gtmtypes("sgmnt_data",41,"type")="int"
	Set gtmtypfldindx("sgmnt_data","last_inc_bkup_last_blk")=41
	Set gtmtypes("sgmnt_data",42,"name")="sgmnt_data.last_com_bkup_last_blk"
	Set gtmtypes("sgmnt_data",42,"off")=236
	Set gtmtypes("sgmnt_data",42,"len")=4
	Set gtmtypes("sgmnt_data",42,"type")="int"
	Set gtmtypfldindx("sgmnt_data","last_com_bkup_last_blk")=42
	Set gtmtypes("sgmnt_data",43,"name")="sgmnt_data.last_rec_bkup_last_blk"
	Set gtmtypes("sgmnt_data",43,"off")=240
	Set gtmtypes("sgmnt_data",43,"len")=4
	Set gtmtypes("sgmnt_data",43,"type")="int"
	Set gtmtypfldindx("sgmnt_data","last_rec_bkup_last_blk")=43
	Set gtmtypes("sgmnt_data",44,"name")="sgmnt_data.reorg_restart_block"
	Set gtmtypes("sgmnt_data",44,"off")=244
	Set gtmtypes("sgmnt_data",44,"len")=4
	Set gtmtypes("sgmnt_data",44,"type")="int"
	Set gtmtypfldindx("sgmnt_data","reorg_restart_block")=44
	Set gtmtypes("sgmnt_data",45,"name")="sgmnt_data.filler_256"
	Set gtmtypes("sgmnt_data",45,"off")=248
	Set gtmtypes("sgmnt_data",45,"len")=8
	Set gtmtypes("sgmnt_data",45,"type")="char"
	Set gtmtypfldindx("sgmnt_data","filler_256")=45
	Set gtmtypes("sgmnt_data",46,"name")="sgmnt_data.now_running"
	Set gtmtypes("sgmnt_data",46,"off")=256
	Set gtmtypes("sgmnt_data",46,"len")=36
	Set gtmtypes("sgmnt_data",46,"type")="char"
	Set gtmtypfldindx("sgmnt_data","now_running")=46
	Set gtmtypes("sgmnt_data",47,"name")="sgmnt_data.filler_owner_node"
	Set gtmtypes("sgmnt_data",47,"off")=292
	Set gtmtypes("sgmnt_data",47,"len")=4
	Set gtmtypes("sgmnt_data",47,"type")="unsigned-int"
	Set gtmtypfldindx("sgmnt_data","filler_owner_node")=47
	Set gtmtypes("sgmnt_data",48,"name")="sgmnt_data.image_count"
	Set gtmtypes("sgmnt_data",48,"off")=296
	Set gtmtypes("sgmnt_data",48,"len")=4
	Set gtmtypes("sgmnt_data",48,"type")="unsigned-int"
	Set gtmtypfldindx("sgmnt_data","image_count")=48
	Set gtmtypes("sgmnt_data",49,"name")="sgmnt_data.freeze"
	Set gtmtypes("sgmnt_data",49,"off")=300
	Set gtmtypes("sgmnt_data",49,"len")=4
	Set gtmtypes("sgmnt_data",49,"type")="unsigned-int"
	Set gtmtypfldindx("sgmnt_data","freeze")=49
	Set gtmtypes("sgmnt_data",50,"name")="sgmnt_data.kill_in_prog"
	Set gtmtypes("sgmnt_data",50,"off")=304
	Set gtmtypes("sgmnt_data",50,"len")=4
	Set gtmtypes("sgmnt_data",50,"type")="int"
	Set gtmtypfldindx("sgmnt_data","kill_in_prog")=50
	Set gtmtypes("sgmnt_data",51,"name")="sgmnt_data.abandoned_kills"
	Set gtmtypes("sgmnt_data",51,"off")=308
	Set gtmtypes("sgmnt_data",51,"len")=4
	Set gtmtypes("sgmnt_data",51,"type")="int"
	Set gtmtypfldindx("sgmnt_data","abandoned_kills")=51
	Set gtmtypes("sgmnt_data",52,"name")="sgmnt_data.unused_freeze_online_filler"
	Set gtmtypes("sgmnt_data",52,"off")=312
	Set gtmtypes("sgmnt_data",52,"len")=4
	Set gtmtypes("sgmnt_data",52,"type")="unsigned-int"
	Set gtmtypfldindx("sgmnt_data","unused_freeze_online_filler")=52
	Set gtmtypes("sgmnt_data",53,"name")="sgmnt_data.filler_320"
	Set gtmtypes("sgmnt_data",53,"off")=316
	Set gtmtypes("sgmnt_data",53,"len")=4
	Set gtmtypes("sgmnt_data",53,"type")="char"
	Set gtmtypfldindx("sgmnt_data","filler_320")=53
	Set gtmtypes("sgmnt_data",54,"name")="sgmnt_data.tn_upgrd_blks_0"
	Set gtmtypes("sgmnt_data",54,"off")=320
	Set gtmtypes("sgmnt_data",54,"len")=8
	Set gtmtypes("sgmnt_data",54,"type")="uint64_t"
	Set gtmtypfldindx("sgmnt_data","tn_upgrd_blks_0")=54
	Set gtmtypes("sgmnt_data",55,"name")="sgmnt_data.desired_db_format_tn"
	Set gtmtypes("sgmnt_data",55,"off")=328
	Set gtmtypes("sgmnt_data",55,"len")=8
	Set gtmtypes("sgmnt_data",55,"type")="uint64_t"
	Set gtmtypfldindx("sgmnt_data","desired_db_format_tn")=55
	Set gtmtypes("sgmnt_data",56,"name")="sgmnt_data.reorg_db_fmt_start_tn"
	Set gtmtypes("sgmnt_data",56,"off")=336
	Set gtmtypes("sgmnt_data",56,"len")=8
	Set gtmtypes("sgmnt_data",56,"type")="uint64_t"
	Set gtmtypfldindx("sgmnt_data","reorg_db_fmt_start_tn")=56
	Set gtmtypes("sgmnt_data",57,"name")="sgmnt_data.reorg_upgrd_dwngrd_restart_block"
	Set gtmtypes("sgmnt_data",57,"off")=344
	Set gtmtypes("sgmnt_data",57,"len")=4
	Set gtmtypes("sgmnt_data",57,"type")="int"
	Set gtmtypfldindx("sgmnt_data","reorg_upgrd_dwngrd_restart_block")=57
	Set gtmtypes("sgmnt_data",58,"name")="sgmnt_data.blks_to_upgrd"
	Set gtmtypes("sgmnt_data",58,"off")=348
	Set gtmtypes("sgmnt_data",58,"len")=4
	Set gtmtypes("sgmnt_data",58,"type")="int"
	Set gtmtypfldindx("sgmnt_data","blks_to_upgrd")=58
	Set gtmtypes("sgmnt_data",59,"name")="sgmnt_data.blks_to_upgrd_subzero_error"
	Set gtmtypes("sgmnt_data",59,"off")=352
	Set gtmtypes("sgmnt_data",59,"len")=4
	Set gtmtypes("sgmnt_data",59,"type")="int"
	Set gtmtypfldindx("sgmnt_data","blks_to_upgrd_subzero_error")=59
	Set gtmtypes("sgmnt_data",60,"name")="sgmnt_data.desired_db_format"
	Set gtmtypes("sgmnt_data",60,"off")=356
	Set gtmtypes("sgmnt_data",60,"len")=4
	Set gtmtypes("sgmnt_data",60,"type")="int"
	Set gtmtypfldindx("sgmnt_data","desired_db_format")=60
	Set gtmtypes("sgmnt_data",61,"name")="sgmnt_data.fully_upgraded"
	Set gtmtypes("sgmnt_data",61,"off")=360
	Set gtmtypes("sgmnt_data",61,"len")=4
	Set gtmtypes("sgmnt_data",61,"type")="boolean_t"
	Set gtmtypfldindx("sgmnt_data","fully_upgraded")=61
	Set gtmtypes("sgmnt_data",62,"name")="sgmnt_data.db_got_to_v5_once"
	Set gtmtypes("sgmnt_data",62,"off")=364
	Set gtmtypes("sgmnt_data",62,"len")=4
	Set gtmtypes("sgmnt_data",62,"type")="boolean_t"
	Set gtmtypfldindx("sgmnt_data","db_got_to_v5_once")=62
	Set gtmtypes("sgmnt_data",63,"name")="sgmnt_data.opened_by_gtmv53"
	Set gtmtypes("sgmnt_data",63,"off")=368
	Set gtmtypes("sgmnt_data",63,"len")=4
	Set gtmtypes("sgmnt_data",63,"type")="boolean_t"
	Set gtmtypfldindx("sgmnt_data","opened_by_gtmv53")=63
	Set gtmtypes("sgmnt_data",64,"name")="sgmnt_data.filler_384"
	Set gtmtypes("sgmnt_data",64,"off")=372
	Set gtmtypes("sgmnt_data",64,"len")=12
	Set gtmtypes("sgmnt_data",64,"type")="char"
	Set gtmtypfldindx("sgmnt_data","filler_384")=64
	Set gtmtypes("sgmnt_data",65,"name")="sgmnt_data.trans_hist"
	Set gtmtypes("sgmnt_data",65,"off")=384
	Set gtmtypes("sgmnt_data",65,"len")=56
	Set gtmtypes("sgmnt_data",65,"type")="th_index"
	Set gtmtypfldindx("sgmnt_data","trans_hist")=65
	Set gtmtypes("sgmnt_data",66,"name")="sgmnt_data.trans_hist.curr_tn"
	Set gtmtypes("sgmnt_data",66,"off")=384
	Set gtmtypes("sgmnt_data",66,"len")=8
	Set gtmtypes("sgmnt_data",66,"type")="uint64_t"
	Set gtmtypfldindx("sgmnt_data","trans_hist.curr_tn")=66
	Set gtmtypes("sgmnt_data",67,"name")="sgmnt_data.trans_hist.early_tn"
	Set gtmtypes("sgmnt_data",67,"off")=392
	Set gtmtypes("sgmnt_data",67,"len")=8
	Set gtmtypes("sgmnt_data",67,"type")="uint64_t"
	Set gtmtypfldindx("sgmnt_data","trans_hist.early_tn")=67
	Set gtmtypes("sgmnt_data",68,"name")="sgmnt_data.trans_hist.last_mm_sync"
	Set gtmtypes("sgmnt_data",68,"off")=400
	Set gtmtypes("sgmnt_data",68,"len")=8
	Set gtmtypes("sgmnt_data",68,"type")="uint64_t"
	Set gtmtypfldindx("sgmnt_data","trans_hist.last_mm_sync")=68
	Set gtmtypes("sgmnt_data",69,"name")="sgmnt_data.trans_hist.filler_8byte"
	Set gtmtypes("sgmnt_data",69,"off")=408
	Set gtmtypes("sgmnt_data",69,"len")=8
	Set gtmtypes("sgmnt_data",69,"type")="char"
	Set gtmtypfldindx("sgmnt_data","trans_hist.filler_8byte")=69
	Set gtmtypes("sgmnt_data",70,"name")="sgmnt_data.trans_hist.mm_tn"
	Set gtmtypes("sgmnt_data",70,"off")=416
	Set gtmtypes("sgmnt_data",70,"len")=8
	Set gtmtypes("sgmnt_data",70,"type")="uint64_t"
	Set gtmtypfldindx("sgmnt_data","trans_hist.mm_tn")=70
	Set gtmtypes("sgmnt_data",71,"name")="sgmnt_data.trans_hist.lock_sequence"
	Set gtmtypes("sgmnt_data",71,"off")=424
	Set gtmtypes("sgmnt_data",71,"len")=4
	Set gtmtypes("sgmnt_data",71,"type")="unsigned-int"
	Set gtmtypfldindx("sgmnt_data","trans_hist.lock_sequence")=71
	Set gtmtypes("sgmnt_data",72,"name")="sgmnt_data.trans_hist.ccp_jnl_filesize"
	Set gtmtypes("sgmnt_data",72,"off")=428
	Set gtmtypes("sgmnt_data",72,"len")=4
	Set gtmtypes("sgmnt_data",72,"type")="unsigned-int"
	Set gtmtypfldindx("sgmnt_data","trans_hist.ccp_jnl_filesize")=72
	Set gtmtypes("sgmnt_data",73,"name")="sgmnt_data.trans_hist.total_blks"
	Set gtmtypes("sgmnt_data",73,"off")=432
	Set gtmtypes("sgmnt_data",73,"len")=4
	Set gtmtypes("sgmnt_data",73,"type")="unsigned-int"
	Set gtmtypfldindx("sgmnt_data","trans_hist.total_blks")=73
	Set gtmtypes("sgmnt_data",74,"name")="sgmnt_data.trans_hist.free_blocks"
	Set gtmtypes("sgmnt_data",74,"off")=436
	Set gtmtypes("sgmnt_data",74,"len")=4
	Set gtmtypes("sgmnt_data",74,"type")="unsigned-int"
	Set gtmtypfldindx("sgmnt_data","trans_hist.free_blocks")=74
	Set gtmtypes("sgmnt_data",75,"name")="sgmnt_data.write_fullblk"
	Set gtmtypes("sgmnt_data",75,"off")=440
	Set gtmtypes("sgmnt_data",75,"len")=4
	Set gtmtypes("sgmnt_data",75,"type")="int"
	Set gtmtypfldindx("sgmnt_data","write_fullblk")=75
	Set gtmtypes("sgmnt_data",76,"name")="sgmnt_data.filler"
	Set gtmtypes("sgmnt_data",76,"off")=444
	Set gtmtypes("sgmnt_data",76,"len")=4
	Set gtmtypes("sgmnt_data",76,"type")="char"
	Set gtmtypfldindx("sgmnt_data","filler")=76
	Set gtmtypes("sgmnt_data",77,"name")="sgmnt_data.flush_time"
	Set gtmtypes("sgmnt_data",77,"off")=448
	Set gtmtypes("sgmnt_data",77,"len")=8
	Set gtmtypes("sgmnt_data",77,"type")="int"
	Set gtmtypfldindx("sgmnt_data","flush_time")=77
	Set gtmtypes("sgmnt_data",77,"dim")=2
	Set gtmtypes("sgmnt_data",78,"name")="sgmnt_data.flush_trigger"
	Set gtmtypes("sgmnt_data",78,"off")=456
	Set gtmtypes("sgmnt_data",78,"len")=4
	Set gtmtypes("sgmnt_data",78,"type")="int"
	Set gtmtypfldindx("sgmnt_data","flush_trigger")=78
	Set gtmtypes("sgmnt_data",79,"name")="sgmnt_data.n_wrt_per_flu"
	Set gtmtypes("sgmnt_data",79,"off")=460
	Set gtmtypes("sgmnt_data",79,"len")=4
	Set gtmtypes("sgmnt_data",79,"type")="int"
	Set gtmtypfldindx("sgmnt_data","n_wrt_per_flu")=79
	Set gtmtypes("sgmnt_data",80,"name")="sgmnt_data.wait_disk_space"
	Set gtmtypes("sgmnt_data",80,"off")=464
	Set gtmtypes("sgmnt_data",80,"len")=4
	Set gtmtypes("sgmnt_data",80,"type")="int"
	Set gtmtypfldindx("sgmnt_data","wait_disk_space")=80
	Set gtmtypes("sgmnt_data",81,"name")="sgmnt_data.defer_time"
	Set gtmtypes("sgmnt_data",81,"off")=468
	Set gtmtypes("sgmnt_data",81,"len")=4
	Set gtmtypes("sgmnt_data",81,"type")="int"
	Set gtmtypfldindx("sgmnt_data","defer_time")=81
	Set gtmtypes("sgmnt_data",82,"name")="sgmnt_data.flush_trigger_top"
	Set gtmtypes("sgmnt_data",82,"off")=472
	Set gtmtypes("sgmnt_data",82,"len")=4
	Set gtmtypes("sgmnt_data",82,"type")="int"
	Set gtmtypfldindx("sgmnt_data","flush_trigger_top")=82
	Set gtmtypes("sgmnt_data",83,"name")="sgmnt_data.mumps_can_bypass"
	Set gtmtypes("sgmnt_data",83,"off")=476
	Set gtmtypes("sgmnt_data",83,"len")=4
	Set gtmtypes("sgmnt_data",83,"type")="boolean_t"
	Set gtmtypfldindx("sgmnt_data","mumps_can_bypass")=83
	Set gtmtypes("sgmnt_data",84,"name")="sgmnt_data.epoch_taper"
	Set gtmtypes("sgmnt_data",84,"off")=480
	Set gtmtypes("sgmnt_data",84,"len")=4
	Set gtmtypes("sgmnt_data",84,"type")="boolean_t"
	Set gtmtypfldindx("sgmnt_data","epoch_taper")=84
	Set gtmtypes("sgmnt_data",85,"name")="sgmnt_data.epoch_taper_time_pct"
	Set gtmtypes("sgmnt_data",85,"off")=484
	Set gtmtypes("sgmnt_data",85,"len")=4
	Set gtmtypes("sgmnt_data",85,"type")="unsigned-int"
	Set gtmtypfldindx("sgmnt_data","epoch_taper_time_pct")=85
	Set gtmtypes("sgmnt_data",86,"name")="sgmnt_data.epoch_taper_jnl_pct"
	Set gtmtypes("sgmnt_data",86,"off")=488
	Set gtmtypes("sgmnt_data",86,"len")=4
	Set gtmtypes("sgmnt_data",86,"type")="unsigned-int"
	Set gtmtypfldindx("sgmnt_data","epoch_taper_jnl_pct")=86
	Set gtmtypes("sgmnt_data",87,"name")="sgmnt_data.asyncio"
	Set gtmtypes("sgmnt_data",87,"off")=492
	Set gtmtypes("sgmnt_data",87,"len")=4
	Set gtmtypes("sgmnt_data",87,"type")="boolean_t"
	Set gtmtypfldindx("sgmnt_data","asyncio")=87
	Set gtmtypes("sgmnt_data",88,"name")="sgmnt_data.reserved_for_upd"
	Set gtmtypes("sgmnt_data",88,"off")=496
	Set gtmtypes("sgmnt_data",88,"len")=4
	Set gtmtypes("sgmnt_data",88,"type")="unsigned-int"
	Set gtmtypfldindx("sgmnt_data","reserved_for_upd")=88
	Set gtmtypes("sgmnt_data",89,"name")="sgmnt_data.avg_blks_per_100gbl"
	Set gtmtypes("sgmnt_data",89,"off")=500
	Set gtmtypes("sgmnt_data",89,"len")=4
	Set gtmtypes("sgmnt_data",89,"type")="unsigned-int"
	Set gtmtypfldindx("sgmnt_data","avg_blks_per_100gbl")=89
	Set gtmtypes("sgmnt_data",90,"name")="sgmnt_data.pre_read_trigger_factor"
	Set gtmtypes("sgmnt_data",90,"off")=504
	Set gtmtypes("sgmnt_data",90,"len")=4
	Set gtmtypes("sgmnt_data",90,"type")="unsigned-int"
	Set gtmtypfldindx("sgmnt_data","pre_read_trigger_factor")=90
	Set gtmtypes("sgmnt_data",91,"name")="sgmnt_data.writer_trigger_factor"
	Set gtmtypes("sgmnt_data",91,"off")=508
	Set gtmtypes("sgmnt_data",91,"len")=4
	Set gtmtypes("sgmnt_data",91,"type")="unsigned-int"
	Set gtmtypfldindx("sgmnt_data","writer_trigger_factor")=91
	Set gtmtypes("sgmnt_data",92,"name")="sgmnt_data.semid"
	Set gtmtypes("sgmnt_data",92,"off")=512
	Set gtmtypes("sgmnt_data",92,"len")=4
	Set gtmtypes("sgmnt_data",92,"type")="int"
	Set gtmtypfldindx("sgmnt_data","semid")=92
	Set gtmtypes("sgmnt_data",93,"name")="sgmnt_data.shmid"
	Set gtmtypes("sgmnt_data",93,"off")=516
	Set gtmtypes("sgmnt_data",93,"len")=4
	Set gtmtypes("sgmnt_data",93,"type")="int"
	Set gtmtypfldindx("sgmnt_data","shmid")=93
	Set gtmtypes("sgmnt_data",94,"name")="sgmnt_data.gt_sem_ctime"
	Set gtmtypes("sgmnt_data",94,"off")=520
	Set gtmtypes("sgmnt_data",94,"len")=8
	Set gtmtypes("sgmnt_data",94,"type")="gtm_time8"
	Set gtmtypfldindx("sgmnt_data","gt_sem_ctime")=94
	Set gtmtypes("sgmnt_data",95,"name")="sgmnt_data.gt_sem_ctime.ctime"
	Set gtmtypes("sgmnt_data",95,"off")=520
	Set gtmtypes("sgmnt_data",95,"len")=8
	Set gtmtypes("sgmnt_data",95,"type")="time_t"
	Set gtmtypfldindx("sgmnt_data","gt_sem_ctime.ctime")=95
	Set gtmtypes("sgmnt_data",96,"name")="sgmnt_data.gt_sem_ctime.filler"
	Set gtmtypes("sgmnt_data",96,"off")=520
	Set gtmtypes("sgmnt_data",96,"len")=8
	Set gtmtypes("sgmnt_data",96,"type")="int"
	Set gtmtypfldindx("sgmnt_data","gt_sem_ctime.filler")=96
	Set gtmtypes("sgmnt_data",96,"dim")=2
	Set gtmtypes("sgmnt_data",97,"name")="sgmnt_data.gt_shm_ctime"
	Set gtmtypes("sgmnt_data",97,"off")=528
	Set gtmtypes("sgmnt_data",97,"len")=8
	Set gtmtypes("sgmnt_data",97,"type")="gtm_time8"
	Set gtmtypfldindx("sgmnt_data","gt_shm_ctime")=97
	Set gtmtypes("sgmnt_data",98,"name")="sgmnt_data.gt_shm_ctime.ctime"
	Set gtmtypes("sgmnt_data",98,"off")=528
	Set gtmtypes("sgmnt_data",98,"len")=8
	Set gtmtypes("sgmnt_data",98,"type")="time_t"
	Set gtmtypfldindx("sgmnt_data","gt_shm_ctime.ctime")=98
	Set gtmtypes("sgmnt_data",99,"name")="sgmnt_data.gt_shm_ctime.filler"
	Set gtmtypes("sgmnt_data",99,"off")=528
	Set gtmtypes("sgmnt_data",99,"len")=8
	Set gtmtypes("sgmnt_data",99,"type")="int"
	Set gtmtypfldindx("sgmnt_data","gt_shm_ctime.filler")=99
	Set gtmtypes("sgmnt_data",99,"dim")=2
	Set gtmtypes("sgmnt_data",100,"name")="sgmnt_data.filler_unixonly"
	Set gtmtypes("sgmnt_data",100,"off")=536
	Set gtmtypes("sgmnt_data",100,"len")=40
	Set gtmtypes("sgmnt_data",100,"type")="char"
	Set gtmtypfldindx("sgmnt_data","filler_unixonly")=100
	Set gtmtypes("sgmnt_data",101,"name")="sgmnt_data.filler_n_retries"
	Set gtmtypes("sgmnt_data",101,"off")=576
	Set gtmtypes("sgmnt_data",101,"len")=20
	Set gtmtypes("sgmnt_data",101,"type")="int"
	Set gtmtypfldindx("sgmnt_data","filler_n_retries")=101
	Set gtmtypes("sgmnt_data",101,"dim")=5
	Set gtmtypes("sgmnt_data",102,"name")="sgmnt_data.filler_n_puts"
	Set gtmtypes("sgmnt_data",102,"off")=596
	Set gtmtypes("sgmnt_data",102,"len")=4
	Set gtmtypes("sgmnt_data",102,"type")="int"
	Set gtmtypfldindx("sgmnt_data","filler_n_puts")=102
	Set gtmtypes("sgmnt_data",103,"name")="sgmnt_data.filler_n_kills"
	Set gtmtypes("sgmnt_data",103,"off")=600
	Set gtmtypes("sgmnt_data",103,"len")=4
	Set gtmtypes("sgmnt_data",103,"type")="int"
	Set gtmtypfldindx("sgmnt_data","filler_n_kills")=103
	Set gtmtypes("sgmnt_data",104,"name")="sgmnt_data.filler_n_queries"
	Set gtmtypes("sgmnt_data",104,"off")=604
	Set gtmtypes("sgmnt_data",104,"len")=4
	Set gtmtypes("sgmnt_data",104,"type")="int"
	Set gtmtypfldindx("sgmnt_data","filler_n_queries")=104
	Set gtmtypes("sgmnt_data",105,"name")="sgmnt_data.filler_n_gets"
	Set gtmtypes("sgmnt_data",105,"off")=608
	Set gtmtypes("sgmnt_data",105,"len")=4
	Set gtmtypes("sgmnt_data",105,"type")="int"
	Set gtmtypfldindx("sgmnt_data","filler_n_gets")=105
	Set gtmtypes("sgmnt_data",106,"name")="sgmnt_data.filler_n_order"
	Set gtmtypes("sgmnt_data",106,"off")=612
	Set gtmtypes("sgmnt_data",106,"len")=4
	Set gtmtypes("sgmnt_data",106,"type")="int"
	Set gtmtypfldindx("sgmnt_data","filler_n_order")=106
	Set gtmtypes("sgmnt_data",107,"name")="sgmnt_data.filler_n_zprevs"
	Set gtmtypes("sgmnt_data",107,"off")=616
	Set gtmtypes("sgmnt_data",107,"len")=4
	Set gtmtypes("sgmnt_data",107,"type")="int"
	Set gtmtypfldindx("sgmnt_data","filler_n_zprevs")=107
	Set gtmtypes("sgmnt_data",108,"name")="sgmnt_data.filler_n_data"
	Set gtmtypes("sgmnt_data",108,"off")=620
	Set gtmtypes("sgmnt_data",108,"len")=4
	Set gtmtypes("sgmnt_data",108,"type")="int"
	Set gtmtypfldindx("sgmnt_data","filler_n_data")=108
	Set gtmtypes("sgmnt_data",109,"name")="sgmnt_data.filler_n_puts_duplicate"
	Set gtmtypes("sgmnt_data",109,"off")=624
	Set gtmtypes("sgmnt_data",109,"len")=4
	Set gtmtypes("sgmnt_data",109,"type")="unsigned-int"
	Set gtmtypfldindx("sgmnt_data","filler_n_puts_duplicate")=109
	Set gtmtypes("sgmnt_data",110,"name")="sgmnt_data.filler_n_tp_updates"
	Set gtmtypes("sgmnt_data",110,"off")=628
	Set gtmtypes("sgmnt_data",110,"len")=4
	Set gtmtypes("sgmnt_data",110,"type")="unsigned-int"
	Set gtmtypfldindx("sgmnt_data","filler_n_tp_updates")=110
	Set gtmtypes("sgmnt_data",111,"name")="sgmnt_data.filler_n_tp_updates_duplicate"
	Set gtmtypes("sgmnt_data",111,"off")=632
	Set gtmtypes("sgmnt_data",111,"len")=4
	Set gtmtypes("sgmnt_data",111,"type")="unsigned-int"
	Set gtmtypfldindx("sgmnt_data","filler_n_tp_updates_duplicate")=111
	Set gtmtypes("sgmnt_data",112,"name")="sgmnt_data.filler_accounting_64_align"
	Set gtmtypes("sgmnt_data",112,"off")=636
	Set gtmtypes("sgmnt_data",112,"len")=4
	Set gtmtypes("sgmnt_data",112,"type")="char"
	Set gtmtypfldindx("sgmnt_data","filler_accounting_64_align")=112
	Set gtmtypes("sgmnt_data",113,"name")="sgmnt_data.staleness"
	Set gtmtypes("sgmnt_data",113,"off")=640
	Set gtmtypes("sgmnt_data",113,"len")=8
	Set gtmtypes("sgmnt_data",113,"type")="int"
	Set gtmtypfldindx("sgmnt_data","staleness")=113
	Set gtmtypes("sgmnt_data",113,"dim")=2
	Set gtmtypes("sgmnt_data",114,"name")="sgmnt_data.ccp_tick_interval"
	Set gtmtypes("sgmnt_data",114,"off")=648
	Set gtmtypes("sgmnt_data",114,"len")=8
	Set gtmtypes("sgmnt_data",114,"type")="int"
	Set gtmtypfldindx("sgmnt_data","ccp_tick_interval")=114
	Set gtmtypes("sgmnt_data",114,"dim")=2
	Set gtmtypes("sgmnt_data",115,"name")="sgmnt_data.ccp_quantum_interval"
	Set gtmtypes("sgmnt_data",115,"off")=656
	Set gtmtypes("sgmnt_data",115,"len")=8
	Set gtmtypes("sgmnt_data",115,"type")="int"
	Set gtmtypfldindx("sgmnt_data","ccp_quantum_interval")=115
	Set gtmtypes("sgmnt_data",115,"dim")=2
	Set gtmtypes("sgmnt_data",116,"name")="sgmnt_data.ccp_response_interval"
	Set gtmtypes("sgmnt_data",116,"off")=664
	Set gtmtypes("sgmnt_data",116,"len")=8
	Set gtmtypes("sgmnt_data",116,"type")="int"
	Set gtmtypfldindx("sgmnt_data","ccp_response_interval")=116
	Set gtmtypes("sgmnt_data",116,"dim")=2
	Set gtmtypes("sgmnt_data",117,"name")="sgmnt_data.ccp_jnl_before"
	Set gtmtypes("sgmnt_data",117,"off")=672
	Set gtmtypes("sgmnt_data",117,"len")=4
	Set gtmtypes("sgmnt_data",117,"type")="boolean_t"
	Set gtmtypfldindx("sgmnt_data","ccp_jnl_before")=117
	Set gtmtypes("sgmnt_data",118,"name")="sgmnt_data.clustered"
	Set gtmtypes("sgmnt_data",118,"off")=676
	Set gtmtypes("sgmnt_data",118,"len")=4
	Set gtmtypes("sgmnt_data",118,"type")="boolean_t"
	Set gtmtypfldindx("sgmnt_data","clustered")=118
	Set gtmtypes("sgmnt_data",119,"name")="sgmnt_data.unbacked_cache"
	Set gtmtypes("sgmnt_data",119,"off")=680
	Set gtmtypes("sgmnt_data",119,"len")=4
	Set gtmtypes("sgmnt_data",119,"type")="boolean_t"
	Set gtmtypfldindx("sgmnt_data","unbacked_cache")=119
	Set gtmtypes("sgmnt_data",120,"name")="sgmnt_data.rc_srv_cnt"
	Set gtmtypes("sgmnt_data",120,"off")=684
	Set gtmtypes("sgmnt_data",120,"len")=4
	Set gtmtypes("sgmnt_data",120,"type")="int"
	Set gtmtypfldindx("sgmnt_data","rc_srv_cnt")=120
	Set gtmtypes("sgmnt_data",121,"name")="sgmnt_data.dsid"
	Set gtmtypes("sgmnt_data",121,"off")=688
	Set gtmtypes("sgmnt_data",121,"len")=4
	Set gtmtypes("sgmnt_data",121,"type")="int"
	Set gtmtypfldindx("sgmnt_data","dsid")=121
	Set gtmtypes("sgmnt_data",122,"name")="sgmnt_data.rc_node"
	Set gtmtypes("sgmnt_data",122,"off")=692
	Set gtmtypes("sgmnt_data",122,"len")=4
	Set gtmtypes("sgmnt_data",122,"type")="int"
	Set gtmtypfldindx("sgmnt_data","rc_node")=122
	Set gtmtypes("sgmnt_data",123,"name")="sgmnt_data.filler_ccp_rc"
	Set gtmtypes("sgmnt_data",123,"off")=696
	Set gtmtypes("sgmnt_data",123,"len")=8
	Set gtmtypes("sgmnt_data",123,"type")="char"
	Set gtmtypfldindx("sgmnt_data","filler_ccp_rc")=123
	Set gtmtypes("sgmnt_data",124,"name")="sgmnt_data.reg_seqno"
	Set gtmtypes("sgmnt_data",124,"off")=704
	Set gtmtypes("sgmnt_data",124,"len")=8
	Set gtmtypes("sgmnt_data",124,"type")="uint64_t"
	Set gtmtypfldindx("sgmnt_data","reg_seqno")=124
	Set gtmtypes("sgmnt_data",125,"name")="sgmnt_data.pre_multisite_resync_seqno"
	Set gtmtypes("sgmnt_data",125,"off")=712
	Set gtmtypes("sgmnt_data",125,"len")=8
	Set gtmtypes("sgmnt_data",125,"type")="uint64_t"
	Set gtmtypfldindx("sgmnt_data","pre_multisite_resync_seqno")=125
	Set gtmtypes("sgmnt_data",126,"name")="sgmnt_data.zqgblmod_tn"
	Set gtmtypes("sgmnt_data",126,"off")=720
	Set gtmtypes("sgmnt_data",126,"len")=8
	Set gtmtypes("sgmnt_data",126,"type")="uint64_t"
	Set gtmtypfldindx("sgmnt_data","zqgblmod_tn")=126
	Set gtmtypes("sgmnt_data",127,"name")="sgmnt_data.zqgblmod_seqno"
	Set gtmtypes("sgmnt_data",127,"off")=728
	Set gtmtypes("sgmnt_data",127,"len")=8
	Set gtmtypes("sgmnt_data",127,"type")="uint64_t"
	Set gtmtypfldindx("sgmnt_data","zqgblmod_seqno")=127
	Set gtmtypes("sgmnt_data",128,"name")="sgmnt_data.repl_state"
	Set gtmtypes("sgmnt_data",128,"off")=736
	Set gtmtypes("sgmnt_data",128,"len")=4
	Set gtmtypes("sgmnt_data",128,"type")="int"
	Set gtmtypfldindx("sgmnt_data","repl_state")=128
	Set gtmtypes("sgmnt_data",129,"name")="sgmnt_data.multi_site_open"
	Set gtmtypes("sgmnt_data",129,"off")=740
	Set gtmtypes("sgmnt_data",129,"len")=4
	Set gtmtypes("sgmnt_data",129,"type")="boolean_t"
	Set gtmtypfldindx("sgmnt_data","multi_site_open")=129
	Set gtmtypes("sgmnt_data",130,"name")="sgmnt_data.filler_seqno"
	Set gtmtypes("sgmnt_data",130,"off")=744
	Set gtmtypes("sgmnt_data",130,"len")=8
	Set gtmtypes("sgmnt_data",130,"type")="uint64_t"
	Set gtmtypfldindx("sgmnt_data","filler_seqno")=130
	Set gtmtypes("sgmnt_data",131,"name")="sgmnt_data.filler_repl"
	Set gtmtypes("sgmnt_data",131,"off")=752
	Set gtmtypes("sgmnt_data",131,"len")=16
	Set gtmtypes("sgmnt_data",131,"type")="char"
	Set gtmtypfldindx("sgmnt_data","filler_repl")=131
	Set gtmtypes("sgmnt_data",132,"name")="sgmnt_data.filler_n_tp_retries"
	Set gtmtypes("sgmnt_data",132,"off")=768
	Set gtmtypes("sgmnt_data",132,"len")=48
	Set gtmtypes("sgmnt_data",132,"type")="int"
	Set gtmtypfldindx("sgmnt_data","filler_n_tp_retries")=132
	Set gtmtypes("sgmnt_data",132,"dim")=12
	Set gtmtypes("sgmnt_data",133,"name")="sgmnt_data.filler_n_tp_retries_conflicts"
	Set gtmtypes("sgmnt_data",133,"off")=816
	Set gtmtypes("sgmnt_data",133,"len")=48
	Set gtmtypes("sgmnt_data",133,"type")="int"
	Set gtmtypfldindx("sgmnt_data","filler_n_tp_retries_conflicts")=133
	Set gtmtypes("sgmnt_data",133,"dim")=12
	Set gtmtypes("sgmnt_data",134,"name")="sgmnt_data.tp_cdb_sc_blkmod"
	Set gtmtypes("sgmnt_data",134,"off")=864
	Set gtmtypes("sgmnt_data",134,"len")=32
	Set gtmtypes("sgmnt_data",134,"type")="int"
	Set gtmtypfldindx("sgmnt_data","tp_cdb_sc_blkmod")=134
	Set gtmtypes("sgmnt_data",134,"dim")=8
	Set gtmtypes("sgmnt_data",135,"name")="sgmnt_data.jnl_alq"
	Set gtmtypes("sgmnt_data",135,"off")=896
	Set gtmtypes("sgmnt_data",135,"len")=4
	Set gtmtypes("sgmnt_data",135,"type")="unsigned-int"
	Set gtmtypfldindx("sgmnt_data","jnl_alq")=135
	Set gtmtypes("sgmnt_data",136,"name")="sgmnt_data.jnl_deq"
	Set gtmtypes("sgmnt_data",136,"off")=900
	Set gtmtypes("sgmnt_data",136,"len")=4
	Set gtmtypes("sgmnt_data",136,"type")="unsigned-int"
	Set gtmtypfldindx("sgmnt_data","jnl_deq")=136
	Set gtmtypes("sgmnt_data",137,"name")="sgmnt_data.jnl_buffer_size"
	Set gtmtypes("sgmnt_data",137,"off")=904
	Set gtmtypes("sgmnt_data",137,"len")=4
	Set gtmtypes("sgmnt_data",137,"type")="int"
	Set gtmtypfldindx("sgmnt_data","jnl_buffer_size")=137
	Set gtmtypes("sgmnt_data",138,"name")="sgmnt_data.jnl_before_image"
	Set gtmtypes("sgmnt_data",138,"off")=908
	Set gtmtypes("sgmnt_data",138,"len")=4
	Set gtmtypes("sgmnt_data",138,"type")="boolean_t"
	Set gtmtypfldindx("sgmnt_data","jnl_before_image")=138
	Set gtmtypes("sgmnt_data",139,"name")="sgmnt_data.jnl_state"
	Set gtmtypes("sgmnt_data",139,"off")=912
	Set gtmtypes("sgmnt_data",139,"len")=4
	Set gtmtypes("sgmnt_data",139,"type")="int"
	Set gtmtypfldindx("sgmnt_data","jnl_state")=139
	Set gtmtypes("sgmnt_data",140,"name")="sgmnt_data.jnl_file_len"
	Set gtmtypes("sgmnt_data",140,"off")=916
	Set gtmtypes("sgmnt_data",140,"len")=4
	Set gtmtypes("sgmnt_data",140,"type")="unsigned-int"
	Set gtmtypfldindx("sgmnt_data","jnl_file_len")=140
	Set gtmtypes("sgmnt_data",141,"name")="sgmnt_data.autoswitchlimit"
	Set gtmtypes("sgmnt_data",141,"off")=920
	Set gtmtypes("sgmnt_data",141,"len")=4
	Set gtmtypes("sgmnt_data",141,"type")="unsigned-int"
	Set gtmtypfldindx("sgmnt_data","autoswitchlimit")=141
	Set gtmtypes("sgmnt_data",142,"name")="sgmnt_data.epoch_interval"
	Set gtmtypes("sgmnt_data",142,"off")=924
	Set gtmtypes("sgmnt_data",142,"len")=4
	Set gtmtypes("sgmnt_data",142,"type")="int"
	Set gtmtypfldindx("sgmnt_data","epoch_interval")=142
	Set gtmtypes("sgmnt_data",143,"name")="sgmnt_data.alignsize"
	Set gtmtypes("sgmnt_data",143,"off")=928
	Set gtmtypes("sgmnt_data",143,"len")=4
	Set gtmtypes("sgmnt_data",143,"type")="unsigned-int"
	Set gtmtypfldindx("sgmnt_data","alignsize")=143
	Set gtmtypes("sgmnt_data",144,"name")="sgmnt_data.jnl_sync_io"
	Set gtmtypes("sgmnt_data",144,"off")=932
	Set gtmtypes("sgmnt_data",144,"len")=4
	Set gtmtypes("sgmnt_data",144,"type")="int"
	Set gtmtypfldindx("sgmnt_data","jnl_sync_io")=144
	Set gtmtypes("sgmnt_data",145,"name")="sgmnt_data.yield_lmt"
	Set gtmtypes("sgmnt_data",145,"off")=936
	Set gtmtypes("sgmnt_data",145,"len")=4
	Set gtmtypes("sgmnt_data",145,"type")="int"
	Set gtmtypfldindx("sgmnt_data","yield_lmt")=145
	Set gtmtypes("sgmnt_data",146,"name")="sgmnt_data.turn_around_point"
	Set gtmtypes("sgmnt_data",146,"off")=940
	Set gtmtypes("sgmnt_data",146,"len")=4
	Set gtmtypes("sgmnt_data",146,"type")="boolean_t"
	Set gtmtypfldindx("sgmnt_data","turn_around_point")=146
	Set gtmtypes("sgmnt_data",147,"name")="sgmnt_data.jnl_eovtn"
	Set gtmtypes("sgmnt_data",147,"off")=944
	Set gtmtypes("sgmnt_data",147,"len")=8
	Set gtmtypes("sgmnt_data",147,"type")="uint64_t"
	Set gtmtypfldindx("sgmnt_data","jnl_eovtn")=147
	Set gtmtypes("sgmnt_data",148,"name")="sgmnt_data.filler_jnl"
	Set gtmtypes("sgmnt_data",148,"off")=952
	Set gtmtypes("sgmnt_data",148,"len")=8
	Set gtmtypes("sgmnt_data",148,"type")="char"
	Set gtmtypfldindx("sgmnt_data","filler_jnl")=148
	Set gtmtypes("sgmnt_data",149,"name")="sgmnt_data.intrpt_recov_resync_seqno"
	Set gtmtypes("sgmnt_data",149,"off")=960
	Set gtmtypes("sgmnt_data",149,"len")=8
	Set gtmtypes("sgmnt_data",149,"type")="uint64_t"
	Set gtmtypfldindx("sgmnt_data","intrpt_recov_resync_seqno")=149
	Set gtmtypes("sgmnt_data",150,"name")="sgmnt_data.intrpt_recov_tp_resolve_time"
	Set gtmtypes("sgmnt_data",150,"off")=968
	Set gtmtypes("sgmnt_data",150,"len")=4
	Set gtmtypes("sgmnt_data",150,"type")="unsigned-int"
	Set gtmtypfldindx("sgmnt_data","intrpt_recov_tp_resolve_time")=150
	Set gtmtypes("sgmnt_data",151,"name")="sgmnt_data.recov_interrupted"
	Set gtmtypes("sgmnt_data",151,"off")=972
	Set gtmtypes("sgmnt_data",151,"len")=4
	Set gtmtypes("sgmnt_data",151,"type")="boolean_t"
	Set gtmtypfldindx("sgmnt_data","recov_interrupted")=151
	Set gtmtypes("sgmnt_data",152,"name")="sgmnt_data.intrpt_recov_jnl_state"
	Set gtmtypes("sgmnt_data",152,"off")=976
	Set gtmtypes("sgmnt_data",152,"len")=4
	Set gtmtypes("sgmnt_data",152,"type")="int"
	Set gtmtypfldindx("sgmnt_data","intrpt_recov_jnl_state")=152
	Set gtmtypes("sgmnt_data",153,"name")="sgmnt_data.intrpt_recov_repl_state"
	Set gtmtypes("sgmnt_data",153,"off")=980
	Set gtmtypes("sgmnt_data",153,"len")=4
	Set gtmtypes("sgmnt_data",153,"type")="int"
	Set gtmtypfldindx("sgmnt_data","intrpt_recov_repl_state")=153
	Set gtmtypes("sgmnt_data",154,"name")="sgmnt_data.before_trunc_total_blks"
	Set gtmtypes("sgmnt_data",154,"off")=984
	Set gtmtypes("sgmnt_data",154,"len")=4
	Set gtmtypes("sgmnt_data",154,"type")="unsigned-int"
	Set gtmtypfldindx("sgmnt_data","before_trunc_total_blks")=154
	Set gtmtypes("sgmnt_data",155,"name")="sgmnt_data.after_trunc_total_blks"
	Set gtmtypes("sgmnt_data",155,"off")=988
	Set gtmtypes("sgmnt_data",155,"len")=4
	Set gtmtypes("sgmnt_data",155,"type")="unsigned-int"
	Set gtmtypfldindx("sgmnt_data","after_trunc_total_blks")=155
	Set gtmtypes("sgmnt_data",156,"name")="sgmnt_data.before_trunc_free_blocks"
	Set gtmtypes("sgmnt_data",156,"off")=992
	Set gtmtypes("sgmnt_data",156,"len")=4
	Set gtmtypes("sgmnt_data",156,"type")="unsigned-int"
	Set gtmtypfldindx("sgmnt_data","before_trunc_free_blocks")=156
	Set gtmtypes("sgmnt_data",157,"name")="sgmnt_data.filler_trunc"
	Set gtmtypes("sgmnt_data",157,"off")=996
	Set gtmtypes("sgmnt_data",157,"len")=4
	Set gtmtypes("sgmnt_data",157,"type")="unsigned-int"
	Set gtmtypfldindx("sgmnt_data","filler_trunc")=157
	Set gtmtypes("sgmnt_data",158,"name")="sgmnt_data.filler_1k"
	Set gtmtypes("sgmnt_data",158,"off")=1000
	Set gtmtypes("sgmnt_data",158,"len")=24
	Set gtmtypes("sgmnt_data",158,"type")="char"
	Set gtmtypfldindx("sgmnt_data","filler_1k")=158
	Set gtmtypes("sgmnt_data",159,"name")="sgmnt_data.jnl_file_name"
	Set gtmtypes("sgmnt_data",159,"off")=1024
	Set gtmtypes("sgmnt_data",159,"len")=256
	Set gtmtypes("sgmnt_data",159,"type")="unsigned-char"
	Set gtmtypfldindx("sgmnt_data","jnl_file_name")=159
	Set gtmtypes("sgmnt_data",160,"name")="sgmnt_data.reorg_restart_key"
	Set gtmtypes("sgmnt_data",160,"off")=1280
	Set gtmtypes("sgmnt_data",160,"len")=256
	Set gtmtypes("sgmnt_data",160,"type")="unsigned-char"
	Set gtmtypfldindx("sgmnt_data","reorg_restart_key")=160
	Set gtmtypes("sgmnt_data",161,"name")="sgmnt_data.machine_name"
	Set gtmtypes("sgmnt_data",161,"off")=1536
	Set gtmtypes("sgmnt_data",161,"len")=256
	Set gtmtypes("sgmnt_data",161,"type")="char"
	Set gtmtypfldindx("sgmnt_data","machine_name")=161
	Set gtmtypes("sgmnt_data",162,"name")="sgmnt_data.encryption_hash"
	Set gtmtypes("sgmnt_data",162,"off")=1792
	Set gtmtypes("sgmnt_data",162,"len")=80
	Set gtmtypes("sgmnt_data",162,"type")="char"
	Set gtmtypfldindx("sgmnt_data","encryption_hash")=162
	Set gtmtypes("sgmnt_data",163,"name")="sgmnt_data.encryption_hash2"
	Set gtmtypes("sgmnt_data",163,"off")=1872
	Set gtmtypes("sgmnt_data",163,"len")=80
	Set gtmtypes("sgmnt_data",163,"type")="char"
	Set gtmtypfldindx("sgmnt_data","encryption_hash2")=163
	Set gtmtypes("sgmnt_data",164,"name")="sgmnt_data.non_null_iv"
	Set gtmtypes("sgmnt_data",164,"off")=1952
	Set gtmtypes("sgmnt_data",164,"len")=4
	Set gtmtypes("sgmnt_data",164,"type")="boolean_t"
	Set gtmtypfldindx("sgmnt_data","non_null_iv")=164
	Set gtmtypes("sgmnt_data",165,"name")="sgmnt_data.encryption_hash_cutoff"
	Set gtmtypes("sgmnt_data",165,"off")=1956
	Set gtmtypes("sgmnt_data",165,"len")=4
	Set gtmtypes("sgmnt_data",165,"type")="int"
	Set gtmtypfldindx("sgmnt_data","encryption_hash_cutoff")=165
	Set gtmtypes("sgmnt_data",166,"name")="sgmnt_data.encryption_hash2_start_tn"
	Set gtmtypes("sgmnt_data",166,"off")=1960
	Set gtmtypes("sgmnt_data",166,"len")=8
	Set gtmtypes("sgmnt_data",166,"type")="uint64_t"
	Set gtmtypfldindx("sgmnt_data","encryption_hash2_start_tn")=166
	Set gtmtypes("sgmnt_data",167,"name")="sgmnt_data.filler_encrypt"
	Set gtmtypes("sgmnt_data",167,"off")=1968
	Set gtmtypes("sgmnt_data",167,"len")=80
	Set gtmtypes("sgmnt_data",167,"type")="char"
	Set gtmtypfldindx("sgmnt_data","filler_encrypt")=167
	Set gtmtypes("sgmnt_data",168,"name")="sgmnt_data.total_buffer_flush_tn"
	Set gtmtypes("sgmnt_data",168,"off")=2048
	Set gtmtypes("sgmnt_data",168,"len")=8
	Set gtmtypes("sgmnt_data",168,"type")="uint64_t"
	Set gtmtypfldindx("sgmnt_data","total_buffer_flush_tn")=168
	Set gtmtypes("sgmnt_data",169,"name")="sgmnt_data.bufct_buffer_flush_tn"
	Set gtmtypes("sgmnt_data",169,"off")=2056
	Set gtmtypes("sgmnt_data",169,"len")=8
	Set gtmtypes("sgmnt_data",169,"type")="uint64_t"
	Set gtmtypfldindx("sgmnt_data","bufct_buffer_flush_tn")=169
	Set gtmtypes("sgmnt_data",170,"name")="sgmnt_data.bufct_buffer_flush_loop_tn"
	Set gtmtypes("sgmnt_data",170,"off")=2064
	Set gtmtypes("sgmnt_data",170,"len")=8
	Set gtmtypes("sgmnt_data",170,"type")="uint64_t"
	Set gtmtypfldindx("sgmnt_data","bufct_buffer_flush_loop_tn")=170
	Set gtmtypes("sgmnt_data",171,"name")="sgmnt_data.wrt_calls_tn"
	Set gtmtypes("sgmnt_data",171,"off")=2072
	Set gtmtypes("sgmnt_data",171,"len")=8
	Set gtmtypes("sgmnt_data",171,"type")="uint64_t"
	Set gtmtypfldindx("sgmnt_data","wrt_calls_tn")=171
	Set gtmtypes("sgmnt_data",172,"name")="sgmnt_data.wrt_count_tn"
	Set gtmtypes("sgmnt_data",172,"off")=2080
	Set gtmtypes("sgmnt_data",172,"len")=8
	Set gtmtypes("sgmnt_data",172,"type")="uint64_t"
	Set gtmtypfldindx("sgmnt_data","wrt_count_tn")=172
	Set gtmtypes("sgmnt_data",173,"name")="sgmnt_data.wrt_blocked_tn"
	Set gtmtypes("sgmnt_data",173,"off")=2088
	Set gtmtypes("sgmnt_data",173,"len")=8
	Set gtmtypes("sgmnt_data",173,"type")="uint64_t"
	Set gtmtypfldindx("sgmnt_data","wrt_blocked_tn")=173
	Set gtmtypes("sgmnt_data",174,"name")="sgmnt_data.wrt_busy_tn"
	Set gtmtypes("sgmnt_data",174,"off")=2096
	Set gtmtypes("sgmnt_data",174,"len")=8
	Set gtmtypes("sgmnt_data",174,"type")="uint64_t"
	Set gtmtypfldindx("sgmnt_data","wrt_busy_tn")=174
	Set gtmtypes("sgmnt_data",175,"name")="sgmnt_data.wrt_noblks_wrtn_tn"
	Set gtmtypes("sgmnt_data",175,"off")=2104
	Set gtmtypes("sgmnt_data",175,"len")=8
	Set gtmtypes("sgmnt_data",175,"type")="uint64_t"
	Set gtmtypfldindx("sgmnt_data","wrt_noblks_wrtn_tn")=175
	Set gtmtypes("sgmnt_data",176,"name")="sgmnt_data.wcb_gds_rundown2_tn"
	Set gtmtypes("sgmnt_data",176,"off")=2112
	Set gtmtypes("sgmnt_data",176,"len")=8
	Set gtmtypes("sgmnt_data",176,"type")="uint64_t"
	Set gtmtypfldindx("sgmnt_data","wcb_gds_rundown2_tn")=176
	Set gtmtypes("sgmnt_data",177,"name")="sgmnt_data.reserved_bgtrcrec2_tn"
	Set gtmtypes("sgmnt_data",177,"off")=2120
	Set gtmtypes("sgmnt_data",177,"len")=8
	Set gtmtypes("sgmnt_data",177,"type")="uint64_t"
	Set gtmtypfldindx("sgmnt_data","reserved_bgtrcrec2_tn")=177
	Set gtmtypes("sgmnt_data",178,"name")="sgmnt_data.reserved_bgtrcrec3_tn"
	Set gtmtypes("sgmnt_data",178,"off")=2128
	Set gtmtypes("sgmnt_data",178,"len")=8
	Set gtmtypes("sgmnt_data",178,"type")="uint64_t"
	Set gtmtypfldindx("sgmnt_data","reserved_bgtrcrec3_tn")=178
	Set gtmtypes("sgmnt_data",179,"name")="sgmnt_data.lost_block_recovery_tn"
	Set gtmtypes("sgmnt_data",179,"off")=2136
	Set gtmtypes("sgmnt_data",179,"len")=8
	Set gtmtypes("sgmnt_data",179,"type")="uint64_t"
	Set gtmtypfldindx("sgmnt_data","lost_block_recovery_tn")=179
	Set gtmtypes("sgmnt_data",180,"name")="sgmnt_data.wc_blocked_onln_rlbk_tn"
	Set gtmtypes("sgmnt_data",180,"off")=2144
	Set gtmtypes("sgmnt_data",180,"len")=8
	Set gtmtypes("sgmnt_data",180,"type")="uint64_t"
	Set gtmtypfldindx("sgmnt_data","wc_blocked_onln_rlbk_tn")=180
	Set gtmtypes("sgmnt_data",181,"name")="sgmnt_data.stale_tn"
	Set gtmtypes("sgmnt_data",181,"off")=2152
	Set gtmtypes("sgmnt_data",181,"len")=8
	Set gtmtypes("sgmnt_data",181,"type")="uint64_t"
	Set gtmtypfldindx("sgmnt_data","stale_tn")=181
	Set gtmtypes("sgmnt_data",182,"name")="sgmnt_data.starved_tn"
	Set gtmtypes("sgmnt_data",182,"off")=2160
	Set gtmtypes("sgmnt_data",182,"len")=8
	Set gtmtypes("sgmnt_data",182,"type")="uint64_t"
	Set gtmtypfldindx("sgmnt_data","starved_tn")=182
	Set gtmtypes("sgmnt_data",183,"name")="sgmnt_data.stale_timer_started_tn"
	Set gtmtypes("sgmnt_data",183,"off")=2168
	Set gtmtypes("sgmnt_data",183,"len")=8
	Set gtmtypes("sgmnt_data",183,"type")="uint64_t"
	Set gtmtypfldindx("sgmnt_data","stale_timer_started_tn")=183
	Set gtmtypes("sgmnt_data",184,"name")="sgmnt_data.stale_timer_pop_tn"
	Set gtmtypes("sgmnt_data",184,"off")=2176
	Set gtmtypes("sgmnt_data",184,"len")=8
	Set gtmtypes("sgmnt_data",184,"type")="uint64_t"
	Set gtmtypfldindx("sgmnt_data","stale_timer_pop_tn")=184
	Set gtmtypes("sgmnt_data",185,"name")="sgmnt_data.stale_process_defer_tn"
	Set gtmtypes("sgmnt_data",185,"off")=2184
	Set gtmtypes("sgmnt_data",185,"len")=8
	Set gtmtypes("sgmnt_data",185,"type")="uint64_t"
	Set gtmtypfldindx("sgmnt_data","stale_process_defer_tn")=185
	Set gtmtypes("sgmnt_data",186,"name")="sgmnt_data.stale_defer_processed_tn"
	Set gtmtypes("sgmnt_data",186,"off")=2192
	Set gtmtypes("sgmnt_data",186,"len")=8
	Set gtmtypes("sgmnt_data",186,"type")="uint64_t"
	Set gtmtypfldindx("sgmnt_data","stale_defer_processed_tn")=186
	Set gtmtypes("sgmnt_data",187,"name")="sgmnt_data.spcfc_buffer_flush_tn"
	Set gtmtypes("sgmnt_data",187,"off")=2200
	Set gtmtypes("sgmnt_data",187,"len")=8
	Set gtmtypes("sgmnt_data",187,"type")="uint64_t"
	Set gtmtypfldindx("sgmnt_data","spcfc_buffer_flush_tn")=187
	Set gtmtypes("sgmnt_data",188,"name")="sgmnt_data.spcfc_buffer_flush_loop_tn"
	Set gtmtypes("sgmnt_data",188,"off")=2208
	Set gtmtypes("sgmnt_data",188,"len")=8
	Set gtmtypes("sgmnt_data",188,"type")="uint64_t"
	Set gtmtypfldindx("sgmnt_data","spcfc_buffer_flush_loop_tn")=188
	Set gtmtypes("sgmnt_data",189,"name")="sgmnt_data.spcfc_buffer_flush_retries_tn"
	Set gtmtypes("sgmnt_data",189,"off")=2216
	Set gtmtypes("sgmnt_data",189,"len")=8
	Set gtmtypes("sgmnt_data",189,"type")="uint64_t"
	Set gtmtypfldindx("sgmnt_data","spcfc_buffer_flush_retries_tn")=189
	Set gtmtypes("sgmnt_data",190,"name")="sgmnt_data.spcfc_buffer_flushed_during_lockwait_tn"
	Set gtmtypes("sgmnt_data",190,"off")=2224
	Set gtmtypes("sgmnt_data",190,"len")=8
	Set gtmtypes("sgmnt_data",190,"type")="uint64_t"
	Set gtmtypfldindx("sgmnt_data","spcfc_buffer_flushed_during_lockwait_tn")=190
	Set gtmtypes("sgmnt_data",191,"name")="sgmnt_data.tp_crit_retries_tn"
	Set gtmtypes("sgmnt_data",191,"off")=2232
	Set gtmtypes("sgmnt_data",191,"len")=8
	Set gtmtypes("sgmnt_data",191,"type")="uint64_t"
	Set gtmtypfldindx("sgmnt_data","tp_crit_retries_tn")=191
	Set gtmtypes("sgmnt_data",192,"name")="sgmnt_data.db_csh_getn_flush_dirty_tn"
	Set gtmtypes("sgmnt_data",192,"off")=2240
	Set gtmtypes("sgmnt_data",192,"len")=8
	Set gtmtypes("sgmnt_data",192,"type")="uint64_t"
	Set gtmtypfldindx("sgmnt_data","db_csh_getn_flush_dirty_tn")=192
	Set gtmtypes("sgmnt_data",193,"name")="sgmnt_data.db_csh_getn_rip_wait_tn"
	Set gtmtypes("sgmnt_data",193,"off")=2248
	Set gtmtypes("sgmnt_data",193,"len")=8
	Set gtmtypes("sgmnt_data",193,"type")="uint64_t"
	Set gtmtypfldindx("sgmnt_data","db_csh_getn_rip_wait_tn")=193
	Set gtmtypes("sgmnt_data",194,"name")="sgmnt_data.db_csh_getn_buf_owner_stuck_tn"
	Set gtmtypes("sgmnt_data",194,"off")=2256
	Set gtmtypes("sgmnt_data",194,"len")=8
	Set gtmtypes("sgmnt_data",194,"type")="uint64_t"
	Set gtmtypfldindx("sgmnt_data","db_csh_getn_buf_owner_stuck_tn")=194
	Set gtmtypes("sgmnt_data",195,"name")="sgmnt_data.db_csh_getn_out_of_design_tn"
	Set gtmtypes("sgmnt_data",195,"off")=2264
	Set gtmtypes("sgmnt_data",195,"len")=8
	Set gtmtypes("sgmnt_data",195,"type")="uint64_t"
	Set gtmtypfldindx("sgmnt_data","db_csh_getn_out_of_design_tn")=195
	Set gtmtypes("sgmnt_data",196,"name")="sgmnt_data.t_qread_buf_owner_stuck_tn"
	Set gtmtypes("sgmnt_data",196,"off")=2272
	Set gtmtypes("sgmnt_data",196,"len")=8
	Set gtmtypes("sgmnt_data",196,"type")="uint64_t"
	Set gtmtypfldindx("sgmnt_data","t_qread_buf_owner_stuck_tn")=196
	Set gtmtypes("sgmnt_data",197,"name")="sgmnt_data.t_qread_out_of_design_tn"
	Set gtmtypes("sgmnt_data",197,"off")=2280
	Set gtmtypes("sgmnt_data",197,"len")=8
	Set gtmtypes("sgmnt_data",197,"type")="uint64_t"
	Set gtmtypfldindx("sgmnt_data","t_qread_out_of_design_tn")=197
	Set gtmtypes("sgmnt_data",198,"name")="sgmnt_data.unused_filler_0031_tn"
	Set gtmtypes("sgmnt_data",198,"off")=2288
	Set gtmtypes("sgmnt_data",198,"len")=8
	Set gtmtypes("sgmnt_data",198,"type")="uint64_t"
	Set gtmtypfldindx("sgmnt_data","unused_filler_0031_tn")=198
	Set gtmtypes("sgmnt_data",199,"name")="sgmnt_data.mlock_wakeups_tn"
	Set gtmtypes("sgmnt_data",199,"off")=2296
	Set gtmtypes("sgmnt_data",199,"len")=8
	Set gtmtypes("sgmnt_data",199,"type")="uint64_t"
	Set gtmtypfldindx("sgmnt_data","mlock_wakeups_tn")=199
	Set gtmtypes("sgmnt_data",200,"name")="sgmnt_data.wc_blocked_wcs_recover_invoked_tn"
	Set gtmtypes("sgmnt_data",200,"off")=2304
	Set gtmtypes("sgmnt_data",200,"len")=8
	Set gtmtypes("sgmnt_data",200,"type")="uint64_t"
	Set gtmtypfldindx("sgmnt_data","wc_blocked_wcs_recover_invoked_tn")=200
	Set gtmtypes("sgmnt_data",201,"name")="sgmnt_data.wc_blocked_wcs_verify_passed_tn"
	Set gtmtypes("sgmnt_data",201,"off")=2312
	Set gtmtypes("sgmnt_data",201,"len")=8
	Set gtmtypes("sgmnt_data",201,"type")="uint64_t"
	Set gtmtypfldindx("sgmnt_data","wc_blocked_wcs_verify_passed_tn")=201
	Set gtmtypes("sgmnt_data",202,"name")="sgmnt_data.wc_blocked_t_qread_db_csh_getn_invalid_blk_tn"
	Set gtmtypes("sgmnt_data",202,"off")=2320
	Set gtmtypes("sgmnt_data",202,"len")=8
	Set gtmtypes("sgmnt_data",202,"type")="uint64_t"
	Set gtmtypfldindx("sgmnt_data","wc_blocked_t_qread_db_csh_getn_invalid_blk_tn")=202
	Set gtmtypes("sgmnt_data",203,"name")="sgmnt_data.wc_blocked_t_qread_db_csh_get_invalid_blk_tn"
	Set gtmtypes("sgmnt_data",203,"off")=2328
	Set gtmtypes("sgmnt_data",203,"len")=8
	Set gtmtypes("sgmnt_data",203,"type")="uint64_t"
	Set gtmtypfldindx("sgmnt_data","wc_blocked_t_qread_db_csh_get_invalid_blk_tn")=203
	Set gtmtypes("sgmnt_data",204,"name")="sgmnt_data.wc_blocked_db_csh_getn_loopexceed_tn"
	Set gtmtypes("sgmnt_data",204,"off")=2336
	Set gtmtypes("sgmnt_data",204,"len")=8
	Set gtmtypes("sgmnt_data",204,"type")="uint64_t"
	Set gtmtypfldindx("sgmnt_data","wc_blocked_db_csh_getn_loopexceed_tn")=204
	Set gtmtypes("sgmnt_data",205,"name")="sgmnt_data.wc_blocked_db_csh_getn_wcsstarvewrt_tn"
	Set gtmtypes("sgmnt_data",205,"off")=2344
	Set gtmtypes("sgmnt_data",205,"len")=8
	Set gtmtypes("sgmnt_data",205,"type")="uint64_t"
	Set gtmtypfldindx("sgmnt_data","wc_blocked_db_csh_getn_wcsstarvewrt_tn")=205
	Set gtmtypes("sgmnt_data",206,"name")="sgmnt_data.wc_blocked_db_csh_get_tn"
	Set gtmtypes("sgmnt_data",206,"off")=2352
	Set gtmtypes("sgmnt_data",206,"len")=8
	Set gtmtypes("sgmnt_data",206,"type")="uint64_t"
	Set gtmtypfldindx("sgmnt_data","wc_blocked_db_csh_get_tn")=206
	Set gtmtypes("sgmnt_data",207,"name")="sgmnt_data.wc_blocked_tp_tend_wcsgetspace_tn"
	Set gtmtypes("sgmnt_data",207,"off")=2360
	Set gtmtypes("sgmnt_data",207,"len")=8
	Set gtmtypes("sgmnt_data",207,"type")="uint64_t"
	Set gtmtypfldindx("sgmnt_data","wc_blocked_tp_tend_wcsgetspace_tn")=207
	Set gtmtypes("sgmnt_data",208,"name")="sgmnt_data.wc_blocked_tp_tend_t1_tn"
	Set gtmtypes("sgmnt_data",208,"off")=2368
	Set gtmtypes("sgmnt_data",208,"len")=8
	Set gtmtypes("sgmnt_data",208,"type")="uint64_t"
	Set gtmtypfldindx("sgmnt_data","wc_blocked_tp_tend_t1_tn")=208
	Set gtmtypes("sgmnt_data",209,"name")="sgmnt_data.wc_blocked_tp_tend_bitmap_tn"
	Set gtmtypes("sgmnt_data",209,"off")=2376
	Set gtmtypes("sgmnt_data",209,"len")=8
	Set gtmtypes("sgmnt_data",209,"type")="uint64_t"
	Set gtmtypfldindx("sgmnt_data","wc_blocked_tp_tend_bitmap_tn")=209
	Set gtmtypes("sgmnt_data",210,"name")="sgmnt_data.wc_blocked_tp_tend_jnl_cwset_tn"
	Set gtmtypes("sgmnt_data",210,"off")=2384
	Set gtmtypes("sgmnt_data",210,"len")=8
	Set gtmtypes("sgmnt_data",210,"type")="uint64_t"
	Set gtmtypfldindx("sgmnt_data","wc_blocked_tp_tend_jnl_cwset_tn")=210
	Set gtmtypes("sgmnt_data",211,"name")="sgmnt_data.wc_blocked_tp_tend_jnl_wcsflu_tn"
	Set gtmtypes("sgmnt_data",211,"off")=2392
	Set gtmtypes("sgmnt_data",211,"len")=8
	Set gtmtypes("sgmnt_data",211,"type")="uint64_t"
	Set gtmtypfldindx("sgmnt_data","wc_blocked_tp_tend_jnl_wcsflu_tn")=211
	Set gtmtypes("sgmnt_data",212,"name")="sgmnt_data.wc_blocked_t_end_hist_tn"
	Set gtmtypes("sgmnt_data",212,"off")=2400
	Set gtmtypes("sgmnt_data",212,"len")=8
	Set gtmtypes("sgmnt_data",212,"type")="uint64_t"
	Set gtmtypfldindx("sgmnt_data","wc_blocked_t_end_hist_tn")=212
	Set gtmtypes("sgmnt_data",213,"name")="sgmnt_data.wc_blocked_t_end_hist1_nullbt_tn"
	Set gtmtypes("sgmnt_data",213,"off")=2408
	Set gtmtypes("sgmnt_data",213,"len")=8
	Set gtmtypes("sgmnt_data",213,"type")="uint64_t"
	Set gtmtypfldindx("sgmnt_data","wc_blocked_t_end_hist1_nullbt_tn")=213
	Set gtmtypes("sgmnt_data",214,"name")="sgmnt_data.wc_blocked_t_end_hist1_nonnullbt_tn"
	Set gtmtypes("sgmnt_data",214,"off")=2416
	Set gtmtypes("sgmnt_data",214,"len")=8
	Set gtmtypes("sgmnt_data",214,"type")="uint64_t"
	Set gtmtypfldindx("sgmnt_data","wc_blocked_t_end_hist1_nonnullbt_tn")=214
	Set gtmtypes("sgmnt_data",215,"name")="sgmnt_data.wc_blocked_t_end_bitmap_nullbt_tn"
	Set gtmtypes("sgmnt_data",215,"off")=2424
	Set gtmtypes("sgmnt_data",215,"len")=8
	Set gtmtypes("sgmnt_data",215,"type")="uint64_t"
	Set gtmtypfldindx("sgmnt_data","wc_blocked_t_end_bitmap_nullbt_tn")=215
	Set gtmtypes("sgmnt_data",216,"name")="sgmnt_data.wc_blocked_t_end_bitmap_nonnullbt_tn"
	Set gtmtypes("sgmnt_data",216,"off")=2432
	Set gtmtypes("sgmnt_data",216,"len")=8
	Set gtmtypes("sgmnt_data",216,"type")="uint64_t"
	Set gtmtypfldindx("sgmnt_data","wc_blocked_t_end_bitmap_nonnullbt_tn")=216
	Set gtmtypes("sgmnt_data",217,"name")="sgmnt_data.wc_blocked_t_end_jnl_cwset_tn"
	Set gtmtypes("sgmnt_data",217,"off")=2440
	Set gtmtypes("sgmnt_data",217,"len")=8
	Set gtmtypes("sgmnt_data",217,"type")="uint64_t"
	Set gtmtypfldindx("sgmnt_data","wc_blocked_t_end_jnl_cwset_tn")=217
	Set gtmtypes("sgmnt_data",218,"name")="sgmnt_data.wc_blocked_t_end_jnl_wcsflu_tn"
	Set gtmtypes("sgmnt_data",218,"off")=2448
	Set gtmtypes("sgmnt_data",218,"len")=8
	Set gtmtypes("sgmnt_data",218,"type")="uint64_t"
	Set gtmtypfldindx("sgmnt_data","wc_blocked_t_end_jnl_wcsflu_tn")=218
	Set gtmtypes("sgmnt_data",219,"name")="sgmnt_data.unused_filler_0052_tn"
	Set gtmtypes("sgmnt_data",219,"off")=2456
	Set gtmtypes("sgmnt_data",219,"len")=8
	Set gtmtypes("sgmnt_data",219,"type")="uint64_t"
	Set gtmtypfldindx("sgmnt_data","unused_filler_0052_tn")=219
	Set gtmtypes("sgmnt_data",220,"name")="sgmnt_data.wc_blocked_tpckh_hist1_nullbt_tn"
	Set gtmtypes("sgmnt_data",220,"off")=2464
	Set gtmtypes("sgmnt_data",220,"len")=8
	Set gtmtypes("sgmnt_data",220,"type")="uint64_t"
	Set gtmtypfldindx("sgmnt_data","wc_blocked_tpckh_hist1_nullbt_tn")=220
	Set gtmtypes("sgmnt_data",221,"name")="sgmnt_data.wc_blocked_tpckh_hist1_nonnullbt_tn"
	Set gtmtypes("sgmnt_data",221,"off")=2472
	Set gtmtypes("sgmnt_data",221,"len")=8
	Set gtmtypes("sgmnt_data",221,"type")="uint64_t"
	Set gtmtypfldindx("sgmnt_data","wc_blocked_tpckh_hist1_nonnullbt_tn")=221
	Set gtmtypes("sgmnt_data",222,"name")="sgmnt_data.jnl_blocked_writer_lost_tn"
	Set gtmtypes("sgmnt_data",222,"off")=2480
	Set gtmtypes("sgmnt_data",222,"len")=8
	Set gtmtypes("sgmnt_data",222,"type")="uint64_t"
	Set gtmtypfldindx("sgmnt_data","jnl_blocked_writer_lost_tn")=222
	Set gtmtypes("sgmnt_data",223,"name")="sgmnt_data.jnl_blocked_writer_stuck_tn"
	Set gtmtypes("sgmnt_data",223,"off")=2488
	Set gtmtypes("sgmnt_data",223,"len")=8
	Set gtmtypes("sgmnt_data",223,"type")="uint64_t"
	Set gtmtypfldindx("sgmnt_data","jnl_blocked_writer_stuck_tn")=223
	Set gtmtypes("sgmnt_data",224,"name")="sgmnt_data.jnl_blocked_writer_blocked_tn"
	Set gtmtypes("sgmnt_data",224,"off")=2496
	Set gtmtypes("sgmnt_data",224,"len")=8
	Set gtmtypes("sgmnt_data",224,"type")="uint64_t"
	Set gtmtypfldindx("sgmnt_data","jnl_blocked_writer_blocked_tn")=224
	Set gtmtypes("sgmnt_data",225,"name")="sgmnt_data.n_jnl_fsyncs_tn"
	Set gtmtypes("sgmnt_data",225,"off")=2504
	Set gtmtypes("sgmnt_data",225,"len")=8
	Set gtmtypes("sgmnt_data",225,"type")="uint64_t"
	Set gtmtypfldindx("sgmnt_data","n_jnl_fsyncs_tn")=225
	Set gtmtypes("sgmnt_data",226,"name")="sgmnt_data.n_jnl_fsync_tries_tn"
	Set gtmtypes("sgmnt_data",226,"off")=2512
	Set gtmtypes("sgmnt_data",226,"len")=8
	Set gtmtypes("sgmnt_data",226,"type")="uint64_t"
	Set gtmtypfldindx("sgmnt_data","n_jnl_fsync_tries_tn")=226
	Set gtmtypes("sgmnt_data",227,"name")="sgmnt_data.n_jnl_fsync_recovers_tn"
	Set gtmtypes("sgmnt_data",227,"off")=2520
	Set gtmtypes("sgmnt_data",227,"len")=8
	Set gtmtypes("sgmnt_data",227,"type")="uint64_t"
	Set gtmtypfldindx("sgmnt_data","n_jnl_fsync_recovers_tn")=227
	Set gtmtypes("sgmnt_data",228,"name")="sgmnt_data.n_db_fsyncs_tn"
	Set gtmtypes("sgmnt_data",228,"off")=2528
	Set gtmtypes("sgmnt_data",228,"len")=8
	Set gtmtypes("sgmnt_data",228,"type")="uint64_t"
	Set gtmtypfldindx("sgmnt_data","n_db_fsyncs_tn")=228
	Set gtmtypes("sgmnt_data",229,"name")="sgmnt_data.n_db_fsyncs_in_crit_tn"
	Set gtmtypes("sgmnt_data",229,"off")=2536
	Set gtmtypes("sgmnt_data",229,"len")=8
	Set gtmtypes("sgmnt_data",229,"type")="uint64_t"
	Set gtmtypfldindx("sgmnt_data","n_db_fsyncs_in_crit_tn")=229
	Set gtmtypes("sgmnt_data",230,"name")="sgmnt_data.n_dbsync_timers_tn"
	Set gtmtypes("sgmnt_data",230,"off")=2544
	Set gtmtypes("sgmnt_data",230,"len")=8
	Set gtmtypes("sgmnt_data",230,"type")="uint64_t"
	Set gtmtypfldindx("sgmnt_data","n_dbsync_timers_tn")=230
	Set gtmtypes("sgmnt_data",231,"name")="sgmnt_data.n_dbsync_writes_tn"
	Set gtmtypes("sgmnt_data",231,"off")=2552
	Set gtmtypes("sgmnt_data",231,"len")=8
	Set gtmtypes("sgmnt_data",231,"type")="uint64_t"
	Set gtmtypfldindx("sgmnt_data","n_dbsync_writes_tn")=231
	Set gtmtypes("sgmnt_data",232,"name")="sgmnt_data.mutex_queue_full_tn"
	Set gtmtypes("sgmnt_data",232,"off")=2560
	Set gtmtypes("sgmnt_data",232,"len")=8
	Set gtmtypes("sgmnt_data",232,"type")="uint64_t"
	Set gtmtypfldindx("sgmnt_data","mutex_queue_full_tn")=232
	Set gtmtypes("sgmnt_data",233,"name")="sgmnt_data.wcb_bt_put_tn"
	Set gtmtypes("sgmnt_data",233,"off")=2568
	Set gtmtypes("sgmnt_data",233,"len")=8
	Set gtmtypes("sgmnt_data",233,"type")="uint64_t"
	Set gtmtypfldindx("sgmnt_data","wcb_bt_put_tn")=233
	Set gtmtypes("sgmnt_data",234,"name")="sgmnt_data.wcb_mutex_salvage_tn"
	Set gtmtypes("sgmnt_data",234,"off")=2576
	Set gtmtypes("sgmnt_data",234,"len")=8
	Set gtmtypes("sgmnt_data",234,"type")="uint64_t"
	Set gtmtypfldindx("sgmnt_data","wcb_mutex_salvage_tn")=234
	Set gtmtypes("sgmnt_data",235,"name")="sgmnt_data.wcb_wcs_flu0_tn"
	Set gtmtypes("sgmnt_data",235,"off")=2584
	Set gtmtypes("sgmnt_data",235,"len")=8
	Set gtmtypes("sgmnt_data",235,"type")="uint64_t"
	Set gtmtypfldindx("sgmnt_data","wcb_wcs_flu0_tn")=235
	Set gtmtypes("sgmnt_data",236,"name")="sgmnt_data.wcb_t_end_sysops_nocr_invcr_tn"
	Set gtmtypes("sgmnt_data",236,"off")=2592
	Set gtmtypes("sgmnt_data",236,"len")=8
	Set gtmtypes("sgmnt_data",236,"type")="uint64_t"
	Set gtmtypfldindx("sgmnt_data","wcb_t_end_sysops_nocr_invcr_tn")=236
	Set gtmtypes("sgmnt_data",237,"name")="sgmnt_data.wcb_t_end_sysops_cr_invcr_tn"
	Set gtmtypes("sgmnt_data",237,"off")=2600
	Set gtmtypes("sgmnt_data",237,"len")=8
	Set gtmtypes("sgmnt_data",237,"type")="uint64_t"
	Set gtmtypfldindx("sgmnt_data","wcb_t_end_sysops_cr_invcr_tn")=237
	Set gtmtypes("sgmnt_data",238,"name")="sgmnt_data.wcb_t_end_sysops_rip_wait_tn"
	Set gtmtypes("sgmnt_data",238,"off")=2608
	Set gtmtypes("sgmnt_data",238,"len")=8
	Set gtmtypes("sgmnt_data",238,"type")="uint64_t"
	Set gtmtypfldindx("sgmnt_data","wcb_t_end_sysops_rip_wait_tn")=238
	Set gtmtypes("sgmnt_data",239,"name")="sgmnt_data.wcb_t_end_sysops_dirtyripwait_tn"
	Set gtmtypes("sgmnt_data",239,"off")=2616
	Set gtmtypes("sgmnt_data",239,"len")=8
	Set gtmtypes("sgmnt_data",239,"type")="uint64_t"
	Set gtmtypfldindx("sgmnt_data","wcb_t_end_sysops_dirtyripwait_tn")=239
	Set gtmtypes("sgmnt_data",240,"name")="sgmnt_data.wcb_gds_rundown1_tn"
	Set gtmtypes("sgmnt_data",240,"off")=2624
	Set gtmtypes("sgmnt_data",240,"len")=8
	Set gtmtypes("sgmnt_data",240,"type")="uint64_t"
	Set gtmtypfldindx("sgmnt_data","wcb_gds_rundown1_tn")=240
	Set gtmtypes("sgmnt_data",241,"name")="sgmnt_data.wcb_wcs_flu1_tn"
	Set gtmtypes("sgmnt_data",241,"off")=2632
	Set gtmtypes("sgmnt_data",241,"len")=8
	Set gtmtypes("sgmnt_data",241,"type")="uint64_t"
	Set gtmtypfldindx("sgmnt_data","wcb_wcs_flu1_tn")=241
	Set gtmtypes("sgmnt_data",242,"name")="sgmnt_data.wcb_mu_back_tn"
	Set gtmtypes("sgmnt_data",242,"off")=2640
	Set gtmtypes("sgmnt_data",242,"len")=8
	Set gtmtypes("sgmnt_data",242,"type")="uint64_t"
	Set gtmtypfldindx("sgmnt_data","wcb_mu_back_tn")=242
	Set gtmtypes("sgmnt_data",243,"name")="sgmnt_data.wcb_t_end_sysops_dirty_invcr_tn"
	Set gtmtypes("sgmnt_data",243,"off")=2648
	Set gtmtypes("sgmnt_data",243,"len")=8
	Set gtmtypes("sgmnt_data",243,"type")="uint64_t"
	Set gtmtypfldindx("sgmnt_data","wcb_t_end_sysops_dirty_invcr_tn")=243
	Set gtmtypes("sgmnt_data",244,"name")="sgmnt_data.wcb_t_end_sysops_wtfini_fail_tn"
	Set gtmtypes("sgmnt_data",244,"off")=2656
	Set gtmtypes("sgmnt_data",244,"len")=8
	Set gtmtypes("sgmnt_data",244,"type")="uint64_t"
	Set gtmtypfldindx("sgmnt_data","wcb_t_end_sysops_wtfini_fail_tn")=244
	Set gtmtypes("sgmnt_data",245,"name")="sgmnt_data.wcb_t_end_sysops_twin_stuck_tn"
	Set gtmtypes("sgmnt_data",245,"off")=2664
	Set gtmtypes("sgmnt_data",245,"len")=8
	Set gtmtypes("sgmnt_data",245,"type")="uint64_t"
	Set gtmtypfldindx("sgmnt_data","wcb_t_end_sysops_twin_stuck_tn")=245
	Set gtmtypes("sgmnt_data",246,"name")="sgmnt_data.db_csh_getn_wrt_latch_stuck_tn"
	Set gtmtypes("sgmnt_data",246,"off")=2672
	Set gtmtypes("sgmnt_data",246,"len")=8
	Set gtmtypes("sgmnt_data",246,"type")="uint64_t"
	Set gtmtypfldindx("sgmnt_data","db_csh_getn_wrt_latch_stuck_tn")=246
	Set gtmtypes("sgmnt_data",247,"name")="sgmnt_data.wcb_secshr_db_clnup_now_crit_tn"
	Set gtmtypes("sgmnt_data",247,"off")=2680
	Set gtmtypes("sgmnt_data",247,"len")=8
	Set gtmtypes("sgmnt_data",247,"type")="uint64_t"
	Set gtmtypfldindx("sgmnt_data","wcb_secshr_db_clnup_now_crit_tn")=247
	Set gtmtypes("sgmnt_data",248,"name")="sgmnt_data.wc_blocked_t_qread_bad_bt_index1_tn"
	Set gtmtypes("sgmnt_data",248,"off")=2688
	Set gtmtypes("sgmnt_data",248,"len")=8
	Set gtmtypes("sgmnt_data",248,"type")="uint64_t"
	Set gtmtypfldindx("sgmnt_data","wc_blocked_t_qread_bad_bt_index1_tn")=248
	Set gtmtypes("sgmnt_data",249,"name")="sgmnt_data.wc_blocked_t_qread_bad_bt_index2_tn"
	Set gtmtypes("sgmnt_data",249,"off")=2696
	Set gtmtypes("sgmnt_data",249,"len")=8
	Set gtmtypes("sgmnt_data",249,"type")="uint64_t"
	Set gtmtypfldindx("sgmnt_data","wc_blocked_t_qread_bad_bt_index2_tn")=249
	Set gtmtypes("sgmnt_data",250,"name")="sgmnt_data.wc_blocked_t_end_crbtmismatch1_tn"
	Set gtmtypes("sgmnt_data",250,"off")=2704
	Set gtmtypes("sgmnt_data",250,"len")=8
	Set gtmtypes("sgmnt_data",250,"type")="uint64_t"
	Set gtmtypfldindx("sgmnt_data","wc_blocked_t_end_crbtmismatch1_tn")=250
	Set gtmtypes("sgmnt_data",251,"name")="sgmnt_data.wc_blocked_t_end_crbtmismatch2_tn"
	Set gtmtypes("sgmnt_data",251,"off")=2712
	Set gtmtypes("sgmnt_data",251,"len")=8
	Set gtmtypes("sgmnt_data",251,"type")="uint64_t"
	Set gtmtypfldindx("sgmnt_data","wc_blocked_t_end_crbtmismatch2_tn")=251
	Set gtmtypes("sgmnt_data",252,"name")="sgmnt_data.wc_blocked_t_end_crbtmismatch3_tn"
	Set gtmtypes("sgmnt_data",252,"off")=2720
	Set gtmtypes("sgmnt_data",252,"len")=8
	Set gtmtypes("sgmnt_data",252,"type")="uint64_t"
	Set gtmtypfldindx("sgmnt_data","wc_blocked_t_end_crbtmismatch3_tn")=252
	Set gtmtypes("sgmnt_data",253,"name")="sgmnt_data.wc_blocked_tp_tend_crbtmismatch1_tn"
	Set gtmtypes("sgmnt_data",253,"off")=2728
	Set gtmtypes("sgmnt_data",253,"len")=8
	Set gtmtypes("sgmnt_data",253,"type")="uint64_t"
	Set gtmtypfldindx("sgmnt_data","wc_blocked_tp_tend_crbtmismatch1_tn")=253
	Set gtmtypes("sgmnt_data",254,"name")="sgmnt_data.wc_blocked_tp_tend_crbtmismatch2_tn"
	Set gtmtypes("sgmnt_data",254,"off")=2736
	Set gtmtypes("sgmnt_data",254,"len")=8
	Set gtmtypes("sgmnt_data",254,"type")="uint64_t"
	Set gtmtypfldindx("sgmnt_data","wc_blocked_tp_tend_crbtmismatch2_tn")=254
	Set gtmtypes("sgmnt_data",255,"name")="sgmnt_data.wc_blocked_tp_tend_crbtmismatch3_tn"
	Set gtmtypes("sgmnt_data",255,"off")=2744
	Set gtmtypes("sgmnt_data",255,"len")=8
	Set gtmtypes("sgmnt_data",255,"type")="uint64_t"
	Set gtmtypfldindx("sgmnt_data","wc_blocked_tp_tend_crbtmismatch3_tn")=255
	Set gtmtypes("sgmnt_data",256,"name")="sgmnt_data.wc_blocked_wcs_wtstart_bad_cr_tn"
	Set gtmtypes("sgmnt_data",256,"off")=2752
	Set gtmtypes("sgmnt_data",256,"len")=8
	Set gtmtypes("sgmnt_data",256,"type")="uint64_t"
	Set gtmtypfldindx("sgmnt_data","wc_blocked_wcs_wtstart_bad_cr_tn")=256
	Set gtmtypes("sgmnt_data",257,"name")="sgmnt_data.wc_blocked_wcs_wtfini_bad_cr_tn"
	Set gtmtypes("sgmnt_data",257,"off")=2760
	Set gtmtypes("sgmnt_data",257,"len")=8
	Set gtmtypes("sgmnt_data",257,"type")="uint64_t"
	Set gtmtypfldindx("sgmnt_data","wc_blocked_wcs_wtfini_bad_cr_tn")=257
	Set gtmtypes("sgmnt_data",258,"name")="sgmnt_data.wc_blocked_bt_get_tn"
	Set gtmtypes("sgmnt_data",258,"off")=2768
	Set gtmtypes("sgmnt_data",258,"len")=8
	Set gtmtypes("sgmnt_data",258,"type")="uint64_t"
	Set gtmtypfldindx("sgmnt_data","wc_blocked_bt_get_tn")=258
	Set gtmtypes("sgmnt_data",259,"name")="sgmnt_data.wc_blocked_wcs_cdb_sc_final_retry_tn"
	Set gtmtypes("sgmnt_data",259,"off")=2776
	Set gtmtypes("sgmnt_data",259,"len")=8
	Set gtmtypes("sgmnt_data",259,"type")="uint64_t"
	Set gtmtypfldindx("sgmnt_data","wc_blocked_wcs_cdb_sc_final_retry_tn")=259
	Set gtmtypes("sgmnt_data",260,"name")="sgmnt_data.wcb_bg_update_lckfail1_tn"
	Set gtmtypes("sgmnt_data",260,"off")=2784
	Set gtmtypes("sgmnt_data",260,"len")=8
	Set gtmtypes("sgmnt_data",260,"type")="uint64_t"
	Set gtmtypfldindx("sgmnt_data","wcb_bg_update_lckfail1_tn")=260
	Set gtmtypes("sgmnt_data",261,"name")="sgmnt_data.wcb_bg_update_lckfail2_tn"
	Set gtmtypes("sgmnt_data",261,"off")=2792
	Set gtmtypes("sgmnt_data",261,"len")=8
	Set gtmtypes("sgmnt_data",261,"type")="uint64_t"
	Set gtmtypfldindx("sgmnt_data","wcb_bg_update_lckfail2_tn")=261
	Set gtmtypes("sgmnt_data",262,"name")="sgmnt_data.wcb_wtstart_lckfail1_tn"
	Set gtmtypes("sgmnt_data",262,"off")=2800
	Set gtmtypes("sgmnt_data",262,"len")=8
	Set gtmtypes("sgmnt_data",262,"type")="uint64_t"
	Set gtmtypfldindx("sgmnt_data","wcb_wtstart_lckfail1_tn")=262
	Set gtmtypes("sgmnt_data",263,"name")="sgmnt_data.wcb_wtstart_lckfail2_tn"
	Set gtmtypes("sgmnt_data",263,"off")=2808
	Set gtmtypes("sgmnt_data",263,"len")=8
	Set gtmtypes("sgmnt_data",263,"type")="uint64_t"
	Set gtmtypfldindx("sgmnt_data","wcb_wtstart_lckfail2_tn")=263
	Set gtmtypes("sgmnt_data",264,"name")="sgmnt_data.wcb_wtstart_lckfail3_tn"
	Set gtmtypes("sgmnt_data",264,"off")=2816
	Set gtmtypes("sgmnt_data",264,"len")=8
	Set gtmtypes("sgmnt_data",264,"type")="uint64_t"
	Set gtmtypfldindx("sgmnt_data","wcb_wtstart_lckfail3_tn")=264
	Set gtmtypes("sgmnt_data",265,"name")="sgmnt_data.wcb_wtstart_lckfail4_tn"
	Set gtmtypes("sgmnt_data",265,"off")=2824
	Set gtmtypes("sgmnt_data",265,"len")=8
	Set gtmtypes("sgmnt_data",265,"type")="uint64_t"
	Set gtmtypfldindx("sgmnt_data","wcb_wtstart_lckfail4_tn")=265
	Set gtmtypes("sgmnt_data",266,"name")="sgmnt_data.wcb_wtfini_lckfail1_tn"
	Set gtmtypes("sgmnt_data",266,"off")=2832
	Set gtmtypes("sgmnt_data",266,"len")=8
	Set gtmtypes("sgmnt_data",266,"type")="uint64_t"
	Set gtmtypfldindx("sgmnt_data","wcb_wtfini_lckfail1_tn")=266
	Set gtmtypes("sgmnt_data",267,"name")="sgmnt_data.wcb_wtfini_lckfail2_tn"
	Set gtmtypes("sgmnt_data",267,"off")=2840
	Set gtmtypes("sgmnt_data",267,"len")=8
	Set gtmtypes("sgmnt_data",267,"type")="uint64_t"
	Set gtmtypfldindx("sgmnt_data","wcb_wtfini_lckfail2_tn")=267
	Set gtmtypes("sgmnt_data",268,"name")="sgmnt_data.wcb_wtfini_lckfail3_tn"
	Set gtmtypes("sgmnt_data",268,"off")=2848
	Set gtmtypes("sgmnt_data",268,"len")=8
	Set gtmtypes("sgmnt_data",268,"type")="uint64_t"
	Set gtmtypfldindx("sgmnt_data","wcb_wtfini_lckfail3_tn")=268
	Set gtmtypes("sgmnt_data",269,"name")="sgmnt_data.wcb_wtfini_lckfail4_tn"
	Set gtmtypes("sgmnt_data",269,"off")=2856
	Set gtmtypes("sgmnt_data",269,"len")=8
	Set gtmtypes("sgmnt_data",269,"type")="uint64_t"
	Set gtmtypfldindx("sgmnt_data","wcb_wtfini_lckfail4_tn")=269
	Set gtmtypes("sgmnt_data",270,"name")="sgmnt_data.wcb_t_end_sysops_dirtystuck1_tn"
	Set gtmtypes("sgmnt_data",270,"off")=2864
	Set gtmtypes("sgmnt_data",270,"len")=8
	Set gtmtypes("sgmnt_data",270,"type")="uint64_t"
	Set gtmtypfldindx("sgmnt_data","wcb_t_end_sysops_dirtystuck1_tn")=270
	Set gtmtypes("sgmnt_data",271,"name")="sgmnt_data.wcb_t_end_sysops_dirtystuck2_tn"
	Set gtmtypes("sgmnt_data",271,"off")=2872
	Set gtmtypes("sgmnt_data",271,"len")=8
	Set gtmtypes("sgmnt_data",271,"type")="uint64_t"
	Set gtmtypfldindx("sgmnt_data","wcb_t_end_sysops_dirtystuck2_tn")=271
	Set gtmtypes("sgmnt_data",272,"name")="sgmnt_data.wcb_secshr_db_clnup_wbuf_dqd_tn"
	Set gtmtypes("sgmnt_data",272,"off")=2880
	Set gtmtypes("sgmnt_data",272,"len")=8
	Set gtmtypes("sgmnt_data",272,"type")="uint64_t"
	Set gtmtypfldindx("sgmnt_data","wcb_secshr_db_clnup_wbuf_dqd_tn")=272
	Set gtmtypes("sgmnt_data",273,"name")="sgmnt_data.dwngrd_refmts_syncio_tn"
	Set gtmtypes("sgmnt_data",273,"off")=2888
	Set gtmtypes("sgmnt_data",273,"len")=8
	Set gtmtypes("sgmnt_data",273,"type")="uint64_t"
	Set gtmtypfldindx("sgmnt_data","dwngrd_refmts_syncio_tn")=273
	Set gtmtypes("sgmnt_data",274,"name")="sgmnt_data.dwngrd_refmts_asyncio_tn"
	Set gtmtypes("sgmnt_data",274,"off")=2896
	Set gtmtypes("sgmnt_data",274,"len")=8
	Set gtmtypes("sgmnt_data",274,"type")="uint64_t"
	Set gtmtypfldindx("sgmnt_data","dwngrd_refmts_asyncio_tn")=274
	Set gtmtypes("sgmnt_data",275,"name")="sgmnt_data.shmpool_refmt_harvests_tn"
	Set gtmtypes("sgmnt_data",275,"off")=2904
	Set gtmtypes("sgmnt_data",275,"len")=8
	Set gtmtypes("sgmnt_data",275,"type")="uint64_t"
	Set gtmtypfldindx("sgmnt_data","shmpool_refmt_harvests_tn")=275
	Set gtmtypes("sgmnt_data",276,"name")="sgmnt_data.shmpool_recovery_tn"
	Set gtmtypes("sgmnt_data",276,"off")=2912
	Set gtmtypes("sgmnt_data",276,"len")=8
	Set gtmtypes("sgmnt_data",276,"type")="uint64_t"
	Set gtmtypfldindx("sgmnt_data","shmpool_recovery_tn")=276
	Set gtmtypes("sgmnt_data",277,"name")="sgmnt_data.shmpool_blkd_by_sdc_tn"
	Set gtmtypes("sgmnt_data",277,"off")=2920
	Set gtmtypes("sgmnt_data",277,"len")=8
	Set gtmtypes("sgmnt_data",277,"type")="uint64_t"
	Set gtmtypfldindx("sgmnt_data","shmpool_blkd_by_sdc_tn")=277
	Set gtmtypes("sgmnt_data",278,"name")="sgmnt_data.shmpool_alloc_bbflush_tn"
	Set gtmtypes("sgmnt_data",278,"off")=2928
	Set gtmtypes("sgmnt_data",278,"len")=8
	Set gtmtypes("sgmnt_data",278,"type")="uint64_t"
	Set gtmtypfldindx("sgmnt_data","shmpool_alloc_bbflush_tn")=278
	Set gtmtypes("sgmnt_data",279,"name")="sgmnt_data.refmt_hvst_blk_released_replaced_tn"
	Set gtmtypes("sgmnt_data",279,"off")=2936
	Set gtmtypes("sgmnt_data",279,"len")=8
	Set gtmtypes("sgmnt_data",279,"type")="uint64_t"
	Set gtmtypfldindx("sgmnt_data","refmt_hvst_blk_released_replaced_tn")=279
	Set gtmtypes("sgmnt_data",280,"name")="sgmnt_data.refmt_hvst_blk_released_io_complete_tn"
	Set gtmtypes("sgmnt_data",280,"off")=2944
	Set gtmtypes("sgmnt_data",280,"len")=8
	Set gtmtypes("sgmnt_data",280,"type")="uint64_t"
	Set gtmtypfldindx("sgmnt_data","refmt_hvst_blk_released_io_complete_tn")=280
	Set gtmtypes("sgmnt_data",281,"name")="sgmnt_data.refmt_hvst_blk_kept_tn"
	Set gtmtypes("sgmnt_data",281,"off")=2952
	Set gtmtypes("sgmnt_data",281,"len")=8
	Set gtmtypes("sgmnt_data",281,"type")="uint64_t"
	Set gtmtypfldindx("sgmnt_data","refmt_hvst_blk_kept_tn")=281
	Set gtmtypes("sgmnt_data",282,"name")="sgmnt_data.refmt_hvst_blk_ignored_tn"
	Set gtmtypes("sgmnt_data",282,"off")=2960
	Set gtmtypes("sgmnt_data",282,"len")=8
	Set gtmtypes("sgmnt_data",282,"type")="uint64_t"
	Set gtmtypfldindx("sgmnt_data","refmt_hvst_blk_ignored_tn")=282
	Set gtmtypes("sgmnt_data",283,"name")="sgmnt_data.refmt_blk_chk_blk_freed_tn"
	Set gtmtypes("sgmnt_data",283,"off")=2968
	Set gtmtypes("sgmnt_data",283,"len")=8
	Set gtmtypes("sgmnt_data",283,"type")="uint64_t"
	Set gtmtypfldindx("sgmnt_data","refmt_blk_chk_blk_freed_tn")=283
	Set gtmtypes("sgmnt_data",284,"name")="sgmnt_data.refmt_blk_chk_blk_kept_tn"
	Set gtmtypes("sgmnt_data",284,"off")=2976
	Set gtmtypes("sgmnt_data",284,"len")=8
	Set gtmtypes("sgmnt_data",284,"type")="uint64_t"
	Set gtmtypfldindx("sgmnt_data","refmt_blk_chk_blk_kept_tn")=284
	Set gtmtypes("sgmnt_data",285,"name")="sgmnt_data.active_lvl_trigger_tn"
	Set gtmtypes("sgmnt_data",285,"off")=2984
	Set gtmtypes("sgmnt_data",285,"len")=8
	Set gtmtypes("sgmnt_data",285,"type")="uint64_t"
	Set gtmtypfldindx("sgmnt_data","active_lvl_trigger_tn")=285
	Set gtmtypes("sgmnt_data",286,"name")="sgmnt_data.new_buff_tn"
	Set gtmtypes("sgmnt_data",286,"off")=2992
	Set gtmtypes("sgmnt_data",286,"len")=8
	Set gtmtypes("sgmnt_data",286,"type")="uint64_t"
	Set gtmtypfldindx("sgmnt_data","new_buff_tn")=286
	Set gtmtypes("sgmnt_data",287,"name")="sgmnt_data.phase2_commit_wait_sleep_in_crit_tn"
	Set gtmtypes("sgmnt_data",287,"off")=3000
	Set gtmtypes("sgmnt_data",287,"len")=8
	Set gtmtypes("sgmnt_data",287,"type")="uint64_t"
	Set gtmtypfldindx("sgmnt_data","phase2_commit_wait_sleep_in_crit_tn")=287
	Set gtmtypes("sgmnt_data",288,"name")="sgmnt_data.phase2_commit_wait_sleep_no_crit_tn"
	Set gtmtypes("sgmnt_data",288,"off")=3008
	Set gtmtypes("sgmnt_data",288,"len")=8
	Set gtmtypes("sgmnt_data",288,"type")="uint64_t"
	Set gtmtypfldindx("sgmnt_data","phase2_commit_wait_sleep_no_crit_tn")=288
	Set gtmtypes("sgmnt_data",289,"name")="sgmnt_data.phase2_commit_wait_pidcnt_tn"
	Set gtmtypes("sgmnt_data",289,"off")=3016
	Set gtmtypes("sgmnt_data",289,"len")=8
	Set gtmtypes("sgmnt_data",289,"type")="uint64_t"
	Set gtmtypfldindx("sgmnt_data","phase2_commit_wait_pidcnt_tn")=289
	Set gtmtypes("sgmnt_data",290,"name")="sgmnt_data.wcb_t_end_sysops_intend_wait_tn"
	Set gtmtypes("sgmnt_data",290,"off")=3024
	Set gtmtypes("sgmnt_data",290,"len")=8
	Set gtmtypes("sgmnt_data",290,"type")="uint64_t"
	Set gtmtypfldindx("sgmnt_data","wcb_t_end_sysops_intend_wait_tn")=290
	Set gtmtypes("sgmnt_data",291,"name")="sgmnt_data.wcb_secshr_db_clnup_phase2_clnup_tn"
	Set gtmtypes("sgmnt_data",291,"off")=3032
	Set gtmtypes("sgmnt_data",291,"len")=8
	Set gtmtypes("sgmnt_data",291,"type")="uint64_t"
	Set gtmtypfldindx("sgmnt_data","wcb_secshr_db_clnup_phase2_clnup_tn")=291
	Set gtmtypes("sgmnt_data",292,"name")="sgmnt_data.wcb_phase2_commit_wait_tn"
	Set gtmtypes("sgmnt_data",292,"off")=3040
	Set gtmtypes("sgmnt_data",292,"len")=8
	Set gtmtypes("sgmnt_data",292,"type")="uint64_t"
	Set gtmtypfldindx("sgmnt_data","wcb_phase2_commit_wait_tn")=292
	Set gtmtypes("sgmnt_data",293,"name")="sgmnt_data.recompute_upd_array_calls_tn"
	Set gtmtypes("sgmnt_data",293,"off")=3048
	Set gtmtypes("sgmnt_data",293,"len")=8
	Set gtmtypes("sgmnt_data",293,"type")="uint64_t"
	Set gtmtypfldindx("sgmnt_data","recompute_upd_array_calls_tn")=293
	Set gtmtypes("sgmnt_data",294,"name")="sgmnt_data.recompute_upd_array_rip_tn"
	Set gtmtypes("sgmnt_data",294,"off")=3056
	Set gtmtypes("sgmnt_data",294,"len")=8
	Set gtmtypes("sgmnt_data",294,"type")="uint64_t"
	Set gtmtypfldindx("sgmnt_data","recompute_upd_array_rip_tn")=294
	Set gtmtypes("sgmnt_data",295,"name")="sgmnt_data.recompute_upd_array_in_tend_tn"
	Set gtmtypes("sgmnt_data",295,"off")=3064
	Set gtmtypes("sgmnt_data",295,"len")=8
	Set gtmtypes("sgmnt_data",295,"type")="uint64_t"
	Set gtmtypfldindx("sgmnt_data","recompute_upd_array_in_tend_tn")=295
	Set gtmtypes("sgmnt_data",296,"name")="sgmnt_data.recompute_upd_array_search_blk_tn"
	Set gtmtypes("sgmnt_data",296,"off")=3072
	Set gtmtypes("sgmnt_data",296,"len")=8
	Set gtmtypes("sgmnt_data",296,"type")="uint64_t"
	Set gtmtypfldindx("sgmnt_data","recompute_upd_array_search_blk_tn")=296
	Set gtmtypes("sgmnt_data",297,"name")="sgmnt_data.recompute_upd_array_new_rec_tn"
	Set gtmtypes("sgmnt_data",297,"off")=3080
	Set gtmtypes("sgmnt_data",297,"len")=8
	Set gtmtypes("sgmnt_data",297,"type")="uint64_t"
	Set gtmtypfldindx("sgmnt_data","recompute_upd_array_new_rec_tn")=297
	Set gtmtypes("sgmnt_data",298,"name")="sgmnt_data.recompute_upd_array_rec_size_tn"
	Set gtmtypes("sgmnt_data",298,"off")=3088
	Set gtmtypes("sgmnt_data",298,"len")=8
	Set gtmtypes("sgmnt_data",298,"type")="uint64_t"
	Set gtmtypfldindx("sgmnt_data","recompute_upd_array_rec_size_tn")=298
	Set gtmtypes("sgmnt_data",299,"name")="sgmnt_data.recompute_upd_array_rec_cmpc_tn"
	Set gtmtypes("sgmnt_data",299,"off")=3096
	Set gtmtypes("sgmnt_data",299,"len")=8
	Set gtmtypes("sgmnt_data",299,"type")="uint64_t"
	Set gtmtypfldindx("sgmnt_data","recompute_upd_array_rec_cmpc_tn")=299
	Set gtmtypes("sgmnt_data",300,"name")="sgmnt_data.recompute_upd_array_blk_fini_tn"
	Set gtmtypes("sgmnt_data",300,"off")=3104
	Set gtmtypes("sgmnt_data",300,"len")=8
	Set gtmtypes("sgmnt_data",300,"type")="uint64_t"
	Set gtmtypfldindx("sgmnt_data","recompute_upd_array_blk_fini_tn")=300
	Set gtmtypes("sgmnt_data",301,"name")="sgmnt_data.recompute_upd_array_blk_split_tn"
	Set gtmtypes("sgmnt_data",301,"off")=3112
	Set gtmtypes("sgmnt_data",301,"len")=8
	Set gtmtypes("sgmnt_data",301,"type")="uint64_t"
	Set gtmtypfldindx("sgmnt_data","recompute_upd_array_blk_split_tn")=301
	Set gtmtypes("sgmnt_data",302,"name")="sgmnt_data.unused_filler_0135_tn"
	Set gtmtypes("sgmnt_data",302,"off")=3120
	Set gtmtypes("sgmnt_data",302,"len")=8
	Set gtmtypes("sgmnt_data",302,"type")="uint64_t"
	Set gtmtypfldindx("sgmnt_data","unused_filler_0135_tn")=302
	Set gtmtypes("sgmnt_data",303,"name")="sgmnt_data.t_qread_ripsleep_nblks_tn"
	Set gtmtypes("sgmnt_data",303,"off")=3128
	Set gtmtypes("sgmnt_data",303,"len")=8
	Set gtmtypes("sgmnt_data",303,"type")="uint64_t"
	Set gtmtypfldindx("sgmnt_data","t_qread_ripsleep_nblks_tn")=303
	Set gtmtypes("sgmnt_data",304,"name")="sgmnt_data.wcs_wtfini_invoked_tn"
	Set gtmtypes("sgmnt_data",304,"off")=3136
	Set gtmtypes("sgmnt_data",304,"len")=8
	Set gtmtypes("sgmnt_data",304,"type")="uint64_t"
	Set gtmtypfldindx("sgmnt_data","wcs_wtfini_invoked_tn")=304
	Set gtmtypes("sgmnt_data",305,"name")="sgmnt_data.wcs_wterror_invoked_tn"
	Set gtmtypes("sgmnt_data",305,"off")=3144
	Set gtmtypes("sgmnt_data",305,"len")=8
	Set gtmtypes("sgmnt_data",305,"type")="uint64_t"
	Set gtmtypfldindx("sgmnt_data","wcs_wterror_invoked_tn")=305
	Set gtmtypes("sgmnt_data",306,"name")="sgmnt_data.wcs_wt_restart_invoked_tn"
	Set gtmtypes("sgmnt_data",306,"off")=3152
	Set gtmtypes("sgmnt_data",306,"len")=8
	Set gtmtypes("sgmnt_data",306,"type")="uint64_t"
	Set gtmtypfldindx("sgmnt_data","wcs_wt_restart_invoked_tn")=306
	Set gtmtypes("sgmnt_data",307,"name")="sgmnt_data.wcs_wt_restart_eagain_tn"
	Set gtmtypes("sgmnt_data",307,"off")=3160
	Set gtmtypes("sgmnt_data",307,"len")=8
	Set gtmtypes("sgmnt_data",307,"type")="uint64_t"
	Set gtmtypfldindx("sgmnt_data","wcs_wt_restart_eagain_tn")=307
	Set gtmtypes("sgmnt_data",308,"name")="sgmnt_data.wcs_wt_restart_reinsert_tn"
	Set gtmtypes("sgmnt_data",308,"off")=3168
	Set gtmtypes("sgmnt_data",308,"len")=8
	Set gtmtypes("sgmnt_data",308,"type")="uint64_t"
	Set gtmtypfldindx("sgmnt_data","wcs_wt_restart_reinsert_tn")=308
	Set gtmtypes("sgmnt_data",309,"name")="sgmnt_data.wcs_wtstart_eagain_tn"
	Set gtmtypes("sgmnt_data",309,"off")=3176
	Set gtmtypes("sgmnt_data",309,"len")=8
	Set gtmtypes("sgmnt_data",309,"type")="uint64_t"
	Set gtmtypfldindx("sgmnt_data","wcs_wtstart_eagain_tn")=309
	Set gtmtypes("sgmnt_data",310,"name")="sgmnt_data.wcs_wtstart_eagain_incrit_tn"
	Set gtmtypes("sgmnt_data",310,"off")=3184
	Set gtmtypes("sgmnt_data",310,"len")=8
	Set gtmtypes("sgmnt_data",310,"type")="uint64_t"
	Set gtmtypfldindx("sgmnt_data","wcs_wtstart_eagain_incrit_tn")=310
	Set gtmtypes("sgmnt_data",311,"name")="sgmnt_data.async_restart_eagain_tn"
	Set gtmtypes("sgmnt_data",311,"off")=3192
	Set gtmtypes("sgmnt_data",311,"len")=8
	Set gtmtypes("sgmnt_data",311,"type")="uint64_t"
	Set gtmtypfldindx("sgmnt_data","async_restart_eagain_tn")=311
	Set gtmtypes("sgmnt_data",312,"name")="sgmnt_data.jnlbuff_phs2cmt_array_full_tn"
	Set gtmtypes("sgmnt_data",312,"off")=3200
	Set gtmtypes("sgmnt_data",312,"len")=8
	Set gtmtypes("sgmnt_data",312,"type")="uint64_t"
	Set gtmtypfldindx("sgmnt_data","jnlbuff_phs2cmt_array_full_tn")=312
	Set gtmtypes("sgmnt_data",313,"name")="sgmnt_data.jnl_phase2_cleanup_if_possible_tn"
	Set gtmtypes("sgmnt_data",313,"off")=3208
	Set gtmtypes("sgmnt_data",313,"len")=8
	Set gtmtypes("sgmnt_data",313,"type")="uint64_t"
	Set gtmtypfldindx("sgmnt_data","jnl_phase2_cleanup_if_possible_tn")=313
	Set gtmtypes("sgmnt_data",314,"name")="sgmnt_data.jnlbuff_phs2cmt_isprcalv_tn"
	Set gtmtypes("sgmnt_data",314,"off")=3216
	Set gtmtypes("sgmnt_data",314,"len")=8
	Set gtmtypes("sgmnt_data",314,"type")="uint64_t"
	Set gtmtypfldindx("sgmnt_data","jnlbuff_phs2cmt_isprcalv_tn")=314
	Set gtmtypes("sgmnt_data",315,"name")="sgmnt_data.recompute_upd_array_paidoff_tn"
	Set gtmtypes("sgmnt_data",315,"off")=3224
	Set gtmtypes("sgmnt_data",315,"len")=8
	Set gtmtypes("sgmnt_data",315,"type")="uint64_t"
	Set gtmtypfldindx("sgmnt_data","recompute_upd_array_paidoff_tn")=315
	Set gtmtypes("sgmnt_data",316,"name")="sgmnt_data.bg_trc_rec_tn_filler"
	Set gtmtypes("sgmnt_data",316,"off")=3232
	Set gtmtypes("sgmnt_data",316,"len")=16
	Set gtmtypes("sgmnt_data",316,"type")="char"
	Set gtmtypfldindx("sgmnt_data","bg_trc_rec_tn_filler")=316
	Set gtmtypes("sgmnt_data",317,"name")="sgmnt_data.total_buffer_flush_cntr"
	Set gtmtypes("sgmnt_data",317,"off")=3248
	Set gtmtypes("sgmnt_data",317,"len")=4
	Set gtmtypes("sgmnt_data",317,"type")="int"
	Set gtmtypfldindx("sgmnt_data","total_buffer_flush_cntr")=317
	Set gtmtypes("sgmnt_data",318,"name")="sgmnt_data.bufct_buffer_flush_cntr"
	Set gtmtypes("sgmnt_data",318,"off")=3252
	Set gtmtypes("sgmnt_data",318,"len")=4
	Set gtmtypes("sgmnt_data",318,"type")="int"
	Set gtmtypfldindx("sgmnt_data","bufct_buffer_flush_cntr")=318
	Set gtmtypes("sgmnt_data",319,"name")="sgmnt_data.bufct_buffer_flush_loop_cntr"
	Set gtmtypes("sgmnt_data",319,"off")=3256
	Set gtmtypes("sgmnt_data",319,"len")=4
	Set gtmtypes("sgmnt_data",319,"type")="int"
	Set gtmtypfldindx("sgmnt_data","bufct_buffer_flush_loop_cntr")=319
	Set gtmtypes("sgmnt_data",320,"name")="sgmnt_data.wrt_calls_cntr"
	Set gtmtypes("sgmnt_data",320,"off")=3260
	Set gtmtypes("sgmnt_data",320,"len")=4
	Set gtmtypes("sgmnt_data",320,"type")="int"
	Set gtmtypfldindx("sgmnt_data","wrt_calls_cntr")=320
	Set gtmtypes("sgmnt_data",321,"name")="sgmnt_data.wrt_count_cntr"
	Set gtmtypes("sgmnt_data",321,"off")=3264
	Set gtmtypes("sgmnt_data",321,"len")=4
	Set gtmtypes("sgmnt_data",321,"type")="int"
	Set gtmtypfldindx("sgmnt_data","wrt_count_cntr")=321
	Set gtmtypes("sgmnt_data",322,"name")="sgmnt_data.wrt_blocked_cntr"
	Set gtmtypes("sgmnt_data",322,"off")=3268
	Set gtmtypes("sgmnt_data",322,"len")=4
	Set gtmtypes("sgmnt_data",322,"type")="int"
	Set gtmtypfldindx("sgmnt_data","wrt_blocked_cntr")=322
	Set gtmtypes("sgmnt_data",323,"name")="sgmnt_data.wrt_busy_cntr"
	Set gtmtypes("sgmnt_data",323,"off")=3272
	Set gtmtypes("sgmnt_data",323,"len")=4
	Set gtmtypes("sgmnt_data",323,"type")="int"
	Set gtmtypfldindx("sgmnt_data","wrt_busy_cntr")=323
	Set gtmtypes("sgmnt_data",324,"name")="sgmnt_data.wrt_noblks_wrtn_cntr"
	Set gtmtypes("sgmnt_data",324,"off")=3276
	Set gtmtypes("sgmnt_data",324,"len")=4
	Set gtmtypes("sgmnt_data",324,"type")="int"
	Set gtmtypfldindx("sgmnt_data","wrt_noblks_wrtn_cntr")=324
	Set gtmtypes("sgmnt_data",325,"name")="sgmnt_data.wcb_gds_rundown2_cntr"
	Set gtmtypes("sgmnt_data",325,"off")=3280
	Set gtmtypes("sgmnt_data",325,"len")=4
	Set gtmtypes("sgmnt_data",325,"type")="int"
	Set gtmtypfldindx("sgmnt_data","wcb_gds_rundown2_cntr")=325
	Set gtmtypes("sgmnt_data",326,"name")="sgmnt_data.reserved_bgtrcrec2_cntr"
	Set gtmtypes("sgmnt_data",326,"off")=3284
	Set gtmtypes("sgmnt_data",326,"len")=4
	Set gtmtypes("sgmnt_data",326,"type")="int"
	Set gtmtypfldindx("sgmnt_data","reserved_bgtrcrec2_cntr")=326
	Set gtmtypes("sgmnt_data",327,"name")="sgmnt_data.reserved_bgtrcrec3_cntr"
	Set gtmtypes("sgmnt_data",327,"off")=3288
	Set gtmtypes("sgmnt_data",327,"len")=4
	Set gtmtypes("sgmnt_data",327,"type")="int"
	Set gtmtypfldindx("sgmnt_data","reserved_bgtrcrec3_cntr")=327
	Set gtmtypes("sgmnt_data",328,"name")="sgmnt_data.lost_block_recovery_cntr"
	Set gtmtypes("sgmnt_data",328,"off")=3292
	Set gtmtypes("sgmnt_data",328,"len")=4
	Set gtmtypes("sgmnt_data",328,"type")="int"
	Set gtmtypfldindx("sgmnt_data","lost_block_recovery_cntr")=328
	Set gtmtypes("sgmnt_data",329,"name")="sgmnt_data.wc_blocked_onln_rlbk_cntr"
	Set gtmtypes("sgmnt_data",329,"off")=3296
	Set gtmtypes("sgmnt_data",329,"len")=4
	Set gtmtypes("sgmnt_data",329,"type")="int"
	Set gtmtypfldindx("sgmnt_data","wc_blocked_onln_rlbk_cntr")=329
	Set gtmtypes("sgmnt_data",330,"name")="sgmnt_data.stale_cntr"
	Set gtmtypes("sgmnt_data",330,"off")=3300
	Set gtmtypes("sgmnt_data",330,"len")=4
	Set gtmtypes("sgmnt_data",330,"type")="int"
	Set gtmtypfldindx("sgmnt_data","stale_cntr")=330
	Set gtmtypes("sgmnt_data",331,"name")="sgmnt_data.starved_cntr"
	Set gtmtypes("sgmnt_data",331,"off")=3304
	Set gtmtypes("sgmnt_data",331,"len")=4
	Set gtmtypes("sgmnt_data",331,"type")="int"
	Set gtmtypfldindx("sgmnt_data","starved_cntr")=331
	Set gtmtypes("sgmnt_data",332,"name")="sgmnt_data.stale_timer_started_cntr"
	Set gtmtypes("sgmnt_data",332,"off")=3308
	Set gtmtypes("sgmnt_data",332,"len")=4
	Set gtmtypes("sgmnt_data",332,"type")="int"
	Set gtmtypfldindx("sgmnt_data","stale_timer_started_cntr")=332
	Set gtmtypes("sgmnt_data",333,"name")="sgmnt_data.stale_timer_pop_cntr"
	Set gtmtypes("sgmnt_data",333,"off")=3312
	Set gtmtypes("sgmnt_data",333,"len")=4
	Set gtmtypes("sgmnt_data",333,"type")="int"
	Set gtmtypfldindx("sgmnt_data","stale_timer_pop_cntr")=333
	Set gtmtypes("sgmnt_data",334,"name")="sgmnt_data.stale_process_defer_cntr"
	Set gtmtypes("sgmnt_data",334,"off")=3316
	Set gtmtypes("sgmnt_data",334,"len")=4
	Set gtmtypes("sgmnt_data",334,"type")="int"
	Set gtmtypfldindx("sgmnt_data","stale_process_defer_cntr")=334
	Set gtmtypes("sgmnt_data",335,"name")="sgmnt_data.stale_defer_processed_cntr"
	Set gtmtypes("sgmnt_data",335,"off")=3320
	Set gtmtypes("sgmnt_data",335,"len")=4
	Set gtmtypes("sgmnt_data",335,"type")="int"
	Set gtmtypfldindx("sgmnt_data","stale_defer_processed_cntr")=335
	Set gtmtypes("sgmnt_data",336,"name")="sgmnt_data.spcfc_buffer_flush_cntr"
	Set gtmtypes("sgmnt_data",336,"off")=3324
	Set gtmtypes("sgmnt_data",336,"len")=4
	Set gtmtypes("sgmnt_data",336,"type")="int"
	Set gtmtypfldindx("sgmnt_data","spcfc_buffer_flush_cntr")=336
	Set gtmtypes("sgmnt_data",337,"name")="sgmnt_data.spcfc_buffer_flush_loop_cntr"
	Set gtmtypes("sgmnt_data",337,"off")=3328
	Set gtmtypes("sgmnt_data",337,"len")=4
	Set gtmtypes("sgmnt_data",337,"type")="int"
	Set gtmtypfldindx("sgmnt_data","spcfc_buffer_flush_loop_cntr")=337
	Set gtmtypes("sgmnt_data",338,"name")="sgmnt_data.spcfc_buffer_flush_retries_cntr"
	Set gtmtypes("sgmnt_data",338,"off")=3332
	Set gtmtypes("sgmnt_data",338,"len")=4
	Set gtmtypes("sgmnt_data",338,"type")="int"
	Set gtmtypfldindx("sgmnt_data","spcfc_buffer_flush_retries_cntr")=338
	Set gtmtypes("sgmnt_data",339,"name")="sgmnt_data.spcfc_buffer_flushed_during_lockwait_cntr"
	Set gtmtypes("sgmnt_data",339,"off")=3336
	Set gtmtypes("sgmnt_data",339,"len")=4
	Set gtmtypes("sgmnt_data",339,"type")="int"
	Set gtmtypfldindx("sgmnt_data","spcfc_buffer_flushed_during_lockwait_cntr")=339
	Set gtmtypes("sgmnt_data",340,"name")="sgmnt_data.tp_crit_retries_cntr"
	Set gtmtypes("sgmnt_data",340,"off")=3340
	Set gtmtypes("sgmnt_data",340,"len")=4
	Set gtmtypes("sgmnt_data",340,"type")="int"
	Set gtmtypfldindx("sgmnt_data","tp_crit_retries_cntr")=340
	Set gtmtypes("sgmnt_data",341,"name")="sgmnt_data.db_csh_getn_flush_dirty_cntr"
	Set gtmtypes("sgmnt_data",341,"off")=3344
	Set gtmtypes("sgmnt_data",341,"len")=4
	Set gtmtypes("sgmnt_data",341,"type")="int"
	Set gtmtypfldindx("sgmnt_data","db_csh_getn_flush_dirty_cntr")=341
	Set gtmtypes("sgmnt_data",342,"name")="sgmnt_data.db_csh_getn_rip_wait_cntr"
	Set gtmtypes("sgmnt_data",342,"off")=3348
	Set gtmtypes("sgmnt_data",342,"len")=4
	Set gtmtypes("sgmnt_data",342,"type")="int"
	Set gtmtypfldindx("sgmnt_data","db_csh_getn_rip_wait_cntr")=342
	Set gtmtypes("sgmnt_data",343,"name")="sgmnt_data.db_csh_getn_buf_owner_stuck_cntr"
	Set gtmtypes("sgmnt_data",343,"off")=3352
	Set gtmtypes("sgmnt_data",343,"len")=4
	Set gtmtypes("sgmnt_data",343,"type")="int"
	Set gtmtypfldindx("sgmnt_data","db_csh_getn_buf_owner_stuck_cntr")=343
	Set gtmtypes("sgmnt_data",344,"name")="sgmnt_data.db_csh_getn_out_of_design_cntr"
	Set gtmtypes("sgmnt_data",344,"off")=3356
	Set gtmtypes("sgmnt_data",344,"len")=4
	Set gtmtypes("sgmnt_data",344,"type")="int"
	Set gtmtypfldindx("sgmnt_data","db_csh_getn_out_of_design_cntr")=344
	Set gtmtypes("sgmnt_data",345,"name")="sgmnt_data.t_qread_buf_owner_stuck_cntr"
	Set gtmtypes("sgmnt_data",345,"off")=3360
	Set gtmtypes("sgmnt_data",345,"len")=4
	Set gtmtypes("sgmnt_data",345,"type")="int"
	Set gtmtypfldindx("sgmnt_data","t_qread_buf_owner_stuck_cntr")=345
	Set gtmtypes("sgmnt_data",346,"name")="sgmnt_data.t_qread_out_of_design_cntr"
	Set gtmtypes("sgmnt_data",346,"off")=3364
	Set gtmtypes("sgmnt_data",346,"len")=4
	Set gtmtypes("sgmnt_data",346,"type")="int"
	Set gtmtypfldindx("sgmnt_data","t_qread_out_of_design_cntr")=346
	Set gtmtypes("sgmnt_data",347,"name")="sgmnt_data.unused_filler_0031_cntr"
	Set gtmtypes("sgmnt_data",347,"off")=3368
	Set gtmtypes("sgmnt_data",347,"len")=4
	Set gtmtypes("sgmnt_data",347,"type")="int"
	Set gtmtypfldindx("sgmnt_data","unused_filler_0031_cntr")=347
	Set gtmtypes("sgmnt_data",348,"name")="sgmnt_data.mlock_wakeups_cntr"
	Set gtmtypes("sgmnt_data",348,"off")=3372
	Set gtmtypes("sgmnt_data",348,"len")=4
	Set gtmtypes("sgmnt_data",348,"type")="int"
	Set gtmtypfldindx("sgmnt_data","mlock_wakeups_cntr")=348
	Set gtmtypes("sgmnt_data",349,"name")="sgmnt_data.wc_blocked_wcs_recover_invoked_cntr"
	Set gtmtypes("sgmnt_data",349,"off")=3376
	Set gtmtypes("sgmnt_data",349,"len")=4
	Set gtmtypes("sgmnt_data",349,"type")="int"
	Set gtmtypfldindx("sgmnt_data","wc_blocked_wcs_recover_invoked_cntr")=349
	Set gtmtypes("sgmnt_data",350,"name")="sgmnt_data.wc_blocked_wcs_verify_passed_cntr"
	Set gtmtypes("sgmnt_data",350,"off")=3380
	Set gtmtypes("sgmnt_data",350,"len")=4
	Set gtmtypes("sgmnt_data",350,"type")="int"
	Set gtmtypfldindx("sgmnt_data","wc_blocked_wcs_verify_passed_cntr")=350
	Set gtmtypes("sgmnt_data",351,"name")="sgmnt_data.wc_blocked_t_qread_db_csh_getn_invalid_blk_cntr"
	Set gtmtypes("sgmnt_data",351,"off")=3384
	Set gtmtypes("sgmnt_data",351,"len")=4
	Set gtmtypes("sgmnt_data",351,"type")="int"
	Set gtmtypfldindx("sgmnt_data","wc_blocked_t_qread_db_csh_getn_invalid_blk_cntr")=351
	Set gtmtypes("sgmnt_data",352,"name")="sgmnt_data.wc_blocked_t_qread_db_csh_get_invalid_blk_cntr"
	Set gtmtypes("sgmnt_data",352,"off")=3388
	Set gtmtypes("sgmnt_data",352,"len")=4
	Set gtmtypes("sgmnt_data",352,"type")="int"
	Set gtmtypfldindx("sgmnt_data","wc_blocked_t_qread_db_csh_get_invalid_blk_cntr")=352
	Set gtmtypes("sgmnt_data",353,"name")="sgmnt_data.wc_blocked_db_csh_getn_loopexceed_cntr"
	Set gtmtypes("sgmnt_data",353,"off")=3392
	Set gtmtypes("sgmnt_data",353,"len")=4
	Set gtmtypes("sgmnt_data",353,"type")="int"
	Set gtmtypfldindx("sgmnt_data","wc_blocked_db_csh_getn_loopexceed_cntr")=353
	Set gtmtypes("sgmnt_data",354,"name")="sgmnt_data.wc_blocked_db_csh_getn_wcsstarvewrt_cntr"
	Set gtmtypes("sgmnt_data",354,"off")=3396
	Set gtmtypes("sgmnt_data",354,"len")=4
	Set gtmtypes("sgmnt_data",354,"type")="int"
	Set gtmtypfldindx("sgmnt_data","wc_blocked_db_csh_getn_wcsstarvewrt_cntr")=354
	Set gtmtypes("sgmnt_data",355,"name")="sgmnt_data.wc_blocked_db_csh_get_cntr"
	Set gtmtypes("sgmnt_data",355,"off")=3400
	Set gtmtypes("sgmnt_data",355,"len")=4
	Set gtmtypes("sgmnt_data",355,"type")="int"
	Set gtmtypfldindx("sgmnt_data","wc_blocked_db_csh_get_cntr")=355
	Set gtmtypes("sgmnt_data",356,"name")="sgmnt_data.wc_blocked_tp_tend_wcsgetspace_cntr"
	Set gtmtypes("sgmnt_data",356,"off")=3404
	Set gtmtypes("sgmnt_data",356,"len")=4
	Set gtmtypes("sgmnt_data",356,"type")="int"
	Set gtmtypfldindx("sgmnt_data","wc_blocked_tp_tend_wcsgetspace_cntr")=356
	Set gtmtypes("sgmnt_data",357,"name")="sgmnt_data.wc_blocked_tp_tend_t1_cntr"
	Set gtmtypes("sgmnt_data",357,"off")=3408
	Set gtmtypes("sgmnt_data",357,"len")=4
	Set gtmtypes("sgmnt_data",357,"type")="int"
	Set gtmtypfldindx("sgmnt_data","wc_blocked_tp_tend_t1_cntr")=357
	Set gtmtypes("sgmnt_data",358,"name")="sgmnt_data.wc_blocked_tp_tend_bitmap_cntr"
	Set gtmtypes("sgmnt_data",358,"off")=3412
	Set gtmtypes("sgmnt_data",358,"len")=4
	Set gtmtypes("sgmnt_data",358,"type")="int"
	Set gtmtypfldindx("sgmnt_data","wc_blocked_tp_tend_bitmap_cntr")=358
	Set gtmtypes("sgmnt_data",359,"name")="sgmnt_data.wc_blocked_tp_tend_jnl_cwset_cntr"
	Set gtmtypes("sgmnt_data",359,"off")=3416
	Set gtmtypes("sgmnt_data",359,"len")=4
	Set gtmtypes("sgmnt_data",359,"type")="int"
	Set gtmtypfldindx("sgmnt_data","wc_blocked_tp_tend_jnl_cwset_cntr")=359
	Set gtmtypes("sgmnt_data",360,"name")="sgmnt_data.wc_blocked_tp_tend_jnl_wcsflu_cntr"
	Set gtmtypes("sgmnt_data",360,"off")=3420
	Set gtmtypes("sgmnt_data",360,"len")=4
	Set gtmtypes("sgmnt_data",360,"type")="int"
	Set gtmtypfldindx("sgmnt_data","wc_blocked_tp_tend_jnl_wcsflu_cntr")=360
	Set gtmtypes("sgmnt_data",361,"name")="sgmnt_data.wc_blocked_t_end_hist_cntr"
	Set gtmtypes("sgmnt_data",361,"off")=3424
	Set gtmtypes("sgmnt_data",361,"len")=4
	Set gtmtypes("sgmnt_data",361,"type")="int"
	Set gtmtypfldindx("sgmnt_data","wc_blocked_t_end_hist_cntr")=361
	Set gtmtypes("sgmnt_data",362,"name")="sgmnt_data.wc_blocked_t_end_hist1_nullbt_cntr"
	Set gtmtypes("sgmnt_data",362,"off")=3428
	Set gtmtypes("sgmnt_data",362,"len")=4
	Set gtmtypes("sgmnt_data",362,"type")="int"
	Set gtmtypfldindx("sgmnt_data","wc_blocked_t_end_hist1_nullbt_cntr")=362
	Set gtmtypes("sgmnt_data",363,"name")="sgmnt_data.wc_blocked_t_end_hist1_nonnullbt_cntr"
	Set gtmtypes("sgmnt_data",363,"off")=3432
	Set gtmtypes("sgmnt_data",363,"len")=4
	Set gtmtypes("sgmnt_data",363,"type")="int"
	Set gtmtypfldindx("sgmnt_data","wc_blocked_t_end_hist1_nonnullbt_cntr")=363
	Set gtmtypes("sgmnt_data",364,"name")="sgmnt_data.wc_blocked_t_end_bitmap_nullbt_cntr"
	Set gtmtypes("sgmnt_data",364,"off")=3436
	Set gtmtypes("sgmnt_data",364,"len")=4
	Set gtmtypes("sgmnt_data",364,"type")="int"
	Set gtmtypfldindx("sgmnt_data","wc_blocked_t_end_bitmap_nullbt_cntr")=364
	Set gtmtypes("sgmnt_data",365,"name")="sgmnt_data.wc_blocked_t_end_bitmap_nonnullbt_cntr"
	Set gtmtypes("sgmnt_data",365,"off")=3440
	Set gtmtypes("sgmnt_data",365,"len")=4
	Set gtmtypes("sgmnt_data",365,"type")="int"
	Set gtmtypfldindx("sgmnt_data","wc_blocked_t_end_bitmap_nonnullbt_cntr")=365
	Set gtmtypes("sgmnt_data",366,"name")="sgmnt_data.wc_blocked_t_end_jnl_cwset_cntr"
	Set gtmtypes("sgmnt_data",366,"off")=3444
	Set gtmtypes("sgmnt_data",366,"len")=4
	Set gtmtypes("sgmnt_data",366,"type")="int"
	Set gtmtypfldindx("sgmnt_data","wc_blocked_t_end_jnl_cwset_cntr")=366
	Set gtmtypes("sgmnt_data",367,"name")="sgmnt_data.wc_blocked_t_end_jnl_wcsflu_cntr"
	Set gtmtypes("sgmnt_data",367,"off")=3448
	Set gtmtypes("sgmnt_data",367,"len")=4
	Set gtmtypes("sgmnt_data",367,"type")="int"
	Set gtmtypfldindx("sgmnt_data","wc_blocked_t_end_jnl_wcsflu_cntr")=367
	Set gtmtypes("sgmnt_data",368,"name")="sgmnt_data.unused_filler_0052_cntr"
	Set gtmtypes("sgmnt_data",368,"off")=3452
	Set gtmtypes("sgmnt_data",368,"len")=4
	Set gtmtypes("sgmnt_data",368,"type")="int"
	Set gtmtypfldindx("sgmnt_data","unused_filler_0052_cntr")=368
	Set gtmtypes("sgmnt_data",369,"name")="sgmnt_data.wc_blocked_tpckh_hist1_nullbt_cntr"
	Set gtmtypes("sgmnt_data",369,"off")=3456
	Set gtmtypes("sgmnt_data",369,"len")=4
	Set gtmtypes("sgmnt_data",369,"type")="int"
	Set gtmtypfldindx("sgmnt_data","wc_blocked_tpckh_hist1_nullbt_cntr")=369
	Set gtmtypes("sgmnt_data",370,"name")="sgmnt_data.wc_blocked_tpckh_hist1_nonnullbt_cntr"
	Set gtmtypes("sgmnt_data",370,"off")=3460
	Set gtmtypes("sgmnt_data",370,"len")=4
	Set gtmtypes("sgmnt_data",370,"type")="int"
	Set gtmtypfldindx("sgmnt_data","wc_blocked_tpckh_hist1_nonnullbt_cntr")=370
	Set gtmtypes("sgmnt_data",371,"name")="sgmnt_data.jnl_blocked_writer_lost_cntr"
	Set gtmtypes("sgmnt_data",371,"off")=3464
	Set gtmtypes("sgmnt_data",371,"len")=4
	Set gtmtypes("sgmnt_data",371,"type")="int"
	Set gtmtypfldindx("sgmnt_data","jnl_blocked_writer_lost_cntr")=371
	Set gtmtypes("sgmnt_data",372,"name")="sgmnt_data.jnl_blocked_writer_stuck_cntr"
	Set gtmtypes("sgmnt_data",372,"off")=3468
	Set gtmtypes("sgmnt_data",372,"len")=4
	Set gtmtypes("sgmnt_data",372,"type")="int"
	Set gtmtypfldindx("sgmnt_data","jnl_blocked_writer_stuck_cntr")=372
	Set gtmtypes("sgmnt_data",373,"name")="sgmnt_data.jnl_blocked_writer_blocked_cntr"
	Set gtmtypes("sgmnt_data",373,"off")=3472
	Set gtmtypes("sgmnt_data",373,"len")=4
	Set gtmtypes("sgmnt_data",373,"type")="int"
	Set gtmtypfldindx("sgmnt_data","jnl_blocked_writer_blocked_cntr")=373
	Set gtmtypes("sgmnt_data",374,"name")="sgmnt_data.n_jnl_fsyncs_cntr"
	Set gtmtypes("sgmnt_data",374,"off")=3476
	Set gtmtypes("sgmnt_data",374,"len")=4
	Set gtmtypes("sgmnt_data",374,"type")="int"
	Set gtmtypfldindx("sgmnt_data","n_jnl_fsyncs_cntr")=374
	Set gtmtypes("sgmnt_data",375,"name")="sgmnt_data.n_jnl_fsync_tries_cntr"
	Set gtmtypes("sgmnt_data",375,"off")=3480
	Set gtmtypes("sgmnt_data",375,"len")=4
	Set gtmtypes("sgmnt_data",375,"type")="int"
	Set gtmtypfldindx("sgmnt_data","n_jnl_fsync_tries_cntr")=375
	Set gtmtypes("sgmnt_data",376,"name")="sgmnt_data.n_jnl_fsync_recovers_cntr"
	Set gtmtypes("sgmnt_data",376,"off")=3484
	Set gtmtypes("sgmnt_data",376,"len")=4
	Set gtmtypes("sgmnt_data",376,"type")="int"
	Set gtmtypfldindx("sgmnt_data","n_jnl_fsync_recovers_cntr")=376
	Set gtmtypes("sgmnt_data",377,"name")="sgmnt_data.n_db_fsyncs_cntr"
	Set gtmtypes("sgmnt_data",377,"off")=3488
	Set gtmtypes("sgmnt_data",377,"len")=4
	Set gtmtypes("sgmnt_data",377,"type")="int"
	Set gtmtypfldindx("sgmnt_data","n_db_fsyncs_cntr")=377
	Set gtmtypes("sgmnt_data",378,"name")="sgmnt_data.n_db_fsyncs_in_crit_cntr"
	Set gtmtypes("sgmnt_data",378,"off")=3492
	Set gtmtypes("sgmnt_data",378,"len")=4
	Set gtmtypes("sgmnt_data",378,"type")="int"
	Set gtmtypfldindx("sgmnt_data","n_db_fsyncs_in_crit_cntr")=378
	Set gtmtypes("sgmnt_data",379,"name")="sgmnt_data.n_dbsync_timers_cntr"
	Set gtmtypes("sgmnt_data",379,"off")=3496
	Set gtmtypes("sgmnt_data",379,"len")=4
	Set gtmtypes("sgmnt_data",379,"type")="int"
	Set gtmtypfldindx("sgmnt_data","n_dbsync_timers_cntr")=379
	Set gtmtypes("sgmnt_data",380,"name")="sgmnt_data.n_dbsync_writes_cntr"
	Set gtmtypes("sgmnt_data",380,"off")=3500
	Set gtmtypes("sgmnt_data",380,"len")=4
	Set gtmtypes("sgmnt_data",380,"type")="int"
	Set gtmtypfldindx("sgmnt_data","n_dbsync_writes_cntr")=380
	Set gtmtypes("sgmnt_data",381,"name")="sgmnt_data.mutex_queue_full_cntr"
	Set gtmtypes("sgmnt_data",381,"off")=3504
	Set gtmtypes("sgmnt_data",381,"len")=4
	Set gtmtypes("sgmnt_data",381,"type")="int"
	Set gtmtypfldindx("sgmnt_data","mutex_queue_full_cntr")=381
	Set gtmtypes("sgmnt_data",382,"name")="sgmnt_data.wcb_bt_put_cntr"
	Set gtmtypes("sgmnt_data",382,"off")=3508
	Set gtmtypes("sgmnt_data",382,"len")=4
	Set gtmtypes("sgmnt_data",382,"type")="int"
	Set gtmtypfldindx("sgmnt_data","wcb_bt_put_cntr")=382
	Set gtmtypes("sgmnt_data",383,"name")="sgmnt_data.wcb_mutex_salvage_cntr"
	Set gtmtypes("sgmnt_data",383,"off")=3512
	Set gtmtypes("sgmnt_data",383,"len")=4
	Set gtmtypes("sgmnt_data",383,"type")="int"
	Set gtmtypfldindx("sgmnt_data","wcb_mutex_salvage_cntr")=383
	Set gtmtypes("sgmnt_data",384,"name")="sgmnt_data.wcb_wcs_flu0_cntr"
	Set gtmtypes("sgmnt_data",384,"off")=3516
	Set gtmtypes("sgmnt_data",384,"len")=4
	Set gtmtypes("sgmnt_data",384,"type")="int"
	Set gtmtypfldindx("sgmnt_data","wcb_wcs_flu0_cntr")=384
	Set gtmtypes("sgmnt_data",385,"name")="sgmnt_data.wcb_t_end_sysops_nocr_invcr_cntr"
	Set gtmtypes("sgmnt_data",385,"off")=3520
	Set gtmtypes("sgmnt_data",385,"len")=4
	Set gtmtypes("sgmnt_data",385,"type")="int"
	Set gtmtypfldindx("sgmnt_data","wcb_t_end_sysops_nocr_invcr_cntr")=385
	Set gtmtypes("sgmnt_data",386,"name")="sgmnt_data.wcb_t_end_sysops_cr_invcr_cntr"
	Set gtmtypes("sgmnt_data",386,"off")=3524
	Set gtmtypes("sgmnt_data",386,"len")=4
	Set gtmtypes("sgmnt_data",386,"type")="int"
	Set gtmtypfldindx("sgmnt_data","wcb_t_end_sysops_cr_invcr_cntr")=386
	Set gtmtypes("sgmnt_data",387,"name")="sgmnt_data.wcb_t_end_sysops_rip_wait_cntr"
	Set gtmtypes("sgmnt_data",387,"off")=3528
	Set gtmtypes("sgmnt_data",387,"len")=4
	Set gtmtypes("sgmnt_data",387,"type")="int"
	Set gtmtypfldindx("sgmnt_data","wcb_t_end_sysops_rip_wait_cntr")=387
	Set gtmtypes("sgmnt_data",388,"name")="sgmnt_data.wcb_t_end_sysops_dirtyripwait_cntr"
	Set gtmtypes("sgmnt_data",388,"off")=3532
	Set gtmtypes("sgmnt_data",388,"len")=4
	Set gtmtypes("sgmnt_data",388,"type")="int"
	Set gtmtypfldindx("sgmnt_data","wcb_t_end_sysops_dirtyripwait_cntr")=388
	Set gtmtypes("sgmnt_data",389,"name")="sgmnt_data.wcb_gds_rundown1_cntr"
	Set gtmtypes("sgmnt_data",389,"off")=3536
	Set gtmtypes("sgmnt_data",389,"len")=4
	Set gtmtypes("sgmnt_data",389,"type")="int"
	Set gtmtypfldindx("sgmnt_data","wcb_gds_rundown1_cntr")=389
	Set gtmtypes("sgmnt_data",390,"name")="sgmnt_data.wcb_wcs_flu1_cntr"
	Set gtmtypes("sgmnt_data",390,"off")=3540
	Set gtmtypes("sgmnt_data",390,"len")=4
	Set gtmtypes("sgmnt_data",390,"type")="int"
	Set gtmtypfldindx("sgmnt_data","wcb_wcs_flu1_cntr")=390
	Set gtmtypes("sgmnt_data",391,"name")="sgmnt_data.wcb_mu_back_cntr"
	Set gtmtypes("sgmnt_data",391,"off")=3544
	Set gtmtypes("sgmnt_data",391,"len")=4
	Set gtmtypes("sgmnt_data",391,"type")="int"
	Set gtmtypfldindx("sgmnt_data","wcb_mu_back_cntr")=391
	Set gtmtypes("sgmnt_data",392,"name")="sgmnt_data.wcb_t_end_sysops_dirty_invcr_cntr"
	Set gtmtypes("sgmnt_data",392,"off")=3548
	Set gtmtypes("sgmnt_data",392,"len")=4
	Set gtmtypes("sgmnt_data",392,"type")="int"
	Set gtmtypfldindx("sgmnt_data","wcb_t_end_sysops_dirty_invcr_cntr")=392
	Set gtmtypes("sgmnt_data",393,"name")="sgmnt_data.wcb_t_end_sysops_wtfini_fail_cntr"
	Set gtmtypes("sgmnt_data",393,"off")=3552
	Set gtmtypes("sgmnt_data",393,"len")=4
	Set gtmtypes("sgmnt_data",393,"type")="int"
	Set gtmtypfldindx("sgmnt_data","wcb_t_end_sysops_wtfini_fail_cntr")=393
	Set gtmtypes("sgmnt_data",394,"name")="sgmnt_data.wcb_t_end_sysops_twin_stuck_cntr"
	Set gtmtypes("sgmnt_data",394,"off")=3556
	Set gtmtypes("sgmnt_data",394,"len")=4
	Set gtmtypes("sgmnt_data",394,"type")="int"
	Set gtmtypfldindx("sgmnt_data","wcb_t_end_sysops_twin_stuck_cntr")=394
	Set gtmtypes("sgmnt_data",395,"name")="sgmnt_data.db_csh_getn_wrt_latch_stuck_cntr"
	Set gtmtypes("sgmnt_data",395,"off")=3560
	Set gtmtypes("sgmnt_data",395,"len")=4
	Set gtmtypes("sgmnt_data",395,"type")="int"
	Set gtmtypfldindx("sgmnt_data","db_csh_getn_wrt_latch_stuck_cntr")=395
	Set gtmtypes("sgmnt_data",396,"name")="sgmnt_data.wcb_secshr_db_clnup_now_crit_cntr"
	Set gtmtypes("sgmnt_data",396,"off")=3564
	Set gtmtypes("sgmnt_data",396,"len")=4
	Set gtmtypes("sgmnt_data",396,"type")="int"
	Set gtmtypfldindx("sgmnt_data","wcb_secshr_db_clnup_now_crit_cntr")=396
	Set gtmtypes("sgmnt_data",397,"name")="sgmnt_data.wc_blocked_t_qread_bad_bt_index1_cntr"
	Set gtmtypes("sgmnt_data",397,"off")=3568
	Set gtmtypes("sgmnt_data",397,"len")=4
	Set gtmtypes("sgmnt_data",397,"type")="int"
	Set gtmtypfldindx("sgmnt_data","wc_blocked_t_qread_bad_bt_index1_cntr")=397
	Set gtmtypes("sgmnt_data",398,"name")="sgmnt_data.wc_blocked_t_qread_bad_bt_index2_cntr"
	Set gtmtypes("sgmnt_data",398,"off")=3572
	Set gtmtypes("sgmnt_data",398,"len")=4
	Set gtmtypes("sgmnt_data",398,"type")="int"
	Set gtmtypfldindx("sgmnt_data","wc_blocked_t_qread_bad_bt_index2_cntr")=398
	Set gtmtypes("sgmnt_data",399,"name")="sgmnt_data.wc_blocked_t_end_crbtmismatch1_cntr"
	Set gtmtypes("sgmnt_data",399,"off")=3576
	Set gtmtypes("sgmnt_data",399,"len")=4
	Set gtmtypes("sgmnt_data",399,"type")="int"
	Set gtmtypfldindx("sgmnt_data","wc_blocked_t_end_crbtmismatch1_cntr")=399
	Set gtmtypes("sgmnt_data",400,"name")="sgmnt_data.wc_blocked_t_end_crbtmismatch2_cntr"
	Set gtmtypes("sgmnt_data",400,"off")=3580
	Set gtmtypes("sgmnt_data",400,"len")=4
	Set gtmtypes("sgmnt_data",400,"type")="int"
	Set gtmtypfldindx("sgmnt_data","wc_blocked_t_end_crbtmismatch2_cntr")=400
	Set gtmtypes("sgmnt_data",401,"name")="sgmnt_data.wc_blocked_t_end_crbtmismatch3_cntr"
	Set gtmtypes("sgmnt_data",401,"off")=3584
	Set gtmtypes("sgmnt_data",401,"len")=4
	Set gtmtypes("sgmnt_data",401,"type")="int"
	Set gtmtypfldindx("sgmnt_data","wc_blocked_t_end_crbtmismatch3_cntr")=401
	Set gtmtypes("sgmnt_data",402,"name")="sgmnt_data.wc_blocked_tp_tend_crbtmismatch1_cntr"
	Set gtmtypes("sgmnt_data",402,"off")=3588
	Set gtmtypes("sgmnt_data",402,"len")=4
	Set gtmtypes("sgmnt_data",402,"type")="int"
	Set gtmtypfldindx("sgmnt_data","wc_blocked_tp_tend_crbtmismatch1_cntr")=402
	Set gtmtypes("sgmnt_data",403,"name")="sgmnt_data.wc_blocked_tp_tend_crbtmismatch2_cntr"
	Set gtmtypes("sgmnt_data",403,"off")=3592
	Set gtmtypes("sgmnt_data",403,"len")=4
	Set gtmtypes("sgmnt_data",403,"type")="int"
	Set gtmtypfldindx("sgmnt_data","wc_blocked_tp_tend_crbtmismatch2_cntr")=403
	Set gtmtypes("sgmnt_data",404,"name")="sgmnt_data.wc_blocked_tp_tend_crbtmismatch3_cntr"
	Set gtmtypes("sgmnt_data",404,"off")=3596
	Set gtmtypes("sgmnt_data",404,"len")=4
	Set gtmtypes("sgmnt_data",404,"type")="int"
	Set gtmtypfldindx("sgmnt_data","wc_blocked_tp_tend_crbtmismatch3_cntr")=404
	Set gtmtypes("sgmnt_data",405,"name")="sgmnt_data.wc_blocked_wcs_wtstart_bad_cr_cntr"
	Set gtmtypes("sgmnt_data",405,"off")=3600
	Set gtmtypes("sgmnt_data",405,"len")=4
	Set gtmtypes("sgmnt_data",405,"type")="int"
	Set gtmtypfldindx("sgmnt_data","wc_blocked_wcs_wtstart_bad_cr_cntr")=405
	Set gtmtypes("sgmnt_data",406,"name")="sgmnt_data.wc_blocked_wcs_wtfini_bad_cr_cntr"
	Set gtmtypes("sgmnt_data",406,"off")=3604
	Set gtmtypes("sgmnt_data",406,"len")=4
	Set gtmtypes("sgmnt_data",406,"type")="int"
	Set gtmtypfldindx("sgmnt_data","wc_blocked_wcs_wtfini_bad_cr_cntr")=406
	Set gtmtypes("sgmnt_data",407,"name")="sgmnt_data.wc_blocked_bt_get_cntr"
	Set gtmtypes("sgmnt_data",407,"off")=3608
	Set gtmtypes("sgmnt_data",407,"len")=4
	Set gtmtypes("sgmnt_data",407,"type")="int"
	Set gtmtypfldindx("sgmnt_data","wc_blocked_bt_get_cntr")=407
	Set gtmtypes("sgmnt_data",408,"name")="sgmnt_data.wc_blocked_wcs_cdb_sc_final_retry_cntr"
	Set gtmtypes("sgmnt_data",408,"off")=3612
	Set gtmtypes("sgmnt_data",408,"len")=4
	Set gtmtypes("sgmnt_data",408,"type")="int"
	Set gtmtypfldindx("sgmnt_data","wc_blocked_wcs_cdb_sc_final_retry_cntr")=408
	Set gtmtypes("sgmnt_data",409,"name")="sgmnt_data.wcb_bg_update_lckfail1_cntr"
	Set gtmtypes("sgmnt_data",409,"off")=3616
	Set gtmtypes("sgmnt_data",409,"len")=4
	Set gtmtypes("sgmnt_data",409,"type")="int"
	Set gtmtypfldindx("sgmnt_data","wcb_bg_update_lckfail1_cntr")=409
	Set gtmtypes("sgmnt_data",410,"name")="sgmnt_data.wcb_bg_update_lckfail2_cntr"
	Set gtmtypes("sgmnt_data",410,"off")=3620
	Set gtmtypes("sgmnt_data",410,"len")=4
	Set gtmtypes("sgmnt_data",410,"type")="int"
	Set gtmtypfldindx("sgmnt_data","wcb_bg_update_lckfail2_cntr")=410
	Set gtmtypes("sgmnt_data",411,"name")="sgmnt_data.wcb_wtstart_lckfail1_cntr"
	Set gtmtypes("sgmnt_data",411,"off")=3624
	Set gtmtypes("sgmnt_data",411,"len")=4
	Set gtmtypes("sgmnt_data",411,"type")="int"
	Set gtmtypfldindx("sgmnt_data","wcb_wtstart_lckfail1_cntr")=411
	Set gtmtypes("sgmnt_data",412,"name")="sgmnt_data.wcb_wtstart_lckfail2_cntr"
	Set gtmtypes("sgmnt_data",412,"off")=3628
	Set gtmtypes("sgmnt_data",412,"len")=4
	Set gtmtypes("sgmnt_data",412,"type")="int"
	Set gtmtypfldindx("sgmnt_data","wcb_wtstart_lckfail2_cntr")=412
	Set gtmtypes("sgmnt_data",413,"name")="sgmnt_data.wcb_wtstart_lckfail3_cntr"
	Set gtmtypes("sgmnt_data",413,"off")=3632
	Set gtmtypes("sgmnt_data",413,"len")=4
	Set gtmtypes("sgmnt_data",413,"type")="int"
	Set gtmtypfldindx("sgmnt_data","wcb_wtstart_lckfail3_cntr")=413
	Set gtmtypes("sgmnt_data",414,"name")="sgmnt_data.wcb_wtstart_lckfail4_cntr"
	Set gtmtypes("sgmnt_data",414,"off")=3636
	Set gtmtypes("sgmnt_data",414,"len")=4
	Set gtmtypes("sgmnt_data",414,"type")="int"
	Set gtmtypfldindx("sgmnt_data","wcb_wtstart_lckfail4_cntr")=414
	Set gtmtypes("sgmnt_data",415,"name")="sgmnt_data.wcb_wtfini_lckfail1_cntr"
	Set gtmtypes("sgmnt_data",415,"off")=3640
	Set gtmtypes("sgmnt_data",415,"len")=4
	Set gtmtypes("sgmnt_data",415,"type")="int"
	Set gtmtypfldindx("sgmnt_data","wcb_wtfini_lckfail1_cntr")=415
	Set gtmtypes("sgmnt_data",416,"name")="sgmnt_data.wcb_wtfini_lckfail2_cntr"
	Set gtmtypes("sgmnt_data",416,"off")=3644
	Set gtmtypes("sgmnt_data",416,"len")=4
	Set gtmtypes("sgmnt_data",416,"type")="int"
	Set gtmtypfldindx("sgmnt_data","wcb_wtfini_lckfail2_cntr")=416
	Set gtmtypes("sgmnt_data",417,"name")="sgmnt_data.wcb_wtfini_lckfail3_cntr"
	Set gtmtypes("sgmnt_data",417,"off")=3648
	Set gtmtypes("sgmnt_data",417,"len")=4
	Set gtmtypes("sgmnt_data",417,"type")="int"
	Set gtmtypfldindx("sgmnt_data","wcb_wtfini_lckfail3_cntr")=417
	Set gtmtypes("sgmnt_data",418,"name")="sgmnt_data.wcb_wtfini_lckfail4_cntr"
	Set gtmtypes("sgmnt_data",418,"off")=3652
	Set gtmtypes("sgmnt_data",418,"len")=4
	Set gtmtypes("sgmnt_data",418,"type")="int"
	Set gtmtypfldindx("sgmnt_data","wcb_wtfini_lckfail4_cntr")=418
	Set gtmtypes("sgmnt_data",419,"name")="sgmnt_data.wcb_t_end_sysops_dirtystuck1_cntr"
	Set gtmtypes("sgmnt_data",419,"off")=3656
	Set gtmtypes("sgmnt_data",419,"len")=4
	Set gtmtypes("sgmnt_data",419,"type")="int"
	Set gtmtypfldindx("sgmnt_data","wcb_t_end_sysops_dirtystuck1_cntr")=419
	Set gtmtypes("sgmnt_data",420,"name")="sgmnt_data.wcb_t_end_sysops_dirtystuck2_cntr"
	Set gtmtypes("sgmnt_data",420,"off")=3660
	Set gtmtypes("sgmnt_data",420,"len")=4
	Set gtmtypes("sgmnt_data",420,"type")="int"
	Set gtmtypfldindx("sgmnt_data","wcb_t_end_sysops_dirtystuck2_cntr")=420
	Set gtmtypes("sgmnt_data",421,"name")="sgmnt_data.wcb_secshr_db_clnup_wbuf_dqd_cntr"
	Set gtmtypes("sgmnt_data",421,"off")=3664
	Set gtmtypes("sgmnt_data",421,"len")=4
	Set gtmtypes("sgmnt_data",421,"type")="int"
	Set gtmtypfldindx("sgmnt_data","wcb_secshr_db_clnup_wbuf_dqd_cntr")=421
	Set gtmtypes("sgmnt_data",422,"name")="sgmnt_data.dwngrd_refmts_syncio_cntr"
	Set gtmtypes("sgmnt_data",422,"off")=3668
	Set gtmtypes("sgmnt_data",422,"len")=4
	Set gtmtypes("sgmnt_data",422,"type")="int"
	Set gtmtypfldindx("sgmnt_data","dwngrd_refmts_syncio_cntr")=422
	Set gtmtypes("sgmnt_data",423,"name")="sgmnt_data.dwngrd_refmts_asyncio_cntr"
	Set gtmtypes("sgmnt_data",423,"off")=3672
	Set gtmtypes("sgmnt_data",423,"len")=4
	Set gtmtypes("sgmnt_data",423,"type")="int"
	Set gtmtypfldindx("sgmnt_data","dwngrd_refmts_asyncio_cntr")=423
	Set gtmtypes("sgmnt_data",424,"name")="sgmnt_data.shmpool_refmt_harvests_cntr"
	Set gtmtypes("sgmnt_data",424,"off")=3676
	Set gtmtypes("sgmnt_data",424,"len")=4
	Set gtmtypes("sgmnt_data",424,"type")="int"
	Set gtmtypfldindx("sgmnt_data","shmpool_refmt_harvests_cntr")=424
	Set gtmtypes("sgmnt_data",425,"name")="sgmnt_data.shmpool_recovery_cntr"
	Set gtmtypes("sgmnt_data",425,"off")=3680
	Set gtmtypes("sgmnt_data",425,"len")=4
	Set gtmtypes("sgmnt_data",425,"type")="int"
	Set gtmtypfldindx("sgmnt_data","shmpool_recovery_cntr")=425
	Set gtmtypes("sgmnt_data",426,"name")="sgmnt_data.shmpool_blkd_by_sdc_cntr"
	Set gtmtypes("sgmnt_data",426,"off")=3684
	Set gtmtypes("sgmnt_data",426,"len")=4
	Set gtmtypes("sgmnt_data",426,"type")="int"
	Set gtmtypfldindx("sgmnt_data","shmpool_blkd_by_sdc_cntr")=426
	Set gtmtypes("sgmnt_data",427,"name")="sgmnt_data.shmpool_alloc_bbflush_cntr"
	Set gtmtypes("sgmnt_data",427,"off")=3688
	Set gtmtypes("sgmnt_data",427,"len")=4
	Set gtmtypes("sgmnt_data",427,"type")="int"
	Set gtmtypfldindx("sgmnt_data","shmpool_alloc_bbflush_cntr")=427
	Set gtmtypes("sgmnt_data",428,"name")="sgmnt_data.refmt_hvst_blk_released_replaced_cntr"
	Set gtmtypes("sgmnt_data",428,"off")=3692
	Set gtmtypes("sgmnt_data",428,"len")=4
	Set gtmtypes("sgmnt_data",428,"type")="int"
	Set gtmtypfldindx("sgmnt_data","refmt_hvst_blk_released_replaced_cntr")=428
	Set gtmtypes("sgmnt_data",429,"name")="sgmnt_data.refmt_hvst_blk_released_io_complete_cntr"
	Set gtmtypes("sgmnt_data",429,"off")=3696
	Set gtmtypes("sgmnt_data",429,"len")=4
	Set gtmtypes("sgmnt_data",429,"type")="int"
	Set gtmtypfldindx("sgmnt_data","refmt_hvst_blk_released_io_complete_cntr")=429
	Set gtmtypes("sgmnt_data",430,"name")="sgmnt_data.refmt_hvst_blk_kept_cntr"
	Set gtmtypes("sgmnt_data",430,"off")=3700
	Set gtmtypes("sgmnt_data",430,"len")=4
	Set gtmtypes("sgmnt_data",430,"type")="int"
	Set gtmtypfldindx("sgmnt_data","refmt_hvst_blk_kept_cntr")=430
	Set gtmtypes("sgmnt_data",431,"name")="sgmnt_data.refmt_hvst_blk_ignored_cntr"
	Set gtmtypes("sgmnt_data",431,"off")=3704
	Set gtmtypes("sgmnt_data",431,"len")=4
	Set gtmtypes("sgmnt_data",431,"type")="int"
	Set gtmtypfldindx("sgmnt_data","refmt_hvst_blk_ignored_cntr")=431
	Set gtmtypes("sgmnt_data",432,"name")="sgmnt_data.refmt_blk_chk_blk_freed_cntr"
	Set gtmtypes("sgmnt_data",432,"off")=3708
	Set gtmtypes("sgmnt_data",432,"len")=4
	Set gtmtypes("sgmnt_data",432,"type")="int"
	Set gtmtypfldindx("sgmnt_data","refmt_blk_chk_blk_freed_cntr")=432
	Set gtmtypes("sgmnt_data",433,"name")="sgmnt_data.refmt_blk_chk_blk_kept_cntr"
	Set gtmtypes("sgmnt_data",433,"off")=3712
	Set gtmtypes("sgmnt_data",433,"len")=4
	Set gtmtypes("sgmnt_data",433,"type")="int"
	Set gtmtypfldindx("sgmnt_data","refmt_blk_chk_blk_kept_cntr")=433
	Set gtmtypes("sgmnt_data",434,"name")="sgmnt_data.active_lvl_trigger_cntr"
	Set gtmtypes("sgmnt_data",434,"off")=3716
	Set gtmtypes("sgmnt_data",434,"len")=4
	Set gtmtypes("sgmnt_data",434,"type")="int"
	Set gtmtypfldindx("sgmnt_data","active_lvl_trigger_cntr")=434
	Set gtmtypes("sgmnt_data",435,"name")="sgmnt_data.new_buff_cntr"
	Set gtmtypes("sgmnt_data",435,"off")=3720
	Set gtmtypes("sgmnt_data",435,"len")=4
	Set gtmtypes("sgmnt_data",435,"type")="int"
	Set gtmtypfldindx("sgmnt_data","new_buff_cntr")=435
	Set gtmtypes("sgmnt_data",436,"name")="sgmnt_data.phase2_commit_wait_sleep_in_crit_cntr"
	Set gtmtypes("sgmnt_data",436,"off")=3724
	Set gtmtypes("sgmnt_data",436,"len")=4
	Set gtmtypes("sgmnt_data",436,"type")="int"
	Set gtmtypfldindx("sgmnt_data","phase2_commit_wait_sleep_in_crit_cntr")=436
	Set gtmtypes("sgmnt_data",437,"name")="sgmnt_data.phase2_commit_wait_sleep_no_crit_cntr"
	Set gtmtypes("sgmnt_data",437,"off")=3728
	Set gtmtypes("sgmnt_data",437,"len")=4
	Set gtmtypes("sgmnt_data",437,"type")="int"
	Set gtmtypfldindx("sgmnt_data","phase2_commit_wait_sleep_no_crit_cntr")=437
	Set gtmtypes("sgmnt_data",438,"name")="sgmnt_data.phase2_commit_wait_pidcnt_cntr"
	Set gtmtypes("sgmnt_data",438,"off")=3732
	Set gtmtypes("sgmnt_data",438,"len")=4
	Set gtmtypes("sgmnt_data",438,"type")="int"
	Set gtmtypfldindx("sgmnt_data","phase2_commit_wait_pidcnt_cntr")=438
	Set gtmtypes("sgmnt_data",439,"name")="sgmnt_data.wcb_t_end_sysops_intend_wait_cntr"
	Set gtmtypes("sgmnt_data",439,"off")=3736
	Set gtmtypes("sgmnt_data",439,"len")=4
	Set gtmtypes("sgmnt_data",439,"type")="int"
	Set gtmtypfldindx("sgmnt_data","wcb_t_end_sysops_intend_wait_cntr")=439
	Set gtmtypes("sgmnt_data",440,"name")="sgmnt_data.wcb_secshr_db_clnup_phase2_clnup_cntr"
	Set gtmtypes("sgmnt_data",440,"off")=3740
	Set gtmtypes("sgmnt_data",440,"len")=4
	Set gtmtypes("sgmnt_data",440,"type")="int"
	Set gtmtypfldindx("sgmnt_data","wcb_secshr_db_clnup_phase2_clnup_cntr")=440
	Set gtmtypes("sgmnt_data",441,"name")="sgmnt_data.wcb_phase2_commit_wait_cntr"
	Set gtmtypes("sgmnt_data",441,"off")=3744
	Set gtmtypes("sgmnt_data",441,"len")=4
	Set gtmtypes("sgmnt_data",441,"type")="int"
	Set gtmtypfldindx("sgmnt_data","wcb_phase2_commit_wait_cntr")=441
	Set gtmtypes("sgmnt_data",442,"name")="sgmnt_data.recompute_upd_array_calls_cntr"
	Set gtmtypes("sgmnt_data",442,"off")=3748
	Set gtmtypes("sgmnt_data",442,"len")=4
	Set gtmtypes("sgmnt_data",442,"type")="int"
	Set gtmtypfldindx("sgmnt_data","recompute_upd_array_calls_cntr")=442
	Set gtmtypes("sgmnt_data",443,"name")="sgmnt_data.recompute_upd_array_rip_cntr"
	Set gtmtypes("sgmnt_data",443,"off")=3752
	Set gtmtypes("sgmnt_data",443,"len")=4
	Set gtmtypes("sgmnt_data",443,"type")="int"
	Set gtmtypfldindx("sgmnt_data","recompute_upd_array_rip_cntr")=443
	Set gtmtypes("sgmnt_data",444,"name")="sgmnt_data.recompute_upd_array_in_tend_cntr"
	Set gtmtypes("sgmnt_data",444,"off")=3756
	Set gtmtypes("sgmnt_data",444,"len")=4
	Set gtmtypes("sgmnt_data",444,"type")="int"
	Set gtmtypfldindx("sgmnt_data","recompute_upd_array_in_tend_cntr")=444
	Set gtmtypes("sgmnt_data",445,"name")="sgmnt_data.recompute_upd_array_search_blk_cntr"
	Set gtmtypes("sgmnt_data",445,"off")=3760
	Set gtmtypes("sgmnt_data",445,"len")=4
	Set gtmtypes("sgmnt_data",445,"type")="int"
	Set gtmtypfldindx("sgmnt_data","recompute_upd_array_search_blk_cntr")=445
	Set gtmtypes("sgmnt_data",446,"name")="sgmnt_data.recompute_upd_array_new_rec_cntr"
	Set gtmtypes("sgmnt_data",446,"off")=3764
	Set gtmtypes("sgmnt_data",446,"len")=4
	Set gtmtypes("sgmnt_data",446,"type")="int"
	Set gtmtypfldindx("sgmnt_data","recompute_upd_array_new_rec_cntr")=446
	Set gtmtypes("sgmnt_data",447,"name")="sgmnt_data.recompute_upd_array_rec_size_cntr"
	Set gtmtypes("sgmnt_data",447,"off")=3768
	Set gtmtypes("sgmnt_data",447,"len")=4
	Set gtmtypes("sgmnt_data",447,"type")="int"
	Set gtmtypfldindx("sgmnt_data","recompute_upd_array_rec_size_cntr")=447
	Set gtmtypes("sgmnt_data",448,"name")="sgmnt_data.recompute_upd_array_rec_cmpc_cntr"
	Set gtmtypes("sgmnt_data",448,"off")=3772
	Set gtmtypes("sgmnt_data",448,"len")=4
	Set gtmtypes("sgmnt_data",448,"type")="int"
	Set gtmtypfldindx("sgmnt_data","recompute_upd_array_rec_cmpc_cntr")=448
	Set gtmtypes("sgmnt_data",449,"name")="sgmnt_data.recompute_upd_array_blk_fini_cntr"
	Set gtmtypes("sgmnt_data",449,"off")=3776
	Set gtmtypes("sgmnt_data",449,"len")=4
	Set gtmtypes("sgmnt_data",449,"type")="int"
	Set gtmtypfldindx("sgmnt_data","recompute_upd_array_blk_fini_cntr")=449
	Set gtmtypes("sgmnt_data",450,"name")="sgmnt_data.recompute_upd_array_blk_split_cntr"
	Set gtmtypes("sgmnt_data",450,"off")=3780
	Set gtmtypes("sgmnt_data",450,"len")=4
	Set gtmtypes("sgmnt_data",450,"type")="int"
	Set gtmtypfldindx("sgmnt_data","recompute_upd_array_blk_split_cntr")=450
	Set gtmtypes("sgmnt_data",451,"name")="sgmnt_data.unused_filler_0135_cntr"
	Set gtmtypes("sgmnt_data",451,"off")=3784
	Set gtmtypes("sgmnt_data",451,"len")=4
	Set gtmtypes("sgmnt_data",451,"type")="int"
	Set gtmtypfldindx("sgmnt_data","unused_filler_0135_cntr")=451
	Set gtmtypes("sgmnt_data",452,"name")="sgmnt_data.t_qread_ripsleep_nblks_cntr"
	Set gtmtypes("sgmnt_data",452,"off")=3788
	Set gtmtypes("sgmnt_data",452,"len")=4
	Set gtmtypes("sgmnt_data",452,"type")="int"
	Set gtmtypfldindx("sgmnt_data","t_qread_ripsleep_nblks_cntr")=452
	Set gtmtypes("sgmnt_data",453,"name")="sgmnt_data.wcs_wtfini_invoked_cntr"
	Set gtmtypes("sgmnt_data",453,"off")=3792
	Set gtmtypes("sgmnt_data",453,"len")=4
	Set gtmtypes("sgmnt_data",453,"type")="int"
	Set gtmtypfldindx("sgmnt_data","wcs_wtfini_invoked_cntr")=453
	Set gtmtypes("sgmnt_data",454,"name")="sgmnt_data.wcs_wterror_invoked_cntr"
	Set gtmtypes("sgmnt_data",454,"off")=3796
	Set gtmtypes("sgmnt_data",454,"len")=4
	Set gtmtypes("sgmnt_data",454,"type")="int"
	Set gtmtypfldindx("sgmnt_data","wcs_wterror_invoked_cntr")=454
	Set gtmtypes("sgmnt_data",455,"name")="sgmnt_data.wcs_wt_restart_invoked_cntr"
	Set gtmtypes("sgmnt_data",455,"off")=3800
	Set gtmtypes("sgmnt_data",455,"len")=4
	Set gtmtypes("sgmnt_data",455,"type")="int"
	Set gtmtypfldindx("sgmnt_data","wcs_wt_restart_invoked_cntr")=455
	Set gtmtypes("sgmnt_data",456,"name")="sgmnt_data.wcs_wt_restart_eagain_cntr"
	Set gtmtypes("sgmnt_data",456,"off")=3804
	Set gtmtypes("sgmnt_data",456,"len")=4
	Set gtmtypes("sgmnt_data",456,"type")="int"
	Set gtmtypfldindx("sgmnt_data","wcs_wt_restart_eagain_cntr")=456
	Set gtmtypes("sgmnt_data",457,"name")="sgmnt_data.wcs_wt_restart_reinsert_cntr"
	Set gtmtypes("sgmnt_data",457,"off")=3808
	Set gtmtypes("sgmnt_data",457,"len")=4
	Set gtmtypes("sgmnt_data",457,"type")="int"
	Set gtmtypfldindx("sgmnt_data","wcs_wt_restart_reinsert_cntr")=457
	Set gtmtypes("sgmnt_data",458,"name")="sgmnt_data.wcs_wtstart_eagain_cntr"
	Set gtmtypes("sgmnt_data",458,"off")=3812
	Set gtmtypes("sgmnt_data",458,"len")=4
	Set gtmtypes("sgmnt_data",458,"type")="int"
	Set gtmtypfldindx("sgmnt_data","wcs_wtstart_eagain_cntr")=458
	Set gtmtypes("sgmnt_data",459,"name")="sgmnt_data.wcs_wtstart_eagain_incrit_cntr"
	Set gtmtypes("sgmnt_data",459,"off")=3816
	Set gtmtypes("sgmnt_data",459,"len")=4
	Set gtmtypes("sgmnt_data",459,"type")="int"
	Set gtmtypfldindx("sgmnt_data","wcs_wtstart_eagain_incrit_cntr")=459
	Set gtmtypes("sgmnt_data",460,"name")="sgmnt_data.async_restart_eagain_cntr"
	Set gtmtypes("sgmnt_data",460,"off")=3820
	Set gtmtypes("sgmnt_data",460,"len")=4
	Set gtmtypes("sgmnt_data",460,"type")="int"
	Set gtmtypfldindx("sgmnt_data","async_restart_eagain_cntr")=460
	Set gtmtypes("sgmnt_data",461,"name")="sgmnt_data.jnlbuff_phs2cmt_array_full_cntr"
	Set gtmtypes("sgmnt_data",461,"off")=3824
	Set gtmtypes("sgmnt_data",461,"len")=4
	Set gtmtypes("sgmnt_data",461,"type")="int"
	Set gtmtypfldindx("sgmnt_data","jnlbuff_phs2cmt_array_full_cntr")=461
	Set gtmtypes("sgmnt_data",462,"name")="sgmnt_data.jnl_phase2_cleanup_if_possible_cntr"
	Set gtmtypes("sgmnt_data",462,"off")=3828
	Set gtmtypes("sgmnt_data",462,"len")=4
	Set gtmtypes("sgmnt_data",462,"type")="int"
	Set gtmtypfldindx("sgmnt_data","jnl_phase2_cleanup_if_possible_cntr")=462
	Set gtmtypes("sgmnt_data",463,"name")="sgmnt_data.jnlbuff_phs2cmt_isprcalv_cntr"
	Set gtmtypes("sgmnt_data",463,"off")=3832
	Set gtmtypes("sgmnt_data",463,"len")=4
	Set gtmtypes("sgmnt_data",463,"type")="int"
	Set gtmtypfldindx("sgmnt_data","jnlbuff_phs2cmt_isprcalv_cntr")=463
	Set gtmtypes("sgmnt_data",464,"name")="sgmnt_data.recompute_upd_array_paidoff_cntr"
	Set gtmtypes("sgmnt_data",464,"off")=3836
	Set gtmtypes("sgmnt_data",464,"len")=4
	Set gtmtypes("sgmnt_data",464,"type")="int"
	Set gtmtypfldindx("sgmnt_data","recompute_upd_array_paidoff_cntr")=464
	Set gtmtypes("sgmnt_data",465,"name")="sgmnt_data.bg_trc_rec_cntr_filler"
	Set gtmtypes("sgmnt_data",465,"off")=3840
	Set gtmtypes("sgmnt_data",465,"len")=8
	Set gtmtypes("sgmnt_data",465,"type")="char"
	Set gtmtypfldindx("sgmnt_data","bg_trc_rec_cntr_filler")=465
	Set gtmtypes("sgmnt_data",466,"name")="sgmnt_data.n_gvcst_srches"
	Set gtmtypes("sgmnt_data",466,"off")=3848
	Set gtmtypes("sgmnt_data",466,"len")=8
	Set gtmtypes("sgmnt_data",466,"type")="db_csh_acct_rec"
	Set gtmtypfldindx("sgmnt_data","n_gvcst_srches")=466
	Set gtmtypes("sgmnt_data",467,"name")="sgmnt_data.n_gvcst_srches.curr_count"
	Set gtmtypes("sgmnt_data",467,"off")=3848
	Set gtmtypes("sgmnt_data",467,"len")=4
	Set gtmtypes("sgmnt_data",467,"type")="int"
	Set gtmtypfldindx("sgmnt_data","n_gvcst_srches.curr_count")=467
	Set gtmtypes("sgmnt_data",468,"name")="sgmnt_data.n_gvcst_srches.cumul_count"
	Set gtmtypes("sgmnt_data",468,"off")=3852
	Set gtmtypes("sgmnt_data",468,"len")=4
	Set gtmtypes("sgmnt_data",468,"type")="int"
	Set gtmtypfldindx("sgmnt_data","n_gvcst_srches.cumul_count")=468
	Set gtmtypes("sgmnt_data",469,"name")="sgmnt_data.n_gvcst_srch_clues"
	Set gtmtypes("sgmnt_data",469,"off")=3856
	Set gtmtypes("sgmnt_data",469,"len")=8
	Set gtmtypes("sgmnt_data",469,"type")="db_csh_acct_rec"
	Set gtmtypfldindx("sgmnt_data","n_gvcst_srch_clues")=469
	Set gtmtypes("sgmnt_data",470,"name")="sgmnt_data.n_gvcst_srch_clues.curr_count"
	Set gtmtypes("sgmnt_data",470,"off")=3856
	Set gtmtypes("sgmnt_data",470,"len")=4
	Set gtmtypes("sgmnt_data",470,"type")="int"
	Set gtmtypfldindx("sgmnt_data","n_gvcst_srch_clues.curr_count")=470
	Set gtmtypes("sgmnt_data",471,"name")="sgmnt_data.n_gvcst_srch_clues.cumul_count"
	Set gtmtypes("sgmnt_data",471,"off")=3860
	Set gtmtypes("sgmnt_data",471,"len")=4
	Set gtmtypes("sgmnt_data",471,"type")="int"
	Set gtmtypfldindx("sgmnt_data","n_gvcst_srch_clues.cumul_count")=471
	Set gtmtypes("sgmnt_data",472,"name")="sgmnt_data.n_clue_used_head"
	Set gtmtypes("sgmnt_data",472,"off")=3864
	Set gtmtypes("sgmnt_data",472,"len")=8
	Set gtmtypes("sgmnt_data",472,"type")="db_csh_acct_rec"
	Set gtmtypfldindx("sgmnt_data","n_clue_used_head")=472
	Set gtmtypes("sgmnt_data",473,"name")="sgmnt_data.n_clue_used_head.curr_count"
	Set gtmtypes("sgmnt_data",473,"off")=3864
	Set gtmtypes("sgmnt_data",473,"len")=4
	Set gtmtypes("sgmnt_data",473,"type")="int"
	Set gtmtypfldindx("sgmnt_data","n_clue_used_head.curr_count")=473
	Set gtmtypes("sgmnt_data",474,"name")="sgmnt_data.n_clue_used_head.cumul_count"
	Set gtmtypes("sgmnt_data",474,"off")=3868
	Set gtmtypes("sgmnt_data",474,"len")=4
	Set gtmtypes("sgmnt_data",474,"type")="int"
	Set gtmtypfldindx("sgmnt_data","n_clue_used_head.cumul_count")=474
	Set gtmtypes("sgmnt_data",475,"name")="sgmnt_data.n_clue_used_same"
	Set gtmtypes("sgmnt_data",475,"off")=3872
	Set gtmtypes("sgmnt_data",475,"len")=8
	Set gtmtypes("sgmnt_data",475,"type")="db_csh_acct_rec"
	Set gtmtypfldindx("sgmnt_data","n_clue_used_same")=475
	Set gtmtypes("sgmnt_data",476,"name")="sgmnt_data.n_clue_used_same.curr_count"
	Set gtmtypes("sgmnt_data",476,"off")=3872
	Set gtmtypes("sgmnt_data",476,"len")=4
	Set gtmtypes("sgmnt_data",476,"type")="int"
	Set gtmtypfldindx("sgmnt_data","n_clue_used_same.curr_count")=476
	Set gtmtypes("sgmnt_data",477,"name")="sgmnt_data.n_clue_used_same.cumul_count"
	Set gtmtypes("sgmnt_data",477,"off")=3876
	Set gtmtypes("sgmnt_data",477,"len")=4
	Set gtmtypes("sgmnt_data",477,"type")="int"
	Set gtmtypfldindx("sgmnt_data","n_clue_used_same.cumul_count")=477
	Set gtmtypes("sgmnt_data",478,"name")="sgmnt_data.n_clue_used_tail"
	Set gtmtypes("sgmnt_data",478,"off")=3880
	Set gtmtypes("sgmnt_data",478,"len")=8
	Set gtmtypes("sgmnt_data",478,"type")="db_csh_acct_rec"
	Set gtmtypfldindx("sgmnt_data","n_clue_used_tail")=478
	Set gtmtypes("sgmnt_data",479,"name")="sgmnt_data.n_clue_used_tail.curr_count"
	Set gtmtypes("sgmnt_data",479,"off")=3880
	Set gtmtypes("sgmnt_data",479,"len")=4
	Set gtmtypes("sgmnt_data",479,"type")="int"
	Set gtmtypfldindx("sgmnt_data","n_clue_used_tail.curr_count")=479
	Set gtmtypes("sgmnt_data",480,"name")="sgmnt_data.n_clue_used_tail.cumul_count"
	Set gtmtypes("sgmnt_data",480,"off")=3884
	Set gtmtypes("sgmnt_data",480,"len")=4
	Set gtmtypes("sgmnt_data",480,"type")="int"
	Set gtmtypfldindx("sgmnt_data","n_clue_used_tail.cumul_count")=480
	Set gtmtypes("sgmnt_data",481,"name")="sgmnt_data.n_t_qreads"
	Set gtmtypes("sgmnt_data",481,"off")=3888
	Set gtmtypes("sgmnt_data",481,"len")=8
	Set gtmtypes("sgmnt_data",481,"type")="db_csh_acct_rec"
	Set gtmtypfldindx("sgmnt_data","n_t_qreads")=481
	Set gtmtypes("sgmnt_data",482,"name")="sgmnt_data.n_t_qreads.curr_count"
	Set gtmtypes("sgmnt_data",482,"off")=3888
	Set gtmtypes("sgmnt_data",482,"len")=4
	Set gtmtypes("sgmnt_data",482,"type")="int"
	Set gtmtypfldindx("sgmnt_data","n_t_qreads.curr_count")=482
	Set gtmtypes("sgmnt_data",483,"name")="sgmnt_data.n_t_qreads.cumul_count"
	Set gtmtypes("sgmnt_data",483,"off")=3892
	Set gtmtypes("sgmnt_data",483,"len")=4
	Set gtmtypes("sgmnt_data",483,"type")="int"
	Set gtmtypfldindx("sgmnt_data","n_t_qreads.cumul_count")=483
	Set gtmtypes("sgmnt_data",484,"name")="sgmnt_data.unused_dsk_reads"
	Set gtmtypes("sgmnt_data",484,"off")=3896
	Set gtmtypes("sgmnt_data",484,"len")=8
	Set gtmtypes("sgmnt_data",484,"type")="db_csh_acct_rec"
	Set gtmtypfldindx("sgmnt_data","unused_dsk_reads")=484
	Set gtmtypes("sgmnt_data",485,"name")="sgmnt_data.unused_dsk_reads.curr_count"
	Set gtmtypes("sgmnt_data",485,"off")=3896
	Set gtmtypes("sgmnt_data",485,"len")=4
	Set gtmtypes("sgmnt_data",485,"type")="int"
	Set gtmtypfldindx("sgmnt_data","unused_dsk_reads.curr_count")=485
	Set gtmtypes("sgmnt_data",486,"name")="sgmnt_data.unused_dsk_reads.cumul_count"
	Set gtmtypes("sgmnt_data",486,"off")=3900
	Set gtmtypes("sgmnt_data",486,"len")=4
	Set gtmtypes("sgmnt_data",486,"type")="int"
	Set gtmtypfldindx("sgmnt_data","unused_dsk_reads.cumul_count")=486
	Set gtmtypes("sgmnt_data",487,"name")="sgmnt_data.n_bgmm_updates"
	Set gtmtypes("sgmnt_data",487,"off")=3904
	Set gtmtypes("sgmnt_data",487,"len")=8
	Set gtmtypes("sgmnt_data",487,"type")="db_csh_acct_rec"
	Set gtmtypfldindx("sgmnt_data","n_bgmm_updates")=487
	Set gtmtypes("sgmnt_data",488,"name")="sgmnt_data.n_bgmm_updates.curr_count"
	Set gtmtypes("sgmnt_data",488,"off")=3904
	Set gtmtypes("sgmnt_data",488,"len")=4
	Set gtmtypes("sgmnt_data",488,"type")="int"
	Set gtmtypfldindx("sgmnt_data","n_bgmm_updates.curr_count")=488
	Set gtmtypes("sgmnt_data",489,"name")="sgmnt_data.n_bgmm_updates.cumul_count"
	Set gtmtypes("sgmnt_data",489,"off")=3908
	Set gtmtypes("sgmnt_data",489,"len")=4
	Set gtmtypes("sgmnt_data",489,"type")="int"
	Set gtmtypfldindx("sgmnt_data","n_bgmm_updates.cumul_count")=489
	Set gtmtypes("sgmnt_data",490,"name")="sgmnt_data.unused_dsk_writes"
	Set gtmtypes("sgmnt_data",490,"off")=3912
	Set gtmtypes("sgmnt_data",490,"len")=8
	Set gtmtypes("sgmnt_data",490,"type")="db_csh_acct_rec"
	Set gtmtypfldindx("sgmnt_data","unused_dsk_writes")=490
	Set gtmtypes("sgmnt_data",491,"name")="sgmnt_data.unused_dsk_writes.curr_count"
	Set gtmtypes("sgmnt_data",491,"off")=3912
	Set gtmtypes("sgmnt_data",491,"len")=4
	Set gtmtypes("sgmnt_data",491,"type")="int"
	Set gtmtypfldindx("sgmnt_data","unused_dsk_writes.curr_count")=491
	Set gtmtypes("sgmnt_data",492,"name")="sgmnt_data.unused_dsk_writes.cumul_count"
	Set gtmtypes("sgmnt_data",492,"off")=3916
	Set gtmtypes("sgmnt_data",492,"len")=4
	Set gtmtypes("sgmnt_data",492,"type")="int"
	Set gtmtypfldindx("sgmnt_data","unused_dsk_writes.cumul_count")=492
	Set gtmtypes("sgmnt_data",493,"name")="sgmnt_data.n_bg_update_creates"
	Set gtmtypes("sgmnt_data",493,"off")=3920
	Set gtmtypes("sgmnt_data",493,"len")=8
	Set gtmtypes("sgmnt_data",493,"type")="db_csh_acct_rec"
	Set gtmtypfldindx("sgmnt_data","n_bg_update_creates")=493
	Set gtmtypes("sgmnt_data",494,"name")="sgmnt_data.n_bg_update_creates.curr_count"
	Set gtmtypes("sgmnt_data",494,"off")=3920
	Set gtmtypes("sgmnt_data",494,"len")=4
	Set gtmtypes("sgmnt_data",494,"type")="int"
	Set gtmtypfldindx("sgmnt_data","n_bg_update_creates.curr_count")=494
	Set gtmtypes("sgmnt_data",495,"name")="sgmnt_data.n_bg_update_creates.cumul_count"
	Set gtmtypes("sgmnt_data",495,"off")=3924
	Set gtmtypes("sgmnt_data",495,"len")=4
	Set gtmtypes("sgmnt_data",495,"type")="int"
	Set gtmtypfldindx("sgmnt_data","n_bg_update_creates.cumul_count")=495
	Set gtmtypes("sgmnt_data",496,"name")="sgmnt_data.n_db_csh_getns"
	Set gtmtypes("sgmnt_data",496,"off")=3928
	Set gtmtypes("sgmnt_data",496,"len")=8
	Set gtmtypes("sgmnt_data",496,"type")="db_csh_acct_rec"
	Set gtmtypfldindx("sgmnt_data","n_db_csh_getns")=496
	Set gtmtypes("sgmnt_data",497,"name")="sgmnt_data.n_db_csh_getns.curr_count"
	Set gtmtypes("sgmnt_data",497,"off")=3928
	Set gtmtypes("sgmnt_data",497,"len")=4
	Set gtmtypes("sgmnt_data",497,"type")="int"
	Set gtmtypfldindx("sgmnt_data","n_db_csh_getns.curr_count")=497
	Set gtmtypes("sgmnt_data",498,"name")="sgmnt_data.n_db_csh_getns.cumul_count"
	Set gtmtypes("sgmnt_data",498,"off")=3932
	Set gtmtypes("sgmnt_data",498,"len")=4
	Set gtmtypes("sgmnt_data",498,"type")="int"
	Set gtmtypfldindx("sgmnt_data","n_db_csh_getns.cumul_count")=498
	Set gtmtypes("sgmnt_data",499,"name")="sgmnt_data.n_db_csh_getn_lcnt"
	Set gtmtypes("sgmnt_data",499,"off")=3936
	Set gtmtypes("sgmnt_data",499,"len")=8
	Set gtmtypes("sgmnt_data",499,"type")="db_csh_acct_rec"
	Set gtmtypfldindx("sgmnt_data","n_db_csh_getn_lcnt")=499
	Set gtmtypes("sgmnt_data",500,"name")="sgmnt_data.n_db_csh_getn_lcnt.curr_count"
	Set gtmtypes("sgmnt_data",500,"off")=3936
	Set gtmtypes("sgmnt_data",500,"len")=4
	Set gtmtypes("sgmnt_data",500,"type")="int"
	Set gtmtypfldindx("sgmnt_data","n_db_csh_getn_lcnt.curr_count")=500
	Set gtmtypes("sgmnt_data",501,"name")="sgmnt_data.n_db_csh_getn_lcnt.cumul_count"
	Set gtmtypes("sgmnt_data",501,"off")=3940
	Set gtmtypes("sgmnt_data",501,"len")=4
	Set gtmtypes("sgmnt_data",501,"type")="int"
	Set gtmtypfldindx("sgmnt_data","n_db_csh_getn_lcnt.cumul_count")=501
	Set gtmtypes("sgmnt_data",502,"name")="sgmnt_data.db_csh_acct_rec_filler_4k"
	Set gtmtypes("sgmnt_data",502,"off")=3944
	Set gtmtypes("sgmnt_data",502,"len")=152
	Set gtmtypes("sgmnt_data",502,"type")="char"
	Set gtmtypfldindx("sgmnt_data","db_csh_acct_rec_filler_4k")=502
	Set gtmtypes("sgmnt_data",503,"name")="sgmnt_data.gvstats_rec_old_now_filler"
	Set gtmtypes("sgmnt_data",503,"off")=4096
	Set gtmtypes("sgmnt_data",503,"len")=496
	Set gtmtypes("sgmnt_data",503,"type")="char"
	Set gtmtypfldindx("sgmnt_data","gvstats_rec_old_now_filler")=503
	Set gtmtypes("sgmnt_data",504,"name")="sgmnt_data.gvstats_rec_filler_4k_plus_512"
	Set gtmtypes("sgmnt_data",504,"off")=4592
	Set gtmtypes("sgmnt_data",504,"len")=16
	Set gtmtypes("sgmnt_data",504,"type")="char"
	Set gtmtypfldindx("sgmnt_data","gvstats_rec_filler_4k_plus_512")=504
	Set gtmtypes("sgmnt_data",505,"name")="sgmnt_data.filler_4k_plus_512"
	Set gtmtypes("sgmnt_data",505,"off")=4608
	Set gtmtypes("sgmnt_data",505,"len")=368
	Set gtmtypes("sgmnt_data",505,"type")="char"
	Set gtmtypfldindx("sgmnt_data","filler_4k_plus_512")=505
	Set gtmtypes("sgmnt_data",506,"name")="sgmnt_data.intrpt_recov_resync_strm_seqno"
	Set gtmtypes("sgmnt_data",506,"off")=4976
	Set gtmtypes("sgmnt_data",506,"len")=128
	Set gtmtypes("sgmnt_data",506,"type")="uint64_t"
	Set gtmtypfldindx("sgmnt_data","intrpt_recov_resync_strm_seqno")=506
	Set gtmtypes("sgmnt_data",506,"dim")=16
	Set gtmtypes("sgmnt_data",507,"name")="sgmnt_data.creation_db_ver"
	Set gtmtypes("sgmnt_data",507,"off")=5104
	Set gtmtypes("sgmnt_data",507,"len")=4
	Set gtmtypes("sgmnt_data",507,"type")="int"
	Set gtmtypfldindx("sgmnt_data","creation_db_ver")=507
	Set gtmtypes("sgmnt_data",508,"name")="sgmnt_data.creation_mdb_ver"
	Set gtmtypes("sgmnt_data",508,"off")=5108
	Set gtmtypes("sgmnt_data",508,"len")=4
	Set gtmtypes("sgmnt_data",508,"type")="int"
	Set gtmtypfldindx("sgmnt_data","creation_mdb_ver")=508
	Set gtmtypes("sgmnt_data",509,"name")="sgmnt_data.certified_for_upgrade_to"
	Set gtmtypes("sgmnt_data",509,"off")=5112
	Set gtmtypes("sgmnt_data",509,"len")=4
	Set gtmtypes("sgmnt_data",509,"type")="int"
	Set gtmtypfldindx("sgmnt_data","certified_for_upgrade_to")=509
	Set gtmtypes("sgmnt_data",510,"name")="sgmnt_data.filler_5k"
	Set gtmtypes("sgmnt_data",510,"off")=5116
	Set gtmtypes("sgmnt_data",510,"len")=4
	Set gtmtypes("sgmnt_data",510,"type")="int"
	Set gtmtypfldindx("sgmnt_data","filler_5k")=510
	Set gtmtypes("sgmnt_data",511,"name")="sgmnt_data.secshr_ops_index_filler"
	Set gtmtypes("sgmnt_data",511,"off")=5120
	Set gtmtypes("sgmnt_data",511,"len")=4
	Set gtmtypes("sgmnt_data",511,"type")="int"
	Set gtmtypfldindx("sgmnt_data","secshr_ops_index_filler")=511
	Set gtmtypes("sgmnt_data",512,"name")="sgmnt_data.secshr_ops_array_filler"
	Set gtmtypes("sgmnt_data",512,"off")=5124
	Set gtmtypes("sgmnt_data",512,"len")=1020
	Set gtmtypes("sgmnt_data",512,"type")="int"
	Set gtmtypfldindx("sgmnt_data","secshr_ops_array_filler")=512
	Set gtmtypes("sgmnt_data",512,"dim")=255
	Set gtmtypes("sgmnt_data",513,"name")="sgmnt_data.next_upgrd_warn"
	Set gtmtypes("sgmnt_data",513,"off")=6144
	Set gtmtypes("sgmnt_data",513,"len")=24
	Set gtmtypes("sgmnt_data",513,"type")="compswap_time_field"
	Set gtmtypfldindx("sgmnt_data","next_upgrd_warn")=513
	Set gtmtypes("sgmnt_data",514,"name")="sgmnt_data.next_upgrd_warn.time_latch"
	Set gtmtypes("sgmnt_data",514,"off")=6144
	Set gtmtypes("sgmnt_data",514,"len")=8
	Set gtmtypes("sgmnt_data",514,"type")="global_latch_t"
	Set gtmtypfldindx("sgmnt_data","next_upgrd_warn.time_latch")=514
	Set gtmtypes("sgmnt_data",515,"name")="sgmnt_data.next_upgrd_warn.time_latch.u"
	Set gtmtypes("sgmnt_data",515,"off")=6144
	Set gtmtypes("sgmnt_data",515,"len")=8
	Set gtmtypes("sgmnt_data",515,"type")="union"
	Set gtmtypfldindx("sgmnt_data","next_upgrd_warn.time_latch.u")=515
	Set gtmtypes("sgmnt_data",516,"name")="sgmnt_data.next_upgrd_warn.time_latch.u.pid_imgcnt"
	Set gtmtypes("sgmnt_data",516,"off")=6144
	Set gtmtypes("sgmnt_data",516,"len")=8
	Set gtmtypes("sgmnt_data",516,"type")="uint64_t"
	Set gtmtypfldindx("sgmnt_data","next_upgrd_warn.time_latch.u.pid_imgcnt")=516
	Set gtmtypes("sgmnt_data",517,"name")="sgmnt_data.next_upgrd_warn.time_latch.u.parts"
	Set gtmtypes("sgmnt_data",517,"off")=6144
	Set gtmtypes("sgmnt_data",517,"len")=8
	Set gtmtypes("sgmnt_data",517,"type")="struct"
	Set gtmtypfldindx("sgmnt_data","next_upgrd_warn.time_latch.u.parts")=517
	Set gtmtypes("sgmnt_data",518,"name")="sgmnt_data.next_upgrd_warn.time_latch.u.parts.latch_pid"
	Set gtmtypes("sgmnt_data",518,"off")=6144
	Set gtmtypes("sgmnt_data",518,"len")=4
	Set gtmtypes("sgmnt_data",518,"type")="int"
	Set gtmtypfldindx("sgmnt_data","next_upgrd_warn.time_latch.u.parts.latch_pid")=518
	Set gtmtypes("sgmnt_data",519,"name")="sgmnt_data.next_upgrd_warn.time_latch.u.parts.latch_word"
	Set gtmtypes("sgmnt_data",519,"off")=6148
	Set gtmtypes("sgmnt_data",519,"len")=4
	Set gtmtypes("sgmnt_data",519,"type")="int"
	Set gtmtypfldindx("sgmnt_data","next_upgrd_warn.time_latch.u.parts.latch_word")=519
	Set gtmtypes("sgmnt_data",520,"name")="sgmnt_data.next_upgrd_warn.hp_latch_space"
	Set gtmtypes("sgmnt_data",520,"off")=6152
	Set gtmtypes("sgmnt_data",520,"len")=16
	Set gtmtypes("sgmnt_data",520,"type")="int"
	Set gtmtypfldindx("sgmnt_data","next_upgrd_warn.hp_latch_space")=520
	Set gtmtypes("sgmnt_data",520,"dim")=4
	Set gtmtypes("sgmnt_data",521,"name")="sgmnt_data.is_encrypted"
	Set gtmtypes("sgmnt_data",521,"off")=6168
	Set gtmtypes("sgmnt_data",521,"len")=4
	Set gtmtypes("sgmnt_data",521,"type")="unsigned-int"
	Set gtmtypfldindx("sgmnt_data","is_encrypted")=521
	Set gtmtypes("sgmnt_data",522,"name")="sgmnt_data.db_trigger_cycle"
	Set gtmtypes("sgmnt_data",522,"off")=6172
	Set gtmtypes("sgmnt_data",522,"len")=4
	Set gtmtypes("sgmnt_data",522,"type")="unsigned-int"
	Set gtmtypfldindx("sgmnt_data","db_trigger_cycle")=522
	Set gtmtypes("sgmnt_data",523,"name")="sgmnt_data.strm_reg_seqno"
	Set gtmtypes("sgmnt_data",523,"off")=6176
	Set gtmtypes("sgmnt_data",523,"len")=128
	Set gtmtypes("sgmnt_data",523,"type")="uint64_t"
	Set gtmtypfldindx("sgmnt_data","strm_reg_seqno")=523
	Set gtmtypes("sgmnt_data",523,"dim")=16
	Set gtmtypes("sgmnt_data",524,"name")="sgmnt_data.save_strm_reg_seqno"
	Set gtmtypes("sgmnt_data",524,"off")=6304
	Set gtmtypes("sgmnt_data",524,"len")=128
	Set gtmtypes("sgmnt_data",524,"type")="uint64_t"
	Set gtmtypfldindx("sgmnt_data","save_strm_reg_seqno")=524
	Set gtmtypes("sgmnt_data",524,"dim")=16
	Set gtmtypes("sgmnt_data",525,"name")="sgmnt_data.freeze_on_fail"
	Set gtmtypes("sgmnt_data",525,"off")=6432
	Set gtmtypes("sgmnt_data",525,"len")=4
	Set gtmtypes("sgmnt_data",525,"type")="boolean_t"
	Set gtmtypfldindx("sgmnt_data","freeze_on_fail")=525
	Set gtmtypes("sgmnt_data",526,"name")="sgmnt_data.span_node_absent"
	Set gtmtypes("sgmnt_data",526,"off")=6436
	Set gtmtypes("sgmnt_data",526,"len")=4
	Set gtmtypes("sgmnt_data",526,"type")="boolean_t"
	Set gtmtypfldindx("sgmnt_data","span_node_absent")=526
	Set gtmtypes("sgmnt_data",527,"name")="sgmnt_data.maxkeysz_assured"
	Set gtmtypes("sgmnt_data",527,"off")=6440
	Set gtmtypes("sgmnt_data",527,"len")=4
	Set gtmtypes("sgmnt_data",527,"type")="boolean_t"
	Set gtmtypfldindx("sgmnt_data","maxkeysz_assured")=527
	Set gtmtypes("sgmnt_data",528,"name")="sgmnt_data.hasht_upgrade_needed"
	Set gtmtypes("sgmnt_data",528,"off")=6444
	Set gtmtypes("sgmnt_data",528,"len")=4
	Set gtmtypes("sgmnt_data",528,"type")="boolean_t"
	Set gtmtypfldindx("sgmnt_data","hasht_upgrade_needed")=528
	Set gtmtypes("sgmnt_data",529,"name")="sgmnt_data.defer_allocate"
	Set gtmtypes("sgmnt_data",529,"off")=6448
	Set gtmtypes("sgmnt_data",529,"len")=4
	Set gtmtypes("sgmnt_data",529,"type")="boolean_t"
	Set gtmtypfldindx("sgmnt_data","defer_allocate")=529
	Set gtmtypes("sgmnt_data",530,"name")="sgmnt_data.filler_ftok_counter_halted"
	Set gtmtypes("sgmnt_data",530,"off")=6452
	Set gtmtypes("sgmnt_data",530,"len")=4
	Set gtmtypes("sgmnt_data",530,"type")="boolean_t"
	Set gtmtypfldindx("sgmnt_data","filler_ftok_counter_halted")=530
	Set gtmtypes("sgmnt_data",531,"name")="sgmnt_data.filler_access_counter_halted"
	Set gtmtypes("sgmnt_data",531,"off")=6456
	Set gtmtypes("sgmnt_data",531,"len")=4
	Set gtmtypes("sgmnt_data",531,"type")="boolean_t"
	Set gtmtypfldindx("sgmnt_data","filler_access_counter_halted")=531
	Set gtmtypes("sgmnt_data",532,"name")="sgmnt_data.lock_crit_with_db"
	Set gtmtypes("sgmnt_data",532,"off")=6460
	Set gtmtypes("sgmnt_data",532,"len")=4
	Set gtmtypes("sgmnt_data",532,"type")="boolean_t"
	Set gtmtypfldindx("sgmnt_data","lock_crit_with_db")=532
	Set gtmtypes("sgmnt_data",533,"name")="sgmnt_data.basedb_fname_len"
	Set gtmtypes("sgmnt_data",533,"off")=6464
	Set gtmtypes("sgmnt_data",533,"len")=4
	Set gtmtypes("sgmnt_data",533,"type")="unsigned-int"
	Set gtmtypfldindx("sgmnt_data","basedb_fname_len")=533
	Set gtmtypes("sgmnt_data",534,"name")="sgmnt_data.basedb_fname"
	Set gtmtypes("sgmnt_data",534,"off")=6468
	Set gtmtypes("sgmnt_data",534,"len")=256
	Set gtmtypes("sgmnt_data",534,"type")="unsigned-char"
	Set gtmtypfldindx("sgmnt_data","basedb_fname")=534
	Set gtmtypes("sgmnt_data",535,"name")="sgmnt_data.read_only"
	Set gtmtypes("sgmnt_data",535,"off")=6724
	Set gtmtypes("sgmnt_data",535,"len")=4
	Set gtmtypes("sgmnt_data",535,"type")="boolean_t"
	Set gtmtypfldindx("sgmnt_data","read_only")=535
	Set gtmtypes("sgmnt_data",536,"name")="sgmnt_data.gvstats_rec"
	Set gtmtypes("sgmnt_data",536,"off")=6728
	Set gtmtypes("sgmnt_data",536,"len")=592
	Set gtmtypes("sgmnt_data",536,"type")="gvstats_rec_csd_t"
	Set gtmtypfldindx("sgmnt_data","gvstats_rec")=536
	Set gtmtypes("sgmnt_data",537,"name")="sgmnt_data.gvstats_rec.n_set"
	Set gtmtypes("sgmnt_data",537,"off")=6728
	Set gtmtypes("sgmnt_data",537,"len")=8
	Set gtmtypes("sgmnt_data",537,"type")="uint64_t"
	Set gtmtypfldindx("sgmnt_data","gvstats_rec.n_set")=537
	Set gtmtypes("sgmnt_data",538,"name")="sgmnt_data.gvstats_rec.n_kill"
	Set gtmtypes("sgmnt_data",538,"off")=6736
	Set gtmtypes("sgmnt_data",538,"len")=8
	Set gtmtypes("sgmnt_data",538,"type")="uint64_t"
	Set gtmtypfldindx("sgmnt_data","gvstats_rec.n_kill")=538
	Set gtmtypes("sgmnt_data",539,"name")="sgmnt_data.gvstats_rec.n_get"
	Set gtmtypes("sgmnt_data",539,"off")=6744
	Set gtmtypes("sgmnt_data",539,"len")=8
	Set gtmtypes("sgmnt_data",539,"type")="uint64_t"
	Set gtmtypfldindx("sgmnt_data","gvstats_rec.n_get")=539
	Set gtmtypes("sgmnt_data",540,"name")="sgmnt_data.gvstats_rec.n_data"
	Set gtmtypes("sgmnt_data",540,"off")=6752
	Set gtmtypes("sgmnt_data",540,"len")=8
	Set gtmtypes("sgmnt_data",540,"type")="uint64_t"
	Set gtmtypfldindx("sgmnt_data","gvstats_rec.n_data")=540
	Set gtmtypes("sgmnt_data",541,"name")="sgmnt_data.gvstats_rec.n_order"
	Set gtmtypes("sgmnt_data",541,"off")=6760
	Set gtmtypes("sgmnt_data",541,"len")=8
	Set gtmtypes("sgmnt_data",541,"type")="uint64_t"
	Set gtmtypfldindx("sgmnt_data","gvstats_rec.n_order")=541
	Set gtmtypes("sgmnt_data",542,"name")="sgmnt_data.gvstats_rec.n_zprev"
	Set gtmtypes("sgmnt_data",542,"off")=6768
	Set gtmtypes("sgmnt_data",542,"len")=8
	Set gtmtypes("sgmnt_data",542,"type")="uint64_t"
	Set gtmtypfldindx("sgmnt_data","gvstats_rec.n_zprev")=542
	Set gtmtypes("sgmnt_data",543,"name")="sgmnt_data.gvstats_rec.n_query"
	Set gtmtypes("sgmnt_data",543,"off")=6776
	Set gtmtypes("sgmnt_data",543,"len")=8
	Set gtmtypes("sgmnt_data",543,"type")="uint64_t"
	Set gtmtypfldindx("sgmnt_data","gvstats_rec.n_query")=543
	Set gtmtypes("sgmnt_data",544,"name")="sgmnt_data.gvstats_rec.n_lock_success"
	Set gtmtypes("sgmnt_data",544,"off")=6784
	Set gtmtypes("sgmnt_data",544,"len")=8
	Set gtmtypes("sgmnt_data",544,"type")="uint64_t"
	Set gtmtypfldindx("sgmnt_data","gvstats_rec.n_lock_success")=544
	Set gtmtypes("sgmnt_data",545,"name")="sgmnt_data.gvstats_rec.n_lock_fail"
	Set gtmtypes("sgmnt_data",545,"off")=6792
	Set gtmtypes("sgmnt_data",545,"len")=8
	Set gtmtypes("sgmnt_data",545,"type")="uint64_t"
	Set gtmtypfldindx("sgmnt_data","gvstats_rec.n_lock_fail")=545
	Set gtmtypes("sgmnt_data",546,"name")="sgmnt_data.gvstats_rec.db_curr_tn"
	Set gtmtypes("sgmnt_data",546,"off")=6800
	Set gtmtypes("sgmnt_data",546,"len")=8
	Set gtmtypes("sgmnt_data",546,"type")="uint64_t"
	Set gtmtypfldindx("sgmnt_data","gvstats_rec.db_curr_tn")=546
	Set gtmtypes("sgmnt_data",547,"name")="sgmnt_data.gvstats_rec.n_dsk_read"
	Set gtmtypes("sgmnt_data",547,"off")=6808
	Set gtmtypes("sgmnt_data",547,"len")=8
	Set gtmtypes("sgmnt_data",547,"type")="uint64_t"
	Set gtmtypfldindx("sgmnt_data","gvstats_rec.n_dsk_read")=547
	Set gtmtypes("sgmnt_data",548,"name")="sgmnt_data.gvstats_rec.n_dsk_write"
	Set gtmtypes("sgmnt_data",548,"off")=6816
	Set gtmtypes("sgmnt_data",548,"len")=8
	Set gtmtypes("sgmnt_data",548,"type")="uint64_t"
	Set gtmtypfldindx("sgmnt_data","gvstats_rec.n_dsk_write")=548
	Set gtmtypes("sgmnt_data",549,"name")="sgmnt_data.gvstats_rec.n_nontp_readwrite"
	Set gtmtypes("sgmnt_data",549,"off")=6824
	Set gtmtypes("sgmnt_data",549,"len")=8
	Set gtmtypes("sgmnt_data",549,"type")="uint64_t"
	Set gtmtypfldindx("sgmnt_data","gvstats_rec.n_nontp_readwrite")=549
	Set gtmtypes("sgmnt_data",550,"name")="sgmnt_data.gvstats_rec.n_nontp_readonly"
	Set gtmtypes("sgmnt_data",550,"off")=6832
	Set gtmtypes("sgmnt_data",550,"len")=8
	Set gtmtypes("sgmnt_data",550,"type")="uint64_t"
	Set gtmtypfldindx("sgmnt_data","gvstats_rec.n_nontp_readonly")=550
	Set gtmtypes("sgmnt_data",551,"name")="sgmnt_data.gvstats_rec.n_nontp_blkwrite"
	Set gtmtypes("sgmnt_data",551,"off")=6840
	Set gtmtypes("sgmnt_data",551,"len")=8
	Set gtmtypes("sgmnt_data",551,"type")="uint64_t"
	Set gtmtypfldindx("sgmnt_data","gvstats_rec.n_nontp_blkwrite")=551
	Set gtmtypes("sgmnt_data",552,"name")="sgmnt_data.gvstats_rec.n_nontp_blkread"
	Set gtmtypes("sgmnt_data",552,"off")=6848
	Set gtmtypes("sgmnt_data",552,"len")=8
	Set gtmtypes("sgmnt_data",552,"type")="uint64_t"
	Set gtmtypfldindx("sgmnt_data","gvstats_rec.n_nontp_blkread")=552
	Set gtmtypes("sgmnt_data",553,"name")="sgmnt_data.gvstats_rec.n_nontp_retries_0"
	Set gtmtypes("sgmnt_data",553,"off")=6856
	Set gtmtypes("sgmnt_data",553,"len")=8
	Set gtmtypes("sgmnt_data",553,"type")="uint64_t"
	Set gtmtypfldindx("sgmnt_data","gvstats_rec.n_nontp_retries_0")=553
	Set gtmtypes("sgmnt_data",554,"name")="sgmnt_data.gvstats_rec.n_nontp_retries_1"
	Set gtmtypes("sgmnt_data",554,"off")=6864
	Set gtmtypes("sgmnt_data",554,"len")=8
	Set gtmtypes("sgmnt_data",554,"type")="uint64_t"
	Set gtmtypfldindx("sgmnt_data","gvstats_rec.n_nontp_retries_1")=554
	Set gtmtypes("sgmnt_data",555,"name")="sgmnt_data.gvstats_rec.n_nontp_retries_2"
	Set gtmtypes("sgmnt_data",555,"off")=6872
	Set gtmtypes("sgmnt_data",555,"len")=8
	Set gtmtypes("sgmnt_data",555,"type")="uint64_t"
	Set gtmtypfldindx("sgmnt_data","gvstats_rec.n_nontp_retries_2")=555
	Set gtmtypes("sgmnt_data",556,"name")="sgmnt_data.gvstats_rec.n_nontp_retries_3"
	Set gtmtypes("sgmnt_data",556,"off")=6880
	Set gtmtypes("sgmnt_data",556,"len")=8
	Set gtmtypes("sgmnt_data",556,"type")="uint64_t"
	Set gtmtypfldindx("sgmnt_data","gvstats_rec.n_nontp_retries_3")=556
	Set gtmtypes("sgmnt_data",557,"name")="sgmnt_data.gvstats_rec.n_tp_readwrite"
	Set gtmtypes("sgmnt_data",557,"off")=6888
	Set gtmtypes("sgmnt_data",557,"len")=8
	Set gtmtypes("sgmnt_data",557,"type")="uint64_t"
	Set gtmtypfldindx("sgmnt_data","gvstats_rec.n_tp_readwrite")=557
	Set gtmtypes("sgmnt_data",558,"name")="sgmnt_data.gvstats_rec.n_tp_readonly"
	Set gtmtypes("sgmnt_data",558,"off")=6896
	Set gtmtypes("sgmnt_data",558,"len")=8
	Set gtmtypes("sgmnt_data",558,"type")="uint64_t"
	Set gtmtypfldindx("sgmnt_data","gvstats_rec.n_tp_readonly")=558
	Set gtmtypes("sgmnt_data",559,"name")="sgmnt_data.gvstats_rec.n_tp_rolledback"
	Set gtmtypes("sgmnt_data",559,"off")=6904
	Set gtmtypes("sgmnt_data",559,"len")=8
	Set gtmtypes("sgmnt_data",559,"type")="uint64_t"
	Set gtmtypfldindx("sgmnt_data","gvstats_rec.n_tp_rolledback")=559
	Set gtmtypes("sgmnt_data",560,"name")="sgmnt_data.gvstats_rec.n_tp_blkwrite"
	Set gtmtypes("sgmnt_data",560,"off")=6912
	Set gtmtypes("sgmnt_data",560,"len")=8
	Set gtmtypes("sgmnt_data",560,"type")="uint64_t"
	Set gtmtypfldindx("sgmnt_data","gvstats_rec.n_tp_blkwrite")=560
	Set gtmtypes("sgmnt_data",561,"name")="sgmnt_data.gvstats_rec.n_tp_blkread"
	Set gtmtypes("sgmnt_data",561,"off")=6920
	Set gtmtypes("sgmnt_data",561,"len")=8
	Set gtmtypes("sgmnt_data",561,"type")="uint64_t"
	Set gtmtypfldindx("sgmnt_data","gvstats_rec.n_tp_blkread")=561
	Set gtmtypes("sgmnt_data",562,"name")="sgmnt_data.gvstats_rec.n_tp_tot_retries_0"
	Set gtmtypes("sgmnt_data",562,"off")=6928
	Set gtmtypes("sgmnt_data",562,"len")=8
	Set gtmtypes("sgmnt_data",562,"type")="uint64_t"
	Set gtmtypfldindx("sgmnt_data","gvstats_rec.n_tp_tot_retries_0")=562
	Set gtmtypes("sgmnt_data",563,"name")="sgmnt_data.gvstats_rec.n_tp_tot_retries_1"
	Set gtmtypes("sgmnt_data",563,"off")=6936
	Set gtmtypes("sgmnt_data",563,"len")=8
	Set gtmtypes("sgmnt_data",563,"type")="uint64_t"
	Set gtmtypfldindx("sgmnt_data","gvstats_rec.n_tp_tot_retries_1")=563
	Set gtmtypes("sgmnt_data",564,"name")="sgmnt_data.gvstats_rec.n_tp_tot_retries_2"
	Set gtmtypes("sgmnt_data",564,"off")=6944
	Set gtmtypes("sgmnt_data",564,"len")=8
	Set gtmtypes("sgmnt_data",564,"type")="uint64_t"
	Set gtmtypfldindx("sgmnt_data","gvstats_rec.n_tp_tot_retries_2")=564
	Set gtmtypes("sgmnt_data",565,"name")="sgmnt_data.gvstats_rec.n_tp_tot_retries_3"
	Set gtmtypes("sgmnt_data",565,"off")=6952
	Set gtmtypes("sgmnt_data",565,"len")=8
	Set gtmtypes("sgmnt_data",565,"type")="uint64_t"
	Set gtmtypfldindx("sgmnt_data","gvstats_rec.n_tp_tot_retries_3")=565
	Set gtmtypes("sgmnt_data",566,"name")="sgmnt_data.gvstats_rec.n_tp_tot_retries_4"
	Set gtmtypes("sgmnt_data",566,"off")=6960
	Set gtmtypes("sgmnt_data",566,"len")=8
	Set gtmtypes("sgmnt_data",566,"type")="uint64_t"
	Set gtmtypfldindx("sgmnt_data","gvstats_rec.n_tp_tot_retries_4")=566
	Set gtmtypes("sgmnt_data",567,"name")="sgmnt_data.gvstats_rec.n_tp_cnflct_retries_0"
	Set gtmtypes("sgmnt_data",567,"off")=6968
	Set gtmtypes("sgmnt_data",567,"len")=8
	Set gtmtypes("sgmnt_data",567,"type")="uint64_t"
	Set gtmtypfldindx("sgmnt_data","gvstats_rec.n_tp_cnflct_retries_0")=567
	Set gtmtypes("sgmnt_data",568,"name")="sgmnt_data.gvstats_rec.n_tp_cnflct_retries_1"
	Set gtmtypes("sgmnt_data",568,"off")=6976
	Set gtmtypes("sgmnt_data",568,"len")=8
	Set gtmtypes("sgmnt_data",568,"type")="uint64_t"
	Set gtmtypfldindx("sgmnt_data","gvstats_rec.n_tp_cnflct_retries_1")=568
	Set gtmtypes("sgmnt_data",569,"name")="sgmnt_data.gvstats_rec.n_tp_cnflct_retries_2"
	Set gtmtypes("sgmnt_data",569,"off")=6984
	Set gtmtypes("sgmnt_data",569,"len")=8
	Set gtmtypes("sgmnt_data",569,"type")="uint64_t"
	Set gtmtypfldindx("sgmnt_data","gvstats_rec.n_tp_cnflct_retries_2")=569
	Set gtmtypes("sgmnt_data",570,"name")="sgmnt_data.gvstats_rec.n_tp_cnflct_retries_3"
	Set gtmtypes("sgmnt_data",570,"off")=6992
	Set gtmtypes("sgmnt_data",570,"len")=8
	Set gtmtypes("sgmnt_data",570,"type")="uint64_t"
	Set gtmtypfldindx("sgmnt_data","gvstats_rec.n_tp_cnflct_retries_3")=570
	Set gtmtypes("sgmnt_data",571,"name")="sgmnt_data.gvstats_rec.n_tp_cnflct_retries_4"
	Set gtmtypes("sgmnt_data",571,"off")=7000
	Set gtmtypes("sgmnt_data",571,"len")=8
	Set gtmtypes("sgmnt_data",571,"type")="uint64_t"
	Set gtmtypfldindx("sgmnt_data","gvstats_rec.n_tp_cnflct_retries_4")=571
	Set gtmtypes("sgmnt_data",572,"name")="sgmnt_data.gvstats_rec.n_ztrigger"
	Set gtmtypes("sgmnt_data",572,"off")=7008
	Set gtmtypes("sgmnt_data",572,"len")=8
	Set gtmtypes("sgmnt_data",572,"type")="uint64_t"
	Set gtmtypfldindx("sgmnt_data","gvstats_rec.n_ztrigger")=572
	Set gtmtypes("sgmnt_data",573,"name")="sgmnt_data.gvstats_rec.n_db_flush"
	Set gtmtypes("sgmnt_data",573,"off")=7016
	Set gtmtypes("sgmnt_data",573,"len")=8
	Set gtmtypes("sgmnt_data",573,"type")="uint64_t"
	Set gtmtypfldindx("sgmnt_data","gvstats_rec.n_db_flush")=573
	Set gtmtypes("sgmnt_data",574,"name")="sgmnt_data.gvstats_rec.n_db_fsync"
	Set gtmtypes("sgmnt_data",574,"off")=7024
	Set gtmtypes("sgmnt_data",574,"len")=8
	Set gtmtypes("sgmnt_data",574,"type")="uint64_t"
	Set gtmtypfldindx("sgmnt_data","gvstats_rec.n_db_fsync")=574
	Set gtmtypes("sgmnt_data",575,"name")="sgmnt_data.gvstats_rec.n_jnl_flush"
	Set gtmtypes("sgmnt_data",575,"off")=7032
	Set gtmtypes("sgmnt_data",575,"len")=8
	Set gtmtypes("sgmnt_data",575,"type")="uint64_t"
	Set gtmtypfldindx("sgmnt_data","gvstats_rec.n_jnl_flush")=575
	Set gtmtypes("sgmnt_data",576,"name")="sgmnt_data.gvstats_rec.n_jnl_fsync"
	Set gtmtypes("sgmnt_data",576,"off")=7040
	Set gtmtypes("sgmnt_data",576,"len")=8
	Set gtmtypes("sgmnt_data",576,"type")="uint64_t"
	Set gtmtypfldindx("sgmnt_data","gvstats_rec.n_jnl_fsync")=576
	Set gtmtypes("sgmnt_data",577,"name")="sgmnt_data.gvstats_rec.n_jbuff_bytes"
	Set gtmtypes("sgmnt_data",577,"off")=7048
	Set gtmtypes("sgmnt_data",577,"len")=8
	Set gtmtypes("sgmnt_data",577,"type")="uint64_t"
	Set gtmtypfldindx("sgmnt_data","gvstats_rec.n_jbuff_bytes")=577
	Set gtmtypes("sgmnt_data",578,"name")="sgmnt_data.gvstats_rec.n_jfile_bytes"
	Set gtmtypes("sgmnt_data",578,"off")=7056
	Set gtmtypes("sgmnt_data",578,"len")=8
	Set gtmtypes("sgmnt_data",578,"type")="uint64_t"
	Set gtmtypfldindx("sgmnt_data","gvstats_rec.n_jfile_bytes")=578
	Set gtmtypes("sgmnt_data",579,"name")="sgmnt_data.gvstats_rec.n_jfile_writes"
	Set gtmtypes("sgmnt_data",579,"off")=7064
	Set gtmtypes("sgmnt_data",579,"len")=8
	Set gtmtypes("sgmnt_data",579,"type")="uint64_t"
	Set gtmtypfldindx("sgmnt_data","gvstats_rec.n_jfile_writes")=579
	Set gtmtypes("sgmnt_data",580,"name")="sgmnt_data.gvstats_rec.n_jrec_logical"
	Set gtmtypes("sgmnt_data",580,"off")=7072
	Set gtmtypes("sgmnt_data",580,"len")=8
	Set gtmtypes("sgmnt_data",580,"type")="uint64_t"
	Set gtmtypfldindx("sgmnt_data","gvstats_rec.n_jrec_logical")=580
	Set gtmtypes("sgmnt_data",581,"name")="sgmnt_data.gvstats_rec.n_jrec_pblk"
	Set gtmtypes("sgmnt_data",581,"off")=7080
	Set gtmtypes("sgmnt_data",581,"len")=8
	Set gtmtypes("sgmnt_data",581,"type")="uint64_t"
	Set gtmtypfldindx("sgmnt_data","gvstats_rec.n_jrec_pblk")=581
	Set gtmtypes("sgmnt_data",582,"name")="sgmnt_data.gvstats_rec.n_jrec_epoch_regular"
	Set gtmtypes("sgmnt_data",582,"off")=7088
	Set gtmtypes("sgmnt_data",582,"len")=8
	Set gtmtypes("sgmnt_data",582,"type")="uint64_t"
	Set gtmtypfldindx("sgmnt_data","gvstats_rec.n_jrec_epoch_regular")=582
	Set gtmtypes("sgmnt_data",583,"name")="sgmnt_data.gvstats_rec.n_jrec_epoch_idle"
	Set gtmtypes("sgmnt_data",583,"off")=7096
	Set gtmtypes("sgmnt_data",583,"len")=8
	Set gtmtypes("sgmnt_data",583,"type")="uint64_t"
	Set gtmtypfldindx("sgmnt_data","gvstats_rec.n_jrec_epoch_idle")=583
	Set gtmtypes("sgmnt_data",584,"name")="sgmnt_data.gvstats_rec.n_jrec_other"
	Set gtmtypes("sgmnt_data",584,"off")=7104
	Set gtmtypes("sgmnt_data",584,"len")=8
	Set gtmtypes("sgmnt_data",584,"type")="uint64_t"
	Set gtmtypfldindx("sgmnt_data","gvstats_rec.n_jrec_other")=584
	Set gtmtypes("sgmnt_data",585,"name")="sgmnt_data.gvstats_rec.n_jnl_extends"
	Set gtmtypes("sgmnt_data",585,"off")=7112
	Set gtmtypes("sgmnt_data",585,"len")=8
	Set gtmtypes("sgmnt_data",585,"type")="uint64_t"
	Set gtmtypfldindx("sgmnt_data","gvstats_rec.n_jnl_extends")=585
	Set gtmtypes("sgmnt_data",586,"name")="sgmnt_data.gvstats_rec.n_db_extends"
	Set gtmtypes("sgmnt_data",586,"off")=7120
	Set gtmtypes("sgmnt_data",586,"len")=8
	Set gtmtypes("sgmnt_data",586,"type")="uint64_t"
	Set gtmtypfldindx("sgmnt_data","gvstats_rec.n_db_extends")=586
	Set gtmtypes("sgmnt_data",587,"name")="sgmnt_data.gvstats_rec.n_crit_success"
	Set gtmtypes("sgmnt_data",587,"off")=7128
	Set gtmtypes("sgmnt_data",587,"len")=8
	Set gtmtypes("sgmnt_data",587,"type")="uint64_t"
	Set gtmtypfldindx("sgmnt_data","gvstats_rec.n_crit_success")=587
	Set gtmtypes("sgmnt_data",588,"name")="sgmnt_data.gvstats_rec.n_crits_in_epch"
	Set gtmtypes("sgmnt_data",588,"off")=7136
	Set gtmtypes("sgmnt_data",588,"len")=8
	Set gtmtypes("sgmnt_data",588,"type")="uint64_t"
	Set gtmtypfldindx("sgmnt_data","gvstats_rec.n_crits_in_epch")=588
	Set gtmtypes("sgmnt_data",589,"name")="sgmnt_data.gvstats_rec.sq_crit_failed"
	Set gtmtypes("sgmnt_data",589,"off")=7144
	Set gtmtypes("sgmnt_data",589,"len")=8
	Set gtmtypes("sgmnt_data",589,"type")="uint64_t"
	Set gtmtypfldindx("sgmnt_data","gvstats_rec.sq_crit_failed")=589
	Set gtmtypes("sgmnt_data",590,"name")="sgmnt_data.gvstats_rec.n_crit_failed"
	Set gtmtypes("sgmnt_data",590,"off")=7152
	Set gtmtypes("sgmnt_data",590,"len")=8
	Set gtmtypes("sgmnt_data",590,"type")="uint64_t"
	Set gtmtypfldindx("sgmnt_data","gvstats_rec.n_crit_failed")=590
	Set gtmtypes("sgmnt_data",591,"name")="sgmnt_data.gvstats_rec.sq_crit_que_slps"
	Set gtmtypes("sgmnt_data",591,"off")=7160
	Set gtmtypes("sgmnt_data",591,"len")=8
	Set gtmtypes("sgmnt_data",591,"type")="uint64_t"
	Set gtmtypfldindx("sgmnt_data","gvstats_rec.sq_crit_que_slps")=591
	Set gtmtypes("sgmnt_data",592,"name")="sgmnt_data.gvstats_rec.n_crit_que_slps"
	Set gtmtypes("sgmnt_data",592,"off")=7168
	Set gtmtypes("sgmnt_data",592,"len")=8
	Set gtmtypes("sgmnt_data",592,"type")="uint64_t"
	Set gtmtypfldindx("sgmnt_data","gvstats_rec.n_crit_que_slps")=592
	Set gtmtypes("sgmnt_data",593,"name")="sgmnt_data.gvstats_rec.sq_crit_yields"
	Set gtmtypes("sgmnt_data",593,"off")=7176
	Set gtmtypes("sgmnt_data",593,"len")=8
	Set gtmtypes("sgmnt_data",593,"type")="uint64_t"
	Set gtmtypfldindx("sgmnt_data","gvstats_rec.sq_crit_yields")=593
	Set gtmtypes("sgmnt_data",594,"name")="sgmnt_data.gvstats_rec.n_crit_yields"
	Set gtmtypes("sgmnt_data",594,"off")=7184
	Set gtmtypes("sgmnt_data",594,"len")=8
	Set gtmtypes("sgmnt_data",594,"type")="uint64_t"
	Set gtmtypfldindx("sgmnt_data","gvstats_rec.n_crit_yields")=594
	Set gtmtypes("sgmnt_data",595,"name")="sgmnt_data.gvstats_rec.n_clean2dirty"
	Set gtmtypes("sgmnt_data",595,"off")=7192
	Set gtmtypes("sgmnt_data",595,"len")=8
	Set gtmtypes("sgmnt_data",595,"type")="uint64_t"
	Set gtmtypfldindx("sgmnt_data","gvstats_rec.n_clean2dirty")=595
	Set gtmtypes("sgmnt_data",596,"name")="sgmnt_data.gvstats_rec.n_wait_for_read"
	Set gtmtypes("sgmnt_data",596,"off")=7200
	Set gtmtypes("sgmnt_data",596,"len")=8
	Set gtmtypes("sgmnt_data",596,"type")="uint64_t"
	Set gtmtypfldindx("sgmnt_data","gvstats_rec.n_wait_for_read")=596
	Set gtmtypes("sgmnt_data",597,"name")="sgmnt_data.gvstats_rec.n_buffer_scarce"
	Set gtmtypes("sgmnt_data",597,"off")=7208
	Set gtmtypes("sgmnt_data",597,"len")=8
	Set gtmtypes("sgmnt_data",597,"type")="uint64_t"
	Set gtmtypfldindx("sgmnt_data","gvstats_rec.n_buffer_scarce")=597
	Set gtmtypes("sgmnt_data",598,"name")="sgmnt_data.gvstats_rec.n_bt_scarce"
	Set gtmtypes("sgmnt_data",598,"off")=7216
	Set gtmtypes("sgmnt_data",598,"len")=8
	Set gtmtypes("sgmnt_data",598,"type")="uint64_t"
	Set gtmtypfldindx("sgmnt_data","gvstats_rec.n_bt_scarce")=598
	Set gtmtypes("sgmnt_data",599,"name")="sgmnt_data.gvstats_rec.f_dbext_wait"
	Set gtmtypes("sgmnt_data",599,"off")=7224
	Set gtmtypes("sgmnt_data",599,"len")=8
	Set gtmtypes("sgmnt_data",599,"type")="uint64_t"
	Set gtmtypfldindx("sgmnt_data","gvstats_rec.f_dbext_wait")=599
	Set gtmtypes("sgmnt_data",600,"name")="sgmnt_data.gvstats_rec.f_bg_wait"
	Set gtmtypes("sgmnt_data",600,"off")=7232
	Set gtmtypes("sgmnt_data",600,"len")=8
	Set gtmtypes("sgmnt_data",600,"type")="uint64_t"
	Set gtmtypfldindx("sgmnt_data","gvstats_rec.f_bg_wait")=600
	Set gtmtypes("sgmnt_data",601,"name")="sgmnt_data.gvstats_rec.f_jnl_wait"
	Set gtmtypes("sgmnt_data",601,"off")=7240
	Set gtmtypes("sgmnt_data",601,"len")=8
	Set gtmtypes("sgmnt_data",601,"type")="uint64_t"
	Set gtmtypfldindx("sgmnt_data","gvstats_rec.f_jnl_wait")=601
	Set gtmtypes("sgmnt_data",602,"name")="sgmnt_data.gvstats_rec.f_mlk_wait"
	Set gtmtypes("sgmnt_data",602,"off")=7248
	Set gtmtypes("sgmnt_data",602,"len")=8
	Set gtmtypes("sgmnt_data",602,"type")="uint64_t"
	Set gtmtypfldindx("sgmnt_data","gvstats_rec.f_mlk_wait")=602
	Set gtmtypes("sgmnt_data",603,"name")="sgmnt_data.gvstats_rec.f_proc_wait"
	Set gtmtypes("sgmnt_data",603,"off")=7256
	Set gtmtypes("sgmnt_data",603,"len")=8
	Set gtmtypes("sgmnt_data",603,"type")="uint64_t"
	Set gtmtypfldindx("sgmnt_data","gvstats_rec.f_proc_wait")=603
	Set gtmtypes("sgmnt_data",604,"name")="sgmnt_data.gvstats_rec.f_trans_wait"
	Set gtmtypes("sgmnt_data",604,"off")=7264
	Set gtmtypes("sgmnt_data",604,"len")=8
	Set gtmtypes("sgmnt_data",604,"type")="uint64_t"
	Set gtmtypfldindx("sgmnt_data","gvstats_rec.f_trans_wait")=604
	Set gtmtypes("sgmnt_data",605,"name")="sgmnt_data.gvstats_rec.f_util_wait"
	Set gtmtypes("sgmnt_data",605,"off")=7272
	Set gtmtypes("sgmnt_data",605,"len")=8
	Set gtmtypes("sgmnt_data",605,"type")="uint64_t"
	Set gtmtypfldindx("sgmnt_data","gvstats_rec.f_util_wait")=605
	Set gtmtypes("sgmnt_data",606,"name")="sgmnt_data.gvstats_rec.f_ws2"
	Set gtmtypes("sgmnt_data",606,"off")=7280
	Set gtmtypes("sgmnt_data",606,"len")=8
	Set gtmtypes("sgmnt_data",606,"type")="uint64_t"
	Set gtmtypfldindx("sgmnt_data","gvstats_rec.f_ws2")=606
	Set gtmtypes("sgmnt_data",607,"name")="sgmnt_data.gvstats_rec.f_ws12"
	Set gtmtypes("sgmnt_data",607,"off")=7288
	Set gtmtypes("sgmnt_data",607,"len")=8
	Set gtmtypes("sgmnt_data",607,"type")="uint64_t"
	Set gtmtypfldindx("sgmnt_data","gvstats_rec.f_ws12")=607
	Set gtmtypes("sgmnt_data",608,"name")="sgmnt_data.gvstats_rec.f_ws15"
	Set gtmtypes("sgmnt_data",608,"off")=7296
	Set gtmtypes("sgmnt_data",608,"len")=8
	Set gtmtypes("sgmnt_data",608,"type")="uint64_t"
	Set gtmtypfldindx("sgmnt_data","gvstats_rec.f_ws15")=608
	Set gtmtypes("sgmnt_data",609,"name")="sgmnt_data.gvstats_rec.f_ws39"
	Set gtmtypes("sgmnt_data",609,"off")=7304
	Set gtmtypes("sgmnt_data",609,"len")=8
	Set gtmtypes("sgmnt_data",609,"type")="uint64_t"
	Set gtmtypfldindx("sgmnt_data","gvstats_rec.f_ws39")=609
	Set gtmtypes("sgmnt_data",610,"name")="sgmnt_data.gvstats_rec.f_ws47"
	Set gtmtypes("sgmnt_data",610,"off")=7312
	Set gtmtypes("sgmnt_data",610,"len")=8
	Set gtmtypes("sgmnt_data",610,"type")="uint64_t"
	Set gtmtypfldindx("sgmnt_data","gvstats_rec.f_ws47")=610
	Set gtmtypes("sgmnt_data",611,"name")="sgmnt_data.filler_8k"
	Set gtmtypes("sgmnt_data",611,"off")=7320
	Set gtmtypes("sgmnt_data",611,"len")=872
	Set gtmtypes("sgmnt_data",611,"type")="char"
	Set gtmtypfldindx("sgmnt_data","filler_8k")=611
	;
	Set gtmtypes("shm_forw_multi_t")="struct"
	Set gtmtypes("shm_forw_multi_t",0)=20
	Set gtmtypes("shm_forw_multi_t","len")=80
	Set gtmtypes("shm_forw_multi_t",1,"name")="shm_forw_multi_t.free_chain"
	Set gtmtypes("shm_forw_multi_t",1,"off")=0
	Set gtmtypes("shm_forw_multi_t",1,"len")=16
	Set gtmtypes("shm_forw_multi_t",1,"type")="que_ent"
	Set gtmtypfldindx("shm_forw_multi_t","free_chain")=1
	Set gtmtypes("shm_forw_multi_t",2,"name")="shm_forw_multi_t.free_chain.fl"
	Set gtmtypes("shm_forw_multi_t",2,"off")=0
	Set gtmtypes("shm_forw_multi_t",2,"len")=8
	Set gtmtypes("shm_forw_multi_t",2,"type")="intptr_t"
	Set gtmtypfldindx("shm_forw_multi_t","free_chain.fl")=2
	Set gtmtypes("shm_forw_multi_t",3,"name")="shm_forw_multi_t.free_chain.bl"
	Set gtmtypes("shm_forw_multi_t",3,"off")=8
	Set gtmtypes("shm_forw_multi_t",3,"len")=8
	Set gtmtypes("shm_forw_multi_t",3,"type")="intptr_t"
	Set gtmtypfldindx("shm_forw_multi_t","free_chain.bl")=3
	Set gtmtypes("shm_forw_multi_t",4,"name")="shm_forw_multi_t.same_hash_chain"
	Set gtmtypes("shm_forw_multi_t",4,"off")=16
	Set gtmtypes("shm_forw_multi_t",4,"len")=16
	Set gtmtypes("shm_forw_multi_t",4,"type")="que_ent"
	Set gtmtypfldindx("shm_forw_multi_t","same_hash_chain")=4
	Set gtmtypes("shm_forw_multi_t",5,"name")="shm_forw_multi_t.same_hash_chain.fl"
	Set gtmtypes("shm_forw_multi_t",5,"off")=16
	Set gtmtypes("shm_forw_multi_t",5,"len")=8
	Set gtmtypes("shm_forw_multi_t",5,"type")="intptr_t"
	Set gtmtypfldindx("shm_forw_multi_t","same_hash_chain.fl")=5
	Set gtmtypes("shm_forw_multi_t",6,"name")="shm_forw_multi_t.same_hash_chain.bl"
	Set gtmtypes("shm_forw_multi_t",6,"off")=24
	Set gtmtypes("shm_forw_multi_t",6,"len")=8
	Set gtmtypes("shm_forw_multi_t",6,"type")="intptr_t"
	Set gtmtypfldindx("shm_forw_multi_t","same_hash_chain.bl")=6
	Set gtmtypes("shm_forw_multi_t",7,"name")="shm_forw_multi_t.token"
	Set gtmtypes("shm_forw_multi_t",7,"off")=32
	Set gtmtypes("shm_forw_multi_t",7,"len")=8
	Set gtmtypes("shm_forw_multi_t",7,"type")="uint64_t"
	Set gtmtypfldindx("shm_forw_multi_t","token")=7
	Set gtmtypes("shm_forw_multi_t",8,"name")="shm_forw_multi_t.time"
	Set gtmtypes("shm_forw_multi_t",8,"off")=40
	Set gtmtypes("shm_forw_multi_t",8,"len")=4
	Set gtmtypes("shm_forw_multi_t",8,"type")="unsigned-int"
	Set gtmtypfldindx("shm_forw_multi_t","time")=8
	Set gtmtypes("shm_forw_multi_t",9,"name")="shm_forw_multi_t.recstat"
	Set gtmtypes("shm_forw_multi_t",9,"off")=44
	Set gtmtypes("shm_forw_multi_t",9,"len")=4
	Set gtmtypes("shm_forw_multi_t",9,"type")="int"
	Set gtmtypfldindx("shm_forw_multi_t","recstat")=9
	Set gtmtypes("shm_forw_multi_t",10,"name")="shm_forw_multi_t.num_reg_total"
	Set gtmtypes("shm_forw_multi_t",10,"off")=48
	Set gtmtypes("shm_forw_multi_t",10,"len")=4
	Set gtmtypes("shm_forw_multi_t",10,"type")="unsigned-int"
	Set gtmtypfldindx("shm_forw_multi_t","num_reg_total")=10
	Set gtmtypes("shm_forw_multi_t",11,"name")="shm_forw_multi_t.num_reg_seen_backward"
	Set gtmtypes("shm_forw_multi_t",11,"off")=52
	Set gtmtypes("shm_forw_multi_t",11,"len")=4
	Set gtmtypes("shm_forw_multi_t",11,"type")="unsigned-int"
	Set gtmtypfldindx("shm_forw_multi_t","num_reg_seen_backward")=11
	Set gtmtypes("shm_forw_multi_t",12,"name")="shm_forw_multi_t.num_reg_seen_forward"
	Set gtmtypes("shm_forw_multi_t",12,"off")=56
	Set gtmtypes("shm_forw_multi_t",12,"len")=4
	Set gtmtypes("shm_forw_multi_t",12,"type")="unsigned-int"
	Set gtmtypfldindx("shm_forw_multi_t","num_reg_seen_forward")=12
	Set gtmtypes("shm_forw_multi_t",13,"name")="shm_forw_multi_t.num_procs"
	Set gtmtypes("shm_forw_multi_t",13,"off")=60
	Set gtmtypes("shm_forw_multi_t",13,"len")=4
	Set gtmtypes("shm_forw_multi_t",13,"type")="unsigned-int"
	Set gtmtypfldindx("shm_forw_multi_t","num_procs")=13
	Set gtmtypes("shm_forw_multi_t",14,"name")="shm_forw_multi_t.hash_index"
	Set gtmtypes("shm_forw_multi_t",14,"off")=64
	Set gtmtypes("shm_forw_multi_t",14,"len")=4
	Set gtmtypes("shm_forw_multi_t",14,"type")="unsigned-int"
	Set gtmtypfldindx("shm_forw_multi_t","hash_index")=14
	Set gtmtypes("shm_forw_multi_t",15,"name")="shm_forw_multi_t.mur_latch"
	Set gtmtypes("shm_forw_multi_t",15,"off")=72
	Set gtmtypes("shm_forw_multi_t",15,"len")=8
	Set gtmtypes("shm_forw_multi_t",15,"type")="global_latch_t"
	Set gtmtypfldindx("shm_forw_multi_t","mur_latch")=15
	Set gtmtypes("shm_forw_multi_t",16,"name")="shm_forw_multi_t.mur_latch.u"
	Set gtmtypes("shm_forw_multi_t",16,"off")=72
	Set gtmtypes("shm_forw_multi_t",16,"len")=8
	Set gtmtypes("shm_forw_multi_t",16,"type")="union"
	Set gtmtypfldindx("shm_forw_multi_t","mur_latch.u")=16
	Set gtmtypes("shm_forw_multi_t",17,"name")="shm_forw_multi_t.mur_latch.u.pid_imgcnt"
	Set gtmtypes("shm_forw_multi_t",17,"off")=72
	Set gtmtypes("shm_forw_multi_t",17,"len")=8
	Set gtmtypes("shm_forw_multi_t",17,"type")="uint64_t"
	Set gtmtypfldindx("shm_forw_multi_t","mur_latch.u.pid_imgcnt")=17
	Set gtmtypes("shm_forw_multi_t",18,"name")="shm_forw_multi_t.mur_latch.u.parts"
	Set gtmtypes("shm_forw_multi_t",18,"off")=72
	Set gtmtypes("shm_forw_multi_t",18,"len")=8
	Set gtmtypes("shm_forw_multi_t",18,"type")="struct"
	Set gtmtypfldindx("shm_forw_multi_t","mur_latch.u.parts")=18
	Set gtmtypes("shm_forw_multi_t",19,"name")="shm_forw_multi_t.mur_latch.u.parts.latch_pid"
	Set gtmtypes("shm_forw_multi_t",19,"off")=72
	Set gtmtypes("shm_forw_multi_t",19,"len")=4
	Set gtmtypes("shm_forw_multi_t",19,"type")="int"
	Set gtmtypfldindx("shm_forw_multi_t","mur_latch.u.parts.latch_pid")=19
	Set gtmtypes("shm_forw_multi_t",20,"name")="shm_forw_multi_t.mur_latch.u.parts.latch_word"
	Set gtmtypes("shm_forw_multi_t",20,"off")=76
	Set gtmtypes("shm_forw_multi_t",20,"len")=4
	Set gtmtypes("shm_forw_multi_t",20,"type")="int"
	Set gtmtypfldindx("shm_forw_multi_t","mur_latch.u.parts.latch_word")=20
	;
	Set gtmtypes("shm_parms")="struct"
	Set gtmtypes("shm_parms",0)=3
	Set gtmtypes("shm_parms","len")=16
	Set gtmtypes("shm_parms",1,"name")="shm_parms.sgmnt_siz"
	Set gtmtypes("shm_parms",1,"off")=0
	Set gtmtypes("shm_parms",1,"len")=8
	Set gtmtypes("shm_parms",1,"type")="ssize_t"
	Set gtmtypfldindx("shm_parms","sgmnt_siz")=1
	Set gtmtypes("shm_parms",2,"name")="shm_parms.shmid"
	Set gtmtypes("shm_parms",2,"off")=8
	Set gtmtypes("shm_parms",2,"len")=4
	Set gtmtypes("shm_parms",2,"type")="int"
	Set gtmtypfldindx("shm_parms","shmid")=2
	Set gtmtypes("shm_parms",3,"name")="shm_parms.key"
	Set gtmtypes("shm_parms",3,"off")=12
	Set gtmtypes("shm_parms",3,"len")=4
	Set gtmtypes("shm_parms",3,"type")="key_t"
	Set gtmtypfldindx("shm_parms","key")=3
	;
	Set gtmtypes("shm_reg_ctl_t")="struct"
	Set gtmtypes("shm_reg_ctl_t",0)=9
	Set gtmtypes("shm_reg_ctl_t","len")=64
	Set gtmtypes("shm_reg_ctl_t",1,"name")="shm_reg_ctl_t.shm_forw_multi"
	Set gtmtypes("shm_reg_ctl_t",1,"off")=0
	Set gtmtypes("shm_reg_ctl_t",1,"len")=8
	Set gtmtypes("shm_reg_ctl_t",1,"type")="addr"
	Set gtmtypfldindx("shm_reg_ctl_t","shm_forw_multi")=1
	Set gtmtypes("shm_reg_ctl_t",2,"name")="shm_reg_ctl_t.consist_jnl_seqno"
	Set gtmtypes("shm_reg_ctl_t",2,"off")=8
	Set gtmtypes("shm_reg_ctl_t",2,"len")=8
	Set gtmtypes("shm_reg_ctl_t",2,"type")="uint64_t"
	Set gtmtypfldindx("shm_reg_ctl_t","consist_jnl_seqno")=2
	Set gtmtypes("shm_reg_ctl_t",3,"name")="shm_reg_ctl_t.jnlext_shm_size"
	Set gtmtypes("shm_reg_ctl_t",3,"off")=16
	Set gtmtypes("shm_reg_ctl_t",3,"len")=8
	Set gtmtypes("shm_reg_ctl_t",3,"type")="size_t"
	Set gtmtypfldindx("shm_reg_ctl_t","jnlext_shm_size")=3
	Set gtmtypes("shm_reg_ctl_t",4,"name")="shm_reg_ctl_t.owning_pid"
	Set gtmtypes("shm_reg_ctl_t",4,"off")=24
	Set gtmtypes("shm_reg_ctl_t",4,"len")=4
	Set gtmtypes("shm_reg_ctl_t",4,"type")="pid_t"
	Set gtmtypfldindx("shm_reg_ctl_t","owning_pid")=4
	Set gtmtypes("shm_reg_ctl_t",5,"name")="shm_reg_ctl_t.err_cnt"
	Set gtmtypes("shm_reg_ctl_t",5,"off")=28
	Set gtmtypes("shm_reg_ctl_t",5,"len")=4
	Set gtmtypes("shm_reg_ctl_t",5,"type")="int"
	Set gtmtypfldindx("shm_reg_ctl_t","err_cnt")=5
	Set gtmtypes("shm_reg_ctl_t",6,"name")="shm_reg_ctl_t.wrn_count"
	Set gtmtypes("shm_reg_ctl_t",6,"off")=32
	Set gtmtypes("shm_reg_ctl_t",6,"len")=4
	Set gtmtypes("shm_reg_ctl_t",6,"type")="int"
	Set gtmtypfldindx("shm_reg_ctl_t","wrn_count")=6
	Set gtmtypes("shm_reg_ctl_t",7,"name")="shm_reg_ctl_t.extr_file_created"
	Set gtmtypes("shm_reg_ctl_t",7,"off")=36
	Set gtmtypes("shm_reg_ctl_t",7,"len")=12
	Set gtmtypes("shm_reg_ctl_t",7,"type")="boolean_t"
	Set gtmtypfldindx("shm_reg_ctl_t","extr_file_created")=7
	Set gtmtypes("shm_reg_ctl_t",7,"dim")=3
	Set gtmtypes("shm_reg_ctl_t",8,"name")="shm_reg_ctl_t.jnlext_shmid"
	Set gtmtypes("shm_reg_ctl_t",8,"off")=48
	Set gtmtypes("shm_reg_ctl_t",8,"len")=4
	Set gtmtypes("shm_reg_ctl_t",8,"type")="int"
	Set gtmtypfldindx("shm_reg_ctl_t","jnlext_shmid")=8
	Set gtmtypes("shm_reg_ctl_t",9,"name")="shm_reg_ctl_t.jnlext_list_size"
	Set gtmtypes("shm_reg_ctl_t",9,"off")=52
	Set gtmtypes("shm_reg_ctl_t",9,"len")=12
	Set gtmtypes("shm_reg_ctl_t",9,"type")="int"
	Set gtmtypfldindx("shm_reg_ctl_t","jnlext_list_size")=9
	Set gtmtypes("shm_reg_ctl_t",9,"dim")=3
	;
	Set gtmtypes("shm_snapshot_t")="struct"
	Set gtmtypes("shm_snapshot_t",0)=633
	Set gtmtypes("shm_snapshot_t","len")=12376
	Set gtmtypes("shm_snapshot_t",1,"name")="shm_snapshot_t.ss_info"
	Set gtmtypes("shm_snapshot_t",1,"off")=0
	Set gtmtypes("shm_snapshot_t",1,"len")=4152
	Set gtmtypes("shm_snapshot_t",1,"type")="snapshot_info_t"
	Set gtmtypfldindx("shm_snapshot_t","ss_info")=1
	Set gtmtypes("shm_snapshot_t",2,"name")="shm_snapshot_t.ss_info.ss_pid"
	Set gtmtypes("shm_snapshot_t",2,"off")=0
	Set gtmtypes("shm_snapshot_t",2,"len")=4
	Set gtmtypes("shm_snapshot_t",2,"type")="unsigned-int"
	Set gtmtypfldindx("shm_snapshot_t","ss_info.ss_pid")=2
	Set gtmtypes("shm_snapshot_t",3,"name")="shm_snapshot_t.ss_info.snapshot_tn"
	Set gtmtypes("shm_snapshot_t",3,"off")=8
	Set gtmtypes("shm_snapshot_t",3,"len")=8
	Set gtmtypes("shm_snapshot_t",3,"type")="uint64_t"
	Set gtmtypfldindx("shm_snapshot_t","ss_info.snapshot_tn")=3
	Set gtmtypes("shm_snapshot_t",4,"name")="shm_snapshot_t.ss_info.db_blk_size"
	Set gtmtypes("shm_snapshot_t",4,"off")=16
	Set gtmtypes("shm_snapshot_t",4,"len")=4
	Set gtmtypes("shm_snapshot_t",4,"type")="unsigned-int"
	Set gtmtypfldindx("shm_snapshot_t","ss_info.db_blk_size")=4
	Set gtmtypes("shm_snapshot_t",5,"name")="shm_snapshot_t.ss_info.free_blks"
	Set gtmtypes("shm_snapshot_t",5,"off")=20
	Set gtmtypes("shm_snapshot_t",5,"len")=4
	Set gtmtypes("shm_snapshot_t",5,"type")="unsigned-int"
	Set gtmtypfldindx("shm_snapshot_t","ss_info.free_blks")=5
	Set gtmtypes("shm_snapshot_t",6,"name")="shm_snapshot_t.ss_info.total_blks"
	Set gtmtypes("shm_snapshot_t",6,"off")=24
	Set gtmtypes("shm_snapshot_t",6,"len")=4
	Set gtmtypes("shm_snapshot_t",6,"type")="unsigned-int"
	Set gtmtypfldindx("shm_snapshot_t","ss_info.total_blks")=6
	Set gtmtypes("shm_snapshot_t",7,"name")="shm_snapshot_t.ss_info.shadow_file"
	Set gtmtypes("shm_snapshot_t",7,"off")=28
	Set gtmtypes("shm_snapshot_t",7,"len")=4097
	Set gtmtypes("shm_snapshot_t",7,"type")="char"
	Set gtmtypfldindx("shm_snapshot_t","ss_info.shadow_file")=7
	Set gtmtypes("shm_snapshot_t",8,"name")="shm_snapshot_t.ss_info.shadow_vbn"
	Set gtmtypes("shm_snapshot_t",8,"off")=4128
	Set gtmtypes("shm_snapshot_t",8,"len")=4
	Set gtmtypes("shm_snapshot_t",8,"type")="int"
	Set gtmtypfldindx("shm_snapshot_t","ss_info.shadow_vbn")=8
	Set gtmtypes("shm_snapshot_t",9,"name")="shm_snapshot_t.ss_info.ss_shmid"
	Set gtmtypes("shm_snapshot_t",9,"off")=4136
	Set gtmtypes("shm_snapshot_t",9,"len")=8
	Set gtmtypes("shm_snapshot_t",9,"type")="long"
	Set gtmtypfldindx("shm_snapshot_t","ss_info.ss_shmid")=9
	Set gtmtypes("shm_snapshot_t",10,"name")="shm_snapshot_t.ss_info.ss_shmsize"
	Set gtmtypes("shm_snapshot_t",10,"off")=4144
	Set gtmtypes("shm_snapshot_t",10,"len")=4
	Set gtmtypes("shm_snapshot_t",10,"type")="int"
	Set gtmtypfldindx("shm_snapshot_t","ss_info.ss_shmsize")=10
	Set gtmtypes("shm_snapshot_t",11,"name")="shm_snapshot_t.failure_errno"
	Set gtmtypes("shm_snapshot_t",11,"off")=4152
	Set gtmtypes("shm_snapshot_t",11,"len")=4
	Set gtmtypes("shm_snapshot_t",11,"type")="unsigned-int"
	Set gtmtypfldindx("shm_snapshot_t","failure_errno")=11
	Set gtmtypes("shm_snapshot_t",12,"name")="shm_snapshot_t.failed_pid"
	Set gtmtypes("shm_snapshot_t",12,"off")=4156
	Set gtmtypes("shm_snapshot_t",12,"len")=4
	Set gtmtypes("shm_snapshot_t",12,"type")="pid_t"
	Set gtmtypfldindx("shm_snapshot_t","failed_pid")=12
	Set gtmtypes("shm_snapshot_t",13,"name")="shm_snapshot_t.in_use"
	Set gtmtypes("shm_snapshot_t",13,"off")=4160
	Set gtmtypes("shm_snapshot_t",13,"len")=4
	Set gtmtypes("shm_snapshot_t",13,"type")="int"
	Set gtmtypfldindx("shm_snapshot_t","in_use")=13
	Set gtmtypes("shm_snapshot_t",14,"name")="shm_snapshot_t.preserve_snapshot"
	Set gtmtypes("shm_snapshot_t",14,"off")=4164
	Set gtmtypes("shm_snapshot_t",14,"len")=4
	Set gtmtypes("shm_snapshot_t",14,"type")="boolean_t"
	Set gtmtypfldindx("shm_snapshot_t","preserve_snapshot")=14
	Set gtmtypes("shm_snapshot_t",15,"name")="shm_snapshot_t.bitmap_latch"
	Set gtmtypes("shm_snapshot_t",15,"off")=4168
	Set gtmtypes("shm_snapshot_t",15,"len")=8
	Set gtmtypes("shm_snapshot_t",15,"type")="global_latch_t"
	Set gtmtypfldindx("shm_snapshot_t","bitmap_latch")=15
	Set gtmtypes("shm_snapshot_t",16,"name")="shm_snapshot_t.bitmap_latch.u"
	Set gtmtypes("shm_snapshot_t",16,"off")=4168
	Set gtmtypes("shm_snapshot_t",16,"len")=8
	Set gtmtypes("shm_snapshot_t",16,"type")="union"
	Set gtmtypfldindx("shm_snapshot_t","bitmap_latch.u")=16
	Set gtmtypes("shm_snapshot_t",17,"name")="shm_snapshot_t.bitmap_latch.u.pid_imgcnt"
	Set gtmtypes("shm_snapshot_t",17,"off")=4168
	Set gtmtypes("shm_snapshot_t",17,"len")=8
	Set gtmtypes("shm_snapshot_t",17,"type")="uint64_t"
	Set gtmtypfldindx("shm_snapshot_t","bitmap_latch.u.pid_imgcnt")=17
	Set gtmtypes("shm_snapshot_t",18,"name")="shm_snapshot_t.bitmap_latch.u.parts"
	Set gtmtypes("shm_snapshot_t",18,"off")=4168
	Set gtmtypes("shm_snapshot_t",18,"len")=8
	Set gtmtypes("shm_snapshot_t",18,"type")="struct"
	Set gtmtypfldindx("shm_snapshot_t","bitmap_latch.u.parts")=18
	Set gtmtypes("shm_snapshot_t",19,"name")="shm_snapshot_t.bitmap_latch.u.parts.latch_pid"
	Set gtmtypes("shm_snapshot_t",19,"off")=4168
	Set gtmtypes("shm_snapshot_t",19,"len")=4
	Set gtmtypes("shm_snapshot_t",19,"type")="int"
	Set gtmtypfldindx("shm_snapshot_t","bitmap_latch.u.parts.latch_pid")=19
	Set gtmtypes("shm_snapshot_t",20,"name")="shm_snapshot_t.bitmap_latch.u.parts.latch_word"
	Set gtmtypes("shm_snapshot_t",20,"off")=4172
	Set gtmtypes("shm_snapshot_t",20,"len")=4
	Set gtmtypes("shm_snapshot_t",20,"type")="int"
	Set gtmtypfldindx("shm_snapshot_t","bitmap_latch.u.parts.latch_word")=20
	Set gtmtypes("shm_snapshot_t",21,"name")="shm_snapshot_t.ss_tn_count"
	Set gtmtypes("shm_snapshot_t",21,"off")=4176
	Set gtmtypes("shm_snapshot_t",21,"len")=8
	Set gtmtypes("shm_snapshot_t",21,"type")="uint64_t"
	Set gtmtypfldindx("shm_snapshot_t","ss_tn_count")=21
	Set gtmtypes("shm_snapshot_t",22,"name")="shm_snapshot_t.shadow_file_header"
	Set gtmtypes("shm_snapshot_t",22,"off")=4184
	Set gtmtypes("shm_snapshot_t",22,"len")=8192
	Set gtmtypes("shm_snapshot_t",22,"type")="sgmnt_data"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header")=22
	Set gtmtypes("shm_snapshot_t",23,"name")="shm_snapshot_t.shadow_file_header.label"
	Set gtmtypes("shm_snapshot_t",23,"off")=4184
	Set gtmtypes("shm_snapshot_t",23,"len")=12
	Set gtmtypes("shm_snapshot_t",23,"type")="unsigned-char"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.label")=23
	Set gtmtypes("shm_snapshot_t",24,"name")="shm_snapshot_t.shadow_file_header.blk_size"
	Set gtmtypes("shm_snapshot_t",24,"off")=4196
	Set gtmtypes("shm_snapshot_t",24,"len")=4
	Set gtmtypes("shm_snapshot_t",24,"type")="int"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.blk_size")=24
	Set gtmtypes("shm_snapshot_t",25,"name")="shm_snapshot_t.shadow_file_header.master_map_len"
	Set gtmtypes("shm_snapshot_t",25,"off")=4200
	Set gtmtypes("shm_snapshot_t",25,"len")=4
	Set gtmtypes("shm_snapshot_t",25,"type")="int"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.master_map_len")=25
	Set gtmtypes("shm_snapshot_t",26,"name")="shm_snapshot_t.shadow_file_header.bplmap"
	Set gtmtypes("shm_snapshot_t",26,"off")=4204
	Set gtmtypes("shm_snapshot_t",26,"len")=4
	Set gtmtypes("shm_snapshot_t",26,"type")="int"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.bplmap")=26
	Set gtmtypes("shm_snapshot_t",27,"name")="shm_snapshot_t.shadow_file_header.start_vbn"
	Set gtmtypes("shm_snapshot_t",27,"off")=4208
	Set gtmtypes("shm_snapshot_t",27,"len")=4
	Set gtmtypes("shm_snapshot_t",27,"type")="int"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.start_vbn")=27
	Set gtmtypes("shm_snapshot_t",28,"name")="shm_snapshot_t.shadow_file_header.acc_meth"
	Set gtmtypes("shm_snapshot_t",28,"off")=4212
	Set gtmtypes("shm_snapshot_t",28,"len")=4
	Set gtmtypes("shm_snapshot_t",28,"type")="int"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.acc_meth")=28
	Set gtmtypes("shm_snapshot_t",29,"name")="shm_snapshot_t.shadow_file_header.max_bts"
	Set gtmtypes("shm_snapshot_t",29,"off")=4216
	Set gtmtypes("shm_snapshot_t",29,"len")=4
	Set gtmtypes("shm_snapshot_t",29,"type")="unsigned-int"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.max_bts")=29
	Set gtmtypes("shm_snapshot_t",30,"name")="shm_snapshot_t.shadow_file_header.n_bts"
	Set gtmtypes("shm_snapshot_t",30,"off")=4220
	Set gtmtypes("shm_snapshot_t",30,"len")=4
	Set gtmtypes("shm_snapshot_t",30,"type")="int"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.n_bts")=30
	Set gtmtypes("shm_snapshot_t",31,"name")="shm_snapshot_t.shadow_file_header.bt_buckets"
	Set gtmtypes("shm_snapshot_t",31,"off")=4224
	Set gtmtypes("shm_snapshot_t",31,"len")=4
	Set gtmtypes("shm_snapshot_t",31,"type")="int"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.bt_buckets")=31
	Set gtmtypes("shm_snapshot_t",32,"name")="shm_snapshot_t.shadow_file_header.reserved_bytes"
	Set gtmtypes("shm_snapshot_t",32,"off")=4228
	Set gtmtypes("shm_snapshot_t",32,"len")=4
	Set gtmtypes("shm_snapshot_t",32,"type")="int"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.reserved_bytes")=32
	Set gtmtypes("shm_snapshot_t",33,"name")="shm_snapshot_t.shadow_file_header.max_rec_size"
	Set gtmtypes("shm_snapshot_t",33,"off")=4232
	Set gtmtypes("shm_snapshot_t",33,"len")=4
	Set gtmtypes("shm_snapshot_t",33,"type")="int"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.max_rec_size")=33
	Set gtmtypes("shm_snapshot_t",34,"name")="shm_snapshot_t.shadow_file_header.max_key_size"
	Set gtmtypes("shm_snapshot_t",34,"off")=4236
	Set gtmtypes("shm_snapshot_t",34,"len")=4
	Set gtmtypes("shm_snapshot_t",34,"type")="int"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.max_key_size")=34
	Set gtmtypes("shm_snapshot_t",35,"name")="shm_snapshot_t.shadow_file_header.lock_space_size"
	Set gtmtypes("shm_snapshot_t",35,"off")=4240
	Set gtmtypes("shm_snapshot_t",35,"len")=4
	Set gtmtypes("shm_snapshot_t",35,"type")="unsigned-int"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.lock_space_size")=35
	Set gtmtypes("shm_snapshot_t",36,"name")="shm_snapshot_t.shadow_file_header.extension_size"
	Set gtmtypes("shm_snapshot_t",36,"off")=4244
	Set gtmtypes("shm_snapshot_t",36,"len")=4
	Set gtmtypes("shm_snapshot_t",36,"type")="unsigned-int"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.extension_size")=36
	Set gtmtypes("shm_snapshot_t",37,"name")="shm_snapshot_t.shadow_file_header.def_coll"
	Set gtmtypes("shm_snapshot_t",37,"off")=4248
	Set gtmtypes("shm_snapshot_t",37,"len")=4
	Set gtmtypes("shm_snapshot_t",37,"type")="unsigned-int"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.def_coll")=37
	Set gtmtypes("shm_snapshot_t",38,"name")="shm_snapshot_t.shadow_file_header.def_coll_ver"
	Set gtmtypes("shm_snapshot_t",38,"off")=4252
	Set gtmtypes("shm_snapshot_t",38,"len")=4
	Set gtmtypes("shm_snapshot_t",38,"type")="unsigned-int"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.def_coll_ver")=38
	Set gtmtypes("shm_snapshot_t",39,"name")="shm_snapshot_t.shadow_file_header.std_null_coll"
	Set gtmtypes("shm_snapshot_t",39,"off")=4256
	Set gtmtypes("shm_snapshot_t",39,"len")=4
	Set gtmtypes("shm_snapshot_t",39,"type")="boolean_t"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.std_null_coll")=39
	Set gtmtypes("shm_snapshot_t",40,"name")="shm_snapshot_t.shadow_file_header.null_subs"
	Set gtmtypes("shm_snapshot_t",40,"off")=4260
	Set gtmtypes("shm_snapshot_t",40,"len")=4
	Set gtmtypes("shm_snapshot_t",40,"type")="int"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.null_subs")=40
	Set gtmtypes("shm_snapshot_t",41,"name")="shm_snapshot_t.shadow_file_header.free_space"
	Set gtmtypes("shm_snapshot_t",41,"off")=4264
	Set gtmtypes("shm_snapshot_t",41,"len")=4
	Set gtmtypes("shm_snapshot_t",41,"type")="unsigned-int"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.free_space")=41
	Set gtmtypes("shm_snapshot_t",42,"name")="shm_snapshot_t.shadow_file_header.mutex_spin_parms"
	Set gtmtypes("shm_snapshot_t",42,"off")=4268
	Set gtmtypes("shm_snapshot_t",42,"len")=16
	Set gtmtypes("shm_snapshot_t",42,"type")="mutex_spin_parms_struct"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.mutex_spin_parms")=42
	Set gtmtypes("shm_snapshot_t",43,"name")="shm_snapshot_t.shadow_file_header.mutex_spin_parms.mutex_hard_spin_count"
	Set gtmtypes("shm_snapshot_t",43,"off")=4268
	Set gtmtypes("shm_snapshot_t",43,"len")=4
	Set gtmtypes("shm_snapshot_t",43,"type")="int"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.mutex_spin_parms.mutex_hard_spin_count")=43
	Set gtmtypes("shm_snapshot_t",44,"name")="shm_snapshot_t.shadow_file_header.mutex_spin_parms.mutex_sleep_spin_count"
	Set gtmtypes("shm_snapshot_t",44,"off")=4272
	Set gtmtypes("shm_snapshot_t",44,"len")=4
	Set gtmtypes("shm_snapshot_t",44,"type")="int"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.mutex_spin_parms.mutex_sleep_spin_count")=44
	Set gtmtypes("shm_snapshot_t",45,"name")="shm_snapshot_t.shadow_file_header.mutex_spin_parms.mutex_spin_sleep_mask"
	Set gtmtypes("shm_snapshot_t",45,"off")=4276
	Set gtmtypes("shm_snapshot_t",45,"len")=4
	Set gtmtypes("shm_snapshot_t",45,"type")="int"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.mutex_spin_parms.mutex_spin_sleep_mask")=45
	Set gtmtypes("shm_snapshot_t",46,"name")="shm_snapshot_t.shadow_file_header.mutex_spin_parms.mutex_que_entry_space_size"
	Set gtmtypes("shm_snapshot_t",46,"off")=4280
	Set gtmtypes("shm_snapshot_t",46,"len")=4
	Set gtmtypes("shm_snapshot_t",46,"type")="int"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.mutex_spin_parms.mutex_que_entry_space_size")=46
	Set gtmtypes("shm_snapshot_t",47,"name")="shm_snapshot_t.shadow_file_header.max_update_array_size"
	Set gtmtypes("shm_snapshot_t",47,"off")=4284
	Set gtmtypes("shm_snapshot_t",47,"len")=4
	Set gtmtypes("shm_snapshot_t",47,"type")="int"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.max_update_array_size")=47
	Set gtmtypes("shm_snapshot_t",48,"name")="shm_snapshot_t.shadow_file_header.max_non_bm_update_array_size"
	Set gtmtypes("shm_snapshot_t",48,"off")=4288
	Set gtmtypes("shm_snapshot_t",48,"len")=4
	Set gtmtypes("shm_snapshot_t",48,"type")="int"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.max_non_bm_update_array_size")=48
	Set gtmtypes("shm_snapshot_t",49,"name")="shm_snapshot_t.shadow_file_header.file_corrupt"
	Set gtmtypes("shm_snapshot_t",49,"off")=4292
	Set gtmtypes("shm_snapshot_t",49,"len")=4
	Set gtmtypes("shm_snapshot_t",49,"type")="boolean_t"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.file_corrupt")=49
	Set gtmtypes("shm_snapshot_t",50,"name")="shm_snapshot_t.shadow_file_header.minor_dbver"
	Set gtmtypes("shm_snapshot_t",50,"off")=4296
	Set gtmtypes("shm_snapshot_t",50,"len")=4
	Set gtmtypes("shm_snapshot_t",50,"type")="int"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.minor_dbver")=50
	Set gtmtypes("shm_snapshot_t",51,"name")="shm_snapshot_t.shadow_file_header.jnl_checksum"
	Set gtmtypes("shm_snapshot_t",51,"off")=4300
	Set gtmtypes("shm_snapshot_t",51,"len")=4
	Set gtmtypes("shm_snapshot_t",51,"type")="unsigned-int"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.jnl_checksum")=51
	Set gtmtypes("shm_snapshot_t",52,"name")="shm_snapshot_t.shadow_file_header.wcs_phase2_commit_wait_spincnt"
	Set gtmtypes("shm_snapshot_t",52,"off")=4304
	Set gtmtypes("shm_snapshot_t",52,"len")=4
	Set gtmtypes("shm_snapshot_t",52,"type")="unsigned-int"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.wcs_phase2_commit_wait_spincnt")=52
	Set gtmtypes("shm_snapshot_t",53,"name")="shm_snapshot_t.shadow_file_header.last_mdb_ver"
	Set gtmtypes("shm_snapshot_t",53,"off")=4308
	Set gtmtypes("shm_snapshot_t",53,"len")=4
	Set gtmtypes("shm_snapshot_t",53,"type")="int"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.last_mdb_ver")=53
	Set gtmtypes("shm_snapshot_t",54,"name")="shm_snapshot_t.shadow_file_header.filler_created"
	Set gtmtypes("shm_snapshot_t",54,"off")=4312
	Set gtmtypes("shm_snapshot_t",54,"len")=52
	Set gtmtypes("shm_snapshot_t",54,"type")="char"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.filler_created")=54
	Set gtmtypes("shm_snapshot_t",55,"name")="shm_snapshot_t.shadow_file_header.createinprogress"
	Set gtmtypes("shm_snapshot_t",55,"off")=4364
	Set gtmtypes("shm_snapshot_t",55,"len")=4
	Set gtmtypes("shm_snapshot_t",55,"type")="boolean_t"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.createinprogress")=55
	Set gtmtypes("shm_snapshot_t",56,"name")="shm_snapshot_t.shadow_file_header.creation_time4"
	Set gtmtypes("shm_snapshot_t",56,"off")=4368
	Set gtmtypes("shm_snapshot_t",56,"len")=4
	Set gtmtypes("shm_snapshot_t",56,"type")="int"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.creation_time4")=56
	Set gtmtypes("shm_snapshot_t",57,"name")="shm_snapshot_t.shadow_file_header.reservedDBFlags"
	Set gtmtypes("shm_snapshot_t",57,"off")=4372
	Set gtmtypes("shm_snapshot_t",57,"len")=4
	Set gtmtypes("shm_snapshot_t",57,"type")="unsigned-int"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.reservedDBFlags")=57
	Set gtmtypes("shm_snapshot_t",58,"name")="shm_snapshot_t.shadow_file_header.max_tn"
	Set gtmtypes("shm_snapshot_t",58,"off")=4376
	Set gtmtypes("shm_snapshot_t",58,"len")=8
	Set gtmtypes("shm_snapshot_t",58,"type")="uint64_t"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.max_tn")=58
	Set gtmtypes("shm_snapshot_t",59,"name")="shm_snapshot_t.shadow_file_header.max_tn_warn"
	Set gtmtypes("shm_snapshot_t",59,"off")=4384
	Set gtmtypes("shm_snapshot_t",59,"len")=8
	Set gtmtypes("shm_snapshot_t",59,"type")="uint64_t"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.max_tn_warn")=59
	Set gtmtypes("shm_snapshot_t",60,"name")="shm_snapshot_t.shadow_file_header.last_inc_backup"
	Set gtmtypes("shm_snapshot_t",60,"off")=4392
	Set gtmtypes("shm_snapshot_t",60,"len")=8
	Set gtmtypes("shm_snapshot_t",60,"type")="uint64_t"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.last_inc_backup")=60
	Set gtmtypes("shm_snapshot_t",61,"name")="shm_snapshot_t.shadow_file_header.last_com_backup"
	Set gtmtypes("shm_snapshot_t",61,"off")=4400
	Set gtmtypes("shm_snapshot_t",61,"len")=8
	Set gtmtypes("shm_snapshot_t",61,"type")="uint64_t"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.last_com_backup")=61
	Set gtmtypes("shm_snapshot_t",62,"name")="shm_snapshot_t.shadow_file_header.last_rec_backup"
	Set gtmtypes("shm_snapshot_t",62,"off")=4408
	Set gtmtypes("shm_snapshot_t",62,"len")=8
	Set gtmtypes("shm_snapshot_t",62,"type")="uint64_t"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.last_rec_backup")=62
	Set gtmtypes("shm_snapshot_t",63,"name")="shm_snapshot_t.shadow_file_header.last_inc_bkup_last_blk"
	Set gtmtypes("shm_snapshot_t",63,"off")=4416
	Set gtmtypes("shm_snapshot_t",63,"len")=4
	Set gtmtypes("shm_snapshot_t",63,"type")="int"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.last_inc_bkup_last_blk")=63
	Set gtmtypes("shm_snapshot_t",64,"name")="shm_snapshot_t.shadow_file_header.last_com_bkup_last_blk"
	Set gtmtypes("shm_snapshot_t",64,"off")=4420
	Set gtmtypes("shm_snapshot_t",64,"len")=4
	Set gtmtypes("shm_snapshot_t",64,"type")="int"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.last_com_bkup_last_blk")=64
	Set gtmtypes("shm_snapshot_t",65,"name")="shm_snapshot_t.shadow_file_header.last_rec_bkup_last_blk"
	Set gtmtypes("shm_snapshot_t",65,"off")=4424
	Set gtmtypes("shm_snapshot_t",65,"len")=4
	Set gtmtypes("shm_snapshot_t",65,"type")="int"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.last_rec_bkup_last_blk")=65
	Set gtmtypes("shm_snapshot_t",66,"name")="shm_snapshot_t.shadow_file_header.reorg_restart_block"
	Set gtmtypes("shm_snapshot_t",66,"off")=4428
	Set gtmtypes("shm_snapshot_t",66,"len")=4
	Set gtmtypes("shm_snapshot_t",66,"type")="int"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.reorg_restart_block")=66
	Set gtmtypes("shm_snapshot_t",67,"name")="shm_snapshot_t.shadow_file_header.filler_256"
	Set gtmtypes("shm_snapshot_t",67,"off")=4432
	Set gtmtypes("shm_snapshot_t",67,"len")=8
	Set gtmtypes("shm_snapshot_t",67,"type")="char"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.filler_256")=67
	Set gtmtypes("shm_snapshot_t",68,"name")="shm_snapshot_t.shadow_file_header.now_running"
	Set gtmtypes("shm_snapshot_t",68,"off")=4440
	Set gtmtypes("shm_snapshot_t",68,"len")=36
	Set gtmtypes("shm_snapshot_t",68,"type")="char"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.now_running")=68
	Set gtmtypes("shm_snapshot_t",69,"name")="shm_snapshot_t.shadow_file_header.filler_owner_node"
	Set gtmtypes("shm_snapshot_t",69,"off")=4476
	Set gtmtypes("shm_snapshot_t",69,"len")=4
	Set gtmtypes("shm_snapshot_t",69,"type")="unsigned-int"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.filler_owner_node")=69
	Set gtmtypes("shm_snapshot_t",70,"name")="shm_snapshot_t.shadow_file_header.image_count"
	Set gtmtypes("shm_snapshot_t",70,"off")=4480
	Set gtmtypes("shm_snapshot_t",70,"len")=4
	Set gtmtypes("shm_snapshot_t",70,"type")="unsigned-int"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.image_count")=70
	Set gtmtypes("shm_snapshot_t",71,"name")="shm_snapshot_t.shadow_file_header.freeze"
	Set gtmtypes("shm_snapshot_t",71,"off")=4484
	Set gtmtypes("shm_snapshot_t",71,"len")=4
	Set gtmtypes("shm_snapshot_t",71,"type")="unsigned-int"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.freeze")=71
	Set gtmtypes("shm_snapshot_t",72,"name")="shm_snapshot_t.shadow_file_header.kill_in_prog"
	Set gtmtypes("shm_snapshot_t",72,"off")=4488
	Set gtmtypes("shm_snapshot_t",72,"len")=4
	Set gtmtypes("shm_snapshot_t",72,"type")="int"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.kill_in_prog")=72
	Set gtmtypes("shm_snapshot_t",73,"name")="shm_snapshot_t.shadow_file_header.abandoned_kills"
	Set gtmtypes("shm_snapshot_t",73,"off")=4492
	Set gtmtypes("shm_snapshot_t",73,"len")=4
	Set gtmtypes("shm_snapshot_t",73,"type")="int"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.abandoned_kills")=73
	Set gtmtypes("shm_snapshot_t",74,"name")="shm_snapshot_t.shadow_file_header.unused_freeze_online_filler"
	Set gtmtypes("shm_snapshot_t",74,"off")=4496
	Set gtmtypes("shm_snapshot_t",74,"len")=4
	Set gtmtypes("shm_snapshot_t",74,"type")="unsigned-int"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.unused_freeze_online_filler")=74
	Set gtmtypes("shm_snapshot_t",75,"name")="shm_snapshot_t.shadow_file_header.filler_320"
	Set gtmtypes("shm_snapshot_t",75,"off")=4500
	Set gtmtypes("shm_snapshot_t",75,"len")=4
	Set gtmtypes("shm_snapshot_t",75,"type")="char"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.filler_320")=75
	Set gtmtypes("shm_snapshot_t",76,"name")="shm_snapshot_t.shadow_file_header.tn_upgrd_blks_0"
	Set gtmtypes("shm_snapshot_t",76,"off")=4504
	Set gtmtypes("shm_snapshot_t",76,"len")=8
	Set gtmtypes("shm_snapshot_t",76,"type")="uint64_t"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.tn_upgrd_blks_0")=76
	Set gtmtypes("shm_snapshot_t",77,"name")="shm_snapshot_t.shadow_file_header.desired_db_format_tn"
	Set gtmtypes("shm_snapshot_t",77,"off")=4512
	Set gtmtypes("shm_snapshot_t",77,"len")=8
	Set gtmtypes("shm_snapshot_t",77,"type")="uint64_t"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.desired_db_format_tn")=77
	Set gtmtypes("shm_snapshot_t",78,"name")="shm_snapshot_t.shadow_file_header.reorg_db_fmt_start_tn"
	Set gtmtypes("shm_snapshot_t",78,"off")=4520
	Set gtmtypes("shm_snapshot_t",78,"len")=8
	Set gtmtypes("shm_snapshot_t",78,"type")="uint64_t"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.reorg_db_fmt_start_tn")=78
	Set gtmtypes("shm_snapshot_t",79,"name")="shm_snapshot_t.shadow_file_header.reorg_upgrd_dwngrd_restart_block"
	Set gtmtypes("shm_snapshot_t",79,"off")=4528
	Set gtmtypes("shm_snapshot_t",79,"len")=4
	Set gtmtypes("shm_snapshot_t",79,"type")="int"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.reorg_upgrd_dwngrd_restart_block")=79
	Set gtmtypes("shm_snapshot_t",80,"name")="shm_snapshot_t.shadow_file_header.blks_to_upgrd"
	Set gtmtypes("shm_snapshot_t",80,"off")=4532
	Set gtmtypes("shm_snapshot_t",80,"len")=4
	Set gtmtypes("shm_snapshot_t",80,"type")="int"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.blks_to_upgrd")=80
	Set gtmtypes("shm_snapshot_t",81,"name")="shm_snapshot_t.shadow_file_header.blks_to_upgrd_subzero_error"
	Set gtmtypes("shm_snapshot_t",81,"off")=4536
	Set gtmtypes("shm_snapshot_t",81,"len")=4
	Set gtmtypes("shm_snapshot_t",81,"type")="int"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.blks_to_upgrd_subzero_error")=81
	Set gtmtypes("shm_snapshot_t",82,"name")="shm_snapshot_t.shadow_file_header.desired_db_format"
	Set gtmtypes("shm_snapshot_t",82,"off")=4540
	Set gtmtypes("shm_snapshot_t",82,"len")=4
	Set gtmtypes("shm_snapshot_t",82,"type")="int"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.desired_db_format")=82
	Set gtmtypes("shm_snapshot_t",83,"name")="shm_snapshot_t.shadow_file_header.fully_upgraded"
	Set gtmtypes("shm_snapshot_t",83,"off")=4544
	Set gtmtypes("shm_snapshot_t",83,"len")=4
	Set gtmtypes("shm_snapshot_t",83,"type")="boolean_t"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.fully_upgraded")=83
	Set gtmtypes("shm_snapshot_t",84,"name")="shm_snapshot_t.shadow_file_header.db_got_to_v5_once"
	Set gtmtypes("shm_snapshot_t",84,"off")=4548
	Set gtmtypes("shm_snapshot_t",84,"len")=4
	Set gtmtypes("shm_snapshot_t",84,"type")="boolean_t"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.db_got_to_v5_once")=84
	Set gtmtypes("shm_snapshot_t",85,"name")="shm_snapshot_t.shadow_file_header.opened_by_gtmv53"
	Set gtmtypes("shm_snapshot_t",85,"off")=4552
	Set gtmtypes("shm_snapshot_t",85,"len")=4
	Set gtmtypes("shm_snapshot_t",85,"type")="boolean_t"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.opened_by_gtmv53")=85
	Set gtmtypes("shm_snapshot_t",86,"name")="shm_snapshot_t.shadow_file_header.filler_384"
	Set gtmtypes("shm_snapshot_t",86,"off")=4556
	Set gtmtypes("shm_snapshot_t",86,"len")=12
	Set gtmtypes("shm_snapshot_t",86,"type")="char"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.filler_384")=86
	Set gtmtypes("shm_snapshot_t",87,"name")="shm_snapshot_t.shadow_file_header.trans_hist"
	Set gtmtypes("shm_snapshot_t",87,"off")=4568
	Set gtmtypes("shm_snapshot_t",87,"len")=56
	Set gtmtypes("shm_snapshot_t",87,"type")="th_index"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.trans_hist")=87
	Set gtmtypes("shm_snapshot_t",88,"name")="shm_snapshot_t.shadow_file_header.trans_hist.curr_tn"
	Set gtmtypes("shm_snapshot_t",88,"off")=4568
	Set gtmtypes("shm_snapshot_t",88,"len")=8
	Set gtmtypes("shm_snapshot_t",88,"type")="uint64_t"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.trans_hist.curr_tn")=88
	Set gtmtypes("shm_snapshot_t",89,"name")="shm_snapshot_t.shadow_file_header.trans_hist.early_tn"
	Set gtmtypes("shm_snapshot_t",89,"off")=4576
	Set gtmtypes("shm_snapshot_t",89,"len")=8
	Set gtmtypes("shm_snapshot_t",89,"type")="uint64_t"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.trans_hist.early_tn")=89
	Set gtmtypes("shm_snapshot_t",90,"name")="shm_snapshot_t.shadow_file_header.trans_hist.last_mm_sync"
	Set gtmtypes("shm_snapshot_t",90,"off")=4584
	Set gtmtypes("shm_snapshot_t",90,"len")=8
	Set gtmtypes("shm_snapshot_t",90,"type")="uint64_t"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.trans_hist.last_mm_sync")=90
	Set gtmtypes("shm_snapshot_t",91,"name")="shm_snapshot_t.shadow_file_header.trans_hist.filler_8byte"
	Set gtmtypes("shm_snapshot_t",91,"off")=4592
	Set gtmtypes("shm_snapshot_t",91,"len")=8
	Set gtmtypes("shm_snapshot_t",91,"type")="char"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.trans_hist.filler_8byte")=91
	Set gtmtypes("shm_snapshot_t",92,"name")="shm_snapshot_t.shadow_file_header.trans_hist.mm_tn"
	Set gtmtypes("shm_snapshot_t",92,"off")=4600
	Set gtmtypes("shm_snapshot_t",92,"len")=8
	Set gtmtypes("shm_snapshot_t",92,"type")="uint64_t"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.trans_hist.mm_tn")=92
	Set gtmtypes("shm_snapshot_t",93,"name")="shm_snapshot_t.shadow_file_header.trans_hist.lock_sequence"
	Set gtmtypes("shm_snapshot_t",93,"off")=4608
	Set gtmtypes("shm_snapshot_t",93,"len")=4
	Set gtmtypes("shm_snapshot_t",93,"type")="unsigned-int"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.trans_hist.lock_sequence")=93
	Set gtmtypes("shm_snapshot_t",94,"name")="shm_snapshot_t.shadow_file_header.trans_hist.ccp_jnl_filesize"
	Set gtmtypes("shm_snapshot_t",94,"off")=4612
	Set gtmtypes("shm_snapshot_t",94,"len")=4
	Set gtmtypes("shm_snapshot_t",94,"type")="unsigned-int"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.trans_hist.ccp_jnl_filesize")=94
	Set gtmtypes("shm_snapshot_t",95,"name")="shm_snapshot_t.shadow_file_header.trans_hist.total_blks"
	Set gtmtypes("shm_snapshot_t",95,"off")=4616
	Set gtmtypes("shm_snapshot_t",95,"len")=4
	Set gtmtypes("shm_snapshot_t",95,"type")="unsigned-int"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.trans_hist.total_blks")=95
	Set gtmtypes("shm_snapshot_t",96,"name")="shm_snapshot_t.shadow_file_header.trans_hist.free_blocks"
	Set gtmtypes("shm_snapshot_t",96,"off")=4620
	Set gtmtypes("shm_snapshot_t",96,"len")=4
	Set gtmtypes("shm_snapshot_t",96,"type")="unsigned-int"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.trans_hist.free_blocks")=96
	Set gtmtypes("shm_snapshot_t",97,"name")="shm_snapshot_t.shadow_file_header.write_fullblk"
	Set gtmtypes("shm_snapshot_t",97,"off")=4624
	Set gtmtypes("shm_snapshot_t",97,"len")=4
	Set gtmtypes("shm_snapshot_t",97,"type")="int"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.write_fullblk")=97
	Set gtmtypes("shm_snapshot_t",98,"name")="shm_snapshot_t.shadow_file_header.filler"
	Set gtmtypes("shm_snapshot_t",98,"off")=4628
	Set gtmtypes("shm_snapshot_t",98,"len")=4
	Set gtmtypes("shm_snapshot_t",98,"type")="char"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.filler")=98
	Set gtmtypes("shm_snapshot_t",99,"name")="shm_snapshot_t.shadow_file_header.flush_time"
	Set gtmtypes("shm_snapshot_t",99,"off")=4632
	Set gtmtypes("shm_snapshot_t",99,"len")=8
	Set gtmtypes("shm_snapshot_t",99,"type")="int"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.flush_time")=99
	Set gtmtypes("shm_snapshot_t",99,"dim")=2
	Set gtmtypes("shm_snapshot_t",100,"name")="shm_snapshot_t.shadow_file_header.flush_trigger"
	Set gtmtypes("shm_snapshot_t",100,"off")=4640
	Set gtmtypes("shm_snapshot_t",100,"len")=4
	Set gtmtypes("shm_snapshot_t",100,"type")="int"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.flush_trigger")=100
	Set gtmtypes("shm_snapshot_t",101,"name")="shm_snapshot_t.shadow_file_header.n_wrt_per_flu"
	Set gtmtypes("shm_snapshot_t",101,"off")=4644
	Set gtmtypes("shm_snapshot_t",101,"len")=4
	Set gtmtypes("shm_snapshot_t",101,"type")="int"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.n_wrt_per_flu")=101
	Set gtmtypes("shm_snapshot_t",102,"name")="shm_snapshot_t.shadow_file_header.wait_disk_space"
	Set gtmtypes("shm_snapshot_t",102,"off")=4648
	Set gtmtypes("shm_snapshot_t",102,"len")=4
	Set gtmtypes("shm_snapshot_t",102,"type")="int"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.wait_disk_space")=102
	Set gtmtypes("shm_snapshot_t",103,"name")="shm_snapshot_t.shadow_file_header.defer_time"
	Set gtmtypes("shm_snapshot_t",103,"off")=4652
	Set gtmtypes("shm_snapshot_t",103,"len")=4
	Set gtmtypes("shm_snapshot_t",103,"type")="int"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.defer_time")=103
	Set gtmtypes("shm_snapshot_t",104,"name")="shm_snapshot_t.shadow_file_header.flush_trigger_top"
	Set gtmtypes("shm_snapshot_t",104,"off")=4656
	Set gtmtypes("shm_snapshot_t",104,"len")=4
	Set gtmtypes("shm_snapshot_t",104,"type")="int"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.flush_trigger_top")=104
	Set gtmtypes("shm_snapshot_t",105,"name")="shm_snapshot_t.shadow_file_header.mumps_can_bypass"
	Set gtmtypes("shm_snapshot_t",105,"off")=4660
	Set gtmtypes("shm_snapshot_t",105,"len")=4
	Set gtmtypes("shm_snapshot_t",105,"type")="boolean_t"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.mumps_can_bypass")=105
	Set gtmtypes("shm_snapshot_t",106,"name")="shm_snapshot_t.shadow_file_header.epoch_taper"
	Set gtmtypes("shm_snapshot_t",106,"off")=4664
	Set gtmtypes("shm_snapshot_t",106,"len")=4
	Set gtmtypes("shm_snapshot_t",106,"type")="boolean_t"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.epoch_taper")=106
	Set gtmtypes("shm_snapshot_t",107,"name")="shm_snapshot_t.shadow_file_header.epoch_taper_time_pct"
	Set gtmtypes("shm_snapshot_t",107,"off")=4668
	Set gtmtypes("shm_snapshot_t",107,"len")=4
	Set gtmtypes("shm_snapshot_t",107,"type")="unsigned-int"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.epoch_taper_time_pct")=107
	Set gtmtypes("shm_snapshot_t",108,"name")="shm_snapshot_t.shadow_file_header.epoch_taper_jnl_pct"
	Set gtmtypes("shm_snapshot_t",108,"off")=4672
	Set gtmtypes("shm_snapshot_t",108,"len")=4
	Set gtmtypes("shm_snapshot_t",108,"type")="unsigned-int"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.epoch_taper_jnl_pct")=108
	Set gtmtypes("shm_snapshot_t",109,"name")="shm_snapshot_t.shadow_file_header.asyncio"
	Set gtmtypes("shm_snapshot_t",109,"off")=4676
	Set gtmtypes("shm_snapshot_t",109,"len")=4
	Set gtmtypes("shm_snapshot_t",109,"type")="boolean_t"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.asyncio")=109
	Set gtmtypes("shm_snapshot_t",110,"name")="shm_snapshot_t.shadow_file_header.reserved_for_upd"
	Set gtmtypes("shm_snapshot_t",110,"off")=4680
	Set gtmtypes("shm_snapshot_t",110,"len")=4
	Set gtmtypes("shm_snapshot_t",110,"type")="unsigned-int"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.reserved_for_upd")=110
	Set gtmtypes("shm_snapshot_t",111,"name")="shm_snapshot_t.shadow_file_header.avg_blks_per_100gbl"
	Set gtmtypes("shm_snapshot_t",111,"off")=4684
	Set gtmtypes("shm_snapshot_t",111,"len")=4
	Set gtmtypes("shm_snapshot_t",111,"type")="unsigned-int"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.avg_blks_per_100gbl")=111
	Set gtmtypes("shm_snapshot_t",112,"name")="shm_snapshot_t.shadow_file_header.pre_read_trigger_factor"
	Set gtmtypes("shm_snapshot_t",112,"off")=4688
	Set gtmtypes("shm_snapshot_t",112,"len")=4
	Set gtmtypes("shm_snapshot_t",112,"type")="unsigned-int"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.pre_read_trigger_factor")=112
	Set gtmtypes("shm_snapshot_t",113,"name")="shm_snapshot_t.shadow_file_header.writer_trigger_factor"
	Set gtmtypes("shm_snapshot_t",113,"off")=4692
	Set gtmtypes("shm_snapshot_t",113,"len")=4
	Set gtmtypes("shm_snapshot_t",113,"type")="unsigned-int"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.writer_trigger_factor")=113
	Set gtmtypes("shm_snapshot_t",114,"name")="shm_snapshot_t.shadow_file_header.semid"
	Set gtmtypes("shm_snapshot_t",114,"off")=4696
	Set gtmtypes("shm_snapshot_t",114,"len")=4
	Set gtmtypes("shm_snapshot_t",114,"type")="int"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.semid")=114
	Set gtmtypes("shm_snapshot_t",115,"name")="shm_snapshot_t.shadow_file_header.shmid"
	Set gtmtypes("shm_snapshot_t",115,"off")=4700
	Set gtmtypes("shm_snapshot_t",115,"len")=4
	Set gtmtypes("shm_snapshot_t",115,"type")="int"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.shmid")=115
	Set gtmtypes("shm_snapshot_t",116,"name")="shm_snapshot_t.shadow_file_header.gt_sem_ctime"
	Set gtmtypes("shm_snapshot_t",116,"off")=4704
	Set gtmtypes("shm_snapshot_t",116,"len")=8
	Set gtmtypes("shm_snapshot_t",116,"type")="gtm_time8"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.gt_sem_ctime")=116
	Set gtmtypes("shm_snapshot_t",117,"name")="shm_snapshot_t.shadow_file_header.gt_sem_ctime.ctime"
	Set gtmtypes("shm_snapshot_t",117,"off")=4704
	Set gtmtypes("shm_snapshot_t",117,"len")=8
	Set gtmtypes("shm_snapshot_t",117,"type")="time_t"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.gt_sem_ctime.ctime")=117
	Set gtmtypes("shm_snapshot_t",118,"name")="shm_snapshot_t.shadow_file_header.gt_sem_ctime.filler"
	Set gtmtypes("shm_snapshot_t",118,"off")=4704
	Set gtmtypes("shm_snapshot_t",118,"len")=8
	Set gtmtypes("shm_snapshot_t",118,"type")="int"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.gt_sem_ctime.filler")=118
	Set gtmtypes("shm_snapshot_t",118,"dim")=2
	Set gtmtypes("shm_snapshot_t",119,"name")="shm_snapshot_t.shadow_file_header.gt_shm_ctime"
	Set gtmtypes("shm_snapshot_t",119,"off")=4712
	Set gtmtypes("shm_snapshot_t",119,"len")=8
	Set gtmtypes("shm_snapshot_t",119,"type")="gtm_time8"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.gt_shm_ctime")=119
	Set gtmtypes("shm_snapshot_t",120,"name")="shm_snapshot_t.shadow_file_header.gt_shm_ctime.ctime"
	Set gtmtypes("shm_snapshot_t",120,"off")=4712
	Set gtmtypes("shm_snapshot_t",120,"len")=8
	Set gtmtypes("shm_snapshot_t",120,"type")="time_t"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.gt_shm_ctime.ctime")=120
	Set gtmtypes("shm_snapshot_t",121,"name")="shm_snapshot_t.shadow_file_header.gt_shm_ctime.filler"
	Set gtmtypes("shm_snapshot_t",121,"off")=4712
	Set gtmtypes("shm_snapshot_t",121,"len")=8
	Set gtmtypes("shm_snapshot_t",121,"type")="int"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.gt_shm_ctime.filler")=121
	Set gtmtypes("shm_snapshot_t",121,"dim")=2
	Set gtmtypes("shm_snapshot_t",122,"name")="shm_snapshot_t.shadow_file_header.filler_unixonly"
	Set gtmtypes("shm_snapshot_t",122,"off")=4720
	Set gtmtypes("shm_snapshot_t",122,"len")=40
	Set gtmtypes("shm_snapshot_t",122,"type")="char"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.filler_unixonly")=122
	Set gtmtypes("shm_snapshot_t",123,"name")="shm_snapshot_t.shadow_file_header.filler_n_retries"
	Set gtmtypes("shm_snapshot_t",123,"off")=4760
	Set gtmtypes("shm_snapshot_t",123,"len")=20
	Set gtmtypes("shm_snapshot_t",123,"type")="int"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.filler_n_retries")=123
	Set gtmtypes("shm_snapshot_t",123,"dim")=5
	Set gtmtypes("shm_snapshot_t",124,"name")="shm_snapshot_t.shadow_file_header.filler_n_puts"
	Set gtmtypes("shm_snapshot_t",124,"off")=4780
	Set gtmtypes("shm_snapshot_t",124,"len")=4
	Set gtmtypes("shm_snapshot_t",124,"type")="int"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.filler_n_puts")=124
	Set gtmtypes("shm_snapshot_t",125,"name")="shm_snapshot_t.shadow_file_header.filler_n_kills"
	Set gtmtypes("shm_snapshot_t",125,"off")=4784
	Set gtmtypes("shm_snapshot_t",125,"len")=4
	Set gtmtypes("shm_snapshot_t",125,"type")="int"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.filler_n_kills")=125
	Set gtmtypes("shm_snapshot_t",126,"name")="shm_snapshot_t.shadow_file_header.filler_n_queries"
	Set gtmtypes("shm_snapshot_t",126,"off")=4788
	Set gtmtypes("shm_snapshot_t",126,"len")=4
	Set gtmtypes("shm_snapshot_t",126,"type")="int"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.filler_n_queries")=126
	Set gtmtypes("shm_snapshot_t",127,"name")="shm_snapshot_t.shadow_file_header.filler_n_gets"
	Set gtmtypes("shm_snapshot_t",127,"off")=4792
	Set gtmtypes("shm_snapshot_t",127,"len")=4
	Set gtmtypes("shm_snapshot_t",127,"type")="int"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.filler_n_gets")=127
	Set gtmtypes("shm_snapshot_t",128,"name")="shm_snapshot_t.shadow_file_header.filler_n_order"
	Set gtmtypes("shm_snapshot_t",128,"off")=4796
	Set gtmtypes("shm_snapshot_t",128,"len")=4
	Set gtmtypes("shm_snapshot_t",128,"type")="int"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.filler_n_order")=128
	Set gtmtypes("shm_snapshot_t",129,"name")="shm_snapshot_t.shadow_file_header.filler_n_zprevs"
	Set gtmtypes("shm_snapshot_t",129,"off")=4800
	Set gtmtypes("shm_snapshot_t",129,"len")=4
	Set gtmtypes("shm_snapshot_t",129,"type")="int"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.filler_n_zprevs")=129
	Set gtmtypes("shm_snapshot_t",130,"name")="shm_snapshot_t.shadow_file_header.filler_n_data"
	Set gtmtypes("shm_snapshot_t",130,"off")=4804
	Set gtmtypes("shm_snapshot_t",130,"len")=4
	Set gtmtypes("shm_snapshot_t",130,"type")="int"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.filler_n_data")=130
	Set gtmtypes("shm_snapshot_t",131,"name")="shm_snapshot_t.shadow_file_header.filler_n_puts_duplicate"
	Set gtmtypes("shm_snapshot_t",131,"off")=4808
	Set gtmtypes("shm_snapshot_t",131,"len")=4
	Set gtmtypes("shm_snapshot_t",131,"type")="unsigned-int"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.filler_n_puts_duplicate")=131
	Set gtmtypes("shm_snapshot_t",132,"name")="shm_snapshot_t.shadow_file_header.filler_n_tp_updates"
	Set gtmtypes("shm_snapshot_t",132,"off")=4812
	Set gtmtypes("shm_snapshot_t",132,"len")=4
	Set gtmtypes("shm_snapshot_t",132,"type")="unsigned-int"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.filler_n_tp_updates")=132
	Set gtmtypes("shm_snapshot_t",133,"name")="shm_snapshot_t.shadow_file_header.filler_n_tp_updates_duplicate"
	Set gtmtypes("shm_snapshot_t",133,"off")=4816
	Set gtmtypes("shm_snapshot_t",133,"len")=4
	Set gtmtypes("shm_snapshot_t",133,"type")="unsigned-int"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.filler_n_tp_updates_duplicate")=133
	Set gtmtypes("shm_snapshot_t",134,"name")="shm_snapshot_t.shadow_file_header.filler_accounting_64_align"
	Set gtmtypes("shm_snapshot_t",134,"off")=4820
	Set gtmtypes("shm_snapshot_t",134,"len")=4
	Set gtmtypes("shm_snapshot_t",134,"type")="char"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.filler_accounting_64_align")=134
	Set gtmtypes("shm_snapshot_t",135,"name")="shm_snapshot_t.shadow_file_header.staleness"
	Set gtmtypes("shm_snapshot_t",135,"off")=4824
	Set gtmtypes("shm_snapshot_t",135,"len")=8
	Set gtmtypes("shm_snapshot_t",135,"type")="int"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.staleness")=135
	Set gtmtypes("shm_snapshot_t",135,"dim")=2
	Set gtmtypes("shm_snapshot_t",136,"name")="shm_snapshot_t.shadow_file_header.ccp_tick_interval"
	Set gtmtypes("shm_snapshot_t",136,"off")=4832
	Set gtmtypes("shm_snapshot_t",136,"len")=8
	Set gtmtypes("shm_snapshot_t",136,"type")="int"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.ccp_tick_interval")=136
	Set gtmtypes("shm_snapshot_t",136,"dim")=2
	Set gtmtypes("shm_snapshot_t",137,"name")="shm_snapshot_t.shadow_file_header.ccp_quantum_interval"
	Set gtmtypes("shm_snapshot_t",137,"off")=4840
	Set gtmtypes("shm_snapshot_t",137,"len")=8
	Set gtmtypes("shm_snapshot_t",137,"type")="int"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.ccp_quantum_interval")=137
	Set gtmtypes("shm_snapshot_t",137,"dim")=2
	Set gtmtypes("shm_snapshot_t",138,"name")="shm_snapshot_t.shadow_file_header.ccp_response_interval"
	Set gtmtypes("shm_snapshot_t",138,"off")=4848
	Set gtmtypes("shm_snapshot_t",138,"len")=8
	Set gtmtypes("shm_snapshot_t",138,"type")="int"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.ccp_response_interval")=138
	Set gtmtypes("shm_snapshot_t",138,"dim")=2
	Set gtmtypes("shm_snapshot_t",139,"name")="shm_snapshot_t.shadow_file_header.ccp_jnl_before"
	Set gtmtypes("shm_snapshot_t",139,"off")=4856
	Set gtmtypes("shm_snapshot_t",139,"len")=4
	Set gtmtypes("shm_snapshot_t",139,"type")="boolean_t"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.ccp_jnl_before")=139
	Set gtmtypes("shm_snapshot_t",140,"name")="shm_snapshot_t.shadow_file_header.clustered"
	Set gtmtypes("shm_snapshot_t",140,"off")=4860
	Set gtmtypes("shm_snapshot_t",140,"len")=4
	Set gtmtypes("shm_snapshot_t",140,"type")="boolean_t"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.clustered")=140
	Set gtmtypes("shm_snapshot_t",141,"name")="shm_snapshot_t.shadow_file_header.unbacked_cache"
	Set gtmtypes("shm_snapshot_t",141,"off")=4864
	Set gtmtypes("shm_snapshot_t",141,"len")=4
	Set gtmtypes("shm_snapshot_t",141,"type")="boolean_t"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.unbacked_cache")=141
	Set gtmtypes("shm_snapshot_t",142,"name")="shm_snapshot_t.shadow_file_header.rc_srv_cnt"
	Set gtmtypes("shm_snapshot_t",142,"off")=4868
	Set gtmtypes("shm_snapshot_t",142,"len")=4
	Set gtmtypes("shm_snapshot_t",142,"type")="int"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.rc_srv_cnt")=142
	Set gtmtypes("shm_snapshot_t",143,"name")="shm_snapshot_t.shadow_file_header.dsid"
	Set gtmtypes("shm_snapshot_t",143,"off")=4872
	Set gtmtypes("shm_snapshot_t",143,"len")=4
	Set gtmtypes("shm_snapshot_t",143,"type")="int"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.dsid")=143
	Set gtmtypes("shm_snapshot_t",144,"name")="shm_snapshot_t.shadow_file_header.rc_node"
	Set gtmtypes("shm_snapshot_t",144,"off")=4876
	Set gtmtypes("shm_snapshot_t",144,"len")=4
	Set gtmtypes("shm_snapshot_t",144,"type")="int"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.rc_node")=144
	Set gtmtypes("shm_snapshot_t",145,"name")="shm_snapshot_t.shadow_file_header.filler_ccp_rc"
	Set gtmtypes("shm_snapshot_t",145,"off")=4880
	Set gtmtypes("shm_snapshot_t",145,"len")=8
	Set gtmtypes("shm_snapshot_t",145,"type")="char"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.filler_ccp_rc")=145
	Set gtmtypes("shm_snapshot_t",146,"name")="shm_snapshot_t.shadow_file_header.reg_seqno"
	Set gtmtypes("shm_snapshot_t",146,"off")=4888
	Set gtmtypes("shm_snapshot_t",146,"len")=8
	Set gtmtypes("shm_snapshot_t",146,"type")="uint64_t"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.reg_seqno")=146
	Set gtmtypes("shm_snapshot_t",147,"name")="shm_snapshot_t.shadow_file_header.pre_multisite_resync_seqno"
	Set gtmtypes("shm_snapshot_t",147,"off")=4896
	Set gtmtypes("shm_snapshot_t",147,"len")=8
	Set gtmtypes("shm_snapshot_t",147,"type")="uint64_t"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.pre_multisite_resync_seqno")=147
	Set gtmtypes("shm_snapshot_t",148,"name")="shm_snapshot_t.shadow_file_header.zqgblmod_tn"
	Set gtmtypes("shm_snapshot_t",148,"off")=4904
	Set gtmtypes("shm_snapshot_t",148,"len")=8
	Set gtmtypes("shm_snapshot_t",148,"type")="uint64_t"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.zqgblmod_tn")=148
	Set gtmtypes("shm_snapshot_t",149,"name")="shm_snapshot_t.shadow_file_header.zqgblmod_seqno"
	Set gtmtypes("shm_snapshot_t",149,"off")=4912
	Set gtmtypes("shm_snapshot_t",149,"len")=8
	Set gtmtypes("shm_snapshot_t",149,"type")="uint64_t"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.zqgblmod_seqno")=149
	Set gtmtypes("shm_snapshot_t",150,"name")="shm_snapshot_t.shadow_file_header.repl_state"
	Set gtmtypes("shm_snapshot_t",150,"off")=4920
	Set gtmtypes("shm_snapshot_t",150,"len")=4
	Set gtmtypes("shm_snapshot_t",150,"type")="int"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.repl_state")=150
	Set gtmtypes("shm_snapshot_t",151,"name")="shm_snapshot_t.shadow_file_header.multi_site_open"
	Set gtmtypes("shm_snapshot_t",151,"off")=4924
	Set gtmtypes("shm_snapshot_t",151,"len")=4
	Set gtmtypes("shm_snapshot_t",151,"type")="boolean_t"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.multi_site_open")=151
	Set gtmtypes("shm_snapshot_t",152,"name")="shm_snapshot_t.shadow_file_header.filler_seqno"
	Set gtmtypes("shm_snapshot_t",152,"off")=4928
	Set gtmtypes("shm_snapshot_t",152,"len")=8
	Set gtmtypes("shm_snapshot_t",152,"type")="uint64_t"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.filler_seqno")=152
	Set gtmtypes("shm_snapshot_t",153,"name")="shm_snapshot_t.shadow_file_header.filler_repl"
	Set gtmtypes("shm_snapshot_t",153,"off")=4936
	Set gtmtypes("shm_snapshot_t",153,"len")=16
	Set gtmtypes("shm_snapshot_t",153,"type")="char"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.filler_repl")=153
	Set gtmtypes("shm_snapshot_t",154,"name")="shm_snapshot_t.shadow_file_header.filler_n_tp_retries"
	Set gtmtypes("shm_snapshot_t",154,"off")=4952
	Set gtmtypes("shm_snapshot_t",154,"len")=48
	Set gtmtypes("shm_snapshot_t",154,"type")="int"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.filler_n_tp_retries")=154
	Set gtmtypes("shm_snapshot_t",154,"dim")=12
	Set gtmtypes("shm_snapshot_t",155,"name")="shm_snapshot_t.shadow_file_header.filler_n_tp_retries_conflicts"
	Set gtmtypes("shm_snapshot_t",155,"off")=5000
	Set gtmtypes("shm_snapshot_t",155,"len")=48
	Set gtmtypes("shm_snapshot_t",155,"type")="int"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.filler_n_tp_retries_conflicts")=155
	Set gtmtypes("shm_snapshot_t",155,"dim")=12
	Set gtmtypes("shm_snapshot_t",156,"name")="shm_snapshot_t.shadow_file_header.tp_cdb_sc_blkmod"
	Set gtmtypes("shm_snapshot_t",156,"off")=5048
	Set gtmtypes("shm_snapshot_t",156,"len")=32
	Set gtmtypes("shm_snapshot_t",156,"type")="int"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.tp_cdb_sc_blkmod")=156
	Set gtmtypes("shm_snapshot_t",156,"dim")=8
	Set gtmtypes("shm_snapshot_t",157,"name")="shm_snapshot_t.shadow_file_header.jnl_alq"
	Set gtmtypes("shm_snapshot_t",157,"off")=5080
	Set gtmtypes("shm_snapshot_t",157,"len")=4
	Set gtmtypes("shm_snapshot_t",157,"type")="unsigned-int"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.jnl_alq")=157
	Set gtmtypes("shm_snapshot_t",158,"name")="shm_snapshot_t.shadow_file_header.jnl_deq"
	Set gtmtypes("shm_snapshot_t",158,"off")=5084
	Set gtmtypes("shm_snapshot_t",158,"len")=4
	Set gtmtypes("shm_snapshot_t",158,"type")="unsigned-int"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.jnl_deq")=158
	Set gtmtypes("shm_snapshot_t",159,"name")="shm_snapshot_t.shadow_file_header.jnl_buffer_size"
	Set gtmtypes("shm_snapshot_t",159,"off")=5088
	Set gtmtypes("shm_snapshot_t",159,"len")=4
	Set gtmtypes("shm_snapshot_t",159,"type")="int"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.jnl_buffer_size")=159
	Set gtmtypes("shm_snapshot_t",160,"name")="shm_snapshot_t.shadow_file_header.jnl_before_image"
	Set gtmtypes("shm_snapshot_t",160,"off")=5092
	Set gtmtypes("shm_snapshot_t",160,"len")=4
	Set gtmtypes("shm_snapshot_t",160,"type")="boolean_t"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.jnl_before_image")=160
	Set gtmtypes("shm_snapshot_t",161,"name")="shm_snapshot_t.shadow_file_header.jnl_state"
	Set gtmtypes("shm_snapshot_t",161,"off")=5096
	Set gtmtypes("shm_snapshot_t",161,"len")=4
	Set gtmtypes("shm_snapshot_t",161,"type")="int"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.jnl_state")=161
	Set gtmtypes("shm_snapshot_t",162,"name")="shm_snapshot_t.shadow_file_header.jnl_file_len"
	Set gtmtypes("shm_snapshot_t",162,"off")=5100
	Set gtmtypes("shm_snapshot_t",162,"len")=4
	Set gtmtypes("shm_snapshot_t",162,"type")="unsigned-int"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.jnl_file_len")=162
	Set gtmtypes("shm_snapshot_t",163,"name")="shm_snapshot_t.shadow_file_header.autoswitchlimit"
	Set gtmtypes("shm_snapshot_t",163,"off")=5104
	Set gtmtypes("shm_snapshot_t",163,"len")=4
	Set gtmtypes("shm_snapshot_t",163,"type")="unsigned-int"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.autoswitchlimit")=163
	Set gtmtypes("shm_snapshot_t",164,"name")="shm_snapshot_t.shadow_file_header.epoch_interval"
	Set gtmtypes("shm_snapshot_t",164,"off")=5108
	Set gtmtypes("shm_snapshot_t",164,"len")=4
	Set gtmtypes("shm_snapshot_t",164,"type")="int"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.epoch_interval")=164
	Set gtmtypes("shm_snapshot_t",165,"name")="shm_snapshot_t.shadow_file_header.alignsize"
	Set gtmtypes("shm_snapshot_t",165,"off")=5112
	Set gtmtypes("shm_snapshot_t",165,"len")=4
	Set gtmtypes("shm_snapshot_t",165,"type")="unsigned-int"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.alignsize")=165
	Set gtmtypes("shm_snapshot_t",166,"name")="shm_snapshot_t.shadow_file_header.jnl_sync_io"
	Set gtmtypes("shm_snapshot_t",166,"off")=5116
	Set gtmtypes("shm_snapshot_t",166,"len")=4
	Set gtmtypes("shm_snapshot_t",166,"type")="int"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.jnl_sync_io")=166
	Set gtmtypes("shm_snapshot_t",167,"name")="shm_snapshot_t.shadow_file_header.yield_lmt"
	Set gtmtypes("shm_snapshot_t",167,"off")=5120
	Set gtmtypes("shm_snapshot_t",167,"len")=4
	Set gtmtypes("shm_snapshot_t",167,"type")="int"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.yield_lmt")=167
	Set gtmtypes("shm_snapshot_t",168,"name")="shm_snapshot_t.shadow_file_header.turn_around_point"
	Set gtmtypes("shm_snapshot_t",168,"off")=5124
	Set gtmtypes("shm_snapshot_t",168,"len")=4
	Set gtmtypes("shm_snapshot_t",168,"type")="boolean_t"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.turn_around_point")=168
	Set gtmtypes("shm_snapshot_t",169,"name")="shm_snapshot_t.shadow_file_header.jnl_eovtn"
	Set gtmtypes("shm_snapshot_t",169,"off")=5128
	Set gtmtypes("shm_snapshot_t",169,"len")=8
	Set gtmtypes("shm_snapshot_t",169,"type")="uint64_t"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.jnl_eovtn")=169
	Set gtmtypes("shm_snapshot_t",170,"name")="shm_snapshot_t.shadow_file_header.filler_jnl"
	Set gtmtypes("shm_snapshot_t",170,"off")=5136
	Set gtmtypes("shm_snapshot_t",170,"len")=8
	Set gtmtypes("shm_snapshot_t",170,"type")="char"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.filler_jnl")=170
	Set gtmtypes("shm_snapshot_t",171,"name")="shm_snapshot_t.shadow_file_header.intrpt_recov_resync_seqno"
	Set gtmtypes("shm_snapshot_t",171,"off")=5144
	Set gtmtypes("shm_snapshot_t",171,"len")=8
	Set gtmtypes("shm_snapshot_t",171,"type")="uint64_t"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.intrpt_recov_resync_seqno")=171
	Set gtmtypes("shm_snapshot_t",172,"name")="shm_snapshot_t.shadow_file_header.intrpt_recov_tp_resolve_time"
	Set gtmtypes("shm_snapshot_t",172,"off")=5152
	Set gtmtypes("shm_snapshot_t",172,"len")=4
	Set gtmtypes("shm_snapshot_t",172,"type")="unsigned-int"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.intrpt_recov_tp_resolve_time")=172
	Set gtmtypes("shm_snapshot_t",173,"name")="shm_snapshot_t.shadow_file_header.recov_interrupted"
	Set gtmtypes("shm_snapshot_t",173,"off")=5156
	Set gtmtypes("shm_snapshot_t",173,"len")=4
	Set gtmtypes("shm_snapshot_t",173,"type")="boolean_t"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.recov_interrupted")=173
	Set gtmtypes("shm_snapshot_t",174,"name")="shm_snapshot_t.shadow_file_header.intrpt_recov_jnl_state"
	Set gtmtypes("shm_snapshot_t",174,"off")=5160
	Set gtmtypes("shm_snapshot_t",174,"len")=4
	Set gtmtypes("shm_snapshot_t",174,"type")="int"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.intrpt_recov_jnl_state")=174
	Set gtmtypes("shm_snapshot_t",175,"name")="shm_snapshot_t.shadow_file_header.intrpt_recov_repl_state"
	Set gtmtypes("shm_snapshot_t",175,"off")=5164
	Set gtmtypes("shm_snapshot_t",175,"len")=4
	Set gtmtypes("shm_snapshot_t",175,"type")="int"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.intrpt_recov_repl_state")=175
	Set gtmtypes("shm_snapshot_t",176,"name")="shm_snapshot_t.shadow_file_header.before_trunc_total_blks"
	Set gtmtypes("shm_snapshot_t",176,"off")=5168
	Set gtmtypes("shm_snapshot_t",176,"len")=4
	Set gtmtypes("shm_snapshot_t",176,"type")="unsigned-int"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.before_trunc_total_blks")=176
	Set gtmtypes("shm_snapshot_t",177,"name")="shm_snapshot_t.shadow_file_header.after_trunc_total_blks"
	Set gtmtypes("shm_snapshot_t",177,"off")=5172
	Set gtmtypes("shm_snapshot_t",177,"len")=4
	Set gtmtypes("shm_snapshot_t",177,"type")="unsigned-int"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.after_trunc_total_blks")=177
	Set gtmtypes("shm_snapshot_t",178,"name")="shm_snapshot_t.shadow_file_header.before_trunc_free_blocks"
	Set gtmtypes("shm_snapshot_t",178,"off")=5176
	Set gtmtypes("shm_snapshot_t",178,"len")=4
	Set gtmtypes("shm_snapshot_t",178,"type")="unsigned-int"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.before_trunc_free_blocks")=178
	Set gtmtypes("shm_snapshot_t",179,"name")="shm_snapshot_t.shadow_file_header.filler_trunc"
	Set gtmtypes("shm_snapshot_t",179,"off")=5180
	Set gtmtypes("shm_snapshot_t",179,"len")=4
	Set gtmtypes("shm_snapshot_t",179,"type")="unsigned-int"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.filler_trunc")=179
	Set gtmtypes("shm_snapshot_t",180,"name")="shm_snapshot_t.shadow_file_header.filler_1k"
	Set gtmtypes("shm_snapshot_t",180,"off")=5184
	Set gtmtypes("shm_snapshot_t",180,"len")=24
	Set gtmtypes("shm_snapshot_t",180,"type")="char"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.filler_1k")=180
	Set gtmtypes("shm_snapshot_t",181,"name")="shm_snapshot_t.shadow_file_header.jnl_file_name"
	Set gtmtypes("shm_snapshot_t",181,"off")=5208
	Set gtmtypes("shm_snapshot_t",181,"len")=256
	Set gtmtypes("shm_snapshot_t",181,"type")="unsigned-char"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.jnl_file_name")=181
	Set gtmtypes("shm_snapshot_t",182,"name")="shm_snapshot_t.shadow_file_header.reorg_restart_key"
	Set gtmtypes("shm_snapshot_t",182,"off")=5464
	Set gtmtypes("shm_snapshot_t",182,"len")=256
	Set gtmtypes("shm_snapshot_t",182,"type")="unsigned-char"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.reorg_restart_key")=182
	Set gtmtypes("shm_snapshot_t",183,"name")="shm_snapshot_t.shadow_file_header.machine_name"
	Set gtmtypes("shm_snapshot_t",183,"off")=5720
	Set gtmtypes("shm_snapshot_t",183,"len")=256
	Set gtmtypes("shm_snapshot_t",183,"type")="char"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.machine_name")=183
	Set gtmtypes("shm_snapshot_t",184,"name")="shm_snapshot_t.shadow_file_header.encryption_hash"
	Set gtmtypes("shm_snapshot_t",184,"off")=5976
	Set gtmtypes("shm_snapshot_t",184,"len")=80
	Set gtmtypes("shm_snapshot_t",184,"type")="char"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.encryption_hash")=184
	Set gtmtypes("shm_snapshot_t",185,"name")="shm_snapshot_t.shadow_file_header.encryption_hash2"
	Set gtmtypes("shm_snapshot_t",185,"off")=6056
	Set gtmtypes("shm_snapshot_t",185,"len")=80
	Set gtmtypes("shm_snapshot_t",185,"type")="char"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.encryption_hash2")=185
	Set gtmtypes("shm_snapshot_t",186,"name")="shm_snapshot_t.shadow_file_header.non_null_iv"
	Set gtmtypes("shm_snapshot_t",186,"off")=6136
	Set gtmtypes("shm_snapshot_t",186,"len")=4
	Set gtmtypes("shm_snapshot_t",186,"type")="boolean_t"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.non_null_iv")=186
	Set gtmtypes("shm_snapshot_t",187,"name")="shm_snapshot_t.shadow_file_header.encryption_hash_cutoff"
	Set gtmtypes("shm_snapshot_t",187,"off")=6140
	Set gtmtypes("shm_snapshot_t",187,"len")=4
	Set gtmtypes("shm_snapshot_t",187,"type")="int"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.encryption_hash_cutoff")=187
	Set gtmtypes("shm_snapshot_t",188,"name")="shm_snapshot_t.shadow_file_header.encryption_hash2_start_tn"
	Set gtmtypes("shm_snapshot_t",188,"off")=6144
	Set gtmtypes("shm_snapshot_t",188,"len")=8
	Set gtmtypes("shm_snapshot_t",188,"type")="uint64_t"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.encryption_hash2_start_tn")=188
	Set gtmtypes("shm_snapshot_t",189,"name")="shm_snapshot_t.shadow_file_header.filler_encrypt"
	Set gtmtypes("shm_snapshot_t",189,"off")=6152
	Set gtmtypes("shm_snapshot_t",189,"len")=80
	Set gtmtypes("shm_snapshot_t",189,"type")="char"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.filler_encrypt")=189
	Set gtmtypes("shm_snapshot_t",190,"name")="shm_snapshot_t.shadow_file_header.total_buffer_flush_tn"
	Set gtmtypes("shm_snapshot_t",190,"off")=6232
	Set gtmtypes("shm_snapshot_t",190,"len")=8
	Set gtmtypes("shm_snapshot_t",190,"type")="uint64_t"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.total_buffer_flush_tn")=190
	Set gtmtypes("shm_snapshot_t",191,"name")="shm_snapshot_t.shadow_file_header.bufct_buffer_flush_tn"
	Set gtmtypes("shm_snapshot_t",191,"off")=6240
	Set gtmtypes("shm_snapshot_t",191,"len")=8
	Set gtmtypes("shm_snapshot_t",191,"type")="uint64_t"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.bufct_buffer_flush_tn")=191
	Set gtmtypes("shm_snapshot_t",192,"name")="shm_snapshot_t.shadow_file_header.bufct_buffer_flush_loop_tn"
	Set gtmtypes("shm_snapshot_t",192,"off")=6248
	Set gtmtypes("shm_snapshot_t",192,"len")=8
	Set gtmtypes("shm_snapshot_t",192,"type")="uint64_t"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.bufct_buffer_flush_loop_tn")=192
	Set gtmtypes("shm_snapshot_t",193,"name")="shm_snapshot_t.shadow_file_header.wrt_calls_tn"
	Set gtmtypes("shm_snapshot_t",193,"off")=6256
	Set gtmtypes("shm_snapshot_t",193,"len")=8
	Set gtmtypes("shm_snapshot_t",193,"type")="uint64_t"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.wrt_calls_tn")=193
	Set gtmtypes("shm_snapshot_t",194,"name")="shm_snapshot_t.shadow_file_header.wrt_count_tn"
	Set gtmtypes("shm_snapshot_t",194,"off")=6264
	Set gtmtypes("shm_snapshot_t",194,"len")=8
	Set gtmtypes("shm_snapshot_t",194,"type")="uint64_t"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.wrt_count_tn")=194
	Set gtmtypes("shm_snapshot_t",195,"name")="shm_snapshot_t.shadow_file_header.wrt_blocked_tn"
	Set gtmtypes("shm_snapshot_t",195,"off")=6272
	Set gtmtypes("shm_snapshot_t",195,"len")=8
	Set gtmtypes("shm_snapshot_t",195,"type")="uint64_t"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.wrt_blocked_tn")=195
	Set gtmtypes("shm_snapshot_t",196,"name")="shm_snapshot_t.shadow_file_header.wrt_busy_tn"
	Set gtmtypes("shm_snapshot_t",196,"off")=6280
	Set gtmtypes("shm_snapshot_t",196,"len")=8
	Set gtmtypes("shm_snapshot_t",196,"type")="uint64_t"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.wrt_busy_tn")=196
	Set gtmtypes("shm_snapshot_t",197,"name")="shm_snapshot_t.shadow_file_header.wrt_noblks_wrtn_tn"
	Set gtmtypes("shm_snapshot_t",197,"off")=6288
	Set gtmtypes("shm_snapshot_t",197,"len")=8
	Set gtmtypes("shm_snapshot_t",197,"type")="uint64_t"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.wrt_noblks_wrtn_tn")=197
	Set gtmtypes("shm_snapshot_t",198,"name")="shm_snapshot_t.shadow_file_header.wcb_gds_rundown2_tn"
	Set gtmtypes("shm_snapshot_t",198,"off")=6296
	Set gtmtypes("shm_snapshot_t",198,"len")=8
	Set gtmtypes("shm_snapshot_t",198,"type")="uint64_t"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.wcb_gds_rundown2_tn")=198
	Set gtmtypes("shm_snapshot_t",199,"name")="shm_snapshot_t.shadow_file_header.reserved_bgtrcrec2_tn"
	Set gtmtypes("shm_snapshot_t",199,"off")=6304
	Set gtmtypes("shm_snapshot_t",199,"len")=8
	Set gtmtypes("shm_snapshot_t",199,"type")="uint64_t"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.reserved_bgtrcrec2_tn")=199
	Set gtmtypes("shm_snapshot_t",200,"name")="shm_snapshot_t.shadow_file_header.reserved_bgtrcrec3_tn"
	Set gtmtypes("shm_snapshot_t",200,"off")=6312
	Set gtmtypes("shm_snapshot_t",200,"len")=8
	Set gtmtypes("shm_snapshot_t",200,"type")="uint64_t"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.reserved_bgtrcrec3_tn")=200
	Set gtmtypes("shm_snapshot_t",201,"name")="shm_snapshot_t.shadow_file_header.lost_block_recovery_tn"
	Set gtmtypes("shm_snapshot_t",201,"off")=6320
	Set gtmtypes("shm_snapshot_t",201,"len")=8
	Set gtmtypes("shm_snapshot_t",201,"type")="uint64_t"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.lost_block_recovery_tn")=201
	Set gtmtypes("shm_snapshot_t",202,"name")="shm_snapshot_t.shadow_file_header.wc_blocked_onln_rlbk_tn"
	Set gtmtypes("shm_snapshot_t",202,"off")=6328
	Set gtmtypes("shm_snapshot_t",202,"len")=8
	Set gtmtypes("shm_snapshot_t",202,"type")="uint64_t"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.wc_blocked_onln_rlbk_tn")=202
	Set gtmtypes("shm_snapshot_t",203,"name")="shm_snapshot_t.shadow_file_header.stale_tn"
	Set gtmtypes("shm_snapshot_t",203,"off")=6336
	Set gtmtypes("shm_snapshot_t",203,"len")=8
	Set gtmtypes("shm_snapshot_t",203,"type")="uint64_t"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.stale_tn")=203
	Set gtmtypes("shm_snapshot_t",204,"name")="shm_snapshot_t.shadow_file_header.starved_tn"
	Set gtmtypes("shm_snapshot_t",204,"off")=6344
	Set gtmtypes("shm_snapshot_t",204,"len")=8
	Set gtmtypes("shm_snapshot_t",204,"type")="uint64_t"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.starved_tn")=204
	Set gtmtypes("shm_snapshot_t",205,"name")="shm_snapshot_t.shadow_file_header.stale_timer_started_tn"
	Set gtmtypes("shm_snapshot_t",205,"off")=6352
	Set gtmtypes("shm_snapshot_t",205,"len")=8
	Set gtmtypes("shm_snapshot_t",205,"type")="uint64_t"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.stale_timer_started_tn")=205
	Set gtmtypes("shm_snapshot_t",206,"name")="shm_snapshot_t.shadow_file_header.stale_timer_pop_tn"
	Set gtmtypes("shm_snapshot_t",206,"off")=6360
	Set gtmtypes("shm_snapshot_t",206,"len")=8
	Set gtmtypes("shm_snapshot_t",206,"type")="uint64_t"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.stale_timer_pop_tn")=206
	Set gtmtypes("shm_snapshot_t",207,"name")="shm_snapshot_t.shadow_file_header.stale_process_defer_tn"
	Set gtmtypes("shm_snapshot_t",207,"off")=6368
	Set gtmtypes("shm_snapshot_t",207,"len")=8
	Set gtmtypes("shm_snapshot_t",207,"type")="uint64_t"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.stale_process_defer_tn")=207
	Set gtmtypes("shm_snapshot_t",208,"name")="shm_snapshot_t.shadow_file_header.stale_defer_processed_tn"
	Set gtmtypes("shm_snapshot_t",208,"off")=6376
	Set gtmtypes("shm_snapshot_t",208,"len")=8
	Set gtmtypes("shm_snapshot_t",208,"type")="uint64_t"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.stale_defer_processed_tn")=208
	Set gtmtypes("shm_snapshot_t",209,"name")="shm_snapshot_t.shadow_file_header.spcfc_buffer_flush_tn"
	Set gtmtypes("shm_snapshot_t",209,"off")=6384
	Set gtmtypes("shm_snapshot_t",209,"len")=8
	Set gtmtypes("shm_snapshot_t",209,"type")="uint64_t"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.spcfc_buffer_flush_tn")=209
	Set gtmtypes("shm_snapshot_t",210,"name")="shm_snapshot_t.shadow_file_header.spcfc_buffer_flush_loop_tn"
	Set gtmtypes("shm_snapshot_t",210,"off")=6392
	Set gtmtypes("shm_snapshot_t",210,"len")=8
	Set gtmtypes("shm_snapshot_t",210,"type")="uint64_t"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.spcfc_buffer_flush_loop_tn")=210
	Set gtmtypes("shm_snapshot_t",211,"name")="shm_snapshot_t.shadow_file_header.spcfc_buffer_flush_retries_tn"
	Set gtmtypes("shm_snapshot_t",211,"off")=6400
	Set gtmtypes("shm_snapshot_t",211,"len")=8
	Set gtmtypes("shm_snapshot_t",211,"type")="uint64_t"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.spcfc_buffer_flush_retries_tn")=211
	Set gtmtypes("shm_snapshot_t",212,"name")="shm_snapshot_t.shadow_file_header.spcfc_buffer_flushed_during_lockwait_tn"
	Set gtmtypes("shm_snapshot_t",212,"off")=6408
	Set gtmtypes("shm_snapshot_t",212,"len")=8
	Set gtmtypes("shm_snapshot_t",212,"type")="uint64_t"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.spcfc_buffer_flushed_during_lockwait_tn")=212
	Set gtmtypes("shm_snapshot_t",213,"name")="shm_snapshot_t.shadow_file_header.tp_crit_retries_tn"
	Set gtmtypes("shm_snapshot_t",213,"off")=6416
	Set gtmtypes("shm_snapshot_t",213,"len")=8
	Set gtmtypes("shm_snapshot_t",213,"type")="uint64_t"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.tp_crit_retries_tn")=213
	Set gtmtypes("shm_snapshot_t",214,"name")="shm_snapshot_t.shadow_file_header.db_csh_getn_flush_dirty_tn"
	Set gtmtypes("shm_snapshot_t",214,"off")=6424
	Set gtmtypes("shm_snapshot_t",214,"len")=8
	Set gtmtypes("shm_snapshot_t",214,"type")="uint64_t"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.db_csh_getn_flush_dirty_tn")=214
	Set gtmtypes("shm_snapshot_t",215,"name")="shm_snapshot_t.shadow_file_header.db_csh_getn_rip_wait_tn"
	Set gtmtypes("shm_snapshot_t",215,"off")=6432
	Set gtmtypes("shm_snapshot_t",215,"len")=8
	Set gtmtypes("shm_snapshot_t",215,"type")="uint64_t"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.db_csh_getn_rip_wait_tn")=215
	Set gtmtypes("shm_snapshot_t",216,"name")="shm_snapshot_t.shadow_file_header.db_csh_getn_buf_owner_stuck_tn"
	Set gtmtypes("shm_snapshot_t",216,"off")=6440
	Set gtmtypes("shm_snapshot_t",216,"len")=8
	Set gtmtypes("shm_snapshot_t",216,"type")="uint64_t"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.db_csh_getn_buf_owner_stuck_tn")=216
	Set gtmtypes("shm_snapshot_t",217,"name")="shm_snapshot_t.shadow_file_header.db_csh_getn_out_of_design_tn"
	Set gtmtypes("shm_snapshot_t",217,"off")=6448
	Set gtmtypes("shm_snapshot_t",217,"len")=8
	Set gtmtypes("shm_snapshot_t",217,"type")="uint64_t"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.db_csh_getn_out_of_design_tn")=217
	Set gtmtypes("shm_snapshot_t",218,"name")="shm_snapshot_t.shadow_file_header.t_qread_buf_owner_stuck_tn"
	Set gtmtypes("shm_snapshot_t",218,"off")=6456
	Set gtmtypes("shm_snapshot_t",218,"len")=8
	Set gtmtypes("shm_snapshot_t",218,"type")="uint64_t"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.t_qread_buf_owner_stuck_tn")=218
	Set gtmtypes("shm_snapshot_t",219,"name")="shm_snapshot_t.shadow_file_header.t_qread_out_of_design_tn"
	Set gtmtypes("shm_snapshot_t",219,"off")=6464
	Set gtmtypes("shm_snapshot_t",219,"len")=8
	Set gtmtypes("shm_snapshot_t",219,"type")="uint64_t"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.t_qread_out_of_design_tn")=219
	Set gtmtypes("shm_snapshot_t",220,"name")="shm_snapshot_t.shadow_file_header.unused_filler_0031_tn"
	Set gtmtypes("shm_snapshot_t",220,"off")=6472
	Set gtmtypes("shm_snapshot_t",220,"len")=8
	Set gtmtypes("shm_snapshot_t",220,"type")="uint64_t"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.unused_filler_0031_tn")=220
	Set gtmtypes("shm_snapshot_t",221,"name")="shm_snapshot_t.shadow_file_header.mlock_wakeups_tn"
	Set gtmtypes("shm_snapshot_t",221,"off")=6480
	Set gtmtypes("shm_snapshot_t",221,"len")=8
	Set gtmtypes("shm_snapshot_t",221,"type")="uint64_t"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.mlock_wakeups_tn")=221
	Set gtmtypes("shm_snapshot_t",222,"name")="shm_snapshot_t.shadow_file_header.wc_blocked_wcs_recover_invoked_tn"
	Set gtmtypes("shm_snapshot_t",222,"off")=6488
	Set gtmtypes("shm_snapshot_t",222,"len")=8
	Set gtmtypes("shm_snapshot_t",222,"type")="uint64_t"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.wc_blocked_wcs_recover_invoked_tn")=222
	Set gtmtypes("shm_snapshot_t",223,"name")="shm_snapshot_t.shadow_file_header.wc_blocked_wcs_verify_passed_tn"
	Set gtmtypes("shm_snapshot_t",223,"off")=6496
	Set gtmtypes("shm_snapshot_t",223,"len")=8
	Set gtmtypes("shm_snapshot_t",223,"type")="uint64_t"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.wc_blocked_wcs_verify_passed_tn")=223
	Set gtmtypes("shm_snapshot_t",224,"name")="shm_snapshot_t.shadow_file_header.wc_blocked_t_qread_db_csh_getn_invalid_blk_tn"
	Set gtmtypes("shm_snapshot_t",224,"off")=6504
	Set gtmtypes("shm_snapshot_t",224,"len")=8
	Set gtmtypes("shm_snapshot_t",224,"type")="uint64_t"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.wc_blocked_t_qread_db_csh_getn_invalid_blk_tn")=224
	Set gtmtypes("shm_snapshot_t",225,"name")="shm_snapshot_t.shadow_file_header.wc_blocked_t_qread_db_csh_get_invalid_blk_tn"
	Set gtmtypes("shm_snapshot_t",225,"off")=6512
	Set gtmtypes("shm_snapshot_t",225,"len")=8
	Set gtmtypes("shm_snapshot_t",225,"type")="uint64_t"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.wc_blocked_t_qread_db_csh_get_invalid_blk_tn")=225
	Set gtmtypes("shm_snapshot_t",226,"name")="shm_snapshot_t.shadow_file_header.wc_blocked_db_csh_getn_loopexceed_tn"
	Set gtmtypes("shm_snapshot_t",226,"off")=6520
	Set gtmtypes("shm_snapshot_t",226,"len")=8
	Set gtmtypes("shm_snapshot_t",226,"type")="uint64_t"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.wc_blocked_db_csh_getn_loopexceed_tn")=226
	Set gtmtypes("shm_snapshot_t",227,"name")="shm_snapshot_t.shadow_file_header.wc_blocked_db_csh_getn_wcsstarvewrt_tn"
	Set gtmtypes("shm_snapshot_t",227,"off")=6528
	Set gtmtypes("shm_snapshot_t",227,"len")=8
	Set gtmtypes("shm_snapshot_t",227,"type")="uint64_t"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.wc_blocked_db_csh_getn_wcsstarvewrt_tn")=227
	Set gtmtypes("shm_snapshot_t",228,"name")="shm_snapshot_t.shadow_file_header.wc_blocked_db_csh_get_tn"
	Set gtmtypes("shm_snapshot_t",228,"off")=6536
	Set gtmtypes("shm_snapshot_t",228,"len")=8
	Set gtmtypes("shm_snapshot_t",228,"type")="uint64_t"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.wc_blocked_db_csh_get_tn")=228
	Set gtmtypes("shm_snapshot_t",229,"name")="shm_snapshot_t.shadow_file_header.wc_blocked_tp_tend_wcsgetspace_tn"
	Set gtmtypes("shm_snapshot_t",229,"off")=6544
	Set gtmtypes("shm_snapshot_t",229,"len")=8
	Set gtmtypes("shm_snapshot_t",229,"type")="uint64_t"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.wc_blocked_tp_tend_wcsgetspace_tn")=229
	Set gtmtypes("shm_snapshot_t",230,"name")="shm_snapshot_t.shadow_file_header.wc_blocked_tp_tend_t1_tn"
	Set gtmtypes("shm_snapshot_t",230,"off")=6552
	Set gtmtypes("shm_snapshot_t",230,"len")=8
	Set gtmtypes("shm_snapshot_t",230,"type")="uint64_t"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.wc_blocked_tp_tend_t1_tn")=230
	Set gtmtypes("shm_snapshot_t",231,"name")="shm_snapshot_t.shadow_file_header.wc_blocked_tp_tend_bitmap_tn"
	Set gtmtypes("shm_snapshot_t",231,"off")=6560
	Set gtmtypes("shm_snapshot_t",231,"len")=8
	Set gtmtypes("shm_snapshot_t",231,"type")="uint64_t"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.wc_blocked_tp_tend_bitmap_tn")=231
	Set gtmtypes("shm_snapshot_t",232,"name")="shm_snapshot_t.shadow_file_header.wc_blocked_tp_tend_jnl_cwset_tn"
	Set gtmtypes("shm_snapshot_t",232,"off")=6568
	Set gtmtypes("shm_snapshot_t",232,"len")=8
	Set gtmtypes("shm_snapshot_t",232,"type")="uint64_t"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.wc_blocked_tp_tend_jnl_cwset_tn")=232
	Set gtmtypes("shm_snapshot_t",233,"name")="shm_snapshot_t.shadow_file_header.wc_blocked_tp_tend_jnl_wcsflu_tn"
	Set gtmtypes("shm_snapshot_t",233,"off")=6576
	Set gtmtypes("shm_snapshot_t",233,"len")=8
	Set gtmtypes("shm_snapshot_t",233,"type")="uint64_t"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.wc_blocked_tp_tend_jnl_wcsflu_tn")=233
	Set gtmtypes("shm_snapshot_t",234,"name")="shm_snapshot_t.shadow_file_header.wc_blocked_t_end_hist_tn"
	Set gtmtypes("shm_snapshot_t",234,"off")=6584
	Set gtmtypes("shm_snapshot_t",234,"len")=8
	Set gtmtypes("shm_snapshot_t",234,"type")="uint64_t"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.wc_blocked_t_end_hist_tn")=234
	Set gtmtypes("shm_snapshot_t",235,"name")="shm_snapshot_t.shadow_file_header.wc_blocked_t_end_hist1_nullbt_tn"
	Set gtmtypes("shm_snapshot_t",235,"off")=6592
	Set gtmtypes("shm_snapshot_t",235,"len")=8
	Set gtmtypes("shm_snapshot_t",235,"type")="uint64_t"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.wc_blocked_t_end_hist1_nullbt_tn")=235
	Set gtmtypes("shm_snapshot_t",236,"name")="shm_snapshot_t.shadow_file_header.wc_blocked_t_end_hist1_nonnullbt_tn"
	Set gtmtypes("shm_snapshot_t",236,"off")=6600
	Set gtmtypes("shm_snapshot_t",236,"len")=8
	Set gtmtypes("shm_snapshot_t",236,"type")="uint64_t"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.wc_blocked_t_end_hist1_nonnullbt_tn")=236
	Set gtmtypes("shm_snapshot_t",237,"name")="shm_snapshot_t.shadow_file_header.wc_blocked_t_end_bitmap_nullbt_tn"
	Set gtmtypes("shm_snapshot_t",237,"off")=6608
	Set gtmtypes("shm_snapshot_t",237,"len")=8
	Set gtmtypes("shm_snapshot_t",237,"type")="uint64_t"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.wc_blocked_t_end_bitmap_nullbt_tn")=237
	Set gtmtypes("shm_snapshot_t",238,"name")="shm_snapshot_t.shadow_file_header.wc_blocked_t_end_bitmap_nonnullbt_tn"
	Set gtmtypes("shm_snapshot_t",238,"off")=6616
	Set gtmtypes("shm_snapshot_t",238,"len")=8
	Set gtmtypes("shm_snapshot_t",238,"type")="uint64_t"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.wc_blocked_t_end_bitmap_nonnullbt_tn")=238
	Set gtmtypes("shm_snapshot_t",239,"name")="shm_snapshot_t.shadow_file_header.wc_blocked_t_end_jnl_cwset_tn"
	Set gtmtypes("shm_snapshot_t",239,"off")=6624
	Set gtmtypes("shm_snapshot_t",239,"len")=8
	Set gtmtypes("shm_snapshot_t",239,"type")="uint64_t"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.wc_blocked_t_end_jnl_cwset_tn")=239
	Set gtmtypes("shm_snapshot_t",240,"name")="shm_snapshot_t.shadow_file_header.wc_blocked_t_end_jnl_wcsflu_tn"
	Set gtmtypes("shm_snapshot_t",240,"off")=6632
	Set gtmtypes("shm_snapshot_t",240,"len")=8
	Set gtmtypes("shm_snapshot_t",240,"type")="uint64_t"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.wc_blocked_t_end_jnl_wcsflu_tn")=240
	Set gtmtypes("shm_snapshot_t",241,"name")="shm_snapshot_t.shadow_file_header.unused_filler_0052_tn"
	Set gtmtypes("shm_snapshot_t",241,"off")=6640
	Set gtmtypes("shm_snapshot_t",241,"len")=8
	Set gtmtypes("shm_snapshot_t",241,"type")="uint64_t"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.unused_filler_0052_tn")=241
	Set gtmtypes("shm_snapshot_t",242,"name")="shm_snapshot_t.shadow_file_header.wc_blocked_tpckh_hist1_nullbt_tn"
	Set gtmtypes("shm_snapshot_t",242,"off")=6648
	Set gtmtypes("shm_snapshot_t",242,"len")=8
	Set gtmtypes("shm_snapshot_t",242,"type")="uint64_t"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.wc_blocked_tpckh_hist1_nullbt_tn")=242
	Set gtmtypes("shm_snapshot_t",243,"name")="shm_snapshot_t.shadow_file_header.wc_blocked_tpckh_hist1_nonnullbt_tn"
	Set gtmtypes("shm_snapshot_t",243,"off")=6656
	Set gtmtypes("shm_snapshot_t",243,"len")=8
	Set gtmtypes("shm_snapshot_t",243,"type")="uint64_t"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.wc_blocked_tpckh_hist1_nonnullbt_tn")=243
	Set gtmtypes("shm_snapshot_t",244,"name")="shm_snapshot_t.shadow_file_header.jnl_blocked_writer_lost_tn"
	Set gtmtypes("shm_snapshot_t",244,"off")=6664
	Set gtmtypes("shm_snapshot_t",244,"len")=8
	Set gtmtypes("shm_snapshot_t",244,"type")="uint64_t"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.jnl_blocked_writer_lost_tn")=244
	Set gtmtypes("shm_snapshot_t",245,"name")="shm_snapshot_t.shadow_file_header.jnl_blocked_writer_stuck_tn"
	Set gtmtypes("shm_snapshot_t",245,"off")=6672
	Set gtmtypes("shm_snapshot_t",245,"len")=8
	Set gtmtypes("shm_snapshot_t",245,"type")="uint64_t"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.jnl_blocked_writer_stuck_tn")=245
	Set gtmtypes("shm_snapshot_t",246,"name")="shm_snapshot_t.shadow_file_header.jnl_blocked_writer_blocked_tn"
	Set gtmtypes("shm_snapshot_t",246,"off")=6680
	Set gtmtypes("shm_snapshot_t",246,"len")=8
	Set gtmtypes("shm_snapshot_t",246,"type")="uint64_t"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.jnl_blocked_writer_blocked_tn")=246
	Set gtmtypes("shm_snapshot_t",247,"name")="shm_snapshot_t.shadow_file_header.n_jnl_fsyncs_tn"
	Set gtmtypes("shm_snapshot_t",247,"off")=6688
	Set gtmtypes("shm_snapshot_t",247,"len")=8
	Set gtmtypes("shm_snapshot_t",247,"type")="uint64_t"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.n_jnl_fsyncs_tn")=247
	Set gtmtypes("shm_snapshot_t",248,"name")="shm_snapshot_t.shadow_file_header.n_jnl_fsync_tries_tn"
	Set gtmtypes("shm_snapshot_t",248,"off")=6696
	Set gtmtypes("shm_snapshot_t",248,"len")=8
	Set gtmtypes("shm_snapshot_t",248,"type")="uint64_t"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.n_jnl_fsync_tries_tn")=248
	Set gtmtypes("shm_snapshot_t",249,"name")="shm_snapshot_t.shadow_file_header.n_jnl_fsync_recovers_tn"
	Set gtmtypes("shm_snapshot_t",249,"off")=6704
	Set gtmtypes("shm_snapshot_t",249,"len")=8
	Set gtmtypes("shm_snapshot_t",249,"type")="uint64_t"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.n_jnl_fsync_recovers_tn")=249
	Set gtmtypes("shm_snapshot_t",250,"name")="shm_snapshot_t.shadow_file_header.n_db_fsyncs_tn"
	Set gtmtypes("shm_snapshot_t",250,"off")=6712
	Set gtmtypes("shm_snapshot_t",250,"len")=8
	Set gtmtypes("shm_snapshot_t",250,"type")="uint64_t"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.n_db_fsyncs_tn")=250
	Set gtmtypes("shm_snapshot_t",251,"name")="shm_snapshot_t.shadow_file_header.n_db_fsyncs_in_crit_tn"
	Set gtmtypes("shm_snapshot_t",251,"off")=6720
	Set gtmtypes("shm_snapshot_t",251,"len")=8
	Set gtmtypes("shm_snapshot_t",251,"type")="uint64_t"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.n_db_fsyncs_in_crit_tn")=251
	Set gtmtypes("shm_snapshot_t",252,"name")="shm_snapshot_t.shadow_file_header.n_dbsync_timers_tn"
	Set gtmtypes("shm_snapshot_t",252,"off")=6728
	Set gtmtypes("shm_snapshot_t",252,"len")=8
	Set gtmtypes("shm_snapshot_t",252,"type")="uint64_t"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.n_dbsync_timers_tn")=252
	Set gtmtypes("shm_snapshot_t",253,"name")="shm_snapshot_t.shadow_file_header.n_dbsync_writes_tn"
	Set gtmtypes("shm_snapshot_t",253,"off")=6736
	Set gtmtypes("shm_snapshot_t",253,"len")=8
	Set gtmtypes("shm_snapshot_t",253,"type")="uint64_t"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.n_dbsync_writes_tn")=253
	Set gtmtypes("shm_snapshot_t",254,"name")="shm_snapshot_t.shadow_file_header.mutex_queue_full_tn"
	Set gtmtypes("shm_snapshot_t",254,"off")=6744
	Set gtmtypes("shm_snapshot_t",254,"len")=8
	Set gtmtypes("shm_snapshot_t",254,"type")="uint64_t"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.mutex_queue_full_tn")=254
	Set gtmtypes("shm_snapshot_t",255,"name")="shm_snapshot_t.shadow_file_header.wcb_bt_put_tn"
	Set gtmtypes("shm_snapshot_t",255,"off")=6752
	Set gtmtypes("shm_snapshot_t",255,"len")=8
	Set gtmtypes("shm_snapshot_t",255,"type")="uint64_t"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.wcb_bt_put_tn")=255
	Set gtmtypes("shm_snapshot_t",256,"name")="shm_snapshot_t.shadow_file_header.wcb_mutex_salvage_tn"
	Set gtmtypes("shm_snapshot_t",256,"off")=6760
	Set gtmtypes("shm_snapshot_t",256,"len")=8
	Set gtmtypes("shm_snapshot_t",256,"type")="uint64_t"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.wcb_mutex_salvage_tn")=256
	Set gtmtypes("shm_snapshot_t",257,"name")="shm_snapshot_t.shadow_file_header.wcb_wcs_flu0_tn"
	Set gtmtypes("shm_snapshot_t",257,"off")=6768
	Set gtmtypes("shm_snapshot_t",257,"len")=8
	Set gtmtypes("shm_snapshot_t",257,"type")="uint64_t"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.wcb_wcs_flu0_tn")=257
	Set gtmtypes("shm_snapshot_t",258,"name")="shm_snapshot_t.shadow_file_header.wcb_t_end_sysops_nocr_invcr_tn"
	Set gtmtypes("shm_snapshot_t",258,"off")=6776
	Set gtmtypes("shm_snapshot_t",258,"len")=8
	Set gtmtypes("shm_snapshot_t",258,"type")="uint64_t"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.wcb_t_end_sysops_nocr_invcr_tn")=258
	Set gtmtypes("shm_snapshot_t",259,"name")="shm_snapshot_t.shadow_file_header.wcb_t_end_sysops_cr_invcr_tn"
	Set gtmtypes("shm_snapshot_t",259,"off")=6784
	Set gtmtypes("shm_snapshot_t",259,"len")=8
	Set gtmtypes("shm_snapshot_t",259,"type")="uint64_t"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.wcb_t_end_sysops_cr_invcr_tn")=259
	Set gtmtypes("shm_snapshot_t",260,"name")="shm_snapshot_t.shadow_file_header.wcb_t_end_sysops_rip_wait_tn"
	Set gtmtypes("shm_snapshot_t",260,"off")=6792
	Set gtmtypes("shm_snapshot_t",260,"len")=8
	Set gtmtypes("shm_snapshot_t",260,"type")="uint64_t"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.wcb_t_end_sysops_rip_wait_tn")=260
	Set gtmtypes("shm_snapshot_t",261,"name")="shm_snapshot_t.shadow_file_header.wcb_t_end_sysops_dirtyripwait_tn"
	Set gtmtypes("shm_snapshot_t",261,"off")=6800
	Set gtmtypes("shm_snapshot_t",261,"len")=8
	Set gtmtypes("shm_snapshot_t",261,"type")="uint64_t"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.wcb_t_end_sysops_dirtyripwait_tn")=261
	Set gtmtypes("shm_snapshot_t",262,"name")="shm_snapshot_t.shadow_file_header.wcb_gds_rundown1_tn"
	Set gtmtypes("shm_snapshot_t",262,"off")=6808
	Set gtmtypes("shm_snapshot_t",262,"len")=8
	Set gtmtypes("shm_snapshot_t",262,"type")="uint64_t"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.wcb_gds_rundown1_tn")=262
	Set gtmtypes("shm_snapshot_t",263,"name")="shm_snapshot_t.shadow_file_header.wcb_wcs_flu1_tn"
	Set gtmtypes("shm_snapshot_t",263,"off")=6816
	Set gtmtypes("shm_snapshot_t",263,"len")=8
	Set gtmtypes("shm_snapshot_t",263,"type")="uint64_t"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.wcb_wcs_flu1_tn")=263
	Set gtmtypes("shm_snapshot_t",264,"name")="shm_snapshot_t.shadow_file_header.wcb_mu_back_tn"
	Set gtmtypes("shm_snapshot_t",264,"off")=6824
	Set gtmtypes("shm_snapshot_t",264,"len")=8
	Set gtmtypes("shm_snapshot_t",264,"type")="uint64_t"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.wcb_mu_back_tn")=264
	Set gtmtypes("shm_snapshot_t",265,"name")="shm_snapshot_t.shadow_file_header.wcb_t_end_sysops_dirty_invcr_tn"
	Set gtmtypes("shm_snapshot_t",265,"off")=6832
	Set gtmtypes("shm_snapshot_t",265,"len")=8
	Set gtmtypes("shm_snapshot_t",265,"type")="uint64_t"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.wcb_t_end_sysops_dirty_invcr_tn")=265
	Set gtmtypes("shm_snapshot_t",266,"name")="shm_snapshot_t.shadow_file_header.wcb_t_end_sysops_wtfini_fail_tn"
	Set gtmtypes("shm_snapshot_t",266,"off")=6840
	Set gtmtypes("shm_snapshot_t",266,"len")=8
	Set gtmtypes("shm_snapshot_t",266,"type")="uint64_t"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.wcb_t_end_sysops_wtfini_fail_tn")=266
	Set gtmtypes("shm_snapshot_t",267,"name")="shm_snapshot_t.shadow_file_header.wcb_t_end_sysops_twin_stuck_tn"
	Set gtmtypes("shm_snapshot_t",267,"off")=6848
	Set gtmtypes("shm_snapshot_t",267,"len")=8
	Set gtmtypes("shm_snapshot_t",267,"type")="uint64_t"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.wcb_t_end_sysops_twin_stuck_tn")=267
	Set gtmtypes("shm_snapshot_t",268,"name")="shm_snapshot_t.shadow_file_header.db_csh_getn_wrt_latch_stuck_tn"
	Set gtmtypes("shm_snapshot_t",268,"off")=6856
	Set gtmtypes("shm_snapshot_t",268,"len")=8
	Set gtmtypes("shm_snapshot_t",268,"type")="uint64_t"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.db_csh_getn_wrt_latch_stuck_tn")=268
	Set gtmtypes("shm_snapshot_t",269,"name")="shm_snapshot_t.shadow_file_header.wcb_secshr_db_clnup_now_crit_tn"
	Set gtmtypes("shm_snapshot_t",269,"off")=6864
	Set gtmtypes("shm_snapshot_t",269,"len")=8
	Set gtmtypes("shm_snapshot_t",269,"type")="uint64_t"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.wcb_secshr_db_clnup_now_crit_tn")=269
	Set gtmtypes("shm_snapshot_t",270,"name")="shm_snapshot_t.shadow_file_header.wc_blocked_t_qread_bad_bt_index1_tn"
	Set gtmtypes("shm_snapshot_t",270,"off")=6872
	Set gtmtypes("shm_snapshot_t",270,"len")=8
	Set gtmtypes("shm_snapshot_t",270,"type")="uint64_t"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.wc_blocked_t_qread_bad_bt_index1_tn")=270
	Set gtmtypes("shm_snapshot_t",271,"name")="shm_snapshot_t.shadow_file_header.wc_blocked_t_qread_bad_bt_index2_tn"
	Set gtmtypes("shm_snapshot_t",271,"off")=6880
	Set gtmtypes("shm_snapshot_t",271,"len")=8
	Set gtmtypes("shm_snapshot_t",271,"type")="uint64_t"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.wc_blocked_t_qread_bad_bt_index2_tn")=271
	Set gtmtypes("shm_snapshot_t",272,"name")="shm_snapshot_t.shadow_file_header.wc_blocked_t_end_crbtmismatch1_tn"
	Set gtmtypes("shm_snapshot_t",272,"off")=6888
	Set gtmtypes("shm_snapshot_t",272,"len")=8
	Set gtmtypes("shm_snapshot_t",272,"type")="uint64_t"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.wc_blocked_t_end_crbtmismatch1_tn")=272
	Set gtmtypes("shm_snapshot_t",273,"name")="shm_snapshot_t.shadow_file_header.wc_blocked_t_end_crbtmismatch2_tn"
	Set gtmtypes("shm_snapshot_t",273,"off")=6896
	Set gtmtypes("shm_snapshot_t",273,"len")=8
	Set gtmtypes("shm_snapshot_t",273,"type")="uint64_t"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.wc_blocked_t_end_crbtmismatch2_tn")=273
	Set gtmtypes("shm_snapshot_t",274,"name")="shm_snapshot_t.shadow_file_header.wc_blocked_t_end_crbtmismatch3_tn"
	Set gtmtypes("shm_snapshot_t",274,"off")=6904
	Set gtmtypes("shm_snapshot_t",274,"len")=8
	Set gtmtypes("shm_snapshot_t",274,"type")="uint64_t"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.wc_blocked_t_end_crbtmismatch3_tn")=274
	Set gtmtypes("shm_snapshot_t",275,"name")="shm_snapshot_t.shadow_file_header.wc_blocked_tp_tend_crbtmismatch1_tn"
	Set gtmtypes("shm_snapshot_t",275,"off")=6912
	Set gtmtypes("shm_snapshot_t",275,"len")=8
	Set gtmtypes("shm_snapshot_t",275,"type")="uint64_t"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.wc_blocked_tp_tend_crbtmismatch1_tn")=275
	Set gtmtypes("shm_snapshot_t",276,"name")="shm_snapshot_t.shadow_file_header.wc_blocked_tp_tend_crbtmismatch2_tn"
	Set gtmtypes("shm_snapshot_t",276,"off")=6920
	Set gtmtypes("shm_snapshot_t",276,"len")=8
	Set gtmtypes("shm_snapshot_t",276,"type")="uint64_t"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.wc_blocked_tp_tend_crbtmismatch2_tn")=276
	Set gtmtypes("shm_snapshot_t",277,"name")="shm_snapshot_t.shadow_file_header.wc_blocked_tp_tend_crbtmismatch3_tn"
	Set gtmtypes("shm_snapshot_t",277,"off")=6928
	Set gtmtypes("shm_snapshot_t",277,"len")=8
	Set gtmtypes("shm_snapshot_t",277,"type")="uint64_t"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.wc_blocked_tp_tend_crbtmismatch3_tn")=277
	Set gtmtypes("shm_snapshot_t",278,"name")="shm_snapshot_t.shadow_file_header.wc_blocked_wcs_wtstart_bad_cr_tn"
	Set gtmtypes("shm_snapshot_t",278,"off")=6936
	Set gtmtypes("shm_snapshot_t",278,"len")=8
	Set gtmtypes("shm_snapshot_t",278,"type")="uint64_t"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.wc_blocked_wcs_wtstart_bad_cr_tn")=278
	Set gtmtypes("shm_snapshot_t",279,"name")="shm_snapshot_t.shadow_file_header.wc_blocked_wcs_wtfini_bad_cr_tn"
	Set gtmtypes("shm_snapshot_t",279,"off")=6944
	Set gtmtypes("shm_snapshot_t",279,"len")=8
	Set gtmtypes("shm_snapshot_t",279,"type")="uint64_t"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.wc_blocked_wcs_wtfini_bad_cr_tn")=279
	Set gtmtypes("shm_snapshot_t",280,"name")="shm_snapshot_t.shadow_file_header.wc_blocked_bt_get_tn"
	Set gtmtypes("shm_snapshot_t",280,"off")=6952
	Set gtmtypes("shm_snapshot_t",280,"len")=8
	Set gtmtypes("shm_snapshot_t",280,"type")="uint64_t"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.wc_blocked_bt_get_tn")=280
	Set gtmtypes("shm_snapshot_t",281,"name")="shm_snapshot_t.shadow_file_header.wc_blocked_wcs_cdb_sc_final_retry_tn"
	Set gtmtypes("shm_snapshot_t",281,"off")=6960
	Set gtmtypes("shm_snapshot_t",281,"len")=8
	Set gtmtypes("shm_snapshot_t",281,"type")="uint64_t"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.wc_blocked_wcs_cdb_sc_final_retry_tn")=281
	Set gtmtypes("shm_snapshot_t",282,"name")="shm_snapshot_t.shadow_file_header.wcb_bg_update_lckfail1_tn"
	Set gtmtypes("shm_snapshot_t",282,"off")=6968
	Set gtmtypes("shm_snapshot_t",282,"len")=8
	Set gtmtypes("shm_snapshot_t",282,"type")="uint64_t"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.wcb_bg_update_lckfail1_tn")=282
	Set gtmtypes("shm_snapshot_t",283,"name")="shm_snapshot_t.shadow_file_header.wcb_bg_update_lckfail2_tn"
	Set gtmtypes("shm_snapshot_t",283,"off")=6976
	Set gtmtypes("shm_snapshot_t",283,"len")=8
	Set gtmtypes("shm_snapshot_t",283,"type")="uint64_t"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.wcb_bg_update_lckfail2_tn")=283
	Set gtmtypes("shm_snapshot_t",284,"name")="shm_snapshot_t.shadow_file_header.wcb_wtstart_lckfail1_tn"
	Set gtmtypes("shm_snapshot_t",284,"off")=6984
	Set gtmtypes("shm_snapshot_t",284,"len")=8
	Set gtmtypes("shm_snapshot_t",284,"type")="uint64_t"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.wcb_wtstart_lckfail1_tn")=284
	Set gtmtypes("shm_snapshot_t",285,"name")="shm_snapshot_t.shadow_file_header.wcb_wtstart_lckfail2_tn"
	Set gtmtypes("shm_snapshot_t",285,"off")=6992
	Set gtmtypes("shm_snapshot_t",285,"len")=8
	Set gtmtypes("shm_snapshot_t",285,"type")="uint64_t"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.wcb_wtstart_lckfail2_tn")=285
	Set gtmtypes("shm_snapshot_t",286,"name")="shm_snapshot_t.shadow_file_header.wcb_wtstart_lckfail3_tn"
	Set gtmtypes("shm_snapshot_t",286,"off")=7000
	Set gtmtypes("shm_snapshot_t",286,"len")=8
	Set gtmtypes("shm_snapshot_t",286,"type")="uint64_t"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.wcb_wtstart_lckfail3_tn")=286
	Set gtmtypes("shm_snapshot_t",287,"name")="shm_snapshot_t.shadow_file_header.wcb_wtstart_lckfail4_tn"
	Set gtmtypes("shm_snapshot_t",287,"off")=7008
	Set gtmtypes("shm_snapshot_t",287,"len")=8
	Set gtmtypes("shm_snapshot_t",287,"type")="uint64_t"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.wcb_wtstart_lckfail4_tn")=287
	Set gtmtypes("shm_snapshot_t",288,"name")="shm_snapshot_t.shadow_file_header.wcb_wtfini_lckfail1_tn"
	Set gtmtypes("shm_snapshot_t",288,"off")=7016
	Set gtmtypes("shm_snapshot_t",288,"len")=8
	Set gtmtypes("shm_snapshot_t",288,"type")="uint64_t"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.wcb_wtfini_lckfail1_tn")=288
	Set gtmtypes("shm_snapshot_t",289,"name")="shm_snapshot_t.shadow_file_header.wcb_wtfini_lckfail2_tn"
	Set gtmtypes("shm_snapshot_t",289,"off")=7024
	Set gtmtypes("shm_snapshot_t",289,"len")=8
	Set gtmtypes("shm_snapshot_t",289,"type")="uint64_t"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.wcb_wtfini_lckfail2_tn")=289
	Set gtmtypes("shm_snapshot_t",290,"name")="shm_snapshot_t.shadow_file_header.wcb_wtfini_lckfail3_tn"
	Set gtmtypes("shm_snapshot_t",290,"off")=7032
	Set gtmtypes("shm_snapshot_t",290,"len")=8
	Set gtmtypes("shm_snapshot_t",290,"type")="uint64_t"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.wcb_wtfini_lckfail3_tn")=290
	Set gtmtypes("shm_snapshot_t",291,"name")="shm_snapshot_t.shadow_file_header.wcb_wtfini_lckfail4_tn"
	Set gtmtypes("shm_snapshot_t",291,"off")=7040
	Set gtmtypes("shm_snapshot_t",291,"len")=8
	Set gtmtypes("shm_snapshot_t",291,"type")="uint64_t"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.wcb_wtfini_lckfail4_tn")=291
	Set gtmtypes("shm_snapshot_t",292,"name")="shm_snapshot_t.shadow_file_header.wcb_t_end_sysops_dirtystuck1_tn"
	Set gtmtypes("shm_snapshot_t",292,"off")=7048
	Set gtmtypes("shm_snapshot_t",292,"len")=8
	Set gtmtypes("shm_snapshot_t",292,"type")="uint64_t"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.wcb_t_end_sysops_dirtystuck1_tn")=292
	Set gtmtypes("shm_snapshot_t",293,"name")="shm_snapshot_t.shadow_file_header.wcb_t_end_sysops_dirtystuck2_tn"
	Set gtmtypes("shm_snapshot_t",293,"off")=7056
	Set gtmtypes("shm_snapshot_t",293,"len")=8
	Set gtmtypes("shm_snapshot_t",293,"type")="uint64_t"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.wcb_t_end_sysops_dirtystuck2_tn")=293
	Set gtmtypes("shm_snapshot_t",294,"name")="shm_snapshot_t.shadow_file_header.wcb_secshr_db_clnup_wbuf_dqd_tn"
	Set gtmtypes("shm_snapshot_t",294,"off")=7064
	Set gtmtypes("shm_snapshot_t",294,"len")=8
	Set gtmtypes("shm_snapshot_t",294,"type")="uint64_t"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.wcb_secshr_db_clnup_wbuf_dqd_tn")=294
	Set gtmtypes("shm_snapshot_t",295,"name")="shm_snapshot_t.shadow_file_header.dwngrd_refmts_syncio_tn"
	Set gtmtypes("shm_snapshot_t",295,"off")=7072
	Set gtmtypes("shm_snapshot_t",295,"len")=8
	Set gtmtypes("shm_snapshot_t",295,"type")="uint64_t"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.dwngrd_refmts_syncio_tn")=295
	Set gtmtypes("shm_snapshot_t",296,"name")="shm_snapshot_t.shadow_file_header.dwngrd_refmts_asyncio_tn"
	Set gtmtypes("shm_snapshot_t",296,"off")=7080
	Set gtmtypes("shm_snapshot_t",296,"len")=8
	Set gtmtypes("shm_snapshot_t",296,"type")="uint64_t"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.dwngrd_refmts_asyncio_tn")=296
	Set gtmtypes("shm_snapshot_t",297,"name")="shm_snapshot_t.shadow_file_header.shmpool_refmt_harvests_tn"
	Set gtmtypes("shm_snapshot_t",297,"off")=7088
	Set gtmtypes("shm_snapshot_t",297,"len")=8
	Set gtmtypes("shm_snapshot_t",297,"type")="uint64_t"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.shmpool_refmt_harvests_tn")=297
	Set gtmtypes("shm_snapshot_t",298,"name")="shm_snapshot_t.shadow_file_header.shmpool_recovery_tn"
	Set gtmtypes("shm_snapshot_t",298,"off")=7096
	Set gtmtypes("shm_snapshot_t",298,"len")=8
	Set gtmtypes("shm_snapshot_t",298,"type")="uint64_t"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.shmpool_recovery_tn")=298
	Set gtmtypes("shm_snapshot_t",299,"name")="shm_snapshot_t.shadow_file_header.shmpool_blkd_by_sdc_tn"
	Set gtmtypes("shm_snapshot_t",299,"off")=7104
	Set gtmtypes("shm_snapshot_t",299,"len")=8
	Set gtmtypes("shm_snapshot_t",299,"type")="uint64_t"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.shmpool_blkd_by_sdc_tn")=299
	Set gtmtypes("shm_snapshot_t",300,"name")="shm_snapshot_t.shadow_file_header.shmpool_alloc_bbflush_tn"
	Set gtmtypes("shm_snapshot_t",300,"off")=7112
	Set gtmtypes("shm_snapshot_t",300,"len")=8
	Set gtmtypes("shm_snapshot_t",300,"type")="uint64_t"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.shmpool_alloc_bbflush_tn")=300
	Set gtmtypes("shm_snapshot_t",301,"name")="shm_snapshot_t.shadow_file_header.refmt_hvst_blk_released_replaced_tn"
	Set gtmtypes("shm_snapshot_t",301,"off")=7120
	Set gtmtypes("shm_snapshot_t",301,"len")=8
	Set gtmtypes("shm_snapshot_t",301,"type")="uint64_t"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.refmt_hvst_blk_released_replaced_tn")=301
	Set gtmtypes("shm_snapshot_t",302,"name")="shm_snapshot_t.shadow_file_header.refmt_hvst_blk_released_io_complete_tn"
	Set gtmtypes("shm_snapshot_t",302,"off")=7128
	Set gtmtypes("shm_snapshot_t",302,"len")=8
	Set gtmtypes("shm_snapshot_t",302,"type")="uint64_t"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.refmt_hvst_blk_released_io_complete_tn")=302
	Set gtmtypes("shm_snapshot_t",303,"name")="shm_snapshot_t.shadow_file_header.refmt_hvst_blk_kept_tn"
	Set gtmtypes("shm_snapshot_t",303,"off")=7136
	Set gtmtypes("shm_snapshot_t",303,"len")=8
	Set gtmtypes("shm_snapshot_t",303,"type")="uint64_t"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.refmt_hvst_blk_kept_tn")=303
	Set gtmtypes("shm_snapshot_t",304,"name")="shm_snapshot_t.shadow_file_header.refmt_hvst_blk_ignored_tn"
	Set gtmtypes("shm_snapshot_t",304,"off")=7144
	Set gtmtypes("shm_snapshot_t",304,"len")=8
	Set gtmtypes("shm_snapshot_t",304,"type")="uint64_t"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.refmt_hvst_blk_ignored_tn")=304
	Set gtmtypes("shm_snapshot_t",305,"name")="shm_snapshot_t.shadow_file_header.refmt_blk_chk_blk_freed_tn"
	Set gtmtypes("shm_snapshot_t",305,"off")=7152
	Set gtmtypes("shm_snapshot_t",305,"len")=8
	Set gtmtypes("shm_snapshot_t",305,"type")="uint64_t"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.refmt_blk_chk_blk_freed_tn")=305
	Set gtmtypes("shm_snapshot_t",306,"name")="shm_snapshot_t.shadow_file_header.refmt_blk_chk_blk_kept_tn"
	Set gtmtypes("shm_snapshot_t",306,"off")=7160
	Set gtmtypes("shm_snapshot_t",306,"len")=8
	Set gtmtypes("shm_snapshot_t",306,"type")="uint64_t"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.refmt_blk_chk_blk_kept_tn")=306
	Set gtmtypes("shm_snapshot_t",307,"name")="shm_snapshot_t.shadow_file_header.active_lvl_trigger_tn"
	Set gtmtypes("shm_snapshot_t",307,"off")=7168
	Set gtmtypes("shm_snapshot_t",307,"len")=8
	Set gtmtypes("shm_snapshot_t",307,"type")="uint64_t"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.active_lvl_trigger_tn")=307
	Set gtmtypes("shm_snapshot_t",308,"name")="shm_snapshot_t.shadow_file_header.new_buff_tn"
	Set gtmtypes("shm_snapshot_t",308,"off")=7176
	Set gtmtypes("shm_snapshot_t",308,"len")=8
	Set gtmtypes("shm_snapshot_t",308,"type")="uint64_t"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.new_buff_tn")=308
	Set gtmtypes("shm_snapshot_t",309,"name")="shm_snapshot_t.shadow_file_header.phase2_commit_wait_sleep_in_crit_tn"
	Set gtmtypes("shm_snapshot_t",309,"off")=7184
	Set gtmtypes("shm_snapshot_t",309,"len")=8
	Set gtmtypes("shm_snapshot_t",309,"type")="uint64_t"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.phase2_commit_wait_sleep_in_crit_tn")=309
	Set gtmtypes("shm_snapshot_t",310,"name")="shm_snapshot_t.shadow_file_header.phase2_commit_wait_sleep_no_crit_tn"
	Set gtmtypes("shm_snapshot_t",310,"off")=7192
	Set gtmtypes("shm_snapshot_t",310,"len")=8
	Set gtmtypes("shm_snapshot_t",310,"type")="uint64_t"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.phase2_commit_wait_sleep_no_crit_tn")=310
	Set gtmtypes("shm_snapshot_t",311,"name")="shm_snapshot_t.shadow_file_header.phase2_commit_wait_pidcnt_tn"
	Set gtmtypes("shm_snapshot_t",311,"off")=7200
	Set gtmtypes("shm_snapshot_t",311,"len")=8
	Set gtmtypes("shm_snapshot_t",311,"type")="uint64_t"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.phase2_commit_wait_pidcnt_tn")=311
	Set gtmtypes("shm_snapshot_t",312,"name")="shm_snapshot_t.shadow_file_header.wcb_t_end_sysops_intend_wait_tn"
	Set gtmtypes("shm_snapshot_t",312,"off")=7208
	Set gtmtypes("shm_snapshot_t",312,"len")=8
	Set gtmtypes("shm_snapshot_t",312,"type")="uint64_t"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.wcb_t_end_sysops_intend_wait_tn")=312
	Set gtmtypes("shm_snapshot_t",313,"name")="shm_snapshot_t.shadow_file_header.wcb_secshr_db_clnup_phase2_clnup_tn"
	Set gtmtypes("shm_snapshot_t",313,"off")=7216
	Set gtmtypes("shm_snapshot_t",313,"len")=8
	Set gtmtypes("shm_snapshot_t",313,"type")="uint64_t"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.wcb_secshr_db_clnup_phase2_clnup_tn")=313
	Set gtmtypes("shm_snapshot_t",314,"name")="shm_snapshot_t.shadow_file_header.wcb_phase2_commit_wait_tn"
	Set gtmtypes("shm_snapshot_t",314,"off")=7224
	Set gtmtypes("shm_snapshot_t",314,"len")=8
	Set gtmtypes("shm_snapshot_t",314,"type")="uint64_t"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.wcb_phase2_commit_wait_tn")=314
	Set gtmtypes("shm_snapshot_t",315,"name")="shm_snapshot_t.shadow_file_header.recompute_upd_array_calls_tn"
	Set gtmtypes("shm_snapshot_t",315,"off")=7232
	Set gtmtypes("shm_snapshot_t",315,"len")=8
	Set gtmtypes("shm_snapshot_t",315,"type")="uint64_t"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.recompute_upd_array_calls_tn")=315
	Set gtmtypes("shm_snapshot_t",316,"name")="shm_snapshot_t.shadow_file_header.recompute_upd_array_rip_tn"
	Set gtmtypes("shm_snapshot_t",316,"off")=7240
	Set gtmtypes("shm_snapshot_t",316,"len")=8
	Set gtmtypes("shm_snapshot_t",316,"type")="uint64_t"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.recompute_upd_array_rip_tn")=316
	Set gtmtypes("shm_snapshot_t",317,"name")="shm_snapshot_t.shadow_file_header.recompute_upd_array_in_tend_tn"
	Set gtmtypes("shm_snapshot_t",317,"off")=7248
	Set gtmtypes("shm_snapshot_t",317,"len")=8
	Set gtmtypes("shm_snapshot_t",317,"type")="uint64_t"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.recompute_upd_array_in_tend_tn")=317
	Set gtmtypes("shm_snapshot_t",318,"name")="shm_snapshot_t.shadow_file_header.recompute_upd_array_search_blk_tn"
	Set gtmtypes("shm_snapshot_t",318,"off")=7256
	Set gtmtypes("shm_snapshot_t",318,"len")=8
	Set gtmtypes("shm_snapshot_t",318,"type")="uint64_t"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.recompute_upd_array_search_blk_tn")=318
	Set gtmtypes("shm_snapshot_t",319,"name")="shm_snapshot_t.shadow_file_header.recompute_upd_array_new_rec_tn"
	Set gtmtypes("shm_snapshot_t",319,"off")=7264
	Set gtmtypes("shm_snapshot_t",319,"len")=8
	Set gtmtypes("shm_snapshot_t",319,"type")="uint64_t"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.recompute_upd_array_new_rec_tn")=319
	Set gtmtypes("shm_snapshot_t",320,"name")="shm_snapshot_t.shadow_file_header.recompute_upd_array_rec_size_tn"
	Set gtmtypes("shm_snapshot_t",320,"off")=7272
	Set gtmtypes("shm_snapshot_t",320,"len")=8
	Set gtmtypes("shm_snapshot_t",320,"type")="uint64_t"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.recompute_upd_array_rec_size_tn")=320
	Set gtmtypes("shm_snapshot_t",321,"name")="shm_snapshot_t.shadow_file_header.recompute_upd_array_rec_cmpc_tn"
	Set gtmtypes("shm_snapshot_t",321,"off")=7280
	Set gtmtypes("shm_snapshot_t",321,"len")=8
	Set gtmtypes("shm_snapshot_t",321,"type")="uint64_t"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.recompute_upd_array_rec_cmpc_tn")=321
	Set gtmtypes("shm_snapshot_t",322,"name")="shm_snapshot_t.shadow_file_header.recompute_upd_array_blk_fini_tn"
	Set gtmtypes("shm_snapshot_t",322,"off")=7288
	Set gtmtypes("shm_snapshot_t",322,"len")=8
	Set gtmtypes("shm_snapshot_t",322,"type")="uint64_t"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.recompute_upd_array_blk_fini_tn")=322
	Set gtmtypes("shm_snapshot_t",323,"name")="shm_snapshot_t.shadow_file_header.recompute_upd_array_blk_split_tn"
	Set gtmtypes("shm_snapshot_t",323,"off")=7296
	Set gtmtypes("shm_snapshot_t",323,"len")=8
	Set gtmtypes("shm_snapshot_t",323,"type")="uint64_t"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.recompute_upd_array_blk_split_tn")=323
	Set gtmtypes("shm_snapshot_t",324,"name")="shm_snapshot_t.shadow_file_header.unused_filler_0135_tn"
	Set gtmtypes("shm_snapshot_t",324,"off")=7304
	Set gtmtypes("shm_snapshot_t",324,"len")=8
	Set gtmtypes("shm_snapshot_t",324,"type")="uint64_t"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.unused_filler_0135_tn")=324
	Set gtmtypes("shm_snapshot_t",325,"name")="shm_snapshot_t.shadow_file_header.t_qread_ripsleep_nblks_tn"
	Set gtmtypes("shm_snapshot_t",325,"off")=7312
	Set gtmtypes("shm_snapshot_t",325,"len")=8
	Set gtmtypes("shm_snapshot_t",325,"type")="uint64_t"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.t_qread_ripsleep_nblks_tn")=325
	Set gtmtypes("shm_snapshot_t",326,"name")="shm_snapshot_t.shadow_file_header.wcs_wtfini_invoked_tn"
	Set gtmtypes("shm_snapshot_t",326,"off")=7320
	Set gtmtypes("shm_snapshot_t",326,"len")=8
	Set gtmtypes("shm_snapshot_t",326,"type")="uint64_t"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.wcs_wtfini_invoked_tn")=326
	Set gtmtypes("shm_snapshot_t",327,"name")="shm_snapshot_t.shadow_file_header.wcs_wterror_invoked_tn"
	Set gtmtypes("shm_snapshot_t",327,"off")=7328
	Set gtmtypes("shm_snapshot_t",327,"len")=8
	Set gtmtypes("shm_snapshot_t",327,"type")="uint64_t"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.wcs_wterror_invoked_tn")=327
	Set gtmtypes("shm_snapshot_t",328,"name")="shm_snapshot_t.shadow_file_header.wcs_wt_restart_invoked_tn"
	Set gtmtypes("shm_snapshot_t",328,"off")=7336
	Set gtmtypes("shm_snapshot_t",328,"len")=8
	Set gtmtypes("shm_snapshot_t",328,"type")="uint64_t"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.wcs_wt_restart_invoked_tn")=328
	Set gtmtypes("shm_snapshot_t",329,"name")="shm_snapshot_t.shadow_file_header.wcs_wt_restart_eagain_tn"
	Set gtmtypes("shm_snapshot_t",329,"off")=7344
	Set gtmtypes("shm_snapshot_t",329,"len")=8
	Set gtmtypes("shm_snapshot_t",329,"type")="uint64_t"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.wcs_wt_restart_eagain_tn")=329
	Set gtmtypes("shm_snapshot_t",330,"name")="shm_snapshot_t.shadow_file_header.wcs_wt_restart_reinsert_tn"
	Set gtmtypes("shm_snapshot_t",330,"off")=7352
	Set gtmtypes("shm_snapshot_t",330,"len")=8
	Set gtmtypes("shm_snapshot_t",330,"type")="uint64_t"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.wcs_wt_restart_reinsert_tn")=330
	Set gtmtypes("shm_snapshot_t",331,"name")="shm_snapshot_t.shadow_file_header.wcs_wtstart_eagain_tn"
	Set gtmtypes("shm_snapshot_t",331,"off")=7360
	Set gtmtypes("shm_snapshot_t",331,"len")=8
	Set gtmtypes("shm_snapshot_t",331,"type")="uint64_t"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.wcs_wtstart_eagain_tn")=331
	Set gtmtypes("shm_snapshot_t",332,"name")="shm_snapshot_t.shadow_file_header.wcs_wtstart_eagain_incrit_tn"
	Set gtmtypes("shm_snapshot_t",332,"off")=7368
	Set gtmtypes("shm_snapshot_t",332,"len")=8
	Set gtmtypes("shm_snapshot_t",332,"type")="uint64_t"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.wcs_wtstart_eagain_incrit_tn")=332
	Set gtmtypes("shm_snapshot_t",333,"name")="shm_snapshot_t.shadow_file_header.async_restart_eagain_tn"
	Set gtmtypes("shm_snapshot_t",333,"off")=7376
	Set gtmtypes("shm_snapshot_t",333,"len")=8
	Set gtmtypes("shm_snapshot_t",333,"type")="uint64_t"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.async_restart_eagain_tn")=333
	Set gtmtypes("shm_snapshot_t",334,"name")="shm_snapshot_t.shadow_file_header.jnlbuff_phs2cmt_array_full_tn"
	Set gtmtypes("shm_snapshot_t",334,"off")=7384
	Set gtmtypes("shm_snapshot_t",334,"len")=8
	Set gtmtypes("shm_snapshot_t",334,"type")="uint64_t"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.jnlbuff_phs2cmt_array_full_tn")=334
	Set gtmtypes("shm_snapshot_t",335,"name")="shm_snapshot_t.shadow_file_header.jnl_phase2_cleanup_if_possible_tn"
	Set gtmtypes("shm_snapshot_t",335,"off")=7392
	Set gtmtypes("shm_snapshot_t",335,"len")=8
	Set gtmtypes("shm_snapshot_t",335,"type")="uint64_t"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.jnl_phase2_cleanup_if_possible_tn")=335
	Set gtmtypes("shm_snapshot_t",336,"name")="shm_snapshot_t.shadow_file_header.jnlbuff_phs2cmt_isprcalv_tn"
	Set gtmtypes("shm_snapshot_t",336,"off")=7400
	Set gtmtypes("shm_snapshot_t",336,"len")=8
	Set gtmtypes("shm_snapshot_t",336,"type")="uint64_t"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.jnlbuff_phs2cmt_isprcalv_tn")=336
	Set gtmtypes("shm_snapshot_t",337,"name")="shm_snapshot_t.shadow_file_header.recompute_upd_array_paidoff_tn"
	Set gtmtypes("shm_snapshot_t",337,"off")=7408
	Set gtmtypes("shm_snapshot_t",337,"len")=8
	Set gtmtypes("shm_snapshot_t",337,"type")="uint64_t"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.recompute_upd_array_paidoff_tn")=337
	Set gtmtypes("shm_snapshot_t",338,"name")="shm_snapshot_t.shadow_file_header.bg_trc_rec_tn_filler"
	Set gtmtypes("shm_snapshot_t",338,"off")=7416
	Set gtmtypes("shm_snapshot_t",338,"len")=16
	Set gtmtypes("shm_snapshot_t",338,"type")="char"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.bg_trc_rec_tn_filler")=338
	Set gtmtypes("shm_snapshot_t",339,"name")="shm_snapshot_t.shadow_file_header.total_buffer_flush_cntr"
	Set gtmtypes("shm_snapshot_t",339,"off")=7432
	Set gtmtypes("shm_snapshot_t",339,"len")=4
	Set gtmtypes("shm_snapshot_t",339,"type")="int"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.total_buffer_flush_cntr")=339
	Set gtmtypes("shm_snapshot_t",340,"name")="shm_snapshot_t.shadow_file_header.bufct_buffer_flush_cntr"
	Set gtmtypes("shm_snapshot_t",340,"off")=7436
	Set gtmtypes("shm_snapshot_t",340,"len")=4
	Set gtmtypes("shm_snapshot_t",340,"type")="int"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.bufct_buffer_flush_cntr")=340
	Set gtmtypes("shm_snapshot_t",341,"name")="shm_snapshot_t.shadow_file_header.bufct_buffer_flush_loop_cntr"
	Set gtmtypes("shm_snapshot_t",341,"off")=7440
	Set gtmtypes("shm_snapshot_t",341,"len")=4
	Set gtmtypes("shm_snapshot_t",341,"type")="int"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.bufct_buffer_flush_loop_cntr")=341
	Set gtmtypes("shm_snapshot_t",342,"name")="shm_snapshot_t.shadow_file_header.wrt_calls_cntr"
	Set gtmtypes("shm_snapshot_t",342,"off")=7444
	Set gtmtypes("shm_snapshot_t",342,"len")=4
	Set gtmtypes("shm_snapshot_t",342,"type")="int"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.wrt_calls_cntr")=342
	Set gtmtypes("shm_snapshot_t",343,"name")="shm_snapshot_t.shadow_file_header.wrt_count_cntr"
	Set gtmtypes("shm_snapshot_t",343,"off")=7448
	Set gtmtypes("shm_snapshot_t",343,"len")=4
	Set gtmtypes("shm_snapshot_t",343,"type")="int"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.wrt_count_cntr")=343
	Set gtmtypes("shm_snapshot_t",344,"name")="shm_snapshot_t.shadow_file_header.wrt_blocked_cntr"
	Set gtmtypes("shm_snapshot_t",344,"off")=7452
	Set gtmtypes("shm_snapshot_t",344,"len")=4
	Set gtmtypes("shm_snapshot_t",344,"type")="int"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.wrt_blocked_cntr")=344
	Set gtmtypes("shm_snapshot_t",345,"name")="shm_snapshot_t.shadow_file_header.wrt_busy_cntr"
	Set gtmtypes("shm_snapshot_t",345,"off")=7456
	Set gtmtypes("shm_snapshot_t",345,"len")=4
	Set gtmtypes("shm_snapshot_t",345,"type")="int"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.wrt_busy_cntr")=345
	Set gtmtypes("shm_snapshot_t",346,"name")="shm_snapshot_t.shadow_file_header.wrt_noblks_wrtn_cntr"
	Set gtmtypes("shm_snapshot_t",346,"off")=7460
	Set gtmtypes("shm_snapshot_t",346,"len")=4
	Set gtmtypes("shm_snapshot_t",346,"type")="int"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.wrt_noblks_wrtn_cntr")=346
	Set gtmtypes("shm_snapshot_t",347,"name")="shm_snapshot_t.shadow_file_header.wcb_gds_rundown2_cntr"
	Set gtmtypes("shm_snapshot_t",347,"off")=7464
	Set gtmtypes("shm_snapshot_t",347,"len")=4
	Set gtmtypes("shm_snapshot_t",347,"type")="int"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.wcb_gds_rundown2_cntr")=347
	Set gtmtypes("shm_snapshot_t",348,"name")="shm_snapshot_t.shadow_file_header.reserved_bgtrcrec2_cntr"
	Set gtmtypes("shm_snapshot_t",348,"off")=7468
	Set gtmtypes("shm_snapshot_t",348,"len")=4
	Set gtmtypes("shm_snapshot_t",348,"type")="int"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.reserved_bgtrcrec2_cntr")=348
	Set gtmtypes("shm_snapshot_t",349,"name")="shm_snapshot_t.shadow_file_header.reserved_bgtrcrec3_cntr"
	Set gtmtypes("shm_snapshot_t",349,"off")=7472
	Set gtmtypes("shm_snapshot_t",349,"len")=4
	Set gtmtypes("shm_snapshot_t",349,"type")="int"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.reserved_bgtrcrec3_cntr")=349
	Set gtmtypes("shm_snapshot_t",350,"name")="shm_snapshot_t.shadow_file_header.lost_block_recovery_cntr"
	Set gtmtypes("shm_snapshot_t",350,"off")=7476
	Set gtmtypes("shm_snapshot_t",350,"len")=4
	Set gtmtypes("shm_snapshot_t",350,"type")="int"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.lost_block_recovery_cntr")=350
	Set gtmtypes("shm_snapshot_t",351,"name")="shm_snapshot_t.shadow_file_header.wc_blocked_onln_rlbk_cntr"
	Set gtmtypes("shm_snapshot_t",351,"off")=7480
	Set gtmtypes("shm_snapshot_t",351,"len")=4
	Set gtmtypes("shm_snapshot_t",351,"type")="int"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.wc_blocked_onln_rlbk_cntr")=351
	Set gtmtypes("shm_snapshot_t",352,"name")="shm_snapshot_t.shadow_file_header.stale_cntr"
	Set gtmtypes("shm_snapshot_t",352,"off")=7484
	Set gtmtypes("shm_snapshot_t",352,"len")=4
	Set gtmtypes("shm_snapshot_t",352,"type")="int"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.stale_cntr")=352
	Set gtmtypes("shm_snapshot_t",353,"name")="shm_snapshot_t.shadow_file_header.starved_cntr"
	Set gtmtypes("shm_snapshot_t",353,"off")=7488
	Set gtmtypes("shm_snapshot_t",353,"len")=4
	Set gtmtypes("shm_snapshot_t",353,"type")="int"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.starved_cntr")=353
	Set gtmtypes("shm_snapshot_t",354,"name")="shm_snapshot_t.shadow_file_header.stale_timer_started_cntr"
	Set gtmtypes("shm_snapshot_t",354,"off")=7492
	Set gtmtypes("shm_snapshot_t",354,"len")=4
	Set gtmtypes("shm_snapshot_t",354,"type")="int"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.stale_timer_started_cntr")=354
	Set gtmtypes("shm_snapshot_t",355,"name")="shm_snapshot_t.shadow_file_header.stale_timer_pop_cntr"
	Set gtmtypes("shm_snapshot_t",355,"off")=7496
	Set gtmtypes("shm_snapshot_t",355,"len")=4
	Set gtmtypes("shm_snapshot_t",355,"type")="int"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.stale_timer_pop_cntr")=355
	Set gtmtypes("shm_snapshot_t",356,"name")="shm_snapshot_t.shadow_file_header.stale_process_defer_cntr"
	Set gtmtypes("shm_snapshot_t",356,"off")=7500
	Set gtmtypes("shm_snapshot_t",356,"len")=4
	Set gtmtypes("shm_snapshot_t",356,"type")="int"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.stale_process_defer_cntr")=356
	Set gtmtypes("shm_snapshot_t",357,"name")="shm_snapshot_t.shadow_file_header.stale_defer_processed_cntr"
	Set gtmtypes("shm_snapshot_t",357,"off")=7504
	Set gtmtypes("shm_snapshot_t",357,"len")=4
	Set gtmtypes("shm_snapshot_t",357,"type")="int"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.stale_defer_processed_cntr")=357
	Set gtmtypes("shm_snapshot_t",358,"name")="shm_snapshot_t.shadow_file_header.spcfc_buffer_flush_cntr"
	Set gtmtypes("shm_snapshot_t",358,"off")=7508
	Set gtmtypes("shm_snapshot_t",358,"len")=4
	Set gtmtypes("shm_snapshot_t",358,"type")="int"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.spcfc_buffer_flush_cntr")=358
	Set gtmtypes("shm_snapshot_t",359,"name")="shm_snapshot_t.shadow_file_header.spcfc_buffer_flush_loop_cntr"
	Set gtmtypes("shm_snapshot_t",359,"off")=7512
	Set gtmtypes("shm_snapshot_t",359,"len")=4
	Set gtmtypes("shm_snapshot_t",359,"type")="int"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.spcfc_buffer_flush_loop_cntr")=359
	Set gtmtypes("shm_snapshot_t",360,"name")="shm_snapshot_t.shadow_file_header.spcfc_buffer_flush_retries_cntr"
	Set gtmtypes("shm_snapshot_t",360,"off")=7516
	Set gtmtypes("shm_snapshot_t",360,"len")=4
	Set gtmtypes("shm_snapshot_t",360,"type")="int"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.spcfc_buffer_flush_retries_cntr")=360
	Set gtmtypes("shm_snapshot_t",361,"name")="shm_snapshot_t.shadow_file_header.spcfc_buffer_flushed_during_lockwait_cntr"
	Set gtmtypes("shm_snapshot_t",361,"off")=7520
	Set gtmtypes("shm_snapshot_t",361,"len")=4
	Set gtmtypes("shm_snapshot_t",361,"type")="int"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.spcfc_buffer_flushed_during_lockwait_cntr")=361
	Set gtmtypes("shm_snapshot_t",362,"name")="shm_snapshot_t.shadow_file_header.tp_crit_retries_cntr"
	Set gtmtypes("shm_snapshot_t",362,"off")=7524
	Set gtmtypes("shm_snapshot_t",362,"len")=4
	Set gtmtypes("shm_snapshot_t",362,"type")="int"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.tp_crit_retries_cntr")=362
	Set gtmtypes("shm_snapshot_t",363,"name")="shm_snapshot_t.shadow_file_header.db_csh_getn_flush_dirty_cntr"
	Set gtmtypes("shm_snapshot_t",363,"off")=7528
	Set gtmtypes("shm_snapshot_t",363,"len")=4
	Set gtmtypes("shm_snapshot_t",363,"type")="int"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.db_csh_getn_flush_dirty_cntr")=363
	Set gtmtypes("shm_snapshot_t",364,"name")="shm_snapshot_t.shadow_file_header.db_csh_getn_rip_wait_cntr"
	Set gtmtypes("shm_snapshot_t",364,"off")=7532
	Set gtmtypes("shm_snapshot_t",364,"len")=4
	Set gtmtypes("shm_snapshot_t",364,"type")="int"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.db_csh_getn_rip_wait_cntr")=364
	Set gtmtypes("shm_snapshot_t",365,"name")="shm_snapshot_t.shadow_file_header.db_csh_getn_buf_owner_stuck_cntr"
	Set gtmtypes("shm_snapshot_t",365,"off")=7536
	Set gtmtypes("shm_snapshot_t",365,"len")=4
	Set gtmtypes("shm_snapshot_t",365,"type")="int"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.db_csh_getn_buf_owner_stuck_cntr")=365
	Set gtmtypes("shm_snapshot_t",366,"name")="shm_snapshot_t.shadow_file_header.db_csh_getn_out_of_design_cntr"
	Set gtmtypes("shm_snapshot_t",366,"off")=7540
	Set gtmtypes("shm_snapshot_t",366,"len")=4
	Set gtmtypes("shm_snapshot_t",366,"type")="int"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.db_csh_getn_out_of_design_cntr")=366
	Set gtmtypes("shm_snapshot_t",367,"name")="shm_snapshot_t.shadow_file_header.t_qread_buf_owner_stuck_cntr"
	Set gtmtypes("shm_snapshot_t",367,"off")=7544
	Set gtmtypes("shm_snapshot_t",367,"len")=4
	Set gtmtypes("shm_snapshot_t",367,"type")="int"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.t_qread_buf_owner_stuck_cntr")=367
	Set gtmtypes("shm_snapshot_t",368,"name")="shm_snapshot_t.shadow_file_header.t_qread_out_of_design_cntr"
	Set gtmtypes("shm_snapshot_t",368,"off")=7548
	Set gtmtypes("shm_snapshot_t",368,"len")=4
	Set gtmtypes("shm_snapshot_t",368,"type")="int"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.t_qread_out_of_design_cntr")=368
	Set gtmtypes("shm_snapshot_t",369,"name")="shm_snapshot_t.shadow_file_header.unused_filler_0031_cntr"
	Set gtmtypes("shm_snapshot_t",369,"off")=7552
	Set gtmtypes("shm_snapshot_t",369,"len")=4
	Set gtmtypes("shm_snapshot_t",369,"type")="int"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.unused_filler_0031_cntr")=369
	Set gtmtypes("shm_snapshot_t",370,"name")="shm_snapshot_t.shadow_file_header.mlock_wakeups_cntr"
	Set gtmtypes("shm_snapshot_t",370,"off")=7556
	Set gtmtypes("shm_snapshot_t",370,"len")=4
	Set gtmtypes("shm_snapshot_t",370,"type")="int"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.mlock_wakeups_cntr")=370
	Set gtmtypes("shm_snapshot_t",371,"name")="shm_snapshot_t.shadow_file_header.wc_blocked_wcs_recover_invoked_cntr"
	Set gtmtypes("shm_snapshot_t",371,"off")=7560
	Set gtmtypes("shm_snapshot_t",371,"len")=4
	Set gtmtypes("shm_snapshot_t",371,"type")="int"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.wc_blocked_wcs_recover_invoked_cntr")=371
	Set gtmtypes("shm_snapshot_t",372,"name")="shm_snapshot_t.shadow_file_header.wc_blocked_wcs_verify_passed_cntr"
	Set gtmtypes("shm_snapshot_t",372,"off")=7564
	Set gtmtypes("shm_snapshot_t",372,"len")=4
	Set gtmtypes("shm_snapshot_t",372,"type")="int"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.wc_blocked_wcs_verify_passed_cntr")=372
	Set gtmtypes("shm_snapshot_t",373,"name")="shm_snapshot_t.shadow_file_header.wc_blocked_t_qread_db_csh_getn_invalid_blk_cntr"
	Set gtmtypes("shm_snapshot_t",373,"off")=7568
	Set gtmtypes("shm_snapshot_t",373,"len")=4
	Set gtmtypes("shm_snapshot_t",373,"type")="int"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.wc_blocked_t_qread_db_csh_getn_invalid_blk_cntr")=373
	Set gtmtypes("shm_snapshot_t",374,"name")="shm_snapshot_t.shadow_file_header.wc_blocked_t_qread_db_csh_get_invalid_blk_cntr"
	Set gtmtypes("shm_snapshot_t",374,"off")=7572
	Set gtmtypes("shm_snapshot_t",374,"len")=4
	Set gtmtypes("shm_snapshot_t",374,"type")="int"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.wc_blocked_t_qread_db_csh_get_invalid_blk_cntr")=374
	Set gtmtypes("shm_snapshot_t",375,"name")="shm_snapshot_t.shadow_file_header.wc_blocked_db_csh_getn_loopexceed_cntr"
	Set gtmtypes("shm_snapshot_t",375,"off")=7576
	Set gtmtypes("shm_snapshot_t",375,"len")=4
	Set gtmtypes("shm_snapshot_t",375,"type")="int"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.wc_blocked_db_csh_getn_loopexceed_cntr")=375
	Set gtmtypes("shm_snapshot_t",376,"name")="shm_snapshot_t.shadow_file_header.wc_blocked_db_csh_getn_wcsstarvewrt_cntr"
	Set gtmtypes("shm_snapshot_t",376,"off")=7580
	Set gtmtypes("shm_snapshot_t",376,"len")=4
	Set gtmtypes("shm_snapshot_t",376,"type")="int"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.wc_blocked_db_csh_getn_wcsstarvewrt_cntr")=376
	Set gtmtypes("shm_snapshot_t",377,"name")="shm_snapshot_t.shadow_file_header.wc_blocked_db_csh_get_cntr"
	Set gtmtypes("shm_snapshot_t",377,"off")=7584
	Set gtmtypes("shm_snapshot_t",377,"len")=4
	Set gtmtypes("shm_snapshot_t",377,"type")="int"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.wc_blocked_db_csh_get_cntr")=377
	Set gtmtypes("shm_snapshot_t",378,"name")="shm_snapshot_t.shadow_file_header.wc_blocked_tp_tend_wcsgetspace_cntr"
	Set gtmtypes("shm_snapshot_t",378,"off")=7588
	Set gtmtypes("shm_snapshot_t",378,"len")=4
	Set gtmtypes("shm_snapshot_t",378,"type")="int"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.wc_blocked_tp_tend_wcsgetspace_cntr")=378
	Set gtmtypes("shm_snapshot_t",379,"name")="shm_snapshot_t.shadow_file_header.wc_blocked_tp_tend_t1_cntr"
	Set gtmtypes("shm_snapshot_t",379,"off")=7592
	Set gtmtypes("shm_snapshot_t",379,"len")=4
	Set gtmtypes("shm_snapshot_t",379,"type")="int"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.wc_blocked_tp_tend_t1_cntr")=379
	Set gtmtypes("shm_snapshot_t",380,"name")="shm_snapshot_t.shadow_file_header.wc_blocked_tp_tend_bitmap_cntr"
	Set gtmtypes("shm_snapshot_t",380,"off")=7596
	Set gtmtypes("shm_snapshot_t",380,"len")=4
	Set gtmtypes("shm_snapshot_t",380,"type")="int"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.wc_blocked_tp_tend_bitmap_cntr")=380
	Set gtmtypes("shm_snapshot_t",381,"name")="shm_snapshot_t.shadow_file_header.wc_blocked_tp_tend_jnl_cwset_cntr"
	Set gtmtypes("shm_snapshot_t",381,"off")=7600
	Set gtmtypes("shm_snapshot_t",381,"len")=4
	Set gtmtypes("shm_snapshot_t",381,"type")="int"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.wc_blocked_tp_tend_jnl_cwset_cntr")=381
	Set gtmtypes("shm_snapshot_t",382,"name")="shm_snapshot_t.shadow_file_header.wc_blocked_tp_tend_jnl_wcsflu_cntr"
	Set gtmtypes("shm_snapshot_t",382,"off")=7604
	Set gtmtypes("shm_snapshot_t",382,"len")=4
	Set gtmtypes("shm_snapshot_t",382,"type")="int"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.wc_blocked_tp_tend_jnl_wcsflu_cntr")=382
	Set gtmtypes("shm_snapshot_t",383,"name")="shm_snapshot_t.shadow_file_header.wc_blocked_t_end_hist_cntr"
	Set gtmtypes("shm_snapshot_t",383,"off")=7608
	Set gtmtypes("shm_snapshot_t",383,"len")=4
	Set gtmtypes("shm_snapshot_t",383,"type")="int"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.wc_blocked_t_end_hist_cntr")=383
	Set gtmtypes("shm_snapshot_t",384,"name")="shm_snapshot_t.shadow_file_header.wc_blocked_t_end_hist1_nullbt_cntr"
	Set gtmtypes("shm_snapshot_t",384,"off")=7612
	Set gtmtypes("shm_snapshot_t",384,"len")=4
	Set gtmtypes("shm_snapshot_t",384,"type")="int"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.wc_blocked_t_end_hist1_nullbt_cntr")=384
	Set gtmtypes("shm_snapshot_t",385,"name")="shm_snapshot_t.shadow_file_header.wc_blocked_t_end_hist1_nonnullbt_cntr"
	Set gtmtypes("shm_snapshot_t",385,"off")=7616
	Set gtmtypes("shm_snapshot_t",385,"len")=4
	Set gtmtypes("shm_snapshot_t",385,"type")="int"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.wc_blocked_t_end_hist1_nonnullbt_cntr")=385
	Set gtmtypes("shm_snapshot_t",386,"name")="shm_snapshot_t.shadow_file_header.wc_blocked_t_end_bitmap_nullbt_cntr"
	Set gtmtypes("shm_snapshot_t",386,"off")=7620
	Set gtmtypes("shm_snapshot_t",386,"len")=4
	Set gtmtypes("shm_snapshot_t",386,"type")="int"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.wc_blocked_t_end_bitmap_nullbt_cntr")=386
	Set gtmtypes("shm_snapshot_t",387,"name")="shm_snapshot_t.shadow_file_header.wc_blocked_t_end_bitmap_nonnullbt_cntr"
	Set gtmtypes("shm_snapshot_t",387,"off")=7624
	Set gtmtypes("shm_snapshot_t",387,"len")=4
	Set gtmtypes("shm_snapshot_t",387,"type")="int"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.wc_blocked_t_end_bitmap_nonnullbt_cntr")=387
	Set gtmtypes("shm_snapshot_t",388,"name")="shm_snapshot_t.shadow_file_header.wc_blocked_t_end_jnl_cwset_cntr"
	Set gtmtypes("shm_snapshot_t",388,"off")=7628
	Set gtmtypes("shm_snapshot_t",388,"len")=4
	Set gtmtypes("shm_snapshot_t",388,"type")="int"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.wc_blocked_t_end_jnl_cwset_cntr")=388
	Set gtmtypes("shm_snapshot_t",389,"name")="shm_snapshot_t.shadow_file_header.wc_blocked_t_end_jnl_wcsflu_cntr"
	Set gtmtypes("shm_snapshot_t",389,"off")=7632
	Set gtmtypes("shm_snapshot_t",389,"len")=4
	Set gtmtypes("shm_snapshot_t",389,"type")="int"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.wc_blocked_t_end_jnl_wcsflu_cntr")=389
	Set gtmtypes("shm_snapshot_t",390,"name")="shm_snapshot_t.shadow_file_header.unused_filler_0052_cntr"
	Set gtmtypes("shm_snapshot_t",390,"off")=7636
	Set gtmtypes("shm_snapshot_t",390,"len")=4
	Set gtmtypes("shm_snapshot_t",390,"type")="int"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.unused_filler_0052_cntr")=390
	Set gtmtypes("shm_snapshot_t",391,"name")="shm_snapshot_t.shadow_file_header.wc_blocked_tpckh_hist1_nullbt_cntr"
	Set gtmtypes("shm_snapshot_t",391,"off")=7640
	Set gtmtypes("shm_snapshot_t",391,"len")=4
	Set gtmtypes("shm_snapshot_t",391,"type")="int"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.wc_blocked_tpckh_hist1_nullbt_cntr")=391
	Set gtmtypes("shm_snapshot_t",392,"name")="shm_snapshot_t.shadow_file_header.wc_blocked_tpckh_hist1_nonnullbt_cntr"
	Set gtmtypes("shm_snapshot_t",392,"off")=7644
	Set gtmtypes("shm_snapshot_t",392,"len")=4
	Set gtmtypes("shm_snapshot_t",392,"type")="int"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.wc_blocked_tpckh_hist1_nonnullbt_cntr")=392
	Set gtmtypes("shm_snapshot_t",393,"name")="shm_snapshot_t.shadow_file_header.jnl_blocked_writer_lost_cntr"
	Set gtmtypes("shm_snapshot_t",393,"off")=7648
	Set gtmtypes("shm_snapshot_t",393,"len")=4
	Set gtmtypes("shm_snapshot_t",393,"type")="int"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.jnl_blocked_writer_lost_cntr")=393
	Set gtmtypes("shm_snapshot_t",394,"name")="shm_snapshot_t.shadow_file_header.jnl_blocked_writer_stuck_cntr"
	Set gtmtypes("shm_snapshot_t",394,"off")=7652
	Set gtmtypes("shm_snapshot_t",394,"len")=4
	Set gtmtypes("shm_snapshot_t",394,"type")="int"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.jnl_blocked_writer_stuck_cntr")=394
	Set gtmtypes("shm_snapshot_t",395,"name")="shm_snapshot_t.shadow_file_header.jnl_blocked_writer_blocked_cntr"
	Set gtmtypes("shm_snapshot_t",395,"off")=7656
	Set gtmtypes("shm_snapshot_t",395,"len")=4
	Set gtmtypes("shm_snapshot_t",395,"type")="int"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.jnl_blocked_writer_blocked_cntr")=395
	Set gtmtypes("shm_snapshot_t",396,"name")="shm_snapshot_t.shadow_file_header.n_jnl_fsyncs_cntr"
	Set gtmtypes("shm_snapshot_t",396,"off")=7660
	Set gtmtypes("shm_snapshot_t",396,"len")=4
	Set gtmtypes("shm_snapshot_t",396,"type")="int"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.n_jnl_fsyncs_cntr")=396
	Set gtmtypes("shm_snapshot_t",397,"name")="shm_snapshot_t.shadow_file_header.n_jnl_fsync_tries_cntr"
	Set gtmtypes("shm_snapshot_t",397,"off")=7664
	Set gtmtypes("shm_snapshot_t",397,"len")=4
	Set gtmtypes("shm_snapshot_t",397,"type")="int"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.n_jnl_fsync_tries_cntr")=397
	Set gtmtypes("shm_snapshot_t",398,"name")="shm_snapshot_t.shadow_file_header.n_jnl_fsync_recovers_cntr"
	Set gtmtypes("shm_snapshot_t",398,"off")=7668
	Set gtmtypes("shm_snapshot_t",398,"len")=4
	Set gtmtypes("shm_snapshot_t",398,"type")="int"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.n_jnl_fsync_recovers_cntr")=398
	Set gtmtypes("shm_snapshot_t",399,"name")="shm_snapshot_t.shadow_file_header.n_db_fsyncs_cntr"
	Set gtmtypes("shm_snapshot_t",399,"off")=7672
	Set gtmtypes("shm_snapshot_t",399,"len")=4
	Set gtmtypes("shm_snapshot_t",399,"type")="int"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.n_db_fsyncs_cntr")=399
	Set gtmtypes("shm_snapshot_t",400,"name")="shm_snapshot_t.shadow_file_header.n_db_fsyncs_in_crit_cntr"
	Set gtmtypes("shm_snapshot_t",400,"off")=7676
	Set gtmtypes("shm_snapshot_t",400,"len")=4
	Set gtmtypes("shm_snapshot_t",400,"type")="int"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.n_db_fsyncs_in_crit_cntr")=400
	Set gtmtypes("shm_snapshot_t",401,"name")="shm_snapshot_t.shadow_file_header.n_dbsync_timers_cntr"
	Set gtmtypes("shm_snapshot_t",401,"off")=7680
	Set gtmtypes("shm_snapshot_t",401,"len")=4
	Set gtmtypes("shm_snapshot_t",401,"type")="int"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.n_dbsync_timers_cntr")=401
	Set gtmtypes("shm_snapshot_t",402,"name")="shm_snapshot_t.shadow_file_header.n_dbsync_writes_cntr"
	Set gtmtypes("shm_snapshot_t",402,"off")=7684
	Set gtmtypes("shm_snapshot_t",402,"len")=4
	Set gtmtypes("shm_snapshot_t",402,"type")="int"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.n_dbsync_writes_cntr")=402
	Set gtmtypes("shm_snapshot_t",403,"name")="shm_snapshot_t.shadow_file_header.mutex_queue_full_cntr"
	Set gtmtypes("shm_snapshot_t",403,"off")=7688
	Set gtmtypes("shm_snapshot_t",403,"len")=4
	Set gtmtypes("shm_snapshot_t",403,"type")="int"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.mutex_queue_full_cntr")=403
	Set gtmtypes("shm_snapshot_t",404,"name")="shm_snapshot_t.shadow_file_header.wcb_bt_put_cntr"
	Set gtmtypes("shm_snapshot_t",404,"off")=7692
	Set gtmtypes("shm_snapshot_t",404,"len")=4
	Set gtmtypes("shm_snapshot_t",404,"type")="int"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.wcb_bt_put_cntr")=404
	Set gtmtypes("shm_snapshot_t",405,"name")="shm_snapshot_t.shadow_file_header.wcb_mutex_salvage_cntr"
	Set gtmtypes("shm_snapshot_t",405,"off")=7696
	Set gtmtypes("shm_snapshot_t",405,"len")=4
	Set gtmtypes("shm_snapshot_t",405,"type")="int"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.wcb_mutex_salvage_cntr")=405
	Set gtmtypes("shm_snapshot_t",406,"name")="shm_snapshot_t.shadow_file_header.wcb_wcs_flu0_cntr"
	Set gtmtypes("shm_snapshot_t",406,"off")=7700
	Set gtmtypes("shm_snapshot_t",406,"len")=4
	Set gtmtypes("shm_snapshot_t",406,"type")="int"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.wcb_wcs_flu0_cntr")=406
	Set gtmtypes("shm_snapshot_t",407,"name")="shm_snapshot_t.shadow_file_header.wcb_t_end_sysops_nocr_invcr_cntr"
	Set gtmtypes("shm_snapshot_t",407,"off")=7704
	Set gtmtypes("shm_snapshot_t",407,"len")=4
	Set gtmtypes("shm_snapshot_t",407,"type")="int"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.wcb_t_end_sysops_nocr_invcr_cntr")=407
	Set gtmtypes("shm_snapshot_t",408,"name")="shm_snapshot_t.shadow_file_header.wcb_t_end_sysops_cr_invcr_cntr"
	Set gtmtypes("shm_snapshot_t",408,"off")=7708
	Set gtmtypes("shm_snapshot_t",408,"len")=4
	Set gtmtypes("shm_snapshot_t",408,"type")="int"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.wcb_t_end_sysops_cr_invcr_cntr")=408
	Set gtmtypes("shm_snapshot_t",409,"name")="shm_snapshot_t.shadow_file_header.wcb_t_end_sysops_rip_wait_cntr"
	Set gtmtypes("shm_snapshot_t",409,"off")=7712
	Set gtmtypes("shm_snapshot_t",409,"len")=4
	Set gtmtypes("shm_snapshot_t",409,"type")="int"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.wcb_t_end_sysops_rip_wait_cntr")=409
	Set gtmtypes("shm_snapshot_t",410,"name")="shm_snapshot_t.shadow_file_header.wcb_t_end_sysops_dirtyripwait_cntr"
	Set gtmtypes("shm_snapshot_t",410,"off")=7716
	Set gtmtypes("shm_snapshot_t",410,"len")=4
	Set gtmtypes("shm_snapshot_t",410,"type")="int"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.wcb_t_end_sysops_dirtyripwait_cntr")=410
	Set gtmtypes("shm_snapshot_t",411,"name")="shm_snapshot_t.shadow_file_header.wcb_gds_rundown1_cntr"
	Set gtmtypes("shm_snapshot_t",411,"off")=7720
	Set gtmtypes("shm_snapshot_t",411,"len")=4
	Set gtmtypes("shm_snapshot_t",411,"type")="int"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.wcb_gds_rundown1_cntr")=411
	Set gtmtypes("shm_snapshot_t",412,"name")="shm_snapshot_t.shadow_file_header.wcb_wcs_flu1_cntr"
	Set gtmtypes("shm_snapshot_t",412,"off")=7724
	Set gtmtypes("shm_snapshot_t",412,"len")=4
	Set gtmtypes("shm_snapshot_t",412,"type")="int"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.wcb_wcs_flu1_cntr")=412
	Set gtmtypes("shm_snapshot_t",413,"name")="shm_snapshot_t.shadow_file_header.wcb_mu_back_cntr"
	Set gtmtypes("shm_snapshot_t",413,"off")=7728
	Set gtmtypes("shm_snapshot_t",413,"len")=4
	Set gtmtypes("shm_snapshot_t",413,"type")="int"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.wcb_mu_back_cntr")=413
	Set gtmtypes("shm_snapshot_t",414,"name")="shm_snapshot_t.shadow_file_header.wcb_t_end_sysops_dirty_invcr_cntr"
	Set gtmtypes("shm_snapshot_t",414,"off")=7732
	Set gtmtypes("shm_snapshot_t",414,"len")=4
	Set gtmtypes("shm_snapshot_t",414,"type")="int"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.wcb_t_end_sysops_dirty_invcr_cntr")=414
	Set gtmtypes("shm_snapshot_t",415,"name")="shm_snapshot_t.shadow_file_header.wcb_t_end_sysops_wtfini_fail_cntr"
	Set gtmtypes("shm_snapshot_t",415,"off")=7736
	Set gtmtypes("shm_snapshot_t",415,"len")=4
	Set gtmtypes("shm_snapshot_t",415,"type")="int"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.wcb_t_end_sysops_wtfini_fail_cntr")=415
	Set gtmtypes("shm_snapshot_t",416,"name")="shm_snapshot_t.shadow_file_header.wcb_t_end_sysops_twin_stuck_cntr"
	Set gtmtypes("shm_snapshot_t",416,"off")=7740
	Set gtmtypes("shm_snapshot_t",416,"len")=4
	Set gtmtypes("shm_snapshot_t",416,"type")="int"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.wcb_t_end_sysops_twin_stuck_cntr")=416
	Set gtmtypes("shm_snapshot_t",417,"name")="shm_snapshot_t.shadow_file_header.db_csh_getn_wrt_latch_stuck_cntr"
	Set gtmtypes("shm_snapshot_t",417,"off")=7744
	Set gtmtypes("shm_snapshot_t",417,"len")=4
	Set gtmtypes("shm_snapshot_t",417,"type")="int"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.db_csh_getn_wrt_latch_stuck_cntr")=417
	Set gtmtypes("shm_snapshot_t",418,"name")="shm_snapshot_t.shadow_file_header.wcb_secshr_db_clnup_now_crit_cntr"
	Set gtmtypes("shm_snapshot_t",418,"off")=7748
	Set gtmtypes("shm_snapshot_t",418,"len")=4
	Set gtmtypes("shm_snapshot_t",418,"type")="int"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.wcb_secshr_db_clnup_now_crit_cntr")=418
	Set gtmtypes("shm_snapshot_t",419,"name")="shm_snapshot_t.shadow_file_header.wc_blocked_t_qread_bad_bt_index1_cntr"
	Set gtmtypes("shm_snapshot_t",419,"off")=7752
	Set gtmtypes("shm_snapshot_t",419,"len")=4
	Set gtmtypes("shm_snapshot_t",419,"type")="int"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.wc_blocked_t_qread_bad_bt_index1_cntr")=419
	Set gtmtypes("shm_snapshot_t",420,"name")="shm_snapshot_t.shadow_file_header.wc_blocked_t_qread_bad_bt_index2_cntr"
	Set gtmtypes("shm_snapshot_t",420,"off")=7756
	Set gtmtypes("shm_snapshot_t",420,"len")=4
	Set gtmtypes("shm_snapshot_t",420,"type")="int"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.wc_blocked_t_qread_bad_bt_index2_cntr")=420
	Set gtmtypes("shm_snapshot_t",421,"name")="shm_snapshot_t.shadow_file_header.wc_blocked_t_end_crbtmismatch1_cntr"
	Set gtmtypes("shm_snapshot_t",421,"off")=7760
	Set gtmtypes("shm_snapshot_t",421,"len")=4
	Set gtmtypes("shm_snapshot_t",421,"type")="int"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.wc_blocked_t_end_crbtmismatch1_cntr")=421
	Set gtmtypes("shm_snapshot_t",422,"name")="shm_snapshot_t.shadow_file_header.wc_blocked_t_end_crbtmismatch2_cntr"
	Set gtmtypes("shm_snapshot_t",422,"off")=7764
	Set gtmtypes("shm_snapshot_t",422,"len")=4
	Set gtmtypes("shm_snapshot_t",422,"type")="int"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.wc_blocked_t_end_crbtmismatch2_cntr")=422
	Set gtmtypes("shm_snapshot_t",423,"name")="shm_snapshot_t.shadow_file_header.wc_blocked_t_end_crbtmismatch3_cntr"
	Set gtmtypes("shm_snapshot_t",423,"off")=7768
	Set gtmtypes("shm_snapshot_t",423,"len")=4
	Set gtmtypes("shm_snapshot_t",423,"type")="int"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.wc_blocked_t_end_crbtmismatch3_cntr")=423
	Set gtmtypes("shm_snapshot_t",424,"name")="shm_snapshot_t.shadow_file_header.wc_blocked_tp_tend_crbtmismatch1_cntr"
	Set gtmtypes("shm_snapshot_t",424,"off")=7772
	Set gtmtypes("shm_snapshot_t",424,"len")=4
	Set gtmtypes("shm_snapshot_t",424,"type")="int"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.wc_blocked_tp_tend_crbtmismatch1_cntr")=424
	Set gtmtypes("shm_snapshot_t",425,"name")="shm_snapshot_t.shadow_file_header.wc_blocked_tp_tend_crbtmismatch2_cntr"
	Set gtmtypes("shm_snapshot_t",425,"off")=7776
	Set gtmtypes("shm_snapshot_t",425,"len")=4
	Set gtmtypes("shm_snapshot_t",425,"type")="int"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.wc_blocked_tp_tend_crbtmismatch2_cntr")=425
	Set gtmtypes("shm_snapshot_t",426,"name")="shm_snapshot_t.shadow_file_header.wc_blocked_tp_tend_crbtmismatch3_cntr"
	Set gtmtypes("shm_snapshot_t",426,"off")=7780
	Set gtmtypes("shm_snapshot_t",426,"len")=4
	Set gtmtypes("shm_snapshot_t",426,"type")="int"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.wc_blocked_tp_tend_crbtmismatch3_cntr")=426
	Set gtmtypes("shm_snapshot_t",427,"name")="shm_snapshot_t.shadow_file_header.wc_blocked_wcs_wtstart_bad_cr_cntr"
	Set gtmtypes("shm_snapshot_t",427,"off")=7784
	Set gtmtypes("shm_snapshot_t",427,"len")=4
	Set gtmtypes("shm_snapshot_t",427,"type")="int"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.wc_blocked_wcs_wtstart_bad_cr_cntr")=427
	Set gtmtypes("shm_snapshot_t",428,"name")="shm_snapshot_t.shadow_file_header.wc_blocked_wcs_wtfini_bad_cr_cntr"
	Set gtmtypes("shm_snapshot_t",428,"off")=7788
	Set gtmtypes("shm_snapshot_t",428,"len")=4
	Set gtmtypes("shm_snapshot_t",428,"type")="int"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.wc_blocked_wcs_wtfini_bad_cr_cntr")=428
	Set gtmtypes("shm_snapshot_t",429,"name")="shm_snapshot_t.shadow_file_header.wc_blocked_bt_get_cntr"
	Set gtmtypes("shm_snapshot_t",429,"off")=7792
	Set gtmtypes("shm_snapshot_t",429,"len")=4
	Set gtmtypes("shm_snapshot_t",429,"type")="int"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.wc_blocked_bt_get_cntr")=429
	Set gtmtypes("shm_snapshot_t",430,"name")="shm_snapshot_t.shadow_file_header.wc_blocked_wcs_cdb_sc_final_retry_cntr"
	Set gtmtypes("shm_snapshot_t",430,"off")=7796
	Set gtmtypes("shm_snapshot_t",430,"len")=4
	Set gtmtypes("shm_snapshot_t",430,"type")="int"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.wc_blocked_wcs_cdb_sc_final_retry_cntr")=430
	Set gtmtypes("shm_snapshot_t",431,"name")="shm_snapshot_t.shadow_file_header.wcb_bg_update_lckfail1_cntr"
	Set gtmtypes("shm_snapshot_t",431,"off")=7800
	Set gtmtypes("shm_snapshot_t",431,"len")=4
	Set gtmtypes("shm_snapshot_t",431,"type")="int"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.wcb_bg_update_lckfail1_cntr")=431
	Set gtmtypes("shm_snapshot_t",432,"name")="shm_snapshot_t.shadow_file_header.wcb_bg_update_lckfail2_cntr"
	Set gtmtypes("shm_snapshot_t",432,"off")=7804
	Set gtmtypes("shm_snapshot_t",432,"len")=4
	Set gtmtypes("shm_snapshot_t",432,"type")="int"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.wcb_bg_update_lckfail2_cntr")=432
	Set gtmtypes("shm_snapshot_t",433,"name")="shm_snapshot_t.shadow_file_header.wcb_wtstart_lckfail1_cntr"
	Set gtmtypes("shm_snapshot_t",433,"off")=7808
	Set gtmtypes("shm_snapshot_t",433,"len")=4
	Set gtmtypes("shm_snapshot_t",433,"type")="int"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.wcb_wtstart_lckfail1_cntr")=433
	Set gtmtypes("shm_snapshot_t",434,"name")="shm_snapshot_t.shadow_file_header.wcb_wtstart_lckfail2_cntr"
	Set gtmtypes("shm_snapshot_t",434,"off")=7812
	Set gtmtypes("shm_snapshot_t",434,"len")=4
	Set gtmtypes("shm_snapshot_t",434,"type")="int"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.wcb_wtstart_lckfail2_cntr")=434
	Set gtmtypes("shm_snapshot_t",435,"name")="shm_snapshot_t.shadow_file_header.wcb_wtstart_lckfail3_cntr"
	Set gtmtypes("shm_snapshot_t",435,"off")=7816
	Set gtmtypes("shm_snapshot_t",435,"len")=4
	Set gtmtypes("shm_snapshot_t",435,"type")="int"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.wcb_wtstart_lckfail3_cntr")=435
	Set gtmtypes("shm_snapshot_t",436,"name")="shm_snapshot_t.shadow_file_header.wcb_wtstart_lckfail4_cntr"
	Set gtmtypes("shm_snapshot_t",436,"off")=7820
	Set gtmtypes("shm_snapshot_t",436,"len")=4
	Set gtmtypes("shm_snapshot_t",436,"type")="int"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.wcb_wtstart_lckfail4_cntr")=436
	Set gtmtypes("shm_snapshot_t",437,"name")="shm_snapshot_t.shadow_file_header.wcb_wtfini_lckfail1_cntr"
	Set gtmtypes("shm_snapshot_t",437,"off")=7824
	Set gtmtypes("shm_snapshot_t",437,"len")=4
	Set gtmtypes("shm_snapshot_t",437,"type")="int"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.wcb_wtfini_lckfail1_cntr")=437
	Set gtmtypes("shm_snapshot_t",438,"name")="shm_snapshot_t.shadow_file_header.wcb_wtfini_lckfail2_cntr"
	Set gtmtypes("shm_snapshot_t",438,"off")=7828
	Set gtmtypes("shm_snapshot_t",438,"len")=4
	Set gtmtypes("shm_snapshot_t",438,"type")="int"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.wcb_wtfini_lckfail2_cntr")=438
	Set gtmtypes("shm_snapshot_t",439,"name")="shm_snapshot_t.shadow_file_header.wcb_wtfini_lckfail3_cntr"
	Set gtmtypes("shm_snapshot_t",439,"off")=7832
	Set gtmtypes("shm_snapshot_t",439,"len")=4
	Set gtmtypes("shm_snapshot_t",439,"type")="int"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.wcb_wtfini_lckfail3_cntr")=439
	Set gtmtypes("shm_snapshot_t",440,"name")="shm_snapshot_t.shadow_file_header.wcb_wtfini_lckfail4_cntr"
	Set gtmtypes("shm_snapshot_t",440,"off")=7836
	Set gtmtypes("shm_snapshot_t",440,"len")=4
	Set gtmtypes("shm_snapshot_t",440,"type")="int"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.wcb_wtfini_lckfail4_cntr")=440
	Set gtmtypes("shm_snapshot_t",441,"name")="shm_snapshot_t.shadow_file_header.wcb_t_end_sysops_dirtystuck1_cntr"
	Set gtmtypes("shm_snapshot_t",441,"off")=7840
	Set gtmtypes("shm_snapshot_t",441,"len")=4
	Set gtmtypes("shm_snapshot_t",441,"type")="int"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.wcb_t_end_sysops_dirtystuck1_cntr")=441
	Set gtmtypes("shm_snapshot_t",442,"name")="shm_snapshot_t.shadow_file_header.wcb_t_end_sysops_dirtystuck2_cntr"
	Set gtmtypes("shm_snapshot_t",442,"off")=7844
	Set gtmtypes("shm_snapshot_t",442,"len")=4
	Set gtmtypes("shm_snapshot_t",442,"type")="int"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.wcb_t_end_sysops_dirtystuck2_cntr")=442
	Set gtmtypes("shm_snapshot_t",443,"name")="shm_snapshot_t.shadow_file_header.wcb_secshr_db_clnup_wbuf_dqd_cntr"
	Set gtmtypes("shm_snapshot_t",443,"off")=7848
	Set gtmtypes("shm_snapshot_t",443,"len")=4
	Set gtmtypes("shm_snapshot_t",443,"type")="int"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.wcb_secshr_db_clnup_wbuf_dqd_cntr")=443
	Set gtmtypes("shm_snapshot_t",444,"name")="shm_snapshot_t.shadow_file_header.dwngrd_refmts_syncio_cntr"
	Set gtmtypes("shm_snapshot_t",444,"off")=7852
	Set gtmtypes("shm_snapshot_t",444,"len")=4
	Set gtmtypes("shm_snapshot_t",444,"type")="int"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.dwngrd_refmts_syncio_cntr")=444
	Set gtmtypes("shm_snapshot_t",445,"name")="shm_snapshot_t.shadow_file_header.dwngrd_refmts_asyncio_cntr"
	Set gtmtypes("shm_snapshot_t",445,"off")=7856
	Set gtmtypes("shm_snapshot_t",445,"len")=4
	Set gtmtypes("shm_snapshot_t",445,"type")="int"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.dwngrd_refmts_asyncio_cntr")=445
	Set gtmtypes("shm_snapshot_t",446,"name")="shm_snapshot_t.shadow_file_header.shmpool_refmt_harvests_cntr"
	Set gtmtypes("shm_snapshot_t",446,"off")=7860
	Set gtmtypes("shm_snapshot_t",446,"len")=4
	Set gtmtypes("shm_snapshot_t",446,"type")="int"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.shmpool_refmt_harvests_cntr")=446
	Set gtmtypes("shm_snapshot_t",447,"name")="shm_snapshot_t.shadow_file_header.shmpool_recovery_cntr"
	Set gtmtypes("shm_snapshot_t",447,"off")=7864
	Set gtmtypes("shm_snapshot_t",447,"len")=4
	Set gtmtypes("shm_snapshot_t",447,"type")="int"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.shmpool_recovery_cntr")=447
	Set gtmtypes("shm_snapshot_t",448,"name")="shm_snapshot_t.shadow_file_header.shmpool_blkd_by_sdc_cntr"
	Set gtmtypes("shm_snapshot_t",448,"off")=7868
	Set gtmtypes("shm_snapshot_t",448,"len")=4
	Set gtmtypes("shm_snapshot_t",448,"type")="int"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.shmpool_blkd_by_sdc_cntr")=448
	Set gtmtypes("shm_snapshot_t",449,"name")="shm_snapshot_t.shadow_file_header.shmpool_alloc_bbflush_cntr"
	Set gtmtypes("shm_snapshot_t",449,"off")=7872
	Set gtmtypes("shm_snapshot_t",449,"len")=4
	Set gtmtypes("shm_snapshot_t",449,"type")="int"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.shmpool_alloc_bbflush_cntr")=449
	Set gtmtypes("shm_snapshot_t",450,"name")="shm_snapshot_t.shadow_file_header.refmt_hvst_blk_released_replaced_cntr"
	Set gtmtypes("shm_snapshot_t",450,"off")=7876
	Set gtmtypes("shm_snapshot_t",450,"len")=4
	Set gtmtypes("shm_snapshot_t",450,"type")="int"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.refmt_hvst_blk_released_replaced_cntr")=450
	Set gtmtypes("shm_snapshot_t",451,"name")="shm_snapshot_t.shadow_file_header.refmt_hvst_blk_released_io_complete_cntr"
	Set gtmtypes("shm_snapshot_t",451,"off")=7880
	Set gtmtypes("shm_snapshot_t",451,"len")=4
	Set gtmtypes("shm_snapshot_t",451,"type")="int"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.refmt_hvst_blk_released_io_complete_cntr")=451
	Set gtmtypes("shm_snapshot_t",452,"name")="shm_snapshot_t.shadow_file_header.refmt_hvst_blk_kept_cntr"
	Set gtmtypes("shm_snapshot_t",452,"off")=7884
	Set gtmtypes("shm_snapshot_t",452,"len")=4
	Set gtmtypes("shm_snapshot_t",452,"type")="int"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.refmt_hvst_blk_kept_cntr")=452
	Set gtmtypes("shm_snapshot_t",453,"name")="shm_snapshot_t.shadow_file_header.refmt_hvst_blk_ignored_cntr"
	Set gtmtypes("shm_snapshot_t",453,"off")=7888
	Set gtmtypes("shm_snapshot_t",453,"len")=4
	Set gtmtypes("shm_snapshot_t",453,"type")="int"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.refmt_hvst_blk_ignored_cntr")=453
	Set gtmtypes("shm_snapshot_t",454,"name")="shm_snapshot_t.shadow_file_header.refmt_blk_chk_blk_freed_cntr"
	Set gtmtypes("shm_snapshot_t",454,"off")=7892
	Set gtmtypes("shm_snapshot_t",454,"len")=4
	Set gtmtypes("shm_snapshot_t",454,"type")="int"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.refmt_blk_chk_blk_freed_cntr")=454
	Set gtmtypes("shm_snapshot_t",455,"name")="shm_snapshot_t.shadow_file_header.refmt_blk_chk_blk_kept_cntr"
	Set gtmtypes("shm_snapshot_t",455,"off")=7896
	Set gtmtypes("shm_snapshot_t",455,"len")=4
	Set gtmtypes("shm_snapshot_t",455,"type")="int"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.refmt_blk_chk_blk_kept_cntr")=455
	Set gtmtypes("shm_snapshot_t",456,"name")="shm_snapshot_t.shadow_file_header.active_lvl_trigger_cntr"
	Set gtmtypes("shm_snapshot_t",456,"off")=7900
	Set gtmtypes("shm_snapshot_t",456,"len")=4
	Set gtmtypes("shm_snapshot_t",456,"type")="int"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.active_lvl_trigger_cntr")=456
	Set gtmtypes("shm_snapshot_t",457,"name")="shm_snapshot_t.shadow_file_header.new_buff_cntr"
	Set gtmtypes("shm_snapshot_t",457,"off")=7904
	Set gtmtypes("shm_snapshot_t",457,"len")=4
	Set gtmtypes("shm_snapshot_t",457,"type")="int"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.new_buff_cntr")=457
	Set gtmtypes("shm_snapshot_t",458,"name")="shm_snapshot_t.shadow_file_header.phase2_commit_wait_sleep_in_crit_cntr"
	Set gtmtypes("shm_snapshot_t",458,"off")=7908
	Set gtmtypes("shm_snapshot_t",458,"len")=4
	Set gtmtypes("shm_snapshot_t",458,"type")="int"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.phase2_commit_wait_sleep_in_crit_cntr")=458
	Set gtmtypes("shm_snapshot_t",459,"name")="shm_snapshot_t.shadow_file_header.phase2_commit_wait_sleep_no_crit_cntr"
	Set gtmtypes("shm_snapshot_t",459,"off")=7912
	Set gtmtypes("shm_snapshot_t",459,"len")=4
	Set gtmtypes("shm_snapshot_t",459,"type")="int"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.phase2_commit_wait_sleep_no_crit_cntr")=459
	Set gtmtypes("shm_snapshot_t",460,"name")="shm_snapshot_t.shadow_file_header.phase2_commit_wait_pidcnt_cntr"
	Set gtmtypes("shm_snapshot_t",460,"off")=7916
	Set gtmtypes("shm_snapshot_t",460,"len")=4
	Set gtmtypes("shm_snapshot_t",460,"type")="int"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.phase2_commit_wait_pidcnt_cntr")=460
	Set gtmtypes("shm_snapshot_t",461,"name")="shm_snapshot_t.shadow_file_header.wcb_t_end_sysops_intend_wait_cntr"
	Set gtmtypes("shm_snapshot_t",461,"off")=7920
	Set gtmtypes("shm_snapshot_t",461,"len")=4
	Set gtmtypes("shm_snapshot_t",461,"type")="int"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.wcb_t_end_sysops_intend_wait_cntr")=461
	Set gtmtypes("shm_snapshot_t",462,"name")="shm_snapshot_t.shadow_file_header.wcb_secshr_db_clnup_phase2_clnup_cntr"
	Set gtmtypes("shm_snapshot_t",462,"off")=7924
	Set gtmtypes("shm_snapshot_t",462,"len")=4
	Set gtmtypes("shm_snapshot_t",462,"type")="int"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.wcb_secshr_db_clnup_phase2_clnup_cntr")=462
	Set gtmtypes("shm_snapshot_t",463,"name")="shm_snapshot_t.shadow_file_header.wcb_phase2_commit_wait_cntr"
	Set gtmtypes("shm_snapshot_t",463,"off")=7928
	Set gtmtypes("shm_snapshot_t",463,"len")=4
	Set gtmtypes("shm_snapshot_t",463,"type")="int"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.wcb_phase2_commit_wait_cntr")=463
	Set gtmtypes("shm_snapshot_t",464,"name")="shm_snapshot_t.shadow_file_header.recompute_upd_array_calls_cntr"
	Set gtmtypes("shm_snapshot_t",464,"off")=7932
	Set gtmtypes("shm_snapshot_t",464,"len")=4
	Set gtmtypes("shm_snapshot_t",464,"type")="int"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.recompute_upd_array_calls_cntr")=464
	Set gtmtypes("shm_snapshot_t",465,"name")="shm_snapshot_t.shadow_file_header.recompute_upd_array_rip_cntr"
	Set gtmtypes("shm_snapshot_t",465,"off")=7936
	Set gtmtypes("shm_snapshot_t",465,"len")=4
	Set gtmtypes("shm_snapshot_t",465,"type")="int"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.recompute_upd_array_rip_cntr")=465
	Set gtmtypes("shm_snapshot_t",466,"name")="shm_snapshot_t.shadow_file_header.recompute_upd_array_in_tend_cntr"
	Set gtmtypes("shm_snapshot_t",466,"off")=7940
	Set gtmtypes("shm_snapshot_t",466,"len")=4
	Set gtmtypes("shm_snapshot_t",466,"type")="int"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.recompute_upd_array_in_tend_cntr")=466
	Set gtmtypes("shm_snapshot_t",467,"name")="shm_snapshot_t.shadow_file_header.recompute_upd_array_search_blk_cntr"
	Set gtmtypes("shm_snapshot_t",467,"off")=7944
	Set gtmtypes("shm_snapshot_t",467,"len")=4
	Set gtmtypes("shm_snapshot_t",467,"type")="int"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.recompute_upd_array_search_blk_cntr")=467
	Set gtmtypes("shm_snapshot_t",468,"name")="shm_snapshot_t.shadow_file_header.recompute_upd_array_new_rec_cntr"
	Set gtmtypes("shm_snapshot_t",468,"off")=7948
	Set gtmtypes("shm_snapshot_t",468,"len")=4
	Set gtmtypes("shm_snapshot_t",468,"type")="int"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.recompute_upd_array_new_rec_cntr")=468
	Set gtmtypes("shm_snapshot_t",469,"name")="shm_snapshot_t.shadow_file_header.recompute_upd_array_rec_size_cntr"
	Set gtmtypes("shm_snapshot_t",469,"off")=7952
	Set gtmtypes("shm_snapshot_t",469,"len")=4
	Set gtmtypes("shm_snapshot_t",469,"type")="int"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.recompute_upd_array_rec_size_cntr")=469
	Set gtmtypes("shm_snapshot_t",470,"name")="shm_snapshot_t.shadow_file_header.recompute_upd_array_rec_cmpc_cntr"
	Set gtmtypes("shm_snapshot_t",470,"off")=7956
	Set gtmtypes("shm_snapshot_t",470,"len")=4
	Set gtmtypes("shm_snapshot_t",470,"type")="int"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.recompute_upd_array_rec_cmpc_cntr")=470
	Set gtmtypes("shm_snapshot_t",471,"name")="shm_snapshot_t.shadow_file_header.recompute_upd_array_blk_fini_cntr"
	Set gtmtypes("shm_snapshot_t",471,"off")=7960
	Set gtmtypes("shm_snapshot_t",471,"len")=4
	Set gtmtypes("shm_snapshot_t",471,"type")="int"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.recompute_upd_array_blk_fini_cntr")=471
	Set gtmtypes("shm_snapshot_t",472,"name")="shm_snapshot_t.shadow_file_header.recompute_upd_array_blk_split_cntr"
	Set gtmtypes("shm_snapshot_t",472,"off")=7964
	Set gtmtypes("shm_snapshot_t",472,"len")=4
	Set gtmtypes("shm_snapshot_t",472,"type")="int"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.recompute_upd_array_blk_split_cntr")=472
	Set gtmtypes("shm_snapshot_t",473,"name")="shm_snapshot_t.shadow_file_header.unused_filler_0135_cntr"
	Set gtmtypes("shm_snapshot_t",473,"off")=7968
	Set gtmtypes("shm_snapshot_t",473,"len")=4
	Set gtmtypes("shm_snapshot_t",473,"type")="int"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.unused_filler_0135_cntr")=473
	Set gtmtypes("shm_snapshot_t",474,"name")="shm_snapshot_t.shadow_file_header.t_qread_ripsleep_nblks_cntr"
	Set gtmtypes("shm_snapshot_t",474,"off")=7972
	Set gtmtypes("shm_snapshot_t",474,"len")=4
	Set gtmtypes("shm_snapshot_t",474,"type")="int"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.t_qread_ripsleep_nblks_cntr")=474
	Set gtmtypes("shm_snapshot_t",475,"name")="shm_snapshot_t.shadow_file_header.wcs_wtfini_invoked_cntr"
	Set gtmtypes("shm_snapshot_t",475,"off")=7976
	Set gtmtypes("shm_snapshot_t",475,"len")=4
	Set gtmtypes("shm_snapshot_t",475,"type")="int"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.wcs_wtfini_invoked_cntr")=475
	Set gtmtypes("shm_snapshot_t",476,"name")="shm_snapshot_t.shadow_file_header.wcs_wterror_invoked_cntr"
	Set gtmtypes("shm_snapshot_t",476,"off")=7980
	Set gtmtypes("shm_snapshot_t",476,"len")=4
	Set gtmtypes("shm_snapshot_t",476,"type")="int"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.wcs_wterror_invoked_cntr")=476
	Set gtmtypes("shm_snapshot_t",477,"name")="shm_snapshot_t.shadow_file_header.wcs_wt_restart_invoked_cntr"
	Set gtmtypes("shm_snapshot_t",477,"off")=7984
	Set gtmtypes("shm_snapshot_t",477,"len")=4
	Set gtmtypes("shm_snapshot_t",477,"type")="int"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.wcs_wt_restart_invoked_cntr")=477
	Set gtmtypes("shm_snapshot_t",478,"name")="shm_snapshot_t.shadow_file_header.wcs_wt_restart_eagain_cntr"
	Set gtmtypes("shm_snapshot_t",478,"off")=7988
	Set gtmtypes("shm_snapshot_t",478,"len")=4
	Set gtmtypes("shm_snapshot_t",478,"type")="int"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.wcs_wt_restart_eagain_cntr")=478
	Set gtmtypes("shm_snapshot_t",479,"name")="shm_snapshot_t.shadow_file_header.wcs_wt_restart_reinsert_cntr"
	Set gtmtypes("shm_snapshot_t",479,"off")=7992
	Set gtmtypes("shm_snapshot_t",479,"len")=4
	Set gtmtypes("shm_snapshot_t",479,"type")="int"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.wcs_wt_restart_reinsert_cntr")=479
	Set gtmtypes("shm_snapshot_t",480,"name")="shm_snapshot_t.shadow_file_header.wcs_wtstart_eagain_cntr"
	Set gtmtypes("shm_snapshot_t",480,"off")=7996
	Set gtmtypes("shm_snapshot_t",480,"len")=4
	Set gtmtypes("shm_snapshot_t",480,"type")="int"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.wcs_wtstart_eagain_cntr")=480
	Set gtmtypes("shm_snapshot_t",481,"name")="shm_snapshot_t.shadow_file_header.wcs_wtstart_eagain_incrit_cntr"
	Set gtmtypes("shm_snapshot_t",481,"off")=8000
	Set gtmtypes("shm_snapshot_t",481,"len")=4
	Set gtmtypes("shm_snapshot_t",481,"type")="int"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.wcs_wtstart_eagain_incrit_cntr")=481
	Set gtmtypes("shm_snapshot_t",482,"name")="shm_snapshot_t.shadow_file_header.async_restart_eagain_cntr"
	Set gtmtypes("shm_snapshot_t",482,"off")=8004
	Set gtmtypes("shm_snapshot_t",482,"len")=4
	Set gtmtypes("shm_snapshot_t",482,"type")="int"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.async_restart_eagain_cntr")=482
	Set gtmtypes("shm_snapshot_t",483,"name")="shm_snapshot_t.shadow_file_header.jnlbuff_phs2cmt_array_full_cntr"
	Set gtmtypes("shm_snapshot_t",483,"off")=8008
	Set gtmtypes("shm_snapshot_t",483,"len")=4
	Set gtmtypes("shm_snapshot_t",483,"type")="int"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.jnlbuff_phs2cmt_array_full_cntr")=483
	Set gtmtypes("shm_snapshot_t",484,"name")="shm_snapshot_t.shadow_file_header.jnl_phase2_cleanup_if_possible_cntr"
	Set gtmtypes("shm_snapshot_t",484,"off")=8012
	Set gtmtypes("shm_snapshot_t",484,"len")=4
	Set gtmtypes("shm_snapshot_t",484,"type")="int"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.jnl_phase2_cleanup_if_possible_cntr")=484
	Set gtmtypes("shm_snapshot_t",485,"name")="shm_snapshot_t.shadow_file_header.jnlbuff_phs2cmt_isprcalv_cntr"
	Set gtmtypes("shm_snapshot_t",485,"off")=8016
	Set gtmtypes("shm_snapshot_t",485,"len")=4
	Set gtmtypes("shm_snapshot_t",485,"type")="int"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.jnlbuff_phs2cmt_isprcalv_cntr")=485
	Set gtmtypes("shm_snapshot_t",486,"name")="shm_snapshot_t.shadow_file_header.recompute_upd_array_paidoff_cntr"
	Set gtmtypes("shm_snapshot_t",486,"off")=8020
	Set gtmtypes("shm_snapshot_t",486,"len")=4
	Set gtmtypes("shm_snapshot_t",486,"type")="int"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.recompute_upd_array_paidoff_cntr")=486
	Set gtmtypes("shm_snapshot_t",487,"name")="shm_snapshot_t.shadow_file_header.bg_trc_rec_cntr_filler"
	Set gtmtypes("shm_snapshot_t",487,"off")=8024
	Set gtmtypes("shm_snapshot_t",487,"len")=8
	Set gtmtypes("shm_snapshot_t",487,"type")="char"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.bg_trc_rec_cntr_filler")=487
	Set gtmtypes("shm_snapshot_t",488,"name")="shm_snapshot_t.shadow_file_header.n_gvcst_srches"
	Set gtmtypes("shm_snapshot_t",488,"off")=8032
	Set gtmtypes("shm_snapshot_t",488,"len")=8
	Set gtmtypes("shm_snapshot_t",488,"type")="db_csh_acct_rec"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.n_gvcst_srches")=488
	Set gtmtypes("shm_snapshot_t",489,"name")="shm_snapshot_t.shadow_file_header.n_gvcst_srches.curr_count"
	Set gtmtypes("shm_snapshot_t",489,"off")=8032
	Set gtmtypes("shm_snapshot_t",489,"len")=4
	Set gtmtypes("shm_snapshot_t",489,"type")="int"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.n_gvcst_srches.curr_count")=489
	Set gtmtypes("shm_snapshot_t",490,"name")="shm_snapshot_t.shadow_file_header.n_gvcst_srches.cumul_count"
	Set gtmtypes("shm_snapshot_t",490,"off")=8036
	Set gtmtypes("shm_snapshot_t",490,"len")=4
	Set gtmtypes("shm_snapshot_t",490,"type")="int"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.n_gvcst_srches.cumul_count")=490
	Set gtmtypes("shm_snapshot_t",491,"name")="shm_snapshot_t.shadow_file_header.n_gvcst_srch_clues"
	Set gtmtypes("shm_snapshot_t",491,"off")=8040
	Set gtmtypes("shm_snapshot_t",491,"len")=8
	Set gtmtypes("shm_snapshot_t",491,"type")="db_csh_acct_rec"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.n_gvcst_srch_clues")=491
	Set gtmtypes("shm_snapshot_t",492,"name")="shm_snapshot_t.shadow_file_header.n_gvcst_srch_clues.curr_count"
	Set gtmtypes("shm_snapshot_t",492,"off")=8040
	Set gtmtypes("shm_snapshot_t",492,"len")=4
	Set gtmtypes("shm_snapshot_t",492,"type")="int"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.n_gvcst_srch_clues.curr_count")=492
	Set gtmtypes("shm_snapshot_t",493,"name")="shm_snapshot_t.shadow_file_header.n_gvcst_srch_clues.cumul_count"
	Set gtmtypes("shm_snapshot_t",493,"off")=8044
	Set gtmtypes("shm_snapshot_t",493,"len")=4
	Set gtmtypes("shm_snapshot_t",493,"type")="int"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.n_gvcst_srch_clues.cumul_count")=493
	Set gtmtypes("shm_snapshot_t",494,"name")="shm_snapshot_t.shadow_file_header.n_clue_used_head"
	Set gtmtypes("shm_snapshot_t",494,"off")=8048
	Set gtmtypes("shm_snapshot_t",494,"len")=8
	Set gtmtypes("shm_snapshot_t",494,"type")="db_csh_acct_rec"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.n_clue_used_head")=494
	Set gtmtypes("shm_snapshot_t",495,"name")="shm_snapshot_t.shadow_file_header.n_clue_used_head.curr_count"
	Set gtmtypes("shm_snapshot_t",495,"off")=8048
	Set gtmtypes("shm_snapshot_t",495,"len")=4
	Set gtmtypes("shm_snapshot_t",495,"type")="int"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.n_clue_used_head.curr_count")=495
	Set gtmtypes("shm_snapshot_t",496,"name")="shm_snapshot_t.shadow_file_header.n_clue_used_head.cumul_count"
	Set gtmtypes("shm_snapshot_t",496,"off")=8052
	Set gtmtypes("shm_snapshot_t",496,"len")=4
	Set gtmtypes("shm_snapshot_t",496,"type")="int"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.n_clue_used_head.cumul_count")=496
	Set gtmtypes("shm_snapshot_t",497,"name")="shm_snapshot_t.shadow_file_header.n_clue_used_same"
	Set gtmtypes("shm_snapshot_t",497,"off")=8056
	Set gtmtypes("shm_snapshot_t",497,"len")=8
	Set gtmtypes("shm_snapshot_t",497,"type")="db_csh_acct_rec"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.n_clue_used_same")=497
	Set gtmtypes("shm_snapshot_t",498,"name")="shm_snapshot_t.shadow_file_header.n_clue_used_same.curr_count"
	Set gtmtypes("shm_snapshot_t",498,"off")=8056
	Set gtmtypes("shm_snapshot_t",498,"len")=4
	Set gtmtypes("shm_snapshot_t",498,"type")="int"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.n_clue_used_same.curr_count")=498
	Set gtmtypes("shm_snapshot_t",499,"name")="shm_snapshot_t.shadow_file_header.n_clue_used_same.cumul_count"
	Set gtmtypes("shm_snapshot_t",499,"off")=8060
	Set gtmtypes("shm_snapshot_t",499,"len")=4
	Set gtmtypes("shm_snapshot_t",499,"type")="int"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.n_clue_used_same.cumul_count")=499
	Set gtmtypes("shm_snapshot_t",500,"name")="shm_snapshot_t.shadow_file_header.n_clue_used_tail"
	Set gtmtypes("shm_snapshot_t",500,"off")=8064
	Set gtmtypes("shm_snapshot_t",500,"len")=8
	Set gtmtypes("shm_snapshot_t",500,"type")="db_csh_acct_rec"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.n_clue_used_tail")=500
	Set gtmtypes("shm_snapshot_t",501,"name")="shm_snapshot_t.shadow_file_header.n_clue_used_tail.curr_count"
	Set gtmtypes("shm_snapshot_t",501,"off")=8064
	Set gtmtypes("shm_snapshot_t",501,"len")=4
	Set gtmtypes("shm_snapshot_t",501,"type")="int"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.n_clue_used_tail.curr_count")=501
	Set gtmtypes("shm_snapshot_t",502,"name")="shm_snapshot_t.shadow_file_header.n_clue_used_tail.cumul_count"
	Set gtmtypes("shm_snapshot_t",502,"off")=8068
	Set gtmtypes("shm_snapshot_t",502,"len")=4
	Set gtmtypes("shm_snapshot_t",502,"type")="int"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.n_clue_used_tail.cumul_count")=502
	Set gtmtypes("shm_snapshot_t",503,"name")="shm_snapshot_t.shadow_file_header.n_t_qreads"
	Set gtmtypes("shm_snapshot_t",503,"off")=8072
	Set gtmtypes("shm_snapshot_t",503,"len")=8
	Set gtmtypes("shm_snapshot_t",503,"type")="db_csh_acct_rec"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.n_t_qreads")=503
	Set gtmtypes("shm_snapshot_t",504,"name")="shm_snapshot_t.shadow_file_header.n_t_qreads.curr_count"
	Set gtmtypes("shm_snapshot_t",504,"off")=8072
	Set gtmtypes("shm_snapshot_t",504,"len")=4
	Set gtmtypes("shm_snapshot_t",504,"type")="int"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.n_t_qreads.curr_count")=504
	Set gtmtypes("shm_snapshot_t",505,"name")="shm_snapshot_t.shadow_file_header.n_t_qreads.cumul_count"
	Set gtmtypes("shm_snapshot_t",505,"off")=8076
	Set gtmtypes("shm_snapshot_t",505,"len")=4
	Set gtmtypes("shm_snapshot_t",505,"type")="int"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.n_t_qreads.cumul_count")=505
	Set gtmtypes("shm_snapshot_t",506,"name")="shm_snapshot_t.shadow_file_header.unused_dsk_reads"
	Set gtmtypes("shm_snapshot_t",506,"off")=8080
	Set gtmtypes("shm_snapshot_t",506,"len")=8
	Set gtmtypes("shm_snapshot_t",506,"type")="db_csh_acct_rec"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.unused_dsk_reads")=506
	Set gtmtypes("shm_snapshot_t",507,"name")="shm_snapshot_t.shadow_file_header.unused_dsk_reads.curr_count"
	Set gtmtypes("shm_snapshot_t",507,"off")=8080
	Set gtmtypes("shm_snapshot_t",507,"len")=4
	Set gtmtypes("shm_snapshot_t",507,"type")="int"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.unused_dsk_reads.curr_count")=507
	Set gtmtypes("shm_snapshot_t",508,"name")="shm_snapshot_t.shadow_file_header.unused_dsk_reads.cumul_count"
	Set gtmtypes("shm_snapshot_t",508,"off")=8084
	Set gtmtypes("shm_snapshot_t",508,"len")=4
	Set gtmtypes("shm_snapshot_t",508,"type")="int"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.unused_dsk_reads.cumul_count")=508
	Set gtmtypes("shm_snapshot_t",509,"name")="shm_snapshot_t.shadow_file_header.n_bgmm_updates"
	Set gtmtypes("shm_snapshot_t",509,"off")=8088
	Set gtmtypes("shm_snapshot_t",509,"len")=8
	Set gtmtypes("shm_snapshot_t",509,"type")="db_csh_acct_rec"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.n_bgmm_updates")=509
	Set gtmtypes("shm_snapshot_t",510,"name")="shm_snapshot_t.shadow_file_header.n_bgmm_updates.curr_count"
	Set gtmtypes("shm_snapshot_t",510,"off")=8088
	Set gtmtypes("shm_snapshot_t",510,"len")=4
	Set gtmtypes("shm_snapshot_t",510,"type")="int"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.n_bgmm_updates.curr_count")=510
	Set gtmtypes("shm_snapshot_t",511,"name")="shm_snapshot_t.shadow_file_header.n_bgmm_updates.cumul_count"
	Set gtmtypes("shm_snapshot_t",511,"off")=8092
	Set gtmtypes("shm_snapshot_t",511,"len")=4
	Set gtmtypes("shm_snapshot_t",511,"type")="int"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.n_bgmm_updates.cumul_count")=511
	Set gtmtypes("shm_snapshot_t",512,"name")="shm_snapshot_t.shadow_file_header.unused_dsk_writes"
	Set gtmtypes("shm_snapshot_t",512,"off")=8096
	Set gtmtypes("shm_snapshot_t",512,"len")=8
	Set gtmtypes("shm_snapshot_t",512,"type")="db_csh_acct_rec"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.unused_dsk_writes")=512
	Set gtmtypes("shm_snapshot_t",513,"name")="shm_snapshot_t.shadow_file_header.unused_dsk_writes.curr_count"
	Set gtmtypes("shm_snapshot_t",513,"off")=8096
	Set gtmtypes("shm_snapshot_t",513,"len")=4
	Set gtmtypes("shm_snapshot_t",513,"type")="int"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.unused_dsk_writes.curr_count")=513
	Set gtmtypes("shm_snapshot_t",514,"name")="shm_snapshot_t.shadow_file_header.unused_dsk_writes.cumul_count"
	Set gtmtypes("shm_snapshot_t",514,"off")=8100
	Set gtmtypes("shm_snapshot_t",514,"len")=4
	Set gtmtypes("shm_snapshot_t",514,"type")="int"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.unused_dsk_writes.cumul_count")=514
	Set gtmtypes("shm_snapshot_t",515,"name")="shm_snapshot_t.shadow_file_header.n_bg_update_creates"
	Set gtmtypes("shm_snapshot_t",515,"off")=8104
	Set gtmtypes("shm_snapshot_t",515,"len")=8
	Set gtmtypes("shm_snapshot_t",515,"type")="db_csh_acct_rec"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.n_bg_update_creates")=515
	Set gtmtypes("shm_snapshot_t",516,"name")="shm_snapshot_t.shadow_file_header.n_bg_update_creates.curr_count"
	Set gtmtypes("shm_snapshot_t",516,"off")=8104
	Set gtmtypes("shm_snapshot_t",516,"len")=4
	Set gtmtypes("shm_snapshot_t",516,"type")="int"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.n_bg_update_creates.curr_count")=516
	Set gtmtypes("shm_snapshot_t",517,"name")="shm_snapshot_t.shadow_file_header.n_bg_update_creates.cumul_count"
	Set gtmtypes("shm_snapshot_t",517,"off")=8108
	Set gtmtypes("shm_snapshot_t",517,"len")=4
	Set gtmtypes("shm_snapshot_t",517,"type")="int"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.n_bg_update_creates.cumul_count")=517
	Set gtmtypes("shm_snapshot_t",518,"name")="shm_snapshot_t.shadow_file_header.n_db_csh_getns"
	Set gtmtypes("shm_snapshot_t",518,"off")=8112
	Set gtmtypes("shm_snapshot_t",518,"len")=8
	Set gtmtypes("shm_snapshot_t",518,"type")="db_csh_acct_rec"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.n_db_csh_getns")=518
	Set gtmtypes("shm_snapshot_t",519,"name")="shm_snapshot_t.shadow_file_header.n_db_csh_getns.curr_count"
	Set gtmtypes("shm_snapshot_t",519,"off")=8112
	Set gtmtypes("shm_snapshot_t",519,"len")=4
	Set gtmtypes("shm_snapshot_t",519,"type")="int"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.n_db_csh_getns.curr_count")=519
	Set gtmtypes("shm_snapshot_t",520,"name")="shm_snapshot_t.shadow_file_header.n_db_csh_getns.cumul_count"
	Set gtmtypes("shm_snapshot_t",520,"off")=8116
	Set gtmtypes("shm_snapshot_t",520,"len")=4
	Set gtmtypes("shm_snapshot_t",520,"type")="int"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.n_db_csh_getns.cumul_count")=520
	Set gtmtypes("shm_snapshot_t",521,"name")="shm_snapshot_t.shadow_file_header.n_db_csh_getn_lcnt"
	Set gtmtypes("shm_snapshot_t",521,"off")=8120
	Set gtmtypes("shm_snapshot_t",521,"len")=8
	Set gtmtypes("shm_snapshot_t",521,"type")="db_csh_acct_rec"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.n_db_csh_getn_lcnt")=521
	Set gtmtypes("shm_snapshot_t",522,"name")="shm_snapshot_t.shadow_file_header.n_db_csh_getn_lcnt.curr_count"
	Set gtmtypes("shm_snapshot_t",522,"off")=8120
	Set gtmtypes("shm_snapshot_t",522,"len")=4
	Set gtmtypes("shm_snapshot_t",522,"type")="int"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.n_db_csh_getn_lcnt.curr_count")=522
	Set gtmtypes("shm_snapshot_t",523,"name")="shm_snapshot_t.shadow_file_header.n_db_csh_getn_lcnt.cumul_count"
	Set gtmtypes("shm_snapshot_t",523,"off")=8124
	Set gtmtypes("shm_snapshot_t",523,"len")=4
	Set gtmtypes("shm_snapshot_t",523,"type")="int"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.n_db_csh_getn_lcnt.cumul_count")=523
	Set gtmtypes("shm_snapshot_t",524,"name")="shm_snapshot_t.shadow_file_header.db_csh_acct_rec_filler_4k"
	Set gtmtypes("shm_snapshot_t",524,"off")=8128
	Set gtmtypes("shm_snapshot_t",524,"len")=152
	Set gtmtypes("shm_snapshot_t",524,"type")="char"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.db_csh_acct_rec_filler_4k")=524
	Set gtmtypes("shm_snapshot_t",525,"name")="shm_snapshot_t.shadow_file_header.gvstats_rec_old_now_filler"
	Set gtmtypes("shm_snapshot_t",525,"off")=8280
	Set gtmtypes("shm_snapshot_t",525,"len")=496
	Set gtmtypes("shm_snapshot_t",525,"type")="char"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.gvstats_rec_old_now_filler")=525
	Set gtmtypes("shm_snapshot_t",526,"name")="shm_snapshot_t.shadow_file_header.gvstats_rec_filler_4k_plus_512"
	Set gtmtypes("shm_snapshot_t",526,"off")=8776
	Set gtmtypes("shm_snapshot_t",526,"len")=16
	Set gtmtypes("shm_snapshot_t",526,"type")="char"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.gvstats_rec_filler_4k_plus_512")=526
	Set gtmtypes("shm_snapshot_t",527,"name")="shm_snapshot_t.shadow_file_header.filler_4k_plus_512"
	Set gtmtypes("shm_snapshot_t",527,"off")=8792
	Set gtmtypes("shm_snapshot_t",527,"len")=368
	Set gtmtypes("shm_snapshot_t",527,"type")="char"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.filler_4k_plus_512")=527
	Set gtmtypes("shm_snapshot_t",528,"name")="shm_snapshot_t.shadow_file_header.intrpt_recov_resync_strm_seqno"
	Set gtmtypes("shm_snapshot_t",528,"off")=9160
	Set gtmtypes("shm_snapshot_t",528,"len")=128
	Set gtmtypes("shm_snapshot_t",528,"type")="uint64_t"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.intrpt_recov_resync_strm_seqno")=528
	Set gtmtypes("shm_snapshot_t",528,"dim")=16
	Set gtmtypes("shm_snapshot_t",529,"name")="shm_snapshot_t.shadow_file_header.creation_db_ver"
	Set gtmtypes("shm_snapshot_t",529,"off")=9288
	Set gtmtypes("shm_snapshot_t",529,"len")=4
	Set gtmtypes("shm_snapshot_t",529,"type")="int"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.creation_db_ver")=529
	Set gtmtypes("shm_snapshot_t",530,"name")="shm_snapshot_t.shadow_file_header.creation_mdb_ver"
	Set gtmtypes("shm_snapshot_t",530,"off")=9292
	Set gtmtypes("shm_snapshot_t",530,"len")=4
	Set gtmtypes("shm_snapshot_t",530,"type")="int"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.creation_mdb_ver")=530
	Set gtmtypes("shm_snapshot_t",531,"name")="shm_snapshot_t.shadow_file_header.certified_for_upgrade_to"
	Set gtmtypes("shm_snapshot_t",531,"off")=9296
	Set gtmtypes("shm_snapshot_t",531,"len")=4
	Set gtmtypes("shm_snapshot_t",531,"type")="int"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.certified_for_upgrade_to")=531
	Set gtmtypes("shm_snapshot_t",532,"name")="shm_snapshot_t.shadow_file_header.filler_5k"
	Set gtmtypes("shm_snapshot_t",532,"off")=9300
	Set gtmtypes("shm_snapshot_t",532,"len")=4
	Set gtmtypes("shm_snapshot_t",532,"type")="int"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.filler_5k")=532
	Set gtmtypes("shm_snapshot_t",533,"name")="shm_snapshot_t.shadow_file_header.secshr_ops_index_filler"
	Set gtmtypes("shm_snapshot_t",533,"off")=9304
	Set gtmtypes("shm_snapshot_t",533,"len")=4
	Set gtmtypes("shm_snapshot_t",533,"type")="int"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.secshr_ops_index_filler")=533
	Set gtmtypes("shm_snapshot_t",534,"name")="shm_snapshot_t.shadow_file_header.secshr_ops_array_filler"
	Set gtmtypes("shm_snapshot_t",534,"off")=9308
	Set gtmtypes("shm_snapshot_t",534,"len")=1020
	Set gtmtypes("shm_snapshot_t",534,"type")="int"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.secshr_ops_array_filler")=534
	Set gtmtypes("shm_snapshot_t",534,"dim")=255
	Set gtmtypes("shm_snapshot_t",535,"name")="shm_snapshot_t.shadow_file_header.next_upgrd_warn"
	Set gtmtypes("shm_snapshot_t",535,"off")=10328
	Set gtmtypes("shm_snapshot_t",535,"len")=24
	Set gtmtypes("shm_snapshot_t",535,"type")="compswap_time_field"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.next_upgrd_warn")=535
	Set gtmtypes("shm_snapshot_t",536,"name")="shm_snapshot_t.shadow_file_header.next_upgrd_warn.time_latch"
	Set gtmtypes("shm_snapshot_t",536,"off")=10328
	Set gtmtypes("shm_snapshot_t",536,"len")=8
	Set gtmtypes("shm_snapshot_t",536,"type")="global_latch_t"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.next_upgrd_warn.time_latch")=536
	Set gtmtypes("shm_snapshot_t",537,"name")="shm_snapshot_t.shadow_file_header.next_upgrd_warn.time_latch.u"
	Set gtmtypes("shm_snapshot_t",537,"off")=10328
	Set gtmtypes("shm_snapshot_t",537,"len")=8
	Set gtmtypes("shm_snapshot_t",537,"type")="union"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.next_upgrd_warn.time_latch.u")=537
	Set gtmtypes("shm_snapshot_t",538,"name")="shm_snapshot_t.shadow_file_header.next_upgrd_warn.time_latch.u.pid_imgcnt"
	Set gtmtypes("shm_snapshot_t",538,"off")=10328
	Set gtmtypes("shm_snapshot_t",538,"len")=8
	Set gtmtypes("shm_snapshot_t",538,"type")="uint64_t"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.next_upgrd_warn.time_latch.u.pid_imgcnt")=538
	Set gtmtypes("shm_snapshot_t",539,"name")="shm_snapshot_t.shadow_file_header.next_upgrd_warn.time_latch.u.parts"
	Set gtmtypes("shm_snapshot_t",539,"off")=10328
	Set gtmtypes("shm_snapshot_t",539,"len")=8
	Set gtmtypes("shm_snapshot_t",539,"type")="struct"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.next_upgrd_warn.time_latch.u.parts")=539
	Set gtmtypes("shm_snapshot_t",540,"name")="shm_snapshot_t.shadow_file_header.next_upgrd_warn.time_latch.u.parts.latch_pid"
	Set gtmtypes("shm_snapshot_t",540,"off")=10328
	Set gtmtypes("shm_snapshot_t",540,"len")=4
	Set gtmtypes("shm_snapshot_t",540,"type")="int"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.next_upgrd_warn.time_latch.u.parts.latch_pid")=540
	Set gtmtypes("shm_snapshot_t",541,"name")="shm_snapshot_t.shadow_file_header.next_upgrd_warn.time_latch.u.parts.latch_word"
	Set gtmtypes("shm_snapshot_t",541,"off")=10332
	Set gtmtypes("shm_snapshot_t",541,"len")=4
	Set gtmtypes("shm_snapshot_t",541,"type")="int"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.next_upgrd_warn.time_latch.u.parts.latch_word")=541
	Set gtmtypes("shm_snapshot_t",542,"name")="shm_snapshot_t.shadow_file_header.next_upgrd_warn.hp_latch_space"
	Set gtmtypes("shm_snapshot_t",542,"off")=10336
	Set gtmtypes("shm_snapshot_t",542,"len")=16
	Set gtmtypes("shm_snapshot_t",542,"type")="int"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.next_upgrd_warn.hp_latch_space")=542
	Set gtmtypes("shm_snapshot_t",542,"dim")=4
	Set gtmtypes("shm_snapshot_t",543,"name")="shm_snapshot_t.shadow_file_header.is_encrypted"
	Set gtmtypes("shm_snapshot_t",543,"off")=10352
	Set gtmtypes("shm_snapshot_t",543,"len")=4
	Set gtmtypes("shm_snapshot_t",543,"type")="unsigned-int"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.is_encrypted")=543
	Set gtmtypes("shm_snapshot_t",544,"name")="shm_snapshot_t.shadow_file_header.db_trigger_cycle"
	Set gtmtypes("shm_snapshot_t",544,"off")=10356
	Set gtmtypes("shm_snapshot_t",544,"len")=4
	Set gtmtypes("shm_snapshot_t",544,"type")="unsigned-int"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.db_trigger_cycle")=544
	Set gtmtypes("shm_snapshot_t",545,"name")="shm_snapshot_t.shadow_file_header.strm_reg_seqno"
	Set gtmtypes("shm_snapshot_t",545,"off")=10360
	Set gtmtypes("shm_snapshot_t",545,"len")=128
	Set gtmtypes("shm_snapshot_t",545,"type")="uint64_t"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.strm_reg_seqno")=545
	Set gtmtypes("shm_snapshot_t",545,"dim")=16
	Set gtmtypes("shm_snapshot_t",546,"name")="shm_snapshot_t.shadow_file_header.save_strm_reg_seqno"
	Set gtmtypes("shm_snapshot_t",546,"off")=10488
	Set gtmtypes("shm_snapshot_t",546,"len")=128
	Set gtmtypes("shm_snapshot_t",546,"type")="uint64_t"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.save_strm_reg_seqno")=546
	Set gtmtypes("shm_snapshot_t",546,"dim")=16
	Set gtmtypes("shm_snapshot_t",547,"name")="shm_snapshot_t.shadow_file_header.freeze_on_fail"
	Set gtmtypes("shm_snapshot_t",547,"off")=10616
	Set gtmtypes("shm_snapshot_t",547,"len")=4
	Set gtmtypes("shm_snapshot_t",547,"type")="boolean_t"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.freeze_on_fail")=547
	Set gtmtypes("shm_snapshot_t",548,"name")="shm_snapshot_t.shadow_file_header.span_node_absent"
	Set gtmtypes("shm_snapshot_t",548,"off")=10620
	Set gtmtypes("shm_snapshot_t",548,"len")=4
	Set gtmtypes("shm_snapshot_t",548,"type")="boolean_t"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.span_node_absent")=548
	Set gtmtypes("shm_snapshot_t",549,"name")="shm_snapshot_t.shadow_file_header.maxkeysz_assured"
	Set gtmtypes("shm_snapshot_t",549,"off")=10624
	Set gtmtypes("shm_snapshot_t",549,"len")=4
	Set gtmtypes("shm_snapshot_t",549,"type")="boolean_t"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.maxkeysz_assured")=549
	Set gtmtypes("shm_snapshot_t",550,"name")="shm_snapshot_t.shadow_file_header.hasht_upgrade_needed"
	Set gtmtypes("shm_snapshot_t",550,"off")=10628
	Set gtmtypes("shm_snapshot_t",550,"len")=4
	Set gtmtypes("shm_snapshot_t",550,"type")="boolean_t"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.hasht_upgrade_needed")=550
	Set gtmtypes("shm_snapshot_t",551,"name")="shm_snapshot_t.shadow_file_header.defer_allocate"
	Set gtmtypes("shm_snapshot_t",551,"off")=10632
	Set gtmtypes("shm_snapshot_t",551,"len")=4
	Set gtmtypes("shm_snapshot_t",551,"type")="boolean_t"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.defer_allocate")=551
	Set gtmtypes("shm_snapshot_t",552,"name")="shm_snapshot_t.shadow_file_header.filler_ftok_counter_halted"
	Set gtmtypes("shm_snapshot_t",552,"off")=10636
	Set gtmtypes("shm_snapshot_t",552,"len")=4
	Set gtmtypes("shm_snapshot_t",552,"type")="boolean_t"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.filler_ftok_counter_halted")=552
	Set gtmtypes("shm_snapshot_t",553,"name")="shm_snapshot_t.shadow_file_header.filler_access_counter_halted"
	Set gtmtypes("shm_snapshot_t",553,"off")=10640
	Set gtmtypes("shm_snapshot_t",553,"len")=4
	Set gtmtypes("shm_snapshot_t",553,"type")="boolean_t"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.filler_access_counter_halted")=553
	Set gtmtypes("shm_snapshot_t",554,"name")="shm_snapshot_t.shadow_file_header.lock_crit_with_db"
	Set gtmtypes("shm_snapshot_t",554,"off")=10644
	Set gtmtypes("shm_snapshot_t",554,"len")=4
	Set gtmtypes("shm_snapshot_t",554,"type")="boolean_t"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.lock_crit_with_db")=554
	Set gtmtypes("shm_snapshot_t",555,"name")="shm_snapshot_t.shadow_file_header.basedb_fname_len"
	Set gtmtypes("shm_snapshot_t",555,"off")=10648
	Set gtmtypes("shm_snapshot_t",555,"len")=4
	Set gtmtypes("shm_snapshot_t",555,"type")="unsigned-int"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.basedb_fname_len")=555
	Set gtmtypes("shm_snapshot_t",556,"name")="shm_snapshot_t.shadow_file_header.basedb_fname"
	Set gtmtypes("shm_snapshot_t",556,"off")=10652
	Set gtmtypes("shm_snapshot_t",556,"len")=256
	Set gtmtypes("shm_snapshot_t",556,"type")="unsigned-char"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.basedb_fname")=556
	Set gtmtypes("shm_snapshot_t",557,"name")="shm_snapshot_t.shadow_file_header.read_only"
	Set gtmtypes("shm_snapshot_t",557,"off")=10908
	Set gtmtypes("shm_snapshot_t",557,"len")=4
	Set gtmtypes("shm_snapshot_t",557,"type")="boolean_t"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.read_only")=557
	Set gtmtypes("shm_snapshot_t",558,"name")="shm_snapshot_t.shadow_file_header.gvstats_rec"
	Set gtmtypes("shm_snapshot_t",558,"off")=10912
	Set gtmtypes("shm_snapshot_t",558,"len")=592
	Set gtmtypes("shm_snapshot_t",558,"type")="gvstats_rec_csd_t"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.gvstats_rec")=558
	Set gtmtypes("shm_snapshot_t",559,"name")="shm_snapshot_t.shadow_file_header.gvstats_rec.n_set"
	Set gtmtypes("shm_snapshot_t",559,"off")=10912
	Set gtmtypes("shm_snapshot_t",559,"len")=8
	Set gtmtypes("shm_snapshot_t",559,"type")="uint64_t"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.gvstats_rec.n_set")=559
	Set gtmtypes("shm_snapshot_t",560,"name")="shm_snapshot_t.shadow_file_header.gvstats_rec.n_kill"
	Set gtmtypes("shm_snapshot_t",560,"off")=10920
	Set gtmtypes("shm_snapshot_t",560,"len")=8
	Set gtmtypes("shm_snapshot_t",560,"type")="uint64_t"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.gvstats_rec.n_kill")=560
	Set gtmtypes("shm_snapshot_t",561,"name")="shm_snapshot_t.shadow_file_header.gvstats_rec.n_get"
	Set gtmtypes("shm_snapshot_t",561,"off")=10928
	Set gtmtypes("shm_snapshot_t",561,"len")=8
	Set gtmtypes("shm_snapshot_t",561,"type")="uint64_t"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.gvstats_rec.n_get")=561
	Set gtmtypes("shm_snapshot_t",562,"name")="shm_snapshot_t.shadow_file_header.gvstats_rec.n_data"
	Set gtmtypes("shm_snapshot_t",562,"off")=10936
	Set gtmtypes("shm_snapshot_t",562,"len")=8
	Set gtmtypes("shm_snapshot_t",562,"type")="uint64_t"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.gvstats_rec.n_data")=562
	Set gtmtypes("shm_snapshot_t",563,"name")="shm_snapshot_t.shadow_file_header.gvstats_rec.n_order"
	Set gtmtypes("shm_snapshot_t",563,"off")=10944
	Set gtmtypes("shm_snapshot_t",563,"len")=8
	Set gtmtypes("shm_snapshot_t",563,"type")="uint64_t"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.gvstats_rec.n_order")=563
	Set gtmtypes("shm_snapshot_t",564,"name")="shm_snapshot_t.shadow_file_header.gvstats_rec.n_zprev"
	Set gtmtypes("shm_snapshot_t",564,"off")=10952
	Set gtmtypes("shm_snapshot_t",564,"len")=8
	Set gtmtypes("shm_snapshot_t",564,"type")="uint64_t"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.gvstats_rec.n_zprev")=564
	Set gtmtypes("shm_snapshot_t",565,"name")="shm_snapshot_t.shadow_file_header.gvstats_rec.n_query"
	Set gtmtypes("shm_snapshot_t",565,"off")=10960
	Set gtmtypes("shm_snapshot_t",565,"len")=8
	Set gtmtypes("shm_snapshot_t",565,"type")="uint64_t"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.gvstats_rec.n_query")=565
	Set gtmtypes("shm_snapshot_t",566,"name")="shm_snapshot_t.shadow_file_header.gvstats_rec.n_lock_success"
	Set gtmtypes("shm_snapshot_t",566,"off")=10968
	Set gtmtypes("shm_snapshot_t",566,"len")=8
	Set gtmtypes("shm_snapshot_t",566,"type")="uint64_t"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.gvstats_rec.n_lock_success")=566
	Set gtmtypes("shm_snapshot_t",567,"name")="shm_snapshot_t.shadow_file_header.gvstats_rec.n_lock_fail"
	Set gtmtypes("shm_snapshot_t",567,"off")=10976
	Set gtmtypes("shm_snapshot_t",567,"len")=8
	Set gtmtypes("shm_snapshot_t",567,"type")="uint64_t"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.gvstats_rec.n_lock_fail")=567
	Set gtmtypes("shm_snapshot_t",568,"name")="shm_snapshot_t.shadow_file_header.gvstats_rec.db_curr_tn"
	Set gtmtypes("shm_snapshot_t",568,"off")=10984
	Set gtmtypes("shm_snapshot_t",568,"len")=8
	Set gtmtypes("shm_snapshot_t",568,"type")="uint64_t"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.gvstats_rec.db_curr_tn")=568
	Set gtmtypes("shm_snapshot_t",569,"name")="shm_snapshot_t.shadow_file_header.gvstats_rec.n_dsk_read"
	Set gtmtypes("shm_snapshot_t",569,"off")=10992
	Set gtmtypes("shm_snapshot_t",569,"len")=8
	Set gtmtypes("shm_snapshot_t",569,"type")="uint64_t"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.gvstats_rec.n_dsk_read")=569
	Set gtmtypes("shm_snapshot_t",570,"name")="shm_snapshot_t.shadow_file_header.gvstats_rec.n_dsk_write"
	Set gtmtypes("shm_snapshot_t",570,"off")=11000
	Set gtmtypes("shm_snapshot_t",570,"len")=8
	Set gtmtypes("shm_snapshot_t",570,"type")="uint64_t"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.gvstats_rec.n_dsk_write")=570
	Set gtmtypes("shm_snapshot_t",571,"name")="shm_snapshot_t.shadow_file_header.gvstats_rec.n_nontp_readwrite"
	Set gtmtypes("shm_snapshot_t",571,"off")=11008
	Set gtmtypes("shm_snapshot_t",571,"len")=8
	Set gtmtypes("shm_snapshot_t",571,"type")="uint64_t"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.gvstats_rec.n_nontp_readwrite")=571
	Set gtmtypes("shm_snapshot_t",572,"name")="shm_snapshot_t.shadow_file_header.gvstats_rec.n_nontp_readonly"
	Set gtmtypes("shm_snapshot_t",572,"off")=11016
	Set gtmtypes("shm_snapshot_t",572,"len")=8
	Set gtmtypes("shm_snapshot_t",572,"type")="uint64_t"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.gvstats_rec.n_nontp_readonly")=572
	Set gtmtypes("shm_snapshot_t",573,"name")="shm_snapshot_t.shadow_file_header.gvstats_rec.n_nontp_blkwrite"
	Set gtmtypes("shm_snapshot_t",573,"off")=11024
	Set gtmtypes("shm_snapshot_t",573,"len")=8
	Set gtmtypes("shm_snapshot_t",573,"type")="uint64_t"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.gvstats_rec.n_nontp_blkwrite")=573
	Set gtmtypes("shm_snapshot_t",574,"name")="shm_snapshot_t.shadow_file_header.gvstats_rec.n_nontp_blkread"
	Set gtmtypes("shm_snapshot_t",574,"off")=11032
	Set gtmtypes("shm_snapshot_t",574,"len")=8
	Set gtmtypes("shm_snapshot_t",574,"type")="uint64_t"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.gvstats_rec.n_nontp_blkread")=574
	Set gtmtypes("shm_snapshot_t",575,"name")="shm_snapshot_t.shadow_file_header.gvstats_rec.n_nontp_retries_0"
	Set gtmtypes("shm_snapshot_t",575,"off")=11040
	Set gtmtypes("shm_snapshot_t",575,"len")=8
	Set gtmtypes("shm_snapshot_t",575,"type")="uint64_t"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.gvstats_rec.n_nontp_retries_0")=575
	Set gtmtypes("shm_snapshot_t",576,"name")="shm_snapshot_t.shadow_file_header.gvstats_rec.n_nontp_retries_1"
	Set gtmtypes("shm_snapshot_t",576,"off")=11048
	Set gtmtypes("shm_snapshot_t",576,"len")=8
	Set gtmtypes("shm_snapshot_t",576,"type")="uint64_t"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.gvstats_rec.n_nontp_retries_1")=576
	Set gtmtypes("shm_snapshot_t",577,"name")="shm_snapshot_t.shadow_file_header.gvstats_rec.n_nontp_retries_2"
	Set gtmtypes("shm_snapshot_t",577,"off")=11056
	Set gtmtypes("shm_snapshot_t",577,"len")=8
	Set gtmtypes("shm_snapshot_t",577,"type")="uint64_t"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.gvstats_rec.n_nontp_retries_2")=577
	Set gtmtypes("shm_snapshot_t",578,"name")="shm_snapshot_t.shadow_file_header.gvstats_rec.n_nontp_retries_3"
	Set gtmtypes("shm_snapshot_t",578,"off")=11064
	Set gtmtypes("shm_snapshot_t",578,"len")=8
	Set gtmtypes("shm_snapshot_t",578,"type")="uint64_t"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.gvstats_rec.n_nontp_retries_3")=578
	Set gtmtypes("shm_snapshot_t",579,"name")="shm_snapshot_t.shadow_file_header.gvstats_rec.n_tp_readwrite"
	Set gtmtypes("shm_snapshot_t",579,"off")=11072
	Set gtmtypes("shm_snapshot_t",579,"len")=8
	Set gtmtypes("shm_snapshot_t",579,"type")="uint64_t"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.gvstats_rec.n_tp_readwrite")=579
	Set gtmtypes("shm_snapshot_t",580,"name")="shm_snapshot_t.shadow_file_header.gvstats_rec.n_tp_readonly"
	Set gtmtypes("shm_snapshot_t",580,"off")=11080
	Set gtmtypes("shm_snapshot_t",580,"len")=8
	Set gtmtypes("shm_snapshot_t",580,"type")="uint64_t"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.gvstats_rec.n_tp_readonly")=580
	Set gtmtypes("shm_snapshot_t",581,"name")="shm_snapshot_t.shadow_file_header.gvstats_rec.n_tp_rolledback"
	Set gtmtypes("shm_snapshot_t",581,"off")=11088
	Set gtmtypes("shm_snapshot_t",581,"len")=8
	Set gtmtypes("shm_snapshot_t",581,"type")="uint64_t"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.gvstats_rec.n_tp_rolledback")=581
	Set gtmtypes("shm_snapshot_t",582,"name")="shm_snapshot_t.shadow_file_header.gvstats_rec.n_tp_blkwrite"
	Set gtmtypes("shm_snapshot_t",582,"off")=11096
	Set gtmtypes("shm_snapshot_t",582,"len")=8
	Set gtmtypes("shm_snapshot_t",582,"type")="uint64_t"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.gvstats_rec.n_tp_blkwrite")=582
	Set gtmtypes("shm_snapshot_t",583,"name")="shm_snapshot_t.shadow_file_header.gvstats_rec.n_tp_blkread"
	Set gtmtypes("shm_snapshot_t",583,"off")=11104
	Set gtmtypes("shm_snapshot_t",583,"len")=8
	Set gtmtypes("shm_snapshot_t",583,"type")="uint64_t"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.gvstats_rec.n_tp_blkread")=583
	Set gtmtypes("shm_snapshot_t",584,"name")="shm_snapshot_t.shadow_file_header.gvstats_rec.n_tp_tot_retries_0"
	Set gtmtypes("shm_snapshot_t",584,"off")=11112
	Set gtmtypes("shm_snapshot_t",584,"len")=8
	Set gtmtypes("shm_snapshot_t",584,"type")="uint64_t"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.gvstats_rec.n_tp_tot_retries_0")=584
	Set gtmtypes("shm_snapshot_t",585,"name")="shm_snapshot_t.shadow_file_header.gvstats_rec.n_tp_tot_retries_1"
	Set gtmtypes("shm_snapshot_t",585,"off")=11120
	Set gtmtypes("shm_snapshot_t",585,"len")=8
	Set gtmtypes("shm_snapshot_t",585,"type")="uint64_t"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.gvstats_rec.n_tp_tot_retries_1")=585
	Set gtmtypes("shm_snapshot_t",586,"name")="shm_snapshot_t.shadow_file_header.gvstats_rec.n_tp_tot_retries_2"
	Set gtmtypes("shm_snapshot_t",586,"off")=11128
	Set gtmtypes("shm_snapshot_t",586,"len")=8
	Set gtmtypes("shm_snapshot_t",586,"type")="uint64_t"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.gvstats_rec.n_tp_tot_retries_2")=586
	Set gtmtypes("shm_snapshot_t",587,"name")="shm_snapshot_t.shadow_file_header.gvstats_rec.n_tp_tot_retries_3"
	Set gtmtypes("shm_snapshot_t",587,"off")=11136
	Set gtmtypes("shm_snapshot_t",587,"len")=8
	Set gtmtypes("shm_snapshot_t",587,"type")="uint64_t"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.gvstats_rec.n_tp_tot_retries_3")=587
	Set gtmtypes("shm_snapshot_t",588,"name")="shm_snapshot_t.shadow_file_header.gvstats_rec.n_tp_tot_retries_4"
	Set gtmtypes("shm_snapshot_t",588,"off")=11144
	Set gtmtypes("shm_snapshot_t",588,"len")=8
	Set gtmtypes("shm_snapshot_t",588,"type")="uint64_t"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.gvstats_rec.n_tp_tot_retries_4")=588
	Set gtmtypes("shm_snapshot_t",589,"name")="shm_snapshot_t.shadow_file_header.gvstats_rec.n_tp_cnflct_retries_0"
	Set gtmtypes("shm_snapshot_t",589,"off")=11152
	Set gtmtypes("shm_snapshot_t",589,"len")=8
	Set gtmtypes("shm_snapshot_t",589,"type")="uint64_t"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.gvstats_rec.n_tp_cnflct_retries_0")=589
	Set gtmtypes("shm_snapshot_t",590,"name")="shm_snapshot_t.shadow_file_header.gvstats_rec.n_tp_cnflct_retries_1"
	Set gtmtypes("shm_snapshot_t",590,"off")=11160
	Set gtmtypes("shm_snapshot_t",590,"len")=8
	Set gtmtypes("shm_snapshot_t",590,"type")="uint64_t"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.gvstats_rec.n_tp_cnflct_retries_1")=590
	Set gtmtypes("shm_snapshot_t",591,"name")="shm_snapshot_t.shadow_file_header.gvstats_rec.n_tp_cnflct_retries_2"
	Set gtmtypes("shm_snapshot_t",591,"off")=11168
	Set gtmtypes("shm_snapshot_t",591,"len")=8
	Set gtmtypes("shm_snapshot_t",591,"type")="uint64_t"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.gvstats_rec.n_tp_cnflct_retries_2")=591
	Set gtmtypes("shm_snapshot_t",592,"name")="shm_snapshot_t.shadow_file_header.gvstats_rec.n_tp_cnflct_retries_3"
	Set gtmtypes("shm_snapshot_t",592,"off")=11176
	Set gtmtypes("shm_snapshot_t",592,"len")=8
	Set gtmtypes("shm_snapshot_t",592,"type")="uint64_t"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.gvstats_rec.n_tp_cnflct_retries_3")=592
	Set gtmtypes("shm_snapshot_t",593,"name")="shm_snapshot_t.shadow_file_header.gvstats_rec.n_tp_cnflct_retries_4"
	Set gtmtypes("shm_snapshot_t",593,"off")=11184
	Set gtmtypes("shm_snapshot_t",593,"len")=8
	Set gtmtypes("shm_snapshot_t",593,"type")="uint64_t"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.gvstats_rec.n_tp_cnflct_retries_4")=593
	Set gtmtypes("shm_snapshot_t",594,"name")="shm_snapshot_t.shadow_file_header.gvstats_rec.n_ztrigger"
	Set gtmtypes("shm_snapshot_t",594,"off")=11192
	Set gtmtypes("shm_snapshot_t",594,"len")=8
	Set gtmtypes("shm_snapshot_t",594,"type")="uint64_t"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.gvstats_rec.n_ztrigger")=594
	Set gtmtypes("shm_snapshot_t",595,"name")="shm_snapshot_t.shadow_file_header.gvstats_rec.n_db_flush"
	Set gtmtypes("shm_snapshot_t",595,"off")=11200
	Set gtmtypes("shm_snapshot_t",595,"len")=8
	Set gtmtypes("shm_snapshot_t",595,"type")="uint64_t"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.gvstats_rec.n_db_flush")=595
	Set gtmtypes("shm_snapshot_t",596,"name")="shm_snapshot_t.shadow_file_header.gvstats_rec.n_db_fsync"
	Set gtmtypes("shm_snapshot_t",596,"off")=11208
	Set gtmtypes("shm_snapshot_t",596,"len")=8
	Set gtmtypes("shm_snapshot_t",596,"type")="uint64_t"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.gvstats_rec.n_db_fsync")=596
	Set gtmtypes("shm_snapshot_t",597,"name")="shm_snapshot_t.shadow_file_header.gvstats_rec.n_jnl_flush"
	Set gtmtypes("shm_snapshot_t",597,"off")=11216
	Set gtmtypes("shm_snapshot_t",597,"len")=8
	Set gtmtypes("shm_snapshot_t",597,"type")="uint64_t"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.gvstats_rec.n_jnl_flush")=597
	Set gtmtypes("shm_snapshot_t",598,"name")="shm_snapshot_t.shadow_file_header.gvstats_rec.n_jnl_fsync"
	Set gtmtypes("shm_snapshot_t",598,"off")=11224
	Set gtmtypes("shm_snapshot_t",598,"len")=8
	Set gtmtypes("shm_snapshot_t",598,"type")="uint64_t"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.gvstats_rec.n_jnl_fsync")=598
	Set gtmtypes("shm_snapshot_t",599,"name")="shm_snapshot_t.shadow_file_header.gvstats_rec.n_jbuff_bytes"
	Set gtmtypes("shm_snapshot_t",599,"off")=11232
	Set gtmtypes("shm_snapshot_t",599,"len")=8
	Set gtmtypes("shm_snapshot_t",599,"type")="uint64_t"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.gvstats_rec.n_jbuff_bytes")=599
	Set gtmtypes("shm_snapshot_t",600,"name")="shm_snapshot_t.shadow_file_header.gvstats_rec.n_jfile_bytes"
	Set gtmtypes("shm_snapshot_t",600,"off")=11240
	Set gtmtypes("shm_snapshot_t",600,"len")=8
	Set gtmtypes("shm_snapshot_t",600,"type")="uint64_t"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.gvstats_rec.n_jfile_bytes")=600
	Set gtmtypes("shm_snapshot_t",601,"name")="shm_snapshot_t.shadow_file_header.gvstats_rec.n_jfile_writes"
	Set gtmtypes("shm_snapshot_t",601,"off")=11248
	Set gtmtypes("shm_snapshot_t",601,"len")=8
	Set gtmtypes("shm_snapshot_t",601,"type")="uint64_t"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.gvstats_rec.n_jfile_writes")=601
	Set gtmtypes("shm_snapshot_t",602,"name")="shm_snapshot_t.shadow_file_header.gvstats_rec.n_jrec_logical"
	Set gtmtypes("shm_snapshot_t",602,"off")=11256
	Set gtmtypes("shm_snapshot_t",602,"len")=8
	Set gtmtypes("shm_snapshot_t",602,"type")="uint64_t"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.gvstats_rec.n_jrec_logical")=602
	Set gtmtypes("shm_snapshot_t",603,"name")="shm_snapshot_t.shadow_file_header.gvstats_rec.n_jrec_pblk"
	Set gtmtypes("shm_snapshot_t",603,"off")=11264
	Set gtmtypes("shm_snapshot_t",603,"len")=8
	Set gtmtypes("shm_snapshot_t",603,"type")="uint64_t"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.gvstats_rec.n_jrec_pblk")=603
	Set gtmtypes("shm_snapshot_t",604,"name")="shm_snapshot_t.shadow_file_header.gvstats_rec.n_jrec_epoch_regular"
	Set gtmtypes("shm_snapshot_t",604,"off")=11272
	Set gtmtypes("shm_snapshot_t",604,"len")=8
	Set gtmtypes("shm_snapshot_t",604,"type")="uint64_t"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.gvstats_rec.n_jrec_epoch_regular")=604
	Set gtmtypes("shm_snapshot_t",605,"name")="shm_snapshot_t.shadow_file_header.gvstats_rec.n_jrec_epoch_idle"
	Set gtmtypes("shm_snapshot_t",605,"off")=11280
	Set gtmtypes("shm_snapshot_t",605,"len")=8
	Set gtmtypes("shm_snapshot_t",605,"type")="uint64_t"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.gvstats_rec.n_jrec_epoch_idle")=605
	Set gtmtypes("shm_snapshot_t",606,"name")="shm_snapshot_t.shadow_file_header.gvstats_rec.n_jrec_other"
	Set gtmtypes("shm_snapshot_t",606,"off")=11288
	Set gtmtypes("shm_snapshot_t",606,"len")=8
	Set gtmtypes("shm_snapshot_t",606,"type")="uint64_t"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.gvstats_rec.n_jrec_other")=606
	Set gtmtypes("shm_snapshot_t",607,"name")="shm_snapshot_t.shadow_file_header.gvstats_rec.n_jnl_extends"
	Set gtmtypes("shm_snapshot_t",607,"off")=11296
	Set gtmtypes("shm_snapshot_t",607,"len")=8
	Set gtmtypes("shm_snapshot_t",607,"type")="uint64_t"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.gvstats_rec.n_jnl_extends")=607
	Set gtmtypes("shm_snapshot_t",608,"name")="shm_snapshot_t.shadow_file_header.gvstats_rec.n_db_extends"
	Set gtmtypes("shm_snapshot_t",608,"off")=11304
	Set gtmtypes("shm_snapshot_t",608,"len")=8
	Set gtmtypes("shm_snapshot_t",608,"type")="uint64_t"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.gvstats_rec.n_db_extends")=608
	Set gtmtypes("shm_snapshot_t",609,"name")="shm_snapshot_t.shadow_file_header.gvstats_rec.n_crit_success"
	Set gtmtypes("shm_snapshot_t",609,"off")=11312
	Set gtmtypes("shm_snapshot_t",609,"len")=8
	Set gtmtypes("shm_snapshot_t",609,"type")="uint64_t"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.gvstats_rec.n_crit_success")=609
	Set gtmtypes("shm_snapshot_t",610,"name")="shm_snapshot_t.shadow_file_header.gvstats_rec.n_crits_in_epch"
	Set gtmtypes("shm_snapshot_t",610,"off")=11320
	Set gtmtypes("shm_snapshot_t",610,"len")=8
	Set gtmtypes("shm_snapshot_t",610,"type")="uint64_t"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.gvstats_rec.n_crits_in_epch")=610
	Set gtmtypes("shm_snapshot_t",611,"name")="shm_snapshot_t.shadow_file_header.gvstats_rec.sq_crit_failed"
	Set gtmtypes("shm_snapshot_t",611,"off")=11328
	Set gtmtypes("shm_snapshot_t",611,"len")=8
	Set gtmtypes("shm_snapshot_t",611,"type")="uint64_t"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.gvstats_rec.sq_crit_failed")=611
	Set gtmtypes("shm_snapshot_t",612,"name")="shm_snapshot_t.shadow_file_header.gvstats_rec.n_crit_failed"
	Set gtmtypes("shm_snapshot_t",612,"off")=11336
	Set gtmtypes("shm_snapshot_t",612,"len")=8
	Set gtmtypes("shm_snapshot_t",612,"type")="uint64_t"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.gvstats_rec.n_crit_failed")=612
	Set gtmtypes("shm_snapshot_t",613,"name")="shm_snapshot_t.shadow_file_header.gvstats_rec.sq_crit_que_slps"
	Set gtmtypes("shm_snapshot_t",613,"off")=11344
	Set gtmtypes("shm_snapshot_t",613,"len")=8
	Set gtmtypes("shm_snapshot_t",613,"type")="uint64_t"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.gvstats_rec.sq_crit_que_slps")=613
	Set gtmtypes("shm_snapshot_t",614,"name")="shm_snapshot_t.shadow_file_header.gvstats_rec.n_crit_que_slps"
	Set gtmtypes("shm_snapshot_t",614,"off")=11352
	Set gtmtypes("shm_snapshot_t",614,"len")=8
	Set gtmtypes("shm_snapshot_t",614,"type")="uint64_t"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.gvstats_rec.n_crit_que_slps")=614
	Set gtmtypes("shm_snapshot_t",615,"name")="shm_snapshot_t.shadow_file_header.gvstats_rec.sq_crit_yields"
	Set gtmtypes("shm_snapshot_t",615,"off")=11360
	Set gtmtypes("shm_snapshot_t",615,"len")=8
	Set gtmtypes("shm_snapshot_t",615,"type")="uint64_t"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.gvstats_rec.sq_crit_yields")=615
	Set gtmtypes("shm_snapshot_t",616,"name")="shm_snapshot_t.shadow_file_header.gvstats_rec.n_crit_yields"
	Set gtmtypes("shm_snapshot_t",616,"off")=11368
	Set gtmtypes("shm_snapshot_t",616,"len")=8
	Set gtmtypes("shm_snapshot_t",616,"type")="uint64_t"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.gvstats_rec.n_crit_yields")=616
	Set gtmtypes("shm_snapshot_t",617,"name")="shm_snapshot_t.shadow_file_header.gvstats_rec.n_clean2dirty"
	Set gtmtypes("shm_snapshot_t",617,"off")=11376
	Set gtmtypes("shm_snapshot_t",617,"len")=8
	Set gtmtypes("shm_snapshot_t",617,"type")="uint64_t"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.gvstats_rec.n_clean2dirty")=617
	Set gtmtypes("shm_snapshot_t",618,"name")="shm_snapshot_t.shadow_file_header.gvstats_rec.n_wait_for_read"
	Set gtmtypes("shm_snapshot_t",618,"off")=11384
	Set gtmtypes("shm_snapshot_t",618,"len")=8
	Set gtmtypes("shm_snapshot_t",618,"type")="uint64_t"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.gvstats_rec.n_wait_for_read")=618
	Set gtmtypes("shm_snapshot_t",619,"name")="shm_snapshot_t.shadow_file_header.gvstats_rec.n_buffer_scarce"
	Set gtmtypes("shm_snapshot_t",619,"off")=11392
	Set gtmtypes("shm_snapshot_t",619,"len")=8
	Set gtmtypes("shm_snapshot_t",619,"type")="uint64_t"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.gvstats_rec.n_buffer_scarce")=619
	Set gtmtypes("shm_snapshot_t",620,"name")="shm_snapshot_t.shadow_file_header.gvstats_rec.n_bt_scarce"
	Set gtmtypes("shm_snapshot_t",620,"off")=11400
	Set gtmtypes("shm_snapshot_t",620,"len")=8
	Set gtmtypes("shm_snapshot_t",620,"type")="uint64_t"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.gvstats_rec.n_bt_scarce")=620
	Set gtmtypes("shm_snapshot_t",621,"name")="shm_snapshot_t.shadow_file_header.gvstats_rec.f_dbext_wait"
	Set gtmtypes("shm_snapshot_t",621,"off")=11408
	Set gtmtypes("shm_snapshot_t",621,"len")=8
	Set gtmtypes("shm_snapshot_t",621,"type")="uint64_t"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.gvstats_rec.f_dbext_wait")=621
	Set gtmtypes("shm_snapshot_t",622,"name")="shm_snapshot_t.shadow_file_header.gvstats_rec.f_bg_wait"
	Set gtmtypes("shm_snapshot_t",622,"off")=11416
	Set gtmtypes("shm_snapshot_t",622,"len")=8
	Set gtmtypes("shm_snapshot_t",622,"type")="uint64_t"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.gvstats_rec.f_bg_wait")=622
	Set gtmtypes("shm_snapshot_t",623,"name")="shm_snapshot_t.shadow_file_header.gvstats_rec.f_jnl_wait"
	Set gtmtypes("shm_snapshot_t",623,"off")=11424
	Set gtmtypes("shm_snapshot_t",623,"len")=8
	Set gtmtypes("shm_snapshot_t",623,"type")="uint64_t"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.gvstats_rec.f_jnl_wait")=623
	Set gtmtypes("shm_snapshot_t",624,"name")="shm_snapshot_t.shadow_file_header.gvstats_rec.f_mlk_wait"
	Set gtmtypes("shm_snapshot_t",624,"off")=11432
	Set gtmtypes("shm_snapshot_t",624,"len")=8
	Set gtmtypes("shm_snapshot_t",624,"type")="uint64_t"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.gvstats_rec.f_mlk_wait")=624
	Set gtmtypes("shm_snapshot_t",625,"name")="shm_snapshot_t.shadow_file_header.gvstats_rec.f_proc_wait"
	Set gtmtypes("shm_snapshot_t",625,"off")=11440
	Set gtmtypes("shm_snapshot_t",625,"len")=8
	Set gtmtypes("shm_snapshot_t",625,"type")="uint64_t"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.gvstats_rec.f_proc_wait")=625
	Set gtmtypes("shm_snapshot_t",626,"name")="shm_snapshot_t.shadow_file_header.gvstats_rec.f_trans_wait"
	Set gtmtypes("shm_snapshot_t",626,"off")=11448
	Set gtmtypes("shm_snapshot_t",626,"len")=8
	Set gtmtypes("shm_snapshot_t",626,"type")="uint64_t"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.gvstats_rec.f_trans_wait")=626
	Set gtmtypes("shm_snapshot_t",627,"name")="shm_snapshot_t.shadow_file_header.gvstats_rec.f_util_wait"
	Set gtmtypes("shm_snapshot_t",627,"off")=11456
	Set gtmtypes("shm_snapshot_t",627,"len")=8
	Set gtmtypes("shm_snapshot_t",627,"type")="uint64_t"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.gvstats_rec.f_util_wait")=627
	Set gtmtypes("shm_snapshot_t",628,"name")="shm_snapshot_t.shadow_file_header.gvstats_rec.f_ws2"
	Set gtmtypes("shm_snapshot_t",628,"off")=11464
	Set gtmtypes("shm_snapshot_t",628,"len")=8
	Set gtmtypes("shm_snapshot_t",628,"type")="uint64_t"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.gvstats_rec.f_ws2")=628
	Set gtmtypes("shm_snapshot_t",629,"name")="shm_snapshot_t.shadow_file_header.gvstats_rec.f_ws12"
	Set gtmtypes("shm_snapshot_t",629,"off")=11472
	Set gtmtypes("shm_snapshot_t",629,"len")=8
	Set gtmtypes("shm_snapshot_t",629,"type")="uint64_t"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.gvstats_rec.f_ws12")=629
	Set gtmtypes("shm_snapshot_t",630,"name")="shm_snapshot_t.shadow_file_header.gvstats_rec.f_ws15"
	Set gtmtypes("shm_snapshot_t",630,"off")=11480
	Set gtmtypes("shm_snapshot_t",630,"len")=8
	Set gtmtypes("shm_snapshot_t",630,"type")="uint64_t"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.gvstats_rec.f_ws15")=630
	Set gtmtypes("shm_snapshot_t",631,"name")="shm_snapshot_t.shadow_file_header.gvstats_rec.f_ws39"
	Set gtmtypes("shm_snapshot_t",631,"off")=11488
	Set gtmtypes("shm_snapshot_t",631,"len")=8
	Set gtmtypes("shm_snapshot_t",631,"type")="uint64_t"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.gvstats_rec.f_ws39")=631
	Set gtmtypes("shm_snapshot_t",632,"name")="shm_snapshot_t.shadow_file_header.gvstats_rec.f_ws47"
	Set gtmtypes("shm_snapshot_t",632,"off")=11496
	Set gtmtypes("shm_snapshot_t",632,"len")=8
	Set gtmtypes("shm_snapshot_t",632,"type")="uint64_t"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.gvstats_rec.f_ws47")=632
	Set gtmtypes("shm_snapshot_t",633,"name")="shm_snapshot_t.shadow_file_header.filler_8k"
	Set gtmtypes("shm_snapshot_t",633,"off")=11504
	Set gtmtypes("shm_snapshot_t",633,"len")=872
	Set gtmtypes("shm_snapshot_t",633,"type")="char"
	Set gtmtypfldindx("shm_snapshot_t","shadow_file_header.filler_8k")=633
	;
	Set gtmtypes("shmpool_blk_hdr")="struct"
	Set gtmtypes("shmpool_blk_hdr",0)=11
	Set gtmtypes("shmpool_blk_hdr","len")=40
	Set gtmtypes("shmpool_blk_hdr",1,"name")="shmpool_blk_hdr.sm_que"
	Set gtmtypes("shmpool_blk_hdr",1,"off")=0
	Set gtmtypes("shmpool_blk_hdr",1,"len")=16
	Set gtmtypes("shmpool_blk_hdr",1,"type")="que_ent"
	Set gtmtypfldindx("shmpool_blk_hdr","sm_que")=1
	Set gtmtypes("shmpool_blk_hdr",2,"name")="shmpool_blk_hdr.sm_que.fl"
	Set gtmtypes("shmpool_blk_hdr",2,"off")=0
	Set gtmtypes("shmpool_blk_hdr",2,"len")=8
	Set gtmtypes("shmpool_blk_hdr",2,"type")="intptr_t"
	Set gtmtypfldindx("shmpool_blk_hdr","sm_que.fl")=2
	Set gtmtypes("shmpool_blk_hdr",3,"name")="shmpool_blk_hdr.sm_que.bl"
	Set gtmtypes("shmpool_blk_hdr",3,"off")=8
	Set gtmtypes("shmpool_blk_hdr",3,"len")=8
	Set gtmtypes("shmpool_blk_hdr",3,"type")="intptr_t"
	Set gtmtypfldindx("shmpool_blk_hdr","sm_que.bl")=3
	Set gtmtypes("shmpool_blk_hdr",4,"name")="shmpool_blk_hdr.blktype"
	Set gtmtypes("shmpool_blk_hdr",4,"off")=16
	Set gtmtypes("shmpool_blk_hdr",4,"len")=4
	Set gtmtypes("shmpool_blk_hdr",4,"type")="int"
	Set gtmtypfldindx("shmpool_blk_hdr","blktype")=4
	Set gtmtypes("shmpool_blk_hdr",5,"name")="shmpool_blk_hdr.blkid"
	Set gtmtypes("shmpool_blk_hdr",5,"off")=20
	Set gtmtypes("shmpool_blk_hdr",5,"len")=4
	Set gtmtypes("shmpool_blk_hdr",5,"type")="int"
	Set gtmtypfldindx("shmpool_blk_hdr","blkid")=5
	Set gtmtypes("shmpool_blk_hdr",6,"name")="shmpool_blk_hdr.use"
	Set gtmtypes("shmpool_blk_hdr",6,"off")=24
	Set gtmtypes("shmpool_blk_hdr",6,"len")=4
	Set gtmtypes("shmpool_blk_hdr",6,"type")="union"
	Set gtmtypfldindx("shmpool_blk_hdr","use")=6
	Set gtmtypes("shmpool_blk_hdr",7,"name")="shmpool_blk_hdr.use.bkup"
	Set gtmtypes("shmpool_blk_hdr",7,"off")=24
	Set gtmtypes("shmpool_blk_hdr",7,"len")=4
	Set gtmtypes("shmpool_blk_hdr",7,"type")="struct"
	Set gtmtypfldindx("shmpool_blk_hdr","use.bkup")=7
	Set gtmtypes("shmpool_blk_hdr",8,"name")="shmpool_blk_hdr.use.bkup.ondsk_blkver"
	Set gtmtypes("shmpool_blk_hdr",8,"off")=24
	Set gtmtypes("shmpool_blk_hdr",8,"len")=4
	Set gtmtypes("shmpool_blk_hdr",8,"type")="int"
	Set gtmtypfldindx("shmpool_blk_hdr","use.bkup.ondsk_blkver")=8
	Set gtmtypes("shmpool_blk_hdr",9,"name")="shmpool_blk_hdr.holder_pid"
	Set gtmtypes("shmpool_blk_hdr",9,"off")=28
	Set gtmtypes("shmpool_blk_hdr",9,"len")=4
	Set gtmtypes("shmpool_blk_hdr",9,"type")="pid_t"
	Set gtmtypfldindx("shmpool_blk_hdr","holder_pid")=9
	Set gtmtypes("shmpool_blk_hdr",10,"name")="shmpool_blk_hdr.valid_data"
	Set gtmtypes("shmpool_blk_hdr",10,"off")=32
	Set gtmtypes("shmpool_blk_hdr",10,"len")=4
	Set gtmtypes("shmpool_blk_hdr",10,"type")="boolean_t"
	Set gtmtypfldindx("shmpool_blk_hdr","valid_data")=10
	Set gtmtypes("shmpool_blk_hdr",11,"name")="shmpool_blk_hdr.image_count"
	Set gtmtypes("shmpool_blk_hdr",11,"off")=36
	Set gtmtypes("shmpool_blk_hdr",11,"len")=4
	Set gtmtypes("shmpool_blk_hdr",11,"type")="int"
	Set gtmtypfldindx("shmpool_blk_hdr","image_count")=11
	;
	Set gtmtypes("shmpool_buff_hdr")="struct"
	Set gtmtypes("shmpool_buff_hdr",0)=640
	Set gtmtypes("shmpool_buff_hdr","len")=8560
	Set gtmtypes("shmpool_buff_hdr",1,"name")="shmpool_buff_hdr.shmpool_crit_latch"
	Set gtmtypes("shmpool_buff_hdr",1,"off")=0
	Set gtmtypes("shmpool_buff_hdr",1,"len")=8
	Set gtmtypes("shmpool_buff_hdr",1,"type")="global_latch_t"
	Set gtmtypfldindx("shmpool_buff_hdr","shmpool_crit_latch")=1
	Set gtmtypes("shmpool_buff_hdr",2,"name")="shmpool_buff_hdr.shmpool_crit_latch.u"
	Set gtmtypes("shmpool_buff_hdr",2,"off")=0
	Set gtmtypes("shmpool_buff_hdr",2,"len")=8
	Set gtmtypes("shmpool_buff_hdr",2,"type")="union"
	Set gtmtypfldindx("shmpool_buff_hdr","shmpool_crit_latch.u")=2
	Set gtmtypes("shmpool_buff_hdr",3,"name")="shmpool_buff_hdr.shmpool_crit_latch.u.pid_imgcnt"
	Set gtmtypes("shmpool_buff_hdr",3,"off")=0
	Set gtmtypes("shmpool_buff_hdr",3,"len")=8
	Set gtmtypes("shmpool_buff_hdr",3,"type")="uint64_t"
	Set gtmtypfldindx("shmpool_buff_hdr","shmpool_crit_latch.u.pid_imgcnt")=3
	Set gtmtypes("shmpool_buff_hdr",4,"name")="shmpool_buff_hdr.shmpool_crit_latch.u.parts"
	Set gtmtypes("shmpool_buff_hdr",4,"off")=0
	Set gtmtypes("shmpool_buff_hdr",4,"len")=8
	Set gtmtypes("shmpool_buff_hdr",4,"type")="struct"
	Set gtmtypfldindx("shmpool_buff_hdr","shmpool_crit_latch.u.parts")=4
	Set gtmtypes("shmpool_buff_hdr",5,"name")="shmpool_buff_hdr.shmpool_crit_latch.u.parts.latch_pid"
	Set gtmtypes("shmpool_buff_hdr",5,"off")=0
	Set gtmtypes("shmpool_buff_hdr",5,"len")=4
	Set gtmtypes("shmpool_buff_hdr",5,"type")="int"
	Set gtmtypfldindx("shmpool_buff_hdr","shmpool_crit_latch.u.parts.latch_pid")=5
	Set gtmtypes("shmpool_buff_hdr",6,"name")="shmpool_buff_hdr.shmpool_crit_latch.u.parts.latch_word"
	Set gtmtypes("shmpool_buff_hdr",6,"off")=4
	Set gtmtypes("shmpool_buff_hdr",6,"len")=4
	Set gtmtypes("shmpool_buff_hdr",6,"type")="int"
	Set gtmtypfldindx("shmpool_buff_hdr","shmpool_crit_latch.u.parts.latch_word")=6
	Set gtmtypes("shmpool_buff_hdr",7,"name")="shmpool_buff_hdr.dskaddr"
	Set gtmtypes("shmpool_buff_hdr",7,"off")=8
	Set gtmtypes("shmpool_buff_hdr",7,"len")=8
	Set gtmtypes("shmpool_buff_hdr",7,"type")="off_t"
	Set gtmtypfldindx("shmpool_buff_hdr","dskaddr")=7
	Set gtmtypes("shmpool_buff_hdr",8,"name")="shmpool_buff_hdr.backup_tn"
	Set gtmtypes("shmpool_buff_hdr",8,"off")=16
	Set gtmtypes("shmpool_buff_hdr",8,"len")=8
	Set gtmtypes("shmpool_buff_hdr",8,"type")="uint64_t"
	Set gtmtypfldindx("shmpool_buff_hdr","backup_tn")=8
	Set gtmtypes("shmpool_buff_hdr",9,"name")="shmpool_buff_hdr.inc_backup_tn"
	Set gtmtypes("shmpool_buff_hdr",9,"off")=24
	Set gtmtypes("shmpool_buff_hdr",9,"len")=8
	Set gtmtypes("shmpool_buff_hdr",9,"type")="uint64_t"
	Set gtmtypfldindx("shmpool_buff_hdr","inc_backup_tn")=9
	Set gtmtypes("shmpool_buff_hdr",10,"name")="shmpool_buff_hdr.tempfilename"
	Set gtmtypes("shmpool_buff_hdr",10,"off")=32
	Set gtmtypes("shmpool_buff_hdr",10,"len")=256
	Set gtmtypes("shmpool_buff_hdr",10,"type")="char"
	Set gtmtypfldindx("shmpool_buff_hdr","tempfilename")=10
	Set gtmtypes("shmpool_buff_hdr",11,"name")="shmpool_buff_hdr.que_free"
	Set gtmtypes("shmpool_buff_hdr",11,"off")=288
	Set gtmtypes("shmpool_buff_hdr",11,"len")=16
	Set gtmtypes("shmpool_buff_hdr",11,"type")="que_ent"
	Set gtmtypfldindx("shmpool_buff_hdr","que_free")=11
	Set gtmtypes("shmpool_buff_hdr",12,"name")="shmpool_buff_hdr.que_free.fl"
	Set gtmtypes("shmpool_buff_hdr",12,"off")=288
	Set gtmtypes("shmpool_buff_hdr",12,"len")=8
	Set gtmtypes("shmpool_buff_hdr",12,"type")="intptr_t"
	Set gtmtypfldindx("shmpool_buff_hdr","que_free.fl")=12
	Set gtmtypes("shmpool_buff_hdr",13,"name")="shmpool_buff_hdr.que_free.bl"
	Set gtmtypes("shmpool_buff_hdr",13,"off")=296
	Set gtmtypes("shmpool_buff_hdr",13,"len")=8
	Set gtmtypes("shmpool_buff_hdr",13,"type")="intptr_t"
	Set gtmtypfldindx("shmpool_buff_hdr","que_free.bl")=13
	Set gtmtypes("shmpool_buff_hdr",14,"name")="shmpool_buff_hdr.que_backup"
	Set gtmtypes("shmpool_buff_hdr",14,"off")=304
	Set gtmtypes("shmpool_buff_hdr",14,"len")=16
	Set gtmtypes("shmpool_buff_hdr",14,"type")="que_ent"
	Set gtmtypfldindx("shmpool_buff_hdr","que_backup")=14
	Set gtmtypes("shmpool_buff_hdr",15,"name")="shmpool_buff_hdr.que_backup.fl"
	Set gtmtypes("shmpool_buff_hdr",15,"off")=304
	Set gtmtypes("shmpool_buff_hdr",15,"len")=8
	Set gtmtypes("shmpool_buff_hdr",15,"type")="intptr_t"
	Set gtmtypfldindx("shmpool_buff_hdr","que_backup.fl")=15
	Set gtmtypes("shmpool_buff_hdr",16,"name")="shmpool_buff_hdr.que_backup.bl"
	Set gtmtypes("shmpool_buff_hdr",16,"off")=312
	Set gtmtypes("shmpool_buff_hdr",16,"len")=8
	Set gtmtypes("shmpool_buff_hdr",16,"type")="intptr_t"
	Set gtmtypfldindx("shmpool_buff_hdr","que_backup.bl")=16
	Set gtmtypes("shmpool_buff_hdr",17,"name")="shmpool_buff_hdr.free_cnt"
	Set gtmtypes("shmpool_buff_hdr",17,"off")=320
	Set gtmtypes("shmpool_buff_hdr",17,"len")=4
	Set gtmtypes("shmpool_buff_hdr",17,"type")="int"
	Set gtmtypfldindx("shmpool_buff_hdr","free_cnt")=17
	Set gtmtypes("shmpool_buff_hdr",18,"name")="shmpool_buff_hdr.backup_cnt"
	Set gtmtypes("shmpool_buff_hdr",18,"off")=324
	Set gtmtypes("shmpool_buff_hdr",18,"len")=4
	Set gtmtypes("shmpool_buff_hdr",18,"type")="int"
	Set gtmtypfldindx("shmpool_buff_hdr","backup_cnt")=18
	Set gtmtypes("shmpool_buff_hdr",19,"name")="shmpool_buff_hdr.reformat_cnt"
	Set gtmtypes("shmpool_buff_hdr",19,"off")=328
	Set gtmtypes("shmpool_buff_hdr",19,"len")=4
	Set gtmtypes("shmpool_buff_hdr",19,"type")="int"
	Set gtmtypfldindx("shmpool_buff_hdr","reformat_cnt")=19
	Set gtmtypes("shmpool_buff_hdr",20,"name")="shmpool_buff_hdr.allocs_since_chk"
	Set gtmtypes("shmpool_buff_hdr",20,"off")=332
	Set gtmtypes("shmpool_buff_hdr",20,"len")=4
	Set gtmtypes("shmpool_buff_hdr",20,"type")="int"
	Set gtmtypfldindx("shmpool_buff_hdr","allocs_since_chk")=20
	Set gtmtypes("shmpool_buff_hdr",21,"name")="shmpool_buff_hdr.total_blks"
	Set gtmtypes("shmpool_buff_hdr",21,"off")=336
	Set gtmtypes("shmpool_buff_hdr",21,"len")=4
	Set gtmtypes("shmpool_buff_hdr",21,"type")="unsigned-int"
	Set gtmtypfldindx("shmpool_buff_hdr","total_blks")=21
	Set gtmtypes("shmpool_buff_hdr",22,"name")="shmpool_buff_hdr.blk_size"
	Set gtmtypes("shmpool_buff_hdr",22,"off")=340
	Set gtmtypes("shmpool_buff_hdr",22,"len")=4
	Set gtmtypes("shmpool_buff_hdr",22,"type")="unsigned-int"
	Set gtmtypfldindx("shmpool_buff_hdr","blk_size")=22
	Set gtmtypes("shmpool_buff_hdr",23,"name")="shmpool_buff_hdr.failed"
	Set gtmtypes("shmpool_buff_hdr",23,"off")=344
	Set gtmtypes("shmpool_buff_hdr",23,"len")=4
	Set gtmtypes("shmpool_buff_hdr",23,"type")="pid_t"
	Set gtmtypfldindx("shmpool_buff_hdr","failed")=23
	Set gtmtypes("shmpool_buff_hdr",24,"name")="shmpool_buff_hdr.backup_errno"
	Set gtmtypes("shmpool_buff_hdr",24,"off")=348
	Set gtmtypes("shmpool_buff_hdr",24,"len")=4
	Set gtmtypes("shmpool_buff_hdr",24,"type")="int"
	Set gtmtypfldindx("shmpool_buff_hdr","backup_errno")=24
	Set gtmtypes("shmpool_buff_hdr",25,"name")="shmpool_buff_hdr.backup_pid"
	Set gtmtypes("shmpool_buff_hdr",25,"off")=352
	Set gtmtypes("shmpool_buff_hdr",25,"len")=4
	Set gtmtypes("shmpool_buff_hdr",25,"type")="unsigned-int"
	Set gtmtypfldindx("shmpool_buff_hdr","backup_pid")=25
	Set gtmtypes("shmpool_buff_hdr",26,"name")="shmpool_buff_hdr.backup_image_count"
	Set gtmtypes("shmpool_buff_hdr",26,"off")=356
	Set gtmtypes("shmpool_buff_hdr",26,"len")=4
	Set gtmtypes("shmpool_buff_hdr",26,"type")="unsigned-int"
	Set gtmtypfldindx("shmpool_buff_hdr","backup_image_count")=26
	Set gtmtypes("shmpool_buff_hdr",27,"name")="shmpool_buff_hdr.shmpool_blocked"
	Set gtmtypes("shmpool_buff_hdr",27,"off")=360
	Set gtmtypes("shmpool_buff_hdr",27,"len")=4
	Set gtmtypes("shmpool_buff_hdr",27,"type")="boolean_t"
	Set gtmtypfldindx("shmpool_buff_hdr","shmpool_blocked")=27
	Set gtmtypes("shmpool_buff_hdr",28,"name")="shmpool_buff_hdr.filler"
	Set gtmtypes("shmpool_buff_hdr",28,"off")=364
	Set gtmtypes("shmpool_buff_hdr",28,"len")=4
	Set gtmtypes("shmpool_buff_hdr",28,"type")="unsigned-int"
	Set gtmtypfldindx("shmpool_buff_hdr","filler")=28
	Set gtmtypes("shmpool_buff_hdr",29,"name")="shmpool_buff_hdr.shadow_file_header"
	Set gtmtypes("shmpool_buff_hdr",29,"off")=368
	Set gtmtypes("shmpool_buff_hdr",29,"len")=8192
	Set gtmtypes("shmpool_buff_hdr",29,"type")="sgmnt_data"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header")=29
	Set gtmtypes("shmpool_buff_hdr",30,"name")="shmpool_buff_hdr.shadow_file_header.label"
	Set gtmtypes("shmpool_buff_hdr",30,"off")=368
	Set gtmtypes("shmpool_buff_hdr",30,"len")=12
	Set gtmtypes("shmpool_buff_hdr",30,"type")="unsigned-char"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.label")=30
	Set gtmtypes("shmpool_buff_hdr",31,"name")="shmpool_buff_hdr.shadow_file_header.blk_size"
	Set gtmtypes("shmpool_buff_hdr",31,"off")=380
	Set gtmtypes("shmpool_buff_hdr",31,"len")=4
	Set gtmtypes("shmpool_buff_hdr",31,"type")="int"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.blk_size")=31
	Set gtmtypes("shmpool_buff_hdr",32,"name")="shmpool_buff_hdr.shadow_file_header.master_map_len"
	Set gtmtypes("shmpool_buff_hdr",32,"off")=384
	Set gtmtypes("shmpool_buff_hdr",32,"len")=4
	Set gtmtypes("shmpool_buff_hdr",32,"type")="int"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.master_map_len")=32
	Set gtmtypes("shmpool_buff_hdr",33,"name")="shmpool_buff_hdr.shadow_file_header.bplmap"
	Set gtmtypes("shmpool_buff_hdr",33,"off")=388
	Set gtmtypes("shmpool_buff_hdr",33,"len")=4
	Set gtmtypes("shmpool_buff_hdr",33,"type")="int"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.bplmap")=33
	Set gtmtypes("shmpool_buff_hdr",34,"name")="shmpool_buff_hdr.shadow_file_header.start_vbn"
	Set gtmtypes("shmpool_buff_hdr",34,"off")=392
	Set gtmtypes("shmpool_buff_hdr",34,"len")=4
	Set gtmtypes("shmpool_buff_hdr",34,"type")="int"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.start_vbn")=34
	Set gtmtypes("shmpool_buff_hdr",35,"name")="shmpool_buff_hdr.shadow_file_header.acc_meth"
	Set gtmtypes("shmpool_buff_hdr",35,"off")=396
	Set gtmtypes("shmpool_buff_hdr",35,"len")=4
	Set gtmtypes("shmpool_buff_hdr",35,"type")="int"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.acc_meth")=35
	Set gtmtypes("shmpool_buff_hdr",36,"name")="shmpool_buff_hdr.shadow_file_header.max_bts"
	Set gtmtypes("shmpool_buff_hdr",36,"off")=400
	Set gtmtypes("shmpool_buff_hdr",36,"len")=4
	Set gtmtypes("shmpool_buff_hdr",36,"type")="unsigned-int"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.max_bts")=36
	Set gtmtypes("shmpool_buff_hdr",37,"name")="shmpool_buff_hdr.shadow_file_header.n_bts"
	Set gtmtypes("shmpool_buff_hdr",37,"off")=404
	Set gtmtypes("shmpool_buff_hdr",37,"len")=4
	Set gtmtypes("shmpool_buff_hdr",37,"type")="int"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.n_bts")=37
	Set gtmtypes("shmpool_buff_hdr",38,"name")="shmpool_buff_hdr.shadow_file_header.bt_buckets"
	Set gtmtypes("shmpool_buff_hdr",38,"off")=408
	Set gtmtypes("shmpool_buff_hdr",38,"len")=4
	Set gtmtypes("shmpool_buff_hdr",38,"type")="int"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.bt_buckets")=38
	Set gtmtypes("shmpool_buff_hdr",39,"name")="shmpool_buff_hdr.shadow_file_header.reserved_bytes"
	Set gtmtypes("shmpool_buff_hdr",39,"off")=412
	Set gtmtypes("shmpool_buff_hdr",39,"len")=4
	Set gtmtypes("shmpool_buff_hdr",39,"type")="int"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.reserved_bytes")=39
	Set gtmtypes("shmpool_buff_hdr",40,"name")="shmpool_buff_hdr.shadow_file_header.max_rec_size"
	Set gtmtypes("shmpool_buff_hdr",40,"off")=416
	Set gtmtypes("shmpool_buff_hdr",40,"len")=4
	Set gtmtypes("shmpool_buff_hdr",40,"type")="int"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.max_rec_size")=40
	Set gtmtypes("shmpool_buff_hdr",41,"name")="shmpool_buff_hdr.shadow_file_header.max_key_size"
	Set gtmtypes("shmpool_buff_hdr",41,"off")=420
	Set gtmtypes("shmpool_buff_hdr",41,"len")=4
	Set gtmtypes("shmpool_buff_hdr",41,"type")="int"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.max_key_size")=41
	Set gtmtypes("shmpool_buff_hdr",42,"name")="shmpool_buff_hdr.shadow_file_header.lock_space_size"
	Set gtmtypes("shmpool_buff_hdr",42,"off")=424
	Set gtmtypes("shmpool_buff_hdr",42,"len")=4
	Set gtmtypes("shmpool_buff_hdr",42,"type")="unsigned-int"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.lock_space_size")=42
	Set gtmtypes("shmpool_buff_hdr",43,"name")="shmpool_buff_hdr.shadow_file_header.extension_size"
	Set gtmtypes("shmpool_buff_hdr",43,"off")=428
	Set gtmtypes("shmpool_buff_hdr",43,"len")=4
	Set gtmtypes("shmpool_buff_hdr",43,"type")="unsigned-int"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.extension_size")=43
	Set gtmtypes("shmpool_buff_hdr",44,"name")="shmpool_buff_hdr.shadow_file_header.def_coll"
	Set gtmtypes("shmpool_buff_hdr",44,"off")=432
	Set gtmtypes("shmpool_buff_hdr",44,"len")=4
	Set gtmtypes("shmpool_buff_hdr",44,"type")="unsigned-int"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.def_coll")=44
	Set gtmtypes("shmpool_buff_hdr",45,"name")="shmpool_buff_hdr.shadow_file_header.def_coll_ver"
	Set gtmtypes("shmpool_buff_hdr",45,"off")=436
	Set gtmtypes("shmpool_buff_hdr",45,"len")=4
	Set gtmtypes("shmpool_buff_hdr",45,"type")="unsigned-int"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.def_coll_ver")=45
	Set gtmtypes("shmpool_buff_hdr",46,"name")="shmpool_buff_hdr.shadow_file_header.std_null_coll"
	Set gtmtypes("shmpool_buff_hdr",46,"off")=440
	Set gtmtypes("shmpool_buff_hdr",46,"len")=4
	Set gtmtypes("shmpool_buff_hdr",46,"type")="boolean_t"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.std_null_coll")=46
	Set gtmtypes("shmpool_buff_hdr",47,"name")="shmpool_buff_hdr.shadow_file_header.null_subs"
	Set gtmtypes("shmpool_buff_hdr",47,"off")=444
	Set gtmtypes("shmpool_buff_hdr",47,"len")=4
	Set gtmtypes("shmpool_buff_hdr",47,"type")="int"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.null_subs")=47
	Set gtmtypes("shmpool_buff_hdr",48,"name")="shmpool_buff_hdr.shadow_file_header.free_space"
	Set gtmtypes("shmpool_buff_hdr",48,"off")=448
	Set gtmtypes("shmpool_buff_hdr",48,"len")=4
	Set gtmtypes("shmpool_buff_hdr",48,"type")="unsigned-int"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.free_space")=48
	Set gtmtypes("shmpool_buff_hdr",49,"name")="shmpool_buff_hdr.shadow_file_header.mutex_spin_parms"
	Set gtmtypes("shmpool_buff_hdr",49,"off")=452
	Set gtmtypes("shmpool_buff_hdr",49,"len")=16
	Set gtmtypes("shmpool_buff_hdr",49,"type")="mutex_spin_parms_struct"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.mutex_spin_parms")=49
	Set gtmtypes("shmpool_buff_hdr",50,"name")="shmpool_buff_hdr.shadow_file_header.mutex_spin_parms.mutex_hard_spin_count"
	Set gtmtypes("shmpool_buff_hdr",50,"off")=452
	Set gtmtypes("shmpool_buff_hdr",50,"len")=4
	Set gtmtypes("shmpool_buff_hdr",50,"type")="int"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.mutex_spin_parms.mutex_hard_spin_count")=50
	Set gtmtypes("shmpool_buff_hdr",51,"name")="shmpool_buff_hdr.shadow_file_header.mutex_spin_parms.mutex_sleep_spin_count"
	Set gtmtypes("shmpool_buff_hdr",51,"off")=456
	Set gtmtypes("shmpool_buff_hdr",51,"len")=4
	Set gtmtypes("shmpool_buff_hdr",51,"type")="int"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.mutex_spin_parms.mutex_sleep_spin_count")=51
	Set gtmtypes("shmpool_buff_hdr",52,"name")="shmpool_buff_hdr.shadow_file_header.mutex_spin_parms.mutex_spin_sleep_mask"
	Set gtmtypes("shmpool_buff_hdr",52,"off")=460
	Set gtmtypes("shmpool_buff_hdr",52,"len")=4
	Set gtmtypes("shmpool_buff_hdr",52,"type")="int"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.mutex_spin_parms.mutex_spin_sleep_mask")=52
	Set gtmtypes("shmpool_buff_hdr",53,"name")="shmpool_buff_hdr.shadow_file_header.mutex_spin_parms.mutex_que_entry_space_size"
	Set gtmtypes("shmpool_buff_hdr",53,"off")=464
	Set gtmtypes("shmpool_buff_hdr",53,"len")=4
	Set gtmtypes("shmpool_buff_hdr",53,"type")="int"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.mutex_spin_parms.mutex_que_entry_space_size")=53
	Set gtmtypes("shmpool_buff_hdr",54,"name")="shmpool_buff_hdr.shadow_file_header.max_update_array_size"
	Set gtmtypes("shmpool_buff_hdr",54,"off")=468
	Set gtmtypes("shmpool_buff_hdr",54,"len")=4
	Set gtmtypes("shmpool_buff_hdr",54,"type")="int"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.max_update_array_size")=54
	Set gtmtypes("shmpool_buff_hdr",55,"name")="shmpool_buff_hdr.shadow_file_header.max_non_bm_update_array_size"
	Set gtmtypes("shmpool_buff_hdr",55,"off")=472
	Set gtmtypes("shmpool_buff_hdr",55,"len")=4
	Set gtmtypes("shmpool_buff_hdr",55,"type")="int"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.max_non_bm_update_array_size")=55
	Set gtmtypes("shmpool_buff_hdr",56,"name")="shmpool_buff_hdr.shadow_file_header.file_corrupt"
	Set gtmtypes("shmpool_buff_hdr",56,"off")=476
	Set gtmtypes("shmpool_buff_hdr",56,"len")=4
	Set gtmtypes("shmpool_buff_hdr",56,"type")="boolean_t"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.file_corrupt")=56
	Set gtmtypes("shmpool_buff_hdr",57,"name")="shmpool_buff_hdr.shadow_file_header.minor_dbver"
	Set gtmtypes("shmpool_buff_hdr",57,"off")=480
	Set gtmtypes("shmpool_buff_hdr",57,"len")=4
	Set gtmtypes("shmpool_buff_hdr",57,"type")="int"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.minor_dbver")=57
	Set gtmtypes("shmpool_buff_hdr",58,"name")="shmpool_buff_hdr.shadow_file_header.jnl_checksum"
	Set gtmtypes("shmpool_buff_hdr",58,"off")=484
	Set gtmtypes("shmpool_buff_hdr",58,"len")=4
	Set gtmtypes("shmpool_buff_hdr",58,"type")="unsigned-int"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.jnl_checksum")=58
	Set gtmtypes("shmpool_buff_hdr",59,"name")="shmpool_buff_hdr.shadow_file_header.wcs_phase2_commit_wait_spincnt"
	Set gtmtypes("shmpool_buff_hdr",59,"off")=488
	Set gtmtypes("shmpool_buff_hdr",59,"len")=4
	Set gtmtypes("shmpool_buff_hdr",59,"type")="unsigned-int"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.wcs_phase2_commit_wait_spincnt")=59
	Set gtmtypes("shmpool_buff_hdr",60,"name")="shmpool_buff_hdr.shadow_file_header.last_mdb_ver"
	Set gtmtypes("shmpool_buff_hdr",60,"off")=492
	Set gtmtypes("shmpool_buff_hdr",60,"len")=4
	Set gtmtypes("shmpool_buff_hdr",60,"type")="int"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.last_mdb_ver")=60
	Set gtmtypes("shmpool_buff_hdr",61,"name")="shmpool_buff_hdr.shadow_file_header.filler_created"
	Set gtmtypes("shmpool_buff_hdr",61,"off")=496
	Set gtmtypes("shmpool_buff_hdr",61,"len")=52
	Set gtmtypes("shmpool_buff_hdr",61,"type")="char"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.filler_created")=61
	Set gtmtypes("shmpool_buff_hdr",62,"name")="shmpool_buff_hdr.shadow_file_header.createinprogress"
	Set gtmtypes("shmpool_buff_hdr",62,"off")=548
	Set gtmtypes("shmpool_buff_hdr",62,"len")=4
	Set gtmtypes("shmpool_buff_hdr",62,"type")="boolean_t"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.createinprogress")=62
	Set gtmtypes("shmpool_buff_hdr",63,"name")="shmpool_buff_hdr.shadow_file_header.creation_time4"
	Set gtmtypes("shmpool_buff_hdr",63,"off")=552
	Set gtmtypes("shmpool_buff_hdr",63,"len")=4
	Set gtmtypes("shmpool_buff_hdr",63,"type")="int"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.creation_time4")=63
	Set gtmtypes("shmpool_buff_hdr",64,"name")="shmpool_buff_hdr.shadow_file_header.reservedDBFlags"
	Set gtmtypes("shmpool_buff_hdr",64,"off")=556
	Set gtmtypes("shmpool_buff_hdr",64,"len")=4
	Set gtmtypes("shmpool_buff_hdr",64,"type")="unsigned-int"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.reservedDBFlags")=64
	Set gtmtypes("shmpool_buff_hdr",65,"name")="shmpool_buff_hdr.shadow_file_header.max_tn"
	Set gtmtypes("shmpool_buff_hdr",65,"off")=560
	Set gtmtypes("shmpool_buff_hdr",65,"len")=8
	Set gtmtypes("shmpool_buff_hdr",65,"type")="uint64_t"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.max_tn")=65
	Set gtmtypes("shmpool_buff_hdr",66,"name")="shmpool_buff_hdr.shadow_file_header.max_tn_warn"
	Set gtmtypes("shmpool_buff_hdr",66,"off")=568
	Set gtmtypes("shmpool_buff_hdr",66,"len")=8
	Set gtmtypes("shmpool_buff_hdr",66,"type")="uint64_t"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.max_tn_warn")=66
	Set gtmtypes("shmpool_buff_hdr",67,"name")="shmpool_buff_hdr.shadow_file_header.last_inc_backup"
	Set gtmtypes("shmpool_buff_hdr",67,"off")=576
	Set gtmtypes("shmpool_buff_hdr",67,"len")=8
	Set gtmtypes("shmpool_buff_hdr",67,"type")="uint64_t"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.last_inc_backup")=67
	Set gtmtypes("shmpool_buff_hdr",68,"name")="shmpool_buff_hdr.shadow_file_header.last_com_backup"
	Set gtmtypes("shmpool_buff_hdr",68,"off")=584
	Set gtmtypes("shmpool_buff_hdr",68,"len")=8
	Set gtmtypes("shmpool_buff_hdr",68,"type")="uint64_t"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.last_com_backup")=68
	Set gtmtypes("shmpool_buff_hdr",69,"name")="shmpool_buff_hdr.shadow_file_header.last_rec_backup"
	Set gtmtypes("shmpool_buff_hdr",69,"off")=592
	Set gtmtypes("shmpool_buff_hdr",69,"len")=8
	Set gtmtypes("shmpool_buff_hdr",69,"type")="uint64_t"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.last_rec_backup")=69
	Set gtmtypes("shmpool_buff_hdr",70,"name")="shmpool_buff_hdr.shadow_file_header.last_inc_bkup_last_blk"
	Set gtmtypes("shmpool_buff_hdr",70,"off")=600
	Set gtmtypes("shmpool_buff_hdr",70,"len")=4
	Set gtmtypes("shmpool_buff_hdr",70,"type")="int"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.last_inc_bkup_last_blk")=70
	Set gtmtypes("shmpool_buff_hdr",71,"name")="shmpool_buff_hdr.shadow_file_header.last_com_bkup_last_blk"
	Set gtmtypes("shmpool_buff_hdr",71,"off")=604
	Set gtmtypes("shmpool_buff_hdr",71,"len")=4
	Set gtmtypes("shmpool_buff_hdr",71,"type")="int"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.last_com_bkup_last_blk")=71
	Set gtmtypes("shmpool_buff_hdr",72,"name")="shmpool_buff_hdr.shadow_file_header.last_rec_bkup_last_blk"
	Set gtmtypes("shmpool_buff_hdr",72,"off")=608
	Set gtmtypes("shmpool_buff_hdr",72,"len")=4
	Set gtmtypes("shmpool_buff_hdr",72,"type")="int"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.last_rec_bkup_last_blk")=72
	Set gtmtypes("shmpool_buff_hdr",73,"name")="shmpool_buff_hdr.shadow_file_header.reorg_restart_block"
	Set gtmtypes("shmpool_buff_hdr",73,"off")=612
	Set gtmtypes("shmpool_buff_hdr",73,"len")=4
	Set gtmtypes("shmpool_buff_hdr",73,"type")="int"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.reorg_restart_block")=73
	Set gtmtypes("shmpool_buff_hdr",74,"name")="shmpool_buff_hdr.shadow_file_header.filler_256"
	Set gtmtypes("shmpool_buff_hdr",74,"off")=616
	Set gtmtypes("shmpool_buff_hdr",74,"len")=8
	Set gtmtypes("shmpool_buff_hdr",74,"type")="char"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.filler_256")=74
	Set gtmtypes("shmpool_buff_hdr",75,"name")="shmpool_buff_hdr.shadow_file_header.now_running"
	Set gtmtypes("shmpool_buff_hdr",75,"off")=624
	Set gtmtypes("shmpool_buff_hdr",75,"len")=36
	Set gtmtypes("shmpool_buff_hdr",75,"type")="char"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.now_running")=75
	Set gtmtypes("shmpool_buff_hdr",76,"name")="shmpool_buff_hdr.shadow_file_header.filler_owner_node"
	Set gtmtypes("shmpool_buff_hdr",76,"off")=660
	Set gtmtypes("shmpool_buff_hdr",76,"len")=4
	Set gtmtypes("shmpool_buff_hdr",76,"type")="unsigned-int"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.filler_owner_node")=76
	Set gtmtypes("shmpool_buff_hdr",77,"name")="shmpool_buff_hdr.shadow_file_header.image_count"
	Set gtmtypes("shmpool_buff_hdr",77,"off")=664
	Set gtmtypes("shmpool_buff_hdr",77,"len")=4
	Set gtmtypes("shmpool_buff_hdr",77,"type")="unsigned-int"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.image_count")=77
	Set gtmtypes("shmpool_buff_hdr",78,"name")="shmpool_buff_hdr.shadow_file_header.freeze"
	Set gtmtypes("shmpool_buff_hdr",78,"off")=668
	Set gtmtypes("shmpool_buff_hdr",78,"len")=4
	Set gtmtypes("shmpool_buff_hdr",78,"type")="unsigned-int"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.freeze")=78
	Set gtmtypes("shmpool_buff_hdr",79,"name")="shmpool_buff_hdr.shadow_file_header.kill_in_prog"
	Set gtmtypes("shmpool_buff_hdr",79,"off")=672
	Set gtmtypes("shmpool_buff_hdr",79,"len")=4
	Set gtmtypes("shmpool_buff_hdr",79,"type")="int"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.kill_in_prog")=79
	Set gtmtypes("shmpool_buff_hdr",80,"name")="shmpool_buff_hdr.shadow_file_header.abandoned_kills"
	Set gtmtypes("shmpool_buff_hdr",80,"off")=676
	Set gtmtypes("shmpool_buff_hdr",80,"len")=4
	Set gtmtypes("shmpool_buff_hdr",80,"type")="int"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.abandoned_kills")=80
	Set gtmtypes("shmpool_buff_hdr",81,"name")="shmpool_buff_hdr.shadow_file_header.unused_freeze_online_filler"
	Set gtmtypes("shmpool_buff_hdr",81,"off")=680
	Set gtmtypes("shmpool_buff_hdr",81,"len")=4
	Set gtmtypes("shmpool_buff_hdr",81,"type")="unsigned-int"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.unused_freeze_online_filler")=81
	Set gtmtypes("shmpool_buff_hdr",82,"name")="shmpool_buff_hdr.shadow_file_header.filler_320"
	Set gtmtypes("shmpool_buff_hdr",82,"off")=684
	Set gtmtypes("shmpool_buff_hdr",82,"len")=4
	Set gtmtypes("shmpool_buff_hdr",82,"type")="char"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.filler_320")=82
	Set gtmtypes("shmpool_buff_hdr",83,"name")="shmpool_buff_hdr.shadow_file_header.tn_upgrd_blks_0"
	Set gtmtypes("shmpool_buff_hdr",83,"off")=688
	Set gtmtypes("shmpool_buff_hdr",83,"len")=8
	Set gtmtypes("shmpool_buff_hdr",83,"type")="uint64_t"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.tn_upgrd_blks_0")=83
	Set gtmtypes("shmpool_buff_hdr",84,"name")="shmpool_buff_hdr.shadow_file_header.desired_db_format_tn"
	Set gtmtypes("shmpool_buff_hdr",84,"off")=696
	Set gtmtypes("shmpool_buff_hdr",84,"len")=8
	Set gtmtypes("shmpool_buff_hdr",84,"type")="uint64_t"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.desired_db_format_tn")=84
	Set gtmtypes("shmpool_buff_hdr",85,"name")="shmpool_buff_hdr.shadow_file_header.reorg_db_fmt_start_tn"
	Set gtmtypes("shmpool_buff_hdr",85,"off")=704
	Set gtmtypes("shmpool_buff_hdr",85,"len")=8
	Set gtmtypes("shmpool_buff_hdr",85,"type")="uint64_t"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.reorg_db_fmt_start_tn")=85
	Set gtmtypes("shmpool_buff_hdr",86,"name")="shmpool_buff_hdr.shadow_file_header.reorg_upgrd_dwngrd_restart_block"
	Set gtmtypes("shmpool_buff_hdr",86,"off")=712
	Set gtmtypes("shmpool_buff_hdr",86,"len")=4
	Set gtmtypes("shmpool_buff_hdr",86,"type")="int"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.reorg_upgrd_dwngrd_restart_block")=86
	Set gtmtypes("shmpool_buff_hdr",87,"name")="shmpool_buff_hdr.shadow_file_header.blks_to_upgrd"
	Set gtmtypes("shmpool_buff_hdr",87,"off")=716
	Set gtmtypes("shmpool_buff_hdr",87,"len")=4
	Set gtmtypes("shmpool_buff_hdr",87,"type")="int"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.blks_to_upgrd")=87
	Set gtmtypes("shmpool_buff_hdr",88,"name")="shmpool_buff_hdr.shadow_file_header.blks_to_upgrd_subzero_error"
	Set gtmtypes("shmpool_buff_hdr",88,"off")=720
	Set gtmtypes("shmpool_buff_hdr",88,"len")=4
	Set gtmtypes("shmpool_buff_hdr",88,"type")="int"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.blks_to_upgrd_subzero_error")=88
	Set gtmtypes("shmpool_buff_hdr",89,"name")="shmpool_buff_hdr.shadow_file_header.desired_db_format"
	Set gtmtypes("shmpool_buff_hdr",89,"off")=724
	Set gtmtypes("shmpool_buff_hdr",89,"len")=4
	Set gtmtypes("shmpool_buff_hdr",89,"type")="int"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.desired_db_format")=89
	Set gtmtypes("shmpool_buff_hdr",90,"name")="shmpool_buff_hdr.shadow_file_header.fully_upgraded"
	Set gtmtypes("shmpool_buff_hdr",90,"off")=728
	Set gtmtypes("shmpool_buff_hdr",90,"len")=4
	Set gtmtypes("shmpool_buff_hdr",90,"type")="boolean_t"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.fully_upgraded")=90
	Set gtmtypes("shmpool_buff_hdr",91,"name")="shmpool_buff_hdr.shadow_file_header.db_got_to_v5_once"
	Set gtmtypes("shmpool_buff_hdr",91,"off")=732
	Set gtmtypes("shmpool_buff_hdr",91,"len")=4
	Set gtmtypes("shmpool_buff_hdr",91,"type")="boolean_t"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.db_got_to_v5_once")=91
	Set gtmtypes("shmpool_buff_hdr",92,"name")="shmpool_buff_hdr.shadow_file_header.opened_by_gtmv53"
	Set gtmtypes("shmpool_buff_hdr",92,"off")=736
	Set gtmtypes("shmpool_buff_hdr",92,"len")=4
	Set gtmtypes("shmpool_buff_hdr",92,"type")="boolean_t"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.opened_by_gtmv53")=92
	Set gtmtypes("shmpool_buff_hdr",93,"name")="shmpool_buff_hdr.shadow_file_header.filler_384"
	Set gtmtypes("shmpool_buff_hdr",93,"off")=740
	Set gtmtypes("shmpool_buff_hdr",93,"len")=12
	Set gtmtypes("shmpool_buff_hdr",93,"type")="char"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.filler_384")=93
	Set gtmtypes("shmpool_buff_hdr",94,"name")="shmpool_buff_hdr.shadow_file_header.trans_hist"
	Set gtmtypes("shmpool_buff_hdr",94,"off")=752
	Set gtmtypes("shmpool_buff_hdr",94,"len")=56
	Set gtmtypes("shmpool_buff_hdr",94,"type")="th_index"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.trans_hist")=94
	Set gtmtypes("shmpool_buff_hdr",95,"name")="shmpool_buff_hdr.shadow_file_header.trans_hist.curr_tn"
	Set gtmtypes("shmpool_buff_hdr",95,"off")=752
	Set gtmtypes("shmpool_buff_hdr",95,"len")=8
	Set gtmtypes("shmpool_buff_hdr",95,"type")="uint64_t"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.trans_hist.curr_tn")=95
	Set gtmtypes("shmpool_buff_hdr",96,"name")="shmpool_buff_hdr.shadow_file_header.trans_hist.early_tn"
	Set gtmtypes("shmpool_buff_hdr",96,"off")=760
	Set gtmtypes("shmpool_buff_hdr",96,"len")=8
	Set gtmtypes("shmpool_buff_hdr",96,"type")="uint64_t"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.trans_hist.early_tn")=96
	Set gtmtypes("shmpool_buff_hdr",97,"name")="shmpool_buff_hdr.shadow_file_header.trans_hist.last_mm_sync"
	Set gtmtypes("shmpool_buff_hdr",97,"off")=768
	Set gtmtypes("shmpool_buff_hdr",97,"len")=8
	Set gtmtypes("shmpool_buff_hdr",97,"type")="uint64_t"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.trans_hist.last_mm_sync")=97
	Set gtmtypes("shmpool_buff_hdr",98,"name")="shmpool_buff_hdr.shadow_file_header.trans_hist.filler_8byte"
	Set gtmtypes("shmpool_buff_hdr",98,"off")=776
	Set gtmtypes("shmpool_buff_hdr",98,"len")=8
	Set gtmtypes("shmpool_buff_hdr",98,"type")="char"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.trans_hist.filler_8byte")=98
	Set gtmtypes("shmpool_buff_hdr",99,"name")="shmpool_buff_hdr.shadow_file_header.trans_hist.mm_tn"
	Set gtmtypes("shmpool_buff_hdr",99,"off")=784
	Set gtmtypes("shmpool_buff_hdr",99,"len")=8
	Set gtmtypes("shmpool_buff_hdr",99,"type")="uint64_t"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.trans_hist.mm_tn")=99
	Set gtmtypes("shmpool_buff_hdr",100,"name")="shmpool_buff_hdr.shadow_file_header.trans_hist.lock_sequence"
	Set gtmtypes("shmpool_buff_hdr",100,"off")=792
	Set gtmtypes("shmpool_buff_hdr",100,"len")=4
	Set gtmtypes("shmpool_buff_hdr",100,"type")="unsigned-int"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.trans_hist.lock_sequence")=100
	Set gtmtypes("shmpool_buff_hdr",101,"name")="shmpool_buff_hdr.shadow_file_header.trans_hist.ccp_jnl_filesize"
	Set gtmtypes("shmpool_buff_hdr",101,"off")=796
	Set gtmtypes("shmpool_buff_hdr",101,"len")=4
	Set gtmtypes("shmpool_buff_hdr",101,"type")="unsigned-int"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.trans_hist.ccp_jnl_filesize")=101
	Set gtmtypes("shmpool_buff_hdr",102,"name")="shmpool_buff_hdr.shadow_file_header.trans_hist.total_blks"
	Set gtmtypes("shmpool_buff_hdr",102,"off")=800
	Set gtmtypes("shmpool_buff_hdr",102,"len")=4
	Set gtmtypes("shmpool_buff_hdr",102,"type")="unsigned-int"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.trans_hist.total_blks")=102
	Set gtmtypes("shmpool_buff_hdr",103,"name")="shmpool_buff_hdr.shadow_file_header.trans_hist.free_blocks"
	Set gtmtypes("shmpool_buff_hdr",103,"off")=804
	Set gtmtypes("shmpool_buff_hdr",103,"len")=4
	Set gtmtypes("shmpool_buff_hdr",103,"type")="unsigned-int"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.trans_hist.free_blocks")=103
	Set gtmtypes("shmpool_buff_hdr",104,"name")="shmpool_buff_hdr.shadow_file_header.write_fullblk"
	Set gtmtypes("shmpool_buff_hdr",104,"off")=808
	Set gtmtypes("shmpool_buff_hdr",104,"len")=4
	Set gtmtypes("shmpool_buff_hdr",104,"type")="int"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.write_fullblk")=104
	Set gtmtypes("shmpool_buff_hdr",105,"name")="shmpool_buff_hdr.shadow_file_header.filler"
	Set gtmtypes("shmpool_buff_hdr",105,"off")=812
	Set gtmtypes("shmpool_buff_hdr",105,"len")=4
	Set gtmtypes("shmpool_buff_hdr",105,"type")="char"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.filler")=105
	Set gtmtypes("shmpool_buff_hdr",106,"name")="shmpool_buff_hdr.shadow_file_header.flush_time"
	Set gtmtypes("shmpool_buff_hdr",106,"off")=816
	Set gtmtypes("shmpool_buff_hdr",106,"len")=8
	Set gtmtypes("shmpool_buff_hdr",106,"type")="int"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.flush_time")=106
	Set gtmtypes("shmpool_buff_hdr",106,"dim")=2
	Set gtmtypes("shmpool_buff_hdr",107,"name")="shmpool_buff_hdr.shadow_file_header.flush_trigger"
	Set gtmtypes("shmpool_buff_hdr",107,"off")=824
	Set gtmtypes("shmpool_buff_hdr",107,"len")=4
	Set gtmtypes("shmpool_buff_hdr",107,"type")="int"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.flush_trigger")=107
	Set gtmtypes("shmpool_buff_hdr",108,"name")="shmpool_buff_hdr.shadow_file_header.n_wrt_per_flu"
	Set gtmtypes("shmpool_buff_hdr",108,"off")=828
	Set gtmtypes("shmpool_buff_hdr",108,"len")=4
	Set gtmtypes("shmpool_buff_hdr",108,"type")="int"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.n_wrt_per_flu")=108
	Set gtmtypes("shmpool_buff_hdr",109,"name")="shmpool_buff_hdr.shadow_file_header.wait_disk_space"
	Set gtmtypes("shmpool_buff_hdr",109,"off")=832
	Set gtmtypes("shmpool_buff_hdr",109,"len")=4
	Set gtmtypes("shmpool_buff_hdr",109,"type")="int"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.wait_disk_space")=109
	Set gtmtypes("shmpool_buff_hdr",110,"name")="shmpool_buff_hdr.shadow_file_header.defer_time"
	Set gtmtypes("shmpool_buff_hdr",110,"off")=836
	Set gtmtypes("shmpool_buff_hdr",110,"len")=4
	Set gtmtypes("shmpool_buff_hdr",110,"type")="int"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.defer_time")=110
	Set gtmtypes("shmpool_buff_hdr",111,"name")="shmpool_buff_hdr.shadow_file_header.flush_trigger_top"
	Set gtmtypes("shmpool_buff_hdr",111,"off")=840
	Set gtmtypes("shmpool_buff_hdr",111,"len")=4
	Set gtmtypes("shmpool_buff_hdr",111,"type")="int"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.flush_trigger_top")=111
	Set gtmtypes("shmpool_buff_hdr",112,"name")="shmpool_buff_hdr.shadow_file_header.mumps_can_bypass"
	Set gtmtypes("shmpool_buff_hdr",112,"off")=844
	Set gtmtypes("shmpool_buff_hdr",112,"len")=4
	Set gtmtypes("shmpool_buff_hdr",112,"type")="boolean_t"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.mumps_can_bypass")=112
	Set gtmtypes("shmpool_buff_hdr",113,"name")="shmpool_buff_hdr.shadow_file_header.epoch_taper"
	Set gtmtypes("shmpool_buff_hdr",113,"off")=848
	Set gtmtypes("shmpool_buff_hdr",113,"len")=4
	Set gtmtypes("shmpool_buff_hdr",113,"type")="boolean_t"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.epoch_taper")=113
	Set gtmtypes("shmpool_buff_hdr",114,"name")="shmpool_buff_hdr.shadow_file_header.epoch_taper_time_pct"
	Set gtmtypes("shmpool_buff_hdr",114,"off")=852
	Set gtmtypes("shmpool_buff_hdr",114,"len")=4
	Set gtmtypes("shmpool_buff_hdr",114,"type")="unsigned-int"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.epoch_taper_time_pct")=114
	Set gtmtypes("shmpool_buff_hdr",115,"name")="shmpool_buff_hdr.shadow_file_header.epoch_taper_jnl_pct"
	Set gtmtypes("shmpool_buff_hdr",115,"off")=856
	Set gtmtypes("shmpool_buff_hdr",115,"len")=4
	Set gtmtypes("shmpool_buff_hdr",115,"type")="unsigned-int"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.epoch_taper_jnl_pct")=115
	Set gtmtypes("shmpool_buff_hdr",116,"name")="shmpool_buff_hdr.shadow_file_header.asyncio"
	Set gtmtypes("shmpool_buff_hdr",116,"off")=860
	Set gtmtypes("shmpool_buff_hdr",116,"len")=4
	Set gtmtypes("shmpool_buff_hdr",116,"type")="boolean_t"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.asyncio")=116
	Set gtmtypes("shmpool_buff_hdr",117,"name")="shmpool_buff_hdr.shadow_file_header.reserved_for_upd"
	Set gtmtypes("shmpool_buff_hdr",117,"off")=864
	Set gtmtypes("shmpool_buff_hdr",117,"len")=4
	Set gtmtypes("shmpool_buff_hdr",117,"type")="unsigned-int"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.reserved_for_upd")=117
	Set gtmtypes("shmpool_buff_hdr",118,"name")="shmpool_buff_hdr.shadow_file_header.avg_blks_per_100gbl"
	Set gtmtypes("shmpool_buff_hdr",118,"off")=868
	Set gtmtypes("shmpool_buff_hdr",118,"len")=4
	Set gtmtypes("shmpool_buff_hdr",118,"type")="unsigned-int"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.avg_blks_per_100gbl")=118
	Set gtmtypes("shmpool_buff_hdr",119,"name")="shmpool_buff_hdr.shadow_file_header.pre_read_trigger_factor"
	Set gtmtypes("shmpool_buff_hdr",119,"off")=872
	Set gtmtypes("shmpool_buff_hdr",119,"len")=4
	Set gtmtypes("shmpool_buff_hdr",119,"type")="unsigned-int"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.pre_read_trigger_factor")=119
	Set gtmtypes("shmpool_buff_hdr",120,"name")="shmpool_buff_hdr.shadow_file_header.writer_trigger_factor"
	Set gtmtypes("shmpool_buff_hdr",120,"off")=876
	Set gtmtypes("shmpool_buff_hdr",120,"len")=4
	Set gtmtypes("shmpool_buff_hdr",120,"type")="unsigned-int"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.writer_trigger_factor")=120
	Set gtmtypes("shmpool_buff_hdr",121,"name")="shmpool_buff_hdr.shadow_file_header.semid"
	Set gtmtypes("shmpool_buff_hdr",121,"off")=880
	Set gtmtypes("shmpool_buff_hdr",121,"len")=4
	Set gtmtypes("shmpool_buff_hdr",121,"type")="int"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.semid")=121
	Set gtmtypes("shmpool_buff_hdr",122,"name")="shmpool_buff_hdr.shadow_file_header.shmid"
	Set gtmtypes("shmpool_buff_hdr",122,"off")=884
	Set gtmtypes("shmpool_buff_hdr",122,"len")=4
	Set gtmtypes("shmpool_buff_hdr",122,"type")="int"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.shmid")=122
	Set gtmtypes("shmpool_buff_hdr",123,"name")="shmpool_buff_hdr.shadow_file_header.gt_sem_ctime"
	Set gtmtypes("shmpool_buff_hdr",123,"off")=888
	Set gtmtypes("shmpool_buff_hdr",123,"len")=8
	Set gtmtypes("shmpool_buff_hdr",123,"type")="gtm_time8"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.gt_sem_ctime")=123
	Set gtmtypes("shmpool_buff_hdr",124,"name")="shmpool_buff_hdr.shadow_file_header.gt_sem_ctime.ctime"
	Set gtmtypes("shmpool_buff_hdr",124,"off")=888
	Set gtmtypes("shmpool_buff_hdr",124,"len")=8
	Set gtmtypes("shmpool_buff_hdr",124,"type")="time_t"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.gt_sem_ctime.ctime")=124
	Set gtmtypes("shmpool_buff_hdr",125,"name")="shmpool_buff_hdr.shadow_file_header.gt_sem_ctime.filler"
	Set gtmtypes("shmpool_buff_hdr",125,"off")=888
	Set gtmtypes("shmpool_buff_hdr",125,"len")=8
	Set gtmtypes("shmpool_buff_hdr",125,"type")="int"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.gt_sem_ctime.filler")=125
	Set gtmtypes("shmpool_buff_hdr",125,"dim")=2
	Set gtmtypes("shmpool_buff_hdr",126,"name")="shmpool_buff_hdr.shadow_file_header.gt_shm_ctime"
	Set gtmtypes("shmpool_buff_hdr",126,"off")=896
	Set gtmtypes("shmpool_buff_hdr",126,"len")=8
	Set gtmtypes("shmpool_buff_hdr",126,"type")="gtm_time8"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.gt_shm_ctime")=126
	Set gtmtypes("shmpool_buff_hdr",127,"name")="shmpool_buff_hdr.shadow_file_header.gt_shm_ctime.ctime"
	Set gtmtypes("shmpool_buff_hdr",127,"off")=896
	Set gtmtypes("shmpool_buff_hdr",127,"len")=8
	Set gtmtypes("shmpool_buff_hdr",127,"type")="time_t"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.gt_shm_ctime.ctime")=127
	Set gtmtypes("shmpool_buff_hdr",128,"name")="shmpool_buff_hdr.shadow_file_header.gt_shm_ctime.filler"
	Set gtmtypes("shmpool_buff_hdr",128,"off")=896
	Set gtmtypes("shmpool_buff_hdr",128,"len")=8
	Set gtmtypes("shmpool_buff_hdr",128,"type")="int"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.gt_shm_ctime.filler")=128
	Set gtmtypes("shmpool_buff_hdr",128,"dim")=2
	Set gtmtypes("shmpool_buff_hdr",129,"name")="shmpool_buff_hdr.shadow_file_header.filler_unixonly"
	Set gtmtypes("shmpool_buff_hdr",129,"off")=904
	Set gtmtypes("shmpool_buff_hdr",129,"len")=40
	Set gtmtypes("shmpool_buff_hdr",129,"type")="char"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.filler_unixonly")=129
	Set gtmtypes("shmpool_buff_hdr",130,"name")="shmpool_buff_hdr.shadow_file_header.filler_n_retries"
	Set gtmtypes("shmpool_buff_hdr",130,"off")=944
	Set gtmtypes("shmpool_buff_hdr",130,"len")=20
	Set gtmtypes("shmpool_buff_hdr",130,"type")="int"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.filler_n_retries")=130
	Set gtmtypes("shmpool_buff_hdr",130,"dim")=5
	Set gtmtypes("shmpool_buff_hdr",131,"name")="shmpool_buff_hdr.shadow_file_header.filler_n_puts"
	Set gtmtypes("shmpool_buff_hdr",131,"off")=964
	Set gtmtypes("shmpool_buff_hdr",131,"len")=4
	Set gtmtypes("shmpool_buff_hdr",131,"type")="int"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.filler_n_puts")=131
	Set gtmtypes("shmpool_buff_hdr",132,"name")="shmpool_buff_hdr.shadow_file_header.filler_n_kills"
	Set gtmtypes("shmpool_buff_hdr",132,"off")=968
	Set gtmtypes("shmpool_buff_hdr",132,"len")=4
	Set gtmtypes("shmpool_buff_hdr",132,"type")="int"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.filler_n_kills")=132
	Set gtmtypes("shmpool_buff_hdr",133,"name")="shmpool_buff_hdr.shadow_file_header.filler_n_queries"
	Set gtmtypes("shmpool_buff_hdr",133,"off")=972
	Set gtmtypes("shmpool_buff_hdr",133,"len")=4
	Set gtmtypes("shmpool_buff_hdr",133,"type")="int"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.filler_n_queries")=133
	Set gtmtypes("shmpool_buff_hdr",134,"name")="shmpool_buff_hdr.shadow_file_header.filler_n_gets"
	Set gtmtypes("shmpool_buff_hdr",134,"off")=976
	Set gtmtypes("shmpool_buff_hdr",134,"len")=4
	Set gtmtypes("shmpool_buff_hdr",134,"type")="int"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.filler_n_gets")=134
	Set gtmtypes("shmpool_buff_hdr",135,"name")="shmpool_buff_hdr.shadow_file_header.filler_n_order"
	Set gtmtypes("shmpool_buff_hdr",135,"off")=980
	Set gtmtypes("shmpool_buff_hdr",135,"len")=4
	Set gtmtypes("shmpool_buff_hdr",135,"type")="int"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.filler_n_order")=135
	Set gtmtypes("shmpool_buff_hdr",136,"name")="shmpool_buff_hdr.shadow_file_header.filler_n_zprevs"
	Set gtmtypes("shmpool_buff_hdr",136,"off")=984
	Set gtmtypes("shmpool_buff_hdr",136,"len")=4
	Set gtmtypes("shmpool_buff_hdr",136,"type")="int"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.filler_n_zprevs")=136
	Set gtmtypes("shmpool_buff_hdr",137,"name")="shmpool_buff_hdr.shadow_file_header.filler_n_data"
	Set gtmtypes("shmpool_buff_hdr",137,"off")=988
	Set gtmtypes("shmpool_buff_hdr",137,"len")=4
	Set gtmtypes("shmpool_buff_hdr",137,"type")="int"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.filler_n_data")=137
	Set gtmtypes("shmpool_buff_hdr",138,"name")="shmpool_buff_hdr.shadow_file_header.filler_n_puts_duplicate"
	Set gtmtypes("shmpool_buff_hdr",138,"off")=992
	Set gtmtypes("shmpool_buff_hdr",138,"len")=4
	Set gtmtypes("shmpool_buff_hdr",138,"type")="unsigned-int"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.filler_n_puts_duplicate")=138
	Set gtmtypes("shmpool_buff_hdr",139,"name")="shmpool_buff_hdr.shadow_file_header.filler_n_tp_updates"
	Set gtmtypes("shmpool_buff_hdr",139,"off")=996
	Set gtmtypes("shmpool_buff_hdr",139,"len")=4
	Set gtmtypes("shmpool_buff_hdr",139,"type")="unsigned-int"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.filler_n_tp_updates")=139
	Set gtmtypes("shmpool_buff_hdr",140,"name")="shmpool_buff_hdr.shadow_file_header.filler_n_tp_updates_duplicate"
	Set gtmtypes("shmpool_buff_hdr",140,"off")=1000
	Set gtmtypes("shmpool_buff_hdr",140,"len")=4
	Set gtmtypes("shmpool_buff_hdr",140,"type")="unsigned-int"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.filler_n_tp_updates_duplicate")=140
	Set gtmtypes("shmpool_buff_hdr",141,"name")="shmpool_buff_hdr.shadow_file_header.filler_accounting_64_align"
	Set gtmtypes("shmpool_buff_hdr",141,"off")=1004
	Set gtmtypes("shmpool_buff_hdr",141,"len")=4
	Set gtmtypes("shmpool_buff_hdr",141,"type")="char"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.filler_accounting_64_align")=141
	Set gtmtypes("shmpool_buff_hdr",142,"name")="shmpool_buff_hdr.shadow_file_header.staleness"
	Set gtmtypes("shmpool_buff_hdr",142,"off")=1008
	Set gtmtypes("shmpool_buff_hdr",142,"len")=8
	Set gtmtypes("shmpool_buff_hdr",142,"type")="int"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.staleness")=142
	Set gtmtypes("shmpool_buff_hdr",142,"dim")=2
	Set gtmtypes("shmpool_buff_hdr",143,"name")="shmpool_buff_hdr.shadow_file_header.ccp_tick_interval"
	Set gtmtypes("shmpool_buff_hdr",143,"off")=1016
	Set gtmtypes("shmpool_buff_hdr",143,"len")=8
	Set gtmtypes("shmpool_buff_hdr",143,"type")="int"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.ccp_tick_interval")=143
	Set gtmtypes("shmpool_buff_hdr",143,"dim")=2
	Set gtmtypes("shmpool_buff_hdr",144,"name")="shmpool_buff_hdr.shadow_file_header.ccp_quantum_interval"
	Set gtmtypes("shmpool_buff_hdr",144,"off")=1024
	Set gtmtypes("shmpool_buff_hdr",144,"len")=8
	Set gtmtypes("shmpool_buff_hdr",144,"type")="int"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.ccp_quantum_interval")=144
	Set gtmtypes("shmpool_buff_hdr",144,"dim")=2
	Set gtmtypes("shmpool_buff_hdr",145,"name")="shmpool_buff_hdr.shadow_file_header.ccp_response_interval"
	Set gtmtypes("shmpool_buff_hdr",145,"off")=1032
	Set gtmtypes("shmpool_buff_hdr",145,"len")=8
	Set gtmtypes("shmpool_buff_hdr",145,"type")="int"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.ccp_response_interval")=145
	Set gtmtypes("shmpool_buff_hdr",145,"dim")=2
	Set gtmtypes("shmpool_buff_hdr",146,"name")="shmpool_buff_hdr.shadow_file_header.ccp_jnl_before"
	Set gtmtypes("shmpool_buff_hdr",146,"off")=1040
	Set gtmtypes("shmpool_buff_hdr",146,"len")=4
	Set gtmtypes("shmpool_buff_hdr",146,"type")="boolean_t"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.ccp_jnl_before")=146
	Set gtmtypes("shmpool_buff_hdr",147,"name")="shmpool_buff_hdr.shadow_file_header.clustered"
	Set gtmtypes("shmpool_buff_hdr",147,"off")=1044
	Set gtmtypes("shmpool_buff_hdr",147,"len")=4
	Set gtmtypes("shmpool_buff_hdr",147,"type")="boolean_t"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.clustered")=147
	Set gtmtypes("shmpool_buff_hdr",148,"name")="shmpool_buff_hdr.shadow_file_header.unbacked_cache"
	Set gtmtypes("shmpool_buff_hdr",148,"off")=1048
	Set gtmtypes("shmpool_buff_hdr",148,"len")=4
	Set gtmtypes("shmpool_buff_hdr",148,"type")="boolean_t"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.unbacked_cache")=148
	Set gtmtypes("shmpool_buff_hdr",149,"name")="shmpool_buff_hdr.shadow_file_header.rc_srv_cnt"
	Set gtmtypes("shmpool_buff_hdr",149,"off")=1052
	Set gtmtypes("shmpool_buff_hdr",149,"len")=4
	Set gtmtypes("shmpool_buff_hdr",149,"type")="int"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.rc_srv_cnt")=149
	Set gtmtypes("shmpool_buff_hdr",150,"name")="shmpool_buff_hdr.shadow_file_header.dsid"
	Set gtmtypes("shmpool_buff_hdr",150,"off")=1056
	Set gtmtypes("shmpool_buff_hdr",150,"len")=4
	Set gtmtypes("shmpool_buff_hdr",150,"type")="int"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.dsid")=150
	Set gtmtypes("shmpool_buff_hdr",151,"name")="shmpool_buff_hdr.shadow_file_header.rc_node"
	Set gtmtypes("shmpool_buff_hdr",151,"off")=1060
	Set gtmtypes("shmpool_buff_hdr",151,"len")=4
	Set gtmtypes("shmpool_buff_hdr",151,"type")="int"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.rc_node")=151
	Set gtmtypes("shmpool_buff_hdr",152,"name")="shmpool_buff_hdr.shadow_file_header.filler_ccp_rc"
	Set gtmtypes("shmpool_buff_hdr",152,"off")=1064
	Set gtmtypes("shmpool_buff_hdr",152,"len")=8
	Set gtmtypes("shmpool_buff_hdr",152,"type")="char"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.filler_ccp_rc")=152
	Set gtmtypes("shmpool_buff_hdr",153,"name")="shmpool_buff_hdr.shadow_file_header.reg_seqno"
	Set gtmtypes("shmpool_buff_hdr",153,"off")=1072
	Set gtmtypes("shmpool_buff_hdr",153,"len")=8
	Set gtmtypes("shmpool_buff_hdr",153,"type")="uint64_t"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.reg_seqno")=153
	Set gtmtypes("shmpool_buff_hdr",154,"name")="shmpool_buff_hdr.shadow_file_header.pre_multisite_resync_seqno"
	Set gtmtypes("shmpool_buff_hdr",154,"off")=1080
	Set gtmtypes("shmpool_buff_hdr",154,"len")=8
	Set gtmtypes("shmpool_buff_hdr",154,"type")="uint64_t"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.pre_multisite_resync_seqno")=154
	Set gtmtypes("shmpool_buff_hdr",155,"name")="shmpool_buff_hdr.shadow_file_header.zqgblmod_tn"
	Set gtmtypes("shmpool_buff_hdr",155,"off")=1088
	Set gtmtypes("shmpool_buff_hdr",155,"len")=8
	Set gtmtypes("shmpool_buff_hdr",155,"type")="uint64_t"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.zqgblmod_tn")=155
	Set gtmtypes("shmpool_buff_hdr",156,"name")="shmpool_buff_hdr.shadow_file_header.zqgblmod_seqno"
	Set gtmtypes("shmpool_buff_hdr",156,"off")=1096
	Set gtmtypes("shmpool_buff_hdr",156,"len")=8
	Set gtmtypes("shmpool_buff_hdr",156,"type")="uint64_t"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.zqgblmod_seqno")=156
	Set gtmtypes("shmpool_buff_hdr",157,"name")="shmpool_buff_hdr.shadow_file_header.repl_state"
	Set gtmtypes("shmpool_buff_hdr",157,"off")=1104
	Set gtmtypes("shmpool_buff_hdr",157,"len")=4
	Set gtmtypes("shmpool_buff_hdr",157,"type")="int"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.repl_state")=157
	Set gtmtypes("shmpool_buff_hdr",158,"name")="shmpool_buff_hdr.shadow_file_header.multi_site_open"
	Set gtmtypes("shmpool_buff_hdr",158,"off")=1108
	Set gtmtypes("shmpool_buff_hdr",158,"len")=4
	Set gtmtypes("shmpool_buff_hdr",158,"type")="boolean_t"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.multi_site_open")=158
	Set gtmtypes("shmpool_buff_hdr",159,"name")="shmpool_buff_hdr.shadow_file_header.filler_seqno"
	Set gtmtypes("shmpool_buff_hdr",159,"off")=1112
	Set gtmtypes("shmpool_buff_hdr",159,"len")=8
	Set gtmtypes("shmpool_buff_hdr",159,"type")="uint64_t"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.filler_seqno")=159
	Set gtmtypes("shmpool_buff_hdr",160,"name")="shmpool_buff_hdr.shadow_file_header.filler_repl"
	Set gtmtypes("shmpool_buff_hdr",160,"off")=1120
	Set gtmtypes("shmpool_buff_hdr",160,"len")=16
	Set gtmtypes("shmpool_buff_hdr",160,"type")="char"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.filler_repl")=160
	Set gtmtypes("shmpool_buff_hdr",161,"name")="shmpool_buff_hdr.shadow_file_header.filler_n_tp_retries"
	Set gtmtypes("shmpool_buff_hdr",161,"off")=1136
	Set gtmtypes("shmpool_buff_hdr",161,"len")=48
	Set gtmtypes("shmpool_buff_hdr",161,"type")="int"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.filler_n_tp_retries")=161
	Set gtmtypes("shmpool_buff_hdr",161,"dim")=12
	Set gtmtypes("shmpool_buff_hdr",162,"name")="shmpool_buff_hdr.shadow_file_header.filler_n_tp_retries_conflicts"
	Set gtmtypes("shmpool_buff_hdr",162,"off")=1184
	Set gtmtypes("shmpool_buff_hdr",162,"len")=48
	Set gtmtypes("shmpool_buff_hdr",162,"type")="int"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.filler_n_tp_retries_conflicts")=162
	Set gtmtypes("shmpool_buff_hdr",162,"dim")=12
	Set gtmtypes("shmpool_buff_hdr",163,"name")="shmpool_buff_hdr.shadow_file_header.tp_cdb_sc_blkmod"
	Set gtmtypes("shmpool_buff_hdr",163,"off")=1232
	Set gtmtypes("shmpool_buff_hdr",163,"len")=32
	Set gtmtypes("shmpool_buff_hdr",163,"type")="int"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.tp_cdb_sc_blkmod")=163
	Set gtmtypes("shmpool_buff_hdr",163,"dim")=8
	Set gtmtypes("shmpool_buff_hdr",164,"name")="shmpool_buff_hdr.shadow_file_header.jnl_alq"
	Set gtmtypes("shmpool_buff_hdr",164,"off")=1264
	Set gtmtypes("shmpool_buff_hdr",164,"len")=4
	Set gtmtypes("shmpool_buff_hdr",164,"type")="unsigned-int"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.jnl_alq")=164
	Set gtmtypes("shmpool_buff_hdr",165,"name")="shmpool_buff_hdr.shadow_file_header.jnl_deq"
	Set gtmtypes("shmpool_buff_hdr",165,"off")=1268
	Set gtmtypes("shmpool_buff_hdr",165,"len")=4
	Set gtmtypes("shmpool_buff_hdr",165,"type")="unsigned-int"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.jnl_deq")=165
	Set gtmtypes("shmpool_buff_hdr",166,"name")="shmpool_buff_hdr.shadow_file_header.jnl_buffer_size"
	Set gtmtypes("shmpool_buff_hdr",166,"off")=1272
	Set gtmtypes("shmpool_buff_hdr",166,"len")=4
	Set gtmtypes("shmpool_buff_hdr",166,"type")="int"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.jnl_buffer_size")=166
	Set gtmtypes("shmpool_buff_hdr",167,"name")="shmpool_buff_hdr.shadow_file_header.jnl_before_image"
	Set gtmtypes("shmpool_buff_hdr",167,"off")=1276
	Set gtmtypes("shmpool_buff_hdr",167,"len")=4
	Set gtmtypes("shmpool_buff_hdr",167,"type")="boolean_t"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.jnl_before_image")=167
	Set gtmtypes("shmpool_buff_hdr",168,"name")="shmpool_buff_hdr.shadow_file_header.jnl_state"
	Set gtmtypes("shmpool_buff_hdr",168,"off")=1280
	Set gtmtypes("shmpool_buff_hdr",168,"len")=4
	Set gtmtypes("shmpool_buff_hdr",168,"type")="int"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.jnl_state")=168
	Set gtmtypes("shmpool_buff_hdr",169,"name")="shmpool_buff_hdr.shadow_file_header.jnl_file_len"
	Set gtmtypes("shmpool_buff_hdr",169,"off")=1284
	Set gtmtypes("shmpool_buff_hdr",169,"len")=4
	Set gtmtypes("shmpool_buff_hdr",169,"type")="unsigned-int"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.jnl_file_len")=169
	Set gtmtypes("shmpool_buff_hdr",170,"name")="shmpool_buff_hdr.shadow_file_header.autoswitchlimit"
	Set gtmtypes("shmpool_buff_hdr",170,"off")=1288
	Set gtmtypes("shmpool_buff_hdr",170,"len")=4
	Set gtmtypes("shmpool_buff_hdr",170,"type")="unsigned-int"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.autoswitchlimit")=170
	Set gtmtypes("shmpool_buff_hdr",171,"name")="shmpool_buff_hdr.shadow_file_header.epoch_interval"
	Set gtmtypes("shmpool_buff_hdr",171,"off")=1292
	Set gtmtypes("shmpool_buff_hdr",171,"len")=4
	Set gtmtypes("shmpool_buff_hdr",171,"type")="int"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.epoch_interval")=171
	Set gtmtypes("shmpool_buff_hdr",172,"name")="shmpool_buff_hdr.shadow_file_header.alignsize"
	Set gtmtypes("shmpool_buff_hdr",172,"off")=1296
	Set gtmtypes("shmpool_buff_hdr",172,"len")=4
	Set gtmtypes("shmpool_buff_hdr",172,"type")="unsigned-int"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.alignsize")=172
	Set gtmtypes("shmpool_buff_hdr",173,"name")="shmpool_buff_hdr.shadow_file_header.jnl_sync_io"
	Set gtmtypes("shmpool_buff_hdr",173,"off")=1300
	Set gtmtypes("shmpool_buff_hdr",173,"len")=4
	Set gtmtypes("shmpool_buff_hdr",173,"type")="int"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.jnl_sync_io")=173
	Set gtmtypes("shmpool_buff_hdr",174,"name")="shmpool_buff_hdr.shadow_file_header.yield_lmt"
	Set gtmtypes("shmpool_buff_hdr",174,"off")=1304
	Set gtmtypes("shmpool_buff_hdr",174,"len")=4
	Set gtmtypes("shmpool_buff_hdr",174,"type")="int"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.yield_lmt")=174
	Set gtmtypes("shmpool_buff_hdr",175,"name")="shmpool_buff_hdr.shadow_file_header.turn_around_point"
	Set gtmtypes("shmpool_buff_hdr",175,"off")=1308
	Set gtmtypes("shmpool_buff_hdr",175,"len")=4
	Set gtmtypes("shmpool_buff_hdr",175,"type")="boolean_t"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.turn_around_point")=175
	Set gtmtypes("shmpool_buff_hdr",176,"name")="shmpool_buff_hdr.shadow_file_header.jnl_eovtn"
	Set gtmtypes("shmpool_buff_hdr",176,"off")=1312
	Set gtmtypes("shmpool_buff_hdr",176,"len")=8
	Set gtmtypes("shmpool_buff_hdr",176,"type")="uint64_t"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.jnl_eovtn")=176
	Set gtmtypes("shmpool_buff_hdr",177,"name")="shmpool_buff_hdr.shadow_file_header.filler_jnl"
	Set gtmtypes("shmpool_buff_hdr",177,"off")=1320
	Set gtmtypes("shmpool_buff_hdr",177,"len")=8
	Set gtmtypes("shmpool_buff_hdr",177,"type")="char"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.filler_jnl")=177
	Set gtmtypes("shmpool_buff_hdr",178,"name")="shmpool_buff_hdr.shadow_file_header.intrpt_recov_resync_seqno"
	Set gtmtypes("shmpool_buff_hdr",178,"off")=1328
	Set gtmtypes("shmpool_buff_hdr",178,"len")=8
	Set gtmtypes("shmpool_buff_hdr",178,"type")="uint64_t"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.intrpt_recov_resync_seqno")=178
	Set gtmtypes("shmpool_buff_hdr",179,"name")="shmpool_buff_hdr.shadow_file_header.intrpt_recov_tp_resolve_time"
	Set gtmtypes("shmpool_buff_hdr",179,"off")=1336
	Set gtmtypes("shmpool_buff_hdr",179,"len")=4
	Set gtmtypes("shmpool_buff_hdr",179,"type")="unsigned-int"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.intrpt_recov_tp_resolve_time")=179
	Set gtmtypes("shmpool_buff_hdr",180,"name")="shmpool_buff_hdr.shadow_file_header.recov_interrupted"
	Set gtmtypes("shmpool_buff_hdr",180,"off")=1340
	Set gtmtypes("shmpool_buff_hdr",180,"len")=4
	Set gtmtypes("shmpool_buff_hdr",180,"type")="boolean_t"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.recov_interrupted")=180
	Set gtmtypes("shmpool_buff_hdr",181,"name")="shmpool_buff_hdr.shadow_file_header.intrpt_recov_jnl_state"
	Set gtmtypes("shmpool_buff_hdr",181,"off")=1344
	Set gtmtypes("shmpool_buff_hdr",181,"len")=4
	Set gtmtypes("shmpool_buff_hdr",181,"type")="int"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.intrpt_recov_jnl_state")=181
	Set gtmtypes("shmpool_buff_hdr",182,"name")="shmpool_buff_hdr.shadow_file_header.intrpt_recov_repl_state"
	Set gtmtypes("shmpool_buff_hdr",182,"off")=1348
	Set gtmtypes("shmpool_buff_hdr",182,"len")=4
	Set gtmtypes("shmpool_buff_hdr",182,"type")="int"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.intrpt_recov_repl_state")=182
	Set gtmtypes("shmpool_buff_hdr",183,"name")="shmpool_buff_hdr.shadow_file_header.before_trunc_total_blks"
	Set gtmtypes("shmpool_buff_hdr",183,"off")=1352
	Set gtmtypes("shmpool_buff_hdr",183,"len")=4
	Set gtmtypes("shmpool_buff_hdr",183,"type")="unsigned-int"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.before_trunc_total_blks")=183
	Set gtmtypes("shmpool_buff_hdr",184,"name")="shmpool_buff_hdr.shadow_file_header.after_trunc_total_blks"
	Set gtmtypes("shmpool_buff_hdr",184,"off")=1356
	Set gtmtypes("shmpool_buff_hdr",184,"len")=4
	Set gtmtypes("shmpool_buff_hdr",184,"type")="unsigned-int"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.after_trunc_total_blks")=184
	Set gtmtypes("shmpool_buff_hdr",185,"name")="shmpool_buff_hdr.shadow_file_header.before_trunc_free_blocks"
	Set gtmtypes("shmpool_buff_hdr",185,"off")=1360
	Set gtmtypes("shmpool_buff_hdr",185,"len")=4
	Set gtmtypes("shmpool_buff_hdr",185,"type")="unsigned-int"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.before_trunc_free_blocks")=185
	Set gtmtypes("shmpool_buff_hdr",186,"name")="shmpool_buff_hdr.shadow_file_header.filler_trunc"
	Set gtmtypes("shmpool_buff_hdr",186,"off")=1364
	Set gtmtypes("shmpool_buff_hdr",186,"len")=4
	Set gtmtypes("shmpool_buff_hdr",186,"type")="unsigned-int"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.filler_trunc")=186
	Set gtmtypes("shmpool_buff_hdr",187,"name")="shmpool_buff_hdr.shadow_file_header.filler_1k"
	Set gtmtypes("shmpool_buff_hdr",187,"off")=1368
	Set gtmtypes("shmpool_buff_hdr",187,"len")=24
	Set gtmtypes("shmpool_buff_hdr",187,"type")="char"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.filler_1k")=187
	Set gtmtypes("shmpool_buff_hdr",188,"name")="shmpool_buff_hdr.shadow_file_header.jnl_file_name"
	Set gtmtypes("shmpool_buff_hdr",188,"off")=1392
	Set gtmtypes("shmpool_buff_hdr",188,"len")=256
	Set gtmtypes("shmpool_buff_hdr",188,"type")="unsigned-char"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.jnl_file_name")=188
	Set gtmtypes("shmpool_buff_hdr",189,"name")="shmpool_buff_hdr.shadow_file_header.reorg_restart_key"
	Set gtmtypes("shmpool_buff_hdr",189,"off")=1648
	Set gtmtypes("shmpool_buff_hdr",189,"len")=256
	Set gtmtypes("shmpool_buff_hdr",189,"type")="unsigned-char"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.reorg_restart_key")=189
	Set gtmtypes("shmpool_buff_hdr",190,"name")="shmpool_buff_hdr.shadow_file_header.machine_name"
	Set gtmtypes("shmpool_buff_hdr",190,"off")=1904
	Set gtmtypes("shmpool_buff_hdr",190,"len")=256
	Set gtmtypes("shmpool_buff_hdr",190,"type")="char"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.machine_name")=190
	Set gtmtypes("shmpool_buff_hdr",191,"name")="shmpool_buff_hdr.shadow_file_header.encryption_hash"
	Set gtmtypes("shmpool_buff_hdr",191,"off")=2160
	Set gtmtypes("shmpool_buff_hdr",191,"len")=80
	Set gtmtypes("shmpool_buff_hdr",191,"type")="char"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.encryption_hash")=191
	Set gtmtypes("shmpool_buff_hdr",192,"name")="shmpool_buff_hdr.shadow_file_header.encryption_hash2"
	Set gtmtypes("shmpool_buff_hdr",192,"off")=2240
	Set gtmtypes("shmpool_buff_hdr",192,"len")=80
	Set gtmtypes("shmpool_buff_hdr",192,"type")="char"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.encryption_hash2")=192
	Set gtmtypes("shmpool_buff_hdr",193,"name")="shmpool_buff_hdr.shadow_file_header.non_null_iv"
	Set gtmtypes("shmpool_buff_hdr",193,"off")=2320
	Set gtmtypes("shmpool_buff_hdr",193,"len")=4
	Set gtmtypes("shmpool_buff_hdr",193,"type")="boolean_t"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.non_null_iv")=193
	Set gtmtypes("shmpool_buff_hdr",194,"name")="shmpool_buff_hdr.shadow_file_header.encryption_hash_cutoff"
	Set gtmtypes("shmpool_buff_hdr",194,"off")=2324
	Set gtmtypes("shmpool_buff_hdr",194,"len")=4
	Set gtmtypes("shmpool_buff_hdr",194,"type")="int"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.encryption_hash_cutoff")=194
	Set gtmtypes("shmpool_buff_hdr",195,"name")="shmpool_buff_hdr.shadow_file_header.encryption_hash2_start_tn"
	Set gtmtypes("shmpool_buff_hdr",195,"off")=2328
	Set gtmtypes("shmpool_buff_hdr",195,"len")=8
	Set gtmtypes("shmpool_buff_hdr",195,"type")="uint64_t"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.encryption_hash2_start_tn")=195
	Set gtmtypes("shmpool_buff_hdr",196,"name")="shmpool_buff_hdr.shadow_file_header.filler_encrypt"
	Set gtmtypes("shmpool_buff_hdr",196,"off")=2336
	Set gtmtypes("shmpool_buff_hdr",196,"len")=80
	Set gtmtypes("shmpool_buff_hdr",196,"type")="char"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.filler_encrypt")=196
	Set gtmtypes("shmpool_buff_hdr",197,"name")="shmpool_buff_hdr.shadow_file_header.total_buffer_flush_tn"
	Set gtmtypes("shmpool_buff_hdr",197,"off")=2416
	Set gtmtypes("shmpool_buff_hdr",197,"len")=8
	Set gtmtypes("shmpool_buff_hdr",197,"type")="uint64_t"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.total_buffer_flush_tn")=197
	Set gtmtypes("shmpool_buff_hdr",198,"name")="shmpool_buff_hdr.shadow_file_header.bufct_buffer_flush_tn"
	Set gtmtypes("shmpool_buff_hdr",198,"off")=2424
	Set gtmtypes("shmpool_buff_hdr",198,"len")=8
	Set gtmtypes("shmpool_buff_hdr",198,"type")="uint64_t"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.bufct_buffer_flush_tn")=198
	Set gtmtypes("shmpool_buff_hdr",199,"name")="shmpool_buff_hdr.shadow_file_header.bufct_buffer_flush_loop_tn"
	Set gtmtypes("shmpool_buff_hdr",199,"off")=2432
	Set gtmtypes("shmpool_buff_hdr",199,"len")=8
	Set gtmtypes("shmpool_buff_hdr",199,"type")="uint64_t"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.bufct_buffer_flush_loop_tn")=199
	Set gtmtypes("shmpool_buff_hdr",200,"name")="shmpool_buff_hdr.shadow_file_header.wrt_calls_tn"
	Set gtmtypes("shmpool_buff_hdr",200,"off")=2440
	Set gtmtypes("shmpool_buff_hdr",200,"len")=8
	Set gtmtypes("shmpool_buff_hdr",200,"type")="uint64_t"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.wrt_calls_tn")=200
	Set gtmtypes("shmpool_buff_hdr",201,"name")="shmpool_buff_hdr.shadow_file_header.wrt_count_tn"
	Set gtmtypes("shmpool_buff_hdr",201,"off")=2448
	Set gtmtypes("shmpool_buff_hdr",201,"len")=8
	Set gtmtypes("shmpool_buff_hdr",201,"type")="uint64_t"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.wrt_count_tn")=201
	Set gtmtypes("shmpool_buff_hdr",202,"name")="shmpool_buff_hdr.shadow_file_header.wrt_blocked_tn"
	Set gtmtypes("shmpool_buff_hdr",202,"off")=2456
	Set gtmtypes("shmpool_buff_hdr",202,"len")=8
	Set gtmtypes("shmpool_buff_hdr",202,"type")="uint64_t"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.wrt_blocked_tn")=202
	Set gtmtypes("shmpool_buff_hdr",203,"name")="shmpool_buff_hdr.shadow_file_header.wrt_busy_tn"
	Set gtmtypes("shmpool_buff_hdr",203,"off")=2464
	Set gtmtypes("shmpool_buff_hdr",203,"len")=8
	Set gtmtypes("shmpool_buff_hdr",203,"type")="uint64_t"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.wrt_busy_tn")=203
	Set gtmtypes("shmpool_buff_hdr",204,"name")="shmpool_buff_hdr.shadow_file_header.wrt_noblks_wrtn_tn"
	Set gtmtypes("shmpool_buff_hdr",204,"off")=2472
	Set gtmtypes("shmpool_buff_hdr",204,"len")=8
	Set gtmtypes("shmpool_buff_hdr",204,"type")="uint64_t"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.wrt_noblks_wrtn_tn")=204
	Set gtmtypes("shmpool_buff_hdr",205,"name")="shmpool_buff_hdr.shadow_file_header.wcb_gds_rundown2_tn"
	Set gtmtypes("shmpool_buff_hdr",205,"off")=2480
	Set gtmtypes("shmpool_buff_hdr",205,"len")=8
	Set gtmtypes("shmpool_buff_hdr",205,"type")="uint64_t"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.wcb_gds_rundown2_tn")=205
	Set gtmtypes("shmpool_buff_hdr",206,"name")="shmpool_buff_hdr.shadow_file_header.reserved_bgtrcrec2_tn"
	Set gtmtypes("shmpool_buff_hdr",206,"off")=2488
	Set gtmtypes("shmpool_buff_hdr",206,"len")=8
	Set gtmtypes("shmpool_buff_hdr",206,"type")="uint64_t"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.reserved_bgtrcrec2_tn")=206
	Set gtmtypes("shmpool_buff_hdr",207,"name")="shmpool_buff_hdr.shadow_file_header.reserved_bgtrcrec3_tn"
	Set gtmtypes("shmpool_buff_hdr",207,"off")=2496
	Set gtmtypes("shmpool_buff_hdr",207,"len")=8
	Set gtmtypes("shmpool_buff_hdr",207,"type")="uint64_t"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.reserved_bgtrcrec3_tn")=207
	Set gtmtypes("shmpool_buff_hdr",208,"name")="shmpool_buff_hdr.shadow_file_header.lost_block_recovery_tn"
	Set gtmtypes("shmpool_buff_hdr",208,"off")=2504
	Set gtmtypes("shmpool_buff_hdr",208,"len")=8
	Set gtmtypes("shmpool_buff_hdr",208,"type")="uint64_t"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.lost_block_recovery_tn")=208
	Set gtmtypes("shmpool_buff_hdr",209,"name")="shmpool_buff_hdr.shadow_file_header.wc_blocked_onln_rlbk_tn"
	Set gtmtypes("shmpool_buff_hdr",209,"off")=2512
	Set gtmtypes("shmpool_buff_hdr",209,"len")=8
	Set gtmtypes("shmpool_buff_hdr",209,"type")="uint64_t"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.wc_blocked_onln_rlbk_tn")=209
	Set gtmtypes("shmpool_buff_hdr",210,"name")="shmpool_buff_hdr.shadow_file_header.stale_tn"
	Set gtmtypes("shmpool_buff_hdr",210,"off")=2520
	Set gtmtypes("shmpool_buff_hdr",210,"len")=8
	Set gtmtypes("shmpool_buff_hdr",210,"type")="uint64_t"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.stale_tn")=210
	Set gtmtypes("shmpool_buff_hdr",211,"name")="shmpool_buff_hdr.shadow_file_header.starved_tn"
	Set gtmtypes("shmpool_buff_hdr",211,"off")=2528
	Set gtmtypes("shmpool_buff_hdr",211,"len")=8
	Set gtmtypes("shmpool_buff_hdr",211,"type")="uint64_t"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.starved_tn")=211
	Set gtmtypes("shmpool_buff_hdr",212,"name")="shmpool_buff_hdr.shadow_file_header.stale_timer_started_tn"
	Set gtmtypes("shmpool_buff_hdr",212,"off")=2536
	Set gtmtypes("shmpool_buff_hdr",212,"len")=8
	Set gtmtypes("shmpool_buff_hdr",212,"type")="uint64_t"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.stale_timer_started_tn")=212
	Set gtmtypes("shmpool_buff_hdr",213,"name")="shmpool_buff_hdr.shadow_file_header.stale_timer_pop_tn"
	Set gtmtypes("shmpool_buff_hdr",213,"off")=2544
	Set gtmtypes("shmpool_buff_hdr",213,"len")=8
	Set gtmtypes("shmpool_buff_hdr",213,"type")="uint64_t"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.stale_timer_pop_tn")=213
	Set gtmtypes("shmpool_buff_hdr",214,"name")="shmpool_buff_hdr.shadow_file_header.stale_process_defer_tn"
	Set gtmtypes("shmpool_buff_hdr",214,"off")=2552
	Set gtmtypes("shmpool_buff_hdr",214,"len")=8
	Set gtmtypes("shmpool_buff_hdr",214,"type")="uint64_t"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.stale_process_defer_tn")=214
	Set gtmtypes("shmpool_buff_hdr",215,"name")="shmpool_buff_hdr.shadow_file_header.stale_defer_processed_tn"
	Set gtmtypes("shmpool_buff_hdr",215,"off")=2560
	Set gtmtypes("shmpool_buff_hdr",215,"len")=8
	Set gtmtypes("shmpool_buff_hdr",215,"type")="uint64_t"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.stale_defer_processed_tn")=215
	Set gtmtypes("shmpool_buff_hdr",216,"name")="shmpool_buff_hdr.shadow_file_header.spcfc_buffer_flush_tn"
	Set gtmtypes("shmpool_buff_hdr",216,"off")=2568
	Set gtmtypes("shmpool_buff_hdr",216,"len")=8
	Set gtmtypes("shmpool_buff_hdr",216,"type")="uint64_t"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.spcfc_buffer_flush_tn")=216
	Set gtmtypes("shmpool_buff_hdr",217,"name")="shmpool_buff_hdr.shadow_file_header.spcfc_buffer_flush_loop_tn"
	Set gtmtypes("shmpool_buff_hdr",217,"off")=2576
	Set gtmtypes("shmpool_buff_hdr",217,"len")=8
	Set gtmtypes("shmpool_buff_hdr",217,"type")="uint64_t"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.spcfc_buffer_flush_loop_tn")=217
	Set gtmtypes("shmpool_buff_hdr",218,"name")="shmpool_buff_hdr.shadow_file_header.spcfc_buffer_flush_retries_tn"
	Set gtmtypes("shmpool_buff_hdr",218,"off")=2584
	Set gtmtypes("shmpool_buff_hdr",218,"len")=8
	Set gtmtypes("shmpool_buff_hdr",218,"type")="uint64_t"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.spcfc_buffer_flush_retries_tn")=218
	Set gtmtypes("shmpool_buff_hdr",219,"name")="shmpool_buff_hdr.shadow_file_header.spcfc_buffer_flushed_during_lockwait_tn"
	Set gtmtypes("shmpool_buff_hdr",219,"off")=2592
	Set gtmtypes("shmpool_buff_hdr",219,"len")=8
	Set gtmtypes("shmpool_buff_hdr",219,"type")="uint64_t"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.spcfc_buffer_flushed_during_lockwait_tn")=219
	Set gtmtypes("shmpool_buff_hdr",220,"name")="shmpool_buff_hdr.shadow_file_header.tp_crit_retries_tn"
	Set gtmtypes("shmpool_buff_hdr",220,"off")=2600
	Set gtmtypes("shmpool_buff_hdr",220,"len")=8
	Set gtmtypes("shmpool_buff_hdr",220,"type")="uint64_t"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.tp_crit_retries_tn")=220
	Set gtmtypes("shmpool_buff_hdr",221,"name")="shmpool_buff_hdr.shadow_file_header.db_csh_getn_flush_dirty_tn"
	Set gtmtypes("shmpool_buff_hdr",221,"off")=2608
	Set gtmtypes("shmpool_buff_hdr",221,"len")=8
	Set gtmtypes("shmpool_buff_hdr",221,"type")="uint64_t"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.db_csh_getn_flush_dirty_tn")=221
	Set gtmtypes("shmpool_buff_hdr",222,"name")="shmpool_buff_hdr.shadow_file_header.db_csh_getn_rip_wait_tn"
	Set gtmtypes("shmpool_buff_hdr",222,"off")=2616
	Set gtmtypes("shmpool_buff_hdr",222,"len")=8
	Set gtmtypes("shmpool_buff_hdr",222,"type")="uint64_t"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.db_csh_getn_rip_wait_tn")=222
	Set gtmtypes("shmpool_buff_hdr",223,"name")="shmpool_buff_hdr.shadow_file_header.db_csh_getn_buf_owner_stuck_tn"
	Set gtmtypes("shmpool_buff_hdr",223,"off")=2624
	Set gtmtypes("shmpool_buff_hdr",223,"len")=8
	Set gtmtypes("shmpool_buff_hdr",223,"type")="uint64_t"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.db_csh_getn_buf_owner_stuck_tn")=223
	Set gtmtypes("shmpool_buff_hdr",224,"name")="shmpool_buff_hdr.shadow_file_header.db_csh_getn_out_of_design_tn"
	Set gtmtypes("shmpool_buff_hdr",224,"off")=2632
	Set gtmtypes("shmpool_buff_hdr",224,"len")=8
	Set gtmtypes("shmpool_buff_hdr",224,"type")="uint64_t"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.db_csh_getn_out_of_design_tn")=224
	Set gtmtypes("shmpool_buff_hdr",225,"name")="shmpool_buff_hdr.shadow_file_header.t_qread_buf_owner_stuck_tn"
	Set gtmtypes("shmpool_buff_hdr",225,"off")=2640
	Set gtmtypes("shmpool_buff_hdr",225,"len")=8
	Set gtmtypes("shmpool_buff_hdr",225,"type")="uint64_t"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.t_qread_buf_owner_stuck_tn")=225
	Set gtmtypes("shmpool_buff_hdr",226,"name")="shmpool_buff_hdr.shadow_file_header.t_qread_out_of_design_tn"
	Set gtmtypes("shmpool_buff_hdr",226,"off")=2648
	Set gtmtypes("shmpool_buff_hdr",226,"len")=8
	Set gtmtypes("shmpool_buff_hdr",226,"type")="uint64_t"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.t_qread_out_of_design_tn")=226
	Set gtmtypes("shmpool_buff_hdr",227,"name")="shmpool_buff_hdr.shadow_file_header.unused_filler_0031_tn"
	Set gtmtypes("shmpool_buff_hdr",227,"off")=2656
	Set gtmtypes("shmpool_buff_hdr",227,"len")=8
	Set gtmtypes("shmpool_buff_hdr",227,"type")="uint64_t"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.unused_filler_0031_tn")=227
	Set gtmtypes("shmpool_buff_hdr",228,"name")="shmpool_buff_hdr.shadow_file_header.mlock_wakeups_tn"
	Set gtmtypes("shmpool_buff_hdr",228,"off")=2664
	Set gtmtypes("shmpool_buff_hdr",228,"len")=8
	Set gtmtypes("shmpool_buff_hdr",228,"type")="uint64_t"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.mlock_wakeups_tn")=228
	Set gtmtypes("shmpool_buff_hdr",229,"name")="shmpool_buff_hdr.shadow_file_header.wc_blocked_wcs_recover_invoked_tn"
	Set gtmtypes("shmpool_buff_hdr",229,"off")=2672
	Set gtmtypes("shmpool_buff_hdr",229,"len")=8
	Set gtmtypes("shmpool_buff_hdr",229,"type")="uint64_t"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.wc_blocked_wcs_recover_invoked_tn")=229
	Set gtmtypes("shmpool_buff_hdr",230,"name")="shmpool_buff_hdr.shadow_file_header.wc_blocked_wcs_verify_passed_tn"
	Set gtmtypes("shmpool_buff_hdr",230,"off")=2680
	Set gtmtypes("shmpool_buff_hdr",230,"len")=8
	Set gtmtypes("shmpool_buff_hdr",230,"type")="uint64_t"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.wc_blocked_wcs_verify_passed_tn")=230
	Set gtmtypes("shmpool_buff_hdr",231,"name")="shmpool_buff_hdr.shadow_file_header.wc_blocked_t_qread_db_csh_getn_invalid_blk_tn"
	Set gtmtypes("shmpool_buff_hdr",231,"off")=2688
	Set gtmtypes("shmpool_buff_hdr",231,"len")=8
	Set gtmtypes("shmpool_buff_hdr",231,"type")="uint64_t"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.wc_blocked_t_qread_db_csh_getn_invalid_blk_tn")=231
	Set gtmtypes("shmpool_buff_hdr",232,"name")="shmpool_buff_hdr.shadow_file_header.wc_blocked_t_qread_db_csh_get_invalid_blk_tn"
	Set gtmtypes("shmpool_buff_hdr",232,"off")=2696
	Set gtmtypes("shmpool_buff_hdr",232,"len")=8
	Set gtmtypes("shmpool_buff_hdr",232,"type")="uint64_t"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.wc_blocked_t_qread_db_csh_get_invalid_blk_tn")=232
	Set gtmtypes("shmpool_buff_hdr",233,"name")="shmpool_buff_hdr.shadow_file_header.wc_blocked_db_csh_getn_loopexceed_tn"
	Set gtmtypes("shmpool_buff_hdr",233,"off")=2704
	Set gtmtypes("shmpool_buff_hdr",233,"len")=8
	Set gtmtypes("shmpool_buff_hdr",233,"type")="uint64_t"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.wc_blocked_db_csh_getn_loopexceed_tn")=233
	Set gtmtypes("shmpool_buff_hdr",234,"name")="shmpool_buff_hdr.shadow_file_header.wc_blocked_db_csh_getn_wcsstarvewrt_tn"
	Set gtmtypes("shmpool_buff_hdr",234,"off")=2712
	Set gtmtypes("shmpool_buff_hdr",234,"len")=8
	Set gtmtypes("shmpool_buff_hdr",234,"type")="uint64_t"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.wc_blocked_db_csh_getn_wcsstarvewrt_tn")=234
	Set gtmtypes("shmpool_buff_hdr",235,"name")="shmpool_buff_hdr.shadow_file_header.wc_blocked_db_csh_get_tn"
	Set gtmtypes("shmpool_buff_hdr",235,"off")=2720
	Set gtmtypes("shmpool_buff_hdr",235,"len")=8
	Set gtmtypes("shmpool_buff_hdr",235,"type")="uint64_t"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.wc_blocked_db_csh_get_tn")=235
	Set gtmtypes("shmpool_buff_hdr",236,"name")="shmpool_buff_hdr.shadow_file_header.wc_blocked_tp_tend_wcsgetspace_tn"
	Set gtmtypes("shmpool_buff_hdr",236,"off")=2728
	Set gtmtypes("shmpool_buff_hdr",236,"len")=8
	Set gtmtypes("shmpool_buff_hdr",236,"type")="uint64_t"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.wc_blocked_tp_tend_wcsgetspace_tn")=236
	Set gtmtypes("shmpool_buff_hdr",237,"name")="shmpool_buff_hdr.shadow_file_header.wc_blocked_tp_tend_t1_tn"
	Set gtmtypes("shmpool_buff_hdr",237,"off")=2736
	Set gtmtypes("shmpool_buff_hdr",237,"len")=8
	Set gtmtypes("shmpool_buff_hdr",237,"type")="uint64_t"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.wc_blocked_tp_tend_t1_tn")=237
	Set gtmtypes("shmpool_buff_hdr",238,"name")="shmpool_buff_hdr.shadow_file_header.wc_blocked_tp_tend_bitmap_tn"
	Set gtmtypes("shmpool_buff_hdr",238,"off")=2744
	Set gtmtypes("shmpool_buff_hdr",238,"len")=8
	Set gtmtypes("shmpool_buff_hdr",238,"type")="uint64_t"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.wc_blocked_tp_tend_bitmap_tn")=238
	Set gtmtypes("shmpool_buff_hdr",239,"name")="shmpool_buff_hdr.shadow_file_header.wc_blocked_tp_tend_jnl_cwset_tn"
	Set gtmtypes("shmpool_buff_hdr",239,"off")=2752
	Set gtmtypes("shmpool_buff_hdr",239,"len")=8
	Set gtmtypes("shmpool_buff_hdr",239,"type")="uint64_t"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.wc_blocked_tp_tend_jnl_cwset_tn")=239
	Set gtmtypes("shmpool_buff_hdr",240,"name")="shmpool_buff_hdr.shadow_file_header.wc_blocked_tp_tend_jnl_wcsflu_tn"
	Set gtmtypes("shmpool_buff_hdr",240,"off")=2760
	Set gtmtypes("shmpool_buff_hdr",240,"len")=8
	Set gtmtypes("shmpool_buff_hdr",240,"type")="uint64_t"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.wc_blocked_tp_tend_jnl_wcsflu_tn")=240
	Set gtmtypes("shmpool_buff_hdr",241,"name")="shmpool_buff_hdr.shadow_file_header.wc_blocked_t_end_hist_tn"
	Set gtmtypes("shmpool_buff_hdr",241,"off")=2768
	Set gtmtypes("shmpool_buff_hdr",241,"len")=8
	Set gtmtypes("shmpool_buff_hdr",241,"type")="uint64_t"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.wc_blocked_t_end_hist_tn")=241
	Set gtmtypes("shmpool_buff_hdr",242,"name")="shmpool_buff_hdr.shadow_file_header.wc_blocked_t_end_hist1_nullbt_tn"
	Set gtmtypes("shmpool_buff_hdr",242,"off")=2776
	Set gtmtypes("shmpool_buff_hdr",242,"len")=8
	Set gtmtypes("shmpool_buff_hdr",242,"type")="uint64_t"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.wc_blocked_t_end_hist1_nullbt_tn")=242
	Set gtmtypes("shmpool_buff_hdr",243,"name")="shmpool_buff_hdr.shadow_file_header.wc_blocked_t_end_hist1_nonnullbt_tn"
	Set gtmtypes("shmpool_buff_hdr",243,"off")=2784
	Set gtmtypes("shmpool_buff_hdr",243,"len")=8
	Set gtmtypes("shmpool_buff_hdr",243,"type")="uint64_t"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.wc_blocked_t_end_hist1_nonnullbt_tn")=243
	Set gtmtypes("shmpool_buff_hdr",244,"name")="shmpool_buff_hdr.shadow_file_header.wc_blocked_t_end_bitmap_nullbt_tn"
	Set gtmtypes("shmpool_buff_hdr",244,"off")=2792
	Set gtmtypes("shmpool_buff_hdr",244,"len")=8
	Set gtmtypes("shmpool_buff_hdr",244,"type")="uint64_t"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.wc_blocked_t_end_bitmap_nullbt_tn")=244
	Set gtmtypes("shmpool_buff_hdr",245,"name")="shmpool_buff_hdr.shadow_file_header.wc_blocked_t_end_bitmap_nonnullbt_tn"
	Set gtmtypes("shmpool_buff_hdr",245,"off")=2800
	Set gtmtypes("shmpool_buff_hdr",245,"len")=8
	Set gtmtypes("shmpool_buff_hdr",245,"type")="uint64_t"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.wc_blocked_t_end_bitmap_nonnullbt_tn")=245
	Set gtmtypes("shmpool_buff_hdr",246,"name")="shmpool_buff_hdr.shadow_file_header.wc_blocked_t_end_jnl_cwset_tn"
	Set gtmtypes("shmpool_buff_hdr",246,"off")=2808
	Set gtmtypes("shmpool_buff_hdr",246,"len")=8
	Set gtmtypes("shmpool_buff_hdr",246,"type")="uint64_t"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.wc_blocked_t_end_jnl_cwset_tn")=246
	Set gtmtypes("shmpool_buff_hdr",247,"name")="shmpool_buff_hdr.shadow_file_header.wc_blocked_t_end_jnl_wcsflu_tn"
	Set gtmtypes("shmpool_buff_hdr",247,"off")=2816
	Set gtmtypes("shmpool_buff_hdr",247,"len")=8
	Set gtmtypes("shmpool_buff_hdr",247,"type")="uint64_t"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.wc_blocked_t_end_jnl_wcsflu_tn")=247
	Set gtmtypes("shmpool_buff_hdr",248,"name")="shmpool_buff_hdr.shadow_file_header.unused_filler_0052_tn"
	Set gtmtypes("shmpool_buff_hdr",248,"off")=2824
	Set gtmtypes("shmpool_buff_hdr",248,"len")=8
	Set gtmtypes("shmpool_buff_hdr",248,"type")="uint64_t"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.unused_filler_0052_tn")=248
	Set gtmtypes("shmpool_buff_hdr",249,"name")="shmpool_buff_hdr.shadow_file_header.wc_blocked_tpckh_hist1_nullbt_tn"
	Set gtmtypes("shmpool_buff_hdr",249,"off")=2832
	Set gtmtypes("shmpool_buff_hdr",249,"len")=8
	Set gtmtypes("shmpool_buff_hdr",249,"type")="uint64_t"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.wc_blocked_tpckh_hist1_nullbt_tn")=249
	Set gtmtypes("shmpool_buff_hdr",250,"name")="shmpool_buff_hdr.shadow_file_header.wc_blocked_tpckh_hist1_nonnullbt_tn"
	Set gtmtypes("shmpool_buff_hdr",250,"off")=2840
	Set gtmtypes("shmpool_buff_hdr",250,"len")=8
	Set gtmtypes("shmpool_buff_hdr",250,"type")="uint64_t"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.wc_blocked_tpckh_hist1_nonnullbt_tn")=250
	Set gtmtypes("shmpool_buff_hdr",251,"name")="shmpool_buff_hdr.shadow_file_header.jnl_blocked_writer_lost_tn"
	Set gtmtypes("shmpool_buff_hdr",251,"off")=2848
	Set gtmtypes("shmpool_buff_hdr",251,"len")=8
	Set gtmtypes("shmpool_buff_hdr",251,"type")="uint64_t"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.jnl_blocked_writer_lost_tn")=251
	Set gtmtypes("shmpool_buff_hdr",252,"name")="shmpool_buff_hdr.shadow_file_header.jnl_blocked_writer_stuck_tn"
	Set gtmtypes("shmpool_buff_hdr",252,"off")=2856
	Set gtmtypes("shmpool_buff_hdr",252,"len")=8
	Set gtmtypes("shmpool_buff_hdr",252,"type")="uint64_t"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.jnl_blocked_writer_stuck_tn")=252
	Set gtmtypes("shmpool_buff_hdr",253,"name")="shmpool_buff_hdr.shadow_file_header.jnl_blocked_writer_blocked_tn"
	Set gtmtypes("shmpool_buff_hdr",253,"off")=2864
	Set gtmtypes("shmpool_buff_hdr",253,"len")=8
	Set gtmtypes("shmpool_buff_hdr",253,"type")="uint64_t"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.jnl_blocked_writer_blocked_tn")=253
	Set gtmtypes("shmpool_buff_hdr",254,"name")="shmpool_buff_hdr.shadow_file_header.n_jnl_fsyncs_tn"
	Set gtmtypes("shmpool_buff_hdr",254,"off")=2872
	Set gtmtypes("shmpool_buff_hdr",254,"len")=8
	Set gtmtypes("shmpool_buff_hdr",254,"type")="uint64_t"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.n_jnl_fsyncs_tn")=254
	Set gtmtypes("shmpool_buff_hdr",255,"name")="shmpool_buff_hdr.shadow_file_header.n_jnl_fsync_tries_tn"
	Set gtmtypes("shmpool_buff_hdr",255,"off")=2880
	Set gtmtypes("shmpool_buff_hdr",255,"len")=8
	Set gtmtypes("shmpool_buff_hdr",255,"type")="uint64_t"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.n_jnl_fsync_tries_tn")=255
	Set gtmtypes("shmpool_buff_hdr",256,"name")="shmpool_buff_hdr.shadow_file_header.n_jnl_fsync_recovers_tn"
	Set gtmtypes("shmpool_buff_hdr",256,"off")=2888
	Set gtmtypes("shmpool_buff_hdr",256,"len")=8
	Set gtmtypes("shmpool_buff_hdr",256,"type")="uint64_t"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.n_jnl_fsync_recovers_tn")=256
	Set gtmtypes("shmpool_buff_hdr",257,"name")="shmpool_buff_hdr.shadow_file_header.n_db_fsyncs_tn"
	Set gtmtypes("shmpool_buff_hdr",257,"off")=2896
	Set gtmtypes("shmpool_buff_hdr",257,"len")=8
	Set gtmtypes("shmpool_buff_hdr",257,"type")="uint64_t"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.n_db_fsyncs_tn")=257
	Set gtmtypes("shmpool_buff_hdr",258,"name")="shmpool_buff_hdr.shadow_file_header.n_db_fsyncs_in_crit_tn"
	Set gtmtypes("shmpool_buff_hdr",258,"off")=2904
	Set gtmtypes("shmpool_buff_hdr",258,"len")=8
	Set gtmtypes("shmpool_buff_hdr",258,"type")="uint64_t"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.n_db_fsyncs_in_crit_tn")=258
	Set gtmtypes("shmpool_buff_hdr",259,"name")="shmpool_buff_hdr.shadow_file_header.n_dbsync_timers_tn"
	Set gtmtypes("shmpool_buff_hdr",259,"off")=2912
	Set gtmtypes("shmpool_buff_hdr",259,"len")=8
	Set gtmtypes("shmpool_buff_hdr",259,"type")="uint64_t"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.n_dbsync_timers_tn")=259
	Set gtmtypes("shmpool_buff_hdr",260,"name")="shmpool_buff_hdr.shadow_file_header.n_dbsync_writes_tn"
	Set gtmtypes("shmpool_buff_hdr",260,"off")=2920
	Set gtmtypes("shmpool_buff_hdr",260,"len")=8
	Set gtmtypes("shmpool_buff_hdr",260,"type")="uint64_t"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.n_dbsync_writes_tn")=260
	Set gtmtypes("shmpool_buff_hdr",261,"name")="shmpool_buff_hdr.shadow_file_header.mutex_queue_full_tn"
	Set gtmtypes("shmpool_buff_hdr",261,"off")=2928
	Set gtmtypes("shmpool_buff_hdr",261,"len")=8
	Set gtmtypes("shmpool_buff_hdr",261,"type")="uint64_t"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.mutex_queue_full_tn")=261
	Set gtmtypes("shmpool_buff_hdr",262,"name")="shmpool_buff_hdr.shadow_file_header.wcb_bt_put_tn"
	Set gtmtypes("shmpool_buff_hdr",262,"off")=2936
	Set gtmtypes("shmpool_buff_hdr",262,"len")=8
	Set gtmtypes("shmpool_buff_hdr",262,"type")="uint64_t"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.wcb_bt_put_tn")=262
	Set gtmtypes("shmpool_buff_hdr",263,"name")="shmpool_buff_hdr.shadow_file_header.wcb_mutex_salvage_tn"
	Set gtmtypes("shmpool_buff_hdr",263,"off")=2944
	Set gtmtypes("shmpool_buff_hdr",263,"len")=8
	Set gtmtypes("shmpool_buff_hdr",263,"type")="uint64_t"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.wcb_mutex_salvage_tn")=263
	Set gtmtypes("shmpool_buff_hdr",264,"name")="shmpool_buff_hdr.shadow_file_header.wcb_wcs_flu0_tn"
	Set gtmtypes("shmpool_buff_hdr",264,"off")=2952
	Set gtmtypes("shmpool_buff_hdr",264,"len")=8
	Set gtmtypes("shmpool_buff_hdr",264,"type")="uint64_t"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.wcb_wcs_flu0_tn")=264
	Set gtmtypes("shmpool_buff_hdr",265,"name")="shmpool_buff_hdr.shadow_file_header.wcb_t_end_sysops_nocr_invcr_tn"
	Set gtmtypes("shmpool_buff_hdr",265,"off")=2960
	Set gtmtypes("shmpool_buff_hdr",265,"len")=8
	Set gtmtypes("shmpool_buff_hdr",265,"type")="uint64_t"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.wcb_t_end_sysops_nocr_invcr_tn")=265
	Set gtmtypes("shmpool_buff_hdr",266,"name")="shmpool_buff_hdr.shadow_file_header.wcb_t_end_sysops_cr_invcr_tn"
	Set gtmtypes("shmpool_buff_hdr",266,"off")=2968
	Set gtmtypes("shmpool_buff_hdr",266,"len")=8
	Set gtmtypes("shmpool_buff_hdr",266,"type")="uint64_t"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.wcb_t_end_sysops_cr_invcr_tn")=266
	Set gtmtypes("shmpool_buff_hdr",267,"name")="shmpool_buff_hdr.shadow_file_header.wcb_t_end_sysops_rip_wait_tn"
	Set gtmtypes("shmpool_buff_hdr",267,"off")=2976
	Set gtmtypes("shmpool_buff_hdr",267,"len")=8
	Set gtmtypes("shmpool_buff_hdr",267,"type")="uint64_t"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.wcb_t_end_sysops_rip_wait_tn")=267
	Set gtmtypes("shmpool_buff_hdr",268,"name")="shmpool_buff_hdr.shadow_file_header.wcb_t_end_sysops_dirtyripwait_tn"
	Set gtmtypes("shmpool_buff_hdr",268,"off")=2984
	Set gtmtypes("shmpool_buff_hdr",268,"len")=8
	Set gtmtypes("shmpool_buff_hdr",268,"type")="uint64_t"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.wcb_t_end_sysops_dirtyripwait_tn")=268
	Set gtmtypes("shmpool_buff_hdr",269,"name")="shmpool_buff_hdr.shadow_file_header.wcb_gds_rundown1_tn"
	Set gtmtypes("shmpool_buff_hdr",269,"off")=2992
	Set gtmtypes("shmpool_buff_hdr",269,"len")=8
	Set gtmtypes("shmpool_buff_hdr",269,"type")="uint64_t"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.wcb_gds_rundown1_tn")=269
	Set gtmtypes("shmpool_buff_hdr",270,"name")="shmpool_buff_hdr.shadow_file_header.wcb_wcs_flu1_tn"
	Set gtmtypes("shmpool_buff_hdr",270,"off")=3000
	Set gtmtypes("shmpool_buff_hdr",270,"len")=8
	Set gtmtypes("shmpool_buff_hdr",270,"type")="uint64_t"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.wcb_wcs_flu1_tn")=270
	Set gtmtypes("shmpool_buff_hdr",271,"name")="shmpool_buff_hdr.shadow_file_header.wcb_mu_back_tn"
	Set gtmtypes("shmpool_buff_hdr",271,"off")=3008
	Set gtmtypes("shmpool_buff_hdr",271,"len")=8
	Set gtmtypes("shmpool_buff_hdr",271,"type")="uint64_t"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.wcb_mu_back_tn")=271
	Set gtmtypes("shmpool_buff_hdr",272,"name")="shmpool_buff_hdr.shadow_file_header.wcb_t_end_sysops_dirty_invcr_tn"
	Set gtmtypes("shmpool_buff_hdr",272,"off")=3016
	Set gtmtypes("shmpool_buff_hdr",272,"len")=8
	Set gtmtypes("shmpool_buff_hdr",272,"type")="uint64_t"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.wcb_t_end_sysops_dirty_invcr_tn")=272
	Set gtmtypes("shmpool_buff_hdr",273,"name")="shmpool_buff_hdr.shadow_file_header.wcb_t_end_sysops_wtfini_fail_tn"
	Set gtmtypes("shmpool_buff_hdr",273,"off")=3024
	Set gtmtypes("shmpool_buff_hdr",273,"len")=8
	Set gtmtypes("shmpool_buff_hdr",273,"type")="uint64_t"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.wcb_t_end_sysops_wtfini_fail_tn")=273
	Set gtmtypes("shmpool_buff_hdr",274,"name")="shmpool_buff_hdr.shadow_file_header.wcb_t_end_sysops_twin_stuck_tn"
	Set gtmtypes("shmpool_buff_hdr",274,"off")=3032
	Set gtmtypes("shmpool_buff_hdr",274,"len")=8
	Set gtmtypes("shmpool_buff_hdr",274,"type")="uint64_t"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.wcb_t_end_sysops_twin_stuck_tn")=274
	Set gtmtypes("shmpool_buff_hdr",275,"name")="shmpool_buff_hdr.shadow_file_header.db_csh_getn_wrt_latch_stuck_tn"
	Set gtmtypes("shmpool_buff_hdr",275,"off")=3040
	Set gtmtypes("shmpool_buff_hdr",275,"len")=8
	Set gtmtypes("shmpool_buff_hdr",275,"type")="uint64_t"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.db_csh_getn_wrt_latch_stuck_tn")=275
	Set gtmtypes("shmpool_buff_hdr",276,"name")="shmpool_buff_hdr.shadow_file_header.wcb_secshr_db_clnup_now_crit_tn"
	Set gtmtypes("shmpool_buff_hdr",276,"off")=3048
	Set gtmtypes("shmpool_buff_hdr",276,"len")=8
	Set gtmtypes("shmpool_buff_hdr",276,"type")="uint64_t"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.wcb_secshr_db_clnup_now_crit_tn")=276
	Set gtmtypes("shmpool_buff_hdr",277,"name")="shmpool_buff_hdr.shadow_file_header.wc_blocked_t_qread_bad_bt_index1_tn"
	Set gtmtypes("shmpool_buff_hdr",277,"off")=3056
	Set gtmtypes("shmpool_buff_hdr",277,"len")=8
	Set gtmtypes("shmpool_buff_hdr",277,"type")="uint64_t"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.wc_blocked_t_qread_bad_bt_index1_tn")=277
	Set gtmtypes("shmpool_buff_hdr",278,"name")="shmpool_buff_hdr.shadow_file_header.wc_blocked_t_qread_bad_bt_index2_tn"
	Set gtmtypes("shmpool_buff_hdr",278,"off")=3064
	Set gtmtypes("shmpool_buff_hdr",278,"len")=8
	Set gtmtypes("shmpool_buff_hdr",278,"type")="uint64_t"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.wc_blocked_t_qread_bad_bt_index2_tn")=278
	Set gtmtypes("shmpool_buff_hdr",279,"name")="shmpool_buff_hdr.shadow_file_header.wc_blocked_t_end_crbtmismatch1_tn"
	Set gtmtypes("shmpool_buff_hdr",279,"off")=3072
	Set gtmtypes("shmpool_buff_hdr",279,"len")=8
	Set gtmtypes("shmpool_buff_hdr",279,"type")="uint64_t"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.wc_blocked_t_end_crbtmismatch1_tn")=279
	Set gtmtypes("shmpool_buff_hdr",280,"name")="shmpool_buff_hdr.shadow_file_header.wc_blocked_t_end_crbtmismatch2_tn"
	Set gtmtypes("shmpool_buff_hdr",280,"off")=3080
	Set gtmtypes("shmpool_buff_hdr",280,"len")=8
	Set gtmtypes("shmpool_buff_hdr",280,"type")="uint64_t"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.wc_blocked_t_end_crbtmismatch2_tn")=280
	Set gtmtypes("shmpool_buff_hdr",281,"name")="shmpool_buff_hdr.shadow_file_header.wc_blocked_t_end_crbtmismatch3_tn"
	Set gtmtypes("shmpool_buff_hdr",281,"off")=3088
	Set gtmtypes("shmpool_buff_hdr",281,"len")=8
	Set gtmtypes("shmpool_buff_hdr",281,"type")="uint64_t"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.wc_blocked_t_end_crbtmismatch3_tn")=281
	Set gtmtypes("shmpool_buff_hdr",282,"name")="shmpool_buff_hdr.shadow_file_header.wc_blocked_tp_tend_crbtmismatch1_tn"
	Set gtmtypes("shmpool_buff_hdr",282,"off")=3096
	Set gtmtypes("shmpool_buff_hdr",282,"len")=8
	Set gtmtypes("shmpool_buff_hdr",282,"type")="uint64_t"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.wc_blocked_tp_tend_crbtmismatch1_tn")=282
	Set gtmtypes("shmpool_buff_hdr",283,"name")="shmpool_buff_hdr.shadow_file_header.wc_blocked_tp_tend_crbtmismatch2_tn"
	Set gtmtypes("shmpool_buff_hdr",283,"off")=3104
	Set gtmtypes("shmpool_buff_hdr",283,"len")=8
	Set gtmtypes("shmpool_buff_hdr",283,"type")="uint64_t"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.wc_blocked_tp_tend_crbtmismatch2_tn")=283
	Set gtmtypes("shmpool_buff_hdr",284,"name")="shmpool_buff_hdr.shadow_file_header.wc_blocked_tp_tend_crbtmismatch3_tn"
	Set gtmtypes("shmpool_buff_hdr",284,"off")=3112
	Set gtmtypes("shmpool_buff_hdr",284,"len")=8
	Set gtmtypes("shmpool_buff_hdr",284,"type")="uint64_t"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.wc_blocked_tp_tend_crbtmismatch3_tn")=284
	Set gtmtypes("shmpool_buff_hdr",285,"name")="shmpool_buff_hdr.shadow_file_header.wc_blocked_wcs_wtstart_bad_cr_tn"
	Set gtmtypes("shmpool_buff_hdr",285,"off")=3120
	Set gtmtypes("shmpool_buff_hdr",285,"len")=8
	Set gtmtypes("shmpool_buff_hdr",285,"type")="uint64_t"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.wc_blocked_wcs_wtstart_bad_cr_tn")=285
	Set gtmtypes("shmpool_buff_hdr",286,"name")="shmpool_buff_hdr.shadow_file_header.wc_blocked_wcs_wtfini_bad_cr_tn"
	Set gtmtypes("shmpool_buff_hdr",286,"off")=3128
	Set gtmtypes("shmpool_buff_hdr",286,"len")=8
	Set gtmtypes("shmpool_buff_hdr",286,"type")="uint64_t"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.wc_blocked_wcs_wtfini_bad_cr_tn")=286
	Set gtmtypes("shmpool_buff_hdr",287,"name")="shmpool_buff_hdr.shadow_file_header.wc_blocked_bt_get_tn"
	Set gtmtypes("shmpool_buff_hdr",287,"off")=3136
	Set gtmtypes("shmpool_buff_hdr",287,"len")=8
	Set gtmtypes("shmpool_buff_hdr",287,"type")="uint64_t"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.wc_blocked_bt_get_tn")=287
	Set gtmtypes("shmpool_buff_hdr",288,"name")="shmpool_buff_hdr.shadow_file_header.wc_blocked_wcs_cdb_sc_final_retry_tn"
	Set gtmtypes("shmpool_buff_hdr",288,"off")=3144
	Set gtmtypes("shmpool_buff_hdr",288,"len")=8
	Set gtmtypes("shmpool_buff_hdr",288,"type")="uint64_t"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.wc_blocked_wcs_cdb_sc_final_retry_tn")=288
	Set gtmtypes("shmpool_buff_hdr",289,"name")="shmpool_buff_hdr.shadow_file_header.wcb_bg_update_lckfail1_tn"
	Set gtmtypes("shmpool_buff_hdr",289,"off")=3152
	Set gtmtypes("shmpool_buff_hdr",289,"len")=8
	Set gtmtypes("shmpool_buff_hdr",289,"type")="uint64_t"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.wcb_bg_update_lckfail1_tn")=289
	Set gtmtypes("shmpool_buff_hdr",290,"name")="shmpool_buff_hdr.shadow_file_header.wcb_bg_update_lckfail2_tn"
	Set gtmtypes("shmpool_buff_hdr",290,"off")=3160
	Set gtmtypes("shmpool_buff_hdr",290,"len")=8
	Set gtmtypes("shmpool_buff_hdr",290,"type")="uint64_t"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.wcb_bg_update_lckfail2_tn")=290
	Set gtmtypes("shmpool_buff_hdr",291,"name")="shmpool_buff_hdr.shadow_file_header.wcb_wtstart_lckfail1_tn"
	Set gtmtypes("shmpool_buff_hdr",291,"off")=3168
	Set gtmtypes("shmpool_buff_hdr",291,"len")=8
	Set gtmtypes("shmpool_buff_hdr",291,"type")="uint64_t"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.wcb_wtstart_lckfail1_tn")=291
	Set gtmtypes("shmpool_buff_hdr",292,"name")="shmpool_buff_hdr.shadow_file_header.wcb_wtstart_lckfail2_tn"
	Set gtmtypes("shmpool_buff_hdr",292,"off")=3176
	Set gtmtypes("shmpool_buff_hdr",292,"len")=8
	Set gtmtypes("shmpool_buff_hdr",292,"type")="uint64_t"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.wcb_wtstart_lckfail2_tn")=292
	Set gtmtypes("shmpool_buff_hdr",293,"name")="shmpool_buff_hdr.shadow_file_header.wcb_wtstart_lckfail3_tn"
	Set gtmtypes("shmpool_buff_hdr",293,"off")=3184
	Set gtmtypes("shmpool_buff_hdr",293,"len")=8
	Set gtmtypes("shmpool_buff_hdr",293,"type")="uint64_t"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.wcb_wtstart_lckfail3_tn")=293
	Set gtmtypes("shmpool_buff_hdr",294,"name")="shmpool_buff_hdr.shadow_file_header.wcb_wtstart_lckfail4_tn"
	Set gtmtypes("shmpool_buff_hdr",294,"off")=3192
	Set gtmtypes("shmpool_buff_hdr",294,"len")=8
	Set gtmtypes("shmpool_buff_hdr",294,"type")="uint64_t"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.wcb_wtstart_lckfail4_tn")=294
	Set gtmtypes("shmpool_buff_hdr",295,"name")="shmpool_buff_hdr.shadow_file_header.wcb_wtfini_lckfail1_tn"
	Set gtmtypes("shmpool_buff_hdr",295,"off")=3200
	Set gtmtypes("shmpool_buff_hdr",295,"len")=8
	Set gtmtypes("shmpool_buff_hdr",295,"type")="uint64_t"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.wcb_wtfini_lckfail1_tn")=295
	Set gtmtypes("shmpool_buff_hdr",296,"name")="shmpool_buff_hdr.shadow_file_header.wcb_wtfini_lckfail2_tn"
	Set gtmtypes("shmpool_buff_hdr",296,"off")=3208
	Set gtmtypes("shmpool_buff_hdr",296,"len")=8
	Set gtmtypes("shmpool_buff_hdr",296,"type")="uint64_t"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.wcb_wtfini_lckfail2_tn")=296
	Set gtmtypes("shmpool_buff_hdr",297,"name")="shmpool_buff_hdr.shadow_file_header.wcb_wtfini_lckfail3_tn"
	Set gtmtypes("shmpool_buff_hdr",297,"off")=3216
	Set gtmtypes("shmpool_buff_hdr",297,"len")=8
	Set gtmtypes("shmpool_buff_hdr",297,"type")="uint64_t"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.wcb_wtfini_lckfail3_tn")=297
	Set gtmtypes("shmpool_buff_hdr",298,"name")="shmpool_buff_hdr.shadow_file_header.wcb_wtfini_lckfail4_tn"
	Set gtmtypes("shmpool_buff_hdr",298,"off")=3224
	Set gtmtypes("shmpool_buff_hdr",298,"len")=8
	Set gtmtypes("shmpool_buff_hdr",298,"type")="uint64_t"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.wcb_wtfini_lckfail4_tn")=298
	Set gtmtypes("shmpool_buff_hdr",299,"name")="shmpool_buff_hdr.shadow_file_header.wcb_t_end_sysops_dirtystuck1_tn"
	Set gtmtypes("shmpool_buff_hdr",299,"off")=3232
	Set gtmtypes("shmpool_buff_hdr",299,"len")=8
	Set gtmtypes("shmpool_buff_hdr",299,"type")="uint64_t"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.wcb_t_end_sysops_dirtystuck1_tn")=299
	Set gtmtypes("shmpool_buff_hdr",300,"name")="shmpool_buff_hdr.shadow_file_header.wcb_t_end_sysops_dirtystuck2_tn"
	Set gtmtypes("shmpool_buff_hdr",300,"off")=3240
	Set gtmtypes("shmpool_buff_hdr",300,"len")=8
	Set gtmtypes("shmpool_buff_hdr",300,"type")="uint64_t"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.wcb_t_end_sysops_dirtystuck2_tn")=300
	Set gtmtypes("shmpool_buff_hdr",301,"name")="shmpool_buff_hdr.shadow_file_header.wcb_secshr_db_clnup_wbuf_dqd_tn"
	Set gtmtypes("shmpool_buff_hdr",301,"off")=3248
	Set gtmtypes("shmpool_buff_hdr",301,"len")=8
	Set gtmtypes("shmpool_buff_hdr",301,"type")="uint64_t"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.wcb_secshr_db_clnup_wbuf_dqd_tn")=301
	Set gtmtypes("shmpool_buff_hdr",302,"name")="shmpool_buff_hdr.shadow_file_header.dwngrd_refmts_syncio_tn"
	Set gtmtypes("shmpool_buff_hdr",302,"off")=3256
	Set gtmtypes("shmpool_buff_hdr",302,"len")=8
	Set gtmtypes("shmpool_buff_hdr",302,"type")="uint64_t"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.dwngrd_refmts_syncio_tn")=302
	Set gtmtypes("shmpool_buff_hdr",303,"name")="shmpool_buff_hdr.shadow_file_header.dwngrd_refmts_asyncio_tn"
	Set gtmtypes("shmpool_buff_hdr",303,"off")=3264
	Set gtmtypes("shmpool_buff_hdr",303,"len")=8
	Set gtmtypes("shmpool_buff_hdr",303,"type")="uint64_t"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.dwngrd_refmts_asyncio_tn")=303
	Set gtmtypes("shmpool_buff_hdr",304,"name")="shmpool_buff_hdr.shadow_file_header.shmpool_refmt_harvests_tn"
	Set gtmtypes("shmpool_buff_hdr",304,"off")=3272
	Set gtmtypes("shmpool_buff_hdr",304,"len")=8
	Set gtmtypes("shmpool_buff_hdr",304,"type")="uint64_t"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.shmpool_refmt_harvests_tn")=304
	Set gtmtypes("shmpool_buff_hdr",305,"name")="shmpool_buff_hdr.shadow_file_header.shmpool_recovery_tn"
	Set gtmtypes("shmpool_buff_hdr",305,"off")=3280
	Set gtmtypes("shmpool_buff_hdr",305,"len")=8
	Set gtmtypes("shmpool_buff_hdr",305,"type")="uint64_t"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.shmpool_recovery_tn")=305
	Set gtmtypes("shmpool_buff_hdr",306,"name")="shmpool_buff_hdr.shadow_file_header.shmpool_blkd_by_sdc_tn"
	Set gtmtypes("shmpool_buff_hdr",306,"off")=3288
	Set gtmtypes("shmpool_buff_hdr",306,"len")=8
	Set gtmtypes("shmpool_buff_hdr",306,"type")="uint64_t"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.shmpool_blkd_by_sdc_tn")=306
	Set gtmtypes("shmpool_buff_hdr",307,"name")="shmpool_buff_hdr.shadow_file_header.shmpool_alloc_bbflush_tn"
	Set gtmtypes("shmpool_buff_hdr",307,"off")=3296
	Set gtmtypes("shmpool_buff_hdr",307,"len")=8
	Set gtmtypes("shmpool_buff_hdr",307,"type")="uint64_t"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.shmpool_alloc_bbflush_tn")=307
	Set gtmtypes("shmpool_buff_hdr",308,"name")="shmpool_buff_hdr.shadow_file_header.refmt_hvst_blk_released_replaced_tn"
	Set gtmtypes("shmpool_buff_hdr",308,"off")=3304
	Set gtmtypes("shmpool_buff_hdr",308,"len")=8
	Set gtmtypes("shmpool_buff_hdr",308,"type")="uint64_t"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.refmt_hvst_blk_released_replaced_tn")=308
	Set gtmtypes("shmpool_buff_hdr",309,"name")="shmpool_buff_hdr.shadow_file_header.refmt_hvst_blk_released_io_complete_tn"
	Set gtmtypes("shmpool_buff_hdr",309,"off")=3312
	Set gtmtypes("shmpool_buff_hdr",309,"len")=8
	Set gtmtypes("shmpool_buff_hdr",309,"type")="uint64_t"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.refmt_hvst_blk_released_io_complete_tn")=309
	Set gtmtypes("shmpool_buff_hdr",310,"name")="shmpool_buff_hdr.shadow_file_header.refmt_hvst_blk_kept_tn"
	Set gtmtypes("shmpool_buff_hdr",310,"off")=3320
	Set gtmtypes("shmpool_buff_hdr",310,"len")=8
	Set gtmtypes("shmpool_buff_hdr",310,"type")="uint64_t"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.refmt_hvst_blk_kept_tn")=310
	Set gtmtypes("shmpool_buff_hdr",311,"name")="shmpool_buff_hdr.shadow_file_header.refmt_hvst_blk_ignored_tn"
	Set gtmtypes("shmpool_buff_hdr",311,"off")=3328
	Set gtmtypes("shmpool_buff_hdr",311,"len")=8
	Set gtmtypes("shmpool_buff_hdr",311,"type")="uint64_t"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.refmt_hvst_blk_ignored_tn")=311
	Set gtmtypes("shmpool_buff_hdr",312,"name")="shmpool_buff_hdr.shadow_file_header.refmt_blk_chk_blk_freed_tn"
	Set gtmtypes("shmpool_buff_hdr",312,"off")=3336
	Set gtmtypes("shmpool_buff_hdr",312,"len")=8
	Set gtmtypes("shmpool_buff_hdr",312,"type")="uint64_t"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.refmt_blk_chk_blk_freed_tn")=312
	Set gtmtypes("shmpool_buff_hdr",313,"name")="shmpool_buff_hdr.shadow_file_header.refmt_blk_chk_blk_kept_tn"
	Set gtmtypes("shmpool_buff_hdr",313,"off")=3344
	Set gtmtypes("shmpool_buff_hdr",313,"len")=8
	Set gtmtypes("shmpool_buff_hdr",313,"type")="uint64_t"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.refmt_blk_chk_blk_kept_tn")=313
	Set gtmtypes("shmpool_buff_hdr",314,"name")="shmpool_buff_hdr.shadow_file_header.active_lvl_trigger_tn"
	Set gtmtypes("shmpool_buff_hdr",314,"off")=3352
	Set gtmtypes("shmpool_buff_hdr",314,"len")=8
	Set gtmtypes("shmpool_buff_hdr",314,"type")="uint64_t"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.active_lvl_trigger_tn")=314
	Set gtmtypes("shmpool_buff_hdr",315,"name")="shmpool_buff_hdr.shadow_file_header.new_buff_tn"
	Set gtmtypes("shmpool_buff_hdr",315,"off")=3360
	Set gtmtypes("shmpool_buff_hdr",315,"len")=8
	Set gtmtypes("shmpool_buff_hdr",315,"type")="uint64_t"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.new_buff_tn")=315
	Set gtmtypes("shmpool_buff_hdr",316,"name")="shmpool_buff_hdr.shadow_file_header.phase2_commit_wait_sleep_in_crit_tn"
	Set gtmtypes("shmpool_buff_hdr",316,"off")=3368
	Set gtmtypes("shmpool_buff_hdr",316,"len")=8
	Set gtmtypes("shmpool_buff_hdr",316,"type")="uint64_t"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.phase2_commit_wait_sleep_in_crit_tn")=316
	Set gtmtypes("shmpool_buff_hdr",317,"name")="shmpool_buff_hdr.shadow_file_header.phase2_commit_wait_sleep_no_crit_tn"
	Set gtmtypes("shmpool_buff_hdr",317,"off")=3376
	Set gtmtypes("shmpool_buff_hdr",317,"len")=8
	Set gtmtypes("shmpool_buff_hdr",317,"type")="uint64_t"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.phase2_commit_wait_sleep_no_crit_tn")=317
	Set gtmtypes("shmpool_buff_hdr",318,"name")="shmpool_buff_hdr.shadow_file_header.phase2_commit_wait_pidcnt_tn"
	Set gtmtypes("shmpool_buff_hdr",318,"off")=3384
	Set gtmtypes("shmpool_buff_hdr",318,"len")=8
	Set gtmtypes("shmpool_buff_hdr",318,"type")="uint64_t"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.phase2_commit_wait_pidcnt_tn")=318
	Set gtmtypes("shmpool_buff_hdr",319,"name")="shmpool_buff_hdr.shadow_file_header.wcb_t_end_sysops_intend_wait_tn"
	Set gtmtypes("shmpool_buff_hdr",319,"off")=3392
	Set gtmtypes("shmpool_buff_hdr",319,"len")=8
	Set gtmtypes("shmpool_buff_hdr",319,"type")="uint64_t"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.wcb_t_end_sysops_intend_wait_tn")=319
	Set gtmtypes("shmpool_buff_hdr",320,"name")="shmpool_buff_hdr.shadow_file_header.wcb_secshr_db_clnup_phase2_clnup_tn"
	Set gtmtypes("shmpool_buff_hdr",320,"off")=3400
	Set gtmtypes("shmpool_buff_hdr",320,"len")=8
	Set gtmtypes("shmpool_buff_hdr",320,"type")="uint64_t"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.wcb_secshr_db_clnup_phase2_clnup_tn")=320
	Set gtmtypes("shmpool_buff_hdr",321,"name")="shmpool_buff_hdr.shadow_file_header.wcb_phase2_commit_wait_tn"
	Set gtmtypes("shmpool_buff_hdr",321,"off")=3408
	Set gtmtypes("shmpool_buff_hdr",321,"len")=8
	Set gtmtypes("shmpool_buff_hdr",321,"type")="uint64_t"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.wcb_phase2_commit_wait_tn")=321
	Set gtmtypes("shmpool_buff_hdr",322,"name")="shmpool_buff_hdr.shadow_file_header.recompute_upd_array_calls_tn"
	Set gtmtypes("shmpool_buff_hdr",322,"off")=3416
	Set gtmtypes("shmpool_buff_hdr",322,"len")=8
	Set gtmtypes("shmpool_buff_hdr",322,"type")="uint64_t"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.recompute_upd_array_calls_tn")=322
	Set gtmtypes("shmpool_buff_hdr",323,"name")="shmpool_buff_hdr.shadow_file_header.recompute_upd_array_rip_tn"
	Set gtmtypes("shmpool_buff_hdr",323,"off")=3424
	Set gtmtypes("shmpool_buff_hdr",323,"len")=8
	Set gtmtypes("shmpool_buff_hdr",323,"type")="uint64_t"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.recompute_upd_array_rip_tn")=323
	Set gtmtypes("shmpool_buff_hdr",324,"name")="shmpool_buff_hdr.shadow_file_header.recompute_upd_array_in_tend_tn"
	Set gtmtypes("shmpool_buff_hdr",324,"off")=3432
	Set gtmtypes("shmpool_buff_hdr",324,"len")=8
	Set gtmtypes("shmpool_buff_hdr",324,"type")="uint64_t"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.recompute_upd_array_in_tend_tn")=324
	Set gtmtypes("shmpool_buff_hdr",325,"name")="shmpool_buff_hdr.shadow_file_header.recompute_upd_array_search_blk_tn"
	Set gtmtypes("shmpool_buff_hdr",325,"off")=3440
	Set gtmtypes("shmpool_buff_hdr",325,"len")=8
	Set gtmtypes("shmpool_buff_hdr",325,"type")="uint64_t"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.recompute_upd_array_search_blk_tn")=325
	Set gtmtypes("shmpool_buff_hdr",326,"name")="shmpool_buff_hdr.shadow_file_header.recompute_upd_array_new_rec_tn"
	Set gtmtypes("shmpool_buff_hdr",326,"off")=3448
	Set gtmtypes("shmpool_buff_hdr",326,"len")=8
	Set gtmtypes("shmpool_buff_hdr",326,"type")="uint64_t"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.recompute_upd_array_new_rec_tn")=326
	Set gtmtypes("shmpool_buff_hdr",327,"name")="shmpool_buff_hdr.shadow_file_header.recompute_upd_array_rec_size_tn"
	Set gtmtypes("shmpool_buff_hdr",327,"off")=3456
	Set gtmtypes("shmpool_buff_hdr",327,"len")=8
	Set gtmtypes("shmpool_buff_hdr",327,"type")="uint64_t"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.recompute_upd_array_rec_size_tn")=327
	Set gtmtypes("shmpool_buff_hdr",328,"name")="shmpool_buff_hdr.shadow_file_header.recompute_upd_array_rec_cmpc_tn"
	Set gtmtypes("shmpool_buff_hdr",328,"off")=3464
	Set gtmtypes("shmpool_buff_hdr",328,"len")=8
	Set gtmtypes("shmpool_buff_hdr",328,"type")="uint64_t"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.recompute_upd_array_rec_cmpc_tn")=328
	Set gtmtypes("shmpool_buff_hdr",329,"name")="shmpool_buff_hdr.shadow_file_header.recompute_upd_array_blk_fini_tn"
	Set gtmtypes("shmpool_buff_hdr",329,"off")=3472
	Set gtmtypes("shmpool_buff_hdr",329,"len")=8
	Set gtmtypes("shmpool_buff_hdr",329,"type")="uint64_t"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.recompute_upd_array_blk_fini_tn")=329
	Set gtmtypes("shmpool_buff_hdr",330,"name")="shmpool_buff_hdr.shadow_file_header.recompute_upd_array_blk_split_tn"
	Set gtmtypes("shmpool_buff_hdr",330,"off")=3480
	Set gtmtypes("shmpool_buff_hdr",330,"len")=8
	Set gtmtypes("shmpool_buff_hdr",330,"type")="uint64_t"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.recompute_upd_array_blk_split_tn")=330
	Set gtmtypes("shmpool_buff_hdr",331,"name")="shmpool_buff_hdr.shadow_file_header.unused_filler_0135_tn"
	Set gtmtypes("shmpool_buff_hdr",331,"off")=3488
	Set gtmtypes("shmpool_buff_hdr",331,"len")=8
	Set gtmtypes("shmpool_buff_hdr",331,"type")="uint64_t"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.unused_filler_0135_tn")=331
	Set gtmtypes("shmpool_buff_hdr",332,"name")="shmpool_buff_hdr.shadow_file_header.t_qread_ripsleep_nblks_tn"
	Set gtmtypes("shmpool_buff_hdr",332,"off")=3496
	Set gtmtypes("shmpool_buff_hdr",332,"len")=8
	Set gtmtypes("shmpool_buff_hdr",332,"type")="uint64_t"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.t_qread_ripsleep_nblks_tn")=332
	Set gtmtypes("shmpool_buff_hdr",333,"name")="shmpool_buff_hdr.shadow_file_header.wcs_wtfini_invoked_tn"
	Set gtmtypes("shmpool_buff_hdr",333,"off")=3504
	Set gtmtypes("shmpool_buff_hdr",333,"len")=8
	Set gtmtypes("shmpool_buff_hdr",333,"type")="uint64_t"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.wcs_wtfini_invoked_tn")=333
	Set gtmtypes("shmpool_buff_hdr",334,"name")="shmpool_buff_hdr.shadow_file_header.wcs_wterror_invoked_tn"
	Set gtmtypes("shmpool_buff_hdr",334,"off")=3512
	Set gtmtypes("shmpool_buff_hdr",334,"len")=8
	Set gtmtypes("shmpool_buff_hdr",334,"type")="uint64_t"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.wcs_wterror_invoked_tn")=334
	Set gtmtypes("shmpool_buff_hdr",335,"name")="shmpool_buff_hdr.shadow_file_header.wcs_wt_restart_invoked_tn"
	Set gtmtypes("shmpool_buff_hdr",335,"off")=3520
	Set gtmtypes("shmpool_buff_hdr",335,"len")=8
	Set gtmtypes("shmpool_buff_hdr",335,"type")="uint64_t"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.wcs_wt_restart_invoked_tn")=335
	Set gtmtypes("shmpool_buff_hdr",336,"name")="shmpool_buff_hdr.shadow_file_header.wcs_wt_restart_eagain_tn"
	Set gtmtypes("shmpool_buff_hdr",336,"off")=3528
	Set gtmtypes("shmpool_buff_hdr",336,"len")=8
	Set gtmtypes("shmpool_buff_hdr",336,"type")="uint64_t"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.wcs_wt_restart_eagain_tn")=336
	Set gtmtypes("shmpool_buff_hdr",337,"name")="shmpool_buff_hdr.shadow_file_header.wcs_wt_restart_reinsert_tn"
	Set gtmtypes("shmpool_buff_hdr",337,"off")=3536
	Set gtmtypes("shmpool_buff_hdr",337,"len")=8
	Set gtmtypes("shmpool_buff_hdr",337,"type")="uint64_t"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.wcs_wt_restart_reinsert_tn")=337
	Set gtmtypes("shmpool_buff_hdr",338,"name")="shmpool_buff_hdr.shadow_file_header.wcs_wtstart_eagain_tn"
	Set gtmtypes("shmpool_buff_hdr",338,"off")=3544
	Set gtmtypes("shmpool_buff_hdr",338,"len")=8
	Set gtmtypes("shmpool_buff_hdr",338,"type")="uint64_t"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.wcs_wtstart_eagain_tn")=338
	Set gtmtypes("shmpool_buff_hdr",339,"name")="shmpool_buff_hdr.shadow_file_header.wcs_wtstart_eagain_incrit_tn"
	Set gtmtypes("shmpool_buff_hdr",339,"off")=3552
	Set gtmtypes("shmpool_buff_hdr",339,"len")=8
	Set gtmtypes("shmpool_buff_hdr",339,"type")="uint64_t"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.wcs_wtstart_eagain_incrit_tn")=339
	Set gtmtypes("shmpool_buff_hdr",340,"name")="shmpool_buff_hdr.shadow_file_header.async_restart_eagain_tn"
	Set gtmtypes("shmpool_buff_hdr",340,"off")=3560
	Set gtmtypes("shmpool_buff_hdr",340,"len")=8
	Set gtmtypes("shmpool_buff_hdr",340,"type")="uint64_t"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.async_restart_eagain_tn")=340
	Set gtmtypes("shmpool_buff_hdr",341,"name")="shmpool_buff_hdr.shadow_file_header.jnlbuff_phs2cmt_array_full_tn"
	Set gtmtypes("shmpool_buff_hdr",341,"off")=3568
	Set gtmtypes("shmpool_buff_hdr",341,"len")=8
	Set gtmtypes("shmpool_buff_hdr",341,"type")="uint64_t"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.jnlbuff_phs2cmt_array_full_tn")=341
	Set gtmtypes("shmpool_buff_hdr",342,"name")="shmpool_buff_hdr.shadow_file_header.jnl_phase2_cleanup_if_possible_tn"
	Set gtmtypes("shmpool_buff_hdr",342,"off")=3576
	Set gtmtypes("shmpool_buff_hdr",342,"len")=8
	Set gtmtypes("shmpool_buff_hdr",342,"type")="uint64_t"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.jnl_phase2_cleanup_if_possible_tn")=342
	Set gtmtypes("shmpool_buff_hdr",343,"name")="shmpool_buff_hdr.shadow_file_header.jnlbuff_phs2cmt_isprcalv_tn"
	Set gtmtypes("shmpool_buff_hdr",343,"off")=3584
	Set gtmtypes("shmpool_buff_hdr",343,"len")=8
	Set gtmtypes("shmpool_buff_hdr",343,"type")="uint64_t"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.jnlbuff_phs2cmt_isprcalv_tn")=343
	Set gtmtypes("shmpool_buff_hdr",344,"name")="shmpool_buff_hdr.shadow_file_header.recompute_upd_array_paidoff_tn"
	Set gtmtypes("shmpool_buff_hdr",344,"off")=3592
	Set gtmtypes("shmpool_buff_hdr",344,"len")=8
	Set gtmtypes("shmpool_buff_hdr",344,"type")="uint64_t"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.recompute_upd_array_paidoff_tn")=344
	Set gtmtypes("shmpool_buff_hdr",345,"name")="shmpool_buff_hdr.shadow_file_header.bg_trc_rec_tn_filler"
	Set gtmtypes("shmpool_buff_hdr",345,"off")=3600
	Set gtmtypes("shmpool_buff_hdr",345,"len")=16
	Set gtmtypes("shmpool_buff_hdr",345,"type")="char"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.bg_trc_rec_tn_filler")=345
	Set gtmtypes("shmpool_buff_hdr",346,"name")="shmpool_buff_hdr.shadow_file_header.total_buffer_flush_cntr"
	Set gtmtypes("shmpool_buff_hdr",346,"off")=3616
	Set gtmtypes("shmpool_buff_hdr",346,"len")=4
	Set gtmtypes("shmpool_buff_hdr",346,"type")="int"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.total_buffer_flush_cntr")=346
	Set gtmtypes("shmpool_buff_hdr",347,"name")="shmpool_buff_hdr.shadow_file_header.bufct_buffer_flush_cntr"
	Set gtmtypes("shmpool_buff_hdr",347,"off")=3620
	Set gtmtypes("shmpool_buff_hdr",347,"len")=4
	Set gtmtypes("shmpool_buff_hdr",347,"type")="int"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.bufct_buffer_flush_cntr")=347
	Set gtmtypes("shmpool_buff_hdr",348,"name")="shmpool_buff_hdr.shadow_file_header.bufct_buffer_flush_loop_cntr"
	Set gtmtypes("shmpool_buff_hdr",348,"off")=3624
	Set gtmtypes("shmpool_buff_hdr",348,"len")=4
	Set gtmtypes("shmpool_buff_hdr",348,"type")="int"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.bufct_buffer_flush_loop_cntr")=348
	Set gtmtypes("shmpool_buff_hdr",349,"name")="shmpool_buff_hdr.shadow_file_header.wrt_calls_cntr"
	Set gtmtypes("shmpool_buff_hdr",349,"off")=3628
	Set gtmtypes("shmpool_buff_hdr",349,"len")=4
	Set gtmtypes("shmpool_buff_hdr",349,"type")="int"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.wrt_calls_cntr")=349
	Set gtmtypes("shmpool_buff_hdr",350,"name")="shmpool_buff_hdr.shadow_file_header.wrt_count_cntr"
	Set gtmtypes("shmpool_buff_hdr",350,"off")=3632
	Set gtmtypes("shmpool_buff_hdr",350,"len")=4
	Set gtmtypes("shmpool_buff_hdr",350,"type")="int"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.wrt_count_cntr")=350
	Set gtmtypes("shmpool_buff_hdr",351,"name")="shmpool_buff_hdr.shadow_file_header.wrt_blocked_cntr"
	Set gtmtypes("shmpool_buff_hdr",351,"off")=3636
	Set gtmtypes("shmpool_buff_hdr",351,"len")=4
	Set gtmtypes("shmpool_buff_hdr",351,"type")="int"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.wrt_blocked_cntr")=351
	Set gtmtypes("shmpool_buff_hdr",352,"name")="shmpool_buff_hdr.shadow_file_header.wrt_busy_cntr"
	Set gtmtypes("shmpool_buff_hdr",352,"off")=3640
	Set gtmtypes("shmpool_buff_hdr",352,"len")=4
	Set gtmtypes("shmpool_buff_hdr",352,"type")="int"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.wrt_busy_cntr")=352
	Set gtmtypes("shmpool_buff_hdr",353,"name")="shmpool_buff_hdr.shadow_file_header.wrt_noblks_wrtn_cntr"
	Set gtmtypes("shmpool_buff_hdr",353,"off")=3644
	Set gtmtypes("shmpool_buff_hdr",353,"len")=4
	Set gtmtypes("shmpool_buff_hdr",353,"type")="int"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.wrt_noblks_wrtn_cntr")=353
	Set gtmtypes("shmpool_buff_hdr",354,"name")="shmpool_buff_hdr.shadow_file_header.wcb_gds_rundown2_cntr"
	Set gtmtypes("shmpool_buff_hdr",354,"off")=3648
	Set gtmtypes("shmpool_buff_hdr",354,"len")=4
	Set gtmtypes("shmpool_buff_hdr",354,"type")="int"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.wcb_gds_rundown2_cntr")=354
	Set gtmtypes("shmpool_buff_hdr",355,"name")="shmpool_buff_hdr.shadow_file_header.reserved_bgtrcrec2_cntr"
	Set gtmtypes("shmpool_buff_hdr",355,"off")=3652
	Set gtmtypes("shmpool_buff_hdr",355,"len")=4
	Set gtmtypes("shmpool_buff_hdr",355,"type")="int"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.reserved_bgtrcrec2_cntr")=355
	Set gtmtypes("shmpool_buff_hdr",356,"name")="shmpool_buff_hdr.shadow_file_header.reserved_bgtrcrec3_cntr"
	Set gtmtypes("shmpool_buff_hdr",356,"off")=3656
	Set gtmtypes("shmpool_buff_hdr",356,"len")=4
	Set gtmtypes("shmpool_buff_hdr",356,"type")="int"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.reserved_bgtrcrec3_cntr")=356
	Set gtmtypes("shmpool_buff_hdr",357,"name")="shmpool_buff_hdr.shadow_file_header.lost_block_recovery_cntr"
	Set gtmtypes("shmpool_buff_hdr",357,"off")=3660
	Set gtmtypes("shmpool_buff_hdr",357,"len")=4
	Set gtmtypes("shmpool_buff_hdr",357,"type")="int"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.lost_block_recovery_cntr")=357
	Set gtmtypes("shmpool_buff_hdr",358,"name")="shmpool_buff_hdr.shadow_file_header.wc_blocked_onln_rlbk_cntr"
	Set gtmtypes("shmpool_buff_hdr",358,"off")=3664
	Set gtmtypes("shmpool_buff_hdr",358,"len")=4
	Set gtmtypes("shmpool_buff_hdr",358,"type")="int"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.wc_blocked_onln_rlbk_cntr")=358
	Set gtmtypes("shmpool_buff_hdr",359,"name")="shmpool_buff_hdr.shadow_file_header.stale_cntr"
	Set gtmtypes("shmpool_buff_hdr",359,"off")=3668
	Set gtmtypes("shmpool_buff_hdr",359,"len")=4
	Set gtmtypes("shmpool_buff_hdr",359,"type")="int"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.stale_cntr")=359
	Set gtmtypes("shmpool_buff_hdr",360,"name")="shmpool_buff_hdr.shadow_file_header.starved_cntr"
	Set gtmtypes("shmpool_buff_hdr",360,"off")=3672
	Set gtmtypes("shmpool_buff_hdr",360,"len")=4
	Set gtmtypes("shmpool_buff_hdr",360,"type")="int"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.starved_cntr")=360
	Set gtmtypes("shmpool_buff_hdr",361,"name")="shmpool_buff_hdr.shadow_file_header.stale_timer_started_cntr"
	Set gtmtypes("shmpool_buff_hdr",361,"off")=3676
	Set gtmtypes("shmpool_buff_hdr",361,"len")=4
	Set gtmtypes("shmpool_buff_hdr",361,"type")="int"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.stale_timer_started_cntr")=361
	Set gtmtypes("shmpool_buff_hdr",362,"name")="shmpool_buff_hdr.shadow_file_header.stale_timer_pop_cntr"
	Set gtmtypes("shmpool_buff_hdr",362,"off")=3680
	Set gtmtypes("shmpool_buff_hdr",362,"len")=4
	Set gtmtypes("shmpool_buff_hdr",362,"type")="int"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.stale_timer_pop_cntr")=362
	Set gtmtypes("shmpool_buff_hdr",363,"name")="shmpool_buff_hdr.shadow_file_header.stale_process_defer_cntr"
	Set gtmtypes("shmpool_buff_hdr",363,"off")=3684
	Set gtmtypes("shmpool_buff_hdr",363,"len")=4
	Set gtmtypes("shmpool_buff_hdr",363,"type")="int"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.stale_process_defer_cntr")=363
	Set gtmtypes("shmpool_buff_hdr",364,"name")="shmpool_buff_hdr.shadow_file_header.stale_defer_processed_cntr"
	Set gtmtypes("shmpool_buff_hdr",364,"off")=3688
	Set gtmtypes("shmpool_buff_hdr",364,"len")=4
	Set gtmtypes("shmpool_buff_hdr",364,"type")="int"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.stale_defer_processed_cntr")=364
	Set gtmtypes("shmpool_buff_hdr",365,"name")="shmpool_buff_hdr.shadow_file_header.spcfc_buffer_flush_cntr"
	Set gtmtypes("shmpool_buff_hdr",365,"off")=3692
	Set gtmtypes("shmpool_buff_hdr",365,"len")=4
	Set gtmtypes("shmpool_buff_hdr",365,"type")="int"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.spcfc_buffer_flush_cntr")=365
	Set gtmtypes("shmpool_buff_hdr",366,"name")="shmpool_buff_hdr.shadow_file_header.spcfc_buffer_flush_loop_cntr"
	Set gtmtypes("shmpool_buff_hdr",366,"off")=3696
	Set gtmtypes("shmpool_buff_hdr",366,"len")=4
	Set gtmtypes("shmpool_buff_hdr",366,"type")="int"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.spcfc_buffer_flush_loop_cntr")=366
	Set gtmtypes("shmpool_buff_hdr",367,"name")="shmpool_buff_hdr.shadow_file_header.spcfc_buffer_flush_retries_cntr"
	Set gtmtypes("shmpool_buff_hdr",367,"off")=3700
	Set gtmtypes("shmpool_buff_hdr",367,"len")=4
	Set gtmtypes("shmpool_buff_hdr",367,"type")="int"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.spcfc_buffer_flush_retries_cntr")=367
	Set gtmtypes("shmpool_buff_hdr",368,"name")="shmpool_buff_hdr.shadow_file_header.spcfc_buffer_flushed_during_lockwait_cntr"
	Set gtmtypes("shmpool_buff_hdr",368,"off")=3704
	Set gtmtypes("shmpool_buff_hdr",368,"len")=4
	Set gtmtypes("shmpool_buff_hdr",368,"type")="int"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.spcfc_buffer_flushed_during_lockwait_cntr")=368
	Set gtmtypes("shmpool_buff_hdr",369,"name")="shmpool_buff_hdr.shadow_file_header.tp_crit_retries_cntr"
	Set gtmtypes("shmpool_buff_hdr",369,"off")=3708
	Set gtmtypes("shmpool_buff_hdr",369,"len")=4
	Set gtmtypes("shmpool_buff_hdr",369,"type")="int"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.tp_crit_retries_cntr")=369
	Set gtmtypes("shmpool_buff_hdr",370,"name")="shmpool_buff_hdr.shadow_file_header.db_csh_getn_flush_dirty_cntr"
	Set gtmtypes("shmpool_buff_hdr",370,"off")=3712
	Set gtmtypes("shmpool_buff_hdr",370,"len")=4
	Set gtmtypes("shmpool_buff_hdr",370,"type")="int"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.db_csh_getn_flush_dirty_cntr")=370
	Set gtmtypes("shmpool_buff_hdr",371,"name")="shmpool_buff_hdr.shadow_file_header.db_csh_getn_rip_wait_cntr"
	Set gtmtypes("shmpool_buff_hdr",371,"off")=3716
	Set gtmtypes("shmpool_buff_hdr",371,"len")=4
	Set gtmtypes("shmpool_buff_hdr",371,"type")="int"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.db_csh_getn_rip_wait_cntr")=371
	Set gtmtypes("shmpool_buff_hdr",372,"name")="shmpool_buff_hdr.shadow_file_header.db_csh_getn_buf_owner_stuck_cntr"
	Set gtmtypes("shmpool_buff_hdr",372,"off")=3720
	Set gtmtypes("shmpool_buff_hdr",372,"len")=4
	Set gtmtypes("shmpool_buff_hdr",372,"type")="int"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.db_csh_getn_buf_owner_stuck_cntr")=372
	Set gtmtypes("shmpool_buff_hdr",373,"name")="shmpool_buff_hdr.shadow_file_header.db_csh_getn_out_of_design_cntr"
	Set gtmtypes("shmpool_buff_hdr",373,"off")=3724
	Set gtmtypes("shmpool_buff_hdr",373,"len")=4
	Set gtmtypes("shmpool_buff_hdr",373,"type")="int"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.db_csh_getn_out_of_design_cntr")=373
	Set gtmtypes("shmpool_buff_hdr",374,"name")="shmpool_buff_hdr.shadow_file_header.t_qread_buf_owner_stuck_cntr"
	Set gtmtypes("shmpool_buff_hdr",374,"off")=3728
	Set gtmtypes("shmpool_buff_hdr",374,"len")=4
	Set gtmtypes("shmpool_buff_hdr",374,"type")="int"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.t_qread_buf_owner_stuck_cntr")=374
	Set gtmtypes("shmpool_buff_hdr",375,"name")="shmpool_buff_hdr.shadow_file_header.t_qread_out_of_design_cntr"
	Set gtmtypes("shmpool_buff_hdr",375,"off")=3732
	Set gtmtypes("shmpool_buff_hdr",375,"len")=4
	Set gtmtypes("shmpool_buff_hdr",375,"type")="int"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.t_qread_out_of_design_cntr")=375
	Set gtmtypes("shmpool_buff_hdr",376,"name")="shmpool_buff_hdr.shadow_file_header.unused_filler_0031_cntr"
	Set gtmtypes("shmpool_buff_hdr",376,"off")=3736
	Set gtmtypes("shmpool_buff_hdr",376,"len")=4
	Set gtmtypes("shmpool_buff_hdr",376,"type")="int"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.unused_filler_0031_cntr")=376
	Set gtmtypes("shmpool_buff_hdr",377,"name")="shmpool_buff_hdr.shadow_file_header.mlock_wakeups_cntr"
	Set gtmtypes("shmpool_buff_hdr",377,"off")=3740
	Set gtmtypes("shmpool_buff_hdr",377,"len")=4
	Set gtmtypes("shmpool_buff_hdr",377,"type")="int"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.mlock_wakeups_cntr")=377
	Set gtmtypes("shmpool_buff_hdr",378,"name")="shmpool_buff_hdr.shadow_file_header.wc_blocked_wcs_recover_invoked_cntr"
	Set gtmtypes("shmpool_buff_hdr",378,"off")=3744
	Set gtmtypes("shmpool_buff_hdr",378,"len")=4
	Set gtmtypes("shmpool_buff_hdr",378,"type")="int"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.wc_blocked_wcs_recover_invoked_cntr")=378
	Set gtmtypes("shmpool_buff_hdr",379,"name")="shmpool_buff_hdr.shadow_file_header.wc_blocked_wcs_verify_passed_cntr"
	Set gtmtypes("shmpool_buff_hdr",379,"off")=3748
	Set gtmtypes("shmpool_buff_hdr",379,"len")=4
	Set gtmtypes("shmpool_buff_hdr",379,"type")="int"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.wc_blocked_wcs_verify_passed_cntr")=379
	Set gtmtypes("shmpool_buff_hdr",380,"name")="shmpool_buff_hdr.shadow_file_header.wc_blocked_t_qread_db_csh_getn_invalid_blk_cntr"
	Set gtmtypes("shmpool_buff_hdr",380,"off")=3752
	Set gtmtypes("shmpool_buff_hdr",380,"len")=4
	Set gtmtypes("shmpool_buff_hdr",380,"type")="int"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.wc_blocked_t_qread_db_csh_getn_invalid_blk_cntr")=380
	Set gtmtypes("shmpool_buff_hdr",381,"name")="shmpool_buff_hdr.shadow_file_header.wc_blocked_t_qread_db_csh_get_invalid_blk_cntr"
	Set gtmtypes("shmpool_buff_hdr",381,"off")=3756
	Set gtmtypes("shmpool_buff_hdr",381,"len")=4
	Set gtmtypes("shmpool_buff_hdr",381,"type")="int"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.wc_blocked_t_qread_db_csh_get_invalid_blk_cntr")=381
	Set gtmtypes("shmpool_buff_hdr",382,"name")="shmpool_buff_hdr.shadow_file_header.wc_blocked_db_csh_getn_loopexceed_cntr"
	Set gtmtypes("shmpool_buff_hdr",382,"off")=3760
	Set gtmtypes("shmpool_buff_hdr",382,"len")=4
	Set gtmtypes("shmpool_buff_hdr",382,"type")="int"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.wc_blocked_db_csh_getn_loopexceed_cntr")=382
	Set gtmtypes("shmpool_buff_hdr",383,"name")="shmpool_buff_hdr.shadow_file_header.wc_blocked_db_csh_getn_wcsstarvewrt_cntr"
	Set gtmtypes("shmpool_buff_hdr",383,"off")=3764
	Set gtmtypes("shmpool_buff_hdr",383,"len")=4
	Set gtmtypes("shmpool_buff_hdr",383,"type")="int"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.wc_blocked_db_csh_getn_wcsstarvewrt_cntr")=383
	Set gtmtypes("shmpool_buff_hdr",384,"name")="shmpool_buff_hdr.shadow_file_header.wc_blocked_db_csh_get_cntr"
	Set gtmtypes("shmpool_buff_hdr",384,"off")=3768
	Set gtmtypes("shmpool_buff_hdr",384,"len")=4
	Set gtmtypes("shmpool_buff_hdr",384,"type")="int"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.wc_blocked_db_csh_get_cntr")=384
	Set gtmtypes("shmpool_buff_hdr",385,"name")="shmpool_buff_hdr.shadow_file_header.wc_blocked_tp_tend_wcsgetspace_cntr"
	Set gtmtypes("shmpool_buff_hdr",385,"off")=3772
	Set gtmtypes("shmpool_buff_hdr",385,"len")=4
	Set gtmtypes("shmpool_buff_hdr",385,"type")="int"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.wc_blocked_tp_tend_wcsgetspace_cntr")=385
	Set gtmtypes("shmpool_buff_hdr",386,"name")="shmpool_buff_hdr.shadow_file_header.wc_blocked_tp_tend_t1_cntr"
	Set gtmtypes("shmpool_buff_hdr",386,"off")=3776
	Set gtmtypes("shmpool_buff_hdr",386,"len")=4
	Set gtmtypes("shmpool_buff_hdr",386,"type")="int"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.wc_blocked_tp_tend_t1_cntr")=386
	Set gtmtypes("shmpool_buff_hdr",387,"name")="shmpool_buff_hdr.shadow_file_header.wc_blocked_tp_tend_bitmap_cntr"
	Set gtmtypes("shmpool_buff_hdr",387,"off")=3780
	Set gtmtypes("shmpool_buff_hdr",387,"len")=4
	Set gtmtypes("shmpool_buff_hdr",387,"type")="int"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.wc_blocked_tp_tend_bitmap_cntr")=387
	Set gtmtypes("shmpool_buff_hdr",388,"name")="shmpool_buff_hdr.shadow_file_header.wc_blocked_tp_tend_jnl_cwset_cntr"
	Set gtmtypes("shmpool_buff_hdr",388,"off")=3784
	Set gtmtypes("shmpool_buff_hdr",388,"len")=4
	Set gtmtypes("shmpool_buff_hdr",388,"type")="int"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.wc_blocked_tp_tend_jnl_cwset_cntr")=388
	Set gtmtypes("shmpool_buff_hdr",389,"name")="shmpool_buff_hdr.shadow_file_header.wc_blocked_tp_tend_jnl_wcsflu_cntr"
	Set gtmtypes("shmpool_buff_hdr",389,"off")=3788
	Set gtmtypes("shmpool_buff_hdr",389,"len")=4
	Set gtmtypes("shmpool_buff_hdr",389,"type")="int"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.wc_blocked_tp_tend_jnl_wcsflu_cntr")=389
	Set gtmtypes("shmpool_buff_hdr",390,"name")="shmpool_buff_hdr.shadow_file_header.wc_blocked_t_end_hist_cntr"
	Set gtmtypes("shmpool_buff_hdr",390,"off")=3792
	Set gtmtypes("shmpool_buff_hdr",390,"len")=4
	Set gtmtypes("shmpool_buff_hdr",390,"type")="int"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.wc_blocked_t_end_hist_cntr")=390
	Set gtmtypes("shmpool_buff_hdr",391,"name")="shmpool_buff_hdr.shadow_file_header.wc_blocked_t_end_hist1_nullbt_cntr"
	Set gtmtypes("shmpool_buff_hdr",391,"off")=3796
	Set gtmtypes("shmpool_buff_hdr",391,"len")=4
	Set gtmtypes("shmpool_buff_hdr",391,"type")="int"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.wc_blocked_t_end_hist1_nullbt_cntr")=391
	Set gtmtypes("shmpool_buff_hdr",392,"name")="shmpool_buff_hdr.shadow_file_header.wc_blocked_t_end_hist1_nonnullbt_cntr"
	Set gtmtypes("shmpool_buff_hdr",392,"off")=3800
	Set gtmtypes("shmpool_buff_hdr",392,"len")=4
	Set gtmtypes("shmpool_buff_hdr",392,"type")="int"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.wc_blocked_t_end_hist1_nonnullbt_cntr")=392
	Set gtmtypes("shmpool_buff_hdr",393,"name")="shmpool_buff_hdr.shadow_file_header.wc_blocked_t_end_bitmap_nullbt_cntr"
	Set gtmtypes("shmpool_buff_hdr",393,"off")=3804
	Set gtmtypes("shmpool_buff_hdr",393,"len")=4
	Set gtmtypes("shmpool_buff_hdr",393,"type")="int"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.wc_blocked_t_end_bitmap_nullbt_cntr")=393
	Set gtmtypes("shmpool_buff_hdr",394,"name")="shmpool_buff_hdr.shadow_file_header.wc_blocked_t_end_bitmap_nonnullbt_cntr"
	Set gtmtypes("shmpool_buff_hdr",394,"off")=3808
	Set gtmtypes("shmpool_buff_hdr",394,"len")=4
	Set gtmtypes("shmpool_buff_hdr",394,"type")="int"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.wc_blocked_t_end_bitmap_nonnullbt_cntr")=394
	Set gtmtypes("shmpool_buff_hdr",395,"name")="shmpool_buff_hdr.shadow_file_header.wc_blocked_t_end_jnl_cwset_cntr"
	Set gtmtypes("shmpool_buff_hdr",395,"off")=3812
	Set gtmtypes("shmpool_buff_hdr",395,"len")=4
	Set gtmtypes("shmpool_buff_hdr",395,"type")="int"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.wc_blocked_t_end_jnl_cwset_cntr")=395
	Set gtmtypes("shmpool_buff_hdr",396,"name")="shmpool_buff_hdr.shadow_file_header.wc_blocked_t_end_jnl_wcsflu_cntr"
	Set gtmtypes("shmpool_buff_hdr",396,"off")=3816
	Set gtmtypes("shmpool_buff_hdr",396,"len")=4
	Set gtmtypes("shmpool_buff_hdr",396,"type")="int"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.wc_blocked_t_end_jnl_wcsflu_cntr")=396
	Set gtmtypes("shmpool_buff_hdr",397,"name")="shmpool_buff_hdr.shadow_file_header.unused_filler_0052_cntr"
	Set gtmtypes("shmpool_buff_hdr",397,"off")=3820
	Set gtmtypes("shmpool_buff_hdr",397,"len")=4
	Set gtmtypes("shmpool_buff_hdr",397,"type")="int"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.unused_filler_0052_cntr")=397
	Set gtmtypes("shmpool_buff_hdr",398,"name")="shmpool_buff_hdr.shadow_file_header.wc_blocked_tpckh_hist1_nullbt_cntr"
	Set gtmtypes("shmpool_buff_hdr",398,"off")=3824
	Set gtmtypes("shmpool_buff_hdr",398,"len")=4
	Set gtmtypes("shmpool_buff_hdr",398,"type")="int"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.wc_blocked_tpckh_hist1_nullbt_cntr")=398
	Set gtmtypes("shmpool_buff_hdr",399,"name")="shmpool_buff_hdr.shadow_file_header.wc_blocked_tpckh_hist1_nonnullbt_cntr"
	Set gtmtypes("shmpool_buff_hdr",399,"off")=3828
	Set gtmtypes("shmpool_buff_hdr",399,"len")=4
	Set gtmtypes("shmpool_buff_hdr",399,"type")="int"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.wc_blocked_tpckh_hist1_nonnullbt_cntr")=399
	Set gtmtypes("shmpool_buff_hdr",400,"name")="shmpool_buff_hdr.shadow_file_header.jnl_blocked_writer_lost_cntr"
	Set gtmtypes("shmpool_buff_hdr",400,"off")=3832
	Set gtmtypes("shmpool_buff_hdr",400,"len")=4
	Set gtmtypes("shmpool_buff_hdr",400,"type")="int"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.jnl_blocked_writer_lost_cntr")=400
	Set gtmtypes("shmpool_buff_hdr",401,"name")="shmpool_buff_hdr.shadow_file_header.jnl_blocked_writer_stuck_cntr"
	Set gtmtypes("shmpool_buff_hdr",401,"off")=3836
	Set gtmtypes("shmpool_buff_hdr",401,"len")=4
	Set gtmtypes("shmpool_buff_hdr",401,"type")="int"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.jnl_blocked_writer_stuck_cntr")=401
	Set gtmtypes("shmpool_buff_hdr",402,"name")="shmpool_buff_hdr.shadow_file_header.jnl_blocked_writer_blocked_cntr"
	Set gtmtypes("shmpool_buff_hdr",402,"off")=3840
	Set gtmtypes("shmpool_buff_hdr",402,"len")=4
	Set gtmtypes("shmpool_buff_hdr",402,"type")="int"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.jnl_blocked_writer_blocked_cntr")=402
	Set gtmtypes("shmpool_buff_hdr",403,"name")="shmpool_buff_hdr.shadow_file_header.n_jnl_fsyncs_cntr"
	Set gtmtypes("shmpool_buff_hdr",403,"off")=3844
	Set gtmtypes("shmpool_buff_hdr",403,"len")=4
	Set gtmtypes("shmpool_buff_hdr",403,"type")="int"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.n_jnl_fsyncs_cntr")=403
	Set gtmtypes("shmpool_buff_hdr",404,"name")="shmpool_buff_hdr.shadow_file_header.n_jnl_fsync_tries_cntr"
	Set gtmtypes("shmpool_buff_hdr",404,"off")=3848
	Set gtmtypes("shmpool_buff_hdr",404,"len")=4
	Set gtmtypes("shmpool_buff_hdr",404,"type")="int"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.n_jnl_fsync_tries_cntr")=404
	Set gtmtypes("shmpool_buff_hdr",405,"name")="shmpool_buff_hdr.shadow_file_header.n_jnl_fsync_recovers_cntr"
	Set gtmtypes("shmpool_buff_hdr",405,"off")=3852
	Set gtmtypes("shmpool_buff_hdr",405,"len")=4
	Set gtmtypes("shmpool_buff_hdr",405,"type")="int"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.n_jnl_fsync_recovers_cntr")=405
	Set gtmtypes("shmpool_buff_hdr",406,"name")="shmpool_buff_hdr.shadow_file_header.n_db_fsyncs_cntr"
	Set gtmtypes("shmpool_buff_hdr",406,"off")=3856
	Set gtmtypes("shmpool_buff_hdr",406,"len")=4
	Set gtmtypes("shmpool_buff_hdr",406,"type")="int"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.n_db_fsyncs_cntr")=406
	Set gtmtypes("shmpool_buff_hdr",407,"name")="shmpool_buff_hdr.shadow_file_header.n_db_fsyncs_in_crit_cntr"
	Set gtmtypes("shmpool_buff_hdr",407,"off")=3860
	Set gtmtypes("shmpool_buff_hdr",407,"len")=4
	Set gtmtypes("shmpool_buff_hdr",407,"type")="int"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.n_db_fsyncs_in_crit_cntr")=407
	Set gtmtypes("shmpool_buff_hdr",408,"name")="shmpool_buff_hdr.shadow_file_header.n_dbsync_timers_cntr"
	Set gtmtypes("shmpool_buff_hdr",408,"off")=3864
	Set gtmtypes("shmpool_buff_hdr",408,"len")=4
	Set gtmtypes("shmpool_buff_hdr",408,"type")="int"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.n_dbsync_timers_cntr")=408
	Set gtmtypes("shmpool_buff_hdr",409,"name")="shmpool_buff_hdr.shadow_file_header.n_dbsync_writes_cntr"
	Set gtmtypes("shmpool_buff_hdr",409,"off")=3868
	Set gtmtypes("shmpool_buff_hdr",409,"len")=4
	Set gtmtypes("shmpool_buff_hdr",409,"type")="int"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.n_dbsync_writes_cntr")=409
	Set gtmtypes("shmpool_buff_hdr",410,"name")="shmpool_buff_hdr.shadow_file_header.mutex_queue_full_cntr"
	Set gtmtypes("shmpool_buff_hdr",410,"off")=3872
	Set gtmtypes("shmpool_buff_hdr",410,"len")=4
	Set gtmtypes("shmpool_buff_hdr",410,"type")="int"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.mutex_queue_full_cntr")=410
	Set gtmtypes("shmpool_buff_hdr",411,"name")="shmpool_buff_hdr.shadow_file_header.wcb_bt_put_cntr"
	Set gtmtypes("shmpool_buff_hdr",411,"off")=3876
	Set gtmtypes("shmpool_buff_hdr",411,"len")=4
	Set gtmtypes("shmpool_buff_hdr",411,"type")="int"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.wcb_bt_put_cntr")=411
	Set gtmtypes("shmpool_buff_hdr",412,"name")="shmpool_buff_hdr.shadow_file_header.wcb_mutex_salvage_cntr"
	Set gtmtypes("shmpool_buff_hdr",412,"off")=3880
	Set gtmtypes("shmpool_buff_hdr",412,"len")=4
	Set gtmtypes("shmpool_buff_hdr",412,"type")="int"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.wcb_mutex_salvage_cntr")=412
	Set gtmtypes("shmpool_buff_hdr",413,"name")="shmpool_buff_hdr.shadow_file_header.wcb_wcs_flu0_cntr"
	Set gtmtypes("shmpool_buff_hdr",413,"off")=3884
	Set gtmtypes("shmpool_buff_hdr",413,"len")=4
	Set gtmtypes("shmpool_buff_hdr",413,"type")="int"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.wcb_wcs_flu0_cntr")=413
	Set gtmtypes("shmpool_buff_hdr",414,"name")="shmpool_buff_hdr.shadow_file_header.wcb_t_end_sysops_nocr_invcr_cntr"
	Set gtmtypes("shmpool_buff_hdr",414,"off")=3888
	Set gtmtypes("shmpool_buff_hdr",414,"len")=4
	Set gtmtypes("shmpool_buff_hdr",414,"type")="int"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.wcb_t_end_sysops_nocr_invcr_cntr")=414
	Set gtmtypes("shmpool_buff_hdr",415,"name")="shmpool_buff_hdr.shadow_file_header.wcb_t_end_sysops_cr_invcr_cntr"
	Set gtmtypes("shmpool_buff_hdr",415,"off")=3892
	Set gtmtypes("shmpool_buff_hdr",415,"len")=4
	Set gtmtypes("shmpool_buff_hdr",415,"type")="int"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.wcb_t_end_sysops_cr_invcr_cntr")=415
	Set gtmtypes("shmpool_buff_hdr",416,"name")="shmpool_buff_hdr.shadow_file_header.wcb_t_end_sysops_rip_wait_cntr"
	Set gtmtypes("shmpool_buff_hdr",416,"off")=3896
	Set gtmtypes("shmpool_buff_hdr",416,"len")=4
	Set gtmtypes("shmpool_buff_hdr",416,"type")="int"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.wcb_t_end_sysops_rip_wait_cntr")=416
	Set gtmtypes("shmpool_buff_hdr",417,"name")="shmpool_buff_hdr.shadow_file_header.wcb_t_end_sysops_dirtyripwait_cntr"
	Set gtmtypes("shmpool_buff_hdr",417,"off")=3900
	Set gtmtypes("shmpool_buff_hdr",417,"len")=4
	Set gtmtypes("shmpool_buff_hdr",417,"type")="int"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.wcb_t_end_sysops_dirtyripwait_cntr")=417
	Set gtmtypes("shmpool_buff_hdr",418,"name")="shmpool_buff_hdr.shadow_file_header.wcb_gds_rundown1_cntr"
	Set gtmtypes("shmpool_buff_hdr",418,"off")=3904
	Set gtmtypes("shmpool_buff_hdr",418,"len")=4
	Set gtmtypes("shmpool_buff_hdr",418,"type")="int"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.wcb_gds_rundown1_cntr")=418
	Set gtmtypes("shmpool_buff_hdr",419,"name")="shmpool_buff_hdr.shadow_file_header.wcb_wcs_flu1_cntr"
	Set gtmtypes("shmpool_buff_hdr",419,"off")=3908
	Set gtmtypes("shmpool_buff_hdr",419,"len")=4
	Set gtmtypes("shmpool_buff_hdr",419,"type")="int"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.wcb_wcs_flu1_cntr")=419
	Set gtmtypes("shmpool_buff_hdr",420,"name")="shmpool_buff_hdr.shadow_file_header.wcb_mu_back_cntr"
	Set gtmtypes("shmpool_buff_hdr",420,"off")=3912
	Set gtmtypes("shmpool_buff_hdr",420,"len")=4
	Set gtmtypes("shmpool_buff_hdr",420,"type")="int"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.wcb_mu_back_cntr")=420
	Set gtmtypes("shmpool_buff_hdr",421,"name")="shmpool_buff_hdr.shadow_file_header.wcb_t_end_sysops_dirty_invcr_cntr"
	Set gtmtypes("shmpool_buff_hdr",421,"off")=3916
	Set gtmtypes("shmpool_buff_hdr",421,"len")=4
	Set gtmtypes("shmpool_buff_hdr",421,"type")="int"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.wcb_t_end_sysops_dirty_invcr_cntr")=421
	Set gtmtypes("shmpool_buff_hdr",422,"name")="shmpool_buff_hdr.shadow_file_header.wcb_t_end_sysops_wtfini_fail_cntr"
	Set gtmtypes("shmpool_buff_hdr",422,"off")=3920
	Set gtmtypes("shmpool_buff_hdr",422,"len")=4
	Set gtmtypes("shmpool_buff_hdr",422,"type")="int"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.wcb_t_end_sysops_wtfini_fail_cntr")=422
	Set gtmtypes("shmpool_buff_hdr",423,"name")="shmpool_buff_hdr.shadow_file_header.wcb_t_end_sysops_twin_stuck_cntr"
	Set gtmtypes("shmpool_buff_hdr",423,"off")=3924
	Set gtmtypes("shmpool_buff_hdr",423,"len")=4
	Set gtmtypes("shmpool_buff_hdr",423,"type")="int"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.wcb_t_end_sysops_twin_stuck_cntr")=423
	Set gtmtypes("shmpool_buff_hdr",424,"name")="shmpool_buff_hdr.shadow_file_header.db_csh_getn_wrt_latch_stuck_cntr"
	Set gtmtypes("shmpool_buff_hdr",424,"off")=3928
	Set gtmtypes("shmpool_buff_hdr",424,"len")=4
	Set gtmtypes("shmpool_buff_hdr",424,"type")="int"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.db_csh_getn_wrt_latch_stuck_cntr")=424
	Set gtmtypes("shmpool_buff_hdr",425,"name")="shmpool_buff_hdr.shadow_file_header.wcb_secshr_db_clnup_now_crit_cntr"
	Set gtmtypes("shmpool_buff_hdr",425,"off")=3932
	Set gtmtypes("shmpool_buff_hdr",425,"len")=4
	Set gtmtypes("shmpool_buff_hdr",425,"type")="int"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.wcb_secshr_db_clnup_now_crit_cntr")=425
	Set gtmtypes("shmpool_buff_hdr",426,"name")="shmpool_buff_hdr.shadow_file_header.wc_blocked_t_qread_bad_bt_index1_cntr"
	Set gtmtypes("shmpool_buff_hdr",426,"off")=3936
	Set gtmtypes("shmpool_buff_hdr",426,"len")=4
	Set gtmtypes("shmpool_buff_hdr",426,"type")="int"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.wc_blocked_t_qread_bad_bt_index1_cntr")=426
	Set gtmtypes("shmpool_buff_hdr",427,"name")="shmpool_buff_hdr.shadow_file_header.wc_blocked_t_qread_bad_bt_index2_cntr"
	Set gtmtypes("shmpool_buff_hdr",427,"off")=3940
	Set gtmtypes("shmpool_buff_hdr",427,"len")=4
	Set gtmtypes("shmpool_buff_hdr",427,"type")="int"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.wc_blocked_t_qread_bad_bt_index2_cntr")=427
	Set gtmtypes("shmpool_buff_hdr",428,"name")="shmpool_buff_hdr.shadow_file_header.wc_blocked_t_end_crbtmismatch1_cntr"
	Set gtmtypes("shmpool_buff_hdr",428,"off")=3944
	Set gtmtypes("shmpool_buff_hdr",428,"len")=4
	Set gtmtypes("shmpool_buff_hdr",428,"type")="int"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.wc_blocked_t_end_crbtmismatch1_cntr")=428
	Set gtmtypes("shmpool_buff_hdr",429,"name")="shmpool_buff_hdr.shadow_file_header.wc_blocked_t_end_crbtmismatch2_cntr"
	Set gtmtypes("shmpool_buff_hdr",429,"off")=3948
	Set gtmtypes("shmpool_buff_hdr",429,"len")=4
	Set gtmtypes("shmpool_buff_hdr",429,"type")="int"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.wc_blocked_t_end_crbtmismatch2_cntr")=429
	Set gtmtypes("shmpool_buff_hdr",430,"name")="shmpool_buff_hdr.shadow_file_header.wc_blocked_t_end_crbtmismatch3_cntr"
	Set gtmtypes("shmpool_buff_hdr",430,"off")=3952
	Set gtmtypes("shmpool_buff_hdr",430,"len")=4
	Set gtmtypes("shmpool_buff_hdr",430,"type")="int"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.wc_blocked_t_end_crbtmismatch3_cntr")=430
	Set gtmtypes("shmpool_buff_hdr",431,"name")="shmpool_buff_hdr.shadow_file_header.wc_blocked_tp_tend_crbtmismatch1_cntr"
	Set gtmtypes("shmpool_buff_hdr",431,"off")=3956
	Set gtmtypes("shmpool_buff_hdr",431,"len")=4
	Set gtmtypes("shmpool_buff_hdr",431,"type")="int"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.wc_blocked_tp_tend_crbtmismatch1_cntr")=431
	Set gtmtypes("shmpool_buff_hdr",432,"name")="shmpool_buff_hdr.shadow_file_header.wc_blocked_tp_tend_crbtmismatch2_cntr"
	Set gtmtypes("shmpool_buff_hdr",432,"off")=3960
	Set gtmtypes("shmpool_buff_hdr",432,"len")=4
	Set gtmtypes("shmpool_buff_hdr",432,"type")="int"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.wc_blocked_tp_tend_crbtmismatch2_cntr")=432
	Set gtmtypes("shmpool_buff_hdr",433,"name")="shmpool_buff_hdr.shadow_file_header.wc_blocked_tp_tend_crbtmismatch3_cntr"
	Set gtmtypes("shmpool_buff_hdr",433,"off")=3964
	Set gtmtypes("shmpool_buff_hdr",433,"len")=4
	Set gtmtypes("shmpool_buff_hdr",433,"type")="int"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.wc_blocked_tp_tend_crbtmismatch3_cntr")=433
	Set gtmtypes("shmpool_buff_hdr",434,"name")="shmpool_buff_hdr.shadow_file_header.wc_blocked_wcs_wtstart_bad_cr_cntr"
	Set gtmtypes("shmpool_buff_hdr",434,"off")=3968
	Set gtmtypes("shmpool_buff_hdr",434,"len")=4
	Set gtmtypes("shmpool_buff_hdr",434,"type")="int"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.wc_blocked_wcs_wtstart_bad_cr_cntr")=434
	Set gtmtypes("shmpool_buff_hdr",435,"name")="shmpool_buff_hdr.shadow_file_header.wc_blocked_wcs_wtfini_bad_cr_cntr"
	Set gtmtypes("shmpool_buff_hdr",435,"off")=3972
	Set gtmtypes("shmpool_buff_hdr",435,"len")=4
	Set gtmtypes("shmpool_buff_hdr",435,"type")="int"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.wc_blocked_wcs_wtfini_bad_cr_cntr")=435
	Set gtmtypes("shmpool_buff_hdr",436,"name")="shmpool_buff_hdr.shadow_file_header.wc_blocked_bt_get_cntr"
	Set gtmtypes("shmpool_buff_hdr",436,"off")=3976
	Set gtmtypes("shmpool_buff_hdr",436,"len")=4
	Set gtmtypes("shmpool_buff_hdr",436,"type")="int"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.wc_blocked_bt_get_cntr")=436
	Set gtmtypes("shmpool_buff_hdr",437,"name")="shmpool_buff_hdr.shadow_file_header.wc_blocked_wcs_cdb_sc_final_retry_cntr"
	Set gtmtypes("shmpool_buff_hdr",437,"off")=3980
	Set gtmtypes("shmpool_buff_hdr",437,"len")=4
	Set gtmtypes("shmpool_buff_hdr",437,"type")="int"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.wc_blocked_wcs_cdb_sc_final_retry_cntr")=437
	Set gtmtypes("shmpool_buff_hdr",438,"name")="shmpool_buff_hdr.shadow_file_header.wcb_bg_update_lckfail1_cntr"
	Set gtmtypes("shmpool_buff_hdr",438,"off")=3984
	Set gtmtypes("shmpool_buff_hdr",438,"len")=4
	Set gtmtypes("shmpool_buff_hdr",438,"type")="int"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.wcb_bg_update_lckfail1_cntr")=438
	Set gtmtypes("shmpool_buff_hdr",439,"name")="shmpool_buff_hdr.shadow_file_header.wcb_bg_update_lckfail2_cntr"
	Set gtmtypes("shmpool_buff_hdr",439,"off")=3988
	Set gtmtypes("shmpool_buff_hdr",439,"len")=4
	Set gtmtypes("shmpool_buff_hdr",439,"type")="int"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.wcb_bg_update_lckfail2_cntr")=439
	Set gtmtypes("shmpool_buff_hdr",440,"name")="shmpool_buff_hdr.shadow_file_header.wcb_wtstart_lckfail1_cntr"
	Set gtmtypes("shmpool_buff_hdr",440,"off")=3992
	Set gtmtypes("shmpool_buff_hdr",440,"len")=4
	Set gtmtypes("shmpool_buff_hdr",440,"type")="int"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.wcb_wtstart_lckfail1_cntr")=440
	Set gtmtypes("shmpool_buff_hdr",441,"name")="shmpool_buff_hdr.shadow_file_header.wcb_wtstart_lckfail2_cntr"
	Set gtmtypes("shmpool_buff_hdr",441,"off")=3996
	Set gtmtypes("shmpool_buff_hdr",441,"len")=4
	Set gtmtypes("shmpool_buff_hdr",441,"type")="int"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.wcb_wtstart_lckfail2_cntr")=441
	Set gtmtypes("shmpool_buff_hdr",442,"name")="shmpool_buff_hdr.shadow_file_header.wcb_wtstart_lckfail3_cntr"
	Set gtmtypes("shmpool_buff_hdr",442,"off")=4000
	Set gtmtypes("shmpool_buff_hdr",442,"len")=4
	Set gtmtypes("shmpool_buff_hdr",442,"type")="int"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.wcb_wtstart_lckfail3_cntr")=442
	Set gtmtypes("shmpool_buff_hdr",443,"name")="shmpool_buff_hdr.shadow_file_header.wcb_wtstart_lckfail4_cntr"
	Set gtmtypes("shmpool_buff_hdr",443,"off")=4004
	Set gtmtypes("shmpool_buff_hdr",443,"len")=4
	Set gtmtypes("shmpool_buff_hdr",443,"type")="int"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.wcb_wtstart_lckfail4_cntr")=443
	Set gtmtypes("shmpool_buff_hdr",444,"name")="shmpool_buff_hdr.shadow_file_header.wcb_wtfini_lckfail1_cntr"
	Set gtmtypes("shmpool_buff_hdr",444,"off")=4008
	Set gtmtypes("shmpool_buff_hdr",444,"len")=4
	Set gtmtypes("shmpool_buff_hdr",444,"type")="int"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.wcb_wtfini_lckfail1_cntr")=444
	Set gtmtypes("shmpool_buff_hdr",445,"name")="shmpool_buff_hdr.shadow_file_header.wcb_wtfini_lckfail2_cntr"
	Set gtmtypes("shmpool_buff_hdr",445,"off")=4012
	Set gtmtypes("shmpool_buff_hdr",445,"len")=4
	Set gtmtypes("shmpool_buff_hdr",445,"type")="int"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.wcb_wtfini_lckfail2_cntr")=445
	Set gtmtypes("shmpool_buff_hdr",446,"name")="shmpool_buff_hdr.shadow_file_header.wcb_wtfini_lckfail3_cntr"
	Set gtmtypes("shmpool_buff_hdr",446,"off")=4016
	Set gtmtypes("shmpool_buff_hdr",446,"len")=4
	Set gtmtypes("shmpool_buff_hdr",446,"type")="int"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.wcb_wtfini_lckfail3_cntr")=446
	Set gtmtypes("shmpool_buff_hdr",447,"name")="shmpool_buff_hdr.shadow_file_header.wcb_wtfini_lckfail4_cntr"
	Set gtmtypes("shmpool_buff_hdr",447,"off")=4020
	Set gtmtypes("shmpool_buff_hdr",447,"len")=4
	Set gtmtypes("shmpool_buff_hdr",447,"type")="int"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.wcb_wtfini_lckfail4_cntr")=447
	Set gtmtypes("shmpool_buff_hdr",448,"name")="shmpool_buff_hdr.shadow_file_header.wcb_t_end_sysops_dirtystuck1_cntr"
	Set gtmtypes("shmpool_buff_hdr",448,"off")=4024
	Set gtmtypes("shmpool_buff_hdr",448,"len")=4
	Set gtmtypes("shmpool_buff_hdr",448,"type")="int"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.wcb_t_end_sysops_dirtystuck1_cntr")=448
	Set gtmtypes("shmpool_buff_hdr",449,"name")="shmpool_buff_hdr.shadow_file_header.wcb_t_end_sysops_dirtystuck2_cntr"
	Set gtmtypes("shmpool_buff_hdr",449,"off")=4028
	Set gtmtypes("shmpool_buff_hdr",449,"len")=4
	Set gtmtypes("shmpool_buff_hdr",449,"type")="int"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.wcb_t_end_sysops_dirtystuck2_cntr")=449
	Set gtmtypes("shmpool_buff_hdr",450,"name")="shmpool_buff_hdr.shadow_file_header.wcb_secshr_db_clnup_wbuf_dqd_cntr"
	Set gtmtypes("shmpool_buff_hdr",450,"off")=4032
	Set gtmtypes("shmpool_buff_hdr",450,"len")=4
	Set gtmtypes("shmpool_buff_hdr",450,"type")="int"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.wcb_secshr_db_clnup_wbuf_dqd_cntr")=450
	Set gtmtypes("shmpool_buff_hdr",451,"name")="shmpool_buff_hdr.shadow_file_header.dwngrd_refmts_syncio_cntr"
	Set gtmtypes("shmpool_buff_hdr",451,"off")=4036
	Set gtmtypes("shmpool_buff_hdr",451,"len")=4
	Set gtmtypes("shmpool_buff_hdr",451,"type")="int"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.dwngrd_refmts_syncio_cntr")=451
	Set gtmtypes("shmpool_buff_hdr",452,"name")="shmpool_buff_hdr.shadow_file_header.dwngrd_refmts_asyncio_cntr"
	Set gtmtypes("shmpool_buff_hdr",452,"off")=4040
	Set gtmtypes("shmpool_buff_hdr",452,"len")=4
	Set gtmtypes("shmpool_buff_hdr",452,"type")="int"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.dwngrd_refmts_asyncio_cntr")=452
	Set gtmtypes("shmpool_buff_hdr",453,"name")="shmpool_buff_hdr.shadow_file_header.shmpool_refmt_harvests_cntr"
	Set gtmtypes("shmpool_buff_hdr",453,"off")=4044
	Set gtmtypes("shmpool_buff_hdr",453,"len")=4
	Set gtmtypes("shmpool_buff_hdr",453,"type")="int"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.shmpool_refmt_harvests_cntr")=453
	Set gtmtypes("shmpool_buff_hdr",454,"name")="shmpool_buff_hdr.shadow_file_header.shmpool_recovery_cntr"
	Set gtmtypes("shmpool_buff_hdr",454,"off")=4048
	Set gtmtypes("shmpool_buff_hdr",454,"len")=4
	Set gtmtypes("shmpool_buff_hdr",454,"type")="int"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.shmpool_recovery_cntr")=454
	Set gtmtypes("shmpool_buff_hdr",455,"name")="shmpool_buff_hdr.shadow_file_header.shmpool_blkd_by_sdc_cntr"
	Set gtmtypes("shmpool_buff_hdr",455,"off")=4052
	Set gtmtypes("shmpool_buff_hdr",455,"len")=4
	Set gtmtypes("shmpool_buff_hdr",455,"type")="int"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.shmpool_blkd_by_sdc_cntr")=455
	Set gtmtypes("shmpool_buff_hdr",456,"name")="shmpool_buff_hdr.shadow_file_header.shmpool_alloc_bbflush_cntr"
	Set gtmtypes("shmpool_buff_hdr",456,"off")=4056
	Set gtmtypes("shmpool_buff_hdr",456,"len")=4
	Set gtmtypes("shmpool_buff_hdr",456,"type")="int"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.shmpool_alloc_bbflush_cntr")=456
	Set gtmtypes("shmpool_buff_hdr",457,"name")="shmpool_buff_hdr.shadow_file_header.refmt_hvst_blk_released_replaced_cntr"
	Set gtmtypes("shmpool_buff_hdr",457,"off")=4060
	Set gtmtypes("shmpool_buff_hdr",457,"len")=4
	Set gtmtypes("shmpool_buff_hdr",457,"type")="int"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.refmt_hvst_blk_released_replaced_cntr")=457
	Set gtmtypes("shmpool_buff_hdr",458,"name")="shmpool_buff_hdr.shadow_file_header.refmt_hvst_blk_released_io_complete_cntr"
	Set gtmtypes("shmpool_buff_hdr",458,"off")=4064
	Set gtmtypes("shmpool_buff_hdr",458,"len")=4
	Set gtmtypes("shmpool_buff_hdr",458,"type")="int"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.refmt_hvst_blk_released_io_complete_cntr")=458
	Set gtmtypes("shmpool_buff_hdr",459,"name")="shmpool_buff_hdr.shadow_file_header.refmt_hvst_blk_kept_cntr"
	Set gtmtypes("shmpool_buff_hdr",459,"off")=4068
	Set gtmtypes("shmpool_buff_hdr",459,"len")=4
	Set gtmtypes("shmpool_buff_hdr",459,"type")="int"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.refmt_hvst_blk_kept_cntr")=459
	Set gtmtypes("shmpool_buff_hdr",460,"name")="shmpool_buff_hdr.shadow_file_header.refmt_hvst_blk_ignored_cntr"
	Set gtmtypes("shmpool_buff_hdr",460,"off")=4072
	Set gtmtypes("shmpool_buff_hdr",460,"len")=4
	Set gtmtypes("shmpool_buff_hdr",460,"type")="int"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.refmt_hvst_blk_ignored_cntr")=460
	Set gtmtypes("shmpool_buff_hdr",461,"name")="shmpool_buff_hdr.shadow_file_header.refmt_blk_chk_blk_freed_cntr"
	Set gtmtypes("shmpool_buff_hdr",461,"off")=4076
	Set gtmtypes("shmpool_buff_hdr",461,"len")=4
	Set gtmtypes("shmpool_buff_hdr",461,"type")="int"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.refmt_blk_chk_blk_freed_cntr")=461
	Set gtmtypes("shmpool_buff_hdr",462,"name")="shmpool_buff_hdr.shadow_file_header.refmt_blk_chk_blk_kept_cntr"
	Set gtmtypes("shmpool_buff_hdr",462,"off")=4080
	Set gtmtypes("shmpool_buff_hdr",462,"len")=4
	Set gtmtypes("shmpool_buff_hdr",462,"type")="int"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.refmt_blk_chk_blk_kept_cntr")=462
	Set gtmtypes("shmpool_buff_hdr",463,"name")="shmpool_buff_hdr.shadow_file_header.active_lvl_trigger_cntr"
	Set gtmtypes("shmpool_buff_hdr",463,"off")=4084
	Set gtmtypes("shmpool_buff_hdr",463,"len")=4
	Set gtmtypes("shmpool_buff_hdr",463,"type")="int"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.active_lvl_trigger_cntr")=463
	Set gtmtypes("shmpool_buff_hdr",464,"name")="shmpool_buff_hdr.shadow_file_header.new_buff_cntr"
	Set gtmtypes("shmpool_buff_hdr",464,"off")=4088
	Set gtmtypes("shmpool_buff_hdr",464,"len")=4
	Set gtmtypes("shmpool_buff_hdr",464,"type")="int"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.new_buff_cntr")=464
	Set gtmtypes("shmpool_buff_hdr",465,"name")="shmpool_buff_hdr.shadow_file_header.phase2_commit_wait_sleep_in_crit_cntr"
	Set gtmtypes("shmpool_buff_hdr",465,"off")=4092
	Set gtmtypes("shmpool_buff_hdr",465,"len")=4
	Set gtmtypes("shmpool_buff_hdr",465,"type")="int"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.phase2_commit_wait_sleep_in_crit_cntr")=465
	Set gtmtypes("shmpool_buff_hdr",466,"name")="shmpool_buff_hdr.shadow_file_header.phase2_commit_wait_sleep_no_crit_cntr"
	Set gtmtypes("shmpool_buff_hdr",466,"off")=4096
	Set gtmtypes("shmpool_buff_hdr",466,"len")=4
	Set gtmtypes("shmpool_buff_hdr",466,"type")="int"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.phase2_commit_wait_sleep_no_crit_cntr")=466
	Set gtmtypes("shmpool_buff_hdr",467,"name")="shmpool_buff_hdr.shadow_file_header.phase2_commit_wait_pidcnt_cntr"
	Set gtmtypes("shmpool_buff_hdr",467,"off")=4100
	Set gtmtypes("shmpool_buff_hdr",467,"len")=4
	Set gtmtypes("shmpool_buff_hdr",467,"type")="int"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.phase2_commit_wait_pidcnt_cntr")=467
	Set gtmtypes("shmpool_buff_hdr",468,"name")="shmpool_buff_hdr.shadow_file_header.wcb_t_end_sysops_intend_wait_cntr"
	Set gtmtypes("shmpool_buff_hdr",468,"off")=4104
	Set gtmtypes("shmpool_buff_hdr",468,"len")=4
	Set gtmtypes("shmpool_buff_hdr",468,"type")="int"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.wcb_t_end_sysops_intend_wait_cntr")=468
	Set gtmtypes("shmpool_buff_hdr",469,"name")="shmpool_buff_hdr.shadow_file_header.wcb_secshr_db_clnup_phase2_clnup_cntr"
	Set gtmtypes("shmpool_buff_hdr",469,"off")=4108
	Set gtmtypes("shmpool_buff_hdr",469,"len")=4
	Set gtmtypes("shmpool_buff_hdr",469,"type")="int"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.wcb_secshr_db_clnup_phase2_clnup_cntr")=469
	Set gtmtypes("shmpool_buff_hdr",470,"name")="shmpool_buff_hdr.shadow_file_header.wcb_phase2_commit_wait_cntr"
	Set gtmtypes("shmpool_buff_hdr",470,"off")=4112
	Set gtmtypes("shmpool_buff_hdr",470,"len")=4
	Set gtmtypes("shmpool_buff_hdr",470,"type")="int"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.wcb_phase2_commit_wait_cntr")=470
	Set gtmtypes("shmpool_buff_hdr",471,"name")="shmpool_buff_hdr.shadow_file_header.recompute_upd_array_calls_cntr"
	Set gtmtypes("shmpool_buff_hdr",471,"off")=4116
	Set gtmtypes("shmpool_buff_hdr",471,"len")=4
	Set gtmtypes("shmpool_buff_hdr",471,"type")="int"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.recompute_upd_array_calls_cntr")=471
	Set gtmtypes("shmpool_buff_hdr",472,"name")="shmpool_buff_hdr.shadow_file_header.recompute_upd_array_rip_cntr"
	Set gtmtypes("shmpool_buff_hdr",472,"off")=4120
	Set gtmtypes("shmpool_buff_hdr",472,"len")=4
	Set gtmtypes("shmpool_buff_hdr",472,"type")="int"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.recompute_upd_array_rip_cntr")=472
	Set gtmtypes("shmpool_buff_hdr",473,"name")="shmpool_buff_hdr.shadow_file_header.recompute_upd_array_in_tend_cntr"
	Set gtmtypes("shmpool_buff_hdr",473,"off")=4124
	Set gtmtypes("shmpool_buff_hdr",473,"len")=4
	Set gtmtypes("shmpool_buff_hdr",473,"type")="int"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.recompute_upd_array_in_tend_cntr")=473
	Set gtmtypes("shmpool_buff_hdr",474,"name")="shmpool_buff_hdr.shadow_file_header.recompute_upd_array_search_blk_cntr"
	Set gtmtypes("shmpool_buff_hdr",474,"off")=4128
	Set gtmtypes("shmpool_buff_hdr",474,"len")=4
	Set gtmtypes("shmpool_buff_hdr",474,"type")="int"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.recompute_upd_array_search_blk_cntr")=474
	Set gtmtypes("shmpool_buff_hdr",475,"name")="shmpool_buff_hdr.shadow_file_header.recompute_upd_array_new_rec_cntr"
	Set gtmtypes("shmpool_buff_hdr",475,"off")=4132
	Set gtmtypes("shmpool_buff_hdr",475,"len")=4
	Set gtmtypes("shmpool_buff_hdr",475,"type")="int"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.recompute_upd_array_new_rec_cntr")=475
	Set gtmtypes("shmpool_buff_hdr",476,"name")="shmpool_buff_hdr.shadow_file_header.recompute_upd_array_rec_size_cntr"
	Set gtmtypes("shmpool_buff_hdr",476,"off")=4136
	Set gtmtypes("shmpool_buff_hdr",476,"len")=4
	Set gtmtypes("shmpool_buff_hdr",476,"type")="int"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.recompute_upd_array_rec_size_cntr")=476
	Set gtmtypes("shmpool_buff_hdr",477,"name")="shmpool_buff_hdr.shadow_file_header.recompute_upd_array_rec_cmpc_cntr"
	Set gtmtypes("shmpool_buff_hdr",477,"off")=4140
	Set gtmtypes("shmpool_buff_hdr",477,"len")=4
	Set gtmtypes("shmpool_buff_hdr",477,"type")="int"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.recompute_upd_array_rec_cmpc_cntr")=477
	Set gtmtypes("shmpool_buff_hdr",478,"name")="shmpool_buff_hdr.shadow_file_header.recompute_upd_array_blk_fini_cntr"
	Set gtmtypes("shmpool_buff_hdr",478,"off")=4144
	Set gtmtypes("shmpool_buff_hdr",478,"len")=4
	Set gtmtypes("shmpool_buff_hdr",478,"type")="int"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.recompute_upd_array_blk_fini_cntr")=478
	Set gtmtypes("shmpool_buff_hdr",479,"name")="shmpool_buff_hdr.shadow_file_header.recompute_upd_array_blk_split_cntr"
	Set gtmtypes("shmpool_buff_hdr",479,"off")=4148
	Set gtmtypes("shmpool_buff_hdr",479,"len")=4
	Set gtmtypes("shmpool_buff_hdr",479,"type")="int"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.recompute_upd_array_blk_split_cntr")=479
	Set gtmtypes("shmpool_buff_hdr",480,"name")="shmpool_buff_hdr.shadow_file_header.unused_filler_0135_cntr"
	Set gtmtypes("shmpool_buff_hdr",480,"off")=4152
	Set gtmtypes("shmpool_buff_hdr",480,"len")=4
	Set gtmtypes("shmpool_buff_hdr",480,"type")="int"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.unused_filler_0135_cntr")=480
	Set gtmtypes("shmpool_buff_hdr",481,"name")="shmpool_buff_hdr.shadow_file_header.t_qread_ripsleep_nblks_cntr"
	Set gtmtypes("shmpool_buff_hdr",481,"off")=4156
	Set gtmtypes("shmpool_buff_hdr",481,"len")=4
	Set gtmtypes("shmpool_buff_hdr",481,"type")="int"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.t_qread_ripsleep_nblks_cntr")=481
	Set gtmtypes("shmpool_buff_hdr",482,"name")="shmpool_buff_hdr.shadow_file_header.wcs_wtfini_invoked_cntr"
	Set gtmtypes("shmpool_buff_hdr",482,"off")=4160
	Set gtmtypes("shmpool_buff_hdr",482,"len")=4
	Set gtmtypes("shmpool_buff_hdr",482,"type")="int"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.wcs_wtfini_invoked_cntr")=482
	Set gtmtypes("shmpool_buff_hdr",483,"name")="shmpool_buff_hdr.shadow_file_header.wcs_wterror_invoked_cntr"
	Set gtmtypes("shmpool_buff_hdr",483,"off")=4164
	Set gtmtypes("shmpool_buff_hdr",483,"len")=4
	Set gtmtypes("shmpool_buff_hdr",483,"type")="int"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.wcs_wterror_invoked_cntr")=483
	Set gtmtypes("shmpool_buff_hdr",484,"name")="shmpool_buff_hdr.shadow_file_header.wcs_wt_restart_invoked_cntr"
	Set gtmtypes("shmpool_buff_hdr",484,"off")=4168
	Set gtmtypes("shmpool_buff_hdr",484,"len")=4
	Set gtmtypes("shmpool_buff_hdr",484,"type")="int"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.wcs_wt_restart_invoked_cntr")=484
	Set gtmtypes("shmpool_buff_hdr",485,"name")="shmpool_buff_hdr.shadow_file_header.wcs_wt_restart_eagain_cntr"
	Set gtmtypes("shmpool_buff_hdr",485,"off")=4172
	Set gtmtypes("shmpool_buff_hdr",485,"len")=4
	Set gtmtypes("shmpool_buff_hdr",485,"type")="int"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.wcs_wt_restart_eagain_cntr")=485
	Set gtmtypes("shmpool_buff_hdr",486,"name")="shmpool_buff_hdr.shadow_file_header.wcs_wt_restart_reinsert_cntr"
	Set gtmtypes("shmpool_buff_hdr",486,"off")=4176
	Set gtmtypes("shmpool_buff_hdr",486,"len")=4
	Set gtmtypes("shmpool_buff_hdr",486,"type")="int"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.wcs_wt_restart_reinsert_cntr")=486
	Set gtmtypes("shmpool_buff_hdr",487,"name")="shmpool_buff_hdr.shadow_file_header.wcs_wtstart_eagain_cntr"
	Set gtmtypes("shmpool_buff_hdr",487,"off")=4180
	Set gtmtypes("shmpool_buff_hdr",487,"len")=4
	Set gtmtypes("shmpool_buff_hdr",487,"type")="int"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.wcs_wtstart_eagain_cntr")=487
	Set gtmtypes("shmpool_buff_hdr",488,"name")="shmpool_buff_hdr.shadow_file_header.wcs_wtstart_eagain_incrit_cntr"
	Set gtmtypes("shmpool_buff_hdr",488,"off")=4184
	Set gtmtypes("shmpool_buff_hdr",488,"len")=4
	Set gtmtypes("shmpool_buff_hdr",488,"type")="int"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.wcs_wtstart_eagain_incrit_cntr")=488
	Set gtmtypes("shmpool_buff_hdr",489,"name")="shmpool_buff_hdr.shadow_file_header.async_restart_eagain_cntr"
	Set gtmtypes("shmpool_buff_hdr",489,"off")=4188
	Set gtmtypes("shmpool_buff_hdr",489,"len")=4
	Set gtmtypes("shmpool_buff_hdr",489,"type")="int"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.async_restart_eagain_cntr")=489
	Set gtmtypes("shmpool_buff_hdr",490,"name")="shmpool_buff_hdr.shadow_file_header.jnlbuff_phs2cmt_array_full_cntr"
	Set gtmtypes("shmpool_buff_hdr",490,"off")=4192
	Set gtmtypes("shmpool_buff_hdr",490,"len")=4
	Set gtmtypes("shmpool_buff_hdr",490,"type")="int"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.jnlbuff_phs2cmt_array_full_cntr")=490
	Set gtmtypes("shmpool_buff_hdr",491,"name")="shmpool_buff_hdr.shadow_file_header.jnl_phase2_cleanup_if_possible_cntr"
	Set gtmtypes("shmpool_buff_hdr",491,"off")=4196
	Set gtmtypes("shmpool_buff_hdr",491,"len")=4
	Set gtmtypes("shmpool_buff_hdr",491,"type")="int"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.jnl_phase2_cleanup_if_possible_cntr")=491
	Set gtmtypes("shmpool_buff_hdr",492,"name")="shmpool_buff_hdr.shadow_file_header.jnlbuff_phs2cmt_isprcalv_cntr"
	Set gtmtypes("shmpool_buff_hdr",492,"off")=4200
	Set gtmtypes("shmpool_buff_hdr",492,"len")=4
	Set gtmtypes("shmpool_buff_hdr",492,"type")="int"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.jnlbuff_phs2cmt_isprcalv_cntr")=492
	Set gtmtypes("shmpool_buff_hdr",493,"name")="shmpool_buff_hdr.shadow_file_header.recompute_upd_array_paidoff_cntr"
	Set gtmtypes("shmpool_buff_hdr",493,"off")=4204
	Set gtmtypes("shmpool_buff_hdr",493,"len")=4
	Set gtmtypes("shmpool_buff_hdr",493,"type")="int"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.recompute_upd_array_paidoff_cntr")=493
	Set gtmtypes("shmpool_buff_hdr",494,"name")="shmpool_buff_hdr.shadow_file_header.bg_trc_rec_cntr_filler"
	Set gtmtypes("shmpool_buff_hdr",494,"off")=4208
	Set gtmtypes("shmpool_buff_hdr",494,"len")=8
	Set gtmtypes("shmpool_buff_hdr",494,"type")="char"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.bg_trc_rec_cntr_filler")=494
	Set gtmtypes("shmpool_buff_hdr",495,"name")="shmpool_buff_hdr.shadow_file_header.n_gvcst_srches"
	Set gtmtypes("shmpool_buff_hdr",495,"off")=4216
	Set gtmtypes("shmpool_buff_hdr",495,"len")=8
	Set gtmtypes("shmpool_buff_hdr",495,"type")="db_csh_acct_rec"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.n_gvcst_srches")=495
	Set gtmtypes("shmpool_buff_hdr",496,"name")="shmpool_buff_hdr.shadow_file_header.n_gvcst_srches.curr_count"
	Set gtmtypes("shmpool_buff_hdr",496,"off")=4216
	Set gtmtypes("shmpool_buff_hdr",496,"len")=4
	Set gtmtypes("shmpool_buff_hdr",496,"type")="int"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.n_gvcst_srches.curr_count")=496
	Set gtmtypes("shmpool_buff_hdr",497,"name")="shmpool_buff_hdr.shadow_file_header.n_gvcst_srches.cumul_count"
	Set gtmtypes("shmpool_buff_hdr",497,"off")=4220
	Set gtmtypes("shmpool_buff_hdr",497,"len")=4
	Set gtmtypes("shmpool_buff_hdr",497,"type")="int"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.n_gvcst_srches.cumul_count")=497
	Set gtmtypes("shmpool_buff_hdr",498,"name")="shmpool_buff_hdr.shadow_file_header.n_gvcst_srch_clues"
	Set gtmtypes("shmpool_buff_hdr",498,"off")=4224
	Set gtmtypes("shmpool_buff_hdr",498,"len")=8
	Set gtmtypes("shmpool_buff_hdr",498,"type")="db_csh_acct_rec"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.n_gvcst_srch_clues")=498
	Set gtmtypes("shmpool_buff_hdr",499,"name")="shmpool_buff_hdr.shadow_file_header.n_gvcst_srch_clues.curr_count"
	Set gtmtypes("shmpool_buff_hdr",499,"off")=4224
	Set gtmtypes("shmpool_buff_hdr",499,"len")=4
	Set gtmtypes("shmpool_buff_hdr",499,"type")="int"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.n_gvcst_srch_clues.curr_count")=499
	Set gtmtypes("shmpool_buff_hdr",500,"name")="shmpool_buff_hdr.shadow_file_header.n_gvcst_srch_clues.cumul_count"
	Set gtmtypes("shmpool_buff_hdr",500,"off")=4228
	Set gtmtypes("shmpool_buff_hdr",500,"len")=4
	Set gtmtypes("shmpool_buff_hdr",500,"type")="int"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.n_gvcst_srch_clues.cumul_count")=500
	Set gtmtypes("shmpool_buff_hdr",501,"name")="shmpool_buff_hdr.shadow_file_header.n_clue_used_head"
	Set gtmtypes("shmpool_buff_hdr",501,"off")=4232
	Set gtmtypes("shmpool_buff_hdr",501,"len")=8
	Set gtmtypes("shmpool_buff_hdr",501,"type")="db_csh_acct_rec"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.n_clue_used_head")=501
	Set gtmtypes("shmpool_buff_hdr",502,"name")="shmpool_buff_hdr.shadow_file_header.n_clue_used_head.curr_count"
	Set gtmtypes("shmpool_buff_hdr",502,"off")=4232
	Set gtmtypes("shmpool_buff_hdr",502,"len")=4
	Set gtmtypes("shmpool_buff_hdr",502,"type")="int"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.n_clue_used_head.curr_count")=502
	Set gtmtypes("shmpool_buff_hdr",503,"name")="shmpool_buff_hdr.shadow_file_header.n_clue_used_head.cumul_count"
	Set gtmtypes("shmpool_buff_hdr",503,"off")=4236
	Set gtmtypes("shmpool_buff_hdr",503,"len")=4
	Set gtmtypes("shmpool_buff_hdr",503,"type")="int"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.n_clue_used_head.cumul_count")=503
	Set gtmtypes("shmpool_buff_hdr",504,"name")="shmpool_buff_hdr.shadow_file_header.n_clue_used_same"
	Set gtmtypes("shmpool_buff_hdr",504,"off")=4240
	Set gtmtypes("shmpool_buff_hdr",504,"len")=8
	Set gtmtypes("shmpool_buff_hdr",504,"type")="db_csh_acct_rec"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.n_clue_used_same")=504
	Set gtmtypes("shmpool_buff_hdr",505,"name")="shmpool_buff_hdr.shadow_file_header.n_clue_used_same.curr_count"
	Set gtmtypes("shmpool_buff_hdr",505,"off")=4240
	Set gtmtypes("shmpool_buff_hdr",505,"len")=4
	Set gtmtypes("shmpool_buff_hdr",505,"type")="int"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.n_clue_used_same.curr_count")=505
	Set gtmtypes("shmpool_buff_hdr",506,"name")="shmpool_buff_hdr.shadow_file_header.n_clue_used_same.cumul_count"
	Set gtmtypes("shmpool_buff_hdr",506,"off")=4244
	Set gtmtypes("shmpool_buff_hdr",506,"len")=4
	Set gtmtypes("shmpool_buff_hdr",506,"type")="int"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.n_clue_used_same.cumul_count")=506
	Set gtmtypes("shmpool_buff_hdr",507,"name")="shmpool_buff_hdr.shadow_file_header.n_clue_used_tail"
	Set gtmtypes("shmpool_buff_hdr",507,"off")=4248
	Set gtmtypes("shmpool_buff_hdr",507,"len")=8
	Set gtmtypes("shmpool_buff_hdr",507,"type")="db_csh_acct_rec"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.n_clue_used_tail")=507
	Set gtmtypes("shmpool_buff_hdr",508,"name")="shmpool_buff_hdr.shadow_file_header.n_clue_used_tail.curr_count"
	Set gtmtypes("shmpool_buff_hdr",508,"off")=4248
	Set gtmtypes("shmpool_buff_hdr",508,"len")=4
	Set gtmtypes("shmpool_buff_hdr",508,"type")="int"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.n_clue_used_tail.curr_count")=508
	Set gtmtypes("shmpool_buff_hdr",509,"name")="shmpool_buff_hdr.shadow_file_header.n_clue_used_tail.cumul_count"
	Set gtmtypes("shmpool_buff_hdr",509,"off")=4252
	Set gtmtypes("shmpool_buff_hdr",509,"len")=4
	Set gtmtypes("shmpool_buff_hdr",509,"type")="int"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.n_clue_used_tail.cumul_count")=509
	Set gtmtypes("shmpool_buff_hdr",510,"name")="shmpool_buff_hdr.shadow_file_header.n_t_qreads"
	Set gtmtypes("shmpool_buff_hdr",510,"off")=4256
	Set gtmtypes("shmpool_buff_hdr",510,"len")=8
	Set gtmtypes("shmpool_buff_hdr",510,"type")="db_csh_acct_rec"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.n_t_qreads")=510
	Set gtmtypes("shmpool_buff_hdr",511,"name")="shmpool_buff_hdr.shadow_file_header.n_t_qreads.curr_count"
	Set gtmtypes("shmpool_buff_hdr",511,"off")=4256
	Set gtmtypes("shmpool_buff_hdr",511,"len")=4
	Set gtmtypes("shmpool_buff_hdr",511,"type")="int"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.n_t_qreads.curr_count")=511
	Set gtmtypes("shmpool_buff_hdr",512,"name")="shmpool_buff_hdr.shadow_file_header.n_t_qreads.cumul_count"
	Set gtmtypes("shmpool_buff_hdr",512,"off")=4260
	Set gtmtypes("shmpool_buff_hdr",512,"len")=4
	Set gtmtypes("shmpool_buff_hdr",512,"type")="int"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.n_t_qreads.cumul_count")=512
	Set gtmtypes("shmpool_buff_hdr",513,"name")="shmpool_buff_hdr.shadow_file_header.unused_dsk_reads"
	Set gtmtypes("shmpool_buff_hdr",513,"off")=4264
	Set gtmtypes("shmpool_buff_hdr",513,"len")=8
	Set gtmtypes("shmpool_buff_hdr",513,"type")="db_csh_acct_rec"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.unused_dsk_reads")=513
	Set gtmtypes("shmpool_buff_hdr",514,"name")="shmpool_buff_hdr.shadow_file_header.unused_dsk_reads.curr_count"
	Set gtmtypes("shmpool_buff_hdr",514,"off")=4264
	Set gtmtypes("shmpool_buff_hdr",514,"len")=4
	Set gtmtypes("shmpool_buff_hdr",514,"type")="int"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.unused_dsk_reads.curr_count")=514
	Set gtmtypes("shmpool_buff_hdr",515,"name")="shmpool_buff_hdr.shadow_file_header.unused_dsk_reads.cumul_count"
	Set gtmtypes("shmpool_buff_hdr",515,"off")=4268
	Set gtmtypes("shmpool_buff_hdr",515,"len")=4
	Set gtmtypes("shmpool_buff_hdr",515,"type")="int"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.unused_dsk_reads.cumul_count")=515
	Set gtmtypes("shmpool_buff_hdr",516,"name")="shmpool_buff_hdr.shadow_file_header.n_bgmm_updates"
	Set gtmtypes("shmpool_buff_hdr",516,"off")=4272
	Set gtmtypes("shmpool_buff_hdr",516,"len")=8
	Set gtmtypes("shmpool_buff_hdr",516,"type")="db_csh_acct_rec"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.n_bgmm_updates")=516
	Set gtmtypes("shmpool_buff_hdr",517,"name")="shmpool_buff_hdr.shadow_file_header.n_bgmm_updates.curr_count"
	Set gtmtypes("shmpool_buff_hdr",517,"off")=4272
	Set gtmtypes("shmpool_buff_hdr",517,"len")=4
	Set gtmtypes("shmpool_buff_hdr",517,"type")="int"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.n_bgmm_updates.curr_count")=517
	Set gtmtypes("shmpool_buff_hdr",518,"name")="shmpool_buff_hdr.shadow_file_header.n_bgmm_updates.cumul_count"
	Set gtmtypes("shmpool_buff_hdr",518,"off")=4276
	Set gtmtypes("shmpool_buff_hdr",518,"len")=4
	Set gtmtypes("shmpool_buff_hdr",518,"type")="int"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.n_bgmm_updates.cumul_count")=518
	Set gtmtypes("shmpool_buff_hdr",519,"name")="shmpool_buff_hdr.shadow_file_header.unused_dsk_writes"
	Set gtmtypes("shmpool_buff_hdr",519,"off")=4280
	Set gtmtypes("shmpool_buff_hdr",519,"len")=8
	Set gtmtypes("shmpool_buff_hdr",519,"type")="db_csh_acct_rec"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.unused_dsk_writes")=519
	Set gtmtypes("shmpool_buff_hdr",520,"name")="shmpool_buff_hdr.shadow_file_header.unused_dsk_writes.curr_count"
	Set gtmtypes("shmpool_buff_hdr",520,"off")=4280
	Set gtmtypes("shmpool_buff_hdr",520,"len")=4
	Set gtmtypes("shmpool_buff_hdr",520,"type")="int"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.unused_dsk_writes.curr_count")=520
	Set gtmtypes("shmpool_buff_hdr",521,"name")="shmpool_buff_hdr.shadow_file_header.unused_dsk_writes.cumul_count"
	Set gtmtypes("shmpool_buff_hdr",521,"off")=4284
	Set gtmtypes("shmpool_buff_hdr",521,"len")=4
	Set gtmtypes("shmpool_buff_hdr",521,"type")="int"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.unused_dsk_writes.cumul_count")=521
	Set gtmtypes("shmpool_buff_hdr",522,"name")="shmpool_buff_hdr.shadow_file_header.n_bg_update_creates"
	Set gtmtypes("shmpool_buff_hdr",522,"off")=4288
	Set gtmtypes("shmpool_buff_hdr",522,"len")=8
	Set gtmtypes("shmpool_buff_hdr",522,"type")="db_csh_acct_rec"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.n_bg_update_creates")=522
	Set gtmtypes("shmpool_buff_hdr",523,"name")="shmpool_buff_hdr.shadow_file_header.n_bg_update_creates.curr_count"
	Set gtmtypes("shmpool_buff_hdr",523,"off")=4288
	Set gtmtypes("shmpool_buff_hdr",523,"len")=4
	Set gtmtypes("shmpool_buff_hdr",523,"type")="int"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.n_bg_update_creates.curr_count")=523
	Set gtmtypes("shmpool_buff_hdr",524,"name")="shmpool_buff_hdr.shadow_file_header.n_bg_update_creates.cumul_count"
	Set gtmtypes("shmpool_buff_hdr",524,"off")=4292
	Set gtmtypes("shmpool_buff_hdr",524,"len")=4
	Set gtmtypes("shmpool_buff_hdr",524,"type")="int"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.n_bg_update_creates.cumul_count")=524
	Set gtmtypes("shmpool_buff_hdr",525,"name")="shmpool_buff_hdr.shadow_file_header.n_db_csh_getns"
	Set gtmtypes("shmpool_buff_hdr",525,"off")=4296
	Set gtmtypes("shmpool_buff_hdr",525,"len")=8
	Set gtmtypes("shmpool_buff_hdr",525,"type")="db_csh_acct_rec"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.n_db_csh_getns")=525
	Set gtmtypes("shmpool_buff_hdr",526,"name")="shmpool_buff_hdr.shadow_file_header.n_db_csh_getns.curr_count"
	Set gtmtypes("shmpool_buff_hdr",526,"off")=4296
	Set gtmtypes("shmpool_buff_hdr",526,"len")=4
	Set gtmtypes("shmpool_buff_hdr",526,"type")="int"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.n_db_csh_getns.curr_count")=526
	Set gtmtypes("shmpool_buff_hdr",527,"name")="shmpool_buff_hdr.shadow_file_header.n_db_csh_getns.cumul_count"
	Set gtmtypes("shmpool_buff_hdr",527,"off")=4300
	Set gtmtypes("shmpool_buff_hdr",527,"len")=4
	Set gtmtypes("shmpool_buff_hdr",527,"type")="int"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.n_db_csh_getns.cumul_count")=527
	Set gtmtypes("shmpool_buff_hdr",528,"name")="shmpool_buff_hdr.shadow_file_header.n_db_csh_getn_lcnt"
	Set gtmtypes("shmpool_buff_hdr",528,"off")=4304
	Set gtmtypes("shmpool_buff_hdr",528,"len")=8
	Set gtmtypes("shmpool_buff_hdr",528,"type")="db_csh_acct_rec"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.n_db_csh_getn_lcnt")=528
	Set gtmtypes("shmpool_buff_hdr",529,"name")="shmpool_buff_hdr.shadow_file_header.n_db_csh_getn_lcnt.curr_count"
	Set gtmtypes("shmpool_buff_hdr",529,"off")=4304
	Set gtmtypes("shmpool_buff_hdr",529,"len")=4
	Set gtmtypes("shmpool_buff_hdr",529,"type")="int"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.n_db_csh_getn_lcnt.curr_count")=529
	Set gtmtypes("shmpool_buff_hdr",530,"name")="shmpool_buff_hdr.shadow_file_header.n_db_csh_getn_lcnt.cumul_count"
	Set gtmtypes("shmpool_buff_hdr",530,"off")=4308
	Set gtmtypes("shmpool_buff_hdr",530,"len")=4
	Set gtmtypes("shmpool_buff_hdr",530,"type")="int"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.n_db_csh_getn_lcnt.cumul_count")=530
	Set gtmtypes("shmpool_buff_hdr",531,"name")="shmpool_buff_hdr.shadow_file_header.db_csh_acct_rec_filler_4k"
	Set gtmtypes("shmpool_buff_hdr",531,"off")=4312
	Set gtmtypes("shmpool_buff_hdr",531,"len")=152
	Set gtmtypes("shmpool_buff_hdr",531,"type")="char"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.db_csh_acct_rec_filler_4k")=531
	Set gtmtypes("shmpool_buff_hdr",532,"name")="shmpool_buff_hdr.shadow_file_header.gvstats_rec_old_now_filler"
	Set gtmtypes("shmpool_buff_hdr",532,"off")=4464
	Set gtmtypes("shmpool_buff_hdr",532,"len")=496
	Set gtmtypes("shmpool_buff_hdr",532,"type")="char"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.gvstats_rec_old_now_filler")=532
	Set gtmtypes("shmpool_buff_hdr",533,"name")="shmpool_buff_hdr.shadow_file_header.gvstats_rec_filler_4k_plus_512"
	Set gtmtypes("shmpool_buff_hdr",533,"off")=4960
	Set gtmtypes("shmpool_buff_hdr",533,"len")=16
	Set gtmtypes("shmpool_buff_hdr",533,"type")="char"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.gvstats_rec_filler_4k_plus_512")=533
	Set gtmtypes("shmpool_buff_hdr",534,"name")="shmpool_buff_hdr.shadow_file_header.filler_4k_plus_512"
	Set gtmtypes("shmpool_buff_hdr",534,"off")=4976
	Set gtmtypes("shmpool_buff_hdr",534,"len")=368
	Set gtmtypes("shmpool_buff_hdr",534,"type")="char"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.filler_4k_plus_512")=534
	Set gtmtypes("shmpool_buff_hdr",535,"name")="shmpool_buff_hdr.shadow_file_header.intrpt_recov_resync_strm_seqno"
	Set gtmtypes("shmpool_buff_hdr",535,"off")=5344
	Set gtmtypes("shmpool_buff_hdr",535,"len")=128
	Set gtmtypes("shmpool_buff_hdr",535,"type")="uint64_t"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.intrpt_recov_resync_strm_seqno")=535
	Set gtmtypes("shmpool_buff_hdr",535,"dim")=16
	Set gtmtypes("shmpool_buff_hdr",536,"name")="shmpool_buff_hdr.shadow_file_header.creation_db_ver"
	Set gtmtypes("shmpool_buff_hdr",536,"off")=5472
	Set gtmtypes("shmpool_buff_hdr",536,"len")=4
	Set gtmtypes("shmpool_buff_hdr",536,"type")="int"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.creation_db_ver")=536
	Set gtmtypes("shmpool_buff_hdr",537,"name")="shmpool_buff_hdr.shadow_file_header.creation_mdb_ver"
	Set gtmtypes("shmpool_buff_hdr",537,"off")=5476
	Set gtmtypes("shmpool_buff_hdr",537,"len")=4
	Set gtmtypes("shmpool_buff_hdr",537,"type")="int"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.creation_mdb_ver")=537
	Set gtmtypes("shmpool_buff_hdr",538,"name")="shmpool_buff_hdr.shadow_file_header.certified_for_upgrade_to"
	Set gtmtypes("shmpool_buff_hdr",538,"off")=5480
	Set gtmtypes("shmpool_buff_hdr",538,"len")=4
	Set gtmtypes("shmpool_buff_hdr",538,"type")="int"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.certified_for_upgrade_to")=538
	Set gtmtypes("shmpool_buff_hdr",539,"name")="shmpool_buff_hdr.shadow_file_header.filler_5k"
	Set gtmtypes("shmpool_buff_hdr",539,"off")=5484
	Set gtmtypes("shmpool_buff_hdr",539,"len")=4
	Set gtmtypes("shmpool_buff_hdr",539,"type")="int"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.filler_5k")=539
	Set gtmtypes("shmpool_buff_hdr",540,"name")="shmpool_buff_hdr.shadow_file_header.secshr_ops_index_filler"
	Set gtmtypes("shmpool_buff_hdr",540,"off")=5488
	Set gtmtypes("shmpool_buff_hdr",540,"len")=4
	Set gtmtypes("shmpool_buff_hdr",540,"type")="int"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.secshr_ops_index_filler")=540
	Set gtmtypes("shmpool_buff_hdr",541,"name")="shmpool_buff_hdr.shadow_file_header.secshr_ops_array_filler"
	Set gtmtypes("shmpool_buff_hdr",541,"off")=5492
	Set gtmtypes("shmpool_buff_hdr",541,"len")=1020
	Set gtmtypes("shmpool_buff_hdr",541,"type")="int"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.secshr_ops_array_filler")=541
	Set gtmtypes("shmpool_buff_hdr",541,"dim")=255
	Set gtmtypes("shmpool_buff_hdr",542,"name")="shmpool_buff_hdr.shadow_file_header.next_upgrd_warn"
	Set gtmtypes("shmpool_buff_hdr",542,"off")=6512
	Set gtmtypes("shmpool_buff_hdr",542,"len")=24
	Set gtmtypes("shmpool_buff_hdr",542,"type")="compswap_time_field"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.next_upgrd_warn")=542
	Set gtmtypes("shmpool_buff_hdr",543,"name")="shmpool_buff_hdr.shadow_file_header.next_upgrd_warn.time_latch"
	Set gtmtypes("shmpool_buff_hdr",543,"off")=6512
	Set gtmtypes("shmpool_buff_hdr",543,"len")=8
	Set gtmtypes("shmpool_buff_hdr",543,"type")="global_latch_t"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.next_upgrd_warn.time_latch")=543
	Set gtmtypes("shmpool_buff_hdr",544,"name")="shmpool_buff_hdr.shadow_file_header.next_upgrd_warn.time_latch.u"
	Set gtmtypes("shmpool_buff_hdr",544,"off")=6512
	Set gtmtypes("shmpool_buff_hdr",544,"len")=8
	Set gtmtypes("shmpool_buff_hdr",544,"type")="union"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.next_upgrd_warn.time_latch.u")=544
	Set gtmtypes("shmpool_buff_hdr",545,"name")="shmpool_buff_hdr.shadow_file_header.next_upgrd_warn.time_latch.u.pid_imgcnt"
	Set gtmtypes("shmpool_buff_hdr",545,"off")=6512
	Set gtmtypes("shmpool_buff_hdr",545,"len")=8
	Set gtmtypes("shmpool_buff_hdr",545,"type")="uint64_t"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.next_upgrd_warn.time_latch.u.pid_imgcnt")=545
	Set gtmtypes("shmpool_buff_hdr",546,"name")="shmpool_buff_hdr.shadow_file_header.next_upgrd_warn.time_latch.u.parts"
	Set gtmtypes("shmpool_buff_hdr",546,"off")=6512
	Set gtmtypes("shmpool_buff_hdr",546,"len")=8
	Set gtmtypes("shmpool_buff_hdr",546,"type")="struct"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.next_upgrd_warn.time_latch.u.parts")=546
	Set gtmtypes("shmpool_buff_hdr",547,"name")="shmpool_buff_hdr.shadow_file_header.next_upgrd_warn.time_latch.u.parts.latch_pid"
	Set gtmtypes("shmpool_buff_hdr",547,"off")=6512
	Set gtmtypes("shmpool_buff_hdr",547,"len")=4
	Set gtmtypes("shmpool_buff_hdr",547,"type")="int"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.next_upgrd_warn.time_latch.u.parts.latch_pid")=547
	Set gtmtypes("shmpool_buff_hdr",548,"name")="shmpool_buff_hdr.shadow_file_header.next_upgrd_warn.time_latch.u.parts.latch_word"
	Set gtmtypes("shmpool_buff_hdr",548,"off")=6516
	Set gtmtypes("shmpool_buff_hdr",548,"len")=4
	Set gtmtypes("shmpool_buff_hdr",548,"type")="int"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.next_upgrd_warn.time_latch.u.parts.latch_word")=548
	Set gtmtypes("shmpool_buff_hdr",549,"name")="shmpool_buff_hdr.shadow_file_header.next_upgrd_warn.hp_latch_space"
	Set gtmtypes("shmpool_buff_hdr",549,"off")=6520
	Set gtmtypes("shmpool_buff_hdr",549,"len")=16
	Set gtmtypes("shmpool_buff_hdr",549,"type")="int"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.next_upgrd_warn.hp_latch_space")=549
	Set gtmtypes("shmpool_buff_hdr",549,"dim")=4
	Set gtmtypes("shmpool_buff_hdr",550,"name")="shmpool_buff_hdr.shadow_file_header.is_encrypted"
	Set gtmtypes("shmpool_buff_hdr",550,"off")=6536
	Set gtmtypes("shmpool_buff_hdr",550,"len")=4
	Set gtmtypes("shmpool_buff_hdr",550,"type")="unsigned-int"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.is_encrypted")=550
	Set gtmtypes("shmpool_buff_hdr",551,"name")="shmpool_buff_hdr.shadow_file_header.db_trigger_cycle"
	Set gtmtypes("shmpool_buff_hdr",551,"off")=6540
	Set gtmtypes("shmpool_buff_hdr",551,"len")=4
	Set gtmtypes("shmpool_buff_hdr",551,"type")="unsigned-int"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.db_trigger_cycle")=551
	Set gtmtypes("shmpool_buff_hdr",552,"name")="shmpool_buff_hdr.shadow_file_header.strm_reg_seqno"
	Set gtmtypes("shmpool_buff_hdr",552,"off")=6544
	Set gtmtypes("shmpool_buff_hdr",552,"len")=128
	Set gtmtypes("shmpool_buff_hdr",552,"type")="uint64_t"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.strm_reg_seqno")=552
	Set gtmtypes("shmpool_buff_hdr",552,"dim")=16
	Set gtmtypes("shmpool_buff_hdr",553,"name")="shmpool_buff_hdr.shadow_file_header.save_strm_reg_seqno"
	Set gtmtypes("shmpool_buff_hdr",553,"off")=6672
	Set gtmtypes("shmpool_buff_hdr",553,"len")=128
	Set gtmtypes("shmpool_buff_hdr",553,"type")="uint64_t"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.save_strm_reg_seqno")=553
	Set gtmtypes("shmpool_buff_hdr",553,"dim")=16
	Set gtmtypes("shmpool_buff_hdr",554,"name")="shmpool_buff_hdr.shadow_file_header.freeze_on_fail"
	Set gtmtypes("shmpool_buff_hdr",554,"off")=6800
	Set gtmtypes("shmpool_buff_hdr",554,"len")=4
	Set gtmtypes("shmpool_buff_hdr",554,"type")="boolean_t"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.freeze_on_fail")=554
	Set gtmtypes("shmpool_buff_hdr",555,"name")="shmpool_buff_hdr.shadow_file_header.span_node_absent"
	Set gtmtypes("shmpool_buff_hdr",555,"off")=6804
	Set gtmtypes("shmpool_buff_hdr",555,"len")=4
	Set gtmtypes("shmpool_buff_hdr",555,"type")="boolean_t"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.span_node_absent")=555
	Set gtmtypes("shmpool_buff_hdr",556,"name")="shmpool_buff_hdr.shadow_file_header.maxkeysz_assured"
	Set gtmtypes("shmpool_buff_hdr",556,"off")=6808
	Set gtmtypes("shmpool_buff_hdr",556,"len")=4
	Set gtmtypes("shmpool_buff_hdr",556,"type")="boolean_t"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.maxkeysz_assured")=556
	Set gtmtypes("shmpool_buff_hdr",557,"name")="shmpool_buff_hdr.shadow_file_header.hasht_upgrade_needed"
	Set gtmtypes("shmpool_buff_hdr",557,"off")=6812
	Set gtmtypes("shmpool_buff_hdr",557,"len")=4
	Set gtmtypes("shmpool_buff_hdr",557,"type")="boolean_t"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.hasht_upgrade_needed")=557
	Set gtmtypes("shmpool_buff_hdr",558,"name")="shmpool_buff_hdr.shadow_file_header.defer_allocate"
	Set gtmtypes("shmpool_buff_hdr",558,"off")=6816
	Set gtmtypes("shmpool_buff_hdr",558,"len")=4
	Set gtmtypes("shmpool_buff_hdr",558,"type")="boolean_t"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.defer_allocate")=558
	Set gtmtypes("shmpool_buff_hdr",559,"name")="shmpool_buff_hdr.shadow_file_header.filler_ftok_counter_halted"
	Set gtmtypes("shmpool_buff_hdr",559,"off")=6820
	Set gtmtypes("shmpool_buff_hdr",559,"len")=4
	Set gtmtypes("shmpool_buff_hdr",559,"type")="boolean_t"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.filler_ftok_counter_halted")=559
	Set gtmtypes("shmpool_buff_hdr",560,"name")="shmpool_buff_hdr.shadow_file_header.filler_access_counter_halted"
	Set gtmtypes("shmpool_buff_hdr",560,"off")=6824
	Set gtmtypes("shmpool_buff_hdr",560,"len")=4
	Set gtmtypes("shmpool_buff_hdr",560,"type")="boolean_t"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.filler_access_counter_halted")=560
	Set gtmtypes("shmpool_buff_hdr",561,"name")="shmpool_buff_hdr.shadow_file_header.lock_crit_with_db"
	Set gtmtypes("shmpool_buff_hdr",561,"off")=6828
	Set gtmtypes("shmpool_buff_hdr",561,"len")=4
	Set gtmtypes("shmpool_buff_hdr",561,"type")="boolean_t"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.lock_crit_with_db")=561
	Set gtmtypes("shmpool_buff_hdr",562,"name")="shmpool_buff_hdr.shadow_file_header.basedb_fname_len"
	Set gtmtypes("shmpool_buff_hdr",562,"off")=6832
	Set gtmtypes("shmpool_buff_hdr",562,"len")=4
	Set gtmtypes("shmpool_buff_hdr",562,"type")="unsigned-int"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.basedb_fname_len")=562
	Set gtmtypes("shmpool_buff_hdr",563,"name")="shmpool_buff_hdr.shadow_file_header.basedb_fname"
	Set gtmtypes("shmpool_buff_hdr",563,"off")=6836
	Set gtmtypes("shmpool_buff_hdr",563,"len")=256
	Set gtmtypes("shmpool_buff_hdr",563,"type")="unsigned-char"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.basedb_fname")=563
	Set gtmtypes("shmpool_buff_hdr",564,"name")="shmpool_buff_hdr.shadow_file_header.read_only"
	Set gtmtypes("shmpool_buff_hdr",564,"off")=7092
	Set gtmtypes("shmpool_buff_hdr",564,"len")=4
	Set gtmtypes("shmpool_buff_hdr",564,"type")="boolean_t"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.read_only")=564
	Set gtmtypes("shmpool_buff_hdr",565,"name")="shmpool_buff_hdr.shadow_file_header.gvstats_rec"
	Set gtmtypes("shmpool_buff_hdr",565,"off")=7096
	Set gtmtypes("shmpool_buff_hdr",565,"len")=592
	Set gtmtypes("shmpool_buff_hdr",565,"type")="gvstats_rec_csd_t"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.gvstats_rec")=565
	Set gtmtypes("shmpool_buff_hdr",566,"name")="shmpool_buff_hdr.shadow_file_header.gvstats_rec.n_set"
	Set gtmtypes("shmpool_buff_hdr",566,"off")=7096
	Set gtmtypes("shmpool_buff_hdr",566,"len")=8
	Set gtmtypes("shmpool_buff_hdr",566,"type")="uint64_t"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.gvstats_rec.n_set")=566
	Set gtmtypes("shmpool_buff_hdr",567,"name")="shmpool_buff_hdr.shadow_file_header.gvstats_rec.n_kill"
	Set gtmtypes("shmpool_buff_hdr",567,"off")=7104
	Set gtmtypes("shmpool_buff_hdr",567,"len")=8
	Set gtmtypes("shmpool_buff_hdr",567,"type")="uint64_t"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.gvstats_rec.n_kill")=567
	Set gtmtypes("shmpool_buff_hdr",568,"name")="shmpool_buff_hdr.shadow_file_header.gvstats_rec.n_get"
	Set gtmtypes("shmpool_buff_hdr",568,"off")=7112
	Set gtmtypes("shmpool_buff_hdr",568,"len")=8
	Set gtmtypes("shmpool_buff_hdr",568,"type")="uint64_t"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.gvstats_rec.n_get")=568
	Set gtmtypes("shmpool_buff_hdr",569,"name")="shmpool_buff_hdr.shadow_file_header.gvstats_rec.n_data"
	Set gtmtypes("shmpool_buff_hdr",569,"off")=7120
	Set gtmtypes("shmpool_buff_hdr",569,"len")=8
	Set gtmtypes("shmpool_buff_hdr",569,"type")="uint64_t"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.gvstats_rec.n_data")=569
	Set gtmtypes("shmpool_buff_hdr",570,"name")="shmpool_buff_hdr.shadow_file_header.gvstats_rec.n_order"
	Set gtmtypes("shmpool_buff_hdr",570,"off")=7128
	Set gtmtypes("shmpool_buff_hdr",570,"len")=8
	Set gtmtypes("shmpool_buff_hdr",570,"type")="uint64_t"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.gvstats_rec.n_order")=570
	Set gtmtypes("shmpool_buff_hdr",571,"name")="shmpool_buff_hdr.shadow_file_header.gvstats_rec.n_zprev"
	Set gtmtypes("shmpool_buff_hdr",571,"off")=7136
	Set gtmtypes("shmpool_buff_hdr",571,"len")=8
	Set gtmtypes("shmpool_buff_hdr",571,"type")="uint64_t"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.gvstats_rec.n_zprev")=571
	Set gtmtypes("shmpool_buff_hdr",572,"name")="shmpool_buff_hdr.shadow_file_header.gvstats_rec.n_query"
	Set gtmtypes("shmpool_buff_hdr",572,"off")=7144
	Set gtmtypes("shmpool_buff_hdr",572,"len")=8
	Set gtmtypes("shmpool_buff_hdr",572,"type")="uint64_t"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.gvstats_rec.n_query")=572
	Set gtmtypes("shmpool_buff_hdr",573,"name")="shmpool_buff_hdr.shadow_file_header.gvstats_rec.n_lock_success"
	Set gtmtypes("shmpool_buff_hdr",573,"off")=7152
	Set gtmtypes("shmpool_buff_hdr",573,"len")=8
	Set gtmtypes("shmpool_buff_hdr",573,"type")="uint64_t"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.gvstats_rec.n_lock_success")=573
	Set gtmtypes("shmpool_buff_hdr",574,"name")="shmpool_buff_hdr.shadow_file_header.gvstats_rec.n_lock_fail"
	Set gtmtypes("shmpool_buff_hdr",574,"off")=7160
	Set gtmtypes("shmpool_buff_hdr",574,"len")=8
	Set gtmtypes("shmpool_buff_hdr",574,"type")="uint64_t"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.gvstats_rec.n_lock_fail")=574
	Set gtmtypes("shmpool_buff_hdr",575,"name")="shmpool_buff_hdr.shadow_file_header.gvstats_rec.db_curr_tn"
	Set gtmtypes("shmpool_buff_hdr",575,"off")=7168
	Set gtmtypes("shmpool_buff_hdr",575,"len")=8
	Set gtmtypes("shmpool_buff_hdr",575,"type")="uint64_t"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.gvstats_rec.db_curr_tn")=575
	Set gtmtypes("shmpool_buff_hdr",576,"name")="shmpool_buff_hdr.shadow_file_header.gvstats_rec.n_dsk_read"
	Set gtmtypes("shmpool_buff_hdr",576,"off")=7176
	Set gtmtypes("shmpool_buff_hdr",576,"len")=8
	Set gtmtypes("shmpool_buff_hdr",576,"type")="uint64_t"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.gvstats_rec.n_dsk_read")=576
	Set gtmtypes("shmpool_buff_hdr",577,"name")="shmpool_buff_hdr.shadow_file_header.gvstats_rec.n_dsk_write"
	Set gtmtypes("shmpool_buff_hdr",577,"off")=7184
	Set gtmtypes("shmpool_buff_hdr",577,"len")=8
	Set gtmtypes("shmpool_buff_hdr",577,"type")="uint64_t"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.gvstats_rec.n_dsk_write")=577
	Set gtmtypes("shmpool_buff_hdr",578,"name")="shmpool_buff_hdr.shadow_file_header.gvstats_rec.n_nontp_readwrite"
	Set gtmtypes("shmpool_buff_hdr",578,"off")=7192
	Set gtmtypes("shmpool_buff_hdr",578,"len")=8
	Set gtmtypes("shmpool_buff_hdr",578,"type")="uint64_t"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.gvstats_rec.n_nontp_readwrite")=578
	Set gtmtypes("shmpool_buff_hdr",579,"name")="shmpool_buff_hdr.shadow_file_header.gvstats_rec.n_nontp_readonly"
	Set gtmtypes("shmpool_buff_hdr",579,"off")=7200
	Set gtmtypes("shmpool_buff_hdr",579,"len")=8
	Set gtmtypes("shmpool_buff_hdr",579,"type")="uint64_t"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.gvstats_rec.n_nontp_readonly")=579
	Set gtmtypes("shmpool_buff_hdr",580,"name")="shmpool_buff_hdr.shadow_file_header.gvstats_rec.n_nontp_blkwrite"
	Set gtmtypes("shmpool_buff_hdr",580,"off")=7208
	Set gtmtypes("shmpool_buff_hdr",580,"len")=8
	Set gtmtypes("shmpool_buff_hdr",580,"type")="uint64_t"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.gvstats_rec.n_nontp_blkwrite")=580
	Set gtmtypes("shmpool_buff_hdr",581,"name")="shmpool_buff_hdr.shadow_file_header.gvstats_rec.n_nontp_blkread"
	Set gtmtypes("shmpool_buff_hdr",581,"off")=7216
	Set gtmtypes("shmpool_buff_hdr",581,"len")=8
	Set gtmtypes("shmpool_buff_hdr",581,"type")="uint64_t"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.gvstats_rec.n_nontp_blkread")=581
	Set gtmtypes("shmpool_buff_hdr",582,"name")="shmpool_buff_hdr.shadow_file_header.gvstats_rec.n_nontp_retries_0"
	Set gtmtypes("shmpool_buff_hdr",582,"off")=7224
	Set gtmtypes("shmpool_buff_hdr",582,"len")=8
	Set gtmtypes("shmpool_buff_hdr",582,"type")="uint64_t"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.gvstats_rec.n_nontp_retries_0")=582
	Set gtmtypes("shmpool_buff_hdr",583,"name")="shmpool_buff_hdr.shadow_file_header.gvstats_rec.n_nontp_retries_1"
	Set gtmtypes("shmpool_buff_hdr",583,"off")=7232
	Set gtmtypes("shmpool_buff_hdr",583,"len")=8
	Set gtmtypes("shmpool_buff_hdr",583,"type")="uint64_t"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.gvstats_rec.n_nontp_retries_1")=583
	Set gtmtypes("shmpool_buff_hdr",584,"name")="shmpool_buff_hdr.shadow_file_header.gvstats_rec.n_nontp_retries_2"
	Set gtmtypes("shmpool_buff_hdr",584,"off")=7240
	Set gtmtypes("shmpool_buff_hdr",584,"len")=8
	Set gtmtypes("shmpool_buff_hdr",584,"type")="uint64_t"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.gvstats_rec.n_nontp_retries_2")=584
	Set gtmtypes("shmpool_buff_hdr",585,"name")="shmpool_buff_hdr.shadow_file_header.gvstats_rec.n_nontp_retries_3"
	Set gtmtypes("shmpool_buff_hdr",585,"off")=7248
	Set gtmtypes("shmpool_buff_hdr",585,"len")=8
	Set gtmtypes("shmpool_buff_hdr",585,"type")="uint64_t"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.gvstats_rec.n_nontp_retries_3")=585
	Set gtmtypes("shmpool_buff_hdr",586,"name")="shmpool_buff_hdr.shadow_file_header.gvstats_rec.n_tp_readwrite"
	Set gtmtypes("shmpool_buff_hdr",586,"off")=7256
	Set gtmtypes("shmpool_buff_hdr",586,"len")=8
	Set gtmtypes("shmpool_buff_hdr",586,"type")="uint64_t"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.gvstats_rec.n_tp_readwrite")=586
	Set gtmtypes("shmpool_buff_hdr",587,"name")="shmpool_buff_hdr.shadow_file_header.gvstats_rec.n_tp_readonly"
	Set gtmtypes("shmpool_buff_hdr",587,"off")=7264
	Set gtmtypes("shmpool_buff_hdr",587,"len")=8
	Set gtmtypes("shmpool_buff_hdr",587,"type")="uint64_t"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.gvstats_rec.n_tp_readonly")=587
	Set gtmtypes("shmpool_buff_hdr",588,"name")="shmpool_buff_hdr.shadow_file_header.gvstats_rec.n_tp_rolledback"
	Set gtmtypes("shmpool_buff_hdr",588,"off")=7272
	Set gtmtypes("shmpool_buff_hdr",588,"len")=8
	Set gtmtypes("shmpool_buff_hdr",588,"type")="uint64_t"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.gvstats_rec.n_tp_rolledback")=588
	Set gtmtypes("shmpool_buff_hdr",589,"name")="shmpool_buff_hdr.shadow_file_header.gvstats_rec.n_tp_blkwrite"
	Set gtmtypes("shmpool_buff_hdr",589,"off")=7280
	Set gtmtypes("shmpool_buff_hdr",589,"len")=8
	Set gtmtypes("shmpool_buff_hdr",589,"type")="uint64_t"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.gvstats_rec.n_tp_blkwrite")=589
	Set gtmtypes("shmpool_buff_hdr",590,"name")="shmpool_buff_hdr.shadow_file_header.gvstats_rec.n_tp_blkread"
	Set gtmtypes("shmpool_buff_hdr",590,"off")=7288
	Set gtmtypes("shmpool_buff_hdr",590,"len")=8
	Set gtmtypes("shmpool_buff_hdr",590,"type")="uint64_t"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.gvstats_rec.n_tp_blkread")=590
	Set gtmtypes("shmpool_buff_hdr",591,"name")="shmpool_buff_hdr.shadow_file_header.gvstats_rec.n_tp_tot_retries_0"
	Set gtmtypes("shmpool_buff_hdr",591,"off")=7296
	Set gtmtypes("shmpool_buff_hdr",591,"len")=8
	Set gtmtypes("shmpool_buff_hdr",591,"type")="uint64_t"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.gvstats_rec.n_tp_tot_retries_0")=591
	Set gtmtypes("shmpool_buff_hdr",592,"name")="shmpool_buff_hdr.shadow_file_header.gvstats_rec.n_tp_tot_retries_1"
	Set gtmtypes("shmpool_buff_hdr",592,"off")=7304
	Set gtmtypes("shmpool_buff_hdr",592,"len")=8
	Set gtmtypes("shmpool_buff_hdr",592,"type")="uint64_t"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.gvstats_rec.n_tp_tot_retries_1")=592
	Set gtmtypes("shmpool_buff_hdr",593,"name")="shmpool_buff_hdr.shadow_file_header.gvstats_rec.n_tp_tot_retries_2"
	Set gtmtypes("shmpool_buff_hdr",593,"off")=7312
	Set gtmtypes("shmpool_buff_hdr",593,"len")=8
	Set gtmtypes("shmpool_buff_hdr",593,"type")="uint64_t"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.gvstats_rec.n_tp_tot_retries_2")=593
	Set gtmtypes("shmpool_buff_hdr",594,"name")="shmpool_buff_hdr.shadow_file_header.gvstats_rec.n_tp_tot_retries_3"
	Set gtmtypes("shmpool_buff_hdr",594,"off")=7320
	Set gtmtypes("shmpool_buff_hdr",594,"len")=8
	Set gtmtypes("shmpool_buff_hdr",594,"type")="uint64_t"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.gvstats_rec.n_tp_tot_retries_3")=594
	Set gtmtypes("shmpool_buff_hdr",595,"name")="shmpool_buff_hdr.shadow_file_header.gvstats_rec.n_tp_tot_retries_4"
	Set gtmtypes("shmpool_buff_hdr",595,"off")=7328
	Set gtmtypes("shmpool_buff_hdr",595,"len")=8
	Set gtmtypes("shmpool_buff_hdr",595,"type")="uint64_t"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.gvstats_rec.n_tp_tot_retries_4")=595
	Set gtmtypes("shmpool_buff_hdr",596,"name")="shmpool_buff_hdr.shadow_file_header.gvstats_rec.n_tp_cnflct_retries_0"
	Set gtmtypes("shmpool_buff_hdr",596,"off")=7336
	Set gtmtypes("shmpool_buff_hdr",596,"len")=8
	Set gtmtypes("shmpool_buff_hdr",596,"type")="uint64_t"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.gvstats_rec.n_tp_cnflct_retries_0")=596
	Set gtmtypes("shmpool_buff_hdr",597,"name")="shmpool_buff_hdr.shadow_file_header.gvstats_rec.n_tp_cnflct_retries_1"
	Set gtmtypes("shmpool_buff_hdr",597,"off")=7344
	Set gtmtypes("shmpool_buff_hdr",597,"len")=8
	Set gtmtypes("shmpool_buff_hdr",597,"type")="uint64_t"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.gvstats_rec.n_tp_cnflct_retries_1")=597
	Set gtmtypes("shmpool_buff_hdr",598,"name")="shmpool_buff_hdr.shadow_file_header.gvstats_rec.n_tp_cnflct_retries_2"
	Set gtmtypes("shmpool_buff_hdr",598,"off")=7352
	Set gtmtypes("shmpool_buff_hdr",598,"len")=8
	Set gtmtypes("shmpool_buff_hdr",598,"type")="uint64_t"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.gvstats_rec.n_tp_cnflct_retries_2")=598
	Set gtmtypes("shmpool_buff_hdr",599,"name")="shmpool_buff_hdr.shadow_file_header.gvstats_rec.n_tp_cnflct_retries_3"
	Set gtmtypes("shmpool_buff_hdr",599,"off")=7360
	Set gtmtypes("shmpool_buff_hdr",599,"len")=8
	Set gtmtypes("shmpool_buff_hdr",599,"type")="uint64_t"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.gvstats_rec.n_tp_cnflct_retries_3")=599
	Set gtmtypes("shmpool_buff_hdr",600,"name")="shmpool_buff_hdr.shadow_file_header.gvstats_rec.n_tp_cnflct_retries_4"
	Set gtmtypes("shmpool_buff_hdr",600,"off")=7368
	Set gtmtypes("shmpool_buff_hdr",600,"len")=8
	Set gtmtypes("shmpool_buff_hdr",600,"type")="uint64_t"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.gvstats_rec.n_tp_cnflct_retries_4")=600
	Set gtmtypes("shmpool_buff_hdr",601,"name")="shmpool_buff_hdr.shadow_file_header.gvstats_rec.n_ztrigger"
	Set gtmtypes("shmpool_buff_hdr",601,"off")=7376
	Set gtmtypes("shmpool_buff_hdr",601,"len")=8
	Set gtmtypes("shmpool_buff_hdr",601,"type")="uint64_t"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.gvstats_rec.n_ztrigger")=601
	Set gtmtypes("shmpool_buff_hdr",602,"name")="shmpool_buff_hdr.shadow_file_header.gvstats_rec.n_db_flush"
	Set gtmtypes("shmpool_buff_hdr",602,"off")=7384
	Set gtmtypes("shmpool_buff_hdr",602,"len")=8
	Set gtmtypes("shmpool_buff_hdr",602,"type")="uint64_t"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.gvstats_rec.n_db_flush")=602
	Set gtmtypes("shmpool_buff_hdr",603,"name")="shmpool_buff_hdr.shadow_file_header.gvstats_rec.n_db_fsync"
	Set gtmtypes("shmpool_buff_hdr",603,"off")=7392
	Set gtmtypes("shmpool_buff_hdr",603,"len")=8
	Set gtmtypes("shmpool_buff_hdr",603,"type")="uint64_t"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.gvstats_rec.n_db_fsync")=603
	Set gtmtypes("shmpool_buff_hdr",604,"name")="shmpool_buff_hdr.shadow_file_header.gvstats_rec.n_jnl_flush"
	Set gtmtypes("shmpool_buff_hdr",604,"off")=7400
	Set gtmtypes("shmpool_buff_hdr",604,"len")=8
	Set gtmtypes("shmpool_buff_hdr",604,"type")="uint64_t"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.gvstats_rec.n_jnl_flush")=604
	Set gtmtypes("shmpool_buff_hdr",605,"name")="shmpool_buff_hdr.shadow_file_header.gvstats_rec.n_jnl_fsync"
	Set gtmtypes("shmpool_buff_hdr",605,"off")=7408
	Set gtmtypes("shmpool_buff_hdr",605,"len")=8
	Set gtmtypes("shmpool_buff_hdr",605,"type")="uint64_t"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.gvstats_rec.n_jnl_fsync")=605
	Set gtmtypes("shmpool_buff_hdr",606,"name")="shmpool_buff_hdr.shadow_file_header.gvstats_rec.n_jbuff_bytes"
	Set gtmtypes("shmpool_buff_hdr",606,"off")=7416
	Set gtmtypes("shmpool_buff_hdr",606,"len")=8
	Set gtmtypes("shmpool_buff_hdr",606,"type")="uint64_t"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.gvstats_rec.n_jbuff_bytes")=606
	Set gtmtypes("shmpool_buff_hdr",607,"name")="shmpool_buff_hdr.shadow_file_header.gvstats_rec.n_jfile_bytes"
	Set gtmtypes("shmpool_buff_hdr",607,"off")=7424
	Set gtmtypes("shmpool_buff_hdr",607,"len")=8
	Set gtmtypes("shmpool_buff_hdr",607,"type")="uint64_t"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.gvstats_rec.n_jfile_bytes")=607
	Set gtmtypes("shmpool_buff_hdr",608,"name")="shmpool_buff_hdr.shadow_file_header.gvstats_rec.n_jfile_writes"
	Set gtmtypes("shmpool_buff_hdr",608,"off")=7432
	Set gtmtypes("shmpool_buff_hdr",608,"len")=8
	Set gtmtypes("shmpool_buff_hdr",608,"type")="uint64_t"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.gvstats_rec.n_jfile_writes")=608
	Set gtmtypes("shmpool_buff_hdr",609,"name")="shmpool_buff_hdr.shadow_file_header.gvstats_rec.n_jrec_logical"
	Set gtmtypes("shmpool_buff_hdr",609,"off")=7440
	Set gtmtypes("shmpool_buff_hdr",609,"len")=8
	Set gtmtypes("shmpool_buff_hdr",609,"type")="uint64_t"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.gvstats_rec.n_jrec_logical")=609
	Set gtmtypes("shmpool_buff_hdr",610,"name")="shmpool_buff_hdr.shadow_file_header.gvstats_rec.n_jrec_pblk"
	Set gtmtypes("shmpool_buff_hdr",610,"off")=7448
	Set gtmtypes("shmpool_buff_hdr",610,"len")=8
	Set gtmtypes("shmpool_buff_hdr",610,"type")="uint64_t"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.gvstats_rec.n_jrec_pblk")=610
	Set gtmtypes("shmpool_buff_hdr",611,"name")="shmpool_buff_hdr.shadow_file_header.gvstats_rec.n_jrec_epoch_regular"
	Set gtmtypes("shmpool_buff_hdr",611,"off")=7456
	Set gtmtypes("shmpool_buff_hdr",611,"len")=8
	Set gtmtypes("shmpool_buff_hdr",611,"type")="uint64_t"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.gvstats_rec.n_jrec_epoch_regular")=611
	Set gtmtypes("shmpool_buff_hdr",612,"name")="shmpool_buff_hdr.shadow_file_header.gvstats_rec.n_jrec_epoch_idle"
	Set gtmtypes("shmpool_buff_hdr",612,"off")=7464
	Set gtmtypes("shmpool_buff_hdr",612,"len")=8
	Set gtmtypes("shmpool_buff_hdr",612,"type")="uint64_t"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.gvstats_rec.n_jrec_epoch_idle")=612
	Set gtmtypes("shmpool_buff_hdr",613,"name")="shmpool_buff_hdr.shadow_file_header.gvstats_rec.n_jrec_other"
	Set gtmtypes("shmpool_buff_hdr",613,"off")=7472
	Set gtmtypes("shmpool_buff_hdr",613,"len")=8
	Set gtmtypes("shmpool_buff_hdr",613,"type")="uint64_t"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.gvstats_rec.n_jrec_other")=613
	Set gtmtypes("shmpool_buff_hdr",614,"name")="shmpool_buff_hdr.shadow_file_header.gvstats_rec.n_jnl_extends"
	Set gtmtypes("shmpool_buff_hdr",614,"off")=7480
	Set gtmtypes("shmpool_buff_hdr",614,"len")=8
	Set gtmtypes("shmpool_buff_hdr",614,"type")="uint64_t"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.gvstats_rec.n_jnl_extends")=614
	Set gtmtypes("shmpool_buff_hdr",615,"name")="shmpool_buff_hdr.shadow_file_header.gvstats_rec.n_db_extends"
	Set gtmtypes("shmpool_buff_hdr",615,"off")=7488
	Set gtmtypes("shmpool_buff_hdr",615,"len")=8
	Set gtmtypes("shmpool_buff_hdr",615,"type")="uint64_t"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.gvstats_rec.n_db_extends")=615
	Set gtmtypes("shmpool_buff_hdr",616,"name")="shmpool_buff_hdr.shadow_file_header.gvstats_rec.n_crit_success"
	Set gtmtypes("shmpool_buff_hdr",616,"off")=7496
	Set gtmtypes("shmpool_buff_hdr",616,"len")=8
	Set gtmtypes("shmpool_buff_hdr",616,"type")="uint64_t"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.gvstats_rec.n_crit_success")=616
	Set gtmtypes("shmpool_buff_hdr",617,"name")="shmpool_buff_hdr.shadow_file_header.gvstats_rec.n_crits_in_epch"
	Set gtmtypes("shmpool_buff_hdr",617,"off")=7504
	Set gtmtypes("shmpool_buff_hdr",617,"len")=8
	Set gtmtypes("shmpool_buff_hdr",617,"type")="uint64_t"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.gvstats_rec.n_crits_in_epch")=617
	Set gtmtypes("shmpool_buff_hdr",618,"name")="shmpool_buff_hdr.shadow_file_header.gvstats_rec.sq_crit_failed"
	Set gtmtypes("shmpool_buff_hdr",618,"off")=7512
	Set gtmtypes("shmpool_buff_hdr",618,"len")=8
	Set gtmtypes("shmpool_buff_hdr",618,"type")="uint64_t"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.gvstats_rec.sq_crit_failed")=618
	Set gtmtypes("shmpool_buff_hdr",619,"name")="shmpool_buff_hdr.shadow_file_header.gvstats_rec.n_crit_failed"
	Set gtmtypes("shmpool_buff_hdr",619,"off")=7520
	Set gtmtypes("shmpool_buff_hdr",619,"len")=8
	Set gtmtypes("shmpool_buff_hdr",619,"type")="uint64_t"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.gvstats_rec.n_crit_failed")=619
	Set gtmtypes("shmpool_buff_hdr",620,"name")="shmpool_buff_hdr.shadow_file_header.gvstats_rec.sq_crit_que_slps"
	Set gtmtypes("shmpool_buff_hdr",620,"off")=7528
	Set gtmtypes("shmpool_buff_hdr",620,"len")=8
	Set gtmtypes("shmpool_buff_hdr",620,"type")="uint64_t"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.gvstats_rec.sq_crit_que_slps")=620
	Set gtmtypes("shmpool_buff_hdr",621,"name")="shmpool_buff_hdr.shadow_file_header.gvstats_rec.n_crit_que_slps"
	Set gtmtypes("shmpool_buff_hdr",621,"off")=7536
	Set gtmtypes("shmpool_buff_hdr",621,"len")=8
	Set gtmtypes("shmpool_buff_hdr",621,"type")="uint64_t"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.gvstats_rec.n_crit_que_slps")=621
	Set gtmtypes("shmpool_buff_hdr",622,"name")="shmpool_buff_hdr.shadow_file_header.gvstats_rec.sq_crit_yields"
	Set gtmtypes("shmpool_buff_hdr",622,"off")=7544
	Set gtmtypes("shmpool_buff_hdr",622,"len")=8
	Set gtmtypes("shmpool_buff_hdr",622,"type")="uint64_t"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.gvstats_rec.sq_crit_yields")=622
	Set gtmtypes("shmpool_buff_hdr",623,"name")="shmpool_buff_hdr.shadow_file_header.gvstats_rec.n_crit_yields"
	Set gtmtypes("shmpool_buff_hdr",623,"off")=7552
	Set gtmtypes("shmpool_buff_hdr",623,"len")=8
	Set gtmtypes("shmpool_buff_hdr",623,"type")="uint64_t"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.gvstats_rec.n_crit_yields")=623
	Set gtmtypes("shmpool_buff_hdr",624,"name")="shmpool_buff_hdr.shadow_file_header.gvstats_rec.n_clean2dirty"
	Set gtmtypes("shmpool_buff_hdr",624,"off")=7560
	Set gtmtypes("shmpool_buff_hdr",624,"len")=8
	Set gtmtypes("shmpool_buff_hdr",624,"type")="uint64_t"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.gvstats_rec.n_clean2dirty")=624
	Set gtmtypes("shmpool_buff_hdr",625,"name")="shmpool_buff_hdr.shadow_file_header.gvstats_rec.n_wait_for_read"
	Set gtmtypes("shmpool_buff_hdr",625,"off")=7568
	Set gtmtypes("shmpool_buff_hdr",625,"len")=8
	Set gtmtypes("shmpool_buff_hdr",625,"type")="uint64_t"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.gvstats_rec.n_wait_for_read")=625
	Set gtmtypes("shmpool_buff_hdr",626,"name")="shmpool_buff_hdr.shadow_file_header.gvstats_rec.n_buffer_scarce"
	Set gtmtypes("shmpool_buff_hdr",626,"off")=7576
	Set gtmtypes("shmpool_buff_hdr",626,"len")=8
	Set gtmtypes("shmpool_buff_hdr",626,"type")="uint64_t"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.gvstats_rec.n_buffer_scarce")=626
	Set gtmtypes("shmpool_buff_hdr",627,"name")="shmpool_buff_hdr.shadow_file_header.gvstats_rec.n_bt_scarce"
	Set gtmtypes("shmpool_buff_hdr",627,"off")=7584
	Set gtmtypes("shmpool_buff_hdr",627,"len")=8
	Set gtmtypes("shmpool_buff_hdr",627,"type")="uint64_t"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.gvstats_rec.n_bt_scarce")=627
	Set gtmtypes("shmpool_buff_hdr",628,"name")="shmpool_buff_hdr.shadow_file_header.gvstats_rec.f_dbext_wait"
	Set gtmtypes("shmpool_buff_hdr",628,"off")=7592
	Set gtmtypes("shmpool_buff_hdr",628,"len")=8
	Set gtmtypes("shmpool_buff_hdr",628,"type")="uint64_t"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.gvstats_rec.f_dbext_wait")=628
	Set gtmtypes("shmpool_buff_hdr",629,"name")="shmpool_buff_hdr.shadow_file_header.gvstats_rec.f_bg_wait"
	Set gtmtypes("shmpool_buff_hdr",629,"off")=7600
	Set gtmtypes("shmpool_buff_hdr",629,"len")=8
	Set gtmtypes("shmpool_buff_hdr",629,"type")="uint64_t"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.gvstats_rec.f_bg_wait")=629
	Set gtmtypes("shmpool_buff_hdr",630,"name")="shmpool_buff_hdr.shadow_file_header.gvstats_rec.f_jnl_wait"
	Set gtmtypes("shmpool_buff_hdr",630,"off")=7608
	Set gtmtypes("shmpool_buff_hdr",630,"len")=8
	Set gtmtypes("shmpool_buff_hdr",630,"type")="uint64_t"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.gvstats_rec.f_jnl_wait")=630
	Set gtmtypes("shmpool_buff_hdr",631,"name")="shmpool_buff_hdr.shadow_file_header.gvstats_rec.f_mlk_wait"
	Set gtmtypes("shmpool_buff_hdr",631,"off")=7616
	Set gtmtypes("shmpool_buff_hdr",631,"len")=8
	Set gtmtypes("shmpool_buff_hdr",631,"type")="uint64_t"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.gvstats_rec.f_mlk_wait")=631
	Set gtmtypes("shmpool_buff_hdr",632,"name")="shmpool_buff_hdr.shadow_file_header.gvstats_rec.f_proc_wait"
	Set gtmtypes("shmpool_buff_hdr",632,"off")=7624
	Set gtmtypes("shmpool_buff_hdr",632,"len")=8
	Set gtmtypes("shmpool_buff_hdr",632,"type")="uint64_t"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.gvstats_rec.f_proc_wait")=632
	Set gtmtypes("shmpool_buff_hdr",633,"name")="shmpool_buff_hdr.shadow_file_header.gvstats_rec.f_trans_wait"
	Set gtmtypes("shmpool_buff_hdr",633,"off")=7632
	Set gtmtypes("shmpool_buff_hdr",633,"len")=8
	Set gtmtypes("shmpool_buff_hdr",633,"type")="uint64_t"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.gvstats_rec.f_trans_wait")=633
	Set gtmtypes("shmpool_buff_hdr",634,"name")="shmpool_buff_hdr.shadow_file_header.gvstats_rec.f_util_wait"
	Set gtmtypes("shmpool_buff_hdr",634,"off")=7640
	Set gtmtypes("shmpool_buff_hdr",634,"len")=8
	Set gtmtypes("shmpool_buff_hdr",634,"type")="uint64_t"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.gvstats_rec.f_util_wait")=634
	Set gtmtypes("shmpool_buff_hdr",635,"name")="shmpool_buff_hdr.shadow_file_header.gvstats_rec.f_ws2"
	Set gtmtypes("shmpool_buff_hdr",635,"off")=7648
	Set gtmtypes("shmpool_buff_hdr",635,"len")=8
	Set gtmtypes("shmpool_buff_hdr",635,"type")="uint64_t"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.gvstats_rec.f_ws2")=635
	Set gtmtypes("shmpool_buff_hdr",636,"name")="shmpool_buff_hdr.shadow_file_header.gvstats_rec.f_ws12"
	Set gtmtypes("shmpool_buff_hdr",636,"off")=7656
	Set gtmtypes("shmpool_buff_hdr",636,"len")=8
	Set gtmtypes("shmpool_buff_hdr",636,"type")="uint64_t"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.gvstats_rec.f_ws12")=636
	Set gtmtypes("shmpool_buff_hdr",637,"name")="shmpool_buff_hdr.shadow_file_header.gvstats_rec.f_ws15"
	Set gtmtypes("shmpool_buff_hdr",637,"off")=7664
	Set gtmtypes("shmpool_buff_hdr",637,"len")=8
	Set gtmtypes("shmpool_buff_hdr",637,"type")="uint64_t"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.gvstats_rec.f_ws15")=637
	Set gtmtypes("shmpool_buff_hdr",638,"name")="shmpool_buff_hdr.shadow_file_header.gvstats_rec.f_ws39"
	Set gtmtypes("shmpool_buff_hdr",638,"off")=7672
	Set gtmtypes("shmpool_buff_hdr",638,"len")=8
	Set gtmtypes("shmpool_buff_hdr",638,"type")="uint64_t"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.gvstats_rec.f_ws39")=638
	Set gtmtypes("shmpool_buff_hdr",639,"name")="shmpool_buff_hdr.shadow_file_header.gvstats_rec.f_ws47"
	Set gtmtypes("shmpool_buff_hdr",639,"off")=7680
	Set gtmtypes("shmpool_buff_hdr",639,"len")=8
	Set gtmtypes("shmpool_buff_hdr",639,"type")="uint64_t"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.gvstats_rec.f_ws47")=639
	Set gtmtypes("shmpool_buff_hdr",640,"name")="shmpool_buff_hdr.shadow_file_header.filler_8k"
	Set gtmtypes("shmpool_buff_hdr",640,"off")=7688
	Set gtmtypes("shmpool_buff_hdr",640,"len")=872
	Set gtmtypes("shmpool_buff_hdr",640,"type")="char"
	Set gtmtypfldindx("shmpool_buff_hdr","shadow_file_header.filler_8k")=640
	;
	Set gtmtypes("show_reply")="struct"
	Set gtmtypes("show_reply",0)=2
	Set gtmtypes("show_reply","len")=257
	Set gtmtypes("show_reply",1,"name")="show_reply.code"
	Set gtmtypes("show_reply",1,"off")=0
	Set gtmtypes("show_reply",1,"len")=1
	Set gtmtypes("show_reply",1,"type")="char"
	Set gtmtypfldindx("show_reply","code")=1
	Set gtmtypes("show_reply",2,"name")="show_reply.line"
	Set gtmtypes("show_reply",2,"off")=1
	Set gtmtypes("show_reply",2,"len")=256
	Set gtmtypes("show_reply",2,"type")="char"
	Set gtmtypfldindx("show_reply","line")=2
	;
	Set gtmtypes("show_request")="struct"
	Set gtmtypes("show_request",0)=7
	Set gtmtypes("show_request","len")=44
	Set gtmtypes("show_request",1,"name")="show_request.code"
	Set gtmtypes("show_request",1,"off")=0
	Set gtmtypes("show_request",1,"len")=1
	Set gtmtypes("show_request",1,"type")="char"
	Set gtmtypfldindx("show_request","code")=1
	Set gtmtypes("show_request",2,"name")="show_request.rnum"
	Set gtmtypes("show_request",2,"off")=1
	Set gtmtypes("show_request",2,"len")=1
	Set gtmtypes("show_request",2,"type")="char"
	Set gtmtypfldindx("show_request","rnum")=2
	Set gtmtypes("show_request",3,"name")="show_request.all"
	Set gtmtypes("show_request",3,"off")=2
	Set gtmtypes("show_request",3,"len")=1
	Set gtmtypes("show_request",3,"type")="char"
	Set gtmtypfldindx("show_request","all")=3
	Set gtmtypes("show_request",4,"name")="show_request.wait"
	Set gtmtypes("show_request",4,"off")=3
	Set gtmtypes("show_request",4,"len")=1
	Set gtmtypes("show_request",4,"type")="char"
	Set gtmtypfldindx("show_request","wait")=4
	Set gtmtypes("show_request",5,"name")="show_request.pid"
	Set gtmtypes("show_request",5,"off")=4
	Set gtmtypes("show_request",5,"len")=4
	Set gtmtypes("show_request",5,"type")="int"
	Set gtmtypfldindx("show_request","pid")=5
	Set gtmtypes("show_request",6,"name")="show_request.nodelength"
	Set gtmtypes("show_request",6,"off")=8
	Set gtmtypes("show_request",6,"len")=1
	Set gtmtypes("show_request",6,"type")="char"
	Set gtmtypfldindx("show_request","nodelength")=6
	Set gtmtypes("show_request",7,"name")="show_request.node"
	Set gtmtypes("show_request",7,"off")=9
	Set gtmtypes("show_request",7,"len")=32
	Set gtmtypes("show_request",7,"type")="char"
	Set gtmtypfldindx("show_request","node")=7
	;
	Set gtmtypes("sib_byte_type")="union"
	Set gtmtypes("sib_byte_type",0)=2
	Set gtmtypes("sib_byte_type","len")=1
	Set gtmtypes("sib_byte_type",1,"name")="sib_byte_type.sib"
	Set gtmtypes("sib_byte_type",1,"off")=0
	Set gtmtypes("sib_byte_type",1,"len")=1
	Set gtmtypes("sib_byte_type",1,"type")="SIB"
	Set gtmtypfldindx("sib_byte_type","sib")=1
	Set gtmtypes("sib_byte_type",2,"name")="sib_byte_type.byte"
	Set gtmtypes("sib_byte_type",2,"off")=0
	Set gtmtypes("sib_byte_type",2,"len")=1
	Set gtmtypes("sib_byte_type",2,"type")="unsigned-char"
	Set gtmtypfldindx("sib_byte_type","byte")=2
	;
	Set gtmtypes("snapshot_context_t")="struct"
	Set gtmtypes("snapshot_context_t",0)=13
	Set gtmtypes("snapshot_context_t","len")=4176
	Set gtmtypes("snapshot_context_t",1,"name")="snapshot_context_t.shdw_fd"
	Set gtmtypes("snapshot_context_t",1,"off")=0
	Set gtmtypes("snapshot_context_t",1,"len")=4
	Set gtmtypes("snapshot_context_t",1,"type")="int"
	Set gtmtypfldindx("snapshot_context_t","shdw_fd")=1
	Set gtmtypes("snapshot_context_t",2,"name")="snapshot_context_t.nl_shmid"
	Set gtmtypes("snapshot_context_t",2,"off")=8
	Set gtmtypes("snapshot_context_t",2,"len")=8
	Set gtmtypes("snapshot_context_t",2,"type")="long"
	Set gtmtypfldindx("snapshot_context_t","nl_shmid")=2
	Set gtmtypes("snapshot_context_t",3,"name")="snapshot_context_t.attach_shmid"
	Set gtmtypes("snapshot_context_t",3,"off")=16
	Set gtmtypes("snapshot_context_t",3,"len")=8
	Set gtmtypes("snapshot_context_t",3,"type")="long"
	Set gtmtypfldindx("snapshot_context_t","attach_shmid")=3
	Set gtmtypes("snapshot_context_t",4,"name")="snapshot_context_t.ss_shmcycle"
	Set gtmtypes("snapshot_context_t",4,"off")=24
	Set gtmtypes("snapshot_context_t",4,"len")=4
	Set gtmtypes("snapshot_context_t",4,"type")="int"
	Set gtmtypfldindx("snapshot_context_t","ss_shmcycle")=4
	Set gtmtypes("snapshot_context_t",5,"name")="snapshot_context_t.total_blks"
	Set gtmtypes("snapshot_context_t",5,"off")=28
	Set gtmtypes("snapshot_context_t",5,"len")=4
	Set gtmtypes("snapshot_context_t",5,"type")="int"
	Set gtmtypfldindx("snapshot_context_t","total_blks")=5
	Set gtmtypes("snapshot_context_t",6,"name")="snapshot_context_t.failure_errno"
	Set gtmtypes("snapshot_context_t",6,"off")=32
	Set gtmtypes("snapshot_context_t",6,"len")=4
	Set gtmtypes("snapshot_context_t",6,"type")="unsigned-int"
	Set gtmtypfldindx("snapshot_context_t","failure_errno")=6
	Set gtmtypes("snapshot_context_t",7,"name")="snapshot_context_t.ss_shm_ptr"
	Set gtmtypes("snapshot_context_t",7,"off")=40
	Set gtmtypes("snapshot_context_t",7,"len")=8
	Set gtmtypes("snapshot_context_t",7,"type")="addr"
	Set gtmtypfldindx("snapshot_context_t","ss_shm_ptr")=7
	Set gtmtypes("snapshot_context_t",8,"name")="snapshot_context_t.start_shmaddr"
	Set gtmtypes("snapshot_context_t",8,"off")=48
	Set gtmtypes("snapshot_context_t",8,"len")=8
	Set gtmtypes("snapshot_context_t",8,"type")="addr"
	Set gtmtypfldindx("snapshot_context_t","start_shmaddr")=8
	Set gtmtypes("snapshot_context_t",9,"name")="snapshot_context_t.bitmap_addr"
	Set gtmtypes("snapshot_context_t",9,"off")=56
	Set gtmtypes("snapshot_context_t",9,"len")=8
	Set gtmtypes("snapshot_context_t",9,"type")="addr"
	Set gtmtypfldindx("snapshot_context_t","bitmap_addr")=9
	Set gtmtypes("snapshot_context_t",10,"name")="snapshot_context_t.shadow_vbn"
	Set gtmtypes("snapshot_context_t",10,"off")=64
	Set gtmtypes("snapshot_context_t",10,"len")=4
	Set gtmtypes("snapshot_context_t",10,"type")="int"
	Set gtmtypfldindx("snapshot_context_t","shadow_vbn")=10
	Set gtmtypes("snapshot_context_t",11,"name")="snapshot_context_t.shadow_file"
	Set gtmtypes("snapshot_context_t",11,"off")=68
	Set gtmtypes("snapshot_context_t",11,"len")=4097
	Set gtmtypes("snapshot_context_t",11,"type")="char"
	Set gtmtypfldindx("snapshot_context_t","shadow_file")=11
	Set gtmtypes("snapshot_context_t",12,"name")="snapshot_context_t.cur_state"
	Set gtmtypes("snapshot_context_t",12,"off")=4168
	Set gtmtypes("snapshot_context_t",12,"len")=4
	Set gtmtypes("snapshot_context_t",12,"type")="int"
	Set gtmtypfldindx("snapshot_context_t","cur_state")=12
	Set gtmtypes("snapshot_context_t",13,"name")="snapshot_context_t.proc_property"
	Set gtmtypes("snapshot_context_t",13,"off")=4172
	Set gtmtypes("snapshot_context_t",13,"len")=4
	Set gtmtypes("snapshot_context_t",13,"type")="unsigned-int"
	Set gtmtypfldindx("snapshot_context_t","proc_property")=13
	;
	Set gtmtypes("snapshot_filhdr_t")="struct"
	Set gtmtypes("snapshot_filhdr_t",0)=13
	Set gtmtypes("snapshot_filhdr_t","len")=5152
	Set gtmtypes("snapshot_filhdr_t",1,"name")="snapshot_filhdr_t.label"
	Set gtmtypes("snapshot_filhdr_t",1,"off")=0
	Set gtmtypes("snapshot_filhdr_t",1,"len")=10
	Set gtmtypes("snapshot_filhdr_t",1,"type")="char"
	Set gtmtypfldindx("snapshot_filhdr_t","label")=1
	Set gtmtypes("snapshot_filhdr_t",2,"name")="snapshot_filhdr_t.ss_info"
	Set gtmtypes("snapshot_filhdr_t",2,"off")=16
	Set gtmtypes("snapshot_filhdr_t",2,"len")=4152
	Set gtmtypes("snapshot_filhdr_t",2,"type")="snapshot_info_t"
	Set gtmtypfldindx("snapshot_filhdr_t","ss_info")=2
	Set gtmtypes("snapshot_filhdr_t",3,"name")="snapshot_filhdr_t.ss_info.ss_pid"
	Set gtmtypes("snapshot_filhdr_t",3,"off")=16
	Set gtmtypes("snapshot_filhdr_t",3,"len")=4
	Set gtmtypes("snapshot_filhdr_t",3,"type")="unsigned-int"
	Set gtmtypfldindx("snapshot_filhdr_t","ss_info.ss_pid")=3
	Set gtmtypes("snapshot_filhdr_t",4,"name")="snapshot_filhdr_t.ss_info.snapshot_tn"
	Set gtmtypes("snapshot_filhdr_t",4,"off")=24
	Set gtmtypes("snapshot_filhdr_t",4,"len")=8
	Set gtmtypes("snapshot_filhdr_t",4,"type")="uint64_t"
	Set gtmtypfldindx("snapshot_filhdr_t","ss_info.snapshot_tn")=4
	Set gtmtypes("snapshot_filhdr_t",5,"name")="snapshot_filhdr_t.ss_info.db_blk_size"
	Set gtmtypes("snapshot_filhdr_t",5,"off")=32
	Set gtmtypes("snapshot_filhdr_t",5,"len")=4
	Set gtmtypes("snapshot_filhdr_t",5,"type")="unsigned-int"
	Set gtmtypfldindx("snapshot_filhdr_t","ss_info.db_blk_size")=5
	Set gtmtypes("snapshot_filhdr_t",6,"name")="snapshot_filhdr_t.ss_info.free_blks"
	Set gtmtypes("snapshot_filhdr_t",6,"off")=36
	Set gtmtypes("snapshot_filhdr_t",6,"len")=4
	Set gtmtypes("snapshot_filhdr_t",6,"type")="unsigned-int"
	Set gtmtypfldindx("snapshot_filhdr_t","ss_info.free_blks")=6
	Set gtmtypes("snapshot_filhdr_t",7,"name")="snapshot_filhdr_t.ss_info.total_blks"
	Set gtmtypes("snapshot_filhdr_t",7,"off")=40
	Set gtmtypes("snapshot_filhdr_t",7,"len")=4
	Set gtmtypes("snapshot_filhdr_t",7,"type")="unsigned-int"
	Set gtmtypfldindx("snapshot_filhdr_t","ss_info.total_blks")=7
	Set gtmtypes("snapshot_filhdr_t",8,"name")="snapshot_filhdr_t.ss_info.shadow_file"
	Set gtmtypes("snapshot_filhdr_t",8,"off")=44
	Set gtmtypes("snapshot_filhdr_t",8,"len")=4097
	Set gtmtypes("snapshot_filhdr_t",8,"type")="char"
	Set gtmtypfldindx("snapshot_filhdr_t","ss_info.shadow_file")=8
	Set gtmtypes("snapshot_filhdr_t",9,"name")="snapshot_filhdr_t.ss_info.shadow_vbn"
	Set gtmtypes("snapshot_filhdr_t",9,"off")=4144
	Set gtmtypes("snapshot_filhdr_t",9,"len")=4
	Set gtmtypes("snapshot_filhdr_t",9,"type")="int"
	Set gtmtypfldindx("snapshot_filhdr_t","ss_info.shadow_vbn")=9
	Set gtmtypes("snapshot_filhdr_t",10,"name")="snapshot_filhdr_t.ss_info.ss_shmid"
	Set gtmtypes("snapshot_filhdr_t",10,"off")=4152
	Set gtmtypes("snapshot_filhdr_t",10,"len")=8
	Set gtmtypes("snapshot_filhdr_t",10,"type")="long"
	Set gtmtypfldindx("snapshot_filhdr_t","ss_info.ss_shmid")=10
	Set gtmtypes("snapshot_filhdr_t",11,"name")="snapshot_filhdr_t.ss_info.ss_shmsize"
	Set gtmtypes("snapshot_filhdr_t",11,"off")=4160
	Set gtmtypes("snapshot_filhdr_t",11,"len")=4
	Set gtmtypes("snapshot_filhdr_t",11,"type")="int"
	Set gtmtypfldindx("snapshot_filhdr_t","ss_info.ss_shmsize")=11
	Set gtmtypes("snapshot_filhdr_t",12,"name")="snapshot_filhdr_t.shadow_file_len"
	Set gtmtypes("snapshot_filhdr_t",12,"off")=4168
	Set gtmtypes("snapshot_filhdr_t",12,"len")=4
	Set gtmtypes("snapshot_filhdr_t",12,"type")="int"
	Set gtmtypfldindx("snapshot_filhdr_t","shadow_file_len")=12
	Set gtmtypes("snapshot_filhdr_t",13,"name")="snapshot_filhdr_t.filler"
	Set gtmtypes("snapshot_filhdr_t",13,"off")=4172
	Set gtmtypes("snapshot_filhdr_t",13,"len")=976
	Set gtmtypes("snapshot_filhdr_t",13,"type")="unsigned-char"
	Set gtmtypfldindx("snapshot_filhdr_t","filler")=13
	;
	Set gtmtypes("snapshot_info_t")="struct"
	Set gtmtypes("snapshot_info_t",0)=9
	Set gtmtypes("snapshot_info_t","len")=4152
	Set gtmtypes("snapshot_info_t",1,"name")="snapshot_info_t.ss_pid"
	Set gtmtypes("snapshot_info_t",1,"off")=0
	Set gtmtypes("snapshot_info_t",1,"len")=4
	Set gtmtypes("snapshot_info_t",1,"type")="unsigned-int"
	Set gtmtypfldindx("snapshot_info_t","ss_pid")=1
	Set gtmtypes("snapshot_info_t",2,"name")="snapshot_info_t.snapshot_tn"
	Set gtmtypes("snapshot_info_t",2,"off")=8
	Set gtmtypes("snapshot_info_t",2,"len")=8
	Set gtmtypes("snapshot_info_t",2,"type")="uint64_t"
	Set gtmtypfldindx("snapshot_info_t","snapshot_tn")=2
	Set gtmtypes("snapshot_info_t",3,"name")="snapshot_info_t.db_blk_size"
	Set gtmtypes("snapshot_info_t",3,"off")=16
	Set gtmtypes("snapshot_info_t",3,"len")=4
	Set gtmtypes("snapshot_info_t",3,"type")="unsigned-int"
	Set gtmtypfldindx("snapshot_info_t","db_blk_size")=3
	Set gtmtypes("snapshot_info_t",4,"name")="snapshot_info_t.free_blks"
	Set gtmtypes("snapshot_info_t",4,"off")=20
	Set gtmtypes("snapshot_info_t",4,"len")=4
	Set gtmtypes("snapshot_info_t",4,"type")="unsigned-int"
	Set gtmtypfldindx("snapshot_info_t","free_blks")=4
	Set gtmtypes("snapshot_info_t",5,"name")="snapshot_info_t.total_blks"
	Set gtmtypes("snapshot_info_t",5,"off")=24
	Set gtmtypes("snapshot_info_t",5,"len")=4
	Set gtmtypes("snapshot_info_t",5,"type")="unsigned-int"
	Set gtmtypfldindx("snapshot_info_t","total_blks")=5
	Set gtmtypes("snapshot_info_t",6,"name")="snapshot_info_t.shadow_file"
	Set gtmtypes("snapshot_info_t",6,"off")=28
	Set gtmtypes("snapshot_info_t",6,"len")=4097
	Set gtmtypes("snapshot_info_t",6,"type")="char"
	Set gtmtypfldindx("snapshot_info_t","shadow_file")=6
	Set gtmtypes("snapshot_info_t",7,"name")="snapshot_info_t.shadow_vbn"
	Set gtmtypes("snapshot_info_t",7,"off")=4128
	Set gtmtypes("snapshot_info_t",7,"len")=4
	Set gtmtypes("snapshot_info_t",7,"type")="int"
	Set gtmtypfldindx("snapshot_info_t","shadow_vbn")=7
	Set gtmtypes("snapshot_info_t",8,"name")="snapshot_info_t.ss_shmid"
	Set gtmtypes("snapshot_info_t",8,"off")=4136
	Set gtmtypes("snapshot_info_t",8,"len")=8
	Set gtmtypes("snapshot_info_t",8,"type")="long"
	Set gtmtypfldindx("snapshot_info_t","ss_shmid")=8
	Set gtmtypes("snapshot_info_t",9,"name")="snapshot_info_t.ss_shmsize"
	Set gtmtypes("snapshot_info_t",9,"off")=4144
	Set gtmtypes("snapshot_info_t",9,"len")=4
	Set gtmtypes("snapshot_info_t",9,"type")="int"
	Set gtmtypfldindx("snapshot_info_t","ss_shmsize")=9
	;
	Set gtmtypes("socket_address")="struct"
	Set gtmtypes("socket_address",0)=6
	Set gtmtypes("socket_address","len")=80
	Set gtmtypes("socket_address",1,"name")="socket_address.sa"
	Set gtmtypes("socket_address",1,"off")=0
	Set gtmtypes("socket_address",1,"len")=8
	Set gtmtypes("socket_address",1,"type")="addr"
	Set gtmtypfldindx("socket_address","sa")=1
	Set gtmtypes("socket_address",2,"name")="socket_address.ai"
	Set gtmtypes("socket_address",2,"off")=8
	Set gtmtypes("socket_address",2,"len")=48
	Set gtmtypes("socket_address",2,"type")="struct"
	Set gtmtypfldindx("socket_address","ai")=2
	Set gtmtypes("socket_address",3,"name")="socket_address.ai_head"
	Set gtmtypes("socket_address",3,"off")=56
	Set gtmtypes("socket_address",3,"len")=8
	Set gtmtypes("socket_address",3,"type")="addr"
	Set gtmtypfldindx("socket_address","ai_head")=3
	Set gtmtypes("socket_address",4,"name")="socket_address.port"
	Set gtmtypes("socket_address",4,"off")=64
	Set gtmtypes("socket_address",4,"len")=2
	Set gtmtypes("socket_address",4,"type")="unsigned-short"
	Set gtmtypfldindx("socket_address","port")=4
	Set gtmtypes("socket_address",5,"name")="socket_address.process"
	Set gtmtypes("socket_address",5,"off")=68
	Set gtmtypes("socket_address",5,"len")=4
	Set gtmtypes("socket_address",5,"type")="pid_t"
	Set gtmtypfldindx("socket_address","process")=5
	Set gtmtypes("socket_address",6,"name")="socket_address.saddr_ip"
	Set gtmtypes("socket_address",6,"off")=72
	Set gtmtypes("socket_address",6,"len")=8
	Set gtmtypes("socket_address",6,"type")="addr"
	Set gtmtypfldindx("socket_address","saddr_ip")=6
	;
	Set gtmtypes("socket_interrupt")="struct"
	Set gtmtypes("socket_interrupt",0)=10
	Set gtmtypes("socket_interrupt","len")=48
	Set gtmtypes("socket_interrupt",1,"name")="socket_interrupt.end_time"
	Set gtmtypes("socket_interrupt",1,"off")=0
	Set gtmtypes("socket_interrupt",1,"len")=16
	Set gtmtypes("socket_interrupt",1,"type")="ABS_TIME"
	Set gtmtypfldindx("socket_interrupt","end_time")=1
	Set gtmtypes("socket_interrupt",2,"name")="socket_interrupt.end_time.at_sec"
	Set gtmtypes("socket_interrupt",2,"off")=0
	Set gtmtypes("socket_interrupt",2,"len")=8
	Set gtmtypes("socket_interrupt",2,"type")="long"
	Set gtmtypfldindx("socket_interrupt","end_time.at_sec")=2
	Set gtmtypes("socket_interrupt",3,"name")="socket_interrupt.end_time.at_usec"
	Set gtmtypes("socket_interrupt",3,"off")=8
	Set gtmtypes("socket_interrupt",3,"len")=8
	Set gtmtypes("socket_interrupt",3,"type")="long"
	Set gtmtypfldindx("socket_interrupt","end_time.at_usec")=3
	Set gtmtypes("socket_interrupt",4,"name")="socket_interrupt.who_saved"
	Set gtmtypes("socket_interrupt",4,"off")=16
	Set gtmtypes("socket_interrupt",4,"len")=4
	Set gtmtypes("socket_interrupt",4,"type")="int"
	Set gtmtypfldindx("socket_interrupt","who_saved")=4
	Set gtmtypes("socket_interrupt",5,"name")="socket_interrupt.max_bufflen"
	Set gtmtypes("socket_interrupt",5,"off")=20
	Set gtmtypes("socket_interrupt",5,"len")=4
	Set gtmtypes("socket_interrupt",5,"type")="int"
	Set gtmtypfldindx("socket_interrupt","max_bufflen")=5
	Set gtmtypes("socket_interrupt",6,"name")="socket_interrupt.bytes_read"
	Set gtmtypes("socket_interrupt",6,"off")=24
	Set gtmtypes("socket_interrupt",6,"len")=4
	Set gtmtypes("socket_interrupt",6,"type")="int"
	Set gtmtypfldindx("socket_interrupt","bytes_read")=6
	Set gtmtypes("socket_interrupt",7,"name")="socket_interrupt.chars_read"
	Set gtmtypes("socket_interrupt",7,"off")=28
	Set gtmtypes("socket_interrupt",7,"len")=4
	Set gtmtypes("socket_interrupt",7,"type")="int"
	Set gtmtypfldindx("socket_interrupt","chars_read")=7
	Set gtmtypes("socket_interrupt",8,"name")="socket_interrupt.end_time_valid"
	Set gtmtypes("socket_interrupt",8,"off")=32
	Set gtmtypes("socket_interrupt",8,"len")=4
	Set gtmtypes("socket_interrupt",8,"type")="boolean_t"
	Set gtmtypfldindx("socket_interrupt","end_time_valid")=8
	Set gtmtypes("socket_interrupt",9,"name")="socket_interrupt.ibfsize_specified"
	Set gtmtypes("socket_interrupt",9,"off")=36
	Set gtmtypes("socket_interrupt",9,"len")=4
	Set gtmtypes("socket_interrupt",9,"type")="boolean_t"
	Set gtmtypfldindx("socket_interrupt","ibfsize_specified")=9
	Set gtmtypes("socket_interrupt",10,"name")="socket_interrupt.newdsocket"
	Set gtmtypes("socket_interrupt",10,"off")=40
	Set gtmtypes("socket_interrupt",10,"len")=8
	Set gtmtypes("socket_interrupt",10,"type")="addr"
	Set gtmtypfldindx("socket_interrupt","newdsocket")=10
	;
	Set gtmtypes("socket_struct")="struct"
	Set gtmtypes("socket_struct",0)=84
	Set gtmtypes("socket_struct","len")=2760
	Set gtmtypes("socket_struct",1,"name")="socket_struct.sd"
	Set gtmtypes("socket_struct",1,"off")=0
	Set gtmtypes("socket_struct",1,"len")=4
	Set gtmtypes("socket_struct",1,"type")="int"
	Set gtmtypfldindx("socket_struct","sd")=1
	Set gtmtypes("socket_struct",2,"name")="socket_struct.temp_sd"
	Set gtmtypes("socket_struct",2,"off")=4
	Set gtmtypes("socket_struct",2,"len")=4
	Set gtmtypes("socket_struct",2,"type")="int"
	Set gtmtypfldindx("socket_struct","temp_sd")=2
	Set gtmtypes("socket_struct",3,"name")="socket_struct.dev"
	Set gtmtypes("socket_struct",3,"off")=8
	Set gtmtypes("socket_struct",3,"len")=8
	Set gtmtypes("socket_struct",3,"type")="addr"
	Set gtmtypfldindx("socket_struct","dev")=3
	Set gtmtypes("socket_struct",4,"name")="socket_struct.passive"
	Set gtmtypes("socket_struct",4,"off")=16
	Set gtmtypes("socket_struct",4,"len")=4
	Set gtmtypes("socket_struct",4,"type")="boolean_t"
	Set gtmtypfldindx("socket_struct","passive")=4
	Set gtmtypes("socket_struct",5,"name")="socket_struct.ioerror"
	Set gtmtypes("socket_struct",5,"off")=20
	Set gtmtypes("socket_struct",5,"len")=4
	Set gtmtypes("socket_struct",5,"type")="boolean_t"
	Set gtmtypfldindx("socket_struct","ioerror")=5
	Set gtmtypes("socket_struct",6,"name")="socket_struct.urgent"
	Set gtmtypes("socket_struct",6,"off")=24
	Set gtmtypes("socket_struct",6,"len")=4
	Set gtmtypes("socket_struct",6,"type")="boolean_t"
	Set gtmtypfldindx("socket_struct","urgent")=6
	Set gtmtypes("socket_struct",7,"name")="socket_struct.delim0containsLF"
	Set gtmtypes("socket_struct",7,"off")=28
	Set gtmtypes("socket_struct",7,"len")=4
	Set gtmtypes("socket_struct",7,"type")="boolean_t"
	Set gtmtypfldindx("socket_struct","delim0containsLF")=7
	Set gtmtypes("socket_struct",8,"name")="socket_struct.state"
	Set gtmtypes("socket_struct",8,"off")=32
	Set gtmtypes("socket_struct",8,"len")=4
	Set gtmtypes("socket_struct",8,"type")="int"
	Set gtmtypfldindx("socket_struct","state")=8
	Set gtmtypes("socket_struct",9,"name")="socket_struct.protocol"
	Set gtmtypes("socket_struct",9,"off")=36
	Set gtmtypes("socket_struct",9,"len")=4
	Set gtmtypes("socket_struct",9,"type")="int"
	Set gtmtypfldindx("socket_struct","protocol")=9
	Set gtmtypes("socket_struct",10,"name")="socket_struct.local"
	Set gtmtypes("socket_struct",10,"off")=40
	Set gtmtypes("socket_struct",10,"len")=80
	Set gtmtypes("socket_struct",10,"type")="socket_address"
	Set gtmtypfldindx("socket_struct","local")=10
	Set gtmtypes("socket_struct",11,"name")="socket_struct.local.sa"
	Set gtmtypes("socket_struct",11,"off")=40
	Set gtmtypes("socket_struct",11,"len")=8
	Set gtmtypes("socket_struct",11,"type")="addr"
	Set gtmtypfldindx("socket_struct","local.sa")=11
	Set gtmtypes("socket_struct",12,"name")="socket_struct.local.ai"
	Set gtmtypes("socket_struct",12,"off")=48
	Set gtmtypes("socket_struct",12,"len")=48
	Set gtmtypes("socket_struct",12,"type")="struct"
	Set gtmtypfldindx("socket_struct","local.ai")=12
	Set gtmtypes("socket_struct",13,"name")="socket_struct.local.ai_head"
	Set gtmtypes("socket_struct",13,"off")=96
	Set gtmtypes("socket_struct",13,"len")=8
	Set gtmtypes("socket_struct",13,"type")="addr"
	Set gtmtypfldindx("socket_struct","local.ai_head")=13
	Set gtmtypes("socket_struct",14,"name")="socket_struct.local.port"
	Set gtmtypes("socket_struct",14,"off")=104
	Set gtmtypes("socket_struct",14,"len")=2
	Set gtmtypes("socket_struct",14,"type")="unsigned-short"
	Set gtmtypfldindx("socket_struct","local.port")=14
	Set gtmtypes("socket_struct",15,"name")="socket_struct.local.process"
	Set gtmtypes("socket_struct",15,"off")=108
	Set gtmtypes("socket_struct",15,"len")=4
	Set gtmtypes("socket_struct",15,"type")="pid_t"
	Set gtmtypfldindx("socket_struct","local.process")=15
	Set gtmtypes("socket_struct",16,"name")="socket_struct.local.saddr_ip"
	Set gtmtypes("socket_struct",16,"off")=112
	Set gtmtypes("socket_struct",16,"len")=8
	Set gtmtypes("socket_struct",16,"type")="addr"
	Set gtmtypfldindx("socket_struct","local.saddr_ip")=16
	Set gtmtypes("socket_struct",17,"name")="socket_struct.remote"
	Set gtmtypes("socket_struct",17,"off")=120
	Set gtmtypes("socket_struct",17,"len")=80
	Set gtmtypes("socket_struct",17,"type")="socket_address"
	Set gtmtypfldindx("socket_struct","remote")=17
	Set gtmtypes("socket_struct",18,"name")="socket_struct.remote.sa"
	Set gtmtypes("socket_struct",18,"off")=120
	Set gtmtypes("socket_struct",18,"len")=8
	Set gtmtypes("socket_struct",18,"type")="addr"
	Set gtmtypfldindx("socket_struct","remote.sa")=18
	Set gtmtypes("socket_struct",19,"name")="socket_struct.remote.ai"
	Set gtmtypes("socket_struct",19,"off")=128
	Set gtmtypes("socket_struct",19,"len")=48
	Set gtmtypes("socket_struct",19,"type")="struct"
	Set gtmtypfldindx("socket_struct","remote.ai")=19
	Set gtmtypes("socket_struct",20,"name")="socket_struct.remote.ai_head"
	Set gtmtypes("socket_struct",20,"off")=176
	Set gtmtypes("socket_struct",20,"len")=8
	Set gtmtypes("socket_struct",20,"type")="addr"
	Set gtmtypfldindx("socket_struct","remote.ai_head")=20
	Set gtmtypes("socket_struct",21,"name")="socket_struct.remote.port"
	Set gtmtypes("socket_struct",21,"off")=184
	Set gtmtypes("socket_struct",21,"len")=2
	Set gtmtypes("socket_struct",21,"type")="unsigned-short"
	Set gtmtypfldindx("socket_struct","remote.port")=21
	Set gtmtypes("socket_struct",22,"name")="socket_struct.remote.process"
	Set gtmtypes("socket_struct",22,"off")=188
	Set gtmtypes("socket_struct",22,"len")=4
	Set gtmtypes("socket_struct",22,"type")="pid_t"
	Set gtmtypfldindx("socket_struct","remote.process")=22
	Set gtmtypes("socket_struct",23,"name")="socket_struct.remote.saddr_ip"
	Set gtmtypes("socket_struct",23,"off")=192
	Set gtmtypes("socket_struct",23,"len")=8
	Set gtmtypes("socket_struct",23,"type")="addr"
	Set gtmtypfldindx("socket_struct","remote.saddr_ip")=23
	Set gtmtypes("socket_struct",24,"name")="socket_struct.lastop"
	Set gtmtypes("socket_struct",24,"off")=200
	Set gtmtypes("socket_struct",24,"len")=4
	Set gtmtypes("socket_struct",24,"type")="unsigned-int"
	Set gtmtypfldindx("socket_struct","lastop")=24
	Set gtmtypes("socket_struct",25,"name")="socket_struct.moreread_timeout"
	Set gtmtypes("socket_struct",25,"off")=204
	Set gtmtypes("socket_struct",25,"len")=4
	Set gtmtypes("socket_struct",25,"type")="unsigned-int"
	Set gtmtypfldindx("socket_struct","moreread_timeout")=25
	Set gtmtypes("socket_struct",26,"name")="socket_struct.handle"
	Set gtmtypes("socket_struct",26,"off")=208
	Set gtmtypes("socket_struct",26,"len")=256
	Set gtmtypes("socket_struct",26,"type")="char"
	Set gtmtypfldindx("socket_struct","handle")=26
	Set gtmtypes("socket_struct",27,"name")="socket_struct.handle_len"
	Set gtmtypes("socket_struct",27,"off")=464
	Set gtmtypes("socket_struct",27,"len")=4
	Set gtmtypes("socket_struct",27,"type")="int"
	Set gtmtypfldindx("socket_struct","handle_len")=27
	Set gtmtypes("socket_struct",28,"name")="socket_struct.bufsiz"
	Set gtmtypes("socket_struct",28,"off")=468
	Set gtmtypes("socket_struct",28,"len")=4
	Set gtmtypes("socket_struct",28,"type")="int"
	Set gtmtypfldindx("socket_struct","bufsiz")=28
	Set gtmtypes("socket_struct",29,"name")="socket_struct.n_delimiter"
	Set gtmtypes("socket_struct",29,"off")=472
	Set gtmtypes("socket_struct",29,"len")=4
	Set gtmtypes("socket_struct",29,"type")="int"
	Set gtmtypfldindx("socket_struct","n_delimiter")=29
	Set gtmtypes("socket_struct",30,"name")="socket_struct.last_recv_errno"
	Set gtmtypes("socket_struct",30,"off")=476
	Set gtmtypes("socket_struct",30,"len")=4
	Set gtmtypes("socket_struct",30,"type")="int"
	Set gtmtypfldindx("socket_struct","last_recv_errno")=30
	Set gtmtypes("socket_struct",31,"name")="socket_struct.delimiter"
	Set gtmtypes("socket_struct",31,"off")=480
	Set gtmtypes("socket_struct",31,"len")=1024
	Set gtmtypes("socket_struct",31,"type")="mstr"
	Set gtmtypfldindx("socket_struct","delimiter")=31
	Set gtmtypes("socket_struct",31,"dim")=64
	Set gtmtypes("socket_struct",32,"name")="socket_struct.delimiter[0].char_len"
	Set gtmtypes("socket_struct",32,"off")=480
	Set gtmtypes("socket_struct",32,"len")=4
	Set gtmtypes("socket_struct",32,"type")="unsigned-int"
	Set gtmtypfldindx("socket_struct","delimiter[0].char_len")=32
	Set gtmtypes("socket_struct",33,"name")="socket_struct.delimiter[0].len"
	Set gtmtypes("socket_struct",33,"off")=484
	Set gtmtypes("socket_struct",33,"len")=4
	Set gtmtypes("socket_struct",33,"type")="int"
	Set gtmtypfldindx("socket_struct","delimiter[0].len")=33
	Set gtmtypes("socket_struct",34,"name")="socket_struct.delimiter[0].addr"
	Set gtmtypes("socket_struct",34,"off")=488
	Set gtmtypes("socket_struct",34,"len")=8
	Set gtmtypes("socket_struct",34,"type")="addr"
	Set gtmtypfldindx("socket_struct","delimiter[0].addr")=34
	Set gtmtypes("socket_struct",35,"name")="socket_struct.idelimiter"
	Set gtmtypes("socket_struct",35,"off")=1504
	Set gtmtypes("socket_struct",35,"len")=1024
	Set gtmtypes("socket_struct",35,"type")="mstr"
	Set gtmtypfldindx("socket_struct","idelimiter")=35
	Set gtmtypes("socket_struct",35,"dim")=64
	Set gtmtypes("socket_struct",36,"name")="socket_struct.idelimiter[0].char_len"
	Set gtmtypes("socket_struct",36,"off")=1504
	Set gtmtypes("socket_struct",36,"len")=4
	Set gtmtypes("socket_struct",36,"type")="unsigned-int"
	Set gtmtypfldindx("socket_struct","idelimiter[0].char_len")=36
	Set gtmtypes("socket_struct",37,"name")="socket_struct.idelimiter[0].len"
	Set gtmtypes("socket_struct",37,"off")=1508
	Set gtmtypes("socket_struct",37,"len")=4
	Set gtmtypes("socket_struct",37,"type")="int"
	Set gtmtypfldindx("socket_struct","idelimiter[0].len")=37
	Set gtmtypes("socket_struct",38,"name")="socket_struct.idelimiter[0].addr"
	Set gtmtypes("socket_struct",38,"off")=1512
	Set gtmtypes("socket_struct",38,"len")=8
	Set gtmtypes("socket_struct",38,"type")="addr"
	Set gtmtypfldindx("socket_struct","idelimiter[0].addr")=38
	Set gtmtypes("socket_struct",39,"name")="socket_struct.odelimiter0"
	Set gtmtypes("socket_struct",39,"off")=2528
	Set gtmtypes("socket_struct",39,"len")=16
	Set gtmtypes("socket_struct",39,"type")="mstr"
	Set gtmtypfldindx("socket_struct","odelimiter0")=39
	Set gtmtypes("socket_struct",40,"name")="socket_struct.odelimiter0.char_len"
	Set gtmtypes("socket_struct",40,"off")=2528
	Set gtmtypes("socket_struct",40,"len")=4
	Set gtmtypes("socket_struct",40,"type")="unsigned-int"
	Set gtmtypfldindx("socket_struct","odelimiter0.char_len")=40
	Set gtmtypes("socket_struct",41,"name")="socket_struct.odelimiter0.len"
	Set gtmtypes("socket_struct",41,"off")=2532
	Set gtmtypes("socket_struct",41,"len")=4
	Set gtmtypes("socket_struct",41,"type")="int"
	Set gtmtypfldindx("socket_struct","odelimiter0.len")=41
	Set gtmtypes("socket_struct",42,"name")="socket_struct.odelimiter0.addr"
	Set gtmtypes("socket_struct",42,"off")=2536
	Set gtmtypes("socket_struct",42,"len")=8
	Set gtmtypes("socket_struct",42,"type")="addr"
	Set gtmtypfldindx("socket_struct","odelimiter0.addr")=42
	Set gtmtypes("socket_struct",43,"name")="socket_struct.buffer_size"
	Set gtmtypes("socket_struct",43,"off")=2544
	Set gtmtypes("socket_struct",43,"len")=8
	Set gtmtypes("socket_struct",43,"type")="size_t"
	Set gtmtypfldindx("socket_struct","buffer_size")=43
	Set gtmtypes("socket_struct",44,"name")="socket_struct.buffered_length"
	Set gtmtypes("socket_struct",44,"off")=2552
	Set gtmtypes("socket_struct",44,"len")=8
	Set gtmtypes("socket_struct",44,"type")="size_t"
	Set gtmtypfldindx("socket_struct","buffered_length")=44
	Set gtmtypes("socket_struct",45,"name")="socket_struct.buffered_offset"
	Set gtmtypes("socket_struct",45,"off")=2560
	Set gtmtypes("socket_struct",45,"len")=8
	Set gtmtypes("socket_struct",45,"type")="size_t"
	Set gtmtypfldindx("socket_struct","buffered_offset")=45
	Set gtmtypes("socket_struct",46,"name")="socket_struct.buffer"
	Set gtmtypes("socket_struct",46,"off")=2568
	Set gtmtypes("socket_struct",46,"len")=8
	Set gtmtypes("socket_struct",46,"type")="addr"
	Set gtmtypfldindx("socket_struct","buffer")=46
	Set gtmtypes("socket_struct",47,"name")="socket_struct.nodelay"
	Set gtmtypes("socket_struct",47,"off")=2576
	Set gtmtypes("socket_struct",47,"len")=4
	Set gtmtypes("socket_struct",47,"type")="boolean_t"
	Set gtmtypfldindx("socket_struct","nodelay")=47
	Set gtmtypes("socket_struct",48,"name")="socket_struct.first_read"
	Set gtmtypes("socket_struct",48,"off")=2580
	Set gtmtypes("socket_struct",48,"len")=4
	Set gtmtypes("socket_struct",48,"type")="boolean_t"
	Set gtmtypfldindx("socket_struct","first_read")=48
	Set gtmtypes("socket_struct",49,"name")="socket_struct.first_write"
	Set gtmtypes("socket_struct",49,"off")=2584
	Set gtmtypes("socket_struct",49,"len")=4
	Set gtmtypes("socket_struct",49,"type")="boolean_t"
	Set gtmtypfldindx("socket_struct","first_write")=49
	Set gtmtypes("socket_struct",50,"name")="socket_struct.def_moreread_timeout"
	Set gtmtypes("socket_struct",50,"off")=2588
	Set gtmtypes("socket_struct",50,"len")=4
	Set gtmtypes("socket_struct",50,"type")="boolean_t"
	Set gtmtypfldindx("socket_struct","def_moreread_timeout")=50
	Set gtmtypes("socket_struct",51,"name")="socket_struct.passtype"
	Set gtmtypes("socket_struct",51,"off")=2592
	Set gtmtypes("socket_struct",51,"len")=4
	Set gtmtypes("socket_struct",51,"type")="int"
	Set gtmtypfldindx("socket_struct","passtype")=51
	Set gtmtypes("socket_struct",52,"name")="socket_struct.filemode"
	Set gtmtypes("socket_struct",52,"off")=2596
	Set gtmtypes("socket_struct",52,"len")=4
	Set gtmtypes("socket_struct",52,"type")="uint"
	Set gtmtypfldindx("socket_struct","filemode")=52
	Set gtmtypes("socket_struct",53,"name")="socket_struct.filemode_mask"
	Set gtmtypes("socket_struct",53,"off")=2600
	Set gtmtypes("socket_struct",53,"len")=4
	Set gtmtypes("socket_struct",53,"type")="uint"
	Set gtmtypfldindx("socket_struct","filemode_mask")=53
	Set gtmtypes("socket_struct",54,"name")="socket_struct.uic"
	Set gtmtypes("socket_struct",54,"off")=2604
	Set gtmtypes("socket_struct",54,"len")=8
	Set gtmtypes("socket_struct",54,"type")="uic_struct_int"
	Set gtmtypfldindx("socket_struct","uic")=54
	Set gtmtypes("socket_struct",55,"name")="socket_struct.uic.mem"
	Set gtmtypes("socket_struct",55,"off")=2604
	Set gtmtypes("socket_struct",55,"len")=4
	Set gtmtypes("socket_struct",55,"type")="uid_t"
	Set gtmtypfldindx("socket_struct","uic.mem")=55
	Set gtmtypes("socket_struct",56,"name")="socket_struct.uic.grp"
	Set gtmtypes("socket_struct",56,"off")=2608
	Set gtmtypes("socket_struct",56,"len")=4
	Set gtmtypes("socket_struct",56,"type")="gid_t"
	Set gtmtypfldindx("socket_struct","uic.grp")=56
	Set gtmtypes("socket_struct",57,"name")="socket_struct.zff"
	Set gtmtypes("socket_struct",57,"off")=2616
	Set gtmtypes("socket_struct",57,"len")=16
	Set gtmtypes("socket_struct",57,"type")="mstr"
	Set gtmtypfldindx("socket_struct","zff")=57
	Set gtmtypes("socket_struct",58,"name")="socket_struct.zff.char_len"
	Set gtmtypes("socket_struct",58,"off")=2616
	Set gtmtypes("socket_struct",58,"len")=4
	Set gtmtypes("socket_struct",58,"type")="unsigned-int"
	Set gtmtypfldindx("socket_struct","zff.char_len")=58
	Set gtmtypes("socket_struct",59,"name")="socket_struct.zff.len"
	Set gtmtypes("socket_struct",59,"off")=2620
	Set gtmtypes("socket_struct",59,"len")=4
	Set gtmtypes("socket_struct",59,"type")="int"
	Set gtmtypfldindx("socket_struct","zff.len")=59
	Set gtmtypes("socket_struct",60,"name")="socket_struct.zff.addr"
	Set gtmtypes("socket_struct",60,"off")=2624
	Set gtmtypes("socket_struct",60,"len")=8
	Set gtmtypes("socket_struct",60,"type")="addr"
	Set gtmtypfldindx("socket_struct","zff.addr")=60
	Set gtmtypes("socket_struct",61,"name")="socket_struct.ozff"
	Set gtmtypes("socket_struct",61,"off")=2632
	Set gtmtypes("socket_struct",61,"len")=16
	Set gtmtypes("socket_struct",61,"type")="mstr"
	Set gtmtypfldindx("socket_struct","ozff")=61
	Set gtmtypes("socket_struct",62,"name")="socket_struct.ozff.char_len"
	Set gtmtypes("socket_struct",62,"off")=2632
	Set gtmtypes("socket_struct",62,"len")=4
	Set gtmtypes("socket_struct",62,"type")="unsigned-int"
	Set gtmtypfldindx("socket_struct","ozff.char_len")=62
	Set gtmtypes("socket_struct",63,"name")="socket_struct.ozff.len"
	Set gtmtypes("socket_struct",63,"off")=2636
	Set gtmtypes("socket_struct",63,"len")=4
	Set gtmtypes("socket_struct",63,"type")="int"
	Set gtmtypfldindx("socket_struct","ozff.len")=63
	Set gtmtypes("socket_struct",64,"name")="socket_struct.ozff.addr"
	Set gtmtypes("socket_struct",64,"off")=2640
	Set gtmtypes("socket_struct",64,"len")=8
	Set gtmtypes("socket_struct",64,"type")="addr"
	Set gtmtypfldindx("socket_struct","ozff.addr")=64
	Set gtmtypes("socket_struct",65,"name")="socket_struct.lastaction"
	Set gtmtypes("socket_struct",65,"off")=2648
	Set gtmtypes("socket_struct",65,"len")=4
	Set gtmtypes("socket_struct",65,"type")="unsigned-int"
	Set gtmtypfldindx("socket_struct","lastaction")=65
	Set gtmtypes("socket_struct",66,"name")="socket_struct.readycycle"
	Set gtmtypes("socket_struct",66,"off")=2652
	Set gtmtypes("socket_struct",66,"len")=4
	Set gtmtypes("socket_struct",66,"type")="unsigned-int"
	Set gtmtypfldindx("socket_struct","readycycle")=66
	Set gtmtypes("socket_struct",67,"name")="socket_struct.pendingevent"
	Set gtmtypes("socket_struct",67,"off")=2656
	Set gtmtypes("socket_struct",67,"len")=4
	Set gtmtypes("socket_struct",67,"type")="boolean_t"
	Set gtmtypfldindx("socket_struct","pendingevent")=67
	Set gtmtypes("socket_struct",68,"name")="socket_struct.howcreated"
	Set gtmtypes("socket_struct",68,"off")=2660
	Set gtmtypes("socket_struct",68,"len")=4
	Set gtmtypes("socket_struct",68,"type")="int"
	Set gtmtypfldindx("socket_struct","howcreated")=68
	Set gtmtypes("socket_struct",69,"name")="socket_struct.parenthandle"
	Set gtmtypes("socket_struct",69,"off")=2664
	Set gtmtypes("socket_struct",69,"len")=8
	Set gtmtypes("socket_struct",69,"type")="addr"
	Set gtmtypfldindx("socket_struct","parenthandle")=69
	Set gtmtypes("socket_struct",70,"name")="socket_struct.obuffer_size"
	Set gtmtypes("socket_struct",70,"off")=2672
	Set gtmtypes("socket_struct",70,"len")=8
	Set gtmtypes("socket_struct",70,"type")="size_t"
	Set gtmtypfldindx("socket_struct","obuffer_size")=70
	Set gtmtypes("socket_struct",71,"name")="socket_struct.obuffer_length"
	Set gtmtypes("socket_struct",71,"off")=2680
	Set gtmtypes("socket_struct",71,"len")=8
	Set gtmtypes("socket_struct",71,"type")="size_t"
	Set gtmtypfldindx("socket_struct","obuffer_length")=71
	Set gtmtypes("socket_struct",72,"name")="socket_struct.obuffer_offset"
	Set gtmtypes("socket_struct",72,"off")=2688
	Set gtmtypes("socket_struct",72,"len")=8
	Set gtmtypes("socket_struct",72,"type")="size_t"
	Set gtmtypfldindx("socket_struct","obuffer_offset")=72
	Set gtmtypes("socket_struct",73,"name")="socket_struct.obuffer_timer_set"
	Set gtmtypes("socket_struct",73,"off")=2696
	Set gtmtypes("socket_struct",73,"len")=4
	Set gtmtypes("socket_struct",73,"type")="boolean_t"
	Set gtmtypfldindx("socket_struct","obuffer_timer_set")=73
	Set gtmtypes("socket_struct",74,"name")="socket_struct.obuffer_output_active"
	Set gtmtypes("socket_struct",74,"off")=2700
	Set gtmtypes("socket_struct",74,"len")=4
	Set gtmtypes("socket_struct",74,"type")="boolean_t"
	Set gtmtypfldindx("socket_struct","obuffer_output_active")=74
	Set gtmtypes("socket_struct",75,"name")="socket_struct.obuffer_flush_time"
	Set gtmtypes("socket_struct",75,"off")=2704
	Set gtmtypes("socket_struct",75,"len")=4
	Set gtmtypes("socket_struct",75,"type")="int"
	Set gtmtypfldindx("socket_struct","obuffer_flush_time")=75
	Set gtmtypes("socket_struct",76,"name")="socket_struct.obuffer_wait_time"
	Set gtmtypes("socket_struct",76,"off")=2708
	Set gtmtypes("socket_struct",76,"len")=4
	Set gtmtypes("socket_struct",76,"type")="int"
	Set gtmtypfldindx("socket_struct","obuffer_wait_time")=76
	Set gtmtypes("socket_struct",77,"name")="socket_struct.obuffer_errno"
	Set gtmtypes("socket_struct",77,"off")=2712
	Set gtmtypes("socket_struct",77,"len")=4
	Set gtmtypes("socket_struct",77,"type")="int"
	Set gtmtypfldindx("socket_struct","obuffer_errno")=77
	Set gtmtypes("socket_struct",78,"name")="socket_struct.obuffer"
	Set gtmtypes("socket_struct",78,"off")=2720
	Set gtmtypes("socket_struct",78,"len")=8
	Set gtmtypes("socket_struct",78,"type")="addr"
	Set gtmtypfldindx("socket_struct","obuffer")=78
	Set gtmtypes("socket_struct",79,"name")="socket_struct.nonblocking"
	Set gtmtypes("socket_struct",79,"off")=2728
	Set gtmtypes("socket_struct",79,"len")=4
	Set gtmtypes("socket_struct",79,"type")="boolean_t"
	Set gtmtypfldindx("socket_struct","nonblocking")=79
	Set gtmtypes("socket_struct",80,"name")="socket_struct.tlsenabled"
	Set gtmtypes("socket_struct",80,"off")=2732
	Set gtmtypes("socket_struct",80,"len")=4
	Set gtmtypes("socket_struct",80,"type")="boolean_t"
	Set gtmtypfldindx("socket_struct","tlsenabled")=80
	Set gtmtypes("socket_struct",81,"name")="socket_struct.tlssocket"
	Set gtmtypes("socket_struct",81,"off")=2736
	Set gtmtypes("socket_struct",81,"len")=8
	Set gtmtypes("socket_struct",81,"type")="addr"
	Set gtmtypfldindx("socket_struct","tlssocket")=81
	Set gtmtypes("socket_struct",82,"name")="socket_struct.tlsreadblocked"
	Set gtmtypes("socket_struct",82,"off")=2744
	Set gtmtypes("socket_struct",82,"len")=4
	Set gtmtypes("socket_struct",82,"type")="boolean_t"
	Set gtmtypfldindx("socket_struct","tlsreadblocked")=82
	Set gtmtypes("socket_struct",83,"name")="socket_struct.tlswriteblocked"
	Set gtmtypes("socket_struct",83,"off")=2748
	Set gtmtypes("socket_struct",83,"len")=4
	Set gtmtypes("socket_struct",83,"type")="boolean_t"
	Set gtmtypfldindx("socket_struct","tlswriteblocked")=83
	Set gtmtypes("socket_struct",84,"name")="socket_struct.tlspolldirection"
	Set gtmtypes("socket_struct",84,"off")=2752
	Set gtmtypes("socket_struct",84,"len")=2
	Set gtmtypes("socket_struct",84,"type")="short"
	Set gtmtypfldindx("socket_struct","tlspolldirection")=84
	;
	Set gtmtypes("source_address")="struct"
	Set gtmtypes("source_address",0)=2
	Set gtmtypes("source_address","len")=8
	Set gtmtypes("source_address",1,"name")="source_address.line"
	Set gtmtypes("source_address",1,"off")=0
	Set gtmtypes("source_address",1,"len")=4
	Set gtmtypes("source_address",1,"type")="unsigned-int"
	Set gtmtypfldindx("source_address","line")=1
	Set gtmtypes("source_address",2,"name")="source_address.column"
	Set gtmtypes("source_address",2,"off")=4
	Set gtmtypes("source_address",2,"len")=4
	Set gtmtypes("source_address",2,"type")="unsigned-int"
	Set gtmtypfldindx("source_address","column")=2
	;
	Set gtmtypes("span_node_integ")="struct"
	Set gtmtypes("span_node_integ",0)=11
	Set gtmtypes("span_node_integ","len")=1064
	Set gtmtypes("span_node_integ",1,"name")="span_node_integ.sn_type"
	Set gtmtypes("span_node_integ",1,"off")=0
	Set gtmtypes("span_node_integ",1,"len")=4
	Set gtmtypes("span_node_integ",1,"type")="unsigned-int"
	Set gtmtypfldindx("span_node_integ","sn_type")=1
	Set gtmtypes("span_node_integ",2,"name")="span_node_integ.span_prev_blk"
	Set gtmtypes("span_node_integ",2,"off")=4
	Set gtmtypes("span_node_integ",2,"len")=4
	Set gtmtypes("span_node_integ",2,"type")="unsigned-int"
	Set gtmtypfldindx("span_node_integ","span_prev_blk")=2
	Set gtmtypes("span_node_integ",3,"name")="span_node_integ.span_blk_cnt"
	Set gtmtypes("span_node_integ",3,"off")=8
	Set gtmtypes("span_node_integ",3,"len")=4
	Set gtmtypes("span_node_integ",3,"type")="unsigned-int"
	Set gtmtypfldindx("span_node_integ","span_blk_cnt")=3
	Set gtmtypes("span_node_integ",4,"name")="span_node_integ.span_tot_blks"
	Set gtmtypes("span_node_integ",4,"off")=12
	Set gtmtypes("span_node_integ",4,"len")=4
	Set gtmtypes("span_node_integ",4,"type")="unsigned-int"
	Set gtmtypfldindx("span_node_integ","span_tot_blks")=4
	Set gtmtypes("span_node_integ",5,"name")="span_node_integ.span_node_sz"
	Set gtmtypes("span_node_integ",5,"off")=16
	Set gtmtypes("span_node_integ",5,"len")=4
	Set gtmtypes("span_node_integ",5,"type")="unsigned-int"
	Set gtmtypfldindx("span_node_integ","span_node_sz")=5
	Set gtmtypes("span_node_integ",6,"name")="span_node_integ.span_frag_off"
	Set gtmtypes("span_node_integ",6,"off")=20
	Set gtmtypes("span_node_integ",6,"len")=4
	Set gtmtypes("span_node_integ",6,"type")="unsigned-int"
	Set gtmtypfldindx("span_node_integ","span_frag_off")=6
	Set gtmtypes("span_node_integ",7,"name")="span_node_integ.key_len"
	Set gtmtypes("span_node_integ",7,"off")=24
	Set gtmtypes("span_node_integ",7,"len")=4
	Set gtmtypes("span_node_integ",7,"type")="unsigned-int"
	Set gtmtypfldindx("span_node_integ","key_len")=7
	Set gtmtypes("span_node_integ",8,"name")="span_node_integ.val_len"
	Set gtmtypes("span_node_integ",8,"off")=28
	Set gtmtypes("span_node_integ",8,"len")=4
	Set gtmtypes("span_node_integ",8,"type")="unsigned-int"
	Set gtmtypfldindx("span_node_integ","val_len")=8
	Set gtmtypes("span_node_integ",9,"name")="span_node_integ.sn_cnt"
	Set gtmtypes("span_node_integ",9,"off")=32
	Set gtmtypes("span_node_integ",9,"len")=4
	Set gtmtypes("span_node_integ",9,"type")="unsigned-int"
	Set gtmtypfldindx("span_node_integ","sn_cnt")=9
	Set gtmtypes("span_node_integ",10,"name")="span_node_integ.sn_blk_cnt"
	Set gtmtypes("span_node_integ",10,"off")=36
	Set gtmtypes("span_node_integ",10,"len")=4
	Set gtmtypes("span_node_integ",10,"type")="unsigned-int"
	Set gtmtypfldindx("span_node_integ","sn_blk_cnt")=10
	Set gtmtypes("span_node_integ",11,"name")="span_node_integ.span_node_buf"
	Set gtmtypes("span_node_integ",11,"off")=40
	Set gtmtypes("span_node_integ",11,"len")=1023
	Set gtmtypes("span_node_integ",11,"type")="unsigned-char"
	Set gtmtypfldindx("span_node_integ","span_node_buf")=11
	;
	Set gtmtypes("span_parms")="struct"
	Set gtmtypes("span_parms",0)=10
	Set gtmtypes("span_parms","len")=64
	Set gtmtypes("span_parms",1,"name")="span_parms.span_status"
	Set gtmtypes("span_parms",1,"off")=0
	Set gtmtypes("span_parms",1,"len")=4
	Set gtmtypes("span_parms",1,"type")="boolean_t"
	Set gtmtypfldindx("span_parms","span_status")=1
	Set gtmtypes("span_parms",2,"name")="span_parms.enable_jnl_format"
	Set gtmtypes("span_parms",2,"off")=4
	Set gtmtypes("span_parms",2,"len")=4
	Set gtmtypes("span_parms",2,"type")="boolean_t"
	Set gtmtypfldindx("span_parms","enable_jnl_format")=2
	Set gtmtypes("span_parms",3,"name")="span_parms.enable_trigger_read_and_fire"
	Set gtmtypes("span_parms",3,"off")=8
	Set gtmtypes("span_parms",3,"len")=4
	Set gtmtypes("span_parms",3,"type")="boolean_t"
	Set gtmtypfldindx("span_parms","enable_trigger_read_and_fire")=3
	Set gtmtypes("span_parms",4,"name")="span_parms.ztval_gvcst_put_redo"
	Set gtmtypes("span_parms",4,"off")=12
	Set gtmtypes("span_parms",4,"len")=4
	Set gtmtypes("span_parms",4,"type")="boolean_t"
	Set gtmtypfldindx("span_parms","ztval_gvcst_put_redo")=4
	Set gtmtypes("span_parms",5,"name")="span_parms.val_forjnl"
	Set gtmtypes("span_parms",5,"off")=16
	Set gtmtypes("span_parms",5,"len")=8
	Set gtmtypes("span_parms",5,"type")="addr"
	Set gtmtypfldindx("span_parms","val_forjnl")=5
	Set gtmtypes("span_parms",6,"name")="span_parms.blk_reserved_bytes"
	Set gtmtypes("span_parms",6,"off")=24
	Set gtmtypes("span_parms",6,"len")=4
	Set gtmtypes("span_parms",6,"type")="int"
	Set gtmtypfldindx("span_parms","blk_reserved_bytes")=6
	Set gtmtypes("span_parms",7,"name")="span_parms.save_msp"
	Set gtmtypes("span_parms",7,"off")=32
	Set gtmtypes("span_parms",7,"len")=8
	Set gtmtypes("span_parms",7,"type")="addr"
	Set gtmtypfldindx("span_parms","save_msp")=7
	Set gtmtypes("span_parms",8,"name")="span_parms.save_mv_chain"
	Set gtmtypes("span_parms",8,"off")=40
	Set gtmtypes("span_parms",8,"len")=8
	Set gtmtypes("span_parms",8,"type")="addr"
	Set gtmtypfldindx("span_parms","save_mv_chain")=8
	Set gtmtypes("span_parms",9,"name")="span_parms.ztold_mval"
	Set gtmtypes("span_parms",9,"off")=48
	Set gtmtypes("span_parms",9,"len")=8
	Set gtmtypes("span_parms",9,"type")="addr"
	Set gtmtypfldindx("span_parms","ztold_mval")=9
	Set gtmtypes("span_parms",10,"name")="span_parms.ztval_mval"
	Set gtmtypes("span_parms",10,"off")=56
	Set gtmtypes("span_parms",10,"len")=8
	Set gtmtypes("span_parms",10,"type")="addr"
	Set gtmtypfldindx("span_parms","ztval_mval")=10
	;
	Set gtmtypes("span_subs")="struct"
	Set gtmtypes("span_subs",0)=3
	Set gtmtypes("span_subs","len")=3
	Set gtmtypes("span_subs",1,"name")="span_subs.b_ctrl"
	Set gtmtypes("span_subs",1,"off")=0
	Set gtmtypes("span_subs",1,"len")=1
	Set gtmtypes("span_subs",1,"type")="unsigned-char"
	Set gtmtypfldindx("span_subs","b_ctrl")=1
	Set gtmtypes("span_subs",2,"name")="span_subs.b_first"
	Set gtmtypes("span_subs",2,"off")=1
	Set gtmtypes("span_subs",2,"len")=1
	Set gtmtypes("span_subs",2,"type")="unsigned-char"
	Set gtmtypfldindx("span_subs","b_first")=2
	Set gtmtypes("span_subs",3,"name")="span_subs.b_sec"
	Set gtmtypes("span_subs",3,"off")=2
	Set gtmtypes("span_subs",3,"len")=1
	Set gtmtypes("span_subs",3,"type")="unsigned-char"
	Set gtmtypfldindx("span_subs","b_sec")=3
	;
	Set gtmtypes("spdesc")="struct"
	Set gtmtypes("spdesc",0)=9
	Set gtmtypes("spdesc","len")=56
	Set gtmtypes("spdesc",1,"name")="spdesc.base"
	Set gtmtypes("spdesc",1,"off")=0
	Set gtmtypes("spdesc",1,"len")=8
	Set gtmtypes("spdesc",1,"type")="addr"
	Set gtmtypfldindx("spdesc","base")=1
	Set gtmtypes("spdesc",2,"name")="spdesc.gtm_free"
	Set gtmtypes("spdesc",2,"off")=8
	Set gtmtypes("spdesc",2,"len")=8
	Set gtmtypes("spdesc",2,"type")="addr"
	Set gtmtypfldindx("spdesc","gtm_free")=2
	Set gtmtypes("spdesc",3,"name")="spdesc.top"
	Set gtmtypes("spdesc",3,"off")=16
	Set gtmtypes("spdesc",3,"len")=8
	Set gtmtypes("spdesc",3,"type")="addr"
	Set gtmtypfldindx("spdesc","top")=3
	Set gtmtypes("spdesc",4,"name")="spdesc.lasttop"
	Set gtmtypes("spdesc",4,"off")=24
	Set gtmtypes("spdesc",4,"len")=8
	Set gtmtypes("spdesc",4,"type")="addr"
	Set gtmtypfldindx("spdesc","lasttop")=4
	Set gtmtypes("spdesc",5,"name")="spdesc.invokestpgcollevel"
	Set gtmtypes("spdesc",5,"off")=32
	Set gtmtypes("spdesc",5,"len")=8
	Set gtmtypes("spdesc",5,"type")="addr"
	Set gtmtypfldindx("spdesc","invokestpgcollevel")=5
	Set gtmtypes("spdesc",6,"name")="spdesc.gcols"
	Set gtmtypes("spdesc",6,"off")=40
	Set gtmtypes("spdesc",6,"len")=4
	Set gtmtypes("spdesc",6,"type")="unsigned-int"
	Set gtmtypfldindx("spdesc","gcols")=6
	Set gtmtypes("spdesc",7,"name")="spdesc.strpllim"
	Set gtmtypes("spdesc",7,"off")=44
	Set gtmtypes("spdesc",7,"len")=4
	Set gtmtypes("spdesc",7,"type")="unsigned-int"
	Set gtmtypfldindx("spdesc","strpllim")=7
	Set gtmtypes("spdesc",8,"name")="spdesc.strpllimwarned"
	Set gtmtypes("spdesc",8,"off")=48
	Set gtmtypes("spdesc",8,"len")=4
	Set gtmtypes("spdesc",8,"type")="boolean_t"
	Set gtmtypfldindx("spdesc","strpllimwarned")=8
	Set gtmtypes("spdesc",9,"name")="spdesc.prvprt"
	Set gtmtypes("spdesc",9,"off")=52
	Set gtmtypes("spdesc",9,"len")=1
	Set gtmtypes("spdesc",9,"type")="unsigned-char"
	Set gtmtypfldindx("spdesc","prvprt")=9
	;
	Set gtmtypes("src_line_struct")="struct"
	Set gtmtypes("src_line_struct",0)=8
	Set gtmtypes("src_line_struct","len")=40
	Set gtmtypes("src_line_struct",1,"name")="src_line_struct.que"
	Set gtmtypes("src_line_struct",1,"off")=0
	Set gtmtypes("src_line_struct",1,"len")=16
	Set gtmtypes("src_line_struct",1,"type")="struct"
	Set gtmtypfldindx("src_line_struct","que")=1
	Set gtmtypes("src_line_struct",2,"name")="src_line_struct.que.fl"
	Set gtmtypes("src_line_struct",2,"off")=0
	Set gtmtypes("src_line_struct",2,"len")=8
	Set gtmtypes("src_line_struct",2,"type")="addr"
	Set gtmtypfldindx("src_line_struct","que.fl")=2
	Set gtmtypes("src_line_struct",3,"name")="src_line_struct.que.bl"
	Set gtmtypes("src_line_struct",3,"off")=8
	Set gtmtypes("src_line_struct",3,"len")=8
	Set gtmtypes("src_line_struct",3,"type")="addr"
	Set gtmtypfldindx("src_line_struct","que.bl")=3
	Set gtmtypes("src_line_struct",4,"name")="src_line_struct.str"
	Set gtmtypes("src_line_struct",4,"off")=16
	Set gtmtypes("src_line_struct",4,"len")=16
	Set gtmtypes("src_line_struct",4,"type")="mstr"
	Set gtmtypfldindx("src_line_struct","str")=4
	Set gtmtypes("src_line_struct",5,"name")="src_line_struct.str.char_len"
	Set gtmtypes("src_line_struct",5,"off")=16
	Set gtmtypes("src_line_struct",5,"len")=4
	Set gtmtypes("src_line_struct",5,"type")="unsigned-int"
	Set gtmtypfldindx("src_line_struct","str.char_len")=5
	Set gtmtypes("src_line_struct",6,"name")="src_line_struct.str.len"
	Set gtmtypes("src_line_struct",6,"off")=20
	Set gtmtypes("src_line_struct",6,"len")=4
	Set gtmtypes("src_line_struct",6,"type")="int"
	Set gtmtypfldindx("src_line_struct","str.len")=6
	Set gtmtypes("src_line_struct",7,"name")="src_line_struct.str.addr"
	Set gtmtypes("src_line_struct",7,"off")=24
	Set gtmtypes("src_line_struct",7,"len")=8
	Set gtmtypes("src_line_struct",7,"type")="addr"
	Set gtmtypfldindx("src_line_struct","str.addr")=7
	Set gtmtypes("src_line_struct",8,"name")="src_line_struct.line"
	Set gtmtypes("src_line_struct",8,"off")=32
	Set gtmtypes("src_line_struct",8,"len")=4
	Set gtmtypes("src_line_struct",8,"type")="int"
	Set gtmtypfldindx("src_line_struct","line")=8
	;
	Set gtmtypes("srch_blk_status")="struct"
	Set gtmtypes("srch_blk_status",0)=15
	Set gtmtypes("srch_blk_status","len")=72
	Set gtmtypes("srch_blk_status",1,"name")="srch_blk_status.cr"
	Set gtmtypes("srch_blk_status",1,"off")=0
	Set gtmtypes("srch_blk_status",1,"len")=8
	Set gtmtypes("srch_blk_status",1,"type")="addr"
	Set gtmtypfldindx("srch_blk_status","cr")=1
	Set gtmtypes("srch_blk_status",2,"name")="srch_blk_status.buffaddr"
	Set gtmtypes("srch_blk_status",2,"off")=8
	Set gtmtypes("srch_blk_status",2,"len")=8
	Set gtmtypes("srch_blk_status",2,"type")="addr"
	Set gtmtypfldindx("srch_blk_status","buffaddr")=2
	Set gtmtypes("srch_blk_status",3,"name")="srch_blk_status.blk_num"
	Set gtmtypes("srch_blk_status",3,"off")=16
	Set gtmtypes("srch_blk_status",3,"len")=4
	Set gtmtypes("srch_blk_status",3,"type")="int"
	Set gtmtypfldindx("srch_blk_status","blk_num")=3
	Set gtmtypes("srch_blk_status",4,"name")="srch_blk_status.tn"
	Set gtmtypes("srch_blk_status",4,"off")=24
	Set gtmtypes("srch_blk_status",4,"len")=8
	Set gtmtypes("srch_blk_status",4,"type")="uint64_t"
	Set gtmtypfldindx("srch_blk_status","tn")=4
	Set gtmtypes("srch_blk_status",5,"name")="srch_blk_status.prev_rec"
	Set gtmtypes("srch_blk_status",5,"off")=32
	Set gtmtypes("srch_blk_status",5,"len")=4
	Set gtmtypes("srch_blk_status",5,"type")="srch_rec_status"
	Set gtmtypfldindx("srch_blk_status","prev_rec")=5
	Set gtmtypes("srch_blk_status",6,"name")="srch_blk_status.prev_rec.offset"
	Set gtmtypes("srch_blk_status",6,"off")=32
	Set gtmtypes("srch_blk_status",6,"len")=2
	Set gtmtypes("srch_blk_status",6,"type")="unsigned-short"
	Set gtmtypfldindx("srch_blk_status","prev_rec.offset")=6
	Set gtmtypes("srch_blk_status",7,"name")="srch_blk_status.prev_rec.match"
	Set gtmtypes("srch_blk_status",7,"off")=34
	Set gtmtypes("srch_blk_status",7,"len")=2
	Set gtmtypes("srch_blk_status",7,"type")="unsigned-short"
	Set gtmtypfldindx("srch_blk_status","prev_rec.match")=7
	Set gtmtypes("srch_blk_status",8,"name")="srch_blk_status.curr_rec"
	Set gtmtypes("srch_blk_status",8,"off")=36
	Set gtmtypes("srch_blk_status",8,"len")=4
	Set gtmtypes("srch_blk_status",8,"type")="srch_rec_status"
	Set gtmtypfldindx("srch_blk_status","curr_rec")=8
	Set gtmtypes("srch_blk_status",9,"name")="srch_blk_status.curr_rec.offset"
	Set gtmtypes("srch_blk_status",9,"off")=36
	Set gtmtypes("srch_blk_status",9,"len")=2
	Set gtmtypes("srch_blk_status",9,"type")="unsigned-short"
	Set gtmtypfldindx("srch_blk_status","curr_rec.offset")=9
	Set gtmtypes("srch_blk_status",10,"name")="srch_blk_status.curr_rec.match"
	Set gtmtypes("srch_blk_status",10,"off")=38
	Set gtmtypes("srch_blk_status",10,"len")=2
	Set gtmtypes("srch_blk_status",10,"type")="unsigned-short"
	Set gtmtypfldindx("srch_blk_status","curr_rec.match")=10
	Set gtmtypes("srch_blk_status",11,"name")="srch_blk_status.cycle"
	Set gtmtypes("srch_blk_status",11,"off")=40
	Set gtmtypes("srch_blk_status",11,"len")=4
	Set gtmtypes("srch_blk_status",11,"type")="int"
	Set gtmtypfldindx("srch_blk_status","cycle")=11
	Set gtmtypes("srch_blk_status",12,"name")="srch_blk_status.level"
	Set gtmtypes("srch_blk_status",12,"off")=44
	Set gtmtypes("srch_blk_status",12,"len")=4
	Set gtmtypes("srch_blk_status",12,"type")="int"
	Set gtmtypfldindx("srch_blk_status","level")=12
	Set gtmtypes("srch_blk_status",13,"name")="srch_blk_status.cse"
	Set gtmtypes("srch_blk_status",13,"off")=48
	Set gtmtypes("srch_blk_status",13,"len")=8
	Set gtmtypes("srch_blk_status",13,"type")="addr"
	Set gtmtypfldindx("srch_blk_status","cse")=13
	Set gtmtypes("srch_blk_status",14,"name")="srch_blk_status.first_tp_srch_status"
	Set gtmtypes("srch_blk_status",14,"off")=56
	Set gtmtypes("srch_blk_status",14,"len")=8
	Set gtmtypes("srch_blk_status",14,"type")="addr"
	Set gtmtypfldindx("srch_blk_status","first_tp_srch_status")=14
	Set gtmtypes("srch_blk_status",15,"name")="srch_blk_status.blk_target"
	Set gtmtypes("srch_blk_status",15,"off")=64
	Set gtmtypes("srch_blk_status",15,"len")=8
	Set gtmtypes("srch_blk_status",15,"type")="addr"
	Set gtmtypfldindx("srch_blk_status","blk_target")=15
	;
	Set gtmtypes("srch_hist")="struct"
	Set gtmtypes("srch_hist",0)=18
	Set gtmtypes("srch_hist","len")=584
	Set gtmtypes("srch_hist",1,"name")="srch_hist.depth"
	Set gtmtypes("srch_hist",1,"off")=0
	Set gtmtypes("srch_hist",1,"len")=4
	Set gtmtypes("srch_hist",1,"type")="int"
	Set gtmtypfldindx("srch_hist","depth")=1
	Set gtmtypes("srch_hist",2,"name")="srch_hist.filler"
	Set gtmtypes("srch_hist",2,"off")=4
	Set gtmtypes("srch_hist",2,"len")=4
	Set gtmtypes("srch_hist",2,"type")="int"
	Set gtmtypfldindx("srch_hist","filler")=2
	Set gtmtypes("srch_hist",3,"name")="srch_hist.h"
	Set gtmtypes("srch_hist",3,"off")=8
	Set gtmtypes("srch_hist",3,"len")=576
	Set gtmtypes("srch_hist",3,"type")="srch_blk_status"
	Set gtmtypfldindx("srch_hist","h")=3
	Set gtmtypes("srch_hist",3,"dim")=8
	Set gtmtypes("srch_hist",4,"name")="srch_hist.h[0].cr"
	Set gtmtypes("srch_hist",4,"off")=8
	Set gtmtypes("srch_hist",4,"len")=8
	Set gtmtypes("srch_hist",4,"type")="addr"
	Set gtmtypfldindx("srch_hist","h[0].cr")=4
	Set gtmtypes("srch_hist",5,"name")="srch_hist.h[0].buffaddr"
	Set gtmtypes("srch_hist",5,"off")=16
	Set gtmtypes("srch_hist",5,"len")=8
	Set gtmtypes("srch_hist",5,"type")="addr"
	Set gtmtypfldindx("srch_hist","h[0].buffaddr")=5
	Set gtmtypes("srch_hist",6,"name")="srch_hist.h[0].blk_num"
	Set gtmtypes("srch_hist",6,"off")=24
	Set gtmtypes("srch_hist",6,"len")=4
	Set gtmtypes("srch_hist",6,"type")="int"
	Set gtmtypfldindx("srch_hist","h[0].blk_num")=6
	Set gtmtypes("srch_hist",7,"name")="srch_hist.h[0].tn"
	Set gtmtypes("srch_hist",7,"off")=32
	Set gtmtypes("srch_hist",7,"len")=8
	Set gtmtypes("srch_hist",7,"type")="uint64_t"
	Set gtmtypfldindx("srch_hist","h[0].tn")=7
	Set gtmtypes("srch_hist",8,"name")="srch_hist.h[0].prev_rec"
	Set gtmtypes("srch_hist",8,"off")=40
	Set gtmtypes("srch_hist",8,"len")=4
	Set gtmtypes("srch_hist",8,"type")="srch_rec_status"
	Set gtmtypfldindx("srch_hist","h[0].prev_rec")=8
	Set gtmtypes("srch_hist",9,"name")="srch_hist.h[0].prev_rec.offset"
	Set gtmtypes("srch_hist",9,"off")=40
	Set gtmtypes("srch_hist",9,"len")=2
	Set gtmtypes("srch_hist",9,"type")="unsigned-short"
	Set gtmtypfldindx("srch_hist","h[0].prev_rec.offset")=9
	Set gtmtypes("srch_hist",10,"name")="srch_hist.h[0].prev_rec.match"
	Set gtmtypes("srch_hist",10,"off")=42
	Set gtmtypes("srch_hist",10,"len")=2
	Set gtmtypes("srch_hist",10,"type")="unsigned-short"
	Set gtmtypfldindx("srch_hist","h[0].prev_rec.match")=10
	Set gtmtypes("srch_hist",11,"name")="srch_hist.h[0].curr_rec"
	Set gtmtypes("srch_hist",11,"off")=44
	Set gtmtypes("srch_hist",11,"len")=4
	Set gtmtypes("srch_hist",11,"type")="srch_rec_status"
	Set gtmtypfldindx("srch_hist","h[0].curr_rec")=11
	Set gtmtypes("srch_hist",12,"name")="srch_hist.h[0].curr_rec.offset"
	Set gtmtypes("srch_hist",12,"off")=44
	Set gtmtypes("srch_hist",12,"len")=2
	Set gtmtypes("srch_hist",12,"type")="unsigned-short"
	Set gtmtypfldindx("srch_hist","h[0].curr_rec.offset")=12
	Set gtmtypes("srch_hist",13,"name")="srch_hist.h[0].curr_rec.match"
	Set gtmtypes("srch_hist",13,"off")=46
	Set gtmtypes("srch_hist",13,"len")=2
	Set gtmtypes("srch_hist",13,"type")="unsigned-short"
	Set gtmtypfldindx("srch_hist","h[0].curr_rec.match")=13
	Set gtmtypes("srch_hist",14,"name")="srch_hist.h[0].cycle"
	Set gtmtypes("srch_hist",14,"off")=48
	Set gtmtypes("srch_hist",14,"len")=4
	Set gtmtypes("srch_hist",14,"type")="int"
	Set gtmtypfldindx("srch_hist","h[0].cycle")=14
	Set gtmtypes("srch_hist",15,"name")="srch_hist.h[0].level"
	Set gtmtypes("srch_hist",15,"off")=52
	Set gtmtypes("srch_hist",15,"len")=4
	Set gtmtypes("srch_hist",15,"type")="int"
	Set gtmtypfldindx("srch_hist","h[0].level")=15
	Set gtmtypes("srch_hist",16,"name")="srch_hist.h[0].cse"
	Set gtmtypes("srch_hist",16,"off")=56
	Set gtmtypes("srch_hist",16,"len")=8
	Set gtmtypes("srch_hist",16,"type")="addr"
	Set gtmtypfldindx("srch_hist","h[0].cse")=16
	Set gtmtypes("srch_hist",17,"name")="srch_hist.h[0].first_tp_srch_status"
	Set gtmtypes("srch_hist",17,"off")=64
	Set gtmtypes("srch_hist",17,"len")=8
	Set gtmtypes("srch_hist",17,"type")="addr"
	Set gtmtypfldindx("srch_hist","h[0].first_tp_srch_status")=17
	Set gtmtypes("srch_hist",18,"name")="srch_hist.h[0].blk_target"
	Set gtmtypes("srch_hist",18,"off")=72
	Set gtmtypes("srch_hist",18,"len")=8
	Set gtmtypes("srch_hist",18,"type")="addr"
	Set gtmtypfldindx("srch_hist","h[0].blk_target")=18
	;
	Set gtmtypes("srch_rec_status")="struct"
	Set gtmtypes("srch_rec_status",0)=2
	Set gtmtypes("srch_rec_status","len")=4
	Set gtmtypes("srch_rec_status",1,"name")="srch_rec_status.offset"
	Set gtmtypes("srch_rec_status",1,"off")=0
	Set gtmtypes("srch_rec_status",1,"len")=2
	Set gtmtypes("srch_rec_status",1,"type")="unsigned-short"
	Set gtmtypfldindx("srch_rec_status","offset")=1
	Set gtmtypes("srch_rec_status",2,"name")="srch_rec_status.match"
	Set gtmtypes("srch_rec_status",2,"off")=2
	Set gtmtypes("srch_rec_status",2,"len")=2
	Set gtmtypes("srch_rec_status",2,"type")="unsigned-short"
	Set gtmtypfldindx("srch_rec_status","match")=2
	;
	Set gtmtypes("st_timer_alloc")="struct"
	Set gtmtypes("st_timer_alloc",0)=2
	Set gtmtypes("st_timer_alloc","len")=16
	Set gtmtypes("st_timer_alloc",1,"name")="st_timer_alloc.addr"
	Set gtmtypes("st_timer_alloc",1,"off")=0
	Set gtmtypes("st_timer_alloc",1,"len")=8
	Set gtmtypes("st_timer_alloc",1,"type")="addr"
	Set gtmtypfldindx("st_timer_alloc","addr")=1
	Set gtmtypes("st_timer_alloc",2,"name")="st_timer_alloc.next"
	Set gtmtypes("st_timer_alloc",2,"off")=8
	Set gtmtypes("st_timer_alloc",2,"len")=8
	Set gtmtypes("st_timer_alloc",2,"type")="addr"
	Set gtmtypfldindx("st_timer_alloc","next")=2
	;
	Set gtmtypes("stack_frame")="struct"
	Set gtmtypes("stack_frame",0)=17
	Set gtmtypes("stack_frame","len")=104
	Set gtmtypes("stack_frame",1,"name")="stack_frame.rvector"
	Set gtmtypes("stack_frame",1,"off")=0
	Set gtmtypes("stack_frame",1,"len")=8
	Set gtmtypes("stack_frame",1,"type")="addr"
	Set gtmtypfldindx("stack_frame","rvector")=1
	Set gtmtypes("stack_frame",2,"name")="stack_frame.l_symtab"
	Set gtmtypes("stack_frame",2,"off")=8
	Set gtmtypes("stack_frame",2,"len")=8
	Set gtmtypes("stack_frame",2,"type")="addr"
	Set gtmtypfldindx("stack_frame","l_symtab")=2
	Set gtmtypes("stack_frame",3,"name")="stack_frame.mpc"
	Set gtmtypes("stack_frame",3,"off")=16
	Set gtmtypes("stack_frame",3,"len")=8
	Set gtmtypes("stack_frame",3,"type")="addr"
	Set gtmtypfldindx("stack_frame","mpc")=3
	Set gtmtypes("stack_frame",4,"name")="stack_frame.ctxt"
	Set gtmtypes("stack_frame",4,"off")=24
	Set gtmtypes("stack_frame",4,"len")=8
	Set gtmtypes("stack_frame",4,"type")="addr"
	Set gtmtypfldindx("stack_frame","ctxt")=4
	Set gtmtypes("stack_frame",5,"name")="stack_frame.literal_ptr"
	Set gtmtypes("stack_frame",5,"off")=32
	Set gtmtypes("stack_frame",5,"len")=8
	Set gtmtypes("stack_frame",5,"type")="addr"
	Set gtmtypfldindx("stack_frame","literal_ptr")=5
	Set gtmtypes("stack_frame",6,"name")="stack_frame.temps_ptr"
	Set gtmtypes("stack_frame",6,"off")=40
	Set gtmtypes("stack_frame",6,"len")=8
	Set gtmtypes("stack_frame",6,"type")="addr"
	Set gtmtypfldindx("stack_frame","temps_ptr")=6
	Set gtmtypes("stack_frame",7,"name")="stack_frame.vartab_ptr"
	Set gtmtypes("stack_frame",7,"off")=48
	Set gtmtypes("stack_frame",7,"len")=8
	Set gtmtypes("stack_frame",7,"type")="addr"
	Set gtmtypfldindx("stack_frame","vartab_ptr")=7
	Set gtmtypes("stack_frame",8,"name")="stack_frame.old_frame_pointer"
	Set gtmtypes("stack_frame",8,"off")=56
	Set gtmtypes("stack_frame",8,"len")=8
	Set gtmtypes("stack_frame",8,"type")="addr"
	Set gtmtypfldindx("stack_frame","old_frame_pointer")=8
	Set gtmtypes("stack_frame",9,"name")="stack_frame.vartab_len"
	Set gtmtypes("stack_frame",9,"off")=64
	Set gtmtypes("stack_frame",9,"len")=2
	Set gtmtypes("stack_frame",9,"type")="short"
	Set gtmtypfldindx("stack_frame","vartab_len")=9
	Set gtmtypes("stack_frame",10,"name")="stack_frame.temp_mvals"
	Set gtmtypes("stack_frame",10,"off")=66
	Set gtmtypes("stack_frame",10,"len")=2
	Set gtmtypes("stack_frame",10,"type")="short"
	Set gtmtypfldindx("stack_frame","temp_mvals")=10
	Set gtmtypes("stack_frame",11,"name")="stack_frame.type"
	Set gtmtypes("stack_frame",11,"off")=68
	Set gtmtypes("stack_frame",11,"len")=2
	Set gtmtypes("stack_frame",11,"type")="unsigned-short"
	Set gtmtypfldindx("stack_frame","type")=11
	Set gtmtypes("stack_frame",12,"name")="stack_frame.flags"
	Set gtmtypes("stack_frame",12,"off")=70
	Set gtmtypes("stack_frame",12,"len")=1
	Set gtmtypes("stack_frame",12,"type")="unsigned-char"
	Set gtmtypfldindx("stack_frame","flags")=12
	Set gtmtypes("stack_frame",13,"name")="stack_frame.dollar_test"
	Set gtmtypes("stack_frame",13,"off")=71
	Set gtmtypes("stack_frame",13,"len")=1
	Set gtmtypes("stack_frame",13,"type")="char"
	Set gtmtypfldindx("stack_frame","dollar_test")=13
	Set gtmtypes("stack_frame",14,"name")="stack_frame.for_ctrl_stack"
	Set gtmtypes("stack_frame",14,"off")=72
	Set gtmtypes("stack_frame",14,"len")=8
	Set gtmtypes("stack_frame",14,"type")="addr"
	Set gtmtypfldindx("stack_frame","for_ctrl_stack")=14
	Set gtmtypes("stack_frame",15,"name")="stack_frame.restart_pc"
	Set gtmtypes("stack_frame",15,"off")=80
	Set gtmtypes("stack_frame",15,"len")=8
	Set gtmtypes("stack_frame",15,"type")="addr"
	Set gtmtypfldindx("stack_frame","restart_pc")=15
	Set gtmtypes("stack_frame",16,"name")="stack_frame.restart_ctxt"
	Set gtmtypes("stack_frame",16,"off")=88
	Set gtmtypes("stack_frame",16,"len")=8
	Set gtmtypes("stack_frame",16,"type")="addr"
	Set gtmtypfldindx("stack_frame","restart_ctxt")=16
	Set gtmtypes("stack_frame",17,"name")="stack_frame.ret_value"
	Set gtmtypes("stack_frame",17,"off")=96
	Set gtmtypes("stack_frame",17,"len")=8
	Set gtmtypes("stack_frame",17,"type")="addr"
	Set gtmtypfldindx("stack_frame","ret_value")=17
	;
	Set gtmtypes("statsDB_deferred_init_que_elem")="struct"
	Set gtmtypes("statsDB_deferred_init_que_elem",0)=3
	Set gtmtypes("statsDB_deferred_init_que_elem","len")=24
	Set gtmtypes("statsDB_deferred_init_que_elem",1,"name")="statsDB_deferred_init_que_elem.next"
	Set gtmtypes("statsDB_deferred_init_que_elem",1,"off")=0
	Set gtmtypes("statsDB_deferred_init_que_elem",1,"len")=8
	Set gtmtypes("statsDB_deferred_init_que_elem",1,"type")="addr"
	Set gtmtypfldindx("statsDB_deferred_init_que_elem","next")=1
	Set gtmtypes("statsDB_deferred_init_que_elem",2,"name")="statsDB_deferred_init_que_elem.baseDBreg"
	Set gtmtypes("statsDB_deferred_init_que_elem",2,"off")=8
	Set gtmtypes("statsDB_deferred_init_que_elem",2,"len")=8
	Set gtmtypes("statsDB_deferred_init_que_elem",2,"type")="addr"
	Set gtmtypfldindx("statsDB_deferred_init_que_elem","baseDBreg")=2
	Set gtmtypes("statsDB_deferred_init_que_elem",3,"name")="statsDB_deferred_init_que_elem.statsDBreg"
	Set gtmtypes("statsDB_deferred_init_que_elem",3,"off")=16
	Set gtmtypes("statsDB_deferred_init_que_elem",3,"len")=8
	Set gtmtypes("statsDB_deferred_init_que_elem",3,"type")="addr"
	Set gtmtypfldindx("statsDB_deferred_init_que_elem","statsDBreg")=3
	;
	Set gtmtypes("storElem")="struct"
	Set gtmtypes("storElem",0)=11
	Set gtmtypes("storElem","len")=32
	Set gtmtypes("storElem",1,"name")="storElem.queueIndex"
	Set gtmtypes("storElem",1,"off")=0
	Set gtmtypes("storElem",1,"len")=1
	Set gtmtypes("storElem",1,"type")="char"
	Set gtmtypfldindx("storElem","queueIndex")=1
	Set gtmtypes("storElem",2,"name")="storElem.state"
	Set gtmtypes("storElem",2,"off")=1
	Set gtmtypes("storElem",2,"len")=1
	Set gtmtypes("storElem",2,"type")="unsigned-char"
	Set gtmtypfldindx("storElem","state")=2
	Set gtmtypes("storElem",3,"name")="storElem.extHdrOffset"
	Set gtmtypes("storElem",3,"off")=2
	Set gtmtypes("storElem",3,"len")=2
	Set gtmtypes("storElem",3,"type")="unsigned-short"
	Set gtmtypfldindx("storElem","extHdrOffset")=3
	Set gtmtypes("storElem",4,"name")="storElem.filler"
	Set gtmtypes("storElem",4,"off")=4
	Set gtmtypes("storElem",4,"len")=4
	Set gtmtypes("storElem",4,"type")="char"
	Set gtmtypfldindx("storElem","filler")=4
	Set gtmtypes("storElem",5,"name")="storElem.realLen"
	Set gtmtypes("storElem",5,"off")=8
	Set gtmtypes("storElem",5,"len")=8
	Set gtmtypes("storElem",5,"type")="size_t"
	Set gtmtypfldindx("storElem","realLen")=5
	Set gtmtypes("storElem",6,"name")="storElem.userStorage"
	Set gtmtypes("storElem",6,"off")=16
	Set gtmtypes("storElem",6,"len")=16
	Set gtmtypes("storElem",6,"type")="union"
	Set gtmtypfldindx("storElem","userStorage")=6
	Set gtmtypes("storElem",7,"name")="storElem.userStorage.deferFreeNext"
	Set gtmtypes("storElem",7,"off")=16
	Set gtmtypes("storElem",7,"len")=8
	Set gtmtypes("storElem",7,"type")="addr"
	Set gtmtypfldindx("storElem","userStorage.deferFreeNext")=7
	Set gtmtypes("storElem",8,"name")="storElem.userStorage.links"
	Set gtmtypes("storElem",8,"off")=16
	Set gtmtypes("storElem",8,"len")=16
	Set gtmtypes("storElem",8,"type")="struct"
	Set gtmtypfldindx("storElem","userStorage.links")=8
	Set gtmtypes("storElem",9,"name")="storElem.userStorage.links.fPtr"
	Set gtmtypes("storElem",9,"off")=16
	Set gtmtypes("storElem",9,"len")=8
	Set gtmtypes("storElem",9,"type")="addr"
	Set gtmtypfldindx("storElem","userStorage.links.fPtr")=9
	Set gtmtypes("storElem",10,"name")="storElem.userStorage.links.bPtr"
	Set gtmtypes("storElem",10,"off")=24
	Set gtmtypes("storElem",10,"len")=8
	Set gtmtypes("storElem",10,"type")="addr"
	Set gtmtypfldindx("storElem","userStorage.links.bPtr")=10
	Set gtmtypes("storElem",11,"name")="storElem.userStorage.userStart"
	Set gtmtypes("storElem",11,"off")=16
	Set gtmtypes("storElem",11,"len")=1
	Set gtmtypes("storElem",11,"type")="unsigned-char"
	Set gtmtypfldindx("storElem","userStorage.userStart")=11
	;
	Set gtmtypes("stringkey")="struct"
	Set gtmtypes("stringkey",0)=6
	Set gtmtypes("stringkey","len")=24
	Set gtmtypes("stringkey",1,"name")="stringkey.str"
	Set gtmtypes("stringkey",1,"off")=0
	Set gtmtypes("stringkey",1,"len")=16
	Set gtmtypes("stringkey",1,"type")="mstr"
	Set gtmtypfldindx("stringkey","str")=1
	Set gtmtypes("stringkey",2,"name")="stringkey.str.char_len"
	Set gtmtypes("stringkey",2,"off")=0
	Set gtmtypes("stringkey",2,"len")=4
	Set gtmtypes("stringkey",2,"type")="unsigned-int"
	Set gtmtypfldindx("stringkey","str.char_len")=2
	Set gtmtypes("stringkey",3,"name")="stringkey.str.len"
	Set gtmtypes("stringkey",3,"off")=4
	Set gtmtypes("stringkey",3,"len")=4
	Set gtmtypes("stringkey",3,"type")="int"
	Set gtmtypfldindx("stringkey","str.len")=3
	Set gtmtypes("stringkey",4,"name")="stringkey.str.addr"
	Set gtmtypes("stringkey",4,"off")=8
	Set gtmtypes("stringkey",4,"len")=8
	Set gtmtypes("stringkey",4,"type")="addr"
	Set gtmtypfldindx("stringkey","str.addr")=4
	Set gtmtypes("stringkey",5,"name")="stringkey.hash_code"
	Set gtmtypes("stringkey",5,"off")=16
	Set gtmtypes("stringkey",5,"len")=4
	Set gtmtypes("stringkey",5,"type")="unsigned-int"
	Set gtmtypfldindx("stringkey","hash_code")=5
	Set gtmtypes("stringkey",6,"name")="stringkey.filler"
	Set gtmtypes("stringkey",6,"off")=20
	Set gtmtypes("stringkey",6,"len")=4
	Set gtmtypes("stringkey",6,"type")="int"
	Set gtmtypfldindx("stringkey","filler")=6
	;
	Set gtmtypes("struct_jrec_align")="struct"
	Set gtmtypes("struct_jrec_align",0)=2
	Set gtmtypes("struct_jrec_align","len")=12
	Set gtmtypes("struct_jrec_align",1,"name")="struct_jrec_align.checksum"
	Set gtmtypes("struct_jrec_align",1,"off")=4
	Set gtmtypes("struct_jrec_align",1,"len")=4
	Set gtmtypes("struct_jrec_align",1,"type")="unsigned-int"
	Set gtmtypfldindx("struct_jrec_align","checksum")=1
	Set gtmtypes("struct_jrec_align",2,"name")="struct_jrec_align.time"
	Set gtmtypes("struct_jrec_align",2,"off")=8
	Set gtmtypes("struct_jrec_align",2,"len")=4
	Set gtmtypes("struct_jrec_align",2,"type")="unsigned-int"
	Set gtmtypfldindx("struct_jrec_align","time")=2
	;
	Set gtmtypes("struct_jrec_blk")="struct"
	Set gtmtypes("struct_jrec_blk",0)=10
	Set gtmtypes("struct_jrec_blk","len")=48
	Set gtmtypes("struct_jrec_blk",1,"name")="struct_jrec_blk.prefix"
	Set gtmtypes("struct_jrec_blk",1,"off")=0
	Set gtmtypes("struct_jrec_blk",1,"len")=24
	Set gtmtypes("struct_jrec_blk",1,"type")="jrec_prefix"
	Set gtmtypfldindx("struct_jrec_blk","prefix")=1
	Set gtmtypes("struct_jrec_blk",2,"name")="struct_jrec_blk.prefix.pini_addr"
	Set gtmtypes("struct_jrec_blk",2,"off")=4
	Set gtmtypes("struct_jrec_blk",2,"len")=4
	Set gtmtypes("struct_jrec_blk",2,"type")="unsigned-int"
	Set gtmtypfldindx("struct_jrec_blk","prefix.pini_addr")=2
	Set gtmtypes("struct_jrec_blk",3,"name")="struct_jrec_blk.prefix.time"
	Set gtmtypes("struct_jrec_blk",3,"off")=8
	Set gtmtypes("struct_jrec_blk",3,"len")=4
	Set gtmtypes("struct_jrec_blk",3,"type")="unsigned-int"
	Set gtmtypfldindx("struct_jrec_blk","prefix.time")=3
	Set gtmtypes("struct_jrec_blk",4,"name")="struct_jrec_blk.prefix.checksum"
	Set gtmtypes("struct_jrec_blk",4,"off")=12
	Set gtmtypes("struct_jrec_blk",4,"len")=4
	Set gtmtypes("struct_jrec_blk",4,"type")="unsigned-int"
	Set gtmtypfldindx("struct_jrec_blk","prefix.checksum")=4
	Set gtmtypes("struct_jrec_blk",5,"name")="struct_jrec_blk.prefix.tn"
	Set gtmtypes("struct_jrec_blk",5,"off")=16
	Set gtmtypes("struct_jrec_blk",5,"len")=8
	Set gtmtypes("struct_jrec_blk",5,"type")="uint64_t"
	Set gtmtypfldindx("struct_jrec_blk","prefix.tn")=5
	Set gtmtypes("struct_jrec_blk",6,"name")="struct_jrec_blk.blknum"
	Set gtmtypes("struct_jrec_blk",6,"off")=24
	Set gtmtypes("struct_jrec_blk",6,"len")=4
	Set gtmtypes("struct_jrec_blk",6,"type")="int"
	Set gtmtypfldindx("struct_jrec_blk","blknum")=6
	Set gtmtypes("struct_jrec_blk",7,"name")="struct_jrec_blk.bsiz"
	Set gtmtypes("struct_jrec_blk",7,"off")=28
	Set gtmtypes("struct_jrec_blk",7,"len")=4
	Set gtmtypes("struct_jrec_blk",7,"type")="unsigned-int"
	Set gtmtypfldindx("struct_jrec_blk","bsiz")=7
	Set gtmtypes("struct_jrec_blk",8,"name")="struct_jrec_blk.ondsk_blkver"
	Set gtmtypes("struct_jrec_blk",8,"off")=32
	Set gtmtypes("struct_jrec_blk",8,"len")=4
	Set gtmtypes("struct_jrec_blk",8,"type")="int"
	Set gtmtypfldindx("struct_jrec_blk","ondsk_blkver")=8
	Set gtmtypes("struct_jrec_blk",9,"name")="struct_jrec_blk.filler"
	Set gtmtypes("struct_jrec_blk",9,"off")=36
	Set gtmtypes("struct_jrec_blk",9,"len")=4
	Set gtmtypes("struct_jrec_blk",9,"type")="int"
	Set gtmtypfldindx("struct_jrec_blk","filler")=9
	Set gtmtypes("struct_jrec_blk",10,"name")="struct_jrec_blk.blk_contents"
	Set gtmtypes("struct_jrec_blk",10,"off")=40
	Set gtmtypes("struct_jrec_blk",10,"len")=1
	Set gtmtypes("struct_jrec_blk",10,"type")="char"
	Set gtmtypfldindx("struct_jrec_blk","blk_contents")=10
	;
	Set gtmtypes("struct_jrec_eof")="struct"
	Set gtmtypes("struct_jrec_eof",0)=8
	Set gtmtypes("struct_jrec_eof","len")=40
	Set gtmtypes("struct_jrec_eof",1,"name")="struct_jrec_eof.prefix"
	Set gtmtypes("struct_jrec_eof",1,"off")=0
	Set gtmtypes("struct_jrec_eof",1,"len")=24
	Set gtmtypes("struct_jrec_eof",1,"type")="jrec_prefix"
	Set gtmtypfldindx("struct_jrec_eof","prefix")=1
	Set gtmtypes("struct_jrec_eof",2,"name")="struct_jrec_eof.prefix.pini_addr"
	Set gtmtypes("struct_jrec_eof",2,"off")=4
	Set gtmtypes("struct_jrec_eof",2,"len")=4
	Set gtmtypes("struct_jrec_eof",2,"type")="unsigned-int"
	Set gtmtypfldindx("struct_jrec_eof","prefix.pini_addr")=2
	Set gtmtypes("struct_jrec_eof",3,"name")="struct_jrec_eof.prefix.time"
	Set gtmtypes("struct_jrec_eof",3,"off")=8
	Set gtmtypes("struct_jrec_eof",3,"len")=4
	Set gtmtypes("struct_jrec_eof",3,"type")="unsigned-int"
	Set gtmtypfldindx("struct_jrec_eof","prefix.time")=3
	Set gtmtypes("struct_jrec_eof",4,"name")="struct_jrec_eof.prefix.checksum"
	Set gtmtypes("struct_jrec_eof",4,"off")=12
	Set gtmtypes("struct_jrec_eof",4,"len")=4
	Set gtmtypes("struct_jrec_eof",4,"type")="unsigned-int"
	Set gtmtypfldindx("struct_jrec_eof","prefix.checksum")=4
	Set gtmtypes("struct_jrec_eof",5,"name")="struct_jrec_eof.prefix.tn"
	Set gtmtypes("struct_jrec_eof",5,"off")=16
	Set gtmtypes("struct_jrec_eof",5,"len")=8
	Set gtmtypes("struct_jrec_eof",5,"type")="uint64_t"
	Set gtmtypfldindx("struct_jrec_eof","prefix.tn")=5
	Set gtmtypes("struct_jrec_eof",6,"name")="struct_jrec_eof.jnl_seqno"
	Set gtmtypes("struct_jrec_eof",6,"off")=24
	Set gtmtypes("struct_jrec_eof",6,"len")=8
	Set gtmtypes("struct_jrec_eof",6,"type")="uint64_t"
	Set gtmtypfldindx("struct_jrec_eof","jnl_seqno")=6
	Set gtmtypes("struct_jrec_eof",7,"name")="struct_jrec_eof.filler"
	Set gtmtypes("struct_jrec_eof",7,"off")=32
	Set gtmtypes("struct_jrec_eof",7,"len")=4
	Set gtmtypes("struct_jrec_eof",7,"type")="unsigned-int"
	Set gtmtypfldindx("struct_jrec_eof","filler")=7
	Set gtmtypes("struct_jrec_eof",8,"name")="struct_jrec_eof.suffix"
	Set gtmtypes("struct_jrec_eof",8,"off")=36
	Set gtmtypes("struct_jrec_eof",8,"len")=4
	Set gtmtypes("struct_jrec_eof",8,"type")="jrec_suffix"
	Set gtmtypfldindx("struct_jrec_eof","suffix")=8
	;
	Set gtmtypes("struct_jrec_epoch")="struct"
	Set gtmtypes("struct_jrec_epoch",0)=13
	Set gtmtypes("struct_jrec_epoch","len")=184
	Set gtmtypes("struct_jrec_epoch",1,"name")="struct_jrec_epoch.prefix"
	Set gtmtypes("struct_jrec_epoch",1,"off")=0
	Set gtmtypes("struct_jrec_epoch",1,"len")=24
	Set gtmtypes("struct_jrec_epoch",1,"type")="jrec_prefix"
	Set gtmtypfldindx("struct_jrec_epoch","prefix")=1
	Set gtmtypes("struct_jrec_epoch",2,"name")="struct_jrec_epoch.prefix.pini_addr"
	Set gtmtypes("struct_jrec_epoch",2,"off")=4
	Set gtmtypes("struct_jrec_epoch",2,"len")=4
	Set gtmtypes("struct_jrec_epoch",2,"type")="unsigned-int"
	Set gtmtypfldindx("struct_jrec_epoch","prefix.pini_addr")=2
	Set gtmtypes("struct_jrec_epoch",3,"name")="struct_jrec_epoch.prefix.time"
	Set gtmtypes("struct_jrec_epoch",3,"off")=8
	Set gtmtypes("struct_jrec_epoch",3,"len")=4
	Set gtmtypes("struct_jrec_epoch",3,"type")="unsigned-int"
	Set gtmtypfldindx("struct_jrec_epoch","prefix.time")=3
	Set gtmtypes("struct_jrec_epoch",4,"name")="struct_jrec_epoch.prefix.checksum"
	Set gtmtypes("struct_jrec_epoch",4,"off")=12
	Set gtmtypes("struct_jrec_epoch",4,"len")=4
	Set gtmtypes("struct_jrec_epoch",4,"type")="unsigned-int"
	Set gtmtypfldindx("struct_jrec_epoch","prefix.checksum")=4
	Set gtmtypes("struct_jrec_epoch",5,"name")="struct_jrec_epoch.prefix.tn"
	Set gtmtypes("struct_jrec_epoch",5,"off")=16
	Set gtmtypes("struct_jrec_epoch",5,"len")=8
	Set gtmtypes("struct_jrec_epoch",5,"type")="uint64_t"
	Set gtmtypfldindx("struct_jrec_epoch","prefix.tn")=5
	Set gtmtypes("struct_jrec_epoch",6,"name")="struct_jrec_epoch.jnl_seqno"
	Set gtmtypes("struct_jrec_epoch",6,"off")=24
	Set gtmtypes("struct_jrec_epoch",6,"len")=8
	Set gtmtypes("struct_jrec_epoch",6,"type")="uint64_t"
	Set gtmtypfldindx("struct_jrec_epoch","jnl_seqno")=6
	Set gtmtypes("struct_jrec_epoch",7,"name")="struct_jrec_epoch.blks_to_upgrd"
	Set gtmtypes("struct_jrec_epoch",7,"off")=32
	Set gtmtypes("struct_jrec_epoch",7,"len")=4
	Set gtmtypes("struct_jrec_epoch",7,"type")="unsigned-int"
	Set gtmtypfldindx("struct_jrec_epoch","blks_to_upgrd")=7
	Set gtmtypes("struct_jrec_epoch",8,"name")="struct_jrec_epoch.free_blocks"
	Set gtmtypes("struct_jrec_epoch",8,"off")=36
	Set gtmtypes("struct_jrec_epoch",8,"len")=4
	Set gtmtypes("struct_jrec_epoch",8,"type")="unsigned-int"
	Set gtmtypfldindx("struct_jrec_epoch","free_blocks")=8
	Set gtmtypes("struct_jrec_epoch",9,"name")="struct_jrec_epoch.total_blks"
	Set gtmtypes("struct_jrec_epoch",9,"off")=40
	Set gtmtypes("struct_jrec_epoch",9,"len")=4
	Set gtmtypes("struct_jrec_epoch",9,"type")="unsigned-int"
	Set gtmtypfldindx("struct_jrec_epoch","total_blks")=9
	Set gtmtypes("struct_jrec_epoch",10,"name")="struct_jrec_epoch.fully_upgraded"
	Set gtmtypes("struct_jrec_epoch",10,"off")=44
	Set gtmtypes("struct_jrec_epoch",10,"len")=4
	Set gtmtypes("struct_jrec_epoch",10,"type")="boolean_t"
	Set gtmtypfldindx("struct_jrec_epoch","fully_upgraded")=10
	Set gtmtypes("struct_jrec_epoch",11,"name")="struct_jrec_epoch.strm_seqno"
	Set gtmtypes("struct_jrec_epoch",11,"off")=48
	Set gtmtypes("struct_jrec_epoch",11,"len")=128
	Set gtmtypes("struct_jrec_epoch",11,"type")="uint64_t"
	Set gtmtypfldindx("struct_jrec_epoch","strm_seqno")=11
	Set gtmtypes("struct_jrec_epoch",11,"dim")=16
	Set gtmtypes("struct_jrec_epoch",12,"name")="struct_jrec_epoch.filler"
	Set gtmtypes("struct_jrec_epoch",12,"off")=176
	Set gtmtypes("struct_jrec_epoch",12,"len")=4
	Set gtmtypes("struct_jrec_epoch",12,"type")="unsigned-int"
	Set gtmtypfldindx("struct_jrec_epoch","filler")=12
	Set gtmtypes("struct_jrec_epoch",13,"name")="struct_jrec_epoch.suffix"
	Set gtmtypes("struct_jrec_epoch",13,"off")=180
	Set gtmtypes("struct_jrec_epoch",13,"len")=4
	Set gtmtypes("struct_jrec_epoch",13,"type")="jrec_suffix"
	Set gtmtypfldindx("struct_jrec_epoch","suffix")=13
	;
	Set gtmtypes("struct_jrec_inctn")="struct"
	Set gtmtypes("struct_jrec_inctn",0)=18
	Set gtmtypes("struct_jrec_inctn","len")=40
	Set gtmtypes("struct_jrec_inctn",1,"name")="struct_jrec_inctn.prefix"
	Set gtmtypes("struct_jrec_inctn",1,"off")=0
	Set gtmtypes("struct_jrec_inctn",1,"len")=24
	Set gtmtypes("struct_jrec_inctn",1,"type")="jrec_prefix"
	Set gtmtypfldindx("struct_jrec_inctn","prefix")=1
	Set gtmtypes("struct_jrec_inctn",2,"name")="struct_jrec_inctn.prefix.pini_addr"
	Set gtmtypes("struct_jrec_inctn",2,"off")=4
	Set gtmtypes("struct_jrec_inctn",2,"len")=4
	Set gtmtypes("struct_jrec_inctn",2,"type")="unsigned-int"
	Set gtmtypfldindx("struct_jrec_inctn","prefix.pini_addr")=2
	Set gtmtypes("struct_jrec_inctn",3,"name")="struct_jrec_inctn.prefix.time"
	Set gtmtypes("struct_jrec_inctn",3,"off")=8
	Set gtmtypes("struct_jrec_inctn",3,"len")=4
	Set gtmtypes("struct_jrec_inctn",3,"type")="unsigned-int"
	Set gtmtypfldindx("struct_jrec_inctn","prefix.time")=3
	Set gtmtypes("struct_jrec_inctn",4,"name")="struct_jrec_inctn.prefix.checksum"
	Set gtmtypes("struct_jrec_inctn",4,"off")=12
	Set gtmtypes("struct_jrec_inctn",4,"len")=4
	Set gtmtypes("struct_jrec_inctn",4,"type")="unsigned-int"
	Set gtmtypfldindx("struct_jrec_inctn","prefix.checksum")=4
	Set gtmtypes("struct_jrec_inctn",5,"name")="struct_jrec_inctn.prefix.tn"
	Set gtmtypes("struct_jrec_inctn",5,"off")=16
	Set gtmtypes("struct_jrec_inctn",5,"len")=8
	Set gtmtypes("struct_jrec_inctn",5,"type")="uint64_t"
	Set gtmtypfldindx("struct_jrec_inctn","prefix.tn")=5
	Set gtmtypes("struct_jrec_inctn",6,"name")="struct_jrec_inctn.detail"
	Set gtmtypes("struct_jrec_inctn",6,"off")=24
	Set gtmtypes("struct_jrec_inctn",6,"len")=16
	Set gtmtypes("struct_jrec_inctn",6,"type")="inctn_detail_t"
	Set gtmtypfldindx("struct_jrec_inctn","detail")=6
	Set gtmtypes("struct_jrec_inctn",7,"name")="struct_jrec_inctn.detail.blknum_struct"
	Set gtmtypes("struct_jrec_inctn",7,"off")=24
	Set gtmtypes("struct_jrec_inctn",7,"len")=16
	Set gtmtypes("struct_jrec_inctn",7,"type")="inctn_detail_blknum_t"
	Set gtmtypfldindx("struct_jrec_inctn","detail.blknum_struct")=7
	Set gtmtypes("struct_jrec_inctn",8,"name")="struct_jrec_inctn.detail.blknum_struct.blknum"
	Set gtmtypes("struct_jrec_inctn",8,"off")=24
	Set gtmtypes("struct_jrec_inctn",8,"len")=4
	Set gtmtypes("struct_jrec_inctn",8,"type")="int"
	Set gtmtypfldindx("struct_jrec_inctn","detail.blknum_struct.blknum")=8
	Set gtmtypes("struct_jrec_inctn",9,"name")="struct_jrec_inctn.detail.blknum_struct.filler_uint4"
	Set gtmtypes("struct_jrec_inctn",9,"off")=28
	Set gtmtypes("struct_jrec_inctn",9,"len")=4
	Set gtmtypes("struct_jrec_inctn",9,"type")="unsigned-int"
	Set gtmtypfldindx("struct_jrec_inctn","detail.blknum_struct.filler_uint4")=9
	Set gtmtypes("struct_jrec_inctn",10,"name")="struct_jrec_inctn.detail.blknum_struct.filler_short"
	Set gtmtypes("struct_jrec_inctn",10,"off")=32
	Set gtmtypes("struct_jrec_inctn",10,"len")=2
	Set gtmtypes("struct_jrec_inctn",10,"type")="unsigned-short"
	Set gtmtypfldindx("struct_jrec_inctn","detail.blknum_struct.filler_short")=10
	Set gtmtypes("struct_jrec_inctn",11,"name")="struct_jrec_inctn.detail.blknum_struct.opcode"
	Set gtmtypes("struct_jrec_inctn",11,"off")=34
	Set gtmtypes("struct_jrec_inctn",11,"len")=2
	Set gtmtypes("struct_jrec_inctn",11,"type")="unsigned-short"
	Set gtmtypfldindx("struct_jrec_inctn","detail.blknum_struct.opcode")=11
	Set gtmtypes("struct_jrec_inctn",12,"name")="struct_jrec_inctn.detail.blknum_struct.suffix"
	Set gtmtypes("struct_jrec_inctn",12,"off")=36
	Set gtmtypes("struct_jrec_inctn",12,"len")=4
	Set gtmtypes("struct_jrec_inctn",12,"type")="jrec_suffix"
	Set gtmtypfldindx("struct_jrec_inctn","detail.blknum_struct.suffix")=12
	Set gtmtypes("struct_jrec_inctn",13,"name")="struct_jrec_inctn.detail.blks2upgrd_struct"
	Set gtmtypes("struct_jrec_inctn",13,"off")=24
	Set gtmtypes("struct_jrec_inctn",13,"len")=16
	Set gtmtypes("struct_jrec_inctn",13,"type")="inctn_detail_blks2upgrd_t"
	Set gtmtypfldindx("struct_jrec_inctn","detail.blks2upgrd_struct")=13
	Set gtmtypes("struct_jrec_inctn",14,"name")="struct_jrec_inctn.detail.blks2upgrd_struct.blks_to_upgrd_delta"
	Set gtmtypes("struct_jrec_inctn",14,"off")=24
	Set gtmtypes("struct_jrec_inctn",14,"len")=4
	Set gtmtypes("struct_jrec_inctn",14,"type")="int"
	Set gtmtypfldindx("struct_jrec_inctn","detail.blks2upgrd_struct.blks_to_upgrd_delta")=14
	Set gtmtypes("struct_jrec_inctn",15,"name")="struct_jrec_inctn.detail.blks2upgrd_struct.filler_uint4"
	Set gtmtypes("struct_jrec_inctn",15,"off")=28
	Set gtmtypes("struct_jrec_inctn",15,"len")=4
	Set gtmtypes("struct_jrec_inctn",15,"type")="unsigned-int"
	Set gtmtypfldindx("struct_jrec_inctn","detail.blks2upgrd_struct.filler_uint4")=15
	Set gtmtypes("struct_jrec_inctn",16,"name")="struct_jrec_inctn.detail.blks2upgrd_struct.filler_short"
	Set gtmtypes("struct_jrec_inctn",16,"off")=32
	Set gtmtypes("struct_jrec_inctn",16,"len")=2
	Set gtmtypes("struct_jrec_inctn",16,"type")="unsigned-short"
	Set gtmtypfldindx("struct_jrec_inctn","detail.blks2upgrd_struct.filler_short")=16
	Set gtmtypes("struct_jrec_inctn",17,"name")="struct_jrec_inctn.detail.blks2upgrd_struct.opcode"
	Set gtmtypes("struct_jrec_inctn",17,"off")=34
	Set gtmtypes("struct_jrec_inctn",17,"len")=2
	Set gtmtypes("struct_jrec_inctn",17,"type")="unsigned-short"
	Set gtmtypfldindx("struct_jrec_inctn","detail.blks2upgrd_struct.opcode")=17
	Set gtmtypes("struct_jrec_inctn",18,"name")="struct_jrec_inctn.detail.blks2upgrd_struct.suffix"
	Set gtmtypes("struct_jrec_inctn",18,"off")=36
	Set gtmtypes("struct_jrec_inctn",18,"len")=4
	Set gtmtypes("struct_jrec_inctn",18,"type")="jrec_suffix"
	Set gtmtypfldindx("struct_jrec_inctn","detail.blks2upgrd_struct.suffix")=18
	;
	Set gtmtypes("struct_jrec_lgtrig")="struct"
	Set gtmtypes("struct_jrec_lgtrig",0)=14
	Set gtmtypes("struct_jrec_lgtrig","len")=56
	Set gtmtypes("struct_jrec_lgtrig",1,"name")="struct_jrec_lgtrig.prefix"
	Set gtmtypes("struct_jrec_lgtrig",1,"off")=0
	Set gtmtypes("struct_jrec_lgtrig",1,"len")=24
	Set gtmtypes("struct_jrec_lgtrig",1,"type")="jrec_prefix"
	Set gtmtypfldindx("struct_jrec_lgtrig","prefix")=1
	Set gtmtypes("struct_jrec_lgtrig",2,"name")="struct_jrec_lgtrig.prefix.pini_addr"
	Set gtmtypes("struct_jrec_lgtrig",2,"off")=4
	Set gtmtypes("struct_jrec_lgtrig",2,"len")=4
	Set gtmtypes("struct_jrec_lgtrig",2,"type")="unsigned-int"
	Set gtmtypfldindx("struct_jrec_lgtrig","prefix.pini_addr")=2
	Set gtmtypes("struct_jrec_lgtrig",3,"name")="struct_jrec_lgtrig.prefix.time"
	Set gtmtypes("struct_jrec_lgtrig",3,"off")=8
	Set gtmtypes("struct_jrec_lgtrig",3,"len")=4
	Set gtmtypes("struct_jrec_lgtrig",3,"type")="unsigned-int"
	Set gtmtypfldindx("struct_jrec_lgtrig","prefix.time")=3
	Set gtmtypes("struct_jrec_lgtrig",4,"name")="struct_jrec_lgtrig.prefix.checksum"
	Set gtmtypes("struct_jrec_lgtrig",4,"off")=12
	Set gtmtypes("struct_jrec_lgtrig",4,"len")=4
	Set gtmtypes("struct_jrec_lgtrig",4,"type")="unsigned-int"
	Set gtmtypfldindx("struct_jrec_lgtrig","prefix.checksum")=4
	Set gtmtypes("struct_jrec_lgtrig",5,"name")="struct_jrec_lgtrig.prefix.tn"
	Set gtmtypes("struct_jrec_lgtrig",5,"off")=16
	Set gtmtypes("struct_jrec_lgtrig",5,"len")=8
	Set gtmtypes("struct_jrec_lgtrig",5,"type")="uint64_t"
	Set gtmtypfldindx("struct_jrec_lgtrig","prefix.tn")=5
	Set gtmtypes("struct_jrec_lgtrig",6,"name")="struct_jrec_lgtrig.token_seq"
	Set gtmtypes("struct_jrec_lgtrig",6,"off")=24
	Set gtmtypes("struct_jrec_lgtrig",6,"len")=8
	Set gtmtypes("struct_jrec_lgtrig",6,"type")="token_seq_t"
	Set gtmtypfldindx("struct_jrec_lgtrig","token_seq")=6
	Set gtmtypes("struct_jrec_lgtrig",7,"name")="struct_jrec_lgtrig.token_seq.jnl_seqno"
	Set gtmtypes("struct_jrec_lgtrig",7,"off")=24
	Set gtmtypes("struct_jrec_lgtrig",7,"len")=8
	Set gtmtypes("struct_jrec_lgtrig",7,"type")="uint64_t"
	Set gtmtypfldindx("struct_jrec_lgtrig","token_seq.jnl_seqno")=7
	Set gtmtypes("struct_jrec_lgtrig",8,"name")="struct_jrec_lgtrig.token_seq.token"
	Set gtmtypes("struct_jrec_lgtrig",8,"off")=24
	Set gtmtypes("struct_jrec_lgtrig",8,"len")=8
	Set gtmtypes("struct_jrec_lgtrig",8,"type")="uint64_t"
	Set gtmtypfldindx("struct_jrec_lgtrig","token_seq.token")=8
	Set gtmtypes("struct_jrec_lgtrig",9,"name")="struct_jrec_lgtrig.strm_seqno"
	Set gtmtypes("struct_jrec_lgtrig",9,"off")=32
	Set gtmtypes("struct_jrec_lgtrig",9,"len")=8
	Set gtmtypes("struct_jrec_lgtrig",9,"type")="uint64_t"
	Set gtmtypfldindx("struct_jrec_lgtrig","strm_seqno")=9
	Set gtmtypes("struct_jrec_lgtrig",10,"name")="struct_jrec_lgtrig.update_num"
	Set gtmtypes("struct_jrec_lgtrig",10,"off")=40
	Set gtmtypes("struct_jrec_lgtrig",10,"len")=4
	Set gtmtypes("struct_jrec_lgtrig",10,"type")="unsigned-int"
	Set gtmtypfldindx("struct_jrec_lgtrig","update_num")=10
	Set gtmtypes("struct_jrec_lgtrig",11,"name")="struct_jrec_lgtrig.filler_short"
	Set gtmtypes("struct_jrec_lgtrig",11,"off")=44
	Set gtmtypes("struct_jrec_lgtrig",11,"len")=2
	Set gtmtypes("struct_jrec_lgtrig",11,"type")="unsigned-short"
	Set gtmtypfldindx("struct_jrec_lgtrig","filler_short")=11
	Set gtmtypes("struct_jrec_lgtrig",12,"name")="struct_jrec_lgtrig.num_participants"
	Set gtmtypes("struct_jrec_lgtrig",12,"off")=46
	Set gtmtypes("struct_jrec_lgtrig",12,"len")=2
	Set gtmtypes("struct_jrec_lgtrig",12,"type")="unsigned-short"
	Set gtmtypfldindx("struct_jrec_lgtrig","num_participants")=12
	Set gtmtypes("struct_jrec_lgtrig",13,"name")="struct_jrec_lgtrig.lgtrig_str"
	Set gtmtypes("struct_jrec_lgtrig",13,"off")=48
	Set gtmtypes("struct_jrec_lgtrig",13,"len")=8
	Set gtmtypes("struct_jrec_lgtrig",13,"type")="jnl_string"
	Set gtmtypfldindx("struct_jrec_lgtrig","lgtrig_str")=13
	Set gtmtypes("struct_jrec_lgtrig",14,"name")="struct_jrec_lgtrig.lgtrig_str.text"
	Set gtmtypes("struct_jrec_lgtrig",14,"off")=52
	Set gtmtypes("struct_jrec_lgtrig",14,"len")=1
	Set gtmtypes("struct_jrec_lgtrig",14,"type")="char"
	Set gtmtypfldindx("struct_jrec_lgtrig","lgtrig_str.text")=14
	;
	Set gtmtypes("struct_jrec_null")="struct"
	Set gtmtypes("struct_jrec_null",0)=9
	Set gtmtypes("struct_jrec_null","len")=48
	Set gtmtypes("struct_jrec_null",1,"name")="struct_jrec_null.prefix"
	Set gtmtypes("struct_jrec_null",1,"off")=0
	Set gtmtypes("struct_jrec_null",1,"len")=24
	Set gtmtypes("struct_jrec_null",1,"type")="jrec_prefix"
	Set gtmtypfldindx("struct_jrec_null","prefix")=1
	Set gtmtypes("struct_jrec_null",2,"name")="struct_jrec_null.prefix.pini_addr"
	Set gtmtypes("struct_jrec_null",2,"off")=4
	Set gtmtypes("struct_jrec_null",2,"len")=4
	Set gtmtypes("struct_jrec_null",2,"type")="unsigned-int"
	Set gtmtypfldindx("struct_jrec_null","prefix.pini_addr")=2
	Set gtmtypes("struct_jrec_null",3,"name")="struct_jrec_null.prefix.time"
	Set gtmtypes("struct_jrec_null",3,"off")=8
	Set gtmtypes("struct_jrec_null",3,"len")=4
	Set gtmtypes("struct_jrec_null",3,"type")="unsigned-int"
	Set gtmtypfldindx("struct_jrec_null","prefix.time")=3
	Set gtmtypes("struct_jrec_null",4,"name")="struct_jrec_null.prefix.checksum"
	Set gtmtypes("struct_jrec_null",4,"off")=12
	Set gtmtypes("struct_jrec_null",4,"len")=4
	Set gtmtypes("struct_jrec_null",4,"type")="unsigned-int"
	Set gtmtypfldindx("struct_jrec_null","prefix.checksum")=4
	Set gtmtypes("struct_jrec_null",5,"name")="struct_jrec_null.prefix.tn"
	Set gtmtypes("struct_jrec_null",5,"off")=16
	Set gtmtypes("struct_jrec_null",5,"len")=8
	Set gtmtypes("struct_jrec_null",5,"type")="uint64_t"
	Set gtmtypfldindx("struct_jrec_null","prefix.tn")=5
	Set gtmtypes("struct_jrec_null",6,"name")="struct_jrec_null.jnl_seqno"
	Set gtmtypes("struct_jrec_null",6,"off")=24
	Set gtmtypes("struct_jrec_null",6,"len")=8
	Set gtmtypes("struct_jrec_null",6,"type")="uint64_t"
	Set gtmtypfldindx("struct_jrec_null","jnl_seqno")=6
	Set gtmtypes("struct_jrec_null",7,"name")="struct_jrec_null.strm_seqno"
	Set gtmtypes("struct_jrec_null",7,"off")=32
	Set gtmtypes("struct_jrec_null",7,"len")=8
	Set gtmtypes("struct_jrec_null",7,"type")="uint64_t"
	Set gtmtypfldindx("struct_jrec_null","strm_seqno")=7
	Set gtmtypes("struct_jrec_null",8,"name")="struct_jrec_null.filler"
	Set gtmtypes("struct_jrec_null",8,"off")=40
	Set gtmtypes("struct_jrec_null",8,"len")=4
	Set gtmtypes("struct_jrec_null",8,"type")="unsigned-int"
	Set gtmtypfldindx("struct_jrec_null","filler")=8
	Set gtmtypes("struct_jrec_null",9,"name")="struct_jrec_null.suffix"
	Set gtmtypes("struct_jrec_null",9,"off")=44
	Set gtmtypes("struct_jrec_null",9,"len")=4
	Set gtmtypes("struct_jrec_null",9,"type")="jrec_suffix"
	Set gtmtypfldindx("struct_jrec_null","suffix")=9
	;
	Set gtmtypes("struct_jrec_pfin")="struct"
	Set gtmtypes("struct_jrec_pfin",0)=7
	Set gtmtypes("struct_jrec_pfin","len")=32
	Set gtmtypes("struct_jrec_pfin",1,"name")="struct_jrec_pfin.prefix"
	Set gtmtypes("struct_jrec_pfin",1,"off")=0
	Set gtmtypes("struct_jrec_pfin",1,"len")=24
	Set gtmtypes("struct_jrec_pfin",1,"type")="jrec_prefix"
	Set gtmtypfldindx("struct_jrec_pfin","prefix")=1
	Set gtmtypes("struct_jrec_pfin",2,"name")="struct_jrec_pfin.prefix.pini_addr"
	Set gtmtypes("struct_jrec_pfin",2,"off")=4
	Set gtmtypes("struct_jrec_pfin",2,"len")=4
	Set gtmtypes("struct_jrec_pfin",2,"type")="unsigned-int"
	Set gtmtypfldindx("struct_jrec_pfin","prefix.pini_addr")=2
	Set gtmtypes("struct_jrec_pfin",3,"name")="struct_jrec_pfin.prefix.time"
	Set gtmtypes("struct_jrec_pfin",3,"off")=8
	Set gtmtypes("struct_jrec_pfin",3,"len")=4
	Set gtmtypes("struct_jrec_pfin",3,"type")="unsigned-int"
	Set gtmtypfldindx("struct_jrec_pfin","prefix.time")=3
	Set gtmtypes("struct_jrec_pfin",4,"name")="struct_jrec_pfin.prefix.checksum"
	Set gtmtypes("struct_jrec_pfin",4,"off")=12
	Set gtmtypes("struct_jrec_pfin",4,"len")=4
	Set gtmtypes("struct_jrec_pfin",4,"type")="unsigned-int"
	Set gtmtypfldindx("struct_jrec_pfin","prefix.checksum")=4
	Set gtmtypes("struct_jrec_pfin",5,"name")="struct_jrec_pfin.prefix.tn"
	Set gtmtypes("struct_jrec_pfin",5,"off")=16
	Set gtmtypes("struct_jrec_pfin",5,"len")=8
	Set gtmtypes("struct_jrec_pfin",5,"type")="uint64_t"
	Set gtmtypfldindx("struct_jrec_pfin","prefix.tn")=5
	Set gtmtypes("struct_jrec_pfin",6,"name")="struct_jrec_pfin.filler"
	Set gtmtypes("struct_jrec_pfin",6,"off")=24
	Set gtmtypes("struct_jrec_pfin",6,"len")=4
	Set gtmtypes("struct_jrec_pfin",6,"type")="unsigned-int"
	Set gtmtypfldindx("struct_jrec_pfin","filler")=6
	Set gtmtypes("struct_jrec_pfin",7,"name")="struct_jrec_pfin.suffix"
	Set gtmtypes("struct_jrec_pfin",7,"off")=28
	Set gtmtypes("struct_jrec_pfin",7,"len")=4
	Set gtmtypes("struct_jrec_pfin",7,"type")="jrec_suffix"
	Set gtmtypfldindx("struct_jrec_pfin","suffix")=7
	;
	Set gtmtypes("struct_jrec_pini")="struct"
	Set gtmtypes("struct_jrec_pini",0)=18
	Set gtmtypes("struct_jrec_pini","len")=208
	Set gtmtypes("struct_jrec_pini",1,"name")="struct_jrec_pini.prefix"
	Set gtmtypes("struct_jrec_pini",1,"off")=0
	Set gtmtypes("struct_jrec_pini",1,"len")=24
	Set gtmtypes("struct_jrec_pini",1,"type")="jrec_prefix"
	Set gtmtypfldindx("struct_jrec_pini","prefix")=1
	Set gtmtypes("struct_jrec_pini",2,"name")="struct_jrec_pini.prefix.pini_addr"
	Set gtmtypes("struct_jrec_pini",2,"off")=4
	Set gtmtypes("struct_jrec_pini",2,"len")=4
	Set gtmtypes("struct_jrec_pini",2,"type")="unsigned-int"
	Set gtmtypfldindx("struct_jrec_pini","prefix.pini_addr")=2
	Set gtmtypes("struct_jrec_pini",3,"name")="struct_jrec_pini.prefix.time"
	Set gtmtypes("struct_jrec_pini",3,"off")=8
	Set gtmtypes("struct_jrec_pini",3,"len")=4
	Set gtmtypes("struct_jrec_pini",3,"type")="unsigned-int"
	Set gtmtypfldindx("struct_jrec_pini","prefix.time")=3
	Set gtmtypes("struct_jrec_pini",4,"name")="struct_jrec_pini.prefix.checksum"
	Set gtmtypes("struct_jrec_pini",4,"off")=12
	Set gtmtypes("struct_jrec_pini",4,"len")=4
	Set gtmtypes("struct_jrec_pini",4,"type")="unsigned-int"
	Set gtmtypfldindx("struct_jrec_pini","prefix.checksum")=4
	Set gtmtypes("struct_jrec_pini",5,"name")="struct_jrec_pini.prefix.tn"
	Set gtmtypes("struct_jrec_pini",5,"off")=16
	Set gtmtypes("struct_jrec_pini",5,"len")=8
	Set gtmtypes("struct_jrec_pini",5,"type")="uint64_t"
	Set gtmtypfldindx("struct_jrec_pini","prefix.tn")=5
	Set gtmtypes("struct_jrec_pini",6,"name")="struct_jrec_pini.process_vector"
	Set gtmtypes("struct_jrec_pini",6,"off")=24
	Set gtmtypes("struct_jrec_pini",6,"len")=176
	Set gtmtypes("struct_jrec_pini",6,"type")="jnl_process_vector"
	Set gtmtypfldindx("struct_jrec_pini","process_vector")=6
	Set gtmtypes("struct_jrec_pini",6,"dim")=2
	Set gtmtypes("struct_jrec_pini",7,"name")="struct_jrec_pini.process_vector[0].jpv_pid"
	Set gtmtypes("struct_jrec_pini",7,"off")=24
	Set gtmtypes("struct_jrec_pini",7,"len")=4
	Set gtmtypes("struct_jrec_pini",7,"type")="unsigned-int"
	Set gtmtypfldindx("struct_jrec_pini","process_vector[0].jpv_pid")=7
	Set gtmtypes("struct_jrec_pini",8,"name")="struct_jrec_pini.process_vector[0].jpv_image_count"
	Set gtmtypes("struct_jrec_pini",8,"off")=28
	Set gtmtypes("struct_jrec_pini",8,"len")=4
	Set gtmtypes("struct_jrec_pini",8,"type")="int"
	Set gtmtypfldindx("struct_jrec_pini","process_vector[0].jpv_image_count")=8
	Set gtmtypes("struct_jrec_pini",9,"name")="struct_jrec_pini.process_vector[0].jpv_time"
	Set gtmtypes("struct_jrec_pini",9,"off")=32
	Set gtmtypes("struct_jrec_pini",9,"len")=8
	Set gtmtypes("struct_jrec_pini",9,"type")="int64_t"
	Set gtmtypfldindx("struct_jrec_pini","process_vector[0].jpv_time")=9
	Set gtmtypes("struct_jrec_pini",10,"name")="struct_jrec_pini.process_vector[0].jpv_login_time"
	Set gtmtypes("struct_jrec_pini",10,"off")=40
	Set gtmtypes("struct_jrec_pini",10,"len")=8
	Set gtmtypes("struct_jrec_pini",10,"type")="int64_t"
	Set gtmtypfldindx("struct_jrec_pini","process_vector[0].jpv_login_time")=10
	Set gtmtypes("struct_jrec_pini",11,"name")="struct_jrec_pini.process_vector[0].jpv_node"
	Set gtmtypes("struct_jrec_pini",11,"off")=48
	Set gtmtypes("struct_jrec_pini",11,"len")=16
	Set gtmtypes("struct_jrec_pini",11,"type")="char"
	Set gtmtypfldindx("struct_jrec_pini","process_vector[0].jpv_node")=11
	Set gtmtypes("struct_jrec_pini",12,"name")="struct_jrec_pini.process_vector[0].jpv_user"
	Set gtmtypes("struct_jrec_pini",12,"off")=64
	Set gtmtypes("struct_jrec_pini",12,"len")=12
	Set gtmtypes("struct_jrec_pini",12,"type")="char"
	Set gtmtypfldindx("struct_jrec_pini","process_vector[0].jpv_user")=12
	Set gtmtypes("struct_jrec_pini",13,"name")="struct_jrec_pini.process_vector[0].jpv_prcnam"
	Set gtmtypes("struct_jrec_pini",13,"off")=76
	Set gtmtypes("struct_jrec_pini",13,"len")=16
	Set gtmtypes("struct_jrec_pini",13,"type")="char"
	Set gtmtypfldindx("struct_jrec_pini","process_vector[0].jpv_prcnam")=13
	Set gtmtypes("struct_jrec_pini",14,"name")="struct_jrec_pini.process_vector[0].jpv_terminal"
	Set gtmtypes("struct_jrec_pini",14,"off")=92
	Set gtmtypes("struct_jrec_pini",14,"len")=15
	Set gtmtypes("struct_jrec_pini",14,"type")="char"
	Set gtmtypfldindx("struct_jrec_pini","process_vector[0].jpv_terminal")=14
	Set gtmtypes("struct_jrec_pini",15,"name")="struct_jrec_pini.process_vector[0].jpv_mode"
	Set gtmtypes("struct_jrec_pini",15,"off")=107
	Set gtmtypes("struct_jrec_pini",15,"len")=1
	Set gtmtypes("struct_jrec_pini",15,"type")="unsigned-char"
	Set gtmtypfldindx("struct_jrec_pini","process_vector[0].jpv_mode")=15
	Set gtmtypes("struct_jrec_pini",16,"name")="struct_jrec_pini.process_vector[0].filler"
	Set gtmtypes("struct_jrec_pini",16,"off")=108
	Set gtmtypes("struct_jrec_pini",16,"len")=4
	Set gtmtypes("struct_jrec_pini",16,"type")="int"
	Set gtmtypfldindx("struct_jrec_pini","process_vector[0].filler")=16
	Set gtmtypes("struct_jrec_pini",17,"name")="struct_jrec_pini.filler"
	Set gtmtypes("struct_jrec_pini",17,"off")=200
	Set gtmtypes("struct_jrec_pini",17,"len")=4
	Set gtmtypes("struct_jrec_pini",17,"type")="int"
	Set gtmtypfldindx("struct_jrec_pini","filler")=17
	Set gtmtypes("struct_jrec_pini",18,"name")="struct_jrec_pini.suffix"
	Set gtmtypes("struct_jrec_pini",18,"off")=204
	Set gtmtypes("struct_jrec_pini",18,"len")=4
	Set gtmtypes("struct_jrec_pini",18,"type")="jrec_suffix"
	Set gtmtypfldindx("struct_jrec_pini","suffix")=18
	;
	Set gtmtypes("struct_jrec_tcom")="struct"
	Set gtmtypes("struct_jrec_tcom",0)=13
	Set gtmtypes("struct_jrec_tcom","len")=56
	Set gtmtypes("struct_jrec_tcom",1,"name")="struct_jrec_tcom.prefix"
	Set gtmtypes("struct_jrec_tcom",1,"off")=0
	Set gtmtypes("struct_jrec_tcom",1,"len")=24
	Set gtmtypes("struct_jrec_tcom",1,"type")="jrec_prefix"
	Set gtmtypfldindx("struct_jrec_tcom","prefix")=1
	Set gtmtypes("struct_jrec_tcom",2,"name")="struct_jrec_tcom.prefix.pini_addr"
	Set gtmtypes("struct_jrec_tcom",2,"off")=4
	Set gtmtypes("struct_jrec_tcom",2,"len")=4
	Set gtmtypes("struct_jrec_tcom",2,"type")="unsigned-int"
	Set gtmtypfldindx("struct_jrec_tcom","prefix.pini_addr")=2
	Set gtmtypes("struct_jrec_tcom",3,"name")="struct_jrec_tcom.prefix.time"
	Set gtmtypes("struct_jrec_tcom",3,"off")=8
	Set gtmtypes("struct_jrec_tcom",3,"len")=4
	Set gtmtypes("struct_jrec_tcom",3,"type")="unsigned-int"
	Set gtmtypfldindx("struct_jrec_tcom","prefix.time")=3
	Set gtmtypes("struct_jrec_tcom",4,"name")="struct_jrec_tcom.prefix.checksum"
	Set gtmtypes("struct_jrec_tcom",4,"off")=12
	Set gtmtypes("struct_jrec_tcom",4,"len")=4
	Set gtmtypes("struct_jrec_tcom",4,"type")="unsigned-int"
	Set gtmtypfldindx("struct_jrec_tcom","prefix.checksum")=4
	Set gtmtypes("struct_jrec_tcom",5,"name")="struct_jrec_tcom.prefix.tn"
	Set gtmtypes("struct_jrec_tcom",5,"off")=16
	Set gtmtypes("struct_jrec_tcom",5,"len")=8
	Set gtmtypes("struct_jrec_tcom",5,"type")="uint64_t"
	Set gtmtypfldindx("struct_jrec_tcom","prefix.tn")=5
	Set gtmtypes("struct_jrec_tcom",6,"name")="struct_jrec_tcom.token_seq"
	Set gtmtypes("struct_jrec_tcom",6,"off")=24
	Set gtmtypes("struct_jrec_tcom",6,"len")=8
	Set gtmtypes("struct_jrec_tcom",6,"type")="token_seq_t"
	Set gtmtypfldindx("struct_jrec_tcom","token_seq")=6
	Set gtmtypes("struct_jrec_tcom",7,"name")="struct_jrec_tcom.token_seq.jnl_seqno"
	Set gtmtypes("struct_jrec_tcom",7,"off")=24
	Set gtmtypes("struct_jrec_tcom",7,"len")=8
	Set gtmtypes("struct_jrec_tcom",7,"type")="uint64_t"
	Set gtmtypfldindx("struct_jrec_tcom","token_seq.jnl_seqno")=7
	Set gtmtypes("struct_jrec_tcom",8,"name")="struct_jrec_tcom.token_seq.token"
	Set gtmtypes("struct_jrec_tcom",8,"off")=24
	Set gtmtypes("struct_jrec_tcom",8,"len")=8
	Set gtmtypes("struct_jrec_tcom",8,"type")="uint64_t"
	Set gtmtypfldindx("struct_jrec_tcom","token_seq.token")=8
	Set gtmtypes("struct_jrec_tcom",9,"name")="struct_jrec_tcom.strm_seqno"
	Set gtmtypes("struct_jrec_tcom",9,"off")=32
	Set gtmtypes("struct_jrec_tcom",9,"len")=8
	Set gtmtypes("struct_jrec_tcom",9,"type")="uint64_t"
	Set gtmtypfldindx("struct_jrec_tcom","strm_seqno")=9
	Set gtmtypes("struct_jrec_tcom",10,"name")="struct_jrec_tcom.filler_short"
	Set gtmtypes("struct_jrec_tcom",10,"off")=40
	Set gtmtypes("struct_jrec_tcom",10,"len")=2
	Set gtmtypes("struct_jrec_tcom",10,"type")="unsigned-short"
	Set gtmtypfldindx("struct_jrec_tcom","filler_short")=10
	Set gtmtypes("struct_jrec_tcom",11,"name")="struct_jrec_tcom.num_participants"
	Set gtmtypes("struct_jrec_tcom",11,"off")=42
	Set gtmtypes("struct_jrec_tcom",11,"len")=2
	Set gtmtypes("struct_jrec_tcom",11,"type")="unsigned-short"
	Set gtmtypfldindx("struct_jrec_tcom","num_participants")=11
	Set gtmtypes("struct_jrec_tcom",12,"name")="struct_jrec_tcom.jnl_tid"
	Set gtmtypes("struct_jrec_tcom",12,"off")=44
	Set gtmtypes("struct_jrec_tcom",12,"len")=8
	Set gtmtypes("struct_jrec_tcom",12,"type")="char"
	Set gtmtypfldindx("struct_jrec_tcom","jnl_tid")=12
	Set gtmtypes("struct_jrec_tcom",13,"name")="struct_jrec_tcom.suffix"
	Set gtmtypes("struct_jrec_tcom",13,"off")=52
	Set gtmtypes("struct_jrec_tcom",13,"len")=4
	Set gtmtypes("struct_jrec_tcom",13,"type")="jrec_suffix"
	Set gtmtypfldindx("struct_jrec_tcom","suffix")=13
	;
	Set gtmtypes("struct_jrec_trunc")="struct"
	Set gtmtypes("struct_jrec_trunc",0)=9
	Set gtmtypes("struct_jrec_trunc","len")=40
	Set gtmtypes("struct_jrec_trunc",1,"name")="struct_jrec_trunc.prefix"
	Set gtmtypes("struct_jrec_trunc",1,"off")=0
	Set gtmtypes("struct_jrec_trunc",1,"len")=24
	Set gtmtypes("struct_jrec_trunc",1,"type")="jrec_prefix"
	Set gtmtypfldindx("struct_jrec_trunc","prefix")=1
	Set gtmtypes("struct_jrec_trunc",2,"name")="struct_jrec_trunc.prefix.pini_addr"
	Set gtmtypes("struct_jrec_trunc",2,"off")=4
	Set gtmtypes("struct_jrec_trunc",2,"len")=4
	Set gtmtypes("struct_jrec_trunc",2,"type")="unsigned-int"
	Set gtmtypfldindx("struct_jrec_trunc","prefix.pini_addr")=2
	Set gtmtypes("struct_jrec_trunc",3,"name")="struct_jrec_trunc.prefix.time"
	Set gtmtypes("struct_jrec_trunc",3,"off")=8
	Set gtmtypes("struct_jrec_trunc",3,"len")=4
	Set gtmtypes("struct_jrec_trunc",3,"type")="unsigned-int"
	Set gtmtypfldindx("struct_jrec_trunc","prefix.time")=3
	Set gtmtypes("struct_jrec_trunc",4,"name")="struct_jrec_trunc.prefix.checksum"
	Set gtmtypes("struct_jrec_trunc",4,"off")=12
	Set gtmtypes("struct_jrec_trunc",4,"len")=4
	Set gtmtypes("struct_jrec_trunc",4,"type")="unsigned-int"
	Set gtmtypfldindx("struct_jrec_trunc","prefix.checksum")=4
	Set gtmtypes("struct_jrec_trunc",5,"name")="struct_jrec_trunc.prefix.tn"
	Set gtmtypes("struct_jrec_trunc",5,"off")=16
	Set gtmtypes("struct_jrec_trunc",5,"len")=8
	Set gtmtypes("struct_jrec_trunc",5,"type")="uint64_t"
	Set gtmtypfldindx("struct_jrec_trunc","prefix.tn")=5
	Set gtmtypes("struct_jrec_trunc",6,"name")="struct_jrec_trunc.orig_total_blks"
	Set gtmtypes("struct_jrec_trunc",6,"off")=24
	Set gtmtypes("struct_jrec_trunc",6,"len")=4
	Set gtmtypes("struct_jrec_trunc",6,"type")="unsigned-int"
	Set gtmtypfldindx("struct_jrec_trunc","orig_total_blks")=6
	Set gtmtypes("struct_jrec_trunc",7,"name")="struct_jrec_trunc.orig_free_blocks"
	Set gtmtypes("struct_jrec_trunc",7,"off")=28
	Set gtmtypes("struct_jrec_trunc",7,"len")=4
	Set gtmtypes("struct_jrec_trunc",7,"type")="unsigned-int"
	Set gtmtypfldindx("struct_jrec_trunc","orig_free_blocks")=7
	Set gtmtypes("struct_jrec_trunc",8,"name")="struct_jrec_trunc.total_blks_after_trunc"
	Set gtmtypes("struct_jrec_trunc",8,"off")=32
	Set gtmtypes("struct_jrec_trunc",8,"len")=4
	Set gtmtypes("struct_jrec_trunc",8,"type")="unsigned-int"
	Set gtmtypfldindx("struct_jrec_trunc","total_blks_after_trunc")=8
	Set gtmtypes("struct_jrec_trunc",9,"name")="struct_jrec_trunc.suffix"
	Set gtmtypes("struct_jrec_trunc",9,"off")=36
	Set gtmtypes("struct_jrec_trunc",9,"len")=4
	Set gtmtypes("struct_jrec_trunc",9,"type")="jrec_suffix"
	Set gtmtypfldindx("struct_jrec_trunc","suffix")=9
	;
	Set gtmtypes("struct_jrec_upd")="struct"
	Set gtmtypes("struct_jrec_upd",0)=14
	Set gtmtypes("struct_jrec_upd","len")=56
	Set gtmtypes("struct_jrec_upd",1,"name")="struct_jrec_upd.prefix"
	Set gtmtypes("struct_jrec_upd",1,"off")=0
	Set gtmtypes("struct_jrec_upd",1,"len")=24
	Set gtmtypes("struct_jrec_upd",1,"type")="jrec_prefix"
	Set gtmtypfldindx("struct_jrec_upd","prefix")=1
	Set gtmtypes("struct_jrec_upd",2,"name")="struct_jrec_upd.prefix.pini_addr"
	Set gtmtypes("struct_jrec_upd",2,"off")=4
	Set gtmtypes("struct_jrec_upd",2,"len")=4
	Set gtmtypes("struct_jrec_upd",2,"type")="unsigned-int"
	Set gtmtypfldindx("struct_jrec_upd","prefix.pini_addr")=2
	Set gtmtypes("struct_jrec_upd",3,"name")="struct_jrec_upd.prefix.time"
	Set gtmtypes("struct_jrec_upd",3,"off")=8
	Set gtmtypes("struct_jrec_upd",3,"len")=4
	Set gtmtypes("struct_jrec_upd",3,"type")="unsigned-int"
	Set gtmtypfldindx("struct_jrec_upd","prefix.time")=3
	Set gtmtypes("struct_jrec_upd",4,"name")="struct_jrec_upd.prefix.checksum"
	Set gtmtypes("struct_jrec_upd",4,"off")=12
	Set gtmtypes("struct_jrec_upd",4,"len")=4
	Set gtmtypes("struct_jrec_upd",4,"type")="unsigned-int"
	Set gtmtypfldindx("struct_jrec_upd","prefix.checksum")=4
	Set gtmtypes("struct_jrec_upd",5,"name")="struct_jrec_upd.prefix.tn"
	Set gtmtypes("struct_jrec_upd",5,"off")=16
	Set gtmtypes("struct_jrec_upd",5,"len")=8
	Set gtmtypes("struct_jrec_upd",5,"type")="uint64_t"
	Set gtmtypfldindx("struct_jrec_upd","prefix.tn")=5
	Set gtmtypes("struct_jrec_upd",6,"name")="struct_jrec_upd.token_seq"
	Set gtmtypes("struct_jrec_upd",6,"off")=24
	Set gtmtypes("struct_jrec_upd",6,"len")=8
	Set gtmtypes("struct_jrec_upd",6,"type")="token_seq_t"
	Set gtmtypfldindx("struct_jrec_upd","token_seq")=6
	Set gtmtypes("struct_jrec_upd",7,"name")="struct_jrec_upd.token_seq.jnl_seqno"
	Set gtmtypes("struct_jrec_upd",7,"off")=24
	Set gtmtypes("struct_jrec_upd",7,"len")=8
	Set gtmtypes("struct_jrec_upd",7,"type")="uint64_t"
	Set gtmtypfldindx("struct_jrec_upd","token_seq.jnl_seqno")=7
	Set gtmtypes("struct_jrec_upd",8,"name")="struct_jrec_upd.token_seq.token"
	Set gtmtypes("struct_jrec_upd",8,"off")=24
	Set gtmtypes("struct_jrec_upd",8,"len")=8
	Set gtmtypes("struct_jrec_upd",8,"type")="uint64_t"
	Set gtmtypfldindx("struct_jrec_upd","token_seq.token")=8
	Set gtmtypes("struct_jrec_upd",9,"name")="struct_jrec_upd.strm_seqno"
	Set gtmtypes("struct_jrec_upd",9,"off")=32
	Set gtmtypes("struct_jrec_upd",9,"len")=8
	Set gtmtypes("struct_jrec_upd",9,"type")="uint64_t"
	Set gtmtypfldindx("struct_jrec_upd","strm_seqno")=9
	Set gtmtypes("struct_jrec_upd",10,"name")="struct_jrec_upd.update_num"
	Set gtmtypes("struct_jrec_upd",10,"off")=40
	Set gtmtypes("struct_jrec_upd",10,"len")=4
	Set gtmtypes("struct_jrec_upd",10,"type")="unsigned-int"
	Set gtmtypfldindx("struct_jrec_upd","update_num")=10
	Set gtmtypes("struct_jrec_upd",11,"name")="struct_jrec_upd.filler_short"
	Set gtmtypes("struct_jrec_upd",11,"off")=44
	Set gtmtypes("struct_jrec_upd",11,"len")=2
	Set gtmtypes("struct_jrec_upd",11,"type")="unsigned-short"
	Set gtmtypfldindx("struct_jrec_upd","filler_short")=11
	Set gtmtypes("struct_jrec_upd",12,"name")="struct_jrec_upd.num_participants"
	Set gtmtypes("struct_jrec_upd",12,"off")=46
	Set gtmtypes("struct_jrec_upd",12,"len")=2
	Set gtmtypes("struct_jrec_upd",12,"type")="unsigned-short"
	Set gtmtypfldindx("struct_jrec_upd","num_participants")=12
	Set gtmtypes("struct_jrec_upd",13,"name")="struct_jrec_upd.mumps_node"
	Set gtmtypes("struct_jrec_upd",13,"off")=48
	Set gtmtypes("struct_jrec_upd",13,"len")=8
	Set gtmtypes("struct_jrec_upd",13,"type")="jnl_string"
	Set gtmtypfldindx("struct_jrec_upd","mumps_node")=13
	Set gtmtypes("struct_jrec_upd",14,"name")="struct_jrec_upd.mumps_node.text"
	Set gtmtypes("struct_jrec_upd",14,"off")=52
	Set gtmtypes("struct_jrec_upd",14,"len")=1
	Set gtmtypes("struct_jrec_upd",14,"type")="char"
	Set gtmtypfldindx("struct_jrec_upd","mumps_node.text")=14
	;
	Set gtmtypes("struct_jrec_ztcom")="struct"
	Set gtmtypes("struct_jrec_ztcom",0)=10
	Set gtmtypes("struct_jrec_ztcom","len")=48
	Set gtmtypes("struct_jrec_ztcom",1,"name")="struct_jrec_ztcom.prefix"
	Set gtmtypes("struct_jrec_ztcom",1,"off")=0
	Set gtmtypes("struct_jrec_ztcom",1,"len")=24
	Set gtmtypes("struct_jrec_ztcom",1,"type")="jrec_prefix"
	Set gtmtypfldindx("struct_jrec_ztcom","prefix")=1
	Set gtmtypes("struct_jrec_ztcom",2,"name")="struct_jrec_ztcom.prefix.pini_addr"
	Set gtmtypes("struct_jrec_ztcom",2,"off")=4
	Set gtmtypes("struct_jrec_ztcom",2,"len")=4
	Set gtmtypes("struct_jrec_ztcom",2,"type")="unsigned-int"
	Set gtmtypfldindx("struct_jrec_ztcom","prefix.pini_addr")=2
	Set gtmtypes("struct_jrec_ztcom",3,"name")="struct_jrec_ztcom.prefix.time"
	Set gtmtypes("struct_jrec_ztcom",3,"off")=8
	Set gtmtypes("struct_jrec_ztcom",3,"len")=4
	Set gtmtypes("struct_jrec_ztcom",3,"type")="unsigned-int"
	Set gtmtypfldindx("struct_jrec_ztcom","prefix.time")=3
	Set gtmtypes("struct_jrec_ztcom",4,"name")="struct_jrec_ztcom.prefix.checksum"
	Set gtmtypes("struct_jrec_ztcom",4,"off")=12
	Set gtmtypes("struct_jrec_ztcom",4,"len")=4
	Set gtmtypes("struct_jrec_ztcom",4,"type")="unsigned-int"
	Set gtmtypfldindx("struct_jrec_ztcom","prefix.checksum")=4
	Set gtmtypes("struct_jrec_ztcom",5,"name")="struct_jrec_ztcom.prefix.tn"
	Set gtmtypes("struct_jrec_ztcom",5,"off")=16
	Set gtmtypes("struct_jrec_ztcom",5,"len")=8
	Set gtmtypes("struct_jrec_ztcom",5,"type")="uint64_t"
	Set gtmtypfldindx("struct_jrec_ztcom","prefix.tn")=5
	Set gtmtypes("struct_jrec_ztcom",6,"name")="struct_jrec_ztcom.token"
	Set gtmtypes("struct_jrec_ztcom",6,"off")=24
	Set gtmtypes("struct_jrec_ztcom",6,"len")=8
	Set gtmtypes("struct_jrec_ztcom",6,"type")="uint64_t"
	Set gtmtypfldindx("struct_jrec_ztcom","token")=6
	Set gtmtypes("struct_jrec_ztcom",7,"name")="struct_jrec_ztcom.filler_8bytes"
	Set gtmtypes("struct_jrec_ztcom",7,"off")=32
	Set gtmtypes("struct_jrec_ztcom",7,"len")=8
	Set gtmtypes("struct_jrec_ztcom",7,"type")="uint64_t"
	Set gtmtypfldindx("struct_jrec_ztcom","filler_8bytes")=7
	Set gtmtypes("struct_jrec_ztcom",8,"name")="struct_jrec_ztcom.filler_short"
	Set gtmtypes("struct_jrec_ztcom",8,"off")=40
	Set gtmtypes("struct_jrec_ztcom",8,"len")=2
	Set gtmtypes("struct_jrec_ztcom",8,"type")="unsigned-short"
	Set gtmtypfldindx("struct_jrec_ztcom","filler_short")=8
	Set gtmtypes("struct_jrec_ztcom",9,"name")="struct_jrec_ztcom.participants"
	Set gtmtypes("struct_jrec_ztcom",9,"off")=42
	Set gtmtypes("struct_jrec_ztcom",9,"len")=2
	Set gtmtypes("struct_jrec_ztcom",9,"type")="unsigned-short"
	Set gtmtypfldindx("struct_jrec_ztcom","participants")=9
	Set gtmtypes("struct_jrec_ztcom",10,"name")="struct_jrec_ztcom.suffix"
	Set gtmtypes("struct_jrec_ztcom",10,"off")=44
	Set gtmtypes("struct_jrec_ztcom",10,"len")=4
	Set gtmtypes("struct_jrec_ztcom",10,"type")="jrec_suffix"
	Set gtmtypfldindx("struct_jrec_ztcom","suffix")=10
	;
	Set gtmtypes("struct_jrec_ztworm")="struct"
	Set gtmtypes("struct_jrec_ztworm",0)=14
	Set gtmtypes("struct_jrec_ztworm","len")=56
	Set gtmtypes("struct_jrec_ztworm",1,"name")="struct_jrec_ztworm.prefix"
	Set gtmtypes("struct_jrec_ztworm",1,"off")=0
	Set gtmtypes("struct_jrec_ztworm",1,"len")=24
	Set gtmtypes("struct_jrec_ztworm",1,"type")="jrec_prefix"
	Set gtmtypfldindx("struct_jrec_ztworm","prefix")=1
	Set gtmtypes("struct_jrec_ztworm",2,"name")="struct_jrec_ztworm.prefix.pini_addr"
	Set gtmtypes("struct_jrec_ztworm",2,"off")=4
	Set gtmtypes("struct_jrec_ztworm",2,"len")=4
	Set gtmtypes("struct_jrec_ztworm",2,"type")="unsigned-int"
	Set gtmtypfldindx("struct_jrec_ztworm","prefix.pini_addr")=2
	Set gtmtypes("struct_jrec_ztworm",3,"name")="struct_jrec_ztworm.prefix.time"
	Set gtmtypes("struct_jrec_ztworm",3,"off")=8
	Set gtmtypes("struct_jrec_ztworm",3,"len")=4
	Set gtmtypes("struct_jrec_ztworm",3,"type")="unsigned-int"
	Set gtmtypfldindx("struct_jrec_ztworm","prefix.time")=3
	Set gtmtypes("struct_jrec_ztworm",4,"name")="struct_jrec_ztworm.prefix.checksum"
	Set gtmtypes("struct_jrec_ztworm",4,"off")=12
	Set gtmtypes("struct_jrec_ztworm",4,"len")=4
	Set gtmtypes("struct_jrec_ztworm",4,"type")="unsigned-int"
	Set gtmtypfldindx("struct_jrec_ztworm","prefix.checksum")=4
	Set gtmtypes("struct_jrec_ztworm",5,"name")="struct_jrec_ztworm.prefix.tn"
	Set gtmtypes("struct_jrec_ztworm",5,"off")=16
	Set gtmtypes("struct_jrec_ztworm",5,"len")=8
	Set gtmtypes("struct_jrec_ztworm",5,"type")="uint64_t"
	Set gtmtypfldindx("struct_jrec_ztworm","prefix.tn")=5
	Set gtmtypes("struct_jrec_ztworm",6,"name")="struct_jrec_ztworm.token_seq"
	Set gtmtypes("struct_jrec_ztworm",6,"off")=24
	Set gtmtypes("struct_jrec_ztworm",6,"len")=8
	Set gtmtypes("struct_jrec_ztworm",6,"type")="token_seq_t"
	Set gtmtypfldindx("struct_jrec_ztworm","token_seq")=6
	Set gtmtypes("struct_jrec_ztworm",7,"name")="struct_jrec_ztworm.token_seq.jnl_seqno"
	Set gtmtypes("struct_jrec_ztworm",7,"off")=24
	Set gtmtypes("struct_jrec_ztworm",7,"len")=8
	Set gtmtypes("struct_jrec_ztworm",7,"type")="uint64_t"
	Set gtmtypfldindx("struct_jrec_ztworm","token_seq.jnl_seqno")=7
	Set gtmtypes("struct_jrec_ztworm",8,"name")="struct_jrec_ztworm.token_seq.token"
	Set gtmtypes("struct_jrec_ztworm",8,"off")=24
	Set gtmtypes("struct_jrec_ztworm",8,"len")=8
	Set gtmtypes("struct_jrec_ztworm",8,"type")="uint64_t"
	Set gtmtypfldindx("struct_jrec_ztworm","token_seq.token")=8
	Set gtmtypes("struct_jrec_ztworm",9,"name")="struct_jrec_ztworm.strm_seqno"
	Set gtmtypes("struct_jrec_ztworm",9,"off")=32
	Set gtmtypes("struct_jrec_ztworm",9,"len")=8
	Set gtmtypes("struct_jrec_ztworm",9,"type")="uint64_t"
	Set gtmtypfldindx("struct_jrec_ztworm","strm_seqno")=9
	Set gtmtypes("struct_jrec_ztworm",10,"name")="struct_jrec_ztworm.update_num"
	Set gtmtypes("struct_jrec_ztworm",10,"off")=40
	Set gtmtypes("struct_jrec_ztworm",10,"len")=4
	Set gtmtypes("struct_jrec_ztworm",10,"type")="unsigned-int"
	Set gtmtypfldindx("struct_jrec_ztworm","update_num")=10
	Set gtmtypes("struct_jrec_ztworm",11,"name")="struct_jrec_ztworm.filler_short"
	Set gtmtypes("struct_jrec_ztworm",11,"off")=44
	Set gtmtypes("struct_jrec_ztworm",11,"len")=2
	Set gtmtypes("struct_jrec_ztworm",11,"type")="unsigned-short"
	Set gtmtypfldindx("struct_jrec_ztworm","filler_short")=11
	Set gtmtypes("struct_jrec_ztworm",12,"name")="struct_jrec_ztworm.num_participants"
	Set gtmtypes("struct_jrec_ztworm",12,"off")=46
	Set gtmtypes("struct_jrec_ztworm",12,"len")=2
	Set gtmtypes("struct_jrec_ztworm",12,"type")="unsigned-short"
	Set gtmtypfldindx("struct_jrec_ztworm","num_participants")=12
	Set gtmtypes("struct_jrec_ztworm",13,"name")="struct_jrec_ztworm.ztworm_str"
	Set gtmtypes("struct_jrec_ztworm",13,"off")=48
	Set gtmtypes("struct_jrec_ztworm",13,"len")=8
	Set gtmtypes("struct_jrec_ztworm",13,"type")="jnl_string"
	Set gtmtypfldindx("struct_jrec_ztworm","ztworm_str")=13
	Set gtmtypes("struct_jrec_ztworm",14,"name")="struct_jrec_ztworm.ztworm_str.text"
	Set gtmtypes("struct_jrec_ztworm",14,"off")=52
	Set gtmtypes("struct_jrec_ztworm",14,"len")=1
	Set gtmtypes("struct_jrec_ztworm",14,"type")="char"
	Set gtmtypfldindx("struct_jrec_ztworm","ztworm_str.text")=14
	;
	Set gtmtypes("super_srch_hist")="struct"
	Set gtmtypes("super_srch_hist",0)=18
	Set gtmtypes("super_srch_hist","len")=1160
	Set gtmtypes("super_srch_hist",1,"name")="super_srch_hist.depth"
	Set gtmtypes("super_srch_hist",1,"off")=0
	Set gtmtypes("super_srch_hist",1,"len")=4
	Set gtmtypes("super_srch_hist",1,"type")="int"
	Set gtmtypfldindx("super_srch_hist","depth")=1
	Set gtmtypes("super_srch_hist",2,"name")="super_srch_hist.filler"
	Set gtmtypes("super_srch_hist",2,"off")=4
	Set gtmtypes("super_srch_hist",2,"len")=4
	Set gtmtypes("super_srch_hist",2,"type")="int"
	Set gtmtypfldindx("super_srch_hist","filler")=2
	Set gtmtypes("super_srch_hist",3,"name")="super_srch_hist.h"
	Set gtmtypes("super_srch_hist",3,"off")=8
	Set gtmtypes("super_srch_hist",3,"len")=1152
	Set gtmtypes("super_srch_hist",3,"type")="srch_blk_status"
	Set gtmtypfldindx("super_srch_hist","h")=3
	Set gtmtypes("super_srch_hist",3,"dim")=16
	Set gtmtypes("super_srch_hist",4,"name")="super_srch_hist.h[0].cr"
	Set gtmtypes("super_srch_hist",4,"off")=8
	Set gtmtypes("super_srch_hist",4,"len")=8
	Set gtmtypes("super_srch_hist",4,"type")="addr"
	Set gtmtypfldindx("super_srch_hist","h[0].cr")=4
	Set gtmtypes("super_srch_hist",5,"name")="super_srch_hist.h[0].buffaddr"
	Set gtmtypes("super_srch_hist",5,"off")=16
	Set gtmtypes("super_srch_hist",5,"len")=8
	Set gtmtypes("super_srch_hist",5,"type")="addr"
	Set gtmtypfldindx("super_srch_hist","h[0].buffaddr")=5
	Set gtmtypes("super_srch_hist",6,"name")="super_srch_hist.h[0].blk_num"
	Set gtmtypes("super_srch_hist",6,"off")=24
	Set gtmtypes("super_srch_hist",6,"len")=4
	Set gtmtypes("super_srch_hist",6,"type")="int"
	Set gtmtypfldindx("super_srch_hist","h[0].blk_num")=6
	Set gtmtypes("super_srch_hist",7,"name")="super_srch_hist.h[0].tn"
	Set gtmtypes("super_srch_hist",7,"off")=32
	Set gtmtypes("super_srch_hist",7,"len")=8
	Set gtmtypes("super_srch_hist",7,"type")="uint64_t"
	Set gtmtypfldindx("super_srch_hist","h[0].tn")=7
	Set gtmtypes("super_srch_hist",8,"name")="super_srch_hist.h[0].prev_rec"
	Set gtmtypes("super_srch_hist",8,"off")=40
	Set gtmtypes("super_srch_hist",8,"len")=4
	Set gtmtypes("super_srch_hist",8,"type")="srch_rec_status"
	Set gtmtypfldindx("super_srch_hist","h[0].prev_rec")=8
	Set gtmtypes("super_srch_hist",9,"name")="super_srch_hist.h[0].prev_rec.offset"
	Set gtmtypes("super_srch_hist",9,"off")=40
	Set gtmtypes("super_srch_hist",9,"len")=2
	Set gtmtypes("super_srch_hist",9,"type")="unsigned-short"
	Set gtmtypfldindx("super_srch_hist","h[0].prev_rec.offset")=9
	Set gtmtypes("super_srch_hist",10,"name")="super_srch_hist.h[0].prev_rec.match"
	Set gtmtypes("super_srch_hist",10,"off")=42
	Set gtmtypes("super_srch_hist",10,"len")=2
	Set gtmtypes("super_srch_hist",10,"type")="unsigned-short"
	Set gtmtypfldindx("super_srch_hist","h[0].prev_rec.match")=10
	Set gtmtypes("super_srch_hist",11,"name")="super_srch_hist.h[0].curr_rec"
	Set gtmtypes("super_srch_hist",11,"off")=44
	Set gtmtypes("super_srch_hist",11,"len")=4
	Set gtmtypes("super_srch_hist",11,"type")="srch_rec_status"
	Set gtmtypfldindx("super_srch_hist","h[0].curr_rec")=11
	Set gtmtypes("super_srch_hist",12,"name")="super_srch_hist.h[0].curr_rec.offset"
	Set gtmtypes("super_srch_hist",12,"off")=44
	Set gtmtypes("super_srch_hist",12,"len")=2
	Set gtmtypes("super_srch_hist",12,"type")="unsigned-short"
	Set gtmtypfldindx("super_srch_hist","h[0].curr_rec.offset")=12
	Set gtmtypes("super_srch_hist",13,"name")="super_srch_hist.h[0].curr_rec.match"
	Set gtmtypes("super_srch_hist",13,"off")=46
	Set gtmtypes("super_srch_hist",13,"len")=2
	Set gtmtypes("super_srch_hist",13,"type")="unsigned-short"
	Set gtmtypfldindx("super_srch_hist","h[0].curr_rec.match")=13
	Set gtmtypes("super_srch_hist",14,"name")="super_srch_hist.h[0].cycle"
	Set gtmtypes("super_srch_hist",14,"off")=48
	Set gtmtypes("super_srch_hist",14,"len")=4
	Set gtmtypes("super_srch_hist",14,"type")="int"
	Set gtmtypfldindx("super_srch_hist","h[0].cycle")=14
	Set gtmtypes("super_srch_hist",15,"name")="super_srch_hist.h[0].level"
	Set gtmtypes("super_srch_hist",15,"off")=52
	Set gtmtypes("super_srch_hist",15,"len")=4
	Set gtmtypes("super_srch_hist",15,"type")="int"
	Set gtmtypfldindx("super_srch_hist","h[0].level")=15
	Set gtmtypes("super_srch_hist",16,"name")="super_srch_hist.h[0].cse"
	Set gtmtypes("super_srch_hist",16,"off")=56
	Set gtmtypes("super_srch_hist",16,"len")=8
	Set gtmtypes("super_srch_hist",16,"type")="addr"
	Set gtmtypfldindx("super_srch_hist","h[0].cse")=16
	Set gtmtypes("super_srch_hist",17,"name")="super_srch_hist.h[0].first_tp_srch_status"
	Set gtmtypes("super_srch_hist",17,"off")=64
	Set gtmtypes("super_srch_hist",17,"len")=8
	Set gtmtypes("super_srch_hist",17,"type")="addr"
	Set gtmtypfldindx("super_srch_hist","h[0].first_tp_srch_status")=17
	Set gtmtypes("super_srch_hist",18,"name")="super_srch_hist.h[0].blk_target"
	Set gtmtypes("super_srch_hist",18,"off")=72
	Set gtmtypes("super_srch_hist",18,"len")=8
	Set gtmtypes("super_srch_hist",18,"type")="addr"
	Set gtmtypfldindx("super_srch_hist","h[0].blk_target")=18
	;
	Set gtmtypes("svn_data_type")="struct"
	Set gtmtypes("svn_data_type",0)=3
	Set gtmtypes("svn_data_type","len")=8
	Set gtmtypes("svn_data_type",1,"name")="svn_data_type.opcode"
	Set gtmtypes("svn_data_type",1,"off")=0
	Set gtmtypes("svn_data_type",1,"len")=4
	Set gtmtypes("svn_data_type",1,"type")="unsigned-int"
	Set gtmtypfldindx("svn_data_type","opcode")=1
	Set gtmtypes("svn_data_type",2,"name")="svn_data_type.can_set"
	Set gtmtypes("svn_data_type",2,"off")=4
	Set gtmtypes("svn_data_type",2,"len")=1
	Set gtmtypes("svn_data_type",2,"type")="char"
	Set gtmtypfldindx("svn_data_type","can_set")=2
	Set gtmtypes("svn_data_type",3,"name")="svn_data_type.os_syst"
	Set gtmtypes("svn_data_type",3,"off")=5
	Set gtmtypes("svn_data_type",3,"len")=1
	Set gtmtypes("svn_data_type",3,"type")="char"
	Set gtmtypfldindx("svn_data_type","os_syst")=3
	;
	Set gtmtypes("symval")="struct"
	Set gtmtypes("symval",0)=32
	Set gtmtypes("symval","len")=176
	Set gtmtypes("symval",1,"name")="symval.ident"
	Set gtmtypes("symval",1,"off")=0
	Set gtmtypes("symval",1,"len")=2
	Set gtmtypes("symval",1,"type")="unsigned-short"
	Set gtmtypfldindx("symval","ident")=1
	Set gtmtypes("symval",2,"name")="symval.sbs_depth"
	Set gtmtypes("symval",2,"off")=2
	Set gtmtypes("symval",2,"len")=2
	Set gtmtypes("symval",2,"type")="unsigned-short"
	Set gtmtypfldindx("symval","sbs_depth")=2
	Set gtmtypes("symval",3,"name")="symval.tp_save_all"
	Set gtmtypes("symval",3,"off")=4
	Set gtmtypes("symval",3,"len")=4
	Set gtmtypes("symval",3,"type")="boolean_t"
	Set gtmtypfldindx("symval","tp_save_all")=3
	Set gtmtypes("symval",4,"name")="symval.xnew_var_list"
	Set gtmtypes("symval",4,"off")=8
	Set gtmtypes("symval",4,"len")=8
	Set gtmtypes("symval",4,"type")="addr"
	Set gtmtypfldindx("symval","xnew_var_list")=4
	Set gtmtypes("symval",5,"name")="symval.xnew_ref_list"
	Set gtmtypes("symval",5,"off")=16
	Set gtmtypes("symval",5,"len")=8
	Set gtmtypes("symval",5,"type")="addr"
	Set gtmtypfldindx("symval","xnew_ref_list")=5
	Set gtmtypes("symval",6,"name")="symval.h_symtab"
	Set gtmtypes("symval",6,"off")=24
	Set gtmtypes("symval",6,"len")=80
	Set gtmtypes("symval",6,"type")="hash_table_mname"
	Set gtmtypfldindx("symval","h_symtab")=6
	Set gtmtypes("symval",7,"name")="symval.h_symtab.base"
	Set gtmtypes("symval",7,"off")=24
	Set gtmtypes("symval",7,"len")=8
	Set gtmtypes("symval",7,"type")="addr"
	Set gtmtypfldindx("symval","h_symtab.base")=7
	Set gtmtypes("symval",8,"name")="symval.h_symtab.top"
	Set gtmtypes("symval",8,"off")=32
	Set gtmtypes("symval",8,"len")=8
	Set gtmtypes("symval",8,"type")="addr"
	Set gtmtypfldindx("symval","h_symtab.top")=8
	Set gtmtypes("symval",9,"name")="symval.h_symtab.spare_base"
	Set gtmtypes("symval",9,"off")=40
	Set gtmtypes("symval",9,"len")=8
	Set gtmtypes("symval",9,"type")="addr"
	Set gtmtypfldindx("symval","h_symtab.spare_base")=9
	Set gtmtypes("symval",10,"name")="symval.h_symtab.entry_passed_thru"
	Set gtmtypes("symval",10,"off")=48
	Set gtmtypes("symval",10,"len")=8
	Set gtmtypes("symval",10,"type")="addr"
	Set gtmtypfldindx("symval","h_symtab.entry_passed_thru")=10
	Set gtmtypes("symval",11,"name")="symval.h_symtab.size"
	Set gtmtypes("symval",11,"off")=56
	Set gtmtypes("symval",11,"len")=4
	Set gtmtypes("symval",11,"type")="unsigned-int"
	Set gtmtypfldindx("symval","h_symtab.size")=11
	Set gtmtypes("symval",12,"name")="symval.h_symtab.initial_size"
	Set gtmtypes("symval",12,"off")=60
	Set gtmtypes("symval",12,"len")=4
	Set gtmtypes("symval",12,"type")="unsigned-int"
	Set gtmtypfldindx("symval","h_symtab.initial_size")=12
	Set gtmtypes("symval",13,"name")="symval.h_symtab.spare_base_size"
	Set gtmtypes("symval",13,"off")=64
	Set gtmtypes("symval",13,"len")=4
	Set gtmtypes("symval",13,"type")="unsigned-int"
	Set gtmtypfldindx("symval","h_symtab.spare_base_size")=13
	Set gtmtypes("symval",14,"name")="symval.h_symtab.count"
	Set gtmtypes("symval",14,"off")=68
	Set gtmtypes("symval",14,"len")=4
	Set gtmtypes("symval",14,"type")="unsigned-int"
	Set gtmtypfldindx("symval","h_symtab.count")=14
	Set gtmtypes("symval",15,"name")="symval.h_symtab.del_count"
	Set gtmtypes("symval",15,"off")=72
	Set gtmtypes("symval",15,"len")=4
	Set gtmtypes("symval",15,"type")="unsigned-int"
	Set gtmtypfldindx("symval","h_symtab.del_count")=15
	Set gtmtypes("symval",16,"name")="symval.h_symtab.exp_trigger_size"
	Set gtmtypes("symval",16,"off")=76
	Set gtmtypes("symval",16,"len")=4
	Set gtmtypes("symval",16,"type")="unsigned-int"
	Set gtmtypfldindx("symval","h_symtab.exp_trigger_size")=16
	Set gtmtypes("symval",17,"name")="symval.h_symtab.cmp_trigger_size"
	Set gtmtypes("symval",17,"off")=80
	Set gtmtypes("symval",17,"len")=4
	Set gtmtypes("symval",17,"type")="unsigned-int"
	Set gtmtypfldindx("symval","h_symtab.cmp_trigger_size")=17
	Set gtmtypes("symval",18,"name")="symval.h_symtab.dont_compact"
	Set gtmtypes("symval",18,"off")=84
	Set gtmtypes("symval",18,"len")=4
	Set gtmtypes("symval",18,"type")="boolean_t"
	Set gtmtypfldindx("symval","h_symtab.dont_compact")=18
	Set gtmtypes("symval",19,"name")="symval.h_symtab.dont_keep_spare_table"
	Set gtmtypes("symval",19,"off")=88
	Set gtmtypes("symval",19,"len")=4
	Set gtmtypes("symval",19,"type")="boolean_t"
	Set gtmtypfldindx("symval","h_symtab.dont_keep_spare_table")=19
	Set gtmtypes("symval",20,"name")="symval.h_symtab.defer_base_release"
	Set gtmtypes("symval",20,"off")=92
	Set gtmtypes("symval",20,"len")=4
	Set gtmtypes("symval",20,"type")="boolean_t"
	Set gtmtypfldindx("symval","h_symtab.defer_base_release")=20
	Set gtmtypes("symval",21,"name")="symval.h_symtab.active"
	Set gtmtypes("symval",21,"off")=96
	Set gtmtypes("symval",21,"len")=4
	Set gtmtypes("symval",21,"type")="boolean_t"
	Set gtmtypfldindx("symval","h_symtab.active")=21
	Set gtmtypes("symval",22,"name")="symval.lv_first_block"
	Set gtmtypes("symval",22,"off")=104
	Set gtmtypes("symval",22,"len")=8
	Set gtmtypes("symval",22,"type")="addr"
	Set gtmtypfldindx("symval","lv_first_block")=22
	Set gtmtypes("symval",23,"name")="symval.lvtree_first_block"
	Set gtmtypes("symval",23,"off")=112
	Set gtmtypes("symval",23,"len")=8
	Set gtmtypes("symval",23,"type")="addr"
	Set gtmtypfldindx("symval","lvtree_first_block")=23
	Set gtmtypes("symval",24,"name")="symval.lvtreenode_first_block"
	Set gtmtypes("symval",24,"off")=120
	Set gtmtypes("symval",24,"len")=8
	Set gtmtypes("symval",24,"type")="addr"
	Set gtmtypfldindx("symval","lvtreenode_first_block")=24
	Set gtmtypes("symval",25,"name")="symval.lv_flist"
	Set gtmtypes("symval",25,"off")=128
	Set gtmtypes("symval",25,"len")=8
	Set gtmtypes("symval",25,"type")="addr"
	Set gtmtypfldindx("symval","lv_flist")=25
	Set gtmtypes("symval",26,"name")="symval.lvtree_flist"
	Set gtmtypes("symval",26,"off")=136
	Set gtmtypes("symval",26,"len")=8
	Set gtmtypes("symval",26,"type")="addr"
	Set gtmtypfldindx("symval","lvtree_flist")=26
	Set gtmtypes("symval",27,"name")="symval.lvtreenode_flist"
	Set gtmtypes("symval",27,"off")=144
	Set gtmtypes("symval",27,"len")=8
	Set gtmtypes("symval",27,"type")="addr"
	Set gtmtypfldindx("symval","lvtreenode_flist")=27
	Set gtmtypes("symval",28,"name")="symval.last_tab"
	Set gtmtypes("symval",28,"off")=152
	Set gtmtypes("symval",28,"len")=8
	Set gtmtypes("symval",28,"type")="addr"
	Set gtmtypfldindx("symval","last_tab")=28
	Set gtmtypes("symval",29,"name")="symval.symvlvl"
	Set gtmtypes("symval",29,"off")=160
	Set gtmtypes("symval",29,"len")=4
	Set gtmtypes("symval",29,"type")="int"
	Set gtmtypfldindx("symval","symvlvl")=29
	Set gtmtypes("symval",30,"name")="symval.trigr_symval"
	Set gtmtypes("symval",30,"off")=164
	Set gtmtypes("symval",30,"len")=4
	Set gtmtypes("symval",30,"type")="boolean_t"
	Set gtmtypfldindx("symval","trigr_symval")=30
	Set gtmtypes("symval",31,"name")="symval.alias_activity"
	Set gtmtypes("symval",31,"off")=168
	Set gtmtypes("symval",31,"len")=4
	Set gtmtypes("symval",31,"type")="boolean_t"
	Set gtmtypfldindx("symval","alias_activity")=31
	Set gtmtypes("symval",32,"name")="symval.filler"
	Set gtmtypes("symval",32,"off")=172
	Set gtmtypes("symval",32,"len")=4
	Set gtmtypes("symval",32,"type")="int"
	Set gtmtypfldindx("symval","filler")=32
	;
	Set gtmtypes("tbp")="struct"
	Set gtmtypes("tbp",0)=4
	Set gtmtypes("tbp","len")=24
	Set gtmtypes("tbp",1,"name")="tbp.que"
	Set gtmtypes("tbp",1,"off")=0
	Set gtmtypes("tbp",1,"len")=16
	Set gtmtypes("tbp",1,"type")="struct"
	Set gtmtypfldindx("tbp","que")=1
	Set gtmtypes("tbp",2,"name")="tbp.que.fl"
	Set gtmtypes("tbp",2,"off")=0
	Set gtmtypes("tbp",2,"len")=8
	Set gtmtypes("tbp",2,"type")="addr"
	Set gtmtypfldindx("tbp","que.fl")=2
	Set gtmtypes("tbp",3,"name")="tbp.que.bl"
	Set gtmtypes("tbp",3,"off")=8
	Set gtmtypes("tbp",3,"len")=8
	Set gtmtypes("tbp",3,"type")="addr"
	Set gtmtypfldindx("tbp","que.bl")=3
	Set gtmtypes("tbp",4,"name")="tbp.bpt"
	Set gtmtypes("tbp",4,"off")=16
	Set gtmtypes("tbp",4,"len")=8
	Set gtmtypes("tbp",4,"type")="addr"
	Set gtmtypfldindx("tbp","bpt")=4
	;
	Set gtmtypes("template_struct")="struct"
	Set gtmtypes("template_struct",0)=5
	Set gtmtypes("template_struct","len")=4104
	Set gtmtypes("template_struct",1,"name")="template_struct.n_subs"
	Set gtmtypes("template_struct",1,"off")=0
	Set gtmtypes("template_struct",1,"len")=2
	Set gtmtypes("template_struct",1,"type")="unsigned-short"
	Set gtmtypfldindx("template_struct","n_subs")=1
	Set gtmtypes("template_struct",2,"name")="template_struct.sub"
	Set gtmtypes("template_struct",2,"off")=8
	Set gtmtypes("template_struct",2,"len")=4096
	Set gtmtypes("template_struct",2,"type")="mstr"
	Set gtmtypfldindx("template_struct","sub")=2
	Set gtmtypes("template_struct",2,"dim")=256
	Set gtmtypes("template_struct",3,"name")="template_struct.sub[0].char_len"
	Set gtmtypes("template_struct",3,"off")=8
	Set gtmtypes("template_struct",3,"len")=4
	Set gtmtypes("template_struct",3,"type")="unsigned-int"
	Set gtmtypfldindx("template_struct","sub[0].char_len")=3
	Set gtmtypes("template_struct",4,"name")="template_struct.sub[0].len"
	Set gtmtypes("template_struct",4,"off")=12
	Set gtmtypes("template_struct",4,"len")=4
	Set gtmtypes("template_struct",4,"type")="int"
	Set gtmtypfldindx("template_struct","sub[0].len")=4
	Set gtmtypes("template_struct",5,"name")="template_struct.sub[0].addr"
	Set gtmtypes("template_struct",5,"off")=16
	Set gtmtypes("template_struct",5,"len")=8
	Set gtmtypes("template_struct",5,"type")="addr"
	Set gtmtypfldindx("template_struct","sub[0].addr")=5
	;
	Set gtmtypes("testpt_struct")="struct"
	Set gtmtypes("testpt_struct",0)=1
	Set gtmtypes("testpt_struct","len")=1
	Set gtmtypes("testpt_struct",1,"name")="testpt_struct.wc_recover"
	Set gtmtypes("testpt_struct",1,"off")=0
	Set gtmtypes("testpt_struct",1,"len")=1
	Set gtmtypes("testpt_struct",1,"type")="char"
	Set gtmtypfldindx("testpt_struct","wc_recover")=1
	;
	Set gtmtypes("textElem")="struct"
	Set gtmtypes("textElem",0)=9
	Set gtmtypes("textElem","len")=32
	Set gtmtypes("textElem",1,"name")="textElem.queueIndex"
	Set gtmtypes("textElem",1,"off")=0
	Set gtmtypes("textElem",1,"len")=4
	Set gtmtypes("textElem",1,"type")="int"
	Set gtmtypfldindx("textElem","queueIndex")=1
	Set gtmtypes("textElem",2,"name")="textElem.state"
	Set gtmtypes("textElem",2,"off")=4
	Set gtmtypes("textElem",2,"len")=4
	Set gtmtypes("textElem",2,"type")="int"
	Set gtmtypfldindx("textElem","state")=2
	Set gtmtypes("textElem",3,"name")="textElem.realLen"
	Set gtmtypes("textElem",3,"off")=8
	Set gtmtypes("textElem",3,"len")=4
	Set gtmtypes("textElem",3,"type")="unsigned-int"
	Set gtmtypfldindx("textElem","realLen")=3
	Set gtmtypes("textElem",4,"name")="textElem.filler"
	Set gtmtypes("textElem",4,"off")=12
	Set gtmtypes("textElem",4,"len")=4
	Set gtmtypes("textElem",4,"type")="int"
	Set gtmtypfldindx("textElem","filler")=4
	Set gtmtypes("textElem",5,"name")="textElem.userStorage"
	Set gtmtypes("textElem",5,"off")=16
	Set gtmtypes("textElem",5,"len")=16
	Set gtmtypes("textElem",5,"type")="union"
	Set gtmtypfldindx("textElem","userStorage")=5
	Set gtmtypes("textElem",6,"name")="textElem.userStorage.links"
	Set gtmtypes("textElem",6,"off")=16
	Set gtmtypes("textElem",6,"len")=16
	Set gtmtypes("textElem",6,"type")="struct"
	Set gtmtypfldindx("textElem","userStorage.links")=6
	Set gtmtypes("textElem",7,"name")="textElem.userStorage.links.fPtr"
	Set gtmtypes("textElem",7,"off")=16
	Set gtmtypes("textElem",7,"len")=8
	Set gtmtypes("textElem",7,"type")="addr"
	Set gtmtypfldindx("textElem","userStorage.links.fPtr")=7
	Set gtmtypes("textElem",8,"name")="textElem.userStorage.links.bPtr"
	Set gtmtypes("textElem",8,"off")=24
	Set gtmtypes("textElem",8,"len")=8
	Set gtmtypes("textElem",8,"type")="addr"
	Set gtmtypfldindx("textElem","userStorage.links.bPtr")=8
	Set gtmtypes("textElem",9,"name")="textElem.userStorage.userStart"
	Set gtmtypes("textElem",9,"off")=16
	Set gtmtypes("textElem",9,"len")=1
	Set gtmtypes("textElem",9,"type")="unsigned-char"
	Set gtmtypfldindx("textElem","userStorage.userStart")=9
	;
	Set gtmtypes("th_index")="struct"
	Set gtmtypes("th_index",0)=9
	Set gtmtypes("th_index","len")=56
	Set gtmtypes("th_index",1,"name")="th_index.curr_tn"
	Set gtmtypes("th_index",1,"off")=0
	Set gtmtypes("th_index",1,"len")=8
	Set gtmtypes("th_index",1,"type")="uint64_t"
	Set gtmtypfldindx("th_index","curr_tn")=1
	Set gtmtypes("th_index",2,"name")="th_index.early_tn"
	Set gtmtypes("th_index",2,"off")=8
	Set gtmtypes("th_index",2,"len")=8
	Set gtmtypes("th_index",2,"type")="uint64_t"
	Set gtmtypfldindx("th_index","early_tn")=2
	Set gtmtypes("th_index",3,"name")="th_index.last_mm_sync"
	Set gtmtypes("th_index",3,"off")=16
	Set gtmtypes("th_index",3,"len")=8
	Set gtmtypes("th_index",3,"type")="uint64_t"
	Set gtmtypfldindx("th_index","last_mm_sync")=3
	Set gtmtypes("th_index",4,"name")="th_index.filler_8byte"
	Set gtmtypes("th_index",4,"off")=24
	Set gtmtypes("th_index",4,"len")=8
	Set gtmtypes("th_index",4,"type")="char"
	Set gtmtypfldindx("th_index","filler_8byte")=4
	Set gtmtypes("th_index",5,"name")="th_index.mm_tn"
	Set gtmtypes("th_index",5,"off")=32
	Set gtmtypes("th_index",5,"len")=8
	Set gtmtypes("th_index",5,"type")="uint64_t"
	Set gtmtypfldindx("th_index","mm_tn")=5
	Set gtmtypes("th_index",6,"name")="th_index.lock_sequence"
	Set gtmtypes("th_index",6,"off")=40
	Set gtmtypes("th_index",6,"len")=4
	Set gtmtypes("th_index",6,"type")="unsigned-int"
	Set gtmtypfldindx("th_index","lock_sequence")=6
	Set gtmtypes("th_index",7,"name")="th_index.ccp_jnl_filesize"
	Set gtmtypes("th_index",7,"off")=44
	Set gtmtypes("th_index",7,"len")=4
	Set gtmtypes("th_index",7,"type")="unsigned-int"
	Set gtmtypfldindx("th_index","ccp_jnl_filesize")=7
	Set gtmtypes("th_index",8,"name")="th_index.total_blks"
	Set gtmtypes("th_index",8,"off")=48
	Set gtmtypes("th_index",8,"len")=4
	Set gtmtypes("th_index",8,"type")="unsigned-int"
	Set gtmtypfldindx("th_index","total_blks")=8
	Set gtmtypes("th_index",9,"name")="th_index.free_blocks"
	Set gtmtypes("th_index",9,"off")=52
	Set gtmtypes("th_index",9,"len")=4
	Set gtmtypes("th_index",9,"type")="unsigned-int"
	Set gtmtypfldindx("th_index","free_blocks")=9
	;
	Set gtmtypes("th_rec")="struct"
	Set gtmtypes("th_rec",0)=10
	Set gtmtypes("th_rec","len")=48
	Set gtmtypes("th_rec",1,"name")="th_rec.tnque"
	Set gtmtypes("th_rec",1,"off")=0
	Set gtmtypes("th_rec",1,"len")=16
	Set gtmtypes("th_rec",1,"type")="struct"
	Set gtmtypfldindx("th_rec","tnque")=1
	Set gtmtypes("th_rec",2,"name")="th_rec.tnque.fl"
	Set gtmtypes("th_rec",2,"off")=0
	Set gtmtypes("th_rec",2,"len")=8
	Set gtmtypes("th_rec",2,"type")="intptr_t"
	Set gtmtypfldindx("th_rec","tnque.fl")=2
	Set gtmtypes("th_rec",3,"name")="th_rec.tnque.bl"
	Set gtmtypes("th_rec",3,"off")=8
	Set gtmtypes("th_rec",3,"len")=8
	Set gtmtypes("th_rec",3,"type")="intptr_t"
	Set gtmtypfldindx("th_rec","tnque.bl")=3
	Set gtmtypes("th_rec",4,"name")="th_rec.tn"
	Set gtmtypes("th_rec",4,"off")=16
	Set gtmtypes("th_rec",4,"len")=8
	Set gtmtypes("th_rec",4,"type")="uint64_t"
	Set gtmtypfldindx("th_rec","tn")=4
	Set gtmtypes("th_rec",5,"name")="th_rec.killtn"
	Set gtmtypes("th_rec",5,"off")=24
	Set gtmtypes("th_rec",5,"len")=8
	Set gtmtypes("th_rec",5,"type")="uint64_t"
	Set gtmtypfldindx("th_rec","killtn")=5
	Set gtmtypes("th_rec",6,"name")="th_rec.blk"
	Set gtmtypes("th_rec",6,"off")=32
	Set gtmtypes("th_rec",6,"len")=4
	Set gtmtypes("th_rec",6,"type")="int"
	Set gtmtypfldindx("th_rec","blk")=6
	Set gtmtypes("th_rec",7,"name")="th_rec.cache_index"
	Set gtmtypes("th_rec",7,"off")=36
	Set gtmtypes("th_rec",7,"len")=4
	Set gtmtypes("th_rec",7,"type")="int"
	Set gtmtypfldindx("th_rec","cache_index")=7
	Set gtmtypes("th_rec",8,"name")="th_rec.flushing"
	Set gtmtypes("th_rec",8,"off")=40
	Set gtmtypes("th_rec",8,"len")=1
	Set gtmtypes("th_rec",8,"type")="char"
	Set gtmtypfldindx("th_rec","flushing")=8
	Set gtmtypes("th_rec",9,"name")="th_rec.filler"
	Set gtmtypes("th_rec",9,"off")=41
	Set gtmtypes("th_rec",9,"len")=3
	Set gtmtypes("th_rec",9,"type")="char"
	Set gtmtypfldindx("th_rec","filler")=9
	Set gtmtypes("th_rec",10,"name")="th_rec.filler_int4"
	Set gtmtypes("th_rec",10,"off")=44
	Set gtmtypes("th_rec",10,"len")=4
	Set gtmtypes("th_rec",10,"type")="int"
	Set gtmtypfldindx("th_rec","filler_int4")=10
	;
	Set gtmtypes("thread_parm_t")="struct"
	Set gtmtypes("thread_parm_t",0)=5
	Set gtmtypes("thread_parm_t","len")=40
	Set gtmtypes("thread_parm_t",1,"name")="thread_parm_t.ntasks"
	Set gtmtypes("thread_parm_t",1,"off")=0
	Set gtmtypes("thread_parm_t",1,"len")=4
	Set gtmtypes("thread_parm_t",1,"type")="int"
	Set gtmtypfldindx("thread_parm_t","ntasks")=1
	Set gtmtypes("thread_parm_t",2,"name")="thread_parm_t.fnptr"
	Set gtmtypes("thread_parm_t",2,"off")=8
	Set gtmtypes("thread_parm_t",2,"len")=8
	Set gtmtypes("thread_parm_t",2,"type")="gtm_pthread_fnptr_t"
	Set gtmtypfldindx("thread_parm_t","fnptr")=2
	Set gtmtypes("thread_parm_t",3,"name")="thread_parm_t.ret_array"
	Set gtmtypes("thread_parm_t",3,"off")=16
	Set gtmtypes("thread_parm_t",3,"len")=8
	Set gtmtypes("thread_parm_t",3,"type")="addr"
	Set gtmtypfldindx("thread_parm_t","ret_array")=3
	Set gtmtypes("thread_parm_t",4,"name")="thread_parm_t.parm_array"
	Set gtmtypes("thread_parm_t",4,"off")=24
	Set gtmtypes("thread_parm_t",4,"len")=8
	Set gtmtypes("thread_parm_t",4,"type")="addr"
	Set gtmtypfldindx("thread_parm_t","parm_array")=4
	Set gtmtypes("thread_parm_t",5,"name")="thread_parm_t.parmElemSize"
	Set gtmtypes("thread_parm_t",5,"off")=32
	Set gtmtypes("thread_parm_t",5,"len")=4
	Set gtmtypes("thread_parm_t",5,"type")="int"
	Set gtmtypfldindx("thread_parm_t","parmElemSize")=5
	;
	Set gtmtypes("tlevel_info")="struct"
	Set gtmtypes("tlevel_info",0)=12
	Set gtmtypes("tlevel_info","len")=72
	Set gtmtypes("tlevel_info",1,"name")="tlevel_info.free_que"
	Set gtmtypes("tlevel_info",1,"off")=0
	Set gtmtypes("tlevel_info",1,"len")=16
	Set gtmtypes("tlevel_info",1,"type")="que_ent"
	Set gtmtypfldindx("tlevel_info","free_que")=1
	Set gtmtypes("tlevel_info",2,"name")="tlevel_info.free_que.fl"
	Set gtmtypes("tlevel_info",2,"off")=0
	Set gtmtypes("tlevel_info",2,"len")=8
	Set gtmtypes("tlevel_info",2,"type")="intptr_t"
	Set gtmtypfldindx("tlevel_info","free_que.fl")=2
	Set gtmtypes("tlevel_info",3,"name")="tlevel_info.free_que.bl"
	Set gtmtypes("tlevel_info",3,"off")=8
	Set gtmtypes("tlevel_info",3,"len")=8
	Set gtmtypes("tlevel_info",3,"type")="intptr_t"
	Set gtmtypfldindx("tlevel_info","free_que.bl")=3
	Set gtmtypes("tlevel_info",4,"name")="tlevel_info.next_tlevel_info"
	Set gtmtypes("tlevel_info",4,"off")=16
	Set gtmtypes("tlevel_info",4,"len")=8
	Set gtmtypes("tlevel_info",4,"type")="addr"
	Set gtmtypfldindx("tlevel_info","next_tlevel_info")=4
	Set gtmtypes("tlevel_info",5,"name")="tlevel_info.tlvl_kill_set"
	Set gtmtypes("tlevel_info",5,"off")=24
	Set gtmtypes("tlevel_info",5,"len")=8
	Set gtmtypes("tlevel_info",5,"type")="addr"
	Set gtmtypfldindx("tlevel_info","tlvl_kill_set")=5
	Set gtmtypes("tlevel_info",6,"name")="tlevel_info.tlvl_kill_used"
	Set gtmtypes("tlevel_info",6,"off")=32
	Set gtmtypes("tlevel_info",6,"len")=4
	Set gtmtypes("tlevel_info",6,"type")="int"
	Set gtmtypfldindx("tlevel_info","tlvl_kill_used")=6
	Set gtmtypes("tlevel_info",7,"name")="tlevel_info.tlvl_jfb_info"
	Set gtmtypes("tlevel_info",7,"off")=40
	Set gtmtypes("tlevel_info",7,"len")=8
	Set gtmtypes("tlevel_info",7,"type")="addr"
	Set gtmtypfldindx("tlevel_info","tlvl_jfb_info")=7
	Set gtmtypes("tlevel_info",8,"name")="tlevel_info.tlvl_tp_hist_info"
	Set gtmtypes("tlevel_info",8,"off")=48
	Set gtmtypes("tlevel_info",8,"len")=8
	Set gtmtypes("tlevel_info",8,"type")="addr"
	Set gtmtypfldindx("tlevel_info","tlvl_tp_hist_info")=8
	Set gtmtypes("tlevel_info",9,"name")="tlevel_info.t_level"
	Set gtmtypes("tlevel_info",9,"off")=56
	Set gtmtypes("tlevel_info",9,"len")=4
	Set gtmtypes("tlevel_info",9,"type")="unsigned-int"
	Set gtmtypfldindx("tlevel_info","t_level")=9
	Set gtmtypes("tlevel_info",10,"name")="tlevel_info.update_trans"
	Set gtmtypes("tlevel_info",10,"off")=60
	Set gtmtypes("tlevel_info",10,"len")=4
	Set gtmtypes("tlevel_info",10,"type")="unsigned-int"
	Set gtmtypfldindx("tlevel_info","update_trans")=10
	Set gtmtypes("tlevel_info",11,"name")="tlevel_info.jnl_list_elems"
	Set gtmtypes("tlevel_info",11,"off")=64
	Set gtmtypes("tlevel_info",11,"len")=4
	Set gtmtypes("tlevel_info",11,"type")="unsigned-int"
	Set gtmtypfldindx("tlevel_info","jnl_list_elems")=11
	Set gtmtypes("tlevel_info",12,"name")="tlevel_info.jfb_list_elems"
	Set gtmtypes("tlevel_info",12,"off")=68
	Set gtmtypes("tlevel_info",12,"len")=4
	Set gtmtypes("tlevel_info",12,"type")="unsigned-int"
	Set gtmtypfldindx("tlevel_info","jfb_list_elems")=12
	;
	Set gtmtypes("token_build")="union"
	Set gtmtypes("token_build",0)=4
	Set gtmtypes("token_build","len")=8
	Set gtmtypes("token_build",1,"name")="token_build.t_piece"
	Set gtmtypes("token_build",1,"off")=0
	Set gtmtypes("token_build",1,"len")=8
	Set gtmtypes("token_build",1,"type")="token_split_t"
	Set gtmtypfldindx("token_build","t_piece")=1
	Set gtmtypes("token_build",2,"name")="token_build.t_piece.local_tn"
	Set gtmtypes("token_build",2,"off")=0
	Set gtmtypes("token_build",2,"len")=4
	Set gtmtypes("token_build",2,"type")="unsigned-int"
	Set gtmtypfldindx("token_build","t_piece.local_tn")=2
	Set gtmtypes("token_build",3,"name")="token_build.t_piece.process_id"
	Set gtmtypes("token_build",3,"off")=4
	Set gtmtypes("token_build",3,"len")=4
	Set gtmtypes("token_build",3,"type")="unsigned-int"
	Set gtmtypfldindx("token_build","t_piece.process_id")=3
	Set gtmtypes("token_build",4,"name")="token_build.token"
	Set gtmtypes("token_build",4,"off")=0
	Set gtmtypes("token_build",4,"len")=8
	Set gtmtypes("token_build",4,"type")="uint64_t"
	Set gtmtypfldindx("token_build","token")=4
	;
	Set gtmtypes("token_seq_t")="union"
	Set gtmtypes("token_seq_t",0)=2
	Set gtmtypes("token_seq_t","len")=8
	Set gtmtypes("token_seq_t",1,"name")="token_seq_t.jnl_seqno"
	Set gtmtypes("token_seq_t",1,"off")=0
	Set gtmtypes("token_seq_t",1,"len")=8
	Set gtmtypes("token_seq_t",1,"type")="uint64_t"
	Set gtmtypfldindx("token_seq_t","jnl_seqno")=1
	Set gtmtypes("token_seq_t",2,"name")="token_seq_t.token"
	Set gtmtypes("token_seq_t",2,"off")=0
	Set gtmtypes("token_seq_t",2,"len")=8
	Set gtmtypes("token_seq_t",2,"type")="uint64_t"
	Set gtmtypfldindx("token_seq_t","token")=2
	;
	Set gtmtypes("token_split_t")="struct"
	Set gtmtypes("token_split_t",0)=2
	Set gtmtypes("token_split_t","len")=8
	Set gtmtypes("token_split_t",1,"name")="token_split_t.local_tn"
	Set gtmtypes("token_split_t",1,"off")=0
	Set gtmtypes("token_split_t",1,"len")=4
	Set gtmtypes("token_split_t",1,"type")="unsigned-int"
	Set gtmtypfldindx("token_split_t","local_tn")=1
	Set gtmtypes("token_split_t",2,"name")="token_split_t.process_id"
	Set gtmtypes("token_split_t",2,"off")=4
	Set gtmtypes("token_split_t",2,"len")=4
	Set gtmtypes("token_split_t",2,"type")="unsigned-int"
	Set gtmtypfldindx("token_split_t","process_id")=2
	;
	Set gtmtypes("toktabtype")="struct"
	Set gtmtypes("toktabtype",0)=4
	Set gtmtypes("toktabtype","len")=28
	Set gtmtypes("toktabtype",1,"name")="toktabtype.name"
	Set gtmtypes("toktabtype",1,"off")=0
	Set gtmtypes("toktabtype",1,"len")=20
	Set gtmtypes("toktabtype",1,"type")="char"
	Set gtmtypfldindx("toktabtype","name")=1
	Set gtmtypes("toktabtype",2,"name")="toktabtype.bo_type"
	Set gtmtypes("toktabtype",2,"off")=20
	Set gtmtypes("toktabtype",2,"len")=4
	Set gtmtypes("toktabtype",2,"type")="unsigned-int"
	Set gtmtypfldindx("toktabtype","bo_type")=2
	Set gtmtypes("toktabtype",3,"name")="toktabtype.uo_type"
	Set gtmtypes("toktabtype",3,"off")=24
	Set gtmtypes("toktabtype",3,"len")=1
	Set gtmtypes("toktabtype",3,"type")="char"
	Set gtmtypfldindx("toktabtype","uo_type")=3
	Set gtmtypes("toktabtype",4,"name")="toktabtype.opr_type"
	Set gtmtypes("toktabtype",4,"off")=26
	Set gtmtypes("toktabtype",4,"len")=2
	Set gtmtypes("toktabtype",4,"type")="unsigned-short"
	Set gtmtypfldindx("toktabtype","opr_type")=4
	;
	Set gtmtypes("tp_frame")="struct"
	Set gtmtypes("tp_frame",0)=34
	Set gtmtypes("tp_frame","len")=184
	Set gtmtypes("tp_frame",1,"name")="tp_frame.restart_pc"
	Set gtmtypes("tp_frame",1,"off")=8
	Set gtmtypes("tp_frame",1,"len")=8
	Set gtmtypes("tp_frame",1,"type")="addr"
	Set gtmtypfldindx("tp_frame","restart_pc")=1
	Set gtmtypes("tp_frame",2,"name")="tp_frame.fp"
	Set gtmtypes("tp_frame",2,"off")=16
	Set gtmtypes("tp_frame",2,"len")=8
	Set gtmtypes("tp_frame",2,"type")="addr"
	Set gtmtypfldindx("tp_frame","fp")=2
	Set gtmtypes("tp_frame",3,"name")="tp_frame.mvc"
	Set gtmtypes("tp_frame",3,"off")=24
	Set gtmtypes("tp_frame",3,"len")=8
	Set gtmtypes("tp_frame",3,"type")="addr"
	Set gtmtypfldindx("tp_frame","mvc")=3
	Set gtmtypes("tp_frame",4,"name")="tp_frame.orig_gv_target"
	Set gtmtypes("tp_frame",4,"off")=32
	Set gtmtypes("tp_frame",4,"len")=8
	Set gtmtypes("tp_frame",4,"type")="addr"
	Set gtmtypfldindx("tp_frame","orig_gv_target")=4
	Set gtmtypes("tp_frame",5,"name")="tp_frame.orig_key"
	Set gtmtypes("tp_frame",5,"off")=40
	Set gtmtypes("tp_frame",5,"len")=8
	Set gtmtypes("tp_frame",5,"type")="addr"
	Set gtmtypfldindx("tp_frame","orig_key")=5
	Set gtmtypes("tp_frame",6,"name")="tp_frame.gd_header"
	Set gtmtypes("tp_frame",6,"off")=48
	Set gtmtypes("tp_frame",6,"len")=8
	Set gtmtypes("tp_frame",6,"type")="addr"
	Set gtmtypfldindx("tp_frame","gd_header")=6
	Set gtmtypes("tp_frame",7,"name")="tp_frame.gd_reg"
	Set gtmtypes("tp_frame",7,"off")=56
	Set gtmtypes("tp_frame",7,"len")=8
	Set gtmtypes("tp_frame",7,"type")="addr"
	Set gtmtypfldindx("tp_frame","gd_reg")=7
	Set gtmtypes("tp_frame",8,"name")="tp_frame.sym"
	Set gtmtypes("tp_frame",8,"off")=64
	Set gtmtypes("tp_frame",8,"len")=8
	Set gtmtypes("tp_frame",8,"type")="addr"
	Set gtmtypfldindx("tp_frame","sym")=8
	Set gtmtypes("tp_frame",9,"name")="tp_frame.vars"
	Set gtmtypes("tp_frame",9,"off")=72
	Set gtmtypes("tp_frame",9,"len")=8
	Set gtmtypes("tp_frame",9,"type")="addr"
	Set gtmtypfldindx("tp_frame","vars")=9
	Set gtmtypes("tp_frame",10,"name")="tp_frame.zgbldir"
	Set gtmtypes("tp_frame",10,"off")=80
	Set gtmtypes("tp_frame",10,"len")=32
	Set gtmtypes("tp_frame",10,"type")="mval"
	Set gtmtypfldindx("tp_frame","zgbldir")=10
	Set gtmtypes("tp_frame",11,"name")="tp_frame.zgbldir.mvtype"
	Set gtmtypes("tp_frame",11,"off")=80
	Set gtmtypes("tp_frame",11,"len")=2
	Set gtmtypes("tp_frame",11,"type")="unsigned-short"
	Set gtmtypfldindx("tp_frame","zgbldir.mvtype")=11
	Set gtmtypes("tp_frame",12,"name")="tp_frame.zgbldir.fnpc_indx"
	Set gtmtypes("tp_frame",12,"off")=83
	Set gtmtypes("tp_frame",12,"len")=1
	Set gtmtypes("tp_frame",12,"type")="unsigned-char"
	Set gtmtypfldindx("tp_frame","zgbldir.fnpc_indx")=12
	Set gtmtypes("tp_frame",13,"name")="tp_frame.zgbldir.utfcgr_indx"
	Set gtmtypes("tp_frame",13,"off")=84
	Set gtmtypes("tp_frame",13,"len")=4
	Set gtmtypes("tp_frame",13,"type")="unsigned-int"
	Set gtmtypfldindx("tp_frame","zgbldir.utfcgr_indx")=13
	Set gtmtypes("tp_frame",14,"name")="tp_frame.zgbldir.m"
	Set gtmtypes("tp_frame",14,"off")=88
	Set gtmtypes("tp_frame",14,"len")=8
	Set gtmtypes("tp_frame",14,"type")="int"
	Set gtmtypfldindx("tp_frame","zgbldir.m")=14
	Set gtmtypes("tp_frame",14,"dim")=2
	Set gtmtypes("tp_frame",15,"name")="tp_frame.zgbldir.str"
	Set gtmtypes("tp_frame",15,"off")=96
	Set gtmtypes("tp_frame",15,"len")=16
	Set gtmtypes("tp_frame",15,"type")="mstr"
	Set gtmtypfldindx("tp_frame","zgbldir.str")=15
	Set gtmtypes("tp_frame",16,"name")="tp_frame.zgbldir.str.char_len"
	Set gtmtypes("tp_frame",16,"off")=96
	Set gtmtypes("tp_frame",16,"len")=4
	Set gtmtypes("tp_frame",16,"type")="unsigned-int"
	Set gtmtypfldindx("tp_frame","zgbldir.str.char_len")=16
	Set gtmtypes("tp_frame",17,"name")="tp_frame.zgbldir.str.len"
	Set gtmtypes("tp_frame",17,"off")=100
	Set gtmtypes("tp_frame",17,"len")=4
	Set gtmtypes("tp_frame",17,"type")="int"
	Set gtmtypfldindx("tp_frame","zgbldir.str.len")=17
	Set gtmtypes("tp_frame",18,"name")="tp_frame.zgbldir.str.addr"
	Set gtmtypes("tp_frame",18,"off")=104
	Set gtmtypes("tp_frame",18,"len")=8
	Set gtmtypes("tp_frame",18,"type")="addr"
	Set gtmtypfldindx("tp_frame","zgbldir.str.addr")=18
	Set gtmtypes("tp_frame",19,"name")="tp_frame.trans_id"
	Set gtmtypes("tp_frame",19,"off")=112
	Set gtmtypes("tp_frame",19,"len")=32
	Set gtmtypes("tp_frame",19,"type")="mval"
	Set gtmtypfldindx("tp_frame","trans_id")=19
	Set gtmtypes("tp_frame",20,"name")="tp_frame.trans_id.mvtype"
	Set gtmtypes("tp_frame",20,"off")=112
	Set gtmtypes("tp_frame",20,"len")=2
	Set gtmtypes("tp_frame",20,"type")="unsigned-short"
	Set gtmtypfldindx("tp_frame","trans_id.mvtype")=20
	Set gtmtypes("tp_frame",21,"name")="tp_frame.trans_id.fnpc_indx"
	Set gtmtypes("tp_frame",21,"off")=115
	Set gtmtypes("tp_frame",21,"len")=1
	Set gtmtypes("tp_frame",21,"type")="unsigned-char"
	Set gtmtypfldindx("tp_frame","trans_id.fnpc_indx")=21
	Set gtmtypes("tp_frame",22,"name")="tp_frame.trans_id.utfcgr_indx"
	Set gtmtypes("tp_frame",22,"off")=116
	Set gtmtypes("tp_frame",22,"len")=4
	Set gtmtypes("tp_frame",22,"type")="unsigned-int"
	Set gtmtypfldindx("tp_frame","trans_id.utfcgr_indx")=22
	Set gtmtypes("tp_frame",23,"name")="tp_frame.trans_id.m"
	Set gtmtypes("tp_frame",23,"off")=120
	Set gtmtypes("tp_frame",23,"len")=8
	Set gtmtypes("tp_frame",23,"type")="int"
	Set gtmtypfldindx("tp_frame","trans_id.m")=23
	Set gtmtypes("tp_frame",23,"dim")=2
	Set gtmtypes("tp_frame",24,"name")="tp_frame.trans_id.str"
	Set gtmtypes("tp_frame",24,"off")=128
	Set gtmtypes("tp_frame",24,"len")=16
	Set gtmtypes("tp_frame",24,"type")="mstr"
	Set gtmtypfldindx("tp_frame","trans_id.str")=24
	Set gtmtypes("tp_frame",25,"name")="tp_frame.trans_id.str.char_len"
	Set gtmtypes("tp_frame",25,"off")=128
	Set gtmtypes("tp_frame",25,"len")=4
	Set gtmtypes("tp_frame",25,"type")="unsigned-int"
	Set gtmtypfldindx("tp_frame","trans_id.str.char_len")=25
	Set gtmtypes("tp_frame",26,"name")="tp_frame.trans_id.str.len"
	Set gtmtypes("tp_frame",26,"off")=132
	Set gtmtypes("tp_frame",26,"len")=4
	Set gtmtypes("tp_frame",26,"type")="int"
	Set gtmtypfldindx("tp_frame","trans_id.str.len")=26
	Set gtmtypes("tp_frame",27,"name")="tp_frame.trans_id.str.addr"
	Set gtmtypes("tp_frame",27,"off")=136
	Set gtmtypes("tp_frame",27,"len")=8
	Set gtmtypes("tp_frame",27,"type")="addr"
	Set gtmtypfldindx("tp_frame","trans_id.str.addr")=27
	Set gtmtypes("tp_frame",28,"name")="tp_frame.extnam_str"
	Set gtmtypes("tp_frame",28,"off")=144
	Set gtmtypes("tp_frame",28,"len")=16
	Set gtmtypes("tp_frame",28,"type")="mstr"
	Set gtmtypfldindx("tp_frame","extnam_str")=28
	Set gtmtypes("tp_frame",29,"name")="tp_frame.extnam_str.char_len"
	Set gtmtypes("tp_frame",29,"off")=144
	Set gtmtypes("tp_frame",29,"len")=4
	Set gtmtypes("tp_frame",29,"type")="unsigned-int"
	Set gtmtypfldindx("tp_frame","extnam_str.char_len")=29
	Set gtmtypes("tp_frame",30,"name")="tp_frame.extnam_str.len"
	Set gtmtypes("tp_frame",30,"off")=148
	Set gtmtypes("tp_frame",30,"len")=4
	Set gtmtypes("tp_frame",30,"type")="int"
	Set gtmtypfldindx("tp_frame","extnam_str.len")=30
	Set gtmtypes("tp_frame",31,"name")="tp_frame.extnam_str.addr"
	Set gtmtypes("tp_frame",31,"off")=152
	Set gtmtypes("tp_frame",31,"len")=8
	Set gtmtypes("tp_frame",31,"type")="addr"
	Set gtmtypfldindx("tp_frame","extnam_str.addr")=31
	Set gtmtypes("tp_frame",32,"name")="tp_frame.old_tp_frame"
	Set gtmtypes("tp_frame",32,"off")=160
	Set gtmtypes("tp_frame",32,"len")=8
	Set gtmtypes("tp_frame",32,"type")="addr"
	Set gtmtypfldindx("tp_frame","old_tp_frame")=32
	Set gtmtypes("tp_frame",33,"name")="tp_frame.restart_ctxt"
	Set gtmtypes("tp_frame",33,"off")=168
	Set gtmtypes("tp_frame",33,"len")=8
	Set gtmtypes("tp_frame",33,"type")="addr"
	Set gtmtypfldindx("tp_frame","restart_ctxt")=33
	Set gtmtypes("tp_frame",34,"name")="tp_frame.active_lv"
	Set gtmtypes("tp_frame",34,"off")=176
	Set gtmtypes("tp_frame",34,"len")=8
	Set gtmtypes("tp_frame",34,"type")="addr"
	Set gtmtypfldindx("tp_frame","active_lv")=34
	;
	Set gtmtypes("tp_region")="struct"
	Set gtmtypes("tp_region",0)=6
	Set gtmtypes("tp_region","len")=40
	Set gtmtypes("tp_region",1,"name")="tp_region.fPtr"
	Set gtmtypes("tp_region",1,"off")=0
	Set gtmtypes("tp_region",1,"len")=8
	Set gtmtypes("tp_region",1,"type")="addr"
	Set gtmtypfldindx("tp_region","fPtr")=1
	Set gtmtypes("tp_region",2,"name")="tp_region.reg"
	Set gtmtypes("tp_region",2,"off")=8
	Set gtmtypes("tp_region",2,"len")=8
	Set gtmtypes("tp_region",2,"type")="addr"
	Set gtmtypfldindx("tp_region","reg")=2
	Set gtmtypes("tp_region",3,"name")="tp_region.file_id"
	Set gtmtypes("tp_region",3,"off")=16
	Set gtmtypes("tp_region",3,"len")=16
	Set gtmtypes("tp_region",3,"type")="unix_file_id"
	Set gtmtypfldindx("tp_region","file_id")=3
	Set gtmtypes("tp_region",4,"name")="tp_region.file_id.inode"
	Set gtmtypes("tp_region",4,"off")=16
	Set gtmtypes("tp_region",4,"len")=8
	Set gtmtypes("tp_region",4,"type")="ino_t"
	Set gtmtypfldindx("tp_region","file_id.inode")=4
	Set gtmtypes("tp_region",5,"name")="tp_region.file_id.device"
	Set gtmtypes("tp_region",5,"off")=24
	Set gtmtypes("tp_region",5,"len")=8
	Set gtmtypes("tp_region",5,"type")="dev_t"
	Set gtmtypfldindx("tp_region","file_id.device")=5
	Set gtmtypes("tp_region",6,"name")="tp_region.fid_index"
	Set gtmtypes("tp_region",6,"off")=32
	Set gtmtypes("tp_region",6,"len")=4
	Set gtmtypes("tp_region",6,"type")="int"
	Set gtmtypfldindx("tp_region","fid_index")=6
	;
	Set gtmtypes("tp_var")="struct"
	Set gtmtypes("tp_var",0)=12
	Set gtmtypes("tp_var","len")=56
	Set gtmtypes("tp_var",1,"name")="tp_var.next"
	Set gtmtypes("tp_var",1,"off")=0
	Set gtmtypes("tp_var",1,"len")=8
	Set gtmtypes("tp_var",1,"type")="addr"
	Set gtmtypfldindx("tp_var","next")=1
	Set gtmtypes("tp_var",2,"name")="tp_var.current_value"
	Set gtmtypes("tp_var",2,"off")=8
	Set gtmtypes("tp_var",2,"len")=8
	Set gtmtypes("tp_var",2,"type")="addr"
	Set gtmtypfldindx("tp_var","current_value")=2
	Set gtmtypes("tp_var",3,"name")="tp_var.save_value"
	Set gtmtypes("tp_var",3,"off")=16
	Set gtmtypes("tp_var",3,"len")=8
	Set gtmtypes("tp_var",3,"type")="addr"
	Set gtmtypfldindx("tp_var","save_value")=3
	Set gtmtypes("tp_var",4,"name")="tp_var.key"
	Set gtmtypes("tp_var",4,"off")=24
	Set gtmtypes("tp_var",4,"len")=24
	Set gtmtypes("tp_var",4,"type")="mname_entry"
	Set gtmtypfldindx("tp_var","key")=4
	Set gtmtypes("tp_var",5,"name")="tp_var.key.var_name"
	Set gtmtypes("tp_var",5,"off")=24
	Set gtmtypes("tp_var",5,"len")=16
	Set gtmtypes("tp_var",5,"type")="mstr"
	Set gtmtypfldindx("tp_var","key.var_name")=5
	Set gtmtypes("tp_var",6,"name")="tp_var.key.var_name.char_len"
	Set gtmtypes("tp_var",6,"off")=24
	Set gtmtypes("tp_var",6,"len")=4
	Set gtmtypes("tp_var",6,"type")="unsigned-int"
	Set gtmtypfldindx("tp_var","key.var_name.char_len")=6
	Set gtmtypes("tp_var",7,"name")="tp_var.key.var_name.len"
	Set gtmtypes("tp_var",7,"off")=28
	Set gtmtypes("tp_var",7,"len")=4
	Set gtmtypes("tp_var",7,"type")="int"
	Set gtmtypfldindx("tp_var","key.var_name.len")=7
	Set gtmtypes("tp_var",8,"name")="tp_var.key.var_name.addr"
	Set gtmtypes("tp_var",8,"off")=32
	Set gtmtypes("tp_var",8,"len")=8
	Set gtmtypes("tp_var",8,"type")="addr"
	Set gtmtypfldindx("tp_var","key.var_name.addr")=8
	Set gtmtypes("tp_var",9,"name")="tp_var.key.hash_code"
	Set gtmtypes("tp_var",9,"off")=40
	Set gtmtypes("tp_var",9,"len")=4
	Set gtmtypes("tp_var",9,"type")="unsigned-int"
	Set gtmtypfldindx("tp_var","key.hash_code")=9
	Set gtmtypes("tp_var",10,"name")="tp_var.key.marked"
	Set gtmtypes("tp_var",10,"off")=44
	Set gtmtypes("tp_var",10,"len")=4
	Set gtmtypes("tp_var",10,"type")="boolean_t"
	Set gtmtypfldindx("tp_var","key.marked")=10
	Set gtmtypes("tp_var",11,"name")="tp_var.var_cloned"
	Set gtmtypes("tp_var",11,"off")=48
	Set gtmtypes("tp_var",11,"len")=4
	Set gtmtypes("tp_var",11,"type")="boolean_t"
	Set gtmtypfldindx("tp_var","var_cloned")=11
	Set gtmtypes("tp_var",12,"name")="tp_var.filler"
	Set gtmtypes("tp_var",12,"off")=52
	Set gtmtypes("tp_var",12,"len")=4
	Set gtmtypes("tp_var",12,"type")="int"
	Set gtmtypfldindx("tp_var","filler")=12
	;
	Set gtmtypes("tr_search_status_t")="struct"
	Set gtmtypes("tr_search_status_t",0)=2
	Set gtmtypes("tr_search_status_t","len")=16
	Set gtmtypes("tr_search_status_t",1,"name")="tr_search_status_t.seqno"
	Set gtmtypes("tr_search_status_t",1,"off")=0
	Set gtmtypes("tr_search_status_t",1,"len")=8
	Set gtmtypes("tr_search_status_t",1,"type")="uint64_t"
	Set gtmtypfldindx("tr_search_status_t","seqno")=1
	Set gtmtypes("tr_search_status_t",2,"name")="tr_search_status_t.prev_seqno"
	Set gtmtypes("tr_search_status_t",2,"off")=8
	Set gtmtypes("tr_search_status_t",2,"len")=8
	Set gtmtypes("tr_search_status_t",2,"type")="uint64_t"
	Set gtmtypfldindx("tr_search_status_t","prev_seqno")=2
	;
	Set gtmtypes("trace_entry")="struct"
	Set gtmtypes("trace_entry",0)=9
	Set gtmtypes("trace_entry","len")=64
	Set gtmtypes("trace_entry",1,"name")="trace_entry.rout_name"
	Set gtmtypes("trace_entry",1,"off")=0
	Set gtmtypes("trace_entry",1,"len")=8
	Set gtmtypes("trace_entry",1,"type")="addr"
	Set gtmtypfldindx("trace_entry","rout_name")=1
	Set gtmtypes("trace_entry",2,"name")="trace_entry.label_name"
	Set gtmtypes("trace_entry",2,"off")=8
	Set gtmtypes("trace_entry",2,"len")=8
	Set gtmtypes("trace_entry",2,"type")="addr"
	Set gtmtypfldindx("trace_entry","label_name")=2
	Set gtmtypes("trace_entry",3,"name")="trace_entry.line_num"
	Set gtmtypes("trace_entry",3,"off")=16
	Set gtmtypes("trace_entry",3,"len")=4
	Set gtmtypes("trace_entry",3,"type")="int"
	Set gtmtypfldindx("trace_entry","line_num")=3
	Set gtmtypes("trace_entry",4,"name")="trace_entry.count"
	Set gtmtypes("trace_entry",4,"off")=20
	Set gtmtypes("trace_entry",4,"len")=4
	Set gtmtypes("trace_entry",4,"type")="unsigned"
	Set gtmtypfldindx("trace_entry","count")=4
	Set gtmtypes("trace_entry",5,"name")="trace_entry.sys_time"
	Set gtmtypes("trace_entry",5,"off")=24
	Set gtmtypes("trace_entry",5,"len")=8
	Set gtmtypes("trace_entry",5,"type")="uint64_t"
	Set gtmtypfldindx("trace_entry","sys_time")=5
	Set gtmtypes("trace_entry",6,"name")="trace_entry.usr_time"
	Set gtmtypes("trace_entry",6,"off")=32
	Set gtmtypes("trace_entry",6,"len")=8
	Set gtmtypes("trace_entry",6,"type")="uint64_t"
	Set gtmtypfldindx("trace_entry","usr_time")=6
	Set gtmtypes("trace_entry",7,"name")="trace_entry.elp_time"
	Set gtmtypes("trace_entry",7,"off")=40
	Set gtmtypes("trace_entry",7,"len")=8
	Set gtmtypes("trace_entry",7,"type")="uint64_t"
	Set gtmtypfldindx("trace_entry","elp_time")=7
	Set gtmtypes("trace_entry",8,"name")="trace_entry.loop_level"
	Set gtmtypes("trace_entry",8,"off")=48
	Set gtmtypes("trace_entry",8,"len")=4
	Set gtmtypes("trace_entry",8,"type")="int"
	Set gtmtypfldindx("trace_entry","loop_level")=8
	Set gtmtypes("trace_entry",9,"name")="trace_entry.raddr"
	Set gtmtypes("trace_entry",9,"off")=56
	Set gtmtypes("trace_entry",9,"len")=8
	Set gtmtypes("trace_entry",9,"type")="addr"
	Set gtmtypfldindx("trace_entry","raddr")=9
	;
	Set gtmtypes("trctbl_entry")="struct"
	Set gtmtypes("trctbl_entry",0)=5
	Set gtmtypes("trctbl_entry","len")=32
	Set gtmtypes("trctbl_entry",1,"name")="trctbl_entry.type"
	Set gtmtypes("trctbl_entry",1,"off")=0
	Set gtmtypes("trctbl_entry",1,"len")=4
	Set gtmtypes("trctbl_entry",1,"type")="int"
	Set gtmtypfldindx("trctbl_entry","type")=1
	Set gtmtypes("trctbl_entry",2,"name")="trctbl_entry.intfld"
	Set gtmtypes("trctbl_entry",2,"off")=4
	Set gtmtypes("trctbl_entry",2,"len")=4
	Set gtmtypes("trctbl_entry",2,"type")="int"
	Set gtmtypfldindx("trctbl_entry","intfld")=2
	Set gtmtypes("trctbl_entry",3,"name")="trctbl_entry.addrfld1"
	Set gtmtypes("trctbl_entry",3,"off")=8
	Set gtmtypes("trctbl_entry",3,"len")=8
	Set gtmtypes("trctbl_entry",3,"type")="addr"
	Set gtmtypfldindx("trctbl_entry","addrfld1")=3
	Set gtmtypes("trctbl_entry",4,"name")="trctbl_entry.addrfld2"
	Set gtmtypes("trctbl_entry",4,"off")=16
	Set gtmtypes("trctbl_entry",4,"len")=8
	Set gtmtypes("trctbl_entry",4,"type")="addr"
	Set gtmtypfldindx("trctbl_entry","addrfld2")=4
	Set gtmtypes("trctbl_entry",5,"name")="trctbl_entry.addrfld3"
	Set gtmtypes("trctbl_entry",5,"off")=24
	Set gtmtypes("trctbl_entry",5,"len")=8
	Set gtmtypes("trctbl_entry",5,"type")="addr"
	Set gtmtypfldindx("trctbl_entry","addrfld3")=5
	;
	Set gtmtypes("treeSrchStatus")="struct"
	Set gtmtypes("treeSrchStatus",0)=3
	Set gtmtypes("treeSrchStatus","len")=24
	Set gtmtypes("treeSrchStatus",1,"name")="treeSrchStatus.lastNodeLookedUp"
	Set gtmtypes("treeSrchStatus",1,"off")=0
	Set gtmtypes("treeSrchStatus",1,"len")=8
	Set gtmtypes("treeSrchStatus",1,"type")="addr"
	Set gtmtypfldindx("treeSrchStatus","lastNodeLookedUp")=1
	Set gtmtypes("treeSrchStatus",2,"name")="treeSrchStatus.lastNodeMin"
	Set gtmtypes("treeSrchStatus",2,"off")=8
	Set gtmtypes("treeSrchStatus",2,"len")=8
	Set gtmtypes("treeSrchStatus",2,"type")="addr"
	Set gtmtypfldindx("treeSrchStatus","lastNodeMin")=2
	Set gtmtypes("treeSrchStatus",3,"name")="treeSrchStatus.lastNodeMax"
	Set gtmtypes("treeSrchStatus",3,"off")=16
	Set gtmtypes("treeSrchStatus",3,"len")=8
	Set gtmtypes("treeSrchStatus",3,"type")="addr"
	Set gtmtypfldindx("treeSrchStatus","lastNodeMax")=3
	;
	Set gtmtypes("triple")="struct"
	Set gtmtypes("triple",0)=50
	Set gtmtypes("triple","len")=136
	Set gtmtypes("triple",1,"name")="triple.opcode"
	Set gtmtypes("triple",1,"off")=0
	Set gtmtypes("triple",1,"len")=4
	Set gtmtypes("triple",1,"type")="unsigned-int"
	Set gtmtypfldindx("triple","opcode")=1
	Set gtmtypes("triple",2,"name")="triple.exorder"
	Set gtmtypes("triple",2,"off")=8
	Set gtmtypes("triple",2,"len")=16
	Set gtmtypes("triple",2,"type")="struct"
	Set gtmtypfldindx("triple","exorder")=2
	Set gtmtypes("triple",3,"name")="triple.exorder.fl"
	Set gtmtypes("triple",3,"off")=8
	Set gtmtypes("triple",3,"len")=8
	Set gtmtypes("triple",3,"type")="addr"
	Set gtmtypfldindx("triple","exorder.fl")=3
	Set gtmtypes("triple",4,"name")="triple.exorder.bl"
	Set gtmtypes("triple",4,"off")=16
	Set gtmtypes("triple",4,"len")=8
	Set gtmtypes("triple",4,"type")="addr"
	Set gtmtypfldindx("triple","exorder.bl")=4
	Set gtmtypes("triple",5,"name")="triple.backptr"
	Set gtmtypes("triple",5,"off")=24
	Set gtmtypes("triple",5,"len")=24
	Set gtmtypes("triple",5,"type")="tbp"
	Set gtmtypfldindx("triple","backptr")=5
	Set gtmtypes("triple",6,"name")="triple.backptr.que"
	Set gtmtypes("triple",6,"off")=24
	Set gtmtypes("triple",6,"len")=16
	Set gtmtypes("triple",6,"type")="struct"
	Set gtmtypfldindx("triple","backptr.que")=6
	Set gtmtypes("triple",7,"name")="triple.backptr.que.fl"
	Set gtmtypes("triple",7,"off")=24
	Set gtmtypes("triple",7,"len")=8
	Set gtmtypes("triple",7,"type")="addr"
	Set gtmtypfldindx("triple","backptr.que.fl")=7
	Set gtmtypes("triple",8,"name")="triple.backptr.que.bl"
	Set gtmtypes("triple",8,"off")=32
	Set gtmtypes("triple",8,"len")=8
	Set gtmtypes("triple",8,"type")="addr"
	Set gtmtypfldindx("triple","backptr.que.bl")=8
	Set gtmtypes("triple",9,"name")="triple.backptr.bpt"
	Set gtmtypes("triple",9,"off")=40
	Set gtmtypes("triple",9,"len")=8
	Set gtmtypes("triple",9,"type")="addr"
	Set gtmtypfldindx("triple","backptr.bpt")=9
	Set gtmtypes("triple",10,"name")="triple.jmplist"
	Set gtmtypes("triple",10,"off")=48
	Set gtmtypes("triple",10,"len")=24
	Set gtmtypes("triple",10,"type")="tbp"
	Set gtmtypfldindx("triple","jmplist")=10
	Set gtmtypes("triple",11,"name")="triple.jmplist.que"
	Set gtmtypes("triple",11,"off")=48
	Set gtmtypes("triple",11,"len")=16
	Set gtmtypes("triple",11,"type")="struct"
	Set gtmtypfldindx("triple","jmplist.que")=11
	Set gtmtypes("triple",12,"name")="triple.jmplist.que.fl"
	Set gtmtypes("triple",12,"off")=48
	Set gtmtypes("triple",12,"len")=8
	Set gtmtypes("triple",12,"type")="addr"
	Set gtmtypfldindx("triple","jmplist.que.fl")=12
	Set gtmtypes("triple",13,"name")="triple.jmplist.que.bl"
	Set gtmtypes("triple",13,"off")=56
	Set gtmtypes("triple",13,"len")=8
	Set gtmtypes("triple",13,"type")="addr"
	Set gtmtypfldindx("triple","jmplist.que.bl")=13
	Set gtmtypes("triple",14,"name")="triple.jmplist.bpt"
	Set gtmtypes("triple",14,"off")=64
	Set gtmtypes("triple",14,"len")=8
	Set gtmtypes("triple",14,"type")="addr"
	Set gtmtypfldindx("triple","jmplist.bpt")=14
	Set gtmtypes("triple",15,"name")="triple.src"
	Set gtmtypes("triple",15,"off")=72
	Set gtmtypes("triple",15,"len")=8
	Set gtmtypes("triple",15,"type")="source_address"
	Set gtmtypfldindx("triple","src")=15
	Set gtmtypes("triple",16,"name")="triple.src.line"
	Set gtmtypes("triple",16,"off")=72
	Set gtmtypes("triple",16,"len")=4
	Set gtmtypes("triple",16,"type")="unsigned-int"
	Set gtmtypfldindx("triple","src.line")=16
	Set gtmtypes("triple",17,"name")="triple.src.column"
	Set gtmtypes("triple",17,"off")=76
	Set gtmtypes("triple",17,"len")=4
	Set gtmtypes("triple",17,"type")="unsigned-int"
	Set gtmtypfldindx("triple","src.column")=17
	Set gtmtypes("triple",18,"name")="triple.rtaddr"
	Set gtmtypes("triple",18,"off")=80
	Set gtmtypes("triple",18,"len")=4
	Set gtmtypes("triple",18,"type")="int"
	Set gtmtypfldindx("triple","rtaddr")=18
	Set gtmtypes("triple",19,"name")="triple.operand"
	Set gtmtypes("triple",19,"off")=88
	Set gtmtypes("triple",19,"len")=32
	Set gtmtypes("triple",19,"type")="oprtype"
	Set gtmtypfldindx("triple","operand")=19
	Set gtmtypes("triple",19,"dim")=2
	Set gtmtypes("triple",20,"name")="triple.operand[0].oprclass"
	Set gtmtypes("triple",20,"off")=88
	Set gtmtypes("triple",20,"len")=4
	Set gtmtypes("triple",20,"type")="int"
	Set gtmtypfldindx("triple","operand[0].oprclass")=20
	Set gtmtypes("triple",21,"name")="triple.operand[0].oprval"
	Set gtmtypes("triple",21,"off")=96
	Set gtmtypes("triple",21,"len")=8
	Set gtmtypes("triple",21,"type")="union"
	Set gtmtypfldindx("triple","operand[0].oprval")=21
	Set gtmtypes("triple",22,"name")="triple.operand[0].oprval.indr"
	Set gtmtypes("triple",22,"off")=96
	Set gtmtypes("triple",22,"len")=8
	Set gtmtypes("triple",22,"type")="addr"
	Set gtmtypfldindx("triple","operand[0].oprval.indr")=22
	Set gtmtypes("triple",23,"name")="triple.operand[0].oprval.tref"
	Set gtmtypes("triple",23,"off")=96
	Set gtmtypes("triple",23,"len")=8
	Set gtmtypes("triple",23,"type")="addr"
	Set gtmtypfldindx("triple","operand[0].oprval.tref")=23
	Set gtmtypes("triple",24,"name")="triple.operand[0].oprval.tsize"
	Set gtmtypes("triple",24,"off")=96
	Set gtmtypes("triple",24,"len")=8
	Set gtmtypes("triple",24,"type")="addr"
	Set gtmtypfldindx("triple","operand[0].oprval.tsize")=24
	Set gtmtypes("triple",25,"name")="triple.operand[0].oprval.lab"
	Set gtmtypes("triple",25,"off")=96
	Set gtmtypes("triple",25,"len")=8
	Set gtmtypes("triple",25,"type")="addr"
	Set gtmtypfldindx("triple","operand[0].oprval.lab")=25
	Set gtmtypes("triple",26,"name")="triple.operand[0].oprval.mlin"
	Set gtmtypes("triple",26,"off")=96
	Set gtmtypes("triple",26,"len")=8
	Set gtmtypes("triple",26,"type")="addr"
	Set gtmtypfldindx("triple","operand[0].oprval.mlin")=26
	Set gtmtypes("triple",27,"name")="triple.operand[0].oprval.mlit"
	Set gtmtypes("triple",27,"off")=96
	Set gtmtypes("triple",27,"len")=8
	Set gtmtypes("triple",27,"type")="addr"
	Set gtmtypfldindx("triple","operand[0].oprval.mlit")=27
	Set gtmtypes("triple",28,"name")="triple.operand[0].oprval.cdlt"
	Set gtmtypes("triple",28,"off")=96
	Set gtmtypes("triple",28,"len")=8
	Set gtmtypes("triple",28,"type")="addr"
	Set gtmtypfldindx("triple","operand[0].oprval.cdlt")=28
	Set gtmtypes("triple",29,"name")="triple.operand[0].oprval.cdidx"
	Set gtmtypes("triple",29,"off")=96
	Set gtmtypes("triple",29,"len")=8
	Set gtmtypes("triple",29,"type")="addr"
	Set gtmtypfldindx("triple","operand[0].oprval.cdidx")=29
	Set gtmtypes("triple",30,"name")="triple.operand[0].oprval.vref"
	Set gtmtypes("triple",30,"off")=96
	Set gtmtypes("triple",30,"len")=8
	Set gtmtypes("triple",30,"type")="addr"
	Set gtmtypfldindx("triple","operand[0].oprval.vref")=30
	Set gtmtypes("triple",31,"name")="triple.operand[0].oprval.temp"
	Set gtmtypes("triple",31,"off")=96
	Set gtmtypes("triple",31,"len")=4
	Set gtmtypes("triple",31,"type")="int"
	Set gtmtypfldindx("triple","operand[0].oprval.temp")=31
	Set gtmtypes("triple",32,"name")="triple.operand[0].oprval.ilit"
	Set gtmtypes("triple",32,"off")=96
	Set gtmtypes("triple",32,"len")=4
	Set gtmtypes("triple",32,"type")="int"
	Set gtmtypfldindx("triple","operand[0].oprval.ilit")=32
	Set gtmtypes("triple",33,"name")="triple.operand[0].oprval.offset"
	Set gtmtypes("triple",33,"off")=96
	Set gtmtypes("triple",33,"len")=4
	Set gtmtypes("triple",33,"type")="int"
	Set gtmtypfldindx("triple","operand[0].oprval.offset")=33
	Set gtmtypes("triple",34,"name")="triple.operand[0].oprval.vreg"
	Set gtmtypes("triple",34,"off")=96
	Set gtmtypes("triple",34,"len")=1
	Set gtmtypes("triple",34,"type")="unsigned-char"
	Set gtmtypfldindx("triple","operand[0].oprval.vreg")=34
	Set gtmtypes("triple",35,"name")="triple.destination"
	Set gtmtypes("triple",35,"off")=120
	Set gtmtypes("triple",35,"len")=16
	Set gtmtypes("triple",35,"type")="oprtype"
	Set gtmtypfldindx("triple","destination")=35
	Set gtmtypes("triple",36,"name")="triple.destination.oprclass"
	Set gtmtypes("triple",36,"off")=120
	Set gtmtypes("triple",36,"len")=4
	Set gtmtypes("triple",36,"type")="int"
	Set gtmtypfldindx("triple","destination.oprclass")=36
	Set gtmtypes("triple",37,"name")="triple.destination.oprval"
	Set gtmtypes("triple",37,"off")=128
	Set gtmtypes("triple",37,"len")=8
	Set gtmtypes("triple",37,"type")="union"
	Set gtmtypfldindx("triple","destination.oprval")=37
	Set gtmtypes("triple",38,"name")="triple.destination.oprval.indr"
	Set gtmtypes("triple",38,"off")=128
	Set gtmtypes("triple",38,"len")=8
	Set gtmtypes("triple",38,"type")="addr"
	Set gtmtypfldindx("triple","destination.oprval.indr")=38
	Set gtmtypes("triple",39,"name")="triple.destination.oprval.tref"
	Set gtmtypes("triple",39,"off")=128
	Set gtmtypes("triple",39,"len")=8
	Set gtmtypes("triple",39,"type")="addr"
	Set gtmtypfldindx("triple","destination.oprval.tref")=39
	Set gtmtypes("triple",40,"name")="triple.destination.oprval.tsize"
	Set gtmtypes("triple",40,"off")=128
	Set gtmtypes("triple",40,"len")=8
	Set gtmtypes("triple",40,"type")="addr"
	Set gtmtypfldindx("triple","destination.oprval.tsize")=40
	Set gtmtypes("triple",41,"name")="triple.destination.oprval.lab"
	Set gtmtypes("triple",41,"off")=128
	Set gtmtypes("triple",41,"len")=8
	Set gtmtypes("triple",41,"type")="addr"
	Set gtmtypfldindx("triple","destination.oprval.lab")=41
	Set gtmtypes("triple",42,"name")="triple.destination.oprval.mlin"
	Set gtmtypes("triple",42,"off")=128
	Set gtmtypes("triple",42,"len")=8
	Set gtmtypes("triple",42,"type")="addr"
	Set gtmtypfldindx("triple","destination.oprval.mlin")=42
	Set gtmtypes("triple",43,"name")="triple.destination.oprval.mlit"
	Set gtmtypes("triple",43,"off")=128
	Set gtmtypes("triple",43,"len")=8
	Set gtmtypes("triple",43,"type")="addr"
	Set gtmtypfldindx("triple","destination.oprval.mlit")=43
	Set gtmtypes("triple",44,"name")="triple.destination.oprval.cdlt"
	Set gtmtypes("triple",44,"off")=128
	Set gtmtypes("triple",44,"len")=8
	Set gtmtypes("triple",44,"type")="addr"
	Set gtmtypfldindx("triple","destination.oprval.cdlt")=44
	Set gtmtypes("triple",45,"name")="triple.destination.oprval.cdidx"
	Set gtmtypes("triple",45,"off")=128
	Set gtmtypes("triple",45,"len")=8
	Set gtmtypes("triple",45,"type")="addr"
	Set gtmtypfldindx("triple","destination.oprval.cdidx")=45
	Set gtmtypes("triple",46,"name")="triple.destination.oprval.vref"
	Set gtmtypes("triple",46,"off")=128
	Set gtmtypes("triple",46,"len")=8
	Set gtmtypes("triple",46,"type")="addr"
	Set gtmtypfldindx("triple","destination.oprval.vref")=46
	Set gtmtypes("triple",47,"name")="triple.destination.oprval.temp"
	Set gtmtypes("triple",47,"off")=128
	Set gtmtypes("triple",47,"len")=4
	Set gtmtypes("triple",47,"type")="int"
	Set gtmtypfldindx("triple","destination.oprval.temp")=47
	Set gtmtypes("triple",48,"name")="triple.destination.oprval.ilit"
	Set gtmtypes("triple",48,"off")=128
	Set gtmtypes("triple",48,"len")=4
	Set gtmtypes("triple",48,"type")="int"
	Set gtmtypfldindx("triple","destination.oprval.ilit")=48
	Set gtmtypes("triple",49,"name")="triple.destination.oprval.offset"
	Set gtmtypes("triple",49,"off")=128
	Set gtmtypes("triple",49,"len")=4
	Set gtmtypes("triple",49,"type")="int"
	Set gtmtypfldindx("triple","destination.oprval.offset")=49
	Set gtmtypes("triple",50,"name")="triple.destination.oprval.vreg"
	Set gtmtypes("triple",50,"off")=128
	Set gtmtypes("triple",50,"len")=1
	Set gtmtypes("triple",50,"type")="unsigned-char"
	Set gtmtypfldindx("triple","destination.oprval.vreg")=50
	;
	Set gtmtypes("tripsize")="struct"
	Set gtmtypes("tripsize",0)=2
	Set gtmtypes("tripsize","len")=16
	Set gtmtypes("tripsize",1,"name")="tripsize.ct"
	Set gtmtypes("tripsize",1,"off")=0
	Set gtmtypes("tripsize",1,"len")=8
	Set gtmtypes("tripsize",1,"type")="addr"
	Set gtmtypfldindx("tripsize","ct")=1
	Set gtmtypes("tripsize",2,"name")="tripsize.size"
	Set gtmtypes("tripsize",2,"off")=8
	Set gtmtypes("tripsize",2,"len")=4
	Set gtmtypes("tripsize",2,"type")="int"
	Set gtmtypfldindx("tripsize","size")=2
	;
	Set gtmtypes("trunc_region")="struct"
	Set gtmtypes("trunc_region",0)=2
	Set gtmtypes("trunc_region","len")=16
	Set gtmtypes("trunc_region",1,"name")="trunc_region.reg"
	Set gtmtypes("trunc_region",1,"off")=0
	Set gtmtypes("trunc_region",1,"len")=8
	Set gtmtypes("trunc_region",1,"type")="addr"
	Set gtmtypfldindx("trunc_region","reg")=1
	Set gtmtypes("trunc_region",2,"name")="trunc_region.next"
	Set gtmtypes("trunc_region",2,"off")=8
	Set gtmtypes("trunc_region",2,"len")=8
	Set gtmtypes("trunc_region",2,"type")="addr"
	Set gtmtypfldindx("trunc_region","next")=2
	;
	Set gtmtypes("tt_interrupt")="struct"
	Set gtmtypes("tt_interrupt",0)=24
	Set gtmtypes("tt_interrupt","len")=136
	Set gtmtypes("tt_interrupt",1,"name")="tt_interrupt.who_saved"
	Set gtmtypes("tt_interrupt",1,"off")=0
	Set gtmtypes("tt_interrupt",1,"len")=4
	Set gtmtypes("tt_interrupt",1,"type")="int"
	Set gtmtypfldindx("tt_interrupt","who_saved")=1
	Set gtmtypes("tt_interrupt",2,"name")="tt_interrupt.buffer_start"
	Set gtmtypes("tt_interrupt",2,"off")=8
	Set gtmtypes("tt_interrupt",2,"len")=8
	Set gtmtypes("tt_interrupt",2,"type")="addr"
	Set gtmtypfldindx("tt_interrupt","buffer_start")=2
	Set gtmtypes("tt_interrupt",3,"name")="tt_interrupt.buffer_32_start"
	Set gtmtypes("tt_interrupt",3,"off")=16
	Set gtmtypes("tt_interrupt",3,"len")=8
	Set gtmtypes("tt_interrupt",3,"type")="addr"
	Set gtmtypfldindx("tt_interrupt","buffer_32_start")=3
	Set gtmtypes("tt_interrupt",4,"name")="tt_interrupt.utf8_more"
	Set gtmtypes("tt_interrupt",4,"off")=24
	Set gtmtypes("tt_interrupt",4,"len")=4
	Set gtmtypes("tt_interrupt",4,"type")="int"
	Set gtmtypfldindx("tt_interrupt","utf8_more")=4
	Set gtmtypes("tt_interrupt",5,"name")="tt_interrupt.dx"
	Set gtmtypes("tt_interrupt",5,"off")=28
	Set gtmtypes("tt_interrupt",5,"len")=4
	Set gtmtypes("tt_interrupt",5,"type")="int"
	Set gtmtypfldindx("tt_interrupt","dx")=5
	Set gtmtypes("tt_interrupt",6,"name")="tt_interrupt.dx_start"
	Set gtmtypes("tt_interrupt",6,"off")=32
	Set gtmtypes("tt_interrupt",6,"len")=4
	Set gtmtypes("tt_interrupt",6,"type")="int"
	Set gtmtypfldindx("tt_interrupt","dx_start")=6
	Set gtmtypes("tt_interrupt",7,"name")="tt_interrupt.dx_instr"
	Set gtmtypes("tt_interrupt",7,"off")=36
	Set gtmtypes("tt_interrupt",7,"len")=4
	Set gtmtypes("tt_interrupt",7,"type")="int"
	Set gtmtypfldindx("tt_interrupt","dx_instr")=7
	Set gtmtypes("tt_interrupt",8,"name")="tt_interrupt.dx_outlen"
	Set gtmtypes("tt_interrupt",8,"off")=40
	Set gtmtypes("tt_interrupt",8,"len")=4
	Set gtmtypes("tt_interrupt",8,"type")="int"
	Set gtmtypfldindx("tt_interrupt","dx_outlen")=8
	Set gtmtypes("tt_interrupt",9,"name")="tt_interrupt.instr"
	Set gtmtypes("tt_interrupt",9,"off")=44
	Set gtmtypes("tt_interrupt",9,"len")=4
	Set gtmtypes("tt_interrupt",9,"type")="int"
	Set gtmtypfldindx("tt_interrupt","instr")=9
	Set gtmtypes("tt_interrupt",10,"name")="tt_interrupt.outlen"
	Set gtmtypes("tt_interrupt",10,"off")=48
	Set gtmtypes("tt_interrupt",10,"len")=4
	Set gtmtypes("tt_interrupt",10,"type")="int"
	Set gtmtypfldindx("tt_interrupt","outlen")=10
	Set gtmtypes("tt_interrupt",11,"name")="tt_interrupt.index"
	Set gtmtypes("tt_interrupt",11,"off")=52
	Set gtmtypes("tt_interrupt",11,"len")=4
	Set gtmtypes("tt_interrupt",11,"type")="int"
	Set gtmtypfldindx("tt_interrupt","index")=11
	Set gtmtypes("tt_interrupt",12,"name")="tt_interrupt.cl"
	Set gtmtypes("tt_interrupt",12,"off")=56
	Set gtmtypes("tt_interrupt",12,"len")=4
	Set gtmtypes("tt_interrupt",12,"type")="int"
	Set gtmtypfldindx("tt_interrupt","cl")=12
	Set gtmtypes("tt_interrupt",13,"name")="tt_interrupt.length"
	Set gtmtypes("tt_interrupt",13,"off")=60
	Set gtmtypes("tt_interrupt",13,"len")=4
	Set gtmtypes("tt_interrupt",13,"type")="int"
	Set gtmtypfldindx("tt_interrupt","length")=13
	Set gtmtypes("tt_interrupt",14,"name")="tt_interrupt.exp_length"
	Set gtmtypes("tt_interrupt",14,"off")=64
	Set gtmtypes("tt_interrupt",14,"len")=4
	Set gtmtypes("tt_interrupt",14,"type")="int"
	Set gtmtypfldindx("tt_interrupt","exp_length")=14
	Set gtmtypes("tt_interrupt",15,"name")="tt_interrupt.insert_mode"
	Set gtmtypes("tt_interrupt",15,"off")=68
	Set gtmtypes("tt_interrupt",15,"len")=4
	Set gtmtypes("tt_interrupt",15,"type")="boolean_t"
	Set gtmtypfldindx("tt_interrupt","insert_mode")=15
	Set gtmtypes("tt_interrupt",16,"name")="tt_interrupt.end_time"
	Set gtmtypes("tt_interrupt",16,"off")=72
	Set gtmtypes("tt_interrupt",16,"len")=16
	Set gtmtypes("tt_interrupt",16,"type")="ABS_TIME"
	Set gtmtypfldindx("tt_interrupt","end_time")=16
	Set gtmtypes("tt_interrupt",17,"name")="tt_interrupt.end_time.at_sec"
	Set gtmtypes("tt_interrupt",17,"off")=72
	Set gtmtypes("tt_interrupt",17,"len")=8
	Set gtmtypes("tt_interrupt",17,"type")="long"
	Set gtmtypfldindx("tt_interrupt","end_time.at_sec")=17
	Set gtmtypes("tt_interrupt",18,"name")="tt_interrupt.end_time.at_usec"
	Set gtmtypes("tt_interrupt",18,"off")=80
	Set gtmtypes("tt_interrupt",18,"len")=8
	Set gtmtypes("tt_interrupt",18,"type")="long"
	Set gtmtypfldindx("tt_interrupt","end_time.at_usec")=18
	Set gtmtypes("tt_interrupt",19,"name")="tt_interrupt.more_ptr"
	Set gtmtypes("tt_interrupt",19,"off")=88
	Set gtmtypes("tt_interrupt",19,"len")=8
	Set gtmtypes("tt_interrupt",19,"type")="addr"
	Set gtmtypfldindx("tt_interrupt","more_ptr")=19
	Set gtmtypes("tt_interrupt",20,"name")="tt_interrupt.zb_ptr"
	Set gtmtypes("tt_interrupt",20,"off")=96
	Set gtmtypes("tt_interrupt",20,"len")=8
	Set gtmtypes("tt_interrupt",20,"type")="addr"
	Set gtmtypfldindx("tt_interrupt","zb_ptr")=20
	Set gtmtypes("tt_interrupt",21,"name")="tt_interrupt.zb_top"
	Set gtmtypes("tt_interrupt",21,"off")=104
	Set gtmtypes("tt_interrupt",21,"len")=8
	Set gtmtypes("tt_interrupt",21,"type")="addr"
	Set gtmtypfldindx("tt_interrupt","zb_top")=21
	Set gtmtypes("tt_interrupt",22,"name")="tt_interrupt.escape_length"
	Set gtmtypes("tt_interrupt",22,"off")=112
	Set gtmtypes("tt_interrupt",22,"len")=2
	Set gtmtypes("tt_interrupt",22,"type")="unsigned-short"
	Set gtmtypfldindx("tt_interrupt","escape_length")=22
	Set gtmtypes("tt_interrupt",23,"name")="tt_interrupt.escape_sequence"
	Set gtmtypes("tt_interrupt",23,"off")=114
	Set gtmtypes("tt_interrupt",23,"len")=16
	Set gtmtypes("tt_interrupt",23,"type")="unsigned-char"
	Set gtmtypfldindx("tt_interrupt","escape_sequence")=23
	Set gtmtypes("tt_interrupt",24,"name")="tt_interrupt.more_buf"
	Set gtmtypes("tt_interrupt",24,"off")=130
	Set gtmtypes("tt_interrupt",24,"len")=5
	Set gtmtypes("tt_interrupt",24,"type")="unsigned-char"
	Set gtmtypfldindx("tt_interrupt","more_buf")=24
	;
	Set gtmtypes("ua_list")="struct"
	Set gtmtypes("ua_list",0)=3
	Set gtmtypes("ua_list","len")=24
	Set gtmtypes("ua_list",1,"name")="ua_list.next_ua"
	Set gtmtypes("ua_list",1,"off")=0
	Set gtmtypes("ua_list",1,"len")=8
	Set gtmtypes("ua_list",1,"type")="addr"
	Set gtmtypfldindx("ua_list","next_ua")=1
	Set gtmtypes("ua_list",2,"name")="ua_list.update_array"
	Set gtmtypes("ua_list",2,"off")=8
	Set gtmtypes("ua_list",2,"len")=8
	Set gtmtypes("ua_list",2,"type")="addr"
	Set gtmtypfldindx("ua_list","update_array")=2
	Set gtmtypes("ua_list",3,"name")="ua_list.update_array_size"
	Set gtmtypes("ua_list",3,"off")=16
	Set gtmtypes("ua_list",3,"len")=4
	Set gtmtypes("ua_list",3,"type")="unsigned-int"
	Set gtmtypfldindx("ua_list","update_array_size")=3
	;
	Set gtmtypes("uic_struct_int")="struct"
	Set gtmtypes("uic_struct_int",0)=2
	Set gtmtypes("uic_struct_int","len")=8
	Set gtmtypes("uic_struct_int",1,"name")="uic_struct_int.mem"
	Set gtmtypes("uic_struct_int",1,"off")=0
	Set gtmtypes("uic_struct_int",1,"len")=4
	Set gtmtypes("uic_struct_int",1,"type")="uid_t"
	Set gtmtypfldindx("uic_struct_int","mem")=1
	Set gtmtypes("uic_struct_int",2,"name")="uic_struct_int.grp"
	Set gtmtypes("uic_struct_int",2,"off")=4
	Set gtmtypes("uic_struct_int",2,"len")=4
	Set gtmtypes("uic_struct_int",2,"type")="gid_t"
	Set gtmtypfldindx("uic_struct_int","grp")=2
	;
	Set gtmtypes("unique_file_id")="union"
	Set gtmtypes("unique_file_id",0)=4
	Set gtmtypes("unique_file_id","len")=24
	Set gtmtypes("unique_file_id",1,"name")="unique_file_id.uid"
	Set gtmtypes("unique_file_id",1,"off")=0
	Set gtmtypes("unique_file_id",1,"len")=16
	Set gtmtypes("unique_file_id",1,"type")="unix_file_id"
	Set gtmtypfldindx("unique_file_id","uid")=1
	Set gtmtypes("unique_file_id",2,"name")="unique_file_id.uid.inode"
	Set gtmtypes("unique_file_id",2,"off")=0
	Set gtmtypes("unique_file_id",2,"len")=8
	Set gtmtypes("unique_file_id",2,"type")="ino_t"
	Set gtmtypfldindx("unique_file_id","uid.inode")=2
	Set gtmtypes("unique_file_id",3,"name")="unique_file_id.uid.device"
	Set gtmtypes("unique_file_id",3,"off")=8
	Set gtmtypes("unique_file_id",3,"len")=8
	Set gtmtypes("unique_file_id",3,"type")="dev_t"
	Set gtmtypfldindx("unique_file_id","uid.device")=3
	Set gtmtypes("unique_file_id",4,"name")="unique_file_id.file_id"
	Set gtmtypes("unique_file_id",4,"off")=0
	Set gtmtypes("unique_file_id",4,"len")=20
	Set gtmtypes("unique_file_id",4,"type")="char"
	Set gtmtypfldindx("unique_file_id","file_id")=4
	;
	Set gtmtypes("unix_db_info")="struct"
	Set gtmtypes("unix_db_info",0)=189
	Set gtmtypes("unix_db_info","len")=1248
	Set gtmtypes("unix_db_info",1,"name")="unix_db_info.s_addrs"
	Set gtmtypes("unix_db_info",1,"off")=0
	Set gtmtypes("unix_db_info",1,"len")=1152
	Set gtmtypes("unix_db_info",1,"type")="sgmnt_addrs"
	Set gtmtypfldindx("unix_db_info","s_addrs")=1
	Set gtmtypes("unix_db_info",2,"name")="unix_db_info.s_addrs.hdr"
	Set gtmtypes("unix_db_info",2,"off")=0
	Set gtmtypes("unix_db_info",2,"len")=8
	Set gtmtypes("unix_db_info",2,"type")="addr"
	Set gtmtypfldindx("unix_db_info","s_addrs.hdr")=2
	Set gtmtypes("unix_db_info",3,"name")="unix_db_info.s_addrs.bmm"
	Set gtmtypes("unix_db_info",3,"off")=8
	Set gtmtypes("unix_db_info",3,"len")=8
	Set gtmtypes("unix_db_info",3,"type")="addr"
	Set gtmtypfldindx("unix_db_info","s_addrs.bmm")=3
	Set gtmtypes("unix_db_info",4,"name")="unix_db_info.s_addrs.wc"
	Set gtmtypes("unix_db_info",4,"off")=16
	Set gtmtypes("unix_db_info",4,"len")=8
	Set gtmtypes("unix_db_info",4,"type")="addr"
	Set gtmtypfldindx("unix_db_info","s_addrs.wc")=4
	Set gtmtypes("unix_db_info",5,"name")="unix_db_info.s_addrs.bt_header"
	Set gtmtypes("unix_db_info",5,"off")=24
	Set gtmtypes("unix_db_info",5,"len")=8
	Set gtmtypes("unix_db_info",5,"type")="addr"
	Set gtmtypfldindx("unix_db_info","s_addrs.bt_header")=5
	Set gtmtypes("unix_db_info",6,"name")="unix_db_info.s_addrs.bt_base"
	Set gtmtypes("unix_db_info",6,"off")=32
	Set gtmtypes("unix_db_info",6,"len")=8
	Set gtmtypes("unix_db_info",6,"type")="addr"
	Set gtmtypfldindx("unix_db_info","s_addrs.bt_base")=6
	Set gtmtypes("unix_db_info",7,"name")="unix_db_info.s_addrs.th_base"
	Set gtmtypes("unix_db_info",7,"off")=40
	Set gtmtypes("unix_db_info",7,"len")=8
	Set gtmtypes("unix_db_info",7,"type")="addr"
	Set gtmtypfldindx("unix_db_info","s_addrs.th_base")=7
	Set gtmtypes("unix_db_info",8,"name")="unix_db_info.s_addrs.ti"
	Set gtmtypes("unix_db_info",8,"off")=48
	Set gtmtypes("unix_db_info",8,"len")=8
	Set gtmtypes("unix_db_info",8,"type")="addr"
	Set gtmtypfldindx("unix_db_info","s_addrs.ti")=8
	Set gtmtypes("unix_db_info",9,"name")="unix_db_info.s_addrs.nl"
	Set gtmtypes("unix_db_info",9,"off")=56
	Set gtmtypes("unix_db_info",9,"len")=8
	Set gtmtypes("unix_db_info",9,"type")="addr"
	Set gtmtypfldindx("unix_db_info","s_addrs.nl")=9
	Set gtmtypes("unix_db_info",10,"name")="unix_db_info.s_addrs.critical"
	Set gtmtypes("unix_db_info",10,"off")=64
	Set gtmtypes("unix_db_info",10,"len")=8
	Set gtmtypes("unix_db_info",10,"type")="addr"
	Set gtmtypfldindx("unix_db_info","s_addrs.critical")=10
	Set gtmtypes("unix_db_info",11,"name")="unix_db_info.s_addrs.shmpool_buffer"
	Set gtmtypes("unix_db_info",11,"off")=72
	Set gtmtypes("unix_db_info",11,"len")=8
	Set gtmtypes("unix_db_info",11,"type")="addr"
	Set gtmtypfldindx("unix_db_info","s_addrs.shmpool_buffer")=11
	Set gtmtypes("unix_db_info",12,"name")="unix_db_info.s_addrs.db_addrs"
	Set gtmtypes("unix_db_info",12,"off")=80
	Set gtmtypes("unix_db_info",12,"len")=16
	Set gtmtypes("unix_db_info",12,"type")="addr"
	Set gtmtypfldindx("unix_db_info","s_addrs.db_addrs")=12
	Set gtmtypes("unix_db_info",12,"dim")=2
	Set gtmtypes("unix_db_info",13,"name")="unix_db_info.s_addrs.mlkctl"
	Set gtmtypes("unix_db_info",13,"off")=96
	Set gtmtypes("unix_db_info",13,"len")=8
	Set gtmtypes("unix_db_info",13,"type")="addr"
	Set gtmtypfldindx("unix_db_info","s_addrs.mlkctl")=13
	Set gtmtypes("unix_db_info",14,"name")="unix_db_info.s_addrs.mlkctl_len"
	Set gtmtypes("unix_db_info",14,"off")=104
	Set gtmtypes("unix_db_info",14,"len")=4
	Set gtmtypes("unix_db_info",14,"type")="unsigned-int"
	Set gtmtypfldindx("unix_db_info","s_addrs.mlkctl_len")=14
	Set gtmtypes("unix_db_info",15,"name")="unix_db_info.s_addrs.dir_tree"
	Set gtmtypes("unix_db_info",15,"off")=112
	Set gtmtypes("unix_db_info",15,"len")=8
	Set gtmtypes("unix_db_info",15,"type")="addr"
	Set gtmtypfldindx("unix_db_info","s_addrs.dir_tree")=15
	Set gtmtypes("unix_db_info",16,"name")="unix_db_info.s_addrs.hasht_tree"
	Set gtmtypes("unix_db_info",16,"off")=120
	Set gtmtypes("unix_db_info",16,"len")=8
	Set gtmtypes("unix_db_info",16,"type")="addr"
	Set gtmtypfldindx("unix_db_info","s_addrs.hasht_tree")=16
	Set gtmtypes("unix_db_info",17,"name")="unix_db_info.s_addrs.next_fenced"
	Set gtmtypes("unix_db_info",17,"off")=128
	Set gtmtypes("unix_db_info",17,"len")=8
	Set gtmtypes("unix_db_info",17,"type")="addr"
	Set gtmtypfldindx("unix_db_info","s_addrs.next_fenced")=17
	Set gtmtypes("unix_db_info",18,"name")="unix_db_info.s_addrs.jnl"
	Set gtmtypes("unix_db_info",18,"off")=136
	Set gtmtypes("unix_db_info",18,"len")=8
	Set gtmtypes("unix_db_info",18,"type")="addr"
	Set gtmtypfldindx("unix_db_info","s_addrs.jnl")=18
	Set gtmtypes("unix_db_info",19,"name")="unix_db_info.s_addrs.sgm_info_ptr"
	Set gtmtypes("unix_db_info",19,"off")=144
	Set gtmtypes("unix_db_info",19,"len")=8
	Set gtmtypes("unix_db_info",19,"type")="addr"
	Set gtmtypfldindx("unix_db_info","s_addrs.sgm_info_ptr")=19
	Set gtmtypes("unix_db_info",20,"name")="unix_db_info.s_addrs.region"
	Set gtmtypes("unix_db_info",20,"off")=152
	Set gtmtypes("unix_db_info",20,"len")=8
	Set gtmtypes("unix_db_info",20,"type")="addr"
	Set gtmtypfldindx("unix_db_info","s_addrs.region")=20
	Set gtmtypes("unix_db_info",21,"name")="unix_db_info.s_addrs.gvt_hashtab"
	Set gtmtypes("unix_db_info",21,"off")=160
	Set gtmtypes("unix_db_info",21,"len")=8
	Set gtmtypes("unix_db_info",21,"type")="addr"
	Set gtmtypfldindx("unix_db_info","s_addrs.gvt_hashtab")=21
	Set gtmtypes("unix_db_info",22,"name")="unix_db_info.s_addrs.miscptr"
	Set gtmtypes("unix_db_info",22,"off")=168
	Set gtmtypes("unix_db_info",22,"len")=8
	Set gtmtypes("unix_db_info",22,"type")="addr"
	Set gtmtypfldindx("unix_db_info","s_addrs.miscptr")=22
	Set gtmtypes("unix_db_info",23,"name")="unix_db_info.s_addrs.next_csa"
	Set gtmtypes("unix_db_info",23,"off")=176
	Set gtmtypes("unix_db_info",23,"len")=8
	Set gtmtypes("unix_db_info",23,"type")="addr"
	Set gtmtypfldindx("unix_db_info","s_addrs.next_csa")=23
	Set gtmtypes("unix_db_info",24,"name")="unix_db_info.s_addrs.encr_key_handle"
	Set gtmtypes("unix_db_info",24,"off")=184
	Set gtmtypes("unix_db_info",24,"len")=8
	Set gtmtypes("unix_db_info",24,"type")="addr"
	Set gtmtypfldindx("unix_db_info","s_addrs.encr_key_handle")=24
	Set gtmtypes("unix_db_info",25,"name")="unix_db_info.s_addrs.encr_key_handle2"
	Set gtmtypes("unix_db_info",25,"off")=192
	Set gtmtypes("unix_db_info",25,"len")=8
	Set gtmtypes("unix_db_info",25,"type")="addr"
	Set gtmtypfldindx("unix_db_info","s_addrs.encr_key_handle2")=25
	Set gtmtypes("unix_db_info",26,"name")="unix_db_info.s_addrs.encr_ptr"
	Set gtmtypes("unix_db_info",26,"off")=200
	Set gtmtypes("unix_db_info",26,"len")=8
	Set gtmtypes("unix_db_info",26,"type")="addr"
	Set gtmtypfldindx("unix_db_info","s_addrs.encr_ptr")=26
	Set gtmtypes("unix_db_info",27,"name")="unix_db_info.s_addrs.ss_ctx"
	Set gtmtypes("unix_db_info",27,"off")=208
	Set gtmtypes("unix_db_info",27,"len")=8
	Set gtmtypes("unix_db_info",27,"type")="addr"
	Set gtmtypfldindx("unix_db_info","s_addrs.ss_ctx")=27
	Set gtmtypes("unix_db_info",28,"name")="unix_db_info.s_addrs.acc_meth"
	Set gtmtypes("unix_db_info",28,"off")=216
	Set gtmtypes("unix_db_info",28,"len")=8
	Set gtmtypes("unix_db_info",28,"type")="union"
	Set gtmtypfldindx("unix_db_info","s_addrs.acc_meth")=28
	Set gtmtypes("unix_db_info",29,"name")="unix_db_info.s_addrs.acc_meth.mm"
	Set gtmtypes("unix_db_info",29,"off")=216
	Set gtmtypes("unix_db_info",29,"len")=4
	Set gtmtypes("unix_db_info",29,"type")="sgmm_addrs"
	Set gtmtypfldindx("unix_db_info","s_addrs.acc_meth.mm")=29
	Set gtmtypes("unix_db_info",30,"name")="unix_db_info.s_addrs.acc_meth.mm.filler"
	Set gtmtypes("unix_db_info",30,"off")=216
	Set gtmtypes("unix_db_info",30,"len")=4
	Set gtmtypes("unix_db_info",30,"type")="int"
	Set gtmtypfldindx("unix_db_info","s_addrs.acc_meth.mm.filler")=30
	Set gtmtypes("unix_db_info",31,"name")="unix_db_info.s_addrs.acc_meth.bg"
	Set gtmtypes("unix_db_info",31,"off")=216
	Set gtmtypes("unix_db_info",31,"len")=8
	Set gtmtypes("unix_db_info",31,"type")="sgbg_addrs"
	Set gtmtypfldindx("unix_db_info","s_addrs.acc_meth.bg")=31
	Set gtmtypes("unix_db_info",32,"name")="unix_db_info.s_addrs.acc_meth.bg.cache_state"
	Set gtmtypes("unix_db_info",32,"off")=216
	Set gtmtypes("unix_db_info",32,"len")=8
	Set gtmtypes("unix_db_info",32,"type")="addr"
	Set gtmtypfldindx("unix_db_info","s_addrs.acc_meth.bg.cache_state")=32
	Set gtmtypes("unix_db_info",33,"name")="unix_db_info.s_addrs.gvstats_rec_p"
	Set gtmtypes("unix_db_info",33,"off")=224
	Set gtmtypes("unix_db_info",33,"len")=8
	Set gtmtypes("unix_db_info",33,"type")="addr"
	Set gtmtypfldindx("unix_db_info","s_addrs.gvstats_rec_p")=33
	Set gtmtypes("unix_db_info",34,"name")="unix_db_info.s_addrs.gvstats_rec"
	Set gtmtypes("unix_db_info",34,"off")=232
	Set gtmtypes("unix_db_info",34,"len")=592
	Set gtmtypes("unix_db_info",34,"type")="gvstats_rec_t"
	Set gtmtypfldindx("unix_db_info","s_addrs.gvstats_rec")=34
	Set gtmtypes("unix_db_info",35,"name")="unix_db_info.s_addrs.gvstats_rec.n_set"
	Set gtmtypes("unix_db_info",35,"off")=232
	Set gtmtypes("unix_db_info",35,"len")=8
	Set gtmtypes("unix_db_info",35,"type")="uint64_t"
	Set gtmtypfldindx("unix_db_info","s_addrs.gvstats_rec.n_set")=35
	Set gtmtypes("unix_db_info",36,"name")="unix_db_info.s_addrs.gvstats_rec.n_kill"
	Set gtmtypes("unix_db_info",36,"off")=240
	Set gtmtypes("unix_db_info",36,"len")=8
	Set gtmtypes("unix_db_info",36,"type")="uint64_t"
	Set gtmtypfldindx("unix_db_info","s_addrs.gvstats_rec.n_kill")=36
	Set gtmtypes("unix_db_info",37,"name")="unix_db_info.s_addrs.gvstats_rec.n_get"
	Set gtmtypes("unix_db_info",37,"off")=248
	Set gtmtypes("unix_db_info",37,"len")=8
	Set gtmtypes("unix_db_info",37,"type")="uint64_t"
	Set gtmtypfldindx("unix_db_info","s_addrs.gvstats_rec.n_get")=37
	Set gtmtypes("unix_db_info",38,"name")="unix_db_info.s_addrs.gvstats_rec.n_data"
	Set gtmtypes("unix_db_info",38,"off")=256
	Set gtmtypes("unix_db_info",38,"len")=8
	Set gtmtypes("unix_db_info",38,"type")="uint64_t"
	Set gtmtypfldindx("unix_db_info","s_addrs.gvstats_rec.n_data")=38
	Set gtmtypes("unix_db_info",39,"name")="unix_db_info.s_addrs.gvstats_rec.n_order"
	Set gtmtypes("unix_db_info",39,"off")=264
	Set gtmtypes("unix_db_info",39,"len")=8
	Set gtmtypes("unix_db_info",39,"type")="uint64_t"
	Set gtmtypfldindx("unix_db_info","s_addrs.gvstats_rec.n_order")=39
	Set gtmtypes("unix_db_info",40,"name")="unix_db_info.s_addrs.gvstats_rec.n_zprev"
	Set gtmtypes("unix_db_info",40,"off")=272
	Set gtmtypes("unix_db_info",40,"len")=8
	Set gtmtypes("unix_db_info",40,"type")="uint64_t"
	Set gtmtypfldindx("unix_db_info","s_addrs.gvstats_rec.n_zprev")=40
	Set gtmtypes("unix_db_info",41,"name")="unix_db_info.s_addrs.gvstats_rec.n_query"
	Set gtmtypes("unix_db_info",41,"off")=280
	Set gtmtypes("unix_db_info",41,"len")=8
	Set gtmtypes("unix_db_info",41,"type")="uint64_t"
	Set gtmtypfldindx("unix_db_info","s_addrs.gvstats_rec.n_query")=41
	Set gtmtypes("unix_db_info",42,"name")="unix_db_info.s_addrs.gvstats_rec.n_lock_success"
	Set gtmtypes("unix_db_info",42,"off")=288
	Set gtmtypes("unix_db_info",42,"len")=8
	Set gtmtypes("unix_db_info",42,"type")="uint64_t"
	Set gtmtypfldindx("unix_db_info","s_addrs.gvstats_rec.n_lock_success")=42
	Set gtmtypes("unix_db_info",43,"name")="unix_db_info.s_addrs.gvstats_rec.n_lock_fail"
	Set gtmtypes("unix_db_info",43,"off")=296
	Set gtmtypes("unix_db_info",43,"len")=8
	Set gtmtypes("unix_db_info",43,"type")="uint64_t"
	Set gtmtypfldindx("unix_db_info","s_addrs.gvstats_rec.n_lock_fail")=43
	Set gtmtypes("unix_db_info",44,"name")="unix_db_info.s_addrs.gvstats_rec.db_curr_tn"
	Set gtmtypes("unix_db_info",44,"off")=304
	Set gtmtypes("unix_db_info",44,"len")=8
	Set gtmtypes("unix_db_info",44,"type")="uint64_t"
	Set gtmtypfldindx("unix_db_info","s_addrs.gvstats_rec.db_curr_tn")=44
	Set gtmtypes("unix_db_info",45,"name")="unix_db_info.s_addrs.gvstats_rec.n_dsk_read"
	Set gtmtypes("unix_db_info",45,"off")=312
	Set gtmtypes("unix_db_info",45,"len")=8
	Set gtmtypes("unix_db_info",45,"type")="uint64_t"
	Set gtmtypfldindx("unix_db_info","s_addrs.gvstats_rec.n_dsk_read")=45
	Set gtmtypes("unix_db_info",46,"name")="unix_db_info.s_addrs.gvstats_rec.n_dsk_write"
	Set gtmtypes("unix_db_info",46,"off")=320
	Set gtmtypes("unix_db_info",46,"len")=8
	Set gtmtypes("unix_db_info",46,"type")="uint64_t"
	Set gtmtypfldindx("unix_db_info","s_addrs.gvstats_rec.n_dsk_write")=46
	Set gtmtypes("unix_db_info",47,"name")="unix_db_info.s_addrs.gvstats_rec.n_nontp_readwrite"
	Set gtmtypes("unix_db_info",47,"off")=328
	Set gtmtypes("unix_db_info",47,"len")=8
	Set gtmtypes("unix_db_info",47,"type")="uint64_t"
	Set gtmtypfldindx("unix_db_info","s_addrs.gvstats_rec.n_nontp_readwrite")=47
	Set gtmtypes("unix_db_info",48,"name")="unix_db_info.s_addrs.gvstats_rec.n_nontp_readonly"
	Set gtmtypes("unix_db_info",48,"off")=336
	Set gtmtypes("unix_db_info",48,"len")=8
	Set gtmtypes("unix_db_info",48,"type")="uint64_t"
	Set gtmtypfldindx("unix_db_info","s_addrs.gvstats_rec.n_nontp_readonly")=48
	Set gtmtypes("unix_db_info",49,"name")="unix_db_info.s_addrs.gvstats_rec.n_nontp_blkwrite"
	Set gtmtypes("unix_db_info",49,"off")=344
	Set gtmtypes("unix_db_info",49,"len")=8
	Set gtmtypes("unix_db_info",49,"type")="uint64_t"
	Set gtmtypfldindx("unix_db_info","s_addrs.gvstats_rec.n_nontp_blkwrite")=49
	Set gtmtypes("unix_db_info",50,"name")="unix_db_info.s_addrs.gvstats_rec.n_nontp_blkread"
	Set gtmtypes("unix_db_info",50,"off")=352
	Set gtmtypes("unix_db_info",50,"len")=8
	Set gtmtypes("unix_db_info",50,"type")="uint64_t"
	Set gtmtypfldindx("unix_db_info","s_addrs.gvstats_rec.n_nontp_blkread")=50
	Set gtmtypes("unix_db_info",51,"name")="unix_db_info.s_addrs.gvstats_rec.n_nontp_retries_0"
	Set gtmtypes("unix_db_info",51,"off")=360
	Set gtmtypes("unix_db_info",51,"len")=8
	Set gtmtypes("unix_db_info",51,"type")="uint64_t"
	Set gtmtypfldindx("unix_db_info","s_addrs.gvstats_rec.n_nontp_retries_0")=51
	Set gtmtypes("unix_db_info",52,"name")="unix_db_info.s_addrs.gvstats_rec.n_nontp_retries_1"
	Set gtmtypes("unix_db_info",52,"off")=368
	Set gtmtypes("unix_db_info",52,"len")=8
	Set gtmtypes("unix_db_info",52,"type")="uint64_t"
	Set gtmtypfldindx("unix_db_info","s_addrs.gvstats_rec.n_nontp_retries_1")=52
	Set gtmtypes("unix_db_info",53,"name")="unix_db_info.s_addrs.gvstats_rec.n_nontp_retries_2"
	Set gtmtypes("unix_db_info",53,"off")=376
	Set gtmtypes("unix_db_info",53,"len")=8
	Set gtmtypes("unix_db_info",53,"type")="uint64_t"
	Set gtmtypfldindx("unix_db_info","s_addrs.gvstats_rec.n_nontp_retries_2")=53
	Set gtmtypes("unix_db_info",54,"name")="unix_db_info.s_addrs.gvstats_rec.n_nontp_retries_3"
	Set gtmtypes("unix_db_info",54,"off")=384
	Set gtmtypes("unix_db_info",54,"len")=8
	Set gtmtypes("unix_db_info",54,"type")="uint64_t"
	Set gtmtypfldindx("unix_db_info","s_addrs.gvstats_rec.n_nontp_retries_3")=54
	Set gtmtypes("unix_db_info",55,"name")="unix_db_info.s_addrs.gvstats_rec.n_tp_readwrite"
	Set gtmtypes("unix_db_info",55,"off")=392
	Set gtmtypes("unix_db_info",55,"len")=8
	Set gtmtypes("unix_db_info",55,"type")="uint64_t"
	Set gtmtypfldindx("unix_db_info","s_addrs.gvstats_rec.n_tp_readwrite")=55
	Set gtmtypes("unix_db_info",56,"name")="unix_db_info.s_addrs.gvstats_rec.n_tp_readonly"
	Set gtmtypes("unix_db_info",56,"off")=400
	Set gtmtypes("unix_db_info",56,"len")=8
	Set gtmtypes("unix_db_info",56,"type")="uint64_t"
	Set gtmtypfldindx("unix_db_info","s_addrs.gvstats_rec.n_tp_readonly")=56
	Set gtmtypes("unix_db_info",57,"name")="unix_db_info.s_addrs.gvstats_rec.n_tp_rolledback"
	Set gtmtypes("unix_db_info",57,"off")=408
	Set gtmtypes("unix_db_info",57,"len")=8
	Set gtmtypes("unix_db_info",57,"type")="uint64_t"
	Set gtmtypfldindx("unix_db_info","s_addrs.gvstats_rec.n_tp_rolledback")=57
	Set gtmtypes("unix_db_info",58,"name")="unix_db_info.s_addrs.gvstats_rec.n_tp_blkwrite"
	Set gtmtypes("unix_db_info",58,"off")=416
	Set gtmtypes("unix_db_info",58,"len")=8
	Set gtmtypes("unix_db_info",58,"type")="uint64_t"
	Set gtmtypfldindx("unix_db_info","s_addrs.gvstats_rec.n_tp_blkwrite")=58
	Set gtmtypes("unix_db_info",59,"name")="unix_db_info.s_addrs.gvstats_rec.n_tp_blkread"
	Set gtmtypes("unix_db_info",59,"off")=424
	Set gtmtypes("unix_db_info",59,"len")=8
	Set gtmtypes("unix_db_info",59,"type")="uint64_t"
	Set gtmtypfldindx("unix_db_info","s_addrs.gvstats_rec.n_tp_blkread")=59
	Set gtmtypes("unix_db_info",60,"name")="unix_db_info.s_addrs.gvstats_rec.n_tp_tot_retries_0"
	Set gtmtypes("unix_db_info",60,"off")=432
	Set gtmtypes("unix_db_info",60,"len")=8
	Set gtmtypes("unix_db_info",60,"type")="uint64_t"
	Set gtmtypfldindx("unix_db_info","s_addrs.gvstats_rec.n_tp_tot_retries_0")=60
	Set gtmtypes("unix_db_info",61,"name")="unix_db_info.s_addrs.gvstats_rec.n_tp_tot_retries_1"
	Set gtmtypes("unix_db_info",61,"off")=440
	Set gtmtypes("unix_db_info",61,"len")=8
	Set gtmtypes("unix_db_info",61,"type")="uint64_t"
	Set gtmtypfldindx("unix_db_info","s_addrs.gvstats_rec.n_tp_tot_retries_1")=61
	Set gtmtypes("unix_db_info",62,"name")="unix_db_info.s_addrs.gvstats_rec.n_tp_tot_retries_2"
	Set gtmtypes("unix_db_info",62,"off")=448
	Set gtmtypes("unix_db_info",62,"len")=8
	Set gtmtypes("unix_db_info",62,"type")="uint64_t"
	Set gtmtypfldindx("unix_db_info","s_addrs.gvstats_rec.n_tp_tot_retries_2")=62
	Set gtmtypes("unix_db_info",63,"name")="unix_db_info.s_addrs.gvstats_rec.n_tp_tot_retries_3"
	Set gtmtypes("unix_db_info",63,"off")=456
	Set gtmtypes("unix_db_info",63,"len")=8
	Set gtmtypes("unix_db_info",63,"type")="uint64_t"
	Set gtmtypfldindx("unix_db_info","s_addrs.gvstats_rec.n_tp_tot_retries_3")=63
	Set gtmtypes("unix_db_info",64,"name")="unix_db_info.s_addrs.gvstats_rec.n_tp_tot_retries_4"
	Set gtmtypes("unix_db_info",64,"off")=464
	Set gtmtypes("unix_db_info",64,"len")=8
	Set gtmtypes("unix_db_info",64,"type")="uint64_t"
	Set gtmtypfldindx("unix_db_info","s_addrs.gvstats_rec.n_tp_tot_retries_4")=64
	Set gtmtypes("unix_db_info",65,"name")="unix_db_info.s_addrs.gvstats_rec.n_tp_cnflct_retries_0"
	Set gtmtypes("unix_db_info",65,"off")=472
	Set gtmtypes("unix_db_info",65,"len")=8
	Set gtmtypes("unix_db_info",65,"type")="uint64_t"
	Set gtmtypfldindx("unix_db_info","s_addrs.gvstats_rec.n_tp_cnflct_retries_0")=65
	Set gtmtypes("unix_db_info",66,"name")="unix_db_info.s_addrs.gvstats_rec.n_tp_cnflct_retries_1"
	Set gtmtypes("unix_db_info",66,"off")=480
	Set gtmtypes("unix_db_info",66,"len")=8
	Set gtmtypes("unix_db_info",66,"type")="uint64_t"
	Set gtmtypfldindx("unix_db_info","s_addrs.gvstats_rec.n_tp_cnflct_retries_1")=66
	Set gtmtypes("unix_db_info",67,"name")="unix_db_info.s_addrs.gvstats_rec.n_tp_cnflct_retries_2"
	Set gtmtypes("unix_db_info",67,"off")=488
	Set gtmtypes("unix_db_info",67,"len")=8
	Set gtmtypes("unix_db_info",67,"type")="uint64_t"
	Set gtmtypfldindx("unix_db_info","s_addrs.gvstats_rec.n_tp_cnflct_retries_2")=67
	Set gtmtypes("unix_db_info",68,"name")="unix_db_info.s_addrs.gvstats_rec.n_tp_cnflct_retries_3"
	Set gtmtypes("unix_db_info",68,"off")=496
	Set gtmtypes("unix_db_info",68,"len")=8
	Set gtmtypes("unix_db_info",68,"type")="uint64_t"
	Set gtmtypfldindx("unix_db_info","s_addrs.gvstats_rec.n_tp_cnflct_retries_3")=68
	Set gtmtypes("unix_db_info",69,"name")="unix_db_info.s_addrs.gvstats_rec.n_tp_cnflct_retries_4"
	Set gtmtypes("unix_db_info",69,"off")=504
	Set gtmtypes("unix_db_info",69,"len")=8
	Set gtmtypes("unix_db_info",69,"type")="uint64_t"
	Set gtmtypfldindx("unix_db_info","s_addrs.gvstats_rec.n_tp_cnflct_retries_4")=69
	Set gtmtypes("unix_db_info",70,"name")="unix_db_info.s_addrs.gvstats_rec.n_ztrigger"
	Set gtmtypes("unix_db_info",70,"off")=512
	Set gtmtypes("unix_db_info",70,"len")=8
	Set gtmtypes("unix_db_info",70,"type")="uint64_t"
	Set gtmtypfldindx("unix_db_info","s_addrs.gvstats_rec.n_ztrigger")=70
	Set gtmtypes("unix_db_info",71,"name")="unix_db_info.s_addrs.gvstats_rec.n_db_flush"
	Set gtmtypes("unix_db_info",71,"off")=520
	Set gtmtypes("unix_db_info",71,"len")=8
	Set gtmtypes("unix_db_info",71,"type")="uint64_t"
	Set gtmtypfldindx("unix_db_info","s_addrs.gvstats_rec.n_db_flush")=71
	Set gtmtypes("unix_db_info",72,"name")="unix_db_info.s_addrs.gvstats_rec.n_db_fsync"
	Set gtmtypes("unix_db_info",72,"off")=528
	Set gtmtypes("unix_db_info",72,"len")=8
	Set gtmtypes("unix_db_info",72,"type")="uint64_t"
	Set gtmtypfldindx("unix_db_info","s_addrs.gvstats_rec.n_db_fsync")=72
	Set gtmtypes("unix_db_info",73,"name")="unix_db_info.s_addrs.gvstats_rec.n_jnl_flush"
	Set gtmtypes("unix_db_info",73,"off")=536
	Set gtmtypes("unix_db_info",73,"len")=8
	Set gtmtypes("unix_db_info",73,"type")="uint64_t"
	Set gtmtypfldindx("unix_db_info","s_addrs.gvstats_rec.n_jnl_flush")=73
	Set gtmtypes("unix_db_info",74,"name")="unix_db_info.s_addrs.gvstats_rec.n_jnl_fsync"
	Set gtmtypes("unix_db_info",74,"off")=544
	Set gtmtypes("unix_db_info",74,"len")=8
	Set gtmtypes("unix_db_info",74,"type")="uint64_t"
	Set gtmtypfldindx("unix_db_info","s_addrs.gvstats_rec.n_jnl_fsync")=74
	Set gtmtypes("unix_db_info",75,"name")="unix_db_info.s_addrs.gvstats_rec.n_jbuff_bytes"
	Set gtmtypes("unix_db_info",75,"off")=552
	Set gtmtypes("unix_db_info",75,"len")=8
	Set gtmtypes("unix_db_info",75,"type")="uint64_t"
	Set gtmtypfldindx("unix_db_info","s_addrs.gvstats_rec.n_jbuff_bytes")=75
	Set gtmtypes("unix_db_info",76,"name")="unix_db_info.s_addrs.gvstats_rec.n_jfile_bytes"
	Set gtmtypes("unix_db_info",76,"off")=560
	Set gtmtypes("unix_db_info",76,"len")=8
	Set gtmtypes("unix_db_info",76,"type")="uint64_t"
	Set gtmtypfldindx("unix_db_info","s_addrs.gvstats_rec.n_jfile_bytes")=76
	Set gtmtypes("unix_db_info",77,"name")="unix_db_info.s_addrs.gvstats_rec.n_jfile_writes"
	Set gtmtypes("unix_db_info",77,"off")=568
	Set gtmtypes("unix_db_info",77,"len")=8
	Set gtmtypes("unix_db_info",77,"type")="uint64_t"
	Set gtmtypfldindx("unix_db_info","s_addrs.gvstats_rec.n_jfile_writes")=77
	Set gtmtypes("unix_db_info",78,"name")="unix_db_info.s_addrs.gvstats_rec.n_jrec_logical"
	Set gtmtypes("unix_db_info",78,"off")=576
	Set gtmtypes("unix_db_info",78,"len")=8
	Set gtmtypes("unix_db_info",78,"type")="uint64_t"
	Set gtmtypfldindx("unix_db_info","s_addrs.gvstats_rec.n_jrec_logical")=78
	Set gtmtypes("unix_db_info",79,"name")="unix_db_info.s_addrs.gvstats_rec.n_jrec_pblk"
	Set gtmtypes("unix_db_info",79,"off")=584
	Set gtmtypes("unix_db_info",79,"len")=8
	Set gtmtypes("unix_db_info",79,"type")="uint64_t"
	Set gtmtypfldindx("unix_db_info","s_addrs.gvstats_rec.n_jrec_pblk")=79
	Set gtmtypes("unix_db_info",80,"name")="unix_db_info.s_addrs.gvstats_rec.n_jrec_epoch_regular"
	Set gtmtypes("unix_db_info",80,"off")=592
	Set gtmtypes("unix_db_info",80,"len")=8
	Set gtmtypes("unix_db_info",80,"type")="uint64_t"
	Set gtmtypfldindx("unix_db_info","s_addrs.gvstats_rec.n_jrec_epoch_regular")=80
	Set gtmtypes("unix_db_info",81,"name")="unix_db_info.s_addrs.gvstats_rec.n_jrec_epoch_idle"
	Set gtmtypes("unix_db_info",81,"off")=600
	Set gtmtypes("unix_db_info",81,"len")=8
	Set gtmtypes("unix_db_info",81,"type")="uint64_t"
	Set gtmtypfldindx("unix_db_info","s_addrs.gvstats_rec.n_jrec_epoch_idle")=81
	Set gtmtypes("unix_db_info",82,"name")="unix_db_info.s_addrs.gvstats_rec.n_jrec_other"
	Set gtmtypes("unix_db_info",82,"off")=608
	Set gtmtypes("unix_db_info",82,"len")=8
	Set gtmtypes("unix_db_info",82,"type")="uint64_t"
	Set gtmtypfldindx("unix_db_info","s_addrs.gvstats_rec.n_jrec_other")=82
	Set gtmtypes("unix_db_info",83,"name")="unix_db_info.s_addrs.gvstats_rec.n_jnl_extends"
	Set gtmtypes("unix_db_info",83,"off")=616
	Set gtmtypes("unix_db_info",83,"len")=8
	Set gtmtypes("unix_db_info",83,"type")="uint64_t"
	Set gtmtypfldindx("unix_db_info","s_addrs.gvstats_rec.n_jnl_extends")=83
	Set gtmtypes("unix_db_info",84,"name")="unix_db_info.s_addrs.gvstats_rec.n_db_extends"
	Set gtmtypes("unix_db_info",84,"off")=624
	Set gtmtypes("unix_db_info",84,"len")=8
	Set gtmtypes("unix_db_info",84,"type")="uint64_t"
	Set gtmtypfldindx("unix_db_info","s_addrs.gvstats_rec.n_db_extends")=84
	Set gtmtypes("unix_db_info",85,"name")="unix_db_info.s_addrs.gvstats_rec.n_crit_success"
	Set gtmtypes("unix_db_info",85,"off")=632
	Set gtmtypes("unix_db_info",85,"len")=8
	Set gtmtypes("unix_db_info",85,"type")="uint64_t"
	Set gtmtypfldindx("unix_db_info","s_addrs.gvstats_rec.n_crit_success")=85
	Set gtmtypes("unix_db_info",86,"name")="unix_db_info.s_addrs.gvstats_rec.n_crits_in_epch"
	Set gtmtypes("unix_db_info",86,"off")=640
	Set gtmtypes("unix_db_info",86,"len")=8
	Set gtmtypes("unix_db_info",86,"type")="uint64_t"
	Set gtmtypfldindx("unix_db_info","s_addrs.gvstats_rec.n_crits_in_epch")=86
	Set gtmtypes("unix_db_info",87,"name")="unix_db_info.s_addrs.gvstats_rec.sq_crit_failed"
	Set gtmtypes("unix_db_info",87,"off")=648
	Set gtmtypes("unix_db_info",87,"len")=8
	Set gtmtypes("unix_db_info",87,"type")="uint64_t"
	Set gtmtypfldindx("unix_db_info","s_addrs.gvstats_rec.sq_crit_failed")=87
	Set gtmtypes("unix_db_info",88,"name")="unix_db_info.s_addrs.gvstats_rec.n_crit_failed"
	Set gtmtypes("unix_db_info",88,"off")=656
	Set gtmtypes("unix_db_info",88,"len")=8
	Set gtmtypes("unix_db_info",88,"type")="uint64_t"
	Set gtmtypfldindx("unix_db_info","s_addrs.gvstats_rec.n_crit_failed")=88
	Set gtmtypes("unix_db_info",89,"name")="unix_db_info.s_addrs.gvstats_rec.sq_crit_que_slps"
	Set gtmtypes("unix_db_info",89,"off")=664
	Set gtmtypes("unix_db_info",89,"len")=8
	Set gtmtypes("unix_db_info",89,"type")="uint64_t"
	Set gtmtypfldindx("unix_db_info","s_addrs.gvstats_rec.sq_crit_que_slps")=89
	Set gtmtypes("unix_db_info",90,"name")="unix_db_info.s_addrs.gvstats_rec.n_crit_que_slps"
	Set gtmtypes("unix_db_info",90,"off")=672
	Set gtmtypes("unix_db_info",90,"len")=8
	Set gtmtypes("unix_db_info",90,"type")="uint64_t"
	Set gtmtypfldindx("unix_db_info","s_addrs.gvstats_rec.n_crit_que_slps")=90
	Set gtmtypes("unix_db_info",91,"name")="unix_db_info.s_addrs.gvstats_rec.sq_crit_yields"
	Set gtmtypes("unix_db_info",91,"off")=680
	Set gtmtypes("unix_db_info",91,"len")=8
	Set gtmtypes("unix_db_info",91,"type")="uint64_t"
	Set gtmtypfldindx("unix_db_info","s_addrs.gvstats_rec.sq_crit_yields")=91
	Set gtmtypes("unix_db_info",92,"name")="unix_db_info.s_addrs.gvstats_rec.n_crit_yields"
	Set gtmtypes("unix_db_info",92,"off")=688
	Set gtmtypes("unix_db_info",92,"len")=8
	Set gtmtypes("unix_db_info",92,"type")="uint64_t"
	Set gtmtypfldindx("unix_db_info","s_addrs.gvstats_rec.n_crit_yields")=92
	Set gtmtypes("unix_db_info",93,"name")="unix_db_info.s_addrs.gvstats_rec.n_clean2dirty"
	Set gtmtypes("unix_db_info",93,"off")=696
	Set gtmtypes("unix_db_info",93,"len")=8
	Set gtmtypes("unix_db_info",93,"type")="uint64_t"
	Set gtmtypfldindx("unix_db_info","s_addrs.gvstats_rec.n_clean2dirty")=93
	Set gtmtypes("unix_db_info",94,"name")="unix_db_info.s_addrs.gvstats_rec.n_wait_for_read"
	Set gtmtypes("unix_db_info",94,"off")=704
	Set gtmtypes("unix_db_info",94,"len")=8
	Set gtmtypes("unix_db_info",94,"type")="uint64_t"
	Set gtmtypfldindx("unix_db_info","s_addrs.gvstats_rec.n_wait_for_read")=94
	Set gtmtypes("unix_db_info",95,"name")="unix_db_info.s_addrs.gvstats_rec.n_buffer_scarce"
	Set gtmtypes("unix_db_info",95,"off")=712
	Set gtmtypes("unix_db_info",95,"len")=8
	Set gtmtypes("unix_db_info",95,"type")="uint64_t"
	Set gtmtypfldindx("unix_db_info","s_addrs.gvstats_rec.n_buffer_scarce")=95
	Set gtmtypes("unix_db_info",96,"name")="unix_db_info.s_addrs.gvstats_rec.n_bt_scarce"
	Set gtmtypes("unix_db_info",96,"off")=720
	Set gtmtypes("unix_db_info",96,"len")=8
	Set gtmtypes("unix_db_info",96,"type")="uint64_t"
	Set gtmtypfldindx("unix_db_info","s_addrs.gvstats_rec.n_bt_scarce")=96
	Set gtmtypes("unix_db_info",97,"name")="unix_db_info.s_addrs.gvstats_rec.f_dbext_wait"
	Set gtmtypes("unix_db_info",97,"off")=728
	Set gtmtypes("unix_db_info",97,"len")=8
	Set gtmtypes("unix_db_info",97,"type")="uint64_t"
	Set gtmtypfldindx("unix_db_info","s_addrs.gvstats_rec.f_dbext_wait")=97
	Set gtmtypes("unix_db_info",98,"name")="unix_db_info.s_addrs.gvstats_rec.f_bg_wait"
	Set gtmtypes("unix_db_info",98,"off")=736
	Set gtmtypes("unix_db_info",98,"len")=8
	Set gtmtypes("unix_db_info",98,"type")="uint64_t"
	Set gtmtypfldindx("unix_db_info","s_addrs.gvstats_rec.f_bg_wait")=98
	Set gtmtypes("unix_db_info",99,"name")="unix_db_info.s_addrs.gvstats_rec.f_jnl_wait"
	Set gtmtypes("unix_db_info",99,"off")=744
	Set gtmtypes("unix_db_info",99,"len")=8
	Set gtmtypes("unix_db_info",99,"type")="uint64_t"
	Set gtmtypfldindx("unix_db_info","s_addrs.gvstats_rec.f_jnl_wait")=99
	Set gtmtypes("unix_db_info",100,"name")="unix_db_info.s_addrs.gvstats_rec.f_mlk_wait"
	Set gtmtypes("unix_db_info",100,"off")=752
	Set gtmtypes("unix_db_info",100,"len")=8
	Set gtmtypes("unix_db_info",100,"type")="uint64_t"
	Set gtmtypfldindx("unix_db_info","s_addrs.gvstats_rec.f_mlk_wait")=100
	Set gtmtypes("unix_db_info",101,"name")="unix_db_info.s_addrs.gvstats_rec.f_proc_wait"
	Set gtmtypes("unix_db_info",101,"off")=760
	Set gtmtypes("unix_db_info",101,"len")=8
	Set gtmtypes("unix_db_info",101,"type")="uint64_t"
	Set gtmtypfldindx("unix_db_info","s_addrs.gvstats_rec.f_proc_wait")=101
	Set gtmtypes("unix_db_info",102,"name")="unix_db_info.s_addrs.gvstats_rec.f_trans_wait"
	Set gtmtypes("unix_db_info",102,"off")=768
	Set gtmtypes("unix_db_info",102,"len")=8
	Set gtmtypes("unix_db_info",102,"type")="uint64_t"
	Set gtmtypfldindx("unix_db_info","s_addrs.gvstats_rec.f_trans_wait")=102
	Set gtmtypes("unix_db_info",103,"name")="unix_db_info.s_addrs.gvstats_rec.f_util_wait"
	Set gtmtypes("unix_db_info",103,"off")=776
	Set gtmtypes("unix_db_info",103,"len")=8
	Set gtmtypes("unix_db_info",103,"type")="uint64_t"
	Set gtmtypfldindx("unix_db_info","s_addrs.gvstats_rec.f_util_wait")=103
	Set gtmtypes("unix_db_info",104,"name")="unix_db_info.s_addrs.gvstats_rec.f_ws2"
	Set gtmtypes("unix_db_info",104,"off")=784
	Set gtmtypes("unix_db_info",104,"len")=8
	Set gtmtypes("unix_db_info",104,"type")="uint64_t"
	Set gtmtypfldindx("unix_db_info","s_addrs.gvstats_rec.f_ws2")=104
	Set gtmtypes("unix_db_info",105,"name")="unix_db_info.s_addrs.gvstats_rec.f_ws12"
	Set gtmtypes("unix_db_info",105,"off")=792
	Set gtmtypes("unix_db_info",105,"len")=8
	Set gtmtypes("unix_db_info",105,"type")="uint64_t"
	Set gtmtypfldindx("unix_db_info","s_addrs.gvstats_rec.f_ws12")=105
	Set gtmtypes("unix_db_info",106,"name")="unix_db_info.s_addrs.gvstats_rec.f_ws15"
	Set gtmtypes("unix_db_info",106,"off")=800
	Set gtmtypes("unix_db_info",106,"len")=8
	Set gtmtypes("unix_db_info",106,"type")="uint64_t"
	Set gtmtypfldindx("unix_db_info","s_addrs.gvstats_rec.f_ws15")=106
	Set gtmtypes("unix_db_info",107,"name")="unix_db_info.s_addrs.gvstats_rec.f_ws39"
	Set gtmtypes("unix_db_info",107,"off")=808
	Set gtmtypes("unix_db_info",107,"len")=8
	Set gtmtypes("unix_db_info",107,"type")="uint64_t"
	Set gtmtypfldindx("unix_db_info","s_addrs.gvstats_rec.f_ws39")=107
	Set gtmtypes("unix_db_info",108,"name")="unix_db_info.s_addrs.gvstats_rec.f_ws47"
	Set gtmtypes("unix_db_info",108,"off")=816
	Set gtmtypes("unix_db_info",108,"len")=8
	Set gtmtypes("unix_db_info",108,"type")="uint64_t"
	Set gtmtypfldindx("unix_db_info","s_addrs.gvstats_rec.f_ws47")=108
	Set gtmtypes("unix_db_info",109,"name")="unix_db_info.s_addrs.dbsync_timer_tn"
	Set gtmtypes("unix_db_info",109,"off")=824
	Set gtmtypes("unix_db_info",109,"len")=8
	Set gtmtypes("unix_db_info",109,"type")="uint64_t"
	Set gtmtypfldindx("unix_db_info","s_addrs.dbsync_timer_tn")=109
	Set gtmtypes("unix_db_info",110,"name")="unix_db_info.s_addrs.our_midnite"
	Set gtmtypes("unix_db_info",110,"off")=832
	Set gtmtypes("unix_db_info",110,"len")=8
	Set gtmtypes("unix_db_info",110,"type")="addr"
	Set gtmtypfldindx("unix_db_info","s_addrs.our_midnite")=110
	Set gtmtypes("unix_db_info",111,"name")="unix_db_info.s_addrs.fullblockwrite_len"
	Set gtmtypes("unix_db_info",111,"off")=840
	Set gtmtypes("unix_db_info",111,"len")=8
	Set gtmtypes("unix_db_info",111,"type")="size_t"
	Set gtmtypfldindx("unix_db_info","s_addrs.fullblockwrite_len")=111
	Set gtmtypes("unix_db_info",112,"name")="unix_db_info.s_addrs.our_lru_cache_rec_off"
	Set gtmtypes("unix_db_info",112,"off")=848
	Set gtmtypes("unix_db_info",112,"len")=8
	Set gtmtypes("unix_db_info",112,"type")="intptr_t"
	Set gtmtypfldindx("unix_db_info","s_addrs.our_lru_cache_rec_off")=112
	Set gtmtypes("unix_db_info",113,"name")="unix_db_info.s_addrs.total_blks"
	Set gtmtypes("unix_db_info",113,"off")=856
	Set gtmtypes("unix_db_info",113,"len")=4
	Set gtmtypes("unix_db_info",113,"type")="unsigned-int"
	Set gtmtypfldindx("unix_db_info","s_addrs.total_blks")=113
	Set gtmtypes("unix_db_info",114,"name")="unix_db_info.s_addrs.prev_free_blks"
	Set gtmtypes("unix_db_info",114,"off")=860
	Set gtmtypes("unix_db_info",114,"len")=4
	Set gtmtypes("unix_db_info",114,"type")="unsigned-int"
	Set gtmtypfldindx("unix_db_info","s_addrs.prev_free_blks")=114
	Set gtmtypes("unix_db_info",115,"name")="unix_db_info.s_addrs.timer"
	Set gtmtypes("unix_db_info",115,"off")=864
	Set gtmtypes("unix_db_info",115,"len")=4
	Set gtmtypes("unix_db_info",115,"type")="unsigned-int"
	Set gtmtypfldindx("unix_db_info","s_addrs.timer")=115
	Set gtmtypes("unix_db_info",116,"name")="unix_db_info.s_addrs.in_wtstart"
	Set gtmtypes("unix_db_info",116,"off")=868
	Set gtmtypes("unix_db_info",116,"len")=4
	Set gtmtypes("unix_db_info",116,"type")="unsigned-int"
	Set gtmtypfldindx("unix_db_info","s_addrs.in_wtstart")=116
	Set gtmtypes("unix_db_info",117,"name")="unix_db_info.s_addrs.now_crit"
	Set gtmtypes("unix_db_info",117,"off")=872
	Set gtmtypes("unix_db_info",117,"len")=4
	Set gtmtypes("unix_db_info",117,"type")="unsigned-int"
	Set gtmtypfldindx("unix_db_info","s_addrs.now_crit")=117
	Set gtmtypes("unix_db_info",118,"name")="unix_db_info.s_addrs.wbuf_dqd"
	Set gtmtypes("unix_db_info",118,"off")=876
	Set gtmtypes("unix_db_info",118,"len")=4
	Set gtmtypes("unix_db_info",118,"type")="unsigned-int"
	Set gtmtypfldindx("unix_db_info","s_addrs.wbuf_dqd")=118
	Set gtmtypes("unix_db_info",119,"name")="unix_db_info.s_addrs.stale_defer"
	Set gtmtypes("unix_db_info",119,"off")=880
	Set gtmtypes("unix_db_info",119,"len")=4
	Set gtmtypes("unix_db_info",119,"type")="unsigned-int"
	Set gtmtypfldindx("unix_db_info","s_addrs.stale_defer")=119
	Set gtmtypes("unix_db_info",120,"name")="unix_db_info.s_addrs.freeze"
	Set gtmtypes("unix_db_info",120,"off")=884
	Set gtmtypes("unix_db_info",120,"len")=4
	Set gtmtypes("unix_db_info",120,"type")="boolean_t"
	Set gtmtypfldindx("unix_db_info","s_addrs.freeze")=120
	Set gtmtypes("unix_db_info",121,"name")="unix_db_info.s_addrs.dbsync_timer"
	Set gtmtypes("unix_db_info",121,"off")=888
	Set gtmtypes("unix_db_info",121,"len")=4
	Set gtmtypes("unix_db_info",121,"type")="boolean_t"
	Set gtmtypfldindx("unix_db_info","s_addrs.dbsync_timer")=121
	Set gtmtypes("unix_db_info",122,"name")="unix_db_info.s_addrs.reorg_last_dest"
	Set gtmtypes("unix_db_info",122,"off")=892
	Set gtmtypes("unix_db_info",122,"len")=4
	Set gtmtypes("unix_db_info",122,"type")="int"
	Set gtmtypfldindx("unix_db_info","s_addrs.reorg_last_dest")=122
	Set gtmtypes("unix_db_info",123,"name")="unix_db_info.s_addrs.jnl_before_image"
	Set gtmtypes("unix_db_info",123,"off")=896
	Set gtmtypes("unix_db_info",123,"len")=4
	Set gtmtypes("unix_db_info",123,"type")="boolean_t"
	Set gtmtypfldindx("unix_db_info","s_addrs.jnl_before_image")=123
	Set gtmtypes("unix_db_info",124,"name")="unix_db_info.s_addrs.read_write"
	Set gtmtypes("unix_db_info",124,"off")=900
	Set gtmtypes("unix_db_info",124,"len")=4
	Set gtmtypes("unix_db_info",124,"type")="boolean_t"
	Set gtmtypfldindx("unix_db_info","s_addrs.read_write")=124
	Set gtmtypes("unix_db_info",125,"name")="unix_db_info.s_addrs.orig_read_write"
	Set gtmtypes("unix_db_info",125,"off")=904
	Set gtmtypes("unix_db_info",125,"len")=4
	Set gtmtypes("unix_db_info",125,"type")="boolean_t"
	Set gtmtypfldindx("unix_db_info","s_addrs.orig_read_write")=125
	Set gtmtypes("unix_db_info",126,"name")="unix_db_info.s_addrs.persistent_freeze"
	Set gtmtypes("unix_db_info",126,"off")=908
	Set gtmtypes("unix_db_info",126,"len")=4
	Set gtmtypes("unix_db_info",126,"type")="boolean_t"
	Set gtmtypfldindx("unix_db_info","s_addrs.persistent_freeze")=126
	Set gtmtypes("unix_db_info",127,"name")="unix_db_info.s_addrs.pblk_align_jrecsize"
	Set gtmtypes("unix_db_info",127,"off")=912
	Set gtmtypes("unix_db_info",127,"len")=4
	Set gtmtypes("unix_db_info",127,"type")="int"
	Set gtmtypfldindx("unix_db_info","s_addrs.pblk_align_jrecsize")=127
	Set gtmtypes("unix_db_info",128,"name")="unix_db_info.s_addrs.min_total_tpjnl_rec_size"
	Set gtmtypes("unix_db_info",128,"off")=916
	Set gtmtypes("unix_db_info",128,"len")=4
	Set gtmtypes("unix_db_info",128,"type")="int"
	Set gtmtypfldindx("unix_db_info","s_addrs.min_total_tpjnl_rec_size")=128
	Set gtmtypes("unix_db_info",129,"name")="unix_db_info.s_addrs.min_total_nontpjnl_rec_size"
	Set gtmtypes("unix_db_info",129,"off")=920
	Set gtmtypes("unix_db_info",129,"len")=4
	Set gtmtypes("unix_db_info",129,"type")="int"
	Set gtmtypfldindx("unix_db_info","s_addrs.min_total_nontpjnl_rec_size")=129
	Set gtmtypes("unix_db_info",130,"name")="unix_db_info.s_addrs.jnl_state"
	Set gtmtypes("unix_db_info",130,"off")=924
	Set gtmtypes("unix_db_info",130,"len")=4
	Set gtmtypes("unix_db_info",130,"type")="int"
	Set gtmtypfldindx("unix_db_info","s_addrs.jnl_state")=130
	Set gtmtypes("unix_db_info",131,"name")="unix_db_info.s_addrs.repl_state"
	Set gtmtypes("unix_db_info",131,"off")=928
	Set gtmtypes("unix_db_info",131,"len")=4
	Set gtmtypes("unix_db_info",131,"type")="int"
	Set gtmtypfldindx("unix_db_info","s_addrs.repl_state")=131
	Set gtmtypes("unix_db_info",132,"name")="unix_db_info.s_addrs.crit_check_cycle"
	Set gtmtypes("unix_db_info",132,"off")=932
	Set gtmtypes("unix_db_info",132,"len")=4
	Set gtmtypes("unix_db_info",132,"type")="unsigned-int"
	Set gtmtypfldindx("unix_db_info","s_addrs.crit_check_cycle")=132
	Set gtmtypes("unix_db_info",133,"name")="unix_db_info.s_addrs.backup_in_prog"
	Set gtmtypes("unix_db_info",133,"off")=936
	Set gtmtypes("unix_db_info",133,"len")=4
	Set gtmtypes("unix_db_info",133,"type")="int"
	Set gtmtypfldindx("unix_db_info","s_addrs.backup_in_prog")=133
	Set gtmtypes("unix_db_info",134,"name")="unix_db_info.s_addrs.snapshot_in_prog"
	Set gtmtypes("unix_db_info",134,"off")=940
	Set gtmtypes("unix_db_info",134,"len")=4
	Set gtmtypes("unix_db_info",134,"type")="boolean_t"
	Set gtmtypfldindx("unix_db_info","s_addrs.snapshot_in_prog")=134
	Set gtmtypes("unix_db_info",135,"name")="unix_db_info.s_addrs.ref_cnt"
	Set gtmtypes("unix_db_info",135,"off")=944
	Set gtmtypes("unix_db_info",135,"len")=4
	Set gtmtypes("unix_db_info",135,"type")="int"
	Set gtmtypfldindx("unix_db_info","s_addrs.ref_cnt")=135
	Set gtmtypes("unix_db_info",136,"name")="unix_db_info.s_addrs.fid_index"
	Set gtmtypes("unix_db_info",136,"off")=948
	Set gtmtypes("unix_db_info",136,"len")=4
	Set gtmtypes("unix_db_info",136,"type")="int"
	Set gtmtypfldindx("unix_db_info","s_addrs.fid_index")=136
	Set gtmtypes("unix_db_info",137,"name")="unix_db_info.s_addrs.filler"
	Set gtmtypes("unix_db_info",137,"off")=952
	Set gtmtypes("unix_db_info",137,"len")=4
	Set gtmtypes("unix_db_info",137,"type")="char"
	Set gtmtypfldindx("unix_db_info","s_addrs.filler")=137
	Set gtmtypes("unix_db_info",138,"name")="unix_db_info.s_addrs.regnum"
	Set gtmtypes("unix_db_info",138,"off")=956
	Set gtmtypes("unix_db_info",138,"len")=4
	Set gtmtypes("unix_db_info",138,"type")="int"
	Set gtmtypfldindx("unix_db_info","s_addrs.regnum")=138
	Set gtmtypes("unix_db_info",139,"name")="unix_db_info.s_addrs.n_pre_read_trigger"
	Set gtmtypes("unix_db_info",139,"off")=960
	Set gtmtypes("unix_db_info",139,"len")=4
	Set gtmtypes("unix_db_info",139,"type")="int"
	Set gtmtypfldindx("unix_db_info","s_addrs.n_pre_read_trigger")=139
	Set gtmtypes("unix_db_info",140,"name")="unix_db_info.s_addrs.jnlpool_validate_check"
	Set gtmtypes("unix_db_info",140,"off")=964
	Set gtmtypes("unix_db_info",140,"len")=4
	Set gtmtypes("unix_db_info",140,"type")="unsigned-int"
	Set gtmtypfldindx("unix_db_info","s_addrs.jnlpool_validate_check")=140
	Set gtmtypes("unix_db_info",141,"name")="unix_db_info.s_addrs.regcnt"
	Set gtmtypes("unix_db_info",141,"off")=968
	Set gtmtypes("unix_db_info",141,"len")=4
	Set gtmtypes("unix_db_info",141,"type")="int"
	Set gtmtypfldindx("unix_db_info","s_addrs.regcnt")=141
	Set gtmtypes("unix_db_info",142,"name")="unix_db_info.s_addrs.t_commit_crit"
	Set gtmtypes("unix_db_info",142,"off")=972
	Set gtmtypes("unix_db_info",142,"len")=4
	Set gtmtypes("unix_db_info",142,"type")="boolean_t"
	Set gtmtypfldindx("unix_db_info","s_addrs.t_commit_crit")=142
	Set gtmtypes("unix_db_info",143,"name")="unix_db_info.s_addrs.wcs_pidcnt_incremented"
	Set gtmtypes("unix_db_info",143,"off")=976
	Set gtmtypes("unix_db_info",143,"len")=4
	Set gtmtypes("unix_db_info",143,"type")="boolean_t"
	Set gtmtypfldindx("unix_db_info","s_addrs.wcs_pidcnt_incremented")=143
	Set gtmtypes("unix_db_info",144,"name")="unix_db_info.s_addrs.incr_db_trigger_cycle"
	Set gtmtypes("unix_db_info",144,"off")=980
	Set gtmtypes("unix_db_info",144,"len")=4
	Set gtmtypes("unix_db_info",144,"type")="boolean_t"
	Set gtmtypfldindx("unix_db_info","s_addrs.incr_db_trigger_cycle")=144
	Set gtmtypes("unix_db_info",145,"name")="unix_db_info.s_addrs.db_trigger_cycle"
	Set gtmtypes("unix_db_info",145,"off")=984
	Set gtmtypes("unix_db_info",145,"len")=4
	Set gtmtypes("unix_db_info",145,"type")="unsigned-int"
	Set gtmtypfldindx("unix_db_info","s_addrs.db_trigger_cycle")=145
	Set gtmtypes("unix_db_info",146,"name")="unix_db_info.s_addrs.db_dztrigger_cycle"
	Set gtmtypes("unix_db_info",146,"off")=988
	Set gtmtypes("unix_db_info",146,"len")=4
	Set gtmtypes("unix_db_info",146,"type")="unsigned-int"
	Set gtmtypfldindx("unix_db_info","s_addrs.db_dztrigger_cycle")=146
	Set gtmtypes("unix_db_info",147,"name")="unix_db_info.s_addrs.hold_onto_crit"
	Set gtmtypes("unix_db_info",147,"off")=992
	Set gtmtypes("unix_db_info",147,"len")=4
	Set gtmtypes("unix_db_info",147,"type")="boolean_t"
	Set gtmtypfldindx("unix_db_info","s_addrs.hold_onto_crit")=147
	Set gtmtypes("unix_db_info",148,"name")="unix_db_info.s_addrs.dse_crit_seize_done"
	Set gtmtypes("unix_db_info",148,"off")=996
	Set gtmtypes("unix_db_info",148,"len")=4
	Set gtmtypes("unix_db_info",148,"type")="boolean_t"
	Set gtmtypfldindx("unix_db_info","s_addrs.dse_crit_seize_done")=148
	Set gtmtypes("unix_db_info",149,"name")="unix_db_info.s_addrs.gbuff_limit"
	Set gtmtypes("unix_db_info",149,"off")=1000
	Set gtmtypes("unix_db_info",149,"len")=4
	Set gtmtypes("unix_db_info",149,"type")="int"
	Set gtmtypfldindx("unix_db_info","s_addrs.gbuff_limit")=149
	Set gtmtypes("unix_db_info",150,"name")="unix_db_info.s_addrs.root_search_cycle"
	Set gtmtypes("unix_db_info",150,"off")=1004
	Set gtmtypes("unix_db_info",150,"len")=4
	Set gtmtypes("unix_db_info",150,"type")="unsigned-int"
	Set gtmtypfldindx("unix_db_info","s_addrs.root_search_cycle")=150
	Set gtmtypes("unix_db_info",151,"name")="unix_db_info.s_addrs.onln_rlbk_cycle"
	Set gtmtypes("unix_db_info",151,"off")=1008
	Set gtmtypes("unix_db_info",151,"len")=4
	Set gtmtypes("unix_db_info",151,"type")="unsigned-int"
	Set gtmtypfldindx("unix_db_info","s_addrs.onln_rlbk_cycle")=151
	Set gtmtypes("unix_db_info",152,"name")="unix_db_info.s_addrs.db_onln_rlbkd_cycle"
	Set gtmtypes("unix_db_info",152,"off")=1012
	Set gtmtypes("unix_db_info",152,"len")=4
	Set gtmtypes("unix_db_info",152,"type")="unsigned-int"
	Set gtmtypfldindx("unix_db_info","s_addrs.db_onln_rlbkd_cycle")=152
	Set gtmtypes("unix_db_info",153,"name")="unix_db_info.s_addrs.reservedDBFlags"
	Set gtmtypes("unix_db_info",153,"off")=1016
	Set gtmtypes("unix_db_info",153,"len")=4
	Set gtmtypes("unix_db_info",153,"type")="unsigned-int"
	Set gtmtypfldindx("unix_db_info","s_addrs.reservedDBFlags")=153
	Set gtmtypes("unix_db_info",154,"name")="unix_db_info.s_addrs.read_only_fs"
	Set gtmtypes("unix_db_info",154,"off")=1020
	Set gtmtypes("unix_db_info",154,"len")=4
	Set gtmtypes("unix_db_info",154,"type")="boolean_t"
	Set gtmtypfldindx("unix_db_info","s_addrs.read_only_fs")=154
	Set gtmtypes("unix_db_info",155,"name")="unix_db_info.s_addrs.crit_probe"
	Set gtmtypes("unix_db_info",155,"off")=1024
	Set gtmtypes("unix_db_info",155,"len")=4
	Set gtmtypes("unix_db_info",155,"type")="boolean_t"
	Set gtmtypfldindx("unix_db_info","s_addrs.crit_probe")=155
	Set gtmtypes("unix_db_info",156,"name")="unix_db_info.s_addrs.canceled_flush_timer"
	Set gtmtypes("unix_db_info",156,"off")=1028
	Set gtmtypes("unix_db_info",156,"len")=4
	Set gtmtypes("unix_db_info",156,"type")="boolean_t"
	Set gtmtypfldindx("unix_db_info","s_addrs.canceled_flush_timer")=156
	Set gtmtypes("unix_db_info",157,"name")="unix_db_info.s_addrs.probecrit_rec"
	Set gtmtypes("unix_db_info",157,"off")=1032
	Set gtmtypes("unix_db_info",157,"len")=56
	Set gtmtypes("unix_db_info",157,"type")="probecrit_rec_t"
	Set gtmtypfldindx("unix_db_info","s_addrs.probecrit_rec")=157
	Set gtmtypes("unix_db_info",158,"name")="unix_db_info.s_addrs.probecrit_rec.t_get_crit"
	Set gtmtypes("unix_db_info",158,"off")=1032
	Set gtmtypes("unix_db_info",158,"len")=8
	Set gtmtypes("unix_db_info",158,"type")="uint64_t"
	Set gtmtypfldindx("unix_db_info","s_addrs.probecrit_rec.t_get_crit")=158
	Set gtmtypes("unix_db_info",159,"name")="unix_db_info.s_addrs.probecrit_rec.p_crit_failed"
	Set gtmtypes("unix_db_info",159,"off")=1040
	Set gtmtypes("unix_db_info",159,"len")=8
	Set gtmtypes("unix_db_info",159,"type")="uint64_t"
	Set gtmtypfldindx("unix_db_info","s_addrs.probecrit_rec.p_crit_failed")=159
	Set gtmtypes("unix_db_info",160,"name")="unix_db_info.s_addrs.probecrit_rec.p_crit_que_slps"
	Set gtmtypes("unix_db_info",160,"off")=1048
	Set gtmtypes("unix_db_info",160,"len")=8
	Set gtmtypes("unix_db_info",160,"type")="uint64_t"
	Set gtmtypfldindx("unix_db_info","s_addrs.probecrit_rec.p_crit_que_slps")=160
	Set gtmtypes("unix_db_info",161,"name")="unix_db_info.s_addrs.probecrit_rec.p_crit_yields"
	Set gtmtypes("unix_db_info",161,"off")=1056
	Set gtmtypes("unix_db_info",161,"len")=8
	Set gtmtypes("unix_db_info",161,"type")="uint64_t"
	Set gtmtypfldindx("unix_db_info","s_addrs.probecrit_rec.p_crit_yields")=161
	Set gtmtypes("unix_db_info",162,"name")="unix_db_info.s_addrs.probecrit_rec.p_crit_que_full"
	Set gtmtypes("unix_db_info",162,"off")=1064
	Set gtmtypes("unix_db_info",162,"len")=8
	Set gtmtypes("unix_db_info",162,"type")="uint64_t"
	Set gtmtypfldindx("unix_db_info","s_addrs.probecrit_rec.p_crit_que_full")=162
	Set gtmtypes("unix_db_info",163,"name")="unix_db_info.s_addrs.probecrit_rec.p_crit_que_slots"
	Set gtmtypes("unix_db_info",163,"off")=1072
	Set gtmtypes("unix_db_info",163,"len")=8
	Set gtmtypes("unix_db_info",163,"type")="uint64_t"
	Set gtmtypfldindx("unix_db_info","s_addrs.probecrit_rec.p_crit_que_slots")=163
	Set gtmtypes("unix_db_info",164,"name")="unix_db_info.s_addrs.probecrit_rec.p_crit_success"
	Set gtmtypes("unix_db_info",164,"off")=1080
	Set gtmtypes("unix_db_info",164,"len")=8
	Set gtmtypes("unix_db_info",164,"type")="uint64_t"
	Set gtmtypfldindx("unix_db_info","s_addrs.probecrit_rec.p_crit_success")=164
	Set gtmtypes("unix_db_info",165,"name")="unix_db_info.s_addrs.lock_crit_with_db"
	Set gtmtypes("unix_db_info",165,"off")=1088
	Set gtmtypes("unix_db_info",165,"len")=4
	Set gtmtypes("unix_db_info",165,"type")="boolean_t"
	Set gtmtypfldindx("unix_db_info","s_addrs.lock_crit_with_db")=165
	Set gtmtypes("unix_db_info",166,"name")="unix_db_info.s_addrs.needs_post_freeze_online_clean"
	Set gtmtypes("unix_db_info",166,"off")=1092
	Set gtmtypes("unix_db_info",166,"len")=4
	Set gtmtypes("unix_db_info",166,"type")="boolean_t"
	Set gtmtypfldindx("unix_db_info","s_addrs.needs_post_freeze_online_clean")=166
	Set gtmtypes("unix_db_info",167,"name")="unix_db_info.s_addrs.needs_post_freeze_flushsync"
	Set gtmtypes("unix_db_info",167,"off")=1096
	Set gtmtypes("unix_db_info",167,"len")=4
	Set gtmtypes("unix_db_info",167,"type")="boolean_t"
	Set gtmtypfldindx("unix_db_info","s_addrs.needs_post_freeze_flushsync")=167
	Set gtmtypes("unix_db_info",168,"name")="unix_db_info.s_addrs.tp_hint"
	Set gtmtypes("unix_db_info",168,"off")=1100
	Set gtmtypes("unix_db_info",168,"len")=4
	Set gtmtypes("unix_db_info",168,"type")="int"
	Set gtmtypfldindx("unix_db_info","s_addrs.tp_hint")=168
	Set gtmtypes("unix_db_info",169,"name")="unix_db_info.s_addrs.tp_in_use"
	Set gtmtypes("unix_db_info",169,"off")=1104
	Set gtmtypes("unix_db_info",169,"len")=4
	Set gtmtypes("unix_db_info",169,"type")="boolean_t"
	Set gtmtypfldindx("unix_db_info","s_addrs.tp_in_use")=169
	Set gtmtypes("unix_db_info",170,"name")="unix_db_info.s_addrs.statsDB_setup_completed"
	Set gtmtypes("unix_db_info",170,"off")=1108
	Set gtmtypes("unix_db_info",170,"len")=4
	Set gtmtypes("unix_db_info",170,"type")="boolean_t"
	Set gtmtypfldindx("unix_db_info","s_addrs.statsDB_setup_completed")=170
	Set gtmtypes("unix_db_info",171,"name")="unix_db_info.s_addrs.gd_instinfo"
	Set gtmtypes("unix_db_info",171,"off")=1112
	Set gtmtypes("unix_db_info",171,"len")=8
	Set gtmtypes("unix_db_info",171,"type")="addr"
	Set gtmtypfldindx("unix_db_info","s_addrs.gd_instinfo")=171
	Set gtmtypes("unix_db_info",172,"name")="unix_db_info.s_addrs.gd_ptr"
	Set gtmtypes("unix_db_info",172,"off")=1120
	Set gtmtypes("unix_db_info",172,"len")=8
	Set gtmtypes("unix_db_info",172,"type")="addr"
	Set gtmtypfldindx("unix_db_info","s_addrs.gd_ptr")=172
	Set gtmtypes("unix_db_info",173,"name")="unix_db_info.s_addrs.jnlpool"
	Set gtmtypes("unix_db_info",173,"off")=1128
	Set gtmtypes("unix_db_info",173,"len")=8
	Set gtmtypes("unix_db_info",173,"type")="addr"
	Set gtmtypfldindx("unix_db_info","s_addrs.jnlpool")=173
	Set gtmtypes("unix_db_info",174,"name")="unix_db_info.s_addrs.mlkhash"
	Set gtmtypes("unix_db_info",174,"off")=1136
	Set gtmtypes("unix_db_info",174,"len")=8
	Set gtmtypes("unix_db_info",174,"type")="addr"
	Set gtmtypfldindx("unix_db_info","s_addrs.mlkhash")=174
	Set gtmtypes("unix_db_info",175,"name")="unix_db_info.s_addrs.mlkhash_shmid"
	Set gtmtypes("unix_db_info",175,"off")=1144
	Set gtmtypes("unix_db_info",175,"len")=4
	Set gtmtypes("unix_db_info",175,"type")="int"
	Set gtmtypfldindx("unix_db_info","s_addrs.mlkhash_shmid")=175
	Set gtmtypes("unix_db_info",176,"name")="unix_db_info.fn"
	Set gtmtypes("unix_db_info",176,"off")=1152
	Set gtmtypes("unix_db_info",176,"len")=8
	Set gtmtypes("unix_db_info",176,"type")="addr"
	Set gtmtypfldindx("unix_db_info","fn")=176
	Set gtmtypes("unix_db_info",177,"name")="unix_db_info.fd"
	Set gtmtypes("unix_db_info",177,"off")=1160
	Set gtmtypes("unix_db_info",177,"len")=4
	Set gtmtypes("unix_db_info",177,"type")="int"
	Set gtmtypfldindx("unix_db_info","fd")=177
	Set gtmtypes("unix_db_info",178,"name")="unix_db_info.owning_gd"
	Set gtmtypes("unix_db_info",178,"off")=1168
	Set gtmtypes("unix_db_info",178,"len")=8
	Set gtmtypes("unix_db_info",178,"type")="addr"
	Set gtmtypfldindx("unix_db_info","owning_gd")=178
	Set gtmtypes("unix_db_info",179,"name")="unix_db_info.fileid"
	Set gtmtypes("unix_db_info",179,"off")=1176
	Set gtmtypes("unix_db_info",179,"len")=16
	Set gtmtypes("unix_db_info",179,"type")="unix_file_id"
	Set gtmtypfldindx("unix_db_info","fileid")=179
	Set gtmtypes("unix_db_info",180,"name")="unix_db_info.fileid.inode"
	Set gtmtypes("unix_db_info",180,"off")=1176
	Set gtmtypes("unix_db_info",180,"len")=8
	Set gtmtypes("unix_db_info",180,"type")="ino_t"
	Set gtmtypfldindx("unix_db_info","fileid.inode")=180
	Set gtmtypes("unix_db_info",181,"name")="unix_db_info.fileid.device"
	Set gtmtypes("unix_db_info",181,"off")=1184
	Set gtmtypes("unix_db_info",181,"len")=8
	Set gtmtypes("unix_db_info",181,"type")="dev_t"
	Set gtmtypfldindx("unix_db_info","fileid.device")=181
	Set gtmtypes("unix_db_info",182,"name")="unix_db_info.semid"
	Set gtmtypes("unix_db_info",182,"off")=1192
	Set gtmtypes("unix_db_info",182,"len")=4
	Set gtmtypes("unix_db_info",182,"type")="int"
	Set gtmtypfldindx("unix_db_info","semid")=182
	Set gtmtypes("unix_db_info",183,"name")="unix_db_info.gt_sem_ctime"
	Set gtmtypes("unix_db_info",183,"off")=1200
	Set gtmtypes("unix_db_info",183,"len")=8
	Set gtmtypes("unix_db_info",183,"type")="time_t"
	Set gtmtypfldindx("unix_db_info","gt_sem_ctime")=183
	Set gtmtypes("unix_db_info",184,"name")="unix_db_info.shmid"
	Set gtmtypes("unix_db_info",184,"off")=1208
	Set gtmtypes("unix_db_info",184,"len")=4
	Set gtmtypes("unix_db_info",184,"type")="int"
	Set gtmtypfldindx("unix_db_info","shmid")=184
	Set gtmtypes("unix_db_info",185,"name")="unix_db_info.gt_shm_ctime"
	Set gtmtypes("unix_db_info",185,"off")=1216
	Set gtmtypes("unix_db_info",185,"len")=8
	Set gtmtypes("unix_db_info",185,"type")="time_t"
	Set gtmtypfldindx("unix_db_info","gt_shm_ctime")=185
	Set gtmtypes("unix_db_info",186,"name")="unix_db_info.ftok_semid"
	Set gtmtypes("unix_db_info",186,"off")=1224
	Set gtmtypes("unix_db_info",186,"len")=4
	Set gtmtypes("unix_db_info",186,"type")="int"
	Set gtmtypfldindx("unix_db_info","ftok_semid")=186
	Set gtmtypes("unix_db_info",187,"name")="unix_db_info.key"
	Set gtmtypes("unix_db_info",187,"off")=1228
	Set gtmtypes("unix_db_info",187,"len")=4
	Set gtmtypes("unix_db_info",187,"type")="key_t"
	Set gtmtypfldindx("unix_db_info","key")=187
	Set gtmtypes("unix_db_info",188,"name")="unix_db_info.raw"
	Set gtmtypes("unix_db_info",188,"off")=1232
	Set gtmtypes("unix_db_info",188,"len")=4
	Set gtmtypes("unix_db_info",188,"type")="boolean_t"
	Set gtmtypfldindx("unix_db_info","raw")=188
	Set gtmtypes("unix_db_info",189,"name")="unix_db_info.db_fs_block_size"
	Set gtmtypes("unix_db_info",189,"off")=1236
	Set gtmtypes("unix_db_info",189,"len")=4
	Set gtmtypes("unix_db_info",189,"type")="unsigned-int"
	Set gtmtypfldindx("unix_db_info","db_fs_block_size")=189
	;
	Set gtmtypes("unix_file_id")="struct"
	Set gtmtypes("unix_file_id",0)=2
	Set gtmtypes("unix_file_id","len")=16
	Set gtmtypes("unix_file_id",1,"name")="unix_file_id.inode"
	Set gtmtypes("unix_file_id",1,"off")=0
	Set gtmtypes("unix_file_id",1,"len")=8
	Set gtmtypes("unix_file_id",1,"type")="ino_t"
	Set gtmtypfldindx("unix_file_id","inode")=1
	Set gtmtypes("unix_file_id",2,"name")="unix_file_id.device"
	Set gtmtypes("unix_file_id",2,"off")=8
	Set gtmtypes("unix_file_id",2,"len")=8
	Set gtmtypes("unix_file_id",2,"type")="dev_t"
	Set gtmtypfldindx("unix_file_id","device")=2
	;
	Set gtmtypes("unix_file_info")="struct"
	Set gtmtypes("unix_file_info",0)=3
	Set gtmtypes("unix_file_info","len")=16
	Set gtmtypes("unix_file_info",1,"name")="unix_file_info.file"
	Set gtmtypes("unix_file_info",1,"off")=0
	Set gtmtypes("unix_file_info",1,"len")=4
	Set gtmtypes("unix_file_info",1,"type")="int"
	Set gtmtypfldindx("unix_file_info","file")=1
	Set gtmtypes("unix_file_info",2,"name")="unix_file_info.fn_len"
	Set gtmtypes("unix_file_info",2,"off")=4
	Set gtmtypes("unix_file_info",2,"len")=4
	Set gtmtypes("unix_file_info",2,"type")="int"
	Set gtmtypfldindx("unix_file_info","fn_len")=2
	Set gtmtypes("unix_file_info",3,"name")="unix_file_info.fn"
	Set gtmtypes("unix_file_info",3,"off")=8
	Set gtmtypes("unix_file_info",3,"len")=8
	Set gtmtypes("unix_file_info",3,"type")="addr"
	Set gtmtypfldindx("unix_file_info","fn")=3
	;
	Set gtmtypes("upd_helper_ctl_struct")="struct"
	Set gtmtypes("upd_helper_ctl_struct",0)=18
	Set gtmtypes("upd_helper_ctl_struct","len")=2088
	Set gtmtypes("upd_helper_ctl_struct",1,"name")="upd_helper_ctl_struct.pre_read_lock"
	Set gtmtypes("upd_helper_ctl_struct",1,"off")=0
	Set gtmtypes("upd_helper_ctl_struct",1,"len")=8
	Set gtmtypes("upd_helper_ctl_struct",1,"type")="global_latch_t"
	Set gtmtypfldindx("upd_helper_ctl_struct","pre_read_lock")=1
	Set gtmtypes("upd_helper_ctl_struct",2,"name")="upd_helper_ctl_struct.pre_read_lock.u"
	Set gtmtypes("upd_helper_ctl_struct",2,"off")=0
	Set gtmtypes("upd_helper_ctl_struct",2,"len")=8
	Set gtmtypes("upd_helper_ctl_struct",2,"type")="union"
	Set gtmtypfldindx("upd_helper_ctl_struct","pre_read_lock.u")=2
	Set gtmtypes("upd_helper_ctl_struct",3,"name")="upd_helper_ctl_struct.pre_read_lock.u.pid_imgcnt"
	Set gtmtypes("upd_helper_ctl_struct",3,"off")=0
	Set gtmtypes("upd_helper_ctl_struct",3,"len")=8
	Set gtmtypes("upd_helper_ctl_struct",3,"type")="uint64_t"
	Set gtmtypfldindx("upd_helper_ctl_struct","pre_read_lock.u.pid_imgcnt")=3
	Set gtmtypes("upd_helper_ctl_struct",4,"name")="upd_helper_ctl_struct.pre_read_lock.u.parts"
	Set gtmtypes("upd_helper_ctl_struct",4,"off")=0
	Set gtmtypes("upd_helper_ctl_struct",4,"len")=8
	Set gtmtypes("upd_helper_ctl_struct",4,"type")="struct"
	Set gtmtypfldindx("upd_helper_ctl_struct","pre_read_lock.u.parts")=4
	Set gtmtypes("upd_helper_ctl_struct",5,"name")="upd_helper_ctl_struct.pre_read_lock.u.parts.latch_pid"
	Set gtmtypes("upd_helper_ctl_struct",5,"off")=0
	Set gtmtypes("upd_helper_ctl_struct",5,"len")=4
	Set gtmtypes("upd_helper_ctl_struct",5,"type")="int"
	Set gtmtypfldindx("upd_helper_ctl_struct","pre_read_lock.u.parts.latch_pid")=5
	Set gtmtypes("upd_helper_ctl_struct",6,"name")="upd_helper_ctl_struct.pre_read_lock.u.parts.latch_word"
	Set gtmtypes("upd_helper_ctl_struct",6,"off")=4
	Set gtmtypes("upd_helper_ctl_struct",6,"len")=4
	Set gtmtypes("upd_helper_ctl_struct",6,"type")="int"
	Set gtmtypfldindx("upd_helper_ctl_struct","pre_read_lock.u.parts.latch_word")=6
	Set gtmtypes("upd_helper_ctl_struct",7,"name")="upd_helper_ctl_struct.pre_read_offset"
	Set gtmtypes("upd_helper_ctl_struct",7,"off")=8
	Set gtmtypes("upd_helper_ctl_struct",7,"len")=4
	Set gtmtypes("upd_helper_ctl_struct",7,"type")="unsigned-int"
	Set gtmtypfldindx("upd_helper_ctl_struct","pre_read_offset")=7
	Set gtmtypes("upd_helper_ctl_struct",8,"name")="upd_helper_ctl_struct.first_done"
	Set gtmtypes("upd_helper_ctl_struct",8,"off")=12
	Set gtmtypes("upd_helper_ctl_struct",8,"len")=4
	Set gtmtypes("upd_helper_ctl_struct",8,"type")="boolean_t"
	Set gtmtypfldindx("upd_helper_ctl_struct","first_done")=8
	Set gtmtypes("upd_helper_ctl_struct",9,"name")="upd_helper_ctl_struct.next_read_offset"
	Set gtmtypes("upd_helper_ctl_struct",9,"off")=16
	Set gtmtypes("upd_helper_ctl_struct",9,"len")=4
	Set gtmtypes("upd_helper_ctl_struct",9,"type")="unsigned-int"
	Set gtmtypfldindx("upd_helper_ctl_struct","next_read_offset")=9
	Set gtmtypes("upd_helper_ctl_struct",10,"name")="upd_helper_ctl_struct.start_helpers"
	Set gtmtypes("upd_helper_ctl_struct",10,"off")=20
	Set gtmtypes("upd_helper_ctl_struct",10,"len")=4
	Set gtmtypes("upd_helper_ctl_struct",10,"type")="unsigned-int"
	Set gtmtypfldindx("upd_helper_ctl_struct","start_helpers")=10
	Set gtmtypes("upd_helper_ctl_struct",11,"name")="upd_helper_ctl_struct.start_n_readers"
	Set gtmtypes("upd_helper_ctl_struct",11,"off")=24
	Set gtmtypes("upd_helper_ctl_struct",11,"len")=4
	Set gtmtypes("upd_helper_ctl_struct",11,"type")="unsigned-int"
	Set gtmtypfldindx("upd_helper_ctl_struct","start_n_readers")=11
	Set gtmtypes("upd_helper_ctl_struct",12,"name")="upd_helper_ctl_struct.start_n_writers"
	Set gtmtypes("upd_helper_ctl_struct",12,"off")=28
	Set gtmtypes("upd_helper_ctl_struct",12,"len")=4
	Set gtmtypes("upd_helper_ctl_struct",12,"type")="unsigned-int"
	Set gtmtypfldindx("upd_helper_ctl_struct","start_n_writers")=12
	Set gtmtypes("upd_helper_ctl_struct",13,"name")="upd_helper_ctl_struct.reap_helpers"
	Set gtmtypes("upd_helper_ctl_struct",13,"off")=32
	Set gtmtypes("upd_helper_ctl_struct",13,"len")=4
	Set gtmtypes("upd_helper_ctl_struct",13,"type")="unsigned-int"
	Set gtmtypfldindx("upd_helper_ctl_struct","reap_helpers")=13
	Set gtmtypes("upd_helper_ctl_struct",14,"name")="upd_helper_ctl_struct.helper_list"
	Set gtmtypes("upd_helper_ctl_struct",14,"off")=36
	Set gtmtypes("upd_helper_ctl_struct",14,"len")=2048
	Set gtmtypes("upd_helper_ctl_struct",14,"type")="upd_helper_entry_struct"
	Set gtmtypfldindx("upd_helper_ctl_struct","helper_list")=14
	Set gtmtypes("upd_helper_ctl_struct",14,"dim")=128
	Set gtmtypes("upd_helper_ctl_struct",15,"name")="upd_helper_ctl_struct.helper_list[0].helper_pid"
	Set gtmtypes("upd_helper_ctl_struct",15,"off")=36
	Set gtmtypes("upd_helper_ctl_struct",15,"len")=4
	Set gtmtypes("upd_helper_ctl_struct",15,"type")="unsigned-int"
	Set gtmtypfldindx("upd_helper_ctl_struct","helper_list[0].helper_pid")=15
	Set gtmtypes("upd_helper_ctl_struct",16,"name")="upd_helper_ctl_struct.helper_list[0].helper_pid_prev"
	Set gtmtypes("upd_helper_ctl_struct",16,"off")=40
	Set gtmtypes("upd_helper_ctl_struct",16,"len")=4
	Set gtmtypes("upd_helper_ctl_struct",16,"type")="unsigned-int"
	Set gtmtypfldindx("upd_helper_ctl_struct","helper_list[0].helper_pid_prev")=16
	Set gtmtypes("upd_helper_ctl_struct",17,"name")="upd_helper_ctl_struct.helper_list[0].helper_type"
	Set gtmtypes("upd_helper_ctl_struct",17,"off")=44
	Set gtmtypes("upd_helper_ctl_struct",17,"len")=4
	Set gtmtypes("upd_helper_ctl_struct",17,"type")="unsigned-int"
	Set gtmtypfldindx("upd_helper_ctl_struct","helper_list[0].helper_type")=17
	Set gtmtypes("upd_helper_ctl_struct",18,"name")="upd_helper_ctl_struct.helper_list[0].helper_shutdown"
	Set gtmtypes("upd_helper_ctl_struct",18,"off")=48
	Set gtmtypes("upd_helper_ctl_struct",18,"len")=4
	Set gtmtypes("upd_helper_ctl_struct",18,"type")="unsigned-int"
	Set gtmtypfldindx("upd_helper_ctl_struct","helper_list[0].helper_shutdown")=18
	;
	Set gtmtypes("upd_helper_entry_struct")="struct"
	Set gtmtypes("upd_helper_entry_struct",0)=4
	Set gtmtypes("upd_helper_entry_struct","len")=16
	Set gtmtypes("upd_helper_entry_struct",1,"name")="upd_helper_entry_struct.helper_pid"
	Set gtmtypes("upd_helper_entry_struct",1,"off")=0
	Set gtmtypes("upd_helper_entry_struct",1,"len")=4
	Set gtmtypes("upd_helper_entry_struct",1,"type")="unsigned-int"
	Set gtmtypfldindx("upd_helper_entry_struct","helper_pid")=1
	Set gtmtypes("upd_helper_entry_struct",2,"name")="upd_helper_entry_struct.helper_pid_prev"
	Set gtmtypes("upd_helper_entry_struct",2,"off")=4
	Set gtmtypes("upd_helper_entry_struct",2,"len")=4
	Set gtmtypes("upd_helper_entry_struct",2,"type")="unsigned-int"
	Set gtmtypfldindx("upd_helper_entry_struct","helper_pid_prev")=2
	Set gtmtypes("upd_helper_entry_struct",3,"name")="upd_helper_entry_struct.helper_type"
	Set gtmtypes("upd_helper_entry_struct",3,"off")=8
	Set gtmtypes("upd_helper_entry_struct",3,"len")=4
	Set gtmtypes("upd_helper_entry_struct",3,"type")="unsigned-int"
	Set gtmtypfldindx("upd_helper_entry_struct","helper_type")=3
	Set gtmtypes("upd_helper_entry_struct",4,"name")="upd_helper_entry_struct.helper_shutdown"
	Set gtmtypes("upd_helper_entry_struct",4,"off")=12
	Set gtmtypes("upd_helper_entry_struct",4,"len")=4
	Set gtmtypes("upd_helper_entry_struct",4,"type")="unsigned-int"
	Set gtmtypfldindx("upd_helper_entry_struct","helper_shutdown")=4
	;
	Set gtmtypes("upd_proc_local_struct")="struct"
	Set gtmtypes("upd_proc_local_struct",0)=12
	Set gtmtypes("upd_proc_local_struct","len")=312
	Set gtmtypes("upd_proc_local_struct",1,"name")="upd_proc_local_struct.upd_proc_pid"
	Set gtmtypes("upd_proc_local_struct",1,"off")=0
	Set gtmtypes("upd_proc_local_struct",1,"len")=4
	Set gtmtypes("upd_proc_local_struct",1,"type")="unsigned-int"
	Set gtmtypfldindx("upd_proc_local_struct","upd_proc_pid")=1
	Set gtmtypes("upd_proc_local_struct",2,"name")="upd_proc_local_struct.upd_proc_pid_prev"
	Set gtmtypes("upd_proc_local_struct",2,"off")=4
	Set gtmtypes("upd_proc_local_struct",2,"len")=4
	Set gtmtypes("upd_proc_local_struct",2,"type")="unsigned-int"
	Set gtmtypfldindx("upd_proc_local_struct","upd_proc_pid_prev")=2
	Set gtmtypes("upd_proc_local_struct",3,"name")="upd_proc_local_struct.read_jnl_seqno"
	Set gtmtypes("upd_proc_local_struct",3,"off")=8
	Set gtmtypes("upd_proc_local_struct",3,"len")=8
	Set gtmtypes("upd_proc_local_struct",3,"type")="uint64_t"
	Set gtmtypfldindx("upd_proc_local_struct","read_jnl_seqno")=3
	Set gtmtypes("upd_proc_local_struct",4,"name")="upd_proc_local_struct.read"
	Set gtmtypes("upd_proc_local_struct",4,"off")=16
	Set gtmtypes("upd_proc_local_struct",4,"len")=8
	Set gtmtypes("upd_proc_local_struct",4,"type")="uint64_t"
	Set gtmtypfldindx("upd_proc_local_struct","read")=4
	Set gtmtypes("upd_proc_local_struct",5,"name")="upd_proc_local_struct.upd_proc_shutdown"
	Set gtmtypes("upd_proc_local_struct",5,"off")=24
	Set gtmtypes("upd_proc_local_struct",5,"len")=4
	Set gtmtypes("upd_proc_local_struct",5,"type")="unsigned-int"
	Set gtmtypfldindx("upd_proc_local_struct","upd_proc_shutdown")=5
	Set gtmtypes("upd_proc_local_struct",6,"name")="upd_proc_local_struct.upd_proc_shutdown_time"
	Set gtmtypes("upd_proc_local_struct",6,"off")=28
	Set gtmtypes("upd_proc_local_struct",6,"len")=4
	Set gtmtypes("upd_proc_local_struct",6,"type")="int"
	Set gtmtypfldindx("upd_proc_local_struct","upd_proc_shutdown_time")=6
	Set gtmtypes("upd_proc_local_struct",7,"name")="upd_proc_local_struct.bad_trans"
	Set gtmtypes("upd_proc_local_struct",7,"off")=32
	Set gtmtypes("upd_proc_local_struct",7,"len")=4
	Set gtmtypes("upd_proc_local_struct",7,"type")="unsigned-int"
	Set gtmtypfldindx("upd_proc_local_struct","bad_trans")=7
	Set gtmtypes("upd_proc_local_struct",8,"name")="upd_proc_local_struct.changelog"
	Set gtmtypes("upd_proc_local_struct",8,"off")=36
	Set gtmtypes("upd_proc_local_struct",8,"len")=4
	Set gtmtypes("upd_proc_local_struct",8,"type")="unsigned-int"
	Set gtmtypfldindx("upd_proc_local_struct","changelog")=8
	Set gtmtypes("upd_proc_local_struct",9,"name")="upd_proc_local_struct.start_upd"
	Set gtmtypes("upd_proc_local_struct",9,"off")=40
	Set gtmtypes("upd_proc_local_struct",9,"len")=4
	Set gtmtypes("upd_proc_local_struct",9,"type")="int"
	Set gtmtypfldindx("upd_proc_local_struct","start_upd")=9
	Set gtmtypes("upd_proc_local_struct",10,"name")="upd_proc_local_struct.log_interval"
	Set gtmtypes("upd_proc_local_struct",10,"off")=44
	Set gtmtypes("upd_proc_local_struct",10,"len")=4
	Set gtmtypes("upd_proc_local_struct",10,"type")="unsigned-int"
	Set gtmtypfldindx("upd_proc_local_struct","log_interval")=10
	Set gtmtypes("upd_proc_local_struct",11,"name")="upd_proc_local_struct.log_file"
	Set gtmtypes("upd_proc_local_struct",11,"off")=48
	Set gtmtypes("upd_proc_local_struct",11,"len")=256
	Set gtmtypes("upd_proc_local_struct",11,"type")="char"
	Set gtmtypfldindx("upd_proc_local_struct","log_file")=11
	Set gtmtypes("upd_proc_local_struct",12,"name")="upd_proc_local_struct.onln_rlbk_flg"
	Set gtmtypes("upd_proc_local_struct",12,"off")=304
	Set gtmtypes("upd_proc_local_struct",12,"len")=4
	Set gtmtypes("upd_proc_local_struct",12,"type")="unsigned-int"
	Set gtmtypfldindx("upd_proc_local_struct","onln_rlbk_flg")=12
	;
	Set gtmtypes("urx_addr")="struct"
	Set gtmtypes("urx_addr",0)=2
	Set gtmtypes("urx_addr","len")=16
	Set gtmtypes("urx_addr",1,"name")="urx_addr.addr"
	Set gtmtypes("urx_addr",1,"off")=0
	Set gtmtypes("urx_addr",1,"len")=8
	Set gtmtypes("urx_addr",1,"type")="addr"
	Set gtmtypfldindx("urx_addr","addr")=1
	Set gtmtypes("urx_addr",2,"name")="urx_addr.next"
	Set gtmtypes("urx_addr",2,"off")=8
	Set gtmtypes("urx_addr",2,"len")=8
	Set gtmtypes("urx_addr",2,"type")="addr"
	Set gtmtypfldindx("urx_addr","next")=2
	;
	Set gtmtypes("urx_labref")="struct"
	Set gtmtypes("urx_labref",0)=4
	Set gtmtypes("urx_labref","len")=24
	Set gtmtypes("urx_labref",1,"name")="urx_labref.addr"
	Set gtmtypes("urx_labref",1,"off")=0
	Set gtmtypes("urx_labref",1,"len")=8
	Set gtmtypes("urx_labref",1,"type")="addr"
	Set gtmtypfldindx("urx_labref","addr")=1
	Set gtmtypes("urx_labref",2,"name")="urx_labref.next"
	Set gtmtypes("urx_labref",2,"off")=8
	Set gtmtypes("urx_labref",2,"len")=8
	Set gtmtypes("urx_labref",2,"type")="addr"
	Set gtmtypfldindx("urx_labref","next")=2
	Set gtmtypes("urx_labref",3,"name")="urx_labref.len"
	Set gtmtypes("urx_labref",3,"off")=16
	Set gtmtypes("urx_labref",3,"len")=4
	Set gtmtypes("urx_labref",3,"type")="unsigned-int"
	Set gtmtypfldindx("urx_labref","len")=3
	Set gtmtypes("urx_labref",4,"name")="urx_labref.name"
	Set gtmtypes("urx_labref",4,"off")=20
	Set gtmtypes("urx_labref",4,"len")=0
	Set gtmtypes("urx_labref",4,"type")="unsigned-char"
	Set gtmtypfldindx("urx_labref","name")=4
	;
	Set gtmtypes("urx_rtnref")="struct"
	Set gtmtypes("urx_rtnref",0)=5
	Set gtmtypes("urx_rtnref","len")=32
	Set gtmtypes("urx_rtnref",1,"name")="urx_rtnref.addr"
	Set gtmtypes("urx_rtnref",1,"off")=0
	Set gtmtypes("urx_rtnref",1,"len")=8
	Set gtmtypes("urx_rtnref",1,"type")="addr"
	Set gtmtypfldindx("urx_rtnref","addr")=1
	Set gtmtypes("urx_rtnref",2,"name")="urx_rtnref.lab"
	Set gtmtypes("urx_rtnref",2,"off")=8
	Set gtmtypes("urx_rtnref",2,"len")=8
	Set gtmtypes("urx_rtnref",2,"type")="addr"
	Set gtmtypfldindx("urx_rtnref","lab")=2
	Set gtmtypes("urx_rtnref",3,"name")="urx_rtnref.next"
	Set gtmtypes("urx_rtnref",3,"off")=16
	Set gtmtypes("urx_rtnref",3,"len")=8
	Set gtmtypes("urx_rtnref",3,"type")="addr"
	Set gtmtypfldindx("urx_rtnref","next")=3
	Set gtmtypes("urx_rtnref",4,"name")="urx_rtnref.len"
	Set gtmtypes("urx_rtnref",4,"off")=24
	Set gtmtypes("urx_rtnref",4,"len")=4
	Set gtmtypes("urx_rtnref",4,"type")="unsigned-int"
	Set gtmtypfldindx("urx_rtnref","len")=4
	Set gtmtypes("urx_rtnref",5,"name")="urx_rtnref.name"
	Set gtmtypes("urx_rtnref",5,"off")=28
	Set gtmtypes("urx_rtnref",5,"len")=0
	Set gtmtypes("urx_rtnref",5,"type")="unsigned-char"
	Set gtmtypfldindx("urx_rtnref","name")=5
	;
	Set gtmtypes("utfcgr")="struct"
	Set gtmtypes("utfcgr",0)=9
	Set gtmtypes("utfcgr","len")=32
	Set gtmtypes("utfcgr",1,"name")="utfcgr.last_str"
	Set gtmtypes("utfcgr",1,"off")=0
	Set gtmtypes("utfcgr",1,"len")=16
	Set gtmtypes("utfcgr",1,"type")="mstr"
	Set gtmtypfldindx("utfcgr","last_str")=1
	Set gtmtypes("utfcgr",2,"name")="utfcgr.last_str.char_len"
	Set gtmtypes("utfcgr",2,"off")=0
	Set gtmtypes("utfcgr",2,"len")=4
	Set gtmtypes("utfcgr",2,"type")="unsigned-int"
	Set gtmtypfldindx("utfcgr","last_str.char_len")=2
	Set gtmtypes("utfcgr",3,"name")="utfcgr.last_str.len"
	Set gtmtypes("utfcgr",3,"off")=4
	Set gtmtypes("utfcgr",3,"len")=4
	Set gtmtypes("utfcgr",3,"type")="int"
	Set gtmtypfldindx("utfcgr","last_str.len")=3
	Set gtmtypes("utfcgr",4,"name")="utfcgr.last_str.addr"
	Set gtmtypes("utfcgr",4,"off")=8
	Set gtmtypes("utfcgr",4,"len")=8
	Set gtmtypes("utfcgr",4,"type")="addr"
	Set gtmtypfldindx("utfcgr","last_str.addr")=4
	Set gtmtypes("utfcgr",5,"name")="utfcgr.ngrps"
	Set gtmtypes("utfcgr",5,"off")=16
	Set gtmtypes("utfcgr",5,"len")=2
	Set gtmtypes("utfcgr",5,"type")="unsigned-short"
	Set gtmtypfldindx("utfcgr","ngrps")=5
	Set gtmtypes("utfcgr",6,"name")="utfcgr.idx"
	Set gtmtypes("utfcgr",6,"off")=18
	Set gtmtypes("utfcgr",6,"len")=2
	Set gtmtypes("utfcgr",6,"type")="unsigned-short"
	Set gtmtypfldindx("utfcgr","idx")=6
	Set gtmtypes("utfcgr",7,"name")="utfcgr.reference"
	Set gtmtypes("utfcgr",7,"off")=20
	Set gtmtypes("utfcgr",7,"len")=4
	Set gtmtypes("utfcgr",7,"type")="boolean_t"
	Set gtmtypfldindx("utfcgr","reference")=7
	Set gtmtypes("utfcgr",8,"name")="utfcgr.entry"
	Set gtmtypes("utfcgr",8,"off")=24
	Set gtmtypes("utfcgr",8,"len")=8
	Set gtmtypes("utfcgr",8,"type")="utfcgr_entry"
	Set gtmtypfldindx("utfcgr","entry")=8
	Set gtmtypes("utfcgr",9,"name")="utfcgr.entry[0].byteidx"
	Set gtmtypes("utfcgr",9,"off")=28
	Set gtmtypes("utfcgr",9,"len")=4
	Set gtmtypes("utfcgr",9,"type")="unsigned-int"
	Set gtmtypfldindx("utfcgr","entry[0].byteidx")=9
	;
	Set gtmtypes("utfcgr_area")="struct"
	Set gtmtypes("utfcgr_area",0)=4
	Set gtmtypes("utfcgr_area","len")=32
	Set gtmtypes("utfcgr_area",1,"name")="utfcgr_area.utfcgrsteal"
	Set gtmtypes("utfcgr_area",1,"off")=0
	Set gtmtypes("utfcgr_area",1,"len")=8
	Set gtmtypes("utfcgr_area",1,"type")="addr"
	Set gtmtypfldindx("utfcgr_area","utfcgrsteal")=1
	Set gtmtypes("utfcgr_area",2,"name")="utfcgr_area.utfcgrmax"
	Set gtmtypes("utfcgr_area",2,"off")=8
	Set gtmtypes("utfcgr_area",2,"len")=8
	Set gtmtypes("utfcgr_area",2,"type")="addr"
	Set gtmtypfldindx("utfcgr_area","utfcgrmax")=2
	Set gtmtypes("utfcgr_area",3,"name")="utfcgr_area.utfcgrs"
	Set gtmtypes("utfcgr_area",3,"off")=16
	Set gtmtypes("utfcgr_area",3,"len")=8
	Set gtmtypes("utfcgr_area",3,"type")="addr"
	Set gtmtypfldindx("utfcgr_area","utfcgrs")=3
	Set gtmtypes("utfcgr_area",4,"name")="utfcgr_area.utfcgrsize"
	Set gtmtypes("utfcgr_area",4,"off")=24
	Set gtmtypes("utfcgr_area",4,"len")=4
	Set gtmtypes("utfcgr_area",4,"type")="unsigned-int"
	Set gtmtypfldindx("utfcgr_area","utfcgrsize")=4
	;
	Set gtmtypes("utfcgr_entry")="struct"
	Set gtmtypes("utfcgr_entry",0)=1
	Set gtmtypes("utfcgr_entry","len")=8
	Set gtmtypes("utfcgr_entry",1,"name")="utfcgr_entry.byteidx"
	Set gtmtypes("utfcgr_entry",1,"off")=4
	Set gtmtypes("utfcgr_entry",1,"len")=4
	Set gtmtypes("utfcgr_entry",1,"type")="unsigned-int"
	Set gtmtypfldindx("utfcgr_entry","byteidx")=1
	;
	Set gtmtypes("utfscan_parseblk")="struct"
	Set gtmtypes("utfscan_parseblk",0)=9
	Set gtmtypes("utfscan_parseblk","len")=48
	Set gtmtypes("utfscan_parseblk",1,"name")="utfscan_parseblk.mv"
	Set gtmtypes("utfscan_parseblk",1,"off")=0
	Set gtmtypes("utfscan_parseblk",1,"len")=8
	Set gtmtypes("utfscan_parseblk",1,"type")="addr"
	Set gtmtypfldindx("utfscan_parseblk","mv")=1
	Set gtmtypes("utfscan_parseblk",2,"name")="utfscan_parseblk.stoponbadchar"
	Set gtmtypes("utfscan_parseblk",2,"off")=8
	Set gtmtypes("utfscan_parseblk",2,"len")=4
	Set gtmtypes("utfscan_parseblk",2,"type")="boolean_t"
	Set gtmtypfldindx("utfscan_parseblk","stoponbadchar")=2
	Set gtmtypes("utfscan_parseblk",3,"name")="utfscan_parseblk.scan_byte_offset"
	Set gtmtypes("utfscan_parseblk",3,"off")=12
	Set gtmtypes("utfscan_parseblk",3,"len")=4
	Set gtmtypes("utfscan_parseblk",3,"type")="int"
	Set gtmtypfldindx("utfscan_parseblk","scan_byte_offset")=3
	Set gtmtypes("utfscan_parseblk",4,"name")="utfscan_parseblk.scan_char_count"
	Set gtmtypes("utfscan_parseblk",4,"off")=16
	Set gtmtypes("utfscan_parseblk",4,"len")=4
	Set gtmtypes("utfscan_parseblk",4,"type")="int"
	Set gtmtypfldindx("utfscan_parseblk","scan_char_count")=4
	Set gtmtypes("utfscan_parseblk",5,"name")="utfscan_parseblk.utfcgr_indx"
	Set gtmtypes("utfscan_parseblk",5,"off")=20
	Set gtmtypes("utfscan_parseblk",5,"len")=4
	Set gtmtypes("utfscan_parseblk",5,"type")="int"
	Set gtmtypfldindx("utfscan_parseblk","utfcgr_indx")=5
	Set gtmtypes("utfscan_parseblk",6,"name")="utfscan_parseblk.scan_char_len"
	Set gtmtypes("utfscan_parseblk",6,"off")=24
	Set gtmtypes("utfscan_parseblk",6,"len")=4
	Set gtmtypes("utfscan_parseblk",6,"type")="int"
	Set gtmtypfldindx("utfscan_parseblk","scan_char_len")=6
	Set gtmtypes("utfscan_parseblk",7,"name")="utfscan_parseblk.scan_char_type"
	Set gtmtypes("utfscan_parseblk",7,"off")=28
	Set gtmtypes("utfscan_parseblk",7,"len")=4
	Set gtmtypes("utfscan_parseblk",7,"type")="int"
	Set gtmtypfldindx("utfscan_parseblk","scan_char_type")=7
	Set gtmtypes("utfscan_parseblk",8,"name")="utfscan_parseblk.badcharstr"
	Set gtmtypes("utfscan_parseblk",8,"off")=32
	Set gtmtypes("utfscan_parseblk",8,"len")=8
	Set gtmtypes("utfscan_parseblk",8,"type")="addr"
	Set gtmtypfldindx("utfscan_parseblk","badcharstr")=8
	Set gtmtypes("utfscan_parseblk",9,"name")="utfscan_parseblk.badchartop"
	Set gtmtypes("utfscan_parseblk",9,"off")=40
	Set gtmtypes("utfscan_parseblk",9,"len")=8
	Set gtmtypes("utfscan_parseblk",9,"type")="addr"
	Set gtmtypfldindx("utfscan_parseblk","badchartop")=9
	;
	Set gtmtypes("util_snapshot_t")="struct"
	Set gtmtypes("util_snapshot_t",0)=3
	Set gtmtypes("util_snapshot_t","len")=24
	Set gtmtypes("util_snapshot_t",1,"name")="util_snapshot_t.master_map"
	Set gtmtypes("util_snapshot_t",1,"off")=0
	Set gtmtypes("util_snapshot_t",1,"len")=8
	Set gtmtypes("util_snapshot_t",1,"type")="addr"
	Set gtmtypfldindx("util_snapshot_t","master_map")=1
	Set gtmtypes("util_snapshot_t",2,"name")="util_snapshot_t.header"
	Set gtmtypes("util_snapshot_t",2,"off")=8
	Set gtmtypes("util_snapshot_t",2,"len")=8
	Set gtmtypes("util_snapshot_t",2,"type")="addr"
	Set gtmtypfldindx("util_snapshot_t","header")=2
	Set gtmtypes("util_snapshot_t",3,"name")="util_snapshot_t.native_size"
	Set gtmtypes("util_snapshot_t",3,"off")=16
	Set gtmtypes("util_snapshot_t",3,"len")=8
	Set gtmtypes("util_snapshot_t",3,"type")="uint64_t"
	Set gtmtypfldindx("util_snapshot_t","native_size")=3
	;
	Set gtmtypes("viewparm")="union"
	Set gtmtypes("viewparm",0)=11
	Set gtmtypes("viewparm","len")=32
	Set gtmtypes("viewparm",1,"name")="viewparm.ident"
	Set gtmtypes("viewparm",1,"off")=0
	Set gtmtypes("viewparm",1,"len")=32
	Set gtmtypes("viewparm",1,"type")="mident_fixed"
	Set gtmtypfldindx("viewparm","ident")=1
	Set gtmtypes("viewparm",2,"name")="viewparm.ident.c"
	Set gtmtypes("viewparm",2,"off")=0
	Set gtmtypes("viewparm",2,"len")=32
	Set gtmtypes("viewparm",2,"type")="char"
	Set gtmtypfldindx("viewparm","ident.c")=2
	Set gtmtypes("viewparm",3,"name")="viewparm.value"
	Set gtmtypes("viewparm",3,"off")=0
	Set gtmtypes("viewparm",3,"len")=8
	Set gtmtypes("viewparm",3,"type")="addr"
	Set gtmtypfldindx("viewparm","value")=3
	Set gtmtypes("viewparm",4,"name")="viewparm.gv_ptr"
	Set gtmtypes("viewparm",4,"off")=0
	Set gtmtypes("viewparm",4,"len")=8
	Set gtmtypes("viewparm",4,"type")="addr"
	Set gtmtypfldindx("viewparm","gv_ptr")=4
	Set gtmtypes("viewparm",5,"name")="viewparm.ni_list"
	Set gtmtypes("viewparm",5,"off")=0
	Set gtmtypes("viewparm",5,"len")=16
	Set gtmtypes("viewparm",5,"type")="noisolation_list"
	Set gtmtypfldindx("viewparm","ni_list")=5
	Set gtmtypes("viewparm",6,"name")="viewparm.ni_list.type"
	Set gtmtypes("viewparm",6,"off")=0
	Set gtmtypes("viewparm",6,"len")=4
	Set gtmtypes("viewparm",6,"type")="int"
	Set gtmtypfldindx("viewparm","ni_list.type")=6
	Set gtmtypes("viewparm",7,"name")="viewparm.ni_list.gvnh_list"
	Set gtmtypes("viewparm",7,"off")=8
	Set gtmtypes("viewparm",7,"len")=8
	Set gtmtypes("viewparm",7,"type")="addr"
	Set gtmtypfldindx("viewparm","ni_list.gvnh_list")=7
	Set gtmtypes("viewparm",8,"name")="viewparm.str"
	Set gtmtypes("viewparm",8,"off")=0
	Set gtmtypes("viewparm",8,"len")=16
	Set gtmtypes("viewparm",8,"type")="mstr"
	Set gtmtypfldindx("viewparm","str")=8
	Set gtmtypes("viewparm",9,"name")="viewparm.str.char_len"
	Set gtmtypes("viewparm",9,"off")=0
	Set gtmtypes("viewparm",9,"len")=4
	Set gtmtypes("viewparm",9,"type")="unsigned-int"
	Set gtmtypfldindx("viewparm","str.char_len")=9
	Set gtmtypes("viewparm",10,"name")="viewparm.str.len"
	Set gtmtypes("viewparm",10,"off")=4
	Set gtmtypes("viewparm",10,"len")=4
	Set gtmtypes("viewparm",10,"type")="int"
	Set gtmtypfldindx("viewparm","str.len")=10
	Set gtmtypes("viewparm",11,"name")="viewparm.str.addr"
	Set gtmtypes("viewparm",11,"off")=8
	Set gtmtypes("viewparm",11,"len")=8
	Set gtmtypes("viewparm",11,"type")="addr"
	Set gtmtypfldindx("viewparm","str.addr")=11
	;
	Set gtmtypes("viewtab_entry")="struct"
	Set gtmtypes("viewtab_entry",0)=4
	Set gtmtypes("viewtab_entry","len")=19
	Set gtmtypes("viewtab_entry",1,"name")="viewtab_entry.keyword"
	Set gtmtypes("viewtab_entry",1,"off")=0
	Set gtmtypes("viewtab_entry",1,"len")=16
	Set gtmtypes("viewtab_entry",1,"type")="unsigned-char"
	Set gtmtypfldindx("viewtab_entry","keyword")=1
	Set gtmtypes("viewtab_entry",2,"name")="viewtab_entry.parm"
	Set gtmtypes("viewtab_entry",2,"off")=16
	Set gtmtypes("viewtab_entry",2,"len")=1
	Set gtmtypes("viewtab_entry",2,"type")="unsigned-char"
	Set gtmtypfldindx("viewtab_entry","parm")=2
	Set gtmtypes("viewtab_entry",3,"name")="viewtab_entry.keycode"
	Set gtmtypes("viewtab_entry",3,"off")=17
	Set gtmtypes("viewtab_entry",3,"len")=1
	Set gtmtypes("viewtab_entry",3,"type")="unsigned-char"
	Set gtmtypfldindx("viewtab_entry","keycode")=3
	Set gtmtypes("viewtab_entry",4,"name")="viewtab_entry.restype"
	Set gtmtypes("viewtab_entry",4,"off")=18
	Set gtmtypes("viewtab_entry",4,"len")=1
	Set gtmtypes("viewtab_entry",4,"type")="unsigned-char"
	Set gtmtypfldindx("viewtab_entry","restype")=4
	;
	Set gtmtypes("vms_lock_sb")="struct"
	Set gtmtypes("vms_lock_sb",0)=4
	Set gtmtypes("vms_lock_sb","len")=24
	Set gtmtypes("vms_lock_sb",1,"name")="vms_lock_sb.cond"
	Set gtmtypes("vms_lock_sb",1,"off")=0
	Set gtmtypes("vms_lock_sb",1,"len")=2
	Set gtmtypes("vms_lock_sb",1,"type")="short"
	Set gtmtypfldindx("vms_lock_sb","cond")=1
	Set gtmtypes("vms_lock_sb",2,"name")="vms_lock_sb.reserved"
	Set gtmtypes("vms_lock_sb",2,"off")=2
	Set gtmtypes("vms_lock_sb",2,"len")=2
	Set gtmtypes("vms_lock_sb",2,"type")="short"
	Set gtmtypfldindx("vms_lock_sb","reserved")=2
	Set gtmtypes("vms_lock_sb",3,"name")="vms_lock_sb.lockid"
	Set gtmtypes("vms_lock_sb",3,"off")=4
	Set gtmtypes("vms_lock_sb",3,"len")=4
	Set gtmtypes("vms_lock_sb",3,"type")="int"
	Set gtmtypfldindx("vms_lock_sb","lockid")=3
	Set gtmtypes("vms_lock_sb",4,"name")="vms_lock_sb.valblk"
	Set gtmtypes("vms_lock_sb",4,"off")=8
	Set gtmtypes("vms_lock_sb",4,"len")=16
	Set gtmtypes("vms_lock_sb",4,"type")="int"
	Set gtmtypfldindx("vms_lock_sb","valblk")=4
	Set gtmtypes("vms_lock_sb",4,"dim")=4
	;
	Set gtmtypes("wcs_conflict_trace_t")="struct"
	Set gtmtypes("wcs_conflict_trace_t",0)=3
	Set gtmtypes("wcs_conflict_trace_t","len")=12
	Set gtmtypes("wcs_conflict_trace_t",1,"name")="wcs_conflict_trace_t.wcs_active_lvl"
	Set gtmtypes("wcs_conflict_trace_t",1,"off")=0
	Set gtmtypes("wcs_conflict_trace_t",1,"len")=4
	Set gtmtypes("wcs_conflict_trace_t",1,"type")="int"
	Set gtmtypfldindx("wcs_conflict_trace_t","wcs_active_lvl")=1
	Set gtmtypes("wcs_conflict_trace_t",2,"name")="wcs_conflict_trace_t.io_in_prog_pid"
	Set gtmtypes("wcs_conflict_trace_t",2,"off")=4
	Set gtmtypes("wcs_conflict_trace_t",2,"len")=4
	Set gtmtypes("wcs_conflict_trace_t",2,"type")="int"
	Set gtmtypfldindx("wcs_conflict_trace_t","io_in_prog_pid")=2
	Set gtmtypes("wcs_conflict_trace_t",3,"name")="wcs_conflict_trace_t.fsync_in_prog_pid"
	Set gtmtypes("wcs_conflict_trace_t",3,"off")=8
	Set gtmtypes("wcs_conflict_trace_t",3,"len")=4
	Set gtmtypes("wcs_conflict_trace_t",3,"type")="int"
	Set gtmtypfldindx("wcs_conflict_trace_t","fsync_in_prog_pid")=3
	;
	Set gtmtypes("wcs_ops_trace_t")="struct"
	Set gtmtypes("wcs_ops_trace_t",0)=11
	Set gtmtypes("wcs_ops_trace_t","len")=56
	Set gtmtypes("wcs_ops_trace_t",1,"name")="wcs_ops_trace_t.db_tn"
	Set gtmtypes("wcs_ops_trace_t",1,"off")=0
	Set gtmtypes("wcs_ops_trace_t",1,"len")=8
	Set gtmtypes("wcs_ops_trace_t",1,"type")="uint64_t"
	Set gtmtypfldindx("wcs_ops_trace_t","db_tn")=1
	Set gtmtypes("wcs_ops_trace_t",2,"name")="wcs_ops_trace_t.process_id"
	Set gtmtypes("wcs_ops_trace_t",2,"off")=8
	Set gtmtypes("wcs_ops_trace_t",2,"len")=4
	Set gtmtypes("wcs_ops_trace_t",2,"type")="unsigned-int"
	Set gtmtypfldindx("wcs_ops_trace_t","process_id")=2
	Set gtmtypes("wcs_ops_trace_t",3,"name")="wcs_ops_trace_t.wc_in_free"
	Set gtmtypes("wcs_ops_trace_t",3,"off")=12
	Set gtmtypes("wcs_ops_trace_t",3,"len")=4
	Set gtmtypes("wcs_ops_trace_t",3,"type")="unsigned-int"
	Set gtmtypfldindx("wcs_ops_trace_t","wc_in_free")=3
	Set gtmtypes("wcs_ops_trace_t",4,"name")="wcs_ops_trace_t.wcs_active_lvl"
	Set gtmtypes("wcs_ops_trace_t",4,"off")=16
	Set gtmtypes("wcs_ops_trace_t",4,"len")=4
	Set gtmtypes("wcs_ops_trace_t",4,"type")="unsigned-int"
	Set gtmtypfldindx("wcs_ops_trace_t","wcs_active_lvl")=4
	Set gtmtypes("wcs_ops_trace_t",5,"name")="wcs_ops_trace_t.wcs_wip_lvl"
	Set gtmtypes("wcs_ops_trace_t",5,"off")=20
	Set gtmtypes("wcs_ops_trace_t",5,"len")=4
	Set gtmtypes("wcs_ops_trace_t",5,"type")="unsigned-int"
	Set gtmtypfldindx("wcs_ops_trace_t","wcs_wip_lvl")=5
	Set gtmtypes("wcs_ops_trace_t",6,"name")="wcs_ops_trace_t.type"
	Set gtmtypes("wcs_ops_trace_t",6,"off")=24
	Set gtmtypes("wcs_ops_trace_t",6,"len")=4
	Set gtmtypes("wcs_ops_trace_t",6,"type")="int"
	Set gtmtypfldindx("wcs_ops_trace_t","type")=6
	Set gtmtypes("wcs_ops_trace_t",7,"name")="wcs_ops_trace_t.blk"
	Set gtmtypes("wcs_ops_trace_t",7,"off")=28
	Set gtmtypes("wcs_ops_trace_t",7,"len")=4
	Set gtmtypes("wcs_ops_trace_t",7,"type")="int"
	Set gtmtypfldindx("wcs_ops_trace_t","blk")=7
	Set gtmtypes("wcs_ops_trace_t",8,"name")="wcs_ops_trace_t.cr_off"
	Set gtmtypes("wcs_ops_trace_t",8,"off")=32
	Set gtmtypes("wcs_ops_trace_t",8,"len")=8
	Set gtmtypes("wcs_ops_trace_t",8,"type")="intptr_t"
	Set gtmtypfldindx("wcs_ops_trace_t","cr_off")=8
	Set gtmtypes("wcs_ops_trace_t",9,"name")="wcs_ops_trace_t.cr_dirty"
	Set gtmtypes("wcs_ops_trace_t",9,"off")=40
	Set gtmtypes("wcs_ops_trace_t",9,"len")=8
	Set gtmtypes("wcs_ops_trace_t",9,"type")="uint64_t"
	Set gtmtypfldindx("wcs_ops_trace_t","cr_dirty")=9
	Set gtmtypes("wcs_ops_trace_t",10,"name")="wcs_ops_trace_t.detail1"
	Set gtmtypes("wcs_ops_trace_t",10,"off")=48
	Set gtmtypes("wcs_ops_trace_t",10,"len")=4
	Set gtmtypes("wcs_ops_trace_t",10,"type")="unsigned-int"
	Set gtmtypfldindx("wcs_ops_trace_t","detail1")=10
	Set gtmtypes("wcs_ops_trace_t",11,"name")="wcs_ops_trace_t.detail2"
	Set gtmtypes("wcs_ops_trace_t",11,"off")=52
	Set gtmtypes("wcs_ops_trace_t",11,"len")=4
	Set gtmtypes("wcs_ops_trace_t",11,"type")="unsigned-int"
	Set gtmtypfldindx("wcs_ops_trace_t","detail2")=11
	;
	Set gtmtypes("wtstart_cr_list_t")="struct"
	Set gtmtypes("wtstart_cr_list_t",0)=3
	Set gtmtypes("wtstart_cr_list_t","len")=16
	Set gtmtypes("wtstart_cr_list_t",1,"name")="wtstart_cr_list_t.numcrs"
	Set gtmtypes("wtstart_cr_list_t",1,"off")=0
	Set gtmtypes("wtstart_cr_list_t",1,"len")=4
	Set gtmtypes("wtstart_cr_list_t",1,"type")="int"
	Set gtmtypfldindx("wtstart_cr_list_t","numcrs")=1
	Set gtmtypes("wtstart_cr_list_t",2,"name")="wtstart_cr_list_t.listsize"
	Set gtmtypes("wtstart_cr_list_t",2,"off")=4
	Set gtmtypes("wtstart_cr_list_t",2,"len")=4
	Set gtmtypes("wtstart_cr_list_t",2,"type")="int"
	Set gtmtypfldindx("wtstart_cr_list_t","listsize")=2
	Set gtmtypes("wtstart_cr_list_t",3,"name")="wtstart_cr_list_t.listcrs"
	Set gtmtypes("wtstart_cr_list_t",3,"off")=8
	Set gtmtypes("wtstart_cr_list_t",3,"len")=8
	Set gtmtypes("wtstart_cr_list_t",3,"type")="addr"
	Set gtmtypfldindx("wtstart_cr_list_t","listcrs")=3
	;
	Set gtmtypes("z_records")="struct"
	Set gtmtypes("z_records",0)=3
	Set gtmtypes("z_records","len")=24
	Set gtmtypes("z_records",1,"name")="z_records.beg"
	Set gtmtypes("z_records",1,"off")=0
	Set gtmtypes("z_records",1,"len")=8
	Set gtmtypes("z_records",1,"type")="addr"
	Set gtmtypfldindx("z_records","beg")=1
	Set gtmtypes("z_records",2,"name")="z_records.gtm_free"
	Set gtmtypes("z_records",2,"off")=8
	Set gtmtypes("z_records",2,"len")=8
	Set gtmtypes("z_records",2,"type")="addr"
	Set gtmtypfldindx("z_records","gtm_free")=2
	Set gtmtypes("z_records",3,"name")="z_records.end"
	Set gtmtypes("z_records",3,"off")=16
	Set gtmtypes("z_records",3,"len")=8
	Set gtmtypes("z_records",3,"type")="addr"
	Set gtmtypfldindx("z_records","end")=3
	;
	Set gtmtypes("zbrk_struct")="struct"
	Set gtmtypes("zbrk_struct",0)=9
	Set gtmtypes("zbrk_struct","len")=56
	Set gtmtypes("zbrk_struct",1,"name")="zbrk_struct.mpc"
	Set gtmtypes("zbrk_struct",1,"off")=0
	Set gtmtypes("zbrk_struct",1,"len")=8
	Set gtmtypes("zbrk_struct",1,"type")="addr"
	Set gtmtypfldindx("zbrk_struct","mpc")=1
	Set gtmtypes("zbrk_struct",2,"name")="zbrk_struct.rtn"
	Set gtmtypes("zbrk_struct",2,"off")=8
	Set gtmtypes("zbrk_struct",2,"len")=8
	Set gtmtypes("zbrk_struct",2,"type")="addr"
	Set gtmtypfldindx("zbrk_struct","rtn")=2
	Set gtmtypes("zbrk_struct",3,"name")="zbrk_struct.lab"
	Set gtmtypes("zbrk_struct",3,"off")=16
	Set gtmtypes("zbrk_struct",3,"len")=8
	Set gtmtypes("zbrk_struct",3,"type")="addr"
	Set gtmtypfldindx("zbrk_struct","lab")=3
	Set gtmtypes("zbrk_struct",4,"name")="zbrk_struct.offset"
	Set gtmtypes("zbrk_struct",4,"off")=24
	Set gtmtypes("zbrk_struct",4,"len")=4
	Set gtmtypes("zbrk_struct",4,"type")="int"
	Set gtmtypfldindx("zbrk_struct","offset")=4
	Set gtmtypes("zbrk_struct",5,"name")="zbrk_struct.count"
	Set gtmtypes("zbrk_struct",5,"off")=28
	Set gtmtypes("zbrk_struct",5,"len")=4
	Set gtmtypes("zbrk_struct",5,"type")="int"
	Set gtmtypfldindx("zbrk_struct","count")=5
	Set gtmtypes("zbrk_struct",6,"name")="zbrk_struct.action"
	Set gtmtypes("zbrk_struct",6,"off")=32
	Set gtmtypes("zbrk_struct",6,"len")=8
	Set gtmtypes("zbrk_struct",6,"type")="addr"
	Set gtmtypfldindx("zbrk_struct","action")=6
	Set gtmtypes("zbrk_struct",7,"name")="zbrk_struct.rtnhdr"
	Set gtmtypes("zbrk_struct",7,"off")=40
	Set gtmtypes("zbrk_struct",7,"len")=8
	Set gtmtypes("zbrk_struct",7,"type")="addr"
	Set gtmtypfldindx("zbrk_struct","rtnhdr")=7
	Set gtmtypes("zbrk_struct",8,"name")="zbrk_struct.m_opcode"
	Set gtmtypes("zbrk_struct",8,"off")=48
	Set gtmtypes("zbrk_struct",8,"len")=1
	Set gtmtypes("zbrk_struct",8,"type")="unsigned-char"
	Set gtmtypfldindx("zbrk_struct","m_opcode")=8
	Set gtmtypes("zbrk_struct",9,"name")="zbrk_struct.filler"
	Set gtmtypes("zbrk_struct",9,"off")=49
	Set gtmtypes("zbrk_struct",9,"len")=7
	Set gtmtypes("zbrk_struct",9,"type")="char"
	Set gtmtypfldindx("zbrk_struct","filler")=9
	;
	Set gtmtypes("zintcmd_active_info")="struct"
	Set gtmtypes("zintcmd_active_info",0)=3
	Set gtmtypes("zintcmd_active_info","len")=24
	Set gtmtypes("zintcmd_active_info",1,"name")="zintcmd_active_info.count"
	Set gtmtypes("zintcmd_active_info",1,"off")=0
	Set gtmtypes("zintcmd_active_info",1,"len")=4
	Set gtmtypes("zintcmd_active_info",1,"type")="int"
	Set gtmtypfldindx("zintcmd_active_info","count")=1
	Set gtmtypes("zintcmd_active_info",2,"name")="zintcmd_active_info.restart_pc_last"
	Set gtmtypes("zintcmd_active_info",2,"off")=8
	Set gtmtypes("zintcmd_active_info",2,"len")=8
	Set gtmtypes("zintcmd_active_info",2,"type")="addr"
	Set gtmtypfldindx("zintcmd_active_info","restart_pc_last")=2
	Set gtmtypes("zintcmd_active_info",3,"name")="zintcmd_active_info.restart_ctxt_last"
	Set gtmtypes("zintcmd_active_info",3,"off")=16
	Set gtmtypes("zintcmd_active_info",3,"len")=8
	Set gtmtypes("zintcmd_active_info",3,"type")="addr"
	Set gtmtypfldindx("zintcmd_active_info","restart_ctxt_last")=3
	;
	Set gtmtypes("zro_ent")="struct"
	Set gtmtypes("zro_ent",0)=9
	Set gtmtypes("zro_ent","len")=48
	Set gtmtypes("zro_ent",1,"name")="zro_ent.type"
	Set gtmtypes("zro_ent",1,"off")=0
	Set gtmtypes("zro_ent",1,"len")=4
	Set gtmtypes("zro_ent",1,"type")="unsigned-int"
	Set gtmtypfldindx("zro_ent","type")=1
	Set gtmtypes("zro_ent",2,"name")="zro_ent.count"
	Set gtmtypes("zro_ent",2,"off")=4
	Set gtmtypes("zro_ent",2,"len")=4
	Set gtmtypes("zro_ent",2,"type")="int"
	Set gtmtypfldindx("zro_ent","count")=2
	Set gtmtypes("zro_ent",3,"name")="zro_ent.str"
	Set gtmtypes("zro_ent",3,"off")=8
	Set gtmtypes("zro_ent",3,"len")=16
	Set gtmtypes("zro_ent",3,"type")="mstr"
	Set gtmtypfldindx("zro_ent","str")=3
	Set gtmtypes("zro_ent",4,"name")="zro_ent.str.char_len"
	Set gtmtypes("zro_ent",4,"off")=8
	Set gtmtypes("zro_ent",4,"len")=4
	Set gtmtypes("zro_ent",4,"type")="unsigned-int"
	Set gtmtypfldindx("zro_ent","str.char_len")=4
	Set gtmtypes("zro_ent",5,"name")="zro_ent.str.len"
	Set gtmtypes("zro_ent",5,"off")=12
	Set gtmtypes("zro_ent",5,"len")=4
	Set gtmtypes("zro_ent",5,"type")="int"
	Set gtmtypfldindx("zro_ent","str.len")=5
	Set gtmtypes("zro_ent",6,"name")="zro_ent.str.addr"
	Set gtmtypes("zro_ent",6,"off")=16
	Set gtmtypes("zro_ent",6,"len")=8
	Set gtmtypes("zro_ent",6,"type")="addr"
	Set gtmtypfldindx("zro_ent","str.addr")=6
	Set gtmtypes("zro_ent",7,"name")="zro_ent.shrlib"
	Set gtmtypes("zro_ent",7,"off")=24
	Set gtmtypes("zro_ent",7,"len")=8
	Set gtmtypes("zro_ent",7,"type")="addr"
	Set gtmtypfldindx("zro_ent","shrlib")=7
	Set gtmtypes("zro_ent",8,"name")="zro_ent.shrsym"
	Set gtmtypes("zro_ent",8,"off")=32
	Set gtmtypes("zro_ent",8,"len")=8
	Set gtmtypes("zro_ent",8,"type")="addr"
	Set gtmtypfldindx("zro_ent","shrsym")=8
	Set gtmtypes("zro_ent",9,"name")="zro_ent.relinkctl_sgmaddr"
	Set gtmtypes("zro_ent",9,"off")=40
	Set gtmtypes("zro_ent",9,"len")=8
	Set gtmtypes("zro_ent",9,"type")="addr"
	Set gtmtypfldindx("zro_ent","relinkctl_sgmaddr")=9
	;
	Set gtmtypes("zro_hist")="struct"
	Set gtmtypes("zro_hist",0)=6
	Set gtmtypes("zro_hist","len")=40
	Set gtmtypes("zro_hist",1,"name")="zro_hist.zroutines_cycle"
	Set gtmtypes("zro_hist",1,"off")=0
	Set gtmtypes("zro_hist",1,"len")=4
	Set gtmtypes("zro_hist",1,"type")="unsigned-int"
	Set gtmtypfldindx("zro_hist","zroutines_cycle")=1
	Set gtmtypes("zro_hist",2,"name")="zro_hist.end"
	Set gtmtypes("zro_hist",2,"off")=8
	Set gtmtypes("zro_hist",2,"len")=8
	Set gtmtypes("zro_hist",2,"type")="addr"
	Set gtmtypfldindx("zro_hist","end")=2
	Set gtmtypes("zro_hist",3,"name")="zro_hist.base"
	Set gtmtypes("zro_hist",3,"off")=16
	Set gtmtypes("zro_hist",3,"len")=24
	Set gtmtypes("zro_hist",3,"type")="zro_validation_entry"
	Set gtmtypfldindx("zro_hist","base")=3
	Set gtmtypes("zro_hist",4,"name")="zro_hist.base[0].cycle"
	Set gtmtypes("zro_hist",4,"off")=16
	Set gtmtypes("zro_hist",4,"len")=4
	Set gtmtypes("zro_hist",4,"type")="unsigned-int"
	Set gtmtypfldindx("zro_hist","base[0].cycle")=4
	Set gtmtypes("zro_hist",5,"name")="zro_hist.base[0].relinkrec"
	Set gtmtypes("zro_hist",5,"off")=24
	Set gtmtypes("zro_hist",5,"len")=8
	Set gtmtypes("zro_hist",5,"type")="addr"
	Set gtmtypfldindx("zro_hist","base[0].relinkrec")=5
	Set gtmtypes("zro_hist",6,"name")="zro_hist.base[0].relinkctl_bkptr"
	Set gtmtypes("zro_hist",6,"off")=32
	Set gtmtypes("zro_hist",6,"len")=8
	Set gtmtypes("zro_hist",6,"type")="addr"
	Set gtmtypfldindx("zro_hist","base[0].relinkctl_bkptr")=6
	;
	Set gtmtypes("zro_search_hist_ent")="struct"
	Set gtmtypes("zro_search_hist_ent",0)=7
	Set gtmtypes("zro_search_hist_ent","len")=64
	Set gtmtypes("zro_search_hist_ent",1,"name")="zro_search_hist_ent.zro_valent"
	Set gtmtypes("zro_search_hist_ent",1,"off")=0
	Set gtmtypes("zro_search_hist_ent",1,"len")=24
	Set gtmtypes("zro_search_hist_ent",1,"type")="zro_validation_entry"
	Set gtmtypfldindx("zro_search_hist_ent","zro_valent")=1
	Set gtmtypes("zro_search_hist_ent",2,"name")="zro_search_hist_ent.zro_valent.cycle"
	Set gtmtypes("zro_search_hist_ent",2,"off")=0
	Set gtmtypes("zro_search_hist_ent",2,"len")=4
	Set gtmtypes("zro_search_hist_ent",2,"type")="unsigned-int"
	Set gtmtypfldindx("zro_search_hist_ent","zro_valent.cycle")=2
	Set gtmtypes("zro_search_hist_ent",3,"name")="zro_search_hist_ent.zro_valent.relinkrec"
	Set gtmtypes("zro_search_hist_ent",3,"off")=8
	Set gtmtypes("zro_search_hist_ent",3,"len")=8
	Set gtmtypes("zro_search_hist_ent",3,"type")="addr"
	Set gtmtypfldindx("zro_search_hist_ent","zro_valent.relinkrec")=3
	Set gtmtypes("zro_search_hist_ent",4,"name")="zro_search_hist_ent.zro_valent.relinkctl_bkptr"
	Set gtmtypes("zro_search_hist_ent",4,"off")=16
	Set gtmtypes("zro_search_hist_ent",4,"len")=8
	Set gtmtypes("zro_search_hist_ent",4,"type")="addr"
	Set gtmtypfldindx("zro_search_hist_ent","zro_valent.relinkctl_bkptr")=4
	Set gtmtypes("zro_search_hist_ent",5,"name")="zro_search_hist_ent.rtnname"
	Set gtmtypes("zro_search_hist_ent",5,"off")=24
	Set gtmtypes("zro_search_hist_ent",5,"len")=32
	Set gtmtypes("zro_search_hist_ent",5,"type")="mident_fixed"
	Set gtmtypfldindx("zro_search_hist_ent","rtnname")=5
	Set gtmtypes("zro_search_hist_ent",6,"name")="zro_search_hist_ent.rtnname.c"
	Set gtmtypes("zro_search_hist_ent",6,"off")=24
	Set gtmtypes("zro_search_hist_ent",6,"len")=32
	Set gtmtypes("zro_search_hist_ent",6,"type")="char"
	Set gtmtypfldindx("zro_search_hist_ent","rtnname.c")=6
	Set gtmtypes("zro_search_hist_ent",7,"name")="zro_search_hist_ent.rtnname_len"
	Set gtmtypes("zro_search_hist_ent",7,"off")=56
	Set gtmtypes("zro_search_hist_ent",7,"len")=4
	Set gtmtypes("zro_search_hist_ent",7,"type")="int"
	Set gtmtypfldindx("zro_search_hist_ent","rtnname_len")=7
	;
	Set gtmtypes("zro_validation_entry")="struct"
	Set gtmtypes("zro_validation_entry",0)=3
	Set gtmtypes("zro_validation_entry","len")=24
	Set gtmtypes("zro_validation_entry",1,"name")="zro_validation_entry.cycle"
	Set gtmtypes("zro_validation_entry",1,"off")=0
	Set gtmtypes("zro_validation_entry",1,"len")=4
	Set gtmtypes("zro_validation_entry",1,"type")="unsigned-int"
	Set gtmtypfldindx("zro_validation_entry","cycle")=1
	Set gtmtypes("zro_validation_entry",2,"name")="zro_validation_entry.relinkrec"
	Set gtmtypes("zro_validation_entry",2,"off")=8
	Set gtmtypes("zro_validation_entry",2,"len")=8
	Set gtmtypes("zro_validation_entry",2,"type")="addr"
	Set gtmtypfldindx("zro_validation_entry","relinkrec")=2
	Set gtmtypes("zro_validation_entry",3,"name")="zro_validation_entry.relinkctl_bkptr"
	Set gtmtypes("zro_validation_entry",3,"off")=16
	Set gtmtypes("zro_validation_entry",3,"len")=8
	Set gtmtypes("zro_validation_entry",3,"type")="addr"
	Set gtmtypfldindx("zro_validation_entry","relinkctl_bkptr")=3
	;
	Set gtmtypes("zs_gv_struct")="struct"
	Set gtmtypes("zs_gv_struct",0)=2
	Set gtmtypes("zs_gv_struct","len")=8
	Set gtmtypes("zs_gv_struct",1,"name")="zs_gv_struct.end"
	Set gtmtypes("zs_gv_struct",1,"off")=0
	Set gtmtypes("zs_gv_struct",1,"len")=4
	Set gtmtypes("zs_gv_struct",1,"type")="int"
	Set gtmtypfldindx("zs_gv_struct","end")=1
	Set gtmtypes("zs_gv_struct",2,"name")="zs_gv_struct.prev"
	Set gtmtypes("zs_gv_struct",2,"off")=4
	Set gtmtypes("zs_gv_struct",2,"len")=4
	Set gtmtypes("zs_gv_struct",2,"type")="int"
	Set gtmtypfldindx("zs_gv_struct","prev")=2
	;
	Set gtmtypes("zs_lv_struct")="struct"
	Set gtmtypes("zs_lv_struct",0)=2
	Set gtmtypes("zs_lv_struct","len")=16
	Set gtmtypes("zs_lv_struct",1,"name")="zs_lv_struct.lvar"
	Set gtmtypes("zs_lv_struct",1,"off")=0
	Set gtmtypes("zs_lv_struct",1,"len")=8
	Set gtmtypes("zs_lv_struct",1,"type")="addr"
	Set gtmtypfldindx("zs_lv_struct","lvar")=1
	Set gtmtypes("zs_lv_struct",2,"name")="zs_lv_struct.child"
	Set gtmtypes("zs_lv_struct",2,"off")=8
	Set gtmtypes("zs_lv_struct",2,"len")=8
	Set gtmtypes("zs_lv_struct",2,"type")="addr"
	Set gtmtypfldindx("zs_lv_struct","child")=2
	;
	Set gtmtypes("zshow_index")="struct"
	Set gtmtypes("zshow_index",0)=2
	Set gtmtypes("zshow_index","len")=2
	Set gtmtypes("zshow_index",1,"name")="zshow_index.offset"
	Set gtmtypes("zshow_index",1,"off")=0
	Set gtmtypes("zshow_index",1,"len")=1
	Set gtmtypes("zshow_index",1,"type")="char"
	Set gtmtypfldindx("zshow_index","offset")=1
	Set gtmtypes("zshow_index",2,"name")="zshow_index.letter"
	Set gtmtypes("zshow_index",2,"off")=1
	Set gtmtypes("zshow_index",2,"len")=1
	Set gtmtypes("zshow_index",2,"type")="char"
	Set gtmtypfldindx("zshow_index","letter")=2
	;
	Set gtmtypes("zshow_out")="struct"
	Set gtmtypes("zshow_out",0)=18
	Set gtmtypes("zshow_out","len")=64
	Set gtmtypes("zshow_out",1,"name")="zshow_out.type"
	Set gtmtypes("zshow_out",1,"off")=0
	Set gtmtypes("zshow_out",1,"len")=1
	Set gtmtypes("zshow_out",1,"type")="char"
	Set gtmtypfldindx("zshow_out","type")=1
	Set gtmtypes("zshow_out",2,"name")="zshow_out.code"
	Set gtmtypes("zshow_out",2,"off")=1
	Set gtmtypes("zshow_out",2,"len")=1
	Set gtmtypes("zshow_out",2,"type")="char"
	Set gtmtypfldindx("zshow_out","code")=2
	Set gtmtypes("zshow_out",3,"name")="zshow_out.curr_code"
	Set gtmtypes("zshow_out",3,"off")=2
	Set gtmtypes("zshow_out",3,"len")=1
	Set gtmtypes("zshow_out",3,"type")="char"
	Set gtmtypfldindx("zshow_out","curr_code")=3
	Set gtmtypes("zshow_out",4,"name")="zshow_out.size"
	Set gtmtypes("zshow_out",4,"off")=4
	Set gtmtypes("zshow_out",4,"len")=4
	Set gtmtypes("zshow_out",4,"type")="int"
	Set gtmtypfldindx("zshow_out","size")=4
	Set gtmtypes("zshow_out",5,"name")="zshow_out.buff"
	Set gtmtypes("zshow_out",5,"off")=8
	Set gtmtypes("zshow_out",5,"len")=8
	Set gtmtypes("zshow_out",5,"type")="addr"
	Set gtmtypfldindx("zshow_out","buff")=5
	Set gtmtypes("zshow_out",6,"name")="zshow_out.ptr"
	Set gtmtypes("zshow_out",6,"off")=16
	Set gtmtypes("zshow_out",6,"len")=8
	Set gtmtypes("zshow_out",6,"type")="addr"
	Set gtmtypfldindx("zshow_out","ptr")=6
	Set gtmtypes("zshow_out",7,"name")="zshow_out.len"
	Set gtmtypes("zshow_out",7,"off")=24
	Set gtmtypes("zshow_out",7,"len")=4
	Set gtmtypes("zshow_out",7,"type")="int"
	Set gtmtypfldindx("zshow_out","len")=7
	Set gtmtypes("zshow_out",8,"name")="zshow_out.displen"
	Set gtmtypes("zshow_out",8,"off")=28
	Set gtmtypes("zshow_out",8,"len")=4
	Set gtmtypes("zshow_out",8,"type")="int"
	Set gtmtypfldindx("zshow_out","displen")=8
	Set gtmtypes("zshow_out",9,"name")="zshow_out.line_num"
	Set gtmtypes("zshow_out",9,"off")=32
	Set gtmtypes("zshow_out",9,"len")=4
	Set gtmtypes("zshow_out",9,"type")="int"
	Set gtmtypfldindx("zshow_out","line_num")=9
	Set gtmtypes("zshow_out",10,"name")="zshow_out.line_cont"
	Set gtmtypes("zshow_out",10,"off")=36
	Set gtmtypes("zshow_out",10,"len")=4
	Set gtmtypes("zshow_out",10,"type")="int"
	Set gtmtypfldindx("zshow_out","line_cont")=10
	Set gtmtypes("zshow_out",11,"name")="zshow_out.flush"
	Set gtmtypes("zshow_out",11,"off")=40
	Set gtmtypes("zshow_out",11,"len")=4
	Set gtmtypes("zshow_out",11,"type")="boolean_t"
	Set gtmtypfldindx("zshow_out","flush")=11
	Set gtmtypes("zshow_out",12,"name")="zshow_out.out_var"
	Set gtmtypes("zshow_out",12,"off")=48
	Set gtmtypes("zshow_out",12,"len")=16
	Set gtmtypes("zshow_out",12,"type")="union"
	Set gtmtypfldindx("zshow_out","out_var")=12
	Set gtmtypes("zshow_out",13,"name")="zshow_out.out_var.lv"
	Set gtmtypes("zshow_out",13,"off")=48
	Set gtmtypes("zshow_out",13,"len")=16
	Set gtmtypes("zshow_out",13,"type")="zs_lv_struct"
	Set gtmtypfldindx("zshow_out","out_var.lv")=13
	Set gtmtypes("zshow_out",14,"name")="zshow_out.out_var.lv.lvar"
	Set gtmtypes("zshow_out",14,"off")=48
	Set gtmtypes("zshow_out",14,"len")=8
	Set gtmtypes("zshow_out",14,"type")="addr"
	Set gtmtypfldindx("zshow_out","out_var.lv.lvar")=14
	Set gtmtypes("zshow_out",15,"name")="zshow_out.out_var.lv.child"
	Set gtmtypes("zshow_out",15,"off")=56
	Set gtmtypes("zshow_out",15,"len")=8
	Set gtmtypes("zshow_out",15,"type")="addr"
	Set gtmtypfldindx("zshow_out","out_var.lv.child")=15
	Set gtmtypes("zshow_out",16,"name")="zshow_out.out_var.gv"
	Set gtmtypes("zshow_out",16,"off")=48
	Set gtmtypes("zshow_out",16,"len")=8
	Set gtmtypes("zshow_out",16,"type")="zs_gv_struct"
	Set gtmtypfldindx("zshow_out","out_var.gv")=16
	Set gtmtypes("zshow_out",17,"name")="zshow_out.out_var.gv.end"
	Set gtmtypes("zshow_out",17,"off")=48
	Set gtmtypes("zshow_out",17,"len")=4
	Set gtmtypes("zshow_out",17,"type")="int"
	Set gtmtypfldindx("zshow_out","out_var.gv.end")=17
	Set gtmtypes("zshow_out",18,"name")="zshow_out.out_var.gv.prev"
	Set gtmtypes("zshow_out",18,"off")=52
	Set gtmtypes("zshow_out",18,"len")=4
	Set gtmtypes("zshow_out",18,"type")="int"
	Set gtmtypfldindx("zshow_out","out_var.gv.prev")=18
	;
	Set gtmtypes("zwr_alias_var")="struct"
	Set gtmtypes("zwr_alias_var",0)=6
	Set gtmtypes("zwr_alias_var","len")=24
	Set gtmtypes("zwr_alias_var",1,"name")="zwr_alias_var.value_printed"
	Set gtmtypes("zwr_alias_var",1,"off")=0
	Set gtmtypes("zwr_alias_var",1,"len")=4
	Set gtmtypes("zwr_alias_var",1,"type")="boolean_t"
	Set gtmtypfldindx("zwr_alias_var","value_printed")=1
	Set gtmtypes("zwr_alias_var",2,"name")="zwr_alias_var.filler"
	Set gtmtypes("zwr_alias_var",2,"off")=4
	Set gtmtypes("zwr_alias_var",2,"len")=4
	Set gtmtypes("zwr_alias_var",2,"type")="int"
	Set gtmtypfldindx("zwr_alias_var","filler")=2
	Set gtmtypes("zwr_alias_var",3,"name")="zwr_alias_var.zwr_var"
	Set gtmtypes("zwr_alias_var",3,"off")=8
	Set gtmtypes("zwr_alias_var",3,"len")=16
	Set gtmtypes("zwr_alias_var",3,"type")="mstr"
	Set gtmtypfldindx("zwr_alias_var","zwr_var")=3
	Set gtmtypes("zwr_alias_var",4,"name")="zwr_alias_var.zwr_var.char_len"
	Set gtmtypes("zwr_alias_var",4,"off")=8
	Set gtmtypes("zwr_alias_var",4,"len")=4
	Set gtmtypes("zwr_alias_var",4,"type")="unsigned-int"
	Set gtmtypfldindx("zwr_alias_var","zwr_var.char_len")=4
	Set gtmtypes("zwr_alias_var",5,"name")="zwr_alias_var.zwr_var.len"
	Set gtmtypes("zwr_alias_var",5,"off")=12
	Set gtmtypes("zwr_alias_var",5,"len")=4
	Set gtmtypes("zwr_alias_var",5,"type")="int"
	Set gtmtypfldindx("zwr_alias_var","zwr_var.len")=5
	Set gtmtypes("zwr_alias_var",6,"name")="zwr_alias_var.zwr_var.addr"
	Set gtmtypes("zwr_alias_var",6,"off")=16
	Set gtmtypes("zwr_alias_var",6,"len")=8
	Set gtmtypes("zwr_alias_var",6,"type")="addr"
	Set gtmtypfldindx("zwr_alias_var","zwr_var.addr")=6
	;
	Set gtmtypes("zwr_hash_table")="struct"
	Set gtmtypes("zwr_hash_table",0)=20
	Set gtmtypes("zwr_hash_table","len")=104
	Set gtmtypes("zwr_hash_table",1,"name")="zwr_hash_table.cleaned"
	Set gtmtypes("zwr_hash_table",1,"off")=0
	Set gtmtypes("zwr_hash_table",1,"len")=4
	Set gtmtypes("zwr_hash_table",1,"type")="boolean_t"
	Set gtmtypfldindx("zwr_hash_table","cleaned")=1
	Set gtmtypes("zwr_hash_table",2,"name")="zwr_hash_table.filler"
	Set gtmtypes("zwr_hash_table",2,"off")=4
	Set gtmtypes("zwr_hash_table",2,"len")=4
	Set gtmtypes("zwr_hash_table",2,"type")="int"
	Set gtmtypfldindx("zwr_hash_table","filler")=2
	Set gtmtypes("zwr_hash_table",3,"name")="zwr_hash_table.h_zwrtab"
	Set gtmtypes("zwr_hash_table",3,"off")=8
	Set gtmtypes("zwr_hash_table",3,"len")=80
	Set gtmtypes("zwr_hash_table",3,"type")="hash_table_addr"
	Set gtmtypfldindx("zwr_hash_table","h_zwrtab")=3
	Set gtmtypes("zwr_hash_table",4,"name")="zwr_hash_table.h_zwrtab.base"
	Set gtmtypes("zwr_hash_table",4,"off")=8
	Set gtmtypes("zwr_hash_table",4,"len")=8
	Set gtmtypes("zwr_hash_table",4,"type")="addr"
	Set gtmtypfldindx("zwr_hash_table","h_zwrtab.base")=4
	Set gtmtypes("zwr_hash_table",5,"name")="zwr_hash_table.h_zwrtab.top"
	Set gtmtypes("zwr_hash_table",5,"off")=16
	Set gtmtypes("zwr_hash_table",5,"len")=8
	Set gtmtypes("zwr_hash_table",5,"type")="addr"
	Set gtmtypfldindx("zwr_hash_table","h_zwrtab.top")=5
	Set gtmtypes("zwr_hash_table",6,"name")="zwr_hash_table.h_zwrtab.spare_base"
	Set gtmtypes("zwr_hash_table",6,"off")=24
	Set gtmtypes("zwr_hash_table",6,"len")=8
	Set gtmtypes("zwr_hash_table",6,"type")="addr"
	Set gtmtypfldindx("zwr_hash_table","h_zwrtab.spare_base")=6
	Set gtmtypes("zwr_hash_table",7,"name")="zwr_hash_table.h_zwrtab.entry_passed_thru"
	Set gtmtypes("zwr_hash_table",7,"off")=32
	Set gtmtypes("zwr_hash_table",7,"len")=8
	Set gtmtypes("zwr_hash_table",7,"type")="addr"
	Set gtmtypfldindx("zwr_hash_table","h_zwrtab.entry_passed_thru")=7
	Set gtmtypes("zwr_hash_table",8,"name")="zwr_hash_table.h_zwrtab.size"
	Set gtmtypes("zwr_hash_table",8,"off")=40
	Set gtmtypes("zwr_hash_table",8,"len")=4
	Set gtmtypes("zwr_hash_table",8,"type")="unsigned-int"
	Set gtmtypfldindx("zwr_hash_table","h_zwrtab.size")=8
	Set gtmtypes("zwr_hash_table",9,"name")="zwr_hash_table.h_zwrtab.initial_size"
	Set gtmtypes("zwr_hash_table",9,"off")=44
	Set gtmtypes("zwr_hash_table",9,"len")=4
	Set gtmtypes("zwr_hash_table",9,"type")="unsigned-int"
	Set gtmtypfldindx("zwr_hash_table","h_zwrtab.initial_size")=9
	Set gtmtypes("zwr_hash_table",10,"name")="zwr_hash_table.h_zwrtab.spare_base_size"
	Set gtmtypes("zwr_hash_table",10,"off")=48
	Set gtmtypes("zwr_hash_table",10,"len")=4
	Set gtmtypes("zwr_hash_table",10,"type")="unsigned-int"
	Set gtmtypfldindx("zwr_hash_table","h_zwrtab.spare_base_size")=10
	Set gtmtypes("zwr_hash_table",11,"name")="zwr_hash_table.h_zwrtab.count"
	Set gtmtypes("zwr_hash_table",11,"off")=52
	Set gtmtypes("zwr_hash_table",11,"len")=4
	Set gtmtypes("zwr_hash_table",11,"type")="unsigned-int"
	Set gtmtypfldindx("zwr_hash_table","h_zwrtab.count")=11
	Set gtmtypes("zwr_hash_table",12,"name")="zwr_hash_table.h_zwrtab.del_count"
	Set gtmtypes("zwr_hash_table",12,"off")=56
	Set gtmtypes("zwr_hash_table",12,"len")=4
	Set gtmtypes("zwr_hash_table",12,"type")="unsigned-int"
	Set gtmtypfldindx("zwr_hash_table","h_zwrtab.del_count")=12
	Set gtmtypes("zwr_hash_table",13,"name")="zwr_hash_table.h_zwrtab.exp_trigger_size"
	Set gtmtypes("zwr_hash_table",13,"off")=60
	Set gtmtypes("zwr_hash_table",13,"len")=4
	Set gtmtypes("zwr_hash_table",13,"type")="unsigned-int"
	Set gtmtypfldindx("zwr_hash_table","h_zwrtab.exp_trigger_size")=13
	Set gtmtypes("zwr_hash_table",14,"name")="zwr_hash_table.h_zwrtab.cmp_trigger_size"
	Set gtmtypes("zwr_hash_table",14,"off")=64
	Set gtmtypes("zwr_hash_table",14,"len")=4
	Set gtmtypes("zwr_hash_table",14,"type")="unsigned-int"
	Set gtmtypfldindx("zwr_hash_table","h_zwrtab.cmp_trigger_size")=14
	Set gtmtypes("zwr_hash_table",15,"name")="zwr_hash_table.h_zwrtab.dont_compact"
	Set gtmtypes("zwr_hash_table",15,"off")=68
	Set gtmtypes("zwr_hash_table",15,"len")=4
	Set gtmtypes("zwr_hash_table",15,"type")="boolean_t"
	Set gtmtypfldindx("zwr_hash_table","h_zwrtab.dont_compact")=15
	Set gtmtypes("zwr_hash_table",16,"name")="zwr_hash_table.h_zwrtab.dont_keep_spare_table"
	Set gtmtypes("zwr_hash_table",16,"off")=72
	Set gtmtypes("zwr_hash_table",16,"len")=4
	Set gtmtypes("zwr_hash_table",16,"type")="boolean_t"
	Set gtmtypfldindx("zwr_hash_table","h_zwrtab.dont_keep_spare_table")=16
	Set gtmtypes("zwr_hash_table",17,"name")="zwr_hash_table.h_zwrtab.defer_base_release"
	Set gtmtypes("zwr_hash_table",17,"off")=76
	Set gtmtypes("zwr_hash_table",17,"len")=4
	Set gtmtypes("zwr_hash_table",17,"type")="boolean_t"
	Set gtmtypfldindx("zwr_hash_table","h_zwrtab.defer_base_release")=17
	Set gtmtypes("zwr_hash_table",18,"name")="zwr_hash_table.h_zwrtab.active"
	Set gtmtypes("zwr_hash_table",18,"off")=80
	Set gtmtypes("zwr_hash_table",18,"len")=4
	Set gtmtypes("zwr_hash_table",18,"type")="boolean_t"
	Set gtmtypfldindx("zwr_hash_table","h_zwrtab.active")=18
	Set gtmtypes("zwr_hash_table",19,"name")="zwr_hash_table.first_zwrzavb"
	Set gtmtypes("zwr_hash_table",19,"off")=88
	Set gtmtypes("zwr_hash_table",19,"len")=8
	Set gtmtypes("zwr_hash_table",19,"type")="addr"
	Set gtmtypfldindx("zwr_hash_table","first_zwrzavb")=19
	Set gtmtypes("zwr_hash_table",20,"name")="zwr_hash_table.zav_flist"
	Set gtmtypes("zwr_hash_table",20,"off")=96
	Set gtmtypes("zwr_hash_table",20,"len")=8
	Set gtmtypes("zwr_hash_table",20,"type")="addr"
	Set gtmtypfldindx("zwr_hash_table","zav_flist")=20
	;
	Set gtmtypes("zwr_sub_lst")="struct"
	Set gtmtypes("zwr_sub_lst",0)=5
	Set gtmtypes("zwr_sub_lst","len")=32
	Set gtmtypes("zwr_sub_lst",1,"name")="zwr_sub_lst.subsc_list"
	Set gtmtypes("zwr_sub_lst",1,"off")=0
	Set gtmtypes("zwr_sub_lst",1,"len")=32
	Set gtmtypes("zwr_sub_lst",1,"type")="struct"
	Set gtmtypfldindx("zwr_sub_lst","subsc_list")=1
	Set gtmtypes("zwr_sub_lst",2,"name")="zwr_sub_lst.subsc_list[0].subsc_type"
	Set gtmtypes("zwr_sub_lst",2,"off")=0
	Set gtmtypes("zwr_sub_lst",2,"len")=1
	Set gtmtypes("zwr_sub_lst",2,"type")="unsigned-char"
	Set gtmtypfldindx("zwr_sub_lst","subsc_list[0].subsc_type")=2
	Set gtmtypes("zwr_sub_lst",3,"name")="zwr_sub_lst.subsc_list[0].actual"
	Set gtmtypes("zwr_sub_lst",3,"off")=8
	Set gtmtypes("zwr_sub_lst",3,"len")=8
	Set gtmtypes("zwr_sub_lst",3,"type")="addr"
	Set gtmtypfldindx("zwr_sub_lst","subsc_list[0].actual")=3
	Set gtmtypes("zwr_sub_lst",4,"name")="zwr_sub_lst.subsc_list[0].first"
	Set gtmtypes("zwr_sub_lst",4,"off")=16
	Set gtmtypes("zwr_sub_lst",4,"len")=8
	Set gtmtypes("zwr_sub_lst",4,"type")="addr"
	Set gtmtypfldindx("zwr_sub_lst","subsc_list[0].first")=4
	Set gtmtypes("zwr_sub_lst",5,"name")="zwr_sub_lst.subsc_list[0].second"
	Set gtmtypes("zwr_sub_lst",5,"off")=24
	Set gtmtypes("zwr_sub_lst",5,"len")=8
	Set gtmtypes("zwr_sub_lst",5,"type")="addr"
	Set gtmtypfldindx("zwr_sub_lst","subsc_list[0].second")=5
	;
	Set gtmtypes("zwr_zav_blk")="struct"
	Set gtmtypes("zwr_zav_blk",0)=4
	Set gtmtypes("zwr_zav_blk","len")=32
	Set gtmtypes("zwr_zav_blk",1,"name")="zwr_zav_blk.zav_base"
	Set gtmtypes("zwr_zav_blk",1,"off")=0
	Set gtmtypes("zwr_zav_blk",1,"len")=8
	Set gtmtypes("zwr_zav_blk",1,"type")="addr"
	Set gtmtypfldindx("zwr_zav_blk","zav_base")=1
	Set gtmtypes("zwr_zav_blk",2,"name")="zwr_zav_blk.zav_free"
	Set gtmtypes("zwr_zav_blk",2,"off")=8
	Set gtmtypes("zwr_zav_blk",2,"len")=8
	Set gtmtypes("zwr_zav_blk",2,"type")="addr"
	Set gtmtypfldindx("zwr_zav_blk","zav_free")=2
	Set gtmtypes("zwr_zav_blk",3,"name")="zwr_zav_blk.zav_top"
	Set gtmtypes("zwr_zav_blk",3,"off")=16
	Set gtmtypes("zwr_zav_blk",3,"len")=8
	Set gtmtypes("zwr_zav_blk",3,"type")="addr"
	Set gtmtypfldindx("zwr_zav_blk","zav_top")=3
	Set gtmtypes("zwr_zav_blk",4,"name")="zwr_zav_blk.next"
	Set gtmtypes("zwr_zav_blk",4,"off")=24
	Set gtmtypes("zwr_zav_blk",4,"len")=8
	Set gtmtypes("zwr_zav_blk",4,"type")="addr"
	Set gtmtypfldindx("zwr_zav_blk","next")=4
	;
	; Structure cross reference (struct topname key to retrieve type
	;
	Set gtmstructs("activelv_dbg_struct")="activelv_dbg_t"
	Set gtmstructs("backup_reg_list_struct")="backup_reg_list"
	Set gtmstructs("block_info_struct")="block_info"
	Set gtmstructs("buddy_list_struct")="buddy_list"
	Set gtmstructs("cache_ent")="cache_entry"
	Set gtmstructs("cache_rec_struct")="cache_rec"
	Set gtmstructs("callin_entry_list")="callin_entry_list"
	Set gtmstructs("clb_stat_struct")="clb_stat"
	Set gtmstructs("cm_lckblklck_struct")="cm_lckblklck"
	Set gtmstructs("cm_lckblkprc_struct")="cm_lckblkprc"
	Set gtmstructs("cm_lckblkreg_struct")="cm_lckblkreg"
	Set gtmstructs("cm_lk_response_struct")="cm_lk_response"
	Set gtmstructs("cm_region_head_struct")="cm_region_head"
	Set gtmstructs("cm_region_list_struct")="cm_region_list"
	Set gtmstructs("cmd_parm_struct")="CLI_PARM"
	Set gtmstructs("cmd_parm_tag")="CLI_ENTRY"
	Set gtmstructs("coll_hdr_struct")="coll_hdr"
	Set gtmstructs("collseq_struct")="collseq"
	Set gtmstructs("compswap_time_field_struct")="compswap_time_field"
	Set gtmstructs("condition_handler_struct")="condition_handler"
	Set gtmstructs("cs_struct")="connection_struct"
	Set gtmstructs("cvs_MD5Context")="cvs_MD5_CTX"
	Set gtmstructs("cw_set_element_struct")="cw_set_element"
	Set gtmstructs("d_socket_struct_type")="d_socket_struct"
	Set gtmstructs("dev_dispatch_struct")="dev_dispatch_struct"
	Set gtmstructs("dev_pairs")="Dev_param_pairs"
	Set gtmstructs("dev_param_pair")="Dev_param_pair"
	Set gtmstructs("dollar_ecode")="dollar_ecode_struct"
	Set gtmstructs("dollar_stack")="dollar_stack_struct"
	Set gtmstructs("enc_handles_struct")="enc_handles"
	Set gtmstructs("err_ctl_struct")="err_ctl"
	Set gtmstructs("err_msg_struct")="err_msg"
	Set gtmstructs("ext_tms_struct")="ext_tms"
	Set gtmstructs("file_control_struct")="file_control"
	Set gtmstructs("fnpc_struct")="fnpc"
	Set gtmstructs("forw_multi_element_struct")="forw_multi_struct"
	Set gtmstructs("freeze_multiproc_state_struct")="freeze_multiproc_state"
	Set gtmstructs("freeze_reg_mp_state_struct")="freeze_reg_mp_state"
	Set gtmstructs("gcall_args_struct")="gcall_args"
	Set gtmstructs("gd_addr_struct")="gd_addr"
	Set gtmstructs("gd_binding_struct")="gd_binding"
	Set gtmstructs("gd_gblname_struct")="gd_gblname"
	Set gtmstructs("gd_id_struct")="unix_file_id"
	Set gtmstructs("gd_inst_info_struct")="gd_inst_info"
	Set gtmstructs("gd_region_struct")="gd_region"
	Set gtmstructs("gd_segment_struct")="gd_segment"
	Set gtmstructs("gdr_name_struct")="gdr_name"
	Set gtmstructs("gld_dbname_list_struct")="gld_dbname_list"
	Set gtmstructs("glist_struct")="glist"
	Set gtmstructs("global_dir_path_struct")="global_dir_path"
	Set gtmstructs("global_list_struct")="global_list"
	Set gtmstructs("global_root_list_struct")="global_root_list"
	Set gtmstructs("global_tlvl_info_struct")="global_tlvl_info"
	Set gtmstructs("gparam_list_struct")="gparam_list"
	Set gtmstructs("gtm_num_range_struct")="gtm_num_range_t"
	Set gtmstructs("gtm_tls_conn_info_struct")="gtm_tls_conn_info"
	Set gtmstructs("gtm_tls_ctx_struct")="gtm_tls_ctx_t"
	Set gtmstructs("gtm_tls_session_struct")="gtm_tls_socket_t"
	Set gtmstructs("gtmsecshr_mesg_struct")="gtmsecshr_mesg"
	Set gtmstructs("gtmsrc_lcl_struct")="gtmsrc_lcl"
	Set gtmstructs("gv_key_struct")="gv_key"
	Set gtmstructs("gv_namehead_struct")="gv_namehead"
	Set gtmstructs("gv_trigger_struct")="gv_trigger_t"
	Set gtmstructs("gvname_info_struct")="gvname_info"
	Set gtmstructs("gvnh_reg_struct")="gvnh_reg_t"
	Set gtmstructs("gvnh_spanreg_struct")="gvnh_spanreg_t"
	Set gtmstructs("gvsavtarg_struct")="gvsavtarg_t"
	Set gtmstructs("gvstats_rec_csd_struct")="gvstats_rec_csd_t"
	Set gtmstructs("gvstats_rec_struct")="gvstats_rec_t"
	Set gtmstructs("gvt_container_struct")="gvt_container"
	Set gtmstructs("gvt_trigger_struct")="gvt_trigger_t"
	Set gtmstructs("gvtr_invoke_parms_struct")="gvtr_invoke_parms_t"
	Set gtmstructs("gvtr_subs_pattern_struct")="gvtr_subs_pattern_t"
	Set gtmstructs("gvtr_subs_point_struct")="gvtr_subs_point_t"
	Set gtmstructs("gvtr_subs_range_struct")="gvtr_subs_range_t"
	Set gtmstructs("gvtr_subs_star_struct")="gvtr_subs_star_t"
	Set gtmstructs("gvzwrite_datablk_struct")="gvzwrite_datablk"
	Set gtmstructs("hash_table_addr_struct")="hash_table_addr"
	Set gtmstructs("hash_table_int4_struct")="hash_table_int4"
	Set gtmstructs("hash_table_int8_struct")="hash_table_int8"
	Set gtmstructs("hash_table_mname_struct")="hash_table_mname"
	Set gtmstructs("hash_table_objcode_struct")="hash_table_objcode"
	Set gtmstructs("hash_table_str_struct")="hash_table_str"
	Set gtmstructs("header_struct_struct")="header_struct"
	Set gtmstructs("i_hdr")="inc_header"
	Set gtmstructs("ihead_struct")="ihdtyp"
	Set gtmstructs("inc_list")="inc_list_struct"
	Set gtmstructs("integ_error_blk_list_struct")="integ_error_blk_list"
	Set gtmstructs("io_desc_struct")="io_desc"
	Set gtmstructs("io_log_name_struct")="io_log_name"
	Set gtmstructs("ipcs_mesg_struct")="ipcs_mesg"
	Set gtmstructs("jnl_ctl_list_struct")="jnl_ctl_list"
	Set gtmstructs("jnl_format_buff_struct")="jnl_format_buffer"
	Set gtmstructs("jnl_private_control_struct")="jnl_private_control"
	Set gtmstructs("jnl_process_vector_struct")="jnl_process_vector"
	Set gtmstructs("jnlext_multi_struct")="jnlext_multi_t"
	Set gtmstructs("jnlpool_addrs_struct")="jnlpool_addrs"
	Set gtmstructs("job_param_str_struct")="job_param_str"
	Set gtmstructs("job_parm_struct")="job_parm"
	Set gtmstructs("joberr_msg_struct")="joberr_msg"
	Set gtmstructs("key_value_struct")="key_cum_value"
	Set gtmstructs("kill_set_struct")="kill_set"
	Set gtmstructs("link_info_struct")="link_info"
	Set gtmstructs("linkage_entry")="linkage_entry"
	Set gtmstructs("long_list_struct")="long_list"
	Set gtmstructs("long_long_list_struct")="long_long_list"
	Set gtmstructs("lvTreeNodeNumStruct")="lvTreeNodeNum"
	Set gtmstructs("lvTreeStruct")="lvTree"
	Set gtmstructs("lv_blk_struct")="lv_blk"
	Set gtmstructs("lv_val_struct")="lv_val"
	Set gtmstructs("lv_xnew_var_struct")="lv_xnew_var"
	Set gtmstructs("lv_xnewref_struct")="lv_xnew_ref"
	Set gtmstructs("lvname_info_struct")="lvname_info"
	Set gtmstructs("lvzwrite_datablk_struct")="lvzwrite_datablk"
	Set gtmstructs("mcalloc_hdr_struct")="mcalloc_hdr"
	Set gtmstructs("mem_list_struct")="mem_list"
	Set gtmstructs("merge_glvn_struct_type")="merge_glvn_struct"
	Set gtmstructs("mlabstruct")="mlabel"
	Set gtmstructs("mlinestruct")="mline"
	Set gtmstructs("mliteralstruct")="mliteral"
	Set gtmstructs("mlk_ctldata_struct")="mlk_ctldata"
	Set gtmstructs("mlk_pvtblk_struct")="mlk_pvtblk"
	Set gtmstructs("mlk_pvtctl_struct")="mlk_pvtctl"
	Set gtmstructs("mlk_shrhash_struct")="mlk_shrhash"
	Set gtmstructs("mlk_stats_struct")="mlk_stats_t"
	Set gtmstructs("mlk_tp_struct")="mlk_tp"
	Set gtmstructs("mmseg_struct")="mmseg"
	Set gtmstructs("mprof_stack_frame_struct")="mprof_stack_frame"
	Set gtmstructs("mprof_tree_struct")="mprof_tree"
	Set gtmstructs("mprof_wrapper_struct")="mprof_wrapper"
	Set gtmstructs("mu_set_reglist")="mu_set_rlist"
	Set gtmstructs("muinc_blk_hdr_struct")="muinc_blk_hdr"
	Set gtmstructs("multi_element_struct")="multi_struct"
	Set gtmstructs("mv_stent_struct")="mv_stent"
	Set gtmstructs("mvarstruct")="mvar"
	Set gtmstructs("mvaxstruct")="mvax"
	Set gtmstructs("node_local_struct")="node_local"
	Set gtmstructs("noisolation_element_struct")="noisolation_element"
	Set gtmstructs("noisolation_list_struct")="noisolation_list"
	Set gtmstructs("omi_cl_stat")="omi_cl_stat"
	Set gtmstructs("omi_cn_stat")="omi_cn_stat"
	Set gtmstructs("omi_conn_ll")="omi_conn_ll"
	Set gtmstructs("omi_err_hdr")="omi_err_hdr"
	Set gtmstructs("omi_hist_rec_t")="omi_hist_rec"
	Set gtmstructs("omi_req_hdr")="omi_req_hdr"
	Set gtmstructs("onln_rlbk_reg_list_struct")="onln_rlbk_reg_list"
	Set gtmstructs("open_relinkctl_struct")="open_relinkctl_sgm"
	Set gtmstructs("open_shlib_struct")="open_shlib"
	Set gtmstructs("oprtypestruct")="oprtype"
	Set gtmstructs("parmblk_struct")="parmblk_struct"
	Set gtmstructs("parse_blk_struct")="parse_blk"
	Set gtmstructs("patstrlit_struct")="pat_strlit"
	Set gtmstructs("pattern_struct")="pattern"
	Set gtmstructs("phase2_wait_trace_struct")="phase2_wait_trace_t"
	Set gtmstructs("pini_list")="pini_list_struct"
	Set gtmstructs("pipe_interrupt_type")="pipe_interrupt"
	Set gtmstructs("plength_struct")="plength"
	Set gtmstructs("probecrit_rec_struct")="probecrit_rec_t"
	Set gtmstructs("pte_csh_struct")="pte_csh"
	Set gtmstructs("ptstr_struct")="ptstr"
	Set gtmstructs("rc_aq_hdr")="rc_aq_hdr"
	Set gtmstructs("rc_dsid_list_struct")="rc_dsid_list"
	Set gtmstructs("rc_hist_rec_t")="rc_hist_rec"
	Set gtmstructs("rc_lknam")="rc_lknam"
	Set gtmstructs("rc_rq_hdr")="rc_rq_hdr"
	Set gtmstructs("rc_sbkey")="rc_sbkey"
	Set gtmstructs("rc_swstr")="rc_swstr"
	Set gtmstructs("rc_xblk_hdr")="rc_xblk_hdr"
	Set gtmstructs("rc_xdsid")="rc_xdsid"
	Set gtmstructs("rc_xnsid")="rc_xnsid"
	Set gtmstructs("redirect_list_struct")="redirect_list"
	Set gtmstructs("redo_root_search_context_struct")="redo_root_search_context"
	Set gtmstructs("reg_ctl_list_struct")="reg_ctl_list"
	Set gtmstructs("relinkctl_data_struct")="relinkctl_data"
	Set gtmstructs("relinkrec_struct")="relinkrec_t"
	Set gtmstructs("relinkshm_hdr")="relinkshm_hdr_t"
	Set gtmstructs("repl_conn_info_struct")="repl_conn_info_t"
	Set gtmstructs("repl_ctl_struct")="repl_ctl_element"
	Set gtmstructs("repl_histinfo_struct")="repl_histinfo"
	Set gtmstructs("repl_inst_hdr_struct")="repl_inst_hdr"
	Set gtmstructs("repl_inst_uuid_struct")="repl_inst_uuid"
	Set gtmstructs("repl_rctl_elem_struct")="repl_rctl_elem_t"
	Set gtmstructs("repl_tls_info_struct")="repl_tls_info_t"
	Set gtmstructs("rhead_struct")="rhdtyp"
	Set gtmstructs("rtnobj_hdr_struct")="rtnobj_hdr_t"
	Set gtmstructs("rtnobjshm_hdr_struct")="rtnobjshm_hdr_t"
	Set gtmstructs("save_xfer_entry_struct")="save_xfer_entry"
	Set gtmstructs("select_list_struct")="select_list"
	Set gtmstructs("semid_queue_elem_t")="semid_queue_elem"
	Set gtmstructs("semwait_status_struct")="semwait_status_t"
	Set gtmstructs("set_jnl_options_struct")="set_jnl_options"
	Set gtmstructs("sgm_info_struct")="sgm_info"
	Set gtmstructs("sgmnt_addrs_struct")="sgmnt_addrs"
	Set gtmstructs("sgmnt_data_struct")="sgmnt_data"
	Set gtmstructs("shm_forw_multi_struct")="shm_forw_multi_t"
	Set gtmstructs("shm_parms_struct")="shm_parms"
	Set gtmstructs("shm_snapshot_struct")="shm_snapshot_t"
	Set gtmstructs("shmpool_blk_hdr_struct")="shmpool_blk_hdr"
	Set gtmstructs("shmpool_buff_hdr_struct")="shmpool_buff_hdr"
	Set gtmstructs("snapshot_context_struct")="snapshot_context_t"
	Set gtmstructs("snapshot_filehdr_struct")="snapshot_filhdr_t"
	Set gtmstructs("snapshot_info_struct")="snapshot_info_t"
	Set gtmstructs("socket_address_type")="socket_address"
	Set gtmstructs("socket_interrupt_type")="socket_interrupt"
	Set gtmstructs("socket_struct_type")="socket_struct"
	Set gtmstructs("span_subs_struct")="span_subs"
	Set gtmstructs("src_line_type")="src_line_struct"
	Set gtmstructs("srch_blk_status_struct")="srch_blk_status"
	Set gtmstructs("st_timer_alloc")="st_timer_alloc"
	Set gtmstructs("stack_frame_struct")="stack_frame"
	Set gtmstructs("statsDB_diqe")="statsDB_deferred_init_que_elem"
	Set gtmstructs("storElemStruct")="storElem"
	Set gtmstructs("struct_spanode_integ")="span_node_integ"
	Set gtmstructs("symval_struct")="symval"
	Set gtmstructs("tag_abs_time")="ABS_TIME"
	Set gtmstructs("tag_ts")="GT_TIMER"
	Set gtmstructs("tbptype")="tbp"
	Set gtmstructs("textElemStruct")="textElem"
	Set gtmstructs("tlevel_info_struct")="tlevel_info"
	Set gtmstructs("token_split_t_struct")="token_split_t"
	Set gtmstructs("tp_frame_struct")="tp_frame"
	Set gtmstructs("tp_region_struct")="tp_region"
	Set gtmstructs("tp_var_struct")="tp_var"
	Set gtmstructs("trctbl_entry_struct")="trctbl_entry"
	Set gtmstructs("treeNodeStruct")="lvTreeNode"
	Set gtmstructs("triplesize")="tripsize"
	Set gtmstructs("tripletype")="triple"
	Set gtmstructs("trunc_reg_struct")="trunc_region"
	Set gtmstructs("ua_list_struct")="ua_list"
	Set gtmstructs("unix_db_info_struct")="unix_db_info"
	Set gtmstructs("unix_file_info_struct")="unix_file_info"
	Set gtmstructs("urx_addr_type")="urx_addr"
	Set gtmstructs("urx_labref_type")="urx_labref"
	Set gtmstructs("urx_rtnref_type")="urx_rtnref"
	Set gtmstructs("utfcgr_entry_struct")="utfcgr_entry"
	Set gtmstructs("utfcgr_struct")="utfcgr"
	Set gtmstructs("util_snapshot_struct")="util_snapshot_t"
	Set gtmstructs("vms_lock_sb_struct")="vms_lock_sb"
	Set gtmstructs("wcs_conflict_trace_struct")="wcs_conflict_trace_t"
	Set gtmstructs("wtstart_cr_list_struct")="wtstart_cr_list_t"
	Set gtmstructs("zro_ent_type")="zro_ent"
	Set gtmstructs("zshow_out_struct")="zshow_out"
	Set gtmstructs("zwr_alias_var_struct")="zwr_alias_var"
	Set gtmstructs("zwr_hash_table_struct")="zwr_hash_table"
	Set gtmstructs("zwr_sub_lst_struct")="zwr_sub_lst"
	Set gtmstructs("zwr_zav_blk_struct")="zwr_zav_blk"
	;
	; Union cross reference (union topname key to retrieve type
	;
	Set gtmunions("gtm_time8_struct")="gtm_time8"
	Set gtmunions("gvtr_subs_struct")="gvtr_subs_t"
	Set gtmunions("omi_li")="omi_li"
	Set gtmunions("omi_si")="omi_si"
	Set gtmunions("omi_vi")="omi_vi"
	Set gtmunions("rc_byte")="rc_byte"
	Set gtmunions("rc_lword")="rc_lword"
	Set gtmunions("rc_q_hdr")="rc_q_hdr"
	Set gtmunions("rc_word")="rc_word"
	;
	Quit
